!***********************************************************************
!
!                     ROUTINE ADPROJ
!
!     This routine calculates the projector on the adiabatic surfaces  
!     and provides files for the potfit routine. Also the adiabatic
!     potential surfaces are calculated.
!
! MRB Jun 05
!
!***********************************************************************

      program adproj

use decimal, only:long,sip,dop
use global
use aglobal
use globmemmod
use maxv, only: maxdim,maxsta,rdmemdim
use dvrdatmod
use griddatmod
use operdef
use datenmod
use compdatmod
use channels
use versions

use op1lib, only: initvxi
use zeigauswmod, only: zeigausw
use memhopsmod, only: memhops
use rdopermod, only: operinfo 

use logdat, only: errormsg
use rddvrmod
use rdopermod, only: rdoper
use iodvrdef, only: wrdvrdef
use vpotpars, only: writevpotpars
use potevalmod, only: calcdiab
use runpfitmod

use module_interface


      implicit none

      integer(long)       :: chkdvr,chkgrd,chkpsi,chkdat,ilbl,chkprp,&
                             potfiles,projfiles,totgrid,m,i,p,zeigpot,&
                             zeigpro,niobuf,proj_i,proj_j,state, &
                             off_i,off_j
      logical(kind=4)     :: lnopot,lproj_ij,lerr,lnoproj,loffdiag
      real(dop)           :: atotgrd
      character(len=c5)   :: filename

      common /zeigerp/ zeigpot,zeigpro,niobuf,totgrid
      common /matelem/ proj_i,proj_j,state,lnopot,lproj_ij,lnoproj
      common /offdiag/ off_i, off_j,loffdiag
      common /filenum/ projfiles

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file.
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.true.
      
!-----------------------------------------------------------------------
! read input options
!-----------------------------------------------------------------------
      call adprojinput
      
!-----------------------------------------------------------------------
! get memory sizes 
!-----------------------------------------------------------------------
      filename=dname
      ilbl=index(filename,' ')-1
      open(idvr,file=filename,form='unformatted',status='old',err=900)
      call rdmemdim(idvr)

      filename=operfile
      ilbl=index(filename,' ')-1
      open(ioper,file=filename,form='unformatted',status='old',err=900)
      call rdmemdim(ioper)

!-----------------------------------------------------------------------
! Allocate mememory 
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_runpfitmod

      chkdvr=1
      call dvrinfo(lerr,chkdvr)
      call chksyserr(filename,lerr)
      if (lerr) go to 1010
      close(idvr)

!-----------------------------------------------------------------------
!  Opening the oper file.
!-----------------------------------------------------------------------
      write(6,'(2a)') ' The operator file is : ',filename(1:ilbl)
      chkdvr=2
      chkgrd=1
      call operinfo(lerr,chkdvr,chkgrd)
      if (lerr) go to 1010
      close(ioper)
      
!-----------------------------------------------------------------------
!  Check input
      
      if(state.gt.nstate) then
         write(6,'(a,i2,a)') &
             'state.gt.nstate, there are only ', nstate, ' states!'
         stop
      endif
      if(proj_i.gt.nstate) then
         write(6,'(a,i2,a)') &
             'proj_i.gt.nstate, there are only ', nstate, ' states!'
         stop
      endif
      if(proj_j.gt.nstate) then
         write(6,'(a,i2,a)') &
             'proj_j.gt.nstate, there are only ', nstate, ' states!'
         stop
      endif
      if(nstate.gt.9) then
         write(6,'(a,i2,a,/a)') 'nstate =',nstate,' must be .le. 9,', &
          'because file names use one-digit numbers for labeling states'
         stop
      endif
      
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.false.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.false.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.false.
      
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! open input and output data files
!
! use standard channels if possible
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=2
      chkgrd=2
      chkpsi=1
      chkdat=1
      chkprp=0
  
!-----------------------------------------------------------------------
! determine total grid dimension. 
!-----------------------------------------------------------------------
      totgrid = 1
      atotgrd = 1.0_dop
      do m = 1, nmode
         atotgrd = atotgrd*subdim(m)
         if( atotgrd .gt. 2147483647.0_dop ) then
            write(6,'(a/,a/,a)')  'The primitive grid is too large!',&
            'Integer overflow. More than 2,147,483,647 grid points.',&
            'Use a coarser grid.'
            do i = m+1, nmode
               atotgrd = atotgrd*subdim(i)
            enddo
            write(6,'(a,f18.0,es14.4)') ' Total griddim =',atotgrd,atotgrd

            stop
         endif
         totgrid = totgrid*subdim(m)
      enddo
!      write(6,'(a,i11,e14.4)')    ' Total griddim = ',totgrid,atotgrd
      write(ilog,'(a,i11,e14.4)') ' Total griddim = ',totgrid,atotgrd
      call myflush(ilog)
     
! --- calculate size of buffer, niobuf
      if(totgrid .le. 8*1024*1024) then
         niobuf = totgrid
      else
         p = totgrid/(1024*1024) + 1
         niobuf = 2*(totgrid/p)
      endif
      
!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw
      
      zeigpot = mrdim
      if(.not.lnopot) then
         potfiles = 1
         if(state.eq.0) then
            potfiles = nstate
         endif
         mrdim = mrdim + niobuf*potfiles
      endif
      
      projfiles = 1
      if(state.eq.0) then
         projfiles = nstate
      endif
      if(.not.lproj_ij) then
         projfiles = projfiles*(nstate*(1+nstate))/2
      endif
      if(lnoproj) then
         projfiles=0
      endif
      if(loffdiag) then
         projfiles=nstate**2
      endif
      
      zeigpro = mrdim
      mrdim = mrdim + niobuf*projfiles
         
!-----------------------------------------------------------------------
! assign work arrays
!-----------------------------------------------------------------------
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
! assign memory for operator
!-----------------------------------------------------------------------
      call memhops

!-----------------------------------------------------------------------
! call routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))

      call aproject(mr(mrhops),mr(zeigpot),mr(zeigpro),niobuf,totgrid,&
           mr(mrort))

deallocate(mc,mr,mi,ml,ms,mf)

      stop
  900 continue
      routine='adpop'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
 1010 write(6,*)' Error reading: '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop

      contains

!#######################################################################

      subroutine aproject(hops,potsrf,adproj,niobuf,totgrid,ort)


      implicit none

      integer(long)       :: proj_i,proj_j,state,projfiles
      integer(long)       :: chkdvr,chkgrd,s,nham,m,i,n,ierr,k,t1,t2,j3,&
                             f,ilbl,endloop,outlaenge,k1,k2,k3,zaehl1,zaehl2,&
                             zaehl3,zaehl4,zaehl5,wfnum,state_low,state_up,&
                             proj_i_low,proj_i_up,proj_j_low,proj_j_up,&
                             off_i,off_j

      integer(long), intent(in)                           :: totgrid,niobuf
      integer(long), dimension(1)                         :: idum
      integer(long), dimension(maxdim)                    :: fpoint,chgdim,chbasis
      integer(long), dimension(maxsta)                    :: outvpot
      integer(long), dimension((maxsta**2)*(1+maxsta)/2)  :: outproj
      real(dop)                                           :: tcpu,tsys,ttot,rotmat
      real(dop), dimension(1)                             :: rdum
      real(dop), dimension(maxdim)                        :: gpoint
      real(dop), dimension(maxsta)                        :: adiab
      real(dop), dimension(ortdim), intent(out)           :: ort
      real(dop), dimension(hopsdim), intent(out)          :: hops
      real(dop), dimension(niobuf*nstate), intent(out)    :: potsrf
      real(dop), dimension(niobuf*nstate*(nstate*(1+nstate))/2), &
                                             intent(out)  :: adproj
      real(dop), dimension(2,maxdim)                      :: chxend
      real(dop), dimension(maxsta,maxsta)                 :: diab,projector,&
                                                             projectmat
      complex(dop), dimension(1)                          :: zdum
      complex(dop), dimension(maxsta,maxsta)              :: rotmatz
      character(len=80),save                              :: buffer
      character(len=2),save                               :: zahlch
      character(len=c5)                                   :: outfile
      character(len=c5)                                   :: filename
      character(len=c1), dimension(maxdim)                :: chmodelabel
      logical(kind=4), dimension(maxdim)                  :: chldvr
      logical(kind=4)                                     :: lnopot,lproj_ij,&
                                                             lnoproj,loffdiag
      real(dop), external                                 :: fcputime,fsystime

      common /matelem/ proj_i,proj_j,state,lnopot,lproj_ij,lnoproj
      common /offdiag/ off_i, off_j,loffdiag
      common /filenum/ projfiles
      

      nham = 1
      wfnum = 0

!-----------------------------------------------------------------------
! read DVR file
!-----------------------------------------------------------------------
      chkdvr=1
      dvrdata(1)=.true.
      call rddvr(ort,rdum,rdum,rdum,zdum,zdum,idum,zdum,zdum,idum,&
                 idum,rdum,chkdvr)

!-----------------------------------------------------------------------
! read operator file
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      call rdoper(hops,chkdvr,chkgrd)
      
!-----------------------------------------------------------------------
!  Determine lower and upper loop limits
!-----------------------------------------------------------------------
      state_low = 1
      state_up  = nstate
      
      if (state.gt.0) then
         state_low = state
         state_up  = state
      endif

      proj_i_low = 1
      proj_j_low = 1
      proj_i_up  = nstate
      proj_j_up  = nstate
      
      if (loffdiag) then
         state_low = 1
         state_up  = 1
      else if (lnoproj) then
         proj_i_low = 1
         proj_i_up  = 0
         proj_j_low = 1
         proj_j_up  = 0
      else if (lproj_ij) then
         proj_i_low = proj_i
         proj_i_up  = proj_i
         proj_j_low = proj_j
         proj_j_up  = proj_j
      endif
      
!-----------------------------------------------------------------------
!  Initialisation of the file arrays
!-----------------------------------------------------------------------   
      if(.not.lnopot) then
         do k=1,state_up-state_low+1
            outvpot(k)=50+k
         enddo
      endif
      
      do k=1,projfiles
         outproj(k)=50+nstate+k
      enddo

!-----------------------------------------------------------------------
!  Removing electronic degree of freedom
!-----------------------------------------------------------------------
      if(feb.gt.0) then
         do f=1,ndof
            chxend(1,f)=xend(1,f)
            chxend(2,f)=xend(2,f)
            chldvr(f)=ldvr(f)
            chbasis(f)=basis(f)
            chgdim(f)=gdim(f)
            chmodelabel(f)=modelabel(f)
         enddo
 
         ndof=ndof-1
                
         do f=feb,ndof
            do i=1,mbaspar
               ipbaspar(i,f)=ipbaspar(i,f+1)
               rpbaspar(i,f)=rpbaspar(i,f+1)
            enddo
            xend(1,f)=xend(1,f+1)
            xend(2,f)=xend(2,f+1)
            ldvr(f)=ldvr(f+1)
            basis(f)=basis(f+1)
            gdim(f)=gdim(f+1)
            modelabel(f)=modelabel(f+1)
         enddo
      endif
      
!-----------------------------------------------------------------------
!  set hoplab
!-----------------------------------------------------------------------
      hoplab(1)='none'
      
!-----------------------------------------------------------------------
!  Opening the output files
!-----------------------------------------------------------------------
      if(loffdiag) then  ! -od option is set
         k3=0
         do k1=1,nstate
            do k2=1,nstate
               k3=k3+1 
               write(outfile,'(a,2i1,a,2i1)') &
                    'apr_od',off_i,off_j,'_',k1,k2
               filename=trim(name)//'/'//trim(outfile)
               if(lovwr) then
                  open(outproj(k3),file=filename,form='unformatted', &
                       status='unknown',err=1030)
               else 
                  open(outproj(k3),file=filename,form='unformatted', &
                       status='new',err=1020)
               endif
               write(outproj(k3),iostat=ierr) progver
                  if (ierr.ne.0) then
                     routine='aproject, adproj.F line 392'
                     message=  'cannot write program version to '//outfile
                     call errormsg
                  endif
               call wrdvrdef(outproj(k3))
               call writevpotpars(outproj(k3),outfile,0)
            enddo
         enddo
         goto 150
      endif

      k3=0
      do k=state_low,state_up
         do k1=proj_i_low,proj_i_up
            if(.not.lproj_ij) then
               proj_j_low = k1
            endif
            do k2=proj_j_low,proj_j_up
               k3=k3+1
               endloop=index(outfile,' ')
               do s=1,endloop
                  outfile(s:s)=' '
               enddo
               if (k.ge.10) then
                  write(outfile,'(a,i2)') 'apr_p',k
               else
                  write(outfile,'(a,i1)') 'apr_p',k
               endif
               outfile(index(outfile,' '):index(outfile,' '))='_'
               zahlch='  '
               if (k1.ge.10) then
                  write(zahlch,'(i2)') k1
                  outfile(index(outfile,' '):index(outfile,' ')+1)=&
                          zahlch(1:2)
               else
                  write(zahlch,'(i1)') k1
                  outfile(index(outfile,' '):index(outfile,' '))=&
                          zahlch(1:1)
               endif
               zahlch='  '
               if (k2.ge.10) then
                  write(zahlch,'(i2)') k2
                  outfile(index(outfile,' '):index(outfile,' ')+1)=&
                          zahlch(1:2)
               else
                  write(zahlch,'(i1)') k2
                  outfile(index(outfile,' '):index(outfile,' '))=&
                          zahlch(1:1)
               endif
               outlaenge=index(outfile,' ')-1
               filename=name(1:laenge)//'/'//outfile(1:outlaenge)
               if(lovwr) then
                  open(outproj(k3),file=filename,form='unformatted', &
                       status='unknown',err=1030)
               else
                  open(outproj(k3),file=filename,form='unformatted', &
                       status='new',err=1020)
               endif

               write(outproj(k3),iostat=ierr) progver
               if (ierr.ne.0) then
                  routine='aproject, adproj.F error 1'
                  message='cannot write program version to '//outfile
                  call errormsg
               endif

               call wrdvrdef(outproj(k3))
               call writevpotpars(outproj(k3),outfile,0)
            enddo
         enddo
      enddo

      if(.not.lnopot) then
         do k=1,state_up-state_low+1
            endloop=index(outfile,' ')
            do s=1,endloop
               outfile(s:s)=' '
            enddo
            if (state_low+k-1.ge.10) then
               write(outfile,'(a,i2)') 'apr_v',state_low+k-1
            else
               write(outfile,'(a,i1)') 'apr_v',state_low+k-1
            endif
            outlaenge=index(outfile,' ')-1
            filename=name(1:laenge)//'/'//outfile(1:outlaenge)
            if(lovwr) then
               open(outvpot(k),file=filename,form='unformatted', &
                    status='unknown',err=1030)
            else
               open(outvpot(k),file=filename,form='unformatted', &
                    status='new',err=1020)
            endif

            write(outvpot(k),iostat=ierr) progver
            if (ierr.ne.0) then
               routine='aproject, adproj.F error 2'
               message='cannot write program version to '//outfile
               call errormsg
            endif

            call wrdvrdef(outvpot(k))
            call writevpotpars(outvpot(k),outfile,0)
         enddo
      endif

  150 continue ! Jump here if loffdiag

!-----------------------------------------------------------------------
!  Inserting electronic degree of freedom
!-----------------------------------------------------------------------
      if(feb.gt.0) then
         ndof=ndof+1
         do f=1,ndof
            xend(1,f)=chxend(1,f)
            xend(2,f)=chxend(2,f)
            ldvr(f)=chldvr(f)
            basis(f)=chbasis(f)
            gdim(f)=chgdim(f)
            modelabel(f)=chmodelabel(f)
         enddo
      endif
!-----------------------------------------------------------------------
! Check for multi-set. 
!-----------------------------------------------------------------------
      if(nstate.eq.1 .or. feb.eq.0) then
         message='adproj works only for multi-set wavefunctions'
         routine='aproject (adproj.F)'
         call errormsg
      endif 

!-----------------------------------------------------------------------
! Write grid dimension to files
!-----------------------------------------------------------------------
      if(.not.lnopot) then
         do zaehl1=1,state_up-state_low+1
            write(outvpot(zaehl1),iostat=ierr) totgrid,niobuf
            if (ierr.ne.0) then
               routine='aproject'
               message='error on writing rest of data to vpotfile'
               call errormsg
               stop
            endif
         enddo
      endif
      
      zaehl4=0
      do zaehl1=state_low,state_up
         do zaehl2=proj_i_low,proj_i_up
            if(.not.lproj_ij .and. .not.loffdiag) then
               proj_j_low = zaehl2
            endif
            do zaehl3=proj_j_low,proj_j_up
               zaehl4=zaehl4+1
               write(outproj(zaehl4),iostat=ierr) totgrid,niobuf
               if (ierr.ne.0) then
                  routine='aproject'
                  message='error on writing totgrid to projfile'
                  call errormsg
                  stop
               endif
            enddo
         enddo
      enddo
      
!-----------------------------------------------------------------------
! Loop over all gridpoints
!-----------------------------------------------------------------------
      j3=0
      call initvxi(fpoint,ndof)

    1 j3=j3+1
!-----------------------------------------------------------------------
! Calculate transformation matrix and adiabatic potential
!-----------------------------------------------------------------------
         do m=1,nmode
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               gpoint(f)=ort(zort(f)-1+fpoint(f))
            enddo
         enddo
         nham=1
         call calcdiab(hops,adiab,diab,rotmatz,fpoint,gpoint,nham)

!-----------------------------------------------------------------------
! copy potential values to the "potential surfaces"
!----------------------------------------------------------------------- 
         if(.not.lnopot) then
            do s=1,state_up-state_low+1
               potsrf(j3+niobuf*(s-1))=adiab(s)
            enddo
         endif
 
!-----------------------------------------------------------------------
! Calculate projection matrices
!-----------------------------------------------------------------------

         if(lnoproj) goto 123
         do s=state_low,state_up
            
!-----------------------------------------------------------------------
! Initialize projection matrix
!----------------------------------------------------------------------- 
            do zaehl1=1,nstate
               do zaehl2=1,nstate
                  projectmat(zaehl1,zaehl2)=0
                  projector(zaehl1,zaehl2)=0
               enddo
            enddo
            if(loffdiag) then
               projector(off_i,off_j)=1.d0
            else
               projector(s,s)=1.d0
            endif
    
!-----------------------------------------------------------------------
! Calculate adiabatic projection matrix
!-----------------------------------------------------------------------
            do zaehl1=1,nstate
               do zaehl2=1,nstate
                  do zaehl3=1,nstate
                     rotmat=dble(rotmatz(zaehl2,zaehl3))
                     projectmat(zaehl1,zaehl2)=projectmat(zaehl1,zaehl2)&
                         +projector(zaehl1,zaehl3)*rotmat
                  enddo
               enddo
            enddo
    
            projector(s,s)=0
            if(loffdiag) projector(off_i,off_j)=0.d0
    
            do zaehl1=1,nstate
               do zaehl2=1,nstate
                  do zaehl3=1,nstate
                     rotmat=dble(rotmatz(zaehl1,zaehl3))
                     projector(zaehl1,zaehl2)=projector(zaehl1,zaehl2)+&
                         rotmat*projectmat(zaehl3,zaehl2)
                  enddo
               enddo
            enddo
    
!-----------------------------------------------------------------------
! copy projection matrices to "projection surfaces"
!-----------------------------------------------------------------------
            zaehl3=0
            do zaehl1=proj_i_low,proj_i_up
               if(.not.lproj_ij .and. .not.loffdiag) then
                  proj_j_low = zaehl1
               endif
               do zaehl2=proj_j_low,proj_j_up
                  zaehl4=zaehl3+((s-state_low)*projfiles)&
                                /(state_up-state_low+1)
                  adproj(j3+niobuf*zaehl4)=projector(zaehl1,zaehl2)
                  zaehl3=zaehl3+1
               enddo
            enddo
 
         enddo
  123    continue  ! if(lnoproj)
 
!-----------------------------------------------------------------------
! Write results if vectors are full
!-----------------------------------------------------------------------
! Potentials
!----------- 
         if (j3.eq.niobuf) then
           
            if(.not.lnopot) then
               do zaehl1=1,state_up-state_low+1
                  write(outvpot(zaehl1),iostat=ierr) &
                      (potsrf(zaehl2+niobuf*(zaehl1-1)),zaehl2=1,niobuf)
                  if (ierr.ne.0) then
                     routine='aproject'
                     message='error on writing data to vpotfile'
                     call errormsg
                     stop
                  endif
               enddo
            endif

!-----------
! Projectors
!-----------    
            zaehl4=0
            do zaehl1=state_low,state_up
               do zaehl2=proj_i_low,proj_i_up
                  if(.not.lproj_ij .and. .not.loffdiag) then
                     proj_j_low = zaehl2
                  endif
                  do zaehl3=proj_j_low,proj_j_up
                     zaehl4=zaehl4+1
                     write(outproj(zaehl4),iostat=ierr) &
                     (adproj(zaehl5+(zaehl4-1)*niobuf),zaehl5=1,niobuf)
                     if (ierr.ne.0) then
                        routine='aproject'
                        message='error on writing data to projfile'
                        call errormsg
                        stop
                     endif
                  enddo
               enddo
            enddo
            j3=0
         endif

!-----------------------------------------------------------------------
! Increment fpoint
!-----------------------------------------------------------------------  
         do f=1,ndof
            if (f.ne.feb) then
               if (fpoint(f).lt.gdim(f)) then
                  fpoint(f)=fpoint(f)+1
                  goto 1
               else
                  fpoint(f)=1
               endif
            endif
         enddo
 
!----------------------------------------------------------------------- 
! Write the rest
!-----------------------------------------------------------------------
! Potentials
!-----------
         if (j3.ne.0) then
 
            if(.not.lnopot) then
               do zaehl1=1,state_up-state_low+1
                  write(outvpot(zaehl1),iostat=ierr) &
                       (potsrf(zaehl2+niobuf*(zaehl1-1)),zaehl2=1,j3)
                  if (ierr.ne.0) then
                     routine='aproject'
                     message='error on writing rest of data to vpotfile'
                     call errormsg
                     stop
                  endif
               enddo
            endif
      
!-----------
! Projectors
!-----------    
            zaehl4=0
            do zaehl1=state_low,state_up
               do zaehl2=proj_i_low,proj_i_up
                  if(.not.lproj_ij) then
                     proj_j_low = zaehl2
                  endif
                  do zaehl3=proj_j_low,proj_j_up
                     zaehl4=zaehl4+1
                     write(outproj(zaehl4),iostat=ierr) &
                     (adproj(zaehl5+(zaehl4-1)*niobuf),zaehl5=1,j3)
                     if (ierr.ne.0) then
                        routine='aproject'
                        message='error on writing totgrid to projfile'
                        call errormsg
                        stop
                     endif
                  enddo
               enddo
            enddo
            j3=0
         endif

!-----------------------------------------------------------------------
! Get overall timing 
!-----------------------------------------------------------------------
      tcpu = fcputime(0)
      tsys = fsystime(0)
      ttot = tcpu+tsys
      t1   = int(ttot/3600)
      ttot = ttot - 3600*t1
      t2   = int(ttot/60)
      ttot = ttot - 60*t2

      write(ilog,'(//,a,/)') ' ***** Overall Timing *****'
      write(ilog,'(a,1x,f12.2)') ' Total cpu time     [s] : ',tcpu
      write(ilog,'(a,1x,f12.2)') ' Total system time  [s] : ',tsys
      write(ilog, 900) ' Total time     [h:m:s] : ',t1,t2,ttot
      write(ilog,*)

! --- WRITE DATE AND MACHINE  ---
      call fpstartup(buffer)
      write(ilog,*)
      write(ilog,'(a2,a78)') ' ',buffer
      

  900 format( a,i4,' :',i3,' :',f5.1)
      stop
      
 1020 write(6,*)' Cannot open the output file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      write(6,*) 'Use option -w to overwrite the output file.'
      stop
      
 1030 write(6,*)' Cannot open the output file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop

      end subroutine aproject
      
!-----------------------------------------------------------------------
!***********************************************************************
!-----------------------------------------------------------------------

      subroutine adprojinput

      implicit none

      integer(long)           :: proj_i,proj_j,state,off_i,off_j
      integer(long), external :: myiargc
      integer(long)           :: i,iarg,ilbl,chkdvr,chkgrd,ierr,&
                                 swap
      logical(kind=4)         :: lerr
      logical(kind=4)         :: lnopot,lproj_ij,lnoproj,loffdiag
      character(len=80)       :: buf, host
      character(len=c5)       :: filename,dvrfile

      common /matelem/ proj_i,proj_j,state,lnopot,lproj_ij,lnoproj
      common /offdiag/ off_i, off_j,loffdiag
      
      lnopot   = .false.
      lnoproj  = .false.
      lproj_ij = .false.
      loffdiag = .false.
      proj_i   = 0
      proj_j   = 0
      off_i    = 0
      off_j    = 0
      state    = 0
            
      operfile = ' '
      dvrfile = ' '
!      fileout  = ' '
      name     = './'
      iarg     = myiargc()
      
! --- WRITE DATE AND MACHINE  ---
!      write(6,'(78a1)')  ('-',i=1,78)
      call fpstartup(host)

      i=1
      if (iarg .le. 0) goto 200
      call mygetarg(0,options)
      optlaenge  = index(options,' ')
      do i = 1, iarg
         call mygetarg(i,buf)
         ilbl  = index(buf,' ')
         ilbl = min(ilbl,100-optlaenge)
         if(ilbl.gt.0) options = options(1:optlaenge)//buf(1:ilbl)
         optlaenge = optlaenge + ilbl
      end do

      i = 0
 100  i=i+1
      if (iarg .lt. i) goto 200
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf .eq. '-i' ) then
            i=i+1
            call mygetarg(i,buf)
            name = buf
         else if( buf .eq. '-p') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         elseif(buf .eq. '-w') then
            lovwr = .true.
         elseif(buf .eq. '-s') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) state
            if (ierr .ne. 0 .or. state.le.0) then
               write(6,'(2a,/a)') 'Expecting a positive integer ',&
                    'after -s, not :', buf
               stop
            endif
         elseif(buf .eq. '-nopot') then
            lnopot = .true.
         elseif(buf .eq. '-noproj') then
            lnoproj = .true.

         elseif(buf .eq. '-me') then
            lproj_ij=.true.
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ilbl) proj_i
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) proj_j
            ierr=ierr+ilbl
            if (ierr .ne. 0 .or. proj_i.le.0 .or. proj_j.le.0) then
               write(6,'(2a,/a)') 'Expecting two positive integers ',&
                    'after -me, not :', buf
               stop
            endif
            if (proj_i .gt. proj_j) then
               swap = proj_i
               proj_i = proj_j
               proj_j = swap
            endif

         elseif(buf .eq. '-od') then
            loffdiag=.true.
            i = i+1   
            call mygetarg(i,buf)
            read(buf,*,iostat=ilbl) off_i
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) off_j
            ierr=ierr+ilbl
            if (ierr .ne. 0 .or. off_i.le.0 .or. off_j.le.0) then
               write(6,'(2a,/a)') 'Expecting two positive integers ', &
                    'after -me, not :', buf
               write(6,'(a,i4,a,i4)') &
                    'off_i =', off_i,',  off-j =', off_j
               stop
            endif
            if (off_i.gt.9 .or. off_j.gt.9) then
               write(6,'(2a,/2i4)') 'Integers after -od ', &
                    'must be one-digit ones, not :', off_i, off_j
               stop
            endif
            if (off_i .eq. off_j) then
               write(6,'(2a,/2i4)') 'Integers after -od ', &
                    'must differ, not :', off_i, off_j
               stop
            endif

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(1x,a,/a)')&
       'Purpose: Calculates the adiabatic potential and the adiabatic',&
       '          projector.'
            write(6,'(1x,2a)')&
       'Usage:   adproj [Options]'
            write(6,'(/,a)') 'Options  : '
            write(6,'(a)') ' -i DIR  : data is stored in directory DIR'
            write(6,'(2a,/,9x,a)') &
                     ' -p FILE : the PES is read '&
                    ,'from FILE rather than from ./oper',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a)') &
                     ' -w      : Overwrite the output files.'
            write(6,'(2a)')&
                     ' -nopot  : Adiabatic PES are not calculated. '
            write(6,'(2a)')&
                     ' -noproj : Projectors will not be calculated.'
            write(6,'(3a)')&
                     ' -s s    : Only adiabatic PES and Projector',&
                     ' Matrix Elements for state s will be calculated.'
            write(6,'(3a)')&
                     ' -me i j : Only Projector Matrix Elements P_i_j',&
                     ' will be calcualted.'
            write(6,'(a,/10x,2a,/10x,a)') &
                 ' -od i j : Only the Projector Matrix Elements of the', &
                     ' electronic density matrix element ij', &
                     ' will be calculated.', &
                     ' Here i and j are unequal one-digit integers.'
            write(6,'(a)')&
                     ' -ver    : Version information about the program '
            write(6,'(a,/a)') &
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      endif
      
  200 continue
      nopt=i-1
      
!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      laenge=index(name,' ')-1
      call abspath(name,laenge)
      filename=name(1:laenge)//'/apr.log'
      open (unit=ilog,file=filename,form='formatted')

      if(operfile .eq. ' ') operfile='oper'
      opflaenge=index(operfile,' ')-1
      call abspath(operfile,opflaenge)

      if(dvrfile .eq. ' ') dvrfile='dvr'
      dlaenge=index(dvrfile,' ')-1
      call abspath(dvrfile,dlaenge)
      dname = dvrfile(1:dlaenge)

!-----Re-set defaults in case of loffdiag
      if(loffdiag) then
         lnopot  = .true.
         lproj_ij= .false.
         lnoproj = .false.
         state   = 0
      endif

!-----------------------------------------------------------------------
!  Write host to log-file, write options to log file.
!-----------------------------------------------------------------------      
      write(ilog,'(a)') 'Current adproj calculation done on:'
      write(ilog,'(/a2,a78)') ' ',host
      if( optlaenge .le. 60 ) then
         write(ilog,'(2a/)')   ' Options : ', options(1:optlaenge)
      else
         write(ilog,'(a,/a/)') ' Options : ', options(1:optlaenge)
      end if

      return 
      
 1000 write(6,*)' Cannot open file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
 1010 write(6,*)' Error reading: '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
       
      end subroutine adprojinput
      
      end program adproj

