!***********************************************************************
!
!                         ANALYSE program CROSSCORR
!
!     This program calculates the overlap between the wave functions
!     of two different calculations, where the first wavefunction is kept 
!     fixed and the second wavefunction is actually a set of wavefunctions 
!     for different times. The cross-correlation function <psi1|psi2(t)>
!     is written to the ascii-file cross.
!
!     The wave functions may differ in
!     the number of spfs but not in the number of grid points of a
!     certain mode. Some additional checks are made in the subroutine
!     'overcheck' to ensure compatibility.
!     The variables concerning the first wave function have the ending
!     'c1', variables concerning the second wave function have the 
!     ending 'c2' or no special ending at all. This is intended to
!     reduce the number of additional variables.
!     The crosscorrelation table is saved as 'cross_X' in the directory 
!     of the second calculation. The X obove stands for the name of the
!     first calculation.
!
!     03/01 HDM
!
!
!***********************************************************************
      
      module crosscorrmod

      use decimal, only : long,dop
      use maxv, only : maxdim, maxsta

!-----------------------------------------------------------------------
! options for overlap calculation
!-----------------------------------------------------------------------
      integer(long), save :: ovtype,ebtype,init,nst1,nst2,sbra,sket,&
                             ovmaxblock,idxbra,idxket

!-----------------------------------------------------------------------
!     pointers for overlap routine
!-----------------------------------------------------------------------
      integer(long), save :: mcdover,mcovr,mcovr1,doverdim,dgldim1,dgldim0, &
                             adim1,dmatdim1,agmatdim1,gsmatdim1,gwpinfodim1, &
                             gdoverdim

      integer(long), dimension(:,:), allocatable, save  :: dovmat,gdovmat
      integer(long), dimension(:,:,:), allocatable,save :: ddovmat
      complex(dop), dimension(:), allocatable, save     :: agmat1

!-----------------------------------------------------------------------
!     real variables for overlap routine
!-----------------------------------------------------------------------
      real(dop), save :: tend

!-----------------------------------------------------------------------
!     logical for overlap routine
!-----------------------------------------------------------------------
      logical(kind=4), save :: lcrsspf,lcrsdiag,lrst,lrst1,lallov,lback, &
                               lcc,lsingle,lshort,lmatrixel,lphasereal,&
                               ldt,lhc

!-----------------------------------------------------------------------
!     logical to read operator
!-----------------------------------------------------------------------
      logical(kind=4), save :: loper

      end module crosscorrmod

!#######################################################################

      program crosscorr

use decimal, only:long,dop,sip
use global
use aglobal
use globmemmod
use maxv, only:maxdim,maxsta
use dvrdatmod
use griddatmod
use psidef
use datenmod
use hpsimod
use operdef
use compdatmod
use compdat1mod
use channels

use op1lib, only: zerovxz, normvxz, cpvxz, zerovxl
use mmlib, only: mmaxzz
use mtlib, only: mtxxzz, mttxzz
use xvlib, only: vvaxzz
use logdat, only: errormsg
use compdat, only: savecsys,savecpsi,getcsys,getcpsi
use iorst, only: rstinfo
use iofile, only: rdpsiinfo
use rddvrmod
use rdopermod
use auswutil, only: savecdat,getcdat
use zeigauswmod, only: zeigausw
use memhopsmod, only: memhops
use zeighpsimod, only: operflag,zeighpsi
use crosscorrmod
use logdat

use module_interface
use versions

use rdfilesmod, only: rdpsi,rdpsigrid
use opwfmod, only: opwfex
use iorst, only: rdrstpsi
use mlanalysis, only: croco_tree
use wftree
use inputmlmod
use multilayer
use mlanalysis, only: overlap_2

use calchamod, only: calcha
use tdkoeffs, only: calctdk
use hphimod, only: hphi
use phihphimod, only: phihphi
 
      implicit none

      integer(long), external              :: slen
      integer(long)                        :: zeig,zeig1,zeig2,m,s,t,&
                                              ilbl,n,chkdvr,chkgrd, &
                                              chkpsi,chkdat,chkprp,i,j
      integer(long)                        :: lentp,lentp2
      integer(long)                        :: nham
      character(len=c5)                    :: filename
      logical(kind=4)                      :: readsys,lerr,linwf,&
                                              ltape,ltree

      integer(long)                              :: aerr
      character(len=20)                          :: cerr
      logical(kind=4), external                  :: int2log

      integer(long), external   :: integersize, realsize, logicalsize

      complex(dop), dimension(:), allocatable  :: workc

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault
      call crscinput

! Now IPSI1 is the file containing the time-dependent wave-function,
! while IPSI is the file containing the time-independent reference
! wave-function.

!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      filename='crscr.log'
      if(laeout.eq.0) filename='/dev/null'
      open (unit=ilog,file=filename,form='formatted')
      logisopen=.true.

!-----------------------------------------------------------------------
! get memory sizes 
!-----------------------------------------------------------------------
      call rdmemdim(ipsi)
      if (loper) then
         filename=operfile
         ilbl=index(filename,' ')-1
         open(ioper,file=filename,form='unformatted',status='old',err=900)
         call rdmemdim(ioper)
         close(ioper)
      endif

!-----------------------------------------------------------------------
! Allocate mememory 
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_psidef
      call alloc_compdat

!-----------------------------------------------------------------------
! set no. of comparison data sets
!-----------------------------------------------------------------------
      ncomp=2

!-----------------------------------------------------------------------
! read data from psi1 file (time-dependent data set, ket).
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      chkprp=0
      if ( lrst1 ) then
         irst=ipsi1
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
         if (citype .gt. 0) then
            routine='crosscorr'
            message='citype .gt.0 not implemented for restart files.'
            call errormsg        
         endif
         filename='restart (-R, time dep. WF)'
         if (lerr) goto 999
         if(normthermal .gt. 0.d0) then
            write(6,'(a,1pe16.9)') &
                'time-dep. data set, ket: normthermal =',normthermal
         endif
      else
         read(ipsi1) filever(ipsi1)
         call rdpsiinfo(ipsi1,chkdvr,chkgrd,chkpsi,chkdat)
         filename='psi1'
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif

!--compare bratree and kettree
   routine='crosscorr'
   if(lrunml) then
      call compare_tapes(tape,ctape(:,2),ltape,ltree,lentp,lentp2)
      if(ltree) then
         do j=1,int(max(lentp,lentp2)),20
            write(ilog,'(/a)') '    j/cross-tape(j)/system-tape(j)'
            write(ilog,'(20i4)') (      i   ,i=j,min(lentp ,19+j))
            write(ilog,'(20i4)') ( tape(i  ),i=j,min(lentp ,19+j))
            write(ilog,'(20i4)') (ctape(i,1),i=j,min(lentp2,19+j))
         enddo
         message='The topologies of the trees cross/system differ.'
         call errormsg
      elseif(ltape) then
         write(ilog,'(2a,2i8)') 'The tapes cross/system are not ',&
           'identical. dgldim:',dgldim,cdgldim(2)
         write(ilog,'(2a,i5)') 'But the topologies of the trees ', &
           'cross/system are identical. lentape =',lentp
      else
         write(ilog,'(a,i5)') 'The tapes cross/system are identical. lentape =',lentp
      endif
   endif

!-----------------------------------------------------------------------
! Use of operator requires identical structures of bra and ket.
!-----------------------------------------------------------------------
   if(loper) then
      if((psitype.ne.cpsitype(2)) .or. (psitype.ne.0 .and. psitype.ne.3)) then
         write(6,'(a,2i5)') ' psitype:', psitype, cpsitype(2)
         message="Both WFs, bra and ket, must be of either MCTDH-type    "// &
         "                  or ML-MCTDH-type if -O option (operator) is used."
         routine='crosscorr'
         call errormsg
      endif

      if (lrunml) then
         if (ltape) then
            message="Tree structures and SPF numbers must be identical if -O option (operator) is used."
            call errormsg
         endif
      else
         filename='reference WF, bra'
         call chksyserr(filename,lerr)
         do s = 1, nstate
            do m = 1, nmode
               if( cdim(m,s,1) .ne. cdim(m,s,2) ) then
                  lerr=.true.
                  write(6,'(a,2i4,a,2i6)') ' s, m', s,m, &
                    ',  dim (bra/ket)',cdim(m,s,1),cdim(m,s,2)
               endif
            enddo
         enddo
         if(lerr.or.readsys) then
            routine='crosscorr'
            message="dim's and states of bra and ket must be "// &
                    "identical if -O option (operator) is used."
            call errormsg
         endif
      endif
   endif

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
      if (loper) then
         chkdvr=2
         chkgrd=2
         call dvrinfo(lerr,chkdvr)
         call operinfo(lerr,chkdvr,chkgrd)
         if (lerr) goto 888

 20      nham=0
         do n=1,hamdim
            if (hamlab(n) .eq. operator) nham=n
         enddo

         if (nham .eq. 0) then
            write(6,'(4a)')'Operator "',&
                 operator(1:slen(operator)),'" not found in ',&
                 operfile(1:slen(operfile))
            write(6,'(1x,a)') 'Name of operator ? Choose from:'
            do n=1,hamdim
               write(6,'(2x,a)') hamlab(n)
            enddo
            read(5,'(a)') operator
            if (operator .eq. ' ') stop
            goto 20
         endif
         write(6,'(3a,i3)') ' Using operator : ',&
               operator(1:slen(operator)),' ,  nham =', nham
         write(ilog,'(a,i3,3a,i4,i2,2x,a,9i3)') 'nham =', nham,&
            ',  operator = ',operator(1:slen(operator)),&
              '  kzahl:',kzahl(nham), nmulpot(nham),'khzahl:',&
             (khzahl(n,1,nham),n=1,nmode)
         if(.not.nodiag(nham)) then
            ilbl=index(hamlab(nham),' ')-1
            routine='crosscorr'
            message='The Hamiltonian '//hamlab(nham)(1:ilbl)//&
                 ' must be of "nodiag" type.'
            call errormsg
         endif
      endif

!-----------------------------------------------------------------------
! save time-dependent data set (ket,ipsi1) as set 2 
!-----------------------------------------------------------------------
      call savecsys(2)
      call savecpsi(2)
      call savecdat(2)

!-----------------------------------------------------------------------
! read data from psi file. (reference data set, bra).
!-----------------------------------------------------------------------
      chkdvr=5
      chkgrd=2
      if ( lrst ) then
         irst=ipsi
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
         if (citype .gt. 0) then
            routine='crosscorr'
            message='citype .gt.0 not implemented for restart files.'
            call errormsg        
         endif
         filename='restart (-r, reference WF)'
         if (lerr) goto 999
         if(normthermal .gt. 0.d0) then
            write(6,'(a,1pe16.9)') &
               'reference data set, bra: normthermal =',normthermal
         endif
      else
         read(ipsi) filever(ipsi)
         if(sbra.gt.0) chkdvr = 6
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
         filename='psi (reference WF)'
         if (chkerr(1) .ne. 0) go to 999
!         if(sbra.gt.0) chkerr(2) = 0
      endif
      call savecpsi(1)

!-----------------------------------------------------------------------
! compare psi and psi1 data to decide on overlap calculation type
!-----------------------------------------------------------------------
      ovtype = 0
      ebtype = 0
      readsys = .false.
      if ((chkerr(2).ne.0 .or. chkerr(3).ne.0) .and. .not.lrst ) then
         if(lignore) then
            write(6,'(a,/a)') ' ############ WARNING #############',&
                              'Different primitive bases'
         else
            routine='crosscorr'
            message='Different primitive bases'
            call errormsg
         endif
      endif
      if (chkerr(4) .ne. 0) then
         ovtype = 1        ! Mode combinations differ
         if (dentype .ne. 0) then
            routine='crosscorr'
            message='ovtype=1 not yet implemented'
            call errormsg
         endif
         readsys=.true.
      endif
      if (chkerr(5) .ne. 0 .and. sbra.eq.0) then
         ebtype = 1        ! Electronic bases differ
         readsys=.true.
      endif
      if(nmode.eq.1 .and. .not.readsys) then
         lexact=.true.
      else
         lexact=.false.
      endif

      if (cpsitype(1)==5 .or. cpsitype(2)==5) then
         if (cpsitype(1) /= cpsitype(2)) then
            message= &
    "crosscorr between ML and non-ML wavefunctions not implemented, use ml2mctdh"
            call errormsg
         endif
         if (lcrsspf) then
            message="option -spf not yet implemented for multi-layer"
            call errormsg
         endif
         if (lcrsdiag) then
            message="option -diag not yet implemented for multi-layer"
            call errormsg
         endif
         ovtype = 3
         lrunml = .true.
      endif

      write(ilog,'(a,i3,a,i3,a,L2)') &
         'ovtype =',ovtype,' ,  ebtype =',ebtype,',  lexact =',lexact

!-----------------------------------------------------------------------
! now read data for psi file. (reference data set, bra).
!-----------------------------------------------------------------------
      rewind(ipsi)
      if(loper) then
         call getcsys(2)
         call getcpsi(2)
         call getcdat(2)
         chkdvr=2
         chkgrd=2
         chkpsi=2
         chkdat=2
      else
         chkdvr=5
         chkgrd=1
         chkpsi=1
         chkdat=1
      endif 
      if ( lrst ) then
         irst=ipsi
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
      else
         if(.not. loper .and. sbra.gt.0)chkdvr=6
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      endif
      if (nmode .eq. 1) then
         do s=1,nstate
            dim(1,s)=1
            do t=1,nstate
               ddim(1,s,t)=1
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! Use of operator requires identical structures of bra and ket.  
!-----------------------------------------------------------------------
      if(loper) then
         if( psitype .ne. 0 .or. cpsitype(2) .ne. 0 ) then
            write(6,*) psitype, cpsitype(2)
            routine='crosscorr'
            message="Both WFs, bra and ket, must be of"// &
                    "MCTDH-type if -O option (operator) is used."
            call errormsg
         endif
         if (lrunml) then
            if (ANY(tape(:) /= ctape(:,2))) then
               routine='crosscorr'
               message= "Tree structures and SPF numbers must be"// &
                    " identical if -O option (operator) is used."
               call errormsg
            endif
         else

            filename='reference WF, bra'
            call chksyserr(filename,lerr)
            if(lignore) lerr=.false.
            do s = 1, nstate
               do m = 1, nmode
                  if( cdim(m,s,1) .ne. cdim(m,s,2) ) then
                     lerr=.true.
                     write(6,'(a,2i4,a,2i6)') ' s, m', s,m,  &
                       ',  dim (bra/ket)',cdim(m,s,1),cdim(m,s,2)
                  endif
               enddo
            enddo
            if(lerr.or.readsys) then
               write(6,'(/a,2i4)') 'ovtype, ebtype =',ovtype,ebtype
               routine='crosscorr'
               message="dim's and states of bra and ket must be "//&
                       "identical if -O option (operator) is used."
               call errormsg
            endif
         endif
      endif


!-----------------------------------------------------------------------
! save reference data (bra,ipsi) as set 1
!-----------------------------------------------------------------------
      call savecsys(1)
      call savecpsi(1)
      call savecdat(1)

!-----------------------------------------------------------------------
! Check if sbra is set (-s option). Only WFs and ovtype=0 are then allowed. 
!-----------------------------------------------------------------------
      if (sbra.gt.0) then
         if(dentype.ne.0) then
            routine='overlap'
            message='-s option is not allowed for density operators'
            call errormsg
         endif  
         if(ovtype .ne. 0) then
            routine='overlap'
            message='-s option requires identical grids'//&
                    ' and combination schemes' 
            call errormsg
         endif  
      endif

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
!-----------------------------------------------------------------------
      if(allocated(dovmat))deallocate(dovmat)
      if(allocated(gdovmat))deallocate(gdovmat)
      if(allocated(dovmat))deallocate(dovmat)
      cerr = 'dovmat  '
      allocate(dovmat(maxdim,maxsta),stat=aerr)
      allocmemory=allocmemory + maxdim*maxsta*realsize()
      if (aerr > 0) go to 1000
      cerr = 'gdovmat  '
      allocate(gdovmat(maxdim,maxsta),stat=aerr)
      allocmemory=allocmemory + maxdim*maxsta*realsize()
      if (aerr > 0) go to 1000
      cerr = 'ddovmat  '
      allocate(ddovmat(maxdim,maxsta,maxsta),stat=aerr)
      allocmemory=allocmemory + maxdim*maxsta*realsize()
      if (aerr > 0) go to 1000

      lpsi    = .true.
      lpsi1   = .true.
      lrdoper = loper
      lort    = loper
      lsphfbr = loper
      ltrafo  = .false.
      ldmat   = .true.
      ldmat1  = .false.
      lgpop   = .false.
      lpsigrd=ovtype.eq.1
      call zeigausw
      call memausw
      ovmaxblock=0

      if (ovtype .eq. 0 .and. dentype .eq. 1) then
         zeig=1
         do s=1,nstate
            do t=1,nstate
               ovmaxblock=max(ovmaxblock,dblock(s,t),cdblock(s,t,2))
               zeig1=cdblock(s,t,2)
               do m=1,nmode
                  ddovmat(m,s,t)=zeig
                  zeig=zeig+ddim(m,s,t)*cddim(m,s,t,2)
                  zeig1=zeig1*cddim(m,s,t,1)/cddim(m,s,t,2)
                  ovmaxblock=max(ovmaxblock,zeig1)
               enddo
            enddo
         enddo
         doverdim=zeig-1
         griddim=1

      else if (ovtype .eq. 0) then
         if(sbra.eq.0)then
            zeig=1
            do s=1,nstate
               ovmaxblock=max(ovmaxblock,block(s),cblock(s,2))
               zeig1=cblock(s,2)
               do m=1,nmode
                  dovmat(m,s)=zeig
                  zeig=zeig+dim(m,s)*cdim(m,s,2)
                  zeig1=zeig1*dim(m,s)/cdim(m,s,2)
                  ovmaxblock=max(ovmaxblock,zeig1)
               enddo
            enddo
            doverdim=zeig-1
            griddim=1
         else
            zeig1=max(cblock(sket,1),cblock(sbra,2))
            do m=1,nmode
               zeig2=zeig1
               zeig1=zeig1*cdim(m,sbra,2)/cdim(m,sket,1)
               ovmaxblock=max(ovmaxblock,zeig1)
               zeig2=zeig2+zeig1
               workcdim=max(workcdim,zeig2)
            enddo
            zeig=1
            zeig1=max(cblock(sbra,2),block(sket))
            do m=1,nmode
               dovmat(m,1)=zeig
               zeig=zeig+cdim(m,sket,1)*cdim(m,sbra,2)
               zeig2=zeig1
               zeig1=zeig1*dim(m,sket)/cdim(m,sbra,2)
               ovmaxblock=max(ovmaxblock,zeig1)
               zeig2=zeig2+zeig1
               workcdim=max(workcdim,zeig2)
            enddo
            workcdim=max(workcdim,2*maxblock)
            doverdim=zeig-1
         endif

      else if (ovtype .eq. 1) then  ! combination schemes differ
         doverdim=1                 ! use full grid (exact WF)
         griddim=nstate
         do m=1,nmode
            griddim=griddim*subdim(m)
         enddo

      else if (ovtype .eq. 3) then  ! ML wavefunctions
         zeig=1
         do s=1,nstate
            zeig1=1
            do m=1,nmode
               dovmat(m,s)=zeig
               zeig=zeig+dim(m,s)*cdim(m,s,2)
            enddo
         enddo
         doverdim=dmatdim  ! needed for lowest layer overlap matrices
         griddim=1

      else
         routine = 'Crosscorr'
         message = 'unimplemented ovtype'
         call errormsg

      endif

!-----------------------------------------------------------------------
! set nst2
!-----------------------------------------------------------------------
      if(nst2.eq.0) then
         nst2 = nstate
      else if(nst2.gt.nstate) then
         write(6,'(a,i3,a,/11x,a,i3,a)') '###  ERROR: The number',&
         nst2,' following the -el option must not be larger',&
         ' then nstate =',nstate,'. (STOP).'
         stop
      endif
 
!-----------------------------------------------------------------------
! check time-steps
!-----------------------------------------------------------------------
      if(lmatrixel .and. abs(out2-cout2(2)).gt.1.d-11) then
         write(6,'(/,a)')'###############  WARNING  ###################'
         write(6,'(a)') 'The time steps (tpsi) of the two WFs differ!'
         write(6,'(a)') 'You will overlap the WFs at different times.'
         write(6,'(a,f15.9,a)') 'Bra:  t_psi_ref =', out2/fs,' fs'
         write(6,'(a,f15.9,a)') 'Ket:  t_psi     =', cout2(2)/fs,' fs'
         write(6,'(a,/)')'#############################################'
         write(ilog,'(a)')'The time steps (tpsi) of the two WFs differ!'
         write(ilog,'(a,f15.9,a)') 'Bra:  t_psi_ref =', out2/fs
         write(ilog,'(a,f15.9,a)') 'Ket:  t_psi     =', cout2(2)/fs
      endif

!-----------------------------------------------------------------------
! assign memory arrays
!-----------------------------------------------------------------------
      call memausw

      if (ovtype .eq. 0) then
         mcovr=mcdim
         mcdim=mcovr+cdmatdim(1)
         mcovr1=mcdim
         mcdim=mcovr1+cdmatdim(2)
         if (dentype .eq. 0) then
            mcdover=mcdim
            mcdim=mcdover+doverdim
         else if (dentype .eq. 1) then
            mcdover=mcdim
            mcdim=mcdover+doverdim
         else if (dentype .eq. 2) then
            mcdover=mcdim
            mcdim=mcdover+2*doverdim
         endif
      else if (ovtype .eq. 1) then
         mcovr=1
         mcovr1=1
         mcdover=1
      else if (ovtype .eq. 3) then  ! ML wavefunctions
         mcovr=1
         mcovr1=1
         mcdover=mcdim
         mcdim=mcdover+doverdim
      endif

!-----------------------------------------------------------------------
! set up various flags and array information needed to use an operator
!-----------------------------------------------------------------------
      if(loper)  then
         call operflag(nham)
         call zeighpsi(nham)

         mchpsi  = mcdim
         mcdim   = mchpsi+hpsidim
         workcdim=max(workcdim,dgldim,3*ovmaxblock)
         if(lexact) workcdim = max(workcdim,2*dgldim+2)
      else
         mchpsi  = 1
      end if

!-----------------------------------------------------------------------
! assign work arrays
!-----------------------------------------------------------------------
      if (ovtype .eq. 0 .and. dentype .eq. 0) then
         workcdim=max(workcdim,2*ovmaxblock)
      else if (ovtype .eq. 0 .and. dentype .eq. 1) then
         workcdim=max(workcdim,2*ovmaxblock)
      else if (ovtype .eq. 0 .and. dentype .eq. 2) then
         workcdim=max(workcdim,2*ovmaxblock*ovmaxblock)
      else if (ovtype .eq. 1) then
         workcdim=max(workcdim,2*(griddim/nstate)+cdgldim(1))
         workcdim=max(workcdim,2*(griddim/nstate)+cdgldim(2))
      endif

!-----------------------------------------------------------------------
! allocate work array
!-----------------------------------------------------------------------
      cerr = 'workc  '
      allocate(workc(workcdim),stat=aerr)
      allocmemory=allocmemory + workcdim*2*realsize()
      if (aerr > 0) go to 1000

!-----------------------------------------------------------------------
! assign memory for operator
!-----------------------------------------------------------------------
      if(loper) call memhops

!-----------------------------------------------------------------------
! call overlap routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))

!-----------------------------------------------------------------------
! get lengths of comparison data arrays
!-----------------------------------------------------------------------
      call getcpsi(2)
      dgldim1=dgldim
      adim1=adim
      dmatdim1=dmatdim
      agmatdim1=agmatdim
      gsmatdim1=gsmatdim
      gwpinfodim1=gwpinfodim
      call getcpsi(1)

allocate(agmat(agmatdim))
allocate(agmat1(agmatdim1))

      if(loper) call corroperin(mc,mr,mi)
      call getcpsi(1)
      call ovrlp(mc(mcpsi),mc(mcpsi1),ms(mspsi),mc(mcdover),&
                 workc,mi(mijindx),mi(mijindx1),&
                 agmat,agmat1,mc(mcadgwp),mc(mcadgwp1))


deallocate(mc,mr,mi,ml,ms,mf)
deallocate(agmat,agmat1)
deallocate(dovmat,gdovmat,ddovmat)
deallocate(workc)

      stop

  888 continue
      routine='crosscorr'
      message= 'Error reading oper-info from:'//operfile(1:opflaenge)
      write(6,'(a)') message
      stop

  900 continue
      routine='crosscorr'
      message= 'Error opening operfile:'//operfile(1:opflaenge)
      write(6,'(a)') message
      stop

  999 continue
      routine='crosscorr'
      message= 'Error reading: '//filename
      call errormsg
 
 1000 write (6,*) 'Error in Memory Allocation of ',cerr,' Error code is ',aerr
      stop

      stop
      contains

!***********************************************************************
!     calculates the overlap
!***********************************************************************

      subroutine ovrlp (psi,psi1,spsi,dover,workc,jindx,jindx1,&
                        agmat,agmat1,adgwp,adgwp1)

      implicit none

      integer(long)                                   :: i,m,s
      integer(long), dimension(1)                     :: idum
      integer(long), dimension(adim), intent(out)     :: jindx
      integer(long), dimension(adim1), intent(out)    :: jindx1
      integer(long), dimension(gwpinfodim)            :: gwpdep
      integer(long), dimension(gwpinfodim1)           :: gwpdep1
      complex(dop)                                    :: ovl2,ovs,phase
      complex(dop), dimension(*), intent(out)         :: dover
      complex(sip), dimension(spsidim), intent(out)   :: spsi
      complex(dop), dimension(dgldim), intent(out)    :: psi
      complex(dop), dimension(dgldim1), intent(out)   :: psi1
      complex(dop), dimension(workcdim), intent(out)  :: workc
      complex(dop), dimension(agmatdim), intent(out)  :: agmat
      complex(dop), dimension(agmatdim1), intent(out) :: agmat1
      complex(dop), dimension(gsmatdim), intent(out)  :: adgwp
      complex(dop), dimension(gsmatdim1), intent(out) :: adgwp1

      real(dop)                                       :: time,dt
      real(dop), dimension(maxdim,maxsta)             :: orbcrs

      logical(kind=4)                                 :: lerr,useunit,&
                                                         lstartup,jump
      character(len=c2)                               :: rdtag

      type(wfnode_t),pointer :: kettree => null()
      type(wfnode_t),pointer :: bratree => null()

      lstartup = .true.

!-----------------------------------------------------------------------
! set up ML
!-----------------------------------------------------------------------
      if (lrunml) then
         call inputml(tape)
      endif

!-----------------------------------------------------------------------
! allocate data structures for ML
!-----------------------------------------------------------------------
      ! bra (ref)
      if (cpsitype(1).eq.5) then
         bratree => new_wftree(ctape(:,1))
         call wftree_set_pointers(bratree,cdgldim(1))
      endif
      ! ket (TD)
      if (cpsitype(2).eq.5) then
         kettree => new_wftree(ctape(:,2))
         call wftree_set_pointers(kettree,cdgldim(2))
      endif

      if (lrunml .and. loper) then
         useunit = .not. nodiag(nham)
         call setoperml(nham,.false.,useunit)
      endif

!-----------------------------------------------------------------------
! Read reference wavefunction (bra).
!-----------------------------------------------------------------------
      call getcsys(1)
      call getcpsi(1)
      if ( lrst ) then
         if (filever(ipsi1) .lt. 11.0_dop) then
            read(ipsi,err=999) time
            read(ipsi,err=999) (psi(i),i=1,dgldim)
         else
            call rdrstpsi(time,psi,jindx,agmat,idum)
         endif
      else
         if(lsingle) init = idxbra
         do i = 1, init-1
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) then
               write(6,'(a,i4,a)') 'ERROR: There are only',&
                    i-1,' WFs on the reference psi-file.'
               stop
            endif
         enddo
         if(ovtype .eq. 1) then
            lrst=.false.
            call rdpsigrid(ipsi,psi,spsi,jindx,workc,lrst)
         else
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
         endif
         if (lend) then
            write(6,'(a,i4,a)') 'ERROR: There are only',&
                  init-1,' WFs on the reference psi-file.'
            stop
         endif
      endif

      if(.not.lmatrixel) close(ipsi)  ! -M is not set.
      if(lhc) then   ! -HC is set.
         do i=1,dgldim
            psi(i)=dconjg(psi(i))
         enddo
      endif

      if(laeout.eq.0) iaus = 6
      if(lcrsspf) then
         if(fs.gt.10.d0) then
            write(iaus,'(a,9x,a,4x,2a,11x,20i8)') &
              '#    time[fs]','Re(cross-correl)', &
              'Im(cross-correl)      Abs(cross-correl)', &
              '       Abs^2',&
              ((dim(m,s),m=1,nmode),s=1,nstate)
         else
            write(iaus,'(a,9x,a,4x,2a,11x,20i8)') &
              '#    time[au]','Re(cross-correl)', &
              'Im(cross-correl)    Abs(cross-correl)', &
              '       Abs^2', &
              ((dim(m,s),m=1,nmode),s=1,nstate)
         endif
      else
         if(fs.gt.10.d0) then
            write(iaus,'(a,9x,a,3x,2a)') &
              '#    time[fs]','Re(cross-correl)', &
              'Im(cross-correl)    Abs(cross-correl)', &
              '       Abs^2'
         else
            write(iaus,'(a,9x,a,3x,2a)') &
              '#    time[au]','Re(cross-correl)', &
              'Im(cross-correl)    Abs(cross-correl)', &
              '       Abs^2'
         endif
      endif
      if(laeout.gt.0) then
         if(laeout.lt.44) then
            write(6,'(2a)')' Cross-correlation written to file : ', &
                             fileout(1:laeout)
         else
            write(6,'(a,/1x,a)')' Cross-correlation written to file:', &
                             fileout(1:laeout)
         end if
      end if

!-----------------------------------------------------------------------
! calculate time step for data
!-----------------------------------------------------------------------
      out2   = cout2(2)
      dt     = out2*step
      time   = tinit
      jump   = .true.

   10 continue       ! Loop over times.
!-----------------------------------------------------------------------
! choose overlap calculation routine
!-----------------------------------------------------------------------
      if (ovtype .eq. 0) then
         call getcpsi(2)
         if(jump) then
            time=time+dt

            if ( lrst1 ) then
               if (filever(ipsi1) .lt. 11.0_dop) then
                  read(ipsi1,err=998) time
                  read(ipsi1,err=998) (psi1(i),i=1,dgldim)
               else
                  call rdrstpsi(time,psi1,jindx1,agmat1,idum)
               endif
               out2=0.0_dop
               dt  =0.0_dop 
            else

               if(lsingle)then
                  do i = 1, idxket - 1
                     call rdpsi(ipsi1,psi1,spsi,jindx1,agmat1,idum,adgwp1,gwpdep1)
                     if (lend) then
                        write(6,'(a,i4,a)') 'ERROR: There are only',&
                             i-1,' WFs on the psi-file.'
                        stop
                     endif
                  enddo
               endif

               call rdpsi(ipsi1,psi1,spsi,jindx1,agmat1,idum,adgwp1,gwpdep1)
               if (lend) goto 77
            endif
         endif
         call getcpsi(1)

         if(lcrsdiag) write(6,'(a,f8.2,a)') '--------------  Time =',&
              time/fs,' -----------------------'
         if(lexact) then
            if(loper) then
               call cpvxz(psi1,workc,dgldim)
               call opwfex(workc,psi1,nham,.true.,.true.)
            endif
            ovl2=0.d0
            do s = 1, nstate
               call vvaxzz(psi(zetf(1,s)),psi1(zetf(1,s)),ovs,subdim(1))
               ovl2 = ovl2 + ovs*dconjg(psi(zpsi(s)))*psi1(zpsi(s))
            enddo
         else if(loper) then
            if(dentype.ne.0 .or. ovtype.ne.0) then
               routine='ovlrp  on analyse/crosscorr.F'
               message='Use of operators is only possible for'//&
               'MCTDH wavefunctions with identical mode combinations.' 
               call errormsg
            endif
            if(lgwp) then
               routine='ovlrp  on analyse/crosscorr.F'
               message='Use of operators is not possible for'//&
               'G-MCTDH wavefunctions.' 
               call errormsg
            endif
            call mtrxelement(ovl2,psi,psi1,workc,time,&
                     mc(mchpsi),hteil,mr(mrhops),&
                     mr(mrort),jindx)

         else if (dentype .eq. 0) then
            if(sbra .gt. 0)then
               call overlap2cs(ovl2,psi,psi1,dover,workc,orbcrs,agmat,&
                    agmat1,sbra,sket)
            else
               call overspf2(psi,psi1,dover,orbcrs,agmat,agmat1)
               call overlap2c(ovl2,psi,psi1,dover,workc,orbcrs,agmat,&
                    agmat1)
            endif
         else if (dentype .eq. 1) then
            call d1overlap2c(ovl2,psi,psi1,dover,workc)
         else if (dentype .eq. 2) then
            call d2overlap2c(ovl2,psi,psi1,dover,workc,orbcrs)
         endif

      else if (ovtype .eq. 1) then
         if(loper) then
            routine='crosscorr'
            message='Operators are not allowed for ovtype.eq.1'
            call errormsg
         endif
         if(lmatrixel) then
            routine='crosscorr'
            message='-M option is not allowed for ovtype.eq.1'
            call errormsg
         endif
         call getcsys(2)
         call getcpsi(2)
         call rdpsigrid(ipsi1,psi1,spsi,jindx1,workc,lrst)
         call getcsys(1)
         call getcpsi(1)
         if (lend) goto 77
         call vvaxzz(psi,psi1,ovl2,griddim)

      else if (ovtype .eq. 3) then  ! ML wavefunctions
         call getcpsi(2)
         if(jump) then
            time=time+dt
            if ( lrst1 ) then
               if (filever(ipsi1) .lt. 11.0_dop) then
                  read(ipsi1,err=998) time
                  read(ipsi1,err=998) (psi1(i),i=1,dgldim)
               else
                  call rdrstpsi(time,psi1,jindx1,agmat1,idum)
               endif
               out2=0.0_dop
               dt  =0.0_dop 
            else

               call rdpsi(ipsi1,psi1,spsi,jindx1,agmat1,idum,adgwp1,gwpdep1)
               if (lend) goto 77
            endif
         endif
         call getcpsi(1)

         if (loper) then
            ! Trees will be identical, tree has been set by inputml.
            ovl2 = croco_tree(nham,time,kettree,psi,psi1)
         else
            call overspf2(psi,psi1,dover,orbcrs,agmat,agmat1)
            call overlap_2(psi,psi1,bratree,kettree,ovl2,dover) ! see analysis.f90
         endif

      endif

      if (lstartup .and. lphasereal) then
         phase = dconjg(ovl2)/abs(ovl2)
         lstartup = .false.
      end if
      if (lphasereal) ovl2 = ovl2*phase

      if(laeout.eq.0) iaus = 6 
      if(lcrsspf) then
         write(iaus,'(1x,g18.9,4(1x,es18.11),2x,20f8.4)') &
              time/fs,dble(ovl2),dimag(ovl2),abs(ovl2),abs(ovl2)**2 &
              ,((orbcrs(m,s),m=1,nmode),s=1,nstate)
      else
         if(abs(ovl2) .le. 9.99999d0 .and. abs(ovl2) .gt. 0.001d0) then
            write(iaus,'(1x,g18.9,3(1x,F18.14),es18.10)') &
              time/fs,dble(ovl2),dimag(ovl2),abs(ovl2),abs(ovl2)**2
         else
            write(iaus,'(1x,g18.9,4(1x,es18.11))') &
              time/fs,dble(ovl2),dimag(ovl2),abs(ovl2),abs(ovl2)**2
         endif
      endif
      if (iaus .ne. 6) call myflush(iaus)

      if(lrst1) goto 77

!-----------------------------------------------------------------------
! Dummy read of psi1 data sets if step>1
!-----------------------------------------------------------------------
      if(step>1 .and. (ovtype==0 .or. ovtype==3)) then
         do i=2,step
            call rdpsi(ipsi1,psi1,spsi,jindx1,agmat1,idum,adgwp1,gwpdep1)
            if (lend) goto 77
         enddo
      endif

!-----------------------------------------------------------------------
!  If option -M is set, read the bra state for the next time point
!-----------------------------------------------------------------------
      if (lmatrixel .and. (jump .or. .not.ldt)) then  !  -M is set.
         if(ldt) time=time+dt
         do i=1,step
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) goto 77
         enddo
         if(lhc) then  !-HC is set.
            do i=1,dgldim
               psi(i)=dconjg(psi(i))
            enddo
         endif
      endif
      jump=.not.(jump.and.ldt)
      call myflush(iaus)

!-----------------------------------------------------------------------
! if more data, go to beginning of loop
!-----------------------------------------------------------------------
      if( time/fs .le. tfin .and. (.not.lsingle) ) goto 10 

 77   continue
      if(lcrsspf) then
         write(iaus,'(a,9x,a,3x,2a,9x,20i8)') &
              '#    time[fs]','Re(cross-correl)',&
              'Im(cross-correl)   Abs(cross-correl)',&
              '   Abs^2' &
              ,((dim(m,s),m=1,nmode),s=1,nstate) 
      else
         write(iaus,'(a,9x,a,3x,2a)') &
              '#    time[fs]','Re(cross-correl)',&
              'Im(cross-correl)   Abs(cross-correl)',&
              '   Abs^2'
      endif
      if(laeout.eq.0) then
         write(6,'(80a)') ('-',i=1,80)
      else
         close(iaus)
      endif
      close(ipsi1)

      return

 998  write(6,*) ' ERROR reading restart file (-R, WF)'
      stop
 999  write(6,*) ' ERROR reading restart file (-r, ref. WF)'
      stop

      end subroutine ovrlp

!#######################################################################
!
! MTRXELEMENT
!
! Calculates the matrix-element of oper between two wavefucntions in 
! MCTDH form with identical mode combinations.
!
!  ovl2 = <psi|oper|psi1>
!
!#######################################################################

      subroutine mtrxelement(ovl2,psi,psi1,workc,time,&
                    hpsi,hteil,hops,ort,jindx)

      implicit none

      integer(long)                                  :: s,ktype
      integer(long), dimension(adim), intent(in)     :: jindx
      complex(dop)                                   :: csum
      complex(dop), intent(out)                      :: ovl2
      complex(dop), dimension(1)                     :: zdum
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim1), intent(in)   :: psi1
      complex(dop), dimension(dgldim)                :: dtpsi
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), dimension(hpsidim), intent(out)  :: hpsi
      complex(dop), dimension(pmatdim), intent(out)  :: hteil
      real(dop), intent(in)                          :: time
      real(dop), dimension(ortdim), intent(in)       :: ort
      real(dop), dimension(hopsdim), intent(in)      :: hops

! --- NOTE: This all works only, if psi and psi1 have exactly the same
! ---       structure. I.e. same mode-combinations, same no of SPFs, etc.

      call zerovxz(hteil,pmatdim)
      call zerovxz(dtpsi,dgldim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
      if (ltdh(nham)) call calctdk(time,nham,psi)

      ktype=0
      call hphi(psi1,hpsi,ktype,nham)
      call phihphi(psi,zdum,hpsi,hteil,nham,.false.)
      call calcha(psi1,dtpsi,hteil,hops,psi1(zetf(1,1)),nham,&
                  psi(zetf(1,1)),jindx,ort)

      ovl2=0.0_dop
      do s=nst1,nst2
         call vvaxzz(psi(zpsi(s)),dtpsi(zpsi(s)),csum,block(s))
         ovl2=ovl2+csum
      enddo

      return
      end subroutine mtrxelement                  ! of mtrxelement



!#######################################################################
!
! OVERLAP2c
!
! Calculates the overlap between two wavefunctions in MCTDH form
! (similar to overlap0 of overlap.F, but no norms evaluated). 
!
!  ovl2 = <psi|psi1>
!
!#######################################################################

      subroutine overspf2(psi,psi1,dover,orbcrs,agmat,agmat1)

use global
use aglobal
use maxv, only:maxdim,maxsta
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

use gwplib, only: diagonal_moments

      implicit none

      integer(long)                                    :: s,m,zeig1,zeig2,swapzeig,i
      complex(dop)                                     :: zdum
      complex(dop), dimension(1)                       :: cdum,cdum1
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim1), intent(in)     :: psi1
      complex(dop), dimension(doverdim), intent(out)   :: dover
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      complex(dop), dimension(agmatdim1), intent(in)   :: agmat1
      real(dop)                                        :: norm
      real(dop), dimension(maxdim,maxsta), intent(out) :: orbcrs

!-----------------------------------------------------------------------
!   calculate density matrices for each degree of freedom defined as 
!   overlap of the single-particle functions rho^m_ij=<phi_i,2|phi_j,1>
!-----------------------------------------------------------------------
      do s=nst1,nst2
         do m=1,nmode
            if (gwpm(m)) then
               call diagonal_moments(m,s,s,nspfdof(m),cgwpdim(m,s,1),cgwpdim(m,s,2),&
                    agmat(czaGmat(m,s,1)),agmat1(czaGmat(m,s,2)),&
                    psi(czxi(m,s,1)),psi1(czxi(m,s,2)),&
                    agmat(czphase(m,s,1)),agmat1(czphase(m,s,2)),&
                    dover(dovmat(m,s)),cdum,cdum1,2,0)

            else
!                        reference         time-dependent
               call mmaxzz(psi(czetf(m,s,1)),psi1(czetf(m,s,2)),&
                    dover(dovmat(m,s)),subdim(m),cdim(m,s,1),&
                    cdim(m,s,2))
            endif
            if(lcrsspf) then
               call normvxz(dover(dovmat(m,s)),norm,&
                    cdim(m,s,1)*cdim(m,s,2))
               orbcrs(m,s) = min(cdim(m,s,1),cdim(m,s,2)) - norm**2
!         call wrdover(m,s,dover(dovmat(m,s)),cdim(m,s,1),cdim(m,s,2))
            endif
            if(lcrsdiag)  write(6,'(2i4,2x,10f8.5,9(/10x,10f8.5))')&
                 m,s,(abs(dover(dovmat(m,s)&
                 + (i-1)*(cdim(m,s,1)+1))),i=1,cdim(m,s,2))
         enddo
      enddo

      return
      end subroutine overspf2                  ! of overspf2

!#######################################################################
!
! OVERLAP2c
!
! Calculates the overlap between two wavefucntions in MCTDH form
! (similar to overlap0 of overlap.F, but no norms evaluated). 
!
!  ovl2 = <psi|psi1>
!
!#######################################################################

      subroutine overlap2c(ovl2,psi,psi1,dover,workc,orbcrs,agmat,&
                 agmat1)

use global
use aglobal
use maxv, only:maxdim,maxsta
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

use gwplib, only: diagonal_moments

      implicit none

      integer(long)                                    :: s,m,zeig1,zeig2,swapzeig,i
      complex(dop)                                     :: zdum
      complex(dop), intent(out)                        :: ovl2
      complex(dop), dimension(1)                       :: cdum,cdum1
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim1), intent(in)     :: psi1
      complex(dop), dimension(doverdim), intent(in)    :: dover
      complex(dop), dimension(workcdim), intent(out)   :: workc
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      complex(dop), dimension(agmatdim1), intent(in)   :: agmat1
      real(dop)                                        :: norm
      real(dop), dimension(maxdim,maxsta), intent(out) :: orbcrs

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      call getcpsi(1)
      ovl2=(0.0_dop,0.0_dop)
      do s=nst1,nst2
         zeig1=1
         zeig2=ovmaxblock+1
         call cpvxz(psi1(czpsi(s,2)),workc(zeig2),cblock(s,2))

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(dover(dovmat(m,s)),workc(zeig1),&
                    workc(zeig2),vdim(m,s),cdim(m,s,2),cndim(m,s,2),&
                    dim(m,s))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* {Sum_L <Phi_J,2|Phi_L,1>A_L,1} within each state
!-----------------------------------------------------------------------
         call vvaxzz(psi(zpsi(s)),workc(zeig2),zdum,block(s))
         ovl2=ovl2+zdum
      enddo

      return
      end subroutine overlap2c                  ! of overlap2c



!#######################################################################
!
! OVERLAP2cs
!
! Calculates the overlap between different states of two wavefunctions 
! in MCTDH form (similar to overlap0s of overlap.F90, but no norms 
! evaluated). GWR 09/12
!
!  ovl2 = <psi|psi1>
!
!#######################################################################

      subroutine overlap2cs(ovl2,psi,psi1,dover,workc,orbcrs,agmat,&
                 agmat1,s2,s1)

use global
use aglobal
use maxv, only:maxdim,maxsta
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

      implicit none

      integer(long)                                    :: s,m,zeig1,zeig2,swapzeig,i
      integer(long), intent(in)                        :: s1,s2
      complex(dop)                                     :: zdum
      complex(dop), intent(out)                        :: ovl2
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim1), intent(in)     :: psi1
      complex(dop), dimension(doverdim), intent(out)   :: dover
      complex(dop), dimension(workcdim), intent(out)   :: workc
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      complex(dop), dimension(agmatdim1), intent(in)   :: agmat1
      real(dop)                                        :: norm
      real(dop), dimension(maxdim,maxsta), intent(out) :: orbcrs

!-----------------------------------------------------------------------
!   calculate density matrices for each degree of freedom defined as 
!   overlap of the single-particle functions rho^m_ij=<phi_i,2|phi_j,1>
!-----------------------------------------------------------------------
         do m=1,nmode
            call mmaxzz(psi(czetf(m,s2,1)),psi1(czetf(m,s1,2)),&
                 dover(dovmat(m,1)),subdim(m),cdim(m,s2,1),&
                 cdim(m,s1,2))

            if(lcrsspf) then
               call normvxz(dover(dovmat(m,1)),norm,&
                    cdim(m,s2,1)*cdim(m,s1,2))
               orbcrs(m,s1) = min(cdim(m,s1,1),cdim(m,s1,2)) - norm**2
!!         call wrdover(m,s,dover(dovmat(m,s)),cdim(m,s1,1),cdim(m,s1,2))
            endif
            if(lcrsdiag)  write(6,'(2i4,2x,10f8.5,9(/10x,10f8.5))')&
                 m,s1,(dble(dover(dovmat(m,s1)&
                 + (i-1)*(cdim(m,s1,1)+1))),i=1,cdim(m,s1,2))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      call getcpsi(1)
      ovl2=(0.0_dop,0.0_dop)
      zeig1=1
      zeig2=ovmaxblock+1
      call cpvxz(psi1(czpsi(s1,2)),workc(zeig2),cblock(s1,2))
      
      do m=1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         call mtxxzz(dover(dovmat(m,1)),workc(zeig1),&
              workc(zeig2),vdim(m,s2),cdim(m,s1,2),cndim(m,s1,2),&
              dim(m,s2))
      enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* {Sum_L <Phi_J,2|Phi_L,1>A_L,1} within each state
!-----------------------------------------------------------------------
      call vvaxzz(psi(zpsi(s2)),workc(zeig2),zdum,block(s2))
      ovl2=ovl2+zdum
 
      return
      end subroutine overlap2cs                  ! of overlap2cs



!#######################################################################
!
! D1OVERLAP2c
!
! Calculates the overlap between type I density matrices.
!
!#######################################################################
      subroutine d1overlap2c(ovl2,psi,psi1,dover,workc)

use global
use aglobal
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

      implicit none

      integer(long)                                    :: s,t,m,zeig1,zeig2,swapzeig
      complex(dop)                                     :: zdum
      complex(dop), intent(out)                        :: ovl2
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim1), intent(in)     :: psi1
      complex(dop), dimension(doverdim), intent(out)   :: dover
      complex(dop), dimension(workcdim), intent(out)   :: workc

!-----------------------------------------------------------------------
!     calculate density matrices for each degree of freedom defined as
!     overlap of the single-particle functions rho^m_ij=<phi_i,2|phi_j
!     ,1>
!-----------------------------------------------------------------------
      do s=nst1,nst2
         do t=nst1,nst2
            do m=1,nmode
               call mmaxzz(psi(cdzetf(m,s,t,1)),psi1(cdzetf(m,s,t,2)),&
                    dover(ddovmat(m,s,t)),subdim(m),cddim(m,s,t,1),&
                    cddim(m,s,t,2))
            enddo
         enddo
      enddo


!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      call getcpsi(1)
      ovl2=(0.0_dop,0.0_dop)
      do s=nst1,nst2
         do t=nst1,nst2
            zeig1=1
            zeig2=ovmaxblock+1
            call cpvxz(psi1(cdzpsi(s,t,2)),workc(zeig2),cdblock(s,t,2))

            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call mtxxzz(dover(ddovmat(m,s,t)),workc(zeig1),&
                    workc(zeig2),vddim(m,s,t),cddim(m,s,t,2),&
                    cnddim(m,s,t,2),ddim(m,s,t))
            enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* Sum_L <Phi_J,1|Phi_L,2>A_L,2 within each state
!-----------------------------------------------------------------------
            call vvaxzz(psi(dzpsi(s,t)),workc(zeig2),zdum,dblock(s,t))
            ovl2=ovl2+zdum
         enddo
      enddo

      return
      end subroutine d1overlap2c                      !of d1overlap2c



!***********************************************************************
!                               D2OVERLAP2c
!
!     Computes tr( rho^(dagger) rho ) 
!
!***********************************************************************
      subroutine d2overlap2c(ovl2,psi,psi1,dover,workc,orbcrs)

use global
use aglobal
use maxv, only:maxdim,maxsta
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

      implicit none


      integer(long)                                                 :: s,m,hzw,t,i
      integer(long), dimension(2)                                   :: zw
      complex(dop), intent(out)                                     :: ovl2
      complex(dop)                                                  :: x
      complex(dop), dimension(dgldim), intent(in)                   :: psi
      complex(dop), dimension(dgldim1), intent(in)                  :: psi1
      complex(dop), dimension(doverdim,2), intent(out)              :: dover
      complex(dop), dimension(ovmaxblock*ovmaxblock,2), intent(out) :: workc
      real(dop)                                                     :: norm
      real(dop), dimension(maxdim,maxsta), intent(out)               :: orbcrs


      do s = nst1,nst2
         do m=1,nmode
            call mmaxzz(psi1(czetf(m,s,2)),psi(zetf(m,s)),&
                 dover(dovmat(m,s),1),subdim(m),cdim(m,s,2),dim(m,s))
            call mmaxzz(psi(zetf(m,s)),psi1(czetf(m,s,2)),&
                 dover(dovmat(m,s),2),subdim(m),dim(m,s),cdim(m,s,2))
            if(lcrsspf) then
               call normvxz(dover(dovmat(m,s),1),norm,&
                    cdim(m,s,1)*cdim(m,s,2))
               orbcrs(m,s) = min(cdim(m,s,1),cdim(m,s,2)) - norm**2
            endif
            if(lcrsdiag) then  
                 write(6,'(a,2i4,2x,20f8.5)') '(1) ',&
                 m,s,(dble(dover(dovmat(m,s)&
                 + (i-1)*(cdim(m,s,1)+1),1)),i=1,cdim(m,s,2))
                 write(6,'(a,2i4,2x,20f8.5)') '(2) ',&
                 m,s,(dble(dover(dovmat(m,s)&
                 + (i-1)*(cdim(m,s,1)+1),2)),i=1,cdim(m,s,2))
            endif 
         enddo
      enddo

      ovl2  = (0.0_dop, 0.0_dop)
      do s = nst1,nst2
         do t = nst1,nst2

            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(s,t)),workc(1,zw(1)),block(s)*block(t))

            do m=1,nmode
               call mtxxzz(dover(dovmat(m,s),1),workc(1,zw(1)),&
                    workc(1,zw(2)),cvdim(m,s,2),dim(m,s),&
                    ndim(m,s)*block(t),cdim(m,s,2))
               hzw = zw(1)
               zw(1) = zw(2)
               zw(2) = hzw
            enddo

            do m=1,nmode
               call mttxzz(dover(dovmat(m,t),2),workc(1,zw(1)),&
                    workc(1,zw(2)),cblock(s,2)*cvdim(m,t,2),dim(m,t),&
                    ndim(m,t),cdim(m,t,2))
               hzw = zw(1)
               zw(1) = zw(2)
               zw(2) = hzw
            enddo

            call vvaxzz(psi1(cdzpsi(s,t,2)),workc(1,zw(1)),x,&
                 cblock(s,2)*cblock(t,2))
            ovl2 = ovl2 + x
         enddo
      enddo

      return
      end  subroutine d2overlap2c              ! of d2overlap2c 

!-----------------------------------------------------------------------
!
! CRSCINPUT
!
! get input parameters and file names for calculation
!
!-----------------------------------------------------------------------

      subroutine crscinput

use global
use aglobal
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod
use channels
use crosscorrmod
use versions

      implicit none

      integer(long), external :: myiargc
      integer(long)           :: iarg,ilbl,llaenge,ierr,i,dvrlaenge,&
                                 chkdvr,chkgrd,slen
      character(len=80)       :: buf
      character(len=c5)       :: filename,flabel
      logical(kind=4)         :: lerr
      character(len=c5), save    :: dvrfile

!-----------------------------------------------------------------------

      lcrsspf = .false.
      lcrsdiag= .false.
      lphasereal = .false.
      lrst1   = .false.
      lrst    = .false.
      loper   = .false.
      lmatrixel= .false.
      lhc      = .false.
      lsingle = .false.
      lshort  = .false.
      filein1 = 'psi'
      filein  = ' '
      fileout = ' '
      operfile= 'oper'
      operator= ' '
      dvrfile = 'dvr'

      sbra = 0
      sket = 0
      init = 1
      nst1 = 1
      nst2 = 0
      optlaenge=1
      options = '   ' 
      tfin = 1.0e99_dop
      iarg = myiargc()
      if (iarg .lt. 1) goto 200 
      
      do i = 1, iarg
         call mygetarg(i,buf)
         ilbl  = index(buf,' ')
         ilbl = min(ilbl,100-optlaenge)
         if(ilbl.gt.0) options = options(1:optlaenge)//buf(1:ilbl) 
         optlaenge = optlaenge + ilbl
      end do

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf(1:3) .eq. '-f ' ) then
            i = i+1
            call mygetarg(i,buf)
            filein1 = buf
         else if( buf(1:3) .eq. '-o ' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
         else if( buf(1:3) .eq. '-i ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) init
            if (ierr .ne. 0  .or.  init .le. 0) then
               write(6,'(2a/1x,a)')&
                 ' The  argument following the option -i',&
                 ' must be a positive integer. Not : ', buf
               stop
            end if
         else if(buf(1:4).eq.'-el ' .or. buf(1:4).eq.'-es ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) nst1
            nst2 = nst1
            if (ierr.ne.0  .or. nst1.le.0) then
               write(6,'(2a/1x,a)')&
                 ' The  argument following the option -el',&
                 ' must be positive integer. Not : ', buf
               stop
            end if
         else if( buf(1:3) .eq. '-s ' ) then 
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) sbra
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) sket
            if (ierr.ne.0  .or.  sbra.le.0 .or. sket.le.0) then
               write(6,'(2a/,a)')&
                 ' The  arguments following the option -s',&
                 ' must be a positive integers. Not : ', buf
               stop
            end if
         else if( buf(1:3) .eq. '-n ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (ierr .ne. 0  .or.  step .le. 0) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -n',&
                 ' must be a positive integer. Not : ', buf
               stop
            end if
         else if( buf .eq. '-nofs' ) then
            fs = 1.d0
         else if( buf(1:3) .eq. '-r ' ) then
            lrst = .true.
            filein = 'restart'
         else if( buf(1:3) .eq. '-R ' ) then
            lrst1 = .true.
            filein1 = 'restart'
         else if( buf(1:3) .eq. '-O ' ) then
            loper = .true.
            i = i+1
            call mygetarg(i,operator)
         else if( buf(1:3) .eq. '-t ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) tfin
            if (ierr .ne. 0 ) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -t',&
                 ' must be a real number. Not : ', buf
               stop
            endif
         else if( buf(1:5) .eq. '-spf ' ) then
            lcrsspf = .true.
         else if( buf(1:6) .eq. '-diag ' ) then
            lcrsdiag = .true.
         else if(buf(1:3) .eq. '-ph') then
            lphasereal = .true.
         else if( buf(1:4) .eq. '-ig ' ) then
            lignore = .true.
         else if( buf(1:3) .eq. '-M ' ) then
            lmatrixel = .true.
         else if( buf(1:4) .eq. '-HC ' ) then
            lhc = .true.
         else if( buf(1:4) .eq. '-DT ' ) then
            ldt = .true.
            lhc = .true.
            lmatrixel = .true.
         else if( buf(1:7) .eq. '-short ' ) then
            lshort = .true.
         elseif( buf(1:8) .eq. '-single ' ) then
            lsingle = .true.
            i = i + 1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) idxbra
            i = i + 1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) idxket
            if (ierr .ne. 0 ) then
               write(6,'(2a/,a)')&
                 ' The  arguments following the option -single',&
                 ' must be integers. Not : ', buf
               stop
            endif
         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(1x,a,/10x,a)')&
       'Purpose: Calculates the cross-correlation function c(t)'&
       , 'c(t) = <psi_ref(0)|psi(t)>'
            write(6,'(/1x,22a)')&
       'Usage  : crosscorr [-f -i -o -n -r -R -O -t -nofs -el -ig', &
       ' -spf -diag -ph -M -HC -DT -short -single -ver -h -?]  psi_ref .'
            write(6,'(/,a)') 'Options : '
            write(6,'(a,/,11x,a,/,11x,a)') &
                     ' -f FILE : The wavefunction (density operator) '&
                    ,'is read from file FILE rather than from ./psi'&
            ,'The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,11x,a/,11x,a/,11x,a)') &
                     ' -o FILE : The output is written to file ',&
                     'FILE rather than to ./cross_XXX',&
             'The string FILE may be a relative or a full path-name.',&
             'If FILE=no, i.e. "-o no", no cross-file will be opened',&
             'and the output will be directed to screen.'
            write(6,'(2a)') ' -r      : Take the restart rather',&
             ' than the psi file as reference WF.'
            write(6,'(2a,/11x,a)') ' -R      : Take the ',&
             'restart rather than the psi file as time-dep. WF.',&
             'In general, this makes sense only when also -r is set.'
            write(6,'(2a)') ' -O oper : Apply operator ',&
                            '"oper" before evaluating the overlap.'
            write(6,'(2a)') ' -t tfin : Compute the overlap',&
             ' only up to tfin.'
            write(6,'(a,a)') &
             ' -nofs   : No transformation to fs. Use when', &
                       " 'time-not-fs' was set in mctdh."
            write(6,'(2a)') ' -n step : Compute the overlap',&
             ' only every step-th output.'
            write(6,'(2a,/11x,a)') ' -i init : Compute the overlap', &
                 ' using the init-th (rather than the first)',&
                 'reference WF. (-i is ignored when -r is set).'
            write(6,'(2a)') &
                     ' -el nst : Compute the cross-correlation',&
                     ' for state nst only.'
            write(6,'(a)') &
                     ' -ig     : Ignore "different bases" error-msg.'
            write(6,'(2a,/11x,a)')&
                     ' -spf    : Print information on overlap of the ',&
                     'two orbital sets.',&
                     'dim - trace(dover^+ * dover) for all m,s.'
            write(6,'(2a,/11x,a)') &
                     ' -diag   : Print the absolute values of diagonal', &
                     ' elements of the', &
                     'overlap matrix of the two orbital sets.'
            write(6,'(2a,/11x,a)') &
                     ' -ph     : Multiply crosscorrelation with a constant',&
                     ' phase such that C(t=0)','is real.'
            write(6,'(2a,/11x,a)') &
                     ' -M      : Matrix-elements. The reference state', &
                     ' is taken as time-dependent.', &
                     'This requires MCTDH WFs of identical structure.'
            write(6,'(2a,/11x,a)') &
                     ' -HC     : Take the complex conjugate of the reference',&
                     ' WF. (symmetric scalar','prod.)'
            write(6,'(2a,/11x,a)') &
                     ' -DT     : Time is doubled, -M and -HC are set.',&
                     ' (For computing correlation','with t/2 trick)'
            write(6, '(2a, /11x,2a)')' -s s s1 : Compute the overlaps',&
                     ' < Psi(t)_s|Psi1(t)_s1>,', 's and s1 denote ',&
                     'the electronic states of bra- and ket-WF.'
            write(6, '(2a, /11x,2a)')' -single i j : Compute the overlap',&
                     ' < Psi(t_i)|Psi1(t_j)>,', 'i and j denote ',&
                     'the positions of WFs in the of bra- and ket-WF files.'
            write(6,'(a)')&
                     ' -ver    : Version information about the program '
            write(6,'(a,/a,/)') &
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(2a/,2a/)') ' The argument psi_ref is the name of',&
                               ' the file containing the', &
             ' (time-independent) comparison wavefunction (density',&
             '  operator).' 
            write(6,'(2a,/2a,/2a)') ' With the aid of the option ',&
                  '-M, crosscorr computes overlaps', &
                  ' <psi_ref(t)|psi(t)>. If in addition -O is set, ', &
                  'matrix-elements', &
                  ' <psi_ref(t)|oper|psi(t)> are calculated.', &
                  ' Note that psi_ref is',' now taken as time-dependent.'
            write(6,'(/2a,/2a,/2a)') &
                 '-O, crosscorr can compute matrix elements',&
                 ' <psi_ref|oper|psi>.',' This, however, requires that ',&
                 'psi_ref and psi are MCTDH-WFs with ', &
                 'identical numbers of SPFs.' 
            write(6,'(/2a,/2a,/2a)') &
                 ' With the aid of the option -DT, which sets -M ad -HC',&
                 ' crosscorr',' computes c(t1+t2)=<psi^*_ref(t1)|psi(t2)>', &
                 ' where t1 and t2 are',' increased alternating. This is',&
                 ' similar to "auto=twice" in MCTDH input.'
            write(6,'(/2a,/a,/2a,/2a)') ' If no argument is given, ',&
                 'the user is prompted for the missing argument.',&
                 ' Note: The option "-o no" directs output to screen.',&
                 ' Note: If the argument ends with a slash "/",',&
                 ' the string "psi" or "restart"','       is appended.',&
                 ' If the argument ends with "restart", -r is set.'
            write(6,'(2a)') ' Note: crosscorr can overlap two MCTDH or ',&
                 'two ML-wavefunctions,',&
                 '       but not a mixture of the two types.'

            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      
      if (iarg .lt. i) goto 200
      call mygetarg(i,filein)

      if (iarg .gt. i) then
         write(6,'(/a)') 'Nothing must follow the argument (ref. WF)'
         write(6,'( a)') 'Place options before argument.'
         write(6,'( a)') '##########################################'
         stop
      endif


 200  continue
      if (filein .eq. ' ') then
         write(6,*) 'Name of reference WF file   :'
         read(5,'(a)') filein
         write(6,*)
      endif

! ---- reference WF ---
      laein=index(filein,' ')-1
      if (filein(laein-6:laein) .eq. 'restart' ) lrst = .true.
      if (filein(laein-2:laein) .eq. 'psi' )     lrst = .false.
      if (lrst) irst = ipsi
      if (filein(laein:laein)   .eq. '/') then 
         if(lrst) then
            filein = filein(1:laein)//'restart'
            laein  = laein + 7 
         else
            filein = filein(1:laein)//'psi'
            laein  = laein + 3
         endif
      end if
      call abspath(filein,laein)

! ---- time-dependent WF ---
      laein1=index(filein1,' ')-1
      if (filein1(laein1:laein1) .eq. '/') then 
         if( lrst1 ) then
            filein1 = filein1(1:laein1)//'restart'
            laein1  = laein1 + 7 
         else
            filein1 = filein1(1:laein1)//'psi'
            laein1  = laein1 + 3
         endif
      end if
      call abspath(filein1,laein1)

!-----------------------------------------------------------------------
! get stem of reference name
! If file name is psi, use directory name.
! Else use file name as extension for fileout.
!-----------------------------------------------------------------------
      llaenge=laein
      ilbl=llaenge
   11 continue
      if (filein(ilbl:ilbl) .eq. '/') then
         continue
      else
        ilbl=ilbl-1
        go to 11
      endif

      if (filein(ilbl+1:laein) .eq. 'psi' .or.&
          filein(ilbl+1:laein) .eq. 'restart' ) then
         llaenge = ilbl
         ilbl=ilbl-1
   12    continue
         if (filein(ilbl:ilbl) .eq. '/') then
            continue
         else
           ilbl=ilbl-1
           go to 12
         endif
      endif
      flabel=filein(ilbl+1:llaenge)
      llaenge=index(flabel,' ' )-1

      if (fileout(1:2) .eq. 'no') then
         laeout = 0
      else
         if (fileout .eq. ' ') then
            fileout = 'cross_'//flabel(1:llaenge)
         endif
         laeout=index(fileout,'  ')-1
         call abspath(fileout,laeout)
      end if


!-----------------------------------------------------------------------
! Opening the files.
! Time-dependent WF (ket) : filein1, psi1, save(2) .
! Reference WF (bra)      : filein , psi , save(1) .
!-----------------------------------------------------------------------
      write(6,'(80a)') ('*',i=1,80)
      filename=filein1(1:laein1)
      open(ipsi1,file=filename,form='unformatted',status='old',&
           err=1000)
      write(6,'(2a)')    ' The time-dep. WF is: ',filein1(1:laein1)

      filename=filein(1:laein)
      open(ipsi,file=filename,form='unformatted',status='old',&
           err=1000)
      write(6,'(2a)')    ' The reference WF is: ',filein(1:laein)

      if(laeout .eq. 0) then
         write(6,'(2a)') ' No outputfile '
      else
         filename=fileout(1:laeout)
         open (iaus,file=filename,form='formatted',status='unknown',&
              err=1000)
         write(6,'(2a)') ' The outputfile is  : ',fileout(1:laeout)

         if(.not.lshort)then
            write(iaus,'(2a)')    '# Cross-correlation of ',&
                                  '(bra, time-independent) '
            ilbl = max(1,laein-77)
            write(iaus,'(2a)')    '# ',filein(ilbl:laein)
            write(iaus,'(a)')     '# and (ket, time-dependent)'
            ilbl = max(1,laein1-77) 
            write(iaus,'(2a)')    '# ',filein1(ilbl:laein1)
            write(iaus,'(a)')     '# '
            if(sbra.gt.0) write(6,'(a,i3,a,/a,i3,a)')&
               ' Electronic state',sbra,' for Bra-WF. (-s option)',&
               ' Electronic state',sket,' for Ket-WF. (-s option)'
         endif
      end if
      dvrlaenge = 0
      opflaenge = 0
      if(loper) then
         opflaenge=index(operfile,' ')-1
         if (operfile(opflaenge:opflaenge) .eq. '/') then
            operfile = operfile(1:opflaenge)//'oper'
            opflaenge = opflaenge + 4
         end if
         call abspath(operfile,opflaenge)
         dvrlaenge=index(dvrfile,' ')-1
         if (dvrfile(dvrlaenge:dvrlaenge) .eq. '/') then
            dvrfile = dvrfile(1:dvrlaenge)//'dvr'
            dvrlaenge = dvrlaenge + 3
         end if
         call abspath(dvrfile,dvrlaenge)
         dlaenge=dvrlaenge
         dname=dvrfile(1:dlaenge)
         filename=dvrfile(1:dvrlaenge)
         open(idvr,file=filename,form='unformatted',&
              status='old',err=1000)
         filename=operfile(1:opflaenge)
         open(ioper,file=filename,form='unformatted',&
              status='old',err=1000)
         chkdvr=1
         call dvrinfo(lerr,chkdvr)
         if (lerr) goto 1010
         chkdvr=2
         chkgrd=1
         call operinfo(lerr,chkdvr,chkgrd)
         if (lerr) goto 1020
         if(laeout .gt. 0) then
            write(iaus,'(a)')     '# DVR and OPER files'
            write(iaus,'(2a)')    '# ',dvrfile(1:dvrlaenge)
            write(iaus,'(2a)')    '# ',operfile(1:opflaenge)
            write(iaus,'(a)')     '# '
         endif
      endif
      if(laeout .gt. 0 .and. .not. lshort) then
         write(iaus,'(2a)')    '# Options: ',options(1:optlaenge)
         write(iaus,'(a)')     '# '
      endif
      return

 1000 write(6,*)' crscinput. Cannot open file : '
      write(6,*) filename(1:slen(filename))
      write(6,*) ' Possible problem: FORTRAN cannot open a file twice.'
      write(6,*) ' Also try "ls '//filename(1:slen(filename))//'".'

      stop

 1010 write(6,'(2a)')&
          'Error reading dvr-info from:',dvrfile(1:dvrlaenge)

 1020 write(6,'(2a)')&
          'Error reading oper-info from:',operfile(1:opflaenge)
      stop
 
      end subroutine crscinput


!##############################################################################


      subroutine wrdover(m,s,dover,dim1,dim2)

      implicit none

      integer(long)                                  :: i,j
      integer(long), intent(in)                      :: m,s,dim1,dim2
      real(dop)                                      :: norm, sum
      complex(dop), dimension(dim1,dim2), intent(in) :: dover
      integer(long)                                  :: count = -1

      if(m.eq.1 .and. s.eq.1) then
         count = count + 1
         write(6,'(/a)')'----------------------------------------------'
      endif

      write(6,'(/a,i3,a,i2,a,i3/)') &
              'SPF-Overlap,  m =',m,' ,  s =', s,' ,  count =',count

      sum = 0.0_dop
      do i = 1, min(dim1,dim2)
         sum = sum + abs(dover(i,i))**2
      enddo
      call normvxz(dover,norm,dim1*dim2)
      write(6,'(3f12.6/)') norm**2, &
           min(dim1,dim2)-norm**2, min(dim1,dim2)-sum

      do i = 1, dim1
         write(6,'(15f9.5)') (dble(dover(i,j)),j=1,dim2)
      enddo

      return
      end subroutine wrdover

!***********************************************************************
!     reads DVR and oper
!***********************************************************************

      subroutine corroperin(mc,mr,mi)

      implicit none

      complex(dop), dimension(*), intent(inout)  :: mc
      real(dop), dimension(*), intent(inout)     :: mr
      integer(long), dimension(*), intent(inout) :: mi

      real(dop), dimension(1)       :: rdum,rdum1,rdum2,rdum3
      integer(long)                 :: chkdvr,chkgrd

!-----------------------------------------------------------------------
! Read DVR data needed to propagate wavefunction (see rddvr.f for
! details)
!-----------------------------------------------------------------------
      call zerovxl(dvrdata,12)
      dvrdata(1) = .true.    ! read ort
      dvrdata(2) = .false.   ! read trafo
      dvrdata(3) = .false.   ! read dvrmat (dif1mat, dif2mat)
      dvrdata(4) = .false.   ! read fftp   (fft momentum)
      dvrdata(5) = .true.    ! read hin (fft)
      dvrdata(6) = .true.    ! read rueck (fft)
      dvrdata(7) = .true.    ! read fftfak
      dvrdata(8) = .true.    ! read jsph
      dvrdata(9) = .true.    ! read msph
      dvrdata(10)= .false.   ! read kinsph
      dvrdata(11)= .true.    ! read exphin
      dvrdata(12)= .true.    ! read exprueck
      chkdvr=1
      call rddvr(mr(mrort),rdum,rdum1,rdum2,mc(mchin),mc(mcrueck),&
                 mi(mifftfak),mc(mcexphin),mc(mcexprueck),&
                 mi(mijsph),mi(mimsph),rdum3,chkdvr)

!-----------------------------------------------------------------------
! Read data needed by the operator
!-----------------------------------------------------------------------
      chkdvr=2
      chkgrd=1
      call rdoper(mr(mrhops),chkdvr,chkgrd)

      return
      end subroutine corroperin


!-----------------------------------------------------------------------


      end program crosscorr

