!***********************************************************************
!
!                        ANALYSE MODULE DDGEOM
!
! Reads the position expectation values from file "check".
! Converts expectation values to Cartesian coordinates and writes to
! file.
! Based on parts of rdcheck and ddtraj.
! 
! 01/16 GWR  
!
!
!***********************************************************************

!-----------------------------------------------------------------------
!> @brief 
!> Generates the expectation values for the geometries of a direct-dynamics simulation.
!>
!> @detail
!> ddgeom computes the expectation values for the geometries of a direct-dynamics calculation.
!>
!> It can also analyse distances, angles and dihedrals for the expectation values of the geometries of a direct-dynamics calculation.
!>
!> OPTIONS:
!>---------
!> 
!>
!>  -f FILE : The data is read from file FILE rather than from ./check 
!>            The string FILE may be a relative or a full path-name.
!>
!>  -o FILE : The geometries are written to file FILE.pl rather than to ./ddgeom.xyz
!>            The string FILE may be a relative or a full path-name.
!>
!>  -dist i1 i2 : 
!>            The distance between the two atoms i1 and i2 is calculated at each timestep
!>            and written to ddgeom_i1i2. i1 and i2 are integers following -dist
!>
!>  -angle i1 i2 i3 : 
!>            The angle between the three atoms i1, i2 and i3 is calculated at each timestep
!>            and written to ddgeom_i1i2i3. i1, i2 and i3 are integers following -angle
!>
!>  -dihedral i1 i2 i3 i4 :
!>            The dihedral angel between i1, i2, i3 and i4 is calculated at each timestep
!>            and written to ddgeom_i1i2i3i4. i1, i2, i3 and i4 are integers following -dihedral
!>
!>  -ver    : Version information about the program.
!>
!>  -h      : Print this help text.
!>
!>  -?      : Print this help text.
!-------------------------------------------------------------------------

      program ddgeom

use decimal, only: long,dop
use global
use aglobal
use versions
use maxv, only: maxsta
use keyunitsmod
use dvrdatmod
use griddatmod
use psidef
use datenmod
use memdimmod
use basismod
use dirdyn
use compdatmod
use compdat1mod

use op1lib, only: zeromxd,cpvxd
use iofile, only: rdchkinfo
use rdfilesmod, only: rdchk,lrdchk
use channels, only: filever, iaus, ioper, idump, iexpect, iexpect1
use gwplib, only: gwp2gpoint
use dd_db, only: ddq2x
use readqc, only: getinam,getatnam

use module_interface

      implicit none

      integer(long)       :: s,i,m,n,f,k,ns,nmd,i6,ilbl,ll,check,ilddtrans,&
                             j,jlbl,gpsize,qrep
      real(dop)                                       :: time,dt,norm,trace,&
                                                         qexp,rdum
      real(dop), dimension(:), allocatable            :: spop,ecorr,etot
      real(dop), dimension(:,:,:), allocatable        :: qdq
      real(dop), dimension(:), allocatable            :: dicht3
      real(dop), dimension(1)                         :: ggp
      logical(kind=4)                                 :: lcheck
      character(len=c5)                               :: mdlbl,filename
      character(len=c2)                               :: rdtag
      character(len=80)                               :: nspf
      integer(long)     :: ierr
      character(len=10) :: cerr
      logical(kind=4)                      :: int2log
      integer(long),dimension(2)           :: atomsDist 
      integer(long),dimension(3)           :: atomsAngle
      integer(long),dimension(4)           :: atomsDihedral
      real(dop), dimension(:), allocatable :: xcenter
      logical(kind=4)                      :: lanalyze,lgauss

!-----------------------------------------------------------------------
!     Set defaults. Set up the input.
!-----------------------------------------------------------------------

      write(6,'(/,78a1,/)')  ('*',i=1,78)
      write(6,'(2a)') ' Analyse program "ddgeom" ',&
                      '(Read check file)'
      call default
      call adefault
      call setunits
      call chkinput(ns,nmd,dt,i6,qrep,atomsDist,atomsAngle,atomsDihedral,&
           lanalyze,lgauss)

      laein  = index(filein, ' ')-7
      laeout = index(fileout,' ')-1

      call rdmemdim(20)

      call alloc_dvrdat
      call alloc_grddat
      call alloc_psidef
      call alloc_dirdyn
      call alloc_dddb

!-----------------------------------------------------------------------
!  Read Check file info and print basis set information.
!-----------------------------------------------------------------------
      check = 1
      call rdchkinfo(20,check,check,check,check)
      tinit = tinit/fs
      out1  = out1/fs

!-----------------------------------------------------------------------
!  Check for legality of the arguments
!-----------------------------------------------------------------------
      if(ns .gt. nstate ) then
         write(6,'(/a,/a,i9,/a,i3,/a)')&
               '######################################################'&
              ,' Illegal value of ns =', ns&
              ,' ns must not be greater than nstate =',nstate    &
              ,'######################################################'
         stop 2
      end if
      if(nmd .gt. nmode ) then
         write(6,'(/a,/a,i9,/a,i3,/a)')&
               '######################################################'&
              ,' Illegal value of nmd =', nmd&
              ,' nmd must not be greater than nmode =',nmode    &
              ,'######################################################'
         stop 3
      end if

      if (i6 .ne. 6) then

         write(6,'(78a1)')  ('-',i=1,78)
         write(6,'(a,i4)') ' Number of degrees of freedom:', ndof       
         write(6,'(a,i4)') ' Number of combined modes    :', nmode
         write(6,'(a,i4)') ' Number of electronic states :', nstate

         write(6,'(/,a)') ' mode dof  dim   DVR     modelabel'
         do m=1,nmode
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               write(6,'(1x,i3,1x,i4,i5,3x,a8,a10)') &
               m, f, gdim(f), cbas(basis(f)), modelabel(f)  
            enddo
         enddo

!-----------------------------------------------------------------------
! write mode information.
!-----------------------------------------------------------------------
         write(6,'(/,a)') ' mode: No. of spf''s per state : modelabels'
         do m=1,nmode
            ll = 1
            mdlbl = '                                                    '
            nspf  = '                                                    '
            do  n=1,nspfdof(m)
               ilbl  = index(modelabel(spfdof(n,m)),' ')
               mdlbl = mdlbl(1:ll)//modelabel(spfdof(n,m))(1:ilbl)
               ll    = ll + ilbl
            end do
            write(nspf,'(95i3)') (dim(m,s),s=1,nstate)
            if(3*nstate+ll .le. 72 ) then
               write(6,'(i4,4a)') m,' :',nspf(1:3*nstate),' : ',&
                                  mdlbl(1:ll)
            else
               write(6,'(i4,3a)') m,' :',nspf(1:3*nstate),' :'
               write(6,'(a,/)') mdlbl(1:ll)
            end if
         enddo

         write(6,'(/,a,f11.3,a,f11.3,a)') &
             ' tinit =', tinit, 'fs,    tout =', out1, 'fs' 
      endif  ! (i6.ne.6)

!-----------------------------------------------------------------------
! get ndofddpes + natm from oper file
!-----------------------------------------------------------------------
      filename=operfile
      open(ioper,file=filename,form='unformatted',status='old',&
           err=1000)
      rewind(ioper)

      do
   10    continue
         read(ioper,err=10,end=20) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'ndofddpes') then
            read(ioper) ndofddpes
         else if (rdtag(ilbl:jlbl) .eq. 'ndofdd') then
            read(ioper) ndofdd
         else if (rdtag(ilbl:jlbl) .eq. 'natm') then
            read(ioper) natm
         else if (rdtag(ilbl:jlbl) .eq. 'xatnam') then
            read(ioper) (xatnam(i),i=1,natm)
         else if (rdtag(ilbl:jlbl) .eq. 'lddtrans') then
            read(ioper) ilddtrans
            lddtrans=int2log(ilddtrans)
         else if (rdtag(ilbl:jlbl) .eq. 'ddxcoo0') then
            read(ioper) (ddxcoo0(i),i=1,ndofddpes)
         else if (rdtag(ilbl:jlbl) .eq. 'ddtransf') then
            read(ioper) ((ddtransf(j,i),j=1,ndofddpes),i=1,ndofdd)
         else if (rdtag(ilbl:jlbl) .eq. 'ddtransb') then
            read(ioper) ((ddtransb(j,i),j=1,ndofdd),i=1,ndofddpes)
         else if (rdtag(ilbl:jlbl) .eq. 'end oper info') then
            exit
         endif
      enddo
   20 continue
      close(ioper)

      cerr = 'spop'
      allocate(spop(nddstate),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ecorr'
      allocate(ecorr(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'etot'
      allocate(etot(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qdq'
      allocate(qdq(4,ndof,nstate),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dicht'
      allocate(dicht3(d3matdim),stat=ierr)
      if (ierr > 0) go to 1000

      spop=0.0_dop
      ecorr=0.0_dop
      etot=0.0_dop
      qdq=0.0_dop
      dicht3=0.0_dop

!
!     read the DATA section of the check file
!
      rewind(20)
      if (filever(20) .lt. 11.0_dop) then
         read(20) filever(20)
         check = 1
         call rdchkinfo(20,check,check,check,check)
      endif
      lend=.false.
      k=0
      lrdchk(1)=.true.
      lrdchk(2)=.true.
      lrdchk(3)=.true.

      gpsize=max(ndofdd,ndofddpes)
      allocate(xcenter(gpsize))

      if(lfileexp)then
         rewind(iexpect)
         do i = 1,7
            read(iexpect,*)
         enddo
      endif

      do 
      
         k=k+1
         time = tinit +(k-1)*out1

         call rdchk(20,ecorr,etot,spop,dicht3,ggp,qdq,norm,trace)

         if(lpntdvr)then
            do f=1,ndofdd
               if (basis(f) .eq. 19) qdq(1,f,1:ns) = rpbaspar(1,f)
            enddo
         endif

         if (lend) exit

         if(lfileexp)then
            read(iexpect,*)rdum,qexp
            qdq(1,qrep,ns) = qexp
         endif

         call plddgeom(qdq(1,1:ndofdd,ns),time,xcenter,lgauss)
         if (lanalyze) call DistAngleDihedral(xcenter,atomsDist,&
                            atomsAngle,atomsDihedral,time)

      end do                    ! of k

      deallocate(spop)
      deallocate(ecorr)
      deallocate(etot)
      deallocate(qdq)
      deallocate(dicht3)

! --- close all open files
      inquire(20,opened=lcheck)
      if (lcheck) close(20)
      inquire(iaus,opened=lcheck)
      if (lcheck) close(iaus)
      inquire(idump,opened=lcheck)
      if (lcheck) close(idump)
      inquire(iexpect,opened=lcheck)
      if (lcheck) close(iexpect)
      inquire(iexpect1,opened=lcheck)
      if (lcheck) close(iexpect1)
     
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_psidef
      call dealloc_dirdyn
      call dealloc_dddb

      stop

 1000 write(ilog,'(a)') 'Error allocating memory in Ddgeom :',cerr
      write(ilog,'(a)') 'Error Code :',ierr
      stop

      contains

!#######################################################################
!>----------------------------------------------------------------------
!> @brief
!> subroutine to generate output, which can be plotted for example with jmol
!>
!>----------------------------------------------------------------------
      subroutine plddgeom(gpoint,time,xcenter,lgauss)

      implicit none

      logical(kind=4), intent(in)                  :: lgauss
      integer(long)                                :: n,i,izero,gpsize
      integer(long), dimension(natm)               :: iatnam,iatiso
      real(dop), dimension(ndofddpes), intent(inout) :: xcenter
      real(dop), intent(in)                        :: time
      real(dop), dimension(ndofdd), intent(in)     :: gpoint
      character(len=c1), dimension(natm)           :: catnam

! gpoint must be ndof for gwp2gpoint, but ndof < ndofddpes with frozen modes,
! otherwise ndof >= ndofddpes 

      izero=0
      catnam(1:natm) = xatnam(1:natm)
      do n = 1, natm
         call getatnam(catnam(n),ilog)   ! ensure atom names are characters
      enddo
      call getinam(catnam(n),iatnam(n),iatiso(n),ilog)

!     Transform
      if (lddtrans) then
         call ddq2x(gpoint,xcenter)
      else
!         call cpvxd(gpoint,xcenter,ndofddpes)
         xcenter = gpoint + ddxcoo0
      endif

! convert to Angst
      do n=1,ndofddpes
         xcenter(n)=xcenter(n)*0.52917725_dop
      enddo

      if(lgauss)then
         write(iaus,'(a,f8.2)') 'Time: ',time
         
         write(iaus,'(/,a)') ' Entering Gaussian System  '

         write(iaus,'(/,a)')&
              '                          Input orientation:   '
         write(iaus,'(2a)')&
              ' ----------------------------------------------',&
              '-----------------------'
         write(iaus,'(2a)')&
              ' Center     Atomic     Atomic             ',&
              ' Coordinates (Angstroms)'
         write(iaus,'(2a)')&
              ' Number     Number      Type              ',&
              'X           Y           Z '
         write(iaus,'(2a)')&
              ' ----------------------------------------------',&
              '-----------------------'
         
         i=1
         do n=1,natm
            write(iaus,'(i5,i11,i14,4x,3f12.6)')&
                 n,iatnam(n),izero,xcenter(i),xcenter(i+1),&
                 xcenter(i+2)
            i=i+3
         enddo
         write(iaus,'(2a)')&
              ' ----------------------------------------------',&
              '-----------------------'
         
         write(iaus,*)
      else
         write(iaus,'(i6)')natm
         write(iaus,*)
         i=1
         do n=1,natm
            write(iaus,'(a5,3f14.8)')&
                 catnam(n)(1:5),xcenter(i),xcenter(i+1),&
                 xcenter(i+2)
            i=i+3
         enddo       
      endif

      return
      end subroutine plddgeom

!-----------------------------------------------------------------
!*****************************************************************
!-----------------------------------------------------------------

!>---------------------------------------------------------------
!> @brief 
!> subroutine that is called if distance, angle or dihedral angle shall be calculated
!>
!>--------------------------------------------------------------

      subroutine DistAngleDihedral(xcenter,atomsDist,atomsAngle,atomsDihedral,time)

      implicit none

      real(dop), dimension(*), intent(in)          :: xcenter
      real(dop), intent(in)                        :: time
      integer(long),dimension(2),intent(in)        :: atomsDist
      integer(long),dimension(3),intent(in)        :: atomsAngle
      integer(long),dimension(4),intent(in)        :: atomsDihedral
      real(dop)                                    :: dist, dihedral, x, y
      integer(long)                                :: i,j
      character(len=120)                           :: strout, strout2
      real(dop),dimension(3)                       :: vec1, vec2, vec3, vec4, vec5, vec6

      strout2=''

      vec1=0.d0
      vec2=0.d0
      vec3=0.d0
      vec4=0.d0
      vec5=0.d0
      dist=0.d0

      if(atomsDist(1)/=0) then
         if(atomsDist(2)/=0) then
            dist=0.d0
            do i=1,3
               dist=dist+(xcenter((atomsDist(2)-1)*3+i)-xcenter((atomsDist(1)-1)*3+i))**2
            end do
            dist=sqrt(dist)
            write(idump,'(2F10.5)') time, dist
          endif
      endif

      if(atomsAngle(1)/=0) then

         do i=1,3
            vec1(i)=xcenter((atomsAngle(1)-1)*3+i)-xcenter((atomsAngle(2)-1)*3+i)
            vec2(i)=xcenter((atomsAngle(3)-1)*3+i)-xcenter((atomsAngle(2)-1)*3+i)
         end do
             
         write(iexpect,'(2F10.5)')  time, angle(vec1,vec2)
             
      endif


      if(atomsDihedral(1)/=0) then
         do i=1,3
            vec3(i)=xcenter((atomsDihedral(2)-1)*3+i)-xcenter((atomsDihedral(1)-1)*3+i)
            vec4(i)=xcenter((atomsDihedral(3)-1)*3+i)-xcenter((atomsDihedral(2)-1)*3+i)
            vec5(i)=xcenter((atomsDihedral(4)-1)*3+i)-xcenter((atomsDihedral(3)-1)*3+i)
         end do
         vec1=cross(vec3,vec4)
         vec1=vec1/length(vec1)
         vec2=cross(vec4,vec5)
         vec2=vec2/length(vec2)
         vec6=cross(vec1,vec4/length(vec4))
         x=dotProd(vec1,vec2)
         y=dotProd(vec6,vec2)
         dihedral=-atan2(y,x)*180.d0/pi
         write(iexpect1,'(F10.5,F15.5)') time, dihedral 
      endif


      end subroutine DistAngleDihedral


!> @brief
!> calculate dot product of two vectors
      FUNCTION dotProd(vec1,vec2)

      implicit none

      real(dop), DIMENSION(3),intent(in) :: vec1, vec2
      real(dop)                          :: dotProd
      integer(long)                      :: i
            
      dotProd=0.d0

      do i=1,3
         dotProd=dotProd+vec1(i)*vec2(i)
      end do
      return

      END FUNCTION dotProd

!> @brief
!> calculate cross product between two vectors
      FUNCTION cross(a, b)

      implicit none

      real(dop), DIMENSION(3)             :: cross
      real(dop), DIMENSION(3), INTENT(IN) :: a, b
      cross(1) = a(2) * b(3) - a(3) * b(2)
      cross(2) = a(3) * b(1) - a(1) * b(3)
      cross(3) = a(1) * b(2) - a(2) * b(1)
      return

      END FUNCTION cross

!> @brief
!> calculate angle between two vectors
      FUNCTION angle(vec1, vec2)

      implicit none

      real(dop)                          :: angle
      real(dop), DIMENSION(3),intent(in) :: vec1,vec2
      integer(long)                      :: i
      angle=0.d0
      do i=1,3
         angle=angle+vec1(i)*vec2(i)
      end do
      angle=angle/length(vec1)/length(vec2)
      angle = acos(angle)*180.d0/pi

      return

      END FUNCTION angle

!> @brief
!> calculate length of a vector
      FUNCTION length(vec1)

      implicit none

      real(dop)                          :: length
      real(dop),DIMENSION(3), intent(in) :: vec1
      integer(long)                      :: i

      length=0.0d0

      do i=1,3
         length=length+vec1(i)**2
      end do
      length=sqrt(length)
      return

      END FUNCTION length

!>----------------------------------------------------------------------
!> @brief
!> subroutine to evaluate input of ddgeom
!>
!>----------------------------------------------------------------------
      subroutine chkinput(ns,nmd,dt,i6,qrep,atomsDist,atomsAngle,&
                 atomsDihedral,lanalyze,lgauss) 

      implicit none

      integer(long), external        :: myiargc
      integer(long)                  :: iarg,ierr,i,j
      integer(long), intent(out)     :: nmd,ns,i6,qrep
      real(dop), intent(out)         :: dt
      character(len=80)              :: buf
      integer(long),dimension(2)     :: atomsDist
      integer(long),dimension(3)     :: atomsAngle
      integer(long),dimension(4)     :: atomsDihedral
      logical, intent(out)           :: lanalyze,lgauss
      character(len=120)             :: strout
      character(len=120)             :: strout2

!-----------------------------------------------------------------------

      atomsDist=0
      atomsAngle=0
      atomsDihedral=0
      lanalyze = .false.
      lgauss = .false.

      name ='./'
      laenge=1
      filein  = '  '
      fileout = '  '
      fileexp = '  '
      ns  = -1
      nmd = -1
      dt  = -1.0_dop
      i6=6
      open(3,file='/dev/null',form='formatted',status='unknown')

      iarg = myiargc() 
      if (iarg .lt. 1) goto 200 

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf .eq. '-f') then
            i=i+1
            call mygetarg(i,buf)
            filein = buf
         else if(buf .eq. '-i') then
            i=i+1
            call mygetarg(i,buf)
            name = buf
         else if( buf(1:3) .eq. '-s ' ) then
            i6=3
         else if( buf(1:3) .eq. '-n ' ) then
            fs = 1.0_dop
         else if( buf(1:5) .eq. '-nofs' ) then
            fs = 1.d0
         else if( buf(1:5) .eq. '-noev' ) then
            ev = 1.d0
         else if( buf(1:5) .eq. '-g' ) then
            lgauss = .true.
         else if( buf(1:2) .eq. '-o ' ) then
            i=i+1
            call mygetarg(i,buf)
            fileout = buf
         else if( buf(1:5) .eq. '-dist' ) then
            do j=1, 2
                i=i+1
                call mygetarg(i,buf)
                read(buf,'(I3)',err=94) atomsDist(j)
            enddo
            lanalyze=.true.
         else if( buf(1:6) .eq. '-angle' ) then
            do j=1, 3
                i=i+1
                call mygetarg(i,buf)
                read(buf,'(I3)',err=95) atomsAngle(j)
            enddo
            lanalyze=.true.
         else if( buf(1:9) .eq. '-dihedral' ) then
            do j=1, 4
                i=i+1
                call mygetarg(i,buf)
                read(buf,'(I3)',err=96) atomsDihedral(j)
            enddo
            lanalyze=.true.
         else if( buf(1:4) .eq. '-exp' ) then
            i=i+1
            call mygetarg(i,buf)
            fileexp = buf
            lfileexp = .true.
            i=i+1
            call mygetarg(i,buf)
            read(buf,*) qrep
         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(/,78a1,/)')  ('*',i=1,78)
            write(6,'(a)') ' Analyse program "ddgeom" ' 
            write(6,'(1x,a/,11x,a)')&
       'Purpose : Reads the check-file.', &
       'Converts position expectation values to molecular geometries.'
            write(6,'(1x,2a)') 'Usage   : ddgeom ',&
       '[-f -oq -g -i -n -ver -h -?] ns nmd'
            write(6,'(/,a)') ' Options : '
            write(6,'(2a,/,9x,a)') &
            ' -f FILE : The data is read '&
           ,'from file FILE rather than from ./check'&
           ,'  The string FILE may be a relative or full path-name.'
            write(6,'(a)') ' -i DIR  : Data is read from file DIR/check'
            write(6,'(2a,/,9x,a/,9x,a/,9x,a)') &
            ' -oq FILE: The geometries ',&
            ' are written to file FILE.xyz','  rather than to ./ddgeom.xyz',&
            '  The string FILE may be a relative or full path-name.'
            write(6,'(2a)')&
            ' -g      : Requests Gaussian output type xyz.'
            write(6,'(2a)')&
            ' -s      : Suppress informative-only part of output',&
            ' (short output).'
            write(6,'(2a)') " -nofs   : No time units.", &
            " Sets fs=1. Use, if 'time-not-fs' is set."
            write(6,'(a)')  " -n      : Same as -nofs"
            write(6,'(a)')&
            ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
            ' -h      : Print this help text. ',&
            ' -?      : Print this help text. '
            write(6,'(/a)') ' There are two arguments: '
            write(6,'(2a,/a)') ' The first argument is ns, i.e. the ',&
                 'state for which the natpops are written.', &
                 ' (ns=0 -> no output files written).' 
            write(6,'(2a)') ' The second argument is nmd, ',&
                 'i.e. the runing number of the mode ',&
                 ' for which the natpops are written.' 
            write(6,'(2a)') ' If not enough arguments are given, ',&
                 'the user is prompted for the',' missing arguments.'  
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      
      read(buf,*,iostat=ierr) ns
      if (ierr .ne. 0) then
         write(6,'(a/,1x,a72)') &
        ' The argument (ntype)  must be int number. Not :',buf
         stop
      end if
   
      i=i+1
      if (iarg .lt. i) goto 200
      call mygetarg(i,buf)
      read(buf,*,iostat=ierr) nmd
      if (ierr .ne. 0) then
       write(6,'(1x,a/,a)')&
       'The second argument (nmd)  must be int number. Not :', buf
       stop
      end if

!-----------------------------------------------------------------------
!  Prompting for missing input
!-----------------------------------------------------------------------

 200  continue
      write(i6,'(/,78a1,/)')  ('*',i=1,78)
      write(i6,'(a)') ' Analyse program "ddgeom" (Read check file)'
      write(i6,'(78a1)')  ('-',i=1,78)

      if(ns .le. 0) then
         write(6,'(a)')&
              ' state =?' 
         read(5,*) ns
      endif

      if(ns.le.0)then
         write(i6,'(a)')'State entered =< than 0'
         stop
      endif

!-----------------------------------------------------------------------
! If filein has not been set, look in name directory.
! Otherwise take name from filein
!-----------------------------------------------------------------------
      if (filein .eq. ' ') then
         laenge=index(name,' ')-1
         call abspath(name,laenge)
         filein=name(1:laenge)//'/check'
         laein=index(filein,' ')-1
      else
         laein=index(filein,' ')-1
         if (filein(laein:laein) .eq. '/') then 
            filein = filein(1:laein)//'check'
            laein = laein + 5
         end if
         call abspath(filein,laein)

         i=laein
   10    continue
         if (filein(i:i) .eq. '/') then
            laenge=i-1
         else
            i=i-1
            go to 10
         endif
         name=filein(1:laenge)
      endif

!-----------------------------------------------------------------------
!  Opening input file.
!-----------------------------------------------------------------------
      open(20,file=filein(1:laein),form='unformatted',&
              status='old',err=97)
      if(laein .le. 58 ) then
         write(i6,'(2a)') ' The inputfile is  : ',filein(1:laein)
      else
         write(i6,'(a)')  ' The inputfile is  :' 
         write(i6,'(a)')   filein(1:laein)
      end if

      read(20) filever(20)
         
!-----------------------------------------------------------------------
!  Opening output files.
!-----------------------------------------------------------------------
      if(fileout(1:2) .eq. '  ') &
           write(fileout,'(2a,i4,a)')name(1:laenge),'/ddgeom-st',ns,'.xyz'
!           fileout=name(1:laenge)//'/ddgeom.xyz'
      call rmblnk(fileout)

      laeout=index(fileout,' ')-1
      if (fileout(laeout:laeout) .eq. '/') then 
         write(fileout,'(2a,i4,a)')name(1:laeout),'/ddgeom-st',ns,'.xyz'
         call rmblnk(fileout)
         laeout=index(fileout,' ')-1
!         fileout = fileout(1:laeout)//'ddgeom.xyz'
!         laeout = laeout + 6
      else if (fileout(laeout-3:laeout) .ne. '.xyz') then 
         fileout = fileout(1:laeout)//'.xyz '
         laeout = laeout + 3
      end if
      call abspath(fileout,laeout)
      open(iaus,file=fileout(1:laeout),form='formatted',&
           status='unknown',err=98)
      write(i6,'(2a)') ' The outputfile is : ',fileout(1:laeout)
     
      if(atomsDist(1)/=0) then
         strout=''
         write(strout,'(a,a,I0,I0)') fileout(1:laeout-3),'_',atomsDist(1),atomsDist(2)
         open(idump,file=trim(strout),form='formatted',&
         status='unknown',err=99)
         write(i6,'(2a)') ' The outputfile for the distance is: ', trim(strout)
         strout='#Outputs of ddgeom'
         write(idump,'(a)') trim(strout)

         write(strout,'(a,I0,a,I0,a)') '#distance between atoms ', atomsDist(1),&
             ' and ', atomsDist(2), ' in Angstrom'
         write(idump,'(a)') trim(strout)
         write(idump,'(a)') '#time   dist'
      end if

      if(atomsAngle(1)/=0) then

         write(strout,'(a,a,I0,I0,I0)') fileout(1:laeout-3),'_',atomsAngle(1),&
             atomsAngle(2),atomsAngle(3)
         open(iexpect,file=trim(strout),form='formatted',&
         status='unknown',err=99)
         write(i6,'(2a)') ' The outputfile for the angle is: ', trim(strout)
         write(iexpect,'(a)') '#Outputs of ddgeom'
         write(strout,'(a,I0,a,I0,a,I0,a)') '#angle between atoms ', atomsAngle(1),&
             ', ', atomsAngle(2), ' and ', atomsAngle(3), ' in degrees'
         write(iexpect,'(a)') trim(strout)
         write(iexpect,'(a)') '#time   angle'

       end if

       if(atomsDihedral(1)/=0) then
         
         write(strout,'(a,a,I0,I0,I0,I0)') fileout(1:laeout-3),'_',atomsDihedral(1),&
             atomsDihedral(2),atomsDihedral(3),atomsDihedral(4)
         open(iexpect1,file=trim(strout),form='formatted',&
            status='unknown',err=99)
         write(i6,'(2a)') ' The outputfile for the dihedral is: ', trim(strout)
         write(iexpect1,'(a)') '#Outputs of ddgeom'
         write(strout,'(a,I0,a,I0,a,I0,a,I0,a)') '#dihedral between atoms ', atomsDihedral(1),&
             ', ', atomsDihedral(2), ', ', atomsDihedral(3), ' and ', atomsDihedral(4), ' in degrees'
         write(iexpect1,'(a)') strout
         write(iexpect1,'(a)') '#time   dihedral'
      end if

!-----------------------------------------------------------------------
!  Opening expectation file if requested.
!-----------------------------------------------------------------------
      if(lfileexp)then
         laeexp = index(fileexp,' ')-1
         open(iexpect,file=fileexp(1:laeexp),form='formatted',&
              status='old',err=93)
         if(laein .le. 55 ) then
            write(i6,'(2a)') ' The expect file is  : ',fileexp(1:laeexp)
         else
            write(i6,'(a)')  ' The expect file is  :' 
            write(i6,'(a)')   fileexp(1:laeexp)
         end if
      endif
         
      return 
      
 93   write(6,*)' Cannot open expectation (chk.pl) file : '
      write(6,*) fileexp
      stop

 94   write(6,*) ' For the calculation of a distance, please provide 2 integers separated by a space'
      stop

 95   write(6,*) ' For the calculation of an angle, please provide 3 integers separated by a space'
      stop

 96   write(6,*) ' For the calculation of a dihedral, please provide 4 integers separated by a space'
      stop

 97   write(6,*)' Cannot open input (check) file : '
      write(6,*) filein
      stop
      
 98   write(6,*)' Cannot open output (chk.pl) file : '
      write(6,*) fileout
      stop
 
 99   write(6,*) ' Cannot open output file for distances, angle or dihedral'
      stop

      end subroutine chkinput

      end program ddgeom


