!***********************************************************************
!
!             Analyse program DENGEN
!
! This program reads the psi-file and calculates a 2D density
!
!     03/05 GW   02/06 HDM
!
! This program is a cut-down version of showsys, and is not general.
! It writes the 2D-densities to an ASCII file, the data of which
! may be plotted with GNUPLOT. 
!
!***********************************************************************

      module dengenmod
 
use decimal, only:long
use versions

      integer(long), save     :: f1,f2
      character(len=16), save :: f1lab,f2lab

      end module dengenmod

      program dengen

use decimal, only:long,dop,sip
use global
use aglobal
use globmemmod
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod
use plot
use channels

use dengenmod
use basismod
use op1lib, only: initvxi, zerovxd
use logdat, only: errormsg
use iorst, only: rstinfo
use iofile, only: rdpsiinfo
use rddvrmod, only: rddvr,dvrinfo
use rdopermod, only: operinfo
use gwplib, only: gwp_ovl, gwp_param
use zeigauswmod, only: zeigausw
use plotutils, only: calcd2d_1ml, calcd2d_2ml
use inputmlmod
use logdat

use module_interface

      implicit none

      integer(long)                        :: chkdvr,chkgrd,chkpsi,chkdat,ilbl,i
      character(len=c5)                    :: filename
      logical(kind=4)                      :: lerr,linwf,lexist

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file. This is set in dengeninput
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.false.

!-----------------------------------------------------------------------
! read input options
!-----------------------------------------------------------------------
      call dengeninput

!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      open (unit=ilog,file=name(1:laenge)//'/dengen.log',form='formatted')
      logisopen=.true.
      write(6,   '(80a1)') ('-',i=1,80) 
      write(ilog,'(80a1)') ('-',i=1,80)
      call fpstartup(filename)
      write(ilog,'(a80)') filename 

!-----------------------------------------------------------------------
!  Opening the PSI file.
!-----------------------------------------------------------------------
      if (lshowrst) then
         filename=psifile
         open(irst,file=filename,form='unformatted',status='old',&
              err=1000)
         ilbl=index(psifile,' ')-1
         write(ilog,'(2a)') ' The restart file is : '//psifile(1:ilbl)
      elseif (lshowpsi) then
         filename=psifile
         open(ipsi,file=filename,form='unformatted',status='old',&
              err=1000)
         ilbl=index(psifile,' ')-1
         write(ilog,'(2a)') ' The psi file is : '//psifile(1:ilbl)
      endif

!-----------------------------------------------------------------------
! get memory sizes
!-----------------------------------------------------------------------
      if (lshowrst) then
         call rdmemdim(irst)
      else
         call rdmemdim(ipsi)
      endif

!-----------------------------------------------------------------------
! Allocate mememory
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_psidef
      call alloc_plot

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
! lfft   : FFT arrays rueck and fftfak
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.false.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.true.
      lfft=lpop2

!-----------------------------------------------------------------------
! read system information from DVR file
!-----------------------------------------------------------------------
      chkdvr=1
      filename=dname(1:dlaenge)//'/dvr'
      ilbl=index(filename,' ')-1
      open(idvr,file=filename,form='unformatted',status='old',err=1000)
      write(ilog,'(2a)') ' The DVR file is : ',filename(1:ilbl)
      call dvrinfo(lerr,chkdvr)
      close(idvr)
      call chksyserr(filename,lerr)
      if (lerr) go to 999

!-----------------------------------------------------------------------
! check arguments (i.e. modes chosen correctly)
!-----------------------------------------------------------------------
      if(f1.eq.0) then
         do i =1,ndof
            if(f1lab.eq.modelabel(i)) f1=i
         enddo
      endif
      if(f1.eq.0) then
         write(6,'(2a)') ' Unknown modelabel: ',f1lab
         stop
      endif
      if(f2.eq.0) then
         do i =1,ndof
            if(f2lab.eq.modelabel(i)) f2=i
         enddo
      endif
      if(f2.eq.0) then
         write(6,'(2a)') ' Unknown modelabel: ',f2lab
         stop
      endif

      if( f1.gt.ndof  .or.  f2.gt.ndof ) then
         write(6,'(a,/a,i4,a,i4,a,i4)') &
             'f1 and f2 must not be larger than ndof',&
             'f1 =', f1,',  f2 =', f2,',  ndof =', ndof
         stop
      endif
      write(6,'(a,i3,3x,a8,1x,a8,i4)')&
          ' First  DOF =',f1,modelabel(f1),cbas(basis(f1)),gdim(f1)
      write(6,'(a,i3,3x,a8,1x,a8,i4)')&
          ' Second DOF =',f2,modelabel(f2),cbas(basis(f2)),gdim(f2)
      if(basis(f1)*basis(f2).eq.0) then
         write(6,'(a)') &
           ' Neither f1 nor f2 must refer to the electronic basis!' 
         stop
      endif
      if(f1.eq.f2) then
         write(6,'(a)') ' f1 und f2 must be different!'
         stop
      endif

!-----------------------------------------------------------------------
!  Opening the output file.
!-----------------------------------------------------------------------
      filename=fileout(1:laeout)//'_'//modelabel(f1)//'_'//modelabel(f2)
      call rmblnk(filename)
      fileout=filename
      laeout=index(fileout,' ')-1
      write(6,'(2a)') ' The output file is : '//fileout(1:laeout)
      write(ilog,'(2a)') ' The out file is : '//fileout(1:laeout)
      if(.not.lovwr) then
         inquire(file=fileout(1:laeout),exist=lexist)
         if(lexist) then
            write(6,'(a)') &
            '### Output file already exist. Use -w to overwrite.'
            write(ilog,'(a)')& 
            '### Output file already exist. Use -w to overwrite.'
            stop
         endif
      endif
      open(iaus,file=filename,form='formatted',status='unknown',&
           err=1000)

      write(ilog,'(a,i3,2x,a8,2x,a8,i4)')&
          ' First  DOF =',f1,modelabel(f1),cbas(basis(f1)),gdim(f1)
      write(ilog,'(a,i3,2x,a8,2x,a8,i4)')&
          ' Second DOF =',f2,modelabel(f2),cbas(basis(f2)),gdim(f2)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! open input and output data files
!
! use standard channels if possible
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------
! psifile is opened in dengeninput if required
!     filename='psi'
!     open(ipsi,file=filename,form='unformatted',status='old',err=1000)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      if (lshowrst) then
         filename='restart'
         read(irst,err=999) filever(irst)
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkdat)
      elseif (lshowpsi) then
         filename='psi'
         read(ipsi,err=999) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      endif
      if(psitype.eq.5) then
         routine='dengen'
         message='Sorry, dengen cannot work with ML-wavefunctions.'
         call errormsg
      endif

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

!-----------------------------------------------------------------------
! here assign any further memory arrays
!-----------------------------------------------------------------------
      mrweight=mrdim
      mrdim=mrweight+ortdim
      mrpldata=1
      mrhops=1
      mrpldata1=mrdim
      mrdim=mrpldata1+maxgdim2*nstate

!-----------------------------------------------------------------------
! assign work arrays
!-----------------------------------------------------------------------
      workcdim=max(workcdim,dgldim+maxspf*2)
      workcdim=max(workcdim,2*maxblock+maxspf*maxspf*2)
      if(lpop2)  workcdim=max(workcdim,totphidim)
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
! call routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(gfgwp))deallocate(gfgwp)
if(allocated(aigwp))deallocate(aigwp)
if(allocated(zgwp))deallocate(zgwp)
if(allocated(gs2))deallocate(gs2)
if(allocated(agmat))deallocate(agmat)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(gfgwp(gfgwpdim))
allocate(aigwp(aigwpdim))
allocate(zgwp(zgwpdim))
allocate(gs2(gs2matdim))
allocate(agmat(agmatdim))

      call subdengen(mr(mrort),mc(mcpsi),&
           ms(mspsi),mi(mijindx),mr(mrpldata1),&
           mr(mrtrafo),mr(mrweight),mc(mcexphin),mc(mcrueck),&
           mi(mifftfak),agmat,gs2,mc(mcadgwp),gfgwp,aigwp,zgwp,&
           mr(mrqcent))

!      return
      
deallocate(mc,mr,mi,ml,ms,mf,agmat,gfgwp,aigwp,zgwp,gs2)

      write(6,   '(80a1)') ('-',i=1,80) 
      write(ilog,'(80a1)') ('-',i=1,80)
      stop
  999 continue
      message= 'Error reading :'//trim(filename)
      write(6,'(a)') message
      stop
 1000 continue
      message = 'Cannot open file: '//trim(filename)
      call errormsg
      stop

      contains

!#######################################################################

      subroutine subdengen(ort,psi,spsi,jindx,wfunc,&
                 trafo,weight,exphin,rueck,fftfak,agmat,gs2,adgwp,gfgwp,&
                 aigwp,zgwp,qcenter)

use global
use aglobal
use griddatmod
use psidef
use datenmod
use operdef
use plot
use channels
use dengenmod

use rddvrmod
use rdfilesmod, only: rdpsi
use psiutil, only: getweight

      implicit none

! standard variables for interactive plots
      integer(long)                               :: ndata,plottask,psicount,n,chkdvr,&
                                                     chkgrd,chkpsi,chkdat,i,nprocessed
      integer(long), dimension(1)                 :: idum,idum1
      integer(long), dimension(adim), intent(out) :: jindx
      integer(long), dimension(*), intent(out)    :: fftfak
      integer(long), dimension(gwpinfodim)        :: gwpdep

      real(dop)                                                 :: time
      real(dop), dimension(1)                                   :: rdum,rdum1,rdum2
      real(dop), dimension(ortdim), intent(out)                 :: ort
      real(dop), dimension(:), allocatable                      :: pltime
      real(dop), dimension(dvrdim), intent(out)                 :: trafo
      real(dop), dimension(ortdim), intent(out)                 :: weight
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(out) :: wfunc

      complex(dop), dimension(1)                      :: zdum,zdum1
      complex(sip), dimension(spsidim), intent(out)   :: spsi
      complex(dop), dimension(dgldim), intent(out)    :: psi
      complex(dop), dimension(fftdim), intent(out)    :: rueck
      complex(dop), dimension(expdim), intent(out)    :: exphin
      real(dop), dimension(qcentdim), intent(out)     :: qcenter
      complex(dop), dimension(agmatdim), intent(out)  :: agmat
      complex(dop), dimension(gsmatdim), intent(out)  :: adgwp
      complex(dop), dimension(gfgwpdim), intent(out)  :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out)  :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)   :: zgwp
      complex(dop), dimension(gs2matdim), intent(out) :: gs2
      logical(kind=4)                                 :: lerr,linwf

      plottask=5
      lshowsys=.true.
      ndata=1
      call zerovxd(weight,ortdim)

!-----------------------------------------------------------------------
! read DVR and operator data
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1

      dvrdata(1)=.true.      ! ort
      dvrdata(2)=.true.      ! trafo
      if(lfft) then
         dvrdata(6)=.true.   ! rueck
         dvrdata(7)=.true.   ! fftfak
         dvrdata(11)=.true.  ! exphin
      endif
      call rddvr(ort,trafo,rdum,rdum1,zdum,rueck,fftfak,exphin,zdum1,&
                 idum,idum1,rdum2,chkdvr)

!-----------------------------------------------------------------------
! set up ML
!-----------------------------------------------------------------------
      if (lrunml) then
         call inputml(tape)
      endif

!-----------------------------------------------------------------------
! Determine the DVR weights.
!-----------------------------------------------------------------------
      call getweight(trafo,weight,ort,.true.)

!-----------------------------------------------------------------------
! Determine no. of data sets and allocate pldata arrays
!-----------------------------------------------------------------------
      if (lshowpsi .and. .not. lshowrst) then
         time=(tinit-out2)/fs
         ndata=0
         rewind(ipsi)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         psicount=-nskip
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)

         do
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if(lend .and. psicount.le.0) then
               write(6,'(/a,i6,a,i6,a)') &
                     'There are only', psicount+nskip,&
                     ' psi functions on file, but you skip over',&
                     nskip,' functions.'
               stop
            endif
            if (lend) exit

            do n=1,step-1
               call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
               if (lend) exit
            enddo

            psicount=psicount+1

            if(psicount.le.0) cycle
            if(psicount.gt.ncut)  exit
            ndata=ndata+1
         enddo
      else
         ndata = 1
      endif
      maxdata = ndata

! fill pltime array with times
      allocate(pltime(maxdata))
      pltime(1) = 0.0_dop
      do n=2,ndata
         pltime(n) = pltime(n-1)+(nskip+1)*out2/fs
      enddo

      if(lshowrst) then
         rewind(irst)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         read(irst) filever(irst)
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkdat)
         read (irst,err=999) time
         if (citype .gt. 0) then
            read(irst,err=999) (jindx(i),i=1,adim)
         endif
         read (irst,err=999) (psi(i),i=1,dgldim)
         time=time/fs 
         if (lgwp) then
            call gwp_param(psi,aGmat,gfgwp,aigwp,zgwp,qcenter,1)
            call gwp_ovl(gs2,gfgwp)
         endif
         call getwf(wfunc,psi,weight,trafo,&
                    exphin,rueck,fftfak,f1,f2,agmat,gs2)
! write to file....
         call outwfunc(iaus,wfunc,ort(zort(f1)),ort(zort(f2)),&
                       gdim(f1),gdim(f2),maxgdim,nstate,time)
      elseif(lshowpsi) then
         time=(tinit-out2)/fs
         rewind(ipsi)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
         do i = 1, nskip
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) go to 77
            time=time+out2/fs
         enddo
         nprocessed=0
         do n=1,ndata
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) go to 77
            if (lgwp) then
               call gwp_param(psi,aGmat,gfgwp,aigwp,zgwp,qcenter,1)
               call gwp_ovl(gs2,gfgwp)
            endif
            time=time+out2/fs
            call getwf(wfunc,psi,weight,trafo,&
                       exphin,rueck,fftfak,f1,f2,agmat,gs2)
! write to file....
            call outwfunc(iaus,wfunc,ort(zort(f1)),ort(zort(f2)),&
                          gdim(f1),gdim(f2),maxgdim,nstate,time)
            nprocessed=nprocessed+1
            do i = 1,step-1
               call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
               if (lend) go to 77
               time=time+out2/fs
            enddo
         enddo
   77    continue
      endif
      write(6,'(a,i5,a)')' There are',nprocessed,' densities computed.'

      deallocate(pltime)

      return

  999 routine='subdengen  (dengen.F)'
      message='Error reading restart file'
      call errormsg

      end subroutine subdengen

!-----------------------------------------------------------------------
!***********************************************************************
!-----------------------------------------------------------------------

      subroutine dengeninput

use global
use aglobal
use griddatmod
use psidef
use datenmod
use compdatmod
use plot
use channels
use dengenmod

      implicit none

      integer(long), external :: myiargc
      integer(long)           :: iarg,ierr,i,psilaenge
      character(len=32)       :: buf


      operfile  = ' '
      psifile  = ' '
      name  = './'
      fileout = 'dens2d'
      iarg = myiargc() 
      ncut = 999999
      nskip= 0
      step = 1

      lshowpsi=.true.
      lshowrst=.false.
      lpop2   =.false.
      lnw     =.false.
      
      i = 0
 100  i=i+1
      if (iarg .lt. i) goto 200
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if(buf .eq. '-i') then
            i = i+1
            call mygetarg(i,buf)
            name = buf
         else if( buf(1:3) .eq. '-f ') then
            i = i+1
            call mygetarg(i,buf)
            psifile = buf
         else if( buf(1:3) .eq. '-p ') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if( buf(1:3) .eq. '-n ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) ncut
            if (ierr .eq. 1) then
               write(6,'(/a)') 'Cannot read argument of -n'
               stop
            endif
         else if( buf(1:5) .eq. '-skip ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) nskip
            if (nskip .lt. 0) then
               write(6,'(/a)') &
               'Argument of -skip must be positive integer'
               stop
            endif
            if (ierr .eq. 1) then
               write(6,'(/a)') 'Cannot read argument of -skip'
               stop
            endif
         else if( buf(1:5) .eq. '-step ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (step .lt. 0) then
               write(6,'(/a)')&
               'Argument of -step must be positive integer'
               stop
            endif
         else if( buf(1:3) .eq. '-o ' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
            
         else if( buf .eq. '-rst') then
            lshowrst=.true.
         else if( buf .eq. '-pop2') then
            lpop2=.true.
         else if( buf .eq. '-nw'  ) then
            lnw=.true.

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(/1x,a)')&
       'Purpose: Generates 2D-densities from the wavefunction.'
            write(6,'(1x,2a)')&
       'Usage:   dengen [-i -f -n -skip -step -o -rst -pop2 -nw ',&
                           '-w -ver -h -?] f1 f2'
            write(6,'(/,a)') 'Options : '
            write(6,'(a)') ' -i DIR  : data is stored in directory DIR'
            write(6,'(2a,/,9x,a)')&
                     ' -f FILE : the wavefunction is read '&
                    ,'from FILE rather than from ./psi',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a)') &
                     ' -o FILE : The ouput is written to file ',&
                     'FILE rather than to ./dens2d.dat',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(a)')&
                     ' -n num  : only num WFs are processed.'
            write(6,'(a)')&
                     ' -skip m : the first m WFs are skipped.'
            write(6,'(2a)')&
                     ' -step step : only every step''s WF '&
                    ,'will be processed.'
            write(6,'(2a)')&
                     ' -rst    : the restart file is read,  '&
                    ,'rather than the psi file.'
            write(6,'(2a)')&
                     ' -nw     : No weights are employed.',&
                     ' Plot the "naked" DVR populations.'
            write(6,'(2a,/10x,2a)')&
                     ' -pop2   : the basis set occupations are '&
                    ,'plotted rather than grid'&
                    ,' populations. For FFT this implies momentum '&
                    ,'space representation.'
            write(6,'(2a)')&
                     ' -w      : Overwrite the output file.'
            write(6,'(a)')&
                     ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(/2x,2a,/2x,a)')&
            'The arguments f1 and f2 denote the DOFs ',&
            'for which a 2D-density is generated.',&
            'The arguments f1 and f2 may be integers or modelabels.'
            write(6,'(/2x,2a,/2x,2a,/2x,2a,/2x,2a,/2x,2a,/2x,2a)')&
            'If there are more WFs on psi file than you ',&
            'want to inspect, use the','options -n, -skip, or -step. ',&
            'Example: "dengen -skip 10 -n 2 rd rv" will',&
            'generate densities of the WFs no. 11 and 12 ',&
            '(i.e. t=10 and 11 if tpsi=1fs),',&
            'and "dengen -step 10 -n 2 rd rv" will gen. ',&
            'dens. of the WFs no. 1 and 11.',&
            'NB1: The option -pop2 is',&
            'ignored for phifbr,sphfbr,k,external bases.',&
            'NB2: To visualise the densities run ', &
            '"gnuplot <dengen-output-file>"', &
            '-----------------------------',&
            '---------------------------------------------------'

            stop
         elseif(buf .eq. '-w') then
            lovwr = .true.
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      else
         if(iarg.ne.i+1) goto 200
         f1lab=trim(buf)
         read(buf,*,iostat=ierr) f1
         if (f1 .le. 0 .and. ierr.eq.0) then
            write(6,'(/a)')&
            'The first Argument f1 must be positive integer'
            stop
         endif
         if (ierr .ne. 0) f1=0
         call mygetarg(i+1,buf)
         f2lab=trim(buf)
         read(buf,*,iostat=ierr) f2
         if (f2 .le. 0 .and. ierr.eq.0) then
            write(6,'(/a)')&
            'The second Argument f2 must be positive integer'
            stop
         endif
         if (ierr .ne. 0) f2=0
      end if
 
!-----------------------------------------------------------------------
! set up file names
!-----------------------------------------------------------------------
      laenge=index(name,' ')-1
      call abspath(name,laenge)

      if (lshowrst) then
         if (psifile .eq. ' ') psifile=name(1:laenge)//'/restart'
         psilaenge=index(psifile,' ')-1
         call abspath(psifile,psilaenge)
         inquire(file=psifile(1:psilaenge),exist=lshowpsi)
         if(.not.lshowpsi) then
            routine='showsysinput'
            message='There is no RESTART file: '//psifile(1:psilaenge)
            call errormsg
         endif
      elseif (lshowpsi) then
         if (psifile .eq. ' ') psifile=name(1:laenge)//'/psi'
         psilaenge=index(psifile,' ')-1
         call abspath(psifile,psilaenge)
         inquire(file=psifile(1:psilaenge),exist=lshowpsi)
      endif

      dname=name
      dlaenge=index(dname,' ')-1
      call abspath(dname,dlaenge)

      laeout=index(fileout,' ')-1
      call abspath(fileout,laeout)

      return 
      
  200 continue
      write(6,'(/2a,/a,i3,a)') 'Input error! There must be ',&
                'exactly two argument (after the options).',&
                'There are', iarg-i+1,' arguments.'
      stop
      end subroutine dengeninput

!#######################################################################

      subroutine outwfunc(iout,wfunc,ort1,ort2,gdim1,gdim2,&
                          maxgdim,nstate,time)

      implicit none

      integer(long)                                            :: g1,g2,s
      integer(long), intent(in)                                :: iout,gdim1,gdim2,maxgdim,nstate
      real(dop), intent(in)                                    :: time
      real(dop), dimension(gdim1), intent(in)                  :: ort1
      real(dop), dimension(gdim2), intent(in)                  :: ort2
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(in) :: wfunc

      write(iout,'(a,f8.3)') '# Time: ',time
      write(iout,'(a,f8.3,a)') 'set title "Time :  ',time,'"'
      write(iout,'(a)') 'splot "-" notitle w l; pause -1'
      do g2=1,gdim2
         do g1=1,gdim1
            write(iout,'(2f12.6,99e16.8)') ort1(g1),ort2(g2),&
                  (wfunc(g1,g2,s),s=1,nstate)
         enddo
         write(iout,*) 
      enddo
      backspace(iout)
      write(iout,'(a)') 'e'
      call myflush(iout)

      return
      end subroutine outwfunc

!#######################################################################

      subroutine getwf(wfunc,psi,weight,trafo,exphin,&
                       rueck,fftfak,f1,f2,agmat,gs2)

use global
use aglobal
use griddatmod
use psidef
use maxv, only: maxdim,maxsta
use compdatmod
use plot
use operdef

use operutil, only: calcd2d
use psiutil, only: showpop2

      implicit none

      integer(long)                            :: f,zeig,m,s,n,ilb1,iub1,&
                                                  ilb2,iub2,g1,g2
      integer(long)                            :: m1,m2,vd,md,nd,gda,gdb,ff
      integer(long), dimension(maxdim)         :: igpoint
      integer(long), dimension(gwpinfodim)     :: idum
      integer(long), intent(in)                :: f1,f2
      integer(long), dimension(fftdim), intent(in)              :: fftfak
      real(dop)                                                 :: adml
      real(dop), dimension(2)                                   :: rdum
      real(dop), dimension(maxsta)                              :: diab
      real(dop), dimension(ortdim), intent(in)                  :: weight
      real(dop), dimension(dvrdim), intent(in)                  :: trafo
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(out) :: wfunc
      complex(dop), dimension(expdim), intent(in)               :: exphin
      complex(dop), dimension(fftdim), intent(in)               :: rueck
      complex(dop), dimension(agmatdim), intent(in)             :: agmat
      complex(dop), dimension(gs2matdim), intent(in)            :: gs2
      complex(dop), dimension(dgldim), intent(inout)            :: psi
      real(dop), dimension(maxdim)                              :: gpoint
      logical(kind=4)                                           :: swap

      call initvxi(igpoint,ndof)
      do m=1,nmode
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            if (icut(f).gt.0) igpoint(f)=icut(f)
         enddo
      enddo

      ilb1=1
      iub1=gdim(f1)
      if (f2 .gt. 0) then
         ilb2=1
         iub2=gdim(f2)
      else
         ilb2=1
         iub2=1
      endif

      if(lpop2) then
         lswap = .true.
         call showpop2(psi,trafo,exphin,rueck,fftfak,f1,f2)
      endif

      if(lrunml) then
         s=1
         m1=dofspf(f1)
         if(f2.eq.0) then
            m2=m1
            ff=f1
         else
            m2=dofspf(f2)
            ff=f2
         endif
         if(m1.eq.m2 .and. f2.gt.0) then  ! both DOFs in same mode
            if(vgdim(f1).gt.vgdim(f2)) then
               swap=.true.  ! swap is true, if f2 is before f1 in mode
               vd=vgdim(f2)
               nd=ngdim(f1)
               gda=gdim(f2)
               gdb=gdim(f1)
            else
               swap=.false.
               vd=vgdim(f1)
               nd=ngdim(f2)
               gda=gdim(f1)
               gdb=gdim(f2)
            endif
            md=subdim(m1)/(vd*nd*gda*gdb)
            call calcd2d_1ml(psi,psi(zetf(m1,s)),vd,gda,md,gdb,nd, &
                 gdim(f1),gdim(f2),dim(m1,s),adml,swap,lnw, &
                 wfunc,maxgdim,weight(zort(f1)),weight(zort(f2)), &
                 ilb2,iub2,ilb1,iub1,agmat)
         else
            zeig=subdim(m1)*dim(m1,s)+1
            call calcd2d_2ml(psi,psi(zetf(m1,s)),psi(zetf(m2,s)), &
                 vgdim(f1),gdim(f1),ngdim(f1),vgdim(ff),gdim(ff), &
                 ngdim(ff),f2,dim(m1,s),dim(m2,s),adml, &
                 lnw,wfunc,maxgdim,weight(zort(f1)), &
                 weight(zort(ff)),ilb2,iub2,ilb1,iub1,agmat)
         endif ! m1.eq.m2

      else

         do g2=ilb2,iub2
            do g1=ilb1,iub1
               igpoint(f1)=g1
               if (f2 .gt. 0) igpoint(f2)=g2

!...........Calculate the 2D density.
               zeig=2*maxblock+1
               call calcd2d(psi,diab,f1,f2,igpoint,gpoint,&
                    agmat,gs2,rdum,idum,weight)
               do s=1,nstate
                  if(lnw) then
                     wfunc(g1,g2,s)=diab(s)
                  else
                     wfunc(g1,g2,s)=diab(s)/weight(zort(f1)-1+g1)
                     if (f2 .gt. 0) then
                        wfunc(g1,g2,s)=wfunc(g1,g2,s)/&
                                      weight(zort(f2)-1+g2)
                     endif
                  endif
               enddo
            enddo
         enddo

      endif  ! if(lrunml)


      return
      end subroutine getwf

!####################   END of DENGEN  #################################
      end program dengen
