!***********************************************************************
!
!             Analyse program EDENGEN
!
! This program reads the psi-file and calculates a 2D paircorrelation
!
!    05/08 Emmerich Tempfli, SZ 2009
!
! This program is a slight modifi!ation of dengen.
! It writes off-diagonal 2D-density matrix elements, <x,x | rho2 | x',x'>,
! to an ASCII file, the data of which may be plotted with GNUPLOT. 
!
!***********************************************************************

      module edengenmod

use decimal, only:long
use versions

      integer(long), save     :: f1,f2
      character(len=16), save :: f1lab,f2lab

      end module edengenmod


      program edengen

use decimal, only: long,dop,sip
use global
use globmemmod
use aglobal
use maxv, only: maxdim, maxsta,maxkoe
use dvrdatmod
use griddatmod
use psidef
use operdef
use datenmod
use plot
use channels
use basismod
use edengenmod

use op1lib, only: initvxi, zerovxd
use logdat, only: errormsg
use rddvrmod
use iorst, only: rstinfo
use iofile, only: rdpsiinfo
use zeigauswmod, only: zeigausw

      implicit none

      integer(long)     :: chkdvr,chkgrd,chkpsi,chkdat,ilbl,i
      real(dop)         :: totmem
      character(len=c5) :: filename
      logical(kind=4)   :: lerr,linwf,lexist

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file. This is set in dengeninput
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.false.

!-----------------------------------------------------------------------
! read input options
!-----------------------------------------------------------------------
      call edengeninput

!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      open (unit=ilog,file=name(1:laenge)//'/dengen.log',form='formatted')
      logisopen=.true.
      write(6,   '(80a1)') ('-',i=1,80)
      write(ilog,'(80a1)') ('-',i=1,80)
      call fpstartup(filename)
      write(ilog,'(a80)') filename

!-----------------------------------------------------------------------
!  Opening the PSI file.
!-----------------------------------------------------------------------
      if (lshowrst) then
         filename=psifile
         open(irst,file=filename,form='unformatted',status='old',&
              err=1000)
         ilbl=index(psifile,' ')-1
         write(ilog,'(2a)') ' The restart file is : '//psifile(1:ilbl)
      elseif (lshowpsi) then
         filename=psifile
         open(ipsi,file=filename,form='unformatted',status='old',&
              err=1000)
         ilbl=index(psifile,' ')-1
         write(ilog,'(2a)') ' The psi file is : '//psifile(1:ilbl)
      endif

!-----------------------------------------------------------------------
! get memory sizes
!-----------------------------------------------------------------------
      if (lshowrst) then
         call rdmemdim(irst)
      else
         call rdmemdim(ipsi)
      endif

!-----------------------------------------------------------------------
! Allocate mememory
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_psidef
      call alloc_plot

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
! lfft   : FFT arrays rueck and fftfak
!-----------------------------------------------------------------------
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.false.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.true.
      lfft=lpop2

!-----------------------------------------------------------------------
! read system information from DVR file
!-----------------------------------------------------------------------
      chkdvr=1
      filename=dname(1:dlaenge)//'/dvr'
      ilbl=index(filename,' ')-1
      open(idvr,file=filename,form='unformatted',status='old',err=1000)
      write(ilog,'(2a)') ' The DVR file is : ',filename(1:ilbl)
      call dvrinfo(lerr,chkdvr)
      close(idvr)
      call chksyserr(filename,lerr)
      if (lerr) go to 999

!-----------------------------------------------------------------------
! check arguments (i.e. modes chosen correctly)
!-----------------------------------------------------------------------
      if(f1.eq.0) then
         do i =1,ndof
            if(f1lab.eq.modelabel(i)) f1=i
         enddo
      endif
      if(f1.eq.0) then
         write(6,'(2a)') ' Unknown modelabel: ',f1lab
         stop
      endif
      if(f2.eq.0) then
         do i =1,ndof
            if(f2lab.eq.modelabel(i)) f2=i
         enddo
      endif
      if(f2.eq.0) then
         write(6,'(2a)') ' Unknown modelabel: ',f2lab
         stop
      endif

      if( f1.gt.ndof  .or.  f2.gt.ndof ) then
         write(6,'(a,/a,i4,a,i4,a,i4)')  &
             'f1 and f2 must not be larger than ndof', &
             'f1 =', f1,',  f2 =', f2,',  ndof =', ndof
         stop
      endif
      write(6,'(a,i3,3x,a8,1x,a8,i4)') &
          ' First  DOF =',f1,modelabel(f1),cbas(basis(f1)),gdim(f1)
      write(6,'(a,i3,3x,a8,1x,a8,i4)') &
          ' Second DOF =',f2,modelabel(f2),cbas(basis(f2)),gdim(f2)
      if(basis(f1)*basis(f2).eq.0) then
         write(6,'(a)') &
           ' Neither f1 nor f2 must refer to the electronic basis!' 
         stop
      endif
      if(f1.eq.f2) then
         write(6,'(a)') ' f1 und f2 must be different!'
         stop
      endif

!-----------------------------------------------------------------------
!  Opening the output file.
!-----------------------------------------------------------------------
      filename=fileout(1:laeout)//'_'//modelabel(f1)//'_'//modelabel(f2)
      call rmblnk(filename)
      fileout=filename
      laeout=index(fileout,' ')-1
      write(6,'(2a)') ' The output file is : '//fileout(1:laeout)
      write(ilog,'(2a)') ' The out file is : '//fileout(1:laeout)
      if(.not.lovwr) then
         inquire(file=fileout(1:laeout),exist=lexist)
         if(lexist) then
            write(6,'(a)') '### Output file already exist. Use -w to overwrite.'
            write(ilog,'(a)') '### Output file already exist. Use -w to overwrite.'
            stop
         endif
      endif
      open(iaus,file=filename,form='formatted',status='unknown',err=1000)

      write(ilog,'(a,i3,2x,a8,2x,a8,i4)') &
          ' First  DOF =',f1,modelabel(f1),cbas(basis(f1)),gdim(f1)
      write(ilog,'(a,i3,2x,a8,2x,a8,i4)') &
          ' Second DOF =',f2,modelabel(f2),cbas(basis(f2)),gdim(f2)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      if (lshowrst) then
         filename='restart'
         read(irst,err=999) filever(irst)
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkdat)
      elseif (lshowpsi) then
         filename='psi'
         read(ipsi,err=999) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      endif

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

!-----------------------------------------------------------------------
! here assign any further memory arrays
!-----------------------------------------------------------------------
      mrweight=mrdim
      mrdim=mrweight+ortdim
      mrpldata=1
      mrpldata1=mrdim
      mrdim=mrpldata1+maxgdim2*nstate

!-----------------------------------------------------------------------
! assign work arrays
!-----------------------------------------------------------------------
      workcdim=max(workcdim,dgldim+maxspf*2)
      workcdim=max(workcdim,2*maxblock+maxspf*maxspf*2)
      if(lpop2)  workcdim=max(workcdim,totphidim)
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
! call subroutine 
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(agmat))deallocate(agmat)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(agmat(agmatdim))

      call subedengen(mr(mrort),mc(mcpsi), &
           mc(mcworkc),ms(mspsi),mi(mijindx),mr(mrpldata1), &
           mr(mrtrafo),mr(mrweight),mc(mcexphin),mc(mcrueck), &
           mi(mifftfak),agmat,mc(mcadgwp))

deallocate(mc,mr,mi,ml,ms,mf)

      stop
  999 continue
      message= 'Error reading :'//trim(filename)
      write(6,'(a)') message
      stop 1
 1000 continue
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//trim(filename)
      call errormsg
      stop 

      contains

!#######################################################################

      subroutine subedengen(ort,psi,workc,spsi,jindx,wfunc, &
                 trafo,weight,exphin,rueck,fftfak,agmat,adgwp)

use rdfilesmod, only: rdpsi
use psiutil, only: getweight
use rddvrmod, only: rddvr
use edengenmod

      implicit none

! standard variables for interactive plots
      integer(long) :: ndata,plottask,psicount,n,chkdvr,chkgrd,chkpsi,chkdat
      integer(long), dimension(1) :: idum
      integer(long), dimension(adim), intent(out) :: jindx
      integer(long), dimension(*), intent(out) :: fftfak
      integer(long)                            :: i,nprocessed
      integer(long), dimension(gwpinfodim)       :: trajst
      integer(long), dimension(gwpinfodim)       :: gwpdep
      real(dop), dimension(ortdim), intent(out)  :: ort
      real(dop), dimension(1)                    :: rdum
      real(dop)                                  :: time
      real(dop), dimension(maxdata) :: pltime
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(out) :: wfunc
      real(dop), dimension(dvrdim), intent(out) :: trafo
      real(dop), dimension(ortdim), intent(out) :: weight
      complex(sip), dimension(spsidim), intent(out) :: spsi
      complex(dop), dimension(dgldim), intent(out)  :: psi
      complex(dop), dimension(workcdim), intent(out)  :: workc
      complex(dop), dimension(fftdim), intent(out)  :: rueck, exphin
      complex(dop), dimension(1)                    :: zdum
      complex(dop), dimension(agmatdim), intent(out)  :: agmat
      complex(dop), dimension(gsmatdim), intent(out) :: adgwp
      logical(kind=4) :: lerr,linwf

      plottask=5
      lshowsys=.true.
      ndata=1
      call zerovxd(pltime,maxdata)
      call zerovxd(weight,ortdim)

!-----------------------------------------------------------------------
! read DVR and operator data
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1

      dvrdata(1)=.true.      ! ort
      dvrdata(2)=.true.      ! trafo
      if(lfft) then
         dvrdata(6)=.true.   ! rueck
         dvrdata(7)=.true.   ! fftfak
         dvrdata(11)=.true.  ! exphin
      endif
      call rddvr(ort,trafo,rdum,rdum,zdum,rueck,fftfak,exphin,zdum, &
                 idum,idum,rdum,chkdvr)

!-----------------------------------------------------------------------
! Determine the DVR weights.
!-----------------------------------------------------------------------
      call getweight(trafo,weight,ort,.true.)

!-----------------------------------------------------------------------
! Determine the number of data-sets (times).
!-----------------------------------------------------------------------
      if (lshowrst) then
         ndata=1
         pltime(ndata)=0.d0
      elseif(lshowpsi) then
         time=(tinit-out2)/fs
         ndata=0
         rewind(ipsi)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         psicount=-nskip
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
 12      continue
         call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)

         if (lend .and. psicount.le.0) then
            write(6,'(/a,i6,a,i6,a)') 'There are only', psicount+nskip, &
                  ' psi functions on file, but you skip over', &
                  nskip,' functions.'
            stop
         endif 
         if (lend) goto 13 
         psicount=psicount+1
         time=time+out2/fs
         if(psicount.le.0) goto 12 
         ndata=ndata+1
         pltime(ndata)=time
         if(psicount.lt.ncut)  goto 12
 13      continue
      endif
  
      if(lshowrst) then
         rewind(irst)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         read(irst) filever(irst)
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkdat)
         read (irst,err=999) time
         read (irst,err=999) (psi(i),i=1,dgldim)
         time=time/fs 
         call egetwf(wfunc,psi,workc,weight,trafo, &
                    exphin,rueck,fftfak,f1,f2,gdim(f1),gdim(f2))
! write to file....
         call eoutwfunc(iaus,wfunc,ort(zort(f1)),ort(zort(f2)), &
                       gdim(f1),gdim(f2),maxgdim,nstate,time)
      elseif(lshowpsi) then
         time=(tinit-out2)/fs
         rewind(ipsi)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
         do i = 1, nskip
            call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
            if (lend) go to 77
            time=time+out2/fs
         enddo
         nprocessed=0
         do n=1,ndata
            call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
            if (lend) go to 77
            time=time+out2/fs
            call egetwf(wfunc,psi,workc,weight,trafo, &
                       exphin,rueck,fftfak,f1,f2,gdim(f1),gdim(f2))
! write to file....
            call eoutwfunc(iaus,wfunc,ort(zort(f1)),ort(zort(f2)), &
                          gdim(f1),gdim(f2),maxgdim,nstate,time)
            nprocessed=nprocessed+1
            do i = 1,step-1
               call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
               if (lend) go to 77
               time=time+out2/fs
            enddo
         enddo
   77    continue
      endif
      write(6,'(a,i5,a)')' There are ',nprocessed,' densities computed.'

      return

  999 routine='subdengen  (dengen.F)'
      message='Error reading restart file'
      call errormsg

      end subroutine subedengen

!-----------------------------------------------------------------------
!***********************************************************************
!-----------------------------------------------------------------------

      subroutine edengeninput

      use edengenmod

      implicit none

      integer(long)              :: iarg,myiargc,ilbl,chkdvr,ierr,i,psilaenge
      logical(kind=4) :: lerr,lexist
      character(len=32) :: buf
      character(len=c5) :: filename !,psifile

      operfile  = ' '
      psifile  = ' '
      name  = './'
      fileout = 'pair2d'
      iarg = myiargc() 
      ncut = 999999
      nskip= 0
      step = 1

      lshowpsi=.true.
      lshowrst=.false.
      lpop2   =.false.
      lnw     =.false.
      
      i = 0
 100  i=i+1
      if (iarg .lt. i) goto 200
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if(buf .eq. '-i') then
            i = i+1
            call mygetarg(i,buf)
            name = buf
         else if( buf(1:3) .eq. '-f ') then
            i = i+1
            call mygetarg(i,buf)
            psifile = buf
         else if( buf(1:3) .eq. '-p ') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if( buf(1:3) .eq. '-n ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) ncut
            if (ierr .eq. 1) then
               write(6,'(/a)') 'Cannot read argument of -n'
               stop
            endif
         else if( buf(1:5) .eq. '-skip ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) nskip
            if (nskip .lt. 0) then
               write(6,'(/a)') 'Argument of -skip must be positive integer'
               stop
            endif
            if (ierr .eq. 1) then
               write(6,'(/a)') 'Cannot read argument of -skip'
               stop
            endif
         else if( buf(1:5) .eq. '-step ') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (step .lt. 0) then
               write(6,'(/a)') 'Argument of -step must be positive integer'
               stop
            endif
         else if( buf(1:3) .eq. '-o ' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
            
         else if( buf .eq. '-rst') then
            lshowrst=.true.
         else if( buf .eq. '-pop2') then
            lpop2=.true.
         else if( buf .eq. '-nw'  ) then
            lnw=.true.

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(/1x,2a)') &
       'Purpose: Generates pair-density matrix <x,x|rho2|y,y> ', &
       'from the wavefunction.'
            write(6,'(1x,2a)') &
       'Usage:   edengen [-i -f -n -skip -step -o -rst -pop2 -nw ', &
                           '-w -ver -h -?] f1 f2'
            write(6,'(/,a)') 'Options : '
            write(6,'(a)') ' -i DIR  : data is stored in directory DIR'
            write(6,'(2a,/,9x,a)') &
                     ' -f FILE : the wavefunction is read ' &
                    ,'from FILE rather than from ./pes', &
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a)') &
                     ' -o FILE : The ouput is written to file ', &
                     'FILE rather than to ./pair2d', &
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(a)') ' -n num  : only num WFs are processed.'
            write(6,'(a)') ' -skip m : the first m WFs are skipped.'
            write(6,'(2a)') ' -step step : only every step''s WF ' &
                    ,'will be processed.'
            write(6,'(2a)') ' -rst    : the restart file is read,  ', &
                     'rather than the psi file.'
            write(6,'(2a)') ' -nw     : No weights are employed.', &
                     ' Plot the "naked" DVR populations.'
            write(6,'(2a,/10x,2a)') ' -pop2   : the basis set occupations are ', &
                     'plotted rather than grid', &
                     ' populations. For FFT this implies momentum ', &
                     'space representation.'
            write(6,'(2a)') ' -w      : Overwrite the output file.'
            write(6,'(a)') ' -ver    : Version information about the program.'
            write(6,'(a,/a)')  &
                     ' -h      : Print this help text. ', &
                     ' -?      : Print this help text. '
            write(6,'(/2x,2a,/2x,2a)') &
            'The arguments f1 and f2 denote the DOFs ', &
            'for which a pair-density matrix is','generated. ', &
            'The arguments f1 and f2 may be integers or modelabels.'  
            write(6,'(/2x,2a,/2x,2a,/2x,2a,/2x,2a,/2x,2a,/2a)') &
            'If there are more WFs on psi file than you ', &
            'want to inspect, use the','options -n, -skip, or -step. ', &
            'Example: "edengen -skip 10 -n 2 rd rv" will', &
            'generate densities of the WFs no. 11 and 12 ', &
            '(i.e. t=10 and 11 if tpsi=1fs),', &
            'and "edengen -step 10 -n 2 rd rv" will gen. ', &
            'dens. of the WFs no. 1 and 11.', &
            'NB: The option -pop2 is', &
            ' ignored for phifbr,sphfbr,k,external bases.', &
            '-----------------------------', &
            '---------------------------------------------------'

            stop
         elseif(buf .eq. '-w') then
            lovwr = .true.
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      else
         if(iarg.ne.i+1) goto 200
         f1lab=trim(buf)
         read(buf,*,iostat=ierr) f1
         if (f1 .le. 0 .and. ierr.eq.0) then
            write(6,'(/a)') 'The first Argument f1 must be positive integer'
            stop
         endif
         if (ierr .ne. 0) f1=0
         call mygetarg(i+1,buf)
         f2lab=trim(buf)
         read(buf,*,iostat=ierr) f2
         if (f2 .le. 0 .and. ierr.eq.0) then
            write(6,'(/a)') 'The second Argument f2 must be positive integer'
            stop
         endif
         if (ierr .ne. 0) f2=0
      end if
 
      laenge=index(name,' ')-1
      call abspath(name,laenge)

!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      open (unit=ilog,file=name(1:laenge)//'/edengen.log',form='formatted')
      logisopen=.true.
      write(6,   '(80a1)') ('-',i=1,80) 
      write(ilog,'(80a1)') ('-',i=1,80)
      call fpstartup(filename)
      write(ilog,'(a80)') filename 

      if (lshowrst) then
         if (psifile .eq. ' ') psifile=name(1:laenge)//'/restart'
         psilaenge=index(psifile,' ')-1
         call abspath(psifile,psilaenge)
         inquire(file=psifile(1:psilaenge),exist=lshowpsi)
         if(.not.lshowpsi) then
            routine='showsysinput'
            message='There is no RESTART file: '//psifile(1:psilaenge)
            call errormsg
         endif
      elseif (lshowpsi) then
         if (psifile .eq. ' ') psifile=name(1:laenge)//'/psi'
         psilaenge=index(psifile,' ')-1
         call abspath(psifile,psilaenge)
         inquire(file=psifile(1:psilaenge),exist=lshowpsi)
      endif

      dname=name
      dlaenge=index(dname,' ')-1
      call abspath(dname,dlaenge)

      laeout=index(fileout,' ')-1
      call abspath(fileout,laeout)

      return 
      
  200 continue
      write(6,'(/2a,/a,i3,a)') 'Input error! There must be ', &
                'exactly two argument (after the options).', &
                'There are', iarg-i+1,' arguments.'
      stop

 1000 write(6,*)' Cannot open file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
 1010 write(6,*)' Error reading: '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
      
      end subroutine edengeninput

!#######################################################################

      subroutine eoutwfunc(iout,wfunc,ort1,ort2,gdim1,gdim2, &
                          maxgdim,nstate,time)

      implicit none

      integer(long), intent(in) :: iout,gdim1,gdim2,maxgdim,nstate
      integer(long)             :: g1,g2,s
      real(dop), dimension(gdim1), intent(in) :: ort1
      real(dop), dimension(gdim2), intent(in) :: ort2
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(in) :: wfunc
      real(dop), intent(in)                                                :: time

      write(iout,'(a,f8.3)') '# Time: ',time
      write(iout,'(a,f8.3,a)') 'set title "Time :  ',time,'"'
      write(iout,'(a)') 'splot "-" notitle w l; pause -1'
      do g2=1,gdim2
         do g1=1,gdim1
            write(iout,'(2f12.6,99e16.8)') ort1(g1),ort2(g2), &
                  (wfunc(g1,g2,s),s=1,nstate)
         enddo
         write(iout,*) 
      enddo
      backspace(iout)
      write(iout,'(a)') 'e'
      call myflush(iout)

      return
      end subroutine eoutwfunc

!#######################################################################

      subroutine egetwf(wfunc,psi,workc,weight,trafo,exphin,rueck, &
                       fftfak,f1,f2,gdim1,gdim2)

use psiutil, only: showpop2, ecalcd2d

      implicit none

      integer(long), intent(in) :: f1,f2,gdim1,gdim2
      integer(long)             :: f,zeig,m,s,n,ilb1,iub1,ilb2,iub2,g1,g2
      integer(long), dimension(fftdim), intent(in) :: fftfak
      integer(long), dimension(maxdim)             :: igpoint
      real(dop), dimension(ortdim), intent(in)     :: weight
      real(dop), dimension(maxgdim,maxgdim,nstate), intent(out) :: wfunc
      real(dop), dimension(dvrdim), intent(in) :: trafo
      real(dop), dimension(maxsta)             :: diab
      complex(dop), dimension(gdim1,gdim2), intent(inout) :: psi
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), dimension(maxspf*maxspf,2)       :: spfpnt
      complex(dop), dimension(expdim), intent(in) :: exphin
      complex(dop), dimension(fftdim), intent(in) :: rueck

      call initvxi(igpoint,ndof)
      do m=1,nmode
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            if (icut(f).gt.0) igpoint(f)=icut(f)
         enddo
      enddo

      ilb1=1
      iub1=gdim(f1)
      if (f2 .gt. 0) then
         ilb2=1
         iub2=gdim(f2)
      else
         ilb2=1
         iub2=1
      endif

      if(lpop2) then
         lswap = .true.
         call showpop2(psi,trafo,exphin,rueck,fftfak,f1,f2)
      endif

      do g2=ilb2,iub2
         do g1=ilb1,iub1
            igpoint(f1)=g1
            if (f2 .gt. 0) igpoint(f2)=g2

!...........!alculate the 2D paircorrelation.
            zeig=2*maxblock+1
            call ecalcd2d(psi,diab,f1,f2,igpoint,spfpnt)
            do s=1,nstate
               if(lnw) then
                  wfunc(g1,g2,s)=diab(s)
               else
                  wfunc(g1,g2,s)=diab(s)/weight(zort(f1)-1+g1)
                  if (f2 .gt. 0) then
                     wfunc(g1,g2,s)=wfunc(g1,g2,s)/weight(zort(f2)-1+g2)
                  endif
               endif
            enddo
         enddo
      enddo

      return
      end subroutine egetwf

      end program edengen

!####################   END of EDENGEN  ################################
