!-----------------------------------------------------------------------
! Finds the optimal matches between a number of lists of eigenvalues
! obtained with the (real) filter program.
! Call as: fdmatch file_1 ... file_k | sort -n | less
! where k is the number of files and file_j is typically the *.eig file
! generated by the filter program.
!-----------------------------------------------------------------------

      program fdmatch

      implicit none

      integer, parameter :: long  = selected_int_kind(9)    
      integer, parameter :: dop = selected_real_kind(15,307) ! real*8 and complex*16

      integer(long), parameter :: maxenergies = 1000, maxdatasets = 20

      logical(kind=4), dimension(1:maxenergies,1:maxdatasets)             :: isout
      integer(long)                                                       :: iargc
      integer(long)                                                       :: datasets,set,set1,&
                                                                             set2,value,pos,file
      integer(long), dimension(1:maxdatasets)                             :: values,filenamelength
      integer(long), dimension(1:maxenergies,1:maxdatasets,1:maxdatasets) :: matches
      real(dop), dimension(1:maxenergies,1:maxdatasets)                   :: energy,intensity
      character(len=256)                                :: argument
      character(len=36*maxdatasets)                     :: line
      character(len=256), dimension(1:maxdatasets)      :: filename

!-----------------------------------------------------------------------
!     Get number of data sets and file names from command line
!-----------------------------------------------------------------------

      datasets = iargc()
      if (datasets .lt. 2) goto 567
      call getarg(1,argument)
      if( argument .eq. '-h' ) goto 567 
      if (datasets .gt. maxdatasets) then
         write(6,'(/a)') 'ERROR! Too many data sets.'
         write(6,'(a,i3,a,i3)') 'datasets =', datasets,&
              ',  maxdatasets =',maxdatasets
         stop
      endif 
      do file = 1,datasets
         call getarg(file,filename(file))
         if( filename(file) .eq. '-h' ) goto 567 
         filenamelength(file) = index(filename(file),' ')-1
      enddo
         
!-----------------------------------------------------------------------
!     Read data from file
!-----------------------------------------------------------------------

      do set = 1,datasets
         open(set+10,file=filename(set)(1:filenamelength(set)),&
              status='old',form='formatted',err=666)
         values(set) = 0
 100     continue
         if (values(set)+1 .gt. maxenergies) goto 888
         read(set+10,*,err=777,end=200)&
            energy(values(set)+1,set), intensity(values(set)+1,set)
         values(set) = values(set)+1
         goto 100
 200     continue
         close(set+10)
      enddo

!-----------------------------------------------------------------------
!     Match all pairs of data sets
!-----------------------------------------------------------------------

      do set1 = 1,datasets
         do set2 = 1,datasets
            do value = 1,values(set2)
               matches(value,set2,set1) = 0
            enddo
         enddo
      enddo
      do set1 = 1,datasets
         do set2 = set1+1,datasets
            call dmatch(energy(1,set1),energy(1,set2),values(set1),&
                        values(set2),matches(1,set1,set2),&
                        matches(1,set2,set1))
         enddo
      enddo

!-----------------------------------------------------------------------
!     Output results
!-----------------------------------------------------------------------

      do set = 1,datasets
         do value = 1,values(set)
            isout(value,set) = .false.
         enddo
      enddo
      do set1 = 1,datasets
         do value = 1,values(set1)
            if (.not. isout(value,set1)) then
               pos = 1
               do set2 = 1,set1-1
                  write(line(pos:pos+36-1),'(a)')&
                     '                                      '
                  pos = pos+36
               enddo
               write(line(pos:pos+36-1),'(f18.9,1pe18.9e2)')&
                  energy(value,set1), intensity(value,set1)
               pos = pos+36
               isout(value,set1) = .true.
               do set2 = set1+1,datasets
                  if ((matches(value,set1,set2) .ne. 0) .and. .not.&
                      isout(matches(value,set1,set2),set2)) then
                     write(line(pos:pos+36-1),'(f18.9,1pe18.9e2)')&
                        energy(matches(value,set1,set2),set2),&
                        intensity(matches(value,set1,set2),set2)
                     pos = pos+36
                     isout(matches(value,set1,set2),set2) = .true.
                  else
                     write(line(pos:pos+36-1),'(a)')&
                        '                                      '
                     pos = pos+36
                  endif
               enddo
               write(6,'(a)') line(1:pos-1)
!               write(6,'(9a)') 
!     +              (line(1+(i-1)*36:19+(i-1)*36),i=1,datasets)
            endif
         enddo
      enddo

      stop

!-----------------------------------------------------------------------
!     Output error message
!-----------------------------------------------------------------------

 567  continue
      write(6,'(/2a,/9x,a)') 'Purpose: Finds the optimal matches ',&
                             'between a number of lists of ',&
             'eigenvalues obtained with the (real) filter program.'

      write(6,'(/2a)') 'Usage  : fdmatch file_1 ... file_k',&
                      ' | sort -n | less '
      write(6,'(/,7x,2a,/7x,a)') 'where file_j (j=1...k) is typically', &
                                 ' a *.eig file generated by the',&
                        'filter program. There may be up to 10 files.'
      write(6,*)

      stop


 666  continue
      write(6,'(a)') 'ERROR: cannot open file:'
      write(6,'(a)') filename(set)(1:filenamelength(set))
      stop

 777  continue
      write(6,'(a)') 'ERROR: cannot read file.'
      stop

 888  continue
      write(6,'(a)') 'ERROR: file too large.'
      stop

      end program fdmatch

