!***********************************************************************
!
!                         AUSWERT ROUTINE FLUX (V86)
!
!     This routine calculates the quantum flux by analysing the
!     matrixelements of the CAP. For definitions and details see
!     the MCTDH-review (Physics Reports (1999)) Eqs(139,190-201).
!
!     Contents: flux, flx
!
! HDM Sep 99
! HDM Nov 00, operators and projectors implemented.
! HDM Oct 2016, ported to F90, works now with ML-WFs.
! HDN Dec 2021, flux for operators which operate on mcap enabled.
!
!***********************************************************************

    program flux

use decimal, only:long,dop,sip
use global
use globmemmod
use aglobal
use maxv, only:maxdim,maxsta
use keyunitsmod
use griddatmod
use psidef
use datenmod
use compdatmod
use operdef
use hpsimod
use channels
use fluxmod
use memdimmod
use dirdyn
use gapmod
use psiutil, only: spfovl

use iofile, only: rdpsiinfo
use rdfilesmod, only: rdpsi
use hphimod, only: hphi,hunphi
use calchamod, only: calcha

use rddvrmod
use op1lib, only: zerovxd, zerovxz, cpvxz, zerovxl
use fftmod, only: fft
use mmlib, only: dmxxdz
use mtlib, only: dtxxdz
use xvlib, only: vvaxzz, dvxxdz
use logdat, only: errormsg
use compdat, only: savecsys,savecpsi 
use auswutil, only: savecdat
use zeigauswmod, only: zeigausw 
use memhopsmod, only: memhops
use zeighpsimod, only: operflag,zeighpsi
use flxinputmod
use flxprj
use flxreduce
use flxutils
use flx2ut
use flxgth
use inputmlmod
use mlanalysis
use multilayer, only: setoperml, get_moduletree
use wftree,  only: new_wftree,dgldim_wftree
use wftree1, only: dispose_wftree

use module_interface

      implicit none

      integer(long), external :: slen
      integer(long)           :: zeig,m,s,t1,t2,n,sdm,np,ilbl,i,k
      integer(long)           :: nthreads
      real(dop), external     :: fcputime,fsystime
      real(dop)               :: tcpu,tsys,ttot
      character(len=100)      :: buffer
      integer(long)           :: zcap,zortf,tm,tm0
      logical(kind=4)         :: lfound

!-----------------------------------------------------------------------
! Get the time when the job is started
!-----------------------------------------------------------------------
    call fpstartup(cstartup)
    call unixzeit(tm0)
    nthreads=0
#ifdef _OPENMP
    !$omp parallel
    nthreads = omp_get_num_threads()
    !$omp end parallel
#endif

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
    call default
    call adefault
    call flxinput(nthreads)

!-----------------------------------------------------------------------
! get memory sizes
!-----------------------------------------------------------------------
      call rdmemdim(ipsi)
      call rdmemdim(ioper)

!-----------------------------------------------------------------------
! Allocate mememory
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      if (ldd.or.lddmctdh) then
         call alloc_dirdyn
         call alloc_dddb
         call alloc_dercp
      endif
      if (lddmctdh.or.lgausspot)then
         call alloc_gapkernel
         if(ldiabdb) call alloc_dirdyn_ddmctdh
      endif
      if (loper) call alloc_hpsimod
      call alloc_psidef
      call alloc_compdat

      if(allocated(zetf_f))deallocate(zetf_f)
      if(allocated(dovmat))deallocate(dovmat)
      allocate(zetf_f(maxdim,maxsta))
      allocate(dovmat(maxdim,maxsta))

      call flxrdinfo

!-----------------------------------------------------------------------
! to allocate memory for psi2 and psi1
! 1. set no. of comparison data sets for 2 wavefunctions
! 2. save information in comparison data sets (used in reduce@flxprj.F)
!-----------------------------------------------------------------------
  if(lovwr .or. ltest) then  ! the corresponding endif is on line 324
    ncomp=2
    call savecsys(1)
    call savecpsi(1)
    call savecdat(1)
    call savecsys(2)
    call savecpsi(2)
    call savecdat(2)

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
    if (loper) then
20     nham=0
       do n=1,hamdim
          if (hamlab(n) .eq. operator) nham=n
       enddo
       lfound=.false.

       if (nham .eq. 0) then
          write(6,'(4a)')'Operator "',trim(operator), &
                         '" not found in ',trim(operfile)
          write(6,'(1x,a)') 'Name of operator ? Choose from:'
          do n=1,hamdim
             write(6,'(2x,a)') hamlab(n)
          enddo
          read(5,'(a)') operator
          if (operator .eq. ' ') stop
          goto 20
       endif
       if(kzahl(nham) .gt. 0  .or. nmulpot(nham) .gt. 0) then
          lcorr = .true.
          do k=zham(nham), zham(nham)+kzahl(nham)-1
             if(hamilton(fcap,k).ne.fcap) lfound=.true.
          enddo
          write(6,'(3a,i3)') ' nodiag: Using correlated operator : ',&
             trim(operator),' ,  nham =', nham
          write(ilog,'(/3a,i3)') ' nodiag: Using correlated operator : ',&
             trim(operator),' ,  nham =', nham
          if(lfound .and. .not.lopsym) then
             routine='flux, flux.F90'
             message='The correlated operator must not operate on the '//&
                     'CAP-DOF as the option -Os is not set.'
             call errormsg
          endif
          if(.not.nodiag(nham)) then
             ilbl=index(hamlab(nham),' ')-1
             routine='flux, flux.F90'
             message='The corelated Hamiltonian '//hamlab(nham)(1:ilbl)//&
                  ' must be created as "nodiag" type.'
             call errormsg
          endif
       elseif(.not.nodiag(nham)) then
          lcorr = .false.
          do s = nst1, nst2
             i=0
             do m = 1, nmode
                if(khzahl(m,s,nham) .gt. 0 ) then
                   if(m.eq.mcap) lfound=.true.
                   i=i+1
                 endif
             enddo
          enddo
          if(ltonly) write(6,'(2a)') 'The -x option is set! ',&
                    'Only the gtheta part is used to evaluate the flux.'
          write(6,'(3a,i3)') ' usediag:  Using uncorrelated operator : ',&
                trim(operator),' ,  nham =', nham
          write(ilog,'(/3a,i3)') ' usediag: Using uncorrelated operator : ',&
                trim(operator),' ,  nham =', nham
          if(lfound) then
             write(6,'(a,i3,a)') ' Acting on',i,' modes, including the CAP-mode.'
             write(ilog,'(a,i3,a)')' Acting on',i,' modes, including the CAP-mode.'
          else
             write(6,'(a,i3,a)') ' Acting on',i,' modes, excluding the CAP-mode.'
             write(ilog,'(a,i3,a)')' Acting on',i,' modes, excluding the CAP-mode.'
          endif

          if(lopsym) then
             write(6,'(3a,i3)')   ' The operator is applied smmetrically (-Os option).'
             write(ilog,'(3a,i3)')' The operator is applied symmetrically (-Os option).'
          endif

          if(.not.lopsym .and. lfound) then
             routine='flux, flux.F'
             message='The operator must not operate on the CAP mode,'//&
                     ' as the -Os option is not set.'
                call errormsg
          endif

       else
          write(6,'(a)') '##### This cannot happen! #####'
          write(6,'(a,l2,a,i5,a,i3)') 'nodiag =',nodiag(nham),' , kzahl =', &
                        kzahl(nham),' ,  nmulpot =',nmulpot(nham)
          write(6,'(a)') '#####  STOP  #####'
          stop
       endif ! kzahl(nham)

       write(ilog,'(a,i5,a,i2,a,99i3)')  &
            '  kzahl:',kzahl(nham),', nmulpot:',nmulpot(nham),&
            ',  khzahl:',(khzahl(n,nst1,nham),n=1,nmode)

    endif  ! loper

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
! lsphfbr : jfbr, mfbr and kinfbr matrices.
!-----------------------------------------------------------------------
       lrddvr  = .true.
       lrdoper = .true.
       ovtype  = 0
       lpsi    = .true.
       lpsi1   = .false.
       lpsigrd = .false.
       lort    = .true.
       ltrafo  = .true.
       lsphfbr = .true.
       ldmat   = .true.
       ldmat1  = .false.
       lgpop   = .false.
       call zeigausw
       lpsi    = .false.
       call memausw
       workcdim=1

       zeig=1
       do s=1,nstate
          do m=1,nmode
             dovmat(m,s)=zeig
             zeig=zeig+dim(m,s)*dim(m,s)
           enddo
        enddo
        doverdim=zeig-1
        griddim=1

!-----------------------------------------------------------------------
! stuff needed for save_CAPspf and mtrxelhun.
! capphidim and mfound are stored on flux.inc
!-----------------------------------------------------------------------
       if(.not.lprojonly) then
          capphidim=0
          do s = nst1, nst2
             capphidim=capphidim+dim(mcap,s)*sdmc
          enddo
          if(loper .and. .not.lcorr) then
             allocate(mfound(nmode))
             do m = 1, nmode
                mfound(m)=.false.
                if(m.eq.mcap .and. .not.lopsym) cycle
                do s = nst1, nst2 ! Check, if one has to operate on mode m
                   if(khzahl(m,s,nham) .gt. 0) then
                      mfound(m)=.true.
                   endif
                enddo ! s
             enddo    ! m
          endif       ! loper .and. .not.lcorr
       endif          ! lprojonly

!-----------------------------------------------------------------------
! set up various flags and array information needed to use projectors
!-----------------------------------------------------------------------
       if(lproj)  then
          zprj(1) = 1
          do np = 2, nprj
             zprj(np) = zprj(np-1) + subdim(prjmode(np-1))
          enddo
          prjdim = zprj(nprj) + subdim(prjmode(nprj))
          mcprj  = mcdim
          mcdim  = mcprj + prjdim
          do np = 1, nprj
             if(prjname(np) .eq. 'eigenf' ) then
                sdm = subdim(prjmode(np))
                workrdim = max(workrdim,2*sdm*sdm+sdm)
                workcdim = max(workcdim,2*sdm*sdm)
             endif
          enddo
          if(.not.lrunml) workcdim=max(workcdim,adim) ! for reduce
          if(lid) workcdim=max(workcdim,maxphidim) ! for mexpand in reduce
       else
          prjdim = 1
          mcprj  = 1
          do np = 1, maxprj
             zprj(np) = 1
          enddo
       endif

!-----------------------------------------------------------------------
! set up various flags and array information needed to use an operator
!-----------------------------------------------------------------------
       if(loper)  then
          call operflag(nham)
          call zeighpsi(nham)
       end if

!-----------------------------------------------------------------------
! set up array information needed for full projection (reduced 1D WF)
!-----------------------------------------------------------------------
       if(full) then
          fulldim = kmax*sdmc*(nst2-nst1+1)
       else
          fulldim = 1
       end if

!-----------------------------------------------------------------------
! set up array information needed for spatial derivative
!-----------------------------------------------------------------------
       if (lderiv) then
          mrfftp = mrdim
          mrdim = mrfftp+fftdim
!         I have no idea why this pointer isn't set from the other
!         routines (memausw/memhops) which actually do set up other
!         FFT-related pointers. Assuming that there is a reason, I
!         won't change those routines, since that would most likely
!         break something else. (FO)
          psi1dim = max(dgldim,fulldim)
          mcpsi1 = mcdim
          mcdim = mcpsi1+psi1dim
       endif

!-----------------------------------------------------------------------
! assign work arrays. The variables m?work? must be set for "memhops".
! NB work arrays must be at end of mc, mr, mi, ml, ms arrays.
!-----------------------------------------------------------------------
       workcdim=max(workcdim,2*maxblock)
       if (lderiv) then
          workcdim=max(workcdim,sdmc)
       endif

       workrdim = max(workrdim,maxgdim*maxgdim)
       workidim = max(workidim,maxgdim)

    else   ! of "if(lovwr .or. ltest) then"
!...Dummy values for lovwr=.false.
       mcdim = 10
       mrdim = 10
       workcdim = nout2
       workrdim = 100
    endif   !   END of "if(lovwr .or. ltest) then" clause.

    mcgt    = mcdim
    mcdim   = mcgt+2*nout2
    mcwm    = mcdim
    mcdim   = mcwm+laewmat
    mcworkc = mcdim
    mcdim   = mcworkc+workcdim

    mrtheta = mrdim
    mrdim   = mrtheta+gdmc
    mrworkr = mrdim
    mrdim   = mrworkr+workrdim

    miworki = midim
    midim   = miworki+workidim

    mlworkl = mldim
    mldim   = mlworkl+workldim

    msworks = msdim
    msdim   = msworks+worksdim

!---------------------------------------------------------------------
!   Assign memory for operator
!---------------------------------------------------------------------
    mrdim   = max(mrdim,4*maxendim) ! Do not move this upwards.

!---------------------------------------------------------------------
!   Assign the memory.
!---------------------------------------------------------------------
      if(lovwr .or. ltest) call memhops
      mrdim   = max(mrdim,4*maxendim) ! Do not move this upwards.

      write(ilog,'(/,80a1,/)') ('-',i=1,80)
      write(ilog,'(a)') '  *** Memory allocated dynamically ***'

      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'(a,i10,/a,i10)') &
       ' workcdim   : ', workcdim,&
       ' workrdim   : ', workrdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(hteil))deallocate(hteil)
if(allocated(agmat))deallocate(agmat)
if(allocated(agmat1))deallocate(agmat1)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(hteil(pmatdim))
allocate(agmat(agmatdim))
allocate(agmat1(agmatdim))

!-----------------------------------------------------------------------
! for GWP integrals use symmetric LHA as default and energy only
!-----------------------------------------------------------------------
      if (lgwp) then
         gwpintm(:)=2
         gwpintordm(:) = 0
         gwpcxm(:) = 0
      endif

    if( lovwr.or.ltest.or.lderiv ) call operin(zcap,zortf)
    if(lprojonly) then
       zcap=1
       zortf=1
    endif

    if( lproj  .and.  (lovwr.or.ltest) ) &
       call projin(mr(mrhops),mr(mrort), &
            mi(mijsph),mi(mimsph),mc(mcprj),mr(mrtrafo))

    if( lovwr .and. .not.full .and. .not.lderiv )               &
       call flx(mc(mcprj),mr(zcap),mr(mrtheta),mr(zortf),mc(mcwm),&
            mc(mcgt),mr(mrhops),agmat,agmat1,mc(mcadgwp),mc(mcadgwp1),tm0)

    if( full .and. .not.lderiv )  &
       call fullflx(mc(mcprj),mr(zcap),             &
                    mr(mrtheta),mr(zortf),mc(mcwm),mc(mcgt))

    if( lderiv )  &
       call fdrvflx(mc(mcprj),mr(zcap),mr(zortf),   &
                    mc(mchin),mc(mcrueck),mi(mifftfak),         &
                    mr(mrfftp),mc(mcwm),mc(mcgt))

    if(.not.ltest .and. .not.lwtt)  &
        call dblint(mc(mcwm),mc(mcgt),mr(1),mr(maxendim+1),     &
                    mr(2*maxendim+1),mr(3*maxendim+1))

    deallocate(mc)
    deallocate(mr)
    deallocate(mi)
    deallocate(ml)
    deallocate(ms)

    deallocate(hteil,agmat,agmat1)

!---------------------------------------------------------------------
! Get timing and write data to screen and "log" file
!---------------------------------------------------------------------
    if(lovwr) then
       t1   = int(tovl/3600.d0)
       tovl = tovl - 3600.d0*t1
       t2   = int(tovl/60.d0)
       tovl = tovl - 60.d0*t2
       write(6,900) 'Overlaps done  [h:m:s] : ',t1,t2,tovl
    end if

    tcpu = fcputime(0)
    tsys = fsystime(0)
    ttot = tcpu+tsys
    t1   = int(ttot/3600.d0)
    ttot = ttot - 3600.d0*t1
    t2   = int(ttot/60.d0)
    ttot = ttot - 60.d0*t2

    write(ilog,'(/a,/)') '***** Overall Timing *****'
    write(ilog,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
    write(ilog,'(a,1x,f12.2)') 'Total system time  [s] : ',tsys
    write(ilog,900) 'Total time     [h:m:s] : ',t1,t2,ttot
    write(6   ,900) 'Total time     [h:m:s] : ',t1,t2,ttot

    call unixzeit(tm)
    tm = tm-tm0
    t1 = tm/3600
    tm = tm - 3600*t1
    t2 = tm/60
    tm = tm - 60*t2
    write(ilog, 901) 'Wall  time     [h:m:s] : ',t1,t2,tm
    write(6   , 901) 'Wall  time     [h:m:s] : ',t1,t2,tm
900 format( a,i4,' :',i3,' :',f6.2)
901 format( a,i4,' :',i3,' :',i3)
    call fpstartup(buffer)
    write(ilog,'(/a)') ' -- flux86 terminated normally --'
    write(ilog,'(a)') trim(buffer)
    write(6,'(80a1)') ('*',s=1,80)
    call myflush(ilog)

    stop
    contains

!#######################################################################


    subroutine flx(projector,wgam,theta,ort,wmat,gtau,hops,agmat1,agmat2,&
               adgwp1,adgwp2,tm0)

!*********************************************************************
!    calculates the matrix elements <psi(t)|W|psi(t')> or
!    <psi(t)|W*O|psi(t')> where O is an operator.
!
! Calls to: getstart, mexpand, reduce,  ovlp, overlap, gtheta, gth0
!           mtrxelmnt, mtrxelhun, multpsi, overlap, ovlp
!*********************************************************************

    implicit none

    integer(long) ::  i,k0,k1,k2,k3,kk,kk1,k01,k02,ks2,chk,jindx(1),&
                      tm1,modus
    integer(long) ::  idum(1)
    integer(long), intent(out) :: tm0
    complex(dop)  ::  ovl
    complex(dop)  ::  workc(workcdim)
    complex(dop)  ::  wmat(laewmat),gtau(2*nout2)
    complex(dop)  ::  projector(prjdim),fullpsi(1)
    real(dop)     ::  tcpu,tsys,time
    real(dop)     ::  wgam(gdmc),theta(gdmc),ort(gdmc),hops(hopsdim)
    real(dop)     ::  fcputime,fsystime
    external      ::  fcputime,fsystime
    logical       ::  useunit
    character(len=6) ::  punit
    character(len=80) :: actual
    integer(long), dimension(gwpinfodim)            :: gwpdep1
    integer(long), dimension(gwpinfodim)            :: gwpdep2
    complex(dop), dimension(agmatdim), intent(out)  :: agmat1
    complex(dop), dimension(agmatdim), intent(out)  :: agmat2
    complex(dop), dimension(gsmatdim), intent(out)  :: adgwp1
    complex(dop), dimension(gsmatdim), intent(out)  :: adgwp2

    complex(sip) ,allocatable  ::  spsi(:)
    complex(dop),allocatable  ::  psi1(:),psi2(:),psi01(:),dtpsi(:),wrk(:)
    complex(dop),allocatable  ::  hunpsi(:),hpsi(:),dover(:),hteil(:),mcap_phi(:)

    type(wfnode_t), pointer :: flxtree


    wrkdim=max(2*adim,maxphidim)
    if(loper .and. lcorr) wrkdim=2*dgldim

    allocate(psi1(dgldim),psi2(dgldim),dover(doverdim),wrk(wrkdim))
    allocate(spsi(spsidim),mcap_phi(capphidim))
    if(loper) then
       if(lcorr) then
          allocate(hpsi(hpsidim),hteil(pmatdim),dtpsi(dgldim))
       else
          allocate(hunpsi(totphidim))
       endif
    endif
    allocate(psi01(dgldim)) ! tmp array

!-----------------------------------------------------------------------*
! Definitions:                                                          *
! mcap = mode of the CAP; fcap = degree of freedom of the CAP.          *
! sdmc = subdim(mcap)                                                   *
! gdmc = gdim(fcap)                                                     *
! wgam = W^gam = CAP.                                                   *
! wtt  = W_tt = <psi(t)|W^gam|psi(t)>                                   *
! IWtt = 2*Int <psi(t)|W^gam|psi(t)> dt + <psi(t)|Theta|psi(t)>         *
! wmat = W_tt' = <psi(t)|W^gam|psi(t')>; t=(k1-1)*out2, t'=(k2-1)*out2  *
! wmat is in symmetric storage mode (upper triangle).                   *
!-----------------------------------------------------------------------*

!-----------------------------------------------------------------------
! Get some additional arguments
!-----------------------------------------------------------------------
    chk   = 0
    tovl  = 0.d0
    if(lprojonly) goto 678

!-----------------------------------------------------------------------
! Modify the CAP Wgam according to the captype (left=1,right=2,both=3)
!-----------------------------------------------------------------------
    if(captype .eq. 1 ) then          !  left  CAP
       do i = gdim(fcap), 2, -1
          if(abs(wgam(i))+abs(wgam(i-1)) .lt. 1.d-15 ) goto 222
          wgam(i) = 0.d0
       enddo
    elseif(captype .eq. 2 ) then      !  right CAP
       do i = 1, gdim(fcap)-1
          if(abs(wgam(i))+abs(wgam(i+1)) .lt. 1.d-15 ) goto 222
          wgam(i) = 0.d0
       enddo
    endif
222 continue

    call getstart(wgam)  ! Determine start1,start2,start

!----------------------------------------------------------------------
!   Build Theta. start1 is the starting point of a left-hand side
!   cap, start2 the one of a right-hand side cap.
!----------------------------------------------------------------------
    if(start .le. gdim(fcap) ) then
       do k1 = 1, start1
             theta(k1) = 1.d0
       enddo
       do k1 = start1+1, start2-1
             theta(k1) = 0.d0
       enddo
       do k1 = start2, gdim(fcap)
             theta(k1) = 1.d0
       enddo
       ostart = ort(start)
    else
       call zerovxd(theta,sdmc)
       ostart = 0.d0
       if(.not.lwtt) write(iflx,'(a)') '# ort = No g_theta evaluation'
       write(iIwtt,'(a)') '# ort = No g_theta evaluation'
    endif

    if(i0s.ge.start-2) then
       write(6,'(a,i5,a,i5)') ' i0s =',i0s,' must be .lt. start-2 =',start-2
       stop
    endif
    if(i0s.gt.0) then
       write(ilog,'(a,i5,a,f9.2)') &
         ' i0s is set! i0s =',i0s,', ort(i0s) =',ort(i0s)
       if(nspfdof(mcap).ne.1) then
          write(6,'(a)') &
            'Sorry, -S option is allowed only for uncombined CAP mode.'
       endif
    endif


!-----------------------------------------------------------------------
! Write w^gam to log file.
!-----------------------------------------------------------------------
    write(ilog,'(/a,6x,a)') '           ort        w^gam[eV] ', &
                            '           ort        w^gam[eV] '

    k0 = max(1,gdim(fcap)-200)
    if(abs(wgam(min(k0+20,gdim(fcap)))) .ge. tiny) k0 = max(1,k0-50)
    if(abs(wgam(min(k0+20,gdim(fcap)))) .ge. tiny) k0 = max(1,k0-50)
    do k1 = k0, (gdim(fcap)+k0)/2
       k2 = k1 + (gdim(fcap)-k0+1)/2
       write(ilog,'(i5,f12.4,f15.8,6x,i5,f12.4,f15.8)')         &
            k1, ort(k1), -wgam(k1)*ev, k2, ort(k2), -wgam(k2)*ev
    enddo
    write(ilog,'(a)')
    call myflush(ilog)

!-----------------------------------------------------------------------
! Write head lines to wtt, IWtt, and flux file.
!-----------------------------------------------------------------------
    punit = '['//trim(unitf)//']   '
    if(laeout.gt.0) then
       write(iwtt, '(/a)') '#   klo    khi   kmax   step  start  ort(start)'
       write(iIwtt,'(/a)') '#   klo    khi   kmax   step  start  ort(start)'
       write(iwtt, '(a2,i5,4i7,f11.4)') '# ',klo,khi,kmax,step,start,ostart
       write(iIwtt,'(a2,i5,4i7,f11.4)') '# ',klo,khi,kmax,step,start,ostart
       if(loper) then
          if(abs(ufac) .gt. 9990.d0 .or. abs(ufac) .lt. 0.1d0) then
             write(iwtt,'(5a,1pe11.4)') '# operator = ',trim(operator),&
               ', unit = ',trim(unitf),' ,       unit-factor =', ufac
             write(iIwtt,'(5a,1pe11.4)') '# operator = ',trim(operator),&
               ', unit = ',trim(unitf),' ,       unit-factor =', ufac
          else
             write(iwtt,'(5a,f9.3)') '# operator = ',trim(operator),&
               ', unit = ',trim(unitf),' ,     unit-factor =', ufac
             write(iIwtt,'(5a,f9.3)') '# operator = ',trim(operator),&
               ', unit = ',trim(unitf),' ,     unit-factor =', ufac
          endif
          write(iwtt,'(/a,6x,2a,8x,a,9x,a,7x,2a,8x,a)') &
              '#  Time[fs]','W_tt',punit,'norm**2','W_tt/norm**2',&
              '<O>',punit,'<O>/norm**2'
          write(iIwtt,'(/a,8x,2a,12x,2a,8x,a)') &
              '#  Time[fs]', 'IW_tt',punit,'Theta',punit,'IW_tt + Theta'
       else
          write(iwtt,'(/a,9x,a,11x,a,9x,a)') '#  Time[fs]',&
              'W_tt','norm**2','W_tt/norm**2'
          write(iIwtt,'(/a,12x,a,16x,a,10x,a)') '#  Time[fs]',    &
              'IW_tt','Theta','IW_tt + Theta'
       endif ! loper
       call myflush(iwtt)
       call myflush(iIwtt)
       if(.not.lwtt) then
          write(iflx, '(/a)') '#   klo    khi   kmax   step  start  ort(start)'
          write(iflx, '(a2,i5,4i7,f11.4)') '# ',klo,khi,kmax,step,start,ostart
          if(loper) then
             if(abs(ufac) .gt. 9990.d0 .or. abs(ufac) .lt. 0.1d0) then
                write(iflx,'(5a,1pe11.4)') '# operator = ',trim(operator),&
                    ',  oper-unit = ',trim(unitf),',  unit-factor =',ufac
             else
                write(iflx,'(5a,f9.3)') '# operator = ',trim(operator),&
                    ',  unit = ',trim(unitf),',  unit-factor =',ufac
             endif
          endif
          call myflush(iflx)
       endif
    endif

678 continue
!-----------------------------------------------------------------------
! Set up the ML-stuff.
!-----------------------------------------------------------------------
    if (lrunml) then
       call inputml(tape) ! tape was read by rdpsiinfo/rdpsidef
       call get_moduletree(flxtree)
       dgldim = dgldim_wftree(flxtree)
       if (loper) then
          useunit = .not. nodiag(nham)
          call setoperml(nham,.false.,useunit)
       endif
    endif

    write(ilog,'(a,i9)') ' dgldim    =', dgldim
    write(ilog,'(a,i9)') ' adim      =', adim
    write(ilog,'(72a1,/)')  ('-',i=1,72)

!-----------------------------------------------------------------------
! Call reduce to compute the reduced wavefunction.
! Note that reduce sets nst1, nst2, and nstate anew.
!-----------------------------------------------------------------------
    if(lproj) then
       if(lrunml) then
          call reduceml(projector,psi2,spsi,dover,flxtree)
       else
          call reduce(projector,psi2,psi1,spsi,jindx,dover,workc,fullpsi)
          write(ilog,'(a,i9)') ' dgldim    =', dgldim
          write(ilog,'(a,i9)') ' adim      =', adim
       endif
       call fzeit(actual)
       i = index(actual,':')
       actual=actual(i-9:i+6)
       write(ilog,'(2a/)') ' Projection done. Time: ',trim(actual)
       call myflush(ilog)
       if(lprojonly) stop
    endif

!-----------------------------------------------------------------------
! Compute data for wt and iwtt files. End if lwtt is .true. (-wtt)
!-----------------------------------------------------------------------
    call gth0(psi2,psi1,theta,wgam,dover,time,hunpsi,&
              hpsi,hteil,hops,dtpsi,spsi,ort,mcap_phi,jindx)
    if(loper) then
       write(iwtt,'(a,6x,2a,8x,a,9x,a,7x,2a,8x,a)') &
           '#  Time[fs]','W_tt',punit,'norm**2','W_tt/norm**2',&
           '<O>',punit,'<O>/norm**2'
       write(iIwtt,'(a,8x,2a,12x,2a,8x,a)') &
           '#  Time[fs]', 'IW_tt',punit,'Theta',punit,'IW_tt + Theta'
    else
       write(iwtt,'(a,9x,a,11x,a,9x,a)') '#  Time[fs]',&
           'W_tt','norm**2','W_tt/norm**2'
       write(iIwtt,'(a,12x,a,16x,a,10x,a)') '#  Time[fs]',    &
           'IW_tt','Theta','IW_tt + Theta'
    endif
    close(iwtt)
    close(iIwtt)
    call fzeit(actual)
    i = index(actual,':')
    actual=actual(i-9:i+6)
    call unixzeit(tm1)
    write(ilog,'(3a,i5,"s"/)') ' Computation of wtt done. Time: ',&
                               trim(actual),',',tm1-tm0
    call myflush(ilog)

    if(lwtt) goto 201

!-----------------------------------------------------------------------
! Step over evaluation of g_w when -v option is set.
!-----------------------------------------------------------------------
    if(lgtt) goto 200

!-----------------------------------------------------------------------
! Start the loops over time-steps.
!-----------------------------------------------------------------------
    call zerovxz(wmat,laewmat)
    rewind (ipsi)
    if (filever(ipsi) .lt. 11.0_dop) then
       read (ipsi)
       call rdpsiinfo(ipsi,chk,chk,chk,chk)
    endif

    time=tinit
    kk1=khi  ! (will be overwritten if .not.lopsym)

    do k1=1,klo-1
       call rdpsi(ipsi,psi1,spsi,jindx,agmat1,idum,adgwp1,gwpdep1)
       if (lend) goto 77
    enddo

    do k1=klo,khi,step
       k01 = (k1-klo)/step + 1
       if(k01.eq.1) call tout(k1,klo,time,tm1)
       if(k01.gt.9 .and. mod(k01,5).eq.2 ) call tout(k1,klo,time,tm1)
       time = (k1-1)*out2

!------------------------------------------------------------------------
! Read Psi1(t=t_k1);  ket-WF, first read.
!------------------------------------------------------------------------
       dgldim=dgldimold
       call rdpsi(ipsi,psi1,spsi,jindx,agmat1,idum,adgwp1,gwpdep1)    
       if (lend) goto 77

!------------------------------------------------------------------------
! Expand psi1 to full MCTDH-length in case of identical particles.
!------------------------------------------------------------------------
       if(lid) call mexpand(psi1)

!---------------------------------------------------------------------
!  Read psi2 for times .le. actual time. (or for all times if lopsym=.true.)
!---------------------------------------------------------------------
       if(.not.lopsym) kk1=k1  ! normal procedure
       rewind (ipsi)
       if (filever(ipsi) .lt. 11.0_dop) then
          read (ipsi)
          call rdpsiinfo(ipsi,chk,chk,chk,chk)
       endif

       do k2=1,klo-1
          call rdpsi(ipsi,psi2,spsi,jindx,agmat2,idum,adgwp2,gwpdep2)
       enddo

       ks2=0
       !$OMP parallel firstprivate(psi1) private(k02,kk,ovl,flxtree) &
       !$OMP private(psi2,psi01,dover,wrk,spsi,hunpsi,hpsi,hteil,dtpsi,mcap_phi)
#ifdef _OPENMP
       if(lrunml) then
          !$OMP CRITICAL
          flxtree => new_wftree(tape,if_spfdof)
          if(lcorr) then
             !call setnewoperml(flxtree,nham,.false.,.false.,.false.)
             call newoperml(flxtree,nham,.false.,.false.,.false.)
          endif
          !$OMP END CRITICAL
       endif
#endif

       !$OMP DO schedule(dynamic,1)
       do k2=klo,kk1,step

!----------------------------------------------------------------------
!   Read Psi2(t'=t_k2)  for t'<=t, bra-WF, second read.
!----------------------------------------------------------------------
          !$OMP CRITICAL
          dgldim=dgldimold
          ks2=ks2+1
          k02=ks2
          call rdpsi(ipsi,psi2,spsi,jindx,agmat2,idum,adgwp2,gwpdep2)
          if (lend) goto 77

          !$OMP END CRITICAL

!------------------------------------------------------------------------
! Expand psi2 to full MCTDH-length in case of identical particles.
!------------------------------------------------------------------------
          if(lid) call mexpand(psi2)
          call multpsi(psi2,wgam) ! Multiply psi2 with wgam.

!-----------------------------------------------------------------------
! Apply operator and do the overlap.  ovl = <wgam*psi2|O|psi1>
!-----------------------------------------------------------------------
          if(loper .and. .not.lcorr) then
             call mtrxelhun(psi2,psi1,ovl,time,hunpsi,dover, &
                  mcap_phi,flxtree)
          elseif(loper .and. lcorr) then
             if(lrunml) then
                !call crocoml(nham,time,psi2,psi1,ovl)
                ovl=croco_tree(nham,time,flxtree,psi2,psi1)
             else
                call mtrxelmnt(psi2,psi1,time,ovl,dtpsi,hteil,&
                               hpsi,hops,mr(mrort),jindx)
             endif
          else ! no operator
             modus=0
             if(lrunml) then
                call spfovl(dover,psi2,psi1,agmat2,agmat1,modus)
                call overlap_1(psi2,psi1,flxtree,ovl,dover)
             else
                call ovlp(ovl,psi2,psi1)
             endif
          endif

          if( k02.le.k01) then          ! Normal procedure
             kk=(k01*(k01-1))/2 + k02
             wmat(kk) = wmat(kk) - ovl
          else                          ! -Os option
             kk=(k02*(k02-1))/2 + k01
             wmat(kk) = wmat(kk) - dconjg(ovl)
          endif

       end do  ! k2-loop  (psi2)
       !$OMP END DO
#ifdef _OPENMP
       call deallocate_wftree(flxtree)
       flxtree => Null()
#endif
       !$OMP end parallel

       k3 = k1+1
       if(lopsym) then   ! re-position psi-file if lopsym=.true.
          rewind (ipsi)
          if (filever(ipsi) .lt. 11.0_dop) then
             read (ipsi)
             call rdpsiinfo(ipsi,chk,chk,chk,chk)
          endif
          do k0=1,k1
             call rdpsi(ipsi,psi1,spsi,jindx,agmat1,idum,adgwp1,gwpdep1) 
             read (ipsi,end=79)
          enddo
       else  ! read over missed steps
          do i = 2, min(step,nout2-k1) 
             call rdpsi(ipsi,psi1,spsi,jindx,agmat1,idum,adgwp1,gwpdep1) 
             if (lend) goto 77
          enddo 
       endif

    end do  ! k1-loop   (psi1)

    call tout(k3,klo,time,tm1)    ! k3 = last k1 +1

    if(lopsym) then
       k0=1
       do kk=1,laewmat
          if(kk.eq.(k0*(k0+1))/2) then
             k0=k0+1
             wmat(kk) = dble(wmat(kk))
          else
             wmat(kk) = 0.5d0*wmat(kk)
          endif
       enddo
   endif

!-----------------------------------------------------------------
! Get the residual g_theta
!-----------------------------------------------------------------
200 if(.not.lwtt) call gtheta(psi2,psi1,time,spsi,dover,&
                              theta,mr(mrort),gtau,hunpsi,hpsi,   &
                              hteil,hops,dtpsi,mcap_phi,jindx)
    call tout(zeroi,klo,time,tm1)

!-----------------------------------------------------------------
! Write start and ostart=ort(start) to screen and log-file.
!-----------------------------------------------------------------
201 if( start .le. gdim(fcap) ) then
       if(captype .eq. 3 ) then
          write(6,'(a,2i4,a,f11.4)')   &
            ' start =',start1,start2,',  ort =', ostart
       else
          write(6,'(a,i4,a,f11.4)') ' start =',start,',  ort =',ostart
       endif
    else
       write(6,'(a,i4,a)')' start =',start,',  No g_theta evaluation!'
    endif
    write(ilog,'(/a,2i4,a,i4,a,f11.4)') &
        ' start1,2 =',start1,start2,',  start =',start,',  ort =',ostart
    if(start .gt. gdim(fcap)) write(ilog,'(a)') ' No g_theta evaluation!'


!---------------------------------------------------------------------
! Get timing. Will be written in main program.
!---------------------------------------------------------------------
    tcpu = fcputime(0)
    tsys = fsystime(0)
    tovl = tcpu+tsys

    deallocate(psi1,psi2,dover,wrk,spsi,mcap_phi)
    if(loper) then
       if(lcorr) then
          deallocate(hpsi,hteil,dtpsi)
       else
          deallocate(psi01,hunpsi)
       endif
    endif

    if(lproj .and. ipsi.eq.ipsi1) then
       close(ipsi,status='delete') ! Delete the reduced WF file.
    else
       close(ipsi,status='keep')
    endif
    return

 77 continue
    write(6   ,'(a)') ' flx: EOF found when reading the psi1 from file!'
    write(ilog,'(a)') ' flx: EOF found when reading the psi1 from file!'
    stop
 79 continue
    write(6   ,'(a)') ' flx with lopsym: ',&
                   'EOF found when reading the psi1 again from file!'
    write(ilog,'(a)') ' flx with lopsym: ',&
                   'EOF found when reading the psi1 again from file!'
    stop
    end subroutine flx

!#######################################################################
    end program flux
