!***********************************************************************
!
!                         gwptransmission program
!
! This computes gwp wave functions, and prints gwp center positions and
! momenta. It also computes the energy and transmission spectrum.
!
!
!***********************************************************************
      
     program gwptransmission

use decimal, only: long,dop,sip
use global
use globmemmod
use versions
use aglobal
use memdimmod
use dvrdatmod
use griddatmod
use operdef
use psidef
use channels
use datenmod
use compdatmod
use compdat1mod

use iofile, only: rdpsiinfo
use rdopermod, only: operinfo
use zeigauswmod, only: zeigausw
use rddvrmod
use memhopsmod, only: memhops
use logdat
use keyunitsmod

      implicit none

      integer(long)     :: chkdvr,chkgrd,chkpsi,chkdat,ilbl 
      character(len=c5) :: filename
      character(len=10) :: stat
      logical(kind=4)   :: lerr
      logical(kind=4)   :: lnumeric

real(dop),allocatable                      :: xplotlims(:,:)
integer(long)                              :: GWPprint

      real(dop)               ::  Emin,Emax,Estep,xlim

!-----------------------------------------------------------------------
! Initialize variables and set defaults
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.true.

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
! lsphfbr : jfbr, mfbr and kinfbr matrices.
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.false.
      ldmat1=.false.
      lort=.false.
      lgpop=.false.
      ltrafo=.false.
      lsphfbr=.false.

!-----------------------------------------------------------------------
! read input parameters and options
! open files
!-----------------------------------------------------------------------
      call GTinput(stat,Emin,Emax,Estep,GWPprint,xlim)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! open input and output data files
!
! use standard channels if possible
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------
      filename=filein
      open(ipsi,file=filename,form='unformatted',status='old',err=1000)

!-----------------------------------------------------------------------
! get memory sizes
!-----------------------------------------------------------------------
      call rdmemdim(ipsi)
      if (lrdoper) then
         filename=name(1:laenge)//'/oper'
         open(ioper,file=filename,form='unformatted',status='old',err=1000)
         call rdmemdim(ioper)
         close(ioper)
      endif

!-----------------------------------------------------------------------
! Allocate memory
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_psidef
      call alloc_operdef

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      filename=filein
      read(ipsi,err=999) filever(ipsi)
      call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)

      if (.not. lgwp) then
         routine = 'Gwptransmission'
         message='No GWP mode present'
         call errormsg
      endif

!-----------------------------------------------------------------------
! read information from system files
!-----------------------------------------------------------------------
      if (lrddvr) then
         chkdvr=2
         filename=dname(1:dlaenge)//'/dvr'
         open(idvr,file=filename,form='unformatted',status='old',&
              err=1000)
         call dvrinfo(lerr,chkdvr)
         close(idvr)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
      if (lrdoper) then
         chkdvr=2
         chkgrd=2
         filename=oname(1:olaenge)//'/oper'
         open(ioper,file=filename,form='unformatted',status='old',&
              err=1000)
         call operinfo(lerr,chkdvr,chkgrd)
         close(ioper)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
         
!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! assign work arrays
! NB work arrays must be at end of mc, mr, mi, ml arrays.
!-----------------------------------------------------------------------
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If oper file read, comment in routine to assign memory for operator
!-----------------------------------------------------------------------
      if (lrdoper) call memhops

!-----------------------------------------------------------------------
! call analysis routine
!-----------------------------------------------------------------------

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(gfgwp))deallocate(gfgwp)
if(allocated(aigwp))deallocate(aigwp)
if(allocated(zgwp))deallocate(zgwp)
if(allocated(gs2))deallocate(gs2)
if(allocated(dicht2))deallocate(dicht2)
if(allocated(agmat))deallocate(agmat)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(gfgwp(gfgwpdim))
allocate(aigwp(aigwpdim))
allocate(zgwp(zgwpdim))
allocate(gs2(gs2matdim))
allocate(dicht2(dmatdim))
allocate(agmat(agmatdim))

      call GTsubanalyse(mc(mcpsi),Emin,Emax,Estep,xplotlims,GWPprint,xlim)

deallocate(mc,mr,mi,ml,ms,mf,gfgwp,aigwp,zgwp,gs2,dicht2,agmat)
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_psidef

      stop
  999 continue
      message= 'Error reading :'//filename
      write(6,'(a)') message
      stop
 1000 continue
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop

      contains

!***********************************************************************
!
!     HERE IS THE SUBROUTINE TO DO THE ANALYSIS
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! change name of subroutine (or else mutil gets confused)
! provide dummy arguments of arrays needed for analysis
!  e.g. psi to match argument mc(mcpsi)
!***********************************************************************

      subroutine GTsubanalyse (psi,Emin,Emax,Estep,xplotlims,GWPprint,xlim)

use global
use aglobal
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

use rddvrmod, only: rddvr
use rdopermod, only: rdoper
use gwplib, only: gwp_cmom,gwp_ovl,gwp_param
use rddvrmod, only: rddvr
use rdfilesmod, only: rdpsi
use psiutil, only: getweight

      implicit none

      integer(long)                :: chkdvr,chkgrd
      integer(long), dimension(1)  :: idum
      real(dop), dimension(1)      :: rdum
      complex(dop), dimension(1)   :: zdum
      complex(dop), dimension(dgldim),target,intent(inout) :: psi



      ! real(dop), dimension(maxdata1)                      :: pltime
      real(dop), dimension(ortdim)           :: ort,weight
      real(dop), dimension(dvrdim)           :: trafo


      integer(long), dimension(gwpinfodim) :: trajst
      integer(long), dimension(adim)       :: jindx
      integer(long), dimension(gwpinfodim) :: gwpdep
      complex(sip), dimension(spsidim)  :: spsi
      complex(dop), dimension(gsmatdim) :: adgwp


      real(dop), dimension(dgldim)          :: qcenter,pcenter

      complex(dop), allocatable :: psiE(:,:),psiE0(:)
      real(dop),intent(in)               ::  Emin,Emax,Estep
      real(dop),intent(in)               :: xplotlims(:,:)
      integer(long),intent(in)           :: GWPprint
      real(dop)                          ::  aE
      integer(long)                      ::  Eind,dimE,Ei0
      complex(dop)                       :: norm,gnorm,gtrm,trm,Eexp


      integer(long)                      :: i,ndata,nplots
      real(dop)                                           :: time
      real(dop),intent(in) :: xlim

      integer(long)                      :: m,s

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! read data from dvr file if desired
! edit which data (set dvrdata(x) to .true. and substitute dummy array)
!-----------------------------------------------------------------------
      if (lrddvr) then
         dvrdata(1) = .true.    ! read ort
         dvrdata(2) = .true.    ! read trafo
         dvrdata(3) = .false.    ! read dvrmat (dif1mat, dif2mat)
         dvrdata(4) = .false.    ! read fftp   (fft momentum)
         dvrdata(5) = .false.    ! read hin    (fft)
         dvrdata(6) = .false.    ! read rueck  (fft)
         dvrdata(7) = .false.    ! read fftfak
         dvrdata(8) = .false.    ! read jsph
         dvrdata(9) = .false.    ! read msph
         dvrdata(10)= .false.    ! read kinsph
         dvrdata(11)= .false.    ! read exphin   (expDVR)
         dvrdata(12)= .false.    ! read exprueck (expDVR)
         chkdvr=2
         call rddvr(ort,ort,rdum,rdum,zdum,zdum,&
                    idum,zdum,zdum,idum,idum,rdum,chkdvr)
!-----------------------------------------------------------------------
! Determine the DVR weights.
!-----------------------------------------------------------------------
         call getweight(trafo,weight,ort,.true.)

      endif


      if (allocated(psiE)) deallocate(psiE)
      if (allocated(psiE0)) deallocate(psiE0)

      dimE=(Emax-Emin)/Estep+1
      allocate(psiE(dimE,2))
      allocate(psiE0(dimE))

      ndata=1
      time=0.0_dop
      nplots=1000
      rewind(ipsi)
      chkdvr=0
      chkgrd=0
      chkpsi=0
      chkdat=0
      read(ipsi) filever(ipsi)
      call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      do i = 1, nskip
         call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
         if (lend) then
            exit
         end if
         time=time+out2/fs
      enddo

      do while (.not. lend)
         call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
         if (lend) exit

         if (lgwp) then
            call gwp_param(psi,aGmat,gfgwp,aigwp,zgwp,qcenter,1)
            call gwp_cmom(psi,pcenter)
            call gwp_ovl(gs2,gfgwp)
            call getnorm(gs2(gs2mat(1,1)),dim(1,1),nspfdof(1),psi(zpsi(1)),psi(zxi(1,1)),&
                 agmat(zagmat(1,1)),gnorm,gtrm,xlim)
            do i=1,qcentdim
               write(idump,*) qcenter(i),ABS(psi(zpsi(1)+i-1)),dimag(psi(zpsi(1)+i-1)),&
                    dimag(psi(zxi(1,1)+i-1))
            enddo
            write(idump,*)
            psiE(:,:)=0.0_dop
            do m=1,nmode
               if (gwpm(m)) then
                  do s=1,nstate
                     call getpsiE1(nspfdof(m),m,dim(m,s),dimE,psiE,psi(zxi(m,s)),agmat(zagmat(m,s)),&
                          agmat(zphase(m,s)),psi(zpsi(s)),Emax,Emin,Estep,time,xlim,xplotlims,GWPprint)
                  end do        !state
               end if           !gwpm
            end do              !mode
            if (ndata==1) then
               psiE0(:)=psiE(:,1)
            end if
            Ei0=2
            Eind=(0-Emin)/Estep+3
            aE=Estep
            norm=0
            Eexp=0
            trm=0
            do while ((aE<=Emax) .and. (Eind<=dimE))
               if (abs(aE) > Estep/10) then
                  if (Eind-Ei0>=1) then
                     psiE(Eind,1)=sqrt(dconjg(psiE(Eind,1))*psiE(Eind,1)+dconjg(psiE(Eind-Ei0,1))*psiE(Eind-Ei0,1))
                     psiE(Eind,2)=sqrt(dconjg(psiE(Eind,2))*psiE(Eind,2)+dconjg(psiE(Eind-Ei0,2))*psiE(Eind-Ei0,2))
                     if (ndata==1) then
                        psiE0(Eind)=sqrt(dconjg(psiE0(Eind))*psiE0(Eind)+dconjg(psiE0(Eind-Ei0))*psiE0(Eind-Ei0))
                     end if
                  end if
                  write(idump,'(4g17.8e3)')aE,abs(psiE(Eind,1)),abs(psiE(Eind,2)),&
                       abs(psiE0(Eind))
                  norm=norm+dconjg(psiE(Eind,1))*psiE(Eind,1)*Estep
                  trm=trm+dconjg(psiE(Eind,2))*psiE(Eind,2)*Estep
                  Eexp=Eexp+dconjg(psiE(Eind,1))*psiE(Eind,1)*Estep*aE
               end if
               Eind=Eind+1
               Ei0=Ei0+2
               aE=aE+Estep
            end do !E
            write(idump,*)
            write(iaus,"(A1,g13.4,5g14.7)") &
              "N",time,dble(norm),dble(trm),dble(Eexp),dble(gnorm),dble(gtrm)


         endif

         do i = 1,step-1
            call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
            if (lend) then
               exit
            end if
            time=time+out2/fs
         enddo
         time=time+out2/fs
         ndata=ndata+1
         if( ndata.gt.nplots ) then
            exit
         end if
      end do
      

      return
      end subroutine GTsubanalyse

      subroutine getnorm(gs2,dim1,gdof,psi,xigmat,agmat,n,t,xlim)
        integer(long),intent(in) :: dim1,gdof
        complex(dop),dimension(dim1,dim1),intent(in) :: gs2
        complex(dop),dimension(dim1),intent(in) :: psi
        complex(dop),dimension(gdof,dim1),intent(in) :: xigmat
        complex(dop),dimension(gdof,gdof,dim1),intent(in) :: agmat
        complex(dop),intent(out) :: n,t

        real(dop),intent(in) :: xlim
        integer(long) :: e1,e2 

        n=0
        t=0
        do e1=1,dim1
           do e2=1,dim1
              n=n+dconjg(psi(e1))*gs2(e1,e2)*psi(e2)
              if ((dble(xigmat(1,e1)/(-2.0_dop*dble(agmat(1,1,e1)))) >xlim) .and. &
                   (dble(xigmat(1,e2)/(-2.0_dop*dble(agmat(1,1,e2)))) >xlim))then
                 t=t+dconjg(psi(e1))*gs2(e1,e2)*psi(e2)
              end if
           end do
        end do
      end subroutine getnorm


      ! compute the transformation of the wave function to energy representation
      ! psiE contains both the full transform and the transform of the transmitted part beyond xlim
      subroutine getpsiE1(gdof,m,dim1,dimE,psiE,xigmat,agmat,phigmat,psi,Emax,Emin,&
                 Estep,time,xlim,xplotlims,GWPprint)

        use errorf, only : dcerfc
        use gwplib, only: printGWPSPF0

        integer(long), intent(in)                 :: gdof,dim1,dimE,m
        complex(dop), dimension(dimE,2),intent(inout)  :: psiE
        real(dop),intent(in)                      ::  Emin,Emax,Estep

        complex(dop), dimension(dim1),intent(in)  :: psi
        complex(dop), dimension(dim1),intent(in)        :: phigmat
        complex(dop), dimension(gdof,dim1),intent(in)      :: xigmat
        complex(dop), dimension(gdof,gdof,dim1),intent(in)    :: aGmat
        real(dop),intent(in) :: xlim

        integer(long)                      :: f,e,Eind
        real(dop),intent(in)               :: time
        real(dop)                          ::  aE,daE
        complex(dop)                       :: allPsiE,alltPsiE,aPsiE,ptrans

        complex(dop), external                      :: fgaover1

        real(dop),dimension(gdof)           :: x
        real(dop),intent(in)                :: xplotlims(:,:)
        integer(long),intent(in)            :: GWPprint



        stop 'Check the xi argument to this call!'
        call printGWPSPF0(GWPprint,gdof,dim1,psi,phigmat,xigmat,agmat,&
             xplotlims=xplotlims,channel=iaus)
        
        do e=1,dim1
           do f=1,gdof
              Eind=1
              aE=Emin
              do while ((aE<=Emax) .and. (Eind <=dimE))

                 ! differential
                 daE= sqrt(sqrt(mass(spfdof(f,m))/(2.0_dop*dabs(aE))))/sqrt(2.0_dop*pi)

                 ! don't compute contribution of the aE==0 element
                 ! this would result in either NaN or a HUGE unphysical contribution
                 allPsiE=0.0_dop
                 alltPsiE=0.0_dop
                 if (abs(aE) > Estep/10.0_dop) then
                    ! this is the (sign corrected) momentum corresponding to aE plus the GWP "momentum"
                    ptrans=xigmat(f,e)-dcmplx(0.0_dop,1.0_dop)*sign(sqrt(2.0_dop*mass(spfdof(f,m))*dabs(aE)),aE)
                    if (lnumeric) then
                       x(f)=xplotlims(f,1)
                       do while (x(f)<=xplotlims(f,2))
                          aPsiE=exp(agmat(f,f,e)*x(f)**2+ptrans*x(f)+phigmat(e))
                          allPsiE=allPsiE+apsiE
                          if (x(f)>xlim) then
                             alltPsiE=alltPsiE+aPsiE
                          end if
                          x(f)=x(f)+xplotlims(f,3)
                       end do
                       allPsiE=allPsiE*xplotlims(f,3)*psi(e)*daE
                       alltPsiE=alltPsiE*xplotlims(f,3)*psi(e)*daE
                    else
                       allPsiE=psi(e)*sqrt(-pi/agmat(f,f,e))*&
                            exp(-ptrans**2/(4.0_dop*agmat(f,f,e))+phigmat(e))*daE

                       aPsiE=sqrt(-agmat(f,f,e))*xlim-ptrans/(2.0_dop*sqrt(-agmat(f,f,e)))
                       call dcerfc(1,aPsiE,aPsiE)
                       ! call cerror(sqrt(-agmat(f,f,e))*xlim-ptrans/(2.0_dop*sqrt(-agmat(f,f,e))),aPsiE)

                       alltPsiE=aPsiE*exp(ptrans**2/(4.0_dop*(-agmat(f,f,e)))+phigmat(e))/&
                            sqrt(-agmat(f,f,e))*sqrt(pi)/2.0_dop*psi(e)*daE
                       ! this check is necessary because there may be overruns in dcerfc due to a huge
                       ! contribution that would ideally be compensated by the exp(ptrans**2) above
                       ! the result would be almost zero anyway -> cutoff
                       ! the negation is used to capute NaNs
                       if (.not. (abs(alltPsiE) > tiny)) then
                          alltPsiE=0.0_dop
                       end if

                       ! this is a REALLY bad approximation, values keep jumping way too much as GWPs
                       ! pass beyond the limit
                       ! if (dble(xigmat(f,e)/(-2.0_dop*dble(agmat(f,f,e)))) >xlim) then
                       !    ! if the Gaussian's center coordinate is larger than the limit
                       !    ! its contribution is added to psiE(:,2)
                       !    alltPsiE=allPsiE
                       ! end if
                    end if
                 end if

                 psiE(Eind,1)=psiE(Eind,1)+allPsiE
                 psiE(Eind,2)=psiE(Eind,2)+alltPsiE
                 Eind=Eind+1
                 aE=aE+Estep
              end do !E
              write(iaus,'(A1,3g16.8)') &
                   'E',time,dimag(xigmat(f,e))**2/(2.0_dop*mass(spfdof(f,m))),&
                   dble(dconjg(psi(e))*psi(e))
           end do  !gdof
        end do     !dim
        write(iaus,'(A1)') 'E'
      end subroutine getpsiE1

function readset(buf,argcount,set)

  character(len=80),intent(in)             :: buf
  integer(long),intent(in) :: argcount
  real(dop),intent(inout) :: set(:)
  integer(long) :: readset

  integer(long) :: j,k,ierr
  
  j=1
  readset=0
  do k=1,argcount-1
     read(buf(j:j+index(buf(j:len(buf)),',')-1),*,iostat=ierr) set(k)
     if (ierr .ne. 0) then
        readset=ierr
        return
     end if
     j=j+index(buf(j:len(buf)),',')
  end do
  read(buf(j:len(buf)),*,iostat=ierr) set(argcount)
  if (ierr .ne. 0) then
     readset=ierr
     return
  end if
end function readset

!-----------------------------------------------------------------------
!***********************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! change name of subroutine (or else mutil gets confused)
!-----------------------------------------------------------------------
      subroutine GTinput(stat,Emin,Emax,Estep,GWPprint,xlim)

use global
use aglobal
use griddatmod
use psidef
use channels
use datenmod
use compdatmod
use compdat1mod

      implicit none

      integer(long)                 :: iarg,myiargc,i,j,next,k,argcount,ierr,iargc
      character(len=80)             :: buf
      character(len=*), intent(out) :: stat
      real(dop),intent(inout)               ::  Emin,Emax,Estep,xlim
      ! real(dop),intent(inout),allocatable   :: xplotlims(:,:)
      integer(long),intent(inout)               ::  GWPprint


!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! provide default input and output filenames 
!-----------------------------------------------------------------------
      filein = ' '
      fileout= 'no'
      name   = './'

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! set default flags
!
! stat : status of output file   
!-----------------------------------------------------------------------
      stat = 'new'
      lgnu = .false.

      xlim=0.9_dop
      Emin=0.0_dop
      Emax=1.0_dop
      lnumeric=.false.
      Estep=(Emax-Emin)/1000.0_dop
      GWPprint=1000
      call keyeunits(Emin,'ev',iarg)
      call keyeunits(Emax,'ev',iarg)
      call keyeunits(Estep,'ev',iarg)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! read options 
!-----------------------------------------------------------------------
      iargc = myiargc() 

      i=1
      do while (i<=iargc)
         call mygetarg(i,buf)
         if( buf(1:1) .eq. '-' ) then
            if( buf .eq. '-f') then
               i=i+1
               call mygetarg(i,buf)
               filein = buf
            else if(buf .eq. '-i') then
               i=i+1
               call mygetarg(i,buf)
               name = buf
            else if( buf .eq. '-inter' ) then
               linter=.true.
            else if( buf .eq. '-numeric' ) then
               lnumeric=.true.
            else if( buf .eq. '-ver' ) then
               call wrversion(6)
               stop

            else if( buf(1:2) .eq. '-o' ) then
               i = i+1
               call mygetarg(i,buf)
               fileout = buf
            else if( buf(1:2) .eq. '-e' ) then
               i = i+1
               call mygetarg(i,buf)
               read(buf,*,iostat=ierr) Emin
               if (ierr .ne. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -E',&
                       ' must be real. Not : ', buf
                  stop
               end if
               i = i+1
               call mygetarg(i,buf)
               call keyeunits(Emin,buf,iarg)
               if (iarg /= 1) then
                  i = i+1
                  call mygetarg(i,buf)
               end if

               read(buf,*,iostat=ierr) Emax
               if (ierr .ne. 0  .or.  (Emax < 0)) then
                  write(6,'(2a/,a)')&
                       ' The second argument following the option -E',&
                       ' must be a positive real. Not : ', buf
                  stop
               end if
               i = i+1
               call mygetarg(i,buf)
               call keyeunits(Emax,buf,iarg)
               if (iarg /= 1) then
                  i = i+1
                  call mygetarg(i,buf)
               end if

               read(buf,*,iostat=ierr) Estep
               if (ierr .ne. 0  .or.  (Estep < 0)) then
                  i = i-1
               else
                  i = i+1
                  call mygetarg(i,buf)
                  call keyeunits(Estep,buf,iarg)
                  if (iarg == 1) then
                     i = i-1
                  end if
               end if

            else if( buf(1:2) .eq. '-p' ) then
               i = i+1
               call mygetarg(i,buf)
               read(buf,*,iostat=ierr) GWPprint
               if (ierr .ne. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -p',&
                       ' must be an integer. Not : ', buf
                  stop
               end if
            else if( buf(1:2) .eq. '-l' ) then
               i = i+1
               call mygetarg(i,buf)
               read(buf,*,iostat=ierr) xlim
               if (ierr .ne. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -l',&
                       ' must be a real. Not : ', buf
                  stop
               end if
            else if( buf(1:2) .eq. '-x' ) then
               i = i+1
               call mygetarg(i,buf)
               argcount=0
               j=1
               next=1
               do while (next>0)
                  next=index(buf(j:len(buf)),',')
                  j=j+next
                  argcount=argcount+1
               end do
               allocate(xplotlims(argcount,3))
               j=1
               ierr=readset(buf,argcount,xplotlims(:,1))
               if (ierr .ne. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -x',&
                       ' must be a comma-seperated set of real numbers. Not : ', buf
                  stop
               end if
               i = i+1
               call mygetarg(i,buf)
               do k=1,argcount
                  call keyeunits(xplotlims(k,1),buf,iarg)
               end do
               if (iarg /= 1) then
                  i = i+1
                  call mygetarg(i,buf)
               end if

               ierr=readset(buf,argcount,xplotlims(:,2))
               if (ierr .ne. 0 ) then
                  write(6,'(2a/,a)')&
                       ' The second argument following the option -x',&
                       ' must be a comma-seperated set of real numbers. Not : ', buf
                  stop
               end if
               i = i+1
               call mygetarg(i,buf)
               do k=1,argcount
                  call keyeunits(xplotlims(k,2),buf,iarg)
               end do
               if (iarg /= 1) then
                  i = i+1
                  call mygetarg(i,buf)
               end if

               ierr=readset(buf,argcount,xplotlims(:,3))
               if (ierr .ne. 0) then
                  i = i-1
                  xplotlims(:,3)=0.0_dop
               else
                  i = i+1
                  call mygetarg(i,buf)
                  do k=1,argcount
                     call keyeunits(xplotlims(k,3),buf,iarg)
                  end do
                  if (iarg == 1) then
                     i = i-2
                  end if
               end if

            else if( buf(1:2) .eq. '-n' ) then
               i = i+1
               call mygetarg(i,buf)
               read(buf,*,iostat=ierr) step
               if (ierr .ne. 0  .or.  step .le. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -n',&
                       ' must be a positive integer. Not : ', buf
                  stop
               end if
            else if( buf(1:2) .eq. '-s' ) then
               i = i+1
               call mygetarg(i,buf)
               read(buf,*,iostat=ierr) nskip
               if (ierr .ne. 0  .or.  nskip .le. 0) then
                  write(6,'(2a/,a)')&
                       ' The  argument following the option -s',&
                       ' must be a positive integer. Not : ', buf
                  stop
               end if
            else if( buf(1:2) .eq. '-w' ) then
               lovwr=.true.
            else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
               write(6,'(78a1)')  ('-',i=1,78)
               write(6,'(1x,a,/10x,a,/1x,2a)')&
                    'Purpose: Prints gwp center positions and momenta. ',&
                             'Also computes the energy and transmission spectrum.',&
                    'Usage  : gwptransmission',&
                    '[-f -i -inter -ver -o -n -p -x -w -h -? -e -l -numeric]'
               write(6,'(/,a)') 'Options : '
               write(6,'(2a,/,11x,a/11x,a)')&
                    ' -f FILE : the data is read '&
                    ,'from file FILE','rather than from ./auto'&
                    ,'The string FILE may be a relative or full path-name.'
               write(6,'(a)') ' -i DIR : data is stored in directory DIR'
               write(6,'(a)') ' -inter : enables interactive plotting'
               write(6,'(a)') ' -numeric : enables numeric Energy integrals'
               write(6,'(a)')&
                    ' -ver    : Version information about the program '
               write(6,'(2a,/,11x,a/,11x,a)') &
                    ' -o FILE : The difference is written to file ',&
                    'FILE.pl rather than to ./XXX',&
                    'The string FILE may be a relative or a full path-name.',&
                    'If FILE=no, i.e. "-o no", no XXX-file will be opened.'
               write(6,'(2a)') ' -n I : Compute the property', &
                    ' only every I-th output.'
               write(6,'(2a,/11x,a)') ' -p I : Compute I', &
                    ' points in position space.', &
                    'If I<0, only plot the reduced densities with |I| points each.'
               write(6,'(2a,/11x,a,/11x,2a)') ' -x R1[,R1..] [unit]',&
                    'R2[,R2..] [unit] [R3[,R3..] [unit]]:', &
                    'Specify position space limits for the plot.',&
                    'If the set R3 is given, ignore the value of the -p',&
                    ' parameter and use R3 as a step size.'
               write(6,'(2a)') ' -l limit : limit after which GWP is considered', &
                    ' to be transmitted.'
               write(6,'(a,/11x,a,2a)') ' -e R1 [unit] R2 [unit] [R3 [unit]]:', &
                    'Compute energy distribution from R1 to R2 with ',&
                    'R3 equidistant steps.'
               write(6,'(2a)') &
                    ' -w      : An existing property file is ',&
                    'overwritten.'
               write(6,'(a,/a)') &
                    ' -h      : Print this help text. ',&
                    ' -?      : Print this help text. '
               write(6,'(78a1)')  ('-',i=1,78)
               stop

            else 
               write(6,'(2a)') ' unknown option : ', buf
               stop
            end if
         end if
         i = i+1
      end do
      
!-----------------------------------------------------------------------
! Open log and output file channels
!-----------------------------------------------------------------------
      open (unit=ilog,file='GWPtrans.log',form='formatted')
      logisopen=.true.
      open (iaus,file='gwptrans', form='formatted')

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If filein has not been set, look in name directory.
! Set XXX to default name
! Otherwise take name from filein
!-----------------------------------------------------------------------
      if (filein .eq. ' ') then
         laenge=index(name,' ')-1
         call abspath(name,laenge)
         filein=name(1:laenge)//'/psi '
         laein=index(filein,' ')-1
      else
         laein=index(filein,' ')-1
         call abspath(filein,laein)

         i=laein
         do while (filein(i:i) /= '/')
            i=i-1
         end do
         name=filein(1:i-1)
      endif

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! get any compulsory arguments if in argument list
!-----------------------------------------------------------------------
!     call mygetarg(i,filein1)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! fill in any arguments not given on command line
!-----------------------------------------------------------------------
!     if (filein1 .eq. ' ') then
!        write(6,*) 'XXX   :'
!        read(5,'(a)') filein1
!        write(6,*)
!     endif

!-----------------------------------------------------------------------
! set status variable to overwrite files
!-----------------------------------------------------------------------
      if (lovwr) stat='unknown'

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! convert filenames to absolute
!-----------------------------------------------------------------------
      if (fileout .ne. 'no') then
         laeout=index(fileout,'  ')-1
         call abspath(fileout,laeout)
      end if


      write(6,'(30A)')filein

      return 
      end subroutine GTinput

!#######################################################################


      end program gwptransmission
