!--------------------------------------------------------------------------
!            JINPOL
!
!  Program performs J-interpolation through the following procedure:
!
!  ( P1=P(J); P2=P(J+N) )
!  1. Calculates A(D)=Intrgral[ P1(E-D)*P1(E+D)dE, {E,Emin,Emax} ]
!     Calculates B(D)=Intrgral[ P2(E-D)*P2(E+D)dE, {E,Emin,Emax} ]
!     Calculates C(D)=Intrgral[ P1(E-D)*P2(E+D)dE, {E,Emin,Emax} ]
!  2. Searches for minimum of F(D) = SQRT(A(D)*B(D)) - C(D) 
!  3. The result: P(J+m)=(1-b)*P1(E-E1) + b*P2(E+E2)
!     where: E1=a*((J+m)*(J+m+1)-J*(J+1))
!            E2=a*((J+N)*(J+N+1)-(J+m)*(J+m+1))
!            a=2*D/((J+N)*(J+N+1)-J*(J+1))
!            a=1/(2*m*R0*R0)    
!            b=m/N
! 
!  See J.Chem.Phys. 116 (2002) 10641 for a description of the algorithm.  
!  Note: 1. The probabilities will be read from files IDIR/NAME{jkJ} and
!           written to files ODIR/OUTFILE{jkJ}
!        2. The arg of -rf is the fixed value of R0 (no minimum will be
!           calculated)
!  Defaults: j0=0; k0=0; r0=1; IDIR=ODIR=./; mass=1 (1837.15); Edim=500
!
! SS and HDM 10/04
!--------------------------------------------------------------------------

      program jinpol

use versions

      implicit none


      integer, parameter :: long  = selected_int_kind(9) 
      integer, parameter :: dop = selected_real_kind(15,307) ! real*8 and complex*16

      integer(long), parameter :: maxendim = 2000,maxj = 99,&
                                  itmax = 50,ddim = 200
      real(dop), parameter    :: xacc = 1.0e-5_dop
      
      integer(long)       :: i,l,edim,j,k,j1,j2,lname,lename,loutfile,ldir,energdim0,&
                             energdim1,energdim2,del1,del2,it,imin,itype,iprint
      real(dop)           :: e1,e2,e,df1,df2,ee1,ee2,eed,emin,mxprb,delta,&
                             cint,d,dd,dmin,dmax,alpha,beta,en1,en2,p1,p2,p,&
                             r0,r0f,mass,r01,err,err1,fac,ff,eshift,pi,cintmin,&
                             alpha0,alphamin,alphamax,emin1,emin2

      real(dop), dimension(maxendim)       :: energ0,energ1,energ2,eearr,rprob1,&
                                              rprob2,aarr,barr,carr
      real(dop), dimension(ddim)           :: acint,ainteg,binteg,cinteg

      character(len=64)         :: name,outfile
      character(len=80)         :: idir,odir
      character(len=120)        :: infile,ename
      character(len=4)          :: fnumj1,fnumj2,fnumi,fnumj,fnumk
      logical(Kind=4)           :: lalpha

      common /jinpi/ j,j1,j2,k,edim,itype,iprint
      common /jinpr/ mass,r0,r0f,e1,e2,eshift,emin1,emin2,mxprb


      pi=4.0_dop*atan(1.0_dop)
      lalpha=.false.
      delta=0.0_dop

! -- Reads the parameters which differ from their default values
      call inpjin(idir,odir,name,ename,outfile,lalpha,maxendim,maxj)


! --- Reads the probabilities P1 and P2

      write(fnumj,'(i4)') j
      write(fnumk,'(i4)') k

      write(fnumj1,'(i4)') j1
      write(fnumj2,'(i4)') j2

      lname=index(name,' ')-1
      ldir=index(idir,' ')-1
      loutfile=index(outfile,' ')-1
      infile=idir(1:ldir)//'/'//name(1:lname)//fnumj//fnumk//fnumj1
      call rmblnk(infile)
      call rdprob(energ1,rprob1,maxendim,energdim1,infile)
      infile=idir(1:ldir)//'/'//name(1:lname)//fnumj//fnumk//fnumj2
      call rmblnk(infile) 
      call rdprob(energ2,rprob2,maxendim,energdim2,infile)

      lename=index(ename,' ')-1
      if(lename .eq. 0) then
         energdim0=energdim1
         do i=1,energdim1
            energ0(i)=energ1(i)
         enddo
      else
         call rdprob(energ0,aarr,maxendim,energdim0,ename)
      endif

! --- Calculates C(D)=integral[P1(E-D)*P2(E+D)dE,{E,E1,E2}]

      if (.not.lalpha) then

         fac=5.0_dop
         err=99.0_dop
         err1=99.0_dop

!........R0 must not be too small.
         if( (j2*(j2+1)-j1*(j1+1))/(2.0_dop*mass*r0*r0) .gt. (e2-e1) ) &
             r0 = sqrt( (j2*(j2+1)-j1*(j1+1))/(2.0_dop*mass*(e2-e1)) )

         if(iprint.gt.0) &
          write(6,'(i4,a,f17.12,20x,f8.3)') 0,' R0=',r0,fac

         do it = 1, itmax

            alpha0=1.0_dop/(2.0_dop*mass*r0*r0)
            alphamin=alpha0/fac
            alphamax=alpha0*fac
 
            dmin=alphamin*(j2*(j2+1)-j1*(j1+1))/2.0_dop
            dmax=alphamax*(j2*(j2+1)-j1*(j1+1))/2.0_dop
            dd=(dmax-dmin)/(ddim-1)

            cintmin=1.0e9_dop
            imin=0 
            do i=1,ddim
               d=dmin+(i-1)*dd
               ee1=e1
               ee2=e2
               eed=(ee2-ee1)/(edim-1)
               do l=1,edim
                  e=ee1+(l-1)*eed
                  call inpol(e-d,p1,energ1,rprob1,energdim1)
                  call inpol(e+d,p2,energ2,rprob2,energdim2)
                  carr(l)=p1*p2
                  aarr(l)=p1*p1
                  barr(l)=p2*p2
                  eearr(l)=e
               enddo

               call integr(cinteg(i),carr,eed,edim)
               call integr(ainteg(i),aarr,eed,edim)
               call integr(binteg(i),barr,eed,edim)

               if(itype.eq.1) then
                  cint=sqrt(ainteg(i)*binteg(i))-cinteg(i)
               elseif(itype.eq.2) then
                  cint=sqrt(ainteg(i)*binteg(i))/cinteg(i)
               elseif(itype.eq.3) then
                  cint=-cinteg(i)
               endif
               acint(i)=cint

!.....Sarch for minimum.
                if(cint .lt. cintmin) then
                   imin    = i
                   delta   = d
                   cintmin = cint
                endif
            enddo

            if(imin.eq.1) then
               delta = delta*2.0_dop/(fac+1.0_dop)
            elseif(imin.eq.ddim) then
               delta = delta*(fac+1.0_dop)/2.0_dop  
            else
               fac=(3.01_dop+fac)/4.0_dop
               df1 = ( acint(imin+1) - acint(imin-1) )/(2.0_dop*dd)
               df2 = ( acint(imin-1) + acint(imin+1) -&
                       acint(imin)*2.0_dop )/(dd*dd)
               delta = delta - df1*df2/(df2*df2+1.0e-16_dop)
            endif
            alpha=2.0_dop*delta/dble(j2*(j2+1)-j1*(j1+1))
            r01=sqrt(1.0_dop/(2.0_dop*mass*alpha))
            err=abs(r0-r01)/(r0+r01)
            r0=r01
            if(iprint.gt.0) write(6,'(i4,a,f17.12,a,e11.3,f8.3,i4)')&
             it, ' R0=',r0,',  error=',err,fac,imin
            if(err+err1.lt.xacc .and. fac.lt.1.5_dop) goto 200
            err1=err
         enddo

         write(6,'(a,i3,a)')'Not converged after',itmax,' iterations!'
         stop

      else
         alpha=1.0_dop/(2.0_dop*mass*r0f*r0f)
         delta=0.0_dop
         imin=1
         ainteg(1)=1.0_dop
         binteg(1)=1.0_dop
         cinteg(1)=1.0_dop
      endif

 200  ff = 1.0_dop - cinteg(imin)/sqrt(ainteg(imin)*binteg(imin))
      write(6,'(a,f11.6,",",i3,a,f11.6,a,f10.6,a)') &
        'R0 =',sqrt(1.0_dop/(2.0_dop*mass*alpha)),it,&
        ' iterations.  Min*1000 =',ff*1.0e3_dop,', delta=',delta,'eV'



! --- Calculates the probabilities
      ldir=index(odir,' ')-1
      do i=j1,j2

         write(fnumi,'(i4)') i

         infile=odir(1:ldir)//'/'&
              //outfile(1:loutfile)//fnumj//fnumk//fnumi
         call rmblnk(infile)
         open(unit=16,file=infile)


         del1=i*(i+1)-j1*(j1+1)
         del2=j2*(j2+1)-i*(i+1)
!        beta=dble(i-j1)/(j2-j1)
         beta=dble(del1)/dble(del1+del2)
         do l=1,energdim1
            e=energ0(l)
            en1=e-alpha*del1
            en2=e+alpha*del2
            call inpol(en1,p1,energ1,rprob1,energdim1)
            call inpol(en2,p2,energ2,rprob2,energdim2)
            if(e   .lt. energ1(1) ) p1=0.d0
            if(en2 .lt. energ2(2) ) p2=0.d0
            p=(1.0_dop-beta)*p1 + beta*p2
!            p=0.5_dop*(1._dop+cos(pi*beta))*p1 +
!     +        0.5_dop*(1._dop-cos(pi*beta))*p2
            emin=(1.0_dop-beta)*emin1 + beta*emin2
            if(e.lt.emin) p=0.0_dop 
            write(16,'(2f20.12)') e,p
         enddo
         close(16)
      enddo

      stop
      contains

!--------------------------------------------------------------------------

      subroutine integr(s,yarr,xd,dim)

!  ---------- Linear integration -----------

      implicit none

      integer(long)                         :: i
      integer(long), intent(in)             :: dim
      real(dop), intent(in)                 :: xd
      real(dop), intent(out)                :: s
      real(dop), dimension(dim), intent(in) :: yarr

      s = 0.5_dop*(yarr(1)+yarr(dim))
      do i = 2, dim-1
         s = s + yarr(i)
      enddo 

      s = s*xd

      return
      end subroutine integr

!--------------------------------------------------------------------------

      subroutine inpol(x,y,xarr,yarr,dim)

!  ---------- Linear interpolation -----------

      implicit none

      integer(long)                         :: n
      integer(long), intent(in)             :: dim
      real(dop)                             :: x1,x2,y1,y2,xd,a
      real(dop), intent(in)                 :: x
      real(dop), intent(out)                :: y
      real(dop), dimension(dim), intent(in) :: xarr,yarr

      if (x.lt.xarr(1)) then
         y=yarr(1)
         return
      elseif (x.gt.xarr(dim)) then
         y=yarr(dim) 
         return   
      endif

      xd=(xarr(dim)-xarr(1))/(dim-1)
      n=(x-xarr(1))/xd + 1

      x1=xarr(n)
      x2=xarr(n+1)
      y1=yarr(n)
      y2=yarr(n+1)

      a=(y2-y1)/(x2-x1)

      y=y1+a*(x-x1) 

      return
      end subroutine inpol

!--------------------------------------------------------------------------

      subroutine rdprob(energ,rprob,maxendim,energdim,infile)

      implicit none

      integer(long)                               :: n
      integer(long), intent(in)                   :: maxendim
      integer(long), intent(out)                  :: energdim
      real(dop)                                   :: mass,r0,r0f,e1,e2,eshift, &
                                                     emin1,emin2,mxprb,ee,prb,&
                                                     dum,sump
      real(dop), dimension(maxendim), intent(out) :: energ,rprob
      character(len=24)                           :: line
      character(len=24), intent(in)               :: infile

      common /jinpr/ mass,r0,r0f,e1,e2,eshift,emin1,emin2,mxprb


      open(unit=15,file=infile,status='old',err=30)

      sump=0.0_dop
      do n=1,250
         read (15,'(a24)',end=5,err=20) line
         if(line(1:1) .eq. '#' .or. line .eq.&
                 '                        ') goto 4     
         read (15,'(f12.6,1x,2f20.9,3x,f20.12)',end=5,err=20)&
              ee,dum,dum,prb
         sump=sump+abs(prb)
         if(sump .gt. 0.0_dop) goto 5
 4       continue
      enddo

 5    rewind 15
      energdim=0
      if( sump .gt. 0.0_dop ) then  ! flux file
         do n=1,20
            read (15,'(a24)') line
            if(line(1:1) .ne. '#' .and. line .ne.&
                 '                        ') goto 6     
         enddo
         write(6,'(2a)') 'Cannot determine file type of ', infile
         stop
 6       backspace 15
          do n=1,maxendim
            read (15,*,end=10,err=25) ee,dum,dum,prb
            energdim=energdim+1
            energ(energdim)=ee+eshift
            rprob(energdim)=max(0.d0,min(prb,mxprb))
            goto 35
 25         backspace 15
            read (15,'(a24)',end=10,err=20) line
            if(line(1:1) .ne. '#' .and. line .ne. &
                 '                        ') goto 20
 35      enddo

      else                      ! rpr file
         do n=1,maxendim
            read (15,*,end=10,err=26) ee, prb
            energdim=energdim+1
            energ(energdim)=ee+eshift
            rprob(energdim)=max(0.d0,min(prb,mxprb))
            goto 36
 26         backspace 15
            read (15,'(a24)',end=10,err=20) line
            if(line(1:1) .ne. '#' .and. line .ne. &
                 '                        ') goto 20
 36      enddo
      endif
      write(6,'(a)') 'Error: Increase maxendim'
      stop
      
 10   close(15)
      return

 20   write(6,'(2a)') 'Error reading file ',infile
      stop

 30   write(6,'(2a)')'Error: Cannot open file for reading ',infile
      stop
   
      end subroutine rdprob

!--------------------------------------------------------------------------
!     inpjin reads the input from comand line

      subroutine inpjin(idir,odir,name,ename,outfile,lalpha,maxendim,maxj)

      implicit none

      integer(long)                  :: i,j,k,j1,j2,length,edim,itype,iprint
      integer(long), intent(in)      :: maxendim,maxj
      real(dop)                      :: e1,e2,r0,r0f,mass,eshift,emin1,emin2,mxprb
      character(len=64), intent(out) :: name, outfile
      character(len=120), intent(out) :: ename
      character(len=80)              :: buf
      character(len=80), intent(out) :: idir,odir
      logical(kind=4), intent(out)   :: lalpha

      common /jinpi/ j,j1,j2,k,edim,itype,iprint
      common /jinpr/ mass,r0,r0f,e1,e2,eshift,emin1,emin2,mxprb

!.....Set Defaults
      mass=1.0_dop
      r0=3.0_dop
      k=0
      j=0
      j1=0
      j2=0
      itype=2
      iprint=0
      edim=500
      e1=999.999e9_dop
      e2=999.999e9_dop
      eshift=0.0_dop 
      emin1=-1.0e9_dop
      emin2=-1.0e9_dop
      mxprb=1.0e9_dop

! .... here write your default paths ............
      idir='.'
      odir='.'
      name=' '
      outfile=' '

      i=0
 10   continue
      i=i+1
      call mygetarg(i,buf)

      if (buf(1:1).eq.' ') goto 100
      if (buf.eq.'-h') then
         write(6,'(/1x,a)') 'Purpose : J-interpolation of flux data.'
         write(6,'(1x,2a,/21x,a)') 'Usage   : ',&
         'jinpol [-d -f -o -j -j1 -j2 -m -r0 -rf -e1 -e2 -ed -es',&
                   '-em -mp -A -B -C -p -ver -h].'
         write(6,'(/,a)') ' Options : '
         write(6,'(2a)') ' -id IDIR: The probabilities files are ',&
                            'read from directory IDIR.'
         write(6,'(2a)') ' -od ODIR: The probabilities files are ',&
                            'written to directory ODIR.'
         write(6,'(2a,/,11x,a)') &
              ' -f FILE : The probabilities  are read ',&
              'from file IDIR/FILEext',&
              'where ext denote the extension jKJ (e.g. 205).'
         write(6,'(2a,/,11x,a)') &
              ' -o FILE : The interpolated probabilities are written ',&
              'to ODIR/FILEext',&
              'where ext denote the extension jKJ (e.g. 205).'
         write(6,'(2a,/,11x,a)') &
              ' -ef FILE: The energy points are taken from file FILE.', &
              ' If -ef is not given,', &
              'the energy points are taken from the first flux file.'
         write(6,'(2a)') &
            ' -j   j  : Search for files with j-value j.',&
                        ' j is the first number in extension.'
         write(6,'(a)') &
            ' -j1  J1 : First J value of interpolation range.'
         write(6,'(a)') &
            ' -j2  J2 : Last  J value of interpolation range.'
         write(6,'(2a)') &
            ' -k   K  : Search for files with K-value K.',&
                        ' K is the second number in extension.'
         write(6,'(2a)') &
              ' -r0  R0 : R0 is the estimated value of the ',&
              'transition-state coordinate.'
         write(6,'(2a,/11x,a)') &
              ' -rf  Rf : Rf is the fixed value of the ',&
              'transition-state coordinate.','(No optimization).'
         write(6,'(a,f4.1,a)') &
            ' -m mass : Reduced mass in AMU (default mass =',&
               mass,').'
         write(6,'(a)') &
            ' -e1  E1 : First energy of fit-interval.'
         write(6,'(a)') &
            ' -e2  E2 : Last  energy of fit-interval.'
         write(6,'(2a,i4,a)')&
            ' -ed edim: Number of energy points used to compute the', &
            ' integrals. (def.=',edim,').'
         write(6,'(2a)')&
            ' -es shft: The energies read from the probability files',&
            ' are shifted by shft.'
         write(6,'(2a)')&
            ' -em1 Emin1: The input probabilities for j1',&
            ' are set to zero for E<Emin1.'   
          write(6,'(2a)')&
            ' -em2 Emin2: The input probabilities for j2',&
            '  are set to zero for E<Emin2.'
         write(6,'(2a)')&
            ' -mp maxp: The input probabilities are set to maxp if',&
            ' prob>maxp.'   
         write(6,'(a)') ' -A      : Minimize F = SQRT(A*B) - C' 
         write(6,'(a)') ' -B      : Minimize F/C (default)' 
         write(6,'(a)') ' -C      : Minimize -C' 
         write(6,'(a)') ' -p      : Print every iteration step.' 
         write(6,'(a)')&
              ' -ver    : Version information about the program '
         write(6,'(a)') &
              ' -h      : Print this help text. '
         write(6,'(/2a,/2a,//2a,/2a,/2a,//2a,/2a)')&
          'The defaults settings are: IDIR=., ODIR=., j=0, k=0, ',&
          'r0=3.0, Edim=500,','Eshift=0.0, Emin=-1.0e9_dop, maxp=1.0e9_dop, ',&
          'mass=1.0 (i.e. 1837.15 au).',&
          'Example: jinpol -f flux -o mflux -j 2 -j1 1 -j2 4 ',&
          '-e1 0.0 -e2 1.2',    &
          'This will read the files flux201 and flux204 from and ',&
          'write the ','files  mflux201, mflux202, mflux203, ',&
          'and mflux204 to current directory.',&
          'NB: The energy interval [E1,E2] is used for determining r0',&
          ' only.','The energy interval of output is the one of ',&
          'the first flux file.'
         write(6,'(78a1)')  ('-',i=1,78)
         stop
      else if( buf(1:4) .eq. '-ver' ) then
         call wrversion(6)
         stop
      endif     
      if (buf(1:1).eq.'-') then
         if (buf(2:3) .eq. 'j ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) j
         elseif (buf(2:3) .eq. 'k ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) k
         elseif (buf(2:4) .eq. 'j1 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) j1
         elseif (buf(2:4) .eq. 'j2 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) j2
         elseif (buf(2:4) .eq. 'id ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) idir
         elseif (buf(2:4) .eq. 'od ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) odir
         elseif (buf(2:3) .eq. 'f ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) name
         elseif (buf(2:3) .eq. 'ef ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) ename
         elseif (buf(2:3) .eq. 'o ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) outfile
         elseif (buf(2:3) .eq. 'm ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) mass
         elseif (buf(2:4) .eq. 'r0 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) r0
         elseif (buf(2:4) .eq. 'rf ') then
            lalpha=.true.
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) r0f
         elseif (buf(2:4) .eq. 'e1 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) e1
         elseif (buf(2:4) .eq. 'e2 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) e2
         elseif (buf(2:4) .eq. 'ed ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) edim
         elseif (buf(2:5) .eq. 'em1 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) emin1
         elseif (buf(2:5) .eq. 'em2 ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) emin2 
         elseif (buf(2:4) .eq. 'mp ') then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) mxprb
         elseif (buf(2:3) .eq. 'A ') then
            itype=1
         elseif (buf(2:3) .eq. 'B ') then
            itype=2
         elseif (buf(2:3) .eq. 'C ') then
            itype=3
         elseif (buf(2:3) .eq. 'p ' ) then
            iprint=1
         elseif (buf(2:4) .eq. 'es ' ) then
            i=i+1
            call mygetarg(i,buf)
            length=index(buf,' ')-1
            read(buf(1:length),*) eshift
         else
            write(6,'(2a)') 'Unknown option tag on input: ',buf(1:32)
            stop
         endif
      endif 
  
      go to 10

 100  continue

! --- Checks the parameters

      if (j.lt.0 ) then
         write(6,'(a)')'Error: j must be >= 0'
         stop
      endif
      if (k.lt.0 .or. k.gt.j) then
         write(6,'(a)')'Warnig: k should be out of 0,...,j'
      endif
      if (k.gt.j1 .or. k.gt.j2) then
         write(6,'(a)')'Warnig: k should be .le. J1 and J2'
      endif
      if (j1.gt.maxj) then
         write(6,'(a)')'Error: J1 must be .le. ',maxj
         stop
      endif
      if (j2.gt.maxj) then
         write(6,'(a)')'Error: J2 must be .le. ',maxj
         stop
      endif
      if (j1.ge.j2) then
         write(6,'(a,2i4)')&
              'Error: J1<J2 is not satisfied, J1, J2 =',j1,j2
         stop
      endif
      if (j2.lt.2) then
         write(6,'(a)')'Error: J2 must be >= 2'
         stop
      endif

      if (edim.gt.maxendim) then
         write(6,'(a)')'Increase in program maxendim'
         stop
      endif
      if (abs(e1-999.999e9_dop) .lt. 1.0_dop) then 
         write(6,'(a)')'Error: e1 is not set'
         stop
      endif
      if (abs(e2-999.999e9_dop) .lt. 1.0_dop) then
         write(6,'(a)')'Error: e2 is not set'
         stop
      endif
      if (e1.ge.e2) then
         write(6,'(a,2f12.6)')&
              'Error: E1<E2 is not satisfied, E1, E2 =',e1,e2
         stop
      endif

      if (name .eq. ' ') then
         write(6,'(a)')'Error: NAME not set'
         stop
      endif
      if (outfile .eq. ' ') then
         write(6,'(a)')'Error: OUTFILE not set'
         stop
      endif
      if (r0.le.0.0_dop) then
         write(6,'(a)')'Error: R0 not set'
         stop
      endif
      
      mass=mass*1837.15_dop/27.2114_dop

      return
      end subroutine inpjin


      end program jinpol

