module oppsimod

      implicit none

      logical(kind=4), save                :: lsp

end module oppsimod

!***********************************************************************
!
!             Analyse program OPPSI
!
! This program reads the psi-file and a specified operator and
! computes and stores the new wavefunction |phi(t) > = operator|psi(t)>.
!
!***********************************************************************

      program oppsi

use decimal, only:long,dop,sip
use global
use globmemmod
use aglobal
use versions
use oppsimod
use memdimmod
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod
use operdef
use dirdyn
use gapmod
use hpsimod
use channels
use runpropmod

use op1lib, only: zerovxz
use logdat, only: errormsg,warnmsg
use iorst, only: rstinfo
use iofile, only: rdpsiinfo,wrpsiinfo
use rddvrmod
use rdopermod, only: operinfo,rdoper
use zeigauswmod, only: zeigausw
use memhopsmod, only: memhops
use zeighpsimod, only: operflag,zeighpsi,d2zeighpsi,d1zeighpsi
use psiutil, only: maxmatwk
use opwfmod, only: wkoperate,operate
use rdfilesmod, only: rdpsi
use tdkoeffs, only: calctdk

use module_interface

      implicit none

      integer(long), external              :: slen
      integer(long)                        :: chkdvr,chkgrd,chkpsi,chkdat,chkprp,ilbl,n,nham 
      character(len=c5)                    :: filename
      character(len=10)                    :: stat
      logical(kind=4)                      :: lerr,lexist,linwf

!-----------------------------------------------------------------------
! Initialize variables and set defaults
!-----------------------------------------------------------------------
      call default
      call adefault
      lsp = .true.

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.true.

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.true.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.true.

!-----------------------------------------------------------------------
! read input parameters and options
! open files
!-----------------------------------------------------------------------
      call oppinput(stat)

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! open input and output data files
!
! use standard channels if possible
!  ipsi : psi file
!  ipsi1 : output file
!-----------------------------------------------------------------------
      filename=filein
      open(ipsi,file=filename,form='unformatted',status='old',err=1000)
      filename=fileout
      open(ipsi1,file=filename,form='unformatted',status=stat,err=1000)

!-----------------------------------------------------------------------
! get memory sizes
!-----------------------------------------------------------------------
      call rdmemdim(ipsi)
      if (lrdoper) then
         filename=operfile
         ilbl=index(filename,' ')-1
         open(ioper,file=filename,form='unformatted',status='old',err=900)
         call rdmemdim(ioper)
         close(ioper)
      endif

!-----------------------------------------------------------------------
! Allocate memory
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_psidef
      call alloc_hpsimod

!-----------------------------------------------------------------------
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      chkprp=0
      filename=filein
      read(ipsi,err=999) filever(ipsi)
      call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      filename='psi'
      call chksyserr(filename,lerr)
      if (lerr) go to 999

!-----------------------------------------------------------------------
! Allocate extra space if DD or DD-GB.
!-----------------------------------------------------------------------
      if (ldd.or.lddmctdh) then
         call alloc_dercp
         call alloc_dirdyn
         call alloc_dddb
         call alloc_gapkernel
      endif

!-----------------------------------------------------------------------
! read information from system files
!-----------------------------------------------------------------------
      if (lrddvr) then
         chkdvr=2
         filename=dname(1:dlaenge)//'/dvr'
         open(idvr,file=filename,form='unformatted',status='old',&
              err=1000)
         call dvrinfo(lerr,chkdvr)
         close(idvr)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
      if (lrdoper) then
         chkdvr=2
         chkgrd=2
         filename=operfile
         open(ioper,file=filename,form='unformatted',status='old',&
              err=1000)
         call operinfo(lerr,chkdvr,chkgrd)
         close(ioper)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
         
!-----------------------------------------------------------------------
! set up new wavefunction file
!-----------------------------------------------------------------------
      lpsiopt(1)=lsp
      lpsiopt(2)=.false.
      lpsiopt(3)=.false.
      lpsiopt(4)=.false.

      write(ipsi1) progver
      call wrpsiinfo(ipsi1)

!-----------------------------------------------------------------------
! get operator name
!-----------------------------------------------------------------------
      if (operator .eq. ' ') then
         write(6,*) 'No operator specified! Use -h option to see usage.'
         write(6,'(1x,a)') 'Name of operator ? Choose from:'
         do n=1,hamdim
            write(6,'(2x,a)') hamlab(n)
         enddo
         read(5,'(a)') operator
         if (operator .eq. ' ') stop
      endif

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
 20   nham=0
      do n=1,hamdim
         if (hamlab(n) .eq. operator) nham=n
      enddo

      if (nham .eq. 0) then
         write(6,'(4a)')'Operator "',&
           operator(1:slen(operator)),'" not found in ',&
           operfile(1:slen(operfile))
         write(6,'(1x,a)') 'Name of operator ? Choose from:'
         do n=1,hamdim
            write(6,'(2x,a)') hamlab(n)
         enddo
         read(5,'(a)') operator
         if (operator .eq. ' ') stop
         goto 20
      endif

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

!-----------------------------------------------------------------------
! set up various flags and array information needed to use an operator
!-----------------------------------------------------------------------
      call operflag(nham)
      if (dentype.eq.0) then
         call zeighpsi(nham)
      else if (dentype.eq.1) then
         call d1zeighpsi(nham)
      else if (dentype.eq.2) then
         call d2zeighpsi(nham)
      endif
      
      mcpsi1=mcdim
      mcdim=mcpsi1+dgldim
      mchpsi=mcdim
      mcdim=mchpsi+hpsidim
      mchunpsi=mcdim
      mcdim=mchunpsi+totphidim
      mcintpsi=mcdim
      mcdim=mcdim+dgldim

      if (lgwp) then
         mrqcent = mrdim
         mrdim   = mrqcent + qcentdim + 1
      else
         mrqcent = mrdim
         mrdim   = mrqcent + 1
      endif                                              

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! assign work arrays
! NB work arrays must be at end of mc, mr, mi, ml arrays.
!-----------------------------------------------------------------------
      call maxmatwk(nham)
      call wkoperate

      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
!  assign memory for operator
!-----------------------------------------------------------------------
      if (lrdoper) call memhops

!-----------------------------------------------------------------------
! call analysis routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(agmat))deallocate(agmat)
if(allocated(gfgwp))deallocate(gfgwp)
if(allocated(aigwp))deallocate(aigwp)
if(allocated(zgwp))deallocate(zgwp)
if(allocated(hteil))deallocate(hteil)
if(allocated(gh))deallocate(gh)
if(allocated(gs2))deallocate(gs2)
if(allocated(hung))deallocate(hung)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(agmat(agmatdim))
allocate(gfgwp(gfgwpdim))
allocate(aigwp(aigwpdim))
allocate(zgwp(zgwpdim))
allocate(hteil(pmatdim))
allocate(gh(ghmatdim))
allocate(gs2(gs2matdim))
allocate(hung(hungdim))

      call suboppsi(mc(mcpsi),mc(mcpsi1),mr(mrort),mc(mchin),&
           mc(mcrueck),mi(mifftfak),mr(mrhops),ms(mspsi),mi(mijindx),&
           mc(mchpsi),mc(mchunpsi),hteil,mc(mcworkc),mi(miworki),&
           agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,&
           hung,mc(mcadgwp),mi(mitrajst),mi(migwpdep))

deallocate(mc,mr,mi,ml,ms,mf)
deallocate(gfgwp,aigwp,zgwp,hteil,gh,gs2,hung)
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_operdef
      call dealloc_psidef
      call dealloc_hpsimod
      call dealloc_dirdyn
      call dealloc_dddb
      call dealloc_dercp
      call dealloc_gapkernel

      stop
  900 continue
      routine='Expect'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
  999 continue
      write(6,'(2a)') 'Error reading :', filename
      stop
 1000 continue
      ilbl=index(filename,' ')-1
      inquire(file=filename(1:ilbl),exist=lexist)
      if(lexist) write(6,'(2a,/a)') &
          ' The output file: ',filename(1:ilbl),' already exists!'&
         ,' overwrite disabled? (Use -w option to enable overwrite)'
      write(6,'(2a)') ' Cannot open file: ',filename(1:ilbl)
      stop
      contains

!***********************************************************************
!
!     HERE IS THE SUBROUTINE TO DO THE ANALYSIS
!
!***********************************************************************
      subroutine suboppsi (psi,dtpsi,ort,hin,rueck,fftfak,hops,&
                 spsi,jindx,hpsi,hunpsi,hteil,workc,worki,&
                 agmat,gfgwp,aigwp,zgwp,qcenter,gh,gs2,&
                 hung,adgwp,trajst,gwpdep)

      implicit none

      complex(dop), dimension(1)                     :: zdum,zdum1
      complex(dop), dimension(*), intent(out)        :: workc
      complex(dop), dimension(dgldim), intent(out)   :: psi,dtpsi
      complex(dop), dimension(fftdim), intent(out)   :: hin,rueck
      complex(dop), dimension(hpsidim), intent(out)  :: hpsi
      complex(dop), dimension(totphidim), intent(out):: hunpsi
      complex(dop), dimension(pmatdim), intent(out)  :: hteil
      complex(dop), dimension(gsmatdim), intent(out) :: adgwp
      complex(dop), dimension(agmatdim), intent(out) :: agmat
      complex(dop), dimension(gfgwpdim), intent(out) :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out) :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)  :: zgwp
      complex(dop), dimension(ghmatdim), intent(out) :: gh
      complex(dop), dimension(gs2matdim), intent(out):: gs2
      complex(dop), dimension(hungdim), intent(out)  :: hung
      complex(sip), dimension(spsidim), intent(out)  :: spsi

      real(dop), external                            :: fcputime,fsystime
      real(dop)                                      :: tcpu,tsys,ttot,time
      real(dop), dimension(1)                        :: rdum,rdum1,rdum2,&
                                                        rdum3
      real(dop), dimension(qcentdim), intent(out)    :: qcenter
      real(dop), dimension(ortdim), intent(out)      :: ort
      real(dop), dimension(hopsdim), intent(out)     :: hops

      integer(long), external                         :: slen
      integer(long)                                   :: t1,t2,i,chkdvr,chkgrd,n,&
                                                         nham,d
      integer(long), dimension(1)                     :: idum,idum1
      integer(long), dimension(fftdim), intent(out)   :: fftfak
      integer(long), dimension(adim), intent(out)     :: jindx
      integer(long), dimension(workidim), intent(out) :: worki
      integer(long), dimension(*), intent(out)        :: trajst,gwpdep

      character(len=120)                              :: gnucmd
      logical(kind=4)                                 :: lexist,limag

      logical(kind=4), dimension(1)                   :: ldum
      logical(kind=4)                                 :: lopnorm
       
!-----------------------------------------------------------------------
! for GWP integrals use symmetric LHA as default and energy only
!-----------------------------------------------------------------------
      if (lgwp) then
         gwpintm(:)=2
         gwpintordm(:) = 0
         gwpcxm(:) = 0
      endif

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! read data from dvr file if desired
! edit which data (set dvrdata(x) to .true. and substitute dummy array)
!-----------------------------------------------------------------------
      if (lrddvr) then
         dvrdata(1)=.true.
         dvrdata(2)=.false.
         dvrdata(4)=.false.
         dvrdata(5)=.true.
         dvrdata(6)=.true.
         dvrdata(7)=.true.
         dvrdata(8)=.false.
         dvrdata(9)=.false.
         dvrdata(11)=.false.
         dvrdata(12)=.false.
         chkdvr=1
         call rddvr(ort,rdum,rdum1,rdum2,hin,&
           rueck,fftfak,zdum,zdum1,idum,idum1,rdum3,chkdvr)
      endif

!-----------------------------------------------------------------------
! read data from oper file.
!-----------------------------------------------------------------------
      if (lrdoper) then
         chkdvr=1
         chkgrd=1
         call rdoper(hops,chkdvr,chkgrd)
      endif

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
      nham=0
      do n=1,hamdim
         if (hamlab(n) .eq. operator) nham=n
      enddo

      write(ilog,'(a,i3,a,l2,3x,a)')  ' nham =',nham,&
           '   nodiag =',nodiag(nham),operator
      write(ilog,'(/,2a)') '# Wavefunction is read from file : ',&
            filein(1:laein)

      time  = tinit
      limag = .false.

!-----------------------------------------------------------------------
! loop over steps
!-----------------------------------------------------------------------
      if (nskip.gt.0)then
         do n=1,nskip
            call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
            if (lend) go to 77
            time=time+out2
         enddo
      endif

   10 continue
      call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
      time=time+out2
      if (lend) goto 77

!-----------------------------------------------------------------------
! write values, continue or terminate the loop over data-sets.
!-----------------------------------------------------------------------
      write(ilog,'(f10.3)') time/fs
      write(6,'(a,f8.2,a)') 'Time = ',time/fs,' fs'

      do n=1,step-1
         call rdpsi(ipsi,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
         if (lend) go to 77
         time=time+out2
      enddo
      if (time .gt. tfin*fs+0.1*out2) go to 77

      lopnorm=.false.
      call operate(psi,nham,lopnorm)

      if (lsp) then
         write(ipsi1) (cmplx(psi(d)), d = 1,dgldim)
      else
         write(ipsi1) (psi(d), d = 1,dgldim)
      endif

      go to 10

 77   continue
      close(ipsi)
      close(ipsi1)

!-----------------------------------------------------------------------
! Get overall timing 
!-----------------------------------------------------------------------
      tcpu = fcputime(0)
      tsys = fsystime(0)
      ttot = tcpu+tsys
      t1   = int(ttot/3600)
      ttot = ttot - 3600*t1
      t2   = int(ttot/60)
      ttot = ttot - 60*t2

      write(ilog,'(//,a,/)') ' ***** Overall Timing *****'
      write(ilog,'(a,1x,f12.2)') ' Total cpu time     [s] : ',tcpu
      write(ilog,'(a,1x,f12.2)') ' Total system time  [s] : ',tsys
      write(ilog, 900) ' Total time     [h:m:s] : ',t1,t2,ttot
      write(ilog,*)
      write(6,'(//,a,/)') '***** Overall Timing *****'
      write(6,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
      write(6,'(a,1x,f12.2)') 'Total system time  [s] : ',tsys
      write(6, 900) 'Total time     [h:m:s] : ',t1,t2,ttot
      write(6,*)
      
  900 format( a,i4,' :',i3,' :',f5.1)
      
      return

      end subroutine suboppsi

!***********************************************************************
!-----------------------------------------------------------------------
! EXPINPUT reads the options and the argument.
!-----------------------------------------------------------------------
      subroutine oppinput(stat)

      implicit none

      integer(long), external      :: myiargc
      integer(long)                :: iarg,i,ierr
      character(len=80)            :: buf
      character(len=*), intent(out):: stat

!-----------------------------------------------------------------------
! provide default input and output filenames 
!-----------------------------------------------------------------------
      filein  = ' '
      operfile = 'oper'
      fileout = 'psi.op'
      operator = ' '
      name='./'
      ddpath = '..'

!-----------------------------------------------------------------------
! set default flags
!
! stat : status of output file   
!-----------------------------------------------------------------------
      stat   = 'new'
      step   = 1
      nskip  = 0
      tfin   = 1.0e+50_dop

!-----------------------------------------------------------------------
! read options 
!-----------------------------------------------------------------------
      iarg = myiargc() 
      if (iarg .lt. 1) goto 200 

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf(1:4) .eq. '-op ') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if(buf(1:3) .eq. '-f ') then
            i = i+1
            call mygetarg(i,buf)
            filein = buf
         else if( buf(1:3) .eq. '-o ' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
         else if(buf(1:3) .eq. '-i ') then
            i=i+1
            call mygetarg(i,buf)
            name = buf
         else if (buf(1:3) .eq. '-t ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*) tfin
         else if( buf(1:4) .eq. '-dp ' ) then
            lsp=.false.
         else if( buf(1:3) .eq. '-n ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (ierr .ne. 0  .or.  step .le. 0) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -n',&
                 ' must be a positive integer. Not : ', buf
               stop
            end if
         else if( buf(1:5) .eq. '-skip' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) nskip
            if (ierr .ne. 0  .or.  nskip .le. 0) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -skip',&
                 ' must be a positive integer. Not : ', buf
               stop
            end if
         else if( buf(1:2) .eq. '-w' ) then
            lovwr=.true.
         else if( buf(1:4) .eq. '-DB ' ) then
            i = i+1
            call mygetarg(i,buf)
            ddpath = buf

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(1x,2a,/1x,2a,/17x,a)')&
       'Purpose: ',&
       'Operates on a wavefunction and stores the result.',&
       'Usage  : ',&
       'oppsi [-op -f -i -o -t -n -skip -dp -pt -w -ver -h -?] [operator].'
            write(6,'(/,a)') 'Options : '
            write(6,'(2a,/,9x,a)') &
                     ' -op FILE : The operator is read '&
                    ,'from file FILE rather than from ./oper'
            write(6,'(2a,/,9x,a)') &
                     ' -f FILE : The wavefunction is read '&
                    ,'from file FILE rather than from ./psi'&
            ,'  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,11x,a/,11x,a/,11x,a)')&
                     ' -o FILE : The output is written to file ',&
                     'FILE.pl rather than to ./expect.pl',&
             'The string FILE may be a relative or a full path-name.',&
             'If FILE=no, i.e. "-o no", no output file will be opened,',&
             'and the output is directed to screen.'
            write(6,'(a)') ' -i DIR : data is stored in directory DIR'
            write(6,'(2a)') 
            write(6,'(2a)')&
                     ' -t tfin : The expectation values are computed',&
                     ' only up to time=tfin.'
            write(6,'(2a)') ' -n step : Compute the property',&
                            ' only every step-th output.'
            write(6,'(2a)') ' -skip nskip : Skip the first nskip',&
                            ' wavefunctions.'
            write(6,'(2a)') ' -dp : Wavefunction saved double precision'
            write(6,'(2a)')&
                     ' -w      : An existing property file is ',&
                     'overwritten.'
            write(6,'(a)')&
                     ' -DB PATH : Sets path to DD DB relative to current &
                     &location. Default ..'
            write(6,'(a)')&
                     ' -ver    : Version information about the program.'
            write(6,'(a,/a)')&
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(/2a/,2a)') ' The argument operator is the name',&
                            ' of the operator, as specified in an',&
                            ' HAMILTONIAN-SECTION_xxx, for which the',&
                            ' expectation value is required.'
            write(6,'(2a)') ' If no argument is given, ',&
                 'the user is prompted for the missing argument.'
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      if (iarg .lt. i) goto 200

!-----------------------------------------------------------------------
! get compulsory argument.
!-----------------------------------------------------------------------
      if (iarg .eq. i) call mygetarg(i,operator)

!-----------------------------------------------------------------------
! End of loop over options and arguments.
!-----------------------------------------------------------------------
 200  continue
 
!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      open (unit=ilog,file='expect.log',form='formatted')
      logisopen=.true.

!-----------------------------------------------------------------------
! set status variable to overwrite files
!-----------------------------------------------------------------------
      if (lovwr) stat='unknown'

!-----------------------------------------------------------------------
! convert filenames to absolute
!-----------------------------------------------------------------------
      laenge=index(name,' ')-1
      call abspath(name,laenge)

      if( filein .eq. ' ') then
         filein='psi'
      endif
      laein=index(filein,' ')-1
      if (filein(max(1,laein-6):laein) .eq. 'restart') irst = ipsi
      if (filein(laein:laein)   .eq. '/') then
         filein = filein(1:laein)//'psi'
         laein  = laein + 3
      end if

      call abspath(filein,laein)

      dname=name
      dlaenge=laenge

      if (operfile .eq. '  ') then
         operfile=name(1:laenge)//'/oper'
         opflaenge=index(operfile,' ')-1
      else
         opflaenge=index(operfile,' ')-1
         call abspath(operfile,opflaenge)
      endif

      if (fileout .eq. '  ') then
         fileout=name(1:laenge)//'/psi.op'
         laeout=index(fileout,'  ')-1
      else if (fileout(1:3) .ne. 'no ') then
         laeout=index(fileout,'  ')-1
         call abspath(fileout,laeout)
      endif

      return

      stop
      end subroutine oppinput


      end program oppsi

