!***********************************************************************
!
!                       PROBSQ
!
!  Computes the sum p_j^2, where p_j denotes the occupation of the
!  j-th eigenstate. Essentially, integral over autocorrelation-squred. 
!  See appendix of JCP 120, 6992 (2004), or WRITEUPS/probsq
! 
!  (HDM 02/03)
!
!***********************************************************************

      program probsq

use decimal, only:long,dop
use versions

use module_interface

      implicit none

      integer(long), parameter          :: maxtp=20480
      integer(long)                     :: i,iautosq
      real(dop)                         :: a,b,c,t,t0,sum1,sum2,sum3,sum4,& 
                                           sum5,pi,pia,tcut,tstart
      real(dop), dimension(maxtp)       :: autosq
      character(len=80)                 :: buffer

!-----------------------------------------------------------------------
!     Set defaults. Set up the input.
!-----------------------------------------------------------------------
      call psqinp(tstart,tcut)

!-----------------------------------------------------------------------
!     Read the autocorrelation function.
!-----------------------------------------------------------------------
 25   read(20,'(a80)',end=31,err=32) buffer
      if( buffer(1:9) .ne. '#    time' ) goto 25
 26   read(20,*,end=31,err=32) t0, a, b, c
      if( t0 .lt. tstart ) goto 26
      autosq(1) = c**2
      read(20,*,end=31,err=32) t, a, b, c
      autosq(2) = c**2
      tstart = tstart - 0.01_dop*(t-t0)
      tcut   = tcut - 0.99_dop*(t-t0)
      
      do i = 3, maxtp
         read(20,*,end=30,err=32) t, a, b, c
         autosq(i) = c**2 
         iautosq = i
         if( t .gt. tcut ) goto 30
      end do
!-----------------------------------------------------------------------
!

 30   continue
      if( iautosq .eq. maxtp ) then
       write(6,'(a/,a)') &
        ' Not all cross-correlation data points may have been read!',&
        ' Increase parameter maxtp and re-compile'
      end if
      write(6,'(a,i6,a,2(f10.3,a))') ' No. of t-points =', iautosq,&
          ',  t-start =', t0,' fs,   t-final =', t,' fs'


!-----------------------------------------------------------------------
! calculate integral
!-----------------------------------------------------------------------
      pi  = 4.0_dop*atan(1.0_dop)
      pia = pi/dble(iautosq)
      
      sum1 = 0.5_dop*autosq(1)
      sum2 = 0.0_dop
      sum3 = 0.5_dop*autosq(1)
      sum4 = 0.0_dop
      sum5 = 0.5_dop*autosq(1)
      do i = 2, iautosq
         sum1 = sum1 + autosq(i)
         sum2 = sum2 + autosq(i)*sin(pia*(i-1))
         sum3 = sum3 + autosq(i)*cos(0.5_dop*pia*(i-1))
         sum4 = sum4 + autosq(i)*sin(pia*(i-1))**2
         sum5 = sum5 + autosq(i)*(1.d0-dble(i-1)/dble(iautosq))
      end do
      sum1 = sum1 - 0.5_dop*autosq(iautosq)
      sum1 = sum1/dble(iautosq-1)
      sum2 = sum2*pi/dble(2*iautosq)
      sum3 = sum3*pi/dble(2*iautosq)
      sum4 = sum4*2.0_dop/dble(iautosq)
      sum5 = sum5*2.d0/dble(iautosq)

      write(6,'(/,23x,a)') &
      'BOX             SIN             COS          SIN**2      (1-t/T)'
      write(6,'(a,5f16.9)') &
      ' Integral :', sum1, sum2, sum3, sum4, sum5
      write(6,'(a,5f16.2)') &
      ' Inverse  :',1.d0/sum1,1.d0/sum2,1.d0/sum3,1.d0/sum4,1.d0/sum5
      write(6,'(a,5f16.6)') &
      ' Log      :', -log(sum1)/log(2.d0),-log(sum2)/log(2.d0), &
       -log(sum3)/log(2.d0),-log(sum4)/log(2.d0),-log(sum5)/log(2.d0)


      stop

 31   write(6,'(a)') ' No data. Check the input (autosq) file.'
      stop
 32   write(6,'(a)') &
         ' Error occurred when reading the input (autosq) file.'
      stop
      contains
      
!#######################################################################

      subroutine psqinp(tstart,tcut)

      implicit none

      integer(long), external  :: myiargc
      integer(long)            :: i,iarg,ierr,laein,laenge
      real(dop), intent(out)   :: tstart,tcut
      character(len=80)        :: buf
      character(len=240)       :: filein
!-----------------------------------------------------------------------

      filein = 'auto'

      tcut = 9.9e+49_dop
      tstart = 0.0_dop
 
      iarg = myiargc() 
      if (iarg .lt. 1) goto 200 

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf(1:2) .eq. '-i' ) then
            i = i+1
            call mygetarg(i,buf)
            filein = buf
         else if( buf .eq. '-a' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) tstart
            if (ierr .ne. 0) then
               write(6,*)'Argument for -a must be real number. Not :'
               write(6,*) buf
               stop
            end if
         else if( buf(1:2) .eq. '-t' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) tcut
            if (ierr .ne. 0) then
               write(6,*)'Argument for -t must be real number. Not :'
               write(6,*) buf
               stop
            end if
         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(/,a)')&
       'Purpose: Computation of the quadratic entropy. sum p_i^2.'

            write(6,'(/,2a,/19x,a)')&
       'Usage : probsq [-i -t -a -ver -h -?]'
            write(6,'(/,a)') 'Options : '
            write(6,'(2a,/,9x,a/,9x,a)') &
               ' -i FILE  : The autocorrelation is read from',&
               ' file FILE','   rather than from ./auto',&
               '   The string FILE may be a relative or full path-name.'
           write(6,'(2a)') &
               ' -a tstart: The autocorrelation function is read only',&
               ' from t=tstart on.'
           write(6,'(2a)') &
               ' -t tcut  : The autocorrelation function is read only',&
               ' up to t=tcut'
            write(6,'(a)')&
               ' -ver     : Version information about the program '
            write(6,'(a,/a)') &
               ' -h       : Print this help text. ',&
               ' -?       : Print this help text. '
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            if(ichar(buf(2:2)).ge.46 .and. ichar(buf(2:2)).le.57) &
                 goto 200     ! Found a negative number, not an option.
            write(6,*) 'unknown option : ', buf(1:50)
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      
 200  continue

!-----------------------------------------------------------------------
! If filein has not been set, look in name directory.
!-----------------------------------------------------------------------
      laein=index(filein,' ')-1
      call abspath(filein,laein)
      
      i=laein
 10   continue
      if (filein(i:i) .eq. '/') then
         laenge=i-1
      else
         i=i-1
         go to 10
      endif

      open(20,file=filein(1:laein),form='formatted',&
              status='old',err=98)
      if(laein .le. 58 ) then
      write(6,'(2a)') ' The  inputfile is : ',filein(1:laein)
      else
      write(6,'(a)') ' The inputfile is : '
      write(6,'(a)')   filein(1:laein)
      end if


      return 
      
 98   write(6,*)' Cannot open input (auto) file : '
      write(6,*) filein
      stop

      end subroutine psqinp

      end program probsq
