!***********************************************************************
!
!  ANALYSE programs module  RDFILES
!
!  This file contains routines to read the files output by the 
!  MCTDH program
!
! rdpsi:  read the wavefunction from the psi file
! rdpsindat: read the no. datasets in the psi file
! rdpsigrid: read the wavefunction from the psi file and transform it
!            to the full grid (primitive basis) representation.
! rdgridpop: read the gridpop file
!
!
!***********************************************************************

      module rdfilesmod
 
use decimal, only: long,dop,sip
use op1lib,  only: zerovxz, cpvxz
use mtlib,   only: mtxxzz

use global
use aglobal
use maxv
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod

use logdat
use gwplib, only: param_from_psi,pqtoxivec
use iofile, only: rdpsiinfo
use channels, only: filever, ipsi
use iopsidef, only: psidat
use wftree, only: new_wftree

      implicit none

!-----------------------------------------------------------------------
! needed to read data from check file. See subroutine rdchk for 
! definition
!-----------------------------------------------------------------------
      logical(kind=4), dimension(8), save :: lrdchk=.false.

      private
      public :: rdpsindat,rdpsi,rdpsirst,rdpsigrid,rdgridpop,rdchk,&
                rewtodata,rdtodata
      public :: lrdchk

      contains

!***********************************************************************
!
!                RDPSINDAT
!
! Called by the various ANALYSE programs in order to read the
! no. of data sets in a file 
!
!***********************************************************************

      subroutine rdpsindat(unit,ndata)

      implicit none

      integer(long), intent(in)     :: unit
      integer(long), intent(out)    :: ndata
      integer(long)                 :: chkdvr,chkgrd,chkpsi,chkdat
      integer(long)                 :: ilbl,jlbl
      character(len=c2)             :: rdtag


      rewind(unit)
      if (filever(unit) .ge. 11.0_dop) then
         call rdtodata(unit)
         do
! read next record - error means it is not a string so ignore and go on to next
   10       continue
            read(unit,err=10,end=100) rdtag
            if (rdtag(1:1) .ne. '<') cycle

            ilbl=2
            jlbl=index(rdtag,'>')-1
            if (rdtag(ilbl:jlbl-9) .eq. 'data set') then
               read(rdtag(jlbl-9:jlbl),*) ndata
            endif
         enddo
  100    continue

      else
         read(ipsi) filever(ipsi)
         chkdvr=0
         chkgrd=0
         chkpsi=0
         chkdat=0
         call rdpsiinfo(unit,chkdvr,chkgrd,chkpsi,chkdat)
         ndata=0
         do
            read(ipsi,end=102,err=998)
            ndata = ndata + 1
         enddo
  102    continue

      endif

      return
 998  continue
      write(message,'(a,i5)') 'Error reading Psi file on unit:',unit
      write(6,'(a)') message
      stop

      end subroutine rdpsindat

!***********************************************************************
!
!                RDPSI
!
! Called by the various ANALYSE programs in order to read the
! psi file data which is formatted differently depending
! on selected options.
!
!***********************************************************************

      subroutine rdpsi(unit,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)

      implicit none

      integer(long), intent(in)                         :: unit
      integer(long), dimension(*), intent(out)          :: jindx
      integer(long), dimension(gwpinfodim), intent(out) :: trajst
      integer(long), dimension(gwpinfodim), intent(out) :: gwpdep
      complex(sip), dimension(dgldim), intent(out)      :: spsi
      complex(dop), dimension(dgldim), intent(out)      :: psi
      complex(dop), dimension(agmatdim), intent(out)    :: agmat
      complex(dop), dimension(gsmatdim), intent(out)    :: adgwp
      logical(kind=4), save                             :: lfirst = .true.


      if (filever(unit) .ge. 11.0_dop) then
! first call read down to data
         if (lfirst) then
            call rdtodata(unit)
            lfirst=.false.
         endif
         call rdpsi11(unit,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)
      else
         call rdpsi_pre11(unit,psi,spsi,jindx,agmat,trajst,adgwp)
      endif

      return
      end subroutine rdpsi

!#######################################################################

      subroutine rdpsi11(unit,psi,spsi,jindx,agmat,trajst,adgwp,gwpdep)

      implicit none

      integer(long)                                     :: dgl,zeig,s,dgldim1,m
      integer(long)                                     :: ilbl,jlbl,ndata
      integer(long)                                     :: tapelength1
      integer(long), dimension(maxsta)                  :: actblock
      integer(long), intent(in)                         :: unit
      integer(long), dimension(*), intent(out)          :: jindx
      integer(long), dimension(gwpinfodim), intent(out) :: trajst
      integer(long), dimension(gwpinfodim), intent(out) :: gwpdep
      real(sip), dimension(dgldim)                      :: rpsi
      complex(sip), dimension(dgldim), intent(out)      :: spsi
      complex(dop), dimension(dgldim), intent(out)      :: psi
      complex(dop), dimension(agmatdim), intent(out)    :: agmat
      complex(dop), dimension(gsmatdim), intent(out)    :: adgwp
      logical(kind=4)                                   :: lpsisp,lpsicm,lselect
      character(len=c2)                                 :: rdtag

      lend = .false.
      lpsisp=lpsiopt(1)
      lpsicm=lpsiopt(2)
      lselect=lpsiopt(4)

      dgldim1=0

      do
! read next record - error means it is not a string so ignore and go on to next
   10    continue
         read(unit,err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:ilbl+7) .eq. 'data set') then
            read(rdtag(ilbl+8:jlbl),'(i10)') ndata
!            print*,'Reading data Set: ',ndata

         else if (rdtag(ilbl:jlbl) .eq. 'compact indices') then
!-----------------------------------------------------------------------
! read indices for compact output or S-MCTDH wavefunction
!-----------------------------------------------------------------------
            dgldim1=dgldim
            zeig=0
            do s=1,nstate
               read (unit,err=999) actblock(s)
               read (unit,err=999) (jindx(zeig+dgl),dgl=1,actblock(s))
               zeig=zeig+actblock(s)
               dgldim1=dgldim1 - (block(s)-actblock(s))
            enddo

         else if (rdtag(ilbl:jlbl) .eq. 'sci indices') then
            read (unit,err=999) ((jindx(zpsi(s)+dgl),dgl=1,block(s)),s=1,nstate)

!-----------------------------------------------------------------------
! read width matrix for FGA
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'agmat') then
            read(unit,err=999) (agmat(dgl),dgl=1,agmatdim)

!-----------------------------------------------------------------------
! read no. of linearly dependent GWPs
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'ngwpdep') then
            read(unit,err=999) ((ngwpdep(m,s),m=1,nmode),s=1,nstate)

!-----------------------------------------------------------------------
! read GWP linear dependencies
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'gwpdep') then
            read(unit,err=999) (gwpdep(dgl),dgl=1,gwpinfodim)

!-----------------------------------------------------------------------
! read GWP -> SPF trafo
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'adgwp') then
            read(unit,err=999) (adgwp(dgl),dgl=1,gsmatdim)

!-----------------------------------------------------------------------
! read state information for trajectory points
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'trajst') then
            read(unit,err=999) (trajst(dgl),dgl=1,gwpinfodim)

!-----------------------------------------------------------------------
! read dynamicWF size
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'dimtape') then
            read(unit,err=999) ((dim(m,s),m=1,nmode),s=1,nstate)
            if (lrunml) then
               read(unit,err=999) tapelength1
               read(unit,err=999) (tape(m),m=1,tapelength1)
               tree => new_wftree(tape)
            endif
            call psidat

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'psi') then
            if (lpsisp .and. psitype == 11) then ! realwp 
               read (unit,err=999) (rpsi(dgl),dgl=1,dgldim)
               do dgl=1,dgldim
                  spsi(dgl)=cmplx(rpsi(dgl),0.0_sip,sip)
                  psi(dgl)=spsi(dgl)
               enddo

            else if (lpsisp) then
               read (unit,err=999) (spsi(dgl),dgl=1,dgldim) 
               do dgl=1,dgldim
                  psi(dgl)=spsi(dgl)
               enddo

            else if (lpsicm) then
               read (unit,err=999) (spsi(dgl),dgl=1,dgldim1)

               zeig=1
               do s=1,nstate
                  call cmavec(psi(zpsi(s)),spsi(zeig),jindx(zeig),&
                       block(s),actblock(s))
                  zeig=zeig+actblock(s)
               enddo

               do dgl=adim+1,dgldim
                  psi(dgl)=spsi(zeig)
                  zeig=zeig+1
               enddo

            else
               read (unit,err=999) (psi(dgl),dgl=1,dgldim)
            endif

         else if (rdtag(ilbl:ilbl+11) .eq. 'end data set') then
            exit

         endif

      enddo

!-----------------------------------------------------------------------
! get width matrix (agmat) if Thawed GWP
! store linear parameters in agmat(xivec)
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call param_from_psi(nspfdof(m),gwpdim(m,s),&
                    agmat(zaGmat(m,s)),psi(zpg(m,s)),psi(zpg(m,s)),&
                    psi(zphi(m,s)),agmat(zphase(m,s)),psi(zxi(m,s)),&
                    1,s,m)
               call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
                    psi(zxi(m,s)),agmat(zxivec(m,s)))
            endif
         enddo
      enddo

! end of data set
      return

! end of file reached
  100 continue
      lend=.true.
      return

  999 continue
      routine='Rdpsi11'
      message='Error Reading PSI file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rdpsi11

!***********************************************************************
!
!                RDPSI
!
! Called by the various ANALYSE programs in order to read the
! psi file data which is formatted differently depending
! on selected options.
!
!***********************************************************************

      subroutine rdpsi_pre11(unit,psi,spsi,jindx,agmat,trajst,adgwp)

      implicit none

      integer(long)                                     :: dgl,zeig,s,dgldim1,m
      integer(long), dimension(maxsta)                  :: actblock
      integer(long), intent(in)                         :: unit
      integer(long), dimension(*), intent(out)          :: jindx
      integer(long), dimension(gwpinfodim), intent(out) :: trajst
      real(sip), dimension(dgldim)                      :: rpsi
      complex(sip), dimension(dgldim), intent(out)      :: spsi
      complex(dop), dimension(dgldim), intent(out)      :: psi
      complex(dop), dimension(agmatdim), intent(out)    :: agmat
      complex(dop), dimension(gsmatdim), intent(out)    :: adgwp
      logical(kind=4)                                   :: lpsisp,lpsicm,lselect

      lend = .false.
      lpsisp=lpsiopt(1)
      lpsicm=lpsiopt(2)
      lselect=lpsiopt(4)

!-----------------------------------------------------------------------
! read indices for compact output or S-MCTDH wavefunction
!-----------------------------------------------------------------------
      dgldim1=0
      if (lpsicm) then
         dgldim1=dgldim
         zeig=0
         do s=1,nstate
            read (unit,end=10) actblock(s)
            read (unit,end=10) (jindx(zeig+dgl),dgl=1,actblock(s))
            zeig=zeig+actblock(s)
            dgldim1=dgldim1 - (block(s)-actblock(s))
         enddo
      else if (lselect) then
         do s=1,nstate
            read (unit,end=10) (jindx(zpsi(s)+dgl),dgl=1,block(s))
         enddo
      endif

!-----------------------------------------------------------------------
! read width matrix for FGA
!-----------------------------------------------------------------------
      if (lgwp .and. gwptype .ge. 3) then
         if (filever(unit) .lt. 10.1d0) then
            read(unit,end=10) (agmat(dgl),dgl=1,agmatdim100)
         else if (filever(unit) .lt. 10.2d0) then
            read(unit,end=10) (agmat(dgl),dgl=1,agmatdim101)
         else
            read(unit,end=10) (agmat(dgl),dgl=1,agmatdim)
         endif
      endif                                                                     

!-----------------------------------------------------------------------
! read GWP -> SPF trafo
!-----------------------------------------------------------------------
      if (lgwp .and. psitype .eq. 0) then
         if (filever(unit) .ge. 9.0004 .and. filever(unit) .lt. 10.1003) then
            read(unit,end=10) (adgwp(dgl),dgl=1,gsmatdim)
         endif
      endif

!-----------------------------------------------------------------------
! read state information for trajectory points
!-----------------------------------------------------------------------
      if (ltraj) then
         read(unit,end=10) (trajst(dgl),dgl=1,gwpinfodim)
      endif

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
      if (lpsisp .and. psitype == 11) then ! realwp 
         read (unit,end=10,err=30) (rpsi(dgl),dgl=1,dgldim)
         do dgl=1,dgldim
            spsi(dgl)=cmplx(rpsi(dgl),0.0_sip,sip)
            psi(dgl)=spsi(dgl)
         enddo

      else if (lpsisp) then
         read (unit,end=10,err=30) (spsi(dgl),dgl=1,dgldim) 
         do dgl=1,dgldim
            psi(dgl)=spsi(dgl)
         enddo

      else if (lpsicm) then
         read (unit,end=10,err=30) (spsi(dgl),dgl=1,dgldim1)

         zeig=1
         do s=1,nstate
            call cmavec(psi(zpsi(s)),spsi(zeig),jindx(zeig),&
                 block(s),actblock(s))
            zeig=zeig+actblock(s)
         enddo

         do dgl=adim+1,dgldim
            psi(dgl)=spsi(zeig)
            zeig=zeig+1
         enddo

      else
         read (unit,end=10,err=30) (psi(dgl),dgl=1,dgldim)

      endif

!-----------------------------------------------------------------------
! get width matrix (agmat) if Thawed GWP
! store linear parameters in agmat(xivec)
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call param_from_psi(nspfdof(m),gwpdim(m,s),&
                    agmat(zaGmat(m,s)),psi(zpg(m,s)),psi(zpg(m,s)),&
                    psi(zphi(m,s)),agmat(zphase(m,s)),psi(zxi(m,s)),1,s,m)
               call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
                    psi(zxi(m,s)),agmat(zxivec(m,s)))
            endif
         enddo
      enddo

      goto 20
 10   lend=.true.
 20   continue

      return

 30   write(6,*) 'ERROR in rdpsi (source/analyse/rdfiles.F).'
      write(6,*) 'Error occurred while reading channel ',unit
      write(6,*) 'psi : unit=8;  psi1 : unit=48'
      write(6,*) 'dgl =', dgl,',  dgldim =', dgldim
      write(6,*) 'lpsicm, lpsisp :',lpsicm, lpsisp 
      stop

      end subroutine rdpsi_pre11

!***********************************************************************
!
!                RDPSIRST
!
! Called by the various ANALYSE programs in order to read the
! psi data from the restart file 
!
!***********************************************************************

      subroutine rdpsirst(unit,psi,agmat)

use global
use aglobal 
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod

use gwplib

      implicit none

      integer(long), intent(in)                      :: unit
      complex(dop), dimension(dgldim), intent(out)   :: psi
      complex(dop), dimension(agmatdim), intent(out) :: agmat

      if (filever(unit) .ge. 11.0_dop) then
         call rdpsirst11(unit,psi,agmat)
      else
         call rdpsirst_pre11(unit,psi,agmat)
      endif

      return
      end subroutine rdpsirst

!-----------------------------------------------------------------------

      subroutine rdpsirst11(unit,psi,agmat)

use global
use aglobal 
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod

use gwplib

      implicit none

      integer(long)                                  :: dgl,s,m
      integer(long), intent(in)                      :: unit
      complex(dop), dimension(dgldim), intent(out)   :: psi
      complex(dop), dimension(agmatdim), intent(out) :: agmat
      integer(long)                                  :: ilbl,jlbl
      character(len=c2)                              :: rdtag

      rewind(unit)

!-----------------------------------------------------------------------
      do
! read next record - error means it is not a string so ignore and go on to next
   10    continue
         read(unit,err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'psi') then
            read (unit,err=30) (psi(dgl),dgl=1,dgldim)
            exit
         endif
      enddo

!-----------------------------------------------------------------------
! get width matrix (agmat) if Thawed GWP
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call param_from_psi(nspfdof(m),gwpdim(m,s),&
                    agmat(zaGmat(m,s)),psi(zpg(m,s)),psi(zpg(m,s)),&
                    psi(zphi(m,s)),agmat(zphase(m,s)),psi(zxi(m,s)),1,s,m)
               call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
                    psi(zxi(m,s)),agmat(zxivec(m,s)))
            endif
         enddo
      enddo

      return

 100  write(6,*) 'ERROR in rdpsirst (source/analyse/rdfiles.F).'
      write(6,*) 'Psi not found on rst file reading channel ',unit
      stop
      
 30   write(6,*) 'ERROR in rdpsirst (source/analyse/rdfiles.F).'
      write(6,*) 'Error occurred while reading channel ',unit
      stop

      end subroutine rdpsirst11

!-----------------------------------------------------------------------

      subroutine rdpsirst_pre11(unit,psi,agmat)

use global
use aglobal 
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod

use gwplib

      implicit none

      integer(long)                                  :: dgl,s,m
      integer(long), intent(in)                      :: unit
      complex(dop), dimension(dgldim), intent(out)   :: psi
      complex(dop), dimension(agmatdim), intent(out) :: agmat

      read (unit,err=30) (psi(dgl),dgl=1,dgldim)

!-----------------------------------------------------------------------
! get width matrix (agmat) if Thawed GWP
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call param_from_psi(nspfdof(m),gwpdim(m,s),&
                    agmat(zaGmat(m,s)),psi(zpg(m,s)),psi(zpg(m,s)),&
                    psi(zphi(m,s)),agmat(zphase(m,s)),psi(zxi(m,s)),1,s,m)
               call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
                    psi(zxi(m,s)),agmat(zxivec(m,s)))
            endif
         enddo
      enddo

      return

 30   write(6,*) 'ERROR in rdpsirst (source/analyse/rdfiles.F).'
      write(6,*) 'Error occurred while reading channel ',unit
      stop

      end subroutine rdpsirst_pre11

!#######################################################################
!
! CMAVEC
!
! copies the truncated A-vector, stored in compact storage mode,
! to the psi vector, filling in blank spaces with 0
!
!#######################################################################

      subroutine cmavec(psi,spsi,jindx,block,actblock)

      implicit none

      integer(long)                                  :: b1,b2
      integer(long), intent(in)                      :: block,actblock
      integer(long), dimension(actblock), intent(in) :: jindx
      complex(sip), dimension(actblock), intent(out) :: spsi
      complex(dop), dimension(block), intent(out)    :: psi

      if (actblock .eq. 0) then
         call zerovxz(psi,block)
         return
      endif

      b1=1
      do b2=1,block
         if (b2 .eq. jindx(b1)) then
            psi(b2)=spsi(b1)
            if (b1 .lt. actblock) b1=b1+1
         else
            psi(b2)=0.0_dop
         endif
      enddo

      return
      end subroutine cmavec

!***********************************************************************
!
!                RDPSIGRID
!
! Called by the various ANALYSE programs in order to read the
! psi file data, and transform the wavefunction to the full grid
! (primitive basis) representation
!
!***********************************************************************

      subroutine rdpsigrid(unit,psi,spsi,jindx,workc,lrst)

use global
use aglobal 
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod
use iorst, only: rdrstpsi
use channels, only: irst

      implicit none

      integer(long)                               :: zeig1,zeig2,zgrd,m,s,&
                                                     block1,swapzeig,vdim1,i
      integer(long), dimension(1)                 :: idum,idum1
      integer(long), intent(in)                   :: unit
      integer(long), dimension(adim), intent(out) :: jindx
      real(dop)                                   :: rdum
      complex(dop), dimension(1)                  :: cdum,cdum1
      complex(sip), dimension(dgldim), intent(out)  :: spsi
      complex(dop), dimension(griddim), intent(out) :: psi
      complex(dop), dimension(dgldim+2*(griddim/nstate)), intent(out) :: workc
      logical(kind=4)                               :: lrst

      if (lgwp) then
         routine='rdpsigrid'
         message='Cannot convert a GWP wavefunction to the full grid'
         call errormsg
      endif

      if(dgldim+2*(griddim/nstate).gt.workcdim) then
         routine='rdpsigrid'
         message='workcdim is too small'
         write(6,*) 'workcdim =                 ', workcdim
         write(6,*) 'dgldim+2*(griddim/nstate) =',&
                     dgldim+2*(griddim/nstate)
         call errormsg
      endif 
      zeig1=dgldim+1
      zeig2=zeig1+griddim/nstate

! read mctdh-wfn into array psi or spsi
      if(lrst) then
         if (filever(irst) .lt. 11.0) then
            read(irst,err=999) (workc(i),i=1,dgldim)
         else
            call rdrstpsi(rdum,workc,jindx,cdum,idum)
         endif
      else
         call rdpsi(unit,workc,spsi,jindx,cdum,idum,cdum1,idum1)
      endif

      zgrd=1
      do s=1,nstate
         call cpvxz(workc(zpsi(s)),workc(zeig2),block(s))

         vdim1=1
         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz (workc(zetf(m,s)),workc(zeig1),&
                 workc(zeig2),vdim1,dim(m,s),ndim(m,s),subdim(m))
            block1=vdim1*subdim(m)*ndim(m,s)
            vdim1=vdim1*subdim(m)
         enddo

         call cpvxz(workc(zeig2),psi(zgrd),griddim/nstate)
         zgrd=zgrd+griddim/nstate
      enddo

      return

 999  write(6,*) ' ERROR reading restart file (-rst)'
      stop

      end subroutine rdpsigrid

!***********************************************************************
!
!                RDGRIDPOP
!
! Called by the various ANALYSE programs in order to read the
! gridpop file data
!
!***********************************************************************

      subroutine rdgridpop(unit,fgpop,gpop1,gpop2)

      implicit none

      integer(long), intent(in)                        :: unit
      real(sip), dimension(2*maxgdim), intent(out)     :: fgpop
      real(dop), dimension(ortdim,nddstate*npacktot), intent(out) &
                                                       :: gpop1,gpop2
      logical(kind=4), save                            :: lfirst = .true.

      gpop1(:,:)=0.0_dop
      gpop2(:,:)=0.0_dop
      if (filever(unit) .ge. 11.0_dop) then
! first call read down to data
         if (lfirst) then
            call rdtodata(unit)
            lfirst=.false.
         endif
         call rdgridpop11(unit,fgpop,gpop1,gpop2)
      else
         call rdgridpop_pre11(unit,fgpop,gpop1,gpop2)
      endif

      return
      end subroutine rdgridpop

!#######################################################################

      subroutine rdgridpop11(unit,fgpop,gpop1,gpop2)

use global
use aglobal 
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod
    
      implicit none

      integer(long)                                    :: g,s,m,n,f,ggdim,ndata
      integer(long), intent(in)                        :: unit
      integer(long)                                    :: ilbl,jlbl
      real(sip), dimension(2*maxgdim), intent(out)     :: fgpop
      real(dop), dimension(ortdim,nddstate*npacktot), intent(out) &
                                                       :: gpop1,gpop2
      character(len=c2)                                :: rdtag

      lend=.false.

      do
! read next record - error means it is not a string so ignore and go on to next
   10    continue
         read(unit,err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:ilbl+7) .eq. 'data set') then
            read(rdtag(ilbl+8:jlbl),'(i10)') ndata
       !     print*,'Reading data Set: ',ndata

!-----------------------------------------------------------------------
! read state populations
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'gpop') then

            if (lgpel) then
               do s=1,nddstate*npacktot
                  do m=1,nmode
                     do n=1,nspfdof(m)
                        f=spfdof(n,m)
                        ggdim = gdim(f)
                        if(basis(f).eq.12 .or. basis(f).eq.20 .or. &
                           basis(f).eq.21) ggdim = ipbaspar(3,f) 
                        read(unit,err=900) (fgpop(g),g=1,gdim(f))&
                           ,(fgpop(g),g=gdim(f)+1,gdim(f)+ggdim)
                        do g=1,gdim(f)
                           gpop1(zort(f)-1+g,s)=fgpop(g)
                        enddo
                        do g=1,ggdim
                           gpop2(zort(f)-1+g,s)=fgpop(gdim(f)+g)
                        enddo
                     enddo
                  enddo
               enddo
            else
               do m=1,nmode
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     ggdim = gdim(f)
                     if(basis(f).eq.12 .or. basis(f).eq.20 .or. &
                        basis(f).eq.21) ggdim = ipbaspar(3,f) 
                        read(unit,err=900) (fgpop(g),g=1,gdim(f))&
                           ,(fgpop(g),g=gdim(f)+1,gdim(f)+ggdim)
                     do g=1,gdim(f)
                        gpop1(zort(f)-1+g,1)=fgpop(g)
                     enddo
                     do g=1,ggdim
                        gpop2(zort(f)-1+g,1)=fgpop(gdim(f)+g)
                     enddo
                  enddo
               enddo
            endif

!-----------------------------------------------------------------------
!       finish set
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:ilbl+11) .eq. 'end data set') then
            exit

         endif

      enddo

! end of data set
      return

! end of file reached
  100 continue
      lend=.true.
      return

      return
  900 continue
      routine='Rdgridpop11'
      message='Error Reading GRIDPOP file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rdgridpop11

!#######################################################################

      subroutine rdgridpop_pre11(unit,fgpop,gpop1,gpop2)

use global
use aglobal 
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod
    
      implicit none

      integer(long)                                    :: g,s,m,n,f,ggdim
      integer(long), intent(in)                        :: unit
      real(sip), dimension(2*maxgdim), intent(out)     :: fgpop
      real(dop), dimension(ortdim,nstate), intent(out) :: gpop1,gpop2

      lend=.false.

      if (lgpel .or. nstate .eq. 1) then
         do s=1,nstate
            do m=1,nmode
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  ggdim = gdim(f)
                  if(basis(f).eq.12 .or. basis(f).eq.20 .or. &
                     basis(f).eq.21) &
                     ggdim = ipbaspar(3,f) 
                  read(unit,err=900,end=999)&
                       (fgpop(g),g=1,gdim(f))&
                     ,(fgpop(g),g=gdim(f)+1,gdim(f)+ggdim)
                  do g=1,gdim(f)
                     gpop1(zort(f)-1+g,s)=fgpop(g)
                  enddo
                  do g=1,ggdim
                     gpop2(zort(f)-1+g,s)=fgpop(gdim(f)+g)
                  enddo
               enddo
            enddo
         enddo
      else
         do m=1,nmode
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               ggdim = gdim(f)
               if(basis(f).eq.12 .or. basis(f).eq.20 .or. &
                  basis(f).eq.21) &
                  ggdim = ipbaspar(3,f) 
               read(unit,err=900,end=999)(fgpop(g),g=1,gdim(f)+ggdim)
               do g=1,gdim(f)
                  gpop1(zort(f)-1+g,1)=fgpop(g)
               enddo
               do g=1,ggdim
                  gpop2(zort(f)-1+g,1)=fgpop(gdim(f)+g)
               enddo
            enddo
         enddo
      endif

      go to 20
  999 lend=.true.
   20 continue

      return
  900 routine='Rdgridpop'
      write(message,'(a,i5)' )&
         'Error reading gridpop file on channel : ',unit
      call errormsg
      end subroutine rdgridpop_pre11

!***********************************************************************
!
!                RDCHK
!
! Called by the various ANALYSE programs in order to read the
! check file data
!
! the array lrdchk(3) enables storage of data in various arrays
!  (see aglobal.inc). See analyse/statepop.F for example.
!
! if lrdchk(n)=.true., following is read and stored
! lrdchk(1) state populations -> spop
! lrdchk(2) natural orbital populations -> dicht3
! lrdchk(3) grid expectation values -> qdq
! lrdchk(4) ggp values -> ggp
! lrdchk(5) energies -> euncorr, etot 
!
! Arrays specified by lrdchk must be correctly assigned in calling 
! routine.
!
!***********************************************************************

      subroutine rdchk(unit,ecorr,etot,spop,dicht3,ggp,qdq,norm,trace)

      implicit none

      integer(long), intent(in)                          :: unit
      integer(long)                                      :: i,f,s
      real(dop), intent(out)                             :: norm,trace
      real(dop), dimension(nddstate), intent(out)        :: spop
      real(dop), dimension(nstate), intent(out)          :: ecorr,etot
      real(dop), dimension(d3matdim), intent(out)        :: dicht3
      real(dop), dimension(gwpinfodim), intent(out)      :: ggp
      real(dop), dimension(4,ndof,nstate), intent(out)   :: qdq
      real(dop), dimension(maxdim,maxsta,4)              :: prop
      logical(kind=4), save                              :: lfirst = .true.

! zero arrays
      norm=0.0_dop
      trace=0.0_dop
      spop(:)=0.0_dop
      ecorr(:)=0.0_dop
      etot(:)=0.0_dop
      dicht3(:)=0.0_dop
      qdq(:,:,:)=0.0_dop

      if (filever(unit) .ge. 11.0_dop) then
! first call read down to data
         if (lfirst) then
            call rdtodata(unit)
            lfirst=.false.
         endif
         call rdchk11(unit,ecorr,etot,spop,dicht3,ggp,qdq,norm,trace)
      else
         call rdchk_pre11(unit,ecorr,etot,spop,dicht3,prop,norm,trace)
! change how qdq is stored to post 11.0 convention
         if (.not. lend) then
            do i=1,4
               do f=1,ndof
                  do s=1,nstate
                     qdq(i,f,s)=prop(f,s,i)
                  enddo
               enddo
            enddo
         endif
      endif
      return
      end subroutine rdchk

!#######################################################################
!
! spop(nddstate): if multi-set, nddstate=nstate so this works for
!                 both single and multi-set
!
! lrdchk controls what is read and must be set in calling routine.
!
!#######################################################################

      subroutine rdchk11(unit,ecorr,etot,spop,dicht3,ggp,qdq,norm,trace)

use global
use aglobal 
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod

      implicit none

      integer(long)                                    :: i,s,t,m,f
      integer(long)                                    :: ilbl,jlbl,ndata
      integer(long)                                    :: tapelength1
      integer(long), intent(in)                        :: unit
      real(dop), intent(out)                           :: norm,trace
      real(dop), dimension(nddstate), intent(out)      :: spop
      real(dop), dimension(nstate), intent(out)        :: ecorr,etot
      real(dop), dimension(d3matdim), intent(out)      :: dicht3
      real(dop), dimension(gwpinfodim), intent(out)    :: ggp
      real(dop), dimension(4,ndof,nstate), intent(out) :: qdq
      character(len=c2)                                :: rdtag

      do
! read next record - error means it is not a string so ignore and go on to next
   10    continue
         read(unit,err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle
!         print*,rdtag

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:ilbl+7) .eq. 'data set') then
            read(rdtag(ilbl+8:jlbl),'(i10)') ndata
!            print*,'Reading data Set: ',ndata

!-----------------------------------------------------------------------
! read dynamicWF size
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'dimtape') then
            read(unit,err=900) ((dim(m,s),m=1,nmode),s=1,nstate)
            if (lrunml) then
               read(unit,err=900) tapelength1
               read(unit,err=900) (tape(m),m=1,tapelength1)
               tree => new_wftree(tape)
            endif
            call psidat

!-----------------------------------------------------------------------
! read state populations and energies
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'sspop' .and. lrdchk(1)) then
            read(unit,err=900) (spop(s),s=1,nddstate)

         else if (rdtag(ilbl:jlbl) .eq. 'spop' .and. lrdchk(1)) then
            read(unit,err=900) (spop(s),s=1,nstate)
         
         else if (rdtag(ilbl:jlbl) .eq. 'ecorr' .and. lrdchk(5)) then
            read(unit,err=900) (ecorr(s),s=1,nstate)

         else if (rdtag(ilbl:jlbl) .eq. 'etot' .and. lrdchk(5)) then
            read(unit,err=900) (etot(s),s=1,nstate)

         else if (rdtag(ilbl:jlbl) .eq. 'decorr' .and. lrdchk(5)) then
            read(unit,err=900) (ecorr(s),s=1,nstate)

         else if (rdtag(ilbl:jlbl) .eq. 'detot' .and. lrdchk(5)) then
            read(unit,err=900) (etot(s),s=1,nstate)

!-----------------------------------------------------------------------
! read natural orbital populations
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'dicht3' .and. lrdchk(2)) then
            read(unit,err=900) &
               (((dicht3(d3mat(m,s)+i),i=0,dim(m,s)-1),m=1,nmode),s=1,nstate)

         else if (rdtag(ilbl:jlbl) .eq. 'ddicht3' .and. lrdchk(2)) then
            read(unit,err=900) &
               ((((dicht3(dd3mat(m,s,t)+i),i=0,ddim(m,s,t)-1),m=1,nmode),&
                    s=1,nstate),t=1,nstate)

!-----------------------------------------------------------------------
! read grid expectation values
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'qdq' .and. lrdchk(3)) then
            do s=1,nstate
               do f=1,ndof
                  if (dofspf(f) .eq. 0) cycle
                  read(unit,err=900) (qdq(i,f,s),i=1,4)
               enddo
            enddo

!-----------------------------------------------------------------------
! read grid expectation values
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'ggp' .and. lrdchk(4)) then
            do s=1,nstate
               do m=1,nmode
                  if (.not. gwpm(m)) cycle
                  if (m .eq. meb) cycle
                  gwpdim(m,s)=dim(m,s)
                  read(unit,err=900) &
                        (ggp(zgwpinfo(m,s)+i),i=0,gwpdim(m,s)-1)
               enddo
            enddo

!-----------------------------------------------------------------------
! read other stuff always
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'norm') then
            read(unit,err=900) norm

         else if (rdtag(ilbl:jlbl) .eq. 'trace') then
            read(unit,err=900) trace
      
!-----------------------------------------------------------------------
! finish set
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:ilbl+11) .eq. 'end data set') then
            exit

         endif

      enddo

! end of data set
      return

! end of file reached
  100 continue
      lend=.true.
      return

      return
  900 continue
      routine='Rdchk11'
      message='Error Reading CHECK file: '//rdtag(ilbl:jlbl)
      call errormsg
      end subroutine rdchk11

!#######################################################################

      subroutine rdchk_pre11(unit,ecorr,etot,spop,dicht3,prop,norm,trace)

use global
use aglobal 
use maxv, only: maxdim,maxsta
use dvrdatmod
use griddatmod
use psidef
use datenmod
use compdatmod 
use compdat1mod

      implicit none

      integer(long)                                      :: i,s,m,n,f,dim1
      integer(long), intent(in)                          :: unit
      real(dop)                                          :: q,dq,expn, & 
                                                            dexpn,rdum
      real(dop), intent(out)                             :: norm,trace
      real(dop), dimension(nddstate), intent(out)        :: spop
      real(dop), dimension(nstate), intent(out)          :: ecorr,etot
      real(dop), dimension(d3matdim), intent(out)        :: dicht3
      real(dop), dimension(maxdim,maxsta,4), intent(out) :: prop

      prop(:,:,:)=0.0_dop
!-----------------------------------------------------------------------
! read state populations
!-----------------------------------------------------------------------
      if (lrdchk(1)) then
         if (leb) then
            read(unit,err=900,end=999) (spop(s),s=1,nddstate)
            read(unit,err=900,end=999) ecorr(1),etot(1)
         else
            read(unit,err=900,end=999) &
                (spop(s),ecorr(s),etot(s),s=1,nstate)
         endif
      else
         read(unit,err=900,end=999) 
      endif

!-----------------------------------------------------------------------
! read natural orbital populations
!-----------------------------------------------------------------------
      if (lrdchk(2)) then
         do s=1,nstate
            do m=1,nmode
               if (filever(unit) .ge. 10.0000_dop .and. ldynspf) then
                  read(unit,err=900,end=999) dim1
               else
                  dim1 = dim(m,s)
               endif
               read(unit,err=900,end=999) (dicht3(d3mat(m,s)+i),i=0,dim1-1)
            enddo
         enddo
      else
         do s=1,nstate
            do m=1,nmode
               if (ldynspf) read(unit,err=900,end=999)
               read(unit,err=900,end=999) 
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! read other properties
!-----------------------------------------------------------------------
! -----  read <q>,<dq>,<n> and <dn>
      if (filever(unit).ge.8.2002_dop) then
         do s=1,nstate
            do m=1,nmode
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  if (ldvr(f) .or. basis(f) .eq. 4 .or. basis(f) .eq. 25) then
                     read(unit,end=999,err=900) q,dq,expn,dexpn
                  else if (basis(f) .eq. 6) then
                     read(unit,end=999,err=900) q,dq,rdum,rdum
                     read(unit,end=999,err=900) expn,dexpn,rdum,rdum
                  else
                     q=0.0d0
                     dq=0.0d0
                     expn=0.0d0
                     dexpn=0.0d0
                  endif
                  if (lrdchk(3)) then
                     prop(f,s,1)=q
                     prop(f,s,2)=dq
                     prop(f,s,3)=expn
                     prop(f,s,4)=dexpn
                  endif
               enddo
            enddo
         enddo
      endif
      
! --- read norm and trace (densities only)
      if (filever(unit).ge.8.3005_dop) then
        if (dentype.ne.0) then
            read(unit,end=999,err=900) norm,trace
        endif
      endif          
      

      go to 20
  999 lend=.true.
   20 continue

      return
  900 routine='Rdchck'
      write(message,'(a,i5)' )&
         'Error reading check file on channel : ',unit
      call errormsg
      end subroutine rdchk_pre11

!#######################################################################
! 
! Reads to first data set in a data file
!
! Why is this needed? rdpsi etc automatically read the next dataset
! in the file, which could be the first.
!#######################################################################

      subroutine rdtodata(unit)

      integer(long), intent(in) :: unit
      integer(long)             :: ilbl,jlbl
      character(len=c2)         :: rdtag

      rewind(unit)
      do
! read next record - error means it is not a string so ignore and go on to next
   10    continue
         read(unit,err=10,end=900) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'data set         1') exit

      enddo

! leave file so that next entry is header for first data set
      backspace(unit)

      return
  900 routine='Rdtodata'
      write(message,'(a,i5)' ) 'Error reading file on channel : ',unit
      write(message(81:94),'(a)' ) 'No data found?'
      call errormsg
      end subroutine rdtodata

!#######################################################################
! 
! Rewinds to read a previous data set (rewinds by nrew, where
! nrew = 1 repositions to reread last data set)
!
!#######################################################################

      subroutine rewtodata(unit,nrew)

      integer(long), intent(in) :: unit,nrew
      integer(long)             :: n
      integer(long)             :: ilbl
      character(len=c2)         :: rdtag

      n=nrew

      do
! backsapce to read previous record - error means it is not a string so ignore and go on to next
   10    continue
         backspace(unit)
         backspace(unit)
         read(unit,err=10,end=900) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         if (rdtag(ilbl:ilbl+8) .eq. 'data set ') n=n-1
         if (n .eq. 0) exit

      enddo

! leave file so that next entry is header for first data set
      backspace(unit)

      return
  900 routine='Rewtodata'
      write(message,'(a,i5)' ) 'Error reading file on channel : ',unit
      write(message(81:94),'(a)' ) 'No data found?'
      call errormsg
      end subroutine rewtodata

      end module rdfilesmod

