!***********************************************************************
!
!           Analyse program:  SHOWSPF
!
! Showspf reads the psi file and writes the absolute value of
! a single-particle function or of a natural-orbital to a file.
! It may call GNUPLOT to visualise the data.
!
!  12/99 GW,  03/00 HDM
!
!
!***********************************************************************
      
      program showspf

use decimal, only: long,dop,sip
use global
use aglobal
use maxv
use globmemmod
use versions
use dvrdatmod
use griddatmod
use psidef
use operdef
use channels
use datenmod
use compdatmod
use compdat1mod
use plot

use op1lib, only: zerovxl, cpmxz
use rmlib, only: rm1hxxxzz
use mmlib, only: mqxczz
use logdat, only: errormsg
use iofile, only: rdpsiinfo
use iorst, only: rstinfo
use rddvrmod, only: dvrinfo
use zeigauswmod, only: zeigausw

use module_interface
use logdat

      implicit none

      integer(long), external               :: slen
      integer(long)                         :: chkdvr,chkgrd,chkpsi,chkdat,&
                                               ilbl,anin
      character(len=c5)                     :: filename
      character(len=10)                     :: stat
      character(len=64)                     :: ccc
      logical(kind=4)                       :: lerr,linwf,check

      integer(long)                         :: dof,state,espf,m,sc,zeig
      common /iplspf/ dof,state,espf

!-----------------------------------------------------------------------
! Initialize variables and set defaults
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.false.

!-----------------------------------------------------------------------
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.true.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.true.

!-----------------------------------------------------------------------
! read input parameters and options
! open files
!-----------------------------------------------------------------------
      call plspfinp(stat)

!-----------------------------------------------------------------------
! open input and output data files
!
! use standard channels if possible
!  ilog : log file
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------
      open(ilog,file='showspf.log',form='formatted')
      logisopen=.true.
      filename=filein
      if (lshowrst) then
         open(irst,file=filename,form='unformatted',status='old',err=900)
      else
         open(ipsi,file=filename,form='unformatted',status='old',err=900)
      endif

      if (fileout .ne. 'no') then
         filename=fileout
         open(iaus,file=filename,form='formatted',&
             status=stat,err=900)
      else
         iaus=6
      endif

!-----------------------------------------------------------------------
! get array dimensions 
!-----------------------------------------------------------------------
      inquire(irst,opened=check)
      if (check) close(irst)
      filename=name(1:laenge)//'/restart'
      ilbl=index(filename,' ')-1
      open(irst,file=filename(1:ilbl),form='unformatted',status='old',err=900)
      call rdmemdim(irst)
      if (.not. lshowrst) close(irst)

!-----------------------------------------------------------------------
! Allocate memory 
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_psidef
      if (lshowrst) call alloc_operdef

!-----------------------------------------------------------------------
! read information from system files
!-----------------------------------------------------------------------
      chkdvr=1
      filename=dname(1:dlaenge)//'/dvr'
      open(idvr,file=filename,form='unformatted',status='old',err=900)
      call dvrinfo(lerr,chkdvr)
      close(idvr)
      call chksyserr(filename,lerr)
      if (lerr) go to 999

!-----------------------------------------------------------------------
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=2
      chkgrd=1
      chkpsi=1
      chkdat=1
      if (lshowrst) then
         filename='restart'
         read(irst,err=999) filever(irst)
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkdat)
      elseif (lshowpsi) then
         filename='psi'
         read(ipsi,err=999) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      endif

!-----------------------------------------------------------------------
! check that dof is not in a combined mode
!-----------------------------------------------------------------------
      m=dofspf(dof)
      if (nspfdof(m) .gt. 1) then
         write(6,'(a)') '###  DOF cannot be in a combined mode  ###'
         write(6,'(2a,i3,a,i3,a,i2)') modelabel(dof)(1:8),&
              ' f =',dof,' , m =', m,' , nspfdof =',nspfdof(m)
         stop
      endif

!-----------------------------------------------------------------------
! write gnuplot info to file
!-----------------------------------------------------------------------
      if (lgnu) then
         anin=1
         if(laein.ge.55) anin=laein-50
         
         if(lnatorb) then
            if(reim .eq. 0 ) then
               ccc = 'Abs(NatOrb), No:'
            else if(reim .eq. 1 ) then
               ccc = 'Re(NatOrb), No:'
            else if(reim .eq. 2 ) then
               ccc = 'Im(NatOrb), No:'
            else if(reim .eq. 3 ) then
               ccc = 'Re(NatOrb), Im(NatOrb), No:'
            else if(reim .eq. 4 ) then
               ccc = 'Abs(NatOrb)^2, No:'
            end if

         else
            if(reim .eq. 0 ) then
               ccc = 'Abs(SPF), No:'
            else if(reim .eq. 1 ) then
               ccc = 'Re(SPF), No:'
            else if(reim .eq. 2 ) then
               ccc = 'Im(NatOrb), No:'
            else if(reim .eq. 3 ) then
               ccc = 'Re(SPF), Im(SPF), No:'
            else if(reim .eq. 4 ) then
               ccc = 'Abs(SPF)^2, No:'
            end if
         end if
         sc = slen(ccc)
         
         write(iaus,'(2a,i3,a,i2,2a)') 'set title "',ccc(1:sc), espf,',&
                  s =',state, filein(anin:laein) ,'"'
         write(iaus,'(a)') 'set style data lines'
         write(iaus,'(a)') 'set nologscale'
         if(lpgrid) write(iaus,'(a)') 'set grid'
         if( xmax .gt. xmin ) then
            write(iaus,'(a,f8.3,a,f8.3,a)') &
               'set xrange [',xmin,':',xmax,']'
         endif

         if (outform .ne. 1) then
            if (reim .eq. 0  .or.  reim .eq. 4) then
               if( ymax .le. 0.0_dop ) then
                  write(iaus,'(a)') 'set yrange [0:*]'
               else
                  write(iaus,'(a,f9.5,a)') 'set yrange [0:',ymax,']'
               end if
            else
               if( ymax .gt. 0.0_dop ) write(iaus,'(a,f9.5,a,f9.5,a)')&
                    'set yrange [',-ymax,':',ymax,']'
            end if

         else if (outform .eq. 1) then
            write(iaus,'(a)') 'set surface'
            write(iaus,'(a)') 'set contour'
            write(iaus,'(a)') 'set cntrparam levels auto'
            write(iaus,'(a)') '#set view 0,0,1,1'
            write(iaus,'(a)') 'set view 60.0,30.0,1,1'
            write(iaus,'(3a)')'set xlabel "',modelabel(dof)(1:9),'"'
            write(iaus,'(a)') 'set ylabel "Time"'
            write(iaus,'(a)') 'set zlabel ""'
            write(iaus,'(a)') 'set origin 0.,0.'
            write(iaus,'(a)') 'set size 1.0, 1.0'
            if(lsmooth)  then
               write(iaus,'(a)')  'set cntrparam cubicspline'
               write(iaus,'(a)')  'set cntrparam points 5'
            else
               write(iaus,'(a)')  'set cntrparam linear'
            end if
            write(iaus,'(a)') 'splot "-" using 1:2:3 ; pause -1'
         endif
      end if


!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw


!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! assign work arrays
! NB work arrays must be at end of mc, mr, mi, ml arrays.
!-----------------------------------------------------------------------
      workcdim=max(workcdim,sdim*maxspf)
      workcdim=max(workcdim,maxdim)
      if (citype .gt. 0) workcdim=max(workcdim,maxblock)
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      if (citype .gt. 0) workidim=max(workidim,maxblock)
      miworki = midim
      midim   = miworki+workidim

!-----------------------------------------------------------------------
! call analysis routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      write(ilog,'(/,2a)') 'psi-file : ', filein (1:laein)
      write(ilog,'(2a)')   'output   : ', fileout(1:laeout)
      write(ilog,'(2a,/)') 'dvr-file : ', dname(1:dlaenge)//'/dvr'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(dicht1))deallocate(dicht1)
if(allocated(dicht2))deallocate(dicht2)
if(allocated(dicht3))deallocate(dicht3)
if(allocated(dicht4))deallocate(dicht4)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(dicht1(dmatdim))
allocate(dicht2(dmatdim))
allocate(dicht3(d3matdim))
allocate(dicht4(dmatdim))

      m=dofspf(dof)
      zeig=sdim*maxspf+1
      call subplspf(mc(mcpsi),ms(mspsi),mi(mijindx),mr(mrort),&
           mc(mcworkc),subdim(m),dim(m,state),mr(mrtrafo),&
           dicht1,dicht2,dicht3,dicht4,mc(zeig),mi(miworki) )

deallocate(mc,mr,mi,ml,ms,mf,dicht1,dicht2,dicht3,dicht4)
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_psidef

      stop
 999  continue
      message= 'Error reading :'//filename(1:c5-15)
      write(6,'(a)') message
      stop

 900  continue
      routine='showspf'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      if(filename(1:ilbl) .eq. fileout(1:ilbl))&
      write(6,'(/a)') ' ###  Use -w option to overwrite! ###' 
      call errormsg
 1000 continue
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg

      stop
      contains

!***********************************************************************
!
!     HERE IS THE SUBROUTINE TO DO THE ANALYSIS
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! provide dummy arguments of arrays needed for analysis
!  e.g. psi to match argument mc(mcpsi)
!***********************************************************************

      subroutine subplspf (psi,spsi,jindx,ort,spfs,subdim1,dim1,&
                 trafo,dicht1,dicht2,dicht3,dicht4,workc,worki)

use global
use aglobal
use griddatmod
use psidef
use channels
use datenmod
use compdatmod
use compdat1mod
use plot

use rdfilesmod, only: rdpsi
use ciselect, only: venpsi
use densitymod, only: denmat
use rddvrmod
use logdat, only: errormsg

      implicit none

      integer(long) :: dof,state,espf
      common /iplspf/ dof,state,espf

      integer(long), external                            :: slen
      integer(long)                                      :: chkdvr,m,s,e,f,g,n,&
                                                            sc,ggdim,iblock,p1,&
                                                            kmin,kmax,kmin2,dk
      integer(long)                                      :: mmmin,mmmax,mmmin2,&
                                                            dm,p2,i
      integer(long), intent(in)                          :: subdim1,dim1
      integer(long), dimension(1)                        :: idum,idum1,idum2
      integer(long), dimension(adim), intent(out)        :: jindx
      integer(long), dimension(workidim), intent(out)    :: worki
      real(dop)                                          :: epsmat,time
      real(dop), dimension(1)                            :: rdum,rdum1,rdum2
      real(dop), dimension(ortdim), intent(out)          :: ort
      real(dop), dimension(dvrdim), intent(out)          :: trafo
      real(dop), dimension(d3matdim), intent(out)        :: dicht3
      complex(dop), dimension(1)                         :: zdum,zdum1,zdum2,zdum3
      complex(dop), dimension(dgldim), intent(out)       :: psi
      complex(dop), dimension(dmatdim), intent(out)      :: dicht1,dicht2
      complex(dop), dimension(dmatdim), intent(out)      :: dicht4
      complex(dop), dimension(workcdim), intent(out)     :: workc
      complex(dop), dimension(subdim1,dim1), intent(out) :: spfs
      complex(sip), dimension(spsidim), intent(out)      :: spsi
      character(len=6)                                   :: fslab
      character(len=80)                                  :: ccc
      logical(kind=4)                                    :: lexist
      integer(long)                                      :: ilbl,jlbl
      character(len=c2)                                  :: rdtag

      fslab = ' fs,  '
      if( fs .lt. 2.0_dop ) fslab = '      '

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! read data from dvr file if desired
! edit which data (set dvrdata(x) to .true. and substitute duumy array)
!-----------------------------------------------------------------------
      if (lrddvr) then
         call zerovxl(dvrdata,12)
         dvrdata(1)=.true.
         dvrdata(2)=.true.
         dvrdata(4)=.false.
         dvrdata(5)=.false.
         dvrdata(6)=.false.
         dvrdata(7)=.false.
         dvrdata(8)=.false.
         dvrdata(9)=.false.
         dvrdata(11)=.false.
         dvrdata(12)=.false.
         chkdvr=1
         call rddvr(ort,trafo,rdum,rdum1,zdum,&
              zdum1,idum,zdum2,zdum3,idum1,idum2,rdum2,chkdvr)
      endif

      if(lsmooth) then
         ccc = 'smooth csplines  notitle ; pause -1'
         if(reim .eq. 0) ccc = 'us 1:2 '//trim(ccc)
         if(reim .eq. 1) ccc = 'us 1:3 '//trim(ccc)
         if(reim .eq. 2) ccc = 'us 1:4 '//trim(ccc)
         if(reim .eq. 3) ccc = 'us 1:3 smooth csplines tit "Re",'&
                   //' "-" us 1:4 smooth csplines tit "Im" ; pause -1'
         if(reim .eq. 4) ccc = 'us 1:2 '//trim(ccc)
      else
         ccc = 'notitle ; pause -1'
         if(reim .eq. 0) ccc = 'us 1:2 '//trim(ccc)
         if(reim .eq. 1) ccc = 'us 1:3 '//trim(ccc)
         if(reim .eq. 2) ccc = 'us 1:4 '//trim(ccc)
         if(reim .eq. 3) ccc = 'us 1:3  tit "Re",'&
                        //' "-" us 1:4 tit "Im" ; pause -1'
         if(reim .eq. 4) ccc = 'us 1:2 '//trim(ccc)
      end if
      sc = slen(ccc)

      time=tinit/fs
      m=dofspf(dof)
      if( m .le. 0) then
         write(6,'(a,i3,a)')'The chosen DOF is not dynamic. Choose again.'
         stop
      end if
      e=espf
      s=state
      if( s .le. 0 .or. s .gt. nstate) then
         write(6,'(a,i3,a)')'The chosen state is out of range. Choose again.'
         stop
      end if
      if (dentype .eq. 1) then
         if( e .le. 0  .or.  e .gt. ddim(m,s,s) ) then
            write(6,'(a,i3,a)')'The orbital number must lie between 1 and',&
                                ddim(m,s,s), '     #####  STOP  #####'
            stop
         end if
      else 
         if( e .le. 0  .or.  e .gt. dim(m,s) ) then
            write(6,'(a,i3,a)')'The orbital number must lie between 1 and',&
                                dim(m,s), '     #####  STOP  #####'
            stop
         end if
      end if
      if (dof .eq. feb .and. febcont .gt. 0 .and. s .gt. elbnd) dof = febcont
      f=dof

      if (outform .eq. 3) then
         time = tinit/fs + indx*step*out2/fs
         write(iaus,'(a,f6.1,3a)') 'set label "',time,fslab,&
              modelabel(dof)(1:12),'" at graph 0.01,0.95'
         write(iaus,'(a,i3,2x,a)') 'plot "-"  index ',indx,ccc(1:sc)
      else if (outform .eq. 4) then
         write(iaus,*) out2/fs
         write(iaus,*) gdim(dof)
         do g=1,gdim(dof)
            write(iaus,*) ort(zort(dof)-1+g)
         enddo
      endif

   10 continue
      if(lshowrst) then
         if (filever(irst) .lt. 11.0_dop) then
            read (irst,err=999) time
            if (citype .gt. 0) then
               read(irst,err=999) (jindx(i),i=1,adim)
            endif
            read (irst,err=999) (psi(i),i=1,dgldim)
         else
            rewind(irst)
            do
  20           continue
               read(irst,iostat=chkerr(1),err=20) rdtag
               if (rdtag(1:1) .ne. '<') cycle

               ilbl=2
               jlbl=index(rdtag,'>')-1

               if (rdtag(ilbl:jlbl) .eq. 'psi') then
                  read (irst,err=999) (psi(i),i=1,dgldim)
               else if (rdtag(ilbl:jlbl) .eq. 'end restart file') then
                  exit
               endif
            enddo
         endif

      else
         call rdpsi(ipsi,psi,spsi,jindx,zdum,idum,zdum1,idum1)
         if (lend) go to 77
      endif

      if (dentype .eq. 1) then
         call cpmxz(psi(dzetf(m,s,s)),spfs,subdim(m),ddim(m,s,s))

      else
         if (lnatorb) then
            epsmat=1.0e-10_dop
            if (citype .gt. 0) then
               iblock=vdim(m,s)*dim(m,s)*ndim(m,s)
               call venpsi(m,jindx(zpsi(s)),worki,jindx(zpsi(s)),&
                    block(s),iblock,block(s),&
                    nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim,&
                    workc,psi(zpsi(s)))
               call rm1hxxxzz(workc,dicht2(dmat(m,s)),vdim(m,s),&
                   dim(m,s),ndim(m,s))
            else 
               call rm1hxxxzz(psi(zpsi(s)),dicht2(dmat(m,s)),vdim(m,s),&
                    dim(m,s),ndim(m,s))
            endif
! spfs used as workspace in denmat
            call beautify(dicht2(dmat(m,s)),dim(m,s))
            call denmat(dicht1(dmat(m,s)),dicht2(dmat(m,s)),&
                 dicht3(d3mat(m,s)),dicht4(dmat(m,s)),dim(m,s),&
                 epsmat)
            call mqxczz(psi(zetf(m,s)),dicht4(dmat(m,s)),spfs,&
                 subdim(m),dim(m,s))        
         else
            call cpmxz(psi(zetf(m,s)),spfs,subdim(m),dim(m,s))
         endif
      endif

      if (outform .eq. 0) then
         write(iaus,'(a)') 'set nolabel'
         if(time .lt. 10.d0) then
            write(iaus,'(a,f6.3,3a)') 'set label "',time,fslab,&
              modelabel(f)(1:12),'" at graph 0.01,0.95'
         else
            write(iaus,'(a,f6.1,3a)') 'set label "',time,fslab,&
              modelabel(f)(1:12),'" at graph 0.01,0.95'
         endif
         write(iaus,'(2a)') 'plot "-" ',ccc(1:sc)

      else if (outform .eq. 2) then
         ccc = ccc(1:sc-2)//'0.3 '
         write(iaus,'(a)') 'set nolabel'
         write(iaus,'(a,f6.1,3a)') 'set label "',time,fslab,&
            modelabel(f)(1:12),'" at graph 0.01,0.95'
         write(iaus,'(2a)') 'plot "-" ',ccc(1:sc+1)
      endif

      ggdim = gdim(f)
      if(basis(f) .eq. 12 .or. basis(f) .eq. 20 .or. basis(f) .eq.21)&
           ggdim = ipbaspar(3,f)
!...Determine the appropriate 'column' of trafo(gdim,gdim,kdim) for KLeg/PLeg
      p1=0
      if( basis(f) .eq. 7 .or. basis(f) .eq. 20) then  ! KLeg
         kmin=ipbaspar(1,f+1)
         kmax=ipbaspar(2,f+1)
         kmin2=ipbaspar(3,f+1)
         p1=(min(kmax,kmin2)-kmin)
         dk=max(ipbaspar(5,f+1),1)
         p1=(p1+dk-1)/dk
!k         p1=p1*gdim(f)**2
         p1=p1*ggdim**2
      elseif( basis(f) .eq. 9 .or. basis(f) .eq. 21) then           ! PLeg
         kmin=ipbaspar(1,f+1)
         kmax=ipbaspar(2,f+1)
         kmin2=ipbaspar(4,f+1)
!k         p1=(min(kmax,kmin2)-kmin)*gdim(f)**2
         p1=(min(kmax,kmin2)-kmin)*ggdim**2
      elseif( basis(f) .eq. 33 ) then          ! Wigner
         kmin=ipbaspar(2,f)
         kmax=ipbaspar(3,f)
         kmin2=ipbaspar(5,f)
         p2=min(kmax,kmin2)-kmin
         if (basis(f+1).eq.8) then
            dk=ipbaspar(5,f+1)
            p2=(p2+dk-1)/dk
         endif
         p2=p2*gdim(f)**2
         mmmin=ipbaspar(6,f)
         mmmax=ipbaspar(7,f)
         mmmin2=ipbaspar(9,f)
         p1=min(mmmax,mmmin2)-mmmin
         if (basis(f+2).eq.8) then
            dm=ipbaspar(5,f+2)
            p1=(p1+dm-1)/dm
         endif
         p1=p1*ipbaspar(4,f)*gdim(f)**2 + p2
      endif

      if (dentype .eq. 1) then
         call outrho1(spfs(1,e),trafo(zdvr(f)+p1),ort(zort(f)),f,nspfdof(m),&
                     vgdim(f),gdim(f),ngdim(f),ggdim,time)
      else
         call outspf(spfs(1,e),trafo(zdvr(f)+p1),ort(zort(f)),f,nspfdof(m),&
                     vgdim(f),gdim(f),ngdim(f),ggdim,time)
      endif

      if (outform .eq. 0 .or. outform .eq. 2) then
         write(iaus,'(a,/)') 'e'
      else if(outform .eq. 1) then
         write(iaus,'(a)')
      else if(outform .eq. 3) then
         write(iaus,'(/,a)')
      endif

      time=time+out2/fs

      do n=1,step-1
         call rdpsi(ipsi,psi,spsi,jindx,zdum,idum,zdum1,idum1)
         if (lend) go to 77
         time=time+out2/fs
      enddo

      if (.not. lshowrst) go to 10
   77 continue

      close(ipsi)
      close(iaus)

      if(lprint) then
         inquire(file="printplot",exist=lexist) 
         open(iaus,file="printplot",form='formatted',status='unknown')
         if(.not.lexist) then
            open(iaus,file="printplot",form='formatted')
            write(iaus,'(a)') 'set terminal postscript'
            write(iaus,'(a)')'set output "| lpr" '
            write(6,'(a)')'File printplot created.'
         end if
         rewind iaus
         read(iaus,'(a1)') gnucmd 
         read(iaus,'(a30)') gnucmd 
         close(iaus)
         write(6,'(a)')'Press enter twice to print. Ctrl C to abort.'
         write(6,'(3a)')'Abort and edit printplot if your printcommand',&
                        ' is not: '//gnucmd(11:29)
      end if
      if(lagnu) then
         gnucmd = 'gnuplot -persist "'//trim(fileout)//'"'
         call excmd(gnucmd)
      end if
      if(lprint) then
         gnucmd = 'gnuplot -persist '//'"printplot"  "'&
              //fileout(1:slen(fileout))//'"'
         call excmd(gnucmd)
      end if

      return
  999 routine='plotsys  (showsys.F)'
      message='Error reading restart file'
      call errormsg

      end subroutine subplspf

!#######################################################################

      subroutine outspf(spfs,trafo,ort,f,nf,vgdim1,gdim1,ngdim1,ggdim,time)

use global
use aglobal
use griddatmod
use psidef
use channels
use datenmod
use compdatmod
use compdat1mod
use plot

      implicit none

      integer(long)                                  :: g,v,n,nf
      integer(long), intent(in)                      :: f,vgdim1,gdim1,&
                                                        ngdim1,ggdim
      real(dop)                                      :: dabsspf
      real(dop), intent(in)                          :: time
      real(dop), dimension(2048), save               :: weight
      real(dop), dimension(gdim1), intent(out)       :: ort
      real(dop), dimension(ggdim,gdim1), intent(out) :: trafo
      complex(dop), dimension(vgdim1,gdim1,ngdim1), intent(in)  :: spfs
      complex(dop), dimension(gdim1)                 :: spfs1d
      integer(long), save                            :: fold = -1

!-----------------------------------------------------------------------
! Determine DVR weights.
!-----------------------------------------------------------------------
      if( fold .ne. f .and. .not.lnw ) then
         fold = f
         if( gdim1 .gt. 2048 ) then
            write(6,'(a,i5,a)') 'gdim =',gdim1,' Grid too large! STOP'
            stop
         end if
         call dvrweights(trafo,ort,gdim1,ggdim,weight,basis(f),&
                         rpbaspar(1,f),ipbaspar(1,f),ipbaspar(1,f+1),&
                         .false.,lnw)
      endif

!-----------------------------------------------------------------------
! Multi-D SPFs are integrated over othe DOFs
!-----------------------------------------------------------------------
      if (nf .gt. 1) then
         spfs1d = 0.0
         do n=1,ngdim1
            do g=1,gdim1
               do v=1,vgdim1
                  spfs1d(g) = spfs1d(g) + &
                                     dconjg(spfs(v,g,n))*spfs(v,g,n)
               enddo
            enddo
         enddo
      else
         do g=1,gdim1
            spfs1d(g) = spfs(1,g,1) 
         enddo
      endif

!-----------------------------------------------------------------------
! Divide out DVR weight from single-particle functions and output
!-----------------------------------------------------------------------
      do g=1,gdim1
         if(.not.lnw) spfs1d(g) = spfs1d(g)/weight(g)
         dabsspf=dble(spfs1d(g))**2 + dimag(spfs1d(g))**2
         if(reim .ne. 4) dabsspf=dsqrt(dabsspf)
         if (outform.eq.0 .or. outform.eq.2 .or. outform.eq.3 ) then
            write(iaus,'(4(1x,f18.10))') ort(g),dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         else if (outform .eq. 1) then
            write(iaus,'(5(1x,f12.8))') ort(g),time,dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         else if (outform .eq. 4) then
            write(iaus,*) dabsspf
         endif
      enddo

      if(reim .eq. 3) then
         write(iaus,'(a)') 'e'
         do g=1,gdim1
            dabsspf=dble(spfs1d(g))**2 + dimag(spfs1d(g))**2
            dabsspf=dsqrt(dabsspf)
            write(iaus,'(4(1x,f18.10))') ort(g),dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         enddo
      end if


      return
      end subroutine outspf

!#######################################################################

      subroutine outrho1(spfs,trafo,ort,f,nf,vgdim1,gdim1,ngdim1,ggdim,time)

use global
use aglobal
use griddatmod
use psidef
use channels
use datenmod
use compdatmod
use compdat1mod
use plot

      implicit none

      integer(long)                                  :: g,v,n,nf
      integer(long), intent(in)                      :: f,vgdim1,gdim1,&
                                                        ngdim1,ggdim
      real(dop)                                      :: dabsspf
      real(dop), intent(in)                          :: time
      real(dop), dimension(2048), save               :: weight
      real(dop), dimension(gdim1), intent(out)       :: ort
      real(dop), dimension(ggdim,gdim1), intent(out) :: trafo
      complex(dop), dimension(vgdim1,gdim1,ngdim1,vgdim1,gdim1,ngdim1), &
                                                      intent(in)  :: spfs
      complex(dop), dimension(gdim1)                 :: spfs1d
      integer(long), save                            :: fold = -1

!-----------------------------------------------------------------------
! Determine DVR weights.
!-----------------------------------------------------------------------
      if( fold .ne. f .and. .not.lnw ) then
         fold = f
         if( gdim1 .gt. 2048 ) then
            write(6,'(a,i5,a)') 'gdim =',gdim1,' Grid too large! STOP'
            stop
         end if
         call dvrweights(trafo,ort,gdim1,ggdim,weight,basis(f),&
                         rpbaspar(1,f),ipbaspar(1,f),ipbaspar(1,f+1),&
                         .false.,lnw)
      endif

!-----------------------------------------------------------------------
! Take 1D trace of Multi-D SPDOs
!-----------------------------------------------------------------------
      spfs1d = 0.0
      do n=1,ngdim1
         do g=1,gdim1
            do v=1,vgdim1
               spfs1d(g) = spfs1d(g) + spfs(v,g,n,v,g,n)
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! Divide out DVR weight from single-particle functions and output
!-----------------------------------------------------------------------
      do g=1,gdim1
         if(.not.lnw) spfs1d(g) = spfs1d(g)/weight(g)
         dabsspf=dble(spfs1d(g))**2 + dimag(spfs1d(g))**2
         if(reim .ne. 4) dabsspf=dsqrt(dabsspf)
         if (outform.eq.0 .or. outform.eq.2 .or. outform.eq.3 ) then
            write(iaus,'(4(1x,f18.10))') ort(g),dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         else if (outform .eq. 1) then
            write(iaus,'(5(1x,f12.8))') ort(g),time,dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         else if (outform .eq. 4) then
            write(iaus,*) dabsspf
         endif
      enddo

      if(reim .eq. 3) then
         write(iaus,'(a)') 'e'
         do g=1,gdim1
            dabsspf=dble(spfs1d(g))**2 + dimag(spfs1d(g))**2
            dabsspf=dsqrt(dabsspf)
            write(iaus,'(4(1x,f18.10))') ort(g),dabsspf,&
                 dble(spfs1d(g)),dimag(spfs1d(g))
         enddo
      end if


      return
      end subroutine outrho1

!#######################################################################

!     This subroutine removes the degeneracies of the density matrix.
!     The natorbs thus become identical to the SPFs for a Hartree product.
  
      subroutine  beautify(dicht2,dim)

      implicit none

      integer(long)                                   :: e
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(inout) :: dicht2

        do e=1,dim
           if( dble(dicht2(e,e)) .lt. 1.0e-8_dop/dble(e))&
               dicht2(e,e)=dicht2(e,e) + 1.0e-8_dop/dble(e)
        enddo

      return
      end subroutine  beautify

!-----------------------------------------------------------------------
!***********************************************************************
!-----------------------------------------------------------------------
      subroutine plspfinp(stat)

use global
use aglobal
use griddatmod
use psidef
use datenmod
use compdatmod
use compdat1mod
use plot

      implicit none

      integer(long) :: dof,state,espf
      common /iplspf/ dof,state,espf

      integer(long), external       :: myiargc
      integer(long)                 :: iarg,ilbl,i,ierr
      character(len=80)             :: buf,string
      character(len=*), intent(out) :: stat

!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! provide default input and output filenames 
!-----------------------------------------------------------------------
      lshowrst = .false.
      lshowpsi = .true.
      filein  = '  '
      fileout = '  '
      name   = '.'
      dname   = '.'

      dof=0
      state=1
      espf=0

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! set default flags
!
! stat : status of output file   
!-----------------------------------------------------------------------
      stat   = 'new'
      lnw    = .false.
      lgnu   = .false.
      lagnu  = .false.
      lprint = .false.
      lpgrid = .false.
      lnatorb= .true.
      lsmooth= .false.
      step   = 1
      xmax   = 0.0_dop
      xmin   = 1.0_dop
      ymax   = 0.0_dop

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! read options 
! comment out or add options as required
!-----------------------------------------------------------------------
      iarg = myiargc() 
      if (iarg .lt. 1) goto 200 

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf(1:3) .eq. '-o ' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
         else if( buf(1:4) .eq. '-no ' ) then
            fileout='no'
         else if( buf(1:3) .eq. '-f ' ) then
            i = i+1
            call mygetarg(i,buf)
            filein = buf
         else if( buf(1:3) .eq. '-i ') then
            i = i+1
            call mygetarg(i,buf)
            name = buf
         else if( buf(1:2) .eq. '-I' ) then
            i = i+1
            outform = 3
            call mygetarg(i,buf)
            read(buf,*,err=999) indx
         else if( buf .eq. '-rst') then
            lshowrst=.true.
         else if( buf(1:3) .eq. '-S ' ) then
            outform = 0
         else if( buf(1:3) .eq. '-T ' ) then
            outform = 1
         else if( buf(1:3) .eq. '-M ' ) then
            outform = 2
         else if( buf(1:3) .eq. '-Ma' ) then
            outform = 4
         else if( buf(1:4) .eq. '-spf' ) then
            lnatorb=.false.
         else if( buf(1:4) .eq. '-nat' ) then
            lnatorb=.true.
         else if( buf(1:4) .eq. '-nw ' ) then
            lnw = .true.
         else if( buf(1:3) .eq. '-G ' ) then
            lpgrid = .true.
         else if( buf(1:3) .eq. '-n ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (ierr .ne. 0  .or.  step .le. 0) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -n',&
                 ' must be a positive integer. Not : ', buf
               stop
            end if
         else if( buf(1:3) .eq. '-x ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,err=998) xmin
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,err=998) xmax
         else if( buf(1:3) .eq. '-y ' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,err=999) ymax
         else if( buf(1:4) .eq. '-sm ' .or. buf(1:7).eq.'-smooth') then
            lgnu = .true.
            lsmooth = .true.
         else if( buf(1:5) .eq. '-abs ' ) then
            reim = 0
         else if( buf(1:4) .eq. '-re ' ) then
            lgnu = .true.
            reim = 1
         else if( buf(1:4) .eq. '-im ' ) then
            lgnu = .true.
            reim = 2
         else if( buf(1:6) .eq. '-reim ' ) then
            lgnu = .true.
            reim=3
         else if( buf(1:6) .eq. '-abs2 ' ) then
            reim = 4
         else if( buf(1:3) .eq. '-w ' ) then
            stat='unknown'
         else if( buf(1:6) .eq. '-nofs ' ) then
            fs = 1.0_dop
         else if( buf(1:3) .eq. '-g ' ) then
            lgnu = .true.
         else if( buf(1:3) .eq. '-a ' ) then
            lgnu  = .true.
            lagnu = .true.
            stat  = 'unknown'
         else if( buf(1:3) .eq. '-p ' ) then
            lgnu   = .true.
            lprint = .true.
         else if( buf(1:3) .eq. '-h ' .or. buf(1:3) .eq. '-? ' ) then
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(1x,a,/1x,a,/21x,a,/21x,a)')&
       'Purpose: To selectively plot (uncombined) natorbs or spf''s.',&
       'Usage  : showspf [-T|-S|-I|-M|-Ma] [-f -i -rst -o -a -g -p -x -y',&
       ' -sm -abs -abs2 -re -im -reim -n -w -nw -nofs',&
       ' -spf -nat -G -ver -h -?]  [fx sx x] .'
            write(6,'(/,a)') 'Output Formats : '
            write(6,'(2a)') &
                    ' -S      : "Step through" by pressing RETURN.',&
                    ' (default).'
            write(6,'(a)') &
                    ' -M      : "Movie", one picture every second.'
            write(6,'(2a)') &
                    ' -I id   : "Index file", id=0 shows the first ',&
                    'picture, id=1 the second, etc.'
            write(6,'(a)') &
                    ' -T      : "3D time file" (grid file).'
            write(6,'(a)') &
                    ' -Ma     : Mathematica file (for dynamic plots).'

            write(6,'(/,a)') 'Options  : '
            write(6,'(2a,/,9x,a)') &
                     ' -f FILE : The wavefunction is read '&
                    ,'from file FILE rather than from ./psi'&
            ,'  The string FILE may be a relative or a full path-name.'
            write(6,'(a)') ' -i DIR  : Use DIR as name directory.'
            write(6,'(2a)')&
                     ' -rst    : the restart file is read, '&
                    ,'rather than the psi file.'
            write(6,'(2a,/,11x,a/,11x,a,/11x,a)') &
                     ' -o FILE : The output is written to file',&
                     ' FILE*_f* rather than','to ./natorb*_f*',&
             'The string FILE may be a relative or a full path-name.',&
             'If FILE=no, i.e. "-o no", no output-file will be opened.'
            write(6,'(2a)') &
                     ' -g      : GNUplot comand lines are written to ',&
                     'the output file(s).'
            write(6,'(2a)') &
                     ' -a      : GNUplot is called automatically.',&
                     ' (sets -g and -w).'
            write(6,'(2a)') &
                     ' -p      : Print GNUplot (sets -g). ',&
                                'Allowed only with -I or -T.' 
            write(6,'(2a)') &
                     ' -sm     : Smooth the plot-data by splines.',&
                     ' (only for GNUplot, sets -g).'
            write(6,'(a)') &
                     ' -abs    : Plot absolute values. (default).'
            write(6,'(a)') &
                     ' -abs2   : Plot absolute values squared. '
            write(6,'(a)') &
           ' -re     : Plot real part (only for GNUplot, sets -g).'
            write(6,'(a)') &
           ' -im     : Plot imaginary part (only for GNUplot, sets -g).'
            write(6,'(2a)') &
                     ' -reim   : Plot both imaginary and real parts',&
                     ' (only for GNUplot, sets -g).'
            write(6,'(2a,/11x,a)') &
                     ' -spf    : Plot the ',&
                     'single-particle functions','rather than natorbs.'
            write(6,'(2a,/11x,2a)') &
                     ' -nat    : Plot the natural ',&
                     'orbitals rather','than spf''s (default).',&
                     ' Note x=1 -> highest occupied natorb.'
            write(6,'(2a)')&
                     ' -nofs   : No transformation to fs.',&
                     " Use when 'time-not-fs' was set in mctdh."
            write(6,'(2a)') &
                     ' -nw     : No weights are employed.',&
                     ' Plot the "naked" DVR values.'
            write(6,'(2a)') ' -n step : Show only every', &
                  ' step-th picture. (Only useful for -S, -M).'
            write(6,'(a)') &
               ' -x xmin xmax : Set the abszisse length.'
            write(6,'(2a)') &
               ' -y ymax : Set the ordinate length (default: auto).',&
               ' Ignored for -T and -Ma.'
            write(6,'(a)') ' -G      : Plot gridlines.'
            write(6,'(a)') &
               ' -w      : An existing output file is overwritten.'
            write(6,'(a,/a)')& 
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(a)') &
                     ' -ver    : Version information about the program '

            write(6,'(/,a)')& 
          ' The arguments select the spf (x), dof (fx) and state (sx) '
            write(6,'(a)') &
            ' Replace x by appropriate number, any order is allowed.'
            write(6,'(2a)') ' If no argument is given, ',&
                 'the user is prompted for the missing argument.'  
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      if (iarg .lt. i) goto 200

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! get any compulsory arguments if in argument list
!-----------------------------------------------------------------------
  150 continue
      call mygetarg(i,string)
      if (string(1:1) .eq. 'f' .or. string(1:1) .eq. 'F') then
         read(string(2:20),*,err=899) dof
      else if (string(1:1) .eq. 's' .or. string(1:1) .eq. 'S') then
         read(string(2:20),*,err=899) state
      else if (string(1:1) .eq. 'x' .or. string(1:1) .eq. 'X') then
         read(string,*,err=899) espf
      else
         read(string,*,err=899) espf
      endif
      i=i+1
      if (i .le. iarg) goto 150

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! fill in any arguments not given on command line
!-----------------------------------------------------------------------
 200  continue
      if (dof .eq. 0) then
         write(6,*) 'DOF  :'
         read(5,*) dof
         write(6,*)
      endif
      if (espf .eq. 0) then
         write(6,*) 'SPF  :'
         read(5,*) espf
         write(6,*)
      endif

!-----------------------------------------------------------------------
! set up output file name
!-----------------------------------------------------------------------
      if (fileout .eq. 'no') go to 500
      if (fileout(1:2) .eq. '  ') then
         if (lnatorb) then
            fileout='natorb'
         else
            fileout='spf'
         endif
      endif
      ilbl = index(fileout,' ')
      if (espf .lt. 10) then
         write(fileout(ilbl:ilbl),'(i1)') espf
         ilbl=ilbl+1
      else if (espf .gt. 10) then
         write(fileout(ilbl:ilbl+1),'(i2)') espf
         ilbl=ilbl+2
      endif
      fileout(ilbl:ilbl+1)='_f'
      ilbl=ilbl+2
      if (dof .lt. 10) then
         write(fileout(ilbl:ilbl),'(i1)') dof
         ilbl=ilbl+1
      else if (dof .gt. 10) then
         write(fileout(ilbl:ilbl+1),'(i2)') dof
         ilbl=ilbl+2
      endif
      if (state .gt. 1) then
         fileout(ilbl:ilbl+1)='_s'
         ilbl=ilbl+2
         if (state .lt. 10) then
            write(fileout(ilbl:ilbl),'(i1)') state
            ilbl=ilbl+1
         else if (state .gt. 10) then
            write(fileout(ilbl:ilbl+1),'(i2)') state
            ilbl=ilbl+2
         endif
      endif
      
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! convert filenames to absolute
!-----------------------------------------------------------------------
  500 continue
      laenge=index(name,' ')-1
      if (filein .eq. ' ') then
         if (lshowrst) then
            filein = name(1:laenge)//'/restart'
         else
            filein = name(1:laenge)//'/psi'
         endif
      endif

      laein=index(filein,' ')-1
      call abspath(filein,laein)

      dname = name
      dlaenge=index(dname,' ')-1
      call abspath(dname,dlaenge)

      if (fileout .ne. 'no') then
         laeout=index(fileout,'  ')-1
         call abspath(fileout,laeout)
      end if

      if (lprint  .and. .not.(outform.eq.1 .or. outform.eq.3)) then
         write(6,'(a)') 'Printing is possible only for I and T formats.'
         write(6,'(a)') ' --- STOP ---'
         stop
      end if

      if (outform .eq. 4) then
         ymax   =  0.0_dop
         lgnu   = .false.
         lagnu  = .false.
         lprint = .false.
      end if
      return 
 899  write(6,*) 'Cannot read argument, ',trim(string)
      stop
 999  write(6,*) 'Integer expected after -I and -y options '
      stop
 998  write(6,*) 'Two real numbers expected after -x  option '
      stop
      end subroutine plspfinp


      end  program showspf

