!***********************************************************************
!
!                        ANALYSE MODULE TRAFODB
!
! Reads and analyses a DB
! 
! 2020 Antonia Freibert
!
!
!***********************************************************************

      module trafodbmod

      use decimal, only : long,dop

!-----------------------------------------------------------------------
! options for finding CoIn
!-----------------------------------------------------------------------
      real(dop), save :: ecoin 

      end module trafodbmod

!-----------------------------------------------------------------------
!> @brief 
!> Plots derivatice couplings, locates intersections, etc, using data in a DB.
!>
!> @detail
!> Plots derivatice couplings, locates intersections, etc, using data in a DB.
!>
!> OPTIONS:
!>---------
!> 
!>
!>  -f FILE : The data is read from file FILE rather than from ./check 
!>            The string FILE may be a relative or a full path-name.
!>
!>  -o FILE : The geometries are written to file FILE.pl rather than to ./ddgeom.pl
!>            The string FILE may be a relative or a full path-name.
!>
!>  -ver    : Version information about the program.
!>
!>  -h      : Print this help text.
!>
!>  -?      : Print this help text.
!-------------------------------------------------------------------------

      program trafodb

use decimal, only: long,dop
use global
use aglobal
use keyunitsmod
use versions
use dvrdatmod
use logdat
use datenmod
use dirdyn
use rdopermod
use iodvrdef
use trafodbmod

use op1lib, only: zeromxd,cpvxd
use channels, only: filever, ioper
use mmlib, only: qqxtdd, qqxxdd,qqtxdd,mmxxdd,mmtxdd
use dd_db, only: dddb_rd, preparedb, getdbnrec, distdb
use dirdyn
use db_sql
use plotutils, only : mode_lab_corr

      implicit none

      integer(long)       :: i,ilbl,ilddtrans,j,jlbl
      integer(long)       :: chkdvr,lenpath,ndof1,nddstate
      integer(long), dimension(:), allocatable :: fdvr
      integer(long)       :: itask=0
      integer(long)       :: nargs
      integer(long), dimension(3)                     :: iargs = -999
      real(dop), dimension(:,:), allocatable :: coonm,nmcoo

      character(len=c5)                    :: filename,dbname,inpname
      character(len=c2)                    :: rdtag
      logical(kind=4)                      :: int2log
      logical                              :: lstep=.false.
      type(SQLITE_DATABASE)                :: dbin
      character(len=c1)       :: moderep
   
!-----------------------------------------------------------------------
!     Set defaults. Set up the input.

!-----------------------------------------------------------------------

      write(6,'(/,78a1,/)')  ('*',i=1,78)
      write(6,'(2a)') ' Analyse program "trafodb" ',&
                      '(Analyse DB files)'
      write(6,'(/,78a1,/)')  ('*',i=1,78)

      call default
      call adefault
      call setunits
      call trafoinp(itask,iargs,nargs)

!-----------------------------------------------------------------------
! open oper file
!
! get memory dimensions and allcoate memory
!-----------------------------------------------------------------------
      filename=operfile
      ilbl=index(filename,' ')-1
      open(ioper,file=filename,form='unformatted',status='old',&
           err=1000)
      write(6,'(2a)') ' The operator file is : ',filename(1:ilbl)
      rewind(ioper)

      call rdmemdim(ioper)

      call alloc_dvrdat
      call alloc_dirdyn
      call alloc_dddb
      
!-----------------------------------------------------------------------
! read the system information
!-----------------------------------------------------------------------
     allocate(fdvr(maxdim))
     chkdvr=1
     rewind(ioper)
     read(ioper,err=999) filever(ioper)
     call rddvrdef(ioper,chkdvr,ndof1,fdvr)
     deallocate(fdvr)

!-----------------------------------------------------------------------
! get ndofddpes + natm from oper file
!-----------------------------------------------------------------------
      rewind(ioper)
      do
   10    continue
         read(ioper,err=10,end=20) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'ndofddpes') then
            read(ioper) ndofddpes
         else if (rdtag(ilbl:jlbl) .eq. 'ndofdd') then
            read(ioper) ndofdd
         else if (rdtag(ilbl:jlbl) .eq. 'nddstate') then
            read(ioper) nddstate
         else if (rdtag(ilbl:jlbl) .eq. 'natm') then
            read(ioper) natm
         else if (rdtag(ilbl:jlbl) .eq. 'xatnam') then
            read(ioper) (xatnam(i),i=1,natm)
         else if (rdtag(ilbl:jlbl) .eq. 'lddtrans') then
            read(ioper) ilddtrans
            lddtrans=int2log(ilddtrans)
         else if (rdtag(ilbl:jlbl) .eq. 'ddxcoo0') then
            read(ioper) (ddxcoo0(i),i=1,ndofddpes)
         else if (rdtag(ilbl:jlbl) .eq. 'ddtransf') then
            read(ioper) ((ddtransf(j,i),j=1,ndofddpes),i=1,ndofdd)
         else if (rdtag(ilbl:jlbl) .eq. 'ddtransb') then
            read(ioper) ((ddtransb(j,i),j=1,ndofdd),i=1,ndofddpes)

         else if (rdtag(ilbl:jlbl) .eq. 'ddlaenge') then
            read(ioper,err=999) ddlaenge0
         else if (rdtag(ilbl:jlbl) .eq. 'ddname') then
            read(ioper,err=999) ddname0

         else if (rdtag(ilbl:jlbl) .eq. 'end oper info') then
            exit
         endif
      enddo
   20 continue
      close(ioper)

! form full path for DD DB
      ddname = ddname0
      ddlaenge = ddlaenge0
      if (ddname(1:1) .ne. '/') then
         lenpath = index(ddpath,' ')-1
         if (lenpath .gt. 0) then
            if (ddpath(lenpath:lenpath) .ne. '/') then
               ddname = ddpath(1:lenpath)//'/'//ddname0
            else
               ddname = ddpath(1:lenpath)//ddname0
            endif
            ddlaenge=index(ddname,' ')-1
            if (ddname(1:1) .ne. '/') call abspath(ddname,ddlaenge)
         else
            call abspath(ddname,ddlaenge)
         endif
      endif
      dbname = ddname(1:ddlaenge) // '/database.sql'
      ilbl=index(dbname,' ')-1
      write(6,'(2a)') ' The DB to be used is : ',dbname(1:ilbl)

!-----------------------------------------------------------------------
! open DB ro
! and store database pointer the_db as local dbin 
!-----------------------------------------------------------------------
      !Needed for NEWSQL
      call makeflagarray
      call makeinfoarray(nddstate)
      call dbopen(dbname,the_db,sqlvers,quanticsver,infoarray, flagarray,0)
      dbin=the_db

!-----------------------------------------------------------------------
! and look what is in DB
!-----------------------------------------------------------------------
      call dbgetentries( dbin, 6, lpesdb, lgradb, lhesdb, lapesdb, lagradb, &
                         lahesdb, ldipdb, ladipdb, lmocodb, lnactdb, &
                         ltransdb, lnmdb, lcasocdb, lsymdb, lcpesdb, &
                         lctransdb, lcivecdb, ltemplate, lqcfreq)

!-----------------------------------------------------------------------
! count records in DB
!-----------------------------------------------------------------------
      call dbgetnrec(dbin,dbnrec)
      write(6,'(a,i5)') ' No. records  : ',dbnrec

      if (dbnrec .le. 0) then
         print*
         print*,'No records found.'
         print*,'STOP'
         stop
      endif

      if (dbnrec .gt. dbmemdim) then
         print*,'Too many records: dbnrec > dbmemdim ',dbnrec,dbmemdim
         stop
      endif

!-----------------------------------------------------------------------
! resize trafo matrices
!-----------------------------------------------------------------------
      allocate(coonm(ndofdd,ndofddpes)) 
      allocate(nmcoo(ndofddpes,ndofdd)) 
      do i=1,ndofdd
         do j=1,ndofddpes
            nmcoo(j,i)=ddtransf(j,i)
            coonm(i,j)=ddtransb(i,j)
         enddo
      enddo

!-----------------------------------------------------------------------
! call routine to do transformation
!-----------------------------------------------------------------------
      call dotrafo(dbin,itask,iargs,nargs,dbnrec,ndofdd,ndofddpes,&
           nmcoo,coonm,ddxcoo0)

      call dealloc_dvrdat
      call dealloc_dirdyn
      call dealloc_dddb
      deallocate(coonm,nmcoo)

!-----------------------------------------------------------------------

      stop

  999 continue
      message= 'Error reading :'//filename(1:c5-15)
      write(6,'(a)') message
      stop
 1000 continue
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop

      contains

!#######################################################################
 
      subroutine dotrafo(dbin,itask,iargs,nargs,recordmax,nmodes,ncoo, & 
                 nmcoo,coonm,xcoo0)

      implicit none

      integer(long), intent(in) :: itask 
      integer(long), intent(in) :: nargs
      integer(long), dimension(*), intent(in) :: iargs
      integer(long)             :: nonact !number of coupterms
      integer(long), intent(in) :: ncoo, nmodes 
      integer(long), intent(in) :: recordmax  !number of structures
      integer                   :: i,m1,m2,s1,s2
      real(dop), dimension(ncoo,nmodes), intent(in) :: nmcoo !forward
      real(dop), dimension(nmodes,ncoo), intent(in) :: coonm !backward
      real(dop), dimension(ncoo), intent(in) :: xcoo0  ! reference geometry
      type(SQLITE_DATABASE), intent(in)      :: dbin

! default look at S0/S1 coupling
      s1 = 1
      if (nddstate .eq. 1) then
         s2 = 1
      else
         s2 = 2
      endif

! no. of NACTs
      nonact = 0
      do i=1,nddstate-1
         do j=i+1,nddstate
            nonact = nonact + 1
         end do
      end do

! Transform NACT
      if (itask .eq. 1) then
         m1 = iargs(1)
         m2 = iargs(2)
         if (nargs .eq. 4) then
            s1 = iargs(3)
            s2 = iargs(4)
         endif
         call nact2nm(dbin,nmodes,ncoo,recordmax,nmcoo,m1,m2,s1,s2)

! Transform Cartesian coords 
      else if (itask .eq. 2) then
         call coo2nm(dbin,nmodes,ncoo,recordmax,coonm,xcoo0)

! Extract and plot mixing in a plane
      else if (itask .eq. 3) then
         m1 = iargs(1)
         m2 = iargs(2)
         call subrotang(dbin,nmodes,recordmax,m1,m2)

! Transform gra
      else if (itask .eq. 4) then
         m1 = iargs(1)
         m2 = iargs(2)
         if (nargs .eq. 4) then
            s1 = iargs(3)
            s2 = iargs(4)
         endif
         call out_dcp(dbin,nmodes,ncoo,recordmax,m1,m2,s1,s2,nmcoo)
       
!create a new nact file where the sign of the vectors are consistent

      else if (itask .eq. 5) then
         m1 = iargs(1)
         m2 = iargs(2)
         call nact_correct(dbin,nmodes,ncoo,nonact,recordmax,m1,m2,nmcoo)
                       

      else if (itask .eq. 6) then
         m1 = iargs(1)
         m2 = iargs(2)
         call adverr(dbin,nmodes,recordmax,m1,m2)

! Transform Hessian 
      else if (itask .eq. 7) then
         call hess2freq(dbin,nmodes,ncoo,recordmax,nmcoo)

! Check Curl condition
      else if (itask .eq. 8) then
         m1 = iargs(1)
         m2 = iargs(2)
         call chkcurl(dbin,nmodes,ncoo,nmcoo,xcoo0,recordmax,m1,m2)

! Find nearest structure
      else if (itask .eq. 9) then
         call getnear(dbin,recordmax)

      end if

      return
      end subroutine dotrafo


!>----------------------------------------------------------------------
!> @brief
!> subroutine to evaluate input of trafodb
!>
!>----------------------------------------------------------------------

      subroutine trafoinp(itask,iargs,nargs)

use compdatmod
use compdat1mod

      implicit none

      integer(long), external        :: myiargc
      integer(long)                  :: iarg,ierr,i,j
      character(len=80)              :: buf
      integer(long), intent(out)     :: itask,nargs
      integer(long), dimension(*), intent(out) :: iargs

!-----------------------------------------------------------------------

      name ='./'
      operfile = ' '
      laenge=1
      filein  = '  '
      fileout = '  '
      fileexp = '  '
      lgnu = .false.
      ddpath = '..'
      inpname  = '  '

      ecoin = 0.001/uev

      iarg = myiargc() 
      if (iarg .lt. 1) then
         i = 1
         goto 200 
      endif

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if(buf .eq. '-i') then
            i=i+1
            call mygetarg(i,buf)
            name = buf
            laenge = index(name,' ')-1
         else if( buf(1:3) .eq. '-o ' ) then
            i=i+1
            call mygetarg(i,buf)
            fileout = buf
         else if( buf(1:4) .eq. '-op ') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if( buf(1:4) .eq. '-DB ' ) then
            i = i+1
            call mygetarg(i,buf)
            ddpath = buf
         else if( buf(1:4) .eq. '-f ') then
            i = i+1
            call mygetarg(i,buf)
            inpname = buf
         else if( buf(1:3) .eq. '-g ' ) then
            lgnu = .true.
         else if( buf(1:6) .eq. '-step ' ) then
            lstep = .true.
         else if ( buf .eq. '-eci' ) then
            i=i+1
            call mygetarg(i,buf)
            read(buf,*) ecoin

            call mygetarg(i+1,buf)
            if (buf .eq. ',') then
               i=i+2
               call mygetarg(i,buf)
               call keylowc(buf)
               call keyeunits(ecoin,buf,ierr)
               if (ierr .eq. 1) then
                  message = 'no energy unit after comma (eci).'
                  call errormsg
               endif
            endif

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(/,78a1,/)')  ('*',i=1,78)
            write(6,'(a)') ' Analyse program "TrafoDB" ' 
            write(6,'(1x,a)')&
       'Purpose : Reads the DB files and analyses the data.'
            write(6,'(1x,2a)') 'Usage   : trafodb ',&
       '[-i -o -op -DB -g -eci R -ver -h -?] Task ' 
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(a)') 'Tasks: '
            write(6,'(9x,a)') 'nact [f1 f2 [s1 s2]] '
            write(6,'(6x,a,3(/,6x,a))') &
              '         Convert NAC from the quantum chemistry data to ',&
              '         normal modes. If f1 and f2 given, NAC is plotted ',&
              '         in the plane of these DOF. State pair can be ',&
              '         selected with s1 s2.                      '
            write(6,'(9x,a)') 'coo  '
            write(6,'(6x,a)') &
              '         Convert coordinates from Cartesian to normal modes'
            write(6,'(9x,a)') 'rotang f1 f2 '
            write(6,'(6x,a,1(/,6x,a))') &
              '         Plot the ADT rotation angle in the plane of f1, f2.',&
              '         Only works for a 2-state problem.'
            write(6,'(9x,a)') 'dcp [f1 f2 [s1 s2]]  '
            write(6,'(6x,a,6(/,6x,a))') &
              '         Transform diabatic gradients to adiabatic picture ',&
              '         and from Cartesian to normal modes. For single state ',&
              '         or if s1 = s2, only transformed forces for that state ',&
              '         output. ',&
              '         If s2 .ne. s1, Plot NAC from the diabatic surfaces data. ',&
              '         If f1 and f2 given, NAC is plotted in the plane ',&
              '         of these DOF. Al .'
            write(6,'(9x,a)') 'correct_nact [f1 f2]'
            write(6,'(6x,a,2(/,6x,a))') &
              '         Corrects the phases of the NAC by locating CoIn. ',&
              '         and making the rotation consistent. If f1,f2 ',&
              '         given the new NAS is plotted. '
            write(6,'(9x,a)') 'adv_error [f1 f2]'
            write(6,'(6x,a,2(/,6x,a))') &
               '         Calculates the error of the adiabatic surfaces ',&
               '         from the raw data and the diabatic Hamiltonian.',&
               '         If f1,f2 given error is plotted in that plane.'
            write(6,'(9x,a)') 'freq  '
            write(6,'(6x,a)') &
               '          Convert the Hessian to normal modes.'
            write(6,'(9x,a)') 'curl f1 f2  '
            write(6,'(6x,a,1(/,6x,a))') &
               '         Calculates the curl error and plots it it the ',&
               '         plane of DOF f1, f2. Only works for 2 state problems.' 
            write(6,'(9x,a)') 'nearest  '
            write(6,'(6x,a,2(/,6x,a))') &
               '         Finds the nearest point in the DB to the structure ',&
               '         in structure.xyz (filename can be changed with -f option).'
            write(6,*)
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(/,a)') ' Options : '
            write(6,'(a)') ' -i DIR  : Working directory is DIR'
            write(6,'(a)') &
            ' -op FILE: The operator file FILE is &
            &read rather than oper'
            write(6,'(a,/,12x,a)') &
            ' -DB PATH : DB is in directory with relative path PATH &
            &to the working directory.','Default is ".."'
            write(6,'(a)') &
            ' -f FILE : Input read from FILE rather than default '  
            write(6,'(2a)') &
            ' -g      : GNUplot command lines are written to ',&
            'the output file(s).'
            write(6,'(a)') &
            ' -step   : Step (pause) between each vector '  
            write(6,'(a)')&
            ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
            ' -h      : Print this help text. ',&
            ' -?      : Print this help text. '
            write(6,'(a)') ' '

            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            write(6,'(9x,a)') 'curl f1 f2 : '
            write(6,'(6x,a,2(/,6x,a))') &
               '         Calculates the curl error and plots it it the ',&
               '         plane of DOF f1, f2. Only works for 2 state problems.' 
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(/,a)') ' Options : '
            write(6,'(a)') ' -i DIR  : Working directory is DIR'
            write(6,'(a)') &
            ' -op FILE: The operator file FILE is &
            &read rather than oper'
            write(6,'(a,/,12x,a)') &
            ' -DB PATH : DB is in directory with relative path PATH &
            &to the working directory.','Default is ".."'
            write(6,'(2a)') &
            ' -g      : GNUplot command lines are written to ',&
            'the output file(s).'
            write(6,'(a)') &
            ' -step   : Step (pause) between each vector '  
            write(6,'(a,2(/,a))')&
            ' -eci R  : States with an energy gap less than R are considered to ',&
            '           be degenerate (i.e. on the CoIn seam). ',&
            '           Units can be given e.g. "R , ev" (note spaces around ,)'
            write(6,'(a)')&
            ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
            ' -h      : Print this help text. ',&
            ' -?      : Print this help text. '
            write(6,'(a)') ' '

            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      end if
      
 200  continue

!-----------------------------------------------------------------------
!  First argument is task
!-----------------------------------------------------------------------
      if (i .le. iarg) then
         call mygetarg(i,buf)
      else
         write(6,'(a)') 'Enter task: '
         read(5,'(a)') buf
      endif

      if (buf .eq. 'nact') then
         itask = 1

      else if (buf .eq. 'coo') then
         itask = 2

      else if (buf .eq. 'rotang') then
         itask = 3

      else if (buf .eq. 'dcp') then
         itask = 4

      else if (buf .eq. 'correct_nact') then
         itask = 5

      else if (buf .eq. 'adv_error') then
         itask = 6

      else if (buf .eq. 'freq') then
         itask = 7

      else if (buf .eq. 'curl') then
         itask = 8

      else if (buf .eq. 'nearest') then
         itask = 9

      else if (buf .eq. 'stop') then
         stop
      else
         write(6,'(a)') 'Unknown task. Try again. Allowed options are:.'
         write(6,'(a)') &
             'nact, coo, rotang, dcp, correct_nact, adv_error, freq, &
             & curl, stop'
         write(6,'(a)') 'NOT:'
         write(6,'(a)') buf
         stop
      endif

!-----------------------------------------------------------------------
!  Get arguments
!-----------------------------------------------------------------------
      j = 0
      do
         i=i+1
         j=j+1
         if (i .gt. iarg) exit
         call mygetarg(i,buf)
         read(buf,*,iostat=ierr) iargs(j)
         if (ierr .ne. 0) then
            write(6,'(a/,1x,a72)') &
           ' The argument must be int number. Not :',buf
            stop
         end if
      enddo
      nargs=j-1

!-----------------------------------------------------------------------
! Set up absolute path for name directory.
! If operfile has not been set, look in name directory.
!-----------------------------------------------------------------------
      call abspath(name,laenge)

      if (operfile .eq. ' ') operfile=name(1:laenge)//'/oper'
      opflaenge=laenge+5

      return 
      
      end subroutine trafoinp

!#######################################################################
!
! Transform NACT from Cartesian to Nmodes
! adiabatic energies from the model Hamiltonian
!
!#######################################################################

      subroutine nact2nm(dbin,nmodes,ncoo,recordmax,nmcoo,m1,m2,&
                 s1,s2)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes,ncoo
      integer(long), intent(in) :: m1,m2,s1,s2
      integer(long)             :: i,j,k,n,nidx,nvec,nplots,natoms
      integer(long)             :: recordref,dercpdim
      real(dop)                                 :: xmin,xmax,ymin,ymax
      real(dop), dimension(ncoo,nmodes), intent(in) :: nmcoo
      real(dop)                                :: diff,mindiff,mine
      type(SQLITE_DATABASE), intent(in)        :: dbin

      real(dop),dimension(nmodes)              :: vnm,refgeo
      real(dop),dimension(nmodes,recordmax)    :: nactq
      real(dop), parameter                     :: pi = 4.0d0*atan(1.0d0)

      logical, dimension(recordmax)            :: lignore

      if (s1 .gt. nddstate) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected state above highest state :',s1,nddstate
         call errormsg
      endif
      if (s2 .gt. nddstate) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected state above highest state :',s2,nddstate
         call errormsg
      endif
      if (m1 .gt. ndof) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected modes not in system:',m1
         call errormsg
      endif
      if (m2 .gt. ndof) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected modes not in system:',m2
         call errormsg
      endif

!-----------------------------------------------------------------------
! open output file
!-----------------------------------------------------------------------
      fileout=name(1:laenge)//'/nactnm.dba'
      open(4,file=fileout,form='formatted',&
           status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The outputfile is : ',fileout(1:ilbl)
      
      lignore(:) = .false.

!-----------------------------------------------------------------------
! read in data
!-----------------------------------------------------------------------
      natoms=ncoo/3
      dercpdim = nddstate*(nddstate-1)/2
      call dbrdnact( dbin, natoms, dercpdim, 1,recordmax, dbdercp)
      call dbrdapes( dbin,nddstate, 1,recordmax, dbadener)
      call dbrdnm( dbin, nmodes, 1,recordmax, dbnm)

! get nact index matching requested states
      nidx = 0
      do i=1,s1
         do j=i+1,s2
            nidx = nidx + 1
         end do
      end do

! get NACT for state pair nidx
      do i = 1,recordmax
         write(4,'(a,i5)') '# Record no. : ',i
         vnm = matmul(dbdercp(1:ncoo,nidx,i),nmcoo)
         write(4,'(f15.10,f15.10,f15.10,f15.10)') vnm
         nactq(1:nmodes,i) = vnm
      end do

! find geometry with lowest energy gap between state1 and state2

      mindiff = 9999.0d0

      call dbrdapes( dbin, nddstate, 1,recordmax, dbadener)
      do i=1,recordmax
         diff = dabs(dbadener(s2,s2,i) - dbadener(s1,s1,i))
         if (diff .lt. mindiff) then
            recordref=i
            mindiff = min(mindiff,diff)
         endif
      enddo

      write(6,'(/,a)') '******'
      write(6,*) 'Looking for CoIn between states : ',s1,s2
      write(6,*) 'Record of lowest E gap is: ', recordref
      write(6,*) 'Energy of states is: ', &
                    dbadener(s1,s1,recordref), dbadener(s2,s2,recordref)
      write(6,'(a,/)') '******'

! read in nmode of reference (lowest E) geometry
      i=recordref
      refgeo(1:nmodes) = dbnm(1:nmodes,i)
      write(6,*) 'Geometry of lowest E gap in normal modes :'
      do n=1,nmodes,10
         j=min(nmodes,n+9)
         write(6,'(10f8.4)') (refgeo(n:j))
      enddo

      if (mindiff .lt. ecoin) then
         mine = dbadener(s1,s1,recordref)
         do i=1,recordmax
            diff = dabs(dbadener(s2,s2,i) - dbadener(s1,s1,i))
            if (diff .lt. ecoin .and. dbadener(s1,s1,i) .lt. mine) then
               recordref=i
               mine = min(mine,dbadener(s1,s1,i))
            endif
         enddo
         write(6,*) 'Record of lowest E on seam is: ', recordref
         write(6,'(a,2f10.6)') 'Energy of states is: ', &
                        dbadener(s1,s1,recordref), dbadener(s2,s2,recordref)
         write(6,'(a,f10.6)') 'Definition of seam is a max gap (eV): ', ecoin*uev
         write(6,'(a,/)') '******'
      endif

! Extract and plot DCP vector field in a plane
      if (m1 .gt. 0 .and. m2 .gt. 0) then
         rewind(4)

         write(fileout,'(2a,4(i5,a))') name(1:laenge),&
                    '/vfdata_',m1,'_',m2,'_',s1,'_',s2,'.pl'
         call rmblnk(fileout)
         open(3,file=fileout,form='formatted',&
              status='unknown',err=98)
         ilbl = index(fileout,' ')-1
         write(6,'(/,2a)') ' The plot file is : ',fileout(1:ilbl)

! igno outside the plane
         lignore(:)=.false.
         do i = 1,recordmax
            do n=1,nmodes
               if (n .eq. m1) cycle
               if (n .eq. m2) cycle

           !    if (abs(q(n) - 0.0) .gt. 0.1) lignore(i) = .true.
            enddo
         enddo

! get aries
         if (lgnu) then
            xmin=999.0
            xmax=-999.0
            ymin=999.0
            ymax=-999.0
            do i = 1,recordmax
               xmin = min(xmin,dbnm(m1,i))
               xmax = max(xmax,dbnm(m1,i))
               ymin = min(ymin,dbnm(m2,i))
               ymax = max(ymax,dbnm(m2,i))
            enddo
            xmin = nint(xmin)-1.0
            xmax = nint(xmax)+1.0
            ymin = nint(ymin)-1.0
            ymax = nint(ymax)+1.0
         endif

! get ts - if stepping then a new plot for each vector
         if (lstep) then
            nplots = recordmax
         else
            nplots = 1
         endif
            
         nvec=0
         do k=1,nplots
            if (lignore(k)) cycle
! get tors to plot
            if (lstep) then
               nvec = nvec+1
            else
               nvec = recordmax
            endif

            if (lgnu) then
               write(3,'(a)') 'set title "DCP vector field"'
               write(3,'(a)') 'set term x11'
               write(3,'(a)') 'set nokey'
               ilbl = index(modelabel(m1),' ')-1
               moderep = modelabel(m1)
               call mode_lab_corr(modelabel(m1),moderep)
               write(3,'(3a)') 'set xlabel "',moderep(1:ilbl),'"'
               ilbl = index(modelabel(m2),' ')-1
               moderep = modelabel(m2)
               call mode_lab_corr(modelabel(m2),moderep)
               write(3,'(3a)') 'set ylabel "',moderep(1:ilbl),'"'
               write(3,'(2(a,f9.4),a)') 'set xrange [',xmin,' : ',xmax,' ]'
               write(3,'(2(a,f9.4),a)') 'set yrange [',ymin,' : ',ymax,' ]'
! add h vectors as a point
               rewind(1)
               rewind(2)
               do i = 1,nvec
                  if (lignore(i)) cycle

                  if (abs(nactq(m1,i)) .lt. 0.00000001 .and. &
                         abs(nactq(m2,i)) .lt. 0.00000001) then
                     write(3,'(2(a,f8.4),a)')&
                        'set label at ',dbnm(m1,i),',',dbnm(m2,i),&
                        ' " " point '
                  end if
               end do
      
               write(3,'(a)') 'factor = 45'
               if (lstep) then
                  write(3,'(a)') 'plot "-" u 1:2:($3*factor):($4*factor) &
                     &with vectors head filled; pause -1'
               else
                  write(3,'(a)') 'plot "-" u 1:2:($3*factor):($4*factor) &
                     &with vectors head filled'
               endif
            endif
! plot
            rewind(1)
            rewind(2)
            do i = 1,nvec
               if (lignore(i)) cycle

               write(3,'(f15.10,f15.10,f15.10,f15.10)')&
                             dbnm(m1,i),dbnm(m2,i),&
                             nactq(m1,i),nactq(m2,i)
            end do
            if (lstep) write(3,'(a)') 'e'
         enddo
         if (.not. lstep) then
            write(3,'(a)') 'e'
            write(3,'(a)') 'pause -1'
         endif

         close(3)
      endif
      close(4)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine nact2nm

!#######################################################################
!
! Transform Coordinates from Cartesian to Nmodes
!
!#######################################################################

      subroutine coo2nm(dbin,nmodes,ncoo,recordmax,coonm,xcoo0)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes,ncoo
      integer(long)             :: i,natoms
      real(dop), dimension(nmodes,ncoo), intent(in) :: coonm
      real(dop), dimension(ncoo), intent(in)        :: xcoo0
      real(dop), dimension(ncoo,1)                  :: vcoo
      real(dop), dimension(ncoo)                    :: xcoo

      real(dop),dimension(nmodes)                   :: vnm
      type(SQLITE_DATABASE), intent(in)             :: dbin

      natoms=ncoo/3
!-----------------------------------------------------------------------
! open output file
!-----------------------------------------------------------------------
      fileout=name(1:laenge)//'/geonm.dba'
      laeout = index(fileout,' ' )-1
      call abspath(fileout,laeout)
      open(2,file=fileout(1:laeout),form='formatted',&
           status='unknown',err=98)
      write(6,'(2a)') ' The outputfile is : ',fileout(1:laeout)

!-----------------------------------------------------------------------
! read in data and transform
!-----------------------------------------------------------------------
      do i = 1,recordmax

         call dbrdgeo( dbin, natoms, i, -1, vcoo)

         xcoo(:) = vcoo(:,1) - xcoo0(:)
         vnm = matmul(coonm,xcoo)

         write(2,'(a,i10)') '#Record No. : ',i
         write(2,'(f15.10,f15.10,f15.10,f15.10)') vnm

      end do

      close(2)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine coo2nm

!#######################################################################
!
!#######################################################################

      subroutine subrotang(dbin,nmodes,recordmax,m1,m2)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes
      integer(long), intent(in) :: m1,m2
      real(dop) :: xmin,xmax,ymin,ymax
      real(dop), dimension(nddstate,nddstate) :: adtmat

      real(dop),dimension(nmodes)  :: q
      real(dop)                    :: rotang,rotang1,rotang2
      real(dop), parameter         :: pi = 4.0d0*atan(1.0d0)
      type(SQLITE_DATABASE), intent(in)       :: dbin

!-----------------------------------------------------------------------
! check input
!-----------------------------------------------------------------------
      if (nddstate .ne. 2) then
         routine = 'Subrotang'
         message = 'Can only plot the ADT rotation angle for a 2-state problem '
         call errormsg
      endif

      if (m1 .le. 0 .or. m2 .le. 0) then
         routine = 'Subrotang'
         message = 'Specify 2 DOFs to plot ADT rotation angle '
         call errormsg
      endif

!-----------------------------------------------------------------------
! Open output file
!-----------------------------------------------------------------------
      write(fileout,'(2a,i5,a,i5,a)') name(1:laenge),'/rotang_',m1,'_',m2,'.pl'
      call rmblnk(fileout)
      open(3,file=fileout,form='formatted',&
           status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The plot file is : ',fileout(1:ilbl)

!-----------------------------------------------------------------------
! Read in data
!-----------------------------------------------------------------------
      call dbrdnm( dbin, nmodes, 1, recordmax, dbnm )
      call dbrdtrans( dbin,nddstate, 1,recordmax, dbtrans )

      if (lgnu) then
         write(3,'(a)') 'set title "NACT mixing angle"'
         write(3,'(a)') 'set term x11'
         write(3,'(a)') 'set nokey'
         ilbl = index(modelabel(m1),' ')-1
         write(3,'(3a)') 'set xlabel "',modelabel(m1)(1:ilbl),'"'
         ilbl = index(modelabel(m2),' ')-1
         write(3,'(3a)') 'set ylabel "',modelabel(m2)(1:ilbl),'"'

! put data points on surface and get axis boundaries
         xmin=999.0
         xmax=-999.0
         ymin=999.0
         ymax=-999.0
         do i = 1,recordmax
            q(1:nmodes) = dbnm(1:nmodes,i)
            adtmat(1:nddstate,1:nddstate) = dbtrans(1:nddstate,1:nddstate,i)

            rotang2 = adtmat(1,2)   ! sin(theta)
            rotang1 = adtmat(1,1)   ! cos(theta)
            if (rotang1 .ge. 0.0 .and. rotang2 .ge. 0.0) then
               rotang = acos(rotang1)
            else if (rotang1 .ge. 0.0 .and. rotang2 .lt. 0.0) then
               rotang = asin(rotang2)
            else if (rotang1 .lt. 0.0 .and. rotang2 .ge. 0.0) then
               rotang = acos(rotang1) 
            else if (rotang1 .lt. 0.0 .and. rotang2 .lt. 0.0) then
               rotang = 2.0*pi-acos(rotang1) 
            endif
            write(3,'(3(a,f8.4),a)')&
               'set label at ',q(m1),',',q(m2),',',rotang,' " " point '

            xmin = min(xmin,q(m1))
            xmax = max(xmax,q(m1))
            ymin = min(ymin,q(m2))
            ymax = max(ymax,q(m2))
         end do
         xmin = nint(xmin)-1.0
         xmax = nint(xmax)+1.0
         ymin = nint(ymin)-1.0
         ymax = nint(ymax)+1.0

         write(3,'(2(a,f9.4),a)') 'set xrange [',xmin,' : ',xmax,' ]'
         write(3,'(2(a,f9.4),a)') 'set yrange [',ymin,' : ',ymax,' ]'
         write(3,'(a)') 'set zrange [* : *]'

! set up surface
         write(3,'(a)') 'set surface'
         write(3,'(a)') 'set dgrid3d 20,20 splines'
         write(3,'(a)') 'set contour'
         write(3,'(a)') 'set clabel "%10.3e"'
         write(3,'(a)') 'set cntrparam cubicspline'
         write(3,'(a)') 'set cntrparam points 5'
         write(3,'(a)') 'set autoscale z'
         write(3,'(a)') 'set cntrparam levels incremental -1.5,0.1,1.5'

         write(3,'(a)') 'splot "-" u 1:2:3 w l'
      endif

!-----------------------------------------------------------------------
! write data
!-----------------------------------------------------------------------
      do i = 1,recordmax
         q(1:nmodes) = dbnm(1:nmodes,i)
         adtmat(1:nddstate,1:nddstate) = dbtrans(1:nddstate,1:nddstate,i)

         rotang2 = adtmat(1,2)   ! sin(theta)
         rotang1 = adtmat(1,1)   ! cos(theta)
         if (rotang1 .ge. 0.0 .and. rotang2 .ge. 0.0) then
            rotang = acos(rotang1)
         else if (rotang1 .ge. 0.0 .and. rotang2 .lt. 0.0) then
            rotang = asin(rotang2)
         else if (rotang1 .lt. 0.0 .and. rotang2 .ge. 0.0) then
            rotang = acos(rotang1) 
         else if (rotang1 .lt. 0.0 .and. rotang2 .lt. 0.0) then
            rotang = 2.0*pi-acos(rotang1) 
         endif

         write(3,'(f15.10,f15.10,f15.10,f15.10)')&
               q(m1),q(m2),rotang

      end do

      if (lgnu) then
         write(3,'(a)') 'e'
         write(3,'(a)') 'pause -1'
      endif

      close(3)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine subrotang

!#######################################################################
!
!
!#######################################################################

      subroutine out_dcp(dbin,nmodes,ncoo,recordmax,m1,m2,s1,s2,nmcoo)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes,ncoo
      integer(long), intent(in) :: m1,m2,s1,s2
      integer(long)             :: k,nplots,nvec

      real(dop) :: xmin,xmax,ymin,ymax

      real(dop), parameter  :: pi = 4.0d0*atan(1.0d0)
      real(dop), dimension(ncoo,nmodes) :: nmcoo !forward
      real(dop), dimension(ncoo,nddstate,nddstate) :: dbgra !vector cartes
      real(dop), dimension(nmodes) :: granm11, granm21, granm22  !vector modes
      real(dop), dimension(nmodes) :: q,coupl1,coupl2,force1,force2
      real(dop), dimension(nddstate,nddstate) :: trans, ttrans, nablaW, coup
      character(len=30)            :: title
      type(SQLITE_DATABASE), intent(in)        :: dbin

!-----------------------------------------------------------------------
! check input
!-----------------------------------------------------------------------
      if (s1 .gt. nddstate) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected state above highest state :',s1,nddstate
         call errormsg
      endif
      if (s2 .gt. nddstate) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected state above highest state :',s2,nddstate
         call errormsg
      endif
      if (m1 .gt. ndof) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected modes not in system:',m1
         call errormsg
      endif
      if (m2 .gt. ndof) then
         routine = 'Nact2nm'
         write(message,'(a,2i5)') 'Selected modes not in system:',m2
         call errormsg
      endif

!-----------------------------------------------------------------------
! open output file
!-----------------------------------------------------------------------
      fileout=name(1:laenge)//'/granm.dba'
      open(3,file=fileout,form='formatted',status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The outputfile is : ',fileout(1:ilbl)
! read gradients
      call dbrdgra(dbin,nddstate,natm,1,recordmax, dbgrad)

      do i = 1,recordmax
         dbgra(1:ncoo,1:nddstate,1:nddstate) = &
                           dbgrad(1:ncoo,1:nddstate,1:nddstate,i) 
         
         write(3,'(a,i10)') '#Record : ',i
! transform to nmodes
         granm11 = matmul(dbgra(:,s1,s1),nmcoo)
         write(3,'(f15.10,f15.10,f15.10,f15.10)') granm11

         if (s2 .ne. s1) then
            granm21 = matmul(dbgra(:,s2,s1),nmcoo)
            granm22 = matmul(dbgra(:,s2,s2),nmcoo)

            write(3,'(f15.10,f15.10,f15.10,f15.10)') granm21
            write(3,'(f15.10,f15.10,f15.10,f15.10)') granm22
         endif

      end do

! Cannot plot NACT for oly 1 state
      if (s2 .eq. s1) return

      fileout=name(1:laenge)//'/coup.dba'
      open(4,file=fileout,form='formatted',status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The outputfile is : ',fileout(1:ilbl)

      rewind(3)
      write(4,'(a)') '#Each row is the 2x2 gradient along a nmode component'
! read in ADT
      call dbrdtrans(dbin, nddstate, 1,recordmax, dbtrans)

      do i = 1,recordmax
         trans(1:nddstate,1:nddstate) = dbtrans(1:nddstate,1:nddstate,i)
         ttrans = transpose(trans)

! read in diabatic gradients in nmode coordinates
         read(3,'(A)') title
         read(3,*) granm11
         read(3,*) granm21
         read(3,*) granm22

         write(4,'(a,i10)') '#Record : ',i

         do j=1,nmodes

            nablaW(1,1)=granm11(j)
            nablaW(1,2)=granm21(j)
            nablaW(2,1)=granm21(j)
            nablaW(2,2)=granm22(j)

! transform to adiabatic picture (gradients and DCP vectors)
            coup = matmul(trans,nablaW)
            coup = matmul(coup,ttrans)

            write(4,'(f15.10,f15.10, f15.10,f15.10)') coup

         end do
      end do

      call dbrdnm( dbin, nmodes, 1,recordmax, dbnm )

!-----------------------------------------------------------------------
! Extract and plot DCP vector field in a plane
!-----------------------------------------------------------------------
      if (m1 .gt. 0 .and. m2 .gt. 0) then
         write(fileout,'(2a,i5,a,i5,a)') name(1:laenge),'/coupdata_',m1,'_',m2,'.pl'
         call rmblnk(fileout)
         open(3,file=fileout,form='formatted',status='unknown',err=98)
         ilbl = index(fileout,' ')-1
         write(6,'(2a)') ' The plot file is : ',fileout(1:ilbl)

! get axis boundaries
         if (lgnu) then
            xmin=999.0
            xmax=-999.0
            ymin=999.0
            ymax=-999.0
            do i = 1,recordmax
! get nmode
               q(1:nmodes) = dbnm(1:nmodes,i)
               xmin = min(xmin,q(m1))
               xmax = max(xmax,q(m1))
               ymin = min(ymin,q(m2))
               ymax = max(ymax,q(m2))
            enddo
            rewind(1)
            xmin = nint(xmin)-1.0
            xmax = nint(xmax)+1.0
            ymin = nint(ymin)-1.0
            ymax = nint(ymax)+1.0
         endif

! get no. of plots - if stepping then a new plot for each vector
         if (lstep) then
            nplots = recordmax
         else
            nplots = 1
         endif
            
         do k=1,nplots
! get no. of vectors to plot
            if (lstep) then
               nvec = k
            else
               nvec = recordmax
            endif

            if (lgnu) then
               write(3,'(a)') 'set title "DCP vector field (gra)"'
               write(3,'(a)') 'set term x11'
               write(3,'(a)') 'set nokey'
               ilbl = index(modelabel(m1),' ')-1
               write(3,'(3a)') 'set xlabel "',modelabel(m1)(1:ilbl),'"'
               ilbl = index(modelabel(m2),' ')-1
               write(3,'(3a)') 'set ylabel "',modelabel(m2)(1:ilbl),'"'
               write(3,'(2(a,f9.4),a)') 'set xrange [',xmin,' : ',xmax,' ]'
               write(3,'(2(a,f9.4),a)') 'set yrange [',ymin,' : ',ymax,' ]'
! add zero length vectors as a point
               rewind(4)
               read(4,*) ! title line
               do i = 1,nvec
                  q(1:nmodes) = dbnm(1:nmodes,i)
                  read(4,'(A)') title
                  do j=1,nmodes
                     read(4,*) force1(j),coupl1(j),coupl2(j),force2(j)
                  end do
                  if (abs(coupl1(m1)) .lt. 0.00000001 .and. &
                      abs(coupl1(m2)) .lt. 0.00000001) then
                     write(3,'(2(a,f8.4),a)')&
                                'set label at ',q(m1),',',q(m2),' " " point '
                  end if
               end do

               write(3,'(a)') 'factor = 25'
               if (lstep) then
                  write(3,'(a)') 'plot "-" u 1:2:($3*factor):($4*factor) with &
                        &vectors head filled; pause -1'
               else
                   write(3,'(a)') 'plot "-" u 1:2:($3*factor):($4*factor) with &
                        &vectors head filled'
               endif
            endif
! plot vectors
            rewind(4)
            read(4,*) ! title line
            do i = 1,nvec
               q(1:nmodes) = dbnm(1:nmodes,i)

               read(4,'(A)') title
               do j=1,nmodes
                  read(4,*) force1(j),coupl1(j),coupl2(j),force2(j)
               end do

               write(3,'(f15.10,f15.10,f15.10,f15.10)')&
                     q(m1),q(m2),coupl1(m1),coupl1(m2)

            end do
            if (lstep) write(3,'(a)') 'e'
         enddo
         if (.not. lstep) then
            write(3,'(a)') 'e'
            write(3,'(a)') 'pause -1'
         endif

         close(3)
      endif
      close(4)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine out_dcp

!#######################################################################
!
!
!#######################################################################

      subroutine nact_correct(dbin,nmodes,ncoo,nonact,recordmax,m1,m2,nmcoo)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes,ncoo,nonact
      integer(long), intent(in) :: m1,m2
      integer, dimension(recordmax) :: ncoin
      integer                   :: n,nc,s,s1,s2,numvec
      integer                   :: natoms

      real(dop), parameter  :: pi = 4.0d0*atan(1.0d0)
      real(dop), dimension(ncoo,nmodes), intent(in) :: nmcoo !forward
      real(dop), dimension(ncoo) :: vcoo   !vector cartes
      real(dop), dimension(nmodes) :: vnm  !vector modes
      real(dop), dimension(nmodes) :: q,nactq
      real(dop), dimension(nddstate,recordmax):: adv
      real(dop)                               :: diff,mindiff
      real(dop), dimension(recordmax) :: x,y,dx,dy
      real(dop), dimension(recordmax) :: dxnew,dynew
      real(dop), dimension(100) :: px,py
      real(dop) :: px2,py2

      real(dop) :: sp
      real :: range_new = 2

      character(len=30)            :: title
      type(SQLITE_DATABASE), intent(in)       :: dbin
             
      fileout=name(1:laenge)//'/nactnm.dba'
      open(4,file=fileout,form='formatted',status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The outputfile is : ',fileout(1:ilbl)

      natoms=ncoo/3
      call dbrdnact( dbin, natoms, dercpdim, 1, recordmax, dbdercp)
      do i = 1,recordmax

         write(4,'(a,i10)') '#Record No. : ',i
         do j = 1,nonact
              vcoo(1:ncoo) = dbdercp(1:ncoo,j,i)

              vnm = matmul(vcoo,nmcoo)

              write(4,'(f15.10,f15.10,f15.10,f15.10)') vnm
         end do
      end do

      if (m1 .gt. 0 .and. m2 .gt. 0) then

         if (nonact .gt. 1) then
            routine = 'NACT_Correct'
            message = 'Can only correct NACT for a 2-state problem'
            call errormsg
         endif

         rewind(1)
         call dbrdapes( dbin,nddstate,1,recordmax,dbadener)
         call dbrdnm( dbin,nmodes,1,recordmax,dbnm)

         rewind(4)
         do i = 1,recordmax

            q(1:nmodes) = dbnm(1:nmodes,i)

            do s=1,nddstate
               adv(s,i) = dbadener(s,s,i)
            enddo

            read(4,'(A)') title
            read(4,*) nactq

            x(i)=q(m1)
            y(i)=q(m2)
            dx(i)=nactq(m1)
            dy(i)=nactq(m2)

!build perpendicular vector
            dxnew(i)=x(i)-dy(i)
            dynew(i)=y(i)+dx(i)

         end do

! ncoin flags the coin assigned to a point. nc is present CoIn.
         ncoin(:) = 0
         nc = 1

print*,'*** Starting Calculation ***'

! Ignore any points with zero NACT
         do i=1,recordmax
            diff = dx(i)*dx(i) + dy(i)*dy(i)
            diff = dsqrt(diff)

            if (diff .lt. 1.0d-6) ncoin(i) = -1
         enddo

  500    continue
! get lowest energy gap as estimate of CoIn

         mindiff = 9999.0d0
         s2=2
         s1=1
         do i=1,recordmax
! if point already assigned to a CoIn, ignore
            if (ncoin(i) .ne. 0) cycle
            diff = dabs(adv(s2,i) - adv(s1,i))
            if (diff .lt. mindiff) then
               j=i
               mindiff = min(mindiff,diff)
            endif
         enddo
         px(nc) = x(j)
         py(nc) = y(j)

print*
print*,'Search for CoIn no: ',nc
print*,'Pass 1. Search around initial guess ',px(nc),py(nc)
         do

            numvec = 0   !counts number of calculated points

            px2 = 0.0d0
            py2 = 0.0d0
            do i=1,recordmax-1
               if (ncoin(i) .ne. 0) cycle
               do j=i+1,recordmax

         !determine intersection of each two vectors pairwise
                  if (abs(x(i)-px(nc)) < range_new &
                         .and. abs(y(i)-py(nc))< range_new) then

                     if (abs(x(j)-px(nc)) < range_new &
                            .and. abs(y(j)-py(nc)) < range_new) then

                       call INTERSECTION(x(i),dxnew(i),x(j),dxnew(j),&
                            y(i),dynew(i),y(j),dynew(j),px2,py2,numvec)

                     end if
                  end if

               end do
            end do

            if (numvec .gt. 1) then

               px2 = px2/numvec
               py2 = py2/numvec

               if ( abs(px(nc)-px2)>5e-1 .and. abs(py(nc)-py2)>5e-1) then
                  print*,'Warning: CoIn found is >0.5 away from guess'
               endif

               exit

            else

               print*,'WARNING: No points found in range '
               range_new = range_new+0.5
            endif
         enddo

         px(nc) = px2
         py(nc) = py2
         print*,'New estimate of CoIn: ',px(nc),py(nc)
		
print*,'Pass 2. Assigning Vectors to CoIn: ',nc
         numvec = 0   !counts number of assigned points
         do i=1,recordmax-1
            if (ncoin(i) .ne. 0) cycle
! add on initial vector
            do j=i+1,recordmax

!determine intersection of each two vectors pairwise

               px2 = 0.0d0
               py2 = 0.0d0
               call INTERSECTION(x(i),dxnew(i),x(j),dxnew(j),&
                       y(i),dynew(i),y(j),dynew(j),px2,py2,n)
       
               if ( abs(px(nc)-px2)<5e-1 .and. abs(py(nc)-py2)<5e-1) then
                  if (ncoin(i) .eq. 0) then
                     ncoin(i) = nc                  
                     numvec = numvec +1
                  else if (ncoin(i) .ne. nc) then
                     print*,'Reassigning point ',i
                  endif
                  if (ncoin(j) .eq. 0) then
                     ncoin(j) = nc                  
                     numvec = numvec +1
                  else if (ncoin(j) .ne. nc) then
                     print*,'Reassigning point ',j
                  endif
               endif

            end do
         end do
print*,'ASSIGNED:',numvec
         if (count(ncoin .eq. -1) .gt. 0) then
            print*,'Zero length:',count(ncoin .eq. -1)
         endif

! If points still not assigned to a CoIn, repeat
         n = count(ncoin .eq.0)
print*,'to be assigned:',n
! if none were assigned in last pass, assign to last CoIn
         if (numvec .eq. 0) then
            do i=1,recordmax
               if (ncoin(i) .eq. 0) then
                  ncoin(i) = nc
               endif
            enddo
! if only 1 point left, assign to last CoIn
         else if (n .eq. 1) then
            nc = nc+1
            do i=1,recordmax
               if (ncoin(i) .eq. 0) then
                  ncoin(i) = nc
                  px(nc) = x(i)
                  py(nc) = y(i)
               endif
            enddo 
! next CoIn
         else if (n .gt. 1) then
            nc = nc+1
            go to 500
         endif

         print*
         print*,'Number of CoIn found: ',nc
         do n=1,nc
            write(6,'(a,i3,2(a,f8.4),a)') &
                'The estimated center of CoIn ',n,  &
                'is at: (', px(n),',', py(n),')'               
         enddo

!create new file
         fileout=name(1:laenge)//'/nact_corrected.dba'
         open(3,file=fileout,form='formatted',status='unknown',err=98)
         ilbl=index(fileout,' ')-1
         write(6,'(2a)') ' The outputfile is : ',fileout(1:ilbl)

         do i = 1,recordmax

            vcoo(1:ncoo) = dbdercp(1:ncoo,1,i)

            write(3,'(a,i10)') '#Record No. : ',i

            n = ncoin(i)
            if (n .gt. 0) then
               sp = (x(i)-px(n))*dy(i)-(y(i)-py(n))*dx(i)
            else
               sp = 1.0
            endif

            if (sp < 0) then
               vcoo = -vcoo
               write(3,'(f20.12,f20.12,f20.12)') vcoo
            else 
               write(3,'(f20.12,f20.12,f20.12)') vcoo
            end if

         end do

         close(3)
      end if

      close(4)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine nact_correct

!#######################################################################

      subroutine INTERSECTION(x_i,dxnew_i,x_j,dxnew_j,&
                              y_i,dynew_i,y_j,dynew_j,px,py,n)

      implicit none
      integer, parameter :: dim_num = 2

      logical :: point_1,point_2
      real(dop) :: p(dim_num)
      real(dop) :: p1(dim_num),p2(dim_num)
      real(dop) :: q1(dim_num),q2(dim_num)
      integer(long) :: n
      real(dop) :: x_i,y_i,x_j,y_j
      real(dop) :: dxnew_i,dynew_i,dxnew_j,dynew_j
      real(dop) :: px,py

      real(dop) :: a1,a2,b1,b2,c1,c2,denom

      p1(1)=x_i
      p1(2)=y_i
      p2(1)=dxnew_i
      p2(2)=dynew_i

      q1(1)=x_j
      q1(2)=y_j
      q2(1)=dxnew_j
      q2(2)=dynew_j

! check wether line is point
      if ( all ( p1(1:dim_num) == p2(1:dim_num) ) ) then
         point_1 = .false.
      else
         point_1 = .true.
      end if
      if ( all ( q1(1:dim_num) == q2(1:dim_num) ) ) then
         point_2 = .false.
      else
         point_2 = .true.
      end if

! determine intersection=center of vector field
      p(:) = 0.0
      if ( point_1 .and. point_2) then

         a1=x_i*dynew_i-y_i*dxnew_i
         a2=x_j*dynew_j-y_j*dxnew_j
         b1=x_i-dxnew_i
         b2=x_j-dxnew_j
         c1=y_i-dynew_i
         c2=y_j-dynew_j
         denom=b1*c2-c1*b2

         if (.not. denom == 0) then
            p(1)=(a1*b2-b1*a2)/denom
            p(2)=(a1*c2-c1*a2)/denom
            n = n + 1
         end if

         px = px + p(1)
         py = py + p(2)

      end if          !point_1 and point_2

      return
      end subroutine INTERSECTION

!#######################################################################
!
! Calculate the difference between the raw ab initio data and the
! adiabatic energies from the model Hamiltonian
!
!#######################################################################

      subroutine adverr(dbin,nmodes,recordmax,m1,m2)

      implicit none

      integer(long), intent(in) :: nmodes,recordmax,m1,m2
      integer(long) :: i,s,s1,info,n

      real(dop), dimension(nmodes)            :: q
      real(dop), dimension(nddstate,nddstate) :: adtmat
      real(dop), dimension(nddstate,recordmax):: adv
      real(dop), dimension(nddstate,nddstate) :: v,tmpav,tmpav1
      real(dop), dimension(3*nddstate)        :: work
      real(dop)                               :: maxerr,averr,err,&
                                                 maxerr1,averr1,err1
      type(SQLITE_DATABASE), intent(in)       :: dbin
      
!-----------------------------------------------------------------------
! read in data
!-----------------------------------------------------------------------
      call dbrdapes( dbin,nddstate, 1, recordmax, dbadener)
      call dbrdpes( dbin,nddstate, 1, recordmax, dbener)
      call dbrdnm( dbin, nmodes, 1, recordmax, dbnm )
      call dbrdtrans( dbin,nddstate, 1, recordmax, dbtrans)

      write(fileout,'(2a,i5,a,i5,a)') name(1:laenge),'/adverr.dat'
      call rmblnk(fileout)
      open(7,file=fileout,form='formatted',&
           status='unknown',err=98)
      ilbl = index(fileout,' ')-1
      write(6,'(2a)') ' The error file is : ',fileout(1:ilbl)

      write(7,'(a)') '#                     Error (eV) '
      write(7,'(a)') '# Dataset   |raw - S^+WS|   |S^+WS - dsyev|'
      maxerr = 0.0d0
      maxerr1 = 0.0d0
      averr = 0.0d0
      averr1 = 0.0d0
      do i = 1,recordmax
         q(1:nmodes) = dbnm(1:nmodes,i)
         do s=1,nddstate
            adv(s,i) = dbadener(s,s,i)
            do s1=1,nddstate
               adtmat(s1,s) = dbtrans(s1,s,i)
               v(s1,s) = dbener(s1,s,i)
            enddo
         enddo

         call qqxtdd(v,adtmat,tmpav1,nddstate)
         tmpav = matmul(adtmat,tmpav1)
         do s=1,nddstate
            tmpav1(s,1) = tmpav(s,s)
         enddo
         call dbubble(tmpav1,tmpav,nddstate,0)

         tmpav = v
         call dsyev('V','U',nddstate,tmpav,nddstate,tmpav1(1,2),work,&
              3*nddstate,info)

         err = 0.0d0
         err1 = 0.0d0
         do s=1,nddstate
            err = err+ dabs(tmpav1(s,1) - adv(s,i))
            err1 = err1+ dabs(tmpav1(s,2) - tmpav1(s,1))
         enddo

         maxerr = max(maxerr,err)
         averr = averr+err
         maxerr1 = max(maxerr1,err1)
         averr1 = averr1+err1

         write(7,'(i7,3x,2f16.12)') i,err,err1

      end do

      print*,'Maximum Error (eV) = ',maxerr
      print*,'Average Error (eV) =', averr/dble(recordmax)

      if (m1 .gt. 0 .and. m2 .gt. 0) then

         write(fileout,'(2a,i5,a,i5,a)') &
                name(1:laenge),'/adverr_',m1,'_',m2,'.pl'
         call rmblnk(fileout)
         open(8,file=fileout,form='formatted',status='unknown',err=98)
         ilbl = index(fileout,' ')-1
         write(6,'(2a)') ' The plot file is : ',fileout(1:ilbl)

         if (lgnu) then
            write(8,'(a)') 'set title "Adiab Error"'
            write(8,'(a)') 'set term x11'
            write(8,'(a)') 'set nokey'
            ilbl = index(modelabel(m1),' ')-1
            write(8,'(3a)') 'set xlabel "',modelabel(m1)(1:ilbl),'"'
            ilbl = index(modelabel(m2),' ')-1
            write(8,'(3a)') 'set ylabel "',modelabel(m2)(1:ilbl),'"'
            write(8,'(3a)') 'set zlabel "eV" rotate offset -2,0'

! put data points on surface and get axis boundaries
            rewind(7)
            read(7,*)
            read(7,*)
            xmin=999.0
            xmax=-999.0
            ymin=999.0
            ymax=-999.0
            do i = 1,recordmax
               q(1:nmodes) = dbnm(1:nmodes,i)
               read(7,*) n,err

               write(8,'(3(a,f8.4),a)')&
                  'set label at ',q(m1),',',q(m2),',',err,' " " point '

               xmin = min(xmin,q(m1))
               xmax = max(xmax,q(m1))
               ymin = min(ymin,q(m2))
               ymax = max(ymax,q(m2))
            end do
            xmin = nint(xmin)-1.0
            xmax = nint(xmax)+1.0
            ymin = nint(ymin)-1.0
            ymax = nint(ymax)+1.0

            write(8,'(2(a,f9.4),a)') 'set xrange [',xmin,' : ',xmax,' ]'
            write(8,'(2(a,f9.4),a)') 'set yrange [',ymin,' : ',ymax,' ]'
            write(8,'(a)') 'set zrange [* : *]'

! set up surface
            write(8,'(a)') 'set surface'
            write(8,'(a)') 'set dgrid3d 20,20 splines'
            write(8,'(a)') 'set contour'
            write(8,'(a)') 'set clabel "%10.3e"'
            write(8,'(a)') 'set cntrparam cubicspline'
            write(8,'(a)') 'set cntrparam points 5'
            write(8,'(a)') 'set autoscale z'
            write(8,'(a)') 'set cntrparam levels incremental -1.5,0.1,1.5'

            write(8,'(a)') 'splot "-" u 1:2:3 w l'
         endif

         rewind(7)
         read(7,*)
         read(7,*)
         do i = 1,recordmax

            q(1:nmodes) = dbnm(1:nmodes,i)

            read(7,*) n,err

            write(8,'(f15.10,f15.10,f15.10,f15.10)')&
                  q(m1),q(m2),err

         end do
         write(8,'(a)') 'e'
         write(8,'(a)') 'pause -1'

         close(8)
      endif
      close(7)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
 
      end subroutine adverr

!#######################################################################
!
! Transform Hessian from Cartesian to Nmodes - i.e. get "frequencies"
!
!#######################################################################

      subroutine hess2freq(dbin,nmodes,ncoo,recordmax,nmcoo)

      implicit none

      integer(long), intent(in) :: recordmax,nmodes,ncoo
      integer(long)             :: i,j,k,s,s1
      real(dop), dimension(ncoo,nmodes), intent(in) :: nmcoo
      real(dop), dimension(nmodes,nmodes,nddstate,nddstate) :: freq
      real(dop), dimension(ncoo,nmodes)            :: tmp
      type(SQLITE_DATABASE), intent(in)            :: dbin

      open(1,file=ddname(1:ddlaenge)//'/hes.dba',form='formatted')

      fileout=name(1:laenge)//'/freqnm.dba'
      laeout = index(fileout,' ' )-1
      call abspath(fileout,laeout)
      open(2,file=fileout(1:laeout),form='formatted',status='unknown',err=98)
      write(6,'(2a)') ' The "frequency (in eV)" file is : ',fileout(1:laeout)

      fileout=name(1:laenge)//'/hesnm.dba'
      laeout = index(fileout,' ' )-1
      call abspath(fileout,laeout)
      open(3,file=fileout(1:laeout),form='formatted',status='unknown',err=98)
      write(6,'(2a)') ' The Hessian file is : ',fileout(1:laeout)
  
      call dbrdhes( dbin,nddstate,natm, 1, recordmax, dbhess )

      do i = 1,recordmax

         do s=1,nddstate
            do s1 = 1,nddstate
               call mmxxdd(dbhess(1:ncoo,1:ncoo,s1,s,i),nmcoo,tmp,ncoo,ncoo,nmodes)
               call mmtxdd(nmcoo,tmp,freq(1,1,s1,s),nmodes,ncoo,nmodes)
            enddo
         enddo

         write(2,'(a,i10)') '#Record : ',i
         do s=1,nddstate
            do s1=s,nddstate
                write(2,'(f15.10,f15.10,f15.10,f15.10)') &
                               (freq(j,j,s1,s)*uev,j=1,nmodes)
                write(2,*)
            enddo
         enddo

         write(3,'(a,i10)') '#Record : ',i
         do s=1,nddstate
            do s1=s,nddstate
                do k=1,nmodes
                   write(3,'(f15.10,f15.10,f15.10,f15.10)') &
                                (freq(j,k,s1,s),j=1,nmodes)
                   write(3,*)
                enddo
                write(3,*) "-----"
            enddo
         enddo

      end do

      close(2)
      close(3)

      return
 98   write(6,*)' Cannot open output file : '
      write(6,*) fileout
      stop
      end subroutine hess2freq

!#######################################################################
!
! Check curl condition of DCP around CoIn
!
!#######################################################################

      subroutine chkcurl(dbin,nmodes,ncoo,nmcoo,xcoo0,recordmax,m1,m2)

      implicit none

      integer(long), intent(in) :: m1,m2
      integer(long), intent(in) :: recordmax
      integer(long), intent(in) :: nmodes,ncoo
      real(dop), dimension(ncoo,nmodes), intent(in) :: nmcoo
      real(dop), dimension(ncoo), intent(in)        :: xcoo0
      real(dop), dimension(nmodes)                  :: vnm

      real(dop), dimension(nmodes)                 :: gpoint
      real(dop), dimension(ncoo)                   :: xgpoint
      real(dop), dimension(nddstate,nddstate)      :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate) &
                                                   :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate) &
                                                   :: deriv2
      real(dop), dimension(ndofddpes)              :: nact,nactm1
      real(dop), dimension(ndofddpes)              :: nactm2
      real(dop), dimension(nmodes)                 :: nactnm
      real(dop), dimension(nmodes)                 ::nactnmm1,nactnmm2
      real(dop), dimension(3,nddstate,nddstate)    :: dipole
      real(dop), dimension(recordmax)              :: curl
      integer(long)                                :: idbrd
      integer(long)                                :: i
      real(dop)                                    :: disp

      type(SQLITE_DATABASE), intent(in)            :: dbin

!-----------------------------------------------------------------------
! Open output file.
! Only for two states
! 2 DOF must be specified
!-----------------------------------------------------------------------
      if (nddstate .gt. 2) then
         print*,'ERROR: Curl calculation only works for 2 state problem'
         stop
      else if (m1 .gt. 0 .and. m2 .gt. 0) then
          write(fileout,'(2a,i5,a,i5,a)') name(1:laenge),'/curl_',m1,'_',m2,'.pl'
          call rmblnk(fileout)
          open(3,file=fileout,form='formatted',status='unknown')
          ilbl = index(fileout,' ')-1
          write(6,'(2a)') ' The plot file is : ',fileout(1:ilbl)
      else
         print*,'ERROR:  Specify two modes to use for the curl '
         stop
      endif
          
!-----------------------------------------------------------------------
! displacement for numerical derivatives
!-----------------------------------------------------------------------
      disp = 0.05

!-----------------------------------------------------------------------
! open DB and read data into memory
! DB to be set up for use by dddb_rd in preparedb must be the_db.
!-----------------------------------------------------------------------
      the_db = dbin
      idbrd = 3
      call preparedb(idbrd,1)

!-----------------------------------------------------------------------
! Calculate curl at each point
!-----------------------------------------------------------------------
      do i = 1,recordmax
         vnm(1:nmodes) = dbnm(1:nmodes,i)

         gpoint(:) = vnm(:)
         xgpoint = matmul(nmcoo,gpoint)
         xgpoint(:) = xgpoint(:) + xcoo0(:)

         call dddb_rd(v,deriv1,deriv2,dipole,xgpoint,.false.,.false.,.false.)

         nact=(deriv1(:,1,2))
         nactnm=matmul(nact,nmcoo)

         gpoint(:) = vnm(:)
         gpoint(m1)=gpoint(m1)+disp
         xgpoint = matmul(nmcoo,gpoint)
         xgpoint(:) = xgpoint(:) + xcoo0(:)

         call dddb_rd(v,deriv1,deriv2,dipole,xgpoint,.false.,.false.,.false.)

         nactm1=(deriv1(:,1,2))
         nactnmm1=matmul(nactm1,nmcoo)

         gpoint(:) = vnm(:)
         gpoint(m2)=gpoint(m2)+disp
         xgpoint = matmul(nmcoo,gpoint)
         xgpoint(:) = xgpoint(:) + xcoo0(:)

         call dddb_rd(v,deriv1,deriv2,dipole,xgpoint,.false.,.false.,.false.)

         nactm2=(deriv1(:,1,2))
         nactnmm2=matmul(nactm2,nmcoo)

!Determine curl
         curl(i) = (nactnmm1(m1)-nactnm(m1))/disp &
                   -(nactnmm2(m2)-nactnm(m2))/disp

         if (.not. lgnu) write(3,'(3f15.10)') vnm(m1),vnm(m2),curl(i)

      end do

!-----------------------------------------------------------------------
! write gnu file
!-----------------------------------------------------------------------
      if (lgnu) then
         write(3,'(a)') 'set title "Curl " '
         write(3,'(a)') 'set term x11'
         write(3,'(a)') 'set nokey'
         ilbl = index(modelabel(m1),' ')-1
         write(3,'(3a)') 'set xlabel "',modelabel(m1)(1:ilbl),'"'
         ilbl = index(modelabel(m2),' ')-1
         write(3,'(3a)') 'set ylabel "',modelabel(m2)(1:ilbl),'"'

! put data points on surface and get axis boundaries
         xmin=999.0
         xmax=-999.0
         ymin=999.0
         ymax=-999.0
         do i = 1,recordmax
            vnm(1:nmodes) = dbnm(1:nmodes,i)
            write(3,'(3(a,f8.4),a)')&
              'set label at ',vnm(m1),',',vnm(m2),',',curl(i),' " " point '

            xmin = min(xmin,vnm(m1))
            xmax = max(xmax,vnm(m1))
            ymin = min(ymin,vnm(m2))
            ymax = max(ymax,vnm(m2))
         enddo

         xmin = nint(xmin)-1.0
         xmax = nint(xmax)+1.0
         ymin = nint(ymin)-1.0
         ymax = nint(ymax)+1.0

         write(3,'(2(a,f9.4),a)') 'set xrange [',xmin,' : ',xmax,' ]'
         write(3,'(2(a,f9.4),a)') 'set yrange [',ymin,' : ',ymax,' ]'
         write(3,'(a)') 'set zrange [* : *]'

! set up surface
         write(3,'(a)') 'set surface'
         write(3,'(a)') 'set dgrid3d 20,20 splines'
         write(3,'(a)') 'set contour'
         write(3,'(a)') 'set clabel "%10.3e"'
         write(3,'(a)') 'set cntrparam cubicspline'
         write(3,'(a)') 'set cntrparam points 5'
         write(3,'(a)') 'set autoscale z'
         write(3,'(a)') 'set cntrparam levels incremental -1.5,0.1,1.5'

         write(3,'(a)') 'splot "-" u 1:2:3 w l'
         do i = 1,recordmax
            vnm(1:nmodes) = dbnm(1:nmodes,i)
            write(3,'(3f15.10)') vnm(m1),vnm(m2),curl(i)
         end do
         write(3,'(a)') 'e'
         write(3,'(a)') 'pause -1'

      endif

      close(3)

      return
      end subroutine chkcurl

!#######################################################################
!
! Finds the nearest point in the DB to the structure in input file
! (xyz format)
!
!#######################################################################

      subroutine getnear(dbin,recordmax)

      implicit none

      integer(long), intent(in) :: recordmax
      integer(long) :: n,ierr,i,j
      integer(long) :: idbloc

      real(dop), dimension(3,natm)            :: instruct
      real(dop)                               :: mindist,mindisp
      real(dop), dimension(recordmax)         :: r1
      real(dop)                               :: angst = 0.529177
      real(dop)                               :: unev = 27.2114
      character(len=c5)                       :: filein
      character(len=10),dimension(natm)       :: alab
      type(SQLITE_DATABASE), intent(in)       :: dbin
      
!-----------------------------------------------------------------------
! read in data
!-----------------------------------------------------------------------
      call dbrdgeo(dbin,natm,1,recordmax,dbgeo)
      call dbrdapes( dbin,nddstate, 1, recordmax, dbadener)

      if (inpname .eq. ' ') inpname = 'structure.xyz'
      jlbl=index(inpname,' ')-1
      write(filein,'(3a)') name(1:laenge),'/',inpname(1:jlbl)
      call rmblnk(filein)
      open(8,file=filein,form='formatted',status='old',err=97)
      jlbl = index(filein,' ')-1

      write(fileout,'(2a)') name(1:laenge),'/nearest.dat'
      call rmblnk(fileout)
      open(7,file=fileout,form='formatted',status='unknown',err=98)
      ilbl = index(fileout,' ')-1

      write(6,'(/,2a)') ' The structure file is : ',filein(1:jlbl)
      write(6,'(2a)') ' The output file is    : ',fileout(1:ilbl)

!----------------------------------------------------------------------- 
! Read structure
!----------------------------------------------------------------------- 
      do
         read(8,*,iostat=ierr,end=99) n
         if (ierr .eq. 0) exit    ! found number of atms
      enddo

      do i=1,n
         read(8,*) alab(i),(instruct(j,i),j=1,3)
      enddo

!-----------------------------------------------------------------------
! Calculate distances to points in the database
!-----------------------------------------------------------------------
      call distdb(instruct,idbloc,r1,mindist,mindisp,recordmax)

      write(6,'(/a,i0)') 'Closest structure in DB: ',idbloc
      write(6,*) 'RMSD: ',mindist

      write(7,'(/a,i0)') 'Closest structure in DB: ',idbloc
      write(7,*) 'RMSD: ',mindist
      write(7,*) 
      write(7,*) 'Adiabatic Energies: ',&
            (dbadener(i,i,idbloc)*unev,i=1,nddstate)
      write(7,*) 
      write(7,*) 'Structure'
      do i=1,n
         write(7,*) alab(i),(dbgeo((i-1)*3+j,idbloc)*angst,j=1,3)
      enddo

      return
 97   write(6,'(/,a)') ' Cannot open input file : '
      write(6,*) filein
      stop
 98   write(6,'(/,a)') ' Cannot open output file : '
      write(6,*) fileout
      stop
 99   write(6,'(/,a)') ' No. of atoms not given in structure file '
      stop
      end subroutine getnear

      end program trafodb


