!***********************************************************************
!
! analyse program TWPROB
!
! calculates reaction probabilities with the method of
! Tannor/Weeks/Garaschuk; see e.g. JCP 110 (1999) 2761, Sect. II
!
! method of operation:
! 1. prepare an input file for a geninwf run, the initial
!    wavepacket should be a direct product of a (narrow) Gaussian
!    in the translational coordinate times an internal state of
!    the reactants that has the quantum numbers for the channel
!    you are interested in
! 2. use crosscorr to calculate the overlap of this (time-independent)
!    WF with your time-dependent WF
! 3. use twprob to calculate the reaction probability for the
!    chosen channel; additional to the crosscorr output, you need
!    the energy distributions of the two WFs you used (best taken
!    from the enerd files)
!
! FO 05/2005
!
!***********************************************************************

      module twprobmod

      use decimal, only : dop,long

      implicit none

      character(len=8), save     :: eunit, eoffunit
      character(len=240), save   :: outfile, ccfile, edstr1, edstr2
      integer(long), save        :: eddim1, eddim2, epoints, ncross, ncos, nexp,&
                                    mredstr1x, mredstr1y, mredstr1y2,mredstr2x, &
                                    mredstr2y, mredstr2y2,mccross
      real(dop), save            :: factor, emin, emax, eoff, tcut, tau
      real(dop), save            :: energysign
      logical(kind=4)            :: ledstr1,ledstr2

      end module twprobmod

      program twprob

use decimal, only: long,dop,sip
use global
use globmemmod
use versions
use channels
use twprobmod
use keyunitsmod

use logdat
use module_interface

      implicit none

      integer(long), external              :: myiargc
      integer(long)                        :: iarg,i,nonopts,len,ierr,ilbl
      character(len=240)                   :: arg
      logical                              :: isopt

!     initialise variables
      call default
      factor  = 1.0_dop
      epoints = 500
      tcut    = 1.d30
      tau     = 0.d0
      outfile = ' '
      nonopts = 0
      ncos    = 0
      nexp    = 0
      ledstr1 = .false.
      ledstr2 = .false.
      energysign = 1.d0

!     parse command line options
      iarg = myiargc()
      if (iarg .le. 0) then
         call twprobhelp
      else
         call mygetarg(0,options)
         optlaenge  = index(options,' ')
         do i = 1, iarg
            call mygetarg(i,arg)
            ilbl  = index(arg,' ')
            ilbl = min(ilbl,100-optlaenge)
            if(ilbl.gt.0) options = options(1:optlaenge)//arg(1:ilbl)
            optlaenge = optlaenge + ilbl
         end do
      endif
      i=1
 10   call mygetarg(i,arg)

!     look for options (starting with '-')
      isopt = .false.
      if (arg(1:1) .eq. '-') then
         isopt = .true.
!        check for a real number, then it's not an option
         read(arg, *, iostat=ierr) emin
         if (ierr .eq. 0)  isopt = .false.
      endif

!     look for options (starting with '-')
      if (isopt) then

!        -o file
         if (arg .eq. '-o') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,outfile)
            len = index(outfile, ' ') - 1
            call abspath(outfile, len)

!        -F factor
         elseif (arg .eq. '-F') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) factor
            if (ierr .ne. 0) then
               write (6,'(/a/)')&
                  '### illegal argument to option -F ###'
               stop
            endif

!        -t time_cut
         elseif (arg .eq. '-t') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) tcut
            if (ierr .ne. 0) then
               write (6,'(/a/)')&
                  '### argument to option -t must be real number ###'
               stop
            endif

!        -cos n
         elseif (arg .eq. '-cos') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) ncos
            if (ierr.ne.0 .or. ncos.le.0) then
               write (6,'(/a/)')  '### argument to option -cos '&
                  //'must be positive integer ###'
               stop
            endif
            
!        -exp tau
         elseif (arg .eq. '-exp') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) tau
            if (ierr.ne.0 .or. tau.le.0) then
               write (6,'(/a/)')  '### argument to option -exp ' &
                  //'must be positive real number ###'
               stop
            endif
            nexp=1

!        -exp2 tau
         elseif (arg .eq. '-exp2') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) tau
            if (ierr.ne.0 .or. tau.le.0) then
               write (6,'(/a/)')  '### argument to option -exp2 ' &
                  //'must be positive real number ###'
               stop
            endif
            nexp=2

!        -p number_of_energy_points
         elseif (arg .eq. '-p') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) epoints
            if (ierr .ne. 0) then
               write (6,'(/a/)')&
                  '### argument to option -p must be integer ###'
               stop
            endif

!        -e value unit
         elseif (arg .eq. '-e') then
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,arg)
            read (arg, *, iostat=ierr) eoff
            if (ierr .ne. 0) then
               write (6,'(/a/)')&
                  '### first argument to -e must be real number ###'
               stop
            endif
            i=i+1
            if (i.gt.iarg) call twprobhelp
            call mygetarg(i,eoffunit)
            call keylowc(eoffunit)
            call keyeunits(eoff, eoffunit, ierr)
            if (ierr .ne. 0) then
               write (6,'(/a/)')&
                    '### 2nd argument to -e must be valid unit ###'
               stop
            endif

!        -minus
         elseif (arg .eq. '-minus') then
            energysign = -1.d0

!        unknown options
         else
            call twprobhelp
         endif

!     non-options are: Emin, Emax, Unit, CrossCorrFile, Edstr1, Edstr2
      else
         if (nonopts .eq. 0) then
!        parse Emin
            read(arg, *, iostat=ierr) emin
            if (ierr .ne. 0) then
               write (6,'(/4a/)')&
                  '### mandatory argument "emin" must be ',&
                  'a real number, not ',arg,' ###'
               stop
            endif
         elseif (nonopts .eq. 1) then
!        parse Emax
            read(arg, *, iostat=ierr) emax
            if (ierr .ne. 0) then
               write (6,'(/4a/)')&
                  '### mandatory argument "emax" must be ',&
                  'a real number, not ',arg,' ###'
               stop
            endif
         elseif (nonopts .eq. 2) then
!        parse Unit, convert emin/emax to a.u.
            eunit = trim(arg)
            call keylowc(eunit)
            call keyeunits(emin, eunit, ierr)
            if (ierr .ne. 0) then
               write (6,'(/4a/)')&
                  '### mandatory argument "unit" must be ',&
                  'a valid unit, not ',arg,' ###'
               stop
            endif
            call keyeunits(emax, eunit, ierr)
         else
!        remaining arguments are files
            len = index(arg, ' ') - 1
            call abspath(arg, len)
            if (nonopts .eq. 3) then
               ccfile = arg(1:)
            elseif (nonopts .eq. 4) then
               edstr1 = arg(1:)
               ledstr1 = .true.
            elseif (nonopts .eq. 5) then
               edstr2 = arg(1:)
               ledstr2 = .true.
            else
               call twprobhelp
               stop
            endif
         endif

         nonopts = nonopts + 1
      endif

!     process next argument
      i=i+1
      if (i.le.iarg) goto 10

!     check if all mandatory parameters have been specified
      if (nonopts.lt.4) call twprobhelp

!     open log file
      open (unit=ilog,file='twprob.log',form='formatted',iostat=ierr)
      if (ierr.ne.0) then
          message='Cannot open file twprob.log'
          call errormsg
      endif
      logisopen=.true.
      write(ilog,'(2a)')   ' Options : ', options(1:optlaenge)

!     count number of lines in crosscorrelation file
      call linecount(ncross, ccfile, "", ierr)
      if (ierr.ne.0) then
         len = index(ccfile, ' ') - 1
         write(6, '(/3a/)')&
            '### Error opening file "',ccfile(1:len),'" ###'
         write(ilog, '(/3a/)')&
            '### Error opening file "',ccfile(1:len),'" ###'
         stop
      endif

!     count number of lines in files edstr[12]
      if (ledstr1) then
         call linecount(eddim1, edstr1, "", ierr)
         if (ierr.ne.0) then
            len = index(edstr1, ' ') - 1
            write(6, '(/3a/)')&
               '### Error opening file "',edstr1(1:len),'" ###'
            write(ilog, '(/3a/)')&
               '### Error opening file "',edstr1(1:len),'" ###'
            stop
         endif
      else
         eddim1 = 0
         edstr1 = "1.000"
      endif

      if (ledstr2) then
         call linecount(eddim2, edstr2, "", ierr)
         if (ierr.ne.0) then
            len = index(edstr2, ' ') - 1
            write(6, '(/3a/)')&
               '### Error opening file "',edstr2(1:len),'" ###'
            write(ilog, '(/3a/)')&
               '### Error opening file "',edstr2(1:len),'" ###'
            stop
         endif
      else
         eddim2 = 0
         edstr2 = "1.000"
      endif

!     calculate workspace
      workrdim = max(eddim1,eddim2)

!     assign pointers to arrays
      mredstr1x = mrdim
      mrdim = mrdim + eddim1
      mredstr1y = mrdim
      mrdim = mrdim + eddim1
      mredstr1y2 = mrdim
      mrdim = mrdim + eddim1

      mredstr2x = mrdim
      mrdim = mrdim + eddim2
      mredstr2y = mrdim
      mrdim = mrdim + eddim2
      mredstr2y2 = mrdim
      mrdim = mrdim + eddim2

      mccross = mcdim
      mcdim = mcdim + ncross

!     assign pointers for workspace
      mrworkr = mrdim
      mrdim = mrdim + workrdim

!     call C routine for memory allocation

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))

call receiver(mc,mr,mi,ml,ms,mf)

deallocate(mc,mr,mi,ml,ms,mf)

      contains


!***********************************************************************

      subroutine receiver(mc,mr,mi,ml,ms,mf)

      implicit none

      complex(dop), dimension(mcdim), intent(inout)  :: mc
      real(dop), dimension(mrdim), intent(inout)     :: mr
      integer(long), dimension(midim), intent(in)    :: mi
      logical(kind=4), dimension(mldim), intent(in)  :: ml
      complex(sip), dimension(msdim), intent(in)     :: ms
      real(sip), dimension(mfdim), intent(in)        :: mf

      integer(long)       :: ierr,i,j,len,cclen,ed1len,ed2len
      real(dop)           :: yp1,ypn,dt,eval,evala,ed1val,ed2val,&
                             prob,t,tend,dele,ed1max,ed2max,eps
      complex(dop)        :: ftval,cc,x

      eps = 0.0_dop

!     read crosscorrelation file into memory
      call rdcross(mc(mccross),ncross,dt,tend,ccfile,ierr)
      cclen = index(ccfile, ' ') - 1
      if (ierr.ne.0) then
         write(6, '(/3a/)')&
            '### Error reading file "',ccfile(1:cclen),'" ###'
         stop
      endif
      tend=min(tcut,tend)
      
!     read energy distribution files into memory
      ed1len = index(edstr1, ' ') - 1
      if (ledstr1) then
         call rdedstr(mr(mredstr1x),mr(mredstr1y),eddim1,edstr1,ierr)
         if (ierr.ne.0) then
            write(6, '(/3a/)')&
               '### Error reading file "',edstr1(1:ed1len),'" ###'
            stop
         endif
      endif

      ed2len = index(edstr2, ' ') - 1
      if (ledstr2) then
         call rdedstr(mr(mredstr2x),mr(mredstr2y),eddim2,edstr2,ierr)
         if (ierr.ne.0) then
            write(6, '(/3a/)')&
               '### Error reading file "',edstr2(1:ed2len),'" ###'
            stop
         endif
      endif
       
!     determine spline coefficients
      if (ledstr1) then
         yp1=(mr(mredstr1y+1)-mr(mredstr1y))/&
             (mr(mredstr1x+1)-mr(mredstr1x))
         ypn=(mr(mredstr1y+eddim1)-mr(mredstr1y+eddim1-1))/&
             (mr(mredstr1x+eddim1)-mr(mredstr1x+eddim1-1))
         call splinenm(mr(mredstr1x),mr(mredstr1y),eddim1,yp1,ypn,&
                       mr(mredstr1y2),mr(mrworkr))
      endif

      if (ledstr2) then
         yp1=(mr(mredstr2y+1)-mr(mredstr2y))/&
             (mr(mredstr2x+1)-mr(mredstr2x))
         ypn=(mr(mredstr2y+eddim2)-mr(mredstr2y+eddim2-1))/&
             (mr(mredstr2x+eddim2)-mr(mredstr2x+eddim2-1))
         call splinenm(mr(mredstr2x),mr(mredstr2y),eddim2,yp1,ypn,&
                       mr(mredstr2y2),mr(mrworkr))
      endif

!     find good energy distribution cutoffs
      ed1max = 0.d0
      if (ledstr1) then
         do i=1,eddim1
            ed1max = max(ed1max, mr(mredstr1y+i-1))
         enddo
      endif
      ed2max = 0.d0
      if (ledstr2) then
         do i=1,eddim2
            ed2max = max(ed2max, mr(mredstr2y+i-1))
         enddo
      endif

!     is output filename set?
      if (outfile.ne.' ') then
!        yes: open the file for writing
         iout=13
         open(iout,file=outfile,form='formatted',err=666)
         endfile   iout
         backspace iout
         len = index(outfile,' ') - 1
         write(iout,'(2a)') '# The output file is: ',outfile(1:len)
      else
!        no: use stdout
         iout=6
      endif
!     write header
      write(iout,'(a/,3(2a/),a,f14.10/,a)')&
       '# Tannor-Weeks probabilities (twprob) for ...',&
       '# cross-correlation file: ', ccfile(1:cclen),&
       '# energy distribution 1 : ', edstr1(1:ed1len),&
       '# energy distribution 2 : ', edstr2(1:ed2len),&
       '# energy shift [eV]     : ', eoff*ev,&
       '#        Energy[eV]         edistr1            edstr2&
             Probability'
      write(ilog,'(a/,3(2a/),a,f14.10/,a)')&
       '# Tannor-Weeks probabilities (twprob) for ...',&
       '# cross-correlation file: ', ccfile(1:cclen),&
       '# energy distribution 1 : ', edstr1(1:ed1len),&
       '# energy distribution 2 : ', edstr2(1:ed2len),&
       '# energy shift [eV]     : ', eoff*ev,&
       '#        Energy[eV]         edistr1            edstr2&
             Probability'

!     set default factor
      factor = factor / (4.0_dop*pi*pi)

!     loop over the specified energy range
      dele = (emax-emin)/dble(epoints)
      do j=0,epoints
         ed1val = 0.d0
         ed2val = 0.d0
         eval = emin + j*dele - eoff

!        check if the E-value is even in the range of the two
!        energy distributions
         evala = eval*ev ! energy index of edstr is in eV
         if (ledstr1) then
            if ((evala.lt.mr(mredstr1x)).or.&
                (evala.gt.mr(mredstr1x-1+eddim1))) then
               prob = 0.d0
               goto 30
            endif
         endif
         if (ledstr2) then
            if ((evala.lt.mr(mredstr2x)).or.&
                (evala.gt.mr(mredstr2x-1+eddim2))) then
               prob = 0.d0
               goto 30
            endif
         endif

!        determine values of energy distributions for this E-value
         if (ledstr1) then
            call splintnm(mr(mredstr1x),mr(mredstr1y),mr(mredstr1y2),&
                          eddim1,evala,ed1val)
         else
            ed1val = 1.d0
         endif
         if (ledstr2) then
            call splintnm(mr(mredstr2x),mr(mredstr2y),mr(mredstr2y2),&
                          eddim2,evala,ed2val)
         else
            ed2val = 1.d0
         endif
!        calculate Fourier transform of crosscorrelation function
!        at this E-value (straight-forward integration with Trapezregel)
         ftval = 0.5_dop*mc(mccross)
         do i=2,ncross
            t = (i-1)*dt     ! here t-unit is fs
            if (t.gt.tcut) goto 20
            cc= mc(mccross+i-1)
!           damp with cos/exp
            if (ncos.gt.0) then
               cc = cc*(cos(0.5*pi*t/tend))**ncos
            endif
            if (nexp.gt.0) then
               cc = cc*exp(-(t/tau)**nexp)
            endif
            x = exp(dcmplx(0.0_dop,eval*t*fs))*cc
            ftval = ftval + x
         enddo
 20      ftval = ftval - .5*x
         ftval = ftval*dt*fs

!        calculate probability(E)
         if ((ed1val .lt. eps*ed1max) .or.&
             (ed2val .lt. eps*ed2max)) then
            prob = 0.0_dop
         else
            prob = factor*abs(ftval)**2/(ed1val*ed2val)
         endif

!        output energy and probability
 30      write(iout, '(1x,f18.14,2x,g18.12,2x,g18.12,4x,e18.12)') &
               (eval+eoff)*ev,ed1val,ed2val,prob
      enddo

      close(iout)
      stop ! finito.

 666  len = index(outfile,' ') - 1
      write(6,'(/3a/)')&
         '### Error opening file "',outfile(1:len),'" for writing ###'
      stop

      end subroutine receiver ! of receiver


!***********************************************************************
!
! subroutine RDCROSS
!
! read a crosscorrelation file into memory
!
! cross : array for storing the crosscorrelation function
! n     : will contain number of values read
! dt    : will contain timestep (as read, i.e. probably in fs)
! tend  : will contain the final time (again in fs)
! filename : guess!
! ierr  : for signaling errors, 0=ok
!
!***********************************************************************

      subroutine rdcross(cross, n, dt, tend, filename, ierr)

      implicit none

      complex(dop), dimension(*), intent(out) :: cross
      real(dop)                               :: t0,t,a,b
      real(dop), intent(out)                  :: dt,tend
      integer(long)                           :: icroco, nline
      integer(long), intent(out)              :: n,ierr
      character(len=*), intent(in)            :: filename
      character(len=80)                       :: buffer

!     open the crosscorrelation file
      icroco = 30
      open(icroco, file=filename, form='formatted', status='old',&
           err=666)

!     skip comments
 10   nline=nline+1
      read(icroco,'(a)',end=667,err=667) buffer
      if (buffer(1:1).eq.'#') goto 10
      if (buffer(2:2).eq.'#') goto 10
      if (buffer     .eq.' ') goto 10

!     process first data line
      read(buffer,*,err=667) t0,a,b
      cross(1) = dcmplx(a,b)

!     process rest of lines
      n = 2
 20   nline=nline+1
      read(icroco,'(a)',end=30,err=667) buffer
      if (buffer(1:1).eq.'#') goto 20
      if (buffer(2:2).eq.'#') goto 20
      if (buffer     .eq.' ') goto 20
      read(buffer,*,err=667) t,a,b
      cross(n) = dcmplx(a,b)
      if(n.eq.2) then
         dt = t-t0
      else if( abs(t-tend-dt) .gt. 5.0d-7 ) then
         write(6,'(/a,/a,f14.8)') &
           ' ### Time points in cross file not equally spaced! ###', &
           ' Time:', t
         stop
      endif
      tend = t
      n = n+1
      goto 20
 30   n = n-1 ! finished

      ierr=0
      close(icroco)
      return

 666  ierr=2 ! error opening file
      return

 667  ierr=3 ! error reading file
      write(6,'(//a,i5,a)') ' ### Error in line',nline, &
                            ' . The wrong line reads:'
      write(6,'(2a)') trim(buffer)
      close(icroco)
      return

      end subroutine rdcross ! of rdcross

!***********************************************************************
!
! subroutine RDEDSTR
!
! read an energy distribution file into memory.
!
! x : array for storing the E-values
! y : array for storing the Delta(E)-values
! n : will contain number of points read
! filename : guess!
! ierr : for signaling errors, 0=ok
!
! code heavily inspired by rdenerd/rdff from flxutils.F
!
!***********************************************************************

      subroutine rdedstr(x, y, n, filename, ierr)

      implicit none

      real(dop)                            :: xdum,ydum
      real(dop), dimension(*), intent(out) :: x,y
      integer(long)                        :: ienerd,len,subdim,g
      integer(long), intent(out)           :: n,ierr
      character(len=*), intent(in)         :: filename
      character(len=80)                    :: buffer

!     open the file
      ienerd = 27
      open(ienerd, file=filename, form='formatted', status='old',&
           err=666)

!     if filename ends with 'enerd' or 'adwkb', it gets special treatment
      len = index(filename,' ') - 1
      if (filename(len-4:len).eq.'enerd' .or.&
          filename(len-4:len).eq.'adwkb') then
!        first line contains number of grid points
         read(ienerd,*,end=667,err=667) subdim
!        skip these lines
         do g=1,subdim
            read(ienerd,*,end=667,err=667) xdum,ydum
         enddo

!     non-enerd files: skip comments at top
      else
 30      read(ienerd,'(a8)',end=667,err=667) buffer
         if (buffer(1:1) .eq. '#') goto 30
         if (buffer .eq. ' ') goto 30
         backspace ienerd
      endif

!     now read 1st and 2nd column
      n = 1
 10   read(ienerd,*,end=20,err=667) x(n),y(n)
      n = n + 1
      goto 10
 20   n = n - 1 ! finished

      ierr=0
      close(ienerd)
      return

 666  ierr=2 ! error opening file
      return

 667  ierr=3 ! error reading file
      close(ienerd)
      return

      end subroutine rdedstr ! of rdedstr



!***********************************************************************
!
! subroutine TWPROBHELP
!
! output help text for twprob to stdout
!
!***********************************************************************

      subroutine twprobhelp

      implicit none

      integer(long)       :: i

      write (6, '(72a)') ('-', i=1,72)
      write (6, '(27(a/),a)')&
       'Purpose: calculate reaction probability with the Tannor-Weeks',&
       '         method (see JCP 110 (1999) 2761.)',&
       ' ',&
       'Usage: twprob [ -o -F -e -p -t -h -? ]',&
       '                emin emax unit file edstr1 edstr2',&
       ' ',&
       'Options:',&
       ' -o FILE: filename for output',&
       ' -F x   : multiply output with a factor x',&
       ' -e R U : shift energy scale in output by value R in units U',&
       ' -p I   : number of energy points is I (default: 500)',&
       ' -t tcut : read crosscorrelation function only up to T=tcut',&
       ' -cos n  : multiply crosscorr.function with cos(pi/2*t/T))**n',&
       ' -exp tau: multiply crosscorr.function with exp(-t/tau)',&
       ' -exp2 tau:multiply crosscorr.function with exp(-(t/tau)**2)',&
       ' -[h?]  : print this help text',&
       ' ',&
       '"emin" and "emax" together with "unit" determine the energy',&
       'range of the output.',&
       '"file" contains the time-dependent overlap of the propagated',&
       'WF with a reference WF (output of crosscorr).',&
       '"edstr[12]" are files that contain the energy distributions of',&
       'the two WFs. This can be an enerd file, or any other file, in',&
       'which case the distribution is determined by columns 1 and 2',&
       '(useful for autospec/cspec/flux files). If an energy ',&
       'distribution is omitted, it is replaced by a constant of 1.0.'
      write (6, '(72a)') ('-', i=1,72)

      stop
      end subroutine twprobhelp ! of twprobhelp


      end program twprob ! of twprob

