!***********************************************************************
!                            MODULE DCORE
!
! MCTDH module containing the core subroutines for the dynamical
! wavefunction code (i.e. keywords initorb and dynspf)
!
! Contains subroutines: funkspawn, dmfields, dmatrizen1, matrizen2b,
! newavec, newavechartree, scaleavechartree, dmatrizen3
!
! Author: David Mendive-Tapia
! Year  : 2019
!
!***********************************************************************

module dcoremod

use global
use globmemmod
use dglobal
use maxv
use dvrdatmod
use griddatmod  ! zpsi(s),zetf(m,s),subdim(m)
use psidef   ! vdim(m,s),dim(m,s),ndim(m,s)
use runpropmod  ! lconst
use timing      ! tspawn
use operdef ! zham(nham),kzahl(nham),kf(k),ki(k)
use hpsimod    ! hpsidim
use channels
use datenmod
use logdat

use hphimod, only: hphi1mk
use tdkoeffs
use projectmod, only: projector,projectorT

use op1lib, only: zerovxz,cpvxz,normvxz
use op2lib, only: swapsxxi,addvxxzo
use mtlib, only: qtxxzz,mttxzz,mtxxzz,qttxzz,mtxxzza,mtaxzz
use xvlib, only: xvxxdzr,xvxxzzo,xvixdzo
use mmlib, only: mmaxzz
use rmlib, only: rmmxxxzz,rm1hxxxzz

      implicit none

      public

      contains

!***********************************************************************
! FUNKSPAWN
!
! Computes the matrix Delta (Eq. 52), for the use of optimal SPFs,
! as described in JCP 153, 234114 (2020)
!
! INPUT
! time      : current time
! m         : mode of interest
! s         : electronic state of interest
! psi       : MCTDH wavefunction
! maxblock2 : maximum size for fmat
!
! mc,mr,mi,ml are standard arrays (see documentation)
!
! OUTPUT
! dltmat    : optimal unoccupied SPFs matrix Delta
! fmat      : mean-fields-like matrix F
! denmat    : density matrix
!
!***********************************************************************

      subroutine funkspawn (time,m,s,psi,dtpsi,dltmat,fmat,denmat,maxblock2)

      implicit none

!     INPUT
!     Wavefunction
      real(dop)    :: time
      complex(dop) :: psi(dgldim),dtpsi(dgldim)
!     Storage arrays
!     Handling of wavefunction
      integer    m,s
!
!     Debugging
!      integer    hktype,nham
!      complex*16 hunpsi(subdim(m)*dim(m,s))
!      complex*16 Idimsub(maxval(dim)*sdim)
!      complex*16 dltmatpp(sdim*sdim)
!      complex*16 MatD(subdim(m)*subdim(m))
!      complex*16 workc(subdim(m)*dim(m,s))
!      complex*16 imagn
!      parameter  (imagn = (0.0d0,1.0d0))
!     Debugging

!     INTERMEDIATE
      complex(dop) :: fmatpp(maxblock2)

!     OUTPUT
!     Optimal SPFs I
      integer(long) :: maxblock2
      complex(dop)  :: dltmat(subdim(m)*subdim(m))
      complex(dop)  :: fmat(maxblock2)
      complex(dop)  :: denmat(dmatdim)

      integer(long), save                        :: tid=0 ! timer ID

!-----------------------------------------------------------------------
!     Start timer
!-----------------------------------------------------------------------
      call get_timer(tid,'Tspawn')
      call start_timer(tid)

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(fmatpp,maxblock2)

!     Debugging
!      call zerovxz(MatD,subdim(m)*subdim(m))
!      call zerovxz(workc,subdim(m)*dim(m,s))
!      call zerovxz(hunpsi,subdim(m)*dim(m,s))
!      call zerovxz(Idimsub,maxval(dim)*sdim)
!      call zerovxz(dltmatpp,sdim*sdim)
!     Debugging

!-----------------------------------------------------------------------
!     Build matrix F and matrix Delta
!-----------------------------------------------------------------------
!     Compute matrix F
      call dmfields(m,s,psi,time,1,fmat,denmat,maxblock2)

!     Debugging
!      call diagnostic1(psi,F,m,s,maxblock2)
!      call diagnostic3(psi(zetf(m,s)),F,vdim(m,s),dim(m,s),ndim(m,s),
!     +     subdim(m))
!     Debugging

!     Debugging
!     Compute matrix F''
!     a) Compute uncorrelated Hamiltonian terms
!      hktype=0
!      nham=1
!      call hunphi1ms (psi(zetf(m,s)),hunpsi,workc,m,s,dim(m,s),
!     +     subdim(m),mc(mchc),mr(mrhr),mi(mihi),ml(mlhl),hktype,nham,
!     +     time)
!     b) Substra!t from dtpsi
!      call cpvxz(dtpsi(zetf(m,s)),workc,subdim(m)*dim(m,s))
!      call xvxxdzr (oner,hunpsi,workc,subdim(m)*dim(m,s))
!     !) Build F''
!     See Eq. 31 in notes do!ument "Optimal initial orbitals"
!     dtpsi(l,j)*Ave!(i,j,k) = F''(i,l,k)
!      call mtxxzz (workc,psi(zpsi(s)),fmatpp,
!     +     vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
!     d) Substra!t uncorrelated terms in F
!     See Eq. 31 in notes do!ument "Optimal initial orbitals"
!     F'(i) - imag*F''(i) = F(i)
!      call xvxxzzr (imagn,fmatpp,fmat,vdim(m,s)*subdim(m)*ndim(m,s))
!     Debugging

!     Compute matrix Delta
!     See Eq. 52 in JCP 153, 234114 (2020)
!     dconjg(F(i,j,k))*F(i,l,k) = Delta(j,l)
      call rm1hxxxzz (fmat,dltmat,vdim(m,s),subdim(m),ndim(m,s))
!     Apply projector (1-P)
      call projectorT (psi(zetf(m,s)),dltmat,dimocc(m,s),subdim(m),subdim(m))
      call projector (psi(zetf(m,s)),dltmat,dimocc(m,s),subdim(m),subdim(m))
! Debugging
!         call cpqaz (dltmat,MatD,subdim(m))
!         call projector (psi(zetf(m,s)),MatD,dim(m,s),subdim(m),subdim(m))
!         call cpvxz (MatD,dltmat,subdim(m)*subdim(m))
! Debugging
! Debugging
!        Compute matrix D''
!        See Eq. 32 in notes document "Optimal initial orbitals"
!        rho(dim,dim)*dtpsi(dim,subdim) = I(dim,subdim)
!         call qmxxzz (mc(mcdicht2),dtpsi(zetf(m,s)),Idimsub,dim(m,s),subdim(m))
!        dconjg(dtpsi(subdim,dim))*D'(dim,subdim) = D''(subdim,subdim)
!         call mmaxzzh (dtpsi(zetf(m,s)),Idimsub,dltmatpp,dim(m,s),subdim(m))
!        Substract uncorrelated terms in D
!        See Eq. 36 in notes document "Optimal initial orbitals"
!        D'(subdim,subdim) - D''(subdim,subdim) = D(subdim,subdim)
!         call submxxzo1(dltmat,dltmatpp,subdim(m),subdim(m))
! Debugging

!-----------------------------------------------------------------------
!     Stop timer
!-----------------------------------------------------------------------
      call stop_timer(tid)

      return
      end subroutine funkspawn 

!#######################################################################
! DMFIELDS
!
! Calculate the mean-field-like matrix F (Eq. 49),
! as described in JCP 153, 234114 (2020)
!
! INPUT
! m         : mode of interest
! s         : electronic state of interest
! psi       : MCTDH wavefunction
! time      : current time
! nham      : "Hamiltonian" operator
! lwork     : logical for memory pointer (dependent on CMF, e.g. lconst)
! maxblock2 : maximum size for fmat
!
! mc,mr,mi,ml are standard arrays (see documentation)
!
! OUTPUT
! fmat      : mean-fields-like matrix F
! denmat    : density matrix
!
!#######################################################################

      subroutine dmfields(m,s,psi,time,nham,fmat,denmat,maxblock2)

      implicit none

!     INPUT
!     Wavefunction
      real(dop)    :: time
      complex(dop) :: psi(dgldim)
!     Handling of wavefunction I
      integer(long) :: m,s,s0,s1
!     Handling of wavefunction II
      integer(long)   :: nham,k,k1,k2

!     INTERMEDIATE
      integer(long) :: maxblock2
      complex(dop) :: hpsi(hpsidim)
      complex(dop) :: I(maxblock2),I2(dim(m,s)*subdim(m))
      complex(dop) :: fmatpp(maxblock2)

!     OUTPUT
      complex(dop) fmat(maxblock2)
      complex(dop) denmat(dmatdim)

      integer(long), save                        :: tid=0 ! timer ID

!-----------------------------------------------------------------------
!     Start timer
!-----------------------------------------------------------------------
      call get_timer(tid,'Tspawnmat')
      call start_timer(tid)

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(fmat,maxblock2)
      call zerovxz(I,maxblock2)
      call zerovxz(I2,dim(m,s)*subdim(m))
      call zerovxz(hpsi,hpsidim)

!-----------------------------------------------------------------------
!     Build matrix F'
!-----------------------------------------------------------------------
!     Correlated terms pointers
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

!     Elements of matrix F' for every kth Hamiltonian term
      do k=k1,k2
         s0=kf(k)
         s1=ki(k)
         if (s1 .ne. s0) then
            call hphi1mk(psi(zetf(m,s1)),hpsi,m,k,dim(m,s1),subdim(m))
            call matrizen2b(k,s0,s1,psi,hteil, &
                 time,hpsi,fmat,m,maxblock2)
         else
!           Single-set
!           Compute hpsi elements
            call hphi1mk(psi(zetf(m,s)),hpsi,m,k,dim(m,s),subdim(m))
            call dmatrizen1(k,s,m,psi,time,hpsi,fmat,maxblock2,hteil)
         endif
      enddo

!-----------------------------------------------------------------------
!     Build matrix F''
!-----------------------------------------------------------------------
      call qttxzz (denmat(dmat(m,s)),psi(zpsi(s)),I,vdim(m,s),dim(m,s),ndim(m,s))
      call rmmxxxzz (psi(zpsi(s)),fmat,I2, vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
      call mttxzz (I2,I,fmatpp,vdim(m,s),dim(m,s),ndim(m,s),subdim(m))

!-----------------------------------------------------------------------
!     Substract matrix F''
!-----------------------------------------------------------------------
      call xvxxdzr (oner,fmatpp,fmat,vdim(m,s)*subdim(m)*ndim(m,s))

!-----------------------------------------------------------------------
!     Stop timer
!-----------------------------------------------------------------------
      call stop_timer(tid)

      return
      end subroutine dmfields

!#######################################################################
! DMATRIZEN1
!
! For Hamiltonian terms involving only a single state,
! calculates the necessary terms in order to build the matrix F (Eq. 49)
! as described in JCP 153, 234114 (2020))
!
! INPUT
! k         : number of correlated Hamiltonian term
! s         : electronic state of interest
! mode      : mode of interest
! psi       : coefficients for the relevant state
! time      : current time
! hpsi      : all products h|phi>
! maxblock2 : maximum size for fmat
!
! hteil, workc are standard arrays (see documentation)
! block,dim,vdim,ndim are standard arrays for one state
!
! OUTPUT
! fmat      : mean-fields-like matrix F for the kth Hamiltonian term
!
!#######################################################################

      subroutine dmatrizen1(k,s,mode,psi,time,hpsi,fmat,maxblock2,hteil)

      implicit none

!     INPUT
!     Wavefunction
      complex(dop)  :: psi(adim)
!     Handling of wavefunction I
      integer(long) :: m,s,mode
      integer(long) :: zeig1,zeig2
!     Handling of wavefunction II
      integer(long) :: k,nham,h
      real(dop)     :: time,c,gpoint(maxdim)
!     Input data
      complex(dop)  :: hpsi(subdim(mode),dim(mode,s)),hteil(pmatdim)
!     Storage arrays
      complex(dop)  :: workc(2*maxblock)

!     OUTPUT
      integer(long) :: maxblock2
      complex(dop)  :: fmat(maxblock2)

      if (diag(mode,k)) return  ! term does not contribute to F

!-----------------------------------------------------------------------
!     calculate value of koeff(k) if time-dependent
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- INITIALISE VARIABLES ---------------------------------------------
!     workc is divided into three parts of length block.
!     The result of a multiplication is stored after zeig2.
!     Before multiplication, the vector to be multiplied is moved to after
!     zeig1.
!-----------------------------------------------------------------------
      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))

!-----------------------------------------------------------------------
!     Loop over modes.
!     First set initial pointers, then multiply wavefunction
!     with single-particle hamiltonian matrices corresponding to
!     dimensions numbers lower than mode
!-----------------------------------------------------------------------
      do m=1,mode-1
         if (.not.diag(m,k)) then
            call swapsxxi(zeig1,zeig2)
            call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         endif
      enddo

!-----------------------------------------------------------------------
!     Multiply array with single-particle hamiltonian matrices for
!     modes higher than m.
!     If H is unitoperator for mode m do not bother
!-----------------------------------------------------------------------
      do m=mode+1,nmode
         if (.not.diag(m,k)) then
            call swapsxxi(zeig1,zeig2)
            call qtxxzz(hteil(pmat(m,k)),workc(zeig1),&
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         endif
      enddo

      m=mode
!-----------------------------------------------------------------------
!     Include Hamitlonian term koefficient
!-----------------------------------------------------------------------
      h=hamilton(fkoe,k)
      if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104)  &
         call xvxxzzo(koeff(k),workc(zeig2),vdim(m,s)*dim(m,s)*ndim(m,s))

!-----------------------------------------------------------------------
!     Now calculate the summation for F.
!-----------------------------------------------------------------------
      call mtxxzza(hpsi,workc(zeig2),fmat,vdim(m,s),dim(m,s),ndim(m,s),subdim(m))

      return
      end subroutine dmatrizen1

!#######################################################################

      subroutine matrizen2b(k,s,s1,psi,hteil,time,hpsi,F,mode,maxblock2)

      implicit none

      integer(long) :: zeig1,zeig2,zeig3,swapzeig,m,m1,k,s,s1,vdim1,&
                       block1,modus,h,nham
      real(dop)     :: time,c,gpoint(maxdim)
      complex(dop)  :: psi(adim),dtpsi(adim),hteil(pmatdim), &
                       hloch(hlochdim),psi1(adim)
      complex(dop)  :: workc(1:3*MaxBlock)
!     Optimal SPFs
      integer(long) :: mode,maxblock2
      complex(dop)  :: hpsi(hpsidim)
      complex(dop)  :: F(maxblock2)
!     Parameters and constants
      complex(dop), parameter ::  imagn = (0.0d0,1.0d0) 

!-----------------------------------------------------------------------
! calculate value of koeff(k) if time-dependent
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- INITIALISE VARIABLES ---------------------------------------------
! workc is divided into three parts of length maxblock, the longest
! possible
! length of the array during the multiplication.
! The result of a multiplication is stored after zeig2.
! Before multiplication, the vector to be multiplied is moved to after
! zeig1.
! After zeig3 is stored the vector multiplied up to mode m-1
!-----------------------------------------------------------------------
      call zerovxz(workc,3*MaxBlock)
      zeig1=maxblock+1
      zeig2=1
      zeig3=2*maxblock+1
      call cpvxz(psi(zpsi(s1)),workc(zeig2),block(s1))
      block1=block(s1)

!-----------------------------------------------------------------------
!  Loop over modes.
!  First set initial pointers, then multiply wavefunction of state s1
!  with single-particle hamiltonian matrices corresponding to
!  dimensions numbers lower than mode
!  as the matrix elements are between different states with different
!   number of coefficients, the following pointers are needed:
!  block1: number of elements in multiplied array.
!  vdim1: equivalent of vdim during multiplication over the modes lower
!  than m.
!  vdim2: equivalent of vdim during multiplication over the modes higher
!  then m.
!  before a multiplication, block=vdim1*dim(m,s1)*ndim(m,s1)
!  after a multiplication, block=vdim1*dim(m,s)*ndim(m,s1)
!  s1 is on RHS, s on LHS of matrix element
!-----------------------------------------------------------------------
      do m=1,mode-1
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         call mtxxzz(hteil(pmat(m,k)),workc(zeig1), &
              workc(zeig2),vdim(m,s),dim(m,s1),ndim(m,s1),dim(m,s))
         block1=vdim(m,s)*dim(m,s)*ndim(m,s1)
      enddo

!-----------------------------------------------------------------------
!  Multiply array with single-particle hamiltonian matrices for
!  modes higher than m.
!  If H is unit-operator for mode m, or basis is complete, do not bother
!  as hloch is never needed.
!-----------------------------------------------------------------------
      vdim1=vdim(mode,s)*dim(mode,s1)  ! add missing m dimension to vdim
      do m=mode+1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig

         call mtxxzz(hteil(pmat(m,k)),workc(zeig1), &
              workc(zeig2),vdim1,dim(m,s1),ndim(m,s1),dim(m,s))
         vdim1=vdim1*dim(m,s)
      enddo

      m=mode
!-----------------------------------------------------------------------
!     Include Hamitlonian term koefficient
!-----------------------------------------------------------------------
      h=hamilton(fkoe,k)
      if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104)  &
         call xvxxzzo(koeff(k),workc(zeig2),vdim(m,s)*dim(m,s)*ndim(m,s))

!-----------------------------------------------------------------------
! Now calculate the summation for F.
!-----------------------------------------------------------------------
      call mtxxzza(hpsi,workc(zeig2),F,vdim(m,s),dim(m,s),ndim(m,s),subdim(m))

      return
      end subroutine matrizen2b

!***********************************************************************
! NEWAVEC
!
! Computes the new A-vector coefficients after enlarging the MCTDH
! wavefunction
!
! INPUT
! m         : mode of interest
! s         : electronic state of interest
! fmat      : mean-fields-like matrix F
! newspf    : additional unoccupied SPFs
! eigenavec : eigenvalues associated with the unoccupied SPFs
! maxblock2 : maximum size for fmat
!
! cvdim,nspf,cndim
!
! OUTPUT
! newas     : new A-vector coefficients stored as array(vdim,dim,ndim)
! psiadd    : new A-vector coefficients stored as array(vdim,nspf,ndim)
!
!***********************************************************************

      subroutine newavec (m,s,fmat,newspf,eigenavec,psiadd,nwas, &
                 maxblock2,nspf,cvdim,cndim,largernspfdim)

      implicit none

!     INPUT
!     Handling of wavefunction
      integer(long) ::  m,s
!     Handling of A-vector
      integer(long) ::  v,e,n
!     Optimal SPFs I
      integer(long) ::  maxblock2
      complex(dop) :: fmat(maxblock2)
      integer(long) ::  cvdim,cndim
!      complex(dop) :: fmat(cvdim,subdim(m),cndim)
!     Optimal SPFs II
      real(dop) :: EigenAvec(imaxspf)
      complex(dop) :: NewSpf(subdim(m)*nspf)

!     OUTPUT
      integer(long) ::  nspf,largernspfdim
      complex(dop) :: psiadd(vdim(m,s),nspf,ndim(m,s))
      complex(dop) :: NwAs(vdim(m,s),largernspfdim,ndim(m,s))

!     DEBUGGING
      logical(kind=4), parameter ::    ldebug = .false.
      complex(dop) :: ovl(nspf,nspf)

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(NwAs,vdim(m,s)*dim(m,s)*ndim(m,s))
      call zerovxz(psiadd,vdim(m,s)*(nspf)*ndim(m,s))

!-----------------------------------------------------------------------
!     Compute new A-vector columns
!-----------------------------------------------------------------------
!     See Eq. 54 in JCP 153, 234114 (2020)
!     phi(l,j)*F(i,j,k) = Avec(i,l,k)
      call mtaxzz (NewSpf,fmat,psiadd,cvdim,subdim(m),cndim,nspf)
!      call mttxzz (NewSpf,F,psiadd,cvdim,subdim(m),cndim,nspf)

!     To sort out later: this could be done somewhere else
!     Consistecy check and data treatment
!      if (.not.linitorb) then
      do e=1,nspf
!        Check coefficient's size
         if (EigenAvec(e).le.1.0d-16) then
!           To sort out later: how small eigenavec should be allowed?
!           Current printing is 1.0d-16
            routine='funkspawn'
            message='WARNING: EigenAvec=0'
            call errormsg
         endif
!        Store psiadd(vdim,nspf,ndim) as NwAs(vdim,dim,ndim)
         if (loptavec2nd) then
            do n=1,ndim(m,s)
               do v=1,vdim(m,s)
                  nwas(v,e,n)=psiadd(v,e,n)
               enddo
            enddo
         endif
      enddo
!      endif

!     To sort out later: this could be done as a diagnostic subroutine
!     Debugging
!     Compute epsilon as a check of consistency
!     See Eq. 55 in JCP 153, 234114 (2020)
!     dconjg(Avec(i))*Avec(i)=mu*tau^2
      if (ldebug) then
         call rmmxxxzz(psiadd,psiadd,ovl,vdim(m,s),nspf,ndim(m,s),nspf)
         do e=1,nspf
            write(ilog,'(a,E12.6)') ' Eigenvalue: ',EigenAvec(e)
            write(ilog,'(a,"(",E12.6,",",E12.6,")")') ' Epsilon   : ', ovl(e,e)
!           ovl(e,e)*SumEigen/EigenAvec(e)
            if (abs(real(ovl(e,e))-betavar).gt.1.0d-10) then
               write(iopteigen,'(a,f12.5,a)') &
               ' WARNING: epsilon check failed at t=',tcurrent/fs,' fs'
               write(iopteigen,'(a,e12.5,a,e12.5)') &
               ' |<Avec|Avec>-epsilon|=', &
               abs(real(ovl(e,e))-betavar),' >',1.0d-10
               write(iopteigen,'(a)')
            endif
      enddo
!     Debugging

      endif

      return
      end subroutine newavec

!***********************************************************************
! NEWOPTHARTREE
!
! Computes the new A-vector coefficients after enlarging the MCTDH
! wavefunction. Only for the special case of a Hartree product.
!
! INPUT
! psi       : MCTDH wavefunction
! hteil     : matrix-elements <phi(i)|h|phi(j)>
! lcavec    : logical (e.g. make new A-vector imaginary or real)
!
! OUTPUT
! psi       : MCTDH wavefunction
!
!***********************************************************************

      subroutine newavechartree (hteil,psi,SumEigen,MaxEigen,EigenTrace,lcavec)

      implicit none

!     INPUT/OUTPUT
!     Wavefunction
      complex(dop) :: psi(dgldim)

!     INPUT
!     Handling of wavefunction
      integer(long) ::    m,s
!     Handling of A-vector
      integer(long) ::    v,e,n
!     Handling of operators
      integer(long) ::    nham,s1
      integer(long) ::    k,k1,k2
!     Matrix-elements <phi(i)|h|phi(j)>
      complex(dop) :: hteil(pmatdim)
!     Optimal A-vector
      logical(kind=4) ::    lcavec
!     Other
      real(dop) ::     SumEigen(nmode),MaxEigen(nmode),EigenTrace(nmode)

!     INTERMEDIATE
      complex(dop) :: tau
      complex(dop) :: avec(adim)
!      complex(dop) :: psiadd(vdim(m,s),nspf,ndim(m,s))
!     Computed quantities
      real(dop) ::     norm
!     Parameters and constants
      complex(dop), parameter :: imagn = (0.0d0,1.0d0)

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(avec,adim)

!-----------------------------------------------------------------------
!     Build matrix F
!-----------------------------------------------------------------------
!     Correlated terms pointers
      nham=1
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

!     Elements of matrix F for every kth Hamiltonian term
      do k=k1,k2
         s=kf(k)
         s1=ki(k)
         if (s1 .ne. s) then
!           Multi-set
            routine='newavechartree'
            message='InitOrb for multi-set not implemented yet!'
            call errormsg
         else
!           Single-set
!           Compute matrix F elements
            call dmatrizen3(k,s,psi,hteil,avec)
         endif
      enddo

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call scaleavechartree(m,s,avec)
         enddo
      enddo
      call normvxz(avec,norm,adim)
      norm=norm**2

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
      if (ioptavec.eq.0) then
!        Definition of tau: individual eigenvalue(s)
! To sort out later: for the moment use SumEigen instead
         tau=sqrt(betavar/maxval(SumEigen))
      else if (ioptavec.eq.1) then
!        Definition of tau: sum of eigenvalues
         tau=sqrt(betavar/maxval(SumEigen))
      else if (ioptavec.eq.2) then
!        Definition of tau: largest eigenvalue
         tau=sqrt(betavar/maxval(MaxEigen))
      else if (ioptavec.eq.3) then
!        Definition of tau: value read from input
         tau=betavar
      else if (ioptavec.eq.4) then
!        Definition of tau: automatic selection
         tau=tauthr/dsqrt(maxval(MaxEigen))
      else if (ioptavec.eq.5) then
!        Definition of tau: automatic selection (trace)
         tau=tauthr/dsqrt(maxval(EigenTrace))
      else if (ioptavec.eq.6) then
!        Definition of tau: automatic selection (trace) + factor
         tau=tauthr/dsqrt(maxval(EigenTrace))
         tau=tau*betavar
      endif

      if (loptavec2nd) psi(1)=sqrt(1.0-tau**2*norm)

!-----------------------------------------------------------------------
!     Output message
!-----------------------------------------------------------------------
      tau=min(0.5*out1,real(tau))
      write(ilog,'(a,E12.6)') ' Tau2: ',real(tau)

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
      if (lcavec) then
         tau=-imagn*tau
      else
         tau=1.0*tau
      endif

!-----------------------------------------------------------------------
!     Add new A-vector columns to psi
!-----------------------------------------------------------------------
      call xvxxzzo (tau,avec,adim)
      call addvxxzo (avec,psi,adim)

!-----------------------------------------------------------------------
!     Re-normalize in the case of Improved-Relaxation
!-----------------------------------------------------------------------
      if (rlaxnum.ge.0.and.lconst) then
         if ((npacket.eq.1).and.(npackts.eq.1)) then
            call normvxz(psi,norm,adim)
!            write(ilog,'(a,f11.8)') ' Norm (A): ',norm
            call xvixdzo(norm,psi,adim)
         endif
      endif

      return
      end subroutine newavechartree 

!***********************************************************************
! SCALEAVECHARTREE
!
!***********************************************************************

      subroutine scaleavechartree(m,s,psi)

      implicit none

!     Handling of wavefunction
      integer(long) ::   m,s
      integer(long) ::    e
!     Wavefunction
      complex(dop) :: psi(vdim(m,s),dim(m,s),ndim(m,s))

!     Set i>1,j>1,k>1 (othogonality requirement)
      do e=1,dim(m,s)
         psi(1,e,1)=0.0
      enddo

      return
      end subroutine scaleavechartree

!***********************************************************************
! DMATRIZEN3
!
! For Hamiltonian terms involving only a single state,
! calculates the necessary terms in order to define the new A-vector
! coefficients. Only for the special case of a Hartree product.
!
! INPUT
! k         : number of correlated Hamiltonian term
! s         : electronic state of interest
! psi       : coefficients for the relevant state
! hteil     : matrix-elements <phi(i)|h|phi(j)>
!
! block,dim,vdim,ndim are standard arrays (see documentation)
!
! OUTPUT
! avec      : new A-vector colums
!
!***********************************************************************
      subroutine dmatrizen3(k,s,psi,hteil,avec)

      implicit none

!     INPUT
!     Wavefunction
      complex(dop) :: psi(adim)
!     Handling of wavefunction
      integer(long) ::    m,s
      integer(long) ::    k,b
      integer(long) ::    zeig1,zeig2
!     Matrix-elements <phi(i)|h|phi(j)>
      complex(dop) :: hteil(pmatdim)

!     INTERMEDIATE
      complex(dop) :: workc(2*maxblock),hteil2(maxspf)

!     OUTPUT
      complex(dop) :: avec(block(s))

! --- INITIALISE VARIABLES ---------------------------------------------
! workc is divided into two parts of length block(s).
! The result of a multiplication is stored after zeig2.
! Before multiplication, the vector to be multiplied is moved to after
! zeig1.
!-----------------------------------------------------------------------
      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))
!      call initvxz(workc(zeig2),block(s))

!-----------------------------------------------------------------------
!  Loop over modes.
!-----------------------------------------------------------------------
      do m=1,nmode
         if (.not.diag(m,k)) then
            call cpvxz(hteil(pmat(m,k)),hteil2,dim(m,s))
            call swapsxxi(zeig1,zeig2)
!            call qtxxzz(hteil(pmat(m,k)),workc(zeig1),
!     +           workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
!            call mtxxzz (hteil2,workc(zeig1),
!     +           workc(zeig2),vdim(m,s),1,ndim(m,s),dim(m,s))
            call mtxxzz (hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),1,ndim(m,s),dim(m,s))
         endif
      enddo

!-----------------------------------------------------------------------
!     Now calculate the summation for X
!-----------------------------------------------------------------------
      do b=1,block(s)
         avec(b)=avec(b)+workc(zeig2+b-1)
      enddo

      return
      end subroutine dmatrizen3

!***********************************************************************
! NEWAVEC2ND
!
! This routine computes the second order correction when computing
! optimal A-vector coefficients.
!
! INPUT
! m         : mode of interest
! s         : electronic state of interest
! psi       : coefficients for the relevant state
! nspf      : number of new unoccupied SPFs
! sumeigen  :
! eigenavec : eigenvalue associated with new SPFs
! nwspfs    : new SPF for this mode (one state)
! denmat    : density matrix
!
! OUTPUT
! ypsi      :
!
!***********************************************************************

      subroutine newavec2nd (s,m,psi,ypsi,nwspfs,psiadd,npsiadd, &
                 nspf,eigenavec,sumeigen,denmat1,normold)

      implicit none

!     INPUT
!     Wavefunction
      complex(dop) :: psi(dgldim),ypsi(adim)
!     Storage arrays
      complex(dop) :: mc(mcdim)
      real(dop) ::     mr(mrdim)
      integer(long) ::    mi(midim)
      logical    ml(mldim)
!     Handling of wavefunction
      integer(long) ::    s,m
!     Handling of operator terms
      integer(long) ::    k,k1,k2
      integer(long) ::    nham,modus,pworkc
      logical    lwork
!     Computed properties
      real(dop) ::     norm,normold
      complex(dop) :: sprod
!     A-vector and optimal SPFs
      real(dop) ::     eigenavec(imaxspf),sumeigen
      complex(dop) :: nwspfs(sdim*imaxspf)
!     Density matrix
      complex(dop) :: denmat1(dmatdim)
!     Other
      integer(long) ::    nspf
      integer(long) ::    npsiadd

!     OUTPUT
      complex(dop) :: psiadd(vdim(m,s),dim(m,s),ndim(m,s))

!     DEBUGGING
      logical(kind=4), parameter :: ldebug=.false.

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(ypsi,adim)

!-----------------------------------------------------------------------
!     Initial set up
!-----------------------------------------------------------------------
!     Workspace
      lwork=lconst
      if (lwork) then
         pworkc=mcintpsi
      else
         pworkc=mcworkc
      endif

!     Hamiltonian
      nham=1
      modus=0

!     Correlated terms
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

!-----------------------------------------------------------------------
!     Compute A-vector second order correction
!-----------------------------------------------------------------------
!     To sort out later: generalize to multi-set
      do k=k1,k2
         call dmatrizen4(k,m,s,psi,ypsi,psiadd,npsiadd,hteil,nspf,&
              nwspfs,sumeigen,denmat1)
      enddo

!-----------------------------------------------------------------------
!     Debugging
!-----------------------------------------------------------------------
      if (ldebug) then
         call normvxz(ypsi,norm,vdim(m,s)*dim(m,s)*ndim(m,s))
         call vvaxzz (psi,ypsi,sprod,vdim(m,s)*dim(m,s)*ndim(m,s))
         write(6,'(a,E12.6,a,E12.6)') &
         ' norm (Y^2): ',norm**2,' 4*SumEigen^2: ',4*SumEigen**2
         write(6,'(a,E12.6)') ' Scalar prod. (A*Y):',abs(sprod)
!         write(6,'(a,"(",E12.6,",",E12.6,")"))') &
!         ' Scalar prod. (A*Y): ',Sprod
!
         call normvxz(psi,normold,vdim(m,s)*dim(m,s)*ndim(m,s))
         write(6,'(a,f11.8)') ' norm (A): ',normold
      endif

      return
      end subroutine newavec2nd

!***********************************************************************
! DMATRIZEN4
!
! This routine computes the second order correction for when computing
! optimal A-vector coefficients.
!
! INPUT
! k         : number of correlated Hamiltonian term
! s         : electronic state of interest
! m         : mode of interest
! psi       : coefficients for the relevant state
! hteil     : matrix-elements <phi(i)|h|phi(j)>
!
! block,dim,vdim,ndim are standard arrays (see documentation)
!
! OUTPUT
! ypsi      : new A-vector colums
!
!***********************************************************************

      subroutine dmatrizen4(k,mode,s,psi,ypsi,psiadd,npsiadd,hteil,nspf, &
                 nwspfs,sumeigen,dicht1)

      implicit none

!     INPUT
!     Wavefunction
      complex(dop) :: psi(dgldim)
!     Storage arrays
!     Handling of modes, A-vector and operator terms
      integer(long) ::    m,s
      integer(long) ::    v,e,n
      integer(long) ::    mode
      integer(long) ::    zeig1,zeig2,k
!     Optimal SPFs
      integer(long) ::    nspf
      complex(dop) :: nwspfs(nspf*subdim(mode))
!     A-vector
      integer(long) ::    npsiadd
      complex(dop) :: psiadd(vdim(mode,s),nspf,ndim(mode,s))
!     Matrix-elements <phi(i)|h|phi(j)>
      complex(dop) :: hteil(pmatdim)
!     Density matrix
      complex(dop) :: dicht1(dmatdim)

!     INTERMEDIATE
      complex(dop) :: xpsi(vdim(mode,s),dim(mode,s),ndim(mode,s))
      complex(dop) :: hpsi(subdim(mode),dim(mode,s)),hteil2(pmatdim)
      complex(dop) :: workc(3*vdim(mode,s)*dim(mode,s)*ndim(mode,s))
      complex(dop) :: I(vdim(mode,s)*dim(mode,s)*ndim(mode,s))
      complex(dop) :: I2(vdim(mode,s)*dim(mode,s)*ndim(mode,s))

!     OUTPUT
      complex(dop) :: ypsi(adim)

!     DEBUGGING
      real(dop) ::     norm,sumeigen

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
!      if (diag(mode,k)) return

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      m=mode
      call zerovxz(workc,3*vdim(m,s)*dim(m,s)*ndim(m,s))
      call zerovxz(hteil2,pmatdim)
      call zerovxz(hpsi,subdim(mode)*dim(m,s))

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
      do v=1,vdim(m,s)
         do e=1,dim(m,s)
            do n=1,ndim(m,s)
               xpsi(v,e,n)=psiadd(v,npsiadd,n)
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
!     INITIALISE VARIABLES
!     workc is divided into two parts of length block.
!     The result of a multiplication is stored after zeig2.
!     Before multiplication, the vector to be multiplied is moved to after
!     zeig1.
!-----------------------------------------------------------------------
      zeig1 = vdim(m,s)*dim(m,s)*ndim(m,s)+1
      zeig2 = 1
      call cpvxz(xpsi,workc(zeig2),vdim(m,s)*dim(m,s)*ndim(m,s))

!     Debugging
!      call normvxz(workc(zeig2),norm,vdim(m,s)*dim(m,s)*ndim(m,s))
!      write(6,'(a,E12.6,a,E12.6)')
!     +' norm (X^2): ',norm**2,' SumEigen: ',SumEigen

!-----------------------------------------------------------------------
!     Loop over modes.
!     First set initial pointers, then multiply wavefunction
!     with single-particle hamiltonian matrices corresponding to
!     dimensions numbers lower than mode
!-----------------------------------------------------------------------
      do m=1,mode-1
         if (.not.diag(m,k)) then
            call swapsxxi(zeig1,zeig2)
            call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         endif
      enddo

!-----------------------------------------------------------------------
!     Multiply array with single-particle hamiltonian matrices for
!     modes higher than m.
!     If H is unitoperator for mode m, or basis is complete, do not bother
!-----------------------------------------------------------------------
      do m=mode+1,nmode
         if (.not.diag(m,k)) then
            call swapsxxi(zeig1,zeig2)
            call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         endif
      enddo

!-----------------------------------------------------------------------
!     Now calculate the vector Y (see notes Eq. 98).
!     Calculation of hloch for final mode has left workc(zeig2) with the
!     product of psi(block) and hteil matrices for all, excluding the final
!     mode.
!-----------------------------------------------------------------------
      m=mode
      if (.not.diag(m,k)) then
!        Compute elements <phi(j)|h|phi(z)>
!        j: occuppied, z: unoccupied
         call hphi1mk(nwspfs,hpsi,m,k,1,subdim(m))
!         call hphi1mk(Nwspfs,hpsi,m,k,nspf,subdim(m),
!     +        mc(mchc),mr(mrhr),mi(mihi),ml(mlhl))
         call mmaxzz(psi(zetf(m,s)),hpsi,hteil2,subdim(m),dim(m,s),dim(m,s))
!        Eq. 98 in Dieter's notes
         call swapsxxi(zeig1,zeig2)
         call mtxxzz (hteil2,workc(zeig1),workc(zeig2), &
              vdim(m,s),dim(m,s),ndim(m,s),dim(m,s))
!        Density
!        To sort out later: this could be moved outside the loop?
!         if (.not.linitorb) then
         call swapsxxi(zeig1,zeig2)
         call qttxzz (dicht1(dmat(m,s)),psi(zpsi(s)),I,vdim(m,s),dim(m,s),ndim(m,s))
         call rmmxxxzz (psi(zpsi(s)),workc(zeig1),I2,vdim(m,s),dim(m,s),ndim(m,s),dim(m,s))
         call mttxzz (I2,I,workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s),dim(m,s))
!         endif

!-----------------------------------------------------------------------
!        Acumulate Y-vector
!-----------------------------------------------------------------------
         call addvxxzo (workc(zeig2),ypsi(zpsi(s)),vdim(m,s)*dim(m,s)*ndim(m,s))

      endif

      return
      end subroutine dmatrizen4

end module dcoremod
