!***********************************************************************
!                            MODULE DOPT
!
! MCTDH module containing utility tools for the dynamical wavefunction
! code (i.e. keywords initorb and dynspf).
!
! Contains subroutines: akscale, avecxtau, diagnostic1, diagnostic2,
! dyndmat, ReSizePsiNmodes, FitSizePsiNmodes, FitSizePsiX, NumAvec,
! zeigspwn, ReadIOps, ml_tape2dim, wrt_eigenvec
!
! Author: David Mendive-Tapia
! Year  : 2019
!
!***********************************************************************

      module doptmod

use global
use globmemmod
use maxv
use dglobal
use datenmod
use channels
use compdatmod
use dvrdatmod
use griddatmod
use psidef
use hpsimod
use logdat
use runpropmod
use iopsidef, only: psidat
use compdat
use psiutil, only: akcopy,akadd
use genphi1, only: inventspfs
use hphimod
use calchamod, only: calcha
use densitymod, only: density
use iorst, only: wrrst
use projectmod

use op1lib, only: cpvxz,cpmxz,zerovxz,cpmxi,trqxz,normvxz
use op2lib, only: addmxxzo,addvxxzo
use xvlib, only: vvaxzz,xqxxzzo,xvxxdzo,xvxxddo,xvxxzzo,xvxxdza
use mmlib, only: mmxczz,mmaxzz,qmxxzz
use mtlib, only: mttxzz
use rmlib, only: rmmxxxzz

      implicit none

      public

contains

!-----------------------------------------------------------------------
! AKSCALE
!
! This routine scales the "A-vector" by a real number "epsilon".
!
!-----------------------------------------------------------------------
      subroutine akscale(psi,vdim,dim,ndim,dim0,epsilon)

      implicit none

!     A-vector pointers
      integer(long) ::    vdim,dim,ndim,dim0
!     Computed quantities
      real(dop) ::     norm,norm1
      complex(dop) :: ctmp
!     Handling of A-vector
      integer(long) ::    v,e,n
!     Input
      real(dop) ::     epsilon
!     Output
      complex(dop) :: psi(vdim,dim,ndim)

!     Loop over SPF's associated elements
      do e=1+dim0,dim
! To sort out later: this is needed for NumAvec. Do in-there.
!        Set first element to zero (othogonality requirement)
         psi(1,e,1)=0.0
!        Compute norm
         norm=0.0d0
         do v=1,vdim
            do n=1,ndim
               ctmp=dconjg(psi(v,e,n))*psi(v,e,n)
               norm=norm+dble(ctmp)
            enddo
         enddo
         norm=sqrt(norm)
!        Scaling
         norm1=norm/sqrt(epsilon)
!        Re-normalize
         do n=1,ndim
            do v=1,vdim
               psi(v,e,n)=psi(v,e,n)/norm1
            enddo
         enddo
      enddo

      return
      end subroutine akscale

!-----------------------------------------------------------------------
! AVECXTAU
!
! This routine scales the "A-vector" by a real number "tau".
! The parameter tau can have several definitions (See Eq. X).
! Input options:
! ioptavec
! 0      : scaling by each individual eigenvalue(s)
! 1      : scaling by the total sum of eigenvalues
! lcavec
! .true. : tau is complex (propagation and imaginary time relaxation)
! .false.: tau is real (Improved relaxation)
!
!-----------------------------------------------------------------------
      subroutine avecxtau(psiadd,tau,EigenAvec,betavar,SumEigen,&
                 vdim,nspf,ndim,imaxspf,ioptavec,lcavec)

      implicit none


!     A-vector pointers
      integer(long) ::    vdim,nspf,ndim
!     Handling of A-vector
      integer(long) ::    v,e,n
!     Input options
      integer(long) ::    ioptavec
      logical(kind=4) ::    lcavec
!     Input parameters
      real(dop) ::     betavar,SumEigen
      integer(long) ::    imaxspf
      real(dop) ::     EigenAvec(maxspf)
!     Output
      real(dop) ::     tau
      complex(dop) :: cfactor
      complex(dop) :: psiadd(vdim,nspf,ndim)
!     Parameters and constants
      complex(dop), parameter :: imagn = (0.0d0,1.0d0)

!     Define complex factor for tau
      if (lcavec) then
!        Complex tau (regular case)
         cfactor=-imagn
      else
!        Real tau (Improved relaxation)
         cfactor=1.0
      endif

!     Define tau according to the options available
      if (ioptavec.eq.0) then
!        Definition of tau: individual eigenvalue(s)
         do e=1,nspf
!           Abort if eigenvalue is numerically too small
!             if (EigenAvec(e).le.1.0d-16) then
!              To sort out later: how small eigenavec should be allowed?
!              Current printing is 1.0d-16
!                routine='avecxtau'
!                message='WARNING: EigenAvec=0'
!                call errormsg
!             endif
!             tau=dsqrt(betavar/EigenAvec(e))
            tau=dsqrt(betavar/SumEigen)
            do v=1,vdim
               do n=1,ndim
                  psiadd(v,e,n)=psiadd(v,e,n)*cfactor*tau
               enddo
            enddo
         enddo
      else if (ioptavec.le.6) then
!        Scale X vector with tau
         cfactor=cfactor*tau
         call xvxxzzo (cfactor,psiadd,vdim*nspf*ndim)
      endif

      return
      end subroutine avecxtau

      subroutine avecxtau2(tau,m,s,EigenAvec,SumEigen,EigenTrace)

      implicit none

!     Input parameters
      integer(long) ::    m, s
      real(dop) ::     SumEigen
      real(dop) ::     EigenAvec(maxspf),EigenTrace
!     Output
      real(dop) ::     tau

!     Define tau according to the options available
      if (ioptavec.le.6) then
         if (ioptavec.eq.1) then
!           Definition of tau: sum of eigenvalues
            tau=dsqrt(betavar/SumEigen)
         else if (ioptavec.eq.2) then
!           Definition of tau: largest eigenvalue
            tau=dsqrt(betavar/maxval(EigenAvec))
         else if (ioptavec.eq.3) then
!           Definition of tau: value read from input
            tau=betavar
         else if (ioptavec.eq.4) then
!           Definition of tau: automatic selection
            tau=tauthr/dsqrt(maxval(EigenAvec))
         else if (ioptavec.eq.5) then
!           Definition of tau: automatic selection
            tau=tauthr/dsqrt(EigenTrace)
         else if (ioptavec.eq.6) then
!           Definition of tau: automatic selection + factor
            tau=tauthr/dsqrt(EigenTrace)
            tau=tau*betavar
         endif
      else
         routine='optspfs'
         message='unrecognized value for ioptavec'
         call errormsg
      endif

!     Output message
      tau=min(0.5*out1,tau)
      write(ilog,'(a,i3,a,i3,a,E12.6)') ' m =',m,', s =',s,', Tau: ',tau

      return
      end subroutine avecxtau2

      subroutine avecytau(workc,ypsi,tau,mspwn,sspwn,SumEigen,EigenAvec)

      implicit none

!     Wavefunction(s)
      complex(dop) :: workc(workcdim)
!       complex(dop) :: ypsi(adim)
      integer(long), intent(in) ::    mspwn,sspwn
      complex(dop) :: ypsi(vdim(mspwn,sspwn),dim(mspwn,sspwn), &
                 ndim(mspwn,sspwn))
!     A-vector and optimal SPFs
      real(dop) ::     EigenAvec(imaxspf),SumEigen
!
      integer(long) ::    m,s
      integer(long) ::    v,e,n
      real(dop) ::     factor,norm,tau
!     Parameter
      logical(kind=4), parameter :: ldebug=.false.

!-----------------------------------------------------------------------
!        Clean arrays
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!     Scaling of Y with tau
!-----------------------------------------------------------------------
      m=mspwn
      s=sspwn
      if (ioptavec.eq.0) then
         do e=1,dim(m,s)
!             factor=(-0.5*(betavar/EigenAvec(e)))
            factor=(-0.5*(betavar/SumEigen))
            do v=1,vdim(m,s)
               do n=1,ndim(m,s)
                  ypsi(v,e,n)=ypsi(v,e,n)*factor
               enddo
            enddo
         enddo
      else if (ioptavec.le.6) then
         factor=-0.5*tau**2
         call xvxxdzo (factor,ypsi,vdim(m,s)*dim(m,s)*ndim(m,s))
      endif

!-----------------------------------------------------------------------
!     Add vector Y to psi
!-----------------------------------------------------------------------
      call addvxxzo (ypsi,workc,vdim(m,s)*dim(m,s)*ndim(m,s))
      call normvxz(workc,norm,vdim(m,s)*dim(m,s)*ndim(m,s))
      if (ldebug) write(6,'(a,f11.8)') ' norm (A+Y): ',norm

      return
      end subroutine avecytau

!-----------------------------------------------------------------------
! DIAGNOSTIC1
!
! This routine checks Eq.43 in Dieter's notes for one mode and state.
! The result is printed in ilog file.
!
!-----------------------------------------------------------------------
      subroutine diagnostic1 (psi,fmat,m,s,maxblock2)

      implicit none

!     Computed properties
      complex(dop) :: ovl(sdim*maxdim)
      real(dop) ::     sumeq43
!     Handling of wavefunction I
      integer(long) ::    m,s
!     Matrix F
      integer(long) ::    maxblock2
      complex(dop) :: fmat(maxblock2)
!     Other
      integer(long) ::    i
!     Wavefunction
      complex(dop) :: psi(dgldim)

!     Clean arrays
      call zerovxz(ovl,dim(m,s)*subdim(m))
      sumeq43=0.0
!     Compute overlap
      call rmmxxxzz(psi(zpsi(s)),fmat,ovl,vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
      do i=1,dim(m,s)*subdim(m)
         sumeq43=sumeq43+abs(ovl(i))
      enddo
!     Print result
      write(ilog,'(a,i3,a,i3,a,E11.5)') ' m: ',m,' s: ',s,' sum: ',sumeq43

      return
      end subroutine diagnostic1

!-----------------------------------------------------------------------
! DIAGNOSTIC2
!
! This routine checks the overlap of the A-vector with itself.
! The result is printed on the screen.
!
!-----------------------------------------------------------------------
      subroutine diagnostic2(psi,vdim,dim,ndim)

      implicit none

!     A-vector pointers
      integer(long) ::    vdim,dim,ndim
!     Computed properties
      complex(dop) :: ovl(dim,dim)
!     Wavefunction
      complex(dop) :: psi(vdim,dim,ndim)
!     Other
      integer(long) ::    i,j

      call rmmxxxzz(psi,psi,ovl,vdim,dim,ndim,dim)

      write(6,*) 'Overlap:'
      do i=1,dim
         write(6,'(6(" (",E12.6,",",E12.6,")"))') (ovl(i,j), j=1,dim)
      enddo

      return
      end subroutine diagnostic2

!-----------------------------------------------------------------------
! DIAGNOSTIC3
!
!-----------------------------------------------------------------------
      subroutine diagnostic3(psi,MatF,vdim,dim,ndim,subdim)

      implicit none

!     A-vector pointers
      integer(long) ::    vdim,dim,ndim
!     Grid pointers
      integer(long) ::    subdim
!     Arrays
      complex(dop) :: MatF2(dim,dim),MatF3(dim,dim),MatD(dim,dim)
      complex(dop) :: MatF(vdim,subdim,ndim)
!     Wavefunction
      complex(dop) :: psi(dim*subdim)
!     Other
      integer(long) ::    i,j

      call mttxzz (psi,MatF,MatF2,vdim,subdim,ndim,dim)
!
      write(iopteigen,'(a)') 'Matrix F (SPF basis):'
      do i=1,dim
         write(iopteigen,'(9999(3x,f10.6,f10.6))') (MatF2(i,j), j=1,dim)
      enddo
!     See Eq. 34 in notes document "Optimal initial orbitals"
!     dconjg(F(i,j,k))*F(i,l,k) = D(j,l)
      call cpvxz(MatF2,MatF3,dim*dim)
      call mmxczz (MatF2,MatF3,MatD,dim,dim,dim)
!
      write(iopteigen,'(a)') 'Matrix D (SPF basis):'
      do i=1,dim
         write(iopteigen,'(9999(3x,f10.6,f10.6))') (MatD(i,j), j=1,dim)
      enddo
      write(iopteigen,'(a)')

      return
      end subroutine diagnostic3

!-----------------------------------------------------------------------
! DIAGNOSTIC4
!
!-----------------------------------------------------------------------
      subroutine diagnostic4(psi,MatD,dim,subdim)

      implicit none

!     Pointers
      integer(long) ::    dim,subdim
!     Arrays
      complex(dop) :: MatD(subdim,subdim),MatD2(dim,dim)
      complex(dop) :: MatI(subdim,dim)
!     Wavefunction
      complex(dop) :: psi(subdim*dim)
!     Other
      integer(long) ::    i,j

      call zerovxz(MatD2,dim*dim)

      call qmxxzz (MatD,psi,MatI,subdim,dim)
      call mmaxzz (MatI,psi,MatD2,subdim,dim,dim)
!
      write(iopteigen,'(a)') 'Matrix D (SPF basis):'
      do i=1,dim
         write(iopteigen,'(9999(3x,f10.6,f10.6))') (MatD2(i,j), j=1,dim)
      enddo
      write(iopteigen,'(a)')

      return
      end subroutine diagnostic4

! To sort out later: call density directly
      subroutine dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)

      implicit none

      integer(long) ::    dgldim
      complex(dop) :: psi(dgldim)
      integer(long) ::    d3matdim,dmatdim,workrdim
      real(dop) ::     denmat3(d3matdim),workr(workrdim)
      complex(dop) :: denmat1(dmatdim),denmat2(dmatdim),denmat4(dmatdim)

      call density(psi,tree,denmat1,denmat2,denmat3,denmat4,mi(mijindx),gs2)

      return
      end subroutine dyndmat

!-----------------------------------------------------------------------
! FITSIZEPSI
!
! This routine expands or contracts a MCTDH wavefunction.
! New A-vector coefficients are set to zero.
! New SPFs are generated with the standard method (e.g. call inventspfs).
!
!
! save set(1) is new WF, save set(2) is old WF
!-----------------------------------------------------------------------
      subroutine FitSizePsi(psi,ort)

      implicit none

!     Wavefunction(s)
      complex(dop), intent(inout)             :: psi(dglmax)
      complex(dop), dimension(:), allocatable :: oldpsi,workc
!     Handling of wavefunction I
      integer(long) ::    m,s
!     Handling of wavefunction II
      integer(long) ::    mblock,dim2,zeig,zeig1,zeig2
!     For subroutine inventspfs
      real(dop) ::     ort(ortdim)
!     Computed properties
      real(dop) ::     norm,norm1

!-----------------------------------------------------------------------
!     Copy old psi into oldpsi
!
! pointers are set for old psi
!-----------------------------------------------------------------------
      call getcsys(2)
      call getcpsi(2)
      call psidat

      allocate(oldpsi(dgldim))

      call cpvxz(psi,oldpsi,dgldim)
      psi(:)=0.0_dop

!-----------------------------------------------------------------------
!     To copy old coefficients to new wavefunction, it is neccesary to
!     correct for any difference in the number of SPFs.
!-----------------------------------------------------------------------
      do s=1,nstate
         allocate(workc(2*cblock(s,1)))   ! cblock(s,1) ..ge. cblock(s,2)

         zeig=1
         zeig1=cblock(s,1)+1
         call cpvxz(oldpsi(zpsi(s)),workc(zeig),block(s))

         do m=1,nmode
            call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1), &
                        cdim(m,s,1),dim(m,s),ndim(m,s))
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo
         call cpvxz(workc(zeig),psi(czpsi(s,1)),cblock(s,1))

         deallocate(workc)

!-----------------------------------------------------------------------
!        Write norm in log file
!-----------------------------------------------------------------------
         write(ilog,'(a/)') '***** Spawning (inventspf) *****'
         call normvxz(oldpsi(zpsi(s)),norm,block(s))
         call normvxz(psi(czpsi(s,1)),norm1,cblock(s,1))
         if (nstate .eq. 1) then
            write(ilog,'(a,f14.9)') 'Old norm : ',norm
            write(ilog,'(a,f14.9)') 'new norm : ',norm1
         else
            write(ilog,'(a,i3,a,f14.9)') 'Old norm, state',s,' : ',norm
            write(ilog,'(a,i3,a,f14.9)') 'New norm, state',s,' : ',norm1
         endif
      enddo

!-----------------------------------------------------------------------
!     Copy single particle functions to relevant part of psi array,
!     discarding or generating higher phi functions as required
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if(cidmode(m,1) .eq. 0) then ! Do nothing for identical modes
               if (cdim(m,s,1) .gt. dim(m,s)) then
                  ! copy all SPFs that we have ...
                  call cpmxz(oldpsi(zetf(m,s)),psi(czetf(m,s,1)),subdim(m),dim(m,s))
                  ! ... and generate more
                  call inventspfs(m,psi(czetf(m,s,1)),dim(m,s), &
                       cdim(m,s,1),subdim(m),ort,mc(mcexphin),mc(mcexprueck))
               else if (cdim(m,s,1) .le. dim(m,s)) then
                     ! copy as many SPFs as possible, discard the rest
                     call cpmxz(oldpsi(zetf(m,s)),psi(czetf(m,s,1)), &
                                subdim(m),cdim(m,s,1))
               endif     ! cdim .le. dim
            endif     ! idmode
         enddo        ! m=1,nmode
      enddo           ! s=1,nstate

!-----------------------------------------------------------------------
!     Restore data for new psi
!-----------------------------------------------------------------------
      call getcsys(1)
      call getcpsi(1)
      call psidat

      return
      end subroutine FitSizePsi

!-----------------------------------------------------------------------
! RESIZEPSINMODES
!
! This routine expands or contracts a MCTDH wavefunction.
! New A-vector coefficients are sent as INPUT argument (NwAs).
! New SPFs are sent as INPUT argument (Nwspfs).
!-----------------------------------------------------------------------
      subroutine ReSizePsiNmodes(psi,NwAs,Nwspfs,dim0,dim1,maxdimak)

      implicit none

!     Wavefunction(s)
      complex(dop) :: psi(dgldim),workc(workcdim)
!     A-vector and optimal SPFs
      integer(long) ::    maxdimak
      complex(dop) :: NwAs(maxdimak*nmode)
      complex(dop) :: Nwspfs(sdim*imaxspf*nmode)
!     Handling of wavefunction I
      integer(long) ::    m,s
!     Handling of wavefunction II
      integer(long) ::    dim0(maxdim,maxsta),dim1(maxdim,maxsta)
!     Handling of wavefunction III
      integer(long) ::    mblock,dim2,zeig,zeig1,zeig2
!     Computed properties
      real(dop) ::     norm,norm1

!-----------------------------------------------------------------------
!     Clean output arrays
!-----------------------------------------------------------------------
      call zerovxz(workc,dgldim)

!-----------------------------------------------------------------------
!     Define pointers for final psi
!-----------------------------------------------------------------------
      call cpmxi(dim1,dim,maxdim,maxsta)
      call psidat
      call savecsys(1)
      call savecpsi(1)

!-----------------------------------------------------------------------
!     Define pointers for initial psi
!-----------------------------------------------------------------------
!     Copy dimold(m,s) into dim(m,s)
      call cpmxi(dim0,dim,maxdim,maxsta)
!     Update pointers
      call psidat
!     Copy old psi into workc
      call cpvxz(psi,workc,dgldim)

!-----------------------------------------------------------------------
!     To copy old coefficients to new wavefunction, it is neccesary to
!     correct for any difference in the number of SPFs.
!-----------------------------------------------------------------------
      do s=1,nstate
         mblock=1
         do m=1,nmode
            dim2=max(dim(m,s),cdim(m,s,1))
            mblock=mblock*dim2
         enddo
         zeig=dgldim+1
         zeig1=zeig+mblock
         call cpvxz(workc(zpsi(s)),workc(zeig),block(s))

         do m=1,nmode
!             call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1),
!      +                  cdim(m,s,1),dim(m,s),ndim(m,s))
            if (dim(m,s).lt.cdim(m,s,1)) then
               call akadd(workc(zeig),workc(zeig1),NwAs(1+(m-1)*maxdimak),&
                    cvdim(m,s,1),cdim(m,s,1),dim(m,s),ndim(m,s))
!              Compute overlaps for debugging
!               call diagnostic2(workc(zeig1),cvdim(m,s,1),
!     +              cdim(m,s,1),cndim(m,s,1))
            else
               call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1), &
                    cdim(m,s,1),dim(m,s),ndim(m,s))
            endif
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo
         call cpvxz(workc(zeig),psi(czpsi(s,1)),cblock(s,1))

!-----------------------------------------------------------------------
!        normalise new wavefunction to be same as old
!-----------------------------------------------------------------------
         if(.not.lreflex) then
            call normvxz(workc(zpsi(s)),norm,block(s))
            call normvxz(psi(czpsi(s,1)),norm1,cblock(s,1))
            if (nstate .eq. 1) then
               write(ilog,'(a,f14.9,",    squared:",f14.9)') &
                  'Old norm (before adapting SPFs): ',norm, norm**2
               write(ilog,'(a,f14.9,",    squared:",f14.9)') &
                  'new norm (before normalisation): ',norm1, norm1**2
            else
               write(ilog,'(a,i3,a,f14.9)') &
               'Old norm                       , state',s,' : ',norm
               write(ilog,'(a,i3,a,f14.9)') &
               'New norm (before normalisation), state',s,' : ',norm1
            endif
!             if (abs(norm-0.0d0) .gt. tiny) then
!                norm1=norm1/norm
!                call xvixdzo(norm1,psi(czpsi(s,1)),cblock(s,1))
!             endif
         endif
      enddo

!-----------------------------------------------------------------------
!     copy single particle functions to relevant part of psi array,
!     discarding or generating higher phi functions as required
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if(cidmode(m,1) .eq. 0) then ! Do nothing for identical modes
               if (cdim(m,s,1) .gt. dim(m,s)) then
                  ! copy all SPFs that we have ...
                  call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)), &
                             subdim(m),dim(m,s))
                  ! ... and generate more
                  call cpvxz(Nwspfs(1+(m-1)*sdim*imaxspf), &
                       psi(czetf(m,s,1)+dim(m,s)*subdim(m)), &
                       subdim(m)*(cdim(m,s,1)-dim(m,s)))
               else if (cdim(m,s,1) .le. dim(m,s)) then
                  ! copy as many SPFs as possible, discard the rest
                  call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)), &
                             subdim(m),cdim(m,s,1))
               endif     ! cdim .le. dim
            endif     ! idmode
         enddo        ! m=1,nmode
      enddo           ! s=1,nstate

!-----------------------------------------------------------------------
!     restore data for new psi
!-----------------------------------------------------------------------
      call getcsys(1)
      call getcpsi(1)
      call psidat
!       maxblock=0
!       call maxmatwk(0)

      return
      end subroutine ReSizePsiNmodes

!-----------------------------------------------------------------------
! FITSIZEPSINMODES
!
! This routine expands or contracts a MCTDH wavefunction.
! New A-vector coefficients are sent as INPUT.
! New SPFs are sent as INPUT.
!-----------------------------------------------------------------------
      subroutine FitSizePsiNmodes(psi,NwAs,Nwspfs,maxdimak)

      implicit none

!     Wavefunction(s)
      complex(dop) :: psi(dgldim),workc(workcdim)
!     A-vector and optimal SPFs
      integer(long) ::    maxdimak
      complex(dop) :: NwAs(maxdimak*nmode)
      complex(dop) :: Nwspfs(sdim*imaxspf*nmode)
!     Handling of wavefunction I
      integer(long) ::    m,s
!     Handling of wavefunction II
      integer(long) ::    mblock,dim2,zeig,zeig1,zeig2
!     Input
      integer(long) ::    mspwn,sspwn
!     Computed properties
      real(dop) ::     norm,norm1

!-----------------------------------------------------------------------
!     Save data for new psi
!-----------------------------------------------------------------------
      call psidat
      call savecsys(1)
      call savecpsi(1)

!-----------------------------------------------------------------------
!     Read old psi into workc and copy olddim(m,s)
!-----------------------------------------------------------------------
!     Copy old psi into workc
      call cpvxz(psi,workc,dgldim)
!     Copy dimold(m,s) into dim(m,s)
      call cpmxi(dimold,dim,maxdim,maxsta)
!     Update pointers
      call psidat

!-----------------------------------------------------------------------
!     Copy A-vector to relevant part of psi array
!-----------------------------------------------------------------------
!     To copy old coefficients to new wavefunction, it is neccesary to
!     correct for any difference in the number of SPFs.
      do s=1,nstate
         mblock=1
         do m=1,nmode
            dim2=max(dim(m,s),cdim(m,s,1))
            mblock=mblock*dim2
         enddo
         zeig=dgldim+1
         zeig1=zeig+mblock
         call cpvxz(workc(zpsi(s)),workc(zeig),block(s))

         do m=1,nmode
            if (dim(m,s).lt.cdim(m,s,1)) then
               call akadd(workc(zeig),workc(zeig1),NwAs(1+(m-1)*maxdimak), &
                    cvdim(m,s,1),cdim(m,s,1),dim(m,s),ndim(m,s))
!              Compute overlaps for debugging
!               call diagnostic2(workc(zeig1),cvdim(m,s,1),
!     +              cdim(m,s,1),cndim(m,s,1))
            else
               call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1), &
                    cdim(m,s,1),dim(m,s),ndim(m,s))
            endif
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo
         call cpvxz(workc(zeig),psi(czpsi(s,1)),cblock(s,1))

! To sort out later: check this is correct
!-----------------------------------------------------------------------
!        Write norm in log file
!-----------------------------------------------------------------------
!          write(ilog,'(a/)') '***** Spawning (optimalspf) *****'
!          call normvxz(workc(zpsi(s)),norm,block(s))
!          call normvxz(psi(czpsi(s,1)),norm1,cblock(s,1))
!          if (nstate .eq. 1) then
!             write(ilog,'(a,f14.9)')
!      +         'Old norm : ',norm
!             write(ilog,'(a,f14.9)')
!      +         'new norm : ',norm1
!          else
!             write(ilog,'(a,i3,a,f14.9)')
!      +      'Old norm, state',s,' : ',norm
!             write(ilog,'(a,i3,a,f14.9)')
!      +      'New norm, state',s,' : ',norm1
!          endif
      enddo

!-----------------------------------------------------------------------
!     Copy single particle functions to relevant part of psi array,
!     discarding or generating higher phi functions as required
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if(cidmode(m,1) .eq. 0) then ! Do nothing for identical modes
               if (cdim(m,s,1) .gt. dim(m,s)) then
                  ! copy all SPFs that we have ...
                  call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)),subdim(m),dim(m,s))
                  ! ... and generate more
                     call cpvxz(Nwspfs(1+(m-1)*sdim*imaxspf), &
                          psi(czetf(m,s,1)+dim(m,s)*subdim(m)), &
                          subdim(m)*(cdim(m,s,1)-dim(m,s)))
               else if (cdim(m,s,1) .le. dim(m,s)) then
!                  if (m.ne.meb) then
                     ! copy as many SPFs as possible, discard the rest
                     call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)), &
                                subdim(m),cdim(m,s,1))
!                  endif
               endif     ! cdim .le. dim
            endif     ! idmode
         enddo        ! m=1,nmode
      enddo           ! s=1,nstate

!-----------------------------------------------------------------------
!     Restore data for new psi
!-----------------------------------------------------------------------
      call getcsys(1)
      call getcpsi(1)
      call psidat

      return
      end subroutine FitSizePsiNmodes

!-----------------------------------------------------------------------
! FITSIZEPSIX
!
! This routine expands or contracts a MCTDH wavefunction for one mode
! and state.
! New A-vector coefficients are sent as INPUT.
! New SPFs are sent as INPUT.
!-----------------------------------------------------------------------
      subroutine FitSizePsiX(psi,NwAs,Nwspfs,mspwn,sspwn)

      implicit none

!     Wavefunction(s)
      complex(dop) :: psi(dgldim),workc(workcdim)
!     A-vector and optimal SPFs
      complex(dop) :: NwAs(adim)
      complex(dop) :: Nwspfs(sdim*imaxspf)
!     Handling of wavefunction I
      integer(long) ::    mspwn,sspwn
!     Handling of wavefunction II
      integer(long) ::    mblock,dim2,zeig,zeig1,zeig2
!     Input
      integer(long) ::    m,s
!     Computed properties
!       real(dop) ::     norm,norm1

!-----------------------------------------------------------------------
!     Save data for new psi
!-----------------------------------------------------------------------
      call psidat
      call savecsys(1)
      call savecpsi(1)

!-----------------------------------------------------------------------
!     Read old psi into workc and copy olddim(m,s)
!-----------------------------------------------------------------------
!     Copy old psi into workc
      call cpvxz(psi,workc,dgldim)
!     Copy dimold(m,s) into dim(m,s)
      call cpmxi(dimold,dim,maxdim,maxsta)
!     Update pointers
      call psidat

!-----------------------------------------------------------------------
!     Copy A-vector to relevant part of psi array
!-----------------------------------------------------------------------
      m=mspwn
      s=sspwn
      if (loptavec) then

         if (.not.lmult) then
            call akadd(workc,psi,NwAs,cvdim(m,s,1),cdim(m,s,1),dim(m,s),ndim(m,s))
!           Compute overlaps for debugging
!            call diagnostic2(psi,cvdim(m,s,1),cdim(m,s,1),cndim(m,s,1))
         else
!-----------------------------------------------------------------------
!           To copy old coefficients to new wavefunction, it is neccesary to
!           correct for any difference in the number of SPFs.
!-----------------------------------------------------------------------
            do s=1,nstate
               mblock=1
               do m=1,nmode
                  dim2=max(dim(m,s),cdim(m,s,1))
                  mblock=mblock*dim2
               enddo
               zeig=dgldim+1
               zeig1=zeig+mblock
               call cpvxz(workc(zpsi(s)),workc(zeig),block(s))

               do m=1,nmode
                  if ((m.eq.mspwn).and.(s.eq.sspwn)) then
                     call akadd(workc(zeig),workc(zeig1),NwAs, &
                          cvdim(m,s,1),cdim(m,s,1),dim(m,s),ndim(m,s))
!                    Compute overlaps for debugging
!                     call diagnostic2(workc(zeig1),cvdim(m,s,1),
!     +                    cdim(m,s,1),cndim(m,s,1))
                  else
                     call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1), &
                          cdim(m,s,1),dim(m,s),ndim(m,s))
                  endif
                  zeig2=zeig
                  zeig=zeig1
                  zeig1=zeig2
               enddo
               call cpvxz(workc(zeig),psi(czpsi(s,1)),cblock(s,1))
            enddo
         endif

      else

         if (.not.lmult) then
            call akcopy(workc,psi,cvdim(m,s,1),cdim(m,s,1),dim(m,s),ndim(m,s))
         else
!-----------------------------------------------------------------------
!           To copy old coefficients to new wavefunction, it is neccesary to
!           correct for any difference in the number of SPFs.
!-----------------------------------------------------------------------
            do s=1,nstate
               mblock=1
               do m=1,nmode
                  dim2=max(dim(m,s),cdim(m,s,1))
                  mblock=mblock*dim2
               enddo
               zeig=dgldim+1
               zeig1=zeig+mblock
               call cpvxz(workc(zpsi(s)),workc(zeig),block(s))

               do m=1,nmode
                  call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,1), &
                              cdim(m,s,1),dim(m,s),ndim(m,s))
                  zeig2=zeig
                  zeig=zeig1
                  zeig1=zeig2
               enddo
               call cpvxz(workc(zeig),psi(czpsi(s,1)),cblock(s,1))
            enddo
         endif

      endif

!-----------------------------------------------------------------------
!     Copy single particle functions to relevant part of psi array,
!     discarding or generating higher phi functions as required
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if(cidmode(m,1) .eq. 0) then ! Do nothing for identical modes
               if (cdim(m,s,1) .gt. dim(m,s)) then
                  ! copy all SPFs that we have ...
                  call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)),subdim(m),dim(m,s))
                  ! ... and generate more
                     call cpvxz(Nwspfs,psi(czetf(m,s,1)+dim(m,s)*subdim(m)), &
                          subdim(m)*(cdim(m,s,1)-dim(m,s)))
               else if (cdim(m,s,1) .le. dim(m,s)) then
!                  if (m.ne.meb) then
                     ! copy as many SPFs as possible, discard the rest
                     call cpmxz(workc(zetf(m,s)),psi(czetf(m,s,1)),subdim(m),&
                          cdim(m,s,1))
!                  endif
               endif     ! cdim .le. dim
            endif     ! idmode
         enddo        ! m=1,nmode
      enddo           ! s=1,nstate

!-----------------------------------------------------------------------
!     Restore data for new psi
!-----------------------------------------------------------------------
      call getcsys(1)
      call getcpsi(1)
      call psidat

      return
      end subroutine FitSizePsiX

!-----------------------------------------------------------------------
! NUMAVEC
!
!-----------------------------------------------------------------------
      subroutine NumAvec(psi,dtpsi,epsilon)

      implicit none

!     Wavefunction
      complex(dop) :: psi(dgldim),dtpsi(dgldim),Kryc(ikryc),HKryc(ikryc)
      real(dop)    :: Krylov(ikrylov),HKrylov(ikrylov)
!     Output
      complex(dop) :: psiadd(adim)
!     Backup
      complex(dop) :: psi2(dgldim),dtpsi2(dgldim)
      real(dop) ::     cp_epsmat,cp_out1,cp_tfinal,cp_tinit,cp_out2
      real(dop) ::     cp_updatetime,cp_tcurrent
!     Density matrix
      real(dop) ::     workr(workrdim)
      complex(dop) :: denmat(maxspf,maxspf)
!
      integer(long) ::    v,e,n
      integer(long) ::    dimold2(nmode,nstate)
!     Computed properties
      real(dop) ::     norm,norm1
!     Writing information to restart file
      complex(dop) :: cdum(1)
      real(dop) ::     rdum(1)
      integer(long) ::    idum
      logical(kind=4) ::    linwf
!     Handling of wavefunction
      integer(long) ::    m,s
      integer(long) ::    i,j
      integer(long) ::    ncount
!     Parameters
      complex(dop), parameter :: imagn = (1.0d0,0.0d0)
!     Debugging
      logical(kind=4), parameter :: ldebug = .false.
!      parameter  (ldebug = .true.)

      real(dop) ::     epsilon
      integer(long) ::    adimold,nham
      logical(kind=4) ::    lrhoepsone,lbackprop,lscaleavec

!     Backup epsmat
      cp_epsmat=epsmat
!     Backup psi
      call cpvxz(psi,psi2,dgldim)
      call cpvxz(dtpsi,dtpsi2,dgldim)
!
      istp=iopteigen
!       iout=iopteigen
      lpsi=.false.

!     Cases:
      if (lconst) then
!        CMF: backward/forward propagation
         if (linitorb) then
!           Like VMF
            lrhoepsone=.true.
            lbackprop=.false.
            lscaleavec=.true.
         else
!           Read A-vector from mc array
            lrhoepsone=.false.
            lbackprop=.true.
            lscaleavec=.false.
!             lscaleavec=.true.
!             epsilon=0.0
         endif
      else
!        VMF: forward propagation with density regularized with 1.0
         lrhoepsone=.true.
         lbackprop=.false.
         lscaleavec=.true.
      endif
!     Definitions above:
!     ldenepsone => compute density with epsmat=1.0
!     lbackprop  => backward propagation
!     lscaleavec => scale a-vector

      if (lrhoepsone) then
!        Compute density with epsilon=1
         epsmat=1.0
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
!        Print out density matrix
!          if (ldebug) then
!             write(6,'(a)') 'Density matrix (epsilon=1.0):'
!             ncount=1
!             do s=1,nstate
!                do m=1,nmode
!                  call cpvxz(mc(mcdicht1+ncount-1),denmat,dim(m,s)*dim(m,s))
!                   write(6,*) 'mode:',m,'state:',s
!                   do i=1,dim(m,s)
!                      write(6,*) (denmat(i,j), j=1,dim(m,s))
!                      ncount=ncount+dim(m,s)*dim(m,s)
!                   enddo
!                enddo
!             enddo
!          endif
      endif

!     Backup settings for propagation
      cp_out1=out1
      cp_out2=out2
      cp_tfinal=tfinal
      cp_tinit=tinit
      cp_updatetime=updatetime
      cp_tcurrent=tcurrent

!     Setting for infinitesimal propagation
      if (lbackprop) then
         call cpvxz(mc(mcpredpsi),psi,adim)
! --- CALCULATE RIGHT HAND SIDE FOR HAMILTONIAN ---
         nham=1
         call calcha(psi,dtpsi,hteil,mr(mrhops), &
              mc(mcphi),nham,mc(mcphi),mi(mijindx),mr(mrort))
         out1=0.01d0*updatetime
         tfinal=tinit+out1
         out2=tfinal
         updatetime=out1
         write(6,*) 'tout:',out1/fs,tfinal/fs
      else
         out1=1.0d-03*fs
         tinit=0.0
         tfinal=tinit+out1
         out2=tfinal
         updatetime=out1
         write(6,*) 'tout:',out1/fs,tfinal/fs
      endif

!     Propagation forward
      call runprop
!     Save numerical A-vector
      call cpvxz(psi,psiadd,adim)

!     Set occupied part to zero
      if (linitorb) then
         psiadd(1)=0.0
!         call xvxxdzr (oner,psi2,psiadd,adim)
      else
         adimold=1
         do s=1,nstate
            do m=1,nmode
              adimold=adimold*dimold(m,s)
            enddo
         enddo
         write(6,*) 'adimold:',adimold,adim
         call zerovxz(psiadd,adimold)
      endif

!     Normalize numerical A-vector according to epsilon
!     To sort out later: dimold2 should be removed. Only one case.
      if (lscaleavec) then
         do s=1,nstate
            do m=1,nmode
              dimold2(m,s)=1
            enddo
         enddo
         do s=1,nstate
            do m=1,nmode
               if (linitorb) then
               call akscale(psiadd,vdim(m,s),dim(m,s),ndim(m,s),&
                    dimold2(m,s),epsilon)
               else
               call akscale(psiadd,vdim(m,s),dim(m,s),ndim(m,s),&
                    dimold(m,s),epsilon)
               endif
            enddo
         enddo

!        Recover psi
         call cpvxz(psi2,psi,dgldim)
         call cpvxz(dtpsi2,dtpsi,dgldim)
!        Add new columns
         call xvxxdza (oner,psiadd,psi2,adim)
         call cpvxz(psi2,psi,adim)
      endif

!       if (lbackprop) then
!          call cpvxz(psi2,psi,adim)
!          call xvxxdza (oner,psiadd,psi,adim)
!          do s=1,nstate
!             do m=1,nmode
!                call cpvxz(psi2(zetf(m,s)),psi(zetf(m,s)),subdim(m))
!             enddo
!          enddo
!       endif

!     Recover settings for normal propagation
      out1=cp_out1
      out2=cp_out2
      tfinal=cp_tfinal
      tinit=cp_tinit
      updatetime=cp_updatetime
      if (.not.linitorb) tinit=cp_tcurrent

!     Compute density with standard epsilon
      if (lrhoepsone) then
         epsmat=cp_epsmat
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
!          write(6,*) 'DENSITY:'
!          do i=1,dmatdim
!             write(6,*) mc(mcdicht1+i-1)
!          enddo
      endif

!-----------------------------------------------------------------------
!     Write wavefunction to restart file
!-----------------------------------------------------------------------
      linwf=.true.
      idum=1
      rdum(1)=1.d0
      cdum(1)=(1.d0,1.d0)
      call wrrst(tcurrent,psi,linwf,cdum,idum,rdum,rdum,mi(mijindx),&
           agmat,mi(mitrajst))

!
      iout=13
      istp=14
      lpsi=.true.
      lnewspf=.true. !So integrator is always initialized

      return
      end subroutine NumAvec

!-----------------------------------------------------------------------
      subroutine zeigspwn(maxblock3,maxdimak,workcdim2)
!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------

      implicit none

      integer(long) ::    m,s
      integer(long) ::    zeig
      integer(long) ::    maxblock3,maxdimak,workcdim2

!-----------------------------------------------------------------------
      routine   = 'Zeigspwn'
      maxblock3 = 0
      maxdimak  = 0
!-----------------------------------------------------------------------
!     totnewspf: total number of new spfs
!     maxdimak : largest A-vector column to be added
!     maxblock3: the largest vector required when operating on psi,
!                e.g. in matrizen1b, matrizen2b
!-----------------------------------------------------------------------
!      do s=1,nstate
!         workcdim=max(workcdim,2*dgldim+cblock(s,1))
!      enddo
!
! cdim(m,s,1) = new psi ;  cdim(m,s,2) = old psi
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if(cidmode(m,1) .eq. 0) then ! Do nothing for identical modes
               if (cdim(m,s,1) .gt. cdim(m,s,2)) then
                  maxdimak= cvdim(m,s,1)*cndim(m,s,1)*(cdim(m,s,1)-cdim(m,s,2)) &
                           +maxdimak
                  maxblock=max(maxblock,cblock(s,1))
                  maxblock3=max(maxblock3,maxblock/cdim(m,s,1)*subdim(m))

               endif
            endif
         enddo
!-----------------------------------------------------------------------
         workcdim2=max(workcdim,2*dgldim+block(s))
!-----------------------------------------------------------------------
      enddo

!-----------------------------------------------------------------------
      if (linitorb) then
         maxdimak=0
         do s=1,nstate
            do m=1,nmode
               if (vdim(m,s)*ndim(m,s)*dim(m,s).gt.maxdimak) then
                  maxdimak=vdim(m,s)*ndim(m,s)*dim(m,s)+maxdimak
               endif
               maxblock3=maxblock/dim(m,s)*subdim(m)+maxblock3
            enddo
         enddo
      endif

!      if (lmult) then
!-----------------------------------------------------------------------
!        Determine pointers to hpsi array
!-----------------------------------------------------------------------
!        zeig=1
!        do m = 1,nmode
!           do s=1,nstate
!              zeig=max(zeig,subdim(m)*dim(m,s))
!           enddo
!        enddo
!        hpsidim = max(hpsidim,zeig)
!     endif
!-----------------------------------------------------------------------
!     To sort out later: write this somwhere in log file
!-----------------------------------------------------------------------

      return
      end subroutine zeigspwn

!-----------------------------------------------------------------------
! READIOPS
!
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! WRT_EIGENVEC
!
!-----------------------------------------------------------------------

      subroutine wrt_eigenvec(psi)

      implicit none

!     Input
!     Wavefunction(s)
      complex(dop) ::     psi(dgldim)
!     Intermediary
!     For subroutine rdrst
      character(len=c5) filename
!     Handling of wavefunction I
      integer(long) ::        m,s
!     Other
      integer(long) ::        i,j
      integer(long) ::        ncount

!-----------------------------------------------------------------------
!     Write eigenvectors into spwnevec
!-----------------------------------------------------------------------
      do m=1,nmode
         do s=1,nstate
            if ((m.ne.meb).and.(m.ne.mpb)) then
!              File: spwneval
               write(filename,'(a,i3,a,i2)') '/spwnevec_',m,'_',s
               call rmblnk(filename)
               open(unit   = ioptveig,file   = name(1:laenge)//filename, &
! To sort out later: why we need "append" here?
!      +              position="append",
                    form   = 'formatted',status = 'old')
               write(ioptveig,'(a,f12.5)') ' Time in fs: ', tcurrent/fs
               write(ioptveig,'(a,i3,a,i3,a,i3)') &
                     ' Optimal SPFs (Lanczos):    mode =',m, &
                     ' , state =',s,' , order =',dim(m,s)
               ncount=1
               do i=zetf(m,s),zetf(m,s)+subdim(m)-1
                  write(ioptveig,*) ncount, &
                  (abs(psi(i+(j-1)*subdim(m))), j=1,dim(m,s))
                  ncount=ncount+1
               enddo
            endif
         enddo
      enddo

      return
      end subroutine wrt_eigenvec

      end module doptmod
