!***********************************************************************
!                            MODULE DSPF
!
!  MCTDH module containing the main driver subroutines for the dynamical
!  wavefunction code (i.e. keywords initorb and dynspf).
!
!  Contains subroutines: runspwn, dynpsi, optpsi, optspfs, optspfslancz.
!
!  Author: David Mendive-Tapia
!  Year  : 2019
!
!***********************************************************************

module dspfmod

use global
use globmemmod
use dglobal
use maxv
use griddatmod  
use psidef   
use runinwfmod  
use runpropmod  
use channels
use datenmod
use compdatmod
use timing
use logdat

use phihphimod, only: phihphi
use iopsidef, only: psidat
use zeighpsimod, only: zeighpsi
use compdat
use densitymod, only: density
use schmidtorthomod, only: schmidtortho
use trafonatmod, only: trafonat
use operdef
use hpsimod
use function, only: funkr
use projectmod, only: projectorT
use iorst
use einsbasmod, only: logsbas
use mempropmod
use rdopermod
use function

use doptmod
use dcoremod

use op1lib, only: zerovxd,zerovxz,cpvxz,normvxz,cpmxi,cpvxd,cpqxz,sum2vxz
use xvlib, only: xvixdzo,xvxxddo,xvxxdzo
use mmlib, only: qmxxzz

      implicit none

      public

      contains

!***********************************************************************
! RUNSPWN
!
! This routine is called from CALLSPWN and drives the extension of the
! MCTDH wavefunction on-the-fly (i.e. spawning).
!
! OLD ROUTINE FROM MCTDH85.1315
!
! psi       : MCTDH wavefunction
! maxdimak  :
! maxblock3 : maximum size for
!
! mc,mr,mi,ml are standard arrays (see documentation)
!
! dim(m,s) : No. spfs of current WF
! dimtape(m,s) : No. spfs of new WF to be created here 
!***********************************************************************

      subroutine runspwn(psi,dtpsi)

      implicit none

      logical(kind=4)     :: linwf

      integer(long)                :: idum
      integer(long), dimension(1)  :: idum1,idum2
      complex(dop), dimension(1)   :: cdum
      real(dop), dimension(1)      :: rdum

!     INPUT
!     Wavefunction
      complex(dop), dimension(dglmax) :: psi,dtpsi
!     Storage arrays

!     INTERMEDIATE I
!     Handling of wavefunction I
      integer(long)             :: s,m
      integer(long)             :: mspwn,sspwn,nmodespwn
      integer(long)             :: maxblock3,maxdimak
      integer(long)             :: maxblock2,totnewspf
!     Handling of wavefunction II
      integer(long)   :: i
      integer(long)   :: p,pp,ierr
!     Parameters controling algorithm
      logical(kind=4) :: lcavec
!     Other
      integer(long), dimension(:,:), allocatable :: dimold2

!     INTERMEDIATE II
!     Optimal SPFs
      integer(long) ::    nspf
      real(dop) ::     sumeigen,eigenavec(maxspf)
      complex(dop), dimension(:), allocatable :: nwas
      complex(dop), dimension(:), allocatable :: nwspfs
      complex(dop), dimension(:), allocatable :: fmat
!     Density matrix for the computation of fmatpp
      complex(dop), dimension(:), allocatable :: denmat1
!     Storage (subroutines density and trafonat)
      real(dop) ::     workr(workrdim)

!     INTERMEDIATE III
!     Computed properties
      real(dop) ::     norm,norm0,norm1

!     Other
      integer(long) ::    nham
      character(len = 52) LineSec

      routine='runspwn'

!-----------------------------------------------------------------------
! Open restart and  oper files
!-----------------------------------------------------------------------
!      filename=name(1:laenge)//'/restart'
!      open(irst,file=filename,form='unformatted',status='old',err=1000)

! Store occupied dimensionalities in dimold(m,s)
      call cpmxi(dim,dimold,maxdim,maxsta)
      call cpmxi(dim,dimocc,maxdim,maxsta)
      call savecsys(2)
      call savecpsi(2)

! Use dimtape(m,s) which contains the new dimensionalities
! Update pointers and allocate new wavefunction
      call cpmxi(dimtape,dim,maxdim,maxsta)
      call psidat
      call savecsys(1)
      call savecpsi(1)

      call zeigspwn(maxblock3,maxdimak,workcdim)

      allocate(dimold2(maxdim,maxsta))
      allocate(nwas(maxdimak*nmode))
      allocate(nwspfs(sdim*maxspf*nmode))
      allocate(fmat(maxblock3))
      allocate(denmat1(dmatdim))

!----------------------------------------------------------------------
!     Relevant info to be used later
!----------------------------------------------------------------------
      totnewspf = 0
      mspwn = 0
      sspwn = 0
      nmodespwn = 0
      dimunocc(:,:)=0
      do s=1,nstate
         do m=1,nmode
            if (dimtape(m,s) .gt. dimold(m,s)) then
               totnewspf=dimtape(m,s)-dimold(m,s)+totnewspf
               mspwn=m
               sspwn=s
               nmodespwn=nmodespwn+1
               dimunocc(m,s) = dimtape(m,s) - dimold(m,s)
            endif
         enddo
      enddo

!----------------------------------------------------------------------
!     Clean arrays
!----------------------------------------------------------------------
       call zerovxz(NwAs,maxdimak*nmode)
       call zerovxz(Nwspfs,sdim*maxspf*nmode)

!----------------------------------------------------------------------
!     Write messages into log and output files
!----------------------------------------------------------------------
      write(ilog,'(/a)') '***** Spawning calculation (i.e. dynspf) *****'
      write(ilog,'(a,i3,a,f10.3)') &
          ' Spawning cycle',idyncyc,' at Time (fs) ;',tcurrent/fs

      write(LineSec,'(a)') &
      '----------------------------------------------------'
      write(iopteigen,'(a,a)') '#-------------------',LineSec
      write(iopteigen,'(a)') '# Computing optimal SPFs'
      write(iopteigen,'(a,a)') '#-------------------',LineSec

!----------------------------------------------------------------------
!     Set defaults
!----------------------------------------------------------------------
      lcavec=.true.
!     Improved relaxation a-vector must be real
      if (rlaxnum.ge.0.and.lconst) lcavec=.false.

!-----------------------------------------------------------------------
!     Schmidt-orthogonalize
!-----------------------------------------------------------------------
      if (lorthopsi) then
         write(ilog,'(a)') 'The initial wavefunction is Schmidt-orthogonalized.'
!        Transform to natural orbitals.
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
         call trafonat(psi,tree,dicht4,mi(mijindx))
      endif
!DMT  Copy density (Size dimold) and store dimold
      call cpvxz(dicht1,denmat1,dmatdim)
      call cpmxi(dimold,dimold2,maxdim,maxsta)
!DMT  Copy density (Size dimold) and store dimold

!----------------------------------------------------------------------
!     Extend the MCTDH wavefunction
! 
! At end, Psi contains new WF and dimensions are for new WF.
!----------------------------------------------------------------------
      if (loptspf) then

!     Restore occupied dimensionalities
         call getcsys(2)
         call getcpsi(2)

!-----------------------------------------------------------------------
!     --- EVALUATE RIGHT HAND SIDE OF ODE ---
!-----------------------------------------------------------------------
         if (lexact) then
            lgetenergy = lrlx
          call funkex(tcurrent,psi(zetf(1,1)),dtpsi(zetf(1,1)))
            lgetenergy = .false.
         else if (lfunkl) then
            call funk(tcurrent,psi,dtpsi)
         else
            call funkr(tcurrent,psi,dtpsi)
         endif

!----------------------------------------------------------------------
!     Initial arrangements
!----------------------------------------------------------------------
!     dimtape(m,s) contains the new dimensionalities,
!     which we put in dim(m,s)
         call cpmxi(dimtape,dim,maxdim,maxsta)
!     Update pointers
         call psidat
         call savecsys(1)
         call savecpsi(1)

!DMT      call cpvxz(mc(mcdicht1),denmat1,dmatdim)

!        Optimal spfs
!        Two cases:
         if (nmodespwn.eq.1) then
!           1) Save data for new psi
            call psidat
            call savecsys(1)
            call savecpsi(1)
!           2) Copy dimold(m,s) into dim(m,s)
            call cpmxi(dimold,dim,maxdim,maxsta)
!           3) Update pointers
            call psidat
!           4) Compute norm before enlargement
            call normvxz(psi,norm0,adim)
            write(ilog,'(a,f11.8)') ' Norm (A): ',norm0
!           4) Only one mode is increased by one spf
            call dynpsi(tcurrent,mspwn,sspwn,psi,dtpsi,fmat,nspf, &
                 SumEigen,EigenAvec,Nwspfs,NwAs,denmat1,lcavec,norm1, &
                 maxblock3,maxblock2,maxdimak)
!           5) Restore data for new psi
            call getcsys(1)
            call getcpsi(1)
            call psidat
!           6) Re-size wavefunction (1 mode)
            call FitSizePsiX(psi,NwAs,Nwspfs,mspwn,sspwn)
!           Debugging
            if (loptavec2nd) then
!              7) Compute norm after enlargement
               call normvxz(psi,norm1,adim)
               write(ilog,'(a,f11.8)') ' Norm (A): ',norm1
               write(ilog,'(a,E12.6)') ' Error: ',(norm1-norm0)
            endif
         else
!           2) More than one mode is increased at once
!            if (iresize.eq.0) then
            if ((iresize.eq.1).or.(iresize.eq.2)) then
!              Default:
!              Case 1: iresize=0
!              All dim(m,s) are dimold(m,s) and we do
!              dim(m,s)=dimtape(m,s) at the very end
!              1) Copy dimold(m,s) into dim(m,s)
               call cpmxi(dimold,dim,maxdim,maxsta)
!              2) Update pointers
               call psidat
!              3) Save data for new psi
               call savecsys(1)
               call savecpsi(1)
!              4) Compute norm before enlargement
               call normvxz(psi,norm0,adim)
               write(ilog,'(a,f11.8)') ' Norm (A): ',norm0
!              Loop
               do s=1,nstate
                  do m=1,nmode
                     if ((m.ne.meb).and.(m.ne.mpb)) then
                        if (dimtape(m,s).gt.dim(m,s)) then
!                          4) Only one mode is increased by one spf
                           call dynpsi(tcurrent,m,s,psi,dtpsi,fmat,nspf,&
                                       SumEigen,EigenAvec,&
                                        Nwspfs(1+(m-1)*sdim*imaxspf),&
                                        NwAs(1+(m-1)*maxdimak),denmat1,lcavec,&
                                norm1,maxblock3,maxblock2,maxdimak)
!                          5) Update density
                           call dyndmat(psi,dgldim,denmat1,dmatdim,&
                                d3matdim,workr,workrdim)
                        endif
                     endif
                  enddo
               enddo
!              6) Re-size wavefunction (all modes)
               call cpmxi(dimtape,dim,maxdim,maxsta)
               call psidat
               call FitSizePsiNmodes(psi,NwAs,Nwspfs,maxdimak)
!              Debugging
               if (loptavec2nd) then
!                 7) Compute norm after enlargement
                  call normvxz(psi,norm1,adim)
                  write(ilog,'(a,f11.8)') ' Norm (A): ',norm1
                  write(ilog,'(a,E12.6)') ' Error: ',(norm1-norm0)
               endif
!            else if ((iresize.eq.1).or.(iresize.eq.2)) then
            else if (iresize.eq.0) then
!              Case 2: iresize=1
!              All dim(m,s) are dimold(m,s) and we do
!              dim(m,s)=dimtape(m,s) one by one
               call cpmxi(dimold,dim,maxdim,maxsta)
               call psidat
!              4) Compute norm before enlargement
               call normvxz(psi,norm0,adim)
               write(ilog,'(a,f11.8)') ' Norm (A): ',norm0
!              Loop
               do s=1,nstate
                  do m=1,nmode
                     if ((m.ne.meb).and.(m.ne.mpb)) then
                        if (dimtape(m,s).gt.dim(m,s)) then
                           dim(m,s)=dimtape(m,s)
!                          1) Save data for new psi
                           call psidat
                           call savecsys(1)
                           call savecpsi(1)
!                          2) Copy dimold(m,s) into dim(m,s)
                           call cpmxi(dimold,dim,maxdim,maxsta)
                           call cpmxi(dimold,dimocc,maxdim,maxsta)
!                          3) Update pointers
                           call psidat
!                          4) Only one mode is increased by one spf
                           call dynpsi(tcurrent,m,s,psi,dtpsi,fmat,nspf,&
                                      SumEigen,EigenAvec,&
                                      Nwspfs(1+(m-1)*sdim*imaxspf),&
                                      NwAs(1+(m-1)*maxdimak),denmat1,lcavec,&
                                      norm1,maxblock3,maxblock2,maxdimak)
!                          5) Restore data for new psi
                           call getcsys(1)
                           call getcpsi(1)
                           call psidat
!                          6) Re-size wavefunction (1 mode)
                           call FitSizePsiX(psi,NwAs(1+(m-1)*maxdimak),&
                                          Nwspfs(1+(m-1)*sdim*imaxspf),&
                                          m,s)
!                          7) Update density
                           dimold(m,s)=dimtape(m,s)
                           call dyndmat(psi,dgldim,denmat1,dmatdim,&
                                        d3matdim,workr,workrdim)
                           if (loptnspf) then
!                              mfields2 needs mc(mchteil)
                               nham=1
                               call zeighpsi(nham)
                               call phihphi(psi,gh,mc(mchpsi), &
                                            hteil,1,usehsym)
                           endif
                        endif
                     endif
                  enddo
               enddo
!              Debugging
               if (loptavec2nd) then
!                 7) Compute norm after enlargement
                  call normvxz(psi,norm1,adim)
                  write(ilog,'(a,f11.8)') ' Norm (A): ',norm1
                  write(ilog,'(a,E12.6)') ' Error: ',(norm1-norm0)
               endif
            else
               routine='runspwn'
!               message = 'unrecognized value for iresize=',iresize
               write(6,*) 'unrecognized value for iresize=',iresize
               call errormsg
            endif
         endif
!        Re-normalize in the case of Improved-Relaxation
         if (rlaxnum.ge.0.and.lconst) then
            if ((npacket.eq.1).and.(npackts.eq.1)) then
               call normvxz(psi,norm1,adim)
               call xvixdzo(norm1,psi,adim)
            endif
         endif

      else
!        Basic case (invent spf)
         call FitSizePsi(psi,mr(mrort))

      endif

!-----------------------------------------------------------------------
!     Schmidt-orthogonalize initial SPFs if desired
!-----------------------------------------------------------------------
      if (lorthopsi) then
         do s=1,nstate
            do m=1,nmode
               if (idmode(m).eq.0) &
                   call schmidtortho(psi(zetf(m,s)),subdim(m),dim(m,s),ierr)
               if (ierr .gt. 0 ) then
                  routine='Getpsi (getpsi.F line 951)'
                  write(message,'(a,i3,a,i3,a,i3)') &
                       'Cannot Schmidt-orthogonalize SPFs for s =',s, &
                       ', m =',m,', e =',ierr
                  call errormsg
               end if
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
!     Normalize A-vectors in case of reflex
!-----------------------------------------------------------------------
      if(lreflex) then
         do p=1,npackts*2
            pp = (p-1)*adim/(2*npackts) + 1
            call normvxz(psi(pp),norm,adim/(2*npackts)) ! compute norm
            if(norm.le.1.d-6) then
               routine='Getpsi (getpsi.F line 969)'
               write(message,'(a,i2,a,e11.4)') &
                  'The norm of the initial WF (pack=',p, &
                  ') is too small. norm =',norm
               call errormsg
            endif
            call xvixdzo(norm,psi(pp),adim/(2*npackts)) ! normalise
            write(ilog,'(2a,i2,a,f14.9)') 'Norm of initial WF ', &
               '(pack=',p,') before re-normalisation:', norm
         enddo
      endif

!DMT  Recover dimold
      call cpmxi(dimold2,dimold,maxdim,maxsta)
!DMT  Recover dimold

      deallocate(dimold2,nwas,nwspfs,fmat,denmat1)

      return
      end subroutine runspwn

!***********************************************************************
!
! New coeffcients (SPFs and A-vector) are computed and added to the
! MCTDH wavefunction
!
! INPUT
! time      : current time
! s         : electronic state of interest
! m         : primitive mode of interest
! psi       : MCTDH wavefunction
! fmat      : mean-fields-like matrix F
! denmat    : density matrix
!
! maxblock3 :
! maxblock2 : maximum size for fmat
! maxdimak  :
!
! mc,mr,mi,ml are standard arrays (see documentation)
!
! OUTPUT
! nspf      : number of new unoccupied SPFs
! sumeigen  :
! trace     : trace of the eigenvalues associated with new SPFs
! eigenavec : eigenvalue associated with new SPFs
! nwspfs    : new SPF for this mode (one state)
! psiadd    :
!
!***********************************************************************

      subroutine dynpsi(time,m,s,psi,dtpsi,fmat,nspf,sumeigen,eigenavec, &
                 nwspfs,psiadd,denmat1,lcavec,normold, &
                 maxblock3,maxblock2,maxdimak)

      implicit none

!     INPUT
!     Wavefunction
      real(dop) ::     time
      complex(dop) :: psi(dgldim),dtpsi(dgldim)
!     Handling of wavefunction
      integer(long) ::    s,m,e
!     Parameters controling algorithm
      logical(kind=4) ::    lcavec
!     Density matrix for the computation of Fpp
      complex(dop) :: denmat1(dmatdim)
!     Second order correction
      integer(long) ::    largernspfdim
      real(dop) ::     normold,SumEigen
      complex(dop) :: xpsi(workcdim),ypsi(workcdim)
!      complex(dop) ::  xpsi(adim),ypsi(adim)

!     INTERMEDIATE
      integer(long) ::    maxblock2,maxblock3
      real(dop) ::     EigenTrace,tau
      real(dop) ::     EigenAvec(maxspf)
      complex(dop) :: fmat(maxblock3)
      complex(dop) :: D(sdim*sdim)

!     OUTPUT
      integer(long) ::    nspf,npsiadd
      integer(long) ::    maxdimak
      complex(dop) :: NwAs(maxdimak)
      complex(dop) :: psiadd(maxdimak)
      complex(dop) :: Nwspfs(sdim*maxspf)

      routine='dynpsi'

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(xpsi,workcdim)
      call zerovxz(ypsi,workcdim)
      call zerovxz(psiadd,maxdimak)

!-----------------------------------------------------------------------
!     Compute optimal SPFs
!-----------------------------------------------------------------------
      call optpsi(time,s,m,psi,dtpsi,D,fmat,denmat1,nspf,&
           Nwspfs,EigenAvec,EigenTrace, &
           maxblock3,maxdimak,maxblock2,imaxspf)

!-----------------------------------------------------------------------
! Re-dimensionalize according to the Eigenvalues of D
!-----------------------------------------------------------------------
      if (loptnspf) then
         dimtape(m,s)=dim(m,s)+nspf
         dim(m,s)=dim(m,s)+nspf
         call psidat
         call savecsys(1)
         call savecpsi(1)
         dim(m,s)=dimold(m,s)
         call psidat
      endif

!-----------------------------------------------------------------------
!     Compute optimal A-vector
!-----------------------------------------------------------------------
      if (loptavec) then
         largernspfdim=max(nspf,dim(m,s))
         call newavec (m,s,fmat,Nwspfs,eigenavec,psiadd,xpsi,maxblock2, &
              dimtape(m,s)-dim(m,s),vdim(m,s),ndim(m,s),largernspfdim)

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
         SumEigen=0.0
         do e=1,nspf
            SumEigen=SumEigen+EigenAvec(e)
         enddo

!-----------------------------------------------------------------------
!     Define the parameter tau
!-----------------------------------------------------------------------
         call avecxtau2(tau,m,s,eigenavec,sumeigen,eigentrace)

!-----------------------------------------------------------------------
!     Second-order correction
!-----------------------------------------------------------------------
         if (loptavec2nd) then
!           Compute A-vector columns one-by-one
            do npsiadd=1,nspf
               call newavec2nd (s,m,psi,ypsi,Nwspfs(1+(npsiadd-1)*subdim(m)), &
                    psiadd,npsiadd,nspf,eigenavec,SumEigen,denmat1,normold)
!              Scaling of Y with tau
               if (loptavec2nd) then
                  call avecytau(psi,ypsi,tau,m,s,SumEigen,EigenAvec)
               endif
            enddo
         endif

!-----------------------------------------------------------------------
!     Scaling of Y with tau
!-----------------------------------------------------------------------
!         call avecytau(psi,ypsi,m,s,SumEigen,EigenAvec)

!-----------------------------------------------------------------------
!     Scaling of X with tau
!-----------------------------------------------------------------------
   !     imag*sqrt(epsilon/lambda)*Avec(i) = Avec(i)
         call avecxtau(psiadd,tau,EigenAvec,betavar,SumEigen, &
                       vdim(m,s),nspf,ndim(m,s),imaxspf,ioptavec,lcavec)
      endif

      return
      end subroutine dynpsi

!***********************************************************************
! OPTPSI
!
! Driver for FUNKSPAWN and OPTSPFS. Contains output printing and change
! of sign trick.
!
! INPUT
! time      : current time
! s         : electronic state of interest
! m         : primitive mode of interest
! psi       : MCTDH wavefunction
! dltmat    : optimal unoccupied SPFs matrix Delta
! fmat      : mean-fields-like matrix F
! denmat    : density matrix
!
! maxnspf   :
! maxdimak  :
! maxblock2 : maximum size for fmat
! maxblock3 :
!
! mc,mr,mi,ml are standard arrays (see documentation)
!
! OUTPUT
! nspf      : number of new unoccupied SPFs
! trace     : trace of the eigenvalues associated with new SPFs
! eigenavec : eigenvalue associated with new SPFs
! nwspfs    : new SPF for this mode (one state)
!
!***********************************************************************

      subroutine optpsi(time,s,m,psi,dtpsi,dltmat,fmat,denmat1,nspf, &
                 nwspfs,eigenavec,trace,maxblock3,maxdimak,maxblock2,maxnspf)

      implicit none

!     INPUT
!     Wavefunction
      real(dop) ::     time
      complex(dop) :: psi(dgldim),dtpsi(dgldim)
!     Handling of wavefunction
      integer(long) ::    m,s,i
!     Parameters controling algorithm
      integer(long) ::    maxit
!     Density matrix for the computation of Fpp
      complex(dop) :: denmat1(dmatdim)

!     INTERMEDIATE
      integer(long) :: maxblock2,maxblock3,maxdimak
      complex(dop) :: fmat(maxblock3)
      complex(dop) :: dltmat(sdim*sdim)

!     OUTPUT
      integer(long), intent(in)  :: maxnspf
      integer(long), intent(out) :: nspf
      real(dop), intent(out) ::   eigenAvec(maxspf)
      real(dop), intent(out) ::   trace
      complex(dop) :: nwspfs(sdim*maxspf)

      routine='optpsi'

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      eigenavec = 0.0d0
      nwspfs = 0.0d0

!-----------------------------------------------------------------------
!     Optimal orbitals
!-----------------------------------------------------------------------
!     To sort out later: this could be in a subroutine with other checks
!     Resize maxblock as maxblock2
      maxblock2=maxblock/dim(m,s)*subdim(m)
      if (maxblock2.gt.maxblock3) then
         routine='optpsi'
         message='WARNING: maxblock2 > maxblock3'
         call errormsg
      else
         write(ilog,'(a,2i8,a,f5.1,a)') ' maxblock2,  maxblock3     :', &
            maxblock2,maxblock3,' (',real(maxblock2)/real(maxblock3)*100,' %)'
      endif
      if (vdim(m,s)*subdim(m)*ndim(m,s).gt.maxblock2) then
         routine='optpsi'
         message='WARNING: vdim(m,s)*subdim(m)*ndim(m,s) > maxblock2'
         call errormsg
      endif

!     Diagonalization
!     First call funkspawn (Generate optimal SPFs)
      call funkspawn(time,m,s,psi,dtpsi,dltmat,fmat,denmat1,maxblock2)

!     Trick: usually after diagonalization one takes the N lowest
!            eigenvalues, but here we care about the largest ones.
!            Thus we change the sign of matrix so after diagonalization
!            the largest eigenvalues come out at the bottom  (e.g. the
!            diagonalization subroutine does not need to be modified).
      if (loptspflancz) then
!        Maximum number of iterations: maxit > dim1+1
         maxit=max(subdim(m),maxnspf+2)
!        Lanczos diagonalization
         call optspfslancz(s,m,dltmat,subdim(m),nspf,trace,eigenavec,&
              nwspfs,maxit,maxnspf)

      else
!        FullDiag
         call xvxxdzo(-oner,dltmat,subdim(m)*subdim(m))
         call optspfs(s,m,dltmat,dim(m,s),subdim(m),nspf,trace,eigenavec,&
              nwspfs,maxnspf)
!        Correct the sign of the eigenvalues
!        Due to the sign trick eigenvalues come out with opposite sign,
!        so they need to be corrected
         call xvxxddo(-oner,eigenavec,maxnspf)
         call xvxxdzo(-oner,dltmat,subdim(m)*subdim(m))
      endif

      return
      end subroutine optpsi

!-----------------------------------------------------------------------
! QOPTSPF
!
!-----------------------------------------------------------------------
      subroutine QOptSpf(psi)

      implicit none
!

!     Wavefunction
      complex(dop) :: psi(dgldim),workc(dgldim)
!     Handling of wavefunction
      integer(long) ::    m,s
      integer(long) ::    dim1,sdm
!     Handling of operators
      integer(long) ::    nham
      integer(long) ::    k,k1,k2
!     Quasi-optimal SPFs
!     Action of the correlated Hamiltonian on the SPFs
      complex(dop) :: Dhpsi(sdim*sdim)
      complex(dop) :: Dhpsi2(sdim*sdim)
      complex(dop) :: hpsi(hpsidim)
      complex(dop) :: SumAlpha
      complex(dop) :: Wkavg(zham(1)+kzahl(1)-1)
      complex(dop) :: Wkavgm(zham(1)+kzahl(1)-1,nmode)
!     Diagonalization
      integer(long) ::    nspf2,maxit
      real(dop) ::     PsdEvec(maxspf)
      complex(dop) :: Psdspfs(sdim*maxspf)
      complex(dop) :: PsdspfsM(sdim*maxspf,nmode)
!     Other
      complex(dop) :: wr
      integer(long) ::    cp_imaxspf,cp_ibeta
      real(dop) ::     cp_betathr,trace
!     Other
      character(len = 52) LineSec
!     Debugging
      logical(kind=4), parameter :: ldebug = .false.
!      parameter  (ldebug = .true.)

!-----------------------------------------------------------------------
!     Message header
!-----------------------------------------------------------------------
      write(LineSec,'(a)') &
      '----------------------------------------------------'
      write(iopteigen,'(a,a)') '#-------------------',LineSec
      write(iopteigen,'(a)') '# Computing initial quasi-optimal SPFs'
      write(iopteigen,'(a,a)') '#-------------------',LineSec

!-----------------------------------------------------------------------
!     Save wavefunction
!-----------------------------------------------------------------------
      call savecsys(1)
      call savecpsi(1)
      call cpvxz(psi,workc,dgldim)

!-----------------------------------------------------------------------
!     Compute Wkavg
!-----------------------------------------------------------------------
      call zerovxz(Dhpsi,sdim*sdim)
      call zerovxz(Dhpsi2,sdim*sdim)
      call zerovxz(Wkavg,zham(1)+kzahl(1)-1)
      call zerovxz(Wkavgm,(zham(1)+kzahl(1)-1)*nmode)

!     Summation over grid (correlated terms)
!     j_\kappa = 1
      nham=1
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      do k=k1,k2
         if (ioptavec.eq.1) then
            Wkavg(k)=1.0*dconjg(koeff(k))*koeff(k)
         else
            Wkavg(k)=1.0
         endif
      enddo

      do k=k1,k2
         do s=1,nstate
            do m=1,nmode
!          if ((m.ne.meb).and.(m.ne.mpb)) then
               dim1=1
               sdm=subdim(m)
               if (.not.diag(m,k)) then
                  call hphi1mk(psi(zetf(m,s)),hpsi,m,k,dim1,sdm)
                  call vvaxzz (hpsi,hpsi,SumAlpha,sdm)
!
                  Wkavgm(k,m)=SumAlpha
                  Wkavg(k)=Wkavg(k)*SumAlpha
!
!                 DEBUGGING
!                   if (ldebug) then
!                      write(iopteigen,*)' m:',m,'k:',k
!                      call mmxczz (hpsi,hpsi,Dhpsi,sdm,1,sdm)
!                    Apply projector
!                      call projectorT (psi(zetf(m,s)),Dhpsi,1,sdm,sdm)
!                      call projector (psi(zetf(m,s)),Dhpsi,1,sdm,sdm)
!                      call diagnostic4(psi(zetf(m,s)),Dhpsi,dim(m,s),
!      +                    subdim(m))
!                   endif
!                 DEBUGGING
               endif
!          endif
            enddo
         enddo
      enddo

!     Product over modes
!       if (mod(iop1,10000)/1000.eq.1) then
!          do k=k1,k2
!             Wkavg(k)=Wkavgm(k,1)
!             do m=2,nmode
!                do s=1,nstate
!                   if (.not.diag(m,k)) Wkavg(k)=Wkavg(k)*Wkavgm(k,m)
!                enddo
!             enddo
!             Wkavg(k)=Wkavg(k)*koeff(k)*koeff(k)
!          enddo
!       endif

!-----------------------------------------------------------------------
!     Compute matrix D
!-----------------------------------------------------------------------
      do m=1,nmode
         do s=1,nstate
            if ((m.eq.meb).and.(m.eq.mpb)) cycle

            if (dimunocc(m,s) .eq. 0) cycle  ! No SPFs to be added

!             if (idmode(m).eq.0 .and. m.ne.mpb) then
!              Clean arrays
            call zerovxz(Dhpsi2,subdim(m)*subdim(m))

!              Wavefunction pointers
            nham=1
            dim1=1
            sdm=subdim(m)
!              Correlated terms pointers
            k1=zham(nham)
            k2=zham(nham)+kzahl(nham)-1
!              Loop over correlated terms
            do k=k1,k2
               if (.not.diag(m,k)) then
                  dim1=1
                  sdm=subdim(m)
                  call hphi1mk(psi(zetf(m,s)),hpsi,m,k,dim1,sdm)
!                      call mmxazz (hpsi,hpsi,Dhpsi,sdm,dim1,sdm)
                  call mmxczz (hpsi,hpsi,Dhpsi,sdm,dim1,sdm)
                  call trqxz (Dhpsi,SumAlpha,sdm)
                  wr=Wkavg(k)/SumAlpha
                  if (ldebug) write(iopteigen,*) 'm:',m,'wr:',wr, &
                  'Wkavg(k):',Wkavg(k),'Tr:',1.0/real(SumAlpha)
                  call xqxxzzo (wr,Dhpsi,sdm)
                  call addmxxzo(Dhpsi,Dhpsi2,sdm,sdm)
               endif
            enddo
!              Debugging
             if (ldebug) call diagnostic4(psi(zetf(m,s)),Dhpsi2, &
                              dim(m,s),subdim(m))
!              Apply projector
             call projectorT (psi(zetf(m,s)),Dhpsi2,1,sdm,sdm)
             call projector (psi(zetf(m,s)),Dhpsi2,1,sdm,sdm)
!              Debugging
             if (ldebug) call diagnostic4(psi(zetf(m,s)),Dhpsi2, &
                              dim(m,s),subdim(m))
!              Diagonalization
             cp_betathr=betathr
             cp_ibeta=ibeta
             cp_imaxspf=imaxspf
!
            if (ioptavec2nd.ne.0) then
               imaxspf=ioptavec2nd
               betathr=0.0
            else if (ioptavec.eq.1) then
               ibeta=1
               betathr=1.0d-09
            else
               ibeta=2
               betathr=1.0d-06
            endif
!              Diagonalization
             if (sdm.gt.10) then
!                 Lanczos
!                 Maximum number of iterations: maxit > dim1+1
                maxit=max(sdm,imaxspf+2)
!                 Lanczos diagonalization
               if (loptnspf) then
                  call optspfslancz(s,m,Dhpsi2,sdm,nspf2,trace,PsdEvec, &
                       PsdspfsM(1,m),maxit,dimunocc(m,s))
               else
                  call optspfslancz(s,m,Dhpsi2,sdm,nspf2,trace,PsdEvec, &
                       Psdspfs,maxit,dimunocc(m,s))
               endif
            else
!              FullDiag
               call xvxxdzo(-oner,Dhpsi2,subdim(m)*subdim(m))
               if (loptnspf) then
                  call optspfs(s,m,Dhpsi2,dim1,subdim(m),nspf2,trace, &
                       PsdEvec,PsdspfsM(1,m),dimunocc(m,s))
               else
                  call optspfs(s,m,Dhpsi2,dim1,subdim(m),nspf2,trace, &
                       PsdEvec,Psdspfs,dimunocc(m,s))
               endif
!              Correct the sign of the eigenvalues
               call xvxxddo(-oner,PsdEvec,dimunocc(m,s))
               call xvxxdzo(-oner,Dhpsi2,subdim(m)*subdim(m))
            endif
!
            betathr=cp_betathr
            ibeta=cp_ibeta
            imaxspf=cp_imaxspf

!           Re-define dim(m,s) sizes
            if (loptnspf) then
!                  dim(m,s)=1+nspf2
!                  if (dim(m,s).gt.maxdimspwn(m,s)) then
!                     routine='QOptSPF'
!                     message='Quasi-SPFs maxdimspwn too small'
!                     call errormsg
!                  endif
               call cpvxz(PsdspfsM(1,m),psi(zetf(m,s)+dimocc(m,s)*sdm), &
                    sdm*(dimunocc(m,s)))
            else
! If keyword "fix" simply copy all quasi-optimal SPFs to unocciped orbs
               call cpvxz(Psdspfs,psi(zetf(m,s)+dimocc(m,s)*sdm), &
                    sdm*(dimunocc(m,s)))
            endif
         enddo
      enddo

      return
      end subroutine QOptSpf

!***********************************************************************
! OPTSPFS
!
! Calls the diagonalisation routine (FullDiag) for the generation of
! optimal SPFs. One state and one mode.
!
! INPUT
! s         : electronic state of interest
! m         : primitive mode of interest
! evek      : matrix to be diagonalized (Delta matrix)
! dim       : number of SPFs already set
! subdim    : number of grid points for mode m
!
! maxnspf   : maximum number of SPFs that can be added
!
! OUTPUT
! nspf      : number of new unoccupied SPFs
! trace     : trace of the eigenvalues associated with new SPFs
! eigen     : eigenvalue associated with new SPFs
! phi       : new SPF for this mode (one state)
!
!***********************************************************************

      subroutine optspfs(s,m,evek,dim,subdim,nspf,trace,eigen,phi,maxnspf)

      implicit none

!     INPUT
!     Matrix to be diagonalized
      integer(long) ::     subdim
      complex(dop) ::  evek(subdim,subdim)
!     For diagonalization (subroutine zheev)
      integer(long) ::     error,e2,dim
      real(dop) ::      eval(subdim,2),work(3*subdim)
      complex(dop) ::  invers(subdim,subdim)
      character(len=4) buf
!     Parameters controling algorithm

!     Handling of wavefunction
      integer(long) ::     s,m,g
!     Other
      integer(long) ::     maxnspf
      real(dop) ::      trace
      complex(dop) ::  cp_evek(subdim,subdim)

!     INTERMEDIATE
      real(dop) ::      beta,tau
      logical(kind=4) ::     laddspf

!     OUTPUT
      integer(long) ::     nspf
      real(dop) ::      eigen(maxnspf)
      complex(dop) ::  phi(subdim*maxnspf)

      integer(long), save                            :: tid=0 ! timer ID

!-----------------------------------------------------------------------
!     Start timer and store evek
!-----------------------------------------------------------------------
      call get_timer(tid, 'tspawndiag')
      call start_timer(tid)

      call cpqxz (evek,cp_evek,subdim)

!-----------------------------------------------------------------------
!     Diagonalize a complex matrix
!     invers       : used as work-space for LAPACK-routine zheev
!     eval         : holds eigenvalues
!     evek (input) : holds on input the matrix to be diagonalised
!     evek (output): holds on output the eigenvectors
!-----------------------------------------------------------------------
      e2 = subdim*subdim
      call zheev('V','U',subdim,evek,subdim,eval,invers,e2,work,error)
      if (error .ne. 0) then
         write(buf,'(i4)') error
         routine = 'matinv'
         message = 'Cannot diagonalize density matrix.'//" Error No.: "//buf
         call errormsg
      endif

!-----------------------------------------------------------------------
!     Print header output
!-----------------------------------------------------------------------
      write(iopteigen,'(a,f12.5)') ' Time in fs: ', tcurrent/fs
      write(iopteigen,'(a,i3,a,i3)') ' Optimal SPFs (FullDiag):   mode =',m, &
                     ' , state =',s
      write(iopteigen,'(a,i2,a,e12.5,a,e12.5)') ' ibeta =',ibeta, &
                     ', betathr =',betathr,', betavar =',betavar

      write(iopteigen,'(a,a,a)') '  spf        eval*tau^2 ', &
                     '             eval[au]       eval(i)/eval(1) ','thr'

!-----------------------------------------------------------------------
!     Definition of tau
!     To sort out later: call here subroutine avecxtau2 and sent tau up
!-----------------------------------------------------------------------
      if (ioptavec.le.1) then
         trace=0.0
         do g=1,subdim
            trace=trace+eval(g,1)
         enddo
         tau=tauthr/dsqrt(abs(trace))
      else if (ioptavec.eq.2) then
         trace=0.0
         do g=1,subdim
            trace=trace+eval(g,1)
         enddo
         tau=tauthr/dsqrt(abs(trace))
      else if (ioptavec.eq.3) then
         tau=betavar
      else if (ioptavec.eq.4) then
         tau=tauthr/dsqrt(abs(eval(1,1)))
      else if (ioptavec.eq.5) then
         trace=0.0
         do g=1,subdim
            trace=trace+eval(g,1)
         enddo
         tau=tauthr/dsqrt(abs(trace))
      else if (ioptavec.eq.6) then
         trace=0.0
         do g=1,subdim
            trace=trace+eval(g,1)
         enddo
         tau=tauthr/dsqrt(abs(trace))
         tau=tau*betavar
      endif

!-----------------------------------------------------------------------
!     Check how many SPFs will be added
!-----------------------------------------------------------------------
      nspf=0
      laddspf=.false.
      do g=1,subdim
!        Definition of beta (e.g. eigenvalues threshold)
         if (ibeta.eq.1) then
            beta=-eval(g,1)*tau**2
         else if (ibeta.eq.2) then
            beta=-eval(g,1)
         else if (ibeta.eq.3) then
            beta=abs(eval(g,1)/eval(1,1))
         endif
!        ErrorCode
         if (ISNAN(beta)) then
            routine='optspfs'
            message = 'Diagonalization failed'
            call errormsg
         endif
!        Check how many SPFs will be added
         if ((beta.gt.betathr).and.(abs(eval(g,1)).gt.1.0d-16).and. &
              (nspf.lt.maxnspf)) then
            laddspf=.true.
            nspf=nspf+1
         endif
!        Write output
         write(iopteigen,'(i5,f18.12,f22.15,f22.15,l4)') &
                g,-eval(g,1)*tau**2,-eval(g,1),abs(eval(g,1)/eval(1,1)),laddspf
         laddspf=.false.
         if(abs(eval(g,1)/eval(1,1)) .lt. 1.d-14) goto 111
      enddo

!-----------------------------------------------------------------------
!     Add a default number of SPFs, when the largest eigenvalue is too
!     small
!-----------------------------------------------------------------------
 111  if (nspf.eq.0) then
         write(iopteigen,'(a)')
         write(iopteigen,'(a)') &
            ' WARNING: none of the eigenvalues is above threshold betathr'
         if (loptnspf) then
!           Add one spf even if nspf=0 (variable number of SPFs)
            nspf=1
            write(iopteigen,'(a)') '          one SPF will be added'
         else
!           Add maxnspf even if nspf=0 (fixed number of SPFs in input)
            write(iopteigen,'(a)')
            write(iopteigen,'(a)') &
            ' WARNING: nspf reduced to maxnspf=',maxnspf
            nspf=maxnspf
         endif
      endif
      write(iopteigen,'(a)')
      call myflush(iopteigen)

!-----------------------------------------------------------------------
!     Save eigenvalues
!-----------------------------------------------------------------------
      call cpvxd(eval,eigen,maxnspf)

!-----------------------------------------------------------------------
!     Copy new SPFs inside phi vector
!-----------------------------------------------------------------------
      call cpvxz(evek,phi,subdim*maxnspf)

!-----------------------------------------------------------------------
!     Stop timer and restore evek
!-----------------------------------------------------------------------
      call cpqxz (cp_evek,evek,subdim)
      call stop_timer(tid)

      return
      end subroutine optspfs

!***********************************************************************
! OPTSPFSLANCZ
!
! Calls the diagonalisation routine (Lanczos) for the generation of
! optimal SPFs. The matrix must be hermitian.
!
! INPUT
! s         : electronic state of interest
! m         : primitive mode of interest
! evek      : matrix to be diagonalized (Delta matrix)
! subdim    : number of grid points for mode m
!
! maxit     : maximum number of Lanczos iterations
! maxnspf   : maximum number of SPFs that can be added
!
! OUTPUT
! nspf      : number of new unoccupied SPFs
! trace     : trace of the eigenvalues associated with new SPFs
! eigen     : eigenvalue associated with new SPFs
! phi       : new SPF for this mode (one state)
!
!***********************************************************************

      subroutine optspfslancz(s,m,evek,subdim,nspf,trace,eigen,phi,maxit,&
                 maxnspf)

      implicit none

!     INPUT
!     Matrix to be diagonalized
      integer(long) :: subdim
      complex(dop)  :: evek(subdim*subdim)
!     For diagonalization (eiglanc)
      integer(long) :: iter,degenerate
      real(dop)     :: offdiag(maxit),wk(2*maxit)
      real(dop)     :: eigenvec(maxit,maxit),eigenval(maxit)
      complex(dop)  :: psi(subdim,maxspf),krylov(subdim,maxit+1)
!     Initial guess
      integer(long) ::    imax
      real(dop) ::     dmax(subdim)
!     Parameters controling algorithm
      integer(long) ::    maxit
!     Handling of psi vector
      integer(long) ::    s,m,g
!     Computed properties
      real(dop) ::     norm
!     Other
      integer(long) ::    maxnspf
      real(dop) ::     trace
      complex(dop) :: evek2(subdim,subdim)

      complex(dop) :: zsum2
      real(dop)    :: sum2

!     INTERMEDIATE
      real(dop) ::     beta,tau
      logical(kind=4) ::    laddspf

!     OUTPUT
      integer(long) ::    nspf
      real(dop) ::     Eigen(maxspf)
      complex(dop) :: phi(subdim*maxspf)

      integer(long), save                            :: tid=0 ! timer ID

! Check if dltmat is non-zero
      call sum2vxz(evek,zsum2,subdim*subdim)
      sum2=abs(zsum2)
      if (sum2 .lt. tiny) then
         write(ilog,'(/,a,i4,a,i4)') &
             'WARNING: F-matrix is zero for mode / state :',m,'/',s
         nspf=0
         return
      endif
      
!-----------------------------------------------------------------------
!     Start timer
!-----------------------------------------------------------------------
      call get_timer(tid, 'tspwnlancz')
      call start_timer(tid)

!-----------------------------------------------------------------------
!     Clean arrays
!-----------------------------------------------------------------------
      call zerovxz(phi,subdim*maxspf)
      call zerovxd(Eigen,maxspf)
      call zerovxd(dmax,subdim)
      call zerovxz(psi,subdim*maxspf)

!-----------------------------------------------------------------------
!     Print header output
!-----------------------------------------------------------------------
!     File: spwneval
      degenerate=0
      write(iopteigen,'(a,f12.5)') ' Time in fs: ', tcurrent/fs
      write(iopteigen,'(a,i3,a,i3,a,i6)') &
            ' Optimal SPFs (Lanczos):    mode =',m, &
            ' , state =',s,' , maxit =',maxit
      write(iopteigen,'(a,i1,a,e12.5,a,e12.5,a,i1)') ' ibeta = ',ibeta, &
                     ', betathr =',betathr,', betavar =',betavar, &
                     ', degenerate = ',degenerate
! Error here with i in eval(i) not defined
!      write(iopteigen,'(a,a,a)') '  spf        eval*tau^2 ', &
!                     '             eval[au]       eval(i)/eval(1) ','thr'

!-----------------------------------------------------------------------
!     Initial guess vector
!-----------------------------------------------------------------------
      call cpvxz(evek,evek2,subdim*subdim)
      do g=1,subdim
         dmax(g)=abs(evek2(g,g))
      enddo
      do g=1,5
         imax=maxloc(dmax,1)
         psi(imax,1)=1.0
         dmax(imax)=0.0
      enddo
      call normvxz(psi(1,1),norm,subdim)
      call xvixdzo(norm,psi(1,1),subdim)

!-----------------------------------------------------------------------
!     Lanczos diagonalization
!-----------------------------------------------------------------------
      call eiglanc(psi,subdim,maxnspf,maxit,krylov,iter,func,eigenvec, &
                   eigenval,offdiag,wk,degenerate,evek,iopteigen)

!     ErrorCode
!      if (error.gt.0) then
!         write(6,*) error
!         routine='optspfslancz'
!         if (error.eq.1) then
!            message = 'illegal integration order'
!         else if (error.eq.2) then
!            message = 'stepsize underflow'
!         else if (error.eq.3) then
!            message = 'diagonalization failed'
!         else if (error.eq.4) then
!            message = 'stepsize too large for restart'
!         endif
!         call errormsg
!      endif

!-----------------------------------------------------------------------
!     Definition of tau
!     To sort out later: call here subroutine avecxtau2 and sent tau up
!-----------------------------------------------------------------------
      if (ioptavec.le.1) then
         trace=0.0
         do g=1,iter
            trace=trace+eigenval(g)
         enddo
         tau=tauthr/dsqrt(trace)
      else if (ioptavec.eq.2) then
         trace=0.0
         do g=1,iter
            trace=trace+eigenval(g)
         enddo
         tau=tauthr/dsqrt(trace)
      else if (ioptavec.eq.3) then
         tau=betavar
      else if (ioptavec.eq.4) then
         tau=tauthr/dsqrt(eigenval(1))
      else if (ioptavec.eq.5) then
         trace=0.0
         do g=1,iter
            trace=trace+eigenval(g)
         enddo
         tau=tauthr/dsqrt(trace)
      else if (ioptavec.eq.6) then
         trace=0.0
         do g=1,iter
            trace=trace+eigenval(g)
         enddo
         tau=tauthr/dsqrt(trace)
         tau=tau*betavar
      endif

!-----------------------------------------------------------------------
!     Check how many SPFs will be added
!-----------------------------------------------------------------------
      laddspf=.false.
      nspf=0
      do g=1,iter
 !       Definition of beta (e.g. eigenvalues threshold)
         if (ibeta.eq.1) then
            beta=eigenval(g)*tau**2
         else if (ibeta.eq.2) then
            beta=eigenval(g)
         else if (ibeta.eq.3) then
            beta=abs(eigenval(g)/eigenval(1))
         endif
 !       Check how many SPFs will be added
         if ((beta.gt.betathr).and. (abs(eigenval(g)).gt.1.0d-16).and.&
             (nspf.lt.maxnspf)) then
            laddspf=.true.
            nspf=nspf+1
         endif
 !       Write output
         write(iopteigen,'(i5,f18.12,f22.15,f22.15,l4)') &
            g,eigenval(g)*tau**2,eigenval(g),abs(eigenval(g)/eigenval(1)),laddspf
         laddspf=.false.
      enddo

!-----------------------------------------------------------------------
!    Internal checks
!-----------------------------------------------------------------------
!!     if (nspf.eq.0) then
!!        write(iopteigen,'(a)')
!!        write(iopteigen,'(a)') &
!!        ' WARNING: none of the eigenvalues is above threshold betathr'
!!        if (loptnspf) then
!           Add one spf even if the largest eigenvalue is too small
!!           nspf=1
!!           write(iopteigen,'(a)') '          one SPF will be added'
!!        else
!!!          Add maxnspf even if the largest eigenvalue is too small
!!        write(iopteigen,'(a)')
!!        write(iopteigen,'(a)') ' WARNING: nspf reduced to maxnspf=',maxnspf
!!           nspf=maxnspf
!!        endif
!!     endif
!!     write(iopteigen,'(a)')
!!     call myflush(iopteigen)

!-----------------------------------------------------------------------
!     Save eigenvalues
!-----------------------------------------------------------------------
      call cpvxd(eigenval,eigen,nspf)

!-----------------------------------------------------------------------
!     Copy new SPF(s) inside phi vector
!-----------------------------------------------------------------------
      call cpvxz(psi,phi,subdim*nspf)

!-----------------------------------------------------------------------
!     Stop timer
!-----------------------------------------------------------------------
      call stop_timer(tid)

      return
      end subroutine optspfslancz

      subroutine func(psi,hpsi,sdm,mc)

      implicit none

      integer(long) ::    sdm
      complex(dop) :: psi(*),hpsi(*), mc(*)

      call qmxxzz (mc,psi,hpsi,sdm,1)

      return 
      end subroutine func

end module dspfmod
