module dwfn

use decimal
use global
use globmemmod
use dglobal
use maxv
use griddatmod
use psidef
use runpropmod
use channels
use logdat

use initorbmod
use dspfmod
use doptmod

use compdat
use iorst, only: wrrst
use opendatmod, only: opendat
use iopsidef, only: psidat,setzetf,setphidim
use getpsimod, only: resizeml
use multilayer, only: setoperml

use op1lib, only: cpmxi,zerovxz,zerovxl,zeromxi,initvxi,zerovxd,&
                  zerovxi,normvxz
use wftree, only: new_wftree,adim_wftree
use wftree1, only: dispose_wftree,allocate_wftree
use trafonatmod, only: reorthopsi

      implicit none

      logical(kind=4), parameter    :: ldebug = .false.

      public

contains

!***********************************************************************
!                       SUBROUTINE CALLSPWN
!
! Top level and memory allocation in the dynamical wavefunction code
! (i.e. keywords initorb and dynspf).
!
! Contains subroutine: runspwn
!
! Author: David Mendive-Tapia
! Year  : 2019
!
!***********************************************************************

      subroutine callspwn(psi,dtpsi)

      implicit none

!     Maximum sizes (defined in program mctdh)
      complex(dop), dimension(dglmax), intent(inout) :: psi,dtpsi

!     Dynamically allocatable arrays
      complex(dop),allocatable :: oldpsi(:)

!     Writing information to restart file
!     Handling of wavefunction
      integer(long)    :: m,s
!     Save accuracy in improved relaxation
!     Dynamically changing alpha
      real(dop)     :: alphastp
!     Other
      integer(long)    :: i
!     Debugging
      type(wfnode_t),pointer      :: oldtree,newtree

      integer(long)                :: idum
      integer(long), dimension(1)  :: idum1,idum2
      complex(dop), dimension(1)   :: cdum
      real(dop), dimension(1)      :: rdum,rdum1

      logical(kind=4)     :: linwf

!-----------------------------------------------------------------------
! Run DynWF loop
!-----------------------------------------------------------------------
      if (.not.lrunml) then
!        Improved relaxation output message
         if (rlaxnum.ge.0.and.lconst) &
             write(iout,'(a,E11.6,a)') ' Conver = ',constop*ev,' eV'

! ---- MCTDH spawning
         ! call runspwn(psi,dtpsi)
         call runinitorb(psi,dtpsi)

  call normvxz(psi,rdum(1),adim)

! ---- Numerical back propagation algorithm
         if (loptavecnum) then
            call NumAvec(psi,dtpsi,betavar)
         endif

      else
! --- ML-MCTDH spawning

! Store old dimensionalities in dimold(m,s)
         call cpmxi(dim,dimold,maxdim,maxsta)
         call savecsys(2)
         call savecpsi(2)
         oldtree => tree

         allocate(oldpsi(dgldim))
         oldpsi(1:dgldim) = psi(1:dgldim)

! Use tape3 which contains the new dimensionalities
! Update pointers and allocate new wavefunction
         i = 1
         do while (itape2(i) .ne. -2)
            tape(itape2(i)) = itape3(i)
            i=i+1
         enddo
         call cpmxi(dimtape,dim,maxdim,maxsta)
         tree => new_wftree(tape)
         mladim  = adim_wftree(tree)
         ml_input_ready = .true.

         call psidat
         call savecsys(1)
         call savecpsi(1)
         newtree => tree

! System data in memory must be for old WF. Tree = Oldtree.
! Oldtree is disposed of at end of resizeml
         call getcpsi(2)
         call resizeml(oldtree,oldpsi,newtree,cdgldim(2),psi,cdgldim(1))
         deallocate(oldpsi)
   
! Restore data for new WF and allocate new tree fully
         call cpmxi(dimtape,dim,maxdim,maxsta)
         tree => newtree
         call getcsys(1)
         call getcpsi(1)
         call psidat
         call allocate_wftree(tree,.true.,dicht1,dicht2,dicht3,dicht4)
         call setoperml(1,.true.,.true.)

         call reorthopsi(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),&
              gs2)

      endif

! Write information to restart file
      linwf=.true.
      idum=1
      rdum(1)=1.d0
      rdum1(1)=1.d0
      cdum(1)=(1.d0,1.d0)
      call wrrst(tcurrent,psi,linwf,cdum,idum,rdum,rdum1,idum1,agmat,idum2)

!-----------------------------------------------------------------------
!     Write end message line to output
!-----------------------------------------------------------------------
      write(iout,'(a)') 'Increase SPF Basis...'
      write(iout,'(78a)') ('-',i=1,78)

!----------------------------------------------------------------------
!        Output array-dimensions in log file
!----------------------------------------------------------------------
      write(ilog,'(/,a,/)') &
         '******* Array-Dimensions (runprop) *******'
      write(ilog,'(a,2i9,a,f5.1,a)') ' dgldim,  dglmax     :', &
         dgldim,dglmax,' (',real(dgldim)/real(dglmax)*100,' %)'

! Improved relaxation
! Increase accuracy every spawning step
      if (rlaxnum.ge.0.and.lconst) then
         if (constop-cp_constop.gt.tiny) then
            if (.not.ldynspf) then
               write(iout,'(a)') &
               ' Extra step in order to reach final convergence'
               ldynspf=.true.
               constop = constop*1.0d-01
            else
               constop = constop*1.0d-01
            endif
         endif
      endif

!----------------------------------------------------------------------
!        Reset flags
!----------------------------------------------------------------------
! Counters
      idyncyc=idyncyc+1
! Logical
      lovwr = .true.
      lstop = .false.
!     Improved relaxation: switch-off autoblock
      lautoblk=.false.
!     Dynamical alpha
      if (idynalpha.ne.0) then
!        New values
         call CntrlAlpha(idynalpha,alphamax,alphathr,alphastp, &
              alpha0,falpha,idyncyc,2)
      else
!        Write output message
         write(ilog,'(a)') ' Alpha parameters :'
         write(ilog,'(a,e12.5)') ' alpha_max:',alphamax
         write(ilog,'(a,e12.5)') ' alpha_thr:',alphathr
         write(ilog,'(a)')
      endif
      if (lrunml) then
         write(ilog,'(a)') 'The ML-tree has been enlarged.'
         write(ilog,'(78a)') ('-',i=1,78)
      endif

      write(ilog,'(a)') &
         ' The number of SPFs is increased for mode(s):'
      do s=1,nstate
         do m=1,nmode
            if (cdim(m,s,1).gt.cdim(m,s,2)) then
               write(ilog,'(a,i3,a,i3,a,i3,a,i3,a,i3)') &
               ' m:',m,' s:',s, &
               ' dim:',cdim(m,s,2),' =>',cdim(m,s,1), &
               ' out of',maxdimspwn(m,s)
            endif
         enddo
      enddo
      write(ilog,'(a)')

      return  
      end subroutine callspwn

!#######################################################################
! Calculation of sizes taken from start of callspwn
!#######################################################################

      subroutine dynwfsizes

      implicit none

!     Save accuracy in improved relaxation
      real*8     :: cp_constop
!     Dynamically changing alpha
      real*8     :: alphastp

!----------------------------------------------------------------------
! Set defaults for propagation
!----------------------------------------------------------------------
      lorthopsi=.false.

!----------------------------------------------------------------------
! Set operator pointers for ML calcs using latest WF size
!----------------------------------------------------------------------
      if (lrunml) then
         call allocate_wftree(tree,.true.,dicht1,dicht2,dicht3,dicht4)
         call setoperml(1,.true.,.true.)
      endif

!----------------------------------------------------------------------
!     Dynamics SPFs (i.e. dynspf)
!----------------------------------------------------------------------
!     Improved relaxation
!     Save final accuracy and change for lower value
!     Both constop and constop0 are read from the input file
      if (rlaxnum.ge.0.and.lconst) then
         cp_constop = constop
         constop = constop0
      endif
!     Step-size for alpha
      call CntrlAlpha(idynalpha,alphamax,alphathr,alphastp,alpha0, &
           falpha,idyncyc,1)

      return
      end subroutine dynwfsizes

!#######################################################################
!
! iopt = 1 : initialisation
!        2 : during propagation
!
!#######################################################################

      subroutine CntrlAlpha(idynalpha,alphamax,alphathr,alphastp,alpha0, &
                 falpha,idyncyc,iopt)

use decimal

      implicit none

      integer(long) :: idynalpha,iopt
      real(dop)  :: alphamax,alphathr,alphastp,alpha0 
                                  ![alphamax],[alphathr],[delta],[alpha0]
      integer(long) :: idyncyc
      real(dop) :: falpha

      if (iopt.eq.1) then
!        Initialization
         ! if [alphathr] is not defined, set alphathr
         if (dabs(alphathr) .lt. tiny) alphathr=alphamax

         if (idynalpha.ne.0) then
            ! Alpha changes in time
            if (idynalpha.eq.1) then
               alphastp=falpha
            else if (idynalpha.eq.2) then
               alphastp=(alphamax-alpha0)/alphamax*falpha
            endif
!           Final values for the thresholds $\alpha_max$ and $\alpha_thr$
            ! If [alpha0] was defined for alphamax, overwrite
            if (dabs(alpha0 - tiny) .gt. 0.0_dop) alphamax=alpha0
            ! If [alpha0] was defined for alphathr, overwrite
            if (dabs(alphathr0 - tiny) .gt. 0.0_dop) alphathr=alphathr0
            ! If [alpha0] was defined for alphamax,
            ! but no [alpha0] was defined for alphathr,
            ! then make them the same
            alphathr=max(alphathr,alphamax)
         endif
      else if (iopt.eq.2) then
!        New values
         alphamax=alphamax+alphastp
         alphathr=alphathr+falpha
!        Write output message
         write(ilog,'(a,i3,a)') &
            ' Alpha parameters for spawning cycle',idyncyc,':'
         write(ilog,'(a,e12.5,a,e12.5,a,e12.5)') &
            ' alpha_max: old =',alphamax-falpha, &
            ' step =',alphastp,' => new=',alphamax
         write(ilog,'(a,e12.5,a,e12.5,a,e12.5)') &
            ' alpha_thr: old =',alphathr-falpha, &
            ' step =',alphastp,' => new=',alphathr
         write(ilog,'(a)')

      endif

      return
      end subroutine CntrlAlpha

!#######################################################################

      subroutine logdynwf

      implicit none

      integer(long) :: m,s
!     Save accuracy in improved relaxation
!     Dynamically changing alpha

      if (lrunml) then
!        Abort when using optimal orbitals and ML
         if (loptspf) then
            routine = 'runspwn'
            message = 'Only std orbitals implemented with ML-spawn'
            call errormsg
         endif

      else
         write(ilog,'(/,a,/)') '****** Dynamical SPFs ******'

!        Report on variables and calculation type
         if (ldebug) then
            write(ilog,'(a)')
            write(ilog,'(a)') 'Summary on InitOrb + DynSpf'
            write(ilog,*) 'linitorb    :',linitorb
            write(ilog,*) 'loptspf     :',loptspf
            write(ilog,*) 'loptnspf    :',loptnspf
            write(ilog,*) 'loptavec    :',loptavec
            write(ilog,*) 'loptspflancz:',loptspflancz
            write(ilog,*) 'iindyninp(1):',imaxspf
            write(ilog,*) 'rindyninp(1):',betathr
            write(ilog,*) 'rindyninp(3):',betavar
            write(ilog,*) 'tauthr',tauthr
            write(ilog,*) 'ldynspf     :',ldynspf
            write(ilog,*) 'lnspf       :',lnspf
            write(ilog,*) 'iindyninp(2):',ibeta
            write(ilog,*) 'alphamax    :',alphamax
            write(ilog,*) 'alphathr    :',alphathr
            write(ilog,*) 'falpha      :',falpha
            write(ilog,*) 'alpha0      :',alpha0
            write(ilog,*) 'ioptavec2nd :',ioptavec2nd
            write(ilog,*) 'ioptavec    :',ioptavec
            write(ilog,*) 'iresize     :',iresize
            write(ilog,*) 'itermax     :',itermax
         else
 !           if (ldynspf.and.(lmult .and. nstate .gt. 1)) then
 !              routine = 'runspwn'
 !              message = '"multi-set" is not yet tested with spawn'
 !              call errormsg
            if (azahl.gt.0) then
               routine = 'runspwn'
               message = '"A-coeff" is not yet tested with spawn'
               call errormsg
            endif
         endif

         write(ilog,'(a)') 'mode state'
         do s=1,nstate
            do m=1,nmode
               write(ilog,'(1x,i3,3x,i3,a,i3,a,i3)') &
               m,s,' => dim:',dim(m,s),' maxdim:', &
               maxdimspwn(m,s)
            enddo
         enddo
      endif

      return
      end subroutine logdynwf

end module dwfn
