!***********************************************************************
!                          MODULE INITORB
!
! MCTDH module for the computation of optimal initial orbitals
! (i.e. keyword initorb).
!
! Contains subroutines: initorb, preiter, resizefmat.
!
! Author: David Mendive-Tapia
! Year  : 2019
!
!***********************************************************************

module initorbmod

use global
use globmemmod
use dglobal
use maxv
use griddatmod
use psidef
!use runinwfmod
use runpropmod
use channels
use logdat
use operdef
use compdatmod
use datenmod
use mpidata
use hpsimod
use memhopsmodmod
use compdatmod
use iopsidef, only: psidat
use phihphimod, only: phihphi
use mempropmod, only: checkmemprop
use rddvrmod, only: rddvr
use iopsidef, only: setzetf, setphidim
use iorst, only: rdrst
use initpropmod, only: initprop
use einsbasmod,  only: logsbas
use trafonatmod, only: trafonat
use genphi1, only: inventspfs
use schmidtorthomod, only: schmidtortho

use dspfmod
use doptmod
use function

use rddvrmod, only: dvrdata
use rdopermod, only: rdoper
use op1lib, only: zerovxd,zerovxz,zerovxi,zerovxl,normvxz,cpvxz,&
                  cpmxi,cpqxz,initvxi,zeromxi

      implicit none

      public

      contains

!----------------------------------------------------------------------
!     Initial optimal orbitals (i.e. initorb)
!----------------------------------------------------------------------

      subroutine runinitorb(psi,dtpsi)

      complex(dop), dimension(:), intent(inout) :: psi,dtpsi
      complex(dop), dimension(:), allocatable   :: oldpsi
      complex(dop), dimension(:), allocatable   :: tmppsi
      real(dop)                                 :: betathr_old

      real(dop)                                 :: rdum

!     Dynwf memory allocation
      integer(long)    :: maxblock3,maxdimak,zeig,zeig1,zeig2

      integer(long)    :: m,s 

      logical(kind=4), parameter    :: ldebug = .false.
!
      if (lrunml) then
         routine = 'Initorb (Initorbmod.f90)'
         message = 'Optimal orbitals cannot be used with ML-MCTDH '
         call errormsg
      endif

      call zeromxi(dimocc,maxdim,maxsta)
      call zeromxi(dimunocc,maxdim,maxsta)

! Switch-off second-order correction with Rlx and IRlx
      ioptavec2nd=0       ! doesn't seem to be used

      betathr_old = betathr
      betathr = betathr0

!-----------------------------------------------------------------------
! Transform to natorbs 
!-----------------------------------------------------------------------
      call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
      call trafonat(psi,tree,dicht4,mi(mijindx))

!-----------------------------------------------------------------------
!     --- EVALUATE RIGHT HAND SIDE OF ODE ---
!-----------------------------------------------------------------------
      if (lfunkl) then
         call funk(tcurrent,psi,dtpsi)
      else
         call funkr(tcurrent,psi,dtpsi)
      endif
  call normvxz(psi,rdum,adim)

!-----------------------------------------------------------------------
! Define no. of occupied (and unoccupied) orbitals
! If call at start (initorb), dim defines total no. of SPFs and need to get
!    dimocc
! If call during propagation (dynorb), dimtape defines total no. of SPFs 
!    and dim no. of occupied.
!-----------------------------------------------------------------------
      if (linitorb) then
! Save final WF size in cdim(1)
         call savecsys(1)
         call savecpsi(1)

         do s=1,nstate
            do m=1,nmode
               if (complete(m)) then
                  dimocc(m,s) = dim(m,s)
               else
                  call getdimocc(dimocc(m,s),dicht3(d3mat(m,s)),dim(m,s))
               endif
               dimunocc(m,s) = dim(m,s) - dimocc(m,s)
            enddo
         enddo
! Save WF size with occupied orbitals only in cdim(2)
         dim(:,:) = dimocc(:,:)
         call psidat
         call savecsys(2)
         call savecpsi(2)

      else
         allocate(oldpsi(dgldim))
         oldpsi(1:dgldim) = psi(1:dgldim)
 
! Save WF size with occupied orbitals only in cdim(2)
         call savecsys(2)
         call savecpsi(2)
         do s=1,nstate
            do m=1,nmode
               dimocc(m,s) = dim(m,s)
               dimunocc(m,s) = dimtape(m,s) - dimocc(m,s)
            enddo
         enddo
  call normvxz(psi,rdum,adim)
! Save final WF size in cdim(1)
         dim(:,:) = dimtape(:,:)
         call psidat
         call savecsys(1)
         call savecpsi(1)

! Expand WF to give space for new SPFs
         allocate(tmppsi(2*adim))
         psi(:) = 0.0_dop
         do s=1,nstate
            zeig=1
            zeig1=zeig+block(s)
            tmppsi(:) = 0.0_dop
            call cpvxz(oldpsi(czpsi(s,2)),tmppsi(zeig),cblock(s,2))

            do m=1,nmode
               call akcopy(tmppsi(zeig),tmppsi(zeig1),cvdim(m,s,1),cdim(m,s,1), &
                    cdim(m,s,2),cndim(m,s,2))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            enddo
            psi(zpsi(s):zpsi(s)+block(s)-1) =  &
                                   tmppsi(zeig:zeig+block(s)-1) 

            do m=1,nmode
               psi(zetf(m,s):zetf(m,s)+cdim(m,s,2)*subdim(m)-1) =  &
                      oldpsi(czetf(m,s,2):czetf(m,s,2)+cdim(m,s,2)*subdim(m)-1) 
            enddo
         enddo

         deallocate(oldpsi,tmppsi)
      endif

! Restore final WF size
      call getcsys(1)
      call getcpsi(1)
      call psidat
  call normvxz(psi,rdum,adim)

! Quasi-optimal orbitals as starting point
      if (lqoptspf) call QOptSpf(psi)

! Call to zeigspwn
! To sort out later: simplify zeigspwn
      call zeigspwn(maxblock3,maxdimak,workcdim)

! Generate new MCTDH wavefunction
      call initorb(psi,dtpsi,maxblock3,maxdimak)

! Write new SPF basis in log file
      call logsbas

      betathr = betathr_old

      return
      end subroutine runinitorb

!***********************************************************************
! see how many SPFs are occupied
!***********************************************************************

      subroutine getdimocc(dimocc,dicht3,dim1)

      implicit none

      integer(long), intent(out) :: dimocc
      integer(long), intent(in)  :: dim1
      integer(long)              :: e
      real(dop), dimension(dim1), intent(in) :: dicht3
      real(dop)                              :: thr = 1.0d-12

      dimocc=0
      do e=1,dim1
         if (dicht3(e) .gt. thr) dimocc = dimocc+1
      enddo
      dimocc=max(dimocc,1)   !  in a multi-set calc there may be 0 
                             !  occupation. Make dim =1

      return
      end subroutine getdimocc

!***********************************************************************
! INITORB
!
! Main driver for the computation of optimal initial orbitals.
!
! INPUT/OUTPUT
!
!***********************************************************************

      subroutine initorb(psi,dtpsi,maxblock3,maxdimak)

      implicit none

!     INPUT/OUTPUT
!     Wavefunction
      complex(dop) ::     psi(dgldim),dtpsi(dgldim)
!     Stotage pointers
      integer(long) ::        maxblock2,maxblock3
!     Handling of wavefunction
      integer(long) ::        m,s
!     For subroutine rdrst

!     INTERMEDIATE
!     Storage
      integer(long) ::        maxdimak,nspf(maxdim,maxsta)
!     Optimal A-vector
      logical(kind=4) ::        lcavec
!     Optimal SPFs I
      complex(dop) :: nwas(maxdimak*nmode)
      complex(dop) :: nwspfs(sdim*maxspf,nmode,nstate)
!     Optimal SPFs II
      complex(dop) :: F(maxblock3)
      complex(dop) :: D(sdim*sdim)
      real(dop) :: eigenavec(maxspf,nmode),eigentrace(nmode)
!     Second order correction I
      integer(long) ::        sumdim
      real(dop) ::         sumeigen(nmode),maxeigen(nmode)
      real(dop) ::         norm0,norm1
!     Second order correction II
      complex(dop) ::     Fpp(maxblock3),dicht1(dmatdim)
      complex(dop) ::     F1(maxblock3)
      complex(dop) ::     psi1(dgldim),psi2(dgldim)
!     Density matrix for the computation of Fpp
      complex(dop) ::     denmat1(dmatdim)
      real(dop) ::         workr(workrdim)
!     Other
      integer(long)       :: ierr
      character(len = 52) :: LineSec

!----------------------------------------------------------------------
!     Clean arrays
!----------------------------------------------------------------------
      call zerovxz(nwspfs,sdim*maxspf*nmode*nstate)
      call zerovxz(nwas,maxdimak*nmode)
      call zerovxd(eigenavec,maxspf*nmode)
      call zerovxz(F1,maxblock3)
      call zerovxz(Fpp,maxblock3)

      nspf(:,:)=0

!----------------------------------------------------------------------
!     Write header in output file
!----------------------------------------------------------------------
      LineSec = '----------------------------------------------------'
      write(iopteigen,'(a,a)') '#-------------------',LineSec
      write(iopteigen,'(a)') '# Computing initial optimal SPFs'
      write(iopteigen,'(a,a)') '#-------------------',LineSec

!-----------------------------------------------------------------------
!     Set defaults
!-----------------------------------------------------------------------
!     Overwrite directory
      lovwr=.true.
!     Optimal SPFs
!      loptspf=.true.
!     A-vector is a complex array
      lcavec=.true.
!     When improved relaxation, A-vector must be real
      if (rlaxnum.ge.0.and.lconst) lcavec=.false.

!-----------------------------------------------------------------------
!     Compute optimal initial orbitals. Algorithm controlled by:
! iresize
!-1: do not enlarge orthogonal space (e.g. fixed WF size - used if not 
!    dynspf 
! 0: enlarge orthogonal space using Eq. derived for a Hartree product
! 1: enlarge orthogonal space by resizing the matrix F (general)
! 2: enlarge only mode 1 with standard orbitals (for debugging)
!
!-----------------------------------------------------------------------
      if (iresize .eq. -1) then
!        Pre-iterations
         if (loptspf) then
            call preiter(psi,nwspfs,eigenavec,nspf,maxblock3, &
                 maxdimak,sumeigen,maxeigen,eigentrace)
         endif
! Copy new SPFs to Psi
! 1. take eigenvalues of F operator
! 2. fill in with invented functions if required.
         do s=1,nstate
            do m=1,nmode
               if (complete(m)) cycle
               if (nspf(m,s) .gt. 0) then
                  call cpvxz(nwspfs(1,m,s),psi(zetf(m,s)+dimocc(m,s)*subdim(m)), &
                       subdim(m)*(nspf(m,s)))
               endif

               if (dimunocc(m,s) .gt. nspf(m,s)) then
                  call inventspfs(m,psi(zetf(m,s)),dimocc(m,s)+nspf(m,s),&
                       dim(m,s),subdim(m),mr(mrort),mc(mcexphin),&
                       mc(mcexprueck))
               endif
               call schmidtortho(psi(zetf(m,s)),subdim(m),dim(m,s),ierr)
               if(ierr.ne.0) then
                  routine = 'Initorb (Initorbmod.f90)'
                  write(message,'(a,2i5,a,i5)') &
                     ' Generation of higher SPFs failed For mode, state. ', m,s, &
                     ' Ierr =', ierr
                  call errormsg
               endif
            enddo
         enddo
         if (loptavec) then
! mfields2 needs mc(mchteil)
            call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
! Compute optimal A-vector
            call newavechartree (hteil,psi,sumeigen, &
                 maxeigen,eigentrace,lcavec)
            call normvxz(psi,norm1,adim)
            write(ilog,'(a,f11.8)') ' Norm (After optimal A-vector): ',norm1
            call xvixdzo(norm1,psi,adim)
         endif

      else if (iresize.eq.0) then

!        Save the enlarged space wfn in dimunocc(m,s)
!         call cpmxi(dim,dimunocc,maxdim,maxsta)
!        Define values inside dimocc(m,s)
!         call cpmxi(dim,dimocc,maxdim,maxsta)
!         do s=1,nstate
!            do m=1,nmode
!               if ((m.ne.meb).and.(m.ne.mpb)) dimocc(m,s)=1
!            enddo
!         enddo
!        Pre-iterations
         call preiter(psi,nwspfs,eigenavec,nspf,maxblock3, &
              maxdimak,sumeigen,maxeigen,eigentrace)
!        Make Hartree product
!        Resize psi (dim (large) => dimocc (small)):
!        1) Pointers for initial psi (e.g. dimold(m,s))
         call cpmxi(dim,dimold,maxdim,maxsta)
!        2) Pointers for final psi (e.g. dimocc(m,s))
!        3) Pointers for memory allocation (e.g. dimunocc(m,s))
!         call cpmxi(dimunocc,dim,maxdim,maxsta)
!         call psidat
!        4) Call ReSizePsiNmodes
         call ReSizePsiNmodes(psi,nwas,nwspfs,dimold,dimocc,maxdimak)

!-----------------------------------------------------------------------
!        Write eigenvectors into spwnevec
!-----------------------------------------------------------------------
         if (ldebugspawn) call wrt_eigenvec(psi)
!        Compute norm before enlargement
         call normvxz(psi,norm0,adim)
         write(ilog,'(a,f11.8)') ' Norm (A): ',norm0
!        6) Re-size wavefunction (all modes)
         call cpmxi(dimocc,dimold,maxdim,maxsta)
         dimtape(:,:)=dimocc(:,:)+nspf(:,:)
         call cpmxi(dimtape,dim,maxdim,maxsta)
         call psidat
         call FitSizePsiNmodes(psi,nwas,nwspfs,maxdimak)
!        7) Optimal A-vector (Hartree product)
         if (loptavec) then
!           mfields2 needs mc(mchteil)
            call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
!           Density matrix
!c            call dyndmat(psi,dgldim,denmat1,dmatdim,
!c     +           d3matdim,workr,workrdim)
!c            call cpvxz(denmat1,mc(mcdicht1),dmatdim)
!           Compute optimal A-vector
            call newavechartree (hteil,psi,sumeigen, &
                 maxeigen,eigentrace,lcavec)
         endif
!        Debugging
         if (loptavec2nd) then
            call normvxz(psi,norm1,adim)
            write(ilog,'(a,f11.8)') ' Norm (A): ',norm1
            write(ilog,'(a,E12.6)') ' Error: ',(norm1-norm0)
         endif

      else if (iresize.eq.1) then
!        Save the enlarged space wfn in psi(1)
         call psidat
         call savecsys(1)
         call savecpsi(1)
         call cpmxi(dim,dimunocc,maxdim,maxsta)
!        Make Hartree product
         call cpmxi(dim,dimocc,maxdim,maxsta)
         do s=1,nstate
            do m=1,nmode
               if ((m.ne.meb).and.(m.ne.mpb)) dimocc(m,s)=1
            enddo
         enddo
!        Pre-iterations
         call preiter(psi,nwspfs,eigenavec,nspf,maxblock3, &
              maxdimak,sumeigen,maxeigen,eigentrace)
         call cpmxi(dim,dimunocc,maxdim,maxsta)
!        Store the enlarged space wfn in psi1
         call cpvxz(psi,psi1,dgldim)
!        Resize psi (dim (large) => dimocc (small)):
!        1) Pointers for initial psi (e.g. dimold(m,s))
         call cpmxi(dim,dimold,maxdim,maxsta)
!        2) Pointers for final psi (e.g. dimocc(m,s))
!        3) Pointers for memory allocation (e.g. dimunocc(m,s))
         call cpmxi(dimunocc,dim,maxdim,maxsta)
         call psidat
!        4) Call ReSizePsiNmodes
         call ReSizePsiNmodes(psi,nwas,nwspfs,dimold,dimocc,maxdimak)
!        5)
!        Compute matrix elements and density with (contracted space)
!        mfields2 needs mc(mchteil)
         call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
!        Density matrix
         call dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)
         call cpvxz(denmat1,dicht1,dmatdim)
!        Save the contracted space wfn in psi(2)
         call savecsys(2)
         call savecpsi(2)
!        Store the contracted space wfn in psi2
         call cpvxz(psi,psi2,dgldim)

!        Write eigenvectors into spwnevec
         if (ldebugspawn) call wrt_eigenvec(psi)
!        Compute norm before enlargement
         call normvxz(psi,norm0,adim)
         write(ilog,'(a,f11.8)') ' Norm (A): ',norm0
!        Loop
         do s=1,nstate
            do m=1,nmode
               if ((m.ne.meb).and.(m.ne.mpb)) then
!                 Restore enlarged space wfn
                  call cpmxi(dimunocc,dim,maxdim,maxsta)
                  call psidat
                  call SameSizeCpSpFs (psi1,psi,dimocc,nwspfs)
!c                  call cpvxz(psi1,psi,dgldim)
!                 Compute matrix elements and density with (current space)
!                 mfields2 needs mc(mchteil)
                  call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
!                 Density matrix
                  call dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)
                  call cpvxz(denmat1,dicht1,dmatdim)

                  call optpsi(tcurrent,s,m,psi,dtpsi,D,F,denmat1,nspf(m,s), &
                       nwspfs(1,m,s), &
                       eigenavec(1,m),eigentrace(m),maxblock3,maxdimak,&
                       maxblock2,&
! Fix for increased eigenvalue
     +                 imaxspf)
!c     +                 dimtape(m,s)-1,
!                 Re-dimensionalize according to the Eigenvalues of D
                  if (loptnspf) then
                     dimtape(m,s)=1+nspf(m,s)
                  endif
                  call resizefmat(m,s,F,F1, &
                       vdim(m,s),subdim(m),ndim(m,s), &
                       cvdim(m,s,2),cndim(m,s,2))
!
                  call cpmxi(dimocc,dim,maxdim,maxsta)
                  call psidat
                  call cpvxz(psi2,psi,dgldim)
!                 mfields2 needs mc(mchteil)
!c                  call phihphi(psi,psi,mc(mchpsi),hteil,mc(mcworkc),1,usehsym)
!c     +                 mi(mihi),ml(mlhl))
!                 Density matrix
                  call dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)
                  call cpvxz(denmat1,dicht1,dmatdim)
!                 4) Only one mode is increased by one spf
                  call dynpsi(tcurrent,m,s,psi,dtpsi,F1,nspf(m,s), &
                       sumeigen(m),eigenavec(1,m), &
                       nwspfs(1,m,s), &
                       nwas(1+(m-1)*maxdimak),denmat1,lcavec,norm1, &
                       maxblock3,maxblock2,maxdimak)
!                 Resize psi (dimocc (small) => dimtape (large)):
!                 1) Pointers for initial psi (e.g. dimocc(m,s))
!                 2) Pointers for final psi (e.g. dim(m,s) + enarged with dimtape(m,s))
                  call cpmxi(dimocc,dimold,maxdim,maxsta)
                  dimold(m,s)=dimtape(m,s)
!                 3) Pointers for memory allocation (e.g. dimunocc(m,s))
                  call cpmxi(dimunocc,dim,maxdim,maxsta)
                  call psidat
!                 4) Call ReSizePsiNmodes
                  call ReSizePsiNmodes(psi,nwas,nwspfs,dimocc,dimold,maxdimak)
!
                  call savecsys(2)
                  call savecpsi(2)
                  call cpvxz(psi,psi2,dgldim)
!
                  dimocc(m,s)=dimtape(m,s)
!                 5) Update density
!c                  call dyndmat(psi,dgldim,denmat1,dmatdim,
!c     +                 d3matdim,workr,workrdim)
!c                   endif
               endif
            enddo
         enddo
!        6) Re-size wavefunction (all modes)
         call cpmxi(dimtape,dim,maxdim,maxsta)
         call psidat
         call FitSizePsiNmodes(psi,nwas,nwspfs,maxdimak)
!        Debugging
         if (loptavec2nd) then
            call normvxz(psi,norm1,adim)
            write(ilog,'(a,f11.8)') ' Norm (A): ',norm1
            write(ilog,'(a,E12.6)') ' Error: ',(norm1-norm0)
         endif

! write wavefunction to restart file
 !        linwf=.true.
 !        idum=1
 !        rdum(1)=1.d0
 !        rdum1(1)=1.d0
 !        cdum(1)=(1.d0,1.d0)
 !        call wrrst(tcurrent,psi,linwf,cdum,idum,rdum,rdum1,&
 !             mi(mijindx),agmat,mi(mitrajst))

      else if (iresize.eq.2) then
!        Save the enlarged space wfn in psi(1)
         call psidat
         call savecsys(1)
         call savecpsi(1)
         call cpmxi(dim,dimunocc,maxdim,maxsta)
!        Pre-iterations
         call preiter(psi,nwspfs,eigenavec,nspf,&
              maxblock3,maxdimak,sumeigen,maxeigen,eigentrace)
!        Compute sum of dimocc(m,s) to identify Hartree product
         sumdim=0
         do s=1,nstate
            do m=1,nmode
               sumdim=sumdim+dimocc(m,s)
            enddo
         enddo
!        Increase orthogonal space if Hartree product
         if (sumdim.eq.nmode) then
!           dimtape(m,s) contains enlarged space
!           (from input, QOptSpf or preiter)
            dimocc(1,1)=dimtape(1,1)
         endif

! Trick: enlarge orthogonal space by enlarging mode=1
!        Resize psi (dim (small) => dimocc (large)):
!        1) Pointers for initial psi (e.g. dimold(m,s))
         call cpmxi(dim,dimold,maxdim,maxsta)
!        2) Pointers for final psi (e.g. dimocc(m,s))
!        3) Pointers for memory allocation (e.g. dimunocc(m,s))
         call cpmxi(dimunocc,dim,maxdim,maxsta)
         call psidat
!        4) Call ReSizePsiNmodes
         call ReSizePsiNmodes(psi,nwas,nwspfs,dimold,dimocc,maxdimak)
!        5) Compute matrix elements and density with (contracted space)
!        mfields2 needs mc(mchteil)
         call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
!        Density matrix
         call dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)
         call cpvxz(denmat1,dicht1,dmatdim)
!
         call cpmxi(dimocc,dimold,maxdim,maxsta)
! To sort out later: organize norm1
         call normvxz(psi,norm1,adim)
         write(ilog,'(a,f11.8)') ' norm2 (A): ',norm1

!        Write eigenvectors into spwnevec
         if (ldebugspawn) call wrt_eigenvec(psi)
!        Enter runspwn with enlarged pointers
         call cpmxi(dimtape,dim,maxdim,maxsta)
         call psidat
!        Generate new MCTDH wavefunction
         iresize=0
         !call runspwn(psi,dtpsi,maxblock3,maxdimak)
         call runspwn(psi,dtpsi)


      else 
         routine = 'Initorb'
         message =' Unknown resize option '
         call errormsg
      endif

!-----------------------------------------------------------------------
!     Flag any complete spf bases
!-----------------------------------------------------------------------
      do m=1,nmode
         do s=1,nstate
            if (subdim(m) .eq. dim(m,s)) then
               complete(m)=.true.
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
!     Write eigenvectors into spwnevec
!-----------------------------------------------------------------------
      if (ldebugspawn) call wrt_eigenvec(psi)

      return
      end subroutine initorb

!########################################################################

!     To sort out later:  this belong in dopt.F
      subroutine SameSizeCpSpFs (workc,psi,dim0,nwspfs)

      implicit none

!     Wavefunction(s)
      complex(dop) :: psi(dgldim),workc(workcdim)
!
      complex(dop) :: nwspfs(sdim*maxspf*nmode)
!     Handling of wavefunction I
      integer(long) ::    m,s
!
      integer(long) ::    dim0(maxdim,maxsta),cp_dim0(maxdim,maxsta)

      call cpvxz(workc,psi,adim)

      do m=1,nmode
         do s=1,nstate
            if ((m.ne.meb).and.(m.ne.mpb)) then
               cp_dim0(m,s)=dim0(m,s)
               dim0(m,s)=1
!              Copy first SPF to psi(dgldim)
               call cpmxz(workc(zetf(m,s)),psi(zetf(m,s)),subdim(m),dim0(m,s))
               if (dim(m,s).gt.1) then
!                 Copy the remaining SPFs to psi(dgldim)
                  call cpvxz(nwspfs(1+(m-1)*sdim*imaxspf), &
                       psi(zetf(m,s)+subdim(m)*dim0(m,s)), &
                       subdim(m)*(dim(m,s)-dim0(m,s)))
               endif
               dim0(m,s)=cp_dim0(m,s)
            endif
         enddo
      enddo

      return
      end subroutine SameSizeCpSpFs 

!***********************************************************************
! PREITER
!
! Loop for the computation of optimal orbitals N cycles before the using
! the final set (e.g. computing A-vector and second-order correction).
!
! INPUT
! itermax
!
! OUTPUT
!
!***********************************************************************

      subroutine preiter(psi,nwspfs,eigenavec,nspf, &
                 maxblock3,maxdimak,sumeigen,maxeigen,eigentrace)

      implicit none

!     INPUT
!     Wavefunction
      complex(dop) ::     psi(dgldim),dtpsi(dgldim)
!     Handling of wavefunction
      integer(long) ::        m,s
      integer(long) ::        nspf(maxdim,maxsta)
!     Density matrix for the computation of Fpp
      complex(dop) ::     denmat1(dmatdim)
!     Other
      integer(long) ::        i,e
      integer(long) ::        maxblock3,maxblock2,maxdimak

!     INTERMEDIATE
      real(dop) ::         workr(workrdim)
      real(dop) ::         sumeigen(nmode),maxeigen(nmode)
      real(dop) ::         eigenavec(maxspf,nmode),eigentrace(nmode)
      complex(dop) ::     fmat(maxblock3)
      complex(dop) :: dltmat(sdim*sdim)

!     DUMMY
!     To sort out later: reorganize so dummy is avoided
      complex(dop) ::     nwas(maxdimak*nmode)

!     OUTPUT
      complex(dop) :: nwspfs(sdim*maxspf,nmode,nstate)

!----------------------------------------------------------------------
!     Clean arrays
!----------------------------------------------------------------------
       call zerovxz(nwspfs,sdim*maxspf*nmode*nstate)
       call zerovxz(nwas,maxdimak*nmode)
       call zerovxd(eigenavec,maxspf*nmode)

       nspf(:,:) = 0

!----------------------------------------------------------------------
!      Pre-iterations
!----------------------------------------------------------------------
       do i=1,itermax+1
!-----------------------------------
!        mfields2 needs mc(mchteil)
!-----------------------------------
         call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)

!-----------------------------------
!        Density
!-----------------------------------
         call dyndmat(psi,dgldim,denmat1,dmatdim,d3matdim,workr,workrdim)
         call cpvxz(denmat1,dicht1,dmatdim)

!-----------------------------------
!        Loop over modes and states
!-----------------------------------
          do m=1,nmode
             do s=1,nstate
                if (dimunocc(m,s) .eq. 0) cycle
! Compute optimal SPFs
                call optpsi(tcurrent,s,m,psi,dtpsi,dltmat,fmat,&
                     denmat1,nspf(m,s),nwspfs(1,m,s),eigenavec(1,m),&
                     eigentrace(m),maxblock3,maxdimak,maxblock2,&
                     dimunocc(m,s))
!                 Save new dimensionality according to the eigenvalues of D
                  if ((i.eq.itermax+1).and.loptnspf) then
!c                  if ((iresize.eq.2).and.loptnspf) then
                      dimtape(m,s)=1+nspf(m,s)
!c                  endif
                  endif
!              Compute overlaps for debugging
!               call diagnostic2(psi,vdim(m,s),dimtape(m,s),ndim(m,s))
            enddo
         enddo
      enddo

!-----------------------------------
!
!-----------------------------------
      do m=1,nmode
         if (nspf(m,1) .eq. 0) cycle
         sumeigen(m)=0.0
         do e=1,nspf(m,1)
            sumeigen(m)=sumeigen(m)+eigenavec(e,m)
         enddo
         maxeigen(m)=0.0
         maxeigen(m)=eigenavec(1,m)
      enddo

      return
      end subroutine preiter

!***********************************************************************
! RESIZEFMAT
!
! Reisize mean-field-like matrix F. This is needed when enlarging the
! orthonal space.
!
! INPUT
! s         : electronic state of interest
! m         : mode of interest
! finp      :
! maxblock2 :
! maxblock3 :
!
! OUTPUT
! fout      :
!
!***********************************************************************

      subroutine resizefmat(m,s,finp,fout,vdim,subdim,ndim,cvdim,cndim)

      implicit none

!     Handling of wavefunction
      integer(long) ::    m,s
!     Handling of tensor
      integer(long) ::    v,e,n
!     Matrices F
      integer(long) ::    subdim
      integer(long) ::    vdim,ndim,cvdim,cndim
      complex(dop) :: finp(vdim,subdim,ndim)
      complex(dop) :: fout(cvdim,subdim,cndim)

!     Debugging
!c      write(6,*) 'TRANSFORM F'
!c      write(6,*) vdim1,subdim,ndim1,maxblock2
!c      write(6,*) cvdim1,cndim1

!     To sort out later: check that cvdim*subdim > vdim*ndim

!     Re-arraingement
      do v=1,cvdim
         do e=1,subdim
            do n=1,cndim
               fout(v,e,n)=finp(v,e,n)
            enddo
         enddo
      enddo

      return
      end subroutine resizefmat

end module initorbmod
