!***********************************************************************
! 
!                Module INIT1
!
! Contents: initho, initrho, initsin, initexp, initfft, initcos,
!           initsphfbr, initlagu, initgwpgrid, initlob 
!
! calls lapack-routine DSTEQR to diagonalise tri-diagonal matrices.
!
! Initialises the arrays trafo, dif2mat and dif1mat for a DVR basis set
!
! variables passed:
!  trafo   : DVR/FBR transformation matrices 
!  ort     : DVR grid points
!  dif1mat : DVR first  derivative matrix
!  dif2mat : DVR second derivative matrix
!  gdim    : number of DVR grid points for this mode
!  xi      : Value of first DVR grid point
!  xf      : Value of last  DVR grid point
!  hofm    : mass * vibrational frequency (homass*hofreq)
!  hoxeq   : equilibrium position of oscillator 
!  workr   : work array 
!
!
!***********************************************************************

      module init1mod

use decimal, only: long, dop
use logdat
use mmlib,   only: qdxxdd, qqxtdd, qqxxdd, qqtxdd
use op1lib,  only: zerovxd, zeromxd

      implicit none
      private
      public :: initho, initexp, initrho, initsin, initsinc, initwsinc, &
                initfft, initsphfbr, initlagu, initcos, initgwpgrid, &
                initlob
      contains

!-----------------------------------------------------------------------
!                Subroutine INITHO
!             Harmonic Oscillator DVR
!-----------------------------------------------------------------------

      subroutine initho(trafo,ort,dif2mat,dif1mat,gdim &
                         ,hofm,hoxeq,dx,hotyp,message)

      implicit none

      integer(long)                                :: g,g1,g2,error
      integer(long), intent(in)                    :: gdim,hotyp
      real(dop)                                    :: x
      real(dop), intent(in)                        :: hoxeq,dx
      real(dop), intent(inout)                     :: hofm
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif1mat,dif2mat
      character(len=*), intent(out)                :: message

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! coordinate operator (note that the diagonal elements are always zero).
!-----------------------------------------------------------------------

! --- Set diagonal elements to 0  

      call zerovxd(ort,gdim)

! --- Set super-diagonal elements for HO-FBR basis.(g=n+1)

      do g=1,gdim-1
         dif2mat(g,1)=sqrt(g/(2.0_dop*hofm))
      enddo


! --- Diagonalize coordinate-matrix to obtain "ort" and "trafo".

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising HO-coordinate-matrix'
         call errormsg
      endif

! --- If xi-xf flag is set
 
      if(hotyp .eq. 1) then
         hofm = dx/(ort(gdim)-ort(1))
         do  g = 1, gdim
            ort(g) = ort(g)*hofm
         end do
         hofm = hofm**(-2)
      end if

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
!  ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g) 
            enddo
         end if
      enddo

!-----------------------------------------------------------------------
! Calculate second derivative matrix in corresponding DVR basis
!
! Harmonic oscillator: H=T+V in FBR basis and V in DVR basis are
! analytically known --> define 2*mass*H in FBR basis, transform to
! DVR basis and subtract 2*mass*V (diagonally) given in DVR basis to
! get  d^2/dQ^2 = -2*mass*T = -2*mass*(H-V) in DVR basis.
!-----------------------------------------------------------------------

      call zeromxd(dif2mat,gdim,gdim)

! --- Total energy in FBR basis (given analytically and diagonal)

      do g=1,gdim
         x=(g-0.5_dop)*hofm*2.0_dop

! --- Transform to DVR basis 

         do g1=1,gdim
            do g2=1,gdim
               dif2mat(g2,g1)=dif2mat(g2,g1)-trafo(g,g2)*x*trafo(g,g1)
            enddo
         enddo
      enddo

! --- Subtract potential contribution for harmonic oscillator

      do g=1,gdim
         dif2mat(g,g) = dif2mat(g,g) + (hofm*ort(g))**2
      enddo
 

!-----------------------------------------------------------------------
! Calculate DVR matrix representation for first derivative.
!
!  The FBR representation of d/dx is antisymmetric tridiagonal with 
!  zero diagonal elements and analytically given as (assuming hbar=1): 
!
!       <ho_n|d/dx|ho_m> = -sqrt(mass*freq/2)*
!                         [ sqrt(m+1) delta_nm+1 - sqrt(m) delta_nm-1]
!
!-----------------------------------------------------------------------

      call zeromxd(dif1mat,gdim,gdim)
      
      do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis

         x = -sqrt(hofm*g/2.0_dop)

! --- Transform from FBR to DVR basis

         do g1=1,gdim
            do g2=1,gdim
               dif1mat(g2,g1)=dif1mat(g2,g1)+trafo(g+1,g2) &
                    *trafo(g,g1)*x-trafo(g,g2)*trafo(g+1,g1)*x
            enddo
         enddo
      enddo


! --- Shift grid points

      do g=1,gdim
         ort(g)=ort(g)+hoxeq
      enddo

      return
      end subroutine initho


!#######################################################################

!-----------------------------------------------------------------------
!                Subroutine INITRHO
!            Radial Harmonic Oscillator DVR
!-----------------------------------------------------------------------

      subroutine initrho(trafo,ort,dif2mat,dif1mat,gdim &
                         ,hofm,hoxeq,xi,dx,hotyp,message)

      implicit none

      integer(long)                                :: g,g1,g2,error
      integer(long), intent(in)                    :: gdim,hotyp
      real(dop)                                    :: x
      real(dop), intent(in)                        :: xi,dx
      real(dop), intent(out)                       :: hoxeq
      real(dop), intent(inout)                     :: hofm
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) ::  trafo,dif2mat,dif1mat
      character(len=*), intent(out)                :: message

!-----------------------------------------------------------------------
! phi_j = sqrt(2)*phi^{HO}_2j-1 ; odd harmonic oscillator functions.
! Calculate symmetric tridiagonal matrix representation X_mn of the
! transformed coordinate operator z = (x-x_0)^2 .
!-----------------------------------------------------------------------

! --- Set superdiagonal elements for RHO-FBR basis.

         do g=1,gdim-1
            dif2mat(g,1)=sqrt(dble(2*g*(2*g+1)))/(2.0_dop*hofm)
         enddo

! --- Set diagonal elements for RHO-FBR basis.

         do g=1,gdim
            ort(g) = (4*g-1)/(2.0_dop*hofm)
         enddo

! --- Diagonalize coordinate-matrix to obtain "ort" and "trafo".

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising HO-coordinate-matrix'
         call errormsg
      endif

! --- Transform ort
      
      do g=1,gdim
         ort(g) = sqrt(ort(g))
      enddo

! --- If xi-xf or x0-xf flag is set
 
      if(hotyp .eq. 1) then
         hofm  = dx/(ort(gdim)-ort(1))
         do  g = 1, gdim
            ort(g) = ort(g)*hofm 
         end do
         hofm  = hofm**(-2)
         hoxeq = xi - ort(1)
      elseif(hotyp .eq. 2) then
         hofm  = dx/ort(gdim)
         do  g = 1, gdim
            ort(g) = ort(g)*hofm
         end do
         hofm  = hofm**(-2)
      end if

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
! ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g) 
            enddo
         end if
      enddo

!-----------------------------------------------------------------------
! Calculate second derivative matrix in corresponding DVR basis
!
! RHarmonic oscillator: H=T+V in FBR basis and V in DVR basis are
! analytically known --> define 2*mass*H in FBR basis, transform to
! DVR basis and subtract 2*mass*V (diagonally) given in DVR basis to
! get  d^2/dQ^2 = -2*mass*T = -2*mass*(H-V) in DVR basis.
!-----------------------------------------------------------------------

      call zeromxd(dif2mat,gdim,gdim)

! --- Total energy in FBR basis (given analytically and diagonal)

      do g=1,gdim
         x=(2*g-0.5_dop)*hofm*2.0_dop

! --- Transform to DVR basis 

         do g1=1,gdim
            do g2=1,gdim
               dif2mat(g2,g1)=dif2mat(g2,g1)-trafo(g,g2)*x*trafo(g,g1)
            enddo
         enddo
      enddo

! --- Subtract potential contribution for harmonic oscillator

      do g=1,gdim
         dif2mat(g,g) = dif2mat(g,g) + (hofm*ort(g))**2
      enddo
 

!-----------------------------------------------------------------------
! Calculate DVR matrix representation for first derivative.
!
!  The FBR representation of 0.5*(x*d/dx+d/dx*x) is antisymmetric 
!  tridiagonal with zero diagonal elements and analytically given as: 
!
!       <ho_{2n-1}|d/dx|ho_{2m-1}> = 
!           sqrt(2m*(2m+1)) delta_nm+1 - sqrt(2m-1)*(2m-2)) delta_nm-1
!
!-----------------------------------------------------------------------

      call zeromxd(dif1mat,gdim,gdim)
      
      do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis

         x = -sqrt(dble(2*g*(2*g+1)))

! --- Transform from FBR to DVR basis

         do g1=1,gdim
            do g2=1,gdim
               dif1mat(g2,g1)=dif1mat(g2,g1)+trafo(g+1,g2) &
                    *trafo(g,g1)*x-trafo(g,g2)*trafo(g+1,g1)*x
            enddo
         enddo
      enddo


! --- Shift grid points

      do g=1,gdim
         ort(g)=ort(g)+hoxeq
      enddo

      return
      end subroutine initrho



!-----------------------------------------------------------------------
!                Subroutine INITSIN
!                   Sine DVR
!-----------------------------------------------------------------------

      subroutine initsin(trafo,ort,dif2mat,dif1mat,gdim,xi,xf,ipar)

      implicit none

      integer(long)                                :: g1,g2
      integer(long), intent(in)                    :: gdim,ipar
      real(dop)                                    :: x,deltax,fac1,Pi
      real(dop), intent(in)                        :: xi,xf
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif2mat,dif1mat
      real(dop), dimension(gdim,gdim)              :: workr

      Pi     = 4.0_dop*atan(1.0_dop)
      deltax = (xf-xi)/dble(gdim-1)
      x      = gdim+1.0_dop
      fac1   = Pi/x

! --- Grid points are analytically given

      do g1=1,gdim
         ort(g1)=xi+(g1-1)*deltax
      enddo         

! --- DVR/FBR transformation matrix is analytically given

      do g1=1,gdim
         do g2=1,gdim
            trafo(g2,g1)= sqrt(2.0_dop/x)*sin(g2*g1*fac1)
         enddo
      enddo

! --- Second derivative in Sine-DVR basis (given analytically)

      do g1=1,gdim
         do g2=1,g1-1
            dif2mat(g2,g1)=-(Pi/deltax)**2 &
                 *2.0_dop*(-1.0_dop)**(g2-g1)/x**2 &
                 *sin(g2*fac1)*sin(g1*fac1) &
                 /(cos(g2*fac1)-cos(g1*fac1))**2
            dif2mat(g1,g2)=dif2mat(g2,g1)
         enddo
         dif2mat(g1,g1)=-(Pi/deltax)**2 &
              *(1.0_dop/3.0_dop+1.0_dop/(6.0_dop*x**2) &
              -1.0_dop/(2.0_dop*x**2*sin(g1*fac1)**2))
      enddo        

!-----------------------------------------------------------------------
! Derivative in FBR basis is analytically given for SIN-DVR
! but is not tridiagonal:
!   i)  all diagonal elements and all elements for which the row
!       and the column index are simultaneously odd or even
!       are zero
!   ii) for all other elements it holds:
!
!           d_ab = 4ab/(a**2-b**2) 1/(gdim+1) 1/deltax
!
!   iii) from ii) it follows that the matrix is still antisymmetric
!-----------------------------------------------------------------------
         
      if(ipar .eq. 0) then
         fac1 = 4.0_dop/((gdim+1)*deltax)
         do g1=1,gdim 
            do g2=1,g1-1
               if (mod(g2+g1,2).eq.0) then
                  dif1mat(g2,g1) = 0.0_dop
               else
                  dif1mat(g2,g1) = fac1*g2*g1/(g2**2-g1**2)
               endif
               dif1mat(g1,g2) = -dif1mat(g2,g1)
            enddo
            dif1mat(g1,g1) = 0.0_dop 
         enddo

!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of 0.5*(sin*d/dq + d/dq*sin)
!-----------------------------------------------------------------------
      elseif(ipar .eq. 1) then            !  "sdq"
         call zeromxd(dif1mat,gdim,gdim)
         fac1 = 0.25_dop*(Pi/((gdim+1)*deltax))  
         do g1=1,gdim-1
            dif1mat(g1,g1+1) = -fac1*(2*g1+1)
            dif1mat(g1+1,g1) =  fac1*(2*g1+1)
         enddo
      
! --- use special spin form, or transform from FBR to DVR basis
      elseif(ipar .eq. 2) then            !  "spin"
         dif1mat(1,1) = 0.0_dop
         dif1mat(1,2) =-0.5_dop
         dif1mat(2,1) = 0.5_dop
         dif1mat(2,2) = 0.0_dop
         dif2mat(1,1) = 0.0_dop
         dif2mat(1,2) = 0.5_dop
         dif2mat(2,1) = 0.5_dop
         dif2mat(2,2) = 0.0_dop
         return
      endif

!-----------------------------------------------------------------------
! Transform first derivative matrix from FBR to DVR form.
!-----------------------------------------------------------------------
      call qqxxdd(dif1mat,trafo,workr,gdim)
      call qqtxdd(trafo,workr,dif1mat,gdim)

      return
      end subroutine initsin


!-----------------------------------------------------------------------
!                Subroutine INITSINC
!                   Sinc DVR
!
!                                       
!                  (-1)^(g1-g2) | - pi^2/3         (i=j)
! (d^2/dx^2)_ij = --------------|
!                   deltax^2    | - 2/(g1-g2)^2    (i<>j)       
!
!  deltax, the gridspacing, i.e.,
!  x(g1)  = g1*deltax (i=1,...,gdim) (Colbert & Miller, JCP 96, 1982 (1992).)
!
!-----------------------------------------------------------------------

      subroutine initsinc(trafo,ort,dif2mat,dif1mat,gdim,xi,xf,ipar)

      implicit none

      integer(long)                                :: g1,g2
      integer(long), intent(in)                    :: gdim,ipar
      real(dop)                                    :: x,deltax,fac1,Pi
      real(dop), intent(in)                        :: xi,xf
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif2mat,dif1mat

      Pi     = 4.0_dop*atan(1.0_dop)
      deltax = (xf-xi)/dble(gdim-1)
      x      = gdim+1.0_dop
      fac1   = Pi/x

! --- Grid points are analytically given

      do g1=1,gdim
         ort(g1)=xi+(g1-1)*deltax
      enddo

! --- DVR/FBR transformation matrix is analytically given

      do g1=1,gdim
         do g2=1,gdim
            trafo(g2,g1)= sqrt(2.0_dop/x)*sin(g2*g1*fac1)
         enddo
      enddo

! --- Second derivative in Sinc-DVR basis (given analytically)

      do g1=1,gdim
         do g2=1,g1-1
            dif2mat(g2,g1)=-(1.0_dop/deltax)**2 &
                           *2.0_dop*(-1.0_dop)**(g2-g1) &
                           *(1.0_dop/(g2-g1)**2)
            dif2mat(g1,g2)=dif2mat(g2,g1)
         enddo
         dif2mat(g1,g1)=-(1.0_dop/deltax)**2*(Pi**2/3.0_dop) 
      enddo

!-----------------------------------------------------------------------
! First derivative in Sinc-DVR is given analytically 
!                 
!              |  0                                (i=j)
! (d/dx)_ij =  | 
!              | (1/deltax)((-1)**(i-j))/(i-j)     (i<>j)       
!
!-----------------------------------------------------------------------

      if(ipar .eq. 0) then
         do g1=1,gdim
            do g2=1,g1-1
               dif1mat(g2,g1)=(1.0_dop/deltax) &
                              *(-1.0_dop)**(g2-g1) &
                              *(1.0_dop/(g2-g1))
               dif1mat(g1,g2)=dif1mat(g2,g1)
            enddo
            dif1mat(g1,g1)=0.0_dop
         enddo
      endif

!-----------------------------------------------------------------------
! Transform first derivative matrix from FBR to DVR form.
!-----------------------------------------------------------------------
!      call qqxxdd(dif1mat,trafo,workr,gdim)
!      call qqtxdd(trafo,workr,dif1mat,gdim)

      return
      end subroutine initsinc


!-----------------------------------------------------------------------
!                Subroutine INITWSINC
!                   Wrapped Sinc DVR
!
!                                       
!                  (-1)^(g1-g2) | - pi^2/3 + 1/2*g1^2           (i=j)
! (d^2/dx^2)_ij = --------------|
!                   deltax^2    | - 2/(g1-g2)^2 + 2/(g1+g2)^2   (i<>j)       
!
!  deltax, the gridspacing, i.e.,
!  x(g1)  = g1*deltax (i=1,...,gdim) (Colbert & Miller, JCP 96, 1982 (1992).)
!
!-----------------------------------------------------------------------

      subroutine initwsinc(trafo,ort,dif2mat,dif1mat,gdim,xi,xf,ipar)

      implicit none

      integer(long)                                :: g1,g2
      integer(long), intent(in)                    :: gdim,ipar
      real(dop)                                    :: x,deltax,fac1,Pi
      real(dop), intent(in)                        :: xi,xf
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif2mat,dif1mat

      Pi     = 4.0_dop*atan(1.0_dop)
      deltax = (xf-xi)/dble(gdim-1)
      x      = gdim+1.0_dop
      fac1   = Pi/x

! --- Grid points are analytically given

      do g1=1,gdim
         ort(g1)=xi+(g1-1)*deltax
      enddo

! --- DVR/FBR transformation matrix is analytically given

      do g1=1,gdim
         do g2=1,gdim
            trafo(g2,g1)= sqrt(2.0_dop/x)*sin(g2*g1*fac1)
         enddo
      enddo

! --- Second derivative in Wrapped Sinc-DVR basis (given analytically)

      do g1=1,gdim
         do g2=1,g1-1
            dif2mat(g2,g1)=-(1.0_dop/deltax)**2 &
                           *2.0_dop*(-1.0_dop)**(g2-g1) &
                           *(1.0_dop/(g2-g1)**2 - 1.0_dop/(g2+g1)**2)
            dif2mat(g1,g2)=dif2mat(g2,g1)
         enddo
         dif2mat(g1,g1)=-(1.0_dop/deltax)**2 &
                        *(Pi**2/3.0_dop-1.0_dop/(2.0_dop*g1**2))
      enddo

!-----------------------------------------------------------------------
! First derivative in Wrapped Sinc-DVR is given analytically 
!                 
!              |  0                                (i=j)
! (d/dx)_ij =  | 
!              | (1/deltax)*(-1)**(i-j)*(1.0/(i-j) - 1.0/(i+j))    (i<>j)       
!
!-----------------------------------------------------------------------

      if(ipar .eq. 0) then
         do g1=1,gdim
            do g2=1,g1-1
               dif1mat(g2,g1)=(1.0_dop/deltax) &
                              *(-1.0_dop)**(g2-g1) &
                              *(1.0_dop/(g2-g1)-1.0_dop/(g2+g1))
               dif1mat(g1,g2)=dif1mat(g2,g1)
            enddo
            dif1mat(g1,g1) = 0.0_dop
         enddo
      endif

!-----------------------------------------------------------------------
! Transform first derivative matrix from FBR to DVR form.
!-----------------------------------------------------------------------
!      call qqxxdd(dif1mat,trafo,workr,gdim)
!      call qqtxdd(trafo,workr,dif1mat,gdim)

      return
      end subroutine initwsinc


!-----------------------------------------------------------------------
!                Subroutine INITEXP
!                  Exponential DVR
!-----------------------------------------------------------------------

      subroutine initexp(trafo,ort,dif2mat,dif1mat,exphin,exprueck,gdim, &
                         xi,xf,kmin,kdim)

      implicit none

      integer(long)                                    :: g,g1,g2,m
      integer(long), intent(in)                        :: gdim,kmin,kdim
      real(dop)                                        :: deltax,fac1,fac2,z,Pi
      real(dop), intent(in)                            :: xi,xf
      real(dop), dimension(gdim), intent(out)          :: ort
      real(dop), dimension(gdim,gdim), intent(out)     :: trafo,dif2mat,dif1mat

      complex(dop), dimension(gdim, kdim), intent(out) :: exphin
      complex(dop), dimension(kdim, gdim), intent(out) :: exprueck


      if( mod(gdim,2) .eq. 0 )  then
         write(6,*) ' ERROR in init1.f : '
         write(6,*) 'Number of grid points must be odd for exp-DVR'
         call stopnow
      end if

      Pi     = 4.0_dop*atan(1.0_dop)
      deltax = (xf-xi)/dble(gdim-1)
      fac1   = Pi/(gdim*deltax)
      fac2   = Pi/dble(gdim)
      z      = 2.0_dop*fac2


! --- Grid points are analytically given

      do g=1,gdim
         ort(g)=xi+(g-1)*deltax
      enddo         

!! --- DVR/FBR transformation matrix is analytically given, however,
!! --- it is complex. Thus the correct trafo is outcommented and 
!! --- trafo is zeroed. Trafo is replaced by exphin and exprueck.
!! --- Note: Hier, opposite to fft, exphin: x->p, exprueck: p->x

      call zeromxd(trafo,gdim,gdim)

!     do g1=1,gdim
!        do g2=1,gdim
!           trafo(g2,g1)= exp(fac2*(0._dop,2._dop)*g1*g2)/sqrt(gdim)
!        enddo
!     enddo

      do g1=1,kdim
         m=g1-1+kmin
         do g2=1,gdim
            exphin(g2,g1)   = exp(dcmplx(0.0_dop,-m*(g2-1)*z)) &
                                 /sqrt(dble(gdim))
            exprueck(g1,g2) = exp(dcmplx(0.0_dop, m*(g2-1)*z)) &
                                 /sqrt(dble(gdim))
         enddo
      enddo

! --- Second derivative in Exponential-DVR basis (given analytically)

      do g1=1,gdim
         do g2=1,g1-1
            g = g1-g2
            dif2mat(g1,g2) = -2.0_dop*fac1**2 *( (-1)**g ) &
                         * cos(fac2*g)/(sin(fac2*g))**2
            dif2mat(g2,g1) =  dif2mat(g1,g2)   
         enddo
         dif2mat(g1,g1) = -fac1**2 *(gdim**2-1)/3.0_dop
      enddo

! --- First derivative in Exponential-DVR basis (given analytically)

      do g1=1,gdim
         do g2=1,g1-1
            g = g1-g2
            dif1mat(g1,g2) =  fac1*( (-1)**g ) /sin(fac2*g)
            dif1mat(g2,g1) =  -dif1mat(g1,g2)   
         enddo
         dif1mat(g1,g1) = 0.0_dop
      enddo

      return
      end subroutine initexp


!#######################################################################

!***********************************************************************
!
!                   Subroutine INITFFT
!
! calls fftinit
!
! Initialises the arrays ort,hin,rueck,fftfak,fftp for FFT 
!
! variables passed for one mode:
!  ort   : FFT grid points
!  hin   : array for position --> momentum FFT transformation
!  rueck : array for momentum --> position FFT transformation
!  fftfak: FFT decomposition factors 
!  fftp  : FFT momentum array
!  gdim  : number of FFT grid points
!  xi    : First point of FFT grid
!  xf    : last point  of FFT grid
!
!   11/96 AJ
!
!***********************************************************************

      subroutine initfft(ort,fftp,gdim,xi,xf)

      implicit none

      integer(long)                               :: g
      integer(long), intent(in)                   :: gdim
      real(dop)                                   :: deltax,length,deltap,pmax,Pi
      real(dop), intent(in)                       :: xi,xf
      real(dop), dimension(0:gdim-1), intent(out) :: ort,fftp

      Pi=4.0_dop*atan(1.0_dop)

!-----------------------------------------------------------------------
! Define grid
!-----------------------------------------------------------------------
      deltax = (xf-xi)/(gdim-1)
      do g=0,gdim-1
         ort(g)=xi+g*deltax
      enddo

!-----------------------------------------------------------------------
! Calculate momentum grid
!-----------------------------------------------------------------------
      length = gdim*deltax
      deltap = 2.0_dop*Pi/length
      pmax   = Pi/deltax 
      if (mod(gdim,2).eq.0) then
         do g=0,gdim/2-1
            fftp(g)=g*deltap
            fftp(g+gdim/2)=-pmax+g*deltap
         enddo
      else
         do g=0,gdim/2
            fftp(g)=g*deltap
         enddo
         do g=0,gdim/2-1
            fftp(g+gdim/2+1)=-pmax+(g+0.5_dop)*deltap
         enddo
      endif      

      return
      end subroutine initfft   


!#######################################################################

!-----------------------------------------------------------------------
!                Subroutine INITCOS
!                  Cosine DVR
!-----------------------------------------------------------------------

      subroutine initcos(trafo,ort,dif2mat,dif1mat,gdim, &
                         xi,xf,message)

      implicit none

      integer(long)                                :: g,g1,g2,error
      integer(long), intent(in)                    :: gdim
      real(dop)                                    :: fac1,x,br,x0,pi
      real(dop), intent(inout)                     :: xi,xf
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif1mat,dif2mat
      real(dop), dimension(gdim,gdim)              :: workr
      character(len=*), intent(out)                :: message

      Pi = 4.0_dop*atan(1.0_dop)
      br = (xf-xi)*gdim/dble(gdim-1)
      x0 = xi - br/(2.0_dop*gdim)
      xi = x0 + br/(2.0_dop*gdim)
      xf = x0 + br - br/(2.0_dop*gdim)

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos(theta) operator (note that the diagonal elements are always zero).
!-----------------------------------------------------------------------


! --- Set diagonal elements of X_mn to 0

      call zerovxd(ort,gdim)

! --- Set subdiagonal elements of X_nm = <n|cos|m>.
! --- Here we actually use -X_nm, such that later the (transformend) 
! --- gridpoints appear in ascending order.

      do g=2,gdim
         dif2mat(g,1)=-0.5_dop
      enddo
      dif2mat(1,1)=-0.5_dop*sqrt(2.0_dop)

! --- Diagonalize cos(theta)-matrix to obtain "ort" and "trafo".
! --- dif2mat contains the sub-diagonal and dif1mat serves as a work array.

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising cos(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
!  ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g)
            enddo
         end if
      enddo

!-----------------------------------------------------------------------
! Rescale grid points cos(theta) --> theta
!  acos returns the arc cosine in the range Pi to 0.
! This is why we diagonalised -Xnm.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=x0+acos(-ort(g))*(br/Pi)
      enddo


!-----------------------------------------------------------------------
!  Calculate second derivative matrix in corresponding DVR basis
!  dif2mat in cos-FBR basis is diagonal and analytically known
!   --> transform to DVR basis
!-----------------------------------------------------------------------

      call zeromxd(dif2mat,gdim,gdim)

      do g=1,gdim
         x=((g-1)*(Pi/br))**2

! --- Transform to DVR basis

         do g1=1,gdim
            do g2=1,gdim
               dif2mat(g2,g1)=dif2mat(g2,g1)-trafo(g,g2)*x*trafo(g,g1)
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of 0.5*(sin*d/dq + d/dq*sin)
!-----------------------------------------------------------------------


      call zeromxd(dif1mat,gdim,gdim)

      fac1 = 0.25_dop*(Pi/br)  
      do g=2,gdim-1
         dif1mat(g,g+1) = -fac1*(2*g-1)
         dif1mat(g+1,g) =  fac1*(2*g-1)
      enddo
      dif1mat(1,2) = -fac1*sqrt(2.0_dop)
      dif1mat(2,1) =  fac1*sqrt(2.0_dop)

      call qqxxdd(dif1mat,trafo,workr,gdim)
      call qqtxdd(trafo,workr,dif1mat,gdim)


      return
      end subroutine initcos


!#######################################################################

!-----------------------------------------------------------------------
!
!                   Subroutine initsphfbr
!                 spherical  harmonics  FBR
!
!
! Initialises the arrays kinsph,jsph,msph for a FBR representation
! in a spherical harmonics basis set.
!
! variables passed for one mode:
!  kinsph : FBR kinetic energy matrix (diagonal matrix)
!  jsph   : azimutal quantum numbers for the basis functions
!  msph   : magnetic quantum numbers for the basis functions
!  gdim   : number of FBR basis functions
!  jmax   : maximum azimutal quantum number for spherical harmonics
!  sphtyp : type of spherical harmonics basis :  1: j,m=all (abs(m)<= jmax)
!           2: j=selected and m=all (j=0,2.. or 1,3..)
!           3: j=selected and m=truncated
!                   (j=0,2.. or 1,3..) and (abs(m) <= mmax)
!           4: j=all  and m=truncated  (abs(m) <= mmax)
!  mmax   : maximum magnetic quantum number for truncated basis
!  mincr  : increment of m's, starting from mmax, if m=truncated
!  f      : fth mode
!
!   08/97 MCH
!
!-----------------------------------------------------------------------

      subroutine initsphfbr(gdim,kinsph,jzsph,jzzsph,jsph,msph, &
                            jmax,sphtyp,mmax,mincr,mmin,ilog)

      implicit none

      integer(long)                               :: m,j0,j,g,xx
      integer(long), intent(in)                   :: gdim,jmax,sphtyp,ilog,mmax,mincr,mmin
      integer(long), dimension(gdim), intent(out) :: jsph,msph
      real(dop), dimension(gdim), intent(out)     :: kinsph,jzsph,jzzsph


!  quantum numbers arrays jsph,msph are defined
!  according to the type of basis.
      g=0
      if (sphtyp.le.2)  then
         if (sphtyp.eq.1) then
            j0=0
         else
            j0=mod(jmax,2)
         endif

         do j=j0,jmax,sphtyp
            do m=-j,j,1
               g=g+1
               jsph(g)=j
               msph(g)=m
            enddo
         enddo

      else

         if (sphtyp.eq.3) then
           xx=2
           j0=mod(jmax,2)
         else
           xx=1
           j0=0
         endif

         do j=j0,jmax,xx
            do m=mmin,mmax,mincr
               if (iabs(m).le.min(mmax,j)) then
                  g=g+1
                  jsph(g)=j
                  msph(g)=m
               endif
            enddo
         enddo

      endif

! to debug
      if (gdim.ne.g) then
         write(ilog,'(a,2i5)') '+++ problem in initsphfbr ',gdim,g
         call stopnow
      endif

! Construction of the j^2, j_z and j_z^2 matrices in the FBR-representation
! diagonal matrices stored as an one-dimensional arrays.

      do g=1,gdim
         kinsph(g)=jsph(g)*(jsph(g)+1)
         jzsph (g)=msph(g)
         jzzsph(g)=msph(g)**2
      enddo

      return
      end subroutine initsphfbr

!#######################################################################

!-----------------------------------------------------------------------
!                Subroutine INITLAGU
!                  Laguerre DVR
! phi_n(x) = (x/n)**(1/2) * exp(-x/2) * L^{1}_{n-1}(x); (alpha=1)
! phi_n(x) = A_n * x**(alpha/2) * exp(-x/(2*breite)) * 
!                                 L^{alpha}_{n-1}(x/breite)
!-----------------------------------------------------------------------

      subroutine initlagu(trafo,ort,dif2mat,dif1mat,gdim, &
                          x0,breite,xi,xf,typ,icut,message,basis)

      implicit none

      integer(long)                                :: g,g1,error,alpha
      integer(long), intent(in)                    :: gdim,typ,icut,basis
      real(dop)                                    :: eigval,cutval
      real(dop), intent(out)                       :: x0
      real(dop), intent(inout)                     :: xi,xf,breite
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif1mat,dif2mat
      real(dop), dimension(gdim,gdim)              :: workr
      character(len=*), intent(out)                :: message
      character(len=5)                             :: clgu


!-----------------------------------------------------------------------
!  Calculate second derivative matrix in corresponding DVR basis
!  (d^2/dx^2 - c/x^2) in Lagu-FBR basis is analytically evaluated. 
!  c = -1/4, 0, 3/4, 2  for alpha = 1,2,3,4 .
!  The centrifugal term c/x^2 is finally removed in DVR representation. 
!-----------------------------------------------------------------------
      alpha=0
      if(basis.eq.15) then
         alpha = 1
         clgu  = 'Lagu1'
         do g=1,gdim
            trafo(g,g) = 0.25_dop*(1-2*g)/breite**2
            do g1=1,g-1
               trafo(g,g1) = -0.5_dop*sqrt(dble(g1)/dble(g))*g1/breite**2
               trafo(g1,g) = trafo(g,g1)
            enddo
         enddo
      elseif(basis.eq.16) then
         alpha = 2
         clgu  = 'Lagu2'
         do g=1,gdim
            trafo(g,g) = dble(1-4*g)/(12.0_dop*breite**2)
            do g1=1,g-1
               trafo(g,g1) = -sqrt(dble(g1*(g1+1))/dble(g*(g+1))) &
                   * dble(2*g1+1)/(6.0_dop*breite**2)
               trafo(g1,g) = trafo(g,g1)
            enddo
         enddo
      elseif(basis.eq.17) then
         alpha = 3
         clgu  = 'Lagu3'
         do g=1,gdim
            trafo(g,g) = -g/(4.0_dop*breite**2)
            do g1=1,g-1
               trafo(g,g1) = sqrt(dble(g1*(g1+1)*(g1+2))/ &
                                  dble(g*(g+1)*(g+2))) &
                   * dble(-g1-1)/(4.0_dop*breite**2)
               trafo(g1,g) = trafo(g,g1)
            enddo
         enddo
      elseif(basis.eq.18) then
         alpha = 4
         clgu  = 'Lagu4'
         do g=1,gdim
            trafo(g,g) = -dble(4*g+1)/(20.0_dop*breite**2)
            do g1=1,g-1
               trafo(g,g1) = sqrt((dble(g1*(g1+1))/dble(g*(g+1)))* &
                          (dble((g1+2)*(g1+3))/dble((g+2)*(g+3)))) &
                   * dble(-2*g1-3)/(10.0_dop*breite**2)
               trafo(g1,g) = trafo(g,g1)
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! In order to remove extremly large eigenvalues of dif2mat, this
! matrix is diagonalized, the largest (negative) eigenvalues are 
! modified and the diagonal matrix is transformed back.
!-----------------------------------------------------------------------
! get all eigenvalues, stored in "ort" and the corresponding normalized
! eigenvectors of the second derivatives matrix "trafo". 
! On output, the eigenvectors are store in "trafo".
!-----------------------------------------------------------------------
      g = gdim*gdim
      call dsyev('V','U',gdim,trafo,gdim,ort,workr,g,error)

      if (error.ne.0) then
         message = 'Error in diagonalising dif2mat (Lagu)'
         call errormsg
      endif

!---- Modify eigenvalues.
      eigval=-ort(1)
      cutval=-ort(icut+1)
      do g = 1, icut
         ort(g) = ort(icut+1)
      enddo

!-----------------------------------------------------------------------
! dif2mat = trafo * ort * trafo-transposed.
!-----------------------------------------------------------------------
      call qdxxdd (trafo,ort,workr,gdim)
      call qqxtdd (trafo,workr,dif2mat,gdim)
      
!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! x operator (note that the diagonal elements are not zero).
!-----------------------------------------------------------------------

! --- Set diagonal elements of X_mn 

      do g=1,gdim
         ort(g) = 2*g + alpha -1
      enddo

! --- Set subdiagonal elements of X_mn.

      do g=1,gdim-1
         workr(g,1) = -sqrt(dble(g*(g+alpha)))
      enddo

! --- Diagonalize x-matrix to obtain "ort" and "trafo".

      call DSTEQR('I',gdim,ort,workr,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising X-matrix (Lagu)'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
!  ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g)
            enddo
         end if
      enddo

!-------------------------------------------------------------------------
!   Transform to DVR basis
!-------------------------------------------------------------------------
      call qqxxdd(dif2mat,trafo,workr,gdim)
      call qqtxdd(trafo,workr,dif2mat,gdim)

!-------------------------------------------------------------------------
! Remove the artifical centrifugal potential c/x^2  (c = -1/4, 0, 3/4, 2).
!-------------------------------------------------------------------------
      if(basis.eq.15) then
         do g=1,gdim
            dif2mat(g,g) = dif2mat(g,g) - 0.25_dop/(ort(g)*breite)**2
         enddo
      elseif(basis.eq.17) then
         do g=1,gdim
            dif2mat(g,g) = dif2mat(g,g) + 0.75_dop/(ort(g)*breite)**2
         enddo
      elseif(basis.eq.18) then
         do g=1,gdim
            dif2mat(g,g) = dif2mat(g,g) + 2.0_dop/(ort(g)*breite)**2
         enddo
      endif

!-------------------------------------------------------------------------
!  Calculate first derivative matrix in corresponding DVR basis
!  d/dx in Lagu-FBR basis is analytically known  
!   --> transform to DVR basis
!-------------------------------------------------------------------------
      if(basis.eq.15) then
         do g=1,gdim
            dif1mat(g,g) = 0.0_dop
            do g1=1,g-1
               dif1mat(g,g1) = 0.5_dop*sqrt(dble(g1)/dble(g))/breite
               dif1mat(g1,g) = -dif1mat(g,g1)
            enddo
         enddo
      elseif(basis.eq.16) then
         do g=1,gdim
            dif1mat(g,g) = 0.0_dop
            do g1=1,g-1
               dif1mat(g,g1) = 0.5_dop*sqrt(dble(g1*(g1+1))/dble(g*(g+1))) &
                                /breite
               dif1mat(g1,g) = -dif1mat(g,g1)
            enddo
         enddo
      elseif(basis.eq.17) then
         do g=1,gdim
            dif1mat(g,g) = 0.0_dop
            do g1=1,g-1
               dif1mat(g,g1) = 0.5_dop*sqrt(dble(g1*(g1+1)*(g1+2))/ &
                                      dble(g*(g+1)*(g+2)))/breite
               dif1mat(g1,g) = -dif1mat(g,g1)
            enddo
         enddo
      elseif(basis.eq.18) then
         do g=1,gdim
            dif1mat(g,g) = 0.0_dop
            do g1=1,g-1
               dif1mat(g,g1) =0.5_dop*sqrt((dble(g1*(g1+1))/dble(g*(g+1))) &
                      * (dble((g1+2)*(g1+3))/dble((g+2)*(g+3))))/breite
               dif1mat(g1,g) = -dif1mat(g,g1)
            enddo
         enddo
      endif

      call qqxxdd(dif1mat,trafo,workr,gdim)
      call qqtxdd(trafo,workr,dif1mat,gdim)

!-----------------------------------------------------------------------
! If xi-xf or x0-xf is set
!-----------------------------------------------------------------------
      if(typ.gt.0) then
         if(typ.eq.1) then
            breite = (xf-xi)/(ort(gdim)-ort(1))
            x0 = xi - ort(1)*breite
         elseif(typ.eq.2) then
            breite = (xf-x0)/ort(gdim)
         endif
         do g=1,gdim
            do g1=1,gdim
               dif1mat(g,g1) = dif1mat(g,g1)/breite
               dif2mat(g,g1) = dif2mat(g,g1)/breite**2
            enddo
         enddo
         eigval=eigval/breite**2
         cutval=cutval/breite**2
      endif

!-----------------------------------------------------------------------
! Print cut information to log file.
!-----------------------------------------------------------------------
      write(2,'(/2a,i3,3(a,1pe10.3))') clgu,': icut =',icut, &
              ', EigVal:',eigval, ', cut :', cutval, &
              ', EigVal/cut:',eigval/cutval

!-----------------------------------------------------------------------
!  Shift grid points 
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g) = ort(g)*breite + x0
      enddo

      xi = ort(1)
      xf = ort(gdim)

      return
      end subroutine initlagu
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!                Subroutine INITLOB
!                   Lobatto DVR
!-----------------------------------------------------------------------

      subroutine initlob(trafo,ort,dif2mat,dif1mat,gdim,xi,xf,noxi,noxf)

      implicit none

      integer(long)                                :: g,e,e1,npnt,g0
      integer(long), intent(in)                    :: gdim,noxi,noxf
      real(dop), intent(in)                        :: xi,xf
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif2mat,dif1mat
      real(dop), dimension(:), allocatable         :: tmport,weight
      real(dop), dimension(:,:), allocatable       :: derlobmat

! get no. of points
      npnt = gdim
      if (noxi .eq. 1) npnt = npnt+1
      if (noxf .eq. 1) npnt = npnt+1

! shift of zero if first Lobatto point not on SPF grid
      g0=0
      if (noxi .eq. 1) g0=1

      allocate(tmport(1:npnt))
      allocate(weight(1:npnt))
      allocate(derlobmat(1:npnt,1:npnt))

! --- Get points and weights. 

      call LobattoAbsWeights(tmport,weight,npnt,xi,xf)
    
      ort(1:gdim) = tmport(g0+1:g0+gdim)

! --- DVR weghts are stored in trafo matrix 

      do e=1,gdim
         do g=1,gdim
            trafo(e,g)= weight(g0+g)
         enddo
      enddo

! --- First derivative matrix

      do e=1,npnt
         do g=1,npnt
            call derLobatto(derLobMat(g,e),e-1,g-1,npnt-1,tmport,weight)
            derLobMat(g,e) = derLobMat(g,e) * 1.0d0/sqrt(weight(g))
         end do
      end do
      dif1mat(1:gdim,1:gdim) = derlobmat(g0+1:g0+gdim,g0+1:g0+gdim)

! --- Second derivative matrix

      dif2mat = 0.0_dop
      do e=1,gdim
         do e1=1,gdim
            do g=1,npnt
               dif2mat(e1,e) = dif2mat(e1,e) - &
                  weight(g)*derlobmat(g,g0+e1)*derlobmat(g,g0+e)
            end do
         end do
      end do

      deallocate(tmport,weight,derlobmat)

      return
      end subroutine initlob

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!
!
!     subroutine LobattoAbsWeights
!
!     Calculates weights and abscissa of Gauss-Lobatto quadrature using 
!     subroutine from Manolopoulos paper:
!
!     Manolopoulos, D. E., Lobatto Shape Functions, Numerical Grid 
!     Methods and their Applications to Schrodingers Equation, 1993 412 
!     57-68
!
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!

      subroutine LobattoAbsWeights(ptr,rw,npntP1,rmin,a0)

      implicit none

      real(dop)           :: gridlength,shift,scalelobatto,weight
      real(dop)           :: z,p1,p2,p3,pi
      real(dop), intent(in)  :: rmin,a0
      real(dop), intent(out) :: rw(1:npntP1),ptr(1:npntP1)
      integer             :: L,k,j,i
      integer, intent(in) :: npntP1

      gridlength = a0-rmin
      L = (npntP1+1)/2
      shift = 0.5d0*(a0+rmin)
      scalelobatto = 0.5d0*gridlength
      weight = gridlength/(npntP1*(npntP1-1d0))
      pi = dacos(-1d0)

      ptr(1)=rmin
      rw(1)=weight

      do k=2,L
         z=cos(pi*(4d0*k-3d0)/(4d0*npntP1-2d0))
         do i=1,51
            p2=0d0;
            p1=1d0;
            do j=1,npntP1-1
               p3=p2
               p2=p1
               p1=((2d0*j-1d0)*z*p2-(j-1d0)*p3)/j
            end do
            p2=(npntP1-1d0)*(p2-z*p1)/(1d0-z*z)
            p3=(2*z*p2-npntP1*(npntP1-1)*p1)/(1d0-z*z)
            z=z-p2/p3
         end do
         ptr(k)=shift-scalelobatto*z
         ptr(npntP1+1-k)=shift+scalelobatto*z
         rw(k)=weight/(p1*p1)
         rw(npntP1+1-k)=rw(k)
      end do

      ptr(npntP1)=a0
      rw(npntP1)=weight

      return

      end subroutine LobattoAbsWeights

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!
!
!     subroutine derLobatto
!
!     Calculates u'_n(r_eta)
!
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%!

      subroutine derLobatto(derLob,n,eta,npnt,rpt,w)

      implicit none

      integer, intent(in) :: n,npnt,eta
      real(dop), intent(in)  :: w(0:npnt), rpt(0:npnt)
      real(dop), intent(out) :: derLob
      real(dop)           :: signrecorder
      real(dop)           :: factor(0:npnt)
      integer             :: Nhalf
      integer             :: j

      Nhalf=(npnt)/2

      if (n==eta) then
         if (n==0) then
            derLob=-1d0/(2d0*w(n))
         else if (n==npnt) then
            derLob=1d0/(2d0*w(n))
         else
            derLob=0d0;
         end if
      else
         derLob=1d0/((rpt(n)-rpt(eta)))
         signrecorder = 1.0
         do j=0,npnt
            if ((j .ne. n) .and. (j .ne. eta)) then
               factor(j) = (rpt(eta)-rpt(j))/(rpt(n)-rpt(j))
               signrecorder = signrecorder * (factor(j) / abs(factor(j)))
               factor(j) = abs(factor(j))
            else
               factor(j) = 1.0d0
            end if
         end do
         do j = 0,npnt
            derLob = derLob * factor(j)
         end do
         derLob = derLob * signrecorder
      end if

      return

      end subroutine derLobatto

!#######################################################################
!
!  INITGWPGRID
!
! Initialises the array ort for the auxilliary grid used for GWP 
! analysis
!
! variables passed for one mode:
!  ort   : grid points
!  gdim  : number of grid points
!  xi    : First point of grid
!  xf    : last point  of grid
!
!#######################################################################

      subroutine initgwpgrid(ort,gdim,xi,xf)

      implicit none

      integer(long)                           :: g
      integer(long), intent(in)               :: gdim
      real(dop)                               :: deltax
      real(dop), intent(in)                   :: xi,xf
      real(dop), dimension(gdim), intent(out) :: ort

!-----------------------------------------------------------------------
! Define grid
!-----------------------------------------------------------------------
      deltax = (xf-xi)/(gdim-1)
      do g=1,gdim
         ort(g)=xi+(g-1)*deltax
      enddo

      return
      end subroutine initgwpgrid    
!-----------------------------------------------------------------------

      end module init1mod

