!***********************************************************************
!
!                Module INIT2
!
! Contents: initleg,initleg,initkleg,initklegr,initk
!
! calls lapack-routine DSTEQR to diagonalise tri-diagonal matrices.
!
! Initialises the arrays trafo, dif2mat and dif1mat for a DVR basis set
!
! variables passed:
!  trafo  : DVR/FBR transformation matrices
!  ort    : DVR grid points
!  dif1mat  : DVR first  derivative matrix
!  dif2mat  : DVR second derivative matrix
!  gdim   : number of DVR grid points for this mode
!  xi     : Value of first DVR grid point
!  xf     : Value of last  DVR grid point
!  ibleg  : symmetry parameter. 0->no sym, 1->ungerade, 2->gerade
!  blz    : magnetic quantum number of associated Legendre fkt. (m=blz).
!  workr  : work array
!
!***********************************************************************

      module init2mod

use decimal, only: long, dop
use global
use logdat
use op1lib,  only: zeromxd, zerovxd, cpvxd
use schmidtorthomod, only: schmidtorthod


      implicit none
      private
      public :: initleg,initlegr,initkleg,initklegr,initk,initextern
      contains

!------------------------------------------------------------------------
!                Subroutine INITLEG
!                   Legendre DVR
!------------------------------------------------------------------------

      subroutine initleg(trafo,ort,dif2mat,dif1mat,gdim &
                         ,blz,ibleg,message)

      implicit none

      integer(long)                                :: g,g1,g2,j,j1,error,jmin
      integer(long), intent(in)                    :: gdim,blz,ibleg
      real(dop)                                    :: x
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: trafo,dif2mat,dif1mat
      character(len=*), intent(out)                :: message

!-----------------------------------------------------------------------
! Use special formulas for the gerade (l=even) or ungerade (l=odd) case.
!-----------------------------------------------------------------------

      if( ibleg .gt. 0 )  goto 100


! ********************** no symmetry ***********************************

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos(theta) operator (note that the diagonal elements are always zero).
!-----------------------------------------------------------------------


! --- Set diagonal elements of X_mn to 0

      call zerovxd(ort,gdim)

! --- Set subdiagonal elements of X_mn.

      do g=2,gdim
         j=g+blz-1
         dif2mat(g-1,1)=sqrt((j**2-blz**2)/(4.0_dop*j*j-1.0_dop))
      enddo

! --- Diagonalize cos(theta)-matrix to obtain "ort" and "trafo".
! --- dif2mat contains the sub-diagonal and dif1mat serves as a work array.

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising cos(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
!  ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g)
            enddo
         end if
      enddo

!-----------------------------------------------------------------------
!  Calculate second derivative matrix in corresponding DVR basis
!  T in Leg-FBR basis is analytically known  --> transform to DVR basis
!  dif2mat := sin^-1(theta) * d/dtheta * sin(theta) * d/dtheta
!              - m^2/sin^2(theta).  (N.B.: m = blz)
!-----------------------------------------------------------------------

      call zeromxd(dif2mat,gdim,gdim)

! ---  -(1/2*I)*Kinetic-energy in Leg-FBR basis (diagonal)

      do g=1,gdim
         x=(g+blz)*(g+blz-1)

! for real wavepacket propagation
         if (lrealwp) then
            dif2mat(g,g) = -x
         else

! --- Transform to DVR basis

            do g1=1,gdim
               do g2=1,gdim
                  dif2mat(g2,g1)=dif2mat(g2,g1)-trafo(g,g2)*x*trafo(g,g1)
               enddo
            enddo
         endif
      enddo

!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of d/dtheta * sin(theta)
!
!  Note: The FBR representation is antisymmetric tridiagonal with zero
!        diagonal elements and  may be symbolically written as:
!
!              D_ab^FBR = x_a [ delta(a,b+1) - delta(a,b-1) ]
!
!        with x_a being the subdiagonal element. The FBR-DVR trafo
!        below makes use of this special form.
!
!-----------------------------------------------------------------------

      call zeromxd(dif1mat,gdim,gdim)

      do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis (j is really j+1).
         j=g+blz
         x=j*sqrt((j**2-blz**2)/(4.0_dop*j**2-1.0_dop))

! --- Transform from FBR to DVR basis

         do g1=1,gdim
            do g2=1,gdim
               dif1mat(g2,g1)=dif1mat(g2,g1)+trafo(g+1,g2) &
                    *trafo(g,g1)*x-trafo(g,g2)*trafo(g+1,g1)*x
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! Rescale grid points cos(theta) --> theta
!  acos returns the arc cosine in the range Pi to 0, i.e.
!  the largest value correspond to g=1.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=acos(ort(g))
      enddo

      return

! ************************  symmetry  **********************************
 100  continue

!-----------------------------------------------------------------------
!       ibleg=1 -> l=ungerade (independent of m),
!       ibleg=2 -> l=gerade (independent of m)
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos^2(theta) operator (note that the diagonal elements are not zero).
!-----------------------------------------------------------------------

!.... set minimum value of j
      jmin = blz + mod(blz+ibleg,2)

! --- Set diagonal elements for X_mn.

      do g = 1, gdim
         j = 2*(g-1) + jmin
         j1 = j + 1
         ort(g) = (j**2-blz**2)/(4.0_dop*j*j-1.0_dop) &
                + (j1**2-blz**2)/(4.0_dop*j1*j1-1.0_dop)
      enddo

! --- Set super-diagonal elements for X_mn.  (j is really j+2).

      do g=2,gdim
         j = 2*(g-1) + jmin
         j1 = j - 1
         dif2mat(g-1,1) = sqrt((j**2-blz**2)/(4.0_dop*j*j-1.0_dop) &
                           * (j1**2-blz**2)/(4.0_dop*j1*j1-1.0_dop))
      enddo

! --- Diagonalize cos^2(theta) matrix to obtain "ort" and "trafo".

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising cos^2(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Change the sign of the eigenvector if necessary, and thus
!  ensure that the DVR-weights are all positive. (See genphi1.f)
!-----------------------------------------------------------------------
      do g=1,gdim
         if(trafo(1,g) .lt. 0.0_dop ) then
            do g1=1,gdim
               trafo(g1,g) = -trafo(g1,g)
            enddo
         end if
      enddo

!-----------------------------------------------------------------------
!  Calculate second derivative matrix in corresponding DVR basis
!  T in Leg-FBR basis is analytically known  --> transform to DVR basis
!  dif2mat := sin^-1(theta) * d/dtheta * sin(theta) * d/dtheta
!              - m^2/sin^2(theta).  (N.B.: m = blz)
!-----------------------------------------------------------------------

      call zeromxd(dif2mat,gdim,gdim)

! ---  -(1/2*I)*Kinetic-energy in Leg-FBR basis (diagonal)

      do g=1,gdim
         j = 2*(g-1) + jmin
         x = j*(j+1)

! for real wavepacket propagation
         if (lrealwp) then
            dif2mat(g,g) = -x
         else

! --- Transform to DVR basis

            do g1=1,gdim
               do g2=1,gdim
                  dif2mat(g2,g1)=dif2mat(g2,g1)-trafo(g,g2)*x*trafo(g,g1)
               enddo
            enddo
         endif
      enddo


!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of :
!  0.5*( cos(theta)*d/dtheta*sin(theta)  + d/dtheta*sin(theta)*cos(theta) )
!
!  Note: The FBR representation is antisymmetric tridiagonal with zero
!        diagonal elements and  may be symbolically written as:
!
!              D_ab^FBR = x_a [ delta(a,b+1) - delta(a,b-1) ]
!
!        with x_a being the subdiagonal element. The FBR-DVR trafo
!        below makes use of this special form.
!
!-----------------------------------------------------------------------

      call zeromxd(dif1mat,gdim,gdim)

      do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis

         j = 2*(g-1) + jmin
         x = (j+1.5_dop)*sqrt(((j+1)**2-blz**2)/(4.0_dop*(j+1)**2-1.0_dop)) &
                      *sqrt(((j+2)**2-blz**2)/(4.0_dop*(j+2)**2-1.0_dop))

! --- Transform from FBR to DVR basis

         do g1=1,gdim
            do g2=1,gdim
               dif1mat(g2,g1)=dif1mat(g2,g1)+trafo(g+1,g2) &
                    *trafo(g,g1)*x-trafo(g,g2)*trafo(g+1,g1)*x
            enddo
         enddo
      enddo


!-----------------------------------------------------------------------
! Rescale grid points cos^2(theta) --> theta
! Note: the grid points lie in the interval (0,Pi/2)
!       the largest value correspond to g=1.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=acos(sqrt(ort(g)))
      enddo

      return
      end subroutine initleg

!#######################################################################


!-----------------------------------------------------------------------
!                Subroutine INITKLEG
!               Extended Legendre DVR
!------------------------------------------------------------------------

      subroutine initkleg(trafo,ort,dif4mat,dif3mat,dif2mat,dif1mat, &
                          gdim,kdim,kmin,kmin2,ibleg,message)

      implicit none

      integer(long)                                     :: g,gg,g1,g2,j,j1,error,m,m1,jmin,jmax
      integer(long), intent(in)                         :: gdim,kmin,kdim,ibleg,kmin2
      real(dop)                                         :: x,w0
      real(dop), dimension(gdim), intent(out)           :: ort
      real(dop), dimension(gdim)                        :: w
      real(dop), dimension(gdim,gdim,kdim), intent(out) :: trafo,dif4mat,dif3mat,dif2mat,dif1mat
      real(dop), dimension(2*gdim)                      :: workr
      character(len=*), intent(out)                     :: message
      logical(kind=4)                                   :: leglog


!-----------------------------------------------------------------------
! kmin2=min(abs(k)).
! ibleg=0 -> no sym, ibleg=1 -> l=ungerade, ibleg=2 l=gerade.
! Use special formulas for the gerade (l=even) or ungerade (l=odd) case.
!-----------------------------------------------------------------------

      if( ibleg .gt. 0 )  goto 100



! ********************** no symmetry ***********************************

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos(theta) operator (note that the diagonal elements are always zero).
!-----------------------------------------------------------------------


! --- Set diagonal elements of X_mn to 0

      call zerovxd(ort,gdim)

! --- Set subdiagonal elements of X_mn.

      do g=2,gdim
! ---    |m|<=j<=|m|+gdim-1
         j=g+kmin2-1
         dif2mat(g-1,1,1)=sqrt((j**2-kmin2**2)/(4.0_dop*j*j-1.0_dop))
      enddo

! --- Diagonalize cos(theta)-matrix to obtain "ort" and trafo(i,j,1).

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)
      if (error.ne.0) then
         message = 'Error in diagonalising cos(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
!  Calculate the weights:
!  w(g)=sqrt(2^(m+1)*m!/(2m+1)!!)*sin^(-m)(theta_g)*trafo(1,g,1),
!  where theta_g=acos(ort(g))  and   m=kmin2.
!-----------------------------------------------------------------------
      do g=1,gdim
         w0 = sqrt(2.0_dop)
         x  = sin(acos(ort(g)))
         do m=1,kmin2
            w0 = w0 * sqrt(2.0_dop*m/dble(2*m+1))/x
         enddo
         w(g) = w0 * abs(trafo(1,g,1))
      enddo

!----------------------------------------------------------------------
!  Calculate matrix trafo :
!  trafo(l,g,m)=w(g)*P_l^m(cos(theta_g)),
!  where P_l^m(x) are L^2-normalized associated Legendre polynomials.
!----------------------------------------------------------------------
      leglog=.true.

      do m1=1,kdim
         m=m1+kmin-1
         do g1=1,gdim
            call algndr(gdim-1+abs(m),abs(m),ort(g1),trafo(1,g1,m1), &
                        leglog)

            if(m.lt.0 .and. mod(-m,2).ne.0) then
               do g2=1,gdim
                  trafo(g2,g1,m1) = -w(g1)*trafo(g2,g1,m1)
               enddo
            else
               do g2=1,gdim
                  trafo(g2,g1,m1) = w(g1)*trafo(g2,g1,m1)
               enddo
            endif
         enddo
      enddo
!----------------------------------------------------------------------
!  Schmidt-orthogonalize the matrix trafo.
!----------------------------------------------------------------------

      do m=1,kdim
         call schmidtorthod(trafo(1,1,m),gdim)
      enddo

!-----------------------------------------------------------------------
!  Calculate second derivative matrix dif2mat (-j^2) in corresponding
!  DVR basis (is analytically known),  --> transform to DVR basis
!  dif2mat := sin^-1(theta) * d/dtheta * sin(theta) * d/dtheta
!              - m^2/sin^2(theta).
!-----------------------------------------------------------------------

      do m1=1,kdim
         m=m1+kmin-1
         call zeromxd(dif2mat(1,1,m1),gdim,gdim)

! ----   -(1/2*I)*Kinetic-energy in Leg-FBR basis (diagonal)

         do g=1,gdim
            j=g+abs(m)-1
            x=j*(j+1)

! for real wavepacket need in FBR
           if (lrealwp) then
              dif2mat(g,g,m1) = -x
           else

! ---    Transform to DVR basis

              do g1=1,gdim
                 do g2=1,gdim
                    dif2mat(g2,g1,m1) = dif2mat(g2,g1,m1)  &
                                        - trafo(g,g2,m1)*x*trafo(g,g1,m1)
                 enddo
              enddo
           endif
         enddo
      enddo

!-----------------------------------------------------------------------
!  Calculate DVR matrix representation of j_+ operator
!  (is analytically known):
!       j_+ := d/dtheta - m*cot(theta).
!-----------------------------------------------------------------------

      if (kdim.ne.1) then
         do m1=1,kdim-1
            m=m1+kmin-1
            call zeromxd(dif1mat(1,1,m1),gdim,gdim)

! ----     j*(j+1)-m*(m+1) in Leg-FBR basis (diagonal)

            do g=1,gdim
               j=g+abs(m)-1
               gg=g+abs(m)-abs(m+1)
               if( gg .gt. 0 .and. gg .le. gdim) then
                 x=sqrt(dble(j*(j+1)-m*(m+1)))

! for real wavepacket need in FBR
                 if (lrealwp) then
                    dif1mat(g,g,m1) = -x
                 else

! ---       Transform to DVR basis

                    do g1=1,gdim
                       do g2=1,gdim
                          dif1mat(g2,g1,m1) = dif1mat(g2,g1,m1) &
                                     - trafo(gg,g2,m1+1)*x*trafo(g,g1,m1)
                       enddo
                    enddo
                 end if
               end if
            enddo
         enddo
         call zeromxd(dif1mat(1,1,kdim),gdim,gdim)
      else
         call zeromxd(dif1mat,gdim,gdim)
      endif


!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of d/dtheta * sin(theta)
!
!  Note: The FBR representation is antisymmetric tridiagonal with zero
!        diagonal elements and  may be symbolically written as:
!
!              D_ab^FBR = x_a [ delta(a,b+1) - delta(a,b-1) ]
!
!        with x_a being the subdiagonal element. The FBR-DVR trafo
!        below makes use of this special form.
!
!-----------------------------------------------------------------------

      do m1=1,kdim
         m=m1+kmin-1
         call zeromxd(dif3mat(1,1,m1),gdim,gdim)

         do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis (j is really j+1).
            j=g+abs(m)
            x=j*sqrt((j**2-m**2)/(4.0_dop*j**2-1.0_dop))

! --- Transform from FBR to DVR basis

            do g1=1,gdim
               do g2=1,gdim
                  dif3mat(g2,g1,m1)=dif3mat(g2,g1,m1) + &
                        trafo(g+1,g2,m1)*trafo(g,g1,m1)*x -  &
                        trafo(g,g2,m1)*trafo(g+1,g1,m1)*x
               enddo
            enddo
         enddo
      enddo


!-----------------------------------------------------------------------
!  Rescale grid points cos(theta) --> theta
!  acos returns the arc cosine in the range Pi to 0, i.e.
!  the largest value correspond to g=1.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=acos(ort(g))
      enddo

!-----------------------------------------------------------------------
!  Finally, store trafo into dif4mat.
!  Not a nice fix, but it saves us from passing trafo all the way
!  to funckleg. HDM 02/09
!-----------------------------------------------------------------------
      call cpvxd(trafo,dif4mat,gdim*gdim*kdim)

      return


! ************************  symmetry  **********************************
 100  continue

!-----------------------------------------------------------------------
!       ibleg=1 -> l=ungerade (independent of m),
!       ibleg=2 -> l=gerade (independent of m)
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos^2(theta) operator (note that the diagonal elements are not zero).
!-----------------------------------------------------------------------

! --- Set diagonal elements for X_mn.

      do g =1,gdim
         j  = 2*(g-1) + kmin2
         j1 = j + 1
         ort(g) = (j**2-kmin2**2)/(4.0_dop*j*j-1.0_dop) &
                + (j1**2-kmin2**2)/(4.0_dop*j1*j1-1.0_dop)
      enddo

! --- Set super-diagonal elements for X_mn.  (j is really j+2).

      do g=2,gdim
         j  = 2*(g-1) + kmin2
         j1 = j - 1
         dif2mat(g-1,1,1) = sqrt((j**2-kmin2**2)/(4.0_dop*j*j-1.0_dop) &
                           * (j1**2-kmin2**2)/(4.0_dop*j1*j1-1.0_dop))
      enddo

! --- Diagonalize cos^2(theta) matrix to obtain "ort".

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)

      if (error.ne.0) then
         message = 'Error in diagonalising cos^2(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
!  Calculate the weights:
!  w(g)=sqrt(2^(m+1)*m!/(2m+1)!!)*sin^(-m)(theta_g)*trafo(1,g,1),
!  where theta_g=acos(sqrt(ort(g)))  and   m=kmin2.
!-----------------------------------------------------------------------
      do g=1,gdim
         w0 = sqrt(2.0_dop)
         x  = sin(acos(sqrt(ort(g))))
         do m=1,kmin2
            w0 = w0 * sqrt(2.0_dop*m/dble(2*m+1))/x
         enddo
         w(g) = w0 * abs(trafo(1,g,1))
      enddo

!----------------------------------------------------------------------
!  Calculate matrix trafo:
!  trafo(l,g,m)=w(g)*P_l^m(cos(theta_g)),
!  where P_l^m(x) are L^2-normalized associated Legendre polynomials.
!----------------------------------------------------------------------
      leglog=.true.

      do m1=1,kdim
         m=m1+kmin-1
         jmin = abs(m) + mod(abs(m)+ibleg,2)
         jmax = 2*(gdim-1) + jmin
         do g1=1,gdim
            call algndr(jmax,abs(m),sqrt(ort(g1)),workr,leglog)
            if(m.lt.0 .and. mod(-m,2).ne.0) then
               do g2=1,gdim
                  j = 2*(g2-1) + jmin
                  trafo(g2,g1,m1) = -w(g1)*workr(j-abs(m)+1)
               enddo
            else
               do g2=1,gdim
                  j = 2*(g2-1) + jmin
                  trafo(g2,g1,m1) = w(g1)*workr(j-abs(m)+1)
               enddo
            endif
         enddo
      enddo

!----------------------------------------------------------------------
!  Schmidt-orthogonalize the matrix trafo.
!----------------------------------------------------------------------

      do m=1,kdim
         call schmidtorthod(trafo(1,1,m),gdim)
      enddo

!-----------------------------------------------------------------------
!  Calculate second derivative matrix dif2mat (-j^2) in corresponding
!  DVR basis (is analytically known),  --> transform to DVR basis
!  dif2mat := sin^-1(theta) * d/dtheta * sin(theta) * d/dtheta
!              - m^2/sin^2(theta).
!-----------------------------------------------------------------------

      do m1=1,kdim
         m=m1+kmin-1
         call zeromxd(dif2mat(1,1,m1),gdim,gdim)

! ---     -(1/2*I)*Kinetic-energy in Leg-FBR basis (diagonal)

         do g=1,gdim
            jmin = abs(m) + mod(abs(m)+ibleg,2)
            j = 2*(g-1) + jmin
            x = j*(j+1)

! for real wavepacket need in FBR
            if (lrealwp) then
               dif2mat(g,g,m1) = -x
            else

! ---    Transform to DVR basis

               do g1=1,gdim
                  do g2=1,gdim
                     dif2mat(g2,g1,m1) = dif2mat(g2,g1,m1) &
                                    - trafo(g,g2,m1)*x*trafo(g,g1,m1)
                  enddo
               enddo
            endif
         enddo
      enddo


!-----------------------------------------------------------------------
!  Calculate DVR matrix representation of j_+ operator (is
!  analytically known).
!       j_+ := d/dtheta - m*cot(theta).
!-----------------------------------------------------------------------

      if (kdim.ne.1) then
         do m1=1,kdim-1
            m=m1+kmin-1
            call zeromxd(dif1mat(1,1,m1),gdim,gdim)

            do g=1,gdim
               jmin = abs(m) + mod(abs(m)+ibleg,2)
               j = 2*(g-1) + jmin
               x = sqrt(dble(j*(j+1)-m*(m+1)))
               gg= g + (abs(m)+ibleg-1)/2 - (abs(m+1)+ibleg-1)/2
               if( gg .gt. 0 .and. gg .le. gdim) then

! for real wavepacket need in FBR
                 if (lrealwp) then
                    dif1mat(g,g,m1) = -x
                 else

! ---       Transform from FBR to DVR basis

                    do g1=1,gdim
                       do g2=1,gdim
                          dif1mat(g2,g1,m1) = dif1mat(g2,g1,m1) &
                                       - trafo(gg,g2,m1+1)*x*trafo(g,g1,m1)
                       enddo
                    enddo
                 end if
               end if
            enddo
         enddo
         call zeromxd(dif1mat(1,1,kdim),gdim,gdim)
      else
         call zeromxd(dif1mat,gdim,gdim)
      endif


!-----------------------------------------------------------------------
! Calculate DVR matrix representation of of :
!  0.5*( cos(theta)*d/dtheta*sin(theta)  +
!  d/dtheta*sin(theta)*cos(theta) )
!
!  Note: The FBR representation is antisymmetric tridiagonal with zero
!        diagonal elements and  may be symbolically written as:
!
!              D_ab^FBR = x_a [ delta(a,b+1) - delta(a,b-1) ]
!
!        with x_a being the subdiagonal element. The FBR-DVR trafo
!        below makes use of this special form.
!
!-----------------------------------------------------------------------

      do m1=1,kdim
         m=m1+kmin-1
         call zeromxd(dif3mat(1,1,m1),gdim,gdim)

         do g=1,gdim-1

! --- Get super-diagonal matrix element in FBR basis.
            jmin = abs(m) + mod(abs(m)+ibleg,2)
            j = 2*(g-1) + jmin
            x = (j+1.5_dop)*sqrt(((j+1)**2-m**2)/(4.0_dop*(j+1)**2-1.0_dop)) &
                         *sqrt(((j+2)**2-m**2)/(4.0_dop*(j+2)**2-1.0_dop))

! --- Transform from FBR to DVR basis

            do g1=1,gdim
               do g2=1,gdim
                  dif3mat(g2,g1,m1)=dif3mat(g2,g1,m1) + &
                        trafo(g+1,g2,m1)*trafo(g,g1,m1)*x - &
                        trafo(g,g2,m1)*trafo(g+1,g1,m1)*x
               enddo
            enddo
         enddo
      enddo


!-----------------------------------------------------------------------
! Rescale grid points cos^2(theta) --> theta
! Note: the grid points lie in the interval (0,Pi/2)
!       the largest value correspond to g=1.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=acos(sqrt(ort(g)))
      enddo

!-----------------------------------------------------------------------
!  Finally, store trafo into dif4mat.
!  Not a nice fix, but it saves us from passing trafo all the way
!  to funckleg. HDM 02/09
!-----------------------------------------------------------------------
      call cpvxd(trafo,dif4mat,gdim*gdim*kdim)

      return
      end subroutine initkleg

!#######################################################################



!-----------------------------------------------------------------------
!                Subroutine INITK
!-----------------------------------------------------------------------

      subroutine initk(ort,gdim,kmin,dk)

      implicit none

      integer(long)                           :: g1,g2
      integer(long), intent(in)               :: gdim,kmin,dk
      real(dop), dimension(gdim), intent(out) :: ort

      do g1=1,gdim
         g2=kmin+(g1-1)*dk
         ort(g1)=dble(g2)
      enddo

      return
      end subroutine initk

!-----------------------------------------------------------------------
!                Subroutine INITEXTRN - For External-DVR
!
!  Reads grid points and dif1,dif2 matrices from channel ifile.
!  The file is already opened for reading.
!
!-----------------------------------------------------------------------
 
      subroutine initextern(ifile,filetype,unitfactor,ort, &
                            dif1mat,dif2mat,gdim)
 
      implicit none
 
      integer(long)                                :: g,g1,g2
      integer(long), intent(in)                    :: ifile,filetype,gdim
      real(dop)                                    :: omx
      real(dop), intent(in)                        :: unitfactor
      real(dop), dimension(gdim), intent(out)      :: ort
      real(dop), dimension(gdim,gdim), intent(out) :: dif1mat,dif2mat


      if (filetype .eq. 1) then
         do g=1,gdim
            read(ifile,*,end=12,err=14) ort(g)
         enddo
         do g2=1,gdim
            read(ifile,*,end=13,err=15) (dif2mat(g1,g2),g1=1,gdim)
         enddo
         do g=1,gdim
            read(ifile,*,end=13,err=16) (dif1mat(g1,g),g1=1,gdim)
         enddo
      else
         do g=1,gdim
            read(ifile,end=12,err=14) ort(g)
         enddo
         do g2=1,gdim
            read(ifile,end=13,err=15) (dif2mat(g1,g2),g1=1,gdim)
         enddo
         do g=1,gdim
            read(ifile,end=13,err=16) (dif1mat(g1,g),g1=1,gdim)
         enddo
      endif

 11   close(ifile)
! --- check for consistency
      omx=abs(ort(1))
      do g=2,gdim
         omx=max(omx,abs(ort(g))) 
      enddo 
      do g=1,gdim-2
         if( (ort(g)-ort(g+1))*(ort(g+1)-ort(g+2)) .le. 0.0_dop ) then
            write(6,'(/2a/a,2i5)') ' ### ERROR: grid points are not ', &
             'in increasing or decreasing order.',' g,gdim =',g+1,gdim
            write(2,'(/2a/a,2i5)') ' ### ERROR: grid points are not ', &
             'in increasing or decreasing order.',' g,gdim =',g+1,gdim
            if(omx.gt.9.99e4_dop) then
               write(6,'(1x,7f11.1)') (ort(g1),g1=1,gdim)
               write(2,'(1x,7f11.1)') (ort(g1),g1=1,gdim)
            elseif(omx.gt.9.99e2_dop) then
               write(6,'(1x,7f11.3)') (ort(g1),g1=1,gdim)
               write(2,'(1x,7f11.3)') (ort(g1),g1=1,gdim)
            elseif(omx.gt.9.99_dop) then
               write(6,'(1x,7f11.5)') (ort(g1),g1=1,gdim)
               write(2,'(1x,7f11.5)') (ort(g1),g1=1,gdim)
            else
               write(6,'(1x,7f11.7)') (ort(g1),g1=1,gdim)
               write(2,'(1x,7f11.7)') (ort(g1),g1=1,gdim)
            endif
            call stopnow
         endif
      enddo 
! --- Change units  
      if(abs(unitfactor-1.0_dop) .gt. 1.0e-9_dop) then
         do g=1,gdim
            ort(g)=ort(g)*unitfactor
         enddo 
      endif

      return

 
 12   if(g.eq.gdim) goto 11
      write(6,*) 'EOF found when reading grid points of Extern-DVR.'
      write(6,*) 'g =', g, ' , gdim =', gdim
      call stopnow

 13   if (g2.eq.1) then
          call zerovxd(dif2mat,gdim*gdim)
          call zerovxd(dif1mat,gdim*gdim)
          go to 11
      elseif (g.eq.1) then
          call zerovxd(dif1mat,gdim*gdim)
          go to 11
      else
          write(6,*) 'End of file for Extern-DVR reached.'
          write(6,*) 'g =', g, ' , g2 =', g2, ' , gdim =', gdim
          call stopnow
      endif

 14   write(6,*) 'Error when reading grid-points of Extern-DVR.'
      call stopnow

 15   write(6,*) 'Error when reading dif2mat of Extern-DVR.'
      call stopnow

 16   write(6,*) 'Error when reading dif1mat of Extern-DVR.'
      call stopnow

      end subroutine initextern


!------------------------------------------------------------------------
!                Subroutine INITLEGR
!              restricted Legendre DVR
!------------------------------------------------------------------------

      subroutine initlegr(trafo,ort,dif2mat,dif1mat,ggdim,gdim,xi,xf)

      implicit none

      integer(long)                                  :: g,g1,g2,shift
      integer(long), intent(in)                      :: ggdim
      integer(long), intent(out)                     :: gdim
      real(dop), intent(out)                         :: xi,xf
      real(dop), dimension(ggdim), intent(out)       :: ort
      real(dop), dimension(ggdim,ggdim), intent(out) :: trafo,dif2mat,dif1mat

!-------------------------------------------------------------------------
! ggdim is the (old) grid-length, i.e. the number of legendre functions.
! gdim is set to the new grid-length, i.e. the restricted number of 
! angle-grid points. 
!-------------------------------------------------------------------------

      g1 = 1
      g2 = ggdim
      do g = 1, ggdim
         if( xf .gt. ort(ggdim+1-g) ) g2 = ggdim+1-g
         if( xi .lt. ort(g) ) g1 = g
      end do
      xi = ort(g1)
      xf = ort(g2)
      shift = g1 - 1
      gdim  = g2 - g1 + 1

      if( shift .gt. 0 ) then
         do g = 1, gdim
            ort(g) = ort(g+shift)
            do g1 = 1, ggdim
               trafo(g1,g) = trafo(g1,g+shift)
            end do
         end do
      end if

      call shiftlegr(dif2mat,dif1mat,gdim,ggdim,shift)

      return
      end subroutine initlegr

!---------------------------------------------------------------------------

      subroutine shiftlegr(dif2mat,dif1mat,gdim,ggdim,shift)
!-------------------------------------------------------------------------
! Shift the restricted grid entries of dif1mat and dif2mat to the upper left
! corner and zero the rest. 
! ggdim is the old (large) grid-dimension, gdim is the new (small) one.
!-------------------------------------------------------------------------

      implicit none

      integer(long)                                  :: g1,g2
      integer(long), intent(in)                      :: gdim,ggdim,shift
      real(dop), dimension(gdim,gdim)                :: d2m,d1m
      real(dop), dimension(ggdim,ggdim), intent(out) :: dif2mat,dif1mat


      do g2 = 1, gdim
         do g1 = 1, gdim
            d1m(g1,g2) = dif1mat(g1+shift,g2+shift)
            d2m(g1,g2) = dif2mat(g1+shift,g2+shift)
         end do
      end do
      call zeromxd(dif1mat,ggdim,ggdim)
      call zeromxd(dif2mat,ggdim,ggdim)

      call cpvxd(d1m,dif1mat,gdim*gdim)
      call cpvxd(d2m,dif2mat,gdim*gdim)

      return
      end subroutine shiftlegr

!------------------------------------------------------------------------
!                Subroutine INITKLEGR
!              restricted KLegendre DVR
!------------------------------------------------------------------------

      subroutine initklegr(trafo,ort,dif3mat,dif2mat,dif1mat,ggdim, &
                           gdim,kdim,xi,xf)

      implicit none

      integer(long)                                       :: g,g1,g2,m1,shift
      integer(long), intent(in)                           :: ggdim,kdim
      integer(long), intent(out)                          :: gdim
      real(dop), intent(out)                              :: xi, xf
      real(dop), dimension(ggdim), intent(out)            :: ort
      real(dop), dimension(ggdim,ggdim,kdim), intent(out) :: trafo,dif3mat, &
                                                             dif2mat,dif1mat

!-------------------------------------------------------------------------
! ggdim is the (old) grid-length, i.e. the number of legendre functions.
! gdim is set to the new grid-length, i.e. the restricted number of 
! angle-grid points. 
!-------------------------------------------------------------------------

      g1 = 1
      g2 = ggdim
      do g = 1, ggdim
         if( xf .gt. ort(ggdim+1-g) ) g2 = ggdim+1-g
         if( xi .lt. ort(g) ) g1 = g
      end do
      xi = ort(g1)
      xf = ort(g2)
      shift = g1 - 1
      gdim  = g2 - g1 + 1

      if( shift .gt. 0 ) then

         do g = 1, gdim
            ort(g) = ort(g+shift)
         enddo
         do m1=1,kdim
            do g = 1, gdim
               do g1 = 1, ggdim
                  trafo(g1,g,m1) = trafo(g1,g+shift,m1)
               end do
            end do
         end do
      end if

      call shiftklegr(dif3mat,dif2mat,dif1mat,gdim,ggdim,kdim,shift)

      return
      end subroutine initklegr

!---------------------------------------------------------------------------

      subroutine shiftklegr(dif3mat,dif2mat,dif1mat,gdim,ggdim,kdim,shift)
!-------------------------------------------------------------------------
! Shift the restricted grid entries of dif1mat, dif2mat and dif3mat 
! to the upper left corner and zero the rest. 

      implicit none

      integer(long)                                       :: g1,g2,m1
      integer(long), intent(in)                           :: gdim,ggdim,kdim,shift
      real(dop), dimension(ggdim,ggdim,kdim), intent(out) :: dif3mat,dif2mat,dif1mat
      real(dop), dimension(gdim,gdim,kdim)                :: d2m,d1m,d3m


      do m1 = 1, kdim
         do g2 = 1, gdim
            do g1 = 1, gdim
               d1m(g1,g2,m1) = dif1mat(g1+shift,g2+shift,m1)
               d2m(g1,g2,m1) = dif2mat(g1+shift,g2+shift,m1)
               d3m(g1,g2,m1) = dif3mat(g1+shift,g2+shift,m1)
            end do
         end do
      end do
      call zerovxd(dif1mat,ggdim*ggdim*kdim)
      call zerovxd(dif2mat,ggdim*ggdim*kdim)
      call zerovxd(dif3mat,ggdim*ggdim*kdim)

      call cpvxd(d1m,dif1mat,gdim*gdim*kdim)
      call cpvxd(d2m,dif2mat,gdim*gdim*kdim)
      call cpvxd(d3m,dif3mat,gdim*gdim*kdim)

      return
      end subroutine shiftklegr

!-------------------------------------------------------------------------
      end module init2mod

