!***********************************************************************
!
!                Module INIT3
!
! Contents: initwigner
!
! calls lapack-routine DSTEQR to diagonalise tri-diagonal matrices.
!
! Initialises the arrays trafo, dif2mat and dif1mat for a DVR basis set
!
! variables passed:
!  trafo  : DVR/FBR transformation matrices
!  ort    : DVR grid points
!  dif1mat  : DVR first  derivative matrix
!  dif2mat  : DVR second derivative matrix
!  gdim   : number of DVR grid points for this mode
!  kdim   : number of k-values
!  ibleg  : symmetry parameter. 0->no sym, 1->ungerade, 2->gerade
!
!***********************************************************************
!#######################################################################

      module init3mod

use decimal, only: long, dop
use logdat
use op1lib, only: zeromxd, cpvxd
use schmidtorthomod, only: schmidtorthod

      implicit none
      private
      public :: initwigner
      contains

!-----------------------------------------------------------------------
!                Subroutine INITWIGNER
!               Wigner function DVR
!------------------------------------------------------------------------

      subroutine initwigner(trafo,ort,dif4mat,dif3mat,dif2mat,dif1mat,&
                          gdim,kdim,mmdim,kmin,mmmin,kmin2,mmmin2,&
                          ibleg,message)

      implicit none

      integer(long)                 :: g,gg,g1,g2,j,error
      integer(long), intent(in)     :: gdim,kdim,ibleg,kmin
      integer(long), intent(inout)  :: kmin2,mmmin2
      integer(long)                 :: m,m1,mm,mm1
      integer(long), intent(in)     :: mmdim,mmmin
      integer(long)                 :: mormm,mormm2,maxms
      real(dop), dimension(gdim), intent(out)                 :: ort
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(out) :: trafo, &
                                                                 dif1mat, & 
                                                                 dif2mat, &
                                                                 dif3mat, &
                                                                 dif4mat
      real(dop), dimension(gdim)    :: w
      real(dop)                     :: x,djmk
      character(len=*), intent(out) :: message


!-----------------------------------------------------------------------
! ibleg=0 -> no sym, ibleg=1 -> l=ungerade, ibleg=2 l=gerade.
! Use special formulas for the gerade (j=even) or ungerade (j=odd) case.
!
! Symmetry has not yet been implement into the Wigner code
!  -P. Thomas 06 July 2012
! So far code only works for kmin2=mmmin2=0
!  -P. Thomas 25 January 2013
!-----------------------------------------------------------------------

      if( ibleg .gt. 0 )  goto 100

! ********************** no symmetry ***********************************

!-----------------------------------------------------------------------
! Calculate symmetric tridiagonal matrix representation X_mn of the
! cos(beta) operator.
!-----------------------------------------------------------------------

      mormm2 = max(kmin2,mmmin2)
! --- Make kmin2 negative if entire k range is negative; same for mmmin2
      if (kmin2.gt.0 .and. kmin.lt.0) then
        kmin2 = -kmin2
      endif
      if (mmmin2.gt.0 .and. mmmin.lt.0) then
        mmmin2 = -mmmin2
      endif

! --- Set diagonal elements of X_mn to k*mm/(j*j+1)

      do g=1,gdim
         j=g-1+mormm2
         if (j.gt.0) then
            ort(g)=kmin2*mmmin2/(j*(j+1.d0))
         else
            ort(g)=0
         endif
      enddo

! --- Set subdiagonal elements of X_mn to
! --- sqrt((j^2-k^2)*(k^2-mm^2))/(j*sqrt(2j+1)*sqrt(2j-1))

      do g=1,gdim-1
         j=g+mormm2
         dif2mat(g,1,1,1)=sqrt(dble((j**2-kmin2**2)*(j**2-mmmin2**2))/ &
                               dble(4*j*j-1))/dble(j)
      enddo

! --- Diagonalize cos(beta)-matrix to obtain "ort" and trafo(i,j,1,1).

      call DSTEQR('I',gdim,ort,dif2mat,trafo,gdim,dif1mat,error)
      if (error.ne.0) then
         message = 'Error in diagonalising cos(theta)-matrix'
         call errormsg
      endif

!-----------------------------------------------------------------------
!  Calculate the Wigner weights^(1/2) for the cos(beta)s:
!  w(g)=absolute value of "1st element of each eigenvector in trafo"
!  divided by "djmn evaluated at the grid point for the minimum values 
!  of j, |k|, and |mm| used in the FBR"
!-----------------------------------------------------------------------

      do g=1,gdim
         w(g)=abs(trafo(1,g,1,1)/djmk(mormm2,mmmin2,kmin2,ort(g)))
      enddo

!----------------------------------------------------------------------
!  Calculate matrix trafo :
!  trafo(g2,g1,m1,mm1)=w(g)*d_j^(mm,m)(cos(beta_g)),
!  where d_j^(mm,m)(x) are Wigner (small) d-functions.
!----------------------------------------------------------------------

      do mm1=1,mmdim
         mm=mm1-1+mmmin
         do m1=1,kdim
            m=m1-1+kmin
            g=max(abs(m),abs(mm))
            do g1=1,gdim
               do g2=1,gdim
                  j=g2-1+g
                  trafo(g2,g1,m1,mm1)=w(g1)*djmk(j,mm,m,ort(g1))
               enddo
            enddo
         enddo
      enddo

!----------------------------------------------------------------------
!  Schmidt-orthogonalize the matrix trafo.
!----------------------------------------------------------------------

      do mm1=1,mmdim
         do m1=1,kdim
           call schmidtorthod(trafo(1,1,m1,mm1),gdim)
         enddo
      enddo

!-----------------------------------------------------------------------
!  Calculate second derivative matrix dif2mat (-j^2) in corresponding
!  DVR basis (analytically known, with eigenvalue: j*(j+1) )
!  --> First, transform to DVR basis, then sum over beta/beta' pairs
!  dif2mat := sin^-1(beta) * d/dbeta * sin(beta) * d/dbeta
!              - m^2/sin^2(beta).
!-----------------------------------------------------------------------

      do mm1=1,mmdim
         mm=mm1-1+mmmin
         do m1=1,kdim
            m=m1-1+kmin
            call zeromxd(dif2mat(1,1,m1,mm1),gdim,gdim)

! ----     Choose m or mm to designate the start of the j-range

            mormm = max(abs(m),abs(mm))

! ----   -(1/2*I)*Kinetic-energy in Leg-FBR basis (diagonal)

            do g=1,gdim
               j=g-1+mormm
               x=j*(j+1)

! ---    Transform to DVR basis

               do g1=1,gdim
                  do g2=1,gdim
                     dif2mat(g2,g1,m1,mm1) = dif2mat(g2,g1,m1,mm1) &
                            - trafo(g,g2,m1,mm1)*x*trafo(g,g1,m1,mm1)
                  enddo
               enddo
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
!  Calculate DVR matrix representation of the molecule BODY-fixed
!  j_+ operator (analytically known, with eigenvalue:
!  sqrt(j*(j+1)-k*(k+1)) when used on D^j_mk(beta) ).
!  Note: this can also be used to represent j_- (call using k lowered
!  by one, and with the beta/beta' indices swapped)
!-----------------------------------------------------------------------

      if (kdim.ne.1) then
         do mm1=1,mmdim
            mm=mm1-1+mmmin
            do m1=1,kdim-1
               m=m1-1+kmin
               maxms=max(abs(m),abs(m+1))
               call zeromxd(dif1mat(1,1,m1,mm1),gdim,gdim)

! ----     Choose m or mm to designate the start of the j-range

               mormm = max(abs(m),abs(mm))

! ----     j*(j+1)-k*(k+1) in Leg-FBR basis (diagonal)

               do g=1,gdim
                  j=g-1+mormm
                  gg=g
                  if(abs(mm).lt.maxms) gg=gg+abs(m)-abs(m+1)
                  if( gg .ge. 1 .and. gg .le. gdim) then
                    x=sqrt(dble(j*(j+1)-m*(m+1)))

! ---      Transform to DVR basis

                    do g1=1,gdim
                       do g2=1,gdim
                          dif1mat(g2,g1,m1,mm1) = dif1mat(g2,g1,m1,mm1) &
                           - trafo(gg,g2,m1+1,mm1)*x*trafo(g,g1,m1,mm1)
                       enddo
                    enddo
                  end if
               enddo
            enddo

! ---      Applying j_+ to the largest k-value gives zero

            call zeromxd(dif1mat(1,1,kdim,mm1),gdim,gdim)

         enddo
      else
         call zeromxd(dif1mat,gdim,gdim)
      endif

!-----------------------------------------------------------------------
!  Calculate DVR matrix representation of the molecule SPACE-fixed
!  js_+ operator (analytically known, with eigenvalue:
!  sqrt(j*(j+1)-m*(m+1)) when used on D^j_mk(beta) ).
!  Note: this can also be used to represent js_- (call using m lowered
!  by one, and with the beta/beta' indices swapped)
!-----------------------------------------------------------------------

      if (mmdim.ne.1) then
         do mm1=1,mmdim-1
            mm=mm1-1+mmmin
            maxms=max(abs(mm),abs(mm+1))
            do m1=1,kdim
               m=m1-1+kmin
               call zeromxd(dif3mat(1,1,m1,mm1),gdim,gdim)

! ----     Choose m or mm to designate the start of the j-range

               mormm = max(abs(m),abs(mm))

! ----     j*(j+1)-m*(m+1) in Leg-FBR basis (diagonal)

               do g=1,gdim
                  j=g-1+mormm
                  gg=g
                  if(abs(m).lt.maxms) gg=gg+abs(mm)-abs(mm+1)
                  if( gg .ge. 1 .and. gg .le. gdim) then
                    x=sqrt(dble(j*(j+1)-mm*(mm+1)))

! ---      Transform to DVR basis

                    do g1=1,gdim
                       do g2=1,gdim
                          dif3mat(g2,g1,m1,mm1) = dif3mat(g2,g1,m1,mm1) &
                           - trafo(gg,g2,m1,mm1+1)*x*trafo(g,g1,m1,mm1)
                       enddo
                    enddo
                  end if
               enddo
            enddo
         enddo

! ---      Applying j_+ to the largest m-value gives zero

         do m1=1,kdim
            call zeromxd(dif3mat(1,1,m1,mmdim),gdim,gdim)
         enddo

      else
         call zeromxd(dif3mat,gdim,gdim)
      endif

!-----------------------------------------------------------------------
!  Store trafo into dif4mat.
!  Not a nice fix, but saves us from passing trafo to funcwig
!-----------------------------------------------------------------------
      call cpvxd(trafo,dif4mat,gdim*gdim*kdim*mmdim)

!-----------------------------------------------------------------------
!  Finally, rescale grid points cos(beta) --> beta
!  acos returns the arc cosine in the range pi to 0, i.e.
!  the largest value corresponds to g=1.
!-----------------------------------------------------------------------
      do g=1,gdim
         ort(g)=acos(ort(g))
      enddo

      return

! ************************  symmetry  **********************************
 100  continue

      message = 'Symmetry is not yet implemented in Wigner DVR'
      call errormsg

      return
      end subroutine initwigner

      end module init3mod

!#######################################################################

