!***********************************************************************
!
! contains: adiabwkb,adiabwkbhh2,wkadiabwkb
!
!***********************************************************************
      module adiabwkbmod

use decimal, only: long, dop
use global
use logdat

      implicit none

      private
      public :: adiabwkb,adiabwkbhh2
     contains

!***********************************************************************
!
!               MCTDH  module  ADIABWKB
!
!
!   computation of energy distribution of the initial wave packet
!    ( call to subr. energdstr )
!   1) without any (a)diabatic correction 
!   2) with  diabatic correction, call to transmf to compute the 
!      mean-field potential ( see restriction in subr. transmf )
!   No general adiabatic correction available
!
!   Adiabatic correction for the H+H2 and related systems is still
!   available by specifying the 'hh2' keyword in the input.
!
!   Originally : 04/97 AJ  
!   Modified   : 03/98 MCH, 10/99 HDM, 06/2003 HDM
!
!***********************************************************************

      subroutine adiabwkb(psi,ort)

use dvrdatmod
use griddatmod
use psidef
use operdef
use runinwfmod
use channels

use transmfmod, only: transmf

use enerdstrmod

      implicit none

      integer(long), parameter :: maxendim=5000, ne=5000

      integer(long)       :: g1,n,s,mtr
      real(dop)           :: e0intern,e0trans,e0trpot,de,xi,xf,emin,vexp,vvar

      real(dop), dimension(sdim)                       :: tmf
      real(dop), dimension(ortdim), intent(in)         :: ort
      complex(dop), dimension(maxendim)                :: edstr
      complex(dop), dimension(dgldim), intent(in)      :: psi
      logical(kind=4)                                  :: lerr

!-----------------------------------------------------------------------
! Initialize variables.
!-----------------------------------------------------------------------
      routine = 'ADIABWKB'
      lerr = .false.

! mtr : translational mode; ftr transl. dof; statetr: transl. state.
      mtr=dofspf(ftr)
      s = statetr
      if(nspfdof(mtr).ne.1) then
         message = 'Translational mode cannot be combined!'
         call errormsg
         stop
      end if

!-----------------------------------------------------------------------
! This is a (very) specialized code --> stop for wrong input parameters
!-----------------------------------------------------------------------
      if( (ledstr.and.lad) .or. (ledstr.and.ldia)  &
                           .or.  (lad.and.ldia) ) then
          message ='use either edstr or dia or ad'
          call warnmsg
      endif

      if (basis(ftr).lt.3.or.basis(ftr).gt.5)  then
          message = 'Equidistant points required for correction.'// &
          ' Use sin, exp or FFT'
          call errormsg
      endif

!      if (sbasis(ftr).ne.1) then
!          message = 'Initial WF for translational mode must be gauss'
!          call errormsg
!      endif

      if (lad) then 
          message ='no adiabatic correction yet'
          call errormsg
      endif

      if (nspfdof(mtr).ne.1) then
         message = 'not verified for a contracted translational mode '
         call errormsg
      endif

      write (ilog,'(a)') '******** (A)Diabatic Corrections ********'
      write (ilog,*)
          
!-----------------------------------------------------------------------
! Calculate DIABATIC translational mean-field 'tmf' needed for WKB
! correction of initial energy distribution.
!-----------------------------------------------------------------------
      if (ldia.or.ledstr)  &
         call transmf(tmf,psi,ort(zort(ftr)),s,mtr,e0intern,e0trans,e0trpot)

!-----------------------------------------------------------------------
! Set zero point of energy to 
! Write adiabatic potential data to 'enerd' file
!-----------------------------------------------------------------------
      open(ienerd,file=name(1:laenge)//'/enerd',form='formatted',err=10)

      write (ienerd,*,err=10) gdim(ftr)
      do g1=1,gdim(ftr)
         if ((.not.lad).and.(.not.ldia)) then
            tmf(g1)=e0intern
         endif
         write (ienerd,*,err=10) ort(zort(ftr)+g1-1),tmf(g1)*eV
      enddo 
      call myflush(ienerd)

!-----------------------------------------------------------------------
! Calculate corrected initial energy distribution. 
! Actually, energdstr  no longer computes the distorted WF by WKB but
! uses the (numerically exact) Numerov method. 
!-----------------------------------------------------------------------
      if(abs(mass(ftr)-1.d0).lt.1.d-9) then
         message='Mass of translational DOF is 1, unset?'
         routine='adiabwkb'
         call warnmsg
      endif
      xi=rpbaspar(1,ftr)
      xf=rpbaspar(2,ftr)
      call energdstr(edstr,de,emin,ne,gdim(ftr),xi,xf,mass(ftr), &
                     tmf,psi(zetf(mtr,s)),ilog,vexp,vvar,ev)

!-----------------------------------------------------------------------
! Write square of diabatically corrected energy distribution of 
! the initial wave packet to file 'enerd'.
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! Output of the energy distribution is given in total energy E
! E = E_translational + E_internal_energy_of_reaction_fragments
!-----------------------------------------------------------------------


      do n=1,ne
         write (ienerd,*,err=10) &
              (emin + n*de)*eV, abs(edstr(n))**2
      enddo
      close (ienerd)

      if ( ldia )then
          write (ilog,'(3a,i3)') &
        'Diabatic correction. Translational dof: ',modelabel(ftr)(1:12)&
        ,'state =',s 
          write (ilog,'(a,/a)') &
        'Translational mean-field and corrected energy distribution' &
        ,'are written to the enerd file.'
      elseif ( ledstr ) then
          write (ilog,'(a)') 'No (a)diabatic correction.'
          write (ilog,'(a)') &
        'Energy distribution is writen in the enerd file.'
      endif

      write (ilog,'(/2a,f13.7,a)') 'Separable internal energy of the', &
            ' reactant fragment:  ', e0intern*ev,' eV'
      write (ilog,'(2a,f13.7,a)') 'Separable translational energy: ', &
            '                     ', e0trans*ev, ' eV'
      write (ilog,'(2a,f13.7,a)') 'Separable translational potential', &
             ' energy shift:      ', e0trpot*ev, ' eV'
      write (ilog,'(2a,f13.7,a)') 'Transl. mean field expectation ', &
             'value <psi|tmf|psi> : ', vexp*ev, ' eV'
      write (ilog,'(2a,f13.7,a)') 'Transl. MF variance,  ', &
             'SQRT(<psi|(tmf-<tmf>)^2|psi>): ', vvar*ev, ' eV'
      write (ilog,'(2a,16x,f13.7,a)') 'Energy shift (E_total - ', &
            'E_kinetic) : ', tmf(gdim(ftr))*ev,' eV'
      write (ilog,'(/2a,/2a,/2a)') &
            'The diabatic correction is likely to fail, if the ', &
            'variance of the','translational mean field ',  &
            'is large (> 1/5 of the mean kinetic energy, say).'  
      write (ilog,'(/2a,/2a,/3a/2a,/2a)') &
            'The total-kinetic energy shift is correct only, if ', &
            'there are no centrifugal','terms for the translational ',  &
            'coordinate. Set J,j=0 to eliminate them.','If the ', &
            'centri. term is separable and is the only separable ', &
            'potential term','for the translational DOF (see op.log), ', &
            'then one may simply subtract the','"Sep. transl. pot. ', &
            'energy shift" term from the "total-kinetic shift" term.' 

      call myflush(ilog)
      return

 10   routine = 'ADIABWKB'
      message = 'Cannot open or write to file enerd'
      call errormsg

      return
      end subroutine adiabwkb

!***********************************************************************
!
!               MCTDH  subroutine  ADIABWKBHH2
!
!   Adiabatic correction for H+H2 system
!
!   04/97 AJ
!
!***********************************************************************

      subroutine adiabwkbhh2(psi,ort,exphin)

use dvrdatmod
use griddatmod
use psidef
use runinwfmod
use operdef
use channels

use adiabmod, only: adiab
use transmfmod, only: transmfhh2

use enerdstrmod

      implicit none

      integer(long), parameter          :: maxendim=1000, ne=1000
      real(dop), parameter              :: erange=4.0_dop

      complex(dop), dimension(maxendim) :: edstr

      integer(long)       :: e1,e2,g1,n,p,ipot2,sl03
      real(dop)           :: de,ez,xi,xf,gmom1,emin,vexp,vvar

      real(dop), dimension(sdim)                       :: tmf
      real(dop), dimension(ortdim), intent(in)         :: ort
      complex(dop), dimension(dgldim), intent(inout)   :: psi
      complex(dop), dimension(expdim), intent(in)      :: exphin

      logical(kind=4)                                  :: lerr

!-----------------------------------------------------------------------
! Initialize variables.
!-----------------------------------------------------------------------
      routine = 'ADIABWKB'
      lerr = .false.

      gmom1=rsbaspar(2,1,1)
      ipot2=isbaspar(1,2,1)
!     sl02=isbaspar(2,2,1)
      if (sbasis(3).eq.2) then
         sl03=isbaspar(2,3,1)
      else if (sbasis(3).eq.5) then
         sl03=isbaspar(1,3,1)
      endif


      write (ilog,'(a)') '******** (A)Diabatic Corrections ********'
      write (ilog,*)

!-----------------------------------------------------------------------
! This is a very specialized code --> stop for wrong input parameters
!-----------------------------------------------------------------------
      if (nstate.ne.1) lerr = .true.
!     if (nmode.eq.2.and.sl02.ne.0) lerr = .true.
      if (nmode.gt.3)  lerr = .true.
      if (azahl.ne.0) then
         if (nmode.eq.2) then
            if (lad) then
               lerr = .true.    
            elseif (ldia) then
               do e2=1,dim(2,1)
                  do e1=2,dim(1,1)
                     p=e1+(e2-1)*dim(1,1)
                     if (abs(psi(p)).gt.0.0_dop) lerr = .true.
                  enddo
               enddo
            else
               lerr = .true. 
            endif
         else
            lerr = .true.
         endif
      endif
      if (basis(1).lt.3.or.basis(1).gt.4)  lerr = .true.
      if (basis(2).lt.3.or.basis(2).gt.4)  lerr = .true.
      if (sbasis(1).ne.1) lerr = .true.
      if (sbasis(2).ne.3) lerr = .true.
!     if (ipot2.ne.0)   lerr = .true.
      if (nmode.eq.3) then
!        if (sl02.ne.sl03) lerr = .true.
         if (sbasis(3).ne.2 .and. sbasis(3).ne.5)   lerr = .true.
         if (subdim(3).gt.gdim(3)) then
            if (basis(3).ne.2 .and. basis(3).ne.7 .and. basis(3).ne.20 &
                .and. basis(3).ne.21 .and. basis(3).ne.9) lerr = .true.
         endif
      endif

      if (lerr) then
         message= &
         'Illegal input parameters. (See adiabwkb.F line 262-275).'
         call errormsg
      endif


!-----------------------------------------------------------------------
! Calculate DIABATIC translational mean-field 'tmf' needed for WKB
! correction of initial energy distribution.
!-----------------------------------------------------------------------
      if (ldia) call transmfhh2(tmf,psi,ort)
              
!-----------------------------------------------------------------------
! Calculate ADIABATIC expansion coefficients 'psi' and ADIABATIC 
! translational mean-field 'tmf' needed for WKB correction of initial 
! energy distribution.
!-----------------------------------------------------------------------
      if (lad) call adiab(tmf,psi,ort,exphin)

      call myflush(ilog)

!-----------------------------------------------------------------------
! Set zero point of energy to 
! Write adiabatic potential data to 'enerd' file
!-----------------------------------------------------------------------
      open(ienerd,file=name(1:laenge)//'/enerd',form='formatted',err=10)

      write (ienerd,*,err=10) subdim(1)
      do g1=1,subdim(1)
         if (lad) then
            tmf(g1)=tmf(g1)-e0eigen
         elseif ((.not.lad).and.(.not.ldia)) then
            tmf(g1)=0.0_dop
         endif
         write (ienerd,*,err=10) ort(zort(1)+g1-1),tmf(g1)*eV
      enddo 
      call myflush(ienerd)

!-----------------------------------------------------------------------
! Calculate WKB corrected initial energy distribution. 
!-----------------------------------------------------------------------
      xi=rpbaspar(1,1)
      xf=rpbaspar(2,1)
      call energdstr(edstr,de,emin,ne,gdim(1),xi,xf,mass(1), &
                     tmf,psi(zetf(1,1)),ilog,vexp,vvar,ev)

!-----------------------------------------------------------------------
! Write square (times Pi) of WKB corrected energy distribution of 
! (adiabatically) corrected initial wave packet to file 'enerd'.
!  ez = minimum of H2 potential (if separated atoms correspond to
!                                  E=zero)
!-----------------------------------------------------------------------
!      ez = -4.74746_dop/eV
      ez = 0.0_dop
      do n=1,ne
         write (ienerd,*,err=10)  &
              (e0eigen-ez+emin+n*de)*eV, abs(edstr(n))**2
      enddo
      close (ienerd)
      return

 10   routine = 'ADIABWKB'
      message = 'Cannot open or write to file enerd'
      call errormsg

      return
      end subroutine adiabwkbhh2

      end module adiabwkbmod


