      module enerdstrmod

use decimal, only: long, dop

      implicit none
      private
      public :: energdstr
      contains

!***********************************************************************
!
!                MCTDH   subroutine ENERGDSTR
!
!  Calculates the (a)diabatically corrected energy distribution
!  edsrt(n) (or DELTA(E)) needed for the flux analysis by overlaping  
!  the initial translational spf with the Jost solution chi. 
!  Here chi is no longer evaluated by WKB but numerically integrated 
!  using the Numerov algorithm. The potential which chi suffers is the 
!  translational mean-field tmf. 
!
!  Note: It is assumed that the translational grid is equidistant!
!        Hence only sine-DVR, exp-DVR or FFT are allowed.
!        It is assumed that the initial translational spf phi is
!        located at large grid points and moves towards smaller grid points.
!
!  INPUT: 
!     phi   : translational wave packet (spf).
!     tmf   : adiabatic potential.
!     ne    : number of energy points for which the energy distribution 
!             is determined. 
!     xi, xf: first and last grid point. 
!     gdim  : number of grid points.
!     mass  : mass for the translational mode.
!
!  OUTPUT: 
!     edstr : energy distribution.
!     emin  : lowest energy for which edstr is evaluated is emin+de. 
!     de    : engergy step. Highest energy for which edstr is evaluated
!             is emin + de*ne.  
!
!
!  10/96 AJ ; New Version 09/99 HDM, 11/01 HDM
!
!***********************************************************************

      subroutine energdstr(edstr,de,emin,ne,gdim,xi,xf,mass, &
                           tmf,phi,ilog,vexp,vvar,ev)

      implicit none

      integer, parameter :: long  = selected_int_kind(9)
      integer, parameter :: dop = selected_real_kind(15,307)

      integer(long)             :: g,n,gmin,gmax,k,gwkb
      integer(long), intent(in) :: gdim,ne,ilog

      real(dop)       :: tmf0,tmf1,vorz,en,dx,pi,c12,c56,p1,p2,p3,p4,&
                         pmax,pmin,msum,ekin,dx2m,pq1,pq2,pq3,pq4,pq5,pq6, &
                         sumphi,pwkb0,pwkb1,pwkb2,pwkb3,pwkb4,&
                         pwi0,pwi1,pwi2,threshold,pmean,pp

      real(dop), dimension(gdim), intent(in) ::  tmf
      real(dop), intent(in)                  ::  xi,xf,mass,ev
      real(dop), intent(out)                 :: de,emin,vexp,vvar

      complex(dop)                           :: chi1,chi2,chi3,chi4,chi5,&
                                                chi6,csum1,csum2,csum3,csum4
      complex(dop),dimension(ne),intent(out) :: edstr
      complex(dop),dimension(gdim),intent(in):: phi

!-----------------------------------------------------------------------
! Compute grid spacing in [au] and Pi and set constants.
!-----------------------------------------------------------------------
      dx   = (xf-xi)/(gdim-1)
      pi   = 4.0_dop*atan(1.0_dop)
      dx2m = 2.0_dop*mass*dx*dx
      c12  = 1.0_dop/12.0_dop/16.0_dop   ! Divison by 16 because there are four
      c56  = 5.0_dop/6.0_dop/16.0_dop    ! Numerov steps per gid interval.
      
!-----------------------------------------------------------------------
! Determine lowest and highest grid point of non-vanishing phi.
!-----------------------------------------------------------------------
      do g = 1, gdim
         gmin = g
         if (abs(phi(g)) .gt. 1.0e-8_dop ) goto 21
      end do
 21   do g = gdim, 1, -1
         gmax = g
         if (abs(phi(g)) .gt. 1.0e-8_dop ) goto 22
      end do

 22   sumphi = 0.0_dop
      do g = gmin, gmax
         sumphi = sumphi + abs(phi(g))
      end do

      if( gmin .le. 2  .or.  gmax .ge. gdim-1 ) then
         write(6,930) gmin, gmax, gdim
         write(ilog,930) gmin, gmax, gdim
 930     format(/'## WARNING! The initial wave packet is too wide',&
            ' for the grid.', &
           /12x,'gmin = ',i3,',  gmax =',i4,',  gdim =', i4 /)  
         gmin = max(gmin,2)
         gmax = min(gmax,gdim-1)
      end if

      write(ilog,'(a,2i4,a,2f9.3)') 'energdstr:  gmin, gmax =',& 
                gmin, gmax,' ,  xmin, xmax =',&
                xi+(gmin-1)*dx,xi+(gmax-1)*dx 

!-----------------------------------------------------------------------
! Determine emin as maximal potential energy on grid points larger gmax-2.
!-----------------------------------------------------------------------
      emin = tmf(gmax-2)
      do g = gmax-1, gdim
         emin = max(emin,tmf(g))
      enddo

!-----------------------------------------------------------------------
! Determine vmin as minimal potential energy between gmin-gmax.
! Determine potential energy expectation value.
!-----------------------------------------------------------------------
      vexp = 0.0_dop
      do g = gmin, gmax
         vexp = vexp + tmf(g)*abs(phi(g))**2
      enddo
      vvar = 0.0_dop
      do g = gmin, gmax
         vvar = vvar + ( (tmf(g)-vexp)*abs(phi(g)) )**2
      enddo
      vvar = sqrt(vvar)

!-----------------------------------------------------------------------
! Determine (approximately, by finite differences) expectation value
! and variance of momentum. This provides direction of momentum (vorz),
! as well a an initial interval for search.
!-----------------------------------------------------------------------
      pmean= 0.d0
      pp   = 0.d0
      do g = gmin+1,gmax-1
         pmean=pmean - aimag(dconjg(phi(g))*(phi(g-1)-phi(g+1)))
         pp = pp - dble(dconjg(phi(g))*(phi(g-1)+phi(g+1)-2.d0*phi(g)))
      enddo
      pmean=pmean/(2.d0*dx)
      pp=pp/(dx*dx)
      vorz= -pmean/abs(pmean)
      pmean=abs(pmean)
      pmax=pmean + 1.3d0*sqrt(abs(pp-pmean*pmean))
      pmax=min(pmax,pi/dx)
      pmin=sqrt(pp)
      if(100.d0*pmin .lt. pmax) then
         write(6,'(/a,/a,/a,f9.4)') '### ERROR (enerdstr.F) ###',&
          ' The width of the initial wavepacket seems to be too small',&
          ' for the given grid spacing: dx =',dx
         write(ilog,'(/a,/a,/a,f9.4)') '### ERROR (enerdstr.F) ###',&
          ' The width of the initial wavepacket seems to be too small',&
          ' for the given grid spacing: dx =',dx
         stop 4
      endif

!-----------------------------------------------------------------------
! Determine ekin as maximal kinetic energy of the wave packet.
! i.e. |DELTA(Ekin)| < 1.d-2 * max(DELTA(E)). 
!-----------------------------------------------------------------------
      threshold = sumphi*1.0e-2_dop
      msum=0.0_dop
      do k = 1, 20
         p1 = 0.2d0*(pmax-pmin) + pmin
         p2 = 0.4d0*(pmax-pmin) + pmin
         p3 = 0.6d0*(pmax-pmin) + pmin
         p4 = 0.8d0*(pmax-pmin) + pmin

!.....Do the fourier integral.
         csum1 = 0.5_dop*( exp(dcmplx(0.0_dop,p1*gmin*dx*vorz))*phi(gmin) +&
                         exp(dcmplx(0.0_dop,p1*gmax*dx*vorz))*phi(gmax) )
         csum2 = 0.5_dop*( exp(dcmplx(0.0_dop,p2*gmin*dx*vorz))*phi(gmin) +&
                         exp(dcmplx(0.0_dop,p2*gmax*dx*vorz))*phi(gmax) )
         csum3 = 0.5_dop*( exp(dcmplx(0.0_dop,p3*gmin*dx*vorz))*phi(gmin) +&
                         exp(dcmplx(0.0_dop,p3*gmax*dx*vorz))*phi(gmax) )
         csum4 = 0.5_dop*( exp(dcmplx(0.0_dop,p4*gmin*dx*vorz))*phi(gmin)+&
                         exp(dcmplx(0.0_dop,p4*gmax*dx*vorz))*phi(gmax))
         do g = gmin+1,gmax-1
            csum1 = csum1 + exp(dcmplx(0.0_dop,p1*g*dx*vorz))*phi(g)
            csum2 = csum2 + exp(dcmplx(0.0_dop,p2*g*dx*vorz))*phi(g)
            csum3 = csum3 + exp(dcmplx(0.0_dop,p3*g*dx*vorz))*phi(g)
            csum4 = csum4 + exp(dcmplx(0.0_dop,p4*g*dx*vorz))*phi(g)
         end do

         msum = max(msum,abs(csum1),abs(csum2),abs(csum3),abs(csum4))
         if( abs(csum4) .gt. threshold ) then
            pmax=min(0.2d0*(pmax+4.d0*pi/dx),2.d0*pmax-pmin)
            pmin = (5.d0*p4 - pmax)*0.25d0
         else if( abs(csum3) .gt. threshold ) then 
            pmin = p2
            pmax = p4
         else if( abs(csum2) .gt. threshold ) then
            pmin = p1
            pmax = p3
         else if( abs(csum1) .gt. threshold ) then
            pmax = p2
         else
             if(abs(csum1) .gt. abs(csum2) .and. &
                abs(csum2) .gt. abs(csum3)) then
                pmin=0.80d0*pmin
                pmax=0.80d0*pmax
             elseif(abs(csum3) .lt. abs(csum4) .and. &
                    abs(csum2) .lt. abs(csum3)) then
                pmin=1.25d0*pmin
                pmax=min(1.25d0*pmax,pi/dx)
             else
                if(p2.lt.0.5d0*pi/dx) then
                   pmin=1.25d0*pmin
                   pmax=min(1.25d0*pmax,pi/dx)
                else
                   pmin=0.80d0*pmin
                   pmax=0.80d0*pmax
                endif
             endif
         end if
!........Stop iteration if converged.
         if(abs(csum4) .lt. threshold .and. &
            abs(csum4) .gt. 0.4d0*threshold) goto 70
         if(abs(csum3) .lt. threshold .and. &
            abs(csum3) .gt. 0.4d0*threshold) goto 70
         if( (pmax-pmin) .lt. 0.01d0*(pmax+pmin) .and. &
              abs(csum4) .lt. threshold .and. &
              msum .gt. 3.d0*threshold ) goto 70
        if(pmin .gt. 0.98d0*pi/dx) goto 70
      end do
      write(6   ,'(a)') 'enerdstr : No convergence!'
      write(ilog,'(a)') 'enerdstr : No convergence!'

 70   if( abs(csum1) .le. threshold ) then
         ekin = 0.5_dop*p1*p1/mass
      else if( abs(csum2) .le. threshold ) then
         ekin = 0.5_dop*p2*p2/mass
      else if( abs(csum3) .le. threshold ) then
         ekin = 0.5d0*p3*p3/mass
      else
         ekin = 0.5d0*p4*p4/mass
      end if

      de = ekin/dble(ne)
      write(ilog,'(a,4f11.3,/a,f10.3,a,2f10.3)') &
       'p1, p2, p3, p4 : ',-vorz*p1, -vorz*p2, -vorz*p3, -vorz*p4, &
       'Ekin_max = ', ekin*ev,' eV,  E_min, E_max =', &
         (emin+de)*ev,(emin+ekin)*ev

!.....Check for possible errors..
      if(p4 .gt. 0.92d0*pi/dx .and. abs(csum4).gt.threshold) then
         write(6,900) dx, ekin*ev
         write(ilog,900) dx, ekin*ev
 900     format(/'## WARNING! The grid spacing dx =',f8.3,/13x, &
            'is too large to fully cover the energy distribution.', &
           /13x,'Ekin_max = ',f9.3,' eV'/)
      else if(min(abs(csum1),abs(csum2),abs(csum3),abs(csum4)) .gt. &
             threshold .or. (abs(csum3).gt.abs(csum2)) .or. &
                            (abs(csum4).gt.abs(csum3)) ) then
        write(6,920) ekin*ev
        write(ilog,920) ekin*ev
 920    format(/'## WARNING! The determination of the maximal energy' &
               /12x,'(subroutine energdstr) did not converge!', &
               /12x,'Ekin_max = ',f9.3,' eV'/)
      end if
      if( msum .lt. 5.d0*threshold .or. abs(csum3).gt.abs(csum2) &
               .or. abs(csum2).gt.abs(csum1) ) then
        write(6,910) ekin*ev
        write(ilog,910) ekin*ev
 910    format(/'## WARNING! Subroutine energdstr may not have found', &
       /12x,'the correct energy interval for the energy distribution.', &
               /12x,'Ekin_max = ',f9.3,' eV'/)
      end if

!***********************************************************************
! End of preliminaries. Actual calculations starts here.
!***********************************************************************

!-----------------------------------------------------------------------
! Loop over energies
!-----------------------------------------------------------------------
      do n = 1, ne
         en = emin + n*de
         
!-----------------------------------------------------------------------
! Determine the energy distribution "edstr" for energy "en".
! First try a WKB solution (Eq. 140, review), if this becomes too
! inaccurate (turning point problem) continue with Numerov-integration
! of the Jost solution.
!-----------------------------------------------------------------------
         pwkb0= sqrt(2.0_dop*mass*(en-tmf(gmax)))
         pwkb1= sqrt(2.0_dop*mass*(en-tmf(gmax-1)))
         edstr(n) = 0.5_dop*phi(gmax)/sqrt(pwkb0) 
         gwkb = gmax
         pwi0 = 0.0_dop
 500     pwkb2= sqrt(2.0_dop*mass*(en-tmf(gwkb-2)))
         if(1.d2*(abs(pwkb0-pwkb1)+abs(pwkb1-pwkb2))+1.0_dop &
            .lt. dx*min(pwkb0,pwkb1,pwkb2)**2 ) then
            gwkb = gwkb - 1
            pwi0 = pwi0 + (pwkb0 + pwkb1)*0.5_dop*dx
            chi2 = exp(dcmplx(0.0_dop,-pwi0*vorz))/sqrt(pwkb1)
            edstr(n) = edstr(n) + phi(gwkb)*chi2 
            pwkb0= pwkb1
            pwkb1= pwkb2
            if(gwkb.gt.gmin) then 
               goto 500
            elseif(gwkb.eq.gmin) then
               edstr(n) = edstr(n) - 0.5_dop*phi(gwkb)*chi2 
               goto 510
            endif
         endif

!-----------------------------------------------------------------------
! Get Jost solution "chi" via Numerov integration when WKB becomes 
! inaccurate. The WKB solutions at the points gwkb-3/4 and gwkb-1  
! serve as the two initial values for the Numerov recursion.
! pwkb = local momentum, pq = (pwkb*dx)**2, pwi = integral[pwkb dx]
!-----------------------------------------------------------------------

         pq1  = dx2m*(en-(3.0_dop*tmf(gwkb-1)+tmf(gwkb))*0.25_dop)
         pq2  = dx2m*(en-tmf(gwkb-1))
         pwkb3= sqrt(2.0_dop*mass*(en-tmf(gwkb)))
         pwkb4= sqrt(2.0_dop*mass*(en-(tmf(gwkb-1)+3.0_dop*tmf(gwkb))*0.25_dop))
         pwkb0= sqrt(2.0_dop*mass*(en-(tmf(gwkb-1)+tmf(gwkb))*0.5_dop))
         pwkb1= sqrt(2.0_dop*mass*(en-(3.0_dop*tmf(gwkb-1)+tmf(gwkb))*0.25_dop))
         pwkb2= sqrt(2.0_dop*mass*(en-tmf(gwkb-1)))
         pwi0 = pwi0 + (pwkb3 + 2.0_dop*pwkb4 + pwkb0)*0.125_dop*dx
         pwi1 = pwi0 + (pwkb0+pwkb1)*0.125_dop*dx
         pwi2 = pwi1 + (pwkb1+pwkb2)*0.125_dop*dx
         chi1 = exp(dcmplx(0.0_dop,-pwi1*vorz))/sqrt(pwkb1)
         chi2 = exp(dcmplx(0.0_dop,-pwi2*vorz))/sqrt(pwkb2)


         tmf1 = min(tmf(gwkb-1),en+0.5_dop/mass)
!........The minimum is taken to avoid excessive groth of the WF
!........in the classical forbidden region.

         do g = gwkb-1, gmin+1, -1
            tmf0 = tmf1
            tmf1 = min(tmf(g-1),en+0.5_dop/mass)
            pq3 = dx2m*(en-(tmf1+3.0_dop*tmf0)*0.25_dop)
            pq4 = dx2m*(en-(tmf1+tmf0)*0.5_dop)
            pq5 = dx2m*(en-(3.0_dop*tmf1+tmf0)*0.25_dop)
            pq6 = dx2m*(en-tmf1)
!...........Mumerov recursion (four steps per grid interval).
            chi3 = ( (2.0_dop-pq2*c56)*chi2 - (1.0_dop+pq1*c12)*chi1 ) &
                    /(1.0_dop+pq3*c12)
            chi4 = ( (2.0_dop-pq3*c56)*chi3 - (1.0_dop+pq2*c12)*chi2 ) &
                    /(1.0_dop+pq4*c12)
            chi5 = ( (2.0_dop-pq4*c56)*chi4 - (1.0_dop+pq3*c12)*chi3 ) &
                    /(1.0_dop+pq5*c12)
            chi6 = ( (2.0_dop-pq5*c56)*chi5 - (1.0_dop+pq4*c12)*chi4 ) &
                    /(1.0_dop+pq6*c12)
!...........Do the integral and shift the solution.
            edstr(n) = edstr(n) + chi2*phi(g)
            pq1  = pq5
            pq2  = pq6
            chi1 = chi5
            chi2 = chi6
         enddo

         edstr(n)=edstr(n) + 0.5_dop*chi2*phi(gmin)
 510     edstr(n)=edstr(n)*sqrt(dx*mass/(2.0_dop*pi))
      enddo

      return
      end subroutine energdstr

      end module enerdstrmod
