!***********************************************************************
!
!               MCTDH   module ORTHOWF
!
! This module orthogonalises the wavefunction against foreign restart files
!
!  orthowf : orthogonalizes the initial psi function
!  rdfrst  : read foreign resart file-info.
!  zeigowf : Zeig routine for orthogonalization of WF.
!
!***********************************************************************
      module orthowfmod

use decimal, only: dop, long
use logdat
use mmlib,   only: mmaxzz
use mtlib,   only: mtxxzz
use xvlib,   only: xvixdzo, vvaxzz, xvxxzzr
use op1lib,  only: normvxz, cpvxz
use global
use griddatmod
use psidef
use crocoflymod
use compdatmod
use runinwfmod
use operdef
use channels

use compdat, only: savecsys,savecpsi,getcsys,getcpsi
use iopsidef, only: psidat
use iorst, only: rstinfo

      implicit none
      private
      public :: orthowf,zeigowf
      contains

!***********************************************************************
!     SUBROUTINE ORTHOWF                                                
!                                                                       
! This routine orthogonalizes the initial psi function (restart)  
! against 'foreign' restart files. Used for improved relaxation.        
!                                                                       
! 02/2004 HDM                                                            
!                                                                       
!***********************************************************************

      subroutine orthowf(chan,psi,psiref,dover)

      implicit none

      complex(dop)                                   :: zdum
      complex(dop), dimension(*), intent(out)        :: psiref
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(doverdim), intent(out) :: dover
      complex(dop), dimension(workcdim)              :: workc
      real(dop)                                      :: time1,norm
      integer(long)                                  :: zeig1,zeig2,swapzeig,&
                                                        s,m,d
      integer(long), intent(in)                      :: chan

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
      call rdfrst(chan,0)
      read (chan,err=999) time1
!      if (citype .gt. 0) then
!         read(chan,err=999) (jindx(b),b=1,adim)
!      endif
      read (chan,err=999) (psiref(d),d=1,cdgldim(2))

!-----------------------------------------------------------------------
! Set dovmat
!-----------------------------------------------------------------------
      zeig1=1
      do s=1,nstate
         do m=1,nmode
            dovmat(m,s)=zeig1
            zeig1=zeig1+cdim(m,s,2)*dim(m,s)
         enddo
      enddo


!-----------------------------------------------------------------------
!   calculate density matrices for each degree of freedom defined as
!   overlap of the single-particle functions rho^m_ij=<phi_i|phi_j,ref>
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call mmaxzz(psi(zetf(m,s)),psiref(czetf(m,s,2)), &
                 dover(dovmat(m,s)),subdim(m),dim(m,s),cdim(m,s,2))
         enddo
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J|Phi_L,ref>A_L,ref within each state
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig2=zpsi(s)
         zeig1=zeig2+ccwcdim
         call cpvxz(psiref(czpsi(s,2)),workc(zeig2),cblock(s,2))

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(dover(dovmat(m,s)),workc(zeig1),workc(zeig2), &
                        vdim(m,s),cdim(m,s,2),cndim(m,s,2),dim(m,s))
         enddo
!--------Move data to first part of workc, if necessary.
         if(zeig1.lt.zeig2) & 
            call cpvxz(workc(zeig2),workc(zeig1),block(s))
      enddo
!-----Now workc contains the transformed A-vector of psiref.

!-----------------------------------------------------------------------
! calculate overlap of transformed reference A-vector with
! original A-vector and orthogonalize.
!-----------------------------------------------------------------------
      call normvxz(workc,norm,adim)       ! compute norm of workc
      call xvixdzo(norm,workc,adim)       ! normalise workc
      write(ilog,'(/a,2f10.6)') & 
        'Norm and norm**2 of transformed ref. A-vector:',norm,norm**2 
      call vvaxzz(workc,psi,zdum,adim)    ! compute overlap (zdum)
      call xvxxzzr(zdum,workc,psi,adim)   ! orthogonalise
      call normvxz(psi,norm,adim)         ! compute norm
      call xvixdzo(norm,psi,adim)         ! normalise
      write(ilog,'(a,i3,a,f10.6,/a,f10.6,a,f10.6,a,f10.6)')  &
          'Norm after orthogonalisation against channel:', chan, &
          ', |ovl|**2 =', abs(zdum)**2,  &
          'norm =', norm, ' , norm**2 =', norm**2, &
          ' , 1-norm**2 =', 1.0_dop-norm**2

!-----Orthonormalize twice, if norm was small
      if(norm .le. 0.2_dop) then
         call vvaxzz(workc,psi,zdum,adim)   ! compute overlap (zdum)
         call xvxxzzr(zdum,workc,psi,adim)  ! orthogonalise
         call normvxz(psi,norm,adim)        ! compute norm
         call xvixdzo(norm,psi,adim)        ! normalise
      endif

!-----------------------------------------------------------------------

      return

  999 continue
      write(message,'(a,i4)') &
           'Error reading foreign restart file from channel',chan
      routine='orthowf'
      call errormsg

      end subroutine orthowf

!***********************************************************************
!
! subroutine RDFRST
!
! Reads the restart-info of a foreign restart file (reference WF).
! Checks if this WF's system definition matches the current system.
! System def. of reference WF is stored into comparison data set 2.
!
! Output variables:
!   (to common blocks)
!   ovtype,ebtype (see "include/crocofly.inc")
!   + everything declared in "include/compdat.inc"
!
! 02/04, HDM  (partly copied from subroutine RDCCREF1 on crocofly.F)
!
!***********************************************************************

      subroutine rdfrst(chan,chk)

      implicit none

      logical(kind=4)           :: linwf, lerr, readsys
      integer(long)             :: chkdvr, chkgrd, chkpsi, chkprp, irst0
      integer(long), intent(in) :: chan, chk

!     save system definition
      call savecsys(1)
      call savecpsi(1)
      irst0=irst
      irst=chan

      if(chk.eq.1) then
!        tell the LOG file what's going to happen
         write(ilog,'(/a,i3,a)') &
              '***** ORTHOWF: reading system definition for '// &
              'reference wavefunction',max(0,chan-380),' *****'

!        check system definition
!        DVR and GRID definitions should match current system (CHK=2)
!        PSI definiton can differ and will be read (CHK=1)
!        PRP definition is read but ignored (CHK=0)
         chkdvr=2
         chkgrd=20
         chkpsi=1
         chkprp=0
      else
         chkdvr=1
         chkgrd=1
         chkpsi=1
         chkprp=0
      endif
      call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

!     check if some checks failed
      readsys = .false.
      if (lerr) then
         routine = 'RDFRST (ORTHOWF)'
!        investigate which checks failed
         if (chkerr(1) .ne. 0) then
!           I/O error: fatal
            message = 'Error reading reference WF'
            call errormsg
         endif
         if (chkerr(2) .ne. 0) then
!           Definitions of DOFs differ: fatal
            message = 'DOF definitions of ref.WF'// &
                      'don''t match system.'
            call errormsg
         endif
         if (chkerr(3) .ne. 0) then
!           Definitions of DVR basis differ: fatal
            message = 'DVR basis definitions of ref.WF'// &
                      'don''t match system.'
            call errormsg
         endif
         if (chkerr(4) .ne. 0) then
!           Mode combination definitions differ: fatal
            message = 'Definition of mode combinations '// &
                      'differs from system.'
            call errormsg
         endif
         if (chkerr(5) .ne. 0) then
!           Electronic basis type differs: fatal
            message = 'Electronic basis type of ref.WF '// &
                      'differs from system.'
            call errormsg
         endif
!        chkerr(6:9) were not affected by the RSTINFO call
         if (chkerr(10) .ne. 0) then
!           Density types differ: fatal
            message = 'Reference WF uses '// &
                 'a density type different from system.'
            call errormsg
         endif
      endif

!     store the system definition of the reference WF
      call psidat
      call savecsys(2)
      call savecpsi(2)

!     restore original system definition
      call getcsys(1)
      call getcpsi(1)
      call psidat
      irst=irst0

      return
      end subroutine rdfrst

!***********************************************************************
!
! subroutine ZEIGOWF
!
! Calculates the size of the (static) memory needed for the
! orthogonalisation of the WF (improved relaxation).
! As the arithmetic is similar to cross-correlation, crocofly.inc is used.
!
! Output variables:
!   (to common blocks)
!   psirefdim : size of reference WF
!   ccjindxdim: size of JINDX array of ref. WF
!   ccwcdim: workc-space needed for this module
!   doverdim,dovmat: needed later in ORTHOWF
!
! 02/04, HDM
!
!***********************************************************************

      subroutine zeigowf

      implicit none

      integer(long) :: zeig,zeig1,m,s,chan,aadim

      ccwcdim=0
      psirefdim=0
      doverdim=0
      aadim=0

      do chan = 381, iorthowf
!        obtain system definition of reference WF from "restart" file
         if(chan.eq.equalfile) then
            call rdfrst(irst,1)
            close(irst)
         else
            call rdfrst(chan,1)
         endif

!        calculate size of reference WF
         psirefdim = max(psirefdim,cdgldim(2))
         aadim = max(aadim,cadim(2))

!        memory needed for JINDX arrray
         if (ccitype(2) .gt. 0) then
            ccjindxdim = cadim(2)
         else
            ccjindxdim = 1
         endif

!        memory needed for the orthogonalize calculation
         zeig=1
         do s=1,nstate
            ccwcdim=max(ccwcdim,block(s),cblock(s,2))
            zeig1=cblock(s,2)
            do m=1,nmode
               dovmat(m,s)=zeig
               zeig=zeig + dim(m,s)*cdim(m,s,2)
               zeig1=zeig1*dim(m,s)/cdim(m,s,2)
               ccwcdim=max(ccwcdim,zeig1)
            enddo
         enddo
         doverdim=max(doverdim,zeig-1)

         ! edit MS (08/2019)
         dmatdim = max(dmatdim,cdmatdim(2))
         d3matdim = max(d3matdim,cd3matdim(2))
         ! end edit MS

      enddo

      workcdim = max(workcdim, 2*ccwcdim+zpsi(nstate))

      return
      end subroutine zeigowf

!***********************************************************************
      end module orthowfmod

