!#######################################################################
! Computation of FBR-integrals in a spherical harmonics basis
! defined by the set of quantum numbers jsph(subdim) and msph(subdim)
! 
! cc3 : array containing the integrals
! cc3t: working array containing the integrals computed with a 
!       smaller number of quadrature points
!
!-----------------------------------------------------------------------
      module fbrintmod

use decimal, only: long, dop
use global
use maxv, only: maxdim
use dvrdatmod
use griddatmod
use operdef
use runopermod
use channels

use potevalmod, only: poteval
use op1lib,  only: zeromxz, cpqxz, zeromxd


      implicit none
      private
      public :: fbrint
      contains 

!-----------------------------------------------------------------------
      subroutine fbrint(h,jsph,msph,jmax,cc3,subdim1,dof1,j25)

      implicit none

      integer(long)                                 :: i,j,ii,jj,li,mi,lj,mj,inum,m,&
                                                       jm,ngg1,ngg2,joff,dof2,idum
      integer(long), intent(in)                     :: h,subdim1,dof1,j25
      integer(long), intent(out)                    :: jmax
      integer(long), dimension(subdim1), intent(in) :: jsph,msph

      real(dop)                                             :: v,fact,thr,tt2
      real(dop), external                                   :: plgndr
      real(dop), dimension(maxdim)                          :: rdum
      real(dop), dimension(maxdim)                          :: gpoint
      complex(dop)                                          :: ket,bra,vz
      complex(dop), dimension(subdim1,subdim1), intent(out) :: cc3
      complex(dop), dimension(subdim1,subdim1)              :: cc3t
      logical(kind=4)                                       :: lintlog

! - dimension is adopted larger than spherical harmonics DVR -
! - separated from main  because of accuracy for quadruture  -
! - jmax,subdim,jsph,msph : same as main routine.

      real(dop), dimension(subdim1)              :: factor,plm
      real(dop), dimension(j25)                  :: tx,tw
      real(dop), dimension(2*j25)                :: ty
      real(dop), dimension(j25,j25)              :: trfo
! -   tx: theta-grid; ty: phi-grid; tw: theta-weigths

      jmax = ipbaspar(2,dof1)
      joff = ipbaspar(3,dof1)
      thr  = rpbaspar(1,dof1)

!  calculation of factor
      do i=1,subdim1
         li=jsph(i)
         mi=msph(i)
         mi=iabs(mi)       
         fact=1.0_dop
         do j=li-mi+1,li+mi
            fact=fact*j
         enddo
         factor(i)=sqrt(dble(2*li+1)/(fact*4.0_dop*pi))
      enddo

!  Integrals are evaluated using more quadrature points than FBR functions 
!  Iterative computation until convergence is achieved, in the limit of
!  nit cycles.
!  inum = current cycle number
!  jm, ngg1, ngg2 : respectively jmax, number of theta values,
!  and number of phi values of the current cycle.
!  jm is always larger than jmax


! - Initialize iteration loop -
      inum=0
      call zeromxz(cc3,subdim1,subdim1)
 
! - Iteration loop -
 1000 continue
      inum=inum+1
      jm=(jmax+joff)+2*(inum-1)
      ngg1=jm+1
      ngg2=2*jm+1
      if (jm.ge.j25) then
         write(ilog,*) 'sphFBR-Integrals not converged! '
         write(6,*)    'sphFBR-Integrals not converged! '
         goto 1100
      endif

      call cpqxz (cc3,cc3t,subdim1)
      call zeromxz(cc3,subdim1,subdim1)

! - integral calculation -
!---------------------------------------------------------------------
! - Determine nodes and weights of Gauss-Legendre quadrature -
      call calsg (trfo,tx,tw,ngg1,ty)

! - phi grid points -
      do i=1,ngg2
         ty(i)=2.0_dop*pi*(i-1)/dble(ngg2)
      enddo

! - multiplication of theta-weights with the weights of the phi mode:
      do i=1,ngg1
         tw(i)=tw(i)*2.0_dop*pi/dble(ngg2)        
      enddo

! - print check -
      lintlog=.false.
      if (lintlog) then
          write(ilog,*) 'iteration :',inum
          write(ilog,'(a,3i3,a)')  &
              '--- jm,ngg1,ngg2=',jm,ngg1,ngg2,' run'
          write(ilog,'(a)') '--- Grid points for theta (sph-DVR) ---'
          write(ilog,'(10f8.4)') (tx(i),i=1,ngg1)
          write(ilog,'(a)') '--- Weights               (sph-DVR) ---'
          write(ilog,'(10f8.4)') (tw(i),i=1,ngg1)
          write(ilog,'(a)') '--- Grid points for phi   (sph-DVR) ---'
          write(ilog,'(10f8.4)') (ty(i),i=1,ngg2)
      endif

! - find dofs in FBR mode
      m=dofspf(dof1)
      dof2=spfdof(2,m)

! - loop -
      do ii=1,ngg1                ! loop over theta grid points
         do i=1,subdim1           ! associated Legendre functions
            li=jsph(i)
            mi=msph(i)
            plm(i)=factor(i)*plgndr(li,iabs(mi),cos(tx(ii)))
            if (mi.lt.0 .and. mod(iabs(mi),2).eq.1) plm(i)=-plm(i)
            if (abs(plm(i)).gt.10.0_dop) then   ! Check of overfolw
               write(ilog,'(a,2i5,f10.5,e16.6)')  &
                    'fbrint.F  warning:',li,mi,tx(ii),plm(i)
               stop
            endif
         enddo

         do jj=1,ngg2             ! loop over phi grid points
            gpoint(dof1)=tx(ii)
            gpoint(dof2)=ty(jj)
            call poteval(h,v,gpoint,vz,rdum,idum)
 
            do i=1,subdim1        ! loop over ket-states
               li=jsph(i)
               mi=msph(i)
               ket=plm(i)*exp(dcmplx(0.0_dop,dble(mi)*ty(jj)))

               do j=1,subdim1     ! loop over bra-states
                  lj=jsph(j)
                  mj=msph(j)
                  bra=plm(j)*exp(dcmplx(0.0_dop,dble(mj)*ty(jj)))

! === INTEGRALS ===

                  cc3(i,j)=cc3(i,j)+dconjg(bra)*ket*vz*tw(ii)
               enddo
! end of j loop over <Ylm|
            enddo
! end of i loop over |Ylm>
         enddo
! end of jj loop over phi grid points
      enddo
! end of ii loop over theta grid points

!---------------------------------------------------------------------

! - result check convergence -
      tt2=0.0_dop
      do i=1,subdim1
         do j=1,subdim1
            tt2=max(tt2,cdabs(cc3(i,j)-cc3t(i,j)))
         enddo
      enddo
      if (tt2.gt.thr*2**inum .and. thr.lt.1.0_dop) then
         if (lintlog) write(ilog,*)'current convergence: ',tt2
         goto 1000
      end if

 1100 write(ilog,'(/,a,1pe10.3)') &
                   ' === Integrals converged with error :',tt2
      write(ilog,*) ' number of iterations performed    :',inum
      write(ilog,*) ' number of grid points theta, phi  :',ngg1,ngg2 
      call myflush(ilog)

      return
      end subroutine fbrint
 
!#######################################################################

      subroutine calsg (trafo,ort,weight,dim,work)
      implicit none

      integer(long)                              :: error,i
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim), intent(out)     :: ort,weight
      real(dop), dimension(2*dim), intent(out)   :: work
      real(dop), dimension(dim,dim), intent(out) :: trafo


! - Diagonalize coordinate matrix -

! - Set diagonal elements -
      call zeromxd(trafo,dim,dim)
      do i=1,dim
         trafo(i,i)=1.0_dop
         ort(i)=0.0_dop
      enddo

! - Set offdiagonal elements -
      do i=1,dim-1
         weight(i)=dsqrt((i**2)/(4.0_dop*i*i-1.0_dop))
      enddo

! - diagonalize -
      call DSTEQR('V',dim,ort,weight,trafo,dim,work,error)
      if (error.ne.0) then
         write (6,*) 'fehler ',error,' beim diagonalisieren'
         stop
      endif

! - calculate grid -
      do i=1,dim
         ort(i)=acos(ort(i))
      enddo

! - calculate weights -
      do i=1,dim
         weight(i)=2.0_dop*trafo(1,i)**2
      enddo

      return
      end subroutine calsg

      end module fbrintmod


