!#######################################################################
!
!              MCTDH  module HEINGABE
!
! MCTDH module to read and sort the data specifying the Hamiltonian
!
!            Subroutine      HEINGABE
!
! Controls the subroutines that read and sort the Hamiltonian data.
! Called from: mctdh/eingabe
! Calls to: warnmsg, ErrorMsg,
!           abspath getpath heinopdef rdinpf gethnum heinterms
!           deinterms heinlab matchlab findlabels sorteb zeigrdf
!           assignhlab
!
!#######################################################################

      module heingabemod

use decimal, only: long, dop
use global
use logdat
use maxv
use dvrdatmod
use griddatmod
use operdef
use runopermod
use runpropmod
use paths
use channels

use heintermmod, only: heinterms
use heinutil, only: gethnum
use matchlabmod, only: matchlab,findlabels
use heintermmod, only: deinterms

use opermod_interface

      implicit none
      private
      public :: heingabe,genopdof
      contains

!#######################################################################

      subroutine heingabe(pass,loperr)

      implicit none

      integer(long), intent(in)       :: pass
      integer(long)                   :: iz,ic,ierr,k,f,nham,hcount
      integer(long), external         :: slen
      logical(kind=4), intent(inout)  :: loperr

      character(len=c3)       :: label

!----------------------------------------------------------------------
! counter for system Hamiltonians (to check that only 1 defined)
! For direct dynamics if reading from input file then system H has
! been set up automatically in cvmcgop
!----------------------------------------------------------------------
      if ((iop .eq. iin) .and. (ldd .or. lddmctdh)) then
         hcount =1
      else
         hcount=0
      endif

!-----------------------------------------------------------------------
! allocate no. of DOFs in operator, and set up space for koeffs, 
! electronic basis etc.
!-----------------------------------------------------------------------
      call genopdof

!-----------------------------------------------------------------------
! Title of operator is read and written in log file
!-----------------------------------------------------------------------
      rewind(iop)
      iz=0
  110 continue

      call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
         routine = 'Heingabe, op_define-section'
         message = inptit
         call errormsg
      endif

      if (keyword(1) .eq. 'op_define-section') then
         if (pass .eq. 2) call heinopdef(loperr,iz)
      else if (keyword(1) .eq. 'end-operator' .or. &
           keyword(1) .eq. 'end-input') then
         continue
         if (pass .eq. 2) then
            if (ldd .or. lddmctdh) then
               write(ilog,'(/a)') '######    DIRECT DYNAMICS OPERATOR   ######'
               write(ilog,'(a)') '          Internally generated           '
            else if(iop.eq.iin) then
               write(ilog,'(2a)') &
                'Operator definitions are being ', &
                'read from input file (no OPERATOR-SECTION).'
               write(ilog,'(/a)') '          Untitled Operator              '
               write(ilog,'(a)') '######    No OP_DEFINE-SECTION !   ######'
            endif
         endif 
      else
         go to 110
      endif

!-----------------------------------------------------------------------
! read hamiltonian terms
!-----------------------------------------------------------------------
      rewind(iop)
      iz=0
  310 continue

      call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
         routine = 'Heingabe, hamiltonian-section'
         message = inptit
         call errormsg
      endif

      if (keyword(1)(1:19) .eq. 'hamiltonian-section') then
         if (keyword(1)(20:20) .eq. '_') then
            label=keyorig(1)(21:lc(1))
            call gethnum(nham,label)
         else
            nham=1
            hcount=hcount+1
         endif
         call heinterms(iop,loperr,iz,nham)
         go to 310
      else if (keyword(1) .eq. 'end-operator' .or. &
           keyword(1) .eq. 'end-input') then
         continue
      else
         go to 310
      endif

      if(hcount.eq.0 .and. frunlevel .eq. 4) then
         routine='heingabe'
         message='No system Hamiltonian-Section defined!'
         call errormsg
      else if (hcount.gt.1) then
         routine='heingabe'
         message='There is more than one system Hamiltonian-Section'&
                //' defined in the operator file.'
         call errormsg
      endif

      if (dentype .eq. 0) go to 415
!-----------------------------------------------------------------------
! read dissipative terms
!-----------------------------------------------------------------------
      rewind(iop)
      iz=0
  410 continue

      call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
         routine = 'Heingabe, dissipative-section'
         message = inptit
         call errormsg
      endif

      if (keyword(1) .eq. 'dissipative-section') then
         nham = 1
         if(ldissin) call deinterms(iop,loperr,iz,nham)
      else if (keyword(1) .eq. 'end-operator' .or. &
           keyword(1) .eq. 'end-input') then
         go to 415
      else
         go to 410
      endif
 415  continue

!-----------------------------------------------------------------------
! read any operator label definitions 
! first from iin then iop
!-----------------------------------------------------------------------
      rewind(iin)
      iz=0
  505 continue

      call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
         routine = 'Heingabe, labels-section'
         message = inptit
         call errormsg
      endif

      if (keyword(1) .eq. 'alter-labels') then
         call heinlab(iin,loperr,iz,pass)
      else if (keyword(1) .eq. 'end-input') then
         continue
      else
         go to 505
      endif

      rewind(iop)
      iz=0
  510 continue

      call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
         routine = 'Heingabe, labels-section'
         message = inptit
         call errormsg
      endif

      if (keyword(1) .eq. 'labels-section') then
         call heinlab(iop,loperr,iz,pass)
      else if (keyword(1) .eq. 'end-operator' .or. &
           keyword(1) .eq. 'end-input') then
         continue
      else
         go to 510
      endif

!-----------------------------------------------------------------------
! close .op file
!-----------------------------------------------------------------------
!     if (iop .ne. iin) close (iop)

!-----------------------------------------------------------------------
! match labels with definitions and add any new operators
!-----------------------------------------------------------------------
      call matchlab(loperr)
      call findlabels              ! for FAST and natpotcut and cpd

!-----------------------------------------------------------------------
! fill in blanks in Hamiltonian with unit operators
!-----------------------------------------------------------------------
      do k=1,min(ktot,maxkoe)
         do f=1,nopdof
            if (hamilton(f,k) .eq. 0) then
               hamilton(f,k)=f
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! After any call of 'warnmsg' (which is also called by 'wrline') the
! logical 'lwarn' is set to true.
!-----------------------------------------------------------------------
      loperr=lwarn.or.loperr
      lwarn=.false.

      return
      end subroutine heingabe

!#######################################################################

      subroutine genopdof

      implicit none

      integer(long)       :: f,n1,m

!-----------------------------------------------------------------------
! add all dynamical modes and DOFS to operator modes and DOFs
!----------------------------------------------------------------------
      nopdof=ndof
      nopmode=nmode

!-----------------------------------------------------------------------
! set up space for H coefficients
!-----------------------------------------------------------------------
      nopdof=nopdof+1
      fkoe=nopdof
      gdim(fkoe)=1
      modelabel(fkoe)='Koeff'
      mass(fkoe)=0.0_dop

      nopmode=nmode+1
      mkoe=nopmode
      subdim(mkoe)=1
      nspfdof(mkoe)=1
      spfdof(1,mkoe)=fkoe
      dofspf(fkoe)=mkoe
      gwpm(mkoe) = .false.

!-----------------------------------------------------------------------
! add el dof to nopmode if lmult and not lebcont
!-----------------------------------------------------------------------
      if (lmult .and. .not. (gdim(feb) .eq.1)) then
         nopmode = nopmode+1
         m = nopmode
         nspfdof(m)=1
         spfdof(1,m)=feb
         dofspf(feb)=m
         gwpm(m) = .false.
         if (.not. lebcont) meb = m
      endif

!-----------------------------------------------------------------------
! add point dvrs to "particle" for coefficient
!-----------------------------------------------------------------------
      if (lpntdvr) then
         do f=1,ndof
            if (basis(f) .eq. 19 .or. basis(f) .eq. 22) then
               nspfdof(mkoe) = nspfdof(mkoe)+1
               n1 = nspfdof(mkoe)
               spfdof(n1,mkoe) = f
               dofspf(f) = mkoe

               vgdim(f) =1
               ngdim(f) =1
            endif
         enddo
      endif

      return
      end subroutine genopdof

!#######################################################################
!
!            Subroutine      HEINOPDEF
!
! Reads the data defining the Hamiltonian.
! Called from: heingabe
! Calls to: warnmsg, errormsg
!
! 03/97 HDM
!
!#######################################################################

      subroutine heinopdef(loperr,iz)

      implicit none

      integer(long)                  :: ic,ierr,i,iz0
      integer(long), intent(in)      :: iz
      logical(kind=4), intent(inout) :: loperr

      routine='Heinopdef'

      iz0 = 0

!-----------------------------------------------------------------------
! READ KEYWORDS:
!-----------------------------------------------------------------------
   10 continue
      call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      message = inptit
      call cntrl(keyword(1),lc(1),ierr)

      i=1
   15 continue
      if (keyword(i) .eq. 'end-op_define-section') then
         goto 20

      else if( keyword(1) .eq. 'title' .or. &
               keyword(1) .eq. 'titel' ) then
         write(ilog,'(/,a,/)')  '***** Hamiltonian title *****'

  100    continue
         call rdinpf(iop,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
         if (ierr .eq. 1) then
            message = 'No "end-title" found'
            call warnmsg
         endif

         if (keyword(1) .eq. 'end-title' .or. &
             keyword(1) .eq. 'end-titel' ) then
            write(ilog,'(/,a)') '***** End Hamiltonian title *****'
            goto 10
         else
            write(ilog,'(a)') inptit(1:80)
            go to 100
         endif

      else
         message = ' unknown keyword : '
         call wrline(i,iz,iz0)

      endif

!-----------------------------------------------------------------------
! check if more keywords are to be read, else read new line
!-----------------------------------------------------------------------
      if (i.lt.ic) then
         i=i+1
         go to 15
      else
         go to 10
      endif

   20 continue

!-----------------------------------------------------------------------
! After any call of 'warnmsg' (which is also called by 'wrline') the
! logical 'lwarn' is set to true.
!-----------------------------------------------------------------------
      loperr=lwarn.or.loperr
      lwarn=.false.

      return
      end subroutine heinopdef

      end module heingabemod
