!***********************************************************************
!
!                  QUANTICS include file CHANNELS.F90
!                                                                     
!  This include file defines the various read/write channels
!                                                                    
! It seems that as defined as parameters, the "use channels" call has
! to be in each subroutine it is used and not in the module header!
!***********************************************************************
      module channels

      use decimal, only : long, dop

      implicit none
      public

!-----------------------------------------------------------------------
! INPUT / OUTPUT CHANNELS
! Integers for input/output channels defined as parameters to prevent 
! them being changed in code accidentally!!
!
! Do not use channel 13, this is reserved for iout when the standard
!                        output is written to a file!
! Do not use channel  5, reserved for keyboard input!
! 
!
!  1  iin     : .inp file 
!  2  ilog    : log file 
!  3  irst    : restart file
!  4  itim    : timing file
!  5          : reserved for keyboard input
!  6  iout    : output file
!  7  idvr    : dvr file
!  8  ipsi    : psi file
!  9  ichk    : check file
! 10  irho    : gridpop file
! 11  iaut    : auto file
! 12  iaute   : auto file (error measure)
! 13  iout    : iout is changed from 6 to 13 when writing to output file.
! 14  istp    : steps file
! 15  istop   : stop file
! 16  iupd    : update file
! 17  ispeed  : speed file
! 18  irlx    : rlx_info file
! 19  iaut1   : auto1 file (contains c_1 = <psi(0)| H |psi(t)>)
! 20  iaut2   : auto2 file (contains c_2 = <psi(0)|H^2|psi(t)>)
! 21  idat    : daten file
! 22  iop     : operator text file
! 23  ipar    : Hamiltonian parameter file
! 24  itmpin  : scratch file used, .eg. in building of operator 
! 25  ioper   : operator file
! 26  ictr    : ctrace file
! 27  ienerd  : (a)diabatically corrected initial energy distribution
! 28  inpot   : 'natpot' file
! 29  iolog   : op.log file in genoper and ddpeserr file in DD runs
! 30  icroco  : cross file (cross-correlation function)
! 31  iexpect : file of expectation values
! 32  iexpect1: second file for expectation values.
! 33  ieigval : file of eigenvalues and intensities
! 34  ilancz  : file of Lanczos vectors
! 35  ieigvec : file of eigenvectors of tridiagonal Lanczos matrix
! 36  ipden   : pdensity file (one particle density)
! 37  iorb    : orben file, orbital energies
! 38  iveig   : veigen file
! 39  iit     : 'iteration' file
! 40  ipw     : 'prodwei' file
! 41  ivpot   : 'vpot' file in POTFIT and ddpes file in DD runs
! 42  ippsi   : 'ppsi' file (un-contracted natpot) 
! 43  iiwtt   : 'integral wtt' file. Int <Psi(t)|W_gamma|Psi(t)> dt
! 44  iwtt    : <Psi(t)|W_gamma|Psi(t)> data
! 45  iflx    : reactive flux data
! 46  iexim   : implicit psi file data
! 47  irp     : reaction probabilities
! 48  ipsi1   : second psi file
! 49  iaus    : general output file used in analyse programs
! 50  idump   : dump file handle for the quick open-close job, close
!               immediately after use!
! 51  ideb    : debug file -only for testing, should not be used in
!               checked-in code!
!   Files for optimal control
! 52  iefifo  : FIFO for the electric field routine (optcntrl)
! 53  ipfifo  : FIFO for the prop.f (optcntrl)
! 54  ie1fifo : FIFO for old electric field (optcntrl, simprop)
! 55  ip1fifo : FIFO for prop.f (optcntrl, simprop)
! MGPF related:
! 59  iextpot   : the ext-natpot file
! 60  iprintrho : the rho file (unused!)
! 61  ioutbin   : binary outputs for natural potentials on the fine grid
! MCPF related:
! 62  isymlog   : logging of the symmetrization
! 63  istat     : logging monte-carlo statistics of natpot
!   Files for local control
! 64  ilct    : local control details
! 65  icontrol: control field for local control
! 66  ilctvib : local control vibrational populations
!
! 67  itimp
! 68  itimmpi
!   Files for Shepard interp
! 69  igin    : unit to read from
! 70  igout   : unit to use to dump to
! 71  itout   : TOUT
! 72  iactive : activegwp file for dynamic DoF id
! DWF  Dynamical WF (i.e. initorb and dynspf)
! 73  ispwn     : read/write of the spf basis section
! 74  inspf     : number of SPFs
! 75  iopteigen : eigenvalues
! 76  ioptveig  : eigevectors
! ML-MCTDH / MLPF
! 78  ilnp    : lowest natural populations (ML-MCTDH) 
!
! 79  igwptrj : data for GWP trajectory run
!------------
! wkoch:
! 80          : Channel for BEL debug
!------------
! 81  itmpop  : scratch file used in operator generatop. Open over 
! 82  iwpot   : potfit weights file
!               runoper hence unique number.
! MCCPD related
! 83  ipcut     : 1-dim or 1-mode cuts through the potential
! 84  ijfunc1   : (V - Vcpd) mode wise
! 85  ijfunc2   : sum c**2
! 86  ijfunc3   : jfunc1 + epsilon*jfunc2

! 100-199 used for:
! 100+f       : Channels for extern-DVR
! 171,172...  : Channels  for readsrf
! 100  iddin  : First channel no. used for reading / writing DD files. 
!               Also used for Zagreb surface hopping input files.
!
! 381,382,..  : Channels for orthogonalise (iorthowf)
!
! 161  iebas
!
! 700 itempl : First channels for reading template file for DD-vMCG. 
! 800 iinput : First channels for input file for DD-vMCG
!            Exact unit used depends on threads if OMP 
! 900 itmorb : First channels for reading template orbitals for DD-vMCG
!1000 iiorb   : First channels for orbital file for DD-vMCG
!              
!-----------------------------------------------------------------------

      integer(long), parameter :: iin    =  1, &
                                  itim   =  4, &
                                  ichk   =  9, &
                                  irho   = 10, &
                                  iaut   = 11, &
                                  iaute  = 12, &
                                  istop  = 15, &
                                  iupd   = 16, &
                                  irlx   = 18, &
                                  iaut1  = 19, &
                                  iaut2  = 20, &
                                  idat   = 21, &
                                  ipar   = 23, &
                                  itmpin = 24, &
                                  ioper  = 25, &
                                  iddin  = 100, &
                                  itempl = 700, &
                                  iinput = 800, &
                                  itmorb = 900, &
                                  iiorb  =1000
      integer(long), parameter :: ictr   = 26, &
                                  ienerd = 27, &
                                  inpot  = 28, &
                                  icroco = 30, &
                                  iexpect= 31, &
                                  iexpect1=32, &
                                  ieigval= 33, &
                                  ilancz = 34, &
                                  ieigvec= 35, &
                                  ipden  = 36, &
                                  iorb   = 37, &
                                  iveig  = 38, &
                                  iit    = 39, &
                                  ipw    = 40, &
                                  ivpot  = 41, &
                                  ippsi  = 42, &
                                  iiwtt  = 43, &
                                  iwtt   = 44, &
                                  iflx   = 45, &
                                  iexim  = 46, &
                                  irp    = 47, &
                                  ipsi1  = 48, &
                                  idump  = 50
      integer(long), parameter :: ideb   = 51, &
                                  iefifo = 52, &    
                                  ipfifo = 53, &
                                  ie1fifo = 54, & 
                                  ip1fifo = 55, & 
                                  iextpot = 59, & 
                                  iprintrho = 60, &
                                  ioutbin = 61, & 
                                  isymlog = 62, & 
                                  istat   = 63, & 
                                  ilct    = 64, &
                                  icontrol= 65, &
                                  ilctvib = 66, &
                                  itimp   = 67, &
                                  itimmpi = 68
      integer(long), parameter :: igin    = 69, &
                                  igout   = 70, &
                                  itout   = 71, &
                                  iactive = 72, &
                                  ispwn    = 73, &
                                  inspf   = 74, &
                                  iopteigen = 75, &
                                  ioptveig  = 76

      integer(long), parameter :: ilnp     = 78, &
                                  igwptrj  = 79, &
                                  iBEL     = 80, &
                                  itmpop   = 81, &
                                  iwpot    = 82, &
                                  ipcut = 83, &
                                  ijfunc1 = 84, &
                                  ijfunc2 = 85, & 
                                  ijfunc3 = 86, &
                                  iebas    = 161

!----------------------------------------------------------------------- 
! Channels that may be changed during run-time.
!
! E.g. iout, ilog and iolog may all be changed to unit 6 if these files
! are written to the screen. 
!
! Take care with these!!
!----------------------------------------------------------------------- 
      integer(long), save :: ilog   =  2, &
                             irst   =  3, &
                             idvr   =  7, &
                             ipsi   =  8, &
                             iout   = 13, &
                             istp   = 14, &
                             ispeed = 17, &
                             iop    = 22, &
                             iolog  = 29, &
                             iaus   = 49

!-----------------------------------------------------------------------
! filever(i): file version number for file on channel i.
!             Defines output formats etc. 
!  NB: filever must be dimensioned larger than largest channel no. used.
!
!-----------------------------------------------------------------------
      integer(long), parameter            :: maxfver = 1200
      
      real(dop), dimension(maxfver), save :: filever

      end module channels
