!***********************************************************************
!                                                                     
!              MCTDH include file  compdat.f90
!                                                                     
!***********************************************************************
      module compdatmod
     
      use decimal, only : long, dop
      use global,  only : c5

      implicit none

! maxcomp must be .ge. 2
      integer(long), parameter  :: maxcomp = 3

      integer(long), save :: ncomp
!-----------------------------------------------------------------------
! DIRECTORY DEFINITION FOR INPUT/OUTPUT FILES
!
! all directories are relative to the startup directory i.e. where the
!  .inp file is held
!
! name:    name of directory in which the output files are stored
! inputfile : path of the inputfile (~.inp)
! options   : string containing the list of options.
! laenge    : length of string name
! inplaenge : length of string inputfile
! olaenge   : length of string options
!-----------------------------------------------------------------------
      character(len=(c5)), dimension(maxcomp), save :: cname
      integer(long), dimension(maxcomp), save       :: claenge,cdentype

!-----------------------------------------------------------------------
! VARIOUS INTEGERS DEFINING SIZE OF WAVEFUNCTION
!
! dgldim   : Length of array containing wavefunction
!            dgldim = block + Sum[dim(m)*subdim(m)]
! dmatdim  : Length of array containing the density matrices
!            dmatdim = Sum[dim(m)*dim(m)]
! d3matdim : Length of array containing dicht3 matrix
!            d3matdim = Sum[2*dim(m)]
! adim     : Cumulative length of all A vectors.
!            adim = Sum[block(s)]
! agmatdim  : length of width matrix arrays (GWPs)
! gsmatdim  : length od overlap matrix arrays (GWPs)
! mladim   : Length of A vector in ML for all layers except the lowest one
!-----------------------------------------------------------------------
      integer(long), dimension(maxcomp), save :: cdgldim,cdmatdim,cd3matdim, &
                                                 cadim,cagmatdim,cgwpinfodim,&
                                                 cgsmatdim,cmladim

!-----------------------------------------------------------------------
! VARIOUS INTEGERS DEFINING SYSTEM
!
! cnmode  : number of spf degrees of freedom
! cndof   : total number of degrees of freedom
! cnstate : number of separate states in calculation
! cnpacket: number of wave functions to be propagated simultaneously
! cnpackts: number of separate (single-set) packets in calculation
! cfeb    : number of degree of freedom described by an electronic basis
! cmeb    : mode including an electronic basis (feb)
! cfpb    : number of degree of freedom described by a packet basis
! cmpb    : mode including a packet basis (fpb)
!
! cmsymmtr: symmetries: (a)sym1d, (a)sym2d sym3d
! cidmode : identical mode (id keyword)
! csymcv  : which DOFs of the A-vector are to be (a)symmetrized
! cbasis  : DVR type for the DOFs
! cipbaspar : DVR basis integer parameters
!-----------------------------------------------------------------------
      integer(long), dimension(maxcomp), save                :: cnmode,cnstate,&
                                                                cndof,cfeb, &
                                                                cmeb,cnpacket, &
                                                                cnpackts,cfpb,cmpb

      integer(long), dimension(:,:), allocatable, save       :: cidmode,csymcv,cbasis
      integer(long), dimension(:,:), allocatable, save       :: cmsymmtr
      integer(long), dimension(:,:,:), allocatable, save     :: cipbaspar

!***********************************************************************
!     SOME VARIABLES SPECIFYING THE FIRST CALCULATION:
!
!     The below listed variables correspond to the ones defined
!     in the files 'psidef.inc'. To each variable name a
!     'c' is added to its beginning to indicate that this variable
!     contains the comparison data sets. The (additional) last
!     array variable switches between the data of the 'first' and
!     'second' calculation.
!***********************************************************************
      integer(long), dimension(:,:), allocatable,save   :: cblock,czpsi
      integer(long),dimension(:,:,:), allocatable, save :: cdim,cvdim, &
                                                           cndim,czetf,&
                                                           czpg,czxi,  &
                                                           czphi,cdmat,&
                                                           cd3mat,czagmat,&
                                                           czxivec,czphase,&
                                                           cgsmat,czdgwp, &
                                                           cgwpdim

      integer(long),dimension(:,:,:), allocatable, save :: cdblock,cdzpsi
      integer(long),dimension(:,:,:,:), allocatable, save :: cddim, cvddim, &
                                                             cnddim,cdzetf, &
                                                             cddmat,cdd3mat

      integer(long), dimension(maxcomp), save      :: cpsitype,ccitype,cgwptype,&
                                                      cgwpphase
      real(dop), dimension(:,:), allocatable, save ::  cciweight
      real(dop), dimension(maxcomp), save          ::  ccicut
      logical(kind=4), dimension(maxcomp), save  :: clgwpphase
      integer(long), dimension(maxcomp), save  :: cdlstate,cdrstate

      logical(kind=4), dimension(:), allocatable, save   :: cldynspf
      integer(long), dimension(:,:,:), allocatable, save :: cmaxdimspwn

!-----------------------------------------------------------------------
! PARAMETERS DEFINING PRIMITIVE BASIS
!
! gdim(f)  : number of grid points for dof f
! vgdim(f) : number of grid points for dofs preceeding f
! ngdim(f) : number of grid points for dofs following f
! subdim(m): number of grid points for mode m
! zort(f)   : pointers to grid
! ortdim    : length of grid array
!-----------------------------------------------------------------------
      integer(long), dimension(maxcomp), save               :: cortdim
      integer(long), dimension(:,:), allocatable, save      :: csubdim,cvgdim, &
                                                               cgdim,cngdim,czort, &
                                                               cnspfdof,cdofspf
      integer(long), dimension(:,:,:), allocatable, save    :: cspfdof

!-----------------------------------------------------------------------
! PARAMETERS DEFINING OPERATORS
!
!  lreal  : Hamiltonian and WF are real (improved relaxation).
!  lcap   : calculation employing CAPs (or complex potentials)
!-----------------------------------------------------------------------
      logical(kind=4), dimension(maxcomp), save :: clcap,clreal,clrreal

!-----------------------------------------------------------------------
! MULTILAYER PARAMETERS
!
! tape : integer array defining the structure of a multilayer wavefunction
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save :: ctape,ctapeM

contains

!#######################################################################

      subroutine alloc_compdat

use maxv
use channels
use dvrdatmod, only: mbaspar

      implicit none

      integer(long)     :: ierr
      character(len=12) :: cerr
      integer(long), external   :: integersize, realsize, logicalsize

! check all arrays are not allocated
      call dealloc_compdat
!
! system arrays
!
      cerr = 'cbasis'
      allocate (cbasis(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cipbaspar'
      allocate (cipbaspar(mbaspar,maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'csubdim'
      allocate (csubdim(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cgdim'
      allocate (cgdim(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cvgdim'
      allocate (cvgdim(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cngdim'
      allocate (cngdim(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cnspfdof'
      allocate (cnspfdof(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cspfdof'
      allocate (cspfdof(maxdim,maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdofspf'
      allocate (cdofspf(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cmsymmtr'
      allocate (cmsymmtr(2*maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + 2*maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cidmode'
      allocate (cidmode(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'csymcv'
      allocate (csymcv(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czort'
      allocate (czort(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
!
! psi arrays
!
      cerr = 'cldynsf'
      allocate (cldynspf(maxcomp),stat=ierr)
      allocmemory=allocmemory + maxcomp*logicalsize()
      if (ierr > 0) go to 1000
      cerr = 'cddim'
      allocate (cddim(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cvddim'
      allocate (cvddim(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cnddim'
      allocate (cnddim(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdblock'
      allocate (cdblock(maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdim'
      allocate (cdim(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cmaxdimspwn'
      allocate (cmaxdimspwn(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cgwpdim'
      allocate (cgwpdim(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cvdim'
      allocate (cvdim(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cndim'
      allocate (cndim(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cblock'
      allocate (cblock(maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czpsi'
      allocate (czpsi(maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdzpsi'
      allocate (cdzpsi(maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdztef'
      allocate (cdzetf(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cddmat'
      allocate (cddmat(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdd3mat'
      allocate (cdd3mat(maxdim,maxsta,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czetf'
      allocate (czetf(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czpg'
      allocate (czpg(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czxi'
      allocate (czxi(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czphi'
      allocate (czphi(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czdgwp'
      allocate (czdgwp(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cdmat'
      allocate (cdmat(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cd3mat'
      allocate (cd3mat(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cxagmat'
      allocate (czagmat(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czxivec'
      allocate (czxivec(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'czphase'
      allocate (czphase(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cgsmat'
      allocate (cgsmat(maxdim,maxsta,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'cciweight'
      allocate (cciweight(maxdim,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxdim*maxcomp*integersize()
      if (ierr > 0) go to 1000
      cerr = 'ctape'
      allocate (ctape(maxtape,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000
      cerr = 'ctapeM'
      allocate (ctapeM(maxtape,maxcomp),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_COMPDAT :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_compdat

!#######################################################################

      subroutine dealloc_compdat

!
! system arrays
!
      if (allocated(cbasis)) deallocate(cbasis)
      if (allocated(cipbaspar)) deallocate(cipbaspar)
      if (allocated(csubdim)) deallocate(csubdim)
      if (allocated(cgdim)) deallocate(cgdim)
      if (allocated(cvgdim)) deallocate(cvgdim)
      if (allocated(cngdim)) deallocate(cngdim)
      if (allocated(cnspfdof)) deallocate(cnspfdof)
      if (allocated(cspfdof)) deallocate(cspfdof)
      if (allocated(cdofspf)) deallocate(cdofspf)
      if (allocated(cmsymmtr)) deallocate(cmsymmtr)
      if (allocated(cidmode)) deallocate(cidmode)
      if (allocated(csymcv)) deallocate(csymcv)
      if (allocated(czort)) deallocate(czort)
!
! psi arrays
!
      if (allocated(cldynspf)) deallocate(cldynspf)
      if (allocated(cddim)) deallocate(cddim)
      if (allocated(cvddim)) deallocate(cvddim)
      if (allocated(cnddim)) deallocate(cnddim)
      if (allocated(cdblock)) deallocate(cdblock)
      if (allocated(cdim)) deallocate(cdim)
      if (allocated(cmaxdimspwn)) deallocate(cmaxdimspwn)
      if (allocated(cgwpdim)) deallocate(cgwpdim)
      if (allocated(cvdim)) deallocate(cvdim)
      if (allocated(cndim)) deallocate(cndim)
      if (allocated(cblock)) deallocate(cblock)
      if (allocated(czpsi)) deallocate(czpsi)
      if (allocated(cdzpsi)) deallocate(cdzpsi)
      if (allocated(cdzetf)) deallocate(cdzetf)
      if (allocated(cddmat)) deallocate(cddmat)
      if (allocated(cdd3mat)) deallocate(cdd3mat)
      if (allocated(czetf)) deallocate(czetf)
      if (allocated(czpg)) deallocate(czpg)
      if (allocated(czxi)) deallocate(czxi)
      if (allocated(czphi)) deallocate(czphi)
      if (allocated(czdgwp)) deallocate(czdgwp)
      if (allocated(cdmat)) deallocate(cdmat)
      if (allocated(cd3mat)) deallocate(cd3mat)
      if (allocated(czagmat)) deallocate(czagmat)
      if (allocated(czxivec)) deallocate(czxivec)
      if (allocated(czphase)) deallocate(czphase)
      if (allocated(cgsmat)) deallocate(cgsmat)
      if (allocated(cciweight)) deallocate(cciweight)
      if (allocated(ctape)) deallocate(ctape)
      if (allocated(ctapeM)) deallocate(ctapeM)

      return
      end subroutine dealloc_compdat

      end module compdatmod

