!#######################################################################
!      datenmod.f90
!
! Module containing options for a calculation that refer to the treatment
! of data in various programs
!
!#######################################################################

      module datenmod
      
      use decimal, only : long,dop
      use global

!-----------------------------------------------------------------------
! LOGICALS TO CONTROL INPUT / OUTPUT OPTIONS
!
!     lpsi     : psi is saved
!     lctr     : correlated trace is written to ctrace file
!     lauto    : autocorrelation data written to file auto
!     lautoe   : autocorrelation without highest natural orbitals
!     lautoout1: first order autocorrelation function <psi(0)|H|psi(t)>
!                is computed
!     lautoout2: 2nd order autocorrelation function <psi(0)|H^2|psi(t)>
!                is computed
!     lintstep : integration step data written to file intstep
!     loutfil  : output to file rather than screen
!     lupdate  : update times are written to file update (only if
!                lconstvar is true)
!     lgpop    : gridpop file is created
!     lpdens   : pdensity file is created (one particle density)
!     lspeed   : controls output to speed file
!     lstopf   : controls whether the stop file is opend
!     llanczvec: the Lanczos vectors are saved in a diagonalisation run
!     leigvec  : the eigenvectors of the tridiagonal Lanczos matrix are
!                saved in a diagonalisation run
!     lcross   : cross-correlation data written to file CROSS
!     lorben   : orbital energies are saved to file orben
!     lquadpop : derives the gridpops, spops,  with quadratic rho
!     lignore  : Ignore "different bases" when reading restart-file.
!     lsheptraj: When running gwp on a Shepard surface, dump trajectories
!     lshepgp  : Include gpoint in sheptraj
!     lshepinit: A hack to allow initialisation of a Shepard surface for restart
!     lactivegp: Identity of active gwp DoFs written to activegwp
!     lveigen  : the eigenvalues and vectors of a 1d-operator are
!                written to the file veigen_modelabel.
!    lsplitrst : Split the restart file of a multi-packet run into individual
!                restart files called rst000, rst001, ...
!     lgraphviz: generate graphviz input file to visualize ML tree
!     lgentrj  : generate trajectory files for gwptrj dynamics
!-----------------------------------------------------------------------
      logical(kind=4), save :: lpsi,lctr,lauto,lintstep,loutfil,lupdate, &
                               lgpop,lpdens,lautoord1,lautoord2,lcross, &
                               lignore,lspeed,lstopf,llanczvec,leigvec, &
                               lorben,lquadpop,ltinitcmd,lveigen,lsplitrst,&
                               lautoe=.false.
      logical(kind=4), save :: lsheptraj=.false.,lshepinit=.false., &
                               lactivegp=.false.,lshepgp=.true.
      logical(kind=4), save :: lgraphviz=.false.
      logical(kind=4), save :: lgentrj=.false.

!-----------------------------------------------------------------------
! TIME VARIABLES
!
!  out1:       Frequency of data output in fs
!  out2:       Frequency of wavefunction output in fs
!  tfinal:     Final time in fs
!  tinit:      Time at start of calculation in fs
!  nout2:      Number of times wavefunction is written to "psi" file
!              (this is needed for array dimensioning in some
!               auswert routines)
!-----------------------------------------------------------------------
      real(dop), save :: out1 = 0.0_dop, &
                         out2 = 0.0_dop, &
                         tfinal = -9.9e99_dop, &
                         tinit = 0.0_dop

!-----------------------------------------------------------------------
! LOGICALS TO CONTROL INPUT / OUTPUT OPTIONS
!
!     lautodoub: autocorrelation function is written twice during period
!                "out1" (only possible for CMF)
!     lpsiopt  : options as to how psi is to be saved
!     lgpel    : if true, gridpop is written separately for each state 
!                (default).
!     lusecpp  : if true CPP has been used on input - delete temp input
!                on close - in closedat.F
!-----------------------------------------------------------------------
      integer(long), parameter                  :: npsiopt = 6
      logical(kind=4), save                     :: lautodoub,lusecpp, &
                                                   louteverystep
      logical(kind=4), save                     :: lgpel = .true.
      logical(kind=4), dimension(npsiopt), save :: lpsiopt

!-----------------------------------------------------------------------
! SOME MORE PARAMETERS NEEDED FOR EXPECTATION VALUES
!
! nexpect     : number of operators to derive the expectation-value
! lreonly     : logical, only real parts will be output to expectation
! nexpect1    : number of operators for file expect1
! lreonly1    : logical, only real parts will be output to expect1
!
! following expectation operator label arrays are allocated if needed 
! in einrun:
! expectlab(nexpect)  : labels for expect file
! expectlab1(nexpect1) : labels for expect1 file
!-----------------------------------------------------------------------
      integer(long), save                          :: nexpect, nexpect1
      logical(kind=4), save                        :: lreonly, lreonly1
      character(len=c2), dimension(:), allocatable, save   :: expectlab
      character(len=c2), dimension(:), allocatable, save   :: expectlab1

!-----------------------------------------------------------------------
! COUNTERS FOR DATA FILES
!
! countpsi : number of output of psi in psi file
!-----------------------------------------------------------------------
      integer(long), save  :: countpsi, countchk, countrho, countpden

      end module datenmod
