!   -*- fortran -*-
!***********************************************************************
!
!                         MCTDH include file DGLOBAL
!
!  MCTDH include file defining various parameters and variables
!  global to the dynamical SPFs code.
!
!***********************************************************************
      module dglobal

      use decimal, only: long,dop
      use global, only: c5
      use maxv, only: maxdim,maxsta,maxtape,allocmemory
      use channels, only: ilog

! Max. length of Psi vector
      integer(long), save :: dglmax

!-----------------------------------------------------------------------
! Thresholds and criteria 
!
! ibeta   : select threshold criterion for selecting optmial SPFs
! betathr0 : threshold for selecting SPFs in initorbs
! betathr : threshold for selecting SPFs
! betavar : 
!
! Thresholds and criteria for spawning
! alphamax:
! alphathr
! alpha0
! falpha
! idynalpha : = 1 dynamical value of alpha
!
!-----------------------------------------------------------------------
      integer(long), save :: ibeta = 1
      real(dop), save     :: betathr0 = 1.0d-9
      real(dop), save     :: betathr = 1.0d-7
      real(dop), save     :: betavar = 0.0d-00

      integer(long), save :: itermax = 0, &
                             imaxspf = 1, &
                             idyncyc = 0, & 
                             iresize = -1, &
                             ioptavec = 5, &
                             ioptavec2nd = 2

      integer(long), save :: idynalpha = 0
      real(dop), save :: alphamax = 0.001 ,&
                         alphathr = 0.0_dop, &
                         alpha0 = 0.0_dop, &
                         alphathr0 = 0.0_dop, &
                         falpha = 0.0_dop
      real(dop), save :: tauthr = 3.0d-03

      logical(kind=4), save :: ldebugspawn = .false.

!-----------------------------------------------------------------------
!     linitorb   : Calculate optimal orbitals at t0 
!-----------------------------------------------------------------------
      logical(kind=4), save :: linitorb = .false.

!-----------------------------------------------------------------------
! Logicals to control choice of unoccupied SPFs    
!     loptspf       : use optimal orbitals
!        if loptspf=.true. following logicals also used
!     loptspflancz  : use Lanczos to get optimal orbitals
!     lqoptspf      : use quasi-optimal orbitals as inital guess
!
! Logicals to control choice of A-coefficients
!     loptavec      : optimise A-vector
!     loptavec2nd   : optimise A-vector using 2nd order algorithm
!     loptavecnum   : numerically optmise A-vector
!-----------------------------------------------------------------------
      logical(kind=4), save :: loptspf = .false., &
                               loptnspf = .false., &
                               loptspflancz = .false., & 
                               lqoptspf = .false., &
                               loptavec = .false., & 
                               loptavec2nd = .false., &
                               loptavecnum = .false.

!-----------------------------------------------------------------------
!     Dynamical WF (i.e. spawn, initorb and dynspf)
!     iindyninp  : integers parameters associated with keywords
!                  "initorb" and "dynspf"
!     ndynspf0   : default no. of initial SPFs for DynWF. 
!     rspawninp  : real parameters associated with keyword "spawn"
!     rindyninp  : real parameters associated with keywords
!                  "initorb" and "dynspf"
!     lindyninp  : logical flags associated with keywords
!                  "initorb" and "dynspf"
!     lnspf      : Output the number of SPFs
!     lnewspf    : New SPFs added before the current integration step
!     lspwnfcall : Initialization integrator after new SPFs
!     tcurrent   : current time step
!     constop0   : initial value when variable convergence threshold
!                  (see constop)
!     ioptnspf    : controls way of selecting orbitals to add 
!             = 0 : Add a fixed number (e.g. in initorb w/o/ dynspf)
!
!     tapeM      : Tape for ML-MCTDH with maximum no. SPFs
!     itape2     : position of SPFs in tape - used to update SPF numbers.
!     itape3     : new no. SPFs during spawning
!    
!-----------------------------------------------------------------------
      integer(long), save   :: ndynspf0 = 3
      logical(kind=4), save :: lnspf,lnewspf,lspwnfcall

      integer(long), dimension(:), allocatable, save  :: tapeM,itape2,&
                                                         itape3
      real(dop), save       :: tcurrent,constop0,cp_constop

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! maxdimspwn : maximum dim(m,s) in dynamical WF
! dimold     : storage of dim(m,s) in dynamical WF (MCTDH)
! dimtape    : storage of dim(m,s) in dynamical WF (ML-MCTDH)
! dimocc     : number of currently occupied orbitals (dynamical WF)
! dimunocc   : total number of orbitals as defined by initorb (dynamical WF)
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save :: maxdimspwn,dimold,&
                                                          dimtape, &
                                                          dimocc,dimunocc

!-----------------------------------------------------------------------
!     Number projection on each SPF
!     lnumberproj  : The Number projection on each SPF to keep the
!                    expectation value of the number operator of each
!                    SPF constant
!-----------------------------------------------------------------------
      logical(kind=4), save :: lnumberproj,lrdnumproj
      character(len=c5), save  :: numprojfile
      integer(long), save   :: npflaenge

contains

      subroutine alloc_dynwf

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr
      integer(long), external :: integersize


! check all arrays are not allocated
      call dealloc_dynwf

      cerr = 'maxdimspwn'
      allocate(maxdimspwn(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dimtape'
      allocate(dimtape(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dimold'
      allocate(dimold(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dimocc'
      allocate(dimocc(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dimunocc'
      allocate(dimunocc(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'tapeM'
      allocate(tapeM(maxtape),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000
      cerr = 'itape2'
      allocate(itape2(maxtape),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000
      cerr = 'itape3'
      allocate(itape3(maxtape),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DYNWF :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      write(6,'(a)') 'Error allocating memory in ALLOC_DYNWF :',cerr
      write(6,'(a,i5)') 'Error Code :',ierr

      end subroutine alloc_dynwf

!#######################################################################

      subroutine dealloc_dynwf

      implicit none

      if (allocated(maxdimspwn)) deallocate(maxdimspwn)
      if (allocated(dimold)) deallocate(dimold)
      if (allocated(dimtape)) deallocate(dimtape)
      if (allocated(dimocc)) deallocate(dimocc)
      if (allocated(dimunocc)) deallocate(dimunocc)
      if (allocated(tapeM)) deallocate(tapeM)
      if (allocated(itape2)) deallocate(itape2)
      if (allocated(itape3)) deallocate(itape3)

      return
      end subroutine dealloc_dynwf
      
      end module dglobal
