
!                                                                     
!              Quantics include file   DIRDYN
!                                                                     
!  file defining and allocating arrays common to several parts of the program.  
!  Declarations independent of modes and states should be placed
!  in include file "global".     
!                                                                     
!***********************************************************************

      module dirdyn

      use decimal, only : long, dop, sip
      use global, only : c1, c2, c3, c4,c5
      use maxv
      use channels, only: ilog
      use sqlite_types, only: SQLITE_DATABASE

      implicit none
    
!-----------------------------------------------------------------------
! flags for direct dynamics options
!
! ddupdate : controls updating of DD surface
!          = 1  : time control
!          = 2  : error control
!          = 0  : never (keep first guess)
! dduperr  : error criteria for DD PES update
! dduptim  : time criteria for DD PES update
!
! dduptim0 : last time at which DB was updated
!
! ddlog : write extra information to log file.
! ddlogfull : write even more information to log file.
!
!-----------------------------------------------------------------------
      integer(long), save   :: ddupdate = 0
      real(dop), save       :: dduperr = 1.0d-6, &
                               dduptim = 0.5*41.3414, &
                               dduptim0 = -99999.0_dop
      logical(kind=4), save :: ddlog = .false., &
                               ddlogfull = .false.

!-----------------------------------------------------------------------
! flag to pass trajectory number in TSH calc
!-----------------------------------------------------------------------
      integer(long), save :: ddtrajnum = 1

!-----------------------------------------------------------------------
! Global to hold "the" SQL database.
! This represents the database in use by the program.
! NB. Routines should copy this value into a local as soon as possible.
! Utility routines should be passed the DB they're working on as an
! argument.  This allows, amongst other things, multiple DBs.
!
! sqldbfile : name of SQLite database
! lsqldb  : if true sqldbfile exists
!-----------------------------------------------------------------------
      type(SQLITE_DATABASE), save :: the_db
      integer(long), save         :: sqlvers = 4

      character(len=c5)     :: sqldbfile
      logical(kind=4), save :: lsqldb  = .false.

!-----------------------------------------------------------------------
! Following flags are true if DB contains the information
! lgeodb   : Geometries (Cartesian)
! lpesdb   : PES (diabatic)
! lgradb   : Gradients (diabatic)
! lhesdb   : Hessians (diabatic)
! lapesdb  : PES (adiabatic)
! lagradb  : Gradients (adiabatic)
! lahesdb  : Hessians (adiabatic)
! ldipdb   : Dipoles (diabatic)
! ladipdb  : Dipoles (adiabatic)
! lmocodb  : MO coefficients (for CAS) 
! ltransdb : ADT transformation matrix
! lnactdb  : NACTs
! lnmdb    : Geometries (Normal mode)
! lcasocdb : SOCs 
! lcsocdb  : diabatic SOCs (used for GB DD)
! lcivecdb : CI vectors
! lsymdb   : Symmetry relationships between geometries
! lcpesdb  : Complex PES (diabatic - imaginary part)
! lctransdb : Complex ADT transformation matrix
! ltemplate : template.dat file
! lqcfreq  : QC frequency file used in coo <-> nm transformation
!-----------------------------------------------------------------------
      logical(kind=4), save :: lgeodb = .false.,&
                               lpesdb = .false.,&
                               lgradb = .false.,&
                               lhesdb = .false.,&
                               lapesdb = .false.,&
                               lagradb = .false.,&
                               lahesdb = .false.,&
                               ldipdb = .false.,&
                               ladipdb = .false.,&
                               lmocodb = .false.,&
                               ltransdb = .false.,&
                               lnactdb = .false.,&
                               lnmdb = .false.,&
                               lcasocdb = .false.,&
                               lcsocdb = .false.,&
                               lcivecdb = .false.,&
                               lsymdb = .false.,&
                               lcpesdb = .false.,&
                               lctransdb = .false.,&
                               ltemplate = .false.,&
                               lqcfreq = .false.

!-----------------------------------------------------------------------
! ldiabdb  : DD uses diabatisation
!-----------------------------------------------------------------------
      logical(kind=4), save :: ldiabdb = .false.

!-----------------------------------------------------------------------
! FLAGS AND PARAMETERS NEEDED FOR CALCULATION OF DIRECT DYNAMICS PES
!
! New version (Benjamin, December 2008)
!
!-----------------------------------------------------------------------
! ndofdd    = no. of MCTDH degrees of freedom for DD.
! ndofddpes = no. of degrees of freedom for DD PES.
! natm = no. of atoms (ndofddpes/3)
! nbasis = total no. of basis functions. Used for MO coeff. storage in CAS
! nelectron = no. of electrons in molecule.
! nocc = no. of occupied orbitals (closed + active in CAS).
! nclosed = no. of closed orbitals in a CAS calculation.
! ncas = no. of MOs in the active space (=nocc-nclosed).
! ncasel = no. of electrons in each configuration (alpha/beta).
! nconf = no. of configurations in the CAS expansion for each multiplicity for each
!         spin.
! nciconf = no. of configurations in the CAS expansion for each multiplicity.
! dercpdim = number of unique non-adiabatic coupling terms (nst*(nst-1)/2).
! lddsoc = Different multiplicities requested so using SOCs.
! if lddtrans = .true., direct dynamics uses transformed coordinates for
!               wavepacket
! if ltshtrans = .true., TSH uses transformed coordinates for trajectories
! dddiab  : Diabatisation scheme (see defqchem (funcqchem.F))
! dde0    : potential energy origin in hartree
! lddwrdb : if DD PES will be written to database
! lddrddb : if DD PES will be read from database
! ddmindb : distance criterion to use database point. If no point closer
!           than ddmindb new data will be calculated.
! dddvmin : energy criterion to use database point for regularisation
! dddbvar : value of variance when determining whether to calc. new point
! dddbhar : value of harmonic deviation for determining when to calc. new point
! lddreord = .true. if set atoms are reordered according to mass families
! ldbsave : whether to keep the DB in memory or not
! ldblocal : take only nearest point in DB
! lusdhes : use approximate updated Hessian or full
! hess0   : Select initial Hessians
!       = 0 : calculate (default)
!       = 1 : take ground-state Hessian for all states.
! lrediab : if wishing to regenerate the diabatic surfaces from the adiabatic
! lrediabnomap : if wishing to regenerate the diabatic surfaces from the adiabatic
!                without re-ordering the points.
! lrediabswap : if wishing to swap DCP vectors to maximise overlap
! ldb2gapot : if wishing to regenerate the diabatic surfaces from the adiabatic
! ddnamedb1 : directory storing a second DB (e.g. to convert to a GAP DB)
! ddlaengedb1 : length of string ddnamedb1
! lshepgrad : use the gradient of the Shepard interpolation
! lgausspot : use the GAP interpolation for the potential.
! lpestest  : use an analytic test potential.
! idpestest : index of analytic test potential.
! ldbvar  : if wishing to use the variance method instead of dbmin
! ldbhar  : if wishing to use the harmonic measure instead of dbmin
! dbinterord : order used in Shepard interpolation of DB points (2p).
! dbinterord1 : first order used in Shepard interpolation of DB points (2q).
! dbnconf  : No. of DB structures inside confidence radius
! dbminwgt : minimum weight of points included in Shepard interpolation
! dbdisp   : if point < dbdisp from a DB point, no interpolation done.
! dbdegen  : parameter defining an energy gap, below which states are "degenerate"
! lnodegen0 : ignore degenaracy of states at Q0
! lnodegen  : if true do take data of degenerate states from QC
!             if false, replace with predicted values
! lchkdcp   : if true, check dcp between neighbouring states for overlap
! mxsinterest : The highest state of interest. Above this degeneracies not checked.
!               Useful if many states included, but only the lowest mxsinterest
!               are needed for the dynamics. 
! nsmult   : no. of different spin multiplicities.
! idxmult  : lists the multiplicities present
! imultblk : no. of states in each spin block (ignoring degeneracy).
! imultoff : offsets to each spin block (ignoring degenerate states).
! stoff    : offsets to spin blocks (accounting for degenerate states)
! stblk    : no. of states in each spin block (accounting for degenerate states)
! ddstoff  : offsets to spin blocks in DD
! ddstblk  : no. of states in each spin block in DD
! nsocstate : No. of split states 
! dercpblk : no. of derivative couplings in each spin block.
! dercpoff : offsets to derivative couplings in each spin block.
! lsocdegen : True if treating degenerate states as single states (multiplicity > 1).
! symdof   : List of non-totally symmetric DOFs.
! cicocut  : CI-coeff. cutoff for ignoring terms when overlapping CAS WFs.
! ldecoup  : Whether to decouple core and AS orbitals when overlapping CAS WFs.
! lrefdbrd : read the diabatisation (DB) reference point from input
!            diabref-section rather than use FC point.
! diabint : Integration scheme for global diabatisation
!           0 = original pseudo-Shepard
!           1 = linear interpolation
!           2 = Shepard
!
!-----------------------------------------------------------------------
      integer(long), save                    :: ndofdd = 0, &
                                                ndofddpes = 0, &
                                                natm = 0, &
                                                ddtransdim = 0
      integer(long), save                    :: dddiab = -999
      integer(long), save                    :: dercpdim = -999
      integer(long), save                    :: nbasis = 0, &
                                                nelectron  = 1,&
                                                nocc = 1, &
                                                nclosed = 1, &
                                                ncas = 1
      integer(long), dimension(:,:), allocatable, save :: nconf,ncasel
      integer(long), dimension(:), allocatable, save   :: nciconf
      integer(long), save                    :: dbinterord = 24, &
                                                dbinterord1 = 4 
      integer(long), save                    :: diabint = 1
      integer(long), save                    :: nactdim = 1
      integer(long), save                    :: idpestest = 0
      integer(long), dimension(:), allocatable, save :: iatsym
      integer(long), dimension(:,:), allocatable, save :: gwpshell,fgwpshell
      real(dop), save                        :: ddmindb = 0.2_dop, &
                                                dde0 = 0.0_dop, &
                                                dddvmin = 0.0_dop, &
                                                dddbvar = 0.0001_dop, &
                                                dddbhar = 0.01_dop
      integer(long), save                    :: dbnconf = 6
      integer(long), save                    :: mxsinterest = 99999
      real(dop), save                        :: dbminwgt = 0.0001_dop, &
                                                dbdisp = 1.0e-6_dop, &
                                                dbdegen = 0.05_dop/27.2114_dop, &
                                                dbmaxe = 999.0_dop
      real(dop), save                        :: dbadreg = 999.00/27.2114
      logical(kind=4), save                  :: lnodegen0=.false., &
                                                lnodegen=.false., &
                                                lchkdcp=.false.
      logical(kind=4), save                  :: lddtrans=.false.,&
                                                lddwrdb=.true.,&
                                                lddrddb=.true.,&
                                                lddreord=.false.,&
                                                ltshtrans=.false.
      logical(kind=4), save                  :: lnbasis
      logical(kind=4), save                  :: ldbsave = .true., &
                                                lupdhes = .true.,&
                                                lshepgrad = .false.,&
                                                lgausspot = .false.,&
                                                ldbvar = .false., &
                                                ldbhar = .false., &
                                                lpestest = .false., &
                                                ldblocal = .false.
      logical(kind=4), save                  :: lrediab = .false.,&
                                                lrediabmap = .true.,&
                                                lrediabswap = .false.
      logical(kind=4), save                  :: ldb2gapot = .false.
      character(len=c5), save :: ddnamedb1=' '
      integer(long), save     :: ddlaengedb1
      integer(long), save     :: hess0 = 0

      logical(kind=4), save                  :: lrefdbrd = .false.
      integer(long), save                    :: nsmult = 1
      integer(long), save                    :: nsocstate 
      integer(long), dimension(:), allocatable, save :: imultblk,imultoff,&
                                                stoff,stblk,ddstoff,ddstblk,&
                                                dercpblk,dercpoff,idxmult
      logical(kind=4), save                  :: lddsoc=.false.,&
                                                lsocdegen = .true.
      integer(long), dimension(:), allocatable, save :: symdof
      real(dop), save                        :: cicocut = 1.0e-8_dop
      logical(kind=4), save                  :: ldecoup=.false.

!-----------------------------------------------------------------------
! ntr      : no. of trivial modes. default 6.
!-----------------------------------------------------------------------
      integer(long), save   :: ntr = 6

!-----------------------------------------------------------------------
! Parameters and Options for TSH calculations
! ndoftsh = no. of degrees of freedom in TSH (nspfdof(1) rounded up to
!           be divisible by 3 as TSH code uses atomic Cartesian format)
! natmtsh = no. of atoms (ndoftsh/3)
!-----------------------------------------------------------------------
      integer(long), save                    :: ndoftsh = 0, &
                                                natmtsh = 0

!-----------------------------------------------------------------------
! options to control quantum chemistry calculation
! lsacas  : use SACAS in place of CAS 
! lorbread: for CAS, read MOs from previous points
! lorbread0: for CAS, read MOs for first points
! lfchk   : use fchk in place of log as GAUSSIAN output file
! fchkcmd : external command to run formchk (GAUSSIAN utility program)
! qcmethod: method to be used HF, CAS
! lsolvent: use a continuum solvent method.      
! solmethod: continuum solvent model be used e.g. PCM
! qcprog  : quantum chemistry program to be used, e.g. GAUSSIAN
! ddextcmd: external command to run quantum chemistry program
! qcsubform: No. of arguments to be taken by submission command
!          = 1  $1 = file name (full path)
!          = 2  $1 = file name   $2 = directory where $1 is located
!          = 3  $1 = file name   $2 = directory where $1 is located,
!               $3 = no of processors to run job
! qcsimulate : if true, QC program not run.
! qcnproc : no of processors to run QC job
! qcscr   : name of the scratch (sub)-folders for Q-Chem jobs
! nroot   : no. of roots in a CAS calculation
! lqcsaveall : if true output files concatenated together
! lqcsaveout : if true output files calculated at tout concatenated 
!              together
! nddmult(sp) : no. of states of multiplicity sp (sp=1,2,3,...)
! istatemap(n) : mapping of states. Nth state in MCTDH is istatemap(n) 
!                in QC, taking SOC into account if not socdegen 
! imultmap(n) : mapping of multiplicity: multiplicity of nth state in MCTDH
! imultqc(n)  : mapping of multiplicity: multiplicity of nth state in MCTDH 
!               ignoring spin-orbit splittings (degenerate states)
!-----------------------------------------------------------------------
      character(len=c3), save               :: ddextcmd,fchkcmd
      character(len=c2), save               :: qcmethod,qcprog,qcscr,&
                                               solmethod, qcbasis="UNKNOWN"
      integer(long), save                   :: qcsubform,qcnproc
      integer(long), save                   :: nroot = -999
      logical(kind=4), save                 :: lsacas = .false.
      logical(kind=4), save                 :: lfchk = .false.
      logical(kind=4), save                 :: lorbread = .true.
      logical(kind=4), save                 :: lorbread0 = .true.
      logical(kind=4), save                 :: lqcsimulate = .false.
      logical(kind=4), save                 :: lqcsaveall = .false.
      logical(kind=4), save                 :: lqcsaveout = .false.
      logical(kind=4), save                 :: lsolvent = .false.

      integer(long), dimension(7), save              :: nddmult=0
      integer(long), dimension(:), allocatable, save :: istatemap, &
                                                        imultmap, &
                                                        imultqc

!-----------------------------------------------------------------------
! Trafo info
! Qcoo refer to coordinates used in Quantum Dynamics. Cartesians refers 
! to coordinates used in QChem calculation, or TSH (where they may not 
! be Cartesians)
!
! DB info
! ddtrfile: file containing DD coordinate trafo info
! trfiletyp: type of calculation in ddtrfile, e.g. RHF or CAS.
!            see eindirdyn.F
! lrde0: if true, E0 is read from trafo file
!
! ddtransf: transforms from Qcoo to Cartesians x = ddtransf*Q
!           (= nmcoo in VCHAM)
! ddtransb: transforms from Cartesians to Qcoo Q = ddtransb*(x - x0)
!           (= coonm in VCHAM)
! ddxcoo0 : x0 required in in backward transformation
!
! dreffile: File containing reference point for regularised or local
!           diabatisation
! coinfile: File containing conical intersection data for regularised 
!           diabatisation or initial BS-like vectors for local diabatisation
!
! xingeo : Cartesian geometry specified in initial-geometry-section
! xatnam : atom names to go with xingeo
! xmass  : masses to go with xingeo
! xcoin  : Conical intersection point used by regularised diabatisation
!          or initial point used by local diabatisation. In Cartesians
! xgd0   : Gradient difference used by regularised or local diabatisation.
!          In Cartesians
! xdcp0  : Derivative Coupling used by regularised or local diabatisation.
!          In Cartesians
! xrefpoint : Reference point used by regularised or local diabatisation.
!             In Cartesians
! xref : Reference point used to define highest symmetry geometry in DB
!             In Cartesians
! atnam: atom names in DD calc to match xref
!
! TSH info
! tshtrfile: file containing TSH coordinate trafo info
! tshtransf: transforms from Qcoo to Cartesians x = ddtransf*Q
!           (= nmcoo in VCHAM)
! tshtransb: transforms from Cartesians to Qcoo Q = ddtransb*(x - x0)
!           (= coonm in VCHAM)
! tshxcoo0 : x0 required in in backward transformation
!
!-----------------------------------------------------------------------
      integer(long), save                           :: dimref,nfam,&
                                                       trfiletyp = 0, &
                                                       coinfiletyp = 0

      real(dop), dimension(:), allocatable, save    :: ddxcoo0,xcoin,&
                                                       xrefpoint,xdcp0,xgd0
      real(dop), dimension(:,:), allocatable, save  :: ddtransf,ddtransb
      real(dop), dimension(:), allocatable, save    :: xmass,mref
      real(dop), dimension(:,:), allocatable, save  :: xref,xingeo

      real(dop), dimension(:,:), allocatable, save  :: tshtransf,tshtransb
      real(dop), dimension(:), allocatable, save    :: tshxcoo0
      real(dop), dimension(:), allocatable, save    :: tshmass
      character(len=c1), dimension(:), allocatable, save :: tshatnam

      logical(kind=4), save                         :: lrde0
      character(len=c1), dimension(:), allocatable, save :: atnam,xatnam
      character(len=c5), save                       :: ddtrfile='    ', &
                                                       coinfile='inp ',&
                                                       dreffile='inp ',&
                                                       tshtrfile='    '

!-----------------------------------------------------------------------
! Symmetry operations to be carried out on structures to make DB
! symmetrical
!
! dbpntgp : integer representing point group. 
!           See lib/utilities/symoplib.f90 for definition.
! dbpntgp0 : point group read from input file (may be different from DB).
! nsymop   : no. of symmetry operations
!
! dbsymmap(i,n) : mapping of atom i after symmetry operation n to
!                 return to initial configuration
!-----------------------------------------------------------------------
      integer(long), save   :: dbpntgp = 0, &
                               nsymop = 1, &
                               dbpntgp0 = 0

      integer(long), parameter   :: maxsymop = 10
      integer(long), dimension(:,:), allocatable, save :: dbsymmap

!-----------------------------------------------------------------------
! ddderiv1 and ddderiv2 required to pass first and second derivs from
! vqchem to derqchem
!-----------------------------------------------------------------------
      real(dop), dimension(:,:,:), allocatable, save   :: ddderiv1
      real(dop), dimension(:,:,:,:), allocatable, save :: ddderiv2

!-----------------------------------------------------------------------
! ddderiv1_gp and ddderiv2_gp required to pass first and second derivs
! from vqchem to derqchem for the dynamic small DB for each GWP
!-----------------------------------------------------------------------      

      real(dop), dimension(:,:,:,:), allocatable, save   :: ddderiv1_gp
      real(dop), dimension(:,:,:,:,:), allocatable, save :: ddderiv2_gp

!-----------------------------------------------------------------------
! db_geo, db_ener, db_grad, db_hess - the (part of) database, stored in
! memory.
! Currently all entries are kept in memory. Currently limited to dbmemdim
! entries.
!
! dbnrec: no. of entries in DB
! dbnrecold: no. of entries in DB prior to current step
! dbignore(irec) = 1  ignore DB entry in operations
! dbinterp(irec)      flags points that are interpolated
!                =-1  QC failed
!                = 0  Not interpolated
!                = 1  Near Degeneracy
!                = 2  Intruder State
!                = 3  State Flip
! dbsymrep: map of symmetry replicas in DB
!
! NB dbmoco is single precision!
!    dbmoco(n,i) : nth MO ith AO
! NB dbsoc only used for GAP and procrustes and not written to DB
!-----------------------------------------------------------------------
      integer(long), save                               :: dbnrec
      integer(long), save                               :: dbnrecold=0
      integer(long), dimension(:), allocatable, save    :: dbignore
      integer(long), dimension(:), allocatable, save    :: dbinterp
      integer(long), dimension(:,:), allocatable, save  :: dbsymrep
      real(dop), dimension(:,:), allocatable, save      :: dbgeo,dbnm,dbcoord
      real(dop), dimension(:,:,:), allocatable, save    :: dbener,dbcener
      real(dop), dimension(:,:,:), allocatable, save    :: dbtrans,dbctrans
      real(dop), dimension(:,:,:,:), allocatable, save  :: dbgrad
      real(dop), dimension(:,:,:,:,:), allocatable, save :: dbhess
      real(dop), dimension(:,:,:), allocatable, save    :: dbdercp
      real(dop), dimension(:,:,:,:), allocatable, save  :: dbdipmom
      real(dop), dimension(:,:,:), allocatable, save    :: dbgapwgt,dbcgapwgt
      real(sip), dimension(:,:,:), allocatable, save    :: dbmoco
      complex(dop), dimension(:,:,:), allocatable, save :: dbadsoc
      complex(dop), dimension(:,:,:), allocatable, save :: dbsoc
      complex(dop), dimension(:,:,:), allocatable, save :: dbsocwgt
      integer(long), dimension(:,:), allocatable, save  :: dbdiabrec
      character(30), dimension(:,:,:), allocatable      :: dbciconfig
      real(dop), dimension(:,:,:,:), allocatable        :: dbcicoeff

!-----------------------------------------------------------------------
! arrays used in dbtogapot DB to GAP conversion
! Allocated in dbtogap (wropermod) and used in dddb_rddb1 (dd_db)
!-----------------------------------------------------------------------
      integer(long)                                      :: db0nrec
      real(dop), dimension(:,:), allocatable, save       :: db0geo
      real(dop), dimension(:,:,:), allocatable, save     :: db0ener
      real(dop), dimension(:,:,:,:), allocatable, save   :: db0grad
      real(dop), dimension(:,:,:,:,:), allocatable, save :: db0hess
      real(dop), dimension(:,:,:,:), allocatable, save   :: db0dipmom

!-----------------------------------------------------------------------
! Dynamic version of database for each GWP 
! dbnrec_gp: no. of entries in DB
!
!-----------------------------------------------------------------------

      integer(long), dimension(:), allocatable, save       :: dbnrec_gp
      integer(long), save                                  :: num_gp
!numrec: number of records in the dynamic small database      
      integer(long), save                                  :: numrec=10
      
      !$omp threadprivate(num_gp)

!----------------------------------------------------------------------
! linked list pointers and arrays      
!-----------------------------------------------------------------------
      type nextptr  ! Derived type 
        integer(long)          :: locDB
        type(nextptr), pointer :: next 
      end type nextptr

     type list
        type(nextptr), Pointer :: locpt
     endtype list

!A list for each GWP containing all the locations from the nearest points       
     type(list), dimension(:) , allocatable,save :: ngp_loc

!-----------------------------------------------------------------------
! The following arrays are used in dddb_rd to store the last point
! calculated.
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable, save       :: xgpold
      real(dop), dimension(:,:), allocatable, save     :: vold
      real(dop), dimension(:,:,:), allocatable, save   :: der1old
      real(dop), dimension(:,:,:,:), allocatable, save :: der2old
      real(dop), dimension(:,:,:), allocatable, save   :: dipold


!-----------------------------------------------------------------------
! Arrays for use whilst re-diabatising a previously computed DB of 
! adiabatic points.
!-----------------------------------------------------------------------
      real(dop), dimension(:,:,:), allocatable, save    :: dbadener
      real(dop), dimension(:,:,:,:), allocatable, save  :: dbadgrad
      real(dop), dimension(:,:,:,:,:), allocatable, save:: dbadhess
      real(dop), dimension(:,:,:,:), allocatable, save  :: dbaddipmom

!-----------------------------------------------------------------------
! dbgeoref, dbenref, dbgraref, dbhesref - reference points energy, gradients and
! hessian used for the hessian updates
! So far up to 5 ref. structures possible
!-----------------------------------------------------------------------
      real(dop), dimension(:,:), allocatable, save       :: dbgeoref
      real(dop), dimension(:,:,:), allocatable, save     :: dbenref
      real(dop), dimension(:,:,:,:), allocatable, save   :: dbgraref
      real(dop), dimension(:,:,:,:,:), allocatable, save :: dbhesref

!-----------------------------------------------------------------------
! Saved rotation angle, point, energy, gradient, Hessian, and
! BS-like vectors for local diabatisation (in system coordinates except
! for reference BS-like vectors in DB coordinates).
! Accumulated rotation angles of local BS-like vectors.
! Set o for previous point  (t); set 2 for current point (t + dt).
! Flags for detection of crossing. Time for consistency test.
! To be improved: (i) possibly redundant with other saved and allocated
! variables whilst useful only when dddiab = 2 (create independent alloc/
! dealloc subroutines?); (ii) BS vectors for two states only at the moment.
! BL 06/16
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable, save         :: diab2ang,diab2tim
      real(dop), dimension(:), allocatable, save         :: diab2agd,diab2adc
      real(dop), dimension(:,:), allocatable, save       :: diabogeo,diab2geo
      real(dop), dimension(:,:,:), allocatable, save     :: diabopes,diab2pes
      real(dop), dimension(:,:,:,:), allocatable, save   :: diabogra,diab2gra
      real(dop), dimension(:,:,:,:,:), allocatable, save :: diabohes,diab2hes
      real(dop), dimension(:,:), allocatable, save       :: diab2bs1,diab2bs2
      integer(long), dimension(:), allocatable, save     :: diab2pre,diab2det
 
!-----------------------------------------------------------------------
! Arrays for information to set up DB 
! (see makeflagarray and makeinfoarray)
!-----------------------------------------------------------------------
      integer(long), dimension(7), save     :: infoarray !Info for setting up DB
      logical(kind=4), dimension(20), save  :: flagarray  !Flags for setting up DB

!-----------------------------------------------------------------------
! Define a derived type to carry the CI vectors from a CAS calculation.
! nconf : the number of determinants in the CI expansion of each 
!         multiplicity.
! config : the active-space orbital occupancy of each config.
! coeff  : the coefficient of each the config. for each state.
!-----------------------------------------------------------------------
       type civec
          integer(long)                            :: nconf
          character(30), dimension(:), allocatable :: config
          real(dop), dimension(:,:), allocatable   :: coeff
       end type civec
      
      contains

!#######################################################################
!
! DIRDYN_ALLOC
!
! memory required for arrays in dirdynmod allocated here.
!
! First check if allocated, then allocate. All arrays must be
! deallocated in
! dirdyn_dealloc
!
!#######################################################################

      subroutine alloc_dirdyn

      implicit none

      integer(long), parameter     :: dbrefdim=1

      integer(long)     :: ierr
      character(len=15) :: cerr

! check all arrays are not allocated
      call dealloc_dirdyn

      cerr = 'istatemap'
      allocate(istatemap(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'imultmap'
      allocate(imultmap(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'imultqc'
      allocate(imultqc(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'iatsym'
      allocate(iatsym(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nconf'
      allocate(nconf(nsmult,2),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nciconf'
      allocate(nciconf(nsmult),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ncasel'
      allocate(ncasel(nsmult,2),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'symdof'
      allocate(symdof(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gwpshell'
      allocate(gwpshell(maxgwp,2),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'fgwpshell'
      allocate(fgwpshell(maxgwp,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddxcoo0'
      allocate(ddxcoo0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xcoin'
      allocate(xcoin(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xrefpoint'
      allocate(xrefpoint(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xdcp0'
      allocate(xdcp0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xgd0'
      allocate(xgd0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddtransf'
      allocate(ddtransf(maxdim,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddtransb'
      allocate(ddtransb(maxdim,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      if (ltraj) then
         cerr = 'tshtransf'
         allocate(tshtransf(maxdim,maxdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'tshtransb'
         allocate(tshtransb(maxdim,maxdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'tshxcoo0'
         allocate(tshxcoo0(maxdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'tshmass'
         allocate(tshmass(maxdim/3),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'tshatnam'
         allocate(tshatnam(maxdim/3),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'xmass'
      allocate(xmass(maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xingeo'
      allocate(xingeo(3,maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xatnam'
      allocate(xatnam(maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddderiv1'
      allocate(ddderiv1(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddderiv2'
      allocate(ddderiv2(maxdim,maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbgeoref'
      allocate(dbgeoref(maxdim,dbrefdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbenref'
      allocate(dbenref(maxsta,maxsta,dbrefdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbgraref'
      allocate(dbgraref(maxdim,maxsta,maxsta,dbrefdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbhesref'
      allocate(dbhesref(maxdim,maxdim,maxsta,maxsta,dbrefdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xgpold'
      allocate(xgpold(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'vold'
      allocate(vold(maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'der1old'
      allocate(der1old(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'der2old'
      allocate(der2old(maxdim,maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dipold'
      allocate(dipold(3,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2ang'
      allocate(diab2ang(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2tim'
      allocate(diab2tim(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2agd'
      allocate(diab2agd(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2adc'
      allocate(diab2adc(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diabogeo'
      allocate(diabogeo(maxdim,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2geo'
      allocate(diab2geo(maxdim,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diabopes'
      allocate(diabopes(maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2pes'
      allocate(diab2pes(maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diabogra'
      allocate(diabogra(maxdim,maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2gra'
      allocate(diab2gra(maxdim,maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diabohes'
      allocate(diabohes(maxdim,maxdim,maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2hes'
      allocate(diab2hes(maxdim,maxdim,maxsta,maxsta,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2bs1'
      allocate(diab2bs1(maxdim,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2bs2'
      allocate(diab2bs2(maxdim,maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2pre'
      allocate(diab2pre(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'diab2det'
      allocate(diab2det(maxgwp),stat=ierr)
      if (ierr > 0) go to 1000
! arrays for small DB version data
      if (ldbsmall) then
         cerr = 'ddderiv1_gp'
         allocate(ddderiv1_gp(maxdim,maxsta,maxsta,maxgwp),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'ddderiv2_gp'
         allocate(ddderiv2_gp(maxdim,maxdim,maxsta,maxsta,maxgwp),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'dbnrec_gp'
         allocate(dbnrec_gp(maxgwp*maxsta),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'ngp_loc'
         allocate(ngp_loc(maxgwp*maxsta),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'dbsymmap'
      allocate(dbsymmap(maxdim,maxsymop),stat=ierr)
      if (ierr > 0) go to 1000

! initialise variables
      istatemap=0
      imultmap=0
      imultqc=0
      iatsym=0
      nconf=1
      nciconf=1
      ncasel=0
      symdof=0
      gwpshell=0
      fgwpshell=0
      ddxcoo0=0
      xcoin=0
      xrefpoint=0
      xdcp0=0
      xgd0=0
      ddtransf=0
      ddtransb=0
      if (ltraj) then
         tshtransf=0
         tshtransb=0
         tshxcoo0=0
      endif
      xmass=0
      xingeo=-999.0
      xatnam=' '
      dbgeoref=0.0_dop
      dbenref=0.0_dop
      dbgraref=0.0_dop
      dbhesref=0.0_dop
      ddderiv1=0.0_dop
      ddderiv2=0.0_dop
      xgpold=0
      vold=0
      der1old=0
      der2old=0
      dipold=0
      diab2ang=0
      diab2tim=0
      diab2agd=0
      diab2adc=0
      diabogeo=0
      diab2geo=0
      diabopes=0
      diab2pes=0
      diabogra=0
      diab2gra=0
      diabohes=0
      diab2hes=0
      diab2bs1=0
      diab2bs2=0
      diab2pre=0
      diab2det=0
      if (ldbsmall) then
         ddderiv1_gp=0.0_dop
         ddderiv2_gp=0.0_dop
         dbnrec_gp=0
      endif
      dbsymmap = 0

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DIRDYN :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_dirdyn

!#######################################################################

      subroutine dealloc_dirdyn

      if (allocated(istatemap)) deallocate(istatemap)
      if (allocated(istatemap)) deallocate(istatemap)
      if (allocated(imultmap)) deallocate(imultmap)
      if (allocated(imultqc)) deallocate(imultqc)
      if (allocated(iatsym)) deallocate(iatsym)
      if (allocated(nconf)) deallocate(nconf)
      if (allocated(nciconf)) deallocate(nciconf)
      if (allocated(ncasel)) deallocate(ncasel)
      if (allocated(symdof)) deallocate(symdof)
      if (allocated(gwpshell)) deallocate(gwpshell)
      if (allocated(fgwpshell)) deallocate(fgwpshell)
      if (allocated(ddxcoo0)) deallocate(ddxcoo0)
      if (allocated(xcoin)) deallocate(xcoin)
      if (allocated(xrefpoint)) deallocate(xrefpoint)
      if (allocated(xdcp0)) deallocate(xdcp0)
      if (allocated(xgd0)) deallocate(xgd0)
      if (allocated(ddtransf)) deallocate(ddtransf)
      if (allocated(ddtransb)) deallocate(ddtransb)
      if (allocated(tshxcoo0)) deallocate(tshxcoo0)
      if (allocated(tshtransf)) deallocate(tshtransf)
      if (allocated(tshtransb)) deallocate(tshtransb)
      if (allocated(xmass)) deallocate(xmass)
      if (allocated(tshmass)) deallocate(tshmass)
      if (allocated(xingeo)) deallocate(xingeo)
      if (allocated(xatnam)) deallocate(xatnam)
      if (allocated(tshatnam)) deallocate(tshatnam)
      if (allocated(dbgeoref)) deallocate(dbgeoref)
      if (allocated(dbenref)) deallocate(dbenref)
      if (allocated(dbgraref)) deallocate(dbgraref)
      if (allocated(dbhesref)) deallocate(dbhesref)
      if (allocated(ddderiv1)) deallocate(ddderiv1)
      if (allocated(ddderiv2)) deallocate(ddderiv2)
      if (allocated(ddderiv1_gp)) deallocate(ddderiv1_gp)
      if (allocated(ddderiv2_gp)) deallocate(ddderiv2_gp)
      if (allocated(xgpold)) deallocate(xgpold)
      if (allocated(vold)) deallocate(vold)
      if (allocated(der1old)) deallocate(der1old)
      if (allocated(der2old)) deallocate(der2old)
      if (allocated(dipold)) deallocate(dipold)
      if (allocated(diab2ang)) deallocate(diab2ang)
      if (allocated(diab2tim)) deallocate(diab2tim)
      if (allocated(diab2agd)) deallocate(diab2agd)
      if (allocated(diab2adc)) deallocate(diab2adc)
      if (allocated(diabogeo)) deallocate(diabogeo)
      if (allocated(diab2geo)) deallocate(diab2geo)
      if (allocated(diabopes)) deallocate(diabopes)
      if (allocated(diab2pes)) deallocate(diab2pes)
      if (allocated(diabogra)) deallocate(diabogra)
      if (allocated(diab2gra)) deallocate(diab2gra)
      if (allocated(diabohes)) deallocate(diabohes)
      if (allocated(diab2hes)) deallocate(diab2hes)
      if (allocated(diab2bs1)) deallocate(diab2bs1)
      if (allocated(diab2bs2)) deallocate(diab2bs2)
      if (allocated(diab2pre)) deallocate(diab2pre)
      if (allocated(diab2det)) deallocate(diab2det)
      if (allocated(dbnrec_gp)) deallocate(dbnrec_gp)
      if (allocated(ngp_loc)) deallocate(ngp_loc)
      if (allocated(dbsymmap)) deallocate(dbsymmap)
      if (allocated(dbsymrep)) deallocate(dbsymrep)

      return
      end subroutine dealloc_dirdyn

!#######################################################################
!
! arrays for DB data  needed in propagation
! SOC is stored in v and does not need own array
!
!#######################################################################

      subroutine alloc_dddb

      implicit none

      integer(long)     :: ierr
      character(len=15) :: cerr

      call dealloc_dddb

! take memory size from input commands if given
      if (dbmemdim0 .gt. 0) dbmemdim = dbmemdim0

      cerr = 'dbgeo'
      allocate(dbgeo(maxdim,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbener'
      allocate(dbener(maxsta,maxsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      if (lddsoc .and. .not. lsocdegen) then
         cerr = 'dbcener'
         allocate(dbcener(maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      if(.not.(lgausspot.or.lddmctdh))then
         cerr = 'dbgrad'
         allocate(dbgrad(maxdim,maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'dbhess'
         allocate(dbhess(maxdim,maxdim,maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'dbadener'
      allocate(dbadener(maxsta,maxsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbdercp'
      allocate(dbdercp(maxdim,maxsta*(maxsta-1)/2,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbadsoc'
      allocate(dbadsoc(maxqcsta,maxqcsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      if(lgausspot.or.lddmctdh)then
         cerr = 'dbsoc'
         allocate(dbsoc(maxqcsta,maxqcsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'dbdipmom'
      allocate(dbdipmom(3,maxsta,maxsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbmoco'
      allocate(dbmoco(maxmoco,maxmoco,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbnm'
      allocate(dbnm(maxdim,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbcoord'
      allocate(dbcoord(maxdim,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbtrans'
      allocate(dbtrans(maxsta,maxsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbinterp'
      allocate(dbinterp(dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbsymrep'
      allocate(dbsymrep(maxsymop,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      if (lcivec) then   ! CI vector for Procrustes
         cerr='dbciconfig'
         allocate(dbciconfig(maxcivec,nsmult,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr='dbcicoeff'
         allocate(dbcicoeff(maxcivec,maxmultblk,nsmult,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      else
         cerr='dbciconfig'
         allocate(dbciconfig(1,1,1),stat=ierr)
         if (ierr > 0) go to 1000
         cerr='dbcicoeff'
         allocate(dbcicoeff(1,1,1,1),stat=ierr)
         if (ierr > 0) go to 1000
      endif

      cerr = 'dbignore'
      allocate(dbignore(dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      if (lsocbas) then
         cerr = 'dbctrans'
         allocate(dbctrans(maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      if (ldb2gapot .or. lddmctdh) then
         cerr = 'dbgapwgt'
         allocate(dbgapwgt(maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'dbcgapwgt'
         allocate(dbcgapwgt(maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'dbsocwgt'
         allocate(dbsocwgt(maxsta,maxsta,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      else
         cerr = 'dbcgapwgt'
         allocate(dbcgapwgt(1,1,dbmemdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'dbdiabrec'
      allocate(dbdiabrec(2,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mref'
      allocate(mref(maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xref'
      allocate(xref(3,maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'atnam'
      allocate(atnam(maxdim/3),stat=ierr)
      if (ierr > 0) go to 1000

      dbgeo=0.0_dop
      dbener=0.0_dop
      if(allocated(dbgrad))dbgrad=0.0_dop
      if(allocated(dbhess))dbhess=0.0_dop
      dbadener=0.0_dop
      dbdercp=0.0_dop
      dbadsoc=cmplx(0.0_dop,0.0_dop,dop)
      if(allocated(dbsoc))dbsoc=0.0_dop
      dbdipmom=0.0_dop
      dbmoco=0.0
      dbnm=0.0_dop
      dbcoord=0.0_dop
      dbtrans=0.0_dop
      dbinterp=0
      dbsymrep=0
      dbignore=0
      dbinterp=0
      if (allocated(dbgapwgt)) dbgapwgt=0.0_dop
      if (allocated(dbsocwgt)) dbsocwgt=cmplx(0.0_dop,0.0_dop,dop)
      dbdiabrec(1,:)=0
      dbdiabrec(2,:)=1

      mref=0
      xref=-999.0
      atnam=' '

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DDDB :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_dddb

!#######################################################################

      subroutine dealloc_dddb

      if (allocated(dbgeo)) deallocate(dbgeo)
      if (allocated(dbener)) deallocate(dbener)
      if (allocated(dbcener)) deallocate(dbcener)
      if (allocated(dbgrad)) deallocate(dbgrad)
      if (allocated(dbhess)) deallocate(dbhess)
      if (allocated(dbadener)) deallocate(dbadener)
      if (allocated(dbdercp)) deallocate(dbdercp)
      if (allocated(dbadsoc)) deallocate(dbadsoc)
      if (allocated(dbsoc)) deallocate(dbsoc)
      if (allocated(dbdipmom)) deallocate(dbdipmom)
      if (allocated(dbmoco)) deallocate(dbmoco)
      if (allocated(dbnm)) deallocate(dbnm)
      if (allocated(dbcoord)) deallocate(dbcoord)
      if (allocated(dbtrans)) deallocate(dbtrans)
      if (allocated(dbctrans)) deallocate(dbctrans)
      if (allocated(dbinterp)) deallocate(dbinterp)
      if (allocated(dbignore)) deallocate(dbignore)
      if (allocated(dbsymrep)) deallocate(dbsymrep)
      if (allocated(dbciconfig)) deallocate(dbciconfig)
      if (allocated(dbcicoeff)) deallocate(dbcicoeff)
      if (allocated(dbgapwgt)) deallocate(dbgapwgt)
      if (allocated(dbcgapwgt)) deallocate(dbcgapwgt)
      if (allocated(dbsocwgt)) deallocate(dbsocwgt)
      if (allocated(dbdiabrec)) deallocate(dbdiabrec)
      if (allocated(mref)) deallocate(mref)
      if (allocated(xref)) deallocate(xref)
      if (allocated(atnam)) deallocate(atnam)

      return
      end subroutine dealloc_dddb

!#######################################################################
!
! allocate arrays needed in addition to those defined in
! alloc_dirdyn to put all the DB data in memory (needed in
! rediabatisation).
!
! nmode.dba does not memory as records are created from ddgeo on-the-fly
!
!#######################################################################

      subroutine alloc_dirdyn_rediab(nrec,nddstate,ndofddpes)

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=15) :: cerr

      integer(long), intent(in) :: nrec,nddstate,ndofddpes

! check all arrays are not allocated
      call dealloc_dirdyn_rediab

      cerr = 'dbadgrad'
      allocate(dbadgrad(ndofddpes,nddstate,nddstate,nrec),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbadhess'
      allocate(dbadhess(ndofddpes,ndofddpes,nddstate,nddstate,nrec),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dbaddipmom'
      allocate(dbaddipmom(3,nddstate,nddstate,nrec),stat=ierr)
      if (ierr > 0) go to 1000
      
      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DIRDYN_REDIAB :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_dirdyn_rediab

!####################################################################

      subroutine dealloc_dirdyn_rediab

      if (allocated(dbadgrad)) deallocate(dbadgrad)
      if (allocated(dbadhess)) deallocate(dbadhess)
      if (allocated(dbaddipmom)) deallocate(dbaddipmom)
      
      return
      end subroutine dealloc_dirdyn_rediab

!#######################################################################

      subroutine alloc_dirdyn_ddmctdh

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_dirdyn_ddmctdh

      cerr = 'dbadener'
      allocate(dbadener(maxsta,maxsta,dbmemdim),stat=ierr)
      if (ierr > 0) go to 1000
!      cerr = 'dbmoco'
!      allocate(dbmoco(nbasis,nbasis,dbmemdim),stat=ierr)
!      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DIRDYN_DDMCTDH :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_dirdyn_ddmctdh

!####################################################################

      subroutine dealloc_dirdyn_ddmctdh

      if (allocated(dbadener)) deallocate(dbadener)
!      if (allocated(dbmoco)) deallocate(dbmoco)

      return
      end subroutine dealloc_dirdyn_ddmctdh

!#######################################################################

      subroutine alloc_dercp

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

      call dealloc_dercp

      cerr = 'idxmult'
      allocate(idxmult(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'imultblk'
      allocate(imultblk(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'imultoff'
      allocate(imultoff(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dercpblk'
      allocate(dercpblk(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dercpoff'
      allocate(dercpoff(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stoff'
      allocate(stoff(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stblk'
      allocate(stblk(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddstoff'
      allocate(ddstoff(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddstblk'
      allocate(ddstblk(maxsta),stat=ierr)
      if (ierr > 0) go to 1000

      imultblk=0
      imultoff=0
      dercpblk=0
      dercpoff=0
      stoff=0
      stblk=0
      ddstoff=0
      ddstblk=0

      return

 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_DERCP :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_dercp

!#######################################################################
      subroutine dealloc_dercp

      if (allocated(idxmult)) deallocate(idxmult)
      if (allocated(imultblk)) deallocate(imultblk)
      if (allocated(imultoff)) deallocate(imultoff)
      if (allocated(dercpblk)) deallocate(dercpblk)
      if (allocated(dercpoff)) deallocate(dercpoff)
      if (allocated(stoff)) deallocate(stoff)
      if (allocated(stblk)) deallocate(stblk)
      if (allocated(ddstoff)) deallocate(ddstoff)
      if (allocated(ddstblk)) deallocate(ddstblk)

      return
      end subroutine dealloc_dercp

!#######################################################################

subroutine makeflagarray

      flagarray(:) = .false.
      flagarray(1) = lpesdb
      flagarray(2) = lapesdb
      flagarray(3) = lnmdb
      flagarray(4) = lnactdb
      flagarray(5) = ltransdb
      flagarray(6) = lgradb
      flagarray(7) = lagradb
      flagarray(8) = lhesdb
      flagarray(9) = lahesdb
      flagarray(10) = lmocodb
      flagarray(11) = lsymdb
      flagarray(12) = ldipdb
      flagarray(13) = ladipdb
      flagarray(14) = lcasocdb
      flagarray(15) = lcivecdb
      flagarray(16) = lcpesdb
      flagarray(17) = lctransdb
      flagarray(18) = lqcfreq
      
end subroutine makeflagarray

subroutine makeinfoarray(nstates)

      integer(long),intent(in)                   ::nstates

      infoarray(1) = natm !natom 
      infoarray(2) = nstates            !nstates 
      infoarray(3) = (infoarray(1)*3)-6 !nmodes based on natom
      infoarray(4) = nbasis             !nbasis
      infoarray(5) = nsymop             !Number of symmetry operations
      infoarray(6) = maxmultblk         !Max. number of states in spin blocks
      infoarray(7) = nsocstate          !nqcstates (including spin degeneracies)
      
end subroutine makeinfoarray

      end module dirdyn

