!***********************************************************************
!                                                                     
!              MCTDH module file griddatmod.f90
!                                                                     
!  MCTDH include file defining array sizes,  parameters etc.
!  which are relevant for the primitive basis (grid).
!                                                                     
! Arrays to be allocated must be assigned in dvrdat_alloc.f90
!
!  11/96 AJ                                                         
!                                                                     
!***********************************************************************
      module dvrdatmod

      use decimal, only : dop,long
      use global, only : c1
      use maxv

      implicit none

!-----------------------------------------------------------------------
! VARIOUS INTEGERS DEFINING SYSTEM
!
! ndof     : total number of degrees of freedom
! feb      : number of degree of freedom described by an electronic basis
! fkoe     : dof containing coefficient information (ndof+1)
! ftime    : dof containing time information (ndof+2)
! fener    : dof containing energy information (ndof+3)
! fcut     : dof containining information on a cut to be plotted
! fpb      : number of degree of freedom described by a packet basis
! gdim(f)  : number of grid points for dof f
! maxgdim  : Max[gdim(f)]
! maxgdim2 : Max[gdim(f)**2]
! maxdof   : Highest no. of DOFs in a particle
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save :: gdim
      integer(long), save                            :: ndof,feb,fkoe,maxgdim,&
                                                        maxgdim2,ftime,fener,&
                                                        fpb,maxdof,fcut

!-----------------------------------------------------------------------
! CONTINUUM ELECTRONIC BASIS
!
!  lebcont    : if true continuum electronic states present
!  sebcont(s) : if true state s is a continuum 
!  febcont    : DOF with continuum electronic state information
!-----------------------------------------------------------------------
      logical(kind=4), save                            :: lebcont = .false.
      integer(long), save                              :: febcont = 0
      logical(kind=4), dimension(:), allocatable, save :: sebcont

!-----------------------------------------------------------------------
! VARIOUS INTEGERS DEFINING SIZE OF DVR
!
! ortdim   : Length of array containing grid positions
!            ortdim = Sum[subdim(m)]
! dvrdim   : Length of array containing DVR representations
!            dvrdim = Sum[subdim(m)*subdim(m)]
!            where sum is over modes using DVR
! fftdim   : Length of array containing FFT representations
!            fftdim = Sum[subdim(m)]
!            where sum is over modes using FFT
! expdim   : Length of complex transformation array of EXP-DVRs
!            expdim = Sum[gdim(f)*kdim]
!            where sum is over modes using EXP-DVR
! sphdim: length of arrays related to spherical harmonics primitive
!         basis: sphdim = Sum[subdim(m)]
!-----------------------------------------------------------------------
      integer(long), parameter :: dvrord = 2

      integer(long), save      :: ortdim,dvrdim,fftdim,expdim,sphdim

!-----------------------------------------------------------------------
! PARAMETERS FOR ARRAY DIMENSIONS
!
!  mbaspar:    maximum number of parameters used to define bases
!-----------------------------------------------------------------------
      integer(long), parameter :: mbaspar = 9 

!-----------------------------------------------------------------------
! PARAMETERS DEFINING PRIMITIVE BASIS
!
! basis(f) : basis type for dof f
! rpbaspar(n,f): real parameters needed to define primitive basis
! ipbaspar(n,f): integer parameters needed to define primitive basis
! eltot,elbnd: if a continuum electronic basis is used, eltot is total 
!              no. of electronic states and elbnd no. of bound states
! lxloc  : one (or more) dofs is local in x (i.e. a trajectory)
! lpntdvr: one (or more) dofs is a single point
! ltpntdvr: one (or more) dofs is a single time-dependent point
! atpntop: the operator label defining the time-dependent point
! itpntop: the operator number defining the time-dependent point
!
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save         :: basis,itpntop
      integer(long), dimension(:,:), allocatable, save       :: ipbaspar
      integer(long), save                                    :: eltot,elbnd

      real(dop), dimension(:,:), allocatable, save           :: rpbaspar
      real(dop), dimension(:,:), allocatable, save           :: xend

      logical(kind=4), save                                  :: lxloc, &
                                                                lpntdvr, &
                                                                ltpntdvr
      logical(kind=4), dimension(:), allocatable, save       :: ldvr
      character(c1), dimension(:), allocatable, save         :: modelabel,atpntop

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! zort(f)    : pointer for ort matrix for dof f
! zdvr(f)    : pointer for dvr matrices (dif1mat, dif2mat, trafo) for dof f
! zfft(f)    : pointers for FFT grid matrices for dof f 
! zexp(f)    : pointers for transformation matrixes for EXP-DVR for 
!              dof f
! zsph(f)    : pointer for jsph, msph, kinsph matrices for mode m
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save :: zort,zdvr,zfft,zexp,zsph

!-----------------------------------------------------------------------
! Special parameters needed in various parts of the program
!
! jtot     : Total angular momentum
! jbf      : (fixed) angular m or K quantum number. (Leg, coupled states).
! csmass   : mass to define a coupled-states centrifugal potential (potfit)
!-----------------------------------------------------------------------
      real(dop), save                            :: jtot,jbf,csmass

      contains

!#######################################################################
!
! DVRDAT_ALLOC
!
! memory required for arrays in dvrdatmod allocated here.
!
! First check if allocated, then allocate. All arrays must be deallocated in 
! dvrdat_dealloc
!
! sum up memory allocated and store in allocmemory
!
!#######################################################################

      subroutine alloc_dvrdat

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr
      integer(long), external   :: integersize, realsize, logicalsize

! check all arrays are not allocated
      call dealloc_dvrdat

      cerr = 'gdim'
      allocate(gdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'modelabel'
      allocate(modelabel(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*c1
      if (ierr > 0) go to 1000
      cerr = 'basis'
      allocate(basis(maxdim),stat=ierr)
      basis=-1
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
! test
      basis(:)=999
! test

      cerr = 'itpntop'
      allocate(itpntop(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'atpntop'
      allocate(atpntop(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*c1
      if (ierr > 0) go to 1000
      cerr = 'ipbaspar'
      allocate(ipbaspar(mbaspar,maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*mbaspar*integersize()
      if (ierr > 0) go to 1000
      cerr = 'rpbaspar'
      allocate(rpbaspar(mbaspar,maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*mbaspar*realsize()
      if (ierr > 0) go to 1000
      cerr = 'xend'
      allocate(xend(2,maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*2*realsize()
      if (ierr > 0) go to 1000
      cerr = 'ldvr'
      allocate(ldvr(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*logicalsize()
      if (ierr > 0) go to 1000
      cerr = 'zort'
      allocate(zort(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*realsize()
      if (ierr > 0) go to 1000
      cerr = 'zdvr'
      allocate(zdvr(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'zfft'
      allocate(zfft(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'zexp'
      allocate(zexp(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'zsph'
      allocate(zsph(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'sebcont'
      allocate(sebcont(maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*logicalsize()
      if (ierr > 0) go to 1000


      return
 1000 write(6,'(a)') 'Error allocating memory in ALLOC_DVRDAT :',cerr
      write(6,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_dvrdat

!#######################################################################

      subroutine dealloc_dvrdat

      implicit none

      if (allocated(gdim)) deallocate(gdim)
      if (allocated(modelabel)) deallocate(modelabel)
      if (allocated(basis)) deallocate(basis)
      if (allocated(itpntop)) deallocate(itpntop)
      if (allocated(atpntop)) deallocate(atpntop)
      if (allocated(ipbaspar)) deallocate(ipbaspar)
      if (allocated(rpbaspar)) deallocate(rpbaspar)
      if (allocated(xend)) deallocate(xend)
      if (allocated(ldvr)) deallocate(ldvr)
      if (allocated(zort)) deallocate(zort)
      if (allocated(zdvr)) deallocate(zdvr)
      if (allocated(zfft)) deallocate(zfft)
      if (allocated(zexp)) deallocate(zexp)
      if (allocated(zsph)) deallocate(zsph)
      if (allocated(sebcont)) deallocate(sebcont)

      return
      end subroutine dealloc_dvrdat

      end module dvrdatmod
