!#######################################################################
!                      GAPMOD.F90
!#######################################################################
      module gapmod

      use decimal, only : long,dop
      use global
      use maxv

      implicit none

!-----------------------------------------------------------------------
! actdof : number of active DOFs.
!-----------------------------------------------------------------------
      integer(long), save   :: actdof=0

!-----------------------------------------------------------------------
! lgaptest : flag to activate the test routine for the GAP.
! iddmtest : allows running of DDMCTDH tests on fitted potentials.
! ddmorder : 0 gives standard GAP, >0 gives order of additive GAP.
! lddmopt  : whether to use optimised potential based on GAP
! lddgbtendec  : whether to use a tesnor decomposed potential, based on GAP.
! ddmoptord : order of optimised GAP.
! lgapderiv : whether to use GAP with derivatives. 
! lgapusesym : whether to use GAP with symmetry. 
! lgapuseinwf : whether to use initial geometry coordinates in frozen 
!               coordinates. 
! lgapsobol : Use of Sobol sequence to sample geometries in GAP.
! lgapallgrd : Use the DVR gridpoints as the GAP sample.
! lgapvarmax : Use of variance maximiser when sampling geometries in GAP.
! gapsample : number of points to sample at each step.
! lgapcomp : flag to run a PES comparison (i.e. between GAP and 
!            ab initio calculated energies).
! ngapcomp : number of points to sample when doing a comparison run.
! lmalon : test flag to call routine to generate potfit data for malonaldehyde.
! lgapethresh : flag to determine whether to use E threshold in determining DB.
! gapethresh : energy threshold above E0 at which energies (S0) are rejected.
! gapeshift  : energy shift used in the GAP fitting to prevent the PES falling 
!              off to 0 at +/- infinity.
! lgapparopt : flag to optimise kernel parameters at end of propagation.
! lgapvarprint : flag to print the GAP variances at the DVR points for
!                exact calculations.
! lgapadiabwf : Adiabatise the initial wavefunction for a read-only
!               calculation.
!-----------------------------------------------------------------------
      logical(kind=4), save :: lgaptest=.false.
      integer(long), save   :: iddmtest=0
      logical(kind=4), save :: lddmopt=.false.
      logical(kind=4), save :: lddgbtendec=.false.
      integer(long), save   :: ddmorder=0
      integer(long), save   :: ddmoptord=1
      logical(kind=4), save :: lgapderiv=.false.
      logical(kind=4), save :: lgapusesym=.false.
      logical(kind=4), save :: lgapuseinwf=.false.
      logical(kind=4), save :: lgapsobol=.true.
      logical(kind=4), save :: lgapallgrd=.false.
      logical(kind=4), save :: lgapvarmax=.false.
      integer(long), save   :: gapsample=100
      logical(kind=4), save :: lgapcomp=.false.
      integer(long), save   :: ngapcomp=1
      logical(kind=4), save :: lmalon=.false.
      logical(kind=4), save :: lgapethresh=.false.
      real(dop), save       :: gapethresh=9999999_dop
      real(dop), save       :: gapeshift=0.0_dop
      logical(kind=4), save :: lgapparopt=.false.
      logical(kind=4), save :: lgapvarprint=.false.
      logical(kind=4), save :: lgapadiabwf=.false.

!-----------------------------------------------------------------------
! Variables and data used in constructing the Sobol sequence needed to 
! sample configuration space non-randomly.
!-----------------------------------------------------------------------
      integer(long), parameter :: maxb=30,maxsob=36,sumpdeg=209
      integer(long), dimension(maxsob), save :: ipoly,polydeg,intx
      integer(long), dimension(maxsob)       :: idnpnt
      integer(long), dimension(sumpdeg)      :: dirninit
      integer(long), allocatable, save       :: dirnno(:,:)
      integer(long), save                    :: nchk,idxn
      real(dop), save                        :: sobfac

      data ipoly /0,1,1,2,1,4,2,4,7,11,13,14,1,13,16,19,22,25,1,4,7,8,14,19,&
                  21,28,31,32,37,41,42,50,55,56,59,62/
      data polydeg /1,2,2*3,2*4,6*5,6*6,18*7/
      data idnpnt /0,1,3,6,9,13,17,22,27,32,37,42,47,53,59,65,71,77,&
                   83,90,97,104,111,118,125,132,139,146,153,160,167,174,&
                   181,188,195,202/
      data dirninit /1,1,1,1,3,7,1,3,3,1,1,3,13,1,1,5,9,1,1,5,13,19,1,1,7,11,&
           23,1,3,5,7,13,1,3,7,13,29,1,1,3,11,31,1,3,7,11,29,1,1,3,7,13,41,&
           1,3,7,11,17,37,1,1,5,11,23,59,1,3,5,7,29,53,1,1,1,11,19,47,&
           1,3,3,13,31,61,1,1,3,7,11,35,95,1,3,7,15,21,61,111,1,3,5,13,29,47,&
           87,1,1,5,9,15,59,67,1,3,3,5,31,51,79,1,1,7,7,17,33,121,1,1,3,5,19,&
           39,101,1,1,7,9,21,41,93,1,3,5,11,25,49,63,1,3,7,13,23,63,75,1,1,1,&
           3,17,53,61,1,1,3,9,15,29,55,1,3,5,15,19,59,97,1,3,7,7,21,47,89,1,&
           3,3,9,25,45,105,1,3,3,13,23,33,77,1,1,5,11,31,55,65,1,1,3,13,17,&
           43,99/
!-----------------------------------------------------------------------
! Variables and DB for Gaussian approximation to the potential.
!
! gapupdate : controls updating of DD surface
!          = 1  : time control
!          = 0  : never (keep first guess)
! gapuptim  : time criterion for GAP PES update
!
! gapuptim0 : last time at which DB was updated
! gapalpha : the coefficient in the GAP exponent
! gapsigma : the regularisation constant in the GAP covariance matrix
! gapkfac : the factor coefficient of the covariance kernel
! gapvartol : the value of the variance above which new DB points are added
! gapsd : no. of SDs around the WP centre from which trial DB points are chosen
! gapdiag : the diagonal term in the A-matrix i.e. the regularised 
!           variance.
! gapker : array indicating the kernel to be used for each DOF.
! gapkerscl : coefficient in the exponent for scaling coordinates when
!             using the scaled Gaussian kernel.
!-----------------------------------------------------------------------
      integer(long), save   :: gapupdate
      real(dop), save       :: gapuptim,gapuptim0
      real(dop), save       :: gapsigma=1.0e-8_dop,&
                               gapvartol=1.0e-6_dop,gapsd=3.0_dop,&
                               optgapalpha=1.0_dop,gapkfac=1.0_dop
      real(dop), save       :: gapdiag
      real(dop), allocatable :: gapalpha(:),gapkerscl(:)
      integer(long), allocatable :: gapker(:)

!-----------------------------------------------------------------------
! achol: Array which holds the lower triangle form of the Cholesky 
!        decomposition of the A-matrix in GAP.
! acholtmp : temporary array which holds the old version of achol whilst
!        the latter is being updated.
! acholdim: Size of the achol array.
! acholtmpdim: Size of the acholtmp array.
!-----------------------------------------------------------------------
      integer(long), save :: acholtmpdim=0,acholdim=0
      real(dop), dimension(:,:), allocatable :: achol,acholtmp

!-----------------------------------------------------------------------
! ARRAY SIZES
!
! pmatgapdim  : Length of hteil and hloch arrays containing <phi|h|phi>
!               matrices
!               pmatdim = Sum[dim(m)*dim(m)*kzahl]
! hpsigapdim  : Length of hpsi array.
!               The actual size depends on the calculation mode.
! hlochgapdim : Length of array containing mean-field.
!               The actual size depends on the calculation mode.
! hopsgapdim  : Size of the GAP hops array.
!-----------------------------------------------------------------------
      integer(long), save :: pmatgapdim,hpsigapdim,hlochgapdim,hopsgapdim

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! pmatgap(m,k,irec)  : pointer to <phi(i)|h|phi(j)> matrices for the mth mode
!                      and kth Hamiltonian coefficient in "HTeil"
! phlochgap(m,k,irec): Pointer to <phi(i)|h|phi(j)> matrices for mth mode and
!                      kth Hamiltonian term in "hloch"
! zhpsigap(m,k,irec) : pointer to start of h|psi> vector
! zhopgap(m,k,irec)  : pointer to start of h vector
!-----------------------------------------------------------------------
      integer(long), dimension(:,:,:), allocatable, save :: zhopgap
      integer(long), dimension(:,:,:), allocatable, save :: zhpsigap
      integer(long), dimension(:,:,:), allocatable, save :: pmatgap,phlochgap

!-----------------------------------------------------------------------
! ACTUAL ARRAYS
!
! hopsgap  : holds the values of the GAP at the gridpoints.
! hpsigap  : product of GAP and SPFs
! hlochgap : mean-field matrices
! hteilgap : SPF matrix elements
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable    :: hopsgap
      complex(dop), dimension(:), allocatable :: hpsigap,hlochgap,hteilgap

!-----------------------------------------------------------------------
! Wavefunction positions and widths (along each DOF) for use with GAP.
!-----------------------------------------------------------------------
      real(dop), dimension(:,:), allocatable, save :: wfposn,wfwidth

!-----------------------------------------------------------------------
! Mapping array for DOFS.
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save :: gridmap    

!-----------------------------------------------------------------------
! Variables for optimised GAP
! ktotog: total number of terms in the list of operator terms.
! ktotcog: total number of correlated terms in the list of operator terms.
!-----------------------------------------------------------------------
      integer(long), save :: ktotog=0,ktotcog=0

!-----------------------------------------------------------------------
! Variables for SVD/Tensor fitted GAP.
! ddgbsing: singular value threshold above which terms are included in 
!           SVD expansion (intra-state).
! ddgbsingss1: singular value threshold above which terms are included in 
!           SVD expansion (inter-state).
! ddgbtens: convergence parameter for the energy residual in the 
!           PARAFAC decomposition procedure.
!-----------------------------------------------------------------------
      real(dop), save :: ddgbsing=1.0e-3_dop,ddgbtens=1.0e-3_dop,&
           ddgbsingss1=1.0e-3_dop

!-----------------------------------------------------------------------
! Derived type used for holding information on the Optimised/SVD GAP 
!-----------------------------------------------------------------------
      type optgappar
         logical(kind=4)                          :: lcorr,lkfirst
         integer(long)                            :: ndofs,k,ki,kf,totgd
         integer(long)                            :: kidx,kcorridx
         integer(long), dimension(:), allocatable :: dofs,gdofs,goffs
         real(dop), dimension(:), allocatable     :: alpha,xi,xf
         real(dop), dimension(:,:), allocatable   :: gaport
         real(dop), dimension(:,:,:), allocatable :: coeff
         type(optgappar), pointer                 :: next,prev
      end type optgappar

      type(optgappar), pointer :: firstogpar
      type(optgappar), pointer :: ogpar,ogpar1

      contains

!#######################################################################
      subroutine ogparinit(initogpar,listogpar)

use channels, only: ilog

      implicit none

      type(optgappar), pointer :: initogpar,listogpar

      integer(long)     :: ierr
      character(len=20) :: cerr

      if(associated(initogpar))call ogpardelall(initogpar,listogpar)
!      if(allocated(initogpar))deallocate(initogpar)

      allocate(initogpar,stat=ierr)
      if(ierr.gt.0)goto 1000

      if(allocated(initogpar%dofs))deallocate(initogpar%dofs)
      cerr = 'dofs'
      allocate(initogpar%dofs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(initogpar%gdofs))deallocate(initogpar%gdofs)
      cerr = 'gdofs'
      allocate(initogpar%gdofs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(initogpar%goffs))deallocate(initogpar%goffs)
      cerr = 'goffs'
      allocate(initogpar%goffs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(initogpar%alpha))deallocate(initogpar%alpha)
      cerr = 'alpha'
      allocate(initogpar%alpha(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(initogpar%xi))deallocate(initogpar%xi)
      cerr = 'xi'
      allocate(initogpar%xi(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(initogpar%xf))deallocate(initogpar%xf)
      cerr = 'xf'
      allocate(initogpar%xf(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

!      if(allocated(initogpar%coeff))deallocate(initogpar%coeff)
!      cerr = 'coeff'
!      allocate(initogpar%coeff(maxgrid,ddmoptord),stat=ierr)
!      if(ierr.gt.0)goto 1001
!      if(allocated(initogpar%gaport))deallocate(initogpar%gaport)
!      cerr = 'gaport'
!      allocate(initogpar%gaport(maxgrid,ddmoptord),stat=ierr)
!      if(ierr.gt.0)goto 1001

      initogpar%lcorr = .false.
      initogpar%ndofs = 0
      initogpar%k = 0
      initogpar%ki = 0
      initogpar%kf = 0
      initogpar%kidx = 0
      initogpar%kcorridx = 0
      initogpar%dofs = 0
      initogpar%gdofs = 0
      initogpar%goffs = 0
      initogpar%alpha = optgapalpha
      initogpar%xi = 0.0_dop
      initogpar%xf = 0.0_dop
      initogpar%totgd = 0
!      initogpar%coeff = 1.0e0_dop
!      initogpar%gaport = 0.0_dop

      listogpar => initogpar
      nullify(listogpar%next)
      nullify(listogpar%prev)

      return

 1000 write(ilog,'(a)') 'Error allocating initial optgap list'
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
 1001 write(ilog,'(a)') 'Error allocating initial optgap list array:',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine ogparinit

!#######################################################################
      subroutine ogparaddcpy(listogpar,vcdim)

      implicit none

      integer(long), intent(in) :: vcdim
      type(optgappar), intent(inout), pointer :: listogpar

      type(optgappar), pointer :: tmpogpar

      if(.not.associated(listogpar%next)) call ogparadd(listogpar)

      listogpar => listogpar%prev

      tmpogpar => listogpar%next

      tmpogpar%ndofs = listogpar%ndofs
      tmpogpar%dofs = listogpar%dofs
      tmpogpar%gdofs = listogpar%gdofs
      tmpogpar%goffs = listogpar%goffs
      tmpogpar%totgd = listogpar%totgd
      tmpogpar%alpha = listogpar%alpha
      tmpogpar%xi = listogpar%xi
      tmpogpar%xf = listogpar%xf
      tmpogpar%lcorr = listogpar%lcorr
      tmpogpar%lkfirst = .false.
      tmpogpar%k = listogpar%k
      tmpogpar%ki = listogpar%ki
      tmpogpar%kf = listogpar%kf
      tmpogpar%kidx = listogpar%kidx + 1
      if(listogpar%lcorr)tmpogpar%kcorridx = listogpar%kcorridx + 1

      call alloc_ogpararr(tmpogpar,vcdim)
      tmpogpar%gaport = listogpar%gaport
      tmpogpar%coeff = listogpar%coeff

      tmpogpar%prev => listogpar
      listogpar%next => tmpogpar

      listogpar => listogpar%next

      nullify(tmpogpar)

      return
      end subroutine ogparaddcpy

!#######################################################################
      subroutine ogparadd(listogpar)

use channels, only: ilog

      implicit none

      type(optgappar), pointer :: listogpar

      integer(long)     :: ierr
      character(len=10) :: cerr

      allocate(listogpar%next,stat=ierr)
      if(ierr.gt.0)goto 1000

      if(allocated(listogpar%next%dofs))deallocate(listogpar%next%dofs)
      cerr = 'dofs'
      allocate(listogpar%next%dofs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(listogpar%next%gdofs))deallocate(listogpar%next%gdofs)
      cerr = 'gdofs'
      allocate(listogpar%next%gdofs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(listogpar%next%goffs))deallocate(listogpar%next%goffs)
      cerr = 'goffs'
      allocate(listogpar%next%goffs(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(listogpar%next%alpha))deallocate(listogpar%next%alpha)
      cerr = 'alpha'
      allocate(listogpar%next%alpha(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(listogpar%next%xi))deallocate(listogpar%next%xi)
      cerr = 'xi'
      allocate(listogpar%next%xi(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

      if(allocated(listogpar%next%xf))deallocate(listogpar%next%xf)
      cerr = 'xf'
      allocate(listogpar%next%xf(ddmoptord),stat=ierr)
      if(ierr.gt.0)goto 1001

!      if(allocated(listogpar%next%coeff))deallocate(listogpar%next%coeff)
!      cerr = 'coeff'
!      allocate(listogpar%next%coeff(maxgrid,ddmoptord),&
!           stat=ierr)
!      if(ierr.gt.0)goto 1001

!      if(allocated(listogpar%next%gaport))deallocate(listogpar%next%gaport)
!      cerr = 'gaport'
!      allocate(listogpar%next%gaport(maxgrid,ddmoptord),&
!           stat=ierr)
!      if(ierr.gt.0)goto 1001

      listogpar%next%prev => listogpar
      listogpar => listogpar%next
      nullify(listogpar%next)

      listogpar%lcorr = .false.
      listogpar%ndofs = 0
      listogpar%k = 0
      listogpar%ki = 0
      listogpar%kf = 0
      listogpar%kidx = 0
      listogpar%kcorridx = 0
      listogpar%dofs = 0
      listogpar%gdofs = 0
      listogpar%goffs = 0
      listogpar%alpha = optgapalpha
      listogpar%xi = 0.0_dop
      listogpar%xf = 0.0_dop
      listogpar%totgd = 0
!      listogpar%coeff = 1.0e0_dop
!      listogpar%gaport = 0.0_dop

      return

 1000 write(ilog,'(a)') 'Error allocating next optgap list'
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
 1001 write(ilog,'(a)') 'Error allocating next optgap list array:',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine ogparadd

!#######################################################################
      subroutine alloc_ogpararr(ogpar1,vcdim)

use channels, only: ilog

      implicit none

      integer(long), intent(in)      :: vcdim
      type(optgappar), intent(inout) :: ogpar1

      integer(long)     :: ierr,maxgrid
      character(len=20) :: cerr

      call dealloc_ogpararr(ogpar1)

      maxgrid = maxval(ogpar1%gdofs(1:ogpar1%ndofs))

      cerr = 'ogpar%coeff'
      allocate(ogpar1%coeff(vcdim,maxgrid,ddmoptord),stat=ierr)
      if(ierr > 0)goto 1000

      cerr = 'ogpar%gaport'
      allocate(ogpar1%gaport(maxgrid,ddmoptord),stat=ierr)
      if(ierr > 0)goto 1000

      ogpar1%coeff = 1.0e-1_dop

      return

1000  write(ilog,'(a)') 'Error allocating memory in ALLOC_OGPARARR',cerr
      write(ilog,'(a,i5)') 'Error code :',ierr
      stop

      end subroutine alloc_ogpararr

!#######################################################################
      subroutine dealloc_ogpararr(ogpar1)

      implicit none

      type(optgappar), intent(inout) :: ogpar1

      if(allocated(ogpar1%coeff)) deallocate(ogpar1%coeff)
      if(allocated(ogpar1%gaport)) deallocate(ogpar1%gaport)

      return
      end subroutine dealloc_ogpararr

!#######################################################################
! Rewind the list of GAP operators
!#######################################################################

      subroutine ogparrew(listogpar)

      implicit none

      type(optgappar), pointer :: listogpar

      do while (associated(listogpar%prev))
         listogpar => listogpar%prev
      enddo

      return
      end subroutine ogparrew

!#######################################################################
      subroutine ogpardelall(initogpar,listogpar)

      implicit none

      type(optgappar), pointer :: initogpar,listogpar

      do while (associated(listogpar%prev))
         if(allocated(listogpar%dofs))deallocate(listogpar%dofs)
         if(allocated(listogpar%gdofs))deallocate(listogpar%gdofs)
         if(allocated(listogpar%goffs))deallocate(listogpar%goffs)
         if(allocated(listogpar%alpha))deallocate(listogpar%alpha)
         if(allocated(listogpar%xi))deallocate(listogpar%xi)
         if(allocated(listogpar%xf))deallocate(listogpar%xf)
         if(allocated(listogpar%coeff))deallocate(listogpar%coeff)
         if(allocated(listogpar%gaport))deallocate(listogpar%gaport)
         listogpar => listogpar%prev
         deallocate(listogpar%next)
         nullify(listogpar%next)
      enddo

      nullify(listogpar)

      if(allocated(initogpar%dofs))deallocate(initogpar%dofs)
      if(allocated(initogpar%gdofs))deallocate(initogpar%gdofs)
      if(allocated(initogpar%goffs))deallocate(initogpar%goffs)
      if(allocated(initogpar%alpha))deallocate(initogpar%alpha)
      if(allocated(initogpar%xi))deallocate(initogpar%xi)
      if(allocated(initogpar%xf))deallocate(initogpar%xf)
      if(allocated(initogpar%coeff))deallocate(initogpar%coeff)
      if(allocated(initogpar%gaport))deallocate(initogpar%gaport)
      if(associated(initogpar%next))nullify(initogpar%next)
      nullify(initogpar)

      return
      end subroutine ogpardelall

!#######################################################################
      subroutine ogpardelone(listogpar)

      implicit none

      type(optgappar), pointer :: listogpar

      if(.not.associated(listogpar%prev))return

      if(allocated(listogpar%dofs))deallocate(listogpar%dofs)
      if(allocated(listogpar%gdofs))deallocate(listogpar%gdofs)
      if(allocated(listogpar%goffs))deallocate(listogpar%goffs)
      if(allocated(listogpar%alpha))deallocate(listogpar%alpha)
      if(allocated(listogpar%xi))deallocate(listogpar%xi)
      if(allocated(listogpar%xf))deallocate(listogpar%xf)
      if(allocated(listogpar%coeff))deallocate(listogpar%coeff)
      if(allocated(listogpar%gaport))deallocate(listogpar%gaport)
      listogpar => listogpar%prev
      deallocate(listogpar%next)
      nullify(listogpar%next)

      return
      end subroutine ogpardelone

!#######################################################################

      subroutine dealloc_gapmod

      implicit none

      call dealloc_gapwfmod
      call dealloc_gappntmod
      call dealloc_gaparrmod
      call dealloc_gapmisc
      call dealloc_gapkernel
      if(lddmopt .or. lddgbtendec)then
         call ogpardelall(firstogpar,ogpar)
      endif

      if(allocated(achol)) deallocate(achol)
      if(allocated(acholtmp)) deallocate(acholtmp)

      return
      end subroutine dealloc_gapmod

!#######################################################################

      subroutine alloc_gappntmod

use maxv
use dirdyn, only: dbnrec
use griddatmod, only: nmode
use operdef, only: zham,kzahl,khtot
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_gappntmod

      if(lexact)then
         cerr = 'zhopgap'
         if(lddmopt .or. lddgbtendec)then
            allocate(zhopgap(nmode,ktotog,1),stat=ierr)
         else
            allocate(zhopgap(nmode,1:khtot(1),1),stat=ierr)
         endif
         if (ierr > 0) go to 1000
      else
         cerr = 'zhpsigap'
         if(lddmopt .or. lddgbtendec)then
            allocate(zhpsigap(nmode,ktotcog,1),stat=ierr)
         else
            allocate(zhpsigap(nmode,zham(1):zham(1)+kzahl(1)-1,dbnrec),&
                 stat=ierr)
         endif
         if (ierr > 0) go to 1000
         cerr = 'pmatgap'
         if(lddmopt .or. lddgbtendec)then
            allocate(pmatgap(nmode,ktotcog+maxsta,1),stat=ierr)
         else
            allocate(pmatgap(nmode,zham(1):zham(1)+kzahl(1)-1+maxsta,dbnrec),&
                 stat=ierr)
         endif
         if (ierr > 0) go to 1000
         cerr = 'phlochgap'
         if(lddmopt .or. lddgbtendec)then
            allocate(phlochgap(nmode,ktotcog,1),stat=ierr)
         else
            allocate(phlochgap(nmode,zham(1):zham(1)+kzahl(1)-1,dbnrec),&
                 stat=ierr)
         endif
         if (ierr > 0) go to 1000
         cerr = 'zhopgap'
         if(lddmopt .or. lddgbtendec)then
            allocate(zhopgap(nmode,ktotog,1),stat=ierr)
         else
            allocate(zhopgap(nmode,1:khtot(1),dbnrec),stat=ierr)
         endif
         if (ierr > 0) go to 1000
      endif

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_GAPPNTMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_gappntmod

!#######################################################################

      subroutine dealloc_gappntmod

      implicit none

      if (allocated(zhpsigap)) deallocate(zhpsigap)
      if (allocated(pmatgap)) deallocate(pmatgap)
      if (allocated(phlochgap)) deallocate(phlochgap)
      if (allocated(zhopgap)) deallocate(zhopgap)

      return
      end subroutine dealloc_gappntmod


!#######################################################################

      subroutine alloc_gapwfmod

use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

      call dealloc_gapwfmod

      cerr = 'wfposn'
      allocate(wfposn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'wfwidth'
      allocate(wfwidth(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_GAPWFMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_gapwfmod

!######################################################################

      subroutine dealloc_gapwfmod

      implicit none

      if (allocated(wfposn)) deallocate(wfposn)
      if (allocated(wfwidth)) deallocate(wfwidth)

      return
      end subroutine dealloc_gapwfmod

!#######################################################################

      subroutine alloc_gaparrmod

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_gaparrmod

      cerr = 'hopsgap'
      allocate(hopsgap(hopsgapdim),stat=ierr)
      if (ierr > 0) go to 1000
      if(.not.lexact)then
         cerr = 'hpsigap'
         allocate(hpsigap(hpsigapdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'hlochgap'
         allocate(hlochgap(hlochgapdim),stat=ierr)
         if (ierr > 0) go to 1000
         cerr = 'hteilgap'
         allocate(hteilgap(pmatgapdim),stat=ierr)
         if (ierr > 0) go to 1000
      endif

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_GAPARRMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_gaparrmod

!#######################################################################

      subroutine dealloc_gaparrmod

      implicit none

      integer(long)     :: ierr

      if (allocated(hopsgap)) deallocate(hopsgap,stat=ierr)
      if (allocated(hpsigap)) deallocate(hpsigap,stat=ierr)
      if (allocated(hlochgap)) deallocate(hlochgap)
      if (allocated(hteilgap)) deallocate(hteilgap)

      return
      end subroutine dealloc_gaparrmod

!#######################################################################
      subroutine alloc_gapmisc

use maxv, only: maxdim        
use channels
      
      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

      call dealloc_gapmisc

      cerr = 'gridmap'
      allocate(gridmap(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_GAPMISC :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_gapmisc

!#############################################################################
      subroutine dealloc_gapmisc

      implicit none

      if (allocated(gridmap)) deallocate(gridmap)

      return
      end subroutine dealloc_gapmisc

!#############################################################################
      subroutine alloc_gapkernel

use maxv, only: maxdim        
use channels
      
      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

      call dealloc_gapkernel

      cerr = 'gapalpha'
      allocate(gapalpha(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gapker'
      allocate(gapker(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gapkerscl'
      allocate(gapkerscl(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

      gapalpha = 0.5_dop
      gapker = 1
      gapkerscl = 0.0_dop

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_GAPKERNEL :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_gapkernel

!#############################################################################
      subroutine dealloc_gapkernel

      implicit none

      if (allocated(gapalpha)) deallocate(gapalpha)
      if (allocated(gapker)) deallocate(gapker)
      if (allocated(gapkerscl)) deallocate(gapkerscl)

      return
      end subroutine dealloc_gapkernel

!#############################################################################
      end module gapmod


