!***********************************************************************
!                                                                     
!              MCTDH module file griddatmod.f90
!                                                                     
!  MCTDH include file defining array sizes,  parameters etc.
!                                                                     
! PARAMETERS DEFINING "COMBINED" GRIDS
! New version (Benjamin, December 2008)
!
!***********************************************************************

      module griddatmod

      use decimal, only : long
      use maxv
      use mltypes, only: wfnode_t

!-----------------------------------------------------------------------
!
! lconm    : calculation contains contracted spf_s
! nmode    : number of spf degrees of freedom
! nmlmode  : in ML, total number of modes in all layers
! meb      : mode including an electronic basis (feb). For a single-state 
!            or multi-set calculation without a continuum, meb = 0 as
!            the electronic basis is not propagated.
! lebgrid  : the electronic DOF includes an SPF basis
! mpb      : mode of the packet basis (fpb)
! mkoe     : "mode" including Hamiltonian constants
! vgdim(f) : number of grid points for dofs preceeding f
! ngdim(f) : number of grid points for dofs following f
! subdim(m): number of grid points for mode m. For GWPs this is the no.
!            of parameters
! supgdim(m): number of grid points for mode m, subdim(m) = supgdim(m)^2, 
!             for densities of type I only
! npg(m)   : no. of GWP parameters (diff. from subdim(m) for PZ method)
! npacket  : number of wavepackets to be propagated simultaneously
! npackts  : number of wavepackets to be propagated simultaneously (single-set)
! npacktot : number of waveppackets (either ss or ms).
! nstate   : number of separate states in calculation. This may not be
!            no. of actual states.
! nddstate : number of actual electronic states (needed in single-state and
!            direct dynamics calculations when nddstate .ne. nstate)
! sgdim    : Max[supgdim(m)] (for densities of type I only)
! sdim     : Max[subdim(m)]
! maxnspfdof : Max[nspfdof(m)]
!-----------------------------------------------------------------------
      logical(kind=4), save                  :: lconm
      integer(long), save                    :: nmode,meb,nstate,sdim,&
                                                sgdim,mkoe,nddstate,mpb,&
                                                nmlmode,maxnspfdof
      integer(long), save                    :: npacket = 1, &
                                                npackts = 1, &
                                                npacktot = 1
      integer(long), dimension(:), allocatable, save :: subdim,supgdim, &
                                                        vgdim,ngdim, &
                                                        npg,nspfdof,dofspf
      integer(long), dimension(:,:), allocatable, save :: spfdof
      logical(kind=4), save                  :: lebgrid=.false.

!-----------------------------------------------------------------------
! GWP information needed to set up no. of parameters (grid length) etc.
! gwpm(m) : if .true. mode is a gwp
! gwptype: 0 = thawed (fully flexible)
!          1 = correlated (fixed diagonal)
!          2 = separable (diagonal)
!          3 = Fixed width (diagonal) DEFAULT
!          4 = Fixed width (not-diagonal)
! gwpphase: -1 = no phase (not propagated) DEFAULT
!           0 = fixed
!           1 = retains normalised GWP 
!           2 = normalised GWP with "1D-action"
!           3 = removes classical terms (cancels X-vector)
! lgwpsep : if .true. GWPs are separable (width matrix stays diagonal)
! lgwpphase : if .false, phase is not stored in psi and is not propagated
!-----------------------------------------------------------------------
      logical(kind=4), dimension(:), allocatable, save :: gwpm
      integer(long), save                       :: gwptype,gwpphase
      logical(kind=4), save                     :: lgwpsep,lgwpphase

!-----------------------------------------------------------------------
! Information on partitioning system (e.g. system - bath)
! lsysshell    : if .true. system is paritioned into "shells"
! isysshell(m) : integer defining shell for mode m (0 = system)
!-----------------------------------------------------------------------
      logical(kind=4), save                          :: lsysshell = .false.
      integer(long), dimension(:), allocatable, save :: isysshell

!-----------------------------------------------------------------------
! HOW ELECTRONIC BASIS IS HANDLED
!
!  leb       : single-set calculation
!  lebonly   : Only an electronic basis (set in einsbas)
!  lmult     : multi-set calculation
!  lmulpack  : two or more wavepackets are propagated simultaneously
!-----------------------------------------------------------------------
      logical(kind=4), save                    :: leb = .false., &
                                                  lmult = .false.,&
                                                  lmulpack = .false.,&
                                                  lebonly = .false.

!-----------------------------------------------------------------------
! MULTILAYER VARIABLES
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save :: tape

      type(wfnode_t), pointer, public, save     :: tree
      type(wfnode_t), pointer, public, save     :: tz
      type(wfnode_t), pointer, public, save     :: treeaux1

      contains

!#######################################################################
!
! GRDDAT_ALLOC
!
! memory required for arrays in griddat allocated here.
!
! First check if allocated, then allocate. All arrays must be deallocated in 
! grddat_dealloc
!
!#######################################################################

      subroutine alloc_grddat

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr
      integer(long), external :: integersize, logicalsize

! check all arrays are not allocated
      call dealloc_grddat

      cerr = 'subdim'
      allocate(subdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'supgdim'
      allocate(supgdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'vgdim'
      allocate(vgdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'ngdim'
      allocate(ngdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'npg'
      allocate(npg(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'nspfdof'
      allocate(nspfdof(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dofspf'
      allocate(dofspf(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'spfdof'
      allocate(spfdof(maxdim,maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'gwpm'
      allocate(gwpm(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*logicalsize()
      if (ierr > 0) go to 1000
      cerr = 'tape'
      allocate(tape(maxtape),stat=ierr)
      allocmemory=allocmemory + maxtape*integersize()
      if (ierr > 0) go to 1000
      cerr = 'isysshell'
      allocate(isysshell(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000

! test
      gwpm(:) = .false.
! test

      return
 1000 write(6,'(a)') 'Error allocating memory in ALLOC_GRDDAT :',cerr
      write(6,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_grddat

!#######################################################################

      subroutine dealloc_grddat

      if (allocated(subdim)) deallocate(subdim)
      if (allocated(supgdim)) deallocate(supgdim)
      if (allocated(vgdim)) deallocate(vgdim)
      if (allocated(ngdim)) deallocate(ngdim)
      if (allocated(npg)) deallocate(npg)
      if (allocated(nspfdof)) deallocate(nspfdof)
      if (allocated(dofspf)) deallocate(dofspf)
      if (allocated(spfdof)) deallocate(spfdof)
      if (allocated(gwpm)) deallocate(gwpm)
      if (allocated(tape)) deallocate(tape)
      if (allocated(isysshell)) deallocate(isysshell)

      return
      end subroutine dealloc_grddat

      end module griddatmod
