!#######################################################################
!                      HPSI.F90
!#######################################################################
      module hpsimod

use decimal, only : long, dop
use maxv
use griddatmod, only: nmode
use channels

      implicit none

!-----------------------------------------------------------------------
! ARRAY SIZES
!
! pmatdim  : Length of hteil and hloch arrays containing <phi|h|phi>
!            matrices
!            pmatdim = Sum[dim(m)*dim(m)*kzahl]
! hpsidim  : Length of hpsi array.
!            The actual size depends on the calculation mode.
! hlochdim : Length of array containing mean-field.
!            The actual size depends on the calculation mode.
! ghmatdim : length of array gh containing Hamiltonian elements
!            for Gaussian mode.
!            [ghmatdim = khtot(1)*subdim(m)*dim(m,s)*dim(m,s1)]
! hunGdim  : length of array hunG, which sums up contributions
!            from uncorrelated Hamiltonian elements for Gaussian
!            mode. [hunGdim = [dim(m,s)*dim(m,s)*nstate]                   
! ghder1dim : length of array containing derivatives for GWP terms
!             [ghder1dim = gdof*maxspf*nmode*maxkoe]
!-----------------------------------------------------------------------
      integer(long), save :: hpsidim,hlochdim
      integer(long), save :: pmatdim,ghmatdim,hunGdim,ghder1dim

!-----------------------------------------------------------------------
! c3matdim : Length of cdvr3 array.
!            c3matdim = Sum[nspfdof(m)*dim(m)]
! c3mat(m,s) : pointer to start of cdvr3 matrices for mode m, state s
!-----------------------------------------------------------------------
      integer(long), save :: c3matdim
      integer(long), dimension(:,:), allocatable, save :: c3mat

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! pmat(m,k)  : pointer to <phi(i)|h|phi(j)> matrices for the mth mode
!              and kth Hamiltonian coefficient in "HTeil"
! pmat2(m,k) : additional pointer, important for dissipative operators
! phloch(m,k): Pointer to <phi(i)|h|phi(j)> matrices for mth mode and
!              kth Hamiltonian term in "hloch"
! zhpsi(m,k) : pointer to start of h|psi> vector
! zhpsi2(m,k): additional pointer, important for dissipative operators
! ghmat(k,m)   : GWP Hamiltonian matrix elements at fixed k and mode m
! hunGmat(m,s,n) : GWP Hamiltonian uncorrelated matrix elements at 
!                mode m, state s and Hamiltonian n
! ghder1(k,m)  : GWP Hamiltonian first derivatives at fixed k and mode m
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save   :: zhpsi,zhpsi2
      integer(long), dimension(:,:), allocatable, save   :: pmat,pmat2,phloch
      integer(long), dimension(:,:), allocatable, save   :: ghmat,ghder1
      integer(long), dimension(:,:,:), allocatable, save :: hunGmat

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS (density operators of type I)
!
! dpmat(m,k,t) : pointer to <phi(i)|h|phi(j)> matrices for the mth mode
!                and kth Hamiltonian coefficient in "HTeil"
! dphloch(m,k,t): Pointer to <phi(i)|h|phi(j)> matrices for mth mode and
!                 kth Hamiltonian term in "hloch"
! dzhpsi(m,k,s) : pointer to start of h|psi> vector
!-----------------------------------------------------------------------
      integer(long), dimension(:,:,:), allocatable, save :: dzhpsi,dpmat, &
                                                              dphloch
!-----------------------------------------------------------------------
! PARAMETERS DEFINING HOW GWP INTEGRALS ARE FORMED
!
! gwpcxm(m): 0   use full C^-1 Y       (no derivatives needed)
!            1   use X + C^-1 Y_corr   (X = dV/dx, add to all)
!            2   use X + C^-1 Y_corr   (X = dV/dx, add to coupled only)
!
! gwpintordm(m): order used for GWP integrals for mode m
! gwpintm(m): integration used for GWP potential matrix elements
!       0: Exact or unsymmetric LHA (from operator type)
!       1: unsymmetric LHA (expand on RHS GWP)
!       2: symmetrised LHA (expand on RHS GWP then make Hermitian)
!       3: symmetric LHA (expand on centre of overlap)
!       4: GPR Gaussian potential
! dermethod: method used for numerical first derivatives
!            0 simple
!            1 using Ridder extrapolation
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save :: gwpcxm
      integer(long), dimension(:), allocatable, save :: gwpintordm
      integer(long), dimension(:), allocatable, save :: gwpintm
      integer(long), save :: dermethod = 1
      real(dop), save     :: fddx = -1.0

      contains

!#######################################################################

      subroutine alloc_hpsimod

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_hpsimod

      cerr = 'zhpsi'
      allocate(zhpsi(maxdim,maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zhpsi2'
      allocate(zhpsi2(maxdim,maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'pmat'
      allocate(pmat(maxdim,maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'pmat2'
      allocate(pmat2(maxdim,maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'phloch'
      allocate(phloch(maxdim,maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      phloch=1
      cerr = 'ghmat'
      allocate(ghmat(maxkoe,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ghder1'
      allocate(ghder1(maxkoe,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'hungmat'
      allocate(hungmat(maxdim,maxkoe,maxham),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dzhpsi'
      allocate(dzhpsi(maxdim,maxkoe,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dpmat'
      allocate(dpmat(maxdim,maxkoe,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dphloch'
      allocate(dphloch(maxdim,maxkoe,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'c3mat'
      allocate(c3mat(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gwpcxm'
      allocate(gwpcxm(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gwpintordm'
      allocate(gwpintordm(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gwpintm'
      allocate(gwpintm(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

! Defaults
      gwpcxm(:) = 1
      gwpintordm(:) = 2
      gwpintm(:) = 1

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_HPSIMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine alloc_hpsimod

!#######################################################################

      subroutine dealloc_hpsimod

      implicit none

      if (allocated(zhpsi)) deallocate(zhpsi)
      if (allocated(zhpsi2)) deallocate(zhpsi2)
      if (allocated(pmat)) deallocate(pmat)
      if (allocated(pmat2)) deallocate(pmat2)
      if (allocated(phloch)) deallocate(phloch)
      if (allocated(ghmat)) deallocate(ghmat)
      if (allocated(ghder1)) deallocate(ghder1)
      if (allocated(hungmat)) deallocate(hungmat)
      if (allocated(dzhpsi)) deallocate(dzhpsi)
      if (allocated(dpmat)) deallocate(dpmat)
      if (allocated(dphloch)) deallocate(dphloch)
      if (allocated(c3mat)) deallocate(c3mat)
      if (allocated(gwpcxm)) deallocate(gwpcxm)
      if (allocated(gwpintordm)) deallocate(gwpintordm)
      if (allocated(gwpintm)) deallocate(gwpintm)

      return
      end subroutine dealloc_hpsimod

      end module hpsimod

