module mltypes

! Definition of types used in ML-MCTDH

use decimal

!------------------------------------------------------------------------------
! T Y P E    W F N O D E
!------------------------------------------------------------------------------
! Every node of the wavefunction tree is a variable of type(wfnode_t).
! Each node holds all information necessary for the propagation of its
! coeffcients and is the basic and most important structure in this
! implementation of the ML-MCTDH algorithm.
!------------------------------------------------------------------------------
    type, public :: wfnode_t

        !----------------------------------------------------------------------
        integer          :: nmodes  ! number of submodes (or primitive coords)
        integer, pointer :: ndim(:) ! number of basis functions in each submode
        integer          :: nsubcoord   ! mode number w.r.t. parent node
        integer          :: navecs      ! number of spfs (avecs)
        integer          :: m           ! if primitive, mode number, 0 otherwise
        integer          :: num         ! mode number (preorder traversal)
        integer          :: layer       ! layer of this node
        integer, pointer :: degf(:)     ! if primitive, the number of each degf
        logical          :: isprimitive
        logical          :: istop
!       integer          :: symmetry    ! 0: no sym; 1 bosonic; -1 fermionic 
!                                       !!! unused
        ! -- FO: corresponds to the following quantities in my notes:
        ! this node   = z
        ! nmodes      = p_z
        ! ndim(kappa) = n_{z,kappa}
        ! nsubcoord   = nu , in z=(z',nu) where z'=parent(z)
        ! navecs      = n_z
        ! degf        = ( f | q_f in Q_z ) but only for primitive z
        !----------------------------------------------------------------------
        integer,pointer           :: z(:,:) ! absolute position of SPFs in psi
        integer,pointer           :: zr(:,:)! relative position of SPFs in psi
        complex(cdbl),allocatable :: phiaux(:)    ! auxiliar phi vector
        integer,allocatable       :: ndimaux(:)   ! auxiliar ndim
        !----------------------------------------------------------------------
        ! hterms(k,m)  : hamiltonian matrix for k-th kterm, m-th mode of this 
        !                node
        ! isunit(k,m)  : true if hterms(k,m) is a unit matrix
        ! allunit(k)   : true if all hterms(k,m) are unit matrices
        ! isuncor(k)   : true if k-th kterm is not correlated with outside nodes
        ! htmtype(k)   : if primitive, type of overall op.term prod_m 
        !                hterms(k,m)
        ! mfields(k,m) : mean-field matrix, indices same as above
        ! dmat(m)      : density matrix of m-th mode in this node
        ! invdmat(m)   : inverse rho of m-th mode
        ! pop(m)       : eigenvalues of rho for m-th mode; grid populations 
        !                if primitive
        ! regpop(m)    : regularized eigenvalues of rho for m-th mode
        ! trfmat(m)    : eigenvectors of dmat(m)
        ! overlap(m)   : overlap matrices used in the product of two 
        !                wavefunctions
        ! mfsum(a,b)   : sum over mean-field matrices x potential operators 
        !                (for CMF)
        ! mtrace(m)    : traces of SPDOs used in analysis of density 
        !                matrix (rho-MCTDH)
        ! shtrace(m)   : single hole traces of SPDOs used in analysis of density 
        !                matrix (rho-MCTDH)
        type(hterm_tp), pointer    :: hterms(:,:) => null()
        logical,pointer            :: isunit(:,:) => null()
        logical,pointer            :: allunit(:)  => null()
        logical,pointer            :: isuncor(:)  => null()
        logical,pointer            :: hasuncor(:,:) => null()
        integer,pointer            :: htmtype(:)  => null()
        type(hterm_tp), pointer    :: mfields(:,:)=> null()
        type(hterm_tp), pointer    :: dmat(:)     => null()
        type(hterm_tp), pointer    :: invdmat(:)  => null()
        type(hterm_tp), pointer    :: pop(:)      => null()
        type(hterm_tp), pointer    :: regpop(:)   => null()
        type(hterm_tp), pointer    :: trfmat(:)   => null()
        type(hterm_tp), pointer    :: overlap(:)  => null()
        type(hterm_tp), pointer    :: mfsum(:,:)  => null()
        type(hterm_tp), pointer    :: mtraces(:)  => null()
        type(hterm_tp), pointer    :: shtraces(:) => null()
        ! -- FO: corresponds to the following quantities in my notes:
        !    [ab] signifies a matrix element, i.e. ~ hterm_t%matrix(a,b)
        ! hterms(r,kappa)[ij]  = hterm^{z,kappa}_{r,ij}
        ! isunit(r,kappa)      = ( hterm^{z,kappa}_{r,ij} == delta_{ij} ) 
        !                         i.e. (\hat{h}^{z,kappa}_r == 1)
        ! allunit(r)           = AND_{kappa=1}^{p_z} isunit(r,kappa) , 
        !                         i.e. (\hat{h}^z_r == 1)
        ! isuncor(r)           = ( \hat{mathcal{H}}^z_r == 1 )
        ! mfields(r,kappa)[ab] = mfterm^{z,kappa}_{r,ab}
        ! dmat(kappa)[ab]      = rho^{z,kappa}_{ab}
        ! invdmat(kappa)[ab]   = {(rho^{z,kappa})^{-1})}_{ab}
        ! overlap(kappa)[ab]   = < phi^{z,kappa}_a | phi^{z,kappa}_b >
        ! trace(kappa)[a]      = Tr(sigma^{z,kappa}_a)
        !----------------------------------------------------------------------
        ! pointers to children and parent nodes, if some
        type(wfnode_tp), pointer   :: submodes(:) => null() ! pointers to 
                                                            ! submodes
        type(wfnode_t), pointer    :: parent      => null() ! pointer to 
                                                            ! parent
        type(wfnode_t), pointer    :: topnode     => null() ! pointer to top 
                                                            ! node
        !----------------------------------------------------------------------
        ! timer IDs for node-specific routines
        integer :: tid_rhs      ! compute_rhs_node
        integer :: tid_mfields  ! compute_mfields_node
        integer :: tid_hterms   ! compute_hterms_node
        integer :: tid_dmat     ! compute_dmat_node
        integer :: tid_opsi     ! compute_opsi_node
        integer :: tid_mfsum    ! compute_mfsum_node
        integer :: tid_uterms   ! compute_uterms_node
        integer :: tid_wterms   ! compute_wterms_node
        !----------------------------------------------------------------------
        ! integrator settings (for CMF split-mode propagation)
        type(integrator_t),pointer :: integrator => null()
        !----------------------------------------------------------------------
        ! The following components will ONLY be defined for the top node. Access
        ! to them in another node will result in unexpected behaviour or the
        ! program crashing...
        ! TODO: turn this part into a tree_t
        integer                  :: tot_nodes   ! number of nodes in tree
        integer                  :: tot_pnodes  ! number of primitive nodes 
                                                ! in tree
        integer                  :: tot_degfs   ! number of degrees of freedom
        integer,allocatable      :: pdim(:)     ! primitive dim, ndim of 
                                                ! primitive nodes
        type(hterm_tp),pointer   :: pgpops(:)   ! populations of primitive 
                                                ! degfs
        integer,pointer          :: tape(:)     => null() ! integer code of
                                                          ! tree struct.
        type(wfnode_tp), pointer :: swap_down(:)=> null()
        type(wfnode_tp), pointer :: swap_up(:)  => null()
        type(wfnode_tp), pointer :: swap_phi(:) => null()
        type(wfnode_tp), pointer :: swap_levelorder(:) => null()
        type(wfnode_tp), pointer :: swap_preorder(:)   => null()
        type(wfnode_tp), pointer :: swap_postorder(:)  => null()
        complex(cdbl),pointer    :: workc(:)    => null() ! complex work array
        real(dbl),pointer        :: workr(:)    => null() ! real work array
        complex(cdbl),pointer    :: workc2(:,:) => null() ! 2D complex work 
                                                          ! array
        integer                  :: workcdim    ! size of workc array
        integer                  :: workcdim2   ! size of workc2 array
        integer                  :: conf_proj = 1       ! projector type
        complex(cdbl)            :: prefactor = -(0.0_dbl, 1.0_dbl)
        logical                  :: isrelax   = .false. ! true if imag. time 
        logical                  :: isrlxt    = .false. ! true if thermal relaxation
        integer                  :: cmftyp    = 0       ! type of CMF 
                                                        ! integration
                                                        ! (0=VMF,
                                                        !  1=CMF/unified,
                                                        !  2=CMF/split)
        logical                  :: isallocated = .false.
        logical                  :: ishallocated = .false.
    end type wfnode_t
!------------------------------------------------------------------------------

    type, public :: wfnode_tp
       type(wfnode_t), pointer :: p => null()
    end type wfnode_tp

!------------------------------------------------------------------------------
!
! ------ hterm_t types (variable %htype) -------
!    1: unit term
!    2: matrix term             (complex)
!    3: diagonal (local) term   (complex)
!    4: matrix term             (real)
!    5: diagonal (local) term   (real)
! ----------------------------------------------
    type, public :: hterm_t
        integer                 :: htype   ! type code of the operator
        integer                 :: hrange  ! size (matrix range) of the operator
        integer                 :: hrange2 ! second size, if not square
        complex(cdbl), pointer  :: matrix(:,:)  => null()
        real(dbl),pointer       :: rmatrix(:,:) => null()
        complex(cdbl), pointer  :: vector(:)    => null()
        real(dbl),pointer       :: rvector(:)   => null()
    end type hterm_t

!------------------------------------------------------------------------------

    type, public :: hterm_tp
        type(hterm_t), pointer  :: p
    end type hterm_tp

!------------------------------------------------------------------------------

   type,public :: integrator_t
   ! This type stores settings and internal data for the various
   ! integration routines.  Such data should not be global, because
   ! different objects in the program can then be propagated with
   ! individual integrator settings.
      integer :: inttyp   ! integrator type (see above)
      logical :: llog     ! .true. if logging is enabled
      logical :: lrhon    ! .true. if density-weighted error-norm is enabled
      integer :: ilog     ! log channel
      logical :: lpost    ! .true. if post-processing is enabled
      integer :: ipost    ! cf procedure int_postprocess
      ! pointers to integrator-specific data, only one of these will be set
!     type(int_abm_t),pointer :: abm => null()
!     type(int_bs_t),pointer  :: bs  => null()
      type(int_rk_t),pointer  :: rk  => null()
      type(int_sil_t),pointer :: sil => null()
      type(int_csil_t),pointer:: csil=> null()
!     type(int_dav_t),pointer :: dav => null()
   end type integrator_t

   type,public :: int_rk_t
      integer   :: order       ! order
      real(dbl) :: accuracy    ! target accuracy
      real(dbl) :: initstep    ! initial step size
      real(dbl) :: laststep    ! last good/optimum step size
      ! statistical data
      integer   :: nsteps      ! total number of steps
      integer   :: naccpt      ! number of accepted steps
      integer   :: nrejct      ! number of rejected steps
      integer   :: nfncns      ! number of RHS function calls
      ! auxiliary vectors
      complex(cdbl),allocatable :: auxpsi(:,:)
   end type int_rk_t

   type,public :: int_sil_t
      integer       :: maxOrder    ! maximum order allowed
      real(dbl)     :: accuracy    ! target accuracy
      complex(cdbl) :: PreFactor   ! function f is actually f = prefactor*H*y
      ! data kept between calls
      integer       :: prevOrder   ! order used in previous step
      real(dbl)     :: prevStep    ! interval integrated in previous step
      real(dbl)     :: prevError   ! error measured in previous step
      ! statistical data
      integer       :: nsteps      ! total number of steps
      integer       :: ngoodst     ! number of good steps
      integer       :: nbadst      ! number of bad steps (order too low)
      integer       :: nfncns      ! number of RHS function calls
      integer       :: realOrder   ! largest order actually used
      ! auxiliary vectors
      complex(cdbl),allocatable :: Krylov(:,:)    ! storage for the Krylov space
      real(dbl),allocatable     :: EigenVector(:,:)
      real(dbl),allocatable     :: Diagonal(:)
      real(dbl),allocatable     :: OffDiag(:)
      real(dbl),allocatable     :: EigenVal(:)
      real(dbl),allocatable     :: OffDg2(:)
   end type int_sil_t

   type,public :: int_csil_t
      integer       :: maxOrder    ! maximum order allowed
      real(dbl)     :: accuracy    ! target accuracy
      complex(cdbl) :: PreFactor   ! function f is actually f = prefactor*H*y
      ! data kept between calls
      integer       :: prevOrder   ! order used in previous step
      real(dbl)     :: prevStep    ! interval integrated in previous step
      real(dbl)     :: prevError   ! error measured in previous step
      ! statistical data
      integer       :: nsteps      ! total number of steps
      integer       :: ngoodst     ! number of good steps
      integer       :: nbadst      ! number of bad steps (order too low)
      integer       :: nfncns      ! number of RHS function calls
      integer       :: realOrder   ! largest order actually used
      ! auxiliary vectors
      complex(cdbl),allocatable :: Krylov(:,:)    ! storage for the Krylov space
      complex(cdbl),allocatable :: Hessenberg(:,:)
      complex(cdbl),allocatable :: EigVec(:,:)
      complex(cdbl),allocatable :: AuxMat(:,:)
      complex(cdbl),allocatable :: EigVal(:)
      complex(cdbl),allocatable :: Aux1(:)
      complex(cdbl),allocatable :: Aux2(:)
      complex(cdbl),allocatable :: Aux3(:)
      complex(cdbl),allocatable :: OldAux1(:)
   end type int_csil_t

!-----------------------------------------------------------------------
! TYPES NEEDED FOR LISTS OF INTEGERS (ILIST, LLIST) 
!-----------------------------------------------------------------------
type :: ilist_t
   integer         :: n = 0          ! length of the list
   integer,pointer :: i(:) => null() ! the list of integers
end type ilist_t

type :: ilist_tp
   type(ilist_t),pointer :: p => null()
end type ilist_tp

type :: llist_node_t
   type(llist_node_t),pointer :: next => null()
   type(llist_node_t),pointer :: prev => null()
   integer                    :: val
end type llist_node_t


type :: llist_int_t
   type(llist_node_t),pointer :: first => null()
   type(llist_node_t),pointer :: last  => null()
   integer                    :: sz  = 0
end type llist_int_t

type :: llist_int_tp
   type(llist_int_t),pointer  :: p => null()
end type llist_int_tp

!-----------------------------------------------------------------------
! TYPES NEEDED FOR TENSORS IN MLOPs
!-----------------------------------------------------------------------
type :: tensor_t
   integer               :: ndim              ! number of dimensions (indices)
   integer,pointer       :: dims(:) => null() ! extent of each dimension
   integer               :: vlen              ! product(dims)
   real(dbl),pointer     :: vr(:) => null()   ! tensor data (real)
end type tensor_t

type :: tensor_tp
   type(tensor_t),pointer :: p => null()
end type tensor_tp

!-----------------------------------------------------------------------
! TYPES NEEDED FOR MLOPs
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
type :: opnode_t
!-----------------------------------------------------------------------
! This contains a list of MLOps. Two kinds are possible:
! isprimitive: The MLOps apply directly to a primitive MCTDH mode.
!    In order to reuse the MCTDH machinery for applying mode-operators
!    to mode-wavefunctions (hcall), we construct (during genoper) a
!    Hamiltonian line (k-term) for each mode-operator. Here we just keep
!    references to these lines.
!    NB: These k-terms are not used in any actual Hamiltonian.
! .not.isprimitive: The MLOps are expressed according to above equation.
!-----------------------------------------------------------------------
   !--- Local node-related data ---
   logical                 :: isprimitive         ! see above
   integer                 :: nops = 0            ! number of MLOps in the list (m_z)
   type(opnode_t),pointer  :: parent => null()    ! pointer to parent node
   !--- Tree-related data ---
   type(optree_t),pointer  :: tree => null()      ! pointer to tree that this node belongs to
   integer                 :: num                 ! internal number of this node
   integer                 :: postnum             ! post-order number of this node
   integer                 :: layer               ! layer (level) of this node inside the tree
   !--- Data for primitive MLOps ---
   integer                 :: m = 0               ! MCTDH mode number
   integer,pointer         :: ks(:) => null()     ! list of Hamiltonian lines, dim=(nops)
   integer,pointer         :: ktypes(:) => null() ! type of k-term for this node (unit, diag, ...)
   !--- Data for hierarchical MLOps ---
   integer                 :: nmodes = 0          ! number of children (p_z)
   type(opnode_tp),pointer :: subops(:) => null() ! pointers to child operators, dim=(nmodes)
   type(tensor_t)          :: tt                  ! transfer tensor
   !--- Data available through tt ---
   !  ndim = nmodes + 1
   !  dims(f) = number of MLOps at the f-th child (f = 1...nmodes)
   !  dims(nmodes+1) = nops
   !  vr -> transfer tensor data
   !--- Data related to wavefunction ---
   type(hterm_tp),pointer  :: uterms(:) => null() ! uterms(c)[i,j] = <phi_i|U_c|phi_j>
   type(hterm_tp),pointer  :: wterms(:) => null() ! wterms(c)[i,j] = <psihole_i|W_c|psihole_j>
   complex(cdbl),pointer   :: vw(:,:) => null()   ! Sum_c W(c,::)*V(c,::)
   logical,pointer         :: early_utm_child(:) => null()
   logical,pointer         :: early_wtm_sibling(:) => null()
      ! .true. if the uterms of the respective child/sibling should be applied
      ! "early" when computing the uterms/wterms/RHS.
end type opnode_t

!-----------------------------------------------------------------------
type :: opnode_tp
!-----------------------------------------------------------------------
   type(opnode_t),pointer :: p => null()
end type opnode_tp

!-----------------------------------------------------------------------
type :: optree_t
!-----------------------------------------------------------------------
! This contains information about the total structure of the tree.
! Mainly used for iterating over the nodes in some way.
!-----------------------------------------------------------------------
   type(opnode_t),pointer  :: topnode => null()      ! the top node
   integer                 :: numnodes               ! total number of nodes in tree
   integer                 :: numleaves              ! total number of leaves in tree
   integer                 :: numlayers              ! number of layers in tree
   type(opnode_tp),pointer :: preorder(:) => null()  ! nodes in pre-order
   type(opnode_tp),pointer :: postorder(:) => null() ! nodes in post-order
   type(opnode_tp),pointer :: leaves(:) => null()    ! leaf nodes
   logical                 :: isallocated = .false.  ! uterms/wterms allocated or not
end type optree_t

!-----------------------------------------------------------------------
type :: optree_tp
!-----------------------------------------------------------------------
   type(optree_t),pointer :: p => null()
end type optree_tp


!-----------------------------------------------------------------------
   type :: vnode_t
!-----------------------------------------------------------------------
      !--- Local node-related data ---
      logical                :: isleaf               ! .true. => children are dofs
      integer                :: nmodes               ! number of submodes/dofs
      type(vnode_tp),pointer :: modes(:) => null()   ! pointers to submodes
      integer,pointer        :: dofnums(:) => null() ! list of dofs
      type(vnode_t),pointer  :: parent   => null()   ! pointer to parent node
      !--- Tree-related data ---
      type(vtree_t),pointer  :: tree => null()       ! pointer to tree that this node belongs to
      integer                :: num                  ! internal number of this node
      integer                :: postnum              ! post-order number of this node
      integer                :: layer                ! layer (level) of this node inside the tree
      !--- MLPF-related data ----
      integer,pointer        :: ndim(:) => null()    ! number of basis tensors/grid points of the children
      integer                :: plen                 ! product(ndim)
      integer                :: nbasis               ! actual number of basis tensors at this node
      integer                :: maxnbasis = 0        ! maximum number of basis tensors at this node
      real(dbl),pointer      :: wghts(:) => null()   ! basis weights, dim=(nbasis)
      real(dbl),pointer      :: basis(:,:) => null() ! basis tensors, dim=(plen,nbasis)
   end type vnode_t

!-----------------------------------------------------------------------
   type :: vnode_tp
!-----------------------------------------------------------------------
      type(vnode_t),pointer  :: p => null()
   end type vnode_tp


!-----------------------------------------------------------------------
   type :: vtree_t
!-----------------------------------------------------------------------
      type(vnode_t),pointer  :: topnode => null()      ! the top node
      integer                :: numnodes               ! total number of nodes in tree
      integer                :: numdofs                ! total number of DOFs in tree
      integer                :: numleaves              ! total number of leaves in tree
      integer                :: numlayers              ! number of layers in tree
      type(vnode_tp),pointer :: preorder(:) => null()  ! nodes in pre-order
      type(vnode_tp),pointer :: postorder(:) => null() ! nodes in post-order
      type(vnode_tp),pointer :: leaves(:) => null()    ! leaf nodes
   end type vtree_t


!-----------------------------------------------------------------------
   type :: vtree_tp
!-----------------------------------------------------------------------
      type(vtree_t),pointer  :: p => null()
   end type vtree_tp

!-----------------------------------------------------------------------
type :: mlpf_t
!-----------------------------------------------------------------------
! This type holds information about a Multi-Layer Potfit (MLPF)
! and its relation to the system, namely how the MLPF DOFs map to
! the system DOFs.
!-----------------------------------------------------------------------
   type(vtree_t),pointer :: vtree => null()
   type(ilist_t),pointer :: fmap  => null()
end type mlpf_t

integer,parameter :: maxnmlpf = 64 ! cf. maxrdf
type(mlpf_t),save :: mlpfs(maxnmlpf)


!-----------------------------------------------------------------------
type :: meta_mlop_t
!-----------------------------------------------------------------------
! This type represents either
! - an actual MLOp, plus the MLPF it was made from (if applicable); or
! - a list of MLOps that will have to be multiplied out to generate a
!   total MLOp. That list is stored as a queue of integer indices into
!   the following array _mmlops_.
!-----------------------------------------------------------------------
   type(optree_t),pointer    :: mlop   => null()
   integer                   :: mlpfnr = 0
   type(llist_int_t),pointer :: ifacq  => null()
end type meta_mlop_t

!-----------------------------------------------------------------------
type,public :: mlham_t
!-----------------------------------------------------------------------
! This type holds all the total MLOps that are part of a specific
! "Hamiltonian". For convenience, the k-term which holds the coefficent/
! time information is also stored.
! As intermediate information, we keep a queue of integer indices into
! _mmlops_ (see genmlham_m)for those total MLOps that have yet to be
! created.
!-----------------------------------------------------------------------
   integer                   :: nmlop    = 0
   type(optree_tp),pointer   :: mlops(:) => null()
   integer,pointer           :: kkoes(:) => null()
   type(llist_int_t),pointer :: immlopq  => null()
end type mlham_t

!-----------------------------------------------------------------------
type,public :: mlham_tp
!-----------------------------------------------------------------------
   type(mlham_t),pointer :: p => null()
end type mlham_tp


  end module mltypes
