!***********************************************************************
!               MPISTRUCTURE.F90
!***********************************************************************

module mpistructure

#ifdef MPI

  use decimal, only : dop, long
  use global,  only : c1,c3,c5
  use mpi
  use mpidata, only : mpierr
  implicit none

  !--------------------------------------------------------------------------
  ! Metadata for QCalcData derived type recording the number of types
  ! and the counts of each. These must be update if members of QCalcData
  ! are changed.
  !--------------------------------------------------------------------------
  integer(long), parameter :: DATA_N_TYPES=4,DATA_N_INTS=19,&
       DATA_N_REALS=2,DATA_N_LOGICALS=4,DATA_N_CHARS_C3=2,&
       DATA_N_CHARS_C5=1
  
  !--------------------------------------------------------------------------
  ! Derived datatype used for MPI communications requesting a quantum
  ! chemistry calculation from a worker process. Contains all system
  ! state information stored as non-allocatable variables from master
  ! process. Data stored in allocatable arrays is communicated
  ! seperately.
  !--------------------------------------------------------------------------
  type :: QCalcData
    integer(long) :: totdim1,s1,hl,ngp,mxmpi,ndofddpes,nddstate&
         &,nsymop,dbnrec,ddlaenge,dbdlength,nroot,maxdim &
         &,qcsubform,nbasis,maxhpar,maxhop,lhaorder,dddiab
     real(dop) :: time,dde0
     logical(kind=4) :: lpestest,lupdhes,ldbsave,lddrddb,lqcsaveall
     character(len=c3) :: ddextcmd, fchkcmd
     character(len=c5) :: ddname
  end type QCalcData

contains
  
  pure function init_qcalc_data(totdim1,s1,hl,ngp,mxmpi,ndofddpes, &
       nddstate,nsymop,dbnrec,ddlaenge,dbdlength,nroot,maxdim, &
       qcsubform,nbasis,maxhpar,maxhop,lhaorder,dddiab,time,dde0, &
       lpestest,lupdhes,ldbsave,lddrddb,lqcsaveall,ddextcmd,fchkcmd,ddname)

    !--------------------------------------------------------------------------
    ! Initialise all values of a QCalcData derived type instance
    !--------------------------------------------------------------------------
 
    integer(long), intent(in) :: totdim1,s1,hl,ngp,mxmpi,ndofddpes &
         &,nddstate,nsymop,dbnrec,ddlaenge,dbdlength,nroot,maxdim &
         &,qcsubform,nbasis,maxhpar,maxhop,lhaorder,dddiab
    real(dop), intent(in) :: time,dde0
    logical(kind=4), intent(in) :: lpestest,lupdhes,ldbsave,lddrddb,lqcsaveall
    character(len=c3), intent(in) :: ddextcmd, fchkcmd
    character(len=c5), intent(in) :: ddname
    type(QCalcData) :: data,init_qcalc_data
    
    data%totdim1 = totdim1
    data%s1 = s1
    data%hl = hl
    data%ngp = ngp
    data%mxmpi = mxmpi
    data%ndofddpes = ndofddpes
    data%nddstate = nddstate
    data%nsymop = nsymop
    data%dbnrec = dbnrec
    data%ddlaenge = ddlaenge
    data%dbdlength = dbdlength
    data%nroot = nroot
    data%maxdim = maxdim
    data%qcsubform = qcsubform
    data%nbasis = nbasis
    data%maxhpar = maxhpar
    data%maxhop = maxhop
    data%lhaorder = lhaorder
    data%dddiab = dddiab
    data%time = time
    data%dde0 = dde0
    data%lpestest = lpestest
    data%lupdhes = lupdhes
    data%ldbsave = ldbsave
    data%lddrddb = lddrddb
    data%lqcsaveall = lqcsaveall
    data%ddextcmd = ddextcmd
    data%fchkcmd = fchkcmd
    data%ddname = ddname
    
    init_qcalc_data = data
  end function init_qcalc_data
  
  subroutine unpack_qcalc_data(data,ndofddpes&
       &,nddstate,ddlaenge,dbdlength,nroot,maxdim,qcsubform,nbasis&
       &,maxhpar,maxhop,lhaorder,dddiab,time,dde0,lpestest,lupdhes&
       &,ldbsave,lddrddb,lqcsaveall,ddextcmd,fchkcmd,ddname)

    !--------------------------------------------------------------------------
    ! Set the value of global variables for worker processes from a
    ! QCalcData instance
    !--------------------------------------------------------------------------
    integer(long), intent(out) :: ndofddpes,nddstate,ddlaenge &
         &,dbdlength,nroot,maxdim,qcsubform,nbasis,maxhpar,maxhop &
         &,lhaorder,dddiab
    real(dop), intent(out) :: time,dde0
    logical(kind=4), intent(out) :: lpestest,lupdhes,ldbsave &
         &,lddrddb,lqcsaveall
    character(len=c3), intent(out) :: ddextcmd, fchkcmd
    character(len=c5), intent(out) :: ddname
    type(QCalcData), intent(in) :: data

    time = data%time
    dde0 = data%dde0
    ndofddpes = data%ndofddpes
    nddstate = data%nddstate
    ddlaenge = data%ddlaenge
    dbdlength = data%dbdlength
    nroot = data%nroot
    maxdim = data%maxdim
    qcsubform = data%qcsubform
    nbasis = data%nbasis
    maxhpar = data%maxhpar
    maxhop = data%maxhop
    lhaorder = data%lhaorder
    dddiab = data%dddiab
    lpestest = data%lpestest
    lupdhes = data%lupdhes
    ldbsave = data%ldbsave
    lddrddb = data%lddrddb
    lqcsaveall = data%lqcsaveall
    ddextcmd = data%ddextcmd
    fchkcmd = data%fchkcmd
    ddname = data%ddname

  end subroutine unpack_qcalc_data
  
  function get_qcalc_mpi_type(qcalc_data)
    !--------------------------------------------------------------------------
    ! Return a new registered MPI type for the QCalcData derived type
    !--------------------------------------------------------------------------
    type(QCalcData)                                 :: qcalc_data
    integer                                         :: get_qcalc_mpi_type, qcalc_type
    integer                                         :: block_lengths(DATA_N_TYPES), types(DATA_N_TYPES)
    integer(kind=MPI_ADDRESS_KIND)                  :: memory_displacements(DATA_N_TYPES)

    call MPI_GET_ADDRESS(qcalc_data%totdim1, memory_displacements(1), mpierr)
    call MPI_GET_ADDRESS(qcalc_data%time, memory_displacements(2), mpierr)
    call MPI_GET_ADDRESS(qcalc_data%lpestest, memory_displacements(3), mpierr)
    call MPI_GET_ADDRESS(qcalc_data%ddextcmd, memory_displacements(4), mpierr)

    memory_displacements(4) = memory_displacements(4) - memory_displacements(1)
    memory_displacements(3) = memory_displacements(3) - memory_displacements(1)
    memory_displacements(2) = memory_displacements(2) - memory_displacements(1)
    memory_displacements(1) = 0
    
    block_lengths(1) = DATA_N_INTS
    block_lengths(2) = DATA_N_REALS
    block_lengths(3) = DATA_N_LOGICALS
    block_lengths(4) = DATA_N_CHARS_C3 * c3 + DATA_N_CHARS_C5 * c5

    types(1) = MPI_INTEGER
    types(2) = MPI_DOUBLE_PRECISION
    types(3) = MPI_LOGICAL
    types(4) = MPI_CHARACTER
    
    call MPI_TYPE_CREATE_STRUCT(DATA_N_TYPES, block_lengths, memory_displacements, types, qcalc_type, mpierr)
    call MPI_TYPE_COMMIT(qcalc_type, mpierr)

    get_qcalc_mpi_type = qcalc_type
  end function get_qcalc_mpi_type
  

#endif

end module mpistructure
