!***********************************************************************
!                                                                     
!              MCTDH include file  PSIDEF.F90
!                                                                     
!  11/96 AJ                                                         
!                                                                     
!***********************************************************************

      module psidef

      use decimal, only : long,dop
      use global, only: c2

      implicit none

!-----------------------------------------------------------------------
! VARIOUS INTEGERS DEFINING SIZE OF WAVEFUNCTION
!
! dgldim   : Length of array containing wavefunction
!            dgldim = block + Sum[dim(m)*subdim(m)]
! dmatdim  : Length of array containing the density matrices
!            dmatdim = Sum[dim(m)*dim(m)]
! d3matdim : Length of array containing dicht3 matrix
!            d3matdim = Sum[2*dim(m)]
! adim     : Cumulative length of all A vectors.
!            adim = Sum[block(s)]. 
!            For ML-MCTDH this is the length of the top-layer vector
! maxspf   : Largest number of single-particle functions
!            maxspf = Max_(m,s) dim(m,s)
! maxjindx : maximum no. of J indices needed in selected CI calculation.
! jindxdim : total no. of J indices needed in the mi array.
! mladim   : the ML-MCTDH length of the A vector (coefficients in
!            all layers except primitive SPFs)
!
!-----------------------------------------------------------------------
      integer(long), save :: dgldim,d3matdim,adim,maxspf,maxjindx,&
                             jindxdim
      integer(long), save :: mladim
      integer(long), save :: dmatdim

!-----------------------------------------------------------------------
! nlayers : No. of layers in an ML-MCTDH calculation. 
!           Default is 2 (normal MCTDH)
!-----------------------------------------------------------------------
      integer(long), save :: nlayers = 2

!-----------------------------------------------------------------------
! ARRAY DIMENSIONS DEPENDENT ON NMODE
!
! PhiDim(M): Length of phi-vector for mode m
!            PhiDim(M) = Sum[Dim(S,M)*SubDim(M)]
! maxphidim: Size of largest phi vector
!            maxphidim = max[phidim(m)]
! totphidim: overall size of single-particle functions (including GBFs)
!            totphidim = Sum_m [phidim(m)] = dgldim-adim
! gbfdim(m): Length of GBF vector for mode m
! maxgbfdim: Size of largest GBF vector
!-----------------------------------------------------------------------
      integer(long), save                            :: maxphidim,totphidim,&
                                                        maxgbfdim
      integer(long), dimension(:), allocatable, save :: phidim,gbfdim
      
!-----------------------------------------------------------------------
! INFORMATION ABOUT WAVEFUNCTION TYPE
!
! psitype: 0 = MCTDH (default) or MCTDH/G (i.e. SPFs expanded in GWP
!              basis)
!          1 = numerically exact (full grid representation)
!          2 = Selected-CI wavefunction
!          3 = swarm of trajectories
!          4 = G-MCTDH (i.e. GWP basis used as SPFs)
!         -4 = G-MCTDH with "plain GWPs"
!          5 = ML-MCTDH
!         11 = numerically exact real wavepacket (full grid representation)
!
! complete(m) : if true, spf basis is "complete" and (1-P)H phi =0
!-----------------------------------------------------------------------
      integer(long), save :: psitype = 0
      logical(kind=4), dimension(:), allocatable , save  :: complete

!-----------------------------------------------------------------------
! PARAMETERS DEFINING SINGLE-PARTICLE FUNCTION BASIS
!
! sbaspar:    maximum number of parameters used to define bases
!
! sbasis: defines the type of initial functions
! isbaspar: integer parameters defining initial SPFs
! rsbaspar: integer parameters defining initial SPFs
! ltdp : true if a time-dependent point SPF is used
!-----------------------------------------------------------------------
      integer(long), parameter :: sbaspar=10

      integer(long), dimension(:), allocatable, save     :: sbasis
      integer(long), dimension(:,:,:), allocatable, save :: isbaspar
      real(dop), dimension(:,:,:), allocatable, save     :: rsbaspar
      logical(kind=4), save                              :: ltdp = .false.

!-----------------------------------------------------------------------
! parameters needed for S-MCTDH
!
! citype   : type of selection used
!            1 = static
! cicut    : cutoff for selection
! ciweight(m) : weight for selection
!-----------------------------------------------------------------------
      integer(long), save                        :: citype
      real(dop), save                            :: cicut
      real(dop), dimension(:), allocatable, save :: ciweight

!-----------------------------------------------------------------------
! parameters for swarm of trajectories
! for grid / GWPs ntraj=0
!
! ntraj0 : number of initial trajectory
! ntraj  : number of trajectories in swarm
!-----------------------------------------------------------------------
      integer(long), save :: ntraj = 0, &
                             ntraj0 = 1
      
!-----------------------------------------------------------------------
! parameters for Cartesian vMCG basis (when using vMCG input)
! ngwp: no. of GWPs
! npargtwp: no. of modes into which the GWP function is partitioned
!-----------------------------------------------------------------------
      integer(long), save :: ngwp
      integer(long), save :: npartgwp = 1

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! dim(m,s)   : number of single particle functions for mode m, state s
! vdim(m,s)  : cumulative number of preceeding degrees of freedom
! ndim(m,s)  : cumulative number of following degrees of freedom
! block(s)   : number of MCTDH A coefficients for state s
! zetf(m,s)  : pointer to start of single particle functions for mode
!              m in psi
! gwpdim(m,s): number of GWP basis functions for mode m, state s
! dimred(m,s): actual number of GWP basis functions for mode m, state s for BEL
! dmat(m,s)  : pointer to start of dichte matrices for mode m, state s
! d3mat(m,s) : pointer to start of dichte3 matrices for mode m, state s
! zpsi(s)    : pointer to wavefunction for state s
! jvdim(m)   : vdim for full jspace
! jdim(m)    : dim for full jspace
! jndim(m)   : ndim for full jspace
! jblock     : total size of full jspace
! maxblock   : Largest size of "block(s)". Also the largest vector 
!              required when operating on psi, e.g. in matrizen2 or 3
!              (see "maxmatwk" zeigprop for the exact size of "maxblock")
! fmblock    : Full maxblock. Similar to maxblock but for the "long"  
!              vector in S-MCTDH (citype=1). For citype=0, fmblock=maxblock.
! msymmtr    : Array specifying (a)symmetrization of DOFs and 2D SPFs
! idmode     : Array specifying identical (by symmetry) modes
! symcv      : Array specifying partial symmetrization of A-vector
!-----------------------------------------------------------------------
      integer(long), save                            :: iofunc,maxblock, &
                                                        fmblock,jblock
      integer(long), dimension(:), allocatable, save :: zpsi,block
      integer(long), dimension(:), allocatable, save :: jdim,jvdim,jndim,&
                                                        idmode,symcv
      integer(long), dimension(:), allocatable, save :: msymmtr
      integer(long), dimension(:,:), allocatable, save :: dim,vdim,ndim,zetf, &
                                                          d3mat,dmat,gwpdim, &
                                                          dimred

!-----------------------------------------------------------------------
! options for generation of Psi
!
!  lautoblk  : Automatic generation of start A-vector for Block-Davidson
!-----------------------------------------------------------------------
      logical(kind=4), save :: lautoblk

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS USED IN OPWF
!-----------------------------------------------------------------------
      integer(long), save                           :: d3matsdim,dmatsdim, &
                                                       workcdims,workrdims
      integer(long), dimension(:,:), allocatable, save :: dims,d3mats,dmats

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS FOR DENSITY OPERATORS
!
! ddim(m,s,t)   : number of single particle functions for mode m
! vddim(m,s,t)  : cumulative number of preceeding degrees of freedom
! nddim(m,s,t)  : cumulative number of following degrees of freedom
! dblock(s,t)   : number of MCTDH A coefficients for state s 
! dzetf(m,s,t)  : pointer to start of single particle functions for mode
!                 m in psi
! ddmat(m,s,t)  : pointer to start of dichte matrices for mode m
! dd3mat(m,s,t) : pointer to start of dichte3 matrices for mode m
! dzpsi(s,t)    : pointer to wavefunction for state s
!-----------------------------------------------------------------------
      integer(long), dimension(:,:,:), allocatable, save :: ddim,vddim, &
                                                            nddim,dzetf,&
                                                            dd3mat,ddmat
      integer(long), dimension(:,:), allocatable, save   :: dzpsi,dblock

!-----------------------------------------------------------------------
! TEMPERATURE, INITIAL STATE AND GROUNDSTATE
!
! ltempmfsc = .true. if frequencies used in the thermalisation are
!                    read from parameters (e.g. if mass-freq scaled
!                    coordinates have been used).
! temperature0 : initial temperature
! temperature_psi : temperature of system 
!-----------------------------------------------------------------------
      real(dop), save         :: temperature0 = -999.0_dop
      real(dop), save         :: temperature_psi = 0.0_dop
      real(dop), save         :: intrace
      complex(dop), save      :: inctr
      integer(long), save     :: dlstate, drstate, ctrnum
      character(len=c2), save :: ctrname
      logical(kind=4), save   :: ltemp_mfsc=.false.

!-----------------------------------------------------------------------
! pointers for gwp portions of psi and dtpsi:
! zphi(m,s),zxi(m,s),zpg(m,s)
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save :: zphi,zxi,zpg,zdgwp

!-----------------------------------------------------------------------
! ARRAY SIZES OF GWP MATRICES
! all sums are exclusively over GWP modes
!
! gs2matdim     : length of array gs2, for Gaussian overlap elements
!                 [gs2matdim = Sum npg(m)*dim(m,s)**2]
! gsmatdim      : length of arrays which have dimensions of an overlap
!                 matrix in Gaussian modes.
!                 [gsmatdim = Sum dim(m,s)**2]
!-----------------------------------------------------------------------
      integer(long), save :: agmatdim,agmatdim100,agmatdim101,qcentdim,&
                             xpmatdim
      integer(long), save :: gs2matdim,gsmatdim,gfgwpdim,aigwpdim,zgwpdim

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! zagmat(m,s)      : pointers for width matrices
! zxivec(m,s)      : pointers for linear parameters
! zcent(m,s)       : pointer for array qcenter
! gs2mat(m,s)      : pointer for overlap matrices
! gsmat(m,s)       :
! zgfgwp(m,s)      :
! zaigwp(m,s)      : pointers for moment generating functions
! zzgwp(m,s)       :
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save   :: zagmat,zxivec,&
                                                            zcent,&
                                                            gs2mat,gsmat,&
                                                            zphase ! yapol
      integer(long), dimension(:,:,:), allocatable, save :: zgfgwp,zaigwp, &
                                                            zzgwp

!-----------------------------------------------------------------------
! Temporary arrays for vMCG symmetrisation
!-----------------------------------------------------------------------
      complex(dop), allocatable, dimension(:), save :: gfgwptmp,aigwptmp,&
                                                       zgwptmp

!-----------------------------------------------------------------------
! GWP INFORMATION ARRAYS
!
! zgwpinfo(m,s): pointer for GWP "info" arrays (eg ddgwpinfo)
! ngwpdep: No. of linearly dependent gwp functions
!
! The following data is needed to generate list of GWP configurations 
! totgp:  total no. of GWP configurations
! nsgp(s): no. of GWP configurations for state s
! zgwp(s): pointer to list of configurations
! vdimgp(m,s), dimgp(m,s), ndimgp(m,s) : sizes for GWP configuration list
!-----------------------------------------------------------------------
      integer(long), save                              :: gwpinfodim=1
      integer(long), dimension(:,:), allocatable, save :: zgwpinfo
      integer(long), dimension(:,:), allocatable, save :: ngwpdep

      integer(long), save                              :: totgp
      integer(long), dimension(:), allocatable, save   :: nsgp,zgp
      integer(long), dimension(:,:), allocatable, save :: vdimgp,dimgp,ndimgp

!-----------------------------------------------------------------------
! total norm of a thermalised wavepacket 
!-----------------------------------------------------------------------
      real(dop), save :: normthermal

      contains

!#######################################################################
!
! PSIDEF_ALLOC
!
! memory required for arrays in psidef allocated here.
!
! First check if allocated, then allocate. All arrays must be deallocated in 
! psidef_dealloc
!
!#######################################################################

      subroutine alloc_psidef

use maxv, only: maxdim,maxsta,lmemold
use global
use dvrdatmod
use channels, only : ilog

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr
      integer(long), external   :: integersize, realsize, logicalsize

! check all arrays are not allocated
      call dealloc_psidef

      cerr = 'complete'
      allocate(complete(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*logicalsize()
      if (ierr > 0) go to 1000
      cerr = 'phidim'
      allocate(phidim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      cerr = 'gbfdim'
      allocate(gbfdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'ciweight'
      allocate(ciweight(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*realsize()
      if (ierr > 0) go to 1000
      cerr = 'dim'
      allocate(dim(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'vdim'
      allocate(vdim(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'ndim'
      allocate(ndim(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'zetf'
      allocate(zetf(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'zpsi'
      allocate(zpsi(maxsta),stat=ierr)
      allocmemory=allocmemory + maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'block'
      allocate(block(maxsta),stat=ierr)
      allocmemory=allocmemory + maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'd3mat'
      allocate(d3mat(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dmat'
      allocate(dmat(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      if (dentype .eq. 1 .or. lmemold) then
         cerr = 'ddim'
         allocate(ddim(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'vddim'
         allocate(vddim(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'nddim'
         allocate(nddim(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'dzetf'
         allocate(dzetf(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'dzpsi'
         allocate(dzpsi(maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'dblock'
         allocate(dblock(maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'dd3mat'
         allocate(dd3mat(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'ddmat'
         allocate(ddmat(maxdim,maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
      endif 
      if (dentype .eq. 2) then
         cerr = 'dzpsi'
         allocate(dzpsi(maxsta,maxsta),stat=ierr)
         allocmemory=allocmemory + maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
      endif 
      cerr = 'jdim'
      allocate(jdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'jvdim'
      allocate(jvdim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'jndim'
      allocate(jndim(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'idmode'
      allocate(idmode(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'symcv'
      allocate(symcv(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'msymmtr'
      allocate(msymmtr(2*maxdim),stat=ierr)
      allocmemory=allocmemory + 2*maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dims'
      allocate(dims(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'd3mats'
      allocate(d3mats(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dmat'
      allocate(dmats(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'gwpdim'
      allocate(gwpdim(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'dimred'
      allocate(dimred(maxdim,maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*integersize()
      if (ierr > 0) go to 1000
      if (lgwp .or. ltraj .or. lmemold) then
         cerr = 'zphi'
         allocate(zphi(maxdim,maxsta),stat=ierr)
         zphi=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zxi'
         allocate(zxi(maxdim,maxsta),stat=ierr)
         zxi=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zpg'
         allocate(zpg(maxdim,maxsta),stat=ierr)
         zpg=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zdgwp'
         allocate(zdgwp(maxdim,maxsta),stat=ierr)
         zdgwp=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zagmat'
         allocate(zagmat(maxdim,maxsta),stat=ierr)
         zagmat=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zxivec'
         allocate(zxivec(maxdim,maxsta),stat=ierr)
         zxivec=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zphase'
         allocate(zphase(maxdim,maxsta),stat=ierr)
         zphase=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zcent'
         allocate(zcent(maxdim,maxsta),stat=ierr)
         zcent=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         if (ierr > 0) go to 1000
         cerr = 'gs2mat'
         allocate(gs2mat(maxdim,maxsta),stat=ierr)
         gs2mat=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'gsmat'
         allocate(gsmat(maxdim,maxsta),stat=ierr)
         gsmat=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zgwpinfo'
         allocate(zgwpinfo(maxdim,maxsta),stat=ierr)
         zgwpinfo=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zgfgwp'
         allocate(zgfgwp(maxdim,maxsta,maxsta),stat=ierr)
         zgfgwp=0
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zaigwp'
         allocate(zaigwp(maxdim,maxsta,maxsta),stat=ierr)
         zaigwp=0
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zzgwp'
         allocate(zzgwp(maxdim,maxsta,maxsta),stat=ierr)
         zzgwp=0
         allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'ngwpdep'
         allocate(ngwpdep(maxdim,maxsta),stat=ierr)
         ngwpdep=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000

         cerr = 'nsgp'
         allocate(nsgp(maxsta),stat=ierr)
         nsgp=0
         allocmemory=allocmemory + maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'zgp'
         allocate(zgp(maxsta),stat=ierr)
         zgp=0
         allocmemory=allocmemory + maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'vdimgp'
         allocate(vdimgp(maxdim,maxsta),stat=ierr)
         vdimgp=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'dimgp'
         allocate(dimgp(maxdim,maxsta),stat=ierr)
         dimgp=0
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000
         cerr = 'ndimgp'
         allocate(ndimgp(maxdim,maxsta),stat=ierr)
         ndimgp=0 
         allocmemory=allocmemory + maxdim*maxsta*integersize()
         if (ierr > 0) go to 1000

      endif
      cerr = 'sbasis'
      allocate(sbasis(maxdim),stat=ierr)
      allocmemory=allocmemory + maxdim*integersize()
      if (ierr > 0) go to 1000
      cerr = 'isbaspar'
      allocate(isbaspar(sbaspar,maxdim,maxsta*maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
      if (ierr > 0) go to 1000
      cerr = 'rsbaspar'
      allocate(rsbaspar(sbaspar,maxdim,maxsta*maxsta),stat=ierr)
      allocmemory=allocmemory + maxdim*maxsta*maxsta*integersize()
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_PSIDEF :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_psidef

!#######################################################################

      subroutine dealloc_psidef

      if (allocated(complete)) deallocate(complete)
      if (allocated(phidim)) deallocate(phidim)
      if (allocated(gbfdim)) deallocate(gbfdim)
      if (allocated(ciweight)) deallocate(ciweight)
      if (allocated(zpsi)) deallocate(zpsi)
      if (allocated(block)) deallocate(block)
      if (allocated(jdim)) deallocate(jdim)
      if (allocated(jvdim)) deallocate(jvdim)
      if (allocated(jndim)) deallocate(jndim)
      if (allocated(idmode)) deallocate(idmode)
      if (allocated(symcv)) deallocate(symcv)
      if (allocated(msymmtr)) deallocate(msymmtr)
      if (allocated(dim)) deallocate(dim)
      if (allocated(vdim)) deallocate(vdim)
      if (allocated(ndim)) deallocate(ndim)
      if (allocated(zetf)) deallocate(zetf)
      if (allocated(d3mat)) deallocate(d3mat)
      if (allocated(dmat)) deallocate(dmat)
      if (allocated(dims)) deallocate(dims)
      if (allocated(d3mats)) deallocate(d3mats)
      if (allocated(dmats)) deallocate(dmats)
      if (allocated(ddim)) deallocate(ddim)
      if (allocated(vddim)) deallocate(vddim)
      if (allocated(nddim)) deallocate(nddim)
      if (allocated(dzetf)) deallocate(dzetf)
      if (allocated(dd3mat)) deallocate(dd3mat)
      if (allocated(ddmat)) deallocate(ddmat)
      if (allocated(dzpsi)) deallocate(dzpsi)
      if (allocated(dblock)) deallocate(dblock)
      if (allocated(gwpdim)) deallocate(gwpdim)
      if (allocated(dimred)) deallocate(dimred)
      if (allocated(zphi)) deallocate(zphi)
      if (allocated(zxi)) deallocate(zxi)
      if (allocated(zpg)) deallocate(zpg)
      if (allocated(zdgwp)) deallocate(zdgwp)
      if (allocated(zagmat)) deallocate(zagmat)
      if (allocated(zxivec)) deallocate(zxivec)
      if (allocated(zphase)) deallocate(zphase)
      if (allocated(zcent)) deallocate(zcent)
      if (allocated(gs2mat)) deallocate(gs2mat)
      if (allocated(gsmat)) deallocate(gsmat)
      if (allocated(zgwpinfo)) deallocate(zgwpinfo)
      if (allocated(ngwpdep)) deallocate(ngwpdep)
      if (allocated(zgfgwp)) deallocate(zgfgwp)
      if (allocated(zaigwp)) deallocate(zaigwp)
      if (allocated(zzgwp)) deallocate(zzgwp)
      if (allocated(sbasis)) deallocate(sbasis)
      if (allocated(isbaspar)) deallocate(isbaspar)
      if (allocated(rsbaspar)) deallocate(rsbaspar)
      if (allocated(nsgp)) deallocate(nsgp)
      if (allocated(zgp)) deallocate(zgp)
      if (allocated(vdimgp)) deallocate(vdimgp)
      if (allocated(dimgp)) deallocate(dimgp)
      if (allocated(ndimgp)) deallocate(ndimgp)

      return
      end subroutine dealloc_psidef

      end module psidef
