module realwpmod

use decimal, only: long,dop

implicit none

!-----------------------------------------------------------------------
! kabs(k)    : terms for calculation of the absorption
! kflux(k)   : terms for the flux calculation
! fabs(k)    : absorption along f
! fflux(k)   : flux along f
! indiabs(f) : number of grid point where the absorption starts along f 
! indiflux(f): number of grid point where the flux is calculating along f 
! normwav    : initial norm of wave function
! timestep   : Time per step      
!-----------------------------------------------------------------------

      integer(long), dimension(:), allocatable, save  :: indiabs,indiflux
      logical(kind=4), dimension(:), allocatable, save:: lrotvec,lrotcoup 
      real(dop), save                         :: normwav,timestep, &
                                                 energsav,escatsav 

! To separate rotational cupling terms (j_p,j_m,jpm,jp^2,jm^2)
      integer(long), dimension(:), allocatable, save  :: nrotcoup

! Chebyshev integration variable
      integer(long), save                     :: norder
      real(dop), save                         :: aunds,bunds

!-----------------------------------------------------------------------
! Energy dependent probability calculations
!
! prne     : Number of energies to be calculated for probability
! femax    : Maximum energy for calculation of E-dep flux
! femin    : Minimum energy for calculation of E-dep flux
! ntimfluxo: Number of points in Fourier transform 
! nfluxf   : Number of coordinates are active for flux calculations
! nabsf    : Number of coordinates are active for absorption 
! inimom   : initial momentum in wave packet
! spread   : initial spread in wave packet
!-----------------------------------------------------------------------
      integer(long), save                     :: fluxsave,prne,ntimfluxo, &
                                                 nfluxf,nabsf
      real(dop), save                         :: femax,femin,inimom,spread
      logical(kind=4), save                   :: lsave!,lcoef
      integer(long), dimension(:), allocatable, save  :: iprob,iunprob

contains

!#######################################################################
!
! REALWPMOD_ALLOC
!
! memory required for arrays in realwpmod allocated here.
!
! First check if allocated, then allocate. All arrays must be deallocated in
! realwpmod_dealloc
!
!#######################################################################

      subroutine alloc_realwpmod

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_realwpmod

      cerr = 'indiabs'
      allocate(indiabs(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'indiflux'
      allocate(indiflux(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'lrotvec'
      allocate(lrotvec(maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'lrotcoup'
      allocate(lrotcoup(maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nrotcoup'
      allocate(nrotcoup(maxkoe),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'iprob'
      allocate(iprob(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'iunprob'
      allocate(iunprob(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_REALWPMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      write(6,'(a)') 'Error allocating memory in ALLOC_REALWPMOD :',cerr
      write(6,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_realwpmod

!#######################################################################

      subroutine dealloc_realwpmod

      if (allocated(indiabs)) deallocate(indiabs)
      if (allocated(indiflux)) deallocate(indiflux)
      if (allocated(lrotvec)) deallocate(lrotvec)
      if (allocated(lrotcoup)) deallocate(lrotcoup)
      if (allocated(nrotcoup)) deallocate(nrotcoup)
      if (allocated(iprob)) deallocate(iprob)
      if (allocated(iunprob)) deallocate(iunprob)

      return
      end subroutine dealloc_realwpmod

end module realwpmod
