!#######################################################################
!     runinwfmod.f90
!#######################################################################

      module runinwfmod

      use global, only: c5
      use decimal, only : long,dop
      use maxv, only : maxdim,maxsta

      implicit none

!-----------------------------------------------------------------------
! PARAMETERS FOR ARRAY DIMENSIONS
!
! sbfound:    logical used to check that all parameters given
!-----------------------------------------------------------------------

      logical(kind=4), dimension(:,:), allocatable, save :: sbfound

!-----------------------------------------------------------------------
! array sizes
!
! spf1ddim: length of spf1d array for 1-D "spf" functions
! adim1d  : no. of coefficients to be generated with 1d spfs
!-----------------------------------------------------------------------
      integer(long), save :: spf1ddim,adim1d

!-----------------------------------------------------------------------
! pointers for arrays for density operators of type I
!
! dzspf1d(f,s,t) : pointers for spf1d array for 1-D "spf" functions
! ddimf(f,s,t)   : no. of 1d spfs to be generated
! dzpsi1d(s,t)   : pointer to state wavefunction of 1d spfs
! dimfran(f)  : no. of states that will be taken into account
!               (e.g. "ran" keyword) in a thermal relaxation run
!-----------------------------------------------------------------------
      integer(long), dimension(:,:,:), allocatable, save :: dzspf1d,ddimf
      integer(long), dimension(:,:), allocatable, save   :: dzpsi1d,dblock1d
      integer(long), dimension(:), allocatable, save     :: dimfran

!-----------------------------------------------------------------------
! pointers for arrays
!
! zspf1d(f,s) : pointers for spf1d array for 1-D "spf" functions
! dimf(f,s)   : no. of 1d spfs to be generated
! zpsi1d(s)   : pointer to state wavefunction of 1d spfs
!-----------------------------------------------------------------------
      integer(long), save                              :: adimf
      integer(long), dimension(:,:), allocatable, save :: zspf1d,dimf, &
                                                          vdimf,ndimf
      integer(long), dimension(:), allocatable, save   :: zpsi1d,block1d

!-----------------------------------------------------------------------
! FLAGS TO CONTROL INWF OPTIONS
!
!  lrestart : the initial wavefunction is read from file(s).
!  inwfop   : no. of operator used to modify an initial wavepacket.
!  lad      : adiabatic corrections
!  ldia     : diabatic corrections
!  lhh2     : special (a)diabatic corrections for H + H2, H + D2
!  lbkmp2   : special (a)diabatic corrections with BKMP2-PES
!  ledstr   : calculate adiabatically corrected energy distribution
!  lbuildwf : the initial wavefunction is build by genphi and gencoeff.
!  lblockspf: the initial SPFs are read from file. For block-imprlx.
!  lblocka  : the initial A-vectors are read from file. For block-imprlx.
!  lorthopsi: wavefunction is Schmidt-orthogonalised at start of restart.
!  lhartreepsi: initial wavefunction is a Hartree product.
!  lrstnorm : wavefunction is normalised after reading in from file
!  lrstcoh  : A coherent superposition is formed in a density matrix
!             calculation on restart (default)
!  lrealpsi : the real part of the wavefunction is Schmidt-orthogonalised
!             and used as the initial wavefunction.
!  lnodirect: Switches off the direct calculation of operate (inwfop).
!             Only the iterative calculation is performed.  
!  pqdistr  : distribution in classical phase-space
!           : 1 = random Wigner distribution
!  pqseed   : seed used for random distributions
!  lnonorm  : Operate normalisation factor will be removed.
!  loportho : Operate with (D-shift), where shift is automatically
!             chosen such that the operated WF is orthogonal to initial WF.
!  lsymorg  : Symmetrisation of two different sets of SPFs
!  lprintmode: Information on how the combined modes are build is written
!              to log-file. (Not for KLeg, there isbaspar(5) is used). 
!-----------------------------------------------------------------------
      integer(long), save   :: pqdistr!,pqseed
      logical(kind=4), save :: lrestart = .false., &
                               lorthopsi = .true., &
                               lhartreepsi = .false., &
                               lrstnorm = .true., &
                               lrstcoh = .true., &
                               lblockspf = .false., &
                               lblocka = .false., &
                               lbuildwf = .false., &
                               lrealpsi2 = .false., &
                               lrealpsi = .false., &
                               lbkmp2,lnodirect, &
                               ladiab,lad,ldia,ledstr,lhh2, &
                               lnonorm,loportho,lprintmode,lsymorb

!-----------------------------------------------------------------------
! Logical which allows to switch off the SPF redundancy check
!-----------------------------------------------------------------------
      logical(kind=4), save :: lredundant = .true.

!-----------------------------------------------------------------------
! FLAGS TO CONTROL THE OPERATION OF AN OPERATOR ON THE INWF
!  inwfop : the operator(s) used to generate an initial wavepacket.
!  inwfopnit : Number of iterations (default=10).
!  inwfoptol : Error tolerance, default = 1.0d-8.
!  operatenorm : normalisation factor of operate. (see lnonorm) 
! FLAGS TO CONTROL THE ORTHOGONALISATION OF THE INWF AGAINST OTHER WF.
!  iorthowf  : Number of orthogonalisations.
! FLAGS TO CONTROL THERMAL STATE WFN
!  mlseed   : seed for random number generator (thermal state)
!-----------------------------------------------------------------------
      integer(long), save                :: inwfopnit,iorthowf
      integer(long), save                :: mlseed = -1
      integer(long), dimension(16), save :: inwfop
      real(dop), save                    :: inwfoptol,operatenorm

!-----------------------------------------------------------------------
! addim  : dim(2,1)*dim(3,1)
! azahl  : number of a-coefficients to be read
! iza    : input file line for start of a-coefficient list
! Beginning of modification (Benjamin, 14 Feb 2007)
! gwpzahl: number of gwp-functions to be read in build_gwp_func
! izgwp  : input file line for start of gwp-function list
! symorb1 : mode number of the first  set of SPFs to be symmetrized
! symorb2 : mode number of the second set of SPFs to be symmetrized
! gwpopit : no. of iterations to be used in optimising GWPs to fit
!             initial WP.
! lgwpop0: if true optimise initial GWPs to fit initial WP
!-----------------------------------------------------------------------
      integer(long), save :: addim,azahl,iza,gwpzahl,izgwp,symorb1,symorb2
      integer(long), save :: gwpopit = 50
      logical(kind=4), save :: lgwpop0 = .false.

!-----------------------------------------------------------------------
! PARAMETERS NEEDED BY VEIGEN, MEIGEN AND ADIABATIC ROUTINES
!-----------------------------------------------------------------------
      integer(long), parameter                :: maxedim=120,maxme=12
      integer(long), save                     :: menum,meham,medim,memode, &
                                                 mestate,meselect,mewrite, &
                                                 mez,mxsdm,meran
      integer(long), dimension(maxme,7), save :: mearray
      real(dop), save                         :: e0eigen
      real(dop), dimension(maxedim), save     :: eeigen

!-----------------------------------------------------------------------
! PARAMETERS DEFINING READING OF AN INITIAL WF
!-----------------------------------------------------------------------
      integer(long), parameter   :: maxreadspf=10
      integer(long), parameter   :: maxreadwf=10

      integer(long), save                                :: nreadspf=0
      integer(long), save                                :: nreadwf=0
      integer(long), save                                :: equalfile=0
      integer(long), dimension(:,:,:), allocatable, save :: smapwf, amapwf
      integer(long), dimension(:,:,:), allocatable, save :: selectspf
      complex(dop), dimension(:), allocatable, save      :: aweight
      character(len=c5), dimension(:), allocatable,save  :: readspffile
      character(len=c5), dimension(:), allocatable,save  :: readwffile

!-----------------------------------------------------------------------
! common block correction routines. (needed in adiab routines and LSTH)
!-----------------------------------------------------------------------
      integer(long), save :: ftr,statetr
      real(dop), save     :: tfac

!-----------------------------------------------------------------------
!     lsymmultkleg = .true. if there are multi-KLeg modes and 
!     symmetrization is used. Set in chkinwf and causes arrays
!     in module multkleg to be set up.
!-----------------------------------------------------------------------
      logical(kind=4), dimension(:), allocatable, save :: lsymmultkleg

      contains

!#######################################################################
!
! RUNINWFMOD_ALLOC
!
! memory required for arrays in runinwfmodmod allocated here.
!
! First check if allocated, then allocate. All arrays must be
! deallocated in
! runinwfmod_dealloc
!
!#######################################################################

      subroutine alloc_runinwfmod

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=15) :: cerr

! check all arrays are not allocated
      call dealloc_runinwfmod

      cerr = 'sbfouond'
      allocate(sbfound(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dzspf1d'
      allocate(dzspf1d(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ddimf'
      allocate(ddimf(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dimfran'
      allocate(dimfran(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dzpsi1d'
      allocate(dzpsi1d(maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dblock1d'
      allocate(dblock1d(maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zspf1d'
      allocate(zspf1d(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dimf'
      allocate(dimf(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'vdimf'
      allocate(vdimf(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ndimf'
      allocate(ndimf(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zpsi1d'
      allocate(zpsi1d(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'block1d'
      allocate(block1d(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'aweight'
      allocate(aweight(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'lsymmultkleg'
      allocate(lsymmultkleg(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      lsymmultkleg=.false.
      cerr = 'selectspf'
      allocate(selectspf(maxreadspf,maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'readspffile'
      allocate(readspffile(maxreadwf),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'readwffile'
      allocate(readwffile(maxreadwf),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'smapwf'
      allocate(smapwf(maxsta,2,maxreadwf),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'amapwf'
      allocate(amapwf(maxsta,2,maxreadwf),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_RUNINWFMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_runinwfmod

!#######################################################################

      subroutine dealloc_runinwfmod

      if (allocated(sbfound)) deallocate(sbfound)
      if (allocated(dzspf1d)) deallocate(dzspf1d)
      if (allocated(ddimf)) deallocate(ddimf)
      if (allocated(dimfran)) deallocate(dimfran)
      if (allocated(dzpsi1d)) deallocate(dzpsi1d)
      if (allocated(dblock1d)) deallocate(dblock1d)
      if (allocated(zspf1d)) deallocate(zspf1d)
      if (allocated(dimf)) deallocate(dimf)
      if (allocated(vdimf)) deallocate(vdimf)
      if (allocated(ndimf)) deallocate(ndimf)
      if (allocated(zpsi1d)) deallocate(zpsi1d)
      if (allocated(block1d)) deallocate(block1d)
      if (allocated(smapwf)) deallocate(smapwf)
      if (allocated(amapwf)) deallocate(amapwf)
      if (allocated(aweight)) deallocate(aweight)
      if (allocated(lsymmultkleg)) deallocate(lsymmultkleg)
      if (allocated(selectspf)) deallocate(selectspf)
      if (allocated(readspffile)) deallocate(readspffile)
      if (allocated(readwffile)) deallocate(readwffile)

      return
      end subroutine dealloc_runinwfmod

      end module runinwfmod
