!***********************************************************************
! 
!  POTFIT include file defining array sizes, common blocks etc. common 
!  to several parts of the program.  
!  defines array sizes, common blocks etc. common to several parts 
!  of the program.
! 
!   10/96 AJ
!
!
!***********************************************************************
      module runpfitmod

use decimal, only: long,dop

      implicit none

!-----------------------------------------------------------------------
! ARRAY SIZES
!
! vpotdim   : Length of array containing exact potential data on grid
!             vpotdim = Prod[gdim(m)]
! cdim      : Length of potential expansion coefficient vector
!             cdim   = Prod[potdim(m)] !!! WARNING: cdim changed to ctdim !!!
! oneddim   : Length of array containing one-dimensional potential curves
!             oneddim = Sum[gdim(m)]
! modcdim   : Number of contracted natural potentials
!             modcdim = Prod_(m.ne.modc)[potdim(m)]
! cvecdim   : Length of array containing uncontracted potential expansion
!             coefficients
!             cvecdim = Prod_(m)[potdim(m)]
! coeffdim  : Length of array containing potential expansion coefficients
!             and later (the equal or larger) array containing
!             the contracted natural potentials
!             coeffdim = Prod_(m.ne.modc)[potdim(m)]*gdim(modc)
! evecdim   : Length of array containing all (even if not used later on)
!             natural potentials
!             evecdim = Sum[psubdim(m)**2]
! evaldim   : Length of array containing all natural eigenvalues
! weightdim : Length of array containing the separable weight vectors
!             weightdim = ortdim
! maxcldim  : Max[cldim(m)]
! sumcldim  : Sum[cldim(m)]
! maxvprojdim: maximum size of projected potentials
! maxnodedim:  maximum size of that part of the remaining/outer grid that
!              is needed on one MPI node
! maxinnerdim: maximum size of projected (inner) grid
! maxprjfdim: maximum size of storage for projector functions
! totprjfdim: total storage size for all projector functions (of all proj.s)
! ortdim1   : Length of coordinate array for function to be subtracted
! hopdim1   : Length of hoplab array for function to be subtracted
!
!!  where Sum/Prod/Max is over all modes
!! WARNING: ctdim was cdim in versions before HDmctdh8.6
!-----------------------------------------------------------------------
      integer(long), save :: vpotdim,evecdim,coeffdim,evaldim,&
                             weightdim,workdim1,workdim2,ctdim,maxcldim,&
                             modcdim,oneddim,sumcldim,maxvprojdim,&
                             maxprjfdim,ortdim1,hopdim1,maxnodedim, &
                             maxinnerdim,totprjfdim,cvecdim

!-----------------------------------------------------------------------
! DIMENSIONS AND POINTERS FOR ARRAYS
!
! potdim(m) : number of relevant natural potentials for mode m
! cldim(f)  : dimension of "long" coefficient vector used to
!             calculate the density matrix for mode f during iteration
! c2v       : holds dimension information used during the transformation of
!             coefficients to potential energies, i.e., during the
!             calculation of a potential energy fit.
! v2c       : holds information used during the transformation of
!             potential energies to contracted coefficients, i.e.,
!             during the calculation of contracted coefficients.
!             col 1: degree of freedom F to be transformed
!             col 2: size of the vector AFTER the transformation of F
!             col 3: size of the dimensions < F BEFORE the trafo
!             col 4: size of dimension F BEFORE the trafo
!             col 5: size of dimensions > F BEFORE the trafo
!             col 6: size of dimension F AFTER the trafo
! wdim(f)    : dimension of the separable weight vector for mode f
! zevec(m)   : relative offset of the eigenvectors of rho(m)
! zweight(f) : relative offset of the separable weight vector of mode f
! zcl(f)     : relative offset of the "long" coefficient vector of mode f
! vgmdim(f)  : "vor" grid-dimension in mode order. 
! vgpdim(f)  : "vor" grid-dimension in dof  order. 
! isppoint(f): Set of DVR indices for a single-point energy calculation
! selmode    : Selected mode along which the surface cut will be performed 
! mctens     : Requested size for the contracted mode SPPs when
!              calculation of the C-tensor is required
! zort1(f)   : no. or grid points for DOF f for function to be subtracted
! gdim1(f)   : length of grid for DOF f for function to be subtracted
! goff(f)    : any offset in the grid between the system and function
!              to be subtracted
!-----------------------------------------------------------------------
      integer(long), save                             :: bufferlength, &
                                                         selmode,mctens
      integer(long), dimension(:), allocatable, save  :: potdim,wdim,zevec, &
                                                         zweight,cldim,&
                                                         zcl,vgpdim,vgmdim, &
                                                         zort1,gdim1,&
                                                         goff,isppoint
      integer(long), dimension(:), allocatable, save ::  c2v
      integer(long), dimension(:,:), allocatable, save :: v2c

!-----------------------------------------------------------------------
! PARAMETERS DEFINING WEIGHT FUNCTIONS
!-----------------------------------------------------------------------
      integer(long), parameter                       :: maxwpar = 6
      integer(long), dimension(:), allocatable, save :: iwei
      real(dop), dimension(:,:), allocatable, save   :: wpar

!-------------------------------------------------------------------
! CONTRACTED MODE
!
! nodbef    : number of modes before the contracted mode
! nodaft    : number of modes after  the contracted mode
! dimbef    : potdim(1)* ... * potdim(nodbef)
! dimmodc   : gdim(modc)
! dimaft    : potdim(nodbef+2)* ... * potdim(ndof)
!
! NOTE:
!    nodbef + 1 + nodaft      = ndof
!    dimbef * dimmodc * dimaft = coeffdim
!-------------------------------------------------------------------
      integer(long), save  :: nodbef,nodaft,dimmodc,dimaft
      integer(long), save  :: dimbef

!-----------------------------------------------------------------------
! VARIABLES FOR 'ITERATION' MODUS
!  relnw[m]: trace-sum of eigenvalues of mode m
!-----------------------------------------------------------------------
      real(dop), save                            :: facit
      real(dop), dimension(:), allocatable, save :: relnw

!-----------------------------------------------------------------------
! VARIABLES SPECIFYING RELEVANT REGIONS
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable, save :: xxmin,xxmax,cxmin,&
                                                    cxmax,sxmin,sxmax

contains

!#######################################################################
!
! RUNPFITMOD_ALLOC
!
! memory required for arrays in runpfitmod allocated here.
!
! First check if allocated, then allocate. All arrays must be deallocated in
! runpfitmod_dealloc
!
!#######################################################################

      subroutine alloc_runpfitmod

use maxv
use channels

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

      integer(long), parameter                   :: maxwpar = 6

! check all arrays are not allocated
      call dealloc_runpfitmod

      cerr = 'potdim'
      allocate(potdim(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'wdim'
      allocate(wdim(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zevec'
      allocate(zevec(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zweight'
      allocate(zweight(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'cldim'
      allocate(cldim(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zcl'
      allocate(zcl(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'vgpdim'
      allocate(vgpdim(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'vgmdim'
      allocate(vgmdim(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zort1'
      allocate(zort1(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gdim1'
      allocate(gdim1(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'goff'
      allocate(goff(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'isppoint'
      allocate(isppoint(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'c2v'
      allocate(c2v(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'v2c'
      allocate(v2c(maxdim,6),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'iwei'
      allocate(iwei(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'wpar'
      allocate(wpar(maxwpar,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'relnw'
      allocate(relnw(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xxmin'
      allocate(xxmin(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xxmax'
      allocate(xxmax(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'cxmin'
      allocate(cxmin(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'cxmax'
      allocate(cxmax(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'sxmin'
      allocate(sxmin(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'sxmax'
      allocate(sxmax(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_RUNPFITMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      write(6,'(a)') 'Error allocating memory in ALLOC_RUNPFITMOD :',cerr
      write(6,'(a,i5)') 'Error Code :',ierr

      stop
      end subroutine alloc_runpfitmod

!#######################################################################

      subroutine dealloc_runpfitmod

      if (allocated(potdim)) deallocate(potdim)
      if (allocated(wdim)) deallocate(wdim)
      if (allocated(zevec)) deallocate(zevec)
      if (allocated(zweight)) deallocate(zweight)
      if (allocated(cldim)) deallocate(cldim)
      if (allocated(zcl)) deallocate(zcl)
      if (allocated(vgpdim)) deallocate(vgpdim)
      if (allocated(vgmdim)) deallocate(vgmdim)
      if (allocated(zort1)) deallocate(zort1)
      if (allocated(gdim1)) deallocate(gdim1)
      if (allocated(goff)) deallocate(goff)
      if (allocated(isppoint)) deallocate(isppoint)
      if (allocated(c2v)) deallocate(c2v)
      if (allocated(v2c)) deallocate(v2c)
      if (allocated(iwei)) deallocate(iwei)
      if (allocated(wpar)) deallocate(wpar)
      if (allocated(relnw)) deallocate(relnw)
      if (allocated(xxmin)) deallocate(xxmin)
      if (allocated(xxmax)) deallocate(xxmax)
      if (allocated(cxmin)) deallocate(cxmin)
      if (allocated(cxmax)) deallocate(cxmax)
      if (allocated(sxmin)) deallocate(sxmin)
      if (allocated(sxmax)) deallocate(sxmax)

      return
      end subroutine dealloc_runpfitmod

!-----------------------------------------------------------------------
      end module runpfitmod

