!#######################################################################
!
!                    runprop.inc
!
!#######################################################################

      module runpropmod

use decimal, only: long, dop
use global, only: c2
use mltypes
use maxv
use griddatmod, only: nmode
use channels
use datenmod
use global, only: lrunml

      implicit none

!-----------------------------------------------------------------------
! Propwf timing identifier
!-----------------------------------------------------------------------
       integer(long), save :: tpropwf

!-----------------------------------------------------------------------
! ARRAY SIZES
!
! intpsidim: Length of complex array needed by the integrators
! intrdim  : Lenght of real array needed by rDAV, SIL
! mfsumdim : Length of array containing the sum over the mean-field
!            matrices times the single-particle operator matrices
!            mfsumdim = Sum_s Sum_m [subdim(m)*dim(m,s)**2]
! spfovrdim: Length of array containing overlaps of spfs
!            spfovrdim = Sum_m Sum_s Sum_s1 [dim(m,s1) dim(m,s)]
! pconsdim : length of arrays holding momentum constraint matices
!-----------------------------------------------------------------------
      integer(long), save :: intpsidim,mfsumdim,spfovrdim,pconsdim,&
                             intrdim

!-----------------------------------------------------------------------
! VARIABLES DEFINING INTEGRATOR ALGORITHM
!
! intord: maximum order of integrator (used when "lconst" is false)
! inttyp : integrator type (used when "lconst" is false)
! IntTypeA/Phi/GWP: integrator used to propagate the A-/phi-vector/GWPs.
! IntOrderA/Phi/GWP: order of integration (the exact meaning of this
!                variable depends on the integrator).
!
! inttyp = N : where N can be:
!   1: Adams-Bashforth-Moulton predictor-corrector
!   2: Bulirsch-Stoer with polynomial extrapolation
!   3: Short iterative Lanczos or Lanczos-Arnoldi with variable order
!   4: Runge-Kutta integrator order 5
!   5: Runge-Kutta integrator order 8
!   6: Davidson "integrator" for improved relaxation.
!   7: LSODA adaptive integrator
!   8: Chebyschev integrator for exact propagations
!   9: ZVODE adaptive integrator
!  10: Verlet fixed step integrator for GBFs and trajectories
!  41: Runge-Kutta integrator order 5 FIXED TIMESTEP    
!  69: SOD fixed step integrator for exact propagations
!  70: SPO fixed step integrator for exact propagations
! maxsilorder: Maximum order that was actually used by SIL integrator
! accuracy: accuracy required for integrator
! accavec: integrator accuracy used for propagation of A-vector
! accphi: integrator accuracy used for propagation of phi-vectors
! acclevel : level of integrator accuracy (0 = default / user specified)
!                                         (1 = tight )
!                                         (2 = medium (tight*10) )
!                                         (3 = medium (tight*100) )
!
! For MULTI-LAYER CMF integrations, each mode can use its own integrator.
! The following arrays hold that information after parsing input.
! This information is later used to initialize the F90 integrators.
! minttyp: integrator type
! mintord: (maximum) integrator order
! mintacc: accuracy
! mintdt : initial stepsize
! mspec  : mode specifications from the INTEGRATOR-SECTION (e.g. "@2");
!          these are parsed later and mapped onto nodes in the tree
! mnspec : number of such mode specifications occuring in inputfile
! minrk5 : minimum step size of RK5 integrator
! maxeigval : maximum eigenvalue of PES. Required for SOD integrator.
! mineigval : minimum eigenvalue of PES. Required for Chebyshev integrator.
!-----------------------------------------------------------------------
      integer(long), save :: inttyp,intord,inttypea,inttypephi,intordera, &
                             intorderphi,maxsilorder,rlaxnum,&
                             phicount,iflagzvode,acclevel
      real(dop), save     :: accuracy,accavec,accphi,minrk5
      logical(kind=4),save:: lzvodecrit,lexpacc
      integer(long), save :: zvodemaxit
      real(dop), save     :: zvodeminstp
      real(dop), save     :: zvodemaxstp
      real(dop), save     :: zvode1ststp
      real(dop), save     :: maxeigval = -999.0d0
      real(dop), save     :: mineigval = 999.0d0

      integer(long), save :: mnspec
      integer(long), dimension(:), allocatable, save     :: minttyp
      integer(long), dimension(:), allocatable, save     :: mintord
      real(dop), dimension(:), allocatable, save         :: mintacc
      real(dop), dimension(:), allocatable, save         :: mintdt
      character(len=64), dimension(:), allocatable, save :: mspec

!-----------------------------------------------------------------------
! lanczorder: Number of Lanczos iterations in a diagonalisation run
! diagham_type   algorithm used to diagonalise Hamiltonian
!              = 0   lanczos
!                1   full diagonalisation
!-----------------------------------------------------------------------
      integer(long), save :: lanczorder = 0
      integer(long), save :: diagham_type = 0

      integer(long), dimension(:), allocatable, save :: node2spec

!-----------------------------------------------------------------------
! TIME VARIABLES
!
!  dt, dtA, dtPhi :  Initial integration step size(s) in fs
!  UpdateTime: Frequency of update of mean-fields in fs
!-----------------------------------------------------------------------
      real(dop), save :: dt,UpdateTime,dtA,dtPhi

!-----------------------------------------------------------------------
! LOGICALS TO CONTROL METHOD OF PROPAGATION
!
! If .true.
!  lconst    : mean-fields etc. are kept constant (= lconstfix .or.
!              lconstvar)
!  lconstfix : mean-fields are kept constant with fixed update time
!  lconstvar : mean-fields are kept constant with variable update time
!  lcmfadaphi: mean-fields are kept constant with adaptive update times
!              and error control only for spfs.
!  lcmfadaa  : mean-fields are kept constant with adaptive update times
!              and error control only for A-coefficients.
!  lfunkl    : mean-fields etc. are NOT kept constant and funk
!              subroutine is employed, i.e. uncorrelated operators 
!              are on the left of the projector
!  lfunkr    : mean-fields etc. are NOT kept constant and funkr
!              subroutine is employed, i.e. uncorrelated operators
!              are on the right of the projector
!  loldsilerr: if true, the standard SIL error criterion is used,
!              otherwise an improved one is taken
!  lnaturcmf : density matrix is re-diagonalized after each CMF-step
!  lenorb    : energy orbitals will be used
!  ltindspf  : SPFs are kept constant (time-independent)
!  lmf       : if true, linear mean-field approach is used,  
!              otherwise the Dirac-Frenkel variational principle
!  limportho : if true, the SPFs' orthonormality is improved after each
!              integrator step (general)
!  lrhonrm   : if true, the error measure of an RK integrator includes
!              the density rho (dicht2).
!  leinint   : read the integrator section, despite "cont" 
!  lwferror  : wavefunction error used for the integration error
!  loptcntrl : Optimal control run. Exchange of data with efield
!  loctpc    : Predictor-Corrector form of OCT (optimal control theory)
!  lsimprop  : Optimal control: do not use psi-file but propagate
!              forward and backward simultaneously (2 runs of MDTCH)
!  loctupdate: If lsimprop is true, loctupdate identifies the instance of
!              MDTCH which is propagated with the new 
!              (to be updated) field (Optimal control).
!  lrealphi  : Takes the real part of the SPFs in a impr.relax. run (rDAV).
!  lmlcmfsplit: In ML-MCTDH CMF propagations, the SPFs for different nodes
!              can be propagated together (.false.) or separately (.true.)
!  liork     : if true, the SPF's orthonormality is improved after each
!              (small) Runge-Kutta integrator step
!  lrktest   : if true, a test for potential problems is made after each
!              (small) Runge-Kutta integrator step
!  allcomplete : Freezes all SPFs
!-----------------------------------------------------------------------
      logical(kind=4), save :: lconst,lfunkl,lfunkr,lconstfix,lconstvar,&
                               lcmfadaphi,loldsilerr,lenorb,lcmfadaa,lmf,&
                               limportho,leinint,loptcntrl,loctpc,lsimprop,&
                               loctupdate,lwferror,lnaturcmf,lrealphi,&
                               lrhonrm = .false.
      logical(kind=4), save :: ltindspf = .false.
      logical(kind=4), save :: liork = .true.
      logical(kind=4), save :: lrktest = .false.
      logical(kind=4), save :: lmlcmfsplit = .true.
      logical(kind=4), save :: allcomplete = .false.

!-----------------------------------------------------------------------
! SPECIAL PARAMETERS PASSED TO FUNK ROUTINES
! rlxenergy:   Energy subtracted from Hamiltonian in relaxation
!              calculation.
! rlxshift :   Energy shift used in subroutine wrtrlxdav
! estart   :   Initial energy
! estart2  :   Second Initial energy (e.g. in type I density matrices
!              this is Tr(p^2 H))
! rlxemin  :   minimal energy bound for selecting states in DAV.
! rlxemax  :   maximal energy bound for selecting states in DAV.
! rlximin  :   minimal energy bound for imaginary part of selected state in cDAV
! lsil:        Tells "funk" routines, that they are called from SIL
! lipictrafo:  If true (and lipic is true), the "funk" routine gives
!              back dtpsi in interaction picture, else in
!              Schroedinger picture.
! lopro:       If true, the new "opro" projector routine is used, else
!              the old "projector" routine is taken.
! currentmode: The mode that is currently integrated (only used when
!              mean-fields are held constant).
! currentnode: The tree for the middle layer node that is currently 
!              integrated (only used when mean-fields are held constant).
! lgetenergy:  If true, the energy is re-computed in "funkex".
! usehsym   :  If true, operator symmetries will be used in building 
!              hteil matrices
!  lquadphi :  Use funkphi2 for improved relaxation.
!  lbackr   :  Back-rotate the SPFs before diagonalisation (impr.relax.). 
!  lquad    :  Diagonalise H^2 rather than H. (DAV impr.relax.). 
!  lortho   :  Re-orthogonalize Lanczos vectors (SIL impr.relax.). 
!  lolsen   :  Use Olsen correction (DAV impr.relax.). 
!  lskip1dav:  Skip the first Davidson diagonalization, start with
!              orbital relaxation.
!  dores    : force restart of integrator at next start of integration
!             loop
!  lsvd     :  Use SVD-regularization. See Meyer and Wang, JCP_148_124105
!-----------------------------------------------------------------------
      real(dop), save       :: rlxenergy,estart,estart2
      logical(kind=4), save :: lipictrafo,lopro,lsil,&
                               usehsym,lquadphi,lbackr,lortho,lquad,lolsen,&
                               lskip1dav,dores,&
                               lsvd=.false.
      logical(kind=4), save :: lgetenergy = .true.
      integer(long), save :: currentmode,precon,precon2
      type(wfnode_t), pointer, public, save     :: currentnode
      real(dop), save       :: rlxemin = -9.9e99_dop, &
                               rlxemax = 9.9e99_dop, &
                               rlximin = -9.9e99_dop, &
                               rlxshift = 0.0_dop

!-----------------------------------------------------------------------
! ACCURACY VARIABLES
!
! epsmat: value used for regularisation of dichte matrix eigenvalues
! epsupdate: accuracy used for adaptive constant mean-fields
! psicut : if psi saved in compact form, coefficient cutoff
! epsgwp: value used for regularisation of GWP C-matrix eigenvalues
! epsdyn: value used to define time-dependent GWP parameters
! epsld:  value used to determine linear dependence in GWP overlap
! linchk: if true linear dependence check is made
! ovlthr: threshold for determining whether eigenvalues of overlap
!         matrix too small
! tikh_fac: factor for Tikhonov regularization
! svd_fac: factor for SVD determination of Lambda dot (vMCG)
! lcysinv: set true to use the old S^{-1} way of generating C and Y
! lsmatovl: true to use S-matrix overlaps to determine Lin. Dep., false
!           to use eigenvalues
! epssvd: value used for regularisation of SVD singular values 
!         (new regularization scheme)
! lrmnoise: If set, small numbers (noise) removed from dtpsi.
!-----------------------------------------------------------------------
      real(dop), save :: epsupdate,psicut
      real(dop), save :: epsmat   = 1.0e-8_dop
      real(dop), save :: epsgwp   = 1.0e-8_dop
      real(dop), save :: epsdyn   = 1.0e-6_dop
      real(dop), save :: epsld    = 5.0e-2_dop
      real(dop), save :: ovlthr   = 0.99_dop
  !    real(dop), save :: tikh_fac = 1.0e-12_dop
  !    real(dop), save :: svd_fac  = 1.0e-10_dop
      real(dop), save :: epsgsmat = 1.0e-8_dop
      real(dop), save :: epssvd   = 1.0e-16_dop

      logical(kind=4), save :: linchk=.false.,lcysinv=.true.,&
                               lsmatovl=.true.
      logical(kind=4), save :: lrmnoise=.false.

!-----------------------------------------------------------------------
!
! GWP OPTIONS
!
! gwpmodel: flag to set model used for GWP propagation
!           Controls propagation through gwpcxm(m), igwpdynm(m), gwpintordm(m)
!           -1 : Mixed models (e.g. some DOF vMCG, some DOF classical)
!                (control passes to gwpmodelm(m))
!            0 : Full vMCG coupling for all GWPs
!            1 : All GWPs follow classical trajectories
!            2 : GWPs are independent (overlaps ignored in propagation)
!            3 : All GWPs are time-independent
!            4 : BEL algorithm used
!            5 : Saddle point GWPs (classical + gwpintordm(m)=0)
! gwpmodelm(m): gwpmodel for mode m. Passes control to gwpcxm, igwpdynm, ...
!
! gwpeoms:   flag to set how EOMS are solved  (C Lamdot = Y)
!          0 Inversion of C-matrix using eigenvalues and regularisation (default)
!          2 LU decomposition
!          3 Conj. Gradient
!          4 Direct inversion of C-matrix
!          5 Tikhonov
!          6 SVD
!          7 Cholesky
!          8 clsq
!          9 Inversion of C-matrix using eigenvalues and projection 
!
! igwpdynm(m): 0 All GWP parameters coupled (C^-1 Y calculated in full)
!            : 1 dynamical allocation of coupled GWP parameters
!              (C^-1 Y only calculated for coupled parameters)
!            : 2 No coupling (C^-1 Y not calculated)
!    In conjunction with gwpcx can generate classical or time-independet GWPs.
! ngwpdyn: No. of dynamic gwp parameters
! gwpdyndim: dimension of gwpdyn array
! lwholeg: "All or nothing" for including parameters associated with a Gaussian
! lresint: When the dynamic parameters change, restart the integrator

! zgwpdyn(m,s): pointer for array flagging "dynamic" GWP parameters
!
! maxtgdyn(m,s): Time at which basis set is fully dynamic
! maxngdyn(m,s): Max. no. of dynamic basis functions
! mincgdyn(m,s): Smallest value of Diag(C) of basis functions
! maxngdep(m,s): Max. no. of linear dependent basis functions found
! maxogdep(m,s): Max. overlap between 2 GWPs during propagation
! maxtgdep(m,s): Time at which maximum overlap found
!
! if lgwpoptyp = .true. GWPs are moved to minimize Y matrix at start
! if lgwpoptdyn = .true. GWPs are optimised using a short initial 
!                        propagation.
! ntotovl : the number of functions that are decoupled due to a large
!           overlap with another GWP
!
! lweightggp : weighted GGPs are output 
! ngwpovr(m,s): counts the no. of low eigenvalues in overlap matrix
!               when using SVD
! ovrinv : 0   invert overlap matrix using regularisation
!        : 1   direct inversion of overlap matrix
!        : 2   invert overlap matrix using SVD
! ltrjsym: If true the EOMs for GWP centres are symmetrised
!
!-----------------------------------------------------------------------
      integer(long), save :: gwpmodel = 0
      integer(long), save :: ovrinv = 2
      integer(long), save :: gwpeoms = 5
      integer(long), save :: gwpdyndim = 1
      integer(long), dimension(:), allocatable, save   :: gwpmodelm,&
                                                          igwpdynm
      integer(long), dimension(:,:), allocatable, save :: ngwpdyn
      integer(long), dimension(:,:), allocatable, save :: zgwpdyn
      integer(long), dimension(:,:), allocatable, save :: maxngdep
      integer(long), dimension(:,:), allocatable, save :: maxngdyn
      real(dop), dimension(:,:), allocatable, save     :: maxogdep
      real(dop), dimension(:,:), allocatable, save     :: maxtgdep
      real(dop), dimension(:,:), allocatable, save     :: mincgdyn
      real(dop), dimension(:,:), allocatable, save     :: maxtgdyn
      integer(long), dimension(:,:), allocatable, save :: ngwpovr
      logical(kind=4), save :: lwholeg = .false.
      logical(kind=4), save :: lresint = .false.
      logical(kind=4), save :: lgwpoptyp = .false.
      logical(kind=4), save :: lgwpoptdyn = .false.
      logical(kind=4), save :: lweightggp = .false.
      logical(kind=4), save :: lrenormgwpa = .true.
      logical(kind=4), save :: ltrjsym = .false.
      logical(kind=4), dimension(:,:,:), allocatable, save :: lcapgwp

!-----------------------------------------------------------------------
! these are for BEL-MCG (WK,TJF, PRL 2012)
!-----------------------------------------------------------------------
      real(dop), save                      :: BELint      ! leap interval (0)
      real(dop), save                      :: BELoff      ! leap offset (0)
      integer(long), save                  :: BELprint    ! number of points to print (debug)
      real(dop), save, allocatable         :: BELplotlims(:,:) ! limits 
      integer(dop), save                   :: reCompDtPsi ! flag whether DtPsi should be recomputed
      complex(dop), dimension(:), allocatable, save :: saveworkc   ! storage for S^-1 H
                  ! this is used to keep workc across funk(r) calls if BELflags(7)/=0
      real(dop), save                               :: BELflags(14)
         ! default values givin in parenthesis
         ! 1 : maximum deviation at end of reassignment (1e-2)
         ! 2 : \epsilon in EQ.(4), i.e. minimum significant density (1e-2)
         ! 3 : \epsilon in EQ.(5), i.e. LHA compliance check (1e-2)
         ! 4 : extra GWPs after limit BELflags(1) is reached (40)
         ! 5 : absolute minimum width of GWPs (1e-3)
         ! 6 : absolute maximum width of GWPs (3e-0)
         ! 7 : reuse S^-1 H matrix in funk calls (1)
         ! 8 : for minimization 3: 1.. minimize only P, 2 minimize P,sigma (2)
         ! 9 : number of steps to take in minimizations (4000)
         ! 10: standard width of gaussians is width of wave packet times this (1)
         ! 11: smallest accepted overlap eigenvalue before switching off projection inversion (1e-5)
         ! 12: weight of deviation contribution to Q selection; omega of (4) of the BEL PRL
         ! 13: number of steps to take in minimizations (10)
         ! 14: calculation of expectation value (0) or skip (BELflags(14)/=0)

!-----------------------------------------------------------------------
! for application of LHA for gwp mode
!
! lforcelha : If true all terms treated using an LHA.
! lha_enerr : cummulative error in energy due to LHA
!
!-----------------------------------------------------------------------
      real(dop), save     :: gwpentol = 1.0d-4
      real(dop), save     :: lha_enerr = 0.0_dop
      real(dop), save     :: avgwpscale = 0.0_dop, &
                             maxgwpscale = 0.0_dop
      integer(long), save :: maxgwpenit = 0, &
                             totgwpenit = 0
      logical(kind=4)     :: lforcelha = .false.

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS
!
! zmfsum(m,s1,s): pointer to start of mfsum for mode m and states s,s1
! zspfovr(m,s): pointer to start of overlap matrix of single-particle
!               functions <phi(m,s)|phi(m,s)>
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save       :: zpcons
      integer(long), dimension(:,:), allocatable, save     :: zspfovr
      integer(long), dimension(:,:,:), allocatable, save   :: zmfsum
      integer(long), dimension(:,:,:,:), allocatable, save :: zmfsum1

!-----------------------------------------------------------------------
! POINTERS FOR ARRAYS (density operators of type I)
!
! dzmfsum(m,s1,s,t): pointer to start of mfsum for mode m and states s,s1
!-----------------------------------------------------------------------
      integer(long),dimension(:,:,:,:), allocatable, save :: dzmfsum

!-----------------------------------------------------------------------
! Energy arrays for density operator programs
!
! detot            : total energy
! decorr(maxsta)   : correlated energy per state
! deuncorr(maxsta) : uncorrelated energy per state
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable, save :: detot,decorr,deuncorr

!----------------------------------------------------------------------
! Character variables. 
! rlxunit is energy-unit used - needed in output and wrtrlxdav 
! rlxtime is time-unit used - needed in output
!-----------------------------------------------------------------------
      character(len=8), save :: rlxunit='ev'
      character(len=8), save :: rlxtime='fs'
      character(len=64), save :: optstring

!-----------------------------------------------------------------------
! edav: Davidson energy passed from Davidson routines to function routines
! kskip: used in Davidson 
!-----------------------------------------------------------------------
      real(dop), save :: edav
      integer(long), save :: kskip

!-----------------------------------------------------------------------
!  Variables for stop control.
!     lnormstop: if true, the keyword 'normstop' in run-section is used
!     nrmstop:   value of normstop (propagation stopped if norm^2 < nrmstop)
!     constop:   value for converged (impr. relax.)
!     lnatpopstop: true if propagation should be stopped when the lowest
!                nat.pop. of one mode (or all modes) exceeds a threshold
!     natpopstop: value of this threshold
!     natpopstopmode/state: number of this mode/state
!        0 means: threshold must be exceeded for all modes/states
!       -1 means: threshold must be exceeded for any one mode/state
!-----------------------------------------------------------------------

      real(dop), save       :: nrmstop = 1.0d-12, &
                               constop, e1stop, e2stop, natpopstop
      integer(long), save   :: natpopstopmode, natpopstopstate
      logical(kind=4), save :: lnormstop = .true., &
                               lstop, lnatpopstop

!-----------------------------------------------------------------------
! Logical to allow read only Srfhop run
!-----------------------------------------------------------------------
      logical(kind=4), save :: lrdsrfhop = .false.

!-----------------------------------------------------------------------
! OPTIONA controlling the initial WF
!  initnorm : initial norm of WF will be initnorm
!-----------------------------------------------------------------------
      real(dop), save :: initnorm = -999.0

!-----------------------------------------------------------------------
! ARRAYS NEEDED FOR EXPECTATION VALUES
!
! expectlab() : labels for the expectation operators
! expvalue()  : expectation values
!-----------------------------------------------------------------------
      complex(dop), dimension(:), allocatable, save :: expvalue,expvalue1

      contains

!#######################################################################
!
! RUNPROPMOD_ALLOC
!
! memory required for arrays in runpropmodmod allocated here.
!
! First check if allocated, then allocate. All arrays must be
! deallocated in
! runpropmod_dealloc
!
!#######################################################################

      subroutine alloc_runpropmod

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_runpropmod

      cerr = 'gwpmodelm'
      allocate(gwpmodelm(nmode),stat=ierr)
      if (ierr > 0) go to 1000
      gwpmodelm(:)=0   ! default is vMCG
      cerr = 'igwpdyn'
      allocate(igwpdynm(nmode),stat=ierr)
      if (ierr > 0) go to 1000
      igwpdynm(:) = 1  ! default is unsymm integrals
      cerr = 'ngwpdyn'
      allocate(ngwpdyn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zgwpdyn'
      allocate(zgwpdyn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'maxngdep'
      allocate(maxngdep(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'maxngdyn'
      allocate(maxngdyn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'maxogdep'
      allocate(maxogdep(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'maxtgdep'
      allocate(maxtgdep(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mincgdyn'
      allocate(mincgdyn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'maxtgdyn'
      allocate(maxtgdyn(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ngwpovr'
      allocate(ngwpovr(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'lcapgwp'
      allocate(lcapgwp(maxdim,maxgwp,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zpcons'
      allocate(zpcons(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zspfovr'
      allocate(zspfovr(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'zmfsum'
      allocate(zmfsum(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      if (dentype .eq. 1) then
         cerr = 'dzmfsum'
         allocate(dzmfsum(maxdim,maxsta,maxsta,maxsta),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      cerr = 'zmfsum1'
      allocate(zmfsum1(maxdim,maxsta,maxsta,maxnhtmshift),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'detot'
      allocate(detot(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'decorr'
      allocate(decorr(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'deuncorr'
      allocate(deuncorr(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      if (nexpect .gt. 0) then
         cerr = 'expvalue'
         allocate(expvalue(nexpect),stat=ierr)
         if (ierr > 0) go to 1000
      endif
      if (nexpect1 .gt. 0) then
         cerr = 'expvalue'
         allocate(expvalue1(nexpect1),stat=ierr)
         if (ierr > 0) go to 1000
      endif

      if (lrunml) then

         cerr = 'minttyp'
         allocate(minttyp(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

         cerr = 'mintord'
         allocate(mintord(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

         cerr = 'mintacc'
         allocate(mintacc(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

         cerr = 'mintdt'
         allocate(mintdt(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

         cerr = 'mspec'
         allocate(mspec(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

         cerr = 'node2spec'
         allocate(node2spec(2*maxdim),stat=ierr)
         if (ierr > 0) go to 1000

      endif

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_RUNPROPMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_runpropmod

!#######################################################################

      subroutine dealloc_runpropmod

      if (allocated(gwpmodelm)) deallocate(gwpmodelm)
      if (allocated(igwpdynm)) deallocate(igwpdynm)
      if (allocated(ngwpdyn)) deallocate(ngwpdyn)
      if (allocated(zgwpdyn)) deallocate(zgwpdyn)
      if (allocated(maxngdep)) deallocate(maxngdep)
      if (allocated(maxngdyn)) deallocate(maxngdyn)
      if (allocated(mincgdyn)) deallocate(mincgdyn)
      if (allocated(maxtgdyn)) deallocate(maxtgdyn)
      if (allocated(ngwpovr)) deallocate(ngwpovr)
      if (allocated(lcapgwp)) deallocate(lcapgwp)
      if (allocated(zpcons)) deallocate(zpcons)
      if (allocated(zspfovr)) deallocate(zspfovr)
      if (allocated(zmfsum)) deallocate(zmfsum)
      if (allocated(zmfsum1)) deallocate(zmfsum1)
      if (allocated(dzmfsum)) deallocate(dzmfsum)
      if (allocated(detot)) deallocate(detot)
      if (allocated(decorr)) deallocate(decorr)
      if (allocated(deuncorr)) deallocate(deuncorr)
      if (allocated(expvalue)) deallocate(expvalue)
      if (allocated(expvalue1)) deallocate(expvalue1)

      if (allocated(minttyp)) deallocate(minttyp)
      if (allocated(mintord)) deallocate(mintord)
      if (allocated(mintacc)) deallocate(mintacc)
      if (allocated(mintdt)) deallocate(mintdt)
      if (allocated(mspec)) deallocate(mspec)

      return
      end subroutine dealloc_runpropmod


      end module runpropmod
