! **********************************************************************
!                                                                      
!                   module PROP
!                                                                      
!  QUANTICS interface module controlling the propagation of the 
!  wavepacket using the ZagHop surface hopping code.
!                                                                      
! **********************************************************************

      module prop_zagreb

use versions
use decimal, only: long, dop
use global
use globmemmod
use paths
use channels
use maxv

use timing
use logdat
use datenmod
use runpropmod
use dvrdatmod
use griddatmod
use operdef
use psidef

use rddvrmod
use rdopermod
use iorst
use iopsidef, only: psidat
use dirdyn
use logintmod

use dd_db, only: getdbnrec,ddq2x, ddx2q, ddv2p
use directdyn, only: getddpes
use prop, only: zeitstop

use zeigproptrmod
use memproptrmod
use opendatmod, only: opendat,writestop
use outputtrmod, only: outputtr
use calcdattrmod, only: epot_shz,ekin_shz
use op1lib, only: cpvxz
use op2lib, only: addvxxdo,subvxxdo1
use xvlib, only: mvtxdd1,mvxxdd1

use omp_lib
use openmpmod

use mpidata

      implicit none

!-----------------------------------------------------------------------
! Options for analysis
!-----------------------------------------------------------------------
      logical(kind=4), dimension(3), save :: lsrfchk=.false.
      real(dop), save                     :: deltae = 0.01 / 27.2114

      private
      public :: runprop_zagreb
      contains

! **********************************************************************
! *** V6.2 MB *********> SUBROUTINE PROPAGATE PSI <*********************
!                       ~~~~~~~~~~~~~~~~~~~~~~~~~~
! Controls the propagation of the wavefunction.
! **********************************************************************

      subroutine runprop_zagreb 

      implicit none

      character(len=80)                          :: actual

      integer(long)                              :: ilbl,jlbl,klbl,&
                                                    runlbl,&
                                                    chkdvr,chkgrd,chkpsi,&
                                                    chkprp
      integer(long)                              :: dirlbl = 0
      integer(long)                              :: ntraj1, nhopsav,nhopsmax

      integer(long)                              :: i,n,n1,ithr
      integer(long)                              :: noutput,istep,nstep
      character(len=c5)                          :: filename,trjdir,trjdir1
      character(len=c5)                          :: runcmd
      character(len=80)                          :: string
      logical(kind=4)                            :: lerr,check,linperr,&
                                                    linwf,lexist
      character(len=c2)                          :: rdtag
      complex(dop), dimension(:), allocatable    :: psi,tmppsi
      real(dop)                                  :: time,tstep,etotavav
      real(dop), dimension(:), allocatable       :: ingeo
      real(dop), dimension(:), allocatable       :: hops

      integer(long)                              :: ic,iz,ierr,aerr,idm
      integer(long), dimension(:), allocatable   :: tmptrajst
      real(dop), dimension(1)                    :: rdm,rdm1
      complex(dop), dimension(1)                 :: cdm

      character(len=20)                          :: cerr,atmp,atmp1
      logical(kind=4)                            :: lovwrtr = .false.
      logical(kind=4), dimension(10)             :: found = .false.
      logical(kind=4)                            :: lnadvec = .false.
      logical(kind=4), dimension(:), allocatable :: lskip

      integer(long), save                        :: tidsh=0 ! timer ID

      integer(long), external   :: integersize,realsize,complexsize,logicalsize
      real(dop), dimension(1)   :: rdum,rdum1
      character(len=72)                          :: stopcommand

!-----------------------------------------------------------------------
!  MPI not yet implemented
!-----------------------------------------------------------------------
#ifdef MPI
      if(mpirank.ne.master)then
        return
      endif
#endif

!-----------------------------------------------------------------------
! Initialise timing arrays
!-----------------------------------------------------------------------
      if (ltime) then
         call timinit
         call get_timer(tidsh, 'SH_Zagreb')
         call start_timer(tidsh)
      endif

      call fzeit(actual)
      i = index(actual,':')
      actual=actual(i-9:i+6)

      write(ilog,'(/,72a1)') ('#',i=1,72)
      if (lrdsrfhop) then
         write(ilog,'(/,a)') '   Reading Zagreb Surface Hopping '
      else
         write(ilog,'(/,a)') '   Running Zagreb Surface Hopping '
      endif
      write(ilog,'(57x,a)')  actual(1:15)
      write(ilog,'(72a1)') ('#',i=1,72)
      write(ilog,'(/,a,i5)') 'No. trajectories = ',ntraj
      if (ntraj0 .gt. 1) then
         write(ilog,'(/,a,i5)') 'Initial trajectory = ',ntraj0
      endif

! --- CHECK ---
      if( tfinal .le. tinit + tiny ) then
         write(ilog,'(a)') '*******************************'
         write(iout,'(/a)') 'Nothing left to do!'
         write(ilog,'(/a)') 'Nothing left to do!'
         write(ilog,'(a,f15.6,a/,a,f15.6,a/)') &
         't-init =', tinit/fs,' fs','tfinal =',tfinal/fs,' fs'
         lstop = .false.
         goto 200
      end if

!-----------------------------------------------------------------------
! Check input file open for reading
!-----------------------------------------------------------------------
      inquire(iin,opened=check)
      if (check) close(iin)
      filename=inputfile(1:inplaenge)
      open(iin,file=filename,status='old',form='formatted',err=900)

!-----------------------------------------------------------------------
! get array dimensions 
!-----------------------------------------------------------------------
      inquire(ioper,opened=check)
      if (check) close(ioper)
      filename=oname(1:olaenge)//'/oper'
      ilbl=index(filename,' ')-1
      open(ioper,file=filename(1:ilbl),form='unformatted',status='old',err=900)
      call rdmemdim(ioper)
      close(ioper)

!-----------------------------------------------------------------------
! Allocate mememory 
!-----------------------------------------------------------------------
      allocmemory=0
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_psidef
      call alloc_dirdyn
      call alloc_dddb
      call alloc_dercp

!-----------------------------------------------------------------------
! Read system / DVR information
!-----------------------------------------------------------------------
      filename=dname(1:dlaenge)//'/dvr'
      ilbl=index(filename,' ')-1
      open(idvr,file=filename,form='unformatted',status='old',err=900)
      chkdvr=1
      call dvrinfo(lerr,chkdvr)
      close(idvr)

!-----------------------------------------------------------------------
! Read data from oper file
!-----------------------------------------------------------------------
      filename=oname(1:olaenge)//'/oper'
      ilbl=index(filename,' ')-1
      open(ioper,file=filename,form='unformatted',status='old',err=900)
      chkdvr=2
      chkgrd=1
      call operinfo(lerr,chkdvr,chkgrd)
      close(ioper)

!-----------------------------------------------------------------------
! Read data from restart file
!-----------------------------------------------------------------------
      filename=rname(1:rlaenge)//'/restart'
      ilbl=index(filename,' ')-1
      open(irst,file=filename,form='unformatted',status='old',err=900)
      chkdvr=2
      chkgrd=2
      chkpsi=1
      chkprp=1
      call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

!-----------------------------------------------------------------------
! Read initial geoms from "wavefunction"
!-----------------------------------------------------------------------
      allocate(psi(dgldim))
      psi=0.0_dop
      rewind(irst)
      do
   10    continue
         read(irst,err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'psi') then
            read (irst,err=999) (psi(i),i=1,dgldim)
         else if (rdtag(ilbl:jlbl) .eq. 'end restart file') then
            exit

         endif
      enddo
  100 continue

      close(irst)

!-----------------------------------------------------------------------
! Log DD protocol
!-----------------------------------------------------------------------
      if (ldd) call logdd

!-----------------------------------------------------------------------
! End short test
!-----------------------------------------------------------------------
      if (test0) then
         write(iout,'(/,a,/)') 'Short Test run was successful.'
         write(ilog,'(/,a,/)') '***** Test run *****'
         write(ilog,'(a,/)')   'Short test run completed.'
         goto 600
      endif

!-----------------------------------------------------------------------
! Set up test
!-----------------------------------------------------------------------
      if (test) then
         ntraj = ntraj0
         nstep = 0
         write(ilog,'(/,a)') '***** Test run *****'
         write(ilog,'(a,/)') 'Single trajectory for 1 step'
      endif

! ------ WRITE RESTART FILE to add prop data
! Hack to write trajst array - was set up in geninwf (?) and should be read
! here rather than simply writing a temporary array.
      allocate(tmptrajst(gwpinfodim))
      tmptrajst(:) = isbaspar(4,feb,1)   ! put all trajectories on initial state
      linwf=.false.            !
      idm=0                    ! just dummies
      rdm(1)=0.0d0             !      "
      rdm1(1)=0.0d0             !      "
      cdm(1)=(0.0d0,0.0d0)     !      "
      call wrrst(time,psi,linwf,cdm,idm,rdm,rdm1,mi(mijindx), &
           agmat,tmptrajst)
      deallocate(tmptrajst)

!-----------------------------------------------------------------------
! For a DD run the DB has to be initialised with the first point
! before starting trajectories
!
! gwpdim, qcentdim needed in getddpes as traj pretends to be a GWP 
!-----------------------------------------------------------------------
      if (ldd .and. lddwrdb) then
         call getdbnrec(dbnrec)
         if (dbnrec .eq. 0) then
            allocate(hops(hopsdim))
            allocate(ingeo(nspfdof(1)))

            operfile=oname(1:olaenge)//'/oper'
            chkdvr=2
            chkgrd=1
            call rdoper(hops,chkdvr,chkgrd)

            gwpdim(1,1) = 1
            zcent(1,1) = 1
            gwpm(1)=.true.
            qcentdim = ndofddpes

            ingeo(:) = dble(psi(zetf(1,1):zetf(1,1)+nspfdof(1)-1))
            time=0.0_dop
            call getddpes(time,ingeo,1,1)
!            deallocate(ingeo,hops)
         endif
      endif
     
!-----------------------------------------------------------------------
!  Read options and
! --- GENERATE INPUT FILE
!-----------------------------------------------------------------------
      ilbl=index(qubindir,' ')-1
      runcmd = qubindir(1:ilbl)//'/run_shzagreb'
      runlbl=index(runcmd,' ')-1

! open Zagreb input file and write system information. Other keywords are
! read from quantics .inp file (dynamics.in section)
      filename=name(1:laenge)//'/dynamics.in'
      open(idump,file=filename,status='unknown',form='formatted',err=900)
      write(idump,'(a)') "$system"
      write(idump,'(a,i5)') " nstate =",nddstate
      write(idump,'(a,i5)') " istate =",isbaspar(4,feb,1)

      rewind(iin)
      linperr=.false.
      iz=0
! read to input section
      do
         iz=iz+1
         call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
         if (ierr .gt. 0) then
            routine = 'Prop_zagreb'
            message = inptit
            call errormsg
         endif

         if (keyword(1) .eq. 'sh_zagreb-section') then
            exit
         else if (keyword(1) .eq. 'end-input') then  ! Use defaults
            go to 150
         endif
      enddo 
            
      readiin: do
         iz=iz+1
         call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
         if (ierr .gt. 0) then
            routine = 'Prop_zagreb'
            message = inptit
            call errormsg
         endif

         if (keyword(1) .eq. 'run_cmd') then
            runcmd = keyorig(3)
            runlbl = index(runcmd,' ')-1
            call abspath(runcmd,runlbl)

         else if (keyword(1) .eq. 'ovwr_tr') then
            lovwrtr = .true.

         else if (keyword(1) .eq. 'ignore_frhops') then
            lsrfchk(1) = .true.
         else if (keyword(1) .eq. 'no_ignore_frhops') then
            lsrfchk(1) = .false.

         else if (keyword(1) .eq. 'ignore_vare') then
            lsrfchk(2) = .true.
            if (keyword(2) .eq. '=') then
               read(keyword(3),*) deltae
               deltae = deltae / ev
            endif
         else if (keyword(1) .eq. 'no_ignore_vare') then
            lsrfchk(2) = .false.

         else if (keyword(1) .eq. 'ignore_deltae') then
            lsrfchk(3) = .true.
         else if (keyword(1) .eq. 'no_ignore_deltae') then
            lsrfchk(3) = .false.

         else if (keyword(1) .eq. 'dynamics.in') then

            do
               read(iin,'(a)') string
               call keylowc(string)
               if (index(string,'end-dynamics.in') .ne. 0) then
                  exit

               else if (index(string,'end-sh_zagreb-section') .ne. 0) then
                  routine = 'Prop_zagreb'
                  message = 'No End-SH_Zagreb-section found '
                  call errormsg

               else if (string(1:1) .eq. '#') then
                  cycle

               else if (index(string,'$dynamics') .ne. 0) then
                  write(idump,'(a)') string
                  found(1) = .true.
                  do
                     read(iin,'(a)') string
                     call keylowc(string)
                     if ((index(string,'$') .ne. 0) .or. &
                        (index(string,'end-dynamics.in') .ne. 0)) then
                        backspace(iin)
                        exit
                     else if (index(string,'tstep') .ne. 0) then
                        found(8) = .true.
                        write(idump,'(a)') string
                        ! read and store tstep 
                        ilbl=index(string,'=')
                        if (ilbl .gt. 0) then   ! string includes "="
                           read(string,*) atmp,atmp1,tstep
                        else
                           read(string,*) atmp,tstep
                        endif

                     else if (index(string,'max_toten_d ') .ne. 0) then
                        found(9) = .true.
                        write(idump,'(a)') string
                     else if (index(string,'max_toten_d_step ') .ne. 0) then
                        found(10) = .true.
                        write(idump,'(a)') string

                     else
                        write(idump,'(a)') string
                     endif
                  enddo

                  ! quantics tfinal and tinit are in au. tstep in fs
                  if (test) then
                     write(idump,'(a)') " max_time = 0.0"
                  else
                     write(idump,'(a,f10.4)') " max_time = ",tfinal/fs
                  endif

                  ! if required keywords not been specified, add here
                  if (.not. found(8)) then
                     write(idump,'(a)') " tstep = 0.1 "
                  endif
                  if (lsrfchk(3) .and. .not. found(9)) then
                     write(idump,'(a)') " max_toten_d = 999.0 "
                  endif
                  if (lsrfchk(3) .and. .not. found(10)) then
                     write(idump,'(a)') " max_toten_d_step = 999.0 "
                  endif

               else if (index(string,'$method') .ne. 0) then
                  write(idump,'(a)') string
                  found(2) = .true.
                  do
                     read(iin,'(a)') string
                     call keylowc(string)
                     if ((index(string,'$') .ne. 0) .or. &
                        (index(string,'end-dynamics.in') .ne. 0)) then
                        backspace(iin)
                        exit
                     else
                        write(idump,'(a)') string
                     endif
                  enddo

               else if (index(string,'$surfhop') .ne. 0) then
                  write(idump,'(a)') string
                  found(3) = .true.
                  do
                     read(iin,'(a)') string
                     call keylowc(string)
                     if ((index(string,'$') .ne. 0) .or. &
                        (index(string,'end-dynamics.in') .ne. 0)) then
                        backspace(iin)
                        exit

                     ! Need either fssh .or. lz
                     else if (index(string,'fssh') .ne. 0) then
                        found(7) = .true.
                        lnadvec = .true.
                        write(idump,'(a)') string
                     else if (index(string,'lz') .ne. 0) then
                        found(7) = .true.
                        write(idump,'(a)') string

                     ! Need either nadvec or overlap
                     else if (index(string,'nadvec') .ne. 0) then
                        found(4) = .true.
                        lnadvec = .true.
                        write(idump,'(a)') string
                     else if (index(string,'overlap') .ne. 0) then
                        found(4) = .true.
                        write(idump,'(a)') string

                     ! Need ssed
                     else if (index(string,'seed') .ne. 0) then
                        found(5) = .true.
                        write(idump,'(a)') string

                     else
                        backspace(iin)
                        read(iin,'(a)') string
                        write(idump,'(a)') string
                     endif
                  enddo

                  ! if required keywords not been specified, add here
                  if (.not. found(7)) then
                     write(idump,'(a)') " fssh adiabatic"
                  endif
                  if (.not. found(4)) then
                     write(idump,'(a)') " nadvec linear"
                     lnadvec = .true.
                  endif
                  if (.not. found(5)) then
                     write(idump,'(a)') " seed -1"
                  endif

               else if (index(string,'$output') .ne. 0) then
                  write(idump,'(a)') string
                  found(6) = .true.
                  do
                     read(iin,'(a)') string
                     call keylowc(string)
                     if ((index(string,'$') .ne. 0) .or. &
                        (index(string,'end-dynamics.in') .ne. 0)) then
                        backspace(iin)
                        exit
                        exit
                     else
                        write(idump,'(a)') string
                     endif
                  enddo

! $end will be added later - ignored here
               else if (index(string,'$end') .ne. 0) then
                  continue

               else
                  write(idump,'(a)') string
               endif
            enddo

         else if (keyword(1) .eq. 'end-sh_zagreb-section') then
            exit

         endif

      enddo readiin

  150 continue
!-----------------------------------------------------------------------
! Add any missing sections to ZagHop input
!-----------------------------------------------------------------------
      if (.not. found(1)) then
         write(idump,'(a)') "$dynamics"
         write(idump,'(a)') " tstep = 0.1 "
         tstep = 0.1
         if (test) then
            write(idump,'(a)') " max_time = 0.0"
         else
            write(idump,'(a,f10.4)') " max_time = ",tfinal/fs
         endif
         if (lsrfchk(3)) then
            write(idump,'(a)') " max_toten_d = 999.0 "
            write(idump,'(a)') " max_toten_d_step = 999.0 "
         endif
      endif
      if (.not. found(2)) then
         write(idump,'(a)') "$method"
         ilbl=index(qubindir,' ')-1
         write(idump,'(a)') &
                " qlib quantics"
      endif
      if (.not. found(3)) then
         write(idump,'(a)') "$surfhop"
         write(idump,'(a)') " fssh adiabatic"
         write(idump,'(a)') " nadvec linear"
         lnadvec =.true.
         write(idump,'(a)') " seed -1"
      endif
      if (.not. found(6) .and. lnadvec) then   ! Needed for Nadvec FSSH
         write(idump,'(a)') "$output"
         write(idump,'(a)') " noprint  7,10,11"
      endif

      write(idump,'(a)') "$end"
      close(idump)

      write(ilog,'(/,a)') '*** Script to run Code ***'
      write(ilog,'(a)') runcmd(1:runlbl)

!-----------------------------------------------------------------------
! If analysis only, skip over running trajectories
!-----------------------------------------------------------------------
      if (lrdsrfhop) then
         goto 200
      endif

!-----------------------------------------------------------------------
! Open stop file
!-----------------------------------------------------------------------
     if (lstopf) then
         filename=name(1:laenge)//'/stop'
         open(istop,file=filename,form='formatted',status='unknown', &
              err=900)

         call writestop
     endif

!-----------------------------------------------------------------------
! check for existence of, or set up, directory for trajectories
!-----------------------------------------------------------------------
      trjdir = name(1:laenge)//'/zagreb_trj'
      lexist=.false.
      dirlbl=index(trjdir,' ')-1
      open(idump,file=trjdir(1:dirlbl)//'/scratch',err=666)
      close(idump,status='delete')
      lexist=.true.
  666 continue
      if (.not. lexist) then
         filename='mkdir '//trjdir(1:dirlbl)//' '//char(zeroi)
         call excmd(filename)
         write(ilog,'(/,a)') '*** Creating Trajectory Data Directory***'
         write(ilog,'(a)') trjdir(1:dirlbl)
      else
         write(ilog,'(/,a)') '*** Writing Trajectory Data to Directory ***'
         write(ilog,'(a)') trjdir(1:dirlbl)
         write(ilog,'(a)') ' '
      endif

!-----------------------------------------------------------------------
! propagate trajectories
!-----------------------------------------------------------------------
      ilbl=index(qubindir,' ')-1
!$omp parallel &
!$omp firstprivate(lstop) &
!$omp private(n,jlbl,klbl,filename,trjdir1,ithr) &
!$omp num_threads(ompthread) if(lompthread)
!$omp do schedule(dynamic)
      do n=ntraj0,ntraj

#ifdef OMP
         ithr = omp_get_thread_num()
#else
         ithr = 0
#endif
! --- STOP IF REQUIRED BY STOP FILE (only on master thread)  ---
         if(lstopf .and. (ithr .eq. 0)) then  
            open(istop,file=name(1:laenge)//'/stop', &
                 form='formatted',status='old',err=676)
            read(istop,'(a)') stopcommand
!...........Go to the end of the stop-file and write the trajectory
!...........number to it in order to 'touch' the stop-file.
            do i = 1, 26
                read(istop,'(a)')
            end do
            write(istop,'(a,i5)') 'Trajectory: ',n
            close(istop)
            goto 677
! --- If the stop-file cannot be opened stop the calculation ---
 676        lstop = .true.
 677        if( stopcommand(1:4) .eq. 'stop' ) then
               lstop = .true.
            else if( stopcommand(4:4).eq.':' .or.  &
                     stopcommand(12:12).eq.':' .or. &
                     stopcommand(23:23).eq.':' ) then
               call zeitstop(stopcommand)
            end if
#ifdef OMP
            if(lstop) cycle ! cannot exit OMP loop 
#else
            if(lstop) exit
#endif
         end if          

! name for directory to contain data
         write(trjdir1,'(a,i7)') trjdir(1:dirlbl)//'/traj.',n
         call rmblnk(trjdir1)
         jlbl=index(trjdir1,' ')-1

! --- CHECK FOR PREVIOUS OUTPUT ---

         filename=trjdir1(1:jlbl)//'/geom'
         klbl=index(filename,' ')-1
         inquire(file=filename(1:klbl),exist=check)
         if (check .and. lovwrtr) then
            filename='rm -rf '//trjdir1(1:jlbl)//' '//char(zeroi)
            call excmd(filename)
         else if (check) then
            write(ilog,'(a,i7)') 'Trajectory already exists: ',n
            cycle
         endif
         write(ilog,'(a,i7)') 'Running trajectory: ',n

! Create directory and cp input file to it
         filename='mkdir '//trjdir1(1:jlbl)//' '//char(zeroi)
         call excmd(filename)

         filename='cp '//name(1:laenge)//'/dynamics.in ' &
                      //trjdir1(1:jlbl)//'/dynamics.in '//char(zeroi)
         call excmd(filename)

         if (ldd) then
            klbl = index(ddname0,'/')-1
! Allow dddata directory in the input file include a / (as in elk_inputs).
            if (klbl .gt. 0) then
               filename=&
                  'mkdir '//trjdir1(1:jlbl)//'/'//ddname0(1:klbl)//char(zeroi)
               call excmd(filename)
               filename=&
                 'cp -r '//ddname(1:ddlaenge)//' ' &
                    //trjdir1(1:jlbl)//'/'//ddname0(1:klbl)//char(zeroi)
               call excmd(filename)
            else
               filename=&
                 'cp -r '//ddname(1:ddlaenge)//' ' //trjdir1(1:jlbl)//char(zeroi)
               call excmd(filename)
            endif
         endif

!-----------------------------------------------------------------------
! generate initial coordinates + veloc
!-----------------------------------------------------------------------
#if OMP
   ithr = omp_get_thread_num()
#else
   ithr = 0
#endif
         call geningeo(trjdir1,psi(zetf(1,1)),nspfdof(1),ntraj,n,ithr)
!-----------------------------------------------------------------------
! --- CALL ZAGREB CODE
!-----------------------------------------------------------------------
         filename= &
             runcmd(1:runlbl+1)//trjdir1(1:jlbl+1)//qubindir(1:ilbl+1)//char(zeroi)
         call excmd(filename)

!-----------------------------------------------------------------------
! If no new data written to DB then delete copy
!-----------------------------------------------------------------------
         if (ldd .and. .not. lddwrdb) then
            klbl=index(ddname0,' ')-1
            filename=&
                 'rm -rf '// trjdir1(1:jlbl)//'/'//ddname0(1:klbl)
            call excmd(filename)
         endif
      enddo
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! --- ZAGREB CODE FINISHED
!-----------------------------------------------------------------------

 200  continue

      if (ltime) then
         call stop_timer(tidsh)
         timers(tidsh)%ncalls=1
         call timout
      endif

! end of trajectories
      call fzeit(actual)
      i = index(actual,':')
      actual=actual(i-9:i+6)

      if (lrdsrfhop) then
! check for existence of directory with trajectories
         trjdir = name(1:laenge)//'/zagreb_trj'
         lexist=.false.
         dirlbl=index(trjdir,' ')-1
         open(idump,file=trjdir(1:dirlbl)//'/scratch',err=667)
         close(idump,status='delete')
         lexist=.true.
  667    continue
         if (.not. lexist) then
            routine = 'Prop_zagreb'
            message = 'No trajectories found'
            call errormsg
         endif

      else
         write(ilog,'(/,a)') '   Finished Zagreb Surface Hopping '
         write(ilog,'(57x,a)')  actual(1:15)
         write(ilog,'(72a1)') ('#',i=1,72)
         write(ilog,'(a)') '   '
      endif

!-----------------------------------------------------------------------
! Close input file
!-----------------------------------------------------------------------
      close(iin)

!-----------------------------------------------------------------------
! allocate pointers to use output routines
!-----------------------------------------------------------------------
      call zeigproptr
      call memproptr

      if(allocated(mc))deallocate(mc)
      if(allocated(mr))deallocate(mr)
      if(allocated(mi))deallocate(mi)
      if(allocated(ml))deallocate(ml)
      if(allocated(dicht1))deallocate(dicht1)
      if(allocated(dicht2))deallocate(dicht2)
      if(allocated(dicht3))deallocate(dicht3)
      if(allocated(dicht4))deallocate(dicht4)

      cerr = 'propzagreb_mc '
      allocate(mc(mcdim+1),stat=aerr)
      allocmemory=allocmemory + mcdim*complexsize()
      if (aerr > 0) go to 1000
      cerr = 'propzagreb_mr '
      allocate(mr(mrdim+1),stat=aerr)
      allocmemory=allocmemory + mrdim*realsize()
      if (aerr > 0) go to 1000
      cerr = 'propzagreb_mi '
      allocate(mi(midim+1),stat=aerr)
      allocmemory=allocmemory + midim*integersize()
      if (aerr > 0) go to 1000
      cerr = 'propzagreb_ml '
      allocate(ml(mldim+1),stat=aerr)
      allocmemory=allocmemory + mldim*logicalsize()
      if (aerr > 0) go to 1000
      cerr = 'runproptr_dicht1  '
      allocate(dicht1(dmatdim+1),stat=aerr)
      allocmemory=allocmemory + dmatdim*complexsize()
      if (aerr > 0) go to 1000
      cerr = 'runproptr_dicht2  '
      allocate(dicht2(dmatdim+1),stat=aerr)
      allocmemory=allocmemory + dmatdim*complexsize()
      if (aerr > 0) go to 1000
      cerr = 'runproptr_dicht3  '
      allocate(dicht3(d3matdim+1),stat=aerr)
      allocmemory=allocmemory + d3matdim*realsize()
      if (aerr > 0) go to 1000
      cerr = 'runproptr_dicht4  '
      allocate(dicht4(dmatdim+1),stat=aerr)
      allocmemory=allocmemory + dmatdim*complexsize()
      if (aerr > 0) go to 1000

!-----------------------------------------------------------------------
! Read DVR data needed to propagate wavefunction (see rddvr.f for
! details)
!-----------------------------------------------------------------------
      dvrdata(:) = .false.
      dvrdata(1)=.true.
      dvrdata(2)=.true.
      chkdvr=1
      call rddvr(mr(mrort),mr(mrtrafo),rdum,mr(mrfftp), &
           mc(mchin),mc(mcrueck),mi(mifftfak),mc(mcexphin), &
           mc(mcexprueck),mi(mijsph),mi(mimsph),rdum1,chkdvr)

!-----------------------------------------------------------------------
! Allocate extra arrays to sort output
!-----------------------------------------------------------------------
      allocate(lskip(ntraj))

!-----------------------------------------------------------------------
! read log files for failed trajectories
! lskip(n) set to true and ntraj1 = no. good trajectories
!-----------------------------------------------------------------------
       write(ilog,'(/,a)') '*** Checking For Failed Trajectories ***'
       ntraj1=ntraj-ntraj0+1
       lskip = .false.
       nhopsav = 0
       nhopsmax = 0
       etotavav = 0.0
       if (ntraj0 .gt. 1) lskip(1:ntraj0-1) = .true.
       do n=ntraj0,ntraj
          write(trjdir1,'(2a,i7)') trjdir(1:dirlbl),'/traj.',n
          call rmblnk(trjdir1)
          jlbl=index(trjdir1,' ')-1

! check trajectory has run
          lexist=.false.
          open(idump,file=trjdir1(1:jlbl)//'/scratch',status='unknown',err=668)
          close(idump,status='delete')

          open(idump,file=trjdir1(1:jlbl)//'/zagreb.log',status='old',err=668)
          close(idump,status='keep')
          lexist=.true.

  668     continue

          if (.not. lexist) then
             write(ilog,'(a,i5)') 'Trajectory not run : ',n
             lskip(n) = .true.
             ntraj1 = ntraj1-1
          else
             call rdtrlog(trjdir1,lskip(n),n,ntraj1,nhopsav,nhopsmax,etotavav)
          endif
       enddo

!-----------------------------------------------------------------------
! if ntraj1 = 0, all directories exist but contain failed runs!
!-----------------------------------------------------------------------
       if (ntraj1 .eq. 0) then
          routine='Runprop_zagreb'
          message= &
     'No trajectories to run and none to read. Check for failed output in zagreb_trj'
          call errormsg
       endif

!-----------------------------------------------------------------------
! Output statistics
!-----------------------------------------------------------------------
       write(ilog,'(/,a)') '*** Trajectory Statistics *** '
       write(ilog,'(a,i5)') 'Maximum no. hops : ',nhopsmax
       write(ilog,'(a,i5)') 'Average no. hops : ',nhopsav/ntraj1
       write(ilog,'(a,f10.6,a)') 'Average Energy   : ',&
                  (etotavav/dble(ntraj1))*ev,' eV'
  
!-----------------------------------------------------------------------
! recalculate psi pointers and resize array
!-----------------------------------------------------------------------
      if ((ntraj1 .lt. ntraj) .or. test) then
! store electronic SPFs
         allocate(tmppsi(subdim(meb)*dim(meb,1)))
         call cpvxz(psi(zetf(meb,1)),tmppsi,subdim(meb)*dim(meb,1))
 
! recalculate dimensions and pointers
         dim(1,1) = ntraj1
         call psidat
         deallocate(psi)
         allocate(psi(dgldim))

! restore electronic SPFs to new location
         call cpvxz(tmppsi,psi(zetf(meb,1)),subdim(meb)*dim(meb,1))
         deallocate(tmppsi)

! log trajectory mapping
         write(ilog,'(/,a,i5)') 'No. Good Trajectories for Output : ',ntraj1
         n1=0
         do n=ntraj0,ntraj
            if (lskip(n)) cycle
            
            n1 = n1+1
            if (ntraj1 .lt. ntraj) then
               write(ilog,'(2(a,i5))') ' Traj ',n,' written as trajectory ',n1
            endif
         enddo

      endif

!-----------------------------------------------------------------------
! Open output files
!-----------------------------------------------------------------------
      call opendat

!-----------------------------------------------------------------------
! Generate psi from good traj files
!-----------------------------------------------------------------------
      if (test) then
         noutput = 1
      else
         noutput = nint((tfinal - tinit)/out1) + 1
      endif
      do i=1,noutput

         time = tinit + (i-1)*out1

! convert tstep to au and get no. of steps done
         istep = nint(time/(tstep*fs))

!-----------------------------------------------------------------------
! read output files
!-----------------------------------------------------------------------
         call trpsi(trjdir,psi(zetf(1,1)),istep,nspfdof(1),ntraj,&
              ntraj1,lskip)

!-----------------------------------------------------------------------
! read energies and
! assign trajectories to states in the "wavefunction"
!-----------------------------------------------------------------------
         epot_shz = 0.0_dop
         ekin_shz = 0.0_dop
         call trpsi_st(trjdir,psi,istep,nddstate,mi(mitrajst),ntraj, &
              ntraj1,lskip)

!-----------------------------------------------------------------------
! And output
!-----------------------------------------------------------------------
         call outputtr(time,psi)

      enddo

!-----------------------------------------------------------------------
! Deallocate memory 
!-----------------------------------------------------------------------
      cerr = 'prop_zagreb'
      deallocate(psi,dicht1,dicht2,dicht3,dicht4,mc,mr,mi,ml,stat=aerr)
      if (aerr > 0) go to 1001
      if (allocated(lskip)) deallocate(lskip)

  600 continue ! End test0 calculation
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_operdef
      call dealloc_psidef
      call dealloc_dirdyn
      call dealloc_dddb
      call dealloc_dercp

      return

  900 continue
      routine='Runprop_zagreb'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop

  999 continue
      routine='Runprop_zagreb'
      message='Error reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg

 1000 write (6,*) 'Error in Memory Allocation of ',cerr,' Error code is ',aerr
      stop

 1001 write (6,*) 'Error in Memory Deallocation of ',cerr,' Error code is ',aerr
      stop

      end subroutine runprop_zagreb

!####################################################################### 

      subroutine geningeo(trjdir1,psi,gdof,ntraj,n,ithr)

      implicit none

      integer(long), intent(in) :: n,ntraj,gdof,ithr
      integer(long)             :: i,j,k,ilbl,iunit
      complex(dop), intent(in)  :: psi(gdof,ntraj)
      character(len=c5)         :: filename
      character(len=c5), intent(in) :: trjdir1

      real(dop), dimension(3*natmtsh) :: qcoo
      real(dop), dimension(3,natmtsh) :: xcoo
      real(dop), dimension(3,natmtsh) :: vcoo
      real(dop), parameter            :: amu = 1822.88849_dop

      iunit = iddin+ithr
      ilbl=index(trjdir1,' ')-1
      filename=trjdir1(1:ilbl)//'/geom'
      open(iunit,file=filename,form='formatted',status='new',err=900)

!-----------------------------------------------------------------------
! extract initial coordinates for trajectory n and transform to "Xcoo"
!-----------------------------------------------------------------------
      qcoo(:) = 0.0_dop
      do i=1,gdof
         qcoo(i) = dble(psi(i,n))
      enddo
      if (ltshtrans) then
         call mvxxdd1(tshtransf,qcoo,xcoo,maxdim,gdof,ndoftsh)
         call addvxxdo(tshxcoo0,xcoo,ndoftsh)
      else
         k=0
         do i=1,natmtsh
            do j=1,3
               k=k+1
               xcoo(j,i) = qcoo(k)
            enddo
         enddo
      endif
            
!-----------------------------------------------------------------------
! write out geom file
! mass in zagreb code is in amu
!-----------------------------------------------------------------------
      do i=1,natmtsh
         write(iunit,'(1x,a4,f10.6,3(1x,f16.12),3x,a1)') &
            tshatnam(i),tshmass(i)/amu,(xcoo(j,i),j=1,3),'q'
      enddo
      close(iunit)
              
      ilbl=index(trjdir1,' ')-1
      filename=trjdir1(1:ilbl)//'/veloc'
      open(iunit,file=filename,form='formatted',status='new',err=900)

!-----------------------------------------------------------------------
! extract initial velocities for trajectory n and transform to "Xcoo"
! need to take care of frozen DOF coordinates
!-----------------------------------------------------------------------
      qcoo(:) = 0.0_dop
      do i=1,gdof
         qcoo(i) = dimag(psi(i,n))
      enddo
      if (ltshtrans) then
         call mvtxdd1(tshtransb,qcoo,vcoo,maxdim,gdof,ndoftsh)

         do i=1,natmtsh
            do j=1,3
               vcoo(j,i) = vcoo(j,i)/tshmass(i)
            enddo
         enddo

      else
         k=0
         do i=1,natmtsh
            do j=1,3
               k=k+1
               vcoo(j,i) = qcoo(k)
            enddo
         enddo

         do i=1,natmtsh
            do j=1,3
               vcoo(j,i) = vcoo(j,i)/tshmass(i)
            enddo
         enddo
      endif
         
!-----------------------------------------------------------------------
! write out veloc file
!-----------------------------------------------------------------------
      do i=1,natmtsh
         write(iunit,'(3(1x,f16.12))')  (vcoo(j,i),j=1,3)
      enddo
      close(iunit)
              

      return

  900 continue
      routine='Geningeo (prop_zagreb)'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop

      end subroutine geningeo

!####################################################################### 
!
! Assign "wavefunction" coefficients, i.e. the trajectory
! weight on the appropriate state.
!
!####################################################################### 

      subroutine trpsi_st(trjdir,psi,istep,nst,trajst,ntraj,ntraj1,lskip)

      implicit none

      integer(long), intent(in)  :: nst,ntraj,ntraj1,istep
      integer(long), intent(out) :: trajst(ntraj1)
      integer(long)              :: i,s,n,istate,dirlbl,ilbl,n1
      complex(dop), intent(out)  :: psi(ntraj1,nst)
      real(dop)                  :: weight,dtmp
      real(dop)                  :: etottmp,epottmp
      character(len=c5)              :: filename
      character(len=c5), intent(in)  :: trjdir 
      character(len=c5)              :: trjdir1 
      logical(kind=4),intent(in)     :: lskip(ntraj)

      psi(:,:)=0.0_dop
      trajst(:)=0
      weight=dble(ntraj1)
      weight = 1.0_dop/dsqrt(weight)
      dirlbl=index(trjdir,' ')-1

      n1=0
      do n=ntraj0,ntraj
         if (lskip(n)) cycle
         n1=n1+1
            
         write(trjdir1,'(2a,i7,a)') trjdir(1:dirlbl),'/traj.',n,'/Results'
         call rmblnk(trjdir1)
         ilbl=index(trjdir1,' ')-1

         filename=trjdir1(1:ilbl)//'/energy.dat'
         open(idump,file=filename,form='formatted',status='old',err=900)

         read(idump,*,err=901,end=902)  ! file header

! find correct data set
         do i=1,istep
            read(idump,*,err=901,end=902) 
         enddo

! current state. Output has convention that S0 = 0
         read(idump,*,err=901,end=902) dtmp,istate,etottmp,epottmp
         trajst(n1) = istate
         epot_shz = epot_shz + epottmp
         ekin_shz = ekin_shz + (etottmp-epottmp)

         s = trajst(n1)
         psi(n1,s) = weight
      enddo

      return
  900 continue
      routine='Trpsi_st (prop_zagreb)'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop
  901 continue
      routine='Trpsi_st (prop_zagreb)'
      message = 'Error reading energy.dat file'
      call errormsg
      stop
  902 continue
      routine='Trpsi_st (prop_zagreb)'
      message = 'Finished , but incomplete energy.dat file? '
      call errormsg
      stop
      end subroutine trpsi_st


!####################################################################### 
!
! coords + mom stored in psi
! energies read and summed. These are passed to calcdatshz through 
! the calcdattrmod module
!
!####################################################################### 

      subroutine trpsi(trjdir,psi,istep,gdof,ntraj,ntraj1,lskip)

      implicit none

      integer(long), intent(in)      :: istep,gdof,ntraj,ntraj1
      integer(long)                  :: i,j,k,ilbl,f,dirlbl,n,n1
      complex(dop), intent(out)      :: psi(gdof,ntraj1)
      character(len=c5)              :: filename
      character(len=c5), intent(in)  :: trjdir 
      character(len=c5)              :: trjdir1 
      logical(kind=4),intent(in)     :: lskip(ntraj)

      real(dop), dimension(3*natmtsh) :: qcoo,pcoo
      real(dop), dimension(3,natmtsh) :: xcoo
      real(dop), parameter            :: amu = 1822.88849_dop

      psi(:,:) = 0.0_dop
      dirlbl=index(trjdir,' ')-1

      n1=0
      do n=ntraj0,ntraj
         if (lskip(n)) cycle
         n1=n1+1
            
         write(trjdir1,'(2a,i7,a)') trjdir(1:dirlbl),'/traj.',n,'/Results'
         call rmblnk(trjdir1)
         ilbl=index(trjdir1,' ')-1

! Read coordinates
         filename=trjdir1(1:ilbl)//'/geometry'
         open(idump,file=filename,form='formatted',status='old',err=900)
! find correct data set
         do i=1,istep
            read(idump,*,err=901,end=902) 
         enddo

         qcoo(:) = 0.0_dop
         do i=1,natmtsh
            read(idump,*,err=901,end=902) (xcoo(j,i),j=1,3)
         enddo
         if (ltshtrans) then
            call subvxxdo1(xcoo,tshxcoo0,ndoftsh)
            call mvxxdd1(tshtransb,xcoo,qcoo,maxdim,ndoftsh,gdof)
         else
            k=0
            do i=1,natmtsh
               do j=1,3
                  k=k+1
                  qcoo(k) = xcoo(j,i)
               enddo
            enddo
         endif
  
         close(idump)

! Read velocities
         filename=trjdir1(1:ilbl)//'/velocity'
         open(idump,file=filename,form='formatted',status='old',err=900)
! find correct data set
         do i=1,istep
            read(idump,*,err=903,end=904) 
         enddo

         pcoo(:) = 0.0_dop
         do i=1,natmtsh
           read(idump,*,err=903,end=904) (xcoo(j,i),j=1,3)
         enddo
         if (ltshtrans) then
            call mvtxdd1(tshtransf,xcoo,pcoo,maxdim,ndoftsh,gdof)
            k=0
            do i=1,natmtsh
               do j=1,3
                  k=k+1
                  pcoo(k) = tshmass(i)*pcoo(k)
               enddo
            enddo

         else
            k=0
            do i=1,natmtsh
               do j=1,3
                  k=k+1
                  pcoo(k) = tshmass(i)*xcoo(j,i)
               enddo
            enddo
         endif

         close(idump)

! compose wavefunction
         do f=1,gdof
            psi(f,n1) = dcmplx(qcoo(f),pcoo(f))
         enddo

      enddo

      return

  900 continue
      routine='Trpsi (prop_zagreb)'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop
  901 continue
      routine='Trpsi (prop_zagreb)'
      message = 'Error reading geometry file'
      call errormsg
      stop
  902 continue
      routine='Trpsi (prop_zagreb)'
      write(message,'(a,i5)') &
            'Finished , but incomplete geometry file? Trajectory:',n
      call errormsg
      stop
  903 continue
      routine='Trpsi (prop_zagreb)'
      message = 'Error reading velocity file'
      call errormsg
      stop
  904 continue
      routine='Trpsi (prop_zagreb)'
      write(message,'(a,i5)') &
            'Finished , but incomplete velocity file? Trajectory:',n
      call errormsg
      stop

      end subroutine trpsi

!#######################################################################

      subroutine rdtrlog(trjdir1,lskip,n,ntraj1,nhopsav,nhopsmax,etotavav)

      implicit none

      integer(long), intent(in)      :: n
      integer(long), intent(out)     :: nhopsav,nhopsmax
      integer(long), intent(inout)   :: ntraj1
      logical(kind=4), intent(inout) :: lskip

      integer(long)                  :: ilbl,i
      integer(long)                  :: nhops,stold,stnew
      character(len=60)              :: string
      character(len=c5)              :: filename
      character(len=c5), intent(in)  :: trjdir1

      real(dop)                      :: time
      real(dop)                      :: etottmp,epottmp
      real(dop)                      :: etotav,etotvar,etot0
      real(dop), intent(out)         :: etotavav
      integer(long)                  :: ndata,istep,nfrhop
      integer(long), dimension(100)  :: ifrstep
      logical(kind=4)                :: lfrhop,lend

!-----------------------------------------------------------------------
! look for frustrated hops
!-----------------------------------------------------------------------
      ilbl=index(trjdir1,' ')-1
      filename=trjdir1(1:ilbl)//'/zagreb.log'
      open(idump,file=filename,form='formatted',status='old',err=900)

      rewind(idump)
      istep = 0
      ifrstep = 0
      nfrhop = 0
      lfrhop = .false.
      lend = .false.
      do 
        read(idump,'(a)',end=100) string

        if (index(string,"Starting step:") .gt. 0) then
           read(string(16:25),*) istep
        else if (index(string,"*** Frustrated Hop ***") .gt. 0) then
           lfrhop = .true.
           nfrhop = nfrhop+1
           ifrstep(nfrhop) = istep
           exit
        else if (index(string,"Drift in total energy too large") .gt.0) then
           exit
        else if (index(string,"Change in total energy too large") .gt.0) then
           exit
        else if (index(string,"FINISHING MAIN PROGRAM LOOP ") .gt. 0) then
           lend = .true.
           exit
        else if (index(string,"NO STEPS TO BE DONE,") .gt. 0) then
           lend = .true.
           exit
        endif
      enddo

100   continue

      close(idump)

!-----------------------------------------------------------------------
! Check energy conservation
!-----------------------------------------------------------------------
      ilbl=index(trjdir1,' ')-1
      filename=trjdir1(1:ilbl)//'/Results/energy.dat'
      open(idump,file=filename,form='formatted',status='old',err=900)

      etotav = 0.0
      etotvar = 0.0
      ndata = 0
      stold = -999
      nhops = 0
      read(idump,'(a40)',end=200) string  ! Header
      read(idump,*,end=200) time,stnew,etottmp,epottmp   ! first energy
      etot0 = etottmp
      backspace(idump)
      do
         read(idump,*,end=200) time,stnew,etottmp,epottmp
         ndata = ndata+1
         etotav = etotav + etottmp
         etotvar = etotvar + etottmp*etottmp
         if (stold .lt. 0) stold = stnew   ! state at first point
         if (stnew .ne. stold) then   ! change of state
            nhops = nhops+1
            stold = stnew
         endif
      enddo

  200 continue
      etotav = etotav / dble(ndata)
      etotvar = etotvar / dble(ndata)
      etotvar = etotvar - etotav*etotav
      etotvar = dsqrt(etotvar)

!-----------------------------------------------------------------------
! Write analysis
!-----------------------------------------------------------------------

      write(ilog,'(a,i5)') '** Traj: ',n
      write(ilog,'(a,3(2x,f10.8))') 'Av Energy / drift / STD : ',&
                 etotav,etottmp-etot0,etotvar
      write(ilog,'(a,2x,i5)') 'No. of hops          : ',nhops
      write(ilog,'(a,2x,i5)') 'No. frustrated hops  : ',nfrhop
      if (nfrhop .gt. 100) then
         write(ilog,'(a)') 'Only first 100 Frustrated Hops listed'
      endif
      if (lfrhop) then
         do i=1,nfrhop
            write(ilog,'(a,i7)') 'Frustrated Hop at step:',ifrstep(nfrhop)
         enddo
      endif
      
      if (.not. lend) &
           write(ilog,'(a,i7)') 'Trajectory Failed to Finish. Last step:',istep

      if (lsrfchk(1) .and. lfrhop) then
         lskip = .true.
         ntraj1 = ntraj1-1
      else if (lsrfchk(2) .and. (etotvar .gt. deltae)) then
         lskip = .true.
         ntraj1 = ntraj1-1
      else if (.not. lend) then
         lskip = .true.
         ntraj1 = ntraj1-1
      endif

!-----------------------------------------------------------------------
! sum tota no. of hops for statistics
!-----------------------------------------------------------------------
      if (.not. lskip) then
         nhopsmax = max(nhops,nhopsmax)
         nhopsav = nhopsav + nhops
         etotavav = etotavav + etotav
      endif

      return
  900 continue
      routine='Rdtrlog (prop_zagreb)'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop
      end subroutine rdtrlog

      end module prop_zagreb

