C **********************************************************************
C *                                                                    *
C *                 SUBROUTINE EIGLANC                                 *
C *                                                                    *
C * Diagonalises a positive semi-definite Hermitian matrix of          *
C * low rank by Lanczos-iterations.                                    *
C * Full re-orthonormalization is used and Eigenvectors are computed.  *
C * This routine is used for generating optimal unoccupied SPFs.       *
C *                                                                    *
C * Input parameters:                                                  *
C *   Psi:       The (complex) initial-value vector.                   *
C *   sdm:       Length of Psi vectors (subdim).                       *
C *   dim1:      Number of eigenvectors to be computed                 *
C *   maxit:     Maximum number of Lanczos iterations.               i  *
C *  degenerate: If degenerate is set to n, then the program           *
C *              tries to find n degenerate eigenvalues.               *
C *              Set degenerate=0 for normal use.                      *
C *  degenerate: If degenerate is set to n, then the program           *
C *              tries to find n degenerate eigenvalues.               *
C *              Set degenerate=0 for normal use.                      *
C *                                                                    *
C * Output parameters:                                                 *
C *   Psi:       Computed eigenvectors. Psi(sdm,dim1)                  *
C *                                                                    *
C *   Krylov:    Complex matrix with dimension (sdm,maxit+1) used      *
C *              to store the Krylov (or Lanczos) vectors |psi>,       *
C *              H|psi>, H^2|psi>, etc. (orthonormalized).             *
C *                                                                    *
C *   iter:      Number of Lanczos iterations actually performed.      *
C *   eigenval:  Computed eigenvalues.                                 *
C *   eigenvec:  Eigenvectors of the tri-diagonal Lanczos matrix.      *
C *                                                                    *
C *                                                                    *
C * Work arrays: offdiag(maxit),wk(2*maxit), all real                  *
C *                                                                    *
C * External routines:                                                 *
C *   Func:      Computes the action of the Hamiltonian on Psi.        *
C *              Called as: Func(Psi,HPsi,sdm)                         *
C *                                                                    *
C *                                                                    *
C * V8.4 10/19 HDM                                                     *
C **********************************************************************

      subroutine eiglanc(psi,sdm,dim1,maxit,krylov,iter,func,
     +                   eigenvec,eigenval,offdiag,wk,degenerate,
     +                   CData,iout)

      Implicit None

      integer    sdm,dim1,maxit,iter,i,g,e,err,degenerate
      real*8     Alpha,Beta,norm,evmax
      real*8     offdiag(maxit),wk(2*maxit)
      real*8     eigenvec(maxit,maxit),eigenval(maxit)
      complex*16 psi(sdm,dim1),krylov(sdm,maxit+1),ovlp
      external   func
C     Matrix to be diagonalized
      complex*16 CData(*)
C     Output channel
      integer    iout

C --- CONSISTENCY CHECK ---
      if(maxit.le.dim1+1) then
         write(6,'(/a,/a,i5,a,i5)')
     +     'ERROR in eiglanc: maxit > dim1+1 not satisfied.',
     +     'maxit =', maxit,',  dim1 =', dim1
         stop
      endif

C --- INITIALIZE VARIABLES ---
      norm=0.d0
      eigenval = 0.0d0
      do g=1,sdm
         norm=norm+dble(DConjg(psi(g,1))*psi(g,1))
      enddo
      do g=1,sdm
         krylov(g,1)=psi(g,1)/sqrt(norm)
      enddo

C --- COMPUTE H|psi>, and determine the first alpha and beta values.
      call func(krylov(1,1),krylov(1,2),sdm,CData)
      Alpha=0.d0
      Do g = 1,sdm
         Alpha = Alpha+dble(DConjg(krylov(g,1))*krylov(g,2))
      EndDo
      Do g = 1,sdm
         krylov(g,2) = krylov(g,2)-Alpha*krylov(g,1)
      EndDo
C --- Re-orthogonalize
      ovlp = (0.d0,0.d0)
      Do g = 1,sdm
         ovlp = ovlp + DConjg(krylov(g,1))*krylov(g,2)
      enddo
      Beta = 0.d0
      Do g = 1,sdm
         krylov(g,2)=krylov(g,2)-ovlp*krylov(g,1)
         Beta = Beta + dble(DConjg(krylov(g,2))*krylov(g,2))
      enddo
      Beta = Sqrt(Beta)
      Do g = 1,sdm
         krylov(g,2)=krylov(g,2)/Beta
      enddo

      OffDiag(1) = -Beta     ! We define the negative tri-diagonal matrix
      eigenval(1) = -Alpha ! to get the eigenvalues ordered non-increasingly
      evmax = Alpha + Beta   ! evmax is an estimate for the largest eigenvalue
      iter = 1

C --- ITERATION LOOP

 100  iter=iter+1
      call func(krylov(1,iter),krylov(1,iter+1),sdm,CData)
      Alpha=0.d0
      Do g = 1,sdm
         Alpha = Alpha+dble(DConjg(krylov(g,iter))*krylov(g,iter+1))
      EndDo
      Do g = 1,sdm
         krylov(g,iter+1) = krylov(g,iter+1) - Alpha*krylov(g,iter)
     +                                       - Beta*krylov(g,iter-1)
      EndDo
C --- Re-orthogonalize
      Do e=1,iter
         ovlp = (0.d0,0.d0)
         Do g = 1,sdm
            ovlp = ovlp + DConjg(krylov(g,e))*krylov(g,iter+1)
         enddo
         Do g = 1,sdm
            krylov(g,iter+1)=krylov(g,iter+1)-ovlp*krylov(g,e)
         enddo
      enddo ! e=1,iter

      Beta=0
      Do g = 1,sdm
         Beta = Beta + dble(DConjg(krylov(g,iter+1))*krylov(g,iter+1))
      enddo
      Beta = Sqrt(Beta)
      if(Beta.gt.1.d-50) then
         Do g = 1,sdm
            krylov(g,iter+1)=krylov(g,iter+1)/Beta
         enddo
      endif
      OffDiag(iter) = -Beta     ! We define the negative tri-diagonal matrix
      eigenval(iter) = -Alpha ! to get the eigenvalues ordered non-increasingly
      evmax = max(evmax,Alpha+Beta) ! evmax is an estimate for the largest eigenvalue
C      WRITE(6,'(a,i4,3e12.3)') 'iter,a,b,evm', iter,Alpha,Beta,evmax

      if(iter.ge.sdm) goto 300
      if(iter.eq.maxit) then
         write(iout,'(/2a)') '### WARNING iter reaches maxit.',
     +                    '  Check convergence!'
         goto 200
      endif
      if(beta .lt. 1.d-12*evmax) goto 300
      if(iter .le. dim1) goto 100
      if(alpha+beta .gt. 1.d-9*evmax) goto 100 ! This is because the operator has low rank


 300  if(degenerate.eq.0) goto 200
C      PRINT *, 'degenerate =',degenerate,', iter =',iter
      degenerate=degenerate-1
      if(iter.ge.maxit-1) then
         write(6,'(/a,i5,a,i5)') '### iter =', iter,',  maxit =',maxit
         write(6,'(2a/)') '### maxit is too small for searching for ',
     +         'degenerate states. Increase maxit.'
         goto 200
      endif
      call random_number(wk)
      do g=1,sdm
         krylov(g,iter+1)=wk(g) - 0.5d0
      enddo
      OffDiag(iter) = 0.d0
C --- Orthonormalize
      do i=1,2   ! Perform the orthonormalization twice.
         Do e=1,iter
            ovlp = (0.d0,0.d0)
            Do g = 1,sdm
               ovlp = ovlp + DConjg(krylov(g,e))*krylov(g,iter+1)
            enddo
            Do g = 1,sdm
               krylov(g,iter+1)=krylov(g,iter+1)-ovlp*krylov(g,e)
            enddo
         enddo ! e=1,iter
C --- Normalize
         norm=0.d0
         do g=1,sdm
               norm=norm+dble(DConjg(krylov(g,iter+1))*krylov(g,iter+1))
         enddo
C         PRINT *, 'norm =',norm
         if(norm .le. 1.d-30) then
            write(6,'(/2a/)') 'eiglanc: ',
     +           'Could not find a new vector for restart!'
            goto 200
         endif
         do g=1,sdm
            krylov(g,iter+1)=krylov(g,iter+1)/sqrt(norm)
         enddo
      enddo  ! i=1,2

      evmax=evmax*0.1d0
      goto 100

C --- DIAGONALIZE LANCZOS MATRIX --- CALL LAPACK ROUTINE ---

 200  call DSTEQR('I',iter,EigenVal,OffDiag,EigenVec,maxit,wk,err)

C --- Correct the sign of the eigenvalues
      Do i = 1,iter
         EigenVal(i)=-EigenVal(i)
      enddo

C --- Build eigenvectors from Lanczos EigenVector and Krylov-vectors.
      Do e = 1,dim1
         Do g = 1,sdm
            Psi(g,e) = 0.d0
         EndDo
         Do i = 1,iter
            Do g = 1,sdm
               Psi(g,e) = Psi(g,e) + Krylov(g,i)*EigenVec(i,e)
            EndDo
         EndDo ! i=1,iter
      EndDo  ! e=1,dim1

      return
      end
