
! **********************************************************************
! *                                                                    *
! *                           LANCZOS (lanczos.f)                      *
! *                                                                    *
! * Library module containing a Lanczos algorithm for diagonalising    *
! * real symmetric or complex Hermitian matrices.                      *
! *                                                                    *
! * References:                                                        *
! *    [1] B. N. Parlett: "The Symmetric Eigenvalue Problem."          *
! *        Prentice-Hall, 1980.                                        *
! *    [2] J. K. Cullum and R. A. Willoughby: "Lanczos Algorithms for  *
! *        Large Symmetric Eigenvalue Computations. Vol. I: Theory."   *
! *        Birkhaeuser, 1985.                                          *
! *                                                                    *
! * Michael Beck, May 1999                                             *
! *                                                                    *
! **********************************************************************
!*** function lancz_sclrprod and subroutine lancz_restartout are now in module diaghammod.f90

      module lanczosmod

use decimal, only: long, dop

      implicit none
      private
      public :: lanczos,lancz_errmsg
      contains


! **********************************************************************
! *                                                                    *
! *                          SUBROUTINE LANCZOS                        *
! *                                                                    *
! * Purpose:                                                           *
! *    Master routine for diagonalising a real symmetric or complex    *
! *    Hermitian matrix.                                               *
! *                                                                    *
! * Input parameters:                                                  *
! *    order:       Number of Lanczos iterations made so far. Set      *
! *                 "order" to 0 for the initial call. (Allows one to  *
! *                 restart the Lanczos routine.)                      *
! *    psi:         Initial vector of length "psidim". If "order" > 0  *
! *                 this array must contain the first Lanczos vector.  *
! *    psidim:      Length of "psi".                                   *
! *    maxorder:    Number of Lanczos iterations to be made.           *
! *    machprec:    Machine precision (smallest x with 1.0+x > 1.0).   *
! *    calceigvec:  If true, the eigenvectors will be computed and     *
! *                 written to file by routine "eigvecout".            *
! *    iworkdim:    Length of "iwork" (must be >= maxorder).           *
! *    rworkdim:    Length of "rwork" (must be >= 6*maxorder-1 if      *
! *                 calceigvec = false, maxorder^2+4*maxorder-1 else). *
! *    cworkdim:    Length of "cwork" (must be >= 2*psidim).           *
! *    rwork:       For "order" > 0 the elements 1,...,order of this   *
! *                 real*8 array must contain the on-diagonal terms    *
! *                 "alpha(1:order)", and the elements order+1,...,    *
! *                 2*order the off-diagonal terms "beta(2:order+1)".  *
! *                 If "order" = 0 the contents of "rwork" is ignored. *
! *    cwork:       For "order" > 0 the first "psidim" elements of     *
! *                 this complex*16 array must contain the second      *
! *                 Lanczos vector. If "order" = 0 the contents of     *
! *                 "cwork" is ignored.                                *
! *                                                                    *
! * Output parameters:                                                 *
! *    psi:         This array is overwritten.                         *
! *    neigs:       Number of accurate eigenvalues found.              *
! *    eigval:      Vector containing the eigenvalues.                 *
! *    eigintens:   Vector containing the corresponding "intensities", *
! *                 i.e. overlaps between "psi" and the eigenvectors.  *
! *    eigerr:      Vector containing the error of the eigenvalues.    *
! *    errflag:     Error/warning flag having the following meaning:   *
! *                  0: everything o.k.                                *
! *                  1: Not enough integer work array                  *
! *                  2: Not enough real work array                     *
! *                  3: Not enough complex work array                  *
! *                  4: Illegal order specified                        *
! *                 11: Diagonalisation of tridiagonal matrix failed   *
! *                                                                    *
! * Other parameters:                                                  *
! *    lhpsi:       Logical data passed to "hpsi" and output routines. *
! *    ihpsi:       Integer data passed to "hpsi" and output routines. *
! *    rhpsi:       Real*8 data passed to "hpsi" and output routines.  *
! *    chpsi:       Complex*16 data passed to "hpsi" and output        *
! *                 routines.                                          *
! *    iwork:       Integer work array of length "iworkdim".           *
! *    rwork:       Real*8 work array of length "rworkdim".            *
! *    cwork:       Complex*16 work array of length "cworkdim".        *
! *                                                                    *
! * External routines:                                                 *
! *    sclrprod:    Complex function computing the scalar product of   *
! *                 two vectors. Called:                               *
! *                 z = sclrprod(bra-vector,ket-vector,vector_length)  *
! *    lanczout:    Routine that writes the current Lanczos vector to  *
! *                 file. Called:                                      *
! *                 call lanczout(lanczos_vector,vector_length)        *
! *    eigvecout:   Routine that writes the current eigenvector to     *
! *                 file. Called:                                      *
! *                 call eigvecout(number_of_eigenvector,eigenvector,  *
! *                                eigenvector_length,lhpsi,ihpsi,     *
! *                                rhpsi,chpsi)                        *
! *    restartout:  Routine that writes two Lanczos vectors and the    *
! *                 on- and off-diagonal elements alpha(1,...,order)   *
! *                 and beta(2,...,order+1) to file. (This data is     *
! *                 needed to restart the Lanczos routine.) Called:    *
! *                 call restartout(lanczos_vector1,lanczos_vector2,   *
! *                                 vector_length,alpha,beta,order,    *
! *                                 lhpsi,ihpsi,rhpsi,chpsi)           *
! *    hpsi:        Routine that multiplies the matrix to be           *
! *                 diagonalised with a vector. Called:                *
! *                 call hpsi(input_vector,resulting_vector,           *
! *                           vector_length,lhpsi,ihpsi,rhpsi,chpsi)   *
! *                                                                    *
! **********************************************************************

      subroutine lanczos (order, psi, psidim, maxorder, machprec,&
                          calceigvec, iworkdim, rworkdim, cworkdim,&
                          neigs, eigval, eigintens, eigerr, errflag,&
                          sclrprod, lanczout, eigvecout, restartout,&
                          hpsi, iwork,&
                          rwork, cwork)


      implicit none

      logical(kind=4), intent(in)                       :: calceigvec
      integer(long), intent(inout)                      :: order
      integer(long), intent(in)                         :: psidim,maxorder,&
                                                           iworkdim,rworkdim,cworkdim
      integer(long), intent(out)                        :: neigs,errflag
      integer(long), dimension(1:iworkdim), intent(out) :: iwork
      real(dop), intent(in)                             :: machprec
      real(dop), dimension(1:maxorder), intent(out)     :: eigval,eigerr,eigintens
      real(dop), dimension(1:rworkdim), intent(inout)   :: rwork
      complex(dop), dimension(1:psidim), intent(inout)  :: psi
      complex(dop), dimension(1:cworkdim), intent(out)  :: cwork

      complex(dop), external                            :: sclrprod
      external                                          :: restartout,hpsi,&
                                                           lanczout,eigvecout

      integer(long)       :: neigvec,iwrk1,iwrk2,rwrk1,rwrk2,rwrk3,rwrk4,rwrk5,&
                             rwrk6,cwrk1,cwrk2,i

! --- CHECK THE ORDER ---

      if ((order .lt. 0) .or. (order .ge. maxorder)) then
         errflag = 4
         return
      endif

! --- COMPUTE POINTERS TO DIFFERENT PARTS OF WORK ARRAYS ---

      iwrk1       = 1
      iwrk2       = iwrk1+maxorder
      rwrk1       = 1
      rwrk2       = rwrk1+maxorder
      rwrk3       = rwrk2+maxorder
      rwrk4       = rwrk3+maxorder-1
      if (.not. calceigvec) then
         rwrk5 = rwrk4+2*maxorder
      else
         rwrk5 = rwrk4+maxorder**2
      endif
      rwrk6       = rwrk5+maxorder
      cwrk1       = 1
      cwrk2       = cwrk1+2*psidim

! --- CHECK THE LENGTH OF THE WORK SPACE ---

      if (iworkdim .lt. iwrk2-iwrk1) then
         errflag = 1
         return
      endif
      if (rworkdim .lt. rwrk6-rwrk1) then
         errflag = 2
         return
      endif
      if (cworkdim .lt. cwrk2-cwrk1) then
         errflag = 3
         return
      endif

! --- IN A RESTART RUN MOVE BETA TO CORRECT POSITION ---

      if (order .gt. 0) then
         do i = order,1,-1
            rwork(rwrk2-1+i) = rwork(rwrk1-1+order+i)
         enddo
      endif

! --- DEFINE HOW MANY COMPONENTS OF AN EIGENVECTOR ARE NEEDED ---

      if (calceigvec) then
         neigvec = maxorder
      else
         neigvec = 2
      endif

! --- CALL LANCZOS ROUTINE ---
      
      call lancz_slave(order,psi,psidim,maxorder,machprec,neigvec,neigs,&
                       eigval,eigintens,eigerr,errflag,sclrprod,&
                       lanczout,eigvecout,restartout,hpsi,&
                       iwork(iwrk1),rwork(rwrk1),&
                       rwork(rwrk2),rwork(rwrk3),rwork(rwrk4),&
                       rwork(rwrk5),cwork(cwrk1))

      return
      end subroutine lanczos

! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE LANCZOS_SLAVE                     *
! *                                                                    *
! * Purpose:                                                           *
! *    Slave routine called by master Lanczos routine.                 *
! *                                                                    *
! * Parameters:                                                        *
! *    For most parameters see "lanczos". Others are given below.      *
! *                                                                    *
! * Input parameters:                                                  *
! *    neigvec: Number of components of eigenvectors to be computed.   *
! *                                                                    *
! * Output parameters:                                                 *
! *    eigstat: "Status" of an eigenvalue. See comment in code for a   *
! *             definition.                                            *
! *    alpha:   On-diagonal elements of tri-diagonal matrix.           *
! *    beta:    Off-diagonal elements of tri-diagonal matrix.          *
! *    eigval2: Eigenvalues of "reduced" tridiagonal matrix; needed    *
! *             for detecting spurious eigenvalues.                    *
! *    eigvec:  Matrix the columns of which contain the eigenvectors.  *
! *             Note that if no eigenvectors are desired, "eigvec"     *
! *             still contains the first and last component of each    *
! *             eigenvector.                                           *
! *    lancz:   Stores two Lanczos vectors (one is held in "psi").     *
! *                                                                    *
! * Other parameters:                                                  *
! *    work:    Real work array of length "maxorder".                  *
! *                                                                    *
! **********************************************************************

      subroutine lancz_slave (order, psi, psidim, maxorder, machprec,&
                              neigvec, neigs, eigval, eigintens,&
                              eigerr, errflag, sclrprod, lanczout,&
                              eigvecout, restartout, hpsi, &
                              eigstat, alpha, beta,&
                              eigval2, eigvec, work, lancz)

      implicit none

      integer(long)                                           :: wrorder
      integer(long), intent(inout)                            :: order
      integer(long), intent(in)                               :: psidim,maxorder,neigvec
      integer(long), intent(out)                              :: neigs,errflag
      integer(long), dimension(1:maxorder), intent(out)       :: eigstat
      real(dop), intent(in)                                   :: machprec
      real(dop), dimension(1:maxorder), intent(out)           :: alpha,eigval,eigerr,&
                                                                 eigintens,work
      real(dop), dimension(2:maxorder+1), intent(out)         :: beta
      real(dop), dimension(1:maxorder-1), intent(out)         :: eigval2
      real(dop), dimension(1:neigvec,1:maxorder), intent(out) :: eigvec
      complex(dop), dimension(1:psidim), intent(inout)        :: psi
      complex(dop), dimension(1:psidim,1:2), intent(out)      :: lancz
      complex(dop), external                                  :: sclrprod
      external   :: eigvecout,lanczout,restartout, hpsi

      integer(long)                   :: p2,p3,swap,neigvec2,order2,i,j
      real(dop)                       :: norm,invnorm,matsum,eps,eigdiff,error
      real(dop), dimension(1:1)       :: dummy

! --- INITIALISE VARIABLES ---

      errflag = 0
      p2      = 2
      p3      = 1
      if(maxorder .le. 1000) then 
         wrorder = max(maxorder/50,2)
      elseif(maxorder .le. 5000) then
         wrorder = maxorder/100
      else
         wrorder = maxorder/200
      endif
      if(wrorder .gt. 200) wrorder=100*((wrorder+33)/100) 
      if(wrorder .gt. 20)  wrorder=10*((wrorder+3)/10) 

! --- COMPUTE NORM OF INITIAL VECTOR ---

      if (order .eq. 0) then
         norm = sqrt(dble(sclrprod(psi,psi,psidim)))
      else
         norm = beta(order+1)
      endif

! --- LANCZOS ITERATION ---

!     This version of the Lanczos scheme is numerically the most stable,
!     see Chap. 2.3 of Ref. [2].

 100  continue
      order = order+1

! --- COMPUTE NORMALISED LANCZOS VECTOR ---

      invnorm = 1.0_dop/norm
      do i = 1,psidim
         lancz(i,p2) = invnorm*psi(i)
      enddo

! --- WRITE LANCZOS VECTOR TO FILE IF DESIRED ---

      call lanczout(lancz(1,p2),psidim)

! --- EVALUATE ACTION OF HAMILTONIAN ON LANCZOS VECTOR ---

      call hpsi(lancz(1,p2),psi,psidim)

! --- DETERMINE ON-DIAGONAL ELEMENT OF LANCZOS MATRIX ---

      alpha(order) = dble(sclrprod(lancz(1,p2),psi,psidim))

! --- GRAM-SCHMIDT-ORTHONORMALISE ON LAST LANCZOS VECTOR ---

      if (order .gt. 1) then
         do i = 1,psidim
            psi(i) = psi(i)-beta(order)*lancz(i,p3)
         enddo
      endif

! --- GRAM-SCHMIDT-ORTHONORMALISE ON LAST BUT ONE LANCZOS VECTOR ---

      do i = 1,psidim
         psi(i) = psi(i)-alpha(order)*lancz(i,p2)
      enddo

! --- DETERMINE OFF-DIAGONAL ELEMENT OF LANCZOS MATRIX ---

      norm = sqrt(dble(sclrprod(psi,psi,psidim)))
      beta(order+1) = norm

! --- SAVE CURRENT LANCZOS VECTOR ---

      swap = p2
      p2   = p3
      p3   = swap

! --- WRITE TO FILE DATA NEEDED FOR A RESTART ---

      call restartout(psi,lancz(1,p3),psidim,alpha,beta,order)

! --- WRITE NUMBER OF ITERATION TO TO OUTPUT FILE

      if(mod(order,wrorder).eq. 0 .or. order.eq.maxorder)&
         call lancz_writer(order,wrorder)

! --- ITERATE UNTIL MAXIMUM ORDER REACHED ---

      if (order .lt. maxorder) goto 100

! --- CALCULATE EIGENVALUES AND -VECTORS OF TRIDIAGONAL MATRIX ---

      call eigtridiag(order,alpha,beta,machprec,neigvec,eigval,eigvec,&
                      errflag,work)
      if (errflag .ne. 0) then
         errflag = errflag+10
         return
      endif

! --- DETERMINE PRECISION PARAMETER ---

!     See Chap. 3.4 of Ref. [2].

      matsum = abs(alpha(1))
      do i = 2,order
         matsum = matsum+abs(alpha(i))+beta(i)
      enddo
      eps = dble(order)*matsum*machprec

! --- UNLESS STATED OTHERWISE ALL EIGENVALUES ARE GOOD ONES ---

!     The status "eigstat" of an eigenvalue is defined as:
!     eigstat =-1: inaccurate eigenvalue
!     eigstat = 0: single accurate eigenvalue
!     eigstat = i: multiple eigenvalue being identical to eigenvalue i
!     Note that multiple eigenvalues are always accurate (see Chap. 4.5
!     of Ref. [2]).

      do i = 1,order
         eigstat(i) = 0
      enddo

! --- DETECT AND LABEL MULTIPLE EIGENVALUES ---

      do i = 1,order
         if (eigstat(i) .eq. 0) then
            do j = i+1,order
               if (abs(eigval(i)-eigval(j)) .lt. eps) then
                  eigstat(i) = i
                  eigstat(j) = i
               endif
            enddo
         endif
      enddo

! --- CALCULATE EIGENVALUES OF REDUCED TRIDIAGONAL MATRIX ---

!     Eigenvalues of the full tridiagonal matrix which are also
!     eigenvalues of the same matrix, but with the first row and column
!     removed, are spurious (see Chap. 4.5 of Ref. [2]).

      neigvec2 = 0
      order2   = order-1
      call eigtridiag(order2,alpha(2),beta(3),machprec,neigvec2,eigval2,&
                      dummy,errflag,work)
      if (errflag .ne. 0) then
         errflag = errflag+10
         return
      endif

! --- DETECT AND LABEL SPURIOUS EIGENVALUES ---

      do i = 1,order
         if (eigstat(i) .eq. 0) then
            do j = 1,order-1
               if (abs(eigval(i)-eigval2(j)) .lt. eps) eigstat(i) = -1
            enddo
         endif
      enddo

! --- ESTIMATE ERROR OF EACH GOOD EIGENVALUE ---

      do i = 1,order
         if (eigstat(i) .ge. 0) then

!       --- DETERMINE INTENSITY ---

!           For multiple eigenvalues the intensities have to be added.

            if ((eigstat(i) .eq. 0) .or. (eigstat(i) .eq. i)) then
               eigintens(i) = eigvec(1,i)**2
            else
               eigintens(eigstat(i)) = eigintens(eigstat(i))&
                                       +eigvec(1,i)**2
            endif

!       --- ESTIMATE ERROR ---

!           See Chap. 13.2 of Ref. [1] for the error formula.

            if (eigstat(i) .eq. 0) then
               eigdiff = 1.0_dop/eps
               do j = 1,order
                  if (j .ne. i) then
                     eigdiff = min(eigdiff,abs(eigval(j)-eigval(i)))
                  endif
               enddo
               error     = beta(order+1)*abs(eigvec(neigvec,i))
               eigerr(i) = max(min(error,(error**2)/eigdiff),&
                               eps/dble(order))
            else
               eigerr(i) = eps/dble(order)
            endif
         endif
      enddo

! --- REMOVE MULTIPLE AND SPURIOUS EIGENVALUES ---

      neigs = 0
      do i = 1,order
         if ((eigstat(i) .eq. 0) .or. (eigstat(i) .eq. i)) then
            neigs = neigs+1
            eigval(neigs)    = eigval(i)
            eigerr(neigs)    = eigerr(i)
            eigintens(neigs) = eigintens(i)
         endif
      enddo

! --- OUTPUT "GOOD" EIGENVECTORS ---

      if (neigvec .eq. maxorder) then
         neigs = 0
         do i = 1,order
            if ((eigstat(i) .eq. 0) .or. (eigstat(i) .eq. i)) then
               neigs = neigs+1
               call eigvecout(eigvec(1,i),neigvec)
            endif
         enddo
      endif

      return
      end subroutine lancz_slave


! **********************************************************************
! *                                                                    *
! *                        SUBROUTINE LANCZ_OUTPUT                     *
! *                                                                    *
! * Writes the current Lanczos vector to file. (In the current form    *
! * this is a dummy routine doing absolutely nothing; it is included   *
! * here for formal reasons but can (if desired) be extended such that *
! * it indeed outputs the Lanczos vector.                              *
! *                                                                    *
! * Input parameters:                                                  *
! *   order:    Current order of the Lanczos iteration.                *
! *   lanczvec: Current Lanczos vector.                                *
! *   dim:      Length of Lanczos vector.                              *
! *                                                                    *
! * Output parameters:                                                 *
! *   none                                                             *
! *                                                                    *
! * Other parameters:                                                  *
! *    lhpsi, ihpsi, rhpsi, chpsi: see above                           *
! *                                                                    *
! **********************************************************************

!     subroutine lancz_output (order, lanczvec, dim, lhpsi, ihpsi,&
!                              rhpsi, chpsi)

!     implicit none

!     logical(kind=4), dimension(*), intent(in)  :: lhpsi
!     integer(long), intent(in)                  :: order,dim
!     integer(long), dimension(*), intent(in)    :: ihpsi
!     real(dop), dimension(*), intent(in)        :: rhpsi
!     real(dop), dimension(1:dim), intent(in)    :: lanczvec
!     complex(dop), dimension(*), intent(in)     :: chpsi


!--- WRITE LANCZOS VECTOR TO FILE ---

! Fill in your personal output commands here if desired.

!     return
!     end subroutine lancz_output

! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE LANCZ_EIGVECOUT                   *
! *                                                                    *
! * Writes the current eigenvector to file. (In the current form       *
! * this is a dummy routine doing absolutely nothing; it is included   *
! * here for formal reasons but can (if desired) be extended such that *
! * it indeed outputs the eigenvectors.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   number: Number of the current eigenvector.                       *
! *   eigvec: Current eigenvector.                                     *
! *   dim:    Length of eigenvector.                                   *
! *                                                                    *
! * Output parameters:                                                 *
! *   none                                                             *
! *                                                                    *
! * Other parameters:                                                  *
! *    lhpsi, ihpsi, rhpsi, chpsi: see above                           *
! *                                                                    *
! **********************************************************************

!     subroutine lancz_eigvecout (number, eigvec, dim, lhpsi, ihpsi,&
!                                 rhpsi, chpsi)

!     implicit none

!     logical(kind=4), dimension(*), intent(in)  :: lhpsi
!     integer(long), intent(in)                  :: number,dim
!     integer(long), dimension(*), intent(in)    :: ihpsi
!     real(dop), dimension(*), intent(in)        :: rhpsi
!     real(dop), dimension(1:dim), intent(in)    :: eigvec
!     complex(dop), dimension(*), intent(in)     :: chpsi

! --- WRITE EIGENVECTOR TO FILE ---

! Fill in your personal output commands here if desired.

!     return
!     end subroutine lancz_eigvecout


! **********************************************************************
! *                                                                    *
! *                        SUBROUTINE EIGTRIDIAG                       *
! *                                                                    *
! * Computes the eigenvalues and the first and last component of the   *
! * eigenvectors of a real symmetric tridiagonal matrix. This is a     *
! * modified version of the "tql2" routine.                            *
! *                                                                    *
! * Input parameters:                                                  *
! *   order:    Order (i.e. size) of the matrix.                       *
! *   ondiag:   Vector with the on-diagonal elements of the matrix.    *
! *   offdiag:  Vector with the off-diagonal elements of the matrix.   *
! *   machprec: Machine precision.                                     *
! *   neigvec:  Number of components of eigenvectors to be computed.   *
! *             Can be "0" (no eigenvectors), "2" (only first and last *
! *             component) or "order" (all components).                *
! *                                                                    *
! * Output parameters:                                                 *
! *   eigval:   Vector containing the eigenvalues of the matrix.       *
! *   eigvec:   Matrix with desired components of the eigenvectors.    *
! *   errflag:  Error flag having the meaning:                         *
! *             0: everything o.k.                                     *
! *             1: diagonalisation failed.                             *
! *             2: illegal value for "neigvec"                         *
! *                                                                    *
! * Other parameters:                                                  *
! *   work:     Real work array of length >= "order".                  *
! *                                                                    *
! **********************************************************************

      subroutine eigtridiag (order, ondiag, offdiag, machprec, neigvec,&
                             eigval, eigvec, errflag, work)

      implicit none

      integer(long), intent(in)                            :: order,neigvec
      integer(long), intent(out)                           :: errflag
      real(dop), intent(in)                                :: machprec
      real(dop), dimension(1:order), intent(in)            :: ondiag
      real(dop), dimension(1:order), intent(out)           :: eigval,work
      real(dop), dimension(1:order-1), intent(in)          :: offdiag
      real(dop), dimension(1:neigvec,1:order), intent(out) :: eigvec

      integer(long)       :: i,j,k,l,m,ii,l1,mml
      real(dop)           :: b,c,f,g,h,p,r,s

! --- INITIALISE VARIABLES ---

      errflag = 0
      do i = 1,order
         eigval(i) = ondiag(i)
      enddo
      do i = 1,order-1
         work(i) = offdiag(i)
      enddo
      work(order) = 0.0_dop
      do i = 1,order
         do j = 1,neigvec
            eigvec(j,i) = 0.0_dop
         enddo
      enddo
      if (neigvec .eq. 2) then
         eigvec(1,1)     = 1.0_dop
         eigvec(2,order) = 1.0_dop
      else
         do i = 1,neigvec
            eigvec(i,i) = 1.0_dop
         enddo
      endif
      b = 0.0_dop
      f = 0.0_dop

! --- RETURN IMMEDIATELY IF MATRIX IS ONE-DIMENSIONAL ---

      if (order .eq. 1) return

! --- CHECK NUMBER OF EIGENVECTOR COMPONENTS ---

      if ((neigvec .ne. 0) .and. (neigvec .ne. 2) .and.&
          (neigvec .ne. order)) then
         errflag = 2
         return
      endif

! --- COMPUTE EIGENVALUES ---

      do l = 1,order
         j = 0
         h = machprec*(abs(eigval(l))+abs(work(l)))
         if (b .lt. h) b = h

!    --- LOOK FOR SMALL SUB-DIAGONAL ELEMENT ---

! changed upper bound to "order-1" from "order" as this is
! subdiagonal and otherwise m=order+1 on exit GAW 26/7
         do m = l,order-1
            if (abs(work(m)) .le. b) goto 120
         enddo
 120     continue
         if (m .eq. l) goto 220
 130     continue
         if (j .eq. 30) then
            errflag = 1
            return
         endif
         j = j+1

!    --- FORM SHIFT ---

         l1 = l+1
         g = eigval(l)
         p = (eigval(l1)-g)/(2.0_dop*work(l))
         r = sqrt(p*p+1.0_dop)
         eigval(l) = work(l)/(p+sign(r,p))
         h = g-eigval(l)
         do i = l1,order
            eigval(i) = eigval(i)-h
         enddo
         f = f+h

!    --- QL TRANSFORMATION ---

         p = eigval(m)
         c = 1.0_dop
         s = 0.0_dop
         mml = m-l

!    --- FOR I = M-1 STEP -1 UNTIL L DO ---

         do ii = 1,mml
            i = m-ii
            g = c*work(i)
            h = c*p
            if (abs(p) .lt. abs(work(i))) goto 150
            c = work(i)/p
            r = sqrt(c*c+1.0_dop)
            work(i+1) = s*p*r
            s = c/r
            c = 1.0_dop/r
            goto 160
 150        continue
            c = p/work(i)
            r = sqrt(c*c+1.0_dop)
            work(i+1) = s*work(i)*r
            s = 1.0_dop/r
            c = c*s
 160        continue
            p = c*eigval(i)-s*g
            eigval(i+1) = h+s*(c*g+s*eigval(i))

!       --- FORM EIGENVECTOR ---

            do k = 1,neigvec
               h = eigvec(k,i+1)
               eigvec(k,i+1) = s*eigvec(k,i)+c*h
               eigvec(k,i) = c*eigvec(k,i)-s*h
            enddo
         enddo
         work(l)   = s*p
         eigval(l) = c*p
         if (abs(work(l)) .gt. b) goto 130
 220     continue
         eigval(l) = eigval(l)+f
      enddo

! --- ORDER EIGENVALUES AND EIGENVECTORS ---

      do ii = 2,order
         i = ii-1
         k = i
         p = eigval(i)
         do j = ii,order
            if (eigval(j) .lt. p) then
               k = j
               p = eigval(j)
            endif
         enddo
         if (k .ne. i) then
            eigval(k) = eigval(i)
            eigval(i) = p
            do j = 1,neigvec
               p = eigvec(j,i)
               eigvec(j,i) = eigvec(j,k)
               eigvec(j,k) = p
            enddo
         endif
      enddo

      return
      end subroutine eigtridiag
      
! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE LANCZ_ERRMSG                      *
! *                                                                    *
! * Generates for a given error number returned by "lanczos" a corres- *
! * ponding error or warning message.                                  *
! *                                                                    *
! * Input parameters:                                                  *
! *   errflag: Error flag returned by "lanczos".                       *
! *                                                                    *
! * Output parameters:                                                 *
! *   errmsg: Error message.                                           *
! *                                                                    *
! **********************************************************************

      subroutine lancz_errmsg (errflag, errmsg)

      implicit none

      integer(long), intent(in)     :: errflag
      character(len=*), intent(out) :: errmsg

! --- GENERATE ERROR MESSAGE ---

      if (errflag .eq. 1) then
         errmsg = 'Not enough integer work array.'
      elseif (errflag .eq. 2) then
         errmsg = 'Not enough real work array.'
      elseif (errflag .eq. 3) then
         errmsg = 'Not enough complex work array.'
      elseif (errflag .eq. 4) then
         errmsg = 'Illegal order specified.'
      elseif (errflag .eq. 11) then
         errmsg = 'Diagonalisation of tridiagonal matrix failed.'
      elseif (errflag .eq. 12) then
         errmsg = 'Illegal number of eigenvector components.'
      else
         errmsg = 'Unknown error occurred.'
      endif

      return
      end subroutine lancz_errmsg

!#######################################################################
!
!       LANCZ_WRITER
!
! Writes number of iterations and CPU-time spent to file "output".
! (Similar to SPEEDOUT)
!#######################################################################

      subroutine lancz_writer(order,wrorder)

use channels

      implicit none

      real(dop)                    :: now,spent
      integer(long)                :: i
      integer(long), intent(in)    :: order
      integer(long), intent(inout) :: wrorder
      character(len=80)            :: actual
      logical(kind=4), save        :: firstcall= .true.
      real(dop), save              :: previous = 0.0_dop
      integer(long), save          :: oldorder = -1000
      real(dop), external          :: fcputime,fsystime

      if( (order-oldorder) .lt. (wrorder+2)/3 ) return
      oldorder=order
!----------------------------------------------------------------------
! determine actual time and CPU-time spent since last call
!----------------------------------------------------------------------
      call fzeit(actual)
      i = index(actual,':')
      actual=actual(i-9:i+6)

      now=fcputime(0)+fsystime(0)
      spent=now-previous
      previous=now

!----------------------------------------------------------------------
! write data to output-file
!----------------------------------------------------------------------
      if (firstcall) then
         firstcall=.false.
         write(iout,'(a,6x,a,8x,a,9x,a)')&
              '#    Order','CPU-time[s]','Delta-CPU','Real time'
         write(iout,'(a,i9,f17.1,26x,a16)')&
              '#', order, now, actual
      else
         write(iout,'(i10,f17.1,f17.3,9x,a16)')&
               order, now, spent, actual
         if(spent .lt. 2.0_dop) then
            wrorder=wrorder*2    ! Enlarge wrorder
            if(wrorder .gt. 300) wrorder=100*((wrorder+33)/100)
            if(wrorder .gt. 30)  wrorder=10*((wrorder+3)/10)
         elseif(spent .gt. 5.d3) then
            wrorder=wrorder/2
            if(wrorder .gt. 300) wrorder=100*((wrorder+33)/100)
            if(wrorder .gt. 30)  wrorder=10*((wrorder+3)/10)
         endif
      end if

      call myflush(iout)



      return
      end subroutine lancz_writer




      end module lanczosmod
