! **********************************************************************
! *                                                                    *
! *                               SVDMOD                               *
! *                                                                    *
! * Program package containing subroutines for performing a singular   *
! * value decomposition of complex symmetric matrices.                 *
! *                                                                    *
! * Michael Beck                                                  4/00 *
! *                                                                    *
! **********************************************************************

      module svdmod

use decimal, only: dop, long
use op1lib,  only: cpmxz, cpqxz, cpmxd
use mmlib, only: mmaxzz, qdxxdz, dqxxdz, mmtxzz, mmxxzz, qdxxzz, dqxxzz, &
                 mmtxdd, mmxxdd, qdxxdd, dqxxdd

      implicit none
      private
      public :: svd, svdz, svdhz, svd_errmsg
      contains

! ----------------------------------------------------------------------
!                              SUBROUTINE SVD
!
! PURPOSE:
!     Performs a singular value decomposition, i.e. it solves the
!     generalised eigenvalue problem A V = B V D, where A and B are real
!     symmetric matrices, V is the eigenvector matrix, and D is a
!     diagonal matrix containing the eigenvalues. B is supposed to be
!     positive definite; if not, B is truncated such that the result is
!     positive definite.
! DEPENDENCIES:
!     Requires LAPACK routine DSYEV.
! INPUT PARAMETERS:
!     a: symmetric matrix as defined above, stored as full matrix
!     b: symmetric matrix as defined above, stored as full matrix
!     n: allocated size of a,b,d
!     cutoff: relative cut-off parameter for the truncation of B, i.e.
!        all eigenvalues smaller than "cutoff" times the largest
!        eigenvalue are considered singular and removed
!     workdim: length of work array; must be >= n**2+4*n-1
! OUTPUT PARAMETERS:
!     a: truncated (n x k) eigenvector matrix V
!     b: is destroyed
!     d: truncated eigenvalue vector D
!     k: size of truncated a,d
!     err: error flag (see subroutine svd_errmsg for details)
! OTHER PARAMETERS:
!     work: real work array of size workdim
! ----------------------------------------------------------------------

      subroutine svd (a,b,n,cutoff,workdim,d,k,err,work)

      implicit none

      integer(long), intent(in)  :: n,workdim
      integer(long), intent(out) :: err,k
      real(dop), intent(in)      :: cutoff
      real(dop), dimension(n,n), intent(inout)     :: a,b
      real(dop), dimension(workdim), intent(inout) :: work
      real(dop), dimension(n), intent(out)         :: d
      integer(long) :: dsyev_wrkdim,umat,wrk,wrkdim,i
      real(dop)     :: threshold

! --- INITIALISE AND CHECK VARIABLES ---

      err = 0
      dsyev_wrkdim = 3*n-1
      if (workdim .lt. dsyev_wrkdim) then
         err = 4
         return
      endif

! --- DIAGONALISE B = U L U^T ---

      call DSYEV('V','U',n,b,n,d,work,workdim,err)
      if (err .ne. 0) then
         err = 1
         return
      endif

! --- CUT-OFF SMALL EIGENVALUES OF L AND TRANSFORM THE OTHERS ---

! In the following it is vital that the eigenvalues are ordered from
! smallest (= d(1)) to largest (= d(n)).

      k = 0
      threshold = cutoff*d(n)
      do i = 1,n
         if (d(i) .ge. threshold) then
            k = k+1
            work(k) = 1.0d0/sqrt(d(i))
         endif
      enddo
      if (k .eq. 0) then
         err = 2
         return
      endif

! --- DETERMINE REMAINING WORKSPACE ---

      umat = k+1
      wrk = umat+n*k
      wrkdim = workdim-wrk+1
      dsyev_wrkdim = 3*k-1
      if (wrkdim .lt. dsyev_wrkdim) then
         err = 4
         return
      endif

! --- SAVE TRUNCATED U ---

      call cpmxd(b(1,n-k+1),work(umat),n,k)

! --- COMPUTE TRUNCATED A = L^(-1/2) U^T A U L^(-1/2) ---

      call mmtxdd(work(umat),a,b,k,n,n)
      call mmxxdd(b,work(umat),a,k,n,k)
      call qdxxdd(a,work(1),b,k)
      call dqxxdd(work(1),b,a,k)

! --- DIAGONALISE TRUNCATED A ---

      call DSYEV('V','U',k,a,k,d,work(wrk),wrkdim,err)
      if (err .ne. 0) then
         err = 3
         return
      endif

! --- COMPUTE TRUNCATED V ---

      call dqxxdd(work(1),a,b,k)
      call mmxxdd(work(umat),b,a,n,k,k)

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine svd

! ----------------------------------------------------------------------
!                              SUBROUTINE SVDZ
!
! PURPOSE:
!     Performs a singular value decomposition, i.e. it solves the
!     generalised eigenvalue problem A V = B V D, where A and B are
!     complex symmetric matrices, V is the eigenvector matrix, and D is
!     a diagonal matrix containing the eigenvalues. B may be singular;
!     if so, B is truncated such that any linear dependencies between
!     its rows or columns are removed.
! DEPENDENCIES:
!     Requires LAPACK routine ZGEEV.
! INPUT PARAMETERS:
!     a: symmetric matrix as defined above, stored as full matrix
!     b: symmetric matrix as defined above, stored as full matrix
!     n: allocated size of a,b,d
!     cutoff: relative cut-off parameter for the truncation of B, i.e.
!        all eigenvalues smaller than "cutoff" times the largest
!        eigenvalue are considered singular and removed
!     workdim: length of work array; must be >= n**2+4*n
! OUTPUT PARAMETERS:
!     a: truncated (n x k) eigenvector matrix V
!     b: is destroyed
!     d: truncated eigenvalue vector D
!     k: size of truncated a,d
!     err: error flag (see subroutine svd_errmsg for details)
! OTHER PARAMETERS:
!     work: complex work array of size workdim
! ----------------------------------------------------------------------

      subroutine svdz (a,b,n,cutoff,workdim,d,k,err,work)

      implicit none

      integer(long), intent(in)  :: n,workdim
      integer(long), intent(out) :: err,k
      real(dop), intent(in)      :: cutoff
      complex(dop), dimension(n,n), intent(inout)     :: a,b
      complex(dop), dimension(workdim), intent(inout) :: work
      complex(dop), dimension(n), intent(out)         :: d

      integer(long) :: eigv,umat,cwrk,rwrk,wrkdim,minpos,i,j
      real(dop)     :: mineig,threshold
      real(dop), dimension(workdim) :: workr
      complex(dop)  :: swap

! --- INITIALISE AND CHECK VARIABLES ---

      err = 0
      eigv = 1
      umat = 1
      cwrk = umat+n**2
      rwrk = cwrk+2*n
      wrkdim = rwrk+n-1
      if (wrkdim .gt. workdim) then
         err = 4
         return
      endif

! --- DIAGONALISE B = U D U^T ---

      call eigcs(n,b,n,n,d,work(umat),err,work(cwrk),workr(rwrk))
      if (err .ne. 0) then
         err = 1
         return
      endif

! --- SORT EIGENVALUES AND INTERCHANGE EIGENVECTORS ACCORDINGLY ---

      call cpqxz(work(umat),b,n)
      do i = 1,n-1
         minpos = i
         mineig = abs(d(minpos))
         do j = i+1,n
            if (abs(d(j)) .lt. mineig) then
               minpos = j
               mineig = abs(d(minpos))
            endif
         enddo
         if (i .ne. minpos) then
            swap = d(i)
            d(i) = d(minpos)
            d(minpos) = swap
            do j = 1,n
               swap = b(j,i)
               b(j,i) = b(j,minpos)
               b(j,minpos) = swap
            enddo
         endif
      enddo

! --- CUT-OFF SMALL EIGENVALUES OF D AND TRANSFORM THE OTHERS ---

! In the following it is assumed that the eigenvalues are ordered from
! smallest (= |d(1)|) to largest (= |d(n)|).

      k = 0
      threshold = cutoff*abs(d(n))
      do i = 1,n
         if (abs(d(i)) .ge. threshold) then
            k = k+1
            work(eigv+k-1) = (1.0d0,0.0d0)/sqrt(d(i))
         endif
      enddo
      if (k .eq. 0) then
         err = 2
         return
      endif

! --- DETERMINE REMAINING WORKSPACE ---

      umat = k+1
      cwrk = umat+n*k
      rwrk = cwrk+2*k
      wrkdim = rwrk+k-1
      if (wrkdim .gt. workdim) then
         err = 4
         return
      endif

! --- SAVE TRUNCATED U ---

      call cpmxz(b(1,n-k+1),work(umat),n,k)

! --- COMPUTE TRUNCATED A = D^(-1/2) U^T A U D^(-1/2) ---

      call mmtxzz(work(umat),a,b,n,k,n)
      call mmxxzz(b,work(umat),a,k,n,k)
      call qdxxzz(a,work(eigv),b,k)
      call dqxxzz(work(eigv),b,a,k)

! --- DIAGONALISE TRUNCATED A ---

      call eigcs(k,a,k,k,d,b,err,work(cwrk),workr(rwrk))
      if (err .ne. 0) then
         err = 3
         return
      endif

! --- SORT EIGENVALUES AND INTERCHANGE EIGENVECTORS ACCORDINGLY ---
 
      do i = 1,k-1
         minpos = i
         mineig = dble(d(minpos))
         do j = i+1,k
            if (dble(d(j)) .lt. mineig) then
               minpos = j
               mineig = dble(d(minpos))
            endif
         enddo
         if (i .ne. minpos) then
            swap = d(i)
            d(i) = d(minpos)
            d(minpos) = swap
            do j = 1,k
               swap = b(j,i)
               b(j,i) = b(j,minpos)
               b(j,minpos) = swap
            enddo
         endif
      enddo

! --- COMPUTE TRUNCATED V ---

      call cpqxz(b,a,k)
      call dqxxzz(work(eigv),a,b,k)
      call mmxxzz(work(umat),b,a,n,k,k)

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine svdz

! ----------------------------------------------------------------------
!                              SUBROUTINE SVDHZ
!
! PURPOSE:
!     Performs a singular value decomposition, i.e. it solves the
!     generalised eigenvalue problem A V = B V D, where A and B are
!     complex hermitian matrices, V is the eigenvector matrix, and D is
!     a diagonal matrix containing the eigenvalues. B may be singular;
!     if so, B is truncated such that any linear dependencies between
!     its rows or columns are removed.
! DEPENDENCIES:
!     Requires LAPACK routine ZHEEV.
! INPUT PARAMETERS:
!     a: complex hermitian matrix as defined above, stored as full matrix
!     b: complex hermitian matrix as defined above, stored as full matrix
!     n: allocated size of a,b,d
!     cutoff: relative cut-off parameter for the truncation of B, i.e.
!        all eigenvalues smaller than "cutoff" times the largest
!        eigenvalue are considered singular and removed
!     workrdim: length of real work array; must be >=4*n
!     workcdim: length of complex work array; must be >= n*n+2*n
! OUTPUT PARAMETERS:
!     a: truncated (n x k) eigenvector matrix V
!     b: is destroyed
!     d: truncated eigenvalue vector D
!     k: size of truncated a,d
!     err: error flag (see subroutine svd_errmsg for details)
! OTHER PARAMETERS:
!     workr: real work array of size workrdim
!     workc: complex work array of size workcdim
! ----------------------------------------------------------------------

      subroutine svdhz (a,b,n,cutoff,d,k,workr,workc,workrdim, &
                        workcdim,err)

      implicit none

      integer(long), intent(in)  :: n,workrdim,workcdim
      integer(long), intent(out) :: err,k
      real(dop), intent(in)      :: cutoff
      real(dop), dimension(workrdim), intent(inout)    :: workr
      complex(dop), dimension(n,n), intent(inout)      :: a,b
      complex(dop), dimension(workcdim), intent(inout) :: workc
      real(dop), dimension(n), intent(out)             :: d
      integer(long) :: eigv,rwrk,cwrk,wrkrdim,wrkcdim,i,umat
      real(dop)     :: threshold


! --- INITIALISE AND CHECK VARIABLES ---

      err = 0
      eigv = 1
      rwrk = eigv+n
      wrkrdim=rwrk+3*n-1
      if (wrkrdim .gt. workrdim) then
         err = 4
         return
      endif
      umat = 1
      cwrk = umat+n*n
      wrkcdim=cwrk+2*n-1
      if (wrkcdim .gt. workcdim) then
         err = 4
         return
      endif

! --- DIAGONALISE B = U D U^+ ---
 
      call ZHEEV('V','U',n,b,n,d,workc(cwrk),2*n,workr(rwrk),err)
      if (err .ne. 0) then
         err = 1
         return
      endif


! --- CUT-OFF SMALL EIGENVALUES OF D AND TRANSFORM THE OTHERS --- 

! In the following it is assumed that the eigenvalues are ordered from
! smallest (= |d(1)|) to largest (= |d(n)|).

      k = 0
      threshold = cutoff*d(n)
      do i = 1,n
         if (d(i) .ge. threshold) then 
            k = k+1
            workr(eigv+k-1) = 1.0d0/sqrt(d(i))
         endif
      enddo
      if (k .eq. 0) then
         err = 2
         return
      endif

! --- SAVE TRUNCATED U ---

      call cpmxz(b(1,n-k+1),workc(umat),n,k)

! --- COMPUTE TRUNCATED A = D^(-1/2) U^+ A U D^(-1/2) ---

      call mmaxzz(workc(umat),a,b,n,k,n)
      call mmxxzz(b,workc(umat),a,k,n,k)
      call qdxxdz(a,workr(eigv),b,k)
      call dqxxdz(workr(eigv),b,a,k)

! --- DIAGONALISE TRUNCATED A ---

      call ZHEEV('V','U',k,a,k,d,workc(cwrk),2*k,workr(rwrk),err)
      if (err .ne. 0) then
         err = 3
         return
      endif

! --- COMPUTE TRUNCATED V ---

      call dqxxdz(workr(eigv),a,b,k)
      call mmxxzz(workc(umat),b,a,n,k,k)


      return
      end subroutine svdhz

! ----------------------------------------------------------------------
!                         SUBROUTINE SVD_ERRMSG
!
! PURPOSE:
!     Generates for a given error number returned by SVD a
!     corresponding error message.
! INPUT PARAMETERS:
!     err: error code returned by SVD
! OUTPUT PARAMETERS:
!     msg: error message
! ----------------------------------------------------------------------

      subroutine svd_errmsg (err,msg)

      implicit none

      integer(long), intent(in)     :: err
      character(len=*), intent(out) :: msg

! --- GENERATE ERROR MESSAGE ---

      if (err .eq. 0) then
         msg = 'no error occurred'
      elseif (err .eq. 1) then
         msg = 'cannot diagonalise matrix B'
      elseif (err .eq. 2) then
         msg = 'all eigenvalues of matrix B are singular'
      elseif (err .eq. 3) then
         msg = 'cannot diagonalise truncated matrix A'
      elseif (err .eq. 4) then
         msg = 'not enough complex workspace'
      else
         msg = 'unknown error occurred'
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine svd_errmsg

      end module svdmod
