!***********************************************************************
!
!     Additional library-like subroutines
!
!   subhc (a,dim)
!       Definition: a(i,j) - conjg(a(j,i)) = a(i,j)
!       Dimensions: a(dim,dim)
!
!   addhc (a,dim)
!       Definition: a(i,j) + conjg(a(j,i)) = a(i,j)
!       Dimensions: a(dim,dim)
!
!   ptr (a,b,c,dim1,dim2)
!       Definition: a(i,j) * b(i) = c(j)
!       Dimensions: a(dim1,dim2), b(dim1), c(dim2)
!
!   ptrc (a,b,c,dim1,dim2)
!       Definition: a(i,j) * b(i) = c(j)
!       Dimensions: a(dim1,dim2), b(dim1), c(dim2)
!
!   ptr2 (a,b,c,dim1,dim2,dim3)
!       Definition: a(i,j,k) * b(j) = c(i,k)
!       Dimensions: a(dim1,dim2,dim3), b(dim2), c(dim1,dim3)
!
!   makeadj (a,b,dim1,dim2)
!       Definition: b(i,j,k) = conjg(a(j,i,k))
!       Dimensions: a(dim1,dim1,dim2), b(dim1,dim1,dim2)
!
!   makeaadj (a,b,dim1,dim2)
!       Definition: b(i,j,k) = -conjg(a(j,i,k))
!       Dimensions: a(dim1,dim1,dim2), b(dim1,dim1,dim2)
!
!   tenstens (a,b,c,supgdim,dim1,dim2)
!        Definition: a(i,j,k) = sum_l b(i,k,l) c(i,j,l)
!        Dimensions: a(supgdim,supgdim,dim2), b(supgdim,dim2,dim1),
!                    c(supgdim,supgdim,dim1)
!
!   tenstens1 (a,b,c,supgdim,dim1,dim2)
!        Definition: a(i,j,k) = a(i,j,k) + sum_l b(i,k,l) c(i,j,l)
!        Dimensions: a(supgdim,supgdim,dim2), b(supgdim,dim2,dim1),
!                    c(supgdim,supgdim,dim1)
!
!   tenstens2 (a,b,c,supgdim,dim1,dim2)
!        Definition: a(i,j,k) = a(i,j,k) + sum_l c(i,j,l) b(j,l,k)
!        Dimensions: a(supgdim,supgdim,dim2), b(supgdim,dim1,dim2),
!                    c(supgdim,supgdim,dim1)
!
!   mktensc:   makes a tensor from the product of a complex matrix with
!              the complex conjugate of a complex vector
!              t(i,j,k)=v(i)*m(j,k)
!
!     cuttens1(tensor,mat,i,j,vdim,dim,ndim):
!         Definition: mat(k,n) = tensor(k,i,j,n)
!         Dimensions: tensor(vdim,dim,dim,ndim),
!                     mat(vdim,ndim)
!
!     filltens1(mat,tensor,i,j,vdim,dim,ndim):
!         Definition: tensor(k,i,j,n) = mat(k,n)
!         Dimensions: tensor(vdim,dim,dim,ndim),
!                     mat(vdim,ndim)
!
!     filltens1c(mat,tensor,i,j,vdim,dim,ndim):
!         Definition: tensor(k,i,j,n) = dconjg(mat(k,n))
!         Dimensions: tensor(vdim,dim,dim,ndim),
!                     mat(vdim,ndim)
!
!     Utility subroutines:
!
!     chktrans
!     mkaherm
!     mkaherm2
!     mktrans
!     vm
!     mv
!     vmv
!     
!***********************************************************************
      module d1tens

      use decimal, only: dop, long
      use openmpmod

      implicit none
      private
      public :: subhc, addhc, ptr1, ptrc, ptr2, makeadj, makeaadj, &
                tenstens, tenstens1, tenstens2, chktrans, mkaherm, &
                mkaherm2, mktrans, vm, vmv, mv, cuttens1, filltens1, &
                filltens1c
      contains

! ----------------------------------------------------------------------
! Library subroutine subhc
!
!     subtracts the hermitian conjugate from a rectangular compl. matrix
!     a(i,j) = a(i,j) - conjg(a(j,i))
!-----------------------------------------------------------------------

      subroutine subhc(a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(inout) :: a

      do j=1,dim
         a(j,j) = a(j,j) - conjg(a(j,j))
         do i=(j+1),dim
            a(i,j) = a(i,j) - conjg(a(j,i))
            a(j,i) = - conjg(a(i,j))
         enddo
      enddo

      return
      end subroutine subhc

! ----------------------------------------------------------------------
! Library subroutine addhc
!
!     adds the hermitian conjugate from a rectangular compl. matrix
!     a(i,j) = a(i,j) + conjg(a(j,i))
!-----------------------------------------------------------------------

      subroutine addhc(a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(inout) :: a

      do j=1,dim
         a(j,j) = a(j,j) + conjg(a(j,j))
         do i=(j+1),dim
            a(i,j) = a(i,j) + conjg(a(j,i))
            a(j,i) = conjg(a(i,j))
         enddo
      enddo

      return
      end subroutine addhc


! ----------------------------------------------------------------------
! Library subroutine ptr1
!
!     calculates some kind of partial trace 
!     a(i,j) * b(i) = c(j)
!-----------------------------------------------------------------------

      subroutine ptr1 (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j 
      integer(long), intent(in)                      :: dim1, dim2
      real(dop), dimension(dim1), intent(in)         :: b
      complex(dop), dimension(dim1,dim2), intent(in) :: a 
      complex(dop), dimension(dim2), intent(out)     :: c

      do j=1,dim2
         c(j) = a(1,j)*b(1)
         do i=2,dim1
            c(j) = c(j) + a(i,j)*b(i)
         enddo
      enddo
      
      return
      end subroutine ptr1

! ----------------------------------------------------------------------
! Library subroutine ptrc
!
!     calculates some kind of partial trace 
!     a(i,j) * b(i) = c(j)
!-----------------------------------------------------------------------

      subroutine ptrc (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim1), intent(in)      :: b
      complex(dop), dimension(dim2), intent(out)     :: c

      do j=1,dim2
         c(j) = a(1,j)*b(1)
         do i=2,dim1
            c(j) = c(j) + a(i,j)*b(i)
         enddo
      enddo
      
      return
      end subroutine ptrc

! ----------------------------------------------------------------------
! Library subroutine ptr2
!
!     calculates some kind of partial trace 
!     a(i,j,k) * b(j) = c(i,k)
!-----------------------------------------------------------------------

      subroutine ptr2 (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, k    
      integer(long), intent(in)                           :: dim1, dim2, dim3
      real(dop), dimension(dim2), intent(in)              :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: a
      complex(dop), dimension(dim1,dim3), intent(out)     :: c

      do k=1,dim3
         do i=1,dim1
            c(i,k) = (0.0d0,0.0d0)
         enddo
      enddo
      do k=1,dim3
         do j=1,dim2
            do i=1,dim1
               c(i,k) = c(i,k)+ a(i,j,k)*b(j)
            enddo
         enddo
      enddo
     
      return
      end subroutine ptr2

!-----------------------------------------------------------------------
! Library subroutine makeadj
!
!     Creates for a set of rectangular complex matrices the corres
!     ponding set of adjoints
!       b(i,j,k) = conjg(a(j,i,k))
!-----------------------------------------------------------------------

      subroutine makeadj (a,b,dim1,dim2)

      implicit none

      integer(long)                                        :: i, j, k  
      integer(long), intent(in)                            :: dim1, dim2
      complex(dop), dimension(dim1,dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim1,dim2), intent(out) :: b

      do k=1,dim2
         do i=1,dim1
            do j=1,dim1
               b(j,i,k) = dconjg(a(i,j,k))
            enddo
         enddo
      enddo

      return
      end subroutine makeadj


!-----------------------------------------------------------------------
! Library subroutine makeaadj
!
!     Creates for a set of rectangular complex matrices the corres
!     ponding set of anti-adjoints
!       b(i,j,k) = -conjg(a(j,i,k))
!-----------------------------------------------------------------------

      subroutine makeaadj (a,b,dim1,dim2)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2
      complex(dop), dimension(dim1,dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim1,dim2), intent(out) :: b

      do k=1,dim2
         do i=1,dim1
            do j=1,dim1
               b(j,i,k) = -conjg(a(i,j,k))
            enddo
         enddo
      enddo

      return
      end subroutine makeaadj


!-----------------------------------------------------------------------
! Library subroutine tenstens
!
! Ananlogon to mattens of MCTDH source code, used in mfsumphi1m
!
!-----------------------------------------------------------------------

      subroutine tenstens (a,b,c,supgdim,dim1,dim2)

      implicit none

      integer(long)                                              :: i, j, k, l
      integer(long), intent(in)                                  :: supgdim, &
                                                                    dim1, dim2
      complex(dop), dimension(supgdim,dim2,dim1), intent(in)     :: b
      complex(dop), dimension(supgdim,supgdim,dim1), intent(in)  :: c
      complex(dop), dimension(supgdim,supgdim,dim2), intent(out) :: a
     
!$omp parallel do num_threads(ompthread) if(lompthread) private(i,j,k)
      do k=1,dim2
         do j=1,supgdim
            do i=1,supgdim
               a(i,j,k) = b(i,k,1)*c(i,j,1)
            enddo
         enddo
      enddo
!$omp end parallel do

      if(.not.lompthread)then
         do l=2,dim1
            do k=1,dim2
               do j=1,supgdim
                  do i=1,supgdim
                     a(i,j,k) = a(i,j,k) + b(i,k,l)*c(i,j,l)
                  enddo
               enddo
            enddo
         enddo
      else
!$omp parallel do private(i,j,k,l) num_threads(ompthread)
         do k=1,dim2
            do l=2,dim1
               do j=1,supgdim
                  do i=1,supgdim
                     a(i,j,k) = a(i,j,k) + b(i,k,l)*c(i,j,l)
                  enddo
               enddo
            enddo
         enddo
!$omp end parallel do
      endif

      return
      end subroutine tenstens


!-----------------------------------------------------------------------
! Library subroutine tenstens1
!
! Ananlogon to mattens of MCTDH source code, used in mfsumphi1m
! for left action
!
!-----------------------------------------------------------------------

      subroutine tenstens1 (a,b,c,supgdim,dim1,dim2)

      implicit none

      integer(long)                                              :: i, j, k, l
      integer(long), intent(in)                                  :: supgdim, &
                                                                    dim1, dim2
      complex(dop), dimension(supgdim,dim2,dim1), intent(in)     :: b
      complex(dop), dimension(supgdim,supgdim,dim1), intent(in)  :: c
      complex(dop), dimension(supgdim,supgdim,dim2), intent(out) :: a
      
      if(.not.lompthread)then
         do l=1,dim1
            do k=1,dim2
               do j=1,supgdim
                  do i=1,supgdim
                     a(i,j,k) = a(i,j,k) + b(i,k,l)*c(i,j,l)
                  enddo
               enddo
            enddo
         enddo
      else
!$omp parallel do num_threads(ompthread) private(i,j,k,l) 
         do k=1,dim2
            do l=1,dim1
               do j=1,supgdim
                  do i=1,supgdim
                     a(i,j,k) = a(i,j,k) + b(i,k,l)*c(i,j,l)
                  enddo
               enddo
            enddo
         enddo
!$omp end parallel do
      endif

      return
      end subroutine tenstens1


!-----------------------------------------------------------------------
! Library subroutine tenstens2
!
! Ananlogon to mattens of MCTDH source code, used in mfsumphi1m,
! for right action
!
!-----------------------------------------------------------------------

      subroutine tenstens2 (a,b,c,supgdim,dim1,dim2)

      implicit none

      integer(long)                                              :: i, j, &
                                                                    k, l     
      integer(long), intent(in)                                  :: supgdim, &
                                                                    dim1, dim2
      complex(dop), dimension(supgdim,dim1,dim2), intent(in)     :: b
      complex(dop), dimension(supgdim,supgdim,dim1), intent(in)  :: c
      complex(dop), dimension(supgdim,supgdim,dim2), intent(out) :: a
      
!$omp parallel do if(lompthread) private(i,j,k,l)
      do k=1,dim2
         do l=1,dim1
            do j=1,supgdim
               do i=1,supgdim
                  a(i,j,k) = a(i,j,k) + c(i,j,l)*b(j,l,k)
               enddo
            enddo
         enddo
      enddo
!$omp end parallel do

      return
      end subroutine tenstens2


!***********************************************************************
!     Utility subroutines:
!
!     chktrans
!     mkherm
!     mkherm2
!     mkaherm
!     mkaherm2
!     mktrans
!     vm
!     mv
!     vmv
!     
!***********************************************************************

      subroutine chktrans(a,b,error,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim2,dim1), intent(in) :: b
      real(dop), intent(out)                         :: error
      real(dop)                                      :: err

      do j=1,dim2
         do i=1,dim1
            err = abs(a(i,j) - b(j,i))
            error = max(err,error)
         enddo
      enddo

      return
      end subroutine chktrans
       

      subroutine mkaherm(a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(inout) :: a

      do i=1,dim
         a(i,i) = (0.0_dop,1.0_dop)*dimag(a(i,i))
         do j=i,dim
            a(j,i) = 0.5d0*(a(j,i) - conjg(a(i,j)))
            a(i,j) = -conjg(a(j,i))
         enddo
      enddo

      return
      end subroutine mkaherm


      subroutine mkaherm2(a,b,dim1,dim2)

      implicit none

      integer(long)                                      :: i, j
      integer(long), intent(in)                          :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(inout)  :: a
      complex(dop), dimension(dim2,dim1), intent(inout)  :: b

      do i=1,dim2
         do j=1,dim1
            a(j,i) = 0.5d0*(a(j,i) - conjg(b(i,j)))
            b(i,j) = -conjg(a(j,i))
         enddo
      enddo

      return
      end subroutine mkaherm2


      subroutine mktrans(a,b,dim1,dim2)

      implicit none

      integer(long)                                      :: i, j
      integer(long), intent(in)                          :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(inout)  :: a
      complex(dop), dimension(dim2,dim1), intent(inout)  :: b

      do i=1,dim2
         do j=1,dim1
            a(j,i) = 0.5d0*(a(j,i) + b(i,j))
            b(i,j) = a(j,i)
         enddo
      enddo

      return
      end subroutine mktrans


      subroutine vm(vec,mat1,mat2,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: vec
      complex(dop), dimension(dim,dim), intent(in)  :: mat1
      complex(dop), dimension(dim,dim), intent(out) :: mat2

      do j=1,dim
         do i=1,dim
            mat2(i,j) = vec(i)*mat1(i,j)
         enddo
      enddo

      return
      end subroutine vm


      subroutine vmv(vec,mat1,mat2,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: vec
      complex(dop), dimension(dim,dim), intent(in)  :: mat1
      complex(dop), dimension(dim,dim), intent(out) :: mat2

      do j=1,dim
         do i=1,dim
            mat2(i,j) = vec(i)*mat1(i,j)*vec(j)
         enddo
      enddo

      return
      end subroutine vmv


      subroutine mv(vec,mat1,mat2,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: vec
      complex(dop), dimension(dim,dim), intent(in)  :: mat1
      complex(dop), dimension(dim,dim), intent(out) :: mat2

      do j=1,dim
         do i=1,dim
            mat2(i,j) = mat1(i,j)*vec(j)
         enddo
      enddo

      return
      end subroutine mv

!**********************************************************************
! projects out part of the tensor (e.g. i,j define a state pair)
!**********************************************************************

      subroutine cuttens1(tensor,matrix,i,j,vdim,dim,ndim)

      implicit none

      integer(long)                                          :: k,n
      integer(long), intent(in)                              :: i,j
      integer(long), intent(in)                              :: vdim, &
                                                                dim, &
                                                                ndim
      complex(dop), dimension(vdim,dim,dim,ndim), intent(in) :: tensor
      complex(dop), dimension(vdim,ndim), intent(out)        :: matrix

      do n=1,ndim
         do k=1,vdim
            matrix(k,n) = tensor(k,i,j,n)
         enddo
      enddo

      return
      end subroutine cuttens1

!**********************************************************************
! fills the tensor from a matrix (e.g. i,j define a state pair)
!**********************************************************************

      subroutine filltens1(matrix,tensor,i,j,vdim,dim,ndim)

      implicit none

      integer(long)                                           :: k,n
      integer(long), intent(in)                               :: i,j
      integer(long), intent(in)                               :: vdim, &
                                                                 dim, &
                                                                 ndim
      complex(dop), dimension(vdim,dim,dim,ndim), intent(out) :: tensor
      complex(dop), dimension(vdim,ndim), intent(in)          :: matrix

      do n=1,ndim
         do k=1,vdim
            tensor(k,i,j,n) = matrix(k,n)
         enddo
      enddo

      return
      end subroutine filltens1

!**********************************************************************
! fills the tensor from the c.c. of a matrix (e.g. i,j define a state pair)
!**********************************************************************

      subroutine filltens1c(matrix,tensor,i,j,vdim,dim,ndim)

      implicit none

      integer(long)                                           :: k,n
      integer(long), intent(in)                               :: i,j
      integer(long), intent(in)                               :: vdim, &
                                                                 dim, &
                                                                 ndim
      complex(dop), dimension(vdim,dim,dim,ndim), intent(out) :: tensor
      complex(dop), dimension(vdim,ndim), intent(in)          :: matrix

      do n=1,ndim
         do k=1,vdim
            tensor(k,i,j,n) = dconjg(matrix(k,n))
         enddo
      enddo

      return
      end subroutine filltens1c

      end module d1tens 



