!***********************************************************************
! 
!                             LA-LIB                                 
!
!  Library module containing general routines for matrix and vector   
!  operations i.e. those not classifiable as matrix*matrix (mm), 
!  matrix*vector (mv) etc.
!
! Contents:
!   mattens:   multiplies a matrix with a tensor to give a matrix
!              b(i,j)=t(i,j,k)*a(i,k)
!   mktens:    makes a tensor from the product of a complex matrix with
!              a complex vector
!              t(i,j,k)=v(i)*m(j,k)
!   mktens_zd: makes a tensor from the product of a complex matrix with
!              a real vector
!              t(i,j,k)=v(i)*m(j,k)
!   normspf:   Normalizes a number of complex vectors (e.g. all 
!              single-particle functions of a specified degree of
!              freedom) simultaneously
!   normspfr:  Normalizes a number of real vectors simultaneously
!   vrminmax:  finds min and max of a real*4 vector simultaneously
!   vdminmax:  finds min and max of a real*8 vector simultaneously
!   simtranfd: similarity transformation C = B A B^T
!   simtranbd: similarity transformation C = B^T A B
!   simtranfz: similarity transformation C = B A B^T
!   simtranbz: similarity transformation C = B^T A B
!   simtranbdz: similarity transformation C = B^T A B. B real, A,C complex
!   simtran1fz: similarity transformation C = B A B^T ; A, C different sizes
!   simtran1bz: similarity transformation C = B^T A B ; A, C different sizes
!   simtran2fz: similarity transformation C = D A B^T ; A, B,C,D  
!                                                       different sizes
!   simtran2bz: similarity transformation C = D^T A B ; A, B, C, D 
!                                                       different sizes
!   simtran1btz: similarity transformation C = B^* A B^T ; A, C 
!                                                          different sizes
!   simtran2btz: similarity transformation C = D^* A B^T ; A, B, C, D 
!                                                          different sizes
!   simtran2tz: similarity transformation C = D A B ; A, B, C, D 
!                                                     different sizes
!   tranqxtzz: transformation C = B^* A B^T
!   tranhxtzz: transformation C = B^* A B^T where A,B and C are Hermitian
!   tranmxtzz: transformation C = B^* A B1^T
!   tranmxxzz: transformation C = B A B1^dag
!   tranmxazz: transformation C = B^dag A B1
!   tran1mxazz: transformation C = B^dag A B1, A, C different sizes
!
!-----------------------------------------------------------------------
!
!***********************************************************************

      module lalib

      use decimal, only: dop, long, sip

      implicit none
      private
      public :: mattens, mktens, mktens_zd, normspf, normspfr, &
                vrminmax, vdminmax, simtranfd, simtranbd, simtranfz, &
                simtranbz, simtranbdz, tranqxtzz, tranhxtzz, tranmxtzz, &
                tranmxxzz, tranmxazz, simtran1fz, simtran2fz,simtran1bz,&
                simtran2bz, simtran1btz, simtran2btz, simtran2tz, tran1mxazz
      contains 

!-----------------------------------------------------------------------
! Library subroutine mattens
!
! Multiplication of a complex matrix with a complex tensor of third
! order:
!           B(i,j) = Sum_k[ T(i,j,k)*A(i,k) ]
!           i=1,...,p;  j=1,...,q;  k=1,...,r
!
! Input-variables:  A     : complex matrix
!                   T     : complex tensor of third order
!                   p,q,r : dimensions of the different arrays
!                   add   : if true, product is added to B
! Output-variables: B     : resulting complex matrix
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in mattens, please do them in dwtmattenspar
! too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine mattens (B,A,T,p,q,r,add)

      implicit none

      integer(long)                                 :: i, j, k 
      integer(long), intent(in)                     :: p, q, r
      complex(dop), dimension(p,r), intent(in)      :: A
      complex(dop), dimension(p,q,r), intent(in)    :: T
      complex(dop), dimension(p,q), intent(inout)   :: B
      logical(kind=4), intent(in)                   :: add

      if (.not. add) then
         do j=1,q
            do i=1,p
               B(i,j)=(0.0d0,0.0d0)
            enddo
         enddo
      endif

      do j=1,q
         do k=1,r
            do i=1,p
               B(i,j)=B(i,j)+T(i,j,k)*A(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine mattens
      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Library subroutine mktens
!
! Makes a tensor of third order by elementwise multiplying a vector with
! a matrix:
!           T(i,j,k) = V(i)*M(j,k)   (no summation)
!           i=1,...,a;  j=1,...,b;  k=1,...,c
!
! Input-variables:  V     : complex vector
!                   M     : complex matrix
!                   a,b,c : dimensions of the different arrays
!                   add   : if true, result is added to T
! Output-variables: T     : resulting complex tensor of third order
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine mktens (T,V,M,a,b,c,add)

use openmpmod

      implicit none

      integer(long)                                 :: i, j, k 
      integer(long), intent(in)                     :: a, b, c
      complex(dop), dimension(a), intent(in)        :: V
      complex(dop), dimension(b,c), intent(in)      :: M
      complex(dop), dimension(a,b,c), intent(inout) :: T
      logical(kind=4), intent(in)                   :: add

      if (.not. add) then
         T=(0.0d0,0.0d0)
      endif

!$omp parallel do if(lompthread) num_threads(ompthread) private(i,j,k)
      do k=1,c
         do j=1,b
            do i=1,a
               T(i,j,k)=T(i,j,k)+V(i)*M(j,k)
            enddo
         enddo
      enddo
!$omp end parallel do

      return
      end subroutine mktens

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Library subroutine mktens_zd
!
! Makes a tensor of third order by elementwise multiplying a real vector by
! a complex matrix:
!           T(i,j,k) = V(i)*M(j,k)   (no summation)
!           i=1,...,a;  j=1,...,b;  k=1,...,c
!
! Input-variables:  V     : complex vector
!                   M     : complex matrix
!                   a,b,c : dimensions of the different arrays
!                   add   : if true, result is added to T
! Output-variables: T     : resulting complex tensor of third order
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine mktens_zd (T,V,M,a,b,c,add)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: a, b, c
      real(dop), dimension(a), intent(in)           :: V
      complex(dop), dimension(b,c), intent(in)      :: M
      complex(dop), dimension(a,b,c), intent(inout) :: T
      logical(kind=4), intent(in)                   :: add

      if (.not. add) then
         do k=1,c
            do j=1,b
               do i=1,a
                  T(i,j,k)=(0.0d0,0.0d0)
               enddo
            enddo
         enddo
      endif
      do k=1,c
         do j=1,b
            do i=1,a
               T(i,j,k)=T(i,j,k)+V(i)*M(j,k)
            enddo
         enddo
      enddo

      return
      end subroutine mktens_zd

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!           Subroutine       NORMSPF
!
!   Normalizes a number of complex vectors (e.g. all single-particle 
!   functions of a specified degree of freedom) simultaneously
!
!   may/96  AJ
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine normspf(psi,subdim,dim)

      implicit none
      
      integer(long)                                      :: e, g
      integer(long), intent(in)                          :: subdim, dim
      real(dop)                                          :: norm
      complex(dop), dimension(subdim,dim), intent(inout) :: psi

      do e=1,dim
         norm=0d0
         do g=1,subdim
            norm=norm+dble(dconjg(psi(g,e))*psi(g,e))
         enddo
         norm=1d0/sqrt(norm)
         do g=1,subdim
            psi(g,e)=norm*psi(g,e)
         enddo
      enddo
 
      return
      end subroutine normspf

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!           Subroutine       NORMSPFR
!
!   Normalizes a number of real vectors simultaneously
!
!   may/96  AJ
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine normspfr(rvec,subdim,dim)

      implicit none
      
      integer(long)                                   :: e, g
      integer(long), intent(in)                       :: subdim, dim
      real(dop), dimension(subdim,dim), intent(inout) :: rvec
      real(dop)                                       :: norm

      do e=1,dim
         norm=0d0
         do g=1,subdim
            norm=norm+rvec(g,e)*rvec(g,e)
         enddo
         norm=1d0/sqrt(norm)
         do g=1,subdim
            rvec(g,e)=norm*rvec(g,e)
         enddo
      enddo
 
      return
      end subroutine normspfr

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!    SUBROUTINE VRMINMAX      
!     finds the minimum and the maximum of a real*4 vector v
!
!     v (i)
!     n (i)
!     vmin, vmax (o)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine vrminmax(v,n,vmin,vmax)

      implicit none

      integer(long)                       :: i, imax
      integer(long), intent(in)           :: n
      real(sip), dimension(n), intent(in) :: v
      real(sip), intent(out)              :: vmin
      real(sip), intent(out)              :: vmax
      real(sip)                           :: x

      imax=1
      vmin=v(1)
      vmax=vmin

      do i=2,n
         x=v(i)
         if (x.lt.vmin) then
            vmin=x
         elseif (x.gt.vmax) then
            imax=i
            vmax=x
         endif
      enddo

      return
      end subroutine vrminmax


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!     SUBROUTINE VDMINMAX
!
!     finds the minimum and the maximum of a real*8 vector v
!
!     v (i)
!     n (i)
!     vmin, vmax (o)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine vdminmax(v,n,vmin,vmax,vavr,vrms)

      implicit none

      integer(long)                       :: i
      integer(long), intent(in)           :: n
      real(dop), dimension(n), intent(in) :: v
      real(dop), intent(out)              :: vmin
      real(dop), intent(out)              :: vmax
      real(dop), intent(out)              :: vavr
      real(dop), intent(out)              :: vrms
      real(dop)                           :: x

      vmin=v(1)
      vmax=vmin
      vavr=vmin
      vrms=vmin**2

      do i=2,n
         x=v(i)
         if (x.lt.vmin) then
            vmin=x
         elseif (x.gt.vmax) then
            vmax=x
         endif
         vavr=vavr+x
         vrms=vrms+x**2
      enddo
      vavr=vavr/dble(n)
      vrms=sqrt(vrms/dble(n))

      return
      end subroutine vdminmax
      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRANFD
!
! similarity transformation of a real matrix by an orthogonal
! matrix, i.e. C = B A B^T
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtranfd(a,b,c,dim)

      implicit none

      integer(long)                              :: i, j, k
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c
      real(dop), dimension(dim,dim)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*b(i,k)
            enddo
         enddo
      enddo
      do i=1,dim
         do k=1,dim
            do j=1,dim
               c(j,i)=c(j,i) + b(j,k)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtranfd

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRANBD
!
! similarity transformation of a real matrix by an orthogonal
! matrix, i.e. C = B^T A B
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtranbd(a,b,c,dim)

      implicit none

      integer(long)                              :: i, j, k
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c
      real(dop), dimension(dim,dim)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim
         do j=1,dim
            do k=1,dim
               c(j,i)=c(j,i) + b(k,j)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtranbd

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRANFZ
!
! similarity transformation of a complex matrix by a complex 
! matrix, i.e. C = B A B^dag
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtranfz(a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c
      complex(dop), dimension(dim,dim)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo
      do i=1,dim
         do k=1,dim
            do j=1,dim
               c(j,i)=c(j,i) + b(j,k)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtranfz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRANBZ
!
! similarity transformation of a complex matrix by a complex
! matrix, i.e. C = B^dag A B
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtranbz(a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c
      complex(dop), dimension(dim,dim)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim
         do j=1,dim
            do k=1,dim
               c(j,i)=c(j,i) + dconjg(b(k,j))*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtranbz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRANBDZ
!
! similarity transformation of a complex matrix by a complex
! matrix, i.e. C = B^dag A B
! B real, A,C complex
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtranbdz(a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)     :: b
      complex(dop), dimension(dim,dim), intent(out) :: c
      complex(dop), dimension(dim,dim)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim
         do j=1,dim
            do k=1,dim
               c(j,i)=c(j,i) + b(k,j)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtranbdz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN1FZ
!
! similarity transformation of a complex matrix by a complex matrix 
! i.e. C(dim,dim) = B(dim,dim1) A(dim1,dim1) B(dim,dim1)^dag 
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran1fz(a,b,c,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      complex(dop), dimension(dim,dim1), intent(in)   :: b
      complex(dop), dimension(dim,dim), intent(out)   :: c
      complex(dop), dimension(dim1,dim)               :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim1
            do j=1,dim1
               tmp(j,i)=tmp(j,i) + a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo
      do i=1,dim
         do k=1,dim1
            do j=1,dim
               c(j,i)=c(j,i) + b(j,k)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran1fz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN2FZ
!
! similarity transformation of a complex matrix by a complex
! matrix with a change of dimensionality, 
! i.e. C(dim2,dim1) = D(dim2,dim) A(dim,dim3) B(dim1,dim3)^dag
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran2fz(d,a,b,c,dim3,dim2,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1,dim2,dim3
      complex(dop), dimension(dim,dim3), intent(in)   :: a
      complex(dop), dimension(dim1,dim3), intent(in)  :: b
      complex(dop), dimension(dim2,dim1), intent(out) :: c
      complex(dop), dimension(dim2,dim), intent(in)   :: d
      complex(dop), dimension(dim,dim1)               :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim1
         do k=1,dim3
            do j=1,dim
               tmp(j,i)=tmp(j,i) + a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo
      do i=1,dim1
         do k=1,dim 
            do j=1,dim2 
               c(j,i)=c(j,i) + d(j,k)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran2fz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN1BZ
!
! similarity transformation of a complex matrix by a complex
! matrix with a change of dimensionality, 
! i.e. C(dim,dim) = B(dim1,dim)^dag A(dim1,dim1) B(dim1,dim)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran1bz(a,b,c,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      complex(dop), dimension(dim1,dim), intent(in)   :: b
      complex(dop), dimension(dim,dim), intent(out)   :: c
      complex(dop), dimension(dim1,dim)               :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim1
            do j=1,dim1
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim
         do j=1,dim
            do k=1,dim1
               c(j,i)=c(j,i) + dconjg(b(k,j))*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran1bz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN1BTZ
!
! similarity transformation of a complex matrix by the transpose of a complex
! matrix with a change of dimensionality, 
! i.e. C(dim,dim) = B(dim,dim1)^dag A(dim1,dim1) B(dim,dim1)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran1btz(a,b,c,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      complex(dop), dimension(dim,dim1), intent(in)   :: b
      complex(dop), dimension(dim,dim), intent(out)   :: c
      complex(dop), dimension(dim1,dim)               :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim
         do k=1,dim1
            do j=1,dim1
               tmp(j,i)=tmp(j,i) + a(j,k)*b(i,k)
            enddo
         enddo
      enddo
      do i=1,dim
         do k=1,dim1
            do j=1,dim
               c(j,i)=c(j,i) + dconjg(b(j,k))*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran1btz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN2BZ
!
! similarity transformation of a complex matrix by the adjoint of a complex
! matrix with a change of dimensionality, 
! i.e. C(dim,dim1) = D(dim2,dim)^* A(dim2,dim3) B(dim3,dim1)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran2bz(d,a,b,c,dim3,dim2,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1,dim2,dim3
      complex(dop), dimension(dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim3,dim1), intent(in)  :: b
      complex(dop), dimension(dim,dim1), intent(out)  :: c
      complex(dop), dimension(dim2,dim), intent(in)   :: d
      complex(dop), dimension(dim2,dim1)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim1
         do k=1,dim3
            do j=1,dim2
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim1
         do j=1,dim
            do k=1,dim2
               c(j,i)=c(j,i) + dconjg(d(k,j))*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran2bz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN2BTZ
!
! similarity transformation of a complex matrix by the transpose of a complex
! matrix with a change of dimensionality, 
! i.e. C(dim,dim1) = D(dim,dim2)^* A(dim2,dim3) B(dim1,dim3)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran2btz(d,a,b,c,dim3,dim2,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1,dim2,dim3
      complex(dop), dimension(dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim1,dim3), intent(in)  :: b
      complex(dop), dimension(dim,dim1), intent(out)  :: c
      complex(dop), dimension(dim,dim2), intent(in)   :: d
      complex(dop), dimension(dim2,dim1)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim1
         do k=1,dim3
            do j=1,dim2
               tmp(j,i)=tmp(j,i) + a(j,k)*b(i,k)
            enddo
         enddo
      enddo
      do i=1,dim1
         do k=1,dim2
            do j=1,dim
               c(j,i)=c(j,i) + dconjg(d(j,k))*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran2btz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! SIMTRAN2TZ
!
! special "similarity transformation" of a complex matrix by a complex
! matrix with a change of dimensionality, 
! i.e. C(dim,dim1) = D(dim2,dim) A(dim2,dim3) B(dim3,dim1)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine simtran2tz(d,a,b,c,dim3,dim2,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim,dim1,dim2,dim3
      complex(dop), dimension(dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim3,dim1), intent(in)  :: b
      complex(dop), dimension(dim,dim1), intent(out)  :: c
      complex(dop), dimension(dim2,dim), intent(in)   :: d
      complex(dop), dimension(dim2,dim1)              :: tmp

      tmp=0.0d0
      c=0.0d0
      do i=1,dim1
         do k=1,dim3
            do j=1,dim2
               tmp(j,i)=tmp(j,i) + a(j,k)*b(k,i)
            enddo
         enddo
      enddo
      do i=1,dim1
         do j=1,dim
            do k=1,dim2
               c(j,i)=c(j,i) + d(k,j)*tmp(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine simtran2tz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRANQXTZZ
!
! similarity transformation of a complex matrix by the transpose of a
! complex matrix, i.e. C = B^* A B^T
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tranqxtzz(a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k, l
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i=1,dim
         do j=1,dim
            c(j,i)=0.0d0
            do k=1,dim
               do l=1,dim
                  c(j,i)=c(j,i) + dconjg(b(j,l))*a(l,k)*b(i,k)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine tranqxtzz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRANHXTZZ
!
! similarity transformation of a hermitian matrix by the transpose of a
! complex matrix, i.e. C = B^* A B^T
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tranhxtzz(a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k, l
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i=1,dim
         c(i,i)=0.0d0
         do k=1,dim
            do l=1,dim
               c(i,i)=c(i,i) + dconjg(b(i,l))*a(l,k)*b(i,k)
            enddo
         enddo

         do j=i+1,dim
            c(j,i)=0.0d0
            do k=1,dim
               do l=1,dim
                  c(j,i)=c(j,i) + dconjg(b(j,l))*a(l,k)*b(i,k)
               enddo
            enddo
            c(i,j)= dconjg(c(j,i))
         enddo
      enddo

      return
      end subroutine tranhxtzz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRANMXTZZ
!
! similarity transformation of a non-symmetric complex matrix by the
! transpose of two complex matrices, i.e. C = B^* A B1^T
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tranmxtzz(b,a,b1,c,dim,dim1)

      implicit none

      integer(long)                                   :: i, j, k, l
      integer(long), intent(in)                       :: dim, dim1
      complex(dop), dimension(dim,dim1), intent(in)   :: a
      complex(dop), dimension(dim,dim), intent(in)    :: b
      complex(dop), dimension(dim1,dim1), intent(in)  :: b1
      complex(dop), dimension(dim,dim1), intent(out)  :: c

      do i=1,dim1
         do j=1,dim
            c(j,i)=0.0d0
            do k=1,dim1
               do l=1,dim
                  c(j,i)=c(j,i) + dconjg(b(j,l))*a(l,k)*b1(i,k)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine tranmxtzz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRANMXXZZ
!
! forward similarity transformation of a non-symmetric complex matrix by
! two complex matrices, i.e. C = B A B1^dag
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tranmxxzz(b,a,b1,c,dim,dim1)

      implicit none

      integer(long)                                   :: i, j, k, l
      integer(long), intent(in)                       :: dim, dim1
      complex(dop), dimension(dim,dim1), intent(in)   :: a
      complex(dop), dimension(dim,dim), intent(in)    :: b
      complex(dop), dimension(dim1,dim1), intent(in)  :: b1
      complex(dop), dimension(dim,dim1), intent(out)  :: c

      do i=1,dim1
         do j=1,dim
            c(j,i)=0.0d0
            do k=1,dim1
               do l=1,dim
                  c(j,i)=c(j,i) + b(j,l)*a(l,k)*dconjg(b1(i,k))
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine tranmxxzz

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRANMXAZZ
!
! backward similarity transformation of a non-symmetric complex matrix by
! two complex matrices, i.e. C = B^dag A B1
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tranmxazz(b,a,b1,c,dim,dim1)

      implicit none

      integer(long)                                   :: i, j, k, l
      integer(long), intent(in)                       :: dim, dim1
      complex(dop), dimension(dim,dim1), intent(in)   :: a
      complex(dop), dimension(dim,dim), intent(in)    :: b
      complex(dop), dimension(dim1,dim1), intent(in)  :: b1
      complex(dop), dimension(dim,dim1), intent(out)  :: c

      do i=1,dim1
         do j=1,dim
            c(j,i)=0.0d0
            do k=1,dim1
               do l=1,dim
                  c(j,i)=c(j,i) + dconjg(b(l,j))*a(l,k)*b1(k,i)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine tranmxazz
      
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! TRAN1MXAZZ
!
! backward similarity transformation of a non-symmetric complex matrix by
! two complex matrices with a change of dimensionality, 
! i.e. C(dim,dim) = B(dim1,dim)^dag A(dim1,dim1) B1(dim1,dim)
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine tran1mxazz(b,a,b1,c,dim1,dim)

      implicit none

      integer(long)                                   :: i, j, k, l
      integer(long), intent(in)                       :: dim, dim1
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      complex(dop), dimension(dim1,dim), intent(in)   :: b
      complex(dop), dimension(dim1,dim), intent(in)   :: b1
      complex(dop), dimension(dim,dim), intent(out)   :: c

      do i=1,dim
         do j=1,dim
            c(j,i)=0.0d0
            do k=1,dim1
               do l=1,dim1
                  c(j,i)=c(j,i) + dconjg(b(l,j))*a(l,k)*b1(k,i)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine tran1mxazz
      
      end module lalib
