! **********************************************************************
!                                                                     
!                              MMLIB                                  
!                                                                     
!  Library module containing linear algebra routines that involve the 
!  multiplication of matrices
!                                                                     
!  Nomenclature:
!    Each name has 6 basic characters:
!    First 2 characters denote the objects being multiplied:
!       q: quadratic matrix
!       m: general (rectangular) matrix
!       h: hermitian matrix
!       p: positive definite matrix
!       s: symmetric matrix
!       u: unitary matrix
!       d: diagonal matrix (only diagonal elements are supplied as a
!          vector)
!       t: tensor of third order
!       v: vector
!       x: scalar
!       e.g. 'qm' denotes the operation (quadratic matrix * rectangular
!       matrix)
!    Character 3 denotes how first object is used:
!       x: unchanged from input
!       t: transpose of input
!       a: adjoint of input
!       c: complex conjugate of input
!       v: as a vector
!    Character 4 denotes how second object is used:
!       see Character 3 above.
!    Character 5, 6 denote data types of first, second object
!       respectively:
!       z: complex double precision (complex*16)
!       c: complex single precision (complex*8)
!       d: real double precision (real*8)
!       r: real single precision (real*4)
!     Further characters, if present, give more informaion:
!       a: the result is added to a further object
!       r: the result is subtracted (removed) from a further object
!       c: the input matrices commute
!       h: the resulting matrix is hermitian
!       h1: the resulting matrix is anti-hermitian
!       s: the resulting matrix is symmetric
!       1: the physical dimensions of the matrices differs from those
!          used.
!   
!  Contents:                                                          
!  In the following list of available subroutines, matrices on the LHS
!  of the definition are input, that on the RHS output. The usual
!  summation convention is used i.e. a sum is made over repeated indices
!  on the LHS (NOTE: there is only elementwise multiplication and 
!  no subsequent summation if a diagonal matrix is involved !!!).
!
!   qqxxdd (a,b,c,dim)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   qqxxzd (a,b,c,dim)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   mmxxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   mmvxzz (a,b,c,dim1,dim2,dim3,p)
!       Definition: a(p,k)*b(k,i) = c(i)
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim3)
!
!   mmxtzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim3,dim2),c(dim1,dim3)
!
!   mmtczz (a,b,c,dim1,dim2,dim3)
!       Definition: a(k,j)*dconjg(b(k,i)) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim3),c(dim2,dim3)
!
!   mmxxzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(k,i) + c(j,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   dmxxzz (a,b,c,dim1,dim2)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim1),b(dim1,dim2),c(dim1,dim2)
!                                                                    
!   dmxxzza (a,b,c,dim1,dim2)
!       Definition: a(j)*b(j,i) + c(j,i) = c(j,i) .
!       Dimensions: a(dim1),b(dim1,dim2),c(dim1,dim2)
!                                                                    
!   dmxxdz (a,b,c,dim1,dim2)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim1),b(dim1,dim2),c(dim1,dim2)
!
!   dmxxdd (a,b,c,dim1,dim2)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim1),b(dim1,dim2),c(dim1,dim2)
!
!   ddxxdd (a,b,c,dim)
!       Definition: a(j)*b(j) = c(j) .
!       Dimensions: a(dim),b(dim),c(dim)
!
!   ddxxdz (a,b,c,dim)
!       Definition: a(j)*b(j) = c(j) .
!       Dimensions: a(dim),b(dim),c(dim)
!
!   dqxxzz (a,b,c,dim)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim),b(dim,dim),c(dim,dim)
!
!   dqxxdd (a,b,c,dim)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim),b(dim,dim),c(dim,dim)
!
!   dqxxdd1 (a,b,c,phdim,dim)
!       Definition: a(j)*b(j,i) = c(j,i) .
!       Dimensions: a(dim),b(phdim,dim),c(phdim,dim)
!
!   dqxxdz (a,v,w,dim)
!       Definition: a(i)*v(i,j) = w(i,j) .
!       Dimensions: a(dim),v(dim,dim),w(dim,dim)
!
!   dqxxdz2 (a,v,w,dim)
!       Definition: a(j)*v(i,j) = w(i,j) .
!       Dimensions: a(dim),v(dim,dim),w(dim,dim)
!
!C  mmxxdz (a,b,c,dim1,dim2,dim3)
!C      Definition: a(j,k)*b(k,i) = c(j,i) .
!C      Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   mmtxdd (a,b,c,dim1,dim2,dim3)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim2,dim1),b(dim2,dim3),c(dim1,dim3)
!
!   mmtxdd1 (a,b,c,phdim,dim1,dim2,dim3)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim2,dim1),b(dim2,dim3),c(dim1,dim3) (= used dim.)
!       Dimensions: a,b,c(phdim,phdim) (= allocated dimension)
!
!   mmxtdd (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim3,dim2),c(dim1,dim3)
!
!   mmxxdd (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   mmxxdd1 (a,b,c,phdim,dim1,dim2,dim3)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3) (= used dim.)
!       Dimensions: a,b,c(phdim,phdim) (= allocated dimension)
!
!   qqxtdd (a,b,c,dim)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   qqxtdd1 (a,b,c,phdim,dim)
!       Definition: a(j,k)*b(i,k) = c(j,i) ; 1 <= i,j,k <= dim .
!       Dimensions: a(phdim,dim),b(phdim,dim),c(phdim,dim)
!
!   qqtxdd (a,b,c,dim)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   qdxxzz (a,b,c,dim)
!       Definition: a(j,i)*b(i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim),c(dim,dim)
!
!   qdxxdd (a,b,c,dim)
!       Definition: a(j,i)*b(i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim),c(dim,dim)
!
!   qdxxdd1 (a,b,c,phdim,dim)
!       Definition: a(j,i)*b(i) = c(j,i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),b(dim),c(phdim,dim)
!
!   hhxtzzc (a,b,c,dim)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   trhhxtzz (a,b,c,dim)
!       Definition: tr(a(j,k)*b(j,k)) = c . (c is real!)
!       Dimensions: a(dim,dim),b(dim,dim),c
!
!   mmaxzzh (a,b,c,dim1,dim2)
!       Definition: dconjg(a(k,j))*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim2),c(dim2,dim2)
!
!   mmaxzzh1 (a,b,c,dim1,dim2)
!       Definition: dconjg(a(k,j))*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim2),c(dim2,dim2)
!
!   mmtxzzs (a,b,c,dim1,dim2)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim2),c(dim2,dim2)
!
!   qmxxzz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim1),b(dim1,dim2),c(dim1,dim2)
!
!   qqxxdd1 (a,b,c,phdim,dim)
!       Definition: a(j,k)*b(k,i) = c(j,i) ; 1 <= i,j,k <= dim .
!       Dimensions: a(phdim,dim),b(phdim,dim),c(phdim,dim)
!
!C  mmxxzz1 (a,b,c,phdim1,phdim2,phdim3,dim1,dim2,dim3)
!C      Definition: a(j,k)*b(k,i) = c(j,i) .
!C      Dimensions: a(phdim1,phdim2),b(phdim2,phdim3),c(phdim1,phdim3)
!
!   mmaxzz (a,b,c,dim1,dim2,dim3)
!       Definition: dconjg(a(k,j))*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim3),c(dim2,dim3)
!
!   mmcxzz (a,b,c,dim1,dim2,dim3)
!       Definition: dconjg(a(j,k))*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   qqcxzz (a,b,c,dim)
!       Definition: dconjg(a(j,k))*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   mmxazz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*dconjg(b(i,k)) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim3,dim2),c(dim1,dim3)
!
!   mmxtzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*b(i,k) +c(j,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim3,dim2),c(dim1,dim3)
!
!   mqxtzza (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(i,k) +c(j,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   mmcazza (a,b,c,dim1,dim2,dim3)
!       Definition: dconjg(a(j,k))*dconjg(b(i,k)) + c(j,i) = c(j,i)
!       Dimensions: a(dim1,dim2),b(dim3,dim2),c(dim1,dim3/)
!
!   mqxxzza (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(k,i) +c(j,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   mqxtzd (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   qqtxzz (a,b,c,dim)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   qqtxdz (a,b,c,dim)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   mqxazz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*dconjg(b(i,k)) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   qmxxdz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim1),b(dim1,dim2),c(dim1,dim2)
!
!   qmtxdz (a,b,c,dim1,dim2)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim1),b(dim1,dim2),c(dim1,dim2)
!
!   qqxxzz (a,b,c,dim)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
!
!   mqxtzz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(i,k) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   mqxxzz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   mmtxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(k,j)*b(k,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim3),c(dim2,dim3)
!
!   mmtxzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(k,j)*b(k,i) + c(j,i) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim1,dim3),c(dim2,dim3)
!                                                                  
!   mmxczz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,k)*dconjg(b(k,i)) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim3),c(dim1,dim3)
!
!   mqxczz (a,b,c,dim1,dim2)
!       Definition: a(j,k)*dconjg(b(k,i)) = c(j,i) .
!       Dimensions: a(dim1,dim2),b(dim2,dim2),c(dim1,dim2)
!
!   qqaxzz (a,b,c,dim)
!       Definition: dconjg(a(k,j))*b(k,i) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)      
!
!   qqxazz (a,b,c,dim)
!       Definition: a(j,k)*dconjg(b(i,k)) = c(j,i) .
!       Dimensions: a(dim,dim),b(dim,dim),c(dim,dim)
! 
!   msqzz (a,b,dim1,dim2,dim3)
!       Definition: Local control: generating vibrational wavefuntion
!       Dimensions: a(dim1,dim2),b(dim1,dim2)
!***********************************************************************

      module mmlib

      use decimal, only: dop, long

      implicit none
      private
      public :: qqxxdd, qqxxzd, mmxxzz, mmvxzz, mmxtzz, mmtczz, &
                mmxxzza, dmxxzz, dmxxzza, dmxxdz, dmxxdd, &
                ddxxdd, ddxxdz,dqxxzz, dqxxdd, dqxxdd1, dqxxdz, dqxxdz2, &
                mmtxdd, mmtxdd1, mmxtdd, mmxxdd, mmxxdd1, &
                qqxtdd, qqxtdd1, qqtxdd, qdxxzz, qdxxdd, qdxxdd1, &
                hhxtzzc, trhhxtzz,mmaxzzh, mmaxzzh1, mmtxzzs, qmxxzz, &
                qqxxdd1, mmaxzz, mmcxzz, qqcxzz, mmxazz, mmxtzza, qdxxdz, &
                mmcazza, mqxxzz, mqxxzza, mqxtzd, qqtxzz, qqtxdz, &
                mqxazz, qmxxdz, qmtxdz, qqxxzz, mqxtzz, mmtxzz, &
                mmtxzza, mmxczz, mqxczz, qqaxzz, qqxazz, msqzz
!               mmxxdz, mmxxzz1, mqxtzza
      contains

!-----------------------------------------------------------------------
! Library subroutine qqxxdd
!
! Multiplication of two real quadratic matrices:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqxxdd (a,b,c,dim)

      implicit none

      integer(long)                              :: i, j, k
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c

      do i=1,dim
         do j=1,dim
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim
         do k = 2,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qqxxzd
!
! Multiplication of a complex quadratic matrix with a real quadratic
! matrix:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqxxzd (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim,dim), intent(in)     :: b
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i=1,dim
         do j=1,dim
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim
         do k = 2,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxxzz
!
! Multiplication of two complex rectangular matrices:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, k
      integer(long), intent(in)                        :: dim1, dim2, dim3
      complex(dop), dimension(dim1,dim2), intent(in)   :: a
      complex(dop), dimension(dim2,dim3), intent(in)   :: b
      complex(dop), dimension(dim1,dim3), intent(out)  :: c

      do i=1,dim3
         do j=1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim3
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! subroutine mmvxzz
!
! Multiplication of two complex rectangular matrices with p fixed (uses
! first matrix as a vector):
! 
!     a(p,k)*b(k,i) = c(i)
!-----------------------------------------------------------------------

      subroutine mmvxzz (a,b,c,dim1,dim2,dim3,p)

      implicit none

      integer(long)                                  :: i, p, k
      integer(long), intent(in)                      :: dim1, dim2, dim3
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim2,dim3), intent(in) :: b
      complex(dop), dimension(dim3), intent(out)     :: c

      do i=1,dim3
         c(i) = a(p,1)*b(1,i)
      enddo
      do i = 1,dim3
         do k = 2,dim2
            c(i) = c(i)+a(p,k)*b(k,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxtzz
!
! Multiplication of a complex rectangular matrix with the transpose of
! a rectangular complex matrix
!     a(j,k)*b(i,k) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxtzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, k
      integer(long), intent(in)                        :: dim1, dim2, dim3
      complex(dop), dimension(dim1,dim2), intent(in)   :: a
      complex(dop), dimension(dim3,dim2), intent(in)   :: b
      complex(dop), dimension(dim1,dim3), intent(out)  :: c

      do i=1,dim3
         do j=1,dim1
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo

      do i = 1,dim3
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmxxzza
!
! Multiplication of a complex rectangular matrix with a rectangular 
! complex matrix, the result of which is added to a further matrix.
!     a(j,k)*b(k,i) +c(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxxzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, k
      integer(long), intent(in)                        :: dim1, dim2, dim3
      complex(dop), dimension(dim1,dim2), intent(in)   :: a
      complex(dop), dimension(dim2,dim3), intent(in)   :: b
      complex(dop), dimension(dim1,dim3), intent(out)  :: c

      do k = 1,dim2
         do i = 1,dim3
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dmxxzz
!
! Multiplication of a diagonal complex matrix with a complex rectangular
! matrix:
!     a(j)*b(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dmxxzz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1), intent(in)       :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine dmxxzza
!
! Multiplication of a diagonal complex matrix with a complex rectangular
! matrix:
!     a(j)*b(j,i) + c(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dmxxzza (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      complex(dop), dimension(dim1), intent(in)         :: a
      complex(dop), dimension(dim1,dim2), intent(in)    :: b
      complex(dop), dimension(dim1,dim2), intent(inout) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = c(j,i) + a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dmxxdz
!
! Multiplication of a diagonal real matrix with a real rectangular
! matrix:
!     a(j)*b(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dmxxdz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      real(dop), dimension(dim1), intent(in)          :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dmxxdd
!
! Multiplication of a diagonal real matrix with a real rectangular
! matrix:
!     a(j)*b(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dmxxdd (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(dop), dimension(dim1), intent(in)       :: a
      real(dop), dimension(dim1,dim2), intent(in)  :: b
      real(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dqxxzz
!
! Multiplication of a diagonal complex matrix with a complex quadratic 
! matrix:
!     a(j)*b(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dqxxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim), intent(in)      :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine ddxxdd
!
! Multiplication of a diagonal real matrix with a diagonal real matrix:
!     a(j)*b(j) = c(j)
!-----------------------------------------------------------------------

      subroutine ddxxdd (a,b,c,dim)

      implicit none

      integer(long)                          :: j
      integer(long), intent(in)              :: dim
      real(dop), dimension(dim), intent(in)  :: a
      real(dop), dimension(dim), intent(in)  :: b
      real(dop), dimension(dim), intent(out) :: c

      do j = 1,dim
         c(j) = a(j)*b(j)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine ddxxdz
!
! Multiplication of a diagonal real matrix with a diagonal complex matrix:
!     a(j)*b(j) = c(j)
!-----------------------------------------------------------------------

      subroutine ddxxdz (a,b,c,dim)

      implicit none

      integer(long)                             :: j
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: a
      complex(dop), dimension(dim), intent(in)  :: b
      complex(dop), dimension(dim), intent(out) :: c

      do j = 1,dim
         c(j) = a(j)*b(j)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dqxxdd
!
! Multiplication of a diagonal real matrix with a real quadratic matrix:
!     a(j)*b(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine dqxxdd (a,b,c,dim)

      implicit none

      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim), intent(in)      :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dqxxdd1
!
! Multiplication of a diagonal real matrix with a real quadratic matrix:
!     a(j)*b(j,i) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine dqxxdd1 (a,b,c,phdim,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim, phdim
      real(dop), dimension(dim), intent(in)        :: a
      real(dop), dimension(phdim,dim), intent(in)  :: b
      real(dop), dimension(phdim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j)*b(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine dqxxdz
!
! Multiplication of a diagonal real matrix with a complex matrix
!     a(i)*v(i,j)=w(i,j)
!-----------------------------------------------------------------------

      subroutine dqxxdz (a,v,w,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: a
      complex(dop), dimension(dim,dim), intent(in)  :: v
      complex(dop), dimension(dim,dim), intent(out) :: w

      do j=1,dim
         do i=1,dim
            w(i,j) = a(i)*v(i,j)
         enddo
      enddo

      return
      end subroutine


! ----------------------------------------------------------------------
! Library subroutine dqxxdz2
!
! Multiplication of a diagonal real matrix with a complex matrix 
!     a(j)*v(i,j)=w(i,j)
!-----------------------------------------------------------------------

      subroutine dqxxdz2 (a,v,w,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: a
      complex(dop), dimension(dim,dim), intent(in)  :: v
      complex(dop), dimension(dim,dim), intent(out) :: w

      do j=1,dim
         do i=1,dim
            w(i,j) = a(j)*v(i,j)
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmxxdz
!
! Multiplication of a rectangular real matrix with a rectangular complex
! matrix:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

!     subroutine mmxxdz (a,b,c,dim1,dim2,dim3)

!     implicit none

!     integer(long)                                   :: i, j, k
!     integer(long), intent(in)                       :: dim1, dim2, dim3
!     real(dop), dimension(dim1,dim2), intent(in)     :: a
!     complex(dop), dimension(dim2,dim3), intent(in)  :: b
!     complex(dop), dimension(dim1,dim3), intent(out) :: c

!     do i=1,dim3
!        do j=1,dim1
!           c(j,i) = a(j,1)*b(1,i)
!        enddo
!     enddo
!     do i = 1,dim3
!        do k = 2,dim2
!           do j = 1,dim1
!              c(j,i) = c(j,i)+a(j,k)*b(k,i)
!           enddo
!        enddo
!     enddo

!     return
!     end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmtxdd
!
! Multiplication of the transpose of a rectangular real matrix with a 
! rectangular real matrix:
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmtxdd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                :: i, j, k
      integer(long), intent(in)                    :: dim1, dim2, dim3
      real(dop), dimension(dim2,dim1), intent(in)  :: a
      real(dop), dimension(dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim1
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim2
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxtdd
!
! Multiplication of the rectangular real matrix with the transpose of a
! rectangular real matrix:
!       a(j,k)*b(i,k) = c(j,i)
!-----------------------------------------------------------------------
 
      subroutine mmxtdd(a,b,c,dim1,dim2,dim3)
 
      implicit none
 
      integer(long)                                :: i, j, k
      integer(long), intent(in)                    :: dim1, dim2, dim3
      real(dop), dimension(dim1,dim2), intent(in)  :: a
      real(dop), dimension(dim3,dim2), intent(in)  :: b
      real(dop), dimension(dim1,dim3), intent(out) :: c
 
      do i = 1,dim3
         do j = 1,dim1
            c(j,i) = a(j,1)*b(i,1)
            do k = 2,dim2
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo
 
      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmtxdd1
!
! Multiplication of the transpose of a rectangular real matrix with a 
! rectangular real matrix:
!     a(k,j)*b(k,i) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim1-3 is used dimension 
!-----------------------------------------------------------------------

      subroutine mmtxdd1 (a,b,c,phdim,dim1,dim2,dim3)

      implicit none

      integer(long)                                  :: i, j, k
      integer(long), intent(in)                      :: phdim, dim1, dim2, &
                                                        dim3
      real(dop), dimension(phdim,phdim), intent(in)  :: a
      real(dop), dimension(phdim,phdim), intent(in)  :: b
      real(dop), dimension(phdim,phdim), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim1
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim2
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxxdd
!
! Multiplication of two real rectangular matrices:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxxdd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                :: i, j, k
      integer(long), intent(in)                    :: dim1, dim2, dim3
      real(dop), dimension(dim1,dim2), intent(in)  :: a
      real(dop), dimension(dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim3), intent(out) :: c

      do i=1,dim3
         do j=1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim3
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxxdd1
!
! Multiplication of two real rectangular matrices:
!     a(j,k)*b(k,i) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim1-3 is used dimension 
!-----------------------------------------------------------------------

      subroutine mmxxdd1 (a,b,c,phdim,dim1,dim2,dim3)

      implicit none

      integer(long)                                  :: i, j, k
      integer(long), intent(in)                      :: phdim, dim1, dim2, &
                                                        dim3
      real(dop), dimension(phdim,phdim), intent(in)  :: a
      real(dop), dimension(phdim,phdim), intent(in)  :: b
      real(dop), dimension(phdim,phdim), intent(out) :: c

      do i=1,dim3
         do j=1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim3
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qqxtdd
!
! Multiplication of a real quadratic matrix with the transpose of a real
! quadratic matrix:
!     a(j,k)*b(i,k) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqxtdd (a,b,c,dim)

      implicit none

      integer(long)                              :: i, j, k
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo
      do k = 2,dim
         do i = 1,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qqxtdd1
!
! Multiplication of a real quadratic matrix with the transpose of a real
! quadratic matrix:
!     a(j,k)*b(i,k) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qqxtdd1 (a,b,c,phdim,dim)

      implicit none

      integer(long)                                :: i, j, k
      integer(long), intent(in)                    :: phdim, dim
      real(dop), dimension(phdim,dim), intent(in)  :: a
      real(dop), dimension(phdim,dim), intent(in)  :: b
      real(dop), dimension(phdim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo
      do k = 2,dim
         do i = 1,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine qqtxdd
!
! Multiplication of the transpose of a real quadratic matrix with a
! real quadratic matrix:
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqtxdd (a,b,c,dim)

      implicit none

      integer(long)                              :: i, j, k
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(in)  :: b
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qdxxzz
!
! Multiplication of a complex quadratic matrix with a diagonal complex 
! matrix:
!     a(j,i)*b(i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qdxxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim), intent(in)      :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)*b(i)
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qdxxdz
!
! Multiplication of a complex quadratic matrix with a diagonal real
! matrix:
!     a(j,i)*b(i) = c(j,i)
!-----------------------------------------------------------------------
 
      subroutine qdxxdz (a,b,c,dim)
 
      implicit none
 
      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim), intent(in)         :: b
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c
 
      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)* b(i)
         enddo
      enddo
 
      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qdxxdd
!
! Multiplication of a real quadratic matrix with a diagonal real matrix:
!     a(j,i)*b(i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qdxxdd (a,b,c,dim)

      implicit none

      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim), intent(in)      :: b
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)*b(i)
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qdxxdd1
!
! Multiplication of a real quadratic matrix with a diagonal real matrix:
!     a(j,i)*b(i) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qdxxdd1 (a,b,c,phdim,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: phdim, dim
      real(dop), dimension(phdim,dim), intent(in)  :: a
      real(dop), dimension(dim), intent(in)        :: b
      real(dop), dimension(phdim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)*b(i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine hhxtzzc
!
! Multiplication of a complex hermitian matrix with the transpose of a 
! complex hermitian matrix, where the two matrices commute:
!     a(j,k)*b(i,k) = c(j,i)
!
! NB The fact that the two matrices commute means that the result of
!    the multiplication is also hermitian
!-----------------------------------------------------------------------

      subroutine hhxtzzc (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = i,dim
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo
      do k = 2,dim
         do i = 1,dim
            do j = i,dim
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo
      do i = 1,dim
         do j = 1,i-1
            c(j,i) = dconjg(c(i,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine trhhxtzz
!
! Trace of the product of a complex hermitian matrix with the transpose
! of a complex hermitian matrix, trace(A*B^T)
!     a(j,k)*b(j,k) = c
!
! NB The fact that the two matrices are hermitian means that the result
!    is real
!-----------------------------------------------------------------------

      subroutine trhhxtzz (a,b,c,dim)

      implicit none

      integer(long)            :: j,k
      integer(long),intent(in) :: dim
      real(dop), intent(out)   :: c
      complex(dop), intent(in) :: a(dim,dim),b(dim,dim)

      c=0.d0
      do j = 2,dim
         do k = 1,j-1
            c = c + 2.d0*dble(a(j,k)*b(j,k))
         enddo
      enddo
      do j = 1,dim
         c = c + dble(a(j,j))*dble(b(j,j))
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmaxzzh
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a rectangular complex matrix, where the result is a hermitian matrix
!     dconjg(a(k,j))*b(k,i) = c(j,i)
!
! NB this routine can be used for the overlap of two sets of "spfs" in
!    the same basis
!-----------------------------------------------------------------------

      subroutine mmaxzzh (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim2,dim2), intent(out) :: c

      do i = 1,dim2
         do j = i,dim2
            c(j,i) = dconjg(a(1,j))*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+dconjg(a(k,j))*b(k,i)
            enddo
         enddo
      enddo
!
! now fill in other half of hermitian matrix
!
      do i=1,dim2
         c(i,i)=dble(c(i,i))
      enddo
      do i=1,dim2
         do j=1,i-1
            c(j,i)=dconjg(c(i,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmaxzzh1
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a rectangular complex matrix, where the result is an anti- hermitian 
! matrix
!     dconjg(a(k,j))*b(k,i) = c(j,i)
!
!-----------------------------------------------------------------------

      subroutine mmaxzzh1 (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      real(dop)                                       :: x
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim2,dim2), intent(out) :: c

      do i = 1,dim2
         do j = i,dim2
            c(j,i) = dconjg(a(1,j))*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+dconjg(a(k,j))*b(k,i)
            enddo
         enddo
      enddo
!
! now fill in other half of anti-hermitian matrix
!
      do i=1,dim2
         x=dimag(c(i,i))
         c(i,i)=dcmplx(0.0d0,x)
      enddo
      do i=1,dim2
         do j=1,i-1
            c(j,i)=-dconjg(c(i,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmtxzzs
!
! Multiplication of a transposed complex rectangular matrix with
! a rectangular complex matrix, where the result is a symmetric matrix
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmtxzzs (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim2,dim2), intent(out) :: c

      do i = 1,dim2
         do j = i,dim2
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo
!
! now fill in other half of symmetric matrix
!
      do i=1,dim2
         do j=1,i-1
            c(j,i)=c(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qmxxzz
!
! Multiplication of a complex quadratic matrix with a complex 
! rectangular matrix:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qmxxzz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim2
         do k = 2,dim1
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qqxxdd1
!
! Multiplication of two real quadratic matrices:
!     a(j,k)*b(k,i) = c(j,i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qqxxdd1 (a,b,c,phdim,dim)

      implicit none

      integer(long)                                :: i, j, k
      integer(long), intent(in)                    :: phdim, dim
      real(dop), dimension(phdim,dim), intent(in)  :: a
      real(dop), dimension(phdim,dim), intent(in)  :: b
      real(dop), dimension(phdim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim
         do k = 2,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxxzz1
!
! Multiplication of two complex rectangular matrices:
!     a(j,k)*b(k,i) = c(j,i)
!
! NB phdims are physical dimensions, dims are used dimensions
!-----------------------------------------------------------------------

!     subroutine mmxxzz1 (phdim1,phdim2,phdim3,dim1,dim2,dim3,a,b,c)

!     implicit none

!     integer(long)                                       :: i, j, k
!     integer(long), intent(in)                           :: phdim1, phdim2, &
!                                                            phdim3, dim1, &
!                                                            dim2, dim3
!     complex(dop), dimension(phdim1,phdim2), intent(in)  :: a
!     complex(dop), dimension(phdim2,phdim3), intent(in)  :: b
!     complex(dop). dimension(phdim1,phdim3), intent(out) :: c

!     do i = 1,dim3
!        do j = i,dim1
!           c(j,i) = a(j,1)*b(1,i)
!        enddo
!     enddo
!     do i = 1,dim3
!        do k = 2,dim2
!           do j = 1,dim1
!              c(j,i) = c(j,i)+a(j,k)*b(k,i)
!           enddo
!        enddo
!     enddo

!     return
!     end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmaxzz
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a rectangular complex matrix
!     dconjg(a(k,j))*b(k,i) = c(j,i)
!
! NB this routine can be used for the overlap of two sets of vectors in
!    different spf bases
!-----------------------------------------------------------------------

      subroutine mmaxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim3), intent(in)  :: b
      complex(dop), dimension(dim2,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim2
            c(j,i) = dconjg(a(1,j))*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+dconjg(a(k,j))*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmtczz
!
! Multiplication of the transpose of a complex rectangular matrix with
! the complex conjugate of a rectangular complex matrix
!     a(k,j)*dconjg(b(k,i)) = c(j,i)
!
!-----------------------------------------------------------------------
      subroutine mmtczz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim3), intent(in)  :: b
      complex(dop), dimension(dim2,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim2
            c(j,i) = a(1,j)*dconjg(b(1,i))
            do k = 2,dim1
               c(j,i) = c(j,i)+a(k,j)*dconjg(b(k,i))
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmaczz
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a adjoint of a rectangular complex matrix
!     dconjg(a(k,j))*dconjg(b(k,i)) = c(j,i)
!
! NB this routine can be used for the overlap of two sets of vectors in
!    different spf bases
!-----------------------------------------------------------------------
!     subroutine mmaczz (a,b,c,dim1,dim2,dim3)

!     implicit none

!     integer(long)                                   :: i, j, k
!     integer(long), intent(in)                       :: dim1, dim2, &
!                                                        dim3
!     complex(dop), dimension(dim1,dim2), intent(in)  :: a
!     complex(dop), dimension(dim1,dim3), intent(in)  :: b
!     complex(dop), dimension(dim2,dim3), intent(out) :: c

!     do i = 1,dim3
!        do j = 1,dim2
!           c(j,i) = dconjg(a(1,j))*dconjg(b(1,i))
!           do k = 2,dim1
!              c(j,i) = c(j,i)+dconjg(a(k,j))*dconjg(b(k,i))
!           enddo
!        enddo
!     enddo

!     return
!     end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmcxzz
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a rectangular complex matrix
!     dconjg(a(j,k))*b(k,i) = c(j,i)
!
! NB this routine can be used for the overlap of two sets of vectors in
!    different spf bases
!-----------------------------------------------------------------------
      subroutine mmcxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim1
            c(j,i) = dconjg(a(j,1))*b(1,i)
            do k = 2,dim2
               c(j,i) = c(j,i)+dconjg(a(j,k))*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine qqcxzz
!
! Multiplication of the adjoint of a complex quadratic matrix with
! a quadratic complex matrix
!     dconjg(a(j,k))*b(k,i) = c(j,i)
!
!-----------------------------------------------------------------------
      subroutine qqcxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = dconjg(a(j,1))*b(1,i)
            do k = 2,dim
               c(j,i) = c(j,i)+dconjg(a(j,k))*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmaxzz
!
! Multiplication of the adjoint of a complex rectangular matrix with
! a rectangular complex matrix
!     a(j,k)*dconjg(b(i,k)) = c(j,i)
!
! NB this routine can be used for the overlap of two sets of vectors in
!    different spf bases
!-----------------------------------------------------------------------

      subroutine mmxazz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim3,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim1
            c(j,i) = 0.0d0
         enddo
      enddo

      do i = 1,dim3
         do j = 1,dim1
            do k = 1,dim2
               c(j,i) = c(j,i)+a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxtzza
!
! Multiplication of a complex rectangular matrix with the transpose of
! a rectangular complex matrix, the result of which is added to a
! further matrix.
!     a(j,k)*b(i,k) + c(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxtzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim3,dim2), intent(in)    :: b
      complex(dop), dimension(dim1,dim3), intent(inout) :: c

      do k = 1,dim2
         do i = 1,dim3
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mmcazza
!
! Multiplication of a conjugated complex rectangular matrix with the 
! adjoint of a quadratic complex matrix, the result of which is added 
! to a further rectangular matrix.
!     dconjg(a(j,k))*dconjg(b(i,k)) + c(j,i) = c(j,i)
!-----------------------------------------------------------------------
      subroutine mmcazza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim3,dim2), intent(in)    :: b
      complex(dop), dimension(dim1,dim3), intent(inout) :: c

      do k = 1,dim2
         do i = 1,dim3
            do j = 1,dim1
               c(j,i) = c(j,i)+dconjg(a(j,k))*dconjg(b(i,k))
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mqxxzza
!
! Multiplication of a complex rectangular matrix with a quadratic
! complex matrix, the result of which is added to a further rectangular
! matrix.
!     a(j,k)*b(k,i) + c(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxxzza (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim2,dim2), intent(in)    :: b
      complex(dop), dimension(dim1,dim2), intent(inout) :: c

      do i = 1,dim2
         do k = 1,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mqxtzd
!
! Multiplication of a rectangular complex matrix with the transpose of 
! a quadratic real matrix:
!     a(j,k)*b(i,k) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxtzd (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      real(dop), dimension(dim2,dim2), intent(in)     :: b
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do j=1,dim1
         do i=1,dim2
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo
      do k=2,dim2
         do j=1,dim1
            do i=1,dim2
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qqtxzz
!
! Multiplication of the transpose of a complex quadratic matrix with a
! complex quadratic matrix:
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqtxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine qqtxdz
!
! Multiplication of the transpose of a real quadratic matrix with a
! complex quadratic matrix:
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqtxdz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim,dim), intent(in)     :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mqxazz
!
! Multiplication of a complex rectangular matrix with the adjoint of 
! a quadratic complex matrix:
!     a(j,k)*dconjg(b(i,k)) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxazz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,1)*dconjg(b(i,1))
         enddo
      enddo
      do k = 2,dim2
         do i = 1,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qmxxdz
!
! Multiplication of a quadratic real matrix with a rectangular complex
! matrix:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qmxxdz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      real(dop), dimension(dim1,dim1), intent(in)     :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i=1,dim2
         do j=1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim2
         do k = 2,dim1
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qmtxdz
!
! Multiplication of the transpose of a quadratic real matrix with a 
! rectangular complex matrix:
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qmtxdz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      real(dop), dimension(dim1,dim1), intent(in)     :: a
      complex(dop), dimension(dim1,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qqxxzz
!
! Multiplication of two complex quadratic matrices:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine qqxxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i=1,dim
         do j=1,dim
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim
         do k = 2,dim
            do j = 1,dim
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine mqxtzz
!
! Multiplication of a complex rectangular matrix with the transpose of
! a quadratic complex matrix:
!     a(j,k)*b(i,k) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxtzz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,1)*b(i,1)
         enddo
      enddo
      do k = 2,dim2
         do i = 1,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(i,k)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mqxxzz
!
! Multiplication of a complex rectangular matrix with a quadratic
! complex matrix:
!     a(j,k)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxxzz (a,b,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,1)*b(1,i)
         enddo
      enddo
      do i = 1,dim2
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmtxzz
!
! Multiplication of a transposed complex rectangular matrix with
! a rectangular complex matrix
!     a(k,j)*b(k,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmtxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim3), intent(in)  :: b
      complex(dop), dimension(dim2,dim3), intent(out) :: c

      do i = 1,dim3
         do j = 1,dim2
            c(j,i) = a(1,j)*b(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmtxzza
!
! Multiplication of a transposed complex rectangular matrix with
! a rectangular complex matrix
!     a(k,j)*b(k,i) + c(j,i) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmtxzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim1,dim3), intent(in)    :: b
      complex(dop), dimension(dim2,dim3), intent(inout) :: c

      do i = 1,dim3
         do j = 1,dim2
            do k = 1,dim1
               c(j,i) = c(j,i)+a(k,j)*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mmxczz
!
! Multiplication of a complex rectangular matrices with the complex
! conjugate of a rectangular matrix:
!     a(j,k)*dconjg(b(k,i)) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mmxczz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim3), intent(out) :: c

      do i=1,dim3
         do j=1,dim1
            c(j,i) = a(j,1)*dconjg(b(1,i))
         enddo
      enddo
      do i = 1,dim3
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*dconjg(b(k,i))
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mqxczz
!
! Multiplication of a complex rectangular matrices with the complex
! conjugate of a quadratic matrix:
!     a(j,k)*dconjg(b(k,i)) = c(j,i)
!-----------------------------------------------------------------------

      subroutine mqxczz (a,b,c,dim1,dim2)

      implicit none
      
      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)  :: b
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i=1,dim2
         do j=1,dim1
            c(j,i) = a(j,1)*dconjg(b(1,i))
         enddo
      enddo
      do i = 1,dim2
         do k = 2,dim2
            do j = 1,dim1
               c(j,i) = c(j,i)+a(j,k)*dconjg(b(k,i))
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
!     Library subroutine qqaxzz                                         
!                                                                       
! Multiplication of the adjoint of a complex quadratic matrix with      
! a another quadratic complex matrix                                    
!     dconjg(a(k,j))*b(k,i) = c(j,i)                                    
!                                                                       
!-----------------------------------------------------------------------

      subroutine qqaxzz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = dconjg(a(1,j))*b(1,i)
            do k = 2,dim
               c(j,i) = c(j,i)+dconjg(a(k,j))*b(k,i)
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
!     Library subroutine qqxazz
!
! Multiplication of a complex quadratic matrix with the adjoint of
! a another quadratic complex matrix
!     a(j,k)*dconjg(b(i,k)) = c(j,i)
!
!-----------------------------------------------------------------------

      subroutine qqxazz (a,b,c,dim)

      implicit none

      integer(long)                                 :: i, j, k
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(in)  :: b
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,1)*dconjg(b(i,1))
            do k = 2,dim
               c(j,i) = c(j,i)+a(j,k)*dconjg(b(i,k))
            enddo
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine msqzz
! Local control: generating vibrational wavefuntion
!-----------------------------------------------------------------------

      subroutine msqzz (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2, &
                                                         dim3
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(out) :: b

      do i = 1,dim2
         do j = 1,dim1
            if (i .ne. dim3) then
               b(j,i) = dcmplx(0.0d0)
            else
               b(j,i) = a(j,i)
            endif
         enddo
      enddo
     
     
      return
      end subroutine

      end module mmlib
