! **********************************************************************
!                                                                     
!                              MTLIB                                  
!                                                                     
!  Library module containing linear algebra routines that involve the 
!  multiplication of a tensor of third order with a matrix.           
!                                                                     
!  Nomenclature:
!    Each name has 6 basic characters:
!    First 2 characters denote the objects being multiplied:
!       q: quadratic matrix
!       m: general (rectangular) matrix
!       h: hermitian matrix
!       p: positive definite matrix
!       s: symmetric matrix
!       u: unitary matrix
!       d: diagonal matrix (only diagonal elements are supplied as a
!          vector)
!       t: 2nd index of tensor of third order
!       t1: 1st index of tensor of third order
!       v: vector
!       x: scalar
!       e.g. 'qt' denotes the operation (quadratic matrix * tensor 2nd
!            index)
!       e.g. 'qt1' denotes the operation (quadratic matrix * tensor 1st
!            index)
!    Character 3 denotes how first object is used:
!       x: unchanged from input
!       t: transpose of input
!       a: adjoint of input
!       c: complex conjugate of input
!       v: as a vector
!    Character 4 denotes how second object is used:
!       see Character 3 above.
!    Character 5, 6 denote data types of first, second object
!       respectively:
!       z: complex double precision (complex*16)
!       c: complex single precision (complex*8)
!       d: real double precision (real*8)
!       r: real single precision (real*4)
!       y: complex matrix stored as two double precision (real*8)
!          matrices
!     Further characters, if present, give more informaion:
!       a: the result is added to a further object
!       r: the result is subtracted (removed) from a further object
!       c: the input matrices commute
!       h: the resulting matrix is hermitian
!       s: the resulting matrix is symmetric
!       1: the physical dimensions of the matrices differs from those
!          used.
!
!  The suffix _s after the name means that the routine works with a
!  "selected" vector", i.e. not all elements are present.
!   
!  Contents:                                                          
!  In the following list of available subroutines, matrices/tensors on
!  the LHS of the definition are input, that on the RHS output. The
!  usual summation convention is used i.e. a sum is made over repeated
!  indices on the LHS.
!
!   addtxxzz (a,b,dim1,dim2,dim3)
!       Definition: a(j) + b(i,j,k) = b(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3)
!
!   qtxxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxzd (a,b,c,dim1,dim2,dim3)
!       Definition: dble(a(l,j))*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxzda (a,b,c,dim1,dim2,dim3)
!       Definition: dble(a(l,j))*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,l)*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxzzs (a,b,c,dim1,dim2,dim3)
!       Definition: c(i,l,k) - a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxzzr (a,b,c,dim1,dim2,dim3)
!       Definition: c(i,l,k) - a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   tqxxzza (a,b,c,dim1,dim2,dim3)
!       Definition: a(i,j,k)*b(j,l) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim1,dim2,dim3),b(dim2,dim2),c(dim1,dim2,dim3)
!
!   tqxazza (a,b,c,dim1,dim2,dim3)
!       Definition: a(i,j,k)*dconjg(b(l,j)) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim1,dim2,dim3),b(dim2,dim2),c(dim1,dim2,dim3)
!
!   tqxazz (a,b,c,dim1,dim2,dim3)
!       Definition: a(i,j,k)*dconjg(b(l,j)) = c(i,l,k) .
!       Dimensions: a(dim1,dim2,dim3),b(dim2,dim2),c(dim1,dim2,dim3)
!
!   mtxxzz (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtxxzza (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtxxzd (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtvxzz (a,b,c,dim1,dim2,dim3,dim4,p)
!       Definition: a(p,j)*b(i,j,k) = c(i,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim3)
!
!   mtcxzz (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: dconjg(a(l,j))*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtxxdz (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mttxzz (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim4),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mttxdd (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim4),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mttxdr (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim4),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtxxdd (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!   mtaxzz (a,b,c,dim1,dim2,dim3,dim4)
!       Definition: dconjg(a(j,l))*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim4),b(dim1,dim2,dim3),c(dim1,dim4,dim3)
!
!C  qtxxdd (a,b,c,dim1,dim2,dim3)
!C      Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!C      Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtcxzz (a,b,c,dim1,dim2,dim3)
!       Definition: dconjg(a(l,j))*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxdd (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxzzo (a,b,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = b(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3)
!
!   dtxxzzr (a,b,c,dim1,dim2,dim3)
!       Definition: c(i,j,k) - a(j)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxdd (a,b,c,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxddo (a,b,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = b(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3)
!
!   dtxxdz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxdzo (a,b,dim1,dim2,dim3)
!       Definition: a(j)*b(i,j,k) = b(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtcxzzo (a,b,dim1,dim2,dim3)
!       Definition: dconjg(a(j))*b(i,j,k) = b(i,j,k) .
!       Dimensions: a(dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtaxzz (a,b,c,dim1,dim2,dim3)
!       Definition: dconjg(a(j,l))*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxdz (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxdza (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qt1xxdz (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,i)*b(i,j,k) = c(l,j,k) .
!       Dimensions: a(dim1,dim1),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dt1xxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(i)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim1),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qt1txdz (a,b,c,dim1,dim2,dim3)
!       Definition: a(i,l)*b(i,j,k) = c(l,j,k) .
!       Dimensions: a(dim1,dim1),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxdz (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxzz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxyz (a,b,c,dim1,dim2,dim3)
!       Definition: a(l,j)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qttxyz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,l)*b(i,j,k) = c(i,l,k) .
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   dtxxyz (a,b,c,dim1,dim2,dim3)
!       Definition: a(j,1)*b(i,j,k) + (0,1)*a(j,2)*b(i,j,k) = c(i,j,k) .
!       Dimensions: a(dim2,2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!
!   qtxxzz_s (a,b,c,dim1,dim2,dim3,index,index1)
!       Definition: index(i,j,k)=x
!                   index1(i,l,k)=x1
!                   a(l,j)*b(x) = c(x1) .
!       Dimensions: a(dim2,dim2),index(dim1,dim2,dim3),
!                   index1(dim1,dim2,dim3),b(max(x)),c(max(x1))
!
!   mtxxzz_s (a,b,c,dim1,dim2,dim3,dim4,dim5,index,index1)
!       Definition: index(i,j,k)=x
!                   index1(i,l,k)=x1
!                   a(l,j)*b(x) = c(x1) .
!       Dimensions: a(dim4,dim5),index(dim1,dim2,dim3),
!                   index1(dim1,dim2,dim3),b(max(x)),c(max(x1))
!
!   qttxzz_s (a,b,c,dim1,dim2,dim3,index,index1)
!       Definition: index(i,j,k)=x
!                   index1(i,l,k)=x1
!                   a(j,l)*b(x) = c(x1) .
!       Dimensions: a(dim2,dim2),index(dim1,dim2,dim3),
!                   index1(dim1,dim2,dim3),b(max(x)),c(max(x1))
!
!   qtxxzza_s (a,b,c,dim1,dim2,dim3,index,index1)
!       Definition: index(i,j,k)=x
!                   index1(i,l,k)=x1
!                   a(l,j)*b(x) + c(x1) = c(x1) .
!       Dimensions: a(dim2,dim2),index(dim1,dim2,dim3),
!                   index1(dim1,dim2,dim3),b(max(x)),c(max(x1))
!
!   qtxxzzr_s (a,b,c,dim1,dim2,dim3,index,index1)
!       Definition: index(i,j,k)=x
!                   index1(i,l,k)=x1
!                   c(x1) - a(l,j)*b(x) = c(x1) .
!       Dimensions: a(dim2,dim2),index(dim1,dim2,dim3),
!                   index1(dim1,dim2,dim3),b(max(x)),c(max(x1))
!
! **********************************************************************
      module mtlib

use decimal, only: dop, long, sip
use openmpmod
use omp_lib

      implicit none
      private
      public :: addtxxzz, qtxxzz, qtxxzd, qtxxzza, qtxxzda, qttxzza, &
                qttxzzs, qtxxzzr, tqxxzza, tqxazza, tqxazz, mtxxzz, &
                mtvxzz, mtxxzd, mtcxzz, mtxxdz, mttxzz, mttxdd, mttxdr, &
                mtxxdd, mtaxzz, qtxxdd, qtcxzz, qttxdd, dtxxzz, &
                dtxxzzo, dtxxzzr, dtxxdd, dtxxddo, dtxxdz, dtxxdzo, dtcxzzo, &
                qtaxzz, qtxxdz, qtxxdza, qt1xxdz, dt1xxzz, qt1txdz, &
                qttxdz, qttxzz, qtxxyz, qttxyz, dtxxyz, qtxxzz_s, &
                mtxxzz_s, qttxzz_s, qtxxzza_s, qtxxzzr_s, mtxxzza

      contains

!-----------------------------------------------------------------------
! Library subroutine addtxxzz
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order:
!           a(j) + b(i,j,k) =  b(i,j,k).
!
! Input-variables:  a      - complex vector 
!                   b      - complex tensor of third order
! Output-variables: b      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine addtxxzz (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                          :: i, j, k
      integer(long), intent(in)                              :: dim1, dim2, &
                                                                dim3
      complex(dop), dimension(dim2), intent(in)              :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(inout) :: b

      do k=1,dim3
         do j=1,dim2
            do i=1,dim1
               b(i,j,k)=b(i,j,k)+a(j)
            enddo
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine qtxxzz
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

#ifdef OMP

!$omp parallel num_threads(ompthread) private(i,j,k,l) if (lompthread)
!$omp do 
      do k=1,dim3
         do i=1,dim1
            do l=1,dim2
               c(i,l,k) = 0.0d0
               do j=1,dim2
                  c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
               enddo
            enddo
         enddo
      enddo
!$omp end do
!$omp end parallel 

#else

      c(:,:,:) = 0.0d0

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

#endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxzd
!
! Multiplication of a complex quadratic matrix with a real tensor
! of third order:
!           dble(a(l,j))*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix (only real part used)
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine qtxxzd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim2,dim3), intent(out) :: c
      complex(dop), dimension(dim2,dim2), intent(in)    :: a

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=dble(a(l,1))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dble(a(l,j))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=dble(a(l,1))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+dble(a(l,j))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=dble(a(l,1))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dble(a(l,j))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=dble(a(l,1))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+dble(a(l,j))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine qtxxzza
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and added to a different tensor:
!           a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input c matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine qtxxzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif
      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxzda
!
! Multiplication of a complex quadratic matrix with a complex tensor
! of third order and added to a different tensor:
!           dble(a(l,j))*b(i,j,k) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix
!                   b      - real tensor of third order
!                   c      - real tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine qtxxzda (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, k, l
      integer(long), intent(in)                           :: dim1, dim2, &
                                                             dim3
      real(dop), dimension(dim1,dim2,dim3), intent(in)    :: b
      real(dop), dimension(dim1,dim2,dim3), intent(inout) :: c
      complex(dop), dimension(dim2,dim2), intent(in)      :: a

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dble(a(l,j))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+dble(a(l,j))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dble(a(l,j))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+dble(a(l,j))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qttxzza
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and added to a different tensor:
!           a(j,l)*b(i,j,k) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine qttxzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxzzs
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and subtracted from a different tensor:
!           c(i,l,k) - a(l,j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

!     subroutine qtxxzzs (a,b,c,dim1,dim2,dim3)

!     implicit none

!     integer(long)                                        :: i, j, k, l
!     integer(long), intent(in)                            :: dim1, dim2, &
!                                                             dim3
!     complex(dop), dimension(dim2,dim2), intent(in)       :: a
!     complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
!     complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

!     if (dim1.ne.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do j=1,dim2
!              do l=1,dim2
!                 do i=1,dim1
!                    c(i,l,k)=c(i,l,k)-a(l,j)*b(i,j,k)
!                 enddo
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do j=1,dim2
!              do l=1,dim2
!                 c(1,l,k)=c(1,l,k)-a(l,j)*b(1,j,k)
!              enddo
!           enddo
!        enddo
!     else if (dim1.ne.1 .and. dim3.eq.1) then
!        do j=1,dim2
!           do l=1,dim2
!              do i=1,dim1
!                 c(i,l,1)=c(i,l,1)-a(l,j)*b(i,j,1)
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.eq.1) then
!        do j=1,dim2
!           do l=1,dim2
!              c(1,l,1)=c(1,l,1)-a(l,j)*b(1,j,1)
!           enddo
!        enddo
!     endif

!     return
!     end subroutine 


!-----------------------------------------------------------------------
! Library subroutine qttxzzs
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and substracted from a different tensor:
!            c(i,l,k) - a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine qttxzzs (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)-a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)-a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)-a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)-a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 


!-----------------------------------------------------------------------
! Library subroutine qtxxzzr
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and subtracted from a different tensor:
!           c(i,l,k) - a(l,j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxzzr (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)-a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)-a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)-a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)-a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine tqxxzza
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and added to a different tensor:
!           a(i,j,k)*b(j,l) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - complex tensor of third order 
!                   b      - complex matrix
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine tqxxzza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)       :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(i,j,k)*b(j,l)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  c(1,l,k)=c(1,l,k)+a(1,j,k)*b(j,l)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(i,j,1)*b(j,l)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               c(1,l,1)=c(1,l,1)+a(1,j,1)*b(j,l)
            enddo
         enddo
      endif

      return
      end subroutine



!-----------------------------------------------------------------------
! Library subroutine tqxazza
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and added to a different tensor:
!           a(i,j,k)*dconjg(b(l,j)) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - complex tensor of third order 
!                   b      - complex matrix
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB input ! matrix is overwritten on output.
!-----------------------------------------------------------------------

      subroutine tqxazza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)       :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(i,j,k)*dconjg(b(l,j))
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  c(1,l,k)=c(1,l,k)+a(1,j,k)*dconjg(b(l,j))
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(i,j,1)*dconjg(b(l,j))
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               c(1,l,1)=c(1,l,1)+a(1,j,1)*dconjg(b(l,j))
            enddo
         enddo
      endif

      return
      end subroutine



!-----------------------------------------------------------------------
! Library subroutine tqxazz
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order:
!           a(i,j,k)*dconjg(b(l,j)) = c(i,l,k).
!
! Input-variables:  a      - complex tensor of third order 
!                   b      - complex matrix
!
! Output-variables: c      - resulting complex tensor
!
!-----------------------------------------------------------------------

      subroutine tqxazz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(in)       :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

       If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(i,1,k)*dconjg(b(l,1))
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(i,j,k)*dconjg(b(l,j))
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(1,1,k)*dconjg(b(l,1))
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(1,j,k)*dconjg(b(l,j))
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(i,1,1)*dconjg(b(l,1))
            enddo
         enddo
         do l=1,dim2
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(i,j,1)*dconjg(b(l,j))
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(1,1,1)*dconjg(b(l,1))
         enddo
         do l=1,dim2
           do j=2,dim2
              c(1,l,1)=c(1,l,1)+a(1,j,1)*dconjg(b(l,j))
            enddo
         enddo
      endif

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mtxxzz
!
! Multiplication of a complex rectangular matrix with a complex tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtxxzz (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4
      complex(dop), dimension(dim4,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim4,dim3), intent(out) :: c

#ifdef OMP
!$omp parallel num_threads(ompthread) private(i,j,k,l) if (lompthread)
!$omp do 
      do k=1,dim3
         do i=1,dim1
            do l=1,dim4
               c(i,l,k) = 0.0d0
               do j=1,dim2
                  c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
               enddo
            enddo
         enddo
      enddo
!$omp end do
!$omp end parallel

#else
      c(:,:,:) = 0.0d0
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

#endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mtxxzza
!
! Multiplication of a complex rectangular matrix with a complex tensor
! of third order and added to a different tensor:
!           a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k) .
!
! Input-variables:  a      - complex matrix
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB tensor c is overwritten on output
!-----------------------------------------------------------------------

      subroutine mtxxzza (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer       dim1,dim2,dim3,dim4, i, j, k, l
      complex*16    a(dim4,dim2),b(dim1,dim2,dim3),c(dim1,dim4,dim3)

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end

!-----------------------------------------------------------------------
! Library subroutine mtvxzz
!
! Multiplication of a complex rectangular matrix with a complex tensor 
! of third order witth p fixed (matrix used as a vector):
!           a(p,j)*b(i,j,k) = c(i,k) .
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtvxzz (a,b,c,dim1,dim2,dim3,dim4,p)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4, p
      complex(dop), dimension(dim4,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim3), intent(out)      :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do i=1,dim1
               c(i,k)=a(p,1)*b(i,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do i=1,dim1
                  c(i,k)=c(i,k)+a(p,j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            c(1,k)=a(p,1)*b(1,1,k)
         enddo
         do k=1,dim3
            do j=2,dim2
               c(1,k)=c(1,k)+a(p,j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do i=1,dim1
            c(i,1)=a(p,1)*b(i,1,1)
         enddo
         do j=2,dim2
            do i=1,dim1
               c(i,1)=c(i,1)+a(p,j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         c(1,1)=a(p,1)*b(1,1,1)
         do j=2,dim2
            c(1,1)=c(1,1)+a(p,j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mtxxzd
!
! Multiplication of a complex rectangular matrix with a real tensor 
! of third order:
!           dble(a(l,j))*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - complex matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine mtxxzd (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, dim4
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim4,dim3), intent(out) :: c
      complex(dop), dimension(dim4,dim2), intent(in)    :: a

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=dble(a(l,1))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dble(a(l,j))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=dble(a(l,1))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+dble(a(l,j))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=dble(a(l,1))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dble(a(l,j))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=dble(a(l,1))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+dble(a(l,j))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mtcxzz
!
! Multiplication of a complex rectangular matrix with a complex tensor 
! of third order:
!           dconjg(a(l,j))*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtcxzz (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4
      complex(dop), dimension(dim4,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=dconjg(a(l,1))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dconjg(a(l,j))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=dconjg(a(l,1))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+dconjg(a(l,j))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=dconjg(a(l,1))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dconjg(a(l,j))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=dconjg(a(l,1))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+dconjg(a(l,j))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 


!-----------------------------------------------------------------------
! Library subroutine mtxxdz
!
! Multiplication of a real rectangular matrix with a complex tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - real rectangular matrix
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtxxdz (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4
      real(dop), dimension(dim4,dim2), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=a(l,1)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=a(l,1)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=a(l,1)*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=a(l,1)*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine mttxzz
!
! Multiplication of a transposed complex rectangular matrix with a 
! complex tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mttxzz (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4
      complex(dop), dimension(dim2,dim4), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim4
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim4
            do j=2,dim2
               c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mttxdd
!
! Multiplication of a transposed real rectangular matrix with a 
! real tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine mttxdd (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, dim4
      real(dop), dimension(dim2,dim4), intent(in)       :: a
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim4
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim4
            do j=2,dim2
               c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mttxdr
!
! Multiplication of a transposed real rectangular matrix with a 
! real tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine mttxdr (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, dim4
      real(sip), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim2,dim4), intent(in)       :: a
      real(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim4
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim4
            do j=2,dim2
               c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mtxxdd
!
! Multiplication of a real rectangular matrix with a real tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine mtxxdd (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, dim4
      real(dop), dimension(dim4,dim2), intent(in)       :: a
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=a(l,1)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=a(l,1)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim4
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
                c(i,l,1)=a(l,1)*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim4
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
        do l=1,dim4
            c(1,l,1)=a(l,1)*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim4
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mtaxzz
!
! Multiplication of the adjoint of a complex rectangular matrix with a 
! complex tensor of third order:
!           dconj(a(j,l))*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtaxzz (a,b,c,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4
      complex(dop), dimension(dim2,dim4), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim4,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do i=1,dim1
                  c(i,l,k)=dconjg(a(1,l))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dconjg(a(j,l))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               c(1,l,k)=dconjg(a(1,l))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim4
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+dconjg(a(j,l))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do i=1,dim1
               c(i,l,1)=dconjg(a(1,l))*b(i,1,1)
            enddo
         enddo
         do l=1,dim4
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dconjg(a(j,l))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            c(1,l,1)=dconjg(a(1,l))*b(1,1,1)
         enddo
         do l=1,dim4
            do j=2,dim2
               c(1,l,1)=c(1,l,1)+dconjg(a(j,l))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qtxxdd
!
! Multiplication of a real quadratic matrix with a real tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k) .
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine qtxxdd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      real(dop), dimension(dim2,dim2), intent(in)       :: a
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(l,1)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(l,1)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(l,1)*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(l,1)*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtcxzz
 
! Multiplication of the complex conjugate of a complex quadratic matrix 
! with a complex tensor of third order:
!           dconjg(a(l,j))*b(i,j,k) = c(i,l,k) .
 
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtcxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=dconjg(a(l,1))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dconjg(a(l,j))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=dconjg(a(l,1))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+dconjg(a(l,j))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=dconjg(a(l,1))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dconjg(a(l,j))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=dconjg(a(l,1))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+dconjg(a(l,j))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qttxdd
!
! Multiplication of a transposed real quadratic matrix with a 
! real tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine qttxdd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k, l
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      real(dop), dimension(dim2,dim2), intent(in)       :: a
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim2
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim2
           do j=2,dim2
              c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dtxxzz
!
! Multiplication of a complex diagonal matrix with a complex tensor 
! of third order. Only diagonal matrix elements are given as a vector:
!           a(j)*b(i,j,k) = c(i,j,k) .
!
! Input-variables:  a      - complex vector with diagonal matrix elements 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2), intent(in)            :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               c(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            c(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine dtxxzzo
!
! Multiplication of a complex diagonal matrix with a complex tensor
! of third order. Only diagonal matrix elements are given as a vector,
! and the result is stored in the input tensor:
!           a(j)*b(i,j,k) = b(i,j,k) .
!
! Input-variables:  a      - complex vector with diagonal matrix elements
!                   b      - complex tensor of third order
! Output-variables: b      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxzzo (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, k
      integer(long), intent(in)                           :: dim1, dim2, &
                                                             dim3
      complex(dop), dimension(dim2), intent(in)           :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(inout) :: b

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  b(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               b(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               b(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            b(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dtxxzzr
!
! Multiplication of a complex diagonal matrix with a complex tensor 
! of third order and subtracted from a different tensor:
!           c(i,l,k) - a(j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex vector with diagonal matrix elements
!                   b      - complex tensor of third order
!                   c      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxzzr (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2), intent(in)            :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=c(i,j,k)-a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               c(1,j,k)=c(1,j,k)-a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=c(i,j,1)-a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            c(1,j,1)=c(1,j,1)-a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine dtxxdd
!
! Multiplication of a real diagonal matrix with a real tensor 
! of third order. Only diagonal matrix elements are given as a vector:
!           a(j)*b(i,j,k) = c(i,j,k) .
!
! Input-variables:  a      - real vector with diagonal matrix elements 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine dtxxdd (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      real(dop), dimension(dim2), intent(in)            :: a
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      real(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               c(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            c(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine dtxxddo
!
! Multiplication of a real diagonal matrix with a real tensor
! of third order. Only diagonal matrix elements are given as a vector,
! and the result is stored in the input tensor:
!           a(j)*b(i,j,k) = b(i,j,k) .
!
! Input-variables:  a      - real vector with diagonal matrix elements
!                   b      - real tensor of third order
! Output-variables: b      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine dtxxddo (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                     :: i, j, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3
      real(dop), dimension(dim2), intent(in)            :: a
      real(dop), dimension(dim1,dim2,dim3), intent(out) :: b

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  b(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               b(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               b(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            b(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dtxxdz
!
! Multiplication of a real diagonal matrix with a complex tensor 
! of third order. Only diagonal matrix elements are given as a vector:
!           a(j)*b(i,j,k) = c(i,j,k) .
!
! Input-variables:  a      - real vector with diagonal matrix elements 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxdz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2), intent(in)               :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               c(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            c(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine dtxxdzo
!
! Multiplication of a real diagonal matrix with a complex tensor
! of third order. Only diagonal matrix elements are given as a vector,
! and the result is stored in the input tensor:
!           a(j)*b(i,j,k) = b(i,j,k) .
!
! Input-variables:  a      - complex vector with diagonal matrix elements
!                   b      - complex tensor of third order
! Output-variables: b      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxdzo (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2), intent(in)               :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: b

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  b(i,j,k)=a(j)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               b(1,j,k)=a(j)*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               b(i,j,1)=a(j)*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            b(1,j,1)=a(j)*b(1,j,1)
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine dtcxzzo
!
! Multiplication of the complex conjugate of a diagonal matrix with a 
! complex tensor of third order. Only diagonal matrix elements are given 
! as a vector, and the result is stored in the input tensor:
!           a(j)*b(i,j,k) = b(i,j,k) .
!
! Input-variables:  a      - complex vector with diagonal matrix elements
!                   b      - complex tensor of third order
! Output-variables: b      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtcxzzo (a,b,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2), intent(in)            :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: b

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  b(i,j,k)=dconjg(a(j))*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               b(1,j,k)=dconjg(a(j))*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               b(i,j,1)=dconjg(a(j))*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            b(1,j,1)=dconjg(a(j))*b(1,j,1)
         enddo
      endif

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine qtaxzz
!
! Multiplication of the adjoint of a complex quadratic matrix with a 
! complex tensor of third order:
!           dconj(a(j,l))*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtaxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=dconjg(a(1,l))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+dconjg(a(j,l))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=dconjg(a(1,l))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+dconjg(a(j,l))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=dconjg(a(1,l))*b(i,1,1)
            enddo
         enddo
         do l=1,dim2
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dconjg(a(j,l))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=dconjg(a(1,l))*b(1,1,1)
         enddo
         do l=1,dim2
            do j=2,dim2
               c(1,l,1)=c(1,l,1)+dconjg(a(j,l))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qtxxdz
!
! Multiplication of a real quadratic matrix with a complex tensor 
! of third order:
!           a(l,j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxdz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,dim2), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(l,1)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(l,1)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(l,1)*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(l,1)*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxdza
!
! Multiplication of a real quadratic matrix with a complex tensor 
! of third order:
!           a(l,j)*b(i,j,k) + c(i,l,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxdza (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,dim2), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(l,j)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+a(l,j)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(l,j)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+a(l,j)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qt1xxdz
!
! Multiplication of a real quadratic matrix with the first index of a 
!  complex tensor of third order:
!           a(l,i)*b(i,j,k) = c(l,j,k) .
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qt1xxdz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim1,dim1), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim2.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim1
                  c(l,j,k)=a(l,1)*b(1,j,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=1,dim2
               do i=2,dim1
                  do l=1,dim1
                     c(l,j,k)=c(l,j,k)+a(l,i)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim2.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim1
               c(l,1,k)=a(l,1)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do i=2,dim1
               do l=1,dim1
                  c(l,1,k)=c(l,1,k)+a(l,i)*b(i,1,k)
               enddo
            enddo
         enddo
      else if (dim2.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim1
               c(l,j,1)=a(l,1)*b(1,j,1)
            enddo
         enddo
         do j=1,dim2
            do i=2,dim1
               do l=1,dim1
                  c(l,j,1)=c(l,j,1)+a(l,i)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim1
            c(l,1,1)=a(l,1)*b(1,1,1)
         enddo
         do i=2,dim1
            do l=1,dim1
               c(l,1,1)=c(l,1,1)+a(l,i)*b(i,1,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine dt1xxzz
!
! Multiplication of a complex diagonal matrix with the first index of a 
!  complex tensor of third order. Only diagonal matrix elements are
!  supplied as  vector:
!           a(i)*b(i,j,k) = c(i,j,k) .
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dt1xxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim1), intent(in)            :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c


      if (dim2.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=a(i)*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim2.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do i=1,dim1
               c(i,1,k)=a(i)*b(i,1,k)
            enddo
         enddo
      else if (dim2.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=a(i)*b(i,j,1)
            enddo
         enddo
      else if (dim2.eq.1 .and. dim3.eq.1) then
         do i=1,dim1
            c(i,1,1)=a(i)*b(i,1,1)
         enddo
      endif

      return
      end subroutine  

!-----------------------------------------------------------------------
! Library subroutine qt1txdz
!
! Multiplication of a transposed real quadratic matrix with the 
!  first index of a complex tensor of third order:
!           a(i,l)*b(i,j,k) = c(l,j,k) .
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qt1txdz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim1,dim1), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim2.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim1
                  c(l,j,k)=a(1,l)*b(1,j,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=1,dim2
               do l=1,dim1
                  do i=2,dim1
                     c(l,j,k)=c(l,j,k)+a(i,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim2.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim1
               c(l,1,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim1
               do i=2,dim1
                  c(l,1,k)=c(l,1,k)+a(i,l)*b(i,1,k)
               enddo
            enddo
         enddo
      else if (dim2.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim1
               c(l,j,1)=a(1,l)*b(1,j,1)
            enddo
         enddo
         do j=1,dim2
            do l=1,dim1
               do i=2,dim1
                  c(l,j,1)=c(l,j,1)+a(i,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim1
            c(l,1,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim1
            do i=2,dim1
               c(l,1,1)=c(l,1,1)+a(i,l)*b(i,1,1)
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qttxdz
!
! Multiplication of a transposed real quadratic matrix with a 
! complex tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qttxdz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,dim2), intent(in)          :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim2
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim2
           do j=2,dim2
              c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qt1xxzz
!
! Multiplication of a real quadratic matrix with the first index of a 
!  complex tensor of third order:
!           a(l,i)*b(i,j,k) = c(l,j,k) .
!
! Input-variables:  a      - real matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

!     subroutine qt1xxzz (a,b,c,dim1,dim2,dim3)

!     implicit none

!     integer(long)                                        :: i, j, k, l
!     integer(long), intent(in)                            :: dim1, dim2, &
!                                                             dim3
!     complex(dop), dimension(dim1,dim1), intent(in)       :: a
!     complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
!     complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

!     if (dim2.ne.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do j=1,dim2
!              do l=1,dim1
!                 c(l,j,k)=a(l,1)*b(1,j,k)
!              enddo
!           enddo
!        enddo
!        do k=1,dim3
!           do j=1,dim2
!              do i=2,dim1
!                 do l=1,dim1
!                    c(l,j,k)=c(l,j,k)+a(l,i)*b(i,j,k)
!                 enddo
!              enddo
!           enddo
!        enddo
!     else if (dim2.eq.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do l=1,dim1
!              c(l,1,k)=a(l,1)*b(1,1,k)
!           enddo
!        enddo
!        do k=1,dim3
!           do i=2,dim1
!              do l=1,dim1
!                 c(l,1,k)=c(l,1,k)+a(l,i)*b(i,1,k)
!              enddo
!           enddo
!        enddo
!     else if (dim2.ne.1 .and. dim3.eq.1) then
!        do j=1,dim2
!           do l=1,dim1
!              c(l,j,1)=a(l,1)*b(1,j,1)
!           enddo
!        enddo
!        do j=1,dim2
!           do i=2,dim1
!              do l=1,dim1
!                 c(l,j,1)=c(l,j,1)+a(l,i)*b(i,j,1)
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.eq.1) then
!        do l=1,dim1
!           c(l,1,1)=a(l,1)*b(1,1,1)
!        enddo
!        do i=2,dim1
!           do l=1,dim1
!              c(l,1,1)=c(l,1,1)+a(l,i)*b(i,1,1)
!           enddo
!        enddo
!     endif

!     return
!     end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qttxzz
!
! Multiplication of a transposed complex quadratic matrix with a 
! complex tensor of third order:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - real matrix 
!                   b      - real tensor of third order
! Output-variables: c      - resulting real tensor
!-----------------------------------------------------------------------

      subroutine qttxzz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      If (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=a(1,l)*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  do i=1,dim1
                     c(i,l,k)=c(i,l,k)+a(j,l)*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=a(1,l)*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do l=1,dim2
               do j=2,dim2
                  c(1,l,k)=c(1,l,k)+a(j,l)*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=a(1,l)*b(i,1,1)
            enddo
         enddo
         do l=1,dim2
            do j=2,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+a(j,l)*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=a(1,l)*b(1,1,1)
         enddo
         do l=1,dim2
           do j=2,dim2
              c(1,l,1)=c(1,l,1)+a(j,l)*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qtxxyz
!
! Multiplication of a complex quadratic matrix with a complex tensor
! of third order, where the complex matrix is stored as two real
! matrices:
!           a(l,j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxyz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,dim2,2), intent(in)        :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=dcmplx(a(l,1,1),a(l,1,2))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                    c(i,l,k)=c(i,l,k)+dcmplx(a(l,j,1),a(l,j,2))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=dcmplx(a(l,1,1),a(l,1,2))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+dcmplx(a(l,j,1),a(l,j,2))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=dcmplx(a(l,1,1),a(l,1,2))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dcmplx(a(l,j,1),a(l,j,2))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=dcmplx(a(l,1,1),a(l,1,2))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+dcmplx(a(l,j,1),a(l,j,2))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qttxyz
!
! Multiplication of the trabspose of a complex quadratic matrix with 
! a complex tensor of third order, where the complex matrix is stored 
! as two real matrices:
!           a(j,l)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qttxyz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k, l
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,dim2,2), intent(in)        :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do i=1,dim1
                  c(i,l,k)=dcmplx(a(1,l,1),a(1,l,2))*b(i,1,k)
               enddo
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  do i=1,dim1
                    c(i,l,k)=c(i,l,k)+dcmplx(a(j,l,1),a(j,l,2))*b(i,j,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               c(1,l,k)=dcmplx(a(1,l,1),a(1,l,2))*b(1,1,k)
            enddo
         enddo
         do k=1,dim3
            do j=2,dim2
               do l=1,dim2
                  c(1,l,k)=c(1,l,k)+dcmplx(a(j,l,1),a(j,l,2))*b(1,j,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do i=1,dim1
               c(i,l,1)=dcmplx(a(1,l,1),a(1,l,2))*b(i,1,1)
            enddo
         enddo
         do j=2,dim2
            do l=1,dim2
               do i=1,dim1
                  c(i,l,1)=c(i,l,1)+dcmplx(a(j,l,1),a(j,l,2))*b(i,j,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            c(1,l,1)=dcmplx(a(1,l,1),a(1,l,2))*b(1,1,1)
         enddo
         do j=2,dim2
            do l=1,dim2
               c(1,l,1)=c(1,l,1)+dcmplx(a(j,l,1),a(j,l,2))*b(1,j,1)
            enddo
         enddo
      endif

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine dtxxyz
!
! Multiplication of a complex quadratic diagonal matrix with a complex
! tensor of third order, where the complex matrix is stored as two real
! matrices:
!           a(j)*b(i,j,k) = c(i,l,k).
!
! Input-variables:  a      - complex matrix
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine dtxxyz (a,b,c,dim1,dim2,dim3)

      implicit none

      integer(long)                                        :: i, j, k
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      real(dop), dimension(dim2,2), intent(in)             :: a
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: b
      complex(dop), dimension(dim1,dim2,dim3), intent(out) :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  c(i,j,k)=dcmplx(a(j,1),a(j,2))*b(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               c(1,j,k)=dcmplx(a(j,1),a(j,2))*b(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               c(i,j,1)=dcmplx(a(j,1),a(j,2))*b(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            c(1,j,1)=dcmplx(a(j,1),a(j,2))*b(1,j,1)
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qtxxzz_s
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order. The tensor is not completely stored, and the indices
! are managed by the index tensors:
!           index(i,j,k)=x
!           index1(i,l,k)=x1
!           a(l,j)*b(x) = c(x1).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qtxxzz_s (a,b,c,dim1,dim2,dim3,index,index1)

      implicit none

      integer(long)                                        :: i, j, k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(*), intent(in)               :: b
      complex(dop), dimension(*), intent(out)              :: c

      do k=1,dim3
         do l=1,dim2
            do i=1,dim1
               x1=index1(i,l,k)
               if (x1 .ne. 0) c(x1)=0.0d0
            enddo
         enddo
      enddo

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        c(x1)=c(x1)+a(l,j)*b(x)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  c(x1)=c(x1)+a(l,j)*b(x)
               endif
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine mtxxzz_s
!
! Multiplication of a complex matrix with a complex tensor 
! of third order. The tensor is not completely stored, and the indices
! are managed by the index tensors:
!           index(i,j,k)=x
!           index1(i,l,k)=x1
!           a(l,j)*b(x) = c(x1).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine mtxxzz_s (a,b,c,dim1,dim2,dim3,dim4,dim5,index,index1)

      implicit none

      integer(long)                                        :: i, j, k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3, dim4, &
                                                              dim5
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(dim5,dim4), intent(in)       :: a
      complex(dop), dimension(*), intent(in)               :: b
      complex(dop), dimension(*), intent(out)              :: c

      do k=1,dim3
         do l=1,dim2
            do i=1,dim1
               x1=index1(i,l,k)
               if (x1 .ne. 0) c(x1)=0.0d0
            enddo
         enddo
      enddo
 
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim4
               do l=1,dim5
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        c(x1)=c(x1)+a(l,j)*b(x)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim4
               do l=1,dim5
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim4
            do l=1,dim5
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim4
            do l=1,dim5
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  c(x1)=c(x1)+a(l,j)*b(x)
               endif
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qttxzz_s
!
! Multiplication of a transposed complex quadratic matrix with a 
! complex tensor c of third order. The tensor is not completely stored, 
! and the indices are managed by the index tensors:
!           index(i,j,k)=x
!           index1(i,l,k)=x1
!           a(j,l)*b(x) = c(x1).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!-----------------------------------------------------------------------

      subroutine qttxzz_s (a,b,c,dim1,dim2,dim3,index,index1)

      implicit none

      integer(long)                                        :: i, j, k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(*), intent(in)               :: b
      complex(dop), dimension(*), intent(out)              :: c

      do k=1,dim3
         do l=1,dim2
            do i=1,dim1
               x1=index1(i,l,k)
               if (x1 .ne. 0) c(x1)=0.0d0
            enddo
         enddo
      enddo

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        c(x1)=c(x1)+a(j,l)*b(x)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(j,l)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(j,l)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  c(x1)=c(x1)+a(j,l)*b(x)
               endif
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxzza_s
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and added to a different tensor. The tensors are not 
! completely stored, and the indices are managed by the index tensors:
!           index(i,j,k)=x
!           index1(i,l,k)=x1
!           a(l,j)*b(x) + c(x1) = c(x1).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB tensor c is overwritten on output
!-----------------------------------------------------------------------

      subroutine qtxxzza_s (a,b,c,dim1,dim2,dim3,index,index1)

      implicit none

      integer(long)                                        :: i, j, k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(*), intent(in)               :: b
      complex(dop), dimension(*), intent(out)              :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        c(x1)=c(x1)+a(l,j)*b(x)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)+a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  c(x1)=c(x1)+a(l,j)*b(x)
               endif
            enddo
         enddo
      endif

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine qtxxzzr_s
!
! Multiplication of a complex quadratic matrix with a complex tensor 
! of third order and subtracted from a different tensor. The tensors are 
! not completely stored, and the indices are managed by the index tensors:
!           index(i,j,k)=x
!           index1(i,l,k)=x1
!           c(x1) - a(l,j)*b(x) = c(x1).
!
! Input-variables:  a      - complex matrix 
!                   b      - complex tensor of third order
! Output-variables: c      - resulting complex tensor
!
! NB tensor c is overwritten on output
!-----------------------------------------------------------------------

      subroutine qtxxzzr_s (a,b,c,dim1,dim2,dim3,index,index1)

      implicit none

      integer(long)                                        :: i, j, k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(dim2,dim2), intent(in)       :: a
      complex(dop), dimension(*), intent(in)               :: b
      complex(dop), dimension(*), intent(out)              :: c

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        c(x1)=c(x1)-a(l,j)*b(x)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim2
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)-a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     c(x1)=c(x1)-a(l,j)*b(x)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim2
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  c(x1)=c(x1)-a(l,j)*b(x)
               endif
            enddo
         enddo
      endif

      return
      end subroutine

      end module mtlib 


      

