!-----------------------------------------------------------------------
!                    OP1LIB
!       SUBROUTINES ACTING ON A SINGLE OBJECT (LEVEL 1 ROUTINES)
!
! NOMENCLATURE:
!    Each of the following routines starts with:
!       sum2: the squares of elements are summed
!       sum:  the elements are summed
!       norm: the norm (sqrt of sum of squares) is calculated
!       tr:   the trace is calculated
!       zero: all elements are set to zero
!       unit: sets up a unit matrix
!       cp:   input matrix/vector is copied to output matrix/vector.
!       init: initialises an object i.e. sets all elements to 1.
!       over: calculates an overlap matrix between two sets of vectors
!       tran: transposes input object
!       cut:  calculate a cut through the source object (the target has
!             one dimension less than the source object)
!    This is followed by four chracters:
!    Character 1 denotes the object on which the action is performed:
!       v: vector
!       q: quadratic matrix
!       m: general (rectangular) matrix
!       h: hermitian matrix
!       t: tensor of third order
!    Character 2 defines how the object is used:
!       t: in sum2, the product of the element with its transpose is
!          summed
!       a: in cp, the adjoint of the object is copied
!       x: this is a blank
!    Character 3 defines the data type of the object:
!       s: real single precision (real*4)
!       d: real double precision (real*8)
!       c: complex single precision (complex*8)
!       z: complex double precision (complex*16)
!       i: integer
!       l: logical
!     Further characters, if present, give more informaion:
!       1: the physical dimensions of the matrices differs from those
!          used.
!       n: (= 1 or 2 or ...) in a cut the dimension to be held constant
!     
! Contents:
! In the following list of available subroutines, objects on the LHS
! of the definition are input, that on the RHS output. The usual
! summation convention is used i.e. a sum is made over repeated indices
! on the LHS
!
!    sum2vxd (vec,sum2,dim)
!        Definition: vec(i)*vec(i) = sum2
!        Dimensions: vec(dim),sum2
!
!    sum2vxz (vec,sum2,dim)
!        Definition: dconjg(vec(i))*vec(i) = sum2
!        Dimensions: vec(dim),sum2
!
!    sum2qxd (mat,sum2,dim)
!        Definition: mat(i,j)*mat(i,j) = sum2
!        Dimensions: mat(dim,dim),sum2
!
!    sum2qxz (mat,sum2,dim)
!        Definition: mat(i,j)*mat(i,j) = sum2
!        Dimensions: mat(dim,dim),sum2
!
!    sum2mxd (mat,sum2,dim1,dim2)
!        Definition: mat(i,j)*mat(i,j) = sum2
!        Dimensions: mat(dim1,dim2),sum2
!
!    sum2qtz (mat,sum2,dim)
!        Definition: mat(i,j)*mat(j,i) = sum2
!        Dimensions: mat(dim,dim),sum2
!
!    sumqxz (mat,sum,dim)
!        Definition: Sum(mat(i,j)) = sum
!        Dimensions: mat(dim,dim),sum
!
!    sumvxz (vec,sum,dim)
!        Definition: Sum(vec(i)) = sum
!        Dimensions: vec(dim),sum
!
!    normvxd (vec,norm,dim)
!        Definition: vec(i)*vec(i) = norm**2
!        Dimensions: vec(dim),norm
!
!    normvxz (vec,norm,dim)
!        Definition: dconjg(vec(i))*vec(i) = norm**2
!        Dimensions: vec(dim),norm
!
!    trhxz (mat,trace,dim)
!        Definition: dble(mat(i,i)) = trace
!        Dimensions: mat(dim,dim),trace
!
!    trqxd (mat,trace,dim)
!        Definition: mat(i,i) = trace
!        Dimensions: mat(dim,dim),trace
!
!    trqxz (mat,trace,dim)
!        Definition: mat(i,i) = trace
!        Dimensions: mat(dim,dim),trace
!
!    trmmcxzz(a,b,trace,dim1,dim2)
!        Definiton: dconjg(a(i,j))*b(j,i) = trace
!        Dimension: a(dim1,dim2), b(dim2,dim1)
!
!    trmmaxzz(a,b,trace,dim1,dim2)
!      Definiton: dconjg(a(j,i))*b(j,i) = trace
!      Dimension: a(dim1,dim2), b(dim1,dim2)
!
!    trmmaxzza(a,b,scal,dim1,dim2)
!        Definiton: dconjg(a(i,j))*b(j,i)) = scal
!        Dimension: a(dim1,dim2), b(dim1,dim2)
!
!    trtxz(ten,vec,dim1,dim2)
!        Definiton: ten(i,j,j)) = vec(i)
!        Dimension: ten(dim1,dim2,dim2), vec(dim1)
!
!    zeromxz (mat,dim1,dim2)
!        Definition: mat(i,j) = (0.0d0,0.0d0)
!        Dimensions: mat(dim1,dim2)
!
!    zeromxd (mat,dim1,dim2)
!        Definition: mat(i,j) = 0.0d0
!        Dimensions: mat(dim1,dim2)
!
!    zeromxs (mat,dim1,dim2)
!        Definition: mat(i,j) = 0.0
!        Dimensions: mat(dim1,dim2)
!
!    zeromxi (mat,dim1,dim2)
!        Definition: mat(i,j) = 0
!        Dimensions: mat(dim1,dim2)
!
!    zeromxl (mat,dim1,dim2)
!        Definition: mat(i,j) = .false.
!        Dimensions: mat(dim1,dim2)
!
!    zerovxz (vec,dim)
!        Definition: vec(i) = (0.0d0,0.0d0)
!        Dimensions: vec(dim)
!
!    zerovxd (vec,dim)
!        Definition: vec(i) = 0.0d0
!        Dimensions: vec(dim)
!
!    zerovxs (vec,dim)
!        Definition: vec(i) = 0.0d0
!        Dimensions: vec(dim)
!
!    zerovxi (vec,dim)
!        Definition: vec(i) = 0
!        Dimensions: vec(dim)
!
!    zerovxl (vec,dim)
!        Definition: vec(i) = .false.
!        Dimensions: vec(dim)
!
!    unitqxz (mat,dim)
!        Definition: mat(i,j) = d(i,j)   (d(i,j) is the kronecker delta)
!        Dimensions: mat(dim,dim)
!
!    unitqxd (mat,dim)
!        Definition: mat(i,j) = d(i,j)   (d(i,j) is the kronecker delta)
!        Dimensions: mat(dim,dim)
!
!    unitqxs (mat,dim)
!        Definition: mat(i,j) = d(i,j)   (d(i,j) is the kronecker delta)
!        Dimensions: mat(dim,dim)
!
!    offdqxz (mat,dim)
!        Definition: mat(i,j) = 1.0 ; i .ne. j
!        Dimensions: mat(dim,dim)
!
!    offiqxz (mat,dim)
!        Definition: mat(i,j) = (0.0,1.0) ; i > j
!        Definition: mat(i,j) = (0.0,-1.0) ; i < j
!        Dimensions: mat(dim,dim)
!
!    cpqxd (a,c,dim)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim,dim),c(dim,dim)
!
!    cpqxd1 (a,c,phdim,dim)
!        Definition: c(j,i) = a(j,i) ; 1 <= i,j <= dim
!        Dimensions: a(phdim,dim),c(phdim,dim)
!
!    cpqxz (a,c,dim)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim,dim),c(dim,dim)
!
!    cpqxdz (a,c,dim)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim,dim),c(dim,dim)
!
!    cpqaz (a,c,dim)
!       Definition: c(i,j) = dconj(a(j,i))
!       Dimensions: a(dim,dim),c(dim,dim)
!
!    cpqtz (a,c,dim)
!        Definition: c(i,j) = a(j,i)
!        Dimensions: a(dim,dim),c(dim,dim)
!
!    cpmaz (a,c,dim1,dim2) 
!        Definition: c(i,j) = dconj(a(j,i))
!        Dimensions: a(dim1,dim2),c(dim2,dim1)

!    cpqvzd (a,c,dim)
!        Definition: c(j) = a(j,j)
!        Dimensions: a(dim,dim),c(dim)
!
!    cpmtz (a,c,dim1,dim2)
!        Definition: c(i,j) = a(j,i)
!        Dimensions: a(dim1,dim2),c(dim2,dim1)
!
!    cpmxz (a,c,dim1,dim2)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim1,dim2),c(dim1,dim2)
!
!    cpmxd (a,c,dim1,dim2)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim1,dim2),c(dim1,dim2)
!
!    cpmxi (a,c,dim1,dim2)
!        Definition: c(j,i) = a(j,i)
!        Dimensions: a(dim1,dim2),c(dim1,dim2)
!
!    cpvxd (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxs (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxz (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvcz (v,w,dim)
!        Definition: w(i) = dconjg(v(i))
!        Dimensions: v(dim), w(dim)
!
!    cpvxdz (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxzd (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxi (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxi2(v,w,dim)
!        Definition: w(i) = v(i) &&  v(i) = w(i)
!        Dimensions: v(dim),w(dim)
!
!    cpvxl (v,w,dim)
!        Definition: w(i) = v(i)
!        Dimensions: v(dim),w(dim)
!
!    initvxz (vec,dim)
!        Definition: vec(i) = (1.0d0,0.0d0)
!        Dimensions: vec(dim)
!
!    initvxd (vec,dim)
!        Definition: vec(i) = 1.0d0
!        Dimensions: vec(dim)
!
!    initvxi (vec,dim)
!        Definition: vec(i) = 1
!        Dimensions: vec(dim)
!
!    initvxl (vec,dim)
!        Definition: vec(i) = .true.
!        Dimensions: vec(dim)
!
!    initmxz (a,dim1,dim2)
!        Definition: a(j,i) = (1.0d0,0.0d0)
!        Dimensions: a(dim1,dim2)
!
!    initmxd (a,dim1,dim2)
!        Definition: a(j,i) = (1.0d0,0.0d0)
!        Dimensions: a(dim1,dim2)
!
!    initmxl (a,dim1,dim2)
!        Definition: a(j,i) = .true.
!        Dimensions: a(dim1,dim2)
!
!    overmxz (a,c,dim1,dim2)
!        Definition: dconjg(a(k,j))*a(k,i) = c(j,i)
!        Dimensions: a(dim1,dim2),c(dim2,dim2)
!
!    overmcz (a,c,dim1,dim2)
!        Definition: a(k,j)*a(k,i) = c(j,i)
!        Dimensions: a(dim1,dim2),c(dim2,dim2)
!
!    tranmxz (a,c,dim1,dim2)
!        Definition: a(k,j) = c(j,k)
!        Dimensions: a(dim1,dim2),c(dim2,dim1)
!
!    tranqxz (a,dim)
!        Definition: a(k,j) = a(j,k)
!        Dimensions: a(dim,dim)
!
!    tranhqxz (a,dim)
!        Definition: a(k,j) = a(j,k)
!        Dimensions: a(dim,dim)
!
!    tranmxd (a,c,dim1,dim2)
!        Definition: a(k,j) = c(j,k)
!        Dimensions: a(dim1,dim2),c(dim2,dim1)
!
!    tranqxd (a,dim)
!        Definition: a(k,j) = a(j,k)
!        Dimentions: a(dim,dim)
!
!    cpvxz_s (v,w,dim1,dim2,index1,index2)
!        Definition: index1(j)=i
!                    index2(k)=i
!                    v(i) = w(i)
!        Dimensions: v(dim1),w(dim2),index1(dim1),index2(dim2)
!     
!    cptxd1 (v,t,dim1,dim2,dim3,jcut1,jcut2)
!        Definition: t(jcut1,i,jcut2)=v(i)
!        Dimensions: t(dim1,dim2,dim3),v(dim2)
!
!    cuttxd1 (t,v,dim1,dim2,dim3,jcut1,jcut2)
!        Definition: v(i)=t(jcut1,i,jcut2)
!        Dimensions: t(dim1,dim2,dim3),v(dim2)
!
!    cuttxd2 (t,m,dim1,dim2,dim3,jcut)
!        Definition: m(i,k)=t(i,jcut,k)
!        Dimensions: t(dim1,dim2,dim3),m(dim1,dim3)
!
!    cuttxz1 (t,v,dim1,dim2,dim3,jcut1,jcut2)
!        Definition: v(i)=t(jcut1,i,jcut2)
!        Dimensions: t(dim1,dim2,dim3),v(dim2)
!
!    cuttxz2 (t,m,dim1,dim2,dim3,jcut)
!        Definition: m(i,k)=t(i,jcut,k)
!        Dimensions: t(dim1,dim2,dim3),m(dim1,dim3)
!
!    cuttxz3 (m,t,dim1,dim2,dim3,jcut)
!        Definition: t(i,jcut,k)=m(i,k)
!        Dimensions: t(dim1,dim2,dim3),m(dim1,dim3)
!
!-----------------------------------------------------------------------

      module op1lib

use decimal, only: dop, long, sip, vlong
use openmpmod

      implicit none
      private
      public :: sum2vxd, sum2vxz, sum2qxd, sum2qxz, sum2mxd, sum2qtz, &
               sumqxz, sumvxz,normvxd, normvxz, trhxz, trqxd, trqxz, &
               trmmcxzz, trmmaxzz, trmmaxzza, trtxz, zeromxz, &
               zeromxd, zeromxs, zeromxi, zeromxl, zerovxz, zerovxz_vlong, &
               zerovxd, zerovxd_vlong, &
!               zeromxd, zeromxs, zeromxi, zeromxl, zerovxz, zerovxd, &
               zerovxs, zerovxi, zerovxl, unitqxz, unitqxd, unitqxs, &
               offdqxz,offiqxz, &
               cpqxd, cpqxd1, cpqxz, cpqxdz, cpqaz, cpqtz, cpqvzd, cpmaz, &
               cpmtz, cpmxz, cpmxd, cpmxi, cpvxd, cpvxs, cpvxz, &
               cpvcz, cpvxdz, cpvxzd, cpvxi, cpvxi2, cpvxl, &
               initvxz, initvxd, initvxi, initvxl, initmxz, initmxd, &
               initmxl, overmxz, overmcz, tranmxz, tranqxz, tranhqxz, &
               tranmxd, tranqxd, cpvxz_s, cptxd1, cuttxd1, cuttxd2, &
               cpvzdz, cpvidz, cuttxz1, cuttxz2, cuttxz3

       contains

!-----------------------------------------------------------------------
! Library subroutine sum2vxd
!
! sums the squares of the elements of a real vector
!        vec(i)*vec(i) = sum2
! i.e. the square of the norm of the vector
!-----------------------------------------------------------------------

      subroutine sum2vxd(vec,sum2,dim)

      implicit none

      integer(long)                         :: i
      integer(long), intent(in)             :: dim
      real(dop), intent(out)                :: sum2
      real(dop), dimension(dim), intent(in) :: vec
      
      sum2=0.
      do i=1,dim
         sum2=sum2+vec(i)*vec(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sum2vxz
!
! sums the squares of the elements of a complex vector
!        dconjg(vec(i))*vec(i) = sum2
! i.e. the square of the norm of the vector
!-----------------------------------------------------------------------

      subroutine sum2vxz(vec,sum2,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      complex(dop), intent(out)                :: sum2
      complex(dop), dimension(dim), intent(in) :: vec
      
      sum2=(0.0d0,0.0d0)
      do i=1,dim
         sum2=sum2+dconjg(vec(i))*vec(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sum2qxd
!
! sums the squares of the elements of a real quadratic matrix
!        mat(i,j)*mat(i,j) = sum2
!-----------------------------------------------------------------------
     
      subroutine sum2qxd(mat,sum2,dim)
      
      implicit none

      integer(long)                             :: i, j
      integer(long), intent(in)                 :: dim
      real(dop), intent(out)                    :: sum2
      real(dop), dimension(dim,dim), intent(in) :: mat

      sum2=0.
      do j=1,dim
         do i=1,dim
            sum2=sum2+mat(i,j)*mat(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sum2qxz
!
! sums the squares of the elements of a complex quadratic matrix
!        mat(i,j)*mat(i,j) = sum2
!-----------------------------------------------------------------------
     
      subroutine sum2qxz(mat,sum2,dim)
      
      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      complex(dop), intent(out)                    :: sum2
      complex(dop), dimension(dim,dim), intent(in) :: mat

      sum2=0.0d0
      do j=1,dim
         do i=1,dim
            sum2=sum2+mat(i,j)*mat(i,j)
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------- 
! Library subroutine sum2mxd
!
! sums the squares of the elements of a real rectangular matrix
!        mat(i,j)*mat(i,j) = sum2
!-----------------------------------------------------------------------
    
      subroutine sum2mxd(mat,sum2,dim1,dim2)
      
      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(dop), intent(out)                       :: sum2
      real(dop), dimension(dim1,dim2), intent(in)  :: mat

      sum2=0.d0
      do j=1,dim2
         do i=1,dim1
            sum2=sum2+mat(i,j)*mat(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sum2qtz
!
! sums the products of the elements of a real quadratic matrix
! with its transpose
!        mat(i,j)*mat(j,i) = sum2
!-----------------------------------------------------------------------
     
      subroutine sum2qtz(mat,sum2,dim)
      
      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      complex(dop), intent(out)                    :: sum2
      complex(dop), dimension(dim,dim), intent(in) :: mat

      sum2=0.d0
      do j=1,dim
         do i=1,dim
            sum2=sum2+mat(i,j)*mat(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sumqxz
!!
! sums the elements of a complex quadratic matrix
!        Sum(mat(i,j)) = sum
!-----------------------------------------------------------------------

      subroutine sumqxz(mat,sum,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      complex(dop), intent(out)                    :: sum 
      complex(dop), dimension(dim,dim), intent(in) :: mat

      sum=0.0d0
      do j=1,dim
         do i=1,dim
            sum=sum+mat(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine sumvxz
!!
! sums the elements of a complex quadratic matrix
!        Sum(vec(i)) = sum
!-----------------------------------------------------------------------

      subroutine sumvxz(vec,sum,dim)

      implicit none

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      complex(dop), intent(out)                    :: sum 
      complex(dop), dimension(dim), intent(in)     :: vec

      sum=0.0d0
      do i=1,dim
         sum=sum+vec(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine normvxd
!
! calculates the norm of a real vector
!        vec(i)*vec(i) = norm**2
!-----------------------------------------------------------------------

      subroutine normvxd(vec,norm,dim)
     
      implicit none

      integer(long)                         :: i
      integer(long), intent(in)             :: dim
      real(dop), intent(out)                :: norm
      real(dop), dimension(dim), intent(in) :: vec
     
      norm=vec(1)*vec(1)
      do i=2,dim
         norm=norm+vec(i)*vec(i)
      enddo
      norm=dsqrt(norm)

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine normvxz
!!
! calculates the norm of a complex vector
!        dconjg(vec(i))*vec(i) = norm**2
!-----------------------------------------------------------------------

      subroutine normvxz(vec,norm,dim)
      
      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop), intent(out)                   :: norm
      complex(dop), dimension(dim), intent(in) :: vec
      
      norm=0.d0
      do i=1,dim
         norm=norm+dble(dconjg(vec(i))*vec(i))
      enddo
      norm=dsqrt(norm)

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine trhxz
!
! calculates the trace of a hermitian matrix
!      dble(mat(i,i)) = trace
!
! NB diagonal elements of complex matrix are real due to hermiticity
!-----------------------------------------------------------------------

      subroutine trhxz (mat,trace,dim)
      
      implicit none

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      real(dop), intent(out)                       :: trace
      complex(dop), dimension(dim,dim), intent(in) :: mat
      
      trace=dble(mat(1,1))
      do i=2,dim
         trace=trace+dble(mat(i,i))
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine trqxd
!
! calculates the trace of a quadratic matrix
!      mat(i,i) = trace
!-----------------------------------------------------------------------

      subroutine trqxd (mat,trace,dim)
      
      implicit none

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      real(dop), intent(out)                    :: trace
      real(dop), dimension(dim,dim), intent(in) :: mat
      
      trace=mat(1,1)
      do i=2,dim
         trace=trace+mat(i,i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine trqxz
!
! calculates the trace of a quadratic matrix
!      mat(i,i) = trace
!-----------------------------------------------------------------------

      subroutine trqxz (mat,trace,dim)
      
      implicit none

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      complex(dop), intent(out)                    :: trace
      complex(dop), dimension(dim,dim), intent(in) :: mat
      
      trace=mat(1,1)
      do i=2,dim
         trace=trace+mat(i,i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine trmmcxzz
!
! calculates the trace of the hermitian product of two complex matrices
!
!      Definiton: conjg(a(i,j))*b(j,i) = trace
!      Dimension: a(dim1,dim2), b(dim2,dim1)
!
! NB diagonal elements of complex matrix are real due to hermiticity
!-----------------------------------------------------------------------

      subroutine trmmcxzz(a,b,trace,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), intent(out)                      :: trace
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim2,dim1), intent(in) :: b

      trace = dconjg(a(1,1))*b(1,1)
      do j=2,dim2
         trace = trace + dconjg(a(1,j))*b(j,1)
      enddo
      do i=2,dim1
         do j=1,dim2
            trace = trace + dconjg(a(i,j))*b(j,i) 
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine trmmcxzz
!
! calculates the trace of the hermitian product of two complex matrices
!
!      Definiton: dconjg(a(j,i))*b(j,i) = trace
!      Dimension: a(dim1,dim2), b(dim1,dim2
!
!-----------------------------------------------------------------------
      subroutine trmmaxzz(a,b,trace,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), intent(out)                      :: trace
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim1,dim2), intent(in) :: b

      trace = dconjg(a(1,1))*b(1,1)
      do j=2,dim1
         trace = trace + dconjg(a(j,1))*b(j,1)
      enddo
      do i=2,dim2
         do j=1,dim1
            trace = trace + dconjg(a(j,i))*b(j,i) 
         enddo
      enddo

      return
      end subroutine



!-----------------------------------------------------------------------
! Library subroutine trmmaxzza
!
! calculates the trace of the product of two complex matrices and adds 
! it to scal
!
!         Definiton: dconjg(a(j,i))*b(j,i)) + scal = scal
!         Dimension: a(dim1,dim2), b(dim1,dim2)
!
!-----------------------------------------------------------------------

      subroutine trmmaxzza(a,b,scal,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), intent(inout)                    :: scal
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim1,dim2), intent(in) :: b

      do j=1,dim2
         do i=1,dim1
            scal = scal + dconjg(a(i,j)) * b(i,j) 
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine trtxz
!
! calculates the trace of tensor, reducing it to a vector.
!
!         Definiton: ten(i,j,j) = vec(i)
!         Dimension: ten(dim1,dim2,dim2), vec(dim1)
!
!-----------------------------------------------------------------------

      subroutine trtxz(ten,vec,dim1,dim2)

      implicit none

      integer(long)                                       :: i, j
      integer(long), intent(in)                           :: dim1, dim2
      complex(dop), dimension(dim1,dim2,dim2), intent(in) :: ten
      complex(dop), dimension(dim1), intent(inout)        :: vec

      call zerovxz(vec,dim1)

      do j=1,dim2
         do i=1,dim1
            vec(i) = vec(i) + ten(i,j,j) 
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zeromxz
!
! makes all the elements of a complex rectangular matrix zero
!      mat(i,j) = (0.0d0,0.0d0)
!-----------------------------------------------------------------------

      subroutine zeromxz(mat,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(out) :: mat

      do j=1,dim2
         do i=1,dim1
            mat(i,j)=(0.0d0,0.0d0)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zeromxd
!!
! makes all the elements of a real*8 rectangular matrix zero
!      mat(i,j) = 0.0d0
!-----------------------------------------------------------------------

      subroutine zeromxd(mat,dim1,dim2)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(dop), dimension(dim1,dim2), intent(out) :: mat

      do j=1,dim2
         do i=1,dim1
            mat(i,j)=0.0d0
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zeromxs
!
! makes all the elements of a real*4 rectangular matrix zero
!      mat(i,j) = 0.0
!-----------------------------------------------------------------------

      subroutine zeromxs(mat,dim1,dim2)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(sip), dimension(dim1,dim2), intent(out) :: mat

      do j=1,dim2
         do i=1,dim1
            mat(i,j)=0.0
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zeromxi
!
! makes all the elements of a rectangular integer matrix zero
!      mat(i,j) = 0
!-----------------------------------------------------------------------

      subroutine zeromxi(mat,dim1,dim2)

      implicit none

      integer(long)                                    :: i, j
      integer(long), intent(in)                        :: dim1, dim2
      integer(long), dimension(dim1,dim2), intent(out) :: mat

      do j=1,dim2
         do i=1,dim1
            mat(i,j)=0
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zeromxl
!
! makes all the elements of a rectangular matrix of logicals .false.
!     mat(i,j) = .false.
!-----------------------------------------------------------------------

      subroutine zeromxl(mat,dim1,dim2)

      implicit none

      integer(long)                                      :: i, j
      integer(long), intent(in)                          :: dim1, dim2
      logical(kind=4), dimension(dim1,dim2), intent(out) :: mat

      do j=1,dim2
         do i=1,dim1
            mat(i,j)=.false.
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zerovxz
!
! makes all the elements of a complex vector zero
!     vec(i)=(0,0d0,0.0d0)
!-----------------------------------------------------------------------

      subroutine zerovxz(vec,dim)

      implicit none

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      complex(dop), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=(0.0d0,0.0d0)
      enddo

      return
      end subroutine

      subroutine zerovxz_vlong(vec,dim)

      implicit none
     
      integer(vlong)                             :: i
      integer(vlong), intent(in)                :: dim
      complex(dop), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=(0.0d0,0.0d0)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zerovxd
!
! makes all the elements of a real*8 vector zero
!     vec(i)=0.0d0
!-----------------------------------------------------------------------
      subroutine zerovxd(vec,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=0.0_dop
      enddo

      return
      end subroutine

      subroutine zerovxd_vlong(vec,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=0.0_dop
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zerovxs
!
! makes all the elements of a real*4 vector zero
!     vec(i)=0.0d0
!-----------------------------------------------------------------------
      subroutine zerovxs(vec,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(sip), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=0.0_sip
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zerovxi
!
! makes all the elements of an integer vector zero
!     vec(i)=0
!-----------------------------------------------------------------------
      subroutine zerovxi(vec,dim)

      implicit none

      integer(long)                              :: i
      integer(long), intent(in)                  :: dim
      integer(long), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=0
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine zerovxl
!
! makes all the elements of a logical vector .false.
!     vec(i)=.false.
!-----------------------------------------------------------------------
      subroutine zerovxl(vec,dim)

      implicit none

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      logical(kind=4), dimension(dim), intent(out) :: vec

      do i=1,dim
         vec(i)=.false.
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine unitqxz
!
! sets up a unit matrix
!     mat(i,j) = 0, if i.ne.j
!     mat(i,j) = 1, if i.eq.j
!-----------------------------------------------------------------------

      subroutine unitqxz(mat,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(out) :: mat

      do j=1,dim
         do i=1,dim
            mat(i,j)=(0.0_dop,0.0_dop)
         enddo
      enddo
      do i=1,dim
         mat(i,i)=(1.0_dop,0.0_dop)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine unitqxd
!
! sets up a unit matrix
!     mat(i,j) = 0, if i.ne.j
!     mat(i,j) = 1, if i.eq.j
!-----------------------------------------------------------------------

      subroutine unitqxd(mat,dim)

      implicit none

      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(out) :: mat

      do j=1,dim
         do i=1,dim
            mat(i,j)=0.0_dop
         enddo
      enddo
      do i=1,dim
         mat(i,i)=1.0_dop
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine unitqxs
!
! sets up a unit matrix
!     mat(i,j) = 0, if i.ne.j
!     mat(i,j) = 1, if i.eq.j
!-----------------------------------------------------------------------

      subroutine unitqxs(mat,dim)

      implicit none

      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(sip), dimension(dim,dim), intent(out) :: mat

      do j=1,dim
         do i=1,dim
            mat(i,j)=0.0_sip
         enddo
      enddo
      do i=1,dim
         mat(i,i)=1.0_sip
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine offdqxz
!
! sets up a unit off-diagonal matrix
!     mat(i,j) = 0, if i.eq.j
!     mat(i,j) = 1, if i.ne.j
!-----------------------------------------------------------------------

      subroutine offdqxz(mat,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(out) :: mat

      do j=1,dim
         mat(j,j)=(0.0_dop,0.0_dop)
         do i=j+1,dim
            mat(i,j)=(1.0_dop,0.0_dop)
            mat(j,i)=(1.0_dop,0.0_dop)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine offiqxz
!
! sets up a hermitian unit imaginary off-diagonal matrix
!     mat(i,j) = 0, if i.eq.j
!     mat(i,j) = (0,1), if i.ne.j
!-----------------------------------------------------------------------

      subroutine offiqxz(mat,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(out) :: mat

      do j=1,dim
         mat(j,j)=(0.0_dop,0.0_dop)
         do i=j+1,dim
            mat(i,j)=(0.0_dop,1.0_dop)
            mat(j,i)=(0.0_dop,-1.0_dop)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cpqxd
!
! copies a real quadratic matrix to another real quadratic matrix
!     c(i,j) = a(i,j)
!-----------------------------------------------------------------------

      subroutine cpqxd (a,c,dim)

      implicit none

      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cpqxd1
!
! copies a real quadratic matrix to another real quadratic matrix
!     c(i,j) = a(i,j)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine cpqxd1 (a,c,phdim,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim, phdim
      real(dop), dimension(phdim,dim), intent(in)  :: a
      real(dop), dimension(phdim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cpqxz
!
! copies a complex quadratic matrix to another complex quadratic matrix
!     c(i,j) = a(i,j)
!-----------------------------------------------------------------------

      subroutine cpqxz (a,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cpqxdz
!
! copies a real quadratic matrix to a complex quadratic matrix
!     c(i,j) = a(i,j)
!-----------------------------------------------------------------------

      subroutine cpqxdz (a,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim,dim), intent(in)     :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpqaz
!
! copies the adjoint of a complex quadratic matrix to a complex 
! quadratic matrix
!     c(j,i) = dconjg(a(i,j)) 
!-----------------------------------------------------------------------

      subroutine cpqaz (a,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = dconjg(a(i,j))
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpqtz
!
! copies the transpose of a complex quadratic matrix to a complex 
! quadratic matrix
!     c(j,i) = a(i,j) 
!-----------------------------------------------------------------------

      subroutine cpqtz (a,c,dim)

      implicit none
      
      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = a(i,j)
         enddo
      enddo
      
      return
      end subroutine


!----------------------------------------------------------------------
! Library subroutine cpmtz
!
! copies the transpose of a complex rectangular atrix to a complex
! rectangular matrix
!     c(j,i) = a(i,j)
!-----------------------------------------------------------------------

      subroutine cpmtz (a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim1), intent(out) :: c

      do i = 1,dim1
         do j = 1,dim2
            c(j,i) = a(i,j)
         enddo
      enddo

      return
      end subroutine


!----------------------------------------------------------------------
! Library subroutine cpmaz
!
! copies the adjoint of a rectangular quadratic matrix to a complex
! rectangular matrix
!     c(j,i) = dconjg(a(i,j))
!-----------------------------------------------------------------------

      subroutine cpmaz (a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim1), intent(out) :: c

      do i = 1,dim1
         do j = 1,dim2
            c(j,i) = dconjg(a(i,j))
         enddo
      enddo

      return
      end subroutine

!----------------------------------------------------------------------
! Library subroutine cpqvzd
!
! copies the diagonal of a quadratic complex matrix to a real vector
!     c(i) = a(i,i)
!-----------------------------------------------------------------------

      subroutine cpqvzd (a,c,dim1)

      implicit none

      integer(long)                                   :: i
      integer(long), intent(in)                       :: dim1
      complex(dop), dimension(dim1,dim1), intent(in)  :: a
      real(dop), dimension(dim1), intent(out)         :: c

!$omp parallel do num_threads(ompthread)
      do i = 1,dim1
         c(i) = dble(a(i,i))
      enddo
!$omp end parallel do

      return
      end subroutine



! ----------------------------------------------------------------------
! Library subroutine cpmxz
!
! copies a complex rectangular matrix to a different complex 
! rectangular matrix
!     c(j,i) = a(j,i)
!-----------------------------------------------------------------------

      subroutine cpmxz (a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(out) :: c

!$omp parallel do num_threads(ompthread)
      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,i)
         enddo
      enddo
!$omp end parallel do

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpmxd
!
! copies a real rectangular matrix to a different real rectangular matrix
!     c(j,i) = a(j,i)
!-----------------------------------------------------------------------

      subroutine cpmxd (a,c,dim1,dim2)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(dop), dimension(dim1,dim2), intent(in)  :: a
      real(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpmxi
!
! copies an integer rectangular matrix to a different integer
! rectangular matrix
!     c(j,i) = a(j,i)
!-----------------------------------------------------------------------

      subroutine cpmxi (a,c,dim1,dim2)

      implicit none

      integer(long)                                    :: i, j
      integer(long), intent(in)                        :: dim1, dim2
      integer(long), dimension(dim1,dim2), intent(in)  :: a
      integer(long), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxd
!
! copies a real vector to a different real vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxd (v,w,dim)

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop), dimension(dim), intent(in)  :: v
      real(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxs
!
! copies a real vector to a different real vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxs (v,w,dim)

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(sip), dimension(dim), intent(in)  :: v
      real(sip), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxz
!
! copies a complex vector to a different complex vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxz (v,w,dim)

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      complex(dop), dimension(dim), intent(in)  :: v
      complex(dop), dimension(dim), intent(out) :: w

!$omp parallel do num_threads(ompthread)
      do i = 1,dim
         w(i) = v(i)
      enddo
!$omp end parallel do 

      return
      end subroutine

! ---------------------------------------------------------------------
! Library subroutine cpvcz
!
! copies a conjugate of a complex vector to a different complex vector
!     w(i) = dconjg(vi))
! ----------------------------------------------------------------------

      subroutine cpvcz (v,w,dim)

      implicit none

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      complex(dop), dimension(dim), intent(in)  :: v
      complex(dop), dimension(dim), intent(out) :: w
      
      do i=1,dim
         w(i) = dconjg(v(i))
      enddo
      
      return
      end subroutine


! ----------------------------------------------------------------------
! Library subroutine cpvxdz
!
! copies a real vector to a  complex vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxdz (v,w,dim)

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: v
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvidz
!
! copies a real vector to the imaginary part of a  complex vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvidz (v,w,dim)

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: v
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = dcmplx(0.d0,v(i))
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvzdz
!
! copies two real vectors to the real and imaginary part of a complex vector
!     w(i) = dcmplx(vr(i),vi(i))
!-----------------------------------------------------------------------

      subroutine cpvzdz (vr,vi,w,dim)

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: vr
      real(dop), dimension(dim), intent(in)     :: vi
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = dcmplx(vr(i),vi(i))
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxzd
!
! copies the real part of a complex vector to a  real vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxzd (v,w,dim)

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      complex(dop), dimension(dim), intent(in)  :: v
      real(dop), dimension(dim), intent(out)    :: w

!$omp parallel do num_threads(ompthread)
      do i = 1,dim
         w(i) = dble(v(i))
      enddo
!$omp end parallel do 

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxi
!
! copies a integer vector to a different integer vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxi (v,w,dim)

      integer(long)                              :: i
      integer(long), intent(in)                  :: dim
      integer(long), dimension(dim), intent(in)  :: v
      integer(long), dimension(dim), intent(out) :: w

!$omp parallel do num_threads(ompthread)
      do i = 1,dim
         w(i) = v(i)
      enddo
!$omp end parallel do 

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxi2
!
! replaces 2 integer vectors
!     w(i) = v(i)  &&  v(i) = w(i)
!-----------------------------------------------------------------------
 
      subroutine cpvxi2 (v,w,dim)
 
      integer(long)                                :: i, a
      integer(long), intent(in)                    :: dim
      integer(long), dimension(dim), intent(inout) :: v
      integer(long), dimension(dim), intent(out)   :: w
 
      do i = 1,dim
         a = w(i)
         w(i) = v(i)
         v(i) = a
      enddo
 
      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine cpvxl
!
! copies a logical vector to a different logical vector
!     w(i) = v(i)
!-----------------------------------------------------------------------

      subroutine cpvxl (v,w,dim)

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      logical(kind=4), dimension(dim), intent(in)  :: v
      logical(kind=4), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initvxz
!
! initialises a complex vector
!    vec(i)=(1.0d0,0.0d0)
!-----------------------------------------------------------------------

      subroutine initvxz(vec,dim)

      implicit none

      integer(long)                              :: i
      integer(long), intent(in)                  :: dim
      complex(dop), dimension(dim), intent(out)  :: vec

      do i=1,dim
         vec(i)=(1.0d0,0.0d0)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initvxd
!
! initialises a real vector
!    vec(i)=1.0d0
!-----------------------------------------------------------------------

      subroutine initvxd(vec,dim)

      implicit none

      integer(long)                           :: i
      integer(long), intent(in)               :: dim
      real(dop), dimension(dim), intent(out)  :: vec

      do i=1,dim
         vec(i)=1.0d0
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initvxi
!
! initialises an integer vector
!    vec(i)=1
!-----------------------------------------------------------------------

      subroutine initvxi(vec,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      integer(long), dimension(dim), intent(out)  :: vec

      do i=1,dim
         vec(i)=1
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initvxl
!
! initialises a logical vector (with .true.)
!    vec(i)=.true.
!-----------------------------------------------------------------------

      subroutine initvxl(vec,dim)

      implicit none

      integer(long)                                 :: i
      integer(long), intent(in)                     :: dim
      logical(kind=4), dimension(dim), intent(out)  :: vec

      do i=1,dim
         vec(i)=.true.
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initmxz
!
! initialises a complex matrix
!    a(j,i)=(1.0d0,0.0d0)
!-----------------------------------------------------------------------

      subroutine initmxz(a,dim1,dim2)

      implicit none

      integer(long)                                    :: i, j
      integer(long), intent(in)                        :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(out)  :: a

      do i=1,dim2
         do j=1,dim1
            a(j,i)=(1.0d0,0.0d0)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initmxd
!
! initialises a real matrix
!    a(j,i)=(1.0d0,0.0d0)
!-----------------------------------------------------------------------

      subroutine initmxd(a,dim1,dim2)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim1, dim2
      real(dop), dimension(dim1,dim2), intent(out)  :: a

      do i=1,dim2
         do j=1,dim1
            a(j,i)=1.0d0
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine initmxl
!
! initialises a logical matrix
!    a(j,i)=.true.
!-----------------------------------------------------------------------

      subroutine initmxl(a,dim1,dim2)

      implicit none

      integer(long)                                       :: i, j
      integer(long), intent(in)                           :: dim1, dim2
      logical(kind=4), dimension(dim1,dim2), intent(out)  :: a

      do i=1,dim2
         do j=1,dim1
            a(j,i)=.true.
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine overmxz
!
! Overlap of a complex set of vectors with itself (i.e. multiplication
! of a matrix with its adjoint)
!     dconjg(a(k,j))*a(k,i) = c(j,i)
! NB resultant matrix is hermitian
!-----------------------------------------------------------------------
      subroutine overmxz (a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(out) :: c

      do i = 1,dim2
         do j = i,dim2
            c(j,i) = dconjg(a(1,j))*a(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+dconjg(a(k,j))*a(k,i)
            enddo
         enddo
      enddo
!
! now fill in other half of hermitian matrix
!
      do i=1,dim2
         c(i,i)=dble(c(i,i))
         do j=1,i-1
            c(j,i)=dconjg(c(i,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine overmcz
!
! Overlap of a complex set of vectors with itself (i.e. multiplication
! of a matrix with its transpose)
!     a(k,j)*a(k,i) = c(j,i)
! NB standard definition takes c.c. of first matrix elements.
! NB resultant matrix is symmetric
!-----------------------------------------------------------------------

      subroutine overmcz (a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j, k
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim2), intent(out) :: c

      do i = 1,dim2
         do j = i,dim2
            c(j,i) = a(1,j)*a(1,i)
            do k = 2,dim1
               c(j,i) = c(j,i)+a(k,j)*a(k,i)
            enddo
         enddo
      enddo
!
! now fill in other half of symmetric matrix
!
      do i=1,dim2
         do j=1,i-1
            c(j,i)=c(i,j)
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine tranmxz
!
! transposes a complex matrix:
!   a(i,j) = c(j,i)
!-----------------------------------------------------------------------

      subroutine tranmxz(a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim2,dim1), intent(out) :: c

      do i=1,dim1
         do j=1,dim2
            c(j,i)=a(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine tranmxd
!
! transposes a real matrix:
!   a(i,j) = c(j,i)
!-----------------------------------------------------------------------

      subroutine tranmxd(a,c,dim1,dim2)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim1, dim2
      real(dop), dimension(dim1,dim2), intent(in)  :: a
      real(dop), dimension(dim2,dim1), intent(out) :: c

      do i=1,dim1
         do j=1,dim2
            c(j,i)=a(i,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine tranqxz
!
! transposes a complex quadratic matrix to itself:
!   a(i,j) = a(j,i)
!-----------------------------------------------------------------------

      subroutine tranqxz(a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop)                                    :: b
      complex(dop), dimension(dim,dim), intent(inout) :: a

      do i=2,dim
         do j=1,i-1
            b=a(i,j)
            a(i,j)=a(j,i)
            a(j,i)=b
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine tranhqxz
!
! transposes a complex hermitian quadratic matrix to itself, enforcing
! hermiticity:
!   a(i,j) = a(j,i)
!-----------------------------------------------------------------------

      subroutine tranhqxz(a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(inout) :: a

      a(1,1)=dble(a(1,1))
      do i=2,dim
         a(i,i)=dble(a(i,i))
         do j=1,i-1
            a(i,j)=a(j,i)
            a(j,i)=dconjg(a(i,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine tranqxd
!
! transposes a real quadratic matrix to itself:
!   a(i,j) = a(j,i)
!-----------------------------------------------------------------------

      subroutine tranqxd(a,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      real(dop)                                    :: b
      real(dop), dimension(dim,dim), intent(inout) :: a

      do i=2,dim
         do j=1,i-1
            b=a(i,j)
            a(i,j)=a(j,i)
            a(j,i)=b
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cpvxz_s
!
! copies a complex vector to a different complex vector, where the 
! vectors are not completely stored, but managed by index arrays
!     w(i) = v(i)
!
!-----------------------------------------------------------------------


      subroutine cpvxz_s(v,w,index1,index2,dim1,dim2)

      implicit none

      integer(long)                              :: b1, b2
      integer(long), intent(in)                  :: dim1, dim2
      integer(long), dimension(dim1), intent(in) :: index1
      integer(long), dimension(dim2), intent(in) :: index2
      complex(dop), dimension(dim1), intent(in)  :: v
      complex(dop), dimension(dim2), intent(out) :: w

      b1=1
      do b2=1,dim2
  100    continue
         if (index2(b2) .eq. index1(b1)) then
            w(b2)=v(b1)
         else if (index2(b2) .gt. index1(b1) .and. b1 .lt. dim1) then
            b1=b1+1
            go to 100
         else
            w(b2)=0.0d0
         endif
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cptxd1
!
! Inserts a real*8 vector into a tensor of third order for fixed given
! values of the first and last index jcut1 and jcut2. 
!     t(jcut1,i,jcut2) = v(i)
!
!-----------------------------------------------------------------------


      subroutine cptxd1(v,t,dim1,dim2,dim3,jcut1,jcut2)

      implicit none

      integer(long)                                      :: i
      integer(long), intent(in)                          :: dim1, dim2, dim3, &
                                                            jcut1, jcut2
      real(dop), dimension(dim2), intent(in)             :: v
      real(dop), dimension(dim1,dim2,dim3), intent(out)  :: t

      do i=1,dim2
         t(jcut1,i,jcut2)=v(i)
      enddo

      return
      end subroutine 

!-----------------------------------------------------------------------
! Library subroutine cuttxd1
!
! calculates a cut through a real*8 tensor of third order for fixed given
! values of the first and last index jcut1 and jcut2. the result is 
! written to a real*8 vector v.
!     v(i) = t(jcut1,i,jcut2)
!
!-----------------------------------------------------------------------

      subroutine cuttxd1(t,v,dim1,dim2,dim3,jcut1,jcut2)

      implicit none

      integer(long)                                      :: i
      integer(long), intent(in)                          :: dim1, dim2, dim3, &
                                                            jcut1, jcut2
      real(dop), dimension(dim1,dim2,dim3), intent(in)   :: t
      real(dop), dimension(dim2), intent(out)            :: v

      do i=1,dim2
         v(i)=t(jcut1,i,jcut2)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cuttxd2
!
! calculates a cut through a real*8 tensor of third order for a fixed given
! value of the second index jcut. the result is written to a real*8 matrix m.
!     m(i,k) = t(i,jcut,k)
!
!-----------------------------------------------------------------------

      subroutine cuttxd2(t,m,dim1,dim2,dim3,jcut)

      implicit none

      integer(long)                                     :: i, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, jcut
      real(dop), dimension(dim1,dim2,dim3), intent(in)  :: t
      real(dop), dimension(dim1,dim3), intent(out)      :: m

      if (dim1.eq.1) then
         do k=1,dim3
            m(1,k)=t(1,jcut,k)
         enddo
      else if (dim3.eq.1) then
         do i=1,dim1
            m(i,1)=t(i,jcut,1)
         enddo
      else
         do k=1,dim3
            do i=1,dim1
               m(i,k)=t(i,jcut,k)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cuttxz1
!
! calculates a cut through a complex*16 tensor of third order for fixed given
! values of the first and last index jcut1 and jcut2. the result is 
! written to a complex*16 vector v.
!     v(i) = t(jcut1,i,jcut2)
!
!-----------------------------------------------------------------------

      subroutine cuttxz1(t,v,dim1,dim2,dim3,jcut1,jcut2)

      implicit none

      integer(long)                                      :: i
      integer(long), intent(in)                          :: dim1, dim2, dim3, &
                                                            jcut1, jcut2
      complex(dop), dimension(dim1,dim2,dim3), intent(in)   :: t
      complex(dop), dimension(dim2), intent(out)            :: v

      do i=1,dim2
         v(i)=t(jcut1,i,jcut2)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cuttxz2
!
! calculates a cut through a complex*16 tensor of third order for a fixed given
! value of the second index jcut. the result is written to a complex*16 matrix m.
!     m(i,k) = t(i,jcut,k)
!
!-----------------------------------------------------------------------

      subroutine cuttxz2(t,m,dim1,dim2,dim3,jcut)

      implicit none

      integer(long)                                     :: i, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, jcut
      complex(dop), dimension(dim1,dim2,dim3), intent(in)  :: t
      complex(dop), dimension(dim1,dim3), intent(out)      :: m

      if (dim1.eq.1) then
         do k=1,dim3
            m(1,k)=t(1,jcut,k)
         enddo
      else if (dim3.eq.1) then
         do i=1,dim1
            m(i,1)=t(i,jcut,1)
         enddo
      else
         do k=1,dim3
            do i=1,dim1
               m(i,k)=t(i,jcut,k)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine cuttxz3
!
! Stores a complex*16 matrix as a cut through a complex*16 tensor of 
! third order with a fixed given ! value of the second index jcut. 
!     t(i,jcut,k) = m(i,k) 
!
!-----------------------------------------------------------------------

      subroutine cuttxz3(m,t,dim1,dim2,dim3,jcut)

      implicit none

      integer(long)                                     :: i, k
      integer(long), intent(in)                         :: dim1, dim2, &
                                                           dim3, jcut
      complex(dop), dimension(dim1,dim2,dim3), intent(out)  :: t
      complex(dop), dimension(dim1,dim3), intent(in)        :: m

      if (dim1.eq.1) then
         do k=1,dim3
            t(1,jcut,k)=m(1,k)
         enddo
      else if (dim3.eq.1) then
         do i=1,dim1
            t(i,jcut,1)=m(i,1)
         enddo
      else
         do k=1,dim3
            do i=1,dim1
               t(i,jcut,k)=m(i,k)
            enddo
         enddo
      endif

      return
      end subroutine

      end module op1lib

