! **********************************************************************
!                                                                     
!                              RMLIB                                  
!                                                                     
!  Library module containing linear algebra routines that involve the 
!  formation of a "density" matrix
!              
!  Nomenclature:
!    Each name is 8 or 9 characters long.
!    First 2 Characters:
!       'rm' denote the operation 'formation of "density" matrix'
!       'rm1' denote the operation 'formation of "density" matrix',
!             where a tensor is multiplied with itself. 
!    Character 3 denotes type of matrix formed:
!       q: quadratic
!       m: general (rectangular) matrix
!       h: hermitian
!       a: anti-hermitian
!       s: symmetric
!       u: unitary
!       d: diagonal (only diagonal elements are saved as a vector)
!    Character 4 denotes how matrix is saved:
!       x: unchanged from calculated
!       t: transpose of calculated
!       a: adjoint of calculated
!       c: complex conjugate of calculated
!    Character 5 denotes how 'bra' tensor is used: 
!       x: complex conjugate of input (standard definition)
!       c: unchanged from input
!    Character 6 denotes how 'ket' tensor is used:
!       x: unchanged from input
!       c: complex conjugate of input
!    Character 7 and 8 denote data types of tensors:
!       z: complex double precision (complex*16)
!       c: complex single precision (complex*8)
!       d: real double precision (real*8)
!       r: real single precision (real*4)
!    Character 9 (optional) denotes data type of matrix:
!       z: complex double precision (complex*16)
!       c: complex single precision (complex*8)
!       d: real double precision (real*8)
!       r: real single precision (real*4)
!   
!  The suffix _s after the name means that the routine works with a
!  "selected" vector", i.e. not all elements are present.
!
!  Contents:                                                          
!  In the following list of available subroutines, matrices/tensors on the 
!  LHS of the definition are input, that on the RHS output. The usual summation
!  convention is used i.e. a sum is made over repeated indices on the LHS.
!
!    rmqxxxzz (bra,ket,mat,dim1,dim2,dim3)
!        Definition: dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmqxcxzz (bra,ket,mat,dim1,dim2,dim3)
!        Definition: bra(i,j,k)*ket(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmmxxxzz (bra,ket,mat,dim1,dim2,dim3,dim4)
!        Definition: dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim4),bra(dim1,dim2,dim3),ket(dim1,dim4,dim3)
!
!    rmmtxxzz (bra,ket,mat,dim1,dim2,dim3,dim4)
!        Definition: dconjg(bra(i,j,k))*ket(i,l,k) = mat(l,j) .
!        Dimensions: mat(dim4,dim2),bra(dim1,dim2,dim3),ket(dim1,dim4,dim3)
!
!    rmhxxxzz (bra,ket,mat,dim1,dim2,dim3)
!        Definition: dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmdxxxzz (bra,ket,mat,dim1,dim2,dim3)
!        Definition: dconjg(bra(i,j,k))*ket(i,j,k) = mat(j) 
!        Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmaxxxzz (bra,ket,mat,dim1,dim2,dim3)
!        Definition: dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmsxcxdd (bra,ket,mat,dim1,dim2,dim3)
!        Definition: bra(i,j,k)*ket(i,l,k) = mat(j,l) .
!     Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmsxcxdd2 (bra,ket,mat,dim1,dim2,dim3)
!        Definition: bra(i,j,k)*ket(i,l,k) + bra(i,l,k)*ket(i,j,k) = mat(j,l).
!     Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!    rmsxcxrrd (bra,ket,mat,dim1,dim2,dim3)
!        Definition: bra(i,j,k)*ket(i,l,k) = mat(j,l) .
!     Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!C   rmqxxxdd (bra,ket,mat,dim1,dim2,dim3)
!C       Definition: bra(i,j,k)*ket(i,l,k) = mat(j,l) .
!C       Dimensions: mat(dim2,dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!
!C   rmdxxxdd (bra,ket,mat,dim1,dim2,dim3)
!C       Definition: bra(i,j,k)*ket(i,j,k) = mat(j) .
!C       Dimensions: mat(dim2),bra(dim1,dim2,dim3),ket(dim1,dim2,dim3)
!                                                                    
!    rm1hxxxzz (ten,mat,dim1,dim2,dim3)
!        Definition: dconjg(ten(i,j,k))*ten(i,l,k) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),ten(dim1,dim2,dim3)
!
!    rm1hxxxzz_s (ten,mat,dim1,dim2,dim3,index)
!        Definition: index(i,j,k)=x
!                    index(i,l,k)=x1
!                    dconjg(ten(x))*ten(x1) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),index(dim1,dim2,dim3),ten(max(x))
!
!    rmhxxxzz_s (bra,ket,mat,dim1,dim2,dim3,index,index1)
!        Definition: index(i,j,k)=x
!                    index1(i,l,k)=x1
!                    dconjg(bra(x))*ket(x1) = mat(j,l) .
!        Dimensions: mat(dim2,dim2),index(dim1,dim2,dim3),
!                    index1(dim1,dim2,dim3),bra(max(x)),ket(max(x1))
!
!    rmmxxxzz_s (bra,ket,mat,dim1,dim2,dim3,dim4,dim5,index,index1)
!        Definition: index(i,j,k)=x
!                    index1(i,l,k)=x1
!                    dconjg(bra(x))*ket(x1) = mat(j,l) .
!        Dimensions: mat(dim4,dim5),index(dim1,dim2,dim3),
!                    index1(dim1,dim2,dim3),bra(max(x)),ket(max(x1))
!
! **********************************************************************

      module rmlib

use decimal, only: dop, sip, long

      implicit none
      private
      public :: rmqxxxzz, rmqxcxzz, rmmxxxzz, rmmtxxzz, rmhxxxzz, &
                rmdxxxzz, rmaxxxzz, rmsxcxdd, rmsxcxdd2, rmsxcxrrd, &
                rm1hxxxzz, rm1hxxxzz_s, rmhxxxzz_s, rmmxxxzz_s
!               rmqxxxdd, rmdxxxdd 

      contains

!-----------------------------------------------------------------------
! Library subroutine rmqxxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmqxxxzz (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      complex(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+dconjg(bra(i,j,k))*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  mat(j,l)=mat(j,l)+dconjg(bra(1,j,k))*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+dconjg(bra(i,j,1))*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               mat(j,l)=mat(j,l)+dconjg(bra(1,j,1))*ket(1,l,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmqxcxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order:
!
!            bra(i,j,k)*ket(i,l,k) = mat(j,l).
!
! NB bra is complex conjugate of standard definition i.e. the c.c. is
!    not taken
!
!-----------------------------------------------------------------------

      subroutine rmqxcxzz (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      complex(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+bra(i,j,k)*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  mat(j,l)=mat(j,l)+bra(1,j,k)*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+bra(i,j,1)*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=2,dim2
               mat(j,l)=mat(j,l)+bra(1,j,1)*ket(1,l,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmmxxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmmxxxzz (bra,ket,mat,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3, dim4
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim4,dim3), intent(in) :: ket
      complex(dop), dimension(dim2,dim4), intent(out)     :: mat

      do l=1,dim4
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do j=1,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+dconjg(bra(i,j,k))*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim4
               do j=1,dim2
                  mat(j,l)=mat(j,l)+dconjg(bra(1,j,k))*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim4
            do j=1,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+dconjg(bra(i,j,1))*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim4
            do j=1,dim2
               mat(j,l)=mat(j,l)+dconjg(bra(1,j,1))*ket(1,l,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmmtxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order, where the transpose of the matrix is stored:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(l,j).
!
!-----------------------------------------------------------------------

      subroutine rmmtxxzz (bra,ket,mat,dim1,dim2,dim3,dim4)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3, dim4
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim4,dim3), intent(in) :: ket
      complex(dop), dimension(dim4,dim2), intent(out)     :: mat

      do j=1,dim2
         do l=1,dim4
            mat(l,j)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  do i=1,dim1
                     mat(l,j)=mat(l,j)+dconjg(bra(i,j,k))*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do l=1,dim4
                  mat(l,j)=mat(l,j)+dconjg(bra(1,j,k))*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               do i=1,dim1
                  mat(l,j)=mat(l,j)+dconjg(bra(i,j,1))*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            do l=1,dim4
               mat(l,j)=mat(l,j)+dconjg(bra(1,j,1))*ket(1,l,1)
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmhxxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order, where the matrix formed is hermitian:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmhxxxzz (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      complex(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+dconjg(bra(i,j,k))*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  mat(j,l)=mat(j,l)+dconjg(bra(1,j,k))*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+dconjg(bra(i,j,1))*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               mat(j,l)=mat(j,l)+dconjg(bra(1,j,1))*ket(1,l,1)
            enddo
         enddo
      endif
!
! now form other half of matrix
!
      do l=1,dim2
         mat(l,l)=dble(mat(l,l))
      enddo
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=dconjg(mat(l,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmdxxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order, forming only the diagonal elements:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(j) 
!
!-----------------------------------------------------------------------

      subroutine rmdxxxzz (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      complex(dop), dimension(dim2), intent(out)          :: mat

      do j=1,dim2
         mat(j)=(0.0d0,0.0d0)
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               do i=1,dim1
                  mat(j)=mat(j)+dconjg(bra(i,j,k))*ket(i,j,k)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do j=1,dim2
               mat(j)=mat(j)+dconjg(bra(1,j,k))*ket(1,j,k)
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do j=1,dim2
            do i=1,dim1
               mat(j)=mat(j)+dconjg(bra(i,j,1))*ket(i,j,1)
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do j=1,dim2
            mat(j)=mat(j)+dconjg(bra(1,j,1))*ket(1,j,1)
         enddo
      endif

      do j=1,dim2
         mat(j)=dble(mat(j))
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmaxxxzz
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order, where the matrix formed is anti-hermitian:
!
!            dconjg(bra(i,j,k))*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmaxxxzz (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      real(dop)                                           :: x 
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      complex(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+dconjg(bra(i,j,k))*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  mat(j,l)=mat(j,l)+dconjg(bra(1,j,k))*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+dconjg(bra(i,j,1))*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               mat(j,l)=mat(j,l)+dconjg(bra(1,j,1))*ket(1,l,1)
            enddo
         enddo
      endif
!
! now form other half of matrix
!
      do l=1,dim2
         x=dimag(mat(l,l))
         mat(l,l)=dcmplx(0.0d0,x)
      enddo
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=-dconjg(mat(l,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmsxcxdd
!
! Formation of density type matrix by multiplication of two real
! tensors of third order, where the matrix formed is symmetric:
!
!            bra(i,j,k)*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmsxcxdd (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, k, l
      integer(long), intent(in)                        :: dim1, dim2, dim3
      real(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      real(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      real(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=0.0d0
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+bra(i,j,k)*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  mat(j,l)=mat(j,l)+bra(1,j,k)*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+bra(i,j,1)*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               mat(j,l)=mat(j,l)+bra(1,j,1)*ket(1,l,1)
            enddo
         enddo
      endif
!
! now form other half of matrix
!
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=mat(l,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmsxcxdd2
!
! Formation of density type matrix by multiplication of two real
! tensors of third order, where the matrix formed is symmetric:
!
!     bra(i,j,k)*ket(i,l,k) + bra(i,l,k)*ket(i,j,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmsxcxdd2 (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, k, l
      integer(long), intent(in)                        :: dim1, dim2, dim3
      real(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
      real(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
      real(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=0.0d0
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+bra(i,j,k)*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=1,dim2
                  mat(j,l)=mat(j,l)+bra(1,j,k)*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+bra(i,j,1)*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=1,dim2
               mat(j,l)=mat(j,l)+bra(1,j,1)*ket(1,l,1)
            enddo
         enddo
      endif
!
! now symmetrize matrix
!
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=mat(j,l) + mat(l,j)
            mat(l,j)=mat(l,j)
         enddo
         mat(l,l)=mat(l,l)*2.d0
      enddo

      return
      end subroutine rmsxcxdd2 

!-----------------------------------------------------------------------
! Library subroutine rmsxcxrrd
!
! Formation of density type matrix by multiplication of two real
! tensors of third order, where the matrix formed is symmetric:
!
!            bra(i,j,k)*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmsxcxrrd (bra,ket,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                    :: i, j, &
                                                          k, l
      integer(long), intent(in)                        :: dim1, &
                                                          dim2, &
                                                          dim3
      real(sip), dimension(dim1,dim2,dim3), intent(in) :: bra
      real(sip), dimension(dim1,dim2,dim3), intent(in) :: ket
      real(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=0.0d0
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+bra(i,j,k)*ket(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  mat(j,l)=mat(j,l)+bra(1,j,k)*ket(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+bra(i,j,1)*ket(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               mat(j,l)=mat(j,l)+bra(1,j,1)*ket(1,l,1)
            enddo
         enddo
      endif
!
! now form other half of matrix
! 
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=mat(l,j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmqxxxdd
!
! Formation of density type matrix by multiplication of two real
! tensors of third order:
!
!            bra(i,j,k)*ket(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

!     subroutine rmqxxxdd (bra,ket,mat,dim1,dim2,dim3)

!     implicit none

!     integer(long)                                    :: i, j, &
!                                                         k, l
!     integer(long), intent(in)                        :: dim1, &
!                                                         dim2, &
!                                                         dim3
!     real(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
!     real(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
!     real(dop), dimension(dim2,dim2), intent(out)     :: mat

!     do l=1,dim2
!        do j=1,dim2
!           mat(j,l)=0.0d0
!        enddo
!     enddo
!     if (dim1.ne.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do l=1,dim2
!              do j=1,dim2
!                 do i=1,dim1
!                    mat(j,l)=mat(j,l)+bra(i,j,k)*ket(i,l,k)
!                 enddo
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do l=1,dim2
!              do j=1,dim2
!                 mat(j,l)=mat(j,l)+bra(1,j,k)*ket(1,l,k)
!              enddo
!           enddo
!        enddo
!     else if (dim1.ne.1 .and. dim3.eq.1) then
!        do l=1,dim2
!           do j=1,dim2
!              do i=1,dim1
!                 mat(j,l)=mat(j,l)+bra(i,j,1)*ket(i,l,1)
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.eq.1) then
!        do l=1,dim2
!           do j=1,dim2
!              mat(j,l)=mat(j,l)+bra(1,j,1)*ket(1,l,1)
!           enddo
!        enddo
!     endif

!     return
!     end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmdxxxdd
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order:
!
!            bra(i,j,k)*ket(i,j,k) = mat(j).
!
!-----------------------------------------------------------------------

!     subroutine rmdxxxdd (bra,ket,mat,dim1,dim2,dim3)

!     implicit none

!     integer       dim1,dim2,dim3, i, j, k
!     real*8        bra(dim1,dim2,dim3), ket(dim1,dim2,dim3), 
!    +              mat(dim2)
!     integer(long)                                    :: i, j, &
!                                                         k
!     integer(long), intent(in)                        :: dim1, &
!                                                         dim2, &
!                                                         dim3
!     real(dop), dimension(dim1,dim2,dim3), intent(in) :: bra
!     real(dop), dimension(dim1,dim2,dim3), intent(in) :: ket
!     real(dop), dimension(dim2,dim2), intent(out)     :: mat

!     do j=1,dim2
!        mat(j)=0.0d0
!     enddo
!     if (dim1.ne.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do j=1,dim2
!              do i=1,dim1
!                 mat(j)=mat(j)+bra(i,j,k)*ket(i,j,k)
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do j=1,dim2
!              mat(j)=mat(j)+bra(1,j,k)*ket(1,j,k)
!           enddo
!        enddo
!     else if (dim1.ne.1 .and. dim3.eq.1) then
!        do j=1,dim2
!           do i=1,dim1
!              mat(j)=mat(j)+bra(i,j,1)*ket(i,j,1)
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.eq.1) then
!        do j=1,dim2
!           mat(j)=mat(j)+dconjg(bra(1,j,1))*ket(1,j,1)
!        enddo
!     endif

!     return
!     end subroutine




!-----------------------------------------------------------------------
! Library subroutine rm1hxxxzz
!
! Formation of density type matrix by multiplication of a complex 
! tensor of third order with itself. The matrix formed is hermitian:
!
!            dconjg(ten(i,j,k))*ten(i,l,k) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rm1hxxxzz (ten,mat,dim1,dim2,dim3)

      implicit none

      integer(long)                                       :: i, j, &
                                                             k, l
      integer(long), intent(in)                           :: dim1, &
                                                             dim2, &
                                                             dim3
      complex(dop), dimension(dim1,dim2,dim3), intent(in) :: ten
      complex(dop), dimension(dim2,dim2), intent(out)     :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo
      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     mat(j,l)=mat(j,l)+dconjg(ten(i,j,k))*ten(i,l,k)
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  mat(j,l)=mat(j,l)+dconjg(ten(1,j,k))*ten(1,l,k)
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  mat(j,l)=mat(j,l)+dconjg(ten(i,j,1))*ten(i,l,1)
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               mat(j,l)=mat(j,l)+dconjg(ten(1,j,1))*ten(1,l,1)
            enddo
         enddo
      endif
!
! now form other half of matrix
!
      do l=1,dim2
         mat(l,l)=dble(mat(l,l))
      enddo
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=dconjg(mat(l,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rm1hxxxzz_s
!
! Formation of density type matrix by multiplication of a complex 
! tensor of third order with itself. The tensor is not completely
! stored, and the indices are managed by the index tensor. The matrix 
! formed is hermitian:
!
!            index(i,j,k)=x 
!            index(i,l,k)=x1 
!            dconjg(ten(x))*ten(x1) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rm1hxxxzz_s (ten,mat,dim1,dim2,dim3,index)

      implicit none

      integer(long)                                        :: i, j, &
                                                              k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, &
                                                              dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      complex(dop), dimension(*), intent(in)               :: ten
      complex(dop), dimension(dim2,dim2), intent(out)      :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        mat(j,l)=mat(j,l)+dconjg(ten(x))*ten(x1)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  x=index(1,j,k)
                  x1=index(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(ten(x))*ten(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(ten(x))*ten(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               x=index(1,j,1)
               x1=index(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  mat(j,l)=mat(j,l)+dconjg(ten(x))*ten(x1)
               endif
            enddo
         enddo
      endif
 
!  now form other half of matrix
 
      do l=1,dim2
         mat(l,l)=dble(mat(l,l))
      enddo
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=dconjg(mat(l,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmhxxxzz_s
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order. The tensors are not completely stored, and the 
! indices are managed by the index tensors. The matrix formed is hermitian:
!
!            index(i,j,k)=x 
!            index1(i,l,k)=x1 
!            dconjg(bra(x))*ket(x1) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmhxxxzz_s (bra,ket,mat,dim1,dim2,dim3,index,index1)

      implicit none

      integer(long)                                        :: i, j, &
                                                              k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, &
                                                              dim2, &
                                                              dim3
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(*), intent(in)               :: bra
      complex(dop), dimension(*), intent(in)               :: ket
      complex(dop), dimension(dim2,dim2), intent(out)      :: mat

      do l=1,dim2
         do j=1,dim2
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim2
               do j=l,dim2
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim2
            do j=l,dim2
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
               endif
            enddo
         enddo
      endif
!
! now form other half of matrix
!
      do l=1,dim2
         mat(l,l)=dble(mat(l,l))
      enddo
      do l=1,dim2
         do j=1,l-1
            mat(j,l)=dconjg(mat(l,j))
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmmxxxzz_s
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order. The tensors are not completely stored, and the 
! indices are managed by the index tensors:
!
!            index(i,j,k)=x 
!            index1(i,l,k)=x1 
!            dconjg(bra(x))*ket(x1) = mat(j,l).
!
!-----------------------------------------------------------------------

      subroutine rmmxxxzz_s (bra,ket,mat,dim1,dim2,dim3,dim4,dim5, &
                             index,index1)

      implicit none

      integer(long)                                        :: i, j, &
                                                              k, l, &
                                                              x, x1
      integer(long), intent(in)                            :: dim1, &
                                                              dim2, &
                                                              dim3, &
                                                              dim4, &
                                                              dim5
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
      integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
      complex(dop), dimension(*), intent(in)               :: bra
      complex(dop), dimension(*), intent(in)               :: ket
      complex(dop), dimension(dim4,dim5), intent(out)      :: mat

      do l=1,dim4
         do j=1,dim5
            mat(j,l)=(0.0d0,0.0d0)
         enddo
      enddo

      if (dim1.ne.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim5
               do j=1,dim4
                  do i=1,dim1
                     x=index(i,j,k)
                     x1=index1(i,l,k)
                     if (x .ne. 0 .and. x1 .ne. 0) then
                        mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.ne.1) then
         do k=1,dim3
            do l=1,dim5
               do j=1,dim4
                  x=index(1,j,k)
                  x1=index1(1,l,k)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.ne.1 .and. dim3.eq.1) then
         do l=1,dim5
            do j=1,dim4
               do i=1,dim1
                  x=index(i,j,1)
                  x1=index1(i,l,1)
                  if (x .ne. 0 .and. x1 .ne. 0) then
                     mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
                  endif
               enddo
            enddo
         enddo
      else if (dim1.eq.1 .and. dim3.eq.1) then
         do l=1,dim5
            do j=1,dim4
               x=index(1,j,1)
               x1=index1(1,l,1)
               if (x .ne. 0 .and. x1 .ne. 0) then
                  mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
               endif
            enddo
         enddo
      endif

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine rmqxxxzz_s
!
! Formation of density type matrix by multiplication of two complex 
! tensors of third order. The tensors are not completely stored, and the 
! indices are managed by the index tensors:
!
!            index(i,j,k)=x 
!            index1(i,l,k)=x1 
!            dconjg(bra(x))*ket(x1) = mat(j,l).
!
!-----------------------------------------------------------------------

!     subroutine rmqxxxzz_s (bra,ket,mat,dim1,dim2,dim3,index,index1)

!     implicit none

!     integer(long)                                       :: i, j, &
!                                                            k, l, &
!                                                            x, x1
!     integer(long), intent(in)                           :: dim1, &
!                                                            dim2, &
!                                                            dim3
!     integer(long), dimension(dim1,dim2,dim3), intent(in) :: index
!     integer(long), dimension(dim1,dim2,dim3), intent(in) :: index1
!     complex(dop), dimension(*), intent(in)               :: bra
!     complex(dop), dimension(*), intent(in)               :: ket
!     complex(dop), dimension(dim2,dim2), intent(out)      :: mat

!     do l=1,dim2
!        do j=1,dim2
!           mat(j,l)=(0.0d0,0.0d0)
!        enddo
!     enddo

!     if (dim1.ne.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do l=1,dim2
!              do j=1,dim2
!                 do i=1,dim1
!                    x=index(i,j,k)
!                    x1=index1(i,l,k)
!                    if (x .ne. 0 .and. x1 .ne. 0) then
!                       mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
!                    endif
!                 enddo
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.ne.1) then
!        do k=1,dim3
!           do l=1,dim2
!              do j=1,dim2
!                 x=index(1,j,k)
!                 x1=index1(1,l,k)
!                 if (x .ne. 0 .and. x1 .ne. 0) then
!                    mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
!                 endif
!              enddo
!           enddo
!        enddo
!     else if (dim1.ne.1 .and. dim3.eq.1) then
!        do l=1,dim2
!           do j=1,dim2
!              do i=1,dim1
!                 x=index(i,j,1)
!                 x1=index1(i,l,1)
!                 if (x .ne. 0 .and. x1 .ne. 0) then
!                    mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
!                 endif
!              enddo
!           enddo
!        enddo
!     else if (dim1.eq.1 .and. dim3.eq.1) then
!        do l=1,dim2
!           do j=1,dim2
!              x=index(1,j,1)
!              x1=index1(1,l,1)
!              if (x .ne. 0 .and. x1 .ne. 0) then
!                 mat(j,l)=mat(j,l)+dconjg(bra(x))*ket(x1)
!              endif
!           enddo
!        enddo
!     endif

!     return
!     end subroutine

      end module rmlib
