!########################################################################
!                                                                       #
!                        SCHMIDTORTHO                                   #
!                                                                       #
! Library module making Schmidt-orthogonalization.                      #
!                                                                       #
! Contains:                                                             #
!       schmidtortho: Schmidt-orthogonalizes a set of complex vectors   #
!                     in column form.                                   #
!                                                                       #
! V7.0 MB                                                               #
!########################################################################

      module schmidtorthomod

use decimal, only: dop, long
use xvlib,   only: xvixdzo, vvaxzz, xvxxzzr,xvixddo, vvtxdd, xvxxddr
use op1lib,  only: normvxd, normvxz, tranqxd

      implicit none
      private
      public :: schmidtortho, schmidtorthod, schmidtortho2, schmidtortho3, &
                schmidtorthor
      contains


!-----------------------------------------------------------------------
!                         SCHMIDTORTHO
!
! Schmidt-orthogonalizes a set of complex vectors,
! (e.g. single-particle functions).
! The vectors are the column-vectors of psi(gdim,dim)
!-----------------------------------------------------------------------

      subroutine schmidtortho (psi,gdim,dim,ierr)

      implicit none

      integer(long)                                    :: e, e1
      integer(long), intent(in)                        :: dim, gdim
      integer(long), intent(out)                       :: ierr
      real(dop)                                        :: norm
      complex(dop)                                     :: overlap
      complex(dop), dimension(gdim,dim), intent(inout) :: psi

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------
      ierr = 0
      do e=1,dim
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxz(psi(1,e),norm,gdim)
         if (norm .le. 1.0e-99_dop ) norm = 1.0e99_dop
         call xvixdzo(norm,psi(1,e),gdim)
      enddo

      do e=1,dim
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxz(psi(1,e),norm,gdim)
         if( norm .le. 0.8_dop ) then
            ierr = e 
            return
         end if
         call xvixdzo(norm,psi(1,e),gdim)
      enddo

      return
      end subroutine


!!----------------------------------------------------------------------
!                            SCHMIDTORTHOD
!
! Schmidt-orthogonalizes (row-orthogonalization) a real quadratic
! matrix psi(dim,dim).
!
! SS 11/98
!-----------------------------------------------------------------------

      subroutine schmidtorthod (psi,dim)

      implicit none

      integer(long)                                :: e, e1, i
      integer(long), intent(in)                    :: dim
      real(dop)                                    :: norm, overlap
      real(dop), dimension(dim,dim), intent(inout) :: psi

      call tranqxd(psi,dim)

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the matrix.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------
      do i=1,2
         do e=1,dim
            do e1=1,e-1
               call vvtxdd(psi(1,e1),psi(1,e),overlap,dim)
               call xvxxddr(overlap,psi(1,e1),psi(1,e),dim)
            enddo
            call normvxd(psi(1,e),norm,dim)
!            if(abs(norm-1.d0) .gt. 0.01d0) norm=1.d99
            call xvixddo(norm,psi(1,e),dim)
         enddo
      enddo

      call tranqxd(psi,dim)

      return
      end subroutine


!-----------------------------------------------------------------------
!                         SCHMIDTORTHO2
!
! Schmidt-orthogonalizes a set of complex vectors,
! (e.g. single-particle functions).
! The vectors are the column-vectors of psi(gdim,dim)
! This routine is identical to schmidtortho on lib/linear/schmidtortho.f
! except for the additional parameter mnorm (minmal norm).
! HDM 10/05
!-----------------------------------------------------------------------

      subroutine schmidtortho2(psi,gdim,dim,ierr,mnorm)

      implicit none

      integer(long)                                    :: e, e1
      integer(long), intent(in)                        :: dim, gdim
      integer(long), intent(out)                       :: ierr
      real(dop)                                        :: norm, mnorm
      complex(dop)                                     :: overlap
      complex(dop), dimension(gdim,dim), intent(inout) :: psi
      
!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------
      mnorm = 1.d9
      ierr = 0
      do e=1,dim
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxz(psi(1,e),norm,gdim)
         mnorm = min(norm,mnorm)
         if (norm .le. 1.0d-99 ) norm = 1.0d99
         call xvixdzo(norm,psi(1,e),gdim)
      enddo

      do e=1,dim
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxz(psi(1,e),norm,gdim)
         if( norm .le. 0.8d0 ) then
            ierr = e
            return
         end if
         call xvixdzo(norm,psi(1,e),gdim)
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
!                         SCHMIDTORTHO3
!
! Schmidt-orthogonalizes a single vector phi against a set of 
! complex vectors, (e.g. single-particle functions).
! The vectors are the column-vectors of psi(gdim,dim)
! This routine is similar to schmidtortho on lib/linear/schmidtortho.f
! HDM 10/06
!-----------------------------------------------------------------------

      subroutine schmidtortho3(psi,phi,gdim,dim,norm)

      implicit none

      integer(long)                                 :: e
      integer(long), intent(in)                     :: dim, gdim
      real(dop)                                     :: norm
      complex(dop)                                  :: overlap
      complex(dop), dimension(gdim,dim), intent(in) :: psi
      complex(dop), dimension(gdim), intent(inout)  :: phi

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made only once.
!-----------------------------------------------------------------------
      do e=1,dim
         call vvaxzz(psi(1,e),phi,overlap,gdim)
         call xvxxzzr(overlap,psi(1,e),phi,gdim)
      enddo
      call normvxz(phi,norm,gdim)
      if (norm .ge. 1.d-50 ) call xvixdzo(norm,phi,gdim)

      return
      end subroutine

!-----------------------------------------------------------------------
!                         SCHMIDTORTHOR
!
! Schmidt-orthogonalizes a set of real vectors,
! (e.g. natpots).
! The vectors are the column-vectors of psi(gdim,dim)
! This routine is identical to schmidtortho2 on lib/linear/schmidtortho.f
! except for real*8 arithmetic.
! HDM 10/06
!-----------------------------------------------------------------------

      subroutine schmidtorthor(psi,gdim,dim,ierr,mnorm)

      implicit none

      integer(long)                                    :: e, e1
      integer(long), intent(in)                        :: dim, gdim
      integer(long), intent(out)                       :: ierr
      real(dop)                                        :: norm, mnorm
      real(dop)                                        :: overlap
      real(dop), dimension(gdim,dim), intent(inout)    :: psi

!     integer    gdim,dim,e,e1,ierr
!     real*8     psi(gdim,dim),overlap
!     real*8     norm, mnorm

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------
      mnorm = 1.d9
      ierr = 0
      do e=1,dim
         do e1=1,e-1
            call vvtxdd(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxddr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxd(psi(1,e),norm,gdim)
         mnorm = min(norm,mnorm)
         if (norm .le. 1.0d-99 ) norm = 1.0d99
         call xvixddo(norm,psi(1,e),gdim)
      enddo

      do e=1,dim
         do e1=1,e-1
            call vvtxdd(psi(1,e1),psi(1,e),overlap,gdim)
            call xvxxddr(overlap,psi(1,e1),psi(1,e),gdim)
         enddo
         call normvxd(psi(1,e),norm,gdim)
         if( norm .le. 0.8d0 ) then
            ierr = e
            return
         end if
         call xvixddo(norm,psi(1,e),gdim)
      enddo

      return
      end subroutine

      end module schmidtorthomod
!#######################################################################
