!#######################################################################
!
! Simultaneous Diagonalisation of multiple matri!es
!
! From Uwe Manthe and Rob Haarvelt
!
! Updated to F90. Library routines changed for Quantics routines.
!######################################################################

!######################################################################
! Simultaneous diagonalization 
!----------------------------------------------------------------------

      subroutine simul_diag (matrix, dim, zahl, trafo)

use xvlib, only: xqxxdzo
use mmlib, only: qqtxzz,qqcxzz

      implicit none

      integer      dim, zahl, iter, maxsteps, i, nr, step
      complex*16   matrix(dim,dim,zahl), trafo(dim,dim)
      complex*16   work(dim,dim)
      real*8       eps, locrel, x
      parameter    (eps=1.d-12, maxsteps=10)

! check for trivial result
      if (dim.eq.1) return
! transform matrices according to initial guess
      do nr=1,zahl
         call qqtxzz(matrix(1,1,nr),trafo,work,dim)
         call qqcxzz(trafo,work,matrix(1,1,nr),dim)
      enddo 

! simulanteous diagonalization by Jacobi algorithm
      iter=20
      call jacobi_sd(matrix, dim, zahl, trafo, eps, iter, locrel)
! try to avoid fix points
      step=1
      do while ((iter.eq.0).and.(step.lt.maxsteps))
         do i=1,zahl
            x=1+cos(step*i*3.14d0/(zahl+1))
            call xqxxdzo(x,matrix,dim)
         enddo
         iter=1
         call jacobi_sd(matrix, dim, zahl, trafo, eps, iter, locrel)
         do i=1,zahl
            x=1+cos(step*i*3.14d0/(zahl+1))
            x=1/x
            call xqxxdzo(x,matrix,dim)
         enddo
         iter=20
         call jacobi_sd(matrix, dim, zahl, trafo, eps, iter, locrel)
         step=step+1
      enddo
  !    if (locrel .gt. 1.0d-4) then
  !       write (6,*) 'Warning: Incomplete simultaneous diagonalization'
  !       write (6,*) 'relative error in localization: ',locrel
  !       write (6,*)
  !    endif 
         
      end 

!-----------------------------------------------------------------------
! Perform simultaneous Jacobi rotations
!----------------------------------------------------------------------

      subroutine jacobi_rot (matrix, dim, zahl, p, q, c, s)

      implicit none

      integer      dim, zahl
      complex*16   matrix(dim,dim,zahl)
      complex*16   c, s, pcopy, qcopy
      integer      i, nr, p, q

! for all simultaneous matrices
      do nr=1,zahl
! first rotatation, A P^H
         do i=1,dim
            pcopy=matrix(i,p,nr)
            qcopy=matrix(i,q,nr)
            matrix(i,p,nr)=c*pcopy+s*qcopy
            matrix(i,q,nr)=c*qcopy-dconjg(s)*pcopy
         enddo
!  second rotation , P ( A P^H)
         do i=1,dim
            pcopy=matrix(p,i,nr)
            qcopy=matrix(q,i,nr)
            matrix(p,i,nr)=c*pcopy+dconjg(s)*qcopy
            matrix(q,i,nr)=c*qcopy-s*pcopy
         enddo
      enddo

      end

!######################################################################
! Consider Jacobi rotation in the trafo matrix 
!----------------------------------------------------------------------

      subroutine trafo_rot (trafo, dim, p, q, c, s)

      implicit none

      integer      dim, i, p, q
      complex*16   trafo(dim,dim), c, s, pcopy, qcopy

! rotate trafo matrix 
      do i=1,dim
         pcopy=trafo(i,p)
         qcopy=trafo(i,q)
         trafo(i,p)=c*pcopy+s*qcopy
         trafo(i,q)=c*qcopy-dconjg(s)*pcopy
      enddo
      
      end

!----------------------------------------------------------------------
! Simultaneously diagonalization using Jacobi algorithm
!----------------------------------------------------------------------

      subroutine jacobi_sd(matrix, dim, zahl, trafo, eps, iter, locrel)

      implicit none
      integer      dim , zahl
      complex*16   matrix (dim,dim,zahl), trafo(dim,dim)
      integer      p, q, iter
      complex*16   c, s
      real*8       eps, loc, loc2, locrel, loc2rel
      logical*4    converged

! set start values
      converged=.false.
      locrel=1d99
      call measureloc (matrix, dim, zahl, loc)
! iterate
      do while ((.not.converged).and.(iter.gt.0))
! update history 
         loc2=loc
         loc2rel=locrel
! sweep through all rotations
         do p=1,dim-1
            do q=p+1,dim
               call jacobi_angles(matrix, dim, zahl, p, q, c, s)
               call jacobi_loc (matrix, dim, zahl, p, q, c, s, loc)
               if (loc.gt.0) then
                  call jacobi_rot (matrix, dim, zahl, p, q, c, s)
                  call trafo_rot (trafo, dim, p, q, c, s)
               endif
            enddo
         enddo
! check convergence
         call measureloc (matrix, dim, zahl, loc)      
         locrel=(loc-loc2)/abs(loc)
         if (locrel.lt.eps) then
            converged=.true.
         else
            iter=iter-1
         endif
! check whether fix point
         if ((loc2rel.lt.locrel).or.(iter.eq.0)) then
            iter=0
         endif
      enddo
      end 

!----------------------------------------------------------------------
! Compute Jacobi rotation angles
!----------------------------------------------------------------------

      subroutine jacobi_angles (matrix, dim, zahl, p, q, c, s)

      implicit none

      integer      dim, zahl, p, q, i, j, nr, ierr
      complex*16   matrix(dim,dim,zahl), c, s, h(3)
      !real*8       g(3,3), ev(3,3), ew(3), eps , work(9)
      real*8       ev(3,3), ew(3), work(9)
      real*8       x, y, z, r, t

      ev(:,:) = 0.0d0
      do nr=1,zahl
         h(1)=matrix(p,p,nr)-matrix(q,q,nr)
         h(2)=matrix(p,q,nr)+matrix(q,p,nr)
         h(3)=(0,-1.d0)*(matrix(p,q,nr)-matrix(q,p,nr))
         do i=1,3
            do j=1,3
               ev(i,j)=ev(i,j)+dble(dconjg(h(i))*h(j))
            enddo
         enddo
      enddo
!     call rDiag(g,3,3,3,ew,ev,3,3,.false.)
      call dsyev('V','U',3,ev,3,ew,work,9,ierr)
      if (ierr .ne. 0) then
         print*,'Jacobi_Angles (simul_diag.f90). Failed to diagonalise matrix'
      endif
      x=ev(1,3)
      y=ev(2,3)
      z=ev(3,3)
      if (x.lt.0) then
         x=-x
         y=-y
         z=-z
      endif 
      r=dsqrt(x**2+y**2+z**2)
      t=dsqrt(2*r*(x+r))
      c=(x+r)/t
      s=(y-(0.d0,1.0d0)*z)/t
      
      end 
     
!----------------------------------------------------------------------
! Evalute effect of Jacobi rotations on the localization
!----------------------------------------------------------------------

      subroutine jacobi_loc (matrix, dim, zahl, p, q, c, s, diffloc)

      implicit none

      integer      dim, zahl
      complex*16   neu(2), matrix(dim,dim,zahl), copy(2,2)
      complex*16   c, s, ctmp
      real*8       diffloc
      integer      nr, p, q

      diffloc=0.d0
! for all simultaneous matrices
      do nr=1,zahl
! first rotatation, A P^H
         copy(1,1)=c*matrix(p,p,nr)+s*matrix(p,q,nr)
         copy(1,2)=c*matrix(p,q,nr)-dconjg(s)*matrix(p,p,nr)
         copy(2,1)=c*matrix(q,p,nr)+s*matrix(q,q,nr)
         copy(2,2)=c*matrix(q,q,nr)-dconjg(s)*matrix(q,p,nr)
! second rotation , P ( A P^H)
         neu(1)=c*copy(1,1)+dconjg(s)*copy(2,1)
         neu(2)=c*copy(2,2)-s*copy(1,2)
! sum difference of the diagonals 
         ctmp=neu(1)*dconjg(neu(1))+neu(2)*dconjg(neu(2))  &
                        -matrix(p,p,nr)*dconjg(matrix(p,p,nr)) &
                        -matrix(q,q,nr)*dconjg(matrix(q,q,nr))
         diffloc=diffloc+dble(ctmp)
      enddo

      end

!----------------------------------------------------------------------
! Measure localization
!----------------------------------------------------------------------

      subroutine measureloc (matrix, dim, zahl, loc)

      implicit none

      integer      dim, zahl, i, nr
      complex*16   matrix(dim,dim,zahl)
      real*8       loc
      
      loc=0.d0
      do nr=1,zahl         
         do i=1,dim
            loc=loc+dble(matrix(i,i,nr)*dconjg(matrix(i,i,nr)))
         enddo         
      enddo      
      end

