! **********************************************************************
!                                                                     
!                              TTLIB                                  
!                                                                     
!  Library module containing linear algebra routines that involve the 
!  multiplication of two tensors of third order.           
!                                                                     
!  Nomenclature:
!    Each name has 6 basic characters:
!                                                   
!    First 2 characters denote the type of tensors
!       d1: tensors of denote operators of type I  
!       d2: tensors of density operators of type II
!                                                      
!    First 3 and 4 character denote the objects being multiplied:
!       t: 2nd index of tensor of third order
!       t1: 1st index of tensor of third order
!    Character 5 denotes how first object is used:
!       x: unchanged from input
!       t: transpose of input
!       a: adjoint of input
!       c: complex conjugate of input
!    Character 6 denotes how second object is used:
!       see Character 5 above.
!    Character 7, 8 and 9 denote the kind of reduction
!       respectively:
!       12:
!       21:
!       12:
!       21:
!    Character 10, 11 and 12 denote the kind of reduction
!       see Characters 7, 8 and 9
!
!   d2ttcxR12r21 (tens1,tens2,mat,dim1,dim2,dim3)
!       Definition: dconjg(tens1(b,i,m,k)) * tens2(i,n,k,b) = c(m,n) 
!       Dimensions: a(dim2,dim2),b(dim1,dim2,dim3),c(dim1,dim2,dim3)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      module ttlib

use decimal, only: dop, long

      implicit none
      private
      public :: d1ttxx, d2ttxx1221, d2ttxx1221a

      contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine d1ttxx(tens1,tens2,mat,vdim,dim,ndim)

      implicit none

      integer(long)                                      :: i, k, m, n
      integer(long), intent(in)                          :: vdim, dim, &
                                                            ndim
      complex(dop), dimension(vdim,dim,ndim), intent(in) :: tens1
      complex(dop), dimension(vdim,dim,ndim), intent(in) :: tens2
      complex(dop), dimension(dim,dim), intent(out)      :: mat

      do m=1,dim
         do n=1,dim
            mat(m,n) = (0.0d0,0.0d0)
         enddo
      enddo

      do i=1,vdim
         do k=1,ndim
            do m=1,dim
               do n=1,dim
                  mat(m,n) = mat(m,n) + tens1(i,m,k)*tens2(i,n,k)
               enddo
            enddo
         enddo
      enddo
      
      return
      end subroutine

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine d2ttxx1221(tens1,tens2,mat,bdim,vdim,dim,ndim)

      implicit none

      integer(long)                                           :: b, i, &
                                                                 k, m, &
                                                                 n
      integer(long), intent(in)                               :: bdim, &
                                                                 vdim, &
                                                                 dim, &
                                                                 ndim
      complex(dop), dimension(bdim,vdim,dim,ndim), intent(in) :: tens1
      complex(dop), dimension(vdim,dim,ndim,bdim), intent(in) :: tens2
      complex(dop), dimension(dim,dim), intent(out)           :: mat


      do m=1,dim
         do n=1,dim
            mat(m,n) = (0.0d0,0.0d0)
         enddo
      enddo

      do b=1,bdim
         do i=1,vdim
            do k=1,ndim
               do m=1,dim
                  do n=1,dim
                     mat(m,n) = mat(m,n) + tens1(b,i,m,k)*tens2(i,n,k,b)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return 
      end subroutine
     
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


      subroutine d2ttxx1221a(tens1,tens2,mat,bdim,vdim,dim,ndim)

      implicit none

      integer(long)                                           :: b, i, &
                                                                 k, m, &
                                                                 n
      integer(long), intent(in)                               :: bdim, &
                                                                 vdim, &
                                                                 dim, &
                                                                 ndim
      complex(dop), dimension(bdim,vdim,dim,ndim), intent(in) :: tens1
      complex(dop), dimension(vdim,dim,ndim,bdim), intent(in) :: tens2
      complex(dop), dimension(dim,dim), intent(out)           :: mat


      do b=1,bdim
         do i=1,vdim
            do k=1,ndim
               do m=1,dim
                  do n=1,dim
                     mat(m,n) = mat(m,n) + tens1(b,i,m,k)*tens2(i,n,k,b)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine

      end module ttlib
