!*******************************************************************
!                                                                     
!                              XVLIB                                  
!                                                                     
!  Library module containing linear algebra routines that involve the 
!  multiplication of vectors or scalars.
!                                                                     
!  Nomenclature:
!    Each name has 6 basic characters:
!    First 2 characters denote the objects being multiplied:
!       q: quadratic matrix
!       m: general (rectangular) matrix
!       h: hermitian matrix
!       p: positive definite matrix
!       s: symmetric matrix
!       u: unitary matrix
!       d: diagonal matrix (only diagonal elements are supplied as a
!          vector)
!       t: tensor of third order
!       v: vector
!       x: scalar
!       e.g. 'xv' denotes the operation (scalar * vector)
!    Character 3 denotes how first object is used:
!       x: unchanged from input
!       t: transpose of input
!       a: adjoint of input
!       c: complex conjugate of input
!       i: inverse (real and complex scalar only)
!    Character 4 denotes how second object is used:
!       see Character 3 above.
!    Character 5, 6 denote data types of first, second object
!       respectively:
!       z: complex double precision (complex*16)
!       c: complex single precision (complex*8)
!       d: real double precision (real*8)
!       r: real single precision (real*4)
!       y: complex matrix stored as two double precision (real*8)
!          matrices
!     Further characters, if present, give more informaion:
!       a: the result is added to a further object
!       r: the result is subtracted (removed) from a further object
!       o: the result is stored (overwritten) in the second input object
!       c: the input matrices commute
!       h: the resulting matrix is hermitian
!       s: the resulting matrix is symmetric
!       1: the physical dimensions of the matrices differs from those
!          used.
!   
!   xqxxzz (x,a,c,dim)
!       Definition: x*a(j,i) = c(j,i)
!       Dimensions: a(dim,dim),c(dim,dim),x
!
!   xmxxzz (x,a,c,dim1,dim2)
!       Definition: x*a(j,i) = c(j,i)
!       Dimensions: a(dim1,dim2),c(dim1,dim2),x
!
!   xmxxzza (x,a,c,dim1,dim2)
!       Definition: x*a(j,i) + c(j,i) = c(j,i)
!       Dimensions: a(dim1,dim2),c(dim1,dim2),x
!
!   xmxxdza (x,a,c,dim1,dim2)
!       Definition: x*a(j,i) + c(j,i) = c(j,i)
!       Dimensions: a(dim1,dim2),c(dim1,dim2),x
!
!   xmxtzza (x,a,c,dim1,dim2)
!       Definition: x*a(i,j) + c(j,i) = c(j,i)
!       Dimensions: a(dim2,dim1),c(dim1,dim2),x
!
!   xqxxdd (x,a,c,dim)
!       Definition: x*a(j,i) = c(j,i)
!       Dimensions: a(dim,dim),c(dim,dim),x
!
!   xqxxdda (x,a,c,dim)
!       Definition: x*a(j,i) + c(j,i) = c(j,i)
!       Dimensions: a(dim,dim),c(dim,dim),x
!
!   xqxxzza (x,a,c,dim)
!       Definition: x*a(j,i) + c(j,i) = c(j,i)
!       Dimensions: a(dim,dim),c(dim,dim),x
!
!   xqxxdda1 (x,a,c,phdim,dim)
!       Definition: x*a(j,i) + c(j,i) = c(j,i) ; 1 <= i,j <= dim
!       Dimensions: a(phdim,dim),c(phdim,dim),x
!
!   xqxxzd (x,a,c,dim)
!       Definition: x*a(j,i) = c(j,i)
!       Dimensions: a(dim,dim),c(dim,dim),x
!
!   xdxxdda (x,a,c,dim)
!       Definition: x*a(i) + c(i,i) = c(i,i)
!       Dimensions: a(dim),c(dim,dim),x
!
!   xdxsdda (x,a,c,dim)
!       Definition: x*a(i)**2 + c(i,i) = c(i,i)
!       Dimensions: a(dim),c(dim,dim),x
!
!   xuxxdda (x,c,dim)
!       Definition: x + c(i,i) = c(i,i)
!       Dimensions: c(dim,dim),x
!
!   xvixddo (x,v,dim)
!       Definition: inv(x)*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvixzzo (x,v,dim)
!       Definition: inv(x)*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvixdzo (x,v,dim)
!       Definition: inv(x)*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvxxzza (x,v,w,dim)
!       Definition: x*v(i) + w(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxzz (x,v,w,dim)
!       Definition: x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxdd (x,v,w,dim)
!       Definition: x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvixdd (x,v,w,dim)
!       Definition: inv(x)*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxczza (x,v,w,dim)
!       Definition: x*dconjg(v(i)) + w(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxdzr (x,v,w,dim)
!       Definition: w(i) - x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxdza (x,v,w,dim)
!       Definition: w(i) + x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxddr (x,v,w,dim)
!       Definition: w(i) - x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxdda (x,v,w,dim)
!       Definition: x*v(i) + w(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxzzo (x,v,dim)
!       Definition: x*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvxxdzo (x,v,dim)
!       Definition: x*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvxxddo (x,v,dim)
!       Definition: x*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xvxxzzr (x,v,w,dim)
!       Definition: w(i) - x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxii (x,v,w,dim)
!       Definition: x*v(i) = w(i)
!       Dimensions: v(dim),w(dim),x
!
!   xvxxiio (x,v,dim)
!       Definition: x*v(i) = v(i)
!       Dimensions: v(dim),x
!
!   xqxxzzo (x,a,dim)
!       Definition: x*a(j,i) = a(j,i)
!       Dimensions: a(dim,dim),x
!
!   xqxxdzo (x,a,dim)
!       Definition: x*a(j,i) = a(j,i)
!       Dimensions: a(dim,dim),x
!
!   xqixdzo (x,a,dim)
!       Definition: inv(x)*a(j,i) = a(j,i)
!       Dimensions: a(dim,dim),x
!
!   xmxxzzo (x,a,dim1,dim2)
!       Definition: x*a(j,i) = a(j,i)
!       Dimensions: a(dim1,dim2),x
!
!   xmxxdzo (x,a,dim1,dim2)
!       Definition: x*a(j,i) = a(j,i)
!       Dimensions: a(dim1,dim2),x
!
!   vvtxdd (u,v,x,dim)
!       Definition: u(i)*v(i) = x
!       Dimensions: u(dim),v(dim),x
!
!   vvtxzz (u,v,x,dim)
!       Definition: u(i)*v(i) = x
!       Dimensions: u(dim),v(dim),x
!
!   vvtxdz (u,v,x,dim)
!       Definition: u(i)*v(i) = x
!       Dimensions: u(dim),v(dim),x
!
!   vvtxzza (u,v,x,dim)
!       Definition: u(i)*v(i) +x = x
!       Dimensions: u(dim),v(dim),x
!
!   vvaxzz (u,v,x,dim)
!       Definition: dconjg(u(i))*v(i) = x
!       Dimensions: u(dim),v(dim),x
!
!   vvxazz (u,v,m,dim)
!       Definition: u(i)*dconjg(v(j)) = m(i,j)
!       Dimensions: u(dim),v(dim),m(dim,dim)
!
!   vvctzz (u,v,m,dim)
!       Definition: dconjg(u(i))*v(j) = m(i,j)
!       Dimensions: u(dim),v(dim),m(dim,dim)
!
!   qvxxdd (a,v,w,dim)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvxxzz (a,v,w,dim)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvxxdd1 (a,v,w,phdim,dim)
!       Definition: a(i,j)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   qvtxdd (a,v,w,dim)
!       Definition: a(j,i)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvtxdd1 (a,v,w,phdim,dim)
!       Definition: a(j,i)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   mvtxdd (a,v,w,dim1,dim2)
!       Definition: a(j,i)*v(j) = w(i)
!       Dimensions: a(dim1,dim2),v(dim1),w(dim2)
!
!   mvaxzz (a,v,w,dim1,dim2)
!       Definition: dconjg(a(j,i))*v(j) = w(i)
!       Dimensions: a(dim1,dim2),v(dim1),w(dim2)
!
!   mvxxdd1 (a,v,w,phdim,dim1,dim2)
!       Definition: a(i,j)*v(j) = w(i)
!       Dimensions: a(dim2,dim1),v(dim1),w(dim2) (= used dimensions)
!       Dimensions: a(phdim,phdim) (= allocated dimensions)
!
!   mvtxdd1 (a,v,w,phdim,dim1,dim2)
!       Definition: a(j,i)*v(j) = w(i)
!       Dimensions: a(dim1,dim2),v(dim1),w(dim2) (= used dimensions)
!       Dimensions: a(phdim,phdim) (= allocated dimensions)
!
!   mvtxzz (a,v,w,dim1,dim2)
!       Definition: a(j,i)*v(j) = w(i) .
!       Dimensions: a(dim1,dim2),v(dim1),w(dim2)
!
!   mvxxzz (a,v,w,dim1,dim2)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim1,dim2),v(dim2),w(dim1)
!
!   mvxxdd (a,v,w,dim1,dim2)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim1,dim2),v(dim2),w(dim1)
!
!   dvxxdd (a,v,w,dim)
!       Definition: a(i)*v(i) = w(i) .
!       Dimensions: a(dim),v(dim),w(dim)
!
!   dvxxdz (a,v,w,dim)
!       Definition: a(i)*v(i) = w(i) .
!       Dimensions: a(dim),v(dim),w(dim)
!
!   dvxxdzo (a,v,dim)
!       Definition: a(i)*v(i) = v(i) .
!       Dimensions: a(dim),v(dim)
!
!   dvxxddo (a,v,dim)
!       Definition: a(i)*v(i) = v(i) .
!       Dimensions: a(dim),v(dim)
!
!   qvxxdz (a,v,w,dim)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvxxdz1 (a,v,w,phdim,dim)
!       Definition: a(i,j)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   qvtxdz (a,v,w,dim)
!       Definition: a(j,i)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvtxdz1 (a,v,w,phdim,dim)
!       Definition: a(j,i)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   qvxxzz1 (a,v,w,phdim,dim)
!       Definition: a(i,j)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   qvxxyz (a,v,w,dim)
!       Definition: a(i,j)*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvaxzz (a,v,w,dim)
!       Definition: dconjg(a(j,i))*v(j) = w(i) .
!       Dimensions: a(dim,dim),v(dim),w(dim)
!
!   qvaxzz1 (a,v,w,phdim,dim)
!       Definition: dconjg(a(j,i))*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
!   qvtxzz1 (a,v,w,phdim,dim)
!       Definition: a(j,i)*v(j) = w(i) ; 1 <= i,j <= dim .
!       Dimensions: a(phdim,dim),v(dim),w(dim)
!
! ----------------------------------------------------------------------
 
      module xvlib

use decimal, only: dop, long
use openmpmod

      implicit none
      private
      public :: xmxxzza, xmxxdza, xmxtzza, xqxxdd, &
                xqxxdda, xqxxzza, xqxxdda1, xqxxzd, xdxxdda, xdxsdda, &
                xuxxdda, xvixddo, xvixzzo, xvixdzo, xvxxzza, xvxxzz, &
                xvxxdd, xvixdd, xvxczza, xvxxdzr, xvxxdza, xvxxddr, &
                xvxxdda, xvxxzzo, xvxxdzo, xvxxddo, xvxxzzr, xvxxii, &
                xvxxiio, xqxxzzo, xqxxdzo, xqixdzo, xmxxzzo, xmxxdzo, &
                vvtxdd, vvtxzz, vvtxdz, vvtxzza, vvaxzz, vvxazz, vvctzz, &
                qvxxdd, qvxxzz, qvxxdd1, qvtxdd, qvtxdd1, mvtxdd, &
                mvxxdd1, mvtxdd1, mvtxzz, mvxxzz, mvxxdd, dvxxdd, dvxxdz, &
                dvxxdzo, dvxxddo, qvxxdz, qvxxdz1, qvtxdz, qvtxdz1, &
                qvxxzz1, qvxxyz, qvaxzz, qvaxzz1, qvtxzz1, mvaxzz, &
                xqxxzz, xmxxzz 

      contains


!-----------------------------------------------------------------------
! Library subroutine xqxxzz
!
! Multiplication of a complex scalar with a complex quadratic matrix:
!     x*a(j,i)=c(j,i)
!-----------------------------------------------------------------------

      subroutine xqxxzz (x,a,c,dim)

      implicit none
    
      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), intent(in)                      :: x
      complex(dop), dimension(dim,dim), intent(in)  :: a
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxxzz
!
! Multiplication of a complex scalar with a complex rectangular matrix:
!     x*a(j,i)=c(j,i)
!-----------------------------------------------------------------------

      subroutine xmxxzz (x,a,c,dim1,dim2)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim1, dim2
      complex(dop), intent(in)                        :: x
      complex(dop), dimension(dim1,dim2), intent(in)  :: a
      complex(dop), dimension(dim1,dim2), intent(out) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxxzza
!
! Multiplication of a complex scalar with a complex rectangular matrix,
! the result of which is added to a further input complex rectangular
! matrix:
!     x*a(j,i) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xmxxzza (x,a,c,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      complex(dop), intent(in)                          :: x
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim1,dim2), intent(inout) :: c


      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = c(j,i)+x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxxdza
!
! Multiplication of a real scalar with a complex rectangular matrix,
! the result of which is added to a further input complex rectangular
! matrix:
!     x*a(j,i) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xmxxdza (x,a,c,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      real(dop), intent(in)                             :: x
      complex(dop), dimension(dim1,dim2), intent(in)    :: a
      complex(dop), dimension(dim1,dim2), intent(inout) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = c(j,i)+x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxtzza
!
! Multiplication of a complex scalar with a transposed complex
! rectangular matrix, the result of which is added to a further input
! complex rectangular matrix:
!     x*a(i,j) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xmxtzza (x,a,c,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      complex(dop), intent(in)                          :: x
      complex(dop), dimension(dim2,dim1), intent(in)    :: a
      complex(dop), dimension(dim1,dim2), intent(inout) :: c

      do i = 1,dim2
         do j = 1,dim1
            c(j,i) = c(j,i)+x*a(i,j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxdd
!
! Multiplication of a real scalar with a real quadratic matrix:
!     x*a(j,i)=c(j,i)
!-----------------------------------------------------------------------

      subroutine xqxxdd (x,a,c,dim)

      implicit none
      
      integer(long)                              :: i, j
      integer(long), intent(in)                  :: dim
      real(dop), intent(in)                      :: x
      real(dop), dimension(dim,dim), intent(in)  :: a
      real(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxdda
!
! Multiplication of a real scalar with a real quadratic matrix,
! the result of which is added to a further input real quadratic matrix:
!     x*a(j,i) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqxxdda (x,a,c,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      real(dop), intent(in)                        :: x
      real(dop), dimension(dim,dim), intent(in)    :: a
      real(dop), dimension(dim,dim), intent(inout) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = c(j,i)+x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxzza
!
! Multiplication of a complex scalar with a complex quadratic matrix,
! the result of which is added to a further input complex quadratic
! matrix:
!     x*a(j,i) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqxxzza (x,a,c,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), intent(in)                        :: x
      complex(dop), dimension(dim,dim), intent(in)    :: a
      complex(dop), dimension(dim,dim), intent(inout) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = c(j,i)+x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxdda1
!
! Multiplication of a real scalar with a real quadratic matrix,
! the result of which is added to a further input real quadratic matrix:
!     x*a(j,i) + c(j,i) = c(j,i)
!
! NB Input c matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqxxdda1 (x,a,c,phdim,dim)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: phdim, dim
      real(dop), intent(in)                          :: x
      real(dop), dimension(phdim,dim), intent(in)    :: a
      real(dop), dimension(phdim,dim), intent(inout) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = c(j,i)+x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxzd
!
! Multiplication of a complex scalar with a real quadratic matrix:
!     x*a(j,i)=c(j,i)
!-----------------------------------------------------------------------

      subroutine xqxxzd (x,a,c,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      real(dop), dimension(dim,dim), intent(in)     :: a
      complex(dop), intent(in)                      :: x
      complex(dop), dimension(dim,dim), intent(out) :: c

      do i = 1,dim
         do j = 1,dim
            c(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xdxxdda
!
! Multiplication of a real scalar with a real diagonal matrix, where
! the result is added to a complex quadratic matrix
!     x*a(i,i) + c(i,i) = c(i,i)
!-----------------------------------------------------------------------

      subroutine xdxxdda (x,a,c,dim)

      implicit none

      integer(long)                                   :: i
      integer(long), intent(in)                       :: dim
      real(dop), intent(in)                           :: x
      real(dop), dimension(dim), intent(in)           :: a
      complex(dop), dimension(dim,dim), intent(inout) :: c

      do i = 1,dim
         c(i,i) = c(i,i) + x*a(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xdxsdda
!
! Multiplication of a real scalar with a squared real diagonal matrix, 
! where the result is added to a complex quadratic matrix
!     x*a(i,i)**2 + c(i,i) = c(i,i)
!-----------------------------------------------------------------------

      subroutine xdxsdda (x,a,c,dim)

      implicit none

      integer(long)                                   :: i
      integer(long), intent(in)                       :: dim
      real(dop), intent(in)                           :: x
      real(dop), dimension(dim), intent(in)           :: a
      complex(dop), dimension(dim,dim), intent(inout) :: c

      do i = 1,dim
         c(i,i) = c(i,i) + x*a(i)**2
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xuxxdda
!
! Multiplication of a real scalar with the unity matrix, where
! the result is added to a complex quadratic matrix
!     x + c(i,i) = c(i,i)
!-----------------------------------------------------------------------

      subroutine xuxxdda (x,c,dim)

      implicit none
      
      integer(long)                                   :: i
      integer(long), intent(in)                       :: dim
      real(dop), intent(in)                           :: x
      complex(dop), dimension(dim,dim), intent(inout) :: c

      do i = 1,dim
         c(i,i) = c(i,i) + x
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvixddo
!
! Multiplication of the inverse of a real scalar with a real vector,
! where the result is stored in the initial array:
!     inv(x)*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!    This routine can be used for the normalisation of a vector if
!    x is the sqrt of the vector scalar product.
!-----------------------------------------------------------------------

      subroutine xvixddo (x,v,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop)                                :: inv
      real(dop), intent(in)                    :: x
      real(dop), dimension(dim), intent(inout) :: v

      inv=1.0d0/x
      do i = 1,dim
         v(i) = inv*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvixzzo
!
! Multiplication of the inverse of a complex scalar with a complex
! vector, where the result is stored in the initial array:
!     inv(x)*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!    This routine can be used for the normalisation of a vector if
!    x is the sqrt of the vector's SYMMETRIC scalar product.
!-----------------------------------------------------------------------

      subroutine xvixzzo (x,v,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop)                                :: inv
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(inout) :: v

      inv=(1.0d0,0.0d0)/x
      do i = 1,dim
         v(i) = inv*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvixdzo
!
! Multiplication of the inverse of a real scalar with a complex vector,
! where the result is stored in the initial array:
!     inv(x)*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!    This routine can be used for the normalisation of a vector if
!    x is the sqrt of the vector scalar product.
!-----------------------------------------------------------------------

      subroutine xvixdzo (x,v,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      real(dop)                                   :: inv
      real(dop), intent(in)                       :: x
      complex(dop), dimension(dim), intent(inout) :: v

      inv=1.0d0/x
      do i = 1,dim
         v(i) = inv*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxzza
!
! Multiplication of a complex scalar with a complex vector,
! where the result is added to a further vector
!     x*v(i) +w(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxzza (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

!$omp parallel do num_threads(ompthread)
      do i = 1,dim
         w(i) = w(i)+x*v(i)
      enddo
!$omp end parallel do 

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxzz 
!
! Multiplication of a complex scalar with a complex vector
!     x*v(i) = w(i)
!
!-----------------------------------------------------------------------

      subroutine xvxxzz (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxdd 
!
! Multiplication of a real scalar with a real vector
!     x*v(i) = w(i)
!
!-----------------------------------------------------------------------

      subroutine xvxxdd (x,v,w,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop), intent(in)                  :: x
      real(dop), dimension(dim), intent(in)  :: v
      real(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvixdd 
!
! Division of a real vector by a real scalar
!     inv(x)*v(i) = w(i)
!
!-----------------------------------------------------------------------

      subroutine xvixdd (x,v,w,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop)                              :: inv
      real(dop), intent(in)                  :: x
      real(dop), dimension(dim), intent(in)  :: v
      real(dop), dimension(dim), intent(out) :: w

      inv=1.0d0/x
      do i = 1,dim
         w(i) = inv*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxczza
!
! Multiplication of a complex scalar with the complex conjugate of a
! complex vector, where the result is added to a further vector
!     x*dconjg(v(i)) +w(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxczza (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)+x*dconjg(v(i))
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxdzr
!
! Multiplication of a real scalar with a complex vector,
! where the result is subtracted from a further vector
!     w(i) - x*v(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxdzr (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      real(dop), intent(in)                       :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)-x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxdza
!
! Multiplication of a real scalar with a complex vector,
! where the result is added to a further vector
!     w(i) + x*v(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxdza (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      real(dop), intent(in)                       :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)+x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxddr
!
! Multiplication of a real scalar with a real vector,
! where the result is subtracted from a further vector
!     w(i) - x*v(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxddr (x,v,w,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop), intent(in)                    :: x
      real(dop), dimension(dim), intent(in)    :: v
      real(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)-x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxdda
!
! Multiplication of a real scalar with a real vector,
! where the result is added to a further vector
!     x*v(i) +w(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxdda (x,v,w,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop), intent(in)                    :: x
      real(dop), dimension(dim), intent(in)    :: v
      real(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)+x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxzzo
!
! Multiplication of a complex scalar with a complex vector,
! where the result is stored in the input vector
!     x*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxzzo (x,v,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxdzo
!
! Multiplication of a real scalar with a complex vector,
! where the result is stored in the input vector
!     x*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxdzo (x,v,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      real(dop), intent(in)                       :: x
      complex(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxddo
!
! Multiplication of a real scalar with a real vector,
! where the result is stored in the input vector
!     x*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxddo (x,v,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop), intent(in)                    :: x
      real(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxzzr
!
! Multiplication of a complex scalar with a complex vector,
! where the result is subtracted from a further vector
!     w(i) - x*v(i) = w(i)
!
! NB Input w vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxzzr (x,v,w,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      complex(dop), intent(in)                    :: x
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(inout) :: w

      do i = 1,dim
         w(i) = w(i)-x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxii
!
! Multiplication of an integer scalar with an integer vector,
!     x*v(i) = w(i)
!
!-----------------------------------------------------------------------

      subroutine xvxxii (x,v,w,dim)

      implicit none

      integer(long)                             :: i, x
      integer(long), intent(in)                 :: dim
      integer(dop), dimension(dim), intent(in)  :: v
      integer(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xvxxiio
!
! Multiplication of an integer scalar with an integer vector,
! where the result is stored in the input vector
!     x*v(i) = v(i)
!
! NB Input v vector is overwritten on output
!-----------------------------------------------------------------------

      subroutine xvxxiio (x,v,dim)

      implicit none

      integer(long)                               :: i, x
      integer(long), intent(in)                   :: dim
      integer(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = x*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxzzo
!
! Multiplication of a complex scalar with a quadratic complex matrix,
! where the result is stored in the input matrix
!     x*a(j,i) = a(j,i)
!
! NB Input a matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqxxzzo (x,a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      complex(dop), intent(in)                        :: x
      complex(dop), dimension(dim,dim), intent(inout) :: a

!$omp parallel do num_threads(ompthread)
      do i = 1,dim
         do j = 1,dim
            a(j,i) = x*a(j,i)
         enddo
      enddo
!$omp end parallel do 

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqxxdzo
!
! Multiplication of a real scalar with a quadratic complex matrix,
! where the result is stored in the input matrix
!     x*a(j,i) = a(j,i)
!
! NB Input a matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqxxdzo (x,a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      real(dop), intent(in)                           :: x
      complex(dop), dimension(dim,dim), intent(inout) :: a

      do i = 1,dim
         do j = 1,dim
            a(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xqixdzo
!
! Multiplication of the inverse of a real scalar with a quadratic 
! complex matrix, where the result is stored in the input matrix
!     x*a(j,i) = a(j,i)
!
! NB Input a matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xqixdzo (x,a,dim)

      implicit none

      integer(long)                                   :: i, j
      integer(long), intent(in)                       :: dim
      real(dop)                                       :: inv
      real(dop), intent(in)                           :: x
      complex(dop), dimension(dim,dim), intent(inout) :: a

      inv=1.0d0/x
      do i = 1,dim
         do j = 1,dim
            a(j,i) = inv*a(j,i)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxxzzo
!
! Multiplication of a complex scalar with a rectangular complex matrix,
! where the result is stored in the input matrix
!     x*a(j,i) = a(j,i)
!
! NB Input a matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xmxxzzo (x,a,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      complex(dop), intent(in)                          :: x
      complex(dop), dimension(dim1,dim2), intent(inout) :: a


!$omp parallel do num_threads(ompthread)
      do i = 1,dim2
         do j = 1,dim1
            a(j,i) = x*a(j,i)
         enddo
      enddo
!$omp end parallel do 

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine xmxxdzo
!
! Multiplication of a real scalar with a rectangular complex matrix,
! where the result is stored in the input matrix
!     x*a(j,i) = a(j,i)
!
! NB Input a matrix is overwritten on output
!-----------------------------------------------------------------------

      subroutine xmxxdzo (x,a,dim1,dim2)

      implicit none

      integer(long)                                     :: i, j
      integer(long), intent(in)                         :: dim1, dim2
      real(dop), intent(in)                             :: x
      complex(dop), dimension(dim1,dim2), intent(inout) :: a

      do i = 1,dim2
         do j = 1,dim1
            a(j,i) = x*a(j,i)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvtxdd
!
! scalar product of two real vectors:
!     u(i)*v(i)=s
!-----------------------------------------------------------------------

      subroutine vvtxdd (u,v,s,dim)

      implicit none

      integer(long)                         :: i
      integer(long), intent(in)             :: dim
      real(dop), dimension(dim), intent(in) :: u
      real(dop), dimension(dim), intent(in) :: v
      real(dop), intent(out)                :: s

      s = u(1)*v(1)
      do i = 2,dim
         s = s+u(i)*v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvtxzz
!
! symmetric product of two complex vectors:
!     u(i)*v(i)=s
!-----------------------------------------------------------------------

      subroutine vvtxzz (u,v,s,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      complex(dop), dimension(dim), intent(in) :: u
      complex(dop), dimension(dim), intent(in) :: v
      complex(dop), intent(out)                :: s

      s = u(1)*v(1)
      do i = 2,dim
         s = s+u(i)*v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvtxdz
!
! symmetric product of two complex vectors:
!     u(i)*v(i)=s
!-----------------------------------------------------------------------

      subroutine vvtxdz (u,v,s,dim)

      implicit none

      integer    dim,i
      real*8     u(dim)
      complex*16 v(dim),s

      s = u(1)*v(1)
      do i = 2,dim
         s = s+u(i)*v(i)
      enddo

      return
      end

!-----------------------------------------------------------------------
! Library subroutine vvtxzza
!
! symmetric product of two complex vectors (result is added to s):
!     u(i)*v(i)=s
!-----------------------------------------------------------------------

      subroutine vvtxzza (u,v,s,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      complex(dop), dimension(dim), intent(in) :: u
      complex(dop), dimension(dim), intent(in) :: v
      complex(dop), intent(inout)              :: s

      do i = 1,dim
         s = s+u(i)*v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvaxzz
!
! scalar product of two complex vectors:
!     dconjg(u(i))*v(i)=s
!-----------------------------------------------------------------------
      subroutine vvaxzz (u,v,s,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      complex(dop), dimension(dim), intent(in) :: u
      complex(dop), dimension(dim), intent(in) :: v
      complex(dop), intent(out)                :: s

      s = dconjg(u(1))*v(1)
      do i = 2,dim
         s = s+dconjg(u(i))*v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvaxzz
!
! scalar product of two complex vectors:
!     dconjg(u(i))*v(i)=s
!-----------------------------------------------------------------------
!     subroutine vvaxzzp(u,v,s,dim,p)

!     implicit none

!     integer(long)                            :: p
!     integer(long), intent(in)                :: dim
!     complex(dop), dimension(dim), intent(in) :: u
!     complex(dop), dimension(dim), intent(in) :: v
!     complex(dop), intent(out)                :: s

!     s = dconjg(u(p))*v(p)

!     return
!     end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvxazz 
!
!       u(i)*dconjg(v(j)) = m(i,j)
!       Dimensions: u(dim),v(dim),m(dim,dim)
!-----------------------------------------------------------------------
      subroutine vvxazz(u,v,m,dim)
 
      implicit none
      
      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim), intent(in)      :: u
      complex(dop), dimension(dim), intent(in)      :: v
      complex(dop), dimension(dim,dim), intent(out) :: m

      do i=1,dim
         do j=1,dim
            m(i,j) = u(i)*dconjg(v(j))
         enddo
      enddo
      
      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine vvctzz
!
!       dconjg(u(i))*v(j) = m(i,j)
!       Dimensions: u(dim),v(dim),m(dim,dim)
!-----------------------------------------------------------------------
      subroutine vvctzz(u,v,m,dim)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: dim
      complex(dop), dimension(dim), intent(in)      :: u
      complex(dop), dimension(dim), intent(in)      :: v
      complex(dop), dimension(dim,dim), intent(out) :: m

      do i=1,dim
         do j=1,dim
            m(i,j) = dconjg(u(i))*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxdd
!
! Multiplication of a real quadratic matrix with a real vector
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------
      subroutine qvxxdd (a,v,w,dim)

      implicit none

      integer(long)                             :: i, j
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: v
      real(dop), dimension(dim,dim), intent(in) :: a
      real(dop), dimension(dim), intent(out)    :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxzz
!
! Multiplication of a complex quadratic matrix with a complex vector
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvxxzz (a,v,w,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      complex(dop), dimension(dim), intent(in)     :: v
      complex(dop), dimension(dim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(out)    :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxdd1
!
! Multiplication of a real quadratic matrix with a real vector
!     a(i,j)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvxxdd1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: phdim, dim
      real(dop), dimension(dim), intent(in)       :: v
      real(dop), dimension(phdim,dim), intent(in) :: a
      real(dop), dimension(dim), intent(out)      :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine qvtxdd
!
! Multiplication of the transpose of a real quadratic matrix with a 
! real vector
!     a(j,i)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvtxdd (a,v,w,dim)

      implicit none

      integer(long)                             :: i, j
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: v
      real(dop), dimension(dim,dim), intent(in) :: a
      real(dop), dimension(dim), intent(out)    :: w

      do i = 1,dim
         w(i) = a(1,i)*v(1)
         do j = 2,dim
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine qvtxdd1
!
! Multiplication of the transpose of a real quadratic matrix with a 
! real vector
!     a(j,i)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvtxdd1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: phdim, dim
      real(dop), dimension(dim), intent(in)       :: v
      real(dop), dimension(phdim,dim), intent(in) :: a
      real(dop), dimension(dim), intent(out)      :: w

      do i = 1,dim
         w(i) = a(1,i)*v(1)
         do j = 2,dim
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine mvtxdd
!
! Multiplication of the transpose of a real rectangular matrix with a
! real vector
!     a(j,i)*v(j)=w(i)
!
!-----------------------------------------------------------------------

      subroutine mvtxdd (a,v,w,dim1,dim2)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: dim1, dim2
      real(dop), dimension(dim1), intent(in)      :: v
      real(dop), dimension(dim1,dim2), intent(in) :: a
      real(dop), dimension(dim2), intent(out)     :: w

      do i = 1,dim2
         w(i) = a(1,i)*v(1)
         do j = 2,dim1
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine mvaxzz
!
! Multiplication of the adjoint of a complex rectangular matrix with a
! complex vector
!     a(j,i)*v(j)=w(i)
!
!-----------------------------------------------------------------------

      subroutine mvaxzz (a,v,w,dim1,dim2)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: dim1, dim2
      complex(dop), dimension(dim1), intent(in)      :: v
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim2), intent(out)     :: w

      do i = 1,dim2
         w(i) = dconjg(a(1,i))*v(1)
         do j = 2,dim1
            w(i) = w(i)+dconjg(a(j,i))*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine mvxxdd1
!
! Multiplication of a real rectangular matrix with a real vector
!     a(i,j)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim1-2 is used dimension 
!-----------------------------------------------------------------------

      subroutine mvxxdd1 (a,v,w,phdim,dim1,dim2)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: phdim, dim1, dim2
      real(dop), dimension(dim1), intent(in)        :: v
      real(dop), dimension(phdim,dim1), intent(in)  :: a
      real(dop), dimension(dim2), intent(out)       :: w

      do i = 1,dim2
         w(i) = a(i,1)*v(1)
         do j = 2,dim1
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine mvtxdd1
!
! Multiplication of the transpose of a real rectangular matrix with a
! real vector
!     a(j,i)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim1-2 is used dimension 
!-----------------------------------------------------------------------

      subroutine mvtxdd1 (a,v,w,phdim,dim1,dim2)

      implicit none

      integer(long)                                 :: i, j
      integer(long), intent(in)                     :: phdim, dim1, dim2
      real(dop), dimension(dim1), intent(in)        :: v
      real(dop), dimension(phdim,dim2), intent(in)  :: a
      real(dop), dimension(dim2), intent(out)       :: w

      do i = 1,dim2
         w(i) = a(1,i)*v(1)
         do j = 2,dim1
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine mvtxzz
!
! Multiplication of the transpose of a complex rectangular matrix with a
! complex vector
!     a(j,i)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine mvtxzz (a,v,w,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), dimension(dim1), intent(in)      :: v
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim2), intent(out)     :: w

      do i = 1,dim2
         w(i) = a(1,i)*v(1)
         do j = 2,dim1
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine


!-----------------------------------------------------------------------
! Library subroutine mvxxzz
!
! Multiplication of a complex rectangular matrix with a complex vector.
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine mvxxzz (a,v,w,dim1,dim2)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: dim1, dim2
      complex(dop), dimension(dim2), intent(in)      :: v
      complex(dop), dimension(dim1,dim2), intent(in) :: a
      complex(dop), dimension(dim1), intent(out)     :: w

      do i = 1,dim1
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim2
         do i = 1,dim1
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine mvxxdd
!
! Multiplication of a real rectangular matrix with a real vector.
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine mvxxdd (a,v,w,dim1,dim2)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: dim1, dim2
      real(dop), dimension(dim2), intent(in)      :: v
      real(dop), dimension(dim1,dim2), intent(in) :: a
      real(dop), dimension(dim1), intent(out)     :: w

      do i = 1,dim1
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim2
         do i = 1,dim1
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine dvxxdd
!
! Multiplication of a diagonal real matrix with a real vector
!     a(i)*v(i)=w(i)
!-----------------------------------------------------------------------

      subroutine dvxxdd (a,v,w,dim)

      implicit none

      integer(long)                          :: i
      integer(long), intent(in)              :: dim
      real(dop), dimension(dim), intent(in)  :: v
      real(dop), dimension(dim), intent(in)  :: a
      real(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = a(i)*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine dvxxdz
!
! Multiplication of a diagonal real matrix with a complex vector
!     a(i)*v(i)=w(i)
!-----------------------------------------------------------------------

      subroutine dvxxdz (a,v,w,dim)

      implicit none

      integer(long)                             :: i
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim), intent(in)     :: a
      complex(dop), dimension(dim), intent(in)  :: v
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = a(i)*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine dvxxdzo
!
! Multiplication of a diagonal real matrix with a complex vector
!     a(i)*v(i)=v(i)
!-----------------------------------------------------------------------

      subroutine dvxxdzo (a,v,dim)

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: dim
      real(dop), dimension(dim), intent(in)       :: a
      complex(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = a(i)*v(i)
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine dvxxddo
!
! Multiplication of a diagonal real matrix with a real vector
!     a(i)*v(i)=v(i)
!-----------------------------------------------------------------------

      subroutine dvxxddo (a,v,dim)

      implicit none

      integer(long)                            :: i
      integer(long), intent(in)                :: dim
      real(dop), dimension(dim), intent(in)    :: a
      real(dop), dimension(dim), intent(inout) :: v

      do i = 1,dim
         v(i) = a(i)*v(i)
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxdz
!
! Multiplication of a real quadratic matrix with a complex vector
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvxxdz (a,v,w,dim)

      implicit none

      integer(long)                             :: i, j
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)  :: v
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxdz1
!
! Multiplication of a real quadratic matrix with a complex vector
!     a(i,j)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvxxdz1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: phdim, dim
      real(dop), dimension(phdim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(out)   :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine
      
!-----------------------------------------------------------------------
! Library subroutine qvtxdz
!
! Multiplication of a real quadratic matrix with a complex vector
!     a(j,i)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvtxdz (a,v,w,dim)

      implicit none

      integer(long)                             :: i, j
      integer(long), intent(in)                 :: dim
      real(dop), dimension(dim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)  :: v
      complex(dop), dimension(dim), intent(out) :: w

      do i = 1,dim
         w(i) = a(1,i)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvtxdz1
!
! Multiplication of a real quadratic matrix with a complex vector
!     a(j,i)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvtxdz1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: phdim, dim
      real(dop), dimension(phdim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(out)   :: w

      do i = 1,dim
         w(i) = a(1,i)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxzz1
!
! Multiplication of a complex quadratic matrix with a complex vector
!     a(i,j)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvxxzz1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: phdim, dim
      complex(dop), dimension(phdim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)       :: v
      complex(dop), dimension(dim), intent(out)      :: w

      do i = 1,dim
         w(i) = a(i,1)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(i,j)*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvxxyz
!
! Multiplication of a complex quadratic matrix with a complex vector,
! where the complex matrix is stored as 2 real matrices.
!     a(i,j)*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvxxyz (a,v,w,dim)

      implicit none

      integer(long)                               :: i, j
      integer(long), intent(in)                   :: dim
      real(dop), dimension(dim,dim,2), intent(in) :: a
      complex(dop), dimension(dim), intent(in)    :: v
      complex(dop), dimension(dim), intent(out)   :: w

      do i = 1,dim
         w(i) = dcmplx(a(i,1,1),a(i,1,2))*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+dcmplx(a(i,j,1),a(i,j,2))*v(j)
         enddo
      enddo

      return
      end subroutine

! ----------------------------------------------------------------------
! Library subroutine qvaxzz
!
! Multiplication of the adjoint of a complex quadratic matrix with a 
! complex vector
!     dconjg(a(j,i))*v(j)=w(i)
!-----------------------------------------------------------------------

      subroutine qvaxzz (a,v,w,dim)

      implicit none

      integer(long)                                :: i, j
      integer(long), intent(in)                    :: dim
      complex(dop), dimension(dim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)     :: v
      complex(dop), dimension(dim), intent(out)    :: w

      do i = 1,dim
         w(i) = dconjg(a(1,i))*v(1)
         do j = 2,dim
            w(i) = w(i)+dconjg(a(j,i))*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvaxzz1
!
! Multiplication of the adjoint of a complex quadratic matrix with a 
! complex vector
!     dconjg(a(j,i))*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvaxzz1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: phdim, dim
      complex(dop), dimension(phdim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)       :: v
      complex(dop), dimension(dim), intent(out)      :: w

      do i = 1,dim
         w(i) = dconjg(a(1,i))*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+dconjg(a(j,i))*v(j)
         enddo
      enddo

      return
      end subroutine

!-----------------------------------------------------------------------
! Library subroutine qvtxzz1
!
! Multiplication of the adjoint of a complex quadratic matrix with a 
! complex vector
!     a(j,i)*v(j)=w(i)
!
! NB phdim is physical (leading) dimension, dim is used dimension 
!-----------------------------------------------------------------------

      subroutine qvtxzz1 (a,v,w,phdim,dim)

      implicit none

      integer(long)                                  :: i, j
      integer(long), intent(in)                      :: phdim, dim
      complex(dop), dimension(phdim,dim), intent(in) :: a
      complex(dop), dimension(dim), intent(in)       :: v
      complex(dop), dimension(dim), intent(out)      :: w

      do i = 1,dim
         w(i) = a(1,i)*v(1)
      enddo
      do j = 2,dim
         do i = 1,dim
            w(i) = w(i)+a(j,i)*v(j)
         enddo
      enddo

      return
      end subroutine

      end module xvlib


