!===============================================================================
module ctensor
!===============================================================================

use gridutil
use mlbase
implicit none
private

public :: create_ctensor,  &
          dispose_ctensor, &
          ctensor_from1d,  &
          ctensor_from2d,  &
          move_ctensor,    &
          reshape_ctensor, &
          ctensor_matmul_r

!-------------------------------------------------------------------------------
!> A type holding a complex tensor,
!! i.e. a multi-dimensional array of complex numbers.
!-------------------------------------------------------------------------------
type,public :: ctensor_t
!-------------------------------------------------------------------------------
   integer                  :: ndim = 0 !< number of dimensions (tensor order)
   integer,allocatable      :: dim(:)   !< extent of each dimension
   complex(dbl),allocatable :: data(:)  !< tensor data 
end type ctensor_t


!-------------------------------------------------------------------------------
type,public :: ctensor_tp
!-------------------------------------------------------------------------------
   type(ctensor_t),pointer :: p => null()
end type ctensor_tp


contains !======================================================================


!-------------------------------------------------------------------------------
!> Initializes a tensor with the given dimensions, with allocated but
!! uninitialized data.
!-------------------------------------------------------------------------------
subroutine create_ctensor(t,dim)
!-------------------------------------------------------------------------------
   type(ctensor_t),intent(out) :: t      !< the resulting tensor
   integer,intent(in)          :: dim(:) !< extent of each dimension
   t%ndim = size(dim)
   t%dim  = dim ! automatic allocation (F2003)
   allocate(t%data(product(dim)))
end subroutine create_ctensor



!-------------------------------------------------------------------------------
!> Destroy a tensor and all data in it.
!-------------------------------------------------------------------------------
subroutine dispose_ctensor(t)
!-------------------------------------------------------------------------------
   type(ctensor_t),intent(inout) :: t !< the tensor to be destroyed
   if (allocated(t%dim)) deallocate(t%dim)
   if (allocated(t%data)) deallocate(t%data)
   t%ndim = 0
end subroutine dispose_ctensor



!-------------------------------------------------------------------------------
!> Initialize a tensor from a given vector.
!-------------------------------------------------------------------------------
subroutine ctensor_from1d(t,a)
!-------------------------------------------------------------------------------
  use op1lib, only: cpvxz

   type(ctensor_t),intent(out) :: t    !< the resulting tensor
   complex(cdbl),intent(in)    :: a(:) !< the given vector
   call create_ctensor(t,shape(a))
   call cpvxz(a,t%data,size(a))
end subroutine ctensor_from1d



!-------------------------------------------------------------------------------
!> Initialize a tensor from a given matrix.
!-------------------------------------------------------------------------------
subroutine ctensor_from2d(t,a)
!-------------------------------------------------------------------------------
  use op1lib, only: cpvxz
   type(ctensor_t),intent(out) :: t      !< the resulting tensor
   complex(cdbl),intent(in)    :: a(:,:) !< the given matrix
   call create_ctensor(t,shape(a))
   call cpvxz(a,t%data,size(a))
end subroutine ctensor_from2d



!-------------------------------------------------------------------------------
!> Moves all data from one tensor to another.
!! The original tensor becomes unallocated.
!-------------------------------------------------------------------------------
subroutine move_ctensor(t,u)
!-------------------------------------------------------------------------------
   type(ctensor_t),intent(inout) :: t !< the original tensor
   type(ctensor_t),intent(out)   :: u !< the target tensor
   u%ndim = t%ndim
   call move_alloc(from=t%dim, to=u%dim)
   call move_alloc(from=t%data, to=u%data)
   t%ndim = 0
end subroutine move_ctensor



!-------------------------------------------------------------------------------
!> Change the shape of a tensor.
!-------------------------------------------------------------------------------
subroutine reshape_ctensor(t,dim)
!-------------------------------------------------------------------------------
   type(ctensor_t),intent(inout) :: t      !< the tensor
   integer,intent(in)            :: dim(:) !< the new shape
   if (product(dim) /= size(t%data)) call stop_code( &
      "ctensor", "reshape_ctensor", &
      "new shape incompatible with tensor size")
   t%ndim = size(dim)
   t%dim  = dim ! automatic reallocation (F2003)
end subroutine reshape_ctensor



!-------------------------------------------------------------------------------
!> Multiply a tensor \f$ T \f$ with a matrix \f$ A \f$ from the right
!! along the mode \f$ m \f$:
!! \f[
!! U_{i_1 \cdots k_m \cdots i_d} := \sum_{i_m=1}^{n_m} 
!! T_{i_1 \cdots i_m \cdots i_d} A_{i_m k_m}
!! \f]
!! If cjg is true, then \f$ A^* \f$ is used instead of \f$ A \f$.
!-------------------------------------------------------------------------------
function ctensor_matmul_r(T,A,m,cjg) result(U)
!-------------------------------------------------------------------------------
  use mtlib, only: mtaxzz,mttxzz
   type(ctensor_t)             :: U   !< the resulting tensor
   type(ctensor_t),intent(in)  :: T   !< the original tensor
   type(ctensor_t),intent(in)  :: A   !< the matrix (as an order-2 tensor)
   integer,intent(in)          :: m   !< the mode along which to multiply
   logical,intent(in),optional :: cjg !< whether to complex-conjugate A

   logical             :: lcjg
   integer,allocatable :: udim(:)
   integer             :: ijk(3)

   ! Check for correct input.
   if (A%ndim /= 2) call stop_code( &
      "ctensor", "ctensor_matmul_r", &
      "tensor used as matrix must have order two")
   if (T%dim(m) /= A%dim(1)) call stop_code( &
      "ctensor", "ctensor_matmul_r", &
      "tensor and matrix have incompatible size")

   if (present(cjg)) then
      lcjg = cjg
   else
      lcjg = .false.
   endif

   ! Allocate new tensor.
   udim = T%dim
   udim(m) = A%dim(2)
   call create_ctensor(U,udim)

   ! Do the multiplication.
   ijk = shape3d(T%dim,m)
   if (lcjg) then
      call mtaxzz(A%data, T%data, U%data, ijk(1), ijk(2), ijk(3), A%dim(2))
   else
      call mttxzz(A%data, T%data, U%data, ijk(1), ijk(2), ijk(3), A%dim(2))
   endif

   return
end function ctensor_matmul_r


!===============================================================================
end module ctensor
!===============================================================================
! vim: set ts=3 sw=3 expandtab:
