!===============================================================================
module ctucker
!===============================================================================

use ctensor
use gridutil
use mlbase
implicit none
private

public :: chosvd


!-------------------------------------------------------------------------------
!> A type holding a complex tensor in Tucker representation.
!-------------------------------------------------------------------------------
type,public :: cttensor_t
!-------------------------------------------------------------------------------
   type(ctensor_t)             :: core     !< core tensor
   type(ctensor_t),allocatable :: basis(:) !< basis vectors
end type cttensor_t


contains !======================================================================


!-------------------------------------------------------------------------------
!> Performs a higher-order singular value decomposition on a complex tensor.
!! The original tensor data is destroyed!
!-------------------------------------------------------------------------------
subroutine chosvd(tt,t,sqerr,mask,nbasis)
!-------------------------------------------------------------------------------
   type(cttensor_t),intent(out)  :: tt        !< resulting Tucker tensor
   type(ctensor_t),intent(inout) :: t         !< original tensor
   real(dbl),intent(inout)       :: sqerr     !< requested/actual squared-error
   logical,intent(in),optional   :: mask(:)   !< whether to perform HOSVD along each mode
   integer,intent(in),optional   :: nbasis(:) !< maximum number of basis vectors for each mode

   integer   :: nsvd,m,nbasis1
   logical   :: lregsvd
   real(dbl) :: sqerr1,sqerr1max

   ! How many SVDs to perform?
   if (present(mask)) then
      nsvd = count(mask)
   else
      nsvd = t%ndim
   endif
   ! In the 2-D case, the HOSVD is a regular SVD.
   lregsvd = (t%ndim==2 .and. nsvd==2)

   ! Distribute error evenly.
   if (lregsvd) then
      sqerr1max = sqerr
   else
      sqerr1max = sqerr/nsvd
   endif
   ! Initialize accumulated squared-error.
   sqerr = 0.d0

   ! Compute basis vectors for all requested modes.
   allocate(tt%basis(t%ndim))
   do m=1,t%ndim
      if (present(mask)) then
         if (.not.mask(m)) cycle
      endif
      if (present(nbasis)) then
         nbasis1 = nbasis(m)
         sqerr1 = 0.d0
      else
         nbasis1 = 0
         sqerr1 = sqerr1max
      endif
      call compute_basis(tt%basis(m),t,m,sqerr1,nbasis1)
      sqerr = sqerr + sqerr1
   enddo
   ! For regular SVD, error must only be counted once.
   if (lregsvd)  sqerr = 0.5d0*sqerr

   ! Compute core tensor.
   call move_ctensor(t,tt%core)
   do m=1,tt%core%ndim
      if (present(mask)) then
         if (.not.mask(m)) cycle
      endif
      t = ctensor_matmul_r(tt%core,tt%basis(m),m,cjg=.true.)
      call move_ctensor(t,tt%core)
   enddo

   return
end subroutine chosvd



!-------------------------------------------------------------------------------
!> Computes the dominant basis vectors of a tensor along the requested mode.
!-------------------------------------------------------------------------------
subroutine compute_basis(b,t,m,sqerr,nbasis)
!-------------------------------------------------------------------------------
   type(ctensor_t),intent(out) :: b      !< the resulting basis vectors
   type(ctensor_t),intent(in)  :: t      !< the original tensor
   integer,intent(in)          :: m      !< mode for which to compute basis
   real(dbl),intent(inout)     :: sqerr  !< requested/actual squared-error
   integer,intent(inout)       :: nbasis !< requested/actual number of basis vectors

   integer                   :: ijk(3),vdim,mdim,ndim,vndim,mindim,info,lwork,i
   complex(cdbl),allocatable :: tmat(:,:),work(:)
   real(dbl),allocatable     :: sigma(:),rwork(:)
   complex(cdbl)             :: cdum,lworkc(1)
   real(dbl)                 :: wsum

   ! Allocate and buld matricization of original tensor.
   ijk = shape3d(t%dim,m)
   vdim = ijk(1)
   mdim = ijk(2)
   ndim = ijk(3)
   vndim = vdim*ndim
   allocate(tmat(vndim,mdim))
   call matricize(t%data,tmat)
   
   ! Compute SVD and right singular vectors of matricized tensor TMAT.

   ! There is a (reported) bug in LAPACK (up to at least 3.6.0) which
   ! violates array bounds if vndim=1.  So we must treat it specially.
   if (vndim==1) then

      ! Here, TMAT has just one row.
      ! The singular value is the 2-norm of this row vector, and the sole
      ! right singular vector is obtained by normalizing TMAT.
      mindim = 1
      allocate(sigma(1))
      sigma(1) = sqrt(dble( dot_product(tmat(1,:), tmat(1,:)) ))
      do i=1,mdim
         tmat(1,i) = tmat(1,i) / sigma(1)
      enddo

   else ! use LAPACK

      mindim = min(mdim,vndim)
      allocate(sigma(mindim))
      allocate(rwork(5*mindim))
      call ZGESVD('N', 'O', vndim, mdim, tmat, vndim, sigma, &
                  cdum, 1, cdum, 1, lworkc, -1, rwork, info) ! workspace query
      lwork = int(lworkc(1))
      allocate(work(lwork))
      call ZGESVD('N', 'O', vndim, mdim, tmat, vndim, sigma, &
                  cdum, 1, cdum, 1, work, lwork, rwork, info)
      if (info /= 0) then
         write (*,*) "ERROR: ZGESVD returned info = ",info
         stop 1
      endif
      deallocate(work)
      deallocate(rwork)

   endif
   ! Now sigma contains the singular values in descending order.
   ! The rows of tmat contain the corresponding right singular vectors.

   if (nbasis==0) then
      ! Determine how many basis vectors to keep: sum up the singular
      ! values squared (starting with the smallest) until the limit
      ! is reached.
      wsum = 0.d0
      i = mindim
      do while (wsum <= sqerr .and. i > 0)
         wsum = wsum + max(0.d0,sigma(i))**2
         i = i-1
      enddo
      nbasis = i+1

   else
      ! nbasis is given by user, but we may not have enough singular
      ! vectors to fulfill the request...
      if (nbasis > mindim) call stop_code( &
         "ctucker_m", "compute_basis", "cannot create enough basis vectors" )
   endif

   ! Determine actual squared-error
   sqerr = 0.d0
   do i = mindim, nbasis+1, -1
      sqerr = sqerr + max(0.d0,sigma(i))**2
   enddo

   ! Copy the dominant basis vectors.
   call ctensor_from2d(b, transpose(tmat(1:nbasis,:)))

   ! Clean up.
   deallocate(work)
   deallocate(rwork)
   deallocate(sigma)
   deallocate(tmat)
   return

   contains

   subroutine matricize(x,xmat)
      complex(cdbl),intent(in)  :: x(vdim,mdim,ndim)
      complex(cdbl),intent(out) :: xmat(vdim,ndim,mdim)
      integer                   :: j,k
      if (ndim==1) then
         xmat(:,1,:) = x(:,:,1)
      else
         do k=1,ndim
            do j=1,mdim
               xmat(:,k,j) = x(:,j,k)
            enddo
         enddo
      endif
   end subroutine matricize

end subroutine compute_basis


!===============================================================================
end module ctucker
!===============================================================================
! vim: set ts=3 sw=3 expandtab:
