!===============================================================================
!> A module for tensors in hierarchical Tucker representation.
!===============================================================================
module htucker
!===============================================================================

use treemod
use ctensor
use ctucker
use mlbase
implicit none
private

public :: dispose_htensor, &
          hiersvd_ltr


!-------------------------------------------------------------------------------
!> A type holding a tensor in hierarchical Tucker representation.
!-------------------------------------------------------------------------------
type,public :: htensor_t
!-------------------------------------------------------------------------------
   type(tree_t),pointer         :: tree      !< the hierarchical structure
   integer,allocatable          :: nbasis(:) !< basis length at each node
   type(ctensor_tp),allocatable :: tt(:)     !< transfer tensors
end type htensor_t



contains !======================================================================


!-------------------------------------------------------------------------------
!> Destroy a HT-tensor and all data in it.
!-------------------------------------------------------------------------------
subroutine dispose_htensor(ht)
!-------------------------------------------------------------------------------
   type(htensor_t),intent(inout) :: ht !< the HT-tensor to be destroyed
   type(ctensor_t),pointer :: tp
   integer :: n
   if (allocated(ht%tt)) then
      do n=1,size(ht%tt)
         tp => ht%tt(n)%p
         if (associated(tp)) then
            call dispose_ctensor(tp)
            deallocate(tp)
         endif
      enddo
      deallocate(ht%tt)
   endif
   if (allocated(ht%nbasis)) deallocate(ht%nbasis)
   ht%tree => null()
end subroutine dispose_htensor



!-------------------------------------------------------------------------------
!> Computes the Hierarchical SVD of the given core tensor,
!! using leaves-to-root truncation.
!-------------------------------------------------------------------------------
subroutine hiersvd_ltr(ht,core,tree,sqerr,nbasis)
!-------------------------------------------------------------------------------
   type(htensor_t),intent(out)    :: ht        !< the resulting tensor in HT format
   type(ctensor_t),intent(inout)  :: core      !< the original core tensor
   type(tree_t),intent(in),target :: tree      !< the hierarchical structure
   real(dbl),intent(inout)        :: sqerr     !< requested/actual squared-error
   integer,intent(in),optional    :: nbasis(:) !< maximum basis length at the nodes

   integer                   :: m,num,nsvdleft,layer,nlnode,plen,c,nsvdl
   type(node_t),pointer      :: node
   real(dbl)                 :: sqerrleft,sqerrl
   type(node_tp),allocatable :: lnode(:)
   integer,allocatable       :: combdim(:)
   logical,allocatable       :: svdmask(:)
   integer,allocatable       :: nbasisl(:)
   type(cttensor_t)          :: tuckt

   ! Check input for consistency.
   if (tree%nleaf /= core%ndim) call stop_code( &
      "htucker_m", "hiersvd_ltr", &
      "number of leaves in tree /= order of core tensor")

   ! Prepare HT-tensor.
   ht%tree => tree
   allocate(ht%nbasis(tree%nnode))
   allocate(ht%tt(tree%nnode))

   ! Set basis length for leaf nodes.
   ht%nbasis(:) = 0
   do m=1,tree%nleaf
      num = tree%leaf(m)%p%num
      ht%nbasis(num) = core%dim(m)
   enddo

   ! Initialize error.
   sqerrleft = sqerr
   nsvdleft = tree%nnode - tree%nleaf - 1
   sqerr = 0.d0

   ! Loop over layers from bottom to top.
   ! The bottom-most layer has been treated elsewhere.
   ! The top-most layer (containng only root) is treated afterwards.
   do layer = tree%nlayer-2, 1, -1

      ! Mode-combine the core tensor according to the current layer.
      call get_layer(tree, layer, lnode)
      nlnode = size(lnode)
      allocate(combdim(nlnode))
      allocate(svdmask(nlnode))
      do m=1,nlnode
         node => lnode(m)%p
         if (node%isleaf) then
            combdim(m) = ht%nbasis(node%num)
            svdmask(m) = .false.
         else
            plen = 1
            do c=1,node%nchild
               plen = plen * ht%nbasis(node%child(c)%p%num)
            enddo
            combdim(m) = plen
            svdmask(m) = .true.
         endif
      enddo
      call reshape_ctensor(core,combdim)

      ! Set the target accuracy.
      if (layer==1 .and. nlnode==2) then
         ! regular SVD
         nsvdl = nsvdleft
         sqerrl = sqerrleft
      else
         ! HOSVD
         nsvdl = count(svdmask)
         sqerrl = (sqerrleft/nsvdleft)*nsvdl
      endif

      ! Do HOSVD on the core tensor.
      if (present(nbasis)) then
         ! Use the specified basis length.
         allocate(nbasisl(nlnode))
         do m=1,nlnode
            num = lnode(m)%p%num
            nbasisl(m) = nbasis(num)
         enddo
         call chosvd(tuckt,core,sqerrl,svdmask,nbasisl)
         deallocate(nbasisl)
      else
         ! Choose basis length according to desired error.
         call chosvd(tuckt,core,sqerrl,svdmask)
      endif

      ! Store the basis.
      do m=1,nlnode
         if (.not.svdmask(m)) cycle
         num = lnode(m)%p%num
         ht%nbasis(num) = tuckt%basis(m)%dim(2)
         allocate(ht%tt(num)%p)
         call move_ctensor(tuckt%basis(m), ht%tt(num)%p)
      enddo
      ! Update the core tensor.
      call move_ctensor(tuckt%core, core)

      ! Update the error.
      sqerr = sqerr + sqerrl
      nsvdleft = nsvdleft - nsvdl

      ! Clean up.
      deallocate(svdmask)
      deallocate(combdim)

   enddo

   ! For the root node, just store the remaining core tensor.
   ! But flatten it first.
   plen = size(core%data)
   call reshape_ctensor(core, [plen,1])
   node => tree%root
   num = node%num
   ht%nbasis(num) = 1
   allocate(ht%tt(num)%p)
   call move_ctensor(core, ht%tt(num)%p)

   ! Done.
   return

end subroutine hiersvd_ltr


!===============================================================================
end module htucker
!===============================================================================
! vim: set ts=3 sw=3 expandtab:
