!=======================================================================
module ilist
!=======================================================================
!
! This is a module for an allocatable list of integers. Really.
!
! Sometimes you want to return an integer list of unknown size from
! a function. So you have to allocate the result. But allocatable
! function results, or allocatable dummy arguments, are only allowed
! since Fortran2003. And they have strange semantics, namely being
! automatically deallocated after the result has been used. Which
! Fortran2003 alleviates by having automatic (re-)allocations of the
! RHS of an assignment. This is all very subtle, and hence best
! avoided. Plus, compiler bugs.
!
! Anyway, occassionally we want to have a list of lists of integers.
! As Fortran doesn't have arrays of pointers (or arrays of arrays),
! we need a defined type anyway.
!
! And while we're at it, this module also has some simple routines
! for treating lists of integers as sets, like union, subset, ...
!
! FO 05/2013
!
!=======================================================================

use mltypes, only: ilist_t, ilist_tp

implicit none
private

public :: new_ilist, dispose_ilist, &
          ilist_union, ilist_subset, ilist_equals


interface new_ilist
   module procedure new_ilist_with_size, new_ilist_from_array
end interface new_ilist

contains !==============================================================


!-----------------------------------------------------------------------
function new_ilist_with_size(n) result(ilist)
!-----------------------------------------------------------------------
   integer,intent(in)    :: n
   type(ilist_t),pointer :: ilist
   allocate(ilist)
   ilist%n = n
   allocate(ilist%i(n))
end function new_ilist_with_size


!-----------------------------------------------------------------------
function new_ilist_from_array(a) result(ilist)
!-----------------------------------------------------------------------
   integer,intent(in)    :: a(:)
   type(ilist_t),pointer :: ilist
   allocate(ilist)
   ilist%n = size(a)
   allocate(ilist%i(ilist%n))
   ilist%i = a
end function new_ilist_from_array


!-----------------------------------------------------------------------
subroutine dispose_ilist(ilist)
!-----------------------------------------------------------------------
   type(ilist_t),pointer :: ilist
   deallocate(ilist%i)
   deallocate(ilist)
end subroutine dispose_ilist


!-----------------------------------------------------------------------
function ilist_union(ilists) result(union)
!-----------------------------------------------------------------------
   type(ilist_tp),intent(in) :: ilists(:)
   type(ilist_t),pointer     :: union
   integer                   :: nl,l,nsum,n,i,j,k
   integer,allocatable       :: utmp(:)
   type(ilist_t),pointer     :: ilist
   nl = size(ilists)
   ! how many elements do all lists have, combined? => nsum
   nsum = 0
   do l=1,nl
      nsum = nsum + ilists(l)%p%n
   enddo
   allocate(utmp(nsum))
   n = 0
   ! loop over all lists
   do l=1,nl
      ilist => ilists(l)%p
      ! loop over all list elements
      elem: do k=1,ilist%n
         i = ilist%i(k)
         ! ignore duplicates
         do j=1,n
            if (i == utmp(j)) cycle elem
         enddo
         ! store element
         n = n+1
         utmp(n) = i
      enddo elem
   enddo
   ! construct result
   union => new_ilist(utmp(1:n))
   deallocate(utmp)
end function ilist_union


!-----------------------------------------------------------------------
pure function ilist_pos(ilist,i) result(pos)
!-----------------------------------------------------------------------
   type(ilist_t),intent(in) :: ilist
   integer,intent(in)       :: i
   integer                  :: pos
   integer                  :: k
   pos = 0
   do k=1,ilist%n
      if (i == ilist%i(k)) then
         pos = k
         return
      endif
   enddo
end function ilist_pos


!-----------------------------------------------------------------------
function ilist_subset(ilist1,ilist2) result(is_subset)
!-----------------------------------------------------------------------
! Checks whether the first list is a subset of the second list, i.e.
! all elements of the first must also be elements of the second list.
!-----------------------------------------------------------------------
   type(ilist_t),intent(in) :: ilist1,ilist2
   logical                  :: is_subset
   integer                  :: k,i,p
   is_subset = .false.
   ! loop over elements in first list
   do k=1,ilist1%n
      i = ilist1%i(k)
      ! see if it can be found in second list
      p = ilist_pos(ilist2,i)
      if (p==0) return
   enddo
   is_subset = .true.
end function ilist_subset


!-----------------------------------------------------------------------
function ilist_equals(ilist1,ilist2,perm) result(isequal)
!-----------------------------------------------------------------------
! Checks whether two lists have the same elements, but not necessarily
! in the same order. If _perm_ is requested, and _isequal_ is true, then
! on exit _perm_ contains the permutation of ilist2 that leads to ilist1,
! i.e. ilist1[k] == ilist2[perm[k]].
!-----------------------------------------------------------------------
   type(ilist_t),intent(in)       :: ilist1,ilist2
   type(ilist_t),pointer,optional :: perm
   logical                        :: isequal
   logical                        :: lperm
   integer                        :: n,k1,i,p
   isequal = .false.
   lperm = present(perm)
   ! lists must have same number of elements
   if (ilist1%n /= ilist2%n) then
      if (lperm) nullify(perm)
      return
   endif
   n = ilist1%n
   if (lperm) perm => new_ilist(n)
   ! loop over elements in first list
   do k1=1,n
      i = ilist1%i(k1)
      ! find element in second list
      p = ilist_pos(ilist2,i)
      if (p==0) then ! not found
         if (lperm) call dispose_ilist(perm)
         return
      endif
      ! store permutation
      if (lperm) perm%i(k1) = p
   enddo
   ! still here? lists must have same elements
   isequal = .true.
end function ilist_equals


end module ilist
! vim: set ts=3 sw=3 expandtab:
