module sort

! ----------------------------------------------------------------------
! Provides routines for sorting.
! (FO 03/2012)
! ----------------------------------------------------------------------
!
! The routines all have this interface:
!   sort(length, order, cmpfunc)
! where
!   length  = size of array to sort
!   order   = integer array which on return contains a permutation of
!             the integers 1..length in the sorted order; must have been
!             allocated by the caller
!   cmpfunc = function taking two integers i1,i2 and compares the array
!             elements at positions i1 and i2; must return integer:
!             < 0 : if 1st element "smaller" than 2nd element
!             > 0 : if 1st element "larger" than 2nd element
!             = 0 : if both elements are "equal"
!
! The meaning of "smaller", "larger", and "equal" is entirely up to
! cmpfunc, but should of course be consistent (i.e. transitive).
!
! You will notice that the array itself is *not* passed to either the
! sort routine or to cmpfunc. This is because Fortran95 has no way to
! pass arrays of arbitrary types. It is assumed that cmpfunc has access
! to the array in question, which e.g. can be achieved by defining the
! array and cmpfunc in the same module.
!
! Example code:
!
!   module example
!       integer,parameter :: adim
!       type(mytype), dimension(adim) :: a
!       contains
!       function acmp(i1,i2) result(c)
!           integer,intent(in) :: i1,i2
!           integer            :: c
!           ... ! compare a(i1) with a(i2) and set c accordingly
!       end function
!       subroutine test
!           integer, dimension(adim) :: order
!           ... ! assume that array a has been filled with data
!           call sort(adim, order, acmp)
!           ! e.g. print array in ascending order
!           do i=1,adim
!               call print_mytype(a(order(i)))
!           end do
!       end subroutine
!   end module
!
! ----------------------------------------------------------------------

    implicit none

    integer,parameter,private :: hybrid_size = 5

    contains

!    subroutine hybrid_sort(n, order, cmp)
!    ! hybrid sort algorithm:
!    !   1. sort small sub-arrays by selection sort
!    !   2. use non-recursive bottom-up merge sort
!    ! runtime = O(n log n)
!    ! needs integer work array of size n
!
!        integer, intent(in)  :: n
!        integer, intent(out) :: order(n)
!        interface
!            function cmp(i1,i2)
!                integer, intent(in) :: i1,i2
!                integer             :: cmp
!            end function cmp
!        end interface
!
!        integer :: i,j,c,s,p,t,m,m2,l,r
!        integer :: work(n)
!
!        ! initialize order as unsorted
!        do i=1,n
!            order(i) = i
!        end do
!
!        ! First, sort small sub-arrays (of size hybrid_size) by selection sort.
!        if (hybrid_size > 1) then
!            s = 1 ! start of sub-array
!            do while (s<n)
!                ! i is the start of the unsorted part of the sub-array
!                do i = s, min(s+hybrid_size-2,n-1)
!                    ! p is the position of the smallest element
!                    p = i
!                    ! go through remainder of array to find the minimum
!                    do j=i+1, min(s+hybrid_size-1,n)
!                        c = cmp(order(j),order(p))
!                        if (c < 0)  p = j
!                    end do
!                    ! if necessary, swap minimum element with position i
!                    if (p /= i) then
!                        t = order(i)
!                        order(i) = order(p)
!                        order(p) = t
!                    end if
!                end do
!                s = s + hybrid_size
!            end do
!        end if
!
!        ! Now switch to merge sort.
!        ! The algorithm makes several passes across the array.
!        ! In each pass, two adjacent sub-arrays (which are already sorted)
!        ! of size m are merged together into a sub-array of size 2*m.
!        m = hybrid_size ! initially, sorted sub-arrays are of this size
!        do while (m<n)
!            m2 = 2*m
!            s = 1 ! start of left sub-array
!            do while (s<=n-m+1)
!                l = s   ! pointer into left sub-array
!                r = s+m ! pointer into right sub-array
!                i = s   ! pointer into merged sub-array
!                do while (i < s+m2 .and. i <= n)
!                    if (l >= s+m) then
!                        ! left sub-array used up, take from right
!                        j = r
!                        r = r+1
!                    elseif (r >= s+m2 .or. r > n) then
!                        ! right sub-array used up, take from left
!                        j = l
!                        l = l+1
!                    else
!                        ! compare elements for left and right
!                        c = cmp(order(l),order(r))
!                        ! and take element from corresponding sub-array
!                        if (c <= 0) then
!                            j = l
!                            l = l+1
!                        else
!                            j = r
!                            r = r+1
!                        endif
!                    endif
!                    ! store the chosen element in the merged sub-array
!                    work(i) = order(j)
!                    i = i+1
!                end do
!                s = s+m2 ! advance to the next sub-arrays
!            end do
!            ! there might be a single left sub-array leftover;
!            ! just leave it alone
!            ! now copy the merged sub-arrays back
!            order(1:i-1) = work(1:i-1)
!            m = m2 ! double m for the next pass
!        end do
!
!    end subroutine hybrid_sort
!
!

    subroutine selection_sort(n, order, cmp)
    ! standard selection sort
    ! runtime = O(n**2)
    ! but no work space needed
    ! this is slower than hybrid_sort for n >= ~8

        integer, intent(in)  :: n
        integer, intent(out) :: order(n)
        interface
            function cmp(i1,i2)
                integer, intent(in) :: i1,i2
                integer             :: cmp
            end function cmp
        end interface

        integer :: i,j,p,c,t

        ! intitialize order as unsorted
        do i=1,n
            order(i) = i
        end do

        ! i is the start of the unsorted part of the array
        do i=1,n-1
            ! p is the position of the smallest element
            p = i
            ! go through remainder of array to find the minimum
            do j=i+1,n
                c = cmp(order(j),order(p))
                if (c < 0)  p = j
            end do
            ! if necessary, swap minimum element with position i
            if (p /= i) then
                t = order(i)
                order(i) = order(p)
                order(p) = t
            end if
        end do

    end subroutine selection_sort

end module sort
