!===============================================================================
!> A module for a bare tree.
!! Such trees contain no data except for the structure of the tree.
!===============================================================================
module treemod
!===============================================================================

implicit none
private

public :: make_leaf, &
          make_node, &
          make_tree, &
          dispose_tree, &
          get_layer


!-------------------------------------------------------------------------------
!> A type holding a bare tree.
!-------------------------------------------------------------------------------
type,public :: tree_t
!-------------------------------------------------------------------------------
   type(node_t),pointer  :: root => null()         !< pointer to root node
   integer               :: nnode                  !< total number of nodes in tree
   integer               :: nleaf                  !< total number of leaves in tree
   integer               :: nlayer                 !< number of layers in tree
   type(node_tp),pointer :: preorder(:) => null()  !< nodes in pre-order
   type(node_tp),pointer :: postorder(:) => null() !< nodes in post-order
   type(node_tp),pointer :: leaf(:) => null()      !< leaf nodes
end type tree_t



!-------------------------------------------------------------------------------
!> A type holding the bare essential of a tree node.
!-------------------------------------------------------------------------------
type,public :: node_t
!-------------------------------------------------------------------------------
   !--- Properties of the node itself ---
   logical               :: isleaf             !< whether this node is a leaf
   integer               :: nchild             !< number of children
   type(node_tp),pointer :: child(:) => null() !< pointers to the children
   type(node_t),pointer  :: parent => null()   !< pointer to the parent
   !--- Properties relating this node to the tree ---
   integer               :: num                !< number of node inside tree
   integer               :: layer              !< layer that this node is on (0 ... nlayer-1)
   type(tree_t),pointer  :: tree => null()     !< pointer to tree that this node belongs to
end type node_t



!-------------------------------------------------------------------------------
type,public :: node_tp
!-------------------------------------------------------------------------------
   type(node_t),pointer  :: p => null()
end type node_tp


contains !======================================================================


!-------------------------------------------------------------------------------
!> Create a leaf.
!-------------------------------------------------------------------------------
function make_leaf() result(leaf)
!-------------------------------------------------------------------------------
   type(node_t),pointer :: leaf !< the resulting leaf
   allocate(leaf)
   ! Mark node as leaf.
   leaf%isleaf = .true.
   ! A leaf has no children.
   leaf%nchild = 0
end function make_leaf



!-------------------------------------------------------------------------------
!> Create a node from its children.
!-------------------------------------------------------------------------------
function make_node(child) result(node)
!-------------------------------------------------------------------------------
   type(node_t),pointer        :: node      !< the resulting node
   type(node_tp),intent(inout) :: child(:)  !< the children of the new node
   integer                     :: nchild,c
   allocate(node)
   ! Mark node as non-leaf.
   node%isleaf = .false.
   ! Link children into node, and set their parent.
   nchild = size(child)
   node%nchild = nchild
   allocate(node%child(nchild))
   do c=1,nchild
      node%child(c)%p => child(c)%p
      child(c)%p%parent => node
   enddo
end function make_node



!-------------------------------------------------------------------------------
!> Create a tree from a root node and its children.
!-------------------------------------------------------------------------------
function make_tree(root) result(tree)
!-------------------------------------------------------------------------------
   type(tree_t),pointer              :: tree !< the resulting tree
   type(node_t),target,intent(inout) :: root !< the root node of the tree
   integer                           :: nnode,nleaf,nlayer,idx,n,l
   type(node_t),pointer              :: node

   allocate(tree)
   ! Set the root node.
   tree%root => root
   ! Count the number of nodes etc. in the tree.
   call count_nodes(root,nnode,nleaf,nlayer)
   tree%nnode  = nnode
   tree%nleaf  = nleaf
   tree%nlayer = nlayer
   ! Record the pre-/post-order sequences of the tree.
   allocate(tree%preorder(nnode))
   idx = 1
   call set_preorder(root,tree%preorder,idx)
   allocate(tree%postorder(nnode))
   idx = 1
   call set_postorder(root,tree%postorder,idx)
   ! Record leaves.
   allocate(tree%leaf(nleaf))
   l = 1
   do n=1,nnode
      node => tree%preorder(n)%p
      if (node%isleaf) then
         tree%leaf(l)%p => node
         l = l+1
      endif
   enddo
   ! Set tree-related data in nodes.
   do n=1,nnode
      node => tree%preorder(n)%p
      node%num = n
      node%tree => tree
   enddo
   call set_layer(root,0)
end function make_tree



!-------------------------------------------------------------------------------
!> Count the total number of nodes, leaves, and layers
!! below a certain node, itself included.
!-------------------------------------------------------------------------------
recursive subroutine count_nodes(node,nnode,nleaf,nlayer)
!-------------------------------------------------------------------------------
   type(node_t),intent(in) :: node
   integer,intent(out)     :: nnode,nleaf,nlayer
   integer                 :: c,chnode,chleaf,chlayer
   type(node_t),pointer    :: child
   if (node%isleaf) then
      nnode = 1
      nleaf = 1
      nlayer = 1
   else
      nnode = 1
      nleaf = 0
      nlayer = 1
      do c=1,node%nchild
         child => node%child(c)%p
         call count_nodes(child,chnode,chleaf,chlayer)
         nnode = nnode + chnode
         nleaf = nleaf + chleaf
         nlayer = max(nlayer,chlayer+1)
      enddo
   endif
end subroutine count_nodes



!-------------------------------------------------------------------------------
recursive subroutine set_preorder(node,seq,idx)
!-------------------------------------------------------------------------------
   type(node_t),target,intent(in) :: node
   type(node_tp),intent(inout)    :: seq(:)
   integer,intent(inout)          :: idx
   type(node_t),pointer           :: child
   integer                        :: c
   ! First record this node.
   seq(idx)%p => node
   idx = idx + 1
   ! Then record all children, if any.
   if (.not.node%isleaf) then
      do c=1,node%nchild
         child => node%child(c)%p
         call set_preorder(child,seq,idx)
      enddo
   endif
end subroutine set_preorder



!-------------------------------------------------------------------------------
recursive subroutine set_postorder(node,seq,idx)
!-------------------------------------------------------------------------------
   type(node_t),target,intent(in) :: node
   type(node_tp),intent(inout)    :: seq(:)
   integer,intent(inout)          :: idx
   type(node_t),pointer           :: child
   integer                        :: c
   ! First record all children, if any.
   if (.not.node%isleaf) then
      do c=1,node%nchild
         child => node%child(c)%p
         call set_postorder(child,seq,idx)
      enddo
   endif
   ! Then record this node.
   seq(idx)%p => node
   idx = idx + 1
end subroutine set_postorder



!-------------------------------------------------------------------------------
recursive subroutine set_layer(node,layer)
!-------------------------------------------------------------------------------
   type(node_t),intent(inout) :: node
   integer,intent(in)         :: layer
   type(node_t),pointer       :: child
   integer                    :: c
   ! Set layer for this node.
   node%layer = layer
   ! Then set layer for the children, if any.
   if (.not.node%isleaf) then
      do c=1,node%nchild
         child => node%child(c)%p
         call set_layer(child,layer+1)
      enddo
   endif
end subroutine set_layer



!-------------------------------------------------------------------------------
!> Destroy a tree and all its children, freeing all used resources.
!-------------------------------------------------------------------------------
subroutine dispose_tree(tree)
!-------------------------------------------------------------------------------
   type(tree_t),pointer :: tree !< the tree to be destroyed
   deallocate(tree%leaf)
   deallocate(tree%postorder)
   deallocate(tree%preorder)
   call dispose_node(tree%root)
   deallocate(tree)
end subroutine dispose_tree



!-------------------------------------------------------------------------------
recursive subroutine dispose_node(node)
!-------------------------------------------------------------------------------
   type(node_t),pointer :: node
   integer              :: c
   ! Destroy all children first.
   if (.not.node%isleaf) then
      do c=1,node%nchild
         call dispose_node(node%child(c)%p)
      enddo
      deallocate(node%child)
   endif
   deallocate(node)
end subroutine dispose_node



!-------------------------------------------------------------------------------
!> Return a list of nodes on a specific layer of a tree.
!! The list will include leaves on upper layers as well.
!-------------------------------------------------------------------------------
subroutine get_layer(tree,layer,nodelist)
!-------------------------------------------------------------------------------
   type(tree_t),intent(in)               :: tree        !< the tree in question
   integer,intent(in)                    :: layer       !< the requested layer
   type(node_tp),allocatable,intent(out) :: nodelist(:) !< the resulting list of nodes
   integer                               :: n,nnode
   integer,allocatable                   :: nodenum(:)
   type(node_t),pointer                  :: node

   allocate(nodenum(tree%nnode))
   nnode = 0
   do n=1,tree%nnode
      node => tree%preorder(n)%p
      if (node%layer > layer) cycle
      if (node%layer < layer .and. .not.node%isleaf) cycle
      nnode = nnode+1
      nodenum(nnode) = n
   enddo
   ! Return list of nodes.
   allocate(nodelist(nnode))
   do n=1,nnode
      nodelist(n)%p => tree%preorder(nodenum(n))%p
   enddo
   ! Clean up.
   deallocate(nodenum)
end subroutine get_layer


!===============================================================================
end module treemod
!===============================================================================
! vim: set ts=3 sw=3 expandtab:
