module integrator

   use mlbase
   use mlrklib
   use mlsillib
   use mlcsillib

   use mltypes
   use mlpostproc

   implicit none
   private

   integer,parameter,public :: inttyp_abm = 1  ! Adams-Bashforth-Moulton
   integer,parameter,public :: inttyp_bs  = 2  ! Bulirsch-Stoer
   integer,parameter,public :: inttyp_sil = 3  ! Short Iterative Lanczos
   integer,parameter,public :: inttyp_rk  = 4  ! Runge-Kutta
   integer,parameter,public :: inttyp_dav = 6  ! Davidson
   integer,parameter,public :: inttyp_csil= 7  ! Short Iterative Lanczos-Arnoldi

   public :: new_integrator,     &
             dispose_integrator, &
             drive_integrator,   &
             show_integrator,    &
             stats_integrator,   &
!            int_abm_init,       &
!            int_bs_init,        &
             int_sil_init,       &
             int_csil_init,      &
!            int_dav_init,       &
             int_rk_init

!   type,public :: integrator_t
!   ! This type stores settings and internal data for the various
!   ! integration routines.  Such data should not be global, because
!   ! different objects in the program can then be propagated with
!   ! individual integrator settings.
!      integer :: inttyp   ! integrator type (see above)
!      logical :: llog     ! .true. if logging is enabled
!      integer :: ilog     ! log channel
!      logical :: lpost    ! .true. if post-processing is enabled
!      integer :: ipost    ! cf procedure int_postprocess
!      ! pointers to integrator-specific data, only one of these will be set
!!     type(int_abm_t),pointer :: abm => null()
!!     type(int_bs_t),pointer  :: bs  => null()
!      type(int_rk_t),pointer  :: rk  => null()
!      type(int_sil_t),pointer :: sil => null()
!      type(int_csil_t),pointer:: csil=> null()
!!     type(int_dav_t),pointer :: dav => null()
!   end type integrator_t


   contains


   function new_integrator(typ,llog,ilog,lpost,ipost,lrhonrm) result (ni)
   ! Initializes a new integrator object.
   ! The caller still needs to initialize the specific integrator settings.
      implicit none
      type(integrator_t),pointer :: ni
      integer,intent(in)         :: typ
      logical,intent(in)         :: llog,lpost ! whether to do logging/post-processing
      integer,intent(in)         :: ilog,ipost ! where to log/how to post-process
      logical,intent(in)         :: lrhonrm    ! whether to use rho (density) weighted error norm
      allocate(ni)
      ni%inttyp = typ
      ni%llog   = llog
      ni%ilog   = ilog
      ni%lpost  = lpost
      ni%ipost  = ipost
      ni%lrhon  = lrhonrm
      select case (typ)
!        case (inttyp_abm)
!           allocate(ni%abm)
!        case (inttyp_bs)
!           allocate(ni%bs)
         case (inttyp_rk)
            allocate(ni%rk)
         case (inttyp_sil)
            allocate(ni%sil)
         case (inttyp_csil)
            allocate(ni%csil)
!        case (inttyp_dav)
!           allocate(ni%dav)
         case default
            call stop_code('integrator_m', 'new_integrator', 'unknown integrator type')
      end select
   end function new_integrator


   subroutine dispose_integrator(i)
   ! Deallocates data stored in the integrator object, so that the
   ! integrator object itself can be disposed.
      implicit none
      type(integrator_t),intent(inout) :: i
      select case (i%inttyp)
         case (inttyp_rk)
            call int_rk_dispose(i%rk)
            deallocate(i%rk)
         case (inttyp_sil)
            call int_sil_dispose(i%sil)
            deallocate(i%sil)
         case (inttyp_csil)
            call int_csil_dispose(i%csil)
            deallocate(i%csil)
         case default
            call stop_code('integrator_m', 'dispose_integrator', 'unknown integrator type')
      end select
   end subroutine dispose_integrator


   subroutine show_integrator(i,desc)
   ! Returns a description of the integrator and possible post-processor in a string
      implicit none
      type(integrator_t),intent(in) :: i
      character(len=*),intent(out)  :: desc
      character(len=80)             :: idesc, ppdesc
      select case (i%inttyp)
         case (inttyp_rk)
            call int_rk_show(i%rk,idesc)
         case (inttyp_sil)
            call int_sil_show(i%sil,idesc)
         case (inttyp_csil)
            call int_csil_show(i%csil,idesc)
         case default
            call stop_code('integrator_m', 'show_integrator', 'unknown integrator type')
      end select
      if (i%lpost) then
         call int_postprocess_show(i%ipost, ppdesc)
         write (desc,'(4a)') trim(idesc), ' (', trim(ppdesc), ')'
      else
         desc = idesc
      endif
   end subroutine show_integrator


   subroutine stats_integrator(itg,iout)
   ! Writes statistics about the integrator to channel iout
      implicit none
      type(integrator_t),intent(in) :: itg
      integer,intent(in)            :: iout
      select case (itg%inttyp)
         case (inttyp_rk)
            call int_rk_stats(itg%rk,iout)
         case (inttyp_sil)
            call int_sil_stats(itg%sil,iout)
         case (inttyp_csil)
            call int_csil_stats(itg%csil,iout)
         case default
            call stop_code('integrator_m', 'stats_integrator', 'unknown integrator type')
      end select
   end subroutine stats_integrator


   subroutine drive_integrator(y,f,dt,t,restart,itg,funk,dmat,dim)
   ! Integrates the system of differential equations  y' = funk(t,y)
   ! over the interval [t,t+dt] with the provided integrator.
      implicit none
      complex(cdbl),intent(in)    :: dmat(:,:) ! density matrix of current node. Dieter Apr2021
      complex(cdbl),intent(inout) :: y(:)      ! dynamic variables (in:initial, out:final)
      complex(cdbl),intent(inout) :: f(:)      ! initial time derivative of dynamic variables
      real(dbl),intent(in)        :: dt        ! integration interval
      real(dbl),intent(in)        :: t         ! initial time
      integer,intent(in)          :: dim       ! number od SPFs, cnode%navecs
      logical,intent(in)          :: restart   ! .true. if integration should start over
      type(integrator_t),intent(inout) :: itg  ! integrator object
      interface
         subroutine funk(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine funk
      end interface
      ! here we just dispatch to the driver for each integrator type
      select case (itg%inttyp)
         case (inttyp_rk)
            call int_rk_drv(y,f,dt,t,restart,itg%rk,funk,itg%llog,itg%ilog,itg%lpost,itg%ipost,itg%lrhon,dmat,dim)
         case (inttyp_sil)
            call int_sil_drv(y,f,dt,t,restart,itg%sil,funk,itg%llog,itg%ilog,itg%lpost,itg%ipost)
         case (inttyp_csil)
            call int_csil_drv(y,f,dt,t,restart,itg%csil,funk,itg%llog,itg%ilog,itg%lpost,itg%ipost)
         case default
            call stop_code('integrator_m', 'drive_integrator', &
               'unknown integrator type')
      end select
   end subroutine drive_integrator


end module integrator

