module mlcsillib
!----------------------------------------------------------------------
! F95 module for the Short Iterative Lanczos-Arnoldi (CSIL) integrator.
!
! FO 12/2012
!----------------------------------------------------------------------

   use global
   use mlbase
   use logging

   use mltypes
   use mlpostproc

   implicit none
   private

   public :: int_csil_init, int_csil_drv, int_csil_dispose, int_csil_show, int_csil_stats

!   type,public :: int_csil_t
!      integer       :: maxOrder    ! maximum order allowed
!      real(dbl)     :: accuracy    ! target accuracy
!      complex(cdbl) :: PreFactor   ! function f is actually f = prefactor*H*y
!      ! data kept between calls
!      integer       :: prevOrder   ! order used in previous step
!      real(dbl)     :: prevStep    ! interval integrated in previous step
!      real(dbl)     :: prevError   ! error measured in previous step
!      ! statistical data
!      integer       :: nsteps      ! total number of steps
!      integer       :: ngoodst     ! number of good steps
!      integer       :: nbadst      ! number of bad steps (order too low)
!      integer       :: nfncns      ! number of RHS function calls
!      integer       :: realOrder   ! largest order actually used
!      ! auxiliary vectors
!      complex(cdbl),allocatable :: Krylov(:,:)      ! storage for the Krylov space
!      complex(cdbl),allocatable :: Hessenberg(:,:)
!      complex(cdbl),allocatable :: EigVec(:,:)
!      complex(cdbl),allocatable :: AuxMat(:,:)
!      complex(cdbl),allocatable :: EigVal(:)
!      complex(cdbl),allocatable :: Aux1(:)
!      complex(cdbl),allocatable :: Aux2(:)
!      complex(cdbl),allocatable :: Aux3(:)
!      complex(cdbl),allocatable :: OldAux1(:)
!   end type int_csil_t

!   real(dbl),parameter :: macheps = epsilon(ONER)  ! F95 std.


   contains


   subroutine int_csil_init(csil,ord,acc,pre)
   ! Initialize the CSIL integrator object.
      implicit none
      type(int_csil_t),intent(inout) :: csil  ! object to initialize
      integer,intent(in)             :: ord   ! maximum order
      real(dbl),intent(in)           :: acc   ! target accuracy
      complex(cdbl),intent(in)       :: pre   ! prefactor

      if (ord < 2) then
         call stop_code('csillib_m', 'int_csil_init', 'order too small')
      endif
      csil%maxorder  = ord
      csil%accuracy  = acc
      csil%prefactor = pre
      csil%prevorder = 0
      csil%prevstep  = ZEROR
      csil%preverror = ZEROR
      csil%nsteps    = 0
      csil%ngoodst   = 0
      csil%nbadst    = 0
      csil%nfncns    = 0
      csil%realorder = 0
      allocate(csil%Hessenberg(0:ord,0:ord))
      allocate(csil%EigVec(0:ord,0:ord))
      allocate(csil%AuxMat(0:ord,0:ord))
      allocate(csil%EigVal(0:ord))
      allocate(csil%Aux1(0:ord))
      allocate(csil%Aux2(0:ord))
      allocate(csil%Aux3(0:ord))
      allocate(csil%OldAux1(0:ord))
   end subroutine int_csil_init


   subroutine int_csil_dispose(csil)
   ! Free memory used by the CSIL integrator object.
      implicit none
      type(int_csil_t),intent(inout) :: csil
      if (allocated(csil%Krylov)) deallocate(csil%Krylov)
      deallocate(csil%OldAux1)
      deallocate(csil%Aux3)
      deallocate(csil%Aux2)
      deallocate(csil%Aux1)
      deallocate(csil%EigVal)
      deallocate(csil%AuxMat)
      deallocate(csil%EigVec)
      deallocate(csil%Hessenberg)
   end subroutine int_csil_dispose


   subroutine int_csil_show(csil,desc)
   ! Return a textual description of the CSIL integrator.
      implicit none
      type(int_csil_t),intent(in)  :: csil
      character(len=*),intent(out) :: desc
      write (desc, '(a,i0,a,es8.2)') &
         'Short Iterative Lanczos-Arnoldi, max.order = ', csil%maxorder, &
         ', accuracy = ', csil%accuracy
   end subroutine int_csil_show


   subroutine int_csil_stats(csil,iout)
   ! Write statistical information about the CSIL integrator to channel iout.
      implicit none
      type(int_csil_t),intent(in) :: csil
      integer,intent(in)          :: iout
      character(len=80)           :: desc
      call int_csil_show(csil,desc)
      write (iout,'(a)') trim(desc)
      write (iout,'(2x,a,i9)') 'Total number of steps: ',csil%nsteps
      write (iout,'(2x,a,i9)') 'Number of good steps:  ',csil%ngoodst
      write (iout,'(2x,a,i9)') 'Number of bad steps:   ',csil%nbadst
      write (iout,'(2x,a,i9)') 'Maximum order used:    ',csil%realorder
   end subroutine int_csil_stats


   subroutine int_csil_drv(y,f,dt,t,restart,csil,funk,llog,ilog,lpost,ipost)
   ! Drives the CSIL integration of the equation dy/dt = funk(y).
   ! funk must be of the form: funk(y) = p*H*y
   !   where H is a linear operator
   !   and p is a prefactor (usually -I)
   ! Replaces the RUNSIL routine.
      implicit none
      complex(cdbl),intent(inout)   :: y(:)    ! initial/final values of dynamic variables
      complex(cdbl),intent(inout)   :: f(:)    ! initially, f=funk(y) [we assume this has already been done]
      real(dbl),intent(in)          :: dt      ! time interval to integrate
      real(dbl),intent(in)          :: t       ! current time
      logical,intent(in)            :: restart ! if .true., reuse previous Krylov space
      logical,intent(in)            :: llog    ! if .true., log information about each step
      integer,intent(in)            :: ilog    ! log channel
      logical,intent(in)            :: lpost   ! if .true., call post-processor after each CSIL step
      integer,intent(in)            :: ipost   ! cf procedure int_postprocess
      type(int_csil_t),intent(inout):: csil    ! object with integrator settings
      interface
         subroutine funk(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine funk
      end interface
      real(dbl)         :: inttime ! time integrated so far
      real(dbl)         :: hmax    ! maximum size of next CSIL step
      real(dbl)         :: hact    ! actual size of last CSIL step
      integer           :: actord  ! actual order used in the CSIL step
      integer           :: ierr    ! error code
      character(len=80) :: msg     ! text buffer

      ! allocate Krylov space if not done previously
      if (.not.allocated(csil%Krylov)) then
         allocate(csil%Krylov(size(y), 2:csil%maxorder))
      endif

      if (llog) then
         write (msg,'(a7,9x,a6,6x,a13,11x,a5,11x,a9)') &
            '#  Step', 'Order', 'Stepsize [fs]', 'Error', 'Time [fs]'
         call write_log(ilog,msg)
      endif

      ! do one CSIL step, actual stepsize returned in hact
      inttime = ZEROR
 10   hmax = dt - inttime
      call csilstepml( csil, &
         funk, size(y), y, f, t+inttime, hmax, restart, &
         hact, actord, ierr, llog, ilog)

      ! check for errorrs
      if (ierr.ne.0) then
         call CSILErrorMsg(ierr, msg)
         call stop_code('csillib_m', 'int_csil_drv', msg)
      endif

      ! update step counters and max. used order
      if (deql(hact,hmax)) then
         csil%ngoodst = csil%ngoodst + 1
      else
         csil%nbadst = csil%nbadst + 1
      endif
      csil%nsteps = csil%nsteps + 1
      csil%realorder = max(csil%realorder, actord)

      ! call post-processor
      if (lpost) call int_postprocess(y, size(y), ipost)

      ! are we finished?
      inttime = inttime + hact
      if (.not.dgeq(inttime,dt)) then
         ! no, we need more steps;
         ! evaluate RHS again
         call funk(t+inttime,y,f)
         csil%nfncns = csil%nfncns + 1
         goto 10
      endif

      return
   end subroutine int_csil_drv



! **********************************************************************
! *                                                                    *
! *                SUBROUTINE CSILSTEP                                 *
! *                                                                    *
! * Integrates a system of complex linear first order differential     *
! * equations with time-independent Hamiltonian (not necessarily       *
! * hermitian) employing the short iterative Lanczos-Arnoldi method.   *
! * The routine runs with both variable step size and order. (First    *
! * the order is increased to achieve the desired accuracy. If this    *
! * failes within the maximum order, the stepsize is reduced.)         *
! * CSILStep makes only one single integration step, so it has to be   *
! * imbedded into a loop that calls CSILStep until the desired time    *
! * interval is integrated. The ODE has the form dPsi/dt = -i H|Psi> = *
! * Func(Time,Psi,DtPsi,...) =: DtPsi.                                 *
! * All computations are performed with double precision.              *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector, i. *
! *              e. -i H|Psi>.                                         *
! *   PsiDim     Length of Psi and DtPsi vectors.                      *
! *   IntPeriod: Lenght of time interval to be integrated.             *
! *   Restart:   Flag for restarting the integrator. If true, the      *
! *              Krylov space is built up before propagation, else the *
! *              old Krylov vectors are used.                          *
! *   OldErrCri: If true, the old (i.e. standard) error criterion is   *
! *              used, otherwise an improved one is taken.             *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   DtPsi:     Normalised first Krylov vector, i. e. (H-<H>)|Psi>.   *
! *   Stepsize:  Time interval that actually has been integrated (can  *
! *              be lower than IntPeriod).                             *
! *   ActualOrder: The order that has actually been used (may be less  *
! *              than IntOrder).                                       *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              2: stepsize underflow,                                *
! *              3: can't diagonalise matrix,                          *
! *              4: stepsize too large for restart,                    *
! *              5: can't solve linear equation.                       *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi. Called *
! *              as Func(Time,Psi,DtPsi).                              *
! *   eigchess:  Computes the eigenvalues and -vectors of a complex    *
! *              upper Hessenberg matrix. Called as:                   *
! *              eigchess(LeadingDimension,Dimension,Matrix,           *
! *              macheps,Eigenvalues,Eigenvectors,ErrorCode).          *
! *   zgesv:     Solves the complex linear equation Ax = b. Called as: *
! *              zgesv(Dimension,A,LeadingDimensionA,PivotIndices,b,   *
! *              LeadingDimensionB,ErrorCode). b is overwriten with x  *
! *              on exit.                                              *
! *                                                                    *
! * V6.0 MB                                                            *
! * V7.0 GW addition of CData,RData,IData,LData arrays                 *
! * V7.0 MB improved error criterion                                   *
! * V8.0 MB routine can handle H|psi> and -i H|psi>                    *
! * V8.5 FO convert to F95 interface; remove StdForm                   *
! *                                                                    *
! **********************************************************************

   subroutine csilstepml( csil, &
      Func, PsiDim, Psi, DtPsi, Time, IntPeriod, Restart, &
      Stepsize, ActualOrder, ErrorCode, llog, ilog)

      implicit none
      type(int_csil_t),intent(inout) :: csil
      interface
         subroutine Func(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine Func
      end interface
      integer,intent(in)          :: PsiDim, ilog
      complex(cdbl),intent(inout) :: Psi(PsiDim), DtPsi(PsiDim)
      real(dbl),intent(in)        :: Time, IntPeriod
      logical,intent(in)          :: Restart, llog
      real(dbl),intent(out)       :: StepSize
      integer,intent(out)         :: ActualOrder, ErrorCode

      real(dbl),parameter :: RelativeMinStep = 1.0d-9
      integer,parameter   :: MinOrder = 2
      logical,parameter   :: OldErrCri = .true. ! TODO: make option

      integer       :: Order,NextOrder,P
      real(dbl)     :: MinStepsize,Error,FormerError,SubDiag,Norm,Inverse,PsiEps
      complex(cdbl) :: CTemp,CInverse
      integer,allocatable ::Auxi(:)

! --- INITIALIZE VARIABLES ---

      Stepsize = IntPeriod
      MinStepsize = Abs(RelativeMinStep*IntPeriod)
      ErrorCode = 0
      ActualOrder = 1
      PsiEps = sqrt(dble(PsiDim))*macheps

! --- RETURN RESULT IF DIFFERENTIAL EQUATION IS NOT A SYSTEM ---

      If (PsiDim .Eq. 1) Then
         Order = 1
         Error = 0.0D0
         Call WriteStep(Error)
         Psi(1) = Exp(StepSize*DtPsi(1)/Psi(1))*Psi(1)
         Return
      EndIf

! --- COMPUTE NORM OF INITIAL WAVEFUNCTION ---

      Norm = Sqrt(Dble(dot_product(Psi,Psi)))

! --- NORMALISE INITIAL WAVEFUNCTIONS ---

      Inverse = 1.0D0/Norm
      If (Restart) Then
         Psi   = Inverse*Psi
         DtPsi = Inverse*DtPsi
      Else
         Psi = Inverse*Psi

! --- SKIP CALCULATION OF KRYLOV VECTORS IF DESIRED (i.e. unless Restart) ---

         If (Abs(Stepsize) .GT. Abs(csil%prevStep)+MinStepsize) Then
            ErrorCode = 4
            Return
         EndIf
         Order = csil%prevOrder
         Error = Abs(csil%prevError*(Stepsize/csil%prevStep)**Order)
         Call WriteStep(Error)
         Goto 200
      EndIf

! --- FURTHER INITIALIZE VARIABLES ---

      csil%prevError = 0.0D0
      csil%prevOrder = 0
      csil%prevStep  = 0.0D0

      FormerError = 1.0D0
      Order = 0
      csil%OldAux1 = CZERO
      csil%Hessenberg = CZERO

! --- BUILD UP THE KRYLOV SPACE ---

 100  Continue

! --- CHECK WHETHER STEPSIZE IS TOO SMALL ---

      If (Abs(Stepsize) .LT. Abs(MinStepsize)) Then
         ErrorCode = 2
         Return
      EndIf

! --- LOOP OVER ALL LOWER KRYLOV VECTORS ---

      NextOrder = Order+1
      Do P = 0,Order

!    --- COMPUTE ELEMENT OF HESSENBERG MATRIX ---

! Note that the first two Krylov vectors are stored in "Psi" and
! "DtPsi", respectively, whereas the higher vectors reside in the
! corresponding columns of the "Krylov" matrix.

         If (P .Eq. 0) Then
            If (Order .Eq. 0) Then
               CTemp = dot_product(Psi, DtPsi)
            Else
               CTemp = dot_product(Psi, csil%Krylov(:,NextOrder))
            EndIf
         ElseIf (P .Eq. 1) Then
            CTemp = dot_product(DtPsi, csil%Krylov(:,NextOrder))
         Else
            CTemp = dot_product(csil%Krylov(:,P), csil%Krylov(:,NextOrder))
         EndIf
         csil%Hessenberg(P,Order) = CTemp

!    --- ORTHOGONALISE CURRENT KRYLOV VECTOR ON LOWER ONES ---

         If (P .Eq. 0) Then
            If (Order .Eq. 0) Then
               DtPsi = DtPsi - CTemp*Psi
            Else
               csil%Krylov(:,NextOrder) = csil%Krylov(:,NextOrder) - CTemp*Psi
            EndIf
         ElseIf (P .Eq. 1) Then
            csil%Krylov(:,NextOrder) = csil%Krylov(:,NextOrder) - CTemp*DtPsi
         Else
            csil%Krylov(:,NextOrder) = csil%Krylov(:,NextOrder) - CTemp*csil%Krylov(:,P)
         EndIf
      EndDo

! --- DETERMINE SUBDIAGONAL ELEMENT OF HESSENBERG MATRIX ---

      If (Order .Eq. 0) Then
         SubDiag = Dble(dot_product(DtPsi, DtPsi))
      Else
         SubDiag = Dble(dot_product(csil%Krylov(:,NextOrder), csil%Krylov(:,NextOrder)))
      EndIf
      SubDiag = Sqrt(SubDiag)
      csil%Hessenberg(NextOrder,Order) = csil%PreFactor * DCmplx(real(SubDiag))

! --- NORMALISE KRYLOV VECTOR ---

      If (SubDiag .LT. PsiEps) Then
         SubDiag = 0.0D0
         CInverse = CZERO
      Else
         CInverse = CONE / csil%Hessenberg(NextOrder,Order)
      EndIf
      If (Order .Eq. 0) Then
         DtPsi = CInverse*DtPsi
      Else
         csil%Krylov(:,NextOrder) = CInverse*csil%Krylov(:,NextOrder)
      EndIf
      Order = NextOrder

! --- ESTIMATE ERROR (OLD ERROR CRITERION)---

      FormerError = Abs(FormerError*Stepsize*SubDiag/Dble(Order))
      If (OldErrCri) Then

!    --- COMPUTE AND OUTPUT ERROR ---

         Error = FormerError
         Call WriteStep(Error)

!    --- CONTINUE UNTIL CONVERGED ---

         If (((Error > csil%accuracy) .And. (Order < csil%maxOrder)) &
            .Or. (Order < MinOrder)) Then

!       --- COMPUTE NEXT KRYLOV VECTOR ---

            If (Order .Eq. 1) Then
               Call Func(Time, DtPsi, csil%Krylov(:,2))
            Else
               Call Func(Time, csil%Krylov(:,Order), csil%Krylov(:,Order+1))
            EndIf
            csil%nfncns = csil%nfncns + 1

            Goto 100
         EndIf
      EndIf
      ActualOrder = Order

! --- ESTIMATE NEXT MATRIX ELEMENTS ---

      csil%Hessenberg(Order-1,Order) = csil%Hessenberg(Order,Order-1)
      csil%Hessenberg(Order,Order) = csil%Hessenberg(Order-1,Order-1)

! --- SAVE HESSENBERG MATRIX ---

      CInverse = CONE/csil%PreFactor
      csil%AuxMat(0:Order,0:Order) = CInverse * csil%Hessenberg(0:Order,0:Order)

! --- DIAGONALISE HESSENBERG MATRIX ---

      Call eigchess(csil%maxOrder+1, Order+1, csil%AuxMat, macheps, &
           csil%EigVal, csil%EigVec, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf

! --- SAVE EIGENVECTORS ---

      csil%AuxMat(0:Order,0:Order) = csil%EigVec(0:Order,0:Order)

! --- SET-UP PSI-VECTOR IN KRYLOV SPACE ---

      csil%Aux1(0) = CONE
      csil%Aux1(1:Order) = CZERO

! --- MULTIPLY INVERSE EIGENVECTOR MATRIX WITH PSI-VECTOR ---

!      Call LinEqz(csil%maxOrder+1, Order+1, csil%AuxMat, csil%Aux1, csil%Aux2, ErrorCode)
      csil%Aux2=csil%Aux1
      allocate(Auxi(Order+1))
      call zgesv(Order+1,1,csil%AuxMat,csil%maxOrder+1,Auxi,csil%Aux2,Order+1,ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 5
         Return
      EndIf
      deallocate(Auxi)

! --- PROPAGATE (OR RELAX) THE PSI-VECTOR IN KRYLOV-SPACE ---

      Do P = 0,Order
         csil%Aux3(P) = Exp(csil%PreFactor * csil%EigVal(P) * Stepsize) * csil%Aux2(P)
      EndDo

! --- MULTIPLY EIGENVECTOR MATRIX WITH PSI-VECTOR ---

      csil%Aux1(0:Order) = MATMUL(csil%EigVec(0:Order,0:Order), csil%Aux3(0:Order))

! --- ESTIMATE THE ERROR (NEW CRITERION)---

      If (.Not. OldErrCri) Then

!    --- COMPUTE AND OUTPUT ERROR ---

         Error = 0.0D0
         If (Order .Eq. 1) Then
            Do P = 0,Order
               csil%OldAux1(P) = csil%Aux1(P)
            EndDo
         Else
            Do P = 0,Order
               Error = Error + Abs(csil%OldAux1(P) - csil%Aux1(P))**2
               csil%OldAux1(P) = csil%Aux1(P)
            EndDo
            Error = Sqrt(Error)
         EndIf
         Call WriteStep(Error)

!    --- CONTINUE UNTIL CONVERGED ---

         If (((Error > csil%accuracy) .And. (Order < csil%maxOrder)) &
             .Or. (Order < MinOrder)) Then

!       --- COMPUTE NEXT KRYLOV VECTOR ---

            If (Order .Eq. 1) Then
               Call Func(Time, DtPsi, csil%Krylov(:,2))
            Else
               Call Func(Time, csil%Krylov(:,Order), csil%Krylov(:,Order+1))
            EndIf
            csil%nfncns = csil%nfncns + 1

            Goto 100
         EndIf
      EndIf

! --- SAVE SOME VALUES IN CASE THAT NEXT CALL IS NON-RESTART CALL ---

      csil%prevOrder = Order
      csil%prevStep  = Stepsize
      csil%prevError = Error

! --- DECREASE STEPSIZE IF LANCZOS DIDN'T CONVERGE ---

      If (Error .LE. csil%accuracy) Goto 300
      Stepsize = Min(Stepsize, Stepsize*(csil%accuracy/FormerError)**(1.0D0/Dble(Order)))
      csil%prevStep = Stepsize
      Call WriteStep(csil%accuracy)

! --- PROPAGATE (OR RELAX) THE PSI-VECTOR IN KRYLOV-SPACE ---

 200  Continue
      Do P = 0,Order
         csil%Aux3(P) = Exp(csil%PreFactor * csil%EigVal(P) * Stepsize) * csil%Aux2(P)
      EndDo

! --- MULTIPLY EIGENVECTOR MATRIX WITH PSI-VECTOR ---

      csil%Aux1(0:Order) = MATMUL(csil%EigVec(0:Order,0:Order), csil%Aux3(0:Order))

! --- PROPAGATE PSI IN HILBERT SPACE ---

! Note again that the Krylov vectors number zero and one aren't stored
! in "Krylov" but in "Psi" and "DtPsi", respectively.

 300  Continue
      Psi = csil%Aux1(0)*Psi + csil%Aux1(1)*DtPsi
      Do P = 2,Order
         Psi = Psi + csil%Aux1(P)*csil%Krylov(:,P)
      EndDo

! --- RE-ESTABLISH THE NORM OF PSI ---

      Psi = Norm*Psi

      ActualOrder = Order
      Return

      contains

         subroutine WriteStep(Error)
            implicit none
            real(dbl) :: Error
            character(len=80) :: msg
            if (llog) then
               write(msg, '(I7,I13,X,2E19.6E2,F18.3)') &
                  csil%nsteps, Order, Stepsize/fs, Error, Time/fs
               call write_log(ilog, msg)
            endif
            return
         end subroutine WriteStep

      End Subroutine CsilStepml


! **********************************************************************
! *                                                                    *
! *                      SUBROUTINE CSILERRORMSG                       *
! *                                                                    *
! * Generates for a given error number returned by "CSILStep/          *
! * CSILStep2" a corresponding error message.                          *
! *                                                                    *
! * Input parameters:                                                  *
! *   Error: Error code returned by CSILStep/CSILStep2.                *
! *                                                                    *
! * Output parameters:                                                 *
! *   Msg:   Error message.                                            *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

   Subroutine CSILErrorMsg (Error,Msg)
      Implicit None
      Integer,intent(in)           :: Error
      Character(len=*),intent(out) :: Msg

! --- GENERATE ERROR MESSAGE ---

      If (Error .Eq. 1) Then
         Msg = 'Illegal integration order'
      ElseIf (Error .Eq. 2) Then
         Msg = 'Stepsize underflow'
      ElseIf (Error .Eq. 3) Then
         Msg = 'Diagonalisation failed'
      ElseIf (Error .Eq. 4) Then
         Msg = 'Illegal initial stepsize'
      ElseIf (Error .Eq. 5) Then
         Msg = 'Cannot solve linear equation'
      Else
         Msg = 'Unknown error occurred'
      EndIf

      Return
   End Subroutine CSILErrorMsg

end module mlcsillib

