module mlrklib
!----------------------------------------------------------------------
! F95 module for the Runge-Kutta integrators.
!----------------------------------------------------------------------

   use global
   use mltypes
   use logging
   use mlbase
   
   use mlpostproc

   implicit none
   private

   public :: int_rk_init, int_rk_drv, int_rk_dispose, int_rk_show, int_rk_stats

!   type,public :: int_rk_t
!      integer   :: order       ! order
!      real(dbl) :: accuracy    ! target accuracy
!      real(dbl) :: initstep    ! initial step size
!      real(dbl) :: laststep    ! last good/optimum step size
!      ! statistical data
!      integer   :: nsteps      ! total number of steps
!      integer   :: naccpt      ! number of accepted steps
!      integer   :: nrejct      ! number of rejected steps
!      integer   :: nfncns      ! number of RHS function calls
!      ! auxiliary vectors
!      complex(cdbl),allocatable :: auxpsi(:,:)
!   end type int_rk_t

!   real(dbl),parameter :: macheps = epsilon(ONER)  ! F95 std.


   contains


   subroutine int_rk_init(rk,ord,acc,stp)
      implicit none
      type(int_rk_t),intent(inout) :: rk     ! object to initialize
      integer,intent(in)           :: ord    ! integration order
      real(dbl),intent(in)         :: acc    ! target accuracy
      real(dbl),intent(in)         :: stp    ! initial stepsize

      if (ord/=5.and.ord/=8) then
         call stop_code('rklib_m', 'int_rk_init', &
            'Runge-Kutta: only order 5 or 8 implemented')
      endif

      rk%order    = ord
      rk%accuracy = acc
      rk%initstep = stp

      rk%laststep = stp
      rk%nsteps   = 0
      rk%naccpt   = 0
      rk%nrejct   = 0
      rk%nfncns   = 0
   end subroutine int_rk_init


   subroutine int_rk_dispose(rk)
      implicit none
      type(int_rk_t),intent(inout):: rk
      deallocate(rk%auxpsi)
   end subroutine int_rk_dispose


   subroutine int_rk_show(rk,desc)
      implicit none
      type(int_rk_t),intent(in)    :: rk
      character(len=*),intent(out) :: desc
      write (desc, '(a,i0,a,es8.2)') &
         'Runge-Kutta, order = ', rk%order, &
         ', accuracy = ', rk%accuracy
   end subroutine int_rk_show


   subroutine int_rk_stats(rk,iout)
      implicit none
      type(int_rk_t),intent(in) :: rk
      integer,intent(in)        :: iout
      character(len=80)         :: desc
      call int_rk_show(rk,desc)
      write (iout,'(a)') trim(desc)
      write (iout,'(2x,a,i9)') 'Total number of steps:   ',rk%nsteps
      write (iout,'(2x,a,i9)') 'Number of accepted steps:',rk%naccpt
      write (iout,'(2x,a,i9)') 'Number of rejected steps:',rk%nrejct
   end subroutine int_rk_stats


   subroutine int_rk_drv(y,f,dt,t,restart,rk,funk,llog,ilog,lpost,ipost,&
              lrhonrm,dmat,dim)
   ! Replaces RUNRK5/RUNRK8/RK5/RK8 routines.
      implicit none
      complex(cdbl),intent(in)    :: dmat(:,:) ! density matrix of current node. 
                                               ! Dieter Apr2021
      complex(cdbl),intent(inout) :: y(:),f(:)
      real(dbl),intent(in)        :: dt,t
      logical,intent(in)          :: restart,llog,lpost,lrhonrm
      integer,intent(in)          :: ilog,ipost,dim
      type(int_rk_t),intent(inout):: rk
      interface
         subroutine funk(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine funk
      end interface
      integer   :: naux,intgfl,ierr
      real(dbl) :: time,tend,hmax,h,hopt,rerr
      character(len=80) :: msg

      if (.not.allocated(rk%auxpsi)) then
         naux=1
         select case (rk%order)
            case (5)
               naux = 7
            case (8)
               naux = 12
            case default
               call stop_code('rklib_m', 'int_rk_init', 'illegal order')
         end select
         allocate(rk%auxpsi(size(y),naux))
      endif

      if (llog) then
         write (msg,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)') &
            '#  Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)', 'Time [fs]'
         call write_log(ilog,msg)
      endif

      time = t
      tend = t + dt
      hmax = dt
      if (restart) then
         h = rk%initstep
      else
         h = rk%laststep
      endif
      hopt = 0.d0
      intgfl = 0

      select case (rk%order)
         case (5)
            call ODP5CRml( rk, funk, size(y), y,                                 &
               time, tend, hmax, h, hopt, rk%accuracy,                         &
               macheps, 0.9d0, 0.04d0, 0.25d0, 8.0d0,                          &
               rk%auxpsi(:,1), rk%auxpsi(:,2), f(:), rk%auxpsi(:,3),           &
               rk%auxpsi(:,4), rk%auxpsi(:,5), rk%auxpsi(:,6), rk%auxpsi(:,7), &
               intgfl, ierr, rerr, lpost, ipost, llog, ilog, lrhonrm, dmat, dim )
         case (8)
            call OD85CRml( rk, funk, size(y), y,                                &
               time, tend, hmax, h, hopt, rk%accuracy,                          &
               macheps, 0.9d0, 0.04d0, 0.25d0, 8.0d0,                           &
               rk%auxpsi(:,1), rk%auxpsi(:,2), rk%auxpsi(:,3), rk%auxpsi(:,4),  &
               f(:), rk%auxpsi(:,5), rk%auxpsi(:,6), rk%auxpsi(:,7), rk%auxpsi(:,8), &
               rk%auxpsi(:,9), rk%auxpsi(:,10), rk%auxpsi(:,11), rk%auxpsi(:,12),    &
               intgfl, ierr, rerr, lpost, ipost, llog, ilog, lrhonrm, dmat, dim )
         case default
            call stop_code('rklib_m', 'int_rk_init', 'illegal order')
      end select

      if (intgfl.ne.1) then
         write (msg,'(a,i0,a,f16.10)') &
            'Error number ',ierr,' parameter ',rerr
         call stop_code('rklib_m', 'int_rk_drv', msg)
      endif

      if (hopt.gt.ZEROR) rk%laststep = hopt

      return
   end subroutine int_rk_drv



!***********************************************************************
!
! SUBROUTINE ODP5CR
!
! Core of the Runge-Kutta integrator of order 5
! (Dormand-Prince formula)
!
! REFERENCE : Hairer/Norsett/Wanner
!             Solving Ordinary Differential Equations I, 2.ed.
!             Springer Verlag, 1993
!
!***********************************************************************
!
! INPUT:
! ------
! rk      INT_RK_T    Integrator object (settings, statistics, ...)
! ODEMOD  external subroutine f that calculates the RHS of the
!         differential equation dy/dt = f(y,t).
!         Must have this interface:
!
!         subroutine ODEMOD(t,y,f)
!         real*8     t       in: time of RHS-evaluation
!         complex*16 y(:)    in: current values of the dynamic variables
!         complex*16 f(:)    out: result of RHS-evaluation
!
! ndgl    INTEGER     number of dynamic variables
! y       COMPLEX*16  initial values of the dynamic variables
! t       REAL*8      start time of integration
! tend    REAL*8      end time of integration
! hmax    REAL*8      maximum allowed stepsize
! h       REAL*8      initial stepsize (for h=0, we will guess one)
! tol     REAL*8      accuracy parameter
!
! uround  REAL*8      machine precision
! safe    REAL*8      safety factor for stepsize prediction
! beta    REAL*8      parameter for Lund-stabilisation
! facdec  REAL*8      factor for maximal decrease of stepsize
! facinc  REAL*8      factor for maximal increase of stepsize
!
! y1      COMPLEX*16  storage for intermediate result
! k1..k6, COMPLEX*16  storage for Runge-Kutta stages
! ys
!
! intgfl  INTEGER     control/error flag
!                     0: this is the first call
!                     1: this is a follow-up call
!
! lpost   LOGICAL     whether to call a post-processing function after
!                     each accepted step
! ipost   INTEGER     parameter for the post-processing function
! llog    LOGICAL     whether to log data about stepsize etc.
! ilog    INTEGER     log channel
!
! OUTPUT:
! -------
! y       COMPLEX*16  final values of dynamic variables
! t       REAL*8      actual time reached in integration
! h       REAL*8      last stepsize
! hopt    REAL*8      last optimal stepsize (not trimmed)
!
! nfcn    INTEGER     number of evaluations of righthand side
! nstep   INTEGER     number of steps
! naccpt  INTEGER     number of accepted steps
! nrejct  INTEGER     number of rejected steps
!
! intgfl  INTEGER     control/error flag
!                        1: OK
!                     else: error!
! ierr    INTEGER     error code
! rerr    REAL*8      error parameter
!
!***********************************************************************

   subroutine ODP5CRml( rk,                         &
      ODEMOD, ndgl, y, t, tend, hmax, h, hopt, tol, &
      uround, safe, beta, facdec, facinc,           &
      y1, k1, k2, k3, k4, k5, k6, ys,               &
      intgfl, ierr, rerr,                           &
      lpost, ipost, llog, ilog, lrhonrm, rho, dim)

      implicit none
      type(int_rk_t),intent(inout) :: rk
      interface
         subroutine ODEMOD(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine ODEMOD
      end interface
      complex(cdbl),intent(in) :: rho(:,:) ! density matrix of current node. 
                                           ! Dieter Apr2021
      integer      :: ndgl, intgfl, ierr, ipost, ilog, dim, subdim
      real(dbl)    :: t, tend, hmax, h, hopt, tol, uround, &
                      safe, beta, facdec, facinc, rerr
      complex(cdbl):: y(ndgl), y1(ndgl), ys(ndgl), k1(ndgl), k2(ndgl), &
                      k3(ndgl), k4(ndgl), k5(ndgl), k6(ndgl)
      logical      :: lpost, llog, lrhonrm

      integer      :: stepnr, iasti, iord
      real(dbl)    :: err, expo1, fac, fac1, facold, facdc1, facin1, &
                      posneg, hnew 
      real(dbl)    :: told = 0.0d0
      logical      :: lastfl, lrjcfl, lcntfl
      character(len=80) :: msg

! --- Coefficients of the method ---
      real(dbl),parameter ::         &
         c2  =      0.2d0          , &
         c3  =      0.3d0          , &
         c4  =      0.8d0          , &
         c5  =      8.d0/9.d0      , &
         a21 =      0.2d0          , &
         a31 =      3.d0/40.d0     , &
         a32 =      9.d0/40.d0     , &
         a41 =     44.d0/45.d0     , &
         a42 = -   56.d0/15.d0     , &
         a43 =     32.d0/9.d0      , &
         a51 =  19372.d0/6561.d0   , &
         a52 = -25360.d0/2187.d0   , &
         a53 =  64448.d0/6561.d0   , &
         a54 = -  212.d0/729.d0    , &
         a61 =   9017.d0/3168.d0   , &
         a62 = -  355.d0/33.d0     , &
         a63 =  46732.d0/5247.d0   , &
         a64 =     49.d0/176.d0    , &
         a65 = - 5103.d0/18656.d0  , &
         b1  =     35.d0/384.d0    , &
         b3  =    500.d0/1113.d0   , &
         b4  =    125.d0/192.d0    , &
         b5  = - 2187.d0/6784.d0   , &
         b6  =     11.d0/84.d0     , &
         e1  =     71.d0/57600.d0  , &
         e3  = -   71.d0/16695.d0  , &
         e4  =     71.d0/1920.d0   , &
         e5  = -17253.d0/339200.d0 , &
         e6  =     22.d0/525.d0    , &
         e7  = -    1.d0/40.d0


      stepnr = 1
      IF (intgfl .EQ. 1) THEN
         err    = .1d0        ! zur Sicherheit
         GOTO 320             ! Zwischenlabel, (kein Sprung in IF-Block)
      ENDIF
! --- Some factors for stepsize-control ---
      expo1   = 0.2d0-beta*0.75d0
      facold  = safe**(1.d0/(expo1-beta))
      facdc1  = ONER/facdec
      facin1  = ONER/facinc
! --- Some more preparations ---
      posneg = SIGN(ONER,tend-t)
      iasti  = 0
      y1(:)  = y(:)
! --- Initialize flags ---
      lastfl = .FALSE.
      lrjcfl = .FALSE.
! --- Initial model evaluation ---
      !can be omitted since k2=dtpsi
      !CALL ODEMOD(t,y1,k2)
      !nfcn = nfcn+1

! === RESTART ===
      lastfl = .FALSE.
! --- Preparations for ODEHIN-CALL ---
      hnew   = h
      hmax   = ABS(hmax)
      iord   = 5

      IF (hnew .EQ. ZEROR) THEN
         call ODEHINml(ODEMOD, ndgl, t, y1, tend, posneg, k2, k3, k4, &
              iord, hmax, uround, tol, rk%nfncns, hnew)
      ENDIF
      hnew = ABS(hnew)*posneg

! ========================================
! === TOP OF THE MAIN INTEGRATION LOOP ===
! ========================================
   1  CONTINUE

         lcntfl = .FALSE.
         h      =  hnew
         IF (lrjcfl) THEN
            y1(:) = y(:)
         ELSE
! ---       Copy result to y, k2 to k1 ---
            y(:) = y1(:)
            k1(:)= k2(:)
         ENDIF
!    ========================
!    === LEAVE INTEGRATOR ===
!    ========================
         IF ((intgfl .NE. 0) .OR. lastfl ) THEN
! ---       Leave the core-integrator ---
            if (lastfl) intgfl = 1
            RETURN
         ENDIF
! ---    Is the stepsize too small ? ---
         IF (0.1d0*ABS(h) .LE. ABS(t)*uround) THEN
            intgfl = -51
            rerr   =  t
            ierr   =  int(LOG10(h))
            GOTO 1
         ENDIF
! ---    Adjust h, if integrator is near the end ---
         IF ((t+1.01d0*h-tend)*posneg .GT. ZEROR) THEN
            h      =  tend - t
            lastfl = .TRUE.
         ENDIF
!    ===============================================
!    === CALCULATE THE 6 STAGES FOR THE NEW STEP ===
!    ===============================================
         rk%nsteps = rk%nsteps + 1
! ---    Stage 1 ---
         y1(:) = y1(:) + (a21*h)*k1(:)

         CALL ODEMOD(t+c2*h,y1,k2)
! ---    Stage 2 ---
         y1(:)= (a31*h)*k1(:) + (a32*h)*k2(:) + y(:)

         CALL ODEMOD(t+c3*h,y1,k3)
! ---    Stage 3 ---
         y1(:)= (a41*h)*k1(:) + (a42*h)*k2(:) + (a43*h)*k3(:) + y(:)

         CALL ODEMOD(t+c4*h,y1,k4)
! ---    Stage 4 ---
         y1(:)= (a51*h)*k1(:) + (a52*h)*k2(:) + (a53*h)*k3(:) &
              + (a54*h)*k4(:) + y(:)

         CALL ODEMOD(t+c5*h,y1,k5)
! ---    Stage 5 ---
         ys(:)= (a61*h)*k1(:) + (a62*h)*k2(:) + (a63*h)*k3(:) &
              + (a64*h)*k4(:) + (a65*h)*k5(:) + y(:)

         CALL ODEMOD(t+h,ys,k6)
! ---    Stage 6 ---
         y1(:)= (b1*h)*k1(:) + (b3*h)*k3(:) + (b4*h)*k4(:) &
              + (b5*h)*k5(:) + (b6*h)*k6(:) + y(:)

         CALL ODEMOD(t+h,y1,k2)
         rk%nfncns = rk%nfncns + 6

! ---    The error-stage ---
         k4(:)= (e4*h)*k4(:) + (e1*h)*k1(:) + (e3*h)*k3(:) &
              + (e5*h)*k5(:) + (e6*h)*k6(:) + (e7*h)*k2(:)
! ---    Estimate the error ---
         if(lrhonrm) then
            err = rhonorm(subdim,dim,k4,rho)/tol
         else
           err = WGTNRM(ndgl,y,y1,k4,tol,uround)
         endif
! ---    Compute hnew ---
         fac1  = err**expo1
! ---    Lund-stabilization ---
         fac   = fac1/facold**beta
! ---    We require  facdec <= hnew/h <= facinc ---
         fac   = MAX(facin1,MIN(facdc1,fac/safe))
         hnew  = h/fac
!    ========================
!    === STEP IS ACCEPTED ===
!    ========================
 320     CONTINUE
         IF (err .LE. ONER) THEN
            IF (intgfl .EQ. 1) THEN
               intgfl = 0
               GOTO 32
            ENDIF
            lrjcfl = .FALSE.
            facold =  MAX(err,1.0d-4)
            rk%naccpt =  rk%naccpt + 1
!       --- Adjust time ---
            told = t
            t    = t + h

   32       CONTINUE
!           write into MCTDH's STEPS-file
            if (llog) then
               write (msg,'(I7,I13,X,2E19.6E2,F18.3)') &
                  stepnr, 1, h/fs, err, told/fs
               call write_log(ilog,msg)
            endif
!           call post-processing function if required
            if (lpost) call int_postprocess(y1,ndgl,ipost)
            stepnr = stepnr+1

! ===       PREPARE NEXT STEP  ===
! ---       Calculation of hnew ---
            IF (ABS(hnew) .GT. hmax) hnew = posneg*hmax
            IF (lrjcfl) hnew = posneg*MIN(ABS(hnew),ABS(h))

!    ========================
!    === STEP IS REJECTED ===
!    ========================
         ELSE
!           write into MCTDH's STEPS-file
            if (llog) then
               write (msg,'(I7,I13,X,2E19.6E2,F18.3)') &
                  stepnr, 0, h/fs, err, t/fs
               call write_log(ilog,msg)
            endif
            stepnr = stepnr+1
! ---       Calculation of hnew ---
            hnew   = h/MIN(facdc1,fac1/safe)
! ---       Adjust the flags and statistics ---
            lrjcfl = .TRUE.
            lastfl = .FALSE.
!           IF (naccpt .GE. 1) nrejct = nrejct + 1
            rk%nrejct = rk%nrejct+1
         ENDIF

! ====================================================
! === HERE IS THE END OF THE MAIN INTEGRATION-LOOP ===
! ====================================================
      IF (.NOT. lastfl) hopt = hnew
      GOTO 1

   end subroutine ODP5CRml



   subroutine OD85CRml( rk,                         &
      ODEMOD, ndgl, y, t, tend, hmax, h, hopt, tol, &
      uround, safe, beta, facdec, facinc,           &
      y1, k1, k2, k3, k4, k5, k6, k7, k8, k9, k10,  &
      kerr, kerr2,                                  &
      intgfl, ierr, rerr,                           &
      lpost, ipost, llog, ilog, lrhonrm, rho, dim)

      implicit none
      type(int_rk_t),intent(inout) :: rk
      interface
         subroutine ODEMOD(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine ODEMOD
      end interface
      complex(cdbl),intent(in) :: rho(:,:) ! density matrix of current node. 
                                           ! Dieter Apr2021
      integer      :: ndgl, intgfl, ierr, ipost, ilog
      real(dbl)    :: t, tend, hmax, h, hopt, tol, uround, &
                      safe, beta, facdec, facinc, rerr
      complex(cdbl):: y(ndgl), y1(ndgl), k1(ndgl), k2(ndgl), k3(ndgl),  &
                      k4(ndgl), k5(ndgl), k6(ndgl), k7(ndgl), k8(ndgl), &
                      k9(ndgl), k10(ndgl), kerr(ndgl), kerr2(ndgl)
      logical      :: lpost, llog, lrhonrm

      integer      :: stepnr, iasti, iord, dim, subdim
      real(dbl)    :: err, expo1, fac, fac1, facold, facdc1, facin1, &
                      posneg, hnew, err2, deno
      real(dbl)    :: told = 0.0d0
      logical      :: lastfl, lrjcfl, lcntfl
      character(len=80) :: msg

! --- Coefficients of the method ---
      real(dbl), parameter :: &
         c2    =  0.526001519587677318785587544488D-01, &
         c3    =  0.789002279381515978178381316732D-01, &
         c4    =  0.118350341907227396726757197510D+00, &
         c5    =  0.281649658092772603273242802490D+00, &
         c6    =  0.333333333333333333333333333333D+00, &
         c7    =  0.25D+00,                             &
         c8    =  0.307692307692307692307692307692D+00, &
         c9    =  0.651282051282051282051282051282D+00, &
         c10   =  0.6D+00,                              &
         c11   =  0.857142857142857142857142857142D+00, &
         b1    =  5.42937341165687622380535766363D-02,  &
         b6    =  4.45031289275240888144113950566D+00,  &
         b7    =  1.89151789931450038304281599044D+00,  &
         b8    = -5.8012039600105847814672114227D+00,   &
         b9    =  3.1116436695781989440891606237D-01,   &
         b10   = -1.52160949662516078556178806805D-01,  &
         b11   =  2.01365400804030348374776537501D-01,  &
         b12   =  4.47106157277725905176885569043D-02,  &
         bh1   =  0.244094488188976377952755905512D+00, &
         bh2   =  0.733846688281611857341361741547D+00, &
         bh3   =  0.220588235294117647058823529412D-01, &
         e1    =  0.1312004499419488073250102996D-01,   &
         e6    = -0.1225156446376204440720569753D+01,   &
         e7    = -0.4957589496572501915214079952D+00,   &
         e8    =  0.1664377182454986536961530415D+01,   &
         e9    = -0.3503288487499736816886487290D+00,   &
         e10   =  0.3341791187130174790297318841D+00,   &
         e11   =  0.8192320648511571246570742613D-01,   &
         e12   = -0.2235530786388629525884427845D-01
      real(dbl), parameter :: &
         a21   =  5.26001519587677318785587544488D-02,  &
         a31   =  1.97250569845378994544595329183D-02,  &
         a32   =  5.91751709536136983633785987549D-02,  &
         a41   =  2.95875854768068491816892993775D-02,  &
         a43   =  8.87627564304205475450678981324D-02,  &
         a51   =  2.41365134159266685502369798665D-01,  &
         a53   = -8.84549479328286085344864962717D-01,  &
         a54   =  9.24834003261792003115737966543D-01,  &
         a61   =  3.7037037037037037037037037037D-02,   &
         a64   =  1.70828608729473871279604482173D-01,  &
         a65   =  1.25467687566822425016691814123D-01,  &
         a71   =  3.7109375D-02,                        &
         a74   =  1.70252211019544039314978060272D-01,  &
         a75   =  6.02165389804559606850219397283D-02,  &
         a76   = -1.7578125D-02,                        &
         a81   =  3.70920001185047927108779319836D-02,  &
         a84   =  1.70383925712239993810214054705D-01,  &
         a85   =  1.07262030446373284651809199168D-01,  &
         a86   = -1.53194377486244017527936158236D-02,  &
         a87   =  8.27378916381402288758473766002D-03,  &
         a91   =  6.24110958716075717114429577812D-01,  &
         a94   = -3.36089262944694129406857109825D+00,  &
         a95   = -8.68219346841726006818189891453D-01,  &
         a96   =  2.75920996994467083049415600797D+01,  &
         a97   =  2.01540675504778934086186788979D+01,  &
         a98   = -4.34898841810699588477366255144D+01
      real(dbl), parameter :: &
         a101  =  4.77662536438264365890433908527D-01,  &
         a104  = -2.48811461997166764192642586468D+00,  &
         a105  = -5.90290826836842996371446475743D-01,  &
         a106  =  2.12300514481811942347288949897D+01,  &
         a107  =  1.52792336328824235832596922938D+01,  &
         a108  = -3.32882109689848629194453265587D+01,  &
         a109  = -2.03312017085086261358222928593D-02,  &
         a111  = -9.3714243008598732571704021658D-01,   &
         a114  =  5.18637242884406370830023853209D+00,  &
         a115  =  1.09143734899672957818500254654D+00,  &
         a116  = -8.14978701074692612513997267357D+00,  &
         a117  = -1.85200656599969598641566180701D+01,  &
         a118  =  2.27394870993505042818970056734D+01,  &
         a119  =  2.49360555267965238987089396762D+00,  &
         a1110 = -3.0467644718982195003823669022D+00,   &
         a121  =  2.27331014751653820792359768449D+00,  &
         a124  = -1.05344954667372501984066689879D+01,  &
         a125  = -2.00087205822486249909675718444D+00,  &
         a126  = -1.79589318631187989172765950534D+01,  &
         a127  =  2.79488845294199600508499808837D+01,  &
         a128  = -2.85899827713502369474065508674D+00,  &
         a129  = -8.87285693353062954433549289258D+00,  &
         a1210 =  1.23605671757943030647266201528D+01,  &
         a1211 =  6.43392746015763530355970484046D-01

! === INITIALIZATIONS ===
      stepnr = 1
      IF (intgfl .EQ. 1) THEN
         err    = .1d0
         GOTO 320
      ENDIF
! --- Some factors for stepsize-control ---
      expo1   = 0.125d0-beta*0.2d0
      facold  = safe**(1.d0/(expo1-beta))
      facdc1  = ONER/facdec
      facin1  = ONER/facinc
! --- Some more preparations ---
      posneg = SIGN(ONER,tend-t)
      iasti  = 0
      k5(:) = y(:)
! --- Initialize flags ---
      lastfl = .FALSE.
      lrjcfl = .FALSE.
! --- Initial model evaluation ---
      !can be omitted since k4 already contains dtpsi
      !CALL ODEMOD(t,k5,k4,mc,mr,mi,ml)
      !nfcn = nfcn+1

! === RESTART ===
      lastfl = .FALSE.
! --- Preparations for ODEHIN-call ---
      hnew   = h
      hmax   = ABS(hmax)
      iord   = 8

      IF (hnew .EQ. ZEROR) THEN
         call ODEHINml(ODEMOD, ndgl, t, k5, tend, posneg, k4, k2, k3, &
              iord, hmax, uround, tol, rk%nfncns, hnew)
      ENDIF
      hnew = ABS(hnew)*posneg

! ========================================
! === TOP OF THE MAIN INTEGRATION LOOP ===
! ========================================
   1  CONTINUE

         lcntfl = .FALSE.
         h      =  hnew
         IF ( .NOT. lrjcfl ) THEN
!      ---  Copy result to y, k4 to k1 ---
            y(:) = k5(:)
            k1(:)= k4(:)
         ENDIF
!    ========================
!    === LEAVE INTEGRATOR ===
!    ========================
         IF ((intgfl .NE. 0) .OR. lastfl ) THEN
!       --- Leave the core-integrator ---
            if (lastfl) intgfl = 1
            RETURN
         ENDIF
!    --- Is the stepsize too small ? ---
         IF (0.1d0*ABS(h) .LE. ABS(t)*uround) THEN
            intgfl = -51
            rerr   =  t
            ierr   =  int(LOG10(h))
            GOTO 1
         ENDIF
!    --- Adjust h, if integrator is near the end ---
         IF ((t+1.01d0*h-tend)*posneg .GT. ZEROR) THEN
            h      =  tend - t
            lastfl = .TRUE.
         ENDIF
!    ================================================
!    === CALCULATE THE 12 STAGES FOR THE NEW STEP ===
!    ================================================
         rk%nsteps = rk%nsteps + 1
!    --- Stage 1 ---
         y1(:) = y(:) + (a21*h)*k1(:)

         CALL ODEMOD(t+c2*h,y1,k2)
!    --- Stage 2 ---
         y1(:) = (a31*h)*k1(:) + (a32*h)*k2(:) + y(:)

         CALL ODEMOD(t+c3*h,y1,k3)
!    --- Stage 3 ---
         y1(:) = (a41*h)*k1(:) + (a43*h)*k3(:) + y(:)

         CALL ODEMOD(t+c4*h,y1,k4)
!    --- Stage 4 ---
         y1(:) = (a51*h)*k1(:) + (a53*h)*k3(:) + (a54*h)*k4(:) + y(:)

         CALL ODEMOD(t+c5*h,y1,k5)
!    --- Stage 5 ---
         y1(:) = (a61*h)*k1(:) + (a64*h)*k4(:) + (a65*h)*k5(:) + y(:)

         CALL ODEMOD(t+c6*h,y1,k6)
!    --- Stage 6 ---
         y1(:) = (a71*h)*k1(:) + (a74*h)*k4(:) + (a75*h)*k5(:) &
               + (a76*h)*k6(:) + y(:)

         CALL ODEMOD(t+c7*h,y1,k7)
!    --- Stage 7 ---
         y1(:) = (a81*h)*k1(:) + (a84*h)*k4(:) + (a85*h)*k5(:) &
               + (a86*h)*k6(:) + (a87*h)*k7(:) + y(:)

         CALL ODEMOD(t+c8*h,y1,k8)
!    --- Stage 8 ---
         y1(:) = (a91*h)*k1(:) + (a94*h)*k4(:) + (a95*h)*k5(:) &
               + (a96*h)*k6(:) + (a97*h)*k7(:) + (a98*h)*k8(:) + y(:)

         CALL ODEMOD(t+c9*h,y1,k9)
!    --- Stage 9 ---
         y1(:) = (a101*h)*k1(:) + (a104*h)*k4(:) + (a105*h)*k5(:) &
               + (a106*h)*k6(:) + (a107*h)*k7(:) + (a108*h)*k8(:) &
               + (a109*h)*k9(:) + y(:)

         CALL ODEMOD(t+c10*h,y1,k10)
!    --- Stage 10 ---
         y1(:) = (a111*h)*k1(:) + (a114*h)*k4(:) + (a115*h)*k5(:) &
               + (a116*h)*k6(:) + (a117*h)*k7(:) + (a118*h)*k8(:) &
               + (a119*h)*k9(:) + (a1110*h)*k10(:) + y(:)

         CALL ODEMOD(t+c11*h,y1,k2)
!    --- Stage 11 ---
         y1(:) = (a121*h)*k1(:) + (a124*h)*k4(:) + (a125*h)*k5(:) &
               + (a126*h)*k6(:) + (a127*h)*k7(:) + (a128*h)*k8(:) &
               + (a129*h)*k9(:) + (a1210*h)*k10(:) + (a1211*h)*k2(:) + y(:)

         CALL ODEMOD(t+h,y1,k3)
         rk%nfncns = rk%nfncns + 11
!    --- Stage 12 ---
         k4(:) = b1*k1(:) + b6*k6(:) + b7*k7(:) + b8*k8(:) + b9*k9(:) &
               + b10*k10(:) + b11*k2(:) + b12*k3(:)
         k5(:) = y(:) + h*k4(:)

!    --- The error-stage ---
         kerr(:) = e1*k1(:) + e6*k6(:) + e7*k7(:) + e8*k8(:) + e9*k9(:) &
                 + e10*k10(:) + e11*k2(:) + e12*k3(:)

!    +-------------------------+
!    +--   Result     -> k5  --+
!    +--   Last stage -> y1  --+
!    +-------------------------+
!    +--   Extra stage -> k4 --+
!    +-------------------------+
!    +--   Later on:         --+
!    +--   RHS(k5)    -> k4  --+
!    +--   RHS(y1)    -> k3  --+
!    +-------------------------+

!    --- Error estimation ---
         if(lrhonrm) then
            err = rhonorm(subdim,dim,kerr,rho)/tol
         else
           err = WGTNRM(ndgl,y,k5,kerr,tol,uround)
         endif

         kerr2(:) = k4(:) + (-bh1)*k1(:) + (-bh2)*k9(:) + (-bh3)*k3(:)

         if(lrhonrm) then
            err2 = rhonorm(subdim,dim,kerr2,rho)/tol
         else
           err2 = WGTNRM(ndgl,y,k5,kerr2,tol,uround)
         endif

         err2 = WGTNRM(ndgl,y,k5,kerr2,tol,uround)
         err   = err**2
         deno  = err + 0.01d0*err2**2
         IF (deno .LE. ZEROR) deno = ONER
         err   = h*(err/SQRT(deno))

!    --- Computation of hnew ---
         fac1  = err**expo1
!    --- LUND-stabilization ---
         fac   = fac1/facold**beta
!    --- We require  facdec <= hnew/h <= facinc ---
         fac   = MAX(facin1,MIN(facdc1,fac/safe))
         hnew  = h/fac
!    ========================
!    === STEP IS ACCEPTED ===
!    ========================
 320     CONTINUE
         IF (err .LE. ONER) THEN
            IF (intgfl .EQ. 1) THEN
               intgfl = 0
               GOTO 32
            ENDIF
            lrjcfl = .FALSE.
!       --- Adjust time ---
            told   = t
            t      = t + h
!       --- New model evaluation  ---
            CALL ODEMOD(t,k5,k4)
            rk%nfncns = rk%nfncns + 1
            facold = MAX(err,1.0d-4)
            rk%naccpt = rk%naccpt + 1

   32       CONTINUE
!           write into MCTDH's STEPS-file
            if (llog) then
               write (msg,'(I7,I13,X,2E19.6E2,F18.3)') &
                  stepnr, 1, h/fs, err, told/fs
               call write_log(ilog,msg)
            endif
!           call post-processor if required
            if (lpost) call int_postprocess(k5,ndgl,ipost)
            stepnr = stepnr+1

!       === PREPARE NEXT STEP  ===
!       --- Calculation of hnew ---
            IF (ABS(hnew) .GT. hmax) hnew = posneg*hmax
            IF (lrjcfl) hnew = posneg*MIN(ABS(hnew),ABS(h))

!    ========================
!    === STEP IS REJECTED ===
!    ========================
         ELSE
!           write into MCTDH's STEPS-file
            if (llog) then
               write (msg,'(I7,I13,X,2E19.6E2,F18.3)') &
                  stepnr, 0, h/fs, err, t/fs
               call write_log(ilog,msg)
            endif
            stepnr = stepnr+1
!       --- Calculation of hnew ---
            hnew   = h/MIN(facdc1,fac1/safe)
!       --- Adjust the flags and statistics ---
            lrjcfl = .TRUE.
            lastfl = .FALSE.
            rk%nrejct = rk%nrejct+1
         ENDIF
! ====================================================
! === HERE IS THE END OF THE MAIN INTEGRATION-LOOP ===
! ====================================================
      IF (.NOT.lastfl)  hopt = hnew
      GOTO 1

   end subroutine OD85CRml



   subroutine ODEHINml(ODEMOD, ndgl, t, y, tend, posneg, dy0, dy1, &
      y1, iord, hmax, small, tol, nfcn, hini)

      implicit none
      interface
         subroutine ODEMOD(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine ODEMOD
      end interface
      integer       :: ndgl, iord, nfcn
      real(dbl)     :: t, tend, posneg, hmax, small, tol, hini
      complex(cdbl) :: y(ndgl), dy0(ndgl), dy1(ndgl), y1(ndgl)
      real(dbl)     :: ynrm, dynrm, h, hmx, ddynrm, dnrmmx, htry

! --- PREPARATIONS ---
      ynrm  = WGTNRM(ndgl,y,y,y,tol,small)
      dynrm = WGTNRM(ndgl,y,y,dy0,tol,small)
! --- COMPUTE STEPSIZE FOR EXPLICIT EULER SUCH THAT   ---
! --- THE INCREMENT IS SMALL COMPARED TO THE SOLUTION ---
      IF ((dynrm .LE. small) .OR. (ynrm .LE. small)) THEN
         h = 1.0d-6
      ELSE
         h = MAX(hmax*small,ABS((ynrm/dynrm)*0.01d0))
      ENDIF
      hmx = MIN(ABS(tend-t),hmax)
      h   = MIN(h,hmx)
      h   = h*posneg
! --- PERFORM AN EXPLICIT EULER STEP ---
      y1(1:ndgl) = y(1:ndgl) + h*dy0(1:ndgl)
! --- EVALUATE ---
      CALL ODEMOD(t+h,y1,dy1)
      nfcn=nfcn+1
! --- ESTIMATE THE SECOND DERIVATIVE OF THE SOLUTION ---
      ddynrm = WGTDST(ndgl,y,y1,dy0,dy1,tol,small)
      ddynrm = ddynrm/h
! --- COMPUTE INITIAL STEP SIZE ---
      dnrmmx = MAX(ABS(ddynrm),dynrm)
      IF (dnrmmx .LE. 1.d-15) THEN
         htry = MAX(1.0d-6,ABS(h)*1.0d-3)
      ELSE
         htry = (0.01d0/dnrmmx)**(ONER/iord)
      ENDIF
      hini = MIN(1.0d2*h,htry,hmx)
      hini = hini*posneg
      RETURN
   END SUBROUTINE ODEHINml



   FUNCTION WGTNRM(ndim,y0,y1,vec,tol,small)
!  computes the weighted norm of a vector
      implicit none
      integer       :: ndim
      complex(cdbl) :: y0(ndim),y1(ndim),vec(ndim)
      real(dbl)     :: tol,small
      integer       :: i
      real(dbl)     :: WGTNRM,wt,tmp

      tmp = 0.d0
      do i = 1,ndim
         wt  = tol + tol*MAX(ABS(y0(i)), ABS(y1(i)), small)
         tmp = tmp + (abs(vec(i))/wt)**2
      enddo
      WGTNRM = SQRT(tmp/ndim)
      return
   END FUNCTION WGTNRM

   FUNCTION rhonorm(subdim,dim,vec,rho)
!  computes a rho-weighted norm of a vector
     IMPLICIT NONE
     INTEGER    ::  i,j,k,dim,subdim
     COMPLEX*16 ::  vec(subdim,dim),rho(dim,dim),ovl
     REAL(dbl)  ::  rhonorm, tmp

     tmp = 0.d0
     DO j = 2, dim
        DO k = 1, j-1
           ovl = 0.d0
           DO i = 1, subdim
              ovl = ovl + dconjg(vec(i,j))*vec(i,k)
           ENDDO
           tmp = tmp + 2.d0*dble(rho(j,k)*ovl)
        ENDDO
     ENDDO

     DO j = 1, dim
        ovl = 0.d0
        DO i = 1, subdim
           ovl = ovl + dconjg(vec(i,j))*vec(i,j)
        ENDDO
        tmp = tmp +  (dble(rho(j,j))+1.d-8)*dble(ovl)
     ENDDO

     rhonorm = SQRT(tmp)*0.5d0

     RETURN
   END FUNCTION rhonorm


   FUNCTION WGTDST(ndim,y0,y1,vec1,vec2,tol,small)
!  computes the weighted distance of two vectors
      implicit none
      integer       :: ndim
      complex(cdbl) :: y0(ndim),y1(ndim),vec1(ndim),vec2(ndim)
      real(dbl)     :: tol,small
      integer       :: i
      real(dbl)     :: WGTDST,wt,tmp

      tmp = 0.d0
      do i = 1,ndim
         wt  = tol + tol*MAX(ABS(y0(i)), ABS(y1(i)), small)
         tmp = tmp + (abs(vec1(i)-vec2(i))/wt)**2
      enddo
      WGTDST = SQRT(tmp/ndim)
      RETURN
   END FUNCTION WGTDST


end module mlrklib
