module mlsillib
!----------------------------------------------------------------------
! F95 module for the Short Iterative Lanczos (SIL) integrator.
!
! FO 12/2012
!----------------------------------------------------------------------

   use global
   use mlbase
   use logging

   use mltypes
   use mlpostproc

   implicit none
   private

   public :: int_sil_init, int_sil_drv, int_sil_dispose, int_sil_show, int_sil_stats

!   type,public :: int_sil_t
!      integer       :: maxOrder    ! maximum order allowed
!      real(dbl)     :: accuracy    ! target accuracy
!      complex(cdbl) :: PreFactor   ! function f is actually f = prefactor*H*y
!      ! data kept between calls
!      integer       :: prevOrder   ! order used in previous step
!      real(dbl)     :: prevStep    ! interval integrated in previous step
!      real(dbl)     :: prevError   ! error measured in previous step
!      ! statistical data
!      integer       :: nsteps      ! total number of steps
!      integer       :: ngoodst     ! number of good steps
!      integer       :: nbadst      ! number of bad steps (order too low)
!      integer       :: nfncns      ! number of RHS function calls
!      integer       :: realOrder   ! largest order actually used
!      ! auxiliary vectors
!      complex(cdbl),allocatable :: Krylov(:,:)      ! storage for the Krylov space
!      real(dbl),allocatable     :: EigenVector(:,:)
!      real(dbl),allocatable     :: Diagonal(:)
!      real(dbl),allocatable     :: OffDiag(:)
!      real(dbl),allocatable     :: EigenVal(:)
!      real(dbl),allocatable     :: OffDg2(:)
!   end type int_sil_t

!   real(dbl),parameter :: macheps = epsilon(ONER)  ! F95 std.


   contains


   subroutine int_sil_init(sil,ord,acc,pre)
   ! Initialize the SIL integrator object.
      implicit none
      type(int_sil_t),intent(inout) :: sil   ! object to initialize
      integer,intent(in)            :: ord   ! maximum order
      real(dbl),intent(in)          :: acc   ! target accuracy
      complex(cdbl),intent(in)      :: pre   ! prefactor

      if (ord < 2) then
         call stop_code('sillib_m', 'int_sil_init', 'order too small')
      endif
      sil%maxorder  = ord
      sil%accuracy  = acc
      sil%prefactor = pre
      sil%prevorder = 0
      sil%prevstep  = ZEROR
      sil%preverror = ZEROR
      sil%nsteps    = 0
      sil%ngoodst   = 0
      sil%nbadst    = 0
      sil%nfncns    = 0
      sil%realorder = 0
      allocate(sil%EigenVector(0:ord, 0:ord+2))
      allocate(sil%Diagonal(0:ord))
      allocate(sil%OffDiag(1:ord))
      allocate(sil%EigenVal(0:ord))
      allocate(sil%OffDg2(1:ord+1))
   end subroutine int_sil_init


   subroutine int_sil_dispose(sil)
   ! Free memory used by the SIL integrator object.
      implicit none
      type(int_sil_t),intent(inout) :: sil
      if (allocated(sil%Krylov)) deallocate(sil%Krylov)
      deallocate(sil%OffDg2)
      deallocate(sil%EigenVal)
      deallocate(sil%OffDiag)
      deallocate(sil%Diagonal)
      deallocate(sil%EigenVector)
   end subroutine int_sil_dispose


   subroutine int_sil_show(sil,desc)
   ! Return a textual description of the SIL integrator.
      implicit none
      type(int_sil_t),intent(in)   :: sil
      character(len=*),intent(out) :: desc
      write (desc, '(a,i0,a,es8.2)') &
         'Short Iterative Lanczos, max.order = ', sil%maxorder, &
         ', accuracy = ', sil%accuracy
   end subroutine int_sil_show


   subroutine int_sil_stats(sil,iout)
   ! Write statistical information about the SIL integrator to channel iout.
      implicit none
      type(int_sil_t),intent(in) :: sil
      integer,intent(in)         :: iout
      character(len=80)          :: desc
      call int_sil_show(sil,desc)
      write (iout,'(a)') trim(desc)
      write (iout,'(2x,a,i9)') 'Total number of steps: ',sil%nsteps
      write (iout,'(2x,a,i9)') 'Number of good steps:  ',sil%ngoodst
      write (iout,'(2x,a,i9)') 'Number of bad steps:   ',sil%nbadst
      write (iout,'(2x,a,i9)') 'Maximum order used:    ',sil%realorder
   end subroutine int_sil_stats


   subroutine int_sil_drv(y,f,dt,t,restart,sil,funk,llog,ilog,lpost,ipost)
   ! Drives the SIL integration of the equation dy/dt = funk(y).
   ! funk must be of the form: funk(y) = p*H*y
   !   where H is a Hermition operator
   !   and p is a prefactor (usually -I)
   ! y should be normalized.
   ! Replaces the RUNSIL routine.
      implicit none
      complex(cdbl),intent(inout)  :: y(:)    ! initial/final values of dynamic variables
      complex(cdbl),intent(inout)  :: f(:)    ! initially, f=funk(y) [we assume this has already been done]
      real(dbl),intent(in)         :: dt      ! time interval to integrate
      real(dbl),intent(in)         :: t       ! current time
      logical,intent(in)           :: restart ! if .true., reuse previous Krylov space
      logical,intent(in)           :: llog    ! if .true., log information about each step
      integer,intent(in)           :: ilog    ! log channel
      logical,intent(in)           :: lpost   ! if .true., call post-processer after each SIL step
      integer,intent(in)           :: ipost   ! cf procedure int_postprocess
      type(int_sil_t),intent(inout):: sil     ! object with integrator settings
      interface
         subroutine funk(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine funk
      end interface
      real(dbl)         :: inttime ! time integrated so far
      real(dbl)         :: hmax    ! maximum size of next SIL step
      real(dbl)         :: hact    ! actual size of last SIL step
      integer           :: actord  ! actual order used in the SIL step
      integer           :: ierr    ! error code
      character(len=80) :: msg     ! text buffer

      ! allocate Krylov space if not done previously
      if (.not.allocated(sil%Krylov)) then
         allocate(sil%Krylov(size(y), 2:sil%maxorder))
      endif

      if (llog) then
         write (msg,'(a7,9x,a6,6x,a13,11x,a5,11x,a9)') &
            '#  Step', 'Order', 'Stepsize [fs]', 'Error', 'Time [fs]'
         call write_log(ilog,msg)
      endif

      ! do one SIL step, actual stepsize returned in hact
      inttime = ZEROR
 10   hmax = dt - inttime
      call silstepml( sil, &
         funk, size(y), y, f, t+inttime, hmax, restart, &
         hact, actord, ierr, llog, ilog)

      ! check for errorrs
      if (ierr.ne.0) then
         call SILErrorMsg(ierr, msg)
         call stop_code('sillib_m', 'int_sil_drv', msg)
      endif

      ! update step counters and max. used order
      if (deql(hact,hmax)) then
         sil%ngoodst = sil%ngoodst + 1
      else
         sil%nbadst = sil%nbadst + 1
      endif
      sil%nsteps = sil%nsteps + 1
      sil%realorder = max(sil%realorder, actord)

      ! call post-processor
      if (lpost) call int_postprocess(y, size(y), ipost)

      ! are we finished?
      inttime = inttime + hact
      if (.not.dgeq(inttime,dt)) then
         ! no, we need more steps;
         ! evaluate RHS again
         call funk(t+inttime,y,f)
         sil%nfncns = sil%nfncns + 1
         goto 10
      endif

      return
   end subroutine int_sil_drv



! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE SILSTEP                                 *
! *                                                                    *
! * Integrates a system of complex linear first order differential     *
! * equations with constant and Hermitian Hamiltonian employing the    *
! * short iterative Lanczos method. The routine runs with both         *
! * variable step size and order. (First it increases the order to     *
! * achieve the desired accuracy. If this failes within the maximum    *
! * order, the stepsize is reduced.) SILStep makes only one single     *
! * integration step, so it has to be imbedded into a loop that calls  *
! * SILStep until the desired time interval is integrated. The ODE is  *
! * of the form dPsi/dt = prefactor*H|Psi> = Func(Time,Psi,DtPsi).     *
! * All computations are performed with double  precision.             *
! *                                                                    *
! * Input parameters:                                                  *
! *   sil:       Integrator object containg settings etc.              *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. prefactor*H|Psi>.                                *
! *   PsiDim:    Length of Psi and DtPsi vectors.                      *
! *   Time:      Current time (only for logging).                      *
! *   IntPeriod: Lenght of time interval to be integrated.             *
! *   Restart:   Flag for restarting the integrator. If true, the      *
! *              Krylov space is built up before propagation, else the *
! *              old Krylov vectors are used.                          *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   DtPsi:     Normalised first Krylov vector, i. e. (H-<H>)|Psi>.   *
! *   Stepsize:  Time interval that actually has been integrated (can  *
! *              be lower than IntPeriod).                             *
! *   ActualOrder: The order that has actually been used (may be less  *
! *              than IntOrder).                                       *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              2: stepsize underflow,                                *
! *              3: diagonalization failed,                            *
! *              4: stepsize too large for restart.                    *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as Func(AbsTime,Psi,DtPsi).                    *
! *                                                                    *
! * V6.0 MB                                                            *
! * V7.0 GW addition of CData,RData,IData,LData arrays                 *
! * V8.2 05/01 HDM  Addition of 'relaxation' to exited states.         *
! *                 Eigenvector-space now dynamically allocated.       *
! * V8.5 12/12 FO   Convert to F95 integrator interface.               *
! *                                                                    *
! **********************************************************************

   subroutine silstepml( sil, &
      Func, PsiDim, Psi, DtPsi, Time, IntPeriod, Restart, &
      Stepsize, ActualOrder, ErrorCode, llog, ilog)

      implicit none
      type(int_sil_t),intent(inout) :: sil
      interface
         subroutine Func(t,y,f)
            use mlbase
            real(dbl),intent(in)        :: t
            complex(cdbl),intent(inout) :: y(:)
            complex(cdbl),intent(out)   :: f(:)
         end subroutine Func
      end interface
      integer,intent(in)          :: PsiDim, ilog
      complex(cdbl),intent(inout) :: Psi(PsiDim), DtPsi(PsiDim)
      real(dbl),intent(in)        :: Time, IntPeriod
      logical,intent(in)          :: Restart, llog
      real(dbl),intent(out)       :: StepSize
      integer,intent(out)         :: ActualOrder, ErrorCode

      real(dbl),parameter :: RelativeMinStep = 1.0d-8
      real(dbl),parameter :: myTiny = 1.0d-18

      integer           :: Order,P,Q
      real(dbl)         :: Beta,Error,MinStepsize,PsiEps
      complex(cdbl)     :: Alpha,CBeta,CSum,CInverse

! --- INITIALIZE VARIABLES ---

      Stepsize = IntPeriod
      MinStepsize = Abs(RelativeMinStep*IntPeriod)
      ErrorCode = 0
      ActualOrder = 1
      PsiEps = sqrt(dble(PsiDim))*macheps

! --- RETURN RESULT IF DIFFERENTIAL EQUATION IS NOT A SYSTEM ---

      If (PsiDim .Eq. 1) Then
         Order = 1
         Error = 0.0D0
         Call WriteStep(Error)
         Psi(1) = Exp(StepSize*DtPsi(1)/Psi(1))*Psi(1)
         Return
      EndIf

! --- SKIP CALCULATION OF KRYLOV VECTORS IF DESIRED ---

      If (.not.Restart) Then
         If (Abs(Stepsize) .GT. Abs(sil%prevStep)+MinStepsize) Then
            ErrorCode = 4
            Return
         EndIf
         Order = sil%prevOrder
         Error = Abs(sil%prevError*(Stepsize/sil%prevStep)**Order)
         Call WriteStep(Error)
         Goto 200
      EndIf

! --- FURTHER INITIALIZE VARIABLES ---

      Order = 1
      sil%prevError = 0.0D0
      sil%prevOrder = 0
      sil%prevStep  = 0.0D0

! --- COMPUTE FIRST DIAGONAL ELEMENT ---

      Alpha = dot_product(Psi,Dtpsi)
      sil%Diagonal(0) = dble(Alpha/sil%PreFactor)

! --- DETERMINE CORRESPONDING BASIS VECTOR AND OFF-DIAGONAL ELEMENT ---

      DtPsi = DtPsi - Alpha*Psi
      Beta  = Sqrt(Dble(dot_product(DtPsi,DtPsi)))
      sil%OffDiag(1) = Beta
      CBeta = sil%PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      If (Beta .LT. PsiEps) Then
         Beta = 0.0D0
         CInverse = (0.0D0,0.0D0)
      Else
         CInverse = 1.0D0/CBeta
      EndIf
      DtPsi = CInverse*DtPsi

! --- COMPUTE ERROR ESTIMATE ---

      Error = Abs(Beta*Stepsize)

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      Call WriteStep(Error)

! --- BUILD UP THE KRYLOV SPACE ---

 100  Continue

! --- CHECK IF STEPSIZE IS TOO SMALL ---

      If (Abs(Stepsize) .LT. Abs(MinStepSize)) Then
         ErrorCode = 2
         Return
      EndIf

! --- RE-INITIALIZE VARIABLES ---

      Order = Order+1

! --- EVALUATE FUNCTION WITH LAST BASIS VECTOR ---

      If (Order .Eq. 2) Then
         Call Func(Time, DtPsi(:), sil%Krylov(:,2))
      Else
         Call Func(Time, sil%Krylov(:,Order-1), sil%Krylov(:,Order))
      EndIf
      sil%nfncns = sil%nfncns + 1

! --- COMPUTE DIAGONAL ELEMENT ---

! Note that the Krylov vectors number zero and one aren't stored in
! "Krylov" but in "Psi" and "DtPsi", respectively.

      If (Order .Eq. 2) Then
         Alpha = dot_product(DtPsi(:), sil%Krylov(:,2))
      Else
         Alpha = dot_product(sil%Krylov(:,Order-1), sil%Krylov(:,Order))
      EndIf
      sil%Diagonal(Order-1) = dble(Alpha/sil%PreFactor)

! --- COMPUTE OFF-DIAGONAL ELEMENT AND BASIS VECTOR ---

      If (Order .Eq. 2) Then
         sil%Krylov(:,2) = sil%Krylov(:,2) - Alpha*DtPsi(:) - CBeta*Psi(:)
      ElseIf (Order .Eq. 3) Then
         sil%Krylov(:,3) = sil%Krylov(:,3) - Alpha*sil%Krylov(:,2) - CBeta*DtPsi(:)
      Else
         sil%Krylov(:,Order) = sil%Krylov(:,Order) - Alpha*sil%Krylov(:,Order-1) &
            - CBeta*sil%Krylov(:,Order-2)
      EndIf
      Beta = Sqrt(Dble(dot_product(sil%Krylov(:,Order), sil%Krylov(:,Order))))
      sil%OffDiag(Order) = Beta
      CBeta = sil%PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      If (Beta .LT. PsiEps) Then
         sil%OffDiag(Order) = 0.0D0
         Beta = 0.0D0
         CInverse = (0.0D0,0.0D0)
      Else
         CInverse = 1.0D0/CBeta
      EndIf

      sil%Krylov(:,Order) = CInverse*sil%Krylov(:,Order)

! --- COMPUTE ERROR ESTIMATE ---

      Error = Abs(Error*Beta*Stepsize/Dble(Order))

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      Call WriteStep(Error)

! --- CONTINUE IF NOT CONVERGED ---

      If((Order < sil%maxOrder) .And. (Error > sil%accuracy) &
         .And. (Beta > PsiEps) )  Goto 100

! --- ESTIMATE NEXT MATRIX ELEMENT ---

      sil%Diagonal(Order) = sil%Diagonal(Order-1)

! --- DECREASE STEPSIZE IF LANCZOS DIDN'T CONVERGE ---

      If (Error > sil%accuracy) Then
         Stepsize = Stepsize*(sil%accuracy/Error)**(1.0D0/Dble(Order))
         Call WriteStep(sil%accuracy)
      EndIf

! --- SAVE SOME VALUES IN CASE THAT NEXT CALL IS NON-RESTART CALL ---

      sil%prevOrder = Order
      sil%prevStep  = Stepsize
      sil%prevError = Error

! --- Save Diagonal and off diagonal values since the diagonalisation ---
! --- routine will overwrite those. ---

      sil%EigenVal(0:Order) = sil%Diagonal(0:Order)
      sil%OffDg2(1:Order)   = sil%OffDiag(1:Order)

! --- DIAGONALIZE LANCZOS MATRIX --- CALL LAPACK ROUTINE ---
! --- EigenVector(*,IntOrder+1) serves as work array
      call DSTEQR('I', Order+1, sil%EigenVal, sil%OffDg2, sil%EigenVector, &
                  sil%maxOrder+1, sil%EigenVector(1,sil%maxOrder+1), ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf

 200  Continue                  ! Jump to here, if not RESTART.

! --- PROPAGATE WAVEFUNCTION ---

! Note again that the Krylov vectors number zero and one aren't stored
! in "Krylov" but in "Psi" and "DtPsi", respectively.

      Do P = 0,Order
         CSum = (0.0D0,0.0D0)
         Do Q = 0,Order
            CSum = CSum + Exp(sil%PreFactor * sil%EigenVal(Q) * Stepsize) * &
                   sil%EigenVector(0,Q) * sil%EigenVector(P,Q)
         EndDo
         If (P .Eq. 0) Then
            Psi(:) = CSum*Psi(:)
         ElseIf (P .Eq. 1) Then
            Psi(:) = Psi(:) + CSum*DtPsi(:)
         Else
            Psi(:) = Psi(:) + Csum*sil%Krylov(:,P)
         EndIf
      EndDo

      ActualOrder = Order
      Return

      contains

         subroutine WriteStep(Error)
            implicit none
            real(dbl) :: Error
            character(len=80) :: msg
            if (llog) then
               write(msg, '(I7,I13,X,2E19.6E2,F18.3)') &
                  sil%nsteps, Order, Stepsize/fs, Error, Time/fs
               call write_log(ilog, msg)
            endif
            return
         end subroutine WriteStep

   End Subroutine SILStepml


! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE SILERRORMSG                       *
! *                                                                    *
! * Generates for a given error number returned by "SILStep/SILStep2/  *
! * SILStep3" a corresponding error message.                           *
! *                                                                    *
! * Input parameters:                                                  *
! *   Error: Error code returned by SILStep/SILStep2.                  *
! *                                                                    *
! * Output parameters:                                                 *
! *   Msg:   Error message.                                            *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

   Subroutine SILErrorMsg (Error,Msg)
      Implicit None
      Integer          :: Error
      Character(len=*) :: Msg

! --- GENERATE ERROR MESSAGE ---

      If (Error .Eq. 1) Then
         Msg = 'Illegal integration order'
      ElseIf (Error .Eq. 2) Then
         Msg = 'Stepsize underflow'
      ElseIf (Error .Eq. 3) Then
         Msg = 'Diagonalisation failed'
      ElseIf (Error .Eq. 4) Then
         Msg = 'Illegal initial stepsize'
      Else
         Msg = 'Unknown error occurred'
      EndIf

      Return
   End Subroutine SILErrorMsg

end module mlsillib

