! **********************************************************************
! *                                                                    *
! *                   ADAMS-BASHFORTH-MOULTON (abmlib.f)               *
! *                                                                    *
! * Library module containing an Adams-Bashforth-Moulton predictor-    *
! * corrector integrator.                                              *
! *                                                                    *
! * Contains:                                                          *
! *   ABM:          The ABM integration routine.                       *
! *   AbsABMError:  Computes the absolute error of the current ABM     *
! *                 integration step.                                  *
! *   RelABMError:  Computes the relative error of the current ABM     *
! *                 integration step.                                  *
! *   WriteABMStep: In the current form WriteABMStep is just a dummy   *
! *                 routine doing absolutely nothing; it is included   *
! *                 here for formal reasons, but can (if desired) be   *
! *                 extended easily such that it writes the size and   *
! *                 error of the current integration step to a file.   *
! *   ABMErrorMsg:  Returns for a given error number a corresponding   *
! *                 error message.                                     *
! *                                                                    *
! **********************************************************************
      module abmlib

use decimal, only: long,dop

      implicit none

      private
      public :: abm,abmerrormsg,absabmerror,writeabmstep

      contains

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE ABM                           *
! *                                                                    *
! * Integrates a system of complex first order differential equations  *
! * employing the Adams-Bashforth-Moulton predictor-corrector method.  *
! * The ABM routine runs with variable step sizes and (except for the  *
! * beginning) a fixed integration order. The ODE is of the form       *
! * dPsi/dt = Func(AbsTime,Psi) =: DtPsi. All computations are         *
! * performed with double precision. The routine allows a previous     *
! * integration to be resumed (e. g. after the integration had been    *
! * stopped to write Psi to a file), as long as the AuxPsi array       *
! * hasn't been overwritten and the integration order and error        *
! * tolerance hasn't been changed. To do so simply set the restart     *
! * flag "RestartABM" to ".false.".                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:           The (complex) initial-value vector.               *
! *   DtPsi:         Time derivative of the initial-value vector (must *
! *                  be passed to the integrator regardless of the     *
! *                  flag "RestartABM").                               *
! *   PsiDim         Length of Psi and DtPsi vectors.                  *
! *   IntPeriod:     Lenght of time interval to be integrated.         *
! *   AbsTime:       Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   IntOrder:      Desired integration order.                        *
! *   InitStep:      Size of first integration step.                   *
! *   TolError:      Maximum error that is tolerated.                  *
! *   RestartABM:    Restart flag; if false, a previous integration is *
! *                  continued, otherwise a new integration is started *
! *                  (see comment above for details).                  *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:           Solution of the ODE at time AbsTime+IntPeriod.    *
! *   DtPsi:         DtPsi is undetermined.                            *
! *   Steps:         Counter for the overall number of integration     *
! *                  steps. (Note that Steps isn't set to zero at the  *
! *                  beginning.)                                       *
! *   RepeatedSteps: Counter for the number of failed and thus         *
! *                  repeated integration steps. (Note that            *
! *                  RepeatedSteps isn't set to zero at the            *
! *                  beginning.)                                       *
! *   ErrorCode:     Error code having the following meaning:          *
! *                  0: everything was o. k.,                          *
! *                  1: illegal integration order,                     *
! *                  2: stepsize underflow.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   AuxPsi:        Auxiliary array of (minimum) size                 *
! *                  PsiDim*(IntOrder+1).                              *
! *                                                                    *
! * External routines:                                                 *
! *   Func:          Computes the time derivative DtPsi of Psi at time *
! *                  AbsTime. Called as Func(AbsTime,Psi,DtPsi)        *
! *   CalcError:     Determines the error of the current integration   *
! *                  step. Called as CalcError(Predicted_Psi,          *
! *                  Corrected_Psi,PsiDim,Error)                       *
! *   WriteStep:     Writes the stepsize and error to a file.          *
! *                  Called as WriteStep(Step,Order,Stepsize,Error).   *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      Subroutine ABM (Psi,DtPsi,PsiDim,IntPeriod,AbsTime,IntOrder,&
                      InitStep,TolError,RestartABM,Steps,RepeatedSteps,&
                      ErrorCode,AuxPsi,Func,CalcError)

use wrintegrat, only: writestep

      Implicit None

      Real(dop), Parameter :: One6th = 1.0_dop/6.0_dop,One30th = 1.0_dop/30.0_dop,&
                              One210th = 1.0_dop/210.0_dop,RelativeMinStep = 1.0e-12_dop

      Integer(long), Parameter :: MaxOrder = 8

      Logical(kind=4), intent(in)                            :: RestartABM
      Integer(long), intent(inout)                           :: Steps,RepeatedSteps
      Integer(long), intent(out)                             :: ErrorCode
      Integer(long), intent(in)                              :: PsiDim,IntOrder
      Real(dop), intent(in)                                  :: IntPeriod,AbsTime,InitStep,TolError
      Complex(dop), dimension(PsiDim), intent(inout)         :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,IntOrder+1), intent(out):: AuxPsi
      External                                               :: Func,CalcError

      Logical(kind=4), save                                 :: StepIsRepeated,CurOrdEqIntOrd,UseOldH1,lcheckH1
      Integer(long)                                         :: D,P,I
      Integer(long), save                                   :: CurrentOrder
      Real(dop), save                                       :: IntError,MinError,Error
      Real(dop)                                             :: Time,NextTime,D2,D3,D4,D5,D6,D7,&
                                                               H1,H1Sqr,MinusH1Cube,MinStepSize
      Real(dop), dimension(MaxOrder)                        :: PredCoef,CorrCoef
      Real(dop), dimension(MaxOrder), save                  :: Distance
      Real(dop), dimension(MaxOrder-1), save                :: H
      Real(dop), dimension(2:MaxOrder-2,2:MaxOrder-1)       :: SumOfD,ProdOfD
      Complex(dop), dimension(MaxOrder)                     :: InterimAuxPsi

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .GT. MaxOrder) Then
         ErrorCode = 1
         Return
      EndIf

! --- INITIALISE VARIABLES ---

      ErrorCode = 0
      Time = AbsTime
      MinStepSize = RelativeMinStep*IntPeriod
      
! --- CONTINUE INTEGRATION ---

      If (.Not. RestartABM) Then
         Goto 200
      EndIf

! --- INITIALISE VARIABLES ---

! "UseOldH1" is a flag set to true if the step size is artificially
! shortened in order to fit it to the remaining integration period
! before leaving the routine. This allows the next stepsize to be
! corrected back to the previous value.
! "LcheckH1" is set to false if the step size is artificially shortened
! so that it is not checked in writestep as it may be below termination
! criteria due to chance.

      IntError = 0.40_dop*TolError
      MinError = 0.01_dop*TolError
      Do P = 1,IntOrder-1
         H(P) = InitStep
      EndDo
      CurrentOrder = 2
      CurOrdEqIntOrd = .False.
      StepIsRepeated = .False.
      UseOldH1 = .False.
      lcheckH1 = .true.

! --- INITIALISE AUXPSI ---

      Do D = 1,PsiDim
         AuxPsi(D,1) = Psi(D)
         AuxPsi(D,2) = DtPsi(D)
      EndDo

! --- INTEGRATION LOOP ---

 100  Continue

! --- CHECK WETHER STEPSIZE IS TOO SMALL ---

      If ((H(1) .LT. MinStepSize) .And.&
         (H(1) .LT. AbsTime+IntPeriod-Time)) Then
         ErrorCode = 2
         Return
      EndIf

! --- CALCULATE PREDICTOR AND CORRECTOR COEFFICIENTS ---

      H1          = H(1)
      H1Sqr       = H1*H1
      MinusH1Cube = -H1Sqr*H1
      D2          = H(2)
      D3          = D2+H(3)
      D4          = D3+H(4)
      D5          = D4+H(5)
      D6          = D5+H(6)
      D7          = D6+H(7)
      Do P = 1,CurrentOrder
         If (P .Eq. 1) Then
            Distance(1) = H1
            PredCoef(1) = H1
            CorrCoef(1) = H1
         ElseIf (P .Eq. 2) Then
            Distance(2) = Distance(1)+H(2)
            PredCoef(2) = 0.5*H1Sqr
            CorrCoef(2) = -PredCoef(2)
         ElseIf (P .Eq. 3) Then
            Distance(3) = Distance(2)+H(3)
            PredCoef(3) = One6th*H1Sqr*(3.0*D2+2.0*H1)
            CorrCoef(3) = One6th*MinusH1Cube
         ElseIf (P .Eq. 4) Then
            Distance(4) = Distance(3)+H(4)
            SumOfD(2,3) = D2+D3
            ProdOfD(2,3) = D2*D3
            PredCoef(4) = One6th*H1Sqr*(3.0*ProdOfD(2,3)&
                          +(2.0*SumOfD(2,3)+1.5*H1)*H1)
            CorrCoef(4) = One6th*MinusH1Cube*(D2+0.5*H1)
         ElseIf (P .Eq. 5) Then
            Distance(5) = Distance(4)+H(5)
            SumOfD(3,3) = D3
            ProdOfD(3,3) = D3
            Do I = 2,3
               SumOfD(I,4)  = SumOfD(I,3)+D4
               ProdOfD(I,4) = ProdOfD(I,3)*D4
            EndDo
            PredCoef(5) = One30th*H1Sqr*(15.0*ProdOfD(2,4)&
                          +(10.0*(D2*SumOfD(3,4)+ProdOfD(3,4))&
                          +(7.5*SumOfD(2,4)+6.0*H1)*H1)*H1)
            CorrCoef(5) = One30th*MinusH1Cube*(5.0*ProdOfD(2,3)&
                          +(2.5*SumOfD(2,3)+1.5*H1)*H1)
         ElseIf (P .Eq. 6) Then
            Distance(6) = Distance(5)+H(6)
            SumOfD(4,4) = D4
            ProdOfD(4,4) = D4
            Do I = 2,4
               SumOfD(I,5)  = SumOfD(I,4)+D5
               ProdOfD(I,5) = ProdOfD(I,4)*D5
            EndDo
            PredCoef(6) = One30th*H1Sqr*(15.0*ProdOfD(2,5)&
                          +(10.0*(D2*(D3*SumOfD(4,5)+ProdOfD(4,5))&
                          +ProdOfD(3,5))&
                          +(7.5*(D2*SumOfD(3,5)+D3*SumOfD(4,5)&
                          +ProdOfD(4,5))&
                          +(6.0*SumOfD(2,5)+5.0*H1)*H1)*H1)*H1)
            CorrCoef(6) = One30th*MinusH1Cube*(5.0*ProdOfD(2,4)&
                          +(2.5*(D2*SumOfD(3,4)+ProdOfD(3,4))&
                          +(1.5*SumOfD(2,4)+H1)*H1)*H1)
         ElseIf (P .Eq. 7) Then
            Distance(7) = Distance(6)+H(7)
            SumOfD(5,5) = D5
            ProdOfD(5,5) = D5
            Do I = 2,5
               SumOfD(I,6)  = SumOfD(I,5)+D6
               ProdOfD(I,6) = ProdOfD(I,5)*D6
            EndDo
            PredCoef(7) = One210th*H1Sqr*(105.0*ProdOfD(2,6)&
                          +(70.0*(D2*(D3*(D4*SumOfD(5,6)+ProdOfD(5,6))&
                          +ProdOfD(4,6))+ProdOfD(3,6))&
                          +(52.5*(D2*(D3*SumOfD(4,6)+D4*SumOfD(5,6)&
                          +ProdOfD(5,6))+D3*(D4*SumOfD(5,6)&
                          +ProdOfD(5,6))+ProdOfD(4,6))&
                          +(42.0*(D2*SumOfD(3,6)+D3*SumOfD(4,6)&
                          +D4*SumOfD(5,6)+ProdOfD(5,6))&
                          +(35.0*SumOfD(2,6)+30.0*H1)*H1)*H1)*H1)*H1)
            CorrCoef(7) = One210th*MinusH1Cube*(35.0*ProdOfD(2,5)&
                          +(17.5*(D2*(D3*SumOfD(4,5)+ProdOfD(4,5))&
                          +ProdOfD(3,5))&
                          +(10.5*(D2*SumOfD(3,5)+D3*SumOfD(4,5)&
                          +ProdOfD(4,5))+(7.0*SumOfD(2,5)&
                          +5.0*H1)*H1)*H1)*H1)
         ElseIf (P .Eq. 8) Then
            SumOfD(6,6) = D6
            ProdOfD(6,6) = D6
            Do I = 2,6
               SumOfD(I,7)  = SumOfD(I,6)+D7
               ProdOfD(I,7) = ProdOfD(I,6)*D7
            EndDo
            PredCoef(8) = One210th*H1Sqr*(105.0*ProdOfD(2,7)&
                          +(60.0*(D2*(D3*(D4*(D5*SumOfD(6,7)&
                          +ProdOfD(6,7))+ProdOfD(5,7))+ProdOfD(4,7))&
                          +ProdOfD(3,7))&
                          +(52.5*(D2*(D3*(D4*SumOfD(5,7)+D5*SumOfD(6,7))&
                          +ProdOfD(6,7))+D4*(D5*SumOfD(6,7)&
                          +ProdOfD(6,7))+ProdOfD(5,7))&
                          +D3*(D4*(D5*SumOfD(6,7)+ProdOfD(6,7)&
                          +ProdOfD(5,7))+ProdOfD(4,7))&
                          +(42.0*(D2*(D3*SumOfD(4,7)+D4*SumOfD(5,7)&
                          +D5*SumOfD(6,7)+ProdOfD(6,7))&
                          +D3*(D4*SumOfD(5,7)+D5*SumOfD(6,7)&
                          +ProdOfD(6,7))+D4*(D5*SumOfD(6,7)&
                          +ProdOfD(6,7))+ProdOfD(5,7))&
                          +(35.0*(D2*SumOfD(3,7)+D3*SumOfD(4,7)&
                          +D4*SumOfD(5,7)+D5*SumOfD(6,7)+ProdOfD(6,7))&
                          +(30.0*SumOfD(2,7)&
                          +26.25*H1)*H1)*H1)*H1)*H1)*H1)
            CorrCoef(8) = One210th*MinusH1Cube*(35.0*ProdOfD(2,6)&
                          +(17.5*(D2*(D3*(D4*SumOfD(5,6)+ProdOfD(5,6))&
                          +ProdOfD(4,6))+ProdOfD(3,6))&
                          +(10.5*(D2*(D3*SumOfD(4,6)+D4*SumOfD(5,6)&
                          +ProdOfD(5,6))+D3*(D4*SumOfD(5,6)&
                          +ProdOfD(5,6))+ProdOfD(4,6))&
                          +(7.0*(D2*SumOfD(3,6)+D3*SumOfD(4,6)&
                          +D4*SumOfD(5,6)+ProdOfD(5,6))+(5.0*SumOfD(2,6)&
                          +3.75*H1)*H1)*H1)*H1)*H1)
         Else
            ErrorCode = 1
            Return
         EndIf
      EndDo

! --- RESTORE ORIGINAL PSI WHEN STEP IS REPEATED ---

      If (StepIsRepeated) Then
         Do D = 1,PsiDim
            Psi(D) = AuxPsi(D,1)
         EndDo
         StepIsRepeated = .False.
      EndIf

! --- PREDICT PSI ---	

! As long as the order p is increased, the predictor must be used as a
! p-step method. Later on, the predictor is run as a (p+1)-step method.

      Do D = 1,PsiDim
         If (CurOrdEqIntOrd) Then
            Do P = 1,CurrentOrder
               Psi(D) = Psi(D)+PredCoef(P)*AuxPsi(D,P+1)
            EndDo
         Else
            Do P = 1,CurrentOrder-1
               Psi(D) = Psi(D)+PredCoef(P)*AuxPsi(D,P+1)
            EndDo
         EndIf
      EndDo
                      
! --- EVALUATE FUNCTION WITH PREDICTED PSI ---

      NextTime = Time+H(1)
      Call Func(NextTime,Psi,DtPsi)

! --- SAVE PREDICTED PSI IN DTPSI AND CORRECT PSI ---

      Do D = 1,PsiDim
         InterimAuxPsi(1) = DtPsi(D)
         Do P = 2,CurrentOrder
            InterimAuxPsi(P) = (InterimAuxPsi(P-1)-AuxPsi(D,P))&
                               /Distance(P-1)
         EndDo
         DtPsi(D) = Psi(D)
         Psi(D) = AuxPsi(D,1)
         Do P = 1,CurrentOrder
            Psi(D) = Psi(D)+CorrCoef(P)*InterimAuxPsi(P)
         EndDo
      EndDo

! --- CALCULATE ERROR AND (IF REQUIRED) NEW STEP SIZE ---

      Call CalcError(DtPsi,Psi,PsiDim,Error)
      If (CurrentOrder .LT. IntOrder) Then
         Error = Error*(2*IntOrder+1.0_dop)/(2*CurrentOrder+1.0_dop)
      EndIf
      Steps = Steps+1
      If (Error .GT. TolError) Then
         H(1) = 0.8_dop*H(1)*(IntError/Error)**(1.0_dop/(CurrentOrder+1))
         RepeatedSteps = RepeatedSteps+1
         StepIsRepeated = .True.
         Goto 100
      EndIf

! --- WRITE STEPSIZE AND ERROR ---

      Call WriteStep(Steps,CurrentOrder,H(1),Error,Time,lcheckH1)

! --- RETURN WHEN FINISHED ---

      Time = Time+H(1)
      If (Time .GE. AbsTime+IntPeriod-MinStepSize) Then
         Return
      EndIf

! --- EVALUATE FUNCTION WITH CORRECTED PSI ---

      Call Func(NextTime,Psi,DtPsi)

 200  Continue

! --- INCREASE INTEGRATION ORDER ---

      If (CurrentOrder .LT. IntOrder) Then
         CurrentOrder = CurrentOrder+1
      Else
         CurOrdEqIntOrd = .True.
      EndIf

! --- COMPUTE NEW DIVIDED DIFFERENCES ---

! As long as the order p is increased, the predictor must be used as a
! p-step method. Later on, the predictor is run as a (p+1)-step method.

      If (CurOrdEqIntOrd) Then
         Do D = 1,PsiDim
            AuxPsi(D,1) = Psi(D)
            InterimAuxPsi(1) = DtPsi(D)
            Do P = 2,CurrentOrder
               InterimAuxPsi(P) = (InterimAuxPsi(P-1)-AuxPsi(D,P))&
                                  /Distance(P-1)
            EndDo
            Do P = 2,CurrentOrder+1
               AuxPsi(D,P) = InterimAuxPsi(P-1)
            EndDo
         EndDo
      Else
         Do D = 1,PsiDim
            AuxPsi(D,1) = Psi(D)
            InterimAuxPsi(1) = DtPsi(D)
            Do P = 2,CurrentOrder-1
               InterimAuxPsi(P) = (InterimAuxPsi(P-1)-AuxPsi(D,P))&
                                  /Distance(P-1)
            EndDo
            Do P = 2,CurrentOrder
               AuxPsi(D,P) = InterimAuxPsi(P-1)
            EndDo
         EndDo
      EndIf

! --- CALCULATE NEW STEP SIZE ---

! If the previous step had been shortened (see above), the next stepsize
! may be set (close) to the stepsize before that one.

      Error = Max(Error,MinError)
      Do P = CurrentOrder-1,2,-1
         H(P) = H(P-1)
      EndDo
      H(1) = H(1)*(IntError/Error)**(1.0_dop/(CurrentOrder+1))
      lcheckH1 = .true.
      If (UseOldH1) Then
         If (H(1) .LT. 0.98_dop*H(3)) Then
            H(1) = 0.98_dop*H(3)
         EndIf
         UseOldH1 = .False.
      EndIf
      If (Time+H(1) .GT. AbsTime+IntPeriod) Then
         H(1) = AbsTime+IntPeriod-Time
         UseOldH1 = .True.
         lcheckH1 = .false.
      EndIf

! --- CONTINUE INTEGRATION ---

      Goto 100
      
      End Subroutine ABM

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE ABSABMERROR                     *
! *                                                                    *
! * Computes the absolute error of the current integration step of the *
! * Adams-Bashforth-Moulton predictor-corrector integrator.            *
! *                                                                    *
! * Input parameters:                                                  *
! *   Predicted: Predicted solution vector.                            *
! *   Corrected: Corrected soultion vector.                            *
! *   PsiDim:    Length of solution vectors.                           *
! *   Cdata:     Complex data array (not used in this error routine)   *
! *                                                                    *
! * Output parameters:                                                 *
! *   Error:     Computed error.                                       *
! *                                                                    *
! * V6.0 MB                                                            *
! * V7.0 GW addition of Data arrays                                    *
! *                                                                    *
! **********************************************************************

      Subroutine absabmerror (Predicted,Corrected,Psidim,Error)

      Implicit None

      Integer(long), intent(in)                   :: PsiDim
      Real(dop), intent(out)                      :: Error
      Complex(dop), dimension(PsiDim), intent(in) :: Predicted,Corrected

      Integer(long)       :: D
      Real(dop)           :: SingleError,Boundary

! --- INITIALISE VARIABLES ---

! If a component of the corrected solution is less than "Boundary" the
! absolute error is taken, else the relative error is computed.

      Boundary = 1.0_dop
      Error = 0.0_dop

! --- COMPUTE ERROR OF BOTH REAL AND IMAGINARY PART ---

      Do D = 1,PsiDim
         SingleError = DAbs(Dble(Predicted(D))-Dble(Corrected(D)))
         If (DAbs(Dble(Corrected(D))) .GT. Boundary) Then
            SingleError = SingleError/DAbs(Dble(Corrected(D)))
         EndIf
         Error = Max(Error,SingleError)
         SingleError = DAbs(DImag(Predicted(D))-DImag(Corrected(D)))
         If (DAbs(DImag(Corrected(D))) .GT. Boundary) Then
            SingleError = SingleError/DAbs(DImag(Corrected(D)))
         EndIf
         Error = Max(Error,SingleError)
      EndDo

      Return
      End Subroutine absabmerror

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE RELABMERROR                     *
! *                                                                    *
! * Computes the relative error of the current integration step of the *
! * Adams-Bashforth-Moulton predictor-corrector integrator.            *
! *                                                                    *
! * Input parameters:                                                  *
! *   Predicted: Predicted solution vector.                            *
! *   Corrected: Corrected soultion vector.                            *
! *   PsiDim:    Length of solution vectors.                           *
! *                                                                    *
! * Output parameters:                                                 *
! *   Error:     Computed error.                                       *
! *                                                                    *
! * V6.0 MB                                                            *
! * V7.0 MB addition of Data arrays                                    *
! *                                                                    *
! **********************************************************************

!     Subroutine RelABMError (Predicted,Corrected,Psidim,Error)

!     Implicit None

!     real(dop), parameter :: Tiny = 1.0e-30_dop

!     Integer(long), intent(in)                   :: PsiDim
!     Real(dop), intent(out)                      :: Error
!     Complex(dop), dimension(PsiDim), intent(in) :: Predicted,Corrected

!     Integer(long)       :: D
!     Real(dop)           :: Maximum

!!--- COMPUTE ERROR OF BOTH REAL AND IMAGINARY PART ---

!     Error = 0.0_dop
!     Maximum = Tiny
!     Do D = 1,PsiDim
!        Maximum = Max(Maximum,DAbs(Dble(Corrected(D))),&
!                      DAbs(DImag(Corrected(D))))
!        Error = Max(Error,DAbs(Dble(Predicted(D))-Dble(Corrected(D))),&
!                    DAbs(DImag(Predicted(D))-DImag(Corrected(D))))
!     EndDo
!     Error = Error/Maximum

!     Return
!     End Subroutine RelABMError

! **********************************************************************
! *                                                                    *
! *                        SUBROUTINE WRITEABMSTEP                     *
! *                                                                    *
! * Writes for the Adams-Bashforth-Moulton predictor-corrector         *
! * integrator the stepsize and error of the current integration step  *
! * to a file. (In the current form WriteABMStep is a dummy routine    *
! * doing absolutely nothing; it is included here for formal reasons,  *
! * but can (if desired) be extended easily such that it writes the    *
! * size and error of the current integration step to a file.          *
! *                                                                    *
! * Input parameters:                                                  *
! *   Step:     Counter for the number of steps.                       *
! *   Order:    Current integration order.                             *
! *   Stepsize: Size of current integration step.                      *
! *   Error:    Error of current integration step.                     *
! *                                                                    *
! * Output parameters:                                                 *
! *   none                                                             *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      Subroutine WriteABMStep (Step,Order,Stepsize,Error)

      Implicit None

      Integer(long), intent(in) :: Step,Order
      Real(dop), intent(in)     :: Stepsize,Error
      Integer(long)             :: dStep,dOrder
      Real(dop)                 :: dStepsize,dError

! --- WRITE STEPSIZE AND ERROR ---
! dummy steps to prevent warning on compilation

      dstep=step
      dorder=order
      dstepsize=stepsize
      derror=error


      Return
      End Subroutine WriteABMStep

! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE ABMERRORMSG                       *
! *                                                                    *
! * Generates for a given error number returned by "ABM" a             *
! * corresponding error message.                                       *
! *                                                                    *
! * Input parameters:                                                  *
! *   Error: Error code returned by ABM.                               *
! *                                                                    *
! * Output parameters:                                                 *
! *   Msg:   Error message.                                            *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine abmerrormsg (Error,Msg)

      Implicit None

      Integer(long), intent(in)     :: Error
      Character(len=*), intent(out) :: Msg

! --- GENERATE ERROR MESSAGE ---

      If (Error .Eq. 1) Then
         Msg = 'Illegal integration order'
      ElseIf (Error .Eq. 2) Then
         Msg = 'Stepsize underflow'
      Else
         Msg = 'Unknown error occurred'
      EndIf

      Return
      end subroutine abmerrormsg

      end module abmlib
