! **********************************************************************
! *                                                                    *
! *                          CHEBLIB (cheblib.f)                       *
! *                                                                    *
! * Program package containing subroutines that propagate wave-packets *
! * employing the chebyshev integration method.                        *
! *                                                                    *
! * CONTAINS:  STILL NEED TO EDIT HERE                                 *
! *    chebstep: Propagates a wave-packet using the second order       *
! *              differencing (SOD) scheme.                            *
! *    bessel:   Computes a stepsize for which the SOD-integrator is   *
! *              stable.                                               *
! *                                                                    *
! **********************************************************************


      module cheblib

      use wrintegrat, only: writestepcheb

      Implicit None

      integer, parameter       :: long  = selected_int_kind(9)
      integer, parameter       :: dop = selected_real_kind(15,307)

      private
      public                   :: chebstep


      contains
!---------------------------------------------------------------------------
!
!                            SUBROUTINE CHEBSTEP 
!
!     The Chebyshev integrator
!
! INPUT PARAMETERS:
!    psi:        Initial values of the dynamic variables
!    hpsi:       Initial time derivative of the dyn. vars
!    psidim = N: Number of dyn. vars
!    time:       Length of integration interval
!    E_max:      Estimate for value of max eigenvalue
!    E_min:      Estimate for value of min eigenvalue
!    accuracy:   Parameter controlling the accuracy
!    hamilton:   Routine for evaluating the RHS
!
! OUTPUT PARAMETERS:
!    psi:      Final values of the dyn. vars
!    hpsi:     Undetermined (i.e. it gets overwritten!)
!
!    bess:     Bessel coefficient
! Bessel function calculation uses a function BESSJ 
! CA, 2016
!
!***********************************************************************

      subroutine chebstep(psi,hpsi,time,psidim,E_max,E_min,hamilton,accuracy)

      Implicit None

      complex(dop), parameter  :: imag=(0.0_dop,1.0_dop)

      external                                       :: hamilton
      integer(long), intent(in)                      :: psidim
      complex(dop), dimension(psidim), intent(inout) :: psi, hpsi
      complex(dop), dimension(psidim)                :: cheby, cheby_save, cheby_old
      complex(dop), dimension(psidim)                :: psi_new
      real(dop), intent(in)                          :: time, accuracy
      real(dop), intent(in)                          :: E_min, E_max

      integer(long)            :: N
      complex(dop)             :: A 
      real(dop)                :: E_grid, bess
      real(dop)                :: alpha

      integer                  :: i
      integer                  :: niter
      real(dop)                :: err

! --- PROPAGATE WAVEFUNCTION ---

      E_grid = dabs(E_max - E_min) 
      alpha = 0.5_dop*E_grid*time
      N = CEILING(alpha) + 10
      A = exp(imag*(0.5_dop*E_grid + E_min)*time)
      
      i=-1
      do 
         i=i+1
! --- bess=bessel_jn(i,alpha) ---
         bess = BESSJ(i,alpha)

         if (i .eq. 0) then
            cheby_old=psi
            psi_new=bess*cheby_old
         else if (i .eq. 1) then
            CALL hamilton(time,psi,hpsi)

! --- NORMALISE Hpsi ---

            cheby=hpsi-(0.5_dop*E_grid+E_min)*psi
            cheby=2.0_dop*cheby/E_grid

! --- FORM CHEBYSCHEV TERM phi_1 ---

            cheby=-imag*cheby
            psi_new = psi_new + 2.0_dop*bess*cheby
         else
            cheby_save=cheby
            CALL hamilton(time,cheby_save,hpsi)

! --- NORMALISE Hphi_n ---

            cheby=hpsi-(0.5_dop*E_grid+E_min)*cheby_save
            cheby=2.0_dop*cheby/E_grid

! --- FORM CHEBYSCHEV TERM phi_n+1 ---

            cheby=-2.0_dop*imag*cheby + cheby_old
            psi_new=psi_new+2*bess*cheby
            cheby_old=cheby_save
         end if
      
         if (abs(bess) .lt. accuracy) then
            err = abs(bess)
            niter = i
            exit
         endif
      end do

! --- MULTIPLY BY PHASE FACTOR TO CORRECT FOR SCALING ---

      psi =A*psi_new

      call writestepcheb(niter,err,time)

      end subroutine chebstep

!---------------------------------------------------------------------------
!
!                            BESSEL FUNCTIONS 
!
! Program to calculate the first kind Bessel function of integer order N,    
!           for any REAL X, using the function BESSJ(N,X).            
!                                              
! SAMPLE RUN:                                                        
!    (Calculate Bessel function for N=2, X=0.75).                       
!    Bessel function of order  2 for X =  0.7500:                       
!         Y =  0.67073997E-01                                           
! REFERENCE: 
!    From Numath Library By Tuan Dang Trong in Fortran 77.    
!    F90 Release 1.0 By J-P Moreau, Paris.   
!        all variables declared           
!          (www.jpmoreau.fr) 
!    Upated for Quantics/E-CAM Library (K.E.Spinlove)
!--------------------------------------------------------------------------- 
!                            FUNCTION bessj
!
! PURPOSE:            
!    This subroutine calculates the first kind modified Bessel function
!    of integer order N, for any REAL X. We use here the classical
!    recursion formula, when X > N. For X < N, the Miller's algorithm
!    is used to avoid overflows. 
! REFERENCE: 
!    C.W.Clenshaw, Chebyshev series for mathematical functions,
!    Mathematical Tables, Vol.5, 1962.
!---------------------------------------------------------------------------

      function bessj (n,x)

      Implicit None
 
      integer, parameter       :: iacc = 40
      real(dop), parameter     :: bigno = 1.d10, bigni = 1.d-10
      
      real(dop)                :: x,bessj,tox,bjm,bj,bjp,sum
      
      integer                  :: m, n, j, jsum

      if (n.eq.0) then
         bessj = bessj0(x)
         return
      endif
      
      if (n.eq.1) then
         bessj = bessj1(x)
         return
      endif

      if (x.eq.0.) then
         bessj = 0.
         return
      endif

      tox = 2./x
      if (x.gt.float(n)) then
         bjm = bessj0(x)
         bj  = bessj1(x)
         do j = 1,n-1
            bjp = j*tox*bj-bjm
            bjm = bj
            bj  = bjp
         enddo   
         bessj = bj

      else
         m = 2*((n+int(sqrt(float(iacc*n))))/2)
         bessj = 0.
         jsum = 0
         sum = 0.
         bjp = 0.
         bj  = 1.
         do j = m,1,-1
            bjm = j*tox*bj-bjp
            bjp = bj
            bj  = bjm
            if (abs(bj).gt.bigno) then
               bj  = bj*bigni
               bjp = bjp*bigni
               bessj = bessj*bigni
               sum = sum*bigni
            endif
            if (jsum.ne.0) sum = sum+bj
            jsum = 1-jsum
            if (j.eq.n) bessj = bjp
         enddo
         sum = 2.*sum-bj
         bessj = bessj/sum
      endif

      return
      end function bessj

!---------------------------------------------------------------------------
!                            FUNCTION bessj0
!
! PURPOSE:
!    This subroutine calculates the First Kind Bessel Function of order 0,
!    for any real number X. The polynomial approximation by series of
!    Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
! REFERENCES:
!    1: M.Abramowitz,I.A.Stegun, Handbook Of Mathematical Functions, 1965.
!    2: C.W.Clenshaw, National Physical Laboratory Mathematical Tables,
!       Vol.5, 1962.
!---------------------------------------------------------------------------

      function bessj0 (x)

      Implicit None

      real(dop)                :: x,bessj0,ax,fr,fs,z,fp,fq,xx
      real(dop)                :: y,p1,p2,p3,p4,p5,r1,r2,r3,r4,r5,r6, &
                                  q1,q2,q3,q4,q5,s1,s2,s3,s4,s5,s6

      data p1,p2,p3,p4,p5 /1.d0,-.1098628627d-2,.2734510407d-4, &
      -.2073370639d-5,.2093887211d-6 /
      data q1,q2,q3,q4,q5 /-.1562499995d-1,.1430488765d-3, &
      -.6911147651d-5,.7621095161d-6,-.9349451520d-7 /
      data r1,r2,r3,r4,r5,r6 /57568490574.d0,-13362590354.d0, &
      651619640.7d0,-11214424.18d0,77392.33017d0,-184.9052456d0 /
      data s1,s2,s3,s4,s5,s6 /57568490411.d0,1029532985.d0, &
      9494680.718d0,59272.64853d0,267.8532712D0,1.d0 /

      if (x.eq.0.d0) goto 1
 
      ax = abs (x)
      if (ax.lt.8.) then
         y = x*x
         fr = r1+y*(r2+y*(r3+y*(r4+y*(r5+y*r6))))
         fs = s1+y*(s2+y*(s3+y*(s4+y*(s5+y*s6))))
         bessj0 = fr/fs
      else
         z = 8./ax
         y = z*z
         xx = ax-.785398164
         fp = p1+y*(p2+y*(p3+y*(p4+y*p5)))
         fq = q1+y*(q2+y*(q3+y*(q4+y*q5)))
         bessj0 = sqrt(.636619772/ax)*(fp*cos(xx)-z*fq*sin(xx))
      endif

      return
 1    bessj0 = 1.d0

      return
      end function

! --------------------------------------------------------------------------
!                            FUNCTION bessj1
!
! PURPOSE:
!    This subroutine calculates the First Kind Bessel Function of order 1,
!    for any real number X. The polynomial approximation by series of
!    Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
! REFERENCES:
!    1. M.Abramowitz,I.A.Stegun, Handbook Of Mathematical Functions, 1965.
!    2. C.W.Clenshaw, National Physical Laboratory Mathematical Tables,
!       Vol.5, 1962.
! --------------------------------------------------------------------------

      function bessj1 (x)

      Implicit None 

      real(dop)                :: x,bessj1,ax,fr,fs,z,fp,fq,xx
      real(dop)                :: y,p1,p2,p3,p4,p5,p6,r1,r2,r3,r4,r5,r6, &
                                  q1,q2,q3,q4,q5,s1,s2,s3,s4,s5,s6

      data p1,p2,p3,p4,p5 /1.d0,.183105d-2,-.3516396496d-4, &
      .2457520174d-5,-.240337019d-6 /,p6 /.636619772d0 /
      data q1,q2,q3,q4,q5 /.04687499995d0,-.2002690873d-3, &
      .8449199096d-5,-.88228987d-6,.105787412d-6 /
      data r1,r2,r3,r4,r5,r6 /72362614232.d0,-7895059235.d0, & 
      242396853.1d0,-2972611.439d0,15704.48260d0,-30.16036606d0 /
      data s1,s2,s3,s4,s5,s6 /144725228442.d0,2300535178.d0, &
      18583304.74d0,99447.43394d0,376.9991397d0,1.d0 /

      ax = abs(x)
      if (ax.lt.8.) then
         y = x*x
         fr = r1+y*(r2+y*(r3+y*(r4+y*(r5+y*r6))))
         fs = s1+y*(s2+y*(s3+y*(s4+y*(s5+y*s6))))
         bessj1 = x*(fr/fs)
      else
         z = 8./ax
         y = z*z
         xx = ax-2.35619491
         fp = p1+y*(p2+y*(p3+y*(p4+y*p5)))
         fq = q1+y*(q2+y*(q3+y*(q4+y*q5)))
         bessj1 = sqrt(p6/ax)*(cos(xx)*fp-z*sin(xx)*fq)*sign(s6,x)
      endif

      return
      end function

end module cheblib

