! **********************************************************************
! *                                                                    *
! *                  DAVIDSON DIAGONALISER (davlib.f)                  *
! *                                                                    *
! * Library module containing a Davidson diaginalisation routine       *
! * for improved relaxation.                                           *
! *                                                                    *
! * Contains:                                                          *
! *    DavStep:      The Davidson routine. Hermitian version           *
! *   cDavStep:      The Davidson routine. complex version             *
! *   DavErrorMsg:  Returns for a given error number a corresponding   *
! *                 error message.                                     *
! *                                                                    *
! **********************************************************************
      module davlib

use decimal, only: long, dop
use op1lib,  only: cpvxd, zerovxd,cpvxz
use timing
use runpropmod, only: edav,kskip

use wrintegrat
use getdiagonalmod
use getdavmatmod
use davutils
use channels
use pardsyevmod

use openmpmod

      implicit none
     
      private
      public :: davstep,rdavstep,cdavstep,daverrormsg,bdavstep,brdavstep
      contains


! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE DAVSTEP                                 *
! *                                                                    *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim).          *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStepDav").                                      *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors.               *
! *   HKrylov:   Matrix containing H*Krylov                            *
! *   rlxemin:   Lower energy to restrict the search for max. ovlp     *
! *   rlxemax:   Upper energy to restrict the search for max. ovlp     *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Time,Psi,DtPsi).                                 *
! *   WriteStep: Writes the stepsize and error to a file.              *
! *              Called as WriteStep(Steps,Order,Stepsize,Error).      *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 06/2003 HDM                                                   *
! * Modifications and OpenMP parallelisation. GWR 09/13                *
! **********************************************************************

      Subroutine DavStep(Psi,DtPsi,PsiDim,IntOrder,TolError,&
                       Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat,&
                       EigenVector,EigenVal,ovlps,&
                       TrueOrder,ErrorCode,Time,rlxemin,rlxemax,&
                       precon,quad,olsen,Func,nmod,lf)

      
      Implicit None

      External                                                   :: Func
      Logical(kind=4)                                            :: jump
      Logical(kind=4), save                                      :: full
      Logical(kind=4), intent(in)                                :: quad,olsen
      Integer(long), intent(in)                                  :: Steps
      Logical(kind=4), intent(in)                                :: lf
      Integer(long), intent(inout)                               :: rlaxnum
      Integer(long), intent(out)                                 :: ErrorCode,TrueOrder
      Integer(long), intent(in)                                  :: PsiDim,IntOrder,&
                                                                    nmod
      Integer(long), intent(inout)                               :: precon
      Integer(long), dimension(precon)                           :: indxdav
      Real(dop), intent(in)                                      :: TolError,Time,&
                                                                    rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps,EigenVal
      Complex(dop), dimension(PsiDim), intent(inout)             :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,IntOrder), intent(out)      :: Krylov
      Complex(dop), dimension(PsiDim,IntOrder), intent(inout)    :: HKrylov
      Complex(dop), dimension(2*PsiDim), intent(out)             :: HDiagonal
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: Tmat
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector
      Complex(dop), dimension(precon,precon+1)                   :: davmat
      Complex(dop)                                               :: cor1,cor2,ovlp

      Integer(long)                                         :: D,P,Q,Q1,i,j,shift,count
      Real(dop)                                             :: Error,olderr,ovl,&
                                                               ovl1,ovj,d2,E,E1,&
                                                               beta,beta1
      Real(dop), dimension(3*IntOrder+4)                    :: workr
      Complex(dop), dimension(16*IntOrder+18)               :: workc

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      Logical(kind=4), save :: FirstCall = .true.

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .LT. 2) Then
         ErrorCode = 1
         write(6,'(a,i4)') 'DavStep: illegal IntOrder =',IntOrder
         Return
      EndIf

! --- INITIALIZE VARIABLES ---

      if(FirstCall) Then
         if(rlaxnum .ge. 19999) then
            write(6,'(a)') 'ortho not allowed for DAV'
            stop
         endif
         if(rlaxnum .ge. 9999) then
            full = .true.
            rlaxnum = rlaxnum - 10000
         else
            full = .false.
         endif
         if(rlaxnum.lt.0 .or.(rlaxnum.gt.920 .and. rlaxnum.ne.930)) then
            write(6,'(a,i5,l3)') ' DAV : Incorrect rlaxnum',rlaxnum,full
            stop
         endif
      endif
      if(quad) then
         write(6,'(a)') ' quad not implemented for DAV, use RDAV'
         stop
      endif
      shift = 0
      count = 1
      d2    = 1.0e-18_dop
      jump  = .true. 
      ErrorCode = 0
      TrueOrder = 1
      Error = 1.0d0
      beta1 = 0
      Q  = 0
      Q1 = 0

!.....davovlp1 sets the second part of Hdiagonal to the transformed 
!.....A-vector of psi_ref.  (DavStep)
      if(rlaxnum.eq.930) then
         call davovlp1(psi,HDiagonal,time)
      endif

!.....Compute the diagonal of the Hamilton matrix (it is real).
      call getdiagonal(HDiagonal)
   
!.....Compute the first trial energy.
      E=0.0_dop
      Do D = 1, PsiDim
         E = E + dble(DConjg(Psi(D))*DtPsi(D))
      enddo

!.....Write very first energy expectation value to rlx_info file.
      if(FirstCall) then
         Q=0
         Q1=0
         E1=E
         ovl =1.0_dop
         ovl1=1.0_dop
         beta=1.0_dop
         call wrtrlxdav(time,Q1,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                        EigenVal,ovlps,FirstCall,IntOrder)
      endif

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
      if( precon .le. 1 ) precon=0
      if( precon .gt. 0 ) then
         call sortdiagonal(dble(HDiagonal),workr,indxdav,E,PsiDim,precon)
         call getdavmath(HDiagonal,davmat,davmat(1,precon+1),&
              indxdav,PsiDim,precon,2*precon-1)
      endif

! --- APPLY INVERSE OF DIAGONAL and ORTHONORMALIZE ---

      olderr=Error
      Error = 0.0_dop
      if( precon .eq. 0 ) then
         Do D = 1, PsiDim
            Error = Error + &
                 dble(DConjg(DtPsi(D)-E*Psi(D))*(DtPsi(D)-E*Psi(D)))
            Krylov(D,1) = (DtPsi(D)-E*Psi(D)) * &
                 dble(HDiagonal(D)-E)/(dble(HDiagonal(D)-E)**2+d2)
         enddo
      else
         Do D = 1, PsiDim
            Krylov(D,1) = DtPsi(D)-E*Psi(D)
         enddo
         call precondith(Krylov(1,1),HDiagonal,davmat, &
              davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error) 

      call orthon1(Psi,Krylov(1,1),ovl,PsiDim,.True.)
      call orthon1(Psi,Krylov(1,1),ovl,PsiDim,.True.)
!
! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
      beta=1.d0
      Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD HKRYLOV FOR FIRST STEP ---
      Call Func(Time,Krylov(1,1),HKrylov(1,1))

! --- BUILD T-MATRIX   (DavStep) ---
      
      Tmat(0,0) = E
      Tmat(1,1) = (0.0_dop,0.0_dop)
      Tmat(0,1) = (0.0_dop,0.0_dop)
      Do D = 1, PsiDim
         Tmat(0,1) = Tmat(0,1) +&
              DConjg(Psi(D))*Hkrylov(D,1)     
         Tmat(1,1) = Tmat(1,1) +&
              DConjg(Krylov(D,1))*HKrylov(D,1)     
      enddo

! --- DIAGONALISE T-MATRIX ---
      
      do j = 0, TrueOrder
         do i = 0, j
            EigenVector(i,j) = Tmat(i,j)
         enddo 
      enddo

      call ZHEEV( 'V', 'U', 2, EigenVector, IntOrder+1,&
                 EigenVal, workc, 16*IntOrder+18, workr, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf

      If(abs(EigenVector(0,0)) .gt. abs(EigenVector(0,1)) ) Then
         Q = 0
      else
         Q = 1 
      endif
      if(rlaxnum.eq.0)  Q = 0 
      if(FirstCall .and. rlaxnum.lt.TrueOrder .and. &
                         rlaxnum.le.900)  Q = rlaxnum 

      E = EigenVal(Q)
      beta = abs(EigenVector(0,Q))**2

      if(rlaxnum.eq.930 .and. .not.FirstCall) then
!........The call to davovlp2c sets "set" in davovlp2c.  
         call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov,&
                        EigenVector,Hdiagonal(PsiDim+1),FirstCall)
!........Add transformed initial state as extra vector.
         TrueOrder = TrueOrder + 1
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Hdiagonal(PsiDim+D)
         enddo
         Error = 1.0_dop
         goto 200
      endif



! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

 100  Continue  !  Iteration Loop (DavStep)
         
! --- RE-INITIALIZE VARIABLES ---

      TrueOrder = TrueOrder+1

!.....Converge to lower states if sought state already converged.
      if(full .and. (TrueOrder .lt. IntOrder-2)) then
         if(Error.lt.TolError*2.0_dop) then
            shift = shift + count
            if(shift .gt. Q) then
               count = -count
               shift = count
            endif
         endif
         Q1 = Q 
         Q  = Q - shift
         E  = EigenVal(Q)
         if(Error.lt.TolError*2.0_dop)  then 
            E1    = EigenVal(Q1)
            beta  = Abs(EigenVector(0,Q))**2
            beta1 = Abs(EigenVector(0,Q1))**2
            call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                           EigenVal,ovlps,FirstCall,IntOrder)
         endif
         Q1 = Q 
      endif

! --- Build Eigenvector ---

      Do D = 1, PsiDim
         Krylov(D,TrueOrder) = EigenVector(0,Q)*(DtPsi(D)-E*Psi(D))
      enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(j,D)
      Do D = 1, PsiDim
         Do j = 1, TrueOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) + &
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end parallel do

! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---

      olderr=Error
      if( precon.eq.0 ) then                          !Letitia
         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + &
                    dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) * &
               dble(HDiagonal(D)-E)/(dble(HDiagonal(D)-E)**2+d2)
         enddo
      else
          call precondith(Krylov(1,TrueOrder),HDiagonal,davmat, &
               davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error)

! --- OLSEN CORRECTION : cor1=<ev|(H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then
! --- Build EigenVector.
         Do D = 1, PsiDim
            workc(2*precon+D) = EigenVector(0,Q) * Psi(D)
         enddo
         Do j = 1, TrueOrder-1
            Do D = 1, PsiDim
               workc(2*precon+D) = workc(2*precon+D) + EigenVector(j,Q) * Krylov(D,j)
            enddo
         enddo
         cor1=(0.d0,0.d0)
         cor2=(0.d0,0.d0)
         if(precon.eq.0) then
            do D = 1, PsiDim
               cor1 = cor1+DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2 = cor2+DConjg(workc(2*precon+D))*workc(2*precon+D)* &
                    dble(HDiagonal(D)-E)/(dble(HDiagonal(D)-E)**2+d2)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * workc(2*precon+D) * &
                    dble(HDiagonal(D)-E)/(dble(HDiagonal(D)-E)**2+d2)
            enddo
         else
            call cpvxz(workc(2*precon+1),HKrylov(1,TrueOrder),PsiDim)
            call precondith(HKrylov(1,TrueOrder),HDiagonal,davmat, &
                 davmat(1,precon+1),indxdav,E,ovj,d2,PsiDim,precon)

            do D = 1, PsiDim
               cor1=cor1+DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2=cor2+DConjg(workc(2*precon+D))*HKrylov(D,TrueOrder)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * HKrylov(D,TrueOrder)
            enddo
         endif
      endif                    !Letitia check till here!
! ---  ORTHONORMALIZE (shortcut for loop) ---

 200  ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,TrueOrder)
      enddo
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*Psi(D)
      enddo

      Do i = 1, TrueOrder-1
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,TrueOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - ovlp*Krylov(D,i)
         enddo
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + &
               dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
      enddo
      if(ovl.le.1.d-39) then     ! If vectors are linearly dependent
         write(6,'(a,3i5,2e12.4)') 'DAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(2,'(a,3i5,2e12.4)') 'DAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         TrueOrder = TrueOrder - 1
         write(6,'(2a)') 'Linearly dependent Davidson vectors. ', &
                         'Convergence assumed'
         goto 310  ! Converged
      endif
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)/sqrt(ovl)
      enddo

!.....Symmetrize coefficients, if desired.
      if(nmod.gt.0) call csymcoeff(Krylov(1,TrueOrder), &
                    HKrylov(1,TrueOrder),PsiDim,nmod,lf,ovl,TrueOrder)

!.....Orthonormalize twice. (DavStep)
      ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,TrueOrder)
      enddo
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*Psi(D)
      enddo

      Do i = 1, TrueOrder-1
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,TrueOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - ovlp*Krylov(D,i)
         enddo
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + &
               dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
      enddo
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)/sqrt(ovl)
      enddo

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
      Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD HKRYLOV  ---
      Call Func(Time,Krylov(1,TrueOrder),HKrylov(1,TrueOrder))

! --- BUILD NEW T-MATRIX ELEMENTS ---
      
      Tmat(0,TrueOrder) = (0.0_dop,0.0_dop)
       Do D = 1, PsiDim
         Tmat(0,TrueOrder) = Tmat(0,TrueOrder) +&
              DConjg(Psi(D))*Hkrylov(D,TrueOrder)     
      enddo

!$omp parallel do if(lompthread.and.TrueOrder.ge.ompthread) num_threads(ompthread) private(j,D)
      Do j = 1, TrueOrder
         Tmat(j,TrueOrder) = (0.0_dop,0.0_dop)
         Do D = 1, PsiDim
            Tmat(j,TrueOrder) = Tmat(j,TrueOrder) +&
            DConjg(Krylov(D,j))*HKrylov(D,TrueOrder)     
         enddo
      enddo
!$omp end parallel do
        
! --- DIAGONALISE T-MATRIX ---
      
      do j = 0, TrueOrder
         do i = 0, TrueOrder
            EigenVector(i,j) = Tmat(i,j)
         enddo 
      enddo

      call ZHEEV( 'V', 'U', TrueOrder+1, EigenVector, IntOrder+1,&
                   EigenVal, workc, 16*IntOrder+18, workr, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf

! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if GS not sought) -

      if(rlaxnum.eq.0) then
         Q = 0 
         beta= abs(EigenVector(0,0))**2
      elseif(FirstCall .and. rlaxnum.lt.TrueOrder .and. &
                             rlaxnum.le.900)  then
         Q = rlaxnum 
         beta= abs(EigenVector(0,Q))**2
      else
         beta=0.0_dop
         Do j = 0, TrueOrder
            beta1= abs(EigenVector(0,j))**2
            If(beta1 .gt. beta) then
               Q = j
               beta = beta1
            endif
         enddo
      endif
      E = EigenVal(Q)
      Q1 = Q
      E1 = E
      beta1 = beta
      ovl=0.0_dop


! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If((TrueOrder.LT.IntOrder) .And.  (full .or.&
        ( (Error+olderr.GT.TolError*10.0_dop) .And. jump)) )  Goto 100


! --- COMPUTE EXACT OVERLAPS  (DavStep)

 310  if(rlaxnum.eq.930) then   ! lock
!........The array ovlps contains the overlaps**2 with startvector.
         call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov,&
                        EigenVector,Hdiagonal(PsiDim+1),FirstCall)
         Q=-1
         do j = 0, TrueOrder
            ovj = ovlps(j)
            if(j.eq.Q1) ovl1 = ovj
            if(EigenVal(j).lt.rlxemin .or. &
               EigenVal(j).gt.rlxemax) ovj=0.0_dop  
            If(ovj .gt. ovl) then
               Q = j
               ovl = ovj
            endif
         enddo
         if( Q .eq. -1 ) then
            E = 0.5_dop*(rlxemin+rlxemax)
            ovj = 1.0e50_dop
            do j = 0, TrueOrder
               if(abs(E-EigenVal(j))/(ovlps(j)+1.0e-4_dop) .lt. ovj) then
                  Q = j
                  ovj=abs(E-EigenVal(j))/(ovlps(j)+1.0e-4_dop)
               endif
            enddo
            ovl = ovlps(Q)
            E = EigenVal(Q)
            write(18,'(a,3f17.6)') '## No eigenvalue in window:',&
                 EigenVal(max(0,Q-1))*27.2114_dop, E*27.2114_dop,&
                 EigenVal(Q+1)*27.2114_dop
         endif
         E = EigenVal(Q)
         beta = abs(EigenVector(0,Q))**2
         If((Q1.ne.Q) .and. (TrueOrder.LT.IntOrder) .and. jump ) then
            jump = .false.  
            call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                           EigenVal,ovlps,FirstCall,IntOrder)
            goto 100
         endif
         If((TrueOrder.LT.IntOrder) .And. (full .or. &
                  (Error.GT.TolError)) ) Goto 100
      endif
      If( TrueOrder .EQ. IntOrder ) Goto 300


! --- Build Eigenvector ---

      TrueOrder = TrueOrder + 1
      Do D = 1, PsiDim
         Krylov(D,TrueOrder) = EigenVector(0,Q)*(DtPsi(D)-E*Psi(D))
      enddo
!$omp parallel do if(lompthread) num_threads(ompthread) private(D,j)
      Do D = 1, PsiDim
         Do j = 1, TrueOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) +&
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end parallel do

! --- DETERMINE NEW ERROR and APPLY INVERSE OF DIAGONAL ---

      olderr= Error
      Error = 0.0_dop
      Do D = 1, PsiDim
         Error = Error + dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
         Krylov(D,TrueOrder) = Krylov(D,TrueOrder) *&
         dble(HDiagonal(D)-E)/(dble(HDiagonal(D)-E)**2 + d2)
      enddo
      Error = Sqrt(Error)
 
      If( Error+olderr .GT. TolError ) Goto 200
 

! --- CONVERGENCE! Build new A-vector  (DavStep) ---

      TrueOrder = TrueOrder - 1
 300  Do D = 1,PsiDim
         Psi(D)   = EigenVector(0,Q)*Psi(D)
         DtPsi(D) = EigenVector(0,Q)*DtPsi(D)
      EndDo

!$omp parallel do if(lompthread) num_threads(ompthread) private(P,D)
      Do D = 1,PsiDim
         Do P = 1,TrueOrder
            Psi(D)   =   Psi(D)+EigenVector(P,Q)*Krylov(D,P)
            DtPsi(D) = DtPsi(D)+EigenVector(P,Q)*HKrylov(D,P)
         EndDo
      EndDo
!$omp end parallel do

      If(Q.ne.Q1) call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,&
                          ovl,ovl1,EigenVal,ovlps,.false.,IntOrder)


      if(FirstCall .and. rlaxnum.ne.930) then
!.....This is for first diag. wrtrlxdav
         do j = 0, TrueOrder
            ovlps(j) = Abs(EigenVector(0,j))**2
         enddo
      endif
      Q1 = Q
      call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                     EigenVal,ovlps,FirstCall,IntOrder)
      FirstCall = .false.
      full      = .false.
      
      Return
      End Subroutine DavStep                                  !  (DavStep)


! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE RDAVSTEP                                *
! *                                                                    *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming real Hamiltonian and real WF                              *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim).          *
! *   IntPeriod: Lenght of time interval to be integrated.             *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStep").                                         *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (real!)       *
! *   HKrylov:   Matrix containing H*Krylov (real!)                    *
! *   rlxemin:   Lower energy to restrict the search for max. ovlp     *
! *   rlxemax:   Upper energy to restrict the search for max. ovlp     *
! *   precon:    Dimension of Pre-Conditioner                          *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Time,Psi,DtPsi).                                 *
! * WriteStepDav: Writes the energy,beta,stepsize and error to a file. *
! *   Called as WriteStepDav(Steps,TrueOrder,E,qual,beta,Error,Q,Time) *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 06/2003 HDM                                                   *
! * Modifications and OpenMP parallelisation. GWR 09/13                *
! **********************************************************************

      Subroutine rDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError,&
                      Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat,Qmat,&
                      EigenVector,EigenVal,ovlps,Rdiagonal,davmat,&
                      TrueOrder,ErrorCode,Time,rlxemin,&
                      rlxemax,precon,quad,olsen,Func,Funcr,lrreal,nmod,lf)

!$ use omp_lib


      Implicit None

      Integer(long), Parameter  :: mxprcn=5000 ! If changed, change also in einrun.

      External                                                        :: Func,Funcr
      Logical(kind=4)                                                 :: jump,runquad
      Logical(kind=4), save                                           :: full
      Logical(kind=4), intent(in)                                     :: quad,olsen,lrreal,lf
      Integer(long), intent(in)                                       :: Steps
      Integer(long), intent(inout)                                    :: rlaxnum,precon,nmod
      Integer(long), intent(out)                                      :: ErrorCode,TrueOrder
      Integer(long), intent(in)                                       :: PsiDim,IntOrder
      Real(dop), intent(in)                                           :: TolError,Time,rlxemin,rlxemax
      Real(dop), dimension(:), allocatable                            :: workr
      Real(dop), dimension(0:IntOrder), intent(inout)                 :: ovlps,EigenVal
      Real(dop), dimension(PsiDim,0:IntOrder), intent(out)            :: Krylov
      Real(dop), dimension(PsiDim,0:IntOrder), intent(out)            :: HKrylov
      Real(dop), dimension(0:IntOrder,0:IntOrder), intent(out)        :: EigenVector
      Real(dop), dimension(PsiDim), intent(out)                       :: RDiagonal
      Real(dop), dimension(((IntOrder+1)*(IntOrder+2))/2), intent(out):: Tmat,Qmat
      Real(dop), dimension(precon,precon+1), intent(out)              :: davmat
      Complex(dop), dimension(PsiDim), intent(inout)                  :: Psi,DtPsi
      Complex(dop), dimension(2*PsiDim), intent(out)                  :: HDiagonal

      Integer(long) :: D,P,Q,Q1,i,j,ij,shift,count,lwork
      Integer(long), dimension(precon)                      :: indxdav
      Real(dop)     :: Error,ovl,ovl1,ovj,d2,E,E1,beta,beta1,&
                       DE,olderr,cor1,cor2
      Real(dop), save     :: edav
      integer(long), save                       :: tid=0 ! timer ID

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      Logical(kind=4), save :: FirstCall = .true.

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .LT. 2) Then
         ErrorCode = 1
         write(6,'(a,i4)') 'rDavStep: IntOrder =',IntOrder
         Return
      EndIf

! --- INITIALIZE VARIABLES ---

      if(FirstCall) Then
         if(rlaxnum .ge. 9999) then
            full = .true.
            rlaxnum = rlaxnum - 10000
         else
            full = .false. 
         endif 
         if(rlaxnum.lt.0 .or.(rlaxnum.gt.920 .and. rlaxnum.ne.930)) then
            write(6,'(a,i5,l3)') ' DAV : Incorrect rlaxnum',rlaxnum,full
            stop
         endif
      endif
      Q     = 0
      Q1    = 0
      shift = 0
      count = 1
      d2    = 1.0e-18_dop
      jump  = .true. 
      ErrorCode = 0
      TrueOrder = 1
      beta1 = 0
      Error = 1.0_dop
      lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
      runquad = quad.and.(.not.full) 

      allocate(workr(lwork))

!.....Compute the diagonal of the Hamilton matrix.
      call getdiagonal(HDiagonal)

!.....Copy complex to real arrays and compute the first trial energy.
      cor1=0.0_dop
      E=0.0_dop
      Do D = 1, PsiDim
         Krylov(D,0)  = dble(  Psi(D))
         HKrylov(D,0) = dble(DtPsi(D))
         Rdiagonal(D) = dble(HDiagonal(D))
         E = E + Krylov(D,0)*HKrylov(D,0)
         cor1=cor1 + abs(dimag(Psi(D))) + abs(dimag(DtPsi(D)))
      enddo
      edav = E

      If(cor1.gt.1.d-11) call realdaverror(cor1)

!.....Write very first energy expectation value to rlx_info file.
      if(FirstCall) then
         E1=E
         ovl =1.0_dop
         ovl1=1.0_dop
         beta=1.0_dop
         call wrtrlxdav(time,Q1,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                        EigenVal,ovlps,FirstCall,IntOrder)
      endif

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
      if( precon .le. 1 ) precon=0
      if( precon .gt. 0 ) then
         call get_timer(tid, 'Precon')
         call start_timer(tid)

         call sortdiagonal(RDiagonal,workr,indxdav,E,PsiDim,precon)
         call getdavmat(RDiagonal,davmat,davmat(1,precon+1),&
                        indxdav,PsiDim,precon,lwork)
         call stop_timer(tid)
      endif

! --- APPLY INVERSE OF DIAGONAL and ORTHONORMALIZE ---

      olderr=Error
      if( precon .eq. 0 ) then
         Error = 0.0_dop
         Do D = 1, PsiDim
            Error = Error + ( HKrylov(D,0)-E*Krylov(D,0) )**2
            Krylov(D,1) = (HKrylov(D,0)-E*Krylov(D,0)) * &
                 (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
         enddo
      else
         Do D = 1, PsiDim
            Krylov(D,1) = HKrylov(D,0)-E*Krylov(D,0) 
         enddo
         call precondit(Krylov(1,1),RDiagonal,davmat,davmat(1,precon+1),&
              indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error) 
      if(Error.lT.0.2*TolError) then ! Startvector is already eigenvector
         beta=1.0_dop
         ovl=0.0_dop
         call wrtrlxdav(time,trueorder,Q,Q,E,E1,beta,beta1,ovl,   &
                        ovl1,EigenVal,ovlps,FirstCall,IntOrder)
         return
      endif

      call orthonr(Krylov,Krylov(1,1),ovl,0,PsiDim)
!.....Do the orthonormalisation twice.
      call orthonr(Krylov,Krylov(1,1),ovl,0,PsiDim)

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
      beta=1.0_dop
      Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD HKRYLOV FOR FIRST STEP ---

      If(lrreal) Then
         Call Funcr(Time,Krylov(1,1),HKrylov(1,1))
      Else
         Do D = 1,PsiDim
            HDiagonal(D) = Krylov(D,1)
         enddo
         Call Func(Time,HDiagonal,DtPsi)
         Do D = 1,PsiDim
            HKrylov(D,1) = dble(DtPsi(D))
         enddo
      Endif

! --- BUILD T-MATRIX  ---
      
      Tmat(1) = E
      Tmat(2) = 0.0_dop
      Tmat(3) = 0.0_dop
      Do D = 1, PsiDim
         Tmat(2) = Tmat(2) + Krylov(D,0)*HKrylov(D,1)     
         Tmat(3) = Tmat(3) + Krylov(D,1)*HKrylov(D,1)     
      enddo
      if(runquad) then
         Qmat(1) = 0.0_dop
         Qmat(2) = 0.0_dop
         Qmat(3) = 0.0_dop
         Do D = 1, PsiDim
            Qmat(1) = Qmat(1) + HKrylov(D,0)*HKrylov(D,0)
            Qmat(2) = Qmat(2) + HKrylov(D,0)*HKrylov(D,1)
            Qmat(3) = Qmat(3) + HKrylov(D,1)*HKrylov(D,1)
         enddo
      endif

! --- DIAGONALISE T-MATRIX ---
      ij = 0
      do j = 0, TrueOrder
         do i = 0, j
            ij = ij + 1
            EigenVector(i,j) = Tmat(ij)
         enddo 
      enddo

      call pardsyev('V','U',TrueOrder+1,EigenVector,IntOrder+1,EigenVal,&
                  workr,lwork,ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workr)
         Return
      EndIf

      If(abs(EigenVector(0,0)) .gt. abs(EigenVector(0,1)) ) Then
         Q = 0
      else
         Q = 1 
      endif
      if(rlaxnum.eq.0)  Q = 0 
      if(FirstCall .and. rlaxnum.lt.TrueOrder .and. &
                         rlaxnum.le.900)  Q = rlaxnum 

      E  = EigenVal(Q)
      beta = EigenVector(0,Q)**2


!.....davovlp1 sets the second part of Hdiagonal to the transformed 
!.....A-vector of psi_ref. The call to davovlp2r sets "set" in davovlp2r. 
      if(rlaxnum.eq.930 .and. .not.FirstCall) then
         call davovlp1(psi,HDiagonal,time)
         call davovlp2r(IntOrder,TrueOrder,ovlps,Krylov,EigenVector,&
                        Hdiagonal(PsiDim+1),FirstCall)
!........Add transformed initial state as extra vector.
         TrueOrder = TrueOrder + 1
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = dble(Hdiagonal(PsiDim+D))
         enddo
         Error = 1.0_dop    ! Dummy Error for lock vector.
         goto 200
      endif


! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

 100  Continue  !  Iteration Loop
         
! --- RE-INITIALIZE VARIABLES ---

      TrueOrder = TrueOrder+1

!.....Converge to lower states if sought state already converged.
      if(full .and. (TrueOrder .lt. IntOrder-2)) then
         if(Error.lt.TolError*2.0_dop) then
            shift = shift + count
            if(shift .gt. Q) then
               count = -count
               shift = count
            endif
         endif
         Q1 = Q 
         Q  = Q - shift
         E  = EigenVal(Q)
         if(Error.lt.TolError*2.0_dop)  then 
            E1    = EigenVal(Q1)
            beta  = EigenVector(0,Q)**2
            beta1 = EigenVector(0,Q1)**2
            call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,&
                           ovl1,EigenVal,ovlps,FirstCall,IntOrder)
         endif
         Q1 = Q 
      endif

! --- Build residual vector |r> ---

      Do D = 1, PsiDim
         Krylov(D,TrueOrder) = &
                 EigenVector(0,Q) * (HKrylov(D,0) - E*Krylov(D,0))
      enddo

!$omp parallel num_threads(ompthread) if(lompthread.and.(trueorder*psidim).ge.2000) 
!$omp do private(D,j)
      Do D = 1, PsiDim
         Do j = 1, TrueOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) + &
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end do
!$omp end parallel

! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      olderr=Error
      if( precon.eq.0 ) then
         Error = 0.0_dop
         Do D = 1, PsiDim
            Error = Error + Krylov(D,TrueOrder)**2
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder)*&
                 (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
         enddo
      else
         call precondit(Krylov(1,TrueOrder),RDiagonal,davmat,&
                        davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error)

! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>    
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then 
! --- Build EigenVector.
         Do D = 1, PsiDim
            workr(2*precon+D) = EigenVector(0,Q) * Krylov(D,0)
         enddo
!$omp parallel num_threads(ompthread) if(lompthread.and.(trueorder*psidim).ge.2000) 
!$omp do private(D,j)
         Do D = 1, PsiDim
            Do j = 1, TrueOrder-1
               workr(2*precon+D) = workr(2*precon+D) +& 
                    EigenVector(j,Q) * Krylov(D,j)
            enddo
         enddo
!$omp end do
!$omp end parallel

         cor1=0.0_dop     
         cor2=0.0_dop
         if(precon.eq.0) then
            do D = 1, PsiDim
               cor1 = cor1 + workr(2*precon+D)*Krylov(D,TrueOrder)
               cor2 = cor2 + workr(2*precon+D)**2 * &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * workr(2*precon+D) *&
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo

         else
            call cpvxd(workr(2*precon+1),HKrylov(1,TrueOrder),PsiDim)
            call precondit(HKrylov(1,TrueOrder),RDiagonal,davmat,&
                           davmat(1,precon+1),indxdav,E,ovj,d2,PsiDim,precon)
            do D = 1, PsiDim
               cor1 = cor1 + workr(2*precon+D)*Krylov(D,TrueOrder)
               cor2 = cor2 + workr(2*precon+D)*HKrylov(D,TrueOrder)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * HKrylov(D,TrueOrder)
            enddo
         endif
      endif

! ---  ORTHONORMALIZE (entry for adding "lock" vector) ---
 
 200  continue
      If((TrueOrder .GE. IntOrder-9) .or.  &
         (Error .LT. 2.0_dop*TolError) ) runquad = .false.

      call orthonr(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)

      if(ovl.le.1.0e-39_dop) then     ! If vectors are linearly dependent
         write(6,'(a,3i5,2e12.4)') 'DAV: Steps, Order, Q, ovl, Error',&
                                    Steps,TrueOrder,Q,ovl,Error
         write(2,'(a,3i5,2e12.4)') 'DAV: Steps, Order, Q, ovl, Error',&
                                    Steps,TrueOrder,Q,ovl,Error
         TrueOrder = TrueOrder - 1
         write(6,'(2a)') 'Linearly dependent Davidson vectors. ',   &
                         'Convergence assumed'
         goto 310  ! Converged
      endif

!.....Symmetrize coefficients, if desired.
      if(nmod.gt.0) call rsymcoeff(Krylov(1,TrueOrder),   &
                    HKrylov(1,TrueOrder),PsiDim,nmod,lf,ovl,TrueOrder)

!.....Do the orthonormalisation twice
      call orthonr(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)

! --- BUILD HKRYLOV  ---
      If(lrreal) Then
         Call Funcr(Time,Krylov(1,TrueOrder),HKrylov(1,TrueOrder))
      Else
         Do D = 1,PsiDim
            HDiagonal(D) = Krylov(D,TrueOrder)
         enddo
         Call Func(Time,HDiagonal,DtPsi)
         Do D = 1,PsiDim
            HKrylov(D,TrueOrder) = dble(DtPsi(D))
         enddo
      EndIf

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
      Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD NEW T-MATRIX ELEMENTS ---

      Do j = 0, TrueOrder
         ij = ij + 1
         Tmat(ij) = 0.0_dop
         if(runquad) then
            Qmat(ij) = 0.0_dop
            Do D = 1, PsiDim
               Tmat(ij) = Tmat(ij) +  Krylov(D,j)*HKrylov(D,TrueOrder)
               Qmat(ij) = Qmat(ij) + HKrylov(D,j)*HKrylov(D,TrueOrder)
            enddo
         else
            Do D = 1, PsiDim
               Tmat(ij) = Tmat(ij) + Krylov(D,j)*HKrylov(D,TrueOrder) 
            enddo
         endif
      enddo

        
! --- DIAGONALISE T-MATRIX ---
      ij = 0
      do j = 0, TrueOrder
         if(runquad) then
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Qmat(ij)-2.0_dop*E*Tmat(ij)
            enddo
            EigenVector(j,j) = EigenVector(j,j) + E*E
         else
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         endif
      enddo

      call pardsyev('V','U',TrueOrder+1,EigenVector,IntOrder+1,EigenVal,&
                     workr,lwork,ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workr)
         Return
      EndIf


! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if GS not sought) -

      if(rlaxnum.eq.0) then
         Q = 0 
         beta= EigenVector(0,0)**2
      elseif(FirstCall .and. rlaxnum.lt.TrueOrder .and. &
                             rlaxnum.le.900)  then
         Q = rlaxnum 
         beta= EigenVector(0,Q)**2
      else
         beta=0.0_dop
         Do j = 0, TrueOrder
            beta1= EigenVector(0,j)**2
            If(beta1 .gt. beta) then
               Q = j
               beta = beta1
            endif
         enddo
      endif
      if(runquad) then
         E = 0.0_dop
         DE = 0
         ij = 0
         do j = 0, TrueOrder
            do i = 0, j
               ij = ij + 1
               DE = EigenVector(i,Q)*Tmat(ij)*EigenVector(j,Q)
               E = E + 2.0_dop*DE
            enddo
            E = E - DE
         enddo
      else
         E = EigenVal(Q)
      endif
      Q1 = Q
      E1 = E
      beta1 = beta
      ovl=0.0_dop


! --- CONTINUE IF NOT CONVERGED --- END OF LOOP if not lock ---
      If( TrueOrder .EQ. PsiDim-1 ) Goto 310
      If((TrueOrder.LT.IntOrder) .And.  (full .or.&
        ( (Error.GT.TolError*10.0_dop) .And. jump)) )  Goto 100


! --- COMPUTE EXACT OVERLAPS

 310  if(rlaxnum.eq.930) then   ! lock
!........The array ovlps contains the overlaps**2 with startvector.
         call davovlp2r(IntOrder,TrueOrder,ovlps,Krylov,EigenVector,&
                        Hdiagonal(PsiDim+1),FirstCall)
         Q=-1
         do j = 0, TrueOrder
            ovj = ovlps(j)
            if(j.eq.Q1) ovl1 = ovj
            if(EigenVal(j).lt.rlxemin .or. &
               EigenVal(j).gt.rlxemax) ovj=0.0_dop  
            If(ovj .gt. ovl) then
               Q = j
               ovl = ovj
            endif
         enddo
         if( Q .eq. -1 ) then
            E = 0.5_dop*(rlxemin+rlxemax)
            ovj = 1.0e50_dop
            do j = 0, TrueOrder
               if(abs(E-EigenVal(j))/(ovlps(j)+1.0e-4_dop) .lt. ovj) then
                  Q = j
                  ovj=abs(E-EigenVal(j))/(ovlps(j)+1.0e-4_dop)
               endif
            enddo
            ovl = ovlps(Q)
            E = EigenVal(Q)
            write(18,'(a,3f17.6)') '## No eigenvalue in window:', &
                 EigenVal(max(0,Q-1))*27.2114_dop, E*27.2114_dop, &
                 EigenVal(Q+1)*27.2114_dop
         endif
         if(runquad) then
            E = 0.0_dop
            DE = 0
            ij = 0
            do j = 0, TrueOrder
               do i = 0, j
                  ij = ij + 1
                  DE = EigenVector(i,Q)*Tmat(ij)*EigenVector(j,Q)
                  E = E + 2.0_dop*DE
               enddo
               E = E - DE
            enddo
         else
            E = EigenVal(Q)
         endif
         beta = EigenVector(0,Q)**2
         If((Q1.ne.Q) .and. (TrueOrder.LT.IntOrder) .and. jump ) then
            jump = .false.  
            call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,&
                           ovl1,EigenVal,ovlps,FirstCall,IntOrder)
            goto 100
         endif
      endif   ! end of lock

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If( TrueOrder .EQ. IntOrder ) Goto 300
      If( runquad .or. full .or. (Error+olderr.GT.TolError) ) Goto 100


! --- CONVERGENCE! Build new A-vector ---

 300  Do D = 1,PsiDim 
         Psi(D) = Krylov(D,0)*EigenVector(0,Q)
         DtPsi(D) = HKrylov(D,0)*EigenVector(0,Q)
      EndDo

!$omp parallel do if(lompthread) num_threads(ompthread) private(D,P)
      Do D = 1,PsiDim
         Do P = 1,TrueOrder
            Psi(D) = Psi(D) + Krylov(D,P)*EigenVector(P,Q)
            DtPsi(D) = DtPsi(D) + HKrylov(D,P)*EigenVector(P,Q)
         EndDo
      EndDo
!$omp end parallel do


      If(Q.ne.Q1) call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,&
                          ovl,ovl1,EigenVal,ovlps,.false.,IntOrder)

      if(FirstCall .and. (rlaxnum.ne.930)) then
!.....This is for first overlaps in wrtrlxdav
         do j = 0, TrueOrder
            ovlps(j) = EigenVector(0,j)**2
         enddo
      endif
      Q1 = Q
      call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,ovl1,&
                     EigenVal,ovlps,FirstCall,IntOrder)

      FirstCall = .false.
      full      = .false.
      edav = E

      deallocate(workr)

      Return
      End Subroutine rDavStep

! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE cDAVSTEP                                *
! *                                                                    *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim).          *
! *   IntPeriod: Lenght of time interval to be integrated.             *
! *   IntOrder:  Maximum iteration order.                              *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStep").                                         *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors.               *
! *   HKrylov:   Matrix containing H*Krylov                            *
! *   rlxemin:   Lower energy to restrict the search for max. ovlp     *
! *   rlxemax:   Upper energy to restrict the search for max. ovlp     *
! *   rlximax:   Lower Im(energy) to restrict the s. for max. ovlp     *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Time,Psi,DtPsi).                                 *
! *   WriteStep: Writes the stepsize and error to a file.              *
! *              Called as WriteStep(Steps,Order,Stepsize,Error).      *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 07/2003 HDM                                                   *
! * Modifications and OpenMP parallelisation. GWR 09/13                *
! * 01/2014 preconditioner added by Letitia                            *
! **********************************************************************

      Subroutine cDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError,&
                          Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat,&
                          EigenVector,EigenVal,ovlps,&
                          TrueOrder,ErrorCode,Time,rlxemin,&
                          rlxemax,rlximin,precon,quad,olsen,Func)

      Implicit None

      External                                                   :: Func
      Logical(kind=4)                                            :: jump
      Logical(kind=4), save                                      :: full
      Logical(kind=4), intent(in)                                :: quad,olsen
      Integer(long), intent(in)                                  :: Steps
      Integer(long), intent(inout)                               :: rlaxnum
      Integer(long), intent(out)                                 :: ErrorCode,&
                                                                    TrueOrder
      Integer(long), intent(in)                                  :: PsiDim,&
                                                                    IntOrder
      Integer(long), intent(inout)                               :: precon
      Real(dop), intent(in)                                      :: TolError,&
                                                                    Time,rlxemin,&
                                                                    rlxemax,rlximin
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps
      Complex(dop), dimension(PsiDim), intent(inout)             :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,IntOrder), intent(out)      :: Krylov
      Complex(dop), dimension(PsiDim,IntOrder), intent(inout)    :: HKrylov
      Complex(dop), dimension(2*PsiDim), intent(out)             :: HDiagonal
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: Tmat
      Complex(dop), dimension(0:IntOrder,0:IntOrder)             :: TTmat
      Complex(dop), dimension(precon,precon+1)                   :: davmat
      Complex(dop), dimension(precon,precon)                     :: leigf,reigf

      Integer(long)                                         :: D,P,Q,Q1,oldQ,&
                                                               i,j,shift,count
      Integer(long), dimension(precon)                      :: indxdav
      Real(dop)                                             :: Error,ovl,ovl1,&
                                                               ovj,d2,beta,beta1,&
                                                               Er,Er1,olderr
      Real(dop), dimension(2*IntOrder+4)                    :: workr
      Complex(dop)                                          :: E,E1,ovlp,cor1,cor2
      Complex(dop), dimension(16*IntOrder+18)               :: workc
      Complex(dop), dimension(0:IntOrder), intent(inout)    :: EigenVal
      Complex(dop), dimension(1,1)                          :: vl

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      Logical(kind=4), save :: FirstCall = .true.

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .LT. 2) Then
         ErrorCode = 1
         write(6,'(a,i4)') 'IntOrder =',IntOrder
         Return
      EndIf

! --- INITIALIZE VARIABLES ---

      if(FirstCall) Then
         if(rlaxnum .ge. 19999) then
            write(6,'(a)') 'ortho not allowed for DAV'
            stop
         endif
         if(rlaxnum .ge. 9999) then
            full = .true.
            rlaxnum = rlaxnum - 10000
         else
            full = .false.
         endif
         if(rlaxnum.lt.0 .or.(rlaxnum.gt.920 .and. rlaxnum.ne.930)) then
            write(6,'(a,i5,l3)') ' DAV : Incorrect rlaxnum',rlaxnum,full
            stop
         endif
      endif
      if(quad) then
         write(6,'(a)') ' quad not implemented for CDAV'
         stop
      endif
      shift = 0
      count = 1
      d2 = 1.0e-18_dop
      jump = .true. 
      ErrorCode = 0
      TrueOrder = 1
      Error = 1.d0
      beta1 = 0

!.....davovlp1 sets the second part of Hdiagonal to the transformed 
!.....A-vector of psi_ref. (cDavStep) 
      if(rlaxnum.eq.930) then
         call davovlp1(psi,HDiagonal,time)
      endif

!.....Compute the diagonal of the Hamilton matrix.
      call getdiagonal(HDiagonal)
   
!.....Compute the first trial energy.
      E=0.0_dop
      Do D = 1, PsiDim
         E = E + DConjg(Psi(D))*DtPsi(D)
      enddo

!.....Write very first energy expectation value to rlx_info file.
      if(FirstCall) then
         Q=0
         Q1=0
         E1=E
         ovl=1.0_dop
         ovl1=1.0_dop
         beta=1.0_dop
         call wrtrlxcdav(time,Q1,Q,Q1,E,E1,beta,beta,ovl,ovl1,&
                           EigenVal,ovlps,FirstCall,IntOrder)
      endif

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
      if( precon .le. 1 ) precon=0
      if( precon .gt. 0 ) then
         call sortdiagonalz(HDiagonal,workr,indxdav,E,PsiDim,precon)
         call getdavmatz(HDiagonal,davmat,leigf,reigf, &
              davmat(1,precon+1),indxdav,PsiDim,precon, &
              2*precon)
      endif

! --- APPLY INVERSE OF DIAGONAL and ORTHONORMALIZE ---

      olderr=Error
      Error = 0.0_dop

      Error = Sqrt(Error) 

      if( precon .eq. 0 ) then
         Do D = 1, PsiDim
            Error = Error + dble(DConjg(DtPsi(D)-E*Psi(D))* &
                                  (DtPsi(D)-E*Psi(D)))
            Krylov(D,1) = (DtPsi(D)-E*Psi(D)) * DConjg &
            (HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E)* &
                                (HDiagonal(D)-E)+d2)
         enddo
      else
         Do D = 1, PsiDim
            Krylov(D,1) = DtPsi(D)-E*Psi(D)
         enddo
         call preconditz(Krylov(1,1),HDiagonal,leigf,reigf, &
              davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error)
      ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,1)
      enddo
      Do D = 1,PsiDim
         Krylov(D,1)=Krylov(D,1)-ovlp*Psi(D)
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + dble(DConjg(Krylov(D,1))*Krylov(D,1))
      enddo
      Do D = 1,PsiDim
         Krylov(D,1)=Krylov(D,1)/sqrt(ovl)
      enddo
!.....Orthonormalize twice.
      ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,1)
      enddo
      Do D = 1,PsiDim
         Krylov(D,1)=Krylov(D,1)-ovlp*Psi(D)
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + dble(DConjg(Krylov(D,1))*Krylov(D,1))
      enddo
      Do D = 1,PsiDim
         Krylov(D,1)=Krylov(D,1)/sqrt(ovl)
      enddo

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
      beta=1.d0
      Call WriteStepCDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD HKRYLOV FOR FIRST STEP ---
      Call Func(Time,Krylov(1,1),HKrylov(1,1))

! --- BUILD T-MATRIX (cDavStep) ---
      
      Tmat(0,0) = E
      Tmat(0,1) = (0.0_dop,0.0_dop)
      Tmat(1,0) = (0.0_dop,0.0_dop)
      Tmat(1,1) = (0.0_dop,0.0_dop)
      Do D = 1, PsiDim
         Tmat(1,0) = Tmat(1,0) +&
              DConjg(Krylov(D,1))*DtPsi(D)     
         Tmat(0,1) = Tmat(0,1) +&
              DConjg(Psi(D))*HKrylov(D,1)     
         Tmat(1,1) = Tmat(1,1) +&
              DConjg(Krylov(D,1))*HKrylov(D,1)     
      enddo

! --- DIAGONALISE T-MATRIX ---
      
      do j = 0, TrueOrder
         do i = 0, TrueOrder
            TTmat(i,j) = Tmat(i,j)
         enddo 
      enddo

      call ZGEEV( 'N', 'V', TrueOrder+1, TTmat, IntOrder+1,&
                   EigenVal, vl, 1, EigenVector, IntOrder+1, &
                   workc, 16*IntOrder+18, workr, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf

      If(abs(EigenVector(0,0)) .gt. abs(EigenVector(0,1)) ) Then
         Q = 0
      else
         Q = 1 
      endif
      if(rlaxnum.eq.0) then
         Q = 0    ! Search for lowest energy (real + abs(im))
         Er=dble(EigenVal(0)) + abs(dimag(EigenVal(0)))
         Do j = 1, TrueOrder
            Er1=dble(EigenVal(j)) + abs(dimag(EigenVal(j)))
            if(Er1.lt.Er) then
               Q=j
               Er=Er1
            endif
         enddo
      endif

      E = EigenVal(Q)
      beta = abs(EigenVector(0,Q))**2

      if(rlaxnum.eq.930 .and. .not.FirstCall) then
!........The call to davovlp2r sets "set" in davovlp2c.  
         call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov,&
                        EigenVector,Hdiagonal(PsiDim+1),FirstCall)
!........Add transformed initial state as extra vector.
         TrueOrder = TrueOrder + 1
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Hdiagonal(PsiDim+D)
         enddo
         Error = 1.0_dop    ! Dummy Error for lock vector.
         goto 200
      endif

! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

 100  Continue  !  Iteration Loop  (cDavStep)
         
! --- RE-INITIALIZE VARIABLES ---

      TrueOrder = TrueOrder+1

!.....Converge to lower states if seeked state already converged.
      if(full .and. (TrueOrder .lt. IntOrder-2)) then
         if(Error.lt.TolError*2.0_dop) then
            shift = shift + count
            if(shift .gt. Q) then
               count = -count
               shift = count
            endif
         endif
         Q1 = Q
         Q  = Q - shift
         E  = EigenVal(Q)
         if(Error.lt.TolError*2.0_dop)  then
            E1    = EigenVal(Q1)
            beta  = dble(EigenVector(0,Q))**2
            beta1 = dble(EigenVector(0,Q1))**2
            call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1,&
                     ovl,ovl1,EigenVal,ovlps,FirstCall,IntOrder)
         endif
         Q1 = Q
      endif

! --- Build residual vector |r> ---

      Do D = 1, PsiDim
         Krylov(D,TrueOrder) = EigenVector(0,Q)*(DtPsi(D)-E*Psi(D))
      enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(j,D)
      Do D = 1, PsiDim
         Do j = 1, TrueOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) + &
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end parallel do

! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      olderr=Error
      if( precon.eq.0 ) then
         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + &
                  dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder)* &
              DConjg(HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E)* &
                                 (HDiagonal(D)-E)+d2)
         enddo
      else
         call preconditz(Krylov(1,TrueOrder),HDiagonal,leigf,reigf, &
              davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)

      endif
      Error = Sqrt(Error)

! --- OLSEN CORRECTION : cor1=<ev|(H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then
! --- Build EigenVector.
         Do D = 1, PsiDim
            workc(2*precon+D) = EigenVector(0,Q) * Psi(D)
         enddo
         Do j = 1, TrueOrder-1
            Do D = 1, PsiDim
               workc(2*precon+D) = workc(2*precon+D) + &
                    EigenVector(j,Q) * Krylov(D,j)
            enddo
         enddo
         cor1=(0.d0,0.d0)
         cor2=(0.d0,0.d0)
         if(precon.eq.0) then
            do D = 1, PsiDim
               cor1 = cor1+DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2 = cor2+DConjg(workc(2*precon+D))*workc(2*precon+D)* &
                    DConjg(HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E)* &
                                              (HDiagonal(D)-E)+d2)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * workc(2*precon+D) * &
                    DConjg(HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E)* &
                                               (HDiagonal(D)-E)+d2)
            enddo
         else
            call cpvxz(workc(2*precon+1),HKrylov(1,TrueOrder),PsiDim)
            call preconditz(HKrylov(1,TrueOrder),HDiagonal,leigf,reigf, &
                 davmat(1,precon+1),indxdav,E,ovj,d2,PsiDim,precon)

            do D = 1, PsiDim
               cor1=cor1+DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2=cor2+DConjg(workc(2*precon+D))*HKrylov(D,TrueOrder)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * HKrylov(D,TrueOrder)
            enddo
         endif
      endif                    !Letitia check till here!



! ---  ORTHONORMALIZE (entry for adding "lock" vector) ---

 200  ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,TrueOrder)
      enddo
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*Psi(D)
      enddo

      Do i = 1, TrueOrder-1
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,TrueOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - ovlp*Krylov(D,i)
         enddo
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + &
               dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
      enddo
      if(ovl.le.1.d-49) then     ! If vectors are linearly dependent
         write(6,'(a,3i5,2e12.4)') 'cDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(2,'(a,3i5,2e12.4)') 'cDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         TrueOrder = TrueOrder - 1
         write(6,'(2a,f11.3)') 'Vectors are linearly dependent.', &
                        ' Convergence assumed. Time =',Time/41.3414d0
         goto 310  ! Converged
      endif
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)/sqrt(ovl)
      enddo
!.....Orthonormalize twice.  (cDavStep)
      ovlp =(0.d0,0.d0)
      Do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,TrueOrder)
      enddo
      Do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*Psi(D)
      enddo

      Do i = 1, TrueOrder-1
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,TrueOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - ovlp*Krylov(D,i)
         enddo
      enddo
      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + &
               dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
      enddo
      Do D = 1,PsiDim
        Krylov(D,TrueOrder)=Krylov(D,TrueOrder)/sqrt(ovl)
      enddo

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
     Call WriteStepCDav(Steps,TrueOrder,E,beta,Error,Q,Time)

! --- BUILD HKRYLOV  ---
      Call Func(Time,Krylov(1,TrueOrder),HKrylov(1,TrueOrder))

! --- BUILD NEW T-MATRIX ELEMENTS ---
      
      Tmat = (0.0_dop,0.0_dop)

      Do D = 1, PsiDim
         Tmat(0,TrueOrder) = Tmat(0,TrueOrder) +&
              DConjg(Psi(D))*HKrylov(D,TrueOrder)     
         Tmat(TrueOrder,0) = Tmat(TrueOrder,0) +&
              DConjg(Krylov(D,TrueOrder))*DtPsi(D)     
         Tmat(TrueOrder,TrueOrder) = Tmat(TrueOrder,TrueOrder) +&
              DConjg(Krylov(D,TrueOrder))*HKrylov(D,TrueOrder)     
      enddo

!$omp parallel do if(lompthread.and.(Trueorder-1).ge.ompthread) num_threads(ompthread) private(j,D)
      do j = 1, TrueOrder-1
         do D = 1, PsiDim
            Tmat(j,TrueOrder) = Tmat(j,TrueOrder) +&
            DConjg(Krylov(D,j))*HKrylov(D,TrueOrder)     
            Tmat(TrueOrder,j) = Tmat(TrueOrder,j) +&
            DConjg(Krylov(D,TrueOrder))*HKrylov(D,j)     
         enddo
      enddo
!$omp end parallel do
            
! --- DIAGONALISE T-MATRIX (cDavStep) ---
      
      do j = 0, TrueOrder
         do i = 0, TrueOrder
            TTmat(i,j) = Tmat(i,j)
         enddo 
      enddo

      call ZGEEV( 'N', 'V', TrueOrder+1, TTmat, IntOrder+1,&
                   EigenVal, vl, 1, EigenVector, IntOrder+1,& 
                   workc, 16*IntOrder+18, workr, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         Return
      EndIf


! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if GS not sought) -

      if(rlaxnum.eq.0) then
         Q = 0    ! Search for lowest energy (real + abs(im))
         Er=dble(EigenVal(0)) + abs(dimag(EigenVal(0)))
         Do j = 1, TrueOrder
            Er1=dble(EigenVal(j)) + abs(dimag(EigenVal(j)))
            if(Er1.lt.Er) then
               Q=j
               Er=Er1
            endif
         enddo
         beta= abs(EigenVector(0,Q))**2
      else
         beta=0.0_dop
         Do j = 0, TrueOrder
            beta1= abs(EigenVector(0,j))**2
            If(beta1 .gt. beta) then
               Q = j
               beta = beta1
            endif
         enddo
      endif
      E = EigenVal(Q)
      Q1 = Q
      E1 = E
      beta1 = beta
      ovl=0.0_dop


! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If((TrueOrder.LT.IntOrder) .And.  (full .or.&
        ( (Error.GT.TolError) .And. jump)) )  Goto 100


! --- COMPUTE EXACT OVERLAPS

 310  if(rlaxnum.eq.930) then   ! lock
!........The array ovlps contains the overlaps**2 with startvector.
         call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov,&
                        EigenVector,Hdiagonal(PsiDim+1),FirstCall)
         oldQ=Q
         Q=-1
         do j = 0, TrueOrder
            ovj = ovlps(j)
            if(j.eq.Q1) ovl1 = ovj
            if(dble(EigenVal(j)).lt.rlxemin .or. &
               dimag(EigenVal(j)).lt.rlximin .or. &
               dble(EigenVal(j)).gt.rlxemax) ovj=0.0_dop  
 
            If(ovj .gt. ovl) then
               Q = j
               ovl = ovj
            endif
         enddo
         if(Q .eq. -1) Q=oldQ  ! no vector i energy window
         E = EigenVal(Q)
         beta = abs(EigenVector(0,Q))**2
         If((Q1.ne.Q) .and. (TrueOrder.LT.IntOrder) .and. jump ) then
            jump = .false.  
            call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1,&
                     ovl,ovl1,EigenVal,ovlps,FirstCall,IntOrder)
            goto 100
         endif
      endif

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If( TrueOrder .EQ. IntOrder ) Goto 300
      If( full .or. (Error+olderr.GT.TolError) ) Goto 100

! --- CONVERGENCE! Build new A-vector (cDavStep) ---

 300  Do D = 1,PsiDim
         Psi(D)   = EigenVector(0,Q)*Psi(D)
         DtPsi(D) = EigenVector(0,Q)*DtPsi(D)
      EndDo

!$omp parallel do if(lompthread) num_threads(ompthread) private(D,P)
      Do D = 1,PsiDim
         Do P = 1,TrueOrder
            Psi(D)   =   Psi(D)+EigenVector(P,Q)*Krylov(D,P)
            DtPsi(D) = DtPsi(D)+EigenVector(P,Q)*HKrylov(D,P)
         EndDo
      EndDo
!$omp end parallel do 

      If(Q.ne.Q1) call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1,&
                             ovl,ovl1,EigenVal,ovlps,.false.,IntOrder)

      if(FirstCall .and. (rlaxnum.ne.930)) then
!.....This is for first diag. wrtrlxdav
         do j = 0, TrueOrder
            ovlps(j) = Abs(EigenVector(0,j))**2
         enddo
      endif
      do j = 0, TrueOrder
         Hdiagonal(j+1) = EigenVal(j)   
      enddo
      Q1 = Q
      call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,ovl1,&
                           Hdiagonal,ovlps,FirstCall,IntOrder)
      FirstCall = .false.
      full      = .false.
      
      Return
      End Subroutine cDavStep                              ! (cDavStep)

! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE BDAVSTEP                                *
! *                                                                    *
! * Block-Davidson diagonalisation. (single-set multi-packet run).     *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming hermitian Hamiltonian and complex WF.                     *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The block of (complex) initial-value vectors.         *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim/npacket)   *
! *   BDim       Block size. BDim=npacket (=gdim(feb)                  *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStepBDAv").                                     *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (complex)     *
! *   HKrylov:   Matrix containing H*Krylov (complex)                  *
! *   precon:    Dimension of Pre-Conditioner                          *
! *   davmat:    Contains the eigenvectors and eigenvalues of the      *
! *              preconditioner.                                       *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Time,Psi,DtPsi).                                 *
! * WriteStepBDav: Writes the energy,beta, and error to a file.        *
! *   Called as WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)  *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 03/2010 HDM                                                   *
! * Modifications and OpenMP parallelisation. GWR 09/13                *
! **********************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in bdavstep, please do them in mpibdavstep too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Subroutine BDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError,          &
                    Steps,rlaxnum,Krylov,HKrylov,HDiagonal,            &
                    EigenVector,EigenVal,ovlps,Rdiagonal,              &
                    TrueOrder,ErrorCode,Time,precon,olsen,             &
                    Func,rlxemin,rlxemax,Bdim,adim,nmod,               &
                    lf,lmult)

      Implicit None

      External                                         :: Func
      Logical(kind=4), save                            :: full
      Logical(kind=4), intent(in)                      :: olsen
      Integer(long), intent(in)                        :: Steps
      Logical(kind=4), intent(in)                      :: lf, lmult
      Integer(long), intent(inout)                     :: rlaxnum
      Integer(long), intent(out)                       :: ErrorCode,TrueOrder
      Integer(long), intent(in)                        :: PsiDim,IntOrder, &
                                                          nmod,Bdim,adim
      Integer(long), intent(inout)                     :: precon
      Real(dop), intent(in)                            :: TolError,Time, &
                                                          rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)  :: ovlps,EigenVal
      Complex(dop), dimension(adim), intent(inout)     :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,0:IntOrder), intent(out)      :: Krylov
      Complex(dop), dimension(PsiDim,0:IntOrder), intent(inout)    :: HKrylov
      Complex(dop), dimension(PsiDim), intent(out)             :: HDiagonal
      Real(dop), dimension(PsiDim), intent(out)                :: RDiagonal
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector

      Integer(long) :: D,P,Q,QQ,i,j,blk
      Integer(long), dimension(precon) :: indxdav
      complex(dop), dimension(precon,precon+1) :: davmat
      Real(dop)     :: Error,E,olderr,ovl,d2,beta,beta1
      Real(dop), dimension(:), allocatable       :: workr
      Complex(dop), dimension(:), allocatable    :: workc

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      Logical(kind=4), save :: FirstCall = .true.

      Integer(long) :: mxcv,border,lwork,qmin
      Integer(long), parameter :: mxbd=320

      Integer(long), dimension(0:8*BDim+9) :: converged, qmap
      Real(dop) :: epse,cor2,ErrorM
      Real(dop), dimension(0:2*mxbd+7) :: olden
      Real(dop), dimension(0:2*mxbd-1), save :: oldenergy
      complex(dop) :: cor1

      mxcv = 8*BDim+9
      epse=0.d0

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .LT. 2*BDim) Then
         ErrorCode = 1
         write(6,'(a,i4)') 'IntOrder must be .GE. 2*packets'
         write(6,'(2(a,i4))') 'IntOrder =',IntOrder,' <',2*BDim
         Return
      EndIf

! --- INITIALIZE VARIABLES ---

      if(FirstCall) Then
         if(rlaxnum .ge. 9999) then
            full = .true.
            rlaxnum = rlaxnum - 10000
         else
            full = .false. 
         endif 
         if(rlaxnum.ne.0 .and. rlaxnum.ne.920) then
            write(6,'(a)')' BDAV: Only rlaxnum=0 or follow allowed'
            stop
         endif
         if(Bdim .gt. mxbd) then
            write(6,'(/a,i5,a,i5)') ' ERROR in bdavlib.f : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd
            write(6,'(a,/a)') ' Increase the parameter mxbd in both',  &
                ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
            write(2,'(/a,i5,a,i5)') ' ERROR in bdavlib.f : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd
            write(2,'(a,/a)') ' Increase the parameter mxbd in both',  &
                ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
            stop
         endif
      endif
      Do i = 0, mxcv
         qmap(i)=i
         converged(i)=0
      enddo 
      d2 = 1.d-18
      ErrorCode = 0
      ErrorM    = 1.d0
      beta1     = 0.d0
      Error     = 1.d0
      TrueOrder = Bdim - 1
      BOrder    = BDim
      lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
      call zerovxd(ovlps,IntOrder)

      allocate(workc(lwork))
      allocate(workr(3*border))

!.....Compute the diagonal of the Hamilton matrix.
      call bgetdiagonal(HDiagonal)

!.....Copy Psi and DtPsi to zero-th column of (complex) Krylov and HKrylov.
      if(lmult) then
         call psi2kryc(Psi,Krylov,0)
         call psi2kryc(DtPsi,HKrylov,0)
      else
         Do Q = 0, BDim-1
            Do D = 1, PsiDim
                Krylov(D,Q) =   Psi(D+PsiDim*Q)
               HKrylov(D,Q) = DtPsi(D+PsiDim*Q)
            enddo
         enddo
      endif
      Do D = 1, PsiDim
         Rdiagonal(D) = dble(HDiagonal(D))
      enddo


! --- BUILD T-MATRIX  ---
      EigenVector = 0.0_dop
!$omp parallel do if(lompthread.and.(BOrder+1).ge.ompthread) num_threads(ompthread) private(i,j,D)
      do j = 0, BOrder - 1
         do i = 0, j 
            do D = 1, PsiDim
               Eigenvector(i,j) = EigenVector(i,j) + &
                    DConjg(Krylov(D,i))*HKrylov(D,j)
            enddo
         enddo
      enddo
!$omp end parallel do

! --- DIAGONALISE T-MATRIX IN THE INITIAL BLOCK OF STARTVECTORS ---
!       call dsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal,
!      $               workr,lwork,ErrorCode)

      call ZHEEV( 'V', 'U', BOrder, EigenVector, IntOrder+1, &
                 EigenVal, workc, lwork, workr, ErrorCode)

      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workc,workr)
         Return
      EndIf

      if(FirstCall) then
         i=1
         beta=1.d0
         beta1=0.d0
         Do Q = 0, BDim-1
            E=EigenVal(Q)
            call wrtrlxbdav(time,Q,Q,Q,E,beta,E,i,EigenVal, &
                            ovlps,FirstCall,IntOrder,BDim)
            oldenergy(Q)=E
            olden(Q)=E
            beta1=beta1+E
         enddo
         E=beta1/dble(BDim)  ! Averaged energy for pre-conditioner
         if(rlxemin.gt.-1.d99) E=rlxemin 
         if(rlxemax.lt. 1.d99) E=rlxemax 
      endif   ! firstcall

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
      if( precon .le. 1 ) precon=0
      if( precon .gt. 0 ) then
         write(6,*) 'Sorry, preconditioner not yet implemented', &
                    ' for BDAV (hermitian)'
         stop 
      endif


! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

  99  continue  ! Block Iteration Loop
      olderr = ErrorM
      ErrorM = 0.d0
      QQ = 0

 100  Continue  !  Iteration Loop

! --- RE-INITIALIZE VARIABLES ---

      TrueOrder = TrueOrder+1
 42   if(converged(QQ).gt.0) then
         if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt.epse) then
            if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. &
                 10.d0*abs(EigenVal(qmap(QQ))-olden(QQ+1))) then ! state missing
               do i = QQ,mxcv-1
                  converged(i) = converged(i+1)
                  olden(i)     = olden(i+1)
               enddo
            elseif(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. &
                 10.d0*abs(EigenVal(qmap(QQ))-olden(max(0,QQ-1)))) then !st. added
               do i = mxcv, QQ+1, -1
                  converged(i) = converged(i-1)
                  olden(i)     = olden(i-1)
               enddo
               converged(QQ)=0
            else
               converged(QQ)=0
            endif
            goto 42  ! Try again, after converged is modified
         endif
         QQ=QQ+1     ! Skip the converged state
         goto 42
      endif
      if(qq.ge.mxcv) then
         write(6,'(a,2i6)')  'BDAV: ## qq, q', qq, q
         write(6,'(a,i6,a)') 'BDAV: Enlarge mxcv=',mxcv,' (also in rundav)'
         WRITE(6,'(a,/(20i4))') 'Qmap:     ',(qmap(i),i=0,mxcv)
         WRITE(6,'(a,/(20i4))') 'Converged:',(converged(i),i=0,mxcv)
         stop 6
      endif
      Q = qmap(QQ)
      E = EigenVal(Q)
      olden(QQ) = E

! --- Build residual vector |r> ---
      Do D = 1, PsiDim
         Krylov(D,TrueOrder) =  EigenVector(0,Q) * (HKrylov(D,0) &
              - E*Krylov(D,0))
      enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(j,D)
      Do D = 1, PsiDim
         Do j = 1, BOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) + &
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end parallel do

! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
     if( precon.eq.0 ) then
         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + abs(Krylov(D,TrueOrder))**2
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder)* &
                 (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
         enddo
      else
         call precondith(Krylov(1,TrueOrder),HDiagonal,davmat, &
              davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error)
      if(QQ.lt.Bdim) ErrorM = max(Error,ErrorM)
      if(Error .lt. 0.5*TolError .and. converged(QQ).eq.0) then
         converged(QQ) = -TrueOrder
      endif
      if(Error .lt. 0.05d0*TolError  .and. converged(QQ).le.0) then
         converged(QQ) = abs(converged(QQ))
      elseif(Error .gt. TolError) then
         converged(QQ) =  0
      endif

! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- |r> = (H-E) * eigenvector
! --- Krylov(TrueOrder)  = (H_0-E)^-1 * |r>    
! --- workc(2*precon+D)  =  eigenvector(D)
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector (used for precon)
! --- Correction : (cor1/cor2)*(H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then
! --- Build EigenVector.
         Do D = 1, PsiDim
            workc(2*precon+D) = EigenVector(0,Q) * Krylov(D,0)
         enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(j,D)
         Do D = 1, PsiDim
            Do j = 1, TrueOrder-1
               workc(2*precon+D) = workc(2*precon+D) + &
                    EigenVector(j,Q) * Krylov(D,j)
            enddo
         enddo
!$omp end parallel do

         cor1=0.d0
         cor2=0.d0
         if (precon .eq. 0) then
            do D = 1, PsiDim
               cor1 = cor1 + DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2 = cor2 + abs(workc(2*precon+D))**2 * &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * workc(2*precon+D) *      &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
         else
            call cpvxz(workc(2*precon+1),HKrylov(1,TrueOrder),PsiDim)
            call precondith(HKrylov(1,TrueOrder),HDiagonal,davmat, &
               davmat(1,precon+1),indxdav,E,ovl,d2,PsiDim,precon)

            do D = 1, PsiDim
               cor1=cor1+DConjg(workc(2*precon+D))*Krylov(D,TrueOrder)
               cor2=cor2+dble(DConjg(workc(2*precon+D))*HKrylov(D,TrueOrder))
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * HKrylov(D,TrueOrder)
            enddo
         endif  ! precon
            
      endif     ! olsen


! ---  ORTHONORMALIZE  ---
      call orthonc(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)
      if(ovl.le.1.d-40) then     ! If vectors are linearly dependent
         write(6,'(/a)') '---------------------------------------------'
         write(6,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(2,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(6,'(a)') 'BDAV: Vectors are linearly dependent!' 
         write(2,'(a)') 'BDAV: Vectors are linearly dependent!' 
         WRITE(6,'(a,20i3)') 'Qmap:     ', (qmap(i),i=0,2*BDim-1)
         WRITE(6,'(a,20i3)') 'Converged:', (converged(i),i=0,2*BDim-1)
         stop 2
      endif
!.....Symmetrize coefficients, if desired.
      if(nmod.gt.0) call csymcoeff(Krylov(1,TrueOrder),  &
                    HKrylov(1,TrueOrder),PsiDim,nmod,lf,ovl,TrueOrder)

!.....Do the orthonormalisation twice
      call orthonc(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)

! --- WRITE STEP AND ERROR IF DESIRED ---
      E = EigenVal(Q)
      beta = Abs(EigenVector(QQ,Q))**2
      Call WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)

      if(TrueOrder .ge. BOrder+BDim-1) goto 41
      QQ = QQ + 1
      goto 100
 41   QQ = 0


! --- BUILD HKRYLOV  ---
      if(lmult) then
         call kryc2psi(Psi,Krylov,BOrder)
      else

!$omp parallel do if(lompthread.and.(BDim+1).ge.ompthread) num_threads(ompthread) private(blk,D)
         Do blk = 0,BDim-1
            Do D = 1,PsiDim
               Psi(D+blk*PsiDim) = Krylov(D,BOrder+blk)
            enddo
         enddo
!$omp end parallel do

      endif
      Call Func(Time,Psi,DtPsi)
      if(lmult) then
         call psi2kryc(DtPsi,HKrylov,BOrder)
      else

!$omp parallel do if(lompthread.and.(BDim+1).ge.ompthread) num_threads(ompthread) private(blk,D)
         Do blk = 0,BDim-1
            Do D = 1,PsiDim
               HKrylov(D,BOrder+blk) = DtPsi(D+blk*PsiDim)
            enddo
         enddo
!$omp end parallel do

      endif

! --- BUILD NEW T-MATRIX ELEMENTS ---
!$omp parallel do if(lompthread.and.(BOrder+1).ge.ompthread) num_threads(ompthread) private(i,j,D)
      do j = BOrder, TrueOrder
         do i = 0, j 
            do D = 1, PsiDim
               Eigenvector(i,j) = EigenVector(i,j) + &
                    DConjg(Krylov(D,i))*HKrylov(D,j)
            enddo
         enddo
      enddo
!$omp end parallel do

! --- DIAGONALISE T-MATRIX ---
      BOrder=TrueOrder+1
!       call pardsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal,
!      $               workr,lwork,ErrorCode)
      call ZHEEV( 'V', 'U', BOrder, EigenVector, IntOrder+1, &
                 EigenVal, workc, lwork, workr, ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workc,workr)
         Return
      EndIf

! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS sought) -

      if(rlaxnum.eq.920) then   ! follow
         qmin=2*BDim
         Do Q = 0, BDim-1
            beta=0.d0
            j = -1
            Do P = 0, TrueOrder
               if(EigenVal(P).ge.rlxemin .and. EigenVal(P).le.rlxemax) then 
                  beta1= Abs(EigenVector(Q,P))**2
                  do i = 0, Q-1
                     if(qmap(i).eq.P) beta1=-1.d0
                  enddo
                  if(beta1 .gt. beta) then
                     j = P
                     beta = beta1
                  endif
               endif
            enddo
            if(j.lt.0) then  ! Energy window too small.
               beta=1.d+99   ! Take vectors with energies closest to window.
               Do P = 0, TrueOrder
                  beta1 = min(abs(rlxemin-EigenVal(P)),abs(EigenVal(P)-rlxemax))
                  if(beta1.lt.beta) then
                     beta=beta1
                     j = P
                  endif 
               enddo
            endif
            qmap(Q) = j
            qmin = min(qmin,j)
         enddo  ! Q

         if(full) qmin=0
         qmin=max(qmin-1,0)               ! Set higher entries of qmap.
         Do Q = BDim,min(mxcv,TrueOrder)  ! They may be used when some
   51       do i = 0, Q-1                 ! vectors converge. 
               if(qmap(i).eq.qmin) then
                  qmin=qmin+1
                  if(qmin.gt.TrueOrder) qmin=0
                  goto 51
               endif
            enddo
            qmap(Q)=qmin
         enddo   !  Q
      elseif(rlaxnum.eq.0 .and. rlxemin.gt.-1.d99) then
         QQ=0
         do Q = 0, TrueOrder
            if(EigenVal(Q).gt.rlxemin .and. QQ.lt.mxcv) then
               qmap(QQ)=Q
               QQ=QQ+1
               i=Q
               if(QQ.eq.Bdim) goto 52
            endif
         enddo
   52    Q=i+1
         if(QQ.lt.Bdim) then
            qmap(QQ)=Q
            QQ=QQ+1
            Q=Q+1
            goto 52
         endif
         if(qmap(0).gt.0) then  ! Take the first state below the block
            qmap(QQ)=qmap(0)-1  ! as additional vector (when others
            QQ=QQ+1             ! have converged).
         endif
   53    if(QQ.lt.mxcv) then    ! Take the states following the block
            qmap(QQ)=Q          ! as additional vectors (when others
            QQ=QQ+1             ! have converged).
            Q=Q+1
            goto 53
         endif
      endif      ! (if rlaxnum.eq.920 ...  elseif rlaxnum.eq.0)
      epse=(EigenVal(qmap(Bdim-1))-EigenVal(qmap(0)))/(Bdim*90.d0)

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP  ---
      If( TrueOrder .GT. IntOrder-BDim ) Goto 300
      If( TrueOrder .GE. PsiDim-BDim   ) Goto 300
      If( full .or. (ErrorM+olderr.GT.TolError) ) Goto 99

! --- CONVERGENCE! Build new A-vector ---

 300  if(lmult) then
         call kryc2finalpsi(Psi,DtPsi,Krylov,HKrylov,EigenVector,TrueOrder,qmap)
      else
!$omp parallel do if(lompthread.and.(BDim+1).ge.ompthread) num_threads(ompthread) private(QQ,Q,D,P)
         Do QQ = 0,BDim-1
            Q = qmap(QQ) 
            Do D = 1,PsiDim 
               Psi(D+QQ*PsiDim)   =  Krylov(D,0)*EigenVector(0,Q)
               DtPsi(D+QQ*PsiDim) = HKrylov(D,0)*EigenVector(0,Q)
            EndDo
            Do P = 1,TrueOrder
               Do D = 1,PsiDim
                  Psi(D+QQ*PsiDim)   = Psi(D+QQ*PsiDim) + &
                                       Krylov(D,P)*EigenVector(P,Q)
                  DtPsi(D+QQ*PsiDim) = DtPsi(D+QQ*PsiDim) + &
                                       HKrylov(D,P)*EigenVector(P,Q)
               EndDo
            EndDo
         EndDo
!$omp end parallel do
      endif


      Do QQ = 0, BDim-1
         Q = qmap(QQ)
         if(FirstCall) then
!........This is for first overlaps in wrtrlxbdav
            do j = 0, TrueOrder
               ovlps(j) = Abs(EigenVector(Q,j))**2
            enddo
         endif
         E=EigenVal(Q)
         beta=Abs(EigenVector(QQ,Q))**2
         j=QQ
         if(beta.lt.0.5d0 .and. TrueOrder.gt.Bdim) then
            beta=0.d0
            do i=0,Bdim+1
               if(Abs(EigenVector(i,Q))**2 .gt. beta) then
                  beta=Abs(EigenVector(i,Q))**2
                  j=i
               endif
            enddo
            if(beta.lt.0.67d0) then
               beta=Abs(EigenVector(QQ,Q))**2
               j=QQ
            endif
         endif
         call wrtrlxbdav(time,trueorder,QQ,Q,E,beta,oldenergy(j),  &
              converged(QQ),EigenVal,ovlps,FirstCall,IntOrder,BDim)
      enddo
      Do QQ = 0, BDim+1
         Q = qmap(QQ)
         oldenergy(QQ)=EigenVal(Q)
      enddo

      FirstCall = .false.
      full      = .false.
      edav = E

      deallocate(workc,workr)

      Return
      End subroutine bdavstep

! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE BRDAVSTEP                               *
! *                                                                    *
! * Block-Davidson diagonalisation. (single-set multi-packet run).     *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming real Hamiltonian and real WF.                             *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The block of (complex) initial-value vectors.         *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim/npacket)   *
! *   BDim       Block size. BDim=npacket (=gdim(feb)                  *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStepBDAv").                                     *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (real!)       *
! *   HKrylov:   Matrix containing H*Krylov (real!)                    *
! *   precon:    Dimension of Pre-Conditioner                          *
! *   davmat:    Contains the eigenvectors and eigenvalues of the      *
! *              preconditioner.                                       *
! *   nmod       Number of modes. If nmod=0 => no symmetrisation       *
! *   lf         lf=.true. for fermions, lf=.false. for bosons         *
! *   lrreal     lrreal=.true. => Funcr is used (rrDAV).               *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Time,Psi,DtPsi).                                 *
! *   Funcr:     Special routine for real arithmetic                   * 
! * WriteStepBDav: Writes the energy,beta, and error to a file.        *
! *   Called as WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)  *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 09/2007 HDM                                                   *
! * Modifications and OpenMP parallelisation. GWR 09/13                *
! **********************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in brdavstep, please do them in mpibrdavstep too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Subroutine BRDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError,         &
                    Steps,rlaxnum,Krylov,HKrylov,HDiagonal,            &  
                    EigenVector,EigenVal,ovlps,Rdiagonal,davmat, &
                    TrueOrder,ErrorCode,Time,precon,olsen,             &
                    Func,Funcr,lrreal,rlxemin,rlxemax,Bdim,adim,nmod,  &
                    lf,lmult)

      Implicit None

      External                                                   :: Func, Funcr
      Logical(kind=4), save                                      :: full
      Logical(kind=4), intent(in)                                :: olsen
      Integer(long), intent(in)                                  :: Steps
      Logical(kind=4), intent(in)                              :: lf, lmult, lrreal
      Integer(long), intent(inout)                               :: rlaxnum
      Integer(long), intent(out)                                 :: ErrorCode,TrueOrder
      Integer(long), intent(in)                                  :: PsiDim,IntOrder,nmod,Bdim,Adim
      Integer(long), intent(inout)                               :: precon
      Real(dop), intent(in)                                      :: TolError,Time,rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps,EigenVal
      Real(dop), dimension(precon,precon+1), intent(inout)       :: davmat
      Complex(dop), dimension(PsiDim), intent(inout)             :: Psi,DtPsi
      Complex(dop), dimension(PsiDim), intent(out)             :: HDiagonal
      Real(dop), dimension(PsiDim,0:IntOrder), intent(out)      :: Krylov
      Real(dop), dimension(PsiDim,0:IntOrder), intent(inout)    :: HKrylov
      Real(dop), dimension(PsiDim), intent(out)                :: RDiagonal
      Real(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector

      Integer(long) :: D,P,Q,QQ,i,j,blk,idown
      Real(dop)     :: Error,olderr,ovl,d2,E,beta,beta1
      Real(dop), dimension(:), allocatable       :: workr
      integer(long), save                        :: tid=0 ! timer ID

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      Logical(kind=4), save :: FirstCall = .true.

      Integer(long) :: mxcv,border,lwork,qmin
      Integer(long), parameter :: mxbd=320

      Integer(long), dimension(precon) :: indxdav
      Integer(long), dimension(0:8*BDim+9) :: converged, qmap
      Real(dop) :: epse,cor1,cor2,ErrorM
      Real(dop), dimension(0:2*mxbd+7) :: olden
      Real(dop), dimension(0:2*mxbd-1,0:2), save :: oldenergy

      mxcv = 8*BDim+9
      epse=0.d0

! --- CHECK INTEGRATION ORDER ---

      If (IntOrder .LT. 2*BDim) Then
         ErrorCode = 1
         write(6,'(a,i4)') 'IntOrder must be .GE. 2*packets'
         write(6,'(2(a,i4))') 'IntOrder =',IntOrder,' <',2*BDim
         Return
      EndIf

! --- INITIALIZE VARIABLES ---

      if(FirstCall) Then
         if(rlaxnum .ge. 9999) then
            full = .true.
            rlaxnum = rlaxnum - 10000
         else
            full = .false. 
         endif 
         if(rlaxnum.ne.0 .and. rlaxnum.ne.920) then
            write(6,'(a)')' BRDAV: Only rlaxnum=0 or follow allowed'
            stop
         endif
         if(Bdim .gt. mxbd) then
            write(6,'(/a,i5,a,i5)') ' ERROR in brdavlib.f : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd            
            write(6,'(a,/a)') ' Increase the parameter mxbd in both',   &
              ' source/lib/ode/brdavlib.f and source/propwf/mpibrdav.F'
            write(2,'(/a,i5,a,i5)') ' ERROR in bdavlib.f : packets =',  &
                                Bdim,' .gt.  mxbd =', mxbd
            write(2,'(a,/a)') ' Increase the parameter mxbd in both',   &
              ' source/lib/ode/brdavlib.f and source/propwf/mpibrdav.F'
            stop
         endif
      endif
      Do i = 0, mxcv
         qmap(i)=i
         converged(i)=0
      enddo 
      d2 = 1.d-18
      ErrorCode = 0
      ErrorM    = 1.d0
      Error     = 1.d0
      beta1     = 0.d0
      TrueOrder = Bdim - 1
      BOrder    = BDim
      lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
      call zerovxd(ovlps,IntOrder)

      allocate(workr(lwork))

!.....Compute the diagonal of the Hamilton matrix.
      call bgetdiagonal(HDiagonal)

!.....Copy complex to real arrays.
      cor1=0.d0
      if(lmult) then
         call psi2kry(Psi,Krylov,0)
         call psi2kry(DtPsi,HKrylov,0)
         Do D = 1, adim
            cor1 = cor1 + abs(dimag(Psi(D))) + abs(dimag(DtPsi(D)))
         enddo
      else
         Do Q = 0, BDim-1
            Do D = 1, PsiDim
                Krylov(D,Q) = dble(  Psi(D+PsiDim*Q))
               HKrylov(D,Q) = dble(DtPsi(D+PsiDim*Q))
               cor1=cor1 + abs(dimag(  Psi(D+PsiDim*Q))) + &
                      abs(dimag(DtPsi(D+PsiDim*Q)))
            enddo
         enddo
      endif
      If(cor1.gt.1.d-11) call realdaverror(cor1)
      Do D = 1, PsiDim
         Rdiagonal(D) = dble(HDiagonal(D))
      enddo

! --- Additional re-orthonormalization of startvectors
      Do Q = 0, BDim-1
         call orthonr(Krylov,Krylov(1,Q),ovl,Q-1,PsiDim)
      enddo

! --- BUILD T-MATRIX  ---
      EigenVector = 0.0_dop
!$omp parallel do if(lompthread.and.(BOrder+1).ge.ompthread) num_threads(ompthread) private(i,j,D)
      do j = 0, BOrder - 1
         do i = 0, j 
            do D = 1, PsiDim
               Eigenvector(i,j) = EigenVector(i,j) + Krylov(D,i)*HKrylov(D,j)
            enddo
         enddo
      enddo
!$omp end parallel do

! --- DIAGONALISE T-MATRIX IN THE INITIAL BLOCK OF STARTVECTORS ---
      call pardsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                     workr,lwork,ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workr)
         Return
      EndIf

      if(FirstCall) then
         i=1
         beta=1.d0
         beta1=0.d0
         Do Q = 0, BDim-1
            E=EigenVal(Q)
            call wrtrlxbdav(time,Q,Q,Q,E,beta,E,i,EigenVal, &
                            ovlps,FirstCall,IntOrder,BDim)
            oldenergy(Q,0)=E
            oldenergy(Q,1)=E
            oldenergy(Q,2)=E
            olden(Q)=E
            beta1=beta1+E
         enddo
         E=beta1/dble(BDim)  ! Averaged energy for pre-conditioner
         if(rlxemin.gt.-1.d99) E=rlxemin 
         if(rlxemax.lt. 1.d99) E=rlxemax 
      endif   ! firstcall

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
      if( precon .le. 1 ) precon=0
      if( precon .gt. 0 ) then
         call get_timer(tid, 'Precon')
         call start_timer(tid)

         call sortdiagonal(RDiagonal,workr,indxdav,E,PsiDim,precon)
         call getdavmat(RDiagonal,davmat,davmat(1,precon+1), &
                        indxdav,PsiDim,precon,lwork)
         call stop_timer(tid)
      endif


! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

  99  continue  ! Block Iteration Loop
      olderr = ErrorM
      ErrorM = 0.d0
      QQ = 0

 100  Continue  !  Iteration Loop

! --- RE-INITIALIZE VARIABLES ---

      TrueOrder = TrueOrder+1
 42   if(converged(QQ).gt.0) then
         if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt.epse) then
            if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. &
                 10.d0*abs(EigenVal(qmap(QQ))-olden(QQ+1))) then ! state missing
               do i = QQ,mxcv-1
                  converged(i) = converged(i+1)
                  olden(i)     = olden(i+1)
               enddo
            elseif(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. &
                 10.d0*abs(EigenVal(qmap(QQ))-olden(max(0,QQ-1)))) then !st. added
               do i = mxcv, QQ+1, -1
                  converged(i) = converged(i-1)
                  olden(i)     = olden(i-1)
               enddo
               converged(QQ)=0
            else
               converged(QQ)=0
            endif
            goto 42  ! Try again, after converged is modified
         endif 
         QQ=QQ+1     ! Skip the converged state
         goto 42
      endif
      if(qq.ge.mxcv) then
         write(6,'(a,2i6)')  'BRDAV: ## qq, q', qq, q
         write(6,'(a,i6,a)') 'BRDAV: Enlarge mxcv=',mxcv,' (also in rundav)'
         WRITE(6,'(a,/(20i4))') 'Qmap:     ',(qmap(i),i=0,mxcv)
         WRITE(6,'(a,/(20i4))') 'Converged:',(converged(i),i=0,mxcv)
         stop 6
      endif
      Q = qmap(QQ)
      E = EigenVal(Q)
      olden(QQ) = E

! --- Build residual vector |r> ---
      Do D = 1, PsiDim
         Krylov(D,TrueOrder) = EigenVector(0,Q) * (HKrylov(D,0) - E*Krylov(D,0))
      enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(D,j)
      Do D = 1, PsiDim
         Do j = 1, BOrder-1
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder) +  &
                 EigenVector(j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo
!$omp end parallel do

! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      if( precon.eq.0 ) then
         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + Krylov(D,TrueOrder)**2
            Krylov(D,TrueOrder) =  &
                 Krylov(D,TrueOrder)*(RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
         enddo
      else
         call precondit(Krylov(1,TrueOrder),RDiagonal,davmat,  &
                        davmat(1,precon+1),indxdav,E,Error,d2,PsiDim,precon)
      endif
      Error = Sqrt(Error)
      if(QQ.lt.Bdim) ErrorM = max(Error,ErrorM)
      if(Error .lt. 0.5*TolError .and. converged(QQ).eq.0) then
         converged(QQ) = -TrueOrder
      endif
      if(Error .lt. 0.05d0*TolError  .and. converged(QQ).le.0) then
         converged(QQ) = abs(converged(QQ)) 
      elseif(Error .gt. TolError) then
         converged(QQ) =  0
      endif

! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>    
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then
! --- Build EigenVector.
         Do D = 1, PsiDim
            workr(2*precon+D) = EigenVector(0,Q) * Krylov(D,0)
         enddo

!$omp parallel do if(lompthread) num_threads(ompthread) private(j,D)
         Do D = 1, PsiDim
            Do j = 1, TrueOrder-1
               workr(2*precon+D) = workr(2*precon+D) +   &
                    EigenVector(j,Q) * Krylov(D,j)
            enddo
         enddo
!$omp end parallel do

         cor1=0.d0
         cor2=0.d0
         if(precon.eq.0) then
            do D = 1, PsiDim
               cor1 = cor1 + workr(2*precon+D)*Krylov(D,TrueOrder)
               cor2 = cor2 +  &
                  workr(2*precon+D)**2*(RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * workr(2*precon+D) *      &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo

         else
            call cpvxd(workr(2*precon+1),HKrylov(1,TrueOrder),PsiDim)
            call precondit(HKrylov(1,TrueOrder),RDiagonal,davmat,  &
                           davmat(1,precon+1),indxdav,E,ovl,d2,PsiDim,precon)
            do D = 1, PsiDim
               cor1 = cor1 + workr(2*precon+D)*Krylov(D,TrueOrder)
               cor2 = cor2 + workr(2*precon+D)*HKrylov(D,TrueOrder)
            enddo
            do D = 1, PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                    (cor1/cor2) * HKrylov(D,TrueOrder)
            enddo
         endif  ! precon
      endif     ! olsen


! ---  ORTHONORMALIZE  ---
      call orthonr(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)
      if(ovl.le.1.d-31) then     ! If vectors are linearly dependent
         write(6,'(/a)') '---------------------------------------------'
         write(6,'(a,3i5,2e12.4)') 'BRDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(2,'(a,3i5,2e12.4)') 'BRDAV: Steps, Order, Q, ovl, Error', &
                                    Steps,TrueOrder,Q,ovl,Error
         write(6,'(a)') 'BRDAV: Vectors are linearly dependent!' 
         write(2,'(a)') 'BRDAV: Vectors are linearly dependent!' 
         WRITE(6,'(a,20i3)') 'Qmap:     ', (qmap(i),i=0,2*BDim-1)
         WRITE(6,'(a,20i3)') 'Converged:', (converged(i),i=0,2*BDim-1)
         stop 2
      endif
!.....Symmetrize coefficients, if desired.
      if(nmod.gt.0) then
         call rsymcoeff(Krylov(1,TrueOrder), &
              HKrylov(1,TrueOrder),PsiDim,nmod,lf,ovl,0)
         if(ovl .le. 1.d0) write(2,'(a,1pe9.2,a,i6)') &
           'rsymcoeff-norm =',ovl,', TrueOrder =',TrueOrder
         if(TrueOrder .gt. Bdim+1) then
            if(ovl.le.1.d-3) then
               TrueOrder=TrueOrder-1
               write(2,'(a,1pe9.2,a,i6)') 'BRDavStep: rsymcoeff-norm =', &
               ovl,',  Iteration stopped at TrueOrder =',TrueOrder
               goto 300
            endif
         else    
            if(ovl.le.1.d-6) then
               write(2,'(a,1pe9.2,a,i6)') 'BRDavStep: rsymcoeff-norm =', &
               ovl,',  Execution stopped! TrueOrder =',TrueOrder
               stop 1
            endif
         endif   
         call orthonr(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)
      endif

!.....Do the orthonormalisation twice
      call orthonr(Krylov,Krylov(1,TrueOrder),ovl,TrueOrder-1,PsiDim)

! --- WRITE STEP AND ERROR IF DESIRED ---
      E = EigenVal(Q)
      beta = EigenVector(QQ,Q)**2
      Call WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)

      if(TrueOrder .ge. BOrder+BDim-1) goto 41
      QQ = QQ + 1
      goto 100
 41   QQ = 0


! --- BUILD HKRYLOV  ---
      If(lrreal) Then 
         if(lmult) then
            call kry2psir(workr,Krylov,BOrder)
         endif
         Call Funcr(Time,Krylov(1,BOrder),HKrylov(1,BOrder))
         if(lmult) then
            call psi2kryr(workr,HKrylov,BOrder)
            call psi2kryr(workr,Krylov,BOrder)
         endif
      Else
         if(lmult) then
            call kry2psi(Psi,Krylov,BOrder)
         else
            Do blk = 0,BDim-1
               Do D = 1,PsiDim
                  Psi(D+blk*PsiDim) = Krylov(D,BOrder+blk)
               enddo
            enddo
         endif
         Call Func(Time,Psi,DtPsi)
         if(lmult) then
            call psi2kry(DtPsi,HKrylov,BOrder)
         else
            Do blk = 0,BDim-1
               Do D = 1,PsiDim
                  HKrylov(D,BOrder+blk) = dble(DtPsi(D+blk*PsiDim))
               enddo
            enddo
         endif
      EndIf

! --- BUILD NEW T-MATRIX ELEMENTS ---
!$omp parallel do if(lompthread.and.(BOrder+1).ge.ompthread) num_threads(ompthread) private(i,j,D)
      do j = BOrder, TrueOrder
         do i = 0, j 
            do D = 1, PsiDim
               Eigenvector(i,j) = EigenVector(i,j) + Krylov(D,i)*HKrylov(D,j)
            enddo
         enddo
      enddo
!$omp end parallel do

! --- DIAGONALISE T-MATRIX ---
      BOrder=TrueOrder+1
      call pardsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                     workr,lwork,ErrorCode)
      If (ErrorCode .NE. 0) Then
         ErrorCode = 3
         deallocate(workr)
         Return
      EndIf


! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS sought) -

      if(rlaxnum.eq.920) then   ! follow
         qmin=2*BDim
         Do Q = 0, BDim-1
            beta=0.d0
            j = -1
            Do P = 0, TrueOrder
               if(EigenVal(P).ge.rlxemin .and. EigenVal(P).le.rlxemax) then 
                  beta1= EigenVector(Q,P)**2
                  do i = 0, Q-1
                     if(qmap(i).eq.P) beta1=-1.d0
                  enddo
                  if(beta1 .gt. beta) then
                     j = P
                     beta = beta1
                  endif
               endif
            enddo
            if(j.lt.0) then  ! Energy window too small.
               beta=1.d+99   ! Take vectors with energies closest to window.
               Do P = 0, TrueOrder
                  beta1 = min(abs(rlxemin-EigenVal(P)),abs(EigenVal(P)-rlxemax))
                  if(beta1.lt.beta) then
                     beta=beta1
                     j = P
                  endif 
               enddo
            endif
            qmap(Q) = j
            qmin = min(qmin,j)
         enddo  ! Q

         if(full) qmin=0
         qmin=max(qmin-1,0)               ! Set higher entries of qmap.
         Do Q = BDim,min(mxcv,TrueOrder)  ! They may be used when some
   51       do i = 0, Q-1                 ! vectors converge. 
               if(qmap(i).eq.qmin) then
                  qmin=qmin+1
                  if(qmin.gt.TrueOrder) qmin=0
                  goto 51
               endif
            enddo
            qmap(Q)=qmin
         enddo   !  Q
      elseif(rlaxnum.eq.0 .and. rlxemin.gt.-1.d99) then
         QQ=0
         do Q = 0, TrueOrder
            if(EigenVal(Q).gt.rlxemin .and. QQ.lt.mxcv) then
               qmap(QQ)=Q
               QQ=QQ+1
               i=Q 
               if(QQ.eq.Bdim) goto 52
            endif
         enddo
   52    Q=i+1
         if(QQ.lt.Bdim) then
            qmap(QQ)=Q
            QQ=QQ+1
            Q=Q+1
            goto 52
         endif
         if(qmap(0).ge.1) then  ! Take the first state below the block
            qmap(QQ)=qmap(0)-1  ! as additional vector (when others
            QQ=QQ+1             ! have converged).
         endif
         if(qmap(0).ge.2) then  ! Take the second state below the block
            qmap(QQ)=qmap(0)-2  ! as additional vector (when others
            QQ=QQ+1             ! have converged).
         endif
         idown=3
   53    if(QQ.lt.mxcv) then    ! Set qmap below and above the block
            if(qmap(0).ge.idown) then
               qmap(QQ)=qmap(0)-idown
               QQ=QQ+1
               idown=idown+1
            endif
            qmap(QQ)=Q          ! Take the states following the block
            QQ=QQ+1             ! as additional vectors (when others
            Q=Q+1               ! have converged).
            goto 53
         endif
      endif      ! (if rlaxnum.eq.920 ...  elseif rlaxnum.eq.0)
      epse=(EigenVal(qmap(Bdim-1))-EigenVal(qmap(0)))/(Bdim*90.d0)

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP  ---
      If( TrueOrder .GT. IntOrder-BDim ) Goto 300
      If( TrueOrder .GE. PsiDim-BDim   ) Goto 300
      If( full .or. (ErrorM+olderr.GT.TolError) ) Goto 99

! --- CONVERGENCE! Build new A-vector ---

 300  if(lmult) then
         call kry2finalpsi(Psi,DtPsi,Krylov,HKrylov,EigenVector,TrueOrder,qmap)
      else
!$omp parallel do if(lompthread.and.(BDim+1).ge.ompthread) num_threads(ompthread) private(QQ,Q,D,P)
         Do QQ = 0,BDim-1
            Q = qmap(QQ) 
            Do D = 1,PsiDim 
               Psi(D+QQ*PsiDim)   =  Krylov(D,0)*EigenVector(0,Q)
               DtPsi(D+QQ*PsiDim) = HKrylov(D,0)*EigenVector(0,Q)
            EndDo
            Do P = 1,TrueOrder
               Do D = 1,PsiDim
                  Psi(D+QQ*PsiDim)   = Psi(D+QQ*PsiDim) + &
                                        Krylov(D,P)*EigenVector(P,Q)
                  DtPsi(D+QQ*PsiDim) = DtPsi(D+QQ*PsiDim) + &
                                       HKrylov(D,P)*EigenVector(P,Q)
               EndDo
            EndDo
         EndDo
!$omp end parallel do
      endif


      Do QQ = 0, BDim-1
         Q = qmap(QQ) 
         if(FirstCall) then
!........This is for first overlaps in wrtrlxbdav
            do j = 0, TrueOrder
               ovlps(j) = EigenVector(Q,j)**2
            enddo
         endif
         E=EigenVal(Q)
         beta=EigenVector(QQ,Q)**2
         j=QQ
         if(beta.lt.0.5d0 .and. TrueOrder.gt.Bdim) then
            beta=0.d0
            do i=0,Bdim+1
               if(EigenVector(i,Q)**2 .gt. beta) then
                  beta=EigenVector(i,Q)**2
                  j=i
               endif
            enddo
            if(beta.lt.0.67d0) then
               beta=EigenVector(QQ,Q)**2
               j=QQ
            endif
         endif
         call wrtrlxbdav(time,trueorder,QQ,Q,E,beta,oldenergy(j,kskip), &
              converged(QQ),EigenVal,ovlps,FirstCall,IntOrder,BDim)
      enddo
      Do QQ = 0, BDim+1
         Q = qmap(QQ)
         oldenergy(QQ,kskip)=EigenVal(Q)
      enddo

      FirstCall = .false.
      full      = .false.
      edav = E

      deallocate(workr)

      Return
      End Subroutine BRDavStep

! **********************************************************************

      end module davlib

