      module davlibpar

use decimal, only: long,dop
use timing
use logdat, only: stopnow

      implicit none

      private
#ifdef MPI
      public :: mpirdavstep,mpicdavstep,mpibdavstep,mpibrdavstep
      contains

! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE MPIRDAVSTEP                             *
! *                                                                    *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming real Hamiltonian and real WF                              *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim).          *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStep").                                         *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (real!)       *
! *   HKrylov:   Matrix containing H*Krylov (real!)                    *
! *   rlxemin:   Lower energy to restrict the search for max. ovlp     *
! *   rlxemax:   Upper energy to restrict the search for max. ovlp     *
! *   precon:    Dimension of Pre-Conditioner                          *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Abs(Time),Psi,DtPsi).                            *
! * WriteStepDav: Writes the energy,beta, and error to a file.         *
! *   Called as WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)      *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 06/2003 HDM                                                   *
! **********************************************************************

      Subroutine mpirDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError, &
                      Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat,Qmat, &
                      EigenVector,EigenVal,ovlps,workr,Rdiagonal,davmat, &
                      indxdav,TrueOrder,ErrorCode,Time,rlxemin, &
                      rlxemax,precon,quad,olsen,Func,Funcr,lrreal,nmod, &
                      lf,worki,workc)

!#ifdef MPI
use mpi
use global
use globmemmod
use mpidata

use getdiagonalmod, only: davovlp1,davovlp2r,getdiagonal
use davutilspar, only: mpiorthonr,mpiovl
use getdavmatmod, only: sortdiagonal,getdavmat,mpigetdavmat
use davutils, only: precondit,orthonr,rsymcoeff,realdaverror
use op1lib, only: zerovxd,cpvxd,zerovxz,cpvxz
use pardsyevmod, only: pardsyev
use mpidsteqmod, only: strtdsteqr
use wrintegrat, only: wrtrlxdav,writestepdav

      Implicit None

      External                                                 :: Func,Funcr
      Logical(kind=4)                                          :: jump,runquad
      Logical(kind=4), save                                    :: full
      Logical(kind=4), intent(in)                              :: quad,olsen, &
                                                                  lrreal
      Logical(kind=4),dimension(5)                             :: mpigoto
      Integer(long), intent(in)                                :: Steps
      Logical(kind=4), intent(in)                              :: lf
      Integer(long), intent(inout)                             :: rlaxnum,precon
      Integer(long), intent(out)                               :: ErrorCode, &
                                                                  TrueOrder
      Integer(long), intent(in)                                :: PsiDim, &
                                                                  IntOrder, &
                                                                  nmod
      Integer(long), dimension(*), intent(out)                 :: worki
      Integer(long),dimension(precon), intent(out)             :: indxdav
      Real(dop), intent(in)                                    :: TolError,Time, &
                                                                  rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)          :: ovlps,EigenVal
      Real(dop), dimension(PsiDim), intent(out)                :: RDiagonal
      Real(dop), dimension(precon,precon+1), intent(out)       :: davmat
      Real(dop), dimension(PsiDim,0:IntOrdermpi), intent(out) :: Krylov
      Real(dop), dimension(PsiDim,0:IntOrdermpi), intent(inout) :: HKrylov
      Real(dop), dimension(((IntOrder+1)*(IntOrder+2))/2), intent(out):: Tmat
      Real(dop), dimension(((IntOrder+1)*(IntOrder+2))/2), intent(out):: Qmat
      Real(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector
      Complex(dop), dimension(2*PsiDim), intent(out)           :: HDiagonal
      Complex(dop), dimension(PsiDim), intent(inout)           :: Psi,DtPsi

      Integer(long)                                         :: D,P,Q,Q1,i,j, &
                                                               shift,count
      Real(dop)                                             :: Error,olderr, &
                                                               ovl,ovl1,ovj,d2, &
                                                               E,E1,beta,beta1,edav
      Real(dop), dimension(3*IntOrder+4), intent(out)       :: workr
      Complex(dop)                                          :: ovlp
      Complex(dop), dimension(*), intent(out)               :: workc

      Logical(kind=4), save :: FirstCall = .true.
      Logical(kind=4)       :: lreturn = .false.

      Integer(long) :: ij,izz,lwork,mpiorder,jbeg,jend,nmpirecv,l,ldz,N, &
                       nmaxit,icompz
      Real(dop) ::     DE
      real(dop) ::     cor1,cor2

      integer(long), dimension(mpi_status_size) :: status
      integer(long), save                       :: tid=0 ! timer ID

!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      mpiorder=0
      mpigoto(1)=.false.
      mpigoto(2)=.false.
      mpigoto(3)=.false.
      mpigoto(4)=.false.
      mpigoto(5)=.false.

!.....Compute the diagonal of the Hamilton matrix.
!     This has been moved here from rank 0
      call getdiagonal(HDiagonal)

! --- CHECK INTEGRATION ORDER ---
      if (mpirank.eq.0) then
         If (IntOrder .LT. 2) Then
            ErrorCode = 1
            write(6,'(a,i4)') 'mpirdavstep: IntOrder =',IntOrder
           lreturn = .true.
         EndIf
      end if

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then
! --- INITIALIZE VARIABLES ---

         if(FirstCall) Then
            if(rlaxnum .ge. 9999) then
               full = .true.
               rlaxnum = rlaxnum - 10000
            else
               full = .false. 
            endif
            if(rlaxnum.lt.0 .or.(rlaxnum.gt.920 .and. rlaxnum.ne.930)) then
               write(6,'(a,i5,l3)') ' DAV : Incorrect rlaxnum',rlaxnum,full
               call stopnow
            endif
         endif
         Q     = 0
         Q1    = 0
         shift = 0
         count = 1
         d2    = 1.e-18_dop
         jump  = .true. 
         ErrorCode = 0
         TrueOrder = 1
         mpiorder = 1
         beta1 = 0
         izz   = 0
         Error = 1.0_dop
         lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
         runquad = quad.and.(.not.full) 

!.....Compute the diagonal of the Hamilton matrix.
! moved above         call getdiagonal(HDiagonal)

!.....Copy complex to real arrays and compute the first trial energy.
         cor1=0.0_dop
         E=0.0_dop
         Do D = 1, PsiDim
            Krylov(D,0)  = dble(  Psi(D))
            HKrylov(D,0) = dble(DtPsi(D))
            Rdiagonal(D) = dble(HDiagonal(D))
            E = E + Krylov(D,0)*HKrylov(D,0)
            cor1=cor1 + abs(dimag(Psi(D))) + abs(dimag(DtPsi(D)))
          enddo
         edav = E

         If(cor1.gt.1.d-14) call realdaverror(cor1)

!.....Write very first energy expectation value to rlx_info file.
         if(FirstCall) then
            E1=E
            ovl =1.0_dop
            ovl1=1.0_dop
            beta=1.0_dop
            call wrtrlxdav(time,Q1,Q,Q1,E,E1,beta,beta,ovl,ovl1, &
                           EigenVal,ovlps,FirstCall,IntOrder)
         endif

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
         if( precon .le. 1 ) precon=0
         if( precon .gt. 0 ) then
            call get_timer(tid, 'Precon')
            call start_timer(tid)

            call sortdiagonal(RDiagonal,workr,indxdav,E,PsiDim,precon)
            call getdavmat(RDiagonal,davmat,davmat(1,precon+1), &
                           indxdav,PsiDim,precon,lwork)
            call stop_timer(tid)
         endif

! --- APPLY INVERSE OF DIAGONAL and ORTHONORMALIZE ---

         olderr=Error
         if( precon .eq. 0 ) then
            Error = 0.0_dop
            Do D = 1, PsiDim
               Error = Error + ( HKrylov(D,0)-E*Krylov(D,0) )**2
               Krylov(D,1) = (HKrylov(D,0)-E*Krylov(D,0)) * &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
         else
            Do D = 1, PsiDim
               Krylov(D,1) = HKrylov(D,0)-E*Krylov(D,0)
            enddo
            call precondit(Krylov(1,1),RDiagonal,davmat, &
                 davmat(1,precon+1),indxdav,E, &
                 Error,d2,PsiDim,precon)
         endif
         Error = Sqrt(Error) 
         if(Error.lT.0.2*TolError) then ! Startvector is already an eigenvector
            lreturn = .true. ! Startvector is already an eigenvector
            beta=1.d0
            Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)
            goto 101
         endif

         call orthonr(Krylov,Krylov(1,1),ovl,0,PsiDim)
!.....Do the orthonormalisation twice.
         call orthonr(Krylov,Krylov(1,1),ovl,0,PsiDim)

! --- WRITE STEP AND ERROR IF DESIRED ---
         beta=1.0_dop
         Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)
      else
         if( precon .gt. 0 ) then
            if(lmpigetdavmat) call mpigetdavmat(RDiagonal,davmat,indxdav, &
                  PsiDim,precon,workr)
            if (precon.gt.2*nmpi.and.lmpidsyev) &
                  call STRTDSTEQR(ldz,l,lendmpi,icompz,n,nmaxit,mr)
         endif
      endif

 101  continue
      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

! --- BUILD HKRYLOV FOR FIRST STEP ---
      If(lrreal) Then
        Call Funcr(Time,Krylov(1,1),HKrylov(1,1))
      Else
         Do D = 1,PsiDim
            HDiagonal(D) = Krylov(D,1)
         enddo
         Call Func(Time,HDiagonal,DtPsi)
         Do D = 1,PsiDim
            HKrylov(D,1) = dble(DtPsi(D))
         enddo
      Endif

      if (mpirank.eq.0) then
! --- BUILD T-MATRIX  ---

         Tmat(1) = E
         Tmat(2) = 0.0_dop
         Tmat(3) = 0.0_dop
         Do D = 1, PsiDim
            Tmat(2) = Tmat(2) + Krylov(D,0)*HKrylov(D,1)
            Tmat(3) = Tmat(3) + Krylov(D,1)*HKrylov(D,1)
         enddo
         if(runquad) then
            Qmat(1) = 0.0_dop
            Qmat(2) = 0.0_dop
            Qmat(3) = 0.0_dop
            Do D = 1, PsiDim
               Qmat(1) = Qmat(1) + HKrylov(D,0)*HKrylov(D,0)
               Qmat(2) = Qmat(2) + HKrylov(D,0)*HKrylov(D,1)
               Qmat(3) = Qmat(3) + HKrylov(D,1)*HKrylov(D,1)
            enddo
         endif

! --- DIAGONALISE T-MATRIX ---
         ij = 0
         do j = 0, TrueOrder
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         enddo
      endif

      if (mpirank.eq.0) then
         call dsyev('V','U',TrueOrder+1,EigenVector,IntOrder+1, &
                                         EigenVal,workr,lwork,ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf
      endif

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then
         If(abs(EigenVector(0,0)) .gt. abs(EigenVector(0,1)) ) Then
            Q = 0
         else
            Q = 1
         endif
         if(rlaxnum.eq.0)  Q = 0 
         if(FirstCall .and. rlaxnum.lt.TrueOrder .and. rlaxnum.le.900) &
              Q = rlaxnum 

         E  = EigenVal(Q)
         beta = EigenVector(0,Q)**2


!.....davovlp1 sets the second part of Hdiagonal to the transformed 
!.....A-vector of psi_ref. The call to davovlp2r sets "set" in davovlp2r. 
         if(rlaxnum.eq.930 .and. .not.FirstCall) then
            call davovlp1(psi,HDiagonal,time)
            call davovlp2r(IntOrder,TrueOrder,ovlps,Krylov,EigenVector, &
                           Hdiagonal(PsiDim+1),FirstCall)
!........Add transformed initial state as extra vector.
            TrueOrder = TrueOrder + 1
            mpiorder = mpiorder + 1
            Do D = 1,PsiDim
               Krylov(D,TrueOrder) = dble(Hdiagonal(PsiDim+D))
            enddo
            Error = 1.d0    ! Dummy Error for lock vector.
            mpigoto(1)=.true.
         endif
      endif

      call MPI_Bcast(runquad,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(mpigoto(1),1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if(mpigoto(1)) then
         mpigoto(1)=.false.
         goto 200
      endif

! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

 100  Continue  !  Iteration Loop

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.trueorder/intordermpi) mpiorder=mpiorder+1

! --- RE-INITIALIZE VARIABLES ---

      if (mpirank.eq.0) then
         TrueOrder = TrueOrder+1
!.....Converge to lower states if sought state already converged.
         if(full .and. (TrueOrder .lt. IntOrder-2)) then
            if(Error.lt.TolError*2.d0) then
               shift = shift + count
               if(shift .gt. Q) then
                  count = -count
                  shift = count
               endif
            endif
            Q1 = Q 
            Q  = Q - shift
            E  = EigenVal(Q)
            if(Error.lt.TolError*2.d0)  then 
               E1    = EigenVal(Q1)
               beta  = EigenVector(0,Q)**2
               beta1 = EigenVector(0,Q1)**2
               call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl, &
                              ovl1,EigenVal,ovlps,FirstCall,IntOrder)
            endif
            Q1 = Q 
         endif
      endif

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

! --- Build residual vector |r> ---
      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(E,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)

 120  call zerovxd(Krylov(1,mpiorder),psidim)
      Do j = 0, mpiorder-1
         Do D = 1, PsiDim
            Krylov(D,mpiorder) = Krylov(D,mpiorder) + &
              EigenVector(intordermpi*mpirank+j,Q) * (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo

      call MPI_Reduce(Krylov(1,MPIOrder),HKrylov(1,MPIOrder),PsiDim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxd(HKrylov(1,MPIOrder),Krylov(1,MPIOrder),PsiDim)


! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      if (mpirank.eq.0) then
         olderr=Error
         if( precon.eq.0 ) then
            Error = 0.d0
            Do D = 1, PsiDim
               Error = Error + Krylov(D,mpiorder)**2
               Krylov(D,mpiorder) = Krylov(D,mpiorder)* &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
         else
            call precondit(Krylov(1,mpiorder),RDiagonal,davmat, &
                           davmat(1,precon+1),indxdav, &
                           E,Error,d2,PsiDim,precon)
         endif
         Error = Sqrt(Error)
      endif


! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction

      if(olsen) then 
! --- Build EigenVector.
         call zerovxd(workr(2*precon+1),psidim)
         Do j = 0, mpiOrder-1
            Do D = 1, PsiDim
               workr(2*precon+D) = workr(2*precon+D) + &
                    EigenVector(intordermpi*mpirank+j,Q) * Krylov(D,j)
            enddo
         enddo
         call MPI_Reduce(workr(2*precon+1),HKrylov(1,mpiorder),PsiDim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         call cpvxd(HKrylov(1,mpiorder),workr(2*precon+1),PsiDim)

         if (mpirank.eq.0) then
            cor1=0.d0
            cor2=0.d0
            if(precon.eq.0) then
               do D = 1, PsiDim
                  cor1 = cor1 + workr(2*precon+D)*Krylov(D,mpiOrder)
                  cor2 = cor2 + workr(2*precon+D)**2 *  &
                       (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
               do D = 1, PsiDim
                  Krylov(D,mpiorder) = Krylov(D,mpiorder) - &
                    (cor1/cor2) * workr(2*precon+D) * &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
            else
               call cpvxd(workr(2*precon+1),HKrylov(1,mpiOrder),PsiDim)
               call precondit(HKrylov(1,mpiOrder),RDiagonal,davmat, &
                              davmat(1,precon+1),indxdav, &
                              E,ovj,d2,PsiDim,precon)
               do D = 1, PsiDim
                  cor1 = cor1 + workr(2*precon+D)*Krylov(D,mpiOrder)
                  cor2 = cor2 + workr(2*precon+D)*HKrylov(D,mpiOrder)
               enddo
               do D = 1, PsiDim
                  Krylov(D,mpiOrder) = Krylov(D,mpiOrder) -  &
                          (cor1/cor2) * HKrylov(D,mpiOrder)
               enddo
            endif
         endif
      endif

! ---  ORTHONORMALIZE (entry for adding "lock" vector) ---

 200  continue
      if (mpirank.eq.0) then
         If((TrueOrder .GE. IntOrder-9) .or. (Error .LT. 2.d0*TolError) ) &
             runquad = .false.
      endif
      call MPI_Bcast(runquad,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if(mpiorder.gt.0) then
         if (mpirank.gt.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonr(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_PRECISION,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                   MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif

      if (mpirank.eq.0) then
         call mpiovl(Krylov(1,MPIOrder),ovl,PsiDim)
         if(ovl.le.1.d-39) then     ! If vectors are linearly dependent
            write(6,'(a,3i5,2e12.4)')'DAV: Steps, Order, Q, ovl, Error', &
                                       Steps,TrueOrder,Q,ovl,Error
            write(2,'(a,3i5,2e12.4)')'DAV: Steps, Order, Q, ovl, Error', &
                                       Steps,TrueOrder,Q,ovl,Error
            TrueOrder = TrueOrder - 1
            if(Error .lt. 0.2d0*TolError) then
               mpigoto(2)=.true.
               goto 666
            endif
            if(izz.gt.2) then
               Q=Q1
               mpigoto(3)=.true.
               goto 666
            endif
            izz=izz+1
            TrueOrder = TrueOrder + 1
            Q = Q + 1
            If( Q .gt. TrueOrder ) Q = Q - 2
            E = EigenVal(Q)
            mpigoto(4)=.true.
            goto 666
         endif
      endif

 666  continue
      call MPI_Bcast(mpigoto,4,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (mpigoto(2)) then
         if (mpirank.eq.trueorder/nmpi+1) mpiorder=mpiorder-1
         mpigoto(2)=.false.
         goto 310
      elseif(mpigoto(3)) then
         if (mpirank.eq.trueorder/nmpi+1) mpiorder=mpiorder-1
         mpigoto(3)=.false.
         goto 300
      elseif(mpigoto(4)) then
         mpigoto(4)=.false.
         goto 120
      endif

!.....Symmetrize coefficients, if desired.
      if (mpirank.eq.0) then
         if(nmod.gt.0) call rsymcoeff(Krylov(1,mpiOrder), &
                   HKrylov(1,mpiOrder),PsiDim,nmod,lf,ovl,TrueOrder)
      endif

!.....Do the orthonormalisation twice
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if(mpiorder.gt.0) then
         if (mpirank.gt.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonr(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_PRECISION,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                    MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif


      if (mpirank.eq.0) call mpiovl(Krylov(1,MPIOrder),ovl,PsiDim)

! --- BUILD HKRYLOV  ---
      call MPI_Bcast(Krylov(1,MPIOrder),psidim,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)

      If(lrreal) Then
         Call Funcr(Time,Krylov(1,mpiOrder),HKrylov(1,mpiOrder))
      Else
         Do D = 1,PsiDim
            HDiagonal(D) = Krylov(D,mpiorder)
         enddo
         Call Func(Time,HDiagonal,DtPsi)
         Do D = 1,PsiDim
            HKrylov(D,mpiorder) = dble(DtPsi(D))
         enddo
      Endif

      call MPI_Bcast(HKrylov(1,MPIOrder),psidim,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)

! --- WRITE STEP AND ERROR IF DESIRED ---
      if (mpirank.eq.0) then
         Call WriteStepDav(Steps,TrueOrder,E,beta,Error,Q,Time)
      endif

! --- BUILD NEW T-MATRIX ELEMENTS ---
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(ij,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      if (mpirank.gt.0) call zerovxd(Tmat,ij)
      jbeg=intordermpi*mpirank
      if (mpirank.lt.trueorder/intordermpi) then
         jend=intordermpi*(mpirank+1)-1
      elseif (mpirank.eq.trueorder/intordermpi) then
         jend=trueorder
      else
         jend=-1
      endif

      Do j = 0, TrueOrder
         ij = ij + 1
         Tmat(ij) = 0.d0
         if(runquad) then
            Qmat(ij) = 0.d0
            if (j.ge.jbeg.and.j.le.jend) then
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij)+  Krylov(D,j-jbeg)*HKrylov(D,mpiOrder)
                  Qmat(ij) = Qmat(ij)+ HKrylov(D,j-jbeg)*HKrylov(D,mpiOrder)
               enddo
            endif
         else
            if (j.ge.jbeg.and.j.le.jend) then
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij) + Krylov(D,j-jbeg)*HKrylov(D,mpiOrder)
               enddo
            endif
         endif
      enddo

      call MPI_Reduce(Tmat,workr,ij,MPI_DOUBLE_PRECISION,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxd(workr,Tmat,ij)

! --- DIAGONALISE T-MATRIX ---
      if (mpirank.eq.0) then
         ij = 0
         do j = 0, TrueOrder
            if(runquad) then
               do i = 0, j
                  ij = ij + 1
                  EigenVector(i,j) = Qmat(ij)-2.d0*E*Tmat(ij)
               enddo
               EigenVector(j,j) = EigenVector(j,j) + E*E
            else
               do i = 0, j
                  ij = ij + 1
                  EigenVector(i,j) = Tmat(ij)
               enddo
            endif
         enddo
      endif

      if (mpirank.eq.0) then
         call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         call pardsyev('V','U',TrueOrder+1,EigenVector,IntOrder+1, &
                       EigenVal,workr,lwork,ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf
      else
         call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         if ((trueorder+1).gt.2*nmpi.and.lmpidsyev) &
            call STRTDSTEQR(ldz,l,lendmpi,icompz,n,nmaxit,mr)
      endif

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS sought) -
      if (mpirank.eq.0) then
         if(rlaxnum.eq.0) then
            Q = 0 
            beta= EigenVector(0,0)**2
         elseif(FirstCall .and. rlaxnum.lt.TrueOrder .and. rlaxnum.le.900)  then
            Q = rlaxnum 
            beta= EigenVector(0,Q)**2
         else
            beta=0.d0
            Do j = 0, TrueOrder
               beta1= EigenVector(0,j)**2
               If(beta1 .gt. beta) then
                  Q = j
                  beta = beta1
               endif
            enddo
         endif
         if(runquad) then
            E = 0.d0
            ij = 0
            do j = 0, TrueOrder
               do i = 0, j
                  ij = ij + 1
                  DE = EigenVector(i,Q)*Tmat(ij)*EigenVector(j,Q)
                  E = E + 2.d0*DE
               enddo
               E = E - DE
            enddo
         else
            E = EigenVal(Q)
         endif
         Q1 = Q
         E1 = E
         beta1 = beta
         ovl=0.d0
      endif

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Error,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(jump,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP if not lock ---
      If( TrueOrder .EQ. PsiDim-1 ) Goto 310
      If((TrueOrder.LT.IntOrder) .And.  (full .or. &
        ( (Error.GT.TolError*10.d0) .And. jump)) )  Goto 100


! --- COMPUTE EXACT OVERLAPS

 310  continue
      if (mpirank.eq.0) then
         if(rlaxnum.eq.930) then   ! lock
!........The array ovlps contains the overlaps**2 with startvector.
            call davovlp2r(IntOrder,TrueOrder,ovlps,Krylov,EigenVector, &
                           Hdiagonal(PsiDim+1),FirstCall)
            Q=-1
            do j = 0, TrueOrder
               ovj = ovlps(j)
               if(j.eq.Q1) ovl1 = ovj
               if(EigenVal(j).lt.rlxemin .or. EigenVal(j).gt.rlxemax) ovj=0.d0  
               If(ovj .gt. ovl) then
                  Q = j
                  ovl = ovj
               endif
            enddo
            if( Q .eq. -1 ) then
               E = 0.5d0*(rlxemin+rlxemax)
               ovj = 1.d50
               do j = 0, TrueOrder
                  if(abs(E-EigenVal(j))/(ovlps(j)+1.d-4) .lt. ovj) then
                     Q = j
                     ovj=abs(E-EigenVal(j))/(ovlps(j)+1.d-4)
                  endif
               enddo
               ovl = ovlps(Q)
               E = EigenVal(Q)
               write(18,'(a,3f17.6)') '## No eigenvalue in window:', &
                    EigenVal(max(0,Q-1))*27.2114d0, E*27.2114d0, &
                    EigenVal(Q+1)*27.2114d0
            endif
            if(runquad) then
               E = 0.d0
               ij = 0
               do j = 0, TrueOrder
                  do i = 0, j
                     ij = ij + 1
                     DE = EigenVector(i,Q)*Tmat(ij)*EigenVector(j,Q)
                     E = E + 2.d0*DE
                  enddo
                  E = E - DE
               enddo
            else
               E = EigenVal(Q)
            endif
            beta = EigenVector(0,Q)**2
            If((Q1.ne.Q) .and. (TrueOrder.LT.IntOrder) .and. jump ) then
               jump = .false.  
               call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl, &
                              ovl1,EigenVal,ovlps,FirstCall,IntOrder)
               mpigoto(5)=.true.
            endif
         endif
      endif

      call MPI_Bcast(mpigoto(5),1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (mpigoto(5)) then
         mpigoto(5)=.false.
         goto 100
      endif

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Error,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(olderr,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)


! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If( TrueOrder .EQ. IntOrder ) Goto 300
      If( TrueOrder .EQ. PsiDim-1 ) Goto 300
      If( runquad .or. full .or. (Error+olderr.GT.TolError) ) Goto 100


! --- CONVERGENCE! Build new A-vector ---
 300  continue
      call zerovxz(Psi,psidim)
      call zerovxz(dtpsi,psidim)
      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)
      Do P = 0,mpiorder-1
         Do D = 1,PsiDim
            Psi(D) = Psi(D) + Krylov(D,P)*EigenVector(intordermpi*mpirank+P,Q)
            DtPsi(D) = DtPsi(D) + HKrylov(D,P)*EigenVector(intordermpi*mpirank+P,Q)
         EndDo
      EndDo
      if (mpirank.eq.0) then
         Do D = 1,PsiDim
            Psi(D) = Psi(D) + Krylov(D,mpiorder)*EigenVector(trueorder,Q)
            DtPsi(D)=DtPsi(D) + HKrylov(D,mpiorder)*EigenVector(trueorder,Q)
         EndDo
      endif

      call MPI_Reduce(Psi,workc,psidim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,Psi,psidim)
      call MPI_Reduce(DtPsi,workc,psidim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,DtPsi,psidim)

      if (mpirank.eq.0) then
         If(Q.ne.Q1) call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1, &
                             ovl,ovl1,EigenVal,ovlps,.false.,IntOrder)

         if(FirstCall .and. (rlaxnum.ne.930)) then
!.....This is for first overlaps in wrtrlxdav
            do j = 0, TrueOrder
               ovlps(j) = EigenVector(0,j)**2
            enddo
         endif
         Q1 = Q
         call wrtrlxdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,ovl1, &
                        EigenVal,ovlps,FirstCall,IntOrder)

         FirstCall = .false.
         full      = .false.
         edav = E
      endif

!#endif

      Return
      End Subroutine mpirDavStep

! **********************************************************************
! *                                                                    *
! *              SUBROUTINE MPIcDAVSTEP                                *
! *                                                                    *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The (complex) initial-value vector.                   *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim).          *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStep").                                         *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors.               *
! *   HKrylov:   Matrix containing H*Krylov                            *
! *   rlxemin:   Lower energy to restrict the search for max. ovlp     *
! *   rlxemax:   Upper energy to restrict the search for max. ovlp     *
! *   rlximax:   Lower Im(energy) to restrict the s. for max. ovlp     *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Abs(Time),Psi,DtPsi)                             *
! *   WriteStep: Writes the stepsize and error to a file.              *
! *              Called as WriteStep(Steps,Order,Stepsize,Error).      *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 07/2003 HDM                                                   *
! **********************************************************************

      Subroutine MPIcDavStep(Psi,DtPsi,PsiDim,IntOrder,TolError, &
                          Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat, &
                          TTmat,EigenVector,EigenVal,workc,workr,ovlps, &
                          Stepsize,TrueOrder,ErrorCode,Time,rlxemin, &
                          rlxemax,rlximin,precon,quad,olsen,Func)

!#ifdef MPI
use mpi
use global
use mpidata

use op1lib, only: zerovxz,cpvxz,zeromxz
use getdiagonalmod, only: getdiagonal,davovlp1,davovlp2c
use wrintegrat, only: wrtrlxcdav,writestepcdav

      Implicit None

      External                                               :: Func
      Logical(kind=4)                                        :: jump
      Logical(kind=4), save                                  :: full
      Logical(kind=4), save                                  :: FirstCall = .true.
      Logical(kind=4)                                        :: lreturn = .false.
      Logical(kind=4)                                        :: runquad,mpigoto(5)
      Logical(kind=4), intent(in)                            :: quad,olsen
      Integer(long), intent(in)                              :: Steps
      Integer(long), intent(inout)                           :: rlaxnum
      Integer(long), intent(out)                             :: ErrorCode, &
                                                                TrueOrder
      Integer(long), intent(in)                              :: PsiDim, &
                                                                IntOrder, &
                                                                precon
      Real(dop), intent(in)                                  :: Stepsize
      Real(dop), intent(in)                                  :: TolError,Time, &
                                                                rlxemin,rlxemax,&
                                                                    rlximin
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps
      Complex(dop), dimension(0:IntOrder), intent(inout)         :: EigenVal
      Complex(dop), dimension(PsiDim), intent(inout)             :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,IntOrdermpi), intent(out)   :: Krylov
      Complex(dop), dimension(PsiDim,IntOrdermpi), intent(inout) :: HKrylov
      Complex(dop), dimension(2*PsiDim), intent(out)             :: HDiagonal
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: Tmat
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: TTmat
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector

      Integer(long)                                         :: D,P,Q,Q1,i,j, &
                                                               shift,count
      Real(dop)                                             :: Error,olderr, &
                                                               ovl,ovl1,ovj,d2, &
                                                               beta,beta1
      Real(dop), dimension(2*IntOrder+4), intent(out)       :: workr
      Complex(dop)                                          :: ovlp,E,E1
      Complex(dop), dimension(*), intent(out)               :: workc

      Integer(long) :: izz,lwork,mpiorder,nmpirecv
      real(dop)     :: cor1,cor2
      Complex(dop), dimension(1,1) :: vl

      integer(long), dimension(mpi_status_size) :: status


!---- WARNING: if(.not.lock) HDiagonal(PsiDim)

      mpiorder=1
      mpigoto(1)=.false.
      mpigoto(2)=.false.
      mpigoto(3)=.false.
      mpigoto(4)=.false.
      mpigoto(5)=.false.

!.....Compute the diagonal of the Hamilton matrix.
!     This has been moved here from rank 0
      call getdiagonal(HDiagonal)

      if (mpirank.eq.0) then
! --- CHECK INTEGRATION ORDER ---

         If (IntOrder .LT. 2) Then
            ErrorCode = 1
            write(6,'(a,i4)') 'IntOrder =',IntOrder
            lreturn = .true.
         EndIf
      end if

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then
! --- INITIALIZE VARIABLES ---

         if(FirstCall) Then
            if(rlaxnum .ge. 19999) then
               write(6,'(a)') 'ortho not allowed for DAV'
               call stopnow
            endif
            if(rlaxnum .ge. 9999) then
               full = .true.
               rlaxnum = rlaxnum - 10000
            else
               full = .false.
            endif
            if(rlaxnum.lt.0 .or. (rlaxnum.gt.920 .and. rlaxnum.ne.930)) then
               write(6,'(a,i5,l3)') ' DAV : Incorrect rlaxnum',rlaxnum,full
               call stopnow
            endif
         endif
         if(quad) then
            write(6,'(a)') ' quad not implemented for CDAV'
            call stopnow
         endif
         if(olsen) then
            write(6,'(a)') ' olsen not implemented for CDAV'
            call stopnow
         endif
         if(precon.gt.0) then
            write(6,'(a)') ' PreConditioner not implemented for CDAV'
            call stopnow
         endif
         shift = 0
         count = 1
         d2 = 1.d-12
         jump = .true. 
         ErrorCode = 0
         TrueOrder = 1
         izz = 0

!.....davovlp1 sets the second part of Hdiagonal to the transformed 
!.....A-vector of psi_ref. (cDavStep) 
         if(rlaxnum.eq.930) then
            call davovlp1(psi,HDiagonal,time)
         endif

!.....Compute the diagonal of the Hamilton matrix.
! moved above         call getdiagonal(HDiagonal)

!.....Compute the first trial energy.
         E=0.d0
         Do D = 1, PsiDim
            E = E + DConjg(Psi(D))*DtPsi(D)
         enddo

!.....Write very first energy expectation value to rlx_info file.
         Q=0
         Q1=0
         if(FirstCall) then
            E1=E
            ovl=1.d0
            ovl1=1.d0
            beta=1.d0
            call wrtrlxcdav(time,Q1,Q,Q1,E,E1,beta,beta,ovl,ovl1, &
                              EigenVal,ovlps,FirstCall,IntOrder)
         endif

! --- APPLY INVERSE OF DIAGONAL and ORTHONORMALIZE ---

         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + DConjg(DtPsi(D)-E*Psi(d))*(DtPsi(D)-E*Psi(D))
            Krylov(D,1) = (DtPsi(D)-E*Psi(D)) * &
                DConjg(HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E) &
                                                 *(HDiagonal(D)-E)+d2)
         enddo
         Error = Sqrt(Error) 
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Psi(D))*Krylov(D,1)
         enddo
         Do D = 1,PsiDim
            Krylov(D,1)=Krylov(D,1)-ovlp*Psi(D)
         enddo
         ovl = 0.d0
         Do D = 1,PsiDim
            ovl = ovl + DConjg(Krylov(D,1))*Krylov(D,1)
         enddo
         Do D = 1,PsiDim
            Krylov(D,1)=Krylov(D,1)/sqrt(ovl)
         enddo
!.....Orthonormalize twice.
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Psi(D))*Krylov(D,1)
         enddo
         Do D = 1,PsiDim
            Krylov(D,1)=Krylov(D,1)-ovlp*Psi(D)
         enddo
         ovl = 0.d0
         Do D = 1,PsiDim
            ovl = ovl + DConjg(Krylov(D,1))*Krylov(D,1)
         enddo
         Do D = 1,PsiDim
            Krylov(D,1)=Krylov(D,1)/sqrt(ovl)
         enddo

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
         beta=1.d0
         Call WriteStepCDav(Steps,TrueOrder,E,beta,Error,Q,Time)
      endif

! --- BUILD HKRYLOV FOR FIRST STEP ---
      Call Func(Time,Krylov(1,1),HKrylov(1,1))

      if(mpirank.eq.0) then
! --- BUILD T-MATRIX (cDavStep) ---

         call zeromxz(Tmat,intorder+1,intorder+1)
         Tmat(0,0) = E
         Tmat(0,1) = (0.d0,0.d0)
         Tmat(1,0) = (0.d0,0.d0)
         Tmat(1,1) = (0.d0,0.d0)
         Do D = 1, PsiDim
            Tmat(1,0) = Tmat(1,0) + DConjg(Krylov(D,1))*DtPsi(D)
            Tmat(0,1) = Tmat(0,1) + DConjg(Psi(D))*HKrylov(D,1)
            Tmat(1,1) = Tmat(1,1) + DConjg(Krylov(D,1))*HKrylov(D,1)
         enddo

! --- DIAGONALISE T-MATRIX ---

         do j = 0, TrueOrder
            do i = 0, TrueOrder
               TTmat(i,j) = Tmat(i,j)
            enddo 
         enddo

         call ZGEEV( 'N', 'V', TrueOrder+1, TTmat, IntOrder+1, &
                      EigenVal, vl, 1, EigenVector, IntOrder+1, &
                      workc, 16*IntOrder+18, workr, ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf

      end if

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then

         If(abs(EigenVector(0,0)) .gt. abs(EigenVector(0,1)) ) Then
            Q = 0
         else
            Q = 1 
         endif
         if(rlaxnum.eq.0)  Q = 0 

         E = EigenVal(Q)
         beta = abs(EigenVector(0,Q))**2

         if(rlaxnum.eq.930 .and. .not.FirstCall) then
!........The call to davovlp2r sets "set" in davovlp2c.  
            call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov, &
                           EigenVector,Hdiagonal(PsiDim+1),FirstCall)
!........Add transformed initial state as extra vector.
            TrueOrder = TrueOrder + 1
            mpiorder  = mpiorder  + 1
            Do D = 1,PsiDim
               Krylov(D,TrueOrder) = dble(Hdiagonal(PsiDim+D))
            enddo
            Error = 1.d0
            mpigoto(1)=.true.
         endif
      endif

      call MPI_Bcast (mpigoto,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if(mpigoto(1)) then
         mpigoto(1)=.false.
         goto 200
      endif

! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

 100  Continue  !  Iteration Loop  (cDavStep)

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.(trueorder-1)/(intordermpi-1)) mpiorder=mpiorder+1

      if (mpirank.eq.0) then
! --- RE-INITIALIZE VARIABLES ---

         TrueOrder = TrueOrder+1

!.....Converge to lower states if seeked state already converged.
         if(full .and. (TrueOrder .lt. IntOrder-2)) then
            if(Error.lt.TolError*2.d0) then
               shift = shift + count
               if(shift .gt. Q) then
                  count = -count
                  shift = count
               endif
            endif
            Q1 = Q
            Q  = Q - shift
            E  = EigenVal(Q)
            if(Error.lt.TolError*2.d0)  then
               E1    = EigenVal(Q1)
               beta  = EigenVector(0,Q)**2
               beta1 = EigenVector(0,Q1)**2
               call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1, &
                        ovl,ovl1,EigenVal,ovlps,FirstCall,IntOrder)
            endif
            Q1 = Q
         endif
      endif
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

! --- Build Eigenvector ---

 120  continue

      call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(E,1,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(EigenVector,(intorder+1)**2,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.0) then
         Do D = 1, PsiDim
            HKrylov(D,mpiOrder) = EigenVector(0,Q)*(DtPsi(D)-E*Psi(D))
         enddo
      else
         call zerovxz(HKrylov(1,mpiOrder),PsiDim)
      endif

      Do j = 1, mpiOrder-1
         Do D = 1, PsiDim
            HKrylov(D,mpiOrder) = HKrylov(D,mpiOrder) + &
                              EigenVector((intordermpi-1)*mpirank+j,Q) * &
                                          (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo

      call MPI_Reduce(HKrylov(1,mpiorder),Krylov(1,mpiOrder),PsiDim, &
                     MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Krylov(1,MPIOrder),psidim,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.0) then
! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---

         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + DConjg(Krylov(D,mpiOrder))*Krylov(D,mpiOrder)
            Krylov(D,mpiOrder) = Krylov(D,mpiOrder)* &
                  DConjg(HDiagonal(D)-E)/(DConjg(HDiagonal(D)-E) &
                                                 *(HDiagonal(D)-E)+d2)
         enddo
         Error = Sqrt(Error)
      endif


 200  continue
      if (mpirank.eq.0) then
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Psi(D))*Krylov(D,mpiOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,mpiOrder)=Krylov(D,mpiOrder)-ovlp*Psi(D)
         enddo
      endif

! ---  ORTHONORMALIZE (shortcut for loop) ---

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if(mpiorder.gt.1) then
         if (mpirank.gt.0.and.trueorder.gt.intordermpi) then
            call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)
         endif

         Do i = 1, mpiorder-1
            ovlp =(0.d0,0.d0)
            Do D = 1,PsiDim
               ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,mpiOrder)
            enddo
            Do D = 1,PsiDim
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - ovlp*Krylov(D,i)
            enddo
         enddo

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if ((trueorder-1).eq.(mpirank+1)*(intordermpi-1)) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                        MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD, &
                                                                 mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                               MPI_DOUBLE_COMPLEX,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                 MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-2)/(intordermpi-1)
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) then
            call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
         endif
      endif

      if (mpirank.eq.0)then
         ovl = 0.d0
         Do D = 1,PsiDim
            ovl = ovl+DConjg(Krylov(D,mpiOrder))*Krylov(D,mpiOrder)
         enddo
         if(ovl.le.1.d-49) then     ! If vectors are linearly dependent
            write(6,'(a,3i5,2e12.4)') 'cDAV: Steps, Order, Q, ovl, Error', &
                                       Steps,TrueOrder,Q,ovl,Error
            write(2,'(a,3i5,2e12.4)') 'cDAV: Steps, Order, Q, ovl, Error', &
                                       Steps,TrueOrder,Q,ovl,Error
            write(6,'(a,i5,f11.3)') 'izz, Time',izz, Time/41.3414d0
            if(izz.gt.2) then
               write(6,'(2a,f11.3)') 'Vectors are linearly dependent.', &
                            ' Assume convergence. Time =',Time/41.3414d0
               Q=Q1
               mpigoto(2)=.true.
               goto 666
            endif
            izz=izz+1
            TrueOrder = TrueOrder - 1
            if(Error .lt. 0.2d0*TolError) then
               mpigoto(3)=.true.
               goto 666
            endif
            TrueOrder = TrueOrder + 1
            Q = Q + 1
            If( Q .gt. TrueOrder ) Q = Q - 2
            E = EigenVal(Q)
            mpigoto(4)=.true.
            goto 666
         endif
      endif
 666  continue
      call MPI_Bcast(mpigoto,4,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (mpigoto(2)) then
         mpigoto(2)=.false.
         goto 300
      elseif (mpigoto(3)) then
         if (mpirank.eq.trueorder/nmpi+1) mpiorder=mpiorder-1
         mpigoto(3)=.false.
         goto 310
      elseif (mpigoto(4)) then
         mpigoto(4)=.false.
         goto 120
      endif
      if (mpirank.eq.0)then
         Do D = 1,PsiDim
            Krylov(D,mpiOrder)=Krylov(D,mpiOrder)/sqrt(ovl)
         enddo
         ovlp =(0.d0,0.d0)
         Do D = 1,PsiDim
            ovlp = ovlp + DConjg(Psi(D))*Krylov(D,mpiOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,mpiOrder)=Krylov(D,mpiOrder)-ovlp*Psi(D)
         enddo
      endif

!.....Orthonormalize twice.  (cDavStep)

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if(mpiorder.gt.1) then
         if (mpirank.gt.0.and.trueorder.gt.intordermpi) then
            call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)
         endif

         Do i = 1, mpiorder-1
            ovlp =(0.d0,0.d0)
            Do D = 1,PsiDim
               ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,mpiOrder)
            enddo
            Do D = 1,PsiDim
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - ovlp*Krylov(D,i)
            enddo
         enddo

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if ((trueorder-1).eq.(mpirank+1)*(intordermpi-1)) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                        MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                               MPI_DOUBLE_COMPLEX,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                 MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-2)/(intordermpi-1)
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) then
            call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
         endif
      endif

      if (mpirank.eq.0)then
         ovl = 0.d0
         Do D = 1,PsiDim
            ovl = ovl+DConjg(Krylov(D,mpiOrder))*Krylov(D,mpiOrder)
         enddo
         Do D = 1,PsiDim
            Krylov(D,mpiOrder)=Krylov(D,mpiOrder)/sqrt(ovl)
         enddo
      endif

      if (mpirank.eq.0) then
! --- WRITE STEPSIZE AND ERROR IF DESIRED ---
         Call WriteStepCDav(Steps,TrueOrder,E,beta,Error,Q,Time)
      endif

      call MPI_Bcast(Krylov(1,MPIOrder),psidim,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)
! --- BUILD HKRYLOV  ---
      Call Func(Time,Krylov(1,mpiOrder),HKrylov(1,mpiOrder))
      call MPI_Bcast(HKrylov(1,MPIOrder),psidim,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)

! --- BUILD NEW T-MATRIX ELEMENTS ---

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      if (mpirank.eq.0) then
         Tmat(0,TrueOrder) = (0.d0,0.d0)
         Tmat(TrueOrder,0) = (0.d0,0.d0)
         Tmat(TrueOrder,TrueOrder) = (0.d0,0.d0)
          Do D = 1, PsiDim
            Tmat(0,TrueOrder) = Tmat(0,TrueOrder) + &
              DConjg(Psi(D))*HKrylov(D,mpiOrder)
            Tmat(TrueOrder,0) = Tmat(TrueOrder,0) + &
              DConjg(Krylov(D,mpiOrder))*DtPsi(D)
            Tmat(TrueOrder,TrueOrder) = Tmat(TrueOrder,TrueOrder) + &
              DConjg(Krylov(D,mpiOrder))*HKrylov(D,mpiOrder)
         enddo
      else
         call zeromxz(Tmat,intorder+1,intorder+1)
      endif

      Do j = 1, mpiOrder-1
         Tmat(mpirank*(intordermpi-1)+j,TrueOrder) = (0.d0,0.d0)
         Tmat(TrueOrder,mpirank*(intordermpi-1)+j) = (0.d0,0.d0)
         Do D = 1, PsiDim
            Tmat(mpirank*(intordermpi-1)+j,TrueOrder) = &
                 Tmat(mpirank*(intordermpi-1)+j,TrueOrder) + &
                 DConjg(Krylov(D,j))*HKrylov(D,mpiOrder)
            Tmat(TrueOrder,mpirank*(intordermpi-1)+j) = &
                 Tmat(TrueOrder,mpirank*(intordermpi-1)+j) + &
                 DConjg(Krylov(D,mpiOrder))*HKrylov(D,j)
         enddo
      enddo

      call MPI_Reduce(Tmat,TTmat,(intorder+1)**2,MPI_DOUBLE_COMPLEX, &
                                        MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxz(TTmat,Tmat,(intorder+1)**2)

      if(mpirank.eq.0) then
! --- DIAGONALISE T-MATRIX (cDavStep) ---

         do j = 0, TrueOrder
            do i = 0, TrueOrder
               TTmat(i,j) = Tmat(i,j)
            enddo
         enddo

         call ZGEEV( 'N', 'V', TrueOrder+1, TTmat, IntOrder+1, &
                      EigenVal, vl, 1, EigenVector, IntOrder+1, &
                      workc, 16*IntOrder+18, workr, ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf

      end if

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then
! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS seeked) -

         if(rlaxnum.eq.0) then
            Q = 0 
            beta= abs(EigenVector(0,0))**2
         else
            beta=0.d0
            Do j = 0, TrueOrder
               beta1= abs(EigenVector(0,j))**2
               If(beta1 .gt. beta) then
                  Q = j
                  beta = beta1
               endif
            enddo
         endif
         E = EigenVal(Q)
         Q1 = Q
         E1 = E
         beta1 = beta
         ovl=0.d0
      endif

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Error,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(jump,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP ---
      If((TrueOrder.LT.IntOrder) .And.  (full .or. &
           ( (Error.GT.TolError) .And. jump)) )  Goto 100

 310  continue

      if(mpirank.eq.0)then
! --- COMPUTE EXACT OVERLAPS

         if(rlaxnum.eq.930) then   ! lock
!........The array ovlps contains the overlaps**2 with startvector.
            call davovlp2c(IntOrder,TrueOrder,ovlps,Psi,Krylov, &
                           EigenVector,Hdiagonal(PsiDim+1),FirstCall)
            do j = 0, TrueOrder
               ovj = ovlps(j)
               if(j.eq.Q1) ovl1 = ovj
               if (dble(EigenVal(j)).lt.rlxemin .or. &
                   dimag(EigenVal(j)).lt.rlximin .or. &
                   dble(EigenVal(j)).gt.rlxemax) ovj=0.d0  
               If(ovj .gt. ovl) then
                  Q = j
                  ovl = ovj
               endif
            enddo
            E = EigenVal(Q)
            beta = abs(EigenVector(0,Q))**2
            If((Q1.ne.Q) .and. (TrueOrder.LT.IntOrder) .and. jump ) then
               jump = .false.  
               call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1, &
                        ovl,ovl1,EigenVal,ovlps,FirstCall,IntOrder)
               mpigoto(5)=.true.
            endif
            If((TrueOrder.LT.IntOrder) .And. (Error.GT.TolError)) mpigoto(5)=.true.
         endif
      endif

      call MPI_Bcast(mpigoto(5),1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if(mpigoto(5)) then
         mpigoto(5)=.false.
         goto 100
      endif

      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Error,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(jump,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)

      If( TrueOrder .EQ. IntOrder ) Goto 300

! --- Build Eigenvector ---

! Hier muss eventuell durch intordermpi-1 dividiert werden
      if (mpirank.eq.trueorder/intordermpi) mpiorder=mpiorder+1

      if (mpirank.eq.0) then
         TrueOrder = TrueOrder + 1
      endif
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.0) then
         Do D = 1, PsiDim
            HKrylov(D,mpiOrder) = EigenVector(0,Q)*(DtPsi(D)-E*Psi(D))
         enddo
      else
         call zerovxz(HKrylov(1,mpiOrder),PsiDim)
      endif

      Do j = 1, mpiOrder-1
         Do D = 1, PsiDim
            HKrylov(D,mpiOrder) = HKrylov(D,mpiOrder) + &
                              EigenVector((intordermpi-1)*mpirank+j,Q) * &
                                          (HKrylov(D,j) - E*Krylov(D,j))
         enddo
      enddo

      call MPI_Reduce(HKrylov(1,mpiorder),Krylov(1,mpiOrder),PsiDim, &
                     MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Krylov(1,MPIOrder),psidim,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)

! --- DETERMINE NEW ERROR and APPLY INVERSE OF DIAGONAL ---

      if (mpirank.eq.0) then
         Error = 0.d0
         Do D = 1, PsiDim
            Error = Error + DConjg(Krylov(D,mpiOrder))*Krylov(D,mpiOrder)
            Krylov(D,mpiOrder) = Krylov(D,mpiOrder)* &
                                 DConjg(HDiagonal(D)-E)/ &
                           (DConjg(HDiagonal(D)-E)*(HDiagonal(D)-E)+d2)
         enddo
         Error = Sqrt(Error)
      endif

      call MPI_Bcast(Krylov(1,MPIOrder),psidim,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Error,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD, &
                                                                 mpierr)

      If( Error .GT. TolError ) Goto 200


! --- CONVERGENCE! Build new A-vector (cDavStep) ---

      if (mpirank.eq.0) then
         TrueOrder = TrueOrder - 1
      endif
 300  continue
      if (mpirank.eq.0) then
         Do D = 1,PsiDim
            Psi(D)   = EigenVector(0,Q)*Psi(D)
            DtPsi(D) = EigenVector(0,Q)*DtPsi(D)
         EndDo
      else
         call zerovxz(psi,psidim)
         call zerovxz(dtpsi,psidim)
      endif
      Do P = 1,mpiOrder-1
         Do D = 1,PsiDim
            Psi(D)   =  Psi(D)+EigenVector(mpirank*(intordermpi-1)+P,Q)*Krylov(D,P)
            DtPsi(D) =DtPsi(D)+EigenVector(mpirank*(intordermpi-1)+P,Q)*HKrylov(D,P)
         EndDo
      EndDo

      call MPI_Reduce(Psi,Krylov(1,mpiOrder),PsiDim, &
                     MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call MPI_Reduce(DtPsi,HKrylov(1,mpiOrder),PsiDim, &
                     MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxz(Krylov(1,mpiorder),Psi,psidim)
      call cpvxz(HKrylov(1,mpiorder),DtPsi,psidim)


      if (mpirank.eq.0) then
         If(Q.ne.Q1) call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta, &
                         beta1,ovl,ovl1,EigenVal,ovlps,.false.,IntOrder)

         if(FirstCall .and. (rlaxnum.ne.930)) then
!.....This is for first diag. wrtrlxdav
            do j = 0, TrueOrder
               ovlps(j) = Abs(EigenVector(0,j))**2
            enddo
         endif
         do j = 0, TrueOrder
            Hdiagonal(j+1) = EigenVal(j)   ! because Eigenval is on workc
         enddo
         Q1 = Q
         call wrtrlxcdav(time,trueorder,Q,Q1,E,E1,beta,beta1,ovl,ovl1, &
                           Hdiagonal,ovlps,FirstCall,IntOrder)
         FirstCall = .false.
         full      = .false.
      endif
!#endif

      Return
      End Subroutine mpicDavStep


! **********************************************************************
! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE MPIBDAVSTEP                             *
! *                                                                    *
! * Block-Davidson diagonalisation. (single-set multi-packet run).     *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming hermitian Hamiltonian and complex WF.                     *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The block of (complex) initial-value vectors.         *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim/npacket)   *
! *   BDim       Block size. BDim=npacket (=gdim(feb)                  *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStepBDAv").                                     *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (complex)     *
! *   HKrylov:   Matrix containing H*Krylov (complex)                  *
! *   precon:    Dimension of Pre-Conditioner                          *
! *   davmat:    Contains the eigenvectors and eigenvalues of the      *
! *              preconditioner.                                       *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Abs(Time),Psi,DtPsi).                            *
! * WriteStepBDav: Writes the energy,beta, and error to a file.        *
! *   Called as WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)  *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.4 03/2010 HDM                                                   *
! **********************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!   If you do some changes in here, please do them in bdavstep too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Subroutine MPIBDavStep(Psi,DtPsi,PsiDim,IntOrder, &
                   TolError,Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat, &
                   EigenVector,EigenVal,ovlps,workr,Rdiagonal,davmat, &
                   TrueOrder,ErrorCode,Time,precon,olsen, &
                   Func,rlxemin,rlxemax,Bdim,nmod,lf, &
                   indxdav,converged,qmap)

!#ifdef MPI
use mpi
use global
use maxv, only: maxdim, maxsta
use griddatmod
use psidef
use mpidata

use op1lib, only: zerovxd,zerovxz,cpvxz
use getdiagonalmod, only: bgetdiagonal
use davutils, only: cprecondit,csymcoeff
use davutilspar, only: mpiorthonc,mpiovlc,mpipsi2kryc,mpikryc2psi,mpikryc2finalpsi
use wrintegrat, only: writestepbdav,wrtrlxbdav

      Implicit None

      External                                                   :: Func
      Integer(long), intent(in)                                  :: PsiDim, &
                                                                    IntOrder, &
                                                                    nmod
      Integer(long), intent(in)                                  :: Steps
      Integer(long), intent(in)                                  :: BDim
      Integer(long), intent(inout)                               :: precon
      Integer(long), dimension(precon), intent(in)               :: indxdav
      Integer(long), dimension(0:8*BDim+9), intent(out)          :: converged,qmap
      Integer(long), intent(inout)                               :: rlaxnum
      Integer(long), intent(out)                                 :: ErrorCode, &
                                                                    TrueOrder
      Complex(dop), dimension(adim), intent(inout)               :: Psi,DtPsi
      Complex(dop), dimension(PsiDim,0:IntOrdermpi), intent(out)   :: Krylov
      Complex(dop), dimension(PsiDim,0:IntOrdermpi), intent(inout) :: HKrylov
      Complex(dop), dimension(PsiDim), intent(out)               :: HDiagonal
      Complex(dop), dimension(((IntOrder+1)*(IntOrder+2))/2), intent(out):: Tmat
      Complex(dop), dimension(0:IntOrder,0:IntOrder), intent(out):: EigenVector
      Complex(dop), dimension(psidim)                            :: workc
      Real(dop), intent(in)                                      :: TolError,Time, &
                                                                    rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps,EigenVal
      Real(dop), dimension(PsiDim), intent(out)                  :: RDiagonal
      Real(dop), dimension(precon,precon+1), intent(out)         :: davmat
      Real(dop), dimension(workrdim), intent(out)                :: workr
      Logical(kind=4), intent(in)                                :: olsen
      Logical(kind=4), intent(in)                                :: lf
 
      Logical(kind=4), save :: FirstCall = .true.
      Logical(kind=4), save :: full
      logical               :: lreturn = .false.

      Integer(long), parameter :: mxbd = 320
      Integer(long)            :: mxcv
      Integer(long)            :: D,P,Q,QQ,i,j,ij,blk,BOrder,lwork,qmin,MPIBOrder
      Integer(long)            :: MPIOrder,recvcount,sendcount,blk2
      Integer(long)            :: nmpirecv,l,nmaxit,icompz,ldz,N
      Real(dop)                :: Error,ovl,d2,E,beta,beta1,olderr,epse
      Real(dop), dimension(0:2*mxbd+7) :: olden
      real(dop)                        :: edav,cor2,ErrorM
      real(dop), dimension(0:mxbd-1), save   :: oldenergy
      complex(dop)                     :: cor1
      integer(long), dimension(mpi_status_size) :: status
      complex(dop), dimension(:), allocatable   :: zwork
      real(dop), dimension(:), allocatable      :: dwork


      MPIOrder =0
      MPIBOrder=0

      if (mpirank.eq.0) then
         mxcv = 8*BDim+9
         epse=0.d0

! --- CHECK INTEGRATION ORDER ---

         If (IntOrder .LT. 2*BDim) Then
            ErrorCode = 1
            write(6,'(a,i4)') 'IntOrder must be .GE. 2*packets'
            write(6,'(2(a,i4))') 'IntOrder =',IntOrder,' <',2*BDim
            lreturn = .true.
         EndIf
      endif

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

! --- INITIALIZE VARIABLES ---

      if (mpirank.eq.0) then
         if(FirstCall) Then
            if(rlaxnum .ge. 9999) then
               full = .true.
               rlaxnum = rlaxnum - 10000
            else
               full = .false. 
            endif 
            if(rlaxnum.ne.0 .and. rlaxnum.ne.920) then
               write(6,'(a)')' BDAV: Only rlaxnum=0 or follow allowed'
               call stopnow
            endif
            if(Bdim .gt. mxbd) then
               write(6,'(/a,i5,a,i5)')' ERROR in mpibdav.F : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd 
               write(6,'(a,/a)') ' Increase the parameter mxbd in both', &
                 ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
               write(2,'(/a,i5,a,i5)') ' mpibdav.F : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd 
               write(2,'(a,/a)') ' Increase the parameter mxbd in both', &
                 ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
               call stopnow
            endif
         endif
         Do i = 0, mxcv
            qmap(i)=i
            converged(i)=0
         enddo 
         d2 = 1.d-18
         ErrorCode = 0
         ErrorM    = 1.d0
         Error     = 1.d0
         olderr    = 1.d0
         beta1     = 0.d0
         TrueOrder = Bdim - 1
         BOrder    = BDim
         lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
         call zerovxd(ovlps,IntOrder)

!.....Compute the diagonal of the Hamilton matrix.
!        call bgetdiagonal(HDiagonal)
      endif
      call bgetdiagonal(HDiagonal)

      call MPI_Bcast(Psi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(DtPsi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)

!.....Copy complex to real arrays.

      if (mpirank.le.(bdim-1)/intordermpi) then
         if(lmult) then
            blk2=mpirank*intordermpi
            call mpipsi2kryc(Psi,Krylov,blk2,mpiborder, &
                            min(intordermpi,bdim-mpirank*intordermpi)-1)
            call mpipsi2kryc(DtPsi,HKrylov,blk2,mpiborder, &
                            min(intordermpi,bdim-mpirank*intordermpi)-1)
         else
            Do Q = 0, min(intordermpi,bdim-mpirank*intordermpi)-1
               Do D = 1, PsiDim
                   Krylov(D,Q) =  Psi(D+PsiDim*(Q+intordermpi*mpirank))
                  HKrylov(D,Q) =DtPsi(D+PsiDim*(Q+intordermpi*mpirank))
               enddo
            enddo
         endif
         mpiorder=min(intordermpi,bdim-mpirank*intordermpi)
      endif

      if(mpirank.eq.0) then
         Do D = 1, PsiDim
            Rdiagonal(D) = dble(HDiagonal(D))
         enddo
      endif

! --- BUILD T-MATRIX  ---
      call zerovxz(Tmat,((IntOrder+1)*(IntOrder+2))/2)

      if (mpiborder.lt.mpiorder) then
         do recvcount=1,mpirank*intordermpi
           if (mpirank.gt.0) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                  mpirank-1,1000+recvcount,MPI_COMM_WORLD,status,mpierr)
           if (mpirank.lt.(bdim-1)/intordermpi) &
           call MPI_Send(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                         mpirank+1,1000+recvcount,MPI_COMM_WORLD,mpierr)
           do j=0,mpiorder-1
              ij=(mpirank*intordermpi+j)*(mpirank*intordermpi+j+1)/2
              ij=ij+recvcount
              Tmat(ij)=0.d0
              Do D = 1, PsiDim
                 Tmat(ij) = Tmat(ij) + DConjg(Krylov(D,mpiorder))*HKrylov(D,j)
              enddo
           enddo
         enddo

         do sendcount=0,mpiorder-1
           if (mpirank.lt.(bdim-1)/intordermpi) &
           call MPI_Send(Krylov(1,sendcount),PsiDim,MPI_DOUBLE_COMPLEX, & 
                mpirank+1,1001+sendcount+mpirank*intordermpi,MPI_COMM_WORLD,mpierr)
           do j=0,sendcount
              ij=(mpirank*intordermpi+sendcount)*(mpirank*intordermpi+sendcount+1)/2
              ij=ij+mpirank*intordermpi+j+1
              Tmat(ij)=0.d0
              Do D = 1, PsiDim
                 Tmat(ij) = Tmat(ij) + DConjg(Krylov(D,j))*HKrylov(D,sendcount)
              enddo
           enddo
         enddo
      endif

! --- join T-mat
      call MPI_Reduce(Tmat,EigenVector,((IntOrder+1)*(IntOrder+2))/2, &
                      MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxz(Eigenvector,Tmat,((IntOrder+1)*(IntOrder+2))/2)

      mpiborder=mpiorder

! --- create EigenVector-Matrix
      if (mpirank.eq.0) then
         ij = 0
         do j = 0, BOrder - 1
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         enddo
      endif

! --- DIAGONALISE T-MATRIX IN THE INITIAL BLOCK OF STARTVECTORS ---
      if (mpirank.eq.0) then
         allocate(zwork(lwork))
         allocate(dwork(3*border))
         call zheev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                                            zwork,lwork,dwork,ErrorCode)
         deallocate(zwork,dwork)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf
      endif

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

      if (mpirank.eq.0) then
         if(FirstCall) then
            i=1
            beta=1.d0
            beta1=0.d0
            Do Q = 0, BDim-1
               E=EigenVal(Q)
               call wrtrlxbdav(time,Q,Q,Q,E,beta,E,i,EigenVal, &
                               ovlps,FirstCall,IntOrder,BDim)
               oldenergy(Q)=E
               olden(Q)=E
               beta1=beta1+E
            enddo
            E=beta1/dble(BDim)  ! Averaged energy for pre-conditioner
            if(rlxemin.gt.-1.d99) E=rlxemin 
            if(rlxemax.lt. 1.d99) E=rlxemax 
         endif   ! firstcall

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
         if( precon .le. 1 ) precon=0
         if( precon .gt. 0 ) then
            write(6,*) 'Sorry, preconditioner not yet implemented', &
                       ' for BDAV (hermitian)'
            call stopnow
!!!            call sortdiagonal(RDiagonal,workr,indxdav,E,PsiDim,precon)
!!!            call getdavmat(RDiagonal,davmat,davmat(1,precon+1), &
!!!                           indxdav,PsiDim,precon,lwork)
         endif
      else
         if( precon .gt. 0 ) then
            write(6,*) 'Sorry, preconditioner not yet implemented', &
                       ' for BDAV (hermitian)'
            call stopnow
!!!            if(lmpigetdavmat) &
!!!               call mpigetdavmat(RDiagonal,davmat,indxdav,PsiDim,precon, &
!!!                                                            workr)
!!!            if (precon.gt.max(2*nmpi*nthread,qrlimit).and.lmpidsyev) &
!!!               call STRTDSTEQR(ldz,l,lendmpi,icompz,n,nmaxit,rdata)
         endif
      endif

! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

  99  continue  ! Block Iteration Loop
      if (mpirank.eq.0) then
         olderr = ErrorM
         ErrorM = 0.d0
         QQ = 0
      endif

 100  Continue  !  Iteration Loop

! --- RE-INITIALIZE VARIABLES ---

      if (mpirank.eq.0) then
         TrueOrder = TrueOrder+1
 42      if(converged(QQ).gt.0) then
            if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt.epse) then
               if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. 10.d0* &
                  abs(EigenVal(qmap(QQ))-olden(QQ+1))) then ! state missing
                  do i = QQ,mxcv-1
                     converged(i) = converged(i+1)
                     olden(i)     = olden(i+1)
                  enddo
               elseif(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. 10.d0* &
                      abs(EigenVal(qmap(QQ))-olden(max(0,QQ-1)))) then !st. added
                  do i = mxcv, QQ+1, -1
                     converged(i) = converged(i-1)
                     olden(i)     = olden(i-1)
                  enddo
                  converged(QQ)=0
               else
                  converged(QQ)=0
               endif
               goto 42  ! Try again, after converged is modified
            endif
            QQ=QQ+1     ! Skip the converged state
            goto 42
         endif
         if(qq.ge.mxcv) then
            write(6,'(a,2i6)')  'BRDAV: ## qq, q', qq, q
            write(6,'(a,i6,a)') 'BRDAV: Enlarge mxcv=',mxcv,' (also in rundav)'
            WRITE(6,'(a,/(20i4))') 'Qmap:     ',(qmap(i),i=0,mxcv)
            WRITE(6,'(a,/(20i4))') 'Converged:',(converged(i),i=0,mxcv)
            call stopnow
         endif
         Q = qmap(QQ)
         E = EigenVal(Q)
         olden(QQ) = E
      endif

      call MPI_Bcast(trueorder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

! --- Build residual vector |r> ---

      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(E,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      call zerovxz(Krylov(1,mpiOrder),psidim)
      if (mpiorder.gt.0) then
         Do j = 0, mpiBOrder-1
            Do D = 1, PsiDim
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder) + &
                                  EigenVector(j+mpirank*intordermpi,Q) * &
                                          (HKrylov(D,j) - E*Krylov(D,j))
            enddo
         enddo
      endif

      call MPI_Reduce(Krylov(1,mpiorder),HKrylov(1,mpiorder),PsiDim, &
                   MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxz(HKrylov(1,mpiorder),Krylov(1,MPIOrder),PsiDim)


! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      if (mpirank.eq.0) then
         if( precon.eq.0 ) then
            Error = 0.d0
            Do D = 1, PsiDim
               Error = Error + abs(Krylov(D,mpiOrder))**2
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder)* &
                    (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
         else
            call cprecondit(Krylov(1,mpiOrder),RDiagonal,davmat, &
                           davmat(1,precon+1),indxdav,workr, &
                           workr(precon+1),E,Error,d2,PsiDim,precon)
         endif
         Error = Sqrt(Error)
         if(QQ.lt.Bdim) ErrorM = max(Error,ErrorM)
         if(Error .lt. 0.5*TolError .and. converged(QQ).eq.0) then
            converged(QQ) = -TrueOrder
         endif
         if(Error .lt. 0.05d0*TolError  .and. converged(QQ).le.0) then
            converged(QQ) = abs(converged(QQ))
         elseif(Error .gt. TolError) then
            converged(QQ) =  0
         endif
      endif

! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- |r> = (H-E) * eigenvector
! --- Krylov(TrueOrder)  = (H_0-E)^-1 * |r>
! --- workc(2*precon+D)  =  eigenvector(D)
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector (used for precon)
! --- Correction : (cor1/cor2)*(H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction


      if(olsen) then
! --- Build EigenVector.
         call zerovxz(workc(2*precon+1),psidim)
         call MPI_Bcast(EigenVector,(IntOrder+1)**2, &
                             MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(E,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

         if (mpiorder.gt.0) then
            Do j = 0, mpiOrder-1
               Do D = 1, PsiDim
                  workc(2*precon+D) = workc(2*precon+D) + &
                      EigenVector(j+mpirank*intordermpi,Q) * Krylov(D,j)
               enddo
            enddo
         endif
         call MPI_Reduce(workc(2*precon+1),HKrylov(1,MPIOrder),PsiDim, &
                   MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         call cpvxz(HKrylov(1,MPIOrder),workc(2*precon+1),PsiDim)
         if (mpirank.eq.0) then
            cor1=0.d0
            cor2=0.d0
!            if(precon.eq.0) then
               do D = 1, PsiDim
                  cor1 = cor1 + DConjg(workc(2*precon+D))*Krylov(D,mpiOrder)
                  cor2 = cor2 + abs(workc(2*precon+D))**2 * &
                         (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
               do D = 1, PsiDim
                  Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - &
                       (cor1/cor2) * workc(2*precon+D) * &
                       (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
!            else
!               call cpvxz(workc(2*precon+1),HKrylov(1,mpiOrder),PsiDim)
!               call precondit(HKrylov(1,mpiOrder),RDiagonal,davmat, &
!                              davmat(1,precon+1),indxdav,workc, &
!                              workc(precon+1),E,ovl,d2,PsiDim,precon)
!               do D = 1, PsiDim
!                  cor1 = cor1 + DConjg(workc(2*precon+D))* &
!                                       Krylov(D,mpiOrder)
!                  cor2 = cor2 + DConjg(workc(2*precon+D))* &
!                                       HKrylov(D,mpiOrder)
!               enddo
!               do D = 1, PsiDim
!                  Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - &
!                       (cor1/cor2) * HKrylov(D,mpiOrder)
!               enddo
!            endif  ! precon
         endif
      endif     ! olsen


! ---  ORTHONORMALIZE  ---

      if(mpiorder.gt.0) then
         if (mpirank.gt.0) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonc(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_COMPLEX,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                    MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
            call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif
      if (mpirank.eq.0) then
         call mpiovlc(Krylov(1,MPIOrder),ovl,PsiDim)
         if(ovl.le.1.d-31) then     ! If vectors are linearly dependent
            write(6,'(/a)') '---------------------------------------------'
            write(6,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                             Steps,TrueOrder,Q,ovl,Error
            write(2,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                             Steps,TrueOrder,Q,ovl,Error
            write(6,'(a)') 'BDAV: Vectors are linearly dependent!' 
            write(2,'(a)') 'BDAV: Vectors are linearly dependent!' 
            WRITE(6,'(a,20i3)') 'Qmap:     ', (qmap(i),i=0,2*BDim-1)
            WRITE(6,'(a,20i3)') 'Converged:',(converged(i),i=0,2*BDim-1)
            call stopnow
         endif
!.....Symmetrize coefficients, if desired.
         if(nmod.gt.0) call csymcoeff(Krylov(1,mpiOrder), &
                   HKrylov(1,mpiOrder),PsiDim,nmod,lf,ovl,TrueOrder)
      endif

!.....Do the orthonormalisation twice
      if(mpiorder.gt.0) then
         if (mpirank.gt.0) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonc(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_COMPLEX,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                    MPI_DOUBLE_COMPLEX,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif

      if (mpirank.eq.0) call mpiovlc(Krylov(1,MPIOrder),ovl,PsiDim)

! --- Send new vector to appropriate processor

      if (trueorder.ge.intordermpi) then
         nmpirecv=trueorder/intordermpi
         if (mpirank.eq.0) then
           call MPI_Send(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                                    nmpirecv,1000,MPI_COMM_WORLD,mpierr)
         endif
         if (mpirank.eq.nmpirecv) then
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_COMPLEX, &
                                    0,1000,MPI_COMM_WORLD,status,mpierr)
           mpiorder=mpiorder+1
         endif
      else
         if (mpirank.eq.0) mpiorder=mpiorder+1
      endif

! --- WRITE STEP AND ERROR IF DESIRED ---
      if (mpirank.eq.0) then
         E = EigenVal(Q)
         beta = Abs(EigenVector(QQ,Q))**2
         Call WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)
      endif

      call MPI_Bcast(BOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(trueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      if(TrueOrder .ge. BOrder+BDim-1) goto 41
      QQ = QQ + 1
      goto 100
 41   QQ = 0

! --- BUILD HKRYLOV  ---

      call zerovxz(Psi,adim)
      if(lmult) then
         blk2=mpirank*intordermpi-border
         call mpikryc2psi(Psi,Krylov,blk2,mpiborder,mpiorder-1)
      else
         Do blk = mpiborder,mpiorder-1
            Do D = 1,PsiDim
               blk2=mpirank*intordermpi-border+blk
               Psi(D+blk2*PsiDim) = Krylov(D,blk)
            enddo
         enddo
      endif
      call MPI_Reduce(Psi,DtPsi,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                               MPI_COMM_WORLD,mpierr)
      call cpvxz(DtPsi,Psi,adim)
      Call Func(Time,Psi,DtPsi)
      call MPI_Bcast(DtPsi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
      if(lmult) then
         blk2=mpirank*intordermpi-border
         call mpipsi2kryc(DtPsi,HKrylov,blk2,mpiborder,mpiorder-1)
      else
         Do blk = mpiborder,mpiorder-1
            Do D = 1,PsiDim
               blk2=mpirank*intordermpi-border+blk
               HKrylov(D,blk) = DtPsi(D+blk2*PsiDim)
            enddo
         enddo
      endif

! --- BUILD NEW T-MATRIX ELEMENTS ---

      if (mpirank.gt.0) then
         call zerovxz(Tmat,((IntOrder+1)*(IntOrder+2))/2)
      else
         call zerovxz(Tmat(ij+1),((IntOrder+1)*(IntOrder+2))/2-ij)
      endif

      if (mpiBorder.lt.mpiorder) then
         do recvcount=1,mpirank*intordermpi
            if (mpirank.gt.0) then
               if (mpirank.eq.border/intordermpi) then
                  call MPI_Recv(Krylov(1,MPIOrder),PsiDim, &
                         MPI_DOUBLE_COMPLEX,(recvcount-1)/intordermpi, &
                            2000+recvcount,MPI_COMM_WORLD,status,mpierr)
               else
                  call MPI_Recv(Krylov(1,MPIOrder),PsiDim, &
                         MPI_DOUBLE_COMPLEX,mpirank-1, &
                            2000+recvcount,MPI_COMM_WORLD,status,mpierr)
               endif
            endif
            if (mpirank.lt.trueorder/intordermpi) then
               call MPI_Send(Krylov(1,mpiorder),PsiDim, &
                         MPI_DOUBLE_COMPLEX,mpirank+1,2000+recvcount, &
                                                  MPI_COMM_WORLD,mpierr)
            endif
            do j=mpiborder,mpiorder-1
               ij=(mpirank*intordermpi+j)*(mpirank*intordermpi+j+1)/2
               ij=ij+recvcount
               Tmat(ij) = 0.d0
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij) + DConjg(Krylov(D,mpiorder))*HKrylov(D,j)
               enddo
            enddo
         enddo
         do j=0,mpiorder-1
            if (mpirank.lt.trueorder/intordermpi) then
               call MPI_Send(Krylov(1,j),PsiDim,MPI_DOUBLE_COMPLEX, &
                       mpirank+1,2000+recvcount+j,MPI_COMM_WORLD,mpierr)
            endif
         enddo
         do i=mpiborder,mpiorder-1
            do j=0,i
               ij=(mpirank*intordermpi+i)*(mpirank*intordermpi+i+1)/2
               ij=ij+recvcount+j
               Tmat(ij) = 0.d0
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij)+ DConjg(Krylov(D,j))*HKrylov(D,i)
               enddo
            enddo
         enddo
      else
        do sendcount=0,mpiorder-1
          call MPI_Send(Krylov(1,sendcount),PsiDim,MPI_DOUBLE_COMPLEX, &
                  border/intordermpi,2001+intordermpi*mpirank+sendcount, &
                                                  MPI_COMM_WORLD,mpierr)
        enddo
      endif

      call MPI_Reduce(Tmat,EigenVector,((IntOrder+1)*(IntOrder+2))/2, &
                   MPI_DOUBLE_COMPLEX,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxz(Eigenvector,Tmat,((IntOrder+1)*(IntOrder+2))/2)

      mpiBOrder=mpiOrder

! --- DIAGONALISE T-MATRIX ---
      if (mpirank.eq.0) then
         ij = 0
         do j = 0, TrueOrder
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         enddo
      endif

      if (mpirank.eq.0) then
         BOrder=TrueOrder+1
         call zheev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                        workc,lwork,workr,ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            lreturn = .true.
         EndIf
      endif

      call MPI_Bcast(lreturn,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      if (lreturn) return

! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS sought) -
      if (mpirank.eq.0) then
         if(rlaxnum.eq.920) then   ! follow
            qmin=2*BDim
            Do Q = 0, BDim-1
               beta=0.d0
               j = -1
               Do P = 0, TrueOrder
                  if(EigenVal(P).ge.rlxemin .and. EigenVal(P).le.rlxemax) then 
                     beta1= Abs(EigenVector(Q,P))**2
                     do i = 0, Q-1
                        if(qmap(i).eq.P) beta1=-1.d0
                     enddo
                     if(beta1 .gt. beta) then
                        j = P
                        beta = beta1
                     endif
                  endif
               enddo
               if(j.lt.0) then  ! Energy window too small.
                  beta=1.d+99   ! Take vectors with energies closest to window.
                  Do P = 0, TrueOrder
                     beta1 = min(abs(rlxemin-EigenVal(P)),abs(EigenVal(P)-rlxemax))
                     if(beta1.lt.beta) then
                        beta=beta1
                        j = P
                     endif 
                  enddo
               endif
               qmap(Q) = j
               qmin = min(qmin,j)
            enddo  ! Q

            if(full) qmin=0
            qmin=max(qmin-1,0)               ! Set higher entries of qmap.
            Do Q = BDim,min(mxcv,TrueOrder)  ! They may be used when some
   51          do i = 0, Q-1                 ! vectors converge. 
                  if(qmap(i).eq.qmin) then
                     qmin=qmin+1
                     if(qmin.gt.TrueOrder) qmin=0
                     goto 51
                  endif
               enddo
               qmap(Q)=qmin
            enddo   !  Q
         elseif(rlaxnum.eq.0 .and. rlxemin.gt.-1.d99) then
            QQ=0
            do Q = 0, TrueOrder
               if(EigenVal(Q).gt.rlxemin .and. QQ.lt.mxcv) then
                  qmap(QQ)=Q
                  QQ=QQ+1
                  i=Q
                  if(QQ.eq.Bdim) goto 52
               endif
            enddo
   52       Q=i+1
            if(QQ.lt.Bdim) then
               qmap(QQ)=Q
               QQ=QQ+1
               Q=Q+1
               goto 52
            endif
            if(qmap(0).gt.0) then  ! Take the first state below the block
               qmap(QQ)=qmap(0)-1  ! as additional vector (when others
               QQ=QQ+1             ! have converged).
            endif
   53       if(QQ.lt.mxcv) then    ! Take the states following the block
               qmap(QQ)=Q          ! as additional vectors (when others
               QQ=QQ+1             ! have converged).
               Q=Q+1
               goto 53
            endif
         endif      ! (if rlaxnum.eq.920 ...  elseif rlaxnum.eq.0)
         epse=(EigenVal(qmap(Bdim-1))-EigenVal(qmap(0)))/(Bdim*90.d0)
      endif

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP  ---
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(ErrorM,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(olderr,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      If( TrueOrder .GT. IntOrder-BDim ) Goto 300
      If( TrueOrder .GE. PsiDim-BDim   ) Goto 300
      If( full .or. (ErrorM+olderr.GT.TolError) ) Goto 99

! --- CONVERGENCE! Build new A-vector ---
 300  continue
      call MPI_Bcast(qmap,8*bdim+10,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_COMPLEX,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call zerovxz(Psi,adim)
      call zerovxz(DtPsi,adim)
      if(lmult) then
         call mpikryc2finalpsi(Psi,DtPsi,Krylov,HKrylov, &
                           EigenVector,TrueOrder,mpiorder,qmap)
      else
         Do QQ = 0,BDim-1
            Q = qmap(QQ)
            Do P = 0,mpiorder-1
               Do D = 1,PsiDim
                  Psi(D+QQ*PsiDim)   = Psi(D+QQ*PsiDim) + &
                        Krylov(D,P)*EigenVector(mpirank*intordermpi+P,Q)
                  DtPsi(D+QQ*PsiDim) = DtPsi(D+QQ*PsiDim) + &
                       HKrylov(D,P)*EigenVector(mpirank*intordermpi+P,Q)
               EndDo
            EndDo
         EndDo
      endif
      call MPI_Reduce(Psi,workc,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,Psi,adim)
      call MPI_Reduce(DtPsi,workc,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,DtPsi,adim)

      if (mpirank.eq.0) then
         Do QQ = 0, BDim-1
            Q = qmap(QQ) 
            if(FirstCall) then
!..............This is for first overlaps in wrtrlxbdav
               do j = 0, TrueOrder
                  ovlps(j) = Abs(EigenVector(Q,j))**2
               enddo
            endif
            E=EigenVal(Q)
            beta=Abs(EigenVector(QQ,Q))**2
            j=QQ
            if(beta.lt.0.5d0 .and. TrueOrder.gt.Bdim) then
               beta=0.d0
               do i=0,Bdim+1
                  if(Abs(EigenVector(i,Q))**2 .gt. beta) then
                     beta=Abs(EigenVector(i,Q))**2
                     j=i
                  endif
               enddo
               if(beta.lt.0.67d0) then
                  beta=Abs(EigenVector(QQ,Q))**2
                  j=QQ
               endif
            endif
            call wrtrlxbdav(time,trueorder,QQ,Q,E,beta,oldenergy(j), &
                 converged(QQ),EigenVal,ovlps,FirstCall,IntOrder,BDim)
         enddo
         Do QQ = 0, BDim+1
            Q = qmap(QQ)
            oldenergy(QQ)=EigenVal(Q)
         enddo

         FirstCall = .false.
         full      = .false.
         edav = E
      endif
!#endif

      Return
      End Subroutine MPIBDavStep

! **********************************************************************
! **********************************************************************
! *                                                                    *
! *                 SUBROUTINE MPIBRDAVSTEP                            *
! *                                                                    *
! * Block-Davidson diagonalisation. (single-set multi-packet run).     *
! * Diagonalises the Hamiltonian represented in the SPFs.              *
! * Assuming real Hamiltonian and real WF.                             *
! *                                                                    *
! *                                                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   Psi:       The block of (complex) initial-value vectors.         *
! *   DtPsi:     Action of Hamiltonian on the initial-value vector,    *
! *              i.e. H|Psi> .                                         *
! *   PsiDim     Length of Psi and DtPsi vectors (i.e. adim/npacket)   *
! *   BDim       Block size. BDim=npacket (=gdim(feb)                  *
! *   IntOrder:  Maximum interation order.                             *
! *   TolError:  Maximum error that is tolerated.                      *
! *   Steps:     Number of steps made so far (Steps is passed to       *
! *              "WriteStepBDAv").                                     *
! *   Krylov:    Matrix with minimum size PsiDim*IntOrder with         *
! *              columns containing the Davison vectors. (real!)       *
! *   HKrylov:   Matrix containing H*Krylov (real!)                    *
! *   precon:    Dimension of Pre-Conditioner                          *
! *   davmat:    Contains the eigenvectors and eigenvalues of the      *
! *              preconditioner.                                       *
! *   nmod       Number of modes. If nmod=0 => no symmetrisation       *
! *   lf         lf=.true. for fermions, lf=.false. for bosons         *
! *   lrreal     lrreal=.true. => Funcr is used (rrDAV).               *
! *                                                                    *
! * Output parameters:                                                 *
! *   Psi:       Propagated Psi.                                       *
! *   Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with     *
! *              columns containing the Krylov vectors H^2|psi> to     *
! *              H^TrueOrder|psi>.                                     *
! *   TrueOrder: The order that has actually been used (may be less    *
! *              than IntOrder).                                       *
! *   Steps:     Same as on entry.                                     *
! *   ErrorCode: Error code having the following meaning:              *
! *              0: everything was o. k.,                              *
! *              1: illegal integration order,                         *
! *              3: diagonalization failed,                            *
! *                                                                    *
! * External routines:                                                 *
! *   Func:      Computes the action of the Hamiltonian on Psi.        *
! *              Called as                                             *
! *              Func(Abs(Time),Psi,DtPsi).                            *
! *   Funcr:     Special routine for real arithmetic                   *
! * WriteStepBDav: Writes the energy,beta, and error to a file.        *
! *   Called as WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)  *
! *                                                                    *
! * Some new Variables:                                                *
! *  rlaxnum : This is the argument of the keyword relaxation.         *
! *            If relaxation has no argument, rlaxnum=-1 .             *
! *            rlaxnum=0   : relaxation to ground-state.               *
! *            rlaxnum=n   : relaxation to n-th state (n=0...900).     *
! *            rlaxnum=920 : relaxation=follow.                        *
! *            rlaxnum=930 : relaxation=lock.                          *
! *                                                                    *
! * V8.3 09/2007 HDM                                                   *
! **********************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes here, please do them in brdavstep too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Subroutine MPIBRDavStep(Psi,DtPsi,PsiDim,IntOrder, &
                   TolError,Steps,rlaxnum,Krylov,HKrylov,HDiagonal,Tmat, &
                   EigenVector,EigenVal,ovlps,workr,Rdiagonal,davmat, &
                   TrueOrder,ErrorCode,Time,precon,olsen, &
                   Func,Funcr,lrreal,rlxemin,rlxemax,Bdim,nmod, &
                   lf,indxdav,converged,qmap)

!#ifdef MPI
use mpi
use global
use globmemmod
use maxv, only: maxdim, maxsta
use psidef
use griddatmod
use mpidata

use op1lib, only: zerovxd,cpvxd,zerovxz,cpvxz
use getdiagonalmod, only: bgetdiagonal
use davutils, only: precondit,rsymcoeff
use davutilspar, only: mpiorthonr,mpiovl,mpipsi2kry,mpikry2psi,mpikry2finalpsi
use getdavmatmod, only: mpigetdavmat,getdavmat,sortdiagonal
use mpidsteqmod, only: strtdsteqr
use pardsyevmod, only: pardsyev
use wrintegrat, only: writestepbdav,wrtrlxbdav

      Implicit None

      External                                                   :: Func,Funcr
      Integer(long), intent(in)                                  :: PsiDim, &
                                                                    IntOrder, &
                                                                    nmod
      Integer(long), intent(in)                                  :: Steps
      Integer(long), intent(in)                                  :: BDim
      Integer(long), intent(inout)                               :: precon
      Integer(long), dimension(precon), intent(out)              :: indxdav
      Integer(long), dimension(0:8*BDim+9), intent(out)          :: converged,&
                                                                    qmap
      Integer(long), intent(inout)                               :: rlaxnum
      Integer(long), intent(out)                                 :: ErrorCode, &
                                                                    TrueOrder
      Complex(dop), dimension(adim), intent(inout)               :: Psi,DtPsi
      Complex(dop), dimension(PsiDim), intent(out)               :: HDiagonal
      Complex(dop), dimension(adim)                              :: workc

      Real(dop), dimension(PsiDim,0:IntOrdermpi), intent(out)    :: Krylov
      Real(dop), dimension(PsiDim,0:IntOrdermpi), intent(inout)  :: HKrylov
      Real(dop), dimension(((IntOrder+1)*(IntOrder+2))/2), intent(out):: Tmat
      Real(dop), dimension(0:IntOrder,0:IntOrder), intent(out)   :: EigenVector
      Real(dop), intent(in)                                      :: TolError,Time, &
                                                                    rlxemin,rlxemax
      Real(dop), dimension(0:IntOrder), intent(inout)            :: ovlps,EigenVal
      Real(dop), dimension(PsiDim), intent(out)                  :: RDiagonal
      Real(dop), dimension(precon,precon+1), intent(out)         :: davmat
      Real(dop), dimension(workrdim), intent(out)                :: workr
      Logical(kind=4), intent(in)                                :: olsen
      Logical(kind=4), intent(in)                                :: lrreal
      Logical(kind=4), intent(in)                                :: lf

      Logical(kind=4), save :: FirstCall = .true.
      Logical(kind=4), save :: full

      Integer(long), parameter :: mxbd = 320
      Integer(long)            :: mxcv,kskip,idown
      Integer(long)            :: D,P,Q,QQ,i,j,ij,blk,BOrder,lwork,qmin,MPIBOrder
      Integer(long)            :: MPIOrder,recvcount,sendcount,blk2
      Integer(long)            :: nmpirecv,l,nmaxit,icompz,ldz,N
      Real(dop)                :: Error,ovl,d2,E,beta,beta1,olderr,epse,edav
      Real(dop), dimension(0:2*mxbd+7) :: olden
      real(dop)                        :: cor1,cor2,ErrorM
      real(dop), dimension(0:mxbd-1,0:2), save   :: oldenergy
      integer(long), dimension(mpi_status_size) :: status
      integer(long), save                       :: tid=0 ! timer ID

! GWR DEBUG
! Not sure what this should be. Defined but not set in the serial part.
      kskip = 0
! GWR DEBUG
      MPIOrder =0
      MPIBOrder=0

      if (mpirank.eq.0) then
         mxcv = 8*BDim+9
         epse=0.d0

! --- CHECK INTEGRATION ORDER ---

         If (IntOrder .LT. 2*BDim) Then
            ErrorCode = 1
            write(6,'(a,i4)') 'IntOrder must be .GE. 2*packets'
            write(6,'(2(a,i4))') 'IntOrder =',IntOrder,' <',2*BDim
            Return
         EndIf

! --- INITIALIZE VARIABLES ---

         if(FirstCall) Then
            if(rlaxnum .ge. 9999) then
               full = .true.
               rlaxnum = rlaxnum - 10000
            else
               full = .false. 
            endif 
            if(rlaxnum.ne.0 .and. rlaxnum.ne.920) then
               write(6,'(a)')' BDAV: Only rlaxnum=0 or follow allowed'
               call stopnow
            endif
            if(Bdim .gt. mxbd) then
               write(6,'(/a,i5,a,i5)')' ERROR in mpibdav.F : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd 
               write(6,'(a,/a)') ' Increase the parameter mxbd in both', &
                 ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
               write(2,'(/a,i5,a,i5)') ' mpibdav.F : packets =', &
                                Bdim,' .gt.  mxbd =', mxbd 
               write(2,'(a,/a)') ' Increase the parameter mxbd in both', &
                 ' source/lib/ode/bdavlib.f and source/propwf/mpibdav.F'
               call stopnow
            endif
         endif
         Do i = 0, mxcv
            qmap(i)=i
            converged(i)=0
         enddo 
         d2    = 1.d-18
         ErrorCode = 0
         ErrorM = 0
         olderr = 1.d0
         TrueOrder = Bdim - 1
         BOrder    = BDim
         beta1 = 0
         Error = 1.d0
         lwork = max(24*max(IntOrder,precon)+48,2*precon+PsiDim)
         call zerovxd(ovlps,IntOrder)

!.....Compute the diagonal of the Hamilton matrix.
!        call bgetdiagonal(HDiagonal)
      endif
      call bgetdiagonal(HDiagonal)


      call MPI_Bcast(Psi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(DtPsi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)

!.....Copy complex to real arrays.
      if (mpirank.le.(bdim-1)/intordermpi) then
         if(lmult) then
            blk2=mpirank*intordermpi
            call mpipsi2kry(Psi,Krylov,blk2,mpiborder, &
                            min(intordermpi,bdim-mpirank*intordermpi)-1)
            call mpipsi2kry(DtPsi,HKrylov,blk2,mpiborder, &
                            min(intordermpi,bdim-mpirank*intordermpi)-1)
         else
            Do Q = 0, min(intordermpi,bdim-mpirank*intordermpi)-1
               Do D = 1, PsiDim
                   Krylov(D,Q) = dble(  Psi(D+PsiDim*(Q+intordermpi*mpirank)))
                  HKrylov(D,Q) = dble(DtPsi(D+PsiDim*(Q+intordermpi*mpirank)))
               enddo
            enddo
         endif
         mpiorder=min(intordermpi,bdim-mpirank*intordermpi)
      endif

      if(mpirank.eq.0) then
         Do D = 1, PsiDim
            Rdiagonal(D) = dble(HDiagonal(D))
         enddo
      endif

! --- BUILD T-MATRIX  ---
      call zerovxd(Tmat,((IntOrder+1)*(IntOrder+2))/2)

      if (mpiborder.lt.mpiorder) then
         do recvcount=1,mpirank*intordermpi
           if (mpirank.gt.0) &
              call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                   mpirank-1,1000+recvcount,MPI_COMM_WORLD,status,mpierr)
           if (mpirank.lt.(bdim-1)/intordermpi) &
              call MPI_Send(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                            mpirank+1,1000+recvcount,MPI_COMM_WORLD,mpierr)
           do j=0,mpiorder-1
              ij=(mpirank*intordermpi+j)*(mpirank*intordermpi+j+1)/2
              ij=ij+recvcount
              Tmat(ij)=0.d0
              Do D = 1, PsiDim
                 Tmat(ij) = Tmat(ij) + Krylov(D,mpiorder)*HKrylov(D,j)
              enddo
           enddo
         enddo

         do sendcount=0,mpiorder-1
           if (mpirank.lt.(bdim-1)/intordermpi) &
             call MPI_Send(Krylov(1,sendcount),PsiDim,MPI_DOUBLE_PRECISION, &
                  mpirank+1,1001+sendcount+mpirank*intordermpi,MPI_COMM_WORLD, &
                                                                 mpierr)
           do j=0,sendcount
              ij=(mpirank*intordermpi+sendcount)*(mpirank*intordermpi+sendcount+1)/2
              ij=ij+mpirank*intordermpi+j+1
              Tmat(ij)=0.d0
              Do D = 1, PsiDim
                 Tmat(ij) = Tmat(ij) + Krylov(D,j)*HKrylov(D,sendcount)
              enddo
           enddo
         enddo
      endif

! --- join T-mat
      call MPI_Reduce(Tmat,EigenVector,((IntOrder+1)*(IntOrder+2))/2, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxd(Eigenvector,Tmat,((IntOrder+1)*(IntOrder+2))/2)

      mpiborder=mpiorder

! --- create EigenVector-Matrix
      if (mpirank.eq.0) then
         ij = 0
         do j = 0, BOrder - 1
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         enddo
      endif

! --- DIAGONALISE T-MATRIX IN THE INITIAL BLOCK OF STARTVECTORS ---
      if (mpirank.eq.0) then
         call dsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                                                  workr,lwork,ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            Return
         EndIf
      endif


      if (mpirank.eq.0) then
         if(FirstCall) then
            i=1
            beta=1.d0
            beta1=0.d0
            Do Q = 0, BDim-1
               E=EigenVal(Q)
               call wrtrlxbdav(time,Q,Q,Q,E,beta,E,i,EigenVal, &
                            ovlps,FirstCall,IntOrder,BDim)
               oldenergy(Q,0)=E
               oldenergy(Q,1)=E
               oldenergy(Q,2)=E
               olden(Q)=E
               beta1=beta1+E
            enddo
            E=beta1/dble(BDim)  ! Averaged energy for pre-conditioner
            if(rlxemin.gt.-1.d99) E=rlxemin 
            if(rlxemax.lt. 1.d99) E=rlxemax 
         endif   ! firstcall

! --- Compute precon x precon Hamiltonian-Sub-Matrix for Preconditioner.
         if( precon .le. 1 ) precon=0
         if( precon .gt. 0 ) then
            call get_timer(tid, 'Precon')
            call start_timer(tid)

            call sortdiagonal(RDiagonal,workr,indxdav,E,PsiDim,precon)
            call getdavmat(RDiagonal,davmat,davmat(1,precon+1), &
                           indxdav,PsiDim,precon,lwork)
            call stop_timer(tid)
         endif
      else
         if( precon .gt. 0 ) then
            if(lmpigetdavmat) call mpigetdavmat(RDiagonal,davmat,indxdav, &
                                                PsiDim,precon,workr)
            if (precon.gt.2*nmpi.and.lmpidsyev) &
               call STRTDSTEQR(ldz,l,lendmpi,icompz,n,nmaxit,mr)
         endif
      endif

! --- LOOP : BUILD UP THE KRYLOV AND HKRYLOV SPACES ---

  99  continue  ! Block Iteration Loop
      if (mpirank.eq.0) then
         olderr = ErrorM
         ErrorM = 0.d0
         QQ = 0
      endif

 100  Continue  !  Iteration Loop

! --- RE-INITIALIZE VARIABLES ---

      if (mpirank.eq.0) then
         TrueOrder = TrueOrder+1
 42      if(converged(QQ).gt.0) then
            if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt.epse) then
               if(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. 10.d0* &
                  abs(EigenVal(qmap(QQ))-olden(QQ+1))) then ! state missing
                  do i = QQ,mxcv-1
                     converged(i) = converged(i+1)
                     olden(i)     = olden(i+1)
                  enddo
               elseif(abs(EigenVal(qmap(QQ))-olden(QQ)).gt. 10.d0* &
                      abs(EigenVal(qmap(QQ))-olden(max(0,QQ-1)))) then !st. added
                  do i = mxcv, QQ+1, -1
                     converged(i) = converged(i-1)
                     olden(i)     = olden(i-1)
                  enddo
                  converged(QQ)=0
               else
                  converged(QQ)=0
               endif
               goto 42  ! Try again, after converged is modified
            endif
            QQ=QQ+1     ! Skip the converged state
            goto 42
         endif
         if(qq.ge.mxcv) then
            write(6,'(a,2i6)')  'BRDAV: ## qq, q', qq, q
            write(6,'(a,i6,a)') 'BRDAV: Enlarge mxcv=',mxcv,' (also in rundav)'
            WRITE(6,'(a,/(20i4))') 'Qmap:     ',(qmap(i),i=0,mxcv)
            WRITE(6,'(a,/(20i4))') 'Converged:',(converged(i),i=0,mxcv)
            call stopnow
         endif
         Q = qmap(QQ)
         E = EigenVal(Q)
         olden(QQ) = E
      endif

      call MPI_Bcast(trueorder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

! --- Build residual vector |r> ---

      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(E,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

      call zerovxd(Krylov(1,mpiOrder),psidim)
      if (mpiorder.gt.0) then
         Do j = 0, mpiBOrder-1
            Do D = 1, PsiDim
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder) + &
                                  EigenVector(j+mpirank*intordermpi,Q) * &
                                          (HKrylov(D,j) - E*Krylov(D,j))
            enddo
         enddo
      endif

      call MPI_Reduce(Krylov(1,mpiorder),HKrylov(1,mpiorder),PsiDim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxd(HKrylov(1,mpiorder),Krylov(1,MPIOrder),PsiDim)


! --- DETERMINE ERROR and APPLY INVERSE OF DIAGONAL ---
      if (mpirank.eq.0) then
         if( precon.eq.0 ) then
            Error = 0.d0
            Do D = 1, PsiDim
               Error = Error + Krylov(D,mpiOrder)**2
               Krylov(D,mpiOrder) = Krylov(D,mpiOrder)* &
                  (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
            enddo
         else
            call precondit(Krylov(1,mpiOrder),RDiagonal,davmat, &
                           davmat(1,precon+1),indxdav, &
                           E,Error,d2,PsiDim,precon)
         endif
         Error = Sqrt(Error)
         if(QQ.lt.Bdim) ErrorM = max(Error,ErrorM)
         if(Error .lt. 0.5*TolError .and. converged(QQ).eq.0) then
            converged(QQ) = -TrueOrder
         endif
         if(Error .lt. 0.05d0*TolError  .and. converged(QQ).le.0) then
            converged(QQ) = abs(converged(QQ))
         elseif(Error .gt. TolError) then
            converged(QQ) =  0
         endif
      endif


! --- OLSEN CORRECTION : cor1=<ev|H_0-E)^-1|r>, cor2=<ev|(H_0-E)^-1|ev>
! --- NOTE: Krylov(TrueOrder) = (H_0-E)^-1 * |r>    
! --- workr(2*precon) =  eigenvector, |r> = (H-E) * eigenvector
! --- HKrylov(TrueOrder) =  (H_0-E)^-1 * eigenvector
! --- Olsen Correction : Krylov(TrueOrder) = (H_0-E)^-1 * |r> - correction


      if(olsen) then
! --- Build EigenVector.
         call zerovxd(workr(2*precon+1),psidim)
         call MPI_Bcast(EigenVector,(IntOrder+1)**2, &
                           MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(E,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD, &
                                                                 mpierr)
         call MPI_Bcast(Q,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

         if (mpiorder.gt.0) then
            Do j = 0, mpiOrder-1
               Do D = 1, PsiDim
                  workr(2*precon+D) = workr(2*precon+D) + &
                      EigenVector(j+mpirank*intordermpi,Q) * Krylov(D,j)
               enddo
            enddo
         endif
         call MPI_Reduce(workr(2*precon+1),HKrylov(1,MPIOrder),PsiDim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         call cpvxd(HKrylov(1,MPIOrder),workr(2*precon+1),PsiDim)
         if (mpirank.eq.0) then
            cor1=0.d0
            cor2=0.d0
            if(precon.eq.0) then
               do D = 1, PsiDim
                  cor1 = cor1 + workr(2*precon+D)*Krylov(D,mpiOrder)
                  cor2 = cor2 + workr(2*precon+D)**2 *  &
                       (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
               do D = 1, PsiDim
                  Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - &
                       (cor1/cor2) * workr(2*precon+D) * &
                       (RDiagonal(D)-E)/( (RDiagonal(D)-E)**2 + d2 )
               enddo
            else
               call cpvxd(workr(2*precon+1),HKrylov(1,mpiOrder),PsiDim)
               call precondit(HKrylov(1,mpiOrder),RDiagonal,davmat, &
                              davmat(1,precon+1),indxdav, &
                              E,ovl,d2,PsiDim,precon)
               do D = 1, PsiDim
                  cor1 = cor1 + workr(2*precon+D)*Krylov(D,mpiOrder)
                  cor2 = cor2 + workr(2*precon+D)*HKrylov(D,mpiOrder)
               enddo
               do D = 1, PsiDim
                  Krylov(D,mpiOrder) = Krylov(D,mpiOrder) - &
                       (cor1/cor2) * HKrylov(D,mpiOrder)
               enddo
            endif  ! precon
         endif
      endif     ! olsen


! ---  ORTHONORMALIZE  ---
      if(mpiorder.gt.0) then
         if (mpirank.gt.0) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonr(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD, mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_PRECISION,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                    MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif

      if (mpirank.eq.0) then
         call mpiovl(Krylov(1,MPIOrder),ovl,PsiDim)
         if(ovl.le.1.d-31) then     ! If vectors are linearly dependent
            write(6,'(/a)') '---------------------------------------------'
            write(6,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                             Steps,TrueOrder,Q,ovl,Error
            write(2,'(a,3i5,2e12.4)') 'BDAV: Steps, Order, Q, ovl, Error', &
                                             Steps,TrueOrder,Q,ovl,Error
            write(6,'(a)') 'BDAV: Vectors are linearly dependent!' 
            write(2,'(a)') 'BDAV: Vectors are linearly dependent!' 
            WRITE(6,'(a,20i3)') 'Qmap:     ', (qmap(i),i=0,2*BDim-1)
            WRITE(6,'(a,20i3)') 'Converged:',(converged(i),i=0,2*BDim-1)
            call stopnow
         endif
!.....Symmetrize coefficients, if desired.
         if (nmod.gt.0) then
            call rsymcoeff(Krylov(1,mpiOrder), &
                   HKrylov(1,mpiOrder),PsiDim,nmod,lf,ovl,TrueOrder)
            if(ovl .le. 1.d0) write(2,'(a,1pe9.2,a,i6)') &
             'rsymcoeff-norm =',ovl,', mpiOrder =',mpiOrder
            if(TrueOrder .gt. Bdim+1) then
               if(ovl.le.1.d-3) then
                  mpiOrder=mpiOrder-1
                  TrueOrder=TrueOrder-1
                   write(2,'(a,1pe9.2,a,2i6)') &
                    'BRDavStep: rsymcoeff-norm =',ovl, &
                  ',  Iteration stopped at TrueOrder,mpiOrder =', &
                  mpiOrder,TrueOrder
                  goto 300
            endif
            else
               if(ovl.le.1.d-6) then
                  write(2,'(a,1pe9.2,a,2i6)') &
                     'BRDavStep: rsymcoeff-norm =',ovl, &
                     ',  Execution stopped! TrueOrder,mpiOrder =', &
                     mpiOrder,TrueOrder
                  stop 1
               endif
            endif
            call mpiorthonr(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)
         endif
      endif


!.....Do the orthonormalisation twice
      if(mpiorder.gt.0) then
         if (mpirank.gt.0) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     mpirank-1,799+mpirank,MPI_COMM_WORLD,status,mpierr)

         call mpiorthonr(Krylov,Krylov(1,MPIOrder),mpiOrder-1,PsiDim)

         if (trueorder.gt.intordermpi) then
            if (mpiorder.eq.intordermpi) then
               if (trueorder.eq.(mpirank+1)*intordermpi) then
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                      MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
               else
                  call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                             MPI_DOUBLE_PRECISION,mpirank+1,800+mpirank, &
                                                  MPI_COMM_WORLD,mpierr)
               endif
            else
               call MPI_Send(Krylov(1,MPIOrder),PsiDim, &
                    MPI_DOUBLE_PRECISION,0,800+mpirank,MPI_COMM_WORLD,mpierr)
            endif
         endif

         nmpirecv=(trueorder-1)/intordermpi
         if (mpirank.eq.0.and.trueorder.gt.intordermpi) &
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                     nmpirecv,800+nmpirecv,MPI_COMM_WORLD,status,mpierr)
      endif

      if (mpirank.eq.0) call mpiovl(Krylov(1,MPIOrder),ovl,PsiDim)

! --- Send new vector to appropriate processor

      if (trueorder.ge.intordermpi) then
         nmpirecv=trueorder/intordermpi
         if (mpirank.eq.0) then
           call MPI_Send(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                                    nmpirecv,1000,MPI_COMM_WORLD,mpierr)
         endif
         if (mpirank.eq.nmpirecv) then
           call MPI_Recv(Krylov(1,MPIOrder),PsiDim,MPI_DOUBLE_PRECISION, &
                                    0,1000,MPI_COMM_WORLD,status,mpierr)
           mpiorder=mpiorder+1
         endif
      else
         if (mpirank.eq.0) mpiorder=mpiorder+1
      endif

! --- WRITE STEP AND ERROR IF DESIRED ---
      if (mpirank.eq.0) then
         E = EigenVal(Q)
         beta = EigenVector(QQ,Q)**2
         Call WriteStepBDav(Steps,TrueOrder,E,beta,Error,Q,QQ,Time)
      endif

      call MPI_Bcast(BOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(trueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      if(TrueOrder .ge. BOrder+BDim-1) goto 41
      QQ = QQ + 1
      goto 100
 41   QQ = 0

! --- BUILD HKRYLOV  ---

      If(lrreal) Then
         if (lmult) then
! --- Collecting and restructuring of the Davidson Vector is done as in
! --- complex case. But since the Funcr-routine needs real arrays and
! --- the complex arrays Psi and DtPsi are not used the data is copied
! --- to real via the array workr and the arrays Psi and DtPsi are
! --- used as real arrays in the Funcr-routine.
            call zerovxz(Psi,adim)
            blk2=mpirank*intordermpi-border
            call mpikry2psi(Psi,Krylov,blk2,mpiborder,mpiorder-1)
            call MPI_Reduce(Psi,DtPsi,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
            call cpvxz(DtPsi,Psi,adim)
            do i=1,adim
               workr(i)=dble(psi(i))
            enddo
!            call cpvxd(workr,psi,adim)
            Call Funcr(Time,workr(1),workr(adim+1))
!            call cpvxd(dtpsi,workr,adim)
            do i=1,adim
               DtPsi(i)=dcmplx(workr(adim+i))
            enddo
            call MPI_Bcast(DtPsi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
            blk2=mpirank*intordermpi-border
            call mpipsi2kry(DtPsi,HKrylov,blk2,mpiborder,mpiorder-1)
         else
! --- adjust length of adim, block(1), and ndim.
            adim=adim/bdim
            block(1)=block(1)/bdim
            do j=1,nmode
               ndim(j,1)=ndim(j,1)/bdim
            enddo
            j=0
            do i=0,bdim-1
               if (border+i.ge.intordermpi) then
                  if (mpirank.eq.(border+i)/intordermpi) then
                     call MPI_Send(Krylov(1,mpiborder+j),PsiDim, &
                           MPI_DOUBLE_PRECISION,0,3000+i,MPI_COMM_WORLD,mpierr)
                  elseif (mpirank.eq.0) then
                     call MPI_Recv(Krylov(1,MPIOrder),PsiDim, &
                        MPI_DOUBLE_PRECISION,(border+i)/intordermpi, &
                                    3000+i,MPI_COMM_WORLD,status,mpierr)
                  endif
               else
                  call cpvxd(Krylov(1,mpiborder+i),Krylov(1,mpiorder),psidim)
               endif
               Call Funcr(Time,Krylov(1,mpiOrder),HKrylov(1,mpiOrder))
               if (border+i.ge.intordermpi) then
                  if (mpirank.eq.(border+i)/intordermpi) then
                     call MPI_Recv(HKrylov(1,mpiborder+j),PsiDim, &
                           MPI_DOUBLE_PRECISION,0,6000+i,MPI_COMM_WORLD, &
                                                          status,mpierr)
                     j=j+1
                  elseif (mpirank.eq.0) then
                     call MPI_Send(HKrylov(1,mpiorder),PsiDim, &
                            MPI_DOUBLE_PRECISION,(border+i)/intordermpi, &
                                           6000+i,MPI_COMM_WORLD,mpierr)
                  endif
               else
                  call cpvxd(HKrylov(1,mpiorder),HKrylov(1,mpiborder+i),psidim)
               endif
            enddo
! --- restore length of adim, block(1), and ndim.
            adim=adim*bdim
            block(1)=block(1)*bdim
            do j=1,nmode
               ndim(j,1)=ndim(j,1)*bdim
            enddo
         endif
      Else
         call zerovxz(Psi,adim)
         if(lmult) then
            blk2=mpirank*intordermpi-border
            call mpikry2psi(Psi,Krylov,blk2,mpiborder,mpiorder-1)
         else
            Do blk = mpiborder,mpiorder-1
               Do D = 1,PsiDim
                  blk2=mpirank*intordermpi-border+blk
                  Psi(D+blk2*PsiDim) = Krylov(D,blk)
               enddo
            enddo
         endif
         call MPI_Reduce(Psi,DtPsi,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
         call cpvxz(DtPsi,Psi,adim)
         Call Func(Time,Psi,DtPsi)
         call MPI_Bcast(DtPsi,adim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD, &
                                                                 mpierr)
         if(lmult) then
            blk2=mpirank*intordermpi-border
            call mpipsi2kry(DtPsi,HKrylov,blk2,mpiborder,mpiorder-1)
         else
            Do blk = mpiborder,mpiorder-1
               Do D = 1,PsiDim
                  blk2=mpirank*intordermpi-border+blk
                  HKrylov(D,blk) = dble(DtPsi(D+blk2*PsiDim))
               enddo
            enddo
         endif
      EndIf

! --- BUILD NEW T-MATRIX ELEMENTS ---


      if (mpirank.gt.0) then
         call zerovxd(Tmat,((IntOrder+1)*(IntOrder+2))/2)
      else
         call zerovxd(Tmat(ij+1),((IntOrder+1)*(IntOrder+2))/2-ij)
      endif

      if (mpiBorder.lt.mpiorder) then
         do recvcount=1,mpirank*intordermpi
            if (mpirank.gt.0) then
               if (mpirank.eq.border/intordermpi) then
                  call MPI_Recv(Krylov(1,MPIOrder),PsiDim, &
                         MPI_DOUBLE_PRECISION,(recvcount-1)/intordermpi, &
                            2000+recvcount,MPI_COMM_WORLD,status,mpierr)
               else
                  call MPI_Recv(Krylov(1,MPIOrder),PsiDim, &
                         MPI_DOUBLE_PRECISION,mpirank-1, &
                            2000+recvcount,MPI_COMM_WORLD,status,mpierr)
               endif
            endif
            if (mpirank.lt.trueorder/intordermpi) then
               call MPI_Send(Krylov(1,mpiorder),PsiDim, &
                         MPI_DOUBLE_PRECISION,mpirank+1,2000+recvcount, &
                                                  MPI_COMM_WORLD,mpierr)
            endif
            do j=mpiborder,mpiorder-1
               ij=(mpirank*intordermpi+j)*(mpirank*intordermpi+j+1)/2
               ij=ij+recvcount
               Tmat(ij) = 0.d0
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij)+ Krylov(D,mpiorder)*HKrylov(D,j)
               enddo
            enddo
         enddo
         do j=0,mpiorder-1
            if (mpirank.lt.trueorder/intordermpi) then
               call MPI_Send(Krylov(1,j),PsiDim,MPI_DOUBLE_PRECISION, &
                       mpirank+1,2000+recvcount+j,MPI_COMM_WORLD,mpierr)
            endif
         enddo
         do i=mpiborder,mpiorder-1
            do j=0,i
               ij=(mpirank*intordermpi+i)*(mpirank*intordermpi+i+1)/2
               ij=ij+recvcount+j
               Tmat(ij) = 0.d0
               Do D = 1, PsiDim
                  Tmat(ij) = Tmat(ij)+ Krylov(D,j)*HKrylov(D,i)
               enddo
            enddo
         enddo
      else
        do sendcount=0,mpiorder-1
          call MPI_Send(Krylov(1,sendcount),PsiDim,MPI_DOUBLE_PRECISION, &
                  border/intordermpi,2001+intordermpi*mpirank+sendcount, &
                                                  MPI_COMM_WORLD,mpierr)
        enddo
      endif

      call MPI_Reduce(Tmat,EigenVector,((IntOrder+1)*(IntOrder+2))/2, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxd(Eigenvector,Tmat,((IntOrder+1)*(IntOrder+2))/2)

      mpiBOrder=mpiOrder

! --- DIAGONALISE T-MATRIX ---
      if (mpirank.eq.0) then
         ij = 0
         do j = 0, TrueOrder
            do i = 0, j
               ij = ij + 1
               EigenVector(i,j) = Tmat(ij)
            enddo
         enddo
      endif

      if (mpirank.eq.0) then
         BOrder=TrueOrder+1
         call MPI_Bcast(border,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         call pardsyev('V','U',BOrder,EigenVector,IntOrder+1,EigenVal, &
                        workr,lwork,ErrorCode)
         If (ErrorCode .NE. 0) Then
            ErrorCode = 3
            Return
         EndIf
      else
         call MPI_Bcast(border,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         if (border.gt.2*nmpi.and.lmpidsyev) &
            call STRTDSTEQR(ldz,l,lendmpi,icompz,n,nmaxit,mr)
      endif

! --- FIND THE EIGENVECTOR WITH THE LARGEST OVERLAP (if not GS is sought) -
      if (mpirank.eq.0) then
         if(rlaxnum.eq.920) then   ! follow
            qmin=2*BDim
            Do Q = 0, BDim-1
               beta=0.d0
               j = -1
               Do P = 0, TrueOrder
                  if(EigenVal(P).ge.rlxemin .and. EigenVal(P).le.rlxemax) then 
                     beta1= EigenVector(Q,P)**2
                     do i = 0, Q-1
                        if(qmap(i).eq.P) beta1=-1.d0
                     enddo
                     if(beta1 .gt. beta) then
                        j = P
                        beta = beta1
                     endif
                  endif
               enddo
               if(j.lt.0) then  ! Energy window too small.
                  beta=1.d+99   ! Take vectors with energies closest to window.
                  Do P = 0, TrueOrder
                     beta1 = min(abs(rlxemin-EigenVal(P)),abs(EigenVal(P)-rlxemax))
                     if(beta1.lt.beta) then
                        beta=beta1
                        j = P
                     endif 
                  enddo
               endif
               qmap(Q) = j
               qmin = min(qmin,j)
            enddo  ! Q

            if(full) qmin=0
            qmin=max(qmin-1,0)               ! Set higher entries of qmap.
            Do Q = BDim,min(mxcv,TrueOrder)  ! They may be used when some
   51          do i = 0, Q-1                 ! vectors converge. 
                  if(qmap(i).eq.qmin) then
                     qmin=qmin+1
                     if(qmin.gt.TrueOrder) qmin=0
                     goto 51
                  endif
               enddo
               qmap(Q)=qmin
            enddo   !  Q
         elseif(rlaxnum.eq.0 .and. rlxemin.gt.-1.d99) then
            QQ=0
            do Q = 0, TrueOrder
               if(EigenVal(Q).gt.rlxemin .and. QQ.lt.mxcv) then
                  qmap(QQ)=Q
                  QQ=QQ+1
                  i=Q
                  if(QQ.eq.Bdim) goto 52
               endif
            enddo
   52       Q=i+1
            if(QQ.lt.Bdim) then
               qmap(QQ)=Q
               QQ=QQ+1
               Q=Q+1
               goto 52
            endif
            if(qmap(0).ge.1) then  ! Take the first state below the block
               qmap(QQ)=qmap(0)-1  ! as additional vector (when others
               QQ=QQ+1             ! have converged).
            endif
            if(qmap(0).ge.2) then  ! Take the second state below the block
               qmap(QQ)=qmap(0)-2  ! as additional vector (when others
               QQ=QQ+1             ! have converged).
            endif
            idown=3
   53       if(QQ.lt.mxcv) then    ! Set qmap below and above the block
               if(qmap(0).ge.idown) then
                  qmap(QQ)=qmap(0)-idown
                  QQ=QQ+1
                  idown=idown+1
               endif
               qmap(QQ)=Q          ! Take the states following the block
               QQ=QQ+1             ! as additional vectors (when others
               Q=Q+1               ! have converged).
               goto 53
            endif
         endif      ! (if rlaxnum.eq.920 ...  elseif rlaxnum.eq.0)
         epse=(EigenVal(qmap(Bdim-1))-EigenVal(qmap(0)))/(Bdim*90.d0)
      endif         ! (mpirank.eq.0)

! --- CONTINUE IF NOT CONVERGED --- END OF LOOP  ---
      call MPI_Bcast(TrueOrder,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(full,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(ErrorM,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(olderr,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
      If( TrueOrder .GT. IntOrder-BDim ) Goto 300
      If( TrueOrder .GE. PsiDim-BDim   ) Goto 300
      If( full .or. (ErrorM+olderr.GT.TolError) ) Goto 99

! --- CONVERGENCE! Build new A-vector ---
 300  continue
      call MPI_Bcast(qmap,8*bdim+10,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(EigenVector,(IntOrder+1)**2,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call zerovxz(Psi,adim)
      call zerovxz(DtPsi,adim)
      if(lmult) then
         call mpikry2finalpsi(Psi,DtPsi,Krylov,HKrylov, &
                           EigenVector,TrueOrder,mpiorder,qmap)
      else
         Do QQ = 0,BDim-1
            Q = qmap(QQ)
            Do P = 0,mpiorder-1
               Do D = 1,PsiDim
                  Psi(D+QQ*PsiDim)   = Psi(D+QQ*PsiDim) +  &
                        Krylov(D,P)*EigenVector(mpirank*intordermpi+P,Q)
                  DtPsi(D+QQ*PsiDim) = DtPsi(D+QQ*PsiDim) + &
                       HKrylov(D,P)*EigenVector(mpirank*intordermpi+P,Q)
               EndDo
            EndDo
         EndDo
      endif
      call MPI_Reduce(Psi,workc,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,Psi,adim)
      call MPI_Reduce(DtPsi,workc,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,0, &
                                                  MPI_COMM_WORLD,mpierr)
      call cpvxz(workc,DtPsi,adim)

      if (mpirank.eq.0) then
         Do QQ = 0, BDim-1
            Q = qmap(QQ) 
            if(FirstCall) then
!...........This is for first overlaps in wrtrlxbdav
               do j = 0, TrueOrder
                  ovlps(j) = EigenVector(Q,j)**2
               enddo
            endif
            E=EigenVal(Q)
            beta=EigenVector(QQ,Q)**2
            j=QQ
            if(beta.lt.0.5d0 .and. TrueOrder.gt.Bdim) then
               beta=0.d0
               do i=0,Bdim+1
                  if(EigenVector(i,Q)**2 .gt. beta) then
                     beta=EigenVector(i,Q)**2
                     j=i
                  endif
               enddo
               if(beta.lt.0.67d0) then
                  beta=EigenVector(QQ,Q)**2
                  j=QQ
               endif
            endif
            call wrtrlxbdav(time,trueorder,QQ,Q,E,beta, &
                   oldenergy(j,kskip),converged(QQ), &
                   EigenVal,ovlps,FirstCall,IntOrder,BDim)
         enddo
         Do QQ = 0, BDim+1
            Q = qmap(QQ)
            oldenergy(QQ,kskip)=EigenVal(Q)
         enddo

         FirstCall = .false.
         full      = .false.
         edav = E
      endif  ! (mpirank.eq.0)

!#endif

      Return
      End Subroutine MPIBRDavStep

#endif
      endmodule davlibpar
