! **********************************************************************        
!                                                                               
!                   module DAVUTILS
!                                                                               
!   MCTDH module containing utility subroutines for Davidson
!                                                          
! Contains:     
!    daverrormsg: generates error messages
!    orthonc: orthonormalises a ser of complex vectors
!    autoblock: automatic generation of start vectors
!    sautoblock: automatic generation of start vectors (symmetrized for 
!                bosons)
!    rsymcoeff: symmetrises coefficients
!    csautoblock: automatic generation of complex start vectors 
!                 (symmetrized for bosons or fermions)
!    csymcoeff: symmetrises coefficients
!
!
! **********************************************************************
!
      module davutils

use decimal, only: long, dop
use xvlib,   only: xvixddo, xvixdzo
use op1lib,  only: zeromxz, zeromxd, normvxz, normvxd, initvxi, cpvxd, cpvxz
use logdat

      implicit none
      private
      public :: daverrormsg,orthonc,autoblock,sautoblock,rsymcoeff,  &
                csautoblock,csymcoeff,realdaverror,orthonr,precondit, &
                precondith,preconditz,cprecondit,orthon1

      contains


! **********************************************************************
! *                                                                    *
! *                       SUBROUTINE DAVERRORMSG                       *
! *                                                                    *
! * Generates for a given error number returned by "SILStep/SILStep2/  *
! * SILStep3" a corresponding error message.                           *
! *                                                                    *
! * Input parameters:                                                  *
! *   Error: Error code returned by DAVStep/RDAVStep/CDAVSTEP          *
! *                                                                    *
! * Output parameters:                                                 *
! *   Msg:   Error message.                                            *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      Subroutine DavErrorMsg (Error,Msg)

      Implicit None

      Integer(long), intent(in) :: Error
      Character(len=*), intent(out) :: Msg

! --- GENERATE ERROR MESSAGE ---

      If (Error .Eq. 1) Then
         Msg = 'Illegal integration order'
      ElseIf (Error .Eq. 2) Then
         Msg = 'Stepsize underflow'
      ElseIf (Error .Eq. 3) Then
         Msg = 'Diagonalisation failed'
      ElseIf (Error .Eq. 4) Then
         Msg = 'Illegal initial stepsize'
      Else
         Msg = 'Unknown error occurred'
      EndIf

      Return
      End subroutine daverrormsg


! **********************************************************************
!    Orthornormalisation of complex vectors.
!    v is orthonormalized against Krylov(0)...Krylov(n)
! ----------------------------------------------------------------------

      Subroutine orthonc(Krylov,v,ovl,n,PsiDim)

      Implicit None

      Integer(long), intent(in) :: PsiDim,n
      Integer(long) :: i,D
      real(dop), intent(out) :: ovl
      complex(dop), dimension(psidim,0:n), intent(in) :: Krylov
      complex(dop), dimension(psidim), intent(out) :: v
      complex(dop) :: ovlc

      Do i = 0, n
         ovlc = (0.d0,0.d0)
         Do D = 1,PsiDim
            ovlc = ovlc + DConjg(Krylov(D,i))*v(D)
         enddo
         Do D = 1,PsiDim
            v(D) = v(D)-ovlc*Krylov(D,i)
         enddo
      enddo

      ovl = 0.d0
      Do D = 1,PsiDim
         ovl = ovl + abs(v(D))**2
      enddo
      ovl=sqrt(ovl)
      if(ovl .le. 1.d-40) return

      Do D = 1,PsiDim
         v(D)=v(D)/ovl
      enddo


      Return
      End subroutine orthonc

! **********************************************************************
!    Orthornormalisation of a single pair of complex vectors.
!    v is orthonormalized against w
! ----------------------------------------------------------------------

      Subroutine orthon1(w,v,ovl,dim,lnorm)

      Implicit None

      logical(kind=4), intent(in) :: lnorm
      Integer(long), intent(in) :: dim
      Integer(long) :: D
      real(dop), intent(out) :: ovl
      complex(dop), dimension(dim), intent(in)  :: w
      complex(dop), dimension(dim), intent(out) :: v
      complex(dop) :: ovlc

      ovlc = (0.d0,0.d0)
      Do D = 1,dim
         ovlc = ovlc + DConjg(w(D))*v(D)
      enddo
      Do D = 1,dim
         v(D) = v(D)-ovlc*w(D)
      enddo

      ovl = 0.d0
      Do D = 1,dim
         ovl = ovl + abs(v(D))**2
      enddo
      ovl=sqrt(ovl)

      if(ovl .le. 1.d-40.and..not.lnorm) return

! Only normalise if required.
      Do D = 1,dim
         v(D)=v(D)/ovl
      enddo

      Return
      End subroutine orthon1

!#######################################################################
!     automatic generation of start vectors for block-davidson 
! ----------------------------------------------------------------------

      subroutine autoblock(Psi,RDiagonal,PsiDim,adim,BDim,rlxemin, &
                           indxdav,workr,nstate,block,nmode,ilog,idx)

      use getdavmatmod, only: bsort, setm

      Implicit None

      Integer(long), intent(in) :: PsiDim,adim,BDim,nmode,ilog,nstate
      Integer(long) :: iz,i,ii,s,bk,jj,m
      Integer(long), dimension(bdim), intent(out) :: indxdav(BDim)
      Integer(long), dimension(nstate), intent(in) :: block
      Integer(long), dimension(nmode+1), intent(out) :: idx
      Real(dop), dimension(psidim), intent(out) :: RDiagonal
      Real(dop), dimension(bdim), intent(out) ::workr
      Real(dop), intent(in) :: rlxemin
      Real(dop) :: diag
      Complex(dop), dimension(adim), intent(out) :: Psi(adim)

      write(ilog,'(/a)')   &
        '******  A-coefficients generated by autoblock  ******'
      if(nstate.gt.1) then
         write(ilog,'(8x,a,4x,a)') 'J   E-diag [au]',  &
                                       's,  j1 j2 j3 j4 j5 ...'
      else
         write(ilog,'(8x,a,4x,a)') 'J   E-diag [au]',   &
                                       'j1 j2 j3 j4 j5 ...'
      endif

      if(rlxemin.gt.-1.d99) then
         do i = 1, PsiDim
            if(RDiagonal(i).lt.rlxemin) RDiagonal(i)=1.d99
         enddo
      endif
      do i = 1, adim
         Psi(i) = (0.d0,0.d0)
      enddo

!.....Similar to subroutine sortdiagonal
      workr(1)=RDiagonal(1)
      indxdav(1)=1

      do i = 2, BDim
         workr(i) = RDiagonal(i)
         indxdav(i) = i
         call bsort(workr,indxdav,i)
      enddo

      do i = BDim+1, PsiDim
         diag = RDiagonal(i)
         if(diag.lt.workr(BDim)) then
            workr(BDim) = diag
            indxdav(BDim) = i
            call bsort(workr,indxdav,BDim)
         endif
      enddo
!.....indxdav contains the indices of the BDim lowest diagonal
!.....elements of the Hamilton matrix. In case of multi-set
!.....electronic states, an additional mapping is necessary.

      if(nstate.gt.1) then  ! multi-set electronic states
         bk = 0
         do i = 1, BDim
            ii=indxdav(i)
            iz=0      ! iz will become zpsi(s)-1
            do s=1,nstate
               if(ii .gt. block(s)/BDim ) then
                  ii = ii - block(s)/BDim
                  iz = iz + block(s)
               else
                  bk = block(s)/BDim
                  goto 220 
               endif
            enddo        ! s=1,nstate
 220        jj = iz+ii+(i-1)*bk
            Psi(jj) = (1.d0,0.d0)
            call setm(jj,idx)
            write(ilog,'(i9,f14.9,i5,", ",35i3)') jj,workr(i),  &
                  idx(nmode+1),(idx(m)+1,m=1,nmode)
         enddo           ! i=1,BDim

      else               ! no multi-set electronic states
         do i = 1, BDim
            jj=indxdav(i)+(i-1)*PsiDim 
            Psi(jj) = (1.d0,0.d0)
            call setm(jj,idx)
            write(ilog,'(i9,f14.9,3x,35i3)') jj,workr(i),(idx(m)+1,m=1,nmode)
         enddo
      endif
      write(ilog,'(a)')

      return
      end subroutine autoblock

!#######################################################################
!     automatic generation of start vectors for block-davidson
!     This version symmetrizes the vectors (bosons or fermions)
! ----------------------------------------------------------------------

      subroutine sautoblock(Psi,RDiagonal,Krylov,PsiDim,adim,BDim,  &
                            nmod,rlxemin,indxdav,workr,lf,ilog,idx)

      use getdavmatmod, only: bsort, setm

      Implicit None

      Integer(long), intent(in) :: PsiDim,adim,BDim,nmod,ilog
      Integer(long) :: i,j,jj,m,nz,nc,nfak
      Integer(long), dimension(bdim), intent(out) :: indxdav(2*BDim)
      Integer(long), dimension(nmod+2), intent(out) :: idx
      Real(dop), dimension(psidim), intent(out) :: RDiagonal
      Real(dop), dimension(2*bdim), intent(out) ::workr
      Real(dop), dimension(psidim,bdim), intent(out) :: Krylov
      Real(dop), intent(in) :: rlxemin
      Real(dop) :: diag,ovl,norm
      Logical(kind=4),intent(in) :: lf
      Logical(kind=4) :: lpr
      Complex(dop), dimension(adim), intent(out) :: Psi(adim)

      lpr=.true. 
      if(lpr) then
         write(2,*)
         lpr=.false.
      endif
      if(rlxemin.gt.-1.d99) then
         do i = 1, PsiDim
            if(RDiagonal(i).lt.rlxemin) RDiagonal(i)=2.d99
         enddo
      endif
      call zeromxd(Krylov,PsiDim,BDim)

      nfak=2
      do i = 3, nmod
         nfak=nfak*i
      enddo

!.....Remove equivalent entries
      do i = 1, min(nmod*BDim,PsiDim)
         do j = i+1, PsiDim
            if(abs(RDiagonal(j)-RDiagonal(i)) .lt. &
                1.d-12*(abs(RDiagonal(i))+abs(RDiagonal(i))+1.d-12)) &
                  RDiagonal(j)=2.d99
         enddo
      enddo

!.....Similar to subroutine sortdiagonal
      nc=1
      if(lf) nc=2
      workr(1)=RDiagonal(nc)
      indxdav(1)=nc

      do i = 2, 2*BDim
  10     nc=nc+1
         if(RDiagonal(nc).gt.1.d99 .and. nc.lt.(PsiDIm-BDim)) goto 10
         workr(i) = RDiagonal(nc)
         indxdav(i) = nc 
         call bsort(workr,indxdav,i)
      enddo

      do i = nc+1, PsiDim
         diag = RDiagonal(i)
         if(diag.lt.workr(2*BDim)) then
            workr(2*BDim) = diag
            indxdav(2*BDim) = i
            call bsort(workr,indxdav,2*BDim)
            do j = i+1, PsiDim
               if(abs(RDiagonal(j)-diag) .lt. &
                   1.d-12*(abs(RDiagonal(i))+abs(diag)+1.d-12)) &
                      RDiagonal(j)=2.d99
            enddo
         endif
      enddo

! --- indxdav is set, now generate the vectors. RDiagonal serves as work
      nz=1
      nc=0
 21   nc=nc+1 
      Krylov(indxdav(nc),nz) = 1.d0
      call rsymcoeff(Krylov,RDiagonal,PsiDim,nmod,lf,ovl,0)
      if(ovl .lt. 1.d-9 .and. nc .le. 2*BDIM) goto 21
      call normvxd(Krylov(1,1),norm,PsiDim) ! compute norm
      call xvixddo(norm,Krylov(1,1),PsiDim) ! normalise
 20   nz=nz+1
      nc=nc+1
      if(nc.ge.2*BDim) then
         write(2,'(a,3i5)')'sautoblock: nc, nz, BDim',nc,nz,BDim
         write(6,'(a,3i5)')'sautoblock: nc, nz, BDim',nc,nz,BDim
         stop 2
      endif
      Krylov(indxdav(nc),nz) = 1.d0
      call rsymcoeff(Krylov(1,nz),RDiagonal,PsiDim,nmod,lf,ovl,0)
      if(ovl .gt. 1.d-9) &    ! the second index of Krylov starts here from 1,
        call orthonr(Krylov,Krylov(1,nz),ovl,nz-2,PsiDim) ! but in orthonr from 0.

      if(ovl .lt. 1.d-9) then
         nz=nz-1
         write(2,'(a,3i4,e10.2)') 'sautoblock: Removed',nc,nz,indxdav(nc),ovl
      else
         call normvxd(Krylov(1,nz),norm,PsiDim) ! compute norm
         write(2,'(a,f9.3,a,i5,a,f15.10)') 'r-symcoeff: norm =',norm,  &
         ',  n! =',nfak,',  1000*diff/n! =',(1.d0-norm/dble(nfak))*1.d3
         call xvixddo(norm,Krylov(1,nz),PsiDim) ! normalise
      endif 
      if(nz.lt.BDim) goto 20
     
! --- Print head lines.
      write(ilog,'(/a)') &
         '******  A-coefficients generated by s-autoblock  ******'
      write(ilog,'(8x,a,12x,a,13x,a)') 'J','A','j1 j2 j3 j4 j5 ...'
 
      do i = 1, BDim
         do j = 1, PsiDim
            jj=j+(i-1)*PsiDim
            Psi(jj) = Krylov(j,i)
            if(abs(Psi(jj)).gt.1.d-8) then
               call setm(jj,idx)
               write(ilog,'(i9,"  (",f9.6,",",f9.6,")",2x,35i3)')  &
                     jj,psi(jj),(idx(m)+1,m=1,nmod+1)
            endif
         enddo
      enddo
      write(ilog,'(a)')

      return
      end subroutine sautoblock

! **********************************************************************
!  Symmetrisation of coefficients.
!  Similar to subroutine symcoeff, but using real arithmetic.
! ----------------------------------------------------------------------

      subroutine rsymcoeff(psi,work,PsiDim,nmod,lf,norm,step)

      implicit none

      integer(long), parameter :: mxp=240
      integer(long), intent(in) :: PsiDim,nmod,step
      integer(long), save :: nfak,nf,dm
      integer(long) :: j,k,m,mm
      integer(long), dimension(mxp), save :: prmttn
      integer(long), dimension(12) ::gpoint
      real(dop), dimension(psidim), intent(out) :: psi,work
      real(dop), intent(out) :: norm
      logical(kind=4), save :: first
      logical(kind=4) :: lf
      data        first /.true./

      if(first) then
         nf=6
         do j = 4, nmod+1
            nf = nf*j
         enddo
         if(2*nf.gt.mxp) then
            write(6,'(a)') 'Too many modes for rsymcoeff (davutils.f)'
            write(6,'(a,i4,a,i6)') 'nmode =',nmod,', (nmode+1)! =',nf
            write(6,'(a,i6)') 'Increase mxp to 2*((nmode+1)!)=',2*nf
            stop
         endif
         nf=nf+1
         call permutation(prmttn,prmttn(nf),nfak,nmod)
         dm=int((PsiDim+0.1d0)**(1.d0/dble(nmod)))  ! dim(1,1)
         first=.false.
      endif

      call cpvxd(psi,work,PsiDim)
      call initvxi(gpoint,nmod)
      j=0

 10   j=j+1
! --- Loop over all permutations (except first one).
      do nf = 2, nfak
         k=0
! --- Compute mapped (composite) k index from transformed mode indices
         do m=nmod, 2, -1
            mm=prmttn(m+(nmod+1)*(nf-1))
            k = dm*(k+gpoint(mm)-1)
         enddo ! m
         mm=prmttn(1+(nmod+1)*(nf-1))
         k = k+gpoint(mm)
         if(lf) then
            psi(j)=psi(j)+work(k)*prmttn(nf*(nmod+1))  ! (fermions)
         else
            psi(j)=psi(j)+work(k)       !(bosons)
         endif
      enddo    ! nf
! --- Set the mode indices for next j
! --- gpoint contains the decoded indices for the encoded index j
      do m=1,nmod
         if(gpoint(m).lt.dm) then
            gpoint(m)=gpoint(m)+1
            goto 10
         else
            gpoint(m)=1
         endif
      enddo    ! m

      call normvxd(psi,norm,PsiDim) ! compute norm
!       write(92,'(a,f11.4,a,i6,a,f18.12)') 'rsymcoeff: norm =',   !!
!      +  norm,',  n! =',nfak,',  diff =', dble(nfak)-norm         !!
!       call myflush(92)                                           !!

      if(step .le. 0) return

      if(norm.lt.nfak*0.5d0) then
         write(2,'(a,f11.4,a,f11.4,a,i5)') 'rsymcoeff: norm=', &
         norm,', diff =',dble(nfak)-norm,', TrueOrder =',step
      endif
      if(norm.lt.1.d-9) then
         if(lf) then
           write(6,*) 'rsymcoeff: Cannot asymmetrize coeffs. norm=',norm
           write(2,*) 'rsymcoeff: Cannot asymmetrize coeffs. norm=',norm
         else
           write(6,*) 'rsymcoeff: Cannot symmetrize coeffs. norm=',norm
           write(2,*) 'rsymcoeff: Cannot symmetrize coeffs. norm=',norm
         endif
!          write(92,*) 'rsymcoeff: Cannot symmetrize coeffs. norm=',norm !
         stop
      endif

      return
      end subroutine rsymcoeff

! **********************************************************************
!#######################################################################
!     automatic generation of start vectors for block-davidson
!     This version symmetrizes the vectors (bosons or fermions)
!     Alomst identical to sautoblock, but with complex Krylov.       
! ----------------------------------------------------------------------

      subroutine csautoblock(Psi,RDiagonal,Krylov,PsiDim,adim,BDim,  &
                            nmod,rlxemin,indxdav,workr,lf,ilog,idx)

      use getdavmatmod, only: bsort, setm

      Implicit None

      Integer(long), intent(in) :: PsiDim,adim,BDim,nmod,ilog
      Integer(long) :: i,j,jj,m,nz,nc,nfak
      Integer(long), dimension(bdim), intent(out) :: indxdav(2*BDim)
      Integer(long), dimension(nmod+2), intent(out) :: idx
      Real(dop), dimension(psidim), intent(out) :: RDiagonal
      Real(dop), dimension(2*bdim), intent(out) ::workr
      Complex(dop), dimension(psidim,bdim+1), intent(out) :: Krylov
      Real(dop), intent(in) :: rlxemin
      Real(dop) :: diag,ovl,norm
      Logical(kind=4),intent(in) :: lf
      Logical(kind=4) :: lpr
      Complex(dop), dimension(adim), intent(out) :: Psi(adim)

      lpr=.true. 
      if(lpr) then
         write(2,*)
         lpr=.false.
      endif
      if(rlxemin.gt.-1.d99) then
         do i = 1, PsiDim
            if(RDiagonal(i).lt.rlxemin) RDiagonal(i)=2.d99
         enddo
      endif
      call zeromxz(Krylov,PsiDim,BDim)

      nfak=2
      do i = 3, nmod
         nfak=nfak*i
      enddo

!.....Remove equivalent entries
      do i = 1, min(nmod*BDim,PsiDim)
         do j = i+1, PsiDim
            if(abs(RDiagonal(j)-RDiagonal(i)) .lt. &
               1.d-12*(abs(RDiagonal(i))+abs(RDiagonal(i))+1.d-12)) &
                  RDiagonal(j)=2.d99
         enddo
      enddo

!.....Similar to subroutine sortdiagonal
      nc=1
      if(lf) nc=2
      workr(1)=RDiagonal(nc)
      indxdav(1)=nc

      do i = 2, 2*BDim
  10     nc=nc+1
         if(RDiagonal(nc).gt.1.d99 .and. nc.lt.(PsiDIm-BDim)) goto 10
         workr(i) = RDiagonal(nc)
         indxdav(i) = nc 
         call bsort(workr,indxdav,i)
      enddo

      do i = nc+1, PsiDim
         diag = RDiagonal(i)
         if(diag.lt.workr(2*BDim)) then
            workr(2*BDim) = diag
            indxdav(2*BDim) = i
            call bsort(workr,indxdav,2*BDim)
            do j = i+1, PsiDim
               if(abs(RDiagonal(j)-diag) .lt. &
                  1.d-12*(abs(RDiagonal(i))+abs(diag)+1.d-12)) &
                    RDiagonal(j)=2.d99
            enddo
         endif
      enddo

! --- indxdav is set, now generate the vectors. Krylov(BDim+1) serves as work
      nz=1
      nc=0
 21   nc=nc+1 
      Krylov(indxdav(nc),nz) = (1.d0,0.d0)
      call csymcoeff(Krylov,Krylov(1,BDim+1),PsiDim,nmod,lf,ovl,0)
      if(ovl .lt. 1.d-9 .and. nc .le. 2*BDIM) goto 21
      call normvxz(Krylov(1,1),norm,PsiDim) ! compute norm
      call xvixdzo(norm,Krylov(1,1),PsiDim) ! normalise
 20   nz=nz+1
      nc=nc+1
      if(nc.ge.2*BDim) then
         write(2,'(a,3i5)')'csautoblock: nc, nz, BDim',nc,nz,BDim
         write(6,'(a,3i5)')'csautoblock: nc, nz, BDim',nc,nz,BDim
         stop 2
      endif
      Krylov(indxdav(nc),nz) = (1.d0,0.d0)
      call csymcoeff(Krylov(1,nz),Krylov(1,BDim+1),PsiDim,nmod,lf,ovl,0)
      if(ovl .gt. 1.d-9) call orthonc(Krylov,Krylov(1,nz),ovl,nz-2,PsiDim)
      if(ovl .lt. 1.d-9) then
         nz=nz-1
         write(2,'(a,3i4,e10.2)') 'csautoblock: Removed', nc,nz,indxdav(nc),ovl
      else
         call normvxz(Krylov(1,nz),norm,PsiDim) ! compute norm
         write(2,'(a,f9.3,a,i5,a,f15.10)') 'c-symcoeff: norm =',norm,  &
            ',  n! =',nfak,',  1000*diff/n! =',(1.d0-norm/dble(nfak))*1.d3
         call xvixdzo(norm,Krylov(1,nz),PsiDim) ! normalise
      endif 
      if(nz.lt.BDim) goto 20
     
! --- Print head lines.
      write(ilog,'(/a)')  &
       '******  A-coefficients generated by cs-autoblock  ******'
      write(ilog,'(8x,a,12x,a,13x,a)') 'J','A','j1 j2 j3 j4 j5 ...'
 
      do i = 1, BDim
         do j = 1, PsiDim
            jj=j+(i-1)*PsiDim
            Psi(jj) = Krylov(j,i)
            if(abs(Psi(jj)).gt.1.d-8) then
               call setm(jj,idx)
               write(ilog,'(i9,"  (",f9.6,",",f9.6,")",2x,35i3)')  &
                     jj,psi(jj),(idx(m)+1,m=1,nmod+1)
            endif
         enddo
      enddo
      write(ilog,'(a)')

      return
      end subroutine csautoblock

! **********************************************************************
!  Symmetrisation of coefficients.
!  Similar to subroutine symcoeff, but using real arithmetic.
! ----------------------------------------------------------------------

      subroutine csymcoeff(psi,work,PsiDim,nmod,lf,norm,step)

      implicit none

      integer(long), parameter :: mxp=240
      integer(long), intent(in) :: PsiDim,nmod,step
      integer(long), save :: nfak,nf,dm
      integer(long) :: j,k,m,mm
      integer(long), dimension(mxp), save :: prmttn
      integer(long), dimension(12) ::gpoint
      complex(dop), dimension(psidim), intent(out) :: psi,work
      real(dop), intent(out) :: norm
      logical(kind=4), save :: first
      logical(kind=4) :: lf
      data        first /.true./

      if(first) then
         nf=6
         do j = 4, nmod+1
            nf = nf*j
         enddo
         if(2*nf.gt.mxp) then
            write(6,'(a)') 'Too many modes for csymcoeff (davutils.f)'
            write(6,'(a,i4,a,i6)') 'nmode =',nmod,', (nmode+1)! =',nf
            write(6,'(a,i6)') 'Increase mxp to 2*((nmode+1)!)=',2*nf
            stop
         endif
         nf=nf+1
         call permutation(prmttn,prmttn(nf),nfak,nmod)
         dm=int((PsiDim+0.1d0)**(1.d0/dble(nmod)))  ! dim(1,1)
         first=.false.
      endif

      call cpvxz(psi,work,PsiDim)
      call initvxi(gpoint,nmod)
      j=0

 10   j=j+1
! --- Loop over all permutations (except first one).
      do nf = 2, nfak
         k=0
! --- Compute mapped (composite) k index from transformed mode indices
         do m=nmod, 2, -1
            mm=prmttn(m+(nmod+1)*(nf-1))
            k = dm*(k+gpoint(mm)-1)
         enddo ! m
         mm=prmttn(1+(nmod+1)*(nf-1))
         k = k+gpoint(mm)
         if(lf) then
            psi(j)=psi(j)+work(k)*prmttn(nf*(nmod+1))  ! (fermions)
         else
            psi(j)=psi(j)+work(k)       !(bosons)
         endif
      enddo    ! nf
! --- Set the mode indices for next j
! --- gpoint contains the decoded indices for the encoded index j
      do m=1,nmod
         if(gpoint(m).lt.dm) then
            gpoint(m)=gpoint(m)+1
            goto 10
         else
            gpoint(m)=1
         endif
      enddo    ! m

      call normvxz(psi,norm,PsiDim) ! compute norm
!       write(92,'(a,f11.4,a,i6,a,f18.12)') 'csymcoeff: norm =', !!
!      +  norm,',  n! =',nfak,',  diff =', dble(nfak)-norm       !! 
!       call myflush(92)                                         !!

      if(step .le. 0) return

      if(norm.lt.nfak*0.5d0) then
         write(2,'(a,f11.4,a,f11.4,a,i5)') 'csymcoeff: norm=',  &
         norm,', diff =',dble(nfak)-norm,', TrueOrder =',step
      endif
      if(norm.lt.1.d-9) then
         if(lf) then
           write(6,*) 'csymcoeff: Cannot asymmetrize coeffs. norm=',norm
           write(6,*) 'csymcoeff: Cannot asymmetrize coeffs. norm=',norm
         else
           write(2,*) 'csymcoeff: Cannot symmetrize coeffs. norm=',norm
           write(2,*) 'csymcoeff: Cannot symmetrize coeffs. norm=',norm
         endif
!          write(92,*) 'csymcoeff: Cannot symmetrize coeffs. norm=',norm
         stop
      endif

      return
      end subroutine csymcoeff

!***********************************************************************

      Subroutine realdaverror(cor1)

use global
use channels

      Implicit None

      real(dop),intent(in) :: cor1


      write(6,'(/a,/a,1pe11.3,/a,/a/a,/a,/a,/a,/a)') &
       '#############################################################', &
       ' A-vector is not real! Imag-part =', cor1,                      &
       ' rDAV and rrDAV require real WF and Hamiltonians.',              &
       ' Check that the intitial WF is real. (Set realpsi in Init_WF).',&
       ' If FFT or PLeg is used in combination with rDAV or rrDAV,',    &
       ' then "realphi" (Run-Section) and "RK8/spf=1.d-9,imp-ortho"',   &
       ' should be set (Integrator-Section).',                          &
       ' Do not use energyorb in this case, better use natorb.',        &
       '#############################################################'

      write(ilog,'(/a,/a,1pe11.3,/a,/a,/a,/a,/a,/a,/a)') &
       '#############################################################', &
       ' A-vector is not real! Imag-part =', cor1,                      &
       ' rDAV and rrDAV require real WF and Hamiltonians.',             &
       ' Check that the intitial WF is real. (Set realpsi in Init_WF).',&
       ' If FFT or PLeg is used in combination with rDAV or rrDAV,',    &
       ' then "realphi" (Run-Section) and "RK8/spf=1.d-9,imp-ortho"',   &
       ' should be set (Integrator-Section).',                          &
       ' Do not use energyorb in this case, better use natorb.',        &
       '#############################################################'

      routine='realdaverror (rdavstep or brdavstep)'
      message='Complex A-vector for real Davidson found.'
      call errormsg

      Return
      End subroutine realdaverror

! **********************************************************************

      Subroutine orthonr(Krylov,v,ovl,n,PsiDim)
!.....Orthornormalisation of real vectors. 
!.....v is orthonormalized against Krylov(0)...Krylov(n)

      Implicit None

      Integer(long)                                :: i,D
      Integer(long), intent(in)                    :: PsiDim,n
      Real(dop), intent(out)                       :: ovl
      Real(dop), dimension(PsiDim), intent(inout)  :: v
      Real(dop), dimension(PsiDim,0:n), intent(in) :: Krylov

      Do i = 0, n
         ovl =0.d0
         Do D = 1,PsiDim
            ovl = ovl + Krylov(D,i)*v(D)
         enddo
         Do D = 1,PsiDim
            v(D) = v(D)-ovl*Krylov(D,i)
         enddo
      enddo
      ovl =0.0_dop
      Do D = 1,PsiDim
         ovl = ovl + v(D)**2
      enddo
      ovl=sqrt(ovl)
      if(ovl .le. 1.0e-40_dop) return
      Do D = 1,PsiDim
         v(D)=v(D)/ovl
      enddo


      Return
      End Subroutine orthonr

! **********************************************************************

      subroutine precondit(vec,RDiagonal,davmat,eigval,indxdav,&
                           E,Error,d2,PsiDim,nn)

!$use omp_lib

use openmpmod

      implicit none

      integer(long)                                  :: i,j,k
      integer(long), intent(in)                      :: nn,PsiDim
      integer(long), dimension(nn), intent(in)       :: indxdav
      real(dop), dimension(nn), intent(in)           :: eigval
      real(dop), dimension(nn)                       :: vv,ww
      real(dop), dimension(nn,nn), intent(in)        :: davmat
      real(dop), dimension(PsiDim), intent(in)       :: RDiagonal
      real(dop), dimension(PsiDim), intent(inout)    :: vec
      real(dop), intent(in)                          :: E,d2
      real(dop), intent(out)                         :: Error

! --- Multiply with inverted sub-Hamilton matrix.
! --- vv = eigvec * (1/eigval) * eigvec^T * vec
      do  i = 1, nn
         vv(i) = vec(indxdav(i))
      enddo

!$omp parallel if(lompthread.and.nn.ge.50) num_threads(ompthread) private(i,j,k)
!$omp do 
      do k = 1, nn
         ww(k) = vv(1)*davmat(1,k)*&
                 (eigval(k)-E)/( (eigval(k)-E)**2 + d2 )
         do j = 2, nn
            ww(k) = ww(k) + vv(j)*davmat(j,k)*&
                   (eigval(k)-E)/( (eigval(k)-E)**2 + d2 )
         enddo
      enddo
!$omp end do

!$omp do
      do i = 1,nn
         vv(i) = davmat(i,1)*ww(1)
      enddo
!$omp end do

!$omp do
      do i = 1, nn
         do k = 2, nn
            vv(i) = vv(i) + davmat(i,k)*ww(k)
         enddo
      enddo
!$omp end do

! --- Determine Error and divide by Diagonal.
      Error = 0.0_dop
!$omp do reduction(+:Error)
      Do i = 1, PsiDim
         Error  = Error + vec(i)**2
         vec(i) = vec(i)*(RDiagonal(i)-E)/( (RDiagonal(i)-E)**2 + d2 )
      enddo
!$omp end do

! --- overwrite the data when obtained via inverted sub-Hamiltonian.
!$omp do
      do  i = 1, nn
         vec(indxdav(i)) = vv(i)
      enddo
!$omp end do
!$omp end parallel

      return
      end subroutine precondit

! **********************************************************************
!    Apply the complex preconditioner.
! ----------------------------------------------------------------------

      subroutine preconditz(vec,HDiagonal,leigf,reigf,eigval, &
                            indxdav,E,Error,d2,PsiDim,nn)

      implicit none

      integer(long), intent(in)                :: nn,PsiDim
      integer(long), dimension(nn), intent(in) :: indxdav
      integer(long)                            ::    i,j,k
      real(dop), intent(in)                    :: d2
      real(dop), intent(out)                   :: Error
      complex(dop)                             :: E
      complex(dop), dimension(nn,nn), intent(in)   :: leigf,reigf
      complex(dop), dimension(nn), intent(in)      :: eigval
      complex(dop), dimension(nn)                  :: vv,ww
      complex(dop), dimension(PsiDim), intent(in)  :: HDiagonal
      complex(dop), dimension(PsiDim), intent(out) :: vec


! --- Multiply with inverted sub-Hamilton matrix.
! --- vv = reigf * (1/eigval) * leigf^+ * vec
      do  i = 1, nn
         vv(i) = vec(indxdav(i))
      enddo

      do k = 1, nn
         ww(k) = vv(1)*(leigf(k,1))* &
                 Dconjg(eigval(k)-E)/(DConjg(eigval(k)-E)* &
                                         (eigval(k)-E)+d2)
         do j = 2, nn
            ww(k) = ww(k) + vv(j)*(leigf(k,j))* &
                   Dconjg(eigval(k)-E)/(DConjg(eigval(k)-E)* &
                                         (eigval(k)-E)+d2)
         enddo
      enddo

      do i = 1, nn
         vv(i) = reigf(i,1)*ww(1)
      enddo

      do k = 2, nn
         do i = 1, nn
            vv(i) = vv(i) + reigf(i,k)*ww(k)
         enddo
      enddo

! --- Determine Error (norm of |r>) and divide by Diagonal.
      Error = 0.d0
      Do i = 1, PsiDim
         Error  = Error + abs(vec(i))**2
         vec(i) = vec(i)*Dconjg(HDiagonal(i)-E)/(DConjg(HDiagonal(i)-E)* &
                                                 (HDiagonal(i)-E)+d2)
      enddo
! --- overwrite the data when obtained via inverted sub-Hamiltonian.
      do  i = 1, nn
         vec(indxdav(i)) = vv(i)
      enddo

      return
      end subroutine preconditz


! **********************************************************************
!    Apply the preconditioner.
! ----------------------------------------------------------------------

      subroutine precondith(vec,HDiagonal,davmat,eigval,indxdav, &
                           E,Error,d2,PsiDim,nn)

      implicit none

      integer(long), intent(in)                :: nn,PsiDim
      integer(long), dimension(nn), intent(in) :: indxdav
      integer(long)                            ::    i,j,k
      real(dop), intent(in)                    :: d2
      real(dop), intent(out)                   :: Error
      real(dop)                                :: E
      complex(dop), dimension(nn,nn), intent(in)   :: davmat
      complex(dop), dimension(nn), intent(in)      :: eigval
      complex(dop), dimension(nn)                  :: vv,ww
      complex(dop), dimension(PsiDim), intent(in)  :: HDiagonal
      complex(dop), dimension(PsiDim), intent(out) :: vec

! --- Multiply with inverted sub-Hamilton matrix.
! --- vv = eigvec * (1/eigval) * eigvec^+ * vec
      do  i = 1, nn
         vv(i) = vec(indxdav(i))
      enddo

      do k = 1, nn
         ww(k) = vv(1)*Dconjg(davmat(1,k))* &
                 dble(eigval(k)-E)/( dble(eigval(k)-E)**2 + d2 )
         do j = 2, nn
            ww(k) = ww(k) + vv(j)*Dconjg(davmat(j,k))* &
                   dble(eigval(k)-E)/(dble(eigval(k)-E)**2 + d2)
         enddo
      enddo

      do i = 1,nn
         vv(i) = davmat(i,1)*ww(1)
      enddo
      do k = 2, nn
         do i = 1, nn
            vv(i) = vv(i) + davmat(i,k)*ww(k)
         enddo
      enddo

! --- Determine Error and divide by Diagonal.
      Error = 0.d0
      Do i = 1, PsiDim
         Error  = Error + abs(vec(i))**2
         vec(i) = vec(i)*(dble(HDiagonal(i))-E)/ &
                 ((dble(HDiagonal(i))-E)**2 + d2)
      enddo

! --- overwrite the data when obtained via inverted sub-Hamiltonian.
      do  i = 1, nn
         vec(indxdav(i)) = vv(i)
      enddo

      return
      end subroutine precondith

! **********************************************************************

      subroutine cprecondit(vec,RDiagonal,davmat,eigval,indxdav,&
                           vv,ww,E,Error,d2,PsiDim,nn)

!$use omp_lib

use openmpmod

      implicit none

      integer(long)                                  :: i,j,k
      integer(long), intent(in)                      :: nn,PsiDim
      integer(long), dimension(nn), intent(in)       :: indxdav
      real(dop), dimension(nn), intent(in)           :: eigval
      real(dop), dimension(nn), intent(out)          :: vv,ww
      real(dop), dimension(nn,nn), intent(in)        :: davmat
      real(dop), dimension(PsiDim), intent(in)       :: RDiagonal
      complex(dop), dimension(PsiDim), intent(inout) :: vec
      real(dop), intent(in)                          :: E,d2
      real(dop), intent(out)                         :: Error

! --- Multiply with inverted sub-Hamilton matrix.
! --- vv = eigvec * (1/eigval) * eigvec^T * vec

!$omp parallel if(lompthread) num_threads(ompthread) private(i,j,k)
!$omp do
      do  i = 1, nn
         vv(i) = dble(vec(indxdav(i)))
      enddo
!$omp end do

!$omp do
      do k = 1, nn
         ww(k) = vv(1)*davmat(1,k)*&
                 (eigval(k)-E)/( (eigval(k)-E)**2 + d2 )
         do j = 2, nn
            ww(k) = ww(k) + vv(j)*davmat(j,k)*&
                   (eigval(k)-E)/( (eigval(k)-E)**2 + d2 )
         enddo
      enddo
!$omp end do

!$omp do
      do i = 1,nn
         vv(i) = davmat(i,1)*ww(1)
      enddo
!$omp end do

!$omp do 
      do i = 1, nn
         do k = 2, nn
            vv(i) = vv(i) + davmat(i,k)*ww(k)
         enddo
      enddo
!$omp end do

! --- Determine Error and divide by Diagonal.
      Error = 0.0_dop
!$omp do reduction(+:Error)
      Do i = 1, PsiDim
         Error  = Error + dble(vec(i))**2
         vec(i) = vec(i)*(RDiagonal(i)-E)/( (RDiagonal(i)-E)**2 + d2 )
      enddo
!$omp end do

! --- overwrite the data when obtained via inverted sub-Hamiltonian.
!$omp do
      do  i = 1, nn
         vec(indxdav(i)) = vv(i)
      enddo
!$omp end do
!$omp end parallel

      return
      end subroutine cprecondit

! **********************************************************************
!-------------------------------------------------------------------------------
 
      end module davutils 
