! **********************************************************************
!                                                                       
!                   module GETDAVMAT
!                                                                       
!   MCTDH module containing subroutines to evaluate some block of
!   the Hamiltonian matrix. This block is used as preconditioner.
!   (only for real Davidson).
!                                                          
! Contains:     
!   bsort:         Performs one sort sweep.
!   sortdiagonal:  Determines the indices, which belong to the block
!                  of the Hamiltonian which is to be inverted.
!   setm:          Determines the mode indices (j_1..j_f) of the
!                  total index J. 
!   getdavmat:     Evaluates davmat, calls davterm and davtermu.
!   davterm:       Evaluates a single correlated term. 
!   davtermu:      Evaluates a single uncorrelated matrix element.    
!   invdavmat:     Inverts davmat, i.e. the selected Hamiltonian block. 
!
!
!  HDM 07/2004
! **********************************************************************
!
      module getdavmatmod

use decimal, only:long, dop
use global
use globmemmod
use logdat
use maxv, only: maxdim,maxsta
use timing
use channels
use distribmod
use psidef
use griddatmod
use operdef
use runpropmod
use mpidata
use pardsyevmod
use hpsimod
#ifdef MPI
      use mpi
#endif

      implicit none
      private
      public :: sortdiagonal,sortdiagonalz,getdavmat,getdavmatz,&
                getdavmath,setm,bsort,davterm
#ifdef MPI
      public :: mpigetdavmat
#endif
      contains

! **********************************************************************

      subroutine bsort(vmin,indx,nn)

      implicit none

      integer(long)                               :: i,p
      integer(long), intent(in)                   :: nn
      integer(long), dimension(nn), intent(inout) :: indx
      real(dop)                                   :: v
      real(dop), dimension(nn), intent(inout)     :: vmin

      do i = nn-1, 1, -1
         if(vmin(i) .gt. vmin(i+1)) then
            v = vmin(i+1)
            vmin(i+1) = vmin(i)
            vmin(i) = v
            p = indx(i+1)
            indx(i+1) = indx(i)
            indx(i) = p
         endif
      enddo

      return
      end subroutine bsort

!#######################################################################

      subroutine sortdiagonal(rdiagonal,wrk,indxdav,e,psidim,nn)

! --- Determines the indices of the nn smallest abs(RDiagonal-E).
 

      implicit none 

      integer(long)                             :: i
      integer(long), intent(in)                 :: psidim,nn
      integer(long), dimension(nn), intent(out) :: indxdav
      real(dop)                                 :: diag
      real(dop), intent(in)                     :: e
      real(dop), dimension(psidim), intent(in)  :: rdiagonal
      real(dop), dimension(nn), intent(out)     :: wrk

      wrk(1)=abs(rdiagonal(1)-e)
      indxdav(1)=1

      do i = 2, nn
         wrk(i) = abs(rdiagonal(i)-e)
         indxdav(i) = i
         call bsort(wrk,indxdav,i)
      enddo

      do i = nn+1, psidim
         diag = abs(rdiagonal(i)-e)
         if(diag.le.wrk(nn)) then
            wrk(nn) = diag
            indxdav(nn) = i
            call bsort(wrk,indxdav,nn)
         endif
      enddo

      return
      end subroutine sortdiagonal

!#######################################################################

      subroutine sortdiagonalz(HDiagonal,wrk,indxdav,E,PsiDim,nn)

! --- Determines the indices of the nn smallest abs(HDiagonal-E).

      implicit none

      integer(long)                               :: i
      integer(long), intent(in)                   :: PsiDim,nn
      integer(long),dimension(nn), intent(out)    :: indxdav(nn)
      real(dop), dimension(nn), intent(out)       :: wrk(nn)
      real(dop)                                   :: diag
      complex(dop), dimension(psidim), intent(in) :: HDiagonal
      complex(dop), intent(in)                    :: E
      integer(long), save                         :: tid=0 ! timer ID

      call get_timer(tid, 'Sortdiagonalz')
      call start_timer(tid)

      wrk(1)=abs(HDiagonal(1)-E)
      indxdav(1)=1

      do i = 2, nn
         wrk(i) = abs(HDiagonal(i)-E)
         indxdav(i) = i
         call bsort(wrk,indxdav,i)
      enddo

      do i = nn+1, PsiDim
         diag = abs(HDiagonal(i)-E)
         if(diag.le.wrk(nn)) then
            wrk(nn) = diag
            indxdav(nn) = i
            call bsort(wrk,indxdav,nn)
         endif
      enddo

      call stop_timer(tid)

      return
      end subroutine sortdiagonalz

!#######################################################################
! setm is older version of adecoder in psiutil 
!#######################################################################

      subroutine setm(ii,mii)

! --- The array mii contains the mode indices (minus 1) for 
! --- the total A-vector index ii.

      implicit none 

      integer(long)                                 :: m,k,s
      integer(long), intent(in)                     :: ii
      integer(long), dimension(maxdim), intent(out) :: mii

      k = ii - 1  
outer:do s = 1, nstate
         mii(nmode+1) = s 
         if(k.lt.block(s)) exit outer
         k = k - block(s)
      enddo outer
     s = mii(nmode+1)
      do m = nmode, 2, -1
         mii(m) = k/vdim(m,s)
         k = k - mii(m)*vdim(m,s)
      enddo
      mii(1) = k 

      return
      end subroutine setm
  
!#######################################################################

      subroutine getdavmat(rdiagonal,davmat,eigval,indxdav, &
                           psidim,nn,lwork)

! --- Evaluates the block of the full Hamiltonian matrix, which is
! --- used as PreConditioner, and diagonalises this block.  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in getdavmat, please do them in mpigetdavmat
! and dwtgetdavmat too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      implicit none

      integer(long)                            :: i,j,ierr
      integer(long), intent(in)                :: psidim,nn,lwork
      integer(long), dimension(nn), intent(in) :: indxdav
      integer(long), dimension(maxdim)         :: mii, mjj

      real(dop), dimension(psidim), intent(in) :: rdiagonal
      real(dop), dimension(nn), intent(out)    :: eigval
      real(dop), dimension(lwork)              :: workr
      real(dop), dimension(nn,nn), intent(out) :: davmat

#ifdef MPI
      if(lmpi.and.lmpigetdavmat.and.lmpistarted) then
         if(.not.lmpidavrun) then
            mpitask=6
! --- Determine Job
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,&
                           mpierr)
            call MPI_Bcast(nn,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(psidim,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         endif
         call mpigetdavmat(RDiagonal,davmat,indxdav,PsiDim,nn,workr)
      else
#endif

         do i = 2, nn
            call setm(indxdav(i),mii)  
            do j = 1, i-1
               call setm(indxdav(j),mjj)  
               davmat(i,j) = 0.0_dop
               call davterm(mii,mjj,davmat(i,j))
               davmat(j,i) = davmat(i,j)
            enddo
!........Diagonal contributions
            davmat(i,i) = rdiagonal(indxdav(i))
         enddo
#ifdef MPI
      endif
#endif

      davmat(1,1) = rdiagonal(indxdav(1))

! --- Diagonalise davmat. davmat then contains the eigenvectors.
      call pardsyev('V','U',nn,davmat,nn,eigval,workr,lwork,ierr)

      if(ierr .ne. 0 ) then
         routine = 'getdavmat'
         message = 'Diagonalisation Error. (Pre-Conditioner)'
         call errormsg
      endif

      write(irlx,'(/a,i5,a,2e16.8)') ' precon =',nn, &
                     ', First and last eigenvalue [eV]:',  &
                       eigval(1)*ev,eigval(nn)*ev 
!       write(6,'(/a)') ' --  Eigenvalues  eV -- ' 
!       write(6,'(8f11.6)') (EigVal(ierr)*ev,ierr=1,nn)      

      return
      end subroutine getdavmat

!#######################################################################

      subroutine getdavmatz(HDiagonal,davmat,leigf,reigf,eigval,&
                 indxdav,PsiDim,nn,lwork)

! --- Evaluates the block of the full Hamiltonian matrix (complex),
! --- which is used as PreConditioner, and diagonalises this block.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in getdavmatz, please do them in mpigetdavmatz
! and dwtgetdavmatz too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      implicit none

#ifdef PARALLEL
#ifdef MPI
use mpi

      integer(long) :: mpierr,mpitask
#endif
      integer(long) :: fsempost,fsemwait,error
#endif

      integer(long)                            :: i,j,ierr
      integer(long), intent(in)                :: psidim,nn,lwork
      integer(long), dimension(nn), intent(in) :: indxdav
      integer(long), dimension(maxdim)         :: mii, mjj
      real(dop), dimension(2*nn)               :: workr
      complex(dop)                                :: normc
      complex(dop), dimension(psidim), intent(in) :: hdiagonal
      complex(dop), dimension(nn,nn), intent(out) :: davmat
      complex(dop), dimension(nn), intent(out)    :: eigval
      complex(dop), dimension(nn,nn), intent(out) :: leigf
      complex(dop), dimension(nn,nn), intent(out) :: reigf
      complex(dop), dimension(lwork)              :: workc
      integer(long), save                         :: tid=0 ! timer ID

      call get_timer(tid, 'Getdavmatz')
      call start_timer(tid)

#ifdef PARALLEL
      if (lthread.and.lgetdavmat.and.lstarted.and..not.lmpi) then
         distribint(1)=nn
         distribint(2)=psidim
         lbcount=1
         ctstep=1
         task=301
         do semcount=1,min(nthread,nn-1)
            error=fsempost(1)
            if (error .ne. 0) then
               write(6,*) error
               routine='getdavmatz'
               message = 'error increasing semaphore 1'
               call errormsg
            endif
         enddo
         do semcount=1,min(nthread,nn-1)
            error=fsemwait(2)
            if (error .ne. 0) then
               write(6,*) error
               routine='getdavmatz'
               message = 'error decreasing semaphore 2'
               call errormsg
            endif
         enddo
#ifdef MPI
      elseif(lmpi.and.lmpigetdavmat.and.lmpistarted) then
         if(.not.lmpidavrun) then
            mpitask=6
! --- Determine Job
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(nn,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(psidim,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         endif
         call mpigetdavmatz(HDiagonal,davmat,indxdav,PsiDim,nn,mc,workc)
#endif
      else
#endif
         do i = 1, nn
            call setm(indxdav(i),mii)
            do j = 1, i-1
               call setm(indxdav(j),mjj)
               davmat(i,j) = (0.d0,0.d0)
               call davtermz(mii,mjj,davmat(i,j))
            enddo
            do j = i+1, nn
               call setm(indxdav(j),mjj)
               davmat(i,j) = (0.d0,0.d0)
               call davtermz(mii,mjj,davmat(i,j))
            enddo
!........Diagonal contributions
            davmat(i,i) = HDiagonal(indxdav(i))
         enddo
#ifdef PARALLEL
      endif
#endif


! --- Diagonalise davmat. davmat then contains the eigenvectors.
      call ZGEEV('V','V',nn,davmat,nn,eigval,leigf,nn,reigf,nn,workc,&
                lwork,workr,ierr)

      ! Change leigf from U to U^+ and bi-orthonormal
      do i=1,nn
        do j=1,nn
           davmat(j,i)=DConjg(leigf(i,j))
        enddo
      enddo
      leigf=davmat

      do i=1,nn
         normc=SUM(leigf(i,:)*reigf(:,i))
         leigf(i,:)=leigf(i,:)/normc
      enddo


      if(ierr .ne. 0 ) then
         routine = 'getdavmat'
         message = 'Diagonalisation Error. (Pre-Conditioner)'
         call errormsg
      endif

      write(irlx,'(/a,i5,a,2e16.8)') ' precon =',nn, &
                     ', First and last eigenvalue [eV]:', &
                       DBLE(eigval(1))*ev,DBLE(eigval(nn))*ev
!       write(6,'(/a)') ' --  Eigenvalues  eV -- '
!       write(6,'(8f11.6)') (EigVal(ierr)*ev,ierr=1,nn)

      call stop_timer(tid)

      return
      end subroutine getdavmatz


!#######################################################################

      subroutine getdavmath(HDiagonal,davmat,eigval,indxdav,PsiDim,nn,lwork)


! --- Evaluates the block of the full Hamiltonian matrix (complex),
! --- which is used as PreConditioner, and diagonalises this block.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in getdavmath, please do them in mpigetdavmath
! and dwtgetdavmath too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      implicit none

#ifdef PARALLEL
#ifdef MPI
use mpi

      integer(long) :: mpierr,mpitask
#endif
      integer(long) :: fsempost,fsemwait,error
#endif
      integer(long)                            :: i,j,ierr
      integer(long), intent(in)                :: psidim,nn,lwork
      integer(long), dimension(nn), intent(in) :: indxdav
      integer(long), dimension(maxdim)         :: mii, mjj
      real(dop), dimension(4*nn-2)               :: workr
      complex(dop), dimension(psidim), intent(in) :: hdiagonal
      complex(dop), dimension(nn,nn), intent(out) :: davmat
      complex(dop), dimension(nn), intent(out)    :: eigval
      complex(dop), dimension(lwork)              :: workc
      integer(long), save                         :: tid=0 ! timer ID

      call get_timer(tid, 'Getdavmath')
      call start_timer(tid)

#ifdef PARALLEL
      if (lthread.and.lgetdavmat.and.lstarted.and..not.lmpi) then
         distribint(1)=nn
         distribint(2)=psidim
         lbcount=1
         ctstep=1
         task=302
         do semcount=1,min(nthread,nn-1)
            error=fsempost(1)
            if (error .ne. 0) then
               write(6,*) error
               routine='getdavmath'
               message = 'error increasing semaphore 1'
               call errormsg
            endif
         enddo
         do semcount=1,min(nthread,nn-1)
            error=fsemwait(2)
            if (error .ne. 0) then
               write(6,*) error
               routine='getdavmatz'
               message = 'error decreasing semaphore 2'
               call errormsg
            endif
         enddo
#ifdef MPI
      elseif(lmpi.and.lmpigetdavmat.and.lmpistarted) then
         if(.not.lmpidavrun) then
            mpitask=6
! --- Determine Job
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(nn,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(psidim,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         endif
         call mpigetdavmatz(HDiagonal,davmat,indxdav,PsiDim,nn,mc,workc)
#endif
      else
#endif
         do i = 1, nn
            call setm(indxdav(i),mii)
            do j = 1, i-1
               call setm(indxdav(j),mjj)
               davmat(i,j) = (0.d0,0.d0)
               call davtermz(mii,mjj,davmat(i,j))
               davmat(j,i)=DConjg(davmat(i,j))
            enddo
!........Diagonal contributions
            davmat(i,i) = HDiagonal(indxdav(i))
         enddo
#ifdef PARALLEL
      endif
#endif

! --- Diagonalise davmat. davmat then contains the eigenvectors.
      call ZHEEV('V','U',nn,davmat,nn,workr(1),workc,lwork,workr(nn+1),ierr)

      ! copy eigenvalues to a complex array (later part of davmat).
      do i=1,nn
         eigval(i)=workr(i)
      enddo

      if(ierr .ne. 0 ) then
         routine = 'getdavmat'
         message = 'Diagonalisation Error. (Pre-Conditioner)'
         call errormsg
      endif

      write(irlx,'(/a,i5,a,2e16.8)') ' precon =',nn, &
                     ', First and last eigenvalue [eV]:', &
                       DBLE(eigval(1))*ev,DBLE(eigval(nn))*ev

      call stop_timer(tid)

      return
      end subroutine getdavmath

!#######################################################################

#ifdef PARALLEL
      subroutine dwtgetdavmat(nn,PsiDim,davmat,indxdav,mc,RDiagonal,mii,mjj)

      implicit none

      integer(long) :: i,j,mutexerror
      integer(long) :: fpthreadmutexlocklb,fpthreadmutexunlocklb
      integer(long), intent(in)                     :: PsiDim,nn
      integer(long), dimension(nn), intent(in)      :: indxdav
      integer(long), dimension(maxdim), intent(out) :: mii, mjj
      real(dop), dimension(psidim), intent(in)      :: RDiagonal
      real(dop), dimension(nn,nn), intent(out)      :: davmat

 1    continue
      mutexerror=fpthreadmutexlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error locking lbmutex'
         call errormsg
      endif

      i=lbcount
      lbcount=lbcount+ctstep

      mutexerror=fpthreadmutexunlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error unlocking lbmutex'
         call errormsg
      endif

      if (i.le.nn) then
         call setm(indxdav(i),mii)
         do j = 1, i-1
            call setm(indxdav(j),mjj)
            davmat(i,j) = 0.d0
            call davterm(mii,mjj,davmat(i,j))
            davmat(j,i) = davmat(i,j)
         enddo
         davmat(i,i) = RDiagonal(indxdav(i))
      endif

      if (i.lt.nn) goto 1

      return
      end subroutine dwtgetdavmat
#endif

!#######################################################################
#ifdef PARALLEL
      subroutine dwtgetdavmatz(nn,PsiDim,davmat,indxdav,mc,HDiagonal,mii,mjj)

      implicit none

      integer(long) :: i,j,mutexerror
      integer(long) :: fpthreadmutexlocklb,fpthreadmutexunlocklb
      integer(long), intent(in)                     :: PsiDim,nn
      integer(long), dimension(nn), intent(in)      :: indxdav
      integer(long), dimension(maxdim), intent(out) :: mii, mjj
      complex(dop), dimension(psidim), intent(in)   :: HDiagonal
      complex(dop), dimension(nn,nn), intent(out)   :: davmat

 9    continue
      mutexerror=fpthreadmutexlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error locking lbmutex'
         call errormsg
      endif

         i=lbcount
         lbcount=lbcount+ctstep

      mutexerror=fpthreadmutexunlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error unlocking lbmutex'
         call errormsg
      endif

      if (i.le.nn) then
         call setm(indxdav(i),mii)
         do j = 1, i-1
            call setm(indxdav(j),mjj)
            davmat(i,j) = (0.d0,0.d0)
            call davtermz(mii,mjj,davmat(i,j))
         enddo
         do j = i+1, nn
            call setm(indxdav(j),mjj)
            davmat(i,j) = (0.d0,0.d0)
            call davtermz(mii,mjj,davmat(i,j))
         enddo
         davmat(i,i) = HDiagonal(indxdav(i))
      endif

      if (i.lt.nn) goto 9


      return
      end subroutine dwtgetdavmatz
#endif

!#######################################################################
#ifdef PARALLEL
      subroutine dwtgetdavmath(nn,PsiDim,davmat,indxdav,mc,HDiagonal,mii,mjj)

      implicit none

      integer(long) :: i,j,mutexerror
      integer(long) :: fpthreadmutexlocklb,fpthreadmutexunlocklb
      integer(long), intent(in)                     :: PsiDim,nn
      integer(long), dimension(nn), intent(in)      :: indxdav
      integer(long), dimension(maxdim), intent(out) :: mii, mjj
      complex(dop), dimension(psidim), intent(in)   :: HDiagonal
      complex(dop), dimension(nn,nn), intent(out)   :: davmat


 8    continue
      mutexerror=fpthreadmutexlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error locking lbmutex'
         call errormsg
      endif

         i=lbcount
         lbcount=lbcount+ctstep

      mutexerror=fpthreadmutexunlocklb()
      if (mutexerror .ne. 0) then
         write(6,*) mutexerror
         routine='dlasr'
         message = 'error unlocking lbmutex'
         call errormsg
      endif

      if (i.le.nn) then
         call setm(indxdav(i),mii)
         do j = 1, i-1
            call setm(indxdav(j),mjj)
            davmat(i,j) = (0.d0,0.d0)
            call davtermz(mii,mjj,davmat(i,j))
            davmat(j,i) = DConjg(davmat(i,j))
         enddo
         davmat(i,i) = HDiagonal(indxdav(i))
      endif

      if (i.lt.nn) goto 8


      return
      end subroutine dwtgetdavmath
#endif

!#######################################################################

      subroutine davterm(mii,mjj,davmat)
! --- Determine correlated terms

use global
use maxv, only: maxdim
use griddatmod
use operdef
use psidef
use hpsimod

      implicit none

      integer(long)                                :: k,mm,m,s,p
      integer(long), dimension(maxdim), intent(in) :: mii,mjj
      real(dop), intent(inout)                     :: davmat
      complex(dop)                                 :: term

! --- Correlated contributions.
outer:do k = zham(1), zham(1)+kzahl(1)-1
         if(kf(k) .ne. mii(nmode+1) .or. &
            ki(k) .ne. mjj(nmode+1) ) then
            cycle outer
         endif 
 
         term = koeff(k)
         do m = 1, nmode
            if ( .not.diag(m,k) ) then
               p = pmat(m,k) + mii(m) + mjj(m)*dim(m,kf(k))
               term = term*hteil(p)
            elseif( mii(m).ne.mjj(m) ) then
               cycle outer
            endif
         enddo
         davmat = davmat + dble(term)
      enddo outer

! --- Uncorrelated contributions.
      k=zham(1)+kzahl(1)-1
      do s = 1, nstate
inner:   do m = 1, nmode
            if( khzahl(m,s,1) .eq.0 ) cycle inner
            do mm = 1, nmode
               if( m.ne.mm ) then
                  if(mii(mm).ne.mjj(mm)) cycle inner
               endif
            enddo 
            p = pmat(m,k+s) + mii(m) + mjj(m)*dim(m,s)
            davmat = davmat + dble(hteil(p))
         enddo inner
      enddo 


      return
      end subroutine davterm

! **********************************************************************

      subroutine davtermz(mii,mjj,davmat)
! --- Determine correlated terms

      implicit none

      integer(long)                                :: k,mm,m,s,p
      integer(long), dimension(maxdim), intent(in) :: mii,mjj
      complex(dop), intent(out)                    :: davmat
      complex(dop)                                 :: term

! --- Correlated contributions.
      do k = zham(1), zham(1)+kzahl(1)-1
         if(kf(k) .ne. mii(nmode+1) .or. ki(k) .ne. mjj(nmode+1) ) then
            goto 60
         endif

         term = koeff(k)
         do m = 1, nmode
            if ( .not.diag(m,k) ) then
               p = pmat(m,k) + mii(m) + mjj(m)*dim(m,kf(k))
               term = term*hteil(p)
            elseif( mii(m).ne.mjj(m) ) then
               goto 60
            endif
         enddo
         davmat = davmat + term
 60      continue
      enddo
! --- Uncorrelated contributions.
      k=zham(1)+kzahl(1)-1
      do s = 1, nstate
         do m = 1, nmode
            if( khzahl(m,s,1) .eq.0 ) goto 70
            do mm = 1, nmode
               if( m.ne.mm ) then
                  if(mii(mm).ne.mjj(mm)) goto 70
               endif
            enddo
            p = pmat(m,k+s) + mii(m) + mjj(m)*dim(m,s)
            davmat = davmat + hteil(p)
  70        continue
         enddo
      enddo

      return
      end subroutine davtermz

! **********************************************************************
!-----------------------------------------------------------------------
!   mpigetdavmat
!-----------------------------------------------------------------------

#ifdef MPI
      subroutine mpigetdavmat(RDiagonal,davmat,indxdav,PsiDim,nn,workr)

use mpi
use maxv
use global
use globmemmod
use timing
use channels
use hpsimod
use mpidata

use op1lib, only: zerovxd

      implicit none

      integer(long), intent(in) :: PsiDim,nn
      integer(long), dimension(nn), intent(in) :: indxdav
      real(dop), dimension(psidim), intent(in) :: RDiagonal
      real(dop), dimension(nn,nn), intent(out) :: davmat
      real(dop), dimension(*), intent(out) :: workr

      integer(long) :: i,j,mii(maxdim),mjj(maxdim)

! --- Get Data
! ... indxdav
      call MPI_Bcast(indxdav,nn,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
! ... Hteil-matrices
      call MPI_Bcast(rdiagonal,psidim,MPI_DOUBLE_PRECISION,0, &
                                                  MPI_COMM_WORLD,mpierr)

      call zerovxd(davmat,nn*nn)

      do i = 2+mpirank, nn, nmpi
         call setm(indxdav(i),mii)
         do j = 1, i-1
            call setm(indxdav(j),mjj)
            davmat(i,j) = 0.d0
            call davterm(mii,mjj,davmat(i,j))
            davmat(j,i) = davmat(i,j)
         enddo
!........Diagonal contributions
         davmat(i,i) = RDiagonal(indxdav(i))
      enddo

! --- Reduce Data
      call mpigetdavmatreduce(davmat,nn,workr)

      return
      end subroutine mpigetdavmat
#endif

!-----------------------------------------------------------------------
!   mpigetdavmatreduce
!-----------------------------------------------------------------------

#ifdef MPI
      subroutine mpigetdavmatreduce(davmat,nn,workr)

use mpi
use global
use timing
use channels
use hpsimod
use mpidata

use op1lib, only: cpvxd

      implicit none

      integer(long), intent(in) :: nn
      real(dop), dimension(nn*nn), intent(out) :: davmat
      real(dop), dimension(*), intent(out) :: workr

      integer(long) :: i,j

! --- Reduce Data
      j=-1
      do i=0,nn*nn/workrdim-1
         j=i
         call MPI_Reduce(davmat(1+i*workrdim),workr,workrdim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         call cpvxd(workr,davmat(1+i*workrdim),workrdim)
      enddo
      j=j+1
      call MPI_Reduce(davmat(1+j*workrdim),workr,nn*nn-j*workrdim, &
                   MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
      call cpvxd(workr,davmat(1+j*workrdim),nn*nn-j*workrdim)

      return
      end subroutine mpigetdavmatreduce
#endif

      end module getdavmatmod

