! **********************************************************************
!                                                                       
!                   module GETDIAGONAL
!                                                                       
!   MCTDH module containing subroutines to evaluate the diagonal of
!   the Hamiltonian matrix. Needed for Davidson. Similar to calcha.
!   This module contains also the overlap routines needed for "lock".
!                                                          
!   Contains:                                                           
!     getdiagonal:  Calculates the diagonal part of the correlated           
!                   Hamiltonian. 
!     getdiag1k:    Calculates the diagonal part of the correlated
!                   Hamiltonian for one coefficient of the inter-state
!                   terms.
!     getdiaguna:   Calculates the diagonal part of the uncorrelated
!                   Hamiltonian.
!     wkgetdiag:    Sets minimal length of work arrays
!
!     davovlp1:     Computes the overlap matrices between the two sets
!                   of SPFs. (cross-SPFs and actual SPFs). 
!     davovlp2r:    Computes the overlap for RDav (real version).
!     davovlp2c:    Computes the overlap for Dav nad CDav (complex).
!     bgetdiagonal: Similar to getdiagonal, but for block-davidson. 
!
!  HDM 06/2003
! **********************************************************************
      module getdiagonalmod

use decimal, only: dop, long
use global
use globmemmod
use logdat
use mmlib,   only: mmaxzz
use mtlib,   only: mtxxzz, addtxxzz, dtxxzz
use xvlib,   only: vvaxzz, xvxxzza
use op1lib,  only: cpvxz, initvxz, zerovxz, normvxz
use op2lib,  only: addvxxzo

      implicit none
      private
      public :: getdiagonal,davovlp1,davovlp2r,davovlp2c,wkgetdiag,&
                bgetdiagonal,kry2psi,psi2kry,kryc2finalpsi,kryc2psi,psi2kryc,&
                psi2kryr,kry2psir,kry2finalpsi,postmpibgetdiag
      contains

!#######################################################################
! *** V8.3 HDM *********> SUBROUTINE  GETDIAGONAL   <*******************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the diagonal of the Hamiltonian for the intra-state terms. 
! Result is returned as a sum over all terms in hdiagonal.
!------------------------------------------------------------------------

      subroutine getdiagonal(hdiagonal)

#ifdef MPI
use mpi
use mpidata
#endif

use dvrdatmod
use psidef
use griddatmod
use operdef
use hpsimod

      implicit none
      integer(long)                               :: k,s,s1,k0,koffset,h, &
                                                     k1,kstep
      complex(dop), dimension(adim), intent(out) :: hdiagonal

! --- CHECK FOR UNALLOWED CONDITIONS ---
!     Only in rank 0
#ifdef MPI
      if (mpirank.eq.0) then
#endif
      if(nmulpot(1).gt.0) then
         routine='getdiagonal'
         message= &
         'No muld-terms allowed for Davidson (improved relaxation).'
         call errormsg
      endif
#ifdef MPI
      endif
#endif

      ! default
      kstep = 1
      k1 = 0

#ifdef MPI
      if (lmpi .and. lmpigetdiag .and. lmpistarted) then
         kstep = nmpi
         k1 = mpirank
      else
         if(mpirank.ne.0) then
            ! do nothing if no-getdiag is set
            return
         end if
      end if

      !uncorrelated terms only in mpi-rank 0
      if (mpirank .eq. 0) then
#endif

! --- DO THE UNCORRELATED TERMS FIRST ---

      call getdiaguna (hdiagonal,hteil,mc(mcworkc),mi(mijindx))  

#ifdef MPI
      else
         ! other MPI ranks initialize with zero
         call zerovxz(hdiagonal,adim)
      end if
#endif

      koffset=zham(1)+kzahl(1)-1
      k0=zham(1)

! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---
! **GR** OPENMP loop? (Threads used in 84.358)

      do k=k0,koffset
         s = kf(k)
         s1= ki(k)
         if (s.eq.s1) then
            call getdiag1k(hteil,mc(mcworkc),k,s,mi(mijindx))

!    --- STORE RIGHT HAND SIDE IN hdiagonal ---

            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then
               call addvxxzo(mc(mcworkc),hdiagonal(zpsi(s)),block(s))
            else
               call xvxxzza(koeff(k),mc(mcworkc),hdiagonal(zpsi(s)),&
                     block(s))
            endif
         endif
      enddo

#ifdef MPI
      ! if MPI is used collect the data from the other processes
      if (lmpi .and. lmpigetdiag .and. lmpistarted) then
         call MPI_REDUCE(hdiagonal,mc(mcworkc),adim,  &
              MPI_DOUBLE_COMPLEX, MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         if (mpirank.eq.0)  then
            call cpvxz(mc(mcworkc),hdiagonal,adim)
         end if
      end if
#endif

      return
      end subroutine getdiagonal

! *** V6.0 GW ************> SUBROUTINE GETDIAG1K <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k.
! Result is returned in workc(1:block(s)).
! See also CALCHA1K 


      subroutine getdiag1k (hteil,workc,k,s,jindx)

use psidef
use griddatmod
use operdef
use hpsimod

      implicit none

      integer(long)                                   :: zeig1,zeig2,zeig3,zeig5,m,i,vdm,ndm
      integer(long), intent(in)                       :: k,s
      integer(long), dimension(adim), intent(in)      :: jindx
      complex(dop), dimension(pmatdim), intent(in)    :: hteil
      complex(dop), dimension(3*fmblock), intent(out) :: workc

! --- COPY PSI VECTOR AND SET UP POINTERS ---
      zeig5 = 2*fmblock+1
      zeig2 = fmblock+1
      zeig1 = 1
      call initvxz(workc(zeig2),fmblock)
      if(citype .gt. 0 ) goto 200

! --- LOOP OVER EACH MODE ---

      do m = 1,nmode
         if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS --- 

            call getdiag(hteil(pmat(m,k)),workc(zeig5),dim(m,s))
            call dtxxzz(workc(zeig5),workc(zeig1),&
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
        endif
      enddo

! --- IF RESULT IS NOT STORED AT START OR WORK ARRAY, NOW MOVE ---
      
      if (zeig2.ne.1) &
         call cpvxz(workc(zeig2),workc,block(s))

      return

!  -------      S-MCTDH  -----------------------
  200 continue   
! ------ compute the "long" vdim and ndim (s-mctdh).
      vdm=1
      ndm=dim(1,s)
      do m=2,nmode
        ndm=ndm*dim(m,s)
      enddo
      do m = 1,nmode
         ndm=ndm/dim(m,s)
         if (.not. diag(m,k)) then
            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3
            call getdiag(hteil(pmat(m,k)),workc(zeig5),dim(m,s))
            call dtxxzz(workc(zeig5),workc(zeig1), &
                 workc(zeig2),vdm,dim(m,s),ndm)
        endif
        vdm=vdm*dim(m,s)
      enddo
! --- Store the diagonal in s-mctdh reduced form 
      do i=0,block(s)-1
         workc(zpsi(s)+i)=workc(jindx(zpsi(s)+i)+zeig2-1)
      enddo
      
      return
      end subroutine getdiag1k

!#######################################################################

      subroutine getdiag(a,b,dim)

      implicit none

      integer(long)                                :: i
      integer(long), intent(in)                    :: dim
      complex(dop), dimension(dim),  intent(out)   :: b
      complex(dop), dimension(dim,dim), intent(in) :: a

      do i=1,dim
         b(i)=a(i,i)
      enddo

      return
      end subroutine getdiag

! *** V8.3 HDM ***********> SUBROUTINE GET DIAG UNA <***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~~~

! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients.

      subroutine getdiaguna (hdiagonal,hteil,workc,jindx)

use psidef
use griddatmod
use operdef
use hpsimod

      implicit none

      integer(long)                                        :: m,s,k,zeig,i,vdm,ndm
      integer(long), dimension(adim), intent(in)           :: jindx
      complex(dop), dimension(adim), intent(out)           :: hdiagonal
      complex(dop), dimension(pmatdim), intent(in)         :: hteil
      complex(dop), dimension(maxspf+fmblock), intent(out) :: workc

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      call zerovxz(hdiagonal,adim)

      k=zham(1)+kzahl(1)-1

      if (citype .eq. 0 ) then              ! normal case
         do s=1,nstate
            do m=1,nmode
               if( khzahl(m,s,1) .gt. 0 ) then 
                  call getdiag(hteil(pmat(m,k+s)),workc,dim(m,s))
                  call addtxxzz(workc,hdiagonal(zpsi(s)),&
                                vdim(m,s),dim(m,s),ndim(m,s))
               endif
            enddo
         enddo

      else                                  ! s-mctdh

         zeig=fmblock+1
         do s=1,nstate
! ------ compute the "long" vdim and ndim (s-mctdh).
            vdm=1
            ndm=dim(1,s)
            do m=2,nmode
               ndm=ndm*dim(m,s)
            enddo        
            call zerovxz(workc,maxspf+fmblock)
            do m=1,nmode
               ndm=ndm/dim(m,s)
               if( khzahl(m,s,1) .gt. 0 ) then 
                  call getdiag(hteil(pmat(m,k+s)),workc(zeig),dim(m,s))
                  call addtxxzz(workc(zeig),workc,vdm,dim(m,s),ndm)
               endif
               vdm=vdm*dim(m,s)
            enddo
! --- Store the diagonal in s-mctdh reduced form 
            do i=0,block(s)-1
               hdiagonal(zpsi(s)+i)=workc(jindx(zpsi(s)+i))
            enddo
         enddo
      endif   

      return
      end subroutine getdiaguna

!#######################################################################
! *** MS 11/2013 ***>  SUBROUTINE  POSTBGETDIAGONAL   <*****************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
!
! post an MPI task for the MPI slaves. This is called from propagatepsi
! when autoblock is set.
!
! NOTE: This routine is not called from the MPI-BDAV routines because
! the MPI slaves are not in the waiting loop (mpiprop) but are
! themselbes already in the respective subroutines. Hence bgetdiagonal
! is called from there directly.
!-----------------------------------------------------------------------

      subroutine postmpibgetdiag

#ifdef MPI
use mpi
use mpidata

      implicit none

      if (lmpi .and. lmpigetdiag .and. lmpistarted .and. mpirank .eq. 0) then
         mpitask=42
         call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      end if
#endif
      end subroutine postmpibgetdiag

!#######################################################################
! *** HDM 09/2007 *****> SUBROUTINE  BGETDIAGONAL   <*******************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the diagonal of the Hamiltonian for Block-Davidson.
! Result is returned as a sum over all terms in hdiagonal.
! Almost identical to GETDIAGONAL.
!------------------------------------------------------------------------

      subroutine bgetdiagonal(hdiagonal)

#ifdef MPI
use mpi
use mpidata
#endif

use dvrdatmod
use psidef
use griddatmod
use operdef
use hpsimod

      implicit none
      integer(long)                               :: k,s,s1,k0,koffset,h,d2,&
                                                     k1,kstep
      complex(dop), dimension(adim), intent(out) :: hdiagonal


! --- CHECK FOR UNALLOWED CONDITIONS ---
!     but only in rank 0

#ifdef MPI
      if (mpirank.eq.0) then
#endif
      if(nmode .ne. mpb) then
         write(6,'(a)') 'bgetdiagonal: packet mode is not last'
         write(6,'(a,2i5)') 'nmode, mpb', nmode, mpb
      endif
      if(nmulpot(1).gt.0) then
         routine='bgetdiagonal'
         message= 'No muld-terms allowed for Davidson (improved relaxation).'
         call errormsg
      endif
#ifdef MPI
      endif
#endif

! --- DO THE UNCORRELATED TERMS FIRST ---
      kstep = 1
      k1 = 0

! if MPI is used, only every nmpi'th k term is evaluated
! in each process. if MPI evaluation is not wanted
! ('no-getdiag') but MPI is used, only execute in rank 0
! otherwise do nothing and return
#ifdef MPI
      if (lmpi .and. lmpigetdiag .and. lmpistarted) then
         kstep = nmpi
         k1 = mpirank
      else
         if(mpirank.ne.0) then
            ! do nothing if no-bgetdiag is set
            return
         end if
      end if

      !unkollelated terms only in mpi-rank 0
      if (mpirank .eq. 0) then
#endif
         call bgetdiaguna (hdiagonal,hteil,mc(mcworkc))

#ifdef MPI
      else
         ! other MPI ranks initialize with zero
         call zerovxz(hdiagonal,adim/npackts)
      end if
#endif


      koffset=zham(1)+kzahl(1)-1
      k0=zham(1)
      s=1

! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

! **GR** OPENMP loop? (Threads used in 84.358)
      do k=k0,koffset
         s = kf(k)
         s1= ki(k)
         if (s.eq.s1) then
            call bgetdiag1k(hteil,mc(mcworkc),k,s)

!    --- STORE RIGHT HAND SIDE IN hdiagonal ---

            d2 = 1 + (zpsi(s)-1)/npackts
            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then
               call addvxxzo(mc(mcworkc),hdiagonal(d2),block(s)/npackts)
            else
               call xvxxzza(koeff(k),mc(mcworkc),hdiagonal(d2), &
                            block(s)/npackts)
            endif
         endif
      enddo

#ifdef MPI
      ! if MPI is used collect the data from the other processes
      if (lmpi .and. lmpigetdiag .and. lmpistarted) then
         call MPI_REDUCE(hdiagonal,mc(mcworkc),adim/npackts, &
              MPI_DOUBLE_COMPLEX, MPI_SUM,0,MPI_COMM_WORLD,mpierr)
         if (mpirank.eq.0)  then
            call cpvxz(mc(mcworkc),hdiagonal,adim/npackts)
         end if
      end if
#endif

      return
      end subroutine bgetdiagonal


! *** 09/2007 HDM *******> SUBROUTINE BGETDIAG1K <*********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k.
! Result is returned in workc(1:block(s)).
! See also CALCHA1K
! Almost identical to GETDIAG1K, only some variables reduced by npackts

      subroutine bgetdiag1k (hteil,workc,k,s)

use psidef
use griddatmod
use operdef
use hpsimod

      implicit none

      integer(long) :: zeig1,zeig2,zeig3,zeig5,m
      integer(long), intent(in)                       :: k,s
      complex(dop), dimension(pmatdim), intent(in)    :: hteil
      complex(dop), dimension(3*maxblock), intent(out) :: workc

! --- SET UP POINTERS AND INITIALISE workc WITH 1 ---
      zeig5 = 2*maxblock/npackts + 1
      zeig2 =   maxblock/npackts + 1
      zeig1 = 1
      call initvxz(workc(zeig2),block(s)/npackts)

! --- LOOP OVER EACH MODE ---

      do m = 1,nmode-1 ! last mode is packet mode.
         if (.not. diag(m,k)) then
         
!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

            call getdiag(hteil(pmat(m,k)),workc(zeig5),dim(m,s))
            call dtxxzz(workc(zeig5),workc(zeig1),&
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s)/npackts)
         endif
      enddo

! --- IF RESULT IS NOT STORED AT START OR WORK ARRAY, NOW MOVE ---

      if (zeig2.ne.1) call cpvxz(workc(zeig2),workc,block(s)/npackts)

      return
      end subroutine bgetdiag1k 



! *** 09/2007 HDM *******> SUBROUTINE B GET DIAG UNA <******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~~

! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients.

      subroutine bgetdiaguna (hdiagonal,hteil,workc)

use psidef
use griddatmod
use operdef
use hpsimod

      implicit none

      integer(long) :: m,s,k
      complex(dop), dimension(adim/npackts), intent(out) :: hdiagonal
      complex(dop), dimension(pmatdim), intent(in) :: hteil
      complex(dop), dimension(maxblock), intent(out) :: workc

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      call zerovxz(hdiagonal,adim/npackts)

      k=zham(1)+kzahl(1)-1

      do s=1,nstate
         do m=1,nmode-1   ! last mode is packet mode
            if( khzahl(m,s,1) .gt. 0 ) then
               call getdiag(hteil(pmat(m,k+s)),workc,dim(m,s))
               call addtxxzz(workc,hdiagonal(1+(zpsi(s)-1)/npackts),&
                         vdim(m,s),dim(m,s),ndim(m,s)/npackts)
            endif
         enddo
      enddo

      return
      end subroutine bgetdiaguna 

!#######################################################################
!
!               davoverlap
!
!#######################################################################

      subroutine davovlp1(psi,workc,time)

use maxv, only: maxdim
use psidef
use griddatmod
use crocoflymod
use channels

      implicit none

      integer(long)                                 :: s,m,zeig1,zeig2,swapzeig
      real(dop)                                     :: norm
      real(dop), intent(in)                         :: time
      real(dop), dimension(maxdim)                  :: orbcrs
      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(2*adim), intent(out)  :: workc
      character(len=64)                             :: str

!-----------------------------------------------------------------------
!   calculate density matrices for each degree of freedom defined as
!   overlap of the single-particle functions rho^m_ij=<phi_i|phi_j,ref>
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call mmaxzz(psi(zetf(m,s)),mc(mcpsiref+zetf(m,s)-1), &
              mc(mcccdover+dovmat(m,s)-1),subdim(m),dim(m,s),dim(m,s))
            call normvxz(mc(mcccdover+dovmat(m,s)-1), &
                         norm,dim(m,s)*dim(m,s))
            orbcrs(m) = dim(m,s) - norm**2
         enddo
         write(str,'(8f8.4)') (orbcrs(m),m=1,min(nmode,8))
         write(irlx,'(f9.3,i3,a64,a3)')  &
               time/fs, s, str,' ""'
         if(nmode.gt.8)  then
            write(str,'(8f8.4)') (orbcrs(m),m=9,min(nmode,16))
            write(irlx,'(12x,a64,a3)') str,' ""'
         endif 
         if(nmode.gt.16) then
            write(str,'(8f8.4)') (orbcrs(m),m=17,min(nmode,24))
          write(irlx,'(12x,a64,a3)') str,' ""'
         endif
      enddo
      call myflush(irlx)

      zeig1=0
      zeig2=0
!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J|Phi_L,ref>A_L,ref within each state
!-----------------------------------------------------------------------
      call cpvxz(mc(mcpsiref),workc,adim)
      do s=1,nstate
         zeig2=zpsi(s)
         zeig1=zeig2+adim

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(mc(mcccdover+dovmat(m,s)-1),workc(zeig1), &
                        workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s), &
                        dim(m,s))
         enddo
      enddo

!-----Move data to second part of workc (=Hdiagonal=A_ref), if necessary
      if(zeig2.lt.zeig1) call cpvxz(workc,workc(adim+1),adim)
      
      return
      end subroutine davovlp1

!-----------------------------------------------------------------------

      subroutine davovlp2r(intorder,order,ovlps,krylov,&
                           eigenvector,a_ref,firstcall)

use psidef

      implicit none

      integer(long)                                           :: i,j
      integer(long), intent(in)                               :: order,intorder
      real(dop), dimension(0:2), save                         :: kovl
      real(dop), dimension(adim,0:intorder), intent(in)       :: krylov
      real(dop), dimension(0:intorder), intent(out)           :: ovlps
      real(dop), dimension(0:intorder,0:intorder), intent(in) :: eigenvector
      complex(dop), dimension(adim), intent(in)               :: a_ref
      logical(kind=4), save                                   :: set
      logical(kind=4), intent(in)                             :: firstcall


      if(firstcall) then
         do j = 0, order
            ovlps(j) = eigenvector(0,j)**2
         enddo
         return
      endif

!-----------------------------------------------------------------------
! calculate Sum_J A_J^* [ Sum_L <Phi_J|Phi_L,ref>A_L,ref ]
! calculate Sum_J Krylov_J^* Aref_J 
! A_ref contains the transformed A-vector of psi_ref. 
! A_ref =second part of Hdiagonal.
!-----------------------------------------------------------------------
      if( order .le. 2 ) set = .true.
      if( set ) then  
         kovl(2) = 0.0_dop
         do j = 0, min(2,order)
            kovl(j) = krylov(1,j)*dble(a_ref(1))
            do i = 2, adim
               kovl(j) = kovl(j) + krylov(i,j)*dble(a_ref(i))
            enddo
         enddo
      endif
      if( order .ge. 2 ) set = .false.

!-----------------------------------------------------------------------
! Transform the overlaps <psi_ref|Krylov> (stored in Kovl) 
! with the eigen-vector matrix. 
!-----------------------------------------------------------------------
      do j = 0, order
         ovlps(j) = 0.0_dop
         do i = 0, 2
            ovlps(j) = ovlps(j) + kovl(i)*eigenvector(i,j)
         enddo
         ovlps(j) = ovlps(j)**2
      enddo 

      return
      end subroutine davovlp2r

!-----------------------------------------------------------------------

      subroutine davovlp2c(intorder,order,ovlps,psi,krylov, &
                           eigenvector,a_ref,firstcall)

use psidef

      implicit none

      integer(long)                                              :: i,j
      integer(long), intent(in)                                  :: order,intorder
      real(dop), dimension(0:order), intent(out)                 :: ovlps
      complex(dop)                                               :: zdum
      complex(dop), dimension(adim), intent(in)                  :: a_ref,psi
      complex(dop), dimension(adim,intorder), intent(in)         :: krylov
      complex(dop), dimension(0:2), save                         :: kovl
      complex(dop), dimension(0:intorder,0:intorder), intent(in) :: eigenvector
      logical(kind=4), save                                      :: set
      logical(kind=4), intent(in)                                :: firstcall


      if(firstcall) then
         do j = 0, order
            ovlps(j) = Abs(eigenvector(0,j))**2
         enddo
         return
      endif


!-----------------------------------------------------------------------
! calculate Sum_J A_J^* [ Sum_L <Phi_J|Phi_L,ref>A_L,ref ]
! calculate Sum_J Krylov_J^* Aref_J
! A_ref contains the transformed A-vector of psi_ref.
! A_ref =second part of Hdiagonal.
!-----------------------------------------------------------------------
      if( order .le. 2 ) set = .true.
      if( set ) then
         call vvaxzz(psi,a_ref,zdum,adim)
         kovl(0) = zdum
         kovl(1) = 0.0_dop
         kovl(2) = 0.0_dop
         do j = 1, min(2,order)
            call vvaxzz(krylov(1,j),a_ref,zdum,adim)
            kovl(j) = zdum
         enddo
      endif
      if( order .ge. 2 ) set = .false.

!-----------------------------------------------------------------------
! Transform the overlaps <psi_ref|Krylov> (stored in Kovl) 
! with the eigen-vector matrix. 
!-----------------------------------------------------------------------
      do j = 0, order
         zdum = (0.0_dop,0.0_dop)
         do i = 0, 2
            zdum = zdum + kovl(i)*eigenvector(i,j)
         enddo
         ovlps(j) = abs(zdum)**2
      enddo

      return
      end subroutine davovlp2c

!************************************************************************

      subroutine kry2finalpsi(Psi,DtPsi,Krylov,HKrylov,&
                              EigenVector,TrueOrder,qmap) 

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long), intent(in) :: TrueOrder
      integer(long), dimension(0:8*npackts+9), intent(in) :: qmap
      integer(long) :: s,q,qq,p,d,d1,d2
      real(dop), dimension(adim/npackts,0:intordera), intent(in) :: Krylov,&
                 HKrylov
      real(dop), dimension(0:intordera,0:intordera), intent(in) :: EigenVector
      complex(dop), dimension(adim), intent(out) :: psi,dtpsi

      do s = 1, nstate
         Do QQ = 0,npackts-1
            Q = qmap(QQ)
            Do d = 1,block(s)/npackts
               d1=zpsi(s)-1+d
               d2=(zpsi(s)-1)/npackts+d
               Psi(d1+QQ*block(s)/npackts)   =  Krylov(d2,0)*EigenVector(0,Q)
               DtPsi(d1+QQ*block(s)/npackts) =  HKrylov(d2,0)*EigenVector(0,Q)
            EndDo
            Do P = 1,TrueOrder
               Do D = 1,block(s)/npackts
                  d1=zpsi(s)-1+d
                  d2=(zpsi(s)-1)/npackts+d
                  Psi(d1+QQ*block(s)/npackts) = Psi(d1+QQ*block(s)/npackts) &
                            + Krylov(d2,P)*EigenVector(P,Q)
                  DtPsi(d1+QQ*block(s)/npackts) = DtPsi(d1+QQ*block(s)/npackts)&
                            +  HKrylov(d2,P)*EigenVector(P,Q)
               EndDo
            EndDo
         EndDo
      enddo


      return
      end subroutine kry2finalpsi


      subroutine psi2kry(Psi,Krylov,border) 

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      real(dop), dimension(adim/npackts,0:intordera), intent(out) :: Krylov
      complex(dop), dimension(adim), intent(in) :: psi

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q) = &
                  dble(Psi(zpsi(s)-1+d+q*(block(s)/npackts)))
            enddo
         enddo
      enddo

      return
      end subroutine psi2kry


      subroutine kry2psi(Psi,Krylov,border)

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      real(dop), dimension(adim/npackts,0:intordera), intent(in) :: Krylov
      complex(dop), dimension(adim), intent(out) :: psi

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               Psi(zpsi(s)-1+d+q*(block(s)/npackts)) = &
                  Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q)
            enddo
         enddo
      enddo

      return
      end subroutine kry2psi


      subroutine psi2kryr(workr,Krylov,border) 

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      real(dop), dimension(adim/npackts,0:intordera), intent(inout) :: Krylov
      real(dop), dimension(adim), intent(inout) :: workr

      do q=0,npackts-1
         do d=1,adim/npackts
            workr(zpsi(1)-1+d+q*(adim/npackts))=Krylov(d,Border+q)
         enddo
      enddo

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q) = &
                  workr(zpsi(s)-1+d+q*(block(s)/npackts))
            enddo
         enddo
      enddo

      return
      end subroutine psi2kryr


      subroutine kry2psir(workr,Krylov,border)

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      real(dop), dimension(adim/npackts,0:intordera), intent(inout) :: Krylov
      real(dop), dimension(adim), intent(inout) :: workr

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               workr(zpsi(s)-1+d+q*(block(s)/npackts)) = &
                  Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q)
            enddo
         enddo
      enddo

      do q=0,npackts-1
         do d=1,adim/npackts
            Krylov(d,Border+q)=workr(zpsi(1)-1+d+q*(adim/npackts))
         enddo
      enddo

      return
      end subroutine kry2psir

!************************************************************************

      subroutine kryc2finalpsi(Psi,DtPsi,Krylov,HKrylov,&
                               EigenVector,TrueOrder,qmap) 

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long), intent(in) :: TrueOrder
      integer(long), dimension(0:8*npackts+9), intent(in) :: qmap
      integer(long) :: s,q,qq,p,d,d1,d2
      complex(dop), dimension(adim/npackts,0:intordera), intent(in) :: Krylov,&
                 HKrylov
      complex(dop), dimension(0:intordera,0:intordera), intent(in) :: EigenVector
      complex(dop), dimension(adim), intent(out) :: psi,dtpsi

      do s = 1, nstate
         Do QQ = 0,npackts-1
            Q = qmap(QQ)
            Do d = 1,block(s)/npackts
               d1=zpsi(s)-1+d
               d2=(zpsi(s)-1)/npackts+d
               Psi(d1+QQ*block(s)/npackts)   =  Krylov(d2,0)*EigenVector(0,Q)
               DtPsi(d1+QQ*block(s)/npackts) =  HKrylov(d2,0)*EigenVector(0,Q)
            EndDo
            Do P = 1,TrueOrder
               Do D = 1,block(s)/npackts
                  d1=zpsi(s)-1+d
                  d2=(zpsi(s)-1)/npackts+d
                  Psi(d1+QQ*block(s)/npackts) = Psi(d1+QQ*block(s)/npackts) &
                            + Krylov(d2,P)*EigenVector(P,Q)
                  DtPsi(d1+QQ*block(s)/npackts) = DtPsi(d1+QQ*block(s)/npackts)&
                            +  HKrylov(d2,P)*EigenVector(P,Q)
               EndDo
            EndDo
         EndDo
      enddo


      return
      end subroutine kryc2finalpsi


      subroutine psi2kryc(Psi,Krylov,border) 

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      complex(dop), dimension(adim/npackts,0:intordera), intent(out) :: Krylov
      complex(dop), dimension(adim), intent(in) :: psi

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q) = &
                  Psi(zpsi(s)-1+d+q*(block(s)/npackts))
            enddo
         enddo
      enddo

      return
      end subroutine psi2kryc


      subroutine kryc2psi(Psi,Krylov,border)

use griddatmod
use psidef
use runpropmod

      implicit none

      integer(long) :: s,q,d
      integer(long), intent(in) :: border
      complex(dop), dimension(adim/npackts,0:intordera), intent(in) :: Krylov
      complex(dop), dimension(adim), intent(out) :: psi

      do s = 1, nstate
         Do q = 0, npackts-1
            Do d = 1, block(s)/npackts
               Psi(zpsi(s)-1+d+q*(block(s)/npackts)) = &
                  Krylov(((zpsi(s)-1)/npackts)+d,BOrder+q)
            enddo
         enddo
      enddo

      return
      end subroutine kryc2psi

!************************************************************************

!#######################################################################
!
!               WKGETDIAG
!
! Here the workspace required for the module is assigned
!
! workcdim, workrdim, workidim and workldim are passed via a common
! block in global.inc, and are checked that they are large enough for
! the module.
!
!#######################################################################

      subroutine wkgetdiag

use psidef

      implicit none

!-----------------------------------------------------------------------
! assign workspace used in module
!-----------------------------------------------------------------------
      workcdim=max(workcdim,3*fmblock,fmblock+maxspf)
      workrdim=workrdim
      workldim=workldim

      return
      end subroutine wkgetdiag

!***********************************************************************
      end module getdiagonalmod

