!***********************************************************************
!
! SUBROUTINE OD85CR
!
! Core of the Runge-Kutta integrator DoPri853.
!
! REFERENCE : Hairer/Norsett/Wanner
!             Solving Ordinary Differential Equations I, 2.ed.
!             Springer Verlag, 1993
!
!***********************************************************************
!
! INPUT:
! ------
! ODEMOD  external subroutine f that calculates the RHS of the
!         differential equation dy/dt = f(y,t).
!         Must have this interface:
!
!         subroutine ODEMOD(t,y,f,mc,mr,mi,ml)
!         real*8     t       in: time of RHS-evaluation
!         complex*16 y(*)    in: current values of the dynamic variables
!         complex*16 f(*)    out: result of RHS-evaluation
!         complex*16 mc(*)   .
!         real*8     mr(*)   . the well-known
!         integer    mi(*)   . MCTDH-arrays
!         logical    ml(*)   .
!
! ndgl    INTEGER     number of dynamic variables
! y       COMPLEX*16  initial values of the dynamic variables
! t       REAL*8      start time of integration
! tend    REAL*8      end time of integration
! hmax    REAL*8      maximum allowed stepsize
! h       REAL*8      initial stepsize (for h=0, we will guess one)
! tol     REAL*8      accuracy parameter
!
! uround  REAL*8      machine precision
! safe    REAL*8      safety factor for stepsize prediction
! beta    REAL*8      parameter for Lund-stabilisation
! facdec  REAL*8      factor for maximal decrease of stepsize
! facinc  REAL*8      factor for maximal increase of stepsize
!
! y1      COMPLEX*16  storage for intermediate result
! k1..k10 COMPLEX*16  storage for Runge-Kutta stages
! kerr,
! kerr2   COMPLEX*16  storage for the error stages
! wt      REAL*8      storage for scaling vector
!
! nfcn    INTEGER     number of evaluations of righthand side
! nstep   INTEGER     number of steps
! naccpt  INTEGER     number of accepted steps
! nrejct  INTEGER     number of rejected steps
!
! intgfl  INTEGER     control/error flag
!                     0: this is the first call
!                     1: this is a follow-up call
!
! lpost   LOGICAL     whether to call a post-processing function after
!                     each accepted step
! POSTFN  EXTERNAL    the post-processing function. Interface:
!                     POSTFN(y)
! ltest   LOGICAL     whether to call a test function after each step
! TESTFN  EXTERNAL    the test function. Interface:
!                     TESTFN(y)
! lerrfn  LOGICAL     whether to use an external routine for error estimate
! ERRFN   EXTERNAL    the error estimate function. Interface:
!                     ERRFN(y)
!
!
! OUTPUT:
! -------
! y       COMPLEX*16  final values of dynamic variables
! t       REAL*8      actual time reached in integration
! h       REAL*8      last stepsize
! hopt    REAL*8      last optimal stepsize (not trimmed)
!
! nfcn    INTEGER     number of evaluations of righthand side
! nstep   INTEGER     number of steps
! naccpt  INTEGER     number of accepted steps
! nrejct  INTEGER     number of rejected steps
!
! intgfl  INTEGER     control/error flag
!                        1: OK
!                     else: error!
! ierr    INTEGER     error code
! rerr    REAL*8      error parameter
!
!***********************************************************************

      subroutine OD85CR(&
           ODEMOD, ndgl, y, t, tend, hmax, h, hopt, tol,&
           uround, safe, beta, facdec, facinc,&
           y1, k1, k2, k3, k4, k5, k6, k7, k8, k9, k10,&
           kerr, kerr2, wt, nfcn, nstep, naccpt, nrejct,&
           intgfl, ierr, rerr, lpost, POSTFN, ltest, TESTFN,&
           lerrfn,errfn)

use decimal, only: long,dop
use wrintegrat, only: writestep
use odeuti, only: setewv,wgtnrm,odehin

      implicit none

      external                                     :: ODEMOD,POSTFN,TESTFN,ERRFN
      integer(long), intent(inout)                 :: nfcn,nstep,naccpt,&
                                                      nrejct,intgfl
      integer(long), intent(out)                   :: ierr
      integer(long), intent(in)                    :: ndgl
      real(dop), intent(out)                       :: rerr
      real(dop), intent(inout)                     :: hmax
      real(dop), intent(in)                        :: tend,tol,uround,safe,&
                                                      beta,facdec,facinc
      real(dop), intent(out)                       :: hopt
      real(dop), intent(inout)                     :: t,h
      real(dop), dimension(ndgl), intent(out)      :: wt
      complex(dop), dimension(ndgl), intent(inout) :: y,k1,k4
      complex(dop), dimension(ndgl), intent(out)   :: y1,kerr,kerr2,k2,k3,&
                                                      k5,k6,k7,k8,k9,k10
      logical(kind=4), intent(in)                  :: lpost,ltest,lerrfn

      integer(long)       :: stepnr,iasti,iord,stepreject
      real(dop)           :: err,expo1,fac,fac1,facold,facdc1,facin1,&
                             posneg,hnew,told,err2,deno
      complex(dop)        :: ztmp
      logical(kind=4)     :: lastfl,lrjcfl,lcntfl,lrecalc

      REAL(dop), PARAMETER :: ONE = 1.0_dop, ZERO = 0.0_dop

!  ----------------------------------
!  --- Coefficients of the method ---
!  ----------------------------------
      REAL(dop), PARAMETER :: &
       c2  = 0.526001519587677318785587544488e-01_dop,&
       c3  = 0.789002279381515978178381316732e-01_dop,&
       c4  = 0.118350341907227396726757197510e+00_dop,&
       c5  = 0.281649658092772603273242802490e+00_dop,&
       c6  = 0.333333333333333333333333333333e+00_dop,&
       c7  = 0.25e+00_dop,&
       c8  = 0.307692307692307692307692307692e+00_dop,&
       c9  = 0.651282051282051282051282051282e+00_dop,&
       c10 = 0.6e+00_dop,&
       c11 = 0.857142857142857142857142857142e+00_dop 

      REAL(dop), PARAMETER :: &
       b1 =   5.42937341165687622380535766363e-02_dop,&
       b6 =   4.45031289275240888144113950566e+00_dop,&
       b7 =   1.89151789931450038304281599044e+00_dop,&
       b8 =  -5.8012039600105847814672114227e+00_dop,&
       b9 =   3.1116436695781989440891606237e-01_dop,&
       b10 = -1.52160949662516078556178806805e-01_dop,&
       b11 =  2.01365400804030348374776537501e-01_dop,&
       b12 =  4.47106157277725905176885569043e-02_dop 

      REAL(dop), PARAMETER :: &
       bh1 = 0.244094488188976377952755905512e+00_dop,&
       bh2 = 0.733846688281611857341361741547e+00_dop,&
       bh3 = 0.220588235294117647058823529412e-01_dop 

      REAL(dop), PARAMETER :: &
       e1 =  0.1312004499419488073250102996e-01_dop,&
       e6 = -0.1225156446376204440720569753e+01_dop,&
       e7 = -0.4957589496572501915214079952e+00_dop,&
       e8 =  0.1664377182454986536961530415e+01_dop,&
       e9 = -0.3503288487499736816886487290e+00_dop,&
       e10 =  0.3341791187130174790297318841e+00_dop,&
       e11 =  0.8192320648511571246570742613e-01_dop,&
       e12 = -0.2235530786388629525884427845e-01_dop 

      REAL(dop), PARAMETER :: &
       a21 =    5.26001519587677318785587544488e-02_dop,&
       a31 =    1.97250569845378994544595329183e-02_dop,&
       a32 =    5.91751709536136983633785987549e-02_dop,&
       a41 =    2.95875854768068491816892993775e-02_dop,&
       a43 =    8.87627564304205475450678981324e-02_dop,&
       a51 =    2.41365134159266685502369798665e-01_dop,&
       a53 =   -8.84549479328286085344864962717e-01_dop,&
       a54 =    9.24834003261792003115737966543e-01_dop,&
       a61 =    3.7037037037037037037037037037e-02_dop,&
       a64 =    1.70828608729473871279604482173e-01_dop,&
       a65 =    1.25467687566822425016691814123e-01_dop,&
       a71 =    3.7109375e-02_dop,&
       a74 =    1.70252211019544039314978060272e-01_dop,&
       a75 =    6.02165389804559606850219397283e-02_dop,&
       a76 =   -1.7578125e-02_dop 

      REAL(dop), PARAMETER :: &
       a81 =    3.70920001185047927108779319836e-02_dop,&
       a84 =    1.70383925712239993810214054705e-01_dop,&
       a85 =    1.07262030446373284651809199168e-01_dop,&
       a86 =   -1.53194377486244017527936158236e-02_dop,&
       a87 =    8.27378916381402288758473766002e-03_dop,&
       a91 =    6.24110958716075717114429577812e-01_dop,&
       a94 =   -3.36089262944694129406857109825e+00_dop,&
       a95 =   -8.68219346841726006818189891453e-01_dop,&
       a96 =    2.75920996994467083049415600797e+01_dop,&
       a97 =    2.01540675504778934086186788979e+01_dop,&
       a98 =   -4.34898841810699588477366255144e+01_dop,&
       a101 =   4.77662536438264365890433908527e-01_dop,&
       a104 =  -2.48811461997166764192642586468e+00_dop,&
       a105 =  -5.90290826836842996371446475743e-01_dop,&
       a106 =   2.12300514481811942347288949897e+01_dop,&
       a107 =   1.52792336328824235832596922938e+01_dop,&
       a108 =  -3.32882109689848629194453265587e+01_dop,&
       a109 =  -2.03312017085086261358222928593e-02_dop 

      REAL(dop), PARAMETER :: &
       a111 =  -9.3714243008598732571704021658e-01_dop,&
       a114 =   5.18637242884406370830023853209e+00_dop,&
       a115 =   1.09143734899672957818500254654e+00_dop,&
       a116 =  -8.14978701074692612513997267357e+00_dop,&
       a117 =  -1.85200656599969598641566180701e+01_dop,&
       a118 =   2.27394870993505042818970056734e+01_dop,&
       a119 =   2.49360555267965238987089396762e+00_dop,&
       a1110 = -3.0467644718982195003823669022e+00_dop,&
       a121 =   2.27331014751653820792359768449e+00_dop,&
       a124 =  -1.05344954667372501984066689879e+01_dop,&
       a125 =  -2.00087205822486249909675718444e+00_dop,&
       a126 =  -1.79589318631187989172765950534e+01_dop,&
       a127 =   2.79488845294199600508499808837e+01_dop,&
       a128 =  -2.85899827713502369474065508674e+00_dop,&
       a129 =  -8.87285693353062954433549289258e+00_dop,&
       a1210 =  1.23605671757943030647266201528e+01_dop,&
       a1211 =  6.43392746015763530355970484046e-01_dop 

      SAVE

!  =======================
!  === INITIALIZATIONS ===
!  =======================
      stepnr = 1
      IF (intgfl .EQ. 1) THEN
         err    = 0.1_dop
         GOTO 320
      ENDIF
      stepreject=0

!  -----------------------------------------
!  --- Some factors for stepsize-control ---
!  -----------------------------------------
      expo1   = 0.125_dop-beta*0.75_dop
      facold  = safe**(1.d0/(expo1-beta))
      facdc1  = ONE/facdec
      facin1  = ONE/facinc
!  ------------------------------
!  --- Some more preparations ---
!  ------------------------------
      posneg = SIGN(ONE,tend-t)
      iasti  = 0
      CALL zcopy(ndgl,y,1,k5,1)
!  ------------------------
!  --- Initialize flags ---
!  ------------------------
      lastfl = .FALSE. 
      lrjcfl = .FALSE.    
!  --------------------------------
!  --- Initial model evaluation ---
!  --------------------------------
!  can be omitted since k4 already contains dtpsi
!      CALL ODEMOD(t,k5,k4)
!      nfcn = nfcn+1       

!  ===============
!  === RESTART ===
!  ===============
      lastfl = .FALSE.
!  ------------------------------------
!  --- Preparations for ODEHIN-call ---  
!  ------------------------------------
      hnew   = h
      hmax   = ABS(hmax)     
      iord   = 8

      IF (hnew .EQ. ZERO) THEN     
         call ODEHIN(ODEMOD, ndgl, t, k5, tend, posneg, k4, k2, k3,&
              iord, hmax, uround, wt, tol, nfcn, hnew)
      ENDIF  
      hnew = ABS(hnew)*posneg

!  ========================================
!  === TOP OF THE MAIN INTEGRATION LOOP ===
!  ========================================
   1  CONTINUE

         lcntfl = .FALSE.    
         h      =  hnew
         IF ( .NOT. lrjcfl ) THEN
!        -----------------------------------
!        ---  Copy result to y, k4 to k1 ---
!        -----------------------------------
            CALL zcopy(ndgl,k5,1, y,1)
            CALL zcopy(ndgl,k4,1,k1,1)
         ENDIF
!     ========================
!     === LEAVE INTEGRATOR ===
!     ========================
         IF ((intgfl .NE. 0) .OR. lastfl ) THEN
!        ---------------------------------
!        --- Leave the core-integrator ---
!        ---------------------------------   
            if (lastfl) intgfl = 1
            RETURN
         ENDIF
!     -----------------------------------      
!     --- Is the stepsize too small ? ---  
!     -----------------------------------
         IF (0.1_dop*ABS(h) .LE. ABS(t)*uround) THEN
            intgfl = -51
            rerr   =  t
            ierr   =  int(LOG10(h))
            GOTO 1
         ENDIF  
!     -----------------------------------------------
!     --- Adjust h, if integrator is near the end --- 
!     -----------------------------------------------
         IF ((t+1.01_dop*h-tend)*posneg .GT. ZERO) THEN
            h      =  tend - t 
            lastfl = .TRUE.
         ENDIF
!     ================================================
!     === CALCULATE THE 12 STAGES FOR THE NEW STEP === 
!     ================================================
         nstep = nstep + 1   
!     ---------------
!     --- Stage 1 ---
!     --------------- 
         CALL zcopy(ndgl,      y,1,y1,1)
         ztmp = DCMPLX(a21*h)
         CALL zaxpy(ndgl,ztmp,k1,1,y1,1)

         CALL ODEMOD(t+c2*h,y1,k2)
!     ---------------
!     --- Stage 2 ---
!     ---------------  
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a31*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a32*h)
         CALL zaxpy(ndgl,ztmp,k2,1,y1,1)
         ztmp = DCMPLX(ONE)
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c3*h,y1,k3)
!     ---------------
!     --- Stage 3 ---
!     ---------------     
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a41*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a43*h)
         CALL zaxpy(ndgl,ztmp,k3,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c4*h,y1,k4)
!     ---------------
!     --- Stage 4 ---
!     ---------------   
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a51*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a53*h)
         CALL zaxpy(ndgl,ztmp,k3,1,y1,1)  
         ztmp = DCMPLX(a54*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c5*h,y1,k5)
!     ---------------
!     --- Stage 5 ---
!     ---------------   
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a61*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a64*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1)  
         ztmp = DCMPLX(a65*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c6*h,y1,k6)
!     ---------------
!     --- Stage 6 ---
!     ---------------  
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a71*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a74*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1) 
         ztmp = DCMPLX(a75*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)  
         ztmp = DCMPLX(a76*h)
         CALL zaxpy(ndgl,ztmp,k6,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c7*h,y1,k7)
!     ---------------
!     --- Stage 7 ---
!     ---------------  
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a81*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a84*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1) 
         ztmp = DCMPLX(a85*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)  
         ztmp = DCMPLX(a86*h)
         CALL zaxpy(ndgl,ztmp,k6,1,y1,1)
         ztmp = DCMPLX(a87*h)
         CALL zaxpy(ndgl,ztmp,k7,1,y1,1) 
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c8*h,y1,k8)
!     ---------------
!     --- Stage 8 ---
!     ---------------  
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a91*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a94*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1) 
         ztmp = DCMPLX(a95*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)  
         ztmp = DCMPLX(a96*h)
         CALL zaxpy(ndgl,ztmp,k6,1,y1,1)
         ztmp = DCMPLX(a97*h)
         CALL zaxpy(ndgl,ztmp,k7,1,y1,1) 
         ztmp = DCMPLX(a98*h)
         CALL zaxpy(ndgl,ztmp,k8,1,y1,1) 
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c9*h,y1,k9)
!     ---------------
!     --- Stage 9 ---
!     --------------- 
         CALL zcopy(ndgl,     k1,1,y1,1) 
         ztmp = DCMPLX(a101*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a104*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1) 
         ztmp = DCMPLX(a105*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)  
         ztmp = DCMPLX(a106*h)
         CALL zaxpy(ndgl,ztmp,k6,1,y1,1)
         ztmp = DCMPLX(a107*h)
         CALL zaxpy(ndgl,ztmp,k7,1,y1,1) 
         ztmp = DCMPLX(a108*h)
         CALL zaxpy(ndgl,ztmp,k8,1,y1,1) 
         ztmp = DCMPLX(a109*h)
         CALL zaxpy(ndgl,ztmp,k9,1,y1,1) 
         ztmp = DCMPLX(ONE   )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c10*h,y1,k10)
!     ----------------
!     --- Stage 10 ---
!     ---------------- 
         CALL zcopy(ndgl,      k1,1,y1,1) 
         ztmp = DCMPLX(a111*h )
         CALL zscal(ndgl,ztmp,      y1,1)
         ztmp = DCMPLX(a114*h )
         CALL zaxpy(ndgl,ztmp, k4,1,y1,1) 
         ztmp = DCMPLX(a115*h )
         CALL zaxpy(ndgl,ztmp, k5,1,y1,1)  
         ztmp = DCMPLX(a116*h )
         CALL zaxpy(ndgl,ztmp, k6,1,y1,1)
         ztmp = DCMPLX(a117*h )
         CALL zaxpy(ndgl,ztmp, k7,1,y1,1) 
         ztmp = DCMPLX(a118*h )
         CALL zaxpy(ndgl,ztmp, k8,1,y1,1) 
         ztmp = DCMPLX(a119*h )
         CALL zaxpy(ndgl,ztmp, k9,1,y1,1) 
         ztmp = DCMPLX(a1110*h)
         CALL zaxpy(ndgl,ztmp,k10,1,y1,1) 
         ztmp = DCMPLX(ONE)
         CALL zaxpy(ndgl,ztmp,  y,1,y1,1) 

         CALL ODEMOD(t+c11*h,y1,k2)
!     ----------------
!     --- Stage 11 ---
!     ----------------  
         CALL zcopy(ndgl,      k1,1,y1,1) 
         ztmp = DCMPLX(a121*h )
         CALL zscal(ndgl,ztmp,      y1,1)
         ztmp = DCMPLX(a124*h )
         CALL zaxpy(ndgl,ztmp, k4,1,y1,1) 
         ztmp = DCMPLX(a125*h )
         CALL zaxpy(ndgl,ztmp, k5,1,y1,1)  
         ztmp = DCMPLX(a126*h )
         CALL zaxpy(ndgl,ztmp, k6,1,y1,1)
         ztmp = DCMPLX(a127*h )
         CALL zaxpy(ndgl,ztmp, k7,1,y1,1) 
         ztmp = DCMPLX(a128*h )
         CALL zaxpy(ndgl,ztmp, k8,1,y1,1) 
         ztmp = DCMPLX(a129*h )
         CALL zaxpy(ndgl,ztmp, k9,1,y1,1) 
         ztmp = DCMPLX(a1210*h)
         CALL zaxpy(ndgl,ztmp,k10,1,y1,1) 
         ztmp = DCMPLX(a1211*h)
         CALL zaxpy(ndgl,ztmp, k2,1,y1,1) 
         ztmp = DCMPLX(ONE)
         CALL zaxpy(ndgl,ztmp,  y,1,y1,1)

         CALL ODEMOD(t+h,y1,k3)
         nfcn = nfcn + 11
!     ----------------
!     --- Stage 12 ---
!     ----------------  
         CALL zcopy(ndgl,      k1,1,k4,1) 
         ztmp = DCMPLX(b1 )
         CALL zscal(ndgl,ztmp,      k4,1)
         ztmp = DCMPLX(b6 )
         CALL zaxpy(ndgl,ztmp, k6,1,k4,1)
         ztmp = DCMPLX(b7 )
         CALL zaxpy(ndgl,ztmp, k7,1,k4,1) 
         ztmp = DCMPLX(b8 )
         CALL zaxpy(ndgl,ztmp, k8,1,k4,1) 
         ztmp = DCMPLX(b9 )
         CALL zaxpy(ndgl,ztmp, k9,1,k4,1) 
         ztmp = DCMPLX(b10)
         CALL zaxpy(ndgl,ztmp,k10,1,k4,1) 
         ztmp = DCMPLX(b11)
         CALL zaxpy(ndgl,ztmp, k2,1,k4,1) 
         ztmp = DCMPLX(b12)
         CALL zaxpy(ndgl,ztmp, k3,1,k4,1) 

         CALL zcopy(ndgl,      y,1,k5,1)
         ztmp = DCMPLX(h)
         CALL zaxpy(ndgl,ztmp,k4,1,k5,1)

!     -----------------------
!     --- The error-stage ---  
!     -----------------------
         CALL zcopy(ndgl,      k1,1,kerr,1) 
         ztmp = DCMPLX(e1 )
         CALL zscal(ndgl,ztmp,      kerr,1)
         ztmp = DCMPLX(e6 )
         CALL zaxpy(ndgl,ztmp, k6,1,kerr,1)
         ztmp = DCMPLX(e7 )
         CALL zaxpy(ndgl,ztmp, k7,1,kerr,1) 
         ztmp = DCMPLX(e8 )
         CALL zaxpy(ndgl,ztmp, k8,1,kerr,1) 
         ztmp = DCMPLX(e9 )
         CALL zaxpy(ndgl,ztmp, k9,1,kerr,1) 
         ztmp = DCMPLX(e10)
         CALL zaxpy(ndgl,ztmp,k10,1,kerr,1) 
         ztmp = DCMPLX(e11)
         CALL zaxpy(ndgl,ztmp, k2,1,kerr,1) 
         ztmp = DCMPLX(e12)
         CALL zaxpy(ndgl,ztmp, k3,1,kerr,1) 

!     +-------------------------+
!     +--   Result     -> k5  --+
!     +--   Last stage -> y1  --+
!     +-------------------------+
!     +--   Extra stage -> k4 --+
!     +-------------------------+
!     +--   Later on:         --+
!     +--   RHS(k5)    -> k4  --+
!     +--   RHS(y1)    -> k3  --+ 
!     +-------------------------+
!     ------------------------
!     --- Error estimation ---
!     ------------------------
         if (lerrfn) then
            call errfn(err,k5)
            err=err/tol
         else
            CALL SETEWV(ndgl,y,k5,tol,uround,wt)
            err   = wgtnrm(ndgl,kerr,wt) 
         endif
         if(err.ne.err) err=1.d6/h ! Fallback for NaN

         CALL zcopy(ndgl,     k4,1,kerr2,1)
         ztmp = DCMPLX(-bh1)
         CALL zaxpy(ndgl,ztmp,k1,1,kerr2,1)
         ztmp = DCMPLX(-bh2)
         CALL zaxpy(ndgl,ztmp,k9,1,kerr2,1)
         ztmp = DCMPLX(-bh3)
         CALL zaxpy(ndgl,ztmp,k3,1,kerr2,1) 

         if (lerrfn) then
            call errfn(err2,k5)
            err2=err2/tol
         else
            err2  = wgtnrm(ndgl,kerr2,wt)  
         endif
         if(err2.ne.err2) err2=1.d0/h ! Fallback for NaN

         err   = err**2
         deno  = err + 0.01_dop*err2**2
         IF (deno .LE. ZERO) deno = ONE 
         err   = h*(err/SQRT(deno))

!     ---------------------------
!     --- Computation of hnew --- 
!     ---------------------------
         fac1  = err**expo1
!     --------------------------
!     --- LUND-stabilization ---
!     --------------------------
         fac   = fac1/facold**beta 
!     ----------------------------------------------
!     --- We require  facdec <= hnew/h <= facinc ---
!     ----------------------------------------------
         fac   = MAX(facin1,MIN(facdc1,fac/safe))
         hnew  = h/fac   
!     ========================  
!     === STEP IS ACCEPTED ===
!     ======================== 
 320     CONTINUE
         IF (err .LE. ONE) THEN
            IF (intgfl .EQ. 1) THEN
               intgfl = 0
               GOTO 32
            ENDIF
            lrjcfl = .FALSE.
!        -------------------
!        --- Adjust time ---
!        -------------------
            told   = t
            t      = t + h 
!        -----------------------------
!        --- New model evaluation  ---
!        -----------------------------
            CALL ODEMOD(t,k5,k4)
            nfcn   = nfcn + 1    
            facold = MAX(err,1.0e-4_dop)
            naccpt = naccpt + 1

   32       CONTINUE
!           write into MCTDH's STEPS-file
            CALL WriteStep(stepnr,1,h,err,told)
!           call post-processor if required
            if (lpost) call POSTFN(k5)
            stepnr = stepnr+1
            stepreject = 0

!        ==========================
!        === PREPARE NEXT STEP  ===
!        ==========================
!        ---------------------------
!        --- Calculation of hnew ---  
!        ---------------------------
            IF (ABS(hnew) .GT. hmax) hnew = posneg*hmax  
            IF (lrjcfl) hnew = posneg*MIN(ABS(hnew),ABS(h))

            if (ltest) then
               lrecalc = .false.
               call TESTFN(k5,lrecalc,t)
               if (lrecalc) then
                   CALL ODEMOD(t,k5,k4)
                   hnew = 0.75d0*h
               endif
            endif

!     ========================
!     === STEP IS REJECTED ===
!     ========================    
         ELSE   
!           write into MCTDH's STEPS-file
            CALL WriteStep(stepnr,0,h,err,t)
            stepnr = stepnr+1
!        ---------------------------
!        --- Calculation of hnew ---  
!        ---------------------------
            hnew   = h/MIN(facdc1,fac1/safe)
            if(err.gt.  3.d0) hnew=hnew*safe   ! HDM May 2021
            if(err.gt. 10.d0) hnew=hnew*safe
            if(err.gt. 30.d0) hnew=hnew*safe
            if(err.gt.100.d0) hnew=hnew*safe
            if(err.gt.300.d0) hnew=hnew*safe
            hnew=hnew*safe**stepreject
            stepreject=stepreject+1

!        ---------------------------------------  
!        --- Adjust the flags and statistics ---
!        ---------------------------------------
            lrjcfl = .TRUE.  
            lastfl = .FALSE.
!           IF (naccpt .GE. 1) nrejct = nrejct + 1   
            nrejct = nrejct+1
         ENDIF         
!  ====================================================
!  === HERE IS THE END OF THE MAIN INTEGRATION-LOOP ===
!  ==================================================== 
      IF (.NOT.lastfl)  hopt = hnew
      GOTO 1

      END subroutine OD85CR               ! of OD85CR
