!***********************************************************************
!
! SUBROUTINE ODP5CRF
!
! Core of the Runge-Kutta integrator of order 5
! (Dormand-Prince formula) with fixed timestep
!
! REFERENCE : Hairer/Norsett/Wanner
!             Solving Ordinary Differential Equations I, 2.ed.
!             Springer Verlag, 1993
!
!***********************************************************************
!
! INPUT:
! ------
! ODEMOD  external subroutine f that calculates the RHS of the
!         differential equation dy/dt = f(y,t).
!         Must have this interface:
!
!         subroutine ODEMOD(t,y,f,mc,mr,mi,ml)
!         real*8     t       in: time of RHS-evaluation
!         complex*16 y(*)    in: current values of the dynamic variables
!         complex*16 f(*)    out: result of RHS-evaluation
!         complex*16 mc(*)   .
!         real*8     mr(*)   . the well-known
!         integer    mi(*)   . MCTDH-arrays
!         logical    ml(*)   .
!
! ndgl    INTEGER     number of dynamic variables
! y       COMPLEX*16  initial values of the dynamic variables
! t       REAL*8      start time of integration
! tend    REAL*8      end time of integration
! hmax    REAL*8      maximum allowed stepsize
! hmin    REAL*8      minimum allowed stepsize
! h       REAL*8      initial stepsize (for h=0, we will guess one)
! tol     REAL*8      accuracy parameter
!
! uround  REAL*8      machine precision
! safe    REAL*8      safety factor for stepsize prediction
! beta    REAL*8      parameter for Lund-stabilisation
! facdec  REAL*8      factor for maximal decrease of stepsize
! facinc  REAL*8      factor for maximal increase of stepsize
!
! y1      COMPLEX*16  storage for intermediate result
! k1..k6, COMPLEX*16  storage for Runge-Kutta stages
! ys
! wt      REAL*8      storage for scaling vector
!
! nfcn    INTEGER     number of evaluations of righthand side
! nstep   INTEGER     number of steps
! naccpt  INTEGER     number of accepted steps
! nrejct  INTEGER     number of rejected steps
!
! intgfl  INTEGER     control/error flag
!                     0: this is the first call
!                     1: this is a follow-up call
!
! lpost   LOGICAL     whether to call a post-processing function after
!                     each accepted step
! POSTFN  EXTERNAL    the post-processing function. Interface:
!                     POSTFN(y)
! ltest   LOGICAL     whether to call a test function after each step
! TESTFN  EXTERNAL    the test function. Interface:
!                     TESTFN(y)
!
!
! OUTPUT:
! -------
! y       COMPLEX*16  final values of dynamic variables
! t       REAL*8      actual time reached in integration
! h       REAL*8      last stepsize
! hopt    REAL*8      last optimal stepsize (not trimmed)
!
! nfcn    INTEGER     number of evaluations of righthand side
! nstep   INTEGER     number of steps
! naccpt  INTEGER     number of accepted steps
! nrejct  INTEGER     number of rejected steps
!
! intgfl  INTEGER     control/error flag
!                        1: OK
!                     else: error!
! ierr    INTEGER     error code
! rerr    REAL*8      error parameter
!
!***********************************************************************

      subroutine ODP5CRF( &
           ODEMOD, ndgl, y, t, tend, hmax, hmin, h, hopt, tol, &
           uround, y1, k1, k2, k3, k4, k5, k6, ys, wt, &
           nfcn, nstep, naccpt, intgfl, ierr, rerr, lpost, POSTFN, &
           ltest, TESTFN)

use decimal, only: long,dop
use wrintegrat, only: writestep
use odeuti, only: setewv,wgtnrm,odehin

      implicit none

      external                                     :: ODEMOD,POSTFN,TESTFN
      integer(long), intent(inout)                 :: nfcn,nstep,naccpt,intgfl
      integer(long), intent(out)                   :: ierr
      integer(long), intent(in)                    :: ndgl
      real(dop), intent(out)                       :: rerr
      real(dop), intent(in)                        :: tend,tol,uround
      real(dop), intent(inout)                     :: hmax
      real(dop), intent(in)                        :: hmin
      real(dop), intent(out)                       :: hopt
      real(dop), intent(inout)                     :: t,h
      real(dop), dimension(ndgl), intent(out)      :: wt
      complex(dop), dimension(ndgl), intent(inout) :: y,k1,k2
      complex(dop), dimension(ndgl), intent(out)   :: y1,ys,k3,k4,k5,k6
      logical(kind=4), intent(in)                  :: lpost,ltest

      integer(long)        :: stepnr,iasti,iord
      real(dop)            :: err,expo1,fac,fac1,facold,facdc1,facin1,&
                              posneg,hnew,told,hin

      complex(dop)         :: ztmp
      logical(kind=4)      :: lastfl,lrjcfl,lcntfl,lrecalc

      REAL(dop), PARAMETER :: ONE = 1.0_dop, ZERO = 0.0_dop

!  ----------------------------------
!  --- Coefficients of the method ---
!  ----------------------------------
      REAL(dop), PARAMETER :: &
                              c2  = 0.2_dop,&
                              c3  = 0.3_dop,&
                              c4  = 0.8_dop,&
                              c5  = 8.0_dop/9.0_dop 

      REAL(dop), PARAMETER :: &
                              a21 =    0.2_dop,&
                              a31 =    3.0_dop/40.0_dop,&
                              a32 =    9.0_dop/40.0_dop,&
                              a41 =   44.0_dop/45.0_dop,&
                              a42 = - 56.0_dop/15.0_dop,&
                              a43 =   32.0_dop/9.0_dop 

      REAL(dop), PARAMETER :: &
                              a51 =  19372.0_dop/6561.0_dop,&
                              a52 = -25360.0_dop/2187.0_dop,&
                              a53 =  64448.0_dop/6561.0_dop,&
                              a54 = -  212.0_dop/729.0_dop,&
                              a61 =   9017.0_dop/3168.0_dop,&
                              a62 = -  355.0_dop/33.0_dop,&
                              a63 =  46732.0_dop/5247.0_dop,&
                              a64 =     49.0_dop/176.0_dop,&
                              a65 = - 5103.0_dop/18656.0_dop 

      Real(dop), PARAMETER :: &
                              b1  =    35.0_dop/384.0_dop,&
                              b3  =   500.0_dop/1113.0_dop,&
                              b4  =   125.0_dop/192.0_dop,&
                              b5  = -2187.0_dop/6784.0_dop,&
                              b6  =    11.0_dop/84.0_dop 

      REAL(dop), PARAMETER :: &
                              e1  =     71.0_dop/57600.0_dop,&
                              e3  = -   71.0_dop/16695.0_dop,&
                              e4  =     71.0_dop/1920.0_dop,&
                              e5  = -17253.0_dop/339200.0_dop,&
                              e6  =     22.0_dop/525.0_dop,&
                              e7  = -    1.0_dop/40.0_dop 

     ! REAL(dop), PARAMETER :: &
     !                         e1  =   5179.0_dop/57600.0_dop,&   !C coefficients for
     !                         e3  =   7571.0_dop/16695.0_dop,&   !C 5th order Dormand-Prince
     !                         e4  =    393.0_dop/640.0_dop,&     !C
     !                         e5  = -92097.0_dop/339200.0_dop,&  !C
     !                         e6  =    187.0_dop/2100.0_dop,&    !C
     !                         e7  =      1.0_dop/40.0_dop        !C
      SAVE

!  =======================
!  === INITIALIZATIONS ===
!  =======================
      ierr = 0
      rerr = 0.0
      stepnr = 1
      IF (intgfl .EQ. 1) THEN
         err    = 0.1_dop        ! zur Sicherheit
         GOTO 320             ! Zwischenlabel, (kein Sprung in IF-Block)
      ENDIF
!  -----------------------------------------
!  --- Some factors for stepsize-control ---
!  -----------------------------------------
!     expo1   = 0.2_dop-beta*0.75_dop
!     facold  = safe**(1.0_dop/(expo1-beta))
!     facdc1  = ONE/facdec
!     facin1  = ONE/facinc
!  ------------------------------
!  --- Some more preparations ---
!  ------------------------------
      posneg = SIGN(ONE,tend-t)
      iasti  = 0
      call zcopy(ndgl,y,1,y1,1)
!  ------------------------
!  --- Initialize flags ---
!  ------------------------
      lastfl = .FALSE.
      lrjcfl = .FALSE.
!  --------------------------------
!  --- Initial model evaluation ---
!  --------------------------------
!  can be omitted since k2=dtpsi
!      CALL ODEMOD(t,y1,k2)
!      nfcn = nfcn+1

!  ===============
!  === RESTART ===
!  ===============
!      lastfl = .FALSE.
!  ------------------------------------
!  --- Preparations for ODEHIN-CALL ---
!  ------------------------------------
      hnew   = h
      hmax   = ABS(hmax)
      iord   = 5

      IF (hnew .EQ. ZERO) THEN
         call ODEHIN(ODEMOD, ndgl, t, y1, tend, posneg, k2, k3, k4,&
              iord, hmax, uround, wt, tol, nfcn, hnew)
      ENDIF

      hnew = max(hnew,hmin)
      hnew = ABS(hnew)*posneg
      hin = hnew

!  ========================================
!  === TOP OF THE MAIN INTEGRATION LOOP ===
!  ========================================
   1  CONTINUE
         lcntfl = .FALSE.
         h      =  hnew
         IF (lrjcfl) THEN
            CALL zcopy(ndgl,y,1,y1,1)
         ELSE
!        -----------------------------------
!        ---  Copy result to y, k2 to k1 ---
!        -----------------------------------
            CALL zcopy(ndgl,y1,1, y,1)
            CALL zcopy(ndgl,k2,1,k1,1)
         ENDIF
!     ========================
!     === LEAVE INTEGRATOR ===
!     ========================
         IF ((intgfl .NE. 0) .OR. lastfl ) THEN
!        ---------------------------------
!        --- Leave the core-integrator ---
!        ---------------------------------
            if (lastfl) intgfl = 1
            RETURN
         ENDIF

!     -----------------------------------------------
!     --- Adjust h, if integrator is near the end ---
!     -----------------------------------------------
         IF ((t+1.01_dop*h-tend)*posneg .GT. ZERO) THEN
            h      =  tend - t
            lastfl = .TRUE.
         ENDIF
!     ===============================================
!     === CALCULATE THE 6 STAGES FOR THE NEW STEP ===
!     ===============================================
         nstep = nstep + 1
!     ---------------
!     --- Stage 1 ---
!     ---------------
         ztmp = DCMPLX(a21*h)
         CALL zaxpy(ndgl,ztmp,k1,1,y1,1)

         CALL ODEMOD(t+c2*h,y1,k2)
!     ---------------
!     --- Stage 2 ---
!     ---------------
         CALL zcopy(ndgl,     k1,1,y1,1)
         ztmp = DCMPLX(a31*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a32*h)
         CALL zaxpy(ndgl,ztmp,k2,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c3*h,y1,k3)
!     ---------------
!     --- Stage 3 ---
!     ---------------
         CALL zcopy(ndgl,     k1,1,y1,1)
         ztmp = DCMPLX(a41*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a42*h)
         CALL zaxpy(ndgl,ztmp,k2,1,y1,1)
         ztmp = DCMPLX(a43*h)
         CALL zaxpy(ndgl,ztmp,k3,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c4*h,y1,k4)
!     ---------------
!     --- Stage 4 ---
!     ---------------
         CALL zcopy(ndgl,     k1,1,y1,1)
         ztmp = DCMPLX(a51*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(a52*h)
         CALL zaxpy(ndgl,ztmp,k2,1,y1,1)
         ztmp = DCMPLX(a53*h)
         CALL zaxpy(ndgl,ztmp,k3,1,y1,1)
         ztmp = DCMPLX(a54*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+c5*h,y1,k5)
!     ---------------
!     --- Stage 5 ---
!     ---------------
         CALL zcopy(ndgl,     k1,1,ys,1)
         ztmp = DCMPLX(a61*h)
         CALL zscal(ndgl,ztmp,     ys,1)
         ztmp = DCMPLX(a62*h)
         CALL zaxpy(ndgl,ztmp,k2,1,ys,1)
         ztmp = DCMPLX(a63*h)
         CALL zaxpy(ndgl,ztmp,k3,1,ys,1)
         ztmp = DCMPLX(a64*h)
         CALL zaxpy(ndgl,ztmp,k4,1,ys,1)
         ztmp = DCMPLX(a65*h)
         CALL zaxpy(ndgl,ztmp,k5,1,ys,1)
         ztmp = DCMPLX(ONE  )
         CALL zaxpy(ndgl,ztmp, y,1,ys,1)

         CALL ODEMOD(t+h,ys,k6)
!     ---------------
!     --- Stage 6 ---
! k7 is stored in k2. y1 is y(k+1)
!     ---------------
         CALL zcopy(ndgl,     k1,1,y1,1)
         ztmp = DCMPLX(b1*h)
         CALL zscal(ndgl,ztmp,     y1,1)
         ztmp = DCMPLX(b3*h)
         CALL zaxpy(ndgl,ztmp,k3,1,y1,1)
         ztmp = DCMPLX(b4*h)
         CALL zaxpy(ndgl,ztmp,k4,1,y1,1)
         ztmp = DCMPLX(b5*h)
         CALL zaxpy(ndgl,ztmp,k5,1,y1,1)
         ztmp = DCMPLX(b6*h)
         CALL zaxpy(ndgl,ztmp,k6,1,y1,1)
         ztmp = DCMPLX(ONE )
         CALL zaxpy(ndgl,ztmp, y,1,y1,1)

         CALL ODEMOD(t+h,y1,k2)
         nfcn = nfcn + 6

!     ---------------
!     -----------------------
!     --- The error-stage ---
!     -----------------------
         ztmp = DCMPLX(e4*h)
         CALL zscal(ndgl,ztmp,     k4,1)
         ztmp = DCMPLX(e1*h)
         CALL zaxpy(ndgl,ztmp,k1,1,k4,1)
         ztmp = DCMPLX(e3*h)
         CALL zaxpy(ndgl,ztmp,k3,1,k4,1)
         ztmp = DCMPLX(e5*h)
         CALL zaxpy(ndgl,ztmp,k5,1,k4,1)
         ztmp = DCMPLX(e6*h)
         CALL zaxpy(ndgl,ztmp,k6,1,k4,1)
         ztmp = DCMPLX(e7*h)
         CALL zaxpy(ndgl,ztmp,k2,1,k4,1)
!     --------------------------
!     --- Estimate the error ---
!     --------------------------
         CALL SETEWV(ndgl,y,y1,tol,uround,wt)
         err = wgtnrm(ndgl,k4,wt)
            
         hnew  = h
!     ========================
!     === STEP IS ACCEPTED ===
!     ========================
 320     CONTINUE
         if (err .gt. 1.0e4_dop) then
            write(6,*) "RK5 ABORT: ERROR OUT OF RANGE:"
            stop
         endif
!        IF (intgfl .EQ. 1) THEN
!           intgfl = 0
!           GOTO 32
!        ENDIF
         lrjcfl = .FALSE.
         naccpt =  naccpt + 1
!        -------------------
!        --- Adjust time ---
!        -------------------
         told = t
         t    = t + h
 32      CONTINUE
!        write into MCTDH's STEPS-file
         CALL WriteStep(stepnr,1,h,err,told)

!           call post-processing function if required
         if (lpost) call POSTFN(y1)
         if (ltest) then
            lrecalc = .false.
            call TESTFN(y1,lrecalc,t)
         endif

         stepnr = stepnr+1

      IF (.NOT. lastfl) hopt = hnew
      GOTO 1

      END subroutine ODP5CRF              ! of ODP5CRF
