!***********************************************************************
      module odeuti

      implicit none

      integer, parameter :: long  = selected_int_kind(9)
      integer, parameter :: dop = selected_real_kind(15,307)

      private
      public :: odehin,setewv,wgtnrm,wgtdst

      contains

!***********************************************************************
!
! SubRoutine ODEHIN
!
! computes an initial stepsize guess for integration
!
! REFERENCE : Hairer/Norsett/Wanner
!             Solving Ordinary Differential Equations I, 2.ed.
!             Springer Verlag, 1993
!
!***********************************************************************
!
! INPUT:
! ------
! ODEMOD  EXTERNAL    function to evaluate the RHS of the ODE
! ndgl    INTEGER     dimension of differential equation system
! t       REAL*8      initial time
! y       COMPLEX*16  initial values of dynamic variables
! tend    REAL*8      final time of integration 
! posneg  REAL*8      direction of integration
! dy0     COMPLEX*16  initial derivatives of dynamics variables
! dy1,y1  COMPLEX*16  additional storage for EULER step
! iord    INTEGER     order of integration method to be used
! hmax    REAL*8      maximum permitted stepsize
! small   REAL*8      underflow indicator
! wt      REAL*8      storage for weight vector
! tol     REAL*8      accuracy parameter
! nfcn    INTEGER     statistics variable
!
! OUPUT:
! ------
! nfcn    INTEGER     statistics variable
! hini    REAL*8      first stepsize to be tried in integration
!
!***********************************************************************

      subroutine ODEHIN(ODEMOD, ndgl, t, y, tend, posneg, dy0, dy1,&
           y1, iord, hmax, small, wt, tol, nfcn, hini)

      implicit none

      external                                  :: ODEMOD
      integer(long), intent(in)                 :: ndgl,iord
      integer(long), intent(inout)              :: nfcn
      real(dop), intent(in)                     :: t,tend,posneg,hmax,small,tol
      real(dop), intent(out)                    :: hini
      real(dop), dimension(ndgl), intent(out)   :: wt
      complex(dop), dimension(ndgl), intent(out):: y1,dy1
      complex(dop), dimension(ndgl), intent(in) :: y,dy0
      
      real(dop)          :: ynrm,dynrm,h,hmx,ddynrm,dnrmmx,htry
      complex(dop)       :: ztmp

      REAL(dop), PARAMETER :: ONE = 1.0_dop

! --------------------
! --- PREPARATIONS ---
! --------------------
      call SETEWV(ndgl,y,y,tol,small,wt)
      ynrm  = wgtnrm(ndgl, y, wt)
      dynrm = wgtnrm(ndgl, dy0, wt)
! -------------------------------------------------------
! --- COMPUTE STEPSIZE FOR EXPLICIT EULER SUCH THAT   ---
! --- THE INCREMENT IS SMALL COMPARED TO THE SOLUTION ---
! -------------------------------------------------------
      IF ((dynrm .LE. small) .OR. (ynrm .LE. small)) THEN
         h = 1.0e-6_dop
      ELSE
         h = MAX(hmax*small,ABS((ynrm/dynrm)*0.01_dop))
      ENDIF
      hmx = MIN(ABS(tend-t),hmax)
      h   = MIN(h,hmx)
      h   = h*posneg
!  --------------------------------------
!  --- PERFORM AN EXPLICIT EULER STEP ---
!  --------------------------------------
      CALL zcopy(ndgl,y,1,y1,1)
      ztmp = DCMPLX(h)
      CALL zaxpy(ndgl,ztmp,dy0,1,y1,1)
!  ----------------
!  --- EVALUATE ---
!  ----------------
      CALL ODEMOD(t+h,y1,dy1)
      nfcn=nfcn+1
!  ------------------------------------------------------
!  --- ESTIMATE THE SECOND DERIVATIVE OF THE SOLUTION ---
!  ------------------------------------------------------
      CALL SETEWV(ndgl,y,y1,tol,small,wt)
      ddynrm = wgtdst(ndgl,dy0,dy1,wt)
      ddynrm = ddynrm/h
! ---------------------------------
! --- COMPUTE INITIAL STEP SIZE ---
! ---------------------------------
      dnrmmx = MAX(ABS(ddynrm),dynrm)
      IF (dnrmmx .LE. 1.0e-15_dop) THEN
         htry = MAX(1.0e-6_dop,ABS(h)*1.0e-3_dop)
      ELSE
         htry = (0.01_dop/dnrmmx)**(ONE/iord)
      ENDIF
      hini = MIN(1.0e2_dop*h,htry,hmx)
      hini = hini*posneg

      RETURN
      END subroutine ODEHIN                   !  of ODEHIN

!***********************************************************************
!
! SubRoutine SETEWV
!
! set the error weight vector according to the given tolerance
!
!***********************************************************************
!
! INPUT:
! ------
! ndim    INTEGER     dimension of the vectors
! y0,y1   COMPLEX*16  scaling vectors
! tol     REAL*8      tolerance parameter
! small   REAL*8      underflow indicator
!
! OUTPUT:
! -------
! wt      REAL*8      error weight vector
!
!***********************************************************************

      subroutine SETEWV(ndim, y0, y1, tol, small, wt)

      implicit none

      integer(long)                             :: i
      integer(long), intent(in)                 :: ndim
      real(dop), intent(in)                     :: tol,small
      real(dop), dimension(ndim), intent(out)   :: wt
      complex(dop), dimension(ndim), intent(in) :: y0,y1

! -----------------------------------
! --- BOTH TOLERANCES ARE SCALARS ---
! -----------------------------------
      do i = 1,ndim
         wt(i) = tol + tol*MAX(ABS(y0(i)), ABS(y1(i)), small)
      enddo

      return
      end subroutine SETEWV             ! of SETEWV

!***********************************************************************
!
! Function WGTNRM
!
! computes the weighted norm of a vector
!
!***********************************************************************
!
! INPUT:
! ------
! ndim    INTEGER     dimension of the vectors
! vec     COMPLEX*16  vector, whose norm is to be computed
! wt      REAL*8      vector of weights
!
! OUTPUT :
! --------
! wgtnrm  REAL*8     function value
!
!***********************************************************************

      REAL(dop) FUNCTION wgtnrm(ndim, vec, wt)

      IMPLICIT NONE

      INTEGER(long)                             :: i
      INTEGER(long), intent(in)                 :: ndim
      REAL(dop)                                 :: tmp
      REAL(dop), dimension(ndim), intent(in)    :: wt
      COMPLEX(dop), dimension(ndim), intent(in) :: vec

      tmp = 0.0_dop
      DO i = 1,ndim
         tmp = tmp + (abs(vec(i))/wt(i))**2
      ENDDO
      wgtnrm = DSQRT(tmp/dble(ndim))

      RETURN
      END FUNCTION wgtnrm                                  ! of WGTNRM

!***********************************************************************
!
! Function WGTDST
!
! computes the weighted distance of two vectors
!
!***********************************************************************
!
! INPUT:
! ------
! ndim       INTEGER     dimension of the vectors
! vec1,vec2  COMPLEX*16  vectors, whose distance is to be computed
! wt         REAL*8      vector of weights
!
! OUTPUT :
! --------
! wgtdst     REAL*8      function value
!
!***********************************************************************

      REAL(dop) FUNCTION wgtdst(ndim, vec1, vec2, wt)

      IMPLICIT NONE

      INTEGER(long)                             :: i
      INTEGER(long), intent(in)                 :: ndim
      REAL(dop)                                 :: tmp
      REAL(dop), dimension(ndim), intent(in)    :: wt
      COMPLEX(dop), dimension(ndim), intent(in) :: vec1,vec2

      tmp = 0.0_dop
      DO i = 1,ndim
         tmp = tmp + (abs(vec1(i)-vec2(i))/wt(i))**2
      ENDDO
      wgtdst = DSQRT(tmp/dble(ndim))

      RETURN
      END FUNCTION wgtdst                    ! of WGTDST


      end module odeuti
