     module rklib

use decimal, only: long,dop
use logdat

      implicit none

      integer(long), save :: rk5steps,rk5accpt,rk5rejct,rk5fncns, &
                             rk8steps,rk8accpt,rk8rejct,rk8fncns

      private
      public :: rk5,rk8,rk5f
      public :: rk5steps,rk5accpt,rk5rejct,rk5fncns, &
                rk8steps,rk8accpt,rk8rejct,rk8fncns

      contains

!***********************************************************************
!
! SUBROUTINE RK5
!
! MCTDH interface to the call of the Runge-Kutta (order 5) integrator
! from the ODESIM project.
!
!***********************************************************************
!
! Input:
!     psi         : initial values of the dynamic variables
!     dtpsi       : initial time derivative of the dyn. vars
!     psidim      : number of dyn. vars
!     intperiod   : length of integration interval
!     abstime     : initial time
!     initstep    : initial stepsize
!     accuracy    : parameter controlling the accuracy
!     auxpsi      : storage for intermediate results, workspace...
!                   size: 7*psidim; can be uninitialized
!     func        : routine for evaluating the RHS
!     mc,mr,mi,ml : the MCTDH-arrays
!     lpost       : whether to call a post-processor
!     postfn      : post-processing routine
!     ltest       : whether to call a test routine at end of step
!     testfn      : test routine
!
! Output:
!     psi         : final values of the dyn. vars
!     dtpsi       : undetermined (i.e. it gets overwritten!)
!     initstep    : optimal stepsize at end of integration
!     error       : errorflag; 0 means OK (currently it's always 0!)
!
! FO, June 2002
!
!***********************************************************************

      subroutine rk5(psi,dtpsi,psidim,intperiod,abstime,initstep,minstep, &
                     accuracy,error,auxpsi,func,lpost,postfn,ltest,testfn, &
                     lerrfn,errfn)

use decimal, only:long,dop
use global
use datenmod

use odemod_interface

      implicit none

      integer(long), intent(in)                        :: psidim
      integer(long), intent(out)                       :: error
      real(dop), intent(in)                            :: intperiod,abstime,&
                                                          accuracy
      real(dop), intent(inout)                         :: initstep
      real(dop), intent(in)                            :: minstep
      real(dop), dimension(psidim)                     :: rk5wt
      complex(dop), dimension(psidim), intent(inout)   :: psi,dtpsi
      complex(dop), dimension(psidim*7), intent(inout) :: auxpsi
      external                                         :: func,postfn,testfn,&
                                                          errfn
      logical(kind=4), intent(in)                      :: lpost,ltest,lerrfn

      real(dop)           :: tend,hmax,h,hopt,rerr,time
      integer(long)       :: intgfl,ierr

      real(dop), parameter  :: safe = 0.85_dop,&  ! 0.9d0   HDM, May 2021
                               beta=0.04_dop,&
                               facdec=0.15_dop,&  ! 0.25d0  HDM, May 2021
                               facinc=10.0_dop    ! 8.d0    HDM, May 2021
      logical(kind=4), save :: initialized = .false.

!     initialize statistics on first call
      if (.not. initialized) then
         rk5steps = 0
         rk5accpt = 0
         rk5rejct = 0
         rk5fncns = 0
         initialized = .true.
      endif

!     set parameters for call to core integrator
      h       = initstep
      time    = abstime
      tend    = abstime+intperiod
      hmax    = out1
      hopt    = 0.0_dop
      intgfl  = 0

!     call core integrator with appropriate post-processor
      call ODP5CR(&
           func, psidim, psi, time, tend, hmax, minstep, h, hopt, accuracy,&
           macheps, safe, beta, facdec, facinc,&
           auxpsi(1), auxpsi(psidim+1), dtpsi, auxpsi(2*psidim+1),&
           auxpsi(3*psidim+1), auxpsi(4*psidim+1), auxpsi(5*psidim+1),&
           auxpsi(6*psidim+1), rk5wt, rk5fncns, rk5steps,&
           rk5accpt, rk5rejct, intgfl, ierr, rerr,&
           lpost, postfn, ltest, testfn, lerrfn, errfn)

!     check error condition
      if (intgfl .ne. 1) then
         routine = 'RK5'
         write (message,100) ierr,rerr
         call errormsg
      else
         error = 0
      endif

!     return the optimal stepsize, but only if the integrator estimated
!     one, which doesn't happen if it finishes in one step
      if (hopt.gt. 0.0_dop) initstep = hopt

      return

 100  format ('Error number ',i2,', parameter ',f16.10)

      end subroutine rk5 

!***********************************************************************
!
! SUBROUTINE RK8
!
! MCTDH interface to the call of the Runge-Kutta (order 8) integrator
! from the ODESIM project.
!
!***********************************************************************
!
! Input:
!     psi         : initial values of the dynamic variables
!     dtpsi       : initial time derivative of the dyn. vars
!     psidim      : number of dyn. vars
!     intperiod   : length of integration interval
!     abstime     : initial time
!     initstep    : initial stepsize
!     accuracy    : parameter controlling the accuracy
!     auxpsi      : storage for intermediate results, workspace...
!                   size: 12*psidim; can be uninitialized
!     func        : routine for evaluating the RHS
!     mc,mr,mi,ml : the MCTDH-arrays
!     lpost       : whether to call a post-processor
!     postfn      : post-processing routine
!
! Output:
!     psi         : final values of the dyn. vars
!     dtpsi       : undetermined (i.e. it gets overwritten!)
!     initstep    : optimal stepsize at end of integration
!     error       : errorflag; 0 means OK (currently it's always 0!)
!
! FO, 08/2002
!
!***********************************************************************

      subroutine rk8(psi,dtpsi,psidim,intperiod,abstime,initstep, &
                     accuracy,error,auxpsi,func,lpost,postfn,ltest,testfn,&
                     lerrfn,errfn)

use decimal, only:long,dop
use global
use datenmod

use odemod_interface

      implicit none

      integer(long), intent(in)                         :: psidim
      integer(long), intent(out)                        :: error
      real(dop), intent(in)                             :: intperiod,abstime,accuracy
      real(dop), intent(inout)                          :: initstep
      real(dop), dimension(psidim)                      :: rk8wt
      complex(dop), dimension(psidim), intent(inout)    :: psi,dtpsi
      complex(dop), dimension(psidim*12), intent(inout) :: auxpsi
      external                                          :: func, postfn, testfn,&
                                                           errfn
      logical(kind=4), intent(in)                       :: lpost,ltest,lerrfn


      real(dop)           :: tend,hmax,h,hopt,rerr,time
      integer(long)       :: intgfl,ierr

      real(dop), parameter  :: safe = 0.85_dop,&   ! 0.9d0   HDM, May 2021
                               beta=0.04_dop,&
                               facdec=0.1_dop,&
                               facinc=8.0_dop      ! 6.d0    HDM, May 2021
      logical(kind=4), save :: initialized = .false.


!     initialize statistics on first call
      if (.not. initialized) then
         rk8steps = 0
         rk8accpt = 0
         rk8rejct = 0
         rk8fncns = 0
         initialized = .true.
      endif

!     set parameters for call to core integrator
      h       = initstep
      time    = abstime
      tend    = abstime+intperiod
      hmax    = out1
      hopt    = 0.0_dop
      intgfl  = 0

!     call core integrator
      call OD85CR( &
           func, psidim, psi, time, tend, hmax, h, hopt, accuracy, &
           macheps, safe, beta, facdec, facinc, &
           auxpsi(1), auxpsi(psidim+1), auxpsi(2*psidim+1), &
           auxpsi(3*psidim+1), dtpsi, auxpsi(4*psidim+1), &
           auxpsi(5*psidim+1), auxpsi(6*psidim+1), auxpsi(7*psidim+1), &
           auxpsi(8*psidim+1), auxpsi(9*psidim+1), auxpsi(10*psidim+1), &
           auxpsi(11*psidim+1), rk8wt, rk8fncns, rk8steps, &
           rk8accpt, rk8rejct, intgfl, ierr, rerr, &
           lpost, postfn, ltest, testfn, lerrfn, errfn)

!     check error condition
      if (intgfl .ne. 1) then
         routine = 'RK8'
         write (message,100) ierr,rerr
         call errormsg
      else
         error = 0
      endif

!     return the optimal stepsizebut only if the integrator estimated
!     one, which doesn't happen if it finishes in one step.
      if (hopt .gt. 0.d0) initstep = hopt

      return

 100  format ('Error number ',i2,', parameter ',f16.10)

      end subroutine rk8

!***********************************************************************
!
! SUBROUTINE RK5F HEREHEREHERE 16/11/17
!
! MCTDH interface to the call of the Runge-Kutta (order 5) FIXED STEPSIZE 
! integrator from the ODESIM project.
!
!***********************************************************************
!
! Input:
!     psi         : initial values of the dynamic variables
!     dtpsi       : initial time derivative of the dyn. vars
!     psidim      : number of dyn. vars
!     intperiod   : length of integration interval
!     abstime     : initial time
!     initstep    : initial stepsize
!     accuracy    : parameter controlling the accuracy
!     auxpsi      : storage for intermediate results, workspace...
!                   size: 7*psidim; can be uninitialized
!     func        : routine for evaluating the RHS
!     mc,mr,mi,ml : the MCTDH-arrays
!     lpost       : whether to call a post-processor
!     postfn      : post-processing routine
!     ltest       : whether to call a test routine at end of step
!     testfn      : test routine
!
! Output:
!     psi         : final values of the dyn. vars
!     dtpsi       : undetermined (i.e. it gets overwritten!)
!     initstep    : optimal stepsize at end of integration
!     error       : errorflag; 0 means OK (currently it's always 0!)
!
! FO, June 2002
!
!***********************************************************************

      subroutine rk5f(psi,dtpsi,psidim,intperiod,abstime,initstep,minstep, &
                     accuracy,error,auxpsi,func,lpost,postfn,ltest,testfn)

use decimal, only:long,dop
use global
use datenmod

use odemod_interface

      implicit none

      integer(long), intent(in)                        :: psidim
      integer(long), intent(out)                       :: error
      real(dop), intent(in)                            :: intperiod,abstime,&
                                                          accuracy
      real(dop), intent(inout)                         :: initstep
      real(dop), intent(in)                            :: minstep
      real(dop), dimension(psidim)                     :: rk5wt
      complex(dop), dimension(psidim), intent(inout)   :: psi,dtpsi
      complex(dop), dimension(psidim*7), intent(inout) :: auxpsi
      external                                         :: func,postfn,testfn
      logical(kind=4), intent(in)                      :: lpost,ltest

      real(dop)           :: tend,hmax,h,hopt,rerr,time
      integer(long)       :: intgfl,ierr

      logical(kind=4), save :: initialized = .false.

!     initialize statistics on first call
      if (.not. initialized) then
         rk5steps = 0
         rk5accpt = 0
         rk5rejct = 0
         rk5fncns = 0
         initialized = .true.
      endif

!     set parameters for call to core integrator
      h       = initstep
      time    = abstime
      tend    = abstime+intperiod
      hmax    = out1
      hopt    = 0.0_dop
      intgfl  = 0

!     call core integrator with appropriate post-processor
      call ODP5CRF(&
           func, psidim, psi, time, tend, hmax, minstep, h, hopt, accuracy,&
           macheps, &
           auxpsi(1), auxpsi(psidim+1), dtpsi, auxpsi(2*psidim+1),&
           auxpsi(3*psidim+1), auxpsi(4*psidim+1), auxpsi(5*psidim+1),&
           auxpsi(6*psidim+1), rk5wt, rk5fncns, rk5steps,&
           rk5accpt, intgfl, ierr, rerr, lpost, postfn, ltest, testfn)

!     check error condition
      if (intgfl .ne. 1) then
         routine = 'RK5F'
         write (message,100) ierr,rerr
         call errormsg
      else
         error = 0
      endif

!     return the optimal stepsize, but only if the integrator estimated
!     one, which doesn't happen if it finishes in one step
      if (hopt.gt. 0.0_dop) initstep = hopt

      return

 100  format ('Error number ',i2,', parameter ',f16.10)

      end subroutine rk5f 

      end module rklib
