
! **************************************************************************
! *                                                                        *
! *                  SHORT ITERATIVE LANCZOS (sillib.f)                    *
! *                                                                        *
! * Library module containing a short iterative Lanczos integrator.        *
! *                                                                        *
! * CONTAINS:                                                              *
! *    SILStep:      The (real) Lanczos integration routine.                *
! *    SILErrorMsg:  Returns for a given error number a corresponding       *
! *                  error message.                                         *
! *    SILDiag:      The (real) Lanczos diagonalisation routine.            *
! *                                                                        *
! **************************************************************************

      module sillib

use decimal, only: long, dop
use op1lib,  only: cpvxd 

      Implicit None

      private
      public                   :: silstep,sildiag,silerrormsg
      contains


! --------------------------------------------------------------------------
!                            SUBROUTINE SILSTEP                                 
!                                                                    
! Integrates a system of complex linear first order differential equations     
! with constant and Hermitian Hamiltonian employing the short iterative   
! Lanczos method. The routine runs with both variable step size and order. 
! (First it increases the order to achieve the desired accuracy. if this 
! fails within the maximum order, the stepsize is reduced.) SILStep makes 
! only one single integration step, so it has to be imbedded into a loop 
! that calls SILStep until the desired time interval is integrated. The ODE 
! is of the form i dPsi/dt = H|Psi> = Func(Time,Psi,DtPsi) =: DtPsi or  
! dPsi/dt = -i H|Psi> = Func(Time,Psi,DtPsi) =: DtPsi, depending on the 
! flag "StdForm". All computations are performed with double precision.
!                                                                    
! INPUT PARAMETERS:                                                  
!    Psi:       The (complex) initial-value vector.                   
!    DtPsi:     Action of Hamiltonian on the initial-value vector, i.e.
!               H|Psi> or -i H|Psi>, depending on "StdForm".     
!    PsiDim     Length of Psi and DtPsi vectors.                      
!    IntPeriod: Length of time interval to be integrated.             
!    IntOrder:  Maximum integration order.                            
!    TolError:  Maximum error that is tolerated.                      
!    Relax:     Flag for relaxation calculation. if true, Psi is relaxed,
!               else propagated.                           
!    Restart:   Flag for restarting the integrator. if true, the Krylov
!               space is built up before propagation, else the old Krylov
!               vectors are used.                          
!    StdForm:   if true, Func(Time,Psi,DtPsi) = -i H|Psi>, else       
!               Func(Time,Psi,DtPsi) = H|Psi>.                        
!    Steps:     Number of steps made so far (Steps is passed to "WriteStep").          
!              
!    Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with columns
!               containing the Krylov vectors H^2|psi> to H^TrueOrder|psi>. 
!               The Krylov vectors are needed on entry only when Restart is 
!               false.    
!                                                                    
! OUTPUT PARAMETERS:                                                 
!    Psi:       Propagated Psi.                                       
!    DtPsi:     Normalised first Krylov vector, i. e. (H-<H>)|Psi>.   
!    Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with columns 
!               containing the Krylov vectors H^2|psi> to H^TrueOrder|psi>.
!    Stepsize:  Time interval that actually has been integrated (can be 
!               lower than IntPeriod).                             
!    TrueOrder: The order that has actually been used (may be less    
!               than IntOrder).                                       
!    Steps:     Same as on entry.                                     
!    ErrorCode: Error code having the following meaning:              
!               0: everything was o.k.,                              
!               1: illegal integration order,                         
!               2: stepsize underflow,                                
!               3: diagonalization failed,                            
!               4: stepsize too large for restart.                    
!                                                                     
! EXTERNAL ROUTINES:                                                 
!    Func:      Computes the action of the Hamiltonian on Psi.        
!               Called as                                             
!               Func(AbsTime,Psi,DtPsiR)                              
!    WriteStep: Writes the stepsize and error to a file.              
!               Called as WriteStep(Steps,Order,Stepsize,Error).      
!                                                                    
! SOME NEW VARIABLES:                                                
!    rlaxnum: This is the argument of the keyword relaxation. if relaxation
!             has no argument, rlaxnum=-1 .             
!    Relax:   Logical set true for a relaxation run.                  
!                                                                    
! V6.0 MB                                                            
! V8.2 05/01 HDM  Addition of 'relaxation' to exited states.         
!                 Eigenvector-space now dynamically allocated.       
!---------------------------------------------------------------------------

      subroutine SILStep(Psi,DtPsi,PsiDim,IntPeriod,IntOrder,TolError,&
                         Relax,Restart,StdForm,Steps,Krylov,Stepsize,&
                         TrueOrder,ErrorCode,Time,Func,&
                         EigenVector,EigenVal,Diagonal,OffDg2,OffDiag)

use wrintegrat, only: writestep 

      Implicit None

      real(dop), parameter     :: RelativeMinStep = 1.0e-10_dop,tiny = 1.0e-18_dop

      external                                                     :: Func
      logical(kind=4), intent(in)                                  :: Relax,Restart,&
                                                                      StdForm
      integer(long), intent(in)                                    :: PsiDim,IntOrder
      integer(long), intent(inout)                                 :: Steps
      integer(long), intent(out)                                   :: TrueOrder,ErrorCode
      complex(dop), dimension(PsiDim), intent(inout)               :: Psi,DtPsi
      complex(dop), dimension(PsiDim,2:IntOrder), intent(inout)    :: Krylov
      real(dop), dimension(0:IntOrder,0:IntOrder+2), intent(inout) :: EigenVector
      real(dop), dimension(0:IntOrder), intent(out)                :: Diagonal,EigenVal
      real(dop), dimension(IntOrder), intent(out)                  :: OffDiag
      real(dop), dimension(IntOrder+1), intent(out)                :: OffDg2
      real(dop), intent(in)                                        :: IntPeriod,&
                                                                      TolError,Time
      real(dop), intent(out)                                       :: Stepsize

      integer(long), save      :: OldOrder
      real(dop), save          :: OldError,OldStepsize
      
      integer(long)            :: D,P,Q
      complex(dop)             :: Alpha,CBeta,PreFactor,CInverse,sum
      real(dop)                :: Beta,Error,MinStepsize,beta1
                                                                      
! --- CHECK INTEGRATION ORDER ---

      if ( IntOrder .lt. 2 ) then
         ErrorCode = 1
         write(6,'(a,i4,a)') 'IntOrder =',IntOrder,' is too small!'
         return
      endif

! --- INITIALIZE VARIABLES ---

      Stepsize = IntPeriod
      MinStepsize = abs(RelativeMinStep*IntPeriod)
      ErrorCode = 0
      if (.not.StdForm) then
         PreFactor = (1.0_dop,0.0_dop)
      elseif (Relax) then
         PreFactor = (-1.0_dop,0.0_dop)
      else
         PreFactor = (0.0_dop,-1.0_dop)
      endif

! --- RETURN RESULT IF DIFFERENTIAL EQUATION IS NOT A SYSTEM ---

      if (PsiDim .Eq. 1) then
         TrueOrder = 1
         Error = 0.0_dop
         call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)
         Psi(1) = exp(StepSize*DtPsi(1)/Psi(1))*Psi(1)
         return
      endif

! --- SKIP CALCULATION OF KRYLOV VECTORS IF DESIRED ---

      if (Restart) then
         OldError=0.0_dop
         OldOrder=0
         OldStepsize=0.0_dop
      else
         if (abs(Stepsize) .gt. abs(OldStepsize)+MinStepsize) then
            ErrorCode = 4
            return
         endif
         TrueOrder = OldOrder
         Error = abs(OldError*(Stepsize/OldStepsize)**TrueOrder)
         call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)
         goto 200
      endif

! --- FURTHER INITIALIZE VARIABLES ---

      Alpha = (0.0_dop,0.0_dop)
      Beta = 0.0_dop
      TrueOrder = 1

! --- COMPUTE FIRST DIAGONAL ELEMENT ---

      do D = 1,PsiDim
         Alpha = Alpha+DConjg(Psi(D))*DtPsi(D)
      enddo
      Diagonal(0) = dble(Alpha/PreFactor)

! --- DETERMINE CORRESPONDING BASIS VECTOR AND OFF-DIAGONAL ELEMENT ---

      do D = 1,PsiDim
         DtPsi(D) = DtPsi(D)-Alpha*Psi(D)
         Beta = Beta+dble(DConjg(DtPsi(D))*DtPsi(D))
      enddo

      Beta = sqrt(Beta)
      OffDiag(1) = Beta
      beta1 = Beta
      CBeta = PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      if (Beta .lt. 1.0e-20_dop) then
         Beta = 0.0_dop
         CInverse = (0.0_dop,0.0_dop)
      else
         CInverse = 1.0_dop/CBeta
      endif
      do D = 1,PsiDim
         DtPsi(D) = CInverse*DtPsi(D)
      enddo

! --- COMPUTE ERROR ESTIMATE ---

      Error = abs(Beta*Stepsize)

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)

! --- BUILD UP THE KRYLOV SPACE ---
 
 100  Continue
        
! --- CHECK IF STEPSIZE IS TOO SMALL ---

      if (abs(Stepsize) .lt. abs(MinStepSize)) then
         ErrorCode = 2
         return
      endif

! --- RE-INITIALIZE VARIABLES ---

      Alpha = (0.0_dop,0.0_dop)
      Beta = 0.0_dop
      TrueOrder = TrueOrder+1


! --- EVALUATE FUNCTION WITH LAST BASIS VECTOR ---

      if (TrueOrder .eq. 2) then
         call Func(Time,DtPsi,Krylov(1,2))
      else
         call Func(Time,Krylov(1,TrueOrder-1),Krylov(1,TrueOrder))
      endif

! --- COMPUTE DIAGONAL ELEMENT ---

! --- Note that the Krylov vectors number zero and one aren't stored in ---
! --- "Krylov" but in "Psi" and "DtPsi", respectively. ---

      if (TrueOrder .Eq. 2) then
         do D = 1,PsiDim
            Alpha = Alpha+DConjg(DtPsi(D))*Krylov(D,2)
         enddo
      else
         do D = 1,PsiDim
            Alpha = Alpha+DConjg(Krylov(D,TrueOrder-1))&
                    *Krylov(D,TrueOrder)
         enddo
      endif
      Diagonal(TrueOrder-1) = dble(Alpha/PreFactor)

! --- COMPUTE OFF-DIAGONAL ELEMENT AND BASIS VECTOR ---

      if (TrueOrder .eq. 2) then
         do D = 1,PsiDim
            Krylov(D,2) = Krylov(D,2)-Alpha*DtPsi(D)-CBeta*Psi(D)
            Beta = Beta+dble(DConjg(Krylov(D,2))*Krylov(D,2))
         enddo
      elseif (TrueOrder .Eq. 3) then
         do D = 1,PsiDim
            Krylov(D,3) = Krylov(D,3)-Alpha*Krylov(D,2)-CBeta*DtPsi(D)
            Beta = Beta+dble(DConjg(Krylov(D,3))*Krylov(D,3))
         enddo
      else
         do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder)&
               -Alpha*Krylov(D,TrueOrder-1)-CBeta*Krylov(D,TrueOrder-2)
            Beta = Beta+dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
         enddo
      endif

      Beta = sqrt(Beta)
      OffDiag(TrueOrder) = Beta
      CBeta = PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      if (Beta .lt. 1.0e-20_dop) then
         OffDiag(TrueOrder) = 0.0_dop
         Beta = 0.0_dop
         CInverse = (0.0_dop,0.0_dop)
      else
         CInverse = 1.0_dop/CBeta
      endif
      
      do D = 1,PsiDim
         Krylov(D,TrueOrder) = CInverse*Krylov(D,TrueOrder)
      enddo

! --- COMPUTE ERROR ESTIMATE ---
      
      Error = abs(Error*Beta*Stepsize/dble(TrueOrder))

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)

! --- CONTINUE IF NOT CONVERGED ---
      if((TrueOrder.lt.IntOrder) .and. (Error.gt.TolError)&
         .and. (Beta.gt.1.0e-20_dop) )  goto 100
            
! --- ESTIMATE NEXT MATRIX ELEMENT ---

      Diagonal(TrueOrder) = Diagonal(TrueOrder-1)
!      Diagonal(TrueOrder) = Diagonal(TrueOrder-1) + beta

! --- DECREASE STEPSIZE IF LANCZOS DIDN'T CONVERGE ---
      
      if (Error .gt. TolError) then
         Stepsize = Stepsize*(TolError/Error)**(1.0_dop/dble(TrueOrder))
         call WriteStep(Steps,TrueOrder,Stepsize,TolError,Time)
      endif


! --- SAVE SOME VALUES IN CASE THAT NEXT CALL IS NON-RESTART CALL ---

      OldOrder = TrueOrder
      OldStepsize = Stepsize
      OldError = Error
    
! --- Save Diagonal and off diagonal values since the diagonalisation ---
! --- routine will overwrite those. ---

      call  cpvxd(Diagonal,EigenVal,TrueOrder+1)
      call  cpvxd(OffDiag,OffDg2,TrueOrder)

! --- DIAGONALIZE LANCZOS MATRIX --- CALL LAPACK ROUTINE --- 
! --- EigenVector(*,IntOrder+1) serves as work array 
      call DSTEQR('I',TrueOrder+1,EigenVal,OffDg2,EigenVector,&
                  IntOrder+1,EigenVector(1,IntOrder+1),ErrorCode)

      if (ErrorCode .ne. 0) then
         ErrorCode = 3
         return
      endif


 200  Continue                  ! Jump to here, if not RESTART.

! --- PROPAGATE WAVEFUNCTION ---

! --- Note again that the Krylov vectors number zero and one aren't stored ---
! --- in "Krylov" but in "Psi" and "DtPsi", respectively. ---

      do P = 0,TrueOrder
         sum = (0.0_dop,0.0_dop)
         do Q = 0,TrueOrder
            sum = sum+exp(PreFactor*EigenVal(Q)*Stepsize)*&
                 EigenVector(0,Q)*EigenVector(P,Q)
         enddo
         if (P .eq. 0) then
            do D = 1,PsiDim
               Psi(D) = sum*Psi(D)
            enddo
         elseif (P .eq. 1) then
            do D = 1,PsiDim
               Psi(D) = Psi(D)+sum*DtPsi(D)
            enddo
         else
            do D = 1,PsiDim
               Psi(D) = Psi(D)+sum*Krylov(D,P)
            enddo
         endif
      enddo

      return
      End Subroutine SILStep


! --------------------------------------------------------------------------
!                       SUBROUTINE SILERRORMSG                       
!                                                                    
! Generates, for a given error number returned by "SILStep/SILStep2/
! SILStep3", a corresponding error message.                           
!                                                                    
! INPUT PARAMETERS:                                                  
!    Error: Error code returned by SILStep/SILStep2                  
!                                                                    
! OUTPUT PARAMETERS:                                                 
!    Msg:   Error message                                            
!                                                                    
! V7.0 MB                                                            
! --------------------------------------------------------------------------

      subroutine silerrormsg (Error,Msg)

      Implicit None

      character(len=*), intent(out) :: Msg
      integer(long), intent(in)     :: Error

! --- GENERATE ERROR MESSAGE ---

      if (Error .eq. 1) then
         Msg = 'Illegal integration order'
      elseif (Error .eq. 2) then
         Msg = 'Stepsize underflow'
      elseif (Error .eq. 3) then
         Msg = 'Diagonalisation failed'
      elseif (Error .eq. 4) then
         Msg = 'Illegal initial stepsize'
      else
         Msg = 'Unknown error occurred'
      endif

      return
      end subroutine silerrormsg


! --------------------------------------------------------------------------
!                            SUBROUTINE SILDIAG                                 
!                                                                    
! Diagonalises a (complex) Hermitian matrix by Lanczos-iterations.   
!                                                                    
! INPUT PARAMETERS:                                                  
!    Psi:       The (complex) initial-value vector.                   
!    DtPsi:     Action of Hamiltonian on the initial-value vector,    
!               i.e. H|Psi> or -i H|Psi>, depending on "StdForm".     
!    PsiDim     Length of Psi and DtPsi vectors.                      
!    IntOrder:  Maximum integration order.                            
!    TolError:  Maximum error that is tolerated.                      
!    StdForm:   if true, Func(Time,Psi,DtPsi) = -i H|Psi>, else       
!               Func(Time,Psi,DtPsi) = H|Psi>.                        
!    Steps:     Number of steps made so far (Steps is passed to       
!               "WriteStep").                                         
!    Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with columns
!               containing the Krylov vectors H^2|psi> to H^TrueOrder|psi>.
!    TrueOrder: On input: the minimal order to be used.               
!    ortho:     If .true., then a full re-orthogonalisation of the Krylov
!               vectors is performed.                          
!                                                                    
! OUTPUT PARAMETERS:                                                 
!    Psi:       Computed eigenvector. The number of the eigenvector to be 
!               chosen is specified by rlaxnum.                 
!    DtPsi:     Normalised first Krylov vector, i. e. (H-<H>)|Psi>.   
!    Krylov:    Matrix with minimum size PsiDim*(IntOrder-1) with columns
!               containing the Krylov vectors H^2|psi> to H^TrueOrder|psi>. 
!    TrueOrder: The order that has actually been used (may be less than
!               IntOrder).                                       
!    Steps:     Same as on entry.                                     
!    ErrorCode: Error code having the following meaning:              
!               0: everything was o. k.,                              
!               1: illegal integration order,                         
!               3: diagonalization failed,                            
!                                                                     
! EXTERNAL ROUTINES:                                                 
!    Func:      Computes the action of the Hamiltonian on Psi.        
!               Called as                                             
!                 Func(AbsTime,Psi,DtPsi)                               
!    WriteStep: Writes the stepsize and error to a file.              
!               Called as 
!                 WriteStep(Steps,Order,Stepsize,Error).      
!                                                                    
! SOME NEW VARIABLES:                                                
!     rlaxnum : This is the argument of the keyword relaxation.         
!               if relaxation has no argument, rlaxnum=-1 .             
!                                                                    
! V8.3 10/03 HDM                                                     
! --------------------------------------------------------------------------

      subroutine SILdiag(Psi,DtPsi,PsiDim,IntOrder,TolError,StdForm,&
                         Steps,Krylov,TrueOrder,ErrorCode,rlaxnum,Time,&
                         ortho,Func,EigenVector,EigenVal,Diagonal,&
                         OffDg2,OffDiag)

use wrintegrat, only: writestep,wrtrlx

      Implicit None

      external                                       :: Func
      logical(kind=4), intent(in)                    :: StdForm,ortho
      integer(long), intent(in)                      :: PsiDim,IntOrder,&
                                                        rlaxnum
      integer(long), intent(inout)                   :: Steps
      integer(long), intent(out)                     :: TrueOrder,ErrorCode
      complex(dop), dimension(PsiDim), intent(inout) :: Psi,DtPsi
      complex(dop), dimension(PsiDim,2:IntOrder), intent(inout)    :: Krylov
      real(dop), dimension(0:IntOrder,0:IntOrder+2), intent(inout) :: &
                                                        EigenVector
      real(dop), dimension(0:IntOrder), intent(out)  :: Diagonal,EigenVal
      real(dop), dimension(IntOrder), intent(out)    :: OffDiag
      real(dop), dimension(IntOrder+1), intent(out)  :: OffDg2
      real(dop), intent(in)                          :: TolError,Time

      logical(kind=4), save    :: full
      logical(kind=4), save    :: FirstCall = .true.
      integer(long), save      :: minorder,relaxnum
      real(dop), save          :: Tinit
      
      integer(long)            :: D,P,Q,i,nd
      complex(dop)             :: Alpha,CBeta,PreFactor,CInverse,ovlp
      real(dop)                :: Beta,Error,mxvc,err1,beta1,beta2
      real(dop)                :: Stepsize


! --- CHECK INTEGRATION ORDER ---

      if ( IntOrder .lt. 2 ) then
         ErrorCode = 1
         write(6,'(a,i4,a)') 'IntOrder =',IntOrder,' is too small!'
         return
      endif

! --- INITIALIZE VARIABLES ---

      nd = 2
      Stepsize = 0.0_dop
      ErrorCode = 0
      full = .false.
      if(FirstCall) then
         Tinit = Time
         
         if(rlaxnum .ge. 9999) then
            full = .true.
            relaxnum = rlaxnum - 10000
            nd = IntOrder+1
         else
            relaxnum = rlaxnum
            full = .false.
         endif
         if (relaxnum .lt. 0) then
            write(6,*) ' relaxnum =',relaxnum, ' is .lt. 0. STOP'
            stop
         endif
         if (.not.StdForm) then  ! This is for meigenf
            minorder  = min(TrueOrder,IntOrder)
         else
            minorder  = 0
         endif
      endif
      if (.not.StdForm) then
         PreFactor = (1.0_dop,0.0_dop)
      else
         PreFactor = (-1.0_dop,0.0_dop)
      endif

! --- FURTHER INITIALIZE VARIABLES ---

      Alpha = (0.0_dop,0.0_dop)
      Beta = 0.0_dop
      TrueOrder = 1

! --- COMPUTE FIRST DIAGONAL ELEMENT ---

      do D = 1,PsiDim
         Alpha = Alpha+DConjg(Psi(D))*DtPsi(D)
      enddo 
      Diagonal(0) = dble(Alpha/PreFactor)

! --- DETERMINE CORRESPONDING BASIS VECTOR AND OFF-DIAGONAL ELEMENT ---

      do D = 1,PsiDim
         DtPsi(D) = DtPsi(D)-Alpha*Psi(D)
         Beta = Beta+dble(DConjg(DtPsi(D))*DtPsi(D))
      enddo 

! --- Re-orthogonalize DtPsi ---
      if(ortho) then
         ovlp = (0.0_dop,0.0_dop)
         do D = 1,PsiDim
            ovlp = ovlp + DConjg(Psi(D))*DtPsi(D)
         enddo
         Beta = 0.0_dop
         do D = 1,PsiDim
            DtPsi(D)=DtPsi(D)-ovlp*Psi(D)
            Beta = Beta+dble(DConjg(DtPsi(D))*DtPsi(D))
         enddo
      endif

      Beta = sqrt(Beta)
      OffDiag(1) = Beta
      beta1 = Beta
      CBeta = PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      if (Beta .lt. 1.0e-20_dop) then 
         Beta = 0.0_dop
         CInverse = (0.0_dop,0.0_dop)
      else
         CInverse = 1.0_dop/CBeta
      endif 
      do D = 1,PsiDim
         DtPsi(D) = CInverse*DtPsi(D)
      enddo 

! --- COMPUTE ERROR ESTIMATE ---

      Error = Beta 

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)

! --- Check for vanishing operator.
      If(abs(Alpha)+Beta .lt. 1.d-30) then
         write(6,'(/a)') 'SILdiag: Zero operator!'
         write(2,'(/a)') 'SILdiag: Zero operator!'
         stop 1
      EndIf

! --- BUILD UP THE KRYLOV SPACE ---
 
 100  Continue
        
! --- RE-INITIALIZE VARIABLES ---

      beta2 = Beta
      Alpha = (0.0_dop,0.0_dop)
      Beta = 0.0_dop
      TrueOrder = TrueOrder+1


! --- EVALUATE FUNCTION WITH LAST BASIS VECTOR ---

      if (TrueOrder .eq. 2) then 
         call Func(Time,DtPsi,Krylov(1,2))
      else
         call Func(Time,Krylov(1,TrueOrder-1),Krylov(1,TrueOrder))
      endif 

! --- COMPUTE DIAGONAL ELEMENT ---

! --- Note that the Krylov vectors number zero and one aren't stored in ---
! --- "Krylov" but in "Psi" and "DtPsi", respectively. ---

      if (TrueOrder .eq. 2) then 
         do D = 1,PsiDim
            Alpha = Alpha+DConjg(DtPsi(D))*Krylov(D,2)
         enddo 
      else
         do D = 1,PsiDim
            Alpha = Alpha+DConjg(Krylov(D,TrueOrder-1))&
                    *Krylov(D,TrueOrder)
         enddo 
      endif 
      Diagonal(TrueOrder-1) = dble(Alpha/PreFactor)

! --- COMPUTE OFF-DIAGONAL ELEMENT AND BASIS VECTOR ---

      if (TrueOrder .eq. 2) then 
         do D = 1,PsiDim
            Krylov(D,2) = Krylov(D,2)-Alpha*DtPsi(D)-CBeta*Psi(D)
            Beta = Beta+dble(DConjg(Krylov(D,2))*Krylov(D,2))
         enddo 
      elseif (TrueOrder .eq. 3) then 
         do D = 1,PsiDim
            Krylov(D,3) = Krylov(D,3)-Alpha*Krylov(D,2)-CBeta*DtPsi(D)
            Beta = Beta+dble(DConjg(Krylov(D,3))*Krylov(D,3))
         enddo 
      else
         do D = 1,PsiDim
            Krylov(D,TrueOrder) = Krylov(D,TrueOrder)&
               -Alpha*Krylov(D,TrueOrder-1)-CBeta*Krylov(D,TrueOrder-2)
            Beta = Beta+dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
         enddo 
      endif 


! --- Re-Orthogonalize against Psi and DtPsi  ---
      ovlp =(0.0_dop,0.0_dop)
      do D = 1,PsiDim
         ovlp = ovlp + DConjg(Psi(D))*Krylov(D,TrueOrder)
      enddo
      do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*Psi(D)
      enddo
      ovlp =(0.0_dop,0.0_dop)
      do D = 1,PsiDim
         ovlp = ovlp + DConjg(DtPsi(D))*Krylov(D,TrueOrder)
      enddo
      do D = 1,PsiDim
         Krylov(D,TrueOrder)=Krylov(D,TrueOrder)-ovlp*DtPsi(D)
      enddo


! --- Re-Orthogonalize against Krylov  ---
      if((ortho.or.full) .and. TrueOrder .ge. 3) then
         do i = 2, TrueOrder-1
            ovlp =(0.0_dop,0.0_dop)
            do D = 1,PsiDim
               ovlp = ovlp + DConjg(Krylov(D,i))*Krylov(D,TrueOrder)
            enddo
            do D = 1,PsiDim
               Krylov(D,TrueOrder) = Krylov(D,TrueOrder) - &
                       ovlp*Krylov(D,i)
            enddo
         enddo

! --- Determine Beta ---
         Beta = 0.0_dop
         do D = 1,PsiDim
            Beta = Beta+dble(DConjg(Krylov(D,TrueOrder))*Krylov(D,TrueOrder))
         enddo
      endif

      Beta = sqrt(Beta)
      OffDiag(TrueOrder) = Beta
      CBeta = PreFactor*Beta

! --- NORMALIZE BASIS VECTOR ---

      if (Beta .lt. 1.0e-20_dop) then 
         OffDiag(TrueOrder) = 0.0_dop
         Beta = 0.0_dop
         CInverse = (0.0_dop,0.0_dop)
      else
         CInverse = 1.0_dop/CBeta
      endif 
      If (beta2+Beta .LT. 1.d-20) MinOrder=0
      
      do D = 1,PsiDim
         Krylov(D,TrueOrder) = CInverse*Krylov(D,TrueOrder)
      enddo 

! --- COMPUTE ERROR ESTIMATE ---
      
      Error = Error*Beta/(0.5d0*Abs(alpha-Diagonal(0))+Beta+1.d-30)

! --- WRITE STEPSIZE AND ERROR IF DESIRED ---

      call WriteStep(Steps,TrueOrder,Stepsize,Error,Time)

! --- CONTINUE IF NOT CONVERGED ---
      
      if((TrueOrder.lt.IntOrder) .and. (full.or.(Error.gt.TolError))&
         .and. Beta.gt.1.0e-20_dop .or. TrueOrder.lt.MinOrder )  goto 100
            
! --- ESTIMATE NEXT MATRIX ELEMENT ---

!     Diagonal(TrueOrder) = Diagonal(TrueOrder-1)
      Diagonal(TrueOrder) = Diagonal(TrueOrder-1) + beta


! --- SET nd TO AVOID EXCESSIVE TRIAL DIAGONALISATIONS ---
      
      if(TrueOrder.lt.IntOrder .and. &
           Beta.gt.1.0e-20_dop .and.  mod(TrueOrder,nd) .ne. 0 ) goto 100
      if(TrueOrder.ge. 20) nd = 5      !  This is to avoid excessive
      if(TrueOrder.ge.100) nd = 10     !  trial diagonalisations.
      if(TrueOrder.ge.200) nd = 20   
      if(TrueOrder.ge.300) nd = 40  
      if(TrueOrder.ge.500) nd = 80  

! --- Save Diagonal and off diagonal values since the diagonalisation ---
! --- routine will overwrite those. ---

      call  cpvxd(Diagonal,EigenVal,TrueOrder+1)
      call  cpvxd(OffDiag,OffDg2,TrueOrder)

! --- DIAGONALIZE LANCZOS MATRIX --- CALL LAPACK ROUTINE --- 
! --- EigenVector(*,IntOrder+1) serves as work array ---
      if(beta.lt.1.0e-18_dop) TrueOrder=TrueOrder-1
      call DSTEQR('I',TrueOrder+1,EigenVal,OffDg2,EigenVector,&
                  IntOrder+1,EigenVector(1,IntOrder+1),ErrorCode)

      if (ErrorCode .ne. 0) then 
         ErrorCode = 3
         Return
      endif 



! --- RELAX WAVEFUNCTION TO EIGENSTATE NO. Q ---
! --- SEARCH FOR THE EIGENVECTOR THAT IS CLOSEST TO PSI ---
      mxvc = 0.0_dop
      do P = 0,TrueOrder
         OffDg2(p+1) = EigenVector(0,P)**2
         if(OffDg2(p+1) .gt. mxvc ) then 
            mxvc = OffDg2(p+1)
            Q = P
         endif 
      enddo 
      
      if(time.eq.tinit .and. relaxnum.le.900) &
           Q = min(relaxnum,TrueOrder) 
      err1  = 1000.0_dop*Beta*EigenVector(TrueOrder,Q)**2
      if(.not.StdForm) &              ! for meigenf
       err1  = err1 + 100.0_dop*Beta*EigenVector(TrueOrder,minorder/4)**2&
                    + 100.0_dop*Beta*EigenVector(TrueOrder,minorder/2)**2 
      error = max(error,err1)
         
      if( TrueOrder.lt.IntOrder .and. Beta.gt.1.0e-20_dop &
           .and. err1.gt.TolError ) goto 100

! --- Build new A-vector. ---
      do D = 1,PsiDim
         Psi(D) = EigenVector(0,Q)*Psi(D)+EigenVector(1,Q)*DtPsi(D)
      enddo 
      do P = 2,TrueOrder
         do D = 1,PsiDim
            Psi(D) = Psi(D)+EigenVector(P,Q)*Krylov(D,P)
         enddo 
      enddo 
      
! --- Write to rlx_info file. EigenVector(*,IntOrder+1) is a work array, ---
! --- containing the error-estimates. --- 
      if(FirstCall) then
         do P = 0,TrueOrder
            EigenVector(p,IntOrder+1) = &
                 2.0_dop*27.2114_dop*Beta*EigenVector(TrueOrder,p)**2
!........The extra factor of 2 is due to the modification, line 358. ---
            if(EigenVector(p,IntOrder+1).lt.1.0e-30_dop)&
                 EigenVector(p,IntOrder+1) = 0.0_dop
         enddo
      else
         EigenVector(Q,IntOrder+1) =&
              2.0_dop*27.2114_dop*Beta*EigenVector(TrueOrder,Q)**2
      endif
      call wrtrlx(Q,TrueOrder,mxvc,beta1,time,EigenVal,&
           EigenVector(0,IntOrder+1),OffDg2,FirstCall,StdForm)
      if(StdForm) then
         FirstCall = .false.    ! Normal case, improved relaxation
      else
         steps = Q              ! This is for meigenf
         do p =0, IntOrder
            EigenVector(p,IntOrder+2)=EigenVal(p)
         enddo 
      endif

      return
      end subroutine SILdiag



      end module sillib
