
! ***************************************************************************
! *                                                                         *
! *                          SODLIB (sodlib.f)                              *
! *                                                                         *
! * Program package containing subroutines that propagate wave-packets      *
! * employing the second order differencing method.                         *
! *                                                                         *
! * CONTAINS:                                                               *
! *    sodstep: Propagates a wave-packet using the second order             *
! *             differencing (SOD) scheme.                                  *
! *    initsod: Determines a second initial vector for the SOD-             *
! *             integrator.                                                 *
! *    sodsize: Computes a stepsize for which the SOD-integrator is         *
! *             stable.                                                     *
! *                                                                         *
! ***************************************************************************

      module sodlib

      Implicit None

      integer, parameter       :: long  = selected_int_kind(9) 
      integer, parameter       :: dop = selected_real_kind(15,307)

      private
      public                   :: sodstep,initsod,sodsize
      contains  


! --------------------------------------------------------------------------
!                            SUBROUTINE SODSTEP
!
! PURPOSE:
!    Propagates a wave-packet one step further using the second order
!    differencing (SOD) scheme. The Hamiltonian may be time-dependent.
!    Since SOD is a two-step method, a second initial wavefunction has
!    to be determined employing subroutine "initsod", for instance.
! INPUT PARAMETERS:
!    psi:       Array containing wavefunction vector at current time step
!    oldpsi:    Array containing wavefunction vector at previous time step 
!    hpsi:      Array containing H|psi> vector of wavefunction psi
!    psidim:    Length of psi arrays
!    abstime:   Absolute time at start of interval
!    intperiod: (constant) The time period over which the interation
!               is performed
!    stepsize:  (constant) Stepsize used for integration (must be the
!               same as in "initsod")
!    hamilton:  Name of routine computing action of Hamiltonian on psi,
!               called as hamilton(abstime,psi,hpsi)
! OUTPUT PARAMETERS:
!    psi:     Array containing propagated wavefunction
!    oldpsi:  Array containing previous wavefunction
!    abstime: New absolute time
!    hpsi:    Array containing H|psi> vector of propagated psi
!
! --------------------------------------------------------------------------

      subroutine sodstep (psi,oldpsi,hpsi,psidim,abstime,intperiod,&
                          stepsize,hamilton)

      Implicit None

      complex(dop), parameter  :: imag=(0.0_dop,1.0_dop)
      real(dop), parameter     :: tiny = 1.0d-8

      external                                       :: hamilton
      integer(long), intent(in)                      :: psidim
      complex(dop), dimension(psidim), intent(inout) :: psi,hpsi,oldpsi
      real(dop), intent(in)                          :: intperiod,stepsize
      real(dop), intent(in)                          :: abstime

      integer(long)            :: d
      complex(dop)             :: swap,factor
      real(dop)                :: inttime,intstep

      inttime = 0.0d0
      intstep = stepsize

! --- START OF PROPAGATION LOOP ---
 100  continue

! --- PROPAGATE WAVEFUNCTION ONE STEP FURTHER ---

      factor = 2.0_dop*imag*intstep
      do d = 1,psidim
         swap = psi(d)
         psi(d) = oldpsi(d)-factor*hpsi(d)
         oldpsi(d) = swap
      enddo

! --- UPDATE TIME VARIABLE ---
      inttime = inttime+intstep

! --- CHECK WHETHER INTEGRATION IS COMPLETE ---
! --- RETURN TO CALLING ROUTINE WHEN FINISHED ---

      if (abs(inttime-intperiod) .lt. tiny) then
         return

      else

! --- COMPUTE ACTION OF HAMILTONIAN ON PROPAGATED WAVEFUNCTION ---

         call hamilton(abstime+inttime,psi,hpsi)

! --- CHECK STEPSIZE AND CONTINUE INTEGRATION ---
         if ((inttime+intstep) .gt. intperiod) then
            intstep = intperiod-inttime-0.01*tiny
         endif

         goto 100
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end  subroutine sodstep

! --------------------------------------------------------------------------
!                            SUBROUTINE INITSOD
!
! PURPOSE:
!    Propagates the initial wavefunction one step forward in order to
!    provide the (two-step) SOD method with a second initial wavefunction.
!    The same effective Hamiltonian as in "sod" is used, namely 
!    psi(t) = exp(-i arcsin(tH)) psi(0). 
!    NOTE - This routine works properly for time-independent Hamiltonian 
!           only.
! INPUT PARAMETERS:
!    psi:      Array containing initial wavefunction vector
!    psidim:   Length of psi arrays
!    abstime:  Absolute time (dummy parameter for time-independent H)
!    stepsize: Maximum stepsize allowed for integration (must be the
!              same as in "sod" and should be computed employing
!              "sodsize")
!    initacc:  Accuracy of initial step, i. e. |norm-1| of propagated
!              wavefunction
!    hamilton: Name of routine computing action of Hamiltonian on psi,
!              called as hamilton(abstime,psi,hpsi)
!    calcnorm: Name of routine computing the norm of the wavefunction,
!              called as calcnorm(psi,psidim,norm)
! OUTPUT PARAMETERS:
!    oldpsi:   Contains the initial wavefunction (same as "psi" in input)
!    psi:      Contains forward propagated wavefunction
!    stepsize: Stepsize for which the computation of the second initial
!              wavefunction converged (can be less than "stepsize" on
!              input)
!    hpsi:     Array containing H|psi> vector of "psi"
!    order:    Integration order used to compute the second initial value
!              (when two or more steps were needed, order is the maximum
!              order of all steps)
!    norm:     Norm of second initial wavefunction
! --------------------------------------------------------------------------

      subroutine initsod (psi,psidim,abstime,stepsize,initacc,hamilton,&
                          calcnorm,oldpsi,hpsi,order,norm,error)

      Implicit None

      complex(dop), parameter  :: imag=(0.0_dop,1.0_dop)

      external                                       :: hamilton,calcnorm
      integer(long), intent(in)                      :: psidim
      integer(long), intent(inout)                   :: order
      integer(long), intent(out)                     :: error
      complex(dop), dimension(psidim), intent(inout) :: psi,oldpsi
      complex(dop), dimension(psidim), intent(inout) :: hpsi
      complex(dop), dimension(psidim)                :: hhpsi
      real(dop), intent(in)                          :: abstime
      real(dop), intent(in)                          :: initacc
      real(dop), intent(in)                          :: stepsize
      real(dop), intent(out)                         :: norm

      integer(long)            :: d,p
      complex(dop)             :: coef
      real(dop)                :: initstep

! --- INITIALISE VARIABLES ---

      initstep = stepsize
      error = 0

! --- SAVE INITIAL WAVEFUNCTION ---

      do d = 1,psidim
         oldpsi(d) = psi(d)
      enddo

! --- PROPAGATE INITIAL WAVEFUNCTION ONE STEP FORWARD ---

! --- Linear term ---

      p = 1
      coef = -imag*stepsize
      psi(:) = psi(:)+coef*hpsi(:)

! --- Higher order terms ---
  100 continue
      if (p .lt. order) then
         p = p+1
      else
         error = 1
         go to 200
      endif

      hhpsi=hpsi
      call hamilton(abstime,hhpsi,hpsi)
      coef = -(1.0d0/dble(p))*imag*stepsize*coef
      psi(:) = psi(:)+coef*hpsi(:)

! --- EXIT LOOP IF CONVERGED ---

      call calcnorm(psi,norm,psidim)
      if (dabs(1.0_dop-norm) .gt. initacc) goto 100

200 continue
      order = p

! --- COMPUTE ACTION OF HAMILTONIAN ON PROPAGATED WAVEFUNCTION ---

      call hamilton(abstime+stepsize,psi,hpsi)

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine initsod

! --------------------------------------------------------------------------
!                            SUBROUTINE SODSIZE
!
! PURPOSE:
!    Computes a stepsize for which the SOD-integrator is stable. 
! INPUT PARAMETERS:
!    intperiod:  Time interval that is to be integrated
!    stepsafety: Safety factor the largest stable stepsize is multiplied
!                with (must be positive and less than one)
!    eigvalmax:  Upper bound for the largest absolute eigenvalue of the
!                Hamiltonian
! OUTPUT PARAMETERS:
!    stepsize:   Stable stepsize for the SOD integrator
! --------------------------------------------------------------------------

      subroutine sodsize (intperiod,stepsafety,eigvalmax,stepsize)

      Implicit None

      real(dop), parameter     :: tiny = 1.0e-8_dop

      real(dop), intent(in)    :: intperiod,stepsafety,eigvalmax
      real(dop), intent(out)   :: stepsize

      integer(long)            :: steps

! --- COMPUTE STEPSIZE ---

      stepsize = min(stepsafety/eigvalmax,intperiod)

! --- FIT STEPSIZE TO INTEGRATION PERIOD ---

      steps = int(intperiod/stepsize-tiny*stepsize)+1
      stepsize = intperiod/steps

      stepsize = 0.5*stepsize

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine sodsize


      end module sodlib
