

! **********************************************************************
! *                                                                    *
! *                             SPLIT OPERATOR (spolib.f90)            *
! *                                                                    *
! * Program package containing subroutines that propagate wave-packets *
! * employing the split operator method.                               *
! *                                                                    *
! * Contains:                                                          *
!         spostep: Propagates a wave-packet using the split operator   *
! *                (SPO) scheme                                        *
!
!            spocoff: 
!
! **********************************************************************

        module spolib

use fftmod

        implicit none

        integer, parameter :: long  = selected_int_kind(9) 
        integer, parameter :: dop = selected_real_kind(15,307)
        complex(dop), dimension(:), allocatable, save :: expke
        complex(dop), dimension(:), allocatable, save :: exppe

        complex(dop), parameter :: imag=(0.0_dop,1.0_dop)
      
        private
        public :: spostep,spoinit

        contains

!**********************************************************************

        subroutine spoinit (timestep,fftp,pe,mass,ndof,gdim,zfft,fftdim,&
                   psidim,eigvalmax,intperiod,stepsafety)

        implicit none

        integer(long)                              :: g,f,steps
        integer(long), intent(in)                  :: ndof,fftdim,psidim
        integer(long), dimension(ndof), intent(in) :: gdim,zfft
        real(dop), intent(in)                      :: stepsafety,&
                                                      eigvalmax,intperiod
        real(dop), intent(inout)                   :: timestep
        real(dop), dimension(fftdim), intent(in)   :: fftp
        real(dop), dimension(psidim), intent(in)   :: pe
        real(dop), dimension(ndof), intent(in)     :: mass

        real(dop), parameter :: tiny = 1.0e-8_dop


! --- COMPUTE STEPSIZE ---

      timestep = min(stepsafety/eigvalmax,intperiod)

! --- FIT STEPSIZE TO INTEGRATION PERIOD ---

      steps = int(intperiod/timestep-tiny*timestep)+1
      timestep = intperiod/steps

! DO WE NEED TO DEALLOCATE??
        allocate(expke(fftdim))
        allocate(exppe(psidim))

        do f=1,ndof
           call setexpke(timestep,expke(zfft(f)),fftp(zfft(f)),mass(f),gdim(f))
        enddo
        do g=1,psidim
           exppe(g) = cdexp(-imag*pe(g)*timestep)
        enddo


        return
        end subroutine spoinit

!**********************************************************************

        subroutine setexpke(timestep,expke1,fftp1,mass1,gdim1)

        implicit none

        integer(long)             :: g
        integer(long), intent(in) :: gdim1
        real(dop), intent(in)     :: timestep
        real(dop), intent(in)     :: mass1
        real(dop), dimension(gdim1), intent(in) :: fftp1
        complex(dop), dimension(gdim1), intent(out) :: expke1

        do g=1,gdim1
           expke1(g) = fftp1(g)*fftp1(g)/(2.0_dop*mass1)
           expke1(g) = cdexp(-0.5_dop*imag*expke1(g)*timestep)
           expke1(g) = expke1(g)
        enddo

        return
        end subroutine setexpke

!**********************************************************************
!                       SUBROUTINE SPOSTEP

!   PURPOSE: 
!       Propagates a wavepacket using the Split Operator Scheme. 

!   INPUT PARAMETERS
!       psi:

!   OUTPUT PARAMETERS
!       psi:

!-----------------------------------------------------------------------
         subroutine spostep (psi,psidim,intperiod,&
                    stepsize,vgdim,gdim,ngdim,hin,rueck,fftfak,&
                    zfft,fftdim,ndof)

         implicit none

        integer(long)                                  :: f,g
        integer(long), intent(in)                      :: psidim,fftdim,ndof
        integer(long), dimension(ndof),intent(in)      :: vgdim,gdim,ngdim,zfft
        integer(long), dimension(fftdim),intent(in)    :: fftfak
        complex(dop), dimension(fftdim),intent(in)     :: hin
        complex(dop), dimension(fftdim),intent(in)     :: rueck
        complex(dop), dimension(psidim), intent(inout) :: psi
        complex(dop), dimension(psidim)                :: hpsi
        complex(dop), parameter                        :: imag=(0.0_dop,1.0_dop)
        real(dop), intent(in)                          :: intperiod,stepsize
        real(dop)                                      :: inttime,intstep
        real(dop), parameter                           :: tiny = 1.0d-8

        inttime = 0.0d0
        intstep = stepsize

! ---   START OF PROPAGATION LOOP ---
100     continue

! --- KE update ---
        do f=1,ndof
           call fftzop(psi,hpsi,vgdim(f),gdim(f),ngdim(f),hin(zfft(f)), &
                rueck(zfft(f)),fftfak(zfft(f)),expke(zfft(f)))
        enddo
        psi=hpsi

! --- PE update ---
        do g=1,psidim
           hpsi(g)= exppe(g)*psi(g)
        enddo
        psi=hpsi

! --- Second KE update ---
        do f=1,ndof
           call fftzop(psi,hpsi,vgdim(f),gdim(f),ngdim(f),hin(zfft(f)), &
                rueck(zfft(f)),fftfak(zfft(f)),expke(zfft(f)))
        enddo
        psi=hpsi


! --- UPDATE TIME VARIABLE ---
        inttime = inttime+intstep

! --- CHECK WHETHER INTEGRATION IS COMPLETE ---
! --- RETURN TO CALLING ROUTINE WHEN FINISHED ---

        if (abs(inttime-intperiod) .lt. tiny) then
           return

        else

! --- CHECK STEPSIZE AND CONTINUE INTEGRATION ---
           if ((inttime+intstep) .gt. intperiod) then
             print*,'SPO: ERROR WITH TIMESTEP'
             stop
           endif

           goto 100
        endif

! --- RETURN TO CALLING ROUTINE ---

        return
        end subroutine spostep

!************************************************************************
! Do FFT of wavefunction vector
!************************************************************************

      subroutine fftzop(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak,fftk)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      complex(dop), dimension(gdim), intent(in)              :: fftk
      complex(dop), dimension(gdim)                          :: vec
      complex(dop), dimension(gdim), intent(in)              :: rueck,hin
      complex(dop), dimension(gdim)                          :: workc
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi

      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=fftk(g)*vec(g)/gdim
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftzop

!--------------------------------------------------------------------- 
        end module spolib
