     module verletlib

use decimal, only: long,dop

      implicit none

      integer(long), save :: versteps

      private
      public :: versteps
      public :: verlet

      contains

!***********************************************************************
!
! SUBROUTINE VERLET
!
!***********************************************************************
!
! Input:
!     psi         : initial values of the dynamic variables
!     dtpsi       : initial time derivative of the dyn. vars
!     psidim      : number of dyn. vars
!     intperiod   : length of integration interval
!     abstime     : initial time
!     initstep    : initial stepsize
!     accuracy    : parameter controlling the accuracy
!     auxpsi      : storage for intermediate results, workspace...
!                   size: 7*psidim; can be uninitialized
!     func        : routine for evaluating the RHS
!     mc,mr,mi,ml : the MCTDH-arrays
!     lpost       : whether to call a post-processor
!     postfn      : post-processing routine
!     ltest       : whether to call a test routine at end of step
!     testfn      : test routine
!
! Output:
!     psi         : final values of the dyn. vars
!     dtpsi       : undetermined (i.e. it gets overwritten!)
!     initstep    : optimal stepsize at end of integration
!     error       : errorflag; 0 means OK (currently it's always 0!)
!
! FO, June 2002
!
!***********************************************************************

      subroutine verlet(psi,dtpsi,gdof,dim1,intperiod,abstime,initstep,func)

use decimal, only:long,dop
use global
use datenmod
use operdef

use logdat,     only: errormsg
use odemod_interface

      implicit none
      
      integer(long), intent(in)                         :: gdof,dim1
      real(dop), intent(in)                             :: intperiod,abstime
      real(dop), intent(inout)                          :: initstep
      complex(dop), dimension(gdof,dim1), intent(inout) :: psi,dtpsi
      complex(dop), dimension(gdof,dim1)                :: psitmp,dtpsitmp
      external                                          :: func,postfn,testfn
      
      real(dop)                         :: tend,time,dt,dtsq,rxold,rxnew,&
                                           vxold,vxnew,axold,axnew
      real(dop), dimension(gdof,dim1)   :: vxmid
      integer(long)                     :: verloop,n,i,f
      logical(kind=4), save             :: initialized = .false.

!     initialize statistics on first call
      if (.not. initialized) then
         versteps = 0
         initialized = .true.
      endif

! set parameters for call to core integrator
      dt      = initstep
      dtsq    = dt*dt
      time    = abstime
      tend    = abstime+intperiod
      verloop = nint(intperiod/dt)
! zero variables
      vxmid = 0.0_dop

! start integration loop
     do n=1, verloop
        versteps = versteps+1
        if (n .eq. verloop) then
           dt=tend-time
           dtsq=dt*dt
        endif
        do i=1,dim1
           do f=1,gdof
              rxold=dble(psi(f,i))
              vxold=dble(dtpsi(f,i))
              axold=dimag(dtpsi(f,i))/mass(f)
              vxmid(f,i)=vxold+0.50*dt*axold
              psitmp(f,i)=rxold+dt*vxold+0.5*dtsq*axold
           enddo  
        enddo

! calculate new accelerations
        call func(time,psitmp,dtpsitmp)
!       dtpsitmp = dtpsi

        do i=1,dim1
           do f=1,gdof
              axnew=dimag(dtpsitmp(f,i))/mass(f)
              vxnew=vxmid(f,i)+0.5*dt*axnew
              vxnew=mass(f)*vxnew
              rxnew=dble(psitmp(f,i))
              psi(f,i)=dcmplx(rxnew,vxnew)
              dtpsi(f,i)=dcmplx(vxnew/mass(f),axnew*mass(f))
           enddo
         enddo 

! Update time         
         time = time+dt
      enddo   
      return

      end subroutine verlet 
      
   end module verletlib
