!----------------------------------------------------------------------------
      module coortransmod

use decimal, only: dop

      implicit none
      private
      public :: coortrans,coortransj,coortransb
      contains

!----------------------------------------------------------------------------
!      Subrutine coortransj transformates the cartesian coordinates
!      of the 4-atomic system to the 6 binding lengths.
!
! On input:  a(3),b(3),c(3),d(3) - cartesian coordinates of the 4 atoms
! On output: rab,rac,rad,rbc,rbd,rcd,cosdi - binding coordinates and the
!                                            cosine of the torsional angle 
!----------------------------------------------------------------------------

      subroutine coortrans(a,b,c,d,rab,rac,rad,rbc,rbd,rcd,cosdi)

      implicit none

      real(dop), dimension(3), intent(in)  :: a,b,c,d
      real(dop), dimension(3)              :: v1,v2,v3,vab,vac,vcd,v1v2,v2v3
      real(dop), intent(out)               :: rab,rac,rad,rbc,rbd,rcd,cosdi
      real(dop)                            :: av1v2,av2v3,th1,th2,sinth1,&
                                              sinth2,costh1,costh2
!
      v1(1)=c(1)-b(1)
      v1(2)=c(2)-b(2)
      v1(3)=c(3)-b(3)
      rbc=sqrt(v1(1)**2+v1(2)**2+v1(3)**2)
!
      v2(1)=d(1)-b(1)
      v2(2)=d(2)-b(2)
      v2(3)=d(3)-b(3)
      rbd=sqrt(v2(1)**2+v2(2)**2+v2(3)**2)
!
      v3(1)=a(1)-d(1)
      v3(2)=a(2)-d(2)
      v3(3)=a(3)-d(3)
      rad=sqrt(v3(1)**2+v3(2)**2+v3(3)**2)
!
      vab(1)=b(1)-a(1)
      vab(2)=b(2)-a(2)
      vab(3)=b(3)-a(3)
      rab=sqrt(vab(1)**2+vab(2)**2+vab(3)**2)
!
      vac(1)=c(1)-a(1)
      vac(2)=c(2)-a(2)
      vac(3)=c(3)-a(3)
      rac=sqrt(vac(1)**2+vac(2)**2+vac(3)**2)
!
      vcd(1)=d(1)-c(1)
      vcd(2)=d(2)-c(2)
      vcd(3)=d(3)-c(3)
      rcd=sqrt(vcd(1)**2+vcd(2)**2+vcd(3)**2)
! v1v2=v1 X v2,  v2v3=v2 X v3
      v1v2(1)=v1(2)*v2(3)-v1(3)*v2(2)
      v1v2(2)=v1(3)*v2(1)-v1(1)*v2(3)
      v1v2(3)=v1(1)*v2(2)-v1(2)*v2(1)
!
      v2v3(1)=v2(2)*v3(3)-v2(3)*v3(2)
      v2v3(2)=v2(3)*v3(1)-v2(1)*v3(3)
      v2v3(3)=v2(1)*v3(2)-v2(2)*v3(1)
!
      av1v2=sqrt(v1v2(1)**2+v1v2(2)**2+v1v2(3)**2)
      av2v3=sqrt(v2v3(1)**2+v2v3(2)**2+v2v3(3)**2)
!     costh3=(v1(1)*v3(1)+v1(2)*v3(2)+v1(3)*v3(3))/(rbc*rad)
! th1
      costh1=(rbc**2+rbd**2-rcd**2)/(2.0_dop*rbc*rbd)
      if(costh1.gt.1.0_dop) costh1=1.0_dop
      if(costh1.lt.-1.0_dop) costh1=-1.0_dop
      th1=acos(costh1)
      sinth1=sin(th1)
! th2
      costh2=((rbd**2+rad**2-rab**2)/(2.0*rbd*rad))
      if(costh2.gt.1.0_dop) costh2=1.0_dop
      if(costh2.lt.-1.0_dop) costh2=-1.0_dop
      th2=acos(costh2)
      sinth2=sin(th2)
! get the dihedral angle
      if(abs(sinth1).le.1.0e-5_dop.or.abs(sinth2).le.1.0e-5_dop) then
        cosdi=1.0_dop
      else
        cosdi=-(v1v2(1)*v2v3(1)+v1v2(2)*v2v3(2)+v1v2(3)*v2v3(3)) &
           /(av1v2*av2v3)
      endif
      if(cosdi.gt.1.0_dop) cosdi=1.0_dop
      if(cosdi.lt.-1.0_dop) cosdi=-1.0_dop
      return
      end subroutine coortrans

!----------------------------------------------------------------------------
!      Subroutine coortransj transformates the Jacobi-coordinates
!      to the cartesian coordinates for the 4-atomic system.
!
! On input:  r1,r2,R,th1,th2,phi - the Jacobi-coordinates
!            tfac1=mb/(ma+mb), tfac2=md/(mc+md), where ma,mb,mc,md are the
!                                                masses of the atoms.
! On output: a(3),b(3),c(3),d(3) - the cartesian coordinates of the atoms
!
! 05.02 SS
! Note: The origin of the cartesian system is taken in the center of mass
!       of the a-b molecule. The Z-axis is directed along the direction of R.
!       a-b molecule lies on the XZ plane.
!----------------------------------------------------------------------------

      subroutine coortransj(R,r1,r2,th1,th2,phi,tfac1,tfac2,a,b,c,d)

      implicit none

      real(dop), intent(in)                :: R,r1,r2,th1,th2,phi,tfac1,tfac2
      real(dop), dimension(3), intent(out) :: a,b,c,d

      a(1) = -tfac1*r1*sin(th1)
      a(2) = 0.0_dop
      a(3) = tfac1*r1*cos(th1)

      b(1) = (1.0_dop-tfac1)*r1*sin(th1)
      b(2) = 0.0_dop
      b(3) = -(1.0_dop-tfac1)*r1*cos(th1)

      c(1) = -tfac2*r2*sin(th2)*cos(phi)
      c(2) = -tfac2*r2*sin(th2)*sin(phi)
      c(3) = R + tfac2*r2*cos(th2)

      d(1) = (1.0_dop-tfac2)*r2*sin(th2)*cos(phi)
      d(2) = (1.0_dop-tfac2)*r2*sin(th2)*sin(phi)
      d(3) = R - (1.0_dop-tfac2)*r2*cos(th2)

      return
      end subroutine coortransj

!----------------------------------------------------------------------------
!      Subroutine coortransb transformates the Binding-coordinates
!      to the cartesian coordinates for the 4-atomic system.
!
! On input:  r1,r2,r3,th1,th2,phi - the Binding-coordinates
! On output: a(3),b(3),c(3),d(3)  - the cartesian coordinates of the atoms
!
! 05.02 SS
! Note: The origin of the cartesian system is taken at the place of b atom.
!       The X-axis is directed along the direction of b-c (r2).
!       a-b-c molecule lies on the XY plane.
!----------------------------------------------------------------------------

      subroutine coortransb(r1,r2,r3,th1,th2,phi,a,b,c,d)

      implicit none

      real(dop), intent(in)                :: r1,r2,r3,th1,th2,phi
      real(dop), dimension(3), intent(out) :: a,b,c,d


      a(1) = r1*cos(th1)
      a(2) = r1*sin(th1)
      a(3) = 0.0_dop

      b(1) = 0.0_dop
      b(2) = 0.0_dop
      b(3) = 0.0_dop

      c(1) = r2
      c(2) = 0.0_dop
      c(3) = 0.0_dop

      d(1) = r2 - r3*cos(th2)
      d(2) = r3*sin(th2)*cos(phi)
      d(3) = r3*sin(th2)*sin(phi)

      return
      end subroutine coortransb


      end module coortransmod
