!#######################################################################
! Libary module fftmod
!
! Contains the FFT-subroutine "fft", the corresponding initialisation
! subroutine "fftinit", and subordinate subroutines.
!
!#######################################################################

      module fftmod

use decimal, only: long, dop

      implicit none
      private
      public :: fft, fftinit, swapfft
      
      contains

!-----------------------------------------------------------------------
! Libary subroutine fft
!
! Calculates the FFT-transform of a complex tensor of third order in the
! middle index. No normalisation is done.
!
! Input-variables:  a      - complex tensor of third order
!                   work   - working array of same size as above array
!                   e      - array with exponentials for the FFT
!                            (calculated by subroutine "fftinit")
!                   faktor - array with factor decomposition for the FFT
!                   dim    - dimension of middle index of the tensor
!                   vor    - dimension of first index of the tensor
!                   nach   - dimension of third index of the tensor
!                   invers - label to mark forward or backward transform
!
! Output-variables: a      - transformed complex tensor of third order
!
! e=rueck and invers=.true. is used in the forward transform (negative
! exponentials), e=hin and inverse=.false. is used in the backward
! transform (positive exponentials).
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
      subroutine fft ( a, work, e, faktor, dim, vor, nach, invers )

      implicit none

      integer(long)                             :: teiler, off, rest, i
      integer(long), dimension(0:*), intent(in) :: faktor
      integer(long), intent(in)                 :: dim, vor, nach
      complex(dop), dimension(dim), intent(in)  :: e
      complex(dop), dimension(dim*vor*nach), intent(out)   :: a
      complex(dop), dimension(dim*vor*nach), intent(out)   :: work
      logical(kind=4), intent(in)               :: invers
      logical(kind=4)                           :: tausch

      off=1
      tausch=.false.
      do i=1,faktor(0)
         teiler=faktor(i)
         rest=dim/(teiler*off)
         if (tausch) then
            call subfft (work,a,e,teiler,vor*off,off,rest,nach,invers)
         else
            call subfft (a,work,e,teiler,vor*off,off,rest,nach,invers)
         endif
         tausch=.not.tausch
         off=off*teiler
      enddo
      if (tausch) then
         do i=1,vor*dim*nach
            a(i)=work(i)
         enddo
      endif
      return
      end subroutine fft

!---------------------------------------------------------------------

      subroutine subfft (a1,b1,e1,faktor,vor,off,rest,nach,invers)

      implicit none

      integer(long)                                   :: i, j, k, n, m, dim, dim1
      integer(long), intent(in)                       :: faktor, off, rest, &
                                                         vor, nach
      real(dop)                                       :: x1, x2, y1, y2
      real(dop), dimension(2,off,0:rest*faktor)       :: e
      real(dop), dimension(2,vor,0:rest-1,0:faktor-1,nach) :: a
      real(dop), dimension(2,vor,0:faktor-1,0:rest-1,nach) :: b
      real(dop), parameter                            :: re3=-0.5d0 
      real(dop), parameter                            :: im3= 0.86602540378443858d0
      logical(kind=4), intent(in)                     :: invers
      complex(dop), dimension(*), intent(in)    :: a1
      complex(dop), dimension(*), intent(in)    :: e1
      complex(dop), dimension(*), intent(out)   :: b1

      dim = nach*faktor*rest*vor
      dim1 = faktor*rest*off
      call reimsplit(a1,a,dim)
      call reimsplit(e1,e,dim1)

      select case (faktor)

!----------------- faktor 2 ------------------------

      case(2)
   
      do j=1,nach
         do k=0,rest-1
            do i=1,vor
               b(1,i,0,k,j)=a(1,i,k,0,j)+a(1,i,k,1,j)
               b(2,i,0,k,j)=a(2,i,k,0,j)+a(2,i,k,1,j)
               b(1,i,1,k,j)=e(1,1,k)*(a(1,i,k,0,j)-a(1,i,k,1,j)) &
                            -e(2,1,k)*(a(2,i,k,0,j)-a(2,i,k,1,j))
               b(2,i,1,k,j)=e(2,1,k)*(a(1,i,k,0,j)-a(1,i,k,1,j)) &
                            +e(1,1,k)*(a(2,i,k,0,j)-a(2,i,k,1,j))
            enddo
         enddo
      enddo
      call reimunsplit(b,b1,dim)
      return

!----------------- faktor 3 ------------------------

      case(3)
      if (invers) then
         do j=1,nach
            do k=0,rest-1
               do i=1,vor
                  x1=re3*(a(1,i,k,1,j)+a(1,i,k,2,j))
                  x2=re3*(a(2,i,k,1,j)+a(2,i,k,2,j))
                  y2=im3*(a(1,i,k,1,j)-a(1,i,k,2,j))
                  y1=-im3*(a(2,i,k,1,j)-a(2,i,k,2,j))
                  b(1,i,0,k,j)=a(1,i,k,0,j)+a(1,i,k,1,j)+a(1,i,k,2,j)
                  b(2,i,0,k,j)=a(2,i,k,0,j)+a(2,i,k,1,j)+a(2,i,k,2,j)
                  b(1,i,1,k,j)=e(1,1,k)  *(a(1,i,k,0,j)+x1-y1) &
                               -e(2,1,k)  *(a(2,i,k,0,j)+x2-y2)
                  b(2,i,1,k,j)=e(2,1,k)  *(a(1,i,k,0,j)+x1-y1) &
                               +e(1,1,k)  *(a(2,i,k,0,j)+x2-y2)
                  b(1,i,2,k,j)=e(1,1,k*2)*(a(1,i,k,0,j)+x1+y1) &
                               -e(2,1,k*2)*(a(2,i,k,0,j)+x2+y2)
                  b(2,i,2,k,j)=e(2,1,k*2)*(a(1,i,k,0,j)+x1+y1) &
                               +e(1,1,k*2)*(a(2,i,k,0,j)+x2+y2)
               enddo
            enddo
         enddo
      else
         do j=1,nach
            do k=0,rest-1
               do i=1,vor
                  x1=re3*(a(1,i,k,1,j)+a(1,i,k,2,j))
                  x2=re3*(a(2,i,k,1,j)+a(2,i,k,2,j))
                  y2=im3*(a(1,i,k,1,j)-a(1,i,k,2,j))
                  y1=-im3*(a(2,i,k,1,j)-a(2,i,k,2,j))
                  b(1,i,0,k,j)=a(1,i,k,0,j)+a(1,i,k,1,j)+a(1,i,k,2,j)
                  b(2,i,0,k,j)=a(2,i,k,0,j)+a(2,i,k,1,j)+a(2,i,k,2,j)
                  b(1,i,1,k,j)=e(1,1,k)  *(a(1,i,k,0,j)+x1+y1) &
                               -e(2,1,k)  *(a(2,i,k,0,j)+x2+y2)
                  b(2,i,1,k,j)=e(2,1,k)  *(a(1,i,k,0,j)+x1+y1) &
                               +e(1,1,k)  *(a(2,i,k,0,j)+x2+y2)
                  b(1,i,2,k,j)=e(1,1,k*2)*(a(1,i,k,0,j)+x1-y1) &
                               -e(2,1,k*2)*(a(2,i,k,0,j)+x2-y2)
                  b(2,i,2,k,j)=e(2,1,k*2)*(a(1,i,k,0,j)+x1-y1) &
                               +e(1,1,k*2)*(a(2,i,k,0,j)+x2-y2)
               enddo
            enddo
         enddo
      endif
      call reimunsplit(b,b1,dim)
      return

      end select

! ---------------- allgemeiner faktor ---------------

      do j=1,nach
         do k=0,rest-1
            do n=0,faktor-1
               do i=1,vor
                  b(1,i,n,k,j)=0.d0
                  b(2,i,n,k,j)=0.d0
               enddo
            enddo
         enddo
      enddo
      do n=0,faktor-1
         do m=0,faktor-1
            do k=0,rest-1
               do j=1,nach
                  do i=1,vor
                     x1=e(1,1,rest*mod(n*m,faktor))
                     x2=e(2,1,rest*mod(n*m,faktor))
                     y1=x1*e(1,1,k*n)-x2*e(2,1,k*n)
                     y2=x2*e(1,1,k*n)+x1*e(2,1,k*n)
                     b(1,i,n,k,j)=b(1,i,n,k,j)+y1*a(1,i,k,m,j)-y2*a(2,i,k,m,j)
                     b(2,i,n,k,j)=b(2,i,n,k,j)+y2*a(1,i,k,m,j)+y1*a(2,i,k,m,j)
                  enddo
               enddo
            enddo
         enddo
      enddo

      call reimunsplit(b,b1,dim)
      return

      end subroutine

!-----------------------------------------------------------------------

      subroutine reimsplit(a1,a,dim)

      implicit none

      integer                      :: i
      integer, intent(in)          :: dim
      complex(dop), dimension(dim) :: a1
      real(dop), dimension(2,dim)  :: a

      do i=1,dim
         a(1,i) = dble(a1(i))
         a(2,i) = dimag(a1(i))
      enddo

      return
      end subroutine reimsplit

!-----------------------------------------------------------------------

      subroutine reimunsplit(a,a1,dim)

      implicit none

      integer                      :: i
      integer, intent(in)          :: dim
      complex(dop), dimension(dim) :: a1
      real(dop), dimension(2,dim)  :: a

      do i=1,dim
         a1(i) = dcmplx(a(1,i),a(2,i))
      enddo

      return
      end subroutine reimunsplit

!-----------------------------------------------------------------------
! Libary subroutine fftinit
!
! Calculates data required for the FFT-transform by subroutine "fft".
!
! Input-variables:  dim        - dimension of the FFT
! Output-variables: faktor     - integer array with factor decomposition
!                   hin, rueck - complex array of lenght dim with 
!                                exponentials for the FFT
!                                (as specified in subroutine "fft")
!-----------------------------------------------------------------------

      subroutine fftinit (dim, faktor, hin, rueck)

      implicit none

      integer(long)                                 :: zahl, n, i
      integer(long), dimension(12)                  :: teiler
      integer(long), intent(in)                     :: dim 
      integer(long), dimension(0:*), intent(out)    :: faktor 
      real(dop)                                     :: pi
      complex(dop), dimension(0:dim-1), intent(out) :: hin
      complex(dop), dimension(0:dim-1), intent(out) :: rueck

      pi=4.0d0*atan(1.0d0)

      do n=0,dim-1
         hin(n)   = exp( dcmplx(0.d0, 2.d0*n*pi/dble(dim)) )
         rueck(n) = exp( dcmplx(0.d0,-2.d0*n*pi/dble(dim)) )
      enddo
      teiler(1)=2
      teiler(2)=3
      teiler(3)=5
      teiler(4)=7
      teiler(5)=11
      teiler(6)=13
      teiler(7)=17
      teiler(8)=19
      teiler(9)=23
      teiler(10)=29
      teiler(11)=31
      teiler(12)=37
      zahl=dim
      n=0
      do i=1,12
         do while (mod(zahl,teiler(i)).eq.0) 
            zahl=zahl/teiler(i)
            n=n+1
            faktor(n)=teiler(i)
         enddo
      enddo
      if (zahl.ne.1) then
         n=n+1
         faktor(n)=zahl
      endif
      faktor(0)=n

      return
      end subroutine




!-----------------------------------------------------------------------

      subroutine swapfft(psi,workc,gdim1,vdim1,ndim1,dim1)

!.....This routine swaps the first and second half of the SPFs
!.....because of the special arrangement of the grid points in
!.....FFT momentum representation.   HDM 03/04

use lalib,  only: normspf
use op1lib, only: cpvxz

      implicit none

      integer(long)                                                  :: g, ev, &
                                                                        en, ga, &
                                                                        gb
      integer(long), intent(in)                                      :: gdim1, &
                                                                        vdim1, &
                                                                        ndim1, &
                                                                        dim1
      complex(dop), dimension(vdim1,gdim1,ndim1*dim1), intent(out)   :: workc
      complex(dop), dimension(vdim1,gdim1,ndim1*dim1), intent(inout) :: psi
      
      call cpvxz(psi,workc,gdim1*vdim1*ndim1*dim1)
      ga = gdim1/2
      gb = (gdim1+1)/2

      do ev = 1, vdim1
         do en = 1, ndim1*dim1
            do g = 1, ga
               psi(ev,g+ga,en) =  workc(ev,g,en)
               psi(ev,g,en) =  workc(ev,gb+g,en)
            enddo
            psi(ev,gdim1,en) =  workc(ev,gb,en)
         enddo
      enddo
      call normspf(psi,vdim1*gdim1*ndim1,dim1)

      return
      end subroutine

      end module fftmod
!-----------------------------------------------------------------------
