!***********************************************************************
!
!           MCTDH  subroutine KEYUNITS
!
!  Rescales 'val' to [au] depending on chosen unit system.
!
!  The subroutines KeyEUnits, KeyMUnits, KeyLUnits and KeyPUnits apply for 
!  energy, mass length and momentum units only. 
!  The entry SetUnits is called from defaults.
!
!***********************************************************************


!-----------------------------------------------------------------------
! PARAMETERS FOR CALCULATIONS
!
! fs : time conversion from a.u. to fs (x[a.u.] = fs*x[fs]).
!
! ev : energy conversion from a.u. to ev (x[a.u.] = x[eV]/ev).
! invcm : energy conversion from a.u. to cm-1 (x[a.u.] = x[cm-1]/invcm).
! kcal  : energy conversion from a.u. to kcal/mol
!        (x[a.u.] = x[kcal/mol]/kcal).
! kjoule: energy conversion from a.u. to kjoule/mol
!        (x[a.u.] = x[kjoule/mol]/kjoule).
! kelvin: energy conversion from a.u. to kelvin
!        (x[a.u.] = x[kelvin]/kelvin).
! nmwl  : energy conversion from a.u. to nanometer-wavelength
!        (x[a.u.] = nmWavelength/x[nmwl]).
! attojoule: energy conversion from a.u. to aJ (= 10^-18 J)
!        (x[a.u.] = x[aJ]/attojoule
!
! atomicmassunit: conversion from atomic mass unit to a.u. 
!                  (m[a.u.] = m[AMU]*atomicmassunit).
! hydrogenmass  : conversion from hydrogen mass to a.u. 
!                  (m[a.u.] = m[H-mass]*hydrogenmass).
! deuteriummass : conversion from deuterium mass to a.u. 
!                  (m[a.u.] = m[D-mass]*deuteriummass).
!
! angstroem  : length conversion from a.u. to angstroem
!                  (x[a.u.] = x[angstroem]/angstroem). 
! picometer  : length conversion from a.u. to picometer
!                  (x[a.u.] = x[picometer]/picometer). 
! nanometer  : length conversion from a.u. to nanometer
!                  (x[a.u.] = x[nanometer]/nanometer). 
! degree     : "length" conversion from degrees to radian
!                  (x[rad] = x[degree]*(pi/180.0)).
!
! debye      : "energy" conversion from debye to au
!                  (x[a.u.] = x[debye]*debye).
!
! The corresponding keywords read :
! au, ev, cm-1, kcal/mol, kJ/mol, kelvin, nmwl, aJ, invev (energy)
! au, amu, H-mass, D-mass                                 (mass)
! au, angst, pm, nm, deg                                  (length)
! au, angst-1, pm-1, nm-1, ev*amu                         (momentum)
! au, fs, ps, invfs                                       (time)
!
! Atomic units (au) is always the default.
! invev and invfs refer to 1/ev and 1/fs respectively.
!
!-----------------------------------------------------------------------

      module keyunitsmod

use decimal, only: dop
use global, only: pi,ev,fs

      implicit none

      real(dop) :: ufs, uev, invcm, kcal, kelvin, kjoule, attojoule, &
                atomicmassunit, hydrogenmass, deuteriummass,debye, &
                angstroem, picometer,nanometer,nmWavelength, &
                protonmass
      logical   :: long
 
      parameter (ufs=41.34137333656d0)
      parameter (uev=27.21138386d0)
      parameter (invcm  = 2.1947463137d+5)
      parameter (kcal   = 6.27503d+2)
      parameter (kjoule = 2.6255d+3)
      parameter (kelvin = 3.15777d+5)
      parameter (nmWavelength=1.d+7/invcm)
      parameter (attojoule = 1.602177d-1*uev)
      parameter (atomicmassunit=1822.88848325d0)
      parameter (protonmass    =1836.15267247d0)
      parameter (hydrogenmass  =1837.15264409d0)
      parameter (deuteriummass =3671.482934845d0)
      parameter (angstroem=0.52917720859d0)
      parameter (picometer=0.52917720859d+2)
      parameter (nanometer=0.52917720859d-1)
      parameter (debye=0.39343d0)

!-----------------------------------------------------------------------
! conversion factor x[cm-1] = x[ev]/invcm2ev
!-----------------------------------------------------------------------
      real(dop), parameter :: invcm2ev = 1.23985e-4_dop

      private
      public :: keyunits, setunits, keylunits, keyeunits, keymunits, &
                keytunits, keypunits

      public :: ufs, uev, invcm, kcal, kelvin, kjoule, attojoule, &
                atomicmassunit, hydrogenmass, deuteriummass,debye, &
                angstroem, picometer,nanometer,nmWavelength, &
                protonmass, invcm2ev

      contains

      subroutine keyunits(val,keyword,ierr)

      implicit none

      integer   ierr
      real*8    val
      character*(*) keyword

      long = .true.
      goto 100

!***********************************************************************

      entry keyeunits(val,keyword,ierr)

      long = .false.
 100  ierr = 0

      if (keyword .eq. 'au') then
         continue
      elseif (keyword .eq. 'mh' ) then
         val = val*1.0d-3
      elseif (keyword .eq. 'ev' ) then
         val = val/ueV
      elseif (keyword .eq. 'mev' ) then
         val = 1.0d-3*val/ueV
      elseif (keyword .eq. 'invev' ) then
         val = val*uev
      elseif (keyword .eq. 'cm-1' ) then  
         val = val/invcm
      elseif (keyword .eq. 'kcal/mol') then  
         val = val/kcal
      elseif (keyword .eq. 'kj/mol' ) then  
         val = val/kjoule
      elseif (keyword .eq. 'kelvin' ) then
         val = val/kelvin
      elseif (keyword .eq. 'nmwl' ) then
         val = nmWavelength/val
      elseif (keyword .eq. 'aj' ) then
         val = val/attojoule
      elseif (keyword .eq. 'debye' ) then
         val = val*debye
      elseif (long)  then
         goto 110
      else
         ierr = 1
      endif

      return

!***********************************************************************

      entry keymunits(val,keyword,ierr)

      long = .false.
 110  ierr = 0

      if (keyword .eq. 'au') then
         continue
      elseif (keyword .eq. 'amu'   ) then
         val = val*atomicmassunit
      elseif(keyword .eq. 'h-mass' ) then  
         val = val*hydrogenmass
      elseif(keyword .eq. 'd-mass' ) then  
         val = val*deuteriummass
      elseif(keyword .eq. 'p-mass' ) then
         val = val*protonmass
      elseif (long)  then
         goto 120
      else
         ierr = 1
      endif

      return

!***********************************************************************

      entry keylunits(val,keyword,ierr)

      long = .false.
 120  ierr = 0

      if (keyword.eq.'au') then
         continue
      else if(keyword .eq. 'angst') then 
         val=val/angstroem
      else if( keyword .eq. 'pm') then 
         val=val/picometer
      else if( keyword .eq. 'nm') then 
         val=val/nanometer
      else if( keyword .eq. 'deg') then 
         val=val*(atan(1d0)/45d0)
      elseif (long)  then
         goto 130
      else
         ierr = 1
      endif

      return

!***********************************************************************

      entry keypunits(val,keyword,ierr)

      long = .false.
 130  ierr = 0

      if (keyword.eq.'au') then
         continue
      else if(keyword .eq. 'angst-1') then 
         val=val*angstroem
      else if( keyword .eq. 'pm-1') then 
         val=val*picometer
      else if( keyword .eq. 'nm-1') then 
         val=val*nanometer
      else if( keyword .eq. 'ev*amu') then 
         val=sqrt(val*2.d0*atomicmassunit/uev)
      elseif (long)  then
         goto 140
      else
         ierr = 1
      endif

      return

!***********************************************************************

      entry keytunits(val,keyword,ierr)

      long = .false.
 140  ierr = 0

      if (keyword.eq.'au') then
         continue
      else if( keyword .eq. 'fs') then
         val=val*ufs
      else if( keyword .eq. 'ps') then
         val=val*ufs*1000.d0
      else if( keyword .eq. 'invfs') then
         val=val/ufs
      else if( keyword .eq. 'invkelvin') then
         val=val*kelvin
      else
         ierr = 1
      endif

      return
      end

!***********************************************************************
! assign values to units stored in global include file
!***********************************************************************

      subroutine setunits

      fs = ufs
      ev = uev
      pi = 4.d0*atan(1.d0)

      return
      end

!***********************************************************************

      end module keyunitsmod
