module mpiutilsmod

use decimal, only: dop, long

implicit none 
private
public :: mpe_decomp1d, mpibcastchunks_cmplx

contains

!-------------------------------------------------------------------
!     MPE_DECOMP1D
!
!     Routine to decompose a 1-D array when given a number of 
!     processors. Values returned assume a global domain of [1:n].
!-------------------------------------------------------------------
 
      subroutine mpe_decomp1d(idx1,idxn,start,end)

use global
use mpidata

      implicit none

      integer(long), intent(in)  :: idx1,idxn
      integer(long), intent(out) :: start, end
      integer(long)              :: nlocal,deficit,length

!     How long is the domain?
      length = idxn - idx1 + 1

!     Get minimum size of local domain.
      nlocal = length/ nmpi

!     Get start index, id of process or time the local domain size 
!     plus 1.
      start = mpirank * nlocal + 1

!     How much of the global domain is unaccounted for?
      deficit = mod(length,nmpi)

!     Account for the fact that some of the deficit may have been
!     added to the earlier blocks (i.e. their sizes increased).
      start = start + min(mpirank,deficit)

!     Increase the size of the local domain if the processor is having
!     some of the deficit allocated to it.
      if(mpirank.lt.deficit)then
         nlocal = nlocal + 1
      endif

!     Get the index of the end of the domain.
      end = start + nlocal - 1
      if(end.gt.length.or.mpirank.eq.nmpi-1) end  = length

! Fianlly account for the original start and end points of the domain.
      start = start + idx1 - 1
      end   = end + idx1 - 1

      return
      end subroutine mpe_decomp1d

!-----------------------------------------------------------------------
!    mpibcastchunks_cmplx - send large arrays in chunks()
!-----------------------------------------------------------------------

      subroutine mpibcastchunks_cmplx(array,arrdim)

use mpidata
#ifdef MPI
  use mpi
#endif

      implicit none
      integer(long), intent(in) :: arrdim
      complex(dop), dimension(arrdim), intent(out) :: array

#ifdef MPI
      integer(long) :: nchunks, chunksize, i,z

      if (arrdim .le. mpimaxchunk) then
         call MPI_Bcast(array,arrdim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)
         return
      end if

      nchunks = arrdim/mpimaxchunk

      z = 1
      do i=1,nchunks
         call MPI_Bcast(array(z),mpimaxchunk,MPI_DOUBLE_COMPLEX,0, &
              MPI_COMM_WORLD,mpierr)
         z = z + mpimaxchunk
      end do

      ! rest
      chunksize = arrdim - nchunks*mpimaxchunk
      if (chunksize .ne. 0) then
         call MPI_Bcast(array(z),chunksize,MPI_DOUBLE_COMPLEX,0, &
              MPI_COMM_WORLD,mpierr)
      end if
#endif

      end subroutine mpibcastchunks_cmplx

end module mpiutilsmod
