
! **********************************************************************
! *                                                                    *
! *                           STRUTIL (strutil.f)                      *
! *                                                                    *
! * Program package containing subroutines for string handling.        *
! * Contains:                                                          *
! *    convertline: removes from a line all blanks and comments        *
! *    countentries: counts the number of entries in a string          *
! *    nextentry: returns the position of the next entry               *
! *    getpara: returns the parameter belonging to a keyword           *
! *    getkeyword: returns the nth keyword of a string                 *
! *    rmblnk: removes blanks between words in a string                *
! *    rmlblnk: removes leading blanks in a string                     *
! *    keylowc: replaces upper case by lower case letters              *
! *    keyupwc: replaces lower case by upper case letters              *
! *    streq: compares two strings, the case of letters and trailing   *
! *           blanks are ignored                                       *
! *    slen: returns length of string ignoring trailing blanks         *
! *    get_tokens: partitions a character string `line' into tokens    *
! *                which are separated by blanks or tabs.              *
! *    rempath: removes the path from the file name                    *
! *    dirpath: removes the file name from the path                    *
! *    cncntrblnk: Concentrates multiple blanks into one blank.        *
! *                Pads end of string with blanks.                     *
! *    formatmsg: Format message for console output.                   *
! *    parsexml: Parse XML type of strings.                            *
! *                                                                    *
! **********************************************************************


! ----------------------------------------------------------------------
!                          SUBROUTINE CONVERTLINE
!
! PURPOSE:
!     Removes from a line all blanks and comments. A comment is
!     everything between "#" and the end of the line, and "!" and the
!     next ",". A comma is placed in front of the line.
! INPUT PARAMETERS:
!     string: string to which the converted line is appended
!     strlen: position after which to append
!     line: line to be worked on
! OUTPUT PARAMETERS:
!     string: string with the converted line appended
!     strlen: new length of "string"
!     error: error code having the following meaning:
!            0: everything is o. k.
!            1: string overflow
! ----------------------------------------------------------------------

      subroutine convertline (string, strlen, line, error)

      implicit none

      integer       strlen, error
      character*(*) string, line

      logical ignore
      integer pos, i

! --- INITIALISE VARIABLES ---

      error = 0
      string(strlen+1:strlen+1) = ','
      pos = strlen+2
      ignore = .false.

! --- LOOP OVER EACH CHARACTER ---

      do i = 1,len(line)
         if (pos .gt. len(string)) then
            error = 1
            return
         endif
         if (line(i:i) .eq. '#') then
            exit !exit loop
         elseif (line(i:i) .eq. '!') then
            ignore = .true.
         elseif (line(i:i) .eq. ',') then
            if (ignore) then
               ignore = .false.
            elseif (string(pos-1:pos-1) .ne. ',') then
               string(pos:pos) = ','
               pos = pos+1
            endif
         elseif ((.not. ignore) .and. line(i:i) .ne. ' ') then
            string(pos:pos) = line(i:i)
            pos = pos+1
         endif
      enddo

! --- DETERMINE NEW LENGTH OF STRING ---

      if (string(pos-1:pos-1) .eq. ',') then
         strlen = pos-2
      else
         strlen = pos-1
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end

! ----------------------------------------------------------------------
!                         SUBROUTINE COUNTENTRIES
!
! PURPOSE:
!     Counts the number of entries in a string, where an entry is a
!     substring separated by a comma.
! INPUT PARAMETERS:
!     string: string the entries of which are counted
!     strlen: length of string
! OUTPUT PARAMETERS:
!     entries: number of entries
! ----------------------------------------------------------------------

      subroutine countentries (string, strlen, entries)

      implicit none

      integer       strlen, entries
      character*(*) string

      integer i

! --- COUNT ENTRIES ---

      entries = 0
      do i = 1,strlen-1
         if (string(i:i) .eq. ',') entries = entries+1
      enddo

! --- RETURN TO CALLING ROUTINE ---

      return
      end

! ----------------------------------------------------------------------
!                          SUBROUTINE NEXTENTRY
!
! PURPOSE:
!     Returns the position of the next entry. An entry is an expression
!     of the form "keyword=parameter". The entries must be separated by
!     commas and must not contain any blanks.
! INPUT PARAMETERS:
!     string: string containing the entries
!     strlen: length of string
!     strpos: current position in "string" (must be 1 on first call,
!             else as returned in the previous call)
! OUTPUT PARAMETERS:
!     strpos: new position in "string" (or 1 if entry was last one)
!     first: position of first character of "keyword"
!     assign: position of "="
!     last: position of last character of "parameter"
! ----------------------------------------------------------------------

      subroutine nextentry (string, strlen, strpos, first, assign, last)

      implicit none

      integer,intent(in) :: strlen
      integer,intent(inout) :: strpos
      integer,intent(out) :: first,assign,last
      character*(*),intent(in) ::  string

! --- FIND POSITION OF KEYWORD AND PARAMETER ---

      first = strpos+1
      assign = strpos+1
      do strpos=first,strlen
          if (string(strpos:strpos) .eq. '=') then
             assign = strpos
          elseif (string(strpos:strpos) .eq. ',') then
             last = strpos-1
             exit
          endif
      enddo

      if(strpos.ge.strlen) then
         last=strlen
         strpos=1
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end
      
! ----------------------------------------------------------------------
!                          SUBROUTINE GETPARA
!
! PURPOSE:
!     Returns the parameter (i.e. rhs of "=") belonging to a keyword
!     (i.e. lhs of "="). The entries (i.e. keyword=parameter) must be
!     separated by commas and must not contain any blanks.
! INPUT PARAMETERS:
!     string: string containing the entries
!     strlen: length of string
!     keyword: the keyword to search for (not case-sensitive)
! OUTPUT PARAMETERS:
!     param: the parameter searched for (as a string)
!     paralen: length of "parameter"
!     parapos: position of parameter (i.e. 1st, 2nd, etc. entry, 0 if
!              not found)
! ----------------------------------------------------------------------

      subroutine getpara (string, strlen, keyword, param, paralen,&
                          parapos)

      implicit none

      integer,intent(in) :: strlen
      integer,intent(out) :: paralen, parapos
      character*(*),intent(in) :: string, keyword
      character*(*),intent(out) :: param

      logical streq
      integer strpos, first, assign, last, entry

! --- INITIALISE VARIABLES ---

      parapos = 0
      strpos  = 1

! --- SEARCH FOR DESIRED KEYWORD ---

      do entry=1,100 !should never arrive to the end
          call nextentry(string,strlen,strpos,first,assign,last)
          if (streq(keyword(1:len(keyword)),string(first:assign-1))) then
             parapos = entry
             paralen = last-assign
             param   = string(assign+1:last)
             exit
          endif
          if (strpos .eq. 1) exit
      enddo
     
! --- RETURN TO CALLING ROUTINE ---

      return
      end

! ----------------------------------------------------------------------
!                          SUBROUTINE GETKEYWORD
!
! PURPOSE:
!     Returns the nth keyword (i.e. lhs of "=") of a string.
!     The entries (i.e. keyword=parameter) must be separated by commas
!     and must not contain any blanks.
! INPUT PARAMETERS:
!     string: string containing the entries
!     strlen: length of string
!     number: the number of the keyword to search for
! OUTPUT PARAMETERS:
!     keyword: the keyword searched for
!     keylen: length of "keyword"
! ----------------------------------------------------------------------

      subroutine getkeyword (string, strlen, number, keyword, keylen)

      implicit none

      integer       strlen, keylen, number
      character*(*) string, keyword

      integer entry, first, last

! --- INITIALISE VARIABLES ---

      first = 1
      entry = 0

! --- FIND THE POSITION OF THE KEYWORD ---

      do
         if (string(first:first) .eq. ',') entry = entry+1
         if ((entry .eq. number) .or. (first .eq. strlen)) exit
         first = first+1
      enddo

! --- GET THE KEYWORD ---

      first   = first+1
      last    = first+min(index(string(first:strlen),'='),&
                          index(string(first:strlen),','))-2
      if (last .lt. first) last = strlen
      keyword = string(first:last)
      keylen  = last-first+1
     
! --- RETURN TO CALLING ROUTINE ---

      return
      end

!#######################################################################
!
!     subroutine rmblnk
!
!     removes blanks between words in a string
!
!#######################################################################

      subroutine rmblnk(string)

      implicit none
      integer i,j,length
      character*(*) string

      length=len(string)

      j=0
      do i=1,length
         if (string(i:i) .ne. ' ') then
            j=j+1
            string(j:j)=string(i:i)
         endif
      enddo
      do i=j+1,length
         string(i:i)=' '
      enddo

      return
      end

!#######################################################################
!
!     subroutine rmlblnk
!
!     removes leading blanks of a string
!
!#######################################################################

      subroutine rmlblnk(string)

      implicit none
      integer i,j,k,length
      character*(*) string

      if (string(1:1).ne.' ') return   ! no leading blanks

      length=len(string)

      do i=2,length
         if (string(i:i).ne.' ') exit
      enddo
      if (i.gt.length) return

      j=0
      do k=i,length
         j=j+1
         string(j:j)=string(k:k)
      enddo
      do i=j+1,length
         string(i:i)=' '
      enddo

      return
      end

!***********************************************************************
! 
!          Subroutine KEYLOWC
!
!   Replaces upper case letters [A-Z] by lower case letters [a-z].
!   All other symbols remain unchanged.
!
!   11/97 AJ
!
!***********************************************************************
      
      subroutine keylowc (string)

      implicit none

      integer i,j,length
      character*(*) string

      length=len(string)

      do i=1,length
         do j=65,90
            if (ichar(string(i:i)).eq.j) string(i:i)=char(j+32)
         enddo
      enddo

      return
      end

!***********************************************************************
! 
!          Subroutine KEYUPC
!
!   Replaces lower case letters [a-z] by upper case letters [A-Z].
!   All other symbols remain unchanged.
!
!   11/97 AJ
!
!***********************************************************************
      
      subroutine keyupc (string)

      implicit none

      integer i,j,length
      character*(*) string

      length=len(string)

      do i=1,length
         do j=97,122
            if (ichar(string(i:i)).eq.j) string(i:i)=char(j-32)
         enddo
      enddo

      return
      end

!***********************************************************************
!
!                           FUNCTION STREQ
!
! Compares two character strings "string1" and "string2". The case of
! letters is ignored, as well as blanks at the end.
! The function value "streq" is true if the strings are equal and false
! otherwise.
!
! 4/98 MB
!
!***********************************************************************

      logical function streq (string1,string2)

      character*(*) string1,string2
      integer       length1,length2,char1,char2,i
      integer       slen

!-----------------------------------------------------------------------
! Determine length of each string (ignoring trailing blanks)
!-----------------------------------------------------------------------
      length1 = slen(string1)
      length2 = slen(string2)

!-----------------------------------------------------------------------
! Compare the strings
!-----------------------------------------------------------------------
      if (length1 .ne. length2) then
         streq = .false.
      else
         streq = .true.
         do i = 1,length1
            char1 = ichar(string1(i:i))
            char2 = ichar(string2(i:i))
            if ((char1 .ge. 65) .and. (char1 .le. 91)) char1 = char1+32
            if ((char2 .ge. 65) .and. (char2 .le. 91)) char2 = char2+32
            if (char1 .ne. char2) then
               streq = .false.
               exit
            endif
         enddo
      endif

      return
      end

!***********************************************************************
!
!                           FUNCTION slen
!
!          returns length of string ignoring trailing blanks
! 
!***********************************************************************

      function slen(string)

      implicit none

      integer slen
      character*(*) string
      integer plength,i

      plength=len(string)
      do i=plength,1,-1
         if (string(i:i).ne.' ') exit
      enddo
      slen=i
      return
      end

!***********************************************************************
!
!                        GET_TOKENS
!
! partitions a character string `line' into tokens which are separated by
! blanks or tabs. returns the number of tokens, their initial and
! final character positions.
!
! line (i)
!    string to be parsed
! max_nof_tokens (i)
!    the program aborts with a message if the number of tokens exeeds
!    this value 
! nof_tokens (o)
!    number of tokens found. 
!    many tokens found. in this case, only the initial max_nof_tokens
!    entries of aof_tbounds are correct.
! aof_tbounds (o)
!    aof_tbounds(i,1) contains the position of the first character of
!    the ith token of string `line', aof_tbounds(i,2) the corresponding
!    last character.
! 
!***********************************************************************

      subroutine get_tokens(line,max_nof_tokens,nof_tokens,aof_tbounds)

      implicit none

      integer,intent(in) :: max_nof_tokens
      character*(*),intent(in) :: line
      integer,intent(out) :: nof_tokens
      integer,intent(out) :: aof_tbounds(max_nof_tokens,2)
      integer i,l
      logical state

      state=.false.
      nof_tokens=0
      i=1
      l=len(line)

      do i=1,l
         if(line(i:i).eq.' '.or.line(i:i).eq.'	') then
!        current character is a blank or a tab
             if (state) aof_tbounds(nof_tokens,2)=i-1
             state=.false.
          elseif (.not.state) then
             state=.true.
             nof_tokens=nof_tokens+1
             if(nof_tokens.gt.max_nof_tokens) then
                write(6,'(a,/,a)') &
                   'get_tokens: too many tokens in the following&
                   & string:',line
                stop
             else
                aof_tbounds(nof_tokens,1)=i
             endif
          endif
      enddo

      if(state) aof_tbounds(nof_tokens,2)=l
         
      return
      end

!######################################################################
!
! rempath
!
! Removes the path from the file name
! pathlen: on return, length of filename after removal of path
!
!#######################################################################

      subroutine rempath(filename,pathlen,error)

      implicit none

      integer pathlen,l,error
      character*(*) filename

      error=0

      pathlen=index(filename,' ')-1
      l=pathlen
      if (filename(l:l) .eq. '/') then
         error=1
      else
         do l=pathlen,1,-1 !loop going from high to low
            if (filename(l:l) .eq. '/') then
               filename=filename(l+1:pathlen)
               pathlen=pathlen-l
               exit
            endif
         enddo
      endif

      return
      end

!#######################################################################
!
! dirpath
!
! Removes the file name from the path, i.e. returns directory-path.
! Note: the final '/' is included in the string filename on return.
!
!#######################################################################

      subroutine dirpath(filename,pathlen)

      implicit none

      integer pathlen
      character*(*) filename

      do pathlen=index(filename,' ')-1,1,-1 !loop going from high to low
         if (filename(pathlen:pathlen).eq.'/') exit
      enddo
      if(pathlen.eq.1 .and. filename(1:1).ne.'/') then
         filename='./'
         pathlen = 2
      endif
      filename=filename(1:pathlen)

      return
      end

!#######################################################################
!
! cncntrblnk
!
! Concentrates multiple blanks into one blank. Pads end of string with
! blanks.
!
!#######################################################################

      subroutine cncntrblnk(string)

      implicit none
      integer i,j,m,length
      character*(*) string

      i=index(string," ")
      if (i .le. 0) return

      j=i+1
      length=len(string)

      do while(j.lt.length)
          ! find next non-blank
          if (string(i:i) .ne. ' ') then
             i=i+1
             j=i+1
             cycle
          elseif (j .lt. length) then
             if (string(j:j) .eq. ' ') then
               j=j+1
               cycle
             end if
          end if
          j=j-1
          ! move end forward
          if (i.ne.j) then
             do m=0,length-j
                string(i+m:i+m) = string(j+m:j+m)
             end do
             ! fill blanks
             string(length:length) = " "
          end if
          i=i+1
          j=i+1
      enddo

      return
      end

!#######################################################################

!=======================================================================
      subroutine formatmsg(lun,lwr6,msg)
!=======================================================================
!
! Write the string _msg_ to the output channel _lun_ but don't use
! lines longer than _llen_ characters. If necessary, break lines, but
! only at spaces or dashes.
! -- FO 10/2013
!
!=======================================================================
      implicit none
      integer           :: lun
      character(len=*)  :: msg
      integer           :: i,j,k,l,j1
      integer,parameter :: llen=80
      logical           :: lwr6

      j = 0
      i = 1
      l = len_trim(msg)

      do
         if (l-j > llen) then
            j1=j
            do k=20,llen-6
               if(msg(i+k:i+k+5) .eq. '      ') then
                  j=i+k    ! 6 blanks found, exit
                  goto 20
               endif
            enddo
            j = scan(msg(i:j+llen), " -", back=.true.) + j
            ! If there is a word longer then llen
            ! then break anyway.
            if (j1.eq.j) j=j+llen
         elseif (l-j > 0) then
            j = l
         else
            exit
         endif
 20      if(lun.gt.0) write (lun,'(a)') trim(msg(i:j))
         if(lwr6)     write (6,'(a)') trim(msg(i:j))
         i = j+1
         msg(i:l)=adjustl(msg(i:l))
         l = len_trim(msg)
      enddo

      end subroutine formatmsg

!=======================================================================
      subroutine FindInList(str,list,nlist,what,nwhat,idx)
!=======================================================================
!     Finds in string 'str' an entry of 'list' and returns which ever is
!     the first of the strings in 'list'. On exit, 'what' contains the
!     substring that is found, 'nwhat' is the position of what in list,
!     and 'idx' is the position of the first character of 'what' in
!     'str'. If none of the strings in 'list' are found, then 'what' is
!     blank and 'nwhat' and 'idx' are 0.
!
!     Input: str, list, nlist
!     Output: what, nwhat, idx
!     -- MS 04/18
!=======================================================================

      implicit none
      integer nlist, nwhat, idx, match, match1, i
      character(len=*) str, list(nlist), what

      match = len(str) + 1
      idx = 0
      nwhat = 0
      what = " "

      do i=1,nlist
         match1 = index(str,list(i))
         if (match1 .gt. 0) then
            if (match1 .lt. match) then
               match = match1
               idx = match1
               what = list(i)
               nwhat = i
            end if
         end if
      end do

      end subroutine FindInList

!#######################################################################
!
! parsexml
!
! Parses XML expressions of the type:
!       <tag>body body body</tag>
! returning whether <tag> was found, the position of the opening <tag>
! in the string, and the body of the string.
! It can also process standalone tags of the form:
!       <tag/>
! in which case returns body='NULL'
!
!#######################################################################

      subroutine parsexml(string,tag,body,pos,found)

      implicit none

      character(len=*),intent(in) :: string,tag
      character(len=360),intent(out) :: body
      integer,intent(out) :: pos
      logical,intent(out) :: found

      character(len=15) :: otag,ctag,satag
      integer :: tag_len,opos,cpos,sapos

      found=.false.
      pos=0

      otag='<'//tag//'>'
      ctag='</'//tag//'>'
      satag='<'//tag//'/>'

      tag_len=len_trim(tag)

      opos=index(string,trim(otag))
      cpos=index(string,trim(ctag),back=.true.)
      sapos=index(string,trim(satag))

      if (opos.ne.0 .and. cpos.ne.0) then
        found=.true.
        pos=opos
        body=string(opos+tag_len+2:cpos-1)
      else if (opos.ne.0 .and. cpos.eq.0) then
        body='ERROR'
        write(6,'(a,/,a,x,a,x,a,x,a)') 'ERROR in subroutine parsexml:',&
              'Found open tag',otag,'but no closing tag',ctag
      else if (sapos.ne.0) then
        found=.true.
        pos=sapos
        body='NULL'
      endif

      return
      end subroutine parsexml

!#######################################################################

