!***********************************************************************
!
! Library routines to set up and apply symmetry operations to a set
! of Cartesian coordinates.
!
!***********************************************************************

      module symoplib

use decimal, only: long, dop

      implicit none


      private 
      public :: fnsymop,fpntgp,dosymop,dosymop2,dosymperm

      real(dop), parameter :: pi = 4.0_dop*datan(1.0_dop)

      contains

!-----------------------------------------------------------------------
! FPNTGP
!
! For a point group label, returns an integer.
!-----------------------------------------------------------------------

      integer function fpntgp (label)

      implicit none

      character(len=*), intent(in) :: label

      if (label .eq. 'c1') then
         fpntgp = 0
      else if (label .eq. 'cs' .or. label .eq. 'cs_xy') then
         fpntgp = 1
      else if (label .eq. 'cs_xz') then
         fpntgp = 2
      else if (label .eq. 'cs_yz') then
         fpntgp = 3
      else if (label .eq. 'c2v' .or. label .eq. 'c2v_z') then
         fpntgp = 4
      else if (label .eq. 'c2v_x') then
         fpntgp = 5
      else if (label .eq. 'c2v_y') then
         fpntgp = 6
      else if (label .eq. 'd2h' .or. label .eq. 'd2h_z') then
         fpntgp = 7
      else if (label .eq. 'd2h_x') then
         fpntgp = 8
      else if (label .eq. 'd2h_y') then
         fpntgp = 9
      else if (label .eq. 'd2d' .or. label .eq. 'd2d_z') then
         fpntgp = 10
      else if (label .eq. 'c2' .or. label .eq. 'c2_z') then
         fpntgp = 11
      else if (label .eq. 'c2_x') then
         fpntgp = 12
      else if (label .eq. 'c2_y') then
         fpntgp = 13

      else 
         fpntgp = -99
      endif

      end function fpntgp

!-----------------------------------------------------------------------
! FNSYMOP
!
! For a point group, pntgrp, returns the no. of symmetry operations
! (including the identity)
!-----------------------------------------------------------------------

      integer function fnsymop (pntgrp)

      implicit none

      integer(long), intent(in) :: pntgrp

      if (pntgrp .eq. 0) then
         fnsymop = 1
      else if (pntgrp .eq. 1) then
         fnsymop = 2
      else if (pntgrp .eq. 2) then
         fnsymop = 2
      else if (pntgrp .eq. 3) then
         fnsymop = 2
      else if (pntgrp .eq. 4) then
         fnsymop = 4
      else if (pntgrp .eq. 5) then
         fnsymop = 4
      else if (pntgrp .eq. 6) then
         fnsymop = 4
      else if (pntgrp .eq. 7) then
         fnsymop = 8
      else if (pntgrp .eq. 8) then
         fnsymop = 8
      else if (pntgrp .eq. 9) then
         fnsymop = 8
      else if (pntgrp .eq. 10) then
         fnsymop = 8
      else if (pntgrp .eq. 11) then
         fnsymop = 2
      else if (pntgrp .eq. 12) then
         fnsymop = 2
      else if (pntgrp .eq. 13) then
         fnsymop = 2

      else 
         fnsymop = -99
      endif

      end function fnsymop 

!-----------------------------------------------------------------------
! DOSYMOP
!
! operate on xin using nth operation of point group pntgp to give xout
!
! NB. If new point groups get added, also add to DOSYMOP2 below
!-----------------------------------------------------------------------

      subroutine dosymop(xin,xout,pntgp,nop,natm,ilog)

      implicit none

      integer(long), intent(in)                 :: pntgp,nop,natm,ilog
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      real(dop), dimension(3,natm)              :: xtmp1,xtmp2
      real(dop)                                 :: theta

      xout(:,:) = 0.0_dop
      select case (pntgp)

      case default

          write(ilog,'(/,a,i5)') &
           'ERROR in symmetry operations (dosymop). Unknown point group : ', &
            pntgp
          stop
         
! Cs (xy plane)
      case (1)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call mirror(xin,xout,natm,3)
         endif

! Cs (xz plane)
      case (2)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call mirror(xin,xout,natm,2)
         endif

! Cs (yz plane)
      case (3)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call mirror(xin,xout,natm,1)
         endif

! C2v (z-axis)
      case (4)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,3)
         else if (nop .eq. 3) then
            call mirror(xin,xout,natm,2)
         else if (nop .eq. 4) then
            call mirror(xin,xout,natm,1)
         endif

! C2v (x-axis)
      case (5)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,1)
         else if (nop .eq. 3) then
            call mirror(xin,xout,natm,2)
         else if (nop .eq. 4) then
            call mirror(xin,xout,natm,3)
         endif

! C2v (y-axis)
      case (6)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,2)
         else if (nop .eq. 3) then
            call mirror(xin,xout,natm,3)
         else if (nop .eq. 4) then
            call mirror(xin,xout,natm,1)
         endif

! D2h (z-axis)
      case (7)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,3)
         else if (nop .eq. 3) then
            call rotate(xin,xout,natm,2,2)
         else if (nop .eq. 4) then
            call rotate(xin,xout,natm,2,1)
         else if (nop .eq. 5) then
            call inversion(xin,xout,natm)
         else if (nop .eq. 6) then
            call mirror(xin,xout,natm,3)
         else if (nop .eq. 7) then
            call mirror(xin,xout,natm,2)
         else if (nop .eq. 8) then
            call mirror(xin,xout,natm,1)
         endif

! D2h (x-axis)
      case (8)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,1)
         else if (nop .eq. 3) then
            call rotate(xin,xout,natm,2,2)
         else if (nop .eq. 4) then
            call rotate(xin,xout,natm,2,3)
         else if (nop .eq. 5) then
            call inversion(xin,xout,natm)
         else if (nop .eq. 6) then
            call mirror(xin,xout,natm,1)
         else if (nop .eq. 7) then
            call mirror(xin,xout,natm,2)
         else if (nop .eq. 8) then
            call mirror(xin,xout,natm,3)
         endif

! D2h (y-axis)
      case (9)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,2)
         else if (nop .eq. 3) then
            call rotate(xin,xout,natm,2,3)
         else if (nop .eq. 4) then
            call rotate(xin,xout,natm,2,1)
         else if (nop .eq. 5) then
            call inversion(xin,xout,natm)
         else if (nop .eq. 6) then
            call mirror(xin,xout,natm,2)
         else if (nop .eq. 7) then
            call mirror(xin,xout,natm,3)
         else if (nop .eq. 8) then
            call mirror(xin,xout,natm,1)
         endif

! D2d (z-axis)
! C2' axis consistent with x axis
      case (10)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call improt(xin,xout,natm,4,3)
         else if (nop .eq. 3) then
            call improt(xin,xtmp1,natm,4,3)
            call improt(xtmp1,xtmp2,natm,4,3)
            call improt(xtmp2,xout,natm,4,3)
         else if (nop .eq. 4) then
            call rotate(xin,xout,natm,2,3)
         else if (nop .eq. 5) then
            call rotate(xin,xout,natm,2,1)
         else if (nop .eq. 6) then
            call rotate(xin,xout,natm,2,2)
         else if (nop .eq. 7) then
            theta = pi/4.0_dop
            call mirrord(xin,xout,natm,theta,3)
         else if (nop .eq. 8) then
            theta = -pi/4.0_dop
            call mirrord(xin,xout,natm,theta,3)
         endif

! C2 (z-axis)
      case (11)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,3)
         endif

! C2 (x-axis)
      case (12)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,1)
         endif

! C2 (y-axis)
      case (13)
         if (nop .eq. 1) then
            xout = xin
         else if (nop .eq. 2) then
            call rotate(xin,xout,natm,2,2)
         endif

      end select

      return
      end subroutine dosymop

!-----------------------------------------------------------------------
! DOSYMOP2
!
! operate on matin using nth operation of point group pntgp to give 
! matout. Like dosymop, but for a tensor rather than a vector
!-----------------------------------------------------------------------

      subroutine dosymop2(matin,matout,pntgp,nop,natm,ilog)

      implicit none

      integer(long), intent(in)                 :: pntgp,nop,natm,ilog
      real(dop), dimension(3,natm,3,natm), intent(in)  :: matin
      real(dop), dimension(3,natm,3,natm), intent(out) :: matout

      select case (pntgp)

      case default

          write(ilog,'(/,a,i5)') &
           'ERROR in symmetry operations (dosymop2). Unknown point group : ', &
            pntgp
          stop
         
! Cs (xy plane)
      case (1)
         if (nop .eq. 1) then
            matout = matin
         else if (nop .eq. 2) then
            call mirror2(matin,matout,natm,3)
         endif

! Cs (xz plane)
      case (2)
         if (nop .eq. 1) then
            matout = matin
         else if (nop .eq. 2) then
            call mirror2(matin,matout,natm,2)
         endif

! Cs (yz plane)
      case (3)
         if (nop .eq. 1) then
            matout = matin
         else if (nop .eq. 2) then
            call mirror2(matin,matout,natm,1)
         endif


      end select

      return
      end subroutine dosymop2

!-----------------------------------------------------------------------
!
! DOSYMPERM
!
! maps one set of atoms onto another
! Used to implement permutations to return axis to original ordering
! after a symmetry operation
!
!-----------------------------------------------------------------------

      subroutine dosymperm(xin,xout,symmap,natm)

      implicit none

      integer(long), intent(in)                  :: natm
      integer(long), dimension(natm), intent(in) :: symmap
      integer(long)                              :: n,n1
      real(dop), dimension(3,natm), intent(in)   :: xin
      real(dop), dimension(3,natm), intent(out)  :: xout

      do n = 1,natm
         n1 = symmap(n)
         xout(:,n) = xin(:,n1)
      enddo

      return
      end subroutine dosymperm

!-----------------------------------------------------------------------
! Reflect a vector of Cartesian coordinates through a plane specified 
! by the "axis" normal to the plane
! iaxis = 1, 2, 3  (x, y, z respectively)
!-----------------------------------------------------------------------

      subroutine mirror(xin,xout,natm,iaxis)

      implicit none

      integer(long), intent(in) :: natm, iaxis
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      integer(long) :: i

      do i=1,3
         if (i .eq. iaxis) then
            xout(i,:) = -xin(i,:)
         else
            xout(i,:) = xin(i,:)
         endif
      enddo

      return
      end subroutine mirror

!-----------------------------------------------------------------------
! Reflct a vector of Cartesian coordinates through a line in the "xy" 
! plane which is angle theta (rad) from the "x-axis". 
! Plane defined by "z-axis":
! iaxis = 1, 2, 3  (x, y, z respectively)
!-----------------------------------------------------------------------

      subroutine mirrord(xin,xout,natm,theta,iaxis)

      implicit none

      integer(long), intent(in) :: natm, iaxis
      real(dop), intent(in)                     :: theta
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      real(dop), dimension(3,3)                 :: rotmat
      integer(long) :: i,j,n

      rotmat = 0.0_dop
      if (iaxis .eq. 1) then
         rotmat(1,1) = 1.0_dop
         rotmat(2,2) = cos(2.0_dop*theta)
         rotmat(3,3) = -rotmat(2,2)
         rotmat(3,2) = sin(2.0_dop*theta)
         rotmat(2,3) = rotmat(3,2)
      else if (iaxis .eq. 2) then
         rotmat(2,2) = 1.0_dop
         rotmat(3,3) = cos(2.0_dop*theta)
         rotmat(1,1) = -rotmat(3,3)
         rotmat(3,1) = sin(2.0_dop*theta)
         rotmat(1,3) = rotmat(3,1)
      else if (iaxis .eq. 3) then
         rotmat(3,3) = 1.0_dop
         rotmat(1,1) = cos(2.0_dop*theta)
         rotmat(2,2) = -rotmat(1,1)
         rotmat(2,1) = sin(2.0_dop*theta)
         rotmat(1,2) = rotmat(2,1)
      endif
      
      do n=1,natm

         do i = 1,3
            xout(i,n) = rotmat(i,1)*xin(1,n)
         enddo
         do j = 2,3
            do i = 1,3
               xout(i,n) = xout(i,n)+rotmat(i,j)*xin(j,n)
            enddo
         enddo

      enddo

      return
      end subroutine mirrord

!-----------------------------------------------------------------------
! Rotate a vector of Cartesian coordinates by 2pi/nang around an axis
! iaxis = 1, 2, 3  (x, y, z respectively)
!-----------------------------------------------------------------------

      subroutine rotate(xin,xout,natm,nang,iaxis)

      implicit none

      integer(long), intent(in) :: natm, iaxis, nang
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      real(dop), dimension(3,3)                 :: rotmat
      real(dop)                                 :: theta
      integer(long) :: i,j,n

      theta = 2.0_dop*pi / dble(nang)
      rotmat = 0.0_dop
      if (iaxis .eq. 1) then
         rotmat(1,1) = 1.0_dop
         rotmat(2,2) = cos(theta)
         rotmat(3,3) = cos(theta)
         rotmat(3,2) = sin(theta)
         rotmat(2,3) = -rotmat(3,2)
      else if (iaxis .eq. 2) then
         rotmat(2,2) = 1.0_dop
         rotmat(1,1) = cos(theta)
         rotmat(3,3) = cos(theta)
         rotmat(3,1) = sin(theta)
         rotmat(1,3) = -rotmat(3,1)
      else if (iaxis .eq. 3) then
         rotmat(3,3) = 1.0_dop
         rotmat(1,1) = cos(theta)
         rotmat(2,2) = cos(theta)
         rotmat(2,1) = sin(theta)
         rotmat(1,2) = -rotmat(2,1)
      endif
      
      do n=1,natm

         do i = 1,3
            xout(i,n) = rotmat(i,1)*xin(1,n)
         enddo
         do j = 2,3
            do i = 1,3
               xout(i,n) = xout(i,n)+rotmat(i,j)*xin(j,n)
            enddo
         enddo

      enddo

      return
      end subroutine rotate


!-----------------------------------------------------------------------
! Improper rotation on a vector of Cartesian coordinates by 2pi/nang 
! around an axis 
! iaxis = 1, 2, 3  (x, y, z respectively)
!-----------------------------------------------------------------------

      subroutine improt(xin,xout,natm,nang,iaxis)

      implicit none

      integer(long), intent(in) :: natm, iaxis, nang
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      real(dop), dimension(3,3)                 :: rotmat
      real(dop)                                 :: theta
      integer(long) :: i,j,n

      theta = 2.0_dop*pi / dble(nang)
      rotmat = 0.0_dop
      if (iaxis .eq. 1) then
         rotmat(1,1) = -1.0_dop
         rotmat(2,2) = cos(theta)
         rotmat(3,3) = cos(theta)
         rotmat(3,2) = sin(theta)
         rotmat(2,3) = -rotmat(3,2)
      else if (iaxis .eq. 2) then
         rotmat(2,2) = -1.0_dop
         rotmat(1,1) = cos(theta)
         rotmat(3,3) = cos(theta)
         rotmat(3,1) = sin(theta)
         rotmat(1,3) = -rotmat(3,1)
      else if (iaxis .eq. 3) then
         rotmat(3,3) = -1.0_dop
         rotmat(1,1) = cos(theta)
         rotmat(2,2) = cos(theta)
         rotmat(2,1) = sin(theta)
         rotmat(1,2) = -rotmat(2,1)
      endif
      
      do n=1,natm

         do i = 1,3
            xout(i,n) = rotmat(i,1)*xin(1,n)
         enddo
         do j = 2,3
            do i = 1,3
               xout(i,n) = xout(i,n)+rotmat(i,j)*xin(j,n)
            enddo
         enddo

      enddo

      return
      end subroutine improt


!-----------------------------------------------------------------------
! Inversion in the origin of a vector of Cartesian coordinates
!-----------------------------------------------------------------------

      subroutine inversion(xin,xout,natm)

      implicit none

      integer(long), intent(in) :: natm
      real(dop), dimension(3,natm), intent(in)  :: xin
      real(dop), dimension(3,natm), intent(out) :: xout
      integer(long) :: i

      do i=1,3
        xout(i,:) = -xin(i,:)
      enddo

      return
      end subroutine inversion

!-----------------------------------------------------------------------
! Reflect a tensor of Cartesian coordinates through a plane specified 
! by the "axis" normal to the plane
! iaxis = 1, 2, 3  (x, y, z respectively)
!-----------------------------------------------------------------------

      subroutine mirror2(matin,matout,natm,iaxis)

      implicit none

      integer(long), intent(in) :: natm, iaxis
      real(dop), dimension(3,natm,3,natm), intent(in)  :: matin
      real(dop), dimension(3,natm,3,natm), intent(out) :: matout
      integer(long) :: i,j

      do i=1,3
         do j=1,3
            if (i .eq. j) then
               matout(j,:,i,:) = matin(j,:,i,:)
            else if (i .eq. iaxis) then
               matout(j,:,i,:) = -matin(j,:,i,:)
            else if (j .eq. iaxis) then
               matout(j,:,i,:) = -matin(j,:,i,:)
            else
               matout(j,:,i,:) = matin(j,:,i,:)
            endif
         enddo
      enddo

      return
      end subroutine mirror2

!-----------------------------------------------------------------------

      end module symoplib
