C***********************************************************************
C
C                Module WIGMAT
C
C Contents: bigwigD, djmk
C
C Calculates Wigner d-matrix elements d^j_{m k} (beta)
C
C***********************************************************************
C#######################################################################

C-----------------------------------------------------------------------
C           Function bigwigD
C           Wigner (big)-D-matrix elements D^j_{m k} (alpha,beta,gamma)
C
C input: angular momentum quantum numbers j, m, k (integers)
C        Euler angles in radians alp_rad, beta_rad, gam_rad
C        set norfcl to .true. to normalize by sqrt((2j+1)/8pi^2)
C-----------------------------------------------------------------------

      complex*16 function bigwigD (j,m,k,alp_rad,beta_rad,gam_rad,
     &     norfcl)

      implicit none

      integer    j,m,k
      real*8     alp_rad,beta_rad,gam_rad,nrfc,Pi,djmk
      complex*16 exp_m,exp_k
      logical    norfcl

      Pi=3.14159265358979323846

      if (norfcl) then ! Normalize the exponential...
         nrfc=1.d0/(2.d0*Pi)
      else             ! ...or cancel the norm from djmk()
         nrfc=sqrt(2.d0/dble(2*j+1))
      endif

      exp_m=exp(dcmplx(0.d0,-dble(m)*alp_rad))
      exp_k=exp(dcmplx(0.d0,-dble(k)*gam_rad))

      bigwigD = nrfc*exp_m*exp_k*djmk(j,m,k,cos(beta_rad))

      return
      end

C#######################################################################
C            Function djmk
C
C  Calculates Wigner small-d matrix elements, d^j_m,k(x), where
C  x=cos(beta) is in the range [-1,1]
C  This function returns the matrix elements in L^2-normalized form
C
C  P. Thomas 04/2013
C-----------------------------------------------------------------------

      real*8 function djmk (j,mm,kk,x)

      implicit none

      integer j,mm,kk,m,k,sym,i,i1,i2,k2,m2,mpo,mmpo,ktmo
      real*8  x,dlo,dhi,dhi2,fac,fc,tbt
      real*8  jpo,jpo2,tjo,sf1,sf2

      if (j.lt.0 .or. abs(mm).gt.j .or. abs(kk).gt.j .or.
     +    abs(x).gt.1) then
         write(6,*) 'ERROR: bad arguments in djmk ',j,mm,kk,x
         write(2,*) 'ERROR: bad arguments in djmk ',j,mm,kk,x
         stop
      endif

      m=mm
      k=kk
      sym=1

C --- Modify indices if required by symmetry

      if (abs(mm).lt.abs(kk)) then
         m=kk
         k=mm
         if (mod(abs(mm-kk),2).eq.1) sym=-sym
      endif
      if (m.lt.0) then
         m=-m
         k=-k
         if (mod(abs(m-k),2).eq.1) sym=-sym
      endif

C --- Case for x=-1 (easy)

      if (abs(x+1.d0).le.1.0d-15) then
         if (k.eq.-m) then
            djmk=sqrt((2*j+1.d0)/2.d0)*(-1.d0)**(j+k)
         else
            djmk=0.d0
         endif
         return
      endif

C --- Calculate d^m_m,k and d^m+1_m,k for 3 term recurrence on j
C     'dlo'  is d^m_m,0     (for k.le.0) or d^m_m,m       (for k.gt.0)
C     'dhi'  is d^m+1,m,0   (for k.le.0) or d^m+1_m,m     (for k.gt.0)
C     'dhi2' is d^m+1,m+1,0 (for k.le.0) or d^m+1_m+1,m+1 (for k.gt.0)
      djmk=1.d0
      if (k.le.0) then   ! Use Legendre-style recurrences, for k=0
         fac=(1.d0+x)*(1.d0-x)
         fc=1.d0
         do i=1,m
            djmk=djmk*fac*fc/(fc+1.d0)
            fc=fc+2.d0
         enddo
         dhi2=sqrt(djmk*fac*fc/(fc+1.d0))
         djmk=sqrt(djmk)
         if (mod(m,2).eq.1) then
            djmk=-djmk
         else
            dhi2=-dhi2
         endif
         k2=0
         fc=x*sqrt(2*m+1.d0)
      else               ! Use Wigner-style recurrences, for k=m
         fac=(1.d0+x)/2.d0
         do i=1,m
            djmk=djmk*fac
         enddo
         dhi2=djmk*fac
         k2=m
         fc=(m*(x-1.d0)+x)
      endif
      dhi=djmk*fc

C --- Calculate d^m_m,k and d^m+1_m,k for k.ne.0 and k.ne.m

      if (k.lt.m .and. k.ne.0) then ! calc d^m_mk from d^m_mm or d^m_m0
         tbt = -sqrt((1.d0-x)/(1.d0+x))
         mpo=m+1
         mmpo=mpo*(m+2)
         if (k.gt.0) dhi2=tbt*dhi2*sqrt(2*(m+1.d0)) ! Need d^m+1,m+1,m
         i1=m+k2
         i2=m-k2+1
         do i=k2,k+1,-1
            ktmo=mmpo-i*(i-1)
            fac=sqrt(dble(i1)/dble(i2))
            djmk=djmk*tbt*fac
            dhi=sqrt(dble(2*mpo)/(dble(ktmo)))*dhi2 +
     +          dble(m+i)/sqrt(dble(ktmo))*tbt*dhi
            dhi2=dhi2*tbt*sqrt((i1+1.d0)/(i2+1.d0))
            i1=i1-1
            i2=i2+1
         enddo
      endif

      if (j.eq.m+1) djmk=dhi

C --- Calculate d^j_mk via 3-term recurrence on j
      if (j.gt.m+1) then
         dlo=djmk
         k2=k*k
         m2=m*m
         do i=m+1,j-1
            jpo=i+1
            jpo2=jpo*jpo
            tjo=2*i+1
            sf1=dble(jpo)/sqrt(dble((jpo2-k2)*(jpo2-m2)))
            sf2=sqrt(dble((i*i-k2)*(i*i-m2)))
            djmk=sf1*(tjo*(x-dble(k*m)/dble(i*jpo))*dhi -
     +           sf2/dble(i)*dlo)
            dlo=dhi
            dhi=djmk
         enddo
      endif

      djmk=sqrt((2*j+1.d0)/2.d0)*sym*djmk

      return
      end

C#######################################################################

