!   -*- f90 -*-
! 
! module conjgrad
!
!
!

module conjgrad
 use conjgrad_m

  implicit none

contains

!*******************************************************************
! conjugate gradients
!
! Solve Omega**T Omega * x = b with:
!
! Omega**T Omega is not known explicitly. Instead a subroutine
! operatung Omega**T Omega is passed.
! x: coefficiens of natpot
! b: Monte-Carlo overlap of SPP with V
!
! On entry, coeff contains the Monte-Carlo overlap of SPP with V
! On exit coeff contains the calculated coefficients.
! Epsilon is the error tolerance that needs to be met by x.
!
! MPI parallelitazion is only done over sampling points
!
!*******************************************************************

subroutine conjgradsingle(coeff,modcdim,epsilon,maxiter,ierr)
  use pftools
  use mpidata
  use global
  use channels
  use logdat
#ifdef MPI
  use mpi
#endif

  implicit none

  external OperateOmega

  integer, intent(IN)             :: modcdim, maxiter
  double precision, intent(INOUT) :: coeff(modcdim)
  double precision, intent(IN)    :: epsilon
  integer, intent(OUT)            :: ierr

  double precision, allocatable   :: rm(:), pm(:), vm(:), mrm(:)
  double precision                :: alpham, alpham1, err
  double precision                :: lambdam, spv

  double precision                :: beta, absb
  integer                         :: iiter
  logical                         :: lstop1,ldone,lskip(1), lpcwarn

  !BLAS
  double precision                :: ddot
  external ddot

#ifdef MPI
  double precision, allocatable   :: mpibuf(:)
#endif

  interface
     subroutine CheckStopFile(lstopnow, stopmessage)
       logical, intent(OUT)                  :: lstopnow
       character(len=*), intent(IN),optional :: stopmessage
     end subroutine CheckStopFile
  end interface

  if (mpirank.eq.0) then
     open(ilancz,file=trim(name)//"/conjgrad.log")
     write(ilancz,'(a)') "#"
     write(ilancz,'(a)') "# Starting conjugate gradients..."

     call myflush(ilancz)
     !write(6,'(a)') " Starting conjugate gradients..."

     absb = dsqrt(dot_product(coeff,coeff))

  end if

  allocate(rm(modcdim))
  allocate(mrm(modcdim))
  allocate(pm(modcdim))
  allocate(vm(modcdim))

  ! start up
  ! we solve Ax = B, where A = omega**T * omega, B = omega**T * Vm
  ! we take coeff == B and x0 = B on startup

  ierr = 0
  iiter = 0
  lskip = .false.
  ldone = .false.

#ifdef MPI
  call MPI_Bcast(coeff,modcdim, MPI_DOUBLE_PRECISION, &
       0,MPI_COMM_WORLD,ierr)
  if (mpirank.eq.0) then
     allocate(mpibuf(modcdim))
  end if
#endif


  call OperateOmega(coeff,rm,modcdim,1,lskip)



#ifdef MPI
  call MPI_Reduce(rm,mpibuf,modcdim,MPI_DOUBLE_PRECISION,&
       MPI_SUM,0,MPI_COMM_WORLD,ierr)

  if (mpirank.eq.0) then
     !$omp parallel workshare
     rm = mpibuf
     !$omp end parallel workshare
  end if
#endif

  call init_mcprecond(lpcwarn)

  if (mpirank.eq.0) then

     if (lpcwarn) then
        routine = "conjgradsingle"
        message = "Preconditioner matrix is not positive definite."
        call errormsg
     end if


     write(ilancz,'(a)') "#"
     write(ilancz,'(a)') "#   ITER  Error"
     !write(6,'(a)') ""
     !write(6,'(a)') "    ITER  Error"

     rm = coeff - rm

  end if

  call operate_mcprecond(rm,mrm,1)

  if (mpirank.eq.0) then
     pm = mrm
     alpham = dot_product(rm,mrm)
  end if

  lstop1 = .false.

  ! main lop
  do while(.not. lstop1)


#ifdef MPI
     call MPI_Bcast(pm,modcdim,MPI_DOUBLE_PRECISION, &
          0,MPI_COMM_WORLD,ierr)
#endif

     call operateOmega(pm,vm,modcdim,1,lskip)


#ifdef MPI
     call MPI_Reduce(vm,mpibuf,modcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank.eq.0) then
        !$omp parallel workshare
        vm = mpibuf
        !$omp end parallel workshare
     end if
#endif


     if (mpirank.eq.0) then

        spv = ddot(modcdim,pm,1,vm,1)

        if (spv .eq. 0.d0) then
           if (ierr .eq. 0) ierr = 1
        elseif (spv .lt. 0.d0) then
           ! marix is not positive definite
           if (ierr .eq. 0) ierr = -1
           ldone = .true.
        end if


        lambdam = alpham / spv

        alpham1 = alpham


        coeff = coeff + lambdam*pm
        rm = rm - lambdam*vm
     end if

     call operate_mcprecond(rm,mrm,1)

     if (mpirank.eq.0) then
        alpham = dot_product(rm,mrm)
        err = dot_product(rm,rm)/absb

        iiter = iiter + 1

        if (err .lt. epsilon) then
           ldone = .true.
        end if

        ! check if the user wants to stop
        message = "CG iteration"
        write(message(13:),'(i8)') iiter
        call CheckStopFile(lstop1, message)

        lstop1 = lstop1 .or. ldone

        if (.not. lstop1) then

           beta = alpham/alpham1
           !$omp parallel workshare
           pm = mrm + beta*pm
           !$omp end parallel workshare
        end if


        lstop1 = lstop1 .or. iiter .ge. maxiter

     end if ! mpirank



#ifdef MPI
     call MPI_Bcast(lstop1,1,MPI_LOGICAL,0, &
          MPI_COMM_WORLD,ierr)
#endif


     if (mpirank.eq.0) then
        write(ilancz,'(i8,1x,e23.16)') iiter, err
        !write(*,'(i8,1x,e23.16)') iiter, err
        call myflush(ilancz)
        if (lstop1) then
           write(ilancz,'(a)') "# Stopping iterations."
        end if


     end if


  end do ! iterations


  ! not converged
  if (ierr.eq.0) then
     if (.not. ldone ) ierr = 2
  end if

  if (mpirank.eq.0) close(ilancz)


  deallocate(rm)
  deallocate(mrm)
  deallocate(pm)
  deallocate(vm)



#ifdef MPI
  if (allocated(mpibuf)) deallocate(mpibuf)
#endif

end subroutine Conjgradsingle


!###################################################################
!*******************************************************************
! conjugate gradients version2
!
! Solve A*x = b with:
!
! A = Omega^T Omega
! x = coefficiens of natpot
! b = Monte-Carlo overlap of SPP with V
!
! coeff: On entry coeff contains b, on exit, coeff contains x
!
! MPI parallelitazion is only done over sampling points so far
!
!*******************************************************************

subroutine BLKconjgrad(coeff,modcdim,NRHS,epsilon,maxiter,ierr)
  use pftools
  use mcpfitparams
  use mpidata
  use global
  use channels
  use logdat
  use schmidtorthomod, only: schmidtorthor
#ifdef MPI
  use mpi
#endif

  implicit none
  integer, intent(IN)             :: modcdim, NRHS, maxiter
  double precision, intent(INOUT) :: coeff(modcdim, NRHS)
  double precision, intent(IN)    :: epsilon
  integer, intent(OUT)            :: ierr

  double precision, allocatable   :: pm(:,:), qm(:,:)
  double precision, allocatable   :: sm(:,:), tm(:,:), xm(:,:)
  double precision, allocatable   :: cm(:,:), zm(:,:), tcm(:,:), cz(:,:)
  double precision, allocatable   :: qz(:,:), Ap(:,:), pf(:,:), zqm(:,:)

  logical, allocatable            :: ldone(:)
  double precision, allocatable   :: err(:)


  double precision                :: rtmp
  integer                         :: iiter, i,j,k,l
  logical                         :: lstop1, lpcwarn


  double precision, allocatable   :: Q(:,:), R(:,:)
  integer, allocatable            :: P(:)

  integer                         :: rank, rank1, minnm, warncount


#ifdef MPI
  double precision, allocatable   :: mpibuf(:,:)
#endif

  interface
     subroutine CheckStopFile(lstopnow, stopmessage)
       logical, intent(OUT)                  :: lstopnow
       character(len=*), intent(IN),optional :: stopmessage
     end subroutine CheckStopFile
  end interface


  lwarn = .false.
  warncount = 0

!!$  Solve: AX=B, preconditionier M ~ A**-1, I = identity
!!$
!!$  X_0 = 0
!!$  R_0 = B
!!$  Z_0 = M R_0
!!$  S_0 = I
!!$  P_0 = 0
!!$
!!$  do k=1,2,3,...
!!$
!!$     P_k C_k = Z_k-1 + P_k-1 S_k-1
!!$
!!$     T_k = P_k^T R_k-1
!!$
!!$     X_k = X_k-1 + P_k T_k
!!$
!!$     Q_k F_k = Q_k-1 - A P_k T_k
!!$
!!$     Z_k = M Q_k
!!$
!!$     C_k = F_k C_k-1
!!$
!!$     S_k = - P_k^T A Z_k
!!$
!!$  done
!!$
!!$  decompositions: P_k are A conjugate, Q_k are orthonormal


  if (mpirank.eq.0) then
     open(ilancz, file=trim(name)//"/conjgrad.log")
     write(ilancz,'(a)') "#"
     write(ilancz,'(a)') "# Starting (block) conjugate gradients."
     call myflush(ilancz)
     !write(6,'(a)') ""
     !write(6,'(a)') " Starting (block) conjugate gradients."
  end if

  ierr = 0

  if (mpirank.eq.0) then

     ! first, we try to reduce the number of right hand sides
     ! by finding redundant vectors using a rank-revealing
     ! QR-Docomposition. This is done with Gram-Schmidt

     write(ilancz,'(a)') '# Trying to deflate the contacted mode...'
     write(ilancz,'(a)') "#"
     !write(*,'(a)')' Trying to deflate the contacted mode...'
     !write(*,'(a)') ""

     minnm = min(modcdim,NRHS)
     allocate(Q(modcdim,minnm))
     allocate(R(NRHS,minnm))
     allocate(P(NRHS))

     !stop
     ! rank revealing QR decomposition of (Ax -b)
     ! to reduce the numner of rhs
     call GramSchmidtDecompose(coeff,Q,R,P,modcdim,NRHS,minnm, &
          rank1,1.d-16,ierr)

     if (ierr.eq.2) then
        routine = "BLKconjgrad"
        message = "Something went terribly wrong." // &
             " This is a softare error."
        call errormsg()
     end if
     if (ierr.eq.1) then
        routine = "BLKConjgrad"
        message = "QR decomposition of the coefficient matrix failed." // &
             " The result is probably garbage."
        call warnmsg()
     end if

     write(ilancz,'(a,i6)') "# Number of independent points " &
          //"on contracted mode:", rank1
     !       write(6,'(a,i6)')' Number of independent points " &
     !            //"on contracted mode:', rank1

     if (cg_inflate .gt. 0) then
        rank = rank1 + cg_inflate
        if (rank.gt. modcdim) rank = modcdim
        !write(6,'(a,i6)') " Added artificial vectors:", rank-rank1
        write(ilancz,'(a,i6)') "# Added artificial vectors:", &
             cg_inflate
        !write(6,'(a,i6)') " Total linear independent vectors:", rank
        write(ilancz,'(a,i6)') "#  Total linear independent vectors",  &
             cg_inflate
     else
        rank = rank1
     end if

     call myflush(ilancz)


  end if

#ifdef MPI
  call MPI_Bcast(rank1,1, MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  call MPI_Bcast(rank,1, MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
  if (mpirank.eq.0) allocate(mpibuf(modcdim,rank))

#endif

!   call init_mcprecond(lpcwarn, Q, rank)
!    if (mpirank.eq.0) then
!       if (lpcwarn) then
!          routine = "BLKconjgrad"
!          message = "Preconditioner matrix is not positive definite."
!          call errormsg
!       end if
!    end if


  ! temporary variables for conj-grad
  ! these must be known in all ranks
  allocate(pf(modcdim, rank))
  allocate(Ap(modcdim, rank))
  allocate(pm(modcdim, rank))
  allocate(zqm(modcdim, rank))
  allocate(qm(modcdim, rank))

  ! startup of conjgrad
  if (mpirank.eq.0) then

     ! temporary variables for conj-grad
     ! these can be left in rank 0 only

     allocate(xm(modcdim, rank))

     allocate(qz(modcdim, rank))


     allocate(cm(rank, rank))
     allocate(cz(rank, rank))
     allocate(sm(rank, rank))
     allocate(tm(rank, rank))
     allocate(zm(rank, rank))
     allocate(tcm(rank, rank))

     allocate(err(rank))

     xm = 0.d0
     qm(:,1:rank1) = Q(:,1:rank1)

     cm = 0.d0
     do i=1,rank
        cm(i,i) = 1.d0
     end do
     sm = 0.d0
     do i=1,rank
        sm(i,i) = 1.d0
     end do
     pm = 0.d0

  end if

  allocate(ldone(rank))
  ldone = .false.

  if (mpirank.eq.0) then
     if (cg_inflate .ne. 0) then

        do j=1,cg_inflate
           do i=1,modcdim
              qm(:,j+rank1) = qm(:,j)*dble(i)
           end do
        end do

        call schmidtorthor(qm,modcdim,rank,ierr,rtmp)

        if (ierr.eq.0) then
           message = "Oththogonalization of the inflated system failed"
           routine = "BlkConjgrad"
           call errormsg
        end if
     end if
  end if

  call init_mcprecond(lpcwarn)

  if (mpirank.eq.0) then

     if (lpcwarn) then
        routine = "BlkConjgrad"
        message = "Preconditioner matrix is not positive definite."
        call errormsg
     end if


     write(ilancz,'(a)') "#"
     write(ilancz,'(a)') "#   ITER  Error(mean)             Error(max)"
     call myflush(ilancz)

     !write(6,'(a)') ""
     !write(6,'(a)') "    ITER  Error(mean)             Error(max)"
  end if


  call operate_mcprecond(qm,zqm,rank)


  lstop1 = .false.
  iiter = 0

  ! MAIN LOOP
  iterations: do while(.not. lstop1)


     if (mpirank.eq.0) then


        ! P F = Z + P S
        pf = zqm

        if (iiter .ne. 0) then
           ! pm == 0 in the 0th iteration

           !$omp parallel do private(i,j)
           do k=1,rank
              do j=1,rank
                 do i=1,modcdim
                    pf(i,k) = pf(i,k) + pm(i,j)*sm(j,k)
                 end do
              end do
           end do
           !$omp end parallel do

        end if ! iiter != 0


     end if ! mpirank == 0

     iiter = iiter + 1

#ifdef MPI

     call MPI_Bcast(pf,modcdim*rank,  &
          MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#endif

     call OperateOmega(pf,Ap,modcdim,rank,ldone)

#ifdef MPI
     call MPI_Reduce(Ap,mpibuf,rank*modcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank.eq.0) then
        Ap = mpibuf
     end if
#endif

     if (mpirank .eq. 0) then

        ! in-situ A-orthonormalization
        call aortho(pf, Ap, modcdim, rank,ierr)

        if (ierr.lt.0) then
           routine = "BLKconjgrad"
           message = "Overlap matrix is singular. Use more sampling points."
           call errormsg
        else if(ierr.gt.0) then
           warncount = warncount + 1
           if (modulo(warncount,5) .eq.0) then
              routine = "BLKconjgrad"
              message = "Linear dependent vectors detected several times during " //&
                   "A-orthogonalization. This is probably fatal. Use more sampling points?"
              call warnmsg
           end if
        end if


        pm = pf

        ! T = P^T Q
        !$omp parallel do private(j)
        do i=1,rank
           do j=1,rank
              tm(j,i) = dot_product(pm(:,j), qm(:,i))
           end do
        end do
        !$omp end parallel do



        tcm = 0.d0
        !$omp parallel do private(j,k)
        do l=1,rank
           do k=1,rank
              do j=1,rank
                 tcm(j,l) =  tcm(j,l) + tm(j,k)*cm(k,l)
              end do
           end do
        end do
        !$omp end parallel do

        ! X = X + P T C
        !$omp parallel do private(i,j)
        do k=1,rank
           do j=1,rank
              do i=1,modcdim
                 xm(i,k) = xm(i,k) + pm(i,j)*tcm(j,k)
              end do
           end do
        end do
        !$omp end parallel do

        ! Q Z =  Q - A P T
        qz = qm
        !$omp parallel do private(i,j)
        do k=1,rank
           do j=1,rank
              do i=1,modcdim
                 qz(i,k) = qz(i,k) - ap(i,j)*tm(j,k)
              end do
           end do
        end do
        !$omp end parallel do

        ! QZ is the residual at this point, however, scaled as
        ! the multoplication with cm is missing, compare:
        !
        !     delta_X =   P T C
        ! ==> delta_R = A P T C
        !
        ! Here delta_R is only  A P T; C is updated
        ! after the QR decomposition

        call QR(qz,qm,zm,modcdim,rank)
     end if ! mpirank 0

     call operate_mcprecond(qm,zqm,rank)

     if (mpirank.eq. 0) then

        cz = 0.d0
        !$omp parallel do private(i,j)
        do k=1,rank
           do j=1,rank
              do i=1,rank
                 cz(i,k) = cz(i,k) + zm(i,j)*cm(j,k)
              end do
           end do
        end do
        !$omp end parallel do
        cm = cz

        ! C is mow updated. It is the coeffitient matrix
        ! of the QR-Decomposed residual, the diagonal
        ! therefore contains the error

        !$omp parallel do
        do i=1,rank
           err(i) = cm(i,i)
        end do
        !$omp end parallel do

        rtmp = 0.d0
        !$omp parallel do reduction(+:rtmp)
        do i=1,rank1
           rtmp = rtmp + err(i)
        end do
        !$omp end parallel do

        write(ilancz,'(i8,1x,e23.16,1x,e23.16)') iiter,  &
             rtmp/rank1, maxval(err(1:rank1))
        !          write(*,'(i8,1x,e23.16,1x,e23.16)') iiter,  &
        !               rtmp/rank1, maxval(err(1:rank1))
        call myflush(ilancz)

        call CheckStopFile(lstop1, message)

        lstop1 = lstop1 .or. all(err(1:rank1) .lt. epsilon)  &
             .or. iiter .ge. maxiter

     end if ! mpirank rank 0

#ifdef MPI
     ! make known in all ranks if we are done
     call MPI_Bcast(lstop1,1,MPI_LOGICAL,0, &
          MPI_COMM_WORLD,ierr)
#endif

     if (lstop1) exit

     if (mpirank .eq. 0) then
        ! S = - P^T A M Q
        sm = 0.d0
        !$omp parallel do private(j)
        do i=1,rank
           do j=1,rank
              sm(j,i) =  -dot_product(ap(:,j), zqm(:,i))
           end do
        end do
        !$omp end parallel do

     end if ! mpirank == 0


  end do iterations


  if (mpirank.eq.0) then

     ldone = err .lt. epsilon
     if (.not. all(ldone)) then
        ierr = 2
        !write(6,'(a)') "NOT converged. Stopped iterations."
        write(ilancz,'(a)') "# NOT converged. Stopped iterations."

        if (iiter .ge. maxiter) then
           !write(6,'(a)') "Reached maximum number of Iterations"
           write(ilancz,'(a)') "# Reached maximum number of Iterations"
        end if
     else
        !write(6,'(a)') "Converged. Stopped iterations."
        write(ilancz,'(a)') "# Converged. Stopped iterations."
     end if
     call myflush(ilancz)
  end if


  deallocate(pf)
  deallocate(Ap)
  deallocate(pm)


  ! startup of conjgrad
  if (mpirank.eq.0) then
     ! free temporary variables for conj-grad
     deallocate(qm)
     deallocate(cm)
     deallocate(sm)
     deallocate(tm)
     deallocate(zm)
     deallocate(zqm)
     deallocate(qz)
     deallocate(tcm)
     deallocate(err)
  end if

  ! undo the initial QR decomposition
  if (mpirank.eq.0) then
     ! re-use gz for a minute..
     allocate(qz(modcdim,NRHS))
     call GramSchmidtReconst(qz,xm,R,P,modcdim,NRHS,rank1)
     coeff = qz
     deallocate(qz)
     deallocate(xm)
     deallocate(Q)
     deallocate(R)
     deallocate(P)
     close(ilancz)
  end if

#ifdef MPI
  if (allocated(mpibuf)) deallocate(mpibuf)
#endif

end subroutine BLKconjgrad

end module conjgrad
