!   -*- f90 -*-
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! module dvrtype
! contains the definition of type dvr_t and basic initialization,
! destruction and IO routines for this type.
!
! Note: dvr_t sticks very close to the original MCTDH data structure.
!       This can probably be changed in the future.
!
!       'export*' routines are compatibility routines for MCTDH-readable
!       files
!
! 09/13 MS
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module dvrtype_m
    use maxv
    use dvrdatmod
    use griddatmod
    use logdvrmod

#ifdef MPI
    use mpidata
    use mpi
#endif


  implicit none

  ! lblen == c1 (global.inc)
  integer, parameter, private :: lblen = 16

  ! type DVR - contains the primitive grid
  type dvr_t
     logical                              :: initialized = .false.

     ! DVRVER
     double precision                     :: filever = 0.d0

     ! DVRDEF
     integer                              :: ndof
     integer                              :: mbaspar
     character(len=lblen), allocatable    :: modelabel(:)
     integer, allocatable                 :: gdim(:)
     integer, allocatable                 :: basis(:)
     logical, allocatable                 :: ldvr(:)
     integer, allocatable                 :: ipbaspar(:,:)
     double precision, allocatable        :: rpbaspar(:,:)
     double precision, allocatable        :: xend(:,:)

     ! DVRINFO
     integer                              :: ortdim = 1, dvrdim = 1
     integer                              :: fftdim = 1, expdim = 1
     integer                              :: sphdim = 1

     integer, allocatable                 :: zort(:), zdvr(:)
     integer, allocatable                 :: zfft(:), zexp(:)
     integer, allocatable                 :: zsph(:)

     ! DVRDATA
     double precision, allocatable        :: ort(:), trafo(:)
     integer                              :: dvrord = 2
     double precision, allocatable        :: dvrmat(:,:)
     double precision, allocatable        :: fftp(:)
     complex*16, allocatable              :: hin(:), rueck(:)
     integer, allocatable                 :: fftfak(:)
     complex*16, allocatable              :: exphin(:), exprueck(:)

     integer, allocatable                 :: jsph(:), msph(:)
     double precision, allocatable        :: kinsph(:)

  end type dvr_t


contains

  function dof_num_by_label(dvr, label) result(num)
    implicit none
    type(dvr_t), intent(IN)      :: dvr
    character(len=*), intent(IN) :: label
    integer                      :: num
    integer                      :: f

    num = 0
    do f=1,dvr%ndof
       if (dvr%modelabel(f) .eq. label) then
          num = f
          exit !loop
       end if
    end do

  end function dof_num_by_label


  subroutine shuffle_dof_order(dvr, neworder)
    ! f1 = neworder(f)
    ! the f'th dof of the input dvr
    ! is the f1'th of the output dvr

    implicit none
    type(dvr_t), intent(INOUT)  :: dvr
    integer, intent(IN)         :: neworder(dvr%ndof)
    integer                     :: astat
    integer                     :: f,f1

    character(len=lblen), allocatable :: modelabel(:)
    integer, allocatable              :: gdim(:), basis(:), ipbaspar(:,:)
    double precision, allocatable     :: rpbaspar(:,:), xend(:,:)
    logical, allocatable              :: ldvr(:)
    integer, allocatable              :: zort(:), zdvr(:)
    integer, allocatable              :: zfft(:), zexp(:)
    integer, allocatable              :: zsph(:)


    allocate(modelabel(dvr%ndof))
    allocate(gdim(dvr%ndof))
    allocate(basis(dvr%ndof))
    allocate(ipbaspar(dvr%mbaspar, dvr%ndof))
    allocate(rpbaspar(dvr%mbaspar, dvr%ndof))
    allocate(xend(2,dvr%ndof))
    allocate(ldvr(dvr%ndof))
    allocate(zort(dvr%ndof))
    allocate(zdvr(dvr%ndof))
    allocate(zfft(dvr%ndof))
    allocate(zexp(dvr%ndof))
    allocate(zsph(dvr%ndof))


    ! We just need to shuffle some things, for the rest we swap pointers

    do f1=1,dvr%ndof
       f = neworder(f1)
       modelabel(f1)  = dvr%modelabel(f)
       gdim(f1)       = dvr%gdim(f)
       basis(f1)      = dvr%basis(f)
       ipbaspar(:,f1) = dvr%ipbaspar(:,f)
       rpbaspar(:,f1) = dvr%rpbaspar(:,f)
       xend(:,f1)     = dvr%xend(:,f)
       ldvr(f1)       = dvr%ldvr(f)
       zort(f1)       = dvr%zort(f)
       zdvr(f1)       = dvr%zdvr(f)
       zfft(f1)       = dvr%zfft(f)
       zexp(f1)       = dvr%zexp(f)
       zsph(f1)       = dvr%zsph(f)
    end do


    do f=1,dvr%ndof
       dvr%modelabel(f)  = modelabel(f)
       dvr%gdim(f)       = gdim(f)
       dvr%basis(f)      = basis(f)
       dvr%ipbaspar(:,f) = ipbaspar(:,f)
       dvr%rpbaspar(:,f) = rpbaspar(:,f)
       dvr%xend(:,f)     = xend(:,f)
       dvr%ldvr(f)       = ldvr(f)
       dvr%zort(f)       = zort(f)
       dvr%zdvr(f)       = zdvr(f)
       dvr%zfft(f)       = zfft(f)
       dvr%zexp(f)       = zexp(f)
       dvr%zsph(f)       = zsph(f)
    end do

    deallocate(modelabel)
    deallocate(gdim)
    deallocate(basis)
    deallocate(ipbaspar)
    deallocate(rpbaspar)
    deallocate(xend)
    deallocate(ldvr)
    deallocate(zort)
    deallocate(zdvr)
    deallocate(zfft)
    deallocate(zexp)
    deallocate(zsph)


  end subroutine shuffle_dof_order



  subroutine copy_dvr_t(dvr_in, dvr_out)
    implicit none
    type(dvr_t), intent(IN)              :: dvr_in
    type(dvr_t), intent(INOUT)           :: dvr_out
    integer                              :: astat,f

    call deallocate_dvr_t(dvr_out)

    ! scalars
    dvr_out%ndof    = dvr_in%ndof
    dvr_out%mbaspar = dvr_in%mbaspar
    dvr_out%ortdim  = dvr_in%ortdim
    dvr_out%dvrdim  = dvr_in%dvrdim
    dvr_out%fftdim  = dvr_in%fftdim
    dvr_out%expdim  = dvr_in%expdim
    dvr_out%sphdim  = dvr_in%sphdim
    dvr_out%dvrord  = dvr_in%dvrord
    dvr_out%filever = dvr_in%filever

    ! allocs
    allocate(dvr_out%zort(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%zort = dvr_in%zort

    allocate(dvr_out%gdim(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%gdim = dvr_in%gdim

    allocate(dvr_out%ldvr(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%ldvr = dvr_in%ldvr

    allocate(dvr_out%zdvr(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%zdvr = dvr_in%zdvr

    allocate(dvr_out%zfft(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%zfft = dvr_in%zfft

    allocate(dvr_out%zexp(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%zexp = dvr_in%zexp

    allocate(dvr_out%zsph(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%zsph = dvr_in%zsph

    allocate(dvr_out%ipbaspar(dvr_out%mbaspar,dvr_out%ndof),stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%ipbaspar = dvr_in%ipbaspar

    allocate(dvr_out%rpbaspar(dvr_out%mbaspar,dvr_out%ndof),stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%rpbaspar = dvr_in%rpbaspar

    allocate(dvr_out%xend(dvr_out%ndof,1), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%xend = dvr_in%xend


    ! DVR
    allocate(dvr_out%ort(dvr_out%ortdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%ort = dvr_in%ort

    allocate(dvr_out%basis(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%basis = dvr_in%basis

    allocate(dvr_out%modelabel(dvr_out%ndof), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    do f=1,dvr_out%ndof
       dvr_out%modelabel(f) = dvr_in%modelabel(f)
    end do

    allocate(dvr_out%trafo(dvr_out%dvrdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%trafo = dvr_in%trafo

    allocate(dvr_out%dvrmat(dvr_out%dvrdim,dvr_out%dvrord), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%dvrmat = dvr_in%dvrmat

    allocate(dvr_out%fftp(dvr_out%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%fftp = dvr_in%fftp


    allocate(dvr_out%hin(dvr_out%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%hin = dvr_in%hin


    allocate(dvr_out%rueck(dvr_out%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%rueck = dvr_in%rueck


    allocate(dvr_out%fftfak(dvr_out%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%fftfak = dvr_in%fftfak


    ! length one if old version
    allocate(dvr_out%exphin(dvr_out%expdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%exphin = dvr_in%exphin


    ! length one if old version
    allocate(dvr_out%exprueck(dvr_out%expdim), stat=astat)
    if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
    dvr_out%exprueck = dvr_in%exprueck

    if (dvr_in%sphdim .gt. 1) then
       allocate(dvr_out%jsph(dvr_out%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
       dvr_out%jsph = dvr_in%jsph


       allocate(dvr_out%msph(dvr_out%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
       dvr_out%msph = dvr_in%msph


       allocate(dvr_out%kinsph(3*dvr_out%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr("routine: copy_dvr_t")
       dvr_out%kinsph = dvr_in%kinsph
    end if

    dvr_out%initialized = .true.

  end subroutine copy_dvr_t


  ! FREE MEMORY USED FOR THE DVR AND RESET SCALARS
  subroutine deallocate_dvr_t(dvr)
    implicit none
    type(dvr_t), intent(INOUT)           :: dvr

    dvr%filever = 0.d0

    ! scalars
    dvr%ndof = 1
    dvr%mbaspar = 9
    dvr%ortdim = 1
    dvr%dvrdim = 1
    dvr%fftdim = 1
    dvr%expdim = 1
    dvr%sphdim = 1
    dvr%dvrord = 1

    if ( allocated(dvr%modelabel) ) deallocate(dvr%modelabel)
    if ( allocated(dvr%gdim) )      deallocate(dvr%gdim)
    if ( allocated(dvr%basis) )     deallocate(dvr%basis)
    if ( allocated(dvr%ldvr) )      deallocate(dvr%ldvr)
    if ( allocated(dvr%ipbaspar) )  deallocate(dvr%ipbaspar)
    if ( allocated(dvr%rpbaspar) )  deallocate(dvr%rpbaspar)
    if ( allocated(dvr%xend) )      deallocate(dvr%xend)
    if ( allocated(dvr%zort) )      deallocate(dvr%zort)
    if ( allocated(dvr%zdvr) )      deallocate(dvr%zdvr)
    if ( allocated(dvr%zfft) )      deallocate(dvr%zfft)
    if ( allocated(dvr%zexp) )      deallocate(dvr%zexp)
    if ( allocated(dvr%zsph) )      deallocate(dvr%zsph)
    if ( allocated(dvr%ort) )       deallocate(dvr%ort)
    if ( allocated(dvr%trafo) )     deallocate(dvr%trafo)
    if ( allocated(dvr%dvrmat) )    deallocate(dvr%dvrmat)
    if ( allocated(dvr%fftp) )      deallocate(dvr%fftp)
    if ( allocated(dvr%hin) )       deallocate(dvr%hin)
    if ( allocated(dvr%rueck) )     deallocate(dvr%rueck)
    if ( allocated(dvr%fftfak) )    deallocate(dvr%fftfak)
    if ( allocated(dvr%exphin) )    deallocate(dvr%exphin)
    if ( allocated(dvr%exprueck) )  deallocate(dvr%exprueck)
    if ( allocated(dvr%jsph) )      deallocate(dvr%jsph)
    if ( allocated(dvr%msph) )      deallocate(dvr%msph)
    if ( allocated(dvr%kinsph) )    deallocate(dvr%kinsph)

    dvr%initialized = .false.

  end subroutine deallocate_dvr_t


  subroutine loaddvr(dvr,fname)
    implicit none
    type (dvr_t), intent(INOUT)     :: dvr
    character(len=*), intent(IN)    :: fname

    integer                         :: astat, idvr
    character(len=100)              :: buffer


    call deallocate_dvr_t(dvr)

    open(newunit=idvr,file=fname,status="old",form='unformatted', &
         action='read',err=998)

    !DVRDEF
    read(idvr,err=999) dvr%filever

    call importDVRdef(dvr,idvr)

    if ((dvr%filever .ge. 8.3016d0) .and. (dvr%filever .lt. 8.40d0)) then
       read(idvr,err=999)
       read(idvr,err=999)
       read(idvr,err=999)
    end if

    ! DVRINFO
    read(idvr,err=999) buffer

    if (dvr%filever .ge. 8.2002d0) then
       read(idvr,err=999) dvr%ortdim, dvr%dvrdim, dvr%fftdim, &
            dvr%expdim, dvr%sphdim
    else
       read(idvr,err=999) dvr%ortdim, dvr%dvrdim, dvr%fftdim, dvr%sphdim
    endif

    allocate(dvr%zort(dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%zort

    allocate(dvr%zdvr(dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%zdvr

    allocate(dvr%zfft(dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%zfft

    allocate(dvr%zexp(dvr%ndof), stat=astat)
    dvr%zexp = 0
    if (dvr%filever .ge. 8.2002d0) then
       read(idvr,err=999) dvr%zexp
    end if

    allocate(dvr%zsph(dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%zsph

    ! DVR
    allocate(dvr%ort(dvr%ortdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%ort

    allocate(dvr%trafo(dvr%dvrdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%trafo

    read(idvr,err=999) dvr%dvrord

    allocate(dvr%dvrmat(dvr%dvrdim,dvr%dvrord), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%dvrmat

    allocate(dvr%fftp(dvr%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%fftp

    allocate(dvr%hin(dvr%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%hin

    allocate(dvr%rueck(dvr%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%rueck

    allocate(dvr%fftfak(dvr%fftdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    read(idvr,err=999) dvr%fftfak

    ! length one if old version
    allocate(dvr%exphin(dvr%expdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    dvr%exphin=dcmplx(0.d0,0.d0)

    ! length one if old version
    allocate(dvr%exprueck(dvr%expdim), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    dvr%exprueck=dcmplx(0.d0,0.d0)

    if (dvr%filever .ge. 8.2002d0) then
       read(idvr,err=999) dvr%exphin
       read(idvr,err=999) dvr%exprueck
    else
       write(*,*) 'WARNING: old DVR-file (before 8.2.2).'
       write(*,*) 'May lead to problems...'
    end if

    if (dvr%sphdim .gt. 1) then
       allocate(dvr%jsph(dvr%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr(fname)
       read(idvr,err=999) dvr%jsph

       allocate(dvr%msph(dvr%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr(fname)
       read(idvr,err=999) dvr%msph

       allocate(dvr%kinsph(3*dvr%sphdim), stat=astat)
       if (astat .ne. 0) call allocerr(fname)
       read(idvr,err=999) dvr%kinsph
    end if

    read(idvr,err=999)
    close(idvr)

    dvr%initialized = .true.



    return

998 continue
    write(*,*) "Unable do open file "//trim(fname)
    stop 1
999 continue
    write(*,*) "error while reading file "//trim(fname)
    stop 1

  end subroutine loaddvr



  subroutine savedvr(dvr,fname,lovwr)
    type(dvr_t), intent(IN)        :: dvr
    character(len=*), intent(IN)   :: fname
    logical, intent(IN),optional   :: lovwr

    character(len=7)               :: fstat
    character(len=100)             :: buffer
    logical                        :: lovwr1
    integer                        :: idvr


    if (.not. dvr%initialized) then
       write(*,*) "ERROR: DVR is not initialized. Aborting"
       stop 1
    end if

    buffer = "DUMMY TEXT. TODO: Port routine 'fpstartup' to Fortran."


    lovwr1 = .false.
    if (present(lovwr)) lovwr1 = lovwr

    fstat='new'
    if (lovwr1) fstat='unknown'

    open(newunit=idvr,file=trim(fname), form='unformatted', status=fstat, err=996)
    write(idvr, err=997) dvr%filever

    ! DVRDEF
    call exportDVRdef(dvr, idvr)

    ! DVR
    write(idvr, err=997) buffer
    write(idvr, err=997) dvr%ortdim,dvr%dvrdim,dvr%fftdim, &
         dvr%expdim,dvr%sphdim

    write(idvr, err=997) dvr%zort
    write(idvr, err=997) dvr%zdvr
    write(idvr, err=997) dvr%zfft
    write(idvr, err=997) dvr%zexp
    write(idvr, err=997) dvr%zsph

    write(idvr, err=997) dvr%ort
    write(idvr, err=997) dvr%trafo
    write(idvr, err=997) dvr%dvrord
    write(idvr, err=997) dvr%dvrmat
    write(idvr, err=997) dvr%fftp
    write(idvr, err=997) dvr%hin
    write(idvr, err=997) dvr%rueck
    write(idvr, err=997) dvr%fftfak
    write(idvr, err=997) dvr%exphin
    write(idvr, err=997) dvr%exprueck
    !if (dvr%sphdim .gt. 1) then
       write(idvr, err=997) dvr%jsph
       write(idvr, err=997) dvr%msph
       write(idvr, err=997) dvr%kinsph
    !end if
    close(idvr)

    return

996 continue
    write(*,*) "Unable do open file "//trim(fname)
    stop 1
997 continue
    write(*,*) "Error while writing file "//trim(fname)
    stop 1

  end subroutine savedvr


  subroutine exportDVRdef(dvr,channel)
    type(dvr_t), intent(IN)     :: dvr
    integer, intent(IN)         :: channel
    integer                     :: log2int
    integer                     :: ildvr(dvr%ndof) ! automatic
    integer                     :: f


    write(channel, err=990) dvr%ndof
    write(channel, err=990) (dvr%modelabel(f),f=1,dvr%ndof)
    write(channel, err=990) dvr%gdim
    write(channel, err=990) dvr%mbaspar
    write(channel, err=990) dvr%basis


    do f=1,dvr%ndof
       ildvr(f) = log2int(dvr%ldvr(f))
    end do
    write(channel, err=990) ildvr

    write(channel, err=990) dvr%ipbaspar
    write(channel, err=990) dvr%rpbaspar
    write(channel, err=990) dvr%xend

    return

990 continue

    write(*,*) "ERROR while exporting DVR_def"
    stop 1

  end subroutine exportDVRdef


  subroutine importDVRdef(dvr,channel)
    type(dvr_t), intent(INOUT)  :: dvr
    integer, intent(IN)         :: channel
    logical                     :: int2log
    integer,allocatable         :: ildvr(:)
    integer                     :: i,f, astat
    character(len=512)          :: fname
    character(len=12)           :: what

    inquire(channel,name=fname)

    what = "ndof"
    read(channel,err=991) dvr%ndof

    if (allocated(dvr%modelabel)) deallocate(dvr%modelabel)
    allocate(dvr%modelabel(dvr%ndof),stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "modelabel"
    read(channel,err=991) (dvr%modelabel(f), f=1,dvr%ndof)


    if (allocated(dvr%gdim)) deallocate(dvr%gdim)
    allocate(dvr%gdim(dvr%ndof),stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "grid dimensions (gdim)"
    read(channel,err=991) dvr%gdim

    what = "bmaspar"
    read(channel,err=991) dvr%mbaspar

    if (allocated(dvr%basis)) deallocate(dvr%basis)
    allocate(dvr%basis(dvr%ndof),stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "basis"
    read(channel,err=991) dvr%basis

    allocate(ildvr(dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)

    if (allocated(dvr%ldvr)) deallocate(dvr%ldvr)
    allocate(dvr%ldvr(dvr%ndof),stat=astat)
    if (astat .ne. 0) call allocerr(fname)

    what = "ildfr"
    read(channel,err=991) ildvr
    do f=1,dvr%ndof
          dvr%ldvr(f) = int2log(ildvr(f))
       end do
    deallocate(ildvr)

    if (allocated(dvr%ipbaspar)) deallocate(dvr%ipbaspar)
    allocate(dvr%ipbaspar(dvr%mbaspar,dvr%ndof),stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "ipbaspar"
    read(channel,err=991) ((dvr%ipbaspar(i,f),i=1,dvr%mbaspar),f=1,dvr%ndof)

    if (allocated(dvr%rpbaspar)) deallocate(dvr%rpbaspar)
    allocate(dvr%rpbaspar(dvr%mbaspar,dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "rpbaspar"
    read(channel,err=991) ((dvr%rpbaspar(i,f),i=1,dvr%mbaspar),f=1,dvr%ndof)

    if (allocated(dvr%xend)) deallocate(dvr%xend)
    allocate(dvr%xend(2,dvr%ndof), stat=astat)
    if (astat .ne. 0) call allocerr(fname)
    what = "xend"
    read(channel,err=991) ((dvr%xend(i,f),i=1,2), f=1, dvr%ndof)


    return

991 continue

    write(*,*) "ERROR while importing DVRdef (importDVRdef)"
    write(*,*) "      routine: importDVRdef"
    write(*,*) "      file:    dvrtype.F90"
    write(*,*) "      what:    "//trim(what)
    stop 1

  end subroutine importDVRdef


  ! compare two dvrs: return chkerr:
  ! chkerr:
  ! chkerr(1) = -1 : dvr1 is empty
  ! chkerr(1) = -2 : dvr2 is empty
  ! chkerr(1) = -3 : both dvr are empty
  ! chkerr    =  0 : dvr1 and dvr2 are equivalent
  ! chkerr(1) = 1 : ndof differs
  ! chkerr(3) = 1 : at least 1 mode label differs
  ! chkerr(3) = 2 : at least 1 mode gdim differs
  ! chkerr(3) = 3 : at least 1 mode basis differs
  ! chkerr(3) = 4 : at least 1 mode ldvr differs
  ! chkerr(3) = 5 : mbaspar differs
  ! chkerr(3) = 6 : at least 1 mode ipbaspar differs
  ! chkerr(3) = 7 : at least 1 mode rpbaspar differs
  ! chkerr(3) = 8 : at least 1 mode xend differs
  subroutine compareDVR(dvr1,dvr2,chkerr, nerr, ilog)
    implicit none
    type(dvr_t), intent(IN)       :: dvr1, dvr2
    integer, intent(IN)           :: nerr
    integer, intent(IN), optional :: ilog
    integer, intent(OUT)          :: chkerr(nerr)
    integer                       :: f, ndof, mbaspar
    logical                       :: lopen

    if (nerr .lt. 10) then
       write(6,*) "Routine compareDVR"
       write(6,*) "ERROR: nerr must .ge. 10, got:", nerr
       stop 1
    end if

    lopen = .false.
    if ( present(ilog)) then
       inquire(unit=ilog, opened=lopen)
    end if


    chkerr = 0
    if (.not.dvr1%initialized) chkerr(1) = chkerr(1) -1
    if (.not.dvr2%initialized) chkerr(1) = chkerr(1) -2

    if (dvr1%ndof .ne. dvr2%ndof) then
       chkerr(2) = 1
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR ndof differs:"
          write(ilog, '(a,i4)') "ndof1: ", dvr1%ndof
          write(ilog, '(a,i4)') 'ndof2: ', dvr2%ndof
          write(ilog,*)
       end if
    end if

    ndof = min(dvr1%ndof, dvr2%ndof)

    if ( any(dvr1%modelabel(1:ndof) .ne. dvr2%modelabel(1:ndof))) then
       chkerr(3) = 1
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR: mode labels differ:"
          write(ilog, '(a4,2a8)') "f","lab1","lab2"
          do f=1,ndof
             write(ilog,'(i4,3x,2a8)') f, dvr1%modelabel(f),  dvr2%modelabel(f)
          end do
          write(ilog,*)
       end if
    end if ! modelabel

    if ( any(dvr1%gdim(1:ndof) .ne. dvr2%gdim(1:ndof))) then
       chkerr(3) = 2
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR: grid dimensions differ:"
          write(ilog, '(a4,3x,2a8)') "f","dim1","dim2"
          do f=1,ndof
             write(ilog,'(i4,3x,2i8)') f, dvr1%gdim(f),  dvr2%gdim(f)
          end do
          write(ilog,*)
       end if
    end if ! gdim

    if ( any(dvr1%basis(1:ndof) .ne. dvr2%basis(1:ndof))) then
       chkerr(3) = 3
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR: grid basis types differ:"
          write(ilog, '(a4,3x,2a8)') "f","typ1","typ2"
          do f=1,ndof
             write(ilog,'(i4,3x,2i8)') f, dvr1%basis(f),  dvr2%basis(f)
          end do
          write(ilog,*)
       end if
    end if ! basis


    if ( any(dvr1%ldvr(1:ndof) .neqv. dvr2%ldvr(1:ndof))) then
       chkerr(3) = 4
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR: grid basis ldvr flag differ:"
          write(ilog, '(a4,3x,2a8)') "f","ldvr1","ldvr2"
          do f=1,ndof
             write(ilog,'(i4,3x,2L8)') f, dvr1%ldvr(f),  dvr2%ldvr(f)
          end do
       end if
    end if ! ldvr

!!$    if (dvr1%mbaspar .ne. dvr2%mbaspar) then
!!$
!!$       if ( present(ilog) .and. lopen) then
!!$          write(ilog, '(a)') "DVR mbaspar differs (ignored):"
!!$          write(ilog, '(a,i4)') "mbaspar1: ", dvr1%mbaspar
!!$          write(ilog, '(a,i4)') 'mbaspar2: ', dvr2%mbaspar
!!$          write(ilog,*)
!!$       end if
!!$    end if

    mbaspar = min(dvr1%mbaspar, dvr2%mbaspar)

    if ( any(dvr1%ipbaspar(1:mbaspar, 1:ndof) .ne.&
             dvr2%ipbaspar(1:mbaspar, 1:ndof)) ) then
       chkerr(3) = 6
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR ipbaspar differs:"
          do f=1,ndof
             if ( any(dvr1%ipbaspar(1:mbaspar, f) .ne.&
                  dvr2%ipbaspar(1:mbaspar, f)) ) then
                write(ilog, '(a,i4,a,10i4)') "dvr1: f, par = ",f,",", &
                     dvr1%ipbaspar(1:mbaspar,f)
                write(ilog, '(a,i4,a,10i4)') "dvr2: f, par = ",f,",", &
                     dvr2%ipbaspar(1:mbaspar,f)
                write(ilog,*)
             end if
          end do
       end if
    end if

    if ( any( abs(dvr1%rpbaspar(1:mbaspar, 1:ndof) - &
                  dvr2%rpbaspar(1:mbaspar, 1:ndof)) .gt. 1.d-9) ) then
       chkerr(3) = 7
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR rpbaspar differs:"
          do f=1,ndof
             if ( any(abs(dvr1%rpbaspar(1:mbaspar,f) - &
                  dvr2%rpbaspar(1:mbaspar,f)) .gt. 1.d-9) ) then
                write(ilog, '(a,i4,a,10f16.8)') "dvr1: f, par = ",f,",", &
                     dvr1%rpbaspar(1:mbaspar,f)
                write(ilog, '(a,i4,a,10f16.8)') "dvr2: f, par = ",f,",", &
                     dvr2%rpbaspar(1:mbaspar,f)
                write(ilog,*)
             end if
          end do

       end if
    end if

    if ( any( abs(dvr1%xend(:, 1:ndof) - &
                  dvr2%xend(:, 1:ndof)) .gt. 1.d-9) ) then
       chkerr(3) = 7
       if ( present(ilog) .and. lopen) then
          write(ilog, '(a)') "DVR grid ranges differ:"
          do f=1,ndof
             if ( any( abs(dvr1%xend(:, f) - &
                  dvr2%xend(:,f)) .gt. 1.d-9) ) then
                write(ilog, '(a,i4,a,10f16.8)') "dvr1: f, xi xf = ",f,",", &
                     dvr1%xend(:,f)
                write(ilog, '(a,i4,a,10f16.8)') "dvr2: f, xi xf = ",f,",", &
                     dvr2%xend(:,f)
                write(ilog,*)
             end if
          end do
       end if
    end if


    call myflush(ilog)

  end subroutine compareDVR



  subroutine logdvrtype(dvr, iounit)
    implicit none

    type(dvr_t), intent(IN)       :: dvr
    integer, intent(IN)           :: iounit
    integer :: mbp

    ndof = min(dvr%ndof,maxdim)
    mbp = min(dvr%mbaspar, mbaspar)

    basis(1:ndof) = dvr%basis(1:ndof)
    modelabel(1:ndof) = dvr%modelabel(1:ndof)
    rpbaspar(1:mbp,1:ndof) = dvr%rpbaspar(1:mbp,1:ndof)
    ipbaspar(1:mbp,1:ndof) = dvr%ipbaspar(1:mbp,1:ndof)
    xend(:,1:ndof) =  dvr%xend(:,1:ndof)
    gdim(1:ndof) = dvr%gdim(1:ndof)

    call logdvr(iounit)

  end subroutine logdvrtype


  subroutine allocerr(message)
    implicit none
    character(len=*) :: message
    write(*,*) "Memory allocation failed while loading file"
    write(*,*) trim(message)
    stop 1
  end subroutine allocerr




  subroutine mpi_bcast_dvr_t(dvr)

    use logdat, only: message, routine, errormsg

    implicit none

    type(dvr_t), intent(INOUT) :: dvr

#ifdef MPI
    integer, parameter         :: mpi_int_nparams = 8
    integer                    :: mpi_int_arr(mpi_int_nparams)
    integer                    :: ierr

    if (mpirank .eq. 0) then

       if (.not. dvr%initialized) then
          routine = "mpi_bcast_dvr_t (MPI_RANK = 0)"
          message = "DVR is not iniialized."
          call errormsg
       end if

       call MPI_Bcast(dvr%filever,1,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       mpi_int_arr(1) = dvr%ndof
       mpi_int_arr(2) = dvr%mbaspar
       mpi_int_arr(3) = dvr%ortdim
       mpi_int_arr(4) = dvr%dvrdim
       mpi_int_arr(5) = dvr%fftdim
       mpi_int_arr(6) = dvr%expdim
       mpi_int_arr(7) = dvr%sphdim
       mpi_int_arr(8) = dvr%dvrord

       call MPI_Bcast(mpi_int_arr,mpi_int_nparams,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_BCast(dvr%modelabel,dvr%ndof*lblen, MPI_CHARACTER, 0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%gdim,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%basis,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%ldvr,dvr%ndof,MPI_LOGICAL,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%ipbaspar,dvr%ndof*dvr%mbaspar,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%rpbaspar,dvr%ndof*dvr%mbaspar,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%xend,2*dvr%ndof,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%zort,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%zdvr,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%zfft,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%zexp,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%zsph,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%ort,dvr%ortdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%trafo,dvr%dvrdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%dvrmat,dvr%dvrdim*dvr%dvrord,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%fftp,dvr%fftdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%hin,dvr%fftdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%rueck,dvr%fftdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%fftfak,dvr%fftdim,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%exphin,dvr%expdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(dvr%exprueck,dvr%expdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       if (dvr%sphdim .gt. 1 ) then

          call MPI_Bcast(dvr%jsph,dvr%sphdim,MPI_INTEGER,0, &
               MPI_COMM_WORLD,ierr)

          call MPI_Bcast(dvr%msph,dvr%sphdim,MPI_INTEGER,0, &
               MPI_COMM_WORLD,ierr)

          call MPI_Bcast(dvr%kinsph,3*dvr%sphdim,MPI_DOUBLE_PRECISION,0, &
               MPI_COMM_WORLD,ierr)
       end if

    else

       call deallocate_dvr_t(dvr)

       call MPI_Bcast(dvr%filever,1,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       call MPI_Bcast(mpi_int_arr,mpi_int_nparams,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       dvr%ndof = mpi_int_arr(1)
       dvr%mbaspar = mpi_int_arr(2)
       dvr%ortdim = mpi_int_arr(3)
       dvr%dvrdim = mpi_int_arr(4)
       dvr%fftdim = mpi_int_arr(5)
       dvr%expdim = mpi_int_arr(6)
       dvr%sphdim = mpi_int_arr(7)
       dvr%dvrord = mpi_int_arr(8)

       allocate(dvr%modelabel(dvr%ndof))
       call MPI_BCast(dvr%modelabel,dvr%ndof*lblen, MPI_CHARACTER, 0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%gdim(dvr%ndof))
       call MPI_Bcast(dvr%gdim,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%basis(dvr%ndof))
       call MPI_Bcast(dvr%basis,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%ldvr(dvr%ndof))
       call MPI_Bcast(dvr%ldvr,dvr%ndof,MPI_LOGICAL,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%ipbaspar(dvr%mbaspar,dvr%ndof))
       call MPI_Bcast(dvr%ipbaspar,dvr%ndof*dvr%mbaspar,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%rpbaspar(dvr%mbaspar,dvr%ndof))
       call MPI_Bcast(dvr%rpbaspar,dvr%ndof*dvr%mbaspar,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%xend(2,dvr%ndof))
       call MPI_Bcast(dvr%xend,2*dvr%ndof,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%zort(dvr%ndof))
       call MPI_Bcast(dvr%zort,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%zdvr(dvr%ndof))
       call MPI_Bcast(dvr%zdvr,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%zfft(dvr%ndof))
       call MPI_Bcast(dvr%zfft,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%zexp(dvr%ndof))
       call MPI_Bcast(dvr%zexp,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%zsph(dvr%ndof))
       call MPI_Bcast(dvr%zsph,dvr%ndof,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%ort(dvr%ortdim))
       call MPI_Bcast(dvr%ort,dvr%ortdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%trafo(dvr%dvrdim))
       call MPI_Bcast(dvr%trafo,dvr%dvrdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%dvrmat(dvr%dvrdim,dvr%dvrord))
       call MPI_Bcast(dvr%dvrmat,dvr%dvrdim*dvr%dvrord,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%fftp(dvr%fftdim))
       call MPI_Bcast(dvr%fftp,dvr%fftdim,MPI_DOUBLE_PRECISION,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%hin(dvr%fftdim))
       call MPI_Bcast(dvr%hin,dvr%fftdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%rueck(dvr%fftdim))
       call MPI_Bcast(dvr%rueck,dvr%fftdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%fftfak(dvr%fftdim))
       call MPI_Bcast(dvr%fftfak,dvr%fftdim,MPI_INTEGER,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%exphin(dvr%expdim))
       call MPI_Bcast(dvr%exphin,dvr%expdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)

       allocate(dvr%exprueck(dvr%expdim))
       call MPI_Bcast(dvr%exprueck,dvr%expdim,MPI_DOUBLE_COMPLEX,0, &
            MPI_COMM_WORLD,ierr)
       if (dvr%sphdim .gt. 1) then

          allocate(dvr%jsph(dvr%sphdim))
          call MPI_Bcast(dvr%jsph,dvr%sphdim,MPI_INTEGER,0, &
               MPI_COMM_WORLD,ierr)

          allocate(dvr%msph(dvr%sphdim))
          call MPI_Bcast(dvr%msph,dvr%sphdim,MPI_INTEGER,0, &
               MPI_COMM_WORLD,ierr)

          allocate(dvr%kinsph(3*dvr%sphdim))
          call MPI_Bcast(dvr%kinsph,3*dvr%sphdim,MPI_DOUBLE_PRECISION,0, &
               MPI_COMM_WORLD,ierr)
       end if

       dvr%initialized = .true.

    end if

#endif

    return

  end subroutine mpi_bcast_dvr_t



  ! Remove one or more DOF in DVR type in place
  ! frem is a list of intergers of DOF
  ! to be removed and nrem is the lenth of
  ! this list.
  ! Note: if one DOF of a 2D or 3D DVR is removed,
  ! also the other DOF must be removed
  ! i.e., all DOF must be in the frem list.
  ! Relies on frem list to contain NO dublicates
  ! and 2D/3D DVR DOFs in consecutive order in DVR type
  ! (can be in different order in frem list)
  subroutine remove_dof_dvr_t(dvr,frem,nrem)
    implicit none
    type(dvr_t), intent(INOUT)   :: dvr
    integer, intent(IN)          :: nrem
    integer, intent(IN)          :: frem(nrem)
    type(dvr_t)                  :: dvr1 ! temporary dvr

    integer                      :: f,f1,f2,zeig,zi,zf,zi1, zf1, dimf,n
    integer, allocatable         :: dvrdimf(:), fftdimf(:), expdimf(:), sphdimf(:)


    if (dvr%ndof .eq. 1) then
       write(*,*) "ERROR while removing DOF from DVR "
       write(*,*) "      routine: remove_dof"
       write(*,*) "      file:    dvrtype.F90"
       write(*,*) "      what:    DVR has only one DOF."
       stop 1
    end if
    if (nrem .ge. dvr%ndof) then
       write(*,*) "ERROR while removing DOF from DVR "
       write(*,*) "      routine: remove_dof"
       write(*,*) "      file:    dvrtype.F90"
       write(*,*) "      what:    No. of DOF to remove is >= No. DOF"
       stop 1
    end if
    if (any(frem .gt. dvr%ndof)) then
       write(*,*) "ERROR while removing DOF from DVR "
       write(*,*) "      routine: remove_dof"
       write(*,*) "      file:    dvrtype.F90"
       write(*,*) "      what:    DOF number in remove list >= No. DOF"
       stop 1
    end if
    if (any(frem .le. 0)) then
       write(*,*) "ERROR while removing DOF from DVR "
       write(*,*) "      routine: remove_dof"
       write(*,*) "      file:    dvrtype.F90"
       write(*,*) "      what:    DOF number in remove list <= 0"
       stop 1
    end if

    ! check if all parts of 2D DVRs are in remove list
    f = 1
    do while (f .le. dvr%ndof) !sphfbr,pleg,kleg
       if (dvr%basis(f) .eq. 6 .or. &
           dvr%basis(f) .eq. 7 .or. &
           dvr%basis(f) .eq. 9) then
          f1 = f+1
          if (any(frem .eq. f) .or. any(frem .eq. f1)) then
             if (.not. any(frem .eq. f) .and. .not. any(frem .eq. f1) ) then
                write(*,*) "ERROR while removing DOF from DVR "
                write(*,*) "      routine: remove_dof"
                write(*,*) "      file:    dvrtype.F90"
                write(*,*) "      what:    If one DOF of a 2D DVR is removed, also the other"
                write(*,*) "               DOF must be removed: Labels: "//trim(dvr%modelabel(f)) &
                     //", "//trim(dvr%modelabel(f+1))//"."
                stop 1
             else
                f = f+2
             end if
          end if
       else if (dvr%basis(f) .eq. 19) then !Wigner
          f1 = f+1
          f2 = f+2
          if (any(frem .eq. f) .or. any(frem .eq. f1) .or. any(frem .eq. f2)) then
             if (.not. any(frem .eq. f) .and. .not. any(frem .eq. f1) .and. .not.  any(frem .eq. f2) ) then
                write(*,*) "ERROR while removing DOF from DVR "
                write(*,*) "      routine: remove_dof"
                write(*,*) "      file:    dvrtype.F90"
                write(*,*) "      what:    If one DOF of a Wigner DVR is removed, also all other"
                write(*,*) "               DOF must also be removed. Labels: "//trim(dvr%modelabel(f)) &
                     //", "//trim(dvr%modelabel(f1))//", "//trim(dvr%modelabel(f2))//"."
                stop 1
             else
                f = f+3
             end if

          end if
       else
          f = f + 1
       end if
    end do




    call copy_dvr_t(dvr,dvr1)
    call deallocate_dvr_t(dvr)

    dvr%ndof    = dvr1%ndof - nrem
    dvr%mbaspar = dvr1%mbaspar
    dvr%filever = dvr1%filever

    allocate(dvr%modelabel(dvr%ndof))
    allocate(dvr%gdim(dvr%ndof))
    allocate(dvr%basis(dvr%ndof))
    allocate(dvr%ldvr(dvr%ndof))
    allocate(dvr%ipbaspar(dvr%mbaspar,dvr%ndof))
    allocate(dvr%rpbaspar(dvr%mbaspar,dvr%ndof))
    allocate(dvr%xend(2,dvr%ndof))

    allocate(dvr%zort(dvr%ndof))
    allocate(dvr%zdvr(dvr%ndof))
    allocate(dvr%zfft(dvr%ndof))
    allocate(dvr%zexp(dvr%ndof))
    allocate(dvr%zsph(dvr%ndof))

    ! things which we can just copy
    ! to the right place
    f = 0
    do f1=1,dvr1%ndof
       if (.not. any(f1 .eq. frem)) then
          f = f + 1
          dvr%modelabel(f)  = dvr1%modelabel(f1)
          dvr%gdim(f)       = dvr1%gdim(f1)
          dvr%basis(f)      = dvr1%basis(f1)
          dvr%ldvr(f)       = dvr1%ldvr(f1)
          dvr%ipbaspar(:,f) = dvr1%ipbaspar(:,f1)
          dvr%rpbaspar(:,f) = dvr1%rpbaspar(:,f1)
          dvr%xend(:,f)     = dvr1%xend(:,f1)
       end if
    end do

    ! pointers we need to re-calculate:

    allocate(dvrdimf(dvr%ndof))
    allocate(fftdimf(dvr%ndof))
    allocate(expdimf(dvr%ndof))
    allocate(sphdimf(dvr%ndof))


    ! ort
    dvr%ortdim = 0
    zeig = 1
    do f=1,dvr%ndof
       dvr%ortdim =  dvr%ortdim + dvr%gdim(f)
       dvr%zort(f) =  zeig
       zeig = zeig + dvr%gdim(f)
    end do

    ! dvr
    zeig=1
    dvrdimf(:) = 0
    do f=1,dvr%ndof
       if (dvr%ldvr(f)) then
          dvr%zdvr(f)=zeig
          if (dvr%basis(f).eq.7) then      ! KLeg
             !......The 2 was introduced (12.10.04,HDM) to account for dif3mat.
             dimf = 2*dvr%gdim(f)*dvr%gdim(f)*dvr%gdim(f+1)
             zeig=zeig+dimf
             dvrdimf(f) = dimf
          elseif (dvr%basis(f).eq.9) then  ! PLeg
             dimf = 2*dvr%gdim(f)*dvr%gdim(f)*dvr%ipbaspar(4,f)
             zeig=zeig + dimf  ! 2*
             dvrdimf(f) = dimf
          elseif (dvr%basis(f).eq.19) then  ! Wigner
             dimf = 2*dvr%gdim(f)*dvr%gdim(f)*dvr%ipbaspar(4,f)*dvr%ipbaspar(8,f)
             zeig = zeig + dimf
             dvrdimf(f) = dimf
          else
             dimf = dvr%gdim(f)*dvr%gdim(f)
             zeig = zeig + dimf
             dvrdimf(f) = dimf
          endif
       else
          dvr%zdvr(f) = 1
       endif
    enddo
    dvr%dvrdim=max(zeig-1,1)

    ! fft
    zeig=1
    fftdimf(:) = 0
    do f=1,dvr%ndof
       if (dvr%basis(f).eq.4) then
          dvr%zfft(f)=zeig
          zeig=zeig+dvr%gdim(f)
          fftdimf(f) = dvr%gdim(f)
       else
          dvr%zfft(f) = 1
       endif
    enddo
    dvr%fftdim=max(zeig-1,1)

    ! exp
    zeig=1
    expdimf(:) = 0
    do f=1,dvr%ndof
       if (dvr%basis(f).eq.5) then
          dvr%zexp(f)=zeig
          if (f.gt.1 .and. dvr%basis(max(1,f-1)).eq.9) then
             dimf = dvr%gdim(f)*dvr%ipbaspar(4,f-1)
             zeig = zeig + dimf
             expdimf(f) = dimf
          elseif (f.gt.1 .and. dvr%basis(max(1,f-1)).eq.19) then
             dimf = dvr%gdim(f)*dvr%ipbaspar(4,f-1)
             zeig = zeig + dimf
             expdimf(f) = dimf
          elseif (f.gt.2 .and. dvr%basis(max(1,f-2)).eq.19) then
             dimf =  dvr%gdim(f)*dvr%ipbaspar(8,f-2)
             zeig = zeig + dimf
             expdimf(f) = dimf
          else
             dimf = dvr%gdim(f)**2
             zeig = zeig + dimf
             expdimf(f) = dimf
          endif
       else
          dvr%zexp(f)=1
       endif
    enddo
    dvr%expdim=max(zeig-1,1)

    ! fbr
    zeig=1
    sphdimf(:) = 0
    do f=1,dvr%ndof
       if (dvr%basis(f).eq.6) then
          dvr%zsph(f)=zeig
          zeig=zeig+dvr%gdim(f)
          sphdimf(f) = dvr%gdim(f)
       else
          dvr%zsph(f) = 1
       endif
    enddo
    dvr%sphdim=max(zeig-1,1)


    ! now copy the remaining stuff
    ! (trafo, diff-mats etc. )

    allocate(dvr%ort(dvr%ortdim))
    allocate(dvr%trafo(dvr%dvrdim))
    allocate(dvr%dvrmat(dvr%dvrdim,dvr%dvrord))
    allocate(dvr%fftp(dvr%fftdim))
    allocate(dvr%hin(dvr%fftdim))
    allocate(dvr%rueck(dvr%fftdim))
    allocate(dvr%fftfak(dvr%fftdim))
    allocate(dvr%exphin(dvr%expdim))
    allocate(dvr%exprueck(dvr%expdim))
    allocate(dvr%jsph(dvr%sphdim))
    allocate(dvr%msph(dvr%sphdim))
    allocate(dvr%kinsph(3*dvr%sphdim))

    f = 0
    do f1=1,dvr1%ndof
       if ( .not. any(frem == f1)) then
          f = f + 1

          ! ort
          zi = dvr%zort(f)
          zf = zi + dvr%gdim(f)-1
          zi1 = dvr1%zort(f1)
          zf1 = zi1 + dvr1%gdim(f1)-1
          dvr%ort(zi:zf) = dvr1%ort(zi1:zf1)

          !fft
          if (dvr%basis(f).eq.4) then
             zi = dvr%zfft(f)
             zf = zi + fftdimf(f)-1
             zi1 = dvr1%zfft(f1)
             zf1 = zi1 + fftdimf(f)-1
             dvr%fftp(zi:zf) = dvr1%fftp(zi1:zf1)
             dvr%hin(zi:zf) = dvr1%hin(zi1:zf1)
             dvr%rueck(zi:zf) = dvr1%rueck(zi1:zf1)
             dvr%fftfak(zi:zf) = dvr1%fftfak(zi1:zf1)
          end if

          !exp
          if (dvr%basis(f).eq.5) then
             zi = dvr%zexp(f)
             zf = zi + expdimf(f)-1
             zi1 = dvr1%zexp(f1)
             zf1 = zi1 + expdimf(f)-1
             dvr%exphin(zi:zf) = dvr1%exphin(zi1:zf1)
             dvr%exprueck(zi:zf) = dvr1%exprueck(zi1:zf1)
          end if

          !fbr
          if (dvr%basis(f).eq.6) then
             zi = dvr%zsph(f)
             zf = zi + sphdimf(f)-1
             zi1 = dvr1%zsph(f1)
             zf1 = zi1 + sphdimf(f)-1
             dvr%jsph(zi:zf) = dvr1%jsph(zi1:zf1)
             dvr%msph(zi:zf) = dvr1%msph(zi1:zf1)
             dvr%kinsph(zi:zf) = dvr1%kinsph(zi1:zf1)
          end if

          zi = dvr%zdvr(f)
          zf = zi + dvrdimf(f) - 1
          zi1 = dvr1%zdvr(f1)
          zf1 = zi1 + dvrdimf(f) - 1

          !trafo, diffmat
          if (dvr%ldvr(f)) then
             dvr%trafo(zi:zf) = dvr1%trafo(zi1:zf1)
             do n=1,dvr%dvrord
                dvr%dvrmat(zi:zf,n) = dvr1%dvrmat(zi1:zf1,n)
             end do
          end if

       end if
    end do

    dvr%initialized = .true.

  end subroutine remove_dof_dvr_t


end module dvrtype_m
