! -*- f90 -*-

! module eximporttools
!
! import and export of tensor formats
! (of potentials) to MCTDH readable format
!
! 10/18 MS



module eximporttools
  use dvrtype_m
  use tensortype
  use natpottype
  use cpdtype_m
  use hcpdtype_m
  use global
  use pglobal
  use dvrdatmod, only: csmass,jtot,jbf
  use griddatmod
  use operdef
  use channels
  use mpidata
  use logdat
  use fitinmod, only: vecin
  use wrfitmod, only: vecout
  use versions, only: progver

  implicit none

  private

  public :: exportMCTDH
  public :: importMCTDH
  public :: exportgrddef

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! Public routines here
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  ! export to an MCTDH-readable file
  subroutine exportMCTDH(tensor, dvr, fname, loverwrite, loperfit)
    implicit none

    class(tensor_t),  intent(IN),target    :: tensor
    type(dvr_t),      intent(IN)           :: dvr
    character(len=*), intent(IN)           :: fname
    logical,          intent(IN)           :: loverwrite
    logical,          intent(IN), optional :: loperfit

    logical                                :: lnatpot,loperfit1
    character(len=7)                       :: stat
    character(len=80)                      :: buffer

    integer                                :: ierr,f

    type(natpot_t), pointer                :: natpot_ptr
    type(natpot_t), target                 :: natpot_contr
    type(cpd_t), pointer                   :: cpd_ptr
    integer                                :: onedipot(tensor%ndof)

    if (mpirank.ne.0) return ! do not write on other ranks that 0


    if (.not. dvr%initialized) then
       write(*,*) "ERROR in exportMCTDH: DVR is not initialized."
       write(*,*) "      Aborting. Nothing was written."
       return
    end if

    if (.not. tensor%initialized) then
       write(*,*) "ERROR in exportMCTDH: tensor is not initialized."
       write(*,*) "      Aborting. Nothing was written."
       return
    end if

    loperfit1 = .false.
    if (present(loperfit)) loperfit1 = loperfit


    ! if tensor is a the natpot
    ! and has no contacted mode
    ! a temporary natpot with an artificial
    ! contracted mode is made and
    ! a pointer is set to this copy.
    ! otherwise the pointer is set to the
    ! original natpot

    select type(tensor)
    type is (natpot_t)
       lnatpot = .true.
       if (tensor%modc .lt. 1) then
          call contractmode(tensor, natpot_contr)
          natpot_ptr => natpot_contr
       else
          natpot_ptr => tensor
       end if
    type is (cpd_t)
       lnatpot = .false.
       cpd_ptr => tensor
    type is (hcpd_t)
       call exporthcpdMCTDH(tensor, dvr, fname, loverwrite)
       return
    end select


    stat = "new"
    if (loverwrite) stat="unknown"
    open(inpot, file=trim(fname),status=stat,form="unformatted",err=990)

    write(inpot,iostat=ierr) progver

    if (lnatpot) then
       ! obsolete - for consistency
       write(inpot) 0,0,dvr%ipbaspar(1,1)
    end if

    call exportDVRdef(dvr,inpot)

    call fpstartup(buffer)
    write(inpot,err=991) buffer

    onedipot = 0

    do f=1,tensor%ndof
       if (tensor%loned(f)) onedipot(f) = 1
    end do


    select type(tensor)
    type is (natpot_t)
       call exportnpinput(natpot_ptr, inpot)

       write(inpot) natpot_ptr%modc,onedipot
       write(inpot) natpot_ptr%lpconm

    type is (cpd_t)
       call exportcpdinput(cpd_ptr, inpot)
       write(inpot) cpd_ptr%rank, onedipot
    end select


    call exportgrddef(tensor, inpot,loperfit1)


    select type(tensor)
    type is (natpot_t)

       ! at this stage modc .ne. 0, but this may change
       if (natpot_ptr%modc .ne. 0) then
          natpot_ptr%potdim(natpot_ptr%modc) =  natpot_ptr%modcdim
       end if
       write(inpot) natpot_ptr%potdim
       if (natpot_ptr%modc .ne. 0) then
          natpot_ptr%potdim(natpot_ptr%modc) = 0
       end if

       call exportnpfit(natpot_ptr, inpot)

    type is (cpd_t)

       call exportcpdfit(cpd_ptr, inpot)

    end select


    call exportvpotpars(tensor, inpot)


    close(inpot)

    ! not dangerous - internal 'allocated' test:
    call deallocate_natpot_t(natpot_contr)

    return

990 continue
    routine = "exportMCTDH"
    message = "ERROR while opening file '"//trim(fname)//"' for writing."
    call errormsg
991 continue
    routine = "exportMCTDH"
    message = "ERROR while writing file '"//trim(fname)//"'."
    call errormsg

  end subroutine exportMCTDH



  subroutine importMCTDH(tensor,dvr,fname)
    implicit none

    class (tensor_t), intent(INOUT) :: tensor
    type(dvr_t), intent(IN)         :: dvr
    character(len=*), intent(IN)    :: fname

    type(natpot_t), pointer         :: natpot
    type(cpd_t), pointer            :: cpd
    type(hcpd_t), pointer           :: hcpd

    if (mpirank .ne. 0) return ! do not read in any other ranks than 0


    select type(tensor)
    type is (natpot_t)
       natpot => tensor
       call importnpotMCTDH(natpot,dvr,fname)
    type is (cpd_t)
       cpd => tensor
       call importcpdMCTDH(cpd,dvr,fname)
    type is (hcpd_t)
       hcpd => tensor
       call importhcpdMCTDH(hcpd,dvr,fname)
    end select

    ! we might want to get rid of this in the future
    ! and actually give the natpot a dvr object
    ! as a class-member...
    tensor%ort = dvr%ort

  end subroutine importMCTDH



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! Private routines here
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  ! import a natpot file from old mctdh format
  ! needs matching dvr as input
  subroutine importnpotMCTDH(natpot,dvr,fname)

    implicit none

    type(natpot_t), intent(INOUT) :: natpot
    type(dvr_t), intent(IN)       :: dvr
    character(len=*), intent(IN)  :: fname

    logical                       :: lexist, lopen
    logical                       :: int2log
    integer                       :: lcount, imult,ieb,imulpack
    integer                       :: i,f,ff,m,n, ierr, ncount
    type(dvr_t)                   :: dvr2

    character(len=80)             :: buffer
    character(len=c5)             :: opts

    external int2log

    call deallocate_natpot_t(natpot)

    inquire(unit=ilog,opened=lopen)
    if (lopen) then
       write(ilog,'(/a)') "Reading natpot file:"
       write(ilog,'(a/)') " "//trim(fname)
       call myflush(ilog)
    end if

    inquire(file=fname,exist=lexist)
    if (.not. lexist ) then
       write(*,*) "File does not exist: "//trim(fname)
       stop 1
    end if

    open(inpot,file=fname,status='OLD',form= 'unformatted', &
         action="read",iostat=ierr)
    if (ierr .ne.0) then
       write(*,*) "Could not open file: "//trim(fname)
       stop 1
    end if

    read(inpot) progver
    if (progver .lt. 8.4d0) then
       write(*,'(a)') "Sorry, natpot file versions older then then" &
            // " 8.4 are not supported."
       write(*,'(a,f8.6)') "Found version ", progver
       stop 1
    end if

    read(inpot) ! unused: 0,0,ipbaspar(1,1)


    call copy_dvr_t(dvr, dvr2)
    call importDVRdef(dvr2,inpot)

    ! check if the DVRs are consistent and if so
    ! copy the natpot
    call compareDVR(dvr,dvr2,chkerr, nerr, ilog)
    if (any(chkerr .ne. 0)) then
       routine = "importMCTDH"
       message = "DVR of natpot differs from system DVR."
       call errormsg
    end if


    ! skip some blabla
    read(inpot) buffer
    read(inpot) lcount
    do i=1,lcount
       read(inpot) buffer
    end do

    allocate(onedipot(dvr2%ndof))
    onedipot = 0
    read(inpot) natpot%modc, onedipot

    read(inpot) natpot%lpconm

    ! start rdgrddef
    read(inpot) dentype

    read(inpot) natpot%ndof,natpot%nmode
    read(inpot) !unused: nstate,npacket,npackts,feb,meb,fpb,mpb

    ! allocate all we can:
    allocate(natpot%zort(natpot%ndof))
    allocate(natpot%gdim(natpot%ndof))
    allocate(natpot%vgdim(natpot%ndof))
    allocate(natpot%ngdim(natpot%ndof))
    allocate(natpot%vgpdim(natpot%ndof))
    allocate(natpot%vgmdim(natpot%ndof))
    allocate(natpot%potdim(natpot%nmode))
    allocate(natpot%potvdim(natpot%nmode))
    allocate(natpot%potndim(natpot%nmode))
    allocate(natpot%subdim(natpot%nmode))
    allocate(natpot%nspfdof(natpot%nmode))
    allocate(natpot%dofspf(natpot%ndof))
    allocate(natpot%modelabel(natpot%ndof))
    allocate(natpot%potdofmap(2,natpot%ndof))

    allocate(natpot%loned(natpot%ndof))


    read(inpot) (natpot%nspfdof(m),m=1,natpot%nmode)
    natpot%maxspfdof = maxval(natpot%nspfdof)

    allocate(natpot%spfdof(natpot%maxspfdof,natpot%nmode))
    natpot%spfdof = 0
    do m=1,natpot%nmode
       read(inpot) (natpot%spfdof(n,m),n=1,natpot%nspfdof(m))
    enddo

    read(inpot) (natpot%dofspf(f),f=1,natpot%ndof)

    read(inpot) imult,ieb,imulpack
    lmult = int2log(imult)
    leb = int2log(ieb)
    lmulpack = int2log(imulpack)
    ! end rdgrddef input - start checks and ptr construction

    ! end rdgrddef checks

    read(inpot) natpot%potdim
    if (natpot%modc .ne. 0) natpot%potdim(natpot%modc) = 0
    natpot%maxspf = maxval(natpot%potdim)


    ! setup pointers now
    natpot%ortdim = dvr%ortdim
    allocate(natpot%ort(natpot%ortdim))
    allocate(natpot%zevec(natpot%nmode,natpot%maxspf))
    ! we need to construct potdofmap
    ! for use of zeig_natpot_t
    do m=1, natpot%nmode
       ff=natpot%nspfdof(m)
       do n=1,ff
          f = natpot%spfdof(n,m)
          natpot%potdofmap(1,f) = m
          natpot%potdofmap(2,f) = n
       end do
    end do
    natpot%modelabel = dvr%modelabel
    natpot%gdim =  dvr%gdim
    do f=1,natpot%ndof
       natpot%loned(f) = onedipot(f) .gt. 0
    end do
    deallocate(onedipot)

    call zeig_natpot_t(natpot)

    ! alloc data
    allocate(natpot%coeff(natpot%cdim))
    allocate(natpot%evec(natpot%evecdim))

    if (any(natpot%loned)) then
       allocate(natpot%oned(natpot%ortdim))
       natpot%oned = 0.d0
       do f=1,natpot%ndof
          if (natpot%loned(f)) then
             n=dvr%zort(f)-1
             read(inpot) (natpot%oned(n+i),i=1,natpot%gdim(f))
          endif
       enddo

    end if

    ! pointer are done. Now import data
    do m=1,natpot%nmode

       if (m.eq.natpot%modc) then
          call vecin(natpot%coeff,natpot%dimbef,natpot%dimmodc, &
               natpot%dimaft,inpot)

       else ! any mode except modc
          call vecin(natpot%evec(natpot%zevec(m,1)),1,natpot%subdim(m), &
               natpot%potdim(m),inpot)
       endif
    enddo

    ! done, the rest is info



    read(inpot) buffer
    read(inpot) ncount

    if (lopen) write(ilog,'(a)') trim(buffer)

    do i=1,ncount
       read(inpot) buffer
       if (lopen) write(ilog,'(a)') trim(buffer)
    end do


    read(inpot) hoplab(1)
    read(inpot) opts
    if (lopen) then
       write(ilog,'(a)') trim(opts)
       call myflush(ilog)
    end if

    ! skip the rest
    close(inpot)

    ! cleanup
    call deallocate_dvr_t(dvr2)

    natpot%initialized = .true.
  end subroutine importnpotMCTDH


  subroutine importcpdMCTDH(cpd,dvr,fname)

    implicit none
    type(cpd_t), intent(INOUT)   :: cpd
    type(dvr_t), intent(IN)      :: dvr
    character(len=*), intent(IN) :: fname

    logical                       :: lexist, lopen
    logical                       :: int2log
    integer                       :: lcount, imult,ieb,imulpack
    integer                       :: i,f,ff,m,n,r, ierr, ncount
    type(dvr_t)                   :: dvr2

    character(len=80)             :: buffer
    character(len=c5)             :: opts

    external int2log

    call deallocate_cpd_t(cpd)

    inquire(unit=ilog,opened=lopen)
    if (lopen) then
       write(ilog,'(/a)') "Reading cpd file:"
       write(ilog,'(a/)') " "//trim(fname)
       call myflush(ilog)
    end if

    inquire(file=fname,exist=lexist)
    if (.not. lexist ) then
       message = "File does not exist: "//trim(fname)
       routine = "importcpdMCTDH"
       call errormsg
    end if

    open(inpot,file=fname,status='OLD',form= 'unformatted', &
         action="read",iostat=ierr)
    if (ierr .ne.0) then
       message = "Could not open file: "//trim(fname)
       routine = "importcpdMCTDH"
       call errormsg
    end if

    read(inpot) progver

    call copy_dvr_t(dvr, dvr2)
    call importDVRdef(dvr2,inpot)

    ! check if the DVRs are consistent and if so
    ! copy the cpd
    call compareDVR(dvr,dvr2,chkerr, nerr, ilog)
    if (any(chkerr .ne. 0)) then
       routine = "importMCTDH"
       message = "DVR of cpd differs from system DVR."
       call errormsg
    end if

    allocate(onedipot(dvr%ndof))

    ! skip some blabla
    read(inpot) buffer
    read(inpot) lcount
    do i=1,lcount
       read(inpot) buffer
    end do

    read(inpot) cpd%rank, onedipot

    ! start rdgrddef
    read(inpot) dentype

    read(inpot) cpd%ndof,cpd%nmode
    read(inpot) !unused: nstate,npacket,npackts,feb,meb,fpb,mpb

    ! allocate all we can:
    allocate(cpd%zort(cpd%ndof))
    allocate(cpd%gdim(cpd%ndof))
    allocate(cpd%subdim(cpd%nmode))
    allocate(cpd%nspfdof(cpd%nmode))
    allocate(cpd%dofspf(cpd%ndof))
    allocate(cpd%modelabel(cpd%ndof))
    allocate(cpd%potdofmap(2,cpd%ndof))
    allocate(cpd%loned(cpd%ndof))

    read(inpot) (cpd%nspfdof(m),m=1,cpd%nmode)
    cpd%maxspfdof = maxval(cpd%nspfdof)

    allocate(cpd%spfdof(cpd%maxspfdof,cpd%nmode))
    cpd%spfdof = 0
    do m=1,cpd%nmode
       read(inpot) (cpd%spfdof(n,m),n=1,cpd%nspfdof(m))
    enddo

    read(inpot) (cpd%dofspf(f),f=1,cpd%ndof)

    read(inpot) imult,ieb,imulpack
    lmult = int2log(imult)
    leb = int2log(ieb)
    lmulpack = int2log(imulpack)
    ! end rdgrddef input - start checks and ptr construction

    ! end rdgrddef checks

    ! setup pointers now
    cpd%ortdim = dvr%ortdim
    allocate(cpd%ort(cpd%ortdim))

    ! we need to construct potdofmap
    ! for use of zeig_cpd_t
    do m=1, cpd%nmode
       ff=cpd%nspfdof(m)
       do n=1,ff
          f = cpd%spfdof(n,m)
          cpd%potdofmap(1,f) = m
          cpd%potdofmap(2,f) = n
       end do
    end do
    cpd%modelabel = dvr%modelabel
    cpd%gdim =  dvr%gdim
    call zeig_cpd_t(cpd)

    do f=1,cpd%ndof
       cpd%loned(f) = onedipot(f) .gt. 0
    end do

    ! alloc data
    allocate(cpd%coeff(cpd%rank))
    allocate(cpd%spp(cpd%nmode))

    do m=1, cpd%nmode
       allocate(cpd%spp(m)%htm(cpd%subdim(m), cpd%rank) )
    end do

    if (any(cpd%loned)) then
       allocate(cpd%oned(cpd%ortdim))
       cpd%oned = 0.d0

       do f=1,cpd%ndof
          if (cpd%loned(f)) then
             n=dvr%zort(f)-1
             read(inpot) (cpd%oned(n+i),i=1,cpd%gdim(f))
          endif
       enddo

    end if

    ! pointer are done. Now import data
    do r=1,cpd%rank
       read (inpot) cpd%coeff(r)
       do m=1, cpd%nmode

          call vecin(cpd%spp(m)%htm(:,r),1,cpd%subdim(m), &
               1,inpot)
       end do

    enddo

    ! done, the rest is info

    read(inpot) buffer
    read(inpot) ncount

    if (lopen) write(ilog,'(a)') trim(buffer)

    do i=1,ncount
       read(inpot) buffer
       if (lopen) write(ilog,'(a)') trim(buffer)
    end do

    read(inpot) hoplab(1)
    read(inpot) opts
    if (lopen) then
       write(ilog,'(a)') trim(opts)
       call myflush(ilog)
    end if

    ! skip the rest
    close(inpot)

    ! cleanup
    call deallocate_dvr_t(dvr2)

    cpd%initialized = .true.

  end subroutine importcpdMCTDH



  subroutine importhcpdMCTDH(hcpd,dvr,fname)

    implicit none
    type(hcpd_t), intent(INOUT)  :: hcpd
    type(dvr_t), intent(IN)      :: dvr
    character(len=*), intent(IN) :: fname

    logical                       :: lexist, lopen
    logical                       :: int2log
    integer                       :: lcount, imult,ieb,imulpack
    integer                       :: i,f,ff,m,n,r, ierr
    type(dvr_t)                   :: dvr2

    character(len=80)             :: buffer

    external int2log

    call deallocate_hcpd_t(hcpd)

    inquire(unit=ilog,opened=lopen)
    if (lopen) then
       write(ilog,'(/a)') "Reading hcpd file:"
       write(ilog,'(a/)') " "//trim(fname)
       call myflush(ilog)
    end if

    inquire(file=fname,exist=lexist)
    if (.not. lexist ) then
       message = "File does not exist: "//trim(fname)
       routine = "importhcpdMCTDH"
       call errormsg
    end if

    open(inpot,file=fname,status='OLD',form= 'unformatted', &
         action="read",iostat=ierr)
    if (ierr .ne.0) then
       message = "Could not open file: "//trim(fname)
       routine = "importhcpdMCTDH"
       call errormsg
    end if

    read(inpot) progver

    call copy_dvr_t(dvr, dvr2)
    call importDVRdef(dvr2,inpot)

    ! check if the DVRs are consistent and if so
    ! copy the cpd
    call compareDVR(dvr,dvr2,chkerr, nerr, ilog)
    if (any(chkerr .ne. 0)) then
       routine = "importMCTDH"
       message = "DVR of hcpd differs from system DVR."
       call errormsg
    end if

    allocate(onedipot(dvr%ndof))

    ! skip some blabla
    read(inpot) buffer
    read(inpot) lcount
    do i=1,lcount
       read(inpot) buffer
    end do

    read(inpot) hcpd%rank, onedipot

    ! start rdgrddef
    read(inpot) dentype

    read(inpot) hcpd%ndof,hcpd%nmode
    read(inpot) !unused: nstate,npacket,npackts,feb,meb,fpb,mpb

    ! allocate all we can:
    allocate(hcpd%zort(hcpd%ndof))
    allocate(hcpd%gdim(hcpd%ndof))
    allocate(hcpd%subdim(hcpd%nmode))
    allocate(hcpd%subdim2(hcpd%nmode))
    allocate(hcpd%nspfdof(hcpd%nmode))
    allocate(hcpd%dofspf(hcpd%ndof))
    allocate(hcpd%modelabel(hcpd%ndof))
    allocate(hcpd%potdofmap(2,hcpd%ndof))
    allocate(hcpd%loned(hcpd%ndof))

    read(inpot) (hcpd%nspfdof(m),m=1,hcpd%nmode)
    hcpd%maxspfdof = maxval(hcpd%nspfdof)

    allocate(hcpd%spfdof(hcpd%maxspfdof,hcpd%nmode))
    hcpd%spfdof = 0
    do m=1,hcpd%nmode
       read(inpot) (hcpd%spfdof(n,m),n=1,hcpd%nspfdof(m))
    enddo

    read(inpot) (hcpd%dofspf(f),f=1,hcpd%ndof)

    read(inpot) imult,ieb,imulpack
    lmult = int2log(imult)
    leb = int2log(ieb)
    lmulpack = int2log(imulpack)
    ! end rdgrddef input - start checks and ptr construction

    ! end rdgrddef checks

    ! setup pointers now
    hcpd%ortdim = dvr%ortdim
    allocate(hcpd%ort(hcpd%ortdim))

    ! we need to construct potdofmap
    ! for use of zeig_cpd_t
    do m=1, hcpd%nmode
       ff=hcpd%nspfdof(m)
       do n=1,ff
          f = hcpd%spfdof(n,m)
          hcpd%potdofmap(1,f) = m
          hcpd%potdofmap(2,f) = n
       end do
    end do
    hcpd%modelabel = dvr%modelabel
    hcpd%gdim =  dvr%gdim


    do m=1,hcpd%nmode
       hcpd%subdim(m) = 1
       do n=1,hcpd%nspfdof(m)
          f=hcpd%spfdof(n,m)
          hcpd%subdim(m) = hcpd%subdim(m)*hcpd%gdim(f)
       enddo
       hcpd%subdim2(m) = hcpd%subdim(m)**2
    end do

    hcpd%vpotdim = product(hcpd%subdim)
    hcpd%rvpotdim = product(dble(hcpd%subdim))
    hcpd%rvpotdim2 = product(dble(hcpd%subdim2))


    hcpd%zort(1) = 1
    do f=2,hcpd%ndof
       hcpd%zort(f) = hcpd%zort(f-1) + hcpd%gdim(f-1)
    end do


    do f=1,hcpd%ndof
       hcpd%loned(f) = onedipot(f) .gt. 0
    end do

    ! alloc data
    allocate(hcpd%coeff(hcpd%rank))
    allocate(hcpd%sph(hcpd%nmode))

    do m=1, hcpd%nmode
       allocate(hcpd%sph(m)%htm(hcpd%subdim(m), hcpd%subdim(m), hcpd%rank) )
    end do

    if (any(hcpd%loned)) then
       allocate(hcpd%oned(hcpd%ortdim))
       hcpd%oned = 0.d0

       do f=1,hcpd%ndof
          if (hcpd%loned(f)) then
             n=dvr%zort(f)-1
             read(inpot) (hcpd%oned(n+i),i=1,hcpd%gdim(f))
          endif
       enddo

    end if

    ! pointer are done. Now import data
    do r=1,hcpd%rank
       read (inpot) hcpd%coeff(r)
       do m=1,hcpd%nmode

          call vecin(hcpd%sph(m)%htm(:,:,r),1,hcpd%subdim2(m), &
               1,inpot)
       end do

    enddo

    ! done, the rest is info

!!$    read(inpot) buffer
!!$    read(inpot) ncount
!!$
!!$    if (lopen) write(ilog,'(a)') trim(buffer)
!!$
!!$    do i=1,ncount
!!$       read(inpot) buffer
!!$       if (lopen) write(ilog,'(a)') trim(buffer)
!!$    end do
!!$
!!$    read(inpot) hoplab(1)
!!$    read(inpot) opts
!!$    if (lopen) then
!!$       write(ilog,'(a)') trim(opts)
!!$       call myflush(ilog)
!!$    end if

    ! skip the rest
    close(inpot)

    ! cleanup
    call deallocate_dvr_t(dvr2)

    hcpd%initialized = .true.

  end subroutine importhcpdMCTDH






  subroutine exporthcpdMCTDH(hcpd,dvr,fname,loverwrite)

    implicit none
    type(hcpd_t), intent(IN)     :: hcpd
    type(dvr_t), intent(IN)      :: dvr
    character(len=*), intent(IN) :: fname
    logical, intent(IN)          :: loverwrite

    character(len=7)             :: stat
    character(len=80)            :: buffer

    integer                      :: onedipot(hcpd%ndof)
    integer                      :: m,r,i,ierr, lcount


    if (.not. dvr%initialized) then
       write(*,*) "ERROR in rwhcpd: DVR is not initialized."
       write(*,*) "      Aborting. Nothing was written."
       return
    end if

    if (.not. hcpd%initialized) then
       write(*,*) "ERROR in rwhcpd: tensor is not initialized."
       write(*,*) "      Aborting. Nothing was written."
       return
    end if

    stat = "new"
    if (loverwrite) stat="unknown"
    open(ioper, file=trim(fname),status=stat,form="unformatted",err=990)

    write(ioper,iostat=ierr) progver
    call exportDVRdef(dvr,ioper)

    call fpstartup(buffer)
    write(ioper,err=991) buffer


    ! input blabla
    lcount = 2
    write(ioper,iostat=ierr) lcount
    write(buffer,'(a,i6)') ' H-Candecomp rank:',hcpd%rank
    write(ioper,iostat=ierr) buffer
    write(buffer,'(a,11(1x,a))') ' Modelabels:', &
         (trim(hcpd%modelabel(i)),i=1,min(hcpd%ndof,11))
    write(ioper,iostat=ierr) buffer

    ! no 1D pots
    onedipot = 0
    write(ioper) hcpd%rank, onedipot

    call exportgrddef(hcpd, ioper,.false.)


    do r=1,hcpd%rank
       write (ioper) hcpd%coeff(r)
       do m=1, hcpd%nmode
          call vecout(hcpd%sph(m)%htm(:,:,r), 1, hcpd%subdim2(m),1, ioper)
       end do
    end do



    close(ioper)

    return

990 continue
    routine = "rwhcpd"
    message = "ERROR while opening file '"//trim(fname)//"' for writing."
    call errormsg
991 continue
    routine = "rwhcpd"
    message = "ERROR while writing file '"//trim(fname)//"'."
    call errormsg
  end subroutine exporthcpdMCTDH




  subroutine exportnpinput(natpot, channel)

    implicit none
    type (natpot_t), intent(IN)    :: natpot
    integer, intent(IN)            :: channel

    character(len=200)             :: line
    integer                        :: i,f, ierr, lcount

    onedipot = 0
    do f=1,natpot%ndof
       if (natpot%loned(f)) onedipot(f) = 1
    end do

    lcount = 4
    write(channel,iostat=ierr) lcount
    write(line,'(a,i3,a,20i6)') ' contracted mode:',natpot%modc, &
         ',  nat.pots:',(natpot%potdim(i),i=1,min(natpot%nmode,20))
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,e12.4,a,18i3)') ' V-cut =', pesmax*27.2114, &
         ' eV,  oned:',(onedipot(i),i=1,min(natpot%ndof,18))
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,e12.4,a,e12.4,a)') ' C-Weights: V-max =', &
         iwvmax*27.2114,' eV,  V-min =',iwvmin*27.2114,' eV'
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,11(1x,a))') ' Modelabels:', &
         (trim(natpot%modelabel(i)),i=1,min(natpot%ndof,11))
    write(channel,iostat=ierr) line(1:80)

  end subroutine exportnpinput


  subroutine exportcpdinput(cpd, channel)

    implicit none
    type (cpd_t), intent(IN)       :: cpd
    integer, intent(IN)            :: channel

    character(len=200)             :: line
    integer                        :: i, ierr, lcount

    lcount = 4
    write(channel,iostat=ierr) lcount
    write(line,'(a,i6)') ' candecomp rank:',cpd%rank
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,ES16.8,a)') ' V-cut =', pesmax*27.2114, ' eV'
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,e12.4,a,e12.4,a)') ' C-Weights: V-max =', &
         iwvmax*27.2114,' eV,  V-min =',iwvmin*27.2114,' eV'
    write(channel,iostat=ierr) line(1:80)
    write(line,'(a,11(1x,a))') ' Modelabels:', &
         (trim(cpd%modelabel(i)),i=1,min(cpd%ndof,11))
    write(channel,iostat=ierr) line(1:80)

  end subroutine exportcpdinput




  subroutine exportgrddef(tensor, channel,loperfit)

    implicit none
    class (tensor_t), intent(IN)   :: tensor
    integer, intent(IN)            :: channel
    logical, intent(IN)            :: loperfit
    integer                        :: m,n,n1,f,log2int

    integer :: ndof1
    integer,allocatable :: nspfdof1(:)
    integer,allocatable :: spfdof1(:,:)
    integer,allocatable :: dofspf1(:)

    allocate(nspfdof1(tensor%nmode))
    allocate(spfdof1(tensor%maxspfdof,tensor%nmode))
    allocate(dofspf1(tensor%ndof))


    if (loperfit) then


       ndof1 = tensor%ndof/2
       nspfdof1(:) = tensor%nspfdof(:)/2
       do m=1,tensor%nmode
          n1 = 1
          do n=1,nspfdof1(m)
             spfdof1(n,m) = tensor%spfdof(n1,m)/2 + 1
             n1 = n1 + 2
          end do
       end do
    else
       ndof1 = tensor%ndof
       nspfdof1(:) = tensor%nspfdof(:)
       spfdof1(:,:) = tensor%spfdof(:,:)
       dofspf1(:) = tensor%dofspf(:)
    end if

    write(channel) dentype
!    write(channel) tensor%ndof,tensor%nmode
!    write(channel) nstate,npacket,npackts,feb,meb,fpb,mpb
!    write(channel) (tensor%nspfdof(m),m=1,tensor%nmode)
!    do m=1,tensor%nmode
!       write(channel) (tensor%spfdof(n,m),n=1,tensor%nspfdof(m))
!    enddo
!    write(channel) (tensor%dofspf(f),f=1,tensor%ndof)


    write(channel) ndof1,tensor%nmode
    write(channel) nstate,npacket,npackts,feb,meb,fpb,mpb
    write(channel) (nspfdof1(m),m=1,tensor%nmode)
    do m=1,tensor%nmode
       write(channel) (spfdof1(n,m),n=1,nspfdof1(m))
    enddo
    write(channel) (dofspf1(f),f=1,ndof1)

    write(channel) log2int(lmult),log2int(leb),log2int(lmulpack)

  end subroutine exportgrddef



  subroutine exportnpfit(natpot, channel)
    implicit none
    type (natpot_t), intent(IN)    :: natpot
    integer, intent(IN)            :: channel
    integer                        :: m,f, zeig,g

    if (any(natpot%loned)) then

       do f=1,natpot%ndof
          if (natpot%loned(f)) then
             zeig=natpot%zort(f)-1
             write(channel) (natpot%oned(zeig+g),g=1,natpot%gdim(f))
          endif
       enddo

    end if


    do m=1,natpot%nmode
       if (m.eq.natpot%modc) then
          call vecout(natpot%coeff,natpot%dimbef,natpot%dimmodc, &
               natpot%dimaft,channel)
       else ! any mode except modc
          call vecout(natpot%evec(natpot%zevec(m,1)),1,natpot%subdim(m), &
               natpot%potdim(m),channel)
       endif
    enddo

  end subroutine exportnpfit


  subroutine exportcpdfit(cpd, channel)
    implicit none
    type (cpd_t), intent(IN)       :: cpd
    integer, intent(IN)            :: channel
    integer                        :: m,r,f,g,zeig


    if (any(cpd%loned)) then

       do f=1,cpd%ndof
          if (cpd%loned(f)) then
             zeig=cpd%zort(f)-1
             write(channel) (cpd%oned(zeig+g),g=1,cpd%gdim(f))
          endif
       enddo

    end if


    do r=1,cpd%rank
       write (channel) cpd%coeff(r)
       do m=1, cpd%nmode
          call vecout(cpd%spp(m)%htm(:,r), 1, cpd%subdim(m),1, channel)
       end do

    end do


  end subroutine exportcpdfit





  subroutine exportvpotpars(tensor, channel)

    implicit none
    class (tensor_t), intent(IN)   :: tensor
    integer, intent(IN)            :: channel

    integer                        :: log2int
    character(len=80)              :: buffer
    character(len=c5)              :: opts

    opts=' '
    if (npesopt.eq.1) then
       if (nopts(1).gt.0) then
          opts=pesopts(1)(1:zpesopts(nopts(npesopt),1))
       endif
    endif

    call fpstartup(buffer)
    write(channel) buffer
    select type(tensor)
    type is (natpot_t)
       call exportnpinput(tensor, channel)
    type is (cpd_t)
       call exportcpdinput(tensor, channel)
    end select


    write(channel) hoplab(1)
    write(channel) opts
    write(channel) log2int(lsp)
    write(channel) log2int(.false.), 0 !lftp=.false., pdofft=0
    write(channel) pesmin, pesmax
    write(channel) log2int(lcspot),jtot,jbf,csmass

  end subroutine exportvpotpars



end module eximporttools
