
! -*- f90 -*-

module mcals_m

  use cpdtype_m
  use cpdtype_s, only: changerank, mpi_bcast_cpd_t, copy_cpd
  use mccuts_m
  use cpdsamplings_m
#ifdef _OPENMP
  use omp_lib
#endif
#ifdef MPI
  use mpi
#endif

  use global
  use pglobal
  use mpidata
  use timing

  use symmops_m
  use mccpdparams, only : lwrfuncs, epsilon_invert, luseblas
  use channels
  use logdat

  use dvrtype_m
  use mccpdparams
  use runfit_m
  use pesselect_m
  use pftools
  use tensortype, only: iscompatible
  use eximporttools
  use mcalsjfunc_m

  implicit none

  double precision, allocatable :: rtmp(:), ctmp(:)  ! for fast rms


  integer, allocatable          :: mgroup(:,:) ! grouped m's
  integer, allocatable          :: nmgroup(:)  ! no. of m's in a group

 ! logical, save                 :: lfixepsilon = .false.
 ! logical, save                 :: lfixepsilonnow = .false.
  integer                       :: curriter

contains

  subroutine init_mcals(nsample,maxrank)
    implicit none

    integer, intent(IN) :: nsample, maxrank

    call free_mcals()

    allocate(rtmp(nsample))
    allocate(ctmp(maxrank))
    curriter = 0

  end subroutine init_mcals


  subroutine free_mcals()
    implicit none

    if (allocated(rtmp))     deallocate(rtmp)
    if (allocated(ctmp))     deallocate(ctmp)
    if (allocated(mgroup))   deallocate(mgroup)
    if (allocated(nmgroup))  deallocate(nmgroup)

  end subroutine free_mcals


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


subroutine mc_als1m_matrices(cpd,mi,ni,B,BT,ovlp,luseblas)
  implicit none

  type(cpd_t), intent(IN)       :: cpd
  integer, intent(IN)           :: mi, ni
  double precision, intent(OUT) :: B(cpd%rank,cpd%subdim(mi),ni)
  double precision, intent(OUT) :: BT(cpd%subdim(mi),cpd%rank,ni)
  double precision, intent(OUT) :: ovlp(cpd%rank,cpd%rank,ni)
  logical, intent(IN)           :: luseblas

  double precision, allocatable :: omega(:,:)
  integer                       :: nsample,n,m,m1,m2,ms
  integer                       :: r,r1,z,z1,ng,MM,KK,NN,LDA
  integer                       :: mz(cpd%nmode-1)
  integer                       :: allocstat
  double precision              :: ddot

  integer(long), save                        :: tid=0 ! timer ID


  external ddot

  ! debug:
  if (ni .ne. nmgroup(mi)) then
     print*, "Error: ni .ne. nmgroup(mi)"
     stop 1
  end if

  nsample = size(basissamplings, 1)


  allocate(omega(nsample,cpd%rank), stat=allocstat)
  if (allocstat.ne.0) then
     routine = "mc_als1m_matrices"
     write(message,'(a,i4)') 'Unable to allocate ALS Omega, mode= ', mi
     call cncntrblnk(message)
     call errormsg
  end if

  ! find out which modes we do the sampling over
  ! ( == all .ne. m)

  ! build omega for all m's in a group
  do ng=1,ni

     if (mpirank.eq.0) then
        call get_timer(tid, 'ptrhs')
        call start_timer(tid)
     endif

     m = mgroup(mi,ng)

     ! modes in omega:
     ms = 1
     do m1=1,cpd%nmode
        if (m.ne.m1) then
           mz(ms)= m1
           ms = ms + 1
        end if
     end do


     if (cpd%nmode .eq. 2) then
        m1 = mz(1)
        !$omp parallel do simd collapse(2) schedule(static)
        do r=1,cpd%rank
           do n=1,nsample
              omega(n,r) =  basissamplings(n,m1,r)
           end do
        end do
        !$omp end parallel do simd
     else

        m1 = mz(1)
        m2 = mz(2)

        !$omp parallel do simd collapse(2) schedule(static)
        do r=1,cpd%rank
           do n=1,nsample
              omega(n,r) =  basissamplings(n,m1,r)*basissamplings(n,m2,r)
           end do
        end do
        !$omp end parallel do simd

        do ms=3,cpd%nmode-1

           m1 = mz(ms)

           !$omp parallel do simd collapse(2) schedule(static)
           do r=1,cpd%rank
              do n=1,nsample
                 omega(n,r) = omega(n,r) *  basissamplings(n,m1,r)
              end do
           end do
           !$omp end parallel do simd

        end do

     end if


#ifdef DEBUG
     if ( any(omega(:,:) .ne. omega(:,:)) ) then
        print*, "Found NaN in omega (mcals.F90)"
        stop 1
     end if
#endif


     ! build overlap matrix (this is the most expensive part)

     if (luseblas) then

        R = cpd%rank
        N = nsample
        call DSYRK("L","T",R,N,1.d0,omega,N,0.d0,ovlp(:,:,ng),R)

     else

        !$omp parallel do private(r1) schedule(dynamic,1)
        do r=1,cpd%rank
           do r1=r,cpd%rank
              ovlp(r1,r,ng) =  ddot(nsample, omega(:,r1),1,omega(:,r),1)
           end do
        end do
        !$omp end parallel do

     end if


     do r=1,cpd%rank
        do r1=r+1,cpd%rank
           ovlp(r,r1,ng) = ovlp(r1,r,ng)
        end do
     end do


     if (luseblas) then

        MM = cpd%subdim(m)
        NN = cpd%rank
        KK = nsample
        LDA = mccuts(m)%mgdim
        call dgemm('N','N',MM,NN,KK,1.d0,mccuts(m)%cuts,&
             LDA,omega,KK, 0.d0,BT(:,:,ng),MM)

     else

        z = 1
        z1 = cpd%subdim(m)

        ! m is a mode of the mode-group
        ! and specified above

        !$omp parallel do private(n) schedule(static)
        do r=1,cpd%rank
           BT(:,r,ng) = mccuts(m)%cuts(z:z1,1)*omega(1,r)
           do n=2,nsample
              BT(:,r,ng) = BT(:,r,ng) + mccuts(m)%cuts(z:z1,n)*omega(n,r)
           end do
        end do
        !$omp end parallel do

     end if

     ! (transpose B)
     B(:,:,ng) = transpose(BT(:,:,ng))


!     do r=1,cpd%rank
!        do n=1,cpd%subdim(m)
!           B(r,n,ng) = BT(n,r,ng)
!        end do
!     end do

     if (mpirank.eq.0) call stop_timer(tid)
  end do ! m in mode groups (ng)


  deallocate(omega)

end subroutine mc_als1m_matrices


! solve system for one effective mode
subroutine mcalsiter1m(cpd, cpd_origgrid, m, ierr, lwrefile, invertmethod)

  implicit none

  type(cpd_t), intent(INOUT)    :: cpd
  type(cpd_t), intent(IN)       :: cpd_origgrid
  integer, intent(IN)           :: m, invertmethod
  integer, intent(INOUT)        :: ierr
  logical, intent(IN)           :: lwrefile

  character(len=c5)             :: fname
  double precision              :: norm2
  integer                       :: m1,r,ng,n1
  integer                       :: allocstat

  double precision, allocatable :: B(:,:,:), BT(:,:,:)
  double precision, allocatable :: ovlp(:,:,:), coeff(:,:)
  double precision, allocatable :: ovlpbak(:,:,:), BTbak(:,:,:)

#ifdef MPI
  double precision, allocatable :: mpibuf(:,:)
  integer                       :: mpierr
  logical                       :: lwarn1
#endif

  double precision              :: dnrm2
  external dnrm2

  integer(long), save                        :: tid=0 ! timer ID

  ng = nmgroup(m)

  allocate(BT(cpd%subdim(m),cpd%rank,ng), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(2a,i4)') 'Unable to allocate ALS transposed ', &
          'right-hand-side (B^T), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if

  allocate(B(cpd%rank,cpd%subdim(m),ng), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(a,i4)') &
          'Unable to allocate ALS right-hand-side (B), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if

  allocate(ovlp(cpd%rank,cpd%rank,ng), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(a,i4)') &
          'Unable to allocate ALS left-hand-side (ovlp), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if

  ! we need to build the Bs,and ovlps
  ! for all m's in the mode-group

  call mc_als1m_matrices(cpd, m, ng, B, BT, ovlp,luseblas)

#ifdef MPI

  ! gather partial results in rank 0
  ! we send chunks of about 100MB (or less)
  allocate(mpibuf(cpd%rank,cpd%rank), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(a,i4)') 'Unable to allocate MPI recv buffer (ovlp), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if


  do n1=1,ng

     call MPI_Reduce(ovlp(:,:,n1),mpibuf,cpd%rank**2,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)

     if (mpirank.eq.0) then
        ovlp(:,:,n1) = mpibuf
     end if

  end do

  deallocate(mpibuf)

  allocate(mpibuf(cpd%rank,cpd%subdim(m)), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(a,i4)') 'Unable to allocate MPI recv buffer (B), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if


  do n1=1,ng

     call MPI_Reduce(B(:,:,n1),mpibuf,cpd%rank*cpd%subdim(m),&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)

     if (mpirank.eq.0) then
        B(:,:,n1) = mpibuf
     end if
  end do
  deallocate(mpibuf)


  allocate(mpibuf(cpd%subdim(m),cpd%rank), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'mcalsiter1m'
     write(message,'(a,i4)') 'Unable to allocate MPI recv buffer (B), mode= ', m
     call cncntrblnk(message)
     call errormsg
  end if


  do n1=1,ng

     call MPI_Reduce(BT(:,:,n1),mpibuf,cpd%rank*cpd%subdim(m),&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)

     if (mpirank.eq.0) then
        BT(:,:,n1) = mpibuf
     end if
  end do
  deallocate(mpibuf)


#endif


  ! if we want to calculate the functional for this mode on the fly,
  ! we need to backup the original matrices
  ! as they will be destroyed in the LAPACK routines
  if (lwrfuncs .and. mpirank .eq.0) then

     allocate(BTbak(cpd%subdim(m),cpd%rank,ng), stat=allocstat)
     if (allocstat.ne.0) then
        routine = 'mcalsiter1m'
        write(message,'(a,i4)') 'Unable to allocate backup of right-hand-side (B^T), mode= ', m
        call cncntrblnk(message)
        call errormsg
     end if

     BTbak(:,:,:) = BT(:,:,:)

     allocate(ovlpbak(cpd%rank,cpd%rank,ng), stat=allocstat)
     if (allocstat.ne.0) then
        routine = 'mcalsiter1m'
        write(message,'(a,i4)') 'Unable to allocate backup of left-hand-side (ovlp), mode= ', m
        call cncntrblnk(message)
        call errormsg
     end if

     ovlpbak(:,:,:) = ovlp(:,:,:)

  end if


  ! we have the matrices, symmetrize if required
  if (lsymmetries .and. mpirank.eq.0) then
     call symmals(ovlp, B, m, ng, cpd%rank,cpd%subdim(m))
  end if


  do n1=1,ng

     m1 = mgroup(m,n1)

     ! regularization
     if (invertmethod .eq. mc_inv_spd .and. mpirank.eq.0) then

        do r=1,cpd%rank
           ovlp(r,r,n1) =  ovlp(r,r,n1) * (1.d0 + epsilon_invert)
        end do

     else if (invertmethod .eq.  mc_inv_eigen  .and. mpirank.eq.0 .and. lwrefile ) then
        write(fname,'(a,i4)') "eigenval_als_mode_", m1
        call rmblnk(fname)
     else
        fname = ""
     end if

     if (mpirank.eq.0) then
        call get_timer(tid, 'ptsolve')
        call start_timer(tid)
     endif

  ! if MPI and SCALAPACK the solvers will
  ! take care of redistributing data from rank 0
  ! so the system to be solved (ovlp and B)
  ! need only be correct in rank 0
  ! Note: results will also be only in rank 0
  ! when sovers are done
  ! - if not SCALAPACK the system will be solved
  ! in rank 0 only


     if (invertmethod.eq.mc_inv_spd) then
        call mcsolve_spd(ovlp(:,:,n1),B(:,:,n1),cpd%rank,cpd%subdim(m),ierr)
     else if (invertmethod.eq.mc_inv_eigen) then

        call mcsolve_eigen(ovlp(:,:,n1),B(:,:,n1),epsilon_invert,.false., .false., &
             cpd%rank,cpd%subdim(m),fname,ierr)
     else
        routine = "mcalsiter1m"
        write(message,'(a,i4)') "Illegal solver: ", invertmethod
        call errormsg
     end if
     if (mpirank.eq.0) call stop_timer(tid)


     if (ierr .ne. 0 .and. mpirank .eq. 0) then
        routine = "mcalsiter1m"
        write(message,'(a,i8,a,i8)') "mcalsiter1m: solver error code:", ierr, " mode: ", m
        call warnmsg
     end if


  end do !n1


  ! results are in rank 0 only at this point

  if (mpirank.eq.0) then

     if (lsymmetries) then
        call symafter(B, m, ng, cpd%rank,cpd%subdim(m))
     end if

     do n1=1,ng
        BT(:,:,n1) = transpose(B(:,:,n1))
     end do



     allocate(coeff(cpd%rank,ng))
     !$omp parallel do private(r,norm2)
     do n1=1,ng
        do r=1, cpd%rank
           norm2 = dnrm2(cpd%subdim(m),BT(:,r,n1),1)
           coeff(r,n1) = norm2 !/cpd%subdim(m))
        end do
     end do
     !$omp end parallel do

     ! geometric mean of the coeff to symmetrize them.
     ! this is not perfect but better then nothing.
     ! the correct coefficients will be obtained
     ! when the next non-swapping mode is optimized.
     if (ng.gt.1) then
        do r=1, cpd%rank
           coeff(r,1) = product(coeff(r,:))
           coeff(r,1) = coeff(r,1)**(1.d0/dble(ng))
        end do
        do n1=2,ng
           coeff(:,n1) = coeff(:,1)
        end do
     end if


     do n1=1,ng

        m1 = mgroup(m,n1)

        !$omp parallel do private(norm2)
        do r=1, cpd%rank

           ! normalize to subdim for nuumerical stability
           if (abs(coeff(r,n1)) .gt. 1.d-12) then
              cpd%coeff(r) = coeff(r,n1)

              cpd%spp(m1)%htm(:,r) =  BT(:,r,n1)/cpd%coeff(r)

           else

              ! Zero terms sometimes happen directly after
              ! growing the rank.
              ! Should be fixed by going through the
              ! next ALS sweep, set random here


              cpd%coeff(r) = 0.d0

              call random_number(cpd%spp(m1)%htm(:,r))
              cpd%spp(m1)%htm(:,r) = cpd%spp(m1)%htm(:,r) - 0.5d0

              norm2 = dnrm2(cpd%subdim(m1),cpd%spp(m1)%htm(:,r),1)
              cpd%spp(m1)%htm(:,r) = cpd%spp(m1)%htm(:,r) / norm2

           end if
        end do
        !$omp end parallel do


        if (lwrfuncs) then
           call calcfuncs1m(m1,ovlpbak(:,:,n1),BTbak(:,:,n1),cpd, epsilon_invert, cpd_origgrid%subdim(m1))
        end if

     end do

  end if


#ifdef MPI


  call MPI_Bcast(cpd%coeff,cpd%rank,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
  do n1=1,ng
     m1 = mgroup(m,n1)
     call MPI_Bcast(cpd%spp(m)%htm,cpd%subdim(m)*cpd%rank,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
  end do


  !collect any warnings
  call MPI_AllReduce(lwarn, lwarn1, 1 , &
       MPI_LOGICAL,MPI_LOR, MPI_COMM_WORLD,mpierr)
  lwarn = lwarn1
#endif


  if (allocated(B)) deallocate(B)
  if (allocated(BT)) deallocate(BT)
  if (allocated(ovlp)) deallocate(ovlp)
  if (allocated(coeff)) deallocate(coeff)
  if (allocated(BTbak)) deallocate(BTbak)
  if (allocated(ovlpbak)) deallocate(ovlpbak)


end subroutine mcalsiter1m





subroutine symmals(ovlp, B, mi, ni, rank, subdim)

  implicit none
  integer, intent(IN)             :: rank,subdim,mi,ni
  double precision, intent(INOUT) :: B(rank,subdim,ni)
  double precision, intent(INOUT) :: ovlp(rank,rank,ni)
  double precision, allocatable   :: btmp(:,:,:),otmp(:,:,:)

  integer                         :: nsym,r,r1,n1,n2
  integer                         :: m,m1,j,z,z1,k,ig,ng,nt
  integer                         :: rpersym, rest
  integer                         :: allocstat

  double precision                :: ddot
  external ddot


  ! MPI rank 0 only
  if (mpirank.ne.0) return

  rpersym = rank/nsymmetries

  allocate(btmp(rank,subdim,ni), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'symmals (mcals.F90)'
     write(message,'(a,i4)') 'Unable to allocate temporary B, mode= ', mi
     call cncntrblnk(message)
     call errormsg
  end if

  ! identity
  btmp = B

  do nsym = 2,nsymmetries

     do ng=1,ni
        m = mgroup(mi,ng)       ! current mode
        m1 = symmmap(m,nsym)    ! target m of mode group

        nt = -11
        do j=1,nmgroup(mi)      ! target n of m1
           if (mgroup(mi,j) .eq. m1) nt = j
        end do

        do r=1,rank

           n1 = (r-1)/rpersym + 1  ! local sym in cpd
           n2 = sym2map(nsym,n1)   ! target sym

           rest = modulo(r,rpersym)
           if (rest .ne. 0) then   ! target r
              r1 = (n2-1)*rpersym + rest
           else
              r1 = (n2)*rpersym
           end if

           do ig=1,subdim
              z = zsymidxmap(m1,nsym)+ig-1
              k = symidxmap(z)
              btmp(r,k,nt)   = btmp(r,k,nt)   + B(r1,ig,ng)
           end do

        end do

     end do

  end do

  B = btmp/nsymmetries
  deallocate(btmp)

  allocate(otmp(rank,rank,ni), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'symmals (mcals.F90)'
     write(message,'(a,i4)') 'Unable to allocate temporary ovlp, mode= ', mi
     call cncntrblnk(message)
     call errormsg
  end if


  otmp = ovlp ! identity

  do nsym = 2,nsymmetries

     do ng=1, ni
        m = mgroup(mi,ng)       ! current mode
        m1 = symmmap(m,nsym)    ! target m of mode group
        nt = -11
        do j=1,ni
           if (mgroup(mi,j) .eq. m1) nt = j
        end do

        do r=1,rank
           n1 = (r-1)/rpersym + 1  ! local sym in chunk
           n2 = sym2map(nsym,n1)   ! target sym
           rest = modulo(r,rpersym)
           if (rest .ne. 0) then   ! target r
              r1 = (n2-1)*rpersym + rest
           else
              r1 = (n2)*rpersym
           end if

           do z=1,rank
              n1 = (z-1)/rpersym + 1  ! local sym in chunk
              n2 = sym2map(nsym,n1)   ! target sym
              rest = modulo(z,rpersym)
              if (rest .ne. 0) then   ! target r
                 z1 = (n2-1)*rpersym + rest
              else
                 z1 = (n2)*rpersym
              end if

              otmp(r1,z1,ng) = otmp(r1,z1,ng) + ovlp(r,z,nt)

           end do
        end do

     end do

  end do

  ovlp = otmp/nsymmetries
  deallocate(otmp)

end subroutine symmals



subroutine symafter(spp, mi, ni, rank, subdim)

  implicit none
  integer, intent(IN)             :: rank,subdim,mi,ni
  double precision, intent(INOUT) :: spp(rank,subdim,ni) ! this is B
  double precision, allocatable   :: spptmp(:,:,:)

  integer                         :: nsym,r,r1,n1,n2
  integer                         :: m,m1,j,z,k,ig,ng,nt
  integer                         :: rpersym, rest

  integer                         :: allocstat

! MPI rank 0 only
  if (mpirank.ne.0) return

  rpersym = rank/nsymmetries

  allocate(spptmp(rank,subdim,ni), stat=allocstat)
  if (allocstat.ne.0) then
     routine = 'symafter (mcals.F90)'
     write(message,'(a,i4)') 'Unable to allocate temporary B, mode= ', mi
     call cncntrblnk(message)
     call errormsg
  end if


  ! identity
  spptmp = spp

  do nsym = 2,nsymmetries

     do ng=1,ni
        m = mgroup(mi,ng)       ! current mode
        m1 = symmmap(m,nsym)    ! target m of mode group

        nt = -11
        do j=1,nmgroup(mi)      ! target n of m1
           if (mgroup(mi,j) .eq. m1) nt = j
        end do

        do r=1,rank

           n1 = (r-1)/rpersym + 1  ! local sym in cpd
           n2 = sym2map(nsym,n1)   ! target sym

           rest = modulo(r,rpersym)
           if (rest .ne. 0) then   ! target r
              r1 = (n2-1)*rpersym + rest
           else
              r1 = (n2)*rpersym
           end if

           do ig=1,subdim
              z = zsymidxmap(m1,nsym)+ig-1
              k = symidxmap(z)
              spptmp(r,k,nt)   = spptmp(r,k,nt)   + spp(r1,ig,ng)
           end do

        end do
     end do
  end do

  spp = spptmp/nsymmetries

  deallocate(spptmp)


end subroutine symafter



subroutine init_mgroups(nmode)
  implicit none
  integer, intent(IN) :: nmode
  integer             :: m,m1,n

  ! effective modes: number of
  ! modes to do: modes which interchange
  ! upon symmetry operations are treated as
  ! one effective mode as the need to be
  ! updated by ALS simultaneously

  if (.not. lsymmetries) then
     allocate(mgroup(nmode, 1))
     allocate(nmgroup(nmode))
     do m=1,nmode
        mgroup(m,1) = m
        nmgroup(m) = 1
     end do
  else
     allocate(nmgroup(nmode))
     allocate(mgroup(nmode,maxval(npermutedmodes)))
     do m=1,nmode
        nmgroup(m) = npermutedmodes(m)
        do n=1,npermutedmodes(m)
           m1 = permutedmodes(m,n)
           mgroup(m,n)= m1
        end do
     end do
  end if

end subroutine init_mgroups



subroutine mc_optimize_cpd(cpd, cpd_origgrid, dvr, invertmethod, lextgrid)
  implicit none

#ifdef MPI
  integer :: mpierr
#endif


  integer, intent(IN)              :: invertmethod
  type(cpd_t), intent(INOUT)       :: cpd
  type(cpd_t), intent(INOUT)       :: cpd_origgrid
  type(dvr_t), intent(IN)          :: dvr
  logical, intent(IN)              :: lextgrid


  integer, parameter               :: nline = 107
  character(len=nline)             :: line
  character(len=c5)                :: iterationsfile
  character(len=c5)                :: fname
  double precision                 :: rms, cf2, jfunc
  integer                          :: startiter, i, itercount
  logical                          :: lstopals, lwarn1, lwrefile
  integer                          :: m,m1,n,ierr, iiter, nsample
  logical                          :: lmupdate(cpd%nmode)
  logical                          :: lmdone(cpd%nmode)

  type(cpd_t)                      :: best_cpd, best_cpd_origgrid
  double precision                 :: best_rms
  integer                          :: best_iter, best_rank
  logical                          :: lnewbestfit

  integer                          :: newrank

  double precision, parameter      :: au2ic =  2.1947463137d+5
  double precision, parameter      :: au2ev =  27.21138386d0
  double precision, parameter      :: au2mev = 27211.38386d0


  ! BLAS
  double precision                 :: ddot
  external ddot

  integer(long), save              :: tid=0 ! timer ID

  itercount = 0

  nsample = size(modeindex,1)

  call init_mcals(nsample,maxrank)
  call init_mgroups(cpd%nmode)
  if (lwrfuncs) call init_jfunc(cpd%nmode)

  ! check for warn-messages
  ! (exit optimization after warning)
  lwarn1 = lwarn
  lwarn = .false.

  if (mpirank.eq.0) then

     do i=1,nline
        line(i:i) = "-"
     end do

     startiter = 0
     iterationsfile = trim(name)//"/iterations"
     if (lcont) then
        call get_als_startiter(startiter, iterationsfile)
        open(iit,file=trim(iterationsfile),iostat=ierr,position='append')
     else
        open(iit,file=trim(iterationsfile),iostat=ierr)
     end if
     if (ierr.ne.0) then
        message = "Cannot open file "//trim(iterationsfile)
        routine = "optimize_cpd (natpot2cpd.F90)"
        call errormsg
     end if

     ! drop the stop-file
     call drop_als_stop()

     iiter = startiter

     write(ilog,'(/a)') "Starting ALS iterations. For details see file:"
     write(ilog,'(a/)') trim(iterationsfile)

     call myflush(ilog)

     ! haeder
     write(*,*)
     write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","|C|^2", "RMS[au]^2 + eps*|C|^2"
     write(*, '(a)') trim(line)
     if (.not. lcont) then
        write(iit, 197) "#","iiter","RMS[cm-1]","RMS[meV]","RMS[au]","|C|^2", "RMS[au]^2 + eps*|C|^2"
        line(1:1) = "#"
        write(iit, '(a)') trim(line)
     end if

  end if !mpirank


#ifdef MPI
  call MPI_Bcast(startiter,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
#endif

  call mccpd_fast_rms(rms,cf2,cpd%coeff,cpd%rank)

  if (mpirank.eq.0) then

     jfunc = rms**2 + cf2*epsilon_invert

     write(*,  203) iiter, rms*au2ic, rms*au2mev, rms, cf2, jfunc,' #guess'
     write(iit,203) iiter, rms*au2ic, rms*au2mev, rms, cf2, jfunc,' #guess'
     call myflush(iit)

     call copy_cpd(cpd, best_cpd)

     if (lextgrid) then
        call copy_cpd(cpd_origgrid, best_cpd_origgrid)
     end if

     best_rms = rms
     best_iter = startiter
     best_rank = best_cpd%rank
     lnewbestfit = .false.
  end if

  ! main loop
  do iiter=startiter+1,alsmaxiter

     ! this is for logging and
     ! fixing epsilon afetr 2 iterations
     curriter = iiter
     itercount = itercount + 1

     ! check if we should stop after this sweep
     if (mpirank.eq.0) then
        call chk_als_stop(lstopals)
     end if

#ifdef MPI
     call MPI_Bcast(lstopals,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
#endif

     if (mpirank.eq.0) then
        call get_timer(tid, 'ptiteration')
        call start_timer(tid)
    endif

     lmdone = .false.

     do m=1,cpd%nmode

        if (lmdone(m)) cycle

        lmupdate = .false.
        lmupdate(m) = .true.
        if (lsymmetries) then
           do n=1,nsymmetries
              m1 = symmmap(m,n)
              lmupdate(m1) = .true.
           end do
        end if

        ! write eigenvalues in last iteration
        lwrefile = ( (iiter .eq. alsmaxiter .or. lstopals) &
             .and. invertmethod  .eq. mc_inv_eigen)


        ! solve for one mode
        call mcalsiter1m(cpd, cpd_origgrid, m,ierr, lwrefile, invertmethod)

        if (ierr .ne. 0) then
           routine = "optimize_cpd(cpd, natpot)"
           write(message,'(a,i6)') "ALS failed with error code", ierr
           call errormsg
        end if


        ! update re-usable stuff
        do m1=1,cpd%nmode
           if (lmupdate(m1)) then
              lmdone(m1) = .true.
              call updatecpdsamplings1m(cpd,modeindex,nsample,m1)
           end if
        end do

     end do ! modes


     call mccpd_fast_rms(rms,cf2,cpd%coeff,cpd%rank)

     if (mpirank.eq.0) then

        jfunc = rms**2 + cf2*epsilon_invert

        write(*,  200) iiter,rms*au2ic, rms*au2mev, rms, cf2, jfunc
        write(iit,200) iiter,rms*au2ic, rms*au2mev, rms, cf2, jfunc
        call myflush(iit)
     end if

     ! detailed functional output if requested
     if (lwrfuncs) then
        call wralsfuncs(iiter)
     end if

     if (mpirank.eq.0) call stop_timer(tid)

     ! save the best solution we have so far
     if (rms .lt. best_rms) then

        call copy_cpd(cpd, best_cpd)
        if (lextgrid) then
           call changerank(cpd_origgrid, cpd%rank, .true.)
           call copy_cpd(cpd_origgrid, best_cpd_origgrid)
           call extcpd2origcpd(cpd,best_cpd_origgrid)
        end if

        best_rms = rms
        best_iter = iiter
        best_rank = best_cpd%rank
        lnewbestfit = .true.
     end if

     if (lwarn) exit    ! iteration loop
     if (lstopals) exit ! iteration loop


     ! grow if requested
     if (lgrow .and. cpd%rank .lt. maxrank) then

        if (modulo(iiter,growiter) .eq. 0 .and. iiter .lt. alsmaxiter) then

           if (cpd%rank .lt. maxrank) then
              newrank = cpd%rank +  growrank
              newrank = min(newrank, maxrank)
           end if

           if (mpirank.eq.0) then

              ! dump a copy of the best cpd for current rank
              if (lnewbestfit) then
                 lnewbestfit = .false.
                 write(fname,'(a,i8)') "best_cpd_rank_", best_cpd%rank
                 call rmblnk(fname)

                 write(ilog,'(a)') ""
                 write(ilog,'(a)') "Dumping a copy of the best fit so far:"
                 write(ilog,'(a)')    "  Filename:   " //  trim(fname)
                 write(ilog,'(a,i4)') "  Iter =      ", best_iter
                 write(ilog,'(a,i4)') "  Rank =      ", best_cpd%rank
                 write(ilog,201)      "  RMS =       ", best_rms*au2ic,  " cm-1"
                 write(ilog,201)      "      =       ", best_rms*au2mev, " meV"
                 write(ilog,202)      "      =       ", best_rms,        " au"
                 write(ilog,'(a)') ""
                 call myflush(ilog)

                 fname = trim(name) // "/" //  trim(fname)
                 if (lextgrid) then
                    call exportMCTDH(best_cpd_origgrid, dvr, fname, lovwr,loperfit)
                 else
                    call exportMCTDH(best_cpd, dvr, fname, lovwr,loperfit)
                 end if
              end if

              write(message,'(a,i8,a,i8)') "Changing rank ", cpd%rank, " --> ", newrank
              call cncntrblnk(message)
              write(*,'(a)') trim(message)
              write(iit,'(a)') "# "//trim(message)

              call myflush(iit)

              ! we use the best so far obtained
              call copy_cpd(best_cpd, cpd)

              if (lsymmetries) then
                 call changerank(cpd, newrank, .true., nsymmetries)
                 call symmetrizecpd(cpd)
              else
                 call changerank(cpd, newrank, .true.)
              end if
           end if

           call mpi_bcast_cpd_t(cpd)
           ! pre-fetch the spp on sampling points
           call mkcpdsamplings(cpd,modeindex,nsample)
        end if


     end if


  end do !iiter

  if (mpirank.eq.0) close(iit)

  if (mpirank.eq.0) then
     if (lstopals) then
        write(ilog,'(/a,i8/)') "Iterations stoped by " &
             //"stop-file at iteration ", iiter
     end if

     write(ilog,'(a)')    "* Using best tensor obtained so far:"
     write(ilog,'(a,i8)') "* iteration: ", best_iter
     write(ilog,201)      "* RMSE     = ", best_rms*au2ic,  " cm-1"
     write(ilog,201)      "*          = ", best_rms*au2mev, " meV"
     write(ilog,202)      "*          = ", best_rms,        " au"
     call myflush(ilog)
  end if

  if (mpirank.eq.0) then
     call copy_cpd(best_cpd, cpd)
  end if
  call mpi_bcast_cpd_t(cpd)

  call deallocate_cpd_t(best_cpd)
  call deallocate_cpd_t(best_cpd_origgrid)

  ! print the header again on screen
  if (mpirank.eq.0) then
     line(1:1) = "-"
     write(*, '(a)') trim(line)
     write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","|C|^2", "RMS[au]^2 + eps*|C|^2"
     write(*, '(a)') trim(line)
  end if

  lwarn = lwarn1 .or. lwarn
  call free_mcals()
  call disposecpdsamplings()
  close(iit)

  return


196 FORMAT (  3x,a,9x,a,7x,a,12x,a,15x,a,6x,a)
197 FORMAT (a,2x,a,9x,a,7x,a,12x,a,15x,a,6x,a)

200 FORMAT (i8,3x,f14.4,1x,f14.4,6x,ES14.8,2(6x,ES14.8))
203 FORMAT (i8,3x,f14.4,1x,f14.4,6x,ES14.8,2(6x,ES14.8),a)
201 FORMAT (a, f14.4, a)
202 FORMAT (a,ES14.8, a)


end subroutine mc_optimize_cpd


subroutine mccpd_fast_rms(rms,cf2,coeff,rank)
  implicit none

  double precision, intent(OUT) :: rms,cf2
  integer, intent(IN)           :: rank
  double precision, intent(IN)  :: coeff(rank)
  integer           :: m,n,r, nmode, nsample
  double precision  :: term, dtmp, cr(rank), rdum
  double precision, allocatable :: tmp(:)

  double precision  :: ddot

#ifdef MPI
  integer           :: mpierr, ibuf
  double precision  :: sbuf, rbuf(rank)
#endif

  external ddot


  nsample = size(basissamplings,1)
  nmode   = size(basissamplings,2)

 if (.not. allocated(rtmp)) then
    allocate(rtmp(nsample))
 else if (size(rtmp,1) .ne. nsample) then
    deallocate(rtmp)
    allocate(rtmp(nsample))
 end if


  !$omp parallel do private(r,term,m,dtmp) schedule(static)
  do n=1,nsample
     dtmp = 0.d0
     do r=1,rank
        ! basissamplings should be transposed. but it can be
        ! really big so one may not want to store a second version.
        term = coeff(r)*basissamplings(n,1,r)*basissamplings(n,2,r)
        do m=3,nmode
           term = term*basissamplings(n,m,r)
        end do
        dtmp = dtmp + term
     end do
     rtmp(n) = dtmp
  end do
  !$omp end parallel do

  rms = 0.d0
  !$omp parallel do simd reduction(+:rms) schedule(static)
  do n=1,nsample
     rms = rms + (energies(n) - rtmp(n))**2
  end do
  !$omp end parallel do simd


  allocate(tmp(nsample))
  do r=1,rank
     !$omp parallel do simd schedule(static)
     do n=1,nsample
        tmp(n) = basissamplings(n,1,r)*basissamplings(n,2,r)
     end do
     !$omp end parallel do simd

     do m=3,nmode
        !$omp parallel do simd schedule(static)
        do n=1,nsample
           tmp(n) = tmp(n)*basissamplings(n,m,r)
        end do
        !$omp end parallel do simd
     end do
     rdum = 0.d0
     !$omp parallel do simd reduction(+:rdum) schedule(static)
     do n=1,nsample
        rdum = rdum +  tmp(n)**2
     end do
     !$omp end parallel do simd

     cr(r) = rdum

  end do
  deallocate(tmp)



!!$  !$omp parallel do private(n,term,m) reduction(+:cf2)
!!$  do r=1,rank
!!$     cr(r) = 0.d0
!!$     do n=1,nsample
!!$        term = basissamplings(n,1,r)*basissamplings(n,2,r)
!!$        do m=3,nmode
!!$           term = term*basissamplings(n,m,r)
!!$        end do
!!$        cr(r) = cr(r) + term**2
!!$     end do
!!$  end do
!!$  !$omp end parallel do


#ifdef MPI

  sbuf = rms

  call MPI_AllReduce(sbuf,rbuf(1),1,&
       MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,mpierr)
  rms = rbuf(1)
  call MPI_AllReduce(cr,rbuf,rank,&
       MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,mpierr)
  cr = rbuf

  call MPI_AllReduce(nsample,ibuf,1,&
       MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,mpierr)
  nsample = ibuf

#endif

  rms = sqrt(rms/nsample)

  cf2 = ddot(rank,cr,1,coeff(:)**2,1)/nsample

!  !$omp parallel do reduction(+:cf2)
!  do r=1,rank
!     cf2 = cf2 + cr(r)*coeff(r)**2
!  end do
!  !$omp end parallel do
!  cf2 = cf2/nsample

end subroutine mccpd_fast_rms







subroutine chk_als_stop(lstopals)

  implicit none

  logical, intent(OUT) :: lstopals

  character(len=c5)    :: stopfile
  character(len=c4)    :: content
  logical              :: lexists
  integer              :: ilbl


  lstopals = .false.

  stopfile = trim(name)//"/stop"
  inquire(file=stopfile, exist=lexists)
  if (.not. lexists) then
     lstopals = .true.
  else
     open(istop, file=stopfile, err=398)
     read(istop,*,err=399) content
     close(istop)
     call cncntrblnk(content)
     content = adjustl(content)
     ilbl = index(content,' ')
     if (content(1:ilbl) .eq. "continue") then
        lstopals = .false.
     else
        lstopals = .true.
     end if
  end if

  return

398 continue
  ! read or open-error
  message = "Could not open stop file: "//trim(stopfile)
  routine = ""
  call warnmsg
  lstopals = .true.
  return

399 continue
  ! read or open-error
  message = "Could not read stop file: "//trim(stopfile)
  routine = ""
  call warnmsg
  lstopals = .true.
  return

end subroutine chk_als_stop



subroutine drop_als_stop()

  implicit none

  character(len=c5)    :: stopfile
  logical              :: lexists
  integer              :: stat

  stopfile = trim(name)//"/stop"
  inquire (file=stopfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "drop_mcals_stop"
     message = "Cannot remove old stopfile "//trim(stopfile)//". Use -w option."
     call errormsg
  else
     open(unit=istop, iostat=stat, file=stopfile, status='unknown')
     if (stat .ne. 0) then
        routine = "drop_mcals_stop"
        message = "Unknown error while opening file "//trim(stopfile)//"."
        call errormsg
     end if
     write(istop,'(a)',iostat=stat) "continue"
     if (stat .ne. 0) then
        routine = "drop_mcals_stop"
        message = "Unknown error while writing file "//trim(stopfile)//"."
        call errormsg
     end if
     close(istop)
  end if




end subroutine drop_als_stop




! search the iterations file
! to get the current iteration number
! for restart jobs
subroutine get_als_startiter(startiter,iterationsfile)

  implicit none

  integer, intent(OUT) :: startiter
  character(len=*), intent(IN) :: iterationsfile

  integer  :: ierr, ierr1

  open(idump, file=iterationsfile, status='old', action='read', iostat=ierr)
  if (ierr.ne.0) then
     routine = "get_als_startiter"
     message = "Unable do get restart iteration number: could not open file: " &
          //trim(iterationsfile)//". Will start at zero."
     call warnmsg
     startiter = 0
     return
  end if

  ierr = 0
  do while(ierr.eq.0)
     read(idump, *, iostat=ierr)

     if (ierr.lt.0) then
        ! we arrived behind the EOF marker
        backspace(idump)
        ! we are just in front of the EOF marker
        backspace(idump)
        ! we are at the begining of last line
        read(idump, *, iostat=ierr1) startiter
        if (ierr1.ne.0) goto 112
        close(idump)
        return
     elseif (ierr.gt.0) then
        ! read error
        goto 112
     else
        continue
     end if

  end do
  close(idump)

  return

112 continue
  routine = "get_als_startiter"
  message = "Unable do get restart iteration number: read error on file: " &
       //trim(iterationsfile)//". Will start at zero"
  call warnmsg
  close(idump)
  startiter = 0
  return

end subroutine get_als_startiter

end module mcals_m
