!  -*- f90 -*-

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Module 'mccomplete'
!
! Create and read/write a vpot file in case complete sampling
! is used.
!
! This is a bit like our red-headed stepchild. Feel free
! to treat it better then we did so far... :)
!
! 12/17 MS
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


module mccomplete
    use global
    use pglobal
    use channels
    use maxv
    use griddatmod
    use combined
    use logdat
    use timing
    use versions
    use vpotpars
    use natpottype
    use dvrtype_m
    use tensortype
    use mcpfitparams
    use pftools
    use pesselect_m
    use mpidata

#ifdef MPI
    use mpi
#endif

  implicit none

  double precision, allocatable,target :: vpot(:)

contains

  subroutine mcmkvpot(natpot, dvr)

    implicit none

    type(natpot_t), intent(IN)  :: natpot
    type(dvr_t), intent(IN)     :: dvr

    integer                     :: i, allocstat
    integer                     :: idx(natpot%ndof)
    integer                     :: midx(natpot%nmode)
    character(len=c5)           :: fname
    logical                     :: lexist
    integer                     :: istart, iend
    integer, allocatable        :: ichunk(:,:)
    double precision, pointer   :: rsendbuf(:)

    integer(long), save                        :: tid=0 ! timer ID

#ifdef MPI
    integer                      :: icn,ir,mpitag,clen
    integer                      :: status(MPI_STATUS_SIZE)
#endif

    if (mpirank.eq.0) then


       if ( abs(natpot%rvpotdim - dble(natpot%vpotdim)) .gt. 0.5) then
          write(message,'(a,E9.3)')"vpotdim is too large to be allocated: ",natpot%rvpotdim
          routine="mcmkvpot"
          call errormsg
       end if

       fname=trim(name)//"/vpot"
       inquire(file=fname, exist=lexist)
       if (lexist .and. .not. lovwr) then
          message = "Cannot overwrite "//trim(fname)//". Use -w option."
          routine = "mcmkvpot"
          call errormsg
       end if

       if (.not. allocated(vpot)) then
          allocate(vpot(natpot%vpotdim), stat=allocstat)
          if (allocstat .ne. 0) then
             routine="mcmkvpot"
             message="Cannot allocate vpot"
             call errormsg
          end if
       elseif (size(vpot) .ne. natpot%vpotdim) then
          message="Allocated vpotdim does not match requested vpotdim. Check the code."
          routine="mcmkvpot"
          call errormsg
       end if
    end if

    allocate(ichunk(2,0:nmpi-1))


#ifdef MPI
    ! In MPI mode each rank gets a chunk to sample.
    ! calculate the chuncs and send them to the tasks
    if (mpirank.eq.0) then
       icn = natpot%vpotdim/nmpi
       ir = modulo(natpot%vpotdim,nmpi)

       istart = 1
       iend = icn

       if (ir.gt.0) then
          iend = iend +1
          ir = ir -1
       end if

       ichunk(1,0) = istart
       ichunk(2,0) = iend

       do i=1,nmpi-1

          ichunk(1,i) = ichunk(2,i-1)+1
          iend = ichunk(1,i) + icn -1
          if (ir.gt.0) then
             iend = iend +1
             ir = ir -1
          end if
          ichunk(2,i) = iend
       end do

    end if

    call mpi_bcast(ichunk,nmpi*2,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)

    istart = ichunk(1,mpirank)
    iend = ichunk(2,mpirank)


#else
    ! If no MPI is used everithing has to
    ! be done in rank 0
    istart = 1
    iend = natpot%vpotdim
#endif

    if (mpirank.eq.0) then
       call get_timer(tid, 'vextiming')
       call start_timer(tid)
    endif


    if (mpirank .eq. 0) then
       rsendbuf => vpot
    else
       allocate(rsendbuf(istart:iend))
    end if


    ! get vpot in mode-order
    !$omp parallel do private(idx,midx) if(lomppotential)
    do i=istart,iend
       call cindex2sindex(i,midx,natpot%subdim,natpot%nmode)
       call m2fdecoder(midx, idx, natpot%nmode, natpot%ndof, natpot%spfdof, &
            natpot%nspfdof, natpot%gdim, natpot%maxspfdof)

       rsendbuf(i) = potential(idx,natpot%ndof)
    end do
    !$omp end parallel do

#ifdef MPI

    ! collect data from the other ranks

    if (mpirank.ne.0) then

       mpitag = 0
       clen = iend - istart + 1

       call mpi_send(rsendbuf, clen, MPI_DOUBLE_PRECISION,0, &
            mpitag,MPI_COMM_WORLD,mpierr)

    else
       do i=1,nmpi-1

          mpitag = 0
          clen = ichunk(2,i) - ichunk(1,i) + 1

          call mpi_recv(vpot(ichunk(1,i)),clen,MPI_DOUBLE_PRECISION,i, &
               mpitag,MPI_COMM_WORLD,status, mpierr)

       end do
    end if
    call mpi_barrier(MPI_COMM_WORLD,mpierr)

#endif

    ! crean up and write to file

    deallocate(ichunk)

    if (mpirank.ne.0) then
       deallocate(rsendbuf)
    else
       call savemcvpot(natpot, dvr)
    end if

    if (mpirank.eq.0) call stop_timer(tid)


  end subroutine mcmkvpot

  ! convert the combined index I (V_I) in mode order to
  ! combined index J (V_J) in dof order
  subroutine modegrid2dofgrid(i,j,subdim,gdim,ndof,nmode,nspfdof,spfdof,maxspfdof)
    implicit none
    integer, intent(IN)  :: i, ndof,nmode, maxspfdof
    integer, intent(IN)  :: gdim(ndof), subdim(nmode), nspfdof(nmode), spfdof(maxspfdof,nmode)
    integer, intent(OUT) :: j
    integer              :: midx(nmode), idx(ndof), tgdim(maxspfdof),tidx(maxspfdof)
    integer              :: m,f,n

    call cindex2sindex(i,midx,subdim,nmode)
    do m=1,nmode
       tgdim = 0
       do n=1,nspfdof(m)
          f = spfdof(n,m)
          tgdim(n) = gdim(f)
       end do
       call cindex2sindex(midx(m),tidx,tgdim,nspfdof(m))
       do n=1,nspfdof(m)
          f = spfdof(n,m)
          idx(f) = tidx(n)
       end do
    end do
    call sindex2cindex(j,idx,gdim,ndof)

  end subroutine modegrid2dofgrid


  ! convert the combined index I (V_I) in dof order to
  ! combined index J (V_J) in mode order
  subroutine dofgrid2modegrid(i,j,subdim,gdim,ndof,nmode,nspfdof,spfdof,maxspfdof)

    implicit none
    integer, intent(IN)  :: i, ndof,nmode, maxspfdof
    integer, intent(IN)  :: gdim(ndof), subdim(nmode), nspfdof(nmode), spfdof(maxspfdof,nmode)
    integer, intent(OUT) :: j
    integer              :: midx(nmode), idx(ndof), tgdim(maxspfdof),tidx(maxspfdof)
    integer              :: m,f,n

    call cindex2sindex(i,idx,gdim,ndof)
    do m=1,nmode
       tgdim = 0
       tidx = 0
       do n=1,nspfdof(m)
          f = spfdof(n,m)
          tidx(n) = idx(f)
          tgdim(n) = gdim(f)
       end do
       call sindex2cindex(midx(m),tidx,tgdim,nspfdof(m))
    end do
    call sindex2cindex(j,midx,subdim,nmode)

  end subroutine dofgrid2modegrid



  ! write the vpot file
  subroutine savemcvpot(natpot, dvr)

    implicit none

    type(natpot_t), intent(IN)  :: natpot
    type(dvr_t), intent(IN)     :: dvr
    character(len=c5)           :: outname
    integer                     :: ierr
    logical                     :: lexist


    outname = trim(name)//"/vpot"

    if (.not.lovwr) then
       inquire(file=outname,exist=lexist)
       if (lexist) then
          routine='savemcvpot'
          message='File already exists: '//trim(outname)
          call errormsg
       endif
    endif

    open(unit=ivpot,status='unknown',file=trim(outname), &
         form='unformatted',iostat=ierr)
    if (ierr.ne.0) then
       message='could not open '//trim(outname)//' for writing'
       routine='savemcvpot'
       call errormsg
    endif
    call myflush(ilog)


    filever(ivpot)=progver
    write(ivpot,iostat=ierr) progver
    if (ierr.ne.0) then
       routine='getpes'
       message='cannot write program version to '//trim(outname)
       call errormsg
    endif

    ! write DVR header"
    call exportDVRdef(dvr,ivpot)

    ! Write vpot parameters (text)"
    call writevpotpars(ivpot,outname,0)

    ! write potential data
    call write_mcvpot(ivpot,outname,natpot)


    close(ivpot)

  end subroutine savemcvpot


  subroutine write_mcvpot(unit,outname, natpot)

    implicit none

    type(natpot_t), intent(IN)    :: natpot
    integer, intent(IN)           :: unit
    character(len=*), intent(IN)  :: outname
    integer                       :: niobuf1, ierr
    integer                       :: nchunks, nrest
    integer                       :: j,jj,j1,j2,j3,k,bmb,f,ff,mm
    character(len=80)             :: actual
    double precision, allocatable :: buffer(:)


    bmb = 256*1024/8*1024 ! 256 MB of real*8
    if (natpot%vpotdim .le. bmb) then
       niobuf = natpot%vpotdim
    else
       niobuf=bmb
       if(natpot%vpotdim .le. 2*bmb) niobuf=bmb/2
       ff = 1
       do f=1,natpot%ndof
          if(natpot%vgpdim(f).le.niobuf) ff=f
       enddo
       mm=niobuf/natpot%vgpdim(ff)
       niobuf=mm*natpot%vgpdim(ff)
    end if

    allocate(buffer(niobuf))


    niobuf1=min(natpot%vpotdim,niobuf)
    write(unit,iostat=ierr) natpot%vpotdim,niobuf1
    if (ierr.ne.0) then
       routine='write_vpot'
       message='error on writing nitems to '//trim(outname)
       call errormsg
    endif

    nchunks = natpot%vpotdim/niobuf
    nrest = modulo(natpot%vpotdim,niobuf)
    j1=1
    if (nchunks.eq.0) then
       j2 = nrest
    else
       j2 = niobuf
    end if

    do k=1,nchunks
       j3 = 0
       do j=j1,j2
          j3 = j3+1
          ! vpot is stored in dof order so we need to fetch
          ! the mode-order j (vpot in in mode-order in RAM)
          call dofgrid2modegrid(j,jj,natpot%subdim,natpot%gdim, &
               natpot%ndof,natpot%nmode,natpot%nspfdof,natpot%spfdof, &
               natpot%maxspfdof)
          buffer(j3) = vpot(jj)

       end do
       write(ivpot) buffer

       call fzeit(actual)
       write(ilog,'(a,i10,i9,3x,a16)') 'Buffer written. j1, j2 =',j1,j2,actual
       call myflush(ilog)
       j1 = j1+niobuf
       j2 = j2+niobuf
    end do

    if (nrest .gt.0) then
       j2 = j1 + nrest -1
       if (j2.ne.natpot%vpotdim) write(*,*) "ERROR ERROR ERROR!!!"

       j3 = 0
       do j=j1,j2
          j3 = j3 + 1
          call dofgrid2modegrid(j,jj,natpot%subdim,natpot%gdim, &
               natpot%ndof,natpot%nmode,natpot%nspfdof,natpot%spfdof, &
               natpot%maxspfdof)
          buffer(j3) = vpot(jj)
       end do

       write(ivpot)buffer(1:nrest)
       call fzeit(actual)
       write(ilog,'(a,i10,i9,3x,a16)') 'Buffer written. j1, j2 =',j1,j2,actual
       call myflush(ilog)
    end if


  end subroutine write_mcvpot


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! load vpot from file
! input   dvr_t (system dvr)
! output  ierr:
!   ierr == 0: OK, vpot is allocated and loaded
!   ierr == 1: file does not exist, vpot is not loaded
!   ierr == 2: system is not compatible, vpot is not loaded
!   ierr == 3: I/O error, vpot is corrupted and not loaded
!
! NOTE!!!!
! loadmcvpot will store the vpot in mode-order, not DOF order
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine loadmcvpot(dvr,tensor,fname,ierr)
    
    implicit none

    type(dvr_t), intent(IN)      :: dvr
    class(tensor_t), intent(IN)  :: tensor
    integer, intent(OUT)         :: ierr
    character(len=*), intent(IN) :: fname


    type(dvr_t)                    :: vpot_dvr

    logical                        :: lexist, lcompatible(7)
    integer                        :: allocstat

    integer                        :: mbaspar1
    integer                        :: i,j,jj,j1,j2,j3,f,k
    integer                        :: nchunks, nrest
    integer                        :: vpotdim, niobuf1
    double precision, allocatable  :: buffer(:)
    double precision               :: pesmin,pesmax,v
    integer(long), save                        :: tid=0 ! timer ID

    ierr = 0

    inquire(file=fname, exist=lexist)
    if (.not. lexist) then
       ierr = 1
       return
    end if

    if (mpirank.eq.0) then
       call get_timer(tid, 'loadv')
       call start_timer(tid)
    endif

    ! copy dvr to include dvrdef
    ! to suppress conflicts
    ndof = min(dvr%ndof,maxdim)

    ldvr(1:ndof) = dvr%ldvr(1:ndof)
    basis(1:ndof) = dvr%basis(1:ndof)
    ipbaspar(1:mbaspar, 1:ndof) = dvr%ipbaspar(1:mbaspar, 1:ndof)
    rpbaspar(1:mbaspar, 1:ndof) = dvr%rpbaspar(1:mbaspar, 1:ndof)
    modelabel(1:ndof) = dvr%modelabel(1:ndof)
    gdim(1:ndof) = dvr%gdim(1:ndof)
    xend(:, 1:ndof) = dvr%xend(:, 1:ndof)


    open(ivpot,file=fname, status="old",action="read",form="unformatted",err=666)
    read(ivpot) vpot_dvr%filever
    call importDVRdef(vpot_dvr,ivpot)
    vpot_dvr%initialized = .true.

    call compareDVR(dvr,vpot_dvr,chkerr, nerr, ilog)

    call myflush(ilog)
    if (any(chkerr .ne. 0)) then
       routine = "loadmcvpot"
       message = "DVR of vpot differs from system DVR."
       call errormsg
    end if

    lcompatible(1:4) = .true.
    lcompatible(5) = all(dvr%ldvr .eqv. vpot_dvr%ldvr)
    lcompatible(6) = .true.
    lcompatible(7) = .true.

    mbaspar1 = min(dvr%mbaspar, vpot_dvr%mbaspar)

    do f=1,min(dvr%ndof,vpot_dvr%ndof)
       do i=1,mbaspar1
          lcompatible(6) =  lcompatible(6) .and. &
               dvr%ipbaspar(i,f) .eq.  vpot_dvr%ipbaspar(i,f)
          lcompatible(7) =  lcompatible(7) .and. &
               abs(dvr%rpbaspar(i,f) -  vpot_dvr%rpbaspar(i,f)) .lt. 1.d-8

       end do
    end do

    write(ilog,*)
    write(ilog,'(a,i4,a,i4)') "No DOF, system:", dvr%ndof, " vpot:", vpot_dvr%ndof
    write(ilog,'(a)') "system  type  gdim | vpot    type  gdim"
    write(ilog,'(a)') "---------------------------------------"
    do f=1,min( dvr%ndof,vpot_dvr%ndof)
       write(ilog,'(a,2i5,2a,2i5)') dvr%modelabel(f)(1:8), dvr%basis(f), dvr%gdim(f), " | ", &
            vpot_dvr%modelabel(f)(1:8), vpot_dvr%basis(f), vpot_dvr%gdim(f)
    end do
    if (dvr%ndof .gt. vpot_dvr%ndof) then
       write(ilog,'(a)') "Further DOF in system:"
       do f=vpot_dvr%ndof, dvr%ndof
          write(ilog,'(2i4)') dvr%modelabel(f)(1:8), dvr%basis(f)
       end do
    end if
    if (vpot_dvr%ndof .gt. dvr%ndof) then
       write(ilog,'(a)') "Further DOF in vpot:"
       do f=dvr%ndof, vpot_dvr%ndof
          write(ilog,'(19x,2i4)') vpot_dvr%modelabel(f)(1:8), vpot_dvr%basis(f)
       end do
    end if

   if ( .not. all(lcompatible)) then
       routine = "loadmcvpot"
       message = "DVR of vpot differs from system. See log file. "
       call warnmsg
       call myflush(ilog)
       ierr = 2
       goto 666
    end if

    !
    !call skipvpotpars(ivpot)
    call pesminmaxfrompars(ivpot,pesmin,pesmax)

    write(ilog,'(a,e20.8)') 'max pes cut (au) : ',pesmax
    write(ilog,'(a,e20.8)') 'min pes cut (au) : ',pesmin

    if (allocated(vpot)) deallocate(vpot)
    allocate(vpot(tensor%vpotdim), stat=allocstat)
    if (allocstat .ne. 0) then
       routine="mcmkvpot"
       message="Cannot allocate vpot"
       call errormsg
    end if


    call skipvpotpars(ivpot)
    read(ivpot) vpotdim,niobuf1

    if (vpotdim .ne. tensor%vpotdim) then
       routine = "loadmcvpot"
       write(message,'(a,2i8)')"vpot dimension differs. System and vpot: ", tensor%vpotdim, vpotdim
       call warnmsg
       goto 666
    end if

    allocate(buffer(niobuf1))




    write(ilog,'(a)',advance='no') "vpot is read from file... "


    nchunks = vpotdim/niobuf1
    nrest = modulo(tensor%vpotdim,niobuf1)
    j1=1
    if (nchunks.eq.0) then
       j2 = nrest
    else
       j2 = niobuf1
    end if

    do k=1,nchunks

       read(ivpot, iostat=ierr) buffer
       if (ierr.ne.0) then
          ierr = 3
          goto 666
       end if

       j3 = 0
       do j=j1,j2
          j3 = j3+1
          ! vpot is stored in dof order so we need to put it
          ! in mode-order j (vpot in in mode-order in RAM)
          call dofgrid2modegrid(j,jj,tensor%subdim,tensor%gdim, &
               tensor%ndof,tensor%nmode,tensor%nspfdof,tensor%spfdof, &
               tensor%maxspfdof)
          vpot(jj) = buffer(j3)
       end do
       j1 = j1+niobuf1
       j2 = j2+niobuf1
    end do

    if (nrest .gt.0) then
       j2 = j1 + nrest -1

       if (j2.ne.tensor%vpotdim) write(*,*) "ERROR ERROR ERROR!!!"

       read(ivpot,iostat=ierr) buffer(1:nrest)
       if (ierr.ne.0) then
          ierr = 3
          goto 666
       end if

       j3 = 0
       do j=j1,j2
          j3 = j3 + 1
          call dofgrid2modegrid(j,jj,tensor%subdim,tensor%gdim, &
               tensor%ndof,tensor%nmode,tensor%nspfdof,tensor%spfdof, &
               tensor%maxspfdof)
          vpot(jj) = buffer(j3)

       end do
    end if

    !$omp parallel do private(v)
    do j=1,tensor%vpotdim
       v=min(vpot(j),pesmax)
       vpot(j)=max(v,pesmin)
    enddo
    !$omp end parallel do

    if (mpirank.eq.0) call stop_timer(tid)

    write(ilog,'(a)') "OK"
    call myflush(ilog)


    ! cleanup
    if (allocated(buffer)) deallocate(buffer)
    call deallocate_dvr_t(vpot_dvr)
    close(ivpot)
    ierr = 0
    return

666 continue

    ! cleanup on error
    if (allocated(buffer)) deallocate(buffer)
    if (allocated(vpot)) deallocate(vpot)
    call deallocate_dvr_t(vpot_dvr)
    close(ivpot)

    return

  end subroutine loadmcvpot








  subroutine pesminmaxfrompars(unit,pesmin1,pesmax1)


    implicit none

    integer, intent(IN)           :: unit
    double precision, intent(OUT) :: pesmin1,pesmax1
    character(len=c5)             :: label,opts

    logical                       :: lsp1,lcspot1,lftp1
    double precision              :: progver1
    double precision              :: jtot1,jbf1,csmass1
    integer                       :: ierr,pdofft1,fdvr(maxdim),chkdvr

    rewind(unit,iostat=ierr)
    if (ierr.ne.0) then
       routine='pesminmaxfrompars'
       message='cannot rewind vpot file'
       call errormsg
    endif

    chkdvr=6
    call readvpotinfo(unit,progver1,label,opts,lsp1, &
         lftp1,pdofft1,pesmin1,pesmax1,chkdvr, &
         lcspot1,jtot1,jbf1,csmass1,fdvr)


  end subroutine pesminmaxfrompars











end module mccomplete
