! -*- f90 -*-


subroutine mccpd_initguessbasis(cpd)
  use global
  use cpdtype_m
  use cpdtype_s, only: normspp1m
  use mccuts_m
  use symmops_m
  implicit none
  type(cpd_t), intent(INOUT) :: cpd

  integer                    :: nstep,r,m,k
  integer                    :: rpersym
  double precision           :: targetnorm
  logical                    :: lfillrand, lwarn1


  if (lsymmetries) then
     rpersym = cpd%rank/nsymmetries
  else
     rpersym = cpd%rank
  end if

  nstep = 1
  lfillrand = .false.

  ! fill cuts as basisvectors
  k = 1
  do r=1,rpersym
     cpd%coeff(r) = 0.d0
     do m=1,cpd%nmode

        if (k.le. mccuts(m)%ncuts) then
           cpd%spp(m)%htm(:,r) = mccuts(m)%cuts(1:cpd%subdim(m), k)
        else
           call random_number(cpd%spp(m)%htm(:,r))
           cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r) - 0.5d0
           lfillrand = .true.
        end if
     end do
     k = k + nstep
  end do



  if (lfillrand) then

     lwarn1 = lwarn
     routine = "mccpd_initguessbasis"
     message = "There are not enough cuts through the PES to populate " &
          //"the initial guess. Filling with random numbers."

#ifdef MPI
     message = "There are not enough cuts through the PES in MPI rank 0 " &
          //"to populate the initial guess. Filling with random numbers."
#endif

     call warnmsg
     lwarn = lwarn1

  end if


  if (lsymmetries) then
     call symmetrizecpd(cpd)
  end if

  do m=1,cpd%nmode
     targetnorm = cpd%subdim(m)
     call normspp1m(cpd,m, targetnorm)
  end do


#ifdef DEBUG
  !NaN is not equal to anything, not even to itself
  do m=1,cpd%nmode
     if( any( cpd%spp(m)%htm .ne. cpd%spp(m)%htm ) ) then
        print*, "NaN in CPD basis functions!"
        stop
     end if
  end do
#endif



end subroutine mccpd_initguessbasis





subroutine mccpd_coeff(cpd, energies, nsample, invertmethod)
  use cpdtype_m
  use cpdsamplings_m
  use symmops_m
  use mccpdparams, only: epsilon_invert,lcoeffoncuts

  use global
  use pglobal
  use channels
  use mpidata
#ifdef MPI
  use mpi
#endif

  implicit none

  type(cpd_t), intent(INOUT)    :: cpd
  integer, intent(IN)           :: nsample, invertmethod
  double precision, intent(IN)  :: energies(nsample)
  integer                       :: r
  integer                       :: allocstat, asize, ierr
  double precision, allocatable :: omega(:,:), ovlp(:,:), B(:)
  character(len=c5)             :: fname


#ifdef MPI
  double precision, allocatable :: mpibuf(:,:)
#endif

  !BLAS
  double precision              :: ddot
  external ddot


  allocate(omega(nsample,cpd%rank), stat=allocstat)
  if (allocstat .ne. 0) then
     asize = cpd%rank*nsample*8/1024/1024
     write(message,'(a,i8,a)') "Could not allocate combined basis sampling array, size = " &
          , asize, " MB."
     call cncntrblnk(message)
     routine = "mccpd_initguesscoeff"
     call errormsg
  end if

  allocate(ovlp(cpd%rank,cpd%rank), stat=allocstat)
  if (allocstat .ne. 0) then
     asize = cpd%rank**2*8/1024/1024
     write(message,'(a,i8,a)') "Could not allocate cpd overlap matrix, size = " &
          , asize, " MB."
     call cncntrblnk(message)
     routine = "mccpd_initguesscoeff"
     call errormsg
  end if

  allocate(B(cpd%rank), stat=allocstat)
  if (allocstat .ne. 0) then
     asize = cpd%rank*8/1024/1024
     write(message,'(a,i8,a)') "Could not allocate left-hand-side, size = " &
          , asize, " MB."
     call cncntrblnk(message)
     routine = "mccpd_initguesscoeff"
     call errormsg
  end if


  call mk_coeff_matrices(cpd,ovlp,B,energies,nsample,lcoeffoncuts)



#ifdef MPI


  ! gather partial results in rank 0

  allocate(mpibuf(cpd%rank,cpd%rank))

  call MPI_Reduce(ovlp(:,:),mpibuf,cpd%rank**2,&
       MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)

  if (mpirank.eq.0) then
     ovlp(:,:) = mpibuf
  end if
  deallocate(mpibuf)

  allocate(mpibuf(cpd%rank,1))
  call MPI_Reduce(B,mpibuf,cpd%rank,&
       MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
  if (mpirank.eq.0) then
     B = mpibuf(:,1)
  end if
  deallocate(mpibuf)

#endif


  if (lsymmetries .and. mpirank.eq.0) then
     call sym_coeff_matrices(ovlp,B,cpd%rank)
  end if


  if (invertmethod .eq. mc_inv_spd ) then

     if (mpirank .eq. 0) then
        !$omp parallel do
        do r=1,cpd%rank
           ovlp(r,r) =  ovlp(r,r) * (1.d0 + epsilon_invert)
        end do
        !$omp end parallel do
     end if

     ! ovlp is redistributed in case of scalapack from rank 0
     call mcsolve_spd(ovlp,B,cpd%rank,1,ierr)

  else if (invertmethod .eq. mc_inv_eigen ) then

     fname = "eigenval_coeff"
     call mcsolve_eigen(ovlp, B, epsilon_invert,.false.,.false., &
           cpd%rank, 1, fname, ierr)
  end if


  if (mpirank.eq.0 .and. ierr.ne.0) then
     routine = "mccpd_initguesscoeff"
     message = "Error while solving for coefficients"
     call errormsg
  end if


  if (mpirank.eq.0) then

     if (lsymmetries) then
        call sym_coeff(B,cpd%rank)
     end if

     do r=1,cpd%rank
        cpd%coeff(r) = B(r)
     end do
  end if

  deallocate(ovlp,B)


#ifdef MPI
  call mpi_bcast(cpd%coeff, cpd%rank, MPI_DOUBLE_PRECISION, &
       0,MPI_COMM_WORLD,mpierr)
#endif


end subroutine mccpd_coeff



subroutine mk_coeff_matrices(cpd,ovlp,B,energies,nsample,lcoeffoncuts)
  use cpdtype_m
  use cpdsamplings_m
  use mccuts_m

  implicit none

  type(cpd_t), intent(IN)         :: cpd
  integer, intent(IN)             :: nsample
  logical, intent(IN)             :: lcoeffoncuts
  double precision, intent(IN)    :: energies(nsample)
  double precision, intent(OUT)   :: ovlp(cpd%rank,cpd%rank)
  double precision, intent(OUT)   :: B(cpd%rank)



  double precision, allocatable   :: omega(:,:), Vovlp(:,:)

  integer                         :: m,r,r1,n

  double precision                :: ddot
  external ddot


  allocate(omega(nsample,cpd%rank))


  if (.not. lcoeffoncuts ) then

     m = 0 ! use all modes
     call mk_coeff_omega(omega, m, cpd%rank, nsample, cpd%nmode)

     !$omp parallel do private(r1) schedule(dynamic,1)
     do r=1,cpd%rank
        do r1=r,cpd%rank
           ovlp(r1,r) =  ddot(nsample, omega(:,r1),1,omega(:,r),1)
        end do
     end do
     !$omp end parallel do


     ! complete symmetric part of ovlp
     do r=1,cpd%rank
        do r1=r+1,cpd%rank
           ovlp(r,r1) = ovlp(r1,r)
        end do
     end do

     !$omp parallel do
     do r=1,cpd%rank
        B(r) = ddot(nsample,omega(:,r),1,energies,1)
     end do
     !$omp end parallel do

  else

     allocate(Vovlp(nsample, cpd%rank))

     do m=1,cpd%nmode

        call mk_coeff_omega(omega, m, cpd%rank, nsample, cpd%nmode)


        if (m.eq.1) then
           !$omp parallel do private(r1) schedule(dynamic,1)
           do r=1,cpd%rank
              do r1=r,cpd%rank
                 ovlp(r1,r) =  ddot(nsample, omega(:,r1),1,omega(:,r),1) &
                      *   ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1), 1)
              end do
           end do
           !$omp end parallel do
        else

           !$omp parallel do private(r1) schedule(dynamic,1)
           do r=1,cpd%rank
              do r1=r,cpd%rank
                 ovlp(r1,r) =  ovlp(r1,r) + ddot(nsample, omega(:,r1),1,omega(:,r),1) &
                      *   ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1), 1)
              end do
           end do
           !$omp end parallel do
        end if



        !$omp parallel do private(n)
        do r=1,cpd%rank
           do n=1,nsample
              Vovlp(n,r) = ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1, mccuts(m)%cuts(:,n),1)
           end do
        end do
        !$omp end parallel do


        if (m.eq.1) then

           !$omp parallel do
           do r=1,cpd%rank
              B(r) = ddot(nsample,omega(:,r),1,Vovlp(:,r),1)
           end do
           !$omp end parallel do

        else

           !$omp parallel do
           do r=1,cpd%rank
              B(r) =  B(r) + ddot(nsample,omega(:,r),1,Vovlp(:,r),1)
           end do
           !$omp end parallel do

        end if

     end do

     deallocate(Vovlp)


  end if

  deallocate(omega)

  return

end subroutine mk_coeff_matrices


subroutine mk_coeff_omega(omega, m, rank, nsample, nmode)
  use cpdsamplings_m
  implicit none
  integer, intent(IN)             :: m, rank, nsample, nmode
  double precision, intent(OUT)   :: omega(nsample,rank)
  integer                         :: mz(nmode)
  integer                         :: m1,m2,m3,ms,n,r


  mz = 0
  ! modes in omega:
  ms = 0
  do m1=1,nmode
     if (m.ne.m1) then
        ms = ms + 1
        mz(ms)= m1
     end if
  end do

  if (ms .eq. 1) then

     m1 = mz(1)
     !$omp parallel do private(n)
     do r=1,rank
        do n=1,nsample
           omega(n,r) =  basissamplings(n,m1,r)
        end do
     end do
     !$omp end parallel do

  else

     m1 = mz(1)
     m2 = mz(2)

     !$omp parallel do private(n) schedule(dynamic,1)
     do r=1,rank
        do n=1,nsample
           omega(n,r) =  basissamplings(n,m1,r)*basissamplings(n,m2,r)
        end do
     end do
     !$omp end parallel do

     do m3=3,ms

        m1 = mz(m3)

        !$omp parallel do private(n) schedule(dynamic,1)
        do r=1,rank
           do n=1,nsample
              omega(n,r) = omega(n,r) *  basissamplings(n,m1,r)
           end do
        end do
        !$omp end parallel do

     end do

  end if



end subroutine mk_coeff_omega





subroutine sym_coeff_matrices(ovlp,B,rank)
  use symmops_m
  implicit none
  integer, intent(IN)             :: rank
  double precision, intent(INOUT) :: ovlp(rank,rank),B(rank)

  integer                         :: nsym,rpersym,rest
  integer                         :: n1,n2,r,r1,z,z1
  double precision, allocatable   :: ovlp1(:,:)


  rpersym = rank/nsymmetries

  allocate(ovlp1(rank,rank))

  ovlp1 = ovlp

  do nsym = 2,nsymmetries

     do r=1,rank
        n1 = (r-1)/rpersym + 1
        n2 = sym2map(nsym,n1)


        rest = modulo(r,rpersym)
        if (rest .ne. 0) then
           r1 = (n2-1)*rpersym + rest
        else
           r1 = (n2)*rpersym
        end if

        do z=1,rank
           n1 = (z-1)/rpersym + 1
           n2 = sym2map(nsym,n1)
           rest = modulo(z,rpersym)
           if (rest .ne. 0) then
              z1 = (n2-1)*rpersym + rest
           else
              z1 = (n2)*rpersym
           end if

           ovlp1(r1,z1) = ovlp1(r1,z1) + ovlp(r,z)
        end do
     end do
  end do


  ovlp = ovlp1/nsymmetries
  deallocate(ovlp1)

  call sym_coeff(B,rank)


end subroutine sym_coeff_matrices




subroutine sym_coeff(B,rank)
  use symmops_m
  implicit none
  integer, intent(IN)             :: rank
  double precision, intent(INOUT) :: B(rank)

  double precision, allocatable   :: B1(:)
  integer                         :: nsym,rpersym,r,n1,n2,rest,r1

  rpersym = rank/nsymmetries


  allocate(B1(rank))
  B1 = B

  do nsym = 2,nsymmetries

     do r=1,rank
        n1 = (r-1)/rpersym + 1
        n2 = sym2map(nsym,n1)

        rest = modulo(r,rpersym)
        if (rest .ne. 0) then
           r1 = (n2-1)*rpersym + rest
        else
           r1 = (n2)*rpersym
        end if

        B1(r1) = B1(r1) + B(r)

     end do
  end do

  B = B1/nsymmetries
  deallocate(B1)

end subroutine sym_coeff
