! -*- f90 -*-
! 
!
!


module mccuts_m

  type d1cuts_t
     integer                       :: mgmode
     integer                       :: mgdim
     integer                       :: ncuts
     double precision, allocatable :: cuts(:,:)
  end type d1cuts_t

  type(d1cuts_t), allocatable      :: mccuts(:)
  logical, save                    :: lcutsdone = .false.
  double precision, allocatable    :: intvcuts2(:)

end module mccuts_m

module mccuts_s

    use global
    use channels
    use tensortype
    use mpidata
    use logdat
    use runfit_m
    use cpdtype_m
    use dvrtype_m
    use pesselect_m
    use pftools
    use symmops_m
    use mccpdparams
    use cpdtype_m
    use mcpmpitools
    use mccpdparams
    use mcsampling, only: ntotsample,nidxfile,lsamecoeffspp, &
         mc_task_spp,mc_task_coeff
    use mccuts_m

    use mpidata
#ifdef MPI
    use mpi
#endif

  implicit none

contains

  subroutine getcutfilename(cutdir,m,nfile,fname)
    implicit none

    character(len=*), intent(IN)    :: cutdir
    integer, intent(IN)             :: m,nfile
    character(len=c5), intent(OUT)  :: fname

    integer                         :: ilbl
    character(len=c4)               :: str

    write(str, '(a,i4,a,i4)') "cuts_mode_",m, "_traj_",nfile
    call rmblnk(str)
    ilbl = len_trim(cutdir)
    if (cutdir(ilbl:ilbl) .eq. "/") then
       fname = trim(cutdir)//trim(str)
    else
       fname = trim(cutdir)//"/"//trim(str)
    end if


  end subroutine getcutfilename


  subroutine free_cuts()
    implicit none
    integer   :: m, ncmodes

    if (allocated(mccuts)) then
       ncmodes = size(mccuts)

       do m=1, ncmodes
          if (allocated(mccuts(m)%cuts)) deallocate(mccuts(m)%cuts)
          mccuts(m)%mgmode = 0
          mccuts(m)%mgdim = 0
          mccuts(m)%ncuts = 0
       end do
       deallocate(mccuts)

    end if
    lcutsdone = .false.

    if (allocated(intvcuts2)) then
       deallocate(intvcuts2)
    end if

  end subroutine free_cuts


  subroutine allocate_cuts(ncutmodes,subdim1,ncuts)
    implicit none


    integer, intent(IN)          :: ncuts, ncutmodes
    integer, intent(IN)          :: subdim1(ncutmodes)
    integer                      :: m
    integer                      :: allocstat, ierr
    integer*8                    :: msize
    character(len=256)           :: str

    call free_cuts()

    allocate(intvcuts2(ncutmodes), stat=ierr)
    if (ierr.ne.0) then
       message = "Allocatig intvcuts2 failed"
       routine = "allocate_cuts"
       call errormsg
    end if
    intvcuts2(:) = 0.d0

    allocate(mccuts(ncutmodes), stat=ierr)
    if (ierr.ne.0) then
       message = "Allocatig cuts failed"
       routine = "allocate_cuts"
       call errormsg
    end if

    if (mpirank.eq.0) then
       write(ilog,'(/a)') "Allocating space for 1D cuts:"
    end if

    do m=1,ncutmodes

       msize = subdim1(m)*ncuts

       if (mpirank.eq.0) then
          if (nmpi.gt.1) then
             write(str,'(a,i4,3x,f14.3,a,i14,a,i12,a)') " Mode  ",m, &
                  dble(msize)/131072," MB, ", msize, " points = ", ncuts, " cuts (per MPI instance)"
          else
             write(str,'(a,i4,3x,f14.3,a,i14,a,i12,a)') " Mode  ",m, &
                  dble(msize)/131072," MB, ", msize, " points = ",  ncuts, " cuts"
          end if
          write(ilog,'(1x,a)') trim(str)

       end if

       allocate(mccuts(m)%cuts(subdim1(m),ncuts), stat=allocstat)

       if (allocstat .ne. 0) then
          write(str,'(a,i4)') 'cuts mode ', m
          call cncntrblnk(str)
          routine = "allocate_cuts"
          message = "cannot allocate "//trim(str)
          call errormsg
       end if

       mccuts(m)%mgmode = m
       mccuts(m)%mgdim = subdim1(m)
       mccuts(m)%ncuts = ncuts

       mccuts(m)%cuts(:,:) = 0.d0
    end do

    if (mpirank.eq.0) then
       write(ilog,'(a)') ""
       call myflush(ilog)
    end if


  end subroutine allocate_cuts



  ! Create the one-d cuts
  ! this outer-loops over the files with sampling points
  ! and inner-loops over modes.
  ! cuts are written to files numbered by modes and sampling-files
  ! so that each sampling file and each mode get their own cuts-file
  ! This allows re-using many files when DVRs of single
  ! modes are changed.
  subroutine mk_mccuts(cpd, dvr)
    implicit none
    type (cpd_t), intent(IN)     :: cpd
    type (dvr_t), intent(IN)      :: dvr

    logical                      :: lrecalc, lconsistent
    integer                      :: m, imc, ig, mgdim, n,f
    integer                      :: nsample, nfile
    integer                      :: idx(cpd%ndof)
    integer                      :: didx(cpd%ndof)
    integer                      :: dgdim(cpd%ndof)
    integer                      :: subdim1(cpd%nmode)
    double precision             :: coord(cpd%ndof)
    character(len=c5)            :: str
    character(len=c5)            :: fname
    logical                      :: lexist


    if (.not. cpd%initialized) then
       print*, "Error, calling mk_mccuts without initialized cpd object."
       stop 1
    end if

    if (mpirank.eq.0) then
       write(ilog, '(//a)') "Generating cuts through the PES tensor:"
    end if


    if (lsymmetries) then
       do m=1,cpd%nmode
          subdim1(m)=pointmaps(m)%npoint
       end do
    else
       do m=1,cpd%nmode
          subdim1(m)=cpd%subdim(m)
       end do
    end if

    do nfile = 1, nidxfile(mc_task_spp)

       call loadruntask(mc_task_spp,dvr,nsample,nfile)
       call dvrindex2modeindex(cpd, nsample)
       call allocate_cuts(cpd%nmode,subdim1,nsample)

       do m=1,cpd%nmode

          if (mpirank.eq.0) then
             write(str, '(a,i4,a,i4)') "mode = ",m,"; trajectory = ", nfile
             call cncntrblnk(str)
             write(ilog,'(a)') "  "//trim(str)
             call myflush(ilog)
          end if

          mgdim = subdim1(m)

          do n=1,cpd%nspfdof(m)
             f = cpd%spfdof(n,m)
             dgdim(n) = cpd%gdim(f)
          end do

          if (mpirank.eq.0) then
             call getcutfilename(mccutsdir,m,nfile,fname)

             if (ltryreusecuts) then
                ! we only check the DVR params here.
                ! If the Sampling points match can only be
                ! evaluated later after all files are loaded

                inquire(file=fname, exist=lexist)
                if (.not. lexist) then
                   lrecalc = .true.
                else
                   open(ipcut, file=fname, action="read", status="old", form='unformatted')
                   call chkdvrheader1m(ipcut,m,mgdim,ntotsample(mc_task_spp,nfile),dvr,cpd,lconsistent)
                   close(ipcut)
                   lrecalc  = .not. lconsistent
                end if

             else
                lrecalc = .true.
             end if

             if (.not. lrecalc) then
                write(ilog,'(a)') "  Try re-using cuts on file "//trim(fname)
             else
                write(ilog,'(a)') "* Creating new cuts on file "//trim(fname)
             end if
             call myflush(ilog)
          end if

#ifdef MPI
          call MPI_BCast(lrecalc,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
#endif
          if (.not. lrecalc) cycle


          !$omp parallel do private(ig,didx,idx,coord,n,f) if(lomppotential) schedule(dynamic,5)
          do imc=1,nsample

             do ig=1,mgdim

                if (cpd%subdim(m) .lt. mccuts(m)%mgdim ) then

                   idx = dvrindex(:,imc)
                   call dvrindex2coord(cpd,idx,coord)

                   do n=1,cpd%nspfdof(m)
                      f = cpd%spfdof(n,m)
                      coord(f) =  pointmaps(m)%points(n,ig)
                   end do
                   mccuts(m)%cuts(ig,imc) = potentialQ(coord, cpd%ndof)

                else

                   call cindex2sindex(ig,didx,dgdim,cpd%nspfdof(m))

                   idx = dvrindex(:,imc)

                   do n=1,cpd%nspfdof(m)
                      f = cpd%spfdof(n,m)
                      idx(f) = didx(n)
                   end do

                   mccuts(m)%cuts(ig,imc) = potential(idx, cpd%ndof)
                end if

             end do ! gridpoints

          end do ! 1d cuts
          !$omp end parallel do

!!#ifdef DEBUG
          if (any(mccuts(m)%cuts(:,:) .ne. mccuts(m)%cuts(:,:))) then
             print*, "NaN in cuts, mode", m
             stop 1
          end if
!!#endif


          call savecutfile(cpd,dvr,nfile,m)


       end do ! modes

       call free_cuts()

    end do ! files / trajectories

    lcutsdone = .true.

    if (mpirank.eq.0) then
       write(ilog, '(a/)') "Generating cuts done."
       call myflush(ilog)
    end if


  end subroutine mk_mccuts




  subroutine wrdvrheader1m(channel,m,subdim1,ncuts,dvr,tensor)

    implicit none

    integer, intent(IN)            :: channel, m, subdim1, ncuts
    type(dvr_t), intent(IN)        :: dvr
    class (tensor_t), intent(IN)   :: tensor

    integer                        :: nspfdof1,mbaspar1
    integer, allocatable           :: ildvr(:)

    integer                        :: n,f,log2int


    nspfdof1 = tensor%nspfdof(m)
    mbaspar1 = dvr%mbaspar

    allocate(gdim(nspfdof1))
    allocate(basis(nspfdof1))
    allocate(ildvr(nspfdof1))
    allocate(ipbaspar(mbaspar1,nspfdof1))
    allocate(modelabel(nspfdof1))
    allocate(rpbaspar(mbaspar1,nspfdof1))
    allocate(xend(2,nspfdof1))

    do n=1,nspfdof1
       f = tensor%spfdof(n,m)

       gdim(n) = dvr%gdim(f)
       basis(n) = dvr%basis(f)
       ildvr(n) = log2int(dvr%ldvr(f))
       modelabel(n) = dvr%modelabel(f)
       ipbaspar(:,n) =  dvr%ipbaspar(:,f)
       rpbaspar(:,n) =  dvr%rpbaspar(:,f)
       xend(:,n) = dvr%xend(:,f)

    end do

    write(channel, err=990) ncuts, subdim1, m
    write(channel, err=990) nspfdof1

    write(channel, err=990) (modelabel(f),f=1,nspfdof1)
    write(channel, err=990) gdim
    write(channel, err=990) mbaspar1
    write(channel, err=990) basis
    write(channel, err=990) ildvr

    write(channel, err=990) ipbaspar
    write(channel, err=990) rpbaspar
    write(channel, err=990) xend

    deallocate(gdim,basis,ildvr,ipbaspar,modelabel,rpbaspar,xend)
    return

990 continue

    write(*,*) "ERROR while exporting DVR_def on cut-file"
    stop 1

  end subroutine wrdvrheader1m



  subroutine chkdvrheader1m(channel,m,subdim2,ncuts,dvr,tensor,lconsistent)

    implicit none

    integer, intent(IN)            :: channel, m, subdim2,ncuts
    type(dvr_t), intent(IN)        :: dvr
    class (tensor_t), intent(IN)   :: tensor
    logical, intent(OUT)           :: lconsistent


    integer                        :: nspfdof1,mbaspar1,m1,subdim1, ncuts1
    integer, allocatable           :: gdim1(:),basis1(:), ildvr1(:), ipbaspar1(:,:)
    character(len=16), allocatable :: modelabel1(:)
    double precision, allocatable  :: rpbaspar1(:,:), xend1(:,:)

    integer                        :: n,f,log2int, ierr


    lconsistent = .true.

    read(channel, iostat=ierr) ncuts1, subdim1,m1
    read(channel, err=991) nspfdof1

    lconsistent = lconsistent .and. (m.eq.m1)  &
         .and. (subdim2.eq. subdim1) &
         .and. (tensor%nspfdof(m) .eq. nspfdof1) &
         .and. (ncuts1 .ge. ncuts)



    if (.not. lconsistent) then
       print*, "cut-file header (1) not consistent with system:"
       print*, "m,m1             ", m,m1
       print*, "subdim2,subdim1   ", subdim2,subdim1
       print*, "nspfdof,nspfdof1 ", tensor%nspfdof(m), nspfdof1
       print*, "ncuts, ncuts1    ", ncuts, ncuts1
       return
    end if

    allocate(gdim1(nspfdof1))
    allocate(basis1(nspfdof1))
    allocate(ildvr1(nspfdof1))
    allocate(modelabel1(nspfdof1))
    allocate(xend1(2,nspfdof1))


    read(channel, err=991) (modelabel1(f),f=1,nspfdof1)
    read(channel, err=991) gdim1
    read(channel, err=991) mbaspar1

    if (dvr%mbaspar .lt. mbaspar1) then
       routine = "chkdvrheader1m"
       message = "Inconsistent 'mbaspar' in dvr and cutfile"
       call errormsg
    end if


    allocate(ipbaspar1(mbaspar1,nspfdof1))
    allocate(rpbaspar1(mbaspar1,nspfdof1))

    read(channel, err=991) basis1
    read(channel, err=991) ildvr1
    read(channel, err=991) ipbaspar1
    read(channel, err=991) rpbaspar1
    read(channel, err=991) xend1


    do n=1,nspfdof1
       f = tensor%spfdof(n,m)
       lconsistent = (gdim1(n) .eq. dvr%gdim(f))              .and. lconsistent
       lconsistent = (basis1(n) .eq. dvr%basis(f))            .and. lconsistent
       lconsistent = (ildvr1(n) .eq. log2int(dvr%ldvr(f)))    .and. lconsistent
       lconsistent = (modelabel1(n) .eq. dvr%modelabel(f))    .and. lconsistent
       !lconsistent = all((ipbaspar1(1:,n) .eq. dvr%ipbaspar(:,f)))  .and. lconsistent
       !lconsistent = all((rpbaspar1(1:mbaspar1,n) .eq. dvr%rpbaspar(1:dvr%mbaspar,f)))  .and. lconsistent
       !lconsistent = all((xend1(:,n) .eq. dvr%xend(:,f)))          .and. lconsistent
       if (.not. lconsistent) exit ! loop
    end do


    if (.not. lconsistent) then
       print*, "cut-file header (1) not consistent with system:"
       print*, "m, n, f=spfdof(n,m)             ", m,n,f
       print*, "gdim1(n),   dvr%gdim(f)         ", gdim1(n),dvr%gdim(f)
       print*, "basis1(n), dvr%basis(f)         ", basis1(n), dvr%basis(f)
       print*, "modelabel1(n), dvr%modelabel(f) ", modelabel1(n), dvr%modelabel(f)
    end if


    deallocate(gdim1,basis1,ildvr1,ipbaspar1,modelabel1,rpbaspar1,xend1)

    return

991 continue

    write(*,*) "ERROR while reading DVR_def from cut-file"
    stop 1

  end subroutine chkdvrheader1m




  subroutine savecutfile(cpd,dvr,nfile,m)
    ! this relies on one sampling file loaded only.
    ! see mkcuts where we load one file
    ! with sampling points from the SPP-Task at the time
    ! and get the cuts. Then we save here for each file:

    implicit none


    integer, intent(IN)           :: nfile,m
    type(cpd_t), intent(IN)       :: cpd
    type(dvr_t), intent(IN)       :: dvr

    character(len=c5)             :: fname

    integer                       :: n, mcnmode, ierr, ig
    logical                       :: lexist
    integer                       :: itot,nsample_remote

#ifdef MPI
    double precision, allocatable :: buffer(:,:)
    integer                       :: ntask, istart, iend, nsample
    integer                       :: buffersize, nchunk, ncutschunk, ichunk, restcuts
    integer, parameter            :: targetbuffersize = 100*1024*1024/8 ! send about 100 MB chunks with mpi

    integer :: mpitag
    integer :: mpistat(mpi_status_size)
#endif


    mcnmode = size(mccuts)

    if (mpirank.eq.0) then

       ! start the file
       call getcutfilename(mccutsdir,m,nfile,fname)

       inquire(file=fname, exist=lexist)
       if (lexist .and. .not. lovwr) then
          routine = "savecutfile"
          message = "Unable to overwrite: use -w option. File: "//trim(fname)
          call errormsg
       end if
       open(ipcut, file=fname, iostat=ierr, form='unformatted', status='unknown')
       if (ierr.ne.0) then
          routine = "savecutfile"
          message = "Could not open file for writing: "//trim(fname)
          call errormsg
       end if

       write(ilog,'(a)', advance='no') '  Writing file '//trim(fname)//" ... "
       call myflush(ilog)
       ! haeder
       call wrdvrheader1m(ipcut,m,mccuts(m)%mgdim,ntotsample(mc_task_spp,nfile),dvr,cpd)

       ! data of rank 0 goes first
       itot = 0
       nsample_remote = mpi_chunksizes(0)
       do n=1,nsample_remote
          write(ipcut, iostat=ierr) (mccuts(m)%cuts(ig,n), ig=1,mccuts(m)%mgdim )
          if (ierr.ne.0) then
             routine = "savecutfile"
             message = "Writing failed in file: "//trim(fname)
             call errormsg
          end if
          itot = itot + 1
       end do



#ifdef MPI

       ! get data from other ranks
       do ntask = 1, nmpi-1

          mpitag = ntask
          nsample_remote = mpi_chunksizes(ntask)               ! samples in sending rank
          ncutschunk = max(targetbuffersize/mccuts(m)%mgdim,1) ! no cuts in chunk
          nchunk = nsample_remote/ncutschunk                   ! no full chunks to send
          restcuts = modulo(nsample_remote,ncutschunk)         ! no cuts in last partly filled chunk
          buffersize = mccuts(m)%mgdim*ncutschunk              ! send buffer size

          allocate(buffer(mccuts(m)%mgdim,ncutschunk), stat=ierr)
          if (ierr.ne.0) then
             message = "Faild to allocate read-buffer for cuts"
             routine = "savecutfile"
             call errormsg
          end if


          do ichunk = 1, nchunk

             call mpi_recv(buffer,buffersize,MPI_DOUBLE_PRECISION,ntask, &
                  mpitag,MPI_COMM_WORLD,mpistat,mpierr)

             do n=1, ncutschunk
                write(ipcut, iostat=ierr) (buffer(ig,n), ig=1, mccuts(m)%mgdim)
                if (ierr.ne.0) then
                   routine = "savecutfile"
                   message = "Writing failed on file: "//trim(fname)
                   call errormsg
                end if
                itot = itot + 1
             end do
          end do

          if (restcuts .ne. 0) then
             call mpi_recv(buffer,restcuts*mccuts(m)%mgdim,MPI_DOUBLE_PRECISION,ntask, &
                  mpitag,MPI_COMM_WORLD,mpistat,mpierr)

             do n=1, restcuts
                write(ipcut, iostat=ierr) (buffer(ig,n), ig=1,mccuts(m)%mgdim)
                if (ierr.ne.0) then
                   routine = "savemccuts"
                   message = "Writing failed on file: "//trim(fname)
                   call errormsg
                end if
                itot = itot + 1
             end do

          end if

          deallocate(buffer)

       end do

       if (itot.ne.ntotsample(mc_task_spp,nfile)) then
          routine = "savecutfile"
          message = "Internal error: no cuts written .ne. sampling points"
          write(*,*) mpirank, itot, ntotsample(mc_task_spp,0)
          call errormsg
       end if


#endif
       write(ilog,'(a)') " OK."
       call myflush(ilog)
       close(ipcut)

    end if ! rank 0, receiving


#ifdef MPI

    ! other ranks, sending
    if (mpirank.ge. 1) then


       mpitag = mpirank
       nsample = mpi_chunksizes(mpirank)                    ! samples in sending rank (this one)
       ncutschunk = max(targetbuffersize/mccuts(m)%mgdim,1) ! no cuts in chunk
       nchunk = nsample/ncutschunk                          ! no full chunks to sent
       restcuts = modulo(nsample,ncutschunk)                ! no cuts in last partly filled chunk
       buffersize = mccuts(m)%mgdim*ncutschunk              ! receive buffer size (must match send buffer)

       mpitag = mpirank

       istart = 1
       iend = ncutschunk

       do ichunk = 1, nchunk

          call mpi_send(mccuts(m)%cuts(:,istart:iend),buffersize,MPI_DOUBLE_PRECISION,0, &
               mpitag,MPI_COMM_WORLD,mpierr)

          istart = istart + ncutschunk
          iend = iend + ncutschunk
       end do

       if (restcuts .ne. 0) then

          iend = istart + restcuts - 1
          call mpi_send(mccuts(m)%cuts(:,istart:iend),mccuts(m)%mgdim*restcuts,MPI_DOUBLE_PRECISION,0, &
               mpitag,MPI_COMM_WORLD,mpierr)

       end if

       if (iend .ne. nsample) then
          routine = "save_mccuts"
          message = "Internal error: no cuts sent .ne. sampling points"
          call errormsg
       end if

    end if ! I am sending rank

#endif



  end subroutine savecutfile





  subroutine load_mccuts(ncutmodes,griddim, dvr, tensor)
    ! this loads cuts for all files and all modes

    implicit none

    integer, intent(IN)           :: ncutmodes
    integer, intent(IN)           :: griddim(ncutmodes)
    type(dvr_t), intent(IN)       :: dvr
    class(tensor_t), intent(IN)   :: tensor

    integer                       :: m,nfile
    integer                       :: ncuts1, subdim1,m1, samplesonfile
    integer                       :: ig, ierr, nsample,jmc,nmc, itot
    integer                       :: targetrank

    character(len=c5)             :: fname

    double precision, allocatable :: buffer(:,:)
    logical                       :: lcompatible, lstartup, lnewfile

    !BLAS
    double precision              :: ddot


#ifdef MPI
    integer :: mpitag, buffersize
    integer :: mpistat(mpi_status_size)
#endif


    if (mpirank.eq.0) then

       if (task_loaded.ne. mc_task_spp) then
          ! produce an internal error here (for later coding):
          ! we need the sampling-spp loaded
          ! as we get the number of points and how they are distributed
          ! among mpi ranks from things that are set in 'loadruntask'
          routine = "load_mccuts"
          message = "Internal: runtask SPP must be loaded."
          call errormsg
       end if

       if (.not. all_taskfiles_loaded ) then
          routine = "load_mccuts"
          message = "Internal: all trajectory files for runtask SPP must be loaded."
          call errormsg
       end if

    end if ! rank 0

    nsample = mpi_chunksizes(mpirank)

    ! in each mpi task we use the same number of cuts as sampling points
    call allocate_cuts(ncutmodes,griddim,nsample)


    do m=1,ncutmodes

       if (mpirank.eq.0) then

          ! start the file for a mode
          ! rank 0 is sender
          ! all others are receiver

          jmc = 0          ! sampling point on current target rank
          nmc = 0          ! sampling point on the current source file
          itot = 0         ! curent total sampling point

          nfile = 0        ! number of current source file
          targetrank = 0   ! target mpi rank

          lstartup = .true.
          lnewfile = .true.


          do targetrank=0,nmpi-1

             allocate(buffer(griddim(m), mpi_chunksizes(targetrank)))

             do jmc=1,mpi_chunksizes(targetrank)


                lnewfile = nmc .ge. ntotsample(mc_task_spp,nfile)
                lnewfile = lnewfile .and. itot .lt. ntotsample(mc_task_spp,0)


                if (lnewfile .or. lstartup) then

                   lnewfile = .false.
                   lstartup = .false.

                   nfile = nfile + 1
                   nmc = 0

                   ! start the new file
                   call getcutfilename(mccutsdir,m,nfile,fname)
                   write(ilog,'(a)')  " Loading  file "//trim(fname)//" ... "
                   call myflush(ilog)

                   open(ipcut, file=fname, iostat=ierr, form='unformatted', &
                        status='old', action="read")
                   if (ierr.ne.0) then
                      routine = "load_mccuts"
                      message = "Could not open file for reading: "//trim(fname)
                      call errormsg
                   end if

                   read(ipcut) ncuts1, subdim1,m1
                   rewind(ipcut)


                   if (nidxfile(mc_task_spp) .eq. 0 .and. lsamecoeffspp) then
                      samplesonfile = ntotsample(mc_task_coeff,nfile)
                   else
                      samplesonfile = ntotsample(mc_task_spp,nfile)
                   end if

                   call chkdvrheader1m(ipcut,m,griddim(m),samplesonfile,dvr,tensor,lcompatible)
                   if (.not. lcompatible) then
                      routine = "load_mccuts"
                      message = "Cutfile incompatible with system. Try re-calculating cuts: "//trim(fname)
                      call errormsg
                   end if
                   !read(ipcut, iostat=ierr) (buffer(ig,jmc), ig=1,griddim(m))
                   !print*, "starting new file: m,buf", m, buffer(1,jmc)
                   !backspace(ipcut)

                end if

                read(ipcut, iostat=ierr) (buffer(ig,jmc), ig=1,griddim(m))
                if (ierr.ne.0) then
                   print*, "buffer:", jmc, mpi_chunksizes(targetrank), size(buffer,2), ierr
                   message = "Faild to read cuts from file "//trim(fname)
                   routine = "load_mccuts"
                   call errormsg
                end if

                nmc = nmc + 1
                itot = itot + 1

                ! this is for calculating the functional later
                ! only use the original grid in case of extended
                if (lwrfuncs) then
                   intvcuts2(m) = intvcuts2(m) + ddot(tensor%subdim(m),buffer(1,jmc),1,buffer(1,jmc),1)
                end if


             end do

#ifdef MPI
             if ( targetrank.ne.0) then
                !print*, "sending: m, target, start, end", m,targetrank, buffer(1,1), buffer(1,mpi_chunksizes(targetrank))

                mpitag = m*10000 + targetrank
                buffersize = mpi_chunksizes(targetrank)*griddim(m)
                call mpi_send(buffer,buffersize,MPI_DOUBLE_PRECISION,targetrank, &
                     mpitag,MPI_COMM_WORLD,ierr)
             end if
#endif

             if ( targetrank.eq.0) then
                !print*, "sending: m, target, start, end", m,targetrank, buffer(1,1), buffer(1,mpi_chunksizes(targetrank))
                mccuts(m)%cuts(:,:) =  buffer(:,:)
             end if




             do jmc = 1, mpi_chunksizes(targetrank)
                if (any(mccuts(m)%cuts(:,jmc) .ne. mccuts(m)%cuts(:,jmc))) then
                   print*, "Detected NaN, m,rank,jmc", m, targetrank,jmc
                end if
             end do
             jmc = 0



             deallocate(buffer)
          end do

       end if




       if (allocated(buffer)) deallocate(buffer)


#ifdef MPI

       ! other ranks, receiving
       if (mpirank .ge. 1) then

          mpitag = m*10000 + mpirank
          buffersize =  mpi_chunksizes(mpirank)*griddim(m)
          call mpi_recv(mccuts(m)%cuts(:,:),buffersize,MPI_DOUBLE_PRECISION,0, &
               mpitag,MPI_COMM_WORLD,mpistat,mpierr)
          !print*, "recv: m,me, start, end", m,mpirank, mccuts(m)%cuts(1,1), mccuts(m)%cuts(1,mpi_chunksizes(targetrank))

       end if ! I am a receiving rank

#endif


       if (any(mccuts(m)%cuts(:,:) .ne. mccuts(m)%cuts(:,:))) then
          print*, "NaN in cuts, mode", m
          stop 1
       end if



    end do ! modes

#ifdef MPI
    call mpi_barrier(MPI_COMM_WORLD,mpierr)
#endif

   ! stop
    lcutsdone = .true.

  end subroutine load_mccuts



subroutine spenergiesfromcuts(nsample,nmode,modeindex,energies,lerr)
  implicit none

  integer, intent(IN)           :: nsample,nmode
  integer, intent(IN)           :: modeindex(nsample,nmode)
  double precision, intent(OUT) :: energies(nsample)
  logical, intent(OUT)          :: lerr
  double precision              :: chk(nmode), rdum
  integer                       :: m
  logical                       :: lok, lok1


  integer    :: n,z


  if (.not. lcutsdone) then
     routine = "getspenergies"
     message = "There are no cuts to get samples from. Call before init?"
     call errormsg
  end if

  if (nmode.ne.size(mccuts,1)) then
     routine = "getspenergies"
     message = "Number of modes of cuts does not match number of modes in system."
     call errormsg
  end if


  lok = .true.

  !$omp parallel do private(z,m,chk,lok1,rdum) reduction(.and.: lok)
  do n=1,nsample
     m = 1
     z = modeindex(n,m)

     rdum  = mccuts(m)%cuts(z,n)
     energies(n) = rdum
     chk(m)      = rdum

     do m=2,nmode
        z = modeindex(n,m)
        chk(m) = mccuts(m)%cuts(z,n)
     end do

     lok1 = .true.
     do m=2,nmode
        if ( abs(chk(m-1) - chk(m)) .gt. 1.d-8 ) then
           lok1 = .false.
        end if
     end do
     lok = lok .and. lok1

  end do
  !$omp end parallel do


#ifdef MPI
  call mpi_allreduce(lok, lok1, 1, MPI_LOGICAL,MPI_LAND,MPI_COMM_WORLD,mpierr)
  lok = lok1
#endif

  lerr = .not. lok


end subroutine spenergiesfromcuts


end module mccuts_s


