!   -*- f90 -*-


module mcdensities_m
  use natpottype, only: natpot_t
  use symmops_m, only: pointmaps, lsymmetries

  ! container for densities
  type densities_t
     integer                       :: dmode, mgdim, ndof
     logical                       :: lholdsevecs
     double precision, allocatable :: density(:,:)
     double precision, allocatable :: npop(:)
     integer, allocatable          :: gdim(:),spfdof(:)

  end type densities_t

  type (densities_t), allocatable  :: densities(:)
  type (densities_t), allocatable  :: densities1(:)
  type (densities_t), allocatable  :: densities2(:)
  integer                          :: ndensities

contains

  subroutine free_densities_arr(dens, ndens)
    implicit none
    integer, intent(IN)               :: ndens
    type (densities_t), intent(INOUT) :: dens(ndens)
    integer                           :: m

    ! free memory
    do m=1,ndens
       if (allocated(dens(m)%density)) deallocate(dens(m)%density)
       if (allocated(dens(m)%npop))    deallocate(dens(m)%npop)
       if (allocated(dens(m)%gdim))    deallocate(dens(m)%gdim)
       if (allocated(dens(m)%spfdof))  deallocate(dens(m)%spfdof)
       dens(m)%lholdsevecs = .false.
       dens(m)%dmode = 0
       dens(m)%mgdim = 0
       dens(m)%ndof = 0
    end do

  end subroutine free_densities_arr



  subroutine alloc_densities_arr(natpot,dens,ndens, allocstat)
    implicit none
    class(natpot_t), intent(IN)       :: natpot
    integer, intent(IN)               :: ndens
    type (densities_t), intent(INOUT) :: dens(ndens)

    integer, intent(OUT)              :: allocstat
    integer                           :: m,n,i,f,mgdim

    call free_densities_arr(dens, ndens)

    allocstat = 0
    i= 0
    do m=1,natpot%nmode

       if (m.ne.natpot%modc) then
          i = i + 1
          if (i.gt.ndens) then
             allocstat = i
             write(*,*) "Internal error, not enough densities allocated"
             stop 1
          end if

          if (lsymmetries) then
             ! we may have an extended grid
             mgdim = pointmaps(m)%npoint
          else
             mgdim = natpot%subdim(m)
          end if


          dens(i)%dmode = m
          dens(i)%mgdim = mgdim

          allocate(dens(i)%density(mgdim,mgdim), stat=allocstat)

          if (allocstat .ne. 0) then
             write(*,*) "Error: could not allocate density matrix for mode", m
             stop 1
          end if

          dens(i)%density = 0.d0

          allocate(dens(i)%npop(mgdim), stat=allocstat)

          if (allocstat .ne. 0) then
             write(*,*) "Error: could non allocate npop array for mode ", m
             stop 1
          end if


          dens(i)%npop = 0.d0

          dens(i)%ndof = natpot%nspfdof(m)
          allocate(dens(i)%gdim(dens(i)%ndof))
          allocate(dens(i)%spfdof(dens(i)%ndof))

          do n=1,natpot%nspfdof(m)
             f = natpot%spfdof(n,m)
             dens(i)%gdim(n) = natpot%gdim(f)
             dens(i)%spfdof(n) = f
          end do

       end if

    end do

  end subroutine alloc_densities_arr

end module mcdensities_m

module mcdensities_s

    use mpidata
    use channels
    use maxv
    use runpfitmod
    use global
    use pglobal
    use logdat
    use diagrsmod
    use mcdensities_m
    use natpottype
    use pesselect_m
    use mcpfitparams
    use mcsampling
    use symmops_m
    use pftools
    use omp_lib
    use mpidata
    use mcpmpitools
    use mccomplete
#ifdef MPI
    use mpi
#endif

  implicit none
  private
  public :: calcspp
  public :: densities_t

contains

  ! main driver routine
  subroutine calcspp(natpot,dvrindex,nsample)
    implicit none

    type(natpot_t),intent(INOUT)   :: natpot
    integer, intent(IN)            :: nsample, dvrindex(:,:)

    double precision               :: sumnw, v

    integer                        :: idx(natpot%ndof)
    integer                        :: m,allocstat
    logical                        :: lloadedevecs
    logical                        :: lalloctmp

    !trigger potential initialization

    idx = natpot%gdim/2
    v = potential(idx,natpot%ndof)

    lalloctmp = (smethod(mc_task_spp,1) .ne. mc_complete) &
         .and. .not. lloadevecs


    sumnw=0.d0
    if (mpirank.eq.0) then
       write(ilog,'(a)')
       write(ilog,'(a)') "Calculating SPPs (see output): "
       call myflush(ilog)
    end if

    ndensities = 0
    ! allocate density arrays
    do m=1,natpot%nmode
       if (natpot%modc.ne.m) ndensities = ndensities + 1
    end do
    allocate(densities(ndensities))
    if (lalloctmp) then
       allocate(densities1(ndensities))
       allocate(densities2(ndensities))
    end if

    call alloc_densities_arr(natpot,densities, ndensities, allocstat)
    if ( lalloctmp ) then
       call alloc_densities_arr(natpot,densities1,ndensities, allocstat)
       call alloc_densities_arr(natpot,densities2,ndensities, allocstat)
    end if

    if (allocstat .ne. 0) then
       routine = "calcspp"
       if (allocstat .eq. natpot%nmode+1) then
          message = "Could not allocate conainer for densities."
       else
          write(message,'(a,i4)') &
               "Could not allocate densities for mode:", allocstat
       end if
       call errormsg
    end if


    if (.not. lloadevecs) then

       if (mpirank.eq. 0) then
          write(ilog,*)
          write(ilog,'(a)') "Calculating densities..."
          call myflush(ilog)
       end if




       do m=1,natpot%nmode

          if (m.ne. natpot%modc) then

             if (mpirank.eq. 0) then
                if (lsymmetries) then
                   write(ilog,'(2(a,i4))') ' mode:', &
                        m, ", points:", pointmaps(m)%npoint
                else
                   write(ilog,'(2(a,i4))') ' mode:', &
                        m, ", points:", natpot%subdim(m)
                end if
                call myflush(ilog)
             end if

             if( smethod(mc_task_spp,1) .eq. mc_complete) then
                if (mpirank .eq. 0) then
                   call mkdensity1mComplete(natpot,m)
                end if
             else
                if (lomegaspp) then
                   call mkdensity1mmem(natpot,m,dvrindex,nsample)
                else
                   call mkdensity1m(natpot,m,dvrindex,nsample)
                end if
             end if

          end if
       end do

       if (mpirank.eq. 0) then
          write(ilog,*)
          write(ilog,'(a)') "Done. Diagonalizing..."
          write(ilog,*)
          call myflush(ilog)
       end if

       if (mpirank.eq. 0) then
          if (nsymmetries .gt. 1 .and. &
               smethod(mc_task_spp,1) .ne. mc_complete ) then
             ! for complete the densities are symmetric automatically
             call symdensities(natpot, densities, ndensities)
             call symdensities(natpot, densities1, ndensities)
             call symdensities(natpot, densities2, ndensities)
          end if

          if (lalloctmp)  call chkevecs(natpot)

       end if

       ! free unneeded memory
       if (lalloctmp) then
          call free_densities_arr(densities1,ndensities)
          call free_densities_arr(densities2,ndensities)
          deallocate(densities1)
          deallocate(densities2)
       end if

       call savedensities(natpot)

    else

       call loaddensities()
       lloadedevecs = .false.
       do m=1,ndensities
          lloadedevecs = lloadedevecs .or. densities(m)%lholdsevecs
       end do
    end if


  ! once we have the SPP we can calculate the sigmas


    call addspps(natpot, sumnw)

    if (.not. lloadedevecs .and. lsaveevecs) then
       ! this now saves evecs
       call savedensities(natpot)
    end if

    call mksigma(natpot,densities,ndensities)


    ! do some final logging
    if (mpirank.eq.0) then
       sumnw=dsqrt(sumnw) ! eV
       call logsumnw(sumnw, iout)
       write(ilog,'(a)')
       call myflush(ilog)
    end if

    ! free memory

    call free_densities_arr(densities,ndensities)
    deallocate(densities)


  end subroutine calcspp




  subroutine addspps(natpot,sumnw)

    implicit none

    type (natpot_t), intent(INOUT)       :: natpot
    double precision, intent(INOUT)      :: sumnw

    integer                              :: m,z1,z2,z3,n,n1,f,ig
    integer                              :: ioformat
    integer                              :: dmode, mgdim
    double precision                     :: relnw, trace
    character(len=80)                    :: buf, fmat
    character(len=4)                     :: smode
    integer                              :: ilbl,i
    character(len=c5)                    :: fname
    integer                              :: gdim(natpot%ndof), idx(natpot%ndof)
    double precision                     :: coord(natpot%ndof)


    double precision, allocatable        :: evecs(:,:), npop(:), mat(:,:)
    logical                              :: lexists

#ifdef MPI
    integer :: ierr
#endif

    if (mpirank.eq.0) then

       do m=1,ndensities

          dmode = densities(m)%dmode
          mgdim = densities(m)%mgdim

          ! some output to the logs
          call getModeLabelsStr(natpot,dmode,buf)
          write(smode,'(i4)') dmode
          smode = adjustl(smode)
          ilbl = len_trim(buf) + 2
          do i=ilbl,80
             buf(i:i) = "*"
          end do
          write(iout,'(a/)') "************* Mode: "//smode//" "//buf(1:46)
          call myflush(iout)

          if (.not. densities(m)%lholdsevecs) then

             gdim = 0
             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                gdim(n) = natpot%gdim(f)
             end do

             write(fmat,'(a,i4,a)') "(", natpot%nspfdof(dmode), 'f8.3,f16.8)'
             call rmblnk(fmat)

             ! allocae temp mem
             allocate(evecs(mgdim,mgdim))
             allocate(npop(mgdim))
             allocate(mat(mgdim,mgdim))

             mat = densities(m)%density

             ioformat = 0

             call diagrs(npop,evecs,mat, &
                  mgdim,natpot%potdim(dmode), &
                  relnw,ioformat)

             sumnw = sumnw + relnw

             densities(m)%density = evecs
             densities(m)%npop = npop
             densities(m)%lholdsevecs = .true.

             do n=1,natpot%potdim(dmode)
                z1 = natpot%zevec(dmode,n)
                z2 = z1 + natpot%subdim(dmode) - 1
                z3 = natpot%subdim(dmode)
                natpot%evec(z1:z2) = evecs(1:z3,n)
             end do

             !check for degeneracies at cut-off
             n = natpot%potdim(dmode)
             if (n.lt.natpot%subdim(dmode)) then
                if ( abs(npop(n+1)/npop(n)-1.d0) .le. 1.d-8 ) then
                   routine = "addspp"
                   write(message,'(a,i4,a)') "It seems only one " &
                        //"of two degenerate SPPs are" &
                        //" taken into account for mode ", dmode, &
                        ". This can leed to a loss of symmetry." &
                        // " Check output and sym.log files."
                   call warnmsg

                   ! we let it run anyways.
                   lwarn = .false.
                end if
             end if


             do n=1,natpot%potdim(dmode)

                write(buf,'(i4,a,i4)') dmode,"_", n
                call rmblnk(buf)
                fname = trim(name)//"/evec_m"//trim(buf)
                inquire(file=trim(fname), EXIST=lexists)
                if (lexists.and..not.lovwr) then
                   routine = "addspps (mcdensities.F90)"
                   message = "File exists, cannot overwrite. Use -w option: "//trim(fname)
                   call errormsg
                end if

                open(idump, file= trim(name)//"/evec_m"//trim(buf), form="formatted")
                call getModeLabelsStr(natpot,dmode,buf)
                buf = "# "//trim(buf)// " SPP"
                write(idump,'(a)') trim(buf)
                do ig=1, mgdim
                   call cindex2sindex(ig,idx,gdim,natpot%nspfdof(dmode))
                   do n1=1,natpot%nspfdof(dmode)
                      coord(n1) = pointmaps(dmode)%points(n1,ig)
                   end do
                   write(idump,fmat)(coord(n1),n1=1,natpot%nspfdof(dmode)), evecs(ig,n)
                end do
                close(idump)

             end do

             deallocate(evecs, mat, npop)

          else

             do n=1,natpot%potdim(dmode)
                z1 = natpot%zevec(dmode,n)
                z2 = z1 + natpot%subdim(dmode) - 1
                z3 = natpot%subdim(dmode)
                natpot%evec(z1:z2) = densities(m)%density(1:z3,n)
             end do

             trace = sum(densities(m)%npop)

             write(iout,'(a)') " SPP and npop read from file:"
             write (iout,'(a,1pe11.4,a,/a,f14.4,a)') &
                  ' Trace of reduced density matrix : ',trace,' au', &
                  ' Reduced trace :', trace*ev*ev/swsum2, ' eV**2'

             ioformat = 0
             call wrdiagrsout(ioformat,mgdim, densities(m)%npop,  &
                  natpot%potdim(dmode), relnw, trace)
             sumnw = sumnw + relnw

             n = natpot%potdim(dmode)
             if (n.lt.natpot%subdim(dmode)) then
                if ( abs(densities(m)%npop(n+1)/densities(m)%npop(n)-1.d0) .le. 1.d-8 ) then
                   routine = "addspp"
                   write(message,'(a,i4,a)') "It seems only one of two degenerate SPPs are" &
                        //" taken into account for mode ", dmode, &
                        ". This can leed to a loss of symmetry." &
                        // " Check output and sym.log files."
                   call warnmsg

                   ! we let it run anyways.
                   lwarn = .false.
                end if
             end if

          end if !loeadevecs

       end do ! modes

    end if ! mpi rank 0

#ifdef MPI

    call MPI_Bcast(natpot%evec,natpot%evecdim, &
         MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)

#endif

  end subroutine addspps





  subroutine logsumnw(sumnw, channel)
    implicit none
    double precision, intent(IN) :: sumnw
    integer, intent(IN)          :: channel
    logical                      :: isopen

    inquire(channel, opened=isopen)
    if (.not. isopen) return

    write(channel,'(/2a)') ' Global (weighted) L^2 error estimated' &
         //' from neglected natural weights:'
    if( sumnw .gt. 0.999d0 ) then
       write(channel,'(1x,f14.4,a,1pe16.4,a,1pe16.4,a/)') &
            sumnw,' eV', sumnw*8065.54353d0, 'cm^-1', &
            sumnw/ev,' a.u.'
    else
       write(channel,'(1x,f14.4,a,f16.3,a,1pe16.4,a/)') &
            sumnw*1.d3,' meV', sumnw*8065.54353d0, ' cm^-1', &
            sumnw/ev,' a.u.'
    endif

    call myflush(channel)

  end subroutine logsumnw



  subroutine mkdensity1mmem(natpot,dmode,dvrindex,nsample)

    implicit none

    type (natpot_t), intent(INOUT)  :: natpot
    integer, intent(IN)             :: dmode,nsample
    integer, intent(IN)             :: dvrindex(natpot%ndof,nsample)
    integer                         :: idx(natpot%ndof)
    double precision                :: coord(natpot%ndof)
    double precision                :: fac
    integer                         :: imc,ig,ig1,n,f,i,i1,i2,m,idens
    integer                         :: mgdim
    integer                         :: nsample1,nsample2,ntotsample1
    integer                         :: didx(natpot%nspfdof(dmode))
    integer                         :: dgdim(natpot%nspfdof(dmode))
    double precision, allocatable   :: V1(:,:), V2(:,:)
    integer                         :: allocstat

    ! BLAS
    double precision                :: ddot

#ifdef MPI
    integer                         :: ierr
    double precision, allocatable   :: tmpdensity(:,:)
#endif

    routine = "mkdensity1mmem"

    if (dmode .eq. natpot%modc) then
       message = "Called addspp1m for contracted mode. Aborting."
       call errormsg
    end if
    if (dmode .gt. natpot%nmode) then
       message = "Called addspp1m for non-existing mode. Aborting."
       call errormsg
    end if

    ! associate pointers
    i = 0
    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then
          i = i+1
          if (densities(i)%dmode .eq. dmode) then
             mgdim = densities(i)%mgdim
             dgdim = densities(i)%gdim
             idens = i

             exit ! loop
          end if
       end if
    end do

    nsample2 = nsample/2
    nsample1 = nsample-nsample2
    ntotsample1 = ntotsample(mc_task_spp,0)

    ! odd
    allocate(V1(nsample1,mgdim), stat=allocstat)
    if (allocstat.ne.0) then
       routine =  "mkdensity1mmem"
       message = "Unable to allocate V(nsamle,mode_gdim),"// &
            " try setting 'no-omega-spp' in RUN-SECTION"
       call errormsg
    end if

    ! even
    allocate(V2(nsample2,mgdim), stat=allocstat)
    if (allocstat.ne.0) then
       routine =  "mkdensity1mmem"
       message = "Unable to allocate V(nsamle,mode_gdim),"// &
            " try setting 'no-omega-spp' in RUN-SECTION"
       call errormsg
    end if


    if (mgdim .eq. natpot%subdim(dmode)) then
       ! density matrix is on primitive grid

       !$omp parallel do private(imc,didx,idx,n,f,i1,i2) &
       !$omp schedule(dynamic,1) if(lomppotential)
       do ig=1,mgdim

          ! odd
          do imc=1,nsample1

             call cindex2sindex(ig,didx,dgdim,natpot%nspfdof(dmode))

             i1 = imc*2-1
             idx = dvrindex(:,i1)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                idx(f) = didx(n)
             end do
             V1(imc,ig) = potential(idx, natpot%ndof)
          end do

          ! even
          do imc=1,nsample2

             call cindex2sindex(ig,didx,dgdim,natpot%nspfdof(dmode))

             i2 = imc*2
             idx = dvrindex(:,i2)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                idx(f) = didx(n)
             end do
             V2(imc,ig) = potential(idx, natpot%ndof)
          end do

       end do
       !$omp end parallel do

    else
       ! density matrix is on extended primitive grid

       !$omp parallel do private(imc,idx,coord,n,f,i1,i2) &
       !$omp schedule(dynamic,1) if(lomppotential)
       do ig=1,mgdim

          ! odd
          do imc=1,nsample1

             i1 = imc*2-1
             idx = dvrindex(:,i1)

             call dvrindex2coord(natpot,idx,coord)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                coord(f) =  pointmaps(m)%points(n,ig)
             end do

             V1(imc,ig) = potentialQ(coord, natpot%ndof)

          end do

          ! even
          do imc=1,nsample2

             i2 = imc*2-1
             idx = dvrindex(:,i2)

             call dvrindex2coord(natpot,idx,coord)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                coord(f) =  pointmaps(m)%points(n,ig)
             end do

             V2(imc,ig) = potentialQ(coord, natpot%ndof)

          end do

       end do
       !$omp end parallel do

    end if


    fac = natpot%rvpotdim/ntotsample1/mgdim


    ! odd
    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig,mgdim
          densities1(idens)%density(ig1,ig) = ddot(nsample1,V1(:,ig1),1,V1(:,ig),1)*fac
       end do
    end do
    !$omp end parallel do

    ! even
    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig,mgdim
          densities2(idens)%density(ig1,ig) = ddot(nsample2,V2(:,ig1),1,V2(:,ig),1)*fac
       end do
    end do
    !$omp end parallel do


    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig+1,mgdim
          densities1(idens)%density(ig,ig1) = &
               densities1(idens)%density(ig1,ig)
       end do
    end do
    !$omp end parallel do


    !$omp parallel do private(ig1) schedule(dynamic)
    do ig=1,mgdim
       do ig1=ig+1,mgdim
          densities2(idens)%density(ig,ig1) = &
               densities2(idens)%density(ig1,ig)
       end do
    end do
    !$omp end parallel do

    ! free space
    deallocate(V1,V2)

#ifdef MPI

    allocate(tmpdensity(mgdim,mgdim))

    call MPI_Reduce(densities2(idens)%density, &
         tmpdensity,mgdim**2,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq. 0) then
       densities2(idens)%density = tmpdensity
    end if

    call MPI_Reduce(densities1(idens)%density, &
         tmpdensity,mgdim**2,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq. 0) then
       densities1(idens)%density = tmpdensity
    end if

    deallocate(tmpdensity)

#endif

    if (mpirank.eq.0) then
       !$omp parallel do private(ig) schedule(static)
       do ig1=1,mgdim
          do ig=1,mgdim
             densities(idens)%density(ig,ig1) = &
                  densities1(idens)%density(ig,ig1) &
                  + densities2(idens)%density(ig,ig1)
          end do
       end do
       !$omp end parallel do
    end if


  end subroutine mkdensity1mmem




  subroutine mkdensity1m(natpot,dmode,dvrindex,nsample)

    implicit none

    type (natpot_t), intent(INOUT)  :: natpot
    integer, intent(IN)             :: dmode,nsample
    integer, intent(IN)             :: dvrindex(natpot%ndof,nsample)
    integer                         :: idx(natpot%ndof)
    double precision                :: coord(natpot%ndof)
    double precision                :: fac
    integer                         :: imc,ig,ig1,n,f,i,m,idens
    integer                         :: mgdim
    integer                         :: nsample1, ntotsample1
    integer                         :: didx(natpot%nspfdof(dmode))
    integer                         :: dgdim(natpot%nspfdof(dmode))
    double precision, allocatable   :: V(:)
    integer                         :: allocstat

#ifdef MPI
    integer                         :: ierr
    double precision, allocatable   :: tmpdensity(:,:)
#endif

    routine = "mkdensity1m"

    if (dmode .eq. natpot%modc) then
       message = "Called addspp1m for contracted mode. Aborting."
       call errormsg
    end if
    if (dmode .gt. natpot%nmode) then
       message = "Called addspp1m for non-existing mode. Aborting."
       call errormsg
    end if

    ! associate pointers
    i = 0
    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then
          i = i+1
          if (densities(i)%dmode .eq. dmode) then
             mgdim = densities(i)%mgdim
             dgdim = densities(i)%gdim
             idens = i

             exit ! loop
          end if
       end if
    end do

    nsample1 = nsample
    ntotsample1 = ntotsample(mc_task_spp,0)

    allocate(V(mgdim), stat=allocstat)
    if (allocstat.ne.0) then
       routine =  "mkdensity1m"
       message = "Unable to allocate V(nsamle,mode_gdim)"
       call errormsg
    end if


    if (mgdim .eq. natpot%subdim(dmode)) then
       ! density matrix is on primitive grid


       do imc=1,nsample

          !$omp parallel do private(didx,idx,n,f) &
          !$omp if(lomppotential)
          do ig=1,mgdim

             call cindex2sindex(ig,didx,dgdim,natpot%nspfdof(dmode))

             idx = dvrindex(:,imc)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                idx(f) = didx(n)
             end do
             v(ig) = potential(idx, natpot%ndof)
          end do
          !$omp end parallel do

          !$omp parallel do private(ig1) schedule(dynamic,1)
          do ig=1,mgdim
             do ig1=ig,mgdim
                if (modulo(imc,2) .eq. 0) then
                   ! even
                   densities2(idens)%density(ig1,ig) =  &
                        densities2(idens)%density(ig1,ig) &
                        + v(ig1)*v(ig)
                else
                   ! odd
                   densities1(idens)%density(ig1,ig) =  &
                        densities1(idens)%density(ig1,ig) &
                        + v(ig1)*v(ig)
                end if
             end do
          end do
          !$omp end parallel do

       end do
    else
       ! density matrix is on extended primitive grid

       do imc=1,nsample

          !$omp parallel do private(idx,coord,n,f) &
          !$omp if(lomppotential)
          do ig=1,mgdim

             idx = dvrindex(:,imc)
             call dvrindex2coord(natpot,idx,coord)

             do n=1,natpot%nspfdof(dmode)
                f = natpot%spfdof(n,dmode)
                coord(f) =  pointmaps(m)%points(n,ig)
             end do

             v(ig) = potentialQ(coord, natpot%ndof)

          end do
          !$omp end parallel do

          !$omp parallel do private(ig1) schedule(dynamic,1)
          do ig=1,mgdim
             do ig1=ig,mgdim
                if (modulo(imc,2) .eq. 0) then
                   ! even
                   densities2(idens)%density(ig1,ig) =  &
                        densities2(idens)%density(ig1,ig) &
                        + v(ig1)*v(ig)
                else
                   ! odd
                   densities1(idens)%density(ig1,ig) =  &
                        densities1(idens)%density(ig1,ig) &
                        + v(ig1)*v(ig)
                end if
             end do
          end do
          !$omp end parallel do

       end do

    end if

    fac = natpot%rvpotdim/ntotsample1/mgdim

    ! EVEN SAMPLING POINTS

    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig,mgdim
          densities2(idens)%density(ig1,ig) =  &
               densities2(idens)%density(ig1,ig) *fac
       end do
    end do
    !$omp end parallel do

    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig+1,mgdim
          densities2(idens)%density(ig,ig1) = &
               densities2(idens)%density(ig1,ig)
       end do
    end do
    !$omp end parallel do


    ! ODD SAMPLING POINTS

    !$omp parallel do private(ig1) schedule(dynamic,1)
    do ig=1,mgdim
       do ig1=ig,mgdim
          densities1(idens)%density(ig1,ig) = &
               densities1(idens)%density(ig1,ig)*fac
       end do
    end do
    !$omp end paralleldo

    !$omp parallel do private(ig1) schedule(dynamic)
    do ig=1,mgdim
       do ig1=ig+1,mgdim
          densities1(idens)%density(ig,ig1) = &
               densities1(idens)%density(ig1,ig)
       end do
    end do
    !$omp end parallel do

    ! free space
    deallocate(V)

#ifdef MPI

    allocate(tmpdensity(mgdim,mgdim))

    call MPI_Reduce(densities2(idens)%density, &
         tmpdensity,mgdim**2,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq. 0) then
       densities2(idens)%density = tmpdensity
    end if

    call MPI_Reduce(densities1(idens)%density, &
         tmpdensity,mgdim**2,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq. 0) then
       densities1(idens)%density = tmpdensity
    end if

    deallocate(tmpdensity)

#endif

    if (mpirank.eq.0) then
       !$omp parallel do private(ig) schedule(static)
       do ig1=1,mgdim
          do ig=1,mgdim
             densities(idens)%density(ig,ig1) = &
                  densities1(idens)%density(ig,ig1) &
                  + densities2(idens)%density(ig,ig1)
          end do
       end do
       !$omp end parallel do
    end if


  end subroutine mkdensity1m







  subroutine mkdensity1mComplete(natpot,dmode)

    type (natpot_t), intent(INOUT) :: natpot
    integer, intent(IN)            :: dmode

    integer                        :: i,j,m
    integer                        :: mgdim,vdim,ndim, idens
    integer                        :: dgdim(natpot%nspfdof(dmode))
    double precision               :: vij


    i = 0
    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then
          i = i+1
          if (densities(i)%dmode .eq. dmode) then
             mgdim = densities(i)%mgdim
             dgdim = densities(i)%gdim
             idens = i
             exit !loop
          end if
       end if
    end do

    vdim = 1
    do m=1,dmode-1
       vdim=vdim*natpot%subdim(m)
    end do
    ndim = 1
    do m=dmode+1, natpot%nmode
       ndim=ndim*natpot%subdim(m)
    end do

    ! this is not optimal, but OK...
    !$omp parallel do private(i,vij) schedule(dynamic)
    do j=1,mgdim
       do i=j,mgdim
          call vdensityij(vpot,i,j,vij,vdim,mgdim,ndim)
          densities(idens)%density(i,j) = vij
       end do
    end do
    !$omp end parallel do

    !$omp parallel do private(i) schedule(dynamic)
    do j=1,mgdim
       do i=j,mgdim
          densities(idens)%density(j,i) = densities(idens)%density(i,j)
       end do
    end do
    !$omp end parallel do


  end subroutine mkdensity1mComplete


  subroutine vdensityij(vpot,i,j,vij,vdim,mgdim,ndim)
    implicit none
    integer, intent(IN)          :: i,j,vdim,mgdim,ndim
    double precision, intent(IN) :: vpot(vdim,mgdim,ndim)
    double precision, intent(OUT):: vij

    integer                      :: iv,in

    vij = 0.d0
    do in=1,ndim
       do iv=1,vdim
          vij = vij + vpot(iv,i,in)*vpot(iv,j,in)
       end do
    end do

  end subroutine vdensityij



  subroutine SaveDensities(natpot)
    implicit none

    type(natpot_t), intent(IN)            :: natpot
    character(len=256)                    :: fname
    character(len=8)                      :: buf
    integer                               :: m

    if (mpirank.ne.0) return

    do m=1,ndensities
       write(buf, '(i4)') densities(m)%dmode
       buf = adjustl(buf)
       if (densities(m)%lholdsevecs) then
          fname = trim(name)//"/evecs_mode_"//trim(buf)
       else
          fname = trim(name)//"/density_mode_"//trim(buf)
       end if
       call SaveDensity(densities(m),natpot,fname,lasciidensity)
    end do

  end subroutine SaveDensities



  subroutine SaveDensity(density,natpot,fname,lascii)
    implicit none

    type(natpot_t), intent(IN)       :: natpot
    character(len=*), intent(IN)     :: fname
    type (densities_t), intent(IN)   :: density
    logical,intent(IN),optional      :: lascii

    logical                          :: lascii1
    character(len=8)                 :: stat
    character(len=32)                :: fmat
    character(len=80)                :: buf
    character(len=100)               :: buf1,buffer
    integer                          :: i,j,jlbl, mgdim,m

    lascii1 = .false.
    if (present(lascii)) lascii1 = lascii

    mgdim = density%mgdim
    m = density%dmode

    buf = " "
    call getModeLabelsStr(natpot,m,buf)
    buf = adjustl(buf)
    buf1 = " "
    call fpstartup(buf1)
    buf1 = adjustl(buf1)
    jlbl = len_trim(buf1)
    buf1(jlbl:jlbl) = " "

    if (lovwr) then
       stat = "unknown"
    else
       stat = "new"
    end if

    if (lascii1) then
       write(fmat,'(i8)') mgdim
       fmat = "("//trim(adjustl(fmat))//"E24.16)"

       open(idump,file=fname,status=stat,form="formatted",err=12)
       write(idump,'(a)')    "# ASCII ASCII ASCII ASCII ASCII "
       write(idump,'(a,i8)') "# Mode: ", m
       write(idump,'(a)')    "# Coordinates: "//trim(buf)
       write(idump,'(a)')    "# "//trim(buf1)
       write(idump,'(a,i8)') "# gdim: ", mgdim
       if (density%lholdsevecs) then
          write(idump,'(a)') "# eigenvectors (SPP) and npop"
       else
          write(idump,'(a)') "# density"
       end if

       do i=1,mgdim
          write(idump,err=13,fmt=fmat) (density%density(j,i), j=1,mgdim)
       end do


       if (density%lholdsevecs) then
          write(idump,err=13,fmt=fmat) (density%npop(j), j=1,mgdim)
       end if

    else
       open(idump,file=fname,status=stat,form="unformatted",err=12)

       buffer =  "# BINARY BINARY BINARY BINARY BINARY"
       write(idump) buffer

       write(buffer,'(a,i8)') "# Mode ", m
       write(idump) buffer

       buffer = "# Coordinates: "//trim(buf)
       write(idump) buffer

       buffer = "# "//trim(buf1)
       write(idump) buffer

       write(buffer,'(a,i8)') "# gdim: ", mgdim
       write(idump) buffer

       if (density%lholdsevecs) then
          buffer = "# eigenvectors (SPP) and npop"
       else
          buffer = "# density"
       end if
       write(idump) buffer

       do i=1,mgdim
          write(idump,err=13) (density%density(j,i), j=1,mgdim)
       end do
       if (density%lholdsevecs) then
          write(idump,err=13) (density%npop(j), j=1,mgdim)
       end if

    end if

    close(idump)

    return

12  continue
    routine = "SaveDensity"
    message = "Unable to open file "//trim(fname)
    call warnmsg
    lwarn = .false.
    return
13  continue
    close(idump)
    routine = "SaveDensity"
    message = "Unable to write file "//trim(fname)
    call warnmsg
    lwarn = .false.
    return

  end subroutine SaveDensity




  subroutine LoadDensities()
    implicit none
    character(len=256)                    :: fname
    character(len=8)                      :: buf
    integer                               :: m, ifail
    logical                               :: lexists

    do m=1,ndensities
       if (mpirank.eq.0) then
          write(buf, '(i4)') densities(m)%dmode
          buf = adjustl(buf)

          fname = trim(name)//"/evecs_mode_"//trim(buf)
          inquire(file=fname,exist=lexists)
          if (.not. lexists) then
             fname = trim(name)//"/density_mode_"//trim(buf)
          end if

          call LoadDensity(densities(m),fname,ifail)
          if (ifail.ne.0) then
             routine = "LoadDensities"
             message = "Failed to load file "//trim(fname)
             call errormsg
          end if
       end if

#ifdef MPI

       call MPI_Bcast(densities(m)%density,densities(m)%mgdim**2, &
            MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ifail)

#endif


    end do


  end subroutine LoadDensities


  subroutine LoadDensity(density,fname,ifail)
    implicit none

    integer, intent(OUT)                   :: ifail
    type (densities_t), intent(INOUT)      :: density
    character(len=*), intent(IN)           :: fname
    integer                                :: i,j, gdim1, mgdim
    logical                                :: lascii, lnext, lheader, levecs
    character(len=100)                     :: buf, coord

    ifail = 0
    lnext=.true.
    mgdim = density%mgdim

     ! read the header, try unformatted
    open(idump,file=fname,status="old", form="unformatted",err=22)
    read(idump,err=21) buf

    ! see if it worked. if not, try formatted read
    if (index(buf, "BINARY") .gt. 0) then
       lnext=.false.
       lascii = .false.
    else if (index(buf,"ASCII") .gt.0) then
       lnext=.false.
       lascii = .true.
    end if


21  continue
    close(idump)

    if (lnext) then
       open(idump,file=fname,status="old", form="formatted")
       read(idump) buf
       close(idump)
       if (index(buf,"ASCII") .gt.0) then
          lascii=.true.
       elseif (index(buf,"BINARY") .gt.0) then
          lascii = .false.
       end if
    end if



    if (lascii) then
       open(idump,file=fname,status="old", form="formatted",err=22)
       ! read the header
       lheader = .true.
       levecs = .false.
       do while(lheader)
          ! check if we are still reading the header
          ! the header starts with "#" in every line
          read(idump,'(a)') buf

          if ( index(buf,"#") .ne. 1) then
             lheader = .false.
             backspace(idump)
             cycle
          end if

          ! get info from header

          i = index(buf,"gdim:",.true.)
          if (i.gt.0) then
             i = i+5
             read(buf(i:i+8),*) gdim1
          end if

          i = index(buf,"eigen",.true.)
          if (i.gt.0) levecs = .true.

          i = index(buf,"Coord",.true.)
          if (i.gt.0) coord=buf(i:)

       end do

       if (gdim1.ne.density%mgdim) then
          routine = "LoadDensity"
          message = "gdim of system does not match gdim in "//trim(fname)
          call errormsg
       end if

       write(ilog,'(a)') "Loading file "//trim(fname)
       write(ilog,'(a,i6)') "Grid dimension:", gdim1
       if (levecs) then
          write(ilog,'(a)') "Contains eigenvectors (SPP) for "//trim(coord)
       else
          write(ilog,'(a)') "Contains density for "//trim(coord)
       end if
       write(ilog,'(a)') ""


       density%lholdsevecs = levecs

       do i=1,mgdim
          read(idump,*,err=23) (density%density(j,i), j=1,mgdim)
       end do

       if (density%lholdsevecs) then
          read(idump,*,err=23) (density%npop(j), j=1,mgdim)
       end if

       close(idump)
       return
    end if

    buf = " "
    open(idump,file=fname,status="old",form="unformatted",err=22)

    levecs = .false.
    lheader = .true.
    do while(lheader)
       ! check if we are still reading the header
       ! the header starts with "#" in every line
       read(idump) buf
       if ( index(buf,"#") .ne. 1) then
          lheader = .false.
          backspace(idump)
          cycle
       end if

       ! get info from header

       i = index(buf,"gdim:",.true.)
       if (i.gt.0) then
          i = i+5
          read(buf(i:i+8),*) gdim1
       end if

       i = index(buf,"eigenvectors",.true.)
       if (i.gt.0) levecs = .true.

       i = index(buf,"Coord",.true.)
       if (i.gt.0) coord=buf(i:)

    end do

    ! read the header

    if (gdim1.ne.mgdim) then
       routine = "LoadDensity"
       message = "gdim of system does not match gdim in "//trim(fname)
       call errormsg
    end if

    write(ilog,'(a)') "Loading file "//trim(fname)
    write(ilog,'(a,i6)') "Grid dimension:", gdim1
    if (levecs) then
       write(ilog,'(a)') "Contains eigenvectors (SPP) for "//trim(coord)
    else
       write(ilog,'(a)') "Contains density for "//trim(coord)
    end if
    write(ilog,'(a)') ""


    density%lholdsevecs = levecs

    do i=1,mgdim
       read(idump,err=23) (density%density(j,i), j=1,mgdim)
    end do

    if (density%lholdsevecs) then
       read(idump,err=23) (density%npop(j), j=1,mgdim)
    end if

    close(idump)

    return

22  continue
    ifail = 1
    density%density = 0.d0
    density%npop = 0.d0
    return
23  continue
    ifail = 2
    density%density = 0.d0
    density%npop = 0.d0
    !write(*,*) i
    return

  end subroutine LoadDensity



  subroutine chkevecs(natpot)
    implicit none
    type(natpot_t),intent(IN)      :: natpot

    integer                        :: mgdim, maxmgdim, maxpotdim, potdim
    integer                        :: m,m1,n,n1,ierr,ig,ig1
    ! LAPACK
    character, parameter           :: JOBZ='V'
    character, parameter           :: UPLO='L'
    integer                        :: LWORK
    double precision               :: fac, trij

    double precision, allocatable  :: WORK(:), quality(:), mat(:,:), npop(:)
    double precision, allocatable  :: evec1(:,:), evec2(:,:), npop1(:), npop2(:)


    ! we do this in rank 0 only
    if (mpirank.ne.0) return

    maxmgdim = -1
    maxpotdim = -1
    do m=1,ndensities
       maxmgdim = max(densities1(m)%mgdim, maxmgdim)
       m1 = densities1(m)%dmode
       maxpotdim = max(maxpotdim,natpot%potdim(m1))
    end do
    allocate(quality(maxpotdim))

    ! Workspace for LAPACK
    LWORK = (maxmgdim+2)*maxmgdim
    allocate(WORK(LWORK))

    do m = 1,ndensities

       m1 = densities1(m)%dmode
       potdim = natpot%potdim(m1)
       mgdim = densities1(m)%mgdim

       allocate(mat(mgdim,mgdim))
       allocate(npop(mgdim))

       allocate(evec1(mgdim, potdim))
       allocate(evec2(mgdim, potdim))
       allocate(npop1(potdim))
       allocate(npop2(potdim))

       !$omp parallel do private(ig)
       do ig1=1,mgdim
          do ig=1,mgdim
             mat(ig,ig1) = densities1(m)%density(ig,ig1)
          end do
       end do
       !$omp end parallel do
       npop = 0.d0
       ierr = 0
       call DSYEV( JOBZ, UPLO, mgdim, mat, mgdim, &
            npop, WORK, LWORK, ierr )

       do n=1,potdim
          evec1(:,n) = mat(:,mgdim-n+1)
          npop1(n) = npop(mgdim-n+1)
       end do

       !$omp parallel do private(ig)
       do ig1=1,mgdim
          do ig=1,mgdim
             mat(ig,ig1) =  densities2(m)%density(ig,ig1)
          end do
       end do
       !$omp end parallel do

       npop = 0.d0
       ierr = 0
       call DSYEV( JOBZ, UPLO, mgdim, mat, mgdim, &
            npop , WORK, LWORK, ierr )

       do n=1,potdim
          evec2(:,n) = mat(:,mgdim-n+1)
          npop2(n) = npop(mgdim-n+1)
       end do

       deallocate(mat,npop)

       !$omp parallel do
       do n=1,potdim
          quality(n) = abs(dot_product(evec1(:,n),evec2(:,n)))
       end do
       !$omp end parallel do

       trij = 0.d0
       !$omp parallel do private(n1) reduction(+:trij) schedule(dynamic)
       do n=1,potdim
          do n1=1,potdim
             trij = trij + dot_product(evec1(:,n),evec2(:,n1))**2
          end do
       end do
       !$omp end parallel do

       write(ilog, '(a,i4,a)') " ** Mode  ",m1," **"
       write(ilog, '(a)') "   Overlaps of SPP obained " &
            //"with odd and even points"
       write(ilog, '(a)') "   of the trajectory:"
       write(ilog, '(a)') "    No.       ovlp          " &
            //"pop(odd) [eV**2]  pop(even) [eV**2]"


       fac = 2.d0*ev**2/natpot%rvpotdim

       ! npops must be multiplied by 2 because
       ! only half the sampling points are used
       do n=1,natpot%potdim(m1)
          write(ilog,'(i8,f16.6,5x,2(e11.4e2,7x))')  &
               n, quality(n), npop1(n)*fac, npop2(n)*fac
       end do

       write(ilog, '(a)') " --------------------------------------------"
       write(ilog, '(a,f11.6)') "     Average ", sum(quality(1:potdim))/potdim
       write(ilog, '(a,f11.6)') " tr{P1 P2}/N ", trij/potdim
       write(ilog, '(a)')
       call myflush(ilog)

       deallocate(npop1,npop2)
       deallocate(evec1,evec2)

    end do

  end subroutine chkevecs




  subroutine symdensities(natpot, densarr, ndens)

    implicit none

    integer, intent(IN)              :: ndens
    class(natpot_t),intent(IN)       :: natpot
    type(densities_t),intent(INOUT)  :: densarr(ndens)

    integer                          :: ig,ig1,z,z1,nsym,k,k1
    integer                          :: m,md,md1,m1,m2, allocstat

    type (densities_t), allocatable  :: tmpdensities(:)

    if (nsymmetries .le. 1) return

    ! workspace
    allocate(tmpdensities(ndens))
    call alloc_densities_arr(natpot,tmpdensities,ndens, allocstat)

    ! do the symmerisation

    do md=1,ndens
       m = densarr(md)%dmode

       do nsym=1,nsymmetries

          m1 = symmmap(m,nsym)
          ! find the mode mapping
          do m2 = 1,ndens
             if (densarr(m2)%dmode .eq. m1) then
                md1 = m2
                exit ! loop
             end if
          end do

          !$omp parallel do private(z,k,k1,z1,ig1)
          do ig=1,densities(md)%mgdim
             z = zsymidxmap(m1,nsym)+ig-1
             k = symidxmap(z)

             do ig1=1,densities(md)%mgdim
                z1 = zsymidxmap(m1,nsym)+ig1-1
                k1 = symidxmap(z1)
                tmpdensities(md)%density(k1,k) = tmpdensities(md)%density(k1,k) &
                     + densarr(md1)%density(ig1,ig)
             end do
          end do
          !$omp end parallel do
       end do

    end do

    ! finally replace the input array

    !$omp parallel do
    do m=1,ndens
       densarr(m)%density = tmpdensities(m)%density/nsymmetries
    end do
    !$omp end parallel do


    ! free workspace
    call free_densities_arr(tmpdensities, ndens)
    deallocate(tmpdensities)

  end subroutine symdensities



end module mcdensities_s


