!  -*- f90 -*-

! solve "omega * coeff = Vn2" for coeff


! this just a wrapper that decides weather to use lapack or scalapack
subroutine mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)

  use mpidata

  implicit none
  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)

#ifdef SCALAPACK
  ! for small matrices or one mpi rank we use the regular lapack
  if (nmpi.eq.1 .or. modcdim .le. 512 ) then
     ! regular lapacck
     if (mpirank.eq.0) then
        call lapack_mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)
     end if
  else
     ! scalapack
     call scalapack_mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)
  end if
#else
  !fallback to regular lapack
  if (mpirank.eq.0) then
     call lapack_mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)
  end if
#endif


end subroutine mcsolve_spd



!###########################################
! solve ovlp * coeff = omega v
!
!
!###########################################
subroutine lapack_mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)
  implicit none
  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)


  character(len=*), parameter      :: UPLO='L'

  call DPOSV( UPLO, modcdim, NRHS, ovlp, modcdim, coeff, &
       modcdim, ierr )


end subroutine lapack_mcsolve_spd



subroutine scalapack_mcsolve_spd(ovlp,coeff,modcdim,NRHS,ierr)

  use global
  use mpidata
#ifdef MPI
  use mpi
#endif

  implicit none

  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)

#ifdef SCALAPACK
#ifdef MPI

  character, parameter             :: UPLO="L"

  integer                          :: MB, mcoeff,movlp, ncoeff,novlp
  integer                          :: DESCovlp(9), DESCovlp2(9)
  integer                          :: DESCcoeff(9), DESCcoeff2(9)
  integer                          :: NPROW,NPCOL,CONTEXT,myrow,mycol
  integer                          :: numroc, mpierr, ovlp2dim,coeff2dim
  double precision, allocatable    :: ovlp2(:,:), coeff2(:,:)

  external numroc
  ierr = 0


  ! we initialize a 2D grid
  ! which is as square as possible
  NPROW = int(dsqrt(dble(nmpi)+1.d-9))
  if (nprow**2 .gt. nmpi) nprow = nprow -1
  NPCOL =  nprow

  ! add an extra row if we have space
  if ((nprow+1)*npcol .le. nmpi) then
     nprow = nprow+1
  end if
  ! add another extra row if we have space
  if ((nprow+1)*npcol .le. nmpi) then
     nprow = nprow+1
  end if


 ! if (mpirank .ge. NPCOL*NPROW) return


  ! initialize the process grid
  CALL BLACS_GET( -1, 0, CONTEXT )
  CALL BLACS_GRIDINIT( CONTEXT, 'R', NPROW, NPCOL )
  call blacs_gridinfo(CONTEXT, nprow, npcol, myrow, mycol)

  ! early exit for all ranks hat
  ! are not part of the process grid
  if (mycol+1 .gt. NPCOL .or. myrow+1 .gt. NPROW) return


  ! create two global arrays descriptors that
  ! describe ovlp and coff as they are distributed
  ! among the MPI ranks now (only in mpirank 1)
  ! setup array descriptors for the matrices omega and coeff as they are now
  CALL DESCINIT(DESCovlp ,modcdim ,modcdim ,modcdim,modcdim,0,0,CONTEXT,modcdim,mpierr)
  CALL DESCINIT(DESCcoeff ,modcdim ,NRHS ,modcdim,NRHS ,0,0,CONTEXT,modcdim,mpierr)

  ! redustribute: calculate local array dimensions
  ! for the re-distributed matrices
  MB = MBdefault
  movlp = NUMROC(modcdim, MB, myrow, 0, NProw )
  mcoeff = NUMROC(modcdim, MB, myrow, 0, NProw )
  novlp = NUMROC(modcdim, MB, mycol, 0, NPcol )
  ncoeff =  NUMROC(NRHS, MB, mycol, 0, NPcol )

  coeff2dim = mcoeff * ncoeff
  ovlp2dim = movlp * novlp

  ! create array descriptors for the re-disribued matrices
  CALL DESCINIT(DESCovlp2 ,modcdim ,modcdim ,MB,MB,0,0,CONTEXT,movlp,mpierr)
  CALL DESCINIT(DESCcoeff2 ,modcdim ,NRHS ,MB,MB ,0,0,CONTEXT,mcoeff,mpierr)

  ! allocate space for the re-disributed matrices
  if (ovlp2dim .gt. 0) then
     allocate(ovlp2(movlp,novlp))
  end if
  if (coeff2dim .gt. 0) then
     allocate(coeff2(mcoeff,ncoeff))
  end if

  ! if modcdim is large, we need
  ! to use a self written distribuion routine
  ! because of an integer overflow in the scalapack
  ! library if not compiled with i8
  ! google "xxmr2d:out of memory"
  call mcdistribute(ovlp, ovlp2,movlp,novlp,DESCovlp2)
  call mcdistribute(coeff,coeff2,mcoeff,ncoeff,DESCcoeff2)

  ! and finally acually re-distribute
!  call pdgemr2d(modcdim, modcdim,ovlp, 1, 1, DESCovlp, ovlp2, 1, 1, DESCovlp2, CONTEXT)
!  call pdgemr2d(modcdim, NRHS, coeff, 1, 1, DESCcoeff, coeff2, 1, 1, DESCcoeff2, CONTEXT)

  ! solve
  call PDPOSV( UPLO, modcdim, NRHS, ovlp2, 1, 1, DESCovlp2, coeff2, 1, 1, DESCcoeff2, ierr )
  if (ierr .ne. 0) then
     write(*,*) "### PDPOSV FATAL ERROR:"
     write(*,*) "### mpirank, ierr", mpirank, ierr
      if (allocated(ovlp2))  deallocate(ovlp2)
      if (allocated(coeff2)) deallocate(coeff2)
     call BLACS_GRIDEXIT(CONTEXT)
     return
  end if

  call  mcgather(coeff,coeff2,mcoeff,ncoeff,DESCcoeff2)
  ! gather the from the distributed matrix back to process-grid (0,0)
  ! call pdgemr2d(modcdim, NRHS, coeff2, 1, 1, desccoeff2, coeff, 1, 1, desccoeff, CONTEXT)

  if (allocated(ovlp2)) deallocate(ovlp2)
  if (allocated(coeff2))deallocate(coeff2)

  ! release the process grid
  call BLACS_GRIDEXIT(CONTEXT)


  return

#endif
  ! no MPI but Scalapack
  message = "ScaLAPACK needs MPI but this binary "// &
       "is compiled without MPI support. This should never have happened."
  routine = "scalapack_mcsolve_spd"
  call errormsg

#endif

end subroutine scalapack_mcsolve_spd



! disribute the matrix A in rank 0 / grid (0,0)
! onto matrices A2 in all ranks of the grid
subroutine mcdistribute(A, A2 ,MA2,NA2,DESCA2)

  use mpidata

  implicit none
  integer, intent(IN)           :: DESCA2(9), MA2,NA2
  double precision, intent(IN)  :: A(DESCA2(3),DESCA2(4))
  double precision, intent(OUT) :: A2(MA2,NA2)
#ifdef SCALAPACK
  integer                       :: M,N,MB,NB, ICTXT, NUMROC
  integer                       :: NPROW, NPCOL, myrow, mycol
  integer                       :: rdest, cdest, nel, mel
  integer                       :: ia,ja, nmb,nnb, i,j
  integer                       :: zmw, znw, msend,nsend, iceil, senddim, recdim
  double precision, allocatable :: work(:,:)

  external  BLACS_GRIDINFO, NUMROC, iceil

  ! if we are a process that does not get a
  ! share of the erray, we are done and return
  recdim = MA2 * NA2
  if (recdim .eq. 0) return


  M = desca2(3)
  N = desca2(4)
  MB = desca2(5)
  NB = desca2(6)

  ! get the grid info
  ICTXT = DESCA2(2)
  CALL BLACS_GRIDINFO( ICTXT, NPROW, NPCOL, MYROW, MYCOL )


  ! if we are rank 0 then we send
  if (myrow .eq. 0 .and. mycol .eq. 0) then

     do rdest = 0, nprow-1
        do cdest = 0, npcol-1

           !calculate the array size of the reciever
           msend = numroc(M,MB,rdest,0,NPROW)
           nsend = numroc(N,NB,cdest,0,NPCOL)

           ! skip sending , if the receiving process does
           ! not get a part of the array
           senddim = msend * nsend
           if (senddim .eq. 0) cycle


           ! allocate a local workspace
           allocate(work(msend,nsend))
           work = 0.d0

           !populate the local workspace

           ! pointer to start in A
           ia = mb*rdest +1
           ja = nb*cdest +1

           !No of blocks to gather
           nmb = iceil(msend,MB)
           nnb = iceil(nsend,NB)

           ! run over blocks
           do j=1,nnb
              nel = min(nsend-(j-1)*nb,nb)-1
              znw = (j-1)*nb + 1
              ia = mb*rdest +1
              do i=1,nmb
                 mel = min(msend-(i-1)*mb,mb)-1
                 zmw = (i-1)*mb + 1

                 work(zmw:zmw + mel, znw:znw+nel) = A(ia:ia+mel, ja:ja+nel)
                 ia = ia + MB*NPROW
              end do
              ja = ja + NB*NPCOL
           end do


           if (rdest .eq. myrow .and. cdest .eq. mycol) then
              ! sender and receiver are identical
              A2 = work
           else

              ! send the local workspace
              call dgesd2d( ictxt, msend, nsend, work, msend, rdest, cdest )

           end if

           deallocate(work)

        end do
     end do

  else
     ! we receive

     call dgerv2d( ictxt, ma2, na2, a2, ma2, 0, 0 )

  end if

#endif

  return

end subroutine mcdistribute


! gather the distributed matrix A2 on the grid
!  into the local matrix A in rank 0 / grid (0,0)
subroutine mcgather(A,A2,MA2,NA2,DESCA2)

  use mpidata

  implicit none
  integer, intent(IN)               :: MA2,NA2, DESCA2(9)
  double precision, intent(OUT)     :: A(DESCA2(3), DESCA2(4))
  double precision, intent(IN)      :: A2(MA2,NA2)

#ifdef SCALAPACK
  integer                       :: M,N,MB,NB, ICTXT, NUMROC
  integer                       :: NPROW, NPCOL, myrow, mycol
  integer                       :: rsrc, csrc, nel, mel
  integer                       :: ia,ja, nmb,nnb, i,j
  integer                       :: zmw, znw, msend,nsend, iceil
  integer                       :: senddim, recdim
  double precision, allocatable :: work(:,:)

  external  BLACS_GRIDINFO, NUMROC, iceil

  ! skip sending, if this process does
  ! not hold a share of the array
  senddim = MA2 * NA2
  if (senddim .eq. 0) return

  M = desca2(3)
  N = desca2(4)
  MB = desca2(5)
  NB = desca2(6)

  ! get the grid info
  ICTXT = DESCA2(2)
  CALL BLACS_GRIDINFO( ICTXT, NPROW, NPCOL, MYROW, MYCOL )


 ! if we are rank 0 then we recieve
  if (myrow .eq. 0 .and. mycol .eq. 0) then

     ! run over the process grid
     do rsrc = 0, nprow-1
        do csrc = 0, npcol-1

           !calculate the array size of the sender
           msend = numroc(M,MB,rsrc,0,NPROW)
           nsend = numroc(N,NB,csrc,0,NPCOL)
           ! skip if sender has no data
           recdim = msend * nsend
           if (recdim .eq. 0) cycle

           ! allocate a local workspace ane recieve
           allocate(work(msend,nsend))
           if (rsrc .eq. myrow .and. csrc .eq. mycol) then
              work = A2
           else
              call dgerv2d( ictxt, msend, nsend, work, msend, rsrc, csrc )
           end if
           !populate the local big matrix

           ! pointer to start in A
           ia = mb*rsrc +1
           ja = nb*csrc +1

           !No of blocks to gather
           nmb = iceil(msend,MB)
           nnb = iceil(nsend,NB)

           ! run over blocks
           do j=1,nnb
              nel = min(nsend-(j-1)*nb,nb)-1
              znw = (j-1)*nb + 1
              ia = mb*rsrc +1
              do i=1,nmb
                 mel = min(msend-(i-1)*mb,mb)-1
                 zmw = (i-1)*mb + 1

                  A(ia:ia+mel, ja:ja+nel) = work(zmw:zmw + mel, znw:znw+nel)
                 ia = ia + MB*NPROW
              end do
              ja = ja + NB*NPCOL
           end do

           deallocate(work)

        end do
     end do
  else
     ! we send
     call dgesd2d( ictxt, MA2, NA2, A2, MA2, 0, 0 )
  end if

#endif

  return

end subroutine mcgather

