!  -*- f90 -*-

module mcmetropolis
    use global
    use channels
    use dvrtype_m
    use mcpfitparams
    use pesselect_s, only: potential
    use mcsampling
    use mcpmpitools

    use mpidata
#ifdef MPI
    use mpi
#endif


  implicit none

contains

  !********************************************************
  ! metropolisstep: do one metropolis step on a grid
  ! until either the step has been acceped or
  ! the number of tries has reached a maximum count
  !
  ! potenial  real*8 funcion that returns the
  !           energy given an index point.
  !
  ! pos       on entry the current position of
  !           the walker, on exit the new position
  !           of the walker
  !
  ! energy    On entry the energy at the current position
  !           of the walker, on exit the energy at
  !           the new position of the walker
  !
  ! beta      1/Kb/T
  !
  ! ndof      number od indices in pos
  !
  ! gdim      size of grid
  !
  ! maxtries  number of tries until considered trapped in
  !           local minimum
  !
  ! ntry      number of tries until sucess
  !
  ! ierr == 0 success
  ! ierr == 1 trapped (maxtries exceeded)
  !
  !
  ! 06/2016 MS
  !********************************************************

  subroutine metropolisstep(potential,pos,energy,beta,ndof,gdim,lperiodic,maxtries,ntry,ierr)

    implicit none

    integer, intent(IN)             :: ndof, maxtries
    integer, intent(INOUT)          :: pos(ndof)
    integer, intent(IN)             :: gdim(ndof)
    integer, intent(OUT)            :: ierr, ntry
    logical, intent(IN)             :: lperiodic(ndof)
    double precision, intent(INOUT) :: energy
    double precision, intent(IN)    :: beta

    integer                         :: newpos(ndof),newpos1(ndof)
    double precision                :: newenergy
    integer                         :: idof, isign
    logical                         :: lfound
    double precision                :: rdum, p
    double precision                :: potential
    external potential
    !interface
    !   function potential(idx,ndof)
    !     double precision             :: potential
    !     integer, intent(IN)          :: ndof
    !     integer, intent(IN)          :: idx(ndof)
    !   end function potential
    !end interface

    lfound = .false.
    ntry = 0
    newpos = pos



    search: do while ((.not. lfound) .and. (ntry .lt. maxtries))

       ntry = ntry + 1

       ! get the direction if the step
       ! from idof = 1..ndof decrese grid index idof
       ! from idof = ndof+1..2ndof increase grid index idof

111    continue
       call random_number(rdum)
       idof = int(rdum*ndof) + 1

       call random_number(rdum)
       if (rdum .lt. 0.5d0) then
          isign = -1
       else
          isign = 1
       end if

       ! metropolis step
       if ( newpos(idof) + isign .gt. gdim(idof)) then
          if (lperiodic(idof)) then
             newpos(idof) = 1
          else
             ! new draw
             goto 111
          end if
       else  if ( newpos(idof) + isign .lt. 1) then
          if (lperiodic(idof)) then
             newpos(idof) = gdim(idof)
          else
             ! new draw
             goto 111
          end if
       else
          newpos(idof) = newpos(idof) + isign
       end if

       newpos1 = newpos
       newenergy =  potential(newpos,ndof)

       if (newenergy .le. energy) then
          lfound = .true.
       else
          p = dexp(-(newenergy-energy)/beta)
          call random_number(rdum)
          if (rdum .le. p) lfound = .true.
       end if

       if (lfound) then
          ! accept and update
          energy = newenergy
          pos(idof) = newpos(idof)
       else
          !reject and reset
          newpos(idof) =  pos(idof)
       end if

    end do search

    ! check error status
    if (.not. lfound) then
       ierr = 1
    else
       ierr = 0
    end if


  end subroutine metropolisstep


  !********************************************************
  ! metropolistrajecory: optian a metropolis trajectory and
  ! write it to a file.
  !
  ! potenial  real*8 funcion that returns the
  !           energy given an index point.
  !
  ! nsample   length of the trajectory
  !
  ! neqlb     skip the first neqlb steps to
  !           equilibrate the walker
  !
  ! nstep     Skip this number of steps between
  !           two acvceped steps to make
  !           the trajectory less correlated
  !
  ! beta      1/Kb/T
  !
  ! ratio     on exit, the ratio of acceped
  !           steps to tried steps
  !
  ! dvr       the primitive grid
  !
  ! filenameidx
  !           The name of the trajectory file
  !
  ! filenameenergies
  !           The name of the eneries file
  !
  ! lbinary   whether the trajectory is written
  !           in binary form (ASCII otherwise)
  !
  ! 06/2016 MS
  !********************************************************
  subroutine MetropolisTrajecotry(dvr, task, nfile, ratio,emean, &
       filenameidx,filenameenergies)

    implicit none

    integer, intent(IN)           :: task, nfile
    type(dvr_t), intent(IN)       :: dvr
    double precision, intent(OUT) :: ratio, emean
    character(len=*),intent(IN)   :: filenameidx
    character(len=*),intent(IN)   :: filenameenergies

    logical                       :: lexist
    integer                       :: i,n,f,j,ierr
    double precision              :: rdum(dvr%ndof), energy, rel, di
    logical                       :: lperiodic(dvr%ndof)

    integer                       :: maxtries, pos(dvr%ndof)
    integer                       :: tottries, ntries
    character(len=c2)             :: fmt,buf
    character(len=20)             :: bar

    double precision, allocatable :: energies(:), tmpenergies(:)
    integer, allocatable          :: dvrindex(:,:), tmpindex(:,:)
    integer                       :: nsample
#ifdef MPI
    integer :: idum
#endif

    ! format for ascii output,
    ! also use in log if walker gets trapped
    write(buf,*) dvr%ndof
    buf=adjustl(buf)
    fmt="("//trim(buf)//"i4)"

    ierr = 0
    ratio = 0.d0
    tottries = 0

    do f=1,dvr%ndof
       if (dvr%basis(f) .eq. 4 .or. dvr%basis(f) .eq. 5) then
          lperiodic(f) = .true.
       else
          lperiodic(f) = .false.
       end if
    end do


    ! if we are not in mpirank 0 we save
    ! the store results in arrays and save them in the end

    call  mpi_set_chunksize(ntotsample(task,nfile),nsample)

    allocate(dvrindex(dvr%ndof,nsample))
    allocate(energies(nsample))

    if (mpirank.eq.0) then
       ! check if files are there and overwrite is allowed
       inquire(file=filenameidx, exist=lexist)
       if (lexist .and. .not. lovwr) then
          routine = "MetropolisTrajecotry"
          message = "File exists, use -w option: "//trim(filenameidx)
          call errormsg
       end if
       inquire(file=filenameenergies, exist=lexist)
       if (lexist .and. .not. lovwr) then
          routine = "MetropolisTrajecotry"
          message = "File exists, use -w option: "//trim(filenameenergies)
          call errormsg
       end if
    end if

    ! we have 2*ndof possibilities to choose a direction.
    maxtries = dvr%ndof * 100
    maxtries = max(maxtries, 100)   ! at least  100
    maxtries = min(maxtries, 10000) ! at most 10000


10  continue

    ! random initial position in the middle
    call random_number(rdum)
    do f=1,dvr%ndof
       pos(f) = int((rdum(f)/2)*dvr%gdim(f)) + dvr%gdim(f)/4 + 1
       if ( pos(f) .gt. dvr%gdim(f) ) pos(f) = dvr%gdim(f)
       if ( pos(f) .lt. 1           ) pos(f) = 1
    end do


    energy = 1.d99
    ! burn-in, all mpiranks do this
    equilibrate: do n=1,nskipfirst(task,nfile)

       call metropolisstep(potential,pos,energy, beta(task,nfile), &
            dvr%ndof,dvr%gdim,lperiodic,maxtries,ntries,ierr)


       if (ierr .ne. 0) exit
    end do equilibrate


    ! check if burn-in went well
    if (ierr .ne. 0) then
       routine = "MetropolisTrajecotry"
       message = "Random walker got trapped during burn-in. "// &
            "Found a hole? Maybe use a higher temperature... Restarting."

       write(*,'(a)') "   f   |.........pos........|   i / dim"
       write(*,'(a)') "---------------------------------------"
       do f=1,dvr%ndof
          di = dble(pos(f)-1)/(dvr%gdim(f)-1)
          rel = di*19.d0
          i = int(rel) + 1
          bar = "...................."
          bar(i:i) = "*"
          write(*,'(i4,3x,a,i4,a,i4,2x,a)') f, "|"//trim(bar)//"|", &
               pos(f),"/", dvr%gdim(f), trim(dvr%modelabel(f))
       end do
       write(*,'(a)') "---------------------------------------"
       write(*,*) "e=", energy
       write(*,*)
       call warnmsg
       lwarn = .false.
       goto 10
    end if


    emean = 0.d0

    ! after burn-in do the sampling that is actually used
    sampling: do n=1,nsample

       ! this is a step we use
       call metropolisstep(potential,pos,energy, beta(task,nfile), &
            dvr%ndof,dvr%gdim,lperiodic,maxtries,ntries,ierr)
       if (ierr .ne. 0) exit

       tottries = tottries + ntries
       emean = emean + energy
       dvrindex(:,n) = pos
       energies(n) = energy

       ! these steps we skip as requested by user
       steps: do j=1,nstep(task,nfile)-1
          call metropolisstep(potential,pos,energy, beta(task,nfile), &
               dvr%ndof,dvr%gdim,lperiodic,maxtries,ntries,ierr) 
          if (ierr .ne. 0) exit

          tottries = tottries + ntries

       end do steps

       if (ierr .ne. 0) exit

    end do sampling

    ! check if burn-in went well
    if (ierr.ne.0) then

       routine = "MetropolisTrajecotry"
       message = "Random walker got trapped during sampling. "// &
            "Found a hole? Maybe use a higher temperature... Restarting."

       write(*,'(a)') "   f   |.........pos........|   i / dim"
       write(*,'(a)') "---------------------------------------"
       do f=1,dvr%ndof
          di = dble(pos(f)-1)/(dvr%gdim(f)-1)
          rel = di*19.d0
          i = int(rel) + 1
          bar = "...................."
          bar(i:i) = "*"
          write(*,'(i4,3x,a,i4,a,i4,2x,a)') f, "|"//trim(bar)//"|", &
               pos(f),"/", dvr%gdim(f), trim(dvr%modelabel(f))
       end do
       write(*,'(a)') "---------------------------------------"
       write(*,*) "e=", energy
       write(*,*)
       call warnmsg
       lwarn = .false.
       goto 10
    end if



#ifdef MPI

    ! we are done. if MPI was used, we need to fetch the
    ! results from the different ranks into rank 0
    ! where IO is handled.

    call MPI_REDUCE(tottries,idum,1,&
         MPI_INTEGER,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
    if (mpirank.eq.0) tottries = idum
    call MPI_Reduce(emean,rdum(1),1,&
         MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
    if (mpirank.eq.0) emean = rdum(1)

#endif


    if (mpirank .eq. 0) then

       ratio = (dble(ntotsample(task,nfile)) + dble(ntotsample(task,nfile)) &
            *dble((nstep(task,nfile)-1)))/tottries
       emean = emean/ntotsample(task,nfile)

       ! open output files
       if (lasciisamplings) then
          open(idump,file=filenameidx,status="unknown", &
               form="formatted",iostat=ierr)
          write(idump,'(a)') "# "//trim(idxfileinfo(task,nfile))
       else
          open(idump,file=filenameidx,status="unknown", &
               form="unformatted",iostat=ierr)
          write(idump) "# "//trim(idxfileinfo(task,nfile))
       end if


       if (ierr.ne.0) then
          routine = "MetropolisTrajecotry"
          !write(message, '(i4,a)') mpirank,"File could not be opened: "//trim(filenameidx)
          message = "File could not be opened: "//trim(filenameidx)

          call errormsg
       end if

       open(ivpot,file=filenameenergies,status="unknown", &
            form="formatted",iostat=ierr)
       write(ivpot,'(a)') "# "//trim(idxfileinfo(task,nfile))

       if (ierr.ne.0) then
          routine = "MetropolisTrajecotry"
          !write(message, '(i4,a)') mpirank, "File could not be opened: "//trim(filenameenergies)
          message = "File could not be opened: "//trim(filenameenergies)
          call errormsg
       end if

    end if


    do i = 0,nmpi-1
       nsample = mpi_chunksizes(i)
       if (mpirank.eq. 0) then
          if (allocated(tmpindex)) deallocate(tmpindex)
          if (allocated(tmpenergies)) deallocate(tmpenergies)
          nsample = mpi_chunksizes(i)
          allocate(tmpindex(dvr%ndof,nsample))
          allocate(tmpenergies(nsample))
       end if

       call mpi_get_chunk(dvrindex,tmpindex,dvr%ndof,nsample,i)
       call mpi_get_energies(energies,tmpenergies,nsample,i)

       if (mpirank.eq. 0) then
          ! first we write rank 0
          do n=1,nsample
             if (lasciisamplings) then
                write(idump,fmt,iostat=ierr) (tmpindex(f,n), f=1,dvr%ndof)
             else
                write(idump,iostat=ierr) (tmpindex(f,n), f=1,dvr%ndof)
             end if
             if (ierr.ne.0) then
                routine = "MetropolisTrajecotry"
                message = "Error while writing to file: "//trim(filenameidx)
                call errormsg
             end if
          end do
          do n=1,nsample
             write(ivpot,'(f20.12)',iostat=ierr) tmpenergies(n)
             if (ierr.ne.0) then
                routine = "MetropolisTrajecotry"
                message = "Error while writing to file: "//trim(filenameenergies)
                call errormsg
             end if
          end do

       end if

    end do


    if (mpirank.eq. 0) then
       close(idump)
       close(ivpot)
    end if
    if (allocated(tmpindex)) deallocate(tmpindex)
    if (allocated(tmpenergies)) deallocate(tmpenergies)



  end subroutine MetropolisTrajecotry





end module mcmetropolis
