!   -*- f90 -*-


module mcoverlap
  use natpottype
  use global
  use pglobal
  use mpidata
  use channels
  use logdat
  use timing

  implicit none


  double precision, allocatable :: omega(:,:)
  double precision, allocatable :: omegaT(:,:)

  ! this is temporary storage for speeding up calculations
  ! mainly to avoid pointer arithmetics  to be
  ! done numerous times and to avoid fetching memory from
  ! remote locations.
  double precision, allocatable :: sppsamplings(:,:)
  double precision, allocatable :: sppsamplingst(:,:)
  integer, allocatable          :: sppdims(:)


  ! we use some global pointers to big things in this module
  type(natpot_t), pointer       :: pnatpot => NULL()
  integer, pointer              :: pdvrindex(:,:) => NULL()

  integer                       :: nsample

  double precision              :: ovlptrace

contains



  !*******************************************************************
  ! chkNpotptr
  !
  ! check if the global natpot pointer is assiciated and allocated
  !
  !*******************************************************************
  subroutine chkNpotptr(callername)

    implicit none
    character(len=*), intent(IN)  :: callername

    if (.not. associated(pnatpot)) then
       routine = callername
       message = "Natpot Pointer is not assiciated."
       call errormsg
    end if

    if (.not. pnatpot%initialized) then
       routine = callername
       message = "Natpot has not been initialized. Aborting."
       call errormsg
    end if

  end subroutine chkNpotptr

end module mcoverlap




!*******************************************************************
! mcsppoverlap
!
! build the overlap matrix without building the omega matrix
!
!*******************************************************************
subroutine mcsppoverlap(ovlp)
  use pftools
  use mcoverlap
  use mcpfitparams
  use mpidata
#ifdef MPI
  use mpi
#endif
  implicit none
  double precision, intent(OUT)  :: ovlp(pnatpot%modcdim,pnatpot%modcdim)

  integer                        :: i,i1,m,n,z1

  integer                        :: midx1(pnatpot%nmode) ! auto alloc
  integer                        :: z(pnatpot%nmode) ! auto alloc
  double precision               :: Oij, r,Vn(nsample) ! auto alloc
  integer                        :: iwork(pnatpot%nmode,2)
  double precision               :: rwork(nsample)
  logical                        :: l1st

  double precision               :: mean, rms, largest, smallest

#ifdef MPI
  double precision               :: mpibuf(pnatpot%modcdim)
#endif

#ifdef DEBUG
  call chkNpotptr("mcsppoverlap")
#endif

  do i1=1,pnatpot%modcdim

     call cindex2sindex(i1,midx1,pnatpot%potdim,pnatpot%nmode)
     z1 = 0
     do m=1, pnatpot%nmode
        if (m .ne. pnatpot%modc) then
           z(m) = z1 + midx1(m)
           z1 = z1 + pnatpot%potdim(m)
        end if
     end do

     Oij = 0.d0
     !$omp parallel do private(r,m,l1st) reduction(+:Oij)
     do n=1,nsample
        l1st = .true.
        do m=1,pnatpot%nmode
           if (m .ne. pnatpot%modc) then
              if (l1st) then
                 r = sppsamplings(z(m),n)
                 l1st = .false.
              else
                 r = r * sppsamplings(z(m),n)
              end if
           end if
        end do
        Vn(n) = r
        Oij = Oij + r**2
     end do
     !$omp end parallel do

     ovlp(i1,i1) = Oij


     !$omp parallel do private(rwork,Oij,iwork) schedule(dynamic)
     do i=i1+1,pnatpot%modcdim

        call operateOmega1J(i, Oij, Vn, rwork,iwork(:,2), iwork(:,1))

        ovlp(i,i1) = Oij
        ovlp(i1,i) = Oij
     end do
     !$omp end parallel do

  end do


#ifdef MPI
  do i1=1,pnatpot%modcdim
     call MPI_Reduce(ovlp(:,i1),mpibuf,pnatpot%modcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,mpierr)
     if (mpirank.eq.0) then
        ovlp(:,i1) = mpibuf
     end if
  end do
#endif

 if (mpirank.eq.0) then
     largest = -1.d99
     smallest = 1.d99
     mean = 0.d0
     rms = 0.d0
     !$omp parallel do reduction(+:mean,rms) &
     !$omp reduction(min:smallest) reduction(max:largest)
     do i=1, pnatpot%modcdim
        mean = mean + ovlp(i,i)
        rms = rms + ovlp(i,i)**2
        largest = max(largest, ovlp(i,i))
        smallest = min(smallest,ovlp(i,i))
     end do
     !$omp end parallel do

     ovlptrace = mean
     mean = mean/ pnatpot%modcdim
     rms = dsqrt(rms/ pnatpot%modcdim)

     write(ilog,'(a)') ""
     write(ilog,'(a)') "Statistics: diagonal of overlap matrix"
     write(ilog,'(a,E24.16)') " mean:", mean
     write(ilog,'(a,E24.16)') " RMS: ", rms
     write(ilog,'(a,E24.16)') " min: ", smallest
     write(ilog,'(a,E24.16)') " max: ", largest
     write(ilog,'(a)') ""

  end if

#ifdef MPI
  call MPI_Bcast(ovlptrace,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
#endif

end subroutine mcsppoverlap



!*******************************************************************
! mcsppoverlapmem
!
! build the overlap matrix from the omega matrix
!
!*******************************************************************
subroutine mcsppoverlapmem(ovlp,modcdim)
  use mcpfitparams
  use mcoverlap
  use mpidata
#ifdef MPI
  use mpi
#endif
  implicit none

  integer, intent(IN)                  :: modcdim
  double precision, intent(OUT)        :: ovlp(modcdim,modcdim)
  integer                              :: i,i1
  double precision                     :: mean, rms, largest, smallest

#ifdef MPI
  integer                              :: ierr
  double precision                     :: mpibuf(modcdim)
#endif

  !$omp parallel do private(i) schedule(dynamic)
  do i1=1,modcdim
     do i=i1,modcdim
        ovlp(i,i1) = dot_product(Omega(:,i),Omega(:,i1))
     end do
  end do
  !$omp end parallel do


  !$omp parallel do private(i) schedule(dynamic)
  do i1=1,modcdim
     do i=i1+1,modcdim
        ovlp(i1,i) = ovlp(i,i1)
     end do
  end do
  !$omp end parallel do

#ifdef MPI
  do i1=1,modcdim
     call MPI_Reduce(ovlp(:,i1),mpibuf,modcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)
     if (mpirank.eq.0) then
        ovlp(:,i1) = mpibuf
     end if
  end do
#endif

  if (mpirank.eq.0) then
     largest = -1.d99
     smallest = 1.d99
     mean = 0.d0
     rms = 0.d0
     !$omp parallel do reduction(+:mean,rms) &
     !$omp reduction(min:smallest) reduction(max:largest)
     do i=1,modcdim
        mean = mean + ovlp(i,i)
        rms = rms + ovlp(i,i)**2
        largest = max(largest, ovlp(i,i))
        smallest = min(smallest,ovlp(i,i))
     end do
     !$omp end parallel do
     ovlptrace = mean

     mean = mean/modcdim
     rms = dsqrt(rms/modcdim)

     write(ilog,'(a)') ""
     write(ilog,'(a)') "Statistics: diagonal of overlap matrix"
     write(ilog,'(a,E24.16)') " mean:", mean
     write(ilog,'(a,E24.16)') " RMS: ", rms
     write(ilog,'(a,E24.16)') " min: ", smallest
     write(ilog,'(a,E24.16)') " max: ", largest
     write(ilog,'(a)') ""

  end if

#ifdef MPI
  call MPI_Bcast(ovlptrace,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#endif

end subroutine mcsppoverlapmem



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! mcsppoverlapMtxElementMem(j,j1,Oij)
! Calculate a single matrix element of the overlap matrix
! (if omega is allocated)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine mcsppoverlapMtxElementMem(j,j1,Oij)
  use mcoverlap
  implicit none
  integer, intent(IN)              :: j,j1
  double precision, intent(OUT)    :: Oij
  double precision                 :: ompddot

#ifdef DEBUG
  call chkNpotptr("mcsppoverlapMtxElement")
#endif

  if (.not. allocated(omega)) then
     routine = "mcsppoverlapMtxElementMem (mcoverlap.F90)"
     message = "Omega matrix is not allocated."
     call errormsg
  end if

  Oij = ompddot(nsample, omega(:,j), omega(:,j1))

end subroutine mcsppoverlapMtxElementMem


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! mcsppoverlapMtxElement(j,j1,Oij)
! Calculate a single matrix element of the overlap matrix
! (if omega is not allocated)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine mcsppoverlapMtxElement(j,j1,Oij,rwork)
  use mcoverlap
  use pftools
  implicit none
  integer, intent(IN)              :: j,j1
  double precision, intent(OUT)    :: Oij
  double precision, intent(INOUT)  :: rwork(nsample,2)
  integer                          :: midx1(pnatpot%nmode), z(pnatpot%nmode)
  integer                          :: n,m,z1
  double precision                 :: r
  logical                          :: l1st
  integer                          :: iwork(pnatpot%nmode,2)

#ifdef DEBUG
  call chkNpotptr("mcsppoverlapMtxElement")
#endif

  call cindex2sindex(j1,midx1,pnatpot%potdim,pnatpot%nmode)
  z1 = 0
  do m=1, pnatpot%nmode
     if (m .ne. pnatpot%modc) then
        z(m) = z1 + midx1(m)
        z1 = z1 + pnatpot%potdim(m)
     end if
  end do

  Oij = 0.d0
  !$omp parallel do private(r,m,l1st) reduction(+:Oij)
  do n=1,nsample
     l1st = .true.
     do m=1,pnatpot%nmode
        if (m .ne. pnatpot%modc) then
           if (l1st) then
              r = sppsamplings(z(m),n)
              l1st = .false.
           else
              r = r * sppsamplings(z(m),n)
           end if
        end if
     end do
     if (j1.ne. j) then
        rwork(n,1) = r
     else
        Oij = Oij + r**2
     end if
  end do
  !$omp end parallel do

  !for j==j1 the result is already in Oij
  if (j1.ne.j) then
     call operateOmega1J(j, Oij, rwork(:,1), rwork(:,2), iwork(:,2), iwork(:,1))
  end if


end subroutine mcsppoverlapMtxElement




!*******************************************************************
! mkOmega
!
! calculates for all sampling points the product of the spp
! and stores the result in Omega(n,I) where I is the combined index
! without the contracted modes and n is the sampling point
!
!*******************************************************************
subroutine mkOmega()
  use pftools
  use mcoverlap
  implicit none
  integer                       :: j,n

  integer                       :: lwork
  double precision, allocatable :: rwork(:)
  integer                       :: iwork(pnatpot%nmode,2)

#ifdef DEBUG
  call chkNpotptr("mkomega")
#endif

  if (allocated(omegaT)) then

     lwork = wk_comb_prod(sppdims,pnatpot%npotmodes)
     allocate(rwork(lwork))

     !$omp parallel do private(rwork)
     do n=1,nsample
        call mkOmega1N(n, omegaT(:,n), lwork, rwork)
     end do
     !$omp end parallel do

     deallocate(rwork)

  else if (allocated(omega)) then

     !$omp parallel do private(iwork)
     do j=1,pnatpot%modcdim
        call mkOmega1J(J, omega(:,j), iwork(:,1),iwork(:,2))
     end do
     !$omp end parallel do

  end if


  if (allocated(omegaT) .and. allocated(omega)) then

     !$omp parallel do private(n)
     do j=1,pnatpot%modcdim
        do n=1,nsample
           omega(n,j) =  omegaT(j,n)
        end do
     end do
     !$omp end parallel do
  end if



end subroutine mkOmega



!*******************************************************************
! mkSppSamplings
!
! fetch the sampling points from the spp and store them
! in the sppsamplings array
!
!*******************************************************************
subroutine mkSppSamplings(dvrindex)
  use pftools
  use mcoverlap
  implicit none
  integer, intent(IN)             :: dvrindex(pnatpot%ndof,nsample)

  integer                         :: m,n,k,j,z
  integer                         :: mpoint(pnatpot%nmode)
  integer                         :: allocstat

#ifdef DEBUG
  call chkNpotptr("mkSppSamplings")
#endif

  if (allocated(sppsamplings)) deallocate(sppsamplings)
  if (allocated(sppdims)) deallocate(sppdims)

  allocate(sppdims(pnatpot%npotmodes))

  k = 1
  do m=1,pnatpot%nmode
     if (m .ne. pnatpot%modc) then
        sppdims(k) = pnatpot%potdim(m)
        k = k+1
     end if
  end do

  allocate(sppsamplings(pnatpot%ntotspp,nsample), stat=allocstat)

  if (allocstat .ne. 0) then
     routine = "mkSppSamplings (prefetcing)"
     message = "Unable to allocate sppsamplings(nspf,nsample)."
     call errormsg
  end if


  !$omp parallel do private(mpoint,k,m,n,j,z)
  do n=1,nsample

     call f2mdecoder(dvrindex(:,n),mpoint,pnatpot%nspfdof, &
          pnatpot%spfdof,pnatpot%gdim,pnatpot%ndof,pnatpot%nmode, &
          pnatpot%maxspfdof)

     k = 1
     do m=1,pnatpot%nmode
        if (m .ne. pnatpot%modc) then
           do j=1,pnatpot%potdim(m)
              z = pnatpot%zevec(m,j) + mpoint(m) - 1
              sppsamplings(k,n) = pnatpot%evec(z)
              k = k+1
           end do
        end if
     end do

  end do
  !$omp end parallel do

  if (allocated(sppsamplingst)) deallocate(sppsamplingst)
  allocate(sppsamplingst(nsample,pnatpot%ntotspp), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "mkSppSamplings (prefetching)"
     message = "Unable to allocate sppsamplingst(nsample,nspp)."
     call errormsg
  end if

  do k=1,pnatpot%ntotspp
     !$omp parallel do
     do n=1,nsample
        sppsamplingst(n,k) = sppsamplings(k,n)
     end do
     !$omp end parallel do
  end do


end subroutine mkSppSamplings



!*******************************************************************
! mkOmega1N(N, omega1N, lwork, work)
!
! Calculates for One sampling point all possiple omega products
! and stores the result Omega1N
!
! For lwork see wk_comb_prod
!
!*******************************************************************
subroutine mkOmega1N(N, omega1N, lwork, work)
  use pftools
  use mcoverlap
  implicit none
  integer, intent(IN)             :: N,lwork
  double precision, intent(OUT)   :: omega1N(pnatpot%modcdim)
  double precision, intent(INOUT) :: work(lwork)

  call comb_prod(pnatpot%npotmodes, sppdims, pnatpot%ntotspp, &
       sppsamplings(:,N), pnatpot%modcdim, omega1N, lwork, work)


end subroutine mkOmega1N



!*******************************************************************
! mkOmega1N(N, omega1N, nsample, coeffidx,z)
!
! Calculates for one configuration the omega product
! for all sampling points and stores the result Omega1J
!
!*******************************************************************
subroutine mkOmega1J(J, omega1J, coeffidx,z)
  use pftools
  use mcoverlap
  implicit none
  integer, intent(IN)             :: J
  double precision, intent(OUT)   :: omega1J(nsample)
  ! iwork
  integer, intent(INOUT)          :: coeffidx(pnatpot%nmode)
  integer, intent(INOUT)          :: z(pnatpot%nmode)
  ! locals
  integer                         :: z1,m,n
  logical                         :: lst

  call cindex2sindex(J,coeffidx,pnatpot%potdim,pnatpot%nmode)
  z1 = 0
  do m=1, pnatpot%nmode
     if (m .ne. pnatpot%modc) then
        z(m) = z1 + coeffidx(m)
        z1 = z1 + pnatpot%potdim(m)
     end if
  end do

  lst = .true.
  do m=1, pnatpot%nmode
     if (m .ne. pnatpot%modc) then
        if (lst) then
           do n=1,nsample
              omega1J(n) = sppsamplingst(n,z(m))
           end do
           lst = .false.
        else
           do n=1,nsample
              omega1J(n) = omega1J(n)*sppsamplingst(n,z(m))
           end do
        end if
     end if
  end do

end subroutine mkOmega1J


!*******************************************************************
! calccoeff
!
! calculate the coefficient vector
!
!*******************************************************************
subroutine calccoeff(natpot, nindex, dvrindex, coeff_ext, coeff_extdim)
  use pftools
  use conjgrad
  use pesselect_m
  use mcpfitparams
  use mcoverlap
  use symmops_m
  use mpidata
#ifdef MPI
  use mpi
#endif

  implicit none
  integer, intent(IN)                    :: nindex, coeff_extdim
  type(natpot_t), intent(INOUT), target  :: natpot
  integer, intent(IN), target            :: dvrindex(natpot%ndof,nindex)
  double precision, intent(OUT)          :: coeff_ext(coeff_extdim)

  integer                                :: i,j,m,n,f
  logical                                :: lgrid, lpseudo, lscale

  ! workspace
  double precision, allocatable          :: coeff(:,:)
  integer, allocatable                   :: modcidx(:)
  integer, allocatable                   :: idx(:)
  integer, allocatable                   :: modcgdim(:)
  integer, allocatable                   :: iwork(:,:)
  double precision, allocatable          :: Vn(:), Vn2(:,:)
  double precision, allocatable          :: ovlp(:,:)
#ifdef MPI
  integer                                :: ierr
  double precision, allocatable          :: recbuf(:)
#endif

  ! local dimensions of extended pfit
  integer                                :: potvdim(natpot%nmode)
  integer                                :: potdim(natpot%nmode)
  integer                                :: potndim(natpot%nmode)

  ! extended grid
  double precision, allocatable          :: modccoord(:)
  double precision                       :: coord(natpot%ndof)

  integer                                :: allocstat
  double precision                       :: rwork(nindex)
  double precision                       :: regularization
  ! LAPACK
  character, parameter                   :: UPLO='L'
  integer                                :: INFO,NRHS

  ! tmp
  double precision                       :: asize, mem, rdum, delta
  double precision                       :: ompddot

  integer(long), save                        :: tid=0 ! timer ID


  external operateomega

  delta = 0.d0

  if (mpirank.eq.0) then
     call get_timer(tid, 'prep coeffs')
     call start_timer(tid)
  end if

  nsample = nindex

  if (mpirank.eq.0) then
     write(ilog,'(a)') "Preparing to calculate coefficients:"
     call myflush(ilog)
  end if

  ! accociate the global pointer
  pnatpot => natpot
  pdvrindex => dvrindex

  ! number of grid points in contracted mode
  lgrid = .true.
  if (natpot%modc .ne. 0) then
     if (lmextgrid(natpot%modc) ) then
        NRHS = pointmaps(natpot%modc)%npoint
        lgrid = .false.
     else
        NRHS = natpot%subdim(natpot%modc)
     end if
  else
     ! otherwise one
     NRHS = 1
  end if

  ! local pointers into coeff_ext
  do m=1,natpot%nmode
     potdim(m) = natpot%potdim(m)
  end do

  potvdim(1) = 1
  do m=2,natpot%nmode
     if (m-1.ne.natpot%modc) then
        potvdim(m) =potvdim(m-1)*potdim(m-1)
     else
        potvdim(m) =potvdim(m-1)*pointmaps(m-1)%npoint
     end if
  end do

  potndim(natpot%nmode)=1
  do m=natpot%nmode-1,1,-1
     if (m+1.ne.natpot%modc) then
        potndim(m)= potndim(m+1)*natpot%potdim(m+1)
     else
        potndim(m)= potndim(m+1)*pointmaps(m+1)%npoint
     end if
  end do


  if (natpot%modc .ne. 0) then
     n = natpot%nspfdof(natpot%modc)
  else
     n = 1
  end if

  routine = "calccoeff"

  ! do all the allocations
  allocate(modcidx(n), STAT=allocstat)
  if (allocstat .ne. 0) then
     message = "Unable to allocate 'modcidx'"
     call errormsg
  end if
  allocate(modcgdim(n), STAT=allocstat)
  if (allocstat .ne. 0) then
     message = "Unable to allocate 'modcgdim'"
     call errormsg
  end if
  allocate(idx(natpot%ndof), STAT=allocstat)
  if (allocstat .ne. 0) then
     message = "Unable to allocate 'idx'"
     call errormsg
  end if
  allocate(modccoord(n), STAT=allocstat)
  if (allocstat .ne. 0) then
     message = "Unable to allocate 'modccoord'"
     call errormsg
  end if

  if (invertmethod .eq. mc_inv_pseudo) then
     allocate(Vn2(nsample,nrhs), STAT=allocstat)
     if (allocstat .ne. 0) then
        message = "Unable to allocate 'Vn2(nsample,NRHS)'"
        call errormsg
     end if
  else
     allocate(Vn(nsample), STAT=allocstat)
     if (allocstat .ne. 0) then
        message = "Unable to allocate 'Vn(nsample)'"
        call errormsg
     end if
  end if
  allocate(coeff(natpot%modcdim,NRHS), STAT=allocstat)
  if (allocstat .ne. 0) then
     message = "Unable to allocate 'coeff(coeffdim)'"
     call errormsg
  end if


  if (mpirank.eq.0) then
     write(ilog,'(a)') "Building SPP sampling arrays..."
     call myflush(ilog)
  end if
  ! get the sampling poins from the spp and
  ! put them in an array
  call mkSppSamplings(dvrindex)


  if (invertmethod .eq. mc_inv_spd &
       .or. invertmethod .eq. mc_inv_eigen ) then
     ! direct inversion - we need to build the overlap matrix

     asize = dble(natpot%modcdim)**2
     mem = asize/131072 !MB

     if (mpirank.eq.0) then
        write(ilog,'(a,f13.0,f13.3,a)') "Allocating ovlp matrix, size: ", &
             asize ,mem," MB"
        call myflush(ilog)
     end if

     allocate(ovlp(natpot%modcdim,natpot%modcdim), STAT=allocstat)
     if (allocstat .ne. 0) then
        message = "Unable to allocate ovlp matrix."
        call errormsg
     end if

  end if

  ! the slow but low-mem variant
  if ( .not. lomega .and. (invertmethod .eq. mc_inv_spd &
       .or. invertmethod .eq. mc_inv_eigen) ) then
     if (mpirank.eq.0) then
        write(ilog,'(a)') "Building Overlap matrix..."
        call myflush(ilog)
     end if
     call mcsppoverlap(ovlp)
  else


     ! the faster memory intensive variant: build omega
     ! and transpose(omega) if needed
     !
     ! for all sampling points the value of the
     ! product of the SPP is calculated for all possible
     ! configurations.

     if (lomega) then
        if (mpirank.eq.0) then
           asize = dble(nsample)*dble(natpot%modcdim)
           mem = asize/131072 !MB
           write(ilog,'(a,f13.0,f13.3,a)') "Allocating Omega, size:       ", &
                asize, mem," MB"
           call myflush(ilog)
        end if

        allocate(omega(nsample,natpot%modcdim), stat=allocstat)
        if (allocstat .ne. 0) then
           message = "Unable to allocate omega."
           call errormsg
        end if
     end if

     ! transpose(omega) can only be used in conj-grad
     ! so we do not allocate it otherwise
     if (lomegat .and. (invertmethod .eq. mc_inv_conjgrad) ) then
        if (mpirank.eq.0) then
           write(ilog,'(a,f13.0,f13.3,a)') "Allocating Omega^T, size:     ", &
                asize, mem," MB"
        end if
        allocate(omegaT(natpot%modcdim,nsample), stat=allocstat)
        if (allocstat .ne. 0) then
           message = "Unable to allocate omega transposed."
           call errormsg
        end if
     end if

     ! build omega - and transpose(omega) if needed
     if (mpirank.eq.0) then
        write(ilog,'(a)') "Building Omega matrix..."
        call myflush(ilog)
     end if
     call mkOmega()

     if (invertmethod .eq. mc_inv_spd &
          .or. invertmethod .eq. mc_inv_eigen) then
        if (mpirank.eq.0) then
           write(ilog,'(a)') "Building Overlap matrix..."
           call myflush(ilog)
        end if
        call mcsppoverlapmem(ovlp,natpot%modcdim)
        if (mpirank.eq.0) then
           write(ilog,'(a)') "Done."
           call myflush(ilog)
        end if
     end if

  end if


  ! grid dimensions of the contacted mode - if any
  if (natpot%modc .ne. 0) then
     do n=1,natpot%nspfdof(natpot%modc)
        f = natpot%spfdof(n,natpot%modc)
        modcgdim(n) = natpot%gdim(f)
     end do
  end if


  if (mpirank.eq.0) then
     write(ilog,'(a)') "Calculating omega V"
     call myflush(ilog)
  end if


  ! This operates omega on V
  do i=1,NRHS

     if (lgrid) then ! the traditional no-exteneded-grid method

        ! covert the combined index i into dvrindex of contracted mode
        if (natpot%modc .ne. 0) then
           call cindex2sindex(i,modcidx,modcgdim,natpot%nspfdof(natpot%modc))
!           modcidx = pointmaps(natpot%modc)%indices(:,i)
        end if


        !$omp parallel do private(f,j,idx) schedule(dynamic) if(lomppotential)
        do n=1,nsample
           do f=1,natpot%ndof
              if (natpot%modc .ne. natpot%dofspf(f) ) then
                 idx(f) = dvrindex(f,n)
              else
                 j = natpot%potdofmap(2,f)
                 idx(f) = modcidx(j)
              end if
           end do

           if (invertmethod.eq.mc_inv_pseudo) then
              Vn2(n,i) = potential(idx,natpot%ndof)
           else
              Vn(n) = potential(idx,natpot%ndof)
           end if

        end do
        !$omp end parallel do

     else ! we have an extended grid
        modccoord = pointmaps(natpot%modc)%points(:,i)

        !$omp parallel do private(f,j,coord, idx) schedule(dynamic) if(lomppotential)
        do n=1,nsample

           idx = dvrindex(:,n)
           call dvrindex2coord(natpot,idx,coord)
           do j=1,natpot%nspfdof(natpot%modc)
              f=natpot%spfdof(j,natpot%modc)
              coord(f) = modccoord(j)
           end do

           if (invertmethod.eq.mc_inv_pseudo) then
              Vn2(n,i) = potentialQ(coord,natpot%ndof)
           else
              Vn(n) = potentialQ(coord,natpot%ndof)
           end if


        end do
        !$omp end parallel do

     end if

     ! skip the rest if pseudoinverse
     if (invertmethod .eq. mc_inv_pseudo) cycle

     if (.not.lomega) then

        allocate(iwork(natpot%nmode,2))
        !$omp parallel do private(rwork,rdum,iwork) schedule(dynamic)
        do j=1,natpot%modcdim
           call operateOmega1J(j, rdum , Vn, rwork, iwork(:,1), iwork(:,2))
           coeff(j,i) = rdum
        end do
        !$omp end parallel do

        deallocate(iwork)

     else
        if (natpot%modcdim .gt.1) then
           !$omp parallel do
           do j=1,natpot%modcdim
              coeff(j,i) = dot_product(Vn,Omega(:,j))
           end do
           !$omp end parallel do
        else
           coeff(1,i) = ompddot(nsample,Vn,Omega(:,1))
        end if

     end if

  end do !NRHS


#ifdef MPI

  if (invertmethod .ne. mc_inv_pseudo) then
     allocate(recbuf(natpot%modcdim))

     do i=1,NRHS
        call MPI_Reduce(coeff(:,i),recbuf,natpot%modcdim,&
             MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if (mpirank .eq. 0) then
           coeff(:,i) = recbuf
        end if
     end do

     deallocate(recbuf)
  end if

#endif


  ! free workspace
  if (allocated(modcidx))  deallocate(modcidx)
  if (allocated(modcgdim)) deallocate(modcgdim)


  if (mpirank.eq.0) then
     write(ilog,'(a)') "Solving 'Omega V = Omega Omega**T C' for C."
     call myflush(ilog)
  end if

  if (mpirank.eq.0) then
     call stop_timer(tid)
  end if


  if (invertmethod .eq. mc_inv_spd) then

     ! we don't need omega any more
     if (allocated(omega)) deallocate(omega)
     if (allocated(omegaT)) deallocate(omegaT)


     call mcsolve_spd(ovlp,coeff,natpot%modcdim,NRHS,INFO)

     if (mpirank.eq.0) then
        routine = "calccoeff"
        if (INFO .gt. 0) then
           message = "Overlap matrix is not positive definite." &
                // " Use more sampling points."
           call errormsg
        else if (INFO .lt. 0) then
           message = "Overlap matrix has illegal values."
           call errormsg
        end if
     end if

  elseif (invertmethod .eq. mc_inv_eigen) then

     ! we don't need omega any more
     if (allocated(omega)) deallocate(omega)
     if (allocated(omegaT)) deallocate(omegaT)

     ! calculated while ovlp was built
     regularization = regularization_coeff*ovlptrace

     lpseudo = .false. ! regularize with eval(i) = eval(i) + eps*exp(-eval(i)/eps)
     lscale = .true.   ! scale eps with trace of overlap matrix

     call mcsolve_eigen(ovlp,coeff,regularization,lpseudo, &
          lscale,natpot%modcdim,NRHS,"eigenvalues",INFO)

     if (mpirank.eq.0) then
        routine = "calccoeff"
        if (INFO .gt. 0) then
           write(message,'(a,i8)') "Diagonalization of SPP overlaps failed. Lapack Err code:",INFO
           call errormsg
        else if (INFO .lt. 0) then
           message = "Overlap matrix has illegal values."
           call errormsg
        end if
     end if

  else if (invertmethod .eq. mc_inv_conjgrad) then

     if (NRHS .eq. 1) then
        call conjgradsingle(coeff, natpot%modcdim, &
             cgtolerance, cgmaxiter,INFO)
     else
        call BLKconjgrad(coeff, natpot%modcdim, NRHS, &
             cgtolerance, cgmaxiter,INFO)
     end if

     if (mpirank.eq.0) then
        routine = "calccoeff"
        if (INFO .eq. 1) then
           message = "Overlap matrix is singular. Use more sampling points."
           call  warnmsg
        else if (INFO .eq. 2) then
           message = "Conjugate Gradients did not converge. " &
                // "See log file. Continuing anyways..."
           call warnmsg
           lwarn = .false.
        else if (INFO .eq. -1) then
           message = "Overlap matrix is not positive definite." &
                // " Use more sampling points."
           call warnmsg
        end if

     end if

  else if (invertmethod .eq. mc_inv_pseudo) then

     coeff = 0.d0

     call mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,natpot%modcdim,delta,INFO)

     if (mpirank.eq.0) then
        routine = "calccoeff"
        if (INFO .gt. 0) then
           message = "Omega matrix does not have full rank." &
                // " Use more sampling points."
           call errormsg
        else if (INFO .lt. 0) then
           message = "Omega matrix has illegal values."
           call errormsg
        end if
     end if


  else
     routine = "calccoeff"
     message = "Unknown invert-method. This should not have happened."
     call errormsg
  end if

  if (mpirank.eq.0) then
     write(ilog,'(a)') "Done."

     if (delta.gt.0.d0) then
        delta = dsqrt(delta)
        write(ilog,'(a)') ""
        write(ilog,'(a)')"least squares error (coeff))= SQRT(sum(delta_coeff**2)):"
        write(ilog,'(3(es11.4e2,a,2x))') delta, " au, ", delta*27.21138386d0, &
             " eV, ", delta*2.1947463137d+5, " cm-1"
        write(ilog,'(a)') ""
        write(iout,'(a)') ""
        write(iout,'(a)')"least squares error (coeff))= SQRT(sum(delta_coeff**2)):"
        write(iout,'(3(es11.4e2,a,2x))') delta, " au, ", delta*27.21138386d0, &
             " eV, ", delta*2.1947463137d+5, " cm-1"
        write(iout,'(a)') ""
     end if


     ! the coeff have been stored in a different order
     ! for the invert routines. copy them to the natpot in the
     ! correct natpot order
     if (natpot%modc .ne. 0) then
        m = natpot%modc

        call vng2vgn2(coeff,natpot%coeff, &
             potvdim(m),NRHS,potndim(m), &
             natpot%potvdim(m), natpot%subdim(m),natpot%potndim(m) )

        call vng2vgn(coeff,coeff_ext,potvdim(m), &
             NRHS, potndim(m))

     else
        natpot%coeff = coeff(:,1)
        coeff_ext = coeff(:,1)
     end if
  end if


#ifdef MPI

  ! a final bcast
  call MPI_Bcast(natpot%coeff,natpot%cdim,  &
       MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
  call MPI_Bcast(coeff_ext,coeff_extdim,  &
       MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#endif


  ! free mem and reset pointers
  pnatpot => NULL()
  pdvrindex => NULL()

  if (allocated(ovlp))         deallocate(ovlp)
  if (allocated(omega))        deallocate(omega)
  if (allocated(omegaT))       deallocate(omegaT)
  if (allocated(sppsamplings)) deallocate(sppsamplings)
  if (allocated(sppsamplingst))deallocate(sppsamplingst)
  if (allocated(sppdims))      deallocate(sppdims)
  if (allocated(modcidx))      deallocate(modcidx)
  if (allocated(modcgdim))     deallocate(modcgdim)
  if (allocated(idx))          deallocate(idx)
  if (allocated(modccoord))    deallocate(modccoord)

end subroutine calccoeff




!*******************************************************************
! y = omega**T * omega * x
!*******************************************************************
subroutine operateOmega(x,y,modcdim,NRHS,lskip)
  use pftools
  use mcpfitparams
  use mcoverlap
  use logdat
  implicit none

  integer, intent(IN)             :: modcdim, NRHS
  double precision, intent(IN)    :: x(modcdim, NRHS)
  double precision, intent(OUT)   :: y(modcdim, NRHS)
  logical, intent(IN)             :: lskip(NRHS)


  double precision                :: z(nsample) ! AUTO ALLOC
  double precision                :: rwork(nsample) ! AUTO ALLOC

  integer                         :: i,j,n

  double precision                :: omega1N(modcdim) ! AUTO ALLOC

  integer                         :: lwork
  double precision, allocatable   :: work(:)
  integer                         :: iwork(pnatpot%nmode,2) ! AUTO ALLOC

  integer(long), save                        :: tid=0 ! timer ID

  if (mpirank.eq. 0) then
     call get_timer(tid, 'ovlp * coeff')
     call start_timer(tid)
  endif

  do i=1,NRHS

     if (.not. lskip(i)) then

        ! OPERATION 1

        ! we first operate omegaT because we make use of the
        ! first index running fastest
        if (allocated(omegaT)) then
           ! optimal

           !$omp parallel do schedule(static)
           do n=1,nsample
              z(n) = dot_product(omegaT(:,n), x(:,i))
           end do
           !$omp end parallel do

        elseif (allocated(omega)) then
           !lots of cache misses but usually still better
           !than re-calulating

           !$omp parallel do schedule(dynamic)
           do n=1,nsample
              z(n) = dot_product(omega(n,:), x(:,i))
           end do
           !$omp end parallel do

        else
           ! re-calculate on-the-fly

           lwork = wk_comb_prod(sppdims,pnatpot%npotmodes)
           allocate(work(lwork))

           !$omp parallel do private(work, omega1N) schedule(dynamic)
           do n=1,nsample
              call mkOmega1N(n, omega1N, lwork, work)
              z(n) = dot_product(omega1N, x(:,i))
           end do
           !$omp end parallel do

           deallocate(work)

        end if


        ! OPERATION 2

        if (allocated(omega)) then

           !$omp parallel do schedule(static)
           do j=1,modcdim
              y(j,i) = dot_product(omega(:,j), z)
           end do
           !$omp end parallel do

        else
           !$omp parallel do private(rwork,iwork) schedule(dynamic)
           do j=1,modcdim
              call operateOmega1J(j,y(j,i),z,rwork,iwork(:,1),iwork(:,2))
           end do
           !$omp end parallel do
        end if


     end if ! lskip

  end do

  if (mpirank.eq. 0) call stop_timer(tid)

end subroutine operateOmega


!*******************************************************************
! operateOmega1J
!
! Calulates sum_n Omega(j,n)*x(n) for one j-Index where
! Omega(j,n) is calculated on the fly.
!
! J: The combined index
!*******************************************************************
subroutine operateOmega1J(J,OmegaX,X,rwork,coeffidx,z)
  use pftools
  use mcoverlap
  implicit none
  integer, intent(IN)             :: J
  double precision, intent(IN)    :: X(nsample)
  double precision, intent(OUT)   :: omegaX

  integer, intent(INOUT)          :: coeffidx(pnatpot%nmode)
  integer, intent(INOUT)          :: z(pnatpot%nmode)
  double precision, intent(INOUT) :: rwork(nsample)
  integer                         :: m,n,z1, mlast
  logical                         :: l1st,llastm
  double precision                :: mddot


  call cindex2sindex(J,coeffidx,pnatpot%potdim,pnatpot%nmode)
  z1 = 0
  do m=1, pnatpot%nmode
     if (m .ne. pnatpot%modc) then
        z(m) = z1 + coeffidx(m)
        z1 = z1 + pnatpot%potdim(m)
     end if
  end do
  if(pnatpot%modc .eq. pnatpot%nmode) then
     mlast = pnatpot%nmode -1
  else
     mlast = pnatpot%nmode
  end if


  l1st = .true.
  do m=1, pnatpot%nmode
     llastm = m .eq. mlast
     if (m .ne. pnatpot%modc) then
        if (l1st .and. .not. llastm) then
           ! first multiplication - init rwork
           do n=1,nsample
              rwork(n) = X(n)*sppsamplingst(n,z(m))
           end do
           l1st = .false.
        elseif (llastm .and. .not. l1st) then
           ! last multiplication - sum up
           omegaX = mddot(nsample,rwork,sppsamplingst(:,z(m)))
        elseif (l1st.and.llastm) then
           ! first and last multiplication: SPP in 1 mode only
           ! sub up directly
           omegaX = mddot(nsample,X,sppsamplingst(:,z(m)))
        else
           ! somewhere between first and last mode
           do n=1,nsample
              rwork(n) = rwork(n)*sppsamplingst(n,z(m))
           end do
        end if
     end if
  end do

end subroutine operateOmega1J




subroutine calccoeffcmplte(natpot)
  use mcpfitparams
  use mccomplete
  use natpottype
  use mtlib, only: mttxdd

  implicit none

  type(natpot_t), intent(INOUT)    :: natpot

  integer                               :: restdim,zevm
  double precision, allocatable,target  :: work1(:), work2(:)

  double precision, pointer             :: pwork(:), ppot(:)
  integer                               :: m, dim1, dim2, dim3, dim4



  restdim = natpot%vpotdim
  ppot => vpot

  dim1 = 1
  ! integrate the modes, this can be
  ! optimized by integrating large modes first
  ! see potfit.
  do m=1,natpot%nmode

     zevm = natpot%zevec(m,1)
     restdim = restdim/natpot%subdim(m)

     dim2 = natpot%subdim(m)
     dim3 = restdim
     if (m.ne. natpot%modc) then
        dim4 = natpot%potdim(m)
     else
        dim4 = natpot%subdim(m)
     end if

     if (modulo(m,2) .eq. 0) then
        if (allocated(work2)) deallocate(work2)
        allocate(work2(dim1*dim4*dim3))
        pwork => work2
     else
        if (allocated(work1)) deallocate(work1)
        allocate(work1(dim1*dim4*dim3))
        pwork => work1
     end if


     if (m.ne. natpot%modc) then
        call mttxdd (natpot%evec(zevm),ppot,pwork,dim1,dim2,dim3,dim4)
     else
        pwork = ppot
     end if

     if (modulo(m,2) .eq. 0) then
        ppot => work2
     else
        ppot => work1
     end if

     if (m .eq. natpot%modc) then
        dim1 = dim1 * natpot%subdim(m)
     else
        dim1 = dim1 * natpot%potdim(m)
     end if

  end do

  natpot%coeff = pwork

  if (allocated(work1)) deallocate(work1)
  if (allocated(work2)) deallocate(work2)

end subroutine calccoeffcmplte






