! vng2vgn              - re-order a 3-way tensor
! mddot                - simplified version of BLAS ddot
! ompddot              - simplified version of BLAS ddot, OpenMP version
! GramSchmidtDecompose - rank revealing QR decomposition
! GramSchmidtreconst   - undo the QR decomposition
! InvertMatrix         - unvert a matrix using LAPACK dsyev


!*******************************************************************
! copy vecin(vdim,ndim,gdim) to vecout(vdim,gdim,ndim)
!*******************************************************************
subroutine vng2vgn(vecin,vecout,vdim,gdim,ndim)
  implicit none
  integer, intent(IN)           :: vdim,gdim,ndim
  double precision, intent(IN)  :: vecin(vdim,ndim,gdim)
  double precision, intent(OUT) :: vecout(vdim,gdim,ndim)
  integer                       :: v,g,n

  !$omp parallel do private(g,v)
  do n=1,ndim
     do g=1,gdim
        do v=1,vdim
           vecout(v,g,n) = vecin(v,n,g)
        end do
     end do
  end do
  !$omp end parallel do

end subroutine vng2vgn

!*******************************************************************
! copy vecin(vdim,gdim,ndim) to vecout(gdim,vdim,ndim)
!*******************************************************************
subroutine vgn2gvn(vecin,vecout,vdim,gdim,ndim)
  implicit none
  integer, intent(IN)           :: vdim,gdim,ndim
  double precision, intent(IN)  :: vecin(vdim,gdim,ndim)
  double precision, intent(OUT) :: vecout(gdim,vdim,ndim)
  integer                       :: v,g,n


  !$omp parallel do private(g,v)
  do n=1,ndim
     do g=1,gdim
        do v=1,vdim
           vecout(g,v,n) = vecin(v,g,n)
        end do
     end do
  end do
  !$omp end parallel do


end subroutine vgn2gvn




!*******************************************************************
! copy vecin(gdim,vdim,ndim) to vecout(vdim,gdim,ndim)
!*******************************************************************
subroutine gvn2vgn(vecin,vecout,vdim,gdim,ndim)
  implicit none
  integer, intent(IN)           :: vdim,gdim,ndim
  double precision, intent(IN)  :: vecin(gdim,vdim,ndim)
  double precision, intent(OUT) :: vecout(vdim,gdim,ndim)
  integer                       :: v,g,n

  !$omp parallel do private(g,v)
  do n=1,ndim
     do g=1,gdim
        do v=1,vdim
           vecout(v,g,n) = vecin(g,v,n)
        end do
     end do
  end do
  !$omp end parallel do


end subroutine gvn2vgn




!*******************************************************************
! copy vecin(vdim,ndim,gdim1) to vecout(vdim,gdim2,ndim)
! copies only min(gdim1,gdim2) points of gdim
!*******************************************************************
subroutine vng2vgn2(vecin,vecout,vdimi,gdimi,ndimi, vdimo,gdimo,ndimo)
  implicit none
  integer, intent(IN)           :: vdimi,gdimi,ndimi
  integer, intent(IN)           :: vdimo,gdimo,ndimo
  double precision, intent(IN)  :: vecin(vdimi,ndimi,gdimi)
  double precision, intent(OUT) :: vecout(vdimo,gdimo,ndimo)
  integer                       :: v,g,n

  !$omp parallel do private(g,v)
  do n=1,ndimo
     do g=1,min(gdimi,gdimo)
        do v=1,vdimo
           vecout(v,g,n) = vecin(v,n,g)
        end do
     end do
  end do
  !$omp end parallel do

end subroutine vng2vgn2


!*******************************************************************
! Less general BLAS ddot (with increments == 1)
!*******************************************************************
double precision function mddot(N,X,Y)
  integer, intent(IN)          :: N
  double precision, intent(IN) :: X(N), Y(N)
  double precision             :: dtemp
  integer                      :: m,i,mp1

  if (N.le.0) then
     mddot = 0.d0
     return
  end if

  dtemp = 0.d0

  m = modulo(N,5)
  if (m.ne.0) then
     do i = 1,m
        dtemp = dtemp + X(i)*Y(i)
     end do
     if (n.lt.5) then
        mddot=dtemp
        return
     end if
  end if

  mp1 = m + 1

  do i = mp1,N,5
     dtemp = dtemp + X(i)*Y(i) + X(i+1)*Y(i+1) &
          + X(i+2)*Y(i+2) + X(i+3)*Y(i+3) + X(i+4)*Y(i+4)
  end do


  mddot = dtemp
  return

end function mddot


!*******************************************************************
! Less general BLAS ddot - OpenMP version(with increments == 1)
!*******************************************************************
double precision function ompddot(N,X,Y)
  integer, intent(IN)          :: N
  double precision, intent(IN) :: X(N), Y(N)
  double precision             :: dtemp
  integer                      :: m,i,mp1

  if (N.le.0) then
     ompddot = 0.d0
     return
  end if

  dtemp = 0.d0

  m = modulo(N,5)
  if (m.ne.0) then
     do i = 1,m
        dtemp = dtemp + X(i)*Y(i)
     end do
     if (n.lt.5) then
        ompddot=dtemp
        return
     end if
  end if

  mp1 = m + 1

  !$omp parallel do reduction(+:dtemp)
  do i = mp1,N,5
     dtemp = dtemp + X(i)*Y(i) + X(i+1)*Y(i+1) &
          + X(i+2)*Y(i+2) + X(i+3)*Y(i+3) + X(i+4)*Y(i+4)
  end do
  !$omp end parallel do

  ompddot = dtemp
  return

end function ompddot


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! aortho(V, AV, N,M)
!
! In-situ A-Orthogonalization of V and AV
! where V(:,i) is a set of vectors and AV(:,i) is
! a set of vectors AV(:,i) = matmul(A,V(:,i))
!
! Results in vectors V(:,i), AV(:,i) such that
!
! dot_procuct(V(:,i), AV(:,j)) == delta_ij
!
! A needs to be positive definite.
!
! V and AV must have full rank , otherwise
! linear dependent columns are nullified
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine aortho(V, AV, N,M,ierr)
  implicit none
  integer, intent(IN)             :: N,M
  double precision, intent(INOUT) :: V(N,M), AV(N,M)
  integer, intent(OUT)            :: ierr
  integer                         :: k,j,i
  double precision                :: c,w


  ierr = 0
  do k=1,M-1

     c = 0.d0
     !$omp parallel do reduction(+:c)
     do j=1,N
        c = c + V(j,k) * AV(j,k)
     end do
     !$omp end parallel do

     if (c.lt.0) then
        ! not positive definite
        ierr = -1
        return
     else if (abs(c) .lt. 1.d-12) then
        ierr = ierr + 1
        c = 0.d0
     else
        c = 1.d0/dsqrt(c)
     end if

     !$omp parallel do
     do j=1,N
        V(j,k) = c*V(j,k)
     end do
     !$omp end parallel do
     !$omp parallel do
     do j=1,N
        AV(j,k) = c*AV(j,k)
     end do
     !$omp end parallel do

     do j=k+1,M
        w = 0.d0
        !$omp parallel do reduction(+:w)
        do i=1,N
           w = w + V(i,k) * AV(i,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do i=1,N
           V(i,j) = V(i,j) - V(i,k)*w
        end do
        !$omp end parallel do
        !$omp parallel do
        do i=1,N
           AV(i,j) = AV(i,j) - AV(i,k)*w
        end do
        !$omp end parallel do
     end do

     do j=k+1,M
        w = 0.d0
        !$omp parallel do reduction(+:w)
        do i=1,N
           w = w + V(i,k) * AV(i,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do i=1,N
           V(i,j) = V(i,j) - V(i,k)*w
        end do
        !$omp end parallel do
        !$omp parallel do
        do i=1,N
           AV(i,j) = AV(i,j) - AV(i,k)*w
        end do
        !$omp end parallel do
     end do

  enddo

  c = 0.d0
  !$omp parallel do reduction(+:c)
  do j=1,N
     c = c + V(j,M) * AV(j,M)
  end do
  !$omp end parallel do


  if (c.lt.0) then
     ! not positive definite
     ierr = -1
     return
  else if (abs(c) .lt. 1.d-12) then
     ierr = ierr + 1
     c = 0.d0
  else
     c = 1.d0/dsqrt(c)
  end if


  !$omp parallel do
  do j=1,N
     V(j,M) = c*V(j,M)
  end do
  !$omp end parallel do
  !$omp parallel do
  do j=1,N
     AV(j,M) = c*AV(j,M)
  end do
  !$omp end parallel do

end subroutine aortho




! A simple gram-schmidt-based QR-Decomposition
! Requires M <= N,
! filling with random vecs if A has not full rank (rank(A)==M).
subroutine qr(A,Q,R,N,M)
  implicit none
  integer, intent(IN)           :: N,M
  double precision, intent(IN)  :: A(N,M)
  double precision, intent(OUT) :: Q(N,M), R(M,M)
  double precision              :: qtmp(N), c
  integer                       :: i,j,k

  Q = 0.d0
  R = 0.d0

  do i=1,M
     !$omp parallel do
     do k=1,N
        qtmp(k) = A(k,i)
     end do
     !$omp end parallel do

     do j=1,i-1

        c = 0.d0
        !$omp parallel do reduction(+:c)
        do k=1,N
           c = c + qtmp(k)*Q(k,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do k=1,N
           qtmp(k) = qtmp(k) - c*Q(k,j)
        end do
        !$omp end parallel do
        R(j,i) = c
     end do

     do j=1,i-1

        c = 0.d0
        !$omp parallel do reduction(+:c)
        do k=1,N
           c = c + qtmp(k)*Q(k,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do k=1,N
           qtmp(k) = qtmp(k) - c*Q(k,j)
        end do
        !$omp end parallel do
        R(j,i) = R(j,i) + c
     end do

     c = 0.d0
     !$omp parallel do reduction(+:c)
     do k=1,N
        c = c + qtmp(k)**2
     end do
     !$omp end parallel do

     ! fill with artificial random vectors
     ! if A is not of full rank
     filling: if (c.lt. 1.d-12) then

        call random_number(qtmp)

        do j=1,i-1
           c = 0.d0
           !$omp parallel do reduction(+:c)
           do k=1,N
              c = c + qtmp(k)*Q(k,j)
           end do
           !$omp end parallel do
           !$omp parallel do
           do k=1,N
              qtmp(k) = qtmp(k) - c*Q(k,j)
           end do
           !$omp end parallel do
        end do

        do j=1,i-1
           c = 0.d0
           !$omp parallel do reduction(+:c)
           do k=1,N
              c = c + qtmp(k)*Q(k,j)
           end do
           !$omp end parallel do
           !$omp parallel do
           do k=1,N
              qtmp(k) = qtmp(k) - c*Q(k,j)
           end do
           !$omp end parallel do
        end do

        c = 0.d0
        !$omp parallel do reduction(+:c)
        do k=1,N
           c = c + qtmp(k)**2
        end do
        !$omp end parallel do

        if (c.lt. 1.d-12) then
           ! the random vector is orthotogonal to all
           ! previous vectors. This is unlikely, unless
           ! the previous vectors are a complete set.
           write(*,*) "Error: QR decomposition failed"
           stop 1
        end if
        c = dsqrt(c)
        Q(:,i) = qtmp/c
        ! the random vector is not from the
        ! original set - set is length to zero
        R(i,i)= 0.d0


     else

        c = dsqrt(c)
        R(i,i) = c
        Q(:,i) = qtmp/c

     end if filling

  end do


end subroutine qr

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! GramSchmidtDecompose(A,Q,R,P,N,M,K,rank,eps,ifail)
!
! Decompose the N by M matrix A into a set of linear
! independent  vectors Q and coefficients R,
! thereby  the rank of the matrix is determined.
!
! On exit, Q(:,1:rank) contain the linear independent
! column vectors of A and Q(:,rank+1:) = 0 such that
! On exit, the coefficients for he i'th column of A
! are found in the P(i)'th column of R
!
!
! Intut
!  N     integer   First dimension of A
!  M     integer   Second dimension of A
!  K     integer   min(N,M)
!  A     real*8    Matrix
!  eps   real*8    Vectors with L2 norm less then eps
!                  after beeing orthogonalized
!                  are considered linear dependent.
!
! Output
!  Q     real*8    Linear independent vectors
!  R     real*8    coefficent vectors
!  P     integer   Column interchange index
!  rank  integer   Rank of A
!  ifail integer   Error flag
!                   0: success
!                   1: orthogonalization failed
!                      or is probably incorrect
!                   2: Fatal: K is to small.
!                  -1: epsilon is <= 0
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine GramSchmidtDecompose(A,Q,R,P,N,M,K,rank,eps,ifail)
  implicit none
  integer, intent(IN)                    :: M,N
  double precision, intent(IN)           :: A(N,M), eps
  double precision, intent(OUT)          :: Q(N,K), R(M,K)
  integer, intent(OUT)                   :: P(M), rank, ifail

  integer                                :: i,j,nlin,k
  double precision                       :: c, qtmp(N), rtmp(K)

  if (eps .lt.0.d0) then
     ifail = -1
     return
  end if

  if (K .lt. min(N,M)) then
     ifail = 2
     return
  end if

  Q = 0.d0
  R = 0.d0

  rank = 0
  nlin = 0
  ifail = 0

  do i=1,M

     do k=1,N
        qtmp(k) = A(k,i)
     end do

     do j=1,rank

        c = 0.d0
        !$omp parallel do reduction(+:c)
        do k=1,N
           c = c + qtmp(k)*Q(k,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do k=1,N
           qtmp(k) = qtmp(k) - c*Q(k,j)
        end do
        !$omp end parallel do
        rtmp(j) = c
     end do

     do j=1,rank

        c = 0.d0
        !$omp parallel do reduction(+:c)
        do k=1,N
           c = c + qtmp(k)*Q(k,j)
        end do
        !$omp end parallel do
        !$omp parallel do
        do k=1,N
           qtmp(k) = qtmp(k) - c*Q(k,j)
        end do
        !$omp end parallel do
        rtmp(j) = rtmp(j) + c
     end do

     c = 0.d0
     !$omp parallel do reduction(+:c)
     do k=1,N
        c = c + qtmp(k)*qtmp(k)
     end do
     !$omp end parallel do
     c = dsqrt(c)

     if (c .ge. eps ) then

        if (rank .ge. min(M,N)) then
           ifail = 1
           P(i) = M-nlin
           nlin = nlin+1
        else
           rank = rank+1
           rtmp(rank) = c
           !$omp parallel do
           do k=1,N
              Q(k,rank) = qtmp(k)/c
           end do
           !$omp end parallel do
           P(i) = rank
        end if
     else
        P(i) = M-nlin
        nlin = nlin+1
     end if


     do j=1,rank
        R(P(i),j) = rtmp(j)
     end do

     if (rank .eq. 0) then
        ifail = 1
     end if

  end do


end subroutine GramSchmidtDecompose

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! GramSchmidtReconst(A,Q,R,P,N,M,rank)
!
! Reconstruct the N by M matrix A which has been
! decomposed with routine GramSchmidtDecompose
! using the matrices Q and R, and the vector P.
!
!
! Input
!  N     integer   First dimension of A
!  M     integer   Second dimension of A
!  rank  integer   Second dimension of Q
!  Q     real*8    Matrix of rank linear
!                  independent vectors
!  R     real*8    Coefficients
!  P     integer   Index interchanges
!
! Output
!  A     real*8    Matrix A
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine GramSchmidtReconst(A,Q,R,P,N,M,rank)
  implicit none
  integer, intent(IN)                    :: M,N,rank
  double precision, intent(OUT)          :: A(N,M)
  double precision, intent(IN)           :: Q(N,rank), R(M,rank)
  integer, intent(IN)                    :: P(M)

  integer                                :: j,k,l
  !$omp parallel do private(l,k)
  do j=1,M
     l = P(j)
     A(:,j) =  Q(:,1)*R(l,1)
     do k=2,rank
        A(:,j) =  A(:,j) +  Q(:,k)*R(l,k)
     end do
  end do
  !$omp end parallel do

end subroutine GramSchmidtReconst


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! trinangleIndex(k,r,r1)
!  r->
!  1  2   4   7    r1
!     3   5   8    |
!         6   9    v
!             10
! retunrns index r,r1 for given index k
! for instace, k=9 above will return r=4,r1=3

subroutine triangleIndex(k,r,r1)
  implicit none
  integer, intent(IN)   :: k
  integer, intent(OUT)  :: r,r1
  double precision      :: d


  d = dsqrt(dble(k*2) + 0.25d0) - 0.5d0

  r = int(d)

  if (r*(r+1) .eq. 2*k) then
     r1 = r
  else
     r1 = k - (r*(r+1))/2
     r = r + 1

     if (r*(r-1)/2 + r1 .ne. k) then
        print*, "Error in subroutine trinangleIndex(k,r,r1)"
        print*, k,r,r1
        stop 1
     end if

  end if

end subroutine triangleIndex


