!   -*- f90 -*-

module mcpmpitools

    use mpidata
    use global
    use logdat
#ifdef MPI
    use mpi
#endif

  implicit none

  integer, allocatable          :: mpi_chunksizes(:)
  integer, allocatable          :: mpi_chunkstart(:)

contains


!*********************************************************************
! mpi_set_chunksize: set the number of sampling poins in each
! mpi rank. Assumes that nsample is known in rank 0 only.
! sets nindex in each rank to the size of its chunk
!
! MS 06/2016
!*********************************************************************

  subroutine mpi_set_chunksize(nsample,nindex)
    implicit none

    integer, intent(IN)    :: nsample
    integer, intent(OUT)   :: nindex
    integer                :: nbase, nleft
    integer                :: i,ierr
    character(len=8)       :: buf

    if (.not. allocated(mpi_chunksizes)) then
       allocate(mpi_chunksizes(0:nmpi-1))
    end if
    if (.not. allocated(mpi_chunkstart)) then
       allocate(mpi_chunkstart(0:nmpi-1))
    end if


    mpi_chunkstart(0) = 1
    if (nmpi .le. 1) then
       nindex = nsample
       mpi_chunksizes(0) = nsample
       return
    end if

#ifdef MPI

    if (mpirank .eq. 0) then

       nbase = nsample/nmpi
       nleft = modulo(nsample,nmpi)
       mpi_chunksizes = nbase

       do i=0,nleft-1
          mpi_chunksizes(i) =  mpi_chunksizes(i) + 1
       end do

    end if

    call MPI_BCAST(mpi_chunksizes,nmpi,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    if (ierr.ne.0) then
       write(buf,*) ierr
       buf=adjustl(buf)
       routine = "mpi_set_chunksize"
       message = "MPI_BCAST failed with error code "//trim(buf)
       call errormsg
    end if

    nindex = mpi_chunksizes(mpirank)

    do i=1,nmpi-1
       mpi_chunkstart(i) =  mpi_chunkstart(i-1) + mpi_chunksizes(i-1)
    end do

#endif

  end subroutine mpi_set_chunksize


!*********************************************************************
! mpi_send_chunk: send a chunck of sampling points (as grid
! indices) from rank 0 to the target rank. The target rank
! needs to know already the size of its chunk. See subroutine
! mpi_set_chunksize. nsample must be set correctly
!
! MS 06/2016
!*********************************************************************
  subroutine mpi_send_chunk(tmpindex,dvrindex,ndof,nsample,targetrank)

    use global
    use mpidata
#ifdef MPI
    use mpi
#endif

    implicit none

    integer, intent(IN)    :: ndof,nsample,targetrank
    integer, intent(IN)    :: tmpindex(ndof,nsample)
    integer, intent(OUT)   :: dvrindex(ndof,nsample)
    integer                :: mpitag, ierr, i,j
    character(len=8)       :: buf, code


#ifdef MPI
    integer                :: status(mpi_status_size)
#endif

    ! exit here if we are neither sender nor reciever
    if (mpirank .ne. 0 .and. mpirank .ne. targetrank) then
       return
    end if

    ! nsample is known in all ranks
    if (nsample .le. 0) then
       write(buf,*) mpirank
       buf=adjustl(buf)
       routine = "mpi_send_chunk (sending samplings)"
       message = "Sampling size is .le. 0 in rank "//trim(buf)
       call errormsg
    end if

    if ( nmpi .eq. 1 .or. (mpirank .eq. 0 .and. targetrank .eq. 0)) then
       !$omp parallel do private(i)
       do j=1,nsample
          do i=1,ndof
             dvrindex(i,j) = tmpindex(i,j)
          end do
       end do
       !$omp end parallel do
       return
    end if


#ifdef MPI

    mpitag=101 + targetrank

    if (mpirank .eq. 0) then

       call MPI_Send(tmpindex, ndof*nsample, &
            MPI_INTEGER,targetrank,mpitag,MPI_COMM_WORLD,ierr)

       if (ierr .ne. 0) then
          write(buf,*) targetrank
          buf=adjustl(buf)
          write(code,*) ierr
          routine = "mpi_send_chunk (sending samplings)"
          message = "MPI_SEND failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    else

       call MPI_Recv(dvrindex,ndof*nsample,MPI_INTEGER, &
            0,mpitag, MPI_COMM_WORLD,status,ierr)

       if (ierr .ne. 0) then
          write(buf,*) mpirank
          buf=adjustl(buf)
          write(code,*) ierr
          code=adjustl(code)
          routine = "mpi_send_chunk (sending samplings)"
          message = "MPI_RECV failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    end if

#endif

  end subroutine mpi_send_chunk



!*********************************************************************
! mpi_get_chunk: recieve a chunck of sampling points (as grid
! indices) from targetrank into rank 0. The target rank and rank 0
! need to know already the size of the chunk. See subroutine
! mpi_set_chunksize.
!
! MS 08/2016
!*********************************************************************
  subroutine mpi_get_chunk(tmpindex,dvrindex,ndof,nsample,targetrank)
    implicit none

    integer, intent(IN)    :: ndof,nsample,targetrank
    integer, intent(IN)    :: tmpindex(ndof,nsample)
    integer, intent(OUT)   :: dvrindex(ndof,nsample)
    integer                :: mpitag, ierr, i,j
    character(len=8)       :: buf, code


#ifdef MPI
    integer                :: status(mpi_status_size)
#endif

    ! exit here if we are neither sender nor reciever
    if (mpirank .ne. 0 .and. mpirank .ne. targetrank) then
       return
    end if

    ! nsample is known in all ranks
    if (nsample .le. 0) then
       write(buf,*) mpirank
       buf=adjustl(buf)
       routine = "mpi_get_chunk (recieving samplings)"
       message = "Sampling size is .le. 0 in rank "//trim(buf)
       call errormsg
    end if

    if ( nmpi .eq. 1 .or. (mpirank .eq. 0 .and. targetrank .eq. 0)) then
       !$omp parallel do private(i)
       do j=1,nsample
          do i=1,ndof
             dvrindex(i,j) = tmpindex(i,j)
          end do
       end do
       !$omp end parallel do
       return
    end if


#ifdef MPI

    mpitag=201 + targetrank


    if (mpirank .eq. targetrank) then

       call MPI_Send(tmpindex, ndof*nsample, &
            MPI_INTEGER,0,mpitag,MPI_COMM_WORLD,ierr)

       if (ierr .ne. 0) then
          write(buf,*) targetrank
          buf=adjustl(buf)
          write(code,*) ierr
          routine = "mpi_get_chunk (sending samplings)"
          message = "MPI_SEND failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    else

       call MPI_Recv(dvrindex,ndof*nsample,MPI_INTEGER, &
            targetrank,mpitag, MPI_COMM_WORLD,status,ierr)

       if (ierr .ne. 0) then
          write(buf,*) mpirank
          buf=adjustl(buf)
          write(code,*) ierr
          code=adjustl(code)
          routine = "mpi_get_chunk (sending samplings)"
          message = "MPI_RECV failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    end if

#endif

  end subroutine mpi_get_chunk



!*********************************************************************
! mpi_get_energies: recieve a chunck of sampling energies
! from targetrank into rank 0. The target rank and rank 0
! need to know already the size of the chunk. See subroutine
! mpi_set_chunksize.
!
! MS 08/2016
!*********************************************************************
  subroutine mpi_get_energies(sendbuf,recbuf,nsample,targetrank)
    implicit none

    integer, intent(IN)    :: nsample,targetrank
    real*8, intent(IN)     :: sendbuf(nsample)
    real*8, intent(OUT)    :: recbuf(nsample)
    integer                :: mpitag, ierr, j
    character(len=8)       :: buf, code

#ifdef MPI
    integer                :: status(mpi_status_size)
#endif

    ! exit here if we are neither sender nor reciever
    if (mpirank .ne. 0 .and. mpirank .ne. targetrank) then
       return
    end if

    ! nsample is known in all ranks
    if (nsample .le. 0) then
       write(buf,*) mpirank
       buf=adjustl(buf)
       routine = "mpi_get_energies (recieving samplings)"
       message = "Sampling size is .le. 0 in rank "//trim(buf)
       call errormsg
    end if

    if ( nmpi .eq. 1 .or. (mpirank .eq. 0 .and. targetrank .eq. 0)) then
       !$omp parallel do
       do j=1,nsample
          recbuf(j) = sendbuf(j)
       end do
       !$omp end parallel do
       return
    end if


#ifdef MPI

    mpitag=301 + targetrank


    if (mpirank .eq. targetrank) then

       call MPI_Send(sendbuf, nsample, &
            MPI_DOUBLE_PRECISION,0,mpitag,MPI_COMM_WORLD,ierr)

       if (ierr .ne. 0) then
          write(buf,*) targetrank
          buf=adjustl(buf)
          write(code,*) ierr
          routine = "mpi_get_energies (sending samplings)"
          message = "MPI_SEND failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    else

       call MPI_Recv(recbuf,nsample,MPI_DOUBLE_PRECISION, &
            targetrank,mpitag, MPI_COMM_WORLD,status,ierr)

       if (ierr .ne. 0) then
          write(buf,*) mpirank
          buf=adjustl(buf)
          write(code,*) ierr
          code=adjustl(code)
          routine = "mpi_get_energies (sending samplings)"
          message = "MPI_RECV failed for rank "//trim(buf)// &
               " with code "//trim(code)
          call errormsg
       end if
    end if

#endif

  end subroutine mpi_get_energies


end module mcpmpitools
