!***********************************************************************
!
!                        MCPOTFIT
!
!  Program package to calculate natural potential fits.
!  using Monte-Carlo samplings
!  Any similarities to the MCTDH package are NOT accidental.
!
!
!   10/13 MS
!
!
!***********************************************************************

      program mcpotfit

use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use timing
use combined
use channels
use mpidata
use logdat
use pdefaultmod
use peingutil, only: pclosedat
use logdvrmod
#ifdef MPI
   use mpi
#endif
use versions

      implicit none


!-----------------------------------------------------------------------
! Define variables
!-----------------------------------------------------------------------
      real(dop)       :: tcpu,tsys,ttot,tcpu0,tsys0
      integer(long)   :: icount,count_rate
      integer(long)   :: ierror

      logical(kind=4) :: linperr

      integer(long) ::  t1,t2

!-----------------------------------------------------------------------
! Get the time when the job is started, set "progname".
!-----------------------------------------------------------------------
      call fpstartup(cstartup)
      call mygetarg(0,progname)
      lprogname=index(progname,' ')-1
      call rempath(progname,lprogname,ierror)
      if(ierror.eq.1) then
         routine='mcpotfit'
         message=progname//' is a directory, not a file name.'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Define default values and initialize variables
!-----------------------------------------------------------------------
      call default
      call pdefault
      call mcdefault
      lmcpf = .true.

!-----------------------------------------------------------------------
! Initialize MPI
!-----------------------------------------------------------------------
#ifdef MPI
      lmpi = .true.
      call mpi_init(mpierr)
      call mpi_comm_size(MPI_COMM_WORLD,nmpi,mpierr)
      call mpi_comm_rank(MPI_COMM_WORLD,mpirank,mpierr)
      if (mpirank.gt.0) then
         call sleep(1)
         call pdefault  
         goto 999
      end if

#else
      nmpi=1
      mpirank=0
#endif

!-----------------------------------------------------------------------
! Open 'inname.inp' and 'name/log' file; Get data from file 'inname.inp'
!-----------------------------------------------------------------------
      linperr = .false.
      call mcpeingabe(linperr)

!-----------------------------------------------------------------------
! Get memory sizes
!-----------------------------------------------------------------------
      call rdmemdim(iin)

!-----------------------------------------------------------------------
! allocate memory for modules used
!-----------------------------------------------------------------------
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_combined
      call alloc_runpfitmod

!-----------------------------------------------------------------------
! if error in input, stop now
!-----------------------------------------------------------------------
      if(ndof.eq.1) then
         routine = 'MCPOTFIT (file.inp)'
         message = 'You are trying to potfit a 1D potential???       '//&
      '                      This is a deep misunderstanding of POTFIT!'
         call errormsg
      elseif(npmode.eq.1.) then
         routine = 'MCPOTFIT (file.inp)'
         message = 'You are trying to potfit a one-mode potential??? '//&
      '                      This is a deep misunderstanding of POTFIT!'
         call errormsg
      endif

      if (linperr) then
         routine = 'MCPOTFIT (file.inp)'
         message = 'Finished with input error. See log file.'
         call errormsg
      endif

      call log_parallelization(ilog)

!-----------------------------------------------------------------------
! pes timing makes only sense if timing is set
!-----------------------------------------------------------------------
      if (.not.ltime) npest=0

!-----------------------------------------------------------------------
! If dvr file exists, get dvr data required for rest of program.
! Else build dvr file
!-----------------------------------------------------------------------
      if (irundvr .eq. 1) then
         call rundvr
      else if (irundvr .eq. 2) then
         call logdvrdat
      endif

!-----------------------------------------------------------------------
! Initialise timing arrays before fitting
!-----------------------------------------------------------------------
      call timinit

      ! mpirank 0 only
      if (lpesfile) then
         call mcpfmemhops()
      end if


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! MCPOTFIT
! the routine runmcpfit (no caller - mem allocated dynamically as needed)
 999  continue

#ifdef MPI
      call MPI_BCast(ldvronly,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
#endif

      if (.not. ldvronly) then
         call runmcpfit()
      end if

      ! skip output if we are not the mpi main rank
      if (mpirank.gt.0) goto 996
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!-----------------------------------------------------------------------
!     Get overall timing and write data to "log" file
!-----------------------------------------------------------------------
      call timout
      call cpu_time(tcpu)
      call system_clock(icount,count_rate)
      tsys = dble(icount)/dble(count_rate)
      tcpu = tcpu-tcpu0
      tsys = tsys-tsys0
      ttot = tsys

      ttot = tcpu
      t1   = int(ttot/3600)
      ttot = ttot - dble(3600*t1)
      t2   = int(ttot/60)
      ttot = ttot - dble(60*t2)

      write(ilog,'(/,a,/)') '***** Overall Timing *****'
      write(iout,'(/,a,/)') '***** Overall Timing *****'
      write(ilog,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
      write(ilog, 900) 'Total time     [h:m:s] : ',t1,t2,ttot
      write(iout, 900) 'Total time     [h:m:s] : ',t1,t2,ttot

      write(ilog,'(/,a,/)') '***** God''s final message *****'
      write(iout,'(/,a,/)') '***** God''s final message *****'
      if (.not. lstop) then
         write(ilog,'(a,/)') 'Program terminated normally.'
         write(iout,'(a,/)') 'Program terminated normally.'
      else
         write(ilog,'(a)') 'Program was stopped.'
         write(ilog,'(a)') 'The stop-command reads:'
         write(ilog,'(a,/)')  trim(stopcommand)
         write(iout,'(a)') 'Program was stopped.'
         write(iout,'(a)') 'The stop-command reads:'
         write(iout,'(a,/)')  trim(stopcommand)
      endif

!-----------------------------------------------------------------------
! Close remaining open files
!-----------------------------------------------------------------------
      call pclosedat

!-----------------------------------------------------------------------
! Shut down MPI.
!-----------------------------------------------------------------------
 996  continue
#ifdef MPI
      call mpi_finalize(mpierr)
#endif

      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_operdef
      call dealloc_combined
      call dealloc_runpfitmod

  900 format( a,i4,' :',i3,' :',f5.1)

      end program mcpotfit
