!   -*- f90 -*-

! initialize and build the preconditioner
subroutine init_mcprecond(lpcwarn)

  use mcprecond_m
  use pftools
  use mcoverlap
  use mcpfitparams
  use mpidata
  use channels
  use global
#ifdef MPI
  use mpi
#endif

  implicit none

  logical, intent(OUT)           :: lpcwarn
  integer                        :: m1,m2,j,j1,j2,m,i,k,k1,k2
  integer                        :: ierr
  integer                        :: LDAB,nmode1
  double precision               :: Ojj
  double precision, allocatable  :: rwork(:,:), rwork2(:,:), rwork3(:,:,:)
  integer, allocatable           :: potdim(:), pdim(:),pidx(:)

#ifdef MPI
  double precision,allocatable   :: mpibuf(:), mpibuf2(:,:), mpibuf3(:,:,:)
#endif

#ifdef DEBUG
  call chkNpotptr("init_conjgrad")
#endif

  call free_mcprecond()
  lpcwarn = .false.

   if (cg_precon_type .eq. cg_precon_none) then
     ! no preconcitioner
     return
  end if

  if (mpirank.eq.0) then
     write(ilog,'(a)') " Preparing the pre-conditioner"
  end if

  if (.not. allocated(omega)) then
     allocate(rwork(nsample,2))
  end if

  ! use the diagonal of omega omega**T (Jacobi preconditionier)
  if (cg_precon_type .eq. cg_precon_diag) then

     pcdim = pnatpot%modcdim
     allocate(precon(pcdim))

     do j=1,pnatpot%modcdim
        if (allocated(omega)) then
           call mcsppoverlapMtxElementMem(j,j,Ojj)
        else
           call mcsppoverlapMtxElement(j,j,Ojj,rwork)
        end if
        precon(j) = Ojj
     end do

#ifdef MPI
     if (mpirank .eq. 0) then
        allocate(mpibuf(pcdim))
     end if
     call MPI_Reduce(precon,mpibuf,pcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank .eq. 0) then
        precon = mpibuf
        deallocate(mpibuf)
     else
        deallocate(precon)
     end if
#endif

     if (mpirank .eq. 0) then

        !$omp parallel do reduction(.or.: lpcwarn)
        do j=1,pnatpot%modcdim
           if (precon(j) .lt. 1.d-16) then
              lpcwarn = .true.
              precon(j) = 1.d16
           else
              precon(j) = 1.d0/precon(j)
           end if
        end do
        !$omp end parallel do
     end if


  else if (cg_precon_type .eq. cg_precon_band) then
     ! Band matrix

     LDAB = cg_precon_size+1
     pcdim = LDAB*pnatpot%modcdim
     allocate(precon2(LDAB,pnatpot%modcdim))

     precon2 = 0.d0

     !$omp parallel do &
     !$omp private(j1,j2,Ojj,m1,m2,rwork) &
     !$omp schedule(dynamic,1)
     do j2=1,pnatpot%modcdim
        do j1=max(j2-cg_precon_size,1),j2

           if (allocated(omega)) then
              call mcsppoverlapMtxElementMem(j1,j2,Ojj)
           else
              call mcsppoverlapMtxElement(j1,j2,Ojj,rwork)
           end if
           m1 = cg_precon_size+1+j1-j2
           m2 = j2

           precon2(m1,m2) = Ojj
        end do
     end do
     !$omp end parallel do

#ifdef MPI
     if (mpirank .eq. 0) then
        allocate(mpibuf2(LDAB,pnatpot%modcdim))
     end if

     call MPI_Reduce(precon2,mpibuf2,pcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank .eq. 0) then
        precon2 = mpibuf2
        deallocate(mpibuf2)
     else
        deallocate(precon2)
     end if
#endif

     if (mpirank.eq.0) then

        ! build a cholesky decomposition of the band matrix
        call DPBTRF( "U", pnatpot%modcdim , cg_precon_size, precon2, LDAB, ierr)
        if (ierr.ne.0) lpcwarn = .false.
     end if



  else if (cg_precon_type .eq. cg_precon_sparse) then


     if (pnatpot%modc .eq. 0) then
        nmode1 = pnatpot%nmode
     else
        nmode1 = pnatpot%nmode - 1
     end if
     allocate(potdim(nmode1))
     allocate(pdim(nmode1))
     allocate(pidx(nmode1))

     k=1
     do m=1,pnatpot%nmode
        if (pnatpot%modc .ne. m) then
           pdim(k) = cg_precon_luse(m)
           potdim(k) = pnatpot%potdim(m)
           if (pdim(k) .gt. potdim(k)) then
              pdim(k) =  potdim(k)
              if (mpirank.eq.0) then
                 write(ilog,'(a,i4,a,i4)') " * Reducing preconditionier" &
                      //" dimension to ",pdim(k), " for mode ", m
              end if
           end if
           k = k+1
        end if
     end do

     call sindex2cindex(pcdim,pdim,pdim,nmode1)

     allocate(rwork2(pcdim,pcdim))
     allocate(precon2(pcdim,pcdim))
     allocate(vmap(pcdim))
     zpcdim = pnatpot%modcdim - pcdim
     allocate(precon(zpcdim))
     allocate(dmap(zpcdim))

     vmap = 0
     dmap = 0

     ! mapping:
     ! these are the indices of the sparse matrix

     do j=1,pcdim
        call cindex2sindex(j,pidx,pdim,nmode1)
        call sindex2cindex(k,pidx,potdim,nmode1)
        vmap(j) = k
     end do


     ! these are the indices of the remaining diagonals
     k = 1
     i = 1
     do j=1,pnatpot%modcdim
        if (vmap(k) .eq.j) then
           if (k .lt. pcdim) k=k+1
        else
           dmap(i) = j
           i = i+1
        end if
     end do


     ! sparse matrix
     do k1=1,pcdim
        j =  vmap(k1)
        if (allocated(omega)) then
           call mcsppoverlapMtxElementMem(j,j,Ojj)
        else
           call mcsppoverlapMtxElement(j,j,Ojj,rwork)
        end if
        rwork2(k1,k1) = Ojj

        do k2=k1+1,pcdim
           j1 =  vmap(k2)

           if (allocated(omega)) then
              call mcsppoverlapMtxElementMem(j,j1,Ojj)
           else
              call mcsppoverlapMtxElement(j,j1,Ojj,rwork)
           end if
           rwork2(k2,k1) = Ojj
        end do
     end do
     do k1=1,pcdim
        do k2=k1+1, pcdim
           rwork2(k1,k2) =  rwork2(k2,k1)
        end do
     end do


     ! remaining diagonals
     do j=1,zpcdim
        j1 = dmap(j)
        if (allocated(omega)) then
           call mcsppoverlapMtxElementMem(j1,j1,Ojj)
        else
           call mcsppoverlapMtxElement(j1,j1,Ojj,rwork)
        end if
        precon(j) = Ojj
     end do


#ifdef MPI

     allocate(mpibuf2(pcdim,pcdim))
     allocate(mpibuf(zpcdim))

     call MPI_Reduce(rwork2,mpibuf2,pcdim**2,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)
     call MPI_Reduce(precon,mpibuf,zpcdim,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank .eq. 0) then
        rwork2 = mpibuf2
        precon = mpibuf
     end if

     deallocate(mpibuf2)
     deallocate(mpibuf)
#endif


     if (mpirank .eq. 0) then

        precon2 = 0.d0
        do i=1,pcdim
           precon2(i,i) = 1.d0
        end do

        ! invert
        call DPOSV( "L", pcdim,pcdim, rwork2, pcdim, precon2, pcdim,ierr )
        if (ierr.ne.0) then
           routine = "init_mcprecond (conjgrad preconditionier)"
           message = "Sparse overlap matrix is not positive definite."
           call errormsg
        end if


        ! now we we have the inverse of reduced overlap in precon2
        do i=1,zpcdim
           if (precon(i) .lt. 1.d-16) then
              routine = "init_mcprecond"
              message = "Overlap matrix is not positive definite"
              write(*,*)precon(i)
              call errormsg
           end if
           precon(i) = 1.d0/precon(i)
        end do


     end if

  else if (cg_precon_type .eq. cg_precon_block) then

     pcdim = pnatpot%modcdim/cg_precon_size
     zpcdim = modulo(pnatpot%modcdim, cg_precon_size)

     allocate(precon3(cg_precon_size,cg_precon_size,pcdim))
     allocate(rwork3(cg_precon_size,cg_precon_size,pcdim))

     do k=1,pcdim
        do i=1,cg_precon_size
           j2 = (k-1)*cg_precon_size + i
           do j=i,cg_precon_size
              j1 = (k-1)*cg_precon_size + j

              if (allocated(omega)) then
                 call mcsppoverlapMtxElementMem(j1,j2,Ojj)
              else
                 call mcsppoverlapMtxElement(j1,j2,Ojj,rwork)
              end if
              rwork3(j,i,k) = Ojj
           end do
        end do
        do i=1,cg_precon_size
           do j=i+1,cg_precon_size
              rwork3(i,j,k) =  rwork3(j,i,k)
           end do
        end do

     end do

    !remaining elements
     if (zpcdim.gt.0) then

        allocate(precon2(zpcdim,zpcdim))
        allocate(rwork2(zpcdim,zpcdim))

        k = cg_precon_size*pcdim

        do i=1,zpcdim
           j1 = k+i
           do j=i,zpcdim
              j2 = k+j

              if (allocated(omega)) then
                 call mcsppoverlapMtxElementMem(j2,j1,Ojj)
              else
                 call mcsppoverlapMtxElement(j2,j1,Ojj,rwork)
              end if
              rwork2(j,i) = Ojj
           end do
        end do
     end if

#ifdef MPI

     allocate(mpibuf3(cg_precon_size,cg_precon_size,pcdim))

     call MPI_Reduce(rwork3,mpibuf3,pcdim*cg_precon_size**2,&
          MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)

     if (mpirank .eq. 0) then
        rwork3 = mpibuf3
     end if
     deallocate(mpibuf3)

     if (zpcdim .gt. 0) then
        allocate(mpibuf2(zpcdim,zpcdim))
        call MPI_Reduce(rwork2,mpibuf2,zpcdim**2,&
             MPI_DOUBLE_PRECISION,MPI_SUM,0,MPI_COMM_WORLD,ierr)
        if (mpirank .eq. 0) then
           rwork2 = mpibuf2
        end if

        deallocate(mpibuf2)
     end if

#endif


     if (mpirank.eq.0) then

        !$omp parallel do private(i, ierr)
        do k=1,pcdim
           precon3(:,:,k) = 0.d0
           do i=1,cg_precon_size
              precon3(i,i,k) = 1.d0
           end do
           call DPOSV( "L", cg_precon_size,cg_precon_size,rwork3(:,:,k), &
                cg_precon_size, precon3(:,:,k),cg_precon_size,ierr )
           if (ierr.ne.0) then
              routine = "init_mcprecond (conjgrad preconditionier)"
              message = "Block overlap matrix is not positive definite."
              call errormsg
           end if
        end do
        !$omp end parallel do

        if (zpcdim .gt. 0) then

           precon2 = 0.d0
           do i=1,zpcdim
              precon2(i,i) = 1.d0
           end do

           call DPOSV( "L", zpcdim,zpcdim,rwork2, &
                zpcdim,precon2,zpcdim,ierr )
           if (ierr.ne.0) then
              routine = "init_mcprecond (conjgrad preconditionier)"
              message = "(Rest) block overlap matrix is not positive definite."
              call errormsg
           end if

        end if


     end if

     if (allocated(rwork2)) deallocate(rwork2)
     if (allocated(rwork3)) deallocate(rwork3)


  else
     routine = "init_mcprecond"
     message = "Unknown preconditionier type. " //&
          "This should not happen. Please file a bugreport."
     call errormsg
  end if

  if (allocated(rwork)) deallocate(rwork)


end subroutine init_mcprecond



subroutine free_mcprecond()
  use mcprecond_m
  implicit none

  if (allocated(precon)) deallocate(precon)
  if (allocated(precon2)) deallocate(precon2)
  if (allocated(precon3)) deallocate(precon3)

end subroutine free_mcprecond


subroutine operate_mcprecond(x,mx,NRHS)
  use mcoverlap
  use mcprecond_m
  use mcpfitparams
  use mpidata
#ifdef MPI
  use mpi
#endif

  implicit none
  integer, intent(IN)           :: NRHS
  double precision, intent(IN)  :: x(pnatpot%modcdim,NRHS)
  double precision, intent(OUT) :: mx(pnatpot%modcdim,NRHS)
  integer                       :: j,i,k,j1,j2
  integer                       :: k1,k2,n
  integer                       :: ierr, LDAB
  double precision              :: rdum

#ifdef DEBUG
  call chkNpotPtr("operate_mcprecon")
#endif


  if (cg_precon_type .eq. cg_precon_none) then
     ! no preconditioner

     ! only in rank 0
     if (mpirank.ne.0) return

     !$omp parallel do private(j)
     do i=1,NRHS
        do j=1,pnatpot%modcdim
           mx(j,i) = x(j,i)
        end do
     end do
     !$omp end parallel do

     return
  else if (cg_precon_type .eq. cg_precon_diag) then
     ! diagonal (Jacobi)

     ! only in rank 0
     if (mpirank.ne.0) return

     !$omp parallel do private(j)
     do i=1,NRHS
        do j=1,pnatpot%modcdim
           mx(j,i) = x(j,i)*precon(j)
        end do
     end do
     !$omp end parallel do
     return

  elseif  (cg_precon_type .eq.  cg_precon_band) then
     ! Band matrix

     ! only in rank 0
     if (mpirank.ne.0) return

     !$omp single
     LDAB = cg_precon_size + 1
     mx = x
     call DPBTRS( "U", pnatpot%modcdim, cg_precon_size, NRHS, &
          precon2, LDAB, mx, pnatpot%modcdim , ierr )
     !$omp end single

     return


  elseif  (cg_precon_type .eq.  cg_precon_sparse) then
     ! invert a sparse overlap  matrix

     ! we do this in rank 0 only
     if (mpirank.ne.0) return

     mx = 0.d0
     do n=1,NRHS

        do j=1,pcdim
           k1 = vmap(j)
           rdum = 0.d0
           !$omp parallel do private(k2) reduction(+:rdum)
           do i=1,pcdim
              k2 = vmap(i)
              rdum = rdum + precon2(i,j)*x(k2,n)
           end do
           !$omp end parallel do
           mx(k1,n) = rdum
        end do

        !$omp parallel do private(k1)
        do j=1,zpcdim
           k1 = dmap(j)
           mx(k1,n) = precon(j)*x(k1,n)
        end do
        !$omp end parallel do

     end do

  elseif  (cg_precon_type .eq.  cg_precon_block) then

    ! we do this in rank 0 only
     if (mpirank.ne.0) return

     mx = 0.d0

     do n=1,NRHS

        do k=1,pcdim

           !$omp parallel do private(j1,j2,j,rdum)
           do i=1,cg_precon_size
              j2 = (k-1)*cg_precon_size + i

              rdum = 0.d0
              do j=1,cg_precon_size
                 j1 = (k-1)*cg_precon_size + j
                 rdum =  rdum + x(j1,n)*precon3(j,i,k)
              end do

              mx(j2,n) = rdum
           end do
           !$omp end parallel do

        end do

        if (zpcdim .gt. 0) then
           k = cg_precon_size*pcdim

           !$omp parallel do private(j1,j2,j,rdum)
           do i=1,zpcdim
              j2 = k + i
              rdum = 0.d0

              do j=1,zpcdim
                 j1 = k+j
                 rdum =  rdum + x(j1,n)*precon2(j,i)
              end do
              mx(j2,n) = rdum

           end do
           !$omp end parallel do

        end if

     end do

  end if


end subroutine operate_mcprecond

