!  -*- f90 -*-

! solve "omega * coeff = Vn2" for coeff


! this just a wrapper that decides weather to use lapack or scalapack
subroutine mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,modcdim,delta,ierr)

  use mpidata

  implicit none
  integer, intent(IN)              :: nsample, NRHS, modcdim
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: Vn2(nsample,NRHS)
  double precision, intent(INOUT)  :: omega(nsample,modcdim)
  double precision, intent(OUT)    :: coeff(modcdim,NRHS)
  double precision, intent(OUT)    :: delta


  ierr = 0

  if (nmpi.eq.1) then
     ! regular lapacck
     call lapack_mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,modcdim,delta,ierr)
  else
     ! scalapack
     call scalapack_mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,modcdim,delta,ierr)
  end if

end subroutine mcsolve_pseudo



!###########################################
! solve omeega* coeff = vn2
!
!
!###########################################
subroutine lapack_mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,modcdim,delta,ierr)
  implicit none
  integer, intent(IN)              :: nsample, NRHS, modcdim
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: Vn2(nsample,NRHS)
  double precision, intent(INOUT)  :: omega(nsample,modcdim)
  double precision, intent(OUT)    :: coeff(modcdim,NRHS)
  double precision, intent(OUT)    :: delta

  character(len=*), parameter      :: TRANS='N'
  integer                          :: LWORK
  double precision, allocatable    :: WORK(:)
  double precision                 :: rdum(2)

  integer                          :: i,j,M,N,LDA,LDB

  M = nsample
  N = modcdim
  LDA = nsample
  LDB = nsample

  ! workspace query
  LWORK = -1
  call DGELS( TRANS, M, N, NRHS, omega, LDA, Vn2, LDB, rdum, LWORK, ierr )
  LWORK = int(rdum(1))
  allocate(WORK(LWORK))

  ! solve
  call DGELS( TRANS, M, N, NRHS, omega, LDA, Vn2, LDB, WORK, LWORK, ierr )

  if (ierr .eq.0) then
     !$omp parallel do
     do i=1,NRHS
        coeff(:,i) = Vn2(1:modcdim,i)
     end do
     !$omp end parallel do
  end if

  ! there is only an error measure if M > N
  if (M .gt. N) then
     delta = 0.d0
     do j = 1, NRHS
        do i = M+1, N
           delta = delta + Vn2(i,j)**2
        end do
     end do
  else
     delta = -1.d0
  end if

  deallocate(WORK)

end subroutine lapack_mcsolve_pseudo



subroutine scalapack_mcsolve_pseudo(Vn2,omega,coeff,nsample,NRHS,modcdim,delta,ierr)

  use global
  use mpidata
#ifdef MPI
  use mpi
#endif

  implicit none
  integer, intent(IN)              :: nsample, NRHS, modcdim
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: Vn2(nsample,NRHS)
  double precision, intent(INOUT)  :: omega(nsample,modcdim)
  double precision, intent(OUT)    :: coeff(modcdim,NRHS)
  double precision, intent(OUT)    :: delta

#ifdef SCALAPACK
#ifdef MPI

  character, parameter             :: TRANS="N"
  integer                          :: LWORK
  double precision, allocatable    :: WORK(:)
  double precision                 :: rdum(2)

  integer                          :: LDA
  integer                          :: DESComg(9), DESCVn2(9), DESCcoeff(9)
  integer                          :: NPROW,NPCOL,CONTEXT,myrow,mycol
  integer                          :: ntotsample, numroc, mpierr

  external numroc
  ierr = 0

  call MPI_Reduce(nsample,ntotsample,1,MPI_INTEGER, &
       MPI_SUM,0,MPI_COMM_WORLD,mpierr)
  call mpi_bcast(ntotsample,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)


  LDA = nsample

  ! we initialize a grid, whoch is 1D grid
  ! wih one column as usually nsample >> modcdim
  ! and create two global arrays descriptors that
  ! describe how omega and Vn2 are distributed
  ! among the MPI ranks

  NPROW = nmpi
  NPCOL = 1
  CALL BLACS_GET( -1, 0, CONTEXT )
  CALL BLACS_GRIDINIT( CONTEXT, 'R', NPROW, NPCOL )
  call blacs_gridinfo(CONTEXT, nprow, npcol, myrow, mycol)

  ! this should give nsample
  LDA = NUMROC(ntotsample, nsample, myrow, 0, NProw )
  if (LDA .ne. nsample) then
     routine = "scalapack_mcsolve_pseudo"
     message = "Something went wrong with the distribution of the sampling points"
     call errormsg
  end if

  ! setup array descriptors for the matrices omega and Vn2 as they are now
  CALL DESCINIT(DESComg ,ntotsample ,modcdim ,nsample,modcdim,0,0,CONTEXT,LDA,mpierr)
  CALL DESCINIT(DESCVn2 ,ntotsample ,NRHS ,nsample,NRHS ,0,0,CONTEXT,LDA,mpierr)


  ! workspace query
  LWORK = -1
  call PDGELS(TRANS,ntotsample,modcdim,NRHS,omega,1,1,DESComg,Vn2,1,1,DESCVn2,rdum,LWORK,ierr)
  LWORK = int(rdum(1))
  allocate(WORK(LWORK))

  ! solve
  call PDGELS(TRANS,ntotsample,modcdim,NRHS,omega,1,1,DESComg,Vn2,1,1,DESCVn2,WORK,LWORK,ierr)
  if (ierr .ne. 0) then
     ! exit here if anything went wrong
     deallocate(work)
     call BLACS_GRIDEXIT(CONTEXT)
     return
  end if

  !initialize a global array for the coefficients
  CALL DESCINIT(DESCcoeff,modcdim ,NRHS ,modcdim,NRHS ,0,0,CONTEXT,modcdim,mpierr)

  ! gather the coeffs: this is on proc-grid (0,0) only, need to bcast later
  call pdgemr2d(modcdim, NRHS, Vn2, 1, 1, descvn2, coeff, 1, 1, desccoeff, CONTEXT)


  delta = -1.d0 ! error estimate not implemented

  ! release the process grid
  call BLACS_GRIDEXIT(CONTEXT)


  return

#endif
  ! no MPI but Scalapack
  message = "ScaLAPACK needs MPI but this binary "// &
       "is compiled without MPI support. This should never have happened."
  routine = "scalapack_mcsolve_pseudo"
  call errormsg

#endif

end subroutine scalapack_mcsolve_pseudo
