! -*- f90 -*-

module mcsampling
  implicit none

  ! some parameters for code readability
  ! methods to get the trajecory and their parameters

  ! smethod: how where to get the sampling trajectories from
  ! 1: load from external file
  ! 2: create a metropolis sampling
  ! 3: create a uniform sampling
  ! 4: complete primitive grid
  integer, parameter         :: mc_nsmethods  = 5
  integer, parameter         :: mc_readfile   = 1
  integer, parameter         :: mc_metropolis = 2
  integer, parameter         :: mc_uniform    = 3
  integer, parameter         :: mc_complete   = 4
  integer, parameter         :: mc_undefined  = -1

  ! enumerate tasks, each task is connected to a
  ! trajectory
  integer, parameter         :: mc_ntasks     = 3
  integer, parameter         :: mc_task_spp   = 1
  integer, parameter         :: mc_task_coeff = 2
  integer, parameter         :: mc_task_test  = 3

  ! each trajectory may be composed if  max_idxfile
  ! sub-trajectories
  integer, parameter         :: max_idxfile = 10
  integer                    :: ntotsample(mc_ntasks,0:max_idxfile)
  integer                    :: nstep(mc_ntasks,max_idxfile)
  integer                    :: nskipfirst(mc_ntasks,max_idxfile)
  integer                    :: smethod(mc_ntasks,max_idxfile)
  double precision           :: beta(mc_ntasks,max_idxfile)

  character(len=240)         :: idxfile(mc_ntasks,max_idxfile)
  character(len=80)          :: idxfileinfo(mc_ntasks,max_idxfile)
  integer                    :: nidxfile(mc_ntasks)

  logical                    :: lasciisamplings
  logical                    :: lesample
  logical                    :: lsamecoeffspp

contains

  logical function allTrajectoriesAreRead()
    implicit none
    integer :: task, ntraj

    allTrajectoriesAreRead = .true.

    do task=1, mc_ntasks
       ntraj = nidxfile(task)
       if (any(smethod(task,1:ntraj) .ne. mc_readfile)) then
          allTrajectoriesAreRead = .false.
          return
       end if
    end do

  end function allTrajectoriesAreRead

end module mcsampling


subroutine mcsampling_default()
  use mcsampling
  implicit none
  integer   :: i,j

  lasciisamplings = .true.
  lesample = .false.
  lsamecoeffspp = .false.

  do j=0,max_idxfile
     do i=1, mc_ntasks
        ntotsample(i,j) = 0
     end do
  end do

  do j=1,max_idxfile
     do i=1, mc_ntasks
        nstep(i,j) = 1
     end do
  end do

  do j=1,max_idxfile
     do i=1, mc_ntasks
        nskipfirst(i,j) = 0
     end do
  end do

  do j=1,max_idxfile
     do i=1, mc_ntasks
        beta(i,j) = 0.d0
     end do
  end do

  do j=1,max_idxfile
     do i=1, mc_ntasks
        smethod(i,j) = mc_undefined
     end do
  end do

  do j=1,max_idxfile
     do i=1, mc_ntasks
        idxfile(i,j) = " "
        idxfileinfo(i,j) = " "
     end do
  end do

  do i=1, mc_ntasks
     nidxfile(i) = 0
  end do


end subroutine mcsampling_default



subroutine logsampling(channel)
  use mcsampling
  implicit none
  integer, intent(IN)           :: channel

  character(len=18)             :: label
  integer                       :: nfile, task
  character(len=8)              :: buf


  do task=1,mc_ntasks

     if (nidxfile(task) .eq. 0) cycle

     write(channel, '(a)')
     select case (task)
     case(mc_task_spp)
        write(channel, '(a,i8)') "Sampling to obtain reduced densities / SPP"
     case(mc_task_coeff)
        write(channel, '(a,i8)') "Sampling to obtain coefficients"
     case(mc_task_test)
        write(channel, '(a,i8)') "Sampling of test trajectory"
     end select

     if (lsamecoeffspp) then
        if (task .eq. mc_task_spp .and. &
             nidxfile(mc_task_spp) .eq. 0 .and. &
             nidxfile(mc_task_coeff) .ne. 0) then
           write(channel, '(a)')    "    Same points as for coefficients."
           return
        else if (task .eq. mc_task_coeff .and. &
             nidxfile(mc_task_coeff) .eq. 0 .and. &
             nidxfile(mc_task_spp) .ne. 0) then
           write(channel, '(a)')    "    Same points as for densities / SPP."
           return
        end if

     end if

     do nfile=1,nidxfile(task)
        write(channel, '(a)') ""
        select case (smethod(task,nfile))
        case (mc_readfile)
           label="read from file"
        case (mc_metropolis)
           label="metropolis"
        case (mc_uniform)
           label="uniform"
        case (mc_complete)
           label="complete grid"
        case default
           label="not defined"
        end select

        write(buf,'(i3)') nfile
        buf = " "//adjustl(buf)

        label = adjustr(label)
        write(channel, '(a)')       buf(1:4)//"Method:              "//trim(label)

        if  (smethod(task,nfile) .ne. mc_undefined) then
           write(label,'(i8)') ntotsample(task,nfile)
           label = adjustr(label)
           write(channel, '(a)')    "    Total points:        "//trim(label)
        end if

        select case (smethod(task,nfile))
        case (mc_readfile)
           write(label,'(i8)') nskipfirst(task,nfile)
           label = adjustr(label)
           write(channel, '(a)')    "    Skip first:          "//trim(label)
           write(label,'(i8)') nstep(task,nfile)
           label = adjustr(label)
           write(channel, '(a,i8)') "    Nstep:               "//trim(label)
           write(channel, '(a,i8)') "    File info:           "//trim(idxfileinfo(task,nfile))
        case (mc_metropolis)
           write(label,'(E16.10)') beta(task,nfile)
           label = adjustr(label)
           write(channel, '(a)')    "    K_b*T (a.u.):        "//trim(label)
           write(label,'(i8)') nskipfirst(task,nfile)
           label = adjustr(label)
           write(channel, '(a)')    "    Skip first (burn-in):"//trim(label)
           write(label,'(i8)') nstep(task,nfile)
           label = adjustr(label)
           write(channel, '(a)')    "    Nstep:               "//trim(label)
        end select
        write(channel, '(a)')       "    File:                "//trim(idxfile(task,nfile))

     end do ! nfile


     write(channel, '(a)')

  end do !  task


end subroutine logsampling




  ! check if the sampling file
  ! has the required number of
  ! samplings (if given) or determine
  ! the number of samplings in the file.
  ! Also set fileinfo is available
subroutine chksamplingfile(task,nfile)
  use mcsampling
  use global
  use channels
  use logdat

  implicit none
  integer, intent(IN) :: task, nfile
  logical             :: isascii
  integer             :: ierr, recordcount, required
  character(len=16)   :: buf(2)
  character(len=80)   :: buf80

  call isASCIIFile(idxfile(task,nfile), isascii, ierr)
  routine = "isASCIIFile called from chksamplingfile"

  if ( ierr .eq. 1) then
     goto 753
  else if ( ierr .eq. 2) then
     goto 754
  else if ( ierr .eq. 3) then
     goto 755
  end if

    ! we know the number of samplings, just check
  ! if they are there
  if (ntotsample(task,nfile) .gt.0) then
     required = nskipfirst(task,nfile) + ntotsample(task,nfile) * nstep(task,nfile)
     if (isascii) then
        call haslinecount(required, recordcount,idxfile(task,nfile), "#", ierr)
        if ( ierr .eq. 1) then
           goto 753
        else if ( ierr .eq. 2) then
           goto 754
        end if
        if (recordcount .lt. required) goto 756
     else
        open(idump,file=idxfile(task,nfile),form="unformatted", err=754)
        recordcount = 0
        do while (recordcount .le. required)
           read(idump, end=756)
           recordcount = recordcount +1
        end do
        close(idump)
     end if

  else
     ! we take all samplings from file
     ! so we scan how many are there
     if (isascii) then

        call linecount(recordcount, idxfile(task,nfile),"#",  ierr)
        if ( ierr .eq. 1) then
           goto 753
        else if ( ierr .eq. 2) then
           goto 754
        end if

     else

        open(idump,file=idxfile(task,nfile),form="unformatted",err=754)
        recordcount = 0
        do while (.true.)
           read(idump, end=777,err=776) buf80
           buf80 = adjustl(buf80)
           if (buf80(1:1) .eq. "#") cycle
776        continue
           recordcount = recordcount +1
        end do
777     continue
        close(idump)
     end if

     ntotsample(task,nfile) = recordcount
  end if

  call GetSamplingFileInfo(idxfile(task,nfile), buf80, isascii)
  idxfileinfo(task,nfile) = buf80

  return

753 continue
  routine = "chksamplingfile"
  message = "File does not exist: "//trim(idxfile(task,nfile))
  call errormsg

754 continue
  routine = "chksamplingfile"
  message = "Could not open file: "//trim(idxfile(task,nfile))
  call errormsg

755 continue
  routine = "chksamplingfile"
  message = "Could not read file: "//trim(idxfile(task,nfile))
  call errormsg

756 continue

  write(buf(1),'(i8)') required
  buf(1) = adjustl(buf(1))
  write(buf(2),'(i8)') recordcount
  buf(2) = adjustl(buf(2))

  routine = "chksamplingfile"
  message = "Found "//trim(buf(2)) // " records/lines but need "// &
       trim(buf(1)) // " in file "//trim(idxfile(task,nfile))
  call errormsg


end subroutine chksamplingfile


! extract the first line from filename and
! store it in buf if line starts with "#"
subroutine GetSamplingFileInfo(filename, buf, lascii)
  use global
  use channels
  implicit none
  character(len=*), intent(IN)    :: filename
  character(len=c4), intent(OUT)  :: buf
  logical, intent(IN)             :: lascii

  buf = " "
  if  (lascii) then
     open(idump,file=filename,form="formatted")
     read(idump,'(a)',err=999) buf
     close(idump)
  else
     open(idump,file=filename,form="unformatted")
     read(idump,err=999) buf
     close(idump)
  end if

  buf = adjustl(buf)
  if (buf(1:1) .eq. "#") then
     buf(1:1) = " "
     buf = adjustl(buf)
  else
     buf = " -- N/A -- "
  end if
  return

999 continue
  close(idump)
  buf = " -- N/A -- "
  return

end subroutine GetSamplingFileInfo




