!  -*- f90 -*-

! solve "ovlp * X = coeff" for X
!
! input:
!   ovlp  : symmetric positive semidefinite, size: modcdim^2
!   coeff : on entry, the right hand side
!           on exit, if ierr == 0, the soluton X
!   epsilon: regilarization:
!            if epsilon > 0 and  lpseudo = .false.
!            then the eigenvalies are e_i are replaced
!            with e_i + epsilon*tr{ovlp} exp(-e_i/[epsilon*tr{ovlp}])
!            if  lpseudo = true.
!            all eigenvalues less then epsilon are not
!            taken into account when calculating the inverse
!   lpseudo: logical, if true, do not invert eigenvalues
!            less then epsilon. If false, regularize small eigenvalies
!   modcdim: number of equations
!   NRHS:    number of right hand sides
!   efile:   if efile is a non-empty string the eigenvalues
!            of ovlp are written to $name/efile
!            if efile is empty, no file is written.
!   ierr:    error flag. if ierr == 0, all went well
!            if ierr <> 0: the LAPACK/SCALAPACK error flag

! this just a wrapper that decides weather to use lapack or scalapack
subroutine mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,modcdim,NRHS,efile,ierr)

  use mpidata
  use channels
  use global
  use logdat

  implicit none

  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)
  double precision, intent(IN)     :: epsilon
  logical, intent(IN)              :: lpseudo, lscale
  character(len=*), intent(IN)     :: efile
  double precision                 :: eval(modcdim)
  double precision                 :: regeval(modcdim)
  character(len=c5)                :: fname
  integer                          :: i

  lwarn = .false.

  if (epsilon.lt.0.d0) then
     routine = "mcsolve_eigen"
     write(message,'(a,ES16.8)') "epsilon must be >= 0.d0. Got: ",epsilon
     call errormsg
  end if

  ierr = 0

#ifdef SCALAPACK

  ! for small matrices or one mpi rank we use the regular lapack
  if (nmpi.eq.1 .or. modcdim .le. 512 ) then
     ! regular lapack
     if (mpirank.eq.0) then
        call lapack_mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,eval, &
             regeval, modcdim,NRHS,ierr)
     end if
  else
     ! scalapack
     call scalapack_mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,eval, &
          regeval,modcdim,NRHS,ierr)
  end if

#else

  !fallback to regular lapack
  if (mpirank.eq.0) then
     call lapack_mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,eval, &
          regeval,modcdim,NRHS,ierr)
  end if

#endif

  ! dump eigevalues to file
  if (mpirank.eq.0 .and. len_trim(efile) .ne. 0) then
     fname = trim(name) // "/" // trim(efile)
     if (lovwr) then
        open(idump,file=fname,status="unknown",err=12)
     else
        open(idump,file=fname,status="new",err=12)
     end if
     write(idump,'(a)') "# eigenvalues, original, regularized, diff"
     do i=1,modcdim
        write(idump,'(3ES23.15)') eval(i), regeval(i),  regeval(i) - eval(i)
     end do
     close(idump)
  end if

  if (lwarn) then
     routine = "mcsolve_eigen"
     message = "Inverting eigenvaliues failed."
     call errormsg
  end if



  return

12 continue
  routine = "mcsolve_eigen"
  message = "Could not write eigenvalues to file. "//trim(fname) &
       //" Use -w option. Continueing anyways."
  call warnmsg()
  lwarn = .false.
  close(idump)


end subroutine mcsolve_eigen



!###########################################
! solve ovlp * coeff = omega v
!
!
!###########################################
subroutine lapack_mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,eval,regeval, &
     modcdim,NRHS,ierr)

  use global
  use logdat

  implicit none

  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)
  double precision, intent(IN)     :: epsilon
  logical, intent(IN)              :: lpseudo, lscale
  double precision, intent(OUT)    :: eval(modcdim)
  double precision, intent(OUT)    :: regeval(modcdim)
  double precision                 :: eval2(modcdim)
  double precision, allocatable    :: coeff2(:,:)
  integer                          :: i,j,k,n, allocstat

  ! LAPACK
  character(len=*), parameter      :: UPLO='L'
  character(len=*), parameter      :: JOBZ='V'
  integer                          :: LWORK
  double precision, allocatable    :: WORK(:)
  double precision                 :: rdum, reg




  LWORK = -1
  ! workspace query
  call DSYEV(JOBZ, UPLO, modcdim, ovlp, modcdim, eval, rdum, LWORK, ierr )
  LWORK = int(rdum)

  allocate(WORK(LWORK), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "lapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate LAPACK workspace."
     call errormsg
  end if

  call DSYEV(JOBZ, UPLO, modcdim, ovlp, modcdim, eval, WORK, LWORK, ierr )
  deallocate(WORK)
  ! ovlp now contains the eigenvectors

  if ( .not. lpseudo) then
     if (epsilon .gt. 0.d0) then
        if (lscale) then
           reg = 0.d0
           !$omp parallel do reduction(+:reg)
           do i=1,modcdim
              reg = reg +  abs(eval(i))
           end do
           !$omp end parallel do
           reg = reg*epsilon/modcdim
        else
           reg = epsilon
        end if

        !$omp parallel do
        do i=1,modcdim
           regeval(i) = eval(i) + reg*dexp(-abs(eval(i))/reg)
        end do
        !$omp end parallel do
     else
        !$omp parallel do
        do i=1,modcdim
           regeval(i) = eval(i)
        end do
        !$omp end parallel do
     end if

     do i=1,modcdim
        if (regeval(i) .ne. 0.d0) then
           eval2(i) = 1.d0/regeval(i)
        else
           message = "Division by zero " &
                //"while inverting overlap matrix. "
           routine = "lapack_mcsolve_eigen"
           call warnmsg
           ierr = -1
           return
        end if
     end do
  else
     do i=1,modcdim
        if ( eval(i) .gt. epsilon) then
           regeval(i) = eval(i)
        else
           regeval(i) = 0.d0
        end if
     end do

     do i=1,modcdim
        if ( eval(i) .gt. epsilon) then
           eval2(i) = 1.d0/regeval(i)
        else
           eval2(i) = 0.d0
        end if
     end do


  end if


  allocate(coeff2(modcdim,NRHS), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "lapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate workspace for coeff inversion"
     call errormsg
  end if


  coeff2 = 0.d0
  do n=1,NRHS
     !$omp parallel do private(k)
     do j=1,modcdim
        do k=1,modcdim
           coeff2(j,n) =   coeff2(j,n) +  ovlp(k,j)* coeff(k,n)*eval2(j)
        end do
     end do
     !$omp end parallel do
  end do

  coeff = 0.d0
  do n=1,NRHS
     !$omp parallel do private(k)
     do j=1,modcdim
        do k=1,modcdim
           coeff(j,n) =   coeff(j,n) +  ovlp(j,k)* coeff2(k,n)
        end do
     end do
     !$omp end parallel do
  end do

  deallocate(coeff2)

end subroutine lapack_mcsolve_eigen



subroutine scalapack_mcsolve_eigen(ovlp,coeff,epsilon,lpseudo,lscale,eval,regeval, &
     modcdim,NRHS,ierr)
  use global
  use mpidata
!  use channels
  use logdat
#ifdef MPI
  use mpi
#endif

  implicit none

  integer, intent(IN)              :: modcdim,NRHS
  integer, intent(OUT)             :: ierr
  double precision, intent(INOUT)  :: ovlp(modcdim,modcdim)
  double precision, intent(INOUT)  :: coeff(modcdim,NRHS)
  double precision, intent(IN)     :: epsilon
  logical, intent(IN)              :: lpseudo,lscale
  double precision, intent(OUT)    :: eval(modcdim)
  double precision, intent(OUT)    :: regeval(modcdim)

#ifdef MPI
#ifdef SCALAPACK

  character, parameter             :: UPLO="L"
  character(len=*), parameter      :: JOBZ='V'
  integer                          :: LWORK
  double precision, allocatable    :: WORK(:)
  double precision                 :: rdum, reg
  double precision                 :: eval2(modcdim)
  integer                          :: i,k,m,n, allocstat
  integer                          :: MB, mcoeff, movlp, ncoeff, novlp
  integer                          :: DESCovlp(9), DESCcoeff(9)
  integer                          :: DESCevec(9), DESCcoeff2(9)
  integer                          :: DESCwork2(9),DESCovlp2(9)
  integer                          :: NPROW,NPCOL,CONTEXT,myrow,mycol
  integer                          :: mpierr
  double precision, allocatable    :: ovlp2(:,:), evec(:,:)
  double precision, allocatable    :: work2(:,:), coeff2(:,:)
  integer                          :: numroc, INDXL2G
  external numroc, INDXL2G

  ierr = 0

  ! we initialize a 2D process grid
  ! which is as square as possible
  NPROW = int(dsqrt(dble(nmpi)+1.d-9))
  if (nprow**2 .gt. nmpi) nprow = nprow -1
  NPCOL =  nprow

  ! add an extra row if we have space
  if ((nprow+1)*npcol .le. nmpi) then
     nprow = nprow+1
  end if
  ! add another extra row if we have space
  if ((nprow+1)*npcol .le. nmpi) then
     nprow = nprow+1
  end if

!  if (mpirank.eq.0) then
!     write(ilog,'(a,i4,a,i4,a)') "Using ScaLAPACK PDSYEV on a ", &
!          NPCOL, " by ", NPROW, " process grid"
!     call myflush(ilog)
!  end if

  ! initialize the process grid
  CALL BLACS_GET( -1, 0, CONTEXT )
  CALL BLACS_GRIDINIT( CONTEXT, 'R', NPROW, NPCOL )
  call blacs_gridinfo(CONTEXT, nprow, npcol, myrow, mycol)

  ! early exit for all ranks hat
  ! are not part of the process grid
  if (mycol+1 .gt. NPCOL .or. myrow+1 .gt. NPROW) return


  ! create two global arrays descriptors that
  ! describe ovlp and coeff as they are distributed
  ! among the MPI ranks now (only in mpirank 1)
  ! setup array descriptors for the matrices omega and coeff as they are now
  CALL DESCINIT(DESCovlp  ,modcdim ,modcdim ,modcdim,modcdim,0,0,CONTEXT,modcdim,mpierr)
  CALL DESCINIT(DESCcoeff ,modcdim ,NRHS ,modcdim,NRHS ,0,0,CONTEXT,modcdim,mpierr)

  ! redustribute: calculate local array dimensions
  ! for the re-distributed matrices
  MB = MBdefault
  movlp =  NUMROC(modcdim, MB, myrow, 0, NProw )
  novlp =  NUMROC(modcdim, MB, mycol, 0, NPcol )
  mcoeff = NUMROC(modcdim, MB, myrow, 0, NProw )
  ncoeff = NUMROC(NRHS, MB, mycol, 0, NPcol )

  ! create array descriptors for the re-disribued overlap matrix
  ! and the eigenvectros
  CALL DESCINIT(DESCovlp2, modcdim ,modcdim ,MB,MB,0,0,CONTEXT,movlp,mpierr)
  CALL DESCINIT(DESCevec,  modcdim ,modcdim ,MB,MB,0,0,CONTEXT,movlp,mpierr)
  CALL DESCINIT(DESCcoeff2 ,modcdim ,NRHS ,MB,MB ,0,0,CONTEXT,mcoeff,mpierr)
  call DESCINIT(DESCwork2  ,modcdim ,NRHS ,MB,MB ,0,0,CONTEXT,mcoeff,mpierr)

  ! allocate space for the disributed matrices
  allocate(ovlp2(movlp,novlp), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "sclapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate distributed ovlp matrix"
     call errormsg
  end if

  ! below could be allocations of zero-sized arrays,
  ! but fortran should be cool with that...
  allocate(evec(movlp,novlp), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "sclapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate distributed evec matrix"
     call errormsg
  end if

  allocate(coeff2(mcoeff,ncoeff), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "sclapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate distributed evec matrix"
     call errormsg
  end if

  allocate(work2(mcoeff,ncoeff), stat=allocstat)
  if (allocstat .ne. 0) then
     routine = "sclapack_mcsolve_eigen (mcsolveeigen.F90)"
     message = "Unable to allocate distributed work2 matrix"
     call errormsg
  end if


  ! if modcdim is large, we need
  ! to use a self written distribuion routine
  ! because of an integer overflow in the scalapack
  ! library. In the long run we should start with
  ! block cyclic arrays in mcoverlap already when calculating
  ! the overlaps...
  ! See also "xxmr2d:out of memory"
  call mcdistribute(ovlp, ovlp2,movlp,novlp,DESCovlp2)
  call mcdistribute(coeff,coeff2,mcoeff,ncoeff,DESCcoeff2)

  ! workspace query
  LWORK = -1
  call PDSYEV( JOBZ, UPLO, modcdim, ovlp2, 1, 1, DESCovlp2, eval, &
       evec, 1, 1, DESCevec, rdum, LWORK, ierr )

  ! alloc WORK
  LWORK = int(rdum)
  allocate(WORK(LWORK))

  ! Eigen-Decomposition
  call PDSYEV( JOBZ, UPLO, modcdim, ovlp2, 1, 1, DESCovlp2, eval, &
       evec, 1, 1, DESCevec, WORK, LWORK, ierr )

  if (ierr .ne. 0) then
     deallocate(ovlp2,evec,work2,work,coeff2)
     call BLACS_GRIDEXIT(CONTEXT)
     return
  end if
  deallocate(WORK)

  if (.not.lpseudo) then
     if (epsilon .gt. 0.d0) then
        if (lscale) then
           reg = 0.d0
           !$omp parallel do reduction(+:reg)
           do i=1,modcdim
              reg  = reg + eval(i)
           end do
           !$omp end parallel do
           reg = reg*epsilon/modcdim
        else
           reg = epsilon
        end if

        !$omp parallel do
        do i=1,modcdim
           regeval(i) = eval(i) + reg*dexp(-eval(i)/reg)
        end do
        !$omp end parallel do
     else
        !$omp parallel do
        do i=1,modcdim
           regeval(i) = eval(i)
        end do
        !$omp end parallel do
     end if


     do i=1,modcdim
        if (regeval(i) .gt. 0.d0) then
           eval2(i) = 1.d0/regeval(i)
        else
           routine = "scalapack_mcsolve_eigen"
           message = "Division by zero or negative eigenvalues " &
                //"while inverting overlap matrix"
           call errormsg
        end if
     end do
  else

     do i=1,modcdim
        if ( eval(i) .gt. epsilon) then
           regeval(i) = eval(i)
        else
           regeval(i) = 0.d0
        end if
     end do

     do i=1,modcdim
        if ( eval(i) .gt. epsilon) then
           eval2(i) = 1.d0/regeval(i)
        else
           eval2(i) = 0.d0
        end if
     end do
  end if


  ! the eigenvectors are in the distributed matrix
  ! evec. We do the the inversion in three steps:
  ! 1) multiply evec**T to coeff and store result in work2

  CALL PDGEMM( 'T', 'N', modcdim, NRHS, modcdim, 1.d0, &
       evec, 1, 1, DESCevec, coeff2, 1, 1, DESCcoeff2,&
       0.d0, work2, 1, 1, DESCwork2)

  ! 2) multiply eval**1 - to work

  !$omp parallel do private(m,k)
  do n=1,ncoeff !NRHS
     do m=1,mcoeff !modcdim
        k = INDXL2G(m, MB, myrow, 0, NProw )
        work2(m,n) = work2(m,n)*eval2(k)
     end do
  end do
  !$omp end parallel do

  ! 3)  multiply evec and work and store the result in coeff2
  CALL PDGEMM( 'N', 'N', modcdim, NRHS, modcdim, 1.d0, &
       evec, 1, 1, DESCovlp2, work2, 1, 1, DESCwork2,&
       0.d0, coeff2, 1, 1, DESCcoeff2)

  ! we are done, now collect the result
  call  mcgather(coeff,coeff2,mcoeff,ncoeff,DESCcoeff2)

  ! Scalapack Gather:
  !  call pdgemr2d(modcdim, NRHS, coeff2, 1, 1, desccoeff2, coeff, &
  !       1, 1, desccoeff, CONTEXT)

  if (allocated(ovlp2)) deallocate(ovlp2)
  if (allocated(coeff2))deallocate(coeff2)
  if (allocated(work2)) deallocate(work2)
  if (allocated(evec)) deallocate(evec)

  ! release the process grid
  call BLACS_GRIDEXIT(CONTEXT)

  return

#endif
#endif

  ! no MPI and no Scalapack
  message = "ScaLAPACK seems not linked. "// &
       " This should never have happened."
  routine = "scalapack_mcsolve_spd"
  call errormsg

end subroutine scalapack_mcsolve_eigen



