!  -*- f90 -*-

module mcstatistics
  use mpidata
  use channels
  use global
  use tensortype
  use natpottype
  use cpdtype_m
  use runfit_m
  use pftools
  use pesselect_m, only: lomppotential
  use pesselect_s, only: potential
  use mcpfitparams
  use mcpmpitools
  use logdat
  use pftools
  use dvrtype_m
  use mcsampling

  implicit none

contains

  subroutine stat_current_traj(tensor, channel, lwrminmax)
    implicit none
    class(tensor_t), intent(IN)   :: tensor
    integer, intent(IN)           :: channel
    logical, intent(IN)           :: lwrminmax

    double precision              :: mean, rms
    double precision              :: vdmin,vdmax,vomin,vomax
    integer                       :: nsampletot, nsample

    nsampletot = sum(mpi_chunksizes)
    nsample = mpi_chunksizes(mpirank)

    select type(tensor)
    type is (cpd_t)
       call testcpd(potential,tensor,dvrindex,nsample, &
            mean,rms,vomin,vomax,vdmin,vdmax,lomppotential)
    type is (natpot_t)
       call testpf(potential,tensor,dvrindex,nsample, &
            mean,rms,vomin,vomax,vdmin,vdmax,lomppotential)
    end select

    if (mpirank.eq.0) then


       call wrstatistics(mean, rms, vdmin, vdmax,  &
            nsampletot, channel)

       if (lwrminmax) then
          call wrvminmax(vomin, vomax, channel)
       end if

       call myflush(channel)

    end if

  end subroutine stat_current_traj




  subroutine stat_complete(natpot, vpot,channel, lwrminmax)
    implicit none
    type(natpot_t), intent(IN)    :: natpot
    integer, intent(IN)           :: channel
    logical, intent(IN)           :: lwrminmax
    double precision, intent(IN)  :: vpot(natpot%vpotdim)

    double precision              :: mean, rms
    double precision              :: vdmin,vdmax,vomin,vomax
    integer                       :: nsampletot


    if (mpirank .eq. 0) then

       nsampletot = natpot%vpotdim
       call testpfcmplte(natpot,vpot,mean,rms,vomin,vomax,vdmin,vdmax)

       call wrstatistics(mean, rms, vdmin, vdmax,  &
            nsampletot, channel)

       if (lwrminmax) then
          call wrvminmax(vomin, vomax, channel)
       end if

       call myflush(channel)

    end if


  end subroutine stat_complete



  subroutine Open_stat_file(filename, lprintheader)

    implicit none

    character(len=*), intent(IN)  :: filename
    character(len=c5)             :: fname
    logical, intent(IN), optional :: lprintheader
    logical                       :: lexists
    logical                       :: lprintheader1

    if (present(lprintheader)) then
       lprintheader1 = lprintheader
    else
       lprintheader1 = .true.
    end if

    if (mpirank.ne.0) return

    fname = trim(name) // "/" // trim(filename)

    inquire(file=fname, EXIST=lexists)
    if (lexists .and..not. lovwr) then
       routine = "Open_stat_file (mcstatistics.F90)"
       message = "Cannot overwrite file "//trim(fname)//". " &
            //"Use -w option."
       call errormsg()
    end if

    open(istat,file=fname, status="unknown")

    if (lprintheader1) then
       write(istat,'(a)')  "  **************************************"
       write(istat,'(a)')  "      STATISTICS FOR THE CONSTITUENT "
       write(istat,'(a)')  "        PARTS OF THE TEST SAMPLING"
       write(istat,'(a)')  "  **************************************"
    end if

  end subroutine Open_stat_file



  subroutine stat_task_files(task,dvr,tensor, channel)
    implicit none
    integer, intent(IN)           :: task, channel
    type(dvr_t), intent(IN)       :: dvr
    class(tensor_t), intent(IN)   :: tensor
    integer                       :: nsampletot, nsample

    integer                       :: nfile
    character(len=18)             :: label

    ! channel must be open

    do nfile=1, nidxfile(task)
       call loadruntask(task,dvr,nsample,nfile)

       if (mpirank.eq.0) then

          write(channel,'(a)') ""
          write(channel,'(a)') ""

          nsampletot = sum(mpi_chunksizes)
          write(channel,'(a)') "File: "//trim(idxfile(task,nfile))

          select case(smethod(task,nfile))
          case (mc_metropolis)
             write(channel,'(a)', advance="no") "Metropolis, "
             write(label,'(i8)') nsampletot
             label = adjustl(label)
             write(channel,'(a)', advance="no") trim(label)//" points, "
             write(label,'(f16.8)') beta(task,nfile)
             label = adjustl(label)
             write(channel,'(a)') "1/beta = "//trim(label)//" a.u."
          case (mc_uniform)
             write(channel,'(a)', advance="no") "Uniform, "
             write(label,'(i8)') nsampletot
             label = adjustl(label)
             write(channel,'(a)') trim(label)//" points"
          case (mc_readfile)
             write(channel,'(a)', advance="no") "External file, "
             write(label,'(i8)') nsampletot
             label = adjustl(label)
             write(channel,'(a)') trim(label)//" points"
             write(channel,'(a)') "File info: "//trim(idxfileinfo(task,nfile))
          end select

       end if

       call stat_current_traj(tensor, channel, .true.)

       if (nfile .ne. nidxfile(task) .and. mpirank.eq.0) then
          write(channel,'(a)')  "*******************************************" &
               //"*********************************"
       end if

    end do



  end subroutine stat_task_files





  subroutine wrstatistics(mean, rms, vmin, vmax, nsample, channel)
    implicit none
    double precision, intent(IN) :: mean, rms, vmin, vmax
    integer, intent(IN)          :: nsample, channel
    character(len=*), parameter  :: fstr="(a,2x,3(es11.4e2,a,2x))"

    double precision, parameter  :: au2ic = 2.1947463137d+5
    double precision, parameter  :: au2ev = 27.21138386d0
    character(len=12)            :: buf
    logical                      :: isopen

    inquire(channel, opened=isopen)
    if (.not. isopen) return

    write(buf,'(i8)') nsample
    call rmblnk(buf)

    write(channel,'(a)') ""
    write(channel,'(a)') " Statistics of the difference " &
         // "(V(fit) - V(exact)) using "//trim(buf)//" sampling points."
    write(channel,fstr) "   Mean    = ", mean, " au, ", &
         mean*au2ev, " eV, ", mean*au2ic, " cm-1"
    write(channel,fstr) "   RMS     = ", rms, " au, ", &
         rms*au2ev, " eV, ", rms*au2ic, " cm-1"
    write(channel,fstr) "   Minval  = ", vmin, " au, ", &
         vmin*au2ev, " eV, ", vmin*au2ic, " cm-1"
    write(channel,fstr) "   Maxval  = ", vmax, " au, ", &
         vmax*au2ev, " eV, ", vmax*au2ic, " cm-1"
    write(channel,'(a)') ""
    call myflush(channel)

  end subroutine wrstatistics



  subroutine wrvminmax(vomin, vomax, channel)
    implicit none
    double precision, intent(IN) :: vomin, vomax
    integer, intent(IN)          :: channel

    double precision, parameter  :: au2ic = 2.1947463137d+5
    double precision, parameter  :: au2ev = 27.21138386d0
    character(len=*), parameter  :: fstr="(a,2x,3(es11.4e2,a,2x))"
    logical                      :: isopen

    inquire(channel, opened=isopen)
    if (.not. isopen) return

    write(channel,'(a)') ""
    write(channel,'(a)') " Minimum and maximum of the exact potential"

    write(channel,fstr) "   Minval  = ", vomin, " au, ", &
         vomin*au2ev, " eV, ", vomin*au2ic, " cm-1"
    write(channel,fstr) "   Maxval  = ", vomax, " au, ", &
         vomax*au2ev, " eV, ", vomax*au2ic, " cm-1"
    write(channel,'(a)') ""
    call myflush(channel)

  end subroutine wrvminmax


end module mcstatistics
