!   -*- f90 -*-
!*******************************************************************
! module mctools
!
! Contains a subroutines using Monte-Carlo techniques
!
! Contents:
!
! setSeed              - set the random seed for intrinsic routine
!                        primitive grid (dvrindex)
! mksamplings          - wrapper that calls the various specific
!                        routines to generate samplings
!
!*******************************************************************




module mctools

    ! MCTDH
    use global
    use channels
    use mcsampling
    use mcpfitparams
    use dvrtype_m
    use mcmetropolis
    use mcuniform
    use mccomplete
    use mpidata
    use pesselect_m
    use pesselect_s, only: test_pes_threadsafe
    use mcpmpitools
    use mcsamplingio

  implicit none

  ! metropolis_maxtry is the number of tries
  ! of the walker to move to a neighboring
  ! grid point in the metropolis algorithm.
  ! If none of metropolis_maxtry consecutive
  ! tries succeeded, then the walker is assumed
  ! to be trapped in a minimum which only consists
  ! of the current grid-point.
  integer, parameter          :: metropolis_maxtry = 1000

contains

  subroutine setSeed(seed)
    implicit none
    integer, intent(IN)  :: seed
    integer              :: rsize, n
    integer, parameter   :: iconst1 = 982451653
    integer, parameter   :: iconst2 = 472882049
    integer, allocatable :: rseed(:)

    ! the input seed is "iseed" from
    ! RUN-SECTION plus MPI-Rank number.
    ! The gfortran RNG creates then
    ! correlated sequenzes in the
    ! different ranks if
    ! one just sets  the seed to continuous
    ! numbers. (At least in some versions
    ! of GCC.) So we mess up the seeds
    ! in the ranks a bit.
    call random_seed(size=rsize)
    allocate(rseed(rsize))
    rseed(1) = modulo(iconst1*seed,iconst2-seed)
    do n=2,rsize
       rseed(n)=modulo(iconst1*rseed(n-1),iconst2-seed)
    end do

    call random_seed(put=rseed(1:rsize))
    deallocate(rseed)

  end subroutine setSeed


  subroutine mksamplings(dvr)
    implicit none
    ! INPUT
    type(dvr_t), intent(IN)        :: dvr

    ! LOCALS
    integer              :: task, nfile
    logical              :: lbinary
    character(len=c5)    :: efname
    character(len=c2)    :: string1,buf
    double precision     :: ratio, mean

    routine = "mksamplings (mctools.f90)"
    lbinary = .not.  lasciisamplings


    ! samplings...

    do task=1,mc_ntasks

       ! skip if there are no sampling points
       ! to be crated for this task
       if (ntotsample(task,0) .eq. 0) cycle

       do nfile=1,nidxfile(task)

          if (nidxfile(task) .gt. 1) then
             write(buf,'(a,i4)') "-",nfile
             call rmblnk(buf)
             buf=adjustl(buf)
          else
             buf=" "
          end if

          ! we re-use SPP from in name-directory
          if (lloadevecs .and. task .eq. mc_task_spp) cycle

          if (smethod(task,nfile) .eq. mc_complete) cycle



          select case (task)
          case (mc_task_spp)
             string1 = "-spp"
          case(mc_task_coeff)
             string1 = "-coeff"
          case(mc_task_test)
             string1 = "-test"
          case default
             message = "Unknown task. This should not happen"
             call errormsg
          end select
          string1 = trim(string1)//trim(buf)

          efname = trim(name)//"/energies"//trim(string1)

          if (smethod(task,nfile) .ne. mc_readfile) then


             if (smethod(task,nfile) .eq. mc_metropolis) then
                ! Metropolis

                write(idxfileinfo(task,nfile),'(a,3(i8,a),f16.8,a)') "Metropolis,", &
                     ntotsample(task,nfile), ", ", &
                     nskipfirst(task,nfile), ", ", &
                     nstep(task,nfile), ", ", &
                     beta(task,nfile), " au"
                call cncntrblnk(idxfileinfo(task,nfile))
                idxfileinfo(task,nfile) = adjustl(idxfileinfo(task,nfile))

                call MetropolisTrajecotry(dvr, task, nfile, ratio,mean, &
                     idxfile(task,nfile),efname)


                if (mpirank.eq.0) then

                   write(*,'(a,f8.3,a)') "Metropolis sampling"//trim(string1) &
                        // " acceptance rate = ", &
                        ratio*100.d0," %."

                   write(ilog,'(a,f8.3,a)') "Metropolis sampling"//trim(string1) &
                        //" acceptance rate = ", &
                        ratio*100.d0," %."

                   write(*,'(a,3(ES12.3,a))') "Mean energy = ", &
                        mean*219474.63137, " cm-1,  ", &
                        mean*27.2113961318, " eV,  ", &
                        mean, "au."
                   write(ilog,'(a,3(ES12.3,a))') "Mean energy = ", &
                        mean*219474.63137, " cm-1,  ", &
                        mean*27.2113961318, " eV,  ", &
                        mean, "au."
                   write(ilog,*)
                   call myflush(ilog)
                end if

             elseif (smethod(task,nfile) .eq. mc_uniform) then
                ! Uniform
                write(idxfileinfo(task,nfile),'(a,i8)') "Uniform, ", ntotsample(task,nfile)
                call cncntrblnk(idxfileinfo(task,nfile))
                idxfileinfo(task,nfile) = adjustl(idxfileinfo(task,nfile))

                call UniformTrajectory(task,nfile,dvr, &
                     idxfile(task,nfile),efname)
             end if
          else
             if (lesample) then
                call sampletraj(efname, task, nfile, dvr)
             end if
          end if

       end do

    end do

  end subroutine mksamplings



  subroutine sampletraj(efname, task, nfile, dvr)
    implicit none

    character(len=*), intent(IN)  :: efname
    integer, intent(IN)           :: task, nfile
    type(dvr_t), intent(IN)       :: dvr

    integer                       :: n,i,nsample, ierr
    integer, allocatable          :: dvrindex(:,:)
    double precision, allocatable :: energies(:)
    double precision, allocatable :: tmpenergies(:)
    logical                       :: lexist

    if (mpirank.eq.0) then
       inquire(file=efname, exist=lexist)
       if (lexist .and. .not.lovwr .and. .not.lreadidx) then
          routine = "sampletraj"
          message = "File exists, use -w option: "//trim(efname)
          call errormsg
       end if
       open(ivpot,file=efname,status="unknown", &
            form="formatted",iostat=ierr)
       write(ivpot,'(a)') "# "//trim(idxfileinfo(task,nfile))
    end if


    call mpi_set_chunksize(ntotsample(task,nfile),nsample)

    allocate(dvrindex(dvr%ndof,nsample))
    allocate(energies(nsample))

    inquire(file=efname, exist=lexist)
    if (lexist .and. .not.lovwr .and. .not.lreadidx) then
       routine = "sampletraj"
       message = "File exists, use -w option: "//trim(efname)
       call errormsg
    end if

    call loaddvrindex(dvrindex,dvr%ndof,dvr%gdim, &
         nsample,ntotsample(task,nfile:nfile), &
         nskipfirst(task,nfile:nfile), &
         nstep(task,nfile:nfile),idxfile(task,nfile:nfile), 1)

    ! does nothing if already tested
    ! or of no OpenMP or if "no-OMPPotenttail"
    ! is set for PES in RUN-SECTION
    call test_pes_threadsafe(dvrindex, dvr%ndof, min(10000,nsample), ilog)


    !$omp parallel do if(lomppotential)
    do n=1,nsample
       energies(n) = potential(dvrindex(:,n),dvr%ndof)
    end do
    !$omp end parallel do


    do i = 0,nmpi-1
       nsample = mpi_chunksizes(i)
       if (mpirank.eq. 0) then
          if (allocated(tmpenergies)) deallocate(tmpenergies)
          allocate(tmpenergies(nsample))
       end if

       call mpi_get_energies(energies,tmpenergies,nsample,i)

       if (mpirank.eq. 0) then

          do n=1,nsample
             write(ivpot,'(f20.12)',iostat=ierr) tmpenergies(n)
             if (ierr.ne.0) then
                routine = "sampletraj"
                message = "Error while writing to file: "//trim(efname)
                call errormsg
             end if
          end do

       end if

    end do
    if (mpirank.eq. 0) then
       close(ivpot)
    end if

    if (allocated(tmpenergies)) deallocate(tmpenergies)
    deallocate(dvrindex,energies)

  end subroutine sampletraj






end module mctools
