

module mcuniform
    use global
    use dvrtype_m
    use pesselect_s, only: potential
    use mcsampling
    use mcpmpitools

    use channels
    use mpidata

  implicit none


contains


  subroutine UniformTrajectory(task, nfile,dvr, &
       filenameidx,filenameenergies)

    implicit none

    integer, intent(IN)           :: task, nfile
    type(dvr_t), intent(IN)       :: dvr
    character(len=*),intent(IN)   :: filenameidx
    character(len=*),intent(IN)   :: filenameenergies


    integer                       :: n,f,ierr
    logical                       :: lexist
    character(len=c2)             :: fmt,buf
    double precision              :: ran(dvr%ndof),v

    double precision, allocatable :: energies(:), tmpenergies(:)
    integer, allocatable          :: dvrindex(:,:), tmpindex(:,:)
    integer                       :: nsample,i



    ! format for ascii output,

    write(buf,*) dvr%ndof
    buf=adjustl(buf)
    fmt="("//trim(buf)//"i4)"


    if (mpirank.eq.0) then

       ! check if files are there and overwrite is allowed
       inquire(file=filenameidx, exist=lexist)
       if (lexist .and. .not. lovwr) then
          routine = "UniformTrajectory"
          message = "File exists, use -w option: "//trim(filenameidx)
          call errormsg
       end if

       if (lesample) then
          inquire(file=filenameenergies, exist=lexist)
          if (lexist .and. .not. lovwr) then
             routine = "UniformTrajectory"
             message = "File exists, use -w option: "//trim(filenameenergies)
             call errormsg
          end if
       end if

    end if


    call mpi_set_chunksize(ntotsample(task,nfile),nsample)

    allocate(dvrindex(dvr%ndof,nsample))


    sampling: do n=1,nsample
       call random_number(ran)
       do f=1,dvr%ndof
          dvrindex(f,n) = int(ran(f)*dvr%gdim(f)) + 1
       end do
    end do sampling


    if (lesample) then
       ! this is the expensive part:
       allocate(energies(nsample))
       v =  potential(dvrindex(:,1),dvr%ndof)
       !$omp parallel do
       do n=1,nsample
          energies(n) = potential(dvrindex(:,n),dvr%ndof)
       end do
       !$omp end parallel do
    end if

    if (mpirank .eq. 0) then

       ! now open output files
       if (lasciisamplings) then
          open(idump,file=filenameidx,status="unknown", &
               form="formatted",iostat=ierr)
          write(idump,'(a)') "# "//trim(idxfileinfo(task,nfile))
       else
          open(idump,file=filenameidx,status="unknown", &
               form="unformatted",iostat=ierr)
          write(idump) "# "//trim(idxfileinfo(task,nfile))
       end if

       if (ierr.ne.0) then
          routine = "UniformTrajectory"
          message = "File could not be opened: "//trim(filenameidx)
          call errormsg
       end if

       if (lesample) then
          open(ivpot,file=filenameenergies,status="unknown", &
            form="formatted",iostat=ierr)

          if (ierr.ne.0) then
             routine = "UniformTrajectory"
             message = "File could not be opened: "//trim(filenameenergies)
             call errormsg
          end if
       end if
    end if


    do i = 0,nmpi-1
       nsample = mpi_chunksizes(i)
       if (mpirank.eq. 0) then
          nsample = mpi_chunksizes(i)
          if (allocated(tmpindex)) deallocate(tmpindex)
          allocate(tmpindex(dvr%ndof,nsample))
          if (lesample) then
             if (allocated(tmpenergies)) deallocate(tmpenergies)
             allocate(tmpenergies(nsample))
          end if
       end if

       call mpi_get_chunk(dvrindex,tmpindex,dvr%ndof,nsample,i)
       if (lesample) call mpi_get_energies(energies,tmpenergies,nsample,i)

       if (mpirank.eq. 0) then
          do n=1,nsample
             if (lasciisamplings) then
                write(idump,fmt,iostat=ierr) (tmpindex(f,n), f=1,dvr%ndof)
             else
                write(idump,iostat=ierr) (tmpindex(f,n), f=1,dvr%ndof)
             end if
             if (ierr.ne.0) then
                routine = "UniformTrajecotry"
                message = "Error while writing to file: "//trim(filenameidx)
                call errormsg
             end if
          end do
          if (lesample) then
             do n=1,nsample
                write(ivpot,'(f16.8)',iostat=ierr) tmpenergies(n)
                if (ierr.ne.0) then
                   routine = "UniformTrajecotry"
                   message = "Error while writing to file: "//trim(filenameenergies)
                   call errormsg
                end if
             end do
          end if
       end if

    end do


    if (mpirank.eq. 0) then
       close(idump)
       if (lesample) close(ivpot)

       write(buf,'(i8)') ntotsample(task,nfile)
       buf = adjustl(buf)
       write(ilog,'(a)') ""
       write(ilog,'(a)') "Wrote "//trim(buf)//" sampling points to file "//trim(filenameidx)
       if (lesample) then
          write(ilog,'(a)') "Wrote matching energies to file "//trim(filenameenergies)
       else
          write(ilog,'(a)') "Matching energies not requested."
       end if
    end if

    if (allocated(tmpindex)) deallocate(tmpindex)
    if (allocated(tmpenergies)) deallocate(tmpenergies)


  end subroutine UniformTrajectory


end module mcuniform
