! -*- f90 -*-


module pesselect_m
  use dvrtype_m

  implicit none

  logical                    :: lomppotential
  logical                    :: lthreadsafetested
  logical                    :: lidxbased
  procedure(pot),pointer     :: potential
  procedure(potQ),pointer    :: potentialQ
  type(dvr_t), pointer       :: pdvr
  double precision, pointer  :: phops(:)

  abstract interface

     function pot(idx,ndof)
       double precision             :: pot
       integer, intent(IN)          :: ndof
       integer, intent(IN)          :: idx(ndof)
     end function Pot

     function potQ(coord,ndof)
       double precision             :: potQ
       integer, intent(IN)          :: ndof
       double precision, intent(IN) :: coord(ndof)
     end function PotQ

  end interface


contains

  subroutine pesselect_default()
    implicit none

    potential => PotZero
    potentialQ => PotZeroQ

    nullify(pdvr)
    nullify(phops)
    potential =>  NULL()
    potentialQ => Null()

    lomppotential = .true.
    lthreadsafetested = .false.
    lidxbased = .false.

  end subroutine pesselect_default

  ! grid version
  double precision function PotZERO(idx,ndof1)
    implicit none
    integer, intent(IN)          :: ndof1
    integer, intent(IN)          :: idx(ndof1)

    PotZERO = 0.d0

  end function PotZERO

  ! coord version
  double precision function PotZEROQ(coord,ndof1)
    implicit none
    integer, intent(IN)          :: ndof1
    double precision, intent(IN) :: coord(ndof1)

    PotZEROQ = 0.d0

  end function PotZEROQ


end module pesselect_m

module pesselect_s
  use global
  use pglobal
  use maxv
  use pesselect_m
  use operdef
  use rshdmr
  use psiutil, only: f2mdecoder
  use callmuldmod, only: callmuld
  use operutil, only: vcalc1st
  use logdat

#ifdef OMP
  use omp_lib
#endif

contains

  subroutine AssignPot(hoplab,dvr,ilog1)
    
    implicit none
    character(len=*), intent(IN)      :: hoplab
    type(dvr_t), intent(IN), target   :: dvr
    integer, intent(IN), optional     :: ilog1

    logical                           :: lwrlog1

    pdvr => dvr

    lwrlog = .false.
    if (present(ilog1)) then
       inquire(unit=ilog1, opened=lwrlog1)
    end if

    select case (hoplab)
    case ("usersurf")
       ! use code from userpot.f90
       potential => UserImplementation
       potentialQ => UserImplementationQ
       if (lwrlog) then
          write(ilog1, '(a)') "Potential is obtained from userpot.f90"
       end if
    case ("usersurf-idx")
       ! use code from userpotidx.f90
       potential => UserImplementationIdx
       potentialQ => UserImplementationIdxQ
       lidxbased = .true.
       if (lwrlog) then
          write(ilog, '(a)') "Operator is obtained from userpot.f90"
       end if

    case ("pes-file")
       ! re-fit a PES file
       potential => PotFromPes
       if (lwrlog1) then
          write(ilog1, '(a)') "Potential re-fit from a PES file."
       end if
    case default
       ! FALLBACK:
       ! a surface from the operator lib
       potential => PotFromMCTDH
       potentialQ => PotFromMCTDHQ
       if (lwrlog) then
          write(ilog1, '(a)') "Potential is MCTDH built-in."
       end if
    end select

    if (lwrlog1) then

#ifdef OMP
       ! only print this if it matters
       if (lomppotential) then
          write(ilog1, '(a)') "Potential evaluation is assumed thread-safe with OpenMP."
       else
          write(ilog1, '(a)') "Potential evaluation is assumed NOT thread-safe with OpenMP."
       end if
#endif
       write(ilog1, '(a)') ""

    end if

  end subroutine AssignPot


  ! call a function from the mctdh operator library
  ! grid version
  double precision function PotFromMCTDH(idx,ndof1)
    implicit none
    integer, intent(IN)          :: ndof1
    integer, intent(IN)          :: idx(ndof1)

    double precision             :: v,vpes,vrshdmr, rshdmrbymidx
    double precision             :: coord(maxdim)
    integer                      :: mpoint(maxdim)

    integer                      :: f,z1,hl,ddst
    external rshdmrbymidx


    do f=1,ndof1
       z1 = pdvr%zort(f) + idx(f) -1
       coord(f) = pdvr%ort(z1)
    end do

    hl=hopdim
    ddst=1
    call callmuld(v,coord,ndof1,hopilab(hl),hopfile(hl), &
         hopipar(1,hl),hoprpar(1,hl),ddst,ddst,hoplab(hl))

    if (lpesfile) then
       if (nmode .eq. 0) then
          routine = "PotFromMCTDH"
          message = "nmode == 0: error while evaluating PES from file."
          call errormsg
       end if
       call f2mdecoder(idx,mpoint)
       call vcalc1st(phops,vpes,mpoint,1)
       v = v - vpes
    end if

    if (lrshdmr .and. lrshdmr_ready) then
       call f2mdecoder(idx,mpoint)
       vrshdmr = rshdmrbymidx(mpoint,nmode)
       v = v - vrshdmr
    end if


    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if

    PotFromMCTDH = v

  end function PotFromMCTDH


  ! call a function from the mctdh operator library
  ! coord version
  double precision function PotFromMCTDHQ(coord,ndof1)
    implicit none
    integer, intent(IN)          :: ndof1
    double precision, intent(IN) :: coord(ndof1)

    double precision             :: v
    integer                      :: hl,ddst

    hl=hopdim
    ddst=1
    call callmuld(v,coord,ndof1,hopilab(hl),hopfile(hl), &
         hopipar(1,hl),hoprpar(1,hl),ddst,ddst,hoplab(hl))

    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if

    PotFromMCTDHQ = v

  end function PotFromMCTDHQ




  ! call a function the user has implemented
  double precision function UserImplementation(idx,ndof1)

    implicit none

    integer, intent(IN)          :: ndof1
    integer, intent(IN)          :: idx(ndof1)


    double precision             :: v, vpes,vrshdmr, rshdmrbymidx
    double precision             :: coord(maxdim)
    integer                      :: mpoint(maxdim)

    ! external function see 'userpot.f90'
    double precision             :: UserPot

    integer                      :: f,z1
    external rshdmrbymidx


    do f=1,ndof1
       z1 = pdvr%zort(f) + idx(f) -1
       coord(f) = pdvr%ort(z1)
    end do

    v = UserPot(coord,ndof1)

    if (lpesfile) then

       call f2mdecoder(idx,mpoint)

       call vcalc1st(phops,vpes,mpoint,1)
       v = v - vpes
    end if

    if (lrshdmr .and. lrshdmr_ready) then
       call f2mdecoder(idx,mpoint)
       vrshdmr = rshdmrbymidx(mpoint,nmode)
       v = v - vrshdmr
    end if


    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if
    UserImplementation = v

  end function UserImplementation


  ! call a function the user has implemented
  double precision function UserImplementationQ(coord,ndof1)
    implicit none
    integer, intent(IN)          :: ndof1
    double precision, intent(IN) :: coord(ndof1)

    double precision             :: v
    ! external function see 'userpot.f90'
    double precision             :: UserPot

    v = UserPot(coord,ndof1)
    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if
    UserImplementationQ = v

  end function UserImplementationQ

 ! call a function the user has implemented in the grid
  double precision function UserImplementationIdx(idx,ndof1)
    implicit none 
       
    integer, intent(IN)          :: ndof1
    integer, intent(IN)          :: idx(ndof1)
          
    double precision             :: v, vpes,vrshdmr, rshdmrbymidx
    integer                      :: mpoint(maxdim)

    ! external function see 'userpotidx.f90'
    double precision             :: UserPotIdx
       
    integer                      :: f,z1
    external rshdmrbymidx

    v = UserPotIdx(idx,ndof1)

    if (lrshdmr .and. lrshdmr_ready) then
       call f2mdecoder(idx,mpoint)
       vrshdmr = rshdmrbymidx(mpoint,nmode)
       v = v - vrshdmr
    end if


    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if
    UserImplementationIdx = v

  end function UserImplementationIdx


 ! call a function the user has implemented in the grid
  double precision function UserImplementationIdxQ(coord,ndof1)
    implicit none

    integer, intent(IN)          :: ndof1
    double precision, intent(IN) :: coord(ndof1)

    print*, "Error, call to UserImplementationIdxQ"
    print*, "But user only provides index routine."
    stop 1

    UserImplementationIdxQ = 0.d0

  end function UserImplementationIdxQ


  ! Call a function that re-fits a PES file
  double precision function PotFromPes(idx,ndof1)
    implicit none

    integer, intent(IN)          :: ndof1
    integer, intent(IN)          :: idx(ndof1)

    double precision             :: v,vrshdmr,rshdmrbymidx
    integer                      :: mpoint(maxdim)
    external rshdmrbymidx

    call f2mdecoder(idx,mpoint)

    call vcalc1st(phops,v,mpoint,1)

    if (lrshdmr .and. lrshdmr_ready) then
       call f2mdecoder(idx,mpoint)
       vrshdmr = rshdmrbymidx(mpoint,nmode)
       v = v - vrshdmr
    end if


    if (v .gt. pesmax) then
       v = pesmax
    elseif (v .lt. pesmin) then
       v = pesmin
    end if

    PotFromPes = v

  end function PotFromPes




! do a number of parallel potential evaluations
! and compare to serial evaluation. Produce an error
! if they do not coincide.
subroutine test_pes_threadsafe(dvrindex, ndof1, ntest,ilog1)
  implicit none
  integer, intent(IN)  :: ndof1, ntest, ilog1
  integer, intent(IN)  :: dvrindex(ndof1,ntest)

#ifdef OMP

  integer, parameter   :: maxfail=5
  double precision     :: eserial(maxfail), eparallel(ntest), eparfail(maxfail)
  integer              :: n, ifail, nthread
  logical              :: lthreadsafe

  ! early exit if no parallelization is used
  ! or test was done before
  if (lthreadsafetested .or. .not. lomppotential) return

  !$omp parallel
  nthread = omp_get_num_threads()
  !$omp end parallel
  if (nthread .eq. 1) return

  !$omp parallel do
  do n=1,ntest
     eparallel(n) = potential(dvrindex(:,n), ndof1)
  end do
  !$omp end parallel do

  lthreadsafe = .true.
  ifail = 1
  do n=1,ntest
     eserial(ifail) = potential(dvrindex(:,n), ndof1)
     if ( abs(eparallel(n)-eserial(ifail)) .gt. 1.d-8) then
        lthreadsafe = .false.
        eparfail(ifail) = eparallel(n)
        ifail = ifail + 1
     end if
     if (ifail.gt. maxfail) exit ! loop
  end do


  if (.not. lthreadsafe) then
     write(ilog1,'(/a)') "Parallel evaluation of PES routine failed:"
     write(ilog1,'(a)') "  fail         Pot(serial)       " &
          //"Pot(parallel)                diff"

     do n=1,ifail - 1
        write(ilog1,'(i6,3f20.10)') n, eserial(n), eparfail(n), &
             eserial(n)-eparfail(n)
     end do
     if (ifail .gt. maxfail) then
        write(ilog1,'(a)') " (skipping more fails...)"

     end if

     message = "PES is not thread-safe. See log. " &
          //"Use the no-OMPPotential flag in the " &
          //"RUN-SECTION or use MPI instead."
     routine = "test_pes_threadsafe"
     call errormsg
  end if

#endif

  lthreadsafetested = .true.

end subroutine test_pes_threadsafe


end module pesselect_s
