!   -*- f90 -*-
!*******************************************************************
! module pftools
!
! Contains a subroutines useful for manupulating natpots. some of
! these are f90 versions of existing subroutines
!
! Contents:
!
! f2mdecoder     - convert a dvr grid point index array to a mode
!                  grid point index array
! m2fdecoder     - convert a mode index point array to a dvr
!                  grid index point array
! cindex2sindex  - convert a combined index to an index array
! sindex2cindex  - convert an index array to the combined index
! pfcoeffbyindex - return the coefficient of a natpot given a mode
!                  index array
! pfcoeffbyindex - return the coefficient of a natpot given a mode
!                  index array (faster if called many times
!                  in a row, but needs extra workspace)
! incrementidx   - increment an index array by one
! pfbyindex      - return the value of a natpot given a
!                  dvr index array
! testpf         - test a natpot along a trajectory
! trsppproj      - for two natpots return the trace of the product
!                  of the projectors on the spp space
! comb_prod      - Build a configuration product for a sampling point
! wk_comb_prod   - Determine workspace needed by comb_prod
!
!*******************************************************************


module pftools
  use global
  use pglobal
  use timing
  use tensortype
  use dvrtype_m
  use natpottype
  use cpdtype_m
  use xvlib, only: mvtxdd
  use op1lib, only: cuttxd2
  use mpidata

#ifdef MPI
  use mpi
#endif

  implicit none

  interface dvrindex2coord
     module procedure dvrindex2coord_dvr
     module procedure dvrindex2coord_tensor
  end interface dvrindex2coord


contains

    ! convert a dvr index point to a coordinate array
  subroutine dvrindex2coord_tensor(tensor,idx,coord)
    implicit none
    class (tensor_t), intent(IN)   :: tensor
    integer, intent(IN)            :: idx(tensor%ndof)
    double precision, intent(OUT)  :: coord(tensor%ndof)
    integer                        :: f,z

    do f=1,tensor%ndof
       z = tensor%zort(f) + idx(f) - 1
       coord(f) = tensor%ort(z)
    end do

  end subroutine dvrindex2coord_tensor

  subroutine dvrindex2coord_dvr(dvr,idx,coord)
    implicit none
    type (dvr_t), intent(IN)       :: dvr
    integer, intent(IN)            :: idx(dvr%ndof)
    double precision, intent(OUT)  :: coord(dvr%ndof)
    integer                        :: f,z

    do f=1,dvr%ndof
       z = dvr%zort(f) + idx(f) - 1
       coord(f) = dvr%ort(z)
    end do

  end subroutine dvrindex2coord_dvr


  ! transform a primitive-basis-index to a combined-mode-index
  subroutine f2mdecoder(fpoint,mpoint,nspfdof,spfdof,gdim,ndof,nmode,maxspfdof)
    implicit none
    integer, intent(IN)            :: ndof,nmode,maxspfdof
    integer, intent(IN)            :: fpoint(ndof),gdim(ndof), nspfdof(nmode)
    integer, intent(IN)            :: spfdof(maxspfdof,nmode)
    integer, intent(OUT)           :: mpoint(nmode)

    integer                        :: m,f,n,vdim1

    do m=1,nmode
       mpoint(m) = 0
       vdim1 = 1
       do n=1,nspfdof(m)
          f = spfdof(n,m)
          mpoint(m) = mpoint(m) + (fpoint(f)-1)*vdim1
          vdim1 = vdim1 * gdim(f)
       end do
       mpoint(m) = mpoint(m) + 1
    end do

  end subroutine f2mdecoder

  ! transform acombined-mode-index to a primitive-basis-index
  subroutine m2fdecoder(mpoint, fpoint, nmode, ndof, spfdof, &
       nspfdof, gdim, maxspfdof)
    implicit none
    integer, intent(IN)        :: nmode, ndof, maxspfdof
    integer, intent(IN)        :: mpoint(nmode), gdim(ndof)
    integer, intent(IN)        :: spfdof(maxspfdof,nmode)
    integer, intent(IN)        :: nspfdof(nmode)
    integer, intent(OUT)       :: fpoint(ndof)

    integer                    :: tmpidx(maxspfdof),f,m,n
    integer                    :: tmpgdim(maxspfdof)


    do m=1,nmode

       do n=1,nspfdof(m)
          f = spfdof(n,m)
          tmpgdim(n) = gdim(f)
       end do

       call cindex2sindex(mpoint(m),tmpidx,tmpgdim,nspfdof(m))

       do n=1,nspfdof(m)
          f = spfdof(n,m)
          fpoint(f) = tmpidx(n)
       end do

    end do

  end subroutine m2fdecoder


  ! combined index to array of single indices
  subroutine cindex2sindex(cindex,sindex,sdim,nmode)
    implicit none
    integer, intent(IN)             :: cindex,nmode
    integer, intent(IN)             :: sdim(nmode)
    integer, intent(OUT)            :: sindex(nmode)
    integer                         :: m,cindex1

    cindex1 = cindex -1

    do m=1,nmode
       if (sdim(m) .ne.0) then ! potdim .eq. 0 -> contr. mode
          sindex(m) = mod(cindex1,sdim(m)) + 1
          cindex1 = cindex1/sdim(m)
       end if
    end do

  end subroutine cindex2sindex

  ! array of single indices to combined index
  subroutine sindex2cindex(cindex,sindex,sdim,nmode)
    implicit none
    integer, intent(IN)             :: nmode
    integer, intent(IN)             :: sindex(nmode),sdim(nmode)
    integer, intent(OUT)            :: cindex
    integer                         :: m, vdim1

    cindex = 0
    vdim1 = 1
    do m=1,nmode
       if (sdim(m) .ne. 0) then ! potdim .eq. 0 -> contr. mode
          cindex = cindex + (sindex(m)-1)*vdim1
          vdim1 = vdim1*sdim(m)
       end if
    end do
    cindex = cindex + 1

  end subroutine sindex2cindex


  ! returns S(j) = A(i,j,k)
  subroutine dgetfiber3d(A,S,dim1,dim2,dim3,i,k)
    implicit none
    integer, intent(IN)           :: dim1,dim2,dim3,i,k
    double precision, intent(IN)  :: A(dim1,dim2,dim3)
    double precision, intent(OUT) :: S(dim2)
    integer                       :: j

    do j=1,dim2
       S(j) = A(i,j,k)
    end do

  end subroutine dgetfiber3d



  ! get a single coefficient by mode index
  double precision function pfcoeffbyindex(natpot,midx)
    implicit none
    type (natpot_t), intent(IN)    :: natpot
    integer, intent(IN)            :: midx(natpot%nmode)
    integer                        :: m, totidx, vdim1

    totidx = 1
    vdim1 = 1
    do m=1,natpot%nmode
       totidx = totidx + (midx(m)-1)*vdim1
       vdim1 = vdim1 * natpot%potdim(m)
    end do
    pfcoeffbyindex = natpot%coeff(totidx)

  end function pfcoeffbyindex


  ! increments an index array by 1
  ! (1-based, first index runs fastest)
  ! gdim(f) is the max value for idx(f)
  ! resets index to (1,1,1,...,1) on overflow
  ! no checks for valid index on entry.
  subroutine incrementidx(idx,gdim,ndof)
    implicit none
    integer, intent(IN)            :: ndof
    integer, intent(IN)            :: gdim(ndof)
    integer, intent(INOUT)         :: idx(ndof)

    integer                        :: f

    idx(1) = idx(1) + 1
    do f=1,ndof-1
       if (idx(f) .gt.gdim(f)) then
          idx(f+1) = idx(f+1) + 1
          idx(f) = 1
       else
          return
       end if
    end do

    if (idx(ndof) .gt.gdim(ndof)) idx(ndof) = 1

  end subroutine incrementidx


  ! increments an index array by 1 under the
  ! constraint that sum(idx) is constant
  ! (1-based, first index runs fastest).
  ! gdim(f) is the max value for idx(f)
  ! sets index to (0,0,0,0,...,0) on overflow
  ! no checks for valid index on entry.
  ! starts with decrementing index 1 and
  ! incrementing index 2 and so forth
  !
  ! This is the quick and dirty method, sorry
  subroutine incrementidxconst(idx,gdim,ndof)
    implicit none
    integer, intent(IN)            :: ndof
    integer, intent(IN)            :: gdim(ndof)
    integer, intent(INOUT)         :: idx(ndof)
    integer                        :: const, const1

    const = sum(idx)
    const1 = const-1
    do while(const1 .ne. const)
       call incrementidx(idx,gdim,ndof)
       const1 = sum(idx)
       if (const1 .eq. ndof) then
          idx = 0
          exit
       end if
    end do

  end subroutine incrementidxconst



  ! get the value of a natpot at a given mode-index
  double precision function pfbymodeindex(natpot,mgidx)
    implicit none
    type (natpot_t), intent(IN)    :: natpot
    integer, intent(IN)            :: mgidx(natpot%nmode)

    integer                        :: cidx(natpot%nmode) !automatic alloc
    integer                        :: mc,m,n,z, vdim1
    double precision               :: spp, sppgpoints(natpot%nmode,natpot%maxspf)


    ! copy spf points to local array
    sppgpoints = 0.d0
    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then
          do n=1,natpot%potdim(m)
             z = natpot%zevec(m,n) + mgidx(m) - 1
             sppgpoints(m,n) = natpot%evec(z)
          end do
       end if
    end do

    ! run over all core
    cidx = 1
    pfbymodeindex = 0.d0

    do mc=1,natpot%modcdim

       spp=1.d0
       do m=1,natpot%nmode
          if (m .ne. natpot%modc) then
             n = cidx(m)
             spp = spp * sppgpoints(m,n)
          end if
       end do

       if (natpot%modc .ne. 0) then
          z = 1
          vdim1 = 1
          do m=1,natpot%nmode

             if (m.eq.natpot%modc) then
                z = z + (mgidx(m)-1)*vdim1
                vdim1 = vdim1 * natpot%subdim(m)
             else
                z = z + (cidx(m)-1)*vdim1
                vdim1 = vdim1 * natpot%potdim(m)
             end if
          end do
       else
          z = mc
       end if

       pfbymodeindex = pfbymodeindex + natpot%coeff(z)*spp

       call incrementidx(cidx,natpot%potdim,natpot%nmode)
    end do

  end function pfbymodeindex


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! wkpfbymodeindex: calculate workspace needed by pfbymodeindex1
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  integer function wkpfbymodeindex(natpot)
    implicit none
    type (natpot_t), intent(IN)            :: natpot
    integer                                :: w1dim, w2dim

    if (natpot%modc .ne. 0) then
       w1dim = natpot%cdim/natpot%subdim(natpot%modc)
       if (natpot%modc .eq. 1) then
          w2dim = w1dim/natpot%potdim(2)
       else
          w2dim = w1dim/natpot%potdim(1)
       end if
    else
       w1dim = natpot%cdim/natpot%potdim(1)
       w2dim = w1dim/natpot%potdim(2)
    end if

    wkpfbymodeindex = w1dim + w2dim


  end function wkpfbymodeindex

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! pfbymodeindex1: calculate vfit given a grid point in combined mode
  !                 indices
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  double precision function pfbymodeindex1(natpot,mgidx, work, workdim)
    implicit none
    type (natpot_t), intent(IN)            :: natpot
    integer, intent(IN)                    :: mgidx(natpot%nmode)
    integer, intent(IN)                    :: workdim
    double precision, intent(OUT)          :: work(workdim)

    integer                                :: z,m,n
    integer                                :: w1ptr, w2ptr, swp
    integer                                :: dim1,dim2,dim3,jcut
    double precision                       :: sppgpoints(natpot%maxspf,natpot%nmode)
    pfbymodeindex1 = 0.d0

    w1ptr = 1

    if (natpot%modc .ne. 0) then
       w2ptr = natpot%cdim/natpot%subdim(natpot%modc) +1
    else
       w2ptr = natpot%cdim/natpot%potdim(1) +1
    end if


    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then
          do n=1,natpot%potdim(m)
             z = natpot%zevec(m,n) + mgidx(m) - 1
             sppgpoints(n,m) = natpot%evec(z)
          end do
       end if
    end do



    if (natpot%modc .ne. 0) then
       m = natpot%modc
       dim1 = natpot%potvdim(m)
       dim2 = natpot%subdim(m)
       dim3 = natpot%potndim(m)
       jcut = mgidx(m)

       call cuttxd2(natpot%coeff, work(w1ptr),dim1,dim2,dim3,jcut)

       dim2 = natpot%potvdim(m) * natpot%potndim(m)

       do m=1, natpot%nmode
          if (m .eq. natpot%modc) cycle

          dim1 = natpot%potdim(m)
          dim2 = dim2/dim1

          call mvtxdd(work(w1ptr),sppgpoints(:,m),work(w2ptr),dim1,dim2)
          swp = w1ptr
          w1ptr = w2ptr
          w2ptr = swp

       end do

       pfbymodeindex1 = work(w1ptr)

    else

       dim1 = natpot%potdim(1)
       dim2 = natpot%potndim(1)

       call mvtxdd(natpot%coeff,sppgpoints(:,1),work(w1ptr),dim1,dim2)

       do m=2, natpot%nmode


          dim1 = natpot%potdim(m)
          dim2 = dim2/dim1

          call mvtxdd(work(w1ptr),sppgpoints(:,m),work(w2ptr),dim1,dim2)
          swp = w1ptr
          w1ptr = w2ptr
          w2ptr = swp

       end do

       pfbymodeindex1 = work(w1ptr)


    end if


  end function pfbymodeindex1


!!$
!!$  double precision function cpdbymodeindex(cpd,mgidx)
!!$    implicit none
!!$
!!$    type(cpd_t), intent(IN) :: cpd
!!$    integer, intent(IN)     :: mgidx(cpd%nmode)
!!$
!!$    integer                 :: r,m,z
!!$    double precision        :: omega(cpd%rank,cpd%nmode)
!!$    double precision        :: term(cpd%rank), spe
!!$    spe = 0.d0
!!$
!!$    !$omp parallel do collapse(2) private(z)
!!$    do m=1,cpd%nmode
!!$       do r=1,cpd%rank
!!$          z = cpd%zspp(r,m) + mgidx(m) - 1
!!$          omega(r,m) = cpd%spp(z)
!!$       end do
!!$    end do
!!$    !$omp end parallel do
!!$
!!$    !$omp parallel do
!!$    do r=1,cpd%rank
!!$       term(r) = cpd%coeff(r)*omega(r,1)
!!$
!!$    end do
!!$    !$omp end parallel do
!!$
!!$    !$omp parallel do collapse(2)
!!$    do m=2,cpd%nmode
!!$       do r=1,cpd%rank
!!$          term(r) = term(r)*omega(r,m)
!!$       end do
!!$    end do
!!$    !$omp end parallel do
!!$
!!$    spe = 0.d0
!!$    !$omp parallel do reduction(+:spe)
!!$    do r=1,cpd%rank
!!$       spe = spe + term(r)
!!$    end do
!!$    !$omp end parallel do
!!$
!!$    cpdbymodeindex = spe
!!$
!!$  end function cpdbymodeindex

  double precision function cpdbymodeindex1(cpd,mgidx)
    implicit none

    type(cpd_t), intent(IN) :: cpd
    integer, intent(IN)     :: mgidx(cpd%nmode)

    integer                 :: r,m,z
    double precision        :: term, spe

    spe = 0.d0

    !$omp parallel do private(term,m,z) reduction(+:spe)
    do r=1,cpd%rank
       term = cpd%coeff(r)
       do m=1,cpd%nmode
          z = mgidx(m) 
          ! this produces a lot of cache misses:
          term = term*cpd%spp(m)%htm(z,r)
       end do
       spe = spe + term
    end do
    !$omp end parallel do

    cpdbymodeindex1 = spe

  end function cpdbymodeindex1


  double precision function cpdbymodeindex(cpd,mgidx)
    implicit none

    type(cpd_t), intent(IN) :: cpd
    integer, intent(IN)     :: mgidx(cpd%nmode)

    integer                 :: m,z
    double precision        :: term, ddot
    double precision, allocatable :: omega(:)

    allocate(omega(cpd%rank))
    m = 1
    z = mgidx(m)
    omega = cpd%spp(m)%htm(z,:)

    do m=2,cpd%nmode
       z = mgidx(m)
       omega = omega*cpd%spp(m)%htm(z,:)
    end do

    cpdbymodeindex = ddot(cpd%rank,cpd%coeff,1,omega,1)

    deallocate(omega)
  end function cpdbymodeindex


  ! get the value of a natpot at a given dvr-index
  double precision function pfbyindex(natpot,fidx)
    implicit none
    type (natpot_t), intent(IN)    :: natpot
    integer, intent(IN)            :: fidx(natpot%ndof)

    integer                        :: mgidx(natpot%nmode) !automatic alloc


    ! get grid index in mode system
    call f2mdecoder(fidx,mgidx,natpot%nspfdof,natpot%spfdof, &
         natpot%gdim,natpot%ndof,natpot%nmode,natpot%maxspfdof)

    pfbyindex = pfbymodeindex(natpot,mgidx)

  end function pfbyindex


  ! get the value of a candecomp at a given dvr-index
  double precision function cpdbyindex(cpd,fidx)
    implicit none
    type (cpd_t), intent(IN)       :: cpd
    integer, intent(IN)            :: fidx(cpd%ndof)

    integer                        :: mgidx(cpd%nmode) !automatic alloc


    ! get grid index in mode system
    call f2mdecoder(fidx,mgidx,cpd%nspfdof,cpd%spfdof, &
         cpd%gdim,cpd%ndof,cpd%nmode,cpd%maxspfdof)

    cpdbyindex = cpdbymodeindex(cpd,mgidx)

  end function cpdbyindex





  ! test a natpot against a vpot
  ! calculates mean, rms, minval and maxval of the difference to
  ! the original routine
  subroutine testpfcmplte(natpot,vpot,mean,rms,vomin,vomax,vmin,vmax)
    implicit none
    type(natpot_t), intent(IN)      :: natpot
    double precision, intent(IN)    :: vpot(natpot%vpotdim)
    double precision, intent(OUT)   :: mean,rms,vmin,vmax,vomin,vomax

    integer                         :: midx(natpot%nmode)
    integer                         :: i, lenwork
    double precision                :: vpf, vdiff
    double precision, allocatable   :: work(:)

    vmin =  1.d99
    vmax = -1.d99
    mean =  0.d0
    rms =  0.d0
    vomin = vpot(1)
    vomax = vpot(1)

    lenwork = wkpfbymodeindex(natpot)
    allocate(work(lenwork))


    !$omp parallel do private(vpf,vdiff,midx,work) &
    !$omp reduction(+:mean,rms) reduction(max:vmax,vomax) &
    !$omp reduction(min:vmin, vomin) schedule(static)
    do i=1,natpot%vpotdim

       call cindex2sindex(i,midx,natpot%subdim,natpot%nmode)

       vpf = pfbymodeindex1(natpot,midx,work,lenwork)
       vdiff =  vpf - vpot(i)
       if (vdiff .lt. vmin) vmin = vdiff
       if (vdiff .gt. vmax) vmax = vdiff
       if (vpot(i) .lt. vomin) vomin = vpot(i)
       if (vpot(i) .gt. vomax) vomax = vpot(i)
       mean = mean + vdiff
       rms = rms + vdiff**2

    end do
    !$omp end parallel do

    deallocate(work)

    mean = mean/natpot%vpotdim
    rms = dsqrt(rms/natpot%vpotdim)

  end subroutine testpfcmplte

  ! test a natpot along a trajectory (dvrindex) of length nsample
  ! against the exact potential from a routine
  ! calculates mean, rms, minval and maxval of the difference to
  ! the original routine
  subroutine testpf(potential,tensor,dvrindex,nsample, &
       mean,rms,vomin,vomax,vmin,vmax,lomppotential)

    implicit none
    class(tensor_t), intent(IN)     :: tensor
    integer, intent(IN)             :: nsample
    integer, intent(IN)             :: dvrindex(tensor%ndof,nsample)
    double precision, intent(OUT)   :: mean,rms,vmin,vmax,vomin,vomax
    logical, intent(IN)             :: lomppotential

    integer                         :: n, nsample1
    double precision                :: vpf, v, vdiff
    double precision                :: potential

    integer(long), save                        :: tid=0 ! timer ID

#ifdef MPI
    integer                         :: ierr
    double precision                :: rdum
#endif

    external potential

    if (mpirank.eq.0) then
       call get_timer(tid, 'ptfiterr')
       call start_timer(tid)
    endif 

    ! we do the first call of potential outside the
    ! parallel region to make the PES routine
    ! call its setup routine if needed.
    ! calling a setup routine should be thread
    ! safe, but you never know when using
    ! third party routines...

    !vmin = 1.d99
    !vmax = -1.d99
    !mean = 0.d0
    !rms = 0.d0

    select type(tensor)
    type is (natpot_t)
       vpf = pfbyindex(tensor,dvrindex(:,1))
    type is (cpd_t)
       vpf = cpdbyindex(tensor,dvrindex(:,1))
    end select

    v = potential(dvrindex(:,1),tensor%ndof)
    vomin = v
    vomax = v
    vdiff = vpf - v
    vmin = vdiff
    vmax = vdiff
    mean = vdiff
    rms = vdiff**2

    ! making the case here is much faster
    ! then inserting an !$omp critical
    ! even the "!$omp if" seems to be slower
    if (lomppotential) then

       !$omp parallel do private(vpf,v,vdiff) &
       !$omp reduction(+:mean,rms) reduction(max:vmax,vomax) &
       !$omp reduction(min:vmin,vomin) schedule(static)
       do n=2,nsample

          select type(tensor)
          type is (natpot_t)
             vpf = pfbyindex(tensor,dvrindex(:,n))
          type is (cpd_t)
             vpf = cpdbyindex(tensor,dvrindex(:,n))
          end select

          v = potential(dvrindex(:,n),tensor%ndof)

          vdiff = vpf - v
          if (vdiff .lt. vmin) vmin = vdiff
          if (vdiff .gt. vmax) vmax = vdiff
          if (v .lt. vomin) vomin = v
          if (v .gt. vomax) vomax = v
          mean = mean + vdiff
          rms = rms + vdiff**2
       end do
       !$omp end parallel do

    else

       !$omp master
       do n=2,nsample

          select type(tensor)
          type is (natpot_t)
             vpf = pfbyindex(tensor,dvrindex(:,n))
          type is (cpd_t)
             vpf = cpdbyindex(tensor,dvrindex(:,n))
          end select

          v = potential(dvrindex(:,n),tensor%ndof)

          vdiff = vpf - v
          if (vdiff .lt. vmin) vmin = vdiff
          if (vdiff .gt. vmax) vmax = vdiff
          if (v .lt. vomin) vomin = v
          if (v .gt. vomax) vomax = v
          mean = mean + vdiff
          rms = rms + vdiff**2
       end do
       !$omp end master
    end if


    nsample1 = nsample


#ifdef MPI

    call MPI_AllReduce(mean,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
    mean = rdum

    call MPI_AllReduce(rms,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
    rms = rdum

    call MPI_AllReduce(vmin,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,ierr)
    vmin = rdum

    call MPI_AllReduce(vomin,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,ierr)
    vomin = rdum

    call MPI_AllReduce(vmax,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,ierr)
    vmax = rdum

    call MPI_AllReduce(vomax,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,ierr)
    vomax = rdum

    call MPI_AllReduce(nsample,nsample1,1,&
         MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)

#endif

    mean = mean/nsample1
    rms = dsqrt(rms/nsample1)

    if (mpirank .eq. 0) call stop_timer(tid)

  end subroutine testpf


  ! test a cpd along a trajectory (dvrindex) of length nsample
  ! against the exact potential from a routine
  ! calculates mean, rms, minval and maxval of the difference to
  ! the original routine
  subroutine testcpd(potential,cpd,dvrindex,nsample, &
       mean,rms,vomin,vomax,vmin,vmax,lomppotential)

    implicit none
    type(cpd_t), intent(IN)         :: cpd
    integer, intent(IN)             :: nsample
    integer, intent(IN)             :: dvrindex(cpd%ndof,nsample)
    double precision, intent(OUT)   :: mean,rms,vmin,vmax,vomin,vomax
    logical, intent(IN)             :: lomppotential

    integer                         :: n, nsample1
    double precision                :: vpf, v(nsample), vdiff
    double precision                :: potential

    integer(long), save             :: tid=0 ! timer ID

#ifdef MPI
    integer                         :: ierr
    double precision                :: rdum
#endif

    external potential

    if (mpirank .eq. 0) then
       call get_timer(tid, 'ptfiterr')
       call start_timer(tid)
    endif

   ! first call potential in
    ! parallel if possible
    ! then the rest is thread safe

    !$omp parallel do if(lomppotential) schedule(static)
    do n=1,nsample
       v(n) = potential(dvrindex(:,n),cpd%ndof)
    end do
    !$omp end parallel do


    vpf = cpdbyindex(cpd,dvrindex(:,1))

    vomin = v(1)
    vomax = v(1)
    vdiff = vpf - v(1)
    vmin = vdiff
    vmax = vdiff
    mean = vdiff
    rms = vdiff**2


    !$omp parallel do private(vpf,vdiff) &
    !$omp reduction(+:mean,rms) reduction(max:vmax,vomax) &
    !$omp reduction(min:vmin,vomin) schedule(static)
    do n=2,nsample

       vpf = cpdbyindex(cpd,dvrindex(:,n))
       vdiff = vpf - v(n)

       if (vdiff .lt. vmin) vmin = vdiff
       if (vdiff .gt. vmax) vmax = vdiff
       if (v(n) .lt. vomin) vomin = v(n)
       if (v(n) .gt. vomax) vomax = v(n)
       mean = mean + vdiff
       rms = rms + vdiff**2
    end do
    !$omp end parallel do

    nsample1 = nsample


#ifdef MPI

    call MPI_AllReduce(mean,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
    mean = rdum

    call MPI_AllReduce(rms,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
    rms = rdum

    call MPI_AllReduce(vmin,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,ierr)
    vmin = rdum

    call MPI_AllReduce(vomin,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MIN,MPI_COMM_WORLD,ierr)
    vomin = rdum

    call MPI_AllReduce(vmax,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,ierr)
    vmax = rdum

    call MPI_AllReduce(vomax,rdum,1,&
         MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,ierr)
    vomax = rdum

    call MPI_AllReduce(nsample,nsample1,1,&
         MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,ierr)

#endif

    mean = mean/nsample1
    rms = dsqrt(rms/nsample1)

    if (mpirank .eq. 0) call stop_timer(tid)

  end subroutine testcpd


  !*******************************************************************
  ! calculate tr{|nm1><nm1| |nm2><nm2|}
  ! where  |nm1><nm1| is the projector on the space spanned
  ! by the single particle potentials of mode m of natpot no. 1
  ! and |nm2><nm2| the projector on the space spanned
  ! by the single particle potentials of mode m of natpot no. 2
  !*******************************************************************
  double precision function trsppproj(natpot1,natpot2,mode)
    implicit none
    type(natpot_t), intent(IN)       :: natpot1,natpot2
    integer, intent(IN)              :: mode

    integer                          :: dim1,dim2,i1,i2,z1,z2,N
    double precision                 :: r,rtmp
    double precision                 :: ddot !BLAS

    !call assert_compatible(natpot1,natpot2)

    dim1 = natpot1%potdim(mode)
    dim2 = natpot2%potdim(mode)

    N = natpot1%subdim(mode)

    rtmp = 0.d0
    if (dim2.ge. dim1) then
       !$omp parallel do private(i1,i2,z1,z2,r) reduction(+:rtmp)
       do i2=1,dim2
          z2= natpot2%zevec(mode,i2)
          do i1=1,dim1
             z1 = natpot1%zevec(mode,i1)
             r = ddot(N, natpot1%evec(z1),1, &
                  natpot2%evec(z2),1)
             rtmp = rtmp + r**2
          end do
       end do
       !$omp end parallel do
    else
       !$omp parallel do private(i1,i2,z1,z2,r) reduction(+:rtmp)
       do i1=1,dim1
          z1= natpot1%zevec(mode,i1)
          do i2=1,dim2
             z2 = natpot2%zevec(mode,i2)
             r = ddot(N, natpot2%evec(z2),1, &
                  natpot1%evec(z1),1)
             rtmp = rtmp + r**2
          end do
       end do
       !$omp end parallel do
    end if

    trsppproj = rtmp
    return

  end function trsppproj


  ! remove blanks from a string
  subroutine rmblnk(string)
    implicit none
    integer          :: i,j,length
    character(len=*) :: string

    length=len(string)

    j=0
    do i=1,length
       if (string(i:i) .ne. ' ') then
          j=j+1
          string(j:j)=string(i:i)
       endif
    enddo
    do i=j+1,length
       string(i:i)=' '
    enddo

  end subroutine rmblnk




  !*******************************************************************
  ! recursively build omega for a single point of the SPP
  ! dims:   like potdim, but modc is missing
  ! vec:    like evec, but only containing a single sampling point
  !         for each SPP
  ! totdim: equal to modcdim
  ! res:    contains omega for one sampling point on exit
  !*******************************************************************
  recursive subroutine comb_prod(ndims, dims, nvec, &
       vec, totdim, res, lwork, work)
    implicit none
    integer, intent(IN)              :: ndims, totdim, nvec, lwork
    integer, intent(IN)              :: dims(ndims)
    double precision, intent(IN)     :: vec(nvec)
    double precision, intent(OUT)    :: res(totdim)
    double precision, intent(INOUT)  :: work(lwork)

    integer                          :: ndiff, i,j,k,z
    integer                          :: ndims1, totdim1, nvec1, lwork1

    ! we end the recursion 1 step early
    if (ndims.eq.2) then
       k = 1
       do j=1,dims(2)
          do i=1,dims(1)
             res(k) =  vec(i)*vec(j+dims(1))
             k = k+1
          end do
       end do
       return
    end if

    !last recursion step - if called for one mode
    if (ndims .eq. 1) then
       res = vec
       return
    end if

    ndims1 = ndims -1
    nvec1 = nvec - dims(1)    !sum(dims(2:ndims))
    ndiff = nvec - nvec1
    totdim1 = totdim/dims(1)  !product(dims(2:ndims))
    z = totdim1 + 1
    lwork1 = lwork - totdim1

    call comb_prod(ndims1, dims(2:ndims), nvec1, &
         vec(ndiff+1:nvec), totdim1, work(1:totdim1), &
         lwork1, work(z:lwork))

    k = 1
    do j=1,totdim1
       do i=1,dims(1)
          res(k) = vec(i)*work(j)
          k = k+1
       end do
    end do


  end subroutine comb_prod

  !*******************************************************************
  !calculate size of workspace for comb_prod
  !*******************************************************************
  function wk_comb_prod(dims, nmode) result(lwork)
    implicit none
    integer, intent(IN) :: nmode
    integer, intent(IN) :: dims(nmode)
    integer             :: lwork, m

    lwork = 0
    do m=nmode, 2, -1
       lwork = lwork + product(dims(m:nmode))
    end do

  end function wk_comb_prod

!!$
!!$  !This is supposed to become a restart file dump
!!$  subroutine wrmcprst(natpot)
!!$    implicit none
!!$    type (natpot_t), intent(IN)       :: natpot
!!$
!!$    include "global.inc"
!!$    include "channels.inc"
!!$    character(len=c5)                 :: filename
!!$
!!$
!!$    filename=name(1:laenge)//'/restart'
!!$    open(irst,file=filename,form='unformatted',status='unknown',err=999)
!!$    !call dump_natpot_data(natpot,irst)
!!$    close(irst)
!!$
!!$    return
!!$
!!$999 continue
!!$    routine = "wrmcprst"
!!$    message = "Unable to open "//filename
!!$    call errormsg
!!$
!!$  end subroutine wrmcprst
!!$



end module pftools
