!  -*- f90 -*-

module rshdmr
  use mpidata
  use channels
  use global
  use logdat

  implicit none
  public


  double precision              :: v0_rshdmr
  double precision, allocatable :: v1_rshdmr(:)
  integer, allocatable          :: z_v1_rshdmr(:)
  integer                       :: v1_len
  logical                       :: lrshdmr
  logical                       :: lrshdmr_ready
  logical                       :: lrshdmr_initialized


end module rshdmr


  subroutine rshdmr_default()
    use rshdmr
    implicit none

    lrshdmr = .false.
    lrshdmr_ready = .false.
    lrshdmr_initialized = .false.

  end subroutine rshdmr_default




  subroutine init_rshdmr(nmode,subdim)
    use rshdmr
    implicit none

    integer, intent(IN)         :: nmode
    integer, intent(IN)         :: subdim(nmode)
    integer                     :: m,j


    v1_len  = 0
    do m=1,nmode
       v1_len = v1_len + subdim(m)
    end do


    allocate(v1_rshdmr(v1_len))
    allocate(z_v1_rshdmr(nmode+1))

    j = 1
    do m=1,nmode
       z_v1_rshdmr(m) = j
       j = j+subdim(m)
    end do
    z_v1_rshdmr(nmode+1) = v1_len + 1

    lrshdmr_ready = .false.
    lrshdmr_initialized = .true.

  end subroutine init_rshdmr


  subroutine free_rshdmr
    use rshdmr
    implicit none

    if (allocated(v1_rshdmr)) deallocate(v1_rshdmr)
    if (allocated(z_v1_rshdmr)) deallocate(z_v1_rshdmr)

    v1_len = 0
    v0_rshdmr = 0.d0
    lrshdmr_ready = .false.
    lrshdmr_initialized = .false.

  end subroutine free_rshdmr


  double precision function rshdmrbymidx(midx,nmode)
    use rshdmr
    implicit none
    integer, intent(IN)       :: nmode
    integer, intent(IN)       :: midx(nmode)

    integer                   :: z,m
    double precision          :: v

    if (.not. lrshdmr_ready) then
       rshdmrbymidx = 0.d0
       return
    end if

    v = v0_rshdmr

    do m = 1,nmode
       z = z_v1_rshdmr(m) + midx(m) - 1
       v = v + v1_rshdmr(z)
    end do

    rshdmrbymidx = v


  end function rshdmrbymidx




  subroutine mkrshdmr(potential, dvrindex,ndof,nmode,nsample,nspfdof, &
       spfdof,gdim,maxspfdof,subdim,lomppotential)
    use rshdmr
    use pftools, only: f2mdecoder,m2fdecoder
    use mpidata
#ifdef MPI
    use mpi
#endif
    implicit none
    integer, intent(IN)           :: ndof,nsample,nmode,maxspfdof
    integer, intent(IN)           :: dvrindex(ndof,nsample)
    integer, intent(IN)           :: gdim(ndof), nspfdof(nmode)
    integer, intent(IN)           :: spfdof(maxspfdof,nmode),subdim(nmode)
    logical, intent(IN)           :: lomppotential
    integer                       :: fpoint(ndof), mpoint(nmode)
    integer                       :: z,z1,m,m1,n,nsample1,ig,mgdim
    double precision              :: v,v1, ovlp(nmode,nmode), alpha(nmode)
    double precision,allocatable  :: vm(:),vtraj(:)
    double precision              :: potential
    integer                       :: ierr

#ifdef MPI
    double precision,allocatable  :: mpibuf(:),mpibuf2(:,:)
#endif
    external potential


    ! the job is:
    ! minimize var(V - V0 - sum_k a_k V1_k)
    ! where var is with respect to a Monte-Carlo
    ! trajectory of points q_n, and V0 is the mean of V(q_n).
    ! V1_k  = mean V1_kn(q_n) where V1_kn(q_n) is
    ! a 1D cut along mode k through point q_n


    if (.not. lrshdmr_initialized) then
       write(*,*) "rshdmr is not initialized"
       stop 1
    end if

    ! we start with v0, thereby
    ! saving the potential values along
    ! the trajectory
    if (mpirank.eq.0) then
       write(ilog,'(a)') ""
       write(ilog,'(a)') "CREATING RSHDMR"
       write(ilog,'(a)') ""
       write(ilog,'(a)',advance="no") "RSHDMR (v0)..."
       call myflush(ilog)
    end if

    allocate(vtraj(nsample))

    v = 0.d0
    !$omp parallel do private(v1) reduction(+:v) if(lomppotential)
    do n=1,nsample
       v1 = potential(dvrindex(:,n), ndof)
       v = v + v1
       vtraj(n) = v1
    end do
    !$omp end parallel do

#ifdef MPI

    ! for some reason, some MPI installations
    ! seem o have a problem with "allreduce"
    ! so we do a reduce and then a bcast

    call MPI_Reduce(v,v0_rshdmr,1,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)
    call MPI_Reduce(nsample,nsample1,1,MPI_INTEGER, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq.0) then
       v0_rshdmr = v0_rshdmr/nsample1
    end if

    call MPI_Bcast(v0_rshdmr,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#else
    v0_rshdmr = v/nsample
#endif

    if (mpirank.eq.0) then
       write(ilog,'(a)') " OK"
       call myflush(ilog)
    end if

    ! v0 is done. Now V1_k

    do m=1,nmode

       if (mpirank.eq.0) then
          write(ilog,'(a,i3,a)',advance="no") "RSHDMR (v1) mode ",m,"..."
          call myflush(ilog)
       end if

       mgdim = subdim(m)


       allocate(vm(mgdim))
       vm(:) = 0.d0

       !$omp parallel do private(fpoint,mpoint,ig,v) &
       !$omp reduction(+:vm) if(lomppotential)
       do n=1,nsample

          fpoint = dvrindex(:,n)
          call f2mdecoder(fpoint,mpoint,nspfdof,spfdof, &
               gdim,ndof,nmode,maxspfdof)

          do ig=1,mgdim
             mpoint(m) = ig
             call m2fdecoder(mpoint,fpoint,nmode,ndof, &
                  spfdof,nspfdof,gdim,maxspfdof)
             v = potential(fpoint,ndof)
             vm(ig) = vm(ig) + v
          end do

       end do
       !$omp end parallel do

#ifdef MPI

       if (mpirank .eq.0) then
          allocate(mpibuf(mgdim))
       end if

       call MPI_Reduce(vm,mpibuf,mgdim,MPI_DOUBLE_PRECISION, &
            MPI_SUM,0,MPI_COMM_WORLD,ierr)

       if (mpirank .eq.0) then
          vm = mpibuf/nsample1 !- v0_rshdmr
          deallocate(mpibuf)
       end if

       call MPI_Bcast(vm,mgdim,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#else
       vm = vm/nsample! - v0_rshdmr
#endif

       !subtract the mean along this tracectorsy

       v = 0.d0
       !$omp parallel do private(fpoint,mpoint,z) reduction(+:v)
       do n=1,nsample
          fpoint = dvrindex(:,n)
          call f2mdecoder(fpoint,mpoint,nspfdof,spfdof, &
               gdim,ndof,nmode,maxspfdof)
          z = mpoint(m)
          v = v + vm(z)
       end do
       !$omp end parallel do

#ifdef MPI
       call MPI_Reduce(v,v1,1,MPI_DOUBLE_PRECISION, &
            MPI_SUM,0,MPI_COMM_WORLD,ierr)
       if (mpirank .eq. 0) then
          v = v1/nsample1
       end if
       call MPI_Bcast(v,1,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#else
       v = v/nsample
#endif

       vm = vm - v
       z = z_v1_rshdmr(m)
       z1 = z + mgdim - 1
       v1_rshdmr(z:z1) = vm

       deallocate(vm)

       if (mpirank.eq.0) then
          write(ilog,'(a)') " OK"
          call myflush(ilog)
       end if

    end do !m

! the mean of the V1_k are now zero wrt the trajectory.
! we need now to minimize the var(V - V0 - sum_k a_k V1_k)
! that is we need the coefficients a_k

    if (mpirank.eq.0) then
       write(ilog,'(a)',advance="no") " Calculating (cross) variances..."
       call myflush(ilog)
    end if

    ovlp = 0.d0
    do m1=1,nmode
       do m = m1,nmode
          v = 0.d0
          !$omp parallel do reduction(+:v) private(fpoint,mpoint,z,z1)
          do n=1,nsample
             fpoint = dvrindex(:,n)
             call f2mdecoder(fpoint,mpoint,nspfdof,spfdof, &
                  gdim,ndof,nmode,maxspfdof)
             z = z_v1_rshdmr(m) + mpoint(m) - 1
             z1 = z_v1_rshdmr(m1) + mpoint(m1) - 1
             v = v + v1_rshdmr(z) *  v1_rshdmr(z1)
          end do
          !$omp end parallel do

          ovlp(m,m1) = v
          if (m.ne.m1) ovlp(m1,m) = v
       end do
    end do

    if (mpirank.eq.0) then
       write(ilog,'(a)') " OK."
       write(ilog,'(a)',advance="no") " Calculating overlaps..."
       call myflush(ilog)
    end if

    do m = 1,nmode

       v = 0.d0
       !$omp parallel do reduction(+:v) private(fpoint,mpoint,z)
       do n=1,nsample
          fpoint = dvrindex(:,n)
          call f2mdecoder(fpoint,mpoint,nspfdof,spfdof, &
               gdim,ndof,nmode,maxspfdof)
          z = z_v1_rshdmr(m) + mpoint(m) - 1
          v = v + v1_rshdmr(z) *  (vtraj(n)-v0_rshdmr)
       end do
       !$omp end parallel do

       alpha(m) = v

    end do

    if (mpirank.eq.0) then
       write(ilog,'(a)') " OK."
       call myflush(ilog)
    end if

#ifdef MPI

    if (mpirank .eq.0) then
       allocate(mpibuf2(nmode,nmode))
       allocate(mpibuf(nmode))
    end if

    call MPI_Reduce(ovlp,mpibuf2,nmode**2,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    call MPI_Reduce(alpha,mpibuf,nmode,MPI_DOUBLE_PRECISION, &
         MPI_SUM,0,MPI_COMM_WORLD,ierr)

    if (mpirank .eq.0) then
       ovlp = mpibuf2/nsample1
       alpha = mpibuf/nsample1
       deallocate(mpibuf2,mpibuf)
    end if

#else
    ovlp = ovlp/nsample
    alpha = alpha/nsample
#endif

    if (mpirank.eq.0) then

       write(ilog,'(a)',advance="no") " Solving for optimal RSHDMR coefficients..."
       call myflush(ilog)

       open(50,file="alpha_etc")
       write(50,'(a)') "Overlap matrix:"
       do m=1,nmode
          write(50,'(8f16.8)') (ovlp(m,m1),m1=1,nmode)
       end do
       write(50,'(a)') " "
       write(50,'(a)') "alpha (before):"
       do m=1,nmode
          write(50,'(f16.8)') alpha(m)
       end do


       call DPOSV( "L",nmode,1,ovlp,nmode,alpha,nmode,ierr)

       if (ierr.ne.0) then
          routine = "mkrshdmr"
          message = "Solving for optimal RSHDMR coefficients failed."
          call errormsg
       end if


       write(50,'(a)') " "
       write(50,'(a)') "alpha (after):"
       do m=1,nmode
          write(50,'(f16.8)') alpha(m)
       end do
       close(50)

       do m = 1,nmode
          z = z_v1_rshdmr(m)
          z1 = z_v1_rshdmr(m+1)-1
          v1_rshdmr(z:z1) =  v1_rshdmr(z:z1)*alpha(m)
       end do

       write(ilog,'(a)') " OK."

    end if

#ifdef MPI
    call  MPI_Bcast(v1_rshdmr,v1_len,MPI_DOUBLE_PRECISION, &
         0,MPI_COMM_WORLD,ierr)
#endif

    deallocate(vtraj)
    lrshdmr_ready = .true.

  end subroutine mkrshdmr


  subroutine laod_rshdmr(nmode,ierr)
     use rshdmr
     use mpidata
#ifdef MPI
    use mpi
#endif
    implicit none

    integer, intent(OUT)   :: ierr
    integer, intent(IN)    :: nmode

    character(len=c5)      :: fname


    ierr = 0
    call free_rshdmr()

    if (mpirank .eq.0) then
       fname = trim(name)//"/rshdmr"
       open(idump,file=fname,status="old", form="unformatted",iostat = ierr)
    end if


#ifdef MPI
    call  MPI_Bcast(ierr,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
#endif

    if (ierr.ne.0) return

    if (mpirank .eq.0) then
       read(idump) v1_len
    end if

#ifdef MPI
    call  MPI_Bcast(v1_len,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
#endif

    allocate(v1_rshdmr(v1_len))
    allocate(z_v1_rshdmr(nmode+1))

    if (mpirank .eq.0) then
       read(idump) v0_rshdmr
       read(idump) v1_rshdmr
       read(idump) z_v1_rshdmr
       close(idump)
    end if

#ifdef MPI
    call  MPI_Bcast(v0_rshdmr,1,MPI_DOUBLE_PRECISION, &
         0,MPI_COMM_WORLD,mpierr)
    call  MPI_Bcast(v1_rshdmr,v1_len,MPI_DOUBLE_PRECISION, &
         0,MPI_COMM_WORLD,mpierr)
    call  MPI_Bcast(z_v1_rshdmr,nmode+1,MPI_INTEGER, &
         0,MPI_COMM_WORLD,mpierr)
#endif



    lrshdmr_ready = .true.
    lrshdmr_initialized = .true.

    return



  end subroutine laod_rshdmr




 subroutine save_rshdmr(nmode)
    use rshdmr
    implicit none

    integer, intent(IN) :: nmode

    character(len=c5)                    :: fname
    integer                              :: m, mgdim,z,i
    character(len=8)                     :: buf


    if (mpirank .eq.0) then
       fname = trim(name)//"/rshdmr"
       open(idump,file=fname,status="unknown", form="unformatted",err=23)
       write(idump) v1_len
       write(idump) v0_rshdmr
       write(idump) v1_rshdmr
       write(idump) z_v1_rshdmr
       close(idump)

       fname = trim(name)//"/rshdmr_v0"
       open(idump,file=fname,status="unknown", form="formatted",err=23)
       write(idump,'(E20.12)') v0_rshdmr
       close(idump)

       do m=1,nmode
          mgdim = z_v1_rshdmr(m+1) - z_v1_rshdmr(m)

          write(buf,'(i8)') m
          buf = adjustl(buf)

          fname = trim(name)//"/rshdmr_v1_mode_"//trim(buf)

          open(idump,file=fname,status="unknown", form="formatted",err=23)
          do i=1,mgdim
             z = z_v1_rshdmr(m)+i-1
             write(idump,'(E20.12)') v1_rshdmr(z)
          end do
          close(idump)
       end do


    end if

    return

23  continue
    message = "Failed saving file "//trim(fname)
    call errormsg

  end subroutine save_rshdmr

