! -*- f90 -*-

module runfit_m
   use dvrtype_m
   use mcsampling
   use channels, only: ilog
   use tensortype
   use mcsamplingio, only: loaddvrindex
   use mpidata
   use mcpmpitools, only: mpi_set_chunksize
   use pftools, only: f2mdecoder

#ifdef MPI
    use mpi
#endif

#ifdef OPENMP
    use omp_lib
#endif

  implicit none

  ! the current single-point indices in
  ! dvr and mode basis and the corresponding
  ! single point energies
  integer, allocatable,target   :: dvrindex(:,:)
  integer, allocatable,target   :: modeindex(:,:)
  double precision, allocatable :: energies(:)


  ! currently loaded number of sampling points in all
  ! MPI ranks
  integer                       :: curr_ntotsample = 0



  integer                       :: task_loaded = 0 ! implicit save
  logical                       :: all_taskfiles_loaded = .false.

contains

  subroutine dispose_current_runtask()
    implicit none
    if (allocated(dvrindex)) deallocate(dvrindex)
    if (allocated(modeindex)) deallocate(modeindex)
    if (allocated(energies)) deallocate(energies)
    curr_ntotsample = 0
    task_loaded = 0
    all_taskfiles_loaded = .false.
  end subroutine dispose_current_runtask


  subroutine loadruntask(task,dvr,nsample,nfile)

    implicit none

    integer, intent(IN)           :: task
    type(dvr_t), intent(IN)       :: dvr
    integer, intent(OUT)          :: nsample
    integer, intent(IN), optional :: nfile

    integer                  :: allocstat,size,n,i
    integer                  :: nfile1, nfile2, nfiles
    character(len=13)        :: buf

    integer                  :: localstep(max_idxfile)
    integer                  :: localskip1st(max_idxfile)
    integer                  :: localnpoints(max_idxfile)

#ifdef MPI
    integer :: mpierr
#endif


    select case(task)
    case (mc_task_spp)
       buf = "densities"
    case (mc_task_coeff)
       buf = "coefficients"
    case (mc_task_test)
       buf = "test"
    end select

    ! nothing to load
    if (smethod(task,1) .eq. mc_complete) return


       ! load sampling points from file
    if (allocated(dvrindex)) deallocate(dvrindex)

    ! this calculates the sampling size in each rank
    ! and sets nsample. works also for non-mpi runs
    if (present(nfile)) then
       call mpi_set_chunksize(ntotsample(task,nfile),nsample)
       nfile1 = nfile
       nfile2 = nfile
       curr_ntotsample = ntotsample(task,nfile)
    else
       call mpi_set_chunksize(ntotsample(task,0),nsample)
       nfile1 = 1
       nfile2 = nidxfile(task)
       curr_ntotsample = ntotsample(task,0)
    end if
    nfiles = nfile2 - nfile1 + 1


#ifdef MPI
    call MPI_BCAST(curr_ntotsample,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
#endif

    allocate(dvrindex(dvr%ndof,nsample),stat=allocstat)
    if (allocstat.ne.0) then
       size = dvr%ndof*nsample
       call mcallocerr('dvrindex', 'integer', size,ilog)
    end if

    localskip1st = -1
    localstep = -1
    localnpoints = -1

    n=1
    do i=nfile1,nfile2
       if (smethod(task,i) .eq. mc_readfile) then
          localstep(n) = nstep(task,i)
          localskip1st(n) = nskipfirst(task,i)
       else
          localstep(n) = 0
          localskip1st(n) = 0
       end if
       localnpoints(n) = ntotsample(task,i)
       n = n+1
    end do


    call loaddvrindex(dvrindex,dvr%ndof,dvr%gdim, &
         nsample,localnpoints,localskip1st, &
         localstep,idxfile(task,nfile1:nfile2),nfiles )



    task_loaded = task
    all_taskfiles_loaded  = (nfiles .eq. nidxfile(task))

  end subroutine loadruntask


  subroutine dvrindex2modeindex(tensor, nsample)
    implicit none

    class (tensor_t), intent(IN) :: tensor
    integer, intent(IN)          :: nsample
    integer                      :: n!, idx(tensor%ndof)


    ! re-allocate if nevessary
    if (.not. allocated(modeindex)) then
       allocate(modeindex(nsample,tensor%nmode))
    else if (size(modeindex,1) .ne. size(dvrindex,2) &
         .or. size(modeindex,2) .ne. tensor%nmode) then
       deallocate(modeindex)
       allocate(modeindex(nsample,tensor%nmode))
    end if


    !$omp parallel do
    do n=1,nsample

       call f2mdecoder(dvrindex(:,n),modeindex(n,:),tensor%nspfdof, &
            tensor%spfdof,tensor%gdim,tensor%ndof,tensor%nmode, &
            tensor%maxspfdof)
    end do
    !$omp end parallel do

  end subroutine dvrindex2modeindex

  subroutine update_energies_dvrindex(potential,lthreadsafe)
    implicit none

    logical, intent(IN), optional :: lthreadsafe

    logical                       :: lomppotential
    integer          :: n, nsample, ndof
    double precision :: potential
    external potential

    if (present(lthreadsafe)) then
       lomppotential = lthreadsafe
    else
       lomppotential = .false.
    end if


    if (.not. allocated(dvrindex)) then
       print*, "Internal error in routine update_energies_dvrindex:"
       print*, "dvrindex is not allocated."
       stop 1
    end if

    ndof = size(dvrindex,1)
    nsample = size(dvrindex,2)



    if (allocated(energies)) deallocate(energies)
    allocate(energies(nsample))

    !$omp parallel do if(lomppotential)
    do n=1,nsample
       energies(n) = potential(dvrindex(:,n), ndof)
    end do
    !$omp end parallel do


  end subroutine update_energies_dvrindex

end module runfit_m







subroutine logparallel(ilog)
  use mpidata
  implicit none
  integer, intent(IN)  :: ilog
  logical              :: logisopen
#ifdef OPENMP
  integer              :: nomp
#endif

  if (mpirank.ne.0) return

  inquire(unit=ilog, opened=logisopen)
  if (.not. logisopen) return

  write(ilog,*)
  write(ilog,*)
  write(ilog,'(a)') '******* PARALLELIZATION ******'
  write(ilog,*)
  if (nmpi .gt. 1) then
     write(ilog, '(a,i6)') " Number of MPI instances:  ", nmpi
  else
     write(ilog, '(a)')    " No MPI parallelization."
  end if

#ifdef OPENMP
  !$omp parallel
  !$omp master
  nomp = omp_get_num_threads()
  !$omp end master
  !$omp end parallel
  write(ilog, '(a,i6)')    " Number of OpenMP threads: ", nomp
  write(ilog,*)
  call myflush(ilog)
  return
#endif

  write(ilog, '(a)')       " No OpenMP parallelization."
  write(ilog,*)
  write(ilog,*)
  call myflush(ilog)
  return


end subroutine logparallel





subroutine mcallocerr(entity, dtype, size, ilog)
  use logdat, only: stopnow
  use mpidata
  implicit none
  character(len=*), intent(IN) :: entity, dtype
  integer, intent(IN)          :: size, ilog

  logical                      :: logisopen

  write(*,'(a)') "ALLOCATION ERROR:"
  write(*,'(a)') " Entity:     "//trim(entity)
  write(*,'(a)') " Type:       "//trim(dtype)
  write(*,'(a,i8)') " Elements:   ",size

#ifdef MPI
  write(*,'(a,i8)') " MPI rank:   ", mpirank
#endif

  inquire(ilog, OPENED=logisopen)
  if (logisopen) then

     ! this will not happen in mpi rank > 0:

     write(ilog,'(a)') "ALLOCATION ERROR:"
     write(ilog,'(a)') " Entity:     "//trim(entity)
     write(ilog,'(a)') " Type:       "//trim(dtype)
     write(ilog,'(a,i8)') " Elements:   ",size
#ifdef MPI
     write(ilog,'(a,i8)') " MPI rank:   ", mpirank
#endif

  end if

  call stopnow()

end subroutine mcallocerr
