!  -*- f90 -*-


! Evaluate symmetries that are defined on the grid
! and skip the rest
subroutine GetSymDVRIdx(idx,gdim,idx1,ndof,isym)
  use symmops_m
  implicit none
  integer, intent(IN)   :: ndof,isym
  integer, intent(IN)   :: idx(ndof), gdim(ndof)
  integer, intent(OUT)  :: idx1(ndof)
  integer               :: f,f1

  idx1 = 0
  do f=1,ndof
     f1= symfmap(f,isym)
     if (symoptablef(f,isym) .eq. op_revert_grid) then
        idx1(f) = gdim(f1) - idx(f1) + 1
     elseif (symoptablef(f,isym) .eq. op_ident_grid) then
        idx1(f) = idx(f1)
     end if
     ! leave the rest undone
  end do

end subroutine GetSymDVRIdx


! Evaluate symmetries that are gefined using coordinates
! and skip the rest
subroutine GetSymDVRCoord(coord_in,coord_out,ndof, isym)
  use symmops_m
  implicit none
  integer, intent(IN)           :: ndof,isym
  double precision, intent(IN)  :: coord_in(ndof)
  double precision, intent(OUT) :: coord_out(ndof)

  integer                       :: f,f1

  coord_out = coord_in
  do f=1,ndof
     f1= symfmap(f,isym)
     if (symoptablef(f,isym) .eq. op_revert_coord) then
        coord_out(f) = -coord_in(f1) + coordshift(f,isym)
     elseif (symoptablef(f,isym) .eq. op_ident_coord)  then
        coord_out(f) =  coord_in(f1) + coordshift(f,isym)
     end if
     ! leave the rest undone

     ! if periodic, put it back in the boundaries
     if (lperiodic(f)) then
        ! note: lperiodic(f) == lperiodic(f1)
        ! as only periodics can swap
        if  (coord_out(f) .lt. periodicbox(1,f)) then
           do while (coord_out(f) .lt. periodicbox(1,f))
              coord_out(f) = coord_out(f) + periodicbox(2,f) - periodicbox(1,f)
           end do
        else if (coord_out(f) .ge. periodicbox(2,f)) then
           do while (coord_out(f) .ge. periodicbox(2,f))
              coord_out(f) = coord_out(f) - (periodicbox(2,f) - periodicbox(1,f))
           end do
        end if
     end if
  end do

end subroutine GetSymDVRCoord


! perform a symmetry operation on one mode m only.
! requires as input coord_in and idx_in.
! coord_in must contain valid data for all DOF that map into
! mode m under symmetry operation nsym with coordinate based
! operations (input as well as usersym)
! idx_in must contain valid data for all DOF that map INTO mode m
! under symmetry operation nsym for grid based operations.
! on exit coord_out contains all symmetric equivalent coordinates,
! idx_out contains the symmetric equivalent dvr points where applicable.
! otherwise -1. Both, coord_out and idx_out are modified only for DOF
! that are members of mode m. all ther DOF are unchanged.
subroutine GetSymPoint1m(coord_in, coord_out, idx_in, idx_out, nsym, m, tensor)
  use tensortype
  use symmops_m
  implicit none
  class(tensor_t), intent(IN)     :: tensor
  integer, intent(IN)             :: idx_in(tensor%ndof), nsym, m
  integer, intent(INOUT)          :: idx_out(tensor%ndof)
  double precision, intent(IN)    :: coord_in(tensor%ndof)
  double precision, intent(INOUT) :: coord_out(tensor%ndof)
  double precision                :: tmpcoord(tensor%ndof)
  integer                         :: tmpidx(tensor%ndof)

  integer                         :: f,z,n,k

  if (symoptablem(m,nsym) .eq. op_usersym) then
     call GetUserSym(coord_in,tmpcoord,tensor%ndof,symlabels(nsym))
  else
     call GetSymDVRCoord(coord_in,tmpcoord,tensor%ndof, nsym)
     call GetSymDVRIdx(idx_in,tensor%gdim,tmpidx,tensor%ndof,nsym)
  end if

  ! convert grids to coords and
  ! try to match coords to grids if possible
  do n=1,tensor%nspfdof(m)
     f = tensor%spfdof(n,m)

     if (lgridbased(f,nsym)) then
        z = tensor%zort(f) + tmpidx(f) -1
        coord_out(f) = tensor%ort(z)
        idx_out(f) =  tmpidx(f)
     else
        coord_out(f) =  tmpcoord(f)
        z = tensor%zort(f)
        idx_out(f) = -999
        do k=1,tensor%gdim(f)
           if (abs(tmpcoord(f) - tensor%ort(z+k-1)) .le. tol_pointsearch) then
              idx_out(f) = k
              exit ! n-loop
           end if
        end do
!!$        if  (idx_out(f) .eq. -999) then
!!$           print*, "code 999 for m,f ",m,f
!!$           do k=1,tensor%gdim(f)
!!$              print*, tmpcoord(f), tensor%ort(z+k-1)
!!$           end do
!!$           stop
!!$        end if
     end if

  end do

end subroutine GetSymPoint1m



! same a GetSymPoint1m, but for all modes
subroutine GetSymPoint(coord_in, coord_out, idx_in, idx_out, nsym, tensor)
  use tensortype
  use symmops_m
  implicit none
  class(tensor_t), intent(IN)    :: tensor
  integer, intent(IN)           :: idx_in(tensor%ndof), nsym
  integer, intent(OUT)          :: idx_out(tensor%ndof)
  double precision, intent(IN)  :: coord_in(tensor%ndof)
  double precision, intent(OUT) :: coord_out(tensor%ndof)

  integer                       :: m

  interface
     subroutine GetSymPoint1m(coord_in, coord_out, idx_in, idx_out, nsym, m, tensor)
       use tensortype
       use symmops_m
       class(tensor_t), intent(IN)     :: tensor
       integer, intent(IN)             :: nsym, m
       double precision, intent(IN)    :: coord_in(tensor%ndof)
       double precision, intent(INOUT) :: coord_out(tensor%ndof)
       integer, intent(IN)             :: idx_in(tensor%ndof)
       integer, intent(INOUT)          :: idx_out(tensor%ndof)
     end subroutine GetSymPoint1m
  end interface

  idx_out = -1
  coord_out = -1.d99

  do m=1,tensor%nmode
     call GetSymPoint1m(coord_in, coord_out, idx_in, idx_out, nsym, m, tensor)
  end do


end subroutine GetSymPoint



! return the closest grid point after symmetry operation 'nsym'
! on exit, if lexact is .true. idx1 contains the exact
! equivalent grid point, otherwise the closest point
! on the primitive grid.
subroutine FindNearestGridpoint(tensor, idx, idx1, nsym, lexact1)
  use tensortype
  use symmops_m
  use pftools
  implicit none
  class(tensor_t), intent(IN)    :: tensor
  integer, intent(IN)           :: idx(tensor%ndof), nsym
  integer, intent(OUT)          :: idx1(tensor%ndof)
  logical, intent(OUT)          :: lexact1
  integer                       :: tmpidx(tensor%ndof)
  double precision              :: qin(tensor%ndof)
  double precision              :: qout(tensor%ndof)
  integer                       :: f,z,i
  integer                       :: iclosest
  double precision              :: delta,closest

  interface
     subroutine GetSymPoint(coord_in, coord_out, idx_in, idx_out, nsym, tensor)
       use tensortype
       use symmops_m
       implicit none
       class(tensor_t), intent(IN)    :: tensor
       integer, intent(IN)           :: idx_in(tensor%ndof), nsym
       integer, intent(OUT)          :: idx_out(tensor%ndof)
       double precision, intent(IN)  :: coord_in(tensor%ndof)
       double precision, intent(OUT) :: coord_out(tensor%ndof)
     end subroutine GetSymPoint
  end interface

  lexact1 = .true.

  call dvrindex2coord(tensor,idx,qin)
  call GetSymPoint(qin, qout, idx, tmpidx, nsym, tensor)

  ! now find the closest DVR point to the equivalent coords
  ! (simple "manhatten norm")
  lexact1 = .true.
  do f=1,tensor%ndof
     z = tensor%zort(f)
     closest = tensor%ort(z)
     iclosest = 1
     delta = abs(closest - qout(f))
     do i=2,tensor%gdim(f)
        if (abs(qout(f) - tensor%ort(z+i-1)) .lt. delta ) then
           closest =  tensor%ort(z+i-1)
           iclosest = i
           delta = abs(closest - qout(f))
        end if
     end do
     idx1(f) = iclosest
     if (delta .gt. tol_pointsearch ) lexact1 = .false.
  end do

end subroutine FindNearestGridpoint




subroutine mksigma(natpot,densities,ndensities)
  use symmops_m
  use mcdensities_m, only: densities_t ! only the type definition
  use natpottype
  use cpdtype_m
  use pftools

  use mpidata
  use global
  use channels

#ifdef MPI
  use mpi
#endif
  implicit none

  integer, intent(IN)           :: ndensities
  type(natpot_t), intent(IN)    :: natpot
  type(densities_t), intent(IN) :: densities(ndensities)

  integer                       :: m,m1,md,md1,n,n1,nsym,z,ig,ig1
  integer                       :: potdim(natpot%nmode), modc
  double precision, allocatable :: symspp(:,:)
  character(len=c5)             :: fname
  logical                       :: lexists

#ifdef MPI
  integer                       :: ierr
#endif

  if (nsymmetries .le. 1) return

  potdim = natpot%potdim
  modc = natpot%modc


  if (allocated(sigma)) deallocate(sigma)
  if (allocated(zsigma)) deallocate(zsigma)

  !use the (mode-)mapping to calculate size of sigma
  allocate(zsigma(natpot%nmode,nsymmetries))
  zsigma = 0

  ! calculate the length of the sigma array
  ! and setup pointers
  sigmadim = 0
  do m=1,natpot%nmode
     if (m.ne. modc) then
        do n=1,nsymmetries
           zsigma(m,n) = sigmadim + 1
           sigmadim = sigmadim + potdim(m)**2
        end do
     end if
  end do

  allocate(sigma(sigmadim))

  if (mpirank.eq.0) then

     do md=1,ndensities
        m = densities(md)%dmode

        allocate(symspp(densities(md)%mgdim,potdim(m)))

        do nsym = 1,nsymmetries
           m1 = symmmap(m,nsym)
           do n=1,ndensities
              if (densities(n)%dmode .eq. m1) then
                 md1 = n
                 exit !loop
              end if
           end do

           do n=1,potdim(m1)
              z = zsymidxmap(m1,nsym)-1
              do ig=1,densities(md1)%mgdim     ! subdim is the same for m and m1
                 z = z+1
                 ig1 = symidxmap(z)

                 symspp(ig1,n) = densities(md1)%density(ig,n)
              end do
           end do

           do n1=1,potdim(m)
              do n=1,potdim(m)

                 z = zsigma(m,nsym) + (n-1)*potdim(m) + n1 -1
                 sigma(z) = dot_product(densities(md)%density(:,n1),symspp(:,n))

                 if ( abs(sigma(z)-1.d0) .lt. 1.d-11) sigma(z) = 1.d0
                 if ( abs(sigma(z)+1.d0) .lt. 1.d-11) sigma(z) = -1.d0
                 if ( abs(sigma(z)) .lt. 1.d-11)      sigma(z) = 0.d0
              end do
           end do

        end do ! nsym

        deallocate(symspp)

     end do

  end if

#ifdef MPI
  call MPI_Bcast(sigma,sigmadim,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
#endif

  if (mpirank.eq.0) then

     fname = trim(name)//"/sigma"
     inquire(file=trim(fname), EXIST=lexists)
     if (lexists.and..not.lovwr) then
        routine = "mksigma, writing sigma matrices:"
        message = "File exists, cannot overwrite. Use -w option: "//trim(fname)
        call errormsg
     end if


     open(idump,file=trim(name)//"/sigma")
     do m=1,natpot%nmode
        if (m.ne.modc) then
           do nsym = 1,nsymmetries

              m1 = symmmap(m,nsym)
              write(idump,'(a,i4,a,i4)') trim(symlabels(nsym)) &
                   //", mode:", m,' <--> ',m1
              write(idump,'(a)') "     n,    n1,     sigma(n,n1)"

              do n1=1,potdim(m)
                 do n=1,potdim(m)
                    z = zsigma(m,nsym) + (n-1)*potdim(m) + n1 -1
                    write(idump,"(2i6,f16.10)") n,n1, sigma(z)
                 end do
              end do
              write(idump,*) "----------------------------------------"

           end do
        end if
     end do
     close(idump)
  end if

end subroutine mksigma




subroutine symmetrizenatpot(natpot, coeff_ext, coeff_extdim)
  use natpottype
  use symmops_m
  use pftools
  use mpidata
  use mtlib, only: qtxxdd
#ifdef MPI
  use mpi
#endif

  implicit none

  type(natpot_t), intent(INOUT)    :: natpot
  integer, intent(IN)              :: coeff_extdim
  double precision, intent(IN)  :: coeff_ext(coeff_extdim)

  double precision, allocatable, target  :: work1(:)
  double precision, allocatable, target  :: work2(:)
  double precision, allocatable          :: coeff(:)
  double precision, pointer              :: pcoeff(:) => null()
  double precision, pointer              :: pwork(:)  => null()

  integer                                :: corig(natpot%nmode)
  integer                                :: cnew(natpot%nmode)
  integer                                :: cgdim(natpot%nmode)

  integer                                :: nsym,m,z, m1, ic, ic1
  integer                                :: dim1,dim2,dim3,dim4
  integer                                :: potvdim(natpot%nmode)
  integer                                :: potdim(natpot%nmode)
  integer                                :: potndim(natpot%nmode)
  integer                                :: dimbef, dimaft

#ifdef MPI
  integer :: ierr
#endif

  ! nothing to do if no symmetry operations are specified
  if (nsymmetries .le. 1) return

  if (mpirank.eq.0) then

     allocate(work1(coeff_extdim))
     allocate(work2(coeff_extdim))
     allocate(coeff(coeff_extdim))


     ! local pointers into coeff_ext
     do m=1,natpot%nmode
        potdim(m) = natpot%potdim(m)
     end do

     potvdim(1) = 1
     do m=2,natpot%nmode
        if (m-1.ne.natpot%modc) then
           potvdim(m) =potvdim(m-1)*potdim(m-1)
        else
           potvdim(m) =potvdim(m-1)*pointmaps(m-1)%npoint
        end if
     end do

     potndim(natpot%nmode)=1
     do m=natpot%nmode-1,1,-1
        if (m+1.ne.natpot%modc) then
           potndim(m)= potndim(m+1)*natpot%potdim(m+1)
        else
           potndim(m)= potndim(m+1)*pointmaps(m+1)%npoint
        end if
     end do

     do m=1,natpot%nmode
        if (m .eq. natpot%modc) then
           cgdim(m) = pointmaps(m)%npoint
        else
           cgdim(m) = natpot%potdim(m)
        end if
     end do

     dimbef = potvdim(natpot%modc)
     dimaft = potndim(natpot%modc)

     coeff =  coeff_ext

     do nsym = 2, nsymmetries

        ! copy the original coeff
        work1 = coeff_ext
        work2 = 0.d0
        pcoeff => work1
        pwork => work2

        ! do the symmetry operations
        ! mode wise, but without mode swapping yet
        do m = 1, natpot%nmode


           if (m.ne.natpot%modc) then

              z = zsigma(m,nsym)
              dim1 = potvdim(m)
              dim2 = potdim(m)
              dim3 = potndim(m)

              call qtxxdd(sigma(z),pcoeff,pwork,dim1,dim2,dim3)

           else
              dim1 = dimbef
              dim2 = pointmaps(m)%npoint
              dim3 = dimaft

              z = zsymidxmap(m,nsym)
              call symmodcmap(pcoeff, pwork,symidxmap(z),dim1,dim2,dim3 )
!
           end if ! modc

           ! swap pointers

           if (modulo(m,2) .eq. 1) then
              pcoeff => work2
              pwork => work1
           else
              pcoeff => work1
              pwork => work2
           end if

        end do


        ! we are done with the mode-internal
        ! symmetry operations
        ! but we may need to exchange modes

        !$omp parallel do private(corig,m,m1,cnew,ic1)
        do ic=1,coeff_extdim
           call cindex2sindex(ic,corig,cgdim,natpot%nmode)
           do m=1,natpot%nmode
              m1 = symmmap(m,nsym)
              cnew(m) = corig(m1)
           end do
           call sindex2cindex(ic1,cnew,cgdim,natpot%nmode)
           pwork(ic1) = pcoeff(ic)
        end do
        !$omp end parallel do

        coeff = coeff + pwork

     end do




     if (natpot%modc .gt. 0) then
        dim1 = natpot%dimbef
        dim2 = natpot%dimmodc
        dim3 = pointmaps(natpot%modc)%npoint
        dim4 = natpot%dimaft

        call coeff_ext2coeff(natpot%coeff,coeff,dim1,dim2,dim3,dim4)
        natpot%coeff = natpot%coeff/nsymmetries
     else

        natpot%coeff = coeff/nsymmetries

     end if

     deallocate(work1, work2, coeff)

  end if ! rank 0



#ifdef MPI
  call  MPI_Bcast(natpot%coeff,natpot%cdim,MPI_DOUBLE_PRECISION, &
       0,MPI_COMM_WORLD,ierr)
#endif


end subroutine symmetrizenatpot


subroutine coeff_ext2coeff(coeff,coeff_ext,vdim,gdim,gdim_ext,ndim)
  implicit none
  integer, intent(IN)            :: vdim,gdim,gdim_ext,ndim
  double precision, intent(IN)   :: coeff_ext(vdim,gdim_ext,ndim)
  double precision, intent(OUT)  :: coeff(vdim,gdim,ndim)
  integer                        :: i,j,k

  !!$omp parallel do private(i,j)
  do k=1,ndim
     do j=1,min(gdim,gdim_ext)
        do i=1,vdim
           coeff(i,j,k) = coeff_ext(i,j,k)
        end do
     end do
  end do
  !!$omp end parallel do

end subroutine coeff_ext2coeff




! symmodcmap: symmetry operation on contracted mode
subroutine symmodcmap(coeffin, coeffout, idmap, dim1,dim2,dim3)
  implicit none
  integer, intent(IN)           :: dim1,dim2,dim3
  double precision, intent(IN)  :: coeffin(dim1,dim2,dim3)
  double precision, intent(OUT) :: coeffout(dim1,dim2,dim3)
  integer, intent(IN)           :: idmap(dim2)

  integer                       :: i1,i2,i3, i

  do i3=1,dim3
     !$omp parallel do private(i,i1)
     do i2=1,dim2
        i = idmap(i2)
        do i1=1,dim1
           coeffout(i1,i,i3) =  coeffin(i1,i2,i3)
        end do
     end do
     !$omp end parallel do
  end do

end subroutine symmodcmap


subroutine testnpotsymmetry(natpot, ntest)
  use natpottype
  use pftools
  use pesselect_s
  use mcpfitparams
  use symmops_m

  use channels
  use global
  use mpidata

  implicit none

  type (natpot_t), intent(IN)       :: natpot
  integer, intent(IN)               :: ntest

  integer                           :: idx(natpot%ndof), idx1(natpot%ndof)
  double precision                  :: vpot, vpotE
  double precision                  :: vorig
  double precision                  :: ran(natpot%ndof)
  logical                           :: lhit
  integer                           :: n,nsym,f, nmiss, nongrid

  interface
     subroutine FindNearestGridpoint(tensor, idx, idx1, nsym, lexact1)
       use tensortype
       use symmops_m
       use pftools
       implicit none
       class(tensor_t), intent(IN)    :: tensor
       integer, intent(IN)           :: idx(tensor%ndof), nsym
       integer, intent(OUT)          :: idx1(tensor%ndof)
       logical, intent(OUT)          :: lexact1
     end subroutine FindNearestGridpoint
  end interface

  ! nothing to do
  if (.not. lsymmetries) return
  if (nsymmetries .le. 1) return
  ! we do this only in rank 0
  if (mpirank.ne.0) return


  write(isymlog,'(a)') &
       " ======= Testing Symmetry of natpot ======="
  write(isymlog,'(a)')
  write(isymlog,'(a)') &
       "  * 50 DVR points are chosen at random."
  write(isymlog,'(a)') &
       "  * All symmetry operations are applied to each point"
  write(isymlog,'(a)') &
       "    leading to another DVR point on which the routine"
  write(isymlog,'(a)') &
       "    and the natpot are probed."
  write(isymlog,'(a)') &
       "  * If a symmetry operation does not lead to a DVR point"
  write(isymlog,'(a)') &
       "    but to an extended grid point, then"
  write(isymlog,'(a)') &
       "    the closest original DVR point is used instead."
  write(isymlog,'(a)') &
       "    This can lead to varying values in the V(routine)"
  write(isymlog,'(a)') &
       "    and V(natpot) columns."
  write(isymlog,'(a)') &
       "    The flag 'on-grid' indicates if the symmetry operation"
  write(isymlog,'(a)') &
       "    stayed on the original DVR grid ('on-grid' = 'T')"
  write(isymlog,'(a)') &
       "    or if not ('on-grid' = 'F')"
  write(isymlog,'(a)') &
       "  * All V(natpot) for 'on-grid' = 'T' should be the same "
  write(isymlog,'(a)') &
       "    within each test set. Likewise all V(routine)."
  write(isymlog,'(a)')
  write(isymlog,'(a)')


  nmiss = 0
  nongrid = 0
  do n=1,ntest

     call random_number(ran)
     do f=1,natpot%ndof
        idx(f) = int(ran(f)*natpot%gdim(f)) + 1
     end do

     write(isymlog,'(a)') ""
     write(isymlog,'(a,i4)')   "Test", n
     write(isymlog,'(a)') &
          "Sym                   V(routine)      V(natpot)  on-grid   DVR-Point"
     write(isymlog,'(a)') &
          "----------------------------------------------------------------------"
     do nsym=1,nsymmetries
        call FindNearestGridpoint(natpot, idx, idx1, nsym, lhit)
        vorig = potential(idx1, natpot%ndof)

        vpot  = pfbyindex(natpot,idx1)
        if (nsym .eq. 1) then
           vpotE = Vpot
        end if

        write(isymlog, '(a, 2f16.8,L8,100i4)') symlabels(nsym), vorig,  &
             vpot, lhit, (idx1(f), f=1,natpot%ndof)

        if (lhit .and. nsym .ne. 1) then
           nongrid = nongrid + 1
           if (abs(vpotE - vpot) .gt. 1.d-6) then
              nmiss = nmiss + 1
           end if
        end if

     end do

  end do

  if (nmiss .gt. 0) then
     message = "Natpot does not observe all implemented symmetries" &
          //" (some degenerate SPP missing?). See sym.log for details."
     routine = "testnpotsymmetry"

     write(isymlog,'(a)')
     write(isymlog,'(a)') "WARNING in subroutine 'testnpotsymmetry':"
     write(isymlog,'(a)') message
     call myflush(isymlog)

     call warnmsg
     lwarn = .false.

  else

     if (nongrid .gt. 0) then
        write(isymlog,'(a)') " "
        write(isymlog,'(a)') "Natpot symmetry test OK."
     else
        message = "Natpot symmetry test undecided. " &
             // "No symmtry stayed on the original " &
             // "grid (testing only there). " &
             // "Please check if the results are plausible."

        write(isymlog,'(a)') " "
        write(isymlog,'(a)') "WARNING in subroutine 'testnpotsymmetry':"
        write(isymlog,'(a)') message
        call myflush(isymlog)

        routine = "testnpotsymmetry"
        call warnmsg
        lwarn = .false.

     end if
  end if


  write(isymlog,'(a)') ""
  write(isymlog,'(a)') " === END testing symmetry ==="
  write(isymlog,'(a)') ""
  close(isymlog)

end subroutine testnpotsymmetry


subroutine testcpdsymmetry(cpd, ntest)
  use cpdtype_m
  use pftools
  use pesselect_s
  use mcpfitparams
  use symmops_m

  use channels
  use global
  use mpidata

  implicit none

  type (cpd_t), intent(IN)       :: cpd
  integer, intent(IN)               :: ntest

  integer                           :: idx(cpd%ndof), idx1(cpd%ndof)
  double precision                  :: vpot, vpotE
  double precision                  :: vorig
  double precision                  :: ran(cpd%ndof)
  logical                           :: lhit
  integer                           :: n,nsym,f, nmiss, nongrid

  interface
     subroutine FindNearestGridpoint(tensor, idx, idx1, nsym, lexact1)
       use tensortype
       use symmops_m
       use pftools
       implicit none
       class(tensor_t), intent(IN)    :: tensor
       integer, intent(IN)           :: idx(tensor%ndof), nsym
       integer, intent(OUT)          :: idx1(tensor%ndof)
       logical, intent(OUT)          :: lexact1
     end subroutine FindNearestGridpoint
  end interface

  ! nothing to do
  if (.not. lsymmetries) return
  if (nsymmetries .le. 1) return
  ! we do this only in rank 0
  if (mpirank.ne.0) return


  write(isymlog,'(a)') &
       " ======= Testing Symmetry of cpd ======="
  write(isymlog,'(a)')
  write(isymlog,'(a)') &
       "  * 50 DVR points are chosen at random."
  write(isymlog,'(a)') &
       "  * All symmetry operations are applied to each point"
  write(isymlog,'(a)') &
       "    leading to another DVR point on which the routine"
  write(isymlog,'(a)') &
       "    and the cpd are probed."
  write(isymlog,'(a)') &
       "  * If a symmetry operation does not lead to a DVR point"
  write(isymlog,'(a)') &
       "    but to an extended grid point, then"
  write(isymlog,'(a)') &
       "    the closest original DVR point is used instead."
  write(isymlog,'(a)') &
       "    This can lead to varying values in the V(routine)"
  write(isymlog,'(a)') &
       "    and V(cpd) columns."
  write(isymlog,'(a)') &
       "    The flag 'on-grid' indicates if the symmetry operation"
  write(isymlog,'(a)') &
       "    stayed on the original DVR grid ('on-grid' = 'T')"
  write(isymlog,'(a)') &
       "    or if not ('on-grid' = 'F')"
  write(isymlog,'(a)') &
       "  * All V(cpd) for 'on-grid' = 'T' should be the same "
  write(isymlog,'(a)') &
       "    within each test set. Likewise all V(routine)."
  write(isymlog,'(a)')
  write(isymlog,'(a)')


  nmiss = 0
  nongrid = 0
  do n=1,ntest

     call random_number(ran)
     do f=1,cpd%ndof
        idx(f) = int(ran(f)*cpd%gdim(f)) + 1
     end do

     write(isymlog,'(a)') ""
     write(isymlog,'(a,i4)')   "Test", n
     write(isymlog,'(a)') &
          "Sym                   V(routine)      V(cpd)     on-grid   DVR-Point"
     write(isymlog,'(a)') &
          "----------------------------------------------------------------------"
     do nsym=1,nsymmetries
        call FindNearestGridpoint(cpd, idx, idx1, nsym, lhit)
        vorig = potential(idx1, cpd%ndof)

        vpot  = cpdbyindex(cpd,idx1)
        if (nsym .eq. 1) then
           vpotE = Vpot
        end if

        write(isymlog, '(a, 2f16.8,L8,100i4)') symlabels(nsym), vorig,  &
             vpot, lhit, (idx1(f), f=1,cpd%ndof)

        if (lhit .and. nsym .ne. 1) then
           nongrid = nongrid + 1
           if (abs(vpotE - vpot) .gt. 1.d-6) then
              nmiss = nmiss + 1
           end if
        end if

     end do

  end do

  if (nmiss .gt. 0) then
     message = "Cpd does not observe all implemented symmetries. " &
          //"Do the symmetries form a group? See sym.log for details."
     routine = "testcpdsymmetry"

     write(isymlog,'(a)')
     write(isymlog,'(a)') "WARNING in subroutine 'testcpdsymmetry':"
     write(isymlog,'(a)') message
     call myflush(isymlog)

     call warnmsg
     lwarn = .false.

  else

     if (nongrid .gt. 0) then
        write(isymlog,'(a)') " "
        write(isymlog,'(a)') "CPD symmetry test OK."
     else
        message = "Cpd symmetry test undecided. " &
             // "No symmtry stayed on the original " &
             // "grid (testing only there). " &
             // "Please check if the results are plausible."

        write(isymlog,'(a)') " "
        write(isymlog,'(a)') "WARNING in subroutine 'testcpdsymmetry':"
        write(isymlog,'(a)') message
        call myflush(isymlog)

        routine = "testnpotsymmetry"
        call warnmsg
        lwarn = .false.

     end if
  end if


  write(isymlog,'(a)') ""
  write(isymlog,'(a)') " === END testing symmetry ==="
  write(isymlog,'(a)') ""
  close(isymlog)

end subroutine testcpdsymmetry

subroutine testextcpdsymmetry(cpd_origgrid, cpd_extgrid, ntest)
  use cpdtype_m
  use pesselect_m
  use mcpfitparams
  use symmops_m
  use pftools
  use channels
  use global
  use mpidata

  implicit none

  type (cpd_t), intent(IN)          :: cpd_origgrid, cpd_extgrid
  integer, intent(IN)               :: ntest

  integer                           :: midx(cpd_origgrid%nmode), midx1(cpd_origgrid%nmode)
  double precision                  :: vpot
  double precision                  :: vref
  double precision                  :: ran(cpd_origgrid%ndof)
  integer                           :: nsym,m,m1,z,n, ndiff
  logical                           :: ldiff

  ! nothing to do
  if (.not. lsymmetries) return
  if (nsymmetries .le. 1) return
  ! we do this only in rank 0
  if (mpirank.ne.0) return

  write(isymlog,'(a)') &
       " ======= Testing Symmetry of cpd on extended grid ======="
  write(isymlog,'(a)')
  write(isymlog,'(a)') &
       "  * 50 mode points are chosen at random."
  write(isymlog,'(a)') &
       "  * All symmetry operations are applied to each point"
  write(isymlog,'(a)') &
       "    leading to another mode-point on which"
  write(isymlog,'(a)') &
       "    cpd on the extended grid is probed."
  write(isymlog,'(a)') &
       "  * All V(cpd_ext) values should be the same within each test set."
   write(isymlog,'(a)')
  write(isymlog,'(a)')


  ndiff = 0

  do n=1,ntest

     call random_number(ran)
     do m=1,cpd_origgrid%nmode
        midx(m) = int(ran(m)*cpd_origgrid%subdim(m)) + 1
     end do

     write(isymlog,'(a)') ""
     write(isymlog,'(a,i4)')   "Test", n
     write(isymlog,'(a)') &
          "Sym                   V(cpd_ext)     Mode-Point"
     write(isymlog,'(a)') &
          "----------------------------------------------------------------------"
     ldiff = .false.
     do nsym=1,nsymmetries

        do m=1,cpd_origgrid%nmode
           m1 = symmmap(m,nsym)
           z = zsymidxmap(m1,nsym)+midx(m)-1
           midx1(m1) = symidxmap(z)
        end do

        vpot  = cpdbyindex(cpd_extgrid, midx1)

        if (nsym .eq. 1) then
           vref = vpot
        else
           if (abs(vpot-vref) .gt. 1.d-6) then
              ldiff = .true.
           end if
        end if

        write(isymlog, '(a, f16.8,100i8)') symlabels(nsym),  &
             vpot, (midx1(m), m=1,cpd_origgrid%nmode)

     end do !nsym

     if (ldiff) ndiff = ndiff + 1

  end do !ntest

  if (ndiff .gt. 0) then

    message = "CPD on extended grid does not observe all implemented symmetries. " &
          //"See sym.log for details."
     routine = "testextcpdsymmetry"

     write(isymlog,'(a)')
     write(isymlog,'(a)') "WARNING in subroutine 'testextcpdsymmetry':"
     write(isymlog,'(a)') message
     call myflush(isymlog)

     call warnmsg
     lwarn = .false.
  end if

  write(isymlog,'(a)') ""
  write(isymlog,'(a)') " === END testing symmetry extended grid ==="
  write(isymlog,'(a)') ""



end subroutine testextcpdsymmetry



! internal symmetrization.
! Only uses the first rank/nsymmetries elements
! and discards the rest.
subroutine symmetrizecpd(cpd)
  use cpdtype_m
  use symmops_m
  implicit none

  type(cpd_t), intent(INOUT) :: cpd
  integer                    :: nsym, r, r1, k, m, m1, ig, z
  integer                    :: rpersym

  if (.not. lsymmetries) return

  rpersym = cpd%rank/nsymmetries

  do nsym=2,nsymmetries
     do r1 = 1, rpersym
        r = rpersym*(nsym-1) + r1
        do m=1,cpd%nmode
           m1 = symmmap(m,nsym)
           do ig=1,cpd%subdim(m)
              z = zsymidxmap(m1,nsym)+ig-1
              k = symidxmap(z)

              cpd%spp(m)%htm(ig,r) =  cpd%spp(m1)%htm(k,r1)
           end do
        end do
     end do
  end do


end subroutine symmetrizecpd
