!  -*- f90 -*-


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! symmops
! contains structures for symmetrizing modes in coordinate
! sbace and coefficients in configuration space
!
! Throughout this module, when doing a symmetry operation
!
! X := R(Y) e.g.  X = -Y + B
!
! f is the result DOF, labeling X, f1 is the source DOF, labeling Y
! m is the result mode with X, m1 is the source mode with Y
!
! TODO: Auto-complete a set of symmetry operations
! if a generating set is given
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module mcpeinsym

  implicit none

  integer                        :: nsymops
  character(len=16), allocatable :: symtable(:,:)
  double precision, allocatable  :: symrshift(:,:)
  logical, allocatable           :: lsymarithmetic(:,:)
  character(len=16), allocatable :: qperiodic(:) ! list of periodic DOF
  integer                        :: nperiodic    ! No. of poeriodif DOF

end module mcpeinsym


module symmops_m

  implicit none

  logical                          :: lsymmetries
  logical                          :: lautoextgrid
  logical                          :: lcomplete


  ! types of symmetry operations
  integer, parameter               :: op_err = -1
  integer, parameter               :: op_ident_grid = 1
  integer, parameter               :: op_revert_grid = 2
  integer, parameter               :: op_ident_coord = 3
  integer, parameter               :: op_revert_coord = 4
  integer, parameter               :: op_usersym = 5
  ! mode operator that can be any combination of
  ! op_ident_grid, op_revert_grid,
  ! op_ident_coord, or op_revert_coord for the mode grids
  ! but must not contain op_usersym for any DOF grid
  integer, parameter               :: op_m_inp = 6

  ! tolerance for searching for extended grit points
  double precision, parameter      :: tol_pointsearch = 1.d-7

  ! Space for the sigma matrices
  double precision, allocatable    :: sigma(:)
  integer, allocatable             :: zsigma(:,:)
  integer                          :: sigmadim

  ! DOF and mode mapping
  integer, allocatable             :: symfmap(:,:) ! dof mapping
  integer, allocatable             :: symmmap(:,:) ! mode mapping


  ! operator tables
  integer, allocatable             :: symoptablef(:,:)! dof operations
  integer, allocatable             :: symoptablem(:,:)! mode operations

  ! mode-grid point mapping:
  ! The symmetry operations are in the end
  ! mappings of the form
  ! nsym: symmetry operation to perform
  ! m1 = symmmap(m,nsym)      : we swap modes m and m1
  ! z =  zsymidxmap(m,nsym)   : pointer to the mode point permutation table
  ! j = symidxmap(z+i-1)      : the i'th point of mode m is swapped with the
  !                             j'th point of mode m1 such that
  ! modepoint(m,i) <==> modepoint(m1,j) for some nsym and m
  !
  integer                          :: symidxmapdim
  integer, allocatable             :: zsymidxmap(:,:)
  integer, allocatable             :: symidxmap(:)

  ! flag if modes m need an extended grid
  logical, allocatable             :: lmextgrid(:)

  ! number of symmetries and teir labels
  integer                          :: nsymmetries
  character(len=16), allocatable   :: symlabels(:)

  ! periodic grids
  logical, allocatable             :: lperiodic(:)     ! is f periodic?
  double precision, allocatable    :: periodicbox(:,:) ! xi, xf, dx of interval

  ! coordinate transformations if not grid based
  double precision, allocatable    :: coordshift(:,:)  ! R for expressions as "f' = f+R"
  logical, allocatable             :: lcoordbased(:,:) ! is expr. like  "f' = f+R"?
  logical, allocatable             :: lgridbased(:,:)  ! is grid based operation?


  ! container for coordinate values and dvr points of the
  ! (possibly extended) grids
  type pointmap_t
     integer                       :: ndof, nlen, npoint
     double precision, allocatable :: points(:,:)
     integer, allocatable          :: indices(:,:)
  end type pointmap_t

  ! this is an array of containers of extended grids
  type (pointmap_t),allocatable,target  :: pointmaps(:)

  ! a symmetry operated on another must be the same as a
  ! third from the group. This table maps this:
  ! n3 = sym2map(n1,n2): R1 R2 V = R3 V
  ! where n1, n2, n3 are integers laeling the symmetries
  integer, allocatable              :: sym2map(:,:)  

  ! n2 = symEmap(n1), such that sym2map(n1,n2) = 1,
  ! i.E., symmetries n1 und n2 operated
  ! consecutively is the identity
  integer, allocatable              :: symEmap(:)

  ! we nees a complete mapping of mode permutations
  ! for mccpd to know which modes must be
  ! updated simultaneously
  integer, allocatable          :: permutedmodes(:,:)
  integer, allocatable          :: npermutedmodes(:)

end module symmops_m


module symmops_s
    use global
    use channels
    use mpidata
    use logdat

    use tensortype
    use dvrtype_m
    use natpottype
    use mcpeinsym
    use pftools
    use pesselect_m
    use mcpfitparams
    use symmops_m
    use pesselect_s, only: lidxbased

    use mpidata
#ifdef MPI
    use mpi
#endif

contains


  subroutine symdefault()

    implicit none

    lsymmetries = .false.
    lautoextgrid = .false.
    lcomplete = .false.
    nsymmetries = 0
    symidxmapdim = 0
    sigmadim = 0

  end subroutine symdefault


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! build_symops
  !
  ! Constract the symmtry mappings from the
  ! input data. This is done in 4 steps:
  !
  ! 1) Get the DOF interchanges and operations
  !    from input directly where possible
  ! 2) Construct the mode interchange mapping
  !    either from the DOF mapping or from user
  !    implemented symmetry operations.
  ! 3) Detect if symmetry operations stay on the
  !    primitive grid and extend the grid if needed
  ! 4) Calulate point mappings on (extended) mode
  !    grids. These mappings are used in the end.
  !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine build_symops(tensor, dvr)

    implicit none
    class(tensor_t), intent(IN)    :: tensor
    type(dvr_t), intent(IN)       :: dvr
    integer                       :: m,n
    integer                       :: nsym
    integer                       :: Efmap(tensor%ndof)
    character(len=c5)             :: fname
    character(len=32)             :: formatstr
    logical                       :: lexists

    lwarn = .false.

    call free_pointmaps()
    allocate(pointmaps(tensor%nmode))
    allocate(lmextgrid(tensor%nmode))
    lmextgrid = .false.

    ! building is done rank 0 only
    ! we just bcast the results so all other
    ! ranks branch to recieve the pointmaps
    ! already here and then exit.
    if (mpirank .ne. 0) then
       call bcast_symmetrymaps(tensor%nmode)
       return
    end if

    nsymmetries = nsymops

    ! allocate memory for the mappings
    call alloc_symmops(tensor)

    ! start symmetry logging
    fname = trim(name)//"/sym.log"
    call opensymlog(fname)

    ! some initial checking and
    ! mapping of the symmetry section
    ! to the order of the PBASIS section
    call chk_symops(tensor,Efmap)

    ! build the DOF mapping and operator table
    ! from the raw input of the SYMMETRY-SECTION
    ! This skips the usersym mapping
    ! which is done later
    call mk_dofmap(tensor,Efmap)

    ! build the MODE operator table from the
    call mk_modemap(tensor)

    ! we now have the mapping of DOFs under symmetry
    ! operations except for those which come from usersym
    ! in file usersym.F90. For those we do mode-mapping only

    ! Detect which modes map onto each other for usersym
    ! thereby make sure no mode-mixing occurs.
    call mk_usermappings(tensor)
    call chk_mapping(tensor,dvr)

    ! now we have the mode-mapping and the DOF mapping
    ! we check now if there are DOF labeled as periodic
    ! and calculate the box boundaries. Also make sure
    ! that periodic DOF are not swapped with non-periodic DOF
    call chk_periodic(tensor, dvr)


    ! if there are user-defined or coordinate based symmetry operations
    ! there might be the need to extend the grids
    ! this we check here and calculate the dimension of the
    ! extended grid. If extgriddim(m) is the same as
    ! tensor%subdim(m) then no extended grid is needed.
    ! This will add extra points to the original dvr points
    call detect_extgrids(tensor)
    ! save the extended grids (if any) to file
    call save_extgrids(tensor)

    ! we have now a complete mapping:
    ! DOF, modes and DVR points

    ! Now check if the user implementation stays on the original
    ! primitive grid and if not, check if the user wants to create
    ! an extended grid
    if (any(lmextgrid) .and. .not. lautoextgrid) then
       routine = "build_symops, checking for extended grids."
       do m=1,tensor%nmode
          if (lmextgrid(m)) then
             write(message, '(a,i4,a)') "Mode ", m, &
                  " does not stay on the primitive grid."
             call warnmsg
          end if
       end do
       message = "Some modes do not stay on the " &
            //"primitive grid and 'auto-extgrid' is not set."
       call errormsg
    end if

    ! From all the information we gathered before
    ! we now calulate the mode-to-mode point-wise mappings
    ! which are the mappings that are actually used
    ! for symmetrization.
    ! allocate space for the index mapping
    if (allocated(zsymidxmap)) deallocate(zsymidxmap)
    allocate(zsymidxmap(tensor%nmode,nsymmetries))
    symidxmapdim = 0
    do nsym=1,nsymmetries
       do m=1,tensor%nmode
          zsymidxmap(m,nsym) = symidxmapdim+1
          symidxmapdim = symidxmapdim + pointmaps(m)%npoint
       end do
    end do
    if (allocated(symidxmap)) deallocate(symidxmap)
    allocate(symidxmap(symidxmapdim))


    ! and build the index maps of mode-grids
    call build_indexmap(tensor)

    ! we are done. Now write the structures to file.
    do m=1,tensor%nmode
       write(fname,'(a,i4)') trim(name)//"/idxmap_mode_", m
       call rmblnk(fname)
       inquire(file=trim(fname), EXIST=lexists)
       if (lexists.and..not.lovwr) then
          routine = "build_symops, writing mode index mapping"
          message = "File exists, cannot overwrite. Use -w option: " &
               //trim(fname)
          call errormsg
       end if
       open(idump,file=trim(fname))
       write(idump,'(a,i4)') "# Equivalent grid points of mode ", m
       write(formatstr,'(a,i4,a)') "(",nsymmetries,"a8)"
       call rmblnk(formatstr)
       write(idump,'(a)',advance="no") "#"
       write(idump,formatstr) (adjustr(symlabels(nsym)(1:8)), &
            nsym=1,nsymmetries)

       do n=1,pointmaps(m)%npoint
          write(idump,'(x,8i8)') (symidxmap(zsymidxmap(m,nsym)+n-1), &
               nsym=1,nsymmetries)
       end do
       close(idump)



    end do

    ! Mapping R1 R2 = R3
    call buildsym2map(tensor%nmode)
    call find_permutingmodes(tensor%nmode)

    ! log what we have before checking
    call logsymmops(tensor)

    ! final check and broadcast
    call chk_sym_mapping(tensor)
    call bcast_symmetrymaps(tensor%nmode)

    ! Free the space used
    ! for input tables
    call clean_mcpeinsym()

  end subroutine build_symops




  subroutine buildsym2map(nmode)
    implicit none
    integer, intent(IN)  :: nmode
    integer              :: n,n1,n2,n3
    integer              :: m,m1,m2,m3
    integer              :: ig,ig1,ig2,ig3
    logical              :: lmatch
    integer              :: nmatch,z, m3a(nmode),m3b(nmode)
    integer, allocatable :: nresultsym(:)

    if (allocated(sym2map)) deallocate(sym2map)
    if (allocated(symEmap)) deallocate(symEmap)

    allocate(sym2map(nsymmetries, nsymmetries))
    allocate(symEmap(nsymmetries))

    allocate(nresultsym(nsymmetries))
    nresultsym(:) = 0

    lcomplete = .true.

    write(isymlog, '(/a/)') "Completeness check:"

    sym2map = 0

    do n2=1,nsymmetries
       do n1=1,nsymmetries

          do n=1,nsymmetries

             nmatch = 0

             do m=1,nmode
                m1 = symmmap(m,n1)
                m2 = symmmap(m1,n2)
                m3a(m) =  m2

                m3 = symmmap(m,n)
                m3b(m) = m3

             end do

             if ( all( m3a(:).eq. m3b(:) ) ) then

                ! we have a candidate, now check grids
                lmatch = .true.
                nmatch = 0

                do m=1,nmode

                   if (.not. lmatch) cycle

                   m1 = symmmap(m,n1)
                   m2 = symmmap(m1,n2)
                   m3 = symmmap(m,n)

                   try_grid: do ig = 1,pointmaps(m)%npoint
                      if (.not. lmatch) cycle

                      z = zsymidxmap(m1,n1)+ig-1
                      ig1 = symidxmap(z)
                      z = zsymidxmap(m2,n2)+ig1-1
                      ig2 = symidxmap(z)

                      z = zsymidxmap(m3,n)+ig-1
                      ig3 = symidxmap(z)

                      if (ig3 .ne. ig2) lmatch = .false.

                   end do try_grid

                   if (lmatch) nmatch = nmatch + 1

                end do

                if (nmatch .eq. nmode) then
                   sym2map(n1,n2) = n
                   exit ! n-loop
                end if

             end if
          end do ! n-loop


       end do
    end do

    lcomplete = .true.

    do n2=1,nsymmetries
       do n1=1,nsymmetries
          n3 = sym2map(n1,n2)
          if (n3 .ne. 0) then
             write(isymlog, '(a)') symlabels(n1) // " X " // symlabels(n2) &
                  //" = " // symlabels(n3)
             nresultsym(n3) = nresultsym(n3) + 1
          else
             write(isymlog, '(a)') symlabels(n1) // " X " // symlabels(n2) &
                  //" = " // "??? "
             lcomplete = .false.
          end if
       end do
    end do

    ! build symEmap
    do n1=1,nsymmetries
       do n2=1,nsymmetries
          n3 = sym2map(n1,n2)
          if (n3 .eq. 1) then
             symEmap(n1) = n2
             cycle
          end if
       end do
    end do

    lcomplete = lcomplete .and. all(nresultsym .eq. nsymmetries)
    deallocate(nresultsym)

    if (.not. lcomplete ) then
       write(isymlog, '(/a)') "The implemented symmetry operations do not form a group."
       message = "The implemented symmetry operations do not form a group. See sym.log."
       routine = "buildsym2map"
       call warnmsg
    else
       write(isymlog, '(/a)') "The set of symmetry operations is complete"
    end if


  end subroutine buildsym2map



  subroutine free_pointmaps()
    implicit none
    integer             :: m

    ! initiallize the point mapping structures
    ! for external grids. This is only a container
    ! array of uninitialized extended grids.
    ! the extended grids themselved are only allocated
    ! and initialized if needed - but the container
    ! needs to be present.
    if (allocated(pointmaps)) then
       do m=1,size(pointmaps)
          pointmaps(m)%nlen = 0
          pointmaps(m)%ndof = 0
          pointmaps(m)%npoint = 0
          if (allocated(pointmaps(m)%points)) then
             deallocate(pointmaps(m)%points)
          end if
          if (allocated(pointmaps(m)%indices)) then
             deallocate(pointmaps(m)%indices)
          end if
       end do
       deallocate(pointmaps)
    end if


  end subroutine free_pointmaps



  subroutine alloc_symmops(tensor)
    implicit none

    class(tensor_t), intent(IN)    :: tensor
    integer                       :: n

    if (allocated(symfmap)) deallocate(symfmap)
    allocate(symfmap(tensor%ndof,nsymmetries))
    symfmap = 0

    if (allocated(symoptablef)) deallocate(symoptablef)
    allocate(symoptablef(tensor%ndof,nsymmetries))
    symoptablef = op_err

    if (allocated(symoptablem)) deallocate(symoptablem)
    allocate(symoptablem(tensor%nmode,nsymmetries))
    symoptablem = op_err

    if (allocated(symlabels)) deallocate(symlabels)
    allocate(symlabels(nsymmetries))
    do n=1,nsymmetries
       symlabels(n) = symtable(0,n)
    end do

    if (allocated(symmmap)) deallocate(symmmap)
    allocate(symmmap(tensor%nmode,nsymmetries))
    symmmap = 0

    if (allocated(lperiodic)) deallocate(lperiodic)
    allocate(lperiodic(tensor%ndof))
    lperiodic = .false.

    if (allocated(periodicbox)) deallocate(periodicbox)
    allocate(periodicbox(3,tensor%ndof))
    periodicbox = 0.d0

    if (allocated(coordshift)) deallocate(coordshift)
    allocate(coordshift(tensor%ndof, nsymmetries))
    coordshift = 0.d0

    if (allocated(lcoordbased)) deallocate(lcoordbased)
    allocate(lcoordbased(tensor%ndof, nsymmetries))
    lcoordbased = .false.

    if (allocated(lgridbased)) deallocate(lgridbased)
    allocate(lgridbased(tensor%ndof, nsymmetries))
    lgridbased = .false.


  end subroutine alloc_symmops



  subroutine mk_dofmap(tensor, Efmap)
    implicit none

    class(tensor_t), intent(IN)    :: tensor
    integer, intent(IN)           :: Efmap(tensor%ndof)
    integer                       :: nsym, f,f1,f2,k,foptype
    character(len=16)             :: buf

    ! build the DOF mapping and operator table
    ! from the raw input of the SYMMETRY-SECTION
    do nsym=1,nsymmetries
       do f=1,tensor%ndof
          f1 = Efmap(f)
          coordshift(f1,nsym) = symrshift(f,nsym)
          lcoordbased(f1,nsym) = lsymarithmetic(f,nsym)


          foptype = op_err
          if (symtable(f,nsym) .eq. "usersym") then
             foptype = op_usersym
          else if (symtable(f,nsym)(1:1) .eq. "-") then
             buf = symtable(f,nsym)(2:)
             if (lsymarithmetic(f,nsym)) then
                foptype = op_revert_coord
             else
                foptype = op_revert_grid
             end if
          else
             buf = symtable(f,nsym)
             if (lsymarithmetic(f,nsym)) then
                foptype = op_ident_coord
             else
                foptype = op_ident_grid
             end if
          end if

          symoptablef(f1,nsym) = foptype

          if (foptype .ne. op_usersym ) then
             k = 0
             do f2=1, tensor%ndof
                if (tensor%modelabel(f2) .eq. buf) then
                   k=f2
                   exit !loop
                end if
             end do
             symfmap(f1,nsym) = k

          end if

       end do
    end do

    do nsym=1,nsymmetries
       do f=1,tensor%ndof
          lgridbased(f,nsym) = .not. lcoordbased(f,nsym) &
               .and. (symoptablef(f,nsym) .ne. op_usersym)
       end do
    end do

    if (any(symoptablef(:,:) .eq. op_err)) then
       routine = "build_symops, DOF mapping"
       message = "An error occured while building the " &
            //" symmetry operations table. Please send a bug report."
       call errormsg
    end if

  end subroutine mk_dofmap



  subroutine mk_modemap(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    integer                       :: nsym,n,m,f,foptype

    ! build a mode based operator table
    ! and check that the user implemented
    ! marked all DOF in a mode as 'usersym'
    do nsym=1,nsymmetries
       do m=1,tensor%nmode

          do n=1,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)

             foptype = symoptablef(f,nsym)

             if (symoptablem(m,nsym) .eq. op_err) then ! initialy it's set to error

                if ( foptype .ne. op_usersym) then
                   symoptablem(m,nsym) = op_m_inp
                else if (foptype .eq. op_usersym) then
                   symoptablem(m,nsym) = op_usersym
                end if
             else
                if (foptype .ne. op_usersym) then
                   if (symoptablem(m,nsym) .eq. op_usersym) then
                      routine = "build_symops, mode mapping"
                      message = "Symmety "//trim(symlabels(nsym)) &
                           //": flag for DOF " &
                           //trim(tensor%modelabel(f)) //" must be 'usersym'."
                      call errormsg
                   end if
                elseif (foptype .eq. op_usersym) then
                   if (symoptablem(m,nsym) .eq.  op_m_inp) then
                      routine = "build_symops, mode mapping"
                      message = "Symmety "//trim(symlabels(nsym)) &
                           //": flag for DOF " &
                           // trim(tensor%modelabel(f)) &
                           //" is 'usersym' but other DOFs in " &
                           // "the mode are not (but must be)."
                      call errormsg
                   end if
                end if
             end if
          end do
       end do
    end do

  end subroutine mk_modemap


  subroutine find_permutingmodes(nmode)
    implicit none
    integer, intent(IN) :: nmode

    integer             :: n,m,m1,n2
    logical             :: lfound

    if (allocated(permutedmodes)) deallocate(permutedmodes)
    allocate(permutedmodes(nmode,nmode))
    permutedmodes = 0
    if (allocated(npermutedmodes)) deallocate(npermutedmodes)
    allocate(npermutedmodes(nmode))

    ! identity
    npermutedmodes(:) = 1
    do m=1,nmode
       permutedmodes(m,1) = m
    end do


    do m=1,nmode

       ! any modes that do not come in the
       ! first symmetry (for instance m1->m2 m2->m3 m3->m1
       ! in cyclic permuations, like C3-symmetry) should come
       ! in the other symmetries (C3^2 must also be implemented)
       ! as the group must be complete...
       do n=1,nsymmetries
          m1 = symmmap(m,n)

          lfound = .false.
          do n2=1,npermutedmodes(m)
             if (m1 .eq. permutedmodes(m,n2)) then
                lfound = .true.
             end if
          end do
          if (.not. lfound) then
             npermutedmodes(m) = npermutedmodes(m) + 1
             permutedmodes(m,npermutedmodes(m)) = m1
          end if
       end do
    end do

  end subroutine find_permutingmodes


  subroutine mk_usermappings(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    integer                       :: nsym,m,m0,n,f,f1,m1

    ! Detect which modes map onto each other
    ! thereby make sure no mode-mixing occurs.
    lwarn = .false.
    do nsym=1,nsymmetries
       do m=1,tensor%nmode
          m0 = -1
          do n=1,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)
             if (symoptablef(f,nsym) .ne. op_usersym) then
                f1 = symfmap(f,nsym)
                m1 = tensor%dofspf(f1)
             else
                call detect_usymmmapping(nsym,f,tensor,m1)
             end if

             if (m0 .lt. 0) then
                m0=m1
             else if (m1.ne.m0) then
                routine = "build_symops, mode mapping"
                write(message, '(a,2i4)') "Mode mixing in symmetry operation " &
                     //trim(symlabels(nsym)) // ", modes: ", m0,m1
                call warnmsg
             end if

          end do
          symmmap(m,nsym) = m1
       end do
    end do
    if (lwarn) then
       message = "Symmetry operations must either stay within one mode or " &
            //"swap two complete modes but cannot mix them."
       call errormsg
    end if

  end subroutine mk_usermappings




  subroutine chk_mapping(tensor,dvr)

    implicit none

    class(tensor_t), intent(IN)    :: tensor
    type(dvr_t), intent(IN)       :: dvr
    integer                       :: f,f1,m,m1,nsym
    logical                       :: lwarn1

    ! Check if exchanged modes have same No. of SPP,
    ! ensure the contracted mode is not exchanged.
    ! withanother mode
    ! and that the modes have the same grids
    do m=1,tensor%nmode
       do nsym=1,nsymmetries
          m1 = symmmap(m,nsym)
          if (m1 .ne. m) then

             select type(tensor)
             type is (natpot_t)
                if (m .eq. tensor%modc .or. m1 .eq. tensor%modc) then
                   routine = "build_symops, mode mapping"
                   message = "Symmetry "//trim(symlabels(nsym)) &
                        //": Contacted mode "// &
                        "cannot be part of a mode-exchanging symmetry operation."
                   call errormsg
                end if
                if (tensor%potdim(m) .ne. tensor%potdim(m1)) then
                   routine = "build_symops, mode mapping"
                   write(message,'(a,i3,a,i3,a)') "Symmetry " &
                        //trim(symlabels(nsym))// &
                        ": modes ",m," and ",m1, &
                        " must have the same number of SPP."
                   call errormsg
                end if
             end select

             if (tensor%subdim(m) .ne. tensor%subdim(m1)) then
                routine = "build_symops, mode mapping"
                write(message,'(a,i3,a,i3,a)') "Symmetry " &
                     //trim(symlabels(nsym))// &
                     ": modes ",m," and ",m1, &
                     " must have the same primitive grids."
                call errormsg
             end if

          end if
       end do
    end do

    lwarn1 = lwarn
    do f=1,tensor%ndof
       do nsym=1,nsymmetries
          if (symoptablef(f,nsym) .ne. op_usersym) then
             f1 = symfmap(f,nsym)
             if (dvr%basis(f) .ne. dvr%basis(f1)) then
                routine = "build_symops, check DVR mapping"
                message = "Symmetry operation "//trim(symlabels(nsym)) // &
                     ": DOFs '"//trim(dvr%modelabel(f)) // "' and '" // &
                     trim(dvr%modelabel(f1))//"' do not have the same DVR type."
                call warnmsg
             end if
          end if
       end do
    end do
    lwarn = lwarn1


    do f=1,tensor%ndof
       if (dvr%basis(f) .eq. 13 ) then ! external DVR
          if (any(symfmap(f,2:nsymmetries) .ne. 0)) then
             routine = "build_symops, check DVR mapping"
             message = "Symmetry operations for external DVRs ("// &
                  trim(tensor%modelabel(f)) //") must be 'usersym'."
             call errormsg
          end if
       end if
    end do


  end subroutine chk_mapping




  subroutine chk_periodic(tensor, dvr)
    implicit none

    class(tensor_t), intent(IN)    :: tensor
    type(dvr_t), intent(IN)       :: dvr
    logical                       :: lfound
    integer                       :: f,f1,f2,z,nsym
    logical                       :: lwarn1
    double precision              :: xi,xf,dx

    do f = 1, nperiodic
       if (qperiodic(f) .ne. " ") then

          lfound = .false.
          do f1=1,tensor%ndof
             if (qperiodic(f) .eq. tensor%modelabel(f1)) then
                lfound = .true.
                f2 = f1
                exit ! loop
             end if
          end do

          if (lfound) then

             lperiodic(f2) = .true.

             if (dvr%basis(f2) .ne. 4 .and. dvr%basis(f2) .ne. 5) then
                routine =  "chk_periodic (symmops.f90)"
                message = "DOF '"//trim(qperiodic(f))// &
                     "' labeled as periodic is not of type FFT or EXP" &
                     //". Continuing anyways."
                call warnmsg
             end if

          else
             routine =  "chk_periodic (symmops.f90)"
             message = "Periodic DOF label not found in primitive basis: " &
                  //trim(qperiodic(f))
             call errormsg

          end if
       end if
    end do

    lwarn1 = lwarn
    lwarn = .false.
    do nsym=1,nsymmetries
       do f = 1, tensor%ndof

          f1 = symfmap(f,nsym)
          if (f1.ne.0) then
             if (lperiodic(f) .NEQV. lperiodic(f1) ) then
                routine =  "chk_periodic (symmops.f90)"
                message = "Permuting a periodic with a non-periodic DOF" &
                     //" in symmetry operation "//trim(symlabels(nsym))//" ." &
                     //"DOFs: "//trim(tensor%modelabel(f))//"," &
                     //trim(tensor%modelabel(f1))//"."
                call warnmsg
             end if
          end if
       end do
    end do
    if (lwarn) then
       routine =  "chk_periodic (reading SYMMETRY-SECTION)"
       message = "Permuting DOFs must both be periodic, or non-periodic." &
            //" See 'periodic' keyword."
       call errormsg
    end if


    do nsym=2,nsymmetries
       do f = 1, tensor%ndof
          if (lperiodic(f) .and. symoptablef(f,nsym) .eq. op_usersym) then
             routine =  "chk_periodic (symmops.f90)"
             message = "Symmetry "//trim(symlabels(nsym))//", " &
                  //"DOF: "//trim(tensor%modelabel(f))//":" &
                  //" 'usersym' operations must take care of " &
                  //" periodic boundaries themselves and cannot be used " &
                  //" together with the 'periodic' keyword."
             call warnmsg
          end if
       end do
    end do
    if (lwarn) then
       routine =  "chk_periodic (reading SYMMETRY-SECTION)"
       message = "Found periodic DOF used with 'usersym'. "
       call errormsg
    end if

    do nsym=2,nsymmetries
       do f = 1, tensor%ndof
          if (lperiodic(f) .and. .not. lcoordbased(f,nsym)) then
             routine =  "chk_periodic (symmops.f90)"
             message = "Symmetry "//trim(symlabels(nsym))//" ." &
                  //"DOF: "//trim(tensor%modelabel(f))//":" &
                  //" Symmetry operations for periodic DOF must " &
                  //" be coordinate-based using '{ }'."
             call warnmsg
          end if
       end do
    end do
    if (lwarn) then
       routine =  "chk_periodic (reading SYMMETRY-SECTION)"
       message = "Found grid based expressions for periodic DOF."
       call errormsg
    end if




    lwarn = lwarn1

    ! calculate box boundaries
    periodicbox = 0.d0
    do f = 1, tensor%ndof
       if (lperiodic(f)) then
          z = tensor%zort(f)
          xi = tensor%ort(z)
          z = tensor%zort(f) +  tensor%gdim(f) - 1
          xf = tensor%ort(z)
          dx = (xf-xi)/(tensor%gdim(f)-1)

          periodicbox(1,f) = xi
          periodicbox(2,f) = xf+dx
          periodicbox(3,f) = dx

       end if
    end do





  end subroutine chk_periodic


  subroutine  bcast_symmetrymaps(nmode1)

    implicit none
    integer, intent(IN) :: nmode1
    integer             :: itmp

#ifdef MPI
    integer             :: m, ierr, size

    call  MPI_Bcast(lmextgrid,nmode1,MPI_LOGICAL,0,MPI_COMM_WORLD,ierr)
    call  MPI_Bcast(nsymmetries,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    call  MPI_Bcast(symidxmapdim,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    call  MPI_Bcast(lcomplete,1,MPI_LOGICAL,0,MPI_COMM_WORLD,ierr)
    do m=1,nmode1
       call  MPI_Bcast(pointmaps(m)%npoint,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
       call  MPI_Bcast(pointmaps(m)%nlen,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
       call  MPI_Bcast(pointmaps(m)%ndof,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    end do

    if (mpirank.ne.0) then
       do m=1,nmode1
          allocate(pointmaps(m)%points(pointmaps(m)%ndof,pointmaps(m)%nlen))
          allocate(pointmaps(m)%indices(pointmaps(m)%ndof,pointmaps(m)%nlen))
       end do
       allocate(zsymidxmap(nmode1,nsymmetries))
       allocate(symidxmap(symidxmapdim))
       allocate(symmmap(nmode1,nsymmetries))
       allocate(permutedmodes(nmode1,nmode1))
       allocate(npermutedmodes(nmode1))
    end if

    ! wa ctually only need pointmaps(m)%points
    ! not pointmaps(m)%indices in mkdensities
    ! but this may be good for later use
    do m=1,nmode1
       size = pointmaps(m)%ndof*pointmaps(m)%nlen
       call  MPI_Bcast(pointmaps(m)%points,size, &
            MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,ierr)
       call  MPI_Bcast(pointmaps(m)%indices,size, &
            MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    end do

    ! mode mapping
    call  MPI_Bcast(symmmap, nmode1*nsymmetries, &
         MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

    ! pointer to mode point mappings
    call  MPI_Bcast(zsymidxmap, nmode1*nsymmetries, &
         MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

    ! mode point mappings
    call  MPI_Bcast(symidxmap,symidxmapdim , &
         MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

    ! flag if extended
    call  MPI_Bcast(lmextgrid, nmode1, MPI_LOGICAL, &
         0,MPI_COMM_WORLD,ierr)


    ! mode permutation counts
    call  MPI_Bcast(permutedmodes, nmode1**2, &
         MPI_INTEGER,0,MPI_COMM_WORLD,ierr)
    call  MPI_Bcast(npermutedmodes, nmode1, &
         MPI_INTEGER,0,MPI_COMM_WORLD,ierr)

#else
    itmp = nmode1 ! avoids unused nmode1 variable warning

#endif

  end subroutine bcast_symmetrymaps




  ! This detects which mode a usersym  operation
  ! maps to by observing which DOF change
  ! when DOF f1 is changed
  subroutine detect_usymmmapping(nsym,f1,tensor,m)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    integer, intent(IN)           :: nsym,f1
    integer, intent(OUT)          :: m

    double precision              :: qin(tensor%ndof)
    double precision              :: qout(tensor%ndof)
    double precision              :: qref(tensor%ndof)
    double precision              :: rdum
    character(len=16)             :: label
    integer                       :: i,n,f,z,f2,m1,m0
    logical                       :: lchange(tensor%ndof)
    integer, parameter            :: ntest = 50

    label = symlabels(nsym)
    lchange = .false.

    ! we do some tests with random grid points
    ! and see where changed coordinates apper
    do n=1,ntest
       ! set some random coordinate values
       do f=1,tensor%ndof
          call random_number(rdum)
          z = int(rdum*dble(tensor%gdim(f))) + tensor%zort(f)
          qin(f) = tensor%ort(z)
       end do
       qref = qin
       call GetUserSym(qin,qref,tensor%ndof,label)

       ! now detect the changes if we vary only one DOF.
       z =  tensor%zort(f1)-1
       do i=1,tensor%gdim(f1)
          qin(f1) =  tensor%ort(z+i)
          qout = qin
          call GetUserSym(qin,qout,tensor%ndof,label)
          do f2=1,tensor%ndof
             lchange(f2) =  lchange(f2) .or.  &
                  (abs(qout(f2) - qref(f2)) .gt. 1.d-7)
          end do
       end do
    end do

    if (.not. any(lchange)) then
       routine = "detect_usymmmapping"
       message = "User defined symmetry operation '"//trim(label)//"' " &
            // "does not do anything for DOF "//trim(tensor%modelabel(f1))
       call warnmsg
    end if


    m0 = -1
    do f2=1,tensor%ndof
       if (lchange(f2)) then
          m1=tensor%dofspf(f2)

          if (m0 .lt. 0) then
             m0=m1
          else if (m1.ne.m0) then
             routine = "detect_usymmmapping"
             write(message, '(a,2i4)') "Mode mixing in user defined " &
                  //"symmetry operation " &
                  //trim(symlabels(nsym)) // ", modes: ", m0,m1
             call warnmsg
          end if
       end if
    end do
    m = m0


  end subroutine detect_usymmmapping


  ! determine the lengths of the extended grids
  ! and add additional points to the mode grid

  subroutine detect_extgrids(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor

    double precision, allocatable :: point(:)
    type (pointmap_t), pointer    :: points_ptr
    integer, allocatable          :: idx(:), idx1(:), gdim(:),gdim1(:)
    integer                       :: npossibleextends(tensor%nmode)
    integer                       :: m,m1,m2,n,f,k,z, nsym
    double precision              :: qin(tensor%ndof)
    double precision              :: qout(tensor%ndof)
    logical                       :: lmsym(nsymmetries,tensor%nmode)
    integer                       :: fidxin(tensor%ndof), fidxout(tensor%ndof)

    interface
       subroutine GetSymPoint1m(coord_in, coord_out, idx_in, idx_out, &
            nsym, m, tensor)
         use tensortype

         class(tensor_t), intent(IN)     :: tensor
         integer, intent(IN)             :: nsym, m
         double precision, intent(IN)    :: coord_in(tensor%ndof)
         double precision, intent(INOUT) :: coord_out(tensor%ndof)
         integer, intent(IN)             :: idx_in(tensor%ndof)
         integer, intent(INOUT)          :: idx_out(tensor%ndof)
       end subroutine GetSymPoint1m
    end interface


    lmsym(:,:) = .false.
    do m=1,tensor%nmode

       ! first check if there is a usersym
       ! or an aritchmetic expression
       ! anywhere for this mode.
       ! and is so count the possible extensions
       ! if not, everything is grid based
       ! and hence will stay on the
       ! original grid for sure
       npossibleextends(m) = 0
       do n=1,tensor%nspfdof(m)
          f = tensor%spfdof(n,m)
          do nsym=1,nsymmetries
             if (symoptablef(f,nsym) .eq. op_usersym &
                  .or. lcoordbased(f,nsym)) then
                npossibleextends(m) = npossibleextends(m) + 1
                lmsym(nsym,m) = .true.
             end if
          end do
       end do
    end do


    ! if there is a "usersym" or a coordinate based trafo we allocate space
    ! for the original points and all possible extra points
    do m=1,tensor%nmode
       if (npossibleextends(m).gt.0) then
          pointmaps(m)%nlen = tensor%subdim(m)*(npossibleextends(m)+1)
          lmextgrid(m) = .true.
       else
          ! otherwise we just set the original numbers
          pointmaps(m)%nlen = tensor%subdim(m)
          lmextgrid(m) = .false.
       end if
       pointmaps(m)%ndof = tensor%nspfdof(m)
       allocate(pointmaps(m)%points(pointmaps(m)%ndof,pointmaps(m)%nlen))
       allocate(pointmaps(m)%indices(pointmaps(m)%ndof,pointmaps(m)%nlen))
    end do



    ! first we fill all pointmaps with the
    ! original primitive grid points
    do m=1,tensor%nmode
       n=tensor%nspfdof(m)
       if (allocated(gdim)) deallocate(gdim)
       allocate(gdim(n))
       if (allocated(idx)) deallocate(idx)
       allocate(idx(n))

       ! primitive grid dimensions of the mode m
       do n=1,tensor%nspfdof(m)
          f = tensor%spfdof(n,m)
          gdim(n) = tensor%gdim(f)
       end do

       ! first fill the lists with the original points
       ! (=='E' symmetry operation)
       do k=1,tensor%subdim(m)
          call cindex2sindex(k,idx,gdim,tensor%nspfdof(m))
          pointmaps(m)%indices(:,k) = idx
          do n=1,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)
             z = tensor%zort(f) + idx(n) - 1
             pointmaps(m)%points(n,k) = tensor%ort(z)
          end do
       end do

       pointmaps(m)%npoint = tensor%subdim(m)

    end do

    ! If there are no usersym keywords and no
    ! analytic symmetries we can exit here
    if (all(npossibleextends(:).eq.0)) return

    lmextgrid = .false.

    ! now we need to scan the user-symetries
    ! and arithmetic expressions
    ! if they produce any extra points
    ! outside the original primitive grid

    do m=1,tensor%nmode
       if (npossibleextends(m).gt.0) then

          n=tensor%nspfdof(m)

          if (allocated(point)) deallocate(point)
          allocate(point(n))
          if (allocated(gdim)) deallocate(gdim)
          allocate(gdim(n))
          if (allocated(gdim1)) deallocate(gdim1)
          allocate(gdim1(n))
          if (allocated(idx)) deallocate(idx)
          allocate(idx(n))
          if (allocated(idx1)) deallocate(idx1)
          allocate(idx1(n))

          ! primitive grid dimensions of the mode m
          do n=1,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)
             gdim(n) = tensor%gdim(f)
          end do

          ! build a default qin in all DOF
          ! and idxin
          do f=1,tensor%ndof
             if (modulo(tensor%gdim(f),2) .eq.1 ) then
                z = tensor%zort(f) + tensor%gdim(f)/2 + 1
                fidxin(f) =  tensor%gdim(f)/2 + 1
             else
                z = tensor%zort(f) + tensor%gdim(f)/2
                fidxin(f) =  tensor%gdim(f)/2
             end if
             qin(f) = tensor%ort(z)
          end do


          ! now we need to find all modes that map
          ! into mode m and transform their primitive
          ! grids with the symmetry operation and
          ! see if the outcome is already a point
          ! on mode m - otherwise add it.
          ! we start with the second symmetry - E is done
          do nsym=2,nsymmetries

             m1 = -1
             do m2=1,tensor%nmode
                if (symmmap(m2,nsym) .eq. m) then
                   m1 = m2
                   exit !m2 loop
                   ! only one mode m1 can map to m per
                   ! symmetry operation
                end if
             end do

             if (m1.eq.-1) then
                routine = "detect_extgrids (symmops.F90)"
                write(message,'(a,i4,a)') "Symmetry '" &
                     //trim(symlabels(nsym))//"':" &
                     //" There seems to be no mode that maps into mode ", m, "."
                call errormsg
             end if

             if (.not.lmsym(nsym,m) .and. .not. lmsym(nsym,m2)) then
                ! if there is no usersym or coordinate based
                ! symmetry operation for this mode for
                ! this symetry operation we can skip the
                ! rest of the loop.
                cycle
             end if


             ! m1 is our 'source' mode
             ! and from this we take all points
             ! which belong to the identity operation 'E'
             points_ptr => pointmaps(m1)

             do n=1,tensor%nspfdof(m1)
                f = tensor%spfdof(n,m1)
                gdim1(n) = tensor%gdim(f)
             end do

             ! now we run over the points
             ! of the mode m1 which maps into m
             do k=1,tensor%subdim(m1)

                call cindex2sindex(k,idx1,gdim1,tensor%nspfdof(m1))

                ! replace input q with values from the
                ! source mode m1...
                do n=1,tensor%nspfdof(m1)
                   f = tensor%spfdof(n,m1)
                   qin(f) = points_ptr%points(n,k)
                   fidxin(f) = idx1(n)
                end do

                ! perform symmetry operation
                call GetSymPoint1m(qin,qout, fidxin, fidxout, nsym, m, tensor)

                ! and pick the values for the
                ! destination  mode m
                ! qout is set for all dof
                do n=1,tensor%nspfdof(m)
                   f = tensor%spfdof(n,m)
                   point(n) = qout(f)
                   idx1(n) = fidxout(f)
                end do


                ! now find out if this point is already in the
                ! points list
                call finddtuple(pointmaps(m)%points, point, &
                     tensor%nspfdof(m), pointmaps(m)%npoint, &
                     tol_pointsearch,n)

                if (n .le. 0) then
                   ! if not found, add the point
                   n = pointmaps(m)%npoint + 1
                   pointmaps(m)%npoint = n
                   pointmaps(m)%points(:,n) = point(:)
                   pointmaps(m)%indices(:,n) = idx1(:)
                end if

             end do ! source mode
          end do !symmetries
       end if ! possible extends

       if  (pointmaps(m)%npoint .eq.  tensor%subdim(m)) then
          ! if no extra points have beem added and
          ! we do not need an extended grid
          lmextgrid(m) = .false.
       else
          lmextgrid(m) = .true.
       end if


    end do ! modes


  end subroutine detect_extgrids



  subroutine save_extgrids(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    logical                       :: lallongrid, lexists
    integer                       :: m,n,f
    character(len=c5)             :: fname
    character(len=c3)             :: formatstr

    lallongrid = .true.
    do m=1,tensor%nmode
       if (pointmaps(m)%npoint .ne. tensor%subdim(m)) then
          lallongrid = .false.

          write(fname,'(a,i4)') trim(name)//"/extgrid_mode_", m
          call rmblnk(fname)
          inquire(file=trim(fname), EXIST=lexists)
          if (lexists.and..not.lovwr) then
             routine = "build_symops, writing usersym that goes off the grid."
             message = "File exists, cannot overwrite. Use -w option: " &
                  //trim(fname)
             call errormsg
          end if

          n = tensor%nspfdof(m)
          write(formatstr,'(a,i4,a)') "(",n,'f16.8)'
          call rmblnk(formatstr)

          open(idump,file=trim(fname), status='unknown')
          ! file header
          write(idump,'(a,i4)') "# extended coordinate values for mode ", m
          write(idump,'(a)', advance="no")"# "
          f = tensor%spfdof(1,m)
          write(idump,'(a14)', advance="no")adjustr( tensor%modelabel(f)(1:14))
          do n=2,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)
             write(idump,'(a16)', advance="no") adjustr(tensor%modelabel(f))
          end do
          write(idump,'(a)') ""
          do n=1,pointmaps(m)%npoint
             write(idump, formatstr) (pointmaps(m)%points(f,n), &
                  f=1,tensor%nspfdof(m))
          end do
          close(idump)

       end if
    end do


  end subroutine save_extgrids


  ! this builds the direct mode-2-mode grid mappings
  subroutine build_indexmap(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    integer                       :: m,m1,f,nsym,z,z1,k1,k,n
    integer                       :: idx(tensor%ndof)
    integer                       :: idx1(tensor%ndof)
    double precision              :: qin(tensor%ndof)
    double precision              :: qout(tensor%ndof)
    double precision              :: point(tensor%maxspfdof)

    interface
       subroutine GetSymPoint1m(coord_in, coord_out, idx_in, idx_out, &
            nsym, m, tensor)
         use tensortype

         class(tensor_t), intent(IN)     :: tensor
         integer, intent(IN)             :: nsym, m
         double precision, intent(IN)    :: coord_in(tensor%ndof)
         double precision, intent(INOUT) :: coord_out(tensor%ndof)
         integer, intent(IN)             :: idx_in(tensor%ndof)
         integer, intent(INOUT)          :: idx_out(tensor%ndof)
       end subroutine GetSymPoint1m
    end interface


    ! identity first
    nsym = 1
    do m=1,tensor%nmode
       z = zsymidxmap(m,nsym)
       do k1=1,pointmaps(m)%npoint
          symidxmap(z+k1-1) = k1
       end do
    end do


    do nsym=2,nsymmetries
       do m=1,tensor%nmode
          z = zsymidxmap(m,nsym)
          m1 = symmmap(m,nsym)

          ! build a default qin in all DOF
          do f=1,tensor%ndof
             if (modulo(tensor%gdim(f),2) .eq.1 ) then
                z1 = tensor%zort(f) + tensor%gdim(f)/2 + 1
                idx(f) =  tensor%gdim(f)/2 + 1
             else
                z1 = tensor%zort(f) + tensor%gdim(f)/2
                idx(f) =  tensor%gdim(f)/2
             end if
             qin(f) = tensor%ort(z1)
          end do

          do k1=1,pointmaps(m1)%npoint

             do n=1,tensor%nspfdof(m1)
                f = tensor%spfdof(n,m1)
                qin(f) = pointmaps(m1)%points(n,k1)
                idx(f) = pointmaps(m1)%indices(n,k1)

             end do

             call GetSymPoint1m(qin, qout, idx, idx1, nsym, m, tensor)

             do n=1,tensor%nspfdof(m)
                f = tensor%spfdof(n,m)
                point(n) = qout(f)
             end do

             call finddtuple(pointmaps(m)%points,point,tensor%nspfdof(m), &
                  pointmaps(m)%npoint,tol_pointsearch,k)

             if (k.eq.-1 .and. .not. lwarn) then
                routine = "symmops, build_indexmap"
                message = "Did not find matching point in intexmaps. See sym.log. " // &           
                     "If you used 'usersym' maybe check if symmetry operations are obayed. "
                call warnmsg
    
             end if

             symidxmap(z+k1-1) = k

          end do
       end do
    end do


  end subroutine build_indexmap



  subroutine chk_symops(tensor,Efmap)
    implicit none

    class(tensor_t), intent(IN)    :: tensor

    integer, intent(OUT)          :: Efmap(tensor%ndof)

    integer                       :: n,f,f1
    character(len=16)             :: buf
    logical                       :: lfound

    lwarn = .false.
    routine = "chk_symops (symmops.F90)"


    ! check for unknown modelabels
    ! and determine the order of the DOF
    Efmap = -1
    do n=1,nsymmetries
       do f=1,tensor%ndof
          if (symtable(f,n) .eq. "usersym") then
             cycle ! f loop
          else if (symtable(f,n) .eq. "-usersym") then
             message = "keyword '-usersym' is not allowed in SYMMETRY-SECTION."
             call warnmsg
          end if
          if (symtable(f,n)(1:1) .eq. "-") then
             buf = symtable(f,n)(2:)
          else
             buf = symtable(f,n)(1:)
          end if

          lfound = .false.
          do f1=1,tensor%ndof
             if (tensor%modelabel(f1) .eq. buf) then
                lfound = .true.
                exit ! f1 loop
             end if
          end do

          if (.not. lfound) then
             message = "Modelabel in SYMMETRY-SECTION does not " &
                  //"match a label of PBASIS: "//trim(buf)
             call warnmsg
             ! lwarn is now .true.
          end if


          ! assign order map based on the "E" operation
          if (n .eq. 1) then
             do f1=1,tensor%ndof
                if (trim(symtable(f,1)) .eq. tensor%modelabel(f1)) then
                   Efmap(f) = f1
                end if
             end do
          end if

       end do
    end do

    if (lwarn) then
       message = "An error was found while checking the symmetry table."
       call errormsg
    end if

  end subroutine chk_symops



  subroutine chk_sym_mapping(tensor)
    implicit none
    class(tensor_t), intent(IN)    :: tensor
    integer                       :: f,f1,nsym,m,m1,z,n,n1
    logical                       :: lseen

    routine = "chk_sym_mapping"

    ! check symfmap
    do nsym=1,nsymmetries
       do f=1,tensor%ndof
          lseen = .false.
          do f1=1,tensor%ndof
             ! this only works because symfmap(f1,nsym)
             ! is 0 for 'usersym' and the f-loop starts at 1.
             if (symfmap(f1,nsym) .eq. f) then
                if (lseen) then
                   message = "Symmetry "//trim(symlabels(nsym)) &
                        //": DOF-label "//trim(tensor%modelabel(f)) &
                        //" is assigned twice."
                   call errormsg
                else
                   lseen = .true.
                end if
             end if
          end do
       end do
    end do

    ! check symmmap
    ! (should be covered by the f-mapping test
    ! but better check twice...)
    do nsym=1,nsymmetries
       do m=1,tensor%nmode
          lseen = .false.
          do m1=1,tensor%nmode
             if (symmmap(m1,nsym) .eq. m) then
                if (lseen) then
                   write(message,'(a,i4,a)') "Symmetry "//trim(symlabels(nsym)) &
                        //": mode", m, &
                        " is assigned twice."
                   call errormsg
                else
                   lseen = .true.
                end if
             end if
          end do
       end do
    end do

    ! indexmapping
    do nsym=1,nsymmetries
       do m=1,tensor%nmode
          z = zsymidxmap(m,nsym)
          do n=1, pointmaps(m)%npoint
             lseen = .false.
             do n1=1, pointmaps(m)%npoint
                if (symidxmap(z+n1-1) .eq. n) then
                   lseen = .true.
                   exit ! inner loop
                end if
             end do
             ! we must see all 'n' in the mapping
             if (.not. lseen) then
                write(message,'(a,i4,a)') "Symmetry "//trim(symlabels(nsym)) &
                     //", mode", m, &
                     ": grid index mapping is not closed. If the symmtries are " &
                     // "correctly implemented than this hints to the existence " &
                     // "of further symmetries that must be added."
                call errormsg
             end if
          end do
       end do
    end do



  end subroutine chk_sym_mapping


  subroutine opensymlog(fname)
    implicit none

    character(len=*), intent(IN) :: fname
    character(len=100)           :: buffer
    logical                      :: lexists

    inquire(file=trim(fname), EXIST=lexists)
    if (lexists.and..not.lovwr) then
       routine = "opensymlog"
       message = "File exists, cannot overwrite. Use -w option: "//trim(fname)
       call errormsg
    end if
    open(isymlog,file=trim(fname))
    write(isymlog,'(a)') "SYMMETRY LOG"
    call fpstartup(buffer)
    write(isymlog,*)
    write(isymlog,'(a80)') buffer
    write(isymlog,*)


  end subroutine opensymlog



  subroutine logsymmops(tensor)
    implicit none

    class(tensor_t), intent(IN)    :: tensor
    character(len=c3)             :: formatstr
    character(len=c5)             :: buf
    character(len=c5)             :: line
    integer                       :: m,f,i,nsym
    integer                       :: idx(tensor%ndof)
    integer                       :: idx1(tensor%ndof)
    double precision              :: ran(tensor%ndof)
    double precision              :: vorig, veqv
    double precision              :: coord(tensor%ndof)
    double precision              :: coord1(tensor%ndof)
    integer                       :: ntest, nmiss, nhit

    interface
       subroutine GetSymPoint(coord_in, coord_out, idx_in, idx_out, &
            nsym, tensor)
         use tensortype
         implicit none
         class(tensor_t), intent(IN)    :: tensor
         integer, intent(IN)           :: idx_in(tensor%ndof), nsym
         integer, intent(OUT)          :: idx_out(tensor%ndof)
         double precision, intent(IN)  :: coord_in(tensor%ndof)
         double precision, intent(OUT) :: coord_out(tensor%ndof)
       end subroutine GetSymPoint

    end interface


    if (mpirank.ne.0) return
    if (.not. lsymmetries) return

    line = ""
    do i=1,nsymmetries*10
       line(i:i) = "-"
    end do


    write(isymlog,*)
    write(isymlog,'(a)') "SYMMETRYSATION"
    write(isymlog,'(a)') "=============="

    if (any(lperiodic)) then
       write(isymlog,*)
       write(isymlog,'(a)') "Periodic DOF"
       write(isymlog,*)
       write(isymlog,'(a)') "DOF         xi*         xf*"
       do f=1,tensor%ndof
          if (lperiodic(f)) then
             write(isymlog,'(a8,2f12.6)') adjustr(tensor%modelabel(f)(1:8)), &
                  periodicbox(1,f),periodicbox(2,f)
          end if
       end do

       write(isymlog,*)
       write(isymlog,'(a)') "*xi (inclusive): start of peripdic interval"
       write(isymlog,'(a)') "  for coordinate based operations"
       write(isymlog,'(a)') "*xf (exlusive):  end of peripdic of interval"
       write(isymlog,'(a)') "  for coordinate based operations "
       write(isymlog,*)

    end if



    write(isymlog,*)
    write(isymlog,'(a)') "DOF permutations from inp. (if usersym then 0)"
    write(isymlog,*)
    write(formatstr,'(a,i4,a)') "(",nsymmetries, "a)"
    call rmblnk(formatstr)
    write(isymlog,formatstr) (adjustr(symlabels(nsym)(1:10)),nsym=1,nsymmetries)
    write(isymlog,'(a)') trim(line)
    write(formatstr,'(a,i4,a)') "(",nsymmetries, "i10)"
    call rmblnk(formatstr)
    do f=1,tensor%ndof
       write(buf,formatstr) (symfmap(f,nsym), nsym=1,nsymmetries)
       write(isymlog,'(a)') trim(buf)
    end do
    write(isymlog,*)

    write(isymlog,*)
    write(isymlog,'(a)') "DOF grid operation from inp."
    write(isymlog,*)
    write(formatstr,'(a,i4,a)') "(",nsymmetries, "a)"
    call rmblnk(formatstr)
    write(isymlog,formatstr) (adjustr(symlabels(nsym)(1:10)),nsym=1,nsymmetries)
    write(isymlog,'(a)') trim(line)
    do f=1,tensor%ndof
       do nsym=1,nsymmetries
          select case(symoptablef(f,nsym))
          case(op_ident_grid)
             write(isymlog,'(a10)',advance="no") "     ident"
          case(op_revert_grid)
             write(isymlog,'(a10)',advance="no") "    invert"
          case(op_usersym)
             write(isymlog,'(a10)',advance="no") "   usersym"
          case(op_ident_coord)
             write(isymlog,'(a10)',advance="no") "   dof + R"
          case(op_revert_coord)
             write(isymlog,'(a10)',advance="no") "  -dof + R"
          case default
             write(isymlog,'(a10)',advance="no") "     ERROR"
          end select
       end do
       write(isymlog,*)
    end do


    if (any(lcoordbased)) then
       write(isymlog,*)
       write(isymlog,*)
       write(isymlog,'(a)') "DOF R shift from inp."
       write(isymlog,*)
       write(formatstr,'(a,i4,a)') "(",nsymmetries, "a)"
       call rmblnk(formatstr)
       write(isymlog,formatstr) (adjustr(symlabels(nsym)(1:10)),nsym=1,nsymmetries)
       write(isymlog,'(a)') trim(line)
       do f=1,tensor%ndof
          do nsym=1,nsymmetries
             if (abs(coordshift(f,nsym)) .gt. 1.d-6) then
                write(isymlog,'(f10.5)',advance="no") coordshift(f,nsym)
             else
                write(isymlog,'(i10)',advance="no") 0
             end if
          end do
          write(isymlog,*)
       end do
    end if

    write(isymlog,*)
    write(isymlog,*)
    write(isymlog,'(a)') "Mode permutations"
    write(isymlog,*)

    write(formatstr,'(a,i4,a)') "(",nsymmetries, "a)"
    call rmblnk(formatstr)
    write(isymlog,formatstr) (adjustr(symlabels(nsym)(1:10)),nsym=1,nsymmetries)
    write(isymlog,'(a)') trim(line)
    write(formatstr,'(a,i4,a)') "(",nsymmetries, "i10)"
    call rmblnk(formatstr)
    do m=1,tensor%nmode
       write(buf,formatstr) (symmmap(m,nsym), nsym=1,nsymmetries)
       write(isymlog,'(a)') trim(buf)
    end do
    write(isymlog,*)


    if ( any(lmextgrid(:))) then
       write(isymlog,*)
       write(isymlog,*)
       write(isymlog,'(a)') "The following extended grids sizes are used:"
       write(isymlog,'(a)')
       write(isymlog, '(a)') "mode   grid_orig    grid_ext"
       write(isymlog, '(a)') "----------------------------"
       do m=1,tensor%nmode
          if (lmextgrid(m)) then
             write(isymlog, '(i4,2i12)')m, tensor%subdim(m), pointmaps(m)%npoint
          end if
       end do
       write(isymlog,*)
    end if
    write(isymlog,*)

    ! we test the symmetries on the original potential
    ! and see if they are correct
    write(isymlog,*)
    write(isymlog,'(a)') &
         " ======= Testing Symmetry implementation on potential routine ======="

    write(isymlog,'(a)')
    write(isymlog,'(a)') &
         "  * 50 DVR points are chosen at random and converted to coordinate values"
    write(isymlog,'(a)') &
         "  * All symmetry operations are applied to each point."
    write(isymlog,'(a)') &
         "  * If the resulting energies are the same for all symmetries"
    write(isymlog,'(a)') &
         "    then the symmetries are implemented correctly"
    write(isymlog,'(a)') &
         "    (a warning will be issued if not)."
    write(isymlog,'(a)') &
         "  * A similar test will be performed on the PES fit. See below."

    write(isymlog,'(a)')
    write(isymlog,'(a)')

    ntest = 50
    nhit = 0
    nmiss = 0
    do i=1,ntest

       call random_number(ran)
       do f=1,tensor%ndof
          idx(f) = int(ran(f)*tensor%gdim(f)) + 1
       end do
       call dvrindex2coord(tensor,idx,coord)

       write(isymlog,*)
       write(isymlog,'(a,i4)')   "Test", i
       write(isymlog,'(a)') &
            "Sym                   V             Coordinates"
       write(isymlog,'(a)') &
            "---------------------------------------------------------"

       if (lidxbased) then
          vorig = potential(idx, tensor%ndof)
          write(isymlog, '(a, f16.8,100i4)') symlabels(1), vorig,  &
               (idx(f), f=1,tensor%ndof)
       else
          vorig = potentialQ(coord, tensor%ndof)
          write(isymlog, '(a, f16.8,100f10.4)') symlabels(1), vorig,  &
               (coord(f), f=1,tensor%ndof)
       end if

       do nsym=2,nsymmetries
          call GetSymPoint(coord, coord1, idx, idx1, nsym, tensor)

          veqv = potentialQ(coord1, tensor%ndof)

          write(isymlog, '(a, f16.8,100f10.4)') symlabels(nsym), veqv,  &
               (coord1(f), f=1,tensor%ndof)

          if(abs(vorig - veqv) .gt. 1.d-6) then
             nmiss = nmiss + 1
          end if

       end do
    end do



    write(isymlog,*)
    write(isymlog,'(a,i6)') "Number of PES values deviating from E: ", nmiss


    if (nmiss .ne. 0) then
       message = "Symmetry operations do not always yield equivalent energies " &
            //"when calling the PES routine. See sym.log for details"
       routine = "logsymmops"

       write(isymlog,'(a)')
       write(isymlog,'(a)') "WARNING:"
       write(isymlog,'(a)') message

       call warnmsg
       lwarn = .false.
    else
       write(isymlog,'(a)') "Check OK."
       write(isymlog,*)
    end if



    call myflush(isymlog)


  end subroutine logsymmops



  subroutine build_simple_sym(tensor)
    implicit none

    class(tensor_t), intent(IN) :: tensor

    integer                    :: n,m,f,k,z
    integer, allocatable       :: gdim(:), idx(:)


    call free_pointmaps()
    allocate(pointmaps(tensor%nmode))
    allocate(lmextgrid(tensor%nmode))
    lmextgrid = .false.

    do m=1,tensor%nmode
       pointmaps(m)%nlen = tensor%subdim(m)
       lmextgrid(m) = .false.
       pointmaps(m)%ndof = tensor%nspfdof(m)
       allocate(pointmaps(m)%points(pointmaps(m)%ndof,pointmaps(m)%nlen))
       allocate(pointmaps(m)%indices(pointmaps(m)%ndof,pointmaps(m)%nlen))
    end do


    do m=1,tensor%nmode
       n=tensor%nspfdof(m)
       if (allocated(gdim)) deallocate(gdim)
       allocate(gdim(n))
       if (allocated(idx)) deallocate(idx)
       allocate(idx(n))

       ! primitive grid dimensions of the mode m
       do n=1,tensor%nspfdof(m)
          f = tensor%spfdof(n,m)
          gdim(n) = tensor%gdim(f)
       end do

       ! first fill the lists with the original points
       ! (=='E' symmetry operation)
       do k=1,tensor%subdim(m)
          call cindex2sindex(k,idx,gdim,tensor%nspfdof(m))
          pointmaps(m)%indices(:,k) = idx
          do n=1,tensor%nspfdof(m)
             f = tensor%spfdof(n,m)
             z = tensor%zort(f) + idx(n) - 1
             pointmaps(m)%points(n,k) = tensor%ort(z)
          end do
       end do

       pointmaps(m)%npoint = tensor%subdim(m)

    end do


  end subroutine build_simple_sym

subroutine dispose_symops
  use symmops_m
  use mcpeinsym

  call clean_mcpeinsym()
  call free_pointmaps()

  if (allocated(sigma)) deallocate(sigma)
  if (allocated(zsigma)) deallocate(zsigma)


  if (allocated(symfmap)) deallocate(symfmap) ! dof mapping
  if (allocated(symmmap)) deallocate(symmmap) ! mode mapping

  if (allocated(symoptablef)) deallocate(symoptablef)! dof operations
  if (allocated(symoptablem)) deallocate(symoptablem)! mode operations


  if (allocated(zsymidxmap)) deallocate(zsymidxmap)
  if (allocated(symidxmap)) deallocate(symidxmap)

  if (allocated(lmextgrid)) deallocate(lmextgrid)


  if (allocated(symlabels)) deallocate(symlabels)


  if (allocated(lperiodic)) deallocate(lperiodic)     ! is f periodic?
  if (allocated(periodicbox)) deallocate(periodicbox) ! xi, xf, dx of interval


  if (allocated(coordshift)) deallocate(coordshift)  ! R for expressions as "f' = f+R"
  if (allocated(lcoordbased)) deallocate(lcoordbased) ! is expr. like  "f' = f+R"?
  if (allocated(lgridbased)) deallocate(lgridbased)  ! is grid based operation?



  if (allocated(sym2map)) deallocate(sym2map)

  if (allocated(symEmap)) deallocate(symEmap)
  if (allocated(permutedmodes)) deallocate(permutedmodes)
  if (allocated(npermutedmodes)) deallocate(npermutedmodes)

  sigmadim = 0
  symidxmapdim = 0
  nsymmetries = 0


end subroutine dispose_symops



subroutine symHermitianOperator(tensor,dvr)
  use mcpeinsym
  use symmops_m
  use dvrtype_m
  use cpdtype_m
  implicit none

  type(cpd_t), intent(IN)        :: tensor
  type(dvr_t), intent(IN)        :: dvr
  integer                        :: f

  nperiodic = 0
  nsymops = 2

  allocate(qperiodic(tensor%ndof))
  allocate(symtable(0:tensor%ndof,nsymops))
  allocate(symrshift(0:tensor%ndof,nsymops))
  allocate(lsymarithmetic(0:tensor%ndof,nsymops))

  lsymarithmetic = .false.
  symrshift = 0.d0

  symtable(0,1) = "E"
  symtable(0,2) = "Hconj"

  do f=1,tensor%ndof
     symtable(f,1) = tensor%modelabel(f)
  end do
  do f=1,tensor%ndof,2
     symtable(f,2) = tensor%modelabel(f+1)
     symtable(f+1,2) = tensor%modelabel(f)
  end do

  call build_symops(tensor, dvr)


end subroutine symHermitianOperator

end module symmops_s
