
module tensortools
  implicit none

  private

  public :: toverlap

contains

  ! polymorphic function toverlap
  ! calculates the overlap of two tensors t1, and t2
  ! which may be of type cpd_t (candecomp)
  ! or natpot_t
  double precision function toverlap (t1,t2)
    use global
    use tensortype
    use natpottype
    use cpdtype_m
    implicit none
    class (tensor_t), intent(IN) :: t1,t2
    double precision             :: ovlp

    routine = "toverlap(t1,t2), (tensortools.f90)"

    if (.not. iscompatible(t1, t2)) then
       ! not all is implemented....
       message = "Incompatible tensors."
       call errormsg
    end if

    select type (t1)
    type is (natpot_t)
       select type (t2)
       type is (natpot_t)
          if (t1%modc .ne. t2%modc) then
             message =  "Natpots have different contracted mode. " &
                  //"Not implemented."
             call errormsg
          end if
          call toverlap_np_np(t1, t2, ovlp)
       type is (cpd_t)
          call toverlap_np_cpd(t1, t2, ovlp)
       class default
          message = "Unknown type of tensor t2"
          call errormsg
       end select
    type is (cpd_t)
       select type (t2)
       type is (natpot_t)
          call toverlap_np_cpd(t2, t1, ovlp)
       type is (cpd_t)
          call toverlap_cpd_cpd(t1, t2, ovlp)
       class default
          message = "Unknown type of tensor t2"
          call errormsg
       end select
    class default
       message = "Ungnown type of tensor t1"
       call errormsg
    end select

    toverlap = ovlp

  end function toverlap


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !    implementations (private)    !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! overlap of two natpots
  subroutine toverlap_np_np(natpot1, natpot2, ovlp)
    use natpottype
    use mtlib, only: mttxdd
    implicit none
    type(natpot_t), intent(IN)    :: natpot1, natpot2
    double precision, intent(OUT) :: ovlp

    double precision, allocatable :: mat(:,:)
    double precision, allocatable, target :: work1(:), work2(:)
    double precision, pointer     :: pwork(:) => NUll()
    double precision, pointer     :: pcoeff(:) => NUll()
    integer                       :: m,n1,n2, z11,z12,z21,z22
    integer                       :: workdim
    integer                       :: dim1,dim2,dim3,dim4
    logical                       :: lstart

    ! we trust that the natpots have the same
    ! structure - only the potdims may be different.
    ! all other must be the same, in particular the
    ! contacted mode

    workdim = 1
    do m=1,natpot1%nmode
       if (m .ne. natpot1%modc) then
          workdim = workdim * max(natpot1%potdim(m),natpot2%potdim(m))
       else
          workdim = workdim * max(natpot1%subdim(m),natpot2%subdim(m))
       end if
    end do

    allocate(work1(workdim))
    allocate(work2(workdim))


    pcoeff => work1
    pwork => work2

    lstart = .true.

    do m=1,natpot1%nmode

       if (m.ne.natpot1%modc) then

          allocate(mat(natpot2%potdim(m), natpot1%potdim(m)))

          ! build overlap matrix
          do n1=1,natpot1%potdim(m)

             z11 = natpot1%zevec(m,n1)
             z12 = z11 + natpot1%subdim(m)-1

             do n2=1,natpot2%potdim(m)

                z21 = natpot2%zevec(m,n2)
                z22 = z21 + natpot2%subdim(m)-1

                mat(n2,n1) = dot_product( &
                     natpot1%evec(z11:z12), &
                     natpot2%evec(z21:z22))

             end do
          end do

          ! operate overlap matrix on coeff
          dim1 = natpot1%potvdim(m)
          dim2 = natpot2%potdim(m)
          dim3 = natpot2%potndim(m)
          dim4 = natpot1%potdim(m)

          if (lstart) then
             call mttxdd (mat,natpot2%coeff,pwork,dim1,dim2,dim3,dim4)
             lstart = .false.
          else
             call mttxdd (mat,pcoeff,pwork,dim1,dim2,dim3,dim4)
          end if

          if (associated(pcoeff, work1)) then
             pcoeff => work2
             pwork => work1
          else
             pcoeff => work1
             pwork => work2
          end if

          deallocate(mat)

       end if

    end do

    ovlp = 0.d0
    !$omp parallel do reduction(+:ovlp)
    do n1=1,natpot1%cdim
       ovlp = ovlp + natpot1%coeff(n1) * pcoeff(n1)
    end do
    !$omp end parallel do


    deallocate(work1, work2)

  end subroutine toverlap_np_np



  ! overlap of a natpot with a candecomp
  subroutine toverlap_np_cpd(natpot, cpd, ovlp)
    use tensortype
    use natpottype
    use cpdtype_m
    use mtlib, only: mttxdd
    implicit none
    type(natpot_t), intent(IN)    :: natpot
    type(cpd_t), intent(IN)       :: cpd
    double precision, intent(OUT) :: ovlp

    double precision, allocatable, target :: work1(:)
    double precision, allocatable, target :: work2(:)
    double precision, pointer     :: pwork(:), pcoeff(:)
    double precision              :: sppovlp(natpot%maxspf)
    integer                       :: r,n,m, zn1,zn2
    integer                       :: dim1,dim2,dim3,dim4
    logical                       :: lmodc

    integer                       :: work1dim, work2dim




    ! first we need to find the minimum
    ! worksize
    work1dim = 0
    work2dim = 0
    lmodc = .false.
    do m=1,natpot%nmode
       if (m .ne. natpot%modc) then

          if (lmodc) then
             dim1 = natpot%subdim(natpot%modc)
          else
             dim1 = 1
          end if
          dim2 = natpot%potdim(m)
          dim3 = natpot%potndim(m)
          dim4 = 1
       else
          lmodc = .true.
       end if

       work2dim = max(work2dim, dim1*dim4*dim3)
       if (m .gt. 1) then
          work1dim = max(work1dim, dim1*dim2*dim3)
       end if

    end do

    ovlp = 0.d0 ! init here - not in parallel region!
    ! thar can go wrong of rank is smaller then num_threads
    ! and thread 0 is not getting a share

    !$omp parallel private(r,work1, work2, pwork, pcoeff,m,n,sppovlp) &
    !$omp private(lmodc,zn1,zn2,dim1,dim2,dim3,dim4)

    allocate(work1(work1dim))
    allocate(work2(work2dim))



    !$omp do reduction(+:ovlp)
    do r=1,cpd%rank

       pcoeff => work1
       pwork => work2

       lmodc = .false.
       do m=1,natpot%nmode

          if (m.ne. natpot%modc) then
             do n=1,natpot%potdim(m)
                zn1 = natpot%zevec(m,n)
                zn2 = zn1 + natpot%subdim(m) - 1
                sppovlp(n) = dot_product(cpd%spp(m)%htm(:,r), natpot%evec(zn1:zn2))
             end do


             if (lmodc) then
                dim1 = natpot%subdim(natpot%modc)
             else
                dim1 = 1
             end if
             dim2 = natpot%potdim(m)
             dim3 = natpot%potndim(m)

             dim4 = 1

             if (m.eq. 1) then
                call mttxdd (sppovlp,natpot%coeff,pwork,dim1,dim2,dim3,dim4)
             else
                call mttxdd (sppovlp,pcoeff,pwork,dim1,dim2,dim3,dim4)
             end if

             if (associated(pcoeff, work1)) then
                pcoeff => work2
                pwork => work1
             else
                pcoeff => work1
                pwork => work2
             end if
          else
             lmodc = .true.
          end if
       end do ! m

       if (natpot%modc .ne. 0) then
          ! if there is a contracted
          ! mode it should now be
          ! concentrated at the beginning
          ! of pcoeff
          m = natpot%modc
          ovlp = ovlp + dot_product(cpd%spp(m)%htm(:,r),pcoeff(1:natpot%subdim(m))) &
               * cpd%coeff(r)
       else
          ! if there is not a contracted mode
          ! the result shoul be in the first element of
          ! pcoeff
          ovlp = ovlp + pcoeff(1)*cpd%coeff(r)
       end if
    end do
    !$omp end do

    deallocate(work1,work2)

    !$omp end parallel



  end subroutine toverlap_np_cpd


  ! overlap of two candecomps
  subroutine toverlap_cpd_cpd(cpd1, cpd2, ovlp)
    use tensortype
    use cpdtype_m
    implicit none
    type(cpd_t), intent(IN), target :: cpd1, cpd2
    double precision, intent(OUT)   :: ovlp

    double precision  :: DDOT, rtmp(cpd1%rank, cpd2%rank)
    integer           :: r1, r2, m

    type(cpd_t), pointer   :: p1,p2
    external DDOT

    p1 => cpd1
    p2 => cpd2

    ovlp = 0.d0

    if (associated(p1,p2)) then
       ! the two cpd's are actually the same objects

       m = 1
       !$omp parallel do private(r1) schedule(guided)
       do r2=1,cpd2%rank
          do r1=r2,cpd1%rank
             rtmp(r1,r2) = cpd1%coeff(r1) * cpd2%coeff(r2)
          end do
       end do
       !$omp end parallel do

       do m=1,cpd1%nmode
          !$omp parallel do private(r1) schedule(guided)
          do r2=1,cpd2%rank
             do r1=r2,cpd1%rank
                rtmp(r1,r2) = rtmp(r1,r2) * DDOT(cpd1%subdim(m),cpd1%spp(m)%htm(:,r1),1,cpd2%spp(m)%htm(:,r2),1)
             end do
          end do
          !$omp end parallel do
       end do

       ovlp = 0.d0
       !$omp parallel do private(r1) reduction(+: ovlp) schedule(guided)
       do r2=1,cpd2%rank
          ovlp = ovlp +  rtmp(r2,r2)
          do r1=r2 + 1,cpd1%rank
             ovlp = ovlp +  rtmp(r1,r2)*2.d0
          end do
       end do
       !$omp end parallel do

    else
       ! the two cpd objects are not the same objects

       m = 1
       !$omp parallel do private(r1) schedule(static)
       do r2=1,cpd2%rank
          do r1=1,cpd1%rank
             rtmp(r1,r2) =  DDOT(cpd1%subdim(m),cpd1%spp(m)%htm(:,r1),1,cpd2%spp(m)%htm(:,r2),1) &
                  * cpd1%coeff(r1) * cpd2%coeff(r2)
          end do
       end do
       !$omp end parallel do

       do m=2,cpd1%nmode
          !$omp parallel do private(r1) schedule(static)
          do r2=1,cpd2%rank
             do r1=1,cpd1%rank
                rtmp(r1,r2) = rtmp(r1,r2) * DDOT(cpd1%subdim(m),cpd1%spp(m)%htm(:,r1),1,cpd2%spp(m)%htm(:,r2),1)
             end do
          end do
          !$omp end parallel do
       end do

       ovlp = 0.d0
       !$omp parallel do private(r1) reduction(+: ovlp) schedule(guided)
       do r2=1,cpd2%rank
          do r1=1,cpd1%rank
             ovlp = ovlp + rtmp(r1,r2)
          end do
       end do
       !$omp end parallel do

    end if

  end subroutine toverlap_cpd_cpd





end module tensortools


