!***********************************************************************
!
!                         QUANTICS TOP LAYER
!
! This program is free software: you can redistribute it and/or modify
!   it under the terms of the GNU General Lesser Public License as 
!   published by the Free Software Foundation, version 3.

!   This program is distributed in the hope that it will be useful,
!   but WITHOUT ANY WARRANTY; without even the implied warranty of
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!   GNU General Lesser Public License for more details.

!   To see a copy of the GNU General Lesser Public License see 
!   <http://www.gnu.org/licenses/>.
!
!***********************************************************************

      program quantics

!$ use omp_lib

#ifdef MPI
use mpi
#endif
#ifdef OMP
use omp_lib
#endif
use mpidata

use maxv
use global
use channels

use eingabemod, only: eingabe

use logdvrmod
use logopmod
use loginwfmod
use logdat
use openmpmod

use prop_zagreb
use dd_db, only: dbfinalise

      implicit none
      
      real(dop)                                  :: tcpu,tsys,ttot,tcpu0,tsys0

      integer(long)                              :: t1,t2,icount,count_rate
      character(len=c5)                          :: filename

!-----------------------------------------------------------------------
! Get the time when the job is started
!-----------------------------------------------------------------------
      call fpstartup(cstartup)
      call cpu_time(tcpu0)
      call system_clock(icount,count_rate)
      tsys0 = dble(icount)/dble(count_rate)

!-----------------------------------------------------------------------
! Initialize MPI-World if distributed memory parallelization is used.
!-----------------------------------------------------------------------
      master = 0
#ifdef MPI
      call MPI_Init(mpierr)
      call mpi_comm_size(MPI_COMM_WORLD,nmpi,mpierr)
      call mpi_comm_rank(MPI_COMM_WORLD,mpirank,mpierr)
#else
      nmpi=1
      mpirank=0
#endif

!-----------------------------------------------------------------------
! Don't allow nested OMP parallelism to retain maximum threads to
! maximum cores asked for
!-----------------------------------------------------------------------
#ifdef OMP
     call omp_set_nested(.false.)
#endif

!-----------------------------------------------------------------------
! Write the PID to the file /tmp/QUANTICS_PID
!-----------------------------------------------------------------------     
      if(mpirank.eq.master)then
         filename='( umask 000 ; echo "$$" > /tmp/QUANTICS_PID )'
         call excmd(filename) 
      endif

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call mdefault

!-----------------------------------------------------------------------
! For MPI runs, only the master processor runs the setup steps
! put slaves to wait 
!-----------------------------------------------------------------------
#ifdef MPI
      if (mpirank .ne. master) then
         call mpislvwait
         go to 999
      endif
#endif

!-----------------------------------------------------------------------
! Get data from file ' .inp'
!-----------------------------------------------------------------------
      runlevel=0
      call eingabe

!-----------------------------------------------------------------------
! Build dvr file or simply read and log dvr
!-----------------------------------------------------------------------
      if (irundvr .eq. 1) then
         runlevel=1
         call rundvr
      else if (irundvr .eq. 2) then
         call logdvrdat
      endif

!-----------------------------------------------------------------------
! Build oper file 
!-----------------------------------------------------------------------
      if (irunoper .eq. 1) then
         runlevel=2
         call runoper
      else if (irunoper .eq. 2) then
         runlevel=3
         call logopdat
      endif

!-----------------------------------------------------------------------
! Generate Initial Wavefunction
!-----------------------------------------------------------------------
      if (iruninwf .eq. 1) then
         runlevel=3
         call runinwf
      else if (iruninwf .eq. 2) then
         call loginwfdat
      endif

#ifdef MPI
!-----------------------------------------------------------------------
! For MPI parallel runs, wake up slaves and bring to here
!-----------------------------------------------------------------------
      mpitask=0
      call mpi_bcast(mpitask,1,mpi_integer,master,mpi_comm_world,mpierr)
999 continue

!-----------------------------------------------------------------------
! Broadcast relevant logical flags to the slaves to allow access to
! propagation routines.
!-----------------------------------------------------------------------
      call mpi_bcast(irunprop,1,mpi_integer,master,mpi_comm_world,mpierr)
      call mpi_bcast(lrundiag,1,mpi_logical,master,mpi_comm_world,mpierr)
      call mpi_bcast(ltraj,1,mpi_logical,master,mpi_comm_world,mpierr)    
      call mpi_bcast(lrealwp,1,mpi_logical,master,mpi_comm_world,mpierr)    
      call mpi_bcast(proptask,1,mpi_integer,master,mpi_comm_world,mpierr)    
#endif

!-----------------------------------------------------------------------
! Propagate wavepacket
!-----------------------------------------------------------------------
      if(irunprop .eq. 1) then
         runlevel=4
         if (proptask .eq. 1 .or. proptask .eq. 4) then
            call runprop

         else if (proptask .eq. 2) then
            if(mpirank.eq.master) call runproprw

         else if (proptask .eq. 3) then
#ifdef MPI
!            call mpirunproptr
#else
            call runproptr
#endif
         else if (proptask .eq. 5) then
            call runprop_zagreb

         endif
      endif

!-----------------------------------------------------------------------
! finish DD calculations
!-----------------------------------------------------------------------
      if (ldd .or. lddmctdh) call dbfinalise

!-----------------------------------------------------------------------
! Get overall timing and write data to "log" file
!-----------------------------------------------------------------------
#ifdef MPI
      if(mpirank.ne.master) goto 998
#endif

      call cpu_time(tcpu)
      call system_clock(icount,count_rate)
      tsys = dble(icount)/dble(count_rate)
      tcpu = tcpu-tcpu0
      tsys = tsys-tsys0
      ttot = tsys
      t1   = int(ttot/3600.0_dop)
      ttot = ttot - dble(3600*t1)
      t2   = int(ttot/60.0_dop)
      ttot = ttot - dble(60*t2)

      write(ilog,*)
      write(ilog,'(a)') '***** Overall Timing *****'
      write(ilog,*)
      write(ilog,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
      write(ilog,'(a,1x,f12.2)') 'Total system time  [s] : ',tsys
      write(ilog, 900) 'Total time      [h:m:s] : ',t1,t2,ttot
      if(frunlevel .eq. 4 .and. proptask .ne. 5) then
         write(iout, 900) 'Total time     [h:m:s] : ',t1,t2,ttot
      endif
  900 format( a,i4,' :',i3,' :',f6.2)
      write(ilog,*)

      write(ilog,'(a)') '***** God''s final message *****'
      write(ilog,'(/,a,/)') 'Program terminated normally.'

     call myflush(ilog)

!-----------------------------------------------------------------------
! finally close files
!-----------------------------------------------------------------------
      call closedat
      
!-----------------------------------------------------------------------
! Finalise MPI. MPI slaves jump to here.
!-----------------------------------------------------------------------
#ifdef MPI
  998  continue
      call MPI_Finalize(mpierr)
#endif

   contains

!#######################################################################
! 
! MPISLVWAIT
!
! puts the MPI slaves in a wait state until given a task
!
!  MPITASKS
!
!  Task  Serial analogon    Description
!-----------------------------------------------------------------------
!  -1    --                 Stop slave
!   0    --                 Return
!-----------------------------------------------------------------------

#ifdef MPI
      subroutine mpislvwait

! --- Get Job
 100  call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)

      if (mpitask.eq.-1) then
         call MPI_Finalize(mpierr)
         stop

      else if (mpitask.eq.0) then
         return

      else
         routine="mpislvwait"
         write(message,'(a,i5)') "Unknown MPI task in Mpislvwait",mpitask
         call errormsg 

      endif

      go to 100

      end subroutine mpislvwait
#endif

      end program quantics


