! **********************************************************************
!                                                                       
!                   module CALCHA
!                                                                       
!   MCTDH module containing subroutines to evaluate the operation of
!   the Hamiltonian matrix on the MCTDH-coefficients
!                                                                       
!   Contains:                                                           
!     calcha:       Calculates the action of the correlated             
!                   Hamiltonian on the MCTDH-coefficients               
!     calcha1k:     Calculates the action of the correlated             
!                   Hamiltonian on the MCTDH-coefficients for one       
!                   coefficient of the inter-state terms.               
!     calcha2k:     Calculates the action of the correlated             
!                   Hamiltonian on the MCTDH-coefficients for one       
!                   coefficient of the intra-state terms.           
!     calcha3k:     Calculates the action of the correlated             
!                   Hamiltonian on the MCTDH-coefficients for one       
!                   coefficient of the multi-dimensional terms
!     calcha4k:     Calculates the action of the correlated             
!                   Hamiltonian on the MCTDH-coefficients for one       
!                   coefficient of the multi-dimensional terms using
!                   direct algorithm
!     hnatfast:     Calculates the action of natpot mean-fields on the 
!                   MCTDH coefficients with a more efficient algorithm.
!     calcha1kl:    Calculates the action of the correlated
!                   Hamiltonian on the MCTDH-coefficients for a single
!                   inter-state term, k escaping mode l.
!     calcha2kl:    The same as calcha1kl, but escaping modes l1 and l2.
!     calcha3kl:    The same as calcha1kl, but escaping modes l1,l2,l3.
!     calcha4kl:    The same as calcha1kl, but escaping modes
!                   l1,l2,l3,l4.
!     
!     calchuna:     Calculates the action of the uncorrelated             
!                   Hamiltonian on the MCTDH-coefficients               
!     calchuna_gwp: Calculates the action of the uncorrelated             
!                   Hamiltonian on the MCTDH-coefficients for only GWP modes
!     calchuna_nogwp: Calculates the action of the uncorrelated             
!                   Hamiltonian on the MCTDH-coefficients ignoring GWP modes
!
! **********************************************************************
      module calchamod

use decimal,  only: long, dop
use logdat
use xvlib,    only: vvtxzz, xvxczza, xvxxzzo, xvxxzza
use mtlib,    only: qtxxzzr, qtxxzza, qtxxzz, mtxxzz, mtaxzz, dtxxdzo,&
                    qtaxzz
use op1lib,   only: cpvxz, cpvxi, cpvxz_s, zerovxz
use op2lib,   only: addvxxzo

#ifdef MPI
use mpi
use mpidata
use mpiutilsmod, only: mpe_decomp1d
#endif
use global
use mltypes, only: hterm_tp
use dvrdatmod
use griddatmod
use operdef
use psidef
use hpsimod
use lctmod
use openmpmod
use ciselect, only: venpsi,venpsi1,venindex1
use potevalmod, only: poteval
use psiutil, only: phipnt
use linear, only: hprod_phi,hunprod_phi
use potevalmod
use runpropmod
use rdopermod

use timing
use channels

use op1lib, only: zerovxz,cpvxz,zerovxi
use op2lib, only: addvxxzo,swapsxxi
use xvlib, only: xvxxzza
use mmlib, only: ddxxdz

      implicit none
      private
      public :: calcha,calchuna,calcha1k,calchuna_gwp,calchuna_nogwp,&
                calcha2k,calcha_ml
#ifdef MPI
      public :: mpicalcha
#endif

      integer(long), dimension(:), allocatable      :: hklha,npklha

      contains

!#######################################################################
!
! Module CALCHA
!
! includes calcha, calcha1k, calcha2k, calcha3k, hnatfast
!
! *** V6.0 GW ************> SUBROUTINE CALCHA   <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for the intra- and inter-state terms.
! Result is returned as a sum over all terms in dtpsi.
!
! nham denotes which operator is to be used. 
! calctdk must be called before calcha is used to calculate time-dependent
! coefficients (required for multi-dimensional terms)
! NB the "fast" option for natpots applies only to SYSTEM Hamiltonian
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in calcha, please do them in dwtcalcha,
! memdwtcalcha and mpicalcha too! WHAT ARE THESE??
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!#######################################################################

      subroutine calcha (psi,dtpsi,hteil,hops,phi,nham,phi1, &
                 jindx,ort)

      implicit none

      integer(long)                                      :: k,s,s1,n,k0, &
                                                            koffset,n1,n2, &
                                                            ithr,h
      integer(long), intent(in)                          :: nham
      integer(long), dimension(jindxdim), intent(in)     :: jindx
      real(dop), dimension(hopsdim), intent(in)          :: hops
      real(dop), dimension(ortdim), intent(in)           :: ort
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(adim), intent(inout)       :: dtpsi
      complex(dop), dimension(:,:), allocatable          :: tmpdtpsi
      complex(dop), dimension(pmatdim), intent(inout)    :: hteil
      complex(dop), dimension(totphidim), intent(in)     :: phi,phi1
      integer(long), save                                :: tid=0 ! timer ID
      integer(long)                                 :: hl,ilbl,iord,nklha
      character(len=c2)                             :: label,nplabel

#ifdef OMP
      integer(long), external  :: omp_get_thread_num, omp_get_num_threads
#endif

      call get_timer(tid, 'Calcha')
      call start_timer(tid)

! --- if FAST is used
      n1=nrdf+1
      do n=1,nrdf
         if (lfast(n)) then
            koffset=nterm(n)-1
            n1=n
            go to 5
         endif
      enddo

      koffset=zham(nham)+kzahl(nham)-1
!  --- if koffset is zero (uncorrelated Hamiltonian) skip to end
      if (koffset .eq. 0) goto 15 

 5    k0=zham(nham)

!-----------------------------------------------------------------------
! Initialize and calculate Hamiltonian moments for correlated LHA
! 1. Get how many potentials to be treated (nklha)
! 2. Allocate arrays
! 3. Load term no. defining LHA potential into hklha
! 4. Calculate potentials for each configuration
!-----------------------------------------------------------------------
      if (lklha) then
         nklha=0
         do k=k0,koffset
            h=hamilton(fkoe,k)
            hl = hterm(1,h)
            label = hoplab(hl)

            ilbl = index(label,'%')+1
            read(label(ilbl:ilbl),*) iord
            if (iord .eq. 0) nklha = nklha+1
         enddo

         allocate(hklha(nklha),npklha(ktot))
         call alloc_klha(nklha)

         n=0
         npklha(:)=0
         do k=k0,koffset
            h=hamilton(fkoe,k)
            hl = hterm(1,h)
            label = hoplab(hl)
            
            ilbl = index(label,'%')+1
            read(label(ilbl:ilbl),*) iord
            if (iord .eq. 0) then
               n=n+1
               hklha(n) = h
               nplabel = label(1:ilbl-1)
               npklha(k) = n
            else if (label(1:ilbl-1) .eq. nplabel(1:ilbl-1)) then
               npklha(k) = n
            endif
         enddo

         do n=1,nklha
            h=hklha(n)
            call klhaconfig(h,n,phi(1),adim+1)
         enddo
      endif

 10   continue

#ifdef MPI
      if(lmpi.and.lmpicalcha.and.lmpistarted) then ! MPI if 1
         if (.not. lmpidavrun) then
            mpitask=2
! --- Determine Job
            call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(lexpectval,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)

            call MPI_Bcast(nham,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         endif

         call mpicalcha(psi,phi,dtpsi,hteil,nham,koffset,k0,jindx)
      else
#endif
!-----------------------------------------------------------------------
!     If no MPI parallelisation is used then an OMP or serial 
!     calculation is performed.
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! allocate memory for omp threads
!-----------------------------------------------------------------------
#ifdef OMP
         if (lompthread) then
            allocate(tmpdtpsi(adim,ompthread-1))
            tmpdtpsi=0.0_dop
         endif
#endif

! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

!$omp parallel num_threads(ompthread) private(k,s,s1,ithr) if(lompthread .and. lompcalcha)
!$omp do schedule(dynamic)
         do k=k0,koffset
! Ignored terms in KLHA
            if (klha(k) .lt. 0) cycle
#ifdef OMP
            ithr = omp_get_thread_num()
#else
            ithr = 0
#endif
            s = kf(k)
            s1= ki(k)
            if (s .eq. s1 .and. ithr .eq. 0) then
               call calcha1k(psi,phi,hteil,dtpsi,k,s,jindx)
            else if (s .eq. s1) then
               call calcha1k(psi,phi,hteil,tmpdtpsi(1,ithr),k,s,jindx)
            else if (s .ne. s1 .and. ithr .eq. 0) then
               call calcha2k(psi,phi,hteil,dtpsi,k,s,s1,jindx)
            else 
               call calcha2k(psi,phi,hteil,tmpdtpsi(1,ithr),k,s,s1,jindx)
            endif

         enddo
!$omp enddo
!$omp end parallel 

#ifdef OMP
         if (lompthread) then
            do ithr = 1,ompthread-1
               call addvxxzo(tmpdtpsi(1,ithr),dtpsi,adim)
            enddo
         endif
#endif

#ifdef MPI
      endif ! MPI endif 1

      if (mpirank.eq.0) then ! MPI if 2

!  ---  if FAST is used

!      if (lmpi.and.lmpicalcha) s = kf(koffset)
#endif
      s = kf(koffset)
      do n=n1,nrdf 
         call hnatfast(psi,dtpsi,hteil,fastprem(zfp(n):), &
                       fastnum(zfn(n)),fastdim(zfd(n)),s,n)
         if (ntermlast(n).eq.zham(nham)+kzahl(nham)+nmulpot(nham)-1) goto 15
         if (nterm(n+1).gt.ntermlast(n)+1 .or. .not.lfast(n+1)) then
            k0=ntermlast(n)+1
            do n2=n+1,nrdf
               if (nterm(n2).eq.ntermlast(n2-1)+1 .and. lfast(n2)) then
                  koffset=nterm(n2)-1
                  n1=n2
                  go to 10
               elseif (n2.eq.nrdf) then
                  koffset=zham(nham)+kzahl(nham)+nmulpot(nham)-1
                  n1=nrdf+1
                  go to 10
               endif
            enddo
         endif
      enddo

 15   continue

!-----------------------------------------------------------------------
! evaluate any multi-dimensional terms
!-----------------------------------------------------------------------
      do n=1,nmulpot(nham)
         k=zham(nham)+kzahl(nham)+n-1
         s=kf(k)
         s1=ki(k)
         if (s.eq.s1 .and. lcdvr.or.ltdvr) then
            call stopnow
!           call calcha5k(psi,cdvr3,cdvr4,k,s)
         else if (ldirmuld) then
            call calcha4k(psi,phi,ort,dtpsi(zpsi(s)),k,s,s1,phi1)
         else
            call calcha3k(psi,phi,hops,dtpsi(zpsi(s)),k,s,s1,phi1,hteil)
         endif

      enddo

#ifdef MPI
      endif    ! MPI endif 2
#endif

      if (lklha) then
         call dealloc_klha
         deallocate(hklha,npklha)
      endif

      call stop_timer(tid)

      return
      end subroutine calcha

! *** V6.0 GW ************> SUBROUTINE CALCHA1K <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single intra-state term, k.
! Result is returned in dtpsi

      subroutine calcha1k (psi,phi,hteil,dtpsi,k,s,jindx)

      implicit none

      integer(long)                                      :: zeig1,zeig2,&
                                                            zeig3,m,block1,&
                                                            block2,h
      integer(long), intent(in)                          :: k,s
      integer(long), dimension(jindxdim), intent(in)     :: jindx
      integer(long), dimension(2*fmblock)                :: worki
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(adim), intent(out)         :: dtpsi
      complex(dop), dimension(2*fmblock)                 :: workc
      complex(dop), dimension(pmatdim), intent(in)       :: hteil
      complex(dop), dimension(adim+1:dgldim), intent(in) :: phi
      complex(dop), dimension(:), allocatable            :: psig
      complex(dop)                                       :: ztmp
      real(dop), dimension(:), allocatable               :: koelha

      if (klha(k) .gt. 0) then
         allocate(koelha(maxblock))
         allocate(psig(maxblock))
      endif

! --- COPY PSI VECTOR AND SET UP POINTERS ---

      zeig2 = fmblock+1             !     maxblock+1
      zeig1 = 1
      call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))

!-----------------------------------------------------------------------
!  calculate coefficients if correlated LHA
! ?Could calculate koelha and save for use after a CMF step?
!-----------------------------------------------------------------------
      if (klha(k) .gt. 0) then
         h=hamilton(fkoe,k)
         call klhacalc(koelha,s,h,npklha(k))

! transform to GWP basis
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call swapsxxi(zeig1,zeig2)
            call qtxxzz(phi(zdgwp(m,s)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         enddo
    
! multiply by LHA coefficients
         call swapsxxi(zeig1,zeig2)
         call ddxxdz(koelha,workc(zeig1),workc(zeig2),block(s))

! save transformed psi
         call cpvxz(workc(zeig2),psig,block(s))
      endif

! --- LOOP OVER EACH MODE ---
      if (citype .gt. 0) then
         block2=block(s)
         call cpvxi(jindx(zpsi(s)),worki(zeig2),block2)

         do m=1,nmode
            if (.not. diag(m,k)) then
               block1=vdim(m,s)*dim(m,s)*ndim(m,s)
               call venpsi(m,jindx(zpsi(s)),worki(zeig1), &
                    worki(zeig2),block(s),block1,block2, &
                    nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                    workc(zeig1),workc(zeig2))
               block2=block1
               call cpvxi(worki(zeig1),worki(zeig2),block1)
               call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo

         zeig3=zeig1
         zeig1=zeig2
         zeig2=zeig3

         block1=block2
         block2=block(s)
         call cpvxz_s(workc(zeig1),workc(zeig2),worki(zeig1), &
                 jindx(zpsi(s)),block1,block2)

      else
         do m = 1,nmode
            if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

               zeig3 = zeig1
               zeig1 = zeig2
               zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

               call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo
      endif

!-----------------------------------------------------------------------
!  If correlated LHA transform back to SPF basis
!-----------------------------------------------------------------------
      if (klha(k) .gt. 0) then
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call swapsxxi(zeig1,zeig2)
            call qtaxzz(phi(zdgwp(m,s)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         enddo
      endif

!    --- STORE RIGHT HAND SIDE IN DTPSI ---

      h=hamilton(fkoe,k)
      if (htmtype(h) .eq. 1) then
         call addvxxzo(workc(zeig2),dtpsi(zpsi(s)),block(s))
      else if (htmtype(h) .eq. 104) then
         ztmp = 0.5_dop
         call xvxxzza(ztmp,workc(zeig2),dtpsi(zpsi(s)),block(s))
      else
         call xvxxzza(koeff(k),workc(zeig2),dtpsi(zpsi(s)),block(s))
      endif

!-----------------------------------------------------------------------
! If correlated LHA, add symmetric part to HA
!-----------------------------------------------------------------------
      if (klha(k) .gt. 0) then
         zeig2 = fmblock+1             !     maxblock+1
         zeig1 = 1
         call cpvxz(psig,workc(zeig2),block(s))

         do m = 1,nmode
            if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

               zeig3 = zeig1
               zeig1 = zeig2
               zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

               call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
           endif
         enddo

!    --- STORE RIGHT HAND SIDE IN DTPSI ---

         call swapsxxi(zeig1,zeig2)
         call ddxxdz(koelha,workc(zeig1),workc(zeig2),block(s))

!-----------------------------------------------------------------------
! and transform back to GWP basis
!-----------------------------------------------------------------------
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call swapsxxi(zeig1,zeig2)
            call qtaxzz(phi(zdgwp(m,s)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         enddo

         ztmp = 0.5_dop
         call xvxxzza(ztmp,workc(zeig2),dtpsi(zpsi(s)),block(s))
      endif

      if (allocated(koelha)) deallocate(koelha)
      if (allocated(psig)) deallocate(psig)

      return
      end subroutine calcha1k


! *********************> SUBROUTINE CALCHA1KL <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k escaping mode l.
!
! Note: Not valid for S-MCTDH.

      subroutine calcha1kl (psi,hteil,dtpsi,k,s,l)

      implicit none

      integer(long)                                    :: zeig1,zeig2,zeig3,m
      integer(long), intent(in)                        :: k,s,l
      complex(dop), dimension(maxblock), intent(in)    :: psi
      complex(dop), dimension(maxblock), intent(out)   :: dtpsi
      complex(dop), dimension(2*maxblock)              :: workc
      complex(dop), dimension(pmatdim), intent(in)     :: hteil


! --- COPY PSI VECTOR AND SET UP POINTERS ---

      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi,workc(zeig2),block(s))


! --- LOOP OVER EACH MODE ---

      do m = 1,nmode
        if (m .eq. l) go to 10

        if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

  !          if(.not.lompthread)then
               call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
  !          else
  !             call qtxxzzomp(hteil(pmat(m,k)),workc(zeig1), &
  !                  workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s),ompthread)
  !          endif
        endif
 10   continue
      enddo

! --- MOVE RESULT TO OUTPUT DTPSI ARRAY ---

      call cpvxz(workc(zeig2),dtpsi,block(s))

      return
      end subroutine calcha1kl


! *********************> SUBROUTINE CALCHA2KL <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k escaping modes 
! l1 and l2.
!
! Note: Not vaild for S-MCTDH.
 
      subroutine calcha2kl (psi,hteil,dtpsi,k,s,l1,l2)

      implicit none
 
      integer(long)                                    :: zeig1,zeig2,zeig3,m
      integer(long), intent(in)                        :: k,s,l1,l2
      complex(dop), dimension(maxblock), intent(in)    :: psi
      complex(dop), dimension(maxblock), intent(out)   :: dtpsi
      complex(dop), dimension(2*maxblock)              :: workc
      complex(dop), dimension(pmatdim), intent(in)     :: hteil

 
! --- COPY PSI VECTOR AND SET UP POINTERS ---
 
      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi,workc(zeig2),block(s))
 
 
! --- LOOP OVER EACH MODE ---
 
      do m = 1,nmode
        if (m.eq.l1 .or. m.eq.l2) go to 10
 
        if (.not. diag(m,k)) then
 
!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---
 
            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3
 
!       --- MULTIPLY WITH MATRIX ELEMENTS ---

    !        if(.not.lompthread)then
               call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
    !        else
    !           call qtxxzzomp(hteil(pmat(m,k)),workc(zeig1), &
    !                workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s),ompthread)
    !        endif
         endif
 10   continue
      enddo
 
! --- MOVE RESULT TO OUTPUT DTPSI ARRAY ---

      call cpvxz(workc(zeig2),dtpsi,block(s))
 
      return
      end subroutine calcha2kl


! *********************> SUBROUTINE CALCHA3KL <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k escaping modes 
! l1, l2 and l3.
!
! Note: Not vaild for S-MCTDH.

      subroutine calcha3kl (psi,hteil,dtpsi,k,s,l1,l2,l3)

      implicit none

      integer(long)                                    :: zeig1,zeig2,zeig3,m
      integer(long), intent(in)                        :: k,s,l1,l2,l3
      complex(dop), dimension(maxblock), intent(in)    :: psi
      complex(dop), dimension(maxblock), intent(out)   :: dtpsi
      complex(dop), dimension(2*maxblock)              :: workc
      complex(dop), dimension(pmatdim), intent(in)     :: hteil


! --- COPY PSI VECTOR AND SET UP POINTERS ---

      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi,workc(zeig2),block(s))


! --- LOOP OVER EACH MODE ---

      do m = 1,nmode
        if (m.eq.l1 .or. m.eq.l2 .or. m.eq.l3) go to 10

        if (.not. diag(m,k)) then
 
!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3
 
!       --- MULTIPLY WITH MATRIX ELEMENTS ---

            call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
        endif
 10   continue
      enddo
 
! --- MOVE RESULT TO OUTPUT DTPSI ARRAY ---

      call cpvxz(workc(zeig2),dtpsi,block(s))

      return
      end subroutine calcha3kl


! *********************> SUBROUTINE CALCHA4KL <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k escaping modes 
! l1, l2 and l3.
!
! Note: Not vaild for S-MCTDH.

      subroutine calcha4kl (psi,hteil,dtpsi,k,s,l1,l2,l3,l4)

      implicit none

      integer(long)                                    :: zeig1,zeig2,zeig3,m
      integer(long), intent(in)                        :: k,s,l1,l2,l3,l4
      complex(dop), dimension(maxblock), intent(in)    :: psi
      complex(dop), dimension(maxblock), intent(out)   :: dtpsi
      complex(dop), dimension(2*maxblock)              :: workc
      complex(dop), dimension(pmatdim), intent(in)     :: hteil


! --- COPY PSI VECTOR AND SET UP POINTERS ---

      zeig1 = block(s)+1
      zeig2 = 1
      call cpvxz(psi,workc(zeig2),block(s))


! --- LOOP OVER EACH MODE ---

      do m = 1,nmode
        if (m.eq.l1 .or. m.eq.l2 .or. m.eq.l3 .or. m.eq.l4) go to 10

        if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

            call qtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         endif
 10   continue
      enddo

! --- MOVE RESULT TO OUTPUT DTPSI ARRAY ---

      call cpvxz(workc(zeig2),dtpsi,block(s))

      return
      end subroutine calcha4kl



! *** V6.0 GW ************> SUBROUTINE CALCHA2K <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for a single inter-state term, k.
! Result is returned in dtpsi

      subroutine calcha2k(psi,phi,hteil,dtpsi,k,s,s1,jindx)


      implicit none

      integer(long)                                   :: zeig1,zeig2,zeig3,m,&
                                                         vdim1,block1,block2,&
                                                         h
      integer(long), intent(in)                       :: k,s,s1
      integer(long), dimension(jindxdim), intent(in)  :: jindx
      integer(long), dimension(2*maxblock)            :: worki
      complex(dop), dimension(adim), intent(in)       :: psi
      complex(dop), dimension(adim), intent(out)      :: dtpsi
      complex(dop), dimension(2*maxblock)             :: workc
      complex(dop), dimension(pmatdim), intent(in)    :: hteil
      complex(dop), dimension(adim+1:dgldim), intent(in) :: phi
      real(dop), dimension(:), allocatable            :: koelha

      if (klha(k) .gt. 0) allocate(koelha(maxblock))

! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

      zeig2 = fmblock+1     !         maxblock+1
      zeig1 = 1

! --- INITIALIZE WORK ARRAY WITH PSI ARRAY ---

      call cpvxz(psi(zpsi(s1)),workc(zeig2),block(s1))

!-----------------------------------------------------------------------
!  calculate coefficients if correlated LHA
! ?Could calculate koelha and save for use ofer a CMD step?
!-----------------------------------------------------------------------
      if (klha(k) .gt. 0) then
         h=hamilton(fkoe,k)
         call klhacalc(koelha,s1,h,npklha(k))

! transform to GWP basis
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call swapsxxi(zeig1,zeig2)
            call qtxxzz(phi(zdgwp(m,s1)),workc(zeig1), &
                 workc(zeig2),vdim(m,s1),dim(m,s1),ndim(m,s1))
         enddo

! multiply by LHA coefficients
         call swapsxxi(zeig1,zeig2)
         call ddxxdz(koelha,workc(zeig1),workc(zeig2),block(s1))
      endif

! --- LOOP OVER EACH MODE ---

      if (citype .gt. 0) then
         block2=block(s1)
         call cpvxi(jindx(zpsi(s1)),worki(zeig2),block2)

         do m=1,nmode
            block1=vdim(m,s)*dim(m,s1)*ndim(m,s1)
            call venpsi1(m,jindx(zpsi(s)),jindx(zpsi(s1)), &
                    worki(zeig1), &
                    worki(zeig2),block(s),block(s1),block1,block2, &
                    nmode,vdim(m,s),dim(m,s1),ndim(m,s1),jvdim, &
                    workc(zeig1),workc(zeig2))
            block2=vdim(m,s)*dim(m,s)*ndim(m,s1)
            call venindex1(m,jindx(zpsi(s)),jindx(zpsi(s1)), &
                    worki(zeig2),block(s),block(s1),block2, &
                    nmode,vdim(m,s),dim(m,s),ndim(m,s1),jvdim)

     !       if(.not.lompthread)then
               call mtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s1),ndim(m,s1), &
                    dim(m,s))
     !       else
     !          call mtxxzzomp(hteil(pmat(m,k)),workc(zeig1), &
     !               workc(zeig2),vdim(m,s),dim(m,s1),ndim(m,s1), &
     !               dim(m,s),ompthread)
     !       endif
         enddo
         zeig3=zeig1
         zeig1=zeig2
         zeig2=zeig3

         block1=block2
         block2=block(s)
         call cpvxz_s(workc(zeig1),workc(zeig2),worki(zeig1), &
                 jindx(zpsi(s)),block1,block2)

      else
         vdim1 = vdim(1,s1)
         do m = 1,nmode

!       --- SWAP POINTERS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

            call zerovxz(workc(zeig2),maxblock)
            call mtxxzz(hteil(pmat(m,k)),workc(zeig1), &
                 workc(zeig2),vdim1,dim(m,s1),ndim(m,s1),dim(m,s))
            vdim1 = vdim1*dim(m,s)
         enddo

      endif

!-----------------------------------------------------------------------
!  If correlated LHA transform back to SPF basis
!-----------------------------------------------------------------------
      if (klha(k) .gt. 0) then
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call swapsxxi(zeig1,zeig2)
            call qtaxzz(phi(zdgwp(m,s)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
         enddo
      endif

!    --- STORE RIGHT HAND SIDE IN DTPSI ---

      h=hamilton(fkoe,k)
      if (htmtype(h) .eq. 1 .or. htmtype(h) .eq. 104) then
         call addvxxzo(workc(zeig2),dtpsi(zpsi(s)),block(s))
      else
         call xvxxzza(koeff(k),workc(zeig2),dtpsi(zpsi(s)),block(s))
      endif

      if (allocated(koelha)) deallocate(koelha)

      return
      end subroutine calcha2k

! *** V6.0 GW ************> SUBROUTINE CALCHA3K <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of multi-dimensional terms of the correlated
! Hamiltonian on the C MCTDH-coefficients for a single inter-state
! term, k.
! Result is returned in dtpsi
! NB phi (containing spfs) is required separately as psi may not
! contain full wavefunction (e.g. lanczos vectors)
!
! < phi1 | H | phi > A
!
!#######################################################################

      subroutine calcha3k (psi,phi,hops,dtpsi,k,s,s1,phi1,hteil)

      implicit none

      integer(long)                                      :: zeig1,zeig2,zeig3,m,muld,&
                                                            vdim1,ndim1,block1,h,f
      integer(long), intent(in)                          :: k,s,s1
      real(dop), dimension(hopsdim), intent(in)          :: hops
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(pmatdim), intent(in)       :: hteil
      complex(dop), dimension(maxblock), intent(out)     :: dtpsi
      complex(dop), dimension(2*maxblock)                :: workc
      complex(dop), dimension(adim+1:dgldim), intent(in) :: phi,phi1


! --- COPY PSI VECTOR AND SET UP POINTERS ---

      zeig1 = maxblock+1
      zeig2 = 1
      call cpvxz(psi(zpsi(s1)),workc(zeig2),block(s1))

!-----------------------------------------------------------------------
! Operate on coeffs by electronic operator (single set)
!-----------------------------------------------------------------------
      if (leb .and. meb .gt. 0) then
         zeig3=zeig1
         zeig1=zeig2
         zeig2=zeig3
         call mtxxzz(hteil(pmat(meb,k)),workc(zeig1), &
              workc(zeig2),vdim(meb,s1),dim(meb,s1),ndim(meb,s1), &
              dim(meb,s))
      endif

!-----------------------------------------------------------------------
!  now transform from spf basis to grid point basis
!  block1 is the number of elements in the transformed array.
!  vdim1 is equivalent to vdim, but takes into account growth of block1.
!  before multiplication, block1=vdim1*dim(m,s1)*ndim(m,s1)
!  after multiplication,  block1 has increased to
!  vdim1*subdim(m)*ndim(m,s1)
!  and vdim1 changes to vdim1*subdim(m) for the next mode.
!-----------------------------------------------------------------------
      f=fterm(k)
      h=hamilton(f,k)
      muld=htmmuld(h)
      vdim1=1
      block1=0

      do m = 1,nmode
         if (.not. diag(m,k)) then   ! Does not work for nodiag
         
!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

    !        if(.not.lompthread)then
               call mtxxzz(phi(zetf(m,s1)),workc(zeig1), &
                    workc(zeig2),vdim1,dim(m,s1),ndim(m,s1),subdim(m))
    !        else
    !           call mtxxzzomp(phi(zetf(m,s1)),workc(zeig1), &
    !                workc(zeig2),vdim1,dim(m,s1),ndim(m,s1),subdim(m),ompthread)
    !        endif

            block1=vdim1*subdim(m)*ndim(m,s1)
            vdim1=vdim1*subdim(m)
        else
            block1=vdim1*dim(m,s1)*ndim(m,s1)
            vdim1=vdim1*dim(m,s1)
        endif
      enddo

!-----------------------------------------------------------------------
! now operate on coeffs by V(q1,....qf)
!-----------------------------------------------------------------------
!      if(lompthread)then
!         call dtxxdzoomp(hops(zhop(h)),workc(zeig2),vdmuld(muld), &
!              muldsize(muld),ndmuld(muld),ompthread)
!      else
         call dtxxdzo(hops(zhop(h)),workc(zeig2),vdmuld(muld), &
              muldsize(muld),ndmuld(muld))
!      endif
!-----------------------------------------------------------------------
!  Transform back from grid points to spfs.
!  block1 now contracts back to block(s).
!  ndim1 is equivalent to ndim, but takes into account contraction of
!  block1.
!  before multiplication, block1=vdim(m,s)*subdim(m)*ndim1
!  after multiplication,  block1 has decreased to
!  vdim(m,s)*dim(m,s)*ndim1
!-----------------------------------------------------------------------
      do m=1,nmode
         if (.not. diag(m,k)) then   ! Does not work for nodiag
            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3

            ndim1=block1/(vdim(m,s)*subdim(m))
            call mtaxzz(phi1(zetf(m,s)),workc(zeig1), &
                 workc(zeig2),vdim(m,s),subdim(m),ndim1,dim(m,s))
            block1=vdim(m,s)*dim(m,s)*ndim1
         else
            ndim1=block1/(vdim(m,s)*dim(m,s))
            block1=vdim(m,s)*dim(m,s)*ndim1
         endif
      enddo

! --- STORE RIGHT HAND SIDE IN DTPSI ---

      h=hamilton(fkoe,k)
      if (htmtype(h) .eq. 1) then
         call addvxxzo(workc(zeig2),dtpsi(zpsi(s)),block(s))
      else
         call xvxxzza(koeff(k),workc(zeig2),dtpsi(zpsi(s)),block(s))
      endif

      return
      end subroutine calcha3k

! *** V8.3 GW ************> SUBROUTINE CALCHA4K <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of multi-dimensional terms of the correlated
! Hamiltonian on the C MCTDH-coefficients for a single inter-state
! term, k. Direct algorithm used so that potential on full grid does
! not need to be held in memory.  
! Result is returned in dtpsi
! NB phi (containing spfs) is required separately as psi may not
! contain full wavefunction (e.g. lanczos vectors)
!
! < phi1 | H | phi > A
! 
!#######################################################################

      subroutine calcha4k (psi,phi,ort,dtpsi,k,s,s1,phi1)

      implicit none

      integer(long)                                      :: zeig,zeig1,m,h,muld,mgpnt,gpnt,&
                                                            p,potdim,n,f,idum,zeig2
      integer(long), intent(in)                          :: k,s,s1
      integer(long), dimension(maxdim)                   :: igpoint
      real(dop)                                          :: v
      real(dop), dimension(maxdim)                       :: rdum
      real(dop), dimension(maxdim)                       :: gpoint
      real(dop), dimension(ortdim), intent(in)           :: ort
      complex(dop)                                       :: psix,vz
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(maxblock), intent(out)     :: dtpsi
      complex(dop), dimension(3*maxblock)                :: workc
      complex(dop), dimension(adim+1:dgldim), intent(in) :: phi,phi1

      f=fterm(k)
      h=hamilton(f,k)
      muld=htmmuld(h)
      potdim=muldsize(muld)

      zeig=1
      zeig1=zeig+block(s)
      zeig2=zeig1+block(s)
      workc = 0.0_dop

      do p=1,potdim
! get coordinates of grid point
         gpnt=p-1
         do m=1,nmode
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               mgpnt=gpnt
               gpnt=gpnt/gdim(f)
               mgpnt=mgpnt-gpnt*gdim(f)
               gpoint(f)=ort(zort(f)+mgpnt)
            enddo
         enddo
! get grid point for each mode
         gpnt=p-1
         do m=1,nmode
            mgpnt=gpnt
            gpnt=gpnt/subdim(m)
            mgpnt=mgpnt-gpnt*subdim(m)
            igpoint(m)=mgpnt+1
         enddo

         call poteval (h,v,gpoint,vz,rdum,idum)
         if (lcutpes) then
            v=min(v,pesmax)
            v=max(v,pesmin)
         endif

         call phipnt(phi,workc(zeig1),igpoint,s1,block(s1))
         call vvtxzz(workc(zeig1),psi(zpsi(s1)),psix,block(s1))

         psix=psix*v

         call phipnt(phi1,workc(zeig),igpoint,s,block(s))
         call xvxczza(psix,workc(zeig),workc(zeig2),block(s))
      enddo

! --- STORE RIGHT HAND SIDE IN DTPSI ---

      h=hamilton(fkoe,k)
      if (htmtype(h) .eq. 1) then
         call addvxxzo(workc(zeig2),dtpsi(zpsi(s)),block(s))
      else
         call xvxxzza(koeff(k),workc(zeig2),dtpsi(zpsi(s)),block(s))
      endif

      return
      end subroutine calcha4k

! ************************* SUBROUTINE HNATFAST *************************
!
! Calculates the action of natpot (n-th) mean-fields on the MCTDH
! coefficients with an efficient algorithm.
! SS
! ***********************************************************************

      subroutine hnatfast(psi,dtpsi,hteil,fastprem1,fastnum1,fastdim1,s,n)

      implicit none

      integer(long)       :: ip,m1,m2,m3,m4,k,k1,k2,k3,k4,h,z1,z2,z3,z4,z5,&
                             j,j1,j2,j3,j4,i,ii,jj2,jj3,jj4,d2,natdim

      integer(long), intent(in)                      :: s,n
      integer(long), dimension(*), intent(in)        :: fastprem1,fastnum1, &
                                                        fastdim1
      complex(dop), dimension(adim), intent(in)      :: psi
      complex(dop), dimension(adim), intent(inout)   :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)   :: hteil
      complex(dop), dimension(5*maxblock)            :: workc


      z1=1
      z2=z1+maxblock
      z3=z2+maxblock
      z4=z3+maxblock
      z5=z4+maxblock

      natdim=mdim(mmodc(n),n)
      d2=natdim
      do i=1,fastorder(n)
         d2=d2/mdimf(natpmode(i,n),n)
      enddo
 
      if (fastorder(n).eq.1) then
! --------- fast - first order

         m1=natpmode(1,n)

         do ip=1,npacket
         do j1=1,mdimf(m1,n)
            k1=fastprem1(j1) + ip-1
            if (.not.diag(m1,k1)) then
               call qtxxzz(hteil(pmat(m1,k1)),psi(zpsi(s)),workc(z1), &
                           vdim(m1,s),dim(m1,s),ndim(m1,s))
            else
               call cpvxz(psi(zpsi(s)),workc(z1),block(s))
            endif
            j=j1
            do i=1,fastdim1(j)
               ii=i+(j-1)*d2
               k=fastnum1(ii) + ip-1
               call calcha1kl(workc(z1),hteil,workc(z2),k,s,m1)
               h=hamilton(fkoe,k)
               if (htmtype(h) .eq. 1) then
                   call addvxxzo(workc(z2),dtpsi(zpsi(s)),block(s))
               else
                   call xvxxzza(koeff(k),workc(z2),dtpsi(zpsi(s)), &
                                block(s))
               endif
            enddo
         enddo
         enddo

      elseif (fastorder(n).eq.2) then
! --------- fast - 2-nd order

         m1=natpmode(1,n)
         m2=natpmode(2,n)

         do ip=1,npacket
         do j1=1,mdimf(m1,n)
            k1=fastprem1(j1) + ip-1
            if (.not.diag(m1,k1)) then
               call qtxxzz(hteil(pmat(m1,k1)),psi(zpsi(s)),workc(z1), &
                           vdim(m1,s),dim(m1,s),ndim(m1,s))
            else
               call cpvxz(psi(zpsi(s)),workc(z1),block(s))
            endif
            do j2=1,mdimf(m2,n)
               jj2=j2+mdimf(m1,n)
               k2=fastprem1(jj2) + ip-1
               if (.not.diag(m2,k2)) then
                  call qtxxzz(hteil(pmat(m2,k2)),workc(z1),workc(z2), &
                              vdim(m2,s),dim(m2,s),ndim(m2,s))
               else
                  call cpvxz(workc(z1),workc(z2),block(s))
               endif
               j=(j2-1)*mdimf(m1,n)+j1
               do i=1,fastdim1(j)
                  ii=i+(j-1)*d2
                  k=fastnum1(ii) + ip-1
                  call calcha2kl(workc(z2),hteil,workc(z3),k,s,m1,m2)
                  h=hamilton(fkoe,k)
                  if (htmtype(h) .eq. 1) then
                      call addvxxzo(workc(z3),dtpsi(zpsi(s)),block(s))
                  else
                      call xvxxzza(koeff(k),workc(z3),dtpsi(zpsi(s)), &
                                   block(s))
                  endif
               enddo
            enddo
         enddo
         enddo

      elseif (fastorder(n).eq.3) then
! --------- fast - 3-rd order

         m1=natpmode(1,n)
         m2=natpmode(2,n)
         m3=natpmode(3,n)

         do ip=1,npacket
         do j1=1,mdimf(m1,n)
            k1=fastprem1(j1) + ip-1
            if (.not.diag(m1,k1)) then
               call qtxxzz(hteil(pmat(m1,k1)),psi(zpsi(s)),workc(z1), &
                           vdim(m1,s),dim(m1,s),ndim(m1,s))
            else
               call cpvxz(psi(zpsi(s)),workc(z1),block(s))
            endif
            do j2=1,mdimf(m2,n)
               jj2=j2+mdimf(m1,n)
               k2=fastprem1(jj2) + ip-1
               if (.not.diag(m2,k2)) then
                  call qtxxzz(hteil(pmat(m2,k2)),workc(z1),workc(z2), &
                              vdim(m2,s),dim(m2,s),ndim(m2,s))
               else
                  call cpvxz(workc(z1),workc(z2),block(s))
               endif
               do j3=1,mdimf(m3,n)
                  jj3=j3+mdimf(m1,n)+mdimf(m2,n)
                  k3=fastprem1(jj3) + ip-1
                  if (.not.diag(m3,k3)) then
                     call qtxxzz(hteil(pmat(m3,k3)),workc(z2),workc(z3), &
                                 vdim(m3,s),dim(m3,s),ndim(m3,s))
                  else
                     call cpvxz(workc(z2),workc(z3),block(s))
                  endif
                  j=(j3-1)*mdimf(m1,n)*mdimf(m2,n)+ &
                    (j2-1)*mdimf(m1,n)+j1
                  do i=1,fastdim1(j)
                     ii=i+(j-1)*d2
                     k=fastnum1(ii) + ip-1
                     call calcha3kl(workc(z3),hteil,workc(z4),k,s, &
                                    m1,m2,m3)
                     h=hamilton(fkoe,k)
                     if (htmtype(h) .eq. 1) then
                        call addvxxzo(workc(z4),dtpsi(zpsi(s)),block(s))
                     else
                        call xvxxzza(koeff(k),workc(z4),dtpsi(zpsi(s)), &
                                     block(s))
                     endif
                  enddo
               enddo
            enddo
         enddo
         enddo

      elseif (fastorder(n).eq.4) then
! --------- fast - 4-th order

         m1=natpmode(1,n)
         m2=natpmode(2,n)
         m3=natpmode(3,n)
         m4=natpmode(4,n)

         do ip=1,npacket
         do j1=1,mdimf(m1,n)
            k1=fastprem1(j1) + ip-1
            if (.not.diag(m1,k1)) then
               call qtxxzz(hteil(pmat(m1,k1)),psi(zpsi(s)),workc(z1), &
                           vdim(m1,s),dim(m1,s),ndim(m1,s))
            else
               call cpvxz(psi(zpsi(s)),workc(z1),block(s))
            endif
            do j2=1,mdimf(m2,n)
               jj2=j2+mdimf(m1,n)
               k2=fastprem1(jj2) + ip-1
               if (.not.diag(m2,k2)) then
                  call qtxxzz(hteil(pmat(m2,k2)),workc(z1),workc(z2), &
                              vdim(m2,s),dim(m2,s),ndim(m2,s))
               else
                  call cpvxz(workc(z1),workc(z2),block(s))
               endif
               do j3=1,mdimf(m3,n)
                  jj3=j3+mdimf(m1,n)+mdimf(m2,n)
                  k3=fastprem1(jj3) + ip-1
                  if (.not.diag(m3,k3)) then
                     call qtxxzz(hteil(pmat(m3,k3)),workc(z2),workc(z3), &
                                 vdim(m3,s),dim(m3,s),ndim(m3,s))
                  else
                     call cpvxz(workc(z2),workc(z3),block(s))
                  endif
                  do j4=1,mdimf(m4,n)
                     jj4=j4+mdimf(m1,n)+mdimf(m2,n)+mdimf(m3,n)
                     k4=fastprem1(jj4) + ip-1
                     if (.not.diag(m4,k4)) then
                        call qtxxzz(hteil(pmat(m4,k4)),workc(z3), &
                                    workc(z4),vdim(m4,s),dim(m4,s), &
                                    ndim(m4,s))
                     else
                        call cpvxz(workc(z3),workc(z4),block(s))
                     endif
                     j=(j4-1)*mdimf(m1,n)*mdimf(m2,n)*mdimf(m3,n)+ &
                       (j3-1)*mdimf(m1,n)*mdimf(m2,n)+ &
                       (j2-1)*mdimf(m1,n)+j1
                     do i=1,fastdim1(j)
                        ii=i+(j-1)*d2
                        k=fastnum1(ii) + ip-1
                        call calcha4kl(workc(z4),hteil,workc(z5),k,s, &
                                       m1,m2,m3,m4)
                        h=hamilton(fkoe,k)
                        if (htmtype(h) .eq. 1) then
                           call addvxxzo(workc(z5),dtpsi(zpsi(s)), &
                                         block(s))
                        else
                           call xvxxzza(koeff(k),workc(z5), &
                                        dtpsi(zpsi(s)),block(s))
                        endif
                     enddo
                  enddo
               enddo
            enddo
         enddo
         enddo

      endif

      return
      end subroutine hnatfast


! **********************************************************************
! *** V6.0 MB ***********> SUBROUTINE CALC HUNA <***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~
!
!calchuna: Calculates the action of the uncorrelated           
!          Hamiltonian on the MCTDH-coefficients.     
!
! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients. 
!
! related routines: calchuna_gwp, calchuna_nogwp
! **********************************************************************

      subroutine calchuna (psi,dtpsi,hteil,nham,jindx)

      implicit none

      integer(long)                                      :: m,s,k,block1,block2,&
                                                            zeig1,zeig2,zeig3
      integer(long), intent(in)                          :: nham
      integer(long), dimension(jindxdim), intent(in)     :: jindx
      integer(long), dimension(2*fmblock)                :: worki
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(adim), intent(inout)       :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)       :: hteil
      complex(dop), dimension(2*fmblock)                 :: workc

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      k=zham(nham)+kzahl(nham)+nmulpot(nham)-1
      if (citype .gt. 0) then    ! Selected CI
         zeig1=1
         zeig2=fmblock+1
         block2=0
         do s=1,nstate
       inner:do m=1,nmode
               if(khzahl(m,s,nham) .eq. 0) cycle inner
               block1=vdim(m,s)*dim(m,s)*ndim(m,s)
               if (m.eq. 1) then
                  call venpsi(m,jindx(zpsi(s)),worki(zeig1), &
                       jindx(zpsi(s)),block(s),block1,block(s), &
                       nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                       workc(zeig1),dtpsi(zpsi(s)))
               else
                  call venpsi(m,jindx(zpsi(s)),worki(zeig1), &
                       worki(zeig2),block(s),block1,block2, &
                       nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                       workc(zeig1),workc(zeig2))
               endif
               zeig3=zeig1
               zeig1=zeig2
               zeig2=zeig3
               block2=block1

               call cpvxz_s(psi(zpsi(s)),workc(zeig1), &
                    jindx(zpsi(s)),worki(zeig2),block(s),block2)

               call qtxxzza(hteil(pmat(m,k+s)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            enddo inner

            call cpvxz_s(workc(zeig2),dtpsi(zpsi(s)), &
                 worki(zeig2),jindx(zpsi(s)),block2,block(s))
         enddo

      else           ! Normal Case 
! Comment in the collapse version when GFortran updated. GWR 08/13

! omp parallel do num_threads(ompthread) if(s.ge.3) private(s,m) collapse(2)
!$omp parallel do num_threads(ompthread) private(s,m) if (lompthread .and. lompcalcha)
         do s=1,nstate
            do m=1,nmode
               if (khzahl(m,s,nham) .gt. 0) then
                  call qtxxzza(hteil(pmat(m,k+s)),psi(zpsi(s)), &
                       dtpsi(zpsi(s)),vdim(m,s),dim(m,s),ndim(m,s))
               endif
            enddo
         enddo
!$omp end parallel do
      endif

      return
      end subroutine calchuna

! **********************************************************************
! *** V6.0 MB ***********> SUBROUTINE CALC HUNA <***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~
!
!calchuna_gwp: Calculates the action of the uncorrelated Hamiltonian for 
!              GWP modes on the MCTDH-coefficients. 
!
! related routines: calchuna, calchuna_nogwp
! **********************************************************************

      subroutine calchuna_gwp (psi,dtpsi,hteil,gfgwp,nham)

      implicit none

      integer(long)                                 :: m,m1,s,k,zeig1,zeig2,zeig3
      integer(long), intent(in)                          :: nham
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(adim), intent(inout)       :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)       :: hteil
      complex(dop), dimension(gfgwpdim), intent(in)      :: gfgwp
      complex(dop), dimension(2*maxblock)                :: workc

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      k=zham(nham)+kzahl(nham)+nmulpot(nham)-1
      do s=1,nstate
         do m=1,nmode
            if (.not. gwpm(m)) cycle

            if (khzahl(m,s,nham) .gt.0) then
               call zerovxz(workc,2*block(s))
               zeig1=1
               zeig2=block(s)+1
               call qtxxzza(hteil(pmat(m,k+s)),psi(zpsi(s)), &
                    workc(zeig1),vdim(m,s),dim(m,s),ndim(m,s))

               do m1=1,nmode
                  if (m1 .eq. m) cycle

                  if (gwpm(m1)) then
                     zeig3=zeig1
                     zeig1=zeig2
                     zeig2=zeig3
              !       if(.not.lompthread)then
                        call qtxxzz(gfgwp(zgfgwp(m1,s,s)),workc(zeig2), &
                             workc(zeig1),vdim(m1,s),dim(m1,s), &
                             ndim(m1,s))
              !       else
              !          call qtxxzzomp(gfgwp(zgfgwp(m1,s,s)),workc(zeig2), &
              !               workc(zeig1),vdim(m1,s),dim(m1,s), &
              !               ndim(m1,s),ompthread)
              !       endif
                  endif
               enddo
               call addvxxzo(workc(zeig1),dtpsi(zpsi(s)),block(s))
            endif
         enddo
      enddo

      return
      end subroutine calchuna_gwp

! **********************************************************************
! *** V6.0 MB ***********> SUBROUTINE CALC HUNA <***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~
!
!calchuna_nogwp: Calculates the action of the uncorrelated Hamiltonian 
!              ignoring GWP modes on the MCTDH-coefficients. 
!
! related routines: calchuna, calchuna_gwp
! **********************************************************************

      subroutine calchuna_nogwp (psi,dtpsi,hteil,nham,jindx)

      implicit none

      integer(long)                                      :: m,s,k,block1,block2,&
                                                            zeig1,zeig2,zeig3
      integer(long), intent(in)                          :: nham
      integer(long), dimension(jindxdim), intent(in)     :: jindx
      integer(long), dimension(2*fmblock)                :: worki
      complex(dop), dimension(adim), intent(in)          :: psi
      complex(dop), dimension(adim), intent(inout)       :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)       :: hteil
      complex(dop), dimension(2*fmblock)                 :: workc

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      k=zham(nham)+kzahl(nham)+nmulpot(nham)-1
      if (citype .gt. 0) then    ! Selected CI
         zeig1=1
         zeig2=fmblock+1
         block2=0
         do s=1,nstate
       inner:do m=1,nmode
               if (gwpm(m)) cycle

               if(khzahl(m,s,nham) .eq. 0) cycle inner
               block1=vdim(m,s)*dim(m,s)*ndim(m,s)
               if (m.eq. 1) then
                  call venpsi(m,jindx(zpsi(s)),worki(zeig1), &
                       jindx(zpsi(s)),block(s),block1,block(s), &
                       nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                       workc(zeig1),dtpsi(zpsi(s)))
               else
                  call venpsi(m,jindx(zpsi(s)),worki(zeig1), &
                       worki(zeig2),block(s),block1,block2, &
                       nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                       workc(zeig1),workc(zeig2))
               endif
               zeig3=zeig1
               zeig1=zeig2
               zeig2=zeig3
               block2=block1

               call cpvxz_s(psi(zpsi(s)),workc(zeig1), &
                    jindx(zpsi(s)),worki(zeig2),block(s),block2)

               call qtxxzza(hteil(pmat(m,k+s)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            enddo inner

            call cpvxz_s(workc(zeig2),dtpsi(zpsi(s)), &
                 worki(zeig2),jindx(zpsi(s)),block2,block(s))
         enddo

      else           ! Normal Case 
         do s=1,nstate
            do m=1,nmode
               if (gwpm(m)) cycle

               if (khzahl(m,s,nham) .gt. 0) then
                  call qtxxzza(hteil(pmat(m,k+s)),psi(zpsi(s)), &
                       dtpsi(zpsi(s)),vdim(m,s),dim(m,s),ndim(m,s))
               endif
            enddo
         enddo
      endif

      return
      end subroutine calchuna_nogwp

!-----------------------------------------------------------------------
!   mpicalcha
!-----------------------------------------------------------------------
#ifdef MPI
      subroutine mpicalcha(psi,phi,dtpsi,hteil,nham,koffset,k0,jindx)

      implicit none

      integer(long), dimension(jindxdim), intent(in)   :: jindx
      integer(long), dimension(2*fmblock)              :: worki
      complex(dop), dimension(adim), intent(in)        :: psi
      complex(dop), dimension(pmatdim), intent(in)     :: hteil
      complex(dop), dimension(adim), intent(out)       :: dtpsi
      complex(dop), dimension(adim)                    :: tmpdtpsi
      complex(dop), dimension(totphidim), intent(in)   :: phi

      integer(long), intent(in)    :: koffset,k0,nham

      integer(long) :: h,i0,i1,k,s,s1

! --- Get Data
      call MPI_BCast(jindx,jindxdim,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
! ... Hteil-matrices
!      call MPI_Bcast(hteil,pmatdim,MPI_DOUBLE_COMPLEX,master,MPI_COMM_WORLD,mpierr)
! ... A-Vector
      call MPI_Bcast(psi,adim,MPI_DOUBLE_COMPLEX,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(koffset,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(k0,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
      if (mpirank .ne. master) call zerovxz(dtpsi,adim)

! Divide up the work.
      call mpe_decomp1d(k0,koffset,i0,i1)

!         do k=k0+mpirank,koffset,nmpi
      do k=i0,i1
         s = kf(k)
         s1= ki(k)
         if (s.eq.s1) then
            call calcha1k(psi,phi,hteil,dtpsi,k,s,jindx)
         else
            call calcha2k(psi,phi,hteil,dtpsi,k,s,s1,jindx)
         endif

      enddo

! --- Reduce Data
      call MPI_Reduce(dtpsi,tmpdtpsi,adim,MPI_DOUBLE_COMPLEX,MPI_SUM,master, &
                                                  MPI_COMM_WORLD,mpierr)
      if(mpirank.eq.master) call cpvxz(tmpdtpsi,dtpsi,adim)

      return
      end subroutine mpicalcha
#endif

!#######################################################################
!
! HA for a multi-layer wavefunction
!
! t: tree for top layer 
! nham : Hamiltonian for which HA is calculated
!#######################################################################

      subroutine calcha_ml(t,psi,dtpsi,nham)

      implicit none
      type(wfnode_t),intent(inout) :: t
      integer(long), intent(in)    :: nham
      complex(dop)                 :: psi(:),dtpsi(:)
      type(hterm_tp),pointer       :: hk(:)
      complex(dop),allocatable     :: work(:)
      integer                      :: k,ki,kf,f
      integer                      :: plen,totplen,h
      logical                      :: lcmf

      plen = product(t%ndim)
      totplen=plen*t%navecs
      dtpsi = CZERO
      lcmf = (t%topnode%cmftyp /= 0)

      ki=zham(nham)
      kf=zham(nham)+kzahl(nham)-1

!$OMP parallel default(shared) private(work,k,hk,f,h) reduction(+:dtpsi)
      allocate(work(plen))
!$OMP do schedule(dynamic)
      do k=ki,kf ! loop over k-terms
         hk=>t%hterms(k,:)
         work = psi
         call hprod_phi(hk,work)
         h=hamilton(fkoe,k)
         if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104) then
            dtpsi = dtpsi + koeff(k)*work
         else
            dtpsi = dtpsi + work
         endif
      end do
!$OMP end do
!$OMP do schedule(dynamic)
      do f=1,t%nmodes
         if (.not. t%hasuncor(f,nham)) cycle

         k=kf+1
         hk=>t%hterms(k,:)
         work = psi
         call hunprod_phi(hk,work,f)
         dtpsi = dtpsi + work
      enddo
!$OMP end do
      deallocate(work)
!$OMP end parallel

      end subroutine calcha_ml

      end module calchamod

