! **********************************************************************
! *                                                                    *
! *                         D1HPHI
! *                                                                    *
! MCTDH module containing subroutines to evaluate the action of the
! Hamiltonian on the single-particle functions
!
! d1hphi:         Computes for all modes and all states the action of    
!                 the correlated Hamiltonian on all the single-particle
!                 functions. 
! d1hphi1m:       Computes the action of the correlated Hamiltonian on 
!                 all single-particle functions for one mode.
! d1hunphi:       Computes for all modes and all states the action of    
!                 the uncorrelated Hamiltonian on all the single-particle
!                 functions.
! d1hunphi1ms:    Computes the action of the  uncorrelated Hamiltonian on
!                 all single-particle functions for one mode and one state.
! d1hphi1mk:      Computes the action of one term of the Hamiltonian on 
!                 the single-particle functions of one mode.
! d1hcall:        Computes the action of one term of the Hamiltonian on 
!                 one single-particle function.
! d1hop:          Computes the action of an operator on a single-particle
!                 functions.
! fftop, fftiop, fftzop: auxilliary routines for hop
!
! **********************************************************************
      module d1hphimod

use decimal, only: dop, long
use global
use globmemmod
use logdat
use mmlib,   only: dqxxdz, dqxxdz2
use mtlib,   only: dtxxdz, dtxxyz, qttxdz, qttxyz, qtxxdz, qtxxyz
use xvlib,   only: xmxxzza
use fftmod,  only: fft
use op1lib,  only: zeromxz, zerovxl, cpvxz
use op2lib,  only: addmxxzo

      implicit none
      private
      public :: d1hphi, d1hcall, d1hunphi, d1hphi1m, d1hunphi1ms, d1hphi1mk
      contains

! **********************************************************************
! *** V6.6 GW ***************> SUBROUTINE HPHI <************************
!                             ~~~~~~~~~~~~~~~~~
! Computes for all modes, all states, and all Hamiltonian terms the
! action of the correlated Hamiltonian on the single-particle functions.
! **********************************************************************

      subroutine d1hphi(psi,hpsi,ktype,nham)

use griddatmod 
use psidef
use hpsimod

      implicit none

      integer(long)                                 :: m
      integer(long), intent(in)                     :: ktype,nham
      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(hpsidim), intent(out) :: hpsi

! --- LOOP OVER EACH MODE AND STATE ---

      do m = 1,nmode
         call d1hphi1m(psi(dzetf(m,1,1)),hpsi,m,dzetf(m,1,1),phidim(m),&
              ktype,nham)
      enddo

      return
      end subroutine d1hphi

! *** V6.6 GW **************> SUBROUTINE D1 HPHI 1M <**********************
!                            ~~~~~~~~~~~~~~~~~~~~

! Computes for one mode the action of the correlated Hamiltonian on the
! single-particle functions.
!
! ktype governs which terms are included:
! ktype = 0 : all terms
! ktype = 1 : Potential terms are excluded (i.e. only kinop terms
!             calculated)
! nham      : Number of operator to be used
!

      subroutine d1hphi1m (psi,hpsi,m,dzetf1,phidim1,ktype,nham)

use global
use maxv, only: maxkoe
use griddatmod 
use operdef
use psidef
use hpsimod
use runpropmod

      implicit none

      integer(long)                                     :: k,k1,&
                                                           k2,kt,s1,t
      integer(long), intent(in)                         :: m,dzetf1,&
                                                           phidim1,nham,&
                                                           ktype
      complex(dop), dimension(dzetf1:dzetf1+phidim1-1), intent(in) :: psi
      complex(dop), dimension(hpsidim), intent(out)                :: hpsi

      logical(kind=4), dimension(maxkoe)                :: kcalc
      logical(kind=4)                                   :: luncorr=.false.


!-----------------------------------------------------------------------
! first find which terms need to be calculated
!
! kact(k)=5 are twin of left-right operators for dentype=1. Calculated in
! d1hcall
!-----------------------------------------------------------------------
      call zerovxl(kcalc,ktot)
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      if (ktype .eq. 0) then
         do k=k1,k2
            if (.not. diag(m,k) .and. k .eq. kfirst(m,k))  &
               kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
            if (kact(k).eq.5) kcalc(k)=.false.
         enddo
      else if (ktype .eq. 1) then
         do k=k1,k2
            if (kinop(m,k) .eq. 1) kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
            if (kact(k).gt.1 .and. kact(k).lt.5) then
               if ((.not.diag(m,k)) .and. (k.eq.kfirst(m,k))) &
                    kcalc(k)=.true.
            endif
            if (kact(k).eq.5) kcalc(k)=.false.
         enddo
      endif

! --- LOOP OVER EACH HAMILTONIAN TERM ---
      do k = k1,k2
         if (kcalc(k)) then
!     pure left acting operators
            if ((kact(k).eq.1) .or. (kact(k).eq.2)) then
               s1=ki(k)
               do t=1,nstate
                  call d1hphi1mk(psi(dzetf(m,s1,t)), &
                       hpsi(dzhpsi(m,k,t)),m,k,ddim(m,s1,t), &
                       subdim(m),supgdim(m),luncorr)
               enddo
!     pure right acting operators
            else if (kact(k).eq.3) then
               s1=kf(k)
               do t=1,nstate
                  call d1hphi1mk(psi(dzetf(m,t,s1)), &
                       hpsi(dzhpsi(m,k,t)),m,k,ddim(m,t,s1), &
                       subdim(m),supgdim(m),luncorr)
               enddo
!     left-right acting operators
            else if ((kact(k).eq.4)) then
               kt = ktwin(k)            
               s1 = ki(k)
               t  = kf(kt)
               call d1hphi1mk(psi(dzetf(m,s1,t)), &
                    hpsi(dzhpsi(m,k,1)),m,k,ddim(m,s1,t), &
                    subdim(m),supgdim(m),luncorr)
           endif
         endif
      enddo

      return
      end subroutine d1hphi1m

! *** V6.6 GW ***************> SUBROUTINE D1HUNPHI <**********************
!                             ~~~~~~~~~~~~~~~~~~~
!
! Computes for all modes and all states the action of the uncorrelated
! Hamiltonian on the single-particle functions.
! Result is returned as a sum of all terms in hunpsi.

      subroutine d1hunphi (psi,hunpsi,hktype,nham)

use global
use griddatmod 
use psidef

      implicit none

      integer(long)                                       :: m,s,t
      integer(long), intent(in)                           :: hktype,nham
      complex(dop), dimension(dgldim), intent(in)         :: psi
      complex(dop), dimension(adim+1:dgldim), intent(out) :: hunpsi


! --- LOOP OVER EACH MODE AND STATE ---

      do m = 1,nmode
! ... exclude Gaussian mode :
         if (gwpm(m)) cycle

         do t = 1,nstate
            do s = 1,nstate
               call d1hunphi1ms(psi(dzetf(m,s,t)),hunpsi(dzetf(m,s,t)),&
                    m,s,t,ddim(m,s,t),subdim(m),hktype,nham)
            enddo
         enddo
      enddo


      return
      end subroutine d1hunphi

! *** V6.6 GW *************> SUBROUTINE D1HUNPHI 1MS <********************
!                           ~~~~~~~~~~~~~~~~~~~~~~~
!
! Computes for one mode and one state the action of the uncorrelated
! Hamiltonian on the single-particle functions.
!
! hktype governs which terms are included:
! hktype = 0 : all terms
! hktype = 1 : exclude CAP terms
! hktype = 2 : only CAP terms
!
! nham      : Number of operator to be used
!
!#######################################################################

      subroutine d1hunphi1ms (psi,hunpsi,m,s,t,dim1,subdim1,hktype,nham)

use global
use maxv, only: maxkoe
use dvrdatmod
use griddatmod
use operdef
use psidef
use runpropmod

      implicit none

      integer(long)                                      :: k,k1,k2,h
      integer(long), intent(in)                          :: hktype
      integer(long), intent(in)                          :: m,s,t,dim1, &
                                                            subdim1,nham
      complex(dop), dimension(subdim1,dim1), intent(in)  :: psi
      complex(dop), dimension(subdim1,dim1), intent(out) :: hunpsi
      complex(dop), dimension(subdim1,dim1)              :: workc
      logical(kind=4), dimension(maxkoe)                 :: kcalc
      logical(kind=4)                                    :: luncorr=.true.

      call zeromxz(hunpsi,subdim(m),ddim(m,s,t))

!-----------------------------------------------------------------------
! first find which terms need to be calculated
! autocap = 0 : normal hermitian operator
!         > 0 : non-active CAP
!         < 0 : active CAP
! nstatecap(k) : state on which CAP is active. If 0 all states.
!-----------------------------------------------------------------------
      if(khzahl(m,s,nham).eq.0) return
      k1=zhun(m,s,nham)
      k2=zhun(m,s,nham)+khzahl(m,s,nham)-1

      do k=k1,k2
         if (hktype .eq. 0) then
            if (autocap(k) .eq. 0) then
               kcalc(k)=.true.
            else if (autocap(k) .lt. 0 .and. nstatecap(k) .eq. 0) then
               kcalc(k)=.true.
            else if (autocap(k) .lt. 0 .and. nstatecap(k) .eq. s) then
               kcalc(k)=.true.
            endif
         else if (hktype .eq. 1 .and. autocap(k) .eq. 0) then
            kcalc(k)=.true.
         else if (hktype .eq. 2 .and. autocap(k) .lt. 0) then
            if (nstatecap(k) .eq. 0 .or. nstatecap(k) .eq. s) kcalc(k)=.true.
         else
            kcalc(k)=.false.
         endif
      enddo

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      do k=k1,k2
         if (kcalc(k)) then
            call d1hphi1mk(psi,workc,m,k,ddim(m,s,t), &
                 subdim(m),supgdim(m),luncorr)

            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then
               call addmxxzo(workc,hunpsi,subdim(m),ddim(m,s,t))
            else
               call xmxxzza(koeff(k),workc,hunpsi,subdim(m),ddim(m,s,t))
            endif

         endif
      enddo

      return
      end subroutine d1hunphi1ms

! *** V6.6 GW **************> SUBROUTINE D1 HPHI 1MK <*********************
!                            ~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and ONE Hamiltonian term the action
! of the correlated Hamiltonian on the single-particle functions.
!
!-----------------------------------------------------------------------

      subroutine d1hphi1mk (psi,hpsi,m,k,dim,subdim,supgdim,luncorr)

      implicit none

      integer(long)                                    :: e
      integer(long), intent(in)                        :: m,k,subdim,dim, &
                                                          supgdim
      complex(dop), dimension(subdim,dim), intent(in)  :: psi
      complex(dop), dimension(subdim,dim), intent(out) :: hpsi
      logical(kind=4), intent(in)                      :: luncorr

! --- LOOP OVER EACH HAMILTONIAN TERM ---

      do e = 1,dim
         call d1hcall(m,k,hpsi(1,e),psi(1,e),subdim,supgdim,luncorr)
      enddo

      return
      end subroutine d1hphi1mk 

!#######################################################################
!
! luncorr=.true. if operator is uncorrelated
!
!#######################################################################

      subroutine d1hcall(m,k,hpsi,psi,subdim1,supgdim1,luncorr)

use global
use griddatmod
use operdef

      implicit none

      integer(long)                                 :: kt,f,n,h
      integer(long), intent(in)                     :: m,k,subdim1,supgdim1
      complex(dop), dimension(subdim1), intent(in)  :: psi
      complex(dop), dimension(subdim1), intent(out) :: hpsi
      complex(dop), dimension(subdim1)              :: tmphpsi
      logical(kind=4), intent(in)                   :: luncorr

!     Uncorrelated terms:
      if (luncorr) then
         n=1
         f=spfdof(n,m)
         call d1hop(f,k,hpsi,psi,mr(mrhops),subdim1,supgdim1, &
              mi(mifftfak),mc(mchin),mc(mcrueck))
         do n=2,nspfdof(m)
            f=spfdof(n,m)
            h=hamilton(f,k)
            if (htmtype(h) .ne. 1) then
               call cpvxz(hpsi,tmphpsi,subdim1)
               call d1hop(f,k,hpsi,tmphpsi,mr(mrhops),subdim1, &
                    supgdim1,mi(mifftfak),mc(mchin),mc(mcrueck))
            endif 
         enddo
!     Correlated terms:
      else
!     left acting operators
         if ((kact(k).eq.1) .or. (kact(k).eq.2) .or. (kact(k).eq.4)) &
              then
            n=1
            f=spfdof(n,m)
            call d1hop(f,k,hpsi,psi,mr(mrhops),subdim1,supgdim1, &
                 mi(mifftfak),mc(mchin),mc(mcrueck))
            do n=2,nspfdof(m)
               f=spfdof(n,m)
               h=hamilton(f,k)
               if (htmtype(h) .ne. 1) then
                  call cpvxz(hpsi,tmphpsi,subdim1)
                  call d1hop(f,k,hpsi,tmphpsi,mr(mrhops),subdim1, &
                       supgdim1,mi(mifftfak),mc(mchin),mc(mcrueck))
               endif 
            enddo
         endif

!     pure right acting operators
         if (kact(k).eq.3) then
            n=1
            f=spfdof(n,m)
            call d1hop2(f,k,hpsi,psi,mr(mrhops),subdim1,supgdim1, &
                 mi(mifftfak),mc(mchin),mc(mcrueck))
            do n=2,nspfdof(m)
               f=spfdof(n,m)
               h=hamilton(f,k)
               if (htmtype(h) .ne. 1) then
                  call cpvxz(hpsi,tmphpsi,subdim1)
                  call d1hop2(f,k,hpsi,tmphpsi,mr(mrhops),subdim1, &
                       supgdim1,mi(mifftfak),mc(mchin),mc(mcrueck))
               endif 
            enddo
         endif

!     right action of left-right acting operators
         if (kact(k).eq.4) then
            kt = ktwin(k)
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               h=hamilton(f,kt)
               if (htmtype(h) .ne. 1) then
                  call cpvxz(hpsi,tmphpsi,subdim1)
                  call d1hop2(f,kt,hpsi,tmphpsi,mr(mrhops),subdim1, &
                       supgdim1,mi(mifftfak),mc(mchin),mc(mcrueck))
               endif 
            enddo
         endif
      endif


      return
      end subroutine d1hcall

!***********************************************************************

      subroutine d1hop (f,k,hpsi,psi,hops,subdim1,supgdim1, &
                 fftfak,hin,rueck)

use global
use dvrdatmod
use griddatmod
use operdef

use hphimod, only: fftop,fftiop,fftzop

      implicit none

      integer(long)                                 :: g,h
      integer(long), intent(in)                     :: f,k,subdim1,supgdim1
      integer(long), dimension(fftdim), intent(in)  :: fftfak
      real(dop), dimension(hopsdim), intent(in)     :: hops
      complex(dop), dimension(subdim1), intent(in)  :: psi
      complex(dop), dimension(subdim1), intent(out) :: hpsi
      complex(dop), dimension(fftdim), intent(in)   :: hin,rueck

!-----------------------------------------------------------------------
! Go to selected operator
!-----------------------------------------------------------------------
      h=hamilton(f,k)
     select case (htmtype(h))

     case default
      routine='d1Hop'
      write(message,'(a,i5)')  'Unknown htmtype :',htmtype(h)
      call errormsg

     case (1)
      call cpvxz(psi,hpsi,subdim1)
      return

     case (2)
      if (htmdof(h) .gt. 0) then
         call dtxxdz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      else
         call dqxxdz(hops(zhop(h)),psi,hpsi,supgdim1)
      endif
      return

     case (3)
      call dtxxdz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (4)
      call dtxxyz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      return

     case (5)
      call fftop(psi,hpsi,vgdim(f),gdim(f),ngdim(f)*supgdim1, &
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)), &
           hops(zhop(h)))   
      return

     case (6)
      call fftiop(psi,hpsi,vgdim(f),gdim(f),ngdim(f)*supgdim1, &
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)), &
           hops(zhop(h)))
      return

     case (7)
      call fftzop(psi,hpsi,vgdim(f),gdim(f),ngdim(f)*supgdim1, &
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)), &
           hops(zhop(h)))
      return

     case (8)
      call qtxxdz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      return

     case (9)
      call qtxxdz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (10)
      call qtxxyz(hops(zhop(h)),psi,hpsi,vgdim(f),gdim(f), &
           ngdim(f)*supgdim1)
      return

     end select

      end subroutine d1hop 

!***********************************************************************

      subroutine d1hop2 (f,k,hpsi,psi,hops,subdim1,supgdim1, &
                 fftfak,hin,rueck)

use global
use dvrdatmod
use griddatmod
use operdef

      implicit none

      integer(long)                                 :: g,h
      integer(long), intent(in)                     :: f,k,subdim1,supgdim1
      integer(long), dimension(fftdim), intent(in)  :: fftfak
      real(dop), dimension(hopsdim), intent(in)     :: hops
      complex(dop), dimension(subdim1), intent(in)  :: psi
      complex(dop), dimension(subdim1), intent(out) :: hpsi
      complex(dop), dimension(fftdim), intent(in)   :: hin,rueck
      complex(dop), dimension(subdim1)              :: vec
      complex(dop), dimension(subdim1)              :: workc


!-----------------------------------------------------------------------
! Go to selected operator
!-----------------------------------------------------------------------
      h=hamilton(f,k)
     select case (htmtype(h))

     case default
      routine='d1Hop2'
      write(message,'(a,i5)')  'Unknown htmtype :',htmtype(h)
      call errormsg

!-----------------------------------------------------------------------
! type 1 and 2 can be replaced by multi-dimensional hop if required
!-----------------------------------------------------------------------
     case (1)
      call cpvxz(psi,hpsi,subdim1)
      return

     case (2)
      call dtxxdz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f), &
           ngdim(f))
      return

     case (3)
      call dtxxdz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f), &
           ngdim(f))
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (4)
      call dtxxyz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f), &
           ngdim(f))
      return

     case (5)
      call fftop2(psi,hpsi,supgdim1*vgdim(f),gdim(f),ngdim(f),&
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)),&
           hops(zhop(h)),vec,workc)   
      return

     case (6)
      call fftiop2(psi,hpsi,supgdim1*vgdim(f),gdim(f),ngdim(f),&
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)),&
           hops(zhop(h)),vec,workc)
      return

     case (7)
      call fftzop2(psi,hpsi,supgdim1*vgdim(f),gdim(f),ngdim(f),&
           hin(zfft(f)),rueck(zfft(f)),fftfak(zfft(f)),&
           hops(zhop(h)),vec,workc)
      return

     case (8)
      call qttxdz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f),&
           ngdim(f))
      return

     case (9)
      call qttxdz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f),&
           ngdim(f))
      do g=1,subdim1
         hpsi(g)=dcmplx(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (10)
      call qttxyz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f),&
           ngdim(f))
      return

     case (11)
      call qttxdz(hops(zhop(h)),psi,hpsi,supgdim1*vgdim(f),gdim(f),&
           ngdim(f))
      return

     case (12)
      call cpvxz(psi,hpsi,subdim1)
      return

     case (13)
      call dqxxdz2(hops(zhop(h)),psi,hpsi,supgdim1)
      return

     end select

      end subroutine d1hop2

!#######################################################################

      subroutine fftop2(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak, &
                 fftk,vec,workc)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim), intent(in)                 :: fftk
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi
      complex(dop), dimension(gdim), intent(in)              :: hin,rueck
      complex(dop), dimension(gdim), intent(out)             :: workc
      complex(dop), dimension(gdim), intent(out)             :: vec

      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=fftk(g)*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftop2

!#######################################################################

      subroutine fftiop2(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak, &
                 fftk,vec,workc)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim), intent(in)                 :: fftk
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi
      complex(dop), dimension(gdim), intent(in)              :: hin,rueck
      complex(dop), dimension(gdim), intent(out)             :: workc
      complex(dop), dimension(gdim), intent(out)             :: vec


      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=conjg(psi(v,g,n))
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=dcmplx(0.0_dop,fftk(g))*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftiop2

!#######################################################################

      subroutine fftzop2(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak, &
                 fftk,vec,workc)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim,2), intent(in)               :: fftk
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi
      complex(dop), dimension(gdim), intent(in)              :: hin,rueck
      complex(dop), dimension(gdim), intent(out)             :: workc
      complex(dop), dimension(gdim), intent(out)             :: vec

      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=dcmplx(fftk(g,1),fftk(g,2))*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftzop2

!-----------------------------------------------------------------------
      end module d1hphimod


