! **********************************************************************
! *                                                                    *
! *                module D2CALCHA
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! * d2calcdissa1:   Calculates the action of the dissipative terms on  *
! *                 the MCTDH-coefficients, no projection is applied   *
! * d2calcdissa2:   as calcdissa1, but projection is applied           *
! * d2calcha/2:     Calculates the action of the correlated            *
! *                 Hamiltonian on the MCTDH-coefficients.             *
! *                 d2calcha2 is used by SIL calculations.             *
! * d2calchuna/2:   Calculates the action of the uncorrelated          *
! *                 Hamiltonian on the MCTDH-coefficients.             *
! *                 d2calchuna2 is used by SIL calculations.           *
! *                                                                    *
! **********************************************************************
      module d2calchamod

use decimal,   only: long, dop
use global
use mltypes, only: hterm_tp
use dvrdatmod
use griddatmod
use operdef
use psidef
use hpsimod
use runpropmod, only: deuncorr,decorr

use logdat
use openmpmod, only: lompthread,ompthread
use mtlib,     only: qtxxzz, qtxxzza, tqxazz, tqxazza, mtxxzz, mttxzz
use xvlib,     only: xvxxzzo, xvxxzza
use op1lib,    only: cpvxz, trhxz, zerovxz
use op2lib,    only: addmxxzo, addmxazo1, submxazo1, submxxzo1
use d2tens, only: mkherm, mkherm2

use linear, only: hprod_phi2,hunprod_phi2

      implicit none
      private
      public :: d2calcdissa1,d2calcdissa2,d2calcha,d2calcha2,d2calchaloch2, &
                d2calchuna,d2calchuna2,d2calchunaloch2,d2calchaloch2_ml,&
                d2calchunaloch2_ml
      contains

! ************************> SUBROUTINE D2CALCDISSA1 <*******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the dissipative terms on the
! DMCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. No projection
!
! **********************************************************************

      subroutine d2calcdissa1 (psi,dtpsi,hteil)

      implicit none

      integer(long)                                             :: m,k,kt,s,si,sf,lsi,lsf,rsi,rsf,hzw,h
      integer(long), dimension(2)                               :: zw
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)              :: hteil
      complex(dop), dimension(maxblock*maxblock,2)              :: workc


! --- LOOP OVER EACH DISSIPATIVE TERM ---

      do k=1,kzahl(1)

! --- PURE LEFT ACTION 
         if ((kact(k).eq.2) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            si = ki(kt)
            sf = kf(k)
            do s=1,nstate
               zw(1) = 1
               zw(2) = 2
               call cpvxz(psi(dzpsi(si,s)),workc(1,zw(1)), &
                    block(si)*block(s))
               do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
                  if ((.not.diag(m,k)) .or. (.not. diag(m,kt))) then
!     Note that si=ki(k), sf=kf(k)
                     call mtxxzz(hteil(pmat2(m,k)),workc(1,zw(1)), &
                          workc(1,zw(2)),vdim(m,sf),dim(m,si), &
                          ndim(m,si)*block(s),dim(m,sf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo

               h=hamilton(fkoe,k)
               if (htmtype(h) .EQ. 1) then
                  call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(sf,s)), &
                       block(sf),block(s))
               else
                  call xvxxzza(koeff(k),workc(1,zw(1)), &
                       dtpsi(dzpsi(sf,s)),block(sf)*block(s))
               endif

            enddo
         endif


! --- PURE RIGHT ACTION 
         if ((kact(k).eq.3) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            si = kf(k)
            sf = ki(kt)
            do s=1,nstate
               zw(1) = 1
               zw(2) = 2
               call cpvxz(psi(dzpsi(s,si)),workc(1,zw(1)), &
                    block(s)*block(si))
               do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
                  if ((.not.diag(m,k)) .or. (.not. diag(m,kt))) then
!     Note that si=kf(k), sf=ki(k)
                     call mttxzz(hteil(pmat2(m,k)),workc(1,zw(1)),&
                          workc(1,zw(2)),block(s)*vdim(m,sf),&
                          dim(m,si),ndim(m,si),dim(m,sf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo

               h=hamilton(fkoe,k)
               if (htmtype(h) .EQ. 1) then
                  call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(s,sf)),&
                       block(s),block(sf))
               else
                  call xvxxzza(koeff(k),workc(1,zw(1)),&
                       dtpsi(dzpsi(s,sf)),block(s)*block(sf))
               endif

            enddo
         endif


! --- LEFT-RIGHT ACTION 
         if ((kact(k).eq.4) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            lsi = ki(k)
            lsf = kf(k)
            rsi = kf(kt)
            rsf = ki(kt)
            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(lsi,rsi)),workc(1,zw(1)),&
                 block(lsi)*block(rsi))

!     Left action
            do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
               if (.not. diag(m,k)) then
!     Note that si=ki(k), sf=kf(k)
                  call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                       workc(1,zw(2)),vdim(m,lsf),dim(m,lsi), &
                       ndim(m,lsi)*block(rsi),dim(m,lsf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo
            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)), &
                 block(lsf)*block(rsi))
            endif

!     Right action
            do m=1,nmode
!     Note that for si .ne. sf always diag(m,kt)=.false.
               if (.not. diag(m,kt)) then
!     Note that rsi=kf(kt), rsf=ki(kt)
                  call mttxzz(hteil(pmat(m,kt)),workc(1,zw(1)), &
                       workc(1,zw(2)),block(lsf)*vdim(m,rsf), &
                       dim(m,rsi),ndim(m,rsi),dim(m,rsf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo

            h=hamilton(fkoe,kt)
            if (htmtype(h) .EQ. 1) then
               call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(lsf,rsf)), &
                    block(lsf),block(rsf))
            else
               call xvxxzza(koeff(kt),workc(1,zw(1)), &
                    dtpsi(dzpsi(lsf,rsf)),block(lsf)*block(rsf))
            endif

         endif

      enddo


      return
      end subroutine d2calcdissa1


! ************************> SUBROUTINE D2CALCDISSA2 <*******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the dissipative terms on the
! DMCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. Projection
!
! **********************************************************************

      subroutine d2calcdissa2 (psi,dtpsi,hteil)

      implicit none

      integer(long)                                             :: m,k,kt,s,lsi,lsf,rsi,rsf,hzw,h
      integer(long), dimension(2)                               :: zw
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)              :: hteil
      complex(dop), dimension(maxblock*maxblock,2)              :: workc


! --- LOOP OVER EACH DISSIPATIVE TERM ---

      do k=1,kzahl(1)

! --- PURE LEFT ACTION 
         if ((kact(k).eq.2) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            lsi = ki(k)
            lsf = kf(k)
            rsi = ki(kt)
            rsf = kf(kt)
            do s=1,nstate
               zw(1) = 1
               zw(2) = 2
               call cpvxz(psi(dzpsi(rsi,s)),workc(1,zw(1)), &
                    block(rsi)*block(s))

               do m=1,nmode
!     Note that for rsi .ne. rsf always diag(m,k)=.false.
                  if (.not. diag(m,kt)) then
!     Note that si=ki(k), sf=kf(k)
                     call mtxxzz(hteil(pmat(m,kt)),workc(1,zw(1)), &
                          workc(1,zw(2)),vdim(m,rsf),dim(m,rsi), &
                          ndim(m,rsi)*block(s),dim(m,rsf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo
!     Note that rsf=lsi
               do m=1,nmode
!     Note that for lsi .ne. lsf always diag(m,k)=.false.
                  if (.not. diag(m,k)) then
                     call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                          workc(1,zw(2)),vdim(m,lsf),dim(m,lsi), &
                          ndim(m,lsi)*block(s),dim(m,lsf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo

               h=hamilton(fkoe,k)
               if (htmtype(h) .EQ. 1) then
                  call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(lsf,s)), &
                       block(lsf),block(s))
               else
                  call xvxxzza(koeff(k),workc(1,zw(1)), &
                       dtpsi(dzpsi(lsf,s)),block(lsf)*block(s))
               endif

            enddo
         endif


! --- PURE RIGHT ACTION 
         if ((kact(k).eq.3) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            lsi = kf(k)
            lsf = ki(k)
            rsi = kf(kt)
            rsf = ki(kt)
            do s=1,nstate
               zw(1) = 1
               zw(2) = 2
               call cpvxz(psi(dzpsi(s,lsi)),workc(1,zw(1)), &
                    block(s)*block(lsi))
               do m=1,nmode
!     Note that for lsi .ne. lsf always diag(m,k)=.false.
                  if (.not. diag(m,k)) then
                     call mttxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                          workc(1,zw(2)),block(s)*vdim(m,lsf), &
                          dim(m,lsi),ndim(m,lsi),dim(m,lsf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo
!     Note that lsf=rsi
               do m=1,nmode
!     Note that for rsi .ne. rsf always diag(m,k)=.false.
                  if (.not. diag(m,kt)) then
                     call mttxzz(hteil(pmat(m,kt)),workc(1,zw(1)), &
                          workc(1,zw(2)),block(s)*vdim(m,rsf), &
                          dim(m,rsi),ndim(m,rsi),dim(m,rsf))
                     hzw = zw(1)
                     zw(1) = zw(2)
                     zw(2) = hzw
                  endif
               enddo

               h=hamilton(fkoe,k)
               if (htmtype(h) .EQ. 1) then
                  call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(s,rsf)), &
                       block(s),block(rsf))
               else
                  call xvxxzza(koeff(k),workc(1,zw(1)), &
                       dtpsi(dzpsi(s,rsf)),block(s)*block(rsf))
               endif

            enddo
         endif


! --- LEFT-RIGHT ACTION 
         if ((kact(k).eq.4) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            lsi = ki(k)
            lsf = kf(k)
            rsi = kf(kt)
            rsf = ki(kt)
            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(lsi,rsi)),workc(1,zw(1)), &
                 block(lsi)*block(rsi))

!     Left action
            do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
               if (.not. diag(m,k)) then
!     Note that si=ki(k), sf=kf(k)
                  call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                       workc(1,zw(2)),vdim(m,lsf),dim(m,lsi), &
                       ndim(m,lsi)*block(rsi),dim(m,lsf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo
            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)), &
                 block(lsf)*block(rsi))
            endif

!     Right action
            do m=1,nmode
!     Note that for si .ne. sf always diag(m,kt)=.false.
               if (.not. diag(m,kt)) then
!     Note that rsi=kf(kt), rsf=ki(kt)
                  call mttxzz(hteil(pmat(m,kt)),workc(1,zw(1)), &
                       workc(1,zw(2)),block(lsf)*vdim(m,rsf), &
                       dim(m,rsi),ndim(m,rsi),dim(m,rsf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo

            h=hamilton(fkoe,k)
            if (htmtype(h) .EQ. 1) then
               call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(lsf,rsf)),&
                    block(lsf),block(rsf))
            else
               call xvxxzza(koeff(kt),workc(1,zw(1)),&
                    dtpsi(dzpsi(lsf,rsf)),block(lsf)*block(rsf))
            endif

         endif

      enddo

      return
      end subroutine d2calcdissa2


! ************************> SUBROUTINE D2CALCHA <***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! DMCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. Additionally calculates the correlated energy.
!
! Assumes hermiticity of psi.
! **********************************************************************

      subroutine d2calcha (psi,dtpsi,hteil,nham)

      implicit none

      integer(long)                                             :: m,k,k1,k2,si,sf,s,hzw,h
      integer(long), dimension(2)                               :: zw
      integer(long), intent(in)                                 :: nham
      real(dop)                                                 :: en
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)              :: hteil
      complex(dop), dimension(maxblock*maxblock,2)              :: workc


! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

      do s=1,nstate
         decorr(s) = 0.0_dop
      enddo

      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

outer:do k=k1,k2

!     Use only hamiltonian terms
         if (kact(k) .ne. 1) cycle outer

         sf = kf(k)
         si = ki(k)

         do s=1,nstate

            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(si,s)),workc(1,zw(1)), &
                 block(si)*block(s))

            do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
               if (.not. diag(m,k)) then
                  call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                       workc(1,zw(2)),vdim(m,sf),dim(m,si), &
                       ndim(m,si)*block(s),dim(m,sf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo
            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)),block(sf)*block(s))
            endif
            call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(sf,s)), &
                 block(sf),block(s))

!     Calculate correlated energy
            if (s .eq. sf) then
               call trhxz(workc(1,zw(1)),en,block(s))
               decorr(s) = decorr(s) + en
            endif

            if (lrlx) then
               call addmxazo1(dtpsi(dzpsi(s,sf)),workc(1,zw(1)), &
                    block(sf),block(s))
            else
               call submxazo1(dtpsi(dzpsi(s,sf)),workc(1,zw(1)), &
                    block(sf),block(s))
            endif

         enddo

      enddo outer

      return
      end subroutine d2calcha ! of d2calcha

! ************************> SUBROUTINE D2CALCHA2<***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! D_MCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. Additionally calculates the correlated energy.
!
! Does not assume hermiticity of psi (i.e. of coefficient matrix).
! Used for SIL integrators.
! **********************************************************************

      subroutine d2calcha2 (psi,dtpsi,hteil,nham)

      implicit none

      integer(long)                                             :: m,k,k1,k2,si,sf,s,hzw,h
      integer(long), dimension(2)                               :: zw
      integer(long), intent(in)                                 :: nham
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)              :: hteil
      complex(dop), dimension(maxblock*maxblock,2)              :: workc


! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

outer:do k=k1,k2

!     Use only hamiltonian terms
!         if (kact(k) .ne. 1) goto 10
         if (kact(k) .ne. 1) cycle outer

         do s=1,nstate

            sf = kf(k)
            si = ki(k)
            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(si,s)),workc(1,zw(1)),&
                 block(si)*block(s))

            do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
               if (.not. diag(m,k)) then
                  call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)),&
                       workc(1,zw(2)),vdim(m,sf),dim(m,si),&
                       ndim(m,si)*block(s),dim(m,sf))
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo
            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)),block(sf)*block(s))
            endif
            call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(sf,s)),&
                    block(sf),block(s))

!     Note that Hteil is (dim(m,sf),dim(m,si))
            sf = ki(k)
            si = kf(k)
            zw(1) = 1
            zw(2) = 2

            call cpvxz(psi(dzpsi(s,si)),workc(1,zw(1)),&
                 block(s)*block(si))

            do m=1,nmode
!     Note that for si .ne. sf always diag(m,k)=.false.
               if (.not. diag(m,k)) then
                  if (si .eq. sf) then
                     call tqxazz(workc(1,zw(1)),hteil(pmat(m,k)),&
                          workc(1,zw(2)),block(s)*vdim(m,sf),&
                          dim(m,si),ndim(m,si))
                  else
                     if(.not.lompthread)then
                        call mttxzz(hteil(pmat(m,k)),workc(1,zw(1)),&
                             workc(1,zw(2)),block(s)*vdim(m,sf),&
                             dim(m,si),ndim(m,si),dim(m,sf))
                     else
                        call mttxzz(hteil(pmat(m,k)),workc(1,zw(1)),&
                             workc(1,zw(2)),block(s)*vdim(m,sf),&
                             dim(m,si),ndim(m,si),dim(m,sf))
                     endif
                  endif
                  hzw = zw(1)
                  zw(1) = zw(2)
                  zw(2) = hzw
               endif
            enddo
            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)),block(s)*block(sf))
            endif

            if (lrlx) then
               call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(s,sf)),&
                    block(s),block(sf))
            else
               call submxxzo1(dtpsi(dzpsi(s,sf)),workc(1,zw(1)),&
                    block(s),block(sf))
            endif
         enddo

! 10      continue
      enddo outer

      return
      end subroutine d2calcha2! of d2calcha2


! ************************> SUBROUTINE D2CALCHALOCH2<***********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! D_MCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. Additionally calculates the correlated energy.
!
! It additionally makes a hole.
!
! Does not assume hermiticity of psi (i.e. of coefficient matrix).
! Used for SIL integrator. 
!
! loch=0: no hole
! loch>0: hole
!
! side=0: left side operation
! side=1: right side operation
!
! nham: acting operator
!
! 11/2001 CC
!
! **********************************************************************
      subroutine d2calchaloch2 (psi,dtpsi,hteil,loch,side,nham,k0)

      implicit none

      integer(long)                                             :: m,k,si,sf,s,hzw,h,k1,k2
      integer(long), dimension(2)                               :: zw
      integer(long), intent(in)                                 :: loch,side,nham
      integer(long), intent(in), optional                       :: k0
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)              :: hteil
      complex(dop), dimension(maxblock*maxblock,2)              :: workc

      routine = 'd2calcaloch2'

      k1 = zham(nham)
      k2 = zham(nham)+kzahl(nham)-1

      if (present(k0) .and. k0 .gt. 0) then
         k1 = k0
         k2 = k0
      endif

! --- LOOP OVER (EACH) CORRELATED HAMILTONIAN TERM ---
outer:do k=k1,k2

!     Use only hamiltonian terms
         if (kact(k) .ne. 1) cycle outer

         do s=1,nstate

            sf = kf(k)
            si = ki(k)
            zw(1) = 1
            zw(2) = 2
            call cpvxz(psi(dzpsi(si,s)),workc(1,zw(1)), &
                 block(si)*block(s))

            do m=1,nmode
               if (m.ne.loch) then
!                  if (htmtype(hamilton(m,k)).ne.1) then 
!     Note that for si .ne. sf always diag(m,k)=.false.
                     if (.not. diag(m,k)) then
                        if (side.eq.2) then
                           call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                                workc(1,zw(2)), &
                                block(s)*vdim(m,sf),dim(m,si), &
                                ndim(m,si),dim(m,sf))
                        else if (side.eq.1) then ! right side operation
                           call mtxxzz(hteil(pmat(m,k)),workc(1,zw(1)), &
                                workc(1,zw(2)),vdim(m,sf),dim(m,si), &
                                ndim(m,si)*block(s),dim(m,sf))
                        else
                           message = 'wrong side integer!'
                           call errormsg
                        endif
                        hzw = zw(1)
                        zw(1) = zw(2)
                        zw(2) = hzw
                     endif
                  endif
!               endif
            enddo


            h=hamilton(fkoe,k)
            if (htmtype(h) .ne. 1) then
               call xvxxzzo(koeff(k),workc(1,zw(1)),block(sf)*block(s))
            endif

            call addmxxzo(workc(1,zw(1)),dtpsi(dzpsi(sf,s)), &
                  block(sf),block(s))

         enddo

! 10      continue
      enddo outer

      return
      end subroutine d2calchaloch2

! ************************> SUBROUTINE D2CALCHALOCH2_ML <********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! D_MCTDH-coefficients, the result is returned as a sum over all 
! terms in dtpsi. Additionally calculates the correlated energy.
!
! It additionally makes a hole.
!
! Does not assume hermiticity of psi (i.e. of coefficient matrix).
! Multi-layer version
!
! loch=0: no hole
! loch>0: hole
!
! side=0: left side operation
! side=1: right side operation
!
! nham: acting operator
!
! 11/2001 CC
!
! **********************************************************************
      subroutine d2calchaloch2_ml (psi,dtpsi,loch,side,nham)

      implicit none

      integer(long)                                             :: k,h,k1,k2
      integer(long), intent(in)                                 :: loch,side,nham
      complex(dop), dimension(adim), intent(in)                 :: psi
      complex(dop), dimension(adim), intent(inout)              :: dtpsi
      complex(dop), dimension(adim)                             :: work
      type(hterm_tp),pointer       :: hk(:)


      routine = 'd2calcaloch2'

      k1 = zham(nham)
      k2 = zham(nham)+kzahl(nham)-1

! --- LOOP OVER (EACH) CORRELATED HAMILTONIAN TERM ---
outer:do k=k1,k2

!     Use only hamiltonian terms
         if (kact(k) .ne. 1) cycle outer

         hk=>tree%hterms(k,:)
         work = psi
         call hprod_phi2(hk,work,loch,side)

         h=hamilton(fkoe,k)
         if (htmtype(h) .ne. 1) then
            call xvxxzzo(koeff(k),work,block(1)*block(1))
         endif

         call addmxxzo(work,dtpsi,block(1),block(1))

      enddo outer

      return
      end subroutine d2calchaloch2_ml

! ***********************> SUBROUTINE D2CALCHUNA<***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients. Note that the coefficients are already included
! in hteil. Additionally calculates the uncorrelated energy.
!
! Does not assume hermiticity of psi (i.e. the coefficent matrix).
! **********************************************************************

      subroutine d2calchuna (psi,dtpsi,hteil,nham)

      implicit none

      integer(long)                                           :: m,s1,s2,k
      integer(long), intent(in)                               :: nham
      complex(dop), dimension(adim), intent(in)               :: psi
      complex(dop), dimension(adim), intent(inout)            :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)            :: hteil
      complex(dop), dimension(maxblock*maxblock)              :: workc


! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      k=zham(nham)+kzahl(nham)-1

      do s2 = 1,nstate
         do s1 = 1,nstate

            if(khzahl(1,s1,nham) .gt. 0) then
               call qtxxzz(hteil(pmat(1,k+s1)),psi(dzpsi(s1,s2)), &
                    workc,vdim(1,s1),dim(1,s1),ndim(1,s1)*block(s2))
            else
               call zerovxz(workc,block(s1)*block(s2))
            endif
            do m = 2,nmode
               if(khzahl(m,s1,nham) .gt. 0) then
                  call qtxxzza(hteil(pmat(m,k+s1)), &
                       psi(dzpsi(s1,s2)),workc,vdim(m,s1),dim(m,s1), &
                       ndim(m,s1)*block(s2))
               endif
            enddo

            call addmxxzo(workc,dtpsi(dzpsi(s1,s2)), &
                 block(s1),block(s2))

!     Calculate uncorrelated energy
            if (s1 .eq. s2) then
               call trhxz(workc,deuncorr(s1),block(s1))
            endif

            if (lrlx) then
               call addmxazo1(dtpsi(dzpsi(s2,s1)),workc, &
                    block(s1),block(s2))
            else
               call submxazo1(dtpsi(dzpsi(s2,s1)),workc, &
                    block(s1),block(s2))
            endif
         enddo
      enddo

      return
      end subroutine d2calchuna! of d2calchuna



! ***********************> SUBROUTINE D2CALCHUNA2<***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients. Note that the coefficients are already included
! in hteil. Additionally calculates the uncorrelated energy.
!
! Does not assume hermiticity of psi (i.e. the coefficient matrix).
! **********************************************************************

      subroutine d2calchuna2 (psi,dtpsi,hteil,nham)

      implicit none

      integer(long)                                           :: m,s1,s2,k
      integer(long), intent(in)                               :: nham
      complex(dop), dimension(adim), intent(in)               :: psi
      complex(dop), dimension(adim), intent(inout)            :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)            :: hteil
      complex(dop), dimension(maxblock*maxblock)              :: workc


! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---
      k=zham(nham)+kzahl(nham)-1

      do s2 = 1,nstate
         do s1 = 1,nstate

            if(khzahl(1,s1,nham).gt.0) then
               call qtxxzz(hteil(pmat(1,k+s1)),psi(dzpsi(s1,s2)),&
                    workc,vdim(1,s1),dim(1,s1),ndim(1,s1)*block(s2))
            else
               call zerovxz(workc,block(s1)*block(s2))
            endif
            do m = 2,nmode
              if(khzahl(m,s1,nham) .gt. 0)then
                 call qtxxzza(hteil(pmat(m,k+s1)),&
                      psi(dzpsi(s1,s2)),workc,vdim(m,s1),dim(m,s1),&
                      ndim(m,s1)*block(s2))
              endif
            enddo
            call addmxxzo(workc,dtpsi(dzpsi(s1,s2)),&
                 block(s1),block(s2))

            if(khzahl(1,s2,nham) .gt. 0) then
               call tqxazz(psi(dzpsi(s1,s2)),hteil(pmat(1,k+s2)),&
                    workc,block(s1)*vdim(1,s2),dim(1,s2),ndim(1,s2))
            else
               call zerovxz(workc,block(s1)*block(s2))
            endif

            do m = 2,nmode
              if(khzahl(m,s2,nham) .gt. 0)then
                 call tqxazza(psi(dzpsi(s1,s2)),&
                      hteil(pmat(m,k+s2)),workc,&
                      block(s1)*vdim(m,s2),dim(m,s2),ndim(m,s2))
              endif
            enddo


            if (lrlx) then
               call addmxxzo(workc,dtpsi(dzpsi(s1,s2)),&
                    block(s1),block(s2))
            else
               call submxxzo1(dtpsi(dzpsi(s1,s2)),workc,&
                    block(s1),block(s2))
            endif


         enddo
      enddo

      return
      end subroutine d2calchuna2 ! d2calchuna2


! ***********************> SUBROUTINE D2CALCHUNALOCH2<***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients. Note that the coefficients are already included
! in hteil. Additionally calculates the uncorrelated energy.
!
! It additionally makes a hole.
!
! Does not assumes hermiticity of psi (i.e. the coefficient matrix).
! **********************************************************************

      subroutine d2calchunaloch2 (psi,dtpsi,hteil,loch,nham)

      implicit none

      integer(long)                                           :: m,s1,s2,k
      integer(long), intent(in)                               :: loch,nham
      complex(dop), dimension(adim), intent(in)               :: psi
      complex(dop), dimension(adim), intent(inout)            :: dtpsi
      complex(dop), dimension(pmatdim), intent(in)            :: hteil
      complex(dop), dimension(maxblock*maxblock)              :: workc


! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---
      k=zham(nham)+kzahl(nham)-1
      do s2 = 1,nstate
         do s1 = 1,nstate

            if ((loch.ne.1).or.(loch.eq.0)) then
               if(khzahl(1,s1,nham) .gt. 0) then
                  call qtxxzz(hteil(pmat(1,k+s1)),psi(dzpsi(s1,s2)), &
                       workc,vdim(1,s1),dim(1,s1),ndim(1,s1)*block(s2))
               else
                  call zerovxz(workc,block(s1)*block(s2))
               endif
               do m = 2,nmode
                  if (m.ne.loch .and. khzahl(m,s1,nham) .gt. 0) then
                     call qtxxzza(hteil(pmat(m,k+s1)), &
                          psi(dzpsi(s1,s2)),workc,vdim(m,s1),dim(m,s1), &
                          ndim(m,s1)*block(s2))
                  endif
               enddo
            else
               if(khzahl(2,s1,nham) .gt. 0) then
                  call qtxxzz(hteil(pmat(2,k+s1)),psi(dzpsi(s1,s2)), &
                       workc,vdim(2,s1),dim(2,s1),ndim(2,s1)*block(s2))
               else
                  call zerovxz(workc,block(s1)*block(s2))
               endif
               do m = 3,nmode
                  if(khzahl(m,s1,nham) .gt. 0) then
                     call qtxxzza(hteil(pmat(m,k+s1)), &
                          psi(dzpsi(s1,s2)),workc,vdim(m,s1),dim(m,s1), &
                          ndim(m,s1)*block(s2))
                  endif
               enddo
            endif
            call addmxxzo(workc,dtpsi(dzpsi(s1,s2)), &
                 block(s1),block(s2))

         enddo
      enddo

      return
      end subroutine d2calchunaloch2

! ***********************> SUBROUTINE D2CALCHUNALOCH2<***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~~
! Calculates the action of the uncorrelated Hamiltonian on the
! MCTDH-coefficients. Note that the coefficients are already included
! in hteil. Additionally calculates the uncorrelated energy.
!
! Multi-layer version.
!
! It additionally makes a hole.
!
! Does not assumes hermiticity of psi (i.e. the coefficient matrix).
! **********************************************************************

      subroutine d2calchunaloch2_ml (psi,dtpsi,loch,nham)

      implicit none

      integer(long)                                           :: m,k
      integer(long), intent(in)                               :: loch,nham
      complex(dop), dimension(adim), intent(in)               :: psi
      complex(dop), dimension(adim), intent(inout)            :: dtpsi
      complex(dop), dimension(adim)                           :: work 
      type(hterm_tp),pointer       :: hk(:)

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---
      k=zham(nham)+kzahl(nham)

      hk=>tree%hterms(k,:)
      work = 0.0_dop
      do m = 1,tree%nmodes
         if (khzahl(m,1,nham) .eq. 0) cycle
         if (m .eq. loch) cycle

         work = psi
         call hunprod_phi2(hk,work,m)
 
         call addmxxzo(work,dtpsi,block(1),block(1))
      enddo


      return
      end subroutine d2calchunaloch2_ml


! **********************************************************************
      end module d2calchamod


