!#######################################################################
!
! MCTDH-module DENSITY
!
! Calculates the reduced density matrices, their inverse matrices and
!  their eigenvalues and eigenvectors.
!
! density: passing routine
! denmat: calculates the reduced density matrices
! matinv: inverts a matrix
!
! contains: density,denmat,matinv,denmatgwp
!#######################################################################

      module densitymod

use decimal, only: dop, long
use global
use logdat
use timing
use mtlib,   only: qtxxzz
use rmlib,   only: rm1hxxxzz, rmhxxxzz
use op1lib,  only: trhxz, cpvxz, zeromxz, unitqxz, normvxz
use eqofmotion1, only: compute_dmat_tree
use openmpmod

use psidef
use griddatmod
use maxv,     only: maxdim,maxsta
use ciselect, only: venpsi
use lalib,    only: tranqxtzz
use runpropmod
use d2tens,     only: rm1_mf, rm2_mf, rm2_mfa, filldens

      implicit none
      private
      public :: density,density0,density1,density2,denmat,chngphs
      contains

!#######################################################################

      subroutine density(psi,tree,dicht1,dicht2,dicht3,dicht4,jindx,&
                 gs2,psi1)

      implicit none

      integer(long), dimension(jindxdim), intent(in)  :: jindx
      real(dop), dimension(d3matdim), intent(out)     :: dicht3
      complex(dop), dimension(dgldim), intent(in)     :: psi
      complex(dop), dimension(dgldim), intent(in), optional :: psi1
      complex(dop), dimension(dmatdim), intent(out)   :: dicht1,dicht2,&
                                                         dicht4
      complex(dop), dimension(gs2matdim), intent(in)  :: gs2
      type(wfnode_t),intent(inout) :: tree

      if (lrunml) then
         call compute_dmat_tree(tree,psi)
      else if (dentype .eq. 0) then
         call density0(psi,dicht1,dicht2,dicht3,dicht4,jindx,gs2,psi1)
      else if (dentype .eq. 1) then
         call density1(psi,dicht1,dicht2,dicht3,dicht4)
      else if (dentype .eq. 2) then
         call density2(psi,dicht1,dicht2,dicht3,dicht4,.false.)
      endif

      return
      end subroutine density

!#######################################################################
! DENSITY
!
! Calculates the reduced density matrix etc. for a single mode.
!
!  psi: wavefunction coefficients
!  dicht1: inverse density matrix
!  dicht2: reduced density matrix
!  dicht3: eigenvalues and regularised eigenvalues of dicht2
!  dicht4: eigenvectors of dicht2
!  dim: number of single particle functions
!  vdim: product of number of single particle functions of proceeding
!        modes
!  ndim: product of number of single particle functions of following
!        modes
!

! V6.3 MB
!#######################################################################

      subroutine density0(psi,dicht1,dicht2,dicht3,dicht4,jindx,gs2,psi1)

      implicit none

      integer(long)                                   :: s,m,iblock,&
                                                         zeig1,zeig2,zeig3,&
                                                         m1,f1
      integer(long), dimension(jindxdim), intent(in)  :: jindx
      integer(long), dimension(maxblock)              :: worki

      real(dop)                                       :: epsrel,norm,trace
      real(dop), dimension(d3matdim), intent(out)     :: dicht3
      complex(dop), dimension(dgldim), intent(in)     :: psi
      complex(dop), dimension(dgldim), intent(in), optional     :: psi1
      complex(dop), dimension(dmatdim), intent(out)   :: dicht1,dicht2,dicht4
      complex(dop), dimension(2*maxblock)             :: workc
      complex(dop), dimension(adim)                   :: psilhs
      complex(dop), dimension(gs2matdim), intent(in)  :: gs2
      integer(long), save                             :: tid=0 ! timer ID

!-----------------------------------------------------------------------
! Allow for calculation of density matrix between different wavefunctions
!-----------------------------------------------------------------------
      if (present(psi1)) then
         psilhs(1:adim) = psi1(1:adim)
      else
         psilhs(1:adim) = psi(1:adim)
      endif

!-----------------------------------------------------------------------
! For ML-MCTDH use different routine
!-----------------------------------------------------------------------
! cash lrunml edit
!      if (lrunml) then
!         call compute_dmat_tree(tree,psi)
!         return
!      endif

      call get_timer(tid, 'Density')
      call start_timer(tid)

!-----------------------------------------------------------------------
! calculate all density matrices
!-----------------------------------------------------------------------
      zeig1=1
      zeig2=maxblock+1
      do s=1,nstate
         do m=1,nmode
            if(idmode(m).eq.0) then
               if (citype .gt. 0) then
                  iblock=vdim(m,s)*dim(m,s)*ndim(m,s)
                  call venpsi(m,jindx(zpsi(s)),worki,jindx(zpsi(s)), &
                       block(s),iblock,block(s), &
                       nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                       workc(zeig1),psi(zpsi(s)))
                  call rm1hxxxzz(workc(zeig1),dicht2(dmat(m,s)), &
                       vdim(m,s),dim(m,s),ndim(m,s))
               else if (abs(psitype) .eq. 4) then
                  call cpvxz(psi(zpsi(s)),workc(zeig1),block(s))
                  do m1=1,nmode
                     f1 = spfdof(1,m1)
                     if (m1 .ne. m .and. gwpm(m1)) then
                        zeig3=zeig1
                        zeig1=zeig2
                        zeig2=zeig3
                        call qtxxzz(gs2(gs2mat(m1,s)),workc(zeig2), &
                             workc(zeig1),vdim(m1,s),dim(m1,s), &
                             ndim(m1,s))
                     endif
                  enddo
                  call rmhxxxzz(psilhs(zpsi(s)),workc(zeig1), &
                       dicht2(dmat(m,s)),vdim(m,s),dim(m,s), &
                       ndim(m,s))

               else 
                   call rmhxxxzz(psilhs(zpsi(s)),psi(zpsi(s)),&
                        dicht2(dmat(m,s)),vdim(m,s),dim(m,s),ndim(m,s))
               endif
            endif

         enddo
      enddo

!-----------------------------------------------------------------------
! compute norm and take a relative regularization parameter
!-----------------------------------------------------------------------
      norm=0.0_dop
      do s=1,nstate
         call trhxz(dicht2(dmat(1,s)),trace,dim(1,s))
         norm=norm+trace
      enddo

      epsrel=norm*epsmat
      if (lsvd)  epsrel=0.d0
      if (lmulpack) epsrel=epsrel/dble(npacket)
      if (mpb.gt.0) epsrel=epsrel/dble(npackts)
      if (lreflex)  epsrel=epsrel/dble(2.d0)

!-----------------------------------------------------------------------
! regularize and invert density matrices for each mode and state
! neither the inverse nor the EVs of the density matrix
! will be needed if psitype == -4, 4. Instead store the diagonal
! elements of the density matrix in dicht3
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (idmode(m) .eq. 0) then
               if (gwpm(m) .and. (abs(psitype) .eq. 4)) then
                  call denmatgwp(dicht2(dmat(m,s)), &
                       dicht3(d3mat(m,s)),dicht4(dmat(m,s)),dim(m,s))
               else 
                  call denmat(dicht1(dmat(m,s)),dicht2(dmat(m,s)), &
                       dicht3(d3mat(m,s)),dicht4(dmat(m,s)),dim(m,s), &
                       epsrel)
               endif
            endif

            call chngphs( dicht4(dmat(m,s)), dim(m,s) )

         enddo
      enddo

      call stop_timer(tid)

      return
      end subroutine density0
      
!#######################################################################
!
! DENSITY 1 (fro pMCTDH(I))
!
!#######################################################################

      subroutine density1(psi,dicht1,dicht2,dicht3,dicht4)

      implicit none

      integer(long)                                   :: s,t,m
      real(dop)                                       :: epsrel,norm,trace
      real(dop), dimension(d3matdim), intent(out)     :: dicht3
      complex(dop), dimension(dgldim), intent(in)     :: psi
      complex(dop), dimension(dmatdim), intent(out)   :: dicht1,dicht2,dicht4
      integer(long)                                   :: tid=0 ! timer ID
 
      call get_timer(tid, 'D1density')
      call start_timer(tid)

!-----------------------------------------------------------------------
! calculate all density matrices
!-----------------------------------------------------------------------
      do s=1,nstate
         do t=1,nstate
            do m=1,nmode
               call rm1hxxxzz(psi(dzpsi(s,t)),dicht2(ddmat(m,s,t)), &
                    vddim(m,s,t),ddim(m,s,t),nddim(m,s,t))
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! compute norm and take a relative regularization parameter
!-----------------------------------------------------------------------
      norm=0.0_dop
      do s=1,nstate
         do t=1,nstate
            call trhxz(dicht2(ddmat(1,s,t)),trace,ddim(1,s,t))
            norm=norm+trace
         enddo
      enddo
      epsrel=norm*epsmat
!      if (lmulpack) epsrel=epsrel/dble(npacket)

!-----------------------------------------------------------------------
! regularize and invert density matrices for each mode and state
!-----------------------------------------------------------------------
      do s=1,nstate
         do t=1,nstate
            do m=1,nmode
               call denmat(dicht1(ddmat(m,s,t)),dicht2(ddmat(m,s,t)),&
                    dicht3(dd3mat(m,s,t)),dicht4(ddmat(m,s,t)),&
                    ddim(m,s,t),epsrel)
            enddo
         enddo
      enddo

      call stop_timer(tid)

      return
      end subroutine density1

!#######################################################################
!
! DENSITY2 (pMCTDH(II))
!
!#######################################################################

      subroutine density2(psi,dicht1,dicht2,dicht3,dicht4,lflag)

      implicit none

      integer(long)                                   :: s,t,m
      real(dop)                                       :: epsrel,norm
      real(dop), dimension(d3matdim), intent(out)     :: dicht3
      complex(dop), dimension(dgldim), intent(in)     :: psi
      complex(dop), dimension(dmatdim), intent(out)   :: dicht1,dicht2,dicht4
      logical(kind=4), intent(in)                     :: lflag
      integer(long)                                   :: tid=0 ! timer ID

      call get_timer(tid, 'D2density')
      call start_timer(tid)

!-----------------------------------------------------------------------
! calculate all density matrices
!-----------------------------------------------------------------------
      do s=1,nstate
         if (lmf.or.lflag) then
!     mean-field approach
            do m=1,nmode
               call rm1_mf(psi(dzpsi(s,s)),dicht2(dmat(m,s)),vdim(m,s),&
                    dim(m,s),ndim(m,s))
            enddo
         else
!     Dirac-Frenkel variational principle
            do m=1,nmode
               call rm2_mf(psi(dzpsi(s,s)),dicht2(dmat(m,s)),&
                    vdim(m,s),dim(m,s),ndim(m,s),block(s))
               do t=1,nstate
                  if (t .eq. s) cycle
                  call rm2_mfa(psi(dzpsi(t,s)),psi(dzpsi(t,s)),&
                       dicht2(dmat(m,s)),vdim(m,s),dim(m,s),&
                       ndim(m,s),block(t))
               enddo
               call filldens(dicht2(dmat(m,s)),dim(m,s))
            enddo
         endif
      enddo


!-----------------------------------------------------------------------
! compute norm and take a relative regularization parameter
!-----------------------------------------------------------------------
      call normvxz(psi,norm,adim)
      epsrel=norm*epsmat
      if (lmulpack) epsrel=epsrel/dble(npacket)

!-----------------------------------------------------------------------
! regularize and invert density matrices for each mode and state
!-----------------------------------------------------------------------
      do s=1,nstate
        do m=1,nmode
          call denmat(dicht1(dmat(m,s)),dicht2(dmat(m,s)),&
               dicht3(d3mat(m,s)),dicht4(dmat(m,s)),dim(m,s),&
               epsrel)
        enddo
      enddo

      call stop_timer(tid)

      return
      end subroutine density2

!#######################################################################
! DENMAT
!
! called by density
! calls to matinv
!
! Calculates the inverse reduced density matrix for a single mode. 
!
!  dicht1: inverse density matrix
!  dicht2: reduced density matrix
!  dicht3: eigenvalues and regularised eigenvalues of dicht2
!  dicht4: eigenvectors of dicht2
!  dim1: number of single particle functions
!
!#######################################################################

      subroutine denmat(dicht1,dicht2,dicht3,dicht4,dim1,epsrel)

      implicit none
      
      integer(long)                                     :: e,e1
      integer(long), intent(in)                         :: dim1
      complex(dop), dimension(dim1,dim1), intent(in)    :: dicht2
      complex(dop), dimension(dim1,dim1), intent(out)   :: dicht1,dicht4
      real(dop), intent(in)                             :: epsrel
      real(dop), dimension(dim1,2), intent(out)         :: dicht3
      
!-----------------------------------------------------------------------
! calculate regularized inverse density matrix
! Notice that negative of dicht2 is inverted, so that eigenvalues are
! ordered with largest at dicht3(1)
!-----------------------------------------------------------------------
      do e=1,dim1
         do e1=1,dim1
            dicht4(e1,e)=-dicht2(e1,e)
         enddo
      enddo

      call matinv(dicht1,dicht3,dicht4,dim1,epsrel)

      return
      end subroutine denmat

!#######################################################################
!
! MATINV
!
! called by denmat
! calls to zheev 
!
! calls diagonalisation routines for density matrix and from the
!  eigenvalues (regularised) and eigenvectors so obtained calculates 
!  the inverse density matrix
! 
! invers: returned as inverted matrix, used as scratch array.
! eval:   eigenvalues of passed matrix
! evek:   eigenvectors of passed matrix; on input: Matrix to be inverted.
! dim:    dimensions of matrices e.g. invers(dim,dim)
! work:   complex workspace         
!
! V5.0 MB
!#######################################################################
  
      subroutine matinv(invers,eval,evek,dim,epsrel)

      implicit none

      integer(long)                                   :: error,e,e1,e2
      integer(long), intent(in)                       :: dim
      complex(dop), dimension(dim,dim), intent(out)   :: invers
      complex(dop), dimension(dim,dim) , intent(inout):: evek
      real(dop), intent(in)                           :: epsrel
      real(dop), dimension(dim,2), intent(out)        :: eval
      real(dop), dimension(3*dim)                     :: work
      character(len=4)                                :: buf

!-----------------------------------------------------------------------
! diagonalize matrix
!-----------------------------------------------------------------------
      e2 = dim*dim
      call zheev('V','U',dim,evek,dim,eval,invers,e2,work,error)
      if (error .ne. 0) then
         write(buf,'(i4)') error
         routine = 'matinv'
         message = 'Cannot diagonalize density matrix'//&
              " Error No.: "//buf
         call errormsg
      endif
!-----------------------------------------------------------------------
! take negative of eigenvalue (due to re-ordering) and regularize.
! The if-statement is introduced to avoid an underflow.
!-----------------------------------------------------------------------
      do e=1,dim
         eval(e,1)=-eval(e,1)
         if( eval(e,1) .lt. 64.0_dop*epsrel) then
            eval(e,2)=eval(e,1)+epsrel*exp(-eval(e,1)/epsrel)
         else
            eval(e,2)=eval(e,1)
         endif
      enddo

!-----------------------------------------------------------------------
! calculate regularized inverse matrix
!-----------------------------------------------------------------------
      call zeromxz(invers,dim,dim)
      do e=1,dim
        do e1=1,dim
          do e2=1,dim
             invers(e1,e)=invers(e1,e) + dconjg(evek(e,e2))*evek(e1,e2)/eval(e2,2)
          enddo
        enddo
        invers(e,e)=dble(invers(e,e))
      enddo

      return
      end subroutine matinv

!#######################################################################

      subroutine denmatgwp(dicht2,dicht3,dicht4,dim1)

      integer, intent(in) :: dim1
      integer             :: e
      complex(dop), dimension(dim1,dim1), intent(in)  :: dicht2
      complex(dop), dimension(dim1,dim1), intent(out) :: dicht4
      real(dop), dimension(dim1), intent(out)         :: dicht3

      do e=1,dim1
         dicht3(e) = dble(dicht2(e,e))
      enddo

      call unitqxz(dicht4,dim1)

      end subroutine denmatgwp

!#######################################################################

      subroutine chngphs( evek, dim )

      implicit none
      integer(long)                                   :: e1, e2
      integer(long), intent(in)                       :: dim
      complex(dop)                                    :: factor
      complex(dop), dimension(dim,dim), intent(inout) :: evek

!-----------------------------------------------------------------------
! Use the phase-convention: evek(e,e).gt.0, i.e. <phi_old|phi_new>.gt.0
!-----------------------------------------------------------------------
      do e2=1,dim
         if( abs(evek(e2,e2)) .gt. 1.0e-9_dop ) then
            factor = abs(evek(e2,e2))/evek(e2,e2)
            do e1=1,dim
               evek(e1,e2) = evek(e1,e2)*factor
            enddo
         endif
      enddo

      return
      end subroutine chngphs


!-----------------------------------------------------------------------
      end module densitymod

