! **********************************************************************
      module evaluate

use decimal, only: long,dop
use maxv
use global
use globmemmod
use dvrdatmod
use griddatmod
use psidef
use operdef
use runpropmod
use logdat

use xvlib,  only: xvxxdzr, xvxxzzo, xmxxzzo, vvaxzz, xvxxzz
use op1lib, only: sum2vxz, trhxz, cpvxz
use op2lib, only: subvxxzo1, addvxxzo
use mtlib,  only: qtxxzz

      implicit none
      private
      public :: prefactor,calcenergy,subenergy,subenergy_ex,calcenergy_ex,&
                getprefactor,calcenergy_g,calcenergy_gs
      contains

! **********************************************************************
!                                                                       
!                           EVALUATION (evaluation.f)                   
!                                                                       
!   MCTDH module containing (auxiliary) subroutines to evaluate the     
!   right hand side of the MCTDH working equations.                     
!                                                                       
!   Contains:                                                           
!     prefactor:    Multiplies the wavefunction derivative with -i for  
!                   propagation or -1 for imaginary time relaxation.    
!     subenergy:    Subtracts the energy from the Hamiltonian, i. e.    
!                   H|psi> is transformed to (H-E)|psi> where E =       
!                   <psi|H|psi>/<psi|psi> (needed for relaxation).      
!                                                                       
! **********************************************************************
! *** V6.0 MB ************> SUBROUTINE PREFACTOR <**********************
!                          ~~~~~~~~~~~~~~~~~~~~~~
! Multiplies wavefunction derivative with -i for normal propagation
! or -1 for imaginary time relaxation.
! **********************************************************************

      subroutine prefactor (dtpsi,dtpsidim)

      implicit none

      integer(long)                                    :: f,m,s
      integer(long), intent(in)                        :: dtpsidim
      complex(dop)                                     :: factor
      complex(dop), dimension(dtpsidim), intent(inout) :: dtpsi


! --- MULTIPLY DTPSI WITH PREFACTOR ---

      if (lfuncnopre) then
         return                                  ! No prefactor :Hpsi

      else if (ldtpsi2hpsi) then
         factor = (0.0_dop,1.0_dop)             ! convert dtpsi to Hpsi
         call xvxxzzo(factor,dtpsi,dtpsidim)

      else if (lrlx) then
         if(inttypea.eq.6 .and. lcap) then  ! complex DAV
            factor = (-0.707107_dop,-0.707107_dop)
         else
            factor = (-1.0_dop,0.0_dop)          ! Relaxation
         endif
         call xvxxzzo(factor,dtpsi,dtpsidim)

      else if (lxloc) then
         factor = (0.0_dop,-1.0_dop)
         call xvxxzzo(factor,dtpsi,adim)
         do s=1,nstate
            do m=1,nmode
               f=spfdof(1,m)
               if (basis(f) .ne. 26 .and. basis(f) .ne. 27) then
                  call xmxxzzo(factor,dtpsi(zetf(m,s)),subdim(m), &
                        dim(m,s))
               endif
            enddo
         enddo

      else 
         factor = (0.0_dop,-1.0_dop)             ! Propagation
         call xvxxzzo(factor,dtpsi,dtpsidim)
      endif

      return
      end subroutine prefactor

! ************************> FUNCTION GETPREFACTOR <**********************
!                          ~~~~~~~~~~~~~~~~~~~~~~

! Return the prefactor to be applied to dtpsi, -i for normal propagation
! or -1 for imaginary time relaxation.

      function getprefactor() result(factor)
        
        implicit none
        
        complex*16 factor
        
        if (lrlx) then
           if(inttypea.eq.6 .and. lcap) then  ! complex DAV
              factor = (-0.707107d0,-0.707107d0)
           else
              factor = (-1.0d0,0.0d0)          ! Relaxation
           endif
        else
           factor = (0.0d0,-1.0d0)             ! Propagation
        endif
        
        return
      end function getprefactor

! **********************************************************************
! *** V6.0 MB ************> SUBROUTINE SUBENERGY <**********************
!                          ~~~~~~~~~~~~~~~~~~~~~~
! Subtracts the energy from the Hamiltonian, i. e. H|psi> is transformed
! to (H-E)|psi> where E = <psi|H|psi>/<psi|psi> (needed for relaxation).
!
! energy is first calculated, e.g. by calcenergy
! orthonormality of spfs assumed
! **********************************************************************

      subroutine subenergy (energy,psi,dtpsi)

      implicit none

      complex(dop)                                 :: ctmp
      complex(dop), dimension(adim), intent(in)    :: psi
      complex(dop), dimension(adim), intent(inout) :: dtpsi
      real(dop)                                    :: norm2,factor
      real(dop), intent(in)                        :: energy

! --- INITIALISE VARIABLES ---

      norm2  = 0.0_dop

! --- If SIL integrator not used, correct energy for norm

      if (inttypea .eq. 3 .or. inttypea.eq. 6) then
         factor = energy

      else
         call sum2vxz(psi,ctmp,adim)
         norm2=dble(ctmp)

         factor = energy/norm2

      endif

! --- SUBTRACT ENERGY TIMES PSI FROM TIME DERIVATIVE ---

      call xvxxdzr(factor,psi,dtpsi,adim)

      return
      end subroutine subenergy

! **********************************************************************
! *** V8.0 MB ***********> SUBROUTINE SUBENERGY_EX <********************
!                         ~~~~~~~~~~~~~~~~~~~~~~~~~
! Subtracts the energy from the Hamiltonian, i. e. H|psi> is transformed
! to (H-E)|psi> where E = <psi|H|psi>/<psi|psi> (needed for numerically
! exact relaxation).
! **********************************************************************

      subroutine subenergy_ex (energy,psi,dtpsi)

      implicit none

      integer(long)                                                 :: s,d
      real(dop)                                                     :: norm2,factor
      real(dop), intent(in)                                         :: energy
      complex(dop), dimension(adim+1:adim+totphidim), intent(in)    :: psi
      complex(dop), dimension(adim+1:adim+totphidim), intent(inout) :: dtpsi

! --- INITIALISE VARIABLES ---

      norm2  = 0.0_dop

! --- COMPUTE <PSI|PSI> AND <PSI|H|PSI> (SUMMED OVER ALL STATES) ---

      do s = 1,nstate
         do d = zetf(1,s),zetf(1,s)+subdim(1)-1
            norm2  = norm2+dble(psi(d))**2+dimag(psi(d))**2
         enddo
      enddo

! --- SUBTRACT ENERGY TIMES PSI FROM TIME DERIVATIVE ---

      factor = energy/norm2
      call xvxxdzr(factor,psi,dtpsi,totphidim)

      return
      end subroutine subenergy_ex

! **********************************************************************
! *                                                                    *
! *                          SUBROUTINE CALCENERGY                     *
! *                                                                    *
! * Computes the energy of the wavefunction (single-particle functions *
! * are assumed to be orthonormal).                                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi: The MCTDH-coefficients vector.                              *
! *   dtpsi: The action of the Hamiltonian on the MCTDH-coefficients.  *
! *                                                                    *
! * Output parameters:                                                 *
! *   energy: The energy of the wavefunction.                          *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! * NB "energy" is complex and is the energy multipled by the          *
! *    prefactor of dtpsi. E.g. for propagation                        *
! *    dtpsi = i psidot and for relaxation dtpsi = -p psidot           *
! **********************************************************************

      subroutine calcenergy (psi,dtpsi,energy,iprefactor)

      implicit none

      integer(long), intent(in)                 :: iprefactor
      real(dop), intent(out)                    :: energy
      complex(dop)                              :: ztmp,prefact
      complex(dop), dimension(adim), intent(in) :: psi,dtpsi

! --- COMPUTE <PSI|H|PSI> (SUMMED OVER ALL STATES) ---

      call vvaxzz(psi,dtpsi,ztmp,adim)

!-----------------------------------------------------------------------
! multiply by prefactor and store i S*Adot in tmp1
!-----------------------------------------------------------------------
      if (iprefactor .eq. 0) then
         continue
      else if (iprefactor .eq. 1) then
         if (lrlx) then
            prefact = (-1.0_dop,0.0_dop)
         else
            prefact = (0.0_dop,1.0_dop)
         endif
         ztmp = prefact*ztmp
      else 
         routine = 'Calcenergy_g (evaluate.f90)'
         write(message,'(a,i5)') 'Unknown prefactor : ',iprefactor
         call errormsg
      endif

      energy = dble(ztmp)

      return
      end  subroutine calcenergy

! **********************************************************************
! *                                                                    *
! *                        SUBROUTINE CALCENERGY_EX                    *
! *                                                                    *
! * Computes the energy of the wavefunction in a numerically exact     *
! * propagation.                                                       *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi: single-particle vector.                                     *
! *   dtpsi: action of the Hamiltonian on single-particle functions.   *
! *                                                                    *
! * Output parameters:                                                 *
! *   energy: The energy of the wavefunction.                          *
! *                                                                    *
! * V8.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine calcenergy_ex (psi,dtpsi,energy)

      implicit none

      integer(long)                                              :: s,d
      real(dop), intent(out)                                     :: energy
      complex(dop), dimension(adim+1:adim+totphidim), intent(in) :: psi,dtpsi

! --- INITIALISE VARIABLES ---

      energy = 0.0_dop

! --- COMPUTE <PSI|H|PSI> (SUMMED OVER ALL STATES) ---

      do s = 1,nstate
         do d = zetf(1,s),zetf(1,s)+subdim(1)-1
            energy = energy+dble (psi(d))*dble (dtpsi(d))&
                           +dimag(psi(d))*dimag(dtpsi(d))
         enddo
      enddo

      return
      end subroutine calcenergy_ex

!#######################################################################
! *                                                                    *
! *                          SUBROUTINE CALCENERGY_G                   *
! *                                                                    *
! * Computes the energy of the wavefunction  when GWP basis is used  s *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi: The MCTDH-coefficients vector.                              *
! *   dtpsi: The action of the Hamiltonian on the MCTDH-coefficients.  *
! *                                                                    *
! * Output parameters:                                                 *
! *   energy: The energy of the wavefunction.                          *
! *                                                                    *
! * If                                                                 *
! * iprefactor = 0, dtpsi = i adot (prop)  -adot (rlx)                 *
! *              1, dtpsi = adot                                       *
! *                                                                    *
!#######################################################################
            
      subroutine calcenergy_g(psi,dtpsi,energy,iprefactor)

      implicit none 
               
      integer(long), intent(in)                     :: iprefactor
      integer(long)                                 :: s
               
      real(dop), intent(out)                          :: energy
      real(dop)                                       :: energys
      complex(dop), dimension(adim), intent(in)       :: dtpsi,psi
               
      energy = 0.0_dop
      do s=1,nstate
         call calcenergy_gs(psi(zpsi(s)),dtpsi(zpsi(s)),energys,s,&
              block(s),iprefactor)
         energy = energy + energys
      enddo

      return
      end subroutine calcenergy_g
              
!#######################################################################

      subroutine calcenergy_gs(psi,dtpsi,energys,s,block1,iprefactor)

      implicit none 
               
      integer(long), intent(in)                     :: s,block1
      integer(long), intent(in)                     :: iprefactor
      integer(long)                                 :: m,m1
      integer(long)                                 :: zeig1,zeig2,zeig3
               
      real(dop), intent(out)                          :: energys
      complex(dop)                                    :: ztmp
      complex(dop), dimension(block1), intent(in)     :: dtpsi,psi
      complex(dop), dimension(2*block1)               :: workc
      complex(dop), dimension(block1)                 :: tmp1,tmp2
               
      workc=0.0_dop
      zeig1=1  
      zeig2=block1+1
      call cpvxz(dtpsi,workc(zeig2),block1)
      do m = 1,nmode
         if (.not. gwpm(m)) cycle
         zeig3=zeig1
         zeig1=zeig2
         zeig2=zeig3
         call qtxxzz(gs2(gs2mat(m,s)),workc(zeig1),workc(zeig2),&
           vdim(m,s),dim(m,s),ndim(m,s))
      enddo

!-----------------------------------------------------------------------
! multiply by prefactor and store i S*Adot in tmp1
!-----------------------------------------------------------------------
      if (iprefactor .eq. 0) then
         call cpvxz(workc(zeig2),tmp1,block1)
      else if (iprefactor .eq. 1) then
         if (lrlx) then
            ztmp = (-1.0_dop,0.0_dop)
         else
            ztmp = (0.0_dop,1.0_dop)
         endif
         call xvxxzz(ztmp,workc(zeig2),tmp1,block1)
      else 
         routine = 'Calcenergy_g (evaluate.f90)'
         write(message,'(a,i5)') 'Unknown prefactor : ',iprefactor
         call errormsg
      endif
     
!-----------------------------------------------------------------------
! store Gdot*A in tmp2
!-----------------------------------------------------------------------
      workc=0.0_dop
      tmp2 = 0.0_dop
      if (.not. lBEL) then
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            call qtxxzz(gsdot(gsmat(m,s)),psi,workc(zeig2),&
                 vdim(m,s),dim(m,s),ndim(m,s))
            do m1=1,nmode
               if (.not. gwpm(m1)) cycle
               if (m1 .eq. m) cycle

               zeig3=zeig1
               zeig1=zeig2
               zeig2=zeig3
               call qtxxzz(gs2(gs2mat(m1,s)),workc(zeig1),workc(zeig2),&
                    vdim(m1,s),dim(m1,s),ndim(m1,s))
            enddo
            call addvxxzo(workc(zeig2),tmp2,block1)

         enddo
      endif

!-----------------------------------------------------------------------
! add i S*Adot + i tau (prop)  OR S*Adot - tau  (rlx)  = HA
!-----------------------------------------------------------------------
      call addvxxzo(tmp2,tmp1,block1)

!-----------------------------------------------------------------------
! form A*HA
!-----------------------------------------------------------------------
      call vvaxzz(psi,tmp1,ztmp,block1)
      energys = dble(ztmp)

      return
      end subroutine calcenergy_gs

!#######################################################################

      end module evaluate
