!#######################################################################
!! GWPLIB
!
! Subroutines relating to GWPs that are needed by different parts 
! of the MCTDH program and ANALYSE programs
!
! gwp_param : routine to extract parameters from GWP and calculates
!             various quantities
!   param_from_psi: extract parameters from wavefunction
!   center  : calculates the centre coordinate of the GWPs
!   diagonal_moments : calculates quantities required to generate 
!                      moments
! normgwp : normalise general GWP
!
! transavec: transforms psi from G-MCTDH to mixed picture
! transavec1: transforms psi from mixed picture to G-MCTDH
!
! contains : gwp_param,gwp_param1m,param_from_psi,center,gwp_cmom,centermom,
! diagonal_moments,diag_zgeev,normgwp,gwp_ovl,overlap_elements,overlap_elements1,
! par2moment,moment2par,check_moments,moment,transavec,transavec1,transadot1,
! gwp2gpoint,gwp2gpoint1,wkgwplib,diag_cont(NU),rotcoeffgwp,rotcoeffgwp1
!
! ovlgspf : gets overlaps of SPFs represented by GWPs
!
!#######################################################################

      module gwplib

use decimal, only: dop, long
use mtlib,   only: qtxxzz
use op1lib,  only: zeromxz, cpqxz, zerovxi, cpvxz, zerovxd, cpmaz, cpmxz, &
                   unitqxz, unitqxd
use mmlib,   only: qqxxzz,qmxxzz,mmxazz,mqxczz,mqxtzz,qqaxzz
use xvlib,   only: qvxxzz
use op2lib,  only: addvxxio
use lalib,  only: simtran1bz

use logdat
use timing
use global
use dvrdatmod
use griddatmod
use psidef
use operdef
use runpropmod
use hpsimod

use gwptrj
use maxv, only: maxdim
use d2tens, only: mkherm,filldens
use xvlib, only: xvixdzo,vvaxzz
use datenmod, only: lgentrj
use BELglobals, only : dgldimmax

      implicit none

      private
      public :: gwp_param,gwp_param_a,gwp_param1m,param_from_psi,center,&
                gwp_cmom,diagonal_moments,normgwp,gwp_ovl,gwp_overlap,&
                gwp_overlap1m,overlap_elements,&
                par2moment,moment2par,moment,transavec,transavec1,&
                gwp2gpoint,gwp2gpoint1,gwp2gpoint2,gwptophi,gwptophi1,&
                transahelp,transahelp1m,&
                renormdgwp,renormdgwp1,getbdot,gwp_param1m_a,pqtoxivec,&
                fgaover,ffgaover,fgaover1,gwpproj,printgwpspf,printgwpspf0,&
                invert_ovr,ovlgspf,gwporthonorm,renormgwpa,&
                rotcoeffgwp,rotcoeffgwp1,renormgwptrja,diagonal_momentsd,&
                diagonal_momentsz,renormgwpBEL,gwp_getwidth,testgs2,&
                gs2redim,gs2redim1,renormgwp,chkpergwp,gwp_distin

      contains 

!+-----------------------------------------------------------------------+
!     calculates Gaussian matrix elements, of overlap and Hamiltonian
!     type
!+-----------------------------------------------------------------------+

      subroutine gwp_param (psi,agmat,gfgwp,aigwp,zgwp,qcenter,modus)

      implicit none

      integer(long)                                  :: m,phidim1
      integer(long),intent(in)                       :: modus
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(agmatdim), intent(inout) :: agmat
      complex(dop), dimension(gfgwpdim), intent(out) :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out) :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)  :: zgwp
      real(dop), dimension(qcentdim), intent(out)    :: qcenter

      do m = 1,nmode
         if (gwpm(m)) then
            phidim1 = phidim(m)+gbfdim(m)
            call gwp_param1m(m,psi(zetf(m,1)),agmat,gfgwp,aigwp,zgwp,&
                 qcenter,zetf(m,1),phidim1,modus)
         endif
      enddo

      return
      end subroutine gwp_param

!#######################################################################

      subroutine gwp_param_a (psi,agmat,gfgwp,aigwp,zgwp,qcenter,modus)

      implicit none

      integer(long)                                  :: m,phidim1
      integer(long),intent(in)                       :: modus
      complex(dop), dimension(dgldim), intent(inout)   :: psi
      complex(dop), dimension(agmatdim), intent(inout) :: agmat
      complex(dop), dimension(gfgwpdim), intent(out)   :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out)   :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)    :: zgwp
      real(dop), dimension(qcentdim), intent(out)      :: qcenter

      do m = 1,nmode
         if (gwpm(m)) then
            phidim1 = phidim(m)+gbfdim(m)
            call gwp_param1m_a(m,psi(zetf(m,1)),agmat, &
                 gfgwp,aigwp,zgwp,qcenter,zetf(m,1),phidim1,modus)
         endif
      enddo

      return
      end subroutine gwp_param_a

!#######################################################################

      subroutine gwp_param1m (m,psi,agmat,gfgwp,aigwp,zgwp, &
                 qcenter,zetf1,phidim1,modus)

      implicit none

      integer(long)             :: sf,si,s
      integer(long), intent(in) :: m,zetf1,phidim1,modus

      complex(dop), dimension(agmatdim), intent(inout)          :: agmat
      complex(dop), dimension(zetf1:zetf1+phidim1-1),intent(inout) :: psi
      complex(dop), dimension(gfgwpdim), intent(out)            :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out)            :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)             :: zgwp
      real(dop), dimension(qcentdim), intent(out)               :: qcenter

!-----------------------------------------------------------------------
! store width matrix in agmat
!-----------------------------------------------------------------------
      do s=1,nstate
         call param_from_psi(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)),&
              psi(zpg(m,s)),psi(zpg(m,s)),psi(zphi(m,s)),agmat(zphase(m,s)),&
              psi(zxi(m,s)),modus,s,m)
      enddo
      
!-----------------------------------------------------------------------
! form linear parameter
!-----------------------------------------------------------------------
      do s=1,nstate
         call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
              psi(zxi(m,s)),agmat(zxivec(m,s)))
      enddo

!+-----------------------------------------------------------------------+
!        center of packet
!+-----------------------------------------------------------------------+
      do s=1,nstate
         call center(nspfdof(m),gwpdim(m,s),psi(zxi(m,s)),qcenter(zcent(m,s)))
      enddo

!+-----------------------------------------------------------------------+
!         determine & store moments:
!+-----------------------------------------------------------------------+
      if (lBEL) then
         do sf=1,nstate
            do si=1,nstate
               call diagonal_moments(m,si,sf,nspfdof(m), &
                       gwpdim(m,sf),gwpdim(m,si), &
                       agmat(zagmat(m,sf)),agmat(zagmat(m,si)), &
                       agmat(zxivec(m,sf)),agmat(zxivec(m,si)), &
                       agmat(zphase(m,sf)),agmat(zphase(m,si)), &
                       gfgwp(zgfgwp(m,sf,si)), &
                       aigwp(zaigwp(m,sf,si)), &
                       zgwp(zzgwp(m,sf,si)),0,0)
            enddo
         enddo
      else
         do sf=1,nstate
            do si=1,nstate
               call diagonal_moments(m,si,sf,nspfdof(m),gwpdim(m,sf),&
                       gwpdim(m,si), &
                       agmat(zagmat(m,sf)),agmat(zagmat(m,si)), &
                       agmat(zxivec(m,sf)),agmat(zxivec(m,si)), &
                       agmat(zphase(m,sf)),agmat(zphase(m,si)), & 
                       gfgwp(zgfgwp(m,sf,si)), &
                       aigwp(zaigwp(m,sf,si)), &
                       zgwp(zzgwp(m,sf,si)),0,1)
            enddo
         enddo
      endif

      return
      end subroutine gwp_param1m

!#######################################################################
!
! version of gwp_param with normalisation of GWPs
!
!#######################################################################

      subroutine gwp_param1m_a (m,psi,agmat,gfgwp,aigwp,zgwp, &
                 qcenter,zetf1,phidim1,modus)

      implicit none

      integer(long)             :: sf,si,s
      integer(long), intent(in) :: m,zetf1,phidim1,modus

      complex(dop), dimension(agmatdim), intent(inout)          :: agmat
      complex(dop), dimension(zetf1:zetf1+phidim1-1),intent(inout) :: psi
      complex(dop), dimension(gfgwpdim), intent(out)            :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out)            :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)             :: zgwp
      real(dop), dimension(qcentdim), intent(out)               :: qcenter

!-----------------------------------------------------------------------
! store width matrix in agmat
!-----------------------------------------------------------------------
      do s=1,nstate
         call param_from_psi(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)),&
              psi(zpg(m,s)),psi(zpg(m,s)),psi(zphi(m,s)),agmat(zphase(m,s)),&
              psi(zxi(m,s)),modus,s,m)
      enddo

!-----------------------------------------------------------------------
! form linear parameter
!-----------------------------------------------------------------------
      do s=1,nstate
         call pqtoxivec(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)), &
              psi(zxi(m,s)),agmat(zxivec(m,s)))
      enddo
      
!-----------------------------------------------------------------------
! set scalar parameter to normalise GWP if required
!-----------------------------------------------------------------------
      do s=1,nstate
         !call normgwp(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)),&
         !     agmat(zxivec(m,s)),agmat(zphase(m,s)))
         if (lgwpphase) then
            call normgwp(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)),&
                 agmat(zxivec(m,s)),agmat(zphase(m,s)))
            psi(zphi(m,s):zphi(m,s)+gwpdim(m,s)-1) = &
                 agmat(zphase(m,s):zphase(m,s)+gwpdim(m,s)-1)
         endif
      enddo

!+-----------------------------------------------------------------------+
!        center of packet
!+-----------------------------------------------------------------------+
      do s=1,nstate
         call center(nspfdof(m),gwpdim(m,s),psi(zxi(m,s)),qcenter(zcent(m,s)))
      enddo

!+-----------------------------------------------------------------------+
!         determine & store moments:
!+-----------------------------------------------------------------------+
      do sf=1,nstate
         do si=1,nstate
            call diagonal_moments(m,si,sf,nspfdof(m),gwpdim(m,sf),gwpdim(m,si), &
                    agmat(zagmat(m,sf)),agmat(zagmat(m,si)), &
                    agmat(zxivec(m,sf)),agmat(zxivec(m,si)), &
                    agmat(zphase(m,sf)),agmat(zphase(m,si)), & ! phase mod yapol
                    gfgwp(zgfgwp(m,sf,si)), &
                    aigwp(zaigwp(m,sf,si)), &
                    zgwp(zzgwp(m,sf,si)),0,1)
         enddo
      enddo

      return
      end subroutine gwp_param1m_a

!#######################################################################
!
! PARAM_FROM_PSI
!
! extract the width matrix from the wavefunction vector if 
! thawed or separable GWPs and store in agmat.
!
!#######################################################################

      subroutine param_from_psi(gdof,gwpdim1,agmat,pgmat,dgmat,phi,phigmat,&
                 xi,modus,s,m) 

!+-----------------------------------------------------------------------+
!     extracts the width matrix, agmat, and the phase, phigmat, from psi
!+-----------------------------------------------------------------------+

      implicit   none

      integer(long)                                        :: ig,i,j
      integer(long), intent(in)                            :: gdof,gwpdim1,&
                                                              modus,s,m
      complex(dop), dimension(gdof,gwpdim1), intent(in)         :: dgmat
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)    :: pgmat
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(inout) :: agmat
      complex(dop), dimension(gdof,gwpdim1), intent(inout)      :: xi
      complex(dop), dimension(gwpdim1), intent(inout)           :: phigmat
      complex(dop), dimension(gwpdim1), intent(inout)           :: phi
      real(dop), dimension(gdof)                                :: p,q
      complex(dop), dimension(gdof)                             :: a

      if (gwptype .le. 1) then
! copy width matrix to agmat
         do ig=1,gwpdim1
            do i=1,gdof
               do j=1,gdof
                  agmat(i,j,ig) = pgmat(i,j,ig)
               enddo
            enddo
         enddo

      else if (gwptype .eq. 2) then
! for diagonal form copy width matrix to agmat
         agmat = 0.0_dop
         do ig=1,gwpdim1
            do i=1,gdof
               agmat(i,i,ig) = dgmat(i,ig)
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! Get the phase
! OR
! Calculate the phase C = -0.25*ln(pi/(2*Re(a))) - a*q^2 - ipq
!-----------------------------------------------------------------------
      if (lgwpphase) then
         phigmat(1:gwpdim1) = phi(1:gwpdim1)

      else
         do ig=1,gwpdim1
            phigmat(ig) = (0.0_dop,0.0_dop)
            do i=1,gdof
               a(i) = -agmat(i,i,ig)
               p(i) = dimag(xi(i,ig))
               q(i) = dble(xi(i,ig))
               phigmat(ig) = phigmat(ig) - a(i)*q(i)**2 &
                    - (0.0_dop,1.0_dop)*p(i)*q(i) &
                    - 0.25_dop*dlog(pi/(2.0_dop*dble(a(i))))
            enddo
         enddo
      endif

      return
      end subroutine param_from_psi

!#######################################################################

      subroutine pqtoxivec(gdof,gwpdim1,agmat,xi,xivec)

!+-----------------------------------------------------------------------+
!     Converts the linear parameters in pq form to xi form
! xi = q + ip 
! xivec = -2Aq + ip
!
!+-----------------------------------------------------------------------+

      implicit   none

      integer(long)                                        :: ig,i,j
      integer(long), intent(in)                            :: gdof,gwpdim1
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)  :: agmat
      complex(dop), dimension(gdof,gwpdim1), intent(out)   :: xivec
      complex(dop), dimension(gdof,gwpdim1), intent(in)    :: xi
      real(dop), dimension(gdof)                           :: p,q
!
! form linear parameter
!
      if (gwptype .le. 1 .or. gwptype .eq. 4) then
         do ig=1,gwpdim1
            do i=1,gdof
               p(i) = dimag(xi(i,ig))
               q(i) = dble(xi(i,ig))
            enddo
            do i=1,gdof
               xivec(i,ig) = dcmplx(0.0_dop,1.0_dop)*p(i)
               do j=1,gdof
                  xivec(i,ig) = xivec(i,ig) - 2.0_dop*agmat(i,j,ig)*q(j)
               enddo
            enddo
         enddo
      else if (gwptype .eq. 2 .or. gwptype .eq. 3) then
         do ig=1,gwpdim1
            do i=1,gdof
               p(i) = dimag(xi(i,ig))
               q(i) = dble(xi(i,ig))
               xivec(i,ig) = -2.0_dop*agmat(i,i,ig)*q(i) &
                    + dcmplx(0.0_dop,1.0_dop)*p(i)
            enddo
         enddo
      endif

      return
      end subroutine pqtoxivec

!#######################################################################

      subroutine center(gdof,gwpdim1,xi,qcenter)

!+-----------------------------------------------------------------------+
!     stores center coordinate q0(ndof) for Gaussian labeled ig
!+-----------------------------------------------------------------------+

      implicit none

      integer(long)                          :: ig,i
      integer(long), intent(in)              :: gdof,gwpdim1
      real(dop), dimension(gdof,gwpdim1), intent(out)     :: qcenter
      complex(dop), dimension(gdof,gwpdim1), intent(in)   :: xi

      do ig=1,gwpdim1
         do i=1,gdof
            qcenter(i,ig) = dble(xi(i,ig))
         enddo
      enddo

      return
      end subroutine center

!#######################################################################
!
! extracts momentum of centre of GWP from wavefunction array
!
!#######################################################################

      subroutine gwp_cmom (psi,pcenter)

      implicit none

      integer(long)                                :: m,s
      complex(dop), dimension(dgldim), intent(in)  :: psi
      real(dop), dimension(qcentdim), intent(out)  :: pcenter

!+-----------------------------------------------------------------------+
!  momentum of center of packet
!+-----------------------------------------------------------------------+
      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call centermom(nspfdof(m),gwpdim(m,s),psi(zxi(m,s)), &
                    pcenter(zcent(m,s)))
            endif
         enddo
      enddo

      return
      end subroutine gwp_cmom

!#######################################################################

      subroutine centermom(gdof,gwpdim1,xigmat,pcenter)

!+-----------------------------------------------------------------------+
!     calculates center momentum p0(ndof) for Gaussian labeled ig
!     (appearing on the r.h.s. of a given matrix element), using the
!     relation xi = - 2 a.q + ip , i.e., p = Im xi
!+-----------------------------------------------------------------------+

      implicit none

      integer(long)                                  :: ig,g
      integer(long), intent(in)                      :: gwpdim1,gdof
      real(dop), dimension(gdof,gwpdim1), intent(out)   :: pcenter
      complex(dop), dimension(gdof,gwpdim1), intent(in) :: xigmat

      do ig=1,gwpdim1
         do g=1,gdof
            pcenter(g,ig) = dimag(xigmat(g,ig))
         enddo
      enddo

      return
      end subroutine centermom

!#######################################################################
!
! modus = 0 generating functions for <g_i | x^n | g_j>
!       = 1 generating functions for <g_i^* | x^n | g_j>
!       = 2 do not use diagonal_momentsd - e.g. for different sets of 
!           functions
!
! moms = 0 only overlap calculated (x^n with n=0). ai and z not used.
!          gf(i,j) = <g_i | g_j>  ( or <g_i^* | g_j> )
!      = 1 generating functions for higher moments calculated and 
!          stored in ai and z
!
!#######################################################################

      subroutine diagonal_moments(m,si,sf,gdof,gwpdim1,gwpdim2, &
                 aleft,aright,xileft,xiright, &
                 phileft,phiright,gf,ai,z,modus,moms)
!+---------------------------------------------------------------------+
!     computes, for a given combination of Gaussians (ig,jg)
!     and of electronic states (si,sf),
!     the eigenvalues and eigenvectors of the width matrix ag,
!     the inverse ag_inv,
!     the linear coefficients xid in the new coordinates,
!     and the jacobian of the transformation.
!+---------------------------------------------------------------------+

      implicit     none

      integer(long)             :: si,sf,ig,jg
      integer(long), intent(in) :: m,gdof,modus,moms,gwpdim1,gwpdim2

      complex(dop), dimension(gwpdim1), intent(in)               :: phileft
      complex(dop), dimension(gwpdim2), intent(in)               :: phiright
      complex(dop), dimension(gdof,gwpdim1), intent(in)          :: xileft
      !complex(dop) :: xilefttmp, xirighttmp
      complex(dop), dimension(gdof,gwpdim2), intent(in)          :: xiright
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)     :: aleft
      complex(dop), dimension(gdof,gdof,gwpdim2), intent(in)     :: aright
      complex(dop), dimension(gwpdim1,gwpdim2), intent(out)      :: gf
      complex(dop)                       :: tmpgf
      complex(dop), dimension(gdof)      :: tmpz
      complex(dop), dimension(gdof,gdof) :: tmpai
      complex(dop), dimension(gdof,gwpdim1,gwpdim2), intent(out) :: z
      complex(dop), dimension(gdof,gdof,gwpdim1,gwpdim2), intent(out) :: ai
      complex(dop), dimension(gdof,gwpdim1) :: xilefttmp
 
!+---------------------------------------------------------------------+
!     for product of Gaussians,
!     write Gaussian parameters in terms of matrix of bilinear terms
!     (ag), vector of linear terms (xig), and complex phase (phig)
!+---------------------------------------------------------------------+
      if ((si .ne. sf) .or. (modus .ge. 1)) then
         do ig = 1,gwpdim1
            do jg = 1,gwpdim2
               call diagonal_momentsz(gdof,aleft(1,1,ig),aright(1,1,jg), &
                    xileft(1,ig),xiright(1,jg),phileft(ig),phiright(jg), &
                    gf(ig,jg),ai(1,1,ig,jg),z(1,ig,jg),modus,moms)
           enddo
         enddo
      else if (si .eq. sf) then
         do ig = 1,gwpdim1
            do jg = 1,ig-1
               call diagonal_momentsz(gdof,aleft(1,1,ig),aright(1,1,jg), &
                    xileft(1,ig),xiright(1,jg),phileft(ig),phiright(jg),&
                    gf(ig,jg),ai(1,1,ig,jg),z(1,ig,jg),modus,moms)
            enddo
            call diagonal_momentsd(gdof,aright(1,1,ig), &
                 xiright(1,ig),phiright(ig),&
                 gf(ig,ig),ai(1,1,ig,ig),z(1,ig,ig),moms)
            do jg = ig+1,gwpdim1
               call diagonal_momentsz(gdof,aleft(1,1,ig),aright(1,1,jg), &
                    xileft(1,ig),xiright(1,jg),phileft(ig),phiright(jg),&
                    gf(ig,jg),ai(1,1,ig,jg),z(1,ig,jg),modus,moms)
           enddo
         enddo

!pergauss 
         if (lpergwp) then
            do jg = 1,ig-1
              call stepgwp(m,gdof,xileft(1,ig),xilefttmp(1,ig))
                  call diagonal_momentsz(gdof,aleft(1,1,ig),aright(1,1,jg), &
                       xilefttmp(1,ig),xiright(1,jg),phileft(ig),phiright(jg),&
                       tmpgf,tmpai,tmpz,modus,moms)
                  gf(ig,jg)=gf(ig,jg)+tmpgf
                   ai(:,:,ig,jg)=ai(:,:,ig,jg)+tmpai
                   z(:,ig,jg)=z(:,ig,jg)+tmpz
            enddo
            do jg = ig+1,gwpdim1
              call stepgwp(m,gdof,xileft(1,ig),xilefttmp(1,ig))
                  call diagonal_momentsz(gdof,aleft(1,1,ig),aright(1,1,jg), &
                       xilefttmp(1,ig),xiright(1,jg),phileft(ig),phiright(jg),&
                       tmpgf,tmpai,tmpz,modus,moms)
                  gf(ig,jg)=gf(ig,jg)+tmpgf
                   ai(:,:,ig,jg)=ai(:,:,ig,jg)+tmpai
                   z(:,ig,jg)=z(:,ig,jg)+tmpz
            enddo
         endif

      endif

      return
      end subroutine diagonal_moments

!#######################################################################

      subroutine diagonal_momentsd(gdof,aright,xiright, &
                 phiright,gf,ai,z,moms)

!+---------------------------------------------------------------------+
!     computes, for a given combination of Gaussians (ig,jg)
!     and of electronic states (si,sf),
!     the eigenvalues and eigenvectors of the width matrix ag,
!     the inverse ag_inv,
!     the linear coefficients xid in the new coordinates,
!     and the jacobian of the transformation.
!+---------------------------------------------------------------------+

      implicit     none

      integer(long)             :: i,j,k,l,info
      integer(long), intent(in) :: gdof,moms

      complex(dop), intent(in)                        :: phiright
      complex(dop), dimension(gdof), intent(in)       :: xiright
      complex(dop), dimension(gdof,gdof), intent(in)  :: aright
      real(dop), dimension(gdof,gdof)                 :: ag,evr
      real(dop), dimension(gdof)                      :: xig,w,xid
      real(dop)                                       :: phig,tmp,ctmp
      real(dop), dimension(3*gdof)                    :: work

      complex(dop), intent(out)                       :: gf
      complex(dop), dimension(gdof), intent(out)      :: z
      complex(dop), dimension(gdof,gdof), intent(out) :: ai
      logical(kind=4) :: adiag

!+---------------------------------------------------------------------+
!     for product of Gaussians,
!     write Gaussian parameters in terms of matrix of bilinear terms
!     (ag), vector of linear terms (xig), and complex phase (phig)
!+---------------------------------------------------------------------+
! ... width parameters
      do k = 1,gdof
         do l = 1,gdof
            ag(k,l) = 2.0_dop*dble(aright(k,l))
         enddo
      enddo

! ... linear parameters in original coordinates :
      do k = 1, gdof
         xig(k) = 2.0_dop*dble(xiright(k))
      enddo

! ... complex phase :
      phig = 2.0_dop*dble(phiright)

!-----------------------------------------------------------------------
! need to check if A-matrix diagonal
!-----------------------------------------------------------------------
      adiag=.false.
      if (gdof .eq. 1 .or. (gwptype .eq. 2 .or. gwptype .eq. 3)) then
         adiag=.true.
      else
         tmp=0.0d0
         do k = 1,gdof
            do l = k+1,gdof
               tmp=tmp+ag(l,k) 
            enddo
         enddo
         if (dabs(tmp) .lt. tiny) adiag=.true.
      endif

!-----------------------------------------------------------------------
! inverse of width matrix
!-----------------------------------------------------------------------
      if (adiag) then
         do k=1,gdof
            w(k)          = ag(k,k)
            if (dabs(w(k)) .lt. tiny) w(k) = -tiny
         enddo
         call unitqxd(evr,gdof)

      else 
!+---------------------------------------------------------------------+
!     determine orthogonal transformation that diagonalizes the
!     complex-symmetric width matrix -> eval,evr
!     using either
!        -- lapack routine zgeev
!        -- Robin's perturbative scheme (diag_cont)
!      ( -- routine CS, by Ilan Bar-On
!           but: this has numerical problems for certain
!           initial conditions )
!      ( -- routine for diagonalization of a real matrix:
!           this applies to all ig=jg !
!           maybe it would be advantageous to distinguish
!           between diagonal/off-diagonal elements )
!+---------------------------------------------------------------------+
         do i=1,gdof
            do j=1,gdof
               evr(i,j)=ag(i,j)
            enddo
         enddo
         call dsyev('V','U',gdof,evr,gdof,w,work,3*gdof,info)

         do i=1,gdof
            if (w(i) .gt. 0.0d0) w(i) = -w(i)
            if (dabs(w(i)) .lt. tiny) w(i) = -w(i)
         enddo
      endif

!+---------------------------------------------------------------------+
!     with the coordinate transformation (evr), determine the
!     linear parameters xid for the diagonal moments:
!+---------------------------------------------------------------------+
      if (adiag) then
         do i = 1,gdof
            xid(i) = xig(i)
         enddo
      else
         do i = 1,gdof
            tmp = 0.0_dop
            do j=1,gdof
               tmp = tmp + xig(j)*evr(j,i)
            enddo
            xid(i) = tmp
         enddo
      endif

!+---------------------------------------------------------------------+
!     for use in function 'moment', compute generating function
!     gf, inverse matrix ai, and vector z = ai * xig
!+---------------------------------------------------------------------+
      tmp=phig
      gf=1.0_dop
      do i = 1,gdof
         tmp=tmp-xiD(i)**2/(4.0_dop*w(i))
         ctmp = dsqrt(-pi/w(i))
         gf=gf*ctmp
      enddo
      gf=gf*dexp(tmp)

!+---------------------------------------------------------------------+
!   calculate inverse of matrix ag from the eigenvalues + eigenvectors
!+---------------------------------------------------------------------+
      if (moms .gt. 0) then
         if (adiag) then
            call zeromxz(ai,gdof,gdof)
            do i=1,gdof
               ai(i,i) = 1.0_dop/w(i)
            enddo

            do i=1,gdof
               z(i) = ai(i,i) * xig(i)
            enddo

         elseif(gdof.gt.1)then
            do k = 1,gdof
               do l = 1,gdof
                  tmp = 0.0_dop
                  do i = 1,gdof
                     tmp = tmp + evr(k,i)*evr(l,i)/w(i)
                  enddo
                  ai(k,l) = tmp
               enddo
            enddo
 
            do i=1,gdof
               tmp = 0.0_dop
               do j = 1,gdof
                  tmp = tmp + dble(ai(i,j) * xig(j))
               enddo
               z(i) = tmp
            enddo
         endif
      endif

      return
      end subroutine diagonal_momentsd

!#######################################################################

      subroutine diagonal_momentsz(gdof,aleft,aright,xileft,xiright, &
                 phileft,phiright,gf,ai,z,modus,moms)

!+---------------------------------------------------------------------+
!     computes, for a given combination of Gaussians (ig,jg)
!     and of electronic states (si,sf),
!     the eigenvalues and eigenvectors of the width matrix ag,
!     the inverse ag_inv,
!     the linear coefficients xid in the new coordinates,
!     and the jacobian of the transformation.
!+---------------------------------------------------------------------+

      implicit     none

      integer(long)             :: i,j,k,l,info,adiag
      integer(long), intent(in) :: gdof,modus,moms
      complex(dop), intent(in)                        :: phileft,phiright
      complex(dop), dimension(gdof), intent(in)       :: xileft,xiright
      complex(dop), dimension(gdof,gdof), intent(in)  :: aleft,aright
      complex(dop), intent(out)                       :: gf
      complex(dop), dimension(gdof), intent(out)      :: z
      complex(dop), dimension(gdof,gdof), intent(out) :: ai
      complex(dop), dimension(gdof,gdof)              :: ag,evr
      complex(dop), dimension(gdof)                   :: xig,w,xid,eval
      complex(dop)                                    :: phig,tmp,ctmp
      real(dop)                                       :: rtmp
      real(dop), dimension(gdof)                      :: reval
      real(dop), dimension(3*gdof)                    :: workr
      real(dop), dimension(gdof,gdof)                 :: revr

!+---------------------------------------------------------------------+
!     for product of Gaussians,
!     write Gaussian parameters in terms of matrix of bilinear terms
!     (ag), vector of linear terms (xig), and complex phase (phig)
!+---------------------------------------------------------------------+
! ... width parameters
      if (modus .eq. 1) then
         do k = 1,gdof
            do l = 1,gdof
               ag(k,l) = aleft(k,l) + aright(k,l)
            enddo
         enddo
      else 
         do k = 1,gdof
            do l = 1,gdof
               ag(k,l) = dconjg(aleft(k,l)) + aright(k,l)
            enddo
         enddo
      endif

! ... linear parameters in original coordinates :
      if (modus .eq. 1) then
         do k = 1, gdof
            xig(k) = xileft(k) + xiright(k)
         enddo
      else 
         do k = 1, gdof
            xig(k) = dconjg(xileft(k)) + xiright(k)
         enddo
      endif

! ... complex phase :
      if (modus .eq. 1) then
         phig = phileft+ phiright
      else 
         phig = dconjg(phileft) + phiright
      endif

!-----------------------------------------------------------------------
! need to check type of width matrix
! adiag = 0 diagonal
!         1 real symmetric
!         2 complex symmetric
!-----------------------------------------------------------------------
      adiag=2
      if (gdof .eq. 1 .or. (gwptype .eq. 2 .or. gwptype .eq. 3)) then
         adiag=0
      else
         tmp=0.0d0
         rtmp=0.0d0
         do k = 1,gdof
            rtmp=rtmp+dimag(ag(k,k))
            do l = k+1,gdof
               rtmp=rtmp+dimag(ag(l,k))
               tmp=tmp+ag(l,k)
            enddo
         enddo
         if (abs(tmp) .lt. tiny) then
            adiag=0
         else if (abs(rtmp) .lt. tiny) then
            adiag=1
         endif
      endif

!-----------------------------------------------------------------------
! inverse of width matrix
!-----------------------------------------------------------------------
      if (adiag .eq. 0) then
         do k=1,gdof
            w(k)          = ag(k,k)
            if (abs(w(k)) .lt. tiny) w(k) = -tiny
         enddo
         call unitqxz(evr,gdof)

      else if(adiag .eq. 1)then

         do i=1,gdof
            do j=1,gdof
               revr(i,j)=dble(ag(i,j))
            enddo
         enddo
         call dsyev('V','U',gdof,revr,gdof,reval,workr,3*gdof,info)

         do i=1,gdof
            if (reval(i) .gt. 0.0d0) reval(i) = -reval(i)
            if (dabs(reval(i)) .lt. tiny) reval(i) = -reval(i)
            w(i) = reval(i)
            do j=1,gdof
               evr(j,i) = revr(j,i)
            enddo
         enddo

      else if(adiag .eq. 2)then

!               call diag_zgeev(gdof,ag,eval,evr)

         call diag_seigen(gdof,ag,eval,evr)

!              call diag_cont (gdof,ag,eval,evr,ig,jg)

!              call diag_CS   (gdof,ag,eval,evr)

!              call diag_real (gdof,ag,eval,evr)

         do i=1,gdof
            w(i) = eval(i)
            rtmp = dble(w(i))*dble(w(i)) + dimag(w(i))*dimag(w(i))
            rtmp = dsqrt(rtmp)
            if (rtmp .lt. tiny) w(i) = -tiny
         enddo
       endif

!+---------------------------------------------------------------------+
!     with the coordinate transformation (evr), determine the
!     linear parameters xid for the diagonal moments:
!+---------------------------------------------------------------------+
      if (adiag .eq. 0) then
         do i = 1,gdof
            xid(i) = xig(i)
         enddo
      else
         do i = 1,gdof
            tmp = dcmplx(0.0_dop,0.0_dop)
            do j=1,gdof
               tmp = tmp + xig(j)*evr(j,i)
            enddo
            xid(i) = tmp
         enddo
      endif

!+---------------------------------------------------------------------+
!     for use in function 'moment', compute generating function
!     gf, inverse matrix ai, and vector z = ai * xig
!+---------------------------------------------------------------------+
      tmp=phig
      gf=1.0_dop
      if (adiag .lt. 2) then
! for real width matrix
         do i = 1,gdof
            rtmp=dble(w(i))
            tmp=tmp-xiD(i)*xiD(i)/(4.0_dop*rtmp)
            rtmp = dsqrt(-pi/rtmp)
            gf=gf*rtmp
         enddo
      else
! for complex width matrix
         do i = 1,gdof
            tmp=tmp-xiD(i)*xiD(i)/(4.0_dop*w(i))
            ctmp = sqrt(-pi/w(i))
            gf=gf*ctmp
         enddo
      endif
      gf=gf*exp(tmp)

!+---------------------------------------------------------------------+
!   calculate inverse of matrix ag from the eigenvalues + eigenvectors
!+---------------------------------------------------------------------+
      if (moms .gt. 0) then
         if (adiag .eq. 0) then
            call zeromxz(ai,gdof,gdof)
            do i=1,gdof
               ai(i,i) = 1.0_dop/w(i)
            enddo
            do i=1,gdof
               z(i) = ai(i,i) * xig(i)
            enddo

         elseif(adiag.gt.0)then
            do k = 1,gdof
               do l = 1,gdof
                  tmp = dcmplx(0.0_dop,0.0_dop)
                  do i = 1,gdof
                     tmp = tmp + evr(k,i)*evr(l,i)/w(i)
                  enddo
                  ai(k,l) = tmp
               enddo
            enddo
 
            do i=1,gdof
               tmp = dcmplx(0.0_dop,0.0_dop)
               do j = 1,gdof
                  tmp = tmp + ai(i,j) * xig(j)
               enddo
               z(i) = tmp
            enddo
         endif
      endif

      return
      end subroutine diagonal_momentsz
!#######################################################################
!pergauss
      subroutine stepgwp(m,gdof,xileft,xilefttmp)
      use global
       implicit none
      integer(long)             :: m,f,gdof,n
      complex(dop), dimension(gdof), intent(in)       :: xileft
      complex(dop), dimension(gdof), intent(out)          :: xilefttmp
      do n=1,nspfdof(m)
        f=spfdof(n,m)
!        write(*,*)"iperiod",ipbaspar(2,f)
!        write(*,*)"normal mode",m
        if (ipbaspar(2,f) .ne. 0) then
          xilefttmp(:) = 4*pi/abs(ipbaspar(2,f)) + xileft(:)
        endif
       enddo 

      return
      end subroutine stepgwp

!#######################################################################

      subroutine diag_seigen(gdof,ag,eval,evr)
!+---------------------------------------------------------------------+
!     calls routine SEigensystem
!     to diagonalize a symmetric complex matrix :
!+---------------------------------------------------------------------+

      implicit     none

      integer(long)             :: n,lda,isort
      integer(long), intent(in) :: gdof

      complex(dop), dimension(gdof,gdof), intent(inout) :: ag
      complex(dop), dimension(gdof,gdof)                :: ag1
      complex(dop), dimension(gdof,gdof), intent(out)   :: evr
      complex(dop), dimension(gdof), intent(out)        :: eval

      n      = gdof
      lda    = n
      isort  = 1

      call cpmxz(ag,ag1,gdof,gdof)
      call seigensystem(n,ag1,lda,eval,evr,lda,isort)

      return
      end subroutine diag_seigen

!#######################################################################

      subroutine normgwp(gdof,gwpdim1,agmat,xi,phig)

!+---------------------------------------------------------------------+
!     normalises a set of general GWPs
!+---------------------------------------------------------------------+

      implicit none

      integer(long)                                       :: i,j,k,l,info,ig
      integer(long), intent(in)                           :: gdof,gwpdim1

      real(dop), dimension(maxdim)                        :: xig,xid
      real(dop), dimension(maxdim,maxdim)                 :: ag
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in) :: agmat
      complex(dop), dimension(gdof,gwpdim1), intent(in)   :: xi
      complex(dop), dimension(gwpdim1), intent(inout)     :: phig

      real(dop)                                           :: tmp,tmp1,imphi,rephi
      real(dop), dimension(gdof)                          :: eval
      real(dop), dimension(3*gdof)                        :: work
      real(dop), dimension(gdof,gdof)                     :: evek
      real(dop), dimension(gdof)                          :: p,q
      complex(dop), dimension(gdof)                       :: a

      if (lgwpsep .and. gwpphase .eq. -1) then
!-----------------------------------------------------------------------
! If scalar parameter not propagated (phase = 0), then set norm
!-----------------------------------------------------------------------
        do ig=1,gwpdim1
            phig(ig) = (0.0_dop,0.0_dop)
            do i=1,gdof
               a(i) = -agmat(i,i,ig)
               p(i) = dimag(xi(i,ig))
               q(i) = dble(xi(i,ig))
               phig(ig) = phig(ig) - a(i)*q(i)**2 &
                    - (0.0_dop,1.0_dop)*p(i)*q(i) &
                    - 0.25_dop*dlog(pi/(2.0_dop*dble(a(i))))
            enddo
         enddo

      else
!+---------------------------------------------------------------------+
!     write Gaussian parameters in terms of matrix of bilinear terms
!     (ag), vector of linear terms (xig), and complex phase (phig)
!+---------------------------------------------------------------------+

         do ig = 1,gwpdim1

! ... width parameters
            do k = 1,gdof
               do l = 1,gdof
                  ag(k,l) = 2.0_dop*dble(agmat(k,l,ig))
               enddo
            enddo
! ... linear parameters in original coordinates :
            do k = 1, gdof
               xig(k) = 2.0_dop*dble(xi(k,ig))
            enddo

            if (gdof .eq. 1 .or. (gwptype .eq. 2 .or. gwptype .eq. 3)) then
               do k=1,gdof
                  eval(k) = ag(k,k)
                  xid(k)  = xig(k)
               enddo

            elseif(gdof.gt.1)then
!+---------------------------------------------------------------------+
!     determine eigenvalues of a matrix
!+---------------------------------------------------------------------+
               do i=1,gdof
                  do j=1,gdof
                     evek(i,j)=ag(i,j)
                  enddo
               enddo
               call dsyev('V','U',gdof,evek,gdof,eval,work,3*gdof,info)

!+---------------------------------------------------------------------+
!     with the coordinate transformation (evr), determine the
!     linear parameters xid for the diagonal moments:
!+---------------------------------------------------------------------+
                do i = 1,gdof
                   tmp = 0.0_dop
                   do j=1,gdof
                      tmp = tmp + xig(j)*evek(j,i)
                   enddo
                   xid(i) = tmp
                enddo
            endif

!+---------------------------------------------------------------------+
! define scalar parameter to normalise function to 1
!+---------------------------------------------------------------------+
            tmp=1.0_dop
            tmp1=0.0_dop
            do i=1,gdof
               tmp = tmp*dsqrt(pi/(-eval(i)))
               tmp1 = tmp1 + xid(i)**2/(4.0_dop*eval(i))
            enddo

! check on NORM
!          tmp3 = 2.0_dop*dble(phig(ig))
!          tmp2=tmp*dexp(tmp3-tmp1)

         
            imphi = dimag(phig(ig))
            rephi = tmp1 - dlog(tmp)
            rephi = 0.5_dop*rephi
            phig(ig) = dcmplx(rephi,imphi)

! check on NORM
!          tmp2 = tmp*dexp(-tmp1) 
!          tmp1 = 2.0_dop*dble(phig(ig))
!          tmp2=tmp2*dexp(tmp1)

         enddo
      endif

      return
      end subroutine normgwp

!#######################################################################
!
!     copies overlap matrices calculated in diagonal_moments, and
!     stored in gfgwp, to gs2 matrices
!
!#######################################################################

      subroutine gwp_ovl (gs2,gfgwp)

      implicit none

      integer(long)                                   :: m,s
      complex(dop), dimension(gs2matdim), intent(out) :: gs2
      complex(dop), dimension(gfgwpdim), intent(in)   :: gfgwp

      do s=1,nstate
         do m = 1,nmode
            if (.not. gwpm(m)) cycle

            if (gwpmodel .eq. 2) then
               call diagovl(gs2(gs2mat(m,s)),gfgwp(zgfgwp(m,s,s)),dim(m,s))
            else
               call cpqxz(gfgwp(zgfgwp(m,s,s)),gs2(gs2mat(m,s)),dim(m,s))
            endif
         enddo
      enddo

      return
      end subroutine gwp_ovl


      subroutine diagovl (gs2,gfgwp,dim1)

      implicit none

      integer(long), intent(in)                       :: dim1
      integer(long)                                   :: e
      complex(dop), dimension(dim1,dim1), intent(out) :: gs2
      complex(dop), dimension(dim1,dim1), intent(in)   :: gfgwp

      gs2=0.0_dop
      do e=1,dim1
         gs2(e,e) = gfgwp(e,e)
      enddo

      return
      end subroutine diagovl


!########################################################################
!
! gwp_param must have called first
!
!########################################################################

      subroutine gwp_overlap (gs2,gsinv,gsval,gsvek,gfgwp,aigwp,&
                 zgwp,modus)

!+-----------------------------------------------------------------------+
!     calculates Gaussian overlap matrix elements
!+-----------------------------------------------------------------------+

      implicit none

      integer(long)                                  :: m
      integer(long), save                            :: tid=0
      integer(long), intent(in)                      :: modus
      real(dop), dimension(gwpinfodim), intent(out)  :: gsval
      complex(dop), dimension(gs2matdim), intent(out):: gs2
      complex(dop), dimension(gsmatdim), intent(out) :: gsinv
      complex(dop), dimension(gsmatdim), intent(out) :: gsvek
      complex(dop), dimension(gfgwpdim), intent(in)  :: gfgwp
      complex(dop), dimension(aigwpdim), intent(in)  :: aigwp
      complex(dop), dimension(zgwpdim), intent(in)   :: zgwp

      call get_timer(tid, 'GWP overlap')
      call start_timer(tid)     

!+-----------------------------------------------------------------------+
!        overlap elements 
!+-----------------------------------------------------------------------+
      do m = 1,nmode
         if (gwpm(m)) then
            call gwp_overlap1m (gs2,gsinv,gsval,gsvek,gfgwp,aigwp,zgwp,m,&
                 modus)
         endif
      enddo

      call stop_timer(tid)

      return
      end subroutine gwp_overlap

!#######################################################################
!
! gwp_param must have called first
!
! modus flag: which elements are to be calculated
!          0: all elements, <g| dg/dlam>
!          1: the zero order elements, <g| g> and S^-1
!          2: do not calculate zero order elements
!          3: only the zero order elements
!
!#######################################################################

      subroutine gwp_overlap1m (gs2,gsinv,gsval,gsvek,gfgwp,aigwp, &
                 zgwp,m,modus)

!+-----------------------------------------------------------------------+
!     calculates Gaussian overlap matrix elements
!+-----------------------------------------------------------------------+

      implicit none

      integer(long)                                      :: s,npg1,npg2
      integer(long), intent(in)                          :: m,modus
      real(dop), dimension(gwpinfodim), intent(out)      :: gsval
      complex(dop), dimension(gs2matdim), intent(out)    :: gs2
      complex(dop), dimension(gsmatdim), intent(out)     :: gsinv
      complex(dop), dimension(gsmatdim), intent(out)     :: gsvek
      complex(dop), dimension(gfgwpdim), intent(in)      :: gfgwp
      complex(dop), dimension(aigwpdim), intent(in)      :: aigwp
      complex(dop), dimension(zgwpdim), intent(in)       :: zgwp

! --- select which elements are to be calculated ---
      if (modus.eq.0 .or. modus .eq. 3) then
         npg1=1
         npg2=npg(m)
      else if (modus.eq.1 .or. modus .eq. 3) then
         npg1=1
         npg2=1
      else if (modus.eq.2) then
         npg1=2
         npg2=npg(m)
      endif

!+-----------------------------------------------------------------------+
!        overlap elements 
!+-----------------------------------------------------------------------+
      do s=1,nstate
         if (lBEL) then
             call cpvxz(gfgwp(zgfgwp(m,s,s)),gs2(gs2mat(m,s)),&
                  gwpdim(m,s)*gwpdim(m,s))
         else
             call overlap_elements (npg1,npg2,nspfdof(m),gwpdim(m,s),&
                  gs2(gs2mat(m,s)),gfgwp(zgfgwp(m,s,s)),aigwp(zaigwp(m,s,s)),&
                  zgwp(zzgwp(m,s,s)))
         endif
      enddo

!-----------------------------------------------------------------------
! inverse of overlap
!-----------------------------------------------------------------------
      if (modus .lt. 2) then
         if (lgentrj .or. gwpmodel .eq. 2) then
! iMCG - S is unit matrix, but still need higher moments.
            do s=1,nstate
               call unitqxz(gsinv(gsmat(m,s)),gwpdim(m,s))
            enddo

         else if ((abs(psitype) .eq. 4) .or. lcysinv) then
! only need inverse for G-MCTDH or MCTDH/G using inverse rather than projector
! built from orthonormal complement
            do s=1,nstate
               call invert_ovr (gs2(gs2mat(m,s)),gsinv(gsmat(m,s)),&
                    gsval(zgwpinfo(m,s)),gsvek(gsmat(m,s)),gwpdim(m,s),m,s)
            enddo
         endif
      endif

      return
      end subroutine gwp_overlap1m

!#######################################################################
!+-----------------------------------------------------------------------+  
!     calculate inverse of overlap matrix: 
!     i.e., inverse of gs2(gs2mat(m,s)) with p1 = 1  : 
!     obtain gs1inv(gsmat(m,s))
!
! modus = 1 : use regularisation
!+-----------------------------------------------------------------------+  

      subroutine invert_ovr (gs2,gsinv,gsval,gsvek,gwpdim1,m,s)

use lineq, only: matinvzhp

      implicit   none

      integer(long)             :: info,e2,lda,e, e1,ierr,lwork,ngwpact,nsing
      integer(long), intent(in) :: gwpdim1,m,s
      complex(dop), dimension(gwpdim1,gwpdim1), intent(out)   :: gsinv
      complex(dop), dimension(gwpdim1,gwpdim1), intent(out)   :: gsvek
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in)    :: gs2

      real(dop)                                               :: sm,noffd,&
                                                                 tol
      real(dop), dimension(gwpdim1), intent(out)              :: gsval
      real(dop), dimension(gwpdim1)                           :: gsvaltmp
      complex(dop), dimension(:,:), allocatable               :: evek,gsinv1
      real(dop), dimension(3*gwpdim1)                         :: work
      complex(dop), dimension(gwpdim1*gwpdim1)                :: workc

!-----------------------------------------------------------------------
! copy overlap matrix
! ngwpact is no. of active functions, ignoring any linearly dependent
! GWPs
!-----------------------------------------------------------------------
      ngwpact=gwpdim1
      allocate(evek(ngwpact,ngwpact))
      allocate(gsinv1(ngwpact,ngwpact))
      evek(1:ngwpact,1:ngwpact)=gs2(1:ngwpact,1:ngwpact)
      noffd = max(1.0_dop,dble((ngwpact-1)*ngwpact/2))

!-----------------------------------------------------------------------
! for trajectory GWPs multiply in electronic overlaps
!-----------------------------------------------------------------------
      if (lgwptrj) then
         do e=1,ngwpact
            evek(e,e) = evek(e,e)*gwptrjeovl(e,e)
            do e1=e+1,ngwpact
               evek(e1,e) = evek(e1,e)*gwptrjeovl(e1,e)
               evek(e,e1) = dconjg(evek(e1,e))
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! direct diagonalisation
!-----------------------------------------------------------------------
      if (ovrinv .eq. 1) then
         tol=1.0e-8_dop
         gsinv=evek
         call zpotrf('L',ngwpact,gsinv,ngwpact,ierr)
         call zpotri('L',ngwpact,gsinv,ngwpact,ierr)
         if (ierr .ne. 0) then
            routine='invert_ovr'
            message='Error in calculating matrix inverse'
            call errormsg
         endif 

         call filldens(gsinv,ngwpact)

         evek=gs2

      else if (ovrinv .eq. 0 .or. ovrinv .eq. 2) then
          
!+-----------------------------------------------------------------------+
!     To calculate inverse of the overlap matrix: 
!     use diagonalization (complex hermitian matrix) and 
!     divide by eigenvalues
!+-----------------------------------------------------------------------+
          gsval=0.0_dop
          e2  = ngwpact*ngwpact
          lda = ngwpact
          lwork=ngwpact*ngwpact
          call zheev('V','U',ngwpact,evek,lda,gsval,workc,lwork,work,info)
          if (info .ne. 0) then
             routine = 'invert_ovr'
             message = 'Cannot diagonalize overlap matrix'
             call errormsg
          endif

!-----------------------------------------------------------------------
! Change the phases of the eigenvectors.
! The phase convention is: Sum_j [j*evek(j)] > 0.
!-----------------------------------------------------------------------
          do e=1,ngwpact
             sm=0.0_dop
             do e1=1,ngwpact
                sm = sm + dble(e1)*dble(evek(e1,e))
             enddo
             if(sm.lt.0.0_dop) then
                do e1=1,ngwpact
                   evek(e1,e) = -evek(e1,e)
                enddo
             endif
          enddo
      
!-----------------------------------------------------------------------
! ovrinv = 0: regularize eigenvalues
! ovrinv = 2: use svd to remove small eigenvalues
!-----------------------------------------------------------------------
          if (ovrinv .eq. 0) then
             do e=1,ngwpact
                if( gsval(e) .lt. 64.0_dop*epsgsmat) then
                   gsvaltmp(e)=gsval(e)+epsgsmat*exp(-gsval(e)/epsgsmat)
                else
                   gsvaltmp(e)=gsval(e)
                endif
                gsvaltmp(e)=1.0_dop/gsvaltmp(e)
             enddo

          else if (ovrinv .eq. 2) then
             nsing = 0
             do e=1,ngwpact
                if( gsval(e) .lt. epsgsmat) then
                   nsing = nsing+1
                   gsvaltmp(e)=0.0_dop
                else
                   gsvaltmp(e)=1.0_dop/gsval(e)
                endif
             enddo

! log no. of singular values < epsgsmat
             ngwpovr(m,s) = max(nsing,ngwpovr(m,s))

          endif

!-----------------------------------------------------------------------
! form inverse of overlap matrix gsinv
!-----------------------------------------------------------------------
          call zeromxz(gsinv1,ngwpact,ngwpact)
          do e=1,ngwpact
             do e1=e,ngwpact
                do e2=1,ngwpact
                   gsinv1(e1,e)=gsinv1(e1,e) +&
                      dconjg(evek(e,e2))*evek(e1,e2)*gsvaltmp(e2)
                enddo
             enddo
          enddo
          do e=1,ngwpact
             gsinv1(e,e)=dble(gsinv1(e,e))
             do e1=e+1,ngwpact
                gsinv1(e,e1)=dconjg(gsinv1(e1,e))
             enddo
          enddo
 
          gsinv=gsinv1
          !call gs2redim1(gsinv,gsvek,gwpdim1,gsinv1,evek,gwpdep1,ngwpact)
          if (.not. lBEL) gsvek=evek

      endif

      call mkherm(gsinv,gwpdim1)

      !call testgs2(gs2,gsinv,gsval,gwpdim1)
      if (allocated(evek)) deallocate(evek)
      if (allocated(gsinv1)) deallocate(gsinv1)

      return
      end subroutine invert_ovr

!#######################################################################
!
! remove a row + column from the overlap matrix 

      subroutine gs2redim(gs2,gwpdim,gs21,gwpdep,ngwpact)

      implicit none

      integer(long) :: i,j,m,n
      integer(long), intent(in)                                 :: gwpdim,&
                                                                   ngwpact
      integer(long), dimension(gwpdim), intent(in)              :: gwpdep
      complex(dop), dimension(gwpdim,gwpdim), intent(in)        :: gs2
      complex(dop), dimension(ngwpact,ngwpact), intent(out)     :: gs21

      m = 0
      do i=1,gwpdim
         if (gwpdep(i) .ne. 0) cycle
         m = m+1
         n = 0
         do j=1,gwpdim
            if (gwpdep(j) .ne. 0) cycle
            n = n + 1
            gs21(n,m) = gs2(j,i)
         enddo
      enddo

      return
      end subroutine gs2redim

!#######################################################################
! 
! add back a column + row of zeros

      subroutine gs2redim1(gsinv,gsvek,gwpdim,gsinv1,evek,gwpdep,ngwpact)

      implicit none

      integer(long), intent(in)                      :: gwpdim,ngwpact
      integer(long), dimension(gwpdim), intent(in)   :: gwpdep
      integer(long)                                  :: i,m,j,n
      complex(dop), dimension(gwpdim,gwpdim), intent(out)   :: gsinv
      complex(dop), dimension(ngwpact,ngwpact), intent(in)  :: gsinv1
      complex(dop), dimension(gwpdim,gwpdim), intent(out)   :: gsvek
      complex(dop), dimension(ngwpact,ngwpact), intent(in)  :: evek

      gsinv = 0.0_dop
      m = 0
      do i=1,gwpdim
         if (gwpdep(i) .ne. 0) cycle
         m = m+1
         n = 0
         do j=1,gwpdim
            if (gwpdep(j) .ne. 0) cycle
            n = n + 1
            gsinv(j,i) = gsinv1(n,m)
            gsvek(j,i) = evek(n,m)
         enddo
      enddo

      return
      end subroutine gs2redim1

!#######################################################################

      subroutine testgs2(gs2,gsinv,eval,gwpdim1)

      implicit none

      integer(long), intent(in) :: gwpdim1
      real(dop), intent(in) :: eval(gwpdim1)
      complex(dop), intent(in), dimension(gwpdim1,gwpdim1) :: gs2,gsinv
      integer(long) :: e,e1,novr
      real(dop) :: gs2max,rtmp,ovlthr
      complex(dop) :: trace,offsum
      complex(dop), dimension(gwpdim1,gwpdim1) :: ztmp
      
      print*,'Lowest GS2 Eval =',eval(1)

      ovlthr = 0.98
      gs2max = 0.0_dop
      novr=0
      do e=1,gwpdim1
         do e1=e+1,gwpdim1
            rtmp = abs(gs2(e1,e))
            if (rtmp .gt. ovlthr) then
               novr = novr+1
               print*,'OVER ',e1,e
            endif
            gs2max = max(gs2max,rtmp)
         enddo
      enddo
      print*,'Maximum GS2 off =',gs2max
      print*,'No. of GS2 off over ovlthr =',novr

      call qqxxzz(gsinv,gs2,ztmp,gwpdim1)
      trace = 0.0_dop
      offsum = 0.0_dop
      do e=1,gwpdim1
         trace = trace+ztmp(e,e)
         do e1=e+1,gwpdim1
            offsum = offsum+ztmp(e1,e)
         enddo
      enddo

      print*,'No. GWPs = ', gwpdim1
      print*,'Trace S^-1 S = ',trace
      print*,'Off-diag sum = ',offsum
      print*
         

      return
      end subroutine testgs2

!#######################################################################

      subroutine overlap_elements (npg1,npg2,gdof,gwpdim1,gs2,gfgwp, &
                 aigwp,zgwp)

!+-----------------------------------------------------------------------+
!     calculates overlap matrices for a set of GWPs
!+-----------------------------------------------------------------------+

      implicit   none

!      integer(long), intent(in)              :: gwpdim1,gdof,gdof1,npg1,npg2
      integer(long), intent(in)              :: gwpdim1,gdof,npg1,npg2
      integer(long)                          :: ig,jg,p1,p2,p1factor,p2factor
      integer(long), dimension(maxdim)       :: pmoment, qmoment

      complex(dop), dimension(gwpdim1,gwpdim1,npg2,npg2), intent(out) :: gs2
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in)            :: gfgwp
      complex(dop), dimension(gdof,gwpdim1,gwpdim1), intent(in)       :: zgwp
      complex(dop), dimension(gdof,gdof,gwpdim1,gwpdim1), intent(in)  :: aigwp
!      complex(dop)                                 :: moment

!
! ... calculate moments diagonal in electronic state s1 :
! ... for overlap elements of type S(alpha,beta) :
!
      do p1 = npg1,npg2
         call par2moment(p1,gdof,pmoment,p1factor,gwptype)

         do p2 = p1,npg2
            call par2moment(p2,gdof,qmoment,p2factor,gwptype)
            call addvxxio(pmoment,qmoment,gdof)

!write(1000,*)"qmoment",qmoment,"source/mctdhlib/gwplib.f90"
            do ig=1,gwpdim1
               do jg=ig,gwpdim1
                  gs2(jg,ig,p2,p1) = p1factor*p2factor* &
                             moment(qmoment,gdof,gfgwp(jg,ig), &
                             aigwp(1,1,jg,ig),zgwp(1,jg,ig))
               enddo
            enddo
         enddo
      enddo
! 
! set symmetry
! 
      do p1 = npg1,npg2
         do ig=1,gwpdim1
            gs2(ig,ig,p1,p1) = dble(gs2(ig,ig,p1,p1))
            do jg=ig+1,gwpdim1
               gs2(ig,jg,p1,p1) = dconjg(gs2(jg,ig,p1,p1))
            enddo
         enddo

         do p2 = p1+1,npg2
            do ig=1,gwpdim1
               gs2(ig,ig,p2,p1) = dble(gs2(ig,ig,p2,p1))
               gs2(ig,ig,p1,p2) = gs2(ig,ig,p2,p1)

               do jg=ig+1,gwpdim1
                  gs2(ig,jg,p2,p1) = dconjg(gs2(jg,ig,p2,p1))
                  gs2(jg,ig,p1,p2) = gs2(jg,ig,p2,p1)
                  gs2(ig,jg,p1,p2) = dconjg(gs2(jg,ig,p1,p2))
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine overlap_elements

!#######################################################################

!      subroutine overlap_elements1 (npg1,npg2,gdof,gwpdim1,gwpdim2,gs2,gfgwp, &
!                 aigwp,zgwp)

!+-----------------------------------------------------------------------+
!     calculates overlap matrices - general routine for overlap between
!     different sets of GWPS
!+-----------------------------------------------------------------------+

!     implicit   none

!     integer(long), intent(in)              :: gwpdim1,gwpdim2,gdof,npg1,npg2
!     integer(long)                          :: ig,jg,p1,p2,p1factor, p2factor
!     integer(long), dimension(maxdim)       :: pmoment, qmoment

!!     complex(dop)                     :: moment
!     complex(dop), dimension(gwpdim1,gwpdim2,npg2,npg2), intent(out) :: gs2
!     complex(dop), dimension(gwpdim1,gwpdim2), intent(in)            :: gfgwp
!     complex(dop), dimension(gdof,gwpdim1,gwpdim2), intent(in)       :: zgwp
!     complex(dop), dimension(gdof,gdof,gwpdim1,gwpdim2), intent(in)  :: aigwp

!!... calculate moments diagonal in electronic state s1 :
!!... for overlap elements of type S(alpha,beta) :
!     do p1 = npg1,npg2
!        call par2moment(p1,gdof,pmoment,p1factor,gwptype)

!        do p2 = p1,npg2
!           call par2moment(p2,gdof,qmoment,p2factor,gwptype)
!           call addvxxio(pmoment,qmoment,gdof)

!           do ig=1,gwpdim2
!              do jg=1,gwpdim1
!                 gs2(jg,ig,p1,p2) = p1factor*p2factor* &
!                            moment(qmoment,gdof,gfgwp(jg,ig), &
!                            aigwp(1,1,jg,ig),zgwp(1,jg,ig))

!                 gs2(jg,ig,p2,p1) = gs2(jg,ig,p1,p2) 
!              enddo
!           enddo
!        enddo
!     enddo

!     return
!     end subroutine overlap_elements1

!#######################################################################

      subroutine par2moment(p1,gdof,imoment,pfactor,gwptype)
!+---------------------------------------------------------------------+
!     gives the integer array imoment(gdof) corresponding to the
!     parameter p1
!+---------------------------------------------------------------------+
      implicit none

      integer(long), intent(in)                   :: p1,gdof,gwptype
      integer(long), intent(out)                  :: pfactor
      integer(long)                               :: i,j,p2,counter 
      integer(long), dimension(gdof), intent(out) :: imoment

! ... note: gdof = nspfdof(m) for gwp mode       

      call zerovxi(imoment,gdof)

! ...  check whether p1 corresponds to phase 
       
      if(p1.eq.1)then
          
         pfactor = 1 

! ...  check whether p1 among linear parameters 

      elseif(p1.gt.1 .and. p1.le.(gdof+1))then

         imoment(p1-1) = 1
         pfactor    = 1 


! ... check whether p1 among quadratic parameters
! ... either diagonal (gwptype =2) or full parameter space

      else if(p1 .gt. gdof+1 .and. gwptype .eq. 2) then 

         i = p1-gdof-1
         imoment(i) = 2
         pfactor    = 1 

      else if(p1 .gt. gdof+1) then 

         p2=p1-gdof-1
      
         counter = 0

! think of A(j,i) stored as lower triangle (i.e. j >= i))

outer:   do i=1,gdof
            do j=i,gdof

               counter = counter + 1
               
               if(counter.eq.p2)then 
                  if(j.ne.i)then
                     imoment(i) = 1
                     imoment(j) = 1
                     pfactor    = 2 
                  elseif(j.eq.i)then
                     imoment(j) = 2
                     pfactor    = 1 
                  endif
                  exit outer
               endif
                  
            enddo
         enddo outer

      endif

      return
      end subroutine par2moment

!#######################################################################

      subroutine moment2par(imoment,gdof,p1,gwptype)
!+---------------------------------------------------------------------+
!     gives the parameter p1 corresponding to the integer array
!     imoment(gdof)
!+---------------------------------------------------------------------+
      implicit none

      integer(long), intent(in)                  :: gdof,gwptype
      integer(long), intent(out)                 :: p1
      integer(long)                              :: i,j,l,iorder,counter
      integer(long), dimension(gdof), intent(in) :: imoment

! ... note: np   = subdim(m,s)
!           gdof = nspfdof(m) for gwp mode

      iorder = 0
      do i=1,gdof
        iorder = iorder + imoment(i)
      enddo

      if(iorder.eq.0)then

        p1 = 1

      elseif(iorder.eq.1)then

        do i=1,gdof
           if (imoment(i).eq.1) p1=i+1
        enddo

      elseif(iorder.eq.2 .and. gwptype .eq. 2)then

        do i=1,gdof
           if (imoment(i).eq.2) p1=gdof+i+1
        enddo

      elseif(iorder.eq.2)then

        do i=1,gdof
           if(imoment(i).eq.2)then
              counter=0
              do j=1,(i-1)
                 counter=counter + (gdof-(j-1))
              enddo
              p1 = counter + 1 + gdof+1
           elseif(imoment(i).eq.1)then
              counter = 0
              do l=1,(i-1)
                 counter=counter + (gdof-(l-1))
              enddo
              do j= (i+1),gdof
                 if (imoment(j).eq.1) then
                    p1 = counter + j - (i-1) + gdof+1
                 endif
              enddo
           endif
        enddo

      endif

!  99  continue  ! not in use!!!!

      return
      end subroutine moment2par

!#######################################################################

!      subroutine check_moments(npg,gdof,gwptype)
!+---------------------------------------------------------------------+
!     checks on routines par2moment and moment2par
!+---------------------------------------------------------------------+

!     implicit none

!     integer(long), intent(in)              :: gdof,gwptype,npg
!     integer(long)                          :: p1,pfactor
!     integer(long), dimension(maxdim)       :: imoment

!     do p1 = 1,npg

!        call par2moment(p1,gdof,imoment,pfactor,gwptype)

!        call moment2par(imoment,gdof,p1,gwptype)

!     enddo

!     return
!     end subroutine check_moments

!#######################################################################

      complex(dop) function moment(imoment,gdof,gf,ai,z)
!+---------------------------------------------------------------------+
!     for a given combination of Gaussians (m,n), and a given
!     integer array imoment, computes the value of the corresponding
!     moment. The method used is based on a generating function (gf). 
!+---------------------------------------------------------------------+
      implicit none 

      integer(long)                                  :: i,j,sum_exp,counter
      integer(long), intent(in)                      :: gdof
      
      integer(long), dimension(5)                    :: im
      integer(long), dimension(gdof), intent(in)     :: imoment

      complex(dop), intent(in)                       :: gf
      complex(dop), dimension(gdof), intent(in)      :: z
      complex(dop), dimension(gdof,gdof), intent(in) :: ai

! ... calculate moments up to 5th order

! ... instead: specify --

      counter = 0
      sum_exp = 0

      do i=1,gdof
        sum_exp = sum_exp + imoment(i)
        if(imoment(i).ne.0)then
           do j=1,imoment(i)
              counter = counter + 1
              im(counter) = i
           enddo
         endif
      enddo

      moment=0
      if(sum_exp.eq.0)then

         moment = gf 

      elseif(sum_exp.eq.1)then 

         moment = -0.5_dop*gf*z(im(1))

      elseif(sum_exp.eq.2)then 
 
         moment = 0.25_dop * gf *(z(im(1))*z(im(2)) &
              - 2.0_dop* ai(im(1),im(2)) )

      elseif(sum_exp.eq.3)then 

         moment = -1.0_dop/8.0_dop * gf *( &
              z(im(1))*z(im(2))*z(im(3)) - &
              2.0_dop* ( &
              ai(im(1),im(2)) * z(im(3)) &
              +   z(im(1)) * ai(im(2),im(3)) &
              +   ai(im(1),im(3)) * z(im(2)) ) ) 

      elseif(sum_exp.eq.4)then         

         moment = 1.0_dop/16.0_dop * gf *( &
                   z(im(1))*z(im(2))*z(im(3))*z(im(4)) - &
               2.0_dop* (&
                   z(im(1))*z(im(2))*ai(im(3),im(4))&
               +   z(im(1))*ai(im(2),im(3))*z(im(4))&
               +   z(im(1))*z(im(3))*ai(im(2),im(4))&
               +   ai(im(1),im(2))*z(im(3))*z(im(4))&
               +   z(im(2))*ai(im(1),im(3))*z(im(4)) &
               +   z(im(2))*z(im(3))*ai(im(1),im(4)) )  + &
               4.0_dop* (&
                   ai(im(1),im(2))*ai(im(3),im(4)) &
               +   ai(im(2),im(3))*ai(im(1),im(4))&
               +   ai(im(1),im(3))*ai(im(2),im(4))  ) ) 

      elseif(sum_exp.eq.5)then
 
          moment = -1.0_dop/32.0_dop * gf *( &
                   z(im(1))*z(im(2))*z(im(3))*z(im(4))*z(im(5)) - &
                   2.0_dop* ( &
                   z(im(1))*z(im(2))*ai(im(3),im(4))*z(im(5)) &
               +   z(im(1))*ai(im(2),im(3))*z(im(4))*z(im(5)) &
               +   z(im(1))*z(im(3))*ai(im(2),im(4))*z(im(5)) &
               +   ai(im(1),im(2))*z(im(3))*z(im(4))*z(im(5)) &
               +   z(im(2))*ai(im(1),im(3))*z(im(4))*z(im(5)) &
               +   z(im(2))*z(im(3))*ai(im(1),im(4)) *z(im(5)) &
               +   z(im(2))*z(im(3))*ai(im(1),im(5)) *z(im(4)) &
               +   z(im(1))*z(im(3))*ai(im(2),im(5)) *z(im(4)) &
               +   z(im(2))*z(im(1))*ai(im(3),im(5)) *z(im(4)) &
               +   z(im(2))*z(im(3))*ai(im(4),im(5)) *z(im(1)) )  + &
               4.0_dop* ( &
                   ai(im(2),im(3))*ai(im(4),im(5))*z(im(1)) &
               +   ai(im(3),im(4))*ai(im(2),im(5))*z(im(1)) &
               +   ai(im(2),im(4))*ai(im(3),im(5))*z(im(1))   &
               +   ai(im(1),im(3))*ai(im(4),im(5))*z(im(2)) &
               +   ai(im(3),im(4))*ai(im(1),im(5))*z(im(2)) &
               +   ai(im(1),im(4))*ai(im(3),im(5))*z(im(2)) &
               +   ai(im(1),im(2))*ai(im(4),im(5))*z(im(3)) &
               +   ai(im(2),im(4))*ai(im(1),im(5))*z(im(3)) &
               +   ai(im(1),im(4))*ai(im(2),im(5))*z(im(3))  &
               +   ai(im(1),im(2))*ai(im(3),im(5))*z(im(4)) &
               +   ai(im(2),im(3))*ai(im(1),im(5))*z(im(4)) &
               +   ai(im(1),im(3))*ai(im(2),im(5))*z(im(4)) &
               +   ai(im(1),im(2))*ai(im(3),im(4))*z(im(5)) &
               +   ai(im(2),im(3))*ai(im(1),im(4))*z(im(5)) &
               +   ai(im(1),im(3))*ai(im(2),im(4))*z(im(5))  )  )  

       elseif(sum_exp.gt.5)then
          routine = 'moment'
          message = 'Higher moments not implemented'
          call errormsg
       endif

       return
       end function moment

!#######################################################################
!
! TRANSAVEC
!
! B = D^dag A
!
! psi : on input A (G-MCTDH picture) on output B (SPF picture)
!
!#######################################################################

      subroutine transavec(psi,adgwp)

      implicit none

      integer(long)                                  :: s,m

      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(gsmatdim), intent(in)  :: adgwp
      complex(dop), dimension(maxblock)              :: workc

!-----------------------------------------------------------------------
! multiply B = D A
!-----------------------------------------------------------------------
      do s=1,nstate
         do m = 1,nmode
            if (gwpm(m)) then
               call qtxxzz(adgwp(gsmat(m,s)),psi(zpsi(s)), &
                    workc,vdim(m,s),dim(m,s),ndim(m,s))
               call cpvxz(workc,psi(zpsi(s)),block(s))
            endif
         enddo
      enddo

      return
      end subroutine transavec

!#######################################################################
!
! TRANSAVEC1
!
! A = D^tilde B
!
! psi : on input B (SPF picture) on output A (G-MCTDH picture)
!
!#######################################################################

      subroutine transavec1(psi)

      implicit none

      integer(long)                                    :: s,m,zeig1,zeig2,zeig3
      complex(dop), dimension(dgldim), intent(inout)   :: psi
      complex(dop), dimension(2*maxblock)              :: workc

!-----------------------------------------------------------------------
! multiply B = D A
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig1 = block(s)+1
         zeig2 = 1
         call cpvxz(psi(zpsi(s)),workc(zeig1),block(s))

         do m = 1,nmode
            if (gwpm(m)) then
               call qtxxzz(psi(zdgwp(m,s)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
               zeig3=zeig2
               zeig2=zeig1
               zeig1=zeig3
            endif
         enddo

         call cpvxz(workc(zeig1),psi(zpsi(s)),block(s))
      enddo

      return
      end subroutine transavec1

!#######################################################################
!
! map Gaussian dof to actual dof in gpoint array
!
! NB gpdof is size of gpoint in calling routine. Ensures it is all zeroed.
! GWR 06/15
!#######################################################################
      subroutine gwp2gpoint(qcenter,gpoint,gdof,gpdof,mode)

      implicit none

      integer(long)                             :: n,f
      integer(long), intent(in)                 :: gdof,mode,gpdof
      real(dop), dimension(gpdof), intent(out)  :: gpoint
      real(dop), dimension(gdof), intent(in)    :: qcenter

      call zerovxd(gpoint,gpdof)
      do n=1,gdof
         f=spfdof(n,mode)
         gpoint(f)=qcenter(n)
      enddo
!
! add coordinates of point DVRs 
!
      if (lpntdvr) then
         do f=1,ndof
            if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
         enddo
      endif

      return
      end subroutine gwp2gpoint

!#######################################################################
!
! map Gaussian dof to actual dof in gpoint array ! changed by yapol
!
      subroutine gwp2gpoint1(qcenter1,qcenter2,gpoint,gdof,gpdof,mode)

      implicit none

      integer(long)                             :: n,f
      integer(long), intent(in)                 :: gdof,gpdof,mode
      real(dop), dimension(gpdof), intent(out)  :: gpoint
      real(dop), dimension(gdof), intent(in)    :: qcenter1,qcenter2

      call zerovxd(gpoint,gpdof)
      do n=1,gdof
         f=spfdof(n,mode)
         gpoint(f)=0.5_dop*(qcenter1(n) + qcenter2(n))
      enddo
!
! add coordinates of point DVRs 
!
      if (lpntdvr) then
         do f=1,ndof
            if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
         enddo
      endif

      return
      end subroutine gwp2gpoint1

!#######################################################################
!
! map Gaussian dof to actual dof in gpoint array
!
      subroutine gwp2gpoint2(qcenter,gpoint,gdof,gpdof)

      implicit none

      integer(long)                             :: n,f,imode,idof
      integer(long), intent(in)                 :: gdof,gpdof
      real(dop), dimension(gpdof), intent(out)  :: gpoint
      real(dop), dimension(gdof), intent(in)    :: qcenter

      call zerovxd(gpoint,gpdof)
      idof = 1
      imode = 1
      do n=1,gdof
         if (idof > nspfdof(imode)) then
            imode = imode + 1
            idof = 1
         endif
         f=spfdof(idof,imode)
         gpoint(f)=qcenter(n)
         idof = idof + 1
      enddo
!
! add coordinates of point DVRs 
!
      if (lpntdvr) then
         do f=1,gpdof
            if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
         enddo
      endif

      return
      end subroutine gwp2gpoint2

!#########################################################################
!+-----------------------------------------------------------------------+
!     calculates transformation matrix for GWPs -> SPFs
!+-----------------------------------------------------------------------+
!#########################################################################

      subroutine gwptophi(gs2,psi)

      implicit   none

      integer(long)                                  :: m,s
      complex(dop), dimension(dgldim), intent(out)   :: psi
      complex(dop), dimension(gs2matdim), intent(in) :: gs2

      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call gwptophi1(gs2(gs2mat(m,s)),psi(zdgwp(m,s)),&
                    gwpdim(m,s),dim(m,s))
            endif
         enddo
      enddo

      return
      end subroutine gwptophi

!#######################################################################
!
! build dim1 orthonormal functions from gwpdim1 GWPs
!
!#######################################################################

      subroutine gwptophi1(gs2,dgwp,gwpdim1,dim1)

      implicit   none

      integer(long)                                   :: e,g,nzero
      integer(long), intent(in)                       :: gwpdim1,dim1
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in) :: gs2
      complex(dop), dimension(gwpdim1,dim1), intent(out)   :: dgwp
      complex(dop), dimension(gwpdim1,gwpdim1)             :: tmpdgwp

      call unitqxz(tmpdgwp,gwpdim1)
      call gwporthonorm(tmpdgwp,gs2,gwpdim1,nzero,tiny)

!
! copy DGWP elements to psi
!
      dgwp=0.0_dop
      do e=1,dim1
         do g=1,gwpdim1
            dgwp(g,e)=tmpdgwp(g,e)
         enddo
      enddo

      return
      end subroutine gwptophi1

!#######################################################################
!
! renormalise DGWP matrices. 
!
!#######################################################################

      subroutine renormdgwp(psi,gs2,dicht4)

      implicit none

      integer(long)                                  :: m,s
      complex(dop), dimension(gs2matdim), intent(in) :: gs2
      complex(dop), dimension(dmatdim), intent(in)   :: dicht4
      complex(dop), dimension(dgldim), intent(inout) :: psi

      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call renormdgwp1(psi(zdgwp(m,s)),gs2(gs2mat(m,s)),&
                    dicht4(dmat(m,s)),dim(m,s))
            endif
         enddo
      enddo

      return
      end subroutine renormdgwp

!#######################################################################

      subroutine renormdgwp1(dgwp,gs2,dicht4,dim1)

      implicit none

      integer(long)                                     :: e,e1,g,g1,i
      integer(long), intent(in)                         :: dim1
      complex(dop), dimension(dim1,dim1), intent(in)    :: gs2
      complex(dop), dimension(dim1,dim1), intent(in)    :: dicht4
      complex(dop), dimension(dim1,dim1)                :: tfm
      complex(dop), dimension(dim1,dim1), intent(inout) :: dgwp
      complex(dop), dimension(dim1,dim1)                :: dgwpno
      complex(dop)                                      :: overlap,norm
      real(dop) :: invnorm,maxovl,doverlap,maxovlo

!-----------------------------------------------------------------------
! transform to natural orbitals
! First check phases of density matrix eigenvectors: 
!      i.e, evek(e,e) > 0 so that <phi_old|pih_new> > 0
!-----------------------------------------------------------------------
      tfm=dicht4

      call mqxczz(dgwp,tfm,dgwpno,dim1,dim1)

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------
       i=0
       maxovlo = 999.0d0
       do 
         i=i+1
         maxovl = 0.0_dop
         do e=1,dim1
            do e1=1,e-1
               overlap=0.0_dop
               do g=1,dim1
                  do g1=1,dim1
                     overlap = overlap + &
                              dconjg(dgwpno(g,e1))*gs2(g,g1)*dgwpno(g1,e)
                  enddo
               enddo
               
               doverlap = abs(overlap)
               doverlap = dsqrt(doverlap)
               maxovl=max(maxovl,doverlap)
               do g=1,dim1
                  dgwpno(g,e) = dgwpno(g,e) - overlap*dgwpno(g,e1)
               enddo
            enddo

            norm=0.0_dop
            do g=1,dim1
               do g1=1,dim1
                  norm = norm + dconjg(dgwpno(g,e))*gs2(g,g1)*dgwpno(g1,e)
               enddo
            enddo
            invnorm = dble(norm)
            invnorm = dsqrt(invnorm)
            if (invnorm .le. 1.0e-99_dop ) invnorm = 1.0e99_dop

            call xvixdzo(invnorm,dgwpno(1,e),dim1)

         enddo
         if (maxovl .lt. maxovlo) then
            if (maxovl .lt. 1.0d-6) then
               exit
            else
               maxovlo = maxovl
            endif
         else if ((i .eq. 10) .or. (maxovl .gt. maxovlo)) then
            if (maxovl .gt. 1.0d-2) then
               routine = 'Renormdgwp1 (gwplib.f90)'
               write(message,'(a,i2,a,a,E12.6,a)') &
                 'Failed to orthonormalise SPFs after', i,  ' attempts. ', &
                 'Max overlap: ',maxovl,'   Linear Dependence?'
               call warnmsg1
            endif
            exit
         endif
      enddo

!-----------------------------------------------------------------------
! transform back to SPFs
!-----------------------------------------------------------------------
      call mqxtzz(dgwpno,tfm,dgwp,dim1,dim1)

      return
      end subroutine renormdgwp1

!#######################################################################
!
! Gram-Schmidt orthonormalisation of a set of GWPs
!
! nzero is no. of "zero sized" orthonormal functions in GWP space
! i.e. GWP space can support gwpdim1 - nzero orhtonormal functions
!
! ldtest: if norm < ldtest, function set to 0
!#######################################################################

      subroutine gwporthonorm(dgwp,gs2,gwpdim1,nzero,ldtest0)

      implicit none

      logical(kind=4)                                   :: lfail,lrst
      integer(long)                                     :: e,g1,g2,i,&
                                                           noffd,est
      integer(long), intent(in)                         :: gwpdim1
      integer(long), intent(out)                        :: nzero
      real(dop)                                         :: norm
      complex(dop)                                      :: ctmp,offd,ond
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in)    :: gs2
      complex(dop), dimension(gwpdim1,gwpdim1), intent(inout) :: dgwp
      complex(dop), dimension(gwpdim1,gwpdim1)          :: tmpdgwp,unitmat,gstmp
      complex(dop), dimension(gwpdim1)                  :: ctmpmat
      real(dop)                                         :: maxoffd,ldtest
      real(dop), intent(in)                             :: ldtest0

! test
!      integer(long) :: maxg1,maxg2
      integer(long), save :: tid=0
      call get_timer(tid, 'GWP orthonorm')
      call start_timer(tid)     
! test

      routine = 'gwporthonorm'
      noffd = max(1,gwpdim1*(gwpdim1-1))
      ldtest = ldtest0
      lfail = .false.

!
! "transform" overlap matrix into guess orthonormal basis set
!
!      call qqaxzz(dgwp,gs2,unitmat,gwpdim1)
      call zgemm('C','N',gwpdim1,gwpdim1,gwpdim1,(1.0_dop,0.0_dop),&
                 dgwp,gwpdim1,gs2,gwpdim1,(0.0_dop,0.0_dop),&
                 unitmat,gwpdim1)
      unitmat=matmul(unitmat,dgwp)
      gstmp = unitmat

      call unitqxz(tmpdgwp,gwpdim1)


! Loops until orthonormalisation is complete.
!
! | phi(e) > = | g( e) > - sum_{e2 < e}  | phi(e2) >< phi(e2) | g(e) >
!
! Normalised
!
! dgwp(e1,e) = S(e1,e2)^-1 < g(e2) | phi(e) >   


! normalise function 1
      ctmp=gstmp(1,1)    
      norm=dble(ctmp)
      norm=dsqrt(norm)
      if (norm .le. 1.0e-12_dop ) then
         message = 'First GWP has zero norm'
         call errormsg
      else
         norm=1.0_dop/norm
      endif
      tmpdgwp(1,1)=norm*tmpdgwp(1,1)        

      if (gwpdim1.eq.1) then
         dgwp(1,1)=tmpdgwp(1,1)        
      
      else
          est = 2
          do i = 1,100
             nzero = 0
             do e=est,gwpdim1
 
                ctmpmat=dcmplx(0.0_dop,0.0_dop)
                do g1 = 1, e-1
                   do g2 = 1, g1
                      ctmpmat(g1) = ctmpmat(g1) + dconjg(tmpdgwp(g2,g1))*gstmp(g2,e)
                   enddo
                enddo
 
                do g1 = 1,e-1
                   ctmp=dcmplx(0.0_dop,0.0_dop)             
                   do g2 = g1, e-1
                      ctmp = ctmp + ctmpmat(g2)*tmpdgwp(g1,g2)
                   enddo
                   tmpdgwp(g1,e) = tmpdgwp(g1,e) - ctmp
                enddo
 
                ctmp=dcmplx(0.0_dop,0.0_dop)
                do g1=1,e
                   do g2=1,e
                      ctmp = ctmp + dconjg(tmpdgwp(g2,e))*gstmp(g2,g1)*tmpdgwp(g1,e)
                   enddo
                enddo
                norm=dble(ctmp)
 
                if (norm .le. ldtest ) then
                   norm = 0.0_dop
                   nzero = nzero + 1
                else
                   norm=dsqrt(norm)
                   norm=1.0_dop/norm
                endif           
                tmpdgwp(:,e)=norm*tmpdgwp(:,e)
 
             enddo ! e-loop

! Update the full transformation matrix.
             dgwp = matmul(dgwp,tmpdgwp)
 
!             call qqaxzz(dgwp,gs2,unitmat,gwpdim1)
             call zgemm('C','N',gwpdim1,gwpdim1,gwpdim1,(1.0_dop,0.0_dop),&
                        dgwp,gwpdim1,gs2,gwpdim1,(0.0_dop,0.0_dop),&
                        unitmat,gwpdim1)
             unitmat=matmul(unitmat,dgwp)
             ond = dcmplx(0.0_dop,0.0_dop)
             offd = dcmplx(0.0_dop,0.0_dop)
             maxoffd = 0.0_dop
 
             lrst = .false.
             do g1 = 1,gwpdim1
                ond = ond + unitmat(g1,g1)
                do g2 = 1,gwpdim1
                   if(g1.ne.g2)then
                      offd = offd + unitmat(g1,g2)
 
                      if(abs(unitmat(g1,g2)) .gt. maxoffd)then
                         maxoffd = abs(unitmat(g1,g2))
                         if(maxoffd.gt.ldtest .and. .not. lrst)then
                            est = g2
                            lrst = .true.
                         endif
                      endif
 
                   endif
                enddo
             enddo
! test
!             write(*,*)
!             write(*,*)dgwp(1:2,2)
!             write(*,*)i,abs(offd)/(dble(noffd)),abs(ond)/(dble(gwpdim1-nzero)),maxoffd,nzero,est
!             write(*,*)maxg1,maxg2,maxoffd,unitmat(maxg1,maxg2)
!             flush(6)
! test

             if(abs(offd)/(dble(noffd)).lt.ldtest .and. maxoffd .lt. dble(gwpdim1)*ldtest)exit
 
             gstmp = unitmat
             call unitqxz(tmpdgwp,gwpdim1)
             if(mod(i,10).eq.0) ldtest = 10.0_dop*ldtest
             lfail = i.eq.100
 
          enddo
      endif

      if(lfail)then
         write(ilog,*)'Failed to orthonormalise GWPs. Avg off-diag = ',&
              offd/(dble(noffd))
         write(ilog,*)'Max off-diag ',maxoffd
         routine = 'gwporthonorm'
         message = 'Failed to orthonormalise GWPs'
         call errormsg
      endif

! test
      call stop_timer(tid)
!      stop
! test

      return
      end subroutine gwporthonorm

!#######################################################################      
!
!  ~D = S^-1 D 
!  adgwp = D^+, i.e. adgwp(dim,gwpdim) = dconjg(D(gwpdim,dim))
!
!#######################################################################      

      subroutine transahelp(psi,adgwp,gs2)

      implicit   none

      integer(long)                                         :: s,m
      complex(dop), dimension(dgldim), intent(in)           :: psi
      complex(dop), dimension(gsmatdim), intent(out)        :: adgwp
      complex(dop), dimension(gs2matdim), intent(in)        :: gs2
      complex(dop), dimension(maxspf*maxspf)                :: workc

!-----------------------------------------------------------------------
! form D from ~D = S^-1 D, then form D^+
!-----------------------------------------------------------------------
     do s=1,nstate
         do m = 1,nmode
            if (gwpm(m)) then
               call qmxxzz(gs2(gs2mat(m,s)),psi(zdgwp(m,s)),workc,&
                    gwpdim(m,s),dim(m,s))
               call cpmaz(workc,adgwp(gsmat(m,s)),gwpdim(m,s),dim(m,s))
            endif
         enddo
      enddo

      return
      end subroutine transahelp

!#######################################################################      
!
!  ~D = S^-1 D 
!  adgwp = D^+
!
!#######################################################################      

      subroutine transahelp1m(currentmode,psi,adgwp,gs2,zetf1, &
                 phidim1)

      implicit   none

      integer(long)             :: s,m
      integer(long), intent(in) :: currentmode,zetf1,phidim1

      complex(dop), dimension(zetf1:zetf1+phidim1), intent(in) :: psi
      complex(dop), dimension(gsmatdim), intent(out)           :: adgwp
      complex(dop), dimension(gs2matdim), intent(in)           :: gs2
      complex(dop), dimension(maxspf*maxspf)                   :: workc

!-----------------------------------------------------------------------
! form D from ~D = S^-1 D, then form D^+
!-----------------------------------------------------------------------
      m = currentmode
      if (gwpm(m)) then
         do s=1,nstate
            call qmxxzz(gs2(gs2mat(m,s)),psi(zdgwp(m,s)),workc,&
                 gwpdim(m,s),dim(m,s))
            call cpmaz(workc,adgwp(gsmat(m,s)),gwpdim(m,s),dim(m,s))
         enddo
      endif

      return
      end subroutine transahelp1m

!#######################################################################

      subroutine getbdot(bdot,agdot,addot,agmat,gwpdim1,gdof,gwptype,&
                 xi,xidot)

      implicit none

      integer(long)                                :: e,g,g1
      integer(long), intent(in)                    :: gwpdim1,gdof,gwptype

      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)  :: agmat
      complex(dop), dimension(gdof,gwpdim1), intent(in)       :: xi
      complex(dop), dimension(gdof,gwpdim1), intent(inout)    :: bdot
      complex(dop), dimension(gdof,gwpdim1), intent(in)       :: addot,xidot
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)  :: agdot

! Get iBdot = dt (i(-2Aq + ip))
! from i xidot = i(qdot + i pdot ), xi = ( q + i p), agdot = i dt A and agmat = A
      if (gwptype .lt. 2) then
         do e=1,gwpdim1
            do g=1,gdof
               bdot(g,e) = dble(xidot(g,e))

               do g1=1,gdof
                  bdot(g,e) = bdot(g,e) + &
                     (-2.0d0)*agdot(g,g1,e)*dble(xi(g1,e))
               enddo

               do g1=1,gdof
                  bdot(g,e) = bdot(g,e) + &
                     (0,-2.0d0)*agmat(g,g1,e)*dimag(xidot(g1,e))
               enddo
            enddo
         enddo

      else if (gwptype .eq. 2) then
         do e=1,gwpdim1
            do g=1,gdof
               bdot(g,e) = &
                     dble(xidot(g,e)) + &
                     (-2.0d0)*addot(g,e)*dble(xi(g,e)) + &
                     (0,-2.0d0)*agmat(g,g,e)*dimag(xidot(g,e)) 
            enddo
         enddo

      else if (gwptype .ge. 3) then
         do e=1,gwpdim1
            do g=1,gdof
               bdot(g,e) = &
                     (0,-2.0d0)*agmat(g,g,e)*dimag(xidot(g,e)) + &
                     dble(xidot(g,e))
            enddo
         enddo

      endif

      return
      end subroutine getbdot

!-----------------------------------------------------------------------
!
! FGAOVER
!
! calculates the overlap matrix between two sets of diagonal
! Gaussian functions
!
! over = <1 | 2>
!
! if g = exp (Ax^2 + Bx + C)
! input is with A = A, xi = q + ip, phi = C;, B = -2Aq + ip
!-----------------------------------------------------------------------

      subroutine fgaover(over,phi1,phi2,xi1,xi2,a1,a2,gwpdim1,gwpdim2,gdof)

      implicit none

      integer(long), intent(in)                            :: gdof,gwpdim1,gwpdim2
      integer(long)                                        :: e1,e2
      complex(dop), dimension(gwpdim1,gwpdim2), intent(out)   :: over
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in)  :: a1
      complex(dop), dimension(gdof,gdof,gwpdim2), intent(in)  :: a2
      complex(dop), dimension(gdof,gwpdim1), intent(in)       :: xi1
      complex(dop), dimension(gdof,gwpdim2), intent(in)       :: xi2
      complex(dop), dimension(gwpdim1), intent(in)            :: phi1
      complex(dop), dimension(gwpdim2), intent(in)            :: phi2

      integer(long), save :: tid=0
      call get_timer(tid, 'fgaover')
      call start_timer(tid)

      do e2=1,gwpdim2
         do e1=1,gwpdim1
            over(e1,e2)=ffgaover(phi1(e1),phi2(e2),xi1(1,e1),xi2(1,e2), &
                        a1(1,1,e1),a2(1,1,e2),gdof)
         enddo
      enddo

      call stop_timer(tid)
      return
      end subroutine fgaover

!-----------------------------------------------------------------------
!
! FFGAOVER
! calculates overlap between two diagonal gwps over = <1 | 2>
!
! xi = q + ip
!
!-----------------------------------------------------------------------

      complex(dop) function ffgaover(phi1,phi2,xi1,xi2,a1,a2,gdof)

      implicit none

      integer(long), intent(in)                      :: gdof
      integer(long)                                  :: f
      complex(dop), intent(in)                       :: phi1,phi2
      complex(dop), dimension(gdof), intent(in)      :: xi1,xi2
      complex(dop), dimension(gdof,gdof), intent(in) :: a1,a2
      complex(dop)                                   :: A,B,C
      real(dop)                                      :: pi

      pi = 4.0d0*atan(1.0d0)

      ffgaover = 1.0d0
      C = phi2+dconjg(phi1)

      do f=1,gdof
         A=a1(f,f)+a2(f,f)
         A = -A

         B = -2.0d0*a1(f,f)*dble(xi1(f)) -2.0d0*a2(f,f)*dble(xi2(f))
         B = B + (0.0d0,1.0d0)*(dimag(xi2(f))-dimag(xi1(f)))

         C = C + B**2/(4.0d0*A)

         ffgaover= ffgaover*sqrt(pi/A)
      enddo

      ffgaover= ffgaover*exp(C)

      return
      end function ffgaover

!-----------------------------------------------------------------------
!
! FGAOVER1
!
! calculates overlap between two diagonal gwps over = <1 | 2>
! if g = exp (Ax^2 + Bx + C) 
!
!-----------------------------------------------------------------------

      complex(dop) function fgaover1(phi1,phi2,xi1,xi2,a1,a2,gdof)

      implicit none

      integer(long), intent(in)                      :: gdof
      integer(long)                                  :: f
      complex(dop), intent(in)                       :: phi1,phi2
      complex(dop), dimension(gdof), intent(in)      :: xi1,xi2
      complex(dop), dimension(gdof,gdof), intent(in) :: a1,a2
      complex(dop)                                   :: gamma
      real(dop)                                      :: alpha,ort,mom,pi

      pi = 4.0d0*atan(1.0d0)

      fgaover1 = 1.0d0
      gamma=phi2+dconjg(phi1)

      do f=1,gdof
         alpha=dble(a1(f,f))+dble(a2(f,f))
!         write(*,*)"a1",dble(a1(f,f)),"a2",dble(a2(f,f))
         alpha=-alpha
         ort=dble(xi1(f))+dble(xi2(f))
         ort=ort/(2.0d0*alpha)
         mom=dimag(xi2(f))-dimag(xi1(f))

         gamma= gamma + alpha*ort*ort +  (0.0d0,1.0d0)*mom*ort
         fgaover1= fgaover1 * dexp(-mom*mom/(4.0d0*alpha))
         fgaover1= fgaover1*dsqrt(pi/alpha)

      enddo

      fgaover1= fgaover1* &
                (dcos(dimag(gamma)) + (0.0d0,1.0d0)*dsin(dimag(gamma)))
      fgaover1= fgaover1 * dexp(dble(gamma))
!      write(*,*)"ort",ort
!      write(*,*)"mom",mom
!      write(*,*)"alpha",alpha
!      write(*,*)dexp(dble(gamma))

      return
      end function fgaover1

!-----------------------------------------------------------------------
!
! GAUSSVAL
!
! returns value of a Gaussian function at the point x, 
!     g(x)= 1/sqrt(2*pi*sigma) * exp [-1/(2*sigma**2) (x-x0)**2]
!
!-----------------------------------------------------------------------

      real(dop) function gaussval(x,sigma,x0)

      implicit none

      real(dop) :: x,sigma,x0,pi

      pi=4.0d0*atan(1.0d0)

      gaussval=(x-x0)*(x-x0)
      gaussval=-gaussval/(2.0d0*sigma*sigma)
      gaussval=dexp(gaussval)
      gaussval=gaussval/dsqrt(2.0d0*pi*sigma)

      return
      end function gaussval

!#######################################################################
! project a set of Gaussians (defined by dimg,phig,xig,ag)
! onto a GWP (defined by dimwf,phiwf,xiwf,awf,psiwf)
! That is, form <g_i|Psi>
! only compute subset from e1 to e2. 1<=e1,e2<=dimg
!#######################################################################
      subroutine gwpproj(e1,e2,gdof, &
           dimg,phig,xig,ag, &
           dimwf,phiwf,xiwf,awf,psiwf, &
           prod)
      integer(long),intent(in) :: e1,e2,gdof,dimwf,dimg
      complex(dop), dimension(dimwf), intent(in)           :: phiwf
      complex(dop), dimension(gdof,dimwf), intent(in)      :: xiwf
      complex(dop), dimension(gdof,gdof,dimwf), intent(in) :: awf
      complex(dop), dimension(dimwf), intent(in)           :: psiwf
      complex(dop), dimension(dimg), intent(in)           :: phig
      complex(dop), dimension(gdof,dimg), intent(in)      :: xig
      complex(dop), dimension(gdof,gdof,dimg), intent(in) :: ag
      integer(long) :: e,f
      complex(dop)  :: apold(gdof,dimwf),axold(gdof,dimwf)
      complex(dop), intent(inout) :: prod(dimg)
      real(dop) :: pi4togdof
      integer(long), save :: tid=0
      call get_timer(tid, 'gwpproj')
      call start_timer(tid)
      pi4togdof=(16.0d0*atan(1.0d0))**(gdof/2.0_dop)
!$omp parallel do private(apold,axold)
      do e=e1,e2
       ! this is the FORTRAN vector version of prod(e)=sum(psiwf(1:dimwf)fgaover(g(e),gwf(1:dimwf)))
       !                                              = <g_e|Psi>
         do f=1,gdof
            apold(f,1:dimwf)=-0.25_dop/(dconjg(ag(f,f,e))+awf(f,f,1:dimwf))
            axold(f,1:dimwf)=dconjg(xig(f,e))+xiwf(f,1:dimwf)
         end do
         prod(e)=&
              sum(sqrt(product(apold(:,1:dimwf),1))*&
              exp(sum(axold(:,1:dimwf)**2*apold(:,1:dimwf),1)+&
              dconjg(phig(e))+phiwf(1:dimwf))*psiwf(1:dimwf))*pi4togdof
      end do
!$omp end parallel do
      call stop_timer(tid)
      return
      end subroutine gwpproj

!#######################################################################

      function freespacegaussian(x,sigma,q0,p0,mass,time)

        complex(dop)         :: freespacegaussian
        real(dop),intent(in) :: x,sigma,q0,p0,mass,time
        
        freespacegaussian=sqrt(sigma/(sqrt(2.0_dop*pi)*(dcmplx(sigma**2,time/(2.0_dop*mass)))))*&
              exp(-p0**2*sigma**2+p0*q0+dcmplx(2.0_dop*sigma**2*p0,x-q0)**2/&
                 dcmplx(4.0_dop*sigma**2,2.0_dop*time/mass))
      end function freespacegaussian

!#######################################################################

      subroutine printGWPSPF(npoints,gdof,Dlower,Dupper, &
                 dim1,psi1,phi1,xi1,a1, &
                 dim0,psi0,phi0,xi0,a0, &
                 xplotlims,channel)

        implicit none
      
        integer(long), intent(in) :: npoints,gdof,dim1
        real(dop),intent(in)::Dlower(*),Dupper(*)

        complex(dop), intent(in)  :: psi1(dim1),phi1(dim1)
        complex(dop), intent(in)  :: xi1(gdof,dim1),a1(gdof,gdof,dim1)

        integer(long), intent(in),optional :: dim0
        complex(dop), intent(in),optional  :: psi0(*),phi0(*)
        complex(dop), intent(in),optional  :: xi0(gdof,*),a0(gdof,gdof,*)

        real(dop),intent(in),optional :: xplotlims(:,:)
        integer(long),intent(in) :: channel


        integer(long)             :: f,f1,e,e1
        complex(dop)              :: psi,allpsi,prefac(dim1,dim1),preexp(dim1,dim1)
        complex(dop),allocatable  :: prefac1(:,:),preexp1(:,:)

        real(dop),dimension(gdof) :: x,dx,xu,xl,dl,du
        logical::test

! dummy copy for unused arguments
        dl = Dlower(1)
        du = Dupper(1)

        if (present(dim0)) then 
          allocate(prefac1(dim0,dim0))
          allocate(preexp1(dim0,dim0))
        else
          allocate(prefac1(1,1))
          allocate(preexp1(1,1))
        endif
        test = .false.
        do f=1,gdof
           test=.false.
           if (present(xplotlims)) test=(size(xplotlims)/3>=f) &
                                .and.(xplotlims(f,1) /= xplotlims(f,2))
           if (test) then
              xl(f)=xplotlims(f,1)
              xu(f)=xplotlims(f,2)
              dx(f)=xplotlims(f,3)
           else
              x(f)=dble(xi1(f,1))/(-2.0_dop*dble(a1(f,f,1)))
              dx(f)=4.0/dsqrt(-4.0_dop*dble(a1(f,f,1)))
              xl(f)=x(f)-dx(f)
              xu(f)=x(f)+dx(f)
              do e=2,dim1
                 x(f)=dble(xi1(f,e))/(-2.0_dop*dble(a1(f,f,e)))
                 dx(f)=4.0/dsqrt(-4.0_dop*dble(a1(f,f,e)))
                 if (x(f)-dx(f)<xl(f)) then
                    xl(f)=x(f)-dx(f)
                 end if
                 if (x(f)+dx(f)>xu(f)) then
                    xu(f)=x(f)+dx(f)
                 end if
              end do
              dx(f)=0.0_dop
           end if
           ! if npoints is negative, only the reduced densities are plotted and
           ! npoints is the number of points along each axis
           ! otherwise put the gdof'th root of npoints number of points in each DOF
           if (dx(f) == 0.0_dop) then
              dx(f)=1.0/abs(npoints)
              if (npoints > 0) then
                 dx(f)=dx(f)**(1.0_dop/gdof)
              end if
              dx(f)=dx(f)*(xu(f)-xl(f))
           end if
           x(f)=xl(f)
        end do

 ! F: old gaussian centres
 ! A (re, im), Q, P, w
        do e=1,dim1
!           write(channel,'(A1,2g16.7e3)',advance='no')'F',dble(psi1(e)),dimag(psi1(e))
!           write(channel,*,advance='no')'F',dble(psi1(e)),dimag(psi1(e))
!           write(channel,'(2g16.7e3)',advance='no')dble(phi1(e)),dimag(phi1(e)) !TM
!           write(channel,*,advance='no')dble(phi1(e)),dimag(phi1(e)) !TM
           do f=1,gdof
!              write(channel,'(3g12.4)',advance='no'),dble(xi1(f,e)/(-2.0_dop*dble(a1(f,f,e)))),&
              write(channel,'(A1,7f26.15)')'F',dble(psi1(e)),dimag(psi1(e)), &
                   dble(phi1(e)),dimag(phi1(e)),dble(xi1(f,e)/(-2.0_dop*dble(a1(f,f,e)))),&
                   dimag(xi1(f,e))-dble(xi1(f,e))/dble(a1(f,f,e))*dimag(a1(f,f,e)),&
                   1.0_dop/sqrt(-4.0_dop*dble(a1(f,f,e)))
           end do
           write(channel,'()')
        end do
        write(channel,'(A1)') 'F'
 ! G: new gaussian centres
 ! A (re, im), (Q, P, w, minX, maxX)
        if (present(dim0)) then
           do e=1,dim0
!              write(channel,'(A1,2g16.7e3)',advance='no')'G',dble(psi0(e)),dimag(psi0(e))
!              write(channel,*,advance='no')'G',dble(psi0(e)),dimag(psi0(e))
!              write(channel,'(2g16.7e3)',advance='no')dble(phi0(e)),dimag(phi0(e)) !TM
!              write(channel,*,advance='no')dble(phi0(e)),dimag(phi0(e)) !TM
              do f=1,gdof
!                 write(channel,'(3g12.4)',advance='no'),dble(xi0(f,e)/(-2.0_dop*dble(a0(f,f,e)))),&
                  write(channel,'(A1,7f26.15)')'G',dble(psi0(e)),dimag(psi0(e)), &
                   dble(phi0(e)),dimag(phi0(e)),dble(xi0(f,e)/(-2.0_dop*dble(a0(f,f,e)))),&
                   dimag(xi0(f,e))-dble(xi0(f,e))/dble(a0(f,f,e))*dimag(a0(f,f,e)),&
                   1.0_dop/sqrt(-4.0_dop*dble(a0(f,f,e)))
!                 write(channel,'(2g12.4)',advance='no')Dlower(f),Dupper(f)
              end do
              write(channel,'()')
           end do
           write(channel,'(A1)') 'G'
        end if

if (.not. lBEL) then
 ! I: old and new wave functions
        if (npoints > 0) then
           f=1
           do while (f .le. gdof)
              allpsi=0.0_dop
              write(channel,'(A1)',advance='no') 'I'
              do e=1,dim1
                 psi=phi1(e)
                 do f=1,gdof
                    psi=psi+x(f)**2*a1(f,f,e)+x(f)*xi1(f,e)
                    if (e == 1) then
                       write(channel,'(g12.4)',advance='no') x(f)
                    end if
                 end do
                 allpsi=allpsi+psi1(e)*exp(psi)
              end do
              write(channel,'(4e16.7e3)',advance='no') dble(allpsi),dimag(allpsi)
              if (present(dim0)) then
                 allpsi=0.0_dop
                 do e=1,dim0 
                    psi=phi0(e)
                    do f=1,gdof
                       psi=psi+x(f)**2*a0(f,f,e)+x(f)*xi0(f,e)
                    end do
                    allpsi=allpsi+psi0(e)*exp(psi)
                 end do
                 write(channel,'(4e16.7e3)',advance='no') dble(allpsi),dimag(allpsi)
              end if
              write(channel,'()')

              ! allpsi=freespacegaussian(x(f),0.0728_dop,0.0_dop,48.98979_dop,30.0_dop,time)
              ! write(channel,'(4e16.7e3)') dble(allpsi),dimag(allpsi)

              do f=1,gdof
                 x(f)=x(f)+dx(f)
                 if (x(f) <= xu(f)+tiny) then
                    exit
                 else
                    x(f)=xl(f)
                    if (f== gdof) then
                       write(channel,'(A3)') 'I E'
                    else
                       write(channel,'(A1)') 'I'
                    end if
                 end if
              end do
           enddo
        end if

        ! compute reduced density for all DOFs
 ! J: old density
        do f=1,gdof
           x(f)=xl(f)
           do e=1,dim1
              do e1=1,dim1
                 preexp(e,e1)=0.0_dop
                 prefac(e,e1)=1.0_dop
                 do f1=1,gdof
                    if (f1==f) then
                       cycle
                    end if
                    preexp(e,e1)=preexp(e,e1)-(dconjg(xi1(f1,e))+xi1(f1,e1))**2/&
                         (4.0_dop*(dconjg(a1(f1,f1,e))+a1(f1,f1,e1)))
                    prefac(e,e1)=-prefac(e,e1)*pi/(dconjg(a1(f1,f1,e))+a1(f1,f1,e1))*&
                         (4.0_dop*a1(f1,f1,e)*a1(f1,f1,e1)/pi**2)**0.25
                 end do
                 prefac(e,e1)=prefac(e,e1)
              end do
           end do
           do while (x(f) <= xu(f)+tiny)
              allpsi=0.0_dop
              do e=1,dim1
                 do e1=1,dim1
                    psi=dconjg(phi1(e))+phi1(e1)+x(f)**2*(dconjg(a1(f,f,e))+a1(f,f,e1))+&
                         x(f)*(dconjg(xi1(f,e))+xi1(f,e1))+preexp(e,e1)
                    allpsi=allpsi+prefac(e,e1)*dconjg(psi1(e))*psi1(e1)*exp(psi)
                 end do
              end do
              write(channel,'(A1,2g16.7e3)')'J',x(f),dble(allpsi)
              x(f)=x(f)+dx(f)
           end do
           write(channel,'(A1)') 'J'
        end do
        write(channel,'(A3)') 'J E'

        !do the same for the reset wave function
 ! K: new density
        if (present(dim0)) then
           do f=1,gdof
              x(f)=xl(f)
              do e=1,dim0
                 do e1=1,dim0
                    preexp1(e,e1)=0.0_dop
                    prefac1(e,e1)=1.0_dop
                    do f1=1,gdof
                       if (f1==f) then
                          cycle
                       end if
                       preexp1(e,e1)=preexp1(e,e1)-(dconjg(xi0(f1,e))+xi0(f1,e1))**2/&
                            (4.0_dop*(dconjg(a0(f1,f1,e))+a0(f1,f1,e1)))
                       prefac1(e,e1)=-prefac1(e,e1)*pi/(dconjg(a0(f1,f1,e))+a0(f1,f1,e1))*&
                         (4.0_dop*a0(f1,f1,e)*a0(f1,f1,e1)/pi**2)**0.25
                    end do
                 end do
              end do
              do while (x(f) <= xu(f)+tiny)
                 allpsi=0.0_dop
                 do e=1,dim0
                    do e1=1,dim0
                       psi=dconjg(phi0(e))+phi0(e1)+x(f)**2*(dconjg(a0(f,f,e))+a0(f,f,e1))+&
                            x(f)*(dconjg(xi0(f,e))+xi0(f,e1))+preexp1(e,e1)
                       allpsi=allpsi+prefac(e,e1)*dconjg(psi0(e))*psi0(e1)*exp(psi)
                    end do
                 end do
                 write(channel,'(A1,2g16.7e3)')'K',x(f),dble(allpsi)
                 x(f)=x(f)+dx(f)
              end do
              write(channel,'(A1)') 'K'
           end do
           write(channel,'(A3)') 'K E'
        end if
endif
        return
      end subroutine printGWPSPF

!#######################################################################

      subroutine printGWPSPF0(npoints,gdof,dim1,psi1,phi1,xi1,a1, &
                 dim0,psi0,phi0,xi0,a0,xplotlims,channel)

        implicit none
      
        integer(long), intent(in) :: npoints,gdof,dim1

        complex(dop), intent(in)  :: psi1(dim1),phi1(dim1)
        complex(dop), intent(in)  :: xi1(gdof,dim1),a1(gdof,gdof,dim1)

        integer(long), intent(in),optional :: dim0
        complex(dop), intent(in),optional  :: psi0(*),phi0(*)
        complex(dop), intent(in),optional  :: xi0(gdof,*),a0(gdof,gdof,*)

        real(dop),intent(in),optional :: xplotlims(:,:)
        integer(long),intent(in) :: channel


        integer(long)             :: f,f1,e,e1
        complex(dop)              :: psi,allpsi,prefac(dim1,dim1),preexp(dim1,dim1)

        real(dop),dimension(gdof) :: x,dx,xu,xl



        do f=1,gdof
           if ((present(xplotlims)) .and. (size(xplotlims)/3>=f) .and.&
                (xplotlims(f,1) /= xplotlims(f,2))) then
              xl(f)=xplotlims(f,1)
              xu(f)=xplotlims(f,2)
              dx(f)=xplotlims(f,3)
           else
              x(f)=dble(xi1(f,1))/(-2.0_dop*dble(a1(f,f,1)))
              dx(f)=4.0/dsqrt(-4.0_dop*dble(a1(f,f,1)))
              xl(f)=x(f)-dx(f)
              xu(f)=x(f)+dx(f)
              do e=2,dim1
                 x(f)=dble(xi1(f,e))/(-2.0_dop*dble(a1(f,f,e)))
                 dx(f)=4.0/dsqrt(-4.0_dop*dble(a1(f,f,e)))
                 if (x(f)-dx(f)<xl(f)) then
                    xl(f)=x(f)-dx(f)
                 end if
                 if (x(f)+dx(f)>xu(f)) then
                    xu(f)=x(f)+dx(f)
                 end if
              end do
              dx(f)=0.0_dop
           end if
           ! if npoints is negative, only the reduced densities are plotted and
           ! npoints is the number of points along each axis
           ! otherwise put the gdof'th root of npoints number of points in each DOF
           if (dx(f) == 0.0_dop) then
              dx(f)=1.0/abs(npoints)
              if (npoints > 0) then
                 dx(f)=dx(f)**(1.0_dop/gdof)
              end if
              dx(f)=dx(f)*(xu(f)-xl(f))
           end if
           x(f)=xl(f)
        end do

        do e=1,dim1
           write(channel,'(A1,2g16.7e3)',advance='no')'F',dble(psi1(e)),dimag(psi1(e))
           do f=1,gdof
              write(channel,'(3g12.4)',advance='no') dble(xi1(f,e)/(-2.0_dop*dble(a1(f,f,e)))),&
                   dimag(xi1(f,e))-dble(xi1(f,e))/dble(a1(f,f,e))*dimag(a1(f,f,e)),&
                   1.0_dop/sqrt(-4.0_dop*dble(a1(f,f,e)))
           end do
           write(channel,'()')
        end do
        write(channel,'(A1)') 'F'
        if (present(dim0)) then
           do e=1,dim0
              write(channel,'(A1,2g16.7e3)',advance='no') 'G',dble(psi0(e)),dimag(psi0(e))
              do f=1,gdof
                 write(channel,'(3g12.4)',advance='no') dble(xi0(f,e)/(-2.0_dop*dble(a0(f,f,e)))),&
                      dimag(xi0(f,e))-dble(xi0(f,e))/dble(a0(f,f,e))*dimag(a0(f,f,e)),&
                      1.0_dop/sqrt(-4.0_dop*dble(a0(f,f,e)))
              end do
              write(channel,'()')
           end do
           write(channel,'(A1)') 'G'
        end if

        if (npoints > 0) then
           f=1
           do while (f .le. gdof)
              allpsi=0.0_dop
              write(channel,'(A1)',advance='no') 'I'
              do e=1,dim1
                 psi=phi1(e)
                 do f=1,gdof
                    psi=psi+x(f)**2*a1(f,f,e)+x(f)*xi1(f,e)
                    if (e == 1) then
                       write(channel,'(g12.4)',advance='no') x(f)
                    end if
                 end do
                 allpsi=allpsi+psi1(e)*exp(psi)
              end do
              write(channel,'(4e16.7e3)',advance='no') dble(allpsi),dimag(allpsi)
              if (present(dim0)) then
                 allpsi=0.0_dop
                 do e=1,dim0 
                    psi=phi0(e)
                    do f=1,gdof
                       psi=psi+x(f)**2*a0(f,f,e)+x(f)*xi0(f,e)
                    end do
                    allpsi=allpsi+psi0(e)*exp(psi)
                 end do
                 write(channel,'(4e16.7e3)',advance='no') dble(allpsi),dimag(allpsi)
              end if
              write(channel,'()')

              ! allpsi=freespacegaussian(x(f),0.0728_dop,0.0_dop,48.98979_dop,30.0_dop,time)
              ! write(channel,'(4e16.7e3)') dble(allpsi),dimag(allpsi)

              do f=1,gdof
                 x(f)=x(f)+dx(f)
                 if (x(f) <= xu(f)+tiny) then
                    exit
                 else
                    x(f)=xl(f)
                    if (f== gdof) then
                       write(channel,'(A3)') 'I E'
                    else
                       write(channel,'(A1)') 'I'
                    end if
                 end if
              end do
           enddo
        end if

        ! compute reduced density for all DOFs
        do f=1,gdof
           x(f)=xl(f)
           do e=1,dim1
              do e1=1,dim1
                 preexp(e,e1)=0.0_dop
                 prefac(e,e1)=1.0_dop
                 do f1=1,gdof
                    if (f1==f) then
                       cycle
                    end if
                    preexp(e,e1)=preexp(e,e1)-(dconjg(xi1(f1,e))+xi1(f1,e1))**2/&
                         (4.0_dop*(dconjg(a1(f1,f1,e))+a1(f1,f1,e1)))
                    prefac(e,e1)=-prefac(e,e1)*pi/(dconjg(a1(f1,f1,e))+a1(f1,f1,e1))*&
                         (4.0_dop*a1(f1,f1,e)*a1(f1,f1,e1)/pi**2)**0.25
                 end do
                 prefac(e,e1)=prefac(e,e1)
              end do
           end do
           do while (x(f) <= xu(f)+tiny)
              allpsi=0.0_dop
              do e=1,dim1
                 do e1=1,dim1
                    psi=dconjg(phi1(e))+phi1(e1)+x(f)**2*(dconjg(a1(f,f,e))+a1(f,f,e1))+&
                         x(f)*(dconjg(xi1(f,e))+xi1(f,e1))+preexp(e,e1)
                    allpsi=allpsi+prefac(e,e1)*dconjg(psi1(e))*psi1(e1)*exp(psi)
                 end do
              end do
              write(channel,'(A1,2g16.7e3)')'J',x(f),dble(allpsi)
              x(f)=x(f)+dx(f)
           end do
           write(channel,'(A1)') 'J'
        end do
        write(channel,'(A3)') 'J E'

        !do the same for the reset wave function
        if (present(dim0)) then
           do f=1,gdof
              x(f)=xl(f)
              do e=1,dim0
                 do e1=1,dim0
                    preexp(e,e1)=0.0_dop
                    prefac(e,e1)=1.0_dop
                    do f1=1,gdof
                       if (f1==f) then
                          cycle
                       end if
                       preexp(e,e1)=preexp(e,e1)-(dconjg(xi0(f1,e))+xi0(f1,e1))**2/&
                            (4.0_dop*(dconjg(a0(f1,f1,e))+a0(f1,f1,e1)))
                       prefac(e,e1)=-prefac(e,e1)*pi/(dconjg(a0(f1,f1,e))+a0(f1,f1,e1))*&
                         (4.0_dop*a0(f1,f1,e)*a0(f1,f1,e1)/pi**2)**0.25
                    end do
                 end do
              end do
              do while (x(f) <= xu(f)+tiny)
                 allpsi=0.0_dop
                 do e=1,dim0
                    do e1=1,dim0
                       psi=dconjg(phi0(e))+phi0(e1)+x(f)**2*(dconjg(a0(f,f,e))+a0(f,f,e1))+&
                            x(f)*(dconjg(xi0(f,e))+xi0(f,e1))+preexp(e,e1)
                       allpsi=allpsi+prefac(e,e1)*dconjg(psi0(e))*psi0(e1)*exp(psi)
                    end do
                 end do
                 write(channel,'(A1,2g16.7e3)')'K',x(f),dble(allpsi)
                 x(f)=x(f)+dx(f)
              end do
              write(channel,'(A1)') 'K'
           end do
           write(channel,'(A3)') 'K E'
        end if

        return
      end subroutine printGWPSPF0

!#######################################################################

!      subroutine gwpspfmatSV(phigmat,xigmat,agmat, &
!                 gwpdim1,gdof,ovr,phi0,xi0,a0,psi0, &
!                 wei,work, gwpdim0)

!     implicit none
!     
!     integer(long),intent(in) :: gdof,gwpdim1,gwpdim0

!     complex(dop), dimension(gwpdim1), intent(in)           :: phigmat
!     complex(dop), dimension(gdof,gwpdim1), intent(in)      :: xigmat
!     complex(dop), dimension(gdof,gdof,gwpdim1), intent(in) :: aGmat

!     complex(dop), external                          :: fgaover1
!     complex(dop), dimension(gwpdim0),intent(in)        :: phi0
!     complex(dop), dimension(gwpdim0),intent(in)        :: psi0
!     complex(dop), dimension(gwpdim1,gwpdim1), intent(out) :: ovr
!     complex(dop), dimension(gwpdim1,gwpdim0)              :: ovr0
!     complex(dop), dimension(gdof,gwpdim0), intent(in)      :: xi0
!     complex(dop), dimension(gwpdim1,2), intent(out)    :: wei
!     complex(dop), dimension(gdof,gdof,gwpdim0), intent(in) :: a0

!     real(dop), dimension(3*gwpdim1),intent(in):: work
!     integer(long) :: ipiv(gwpdim1),info

!-----------------------------------------------------------------------
! get spf basis parameters
!-----------------------------------------------------------------------
!     call fgaover(ovr,phigmat,phigmat,xigmat,xigmat,agmat,agmat, &
!          gwpdim1,gwpdim1,gdof)

!     call fgaover(ovr0,phigmat,phi0,xigmat,xi0,agmat,a0,gwpdim1,gwpdim0,gdof)

!     call zgemm('N','N',gwpdim1,1,gwpdim0,dcmplx(1.0_dop,0.0_dop),&
!          ovr0,gwpdim1,psi0,gwpdim0,dcmplx(0.0_dop,0.0_dop),wei,gwpdim1)

!     call zhesv('U',gwpdim1,1,ovr,gwpdim1,ipiv,wei,gwpdim1,work,3*gwpdim1,info)

!     return
!     end subroutine gwpspfmatSV

!#######################################################################

      subroutine ovlgspf(ovl,dgwp,gs2,gwpdim1,dim1)

      implicit none

      integer(long), intent(in)      :: gwpdim1,dim1
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in) :: gs2
      complex(dop), dimension(gwpdim1,dim1), intent(in)    :: dgwp
      complex(dop), dimension(dim1,dim1), intent(out)      :: ovl
      integer(long)                  :: e,e1,g,g1

      ovl=0.0_dop

      do e1=1,dim1
         do e=1,dim1
            do g1=1,gwpdim1
               do g=1,gwpdim1
                  ovl(e,e1) = ovl(e,e1) + &
                      dconjg(dgwp(g,e))*gs2(g,g1)*dgwp(g1,e1)
               enddo
            enddo
         enddo
      enddo

      end subroutine ovlgspf

!***********************************************************************

      subroutine renormgwp(psi,agmat)

      implicit none

      integer(long)                                  :: s,m
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(agmatdim), intent(inout)  :: agmat ! yapol

      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
               call normgwp(nspfdof(m),gwpdim(m,s),agmat(zagmat(m,s)),&
                    agmat(zxivec(m,s)),agmat(zphase(m,s)))
               if (lgwpphase) then
                  psi(zphi(m,s):zphi(m,s)+gwpdim(m,s)-1) = &
                              agmat(zphase(m,s):zphase(m,s)+gwpdim(m,s)-1)
               endif
            endif
         enddo
      enddo

      return
      end subroutine renormgwp

!***********************************************************************
! 
! RENORMGWPA
!
! renormalises A-vector for a G-MCTDH WF
!
!***********************************************************************

      subroutine renormgwpa(psi,gs2)

      implicit none

      integer(long)                                  :: s,m,zeig,zeig1,zeig2
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(gs2matdim), intent(in) :: gs2
      complex(dop), dimension(2*maxblock)            :: workc
      complex(dop)                                   :: sum
      real(dop)                                      :: norm

!-----------------------------------------------------------------------
! Calculate norm
!-----------------------------------------------------------------------
      norm=0.0_dop
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         call cpvxz(psi(zpsi(s)),workc(zeig),block(s))
         do m=1,nmode
            if(gwpm(m))then
              call qtxxzz(gs2(gs2mat(m,s)), &
                          workc(zeig),workc(zeig1), &
                          vdim(m,s),dim(m,s),ndim(m,s))
              zeig2=zeig
              zeig=zeig1
              zeig1=zeig2
            endif
         enddo
         call vvaxzz(psi(zpsi(s)),workc(zeig),sum,block(s))
         norm=norm+dble(sum)
      enddo
      norm = dsqrt(norm)
      if(abs(norm).lt.1.0e-12_dop)then
         routine='renormgwpa'
         message='A-vector has zero norm'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Normalise. 
! If CAP, only normalise if norm > 1
!-----------------------------------------------------------------------
      if (lcap .and. (norm .lt. 1.0_dop)) then  
         continue
      else
         call xvixdzo(norm,psi,adim)
      endif

      return
      end subroutine renormgwpa

!-----------------------------------------------------------------------

      subroutine renormgwpBEL(psi,gs2)

      implicit none

      integer(long)                                  :: s,m,zeig,zeig1,zeig2
      complex(dop), dimension(dgldimmax), intent(inout) :: psi
      complex(dop), dimension(gs2matdim), intent(in) :: gs2
      complex(dop), dimension(2*maxblock)            :: workc
      complex(dop)                                   :: sum
      real(dop)                                      :: norm


      norm=0.0_dop
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         call cpvxz(psi(zpsi(s)),workc(zeig),block(s))
         do m=1,nmode
            if(gwpm(m))then
              call qtxxzz(gs2(gs2mat(m,s)), &
                          workc(zeig),workc(zeig1), &
                          vdim(m,s),dim(m,s),ndim(m,s))
              zeig2=zeig
              zeig=zeig1
              zeig1=zeig2
            endif
         enddo
         call vvaxzz(psi(zpsi(s)),workc(zeig),sum,block(s))
         norm=norm+dble(sum)
      enddo
      norm = dsqrt(norm)
      if(abs(norm).lt.1.0e-12_dop)then
         routine='renormgwpa'
         message='A-vector has zero norm'
         call errormsg
      endif

      call xvixdzo(norm,psi,adim)
      write(1000,*)"norm",norm,"/source/mctdhlib/gwplib.f90"
      return
      end subroutine renormgwpBEL

!***********************************************************************

      subroutine renormgwptrja(psi,gs2,gwpdim1)

      implicit none

      integer(long)                                  :: zeig,zeig1,zeig2
      integer(long)                                  :: e,e1
      integer(long), intent(in)                      :: gwpdim1
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(2*maxblock)            :: workc
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in) :: gs2
      complex(dop), dimension(gwpdim1,gwpdim1)             :: ovl
      complex(dop)                                   :: sum
      real(dop)                                      :: norm

      do e=1,gwpdim1
         ovl(e,e) = gs2(e,e)*gwptrjeovl(e,e)
         do e1=e+1,gwpdim1
            ovl(e1,e) = gs2(e1,e)*gwptrjeovl(e1,e)
            ovl(e,e1) = dconjg(ovl(e1,e))
         enddo
      enddo

      norm=0.0_dop
      zeig=1
      zeig1=block(1)+1
      call cpvxz(psi(1),workc(zeig),block(1))
      call qvxxzz(ovl,workc(zeig),workc(zeig1),gwpdim1)
      zeig2=zeig
      zeig=zeig1
      zeig1=zeig2
      call vvaxzz(psi(zpsi(1)),workc(zeig),sum,block(1))
      norm=norm+dble(sum)
      
      norm = dsqrt(norm)
      if(abs(norm).lt.1.0e-12_dop)then
         routine='renormgwptrja'
         message='A-vector has zero norm'
         call errormsg
      endif

      call xvixdzo(norm,psi,adim)

      return
      end subroutine renormgwptrja

!#######################################################################
!
! ROTCOEFFGWP / ROTCOEFFGWP1
!
! For local diabatisation:
!   A^(adia) = U^dag A^(dia).
!
! psi: on input A^(dia) and on output A^(adia).
!
! Subroutine rotcoeffgwp calls subroutine rotcoeffgwp1.
! It works together with diabat2 (opfuncs/diabatmod).
!
! A^(dia) can can occur to be non-normalised (because diabatic states have
! different rotation angles and are not orthogonal) but A^(adia) should
! be normalised (because adiabatic states are orthogonal).
! Renormalisation is performed after rotation back to the adiabatic
! representation rather than before for this reason.
!
! To be improved:
! - gpoint and diab2geo are identical at the moment but this could be
!   generalised upon using extrapolations and shifts(?);
! - could be generalised to other cases than crude adiabatic;
! - a generalised overlap (GWP and electronic states) could be used
!   to account for non-orthogonal states;
! - currently works only for single set (compulsory), two states,
!   and a single fully-combined mode (could be generalised);
! - uses saved PES (to second order) at previous point for each GWP but
!   DB could be used (extrapolations and shifts?);
! - decision on adding/removing pi (Berry phase) could be made
!   based on the sign of the derivative of the angle (would require
!   storing angle at previous step);
! - determine which "use only" are required.
! BL 06/16
!#######################################################################

      subroutine rotcoeffgwp(time,psi,adim1)

use global
use psidef
use griddatmod

      implicit none

      integer(long), intent(in)                        :: adim1
      real(dop), intent(in)                            :: time
      complex(dop), dimension(adim1), intent(inout)    :: psi

      integer(long)                                    :: m,s
      real(dop), dimension(qcentdim)                   :: qcenter
      complex(dop), dimension(agmatdim)                :: agmat
      complex(dop), dimension(gfgwpdim)                :: gfgwp
      complex(dop), dimension(aigwpdim)                :: aigwp
      complex(dop), dimension(zgwpdim)                 :: zgwp

!-----------------------------------------------------------------------
! Get new geometry for each GWP.
! Note: this is redundant with saved geometries but could be useful
!       for further generalisation.
!-----------------------------------------------------------------------
      call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,qcenter,0)

!-----------------------------------------------------------------------
! Rotate A-coefficients for m and s (in practice, nstate=1 and nmode=1).
!-----------------------------------------------------------------------
      do s=1,nstate
         do m = 1,nmode
            if (gwpm(m)) then
               call rotcoeffgwp1(time,psi(zpsi(s)),block(s), &
                                 qcenter(zcent(m,s)),nspfdof(m),dim(m,s),m)
            endif
         enddo
      enddo

      return

      end subroutine rotcoeffgwp

!#######################################################################

      subroutine rotcoeffgwp1(time,psi1,block1,qcenter1,nspfdof1,dim1,m1)

use global
use channels
use dirdyn
use griddatmod

      implicit none

      integer(long), intent(in)                       :: block1,nspfdof1,dim1,m1
      real(dop), intent(in)                           :: time
      real(dop), dimension(nspfdof1,dim1), intent(in) :: qcenter1
      complex(dop), dimension(block1), intent(inout)  :: psi1

      integer(long)                                   :: e,s,s1,g,g1
      real(dop)                                       :: dr,dr1,x,y,z,d,x0,y0,z0,d0,dv,phi
      real(dop), dimension(ndofdd)                    :: gpoint
      real(dop), dimension(nddstate)                  :: v,v0
      real(dop), dimension(nddstate,nddstate)         :: h
      complex(dop)                                    :: a1,a2

!
! Check time is consistent with diabat2.
!
      do e=1,dim1
         if (abs(time-diab2tim(e)) .ge. tiny) then
            if (ddlog) write(ilog,'(a,2i7,f12.6)') &
               'Inconsistent diabatisation time for GWP, mode, time: ',e,m1,time/fs
            routine = 'rotcoeffgwp1'
            message = 'Inconsistent diabatisation time'
            call errormsg
         endif
      enddo

!-----------------------------------------------------------------------
! Diabatic Hamiltonian centred at old point (t) and extrapolated to new
! point (t + dt being the current time here).
! For each GWP, get new point in system coordinates, estimate diabatic
! Hamiltonian matrix elements (gradient and Hessian in system coordinates),
! update diabatic-to-adiabatic angle, and rotate A-coefficients.
! Possible change: DB coordinates could be used instead of system coordinates.
!-----------------------------------------------------------------------
      h(:,:)=0.0_dop
      gpoint(:)=0.0_dop
      do e=1,dim1
         call gwp2gpoint(qcenter1(m1,e),gpoint,nspfdof1,ndofdd,m1)
!
! Estimate diabatic Hamiltonian at current point
! (quadratic extrapolation around previous point).
!
         do s=1,nddstate
            do s1=s,nddstate
               h(s1,s)=diabopes(s1,s,e)
               do g=1,ndofdd
                  dr=gpoint(g)-diabogeo(g,e)
                  h(s1,s)=h(s1,s)+diabogra(g,s1,s,e)*dr
                  do g1=1,ndofdd
                     dr1=gpoint(g1)-diabogeo(g1,e)
                     h(s1,s)=h(s1,s)+0.5_dop*dr1*diabohes(g1,g,s1,s,e)*dr
                  enddo
               enddo
            enddo
         enddo
         do s=1,nddstate-1
            do s1=s+1,nddstate
               h(s,s1)=h(s1,s)
            enddo
         enddo
!
! Check errors on current eigenvalues (for two states only).
! N.B.: - diab2pes is diagonal at the moment but the current procedure
!         sould be able to complies with more general situations;
!       - no need to regularise h because this would not alter the
!         value of the angle (same factor for x and y).
! Possible change: add test if small energy difference.
!
         x=0.5_dop*(h(2,2)-h(1,1))
         y=h(1,2)
         z=0.5_dop*(h(2,2)+h(1,1))
         d=sqrt(x**2+y**2)
         v(1)=z-d
         v(2)=z+d
         x0=0.5_dop*(diab2pes(2,2,e)-diab2pes(1,1,e))
         y0=diab2pes(1,2,e)
         z0=0.5_dop*(diab2pes(2,2,e)+diab2pes(1,1,e))
         d0=sqrt(x0**2+y0**2)
         v0(1)=z0-d0
         v0(2)=z0+d0
         do s=1,nddstate
            if ((abs(v(s)-v0(s)) .ge. 1.d-3) .and. ddlog) &
               write(ilog,'(a,2i7,f12.6,i7)') &
               'Diabatisation producing large eigenvalue error for GWP, mode, time, state: ', &
               e,m1,time/fs,s
         enddo
!
! Calculate rotation angle, check for crossing, and correct if predicted and
! confirmed crossings do not match.
! Energy difference should never be negative (test on permuted states) at the
! moment but this may change in more general procedures.
! Possible changes: use previously saved angle to determine if Berry phase
!                   should be added; re-orient BS if undefined angle;
!                   check when cumulative angles reaches multiples of pi/2.
!
         dv=0.5_dop*(diabopes(2,2,e)-diabopes(1,1,e))
         if (dv .lt. 0.0_dop) then
            x=-x
            if (ddlog) write(ilog,'(a,2i7,f12.6)') &
               'Permuted states for GWP, mode, time: ',e,m1,time/fs
         endif
         if ((abs(x) .lt. 1.d-7) .and. (abs(y) .lt. 1.d-7)) then
            phi=0.0_dop
            if (ddlog) write(ilog,'(a,2i7,f12.6)') &
               'Undefined angle for GWP, mode, time: ',e,m1,time/fs
         else
            phi=-0.5_dop*atan2(y,x)
         endif
         if (x .lt. 0.0_dop) then
            diab2pre(e)=-diab2pre(e)
            if (ddlog) write(ilog,'(a,2i7,f12.6)') &
               'Crossing predicted for GWP, mode, time: ',e,m1,time/fs
         endif
         if (diab2det(e)*diab2pre(e) .lt. 0) then
            if (y .lt. 0.0_dop) then
               phi=-phi+0.5_dop*pi
            else
               phi=-phi-0.5_dop*pi
            endif
            diab2pre(e)=-diab2pre(e)
            if (ddlog) write(ilog,'(a,2i7,f12.6)') &
               'Crossing fixed for GWP, mode, time: ',e,m1,time/fs
         endif
!
! Rotate coefficients.
!
         a1=cos(phi)*psi1(e)+sin(phi)*psi1(e+dim1)
         a2=-sin(phi)*psi1(e)+cos(phi)*psi1(e+dim1)
         psi1(e)=a1
         psi1(e+dim1)=a2
         diab2ang(e)=diab2ang(e)+phi
         if (ddlog) write(ilog,'(a,2i7,f12.6,f16.2)') &
            'Cumulative angle (in deg.) for GWP, mode, time:', &
            e,m1,time/fs,diab2ang(e)*180.0_dop/pi
      enddo
!
! Save current geometry and diabatic Hamiltonian to be used on next time step.
!
      diabogeo=diab2geo
      diabopes=diab2pes
      diabogra=diab2gra
      diabohes=diab2hes

      return

      end subroutine rotcoeffgwp1

!###########################################################################
! Subroutine to extract a single element of the width matrix.
! Helps by converting 1D AGMat in mc array to 3D array.
! GWR 12/17
!###########################################################################

      subroutine gwp_getwidth(agmat,ag,gdof,gwpdim1,f,f1,ig)

      implicit none

      integer(long), intent(in) :: gdof,gwpdim1,f,f1,ig
      complex(dop), intent(out) :: ag
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in) :: agmat

      ag = agmat(f,f1,ig)

      return

      end subroutine gwp_getwidth

!#######################################################################
! Check GWPs with periodic boundaries. If outside range move back in.
!#######################################################################

      subroutine chkpergwp(psi)

      implicit none

      integer(long)    :: m,s
      complex(dop), dimension(dgldim), intent(inout) :: psi

      do m=1,nmode
         do s=1,nstate
            call chkpergwp1m(psi(zxi(m,s)),m,nspfdof(m),gwpdim(m,s))
         enddo
      enddo

      return
      end subroutine chkpergwp

!#######################################################################

      subroutine chkpergwp1m(gwpxi,m,gdof,dim1)

      implicit none

      integer(long), intent(in)  :: m,gdof,dim1
      integer(long)              :: n,f,e
      real(dop)                  :: xi,xf,xrange,qgwp
      complex(dop), dimension(dgldim), intent(inout) :: gwpxi(gdof,dim1)

      do n=1,nspfdof(m)
         f = spfdof(n,m)
         if (ipbaspar(2,f) .eq. 0) cycle

         xi = rpbaspar(1,f)
         xf = rpbaspar(2,f)
         xrange = xf-xi

         do e = 1,dim1
            qgwp = dble(gwpxi(f,e))
            if (qgwp .gt. xf) then
               gwpxi(f,e) = gwpxi(f,e) - xrange
            else if (qgwp .lt. xi) then
               gwpxi(f,e) = gwpxi(f,e) + xrange
            endif
         enddo
      enddo

      return
      end subroutine chkpergwp1m

!#######################################################################
! 
! GWP_DISTIN
!
! List of possible distributions for GWP basis read from input
! called from eb_gwp, eb_hogwp and einingeom
!
!#######################################################################

      subroutine gwp_distin(i,typ,np,nq,noq0)

      implicit none

      integer(long), intent(inout) :: i,typ,np,nq,noq0
      integer(long)                :: ierr,ierr1

      noq0 = 0
      if (keyword(i).eq.'seqex') then
         typ = 1
      else if (keyword(i).eq.'seqex-noq0') then
         typ = 1
         noq0 = 1
      else if (keyword(i).eq.'simex') then
         typ = 2
      else if (keyword(i).eq.'shell') then
         typ = 3
      else if (keyword(i).eq.'shell-noq0') then
         typ = 3
         noq0 = 1
      else if (keyword(i).eq.'readmd') then
         typ = 4
      else if (keyword(i).eq.'wigner' .or. keyword(i) .eq. 'random_pq') then
         typ = 5
      else if (keyword(i).eq.'random') then
         typ = 6
      else if &
        (keyword(i).eq.'wigner_sym' .or. keyword(i) .eq. 'symmetric_pq') then
         typ = 7
      else if (keyword(i)(1:6).eq.'pqgrid') then
         typ = 8
         i=i+1
         if(keyword(i) .eq. ',' .and. keyword(i+2) .eq. ',') then
            read(keyword(i+1),*,iostat=ierr) nq
            read(keyword(i+3),*,iostat=ierr1) np
            if (ierr .ne. 0  .or. ierr1 .ne. 0) then
               routine='gwp_dist (eininwfmod.F90)'
               message = 'Incorrect definition for pqgrid'
               call warnmsg
            endif
            i = i+3
         else
            routine='eb_gwp (eininwf_util.F)'
            message = "Need grid dimensions with gwp pqgrid keyword"
            call errormsg
         endif

      else
         routine='eb_gwp (eininwf_util.F)'
         write(message,'(2a)') &
             "Unknown GWP distribution type: ",trim(keyword(i))
         call errormsg
      endif

      end subroutine gwp_distin


      end module gwplib

