!***********************************************************************
!                               
!                           HPHI 
!                              
! MCTDH module containing subroutines to evaluate the action of the
! Hamiltonian on the single-particle functions
!
! hphi:         Computes for all modes and all states the action of    
!               the correlated Hamiltonian on all the single-particle
!               functions. 
! hphi1m:       Computes the action of the correlated Hamiltonian on 
!               all single-particle functions for one mode.
! hunphi:       Computes for all modes and all states the action of    
!               the uncorrelated Hamiltonian on all the single-particle
!               functions.
! hunphi1ms:    Computes the action of the  uncorrelated Hamiltonian on
!               all single-particle functions for one mode and one state.
! hphi1mk:      Computes the action of one term of the Hamiltonian on 
!               the single-particle functions of one mode.
! hcall:        Computes the action of one term of the Hamiltonian on 
!               one single-particle function.
! hop:          Computes the action of an operator on a single-particle
!               functions.
! fftop, fftiop, fftzop: auxilliary routines for hop
!
!***********************************************************************

      module hphimod

use decimal, only: dop,long
use global
use globmemmod
use logdat
use dvrdatmod
use griddatmod
use psidef
use hpsimod
use operdef

use timing
use xvlib,   only: xmxxzza
use mtlib,   only: mttxzz, qtxxyz, qtxxdz, dtxxyz, dtxxdz
use fftmod,  only: fft
use op1lib,  only: zerovxl, zeromxz, cpvxz
use sdlib,   only: otxxyz, otxxdz
use op2lib,  only: addmxxzo

      implicit none
      private
      public :: hphi,hunphi,hunphi1ms,hphi1m,hphi1mk,hcall,fftop,fftiop, &
                fftzop,shiftk
      contains

! *** V6.6 GW ***************> SUBROUTINE HPHI <************************
!                             ~~~~~~~~~~~~~~~~~
! Computes for all modes, all states, and all Hamiltonian terms the
! action of the correlated Hamiltonian on the single-particle functions.
!***********************************************************************

      subroutine hphi(psi,hpsi,ktype,nham)

use griddatmod
use psidef
use hpsimod
use openmpmod

      implicit none

      integer(long)                                 :: m
      integer(long),save                            :: tid=0
      integer(long), intent(in)                     :: ktype,nham
      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(hpsidim), intent(out) :: hpsi

      call get_timer(tid,'HPhi')
      call start_timer(tid)

! --- LOOP OVER EACH MODE AND STATE ---

      do m = 1,nmode
         call hphi1m(psi(zetf(m,1)),hpsi,m,zetf(m,1), &
              phidim(m),ktype,nham)
      enddo

      call stop_timer(tid)

      return
      end subroutine hphi

! *** V6.6 GW **************> SUBROUTINE HPHI 1M <**********************
!                            ~~~~~~~~~~~~~~~~~~~~

! Computes for one mode the action of the correlated Hamiltonian on the
! single-particle functions.
!
! ktype governs which terms are included:
! ktype = 0 : all terms
! ktype = 1 : Potential terms are excluded (i.e. only kinop terms 
!             calculated)
! nham      : Number of operator to be used
!

      subroutine hphi1m (psi,hpsi,m,zetf1,phidim1,ktype,nham)

use maxv, only: maxkoe
use dvrdatmod
use griddatmod
use psidef
use operdef
use hpsimod
use runpropmod

      implicit none

      integer(long)                                             :: k,k1,k2, &
                                                                   k3,f,s1
      integer(long), intent(in)                                 :: m,ktype, &
                                                                   nham,zetf1,&
                                                                   phidim1
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(in):: psi
      complex(dop), dimension(hpsidim), intent(out)             :: hpsi
      logical(kind=4), dimension(maxkoe)                        :: kcalc

! --- IF GWP or X-local mode, do not calculate

      f = spfdof(1,m)
      if(basis(f).eq.25 .or. basis(f) .eq. 26 .or. basis(f) .eq. 27) &
         return

!-----------------------------------------------------------------------
! first find which terms need to be calculated
!-----------------------------------------------------------------------
      k1=zham(nham)      
      k2=zham(nham)+kzahl(nham)+nmulpot(nham)-1
      call zerovxl(kcalc(k1),khtot(nham))

      if (ktype .eq. 0) then
         do k=k1,k2
            if ((pmat(m,k) .gt. 0) .and. (k .eq. kfirst(m,k)))  &
               kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
         enddo
      else if (ktype .eq. 1) then
         do k=k1,k2
            if (kinop(m,k) .eq. 1) kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
         enddo
      endif

! --- LOOP OVER EACH HAMILTONIAN TERM ---

      do k=k1,k2
         s1=ki(k)
         if (kcalc(k)) then
            call hphi1mk(psi(zetf(m,s1)),hpsi(zhpsi(m,k)), &
                 m,k,dim(m,s1),subdim(m))
         endif
      enddo

      if (dentype .eq. 2) then
!     pure-left and pure-right acting terms
         call zerovxl(kcalc(k1),kzahl(nham))
         if (ktype .eq. 1) then
!     for spf-integration only
            do k=k1,k2
               if ((kact(k).eq.2) .and. (k.lt.ktwin(k))) then
                  k3 = ktwin(k)
                  if (kinop(m,k).eq.1 .or.kinop(m,k3).eq.1) &
                     kcalc(k)=.true.
               endif
            enddo
         else
            if (ldproj) then
               do k=k1,k2
                  if ((kact(k).eq.2) .and. (k.lt.ktwin(k)))  &
                      kcalc(k) = .true.
               enddo
            else
               do k=k1,k2
                  if ((kact(k).eq.2) .and. (k.lt.ktwin(k)))  &
                       kcalc(k) = .true.
                  if ((kact(k).eq.3) .and. (k.lt.ktwin(k)))  &
                       kcalc(k) = .true.
               enddo
            endif
         endif

         do k=k1,k2
           if (kcalc(k)) then
               k3 = ktwin(k)
!     check if already calculated
               if ((.not. diag(m,k3)) .and. (.not. diag(m,k))) then
                  s1 = ki(k3)
                  call hphi1mk(hpsi(zhpsi(m,k3)), &
                       hpsi(zhpsi2(m,k)),m,k,dim(m,s1),subdim(m))
               endif
            endif 
         enddo              
      endif

      return
      end subroutine hphi1m

!#######################################################################
!
! Module HUNPHI
!
! contains hunphi, hunphi1ms
!
!
! hktype = 0 : all terms
!        = 1 : only hermitian terms
!        = 2 : CAP terms
!
!#######################################################################

! *** V6.6 GW ***************> SUBROUTINE HUNPHI <**********************
!                             ~~~~~~~~~~~~~~~~~~~
!
! Computes for all modes and all states the action of the uncorrelated
! Hamiltonian on the single-particle functions.
! Result is returned as a sum of all terms in hunpsi.

      subroutine hunphi (psi,hunpsi,hktype,nham)

use griddatmod
use psidef
use lctmod
use operdef

      implicit none

      integer(long)                                       :: f,m,s
      integer(long), save                                 :: tid=0
      integer(long), intent(in)                           :: hktype,nham
      complex(dop), dimension(dgldim), intent(in)         :: psi
      complex(dop), dimension(adim+1:dgldim), intent(out) :: hunpsi

      call get_timer(tid,'HunPhi')
      call start_timer(tid)

! --- LOOP OVER EACH MODE AND STATE ---

      do m = 1,nmode

         if (m .eq. mpb) cycle
! ... exclude Gaussian mode :
         if (gwpm(m)) cycle

         f = spfdof(1,m)
         if(idmode(m) .eq. 0) then
            do s = 1,nstate
               call hunphi1ms(psi(zetf(m,s)),hunpsi(zetf(m,s)), &
                    m,s,dim(m,s),subdim(m),hktype,nham)
            enddo
         endif
      enddo

      call stop_timer(tid)

      return
      end subroutine hunphi

! *** V6.6 GW *************> SUBROUTINE HUNPHI 1MS <********************
!                           ~~~~~~~~~~~~~~~~~~~~~~~
!
! Computes for one mode and one state the action of the uncorrelated
! Hamiltonian on the single-particle functions.
!
! hktype governs which terms are included:
! hktype = 0 : all terms
! hktype = 1 : exclude CAP terms
! hktype = 2 : only CAP terms
!
! nham      : Number of operator to be used
!
!#######################################################################

      subroutine hunphi1ms (psi,hunpsi,m,s,dim1,sdm,hktype,nham)

use maxv, only: maxkoe
use dvrdatmod
use griddatmod
use psidef
use operdef

      implicit none

      integer(long)                                  :: k,k1,k2,h
      integer(long), intent(in)                      :: m,s,dim1,sdm,hktype,nham
      complex(dop), dimension(sdm,dim1), intent(in)  :: psi
      complex(dop), dimension(sdm,dim1), intent(out) :: hunpsi
      complex(dop), dimension(sdm,dim1)              :: tmphpsi
      logical(kind=4), dimension(maxkoe)             :: kcalc

      if (gwpm(m)) return

!     sdm = subdim(m), dim1=dim(m,s)
      call zeromxz(hunpsi,sdm,dim1)

!-----------------------------------------------------------------------
! first find which terms need to be calculated
! autocap = 0 : normal hermitian operator
!         > 0 : non-active CAP 
!         < 0 : active CAP 
! nstatecap(k) : state on which CAP is active. If 0 all states.
!-----------------------------------------------------------------------
      if(khzahl(m,s,nham).eq.0) return
      k1=zhun(m,s,nham)
      k2=zhun(m,s,nham)+khzahl(m,s,nham)-1

      do k=k1,k2
         if (hktype .eq. 0) then
            if (autocap(k) .eq. 0) then
               kcalc(k)=.true.
            else if (autocap(k) .lt. 0 .and. nstatecap(k) .eq. 0) then
               kcalc(k)=.true.
            else if (autocap(k) .lt. 0 .and. nstatecap(k) .eq. s) then
               kcalc(k)=.true.
            endif
         else if (hktype .eq. 1 .and. autocap(k) .eq. 0) then
            kcalc(k)=.true.
         else if (hktype .eq. 2 .and. autocap(k) .lt. 0) then
            if (nstatecap(k) .eq. 0 .or. nstatecap(k) .eq. s) kcalc(k)=.true.
         else
            kcalc(k)=.false.
         endif
      enddo

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---
      do k=k1,k2
         if (kcalc(k)) then
            call hphi1mk(psi,tmphpsi,m,k,dim1,sdm)

            h=hamilton(fkoe,k)

            if (htmtype(h) .eq. 1) then
               call addmxxzo(tmphpsi,hunpsi,sdm,dim1)
            else
               call xmxxzza(koeff(k),tmphpsi,hunpsi,sdm,dim1)
            endif

         endif
      enddo

      return
      end subroutine hunphi1ms

!#######################################################################

! *** V6.6 GW **************> SUBROUTINE HPHI 1MK <*********************
!                            ~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and ONE Hamiltonian term the action
! of the correlated Hamiltonian on the single-particle functions.

      subroutine hphi1mk (psi,hpsi,m,k,dim,subdim)

      implicit none

      integer(long)                                    :: e
      integer(long), intent(in)                        :: k,dim,subdim,m
      complex(dop), dimension(subdim,dim), intent(in)  :: psi
      complex(dop), dimension(subdim,dim), intent(out) :: hpsi

! --- LOOP OVER EACH HAMILTONIAN TERM ---

      do e = 1,dim
         call hcall(m,k,hpsi(1,e),psi(1,e),subdim)
      enddo

      return
      end subroutine hphi1mk

!#######################################################################

      subroutine hcall(m,k,hpsi,psi,subdim1)

      implicit none

      integer(long)                                 :: f,n,n1,h,jLMR
      integer(long), intent(in)                     :: m,k,subdim1
      complex(dop), dimension(subdim1), intent(in)  :: psi
      complex(dop), dimension(subdim1), intent(out) :: hpsi
      complex(dop), dimension(subdim1)              :: tmphpsi

! do not calculate non-active CAP terms
      if (autocap(k) .gt. 0) then
         hpsi=0.0
         return
      endif

      n=1
      f=spfdof(n,m)
      h=hamilton(f,k)

      if(h.lt.0) then  ! an LMR term is found
         jLMR = -h
         h=hamLMR(jLMR,1)
         call hop(f,h,k,hpsi,psi,mr(mrhops),subdim1, &
                  mi(mifftfak),mc(mchin),mc(mcrueck),mc(mcexphin), &
                  mc(mcexprueck))
         h=hamLMR(jLMR,2)
         if(h.gt.0) then
            call cpvxz(hpsi,tmphpsi,subdim1)
            call hop(f,h,k,hpsi,tmphpsi,mr(mrhops),subdim1, &
                  mi(mifftfak),mc(mchin),mc(mcrueck),mc(mcexphin), &
                  mc(mcexprueck))
         endif
         h=hamLMR(jLMR,3)
         call cpvxz(hpsi,tmphpsi,subdim1)
         call hop(f,h,k,hpsi,tmphpsi,mr(mrhops),subdim1, &
                  mi(mifftfak),mc(mchin),mc(mcrueck),mc(mcexphin), &
                  mc(mcexprueck))

      else    ! Normal procedure, no LMR
         call hop(f,h,k,hpsi,psi,mr(mrhops),subdim1,&
                   mi(mifftfak),mc(mchin),mc(mcrueck),mc(mcexphin),&
                   mc(mcexprueck))

!     if( htmdof(h) .eq. 0 ) return
!     NEWKLEG: the above is just an optimization,
!     for multi-mode operators the following loop won't actually
!     do anything because hamilton(f,k) is the unit operator for
!     all but the first DOF of the mode.
!     But if multiple KLegs are in one mode, this optimization
!     would skip the "later" KLeg operators.

         do n1=2,nspfdof(m)
            f=spfdof(n1,m)
            h=hamilton(f,k)
            if (htmtype(h) .ne. 1) then
               call cpvxz(hpsi,tmphpsi,subdim1)
               call hop(f,h,k,hpsi,tmphpsi,mr(mrhops),subdim1,&
                        mi(mifftfak),mc(mchin),mc(mcrueck),mc(mcexphin),&
                        mc(mcexprueck))
            endif 
         enddo
      endif 

      return
      end subroutine hcall

!***********************************************************************

      subroutine hop (f,h,k,hpsi,psi,hops,subdim1,fftfak,hin,rueck, &
                 exphin,exprueck)

use dvrdatmod
use griddatmod
use psidef
use operdef

use operutil, only: getvgn

      implicit none

      integer(long)                                 :: g,h,m,dm, &
                                                       z1,z2,z3,vd,gd,nd
      integer(long), intent(in)                     :: subdim1,f,k
      integer(long), dimension(fftdim), intent(in)  :: fftfak
      real(dop), dimension(hopsdim), intent(in)     :: hops
      complex(dop), dimension(subdim1), intent(in)  :: psi
      complex(dop), dimension(subdim1), intent(out) :: hpsi
      complex(dop), dimension(fftdim), intent(in)   :: hin,rueck
      complex(dop), dimension(expdim), intent(in)   :: exphin,exprueck

!-----------------------------------------------------------------------
! Go to selected operator
!-----------------------------------------------------------------------
      m=dofspf(f)

     select case (htmtype(h))

     case default
      routine='Hop'
      write(message,'(a,i3,a,3i5)') 'Unknown htmtype :',htmtype(h), &
                                    ',  h, k, f', h,k,f
      call errormsg

     case (1)!     unit matrix
      call cpvxz(psi,hpsi,subdim1)
      return

     case (2)!     real vector (diagonal matrix)
      call getvgn(h,vd,gd,nd)
      call dtxxdz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      return

     case (3)!     imaginary vector (imaginary part stored as real)
      call getvgn(h,vd,gd,nd)
      call dtxxdz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (4)!     complex vector  (diagonal matrix)
      call getvgn(h,vd,gd,nd)
      call dtxxyz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      return

     case (5)!     FFT, real vector
      call fftop(psi,hpsi,vgdim(f),gdim(f),ngdim(f),hin(zfft(f)), &
              rueck(zfft(f)),fftfak(zfft(f)),hops(zhop(h)))
      return

     case (6)!     FFT, imaginary vector
      call fftiop(psi,hpsi,vgdim(f),gdim(f),ngdim(f),hin(zfft(f)), &
              rueck(zfft(f)),fftfak(zfft(f)),hops(zhop(h)))
      return

     case (7)!     FFT, complex vector
      call fftzop(psi,hpsi,vgdim(f),gdim(f),ngdim(f),hin(zfft(f)), &
              rueck(zfft(f)),fftfak(zfft(f)),hops(zhop(h)))
      return

     case (8)!     real matrix 
      call getvgn(h,vd,gd,nd)

      call qtxxdz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      return

     case (9)!     imaginary matrix (imaginary part stored as real)
      call getvgn(h,vd,gd,nd)
      call qtxxdz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (10)!     complex matrix
      call getvgn(h,vd,gd,nd)
      call qtxxyz(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      return

     case (11)!     Fourier transformed natural potential -- OBSOLETE
      routine = 'hop (htmtype=11)'
      message = 'Support for FT-natpots has been removed.'
      call errormsg
      return

     case (12)!     real vector (side-diagonal matrix)
      dm=htmshift(h)
      call getvgn(h,vd,gd,nd)
      call otxxdz(hops(zhop(h)),dm,psi,hpsi,vd,gd,nd)
      return

     case (13)!     imaginary vector (side-diagonal matrix; stored as real)
      dm=htmshift(h)
      call getvgn(h,vd,gd,nd)
      call otxxdz(hops(zhop(h)),dm,psi,hpsi,vd,gd,nd)
      do g=1,subdim1
         hpsi(g)=(0.0_dop,1.0_dop)*hpsi(g)
      enddo
      return

     case (14)!     complex vector (side-diagonal matrix)
      dm=htmshift(h)
      call getvgn(h,vd,gd,nd)
      call otxxyz(hops(zhop(h)),dm,psi,hpsi,vd,gd,nd)
      return

     case (16)!!     real 3D tensor (for jpm)
      call apply_jfunk(jplusminuspsi,1)
      return

     case (15,17) ! real 3D tensor (for dth1, dth2) and real 3D tensor (for j^2)
      call apply_jfunk(j2psi,0)
      return

     case (18)!!     real 3D tensor (for j_p and sJm and sJmk)
      call apply_jfunk(jpluspsi,1)
      return

     case (19)!!     real 3D tensor (for j_m and sJp and sJpk)
      call apply_jfunk(jminuspsi,1)
      return

     case (20)!!     real 3D tensor (for jp^2)
      call apply_jfunk(jpluspsi,2)
      return

     case (21)!!     real 3D tensor (for jm^2)
      call apply_jfunk(jminuspsi,2)
      return

     case(22) !! real 3D tensor (for jpjm)
      call apply_jfunk(jpluspsi,0)
      return

     case(23) !! real 3D tensor (for jmjp)
      call apply_jfunk(jpluspsi,0)
      return

     case (24)
      return

     case (25) ! readmap
      call getvgn(h,vd,gd,nd)
      call applyreadmap(hops(zhop(h)),psi,hpsi,vd,gd,nd)
      return

     case (26:29)
      return

     case (30)
!     (for cjpm)
!     memory layout for the corresponding hops array:
!       hops = | cplus | j1plus | j2plus |
!       dim(cplus)  = kdim1+kdim2
!       dim(j1plus) = gdim1^2*kdim1
!       dim(j2plus) = gdim2^2*kdim2
!     where
!       gdim1 = gdim(f)   = size of theta1 grid
!       kdim1 = gdim(f+1) = size of phi1 grid
!       gdim2 = gdim(f+2) = size of theta2 grid
!       kdim2 = gdim(f+3) = size of phi2 grid
      if (basis(f).eq.7 .or. basis(f) .eq.20) then
         z1 = zhop(h)
         z2 = z1 + gdim(f+1)+gdim(f+3)
         z3 = z2 + gdim(f)*gdim(f)*gdim(f+1)
         call cjpmpsi(hops(z1),hops(z2),hops(z3),psi,hpsi,&
              vgdim(f),gdim(f),gdim(f+1),gdim(f+2),gdim(f+3),ngdim(f+3))
      else
         routine = 'hop (cjpm)'
         message = 'cjpm only implemented for KLeg'
         call errormsg
      endif
      return

      case (31) !     Coordinate reflection (for Rf)
      call rflx(psi,hpsi,vgdim(f),gdim(f),ngdim(f))
      return

      case (32) !     Coordinate reflection and sign change (for Rfm)
      call rflxm(psi,hpsi,vgdim(f),gdim(f),ngdim(f))
      return

     case (33:39)
      return

     case (40) !! real 4D tensor (for j^2wig)
      call apply_wigfunk(j2psiwig,0)
      return

     case (41,55,57) !! real 4D tensor (for j_pwig)
                     !! real 4D tensor (for jpjzwig, body-fixed)
                     !! real 4D tensor (for jzjpwig, body-fixed)
      call apply_wigfunk(jplpsiwig,1)
      return

     case (42,56,58) !! real 4D tensor (for j_mwig)
                     !! real 4D tensor (for jmjzwig, body-fixed)
                     !! real 4D tensor (for jzjmwig, body-fixed)
      call apply_wigfunk(jminpsiwig,1)
      return

     case (43) !! real 4D tensor (for jp2wig)
      call apply_wigfunk(jplpsiwig,2)
      return

     case (44) !! real 4D tensor (for jm2wig)
      call apply_wigfunk(jminpsiwig,2)
      return

     case (45) !! real 4D tensor (for jpmwig)
      call apply_wigfunk(jpmpsiwig,1)
      return

     case (46,59,61) !! real 4D tensor (for j_pswig)
                     !! real 4D tensor (for jpjzswig, space-fixed)
                     !! real 4D tensor (for jzjpswig, space-fixed)
      call apply_wigfunk(jsplpsiwig,1)
      return

     case (47,60,62) !! real 4D tensor (for j_mswig)
                     !! real 4D tensor (for jmjzswig, space-fixed)
                     !! real 4D tensor (for jzjmswig, space-fixed)
      call apply_wigfunk(jsminpsiwig,1)
      return

     case (48) !! real 4D tensor (for jp2swig)
      call apply_wigfunk(jsplpsiwig,2)
      return

     case (49) !! real 4D tensor (for jm2swig)
      call apply_wigfunk(jsminpsiwig,2)
      return

     case (50) !! real 4D tensor (for jpmswig)
      call apply_wigfunk(jspmpsiwig,1)
      return

     case (51:52) !! 51 real 4D tensor (for jpjmwig, body-fixed)
                  !! 52 real 4D tensor (for jmjpwig, body-fixed)
      call apply_wigfunk(jplpsiwig,0)
      return

     case (53:54) !! 53 real 4D tensor (for jpjmswig, space-fixed)
                  !! 54 real 4D tensor (for jmjpswig, space-fixed)
      call apply_wigfunk(jsplpsiwig,0)
      return

      case (63:69)
      return

      end select

    contains

!-----------------------------------------------------------------------
!
! Assumes Kleg or Pleg DVR
!
!-----------------------------------------------------------------------

      subroutine apply_jfunk(jfunk,m)

         external jfunk

         integer(long), intent(in) :: m
         integer(long)             :: kdim,zeigk
         complex(dop),allocatable  :: workc2(:,:)

         if (basis(f).eq.7) then
            call jfunk(hops(zhop(h)),psi,hpsi, &
                 vgdim(f),gdim(f),gdim(f+1),ngdim(f+1),m)

         elseif (basis(f).eq.9) then
            kdim=ipbaspar(3,f+1)
            zeigk=vgdim(f+1)*kdim*ngdim(f+1)
            allocate(workc2(zeigk,2))
            call mttxzz(exphin(zexp(f+1)),psi,workc2(1,1),vgdim(f+1), &
                        gdim(f+1),ngdim(f+1),kdim)
            call jfunk(hops(zhop(h)),workc2(1,1),workc2(1,2), &
                       vgdim(f),gdim(f),kdim,ngdim(f+1),m)
            call mttxzz(exprueck(zexp(f+1)),workc2(1,2),hpsi,vgdim(f+1), &
                        kdim,ngdim(f+1),gdim(f+1))
            deallocate(workc2)

         else
            write(message,'(a,i0,a,i0,a,i0,a,i0,a)') &
               'Illegal basis ',basis(f),' for htmtype ',htmtype(h), &
               ' (h=',h,', f=',f,')'
            call errormsg
         endif

      end subroutine apply_jfunk

!-----------------------------------------------------------------------

      subroutine apply_wigfunk(wigfunk,m)

         external wigfunk
         integer(long), intent(in) :: m
         integer(long)             :: kdim,mmdim,zeigk
         complex(dop),allocatable  :: workc2(:,:)

         if (basis(f).ne.33) then
            write(message,'(a,i0,a,i0,a,i0,a,i0,a)') &
               'Illegal basis ',basis(f),' for htmtype ',htmtype(h), &
               ' (h=',h,', f=',f,')'
            call errormsg
         endif

         if (basis(f+1).eq.5 .and. basis(f+2).eq.5) then
            kdim =ipbaspar(4,f)
            mmdim=ipbaspar(8,f)
            zeigk=vgdim(f+1)*kdim*ngdim(f+1)
            allocate(workc2(zeigk,2))
            call mttxzz(exphin(zexp(f+1)),psi,workc2(1,1), &
                 vgdim(f+1),gdim(f+1),ngdim(f+1),kdim)
            call mttxzz(exphin(zexp(f+2)),workc2(1,1),workc2(1,2), &
                 vgdim(f+1)*kdim,gdim(f+2),ngdim(f+2),mmdim)
            call wigfunk(hops(zhop(h)),workc2(1,2),workc2(1,1), &
                 vgdim(f),gdim(f),kdim,mmdim,ngdim(f+2),m)
            call mttxzz(exprueck(zexp(f+2)),workc2(1,1),workc2(1,2), &
                 vgdim(f+1)*kdim,mmdim,ngdim(f+2),gdim(f+2))
            call mttxzz(exprueck(zexp(f+1)),workc2(1,2),hpsi, &
                 vgdim(f+1),kdim,ngdim(f+1),gdim(f+1))
            deallocate(workc2)

         elseif (basis(f+1).eq.8 .and. basis(f+2).eq.8) then
            call wigfunk(hops(zhop(h)),psi,hpsi, &
                 vgdim(f),gdim(f),gdim(f+1),gdim(f+2),ngdim(f+2),m)

         elseif (basis(f+1).eq.8 .and. basis(f+2).eq.5) then
            mmdim=ipbaspar(8,f)
            zeigk=vgdim(f+2)*mmdim*ngdim(f+2)
            allocate(workc2(zeigk,2))
            call mttxzz(exphin(zexp(f+2)),psi,workc2(1,1), &
                 vgdim(f+2),gdim(f+2),ngdim(f+2),mmdim)
            call wigfunk(hops(zhop(h)),workc2(1,1),workc2(1,2), &
                 vgdim(f),gdim(f),gdim(f+1),mmdim,ngdim(f+2),m)
            call mttxzz(exprueck(zexp(f+2)),workc2(1,2),hpsi, &
                 vgdim(f+2),mmdim,ngdim(f+2),gdim(f+2))
            deallocate(workc2)

         elseif (basis(f+1).eq.5 .and. basis(f+2).eq.8) then
            kdim =ipbaspar(4,f)
            zeigk=vgdim(f+1)*kdim*ngdim(f+1)
            allocate(workc2(zeigk,2))
            call mttxzz(exphin(zexp(f+1)),psi,workc2(1,1), &
                 vgdim(f+1),gdim(f+1),ngdim(f+1),kdim)
            call wigfunk(hops(zhop(h)),workc2(1,1),workc2(1,2), &
                 vgdim(f),gdim(f),kdim,gdim(f+2),ngdim(f+2),m)
            call mttxzz(exprueck(zexp(f+1)),workc2(1,2),hpsi, &
                 vgdim(f+1),kdim,ngdim(f+1),gdim(f+1))
            deallocate(workc2)
         else
            write(message,'(a)') 'Only exp/K can complement Wigner'
            call errormsg
         endif

      end subroutine apply_wigfunk

      end subroutine hop

!************************************************************************

      subroutine fftop(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak,fftk)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim), intent(in)                 :: fftk
      complex(dop), dimension(gdim)                          :: vec
      complex(dop), dimension(gdim), intent(in)              :: rueck,hin
      complex(dop), dimension(gdim)                          :: workc
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi


      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=fftk(g)*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftop

!#######################################################################

      subroutine fftiop(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak,fftk)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim), intent(in)                 :: fftk
      complex(dop), dimension(gdim)                          :: vec
      complex(dop), dimension(gdim), intent(in)              :: rueck,hin
      complex(dop), dimension(gdim)                          :: workc
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi

      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=dcmplx(0.0_dop,fftk(g))*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftiop

!#######################################################################

      subroutine fftzop(psi,hpsi,vgdim,gdim,ngdim,hin,rueck,fftfak,fftk)

      implicit none

      integer(long)                                          :: g,v,n
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), dimension(gdim), intent(in)             :: fftfak
      real(dop), dimension(gdim,2), intent(in)               :: fftk
      complex(dop), dimension(gdim)                          :: vec
      complex(dop), dimension(gdim), intent(in)              :: rueck,hin
      complex(dop), dimension(gdim)                          :: workc
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi

      do n=1,ngdim
         do v=1,vgdim
            do g=1,gdim
               vec(g)=psi(v,g,n)
            enddo
            call fft(vec,workc,rueck,fftfak,gdim,1,1,.true.)
            do g=1,gdim
               vec(g)=dcmplx(fftk(g,1),fftk(g,2))*vec(g)
            enddo
            call fft(vec,workc,hin,fftfak,gdim,1,1,.false.)
            do g=1,gdim
               hpsi(v,g,n)=vec(g)
            enddo
         enddo
      enddo

      return
      end subroutine fftzop

!-----------------------------------------------------------------------
!  The following subroutines: j2psi,jpluspsi,jminuspsi, jplusminuspsi
!  calculate the action of (KLeg/Pleg) j^2,j_p,j_m,jpm operators on
!  wave function psi.
!
! m is ignored
!-----------------------------------------------------------------------

      subroutine j2psi (j,psi,jpsi,vgdim,gdim,kdim,ngdim,m)

      implicit none

      integer(long)                               :: g,g2,v,n,k,m1
      integer(long), intent(in)                   :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(gdim,gdim,kdim), intent(in)            :: j
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jpsi

! m1=m to prevent unused variable
      m1=m
      do n=1,ngdim
         do k=1,kdim
            do g=1,gdim
               do v=1,vgdim
                  jpsi(v,g,k,n)=j(g,1,k)*psi(v,1,k,n)
               enddo
            enddo
         enddo
      enddo
      do n=1,ngdim
         do k=1,kdim
            do g=1,gdim
               do g2=2,gdim
                  do v=1,vgdim
                     jpsi(v,g,k,n)=jpsi(v,g,k,n)+j(g,g2,k)*psi(v,g2,k,n)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine j2psi

!-----------------------------------------------------------------------

      subroutine jpluspsi (jpl,psi,jplpsi,vgdim,gdim,kdim,ngdim,m)

      implicit none

      integer(long)                                               :: g,g2,v,n,k
      integer(long), intent(in)                                   :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(gdim,gdim,kdim), intent(in)            :: jpl
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jplpsi


      do n=1,ngdim
         do k=1,m
            do g=1,gdim
               do v=1,vgdim
                  jplpsi(v,g,k,n)=(0.0_dop,0.0_dop)
               enddo
            enddo
         enddo
      enddo

      do n=1,ngdim
         do k=m+1,kdim
            do g=1,gdim
               do v=1,vgdim
                  jplpsi(v,g,k,n)=jpl(g,1,k-m)*psi(v,1,k-m,n)
               enddo
            enddo
         enddo
      enddo
      do n=1,ngdim
         do k=m+1,kdim
            do g=1,gdim
               do g2=2,gdim
                  do v=1,vgdim
                     jplpsi(v,g,k,n)=jplpsi(v,g,k,n)+ &
                                     jpl(g,g2,k-m)*psi(v,g2,k-m,n)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine jpluspsi

!-----------------------------------------------------------------------

      subroutine jminuspsi (jmin,psi,jminpsi,vgdim,gdim,kdim,ngdim,m)

      implicit none

      integer(long)                                               :: g,g2,v,n,k
      integer(long), intent(in)                                   :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(gdim,gdim,kdim), intent(in)            :: jmin
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jminpsi


      do n=1,ngdim
         do k=1,kdim-m
            do g=1,gdim
               do v=1,vgdim
                  jminpsi(v,g,k,n)=jmin(g,1,k+m)*psi(v,1,k+m,n)
               enddo
            enddo
         enddo
      enddo
      do n=1,ngdim
         do k=1,kdim-m
            do g=1,gdim
               do g2=2,gdim
                  do v=1,vgdim
                     jminpsi(v,g,k,n)=jminpsi(v,g,k,n)+&
                                      jmin(g,g2,k+m)*psi(v,g2,k+m,n)
                  enddo
               enddo
            enddo
         enddo
      enddo

      do n=1,ngdim
         do k=kdim-m+1,kdim
            do g=1,gdim
               do v=1,vgdim
                  jminpsi(v,g,k,n)=(0.0_dop,0.0_dop)
               enddo
            enddo
         enddo
      enddo


      return
      end subroutine jminuspsi

!-----------------------------------------------------------------------

      subroutine jplusminuspsi (jplmin,psi,jplminpsi,vgdim,gdim,kdim, &
                                ngdim,m)

      implicit none

      integer(long)                                               :: g,g2,v,n,k
      integer(long), intent(in)                                   :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(gdim,gdim,kdim), intent(in)            :: jplmin
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jplminpsi


!---- for k=1,...,m  and  k=kdim,...,kdim-m+1
      do n=1,ngdim
         do k=1,m
            do g=1,gdim
               do v=1,vgdim
                  jplminpsi(v,g,k,n)=jplmin(1,g,k)*psi(v,1,k+1,n)
                  jplminpsi(v,g,kdim-k+1,n)=jplmin(g,1,kdim-k)*&
                                            psi(v,1,kdim-k,n)
               enddo
            enddo
         enddo
      enddo
      do n=1,ngdim
         do k=1,m
            do g=1,gdim
               do g2=2,gdim
                  do v=1,vgdim
                     jplminpsi(v,g,k,n)=jplminpsi(v,g,k,n) + &
                                        jplmin(g2,g,k)*psi(v,g2,k+1,n)
                     jplminpsi(v,g,kdim-k+1,n)=jplminpsi(v,g,kdim-k+1,n)&
                                + jplmin(g,g2,kdim-k)*psi(v,g2,kdim-k,n)
                  enddo
               enddo
            enddo
         enddo
      enddo

!---- for k=m+1,...,kdim-m
      do n=1,ngdim
         do k=m+1,kdim-m
            do g=1,gdim
               do v=1,vgdim
                  jplminpsi(v,g,k,n)=jplmin(g,1,k-1)*psi(v,1,k-1,n) + &
                                     jplmin(1,g,k)  *psi(v,1,k+1,n)
               enddo
            enddo
         enddo
      enddo
      do n=1,ngdim
         do k=m+1,kdim-m
            do g=1,gdim
               do g2=2,gdim
                  do v=1,vgdim
                     jplminpsi(v,g,k,n)=jplminpsi(v,g,k,n) + &
                                      jplmin(g,g2,k-1)*psi(v,g2,k-1,n) +&
                                      jplmin(g2,g,k)  *psi(v,g2,k+1,n)
                  enddo
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine jplusminuspsi 

!-----------------------------------------------------------------------
!  The next subroutines: j2psiwig,jplpsiwig,jminpsiwig,jpmpsiwig,
!  jsplpsiwig,jsminpsiwig,jspmpsiwig, calculate the action of (Wigner)
!  j^2,j_p,j_m,jpm,js_p,js_m,jspm operators on wave function psi.
!
! m is ignored
!-----------------------------------------------------------------------

      subroutine j2psiwig (j,psi,jpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y,m1
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: j
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jpsi

! m1=m to prevent unused variable
      m1=m
      do n=1,ngdim
        do y=1,mmdim
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jpsi(v,g,k,y,n)=j(g,1,k,y)*psi(v,1,k,y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim
          do k=1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jpsi(v,g,k,y,n)=jpsi(v,g,k,y,n) + j(g,g2,k,y)*psi(v,g2,k,y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine j2psiwig 

!-----------------------------------------------------------------------
!     Molecule body-fixed raising and lowering operators:
!     j_p,j_m,jpm (j_m raises, j_p lowers convention, to ensure
!     j+ = jx + i*jy and j- = jx - i*jy)

      subroutine jminpsiwig (jmin,psi,jminpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jmin
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jminpsi

      do n=1,ngdim
        do y=1,mmdim
          do k=1,m
            do g=1,gdim
              do v=1,vgdim
                jminpsi(v,g,k,y,n)=(0.d0,0.d0)
              enddo
            enddo
          enddo
        enddo
      enddo

      do n=1,ngdim
        do y=1,mmdim
          do k=m+1,kdim
            do g=1,gdim
              do v=1,vgdim
                jminpsi(v,g,k,y,n)=jmin(g,1,k-m,y)*psi(v,1,k-m,y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim
          do k=m+1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jminpsi(v,g,k,y,n) = jminpsi(v,g,k,y,n)+ &
                                      jmin(g,g2,k-m,y)*psi(v,g2,k-m,y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jminpsiwig 

!-----------------------------------------------------------------------

      subroutine jplpsiwig (jpl,psi,jplpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jpl
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jplpsi

      do n=1,ngdim
        do y=1,mmdim
          do k=1,kdim-m
            do g=1,gdim
              do v=1,vgdim
                jplpsi(v,g,k,y,n)=jpl(g,1,k+m,y)*psi(v,1,k+m,y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim
          do k=1,kdim-m
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jplpsi(v,g,k,y,n)=jplpsi(v,g,k,y,n) + &
                                   jpl(g,g2,k+m,y)*psi(v,g2,k+m,y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      do n=1,ngdim
        do y=1,mmdim
          do k=kdim-m+1,kdim
            do g=1,gdim
              do v=1,vgdim
                jplpsi(v,g,k,y,n)=(0.d0,0.d0)
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jplpsiwig 

!-----------------------------------------------------------------------

      subroutine jpmpsiwig (jplmin,psi,jplminpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jplmin
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jplminpsi


!---- for k=1,...,m  and  k=kdim,...,kdim-m+1
      do n=1,ngdim
        do y=1,mmdim
          do k=1,m
            do g=1,gdim
              do v=1,vgdim
                jplminpsi(v,g,k,y,n)=jplmin(1,g,k,y)*psi(v,1,k+1,y,n)
                jplminpsi(v,g,kdim-k+1,y,n)=jplmin(g,1,kdim-k,y)* &
                                          psi(v,1,kdim-k,y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim
          do k=1,m
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                 jplminpsi(v,g,k,y,n)=jplminpsi(v,g,k,y,n) + &
                                     jplmin(g2,g,k,y)*psi(v,g2,k+1,y,n)
                 jplminpsi(v,g,kdim-k+1,y,n)=jplminpsi(v,g,kdim-k+1,y,n) &
                            + jplmin(g,g2,kdim-k,y)*psi(v,g2,kdim-k,y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

!---- for k=m+1,...,kdim-m
      do n=1,ngdim
        do y=1,mmdim
          do k=m+1,kdim-m
            do g=1,gdim
              do v=1,vgdim
               jplminpsi(v,g,k,y,n)=jplmin(g,1,k-1,y)*psi(v,1,k-1,y,n) + &
                                   jplmin(1,g,k,y)*psi(v,1,k+1,y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim
          do k=m+1,kdim-m
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jplminpsi(v,g,k,y,n)=jplminpsi(v,g,k,y,n) + &
                                  jplmin(g,g2,k-1,y)*psi(v,g2,k-1,y,n) + &
                                  jplmin(g2,g,k,y)*psi(v,g2,k+1,y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jpmpsiwig 

!-----------------------------------------------------------------------
!     Space-fixed raising and lowering operators:
!     j_ps,j_ms,jpms

      subroutine jsplpsiwig (jpl,psi,jplpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jpl
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jplpsi

      do n=1,ngdim
        do y=1,m
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jplpsi(v,g,k,y,n)=(0.d0,0.d0)
              enddo
            enddo
          enddo
        enddo
      enddo

      do n=1,ngdim
        do y=m+1,mmdim
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jplpsi(v,g,k,y,n)=jpl(g,1,k,y-m)*psi(v,1,k,y-m,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=m+1,mmdim
          do k=1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jplpsi(v,g,k,y,n)=jplpsi(v,g,k,y,n)+ &
                                  jpl(g,g2,k,y-m)*psi(v,g2,k,y-m,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jsplpsiwig 

!-----------------------------------------------------------------------

      subroutine jsminpsiwig (jmin,psi,jminpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jmin
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jminpsi

      do n=1,ngdim
        do y=1,mmdim-m
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jminpsi(v,g,k,y,n)=jmin(g,1,k,y+m)*psi(v,1,k,y+m,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,mmdim-m
          do k=1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jminpsi(v,g,k,y,n)=jminpsi(v,g,k,y,n)+ &
                                   jmin(g,g2,k,y+m)*psi(v,g2,k,y+m,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      do n=1,ngdim
        do y=mmdim-m+1,mmdim
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jminpsi(v,g,k,y,n)=(0.d0,0.d0)
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jsminpsiwig 

!-----------------------------------------------------------------------

      subroutine jspmpsiwig (jplmin,psi,jplminpsi,vgdim,gdim,kdim,mmdim,ngdim,m)

      implicit none

      integer(long), intent(in) :: m,gdim,kdim,mmdim,vgdim,ngdim
      integer(long)             :: g,g2,v,n,k,y
      real(dop), dimension(gdim,gdim,kdim,mmdim), intent(in)            :: jplmin
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,kdim,mmdim,ngdim), intent(out) :: jplminpsi


!---- for mm=1,...,m  and  m=mmdim,...,mmdim-m+1
      do n=1,ngdim
        do y=1,m
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
                jplminpsi(v,g,k,y,n)=jplmin(1,g,k,y)*psi(v,1,k,y+1,n)
                jplminpsi(v,g,k,mmdim-y+1,n)=jplmin(g,1,k,mmdim-y)* &
                                                psi(v,1,k,mmdim-y,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=1,m
          do k=1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                 jplminpsi(v,g,k,y,n)=jplminpsi(v,g,k,y,n) + &
                                     jplmin(g2,g,k,y)*psi(v,g2,k,y+1,n)
               jplminpsi(v,g,k,mmdim-y+1,n)=jplminpsi(v,g,k,mmdim-y+1,n) &
                          + jplmin(g,g2,k,mmdim-y)*psi(v,g2,k,mmdim-y,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

!---- for mm=m+1,...,mmdim-m
      do n=1,ngdim
        do y=m+1,mmdim-m
          do k=1,kdim
            do g=1,gdim
              do v=1,vgdim
               jplminpsi(v,g,k,y,n)=jplmin(g,1,k,y-1)*psi(v,1,k,y-1,n) + &
                                   jplmin(1,g,k,y)*psi(v,1,k,y+1,n)
              enddo
            enddo
          enddo
        enddo
      enddo
      do n=1,ngdim
        do y=m+1,mmdim-m
          do k=1,kdim
            do g=1,gdim
              do g2=2,gdim
                do v=1,vgdim
                  jplminpsi(v,g,k,y,n)=jplminpsi(v,g,k,y,n) + &
                                  jplmin(g,g2,k,y-1)*psi(v,g2,k,y-1,n) + &
                                  jplmin(g2,g,k,y)*psi(v,g2,k,y+1,n)
                enddo
              enddo
            enddo
          enddo
        enddo
      enddo

      return
      end subroutine jspmpsiwig 

!-----------------------------------------------------------------------
!
! CJPMPSI: calculates action of the operator
!          C+(J,K)[j1+ + j2+] + C-(J,K)[j1- + j2-]
!          on psi (where K=k1+k2)
!
! Note: C-(J,K)   = C+(J,K-1)
!       j-(g,j,k) = j+(j,g,k-1)
!
! cpl : contains C+
! jpl1: contains j1+
! jpl2: contains j2+
! psi : particle wavefunction
! hpsi: result
!
! 06/2006 FO
!
!-----------------------------------------------------------------------

      subroutine cjpmpsi(cpl,jpl1,jpl2,psi,hpsi, &
                         vgdim,gdim1,kdim1,gdim2,kdim2,ngdim)

      implicit none

      integer(long)                                                             :: v,g1,j1,k1,g2,j2,k2,k,n
      integer(long), intent(in)                                                 :: vgdim,gdim1,kdim1,gdim2,kdim2,ngdim
      real(dop), dimension(kdim1+kdim2), intent(in)                             :: cpl
      real(dop), dimension(gdim1,gdim1,kdim1), intent(in)                       :: jpl1
      real(dop), dimension(gdim2,gdim2,kdim2), intent(in)                       :: jpl2
      complex(dop), dimension(vgdim,gdim1,kdim1,gdim2,kdim2,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim1,kdim1,gdim2,kdim2,ngdim), intent(out) :: hpsi
      complex(dop)                                                              :: tmp

      do n=1,ngdim
       do k2=1,kdim2
        do g2=1,gdim2
         do k1=1,kdim1
          k=k1+k2
          do g1=1,gdim1
           do v=1,vgdim
              tmp = (0.0_dop, 0.0_dop)
!             C+ j1+
              if (k1.gt.1) then
                 do j1=1,gdim1
                    tmp = tmp +&
                        cpl(k-1)*jpl1(g1,j1,k1-1)*psi(v,j1,k1-1,g2,k2,n)
                 enddo
              endif
!             C- j1-
              if (k1.lt.kdim1) then
                 do j1=1,gdim1
                    tmp = tmp +&
                        cpl(k)*jpl1(j1,g1,k1)*psi(v,j1,k1+1,g2,k2,n)
                 enddo
              endif
!             C+ j2+
              if (k2.gt.1) then
                 do j2=1,gdim2
                    tmp = tmp +&
                        cpl(k-1)*jpl2(g2,j2,k2-1)*psi(v,g1,k1,j2,k2-1,n)
                 enddo
              endif
!             C- j2-
              if (k2.lt.kdim2) then
                 do j2=1,gdim2
                    tmp = tmp +&
                        cpl(k)*jpl2(j2,g2,k2)*psi(v,g1,k1,j2,k2+1,n)
                 enddo
              endif
              hpsi(v,g1,k1,g2,k2,n) = tmp
           enddo
          enddo
         enddo
        enddo
       enddo
      enddo
                     
      return
      end subroutine cjpmpsi! of cjpmpsi


!#######################################################################
!   SUBROUTINE SHIFTK
!
!   Shifts the multi-dimensional wavefunction:
!           K(m)*psi(x1,...,k,...,xn)=psi(x1,...,k+m,...,xn)
! not in use
!#######################################################################
 
      subroutine shiftk(m,psi,hpsi,vgdim,gdim,ngdim)
 
      implicit none
 
      integer(long)                                          :: i,k,j
      integer(long), intent(in)                              :: vgdim,gdim,ngdim
      integer(long), intent(inout)                           :: m
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi
 
 
! --- for m=0
      if (m .eq. 0) then
         call cpvxz(psi,hpsi,vgdim*gdim*ngdim)
! --- for m>0
      elseif (m .gt. 0) then
!         --- for psi(x1,...,k)
         if (vgdim.ne.1 .and. ngdim.eq.1) then
            do k=1,gdim-m
               do i=1,vgdim
                  hpsi(i,k,1) = psi(i,k+m,1)
               enddo
            enddo
            do k=gdim-m+1,gdim
               do i=1,vgdim
                  hpsi(i,k,1) = (0.0_dop,0.0_dop)
               enddo
            enddo
!         --- for psi(k,x1,...)
         elseif (vgdim.eq.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,gdim-m
                  hpsi(1,k,j) = psi(1,k+m,j)
               enddo
            enddo
            do j=1,ngdim
               do k=gdim-m+1,gdim
                  hpsi(1,k,j) = (0.0_dop,0.0_dop)
               enddo
            enddo
!         --- for psi(x1,...,k,...,xn)
         elseif (vgdim.ne.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,gdim-m
                  do i=1,vgdim
                     hpsi(i,k,j) = psi(i,k+m,j)
                  enddo
               enddo
            enddo
            do j=1,ngdim
               do k=gdim-m+1,gdim
                  do i=1,vgdim
                     hpsi(i,k,j) = (0.0_dop,0.0_dop)
                  enddo
               enddo
            enddo
         endif 
! --- for m<0
      else
         m=abs(m)
!         --- for psi(x1,...,k)
         if (vgdim.ne.1 .and. ngdim.eq.1) then
            do k=1,m
               do i=1,vgdim
                  hpsi(i,k,1) = (0.0_dop,0.0_dop)
               enddo
            enddo
            do k=m+1,gdim
               do i=1,vgdim
                  hpsi(i,k,1) = psi(i,k-m,1)
               enddo
            enddo
!         --- for psi(k,x1,...)
         elseif (vgdim.eq.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,m
                  hpsi(1,k,j) = (0.0_dop,0.0_dop)
               enddo
            enddo
            do j=1,ngdim
               do k=m+1,gdim
                  hpsi(1,k,j) = psi(1,k-m,j)
               enddo
            enddo
!         --- for psi(x1,...,k,...,xn)
         elseif (vgdim.ne.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,m
                  do i=1,vgdim
                     hpsi(i,k,j) = (0.0_dop,0.0_dop)
                  enddo
               enddo
            enddo
            do j=1,ngdim
               do k=m+1,gdim
                  do i=1,vgdim
                     hpsi(i,k,j) = psi(i,k-m,j)
                  enddo
               enddo
            enddo
         endif
      endif
 
      return
      end subroutine shiftk

!#######################################################################
!   SUBROUTINE RFLX
!
!   Coordinate reflection: psi(x1,..,xk,..,xn) => psi(x1,..,-xk,..,xn)
!   Actually the grid point is mapped: i => N-i+1
!#######################################################################

      subroutine rflx(psi,hpsi,vgdim,gdim,ngdim)

      implicit none

      integer(long), intent(in) :: vgdim,gdim,ngdim
      integer(long)             :: k,j,i
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi

!         --- for psi(x)
         if (vgdim.eq.1 .and. ngdim.eq.1) then
            do k=1,gdim
               hpsi(1,k,1) = psi(1,gdim+1-k,1)
            enddo

!         --- for psi(x,...)
         elseif (vgdim.eq.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,gdim
                  hpsi(1,k,j) = psi(1,gdim+1-k,j)
               enddo
            enddo

!         --- for psi(...,x)
         elseif (vgdim.ne.1 .and. ngdim.eq.1) then
            do k=1,gdim
               do i=1,vgdim
                  hpsi(i,k,1) = psi(i,gdim+1-k,1)
               enddo
            enddo

!         --- for psi(...,x,...)
         else
            do j=1,ngdim
               do k=1,gdim
                  do i=1,vgdim
                     hpsi(i,k,j) = psi(i,gdim+1-k,j)
                  enddo
               enddo
            enddo
         endif

      return
      end subroutine rflx

!#######################################################################
!   SUBROUTINE RFLXM
!
!   Coordinate reflection: psi(x1,..,xk,..,xn) => -psi(x1,..,-xk,..,xn)
!   Actually the grid point is mapped: i => N-i+1 and psi => -psi
!#######################################################################

      subroutine rflxm(psi,hpsi,vgdim,gdim,ngdim)

      implicit none

      integer(long), intent(in) :: vgdim,gdim,ngdim
      integer(long)             :: k,j,i
      complex(dop), dimension(vgdim,gdim,ngdim), intent(in)  :: psi
      complex(dop), dimension(vgdim,gdim,ngdim), intent(out) :: hpsi

!         --- for psi(x)
         if (vgdim.eq.1 .and. ngdim.eq.1) then
            do k=1,gdim
               hpsi(1,k,1) = -psi(1,gdim+1-k,1)
            enddo

!         --- for psi(x,...)
         elseif (vgdim.eq.1 .and. ngdim.ne.1) then
            do j=1,ngdim
               do k=1,gdim
                  hpsi(1,k,j) = -psi(1,gdim+1-k,j)
               enddo
            enddo

!         --- for psi(...,x)
         elseif (vgdim.ne.1 .and. ngdim.eq.1) then
            do k=1,gdim
               do i=1,vgdim
                  hpsi(i,k,1) = -psi(i,gdim+1-k,1)
               enddo
            enddo

!         --- for psi(...,x,...)
         else
            do j=1,ngdim
               do k=1,gdim
                  do i=1,vgdim
                     hpsi(i,k,j) = -psi(i,gdim+1-k,j)
                  enddo
               enddo
            enddo
         endif

      return
      end subroutine rflxm

!#######################################################################
!   SUBROUTINE APPLYREADMAP
!
!   Applies mapping operator from readmap
!#######################################################################

      subroutine applyreadmap(hops,psi,hpsi,vd,gd,nd)

      implicit none

      integer(long) :: i,j,nel,irow,icol,iv,in
      integer(long), intent(in) :: vd,gd,nd
      real(dop) :: val
      real(dop), dimension(*), intent(in) :: hops
      complex(dop), dimension(vd,gd,nd), intent(in)  :: psi
      complex(dop), dimension(vd,gd,nd), intent(out) :: hpsi

      nel=nint(hops(1))

      if(vd.eq.1 .and. nd.eq.1) then
         do j=1,gd
            hpsi(1,j,1)=(0.d0,0.d0)
         enddo
         do i=1,nel
            irow=nint(hops(i+1))
            icol=nint(hops(nel+i+1))
            val=hops(2*nel+i+1)
!           hpsi(1,irow,1)=val*psi(1,icol,1)
            hpsi(1,irow,1)=hpsi(1,irow,1)+val*psi(1,icol,1)
         enddo

      else
         do in=1,nd
            do j=1,gd
               do iv=1,vd
                  hpsi(iv,j,in)=(0.d0,0.d0)
               enddo
             enddo
          enddo

         do in=1,nd
            do i=1,nel
               irow=nint(hops(i+1))
               icol=nint(hops(nel+i+1))
               val=hops(2*nel+i+1)
               do iv=1,vd
!                 hpsi(iv,irow,in)=val*psi(iv,icol,in)
                  hpsi(iv,irow,in)=hpsi(iv,irow,in)+val*psi(iv,icol,in)
               enddo
            enddo
         enddo
      endif

      return
      end subroutine applyreadmap

!#######################################################################
      end module hphimod

