!######################################################################
!
!                      MCTDH module IORST
!
! Controls the input and output of the wavefunction to the restart 
! file.
!
! subroutines :
!    wrrst    : writes restart file
!    rdrst    : reads restart file
!    rstinfo  : reads information part of restart file
!    splitrst : splits the restart file of a block-Davidson run
!               into individual restart files 
!    wrrstrw    : writes restart file for real WP
!    rdrstrw    : reads restart file for real WP
!    rstinforw  : reads information part of restart file for real WP
!
!  V6.4 GW
!
!######################################################################
      module iorst

use decimal, only: long, dop
use logdat
use versions, only: progver
use op1lib,  only: zerovxi, cpvxz

use global
use dglobal
use dvrdatmod
use griddatmod
use datenmod
use operdef
use dirdyn
use hpsimod, only: gwpcxm,gwpintordm,gwpintm
use runpropmod
use channels
use psidef
use oct
use gapmod
use realwpmod

use iogrddef, only:wrgrddef,rdgrddef,griddat
use iopsidef, only:wrpsidef,rdpsidef,psidat,tapelength
use iodvrdef, only:wrdvrdef,rddvrdef
use wftree, only: new_wftree

use module_interface

      implicit none
      private
      public :: wrrst,rdrst,rstinfo,splitrst,wrrstrw,rdrstrw,rstinforw,&
                rdrstpsi,chkrsttyp
      contains

!#######################################################################
! WRRST
!
! writes information to restart file needed for continuation runs etc.
!
! if linwf is .true., data is for an initial wavefunction, and
!    no information about integrator is written
!
! V8.0 GW
!#######################################################################

      subroutine wrrst(time,psi,linwf,lanczvec,lanczord, &
                       alpha,beta,jindx,agmat,trajst)

      implicit none

      integer(long)                                    :: d,b,i,m,s
      integer(long), external                          :: log2int
      integer(long), intent(in)                        :: lanczord
      integer(long), dimension(jindxdim), intent(in)   :: jindx
      integer(long), dimension(gwpinfodim), intent(in) :: trajst

      real(dop), intent(in)                            :: time
      real(dop), dimension(1:lanczord), intent(in)     :: alpha,beta
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(1:totphidim), intent(in) :: lanczvec
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      character(len=c5)                                :: filename
      character(len=100)                               :: buffer
      logical(kind=4), intent(in)                      :: linwf
      character(len=c2)                                :: wrtag
      logical(kind=4)                                  :: lcheck

      routine='Wrrst'

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Wrrst (iorst.f90)'
         message = 'Psi has become corrupted. Restart file not written. '
         call errormsg
      endif

      if(irst.ge.50) then
         irst=irst-50   ! use filenames "rst000, rst001" etc
         write(filename,'(2a,i3.3)') name(1:laenge),'/rst',irst
         irst=50
      else
         filename=name(1:laenge)//'/restart'
      endif
      open(irst,file=filename,form='unformatted',status='unknown',err=999)

      write(irst) progver

!-----------------------------------------------------------------------
! write title
!-----------------------------------------------------------------------
      wrtag= '<runtitle>'
      write(irst) wrtag
      write(irst) runtitle

!-----------------------------------------------------------------------
! write date and time of creation
!-----------------------------------------------------------------------
      call fpstartup(buffer)
      wrtag= '<creation date>'
      write(irst) wrtag
      write(irst) buffer

      wrtag= '<end rsttitle>'
      write(irst) wrtag

!-----------------------------------------------------------------------
! array sizes
!-----------------------------------------------------------------------
      call wrmemdim(irst)

!-----------------------------------------------------------------------
! write system definition
!-----------------------------------------------------------------------
      call wrdvrdef(irst)
      call wrgrddef(irst)
      call wrpsidef(irst)

      wrtag= '<timeunit>'
      write(irst) wrtag
      write(irst) fs
      wrtag= '<energyunit>'
      write(irst) wrtag
      write(irst) ev
      wrtag= '<rlxunit>'
      write(irst) wrtag
      write(irst) rlxunit
      wrtag= '<rlxtime>'
      write(irst) wrtag
      write(irst) rlxtime
      wrtag= '<temperature0>'
      write(irst) wrtag
      write(irst) temperature0
      wrtag= '<normthermal>'
      write(irst) wrtag
      write(irst) normthermal
      wrtag= '<timecap>'
      write(irst) wrtag
      write(irst) (timecap(d),d=1,maxdim)

!-----------------------------------------------------------------------
! write out integrator information if in propagation stage
!-----------------------------------------------------------------------
      if (.not. linwf) then
         wrtag= '<start integrator info>'
         write(irst) wrtag

         wrtag= '<proptask>'
         write(irst) wrtag
         write(irst) proptask
         wrtag= '<lrlx>'
         write(irst) wrtag
         write(irst) log2int(lrlx)
         wrtag= '<irunprop>'
         write(irst) wrtag
         write(irst) irunprop
         wrtag= '<lrundiag>'
         write(irst) wrtag
         write(irst) log2int(lrundiag)
         wrtag= '<out1>'
         write(irst) wrtag
         write(irst) out1
         wrtag= '<out2>'
         write(irst) wrtag
         write(irst) out2

         wrtag= '<lconst>'
         write(irst) wrtag
         write(irst) log2int(lconst)
         wrtag= '<lfunkl>'
         write(irst) wrtag
         write(irst) log2int(lfunkl)
         wrtag= '<lfunkr>'
         write(irst) wrtag
         write(irst) log2int(lfunkr)
         wrtag= '<lexact>'
         write(irst) wrtag
         write(irst) log2int(lexact)
         wrtag= '<lrealwp>'
         write(irst) wrtag
         write(irst) log2int(lrealwp)
         wrtag= '<lrunml>'
         write(irst) wrtag
         write(irst) log2int(lrunml)
         wrtag= '<lconstfix>'
         write(irst) wrtag
         write(irst) log2int(lconstfix)
         wrtag= '<lconstvar>'
         write(irst) wrtag
         write(irst) log2int(lconstvar)
         wrtag= '<lcmfadaphi>'
         write(irst) wrtag
         write(irst) log2int(lcmfadaphi)
         wrtag= '<lcmfadaa>'
         write(irst) wrtag
         write(irst) log2int(lcmfadaa)
         wrtag= '<inttyp>'
         write(irst) wrtag
         write(irst) inttyp
         wrtag= '<inttypea>'
         write(irst) wrtag
         write(irst) inttypea
         wrtag= '<inttypephi>'
         write(irst) wrtag
         write(irst) inttypephi
         wrtag= '<intord>'
         write(irst) wrtag
         write(irst) intord
         wrtag= '<intordera>'
         write(irst) wrtag
         write(irst) intordera
         wrtag= '<intorderphi>'
         write(irst) wrtag
         write(irst) intorderphi
         wrtag= '<accuracy>'
         write(irst) wrtag
         write(irst) accuracy
         wrtag= '<accavec>'
         write(irst) wrtag
         write(irst) accavec
         wrtag= '<accphi>'
         write(irst) wrtag
         write(irst) accphi
         wrtag= '<dt>'
         write(irst) wrtag
         write(irst) dt
         wrtag= '<dta>'
         write(irst) wrtag
         write(irst) dta
         wrtag= '<dtphi>'
         write(irst) wrtag
         write(irst) dtphi
         wrtag= '<minrk5>'
         write(irst) wrtag
         write(irst) minrk5
         wrtag= '<updatetime>'
         write(irst) wrtag
         write(irst) updatetime
         wrtag= '<epsupdate>'
         write(irst) wrtag
         write(irst) epsupdate
         wrtag= '<macheps>'
         write(irst) wrtag
         write(irst) macheps
         wrtag= '<epsmat>'
         write(irst) wrtag
         write(irst) epsmat
         wrtag= '<epssvd>'
         write(irst) wrtag
         write(irst) epssvd
         wrtag= '<lopro>'
         write(irst) wrtag
         write(irst) log2int(lopro)
         wrtag= '<liork>'
         write(irst) wrtag
         write(irst) log2int(liork)
         wrtag= '<limportho>'
         write(irst) wrtag
         write(irst) log2int(limportho)
         wrtag= '<lrktest>'
         write(irst) wrtag
         write(irst) log2int(lrktest)
         wrtag= '<lnaturcmf>'
         write(irst) wrtag
         write(irst) log2int(lnaturcmf)
         wrtag= '<lcap>'
         write(irst) wrtag
         write(irst) log2int(lcap)
!         wrtag= '<autocap>'
!         write(irst) wrtag
!         write(irst) (autocap(k),k=1,ktot)
         wrtag= '<lreal>'
         write(irst) wrtag
         write(irst) log2int(lreal)
         wrtag= '<lrreal>'
         write(irst) wrtag
         write(irst) log2int(lrreal)
         wrtag= '<estart>'
         write(irst) wrtag
         write(irst) estart
         wrtag= '<ktot>'
         write(irst) wrtag
         write(irst) ktot
         wrtag= '<lmlcmfsplit>'
         write(irst) wrtag
         write(irst) log2int(lmlcmfsplit)
         wrtag= '<mnspec>'
         write(irst) wrtag
         write(irst) mnspec
         if (lrunml) then
            wrtag= '<mspec>'
            write(irst) wrtag
            write(irst) (mspec(d),d=1,2*maxdim)
            wrtag= '<minttyp>'
            write(irst) wrtag
            write(irst) (minttyp(d),d=1,2*maxdim)
            wrtag= '<mintord>'
            write(irst) wrtag
            write(irst) (mintord(d),d=1,2*maxdim)
            wrtag= '<mintacc>'
            write(irst) wrtag
            write(irst) (mintacc(d),d=1,2*maxdim)
            wrtag= '<mintdt>'
            write(irst) wrtag
            write(irst) (mintdt(d), d=1,2*maxdim)
         endif
         wrtag= '<lrealphi>'
         write(irst) wrtag
         write(irst) log2int(lrealphi)
         wrtag= '<lrhonrm>'
         write(irst) wrtag
         write(irst) log2int(lrhonrm)
         wrtag= '<lenorb>'
         write(irst) wrtag
         write(irst) log2int(lenorb)
         wrtag= '<usehsym>'
         write(irst) wrtag
         write(irst) log2int(usehsym)
         wrtag= '<ldirmuld>'
         write(irst) wrtag
         write(irst) log2int(ldirmuld)

         if (nexpect .gt. 0) then
            wrtag= '<nexpect>'
            write(irst) wrtag
            write(irst) nexpect
            wrtag= '<lreonly>'
            write(irst) wrtag
            write(irst) log2int(lreonly)
            wrtag= '<expectlab>'
            write(irst) wrtag
            do i=1,nexpect
               write(irst) expectlab(i)
            enddo
         endif
         if (nexpect1 .gt. 0) then
            wrtag= '<nexpect1>'
            write(irst) wrtag
            write(irst) nexpect1
            wrtag= '<lreonly1>'
            write(irst) wrtag
            write(irst) log2int(lreonly1)
            wrtag= '<expectlab1>'
            write(irst) wrtag
            do i=1,nexpect
               write(irst) expectlab1(i)
            enddo
         endif
         if (lgwp) then
            wrtag= '<gwpmodel>'
            write(irst) wrtag
            write(irst) gwpmodel
            wrtag= '<gwpintm>'
            write(irst) wrtag
            write(irst) (gwpintm(m),m=1,nmode)
            wrtag= '<gwpmodelm>'
            write(irst) wrtag
            write(irst) (gwpmodelm(m),m=1,nmode)
            wrtag= '<gwpintordm>'
            write(irst) wrtag
            write(irst) (gwpintordm(m),m=1,nmode)
            wrtag= '<igwpdynm>'
            write(irst) wrtag
            write(irst) (igwpdynm(m),m=1,nmode)
            wrtag= '<gwpcxm>'
            write(irst) wrtag
            write(irst) (gwpcxm(m),m=1,nmode)
            wrtag= '<epsgwp>'
            write(irst) wrtag
            write(irst) epsgwp
            wrtag= '<epsgsmat>'
            write(irst) wrtag
            write(irst) epsgsmat
            wrtag= '<epsdyn>'
            write(irst) wrtag
            write(irst) epsdyn
            wrtag= '<epsld>'
            write(irst) wrtag
            write(irst) epsld
            wrtag= '<linchk>'
            write(irst) wrtag
            write(irst) log2int(linchk)
            wrtag= '<lcysinv>'
            write(irst) wrtag
            write(irst) log2int(lcysinv)
            wrtag= '<lsmatovl>'
            write(irst) wrtag
            write(irst) log2int(lsmatovl)
            wrtag= '<lwholeg>'
            write(irst) wrtag
            write(irst) log2int(lwholeg)
            wrtag= '<lresint>'
            write(irst) wrtag
            write(irst) log2int(lresint)
            wrtag= '<gwpeoms>'
            write(irst) wrtag
            write(irst) gwpeoms

         endif
         if (ldd.or.lddmctdh) then
            wrtag= '<ndofdd>'
            write(irst) wrtag
            write(irst) ndofdd
            wrtag= '<ndofddpes>'
            write(irst) wrtag
            write(irst) ndofddpes
            wrtag= '<ddtransdim>'
            write(irst) wrtag
            write(irst) ddtransdim
            wrtag= '<dddiab>'
            write(irst) wrtag
            write(irst) dddiab
            wrtag= '<ddupdate>'
            write(irst) wrtag
            write(irst) ddupdate
            wrtag= '<lupdhes>'
            write(irst) wrtag
            write(irst) log2int(lupdhes)
            wrtag= '<ddlog>'
            write(irst) wrtag
            write(irst) log2int(ddlog)
            wrtag= '<ddlogfull>'
            write(irst) wrtag
            write(irst) log2int(ddlogfull)
            wrtag= '<ldbsave>'
            write(irst) wrtag
            write(irst) log2int(ldbsave)
            wrtag= '<lddwrdb>'
            write(irst) wrtag
            write(irst) log2int(lddwrdb)
            wrtag= '<lddrddb>'
            write(irst) wrtag
            write(irst) log2int(lddrddb)
            wrtag= '<numrec>'
            write(irst) wrtag
            write(irst) numrec
            wrtag= '<dbinterord>'
            write(irst) wrtag
            write(irst) dbinterord
            wrtag= '<dbinterord1>'
            write(irst) wrtag
            write(irst) dbinterord1
            wrtag= '<dbnconf>'
            write(irst) wrtag
            write(irst) dbnconf
         endif
         if(lddmctdh.or.lgausspot)then
            wrtag= '<gapupdate>'
            write(irst) wrtag
            write(irst) gapupdate
            wrtag= '<gapuptim>'
            write(irst) wrtag
            write(irst) gapuptim
         endif

         wrtag= '<end integrator info>'
         write(irst) wrtag

      else
         wrtag= '<no integrator info>'
         write(irst) wrtag

      endif

!-----------------------------------------------------------------------
! write out wavefunction 
!-----------------------------------------------------------------------
      wrtag= '<time>'
      write(irst) wrtag
      write(irst) time
      wrtag= '<temperature>'
      write(irst) wrtag
      write(irst) temperature_psi
      if (citype .gt. 0) then
         wrtag= '<jindx>'
         write(irst) wrtag
         write(irst) (jindx(b),b=1,adim)
      endif
! write instantaneous dim and tape for dynamicWF
      if (ldynspf) then
         wrtag= '<dimtape>'
         write(irst) wrtag
         write (irst) ((dim(m,s),m=1,nmode),s=1,nstate)
         if (lrunml) then
            write(irst) tapelength()
            write(irst) (tape(m),m=1,tapelength())
         endif
      endif
      wrtag= '<psi>'
      write(irst) wrtag
      write (irst) (psi(d),d=1,dgldim)

!-----------------------------------------------------------------------
! Write out data needed to restart Lanczos routine
!-----------------------------------------------------------------------
      if (lrundiag .and. (.not. linwf)) then
         wrtag= '<lanczvec>'
         write(irst) wrtag
         write(irst) (lanczvec(d),d=1,totphidim)
         wrtag= '<lanczord>'
         write(irst) wrtag
         write(irst) lanczord
         wrtag= '<alpha>'
         write(irst) wrtag
         write(irst) (alpha(d),d=1,lanczord)
         wrtag= '<beta>'
         write(irst) wrtag
         write(irst) (beta(d), d=1,lanczord)
      endif

!-----------------------------------------------------------------------
! for FG GWP, write out width matrices
!-----------------------------------------------------------------------
      if (lgwp .and. gwptype .ge. 3) then
         wrtag= '<agmat>'
         write(irst) wrtag
         write(irst) (agmat(d),d=1,agmatdim)
      endif

!-----------------------------------------------------------------------
! for trajectories, write out state associated with each point
!-----------------------------------------------------------------------
      if (ltraj) then
         wrtag= '<trajst>'
         write(irst) wrtag
         write(irst) (trajst(d),d=1,gwpinfodim)
      endif

      wrtag= '<end restart file>'
      write(irst) wrtag
      close(irst)

      return
  999 continue
      write(message,'(a,f8.3,a)') &
         'Error opening restart file at t = ',time/fs,'fs'
      call errormsg
      end subroutine wrrst

!#######################################################################
! RDRST
!
! reads restart file 
!
! before calling rdrst set the logicals
!     chkdvr
!     chkgrd
!     chkpsi
!     chkprp = 1 read integrator info
!
! also set
!    iflag = 0 read wavefunction only
!          = 1 read wavefunction and lanczos vectors (if appropriate)
!
! if linwf is .true., data is for an initial wavefunction, and
!    no information about integrator or lanczos vetcors is present 
!    in file
!
! V8.0 GW
!#######################################################################

      subroutine rdrst(time,psi,linwf,chkdvr,chkgrd,chkpsi,chkprp, &
                       lanczvec,lanczord,tridiag,iflag,jindx,agmat, &
                       trajst)

      implicit none

      integer(long), intent(inout)                     :: chkgrd
      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp,iflag
      integer(long), intent(out)                       :: lanczord
      integer(long), dimension(jindxdim), intent(out)  :: jindx
      integer(long), dimension(gwpinfodim), intent(out):: trajst

      real(dop), intent(out)                           :: time
      real(dop), dimension(*), intent(out)             :: tridiag
      complex(dop), dimension(dgldim), intent(out)     :: psi
      complex(dop),dimension(1:totphidim),intent(out)  :: lanczvec
      complex(dop), dimension(agmatdim), intent(out)   :: agmat
      logical(kind=4), intent(out)                     :: linwf

      rewind(irst)
      read(irst) filever(irst)

      if (filever(irst) .lt. 11.0_dop) then
         call rdrst_pre11(time,psi,linwf,chkdvr,chkgrd,chkpsi,chkprp, &
                          lanczvec,lanczord,tridiag,iflag,jindx,agmat, &
                          trajst)
      else
         call rdrst11(time,psi,chkdvr,chkgrd,chkpsi,chkprp, &
                      lanczvec,lanczord,tridiag,jindx,agmat, &
                      trajst)
      endif

      return
      end subroutine rdrst

!#######################################################################
!
! RDRST
!
! For a DynamicWF, before reading psi the instantaneous no. of SPFs
! are read and psi dimensions set up. These overwrite the dimensions
! on entry which were for the maximum system.
!#######################################################################

      subroutine rdrst11(time,psi,chkdvr,chkgrd,chkpsi,chkprp, &
                       lanczvec,lanczord,tridiag,jindx,agmat, &
                       trajst)

      implicit none

      integer(long)                                    :: d,ilbl,jlbl,b,m,s
      integer(long)                                    :: tapelength1
      integer(long), intent(inout)                     :: chkgrd
      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp
      integer(long), intent(out)                       :: lanczord
      integer(long), dimension(jindxdim), intent(out)  :: jindx
      integer(long), dimension(gwpinfodim), intent(out):: trajst

      real(dop), intent(out)                           :: time
      real(dop), dimension(*), intent(out)             :: tridiag
      complex(dop), dimension(dgldim), intent(out)     :: psi
      complex(dop),dimension(1:totphidim),intent(out)  :: lanczvec
      complex(dop), dimension(agmatdim), intent(out)   :: agmat
      character(len=c1)                                :: file
      logical(kind=4)                                  :: lerr
      character(len=c2)                                :: rdtag

      routine='Rdrst'
      file='Restart'
      call rstinfo11(lerr,chkdvr,chkgrd,chkpsi,chkprp)

      routine='Rdrst'
      if (lerr) then
         message='Error reading RESTART file'
         call errormsg
      endif

      rewind(irst)
      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
         if (rdtag(ilbl:jlbl) .eq. 'time') then
            read (irst,err=999) time
         else if (rdtag(ilbl:jlbl) .eq. 'temperature') then
            read (irst,err=999) temperature_psi
         else if (rdtag(ilbl:jlbl) .eq. 'jindx') then
            read(irst,err=999) (jindx(b),b=1,adim)
! read dynamicWF size
         else if (rdtag(ilbl:jlbl) .eq. 'dimtape') then
            read(irst,err=999) ((dim(m,s),m=1,nmode),s=1,nstate)
            if (lrunml) then
               read(irst,err=999) tapelength1
               read(irst,err=999) (tape(m),m=1,tapelength1)
               tree => new_wftree(tape)
            endif
            call psidat
         else if (rdtag(ilbl:jlbl) .eq. 'psi') then
            read (irst,err=999) (psi(d),d=1,dgldim)

!-----------------------------------------------------------------------
! Read data needed to restart Lanczos routine
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'lanczvec') then
            read(irst,err=999) (lanczvec(d),d=1,totphidim)
         else if (rdtag(ilbl:jlbl) .eq. 'lanczord') then
            read(irst,err=999) lanczord
         else if (rdtag(ilbl:jlbl) .eq. 'alpha') then
            read(irst,err=999) (tridiag(d),d=1,lanczord)
         else if (rdtag(ilbl:jlbl) .eq. 'beta') then
            read(irst,err=999) (tridiag(d),d=lanczord+1,2*lanczord)

!-----------------------------------------------------------------------
! for FG, read in width matrices. 
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'agmat') then
            read(irst,err=999) (agmat(d),d=1,agmatdim)

!-----------------------------------------------------------------------
! for trajectories, read in state associated with each point
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'trajst') then
            read(irst,err=999) (trajst(d),d=1,gwpinfodim)

!-----------------------------------------------------------------------
! end of file
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'end restart file') then
            exit

         endif

      enddo

  100 continue

      close(irst)
      return

  999 continue
      routine='Rdrst11'
      message='Error reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rdrst11

!#######################################################################
!
! RDRST: obtain psi only
!
!#######################################################################

      subroutine rdrstpsi(time,psi,jindx,agmat,trajst)

      implicit none

      integer(long)                                    :: d,ilbl,jlbl,b
      integer(long), dimension(jindxdim), intent(out)  :: jindx
      integer(long), dimension(gwpinfodim), intent(out):: trajst

      real(dop), intent(out)                           :: time
      complex(dop), dimension(dgldim), intent(out)     :: psi
      complex(dop), dimension(agmatdim), intent(out)   :: agmat
      character(len=c1)                                :: file
      character(len=c2)                                :: rdtag

      routine='Rdrstpsi'
      file='Restart'

      rewind(irst)
      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
         if (rdtag(ilbl:jlbl) .eq. 'time') then
            read (irst,err=999) time
         else if (rdtag(ilbl:jlbl) .eq. 'temperature') then
            read(irst,err=999) temperature_psi
         else if (rdtag(ilbl:jlbl) .eq. 'jindx') then
            read(irst,err=999) (jindx(b),b=1,adim)
         else if (rdtag(ilbl:jlbl) .eq. 'psi') then
            read (irst,err=999) (psi(d),d=1,dgldim)

!-----------------------------------------------------------------------
! for FG, read in width matrices. 
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'agmat') then
            read(irst,err=999) (agmat(d),d=1,agmatdim)

!-----------------------------------------------------------------------
! for trajectories, read in state associated with each point
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'trajst') then
            read(irst,err=999) (trajst(d),d=1,gwpinfodim)

!-----------------------------------------------------------------------
! end of file
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'end restart file') then
            exit

         endif

      enddo

  100 continue

      close(irst)
      return

  999 continue
      routine='Rdrst11'
      message='Error reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rdrstpsi

!#######################################################################
!
! RDRST
!
!#######################################################################

      subroutine rdrst_pre11(time,psi,linwf,chkdvr,chkgrd,chkpsi,chkprp, &
                       lanczvec,lanczord,tridiag,iflag,jindx,agmat, &
                       trajst)

      implicit none

      integer(long)                                    :: d,ilbl,idum,b
      integer(long), intent(inout)                     :: chkgrd
      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp,iflag
      integer(long), intent(out)                       :: lanczord
      integer(long), dimension(jindxdim), intent(out)  :: jindx
      integer(long), dimension(gwpinfodim), intent(out):: trajst

      real(dop), intent(out)                           :: time
      real(dop), dimension(*), intent(out)             :: tridiag
      complex(dop), dimension(dgldim), intent(out)     :: psi
      complex(dop),dimension(1:totphidim),intent(out)  :: lanczvec
      complex(dop), dimension(agmatdim), intent(out)   :: agmat
      logical(kind=4), intent(out)                     :: linwf
      character(len=c1)                                :: file
      logical(kind=4)                                  :: lerr

      routine='Rdrst'
      file='Restart'
      call rstinfo_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      routine='Rdrst'
      if (lerr) then
         message='Error reading RESTART file'
         call errormsg
      endif

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
      read (irst,err=999) time
      if (citype .gt. 0) then
         read(irst,err=999) (jindx(b),b=1,adim)
      endif
      read (irst,err=999) (psi(d),d=1,dgldim)
      if (iflag .eq. 0) go to 100

!-----------------------------------------------------------------------
! Read data needed to restart Lanczos routine
!-----------------------------------------------------------------------
      if (lrundiag .and. (.not. linwf)) then
         read(irst,err=999) idum
         read(irst,err=999) (lanczvec(d),d=1,totphidim)
         read(irst,err=999) lanczord
         read(irst,err=999) (tridiag(d),d=1,lanczord)
         read(irst,err=999) (tridiag(d),d=lanczord+1,2*lanczord)
      endif

 100  continue

!-----------------------------------------------------------------------
! for FG, read in width matrices. 
!-----------------------------------------------------------------------
      if (lgwp .and. gwptype .ge. 3) then
         if (filever(irst) .lt. 10.1d0) then
            read(irst,err=999) (agmat(d),d=1,agmatdim100)
         else if (filever(irst) .lt. 10.2d0) then
            read(irst,err=999) (agmat(d),d=1,agmatdim101)
         else
            read(irst,err=999) (agmat(d),d=1,agmatdim)
         endif
      endif

!-----------------------------------------------------------------------
! for direct dynamics, read in present surface information (redundant)
!-----------------------------------------------------------------------
      if (ldd .and. (.not. linwf)) then
         read(irst,err=999) 
      endif

!-----------------------------------------------------------------------
! for trajectories, read in state associated with each point
!-----------------------------------------------------------------------
      if (ltraj) then
         read(irst,err=999) (trajst(d),d=1,gwpinfodim)
      endif

      close(irst)
      return

  999 continue
      ilbl=index(file,' ')-1
      message='Error reading '//file(1:ilbl)//' file'
      call errormsg

      end subroutine rdrst_pre11

!#######################################################################
! RSTINFO
!
! reads first part of restart file, which contains information on
! size of wavefunction and related parameters.
!
! V8.0 GW
!#######################################################################

      subroutine rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long), intent(in)    :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout) :: chkgrd
      logical(kind=4), intent(out) :: lerr,linwf

      rewind(irst)
      read(irst) filever(irst)

      if (filever(irst) .lt. 11.00_dop) then
         call rstinfo_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
      else
         call rstinfo11(lerr,chkdvr,chkgrd,chkpsi,chkprp)
      endif

      return

      end subroutine rstinfo

!#######################################################################

      subroutine rstinfo11(lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long) :: ndof1,ilrlx,ilconst,ilfunkl,ilfunkr, &
                       ilexact,ilconstfix,ilconstvar,ilcmfadaphi,ilopro,&
                       iliork,ilimportho,ilrktest,ilrunprop,ilrundiag,ktot1,&
                       ilinchk,nmode1,ilcmfadaa,i,m,d,&
                       ilcap,ilreal,ilrreal,ierr,ilnaturcmf,ilwholeg,&
                       ilresint,ilrealwp,ilreonly,ilreonly1,ilrunml,&
                       icysinv,ismatovl,ilupdhes,iddlog,iddlogfull,ildbsave,&
                       ilddwrdb,ilddrddb
      integer(long) :: gwpcx,gwpintord,gwpint,igwpdyn

      integer(long), intent(in)               :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout)            :: chkgrd
      integer(long)                           :: ilbl,jlbl

      integer(long), dimension(maxdim)        :: fdvr,mdvr,mmode,nspfdof1
      integer(long), dimension(maxdim,maxdim) :: spfdof1

      integer(long)                           :: mnspec1
      integer(long)                           :: imlcmfsplit, irealphi, &
                                                 irhonrm, ienorb, &
                                                 iusehsym, idirmuld
      integer(long), dimension(2*maxdim)      :: mspec1, minttyp1, &
                                                 mintord1, mintacc1, &
                                                 mintdt1

      character(len=c4)                       :: file
      character(len=c5)                       :: title,fname
      character(len=100)                      :: buffer
      logical(kind=4), external               :: int2log
      logical(kind=4), intent(out)            :: lerr
      character(len=c2)                       :: rdtag


!-----------------------------------------------------------------------
! Title
!-----------------------------------------------------------------------
      rewind(irst)
      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle
         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'runtitle') then
            read(irst,err=999) title

         else if (rdtag(ilbl:jlbl) .eq. 'creation date') then
            read(irst,err=999) buffer

         else if (rdtag(ilbl:jlbl) .eq. 'end rsttitle') then
            exit

         endif

      enddo

  100 continue
!     lwrlog: avoid flooding log file with RESTART information
      if (lwrlog) then
         write(ilog,'(/,a)') &
              '**** RSTINFO: Reading RESTART file ****'
         inquire(irst, name=fname)
         write(ilog,'(a)') 'File name: '//trim(fname)
         if(title .ne. ' ') write(ilog,'(a80)') trim(title)
         if (dentype.eq.0) then
            write(ilog,'(a,f9.5)') 'Type : wavefunction, file-version =', &
                                 filever(irst)
         else if (dentype.eq.1) then
            write(ilog,'(a)') 'Type : density operator I '
         else if (dentype.eq.2) then
            write(ilog,'(a)') 'Type : density operator II'
         endif
         write(ilog,'(a)') 'Created on:'
         write(ilog,'(a)') trim(buffer)
      end if

!-----------------------------------------------------------------------
! Read DVR, GRD and PSI data
!-----------------------------------------------------------------------
      call zerovxi(chkerr,nerr)
      call rddvrdef(irst,chkdvr,ndof1,fdvr)
      call rdgrddef(irst,chkgrd,fdvr,mdvr,mmode,nspfdof1,spfdof1,nmode1)
      if(chkdvr.eq.6) then 
         chkerr(2)=0
         chkerr(5)=0
      endif
      call rdpsidef(irst,chkpsi)

      file='Restart'
      call chksyserr(file,lerr)
!.....Ignore different basis if lignore=.true. 
      if(lignore.and.lerr) then
         lerr=.false.
         write(6,'(/a,/a)') ' #############  WARNING  ##############', &
                     'Different primitive bases: system/restart-file'
         write(ilog,'(/a,/a)') ' ############  WARNING  #############', &
                     'Different primitive bases: system/restart-file'
      endif
      if (lerr) return
      routine='Rstinfo'

!     Avoid floading the efield-logfile when doing optimal control.   
!     This has been added because the restart file is re-opened 
!     for every time step by the efield routine.
!     Shifting the file pointer back and wait 
!     for update of the restart file does not always work
!     since the file pointer might become invalid. 

!-----------------------------------------------------------------------
! read restart information 
!-----------------------------------------------------------------------
      ierr=2
      if (chkprp .eq. 1) then
         rewind(irst)

         do
   20       continue
            read(irst,err=20,end=200) rdtag
            if (rdtag(1:1) .ne. '<') cycle

            ilbl=2
            jlbl=index(rdtag,'>')-1
            if (rdtag(ilbl:jlbl) .eq. 'timeunit') then
               read (irst,err=999) fs
            else if (rdtag(ilbl:jlbl) .eq. 'energyunit') then
               read (irst,err=999) ev
            else if (rdtag(ilbl:jlbl) .eq. 'rlxunit') then
               read (irst,err=999) rlxunit
            else if (rdtag(ilbl:jlbl) .eq. 'rlxtime') then
               read (irst,err=999) rlxtime
            else if (rdtag(ilbl:jlbl) .eq. 'temperature0') then
               read (irst,err=999) temperature0
            else if (rdtag(ilbl:jlbl) .eq. 'normthermal') then
               read (irst,err=999) normthermal
            else if (rdtag(ilbl:jlbl) .eq. 'timecap') then
               read (irst,err=999) (timecap(d),d=1,maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'no integrator info') then
               exit

            else if (rdtag(ilbl:jlbl) .eq. 'end integrator info') then
               exit

! test
            else if (rdtag(ilbl:jlbl) .eq. 'proptask') then
               read (irst,err=999) proptask
! test
            else if (rdtag(ilbl:jlbl) .eq. 'lrlx') then
               read (irst,err=999) ilrlx
               lrlx = int2log(ilrlx)
            else if (rdtag(ilbl:jlbl) .eq. 'lrunprop') then
               read (irst,err=999) ilrunprop
               irunprop = ilrunprop
            else if (rdtag(ilbl:jlbl) .eq. 'lrundiag') then
               read (irst,err=999) ilrundiag
               lrundiag = int2log(ilrundiag)
            else if (rdtag(ilbl:jlbl) .eq. 'out1') then
               read (irst,err=999) out1
            else if (rdtag(ilbl:jlbl) .eq. 'out2') then
               read (irst,err=999) out2
            else if (rdtag(ilbl:jlbl) .eq. 'lconst') then
               read (irst,err=999) ilconst
               lconst = int2log(ilconst)
            else if (rdtag(ilbl:jlbl) .eq. 'lfunkl') then
               read (irst,err=999) ilfunkl
               lfunkl = int2log(ilfunkl)
            else if (rdtag(ilbl:jlbl) .eq. 'lfunkr') then
               read (irst,err=999) ilfunkr
               lfunkr = int2log(ilfunkr)
            else if (rdtag(ilbl:jlbl) .eq. 'lexact') then
               read (irst,err=999) ilexact
               lexact = int2log(ilexact)
            else if (rdtag(ilbl:jlbl) .eq. 'lrealwp') then
               read (irst,err=999) ilrealwp
               lrealwp= int2log(ilrealwp)
            else if (rdtag(ilbl:jlbl) .eq. 'lrunml') then
               read (irst,err=999) ilrunml
               lrunml= int2log(ilrunml)

            else if (rdtag(ilbl:jlbl) .eq. 'lconstfix') then
               read (irst,err=999) ilconstfix
               lconstfix  = int2log(ilconstfix)
            else if (rdtag(ilbl:jlbl) .eq. 'lconstvar') then
               read (irst,err=999) ilconstvar
               lconstvar  = int2log(ilconstvar)
            else if (rdtag(ilbl:jlbl) .eq. 'lcmfadaphi') then
               read (irst,err=999) ilcmfadaphi
               lcmfadaphi = int2log(ilcmfadaphi)
            else if (rdtag(ilbl:jlbl) .eq. 'lcmfadaa') then
               read (irst,err=999) ilcmfadaa
               lcmfadaa = int2log(ilcmfadaa)
            else if (rdtag(ilbl:jlbl) .eq. 'inttyp') then
               read (irst,err=999) inttyp
            else if (rdtag(ilbl:jlbl) .eq. 'inttypea') then
               read (irst,err=999) inttypea
            else if (rdtag(ilbl:jlbl) .eq. 'inttypephi') then
               read (irst,err=999) inttypephi
            else if (rdtag(ilbl:jlbl) .eq. 'intord') then
               read (irst,err=999) intord
            else if (rdtag(ilbl:jlbl) .eq. 'intordera') then
               read (irst,err=999) intordera
            else if (rdtag(ilbl:jlbl) .eq. 'intorderphi') then
               read (irst,err=999) intorderphi
            else if (rdtag(ilbl:jlbl) .eq. 'accuracy') then
               read (irst,err=999) accuracy
            else if (rdtag(ilbl:jlbl) .eq. 'accavec') then
               read (irst,err=999) accavec
            else if (rdtag(ilbl:jlbl) .eq. 'accphi') then
               read (irst,err=999) accphi
            else if (rdtag(ilbl:jlbl) .eq. 'dt') then
               read (irst,err=999) dt
            else if (rdtag(ilbl:jlbl) .eq. 'dta') then
               read (irst,err=999) dta
            else if (rdtag(ilbl:jlbl) .eq. 'dtphi') then
               read (irst,err=999) dtphi
            else if (rdtag(ilbl:jlbl) .eq. 'minrk5') then
               read (irst,err=999) minrk5
            else if (rdtag(ilbl:jlbl) .eq. 'updatetime') then
               read (irst,err=999) updatetime
            else if (rdtag(ilbl:jlbl) .eq. 'epsupdate') then
               read (irst,err=999) epsupdate
            else if (rdtag(ilbl:jlbl) .eq. 'macheps') then
               read (irst,err=999) macheps
            else if (rdtag(ilbl:jlbl) .eq. 'epsmat') then
               read (irst,err=999) epsmat
            else if (rdtag(ilbl:jlbl) .eq. 'epssvd') then
               read (irst,err=999) epssvd
            else if (rdtag(ilbl:jlbl) .eq. 'lopro') then
               read (irst,err=999) ilopro
               lopro     = int2log(ilopro)
            else if (rdtag(ilbl:jlbl) .eq. 'liork') then
               read (irst,err=999) iliork
               liork     = int2log(iliork)
            else if (rdtag(ilbl:jlbl) .eq. 'limportho') then
               read (irst,err=999) ilimportho
               limportho = int2log(ilimportho)
            else if (rdtag(ilbl:jlbl) .eq. 'lrktest') then
               read (irst,err=999) ilrktest
               lrktest = int2log(ilrktest)
            else if (rdtag(ilbl:jlbl) .eq. 'lnaturcmf') then
               read (irst,err=999) ilnaturcmf
               lnaturcmf = int2log(ilnaturcmf)
            else if (rdtag(ilbl:jlbl) .eq. 'lcap') then
               read(irst,err=999) ilcap
               lcap   = int2log(ilcap)
!            else if (rdtag(ilbl:jlbl) .eq. 'autocap') then
!               read(irst,err=999) (autocap(k),k=1,ktot1)
            else if (rdtag(ilbl:jlbl) .eq. 'lreal') then
               read(irst,err=999) ilreal
               lreal  = int2log(ilreal)
            else if (rdtag(ilbl:jlbl) .eq. 'lrreal') then
               read(irst,err=999) ilrreal
               lrreal = int2log(ilrreal)
            else if (rdtag(ilbl:jlbl) .eq. 'estart') then
               read (irst,err=999) estart
            else if (rdtag(ilbl:jlbl) .eq. 'ktot') then
               read(irst,err=999) ktot1

            else if (rdtag(ilbl:jlbl) .eq. 'lmlcmfsplit') then
               read(irst,err=999) imlcmfsplit
               lmlcmfsplit = int2log(imlcmfsplit)
            else if (rdtag(ilbl:jlbl) .eq. 'mnspec') then
               read(irst,err=999) mnspec1
            else if (rdtag(ilbl:jlbl) .eq. 'mspec') then
               read(irst,err=999) (mspec1(d),d=1,2*maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'minttyp') then
               read(irst,err=999) (minttyp1(d),d=1,2*maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'mintord') then
               read(irst,err=999) (mintord1(d),d=1,2*maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'mintacc') then
               read(irst,err=999) (mintacc1(d),d=1,2*maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'mintdt') then
               read(irst,err=999) (mintdt1(d),d=1,2*maxdim)
            else if (rdtag(ilbl:jlbl) .eq. 'lrealphi') then
               read(irst,err=999) irealphi
               lrealphi = int2log(irealphi)
            else if (rdtag(ilbl:jlbl) .eq. 'lrhonrm') then
               read(irst,err=999) irhonrm
               lrhonrm = int2log(irhonrm)
            else if (rdtag(ilbl:jlbl) .eq. 'lenorb') then
               read(irst,err=999) ienorb
               lenorb = int2log(ienorb)
            else if (rdtag(ilbl:jlbl) .eq. 'usehsym') then
               read(irst,err=999) iusehsym
               usehsym = int2log(iusehsym)
            else if (rdtag(ilbl:jlbl) .eq. 'ldirmuld') then
               read(irst,err=999) idirmuld
               ldirmuld = int2log(idirmuld)

            else if (rdtag(ilbl:jlbl) .eq. 'nexpect') then
               read(irst,err=999) nexpect
            else if (rdtag(ilbl:jlbl) .eq. 'lreonly') then
               read(irst,err=999) ilreonly
               lreonly = int2log(ilreonly)
            else if (rdtag(ilbl:jlbl) .eq. 'expectlab') then
               if (.not. allocated(expectlab)) allocate(expectlab(nexpect))
               if (.not. allocated(expvalue)) allocate(expvalue(nexpect))
               do i=1,nexpect
                  read(irst,err=999) expectlab(i)
               enddo
            else if (rdtag(ilbl:jlbl) .eq. 'nexpect1') then
               read(irst,err=999) nexpect1
            else if (rdtag(ilbl:jlbl) .eq. 'lreonly1') then
               read(irst,err=999) ilreonly1
               lreonly1 = int2log(ilreonly1)
            else if (rdtag(ilbl:jlbl) .eq. 'expectlab1') then
               if (.not. allocated(expectlab1)) allocate(expectlab1(nexpect1))
               if (.not. allocated(expvalue1)) allocate(expvalue1(nexpect1))
               do i=1,nexpect1
                  read(irst,err=999) expectlab1(i)
               enddo

            else if (rdtag(ilbl:jlbl) .eq. 'gwpmodel') then
               read(irst,err=999) gwpmodel
            else if (rdtag(ilbl:jlbl) .eq. 'gwpmodelm') then
               read(irst,err=999) (gwpmodelm(m),m=1,nmode)
! Earlier versions had a single gwpintord. Read and set gwpintordm
            else if (rdtag(ilbl:jlbl) .eq. 'gwpintord' .or. &   ! older versions
                     rdtag(ilbl:jlbl) .eq. 'lhaorder') then     ! used lhaorder
               read(irst,err=999) gwpintord
               gwpintordm(:) = gwpintord                        
            else if (rdtag(ilbl:jlbl) .eq. 'gwpintordm') then
               read(irst,err=999) (gwpintordm(m),m=1,nmode)
            else if (rdtag(ilbl:jlbl) .eq. 'gwpint') then
               read(irst,err=999) gwpint
               gwpintm(:) = gwpint
! Earlier versions had a single gwpcx. Read and set gwpcxm
            else if (rdtag(ilbl:jlbl) .eq. 'gwpcx') then
               read(irst,err=999) gwpcx
               gwpcxm(:)=gwpcx
            else if (rdtag(ilbl:jlbl) .eq. 'gwpcxm') then
               read(irst,err=999) (gwpcxm(m),m=1,nmode)
            else if (rdtag(ilbl:jlbl) .eq. 'igwpdyn') then
               read(irst,err=999) igwpdyn
               igwpdynm(:) = igwpdyn
            else if (rdtag(ilbl:jlbl) .eq. 'igwpdynm') then
               read(irst,err=999) (igwpdynm(m),m=1,nmode)
            else if (rdtag(ilbl:jlbl) .eq. 'epsgwp') then
               read(irst,err=999) epsgwp
            else if (rdtag(ilbl:jlbl) .eq. 'epsgsmat') then
               read(irst,err=999) epsgsmat
            else if (rdtag(ilbl:jlbl) .eq. 'epsdyn') then
               read(irst,err=999) epsdyn
            else if (rdtag(ilbl:jlbl) .eq. 'epsld') then
               read(irst,err=999) epsld
            else if (rdtag(ilbl:jlbl) .eq. 'linchk') then
               read(irst,err=999) ilinchk
               linchk=int2log(ilinchk)
            else if (rdtag(ilbl:jlbl) .eq. 'lcysinv') then
               read(irst,err=999) icysinv
               lcysinv=int2log(icysinv)
            else if (rdtag(ilbl:jlbl) .eq. 'lsmatovl') then
               read(irst,err=999) ismatovl
               lsmatovl=int2log(ismatovl)
            else if (rdtag(ilbl:jlbl) .eq. 'lwholeg') then
               read(irst,err=999) ilwholeg
               lwholeg=int2log(ilwholeg)
            else if (rdtag(ilbl:jlbl) .eq. 'lresint') then
               read(irst,err=999) ilresint
               lresint=int2log(ilresint)
            else if (rdtag(ilbl:jlbl) .eq. 'gwpeoms') then
               read(irst,err=999) gwpeoms
            else if (rdtag(ilbl:jlbl) .eq. 'ndofdd') then
               read(irst,err=999) ndofdd
            else if (rdtag(ilbl:jlbl) .eq. 'ndofddpes') then
               read(irst,err=999) ndofddpes
            else if (rdtag(ilbl:jlbl) .eq. 'ddtransdim') then
               read(irst,err=999) ddtransdim
            else if (rdtag(ilbl:jlbl) .eq. 'dddiab') then
               read(irst,err=999) dddiab
            else if (rdtag(ilbl:jlbl) .eq. 'ddupdate') then
               read(irst,err=999) ddupdate
            else if (rdtag(ilbl:jlbl) .eq. 'lupdhes') then
               read(irst,err=999) ilupdhes
               lupdhes=int2log(ilupdhes)
            else if (rdtag(ilbl:jlbl) .eq. 'ddlog') then
               read(irst,err=999) iddlog
               ddlog=int2log(iddlog)
            else if (rdtag(ilbl:jlbl) .eq. 'ddlogfull') then
               read(irst,err=999) iddlogfull
               ddlogfull=int2log(iddlogfull)
            else if (rdtag(ilbl:jlbl) .eq. 'ldbsave') then
               read(irst,err=999) ildbsave
               ldbsave=int2log(ildbsave)
            else if (rdtag(ilbl:jlbl) .eq. 'lddwrdb') then
               read(irst,err=999) ilddwrdb
               lddwrdb=int2log(ilddwrdb)
            else if (rdtag(ilbl:jlbl) .eq. 'lddrddb') then
               read(irst,err=999) ilddrddb
               lddrddb=int2log(ilddrddb)

            else if (rdtag(ilbl:jlbl) .eq. 'numrec') then
               read(irst,err=999) numrec
            else if (rdtag(ilbl:jlbl) .eq. 'dbinterord') then
               read(irst,err=999) dbinterord
            else if (rdtag(ilbl:jlbl) .eq. 'dbinterord1') then
               read(irst,err=999) dbinterord1
            else if (rdtag(ilbl:jlbl) .eq. 'dbnconf') then
               read(irst,err=999) dbnconf

            else if (rdtag(ilbl:jlbl) .eq. 'gapupdate') then
               read(irst,err=999) gapupdate
            else if (rdtag(ilbl:jlbl) .eq. 'gapuptim') then
               read(irst,err=999) gapuptim
            endif

         enddo
  200    continue

      endif

      return
  999 continue
      routine='Rstinfo11'
      message='Error Reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg
      end subroutine rstinfo11

!#######################################################################

      subroutine rstinfo_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long) :: ndof1,idum,ilinwf,ilrlx,ilconst,ilfunkl,ilfunkr, &
                       ilexact,ilconstfix,ilconstvar,ilcmfadaphi,ilopro,&
                       ilrunprop,ilrundiag,ktot1,k,nmode1,ilcmfadaa,i,s,&
                       ilcap,ilreal,ilrreal,ierr,ilnaturcmf,ilwholeg,&
                       ilresint,ilrealwp,ivardim,ilrunml,irr,d,i1,i2,mxdm
      integer(long) :: gwpcx,gwpintord,igwpdyn,gwpint

      integer(long), intent(in)                     :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout)                  :: chkgrd

      integer(long), dimension(maxdim)              :: fdvr,mdvr,mmode,nspfdof1
      integer(long), dimension(maxdim,maxdim)       :: spfdof1

      real(dop)                                     :: rdum
      complex(dop)                                  :: cdum
      character(len=c4)                             :: file
      character(len=c5)                             :: title
      character(len=100)                            :: buffer
      logical(kind=4), external                     :: int2log
      logical(kind=4), intent(out)                  :: lerr,linwf

      call zerovxi(chkerr,nerr)
      call rddvrdef(irst,chkdvr,ndof1,fdvr)
      call rdgrddef(irst,chkgrd,fdvr,mdvr,mmode,nspfdof1,spfdof1,nmode1)
      if(chkdvr.eq.6) then 
         chkerr(2)=0
         chkerr(5)=0
      endif
      call rdpsidef(irst,chkpsi)

      file='Restart'
      call chksyserr(file,lerr)
!.....Ignore different basis if lignore=.true. 
      if(lignore.and.lerr) then
         lerr=.false.
         write(6,'(/a,/a)') ' #############  WARNING  ##############', &
                     'Different primitive bases: system/restart-file'
         write(ilog,'(/a,/a)') ' ############  WARNING  #############', &
                     'Different primitive bases: system/restart-file'
      endif
      if (lerr) return
      routine='Rstinfo'

      ierr=1
      read(irst,err=999) title
      read(irst,err=999) buffer
      
!     Avoid floading the efield-logfile when doing optimal control.   
!     This has been added because the restart file is re-opened 
!     for every time step by the efield routine.
!     Shifting the file pointer back and wait 
!     for update of the restart file does not always work
!     since the file pointer might become invalid. 

      if (lwrlog) then
         write(ilog,'(/,a)') &
              '**** RSTINFO: Reading RESTART file ****'
         if(title .ne. ' ') write(ilog,'(a80)') title
         if (dentype.eq.0) then
            write(ilog,'(a,f9.5)') 'Type : wavefunction, file-version =', &
                                 filever(irst)
         else if (dentype.eq.1) then
            write(ilog,'(a)') 'Type : density operator I '
         else if (dentype.eq.2) then
            write(ilog,'(a)') 'Type : density operator II'
         endif
         write(ilog,'(a)') 'Created on:'
         write(ilog,'(a)') buffer
      end if

!-----------------------------------------------------------------------
! read restart information 
!-----------------------------------------------------------------------
      ierr=2
      if (filever(min(irst,maxfver)) .ge. 8.60060d0) then
         read(irst,iostat=irr) ilinwf, normthermal
         if(irr.gt.0) then
            normthermal=0
            backspace(irst)
            read (irst,err=999) ilinwf
         else
            read(irst,iostat=irr) (timecap(d),d=1,ndof1)
            if(irr.gt.0)  backspace(irst)
         endif
      else
         normthermal=0
         read(irst,err=999) ilinwf
      endif

      linwf = int2log(ilinwf)
      if ((.not. linwf) .and. chkprp .eq. 0) then
         ierr=3
         if (filever(irst) .ge. 8.1005) then
            read (irst,err=999) idum,idum,idum
         else
            read (irst,err=999) idum
         endif
         ierr=4
         read (irst,err=999) rdum,rdum
         if (filever(irst) .ge. 8.5000 .and. &
                     filever(irst) .lt. 9.0000_dop) then
             read (irst,err=999) idum,idum,idum,idum,idum
         else
             read (irst,err=999) idum,idum,idum,idum
         end if
         read (irst,err=999) idum,idum,idum
         if (filever(irst) .ge. 9.0001_dop) then
            read (irst,err=999) idum
         else if (filever(irst) .ge. 9.0000_dop) then
            read (irst,err=999) idum,idum
         endif
         read (irst,err=999) idum,idum,idum
         read (irst,err=999) idum,idum,idum
         read (irst,err=999) rdum,rdum,rdum
         read (irst,err=999) rdum,rdum,rdum
         if (filever(irst) .ge. 10.2000_dop) then
            read (irst,err=999) rdum
         endif
         read (irst,err=999) rdum,rdum
         if (filever(irst) .ge. 9.0015_dop) then
            read (irst,err=999) rdum,rdum
         else 
            read (irst,err=999) rdum,rdum,rdum
         endif
         if (filever(irst) .ge. 9.0015_dop) then
            read (irst,err=999) idum
         else if (filever(irst) .ge. 9.0014_dop) then
            read (irst,err=999) idum,idum,idum
         else
            read (irst,err=999) idum,idum,idum,idum
         endif
         ierr=5
         if (filever(irst) .ge. 10.0001_dop) then
            read (irst,err=999) idum
         endif
         if(filever(irst) .ge. 8.3012_dop) then
            read (irst,err=999) idum,idum,idum
         endif
         ierr=6
         read (irst,err=999) rdum
         if (filever(irst) .ge. 8.200_dop) then
            read (irst,err=999) idum,idum
         endif
         if(filever(irst) .ge. 8.5015d0 .and. &
            filever(irst) .le. 8.6015d0) then
            ierr=14
            read(irst,err=999) idum, mxdm
            mxdm=min(mxdm,maxdim)
            ierr=15
            read(irst,err=999,end=999) (cdum,d=1,2*mxdm)
            read(irst,err=999,end=999) (idum,d=1,2*mxdm)
            read(irst,err=999,end=999) (idum,d=1,2*mxdm)
            read(irst,err=999,end=999) (rdum,d=1,2*mxdm)
            read(irst,err=999,end=999) (rdum,d=1,2*mxdm)
            ierr=16
            read(irst,err=999) idum, idum
            read(irst,err=999) idum, idum
            read(irst,err=999) idum, idum
            read(irst,err=999) idum, idum
         endif
         if (lgwp) then
            read(irst,err=999) idum,idum,rdum,rdum
            if (filever(irst) .ge. 10.0002_dop) then
               read(irst,err=999) idum,rdum,idum,idum
            endif
            if (filever(irst) .ge. 10.0004_dop) then
               read(irst,err=999) idum
               read(irst,err=999) idum,idum
            endif
         endif
         if (ldd) then
            read(irst,err=999) idum,idum,idum
            read(irst,err=999) idum,idum,idum
            read(irst,err=999) idum
            read(irst,err=999) idum
         endif
      else if ((.not. linwf) .and. chkprp .eq. 1) then
         ierr=7
         if (filever(irst) .ge. 8.1005) then
            read (irst,err=999) ilrlx,ilrunprop,ilrundiag
            lrlx = int2log(ilrlx)
            irunprop = ilrunprop
            lrundiag = int2log(ilrundiag)
         else
            read (irst,err=999) ilrlx
            lrlx = int2log(ilrlx)
         endif
         ierr=8
         read (irst,err=999) out1,out2
         if (filever(irst) .ge. 8.5000 .and. &
                     filever(irst) .lt. 9.0000_dop) then
            read (irst,err=999) ilconst,ilfunkl,ilfunkr,ilexact,ilrunml
            lrunml = int2log(ilrunml)
         else if (filever(irst) .ge. 9.0001_dop) then
            read (irst,err=999) ilconst,ilfunkl,ilfunkr,ilexact,ilrealwp,ilrunml
            lrunml = .false.
         else
            read (irst,err=999) ilconst,ilfunkl,ilfunkr,ilexact
            lrunml = .false.
         end if

         lconst = int2log(ilconst)
         lfunkl = int2log(ilfunkl)
         lfunkr = int2log(ilfunkr)
         lexact = int2log(ilexact)
         lrealwp= int2log(ilrealwp)
         lrunml = int2log(ilrunml)
         ierr=9
         read (irst,err=999) ilconstfix,ilconstvar,ilcmfadaphi
         lconstfix  = int2log(ilconstfix)
         lconstvar  = int2log(ilconstvar)
         lcmfadaphi = int2log(ilcmfadaphi)
         if (filever(irst) .ge. 9.0001_dop) then
            read (irst,err=999) ilcmfadaa
            lcmfadaa = int2log(ilcmfadaa)
         else if (filever(irst) .ge. 9.0000_dop) then
            read (irst,err=999) ilcmfadaa,idum
            lcmfadaa = int2log(ilcmfadaa)
         endif
         ierr=10
         read (irst,err=999) inttyp,inttypea,inttypephi
         read (irst,err=999) intord,intordera,intorderphi
         read (irst,err=999) accuracy,accavec,accphi
         read (irst,err=999) dt,dta,dtphi
         if (filever(irst) .ge. 10.2000_dop) then
            read (irst,err=999) minrk5
         endif
         read (irst,err=999) updatetime,epsupdate
         if (filever(irst) .ge. 9.0015_dop) then
            read (irst,err=999) macheps,epsmat
         else
            read (irst,iostat=irr) macheps,rdum,epsmat,epssvd
            if(irr.gt.0) then
               backspace(irst)
               read (irst,err=999) macheps,rdum,epsmat
            endif
         endif
         if (filever(irst) .ge. 9.0015_dop) then
            read (irst,err=999) ilopro
         else if (filever(irst) .ge. 9.0014_dop) then
            read (irst,err=999) idum,idum,ilopro
         else
            read (irst,err=999) idum,idum,idum,ilopro
         endif
         lopro     = int2log(ilopro)
         if (filever(irst) .ge. 10.0001_dop) then
            read (irst,err=999) ilnaturcmf
         endif
         lnaturcmf = int2log(ilnaturcmf)
         
         ierr=11
         if(filever(irst) .ge. 8.3012_dop) then
            read(irst,err=999) ilcap,ilreal,ilrreal
            lcap   = int2log(ilcap)
            lreal  = int2log(ilreal)
            lrreal = int2log(ilrreal)
         endif
         ierr=12
         read (irst,err=999) estart
         ierr=13
         if (filever(irst) .ge. 8.200_dop) then
            read(irst,err=999) ktot1,(autocap(k),k=1,ktot1)
         endif
         if(filever(irst) .ge. 8.5015d0 .and. &
            filever(irst) .le. 8.6015d0) then
            ierr=14
            read(irst,err=999) mnspec,mxdm
            mxdm=min(mxdm,maxdim)
            ierr=15
            read(irst,err=999) (mspec(d),  d=1,2*mxdm)
            read(irst,err=999) (minttyp(d),d=1,2*mxdm)
            read(irst,err=999) (mintord(d),d=1,2*mxdm)
            read(irst,err=999) (mintacc(d),d=1,2*mxdm)
            read(irst,err=999) (mintdt(d), d=1,2*mxdm)
            ierr=16
            read(irst,err=999) i1,i2
            lmlcmfsplit=int2log(i1)
            liork=int2log(i2)
            read(irst,err=999) i1,i2
            lrealphi=int2log(i1)
            lrhonrm=int2log(i2)
            read(irst,err=999) i1,i2
            limportho=int2log(i1)
            lenorb=int2log(i2)
            read(irst,err=999) i1,i2
            usehsym=int2log(i1)
            ldirmuld=int2log(i2)
         endif
         if (lgwp) then
            read(irst,err=999) gwpint,gwpintord,epsgwp,rdum
            gwpintordm(:) = gwpintord
            gwpintm(:) = gwpint
            if (filever(irst) .ge. 10.0002_dop) then
               read(irst,err=999) igwpdyn,epsdyn,ilwholeg,ilresint
               igwpdynm = igwpdyn
               lwholeg=int2log(ilwholeg)
               lresint=int2log(ilresint)
               igwpdynm(:) = igwpdyn
            endif
            if (filever(irst) .ge. 10.0004_dop) then
               read(irst,err=999) ivardim
               read(irst,err=999) gwpeoms,gwpcx
               gwpcxm(:) = gwpcx
            endif
         endif
         if (ldd) then
            read(irst,err=999) idum,idum,(idum,i=1,3), &
                        (idum,s=1,nstate)
            read(irst,err=999) ndofdd,ndofddpes,ddtransdim
            read(irst,err=999) dddiab
            read(irst,err=999) ddupdate
         endif
      endif

      return
  999 continue
      if(ierr.eq.1) write(6,'(/a)') '### Is this a restart file? ###'
      routine='rstinfo (iorst.F line 366)'
      write(message,'(a,i4)') &
          'Error reading RESTART file (RSTINFO). Error No.',ierr
      call errormsg
      end subroutine rstinfo_pre11

!************************************************************************

      subroutine splitrst(psi,dtpsi,dgldim1,time,jindx,agmat,trajst)

! --- This routine splits the resatrt file of a block-Davidson run
! --- into individual restart files (without the "electronic" state)
! --- This routine should only be called at the very end of a calculation
! --- because it resets some parameters.

      implicit none

      integer(long), intent(in)                        ::  dgldim1

      integer(long)                                    :: i,iz,s,idum,mpb1,&
                                                          fpb1,npack1,adim1
      real(dop), dimension(1)                          :: rdum
      real(dop)                                        :: time
      complex(dop),dimension(dgldim1), intent(in)      :: psi
      complex(dop),dimension(dgldim1), intent(out)     :: dtpsi
      complex(dop),dimension(1)                        :: cdum
      logical(kind=4)                                  :: linwf

      integer(long), dimension(jindxdim), intent(in)   :: jindx
      integer(long), dimension(gwpinfodim), intent(in) :: trajst
      complex(dop), dimension(agmatdim), intent(in)    :: agmat


! --- reset logicals and parameters to fit the new structure
      lmulpack=.false.
      mpb1=mpb
      fpb1=fpb
      mpb=0
      fpb=0
      nmode=nmode-1
      ndof=ndof-1
      maxdim=maxdim-1
      adim1=adim
      adim=adim/npackts
      npack1=npackts
      npackts=1
      totphidim=totphidim-npack1**2
      if(lreflex) then
         ndof=ndof-1
         nmode=nmode-1
         adim=adim/2
         leb=.false.
         meb=0
         feb=0
         npack1=npack1*2
         totphidim=totphidim-4
      endif
      dgldim=adim+totphidim
      linwf=.true.
      idum=0
      rdum(1)=0.d0
      cdum(1)=0.d0
      do s=1,nstate
         block(s)=block(s)/npack1
      enddo
      zpsi(1) = 1
      do s=1,nstate-1
         zpsi(s+1) = zpsi(s) + block(s)
      enddo
      if(lmult) dofspf(feb) = 0
      call griddat

! --- copy orbitals
      call cpvxz(psi(adim1+1),dtpsi(adim+1),totphidim)

! --- copy A-vector and write rst file
      if(lmult) then
         do i = 1, npack1
            irst=49+i
            iz=1    ! iz = zpsi of block WF
            do s=1,nstate
               call cpvxz(psi(iz+(i-1)*block(s)),dtpsi(zpsi(s)),block(s))
               iz = iz + block(s)*npack1
            enddo
            call wrrst(time,dtpsi,linwf,cdum,idum,rdum,rdum,jindx,agmat,  &
                 trajst)
         enddo
      else
         do i = 1, npack1
            irst=49+i
            call cpvxz(psi(1+(i-1)*adim),dtpsi,adim)
            call wrrst(time,dtpsi,linwf,cdum,idum,rdum,rdum,jindx,agmat,  &
                 trajst)
         enddo
      endif

      maxdim=maxdim+1

      return
      end subroutine splitrst

!#######################################################################
! WRRSTRW
!
! writes information to restart file needed for continuation runs etc.
!
! if linwf is .true., data is for an initial wavefunction, and
!    no information about integrator is written
!
! V8.0 GW
!#######################################################################

      subroutine wrrstrw(time,psit1,psit2,prplt,prfluxt,linwf)

      implicit none

      integer(long)                                    :: d,k,f
      integer(long), external                          :: log2int

      real(dop), intent(in)                            :: time
      real(dop), dimension(totphidim), intent(in)      :: psit1,psit2
      real(dop), dimension(ndof), intent(in)           :: prplt,prfluxt
      character(len=c5)                                :: filename
      character(len=100)                               :: buffer
      logical(kind=4), intent(in)                      :: linwf
      character(len=c2)                                :: wrtag

      routine='Wrrstrw'

      if(irst.ge.50) then
         irst=irst-50   ! use filenames "rst000, rst001" etc
         write(filename,'(2a,i3.3)') name(1:laenge),'/rst',irst
         irst=50
      else
         filename=name(1:laenge)//'/restart'
      endif
      open(irst,file=filename,form='unformatted',status='unknown',err=999)

      write(irst) progver
!-----------------------------------------------------------------------
! write title
!-----------------------------------------------------------------------
      wrtag= '<runtitle>'
      write(irst) wrtag
      write(irst) runtitle

!-----------------------------------------------------------------------
! write date and time of creation
!-----------------------------------------------------------------------
      call fpstartup(buffer)
      wrtag= '<creation date>'
      write(irst) wrtag
      write(irst) buffer

      wrtag= '<end rsttitle>'
      write(irst) wrtag

!-----------------------------------------------------------------------
! write system definition
!-----------------------------------------------------------------------
      call wrdvrdef(irst)
      call wrgrddef(irst)
      call wrpsidef(irst)

!-----------------------------------------------------------------------
! write out restart information if in propagation stage
!-----------------------------------------------------------------------
      if (.not. linwf) then
         wrtag= '<start integrator info>'
         write(irst) wrtag

         wrtag= '<lrlx>'
         write(irst) wrtag
         write(irst) log2int(lrlx)
         wrtag= '<lrunprop>'
         write(irst) wrtag
         write(irst) irunprop
         wrtag= '<lrundiag>'
         write(irst) wrtag
         write(irst) log2int(lrundiag)
         wrtag= '<out1>'
         write(irst) wrtag
         write(irst) out1
         wrtag= '<out2>'
         write(irst) wrtag
         write(irst) out2

         wrtag= '<lexact>'
         write(irst) wrtag
         write(irst) log2int(lexact)
         wrtag= '<lrealwp>'
         write(irst) wrtag
         write(irst) log2int(lrealwp)

         wrtag= '<normwav>'
         write(irst) wrtag
         write(irst) normwav
         wrtag= '<timestep>'
         write(irst) wrtag
         write(irst) timestep
         wrtag= '<energsav>'
         write(irst) wrtag
         write(irst) energsav
         wrtag= '<escatsav>'
         write(irst) wrtag
         write(irst) escatsav
         wrtag= '<aunds>'
         write(irst) wrtag
         write(irst) aunds
         wrtag= '<bunds>'
         write(irst) wrtag
         write(irst) bunds
         wrtag= '<pesmax>'
         write(irst) wrtag
         write(irst) pesmax
         wrtag= '<fluxsave>'
         write(irst) wrtag
         write(irst) fluxsave
         wrtag= '<prne>'
         write(irst) wrtag
         write(irst) prne
         wrtag= '<ntimfluxo>'
         write(irst) wrtag
         write(irst) ntimfluxo
         wrtag= '<femax>'
         write(irst) wrtag
         write(irst) femax
         wrtag= '<femin>'
         write(irst) wrtag
         write(irst) femin
         wrtag= '<inimom>'
         write(irst) wrtag
         write(irst) inimom
         wrtag= '<spread>'
         write(irst) wrtag
         write(irst) spread
         wrtag= '<macheps>'
         write(irst) wrtag
         write(irst) macheps
         wrtag= '<epsmat>'
         write(irst) wrtag
         write(irst) epsmat
         wrtag= '<lreal>'
         write(irst) wrtag
         write(irst) log2int(lreal)
         wrtag= '<lrreal>'
         write(irst) wrtag
         write(irst) log2int(lrreal)
         wrtag= '<estart>'
         write(irst) wrtag
         write(irst) estart
         wrtag= '<ktot>'
         write(irst) wrtag
         write(irst) ktot
         wrtag= '<autoabs>'
         write(irst) wrtag
         write(irst) (autoabs(k),k=1,ktot)
         wrtag= '<autoflux>'
         write(irst) wrtag
         write(irst) (autoflux(k),k=1,ktot)

         wrtag= '<end integrator info>'
         write(irst) wrtag

      else
         wrtag= '<no integrator info>'
         write(irst) wrtag

      endif

!-----------------------------------------------------------------------
! write out wavefunction and probability
!-----------------------------------------------------------------------
      wrtag= '<time>'
      write(irst) wrtag
      write(irst) time
      wrtag= '<psi>'
      write(irst) wrtag
      write (irst) (psit1(d),d=1,totphidim)
      write (irst) (psit2(d),d=1,totphidim)
      if (.not. linwf) then
         if (count(autoabs > 0) .gt. 0) then
            wrtag= '<prplt>'
            write(irst) wrtag
            do k=1,ktot
               if (autoabs(k) > 0) then
                  f = autoabs(k)
                  write (irst) prplt(f)
               endif
            enddo

            wrtag= '<prflux>'
            write(irst) wrtag
            do k=1,ktot
               if (autoflux(k) > 0) then
                  f = autoflux(k)
                  write (irst) prfluxt(f)
               endif
            enddo
         endif
      endif

      close(irst)

      return
  999 continue
      write(message,'(a,f8.3,a)') &
         'Error opening restart file at t = ',time/fs,'fs'
      call errormsg
      end subroutine wrrstrw

!#######################################################################
! RDRSTRW
!
! reads restart file 
!
! before calling rdrst set the logicals
!     chkdvr
!     chkgrd
!     chkpsi
!     chkprp
!
! also set
!    iflag = 0 read wavefunction only
!          = 1 read wavefunction and lanczos vectors (if appropriate)
!
! if linwf is .true., data is for an initial wavefunction, and
!    no information about integrator or lanczos vetcors is present 
!    in file
!
! V8.0 GW
!#######################################################################

      subroutine rdrstrw(time,psit1,psit2,prplt,prfluxt,linwf,chkdvr, &
                         chkgrd,chkpsi,chkprp)

      implicit none

      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout)                     :: chkgrd
      real(dop), intent(out)                           :: time
      real(dop), dimension(totphidim), intent(out)     :: psit1,psit2
      real(dop), dimension(ndof), intent(out)          :: prplt,prfluxt
      logical(kind=4), intent(out)                     :: linwf

      rewind(irst)
      read(irst) filever(irst)

      if (filever(irst) .ge. 11.0_dop) then
         call rdrstrw11(time,psit1,psit2,prplt,prfluxt,chkdvr, &
                         chkgrd,chkpsi,chkprp)
      else
         call rdrstrw_pre11(time,psit1,psit2,prplt,prfluxt,linwf,chkdvr, &
                         chkgrd,chkpsi,chkprp)
      endif

      return
      end subroutine rdrstrw

!#######################################################################

      subroutine rdrstrw11(time,psit1,psit2,prplt,prfluxt,chkdvr, &
                         chkgrd,chkpsi,chkprp)

      implicit none

      integer(long)                                    :: d,k,f,ilbl,jlbl
      integer(long), intent(inout)                     :: chkgrd
      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp

      real(dop), intent(out)                           :: time
      real(dop), dimension(totphidim), intent(out)     :: psit1,psit2
      real(dop), dimension(ndof), intent(out)          :: prplt,prfluxt
      character(len=c1)                                :: file
      logical(kind=4)                                  :: lerr
      character(len=c2)                                :: rdtag

      routine='Rdrstrw'
      file='Restart'
      call rstinforw11(lerr,chkdvr,chkgrd,chkpsi,chkprp)

      routine='Rdrstrw'
      if (lerr) then
         message='Error reading RESTART file'
         call errormsg
      endif

      rewind(irst)
      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
         if (rdtag(ilbl:jlbl) .eq. 'time') then
            read (irst,err=999) time

         else if (rdtag(ilbl:jlbl) .eq. 'psi') then
            read (irst,err=999) (psit1(d),d=1,totphidim)
            read (irst,err=999) (psit2(d),d=1,totphidim)

         else if (rdtag(ilbl:jlbl) .eq. 'prplt') then
            do k=1,ktot
               if (autoabs(k) > 0) then
                  f = autoabs(k)
                  read (irst,err=999) prplt(f)
               endif
            enddo

         else if (rdtag(ilbl:jlbl) .eq. 'prflux') then
            do k=1,ktot
               if (autoflux(k) > 0) then
                  f = autoflux(k)
                  read (irst,err=999) prfluxt(f)
               endif
            enddo

!-----------------------------------------------------------------------
! end of file
!-----------------------------------------------------------------------
         else if (rdtag(ilbl:jlbl) .eq. 'end restart file') then
            exit

         endif

      enddo

  100 continue

      close(irst)
      return

  999 continue
      routine='Rdrst11'
      message='Error reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rdrstrw11

!#######################################################################

      subroutine rdrstrw_pre11(time,psit1,psit2,prplt,prfluxt,linwf,chkdvr, &
                         chkgrd,chkpsi,chkprp)

      implicit none

      integer(long)                                    :: d,ilbl,k,f
      integer(long), intent(inout)                     :: chkgrd
      integer(long), intent(in)                        :: chkdvr,chkpsi,chkprp

      real(dop), intent(out)                           :: time
      real(dop), dimension(totphidim), intent(out)     :: psit1,psit2
      real(dop), dimension(ndof), intent(out)          :: prplt,prfluxt
      logical(kind=4), intent(out)                     :: linwf
      character(len=c1)                                :: file
      logical(kind=4)                                  :: lerr

      routine='Rdrstrw'
      file='Restart'
      call rstinforw_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      routine='Rdrstrw'
      if (lerr) then
         message='Error reading RESTART file'
         call errormsg
      endif

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
      read (irst,err=999) time
!      write (*,*) 'Starting propagation at point ',time+1.0_dop
      read (irst,err=999) (psit1(d),d=1,totphidim)
      read (irst,err=999) (psit2(d),d=1,totphidim)

      if (.not. linwf) then
         do k=1,ktot
            if (autoabs(k) > 0) then
               f = autoabs(k)
               read (irst,err=999) prplt(f)
            endif
         enddo

         do k=1,ktot
            if (autoflux(k) > 0) then
               f = autoflux(k)
               read (irst,err=999) prfluxt(f)
            endif
         enddo
      endif

      close(irst)
      return

  999 continue
      ilbl=index(file,' ')-1
      message='Error reading '//file(1:ilbl)//' file'
      call errormsg

      end subroutine rdrstrw_pre11

!#######################################################################
! RSTINFORW
!
! reads first part of restart file, which contains information on
! size of wavefunction and related parameters.
!
! V8.0 GW
!#######################################################################

      subroutine rstinforw(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long), intent(in)    :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout) :: chkgrd
      logical(kind=4), intent(out) :: lerr,linwf

      rewind(irst)
      read(irst) filever(irst)

      if (filever(irst) .lt. 11.00_dop) then
         call rstinforw_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
      else
         call rstinforw11(lerr,chkdvr,chkgrd,chkpsi,chkprp)
      endif

      return

      end subroutine rstinforw

!#######################################################################

      subroutine rstinforw11(lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long) :: ndof1,ilrlx,ilexact,ilrunprop,ilrundiag,k,nmode1,&
                       ilreal,ilrreal,ierr,ilrealwp
      integer(long) :: ilbl,jlbl

!      integer(long), dimension(maxkoe)        :: ilrotvec
      integer(long), intent(in)                :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout)             :: chkgrd

      integer(long), dimension(maxdim)         :: fdvr,mdvr,mmode,nspfdof1
      integer(long), dimension(maxdim,maxdim)  :: spfdof1

      character(len=c4)                        :: file
      character(len=c5)                        :: title
      character(len=100)                       :: buffer
      logical(kind=4), external                :: int2log
      logical(kind=4), intent(out)             :: lerr
      character(len=c2)                        :: rdtag

!-----------------------------------------------------------------------
! Title
!-----------------------------------------------------------------------
      rewind(irst)
      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1
         if (rdtag(ilbl:jlbl) .eq. 'runtitle') then
            read(irst,err=999) title

         else if (rdtag(ilbl:jlbl) .eq. 'creation date') then
            read(irst,err=999) buffer

         else if (rdtag(ilbl:jlbl) .eq. 'end rsttitle') then
            exit

         endif

      enddo

  100 continue

      if (lwrlog) then
         write(ilog,'(/,a)') &
              '**** RSTINFO: Reading RESTART file ****'
         if(title .ne. ' ') write(ilog,'(a80)') title
         write(ilog,'(a)') 'Real WP '
      end if

!-----------------------------------------------------------------------
! Read DVR, GRD and PSI data
!-----------------------------------------------------------------------
      call zerovxi(chkerr,nerr)
      call rddvrdef(irst,chkdvr,ndof1,fdvr)
      call rdgrddef(irst,chkgrd,fdvr,mdvr,mmode,nspfdof1,spfdof1,nmode1)
      if(chkdvr.eq.6) then
         chkerr(2)=0
         chkerr(5)=0
      endif
      call rdpsidef(irst,chkpsi)

      file='Restart'
      call chksyserr(file,lerr)
!.....Ignore different basis if lignore=.true.
      if(lignore.and.lerr) then
         lerr=.false.
         write(6,'(/a,/a)') ' #############  WARNING  ##############', &
                     'Different primitive bases: system/restart-file'
         write(ilog,'(/a,/a)') ' ############  WARNING  #############', &
                     'Different primitive bases: system/restart-file'
      endif
      if (lerr) return
      routine='Rstinfo'

!     Avoid floading the efield-logfile when doing optimal control.
!     This has been added because the restart file is re-opened
!     for every time step by the efield routine.
!     Shifting the file pointer back and wait
!     for update of the restart file does not always work
!     since the file pointer might become invalid.

!-----------------------------------------------------------------------
! read restart information 
!-----------------------------------------------------------------------
      ierr=2
      if (chkprp .eq. 1) then
         rewind(irst)

         do
   20       continue
            read(irst,err=20,end=200) rdtag
            if (rdtag(1:1) .ne. '<') cycle

            ilbl=2
            jlbl=index(rdtag,'>')-1
            if (rdtag(ilbl:jlbl) .eq. 'no integrator info') then
               exit

            else if (rdtag(ilbl:jlbl) .eq. 'lrlx') then
               read (irst,err=999) ilrlx
               lrlx = int2log(ilrlx)
            else if (rdtag(ilbl:jlbl) .eq. 'lrunprop') then
               read (irst,err=999) ilrunprop
       !        lrunprop = int2log(ilrunprop)
                irunprop = ilrunprop
            else if (rdtag(ilbl:jlbl) .eq. 'lrundiag') then
               read (irst,err=999) ilrundiag
               lrundiag = int2log(ilrundiag)
            else if (rdtag(ilbl:jlbl) .eq. 'out1') then
               read (irst,err=999) out1
            else if (rdtag(ilbl:jlbl) .eq. 'out2') then
               read (irst,err=999) out2

            else if (rdtag(ilbl:jlbl) .eq. 'lexact') then
               read (irst,err=999) ilexact
               lexact = int2log(ilexact)
            else if (rdtag(ilbl:jlbl) .eq. 'lrealwp') then
               read (irst,err=999) ilrealwp
               lrealwp= int2log(ilrealwp)

            else if (rdtag(ilbl:jlbl) .eq. 'normwav') then
               read (irst,err=999) normwav
            else if (rdtag(ilbl:jlbl) .eq. 'timestep') then
               read (irst,err=999) timestep
            else if (rdtag(ilbl:jlbl) .eq. 'energsav') then
               read (irst,err=999) energsav
            else if (rdtag(ilbl:jlbl) .eq. 'escatsav') then
               read (irst,err=999) escatsav
            else if (rdtag(ilbl:jlbl) .eq. 'escatsav') then
               read (irst,err=999) aunds,bunds,pesmax
            else if (rdtag(ilbl:jlbl) .eq. 'aunds') then
               read (irst,err=999) aunds
            else if (rdtag(ilbl:jlbl) .eq. 'bunds') then
               read (irst,err=999) bunds
            else if (rdtag(ilbl:jlbl) .eq. 'pesmax') then
               read (irst,err=999) pesmax
            else if (rdtag(ilbl:jlbl) .eq. 'fluxsave') then
               read (irst,err=999) fluxsave
            else if (rdtag(ilbl:jlbl) .eq. 'prne') then
               read (irst,err=999) prne
            else if (rdtag(ilbl:jlbl) .eq. 'ntimfluxo') then
               read (irst,err=999) ntimfluxo
            else if (rdtag(ilbl:jlbl) .eq. 'femax') then
               read (irst,err=999) femax
            else if (rdtag(ilbl:jlbl) .eq. 'femin') then
               read (irst,err=999) femin
            else if (rdtag(ilbl:jlbl) .eq. 'inimom') then
               read (irst,err=999) inimom
            else if (rdtag(ilbl:jlbl) .eq. 'spread') then
               read (irst,err=999) spread
            else if (rdtag(ilbl:jlbl) .eq. 'macheps') then
               read (irst,err=999) macheps
            else if (rdtag(ilbl:jlbl) .eq. 'epsmat') then
               read (irst,err=999) epsmat
            else if (rdtag(ilbl:jlbl) .eq. 'lreal') then
               read(irst,err=999) ilreal
               lreal  = int2log(ilreal)
            else if (rdtag(ilbl:jlbl) .eq. 'lrreal') then
               read(irst,err=999) ilrreal
               lrreal = int2log(ilrreal)
            else if (rdtag(ilbl:jlbl) .eq. 'estart') then
               read (irst,err=999) estart
            else if (rdtag(ilbl:jlbl) .eq. 'ktot') then
               read(irst,err=999) ktot,(autoabs(k),k=1,ktot)
            else if (rdtag(ilbl:jlbl) .eq. 'autoabs') then
               read(irst,err=999) (autoabs(k),k=1,ktot)
            else if (rdtag(ilbl:jlbl) .eq. 'autoflux') then
               read(irst,err=999) (autoflux(k),k=1,ktot)

            endif

         enddo
  200    continue

      endif

      return
  999 continue
      routine='Rstinforw11'
      message='Error Reading Restart file: '//rdtag(ilbl:jlbl)
      call errormsg

      end subroutine rstinforw11

!#######################################################################

      subroutine rstinforw_pre11(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

      implicit none

      integer(long) :: ndof1,ilinwf,ilrlx,ilexact,&
                       ilrunprop,ilrundiag,k,nmode1,&
                       ilreal,ilrreal,ierr,iver,ilrealwp

!      integer(long), dimension(maxkoe)         :: ilrotvec
      integer(long), intent(in)                :: chkdvr,chkpsi,chkprp
      integer(long), intent(inout)             :: chkgrd

      integer(long), dimension(maxdim)         :: fdvr,mdvr,mmode,nspfdof1
      integer(long), dimension(maxdim,maxdim)  :: spfdof1

      real(dop)                                :: rdum
      character(len=c4)                        :: file
      character(len=c5)                        :: title
      character(len=100)                       :: buffer
      logical(kind=4), external                :: int2log
      logical(kind=4), intent(out)             :: lerr,linwf

      iver=min(irst,maxfver)

      call zerovxi(chkerr,nerr)
      call rddvrdef(irst,chkdvr,ndof1,fdvr)
      call rdgrddef(irst,chkgrd,fdvr,mdvr,mmode,nspfdof1,spfdof1,nmode1)
      if(chkdvr.eq.6) then 
         chkerr(2)=0
         chkerr(5)=0
      endif
      call rdpsidef(irst,chkpsi)

      file='Restart'
      call chksyserr(file,lerr)
!.....Ignore different basis if lignore=.true. 
      if(lignore.and.lerr) then
         lerr=.false.
         write(6,'(/a,/a)') ' #############  WARNING  ##############', &
                     'Different primitive bases: system/restart-file'
         write(ilog,'(/a,/a)') ' ############  WARNING  #############', &
                     'Different primitive bases: system/restart-file'
      endif
      if (lerr) return
      routine='Rstinforw'

      ierr=1
      read(irst,err=999) title
      read(irst,err=999) buffer
      
!     Avoid floading the efield-logfile when doing optimal control.   
!     This has been added because the restart file is re-opened 
!     for every time step by the efield routine.
!     Shifting the file pointer back and wait 
!     for update of the restart file does not always work
!     since the file pointer might become invalid. 

      if (lwrlog) then
         write(ilog,'(/,a)') &
              '**** RSTINFO: Reading RESTART file ****'
         if(title .ne. ' ') write(ilog,'(a80)') title
         if (dentype.eq.0) then
            write(ilog,'(a,f9.5)') 'Type : wavefunction, file-version =', &
                                 filever(iver)
         else if (dentype.eq.1) then
            write(ilog,'(a)') 'Type : density operator I '
         else if (dentype.eq.2) then
            write(ilog,'(a)') 'Type : density operator II'
         endif
         write(ilog,'(a)') 'Created on:'
         write(ilog,'(a)') buffer
      end if

!-----------------------------------------------------------------------
! read restart information 
!-----------------------------------------------------------------------
      ierr=2
      read(irst,err=999) ilinwf
      linwf = int2log(ilinwf)
      if ((.not. linwf) .and. chkprp .eq. 1) then
         ierr=3
         if (filever(iver) .ge. 8.1005) then
            read (irst,err=999) ilrlx,ilrunprop,ilrundiag
            lrlx = int2log(ilrlx)
            !lrunprop = int2log(ilrunprop)
            irunprop = ilrunprop
            lrundiag = int2log(ilrundiag)
         else
            read (irst,err=999) ilrlx
            lrlx = int2log(ilrlx)
         endif
         ierr=4
         read (irst,err=999) out1,out2
         read (irst,err=999) ilexact,ilrealwp
         lexact = int2log(ilexact)
         lrealwp= int2log(ilrealwp)
         ierr=5
         ierr=6
         read (irst,err=999) normwav,timestep,energsav,escatsav
         write (*,*) 'The initial norm of the wave packet was',normwav
         write (*,*) 'The (initial) energy was ',energsav
         read (irst,err=999) aunds,bunds,pesmax
         read (irst,err=999) fluxsave,prne,ntimfluxo
         read (irst,err=999) femax,femin,inimom,spread
         if (filever(iver) .ge. 9.0015_dop) then
            read (irst,err=999) macheps,epsmat
         else
            read (irst,err=999) macheps,rdum,epsmat
         endif
         
         ierr=7
         if(filever(iver) .ge. 8.3012_dop) then
            read(irst,err=999) ilreal,ilrreal
            lreal  = int2log(ilreal)
            lrreal = int2log(ilrreal)
         endif
         ierr=8
         read (irst,err=999) estart
         ierr=9
         if (filever(iver) .ge. 8.200_dop) then
            read(irst,err=999) ktot,(autoabs(k),k=1,ktot)
            read(irst,err=999) (autoflux(k),k=1,ktot)
         endif
      endif

      return
  999 continue
      routine='rstinforw_pre11 (iorstrw.f90)'
      write(message,'(a,i4)') &
          'Error reading RESTART file (RSTINFO). Error No.',ierr
      call errormsg
      end subroutine rstinforw_pre11

!#######################################################################
!
! Basic check that restart file is appropriate for a calculation,
! E.g. psitype and dentype are the same as those defined in input
!
!#######################################################################

      subroutine chkrsttyp

      implicit none

      integer(long)         :: ilbl,jlbl
      integer(long)         :: psitype1,dentype1
      character(len=c2)     :: rdtag
      logical(kind=4)       :: lerr = .false.

      routine='Chkrsttyp'

      rewind(irst)
      read(irst) filever(irst)

! Checks only work for versions > 11.0
      if (filever(irst) .lt. 11.0d0) return

      do
  10     continue
         read(irst,iostat=chkerr(1),err=10,end=100) rdtag
         if (rdtag(1:1) .ne. '<') cycle

         ilbl=2
         jlbl=index(rdtag,'>')-1

!-----------------------------------------------------------------------
! read wavefunction
!-----------------------------------------------------------------------
         if (rdtag(ilbl:jlbl) .eq. 'psitype') then
            read (irst,err=999) psitype1
         else if (rdtag(ilbl:jlbl) .eq. 'dentype') then
            read (irst,err=999) dentype1
         endif
      enddo

  100 continue
      if (psitype1 .eq. 0 .and. psitype .eq. 5) then
         continue
      else if (psitype1 .eq. 5 .and. psitype .eq. 0) then
         continue
      else if (psitype1 .ne. psitype) then
         message = 'Psi type in restart file differs from system '
         call warnmsg
         lerr = .true.
      endif
      if (dentype1 .ne. dentype) then
         message = 'Density type in restart file differs from system '
         call warnmsg
         lerr = .true.
      endif

      if (lerr) then
         message = 'Wrong restart file for calculation.'
         call errormsg
      endif

      return
  999 continue
      message='Error reading Restart file '
      call errormsg

      end subroutine chkrsttyp

!************************************************************************

      end module iorst



