! **********************************************************************
!                                                                       
!                    module PHIHPHI
!                                                                       
!   MCTDH module containing (auxiliary) subroutines to evaluate the     
!   matrix elements of an operator in a spf basis
!                                                                       
!   Contains:                                                           
!     phihphi:      Computes for all modes the "correlated"             
!                   matrix elements <phi|h|phi>.                        
!     phihphi1m:    Computes for one mode the "correlated"              
!                   matrix elements <phi|h|phi>.                        
!     phihunphi:    Computes for all modes and states the               
!                   "uncorrelated" matrix elements <phi|h|phi>.         
!     dwtphihphi1m: Computes the loop over the kterms in parallel.
!                                                                       
! **********************************************************************
      module phihphimod

use decimal, only: dop,long
use global
use logdat
use d2tens,  only: mkherm, mkaherm
use mmlib,   only: mmaxzz, mmaxzzh, mmaxzzh1
use op1lib,  only: cpmxz, cpqxz, zeromxz,zerovxz,cpvxz
use timing

use mltypes
use linear
use operdef

#ifdef MPI
use mpi
use mpiutilsmod, only: mpe_decomp1d
#endif
use mpidata
use openmpmod
use dvrdatmod
use griddatmod
use psidef
use hpsimod
use operdef

use xlocphihphimod, only : xlocphihphi
use hphimod, only: hphi1mk
use xlocphihphimod, only : xlocphihphi
use op2lib, only: addmxxzo

      implicit none
      private
      public :: phihphi,phihunphi
#ifdef MPI
      public :: mpiphihphi
#endif

      contains

! *** V6.0 GW **************> SUBROUTINE PHIHPHIPAR <*******************
!                            ~~~~~~~~~~~~~~~~~~~~

! Computes for all modes the correlated Hamiltonian matrix elements
! <phi|h|phi>.
!
! nham : which operator is being used.
! lsym : if true use symmetries of operators in building matrices.
!
! NB dentype 2 operations have been moved to phihphi - to be checked,
! see d2phihphi below
!
! For CMF (psitype = 0/5) all moments of hteil need to be stored. For
! VMF only the zero moment is needed.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine phihphi (psi,gh,hpsi,hteil,nham,lsym)

      implicit none

      integer(long)             :: k,m,s,s1,k1,k2,f,h
      integer(long), intent(in) :: nham
      integer(long), save       :: tid=0
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(hpsidim), intent(in)   :: hpsi
      complex(dop), dimension(pmatdim), intent(out)  :: hteil
      complex(dop), dimension(ghmatdim), intent(in)  :: gh
      logical(kind=4), intent(in)                    :: lsym

      call get_timer(tid, 'PhiHPhi')
      call start_timer(tid)

! --- COMPUTE <PHI|H|PHI> MATRICES FOR EACH HAMILTONIAN TERM---

      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)+nmulpot(nham)-1

!-----------------------------------------------------------------------
#ifdef MPI
      if (lmpi.and.lmpiphihphi.and.lmpistarted) then
         mpitask=8
! --- Determine Job
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call mpiphihphi(psi,hpsi,hteil,nham,lsym,gh)
      else
#endif

!-----------------------------------------------------------------------
!     If no parallelisation is used then a usual calculation is 
!     performed.
!-----------------------------------------------------------------------
! --- LOOP OVER EACH MODE AND STATE ---
!$omp parallel num_threads(ompthread) if(lompthread .and. lompphihphi)
!$omp do private(m,f,k,s,s1) schedule(dynamic)
         do k = k1,k2
            s=kf(k)
            s1=ki(k)

            do m = 1,nmode
               if (k .ne. kfirst(m,k) ) cycle
               if (pmat(m,k) .eq. 0) cycle

               f = spfdof(1,m)
               if (gwpm(m)) then
                  if ((psitype .eq. 0 .or. psitype .eq. 5) .and. &
                       (klha(k) .eq. 0)) then
                     call gh2hteil(gh(ghmat(k,m)),hteil(pmat(m,k)),&
                          psi(zdgwp(m,s)),psi(zdgwp(m,s1)),&
                          dim(m,s),dim(m,s1))
                  else
                     call cpmxz(gh(ghmat(k,m)),hteil(pmat(m,k)),&
                          dim(m,s),dim(m,s1))
                  endif
               endif

               if (hsym(m,k) .eq. 0 .or. .not. lsym) then
                  if (gwpm(m)) then
                     continue
                  else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     routine='phihphi'
                     write(message,'(a,i5)') &
                          'Non-potential term on x-local DOF:',f
                     call errormsg
                  else
                     call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s),dim(m,s1))
                  endif

               else if (hsym(m,k) .eq. 1) then
                  if (gwpm(m)) then
                     call mkherm(hteil(pmat(m,k)),dim(m,s))

                  else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     h=hamilton(f,k)
                     call xlocphihphi(psi(zetf(m,1)),hteil(pmat(m,k)),h,f,&
                          3,dim(m,1))
                  else
                     call mmaxzzh(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s))
                  endif
               else if (hsym(m,k) .eq. -1) then
                  if (gwpm(m))then
                     call mkaherm(hteil(pmat(m,k)),dim(m,s))
                  else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     routine='phihphi'
                     write(message,'(a,i5)') &
                          'Imaginary term on x-local DOF:',f
                     call errormsg
                  else
                     call mmaxzzh1(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s))
                  endif
               endif
               
!     Dissipative operators
               if ( (k .lt. ktwin(k)) .and. &
                       ((kact(k) .eq. 2) .or. ((kact(k).eq.3) .and. ldproj)) ) then
                  k1 = ktwin(k)
!     Note that for si .ne. sf always diag(m,k)=.false.
                  if ((.not. diag(m,k)) .and. (.not. diag(m,k1))) then
                     s1 = ki(k1)
                     s  = kf(k)
                     call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi2(m,k)),&
                          hteil(pmat2(m,k)),subdim(m),dim(m,s),dim(m,s1))
                  endif
               endif

            enddo
         enddo
!$omp end do
!$omp end parallel 
#ifdef MPI
      endif
! just make hteil known in all MPI slaves as soon as it is 
! newly calculated so we dont have to send it in various 
! subroutines 
      if (lmpi.and.lmpistarted) then
         mpitask=22
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(hteil,pmatdim,MPI_DOUBLE_COMPLEX,master, &
              MPI_COMM_WORLD,mpierr)
      end if
#endif

      call stop_timer(tid)

      return
      end subroutine phihphi

! *** V6.0 MB *************> SUBROUTINE PHIHUNPHI <*********************
!                           ~~~~~~~~~~~~~~~~~~~~~~
! Computes for all modes and states the uncorrelated matrix elements
! <phi|h_uncorr|phi>.
! Converts and saves hung as hunteil <phi|h_uncorrX|phi>.
!
! Required by CMF integrator
!
!#######################################################################

      subroutine phihunphi (psi,hunpsi,hung,gh,hteil,nham)

      implicit none

      integer(long)                                      :: m,s,kz,k,k1,k2
      integer(long), save                                :: tid=0
      integer(long), intent(in)                          :: nham
      complex(dop), dimension(dgldim), intent(in)        :: psi
      complex(dop), dimension(adim+1:dgldim), intent(in) :: hunpsi
      complex(dop), dimension(pmatdim), intent(out)      :: hteil
      complex(dop), dimension(hungdim), intent(in)       :: hung
      complex(dop), dimension(ghmatdim), intent(in)      :: gh
      complex(dop), dimension(maxspf*maxspf)             :: ztmp

      call get_timer(tid,'PhiHunPhi')
      call start_timer(tid)

! --- LOOP OVER EACH MODE AND STATE ---

      kz=zham(nham)+kzahl(nham)+nmulpot(nham)-1

!$omp parallel num_threads(ompthread) if(lompthread .and. lompphihphi)  
!$omp do private(m,s)
      do m = 1,nmode

!        Gaussian mode :

         if(gwpm(m))then
            do s = 1,nstate
               if(khzahl(m,s,nham) .gt. 0 ) then
                  if (psitype .eq. 0 .or. psitype .eq. 5) then
                     call gh2hteil(hung(hungmat(m,s,nham)),hteil(pmat(m,kz+s)), &
                          psi(zdgwp(m,s)),psi(zdgwp(m,s)),dim(m,s),dim(m,s))
                  else
                     call cpqxz(hung(hungmat(m,s,nham)),hteil(pmat(m,kz+s)),&
                          dim(m,s))
                  endif
                  call mkherm(hteil(pmat(m,kz+s)),dim(m,s))
               endif

! Add in non-hermititan CAP terms
               if (lcap) then
                  k1=zhun(m,s,nham)
                  k2=zhun(m,s,nham)+khzahl(m,s,nham)-1
                  do k=k1,k2
                     if (autocap(k) .ge. 0) cycle

                     if (psitype .eq. 0 .or. psitype .eq. 5) then
                         call gh2hteil(gh(ghmat(k,m)),ztmp, &
                              psi(zdgwp(m,s)),psi(zdgwp(m,s)),dim(m,s),dim(m,s))
                        call addmxxzo(ztmp,hteil(pmat(m,kz+s)),&
                             gwpdim(m,s),dim(m,s))
                     else
                        call addmxxzo(gh(ghmat(k,m)),hteil(pmat(m,kz+s)),&
                             dim(m,s),dim(m,s))
                     endif
                  enddo
               endif

            enddo
         else

            do s = 1,nstate
               if(khzahl(m,s,nham) .gt. 0 ) then
                  call mmaxzz(psi(zetf(m,s)),hunpsi(zetf(m,s)),&
                     hteil(pmat(m,kz+s)),subdim(m),dim(m,s),dim(m,s))
               endif
            enddo
         endif
      enddo
!$omp enddo
!$omp end parallel 


#ifdef MPI
! just make hteil known in all MPI slaves as soon as it is 
! newly calculated so we dont have to send it in various 
! subroutines 
      if (lmpi.and.lmpistarted) then
         mpitask=22
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(hteil,pmatdim,MPI_DOUBLE_COMPLEX,master, &
              MPI_COMM_WORLD,mpierr)
      end if
#endif

      call stop_timer(tid)

      return
      end subroutine phihunphi

! *** V6.0 GW ************> SUBROUTINE D2PHIHPHI1M <********************
!                          ~~~~~~~~~~~~~~~~~~~~~~~
! Computes for one mode, the correlated Hamiltonian matrix elements
! <phi|h|phi>.
!#######################################################################

      subroutine d2phihphi1m (psi,hpsi,hteil,m,zetf1,phidim1,nham, &
                 lsym)

      implicit none

      integer(long)                                              :: k,k1,k2,k3,s,s1
      integer(long), intent(in)                                  :: m,zetf1,phidim1,nham
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(in) :: psi
      complex(dop), dimension(hpsidim), intent(in)               :: hpsi
      complex(dop), dimension(pmatdim), intent(out)              :: hteil
      logical(kind=4), intent(in)                                :: lsym

! --- COMPUTE <PHI|H|PHI> MATRICES FOR EACH HAMILTONIAN TERM---
      k2=zham(nham)
      k3=zham(nham)+kzahl(nham)+nmulpot(nham)-1

!$omp parallel num_threads(ompthread) if(lompthread .and. lompphihphi)
!$omp do private(k,k1,s,s1) schedule (dynamic)
      do k = k2,k3
         if (.not. diag(m,k) .and. k .eq. kfirst(m,k) ) then
            s=kf(k)
            s1=ki(k)
            if ( (s.eq.s1) .and. lsym .and.&
                 ( (kact(k).eq.1).or.(kact(k).eq.6) ) ) then
               call mmaxzzh(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                    hteil(pmat(m,k)),subdim(m),dim(m,s))
            else
               call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                    hteil(pmat(m,k)),subdim(m),dim(m,s),dim(m,s1))
            endif
         endif

!     Dissipative operators
         if (ldproj) then
            if ((kact(k).eq.2) .and. (k.lt.ktwin(k))) then
               k1 = ktwin(k)
               if ((.not. diag(m,k)) .and. (.not. diag(m,k1))) then
                  s1 = ki(k1)
                  s  = kf(k)
                  call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi2(m,k)),&
                       hteil(pmat2(m,k)),subdim(m),dim(m,s),dim(m,s1))
               endif
            endif
         else
            if (((kact(k).eq.2).or.(kact(k).eq.3)) .and.&
                 (k.lt.ktwin(k))) then
               k1 = ktwin(k)
!     Note that for si .ne. sf always diag(m,k)=.false.
               if ((.not. diag(m,k)) .and. (.not. diag(m,k1))) then
                  s1 = ki(k1)
                  s  = kf(k)
                  call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi2(m,k)),&
                       hteil(pmat2(m,k)),subdim(m),dim(m,s),dim(m,s1))
               endif
            endif
         endif

      enddo
!$omp end do
!$omp end parallel 

      return
      end subroutine d2phihphi1m

!-----------------------------------------------------------------------
!   mpiphihphi
!-----------------------------------------------------------------------

#ifdef MPI
      subroutine mpiphihphi(psi,hpsi,hteil,nham,lsym,gh)

      implicit none

      integer(long), intent(in)                      :: nham
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(hpsidim), intent(in)   :: hpsi
      complex(dop), dimension(ghmatdim), intent(in)  :: gh
      complex(dop), dimension(pmatdim), intent(out)  :: hteil
      logical(kind=4), intent(in)                    :: lsym
      integer(long) :: m,i1,i2,k,k1,k2,s,s1,f,h
      complex(dop), dimension(pmatdim)               :: tmphteil

! --- Get Data
      call MPI_Bcast(nham,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(lsym,1,MPI_LOGICAL,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(psi,dgldim,MPI_DOUBLE_COMPLEX,master, &
                     MPI_COMM_WORLD,mpierr)
      if (lgwp) then
         call MPI_Bcast(gh,ghmatdim,MPI_DOUBLE_COMPLEX,master, &
                        MPI_COMM_WORLD,mpierr)
      endif
      call MPI_Bcast(hpsi,hpsidim,MPI_DOUBLE_COMPLEX,master, &
                     MPI_COMM_WORLD,mpierr)

      hteil(:) = 0.0_dop

! --- COMPUTE <PHI|H|PHI> MATRICES FOR EACH HAMILTONIAN TERM---

      i1=zham(nham)
      i2=zham(nham)+kzahl(nham)+nmulpot(nham)-1

! Divide up the work.
      call mpe_decomp1d(i1,i2,k1,k2)

! --- LOOP OVER EACH MODE AND STATE ---
      do k = k1,k2
         s=kf(k)
         s1=ki(k)
         do m = 1,nmode
            if (.not. diag(m,k) .and. k .eq. kfirst(m,k) ) then
               f = spfdof(1,m)
               if (gwpm(m)) then
                  if (psitype .eq. 0 .or. psitype .eq. 5) then
                     call gh2hteil(gh(ghmat(k,m)),hteil(pmat(m,k)),&
                          psi(zdgwp(m,s)),psi(zdgwp(m,s1)),&
                          dim(m,s),dim(m,s1))
                  else
                     call cpmxz(gh(ghmat(k,m)),hteil(pmat(m,k)),&
                          dim(m,s),dim(m,s1))
                  endif
               endif

               if (hsym(m,k) .eq. 0 .or. .not. lsym) then
                  if (gwpm(m)) then
                     continue
                  else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     routine='phihphi'
                     write(message,'(a,i5)') &
                          'Non-potential term on x-local DOF:',f
                     call errormsg
                  else
                     call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s),dim(m,s1))
                  endif

               else if (hsym(m,k) .eq. 1) then
                  if (gwpm(m)) then
                     call mkherm(hteil(pmat(m,k)),dim(m,s))

                 else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     h=hamilton(f,k)
                     call xlocphihphi(psi(zetf(m,1)),hteil(pmat(m,k)),h,f,&
                          3,dim(m,1))
                  else
                     call mmaxzzh(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s))
                  endif
               else if (hsym(m,k) .eq. -1) then
                  if (gwpm(m))then
                     call mkaherm(hteil(pmat(m,k)),dim(m,s))
                  else if(basis(f).eq.26 .or. basis(f) .eq. 27)then
                     routine='phihphi'
                     write(message,'(a,i5)') &
                          'Imaginary term on x-local DOF:',f
                     call errormsg
                  else
                     call mmaxzzh1(psi(zetf(m,s)),hpsi(zhpsi(m,k)),&
                          hteil(pmat(m,k)),subdim(m),dim(m,s))
                  endif
               endif
            endif
!     Dissipative operators
            if ( (k .lt. ktwin(k)) .and. &
                    ((kact(k) .eq. 2) .or. ((kact(k).eq.3) .and. ldproj)) ) then
               k1 = ktwin(k)
!     Note that for si .ne. sf always diag(m,k)=.false.
               if ((.not. diag(m,k)) .and. (.not. diag(m,k1))) then
                  s1 = ki(k1)
                  s  = kf(k)
                  call mmaxzz(psi(zetf(m,s)),hpsi(zhpsi2(m,k)),&
                       hteil(pmat2(m,k)),subdim(m),dim(m,s),dim(m,s1))
               endif
            endif

         enddo
      enddo

! --- Reduce Data
      call MPI_Reduce(hteil,tmphteil,pmatdim,MPI_DOUBLE_COMPLEX,MPI_SUM,master, &
                                                  MPI_COMM_WORLD,mpierr)
      if(mpirank.eq.master) call cpvxz(tmphteil,hteil,pmatdim)

      return
      end subroutine mpiphihphi
#endif

!#######################################################################
!
! Convert Gh to hteil (GWP -> SPF basis)
!
!#######################################################################

      subroutine gh2hteil(gh,hteil,dgwp1,dgwp2,dim1,dim2)

      implicit none

      integer(long)                                   :: e,e1,e2,e3
      integer(long), intent(in)                       :: dim1,dim2
      complex(dop), dimension(dim1,dim2), intent(in)  :: gh
      complex(dop), dimension(dim1,dim2), intent(out) :: hteil
      complex(dop), dimension(dim1,dim1), intent(in)  :: dgwp1
      complex(dop), dimension(dim2,dim2), intent(in)  :: dgwp2
      complex(dop), dimension(dim1,dim2)              :: ztmp

      hteil(:,:)=0.0_dop
      ztmp(:,:)=0.0_dop
      do e3=1,dim2
         do e1=1,dim2
            do e2=1,dim1
               ztmp(e2,e1) = ztmp(e2,e1)+ gh(e2,e3)*dgwp2(e3,e1)
            enddo
         enddo
      enddo
      do e2=1,dim1
         do e1=1,dim2
            do e=1,dim1
               hteil(e,e1) = hteil(e,e1) + dconjg(dgwp1(e2,e))*ztmp(e2,e1)
            enddo
         enddo
      enddo

      return
      end subroutine gh2hteil

!#######################################################################

      end module phihphimod
