!#######################################################################
!                    calctdk  (mctdhlib/calctdk.F)
!#######################################################################

      module tdkoeffs

use decimal, only: long, dop
use global
use globmemmod
use logdat

use dvrdatmod
use griddatmod
use operdef
use psidef
use runinwfmod
use lctmod
use datenmod
use runpropmod

use potevalmod, only: poteval,poteval1d
use expectval, only: exexpectvalue,expectvalue

      implicit none
      private
      public :: calctdk,calctdp,calctdp1,koeff_ml

      contains

!#######################################################################
!
! Calculate time-dependent coefficients
!
! Only possible during
! propagation so are not needed for calls from e.g. analysis routines.
!
! Any changes not involving lct may be needed in calctdk_nolct (expectval)
! as well.
!#######################################################################

      subroutine calctdk(time,nham,psi)

      implicit none

      integer(long)                               :: h,k,kl,kh,idum
      integer(long), intent(in)                   :: nham
      real(dop)                                   :: c
      real(dop), dimension(maxdim)                :: rdum
      real(dop), intent(in)                       :: time
      real(dop), dimension(maxdim)                :: gpoint
      complex(dop)                                :: cz
      complex(dop), dimension(dgldim), intent(inout) :: psi


      kl=zham(nham)
      kh=zham(nham)+khtot(nham)-1

      do k=kl,kh
         h=hamilton(fkoe,k)
         if (htmtype(h) .eq. 101) then
            gpoint(fkoe)=1.0_dop
            gpoint(ftime)=time
            call poteval(h,c,gpoint,cz,rdum,idum)
            koeff(k) = cz
         else if (htmtype(h) .eq. 151) then
            gpoint(fkoe)=1.0_dop
            gpoint(ftime)=time
            call funclct(h,gpoint,psi,cz)
!            vfield=dble(cz)
            koeff(k) = cz
         endif
      enddo

      return
      end subroutine calctdk

!#######################################################################
!
! multilayer driver for calctdk
!
!#######################################################################

    subroutine koeff_ml(k,time,l,psi)
        implicit none
        integer,intent(in)          :: k
        real(dbl),intent(in)        :: time
        integer,intent(in)          :: l
        complex(cdbl),intent(inout) :: psi(l)
        integer                     :: ktype,h,nham
        complex(cdbl)               :: cz

        h=hamilton(fkoe,k) ! Inquire type of coefficient
        ktype=htmtype(h)

        if(ktype .eq. 101) then  ! TD coefficient
           nham=1
           call calctdk(time,nham,psi)
        end if

        cz = koeff(k)        ! Read koeff value

        if(ktype/=1) then
            psi = psi * cz
        end if
    end subroutine koeff_ml

!#######################################################################
!
! Calculates the single particle function for a driven (TDpoint) basis 
! set. Finds relevant DOFs and passes arrays down to calctdp1
!
! variables passed for one mode:
!  time: time
!  ort: value of grid coordinates
!
! returned: 
!  psi: a delta function on a single point determined by the value of 
!       the function nham
!#######################################################################

      subroutine calctdp(time,psi,ort)

      implicit none

      integer(long)                                :: nham
      integer(long)                                :: f,m,s
      real(dop), intent(in)                        :: time
      real(dop), dimension(ortdim), intent(in)     :: ort
      complex(dop), dimension(dgldim), intent(out) :: psi

      do f=1,ndof
         if (sbasis(f) .ne. 11) cycle

         m=dofspf(f)
         nham = isbaspar(1,f,1)
         do s=1,nstate
            call calctdp1(time,s,nham,psi(zetf(m,s)),ort(zort(f)),&
                 gdim(f))
         enddo
      enddo

      end subroutine calctdp

!#######################################################################
!
! Calculates the single particle function for a driven (TDpoint) basis 
! set.
!
! variables passed for one mode:
!  time: time
!  nham: number of function to be evaluated
!  gdim: No. of grid points
!  ort: value of grid coordinates
!
! returned: 
!  psi: a delta function on a single point determined by the value of 
!       the function nham
!#######################################################################

      subroutine calctdp1(time,s1,nham,psi,ort,gdim1)

      implicit none

      integer(long), intent(in)                   :: nham,gdim1,s1
      integer(long)                               :: h,k,kl,kh,idum,g,g1
      real(dop), intent(in)                       :: time
      real(dop), dimension(gdim1), intent(in)     :: ort
      real(dop), dimension(maxdim)                :: rdum
      real(dop)                                   :: c
      real(dop)                                   :: fval,diff,mindiff
      real(dop), dimension(maxdim)                :: gpoint
      complex(dop), dimension(gdim1), intent(out) :: psi
      complex(dop)                                :: cz

      psi(:) = 0.0_dop

!-----------------------------------------------------------------------
! fval is the sum of terms and for a simple (real) TD function returns the
! total function value at time
!-----------------------------------------------------------------------
      fval = 0.0_dop
      if (.not. lrlx) then
         kl=zham(nham)
         kh=zham(nham)+khtot(nham)-1

         do k=kl,kh
            if (ki(k) .ne. s1) cycle

            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 101) then
               gpoint(fkoe)=1.0_dop
               gpoint(ftime)=time
               call poteval(h,c,gpoint,cz,rdum,idum)
               fval = fval + dble(c)
            endif
         enddo
      endif

!-----------------------------------------------------------------------
! find nearest point to function value
!-----------------------------------------------------------------------
      g1=0
      mindiff = 99999.0
      do g=1,gdim1
         diff = abs(ort(g) - fval)
         if (diff .lt. mindiff) then
            mindiff = diff
            g1=g
         endif
      enddo

!-----------------------------------------------------------------------
! psi is a delta function on point g1
!-----------------------------------------------------------------------
      psi(g1) = 1.0_dop

      return
      end subroutine calctdp1

!#######################################################################
! Local control functions. Called from main calctdk.
!
! the local control field is stored in vfield for writing to
! field.lct file during output. This routine may be called
! a number of times if the field interacts with different terms
! in the Hamiltonian but each time the field should be the same.
! However it may calculate a zero field if a step function is used 
! to toggle the control. Hence save any non-zero field.
! 
!#######################################################################

      subroutine funclct(h,x,psi,vz)

      implicit none

      integer(long)                               :: ifunc,hl,ifile,f,n,h1,&
                                                     pottyp
      integer(long), intent(in)                   :: h
      real(dop)                                   :: time,rfac
      real(dop), dimension(nopdof), intent(in)    :: x
      complex(dop), intent(out)                   :: vz
      complex(dop), dimension(dgldim), intent(inout):: psi
      complex(dop)                                :: zfac,zterm
      real(dop), dimension(maxdim)                :: gpoint,ppoint

      time=x(ftime)

! find file containing operator or calculate coefficient.
      zterm=1.0_dop
      h1=0
      do n=1,htmfac(h)
         hl=hterm(n,h)
         f=htmdof(hl)
         if (f .eq. ftime) then
            h1=hterm(1,hl)
            ifile=hopfile(h1)
            ifunc=hopilab(h1)
         else if (f .eq. fkoe) then
            pottyp=0
            gpoint(fkoe)=1.0_dop
            rfac=0.0_dop
            zfac=0.0_dop
            call poteval1d(hl,rfac,gpoint,zfac,ppoint,pottyp)
            zterm=zterm*zfac
         endif
      enddo
      iofunc=ifunc

! go to selected ifunc function
     select case (ifunc)

     case default
        routine='funclct (opfuncs/funclct.F)'
        write(message,'(a,i5)') 'Goto Error, Function not found :',ifunc
        call errormsg

     case (1)! rdfield.lct (main part in propwf/output.F)
        vz=vfield
       vz = zterm*vz
        return

     case (2)
        call lct_elec(vz,time) 
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
        return

     case (3)
        call lct_vibr(vz,time)
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
        return

     case (4)
        call lct_nucl(vz,time)
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
        return

     case (5)
       call lct(vz,time,h1,hl,psi)
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
       return

     case (6)
       call lcttd(vz,time,h1,psi)
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
       return

     case (7)
       call lct_2d(vz,time,h1,psi)
       if (abs(vz) .gt. tiny) vfield=dble(vz)
       vz = zterm*vz
       return

! Add more functions here
     end select

! multiply by coefficient

      end subroutine funclct

!#######################################################################
! Calculates the pulse for maximising electronic transfer 

      subroutine lct_elec(vz,time)

      implicit none

      real(dop), intent(in)                    :: time
      real(dop), dimension(2)                  :: lambda
      complex(dop), intent(out)                :: vz

      if (lctvibpop .eqv. .true.) then
        print*,"Wrong LCT routine- you need lct_vib in .op."
        stop
      endif

      vz=0.0_dop
      lambda(1)=hoppar(1,1)
      lambda(2)=hoppar(2,1)
! Switching
      tog=1
      if (ltoggle .eqv. .true.) then
         if (time .gt. switch(nmb)) then
            nmb=nmb+1
            if (tog .eq. 2) then
              tog=1
            else
              tog=2
            endif
         endif
      endif
! Creating pulse
      vz=lambda(tog)*imag(expvalue(tog))
      if (time .lt. 0.0) then
         vz=0.01
      endif
! Smoothing Pulse
      if (lgpulse .eqv. .false.) then
         if (time .gt. 0.0_dop) then
            call gausspulse(time,vz)
         endif
      endif

      return
      end subroutine lct_elec

!#######################################################################
! Calculates the pulses for heating and cooling within vibrational levels

      subroutine lct_vibr(vz,time)

      implicit none

      real(dop), intent(in)                    :: time
      real(dop), dimension(2)                  :: lambda
      complex(dop), intent(out)                :: vz

      vz=0.0_dop
      lambda(1)=hoppar(1,1)
      lambda(2)=hoppar(2,1)
      tog=1
      vz=lambda(tog)*imag(expvalue(tog))
      if (time .lt. 0.0) then
         vz=0.0001
      endif
      return
      end subroutine lct_vibr

!#######################################################################
! Calculates the pulses for controlling momentum of wavepacket 

      subroutine lct_nucl(vz,time)

      implicit none

      real(dop), intent(in)                    :: time
      real(dop), dimension(2)                  :: lambda
      complex(dop), intent(out)                :: vz

      if (lctvibpop .eqv. .true.) then
        print*,"Wrong LCT routine- you need lct_vib in .op."
        stop
      endif

      vz=0.0_dop
      lambda(1)=hoppar(1,1)
! Switching
      if (ltoggle .eqv. .true.) then
         if (time .gt. switch(nmb)) then
            nmb=nmb+1
            if (tog .eq. 2) then
              tog=1
            else
              tog=2
            endif
         endif
      endif
! Creating pulse
!     vz=lambda(tog)*expvalue(tog+1)*(1-(expvalue(1)/targ))
      vz=lambda(1)*expvalue(1)
      if (time .lt. 0.0) then
         vz=0.01
      endif
! Smoothing Pulse
      if (lgpulse .eqv. .false.) then
         if (time .gt. 0.0_dop) then
            call gausspulse(time,vz)
         endif
      endif

      return
      end subroutine lct_nucl

!#######################################################################
! Ensures that calculated pulse is physical and reproducable

      subroutine gausspulse(time,vz)

      implicit none

      real(dop)                   :: r
      real(dop), intent(in)       :: time
      complex(dop), intent(inout) :: vz

! Adding initial gaussian smoothing function (always first 20fs)
      if (time/fs .gt. 0) then 
         if (time/fs .lt. 20) then
            r=time-(20*fs)
            vz=vz*(exp((r*r)/(-2000*fs)))
         endif
      endif
! Switching Function (inverse gaussian) 20fs switch off/on.
      if (nmb .eq. 1) then      
         if (abs((switch(nmb)-time)) .lt. (20*fs)) then
            r=(switch(nmb)-(20*fs))-time
            vz=vz*(exp((r*r)/(-2000*fs)))
         endif
      endif

      if (nmb .gt. 1) then
         if (time .gt. (switch(nmb)-(20*fs))) then 
            r=(switch(nmb)-(20*fs))-time
            vz=vz*(exp((r*r)/(-2000*fs)))
         endif
         if (abs((switch(nmb-1)-time)) .lt. (20*fs)) then
            r=(switch(nmb-1)+(20*fs))-time
            vz=vz*(exp((r*r)/(-2000*fs)))
         endif
      endif

      return
      end subroutine gausspulse


!#######################################################################
! Calculates the pulse for a general operator
!

      subroutine lct(vz,time,h1,hl,psi)

      implicit none

      integer(long)                            :: nham,ire
      integer(long), intent(in)                :: h1,hl
      real(dop)                                :: lambda,steppos
      real(dop), intent(in)                    :: time
      complex(dop)                             :: evalue,svalue
      complex(dop), intent(out)                :: vz
      complex(dop), dimension(*), intent(inout):: psi

      svalue=(1.0_dop,0.0_dop)

      vz=0.0_dop
      lambda=hoprpar(2,h1)
      nham=hopipar(2,h1)
      ire=hopipar(3,h1)

      ! Creating pulse
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham),evalue)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif

! Get the step function of an expectation value.
      if(hterm(2,hl).gt.0.and.hopfile(hterm(2,hl)).eq.309)then
         nham=hopipar(2,hterm(2,hl))

! Get the expectation value (svalue)
         if (lexact) then
            call exexpectvalue(psi,time,hamlab(nham),svalue)
         else 
             call expectvalue(psi,time,hamlab(nham),evalue,&
                  mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
                  agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
                  mr(mrgsval),gsvek,ghderiv1)
         endif

! Evaluate the step
         steppos=hoprpar(2,hterm(2,hl))
         if(abs(imag(svalue)).gt.1.0d-10)then
            routine = 'lct'
            write(message,'(a,d21.16)')'Complex expectation',imag(svalue)
            call errormsg
         endif

         if(dble(svalue).le.steppos)then
            svalue = 0.0_dop
         else
            svalue = 1.0_dop
         endif
      endif

      if (ire .eq. 0) then
         vz=lambda*dble(evalue)*dble(svalue)!*envelope
      else
         vz=lambda*dimag(evalue)*dble(svalue)
      endif

! Smoothing Pulse
!      if (lgpulse .eqv. .false.) then
!         if (time .gt. 0.0_dop) then
!            call gausspulse(time,vz)
!         endif
!      endif

      return
      end subroutine lct

!#######################################################################
! Calculates the pulse for a general operator
!
      subroutine lcttd(vz,time,hl,psi)

      implicit none

      integer(long)                            :: nham,ire,nham1,nham2
      integer(long), intent(in)                :: hl
      real(dop)                                :: revalue,rfvalue, &
                                                  diff,r,r2, &
                                                  gradact,lambdaopt,envelope, &
                                                  revalue1
      real(dop), save                          :: lambda,grad
      real(dop), intent(in)                    :: time
      complex(dop)                             :: evalue,fvalue,currmom
      complex(dop), intent(out)                :: vz
      complex(dop), dimension(*), intent(inout):: psi

      logical(kind=4), save :: initial = .true.

       if (initial) then
          initial=.false.
          lambda = 0.5
          grad=1.0
       endif

      vz=0.0_dop
      nham=hopipar(2,hl)
      ire=hopipar(3,hl)
      nham1=hopipar(4,hl)
      nham2=hopipar(5,hl)

! Creating pulse. Get expectation to define field.
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham),evalue)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif

! Get expectation of "forces" operator 
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham1),fvalue)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif

! Get expectation of condition operator 
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham2),currmom)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif

      revalue = 0.0_dop
      rfvalue = 0.0_dop
      if (ire .eq. 0) then
         revalue = dble(evalue)
         rfvalue = dble(fvalue)
      else
         print*, 'Real part has to be taken'
      endif

! Pulse envelope
! ilctenv = 1 Gaussian
      r=time-lctenvtime
      r2=r*r
      envelope=0.0_dop
      if (ilctenv .eq. 1) then
         envelope=exp(-r2/(2.0d0*lctenvwidth*lctenvwidth))
      endif

! Approximate gradient
      r=time-lctftime
      if (dabs(r) .gt. 10.0) grad=(lctevalue-dble(currmom))/dabs(r)
      if (dble(currmom) .gt. lctevalue) grad=-grad

! Actual gradient
      gradact = lambda*revalue*revalue*envelope + rfvalue

! Best guess lambda
      diff = grad - gradact 
      revalue1=abs(revalue)
      revalue1=max(revalue1,1.0d-2)
      lambdaopt = tanh(diff) / revalue1
      diff = (lambdaopt-lambda) 
      if (abs(diff) .gt. 0.1*lambda) then
         if (lambdaopt .lt. lambda) then
            lambda = lambda -0.1*lambda
         else if (lambdaopt .gt. lambda) then
            lambda = lambda +0.1*lambda
         endif
      else 
         lambda = lambdaopt
      endif
      lambda = min(lambda,1000.0d0)
!      print*,time/fs,lambdaopt,lambda

! Value of field
      vz = lambda * revalue * envelope

      return
      end subroutine lcttd

!#######################################################################
! Calculates the pulse for a general operator
!

      subroutine lct_2d(vz,time,hl,psi)

      implicit none

      integer(long)                            :: nham,ire,nham1
      integer(long), intent(in)                :: hl
      real(dop)                                :: lambda
      real(dop), intent(in)                    :: time
      complex(dop)                             :: evalue,fvalue
      complex(dop), intent(out)                :: vz
      complex(dop), dimension(*), intent(inout):: psi


      vz=0.0_dop
      nham=hopipar(2,hl)
      ire=hopipar(3,hl)
      nham1=hopipar(4,hl)
      lambda=hoprpar(2,hl)

! Creating pulse. Get expectation to define field.
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham),evalue)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif

! Get expectation of condition operator 
      if (lexact) then
         call exexpectvalue(psi,time,hamlab(nham1),fvalue)
      else 
         call expectvalue(psi,time,hamlab(nham),evalue,&
              mr(mrhops),mc(mchpsi),hteil,mi(mijindx),mr(mrort),&
              agmat,gfgwp,aigwp,zgwp,mr(mrqcent),gh,gs2,hung,&
              mr(mrgsval),gsvek,ghderiv1)
      endif


      if (ire .eq. 0) then
         if (dble(evalue) .lt. 0.01.and.dble(evalue).ge.0.0) evalue = 0.01
         if (dble(evalue) .gt.-0.01.and.dble(evalue).le.0.0) evalue =-0.01
         vz=lambda*dble(fvalue/evalue)
!         vz=lambda*dble(evalue)
      else
         if (dimag(evalue) .lt. 0.001) evalue = dcmplx(0,0.001)
         vz=lambda*dimag(fvalue/evalue)
      endif
      print*,time/fs,dble(evalue),dble(fvalue)

!      if (dble(vz).lt.-0.1) vz = -0.1
!      if (dble(vz).gt.0.1) vz = 0.1

! Smoothing Pulse
      if (lgpulse .eqv. .false.) then
         if (time .gt. 0.0_dop) then
            call gausspulse(time,vz)
         endif
      endif

      return
      end subroutine lct_2d

      end module tdkoeffs

