!#######################################################################
!
!              MCTDH-module TIMING
! 
! analyses the time needed in the various sections of the program
!
!  contains subroutines: timinit, timout, strttim, stoptim, readtim
!
!#######################################################################

      module timing

use decimal, only: long, dop, vlong
use mltypes, only: wfnode_t
use op1lib,  only: zerovxi, zeromxd
use logdat

      implicit none

      public :: timinit,timout,readtim,get_timer,start_timer,stop_timer, &
                cmp_totaltime,init_node_timer,print_timers,walltime
#ifdef MPI
      public :: mpitiming
#endif

      integer, private, parameter :: timer_desc_len = 15

   type, public :: timer_t
        real(dop)                     :: cputime  ! accumulated CPU time
        real(dop)                     :: systime  ! accumulated system time
        real(dop)                     :: realtime ! accumulated real time
        integer                       :: ncalls   ! number of invocations
        character(len=timer_desc_len) :: desc     ! description of timer
    end type

      integer(long), private, save               :: max_timers = 0
      type(timer_t), allocatable, public, save   :: timers(:)

      integer(vlong), private, save              :: count_rate
      real(dop), private, save                   :: initsys
      real(dop), public, save                    :: initcpu
! wall clock time used in previous calculation if restart
      integer(long), private, dimension(4), save :: rstwall

      private

      contains

!***********************************************************************
!
!                   TIMINIT
!
! called from main programs.
! opens timing file and reads old data if it exists.
!
! sets up the arrays needed for the timing analysis
!
! V6.0 AJ
!
!***********************************************************************
 
      subroutine timinit

#ifdef MPI
  use mpi
#endif
use global
use dvrdatmod
use griddatmod
use channels

use mpidata

      implicit none

#ifdef MPI
      character(len=4) cmpirank
      character(len=12) pfile
#endif
      logical(kind=4)   :: loldtime,loptime
      character(len=c5) :: filename

      real(dop)         :: cpu_now
      integer(vlong)    :: icount
! logical to flag initial entry
      logical(kind=4), save :: initial = .true.

#ifdef MPI
! TIMING OF MPI TO BE SORTED. Presently only on master.
      if (mpirank .gt. 0) return
#endif

!-----------------------------------------------------------------------
! initialise overall times and get count_rate for system clock
!-----------------------------------------------------------------------
      if (initial) then
         call cpu_time(cpu_now) ! F95 intrinsic
         call system_clock(icount,count_rate) ! F90 intrinsic
         initcpu = cpu_now
         initsys = dble(icount)/dble(count_rate)
         rstwall = 0
      endif

!-----------------------------------------------------------------------
! open timing file, and read it if continuation run
! if first entry and overwrite then open new file
!-----------------------------------------------------------------------
      if (ltime) then
         loldtime=.false.
         if (mpirank.eq.0 .and. ltime) then
            filename=name(1:laenge)//'/timing'
            inquire(file=filename,exist=loldtime)
         endif

         if (initial .and. lovwr) then
            open(itim,file=name(1:laenge)//'/timing',status='unknown', &
                   form='formatted',err=999)
            write(ilog,'(/,a)') 'Timing file opened'

         else if (loldtime) then
            inquire(file=filename,opened=loptime)
            if (loptime) then
               rewind(itim)
            else
               open(itim,file=name(1:laenge)//'/timing',status='old', &
                   form='formatted',err=999)
               write(ilog,'(/,a)') 'Timing file opened'
            endif
#ifdef MPI
            if (lmpi) then
               open(itimmpi,file=name(1:laenge)//'/mpitiming', &
                    status='unknown',form='formatted',err=999)
            endif
#endif
            call readtim

         else if (initial) then
            open(itim,file=name(1:laenge)//'/timing',status='new', &
                   form='formatted',err=999)
            write(ilog,'(/,a)') 'Timing file opened'

         else
            routine='Tinit'
            message = 'Timing file not found'
            call errormsg
         endif

      endif
      initial=.false.

      return
  999 continue
      routine='Tinit'
      message = 'Cannot open timing file'
      call errormsg

      end subroutine timinit

!***********************************************************************
!         TIMOUT
!
! called from datout
! calls to system C routines fcputime and fsystime
!
! write timing analysis to timing file
!
!  V6.0 AJ
!
!***********************************************************************

      subroutine timout

#ifdef MPI
   use mpi
#endif

use global
use channels

use mpidata

use op1lib, only: zerovxd

      implicit none

      real(dop)                              :: percpu
      real(dop), external                    :: fcputime,fsystime
!      external fcputime,fsystime
#ifdef MPI
      integer(long)                          :: t
      real(dop), dimension(2,nmpi)           :: mpicpu
      real(dop)                              :: mpiall
#endif
      integer(long)      :: i
      real(dop)          :: cpu_now,totcpu,totwall
      integer(vlong)     :: icount
      integer(long)      :: nhrs,nmins,nsecs
      character(len=100) :: buffer

      if (mpirank .gt. 0) return

!-----------------------------------------------------------------------
! get overall times
! endwall (wall time) has following values
! 1 - year
! 2 - month
! 3 - day
! 4 - time wrt UTC
! 5 - hrs
! 6 - mins
! 7 - secs
! 8 - millisecs
!-----------------------------------------------------------------------
      call cpu_time(cpu_now) ! F95 intrinsic
      call system_clock(icount) ! F90 intrinsic
      totcpu = cpu_now - initcpu
      totwall = dble(icount)/dble(count_rate) - initsys

      call getwallclk(totwall,nhrs,nmins,nsecs)

! total wall time in s

!-----------------------------------------------------------------------
! write file header
!-----------------------------------------------------------------------
      rewind(itim)
      write(itim,'(1x,a72)') cstartup(1:72)
      write(itim,'(1x,a)') name(1:laenge)
      write(itim,'(1x,75a)') ('-',i=1,75)

      write(itim,*) ' '
      write(itim,'(1x,a15,4x,a7,7x,a5,10x,a3,5x,a4,8x,a5)') &
         'Subroutine     ','Calls N','cpu/N','cpu','%cpu','Clock'

!-----------------------------------------------------------------------
! output timing analysis
!-----------------------------------------------------------------------
      if (totcpu.gt.0.0_dop) then
         percpu=100.0_dop/totcpu
      else
         percpu=0.0_dop
      endif

! new output code
      call print_timers(itim, percpu)

      write(itim,21) totcpu,totwall,nhrs,nmins,nsecs

!-----------------------------------------------------------------------
! write file tail
!-----------------------------------------------------------------------
      call fpstartup(buffer)
      write(itim,'(/,1x,75a)') ('-',i=1,75)
      write(itim,'(1x,a72)') buffer(1:72)

      endfile(itim)
      call myflush(itim)

   21 format(1x,/'Total (cpu,wall) ',f12.3,1x,f12.3,1x, &
            '  (',i6,'h :',i3,'m :',i4,'s)',/)

#ifdef MPI
      if (lmpi.and.mpirank.eq.0.and.lmpistarted.and.lpartim) then
         mpitask=999
         call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         call mpitiming(mpicpu)
         mpicpu(1,1)=totcpu
         mpiall=0.0
         do t=1,nmpi
            mpiall=mpiall+mpicpu(1,t)
         enddo
         rewind(itimmpi)
         write(itimmpi,'(1x,a7,10x,a3,10x,a4)')'Process','cpu','%cpu'
         write(itimmpi,'(40(a1))')('-',t=1,40)
         do t=1,nmpi
            write(itimmpi,'(1x,i4,5x,f12.1,7x,f6.2)') &
                 t-1,mpicpu(1,t),mpicpu(1,t)/mpiall*100.0
         enddo
         endfile(itimmpi)
         call myflush(itimmpi)
      endif
#endif

      return
      end subroutine timout

!#######################################################################
! wallclk is no. of secs elapsed
!#######################################################################

      subroutine getwallclk(wallclk,nhrs,nmins,nsecs)

      implicit none 

      real(dop), intent(in)       :: wallclk
      real(dop)                   :: tmpwall
      integer(long),intent(out)   :: nhrs,nmins,nsecs

      nhrs = int(wallclk/3600.0)
      tmpwall = wallclk - nhrs*3600.0
      nmins = int(tmpwall/60.0)
      tmpwall = tmpwall - nmins*60.0
      nsecs = nint(tmpwall)

      return
      end subroutine getwallclk

!#######################################################################
! READTIM
!
! reads timing file. Needed for continuation runs.
!
!#######################################################################

      subroutine readtim

use global
use channels

      implicit none

      real(dop)               :: rsttotcpu,rstclksys
      real(dop)               :: cpu_now
      integer(vlong)          :: icount
      integer(long)           :: ierr
      character(len=12)       :: string

      rewind(itim)
! read blank line and column headings 
      do
        read(itim,'(a12)',end=30) string
        if (string .eq. ' Subroutine ') exit
      enddo

      call read_timers(itim)

      read(itim,'(17x,f12.3,1x,f12.3)',iostat=ierr) rsttotcpu,rstclksys
      if (ierr .gt. 0) then
! try old format
         read(itim,'(17x,f12.3,1x,f12.3)',err=999) rsttotcpu,rstclksys
      endif
      
! add already used times to final total (totcpu = present - initcpu)
      call cpu_time(cpu_now) ! F95 intrinsic
      initcpu = cpu_now-rsttotcpu 
      call system_clock(icount,count_rate) ! F90 intrinsic
      initsys = dble(icount)/dble(count_rate) - rstclksys

      write (ilog,'(a)') 'Timing file read'

   30 continue
      return
  999 continue
      routine='Readtim'
      message='Error reading timing file. Check start type'
      call errormsg

      end subroutine readtim

!-----------------------------------------------------------------------
!  MPITIMING
!-----------------------------------------------------------------------
#ifdef MPI
      subroutine mpitiming(mpicpu)

use mpi
use global
use maxv, only: maxdim
use mpidata

      implicit none

      real(dop),dimension(2,nmpi)  ::  mpicpu

      integer(long),dimension(mpi_status_size)  :: status

      integer(long)                             :: i
      real(dop)                                 :: sumreal

      if (mpirank.eq.0) then
         do i=2,nmpi
            call MPI_Recv(mpicpu(1,i),1,MPI_DOUBLE_PRECISION,i-2,498+i, &
                                           MPI_COMM_WORLD,status,mpierr)
         enddo
      else
!         call MPI_Send(tcalls(1,tmpi),1,MPI_DOUBLE_PRECISION,0, &
!                                      500+mpirank,MPI_COMM_WORLD,mpierr)
      endif
      return
      end subroutine mpitiming
#endif

!***********************************************************************
! print timers
!***********************************************************************
    subroutine print_timers(itim, tscale)
        implicit none
        integer, intent(in) :: itim  ! i/o channel
        real(dop) :: tscale     ! percentage scale
        integer   :: order(max_timers)
        integer   :: i,j

        call hybrid_sort(max_timers, order, cmp_totaltime)
        do j=1,max_timers
            i = order(j)
            if (timers(i)%ncalls == 0) cycle ! skip unused timers

! if format is changed, change that in read_timers
            write(itim, '(1x,a15,1x,i10,1x,f11.4,1x,f14.3,1x,f7.1,1x,f14.3)') &
                timers(i)%desc, &
                timers(i)%ncalls, &
                timers(i)%cputime/timers(i)%ncalls, &
                timers(i)%cputime, &
                timers(i)%cputime*tscale, &
                timers(i)%systime
         end do
    end subroutine print_timers

!***********************************************************************

    ! read timers
    subroutine read_timers(itim)
        implicit none
        integer, intent(in) :: itim  ! i/o channel
        character(len=timer_desc_len) :: desc
        integer       :: tid, ncalls
        real(dop)     :: cputime, walltime

        do
            ! an empty line or "Total" means that all timers have been processed
            read(itim, '(1x,a15)') desc
            if (desc == '      ') exit
            if (desc == 'otal (cpu,wall)') then
               backspace(itim)
               exit
            endif

            backspace(itim)
! format must match that in print_timers
            read(itim, '(1x,a15,1x,i10,13x,f14.3,9x,f14.3)') &
                desc, ncalls, cputime,walltime
            ! look for matching timer or create a new one
            tid = 0
            call get_timer(tid, desc)
            ! and add data to that timer
            timers(tid)%cputime = cputime
            timers(tid)%systime = walltime
            timers(tid)%ncalls  = ncalls
        end do
    end subroutine read_timers

!***********************************************************************

    ! compare two timers by total (cpu+sys) time
    function cmp_totaltime(i1,i2) result(cmp)
        integer, intent(in) :: i1,i2
        integer             :: cmp
        real(dop)      :: t1,t2
        t1 = timers(i1)%cputime + timers(i1)%systime
        t2 = timers(i2)%cputime + timers(i2)%systime
        if (t1 > t2 ) then
            cmp = -1
        elseif (t1 < t2) then
            cmp = 1
        else
            cmp = 0
        end if
    end function cmp_totaltime

!***********************************************************************

    subroutine hybrid_sort(n, order, cmp)
    ! hybrid sort algorithm:
    !   1. sort small sub-arrays by selection sort
    !   2. use non-recursive bottom-up merge sort
    ! runtime = O(n log n)
    ! needs integer work array of size n

        integer, intent(in)  :: n
        integer, intent(out) :: order(n)
        integer,parameter :: hybrid_size = 5
        interface
            function cmp(i1,i2)
                integer, intent(in) :: i1,i2
                integer             :: cmp
            end function cmp
        end interface

        integer :: i,j,c,s,p,t,m,m2,l,r
        integer :: work(n)

        ! initialize order as unsorted
        do i=1,n
            order(i) = i
        end do

        ! First, sort small sub-arrays (of size hybrid_size) by selection sort.
        if (hybrid_size > 1) then
            s = 1 ! start of sub-array
            do while (s<n)
                ! i is the start of the unsorted part of the sub-array
                do i = s, min(s+hybrid_size-2,n-1)
                    ! p is the position of the smallest element
                    p = i
                    ! go through remainder of array to find the minimum
                    do j=i+1, min(s+hybrid_size-1,n)
                        c = cmp(order(j),order(p))
                        if (c < 0)  p = j
                    end do
                    ! if necessary, swap minimum element with position i
                    if (p /= i) then
                        t = order(i)
                        order(i) = order(p)
                        order(p) = t
                    end if
                end do
                s = s + hybrid_size
            end do
        end if

        ! Now switch to merge sort.
        ! The algorithm makes several passes across the array.
        ! In each pass, two adjacent sub-arrays (which are already sorted)
        ! of size m are merged together into a sub-array of size 2*m.
        m = hybrid_size ! initially, sorted sub-arrays are of this size
        do while (m<n)
            m2 = 2*m
            s = 1 ! start of left sub-array
            do while (s<=n-m+1)
                l = s   ! pointer into left sub-array
                r = s+m ! pointer into right sub-array
                i = s   ! pointer into merged sub-array
                do while (i < s+m2 .and. i <= n)
                    if (l >= s+m) then
                        ! left sub-array used up, take from right
                        j = r
                        r = r+1
                    elseif (r >= s+m2 .or. r > n) then
                        ! right sub-array used up, take from left
                        j = l
                        l = l+1
                    else
                        ! compare elements for left and right
                        c = cmp(order(l),order(r))
                        ! and take element from corresponding sub-array
                        if (c <= 0) then
                            j = l
                            l = l+1
                        else
                            j = r
                            r = r+1
                        endif
                    endif
                    ! store the chosen element in the merged sub-array
                    work(i) = order(j)
                    i = i+1
                end do
                s = s+m2 ! advance to the next sub-arrays
            end do
            ! there might be a single left sub-array leftover;
            ! just leave it alone
            ! now copy the merged sub-arrays back
            order(1:i-1) = work(1:i-1)
            m = m2 ! double m for the next pass
        end do

    end subroutine hybrid_sort

!***********************************************************************

    ! get timer ID of an existing timer matching the description,
    ! or create a new timer and return its ID
    subroutine get_timer(id, desc)
        implicit none
        character(len=*), intent(in) :: desc  ! description of timer
        integer, intent(inout)       :: id    ! id of timer
        integer                      :: i
        type(timer_t), allocatable   :: tmptimers(:)


        ! if id is already set, do nothing
        if (id /= 0) return

        ! look for matching timer
        do i=1,max_timers
            if (timers(i)%desc == desc) then
                id = i
                return
            endif
        end do

        ! not found, create new timer
        if (.not. allocated(timers)) then
           max_timers=1
           allocate(timers(max_timers))
        else
           allocate(tmptimers(max_timers))
           tmptimers=timers

           deallocate(timers)
           max_timers=max_timers+1
           allocate(timers(max_timers))

           timers(1:max_timers-1)=tmptimers
           deallocate(tmptimers)
        endif

        id = max_timers
        timers(id)%desc = desc
        timers(id)%cputime = 0.0d0
        timers(id)%systime = 0.0d0
        timers(id)%realtime = 0.0d0
        timers(id)%ncalls = 0
    end subroutine get_timer

!***********************************************************************

    ! start a timer
    subroutine start_timer(id)
        implicit none
        integer, intent(in) :: id
        integer(vlong)      :: icount
        real(dop)           :: cpu_now 
        real(dop)           :: wtim

        call cpu_time(cpu_now)    ! F95 intrinsic
        call system_clock(icount) ! F90 intrinsic
        call walltime(wtim)

        timers(id)%cputime = timers(id)%cputime - cpu_now
        timers(id)%systime = timers(id)%systime - dble(icount)/dble(count_rate)
        timers(id)%realtime = timers(id)%realtime - wtim
        timers(id)%ncalls  = timers(id)%ncalls + 1
    end subroutine start_timer

!***********************************************************************

    ! stop a timer
    subroutine stop_timer(id)
        implicit none
        integer, intent(in) :: id
        integer(vlong)      :: icount
        real(dop)           :: cpu_now
        real(dop)           :: wtim

        call cpu_time(cpu_now) ! F95 intrinsic
        call system_clock(icount) ! F90 intrinsic
        call walltime(wtim)

        timers(id)%cputime = timers(id)%cputime + cpu_now
        timers(id)%systime = timers(id)%systime + dble(icount)/dble(count_rate)
        timers(id)%realtime = timers(id)%realtime + wtim

    end subroutine stop_timer

!***********************************************************************

   subroutine init_node_timer(no,tid,what)
        implicit none
        type(wfnode_t),intent(inout)  :: no   ! node
        integer,intent(inout)         :: tid  ! which timer (part of node!)
        character(len=*)              :: what ! for timer description
        integer, parameter :: timer_desc_len = 20
        character(len=timer_desc_len) :: tdesc
        write(tdesc,'(2a,i0)') trim(what),'_n',no%num
        call get_timer(tid, tdesc)
    end subroutine init_node_timer

!***********************************************************************

! ----------------------------------------------------------------------
! Returns the wall-time in seconds as a real*8 number
! (HDM 03/2023)
! ----------------------------------------------------------------------

   subroutine walltime(rtime)
   implicit none
   real*8, intent(out)  :: rtime
   integer              :: hour
   integer, save        :: addhour=0, oldhour=0
   integer,dimension(8) :: values

   call date_and_time (VALUES=values)

   hour=values(5)
   if(hour .lt. oldhour)  addhour=addhour+24
   oldhour=hour
   rtime=3600.d0*(hour+addhour)+60.d0*values(6)+dble(values(7))+1.d-3*values(8)
   return
   end


      end module timing



