!#############################################################################
!                 --- Heidelberg MCTDH ---                                   #
!                     Module: multilayer_m                                   #
!                     Author: Oriol Vendrell (oriol.vendrell@gmail.com)      #
!                     Year  : 2008                                           #
!#############################################################################

! Driver on top of the ML machinery. Together with the files interface.F and
! mlservices.F, it provides the necessary infrastructure for the interaction of
! the Heidelberg MCTDH code with the ML modules.

module inputmlmod

    use decimal
    use global
    use wftree
    use mlbase
    use mlanalysis
    use integrator
    use timing
    use if_grid
    use ctensor
    use htucker
    use treemod

    use mltypes
    use gridutil
    use linear
    use griddatmod, only: tree
    use psidef
    use logdat
    use psiutil, only: akcopy2

    private

    ! Public module methods
    public  :: inputml,          &
               truncate_avec,    &
               copy_avecs_tree


    contains

    subroutine inputml(tape)
      use maxv, only: maxtape
        implicit none
        integer, intent(in)      :: tape(maxtape)

        tree => new_wftree(tape,if_spfdof)
        
        ! Calculate size of all ML-coeffs in psi
        mladim  = adim_wftree(tree)

        ml_input_ready = .true.
        if(ml_input_ready.and.&
           ml_oper_ready.and.&
           ml_prop_ready) then
           ml_ready=.true.
        end if

        if (tree%istop .and. tree%isprimitive) then
           routine="Inputml (inputml.f90)"
           message = 'Exact calculations with ML code not implemented'
           call errormsg
        endif

    end subroutine inputml


!#######################################################################

subroutine truncate_avec(phi,ni,nj,nk,nn)
   integer,intent(in)          :: ni,nj,nk,nn
   complex(cdbl),intent(inout) :: phi(ni,nj,nk)
   integer                     :: i,j,k
   do k=1,nk
      do j=nn+1,nj
         do i=1,ni
            phi(i,j,k) = CZERO
         enddo
      enddo
   enddo
end subroutine truncate_avec

!#######################################################################

!-----------------------------------------------------------------------
subroutine copy_avecs_tree(tree,psi,tree1,psi1)
!-----------------------------------------------------------------------
! Copies all avecs from tree to tree1, taking into account that the
! numbers of SPFs can be different.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in) :: tree,tree1
   complex(cdbl),intent(in)  :: psi(:)
   complex(cdbl),intent(out) :: psi1(:)
   type(wfnode_t),pointer    :: node,node1
   integer                   :: m,i,z0i,z0f,z1i,z1f
   do m=1,tree%tot_nodes
      node => tree%swap_up(m)%p
      node1 => tree1%swap_up(m)%p
      do i=1,min(node%navecs,node1%navecs)
         z0i = node%z(i,1)
         z0f = node%z(i,2)
         z1i = node1%z(i,1)
         z1f = node1%z(i,2)
         if (node%istop .and. dentype .eq. 2) then
            call copy_d2tensor(psi(z0i:z0f),node%ndim,psi1(z1i:z1f),&
                 node1%ndim,z0f,z1f,node%nmodes)
         else
            call copy_tensor(psi(z0i:z0f),node%ndim,psi1(z1i:z1f),node1%ndim)
         endif
      enddo
   enddo
end subroutine copy_avecs_tree

!#######################################################################

subroutine copy_tensor(a,adim,b,bdim)
   complex(cdbl),intent(in)  :: a(:)
   integer,intent(in)        :: adim(:)
   complex(cdbl),intent(out) :: b(:)
   integer,intent(in)        :: bdim(:)
   integer                   :: bidx(size(bdim))
   integer                   :: aa,bb
   bidx = 1
   do
      bb = flat_index(bidx,bdim)
      if (any(bidx>adim)) then
         b(bb) = CZERO
      else
         aa = flat_index(bidx,adim)
         b(bb) = a(aa)
      endif
      if (.not.next_point(bidx,bdim)) exit
   enddo
end subroutine copy_tensor

!#######################################################################
!
! Copy and expand top layer vector for density2
!
!#######################################################################

subroutine copy_d2tensor(a,adim,b,bdim,mblock,mblock1,nmodes)

   implicit none

   complex(cdbl),intent(in)  :: a(:)
   integer,intent(in)        :: mblock,mblock1,nmodes
   integer,intent(in)        :: adim(:)
   complex(cdbl),intent(out) :: b(:)
   complex(cdbl)             :: work(2*mblock1)
   integer,intent(in)        :: bdim(:)
   integer                   :: zeig,zeig1,zeig2,m
   integer                   :: ijk(3),ijk1(3)

   work(1:mblock) = a(1:mblock)

   zeig=1
   zeig1=mblock1+1
   do m=1,nmodes
      ijk=ijk_shape(m,adim)
      ijk1=ijk_shape(m,bdim)
      call akcopy2(work(zeig),work(zeig1),ijk1(1),&
           ijk1(2),ijk(2),ijk(3),ijk1(1),&
           ijk1(2),ijk(2),ijk(3))
      zeig2=zeig
      zeig=zeig1
      zeig1=zeig2
   enddo
      
   b(1:mblock1) = work(1:mblock1)

end subroutine copy_d2tensor


end module inputmlmod
