!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: eqofmotion.f90 518 2013-11-05 18:32:34Z frank $


module eqofmotion

    use decimal
    use global
    use globmemmod
    use wftree
    use linear
    use mltensor
    use timing

    use mltypes
    use psiutil, only: akcopy
    use hphimod, only: hcall
    use logdat
    use griddatmod
    use operdef
    use psidef, only: dmat,d3mat,complete,intrace
    use hphimod, only: hcall
    use dvrdatmod, only: fkoe
    use eqofmotion1
    use mlop_eom

    use op1lib, only: trhxz
    use op2lib
    use tdkoeffs, only: calctdk
    use runpropmod, only: decorr,deuncorr

    use openmpmod

#ifdef OMP
use omp_lib
#endif

    private
    public  :: funktionml,                  &
               compute_rhs_tree,            &
               compute_rhs_treer,           &
               compute_mlop_noder,          &
               compute_rhs_node,            &
               compute_rhs_a,               &
               compute_rhs_allphi,          &
               koeff_ml

    logical,save   :: debug = .false.

    contains

    subroutine funktionml(time,psi,dtpsi,pdim,cntrl,node)
        implicit none
        ! ---
        ! Call various computation routines needed for the evaluation
        ! of the derivatives of the wavefunction coefficients.
        !
        ! time: actual time in the propagation
        ! psi,dtpsi: complex arrays with the wavefunction and its derivative
        ! plen: length of psi and dtpsi. Must be given explicitly since this
        !       routine is usually called from outside a module, which makes
        !       the interface non-explicit, and therefore assumed-shape does not
        !       work.
        !
        ! cntrl: control which specific subroutines are active in a certain
        ! call. This is to make the implementation of CMF-like schemes
        ! possible (and easy). The various possible choices for the cntrl
        ! variable emulate the traditional funk0, funk1, funka, ...
        !
        ! cntrl | hterms | rho+MF | MFsum | dtA(top) | dtPhi || emulates
        !--------------------------------------------------------------------
        !   0       X        X        -        X         -      funk0
        !   1       -        -        -        -         X      funk1
        !   2       X        -        -        -         -      funk2
        !   3       -        X        X        X         X      funk3
        !   4       X        X        X        -         -      funk0 (no dtA)
        !  10       X        X        -        X         X      funk
        !  11       -        -        -        X         -      funkA
        !  12       -        -        -        -         X      funkphi
        !  13       -        -        -        X         X      funkA+funkphi
        !--------------------------------------------------------------------
        !
        ! node: node for which the derivative is taken if cntrl==12.
        !       ignored otherwise.
        !       Important note: the numbering assumed is the one
        !                       in the swap_phi list (see node_m.f90)
        ! ---
        real(dbl), intent(in)     :: time   ! actual time (for TD operators)
        integer(long), intent(in) :: pdim
        complex(cdbl)             :: psi(pdim)
        complex(cdbl)             :: dtpsi(pdim)
        integer                   :: cntrl,node
        integer                   :: z0,zf

        if(.not.ml_ready) then
            write(6,*) 'Multilayer has not been properly initialized'
            write(6,*) 'Execution will abort'
            stop
        end if

        select case(cntrl)

            case(0)
                call compute_hterms_tree(time,tree,psi,1)
                call compute_all_uterms_tree(time,1,tree,psi)
                call compute_dmat_tree(tree,psi)
                call compute_mfields_tree(tree,psi)
                call compute_all_wterms_tree(time,1,tree,psi)
                tz => tree%swap_down(1)%p
                z0 = tz%zr(1,1)
                zf = tz%zr(1,2)
                call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))

            case(1)
                call compute_rhs_allphi(tree,psi,dtpsi)

            case(2)
                call compute_hterms_tree(time,tree,psi,1)
                call compute_all_uterms_tree(time,1,tree,psi)

            case(3)
                call compute_dmat_tree(tree,psi)
                call compute_mfields_tree(tree,psi)
                call compute_all_wterms_tree(time,1,tree,psi)
                call compute_mfsum_tree(time,tree)
                call compute_rhs_tree(tree,psi,dtpsi)

            case(4)
                call compute_hterms_tree(time,tree,psi,1)
                call compute_all_uterms_tree(time,1,tree,psi)
                call compute_dmat_tree(tree,psi)
                call compute_mfields_tree(tree,psi)
                call compute_all_wterms_tree(time,1,tree,psi)
                call compute_mfsum_tree(time,tree)

            case(10)
                call compute_hterms_tree(time,tree,psi,1)
                call compute_all_uterms_tree(time,1,tree,psi)
                call compute_dmat_tree(tree,psi)
                call compute_mfields_tree(tree,psi)
                call compute_all_wterms_tree(time,1,tree,psi)
                call compute_rhs_tree(tree,psi,dtpsi)

            case(11)
                tz => tree%swap_down(1)%p
                z0 = tz%zr(1,1)
                zf = tz%zr(1,2)
                call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))

            case(12)
                if(node<1 .or. node>tree%tot_nodes) then
                    write(6,*) 'funktionml: node number out of range,',node
                    stop
                end if
                tz => tree%swap_phi(node)%p
                z0 = tz%zr(1,1)
                zf = tz%zr(tz%navecs,2)
                call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))

            case(13)
                call compute_rhs_tree(tree,psi,dtpsi)

            case default
                write(6,*) 'funktionml: control case not implemented,',cntrl
                stop

        end select

        ! Do additional things to dtpsi (but not for cntrl=2, where funktionml
        ! was called with dtpsi=psi, and not for cntrl=4, where we don't
        ! compute a RHS)
        if (cntrl.ne.2 .and. cntrl.ne.4) then
           ! For relaxation, do top layer projector;
           ! but not for ml-cmf=split where we renormalize instead
           ! TODO: maybe this shouldn't be here
           if(tree%isrelax .and. .not.tree%isrlxt .and. tree%cmftyp /= 2) then
              call projectout_psi(tree,psi,dtpsi)
           end if
           ! Apply prefactor
           dtpsi = tree%prefactor * dtpsi
        endif

    end subroutine funktionml

    subroutine compute_rhs_tree(tree,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop)                :: psi(:),dtpsi(:)
        integer                  :: i,z0,zf
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' RHS TREE'
        call get_timer(tid, 'rhs_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            z0 = cnode%z(1,1)
            zf = cnode%z(cnode%navecs,2)
            call compute_rhs_node(cnode,psi(z0:zf),dtpsi(z0:zf))
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END RHS TREE'
    end subroutine compute_rhs_tree

    subroutine compute_rhs_treer(tree,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop)                :: psi(:),dtpsi(:)
        integer                  :: i,z0,zf
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' RHS TREE'
        call get_timer(tid, 'rhs_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            if (cnode%isprimitive) cycle

            z0 = cnode%z(1,1)
            zf = cnode%z(cnode%navecs,2)
            call compute_rhs_node(cnode,psi(z0:zf),dtpsi(z0:zf))
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END RHS TREE'
    end subroutine compute_rhs_treer

    subroutine compute_mlop_noder(tree,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop)                :: psi(:),dtpsi(:)
        integer                  :: i,z0,zf
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' RHS TREE'
        call get_timer(tid, 'mlop_noder')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            if (cnode%isprimitive) then
! lowest layer MLOp contributions
               z0 = cnode%z(1,1)
               zf = cnode%z(cnode%navecs,2)
               call compute_all_mlop_rhs_node(cnode,psi(z0:zf),dtpsi(z0:zf))
            endif
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END RHS MLOP'
    end subroutine compute_mlop_noder


    subroutine compute_rhs_allphi(tree,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop)                :: psi(:),dtpsi(:)
        integer                  :: i,z0,zf
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' RHS PHI'
        call get_timer(tid, 'rhs_allphi')
        call start_timer(tid)
        do i=2,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            if (cnode%isprimitive) cycle
            z0 = cnode%z(1,1)
            zf = cnode%z(cnode%navecs,2)
            call compute_rhs_node(cnode,psi(z0:zf),dtpsi(z0:zf))
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END RHS PHI'
    end subroutine compute_rhs_allphi


    subroutine compute_rhs_node(t,psi,dtpsi)
    ! Eq.(26-27), Manthe,[JCP,128,164116,(2008)]
    ! Schematically, this routine does:
    !   dt_|phi> = (1-P) [ h_uncor + (rho)^-1 * <H>*h_cor ] |phi>
    ! OVR: Note added on 19.08.10
    !   The routine does H|Psi>. No prefactor is multiplied and
    !   the projector E|Psi> is not substracted for relaxations.
    !   This must be taken care of after completion of this call.
        implicit none
        type(wfnode_t),intent(inout) :: t
        complex(dop)                 :: psi(:),dtpsi(:)
        real(dop)                    :: en
        type(wfnode_t),pointer       :: tp
        type(hterm_tp),pointer       :: hk(:)
        complex(dop),pointer         :: mfield(:,:),invdmat(:,:)
        complex(dop),allocatable     :: work(:),huncorr(:),tmpdtpsi(:,:)
        integer                      :: k,ns,a,b,f,ithr
        integer                      :: z0,zf,z00,zff,plen,totplen,h
        logical                      :: lcmf

        plen = product(t%ndim)
        totplen=plen*t%navecs
        dtpsi = CZERO
        lcmf = (t%topnode%cmftyp /= 0)

        if (t%tid_rhs == 0) call init_node_timer(t, t%tid_rhs, 'rhs')
        call start_timer(t%tid_rhs)

        if(t%istop) then  ! MCTDH FIRST LAYER rhs equation

            z0 = t%z(1,1)
            zf = t%z(1,2)
            if (dentype .eq. 2) then
               decorr(1) = 0.0_dop
               deuncorr(1) = 0.0_dop
            endif

            if (dentype .eq. 2) then
               allocate(tmpdtpsi(plen*plen,ompthread))
            else
               allocate(tmpdtpsi(plen,ompthread))
            endif
            tmpdtpsi = CZERO
!$OMP parallel default(shared) private(work,tmpdtpsi,k,hk,f,h,ithr) 
            if (dentype .eq. 2) then
               allocate(work(plen*plen))
            else
               allocate(work(plen))
            endif
!$OMP do schedule(runtime)
            do k=1,kzahl(1) ! loop over k-terms
#ifdef OMP
                ithr = omp_get_thread_num()+1
#else
                ithr = 1
#endif
                hk=>t%hterms(k,:)
                work = psi
                if (dentype .eq. 2) then
                   call hprod_phi2(hk,work)
                else
                   call hprod_phi(hk,work)
                endif
                h=hamilton(fkoe,k)
                if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104) then
                   work =  koeff(k)*work
                endif
                tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + work(:)

                if (dentype .eq. 1) then
                   if (lrlx) then
                      call addvcxzo(work,tmpdtpsi(1,ithr),plen)
                   else
                      call subvcxzo(work,tmpdtpsi(1,ithr),plen)
                   endif
                else if (dentype .eq. 2) then
                   if (lrlx) then
                      call addmxazo1(tmpdtpsi(1,ithr),work,plen,plen)
                   else
                      call submxazo1(tmpdtpsi(1,ithr),work,plen,plen)
                   endif
               endif
               if (dentype .eq. 2) then
                  call trhxz(work,en,plen)
                  decorr(1) = decorr(1) + en
               endif

            end do
!$OMP end do
!$OMP do schedule(runtime)
            do f=1,t%nmodes
#ifdef OMP
                ithr = omp_get_thread_num()+1
#else
                ithr = 1
#endif
                if (.not. t%hasuncor(f,1)) cycle
                k=kzahl(1)+1
                hk=>t%hterms(k,:)
                work = psi
                if (dentype .eq. 2) then
                   call hunprod_phi2(hk,work,f)
                else
                   call hunprod_phi(hk,work,f)
                endif
                tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + work(:)

                if (dentype .eq. 1) then
                   if (lrlx) then
                      call addvcxzo(work,tmpdtpsi(1,ithr),plen)
                   else
                      call subvcxzo(work,tmpdtpsi(1,ithr),plen)
                   endif
                else if (dentype .eq. 2) then
                   if (lrlx) then
                      call addmxazo1(tmpdtpsi(1,ithr),work,plen,plen)
                   else
                      call submxazo1(tmpdtpsi(1,ithr),work,plen,plen)
                   endif
                endif
                if (dentype .eq. 2) then
                   call trhxz(work,en,plen)
                   deuncorr(1) = deuncorr(1) + en
                endif
            enddo
!$OMP end do
            deallocate(work)
!$OMP end parallel
            do ithr = 1,ompthread
               dtpsi(:) = dtpsi(:) + tmpdtpsi(:,ithr)
            enddo
            deallocate(tmpdtpsi)

            ! MLOp CONTRIBUTIONS
            call compute_all_mlop_rhs_node(t,psi,dtpsi)

        else  ! ML-MCTDH right-hand-side equation

            ! Note: parent node holds mean fields and density matrices entering
            !       the equations of motion of the coefficients here.

            ns=t%nsubcoord
            tp=>t%parent
            invdmat=>tp%invdmat(ns)%p%matrix

            ! CORRELATED k-TERMS

            allocate(tmpdtpsi(totplen,ompthread))
            tmpdtpsi = CZERO
!$OMP parallel default(shared) private(work,k,hk,mfield,a,b,z0,zf,z00,zff,f,huncorr,ithr) 
            allocate(work(totplen))
!$OMP do schedule(runtime)
            do k=1,kzahl(1) ! loop over oper products
#ifdef OMP
                ithr = omp_get_thread_num()+1
#else
                ithr = 1
#endif
                if(tp%isuncor(k)) cycle  ! is uncorrelated

                mfield=>tp%mfields(k,ns)%p%matrix

                ! - Do the matrix part of the MEAN-FIELD; h_k|phi_a>
                if(t%isprimitive) then
                    ! - for CMF, skip summed-up terms
                    if (lcmf) then
                        if (t%htmtype(k).ge.0) cycle
                    endif
                    do a=1,t%navecs
                        z0  = t%zr(a,1)
                        zf  = t%zr(a,2)
                        call hcall(t%m,k,work(z0:zf),psi(z0:zf),plen)
                    end do
                else
                    hk => t%hterms(k,:)
                    work = psi
                    do a=1,t%navecs
                        z0 = t%zr(a,1)
                        zf = t%zr(a,2)
                        call hprod_phi(hk,work(z0:zf))
                    end do
                end if

                ! - Apply coefficient (if not 1) to SPF's
                h=hamilton(fkoe,k)
                if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104)  &
                        work(1:totplen)=koeff(k)*work(1:totplen)

                ! - Apply MEAN-FIELD coefficients. Result accumulated in dtpsi.
                do a=1,t%navecs
                    z0  = t%zr(a,1)
                    zf  = t%zr(a,2)
                    do b=1,t%navecs
                        z00 = t%zr(b,1)
                        zff = t%zr(b,2)
                        tmpdtpsi(z0:zf,ithr) = tmpdtpsi(z0:zf,ithr) + &
                                                  mfield(a,b)*work(z00:zff)
                    end do
                end do

                ! - Form Liouvillian for density type I
                if (dentype .eq. 1 .and. lrlx) then
                   do a=1,t%navecs
                       z0  = t%zr(a,1)
                       zf  = t%zr(a,2)
                       do b=1,t%navecs
                           z00 = t%zr(b,1)
                           zff = t%zr(b,2)
                           tmpdtpsi(z0:zf,ithr) = tmpdtpsi(z0:zf,ithr) + &
                                             dconjg(mfield(a,b)*work(z00:zff))
                       end do
                   end do
                else if (dentype .eq. 1) then
                   do a=1,t%navecs
                       z0  = t%zr(a,1)
                       zf  = t%zr(a,2)
                       do b=1,t%navecs
                           z00 = t%zr(b,1)
                           zff = t%zr(b,2)
                           tmpdtpsi(z0:zf,ithr) = tmpdtpsi(z0:zf,ithr) - &
                                             dconjg(mfield(a,b)*work(z00:zff))
                       end do
                   end do
                endif
            end do
!$OMP end do

            ! MLOp CONTRIBUTIONS

            call compute_all_mlop_rhs_node(t,psi,dtpsi)


            ! Apply INVERSE DENSITY MATRIX

#ifdef OMP
            ithr = omp_get_thread_num()+1
#else
            ithr = 1
#endif
            work(:)  = tmpdtpsi(:,ithr)
            tmpdtpsi(:,ithr) = CZERO
            do a=1,t%navecs
                z0  = t%zr(a,1)
                zf  = t%zr(a,2)
                do b=1,t%navecs
                    z00 = t%zr(b,1)
                    zff = t%zr(b,2)
                    tmpdtpsi(z00:zff,ithr) = tmpdtpsi(z00:zff,ithr) + &
                                               invdmat(b,a)*work(z0:zf)
                end do
            end do

            ! - UNCORRELATED k-TERMS

!$OMP do schedule(runtime)
            do k=1,kzahl(1) ! loop over oper products
#ifdef OMP
                ithr = omp_get_thread_num()+1
#else
                ithr = 1
#endif

                if(.not.tp%isuncor(k)) cycle  ! is uncorrelated
                if(t%allunit(k)) cycle        ! but motion is in another branch

                if(t%isprimitive) then
                    ! - for CMF, skip summed-up terms
                    if (lcmf) then
                        if (t%htmtype(k).ge.0) cycle
                    endif
                    do a=1,t%navecs
                        z0  = t%zr(a,1)
                        zf  = t%zr(a,2)
                        call hcall(t%m,k,work(z0:zf),psi(z0:zf),plen)
                    end do
                else
                    hk => t%hterms(k,:)
                    work = psi
                    do a=1,t%navecs
                        z0  = t%zr(a,1)
                        zf  = t%zr(a,2)
                        call hprod_phi(hk,work(z0:zf))
                    end do
                end if

                ! - Apply coefficient (if not 1) to SPF's
                h=hamilton(fkoe,k)
                if (htmtype(h) .ne. 1 .and. htmtype(h) .ne. 104) then
                   work =  koeff(k)*work
                endif

                ! - and summ
                tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + work(:)

                if (dentype .eq. 1 .and. kact(k) .eq. 1) then
                   if (lrlx) then
                      tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + dconjg(work)
                   else
                      tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) - dconjg(work)
                   endif
               endif
            end do
!$OMP end do

! and finally explicitely uncorrelated terms

!$OMP do schedule(runtime)
            do f=1,t%nmodes
#ifdef OMP
               ithr = omp_get_thread_num()+1
#else
               ithr = 1
#endif
               if (t%hasuncor(f,1)) then
                  if (t%isprimitive) then
                      allocate(huncorr(totplen))
                      do a=1,t%navecs
                          huncorr=0.0_dop
                          z0  = t%zr(a,1)
                          zf  = t%zr(a,2)
                          do k=kzahl(1)+1,khtot(1)
                             if(t%isunit(k,f)) cycle

                             call hcall(t%m,k,work(z0:zf),psi(z0:zf),plen)
                             huncorr(z0:zf)=huncorr(z0:zf)+koeff(k)*work(z0:zf)
                          enddo
                          work(z0:zf)=huncorr(z0:zf)

                      end do
                      deallocate(huncorr)
                  else
                      k=kzahl(1)+1
                      hk => t%hterms(k,:)
                      work = psi
                      do a=1,t%navecs
                         z0  = t%zr(a,1)
                         zf  = t%zr(a,2)
                         call hunprod_phi(hk,work(z0:zf),f)
                      end do
                  end if

                  ! - and summ
                  tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + work(:)

                   if (dentype .eq. 1 .and. kact(k) .eq. 1) then
                      if (lrlx) then
                          tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) + dconjg(work(:))
                      else
                          tmpdtpsi(:,ithr) = tmpdtpsi(:,ithr) - dconjg(work(:))
                      endif
                   endif
               endif
            enddo
!$OMP end do

            deallocate(work)

!$OMP end parallel

            do ithr = 1,ompthread
               dtpsi(:) = dtpsi(:) + tmpdtpsi(:,ithr)
            enddo
            deallocate(tmpdtpsi)

            ! APPLY PROJECTOR

            call projector_ml(psi,dtpsi,t%navecs,plen)

        end if

        call stop_timer(t%tid_rhs)

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if(debug) then
            write(6,*) 'PSI, node:',t%num
            do a=1,t%navecs
                write(6,*) 'PHI',a
                z0 = t%zr(a,1)
                write(6,10) (psi(z0+b-1),b=1,plen)
            end do
            write(6,*) 'DTPSI, node:',t%num
            do a=1,t%navecs
                write(6,*) 'DTPHI',a
                z0 = t%zr(a,1)
                write(6,10) (dtpsi(z0+b-1),b=1,plen)
            end do
        end if
        10 format(7(2x,'(',1PE11.4,1x,1PE11.4,')'))
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    end subroutine compute_rhs_node


    subroutine koeff_ml(k,time,l,psi)
        implicit none
        integer,intent(in)          :: k
        real(dbl),intent(in)        :: time
        integer,intent(in)          :: l
        complex(cdbl),intent(inout) :: psi(l)
        integer                     :: ktype,h,nham
        complex(cdbl)               :: cz

        h=hamilton(fkoe,k) ! Inquire type of coefficient
        ktype=htmtype(h)

        if(ktype .eq. 101) then  ! TD coefficient
           nham=1
           call calctdk(time,nham,psi)
        end if

        cz = koeff(k)        ! Read koeff value

        if(ktype/=1) then
            psi = psi * cz
        end if
    end subroutine koeff_ml


end module eqofmotion
