!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: eqofmotion.f90 518 2013-11-05 18:32:34Z frank $


module eqofmotion1

    use decimal
    use global
    use mltypes
    use globmemmod
    use wftree
    use linear
    use mltensor
    use timing

    use hphimod, only: hcall
    use mlop_eom, only: compute_all_mlop_mfsum,compute_all_mlop_etot
    use logdat
    use griddatmod
    use operdef
    use psidef, only: dmat,d3mat,complete
    use dvrdatmod, only: fkoe
    use d2mfieldsmod, only: d2mfieldsml
    use runpropmod, only: epsmat

    use op2lib

    private
    public  :: compute_hterms_tree,         &
               compute_hterms_treer,        &
               compute_mfields_tree,        &
               compute_dmat_tree,           &
               compute_mfsum_tree,          &
               ecorrml,                     &
               euncorrml,                   &
               projectout_psi,              &
               init_node_timer,             &
               projector_ml

    logical,save   :: debug = .false.

    contains

! -----------------------------------------------------------------------------
! TREE COMPUTATION PROCEDURES
!
! subroutines and functions that work at the tree level. They contain usually
! calls to node routines, which are the ones doing the real work.
! -----------------------------------------------------------------------------
    subroutine compute_hterms_tree(time,tree,psi,nham)
        implicit none
        real(dop)                    :: time
        type(wfnode_t),intent(inout) :: tree
        integer(long),intent(in)     :: nham
        complex(dop)                :: psi(:)
        integer                  :: i
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' HTERMS TREE'
        call get_timer(tid, 'hterms_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_up(i)%p
            call compute_hterms_node(time,cnode,psi,.false.,nham)
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END HTERMS TREE'
    end subroutine compute_hterms_tree

    subroutine compute_hterms_treer(time,tree,psi,nham)
        implicit none
        real(dop)                    :: time
        type(wfnode_t),intent(inout) :: tree
        integer(long),intent(in)     :: nham
        complex(dop)                :: psi(:)
        integer                  :: i
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' HTERMS TREE'
        call get_timer(tid, 'hterms_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_up(i)%p
            if (cnode%isprimitive) cycle
            call compute_hterms_node(time,cnode,psi,.true.,nham)
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END HTERMS TREE'
    end subroutine compute_hterms_treer


    subroutine compute_mfields_tree(tree,psi)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop),intent(in)     :: psi(:)
        integer                      :: i
        type(wfnode_t),pointer       :: cnode
        integer, save                :: tid = 0 ! timer ID
        if(debug) write(6,*) ' MFIELDS TREE'
        call get_timer(tid, 'mfields_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            call compute_mfields_node(cnode,psi)
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END MFIELDS TREE'
    end subroutine compute_mfields_tree


    subroutine compute_dmat_tree(tree,psi,psi1)
        implicit none
        type(wfnode_t),intent(inout) :: tree
        complex(dop), intent(in)    :: psi(:)
        complex(dop), intent(in), optional :: psi1(:)
        integer                  :: i
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        if(debug) write(6,*) ' DMAT TREE'
        call get_timer(tid, 'dmat_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            call compute_dmat_node(cnode,psi,psi1)
        end do
        call stop_timer(tid)
        if(debug) write(6,*) ' END DMAT TREE'
    end subroutine compute_dmat_tree

    subroutine compute_mfsum_tree(time,tree)
        implicit none
        real(dbl),intent(in)         :: time
        type(wfnode_t),intent(inout) :: tree
        type(wfnode_t),pointer :: cnode
        integer                :: i
        integer,save           :: tid=0
        if (tree%cmftyp == 0) then
            write(*,*) "WARNING: mfsum_tree called for VMF calculation!"
            return
        endif
        call get_timer(tid, 'mfsum_tree')
        call start_timer(tid)
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            call compute_mfsum_node(time,cnode)
        end do
        call stop_timer(tid)
    end subroutine compute_mfsum_tree

! -----------------------------------------------------------------------------
! NODE COMPUTATION PROCEDURES
!
! subroutines and functions that work at the node level
! -----------------------------------------------------------------------------

    subroutine compute_hterms_node(time,t,psi,noprim,nham)
    ! Eq.(33), Manthe,[JCP,128,164116,(2008)]
        implicit none
        real(dop), intent(in)        :: time
        real(dop)                    :: time1
        type(wfnode_t),intent(inout) :: t
        integer(long), intent(in)    :: nham
        complex(dop)                 :: psi(:)
        type(wfnode_t),pointer       :: childf
        complex(dop),allocatable     :: work(:),huncorr(:)
        integer                      :: ndim,subdim,k,f,i,j,f1
        integer                      :: z0,zf
! if true, hterms for primitive nodes not calculated
        logical, intent(in)          :: noprim

! time argument is not used
        time1=time

! no hterms calculated for primitive nodes - stored in layer above 
        if(t%isprimitive) return

        ! loop over modes
        do f=1,t%nmodes
            childf => t%submodes(f)%p
            if (noprim .and. childf%isprimitive) cycle

            if (childf%tid_hterms == 0) then
               call init_node_timer(childf, childf%tid_hterms, 'hterms')
            endif
            call start_timer(childf%tid_hterms)
            ndim   =  t%ndim(f)
            subdim =  product(childf%ndim)

!$OMP       parallel default(shared) private(work,k,i,z0,zf,j,f1,huncorr)
            allocate(work(subdim))
!$OMP       do schedule(runtime)
! loop over h_k oper products (k-terms)
            do k=zham(nham),zham(nham)+kzahl(nham)-1  
                if(associated(t%isunit)) then
                    if(t%isunit(k,f)) cycle ! skip h_f^k
                end if
                do i=1,ndim ! loop over elements h_f^k(i,j)
                    z0 = childf%z(i,1)
                    zf = childf%z(i,2)
                    ! compute h_k|phi_i>
                    if(childf%isprimitive) then
                        call hcall(childf%m,k,work,psi(z0:zf),subdim)
                    else
                        work = psi(z0:zf)
                        call hprod_phi(childf%hterms(k,:),work)
                    end if
                    ! do matrix elements <phi_j|h|phi_i>.
                    ! No symmetry (hermitian, antihermitian,...) is assumed.
                    do j=1,ndim
                        z0 = childf%z(j,1)
                        zf = childf%z(j,2)
                        t%hterms(k,f)%p%matrix(j,i) = phi_phi(psi(z0:zf),work)
                    end do
                end do
            end do
!$OMP       end do

            if (t%hasuncor(f,nham)) then
               allocate(huncorr(subdim))
!$OMP          do schedule(runtime)
               do i=1,ndim ! loop over elements h_f^k(i,j)
                  z0 = childf%z(i,1)
                  zf = childf%z(i,2)
                  ! compute h_k|phi_i>
                  if (childf%isprimitive) then
                     huncorr=0.0_dop
                     do k=zham(nham)+kzahl(nham),zham(nham)+khtot(nham)-1
                        if(t%isunit(k,f)) cycle

                        call hcall(childf%m,k,work,psi(z0:zf),subdim)
                        huncorr=huncorr+koeff(k)*work
                     enddo
                     work=huncorr

                  else
                     huncorr=0.0_dop
! cycle over child nodes and add on any uncorrelated terms. These are uncorrelated
! for the present node.
                     do f1=1,childf%nmodes
                        if (.not. childf%hasuncor(f1,nham)) cycle

                        work = psi(z0:zf)
                        call hunprod_phi(childf%hterms(kzahl(1)+1,:),work,f1)
                        huncorr=huncorr+work
                     enddo
                     work=huncorr
               
                  end if
                    ! do matrix elements <phi_j|h|phi_i>.
                    ! No symmetry (hermitian, antihermitian,...) is assumed.
                  k=zham(nham)+kzahl(nham)
                  do j=1,ndim
                      z0 = childf%z(j,1)
                      zf = childf%z(j,2)
                      t%hterms(k,f)%p%matrix(j,i) = phi_phi(psi(z0:zf),work)
                  end do
               end do
!$OMP end do
               deallocate(huncorr)
            endif
            deallocate(work)
!$OMP       end parallel

            call stop_timer(childf%tid_hterms)
        end do

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if(debug) then
            do f=1,t%nmodes
                ndim = t%ndim(f)
                do k=zham(nham),zham(nham)+khtot(nham)-1
                    if(t%isunit(k,f)) cycle ! skip h_f^k
                    write(6,*) 'hterm: nodenum,f,k', t%num,f,k
                    do j=1,ndim
                        write(6,10) (t%hterms(k,f)%p%matrix(j,i),i=1,ndim)
                    end do
                end do
            end do
        end if
        10 format(8(2x,'(',1PE11.4,1x,1PE11.4,')'))
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    end subroutine compute_hterms_node


    subroutine compute_mfields_node(t,psi)
    ! Eq.(36), Manthe,[JCP,128,164116,(2008)]

! NB the mfields calculated here do not include the Hamiltonian term
! coefficient, unlike in usual MCTDH.
        implicit none
        type(wfnode_t),intent(inout)  :: t
        complex(dop),intent(in)    :: psi(:)

        type(hterm_tp),pointer     :: hk(:)
        complex(dop),pointer       :: mfparent(:,:),mfcurrent(:,:)
        type(wfnode_t),pointer     :: childf
        integer                    :: k,f,p,r,ns
        complex(dop),allocatable   :: work(:)
        complex(dop),allocatable   :: work2(:,:)
        complex(dop),allocatable   :: mftmp(:,:)
        integer                    :: z0,zf,plen

        if(t%isprimitive) return

        if (t%tid_mfields == 0) &
            call init_node_timer(t, t%tid_mfields, 'mfields')
        call start_timer(t%tid_mfields)

        plen = product(t%ndim)

        if(t%istop) then ! NORMAL MCTDH mean fields

        z0 = t%z(1,1)
        zf = t%z(1,2)
!$OMP   parallel default(shared) private(k,hk,f,mfcurrent,work)
        allocate(work(plen))
!$OMP   do schedule(runtime)
        do k=1,kzahl(1) ! loop over oper products
            if(t%isuncor(k)) cycle
            
            if (dentype .eq. 2) then
               call d2mfieldsml(psi(z0:zf),t,k,t%nmodes)
            else
               hk => t%hterms(k,:)
               do f=1,t%nmodes ! loop over modes
                   if(t%isunit(k,f)) cycle ! skip 
                   childf => t%submodes(f)%p
                   if (childf%isprimitive)then
                     if (complete(childf%m)) cycle ! skip for complete prim basis 
                   endif

                   mfcurrent => t%mfields(k,f)%p%matrix
                   work = psi(z0:zf)
                   call hprod_phi(hk,work,f)
                   call compute_dmat(mfcurrent,f,t%ndim,psi(z0:zf),work)
               end do
            endif
        end do
!$OMP   end do
        deallocate(work)
!$OMP   end parallel

        else ! MULTILAYER mean fields

        ns=t%nsubcoord
!$OMP   parallel default(shared) private(k,hk,mfparent,work2,r,z0,zf,p,f,mftmp,mfcurrent,work)
        allocate(work2(plen,t%navecs))
        allocate(work(plen))
!$OMP   do schedule(runtime)
        do k=1,kzahl(1) ! loop over oper products
            if(t%isuncor(k)) cycle
            if(t%allunit(k)) cycle ! skip 

            hk => t%hterms(k,:)
            if(t%parent%isuncor(k)) then
                ! MF are needed here but this k-term is separable one layer
                ! above. The MF above are identical to the density matrices.
                mfparent => t%parent%dmat(ns)%p%matrix
            else
                mfparent => t%parent%mfields(k,ns)%p%matrix
            end if
            ! Apply parent MF to list of A-vectors.
            work2 = CZERO
            do r=1,t%navecs
                z0 = t%z(r,1)
                zf = t%z(r,2)
                do p=1,t%navecs
                    work2(:,p) = work2(:,p) + mfparent(p,r)*psi(z0:zf)
                enddo
            enddo
            do f=1,t%nmodes ! loop over modes
                if(t%isunit(k,f)) cycle ! skip 
                childf => t%submodes(f)%p
                if (childf%isprimitive)then
                  if (complete(childf%m)) cycle ! skip for complete prim basis 
                endif

                allocate(mftmp(t%ndim(f),t%ndim(f)))
                mfcurrent => t%mfields(k,f)%p%matrix
                mfcurrent = CZERO
                do r=1,t%navecs
                    z0 = t%z(r,1)
                    zf = t%z(r,2)
                    work(:) = work2(:,r)
                    call hprod_phi(hk, work, f)
                    call compute_dmat(mftmp, f, t%ndim, psi(z0:zf), work)
                    mfcurrent = mfcurrent + mftmp
                enddo
                deallocate(mftmp)
            end do
        end do

!$OMP   end do
        deallocate(work)
        deallocate(work2)
!$OMP   end parallel

        end if
        call stop_timer(t%tid_mfields)

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if(debug) then
            do f=1,t%nmodes
                do k=1,kzahl(1)
                    if(t%isuncor(k)) cycle
                    mfcurrent => t%mfields(k,f)%p%matrix
                    write(6,*) 'mean-field: nodenum,f,k', t%num,f,k
                    do r=1,t%ndim(f)
                      write(6,10) (mfcurrent(r,p),p=1,t%ndim(f))
                    end do
                end do
            end do
        end if
        10 format(8(2x,'(',1PE11.4,1x,1PE11.4,')'))
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    end subroutine compute_mfields_node


    subroutine compute_dmat_node(t,psi,psi1)
    ! Compute density matrices using an equivalent expression
    ! to Eq.(36), Manthe,[JCP,128,164116,(2008)]. The calculation of the
    ! density matrices strongly resembles that of the mean-fields
        implicit none
        type(wfnode_t),intent(inout)  :: t
        complex(dop), intent(in)     :: psi(:)
        complex(dop), intent(in), optional :: psi1(:)
        complex(dop),pointer       :: dmparent(:,:),dmcurrent(:,:)
        complex(dop),pointer       :: idm(:,:),trfmat(:,:)
        real(dop),pointer           :: pop(:),regpop(:)
        complex(dop),allocatable   :: work2(:,:)
        complex(dop),allocatable   :: dmtmp(:,:)
        integer                     :: f,p,r,ns
        integer                     :: z0,zf,plen
        type(wfnode_t),pointer      :: childf

        if(t%isprimitive) return

        plen = product(t%ndim)

        if(t%istop) then ! top node; NORMAL MCTDH density matrix

        z0 = t%z(1,1)
        zf = t%z(1,2)
        do f=1,t%nmodes ! loop over modes
            childf => t%submodes(f)%p
            if (childf%tid_dmat == 0) then
                call init_node_timer(childf, childf%tid_dmat, 'dmat')
            endif
            call start_timer(childf%tid_dmat)
            dmcurrent => t%dmat(f)%p%matrix
            idm       => t%invdmat(f)%p%matrix
            pop       => t%pop(f)%p%rvector
            regpop    => t%regpop(f)%p%rvector
            trfmat    => t%trfmat(f)%p%matrix
            if (dentype .eq. 2) then
               call compute_d2mat(dmcurrent,f,t%ndim,psi(z0:zf),psi1(z0:zf))
            else
               call compute_dmat(dmcurrent,f,t%ndim,psi(z0:zf),psi1(z0:zf))
            endif
            call invdmat(dmcurrent,idm,trfmat,pop,regpop,epsmat)
            if (t%submodes(f)%p%isprimitive) then
                call cpdmat(pop,regpop,trfmat,size(pop),t%submodes(f)%p%m)
            end if
            call stop_timer(childf%tid_dmat)
        end do

        else ! MULTILAYER density matrix

        ns=t%nsubcoord ! which mode am I with respect to the parent?
        dmparent => t%parent%dmat(ns)%p%matrix
        allocate(work2(plen,t%navecs))
        work2 = CZERO
        if (present(psi1)) then
           do r=1,t%navecs
              z0 = t%z(r,1)
              zf = t%z(r,2)
              do p=1,t%navecs
                 work2(:,p) = work2(:,p) + dmparent(p,r)*psi1(z0:zf)
              enddo
           enddo
        else
           do r=1,t%navecs
              z0 = t%z(r,1)
              zf = t%z(r,2)
              do p=1,t%navecs
                 work2(:,p) = work2(:,p) + dmparent(p,r)*psi(z0:zf)
              enddo
           enddo
        endif
        do f=1,t%nmodes ! loop over modes
            childf => t%submodes(f)%p
            if (childf%tid_dmat == 0) then
                call init_node_timer(childf, childf%tid_dmat, 'dmat')
            endif
            call start_timer(childf%tid_dmat)
            dmcurrent => t%dmat(f)%p%matrix
            dmcurrent = CZERO
            idm       => t%invdmat(f)%p%matrix
            pop       => t%pop(f)%p%rvector
            regpop    => t%regpop(f)%p%rvector
            trfmat    => t%trfmat(f)%p%matrix
            allocate(dmtmp(t%ndim(f),t%ndim(f)))
            do p=1,t%navecs
               z0 = t%z(p,1)
               zf = t%z(p,2)
               call compute_dmat(dmtmp, f, t%ndim, psi(z0:zf), work2(:,p))
               dmcurrent = dmcurrent + dmtmp
            enddo
            deallocate(dmtmp)
            call invdmat(dmcurrent,idm,trfmat,pop,regpop,epsmat)
        !    if (t%submodes(f)%p%isprimitive) then
        !        call cpdmat(pop,regpop,trfmat,size(pop),t%submodes(f)%p%m)
        !    end if
            call stop_timer(childf%tid_dmat)
        end do

        deallocate(work2)

        end if
    end subroutine compute_dmat_node

    subroutine compute_mfsum_node(time,t)
    ! Compute (for one node) the sum over the product of mean-field matrix
    ! times operator term, and apply the inverse density matrix to this.
    ! Only diagonal operator terms are included in the sum.  This mostly
    ! makes sense for primitive nodes, so for others we currently don't
    ! bother.
    ! This summation is needed to speed up CMF calculations.
        implicit none
        real(dbl),intent(in)         :: time
        real(dbl)                    :: time1
        type(wfnode_t),intent(inout) :: t
        type(wfnode_t),pointer      :: tp
        complex(cdbl),pointer       :: mf(:,:)
        integer                     :: k,plen,ns,a,b,i,h
        complex(cdbl),allocatable   :: phi(:),hphi(:)
        complex(cdbl),allocatable   :: mfs(:,:,:)

! time argument is not used
        time1=time

        if (.not. t%isprimitive) return
        if (gwpm(t%m)) return
        if (t%tid_mfsum == 0) call init_node_timer(t, t%tid_mfsum, 'mfsum')
        call start_timer(t%tid_mfsum)

        plen = product(t%ndim)
        allocate(phi(plen))
        allocate(mfs(plen,t%navecs,t%navecs))

        ! mean-fields and inverse density matrices are stored at the parent
        tp => t%parent
        ns =  t%nsubcoord
        ! initialize arrays
        mfs = CZERO
        phi = CONE
!$OMP parallel default(shared) private(k,hphi,mf,a,b) reduction(+:mfs)
        allocate(hphi(plen))
!$OMP do schedule(runtime)

        do k=1,kzahl(1)
            ! skip non-diagonal terms
            if (t%htmtype(k) < 0) cycle
            ! evaluate the diagonal hterm
            call hcall(t%m,k,hphi,phi,plen)
            ! apply coefficient (if not 1)
            h=hamilton(fkoe,k)
            if (t%isuncor(k)) then
                ! uncorrelated part
                do a=1,t%navecs
                   do i=1,plen
                      mfs(i,a,a) = mfs(i,a,a) + hphi(i)
                   end do
                end do
            else
                ! correlated part
                mf => tp%mfields(k,ns)%p%matrix
                do b=1,t%navecs
                   do a=1,t%navecs
                      do i=1,plen
                         mfs(i,a,b) = mfs(i,a,b) + mf(a,b)*hphi(i)
                      end do
                   end do
                end do
            end if
        end do
!$OMP end do
        deallocate(hphi)
!$OMP end parallel

        ! add MLOp contributions
        call compute_all_mlop_mfsum(t,phi,mfs)

        ! store
        do b=1,t%navecs
            do a=1,t%navecs
                t%mfsum(a,b)%p%vector(:) = mfs(:,a,b)
            end do
        end do

        ! clean up
        deallocate(mfs)
        deallocate(phi)
        call stop_timer(t%tid_mfsum)
    end subroutine compute_mfsum_node

    subroutine cpdmat(pop,rpop,trfmat,l,m)
        implicit none
        real(dbl),intent(in)     :: pop(l)
        real(dbl),intent(in)     :: rpop(l)
        complex(cdbl),intent(in) :: trfmat(l*l)
        integer,intent(in)       :: l
        integer,intent(in)       :: m
!        real(dbl)                :: work(2*l)

!        work(1:l)     = pop
!        work(l+1:2*l) = rpop
!        call cpvxd(work,mr(mrdicht3+d3mat(m,1)-1),2*l)
!        call cpvxz(trfmat,dicht4(dmat(m,1)),l*l)
! copy nat pops and regularized natpops to dicht3
        dicht3(d3mat(m,1):d3mat(m,1)-1+l) = pop(1:l)
        dicht3(d3mat(m,1)+l:d3mat(m,1)-1+2*l) = rpop(1:l)
! copy nat orbs to dicht4
        dicht4(dmat(m,1):dmat(m,1)-1+l*l) = trfmat(1:l*l)

    end subroutine cpdmat

    subroutine projectout_psi(this,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(in)    :: this
        complex(dop),intent(in)     :: psi(:)
        complex(dop),intent(inout)  :: dtpsi(:)
        real(dop)                   :: norm2,energy
        integer                      :: z0,zf
        ! Return  H|Psi> - |Psi><Psi|H|Psi> which is equivalent
        ! to (H-E)|Psi>
        ! This modified derivative keeps the normalization of the
        ! top layer A vector, and therefore of the whole wavefunction,
        ! during relaxation runs
        !
        ! this  : top tree node
        ! psi   : complete psi array
        ! dtpsi : complete dtpsi array
        z0 = this%z(1,1)
        zf = this%z(1,2)
        norm2  = dble(dot_product(psi(z0:zf),psi(z0:zf)))
        energy = dble(dot_product(psi(z0:zf),dtpsi(z0:zf)))
        energy = energy/norm2
        dtpsi(z0:zf) = dtpsi(z0:zf) - energy*psi(z0:zf)
    end subroutine projectout_psi


    subroutine ecorrml(time,t,psi,ecorr)
    !   Compute E = <A|H|A>
        implicit none
        type(wfnode_t),intent(in)   :: t
        complex(dop),intent(in)     :: psi(:)
        complex(dop),intent(out)    :: ecorr
        type(hterm_tp),pointer      :: hk(:)
        complex(dop),allocatable    :: work(:)
        complex(dop)                :: hpsi(product(t%ndim))
        integer                     :: k,z0,zf,plen,nham
        real(dop),intent(in)        :: time

        if(.not.t%istop.or.t%isprimitive) then
            write(6,*) 'Etot can only be computed for the top node'
            stop
        end if

        plen = product(t%ndim)

        hpsi = CZERO
        z0 = t%z(1,1)
        zf = t%z(1,2)
        !$omp parallel default(shared) private(k,work,hk) reduction(+:hpsi)
        allocate(work(product(t%ndim)))
        !$omp do
        do k=1,kzahl(1) ! loop over k-terms
            hk=>t%hterms(k,:)
            work = psi(z0:zf)
            call hprod_phi(hk,work)
            hpsi = hpsi + koeff(k)*work
        end do
        !$omp end do
        deallocate(work)
        !$omp end parallel

        ecorr = dot_product(psi(z0:zf),hpsi)

        ! Add contribution from MLOPs.
        nham=1
        ecorr = ecorr + compute_all_mlop_etot(time,nham)

    end subroutine ecorrml

    subroutine euncorrml(t,psi,euncorr)
    !   Compute E = sum <A|h_kappa|A>
        implicit none
        type(wfnode_t),intent(in)   :: t
        complex(dop),intent(in)     :: psi(:)
        complex(dop),intent(out)    :: euncorr
        complex(dop)                :: work(product(t%ndim))
        complex(dop)                :: hpsi(product(t%ndim))
        integer                     :: z0,zf,plen,f

        if(.not.t%istop.or.t%isprimitive) then
            write(6,*) 'Etot can only be computed for the top node'
            stop
        end if

        plen = product(t%ndim)

        hpsi = CZERO
        z0 = t%z(1,1)
        zf = t%z(1,2)
        do f=1,t%nmodes
           if (.not. t%hasuncor(f,1)) cycle

            work = psi(z0:zf)
            call hunprod_phi(t%hterms(kzahl(1)+1,:),work,f)
            hpsi = hpsi + work
        enddo

        euncorr = dot_product(psi(z0:zf),hpsi)
    end subroutine euncorrml

subroutine projector_ml(psi,dtpsi,nvecs,plen)
    use mlbase
    use timing
    implicit none
    integer,intent(in)        ::   nvecs
    integer,intent(in)        ::   plen
    complex(cdbl),intent(in)  ::   psi(plen,nvecs)
    complex(cdbl),intent(out) ::   dtpsi(plen,nvecs)
    complex(cdbl)             ::   overlap
    integer                   ::   e,e1
    integer,save              ::   tid=0

    call get_timer(tid,'projector_ml')
    call start_timer(tid)
!$OMP parallel default(shared) private(e,e1,overlap)
!$OMP do schedule(runtime)
    do e = 1,nvecs
       do e1 = 1,nvecs
          overlap = dot_product(psi(:,e1),dtpsi(:,e))
          dtpsi(:,e) = dtpsi(:,e) - overlap*psi(:,e1)
       enddo
    enddo
!$OMP end do
!$OMP do schedule(runtime)
    do e = 1,nvecs
       do e1 = 1,nvecs
          overlap = dot_product(psi(:,e1),dtpsi(:,e))
          dtpsi(:,e) = dtpsi(:,e) - overlap*psi(:,e1)
       enddo
    enddo
!$OMP end do
!$OMP end parallel
    call stop_timer(tid)
end subroutine projector_ml


end module eqofmotion1
