!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: eqofmotion.f90 518 2013-11-05 18:32:34Z frank $


module eqofmotion1

    use decimal
    use global
    use globmemmod
    use wftree
    use timing

    use mltypes
    use operdef
    use dvrdatmod, only: fkoe
    use tdkoeffs, only: calctdk

    private
    public  :: compute_opsi

    logical,save   :: debug = .false.
    real(dop),save :: eps_inv = 1.0d-8

    contains


    subroutine compute_opsi_tree(this,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout) :: this
        complex(dop)                :: psi(:),dtpsi(:)
        integer                  :: i,z0,zf
        type(wfnode_t),pointer   :: cnode
        integer, save            :: tid = 0 ! timer ID
        call get_timer(tid, 'opsi_tree')
        call start_timer(tid)
        do i=1,this%tot_nodes
            cnode => this%swap_down(i)%p
            z0 = cnode%z(1,1)
            zf = cnode%z(cnode%navecs,2)
            call compute_opsi_node(cnode,psi(z0:zf),dtpsi(z0:zf))
        end do
        call stop_timer(tid)
    end subroutine compute_opsi_tree


    subroutine compute_opsi_node(this,psi,dtpsi)
        implicit none
        type(wfnode_t),intent(inout)       :: this
        complex(dop),intent(in)           :: psi(:)
        complex(dop),intent(out)          :: dtpsi(:)
        type(hterm_tp),pointer       :: hk(:)
        complex(dop),pointer        :: work(:)
        integer                      :: k,plen,totplen

        if (this%tid_opsi == 0) then
            call init_node_timer(this, this%tid_opsi, 'opsi')
        endif
        call start_timer(this%tid_opsi)

        plen = product(this%ndim)
        totplen=plen*this%navecs
        dtpsi = CZERO

        if(this%istop) then

            if(this%isprimitive) then  ! exact calc. using ML code

                work => this%topnode%workc(1:plen)
                do k=1,khtot(1) ! loop over k-terms
                    call hcall(this%m,k,work,psi,plen)
                    call koeff_ml(k,0.d0,plen,work)
                    dtpsi = dtpsi + work
                end do
                work => null()

            else

                work => this%topnode%workc(1:plen)
                do k=1,khtot(1) ! loop over k-terms
                    hk=>this%hterms(k,:)
                    work = psi
                    call hprod_phi(hk,work)
                    call koeff_ml(k,0.d0,plen,work)
                    dtpsi = dtpsi + work
                end do
                work => null()

            end if

        end if
        call stop_timer(this%tid_opsi)

    end subroutine compute_opsi_node


    subroutine compute_opsi(this,psi,dtpsi)
        ! Constructs the hterms at all layers for some operator and applies the
        ! matrix representation to the top layer A vector.  This routine is used
        ! in expectation value or cross-correlation evaluations
        implicit none
        type(wfnode_t),intent(inout)       :: this
        complex(dop),intent(in)           :: psi(:)
        complex(dop),intent(out)          :: dtpsi(:)

        call compute_hterms_tree(0.d0,this,psi,1)
        call compute_opsi_tree(this,psi,dtpsi)
    end subroutine compute_opsi

    subroutine koeff_ml(k,time,l,psi)
        implicit none
        integer,intent(in)          :: k
        real(dbl),intent(in)        :: time
        integer,intent(in)          :: l
        complex(cdbl),intent(inout) :: psi(l)
        integer                     :: ktype,h,nham
        complex(cdbl)               :: cz

        h=hamilton(fkoe,k) ! Inquire type of coefficient
        ktype=htmtype(h)

        if(ktype .eq. 101) then  ! TD coefficient
           nham=1
           call calctdk(time,nham,psi)
        end if

        cz = koeff(k)        ! Read koeff value

        if(ktype/=1) then
            psi = psi * cz
        end if
    end subroutine koeff_ml


end module eqofmotion1
