!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: linear.f90 508 2013-07-30 13:28:52Z dieter $

module linear

    use decimal
    use mltypes
    use global
    use mlbase
    use mltensor
    use d2tens
    use logdat
    use runpropmod, only: lmf

    private
    public  :: hprod_phi,          &
               hprod_phi2,         &
               hunprod_phi,        &
               hunprod_phi2,       &
               trhprod_phi,        &
               trhprod_phi2,       &
               trhunprod_phi,      &
               chprod_phi2,        &
               phi_hprod_phi,      &
               phi_phi,            &
               phi_ext_phi,        &
               ijk_shape,          &
               herm_matrix_invert, &
               herm_matrix_diag,   &
               compute_dmat,       &
               compute_d2mat,      &
               intgr_dens,         &
               invdmat,            &
               proj_simple,        &
               proj_inv,           &
               matprod_tensor,     &
               decontract_index,   &
               gramschmidt,        &
               dgramschmidt2,      &
               zqr


    contains

    subroutine hprod_phi(hl,phi,fhole)  ! ha.hb...|P>
        ! Apply hterms in the hl list to phi successively.
        ! Each term in hl is an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fhole can be specified in order to skip one of the degrees
        ! of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in),optional           :: fhole

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l,fh
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        ! -> here the assertion philen==product(ndim) could (should) be tested

        fh=0
        if(present(fhole)) fh=fhole
        do l=1,nmodes
            if(l==fh) cycle
            select case(hl(l)%p%htype)
                case(1)
                    cycle ! phi is unchanged
                case(2)
                    ijk=ijk_shape(l,ndim)
                    call matrix_tensor_compl(hl(l)%p%matrix,phi,ijk)
                case(3)
                    ijk=ijk_shape(l,ndim)
                    call vector_tensor_compl(hl(l)%p%vector,phi,ijk)
                case(4)
                    ijk=ijk_shape(l,ndim)
                    call dmatrix_tensor_compl(hl(l)%p%rmatrix,phi,ijk)
                case(5)
                    ijk=ijk_shape(l,ndim)
                    call dvector_tensor_compl(hl(l)%p%rvector,phi,ijk)
                case default
                    write(6,*) 'module: linear'
                    write(6,*) 'hterm of type ',hl(l)%p%htype, ' cannot be handled'
                    stop
            end select
        end do
    end subroutine hprod_phi

    subroutine hprod_phi2(hl,phi,fhole,side)  ! ha.hb...|P>
        ! Apply hterms in the hl list to phi successively.
        ! Special case of density(II) top layer.
        ! Each term in hl is an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fhole can be specified in order to skip one of the degrees
        ! of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in),optional           :: fhole,side

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l,fh,plen,fs
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        plen = product(ndim)
        ! -> here the assertion philen==product(ndim) could (should) be tested

        fh=0
        if(present(fhole)) fh=fhole
! side = 1 acts RHS
! side = 2 acts LHS
        fs=1
        if(present(side)) fs=side
        do l=1,nmodes
            if(l==fh) cycle
            ijk=ijk_shape(l,ndim)
            if (fs .eq. 1) then
               ijk(3) = ijk(3)*plen
            else if (fs .eq. 2) then
               ijk(1) = ijk(1)*plen
            endif
            select case(hl(l)%p%htype)
                case(1)
                    cycle
                case(2)
                    call matrix_tensor_compl(hl(l)%p%matrix,phi,ijk)
                case(3)
                    call vector_tensor_compl(hl(l)%p%vector,phi,ijk)
                case(4)
                    call dmatrix_tensor_compl(hl(l)%p%rmatrix,phi,ijk)
                case(5)
                    call dvector_tensor_compl(hl(l)%p%rvector,phi,ijk)
                case default
                    write(6,*) 'module: linear'
                    write(6,*) 'hterm of type ',hl(l)%p%htype, ' cannot be handled'
                    stop
            end select
        end do
    end subroutine hprod_phi2

    subroutine hunprod_phi(hl,phi,fh)  ! hf|P>
        ! Apply uncorrelated hterm to phi.
        ! Special case of density(II) top layer.
        ! A term in hl is an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fh defines the degrees of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in)                    :: fh

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        ! -> here the assertion philen==product(ndim) could (should) be tested

        select case(hl(fh)%p%htype)
            case(1)
                continue
            case(2)
                ijk=ijk_shape(fh,ndim)
                call matrix_tensor_compl(hl(fh)%p%matrix,phi,ijk)
            case(3)
                ijk=ijk_shape(l,ndim)
                call vector_tensor_compl(hl(fh)%p%vector,phi,ijk)
            case default
                write(6,*) 'module: linear'
                write(6,*) 'hterm of type ',hl(fh)%p%htype, ' cannot be handled'
                stop
        end select
    end subroutine hunprod_phi

    subroutine hunprod_phi2(hl,phi,fh)  ! hf|P>
        ! Apply uncorrelated hterm to phi.
        ! A term in hl is an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fh defines the degrees of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in)                    :: fh

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l,plen
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        plen = product(ndim)
        ! -> here the assertion philen==product(ndim) could (should) be tested

        select case(hl(fh)%p%htype)
            case(1)
                continue
            case(2)
                ijk=ijk_shape(fh,ndim)
                ijk(3) = ijk(3)*plen
                call matrix_tensor_compl(hl(fh)%p%matrix,phi,ijk)
            case(3)
                ijk=ijk_shape(l,ndim)
                ijk(3) = ijk(3)*plen
                call vector_tensor_compl(hl(fh)%p%vector,phi,ijk)
            case default
                write(6,*) 'module: linear'
                write(6,*) 'hterm of type ',hl(fh)%p%htype, ' cannot be handled'
                stop
        end select
    end subroutine hunprod_phi2

    subroutine trhprod_phi(hl,phi,fhole)  ! ha.hb...|P>
        ! Apply hterms in the hl list to phi successively.
        ! Each term in hl is the transpose of an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fhole can be specified in order to skip one of the degrees
        ! of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in),optional           :: fhole

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l,fh
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        ! -> here the assertion philen==product(ndim) could (should) be tested

        fh=0
        if(present(fhole)) fh=fhole
        do l=1,nmodes
            if(l==fh) cycle
            select case(hl(l)%p%htype)
                case(1)
                    cycle
                case(2)
                    ijk=ijk_shape(l,ndim)
                    call trmatrix_tensor_compl(hl(l)%p%matrix,phi,ijk)
                case(3)
                    ijk=ijk_shape(l,ndim)
                    call vector_tensor_compl(hl(l)%p%vector,phi,ijk)
                case(4)
                    ijk=ijk_shape(l,ndim)
                    call dtrmatrix_tensor_compl(hl(l)%p%rmatrix,phi,ijk)
                case(5)
                    ijk=ijk_shape(l,ndim)
                    call dvector_tensor_compl(hl(l)%p%rvector,phi,ijk)
                case default
                    write(6,*) 'module: linear'
                    write(6,*) 'hterm of type ',hl(l)%p%htype, ' cannot be handled'
                    stop
            end select
        end do
    end subroutine trhprod_phi

    subroutine trhunprod_phi(hl,phi,fh)  ! hf|P>
        ! Apply uncorrelated hterm to phi.
        ! A term in hl is the transpose of an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fh defines the degrees of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in)                    :: fh

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        ! -> here the assertion philen==product(ndim) could (should) be tested

        select case(hl(fh)%p%htype)
            case(1)
                continue
            case(2)
                ijk=ijk_shape(fh,ndim)
                call trmatrix_tensor_compl(hl(fh)%p%matrix,phi,ijk)
            case(3)
                ijk=ijk_shape(l,ndim)
                call vector_tensor_compl(hl(fh)%p%vector,phi,ijk)
            case default
                write(6,*) 'module: linear'
                write(6,*) 'hterm of type ',hl(fh)%p%htype, ' cannot be handled'
                stop
        end select
    end subroutine trhunprod_phi

    subroutine chprod_phi2(hl,phi,fhole)  ! ha.hb...|P>
        ! Apply hterms in the hl list to phi successively.
        ! Special case of density(II) top layer:
        !          Adj of matrix acting on LHS.
        ! Each term in hl is the complex conjugate of an operator (matrix) that
        ! operates on one of the dimensions of phi.
        ! Therefore this is just a matrix-vector multiplication in
        ! which the operator factorizes into a product of operators
        ! and the tensorial structure is used.
        ! The result is accumulated in phi, i.e. phi is modified in place.
        ! fhole can be specified in order to skip one of the degrees
        ! of freedom.
        implicit none
        type(hterm_tp),intent(in)             :: hl(:)
        complex(cdbl),intent(inout)           :: phi(:)
        integer,intent(in),optional           :: fhole

        integer                 :: ndim(size(hl)) ! automatic array
        integer                 :: nmodes,philen,l,fh,plen
        integer                 :: ijk(3)

        ! Note: nmodes and ndim could be passed as arguments since they are
        ! known (or can be computed) already in the calling units. We re-compute
        ! them here to keep a cleaner inteface. This can be changed if it turns
        ! out to produce noticeable overhead.
        philen=size(phi)
        nmodes=size(hl)
        do l=1,nmodes
            ndim(l)=hl(l)%p%hrange
        end do
        plen = product(ndim)
        ! -> here the assertion philen==product(ndim) could (should) be tested

        fh=0
        if(present(fhole)) fh=fhole
        do l=1,nmodes
            if(l==fh) cycle
            select case(hl(l)%p%htype)
                case(1)
                    cycle
                case(2)
                    ijk=ijk_shape(l,ndim)
                    ijk(1) = ijk(1)*plen
                    call cmatrix_tensor_compl(hl(l)%p%matrix,phi,ijk)
                case(3)
                    ijk=ijk_shape(l,ndim)
                    ijk(1) = ijk(1)*plen
                    call cvector_tensor_compl(hl(l)%p%vector,phi,ijk)
                case(4)
                    ijk=ijk_shape(l,ndim)
                    ijk(1) = ijk(1)*plen
                    call dmatrix_tensor_compl(hl(l)%p%rmatrix,phi,ijk)
                case(5)
                    ijk=ijk_shape(l,ndim)
                    ijk(1) = ijk(1)*plen
                    call dvector_tensor_compl(hl(l)%p%rvector,phi,ijk)
                case default
                    write(6,*) 'module: linear'
                    write(6,*) 'hterm of type ',hl(l)%p%htype, ' cannot be handled'
                    stop
            end select
        end do
    end subroutine chprod_phi2


    function phi_hprod_phi(p1,hl,p2) result(r)  ! <P1|ha.hb...|P2>
        implicit none
        complex(cdbl),intent(in)             :: p1(:),p2(:)
        type(hterm_tp),intent(in)            :: hl(:)
        complex(cdbl)                        :: r
        complex(cdbl)                        :: work(size(p1))
        work=p2
        call hprod_phi(hl,work)
        r=phi_phi(p1,work)
    end function phi_hprod_phi


    function phi_phi(p1,p2,cp) result(r)  ! <P1|P2>
        implicit none
        complex(cdbl),intent(in)    ::  p1(:),p2(:)
        complex(cdbl)               ::  r
        integer,intent(in),optional ::  cp ! if 1, complex symmetric product
        if(.not.present(cp)) then
            r = dot_product(p1,p2) ! intrinsic dot_product takes the usual conj(p1)*p2 of quantum mechanics
        else
            if(cp.eq.1) then
                r = dot_product(dconjg(p1),p2)
            else
                r = dot_product(p1,p2)
            end if
        end if
    end function phi_phi


    subroutine phi_ext_phi(p1,p2,mat)  ! |P1><P2|
        implicit none
        complex(cdbl),intent(in)    :: p1(:),p2(:)
        complex(cdbl),intent(out)   :: mat(:,:)
        integer               :: n1,n2,i,j
        n1=size(p1)
        n2=size(p2)
        mat=(0.d0,0.d0)
        do i=1,n1
            do j=1,n2
                mat(i,j) = p1(i)*dconjg(p2(j))
            end do
        end do
    end subroutine phi_ext_phi


    subroutine compute_dmat(dmat,k,ndim,av1,av2)
        ! Compute the density-matrix for the general
        ! case of two different A-vectors for degree of freedom (mode) k
        !
        ! dmat: output, the density matrix  => dmat(ndim(k),ndim(k))
        ! k: degree of freedom over which the density is computed
        ! ndim: shape of the a-vectors (integer array)
        ! av1,av2: a-vectors, rank-1 arrays => size(avX).eq.product(ndim)
        implicit none

        complex(cdbl),intent(out)         :: dmat(:,:)
        integer,intent(in)                :: k,ndim(:)
        complex(cdbl),intent(in)          :: av1(:)
        complex(cdbl),intent(in),optional :: av2(:)
        integer                           :: ijk(3)

        ijk=ijk_shape(k,ndim)

        if(present(av2)) then
            call dmat_tensor_tensor_compl(dmat,av1,av2,ijk)
        else
            call dmat_tensor_tensor_compl(dmat,av1,av1,ijk)
        end if

    end subroutine compute_dmat

    subroutine compute_d2mat(dmat,k,ndim,av1,av2)
        ! Compute the density-matrix for the general
        ! case of two different A-vectors for degree of freedom (mode) k
        !
        ! dmat: output, the density matrix  => dmat(ndim(k),ndim(k))
        ! k: degree of freedom over which the density is computed
        ! ndim: shape of the a-vectors (integer array)
        ! av1,av2: a-vectors, rank-1 arrays => size(avX).eq.product(ndim)
        implicit none

        complex(cdbl),intent(out)         :: dmat(:,:)
        integer,intent(in)                :: k,ndim(:)
        complex(cdbl),intent(in)          :: av1(:)
        complex(cdbl),intent(in),optional :: av2(:)
        integer                           :: ijk(3),plen

        ijk=ijk_shape(k,ndim)

        if(present(av2)) then
          routine = 'Compute_dmat (linear.f90) '
          message = 'Density (II) not coded ....'
          call errormsg
        endif
!     mean-field approach
        if (lmf) then
           call rm1_mf(av1,dmat,ijk(1),ijk(2),ijk(3))
        else
!     Dirac-Frenkel variational principle
           plen=ijk(1)*ijk(2)*ijk(3)
           call rm2_mf(av1,dmat,ijk(1),ijk(2),ijk(3),plen)
           call filldens(dmat,ijk(2))
        endif

    end subroutine compute_d2mat


    subroutine intgr_dens(d1d,k,ndim,dens)
        ! Integrate a multidimensional density function for all except one
        ! coordinate. Returns a 1D density
        !
        ! d1d: output 1D density
        ! k: degree of freedom over which the density is computed
        ! ndim: shape of the a-vectors (integer array)
        ! dens: input multidimensional density
        implicit none

        real(dbl),intent(out)              :: d1d(:)
        integer,intent(in)                 :: k,ndim(:)
        real(dbl),intent(in)               :: dens(:)
        integer                            :: ijk(3)

        ijk=ijk_shape(k,ndim)
        call dens_tensor_real(d1d,dens,ijk)
    end subroutine intgr_dens


    subroutine matprod_tensor(ml,tens1,tens2)
        ! ml    - a list of rectangular matrices.
        ! tens1 - complex array, its shape conforms to the lengths of the rows
        !         of the matrices of ml
        ! tens2 - complex array, its shape conforms to the lengths of the columns
        !         of the matrices of ml
        !
        ! The matrices in ml are successively multiplied to the different
        ! indices of tens1. The result of all operations is finally stored in
        ! tens2.
        ! Note: this routine is similar to hprod_phi, which however corresponds
        ! to the multiplication with a list of square matrices and can be coded
        ! more efficiently. This is because here the shape and size of the
        ! output array changes at each cycle and a lot of allocation and
        ! deallocation becomes necessary.
        ! TODO: this routine can probably be optimized by only reallocating the
        ! work arrays when their current size is not sufficient, and keeping
        ! track of their actual used size.
        implicit none
        type(hterm_tp),intent(in)        :: ml(:)
        complex(cdbl),intent(in)         :: tens1(:)
        complex(cdbl),intent(out)        :: tens2(:)

        integer                  :: ndim1(size(ml))
        integer                  :: ndim2(size(ml))
        integer                  :: ntmp1(size(ml))
        integer                  :: ntmp2(size(ml))
        integer                  :: nindex,i,ijk(3),plen
        complex(cdbl),allocatable,target :: work1(:),work2(:)

        nindex = size(ml)
        do i=1,nindex
            ndim1(i) = size(ml(i)%p%matrix,2)
            ndim2(i) = size(ml(i)%p%matrix,1)
        end do

        allocate(work1(product(ndim1)))
        work1 = tens1

        ntmp2 = ndim1
        do i=1,nindex
            ntmp1 = ntmp2
            ntmp2(i) = ndim2(i)
            plen = product(ntmp2)
            allocate(work2(plen))
            ijk = ijk_shape(i,ntmp1)

            call recmatrix_tensor_compl(ml(i)%p%matrix,work1,work2,ijk,ndim2(i))

            deallocate(work1)
            allocate(work1(plen))
            work1 = work2
            deallocate(work2)
        end do

        tens2 = work1
        deallocate(work1)
    end subroutine matprod_tensor


    subroutine decontract_index(anew,aold,dimnew,dimold,spfs,i)
        ! Decontract spfs expansion in index i of aold.
        !
        ! New coefficients are stored in anew
        ! anew: array conformable to dimnew
        ! aold: array conformable to dimold
        ! spfs: spfs vectors, conformable to (dimold(i),dimnew(i))
        ! i: transformed index
        complex(cdbl),intent(out)       :: anew(:)
        complex(cdbl),intent(in)        :: aold(:)
        integer,intent(in)              :: dimnew(:),dimold(:),i
        complex(cdbl),intent(in)        :: spfs(:)
        integer                         :: ijk(3),lnew

        ijk = ijk_shape(i,dimold)
        lnew = dimnew(i)
        call recmatrix_tensor_compl(spfs,aold,anew,ijk,lnew)
    end subroutine decontract_index


    subroutine invdmat(dmat,inv,evecs,evals,evalsreg,epsrel)
        implicit none
        complex(cdbl),intent(in)      :: dmat(:,:)
        complex(cdbl),intent(out)     :: inv(:,:)
        complex(cdbl),intent(out)     :: evecs(:,:)
        real(dbl),intent(out)         :: evals(:)
        real(dbl),intent(out)         :: evalsreg(:)
        real(dbl),intent(in)          :: epsrel

        real(dbl)                     :: evalstmp(size(dmat,1),2)
        real(cdbl)                    :: work(3*size(dmat,1))
        integer                       :: l

        l=size(dmat,1)
        evecs = -1.0d0*dmat
        call matinv_ml(inv,evalstmp,evecs,l,epsrel,work)
        evals    = evalstmp(:,1)
        evalsreg = evalstmp(:,2)
    end subroutine invdmat


    subroutine gramschmidt(vecs,veclen,vecnum,ierr)
        implicit none
        complex(cdbl),intent(inout) :: vecs(:)
        integer,intent(in)          :: veclen
        integer,intent(in)          :: vecnum
        integer,intent(out)         :: ierr
        integer                     :: i,j,z0i,zfi,z0j,zfj
        real(dbl)                   :: norm
        ! Gram-Schmidt othogonalization of vecs in place
        ! vecs contains the vectors one after the other,
        ! the total length of vecs is veclen*vecnum
        ierr = 0

        do i=1,vecnum
            z0i = (i-1)*veclen + 1
            zfi = z0i + veclen - 1
            do j=1,i-1
                z0j = (j-1)*veclen + 1
                zfj = z0j + veclen - 1
                vecs(z0i:zfi) = vecs(z0i:zfi) - &
                       dot_product(vecs(z0j:zfj),vecs(z0i:zfi))*vecs(z0j:zfj)
            end do
            norm = sqrt( dble( dot_product(vecs(z0i:zfi),vecs(z0i:zfi)) ) )
            if (norm .le. 1.0d-99 ) norm = 1.0d99
            vecs(z0i:zfi) = vecs(z0i:zfi)/norm
        end do

        do i=1,vecnum
            z0i = (i-1)*veclen + 1
            zfi = z0i + veclen - 1
            do j=1,i-1
                z0j = (j-1)*veclen + 1
                zfj = z0j + veclen - 1
                vecs(z0i:zfi) = vecs(z0i:zfi) - &
                       dot_product(vecs(z0j:zfj),vecs(z0i:zfi))*vecs(z0j:zfj)
            end do
            norm = sqrt( dble(dot_product(vecs(z0i:zfi),vecs(z0i:zfi)) ))
            if( norm .le. 0.8d0 ) then
                ierr = i
                return
            end if
            vecs(z0i:zfi) = vecs(z0i:zfi)/norm
        end do
    end subroutine gramschmidt

    ! Similar to gramschmidt but for real
    ! vectors. Also removes
    ! all linear dependent vectors.
    ! returns vecnum as the number of
    ! as the new number of linear independent
    ! vectors
    subroutine dgramschmidt2(vecs,veclen,vecnum, ierr)
      implicit none
      integer,intent(IN)          :: veclen
      integer,intent(INOUT)       :: vecnum
      integer, intent(OUT)        :: ierr
      real(dbl),intent(inout)     :: vecs(veclen,vecnum)
      logical                     :: ldelete(vecnum)
      real(dbl)                   :: norm, ovlp, ddot, dnrm2
      integer                     :: i,j
      ! Gram-Schmidt othogonalization of vecs in place

      ! first round, mark
      ! linear dependents
      ldelete = .false.
      do i=1,vecnum
         do j=1,i-1
            if (.not. ldelete(j)) then
               ovlp = ddot(veclen,vecs(:,j),1,vecs(:,i),1)
               vecs(:,i) = vecs(:,i) - ovlp * vecs(:,j)
            end if
         end do

         norm = dnrm2(veclen,vecs(:,i),1)
         if (norm .le. 1.0d-16 ) then
            ldelete(i) = .true.
            vecs(:,i) = 0.d0
         else
            vecs(:,i) = vecs(:,i)/norm
         end if
      end do

      ! remove
      do while (any(ldelete))
         do i=1,vecnum
            if (ldelete(i)) then
               do j=i+1, vecnum
                  ldelete(j-1) = ldelete(j)
                  vecs(:,j-1) = vecs(:,j)
               end do
               ldelete(vecnum) = .false.
               vecs(:,vecnum) = 0.d0
               vecnum = vecnum - 1
               exit ! i-loop
            end if
         end do
      end do

      ierr = 0
      do i=1,vecnum
         do j=1,i-1
            ovlp = ddot(veclen,vecs(:,j),1,vecs(:,i),1)
            vecs(:,i) = vecs(:,i) - ovlp * vecs(:,j)
         end do

         norm = dnrm2(veclen,vecs(:,i),1)
         if (norm .le. 1.0d-16 ) then
            ierr = i
            return
         else
            vecs(:,i) = vecs(:,i)/norm
         end if
      end do

    end subroutine dgramschmidt2


    !-------------------------------------------------------------------
    subroutine zqr(m,n,q,r,info)
    !-------------------------------------------------------------------
    ! Convenience routine for computing the QR decomposition of an
    ! mxn-matrix A. On entry, Q=A. On exit, Q is a mxn-matrix with
    ! orthonormal columns, and R is an upper triangular nxn-matrix,
    ! such that A=QR. On success, info==0.
    !-------------------------------------------------------------------
       integer,intent(in)                 :: m,n
       complex(cdbl),intent(inout)        :: q(m,n)
       complex(cdbl),intent(out),optional :: r(n,n)
       integer,intent(out)                :: info
       complex(cdbl),allocatable          :: tau(:)
       complex(cdbl),allocatable          :: work(:)
       complex(cdbl)                      :: cwork(1)
       integer                            :: mn,lwork,i,j

       mn = min(m,n)
       allocate(tau(mn))
       ! Compute QR factorization, LAPACK style.
       call ZGEQRF(m,n,q,m,tau,cwork,-1,info) ! workspace query
       lwork = int(cwork(1))
       allocate(work(lwork))
       call ZGEQRF(m,n,q,m,tau,work,lwork,info)
       if (info /= 0) return
       ! Copy upper triangular part to R.
       if (present(r)) then
          r = CZERO
          do j=1,n
             do i=1,min(j,mn)
                r(i,j) = q(i,j)
             enddo
          enddo
       endif
       ! Build the matrix Q.
       call ZUNGQR(m,n,mn,q,m,tau,cwork,-1,info) ! workspace query
       lwork = int(cwork(1))
       if (lwork > size(work)) then
          deallocate(work)
          allocate(work(lwork))
       endif
       call ZUNGQR(m,n,mn,q,m,tau,work,lwork,info)
       ! Clean up.
       deallocate(work)
       deallocate(tau)
    end subroutine zqr



    function ijk_shape(l,ndim)
        implicit none
        integer :: l
        integer :: ndim(:)
        integer :: ijk_shape(3)
        integer :: nmodes
        nmodes=size(ndim)
        ijk_shape=(/1,1,1/)
        ijk_shape(2)=ndim(l)
        if(l>1)      ijk_shape(1)=product(ndim(1:l-1))
        if(l<nmodes) ijk_shape(3)=product(ndim(l+1:nmodes))
    end function ijk_shape


!#######################################################################
!
! MATINV
!
! called by invdmat (which is called by compute_dmat_node)
! calls to zheev
!
! calls diagonalisation routines for density matrix and from the
!  eigenvalues (regularised) and eigenvectors so obtained calculates
!  the inverse density matrix
!
! invers: returned as inverted matrix, used as scratch array.
! eval:   eigenvalues of passed matrix
! evek:   eigenvectors of passed matrix; on input: Matrix to be inverted.
! l:      dimensions of matrices e.g. invers(l,l)
! work:   complex workspace
!
!#######################################################################
    subroutine matinv_ml(invers,eval,evek,l,epsrel,work)
        implicit none
        integer       :: error,e,e1,e2,l,ichn
        complex(cdbl) :: invers(l,l),evek(l,l)
        real(dbl)     :: eval(l,2),epsrel,work(3*l)

!-----------------------------------------------------------------------
! diagonalize matrix
!-----------------------------------------------------------------------
        e2 = l*l
        call zheev('V','U',l,evek,l,eval,invers,e2,work,error)
        if (error .ne. 0) then
           ichn=6
 100       write(ichn,*)
           write(ichn,*) '###################################################'
           write(ichn,*) 'matinv_ml: error diagonalizing density matrix'
           write(ichn,*) '  dim =',l
           write(ichn,*) 'error =',error
           write(ichn,*) 'This error may result from a too inaccurate integrator setting.'
           write(ichn,*) '###################################################'
           ichn=ichn-4
           if(ichn > 0) goto 100
           stop 1
        endif

!-----------------------------------------------------------------------
! take negative of eigenvalue (due to re-ordering) and regularize.
! The if-statement is introduced to avoid an underflow.
!-----------------------------------------------------------------------
        do e=1,l
           eval(e,1)=-eval(e,1)
           If(eval(e,1).le.2.d-17) eval(e,1)=0.d0
           if( eval(e,1) .lt. 64.d0*epsrel) then
              eval(e,2)=eval(e,1)+epsrel*exp(-eval(e,1)/epsrel)
           else
              eval(e,2)=eval(e,1)
           endif
        enddo

!-----------------------------------------------------------------------
! calculate regularized inverse matrix
!-----------------------------------------------------------------------
        invers = CZERO
        do e=1,l
          do e1=1,l
            do e2=1,l
               invers(e1,e)=invers(e1,e)+ &
                    dconjg(evek(e,e2))*evek(e1,e2)/eval(e2,2)
            enddo
          enddo
          invers(e,e)=dble(invers(e,e))
        enddo
    end subroutine matinv_ml

end module linear
