!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: analysis.f90 518 2013-11-05 18:32:34Z frank $


module mlanalysis

    use global
    use wftree
    use linear
    use global
    use dmatmod
    use psidef
    use d1lib
    use op1lib
    use operdef
    use dvrdatmod

    use mltypes
    use mlop_eom
    use hphimod, only: hcall
    use psiutil, only: spfovl

    private
    public  :: expect,            &
               overlap_1,         &
               overlap_2,         &
               overlap_3,         &
               overlap_toplayer,  &
               compute_overlap_tree, &
               compute_mtraces_tree, &
               compute_shtraces_tree, &
               gridpop,           &
               print_wftree,      &
               headline_tree,     &
               errorphi_tree,     &
               ml_to_mctdh,       &
               croco_tree,        &
               printA_tree,       &
               reset_pointer_z,   &
               norm0_tree



    interface compute_overlap_node
        module procedure compute_overlap_1node, &
                         compute_overlap_2node
    end interface compute_overlap_node

    interface compute_overlap_tree
        module procedure compute_overlap_1tree, &
                         compute_overlap_2tree
    end interface compute_overlap_tree



    contains



! ----------------------------------------------------------------------------
    subroutine overlap_1(psi1,psi2,tree,ovr,ovlmat,cs)
        implicit none
        ! compute <P1|P2> for two ML wavefunctions having
        ! EXACTLY the same tree structure.
        ! The number of SPF's per layer must be the SAME.
        !
        ! ovlmat are the matrices containing the overlaps of the
        ! primitive layer SPFs
        complex(cdbl), intent(in)    :: psi1(:),psi2(:)
        type(wfnode_t),intent(inout) :: tree
        complex(cdbl), intent(out)   :: ovr
        complex(cdbl),intent(in)     :: ovlmat(dmatdim)
        integer,intent(in),optional  :: cs ! if 1, complex symmetric product

        integer                      :: cpl
        complex(cdbl), allocatable   :: work(:)

        cpl = 0 ! default to complex conjugate product
        if(present(cs)) cpl = cs
        ovr = CZERO

        ! Initialize
        allocate(work(product(tree%ndim)))
        call allocate_overlap_matrices(tree,tree,ovlmat)

        ! Compute overlap matrices and transform psi2 to the basis
        ! of psi1 from below recursively.
        call compute_overlap_tree(psi1,psi2,tree,cpl)

        ! Transform the top vector of psi2 to the basis of psi1
        work = psi2(tree%z(1,1):tree%z(1,2))
        if (dentype .eq. 2) then
           call chprod_phi2(tree%overlap,work)
           call hprod_phi2(tree%overlap,work)
        else
           call hprod_phi(tree%overlap,work)
        endif

        ! Take the final overlap
        if (dentype .eq. 2) then
           !call trmmcxzz(psi1,work,ovr,block(1),block(1))
           ovr = phi_phi(psi1(tree%z(1,1):tree%z(1,2)),work,cpl)
        else
           ovr = phi_phi(psi1(tree%z(1,1):tree%z(1,2)),work,cpl)
        endif

        ! Terminate
        call deallocate_overlap_matrices(tree)
        deallocate(work)
    end subroutine overlap_1


    subroutine overlap_2(psi1,psi2,tree1,tree2,ovr,ovlmat,cs)
        implicit none
        ! compute <P1|P2> for two ML wavefunctions having
        ! EXACTLY the same tree structure.
        ! The number of SPF's per layer may vary.
        !
        ! ovlmat are the matrices containing the overlaps of the
        ! primitive layer SPFs
        complex(cdbl), intent(in)      :: psi1(:),psi2(:)
        type(wfnode_t), intent(inout)  :: tree1
        type(wfnode_t), intent(inout)  :: tree2
        complex(cdbl), intent(out)     :: ovr
        complex(cdbl),intent(in)       :: ovlmat(dmatdim)
        integer,intent(in),optional    :: cs ! if 1, complex symmetric product

        integer                        :: cpl
        complex(cdbl)                  :: work(product(tree1%ndim))

        cpl = 0 ! default to complex conjugate product
        if(present(cs)) cpl = cs

        ovr = CZERO

        ! Initialize
        call allocate_overlap_matrices(tree1,tree2,ovlmat)

        ! Compute overlap matrices and transform psi2 to the basis
        ! of psi1 from below recursively.
        call compute_overlap_tree(psi1,psi2,tree1,tree2,cpl)

        ! Transform the top vector of psi2 to the basis of psi1
        call matprod_tensor(tree2%overlap,                  &
                            psi2(tree2%z(1,1):tree2%z(1,2)),&
                            work)

        ! Take the final overlap
        ovr = phi_phi(psi1(tree1%z(1,1):tree1%z(1,2)),work,cpl)

        ! Terminate
        call deallocate_overlap_matrices(tree2)
    end subroutine overlap_2


    subroutine overlap_3(psi1,psi2,tree1,tree2,ovr,delta,ovlmat,cs)
        implicit none
        ! Like overlap_2, but additionally compute an error measure
        ! for each node, and return them in the array delta.
        !
        ! ovlmat are the matrices containing the overlaps of the
        ! primitive layer SPFs
        complex(cdbl),intent(in)     :: psi1(:),psi2(:)
        type(wfnode_t),intent(inout) :: tree1
        type(wfnode_t),intent(inout) :: tree2
        complex(cdbl), intent(out)   :: ovr
        real(dbl),intent(out)        :: delta(tree1%tot_nodes)
        complex(cdbl),intent(in)     :: ovlmat(dmatdim)
        integer,intent(in),optional  :: cs ! if 1, complex symmetric product

        integer                     :: cpl
        complex(cdbl)               :: work(product(tree1%ndim))

        cpl = 0 ! default to complex conjugate product
        if(present(cs)) cpl = cs

        ovr = CZERO

        ! Initialize
        call allocate_overlap_matrices(tree1,tree2,ovlmat)

        ! Compute overlap matrices and transform psi2 to the basis
        ! of psi1 from below recursively.
        call compute_overlap_tree(psi1,psi2,tree1,tree2,cpl)

        ! Transform the top vector of psi2 to the basis of psi1
        call matprod_tensor(tree2%overlap,                  &
                            psi2(tree2%z(1,1):tree2%z(1,2)),&
                            work)

        ! Take the final overlap
        ovr = phi_phi(psi1(tree1%z(1,1):tree1%z(1,2)),work,cpl)

        ! Compute the per-node error measures.
        ! These need the density matrices.
        call compute_just_dmat_tree(tree1,psi1)
        call compute_just_dmat_tree(tree2,psi2)
        call errorphi_tree(tree1,tree2,delta)

        ! Terminate
        call deallocate_overlap_matrices(tree2)
    end subroutine overlap_3


!   subroutine compute_deltaphi_tree(tree,psi1,psi2,deltaphi)
!       implicit none
!       type(wfnode_t),intent(inout):: tree
!       complex(cdbl),intent(in)  :: psi1(:),psi2(:)
!       real(dbl),intent(out)     :: deltaphi
!       integer                   :: i
!       real(dbl)                 :: deltanode
!       type(wfnode_t),pointer    :: cnode
!       deltaphi=0.d0
!       do i=2,tree%tot_nodes
!           cnode => tree%swap_down(i)%p
!           call compute_errorphi_node(cnode,psi1,psi2,deltanode)
!           deltaphi = deltaphi + deltanode
!       end do
!   end subroutine compute_deltaphi_tree


    subroutine overlap_toplayer(psi1,psi2,tree,ovr,cs)
        implicit none
        ! Compute <P1|P2> only for the top layer and assuming that
        ! the two A-vectors have the same length.
        complex(cdbl), intent(in)    :: psi1(:),psi2(:)
        type(wfnode_t), intent(in)   :: tree
        complex(cdbl), intent(out)   :: ovr
        integer,intent(in),optional  :: cs ! if 1, complex symmetric product
        integer   :: z0,zf,cpl

        cpl = 0 ! default to complex conjugate product
        if(present(cs)) cpl = cs

        z0 = tree%z(1,1)
        zf = tree%z(1,2)
        ovr = phi_phi(psi1(z0:zf),psi2(z0:zf),cpl)
    end subroutine overlap_toplayer

! ----------------------------------------------------------------------------
    subroutine compute_overlap_1tree(psi1,psi2,tree,cp)
        implicit none
        ! Compute overlap matrices between SPF's of two different
        ! wavefunctions.
        complex(cdbl),intent(in)        :: psi1(:)
        complex(cdbl),intent(in)        :: psi2(:)
        type(wfnode_t),intent(inout)    :: tree
        integer,intent(in),optional     :: cp ! if 1, complex symmetric prod.
        integer                         :: cpl
        type(wfnode_t),pointer          :: cnode
        integer                         :: i

        cpl = 0 ! default to complex conjugate product
        if(present(cp)) cpl = cp

        do i=1,tree%tot_nodes
            cnode => tree%swap_up(i)%p
            if(cnode%isprimitive) cycle
            call compute_overlap_node(psi1,psi2,cnode,cpl)
        end do
    end subroutine compute_overlap_1tree


    subroutine compute_overlap_2tree(psi1,psi2,tree1,tree2,cp)
        implicit none
        ! Compute overlap matrices between SPF's of two different
        ! wavefunctions. They are stored in tree2
        complex(cdbl),intent(in)        :: psi1(:)
        complex(cdbl),intent(in)        :: psi2(:)
        type(wfnode_t),intent(inout)    :: tree1
        type(wfnode_t),intent(inout)    :: tree2
        integer,intent(in),optional     :: cp ! if 1, complex symmetric prod.
        integer                         :: cpl
        type(wfnode_t),pointer          :: cnode1,cnode2
        integer                         :: i

        cpl = 0 ! default to complex conjugate product
        if(present(cp)) cpl = cp

        do i=1,tree1%tot_nodes
            cnode1 => tree1%swap_up(i)%p
            cnode2 => tree2%swap_up(i)%p
            if(cnode1%isprimitive) cycle
            call compute_overlap_node(psi1,psi2,cnode1,cnode2,cpl)
        end do
    end subroutine compute_overlap_2tree


    subroutine compute_overlap_1node(psi1,psi2,cnode,cp)
        implicit none
        complex(cdbl),intent(in)        :: psi1(:)
        complex(cdbl),intent(in)        :: psi2(:)
        type(wfnode_t),intent(inout)    :: cnode
        integer,intent(in),optional     :: cp ! if 1, complex symmetric prod.

        integer                         :: cpl
        type(wfnode_t),pointer          :: child
        complex(cdbl),pointer           :: ovrmat(:,:)
        complex(cdbl),allocatable       :: work(:)
        integer                         :: i,j,f

        cpl = 0 ! default to complex conjugate product
        if(present(cp)) cpl = cp

!$OMP parallel default(shared) private(f,work,ovrmat,child)
!$OMP do schedule(runtime)
        do f=1,cnode%nmodes
            ovrmat => cnode%overlap(f)%p%matrix
            child  => cnode%submodes(f)%p

! primitive overlaps already done
            if (child%isprimitive) cycle

            allocate(work(product(child%ndim)))
            do j=1,cnode%ndim(f)
                do i=1,cnode%ndim(f)
                   work = psi2(child%z(j,1):child%z(j,2))
                   call hprod_phi(child%overlap,work)
                   ovrmat(i,j) = phi_phi(psi1(child%z(i,1):child%z(i,2)),&
                                         work, cpl)
                end do
            end do
            deallocate(work)
        end do
!$OMP end do
!$OMP end parallel

    end subroutine compute_overlap_1node


    subroutine compute_overlap_2node(psi1,psi2,node1,node2,cp)
        implicit none
        complex(cdbl),intent(in)        :: psi1(:)
        complex(cdbl),intent(in)        :: psi2(:)
        type(wfnode_t),intent(inout)    :: node1
        type(wfnode_t),intent(inout)    :: node2
        integer,intent(in),optional     :: cp ! if 1, complex symmetric prod.

        integer                         :: cpl
        type(wfnode_t),pointer          :: child1,child2
        complex(cdbl),pointer           :: ovrmat(:,:)
        complex(cdbl),allocatable       :: work(:)
        integer                         :: i,j,f

        cpl = 0 ! default to complex conjugate product
        if(present(cp)) cpl = cp

        do f=1,node1%nmodes
            ovrmat => node2%overlap(f)%p%matrix
            child1 => node1%submodes(f)%p
            child2 => node2%submodes(f)%p

! primitive overlaps already done
            if (child1%isprimitive) cycle

!$OMP parallel default(shared) private(j,i,work)
!$OMP do schedule(runtime)
            do j=1,node2%ndim(f)
                do i=1,node1%ndim(f)

                    allocate(work(product(child1%ndim)))

                    call matprod_tensor(                         &
                               child2%overlap,                   &
                               psi2(child2%z(j,1):child2%z(j,2)),&
                               work)
                    ovrmat(i,j) = phi_phi(                            &
                                psi1(child1%z(i,1):child1%z(i,2)),&
                                work,cpl)

                    deallocate(work)
                end do
            end do
!$OMP end do
!$OMP end parallel

        end do
    end subroutine compute_overlap_2node

! ----------------------------------------------------------------------------
    subroutine compute_mtraces_tree(psi,tree,cmpl)
        implicit none
        ! Compute trace vectorss of SPDOs of a type I density matrix
        complex(cdbl),intent(in)        :: psi(:)
        type(wfnode_t),intent(inout)    :: tree
        type(wfnode_t),pointer          :: cnode
        integer, intent(in), optional   :: cmpl   ! if cmpl=1, complex traces
        integer                         :: i,iflag

        if (present(cmpl)) then
           iflag = cmpl
        else
           iflag = 0
        endif
        
        if (iflag .eq. 0) then
           do i=1,tree%tot_nodes
               cnode => tree%swap_up(i)%p
               if(cnode%isprimitive) cycle
               call compute_mtraces_node(psi,cnode)
           end do
        else
           do i=1,tree%tot_nodes
               cnode => tree%swap_up(i)%p
               if(cnode%isprimitive) cycle
               call compute_zmtraces_node(psi,cnode)
           end do
        endif

    end subroutine compute_mtraces_tree


    subroutine compute_mtraces_node(psi,cnode)
        implicit none
        complex(cdbl),intent(in)        :: psi(:)
        type(wfnode_t),intent(inout)    :: cnode

        type(wfnode_t),pointer          :: child
        complex(cdbl),allocatable       :: work(:)
        integer                         :: i,j,f,adim

        do f=1,cnode%nmodes
            child  => cnode%submodes(f)%p
            adim = product(child%ndim)

! lowest layer SPDO traces already done
            if (child%isprimitive) cycle

            cnode%mtraces(f)%p%rvector(:) = CZERO
!$OMP parallel default(shared) private(j,i,work)
            allocate(work(adim))
!$OMP do schedule(runtime)
            do j=1,cnode%ndim(f)
               work = psi(child%z(j,1):child%z(j,2))
               call hprod_phi(child%mtraces,work)

               do i = 1,adim
                  cnode%mtraces(f)%p%rvector(j)  = cnode%mtraces(f)%p%rvector(j) &
                                                   + dble(work(i))
               enddo

            end do
!$OMP end do
            deallocate(work)
!$OMP end parallel

        end do
    end subroutine compute_mtraces_node

    subroutine compute_zmtraces_node(psi,cnode)
        implicit none
        complex(cdbl),intent(in)        :: psi(:)
        type(wfnode_t),intent(inout)    :: cnode

        type(wfnode_t),pointer          :: child
        complex(cdbl),allocatable       :: work(:)
        integer                         :: i,j,f,adim

        do f=1,cnode%nmodes
            child  => cnode%submodes(f)%p
            adim = product(child%ndim)

! lowest layer SPDO traces already done
            if (child%isprimitive) cycle

            cnode%mtraces(f)%p%vector(:) = CZERO
!$OMP parallel default(shared) private(j,i,work)
            allocate(work(adim))
!$OMP do schedule(runtime)
            do j=1,cnode%ndim(f)
               work = psi(child%z(j,1):child%z(j,2))
               call hprod_phi(child%mtraces,work)

               do i = 1,adim
                  cnode%mtraces(f)%p%vector(j)  = cnode%mtraces(f)%p%vector(j) &
                                                   + work(i)
               enddo

            end do
!$OMP end do
            deallocate(work)
!$OMP end parallel

        end do
    end subroutine compute_zmtraces_node

! ----------------------------------------------------------------------------

    subroutine compute_shtraces_tree(psi,tree)
        implicit none
        ! Compute trace vectorss of SPDOs of a type I density matrix
        complex(cdbl),intent(in)        :: psi(:)
        type(wfnode_t),intent(inout)    :: tree
        type(wfnode_t),pointer          :: cnode
        integer                         :: i

        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            if(cnode%isprimitive) cycle
            call compute_shtraces_node(psi,cnode)
        end do

    end subroutine compute_shtraces_tree


    subroutine compute_shtraces_node(psi,cnode)
        implicit none
        complex(cdbl),intent(in)        :: psi(:)
        type(wfnode_t),intent(inout)    :: cnode

        complex(cdbl),allocatable       :: work(:)
        complex(cdbl),allocatable       :: work2(:,:)
        real(dbl),allocatable           :: shttmp(:)
        integer                         :: f,adim,r
        integer                         :: z0,zf,ns
        integer                         :: ijk(3)
        real(dbl), pointer              :: shtparent(:),shtcurrent(:)

        if(cnode%isprimitive) return

        adim = product(cnode%ndim)

        if(cnode%istop) then ! NORMAL MCTDH mean fields

           z0 = cnode%z(1,1)
           zf = cnode%z(1,2)

!$OMP parallel default(shared) private(f,work,ijk)
!$OMP do schedule(runtime)
           do f=1,cnode%nmodes
               allocate(work(adim))
               cnode%shtraces(f)%p%rvector(:) = CZERO
               work = psi(z0:zf)
               call hprod_phi(cnode%mtraces,work,f)

               ijk = ijk_shape(f,cnode%ndim)
               call travec(work,cnode%shtraces(f)%p%rvector,ijk(1),ijk(2),ijk(3))

               deallocate(work)
           end do
!$OMP end do
!$OMP end parallel

        else

           ns=cnode%nsubcoord
           shtparent => cnode%parent%shtraces(ns)%p%rvector
           allocate(work2(adim,cnode%navecs))
           allocate(work(adim))

           work2 = CZERO
           do r=1,cnode%navecs
              z0 = cnode%z(r,1)
              zf = cnode%z(r,2)
              work2(:,r) = work2(:,r) + shtparent(r)*psi(z0:zf)
           enddo
           do f=1,cnode%nmodes ! loop over modes

              allocate(shttmp(cnode%ndim(f)))
              shtcurrent => cnode%shtraces(f)%p%rvector
              shtcurrent = CZERO
              do r=1,cnode%navecs
                 z0 = cnode%z(r,1)
                 zf = cnode%z(r,2)
                 work(:) = work2(:,r)
                 call hprod_phi(cnode%mtraces, work, f)
                 ijk = ijk_shape(f,cnode%ndim)
                 call travec(work,shttmp,ijk(1),ijk(2),ijk(3))
                 shtcurrent = shtcurrent + shttmp
              enddo
              deallocate(shttmp)
           end do

        endif
    end subroutine compute_shtraces_node


    subroutine gridpop(this,psi)
        implicit none
        type(wfnode_t),intent(inout)    :: this        ! this node
        complex(cdbl),intent(in)        :: psi(:)      ! whole psi array
        type(wfnode_t),pointer          :: cnode
        integer                         :: i
        do i=1,this%tot_nodes
            cnode => this%swap_down(i)%p
            if(cnode%isprimitive) then
                call gridpop_node(cnode,psi)
            end if
        end do
    end subroutine gridpop


    subroutine gridpop_node(this,psi)
        implicit none
        ! Compute grid populations
        ! For non-primitive nodes the populations refer to the basis one level
        ! below and are not related to the populations with respect to the
        ! underlying time-independent basis. The name gridpop is due to the fact
        ! that when using a DVR in position representation the grid populations
        ! indicate the population at each grid point along a coordinate.
        type(wfnode_t),intent(inout)    :: this        ! this node
        complex(cdbl),target,intent(in) :: psi(:)      ! whole psi array
        complex(cdbl),pointer           :: dmat(:,:)   ! dmat, parent
        complex(cdbl),pointer           :: phi1(:),phi2(:) ! spfs of this node
        complex(cdbl),pointer           :: workc(:)
        real(dbl),pointer               :: workr(:)
        integer                         :: ns,p,q,plen,d

        if(.not. this%isprimitive) return

        plen = product(this%ndim)
        ! workc holds the density grid population of the whole node.
        ! If the mode is combined the degf populations are computed
        ! afterwards from partial integration of workc
        workc => this%topnode%workc(1:plen)
        workr => this%topnode%workr(1:plen)
        workc = CZERO

        ! Calculate the probability density for the whole node
        if(this%istop) then
            ! exact calculation, construct density from total psi
            workc = dconjg(psi)*psi
        else
            ! normal gridpop, contruct density using parent's dmat
            ns = this%nsubcoord
            dmat => this%parent%dmat(ns)%p%matrix
            do p=1,this%navecs
                phi1 => psi(this%z(p,1):this%z(p,2))
                do q=1,this%navecs
                    phi2 => psi(this%z(q,1):this%z(q,2))
                    workc = workc + dmat(p,q)*(dconjg(phi1)*phi2)
                end do
            end do
        end if

        ! Calculate the density for each degree of freedom
        workr = dble(workc)
        do p=1,this%nmodes
            d=this%degf(p)
            call intgr_dens(this%topnode%pgpops(d)%p%rvector,p,this%ndim,workr)
        end do

        workc => null()
        workr => null()
    end subroutine gridpop_node


! ----------------------------------------------------------------------------
    subroutine ml_to_mctdh(self,psi,psimctdh,cntrl)
        ! Convert a ML wavefunction to an MCTDH one
        !
        ! This can be done in different ways that result in different MCTDH
        ! wavefunctions (2 layers of TD coeffs) representing the same original ML
        ! wavefunctions (2 < layers of TD coeffs):
        !
        ! 1) Grow the nodes below the top node until they contain all
        !    primitive degfs of their branch.
        ! 2) Grow the A-vector at the top layer until it contains the expansion
        !    coefficients for the basis in the primitive nodes of the wavefunction.
        ! 3) The top node has two logical coordinates, system (Q) and electronic
        !    (el) coordinates. In this case a multi-state MCTDH wavefunction can
        !    be generated.
        !
        ! Options 2) and 3) result in a large A-vector, and are likely the desired
        ! options if we want to compare to similar MCTDH calculations. Option
        ! 1) is likely to result in overcombined modes, which would correspond
        ! to an hipothetical MCTDH calculation likely to have been inefficient
        ! or even undoable.
        ! The optional argument cntrl=(1,2,3) controls which option is taken.

        implicit none
        type(wfnode_t),intent(inout)    :: self   ! top node in tree
        complex(cdbl),intent(in),target :: psi(:)
        complex(cdbl),intent(out)       :: psimctdh(:)
        integer,intent(in),optional     :: cntrl

        integer                      :: i,s,l,al,zi,zf,zia,zfa,lcntrl,nstates
        integer                      :: indx(self%ndim(2))
        type(wfnode_t),pointer       :: cnode      => null()
        complex(cdbl),pointer        :: spfs(:)    => null()
        complex(cdbl),allocatable    :: astate(:)
        real(dbl)                    :: big

        if (.not. self%istop) then
            write(6,*) 'routine: ml_to_mctdh; called for node other than top'
            stop
        end if

        if (present(cntrl)) then
            lcntrl = cntrl
        else
            lcntrl = 2
        end if

        ! Compute MCTDH coefficients
        do i=1,self%tot_nodes
            cnode => self%swap_up(i)%p
            if(cnode%isprimitive) cycle ! never undo primitive basis expansion
            if(cnode%istop .and. (lcntrl==1 .or. lcntrl==3)) cycle
            call decontract_node(cnode,psi,lcntrl)
        end do

        ! Copy MCTDH coefficients to psimctdh
        select case(lcntrl)

            !*********
            case(1)  ! highly combined modes, original A-vector(top) kept
                zi = 1
                zf = product(self%ndim)*self%navecs
                psimctdh(zi:zf) = psi(zi:zf)
                zi = zf + 1
                do i=1,self%nmodes
                    cnode => self%submodes(i)%p
                    if (allocated(cnode%phiaux)) then
                        l = size(cnode%phiaux)
                        zf = zi + l - 1
                        psimctdh(zi:zf) = cnode%phiaux
                    else
                        l = product(cnode%ndim)*cnode%navecs
                        zf = zi + l - 1
                        psimctdh(zi:zf) = &
                             psi(cnode%z(1,1):cnode%z(cnode%navecs,2))
                    end if
                    zi = zf + 1
                end do

            !*********
            case(2)  ! large A-vector, original primitive modes kept
                zi = 1
                do i=1,self%tot_nodes
                    cnode => self%swap_down(i)%p
                    if(.not.cnode%isprimitive .and. .not.cnode%istop) cycle
                    if (allocated(cnode%phiaux)) then
                        l = size(cnode%phiaux)
                        zf = zi + l - 1
                        psimctdh(zi:zf) = cnode%phiaux
                    else
                        l = product(cnode%ndim)*cnode%navecs
                        zf = zi + l - 1
                        psimctdh(zi:zf) = &
                             psi(cnode%z(1,1):cnode%z(cnode%navecs,2))
                    end if
                    zi = zf + 1
                end do

            !*********
            case(3)  ! muli-state MCTDH wfn is generated
                ! Sanity checks
                if (self%nmodes /= 2) then
                    write(6,*) 'ml2mctdh: top node must have 2 submodes'
                    stop
                end if
                if (self%ndim(1) /= self%ndim(2)) then
                    write(6,*) 'ml2mctdh: number of electronic states and system '
                    write(6,*) 'states differ at the top layer, unable to transform'
                    stop
                end if

                cnode => self%submodes(1)%p  ! node with un-processed A-vectors
                if (allocated(cnode%phiaux)) then
                    l = size(cnode%phiaux)
                    al = product(cnode%ndimaux)
                    spfs => cnode%phiaux
                else
                    l = product(cnode%ndim)*cnode%navecs
                    al = product(cnode%ndim)
                    spfs => psi(cnode%z(1,1):cnode%z(cnode%navecs,2))
                end if

                ! Allocate space for new A-vectors and multiply with coeffs of first layer
                nstates = cnode%navecs
                allocate(astate(l))
                astate = CZERO
                zia = 1
                do s=1,nstates
                    zfa = zia + al - 1
                    zi = 1
                    do i=1,nstates
                        zf = zi + al - 1
                        astate(zia:zfa) = astate(zia:zfa) + &
                           psi( (s-1)*nstates + i ) * spfs(zi:zf)
                        zi = zf + 1
                    end do
                    zia = zfa + 1
                end do
                spfs => null()

                ! Find out the internal ordering of the electronic states
                indx = 0 ! reset indx array
                cnode => self%submodes(2)%p ! node with electronic degf
                spfs => psi(cnode%z(1,1):cnode%z(cnode%navecs,2))
                do s=1,nstates
                    big = 0.d0
                    do i=1,nstates
                        if (abs(spfs( (s-1)*nstates + i )) .gt. big) then
                            big = abs(spfs( (s-1)*nstates + i ))
                            indx(s) = i
                        end if
                    end do
                end do
                spfs => null()

                ! Copy A(s) blocks
                zia = 1
                do s=1,nstates
                    zi = (indx(s) - 1)*al + 1
                    zf = zi + al - 1
                    zfa = zia + al - 1
                    psimctdh(zi:zf) = astate(zia:zfa)
                    zia = zfa + 1
                end do

                zi = al*nstates + 1
                ! Copy phi(m,s) modes
                do i=1,self%tot_nodes
                    cnode => self%swap_down(i)%p
                    if(.not.cnode%isprimitive) cycle
                    if(cnode%nsubcoord==2 .and. cnode%parent%istop) cycle
                    do s=1,nstates  ! copy phi(s) blocks
                        if (allocated(cnode%phiaux)) then
                            l = size(cnode%phiaux)
                            zf = zi + l - 1
                            psimctdh(zi:zf) = cnode%phiaux
                        else
                            l = product(cnode%ndim)*cnode%navecs
                            zf = zi + l - 1
                            psimctdh(zi:zf) = &
                                 psi(cnode%z(1,1):cnode%z(cnode%navecs,2))
                        end if
                        zi = zf + 1
                    end do
                end do

                deallocate(astate)

        end select

        ! Final cleanup
        do i=1,self%tot_nodes
            cnode => self%swap_up(i)%p
            if (allocated(cnode%phiaux))  deallocate(cnode%phiaux)
            if (allocated(cnode%ndimaux)) deallocate(cnode%ndimaux)
        end do
    end subroutine ml_to_mctdh


    subroutine decontract_node(no,psi,cntrl)
        ! Undo the basis expansion of a given node
        !
        ! At the end, the node stores in phiaux the vectors of coefficients
        ! indexed by the indices of its submodes, meaning that at this level
        ! the basis expansion has been undone.
        !
        ! cntrl = 1: combine modes. number of indices in this node remains constant.
        ! cntrl = 2: enlarge dimensionality of A-vec. Number of indices grows.
        implicit none
        type(wfnode_t),intent(inout)    :: no
        complex(cdbl),intent(in),target :: psi(:)
        integer,intent(in)              :: cntrl

        type(wfnode_t),pointer       :: child
        integer                      :: i,j,zio,zfo,zin,zfn
        integer                      :: dimnew(no%nmodes)
        integer                      :: dimold(no%nmodes)
        complex(cdbl),pointer        :: spfs(:) => null()
        complex(cdbl),allocatable    :: aold(:),anew(:)

        dimnew = no%ndim
        dimold = dimnew
        allocate(aold(product(dimold)*no%navecs))
        aold = psi(no%z(1,1):no%z(no%navecs,2))

        ! run over submodes and decontract the basis expansion
        do i=1,no%nmodes

            child => no%submodes(i)%p
            if (child%isprimitive .and. (cntrl==2 .or. cntrl==3)) cycle

            if(allocated(child%phiaux)) then
                spfs => child%phiaux
                dimnew(i) = product(child%ndimaux)
            else
                spfs => psi(child%z(1,1):child%z(child%navecs,2))
                dimnew(i) = product(child%ndim)
            end if

            allocate(anew(product(dimnew)*no%navecs))
            do j=1,no%navecs
                zio = (j-1)*product(dimold) + 1
                zfo = j*product(dimold)
                zin = (j-1)*product(dimnew) + 1
                zfn = j*product(dimnew)
                call decontract_index(anew(zin:zfn),   &
                                      aold(zio:zfo),   &
                                      dimnew,          &
                                      dimold,          &
                                      spfs,            &
                                      i)
            end do
            deallocate(aold)
            dimold = dimnew
            allocate(aold(product(dimold)*no%navecs))
            aold = anew
            deallocate(anew)

            spfs => null()
            if(allocated(child%phiaux)) then
                deallocate(child%phiaux)
                deallocate(child%ndimaux)
            end if

        end do  ! submodes

        allocate(no%phiaux(product(dimnew)*no%navecs))
        no%phiaux = aold
        deallocate(aold)

        ! set ndimaux, which holds the new dimensions for this node
        allocate(no%ndimaux(size(dimnew)))
        no%ndimaux = dimold

!       select case(cntrl)
!           case(1)
!               allocate(no%ndimaux(size(dimnew)))
!               no%ndimaux = dimold
!           case(2,3)
!               allocate(no%ndimaux(nindx))
!               zio = 1
!               do i=1,no%nmodes
!                   child => no%submodes(i)%p
!                   if (child%isprimitive) then
!                       zfo = zio
!                       no%ndimaux(zio) = no%ndim(i)
!                   else
!                       if(allocated(child%ndimaux)) then
!                           zfo = zio + size(child%ndimaux) - 1
!                           no%ndimaux(zio:zfo) = child%ndimaux
!                       else
!                           zfo = zio + size(child%ndim) - 1
!                           no%ndimaux(zio:zfo) = child%ndim
!                       end if
!                   end if
!                   zio = zfo + 1
!               end do
!       end select
    end subroutine decontract_node


    subroutine headline_tree(iout,tree,ltop,pre,post,buf)
      implicit none
      integer,intent(in)             :: iout
      type(wfnode_t),intent(in)      :: tree
      logical,intent(in)             :: ltop      ! include top node?
      character(len=*),intent(in)    :: pre,post
      character(len=*),intent(inout) :: buf
      type(wfnode_t),pointer         :: no
      integer                        :: m
      write (iout,'(a)',advance="no") pre
      do m=1,tree%tot_nodes
         no => tree%swap_down(m)%p
         if (no%istop .and. .not.ltop) cycle
         write (buf,'(a,i0)') '@',no%num
         write (iout,'(a)',advance="no") adjustr(buf)
      enddo
      write (iout,'(a)') post
    end subroutine headline_tree


   subroutine print_wftree(iout,tree,mlbl,lcomment)
      ! Writes a textual description of the tree structure to the
      ! output channel. Used in overlap.F90, ortho.F90, rdacoeff.F90, norm.F90
      implicit none
      integer,intent(in)         :: iout     ! output channel
      type(wfnode_t),intent(in)  :: tree     ! the tree
      character(len=*),intent(in):: mlbl(:)  ! modelabels
      logical,intent(in),optional:: lcomment ! prefix with # or not
      logical                    :: lc
      integer                    :: m,f
      type(wfnode_t),pointer     :: no,ch

      if (present(lcomment)) then
         lc=lcomment
      else
         lc=.false.
      endif

      if (lc) write (iout,'(a)',advance='no') '# '
      write (iout,'(a)') 'node : no. of SPFs : submodes'
      do m=1,tree%tot_nodes
         if (lc) write (iout,'(a)',advance='no') '# '
         no => tree%swap_down(m)%p
         write (iout,'(i4,a,i4,a)',advance='no') &
            no%num,' : ',no%navecs,' : '
         if (no%isprimitive) then
            do f=1,no%nmodes
               write (iout,'(a,1x)',advance='no') trim(mlbl(no%degf(f)))
            enddo
         else
            do f=1,no%nmodes
               ch => no%submodes(f)%p
               write (iout,'(a,i0,1x)',advance='no') '@',ch%num
            enddo
         endif
         write (iout,*)
      enddo
      return
   end subroutine print_wftree

   subroutine reset_pointer_z(tree,mode)
   ! This subroutine is called from subroutine reduceml.
   ! The latter is part of flux and wfprof.
      implicit none
      type(wfnode_t),intent(in)  :: tree
      integer,intent(in)         :: mode
      integer                    :: m,f
      type(wfnode_t),pointer     :: no

      do m=1,tree%tot_nodes
         no => tree%swap_down(m)%p
         if (no%isprimitive) then
            if(no%m .eq. mode) then
               do f=1,no%navecs
                  no%z(f,2)=no%z(f,1)
               enddo
            endif
         endif
      enddo
   end subroutine reset_pointer_z

   subroutine errorphi_tree(t1,t2,delta)
   ! Computes a per-node error measure between two ML wavefunctions,
   ! and returns them in the array delta.
   ! The overlap matrices between the wavefunctions as well as the
   ! density matrices for t1 and t2 must already have been computed,
   ! i.e. this procedure must be called *after*
   !   compute_overlap_tree
   !   compute_just_dmat_tree OR compute_dmat_tree
   ! The overlap matrices must be stored in t2.
      implicit none
      type(wfnode_t),intent(in) :: t1,t2
      real(dbl),intent(out)     :: delta(t1%tot_nodes)
      type(wfnode_t),pointer    :: n1,n2
      integer                   :: i
      real(dbl)                 :: delta1,delta2
      delta = ZEROR
      do i=1,t1%tot_nodes
         n1 => t1%swap_down(i)%p
         n2 => t2%swap_down(i)%p
         if (n1%istop) cycle
         call errorphi_node(n1,n2,delta1)   ! using \rho of 1st WF
         call errorphi_node_c(n1,n2,delta2) ! using \rho of 2nd WF
         delta(n1%num) = 0.5d0*(delta1+delta2)
      enddo
   end subroutine errorphi_tree


   subroutine errorphi_node(n1,n2,delta)
   ! Computes for one node the error measure
   !   1 - tr(O^\dagger \rho^T O) / tr(\rho)
   ! where O is the overlap matrix between the SPFs,
   !   O = \phi1^\dagger \phi2 ,
   ! and \rho is the density matrix belonging to n1.
      implicit none
      type(wfnode_t),intent(in) :: n1,n2
      real(dbl),intent(out)     :: delta
      type(wfnode_t),pointer    :: p1,p2
      complex(cdbl),pointer     :: ov(:,:),dm(:,:)
      integer                   :: f,i,j,k
      complex(cdbl)             :: er,tr

      f  = n1%nsubcoord
      p1 => n1%parent
      p2 => n2%parent
      ov => p2%overlap(f)%p%matrix
      dm => p1%dmat(f)%p%matrix
      tr = CZERO
      ! compute tr(\rho)
      do j=1,n1%navecs
         tr = tr + dm(j,j)
      enddo
      ! compute tr(O^\dagger \rho^T O)
      er = CZERO
      do i=1,n2%navecs
         do j=1,n1%navecs
            do k=1,n1%navecs
               er = er + dconjg(ov(j,i))*dm(k,j)*ov(k,i)
            enddo
         enddo
      enddo
      delta = max(1.d0 - dble(er)/dble(tr), 0.d0)
   end subroutine errorphi_node


   subroutine errorphi_node_c(n1,n2,delta)
   ! Computes for one node the error measure
   !   1 - tr(O \rho^T O^\dagger) / tr(\rho)
   ! where O is the overlap matrix between the SPFs,
   !   O = \phi1^\dagger \phi2 ,
   ! and \rho is the density matrix belonging to n2.
   ! The difference to errorphi_node is that here we use the density
   ! matrices from the 2nd WF, but as we want to use the same overlap
   ! matrices as there, O is replaced by O^\dagger.
      implicit none
      type(wfnode_t),intent(in) :: n1,n2
      real(dbl),intent(out)     :: delta
      type(wfnode_t),pointer    :: p1,p2
      complex(cdbl),pointer     :: ov(:,:),dm(:,:)
      integer                   :: f,i,j,k
      complex(cdbl)             :: er,tr

      f  = n1%nsubcoord
      p1 => n1%parent
      p2 => n2%parent
      ov => p2%overlap(f)%p%matrix
      dm => p2%dmat(f)%p%matrix
      tr = CZERO
      ! compute tr(\rho)
      do j=1,n2%navecs
         tr = tr + dm(j,j)
      enddo
      ! compute tr(O \rho^T O^\dagger)
      er = CZERO
      do j=1,n2%navecs
         do k=1,n2%navecs
            do i=1,n1%navecs
               er = er + ov(i,j)*dm(k,j)*dconjg(ov(i,k))
            enddo
         enddo
      enddo
      delta = max(1.d0 - dble(er)/dble(tr), 0.d0)
   end subroutine errorphi_node_c

    subroutine printA_tree(iout,tree,psi,absi,thr,mx,lcnt)
!   This subroutine is used in rdacoeff.F90
      implicit none
      integer,intent(in)          :: iout   ! output channel
      type(wfnode_t),intent(in)   :: tree   ! the tree
      complex(cdbl),intent(in)    :: psi(:) ! wavefunction

      type(wfnode_t),pointer      :: ch
      type(wfnode_t),pointer      :: no
      integer                     :: k,f,i,j
      integer absi  ! flag -abs -abs2 -re -im (see rdacoeff)
      real*8 thr,mx ! max and min thresholds (see rdacoeff)
      real*8  norm  ! norm (see rdacoeff)
      integer cnt   ! counter (see rdacoeff)
      logical lcnt  ! flag -count (see rdacoeff)

!        Extra printing for debugging
!!         Write(iout,'(A,I6,I6)') 'Size psi:',dgldim_wftree(tree),&
!!         tree%tot_coefs
!!         Write(iout,'(A,I6)') 'Size A: ',adim_wftree(tree)
!!         Write(iout,'(A,I6)') 'Size Atop: ',atopdim_wftree(tree)
!!         Write(iout,'(A,I6)') 'Total nodes: ',tree%tot_nodes
!!         Write(iout,'(A,I1)') 'absi: ',absi

!     Main loop over ML nodes
      do k=1,tree%tot_nodes
         no => tree%swap_down(k)%p
         if(.not. no%isprimitive) then
!           Print node information (non-primitive)
            write (iout,'(a)',advance='no') '# '
            write (iout,'(a)') 'mode : no. of SPFs : submodes'
            write (iout,'(a)',advance='no') '# '
            write (iout,'(i4,a,i4,a)',advance='no') &
            no%num,' : ',no%navecs,' : '
            do f=1,no%nmodes
               ch => no%submodes(f)%p
               write (iout,'(a,i0,1x)',advance='no') '@',ch%num
            enddo
            write (iout,*)
!           Print header
            if (.not.lcnt) then
               write(iout,'(2x,a,5x,a,12x,a,13x,a)')&
                    'J','m','A','j1 j2 j3 j4 j5 ...'
               write(iout,'(2a)') '  -----------------',&
                    '--------------------------------------------------------'
            endif
!           Loop over A vectors
            do i=1,no%navecs
               norm=0.0
               cnt=0
               do j=no%z(i,1),no%z(i,2)
                  if ( abs(psi(j)) .gt. thr .and. abs(psi(j)) .le. mx) then
                  cnt = cnt + 1
                  norm = norm + dble(psi(j))**2 + dimag(psi(j))**2
!                    Print A coeff.
                     if (.not.lcnt) then
                        if (absi.eq.1) then
                           write(iout,'(i4,x,i4,f21.12,2x,2x,35i3)')&
                           j,i,abs(psi(j))
                        else if (absi.eq.2) then
                           write(iout,'(i4,x,i4,f21.12,2x,2x,35i3)')&
                           j,i,abs(psi(j))**2
                        else if (absi.eq.3) then
                           write(iout,'(i4,x,i4,f21.12,2x,2x,35i3)')&
                           j,i,dble(psi(j))
                        else if (absi.eq.4) then
                           write(iout,'(i4,x,i4,f21.12,2x,2x,35i3)')&
                           j,i,dimag(psi(j))
                        else
                           write(iout,'(i4,x,i4,"  (",f9.6,",",f9.6,")",2x,35i3)')&
                           j,i,psi(j)
                        endif
                     endif
                  endif
               enddo
               norm = dsqrt(norm)
!              Print summary tale
               write (iout,*)
               if (lcnt) write (iout,'(x,a,i4)') 'For m=',i
               write(iout,'(a,1p2e9.2," :",i8)')&
               ' Number of A-coeffs with absolute values between',thr,mx,cnt
               write(iout,'(a,f9.6)')&
               ' Norm of these A-coeffs (sqrt[sum|A|^2]):',norm
               if (.not.lcnt)  write (iout,*)
            enddo
            if (lcnt) write (iout,*)
         end if
      enddo
   end subroutine printA_tree

    subroutine norm0_tree(iout,tree,mlbl,psi,j,time,anorm2,ovl,fs,lhead,agmat)
!   This subroutine is used in norm.F90 . DMT
      implicit none
      integer,intent(in)          :: iout,j   ! output channel, node-counter
      type(wfnode_t),intent(inout) :: tree   ! the tree
      character(len=*),intent(in) :: mlbl(:) ! modelabels
      complex(cdbl),intent(in)    :: psi(:) ! wavefunction

      type(wfnode_t),pointer      :: ch
      type(wfnode_t),pointer      :: no
      integer                     :: f,l

      integer,intent(in)          :: lhead ! header print option
!     lhead=0 (only header);lhead=1 (header+norm);lhead=2 (only norm)
      real*8,intent(in)           :: time,fs
      real*8,intent(out)          :: anorm2
      real*8                      :: norm
      complex(cdbl),intent(out)   :: ovl
      complex(cdbl)               :: ovlmat(dmatdim)
      complex(cdbl), intent(in)   :: agmat(agmatdim)

      no => tree%swap_down(j)%p

      if (lhead.le.1) then
!        Extra printing for debugging
!!         write(iout,*)
!!         write(iout,*) '-----------------------------------------'
!!         write (iout,'(a,i4)') '(node) num =',no%num
!!         write (iout,'(a,i4,a,i4)') 'navecs =',no%navecs,',  nmodes =',no%nmodes
!!         write (iout,'(a,5i4)') '(subdims) ndim =',(no%ndim(f),f=1,no%nmodes)
!!         write (iout,'(a,i6,x,i6)') 'lenght=',&
!!          product(no%ndim)*no%navecs,no%z(no%navecs,2)-no%z(no%navecs,1)+1
!!         write (iout,'(a,i7,x,i7)') 'z0 zf=',no%z(no%navecs,1),no%z(no%navecs,2)
!        Print node information (primitive and non-primitive)
!        END Extra printing for debugging

         if(j.gt.1) write(iout,*)
         write (iout,'(a)') '# node : no. of SPFs : submodes : subdims'
         write (iout,'(a)',advance='no') '# '
         write (iout,'(i4,a,i6,a)',advance='no') no%num,' : ',no%navecs,' : '
         if (no%isprimitive) then
            do f=1,no%nmodes
               write (iout,'(a,1x)',advance='no') trim(mlbl(no%degf(f)))
            enddo
         else
            do f=1,no%nmodes
               ch => no%submodes(f)%p
               write (iout,'(a,i0,1x)',advance='no') '@',ch%num
            enddo
         endif
         write (iout,'(a,9i4)',advance='no') ' : ',(no%ndim(f),f=1,no%nmodes)
         write (iout,*)
!        Print header for ML
         if(fs.gt.2.d0) then
            write(iout,'(a)',advance='no') '    Time[fs]'
         else
            write(iout,'(a)',advance='no') '    Time[au]'
         endif
         if(j.eq.1) then
            write(iout,'(a,i0,a)',advance='no') '     ||psi||^2     ||A1||^2'
            write(iout,'(a,i0,a)')              '       ||psi||       ||A1||'
         else
            do f=1,no%navecs
               if (no%isprimitive) then
                  write(iout,'(a,i0,a)',advance='no') '     ||SPF',f,'|| '
               else
                  write(iout,'(a,i0,a)',advance='no') '      ||A',f,'||  '
               endif
            enddo
            write (iout,*)
         endif
      endif ! lhead.le.1

!     Calculate norm
      if (lhead.ge.1) then
         if(j.eq.1) then
            anorm2=0.d0
            do l=no%z(1,1),no%z(1,2)
               anorm2 = anorm2 + dble(psi(l))**2 + dimag(psi(l))**2
            enddo
            call spfovl(ovlmat,psi,psi,agmat,agmat,0)
            call overlap_1(psi,psi,tree,ovl,ovlmat)
         else
             write (iout,'(f12.4)',advance='no') time
!            Loop over A vectors
             do f=1,no%navecs
                norm=0.d0
                do l=no%z(f,1),no%z(f,2)
                   norm = norm + dble(psi(l))**2 + dimag(psi(l))**2
                enddo
                norm = dsqrt(norm)
                write (iout,'(1x,f13.9)',advance='no') norm
             enddo
             write (iout,*)
         endif
      endif ! lhead.ge.1

      return
    end subroutine norm0_tree

   function croco_tree(nham,time,tree,psi1,psi2) result(cc)
      ! compute <psi1|O|psi2> for two ML wavefunctions having
      ! EXACTLY the same tree structure.
      ! The number of SPF's per layer must be the SAME.
      ! The operator O must have been set by setoperml.

      implicit none 

      complex(cdbl)                :: cc
      integer,intent(in)           :: nham
      real(dbl),intent(in)         :: time
      type(wfnode_t),intent(inout) :: tree
      complex(cdbl),intent(in)     :: psi1(:),psi2(:)

      integer                   :: i,k,sdm,z0,zf,ki,kf,h
      type(wfnode_t),pointer    :: no
      complex(cdbl),allocatable :: work(:),hpsi2(:)

      ! compute <psi1|O|psi2> terms for all nodes in bottom-up order
      do i=1,tree%tot_nodes
         no => tree%swap_up(i)%p
         call compute_cc_node(no,psi1,psi2)
      enddo

      ! compute <psi1|O|psi2> for top-node
      sdm = product(tree%ndim)

      allocate(hpsi2(sdm))
      z0 = tree%z(1,1)
      zf = tree%z(1,2)
      ki = zham(nham)
      kf = zham(nham)+kzahl(nham)-1
      hpsi2 = CZERO

!$OMP parallel default(shared) private(k,work)
      allocate(work(sdm))

!$OMP do reduction(+:hpsi2)
      do k=ki,kf
         work = psi2(z0:zf)
         call hprod_phi(tree%hterms(k,:),work)
         h = hamilton(fkoe,k)
         if (htmtype(h) .ne. 1) then
            work(:) = koeff(k)*work(:)
         endif
         hpsi2 = hpsi2 + work
      enddo
!$OMP end do
      deallocate(work)

!$OMP WORKSHARE
      cc = dot_product(psi1(z0:zf),hpsi2)
!$OMP END WORKSHARE
!$OMP END PARALLEL

      ! add contributions from MLOps
      call compute_all_uterms2_tree(time,nham,tree,psi1,psi2)
      cc = cc + compute_all_mlop_etot(time,nham)

      deallocate(hpsi2)
      deallocate(work)
   end function croco_tree

   subroutine compute_cc_node(no,psi1,psi2)
   ! Like compute_hterms_node, but with different bra and ket.
      type(wfnode_t),intent(inout) :: no
      complex(cdbl),intent(in)     :: psi1(:),psi2(:)

      integer                   :: m,ndim,sdm,k,i,j,z0,zf
      integer                   :: ki,kf,nham
      logical                   :: lunit
      type(wfnode_t),pointer    :: ch
      complex(cdbl),allocatable :: work(:)

      if (no%isprimitive) return

! assumes system Hamiltonian
      nham = 1
      ki = zham(nham)
      kf = zham(nham)+kzahl(nham)-1
      do m=1,no%nmodes
         ch => no%submodes(m)%p
         ndim = no%ndim(m)
         sdm  = product(ch%ndim)
!$OMP    parallel default(shared) private(work,k,lunit,i,z0,zf,j)
         allocate(work(sdm))
!$OMP    do schedule(runtime)
         do k=ki,kf ! loop over k-terms
            lunit = .false.
            if (associated(no%isunit)) then
               if(no%isunit(k,m)) lunit=.true.
            endif
            do i=1,ndim
               z0 = ch%z(i,1)
               zf = ch%z(i,2)
               ! compute h_k|phi2_i>
               if(ch%isprimitive) then
                  if (lunit) then
                     work = psi2(z0:zf)
                  else
                     call hcall(k,ch%m,work,psi2(z0:zf),sdm)
                  endif
               else
                  work = psi2(z0:zf)
                  call hprod_phi(ch%hterms(k,:),work)
               endif
               ! do matrix elements <phi1_j|h_k|phi2_i>.
               ! No symmetry (hermitian, antihermitian,...) is assumed.
               do j=1,ndim
                  z0 = ch%z(j,1)
                  zf = ch%z(j,2)
                  no%hterms(k,m)%p%matrix(j,i) = dot_product(psi1(z0:zf),work)
               enddo
            enddo
         enddo
!$OMP    enddo
         deallocate(work)
!$OMP    end parallel
      enddo

   end subroutine compute_cc_node


end module mlanalysis
