!=======================================================================
module wftrafo
!=======================================================================
!
! Tranformations of the multi-layer wavefunction.
!
! -- FO 07/2013
!
!=======================================================================

use wftree
use htermmod
use linear
use mltensor
use gridutil
use timing
use global
use globmemmod
use psidef, only: dmatdim

use mltypes
use mlbase

implicit none
private

public :: natorb_tree,     &
          renorm_tree,     &
          reortho_tree

contains !==============================================================


!-----------------------------------------------------------------------
subroutine natorb_tree(tree,psi)
!-----------------------------------------------------------------------
! Transform the whole multi-layer wavefunction to natural orbitals.
! NOTE: This needs the eigenvectors of the density matrices, so make
! sure that compute_dmat_tree has been called before this.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(inout) :: tree
   complex(cdbl),intent(inout)  :: psi(:)
   type(wfnode_t),pointer       :: node
   integer                      :: m
   integer,save                 :: tid = 0

   call get_timer(tid,"natorb_tree")
   call start_timer(tid)
   do m=1,tree%tot_nodes
      node => tree%swap_postorder(m)%p
      call natorb_node(node,psi)
   enddo
   call stop_timer(tid)

end subroutine natorb_tree


!-----------------------------------------------------------------------
subroutine natorb_node(node,psi)
!-----------------------------------------------------------------------
! Transform the SPFs of the *children* of this node to natural orbitals,
! and adjust the A-vectors at this node accordingly.
! See Chapter 2.1 in my notes.
!-----------------------------------------------------------------------
  use mmlib, only: mqxczz
   type(wfnode_t),intent(inout) :: node
   complex(cdbl),intent(inout)  :: psi(:)
   type(wfnode_t),pointer       :: child
   integer                      :: m,plen,nspf,z0,zf,i
   complex(cdbl),pointer        :: trfmat(:,:)
   complex(cdbl),allocatable    :: work(:)
   integer                      :: ijk(3)

   ! SPFs at primitive nodes are expressed in DVR bases, so there is
   ! nothing to do.
   if (node%isprimitive) return
   ! For each child node, use the eigenvectors of the density matrix
   ! to transform the SPFs.
   do m=1,node%nmodes
      child => node%submodes(m)%p
      plen = product(child%ndim)
      nspf = child%navecs
      allocate(work(plen*nspf))
      trfmat => node%trfmat(m)%p%matrix
      z0 = child%z(1,1)
      zf = child%z(nspf,2)
      call mqxczz(psi(z0:zf),trfmat,work,plen,nspf)
      psi(z0:zf) = work
      deallocate(work)
   enddo
   ! Do the inverse transform to the A-vectors of this node.
   do i=1,node%navecs
      z0 = node%z(i,1)
      zf = node%z(i,2)
      if (node%istop .and. dentype .eq. 2) then
         plen = product(node%ndim)
         do m=1,node%nmodes
            trfmat => node%trfmat(m)%p%matrix
! act to LHS with adjoint
            ijk = ijk_shape(m,node%ndim)
            ijk(1)=ijk(1)*plen
            call matrix_tensor_compl_ad(trfmat, psi(z0:zf), ijk)
! act to RHS with transpose
            ijk = ijk_shape(m,node%ndim)
            ijk(3)=ijk(3)*plen
            call matrix_tensor_compl_tp(trfmat, psi(z0:zf), ijk)
         enddo
      else
! (This is like hprod_phi, but with transposed matrices.)
         do m=1,node%nmodes
            ijk = ijk_shape(m,node%ndim)
            trfmat => node%trfmat(m)%p%matrix
            call matrix_tensor_compl_tp(trfmat, psi(z0:zf), ijk)
         enddo
      endif
   enddo

end subroutine natorb_node


!-----------------------------------------------------------------------
subroutine renorm_tree(tree,psi,norm)
!-----------------------------------------------------------------------
! Normalizes the total WF to the given norm (default: one). As the
! norm is given by the (single) avec of the topnode (assuming
! orthonormal SPFs), this just normalizes this avec.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)     :: tree
   complex(cdbl),intent(inout)   :: psi(:)
   real(dbl),intent(in),optional :: norm
   integer                       :: z0,zf
   real(dbl)                     :: norm1,invnorm

   if (.not. tree%istop) return
   if (present(norm)) then
      norm1 = norm
   else
      norm1 = 1.d0
   endif
   z0 = tree%z(1,1)
   zf = tree%z(1,2)
   invnorm = norm1/sqrt(dble(dot_product(psi(z0:zf),psi(z0:zf))))
   psi(z0:zf) = invnorm*psi(z0:zf)

end subroutine renorm_tree


!-----------------------------------------------------------------------
subroutine reortho_tree(tree,psi)
!-----------------------------------------------------------------------
! Makes the SPFs throughout the tree orthonormal.
! More radical than the simle Gram-Schmidt and WF is significantly 
! changed.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)   :: tree
   complex(cdbl),intent(inout) :: psi(:)
   type(wfnode_t),pointer      :: node
   integer                     :: m
   integer,save                :: tid = 0

   call get_timer(tid,"reortho_tree")
   call start_timer(tid)
   ! Orthonormalize nodes from bottom to top.
   do m=1,tree%tot_nodes
      node => tree%swap_postorder(m)%p
      call reortho_node(node,psi)
   enddo
   ! For the top node, one could only renormalize the avec, but maybe
   ! that is not wanted. We won't do it here. See renorm_tree.
   call stop_timer(tid)

end subroutine reortho_tree


!-----------------------------------------------------------------------
subroutine reortho_node(node,psi)
!-----------------------------------------------------------------------
! Makes the SPFs at each child of this node orthonormal, and adjusts
! the avecs at this node accordingly.
! See Chapter 2.2 in my notes.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)   :: node
   complex(cdbl),intent(inout) :: psi(:)
   type(wfnode_t),pointer      :: child
   integer                     :: m,plen,nspf,info,i,z0,zf
   type(hterm_tp),allocatable  :: rterm(:)

   ! Leaf nodes have no children.
   if (node%isprimitive) return
   ! Allocate space for transformation matrices.
   allocate(rterm(node%nmodes))
   do m=1,node%nmodes
      rterm(m)%p => new_hterm(2,node%ndim(m))
   enddo
   ! Loop over children.
   do m=1,node%nmodes
      child => node%submodes(m)%p
      plen = product(child%ndim)
      nspf = child%navecs
      z0 = child%z(1,1)
      zf = child%z(nspf,2)
      ! Interpret the array of SPFs at this child as a matrix,
      ! and perform a QR decomposition of this matrix. The Q matrix
      ! will have orthonormal columns, which we will use as the new
      ! SPFs.
      call zqr(plen,nspf,psi(z0:zf),rterm(m)%p%matrix,info)
      if (info /= 0) then
         write (*,'(a,i0,a,i0)') "node%num = ",child%num,", info =",info
         call stop_code("wftrafo_m", "reortho_node", "QR decomposition failed")
      endif
      ! The child's SPFs have been overwritten with the new,
      ! orthonormal ones (from Q).
   enddo
   ! Transform the A-vectors at this node using the Rs of the children.
   do i=1,node%navecs
      z0 = node%z(i,1)
      zf = node%z(i,2)
      if (node%istop .and. dentype .eq. 2) then
          call chprod_phi2(rterm,psi(z0:zf))
          call hprod_phi2(rterm,psi(z0:zf))
      else
          call hprod_phi(rterm,psi(z0:zf))
      endif
   enddo
   ! Now the SPFs at this node are *unchanged* -- only their representation
   ! (avecs + child-SPFs) has changed.

   ! Clean up.
   do m=1,node%nmodes
      call deallocate_hterm(rterm(m)%p)
   enddo
   deallocate(rterm)

end subroutine reortho_node

!=======================================================================
end module wftrafo
