!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: wftree.f90 518 2013-11-05 18:32:34Z frank $

module wftree

    use global
    use globmemmod
    use channels
    use treemod
    use linear
    use logging

    use mltypes
    use dvrdatmod
    use griddatmod
    use operdef
    use hpsimod
    use logdat

    private
    public  :: new_wftree,                    &
               dgldim_wftree,                 &
               dgldim_wftree_mctdh,           &
               adim_wftree,                   &
               atopdim_wftree,                &
               compare_wftrees,               &
               tree_from_wftree,              &
               wftree_from_tree,              &
               wftree_set_navecs,             &
               wftree_init_pmodes,            &
               wftree_set_pointers,           &
               compare_tapes

! unused routine
    public  :: totphidim_wftree

    logical,private  :: debug=.false.

    contains

! -----------------------------------------------------------------------------
! TREE RELATED PROCEDURES
! -----------------------------------------------------------------------------
    function new_wftree(tape,dof_order) result(ttop)
    ! - input:
    !   tape: array of integers. This array contains all information related to
    !           the tree structure in a convenient codification for further
    !           processing.
    ! - returns:
    !     type node, pointer to a tree structure
        implicit none
        integer,intent(in)   :: tape(:)
        interface
            function dof_order(pos,mode) result(ndof)
                use global
                integer,intent(in)  :: pos
                integer,intent(in)  :: mode
                integer             :: ndof
            end function dof_order
        end interface
        optional                :: dof_order
        type(wfnode_t), pointer :: ttop
        integer :: i,it

        if(debug) then
            write(6,*) 'TAPE:'
            it=0
            do i=1,size(tape)
                if(tape(i)==-2) it=i
            end do
            write(6,10) (tape(i),i=1,it)
        end if

        ttop => new_wfnode()

        call link_tree_structure(ttop,tape)
        call swap_lists(ttop)
        call count_modes_tree(ttop,dof_order)
        call workspace(ttop)
        call swap_lists_phi(ttop)
        if(debug) write(6,*) 'T R E E     D O N E'
        10 format(12I3)
    end function new_wftree

    function new_wfnode() result(nt)
    ! return pointer to a new node
        implicit none
        type(wfnode_t),pointer :: nt
        allocate(nt)
        nt%nmodes=1
        nt%navecs=1
        nt%tot_nodes=1
        nt%nsubcoord=0
        nt%isprimitive=.false.
        nt%istop=.false.
        nt%submodes => null()
        nt%parent   => null()
        nt%z        => null()
        nt%hterms   => null()
        nt%isunit   => null()
        nt%allunit  => null()
        nt%isuncor  => null()
        nt%hasuncor => null()
        nt%htmtype  => null()
        nt%mfields  => null()
        nt%dmat     => null()
        nt%invdmat  => null()
        nt%pop      => null()
        nt%regpop   => null()
        nt%trfmat   => null()
        nt%overlap  => null()
        nt%mfsum    => null()
        nt%mtraces  => null()
        nt%shtraces => null()
        nt%swap_down=> null()
        nt%swap_up  => null()
        nt%swap_phi => null()
        nt%swap_preorder   => null()
        nt%swap_preorder   => null()
        nt%swap_levelorder => null()
        nt%tid_rhs     = 0
        nt%tid_mfields = 0
        nt%tid_hterms  = 0
        nt%tid_dmat    = 0
        nt%tid_opsi    = 0
        nt%tid_mfsum   = 0
        nt%tid_uterms  = 0
        nt%tid_wterms  = 0
    end function new_wfnode

!----- T R E E    D O N E !!
! -----------------------------------------------------------------------------


    subroutine link_tree_structure(ttop,tape)
        ! Process TAPE and provide a new tree structure
        implicit none
        type(wfnode_t),intent(inout),target :: ttop
        integer,intent(in)                  :: tape(:)
        integer                             :: i,j,total_nodes
        type(wfnode_t), pointer             :: tnew,tcurrent

        total_nodes = 1
        ttop%layer  = 0

        tnew     => ttop
        tcurrent => tnew
        tnew%topnode => ttop

        allocate(ttop%tape(size(tape)))

        ttop%tape = tape
        ttop%istop = .true.
        ttop%num = total_nodes

        j=1
        do while(tape(j)/=-2)
            select case(tape(j))
                case(1:) ! ndim of current node is specified
                    tcurrent%nmodes = tape(j)
                    allocate(tcurrent%ndim(tcurrent%nmodes))
                    do i=1,tcurrent%nmodes
                        tcurrent%ndim(i)=tape(j+i)
                    end do
                    j = j + tcurrent%nmodes + 1

                case(:0) ! control instruction
                    select case(tape(j))
                        case(-1) ! go one layer down creating a new node
                            if (.not. associated(tcurrent%submodes)) then
                                 allocate(tcurrent%submodes(tcurrent%nmodes))
                                 do i=1,tcurrent%nmodes
                                     tcurrent%submodes(i)%p => null()
                                 end do
                            end if
                            tnew         => new_wfnode()
                            tnew%parent  => tcurrent
                            tnew%topnode => tcurrent%topnode
                            total_nodes  =  total_nodes + 1
                            tnew%num     =  total_nodes
                            tnew%layer   =  tnew%parent%layer + 1
                            j=j+1
                            tnew%navecs = tcurrent%ndim(tape(j))
                            tcurrent%submodes(tape(j))%p => tnew
                            tcurrent => tnew
                            j=j+1

                        case(0) ! jump one layer up
                            tcurrent => tcurrent%parent
                            j=j+1

                        case(-2)
                            exit
                        case default
                          write(6,*) 'tape instruction out of bounds: ',tape(j)
                          exit
                    end select
           end select
        end do
        ttop%tot_nodes=total_nodes
    end subroutine link_tree_structure


    subroutine count_modes_tree(ttop,dof_order)
        implicit none
        ! Identify primitive modes and degrees of freedom.
        ! For nodes with terminal (primitive) components: set degf array
        ! according to spfdof(n,m) in griddat.inc
        ! For the rest of nodes its components are set to 0
        ! Allocate and set pdim(:) in the top node

        type(wfnode_t),intent(inout):: ttop
        interface
            function dof_order(pos,mode) result(ndof)
                use global
                integer,intent(in)  :: pos
                integer,intent(in)  :: mode
                integer             :: ndof
            end function dof_order
        end interface
        optional                    :: dof_order
        type(wfnode_t),pointer      :: tcurrent
        integer                     :: numdegf
        integer                     :: nummod,numnod
        integer                     :: i,j

        ! Find primitive nodes and primitive degfs and count them
        ! non-primitive nodes get a "mode" number > nmodes
        numdegf = 1
        nummod = 1
        numnod = nmode +1 
        do i=1,ttop%tot_nodes
            tcurrent => ttop%swap_down(i)%p
            allocate(tcurrent%degf(tcurrent%nmodes))
            if(associated(tcurrent%submodes)) then ! sublayers for ALL or SOME modes exist
                tcurrent%isprimitive = .false.
                tcurrent%m = numnod
                numnod = numnod + 1
                do j=1,tcurrent%nmodes
                    ! check for mixed prim./non-prim. nodes and abort if found
                    if(.not.associated(tcurrent%submodes(j)%p)) then ! j is primitive
                        write(ilog,*) 'Attempt to define a mixed primitive, non-primitive node'
                        write(ilog,*) 'This feature is not implemented!'
                        write(ilog,*) 'Erroneous node: ',i, j
                        write(ilog,*) 'Stop execution'
                        stop
                    else ! set degf=0 for tcurrent and tell child its number
                        tcurrent%degf(j)=0
                        tcurrent%submodes(j)%p%nsubcoord=j
                    end if
                end do
            else ! primtive node
                tcurrent%isprimitive = .true.
                tcurrent%m = nummod
                nummod = nummod + 1
                do j=1,tcurrent%nmodes
                    if (present(dof_order)) then
                        tcurrent%degf(j) = dof_order(j,tcurrent%m)
                    else
                        tcurrent%degf(j) = numdegf
                    end if
                    numdegf = numdegf + 1
                end do
            end if
        end do
        ttop%tot_degfs = numdegf - 1
        ttop%tot_pnodes = nummod - 1

        ! Set pdim in the top node
        allocate(ttop%pdim(ttop%tot_degfs))
        numdegf = 1
        do i=1,ttop%tot_nodes
            tcurrent => ttop%swap_down(i)%p
            if(tcurrent%isprimitive) then
                do j=1,tcurrent%nmodes
                    ttop%pdim(numdegf) = tcurrent%ndim(j)
                    numdegf = numdegf + 1
                end do
            end if
        end do
    end subroutine count_modes_tree


    function dgldim_wftree(self) result(l)
    ! compute the length of psi (total number of coefficients in
    ! all layers) from a given tree structure
    ! NB This does not take into account a GWP basis.
        implicit none
        type(wfnode_t),intent(in) :: self
        type(wfnode_t),pointer    :: tp
        integer                   :: l,i,j
        l=0
        do i=1,self%tot_nodes
            tp => self%swap_down(i)%p
            if (tp%istop .and. dentype .eq. 2) then
               j = product(tp%ndim)
               l = l + j*j*tp%navecs
            else if (tp%isprimitive .and. dentype .eq. 1) then
               j = product(tp%ndim)
               l = l + j*j*tp%navecs
            else
               l = l + product(tp%ndim)*tp%navecs
            endif
        end do
    end function dgldim_wftree


    function dgldim_wftree_mctdh(self,cntrl) result(l)
    ! compute the length of psi that corresponds to compacting to an MCTDH WF
    ! cntrl = 1: modes are combined, A vector keeps dimensionality
    ! cntrl = 2: primitive modes conserved, A vector grows
    ! cntrl = 3: multistate MCTDH
        implicit none
        type(wfnode_t),intent(in) :: self
        integer,intent(in)        :: cntrl
        type(wfnode_t),pointer    :: tp
        integer                   :: i,l,lphi,la
        select case(cntrl)

            case(1)
                write(6,*) 'dgldim_wftree_mctdh case(1) not implemented'
                stop 1

            case(2)
                lphi = 0
                la = 1
                do i=1,self%tot_nodes
                    tp => self%swap_down(i)%p
                    if(tp%isprimitive) then
                        lphi = lphi + product(tp%ndim)*tp%navecs
                        la = la*tp%navecs
                    end if
                end do
                l = lphi + la

            case(3)
                lphi = 0
                la = 1
                do i=1,self%tot_nodes
                    tp => self%swap_down(i)%p
                    if(tp%isprimitive) then
                        if (tp%parent%istop .and. tp%nsubcoord==2) cycle
                        lphi = lphi + product(tp%ndim)*tp%navecs
                        la = la*tp%navecs
                    end if
                end do
                l = (la + lphi)*self%submodes(1)%p%navecs

            case default
                write(6,*) 'dgldim_wftree_mctdh unknown value cntrl =',cntrl
                stop 1

        end select
    end function dgldim_wftree_mctdh


    function adim_wftree(t) result(l)
    ! compute the length of the A vector (coefficients in
    ! all layers except primitive SPFs)
        implicit none
        type(wfnode_t),intent(in) :: t
        type(wfnode_t),pointer    :: tp
        integer                   :: l,i,j
        l=0
        do i=1,t%tot_nodes
            tp => t%swap_down(i)%p
            if (tp%istop .and. dentype .eq. 2) then
               j = product(tp%ndim)
               l = l + j*j*tp%navecs
            else if(.not. tp%isprimitive) then
                l = l+product(tp%ndim)*tp%navecs
            end if
        end do
    end function adim_wftree


    function atopdim_wftree(t) result(l)
    ! compute the length of the top layer A vector (block(1))
        implicit none
        type(wfnode_t),intent(in) :: t
        integer                   :: l
        l = product(t%ndim)*t%navecs
    end function atopdim_wftree


    function totphidim_wftree(t) result(l)
    ! compute the length of the SPF part of psi
        implicit none
        type(wfnode_t),intent(in) :: t
        type(wfnode_t),pointer    :: tp
        integer                   :: l,i
        l=0
        do i=1,t%tot_nodes
            tp => t%swap_down(i)%p
            if(tp%isprimitive) then
                l = l+product(tp%ndim)*tp%navecs
            end if
        end do
    end function totphidim_wftree


    subroutine workspace(t)
    ! compute the needed workspace for the calculation and set t%workcdim
    ! t: top node of a tree
        implicit none
        type(wfnode_t),intent(inout) :: t
        type(wfnode_t),pointer       :: tp
        integer                      :: i,f,tmpw,tmpw2,j
        tmpw  = 0
        tmpw2 = 0
        do i=1,t%tot_nodes
            tp => t%swap_down(i)%p
            j = product(tp%ndim)
            if (tp%istop .and. dentype .eq. 2) then
               j = j**2
            endif

            tmpw = max(tmpw,product(tp%ndim)*tp%navecs)
            if(.not.tp%isprimitive) then
                do f=1,tp%nmodes
                    tmpw2 = max(tmpw2,tp%ndim(f))
                end do
            end if
        end do
        t%workcdim  = 2*tmpw
        t%workcdim2 = tmpw2
    end subroutine workspace


    subroutine wftree_set_pointers(ttop,dgldim)
     ! for each node of the tree start and end positions
     ! of its coefficients along w%psi are computed and stored
         implicit none
         type(wfnode_t),intent(inout) :: ttop
         type(wfnode_t),pointer       :: tcur
         integer,intent(out)          :: dgldim
         integer                      :: point,i,j
         integer                      :: philen
         point=1
         ! non-primitive nodes
         do i=1,ttop%tot_nodes
             tcur => ttop%swap_down(i)%p
             if(tcur%isprimitive) cycle
             allocate(tcur%z(tcur%navecs,2))
             allocate(tcur%zr(tcur%navecs,2))
             if (tcur%istop .and. dentype .eq. 2) then
                philen = product(tcur%ndim)
                philen=philen**2
             else
                philen = product(tcur%ndim)
             endif
             do j=1,tcur%navecs
                 tcur%z(j,1) = point
                 tcur%z(j,2) = point + philen - 1
                 point = point + philen
             end do
             tcur%zr = tcur%z - tcur%z(1,1) + 1
         end do
         ! primitive nodes
         do i=1,ttop%tot_nodes
             tcur => ttop%swap_down(i)%p
             if(.not.tcur%isprimitive) cycle
             allocate(tcur%z(tcur%navecs,2))
             allocate(tcur%zr(tcur%navecs,2))
             philen = subdim(tcur%m)
             if (gwpm(tcur%m)) then
                philen = philen + tcur%ndim(1)
             endif
             do j=1,tcur%navecs
                 tcur%z(j,1) = point
                 tcur%z(j,2) = point + philen - 1
                 point = point + philen
             end do
             tcur%zr = tcur%z - tcur%z(1,1) + 1
         end do
         dgldim = point-1
     end subroutine wftree_set_pointers

! -----------------------------------------------------------------------------
! TREE TRAVERSAL ROUTINES
!
! Different ways to traverse the tree are coded here. Instead of really
! following the tree in a certain order by using the pointers between nodes, we
! build lists that contain pointers to the nodes in the desired order and then
! loop over these lists when needed.
! -----------------------------------------------------------------------------
    subroutine swap_lists(t)
    ! set swap lists to traverse the tree top-to-bottom and bottom-to-top
        implicit none
        type(wfnode_t),intent(inout),target :: t
        integer                             :: i

        allocate(t%swap_down(t%tot_nodes))
        allocate(t%swap_up(t%tot_nodes))
        allocate(t%swap_levelorder(t%tot_nodes))
        allocate(t%swap_preorder(t%tot_nodes))
        allocate(t%swap_postorder(t%tot_nodes))
        do i=1,t%tot_nodes
            t%swap_down(i)%p        => null()
            t%swap_up(i)%p          => null()
            t%swap_levelorder(i)%p  => null()
            t%swap_preorder(i)%p    => null()
            t%swap_postorder(i)%p   => null()
        end do

        call preorder_traversal(t,t)
        call postorder_traversal(t,t)

        t%swap_levelorder(1)%p => t
        call levelorder_traversal(t,t)

        ! Set up the up and down lists.
        ! down -> preorder (depth-first)
        ! up   -> postorder
        do i=1,t%tot_nodes
            t%swap_down(i)%p => t%swap_preorder(i)%p
            t%swap_up(i)%p   => t%swap_postorder(i)%p
        end do
    end subroutine swap_lists


    subroutine swap_lists_phi(t)
        implicit none
        type(wfnode_t),intent(inout),target :: t
        integer                             :: i,m
        type(wfnode_t), pointer             :: tz

        allocate(t%swap_phi(t%tot_nodes))
        do i=1,t%tot_nodes
            t%swap_phi(i)%p => null()
        end do
        ! Set up the phi list. The ordering is:
        !  - Primitive modes (the usual mctdh ones)
        !  - Non-primitive modes in intermediate layers
        !  - Top node
        ! This list is used when computing the RHS of the
        ! equations of motion for specific modes in CMF schemes.
        m=1
        do i=1,t%tot_nodes
            tz => t%swap_down(i)%p
            if(tz%isprimitive) then
                t%swap_phi(m)%p => tz
                m=m+1
            end if
        end do
        do i=1,t%tot_nodes
            tz => t%swap_down(i)%p
            if(tz%isprimitive .or. tz%istop) cycle
            t%swap_phi(m)%p => tz
            m=m+1
        end do
        t%swap_phi(m)%p => t%swap_down(1)%p
    end subroutine swap_lists_phi


    recursive subroutine preorder_traversal(t,ttop)
        ! depth-first traversal, recursively:
        !   1. visit node
        !   2. traverse subnodes left to right
        implicit none
        type(wfnode_t),target  :: t
        type(wfnode_t)         :: ttop
        integer                :: i,j
        ! search next free spot in list
        i=1
        do while(associated(ttop%swap_preorder(i)%p))
            i=i+1
        end do
        ! visit (node is recorded)
        ttop%swap_preorder(i)%p => t
        ! traverse left to right
        if(associated(t%submodes)) then ! there are sublayers for all or SOME modes
            do j=1,t%nmodes
                if(associated(t%submodes(j)%p)) then ! there is a sublayer below mode j
                    call preorder_traversal(t%submodes(j)%p,ttop) ! traverse
                end if
            end do
        end if
    end subroutine preorder_traversal


    recursive subroutine postorder_traversal(t,ttop)
        ! postorder traversal, recursively:
        !   1. traverse subnodes left to right
        !   2. visit node
        implicit none
        type(wfnode_t),target :: t
        type(wfnode_t)        :: ttop
        integer               :: i,j
        ! traverse left to right
        if(associated(t%submodes)) then ! there are sublayers for all or SOME modes
            do j=1,t%nmodes
                if(associated(t%submodes(j)%p)) then ! there is a sublayer below mode j
                    call postorder_traversal(t%submodes(j)%p,ttop)
                end if
            end do
        end if
        ! search next free spot in list
        i=1
        do while(associated(ttop%swap_postorder(i)%p))
            i=i+1
        end do
        ! visit (node is recorded)
        ttop%swap_postorder(i)%p => t
    end subroutine postorder_traversal


    recursive subroutine levelorder_traversal(t,ttop)
        ! breadth-first traversal, recursively
        !   1. visit all children left to right (if any)
        !   2. traverse left to right
        !   note: the present node has been visited in the
        !         previous recursive call. Therefore the top
        !         node must be visited outside this routine
        implicit none
        type(wfnode_t),intent(in),target :: t
        type(wfnode_t),intent(inout)     :: ttop
        integer                          :: i,j
        ! search next free spot in list
        i=1
        do while(associated(ttop%swap_down(i)%p))
            i=i+1
        end do
        ! visit subnodes left to right
        if(associated(t%submodes)) then ! there are sublayers for all or SOME modes
            do j=1,t%nmodes
                if(associated(t%submodes(j)%p)) then ! there is a sublayer below mode j
                    ttop%swap_down(i)%p => t%submodes(j)%p
                    i=i+1
                end if
            end do
            ! traverse subnodes left to right
            do j=1,t%nmodes
                if(associated(t%submodes(j)%p)) then ! there is a sublayer below mode j
                    call levelorder_traversal(t%submodes(j)%p,ttop)
                end if
            end do
        end if
    end subroutine levelorder_traversal


   function compare_wftrees(tree1,tree2) result(match)
      ! Compare two trees regarding structure and number of SPFs.
      ! Result: match=0 - trees have different structure
      !         match=1 - same structure, different SPF no.s
      !         match=2 - identical trees
      type(wfnode_t),intent(in) :: tree1,tree2
      integer                   :: match
      match = compare_nodes(tree1,tree2)
   end function compare_wftrees


   recursive function compare_nodes(node1,node2) result(match)
      type(wfnode_t),intent(in) :: node1,node2
      integer                   :: match
      type(wfnode_t),pointer    :: ch1,ch2
      integer                   :: m,chmatch
      match=2
      if (node1%navecs /= node2%navecs) match=1
      if (node1%isprimitive) then
         if (node2%isprimitive) then
            ! both primitive
            if (node1%m /= node2%m) match=0
         else
            ! 1 primitive, 2 not
            match=0
         endif
      else
         if (node2%isprimitive) then
            ! 2 primitive, 1 not
            match=0
         else
            ! both not primitive
            if (node1%nmodes /= node2%nmodes) then
               match=0
            else
               ! recurse
               do m=1,node1%nmodes
                  ch1 => node1%submodes(m)%p
                  ch2 => node2%submodes(m)%p
                  chmatch = compare_nodes(ch1,ch2)
                  match = min(match,chmatch)
                  if (match==0) exit
               enddo
            endif
         endif
      endif
   end function compare_nodes

!===============================================================================


!-------------------------------------------------------------------------------
   function tree_from_wftree(wftree) result(tree)
!-------------------------------------------------------------------------------
      type(tree_t),pointer      :: tree
      type(wfnode_t),intent(in) :: wftree
      type(node_t),pointer      :: root
      root => node_from_wfnode(wftree)
      tree => make_tree(root)
   end function tree_from_wftree


   recursive function node_from_wfnode(wfnode) result(node)
      type(node_t),pointer      :: node
      type(wfnode_t),intent(in) :: wfnode
      type(node_tp),allocatable :: child(:)
      type(wfnode_t),pointer    :: wfchild
      integer                   :: c
      if (wfnode%isprimitive) then
         node => make_leaf()
      else
         allocate(child(wfnode%nmodes))
         do c=1,wfnode%nmodes
            wfchild => wfnode%submodes(c)%p
            child(c)%p => node_from_wfnode(wfchild)
         enddo
         node => make_node(child)
         deallocate(child)
      endif
   end function node_from_wfnode


!-------------------------------------------------------------------------------
   function wftree_from_tree(tree) result(wftree)
!-------------------------------------------------------------------------------
      type(wfnode_t),pointer  :: wftree
      type(tree_t),intent(in) :: tree
      integer                 :: n
      type(wfnode_t),pointer  :: wfnode
      type(node_t),pointer    :: node

      ! Clone the tree skeleton.
      wftree => wfnode_from_node(tree%root)
      wftree%istop = .true.

      ! Set node counts.
      wftree%tot_nodes = tree%nnode
      wftree%tot_pnodes = tree%nleaf

      ! Initialize swap lists.
      call swap_lists(wftree)
      call swap_lists_phi(wftree)

      ! Set tree-related information in nodes
      do n=1,tree%nnode
         wfnode => wftree%swap_preorder(n)%p
         node => tree%preorder(n)%p
         wfnode%num = node%num
         wfnode%layer = node%layer
         wfnode%topnode => wftree
      enddo
      
      return
   end function wftree_from_tree


   recursive function wfnode_from_node(node) result(wfnode)
      type(wfnode_t),pointer  :: wfnode
      type(node_t),intent(in) :: node
      integer                 :: c
      type(node_t),pointer    :: child
      type(wfnode_t),pointer  :: wfchild

      wfnode => new_wfnode()
      wfnode%m = 0
      wfnode%degf => null()
      wfnode%navecs = 0

      if (node%isleaf) then
         wfnode%isprimitive = .true.
         ! No information about primitive mode available here.
         wfnode%nmodes = 0
         wfnode%ndim => null()

      else
         wfnode%isprimitive = .false.
         wfnode%nmodes = node%nchild
         allocate(wfnode%ndim(node%nchild))
         wfnode%ndim = 0
         allocate(wfnode%submodes(node%nchild))
         do c=1,node%nchild
            child => node%child(c)%p
            wfchild => wfnode_from_node(child)
            wfnode%submodes(c)%p => wfchild
            wfchild%nsubcoord = c
            wfchild%parent => wfnode
         enddo

      endif
      return
   end function wfnode_from_node


!-------------------------------------------------------------------------------
subroutine wftree_set_navecs(tree,navecs)
!-------------------------------------------------------------------------------
   type(wfnode_t),intent(inout) :: tree
   integer,intent(in)           :: navecs(:)
   integer                      :: n,c
   type(wfnode_t),pointer       :: node
   do n=1,tree%tot_nodes
      node => tree%swap_postorder(n)%p
      node%navecs = navecs(node%num)
      if (associated(node%submodes)) then
         do c=1,node%nmodes
            node%ndim(c) = node%submodes(c)%p%navecs
         enddo
      endif
   enddo
end subroutine wftree_set_navecs


!-------------------------------------------------------------------------------
subroutine wftree_init_pmodes(tree,dofsize,dofnums)
!-------------------------------------------------------------------------------
   type(wfnode_t),intent(inout) :: tree
   interface
      function dofsize(f) result(gdm)
         integer,intent(in) :: f
         integer            :: gdm
      end function dofsize
      function dofnums(m) result(fs)
         integer,intent(in)  :: m
         integer,allocatable :: fs(:)
      end function dofnums
   end interface
   integer :: n,nmode,ndof,f
   type(wfnode_t),pointer :: node
   integer,allocatable :: fs(:)

   nmode = 0
   ndof = 0
   do n=1,tree%tot_nodes
      ! Loop over all primitive nodes.
      node => tree%swap_preorder(n)%p
      if (.not.node%isprimitive) cycle
      ! Store mode number.
      nmode = nmode + 1
      node%m = nmode
      ! Get DOFs for this mode from MCTDH land, and store them in this node.
      fs = dofnums(nmode)
      node%nmodes = size(fs)
      allocate(node%degf(node%nmodes))
      node%degf = fs
      ! Store DOF grid sizes.
      allocate(node%ndim(node%nmodes))
      do f=1,node%nmodes
         node%ndim(f) = dofsize(fs(f))
      enddo
      ! Count total number of DOFs.
      ndof = ndof + node%nmodes
   enddo
   tree%tot_degfs = ndof

   ! Set pdim in the top node.
   allocate(tree%pdim(ndof))
   ndof = 0
   do n=1,tree%tot_nodes
      node => tree%swap_preorder(n)%p
      if(.not.node%isprimitive) cycle
      do f=1,node%nmodes
         ndof = ndof + 1
         tree%pdim(ndof) = node%ndim(f)
      enddo
   enddo

   return
end subroutine wftree_init_pmodes

  subroutine compare_tapes(tape,tape2,ltape,ltree,lentape,lentape2)

      implicit none
      integer             :: i,j
      integer,intent(out) :: lentape,lentape2
      integer,intent(in)  :: tape(maxtape),tape2(maxtape)
      logical,intent(out) :: ltape,ltree

      ltape=.false. ! ltape is true if the tapes differ.
      ltree=.false. ! ltree is true if topologies of the trees differ.
      lentape=0
      lentape2=0

      do i=1,maxtape
         if(tape(i).ne.tape2(i)) ltape=.true.
         if(tape(i).eq. -2) then
            lentape=i
         endif
         if(tape2(i).eq. -2) then
            lentape2=i
         endif
         if(lentape.gt.0 .and. lentape2.gt.0) goto 101
      enddo
      write(6,'(/a)') 'compare_tapes: No end of tape found!'
      write(6,'(a,2i7)')  'lentape, lentape2:', lentape,lentape2
      stop

 101  if(.not.ltape) return  ! tapes are identical.

      ltree=.true.
      ! tree topologies cannot be identical when the tape lengths differ.
      if(lentape.ne.lentape2) return

      ! check if the tree topologies are identical.
      j=1
 doj: do while(j.le.lentape)
         select case(tape(j))
            case(1:) ! ndim of current node is specified
               if(tape(j).ne.tape2(j)) exit doj
               i=tape(j) ! number of modes
               j=j+i+1   ! skip over numbers of modes or SPFs
            case(0)
               if(tape(j).ne.tape2(j)) exit doj
               j=j+1
            case(-1) ! go one layer down
               if(tape(j).ne.tape2(j)) exit doj
               j=j+1
               if(tape(j).ne.tape2(j)) exit doj
               j=j+1
            case(-2)
               ltree=.false. ! No problems found. Set ltree=.false.
               exit doj
         end select
      enddo doj

   end subroutine compare_tapes


end module wftree
