!!  Copyright (C) 2009 O. Vendrell
!!  MLCORE, Multilayer-MCTDH core modules for quantum dynamics
!!
!!  This file is part of MLCORE.
!!
!!  MLCORE is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  $Id: wftree.f90 518 2013-11-05 18:32:34Z frank $
!!
!! These routines require system information so have been split from wftree.f90

module wftree1

    use global
    use dglobal, only: dimtape,ldebugspawn
    use globmemmod
    use channels
    use treemod
    use htermmod
    use linear
    use logging

    use mltypes
    use dvrdatmod
    use griddatmod
    use operdef
    use psidef
    use hpsimod
    use logdat
    use psiutil, only: akcopy

    private
    public  :: allocate_wftree,               &
               allocate_wftree_h,             &
               allocate_wfnode,               &
               dispose_wftree,                &
               dispose_wftree_h,              &
               allocate_work,                 &
               dispose_work,                  &
               allocate_hterm_matrices,       &
               deallocate_hterm_matrices,     &
               allocate_overlap_matrices,     &
               deallocate_overlap_matrices,   &
               allocate_mtraces_matrices,     &
               allocate_mtraces_matrices2,    &
               allocate_zmtraces_matrices,    &
               deallocate_mtraces_matrices,   &
               allocate_shtraces_matrices,    &
               deallocate_shtraces_matrices,  &
               set_prefactor,                 &
               set_isrelax,                   &
               set_isrlxt,                    &
               set_hartree_avec,              &
               set_wfnode_htmtype,            &
               get_mode_htmtype,              &
               outpoptree,                    &
               outlowpop,                     &
               mlspwn_wftree,                 &
               spwn_wftree,                   &
               copy_wftree

    logical,private  :: debug=.false.

    contains

    subroutine allocate_wftree(t,ldmat,dicht1,dicht2,dicht3,dicht4)
        ! Allocate space for density matrices...
        ! t: top level node of tree structure
        ! nk: number of products of the operator
        ! lfields: if true allocate space for mfields and dmat
        implicit none
        type(wfnode_t),intent(inout) :: t
        logical, intent(in)          :: ldmat
        complex(dop), dimension(*)   :: dicht1,dicht2,dicht4
        real(dop), dimension(*)      :: dicht3
        type(wfnode_t),pointer   :: tp
        integer                  :: i

        if(t%isallocated) return

        ! space for grid populations
        allocate(t%pgpops(t%tot_degfs))
        do i=1,t%tot_degfs
            t%pgpops(i)%p => new_hterm(5,t%pdim(i))
        end do

        do i=1,t%tot_nodes
            tp => t%swap_up(i)%p
            call allocate_wfnode(tp,ldmat,dicht1,dicht2,dicht3,dicht4)
        end do

        t%isallocated = .true.

    end subroutine allocate_wftree

    subroutine allocate_wftree_h(t,lfields,useunit)
        ! Allocate space for hterms, mfields, ...
        ! according to the number of products (lines) of the operator
        ! t: top level node of tree structure
        ! nk: number of products of the operator
        ! lfields: if true allocate space for mfields and dmat
        implicit none
        type(wfnode_t),intent(inout) :: t
        logical, intent(in)          :: lfields,useunit
        type(wfnode_t),pointer   :: tp
        integer                  :: i

        if(t%ishallocated) return

        ! Each node holds h matrices that operate on its A vectors.
        ! The tensorial structure is used:
        ! Instead of holding a single h-matrix that operates on the
        ! whole A vector, nmodes matrices are stored, each operating
        ! on each of its dimensions.
        ! The nodes at the bottom of the tree must contain h-matrices that
        ! operate on the coefficients of the time-independent primitive basis.

        call allocate_work(t)

        do i=1,t%tot_nodes
           tp => t%swap_up(i)%p
           call set_wfnode_hunit(tp)
        end do

        do i=1,t%tot_nodes
            tp => t%swap_down(i)%p
            call set_wfnode_uncorrelated(tp)
        end do

        call allocate_hterm_matrices(t,useunit)
        if (lfields) call allocate_mfields(t)

        t%ishallocated = .true.
    end subroutine allocate_wftree_h

    subroutine allocate_work(this)
        ! allocate work space for this tree.
        implicit none
        type(wfnode_t),intent(inout) :: this
        if(.not. this%istop) return
        allocate(this%workc(this%workcdim))
        allocate(this%workr(this%workcdim))
        allocate(this%workc2(this%workcdim2,this%workcdim2))
    end subroutine allocate_work


    subroutine dispose_wftree(this)
        implicit none
        type(wfnode_t),intent(inout) :: this
        type(wfnode_t),pointer       :: np
        integer                      :: i
        if(.not. this%isallocated) return

        call dispose_work(this)

        do i=1,this%tot_nodes
            np => this%swap_up(i)%p
            call dispose_wfnode(np)
        end do
        this%isallocated = .false.
    end subroutine dispose_wftree

    subroutine dispose_wftree_h(this)
        implicit none
        type(wfnode_t),intent(inout) :: this
        type(wfnode_t),pointer       :: np
        integer                      :: i
        if(.not. this%ishallocated) return

        call dispose_work(this)

        call deallocate_hterm_matrices(this)
        call deallocate_mfields(this)

        do i=1,this%tot_nodes
           np => this%swap_up(i)%p
           deallocate(np%isunit)
           deallocate(np%allunit)
           deallocate(np%isuncor)
           deallocate(np%hasuncor)
        end do

        this%ishallocated = .false.
    end subroutine dispose_wftree_h


    subroutine dispose_work(this)
        ! deallocate work space for this tree.
        implicit none
        type(wfnode_t),intent(inout) :: this
        if(.not. this%istop) return
        if (associated(this%workc)) deallocate(this%workc)
        if (associated(this%workr)) deallocate(this%workr)
        if (associated(this%workc2)) deallocate(this%workc2)
    end subroutine dispose_work

! -----------------------------------------------------------------------------
! NODE RELATED PROCEDURES
! -----------------------------------------------------------------------------

    subroutine set_wfnode_hunit(no)
        ! Set which h-terms are unit operators in a node.
        ! This depends on the nodes below, so this routine must
        ! be called in a swap_up loop.
        implicit none
        type(wfnode_t),intent(inout) :: no
        type(wfnode_t),pointer        :: childf
        integer                       :: k,f,m
        integer                       :: nham
        allocate(no%isunit(ktot,no%nmodes))
        allocate(no%allunit(ktot))

      no%isunit  = .true.
      no%allunit = .true.

      do nham = 1,hamdim
!
! if the operator uses explicit evaluation of unit operators, set
! isunit=.false. ELSE set isunit according to operator type
! 
        if (nodiag(nham)) then
           do k=zham(nham),zham(nham)+khtot(nham)-1
              no%isunit(k,:)  = .false.
              no%allunit(k) = .false.
           enddo
           cycle 
        endif


        if(no%isprimitive) then

            do k=zham(nham),zham(nham)+khtot(nham)-1
! take isunit flags from diag flag of primitive modes.               
                m = no%m
                if (.not. diag(m,k)) then
                   do f=1,no%nmodes
                      no%isunit(k,f) = .false.
                      no%allunit(k)  = .false.
                   enddo
                end if

            end do

        else

            do k=zham(nham),zham(nham)+khtot(nham)-1
            !do k=1,ktot
                do f=1,no%nmodes
                    childf => no%submodes(f)%p
                    if(.not.childf%allunit(k)) then
                        no%isunit(k,f) = .false.
                        no%allunit(k)  = .false.
                    end if
                end do
                if(debug) then
                    do f=1,no%nmodes
                        write(6,*) 'k,nmode,isunit:',k,f,no%isunit(k,f)
                    end do
                end if
            end do

        end if

      enddo
    end subroutine set_wfnode_hunit


    subroutine set_wfnode_htmtype(no)
        ! For primitive nodes, set type of overall operator terms
        ! i.e. of prod_m hterms(k,f)
        ! This information is used for summing the mean-field terms
        ! in CMF calculations.
        ! Meaningful values:
        !   0 = zero
        !   1 = unit
        !   2 = diagonal
        !  -1 = anything else
        implicit none
        type(wfnode_t),intent(inout)  :: no
        integer                       :: k

        if (no%isprimitive) then
            allocate(no%htmtype(ktot))
            do k=1,ktot
                no%htmtype(k) = get_mode_htmtype(no,k)
            end do
        endif
    end subroutine set_wfnode_htmtype

    function get_mode_htmtype(no,k) result(kt)
        integer                   :: kt,h
        type(wfnode_t),intent(in) :: no
        integer,intent(in)        :: k
        integer :: f,kft
        kt = 1
        do f=1,no%nmodes
            h=hamilton(no%degf(f),k)
            kft = htmtype(h)
            select case (kft)
            case (0)
                kt = 0
                exit
            case (1)
                continue
            case (2,3,4)
                kt = 2
            case default
                kt = -1
                exit
            end select
        end do
    end function get_mode_htmtype

    subroutine set_wfnode_uncorrelated(no)
        ! Set which mode in a node (if any) is separable from the rest
        ! This depends also on the nodes above, so this routine must
        ! be called in a swap_down loop.
        implicit none
        type(wfnode_t),intent(inout)  :: no
        integer                       :: k,f,nh,fun,n,kcorr

        allocate(no%isuncor(ktot))
        allocate(no%hasuncor(no%nmodes,hamdim))

        no%isuncor = .false.
        no%hasuncor = .false.
      do n=1,hamdim
        kcorr = zham(n)+kzahl(n)-1
        do k=zham(n),zham(n)+khtot(n)-1
            nh=0

! if k is uncorrelated, fun is uncorrelated node / mode
            fun=0
            do f=1,no%nmodes
                if (.not.no%isunit(k,f)) then
                   nh = nh + 1
                   fun = f
                endif
            end do

            if(nh<=1) then
                if(no%istop) then
                    no%isuncor(k) = .true.
                    if (k .gt. kcorr) no%hasuncor(fun,n)=.true.
                else
                    if(no%parent%isuncor(k)) then
                        no%isuncor(k) = .true.
                        if (k .gt. kcorr .and. fun .gt. 0) then
                           no%hasuncor(fun,n)=.true.
                        endif
                    else
                        no%isuncor(k) = .false.
                    end if
                end if
            else
                no%isuncor(k) = .false.
            end if
            if(debug) then
                write(6,*) 'k,isuncor',k,no%isuncor(k)
            end if
        end do
      enddo
        if(debug) write(6,*) ' '
    end subroutine set_wfnode_uncorrelated


    subroutine allocate_wfnode(no,ldmat,dicht1,dicht2,dicht3,dicht4)
        ! allocate memory for some of the data structures of a node
        !
        ! input:
        !  no -- node being allocated
        !  nk  -- number of operator terms
        !  ldmat -- true(default): dmat are allocated
        !
        ! notes: storage has to be allocated in swap_up direction (in calling
        ! routine). The storage allocated depends on which operators are
        ! unit-operators in the nodes below.
        implicit none
        type(wfnode_t),intent(inout)  :: no
        type(wfnode_t),pointer        :: childf
        logical, intent(in), optional :: ldmat
        complex(dop), dimension(*)    :: dicht1,dicht2,dicht4
        real(dop), dimension(*)       :: dicht3
        integer                       :: f,alen,m
        logical                       :: lf

        lf = .true.
        if(present(ldmat)) lf = ldmat

        ! for primitive nodes allocate 1d density arrays
        if(no%isprimitive) then
            ! space for grid populations
            allocate(no%pop(no%nmodes))
            do f=1,no%nmodes
                alen = no%ndim(f)
                no%pop(f)%p => new_hterm(5,alen)  ! real vector
            end do
            return
        end if

        if (.not. lf) return

        ! allocate matrices of pointers
        allocate(no%dmat(no%nmodes))
        allocate(no%invdmat(no%nmodes))
        allocate(no%pop(no%nmodes))
        allocate(no%regpop(no%nmodes))
        allocate(no%trfmat(no%nmodes))

        ! allocate matrices of numbers for non-primitive nodes
        do f=1,no%nmodes
           childf => no%submodes(f)%p
           alen = no%ndim(f)

           if (childf%isprimitive) then
              m=childf%m
              if (alen .ne. dim(m,1)) then
                 routine="Allocate_wfnode"
                 write(message,'(a,i5)') &
        "No. of SPFs defined in ML-MCTDH lowest layer inconsistent. Mode :",m
                 call errormsg
              endif

               no%dmat(f)%p    => &
                         set_hterm(2,dicht2(dmat(m,1)),alen,alen)
               no%invdmat(f)%p => &
                         set_hterm(2,dicht1(dmat(m,1)),alen,alen)
               no%pop(f)%p     => &
                         set_htermdv(5,dicht3(d3mat(m,1)),alen)
               no%regpop(f)%p  => &
                         set_htermdv(5,dicht3(d3mat(m,1)+alen),alen)
               no%trfmat(f)%p  => &
                         set_hterm(2,dicht4(dmat(m,1)),alen,alen)
           else
               no%dmat(f)%p    => new_hterm(2,alen)
               no%invdmat(f)%p => new_hterm(2,alen)
               no%pop(f)%p     => new_hterm(5,alen)
               no%regpop(f)%p  => new_hterm(5,alen)
               no%trfmat(f)%p  => new_hterm(2,alen)
           endif
        enddo

    end subroutine allocate_wfnode


    subroutine dispose_wfnode(n)
        implicit none
        ! deallocate all storage of one node, so that it
        ! can be safely deallocated/disassociated
        type(wfnode_t), intent(inout) :: n
        type(wfnode_t), pointer       :: childf
        integer                       :: j

        if(.not.associated(n%dmat)) return  !! HDM Feb 2022

        if(n%isprimitive) then
            do j=1,n%nmodes
                call deallocate_hterm(n%pop(j)%p)
            end do
            deallocate(n%pop)
            return
        end if

        ! deallocate storage
        do j=1,n%nmodes
            childf => n%submodes(j)%p
            if (childf%isprimitive) cycle

            call deallocate_hterm(n%dmat(j)%p)
            call deallocate_hterm(n%pop(j)%p)
            call deallocate_hterm(n%regpop(j)%p)
            call deallocate_hterm(n%invdmat(j)%p)
            call deallocate_hterm(n%trfmat(j)%p)
        end do

        ! deallocate matrices of pointers
        deallocate(n%dmat)
        deallocate(n%invdmat)
        deallocate(n%pop)
        deallocate(n%regpop)
        deallocate(n%trfmat)

    end subroutine dispose_wfnode


    subroutine allocate_overlap_matrices(t1,t2,ovlmat)
        implicit none
        ! Given two trees (which could be both the same) the overlap
        ! matrices of t2 are initialized so that they can store matrix
        ! elements of the SPF's of a wavefunction conforming to t1 and
        ! a wavefunction conforming to t2.
        ! The tree structure between t1 and t1 must be identical, the
        ! number of SPF's may vary from node to node (and that's why
        ! the overlap matrices are allowed to be rectangular)
        type(wfnode_t),intent(inout) :: t1
        type(wfnode_t),intent(inout) :: t2
        complex(cdbl),intent(in)     :: ovlmat(dmatdim)
        integer                      :: i,f,m
        type(wfnode_t),pointer       :: cnode1,cnode2,childf

        do i=1,t1%tot_nodes
            cnode1 => t1%swap_up(i)%p
            cnode2 => t2%swap_up(i)%p

            if(cnode1%isprimitive) cycle

            allocate(cnode2%overlap(cnode1%nmodes))

            do f=1,cnode1%nmodes
                childf => cnode1%submodes(f)%p
                if (childf%isprimitive) then
                   m=childf%m
                   cnode2%overlap(f)%p => &
                     set_hterm(2,ovlmat(dmat(m,1)),cnode1%ndim(f),cnode2%ndim(f))
                else
                   cnode2%overlap(f)%p => &
                     new_hterm(2,cnode1%ndim(f),cnode2%ndim(f))
                endif
            end do
        end do
    end subroutine allocate_overlap_matrices


    subroutine deallocate_overlap_matrices(t)
        implicit none
        type(wfnode_t),intent(inout) :: t
        type(wfnode_t),pointer       :: cnode,childf
        integer                      :: i,j
        do i=1,t%tot_nodes
            cnode => t%swap_up(i)%p
            if(cnode%isprimitive) cycle
            do j=1,cnode%nmodes
                childf => cnode%submodes(j)%p
                if (.not. childf%isprimitive) then
                   call deallocate_hterm(cnode%overlap(j)%p)
                endif
                deallocate(cnode%overlap(j)%p)
            end do
            deallocate(cnode%overlap)
        end do
    end subroutine deallocate_overlap_matrices

    subroutine allocate_mtraces_matrices(t1,ptraces)
        implicit none
        ! Given a tree the mode traces and single-hole traces for
        ! a ML density matrix are initialized.
        ! ptraces are the vectors for the traces of the lowest layer SPDOs
        type(wfnode_t),intent(inout) :: t1
        real(dbl),intent(in)         :: ptraces(d3matdim)
        integer                      :: i,f,m
        type(wfnode_t),pointer       :: cnode1,childf

        do i=1,t1%tot_nodes
            cnode1 => t1%swap_up(i)%p

            if(cnode1%isprimitive) cycle

            allocate(cnode1%mtraces(cnode1%nmodes))

            do f=1,cnode1%nmodes
                childf => cnode1%submodes(f)%p
                if (childf%isprimitive) then
                   m=childf%m
                   cnode1%mtraces(f)%p => &
                     set_htermdv(5,ptraces(dd3mat(m,1,1)),cnode1%ndim(f))
                else
                   cnode1%mtraces(f)%p => &
                     new_hterm(5,cnode1%ndim(f))
                endif
            end do
        end do
    end subroutine allocate_mtraces_matrices

    subroutine allocate_mtraces_matrices2(t1,ptraces)
        implicit none
        ! Given a tree the mode traces and single-hole traces for
        ! a ML density matrix are initialized.
        ! ptraces are the vectors for the traces of the lowest layer SPDOs
        ! As allocate_mtraces_matrices but uses d3mat in place of dd3mat for 
        ! allocating lowest layer (type II properties).
        type(wfnode_t),intent(inout) :: t1
        real(dbl),intent(in)         :: ptraces(d3matdim)
        integer                      :: i,f,m
        type(wfnode_t),pointer       :: cnode1,childf

        do i=1,t1%tot_nodes
            cnode1 => t1%swap_up(i)%p

            if(cnode1%isprimitive) cycle

            allocate(cnode1%mtraces(cnode1%nmodes))

            do f=1,cnode1%nmodes
                childf => cnode1%submodes(f)%p
                if (childf%isprimitive) then
                   m=childf%m
                   cnode1%mtraces(f)%p => &
                     set_htermdv(5,ptraces(d3mat(m,1)),cnode1%ndim(f))
                else
                   cnode1%mtraces(f)%p => &
                     new_hterm(5,cnode1%ndim(f))
                endif
            end do
        end do
    end subroutine allocate_mtraces_matrices2

    subroutine allocate_zmtraces_matrices(t1,ptraces)
        implicit none
        ! Given a tree the mode traces and single-hole traces for
        ! a ML density matrix are initialized.
        ! ptraces are the vectors for the traces of the lowest layer SPDOs
        type(wfnode_t),intent(inout) :: t1
        complex(dbl),intent(in)         :: ptraces(d3matdim)
        integer                      :: i,f,m
        type(wfnode_t),pointer       :: cnode1,childf

        do i=1,t1%tot_nodes
            cnode1 => t1%swap_up(i)%p

            if(cnode1%isprimitive) cycle

            allocate(cnode1%mtraces(cnode1%nmodes))
            f=cnode1%nmodes

            do f=1,cnode1%nmodes
                childf => cnode1%submodes(f)%p
                if (childf%isprimitive) then
                   m=childf%m
                   cnode1%mtraces(f)%p => &
                     set_htermv(3,ptraces(dd3mat(m,1,1)),cnode1%ndim(f))
                else
                   cnode1%mtraces(f)%p => &
                     new_hterm(3,cnode1%ndim(f))
                endif
            end do
        end do
    end subroutine allocate_zmtraces_matrices

    subroutine deallocate_mtraces_matrices(t)
        implicit none
        type(wfnode_t),intent(inout) :: t
        type(wfnode_t),pointer       :: cnode,childf
        integer                      :: i,j
        do i=1,t%tot_nodes
            cnode => t%swap_up(i)%p
            if(cnode%isprimitive) cycle
            do j=1,cnode%nmodes
                childf => cnode%submodes(j)%p
                if (.not. childf%isprimitive) then
                   call deallocate_hterm(cnode%mtraces(j)%p)
                endif
                deallocate(cnode%mtraces(j)%p)
            end do
        end do
    end subroutine deallocate_mtraces_matrices

    subroutine allocate_shtraces_matrices(t1,shtraces)
        implicit none
        ! Given a tree the single-hole traces for
        ! a ML density matrix are initialized.
        ! ptraces are the vectors for the traces of the lowest layer SPDOs
        type(wfnode_t),intent(inout) :: t1
        real(dbl),intent(in)         :: shtraces(d3matdim)
        integer                      :: i,f,m
        type(wfnode_t),pointer       :: cnode1,childf

        do i=1,t1%tot_nodes
            cnode1 => t1%swap_up(i)%p

            if(cnode1%isprimitive) cycle

            allocate(cnode1%shtraces(cnode1%nmodes))

            do f=1,cnode1%nmodes
                childf => cnode1%submodes(f)%p
                if (childf%isprimitive) then
                   m=childf%m
                   cnode1%shtraces(f)%p => &
                     set_htermdv(5,shtraces(dd3mat(m,1,1)),cnode1%ndim(f))
                else
                   cnode1%shtraces(f)%p => &
                     new_hterm(5,cnode1%ndim(f))
                endif
            end do
        end do
    end subroutine allocate_shtraces_matrices

    subroutine deallocate_shtraces_matrices(t)
        implicit none
        type(wfnode_t),intent(inout) :: t
        type(wfnode_t),pointer       :: cnode,childf
        integer                      :: i,j
        do i=1,t%tot_nodes
            cnode => t%swap_up(i)%p
            if(cnode%isprimitive) cycle
            do j=1,cnode%nmodes
                childf => cnode%submodes(j)%p
                if (.not. childf%isprimitive) then
                   call deallocate_hterm(cnode%shtraces(j)%p)
                endif
                deallocate(cnode%shtraces(j)%p)
            end do
        end do
    end subroutine deallocate_shtraces_matrices


    subroutine allocate_hterm_matrices(this,useunit)
        ! Allocate hterm matrices in all nodes
        !
        ! this - top node of the tree
        ! nkterms - number of product lines to be allocated
        ! useunit - bool, wether to use information in
        !           node%isunit to decide the type of hterm to
        !           allocate. It is useful if allocating space
        !           for the Hamiltonian and node%isunit has been
        !           computed. For small operators for analysis
        !           purposes this will usually not be the case.
        implicit none
        type(wfnode_t),intent(inout) :: this
        logical,intent(in)           :: useunit
        type(wfnode_t),pointer       :: cnode
        type(wfnode_t),pointer       :: childf
        integer                      :: i,f,k,alen,m,n
        do i=1,this%tot_nodes
            cnode => this%swap_up(i)%p
            if(cnode%isprimitive) cycle
            allocate( cnode%hterms(ktot,cnode%nmodes) )

            do f=1,cnode%nmodes
               childf => cnode%submodes(f)%p
               alen = cnode%ndim(f)
               if (childf%isprimitive) then
                  m=childf%m
                  if (alen .ne. dim(m,1)) then
                     routine="Allocate_hterm_matrices"
                     write(message,'(a,i5)') &
        "No. of SPFs defined in ML-MCTDH lowest layer inconsistent. Mode :",m
                     call errormsg
                  endif
               endif

               do n=1,hamdim
                 if(useunit) then
                    do k=zham(n),zham(n)+kzahl(n)-1
                       if(cnode%isunit(k,f)) then
                          cnode%hterms(k,f)%p => new_hterm(1,alen)
                       else if (childf%isprimitive) then
                          if (pmat(m,k) .gt. 0) &
                             cnode%hterms(k,f)%p => &
                               set_hterm(2,hteil(pmat(m,k)),alen,alen)
                       else
                           cnode%hterms(k,f)%p => new_hterm(2,alen)
                       endif
                    enddo
                 else
                    if (childf%isprimitive) then
                       do k=zham(n),zham(n)+kzahl(n)-1
                          if (pmat(m,k) .gt. 0) &
                             cnode%hterms(k,f)%p => &
                               set_hterm(2,hteil(pmat(m,k)),alen,alen)
                       end do
                    else
                       do k=zham(n),zham(n)+kzahl(n)-1
                          cnode%hterms(k,f)%p => new_hterm(2,alen)
                       end do
                    endif
                 end if

! assign space for any explicitely uncorrelated terms
                 if (cnode%hasuncor(f,n)) then
! DOF with uncorrelated operator
                    k=zham(n)+kzahl(n)
                    if (childf%isprimitive) then
                       if (pmat(m,kzahl(n)+1) .gt. 0) &
                          cnode%hterms(k,f)%p => &
                             set_hterm(2,hteil(pmat(m,kzahl(n)+1)),alen,alen)
                    else
                       cnode%hterms(k,f)%p => new_hterm(2,alen)
                    endif
                 else if (khtot(n) .gt. kzahl(n)) then
! DOF with unit operator
                    k=zham(n)+kzahl(n)
                    cnode%hterms(k,f)%p => new_hterm(1,alen)
                 endif
              enddo
            enddo

        end do

    end subroutine allocate_hterm_matrices


    subroutine deallocate_hterm_matrices(this)
        implicit none
        type(wfnode_t),intent(inout) :: this
        type(wfnode_t),pointer       :: cnode,childf
        integer                      :: i,f,k
        do i=1,this%tot_nodes
            cnode => this%swap_up(i)%p
            if(cnode%isprimitive) cycle
            if(.not.associated(cnode%hterms)) cycle  !! HDM Feb 2022
            do f=1,cnode%nmodes
                childf => cnode%submodes(f)%p
                if (childf%isprimitive) cycle
                do k=1,khtot(1)
                    call deallocate_hterm(cnode%hterms(k,f)%p)
                end do
            end do
            deallocate(cnode%hterms)
        end do
    end subroutine deallocate_hterm_matrices

    subroutine allocate_mfields(t)
       ! Allocate mean-field matrices for all nodes.
       ! t: top node of tree
       type(wfnode_t),intent(inout) :: t
       type(wfnode_t),pointer       :: no,childf
       integer                      :: i,f,k,alen
       do i=1,t%tot_nodes
          no => t%swap_up(i)%p
          if (no%isprimitive) cycle

          ! allocate matrices of pointers
          allocate(no%mfields(1:kzahl(1), no%nmodes))

          ! allocate matrices of numbers
          do f=1,no%nmodes
             alen = no%ndim(f)
             childf => no%submodes(f)%p
             if (childf%isprimitive) then
                m=childf%m
                if (alen .ne. dim(m,1)) then
                   routine="Allocate_mfields"
                   write(message,'(a,i5)') &
        "No. of SPFs defined in ML-MCTDH lowest layer inconsistent. Mode :",m
                   call errormsg
                endif
                do k=1,kzahl(1)
                   no%mfields(k,f)%p => &
                          set_hterm(2,mc(mchloch+phloch(m,k)-1),alen,alen)
                end do
             else
                do k=1,kzahl(1)
                   if(no%isuncor(k)) then  ! separable term, no MF needed
                      no%mfields(k,f)%p => new_hterm(1,alen)
                   else
                      no%mfields(k,f)%p => new_hterm(2,alen)
                   endif
                enddo
             endif
          enddo
       enddo
    end subroutine allocate_mfields


    subroutine deallocate_mfields(t)
       ! Deallocate storage for mean-field matrices.
       ! t: top node of tree
       type(wfnode_t),intent(inout) :: t
       type(wfnode_t),pointer       :: no,childf
       integer                      :: i,f,k
       do i=1,t%tot_nodes
          no => t%swap_up(i)%p
          if (no%isprimitive) cycle
          if (associated(no%mfields)) then
             do f=1,no%nmodes
                childf => no%submodes(f)%p
                if (childf%isprimitive) cycle

                do k=1,kzahl(1)
                   call deallocate_hterm(no%mfields(k,f)%p)
                enddo
             enddo
             deallocate(no%mfields)
          endif
       enddo
    end subroutine deallocate_mfields

!-----------------------------------------------------------------------
! Routine to set up initial vectors etc.
!-----------------------------------------------------------------------
    subroutine set_hartree_avec(psi,t)
        implicit none
        complex(cdbl),intent(inout)      :: psi(:)
        type(wfnode_t),intent(in),target :: t
        type(wfnode_t),pointer           :: cnode
        integer                          :: i,j,gdim,ierr,sqdim
        integer, dimension(35)           :: seed
        real(dbl)                        :: rand
        logical                          :: rlx


        rlx = lrlx .and. .not. lrlxt .and. .not. lrlxun  ! Relaxation
        if (rlx) then  ! Relaxation
           do j=1,35
              seed(j)=j
           enddo
           call random_seed(put=seed)
           call random_number(rand)
  !        PRINT *, ' rand =', rand
        endif

        do i=1,t%tot_nodes

            cnode => t%swap_up(i)%p
            if(cnode%isprimitive) cycle
            gdim = cnode%z(1,2) - cnode%z(1,1) + 1
            sqdim=int(sqrt(gdim+0.1d0))       ! Dieter 27.12.2020
            ierr = 0

!-----------------------------------------------------------------------
! Initial vectors set to be |1> = phi1 (to give Hartree product for 
! full WP) and ! higher vectors include all functions from layer below
!
! special case for 3 vecs + 3 spfs  - initial guess |2> = (0,1,-1), 
! |3> = (0,-1,1) leads to |3> = 0 after orthonormalisation
!-----------------------------------------------------------------------
            psi(cnode%z(1,1):cnode%z(cnode%navecs,2)) = CZERO
            if (rlx) then  ! Relaxation
               do j=1,gdim-1                ! Dieter 27.12.2020
                  call random_number(rand)
                  psi(cnode%z(1,1)+j) = 5.d-4*(rand-0.5) ! CONE*1.0d-4
               enddo
               do j=1,sqdim-1    ! This is strictly correct for two siblings only!
                  psi(cnode%z(1,1)+j*(sqdim+1)) = CONE*1.0d-3
               enddo                        ! Dieter 27.12.2020
            endif
            psi(cnode%z(1,1)) = CONE
            if(cnode%navecs == 3 .and. gdim == 3) then
               psi(cnode%z(2,1)+1) = CONE
               psi(cnode%z(3,2)) = CONE
            else if(cnode%navecs>1) then
                do j=2,cnode%navecs
                    psi(cnode%z(j,1)+1:cnode%z(j,2)) = CONE
                    psi(cnode%z(j,1)+j-1) = -CONE
                end do
            end if
            call gramschmidt(psi(cnode%z(1,1):cnode%z(cnode%navecs,2)),&
                             gdim,cnode%navecs,ierr)
        end do
    end subroutine set_hartree_avec

    subroutine set_prefactor(self,prf)
        implicit none
        type(wfnode_t),intent(inout) :: self
        complex(cdbl),intent(in)     :: prf
        self%prefactor = prf
    end subroutine set_prefactor


    subroutine set_isrelax(self,isrlx)
        implicit none
        type(wfnode_t),intent(inout) :: self
        logical,intent(in)           :: isrlx
        self%isrelax = isrlx
    end subroutine set_isrelax

    subroutine set_isrlxt(self,isrlxt)
        implicit none
        type(wfnode_t),intent(inout) :: self
        logical,intent(in)           :: isrlxt
        self%isrlxt = isrlxt
    end subroutine set_isrlxt

!-----------------------------------------------------------------------
! Output routines
!-----------------------------------------------------------------------

    subroutine outpoptree(t,iout)
        implicit none
        type(wfnode_t),intent(in),target :: t
        type(wfnode_t),pointer           :: cnode
        integer                          :: ndim,i,f,l,iout,f1,m1
        integer                          :: maxlayer=0
        do i=1,t%tot_nodes
            cnode => t%swap_down(i)%p
            if(cnode%layer>maxlayer) maxlayer=cnode%layer
        end do
        write(iout,*)
        write(iout,'(a)') ' Natural weights *1000 :'
        write(iout,*)
        do l=0,maxlayer
            do i=1,t%tot_nodes
                cnode => t%swap_down(i)%p
                if(cnode%isprimitive) cycle
                if(.not.cnode%layer==l) cycle
                write(iout,'(1x,a,i4,3x,a,i3)') &
                      'node:',cnode%num, ' layer:',cnode%layer
                do f=1,cnode%nmodes
                    ndim = cnode%ndim(f)
                    if (cnode%submodes(f)%p%isprimitive) then
                       m1 = cnode%submodes(f)%p%m
                       f1 = spfdof(1,m1)
                       if (nspfdof(m1) .eq. 1) then
                          write(iout,'(a12,a)',advance="no") modelabel(f1),':'
                       else
                          write(iout,'(a8,a,i3,a)',advance="no") &
                                modelabel(f1),'C',nspfdof(m1),':'
                       endif
                    else
                       m1 = cnode%submodes(f)%p%num
                       write(iout,'(1x,a,i2,a)',advance="no") ' node',m1,'    :'
                    endif
                    select case(ndim)
                        case(:7)
                          write(iout,'(7(1x,f9.3))') &
                             cnode%pop(f)%p%rvector*1000.d0
                        case default
                          write(iout,'(7(f9.3))') &
                               cnode%pop(f)%p%rvector(1:7)*1000.d0
                          write(iout,'(13x,7(f9.3))') &
                               cnode%pop(f)%p%rvector(8:)*1000.d0
                    end select
                end do
                write(iout,*)
            end do
        end do

    end subroutine outpoptree


    subroutine outlowpop(t,iout,time,lhead)
        implicit none
        type(wfnode_t),intent(in),target :: t ! tree
        integer,intent(in)               :: iout
        real(dbl),intent(in)             :: time
        logical,intent(in)               :: lhead

        type(wfnode_t),pointer           :: cnode
        real(dbl)                        :: pop
        integer                          :: i,f,ndim,l
        character(len=10)                :: text
        integer,parameter                :: fileversion=1

        ! print header if requested
        if (lhead) then
            write(iout, '(a,i0,a)') '# lownatpop:version=', fileversion, &
                ' -- lowest natural populations for all modes over time'
            do l=1,3
                if (l==1) then
                    write(iout, '(a)', advance="no") '# time[fs]  '
                else
                    write(iout, '(a)', advance="no") '#           '
                end if
                do i=1,t%tot_nodes
                    cnode => t%swap_down(i)%p
                    if (cnode%isprimitive) cycle
                    do f=1,cnode%nmodes
                        if (l==1) then
                            write(text, '(a,i0)') 'layer:',cnode%layer
                        elseif (l==2) then
                            write(text, '(a,i0)') 'node:',cnode%num
                        else
                            write(text, '(a,i0)') 'mode:',f
                        end if
                        write(iout, '(1x,a10)', advance="no") text
                    end do
                end do
                write(iout,*)
            end do
        end if

        ! the following will be printed on all calls
        write(iout, '(g11.6,1x)', advance="no") time
        do i=1,t%tot_nodes
            cnode => t%swap_down(i)%p
            if (cnode%isprimitive) cycle
            do f=1,cnode%nmodes
                ndim = cnode%ndim(f)
                pop  = cnode%pop(f)%p%rvector(ndim)
                if (pop<1.d-99) pop=0.d0
                write(iout, '(e10.5,1x)', advance="no") pop
            end do
        end do
        write(iout,*)
        call myflush(iout)
    end subroutine outlowpop

    subroutine mlspwn_wftree(ich,ispwn,time,lhead,lnspf,lstop,spawnopt,&
               imaxspf,spawnthr,maxtape,itape3,itape2,tape,tapeM)
        implicit none
        integer,intent(in)   :: ich,imaxspf
        real(dbl),intent(in) :: time
        logical,intent(in)   :: lhead

        logical              :: lstop,lnspf
        integer              :: spawnopt
        real(dbl)            :: spawnthr
!       Read maxtree
        integer              :: maxtape
        integer              :: itape3(maxtape),itape2(maxtape),&
                                tapeM(maxtape),tape(maxtape)
!       Write new tree in spawn-file
        type(wfnode_t),pointer :: cnode,childf
        integer                :: f,i,j,ncount,ndim,m
        integer                :: ispwn
        real(dbl)              :: pop
        character(len=16)      :: string

!       Monitor natural orbital population
! tape: present tree
! tapeM: tree with max. no. SPFs
! itape3: no. spfs needed according to nat. pops.
! itape2: position in tree of SPFs recorded in itape3
        ncount=0
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            if (.not. cnode%isprimitive) then
               do f=1,cnode%nmodes
                  ncount=ncount+1

!                 spawning adding spfs criteria
                  ndim = cnode%ndim(f)
                  pop  = cnode%pop(f)%p%rvector(ndim)
                  if (pop*1000 > spawnthr) then
                     if (spawnopt.eq.0) then
                        itape3(ncount)=ndim+imaxspf
                        itape3(ncount)=min(itape3(ncount),tapeM(itape2(ncount)))
                     endif
                  else
                     itape3(ncount)=ndim
                  end if

! set dim(m,s) for primitive modes
                  childf => cnode%submodes(f)%p
                  if (childf%isprimitive) then
                     m=childf%m
                     dimtape(m,1) = itape3(ncount)
                  endif

                  if (itape3(ncount) .gt. tape(itape2(ncount))) lstop = .true.
               enddo

            end if
        end do

        call spwn_wftree(tree,ich,time,lhead,lnspf)

!       Write new tree in spawn-file
        if (ldebugspawn) then
           ncount=1
           rewind(ispwn)
           write(ispwn,'(a,f13.6,a)') 'time: ',time,' fs'
           write(ispwn,'(a)') 'ML-Basis-Section'
           do i=1,tree%tot_nodes
               cnode => tree%swap_down(i)%p
!              write nodes
               do j=1,cnode%layer*2
                  write(ispwn,'(x)',advance='no')
               end do
               write(string,'(i4)') cnode%layer
               write(ispwn,'(a)',advance='no') trim(adjustl(string))
               write(ispwn,'(a)',advance='no') '>'
!              write SPFs or modes
               if (cnode%isprimitive) then
                  write(ispwn,'(a)',advance='no') ' ['
                  do f=1,cnode%nmodes
                     if (f.ne.1) write(ispwn,'(x)',advance='no')
                     write(ispwn,'(a)',advance='no') trim(modelabel(f))
                  end do
                  write(ispwn,'(a)',advance='no') ']'
               else
                  do f=1,cnode%nmodes
!                    spawning adding spfs criteria
                     ndim = cnode%ndim(f)
                     write(string,'(i4)') ndim
                     write(ispwn,'(x,a)',advance='no') trim(adjustl(string))
                     ncount=ncount+1
                  end do
               end if
               write(ispwn,*)
           end do
           write(ispwn,'(a)') 'end-ML-Basis-Section'
        endif

    end subroutine mlspwn_wftree

    subroutine spwn_wftree(t,iout,time,lhead,lnspf)
        implicit none
        type(wfnode_t),intent(in),target :: t ! tree
        integer,intent(in)               :: iout
        real(dbl),intent(in)             :: time
        logical,intent(in)               :: lhead

        type(wfnode_t),pointer           :: cnode
        integer                          :: i,f,ndim,l
        character(len=10)                :: text
        integer,parameter                :: fileversion=1
!       Spawning stopping criteria
        logical                          :: lnspf

!       Print header if requested
        if (lhead) then
            write(iout, '(a,i0,a)') '# nspf:version=', fileversion, &
                ' -- number of single particle functions for all modes over time'
            do l=1,3
                if (l==1) then
                    write(iout, '(a)', advance="no") '# time[fs]  '
                else
                    write(iout, '(a)', advance="no") '#           '
                end if
                do i=1,t%tot_nodes
                    cnode => t%swap_down(i)%p
                    if (cnode%isprimitive) cycle
                    do f=1,cnode%nmodes
                        if (l==1) then
                            write(text, '(a,i0)') 'layer:',cnode%layer
                        elseif (l==2) then
                            write(text, '(a,i0)') 'node:',cnode%num
                        else
                            write(text, '(a,i0)') 'mode:',f
                        end if
                        write(iout, '(1x,a10)', advance="no") text
                    end do
                end do
                write(iout,*)
            end do
        end if
!       Number of spf will be printed after every spawning
        if (lnspf) then
           write(iout, '(g11.6,1x)', advance="no") time
           do i=1,t%tot_nodes
               cnode => t%swap_down(i)%p
               if (cnode%isprimitive) cycle
               do f=1,cnode%nmodes
                  ndim = cnode%ndim(f)
                  write(iout, '(i4,1x)', advance="no") ndim
              end do
          end do
          write(iout,*)
          lnspf=.false.
        end if

        call myflush(iout)
    end subroutine spwn_wftree

! -----------------------------------------------------------------------------
! TREE & NODE COPY PROCEDURES
! -----------------------------------------------------------------------------
    subroutine copy_wftree(oldpsi,oldtree,newpsi,newtree)
        implicit none
        ! Copy oldpsi to newpsi. Both trees must have the same structure, but
        ! the number of SPF's may vary.
        ! In principle this has been already checked, so no further consistency
        ! checks here.
        ! Only the A vectors of non-primitive nodes are handled here. This is
        ! because the SPF's are handled in the getpsi subroutine by the normal
        ! MCTDH.
        complex(dop)                  :: oldpsi(*)
        type(wfnode_t),pointer         :: oldtree
        complex(dop),intent(inout)    :: newpsi(:)
        type(wfnode_t),pointer         :: newtree
        type(wfnode_t),pointer         :: cold,cnew
        integer                        :: i
        do i=1,newtree%tot_nodes
            cold => oldtree%swap_down(i)%p
            cnew => newtree%swap_down(i)%p
            if(cnew%isprimitive) cycle
            call copy_wfnode(oldpsi,cold,newpsi,cnew)
        end do
    end subroutine copy_wftree

    subroutine copy_wfnode(oldpsi,oldnode,newpsi,newnode)
        implicit none
        ! Copy Avecs from oldpsi to newpsi for nodes oldnode to newnode
        complex(dop)                    :: oldpsi(*)
        type(wfnode_t),intent(in)        :: oldnode
        complex(dop),intent(inout)      :: newpsi(:)
        type(wfnode_t),intent(in)        :: newnode
        integer                          :: i,j,navecs,tlen
        integer                          :: tdim(newnode%nmodes)
        complex(dop),allocatable,target :: tmppsi1(:),tmppsi2(:)
        complex(dop),pointer            :: tmppsia(:),tmppsib(:),tmppsic(:)
        integer                          :: oshape(3),nshape(3)
        integer                          :: zo0,zof,zn0,znf,ierr

        do i=1,newnode%nmodes
            tdim(i)=max(oldnode%ndim(i),newnode%ndim(i))
        end do
        tlen = product(tdim)
        allocate(tmppsi1(tlen))
        allocate(tmppsi2(tlen))

        ! transfer avec coeffs from old to new
        navecs=min(oldnode%navecs,newnode%navecs) ! navecs can be safely copied
        do i=1,navecs
            zo0 = oldnode%z(i,1)
            zof = oldnode%z(i,2)
            zn0 = newnode%z(i,1)
            znf = newnode%z(i,2)
            tmppsi1=CZERO
            tmppsi2=CZERO
            tmppsia => tmppsi1
            tmppsib => tmppsi2
            tmppsia(1:zof-zo0+1) = oldpsi(zo0:zof)
            do j=1,newnode%nmodes
                oshape=ijk_shape(j,oldnode%ndim)
                nshape=ijk_shape(j,newnode%ndim)
                call akcopy(tmppsia,tmppsib,nshape(1),nshape(2),&
                            oshape(2),oshape(3))
                tmppsic => tmppsia
                tmppsia => tmppsib
                tmppsib => tmppsic
            end do
            newpsi(zn0:znf) = tmppsia(1:znf-zn0+1)
        end do

        ! generate new avecs if necessary
        if(newnode%navecs>oldnode%navecs) then
            do i=oldnode%navecs+1,newnode%navecs
                zn0 = newnode%z(i,1)
                znf = newnode%z(i,2)
                newpsi(zn0:znf) =  CONE
                newpsi(zn0+i-1) = -CONE
            end do
        end if

        ! Schmidtt orthogonalize
        tlen = product(newnode%ndim)
        call gramschmidt(newpsi(newnode%z(1,1):newnode%z(newnode%navecs,2)),&
                                      tlen,newnode%navecs,ierr)

        tmppsia => null()
        tmppsib => null()
        tmppsic => null()
        deallocate(tmppsi1)
        deallocate(tmppsi2)
    end subroutine copy_wfnode

!===============================================================================

end module wftree1
