!=======================================================================
module genmlop
!=======================================================================

use mltypes
use optree
use wftree
use ilist
use tensors
implicit none
private

public :: make_optree_from_mlwf, mlop_outer_product, &
          make_optree_from_op1d


contains !==============================================================

!-----------------------------------------------------------------------
function make_optree_from_op1d(m,k) result(optree)
!-----------------------------------------------------------------------
   type(optree_t),pointer :: optree ! created here
   integer,intent(in)     :: m      ! the MCTDH mode that the 1D operator applies to
   integer,intent(in)     :: k      ! the k-term which stores the 1D operator
   type(opnode_t),pointer :: opnode
   allocate(opnode)
   opnode%isprimitive = .true.
   opnode%nops = 1
   opnode%m = m
   allocate(opnode%ks(1))
   opnode%ks(1) = k
   optree => make_optree(opnode)
end function make_optree_from_op1d


!-----------------------------------------------------------------------
function make_optree_from_mlwf(wftree) result(optree)
!-----------------------------------------------------------------------
! This (and the following function) creates an MLOp with the same tree
! structure as the ML-wavefunction. The actual data (V-tensors and
! k-terms) is filled in later.
!-----------------------------------------------------------------------
   type(optree_t),pointer    :: optree ! created here
   type(wfnode_t),intent(in) :: wftree ! top node of ML-WF tree
   type(opnode_t),pointer    :: optopnode
   optopnode => make_opnode_from_mlwf(wftree)
   optree => make_optree(optopnode)
end function make_optree_from_mlwf


!-----------------------------------------------------------------------
recursive function make_opnode_from_mlwf(wfnode) result(opnode)
!-----------------------------------------------------------------------
   type(opnode_t),pointer      :: opnode
   type(wfnode_t),intent(in)   :: wfnode
   integer                     :: nmodes,m
   type(opnode_tp),allocatable :: opchildren(:)
   type(wfnode_t),pointer      :: wfchild
   if (wfnode%isprimitive) then
      opnode => make_opleaf(wfnode%m)
   else
      nmodes = wfnode%nmodes
      allocate(opchildren(nmodes))
      do m=1,nmodes
         wfchild => wfnode%submodes(m)%p
         opchildren(m)%p => make_opnode_from_mlwf(wfchild)
      enddo
      opnode => make_opnode(opchildren)
      deallocate(opchildren)
   endif
end function make_opnode_from_mlwf


!-----------------------------------------------------------------------
recursive subroutine mlop_outer_product(tgt,ops,ierr,errmsg)
!-----------------------------------------------------------------------
! Consider an outer product of a set of MLOps. Each MLOp has a tree
! structure, so what should be the tree structure of the resulting
! MLOp? Here we consider that the tree structure of the resulting MLOp
! is already given (_tgt_). In general, the tree structures of the
! factors (_ops_) in the outer product will not be compatible with
! the target tree. But it might be possible to use the hierarchical
! expansion of an MLOp to alleviate this problem. Though this usually
! requires transposing and/or extending the transfer tensor. This is
! what this routine does.
! The recursive solution of the mentioned problem requires that we
! consider a more general problem, namely making a list of outer
! products of lists of MLOps.
!-----------------------------------------------------------------------
   type(opnode_t),intent(inout)   :: tgt
   type(opnode_tp),intent(in)     :: ops(:)
   integer,intent(out)            :: ierr
   character(len=240),intent(out) :: errmsg
   type(opnode_t),pointer         :: ch,op,subop
   integer                        :: r,nr,m,nm,m1,l,nsubm,map(2)
   type(ilist_t),pointer          :: tgtmodes, opmodes_union, subopmodes
   type(ilist_tp),pointer         :: opmodes(:),chmodes(:)
   integer,allocatable            :: chnr(:),chmap(:)
   type(opnode_tp),allocatable    :: chops(:,:)
   type(tensor_t)                 :: tt,ttmp,kron
   integer,parameter              :: maxchnr = 32

   ! Clear error.
   ierr = 0
   errmsg = ""

   ! Number of MLOp-Lists being multiplied.
   nr = size(ops)
   ! Sanity check: need at least one factor.
   if (nr < 1) then
      ierr = 1
      write (errmsg,'(2a,i0,a)') &
         'no MLOp assigned to target ',&
         '(target node =', tgt%num, ')'
      return
   endif

   ! If tgt is primitive, the recursion must terminate. There must be
   ! only one MLOp-List, it must also be primitive, and its mode must
   ! match tgt's mode.
   if (tgt%isprimitive) then
      if (nr /= 1) then
         ierr = 1
         write (errmsg,'(2a,i0,a)') &
            'cannot multiply more than one MLOp into a primitive mode ', &
            '(target node = ', tgt%num, ')'
         return
      endif
      op => ops(1)%p
      if (.not. op%isprimitive) then
         ierr = 1
         write (errmsg,'(2a,i0,a)') &
            'cannot fit a non-primitive MLOp into a primitive mode ', &
            '(target node = ', tgt%num, ')'
         return
      endif
      if (tgt%m /= op%m) then
         ierr = 1
         write (errmsg,'(2a,3(i0,a))') &
            'MLOp mode does not match system mode ', &
            '(target node = ', tgt%num, &
            'system mode = ', tgt%m, 'MLOp mode = ', op%m, ')'
         return
      endif
      ! Just copy the k-terms. Done.
      call set_opleaf_kterms(tgt,op%ks)
      return
   endif

   !--------------------------------------------------------------------
   ! Now, tgt is not primitive. We will later deal with the children of
   ! tgt (recursive descent). But for this, we need to figure out which
   ! MLOp-List belongs to which tgt-child. This will also determine how
   ! to build the transfer tensor of tgt.
   !--------------------------------------------------------------------

   ! Get the list of modes that tgt acts upon.
   tgtmodes => mlop_modenums(tgt)
   ! For each tgt-child, get the list of modes that it acts upon.
   nm = tgt%nmodes
   allocate(chmodes(nm))
   do m=1,nm
      chmodes(m)%p => mlop_modenums(tgt%subops(m)%p)
   enddo
   ! For each MLOp-List, get the list of modes that it acts upon.
   allocate(opmodes(nr))
   do r=1,nr
      opmodes(r)%p => mlop_modenums(ops(r)%p)
   enddo
   ! Sanity check: the union of all the modes that all the MLOps act
   ! upon must exactly be the list of modes that tgt acts upon.
   opmodes_union => ilist_union(opmodes)
   if (.not. ilist_equals(opmodes_union, tgtmodes)) then
      ierr = 1
      write (errmsg,'(2a,i0,a)') &
         'product of MLOps is not congruent with target mode ',&
         '(target node =', tgt%num, ')'
      return
   endif
   ! Clean up a bit.
   call dispose_ilist(opmodes_union)
   call dispose_ilist(tgtmodes)

   !--------------------------------------------------------------------
   ! Based on the information about the mode-lists which we just
   ! gathered, we can determine which MLOp-List should go to which
   ! child of tgt.
   !--------------------------------------------------------------------

   ! chnr(m) = number of MLOp-Lists that belong to m-th tgt-child
   ! chops(:,m) = list of MLOp-Lists that belong there
   allocate(chnr(nm))
   allocate(chops(maxchnr,nm))
   chnr = 0

   ! Start with a unit transfer tensor.
   ! It has one index for each tgt-child, and an additional index
   ! (the last one) for each element of the resulting MLOp-List.
   tt = unit_tensor(nm+1)

   ! Loop over all factors (MLOp-Lists).
   do r=1,nr
      op => ops(r)%p
      ! Check if it belongs to any of the tgt-children.
      ! If yes, it's the m1-th tgt-child. If no, m1==0.
      m1 = 0
      do m=1,nm
         if (ilist_subset(opmodes(r)%p, chmodes(m)%p)) then
            m1 = m
            exit
         endif
      enddo
      if (m1 /= 0) then
         !-----------------------------------------------------------------
         ! MLOp-List #r can be assigned to a tgt-child.
         !-----------------------------------------------------------------

         ! Queue current MLOp-List up for tgt-child.
         chnr(m1) = chnr(m1) + 1
         if (chnr(m1) > maxchnr) then
            ierr = 3
            errmsg = 'mlop_outer_product: maxchnr is too small'
            return
         endif
         chops(chnr(m1),m1)%p => op

         ! Update transfer tensor.
         if (op%nops > 1) then
            ! TODO: explain this
            kron = krondelta_tensor(op%nops)
            map = [m1,nm+1]
            ttmp = tt
            tt = tensor_extend(ttmp,kron,map)
            ! TODO: warn if size increases very much
            ! Clean up.
            call dispose_tensor(ttmp)
            call dispose_tensor(kron)
         else
            ! transfer tensor unchanged
            continue
         endif

      else
         !-----------------------------------------------------------------
         ! MLOp-List #r cannot be assigned to any tgt-child, because it
         ! acts on modes that lie in different tgt-children. Hence we must
         ! decompose this MLOp-List and try to assign its subops to the
         ! tgt-children.
         !-----------------------------------------------------------------

         if (op%isprimitive) then
            ierr = 1
            write (errmsg, '(2a,i0,a)') &
               'cannot decompose primitive MLOp ', &
               '(target node = ', tgt%num, ')'
            return
         endif

         ! Loop over subops of this MLOp-List.
         nsubm = op%nmodes
         allocate(chmap(nsubm+1))
         do l=1,nsubm
            subop => op%subops(l)%p
            ! Check to which tgt-child this subop belongs.
            subopmodes => mlop_modenums(subop)
            m1 = 0
            do m=1,nm
               if (ilist_subset(subopmodes, chmodes(m)%p)) then
                  m1 = m
                  exit
               endif
            enddo
            call dispose_ilist(subopmodes)
            ! No matching tgt-child: the MLOp is more combined than
            ! the target. This cannot work.
            if (m1==0) then
               ierr = 1
               write (errmsg, '(2a,i0,a)') &
                  'MLOp structure is incompatible with target ', &
                  '(target node = ', tgt%num, ')'
               return
            endif
            ! Otherwise queue the subop up for its tgt-child.
            chnr(m1) = chnr(m1) + 1
            if (chnr(m1) > maxchnr) then
               ierr = 3
               errmsg = 'mlop_outer_product: maxchnr is too small'
               return
            endif
            chops(chnr(m1),m1)%p => subop
            ! Remember the mapping subop -> tgt-child. This is needed for
            ! updating/transposing the transfer tensor.
            chmap(l) = m1
         enddo
         chmap(nsubm+1) = nm+1

         ! Update the transfer tensor.
         ! TODO: explain this
         ttmp = tt
         tt = tensor_extend(ttmp,op%tt,chmap)
         ! TODO: warn if tt size increases too much

         ! Clean up.
         call dispose_tensor(ttmp)
         deallocate(chmap)
      endif
   enddo

   ! Store the transfer tensor and associated data.
   tgt%tt = tt
   tgt%nops = tgt%tt%dims(nm+1)

   ! Clean up some more.
   do r=1,nr
      call dispose_ilist(opmodes(r)%p)
   enddo
   deallocate(opmodes)

   ! Now we can process all children.
   do m=1,nm
      ch => tgt%subops(m)%p
      call mlop_outer_product(ch, chops(1:chnr(m),m), ierr, errmsg)
      if (ierr /= 0) return
   enddo

   ! Check for consistency
   do m=1,nm
      ch => tgt%subops(m)%p
      if (tgt%tt%dims(m) /= ch%nops) then
         ierr = 10
         write (errmsg,'(a,5(i0,a))') &
            'inconsistency detected: tgt%dims(',m,') = ',tgt%tt%dims(m), &
            ' BUT tgt%subops(',m,')%nops = ',ch%nops, &
            ' (target node = ', tgt%num, ')'
         return
      endif
   enddo

   ! Clean up.
   deallocate(chops)
   deallocate(chnr)
   do m=1,nm
      call dispose_ilist(chmodes(m)%p)
   enddo
   deallocate(chmodes)

end subroutine mlop_outer_product

end module genmlop
! vim: set ts=3 sw=3 expandtab:
