!=======================================================================
module mlop_eom
!=======================================================================

use global
use maxv, only: maxdim
use mlham
use mltypes
use wftree
use wftree1
use optree
use htermmod
use linear
use timing
use partitions
use gridutil
use mlbase
use hphimod, only: hcall
use xvlib, only: vvtxdz, vvtxzz
use mltensor
use dvrdatmod, only: fkoe,ftime
use operdef
use potevalmod, only: poteval

implicit none


contains !==============================================================


!-----------------------------------------------------------------------
subroutine alloc_all_mlop_tree(wf,nham)
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in) :: wf
   integer,intent(in)        :: nham
   interface
      function ktypefunc(k,dof) result(typeoper)
         integer,intent(in)   :: k
         integer,intent(in)   :: dof
         integer              :: typeoper
      end function ktypefunc
   end interface
   integer                :: i
   type(mlham_t),pointer  :: mlham
   type(optree_t),pointer :: mlop
   integer :: m
   type(wfnode_t),pointer :: wfno
   type(opnode_t),pointer :: opno

   mlham => mlham_get(nham)           ! the Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps

   ! Loop over all MLOps in the Hamiltonian.
   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      ! Allocate space for uterms/wterms/etc.
      call allocate_opterms_tree(wf,mlop)
      ! For CMF, find whether summing up is possible (~mfsum)
      if (wf%cmftyp > 0) then
         do m=1,wf%tot_nodes
            wfno => wf%swap_postorder(m)%p
            opno => mlop%postorder(m)%p
            call set_opnode_htmtype(wfno,opno)
         enddo
      endif
   enddo
end subroutine alloc_all_mlop_tree


!-----------------------------------------------------------------------
function compute_all_mlop_etot(time,nham) result(etot)
!-----------------------------------------------------------------------
   real(dbl),intent(in)    :: time
   integer,intent(in)      :: nham
   complex(cdbl)           :: etot
   type(mlham_t),pointer   :: mlham
   type(optree_t),pointer  :: mlop
   integer                 :: i,kkoe

   etot = CZERO
   mlham => mlham_get(nham)           ! the Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps

   ! Add contributions from all MLOps in the Hamiltonian.
   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      kkoe = mlham%kkoes(i)
      etot = etot + compute_mlop_etot(time,kkoe,mlop)
   enddo
end function compute_all_mlop_etot



!-----------------------------------------------------------------------
subroutine compute_all_uterms_tree(time,nham,tree,psi)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   integer,intent(in)           :: nham
   type(wfnode_t),intent(inout) :: tree
   complex(cdbl),intent(in)     :: psi(:)

   integer                :: i
   type(mlham_t),pointer  :: mlham
   type(optree_t),pointer :: mlop

   mlham => mlham_get(nham)           ! the Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps
   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      call compute_uterms_tree(time,tree,psi,mlop)
   enddo
end subroutine compute_all_uterms_tree


!-----------------------------------------------------------------------
subroutine compute_all_uterms2_tree(time,nham,tree,psi1,psi2)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   integer,intent(in)           :: nham
   type(wfnode_t),intent(inout) :: tree
   complex(cdbl),intent(in)     :: psi1(:),psi2(:)

   integer                :: i
   type(mlham_t),pointer  :: mlham
   type(optree_t),pointer :: mlop

   mlham => mlham_get(nham)           ! the Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps
   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      call compute_uterms2_tree(time,tree,psi1,psi2,mlop)
   enddo
end subroutine compute_all_uterms2_tree


!-----------------------------------------------------------------------
subroutine compute_all_wterms_tree(time,nham,tree,psi)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   integer,intent(in)           :: nham
   type(wfnode_t),intent(inout) :: tree
   complex(cdbl),intent(in)     :: psi(:)
   integer                      :: i
   type(mlham_t),pointer        :: mlham
   type(optree_t),pointer       :: mlop
   integer                      :: kkoe

   mlham => mlham_get(nham)           ! the Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps

   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      kkoe = mlham%kkoes(i)
      call compute_wterms_tree(time,kkoe,tree,psi,mlop)
   enddo
end subroutine compute_all_wterms_tree


!-----------------------------------------------------------------------
subroutine compute_all_mlop_mfsum(wfno,phi,mfsum)
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)   :: wfno
   complex(cdbl),intent(in)    :: phi(:)
   complex(cdbl),intent(inout) :: mfsum(:,:,:)
   type(mlham_t),pointer  :: mlham
   type(optree_t),pointer :: mlop
   type(opnode_t),pointer :: opno
   integer                :: i

   mlham => mlham_get(1)              ! system Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps

   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      ! Find matching opnode from wfnode.
      opno => mlop%preorder(wfno%num)%p
      ! Compute and add contribution of this MLOp to mfsum.
      call mlop_mfsum_node(opno,wfno,phi,mfsum)
   enddo
end subroutine compute_all_mlop_mfsum


!-----------------------------------------------------------------------
subroutine compute_all_mlop_rhs_node(wfno,psi,dtpsi)
!-----------------------------------------------------------------------
   type(wfnode_t),intent(inout) :: wfno
   complex(cdbl),intent(in)     :: psi(:)
   complex(cdbl),intent(inout)  :: dtpsi(:)
   type(mlham_t),pointer  :: mlham
   type(optree_t),pointer :: mlop
   type(opnode_t),pointer :: opno
   integer                :: i

   mlham => mlham_get(1)              ! system Hamiltonian
   if (.not.associated(mlham)) return ! no MLOps

! parallel if(mlham%nmlop > 1) default(shared) private(mlop,opno) reduction(+:dtpsi)
! do schedule(runtime)
   do i=1,mlham%nmlop
      mlop => mlham%mlops(i)%p
      ! Find matching opnode from wfnode.
      opno => mlop%preorder(wfno%num)%p
      ! Compute contribution of this MLOp to dtpsi.
      call mlop_rhs_node(opno,wfno,psi,dtpsi)
   enddo
! end do
! end parallel
end subroutine compute_all_mlop_rhs_node


!-----------------------------------------------------------------------
function compute_mlop_etot(time,kkoe,op) result(etot)
!-----------------------------------------------------------------------
   complex(cdbl)             :: etot
   real(dbl),intent(in)      :: time
   integer,intent(in)        :: kkoe
   type(optree_t),intent(in) :: op
   integer                   :: ktype
   complex(cdbl)             :: coe
   type(opnode_t),pointer    :: opno
   ! <Psi|U^Ø|Psi> is just the uterm of the top node, times the
   ! coefficient. The uterms must already have been computed, by
   ! compute_uterms_tree or compute_uterms2_tree.
   call if_koeff(time,kkoe,ktype,coe)
   opno => op%topnode
   etot = coe * opno%uterms(1)%p%matrix(1,1)
end function compute_mlop_etot


!-----------------------------------------------------------------------
subroutine compute_uterms_tree(time,wf,psi,op)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   type(wfnode_t),intent(inout) :: wf
   complex(cdbl),intent(in)     :: psi(:)
   type(optree_t),intent(inout) :: op

   integer                :: i
   type(wfnode_t),pointer :: wfno
   type(opnode_t),pointer :: opno
   integer,save           :: tid = 0

   call get_timer(tid,"uterms_tree")
   call start_timer(tid)
   do i=1,wf%tot_nodes
      wfno => wf%swap_up(i)%p
      opno => op%postorder(i)%p
      call compute_uterms_node(time,wfno,psi,opno)
   enddo
   call stop_timer(tid)
end subroutine compute_uterms_tree


!-----------------------------------------------------------------------
subroutine compute_uterms2_tree(time,wf,psi1,psi2,op)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   type(wfnode_t),intent(inout) :: wf
   complex(cdbl),intent(in)     :: psi1(:),psi2(:)
   type(optree_t),intent(inout) :: op

   integer                :: i
   type(wfnode_t),pointer :: wfno
   type(opnode_t),pointer :: opno
   integer,save           :: tid = 0

   call get_timer(tid,"uterms2_tree")
   call start_timer(tid)
   do i=1,wf%tot_nodes
      wfno => wf%swap_up(i)%p
      opno => op%postorder(i)%p
      call compute_uterms2_node(time,wfno,psi1,psi2,opno)
   enddo
   call stop_timer(tid)
end subroutine compute_uterms2_tree


!-----------------------------------------------------------------------
subroutine compute_wterms_tree(time,kkoe,wf,psi,op)
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   integer,intent(in)           :: kkoe
   type(wfnode_t),intent(inout) :: wf
   complex(cdbl),intent(in)     :: psi(:)
   type(optree_t),intent(inout) :: op
   integer                      :: i
   type(wfnode_t),pointer       :: wfno
   type(opnode_t),pointer       :: opno
   integer,save                 :: tid = 0
   complex(cdbl)                :: coe
   integer                      :: ktype

   call get_timer(tid,"wterms_tree")
   call start_timer(tid)
   ! Store (time-dep) coefficient in wterm of top node.
   call if_koeff(time,kkoe,ktype,coe)
   op%topnode%wterms(1)%p%matrix(1,1) = coe
   ! Compute wterms down-recursively.
   do i=1,wf%tot_nodes
      wfno => wf%swap_down(i)%p
      opno => op%preorder(i)%p
      call compute_wterms_node(wfno,psi,opno)
   enddo
   call stop_timer(tid)
end subroutine compute_wterms_tree


!-----------------------------------------------------------------------
subroutine allocate_opterms_tree(wf,op)
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)    :: wf
   type(optree_t),intent(inout) :: op
   type(wfnode_t),pointer       :: wfno,wfch
   type(opnode_t),pointer       :: opno,opch
   integer                      :: i,m,c,blen,xlen
   if (op%isallocated) return
   ! Loop over all nodes in the tree.
   do i=1,wf%tot_nodes
      wfno => wf%swap_down(i)%p
      opno => op%preorder(i)%p
      ! Allocate wterms of this node.
      allocate(opno%wterms(opno%nops))
      do c=1,opno%nops
         opno%wterms(c)%p => new_hterm(2,wfno%navecs)
      enddo
      ! V*W is computed when doing the wterms, and can be reused for the RHS
      if (.not.wfno%isprimitive) then
         blen = product(opno%tt%dims(1:opno%nmodes))
         xlen = wfno%navecs**2
         allocate(opno%vw(xlen,blen))
      endif
      ! Allocate uterms of the children of this node.
      ! (This is supposed to improve memory locality.)
      if (.not.wfno%isprimitive) then
         do m=1,wfno%nmodes
            wfch => wfno%submodes(m)%p
            opch => opno%subops(m)%p
            allocate(opch%uterms(opch%nops))
            do c=1,opch%nops
               ! TODO: use unit hterm if possible (?)
               opch%uterms(c)%p => new_hterm(2,wfch%navecs)
            enddo
         enddo
      endif
   enddo
   ! Top node: allocate uterm & set wterm to one.
   opno => op%topnode
   allocate(opno%uterms(1))
   opno%uterms(1)%p => new_hterm(2,1)
   opno%wterms(1)%p => new_hterm(2,1)
   opno%wterms(1)%p%matrix(1,1) = CONE
   ! Precompute some optimization.
   do i=1,wf%tot_nodes
      opno => op%preorder(i)%p
      call optimize_opterms_node(opno)
   enddo
   ! Don't do this again.
   op%isallocated = .true.
end subroutine allocate_opterms_tree


!-----------------------------------------------------------------------
subroutine set_opnode_htmtype(wfno,opno)
!-----------------------------------------------------------------------
   type(wfnode_t),intent(in)    :: wfno
   type(opnode_t),intent(inout) :: opno

   integer :: c,k
   if (wfno%isprimitive) then
      ! opno%ktypes has already been allocated
      do c=1,opno%nops
         k = opno%ks(c)
         opno%ktypes(c) = get_mode_htmtype(wfno,k)
      enddo
   endif
end subroutine set_opnode_htmtype


!-----------------------------------------------------------------------
subroutine optimize_opterms_node(opno)
!-----------------------------------------------------------------------
! For reducing the size of the workspace needed in the computation
! of the uterms, wterms, and RHS, the application of the uterms of
! the node's children (for uterms/RHS) or siblings (for wterms) is
! split into two parts.  Here we find the split which minimizes the
! workspace.
!-----------------------------------------------------------------------
   type(opnode_t),pointer :: opno
   type(opnode_t),pointer :: child
   integer                :: nchld,mm,p
   integer,pointer        :: chlddim(:)
   logical,allocatable    :: usesib(:)
   if (opno%isprimitive) return

   ! Optimize which children's uterms to apply in the first half
   ! of the uterms/RHS computation.

   nchld = opno%nmodes
   chlddim => opno%tt%dims(1:nchld)
   allocate(opno%early_utm_child(nchld))
   call minimize_sumprod(nchld, chlddim, opno%early_utm_child)

   ! Optimize which siblings' uterms to apply in step 2 of the
   ! wterms computation.

   allocate(usesib(nchld))
   ! Loop over children.
   do mm=1,nchld
      child => opno%subops(mm)%p
      ! Find optimal split.
      call minimize_sumprod(nchld,chlddim,usesib)
      ! Never apply own uterms!
      if (usesib(mm)) usesib=.not.usesib
      ! Store the split.
      allocate(child%early_wtm_sibling(nchld))
      child%early_wtm_sibling = usesib
   enddo
   ! Clean up.
   deallocate(usesib)

   contains !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   subroutine minimize_sumprod(ndim,dims,ptmin)
      integer,intent(in)  :: ndim
      integer,intent(in)  :: dims(ndim)
      logical,intent(out) :: ptmin(ndim)
      logical             :: pt(ndim)
      integer             :: spmin,sp
      ptmin = .false.
      spmin = product(dims)+1
      do p=1,partition_count(ndim)
         pt = partition_get(ndim,p)
         sp = product(dims,mask=pt) + product(dims,mask=.not.pt)
         if (sp<spmin) then
            ptmin = pt
            spmin = sp
         endif
      enddo
      if (product(dims,mask=pt) < product(dims,mask=.not.pt)) &
         ptmin = .not.ptmin
      where(dims==1) ptmin=.true.
   end subroutine minimize_sumprod

end subroutine optimize_opterms_node


!-----------------------------------------------------------------------
subroutine compute_uterms_node(time,wfno,psi,opno)
!-----------------------------------------------------------------------
! Eq.(82) in my notes.
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   real(dbl)                    :: time1
   type(wfnode_t),intent(inout) :: wfno
   complex(cdbl),intent(in)     :: psi(:)
   type(opnode_t),intent(inout) :: opno

   integer                   :: c,i,j,k,z0,zf,philen,m,nm
   integer                   :: bdim(wfno%nmodes)
   integer                   :: blen,bb,bb1,bb2,ij,ztt
   integer                   :: b(wfno%nmodes)
   complex(cdbl),allocatable :: work(:)
   complex(cdbl),pointer     :: upsi1(:,:,:)
   complex(cdbl),pointer     :: upsi2(:,:,:)
   complex(cdbl),allocatable :: psiupsi(:,:)
   complex(cdbl),pointer     :: uterm(:,:)
   type(opnode_t),pointer    :: opch
   logical,pointer           :: mmask(:)

! dummy call to use argument time
   time1=time

   if (wfno%tid_uterms==0) &
      call init_node_timer(wfno, wfno%tid_uterms, 'uterms')
   call start_timer(wfno%tid_uterms)

   philen=product(wfno%ndim)

   if (wfno%isprimitive) then
      allocate(work(philen))
      do c=1,opno%nops
         k=opno%ks(c)
         ! TODO: skip unit k-terms
         do j=1,wfno%navecs
            ! Apply primitive mode operator (encoded in kterm) to j-th SPF.
            z0 = wfno%z(j,1)
            zf = wfno%z(j,2)
            call hcall(wfno%m, k, work, psi(z0:zf), philen)
            do i=1,wfno%navecs
               ! Do matrix element <phi_i|U_c|phi_j>.
               z0 = wfno%z(i,1)
               zf = wfno%z(i,2)
               opno%uterms(c)%p%matrix(i,j) = phi_phi(psi(z0:zf),work)
            enddo
         enddo
      enddo
      deallocate(work)

   else

      nm = wfno%nmodes
      bdim = opno%tt%dims(1:nm)
      blen = product(bdim)
      mmask => opno%early_utm_child

      ! The uterms of the children must be applied to the avecs.
      ! It takes less space and time to apply half of the uterms
      ! to A, and the other half to A*.

      ! Copy A and A*.
      allocate(upsi1(philen,wfno%navecs,1))
      allocate(upsi2(philen,wfno%navecs,1))
      do i=1,wfno%navecs
         z0 = wfno%z(i,1)
         zf = wfno%z(i,2)
         upsi1(:,i,1) = psi(z0:zf)
         upsi2(:,i,1) = dconjg(psi(z0:zf))
      enddo

      ! Apply half of the uterms to A.
      do m=1,nm
         if (.not.mmask(m)) cycle
         opch => opno%subops(m)%p
         call u_expand(opch%uterms,m,upsi1,wfno%ndim)
      enddo

      ! Apply the remaining uterms to A*.
      do m=1,nm
         if (mmask(m)) cycle
         opch => opno%subops(m)%p
         call u_expand(opch%uterms,m,upsi2,wfno%ndim,.true.)
      enddo

      ! TODO: if the transfer tensor is sparse, it should be cheaper to
      ! do the following two steps together, as the contraction can be
      ! omitted for those elements of tt that are negligible.

      ! Contraction:
      ! psiupsi(B,i j) := Sum_L upsi2(L,i,b2) upsi1(L,j,b1)
      allocate(psiupsi(blen,wfno%navecs**2))
      b = 1
  20  bb2 = flat_index(b,bdim,.not.mmask)
  10  bb1 = flat_index(b,bdim,mmask)
      bb  = flat_index(b,bdim)
      ij = 1
      do j=1,wfno%navecs
         do i=1,wfno%navecs
            call vvtxzz(upsi2(1:philen,i,bb2), upsi1(1:philen,j,bb1), &
                 psiupsi(bb,ij), philen)
            ij = ij+1
         enddo
      enddo
      if (next_point(b,bdim,mmask)) goto 10
      if (next_point(b,bdim,.not.mmask)) goto 20

      ! Build the uterms.
      ! uterms(c)[i,j] = Sum_B tt(B,c) psiupsi(B,i j)
      ztt = 1
      do c=1,opno%nops
         uterm => opno%uterms(c)%p%matrix
         ij = 1
         do j=1,wfno%navecs
            do i=1,wfno%navecs
               call vvtxdz(opno%tt%vr(ztt:ztt+blen-1), psiupsi(1,ij),&
                    uterm(i,j), blen)
               ij = ij+1
            enddo
         enddo
         ztt = ztt+blen
      enddo

      ! Clean up.
      deallocate(psiupsi)
      deallocate(upsi2)
      deallocate(upsi1)

   endif
   call stop_timer(wfno%tid_uterms)

end subroutine compute_uterms_node


!-----------------------------------------------------------------------
subroutine compute_uterms2_node(time,wfno,psi1,psi2,opno)
!-----------------------------------------------------------------------
! Chapter 4.3 in my notes.
!-----------------------------------------------------------------------
   real(dbl),intent(in)         :: time
   real(dbl)                    :: time1
   type(wfnode_t),intent(inout) :: wfno
   complex(cdbl),intent(in)     :: psi1(:),psi2(:)
   type(opnode_t),intent(inout) :: opno

   integer                   :: c,i,j,k,z0,zf,philen,m,nm
   integer                   :: bdim(wfno%nmodes)
   integer                   :: blen,bb,bb1,bb2,ij,ztt
   integer                   :: b(wfno%nmodes)
   complex(cdbl),allocatable :: work(:)
   complex(cdbl),pointer     :: upsi1(:,:,:)
   complex(cdbl),pointer     :: upsi2(:,:,:)
   complex(cdbl),allocatable :: psi1upsi2(:,:)
   complex(cdbl),pointer     :: uterm(:,:)
   type(opnode_t),pointer    :: opch
   logical,pointer           :: mmask(:)

! dummy call to use argument time
   time1=time

   if (wfno%tid_uterms==0) &
      call init_node_timer(wfno, wfno%tid_uterms, 'uterms2')
   call start_timer(wfno%tid_uterms)

   philen=product(wfno%ndim)

   if (wfno%isprimitive) then
      allocate(work(philen))
      do c=1,opno%nops
         k=opno%ks(c)
         do j=1,wfno%navecs
            ! Apply primitive mode operator (encoded in kterm) to j-th SPF of psi2.
            z0 = wfno%z(j,1)
            zf = wfno%z(j,2)
            call hcall(wfno%m, k, work, psi2(z0:zf), philen)
            do i=1,wfno%navecs
               ! Do matrix element <phi1_i|U_c|phi2_j>.
               z0 = wfno%z(i,1)
               zf = wfno%z(i,2)
               opno%uterms(c)%p%matrix(i,j) = phi_phi(psi1(z0:zf),work)
            enddo
         enddo
      enddo
      deallocate(work)

   else

      nm = wfno%nmodes
      bdim = opno%tt%dims(1:nm)
      blen = product(bdim)
      mmask => opno%early_utm_child

      ! The uterms of the children must be applied to the avecs.
      ! It takes less space and time to apply half of the uterms
      ! to A2, and the other half to A1*.

      ! Copy A2 and A1*.
      allocate(upsi2(philen,wfno%navecs,1))
      allocate(upsi1(philen,wfno%navecs,1))
      do i=1,wfno%navecs
         z0 = wfno%z(i,1)
         zf = wfno%z(i,2)
         upsi2(:,i,1) = psi2(z0:zf)
         upsi1(:,i,1) = dconjg(psi1(z0:zf))
      enddo

      ! Apply half of the uterms to A2.
      do m=1,nm
         if (.not.mmask(m)) cycle
         opch => opno%subops(m)%p
         call u_expand(opch%uterms,m,upsi2,wfno%ndim)
      enddo

      ! Apply the remaining uterms to A1*.
      do m=1,nm
         if (mmask(m)) cycle
         opch => opno%subops(m)%p
         call u_expand(opch%uterms,m,upsi1,wfno%ndim,.true.)
      enddo

      ! TODO: if the transfer tensor is sparse, it should be cheaper to
      ! do the following two steps together, as the contraction can be
      ! omitted for those elements of tt that are negligible.

      ! Contraction:
      ! psi1upsi2(B,i j) := Sum_L upsi1(L,i,b2) upsi2(L,j,b1)
      allocate(psi1upsi2(blen,wfno%navecs**2))
      b = 1
  20  bb2 = flat_index(b,bdim,.not.mmask)
  10  bb1 = flat_index(b,bdim,mmask)
      bb  = flat_index(b,bdim)
      ij = 1
      do j=1,wfno%navecs
         do i=1,wfno%navecs
       !GW!!     call vvtxzz(upsi1(1,i,bb2), upsi2(1,j,bb1), psi1upsi2(bb,ij), philen)
            ij = ij+1
         enddo
      enddo
      if (next_point(b,bdim,mmask)) goto 10
      if (next_point(b,bdim,.not.mmask)) goto 20

      ! Build the uterms.
      ! uterms(c)[i,j] = Sum_B tt(B,c) psi1upsi2(B,i j)
      ztt = 1
      do c=1,opno%nops
         uterm => opno%uterms(c)%p%matrix
         ij = 1
         do j=1,wfno%navecs
            do i=1,wfno%navecs
        ! GW!!      call vvtxdz( opno%tt%vr(ztt), psi1upsi2(1,ij), uterm(i,j), blen)
               ij = ij+1
            enddo
         enddo
         ztt = ztt+blen
      enddo

      ! Clean up.
      deallocate(psi1upsi2)
      deallocate(upsi1)
      deallocate(upsi2)

   endif
   call stop_timer(wfno%tid_uterms)

end subroutine compute_uterms2_node


!-----------------------------------------------------------------------
subroutine compute_wterms_node(wfno,psi,opno)
!-----------------------------------------------------------------------
! Eq.(84) in my notes.
! This computes the wterms for all children of the current node.
! The (maybe time-dependent) coefficient has been stored in the wterm
! of the topnode (which would otherwise be 1), so that it propagates
! to all wterms.
!-----------------------------------------------------------------------
   type(wfnode_t),intent(inout) :: wfno
   complex(cdbl),intent(in)     :: psi(:)
   type(opnode_t),intent(inout) :: opno
   integer                      :: philen,blen,b1len,b2len
   integer                      :: c,zc,cdim
   integer                      :: m,m1,m2,nm,nm1,nm2
   integer                      :: mm,i,z0,zf
   complex(cdbl),pointer        :: upsi(:,:,:)
   complex(cdbl),pointer        :: uvwpsi(:,:,:)
   integer                      :: bdim(wfno%nmodes)
   integer                      :: b1dim(wfno%nmodes)
   integer                      :: b2dim(wfno%nmodes)
   type(opnode_t),pointer       :: opch,opsib

   if (wfno%isprimitive) return

   ! Let p be the number of submodes combined in this node.
   ! To compute the wterms for submode m, we need to contract
   ! 3 tensors of order p+1 (A,A*,V) and p tensors of order 3
   ! (W, U^1 ... U^p omitting U^m) over all common indices.
   ! The optimal order for performing these contractions (regarding
   ! computational cost, but constrained by available workspace) in
   ! general depends on the extents of all the dimensions involved.
   ! Finding this order is a formidable problem in itself.

   ! Based on experimentation with a simple solver for the mentioned
   ! optimization problem, here the following general strategy will
   ! be used. While not being generally optimal, it shouldn't be too
   ! bad for the majority of practical cases.
   ! 1) compute W*V
   ! 2) apply (the bigger) half of the U to A
   ! 3) contract with W*V  <== expected bottleneck
   ! 4) apply rest of the U
   ! 5) contract with A*
   ! The major contribution to the size of the required workspace is
   ! size(A)*[ product{ subops(m')%nops | U^m' used in step 2 } +
   !           product{ subops(m')%nops | U^m' not used in step 2, incl. U^m } ]
   ! This can be minimized by choosing which U^m' to use in step 2.

   if (wfno%tid_wterms==0) &
      call init_node_timer(wfno, wfno%tid_wterms, 'wterms')
   call start_timer(wfno%tid_wterms)

   ! 1) VW[il,B] = Sum_c W[il,c] V[B,c]
   nm = wfno%nmodes
   bdim(1:nm) = opno%tt%dims(1:nm)
   blen = product(bdim)
   cdim = opno%nops
   opno%vw = CZERO
   zc = 1
   do c=1,cdim
      call addvw(opno%vw, opno%wterms(c)%p%matrix, opno%tt%vr(zc:zc+blen-1), wfno%navecs**2, blen)
      zc = zc+blen
   enddo

   ! Loop over children.
   philen = product(wfno%ndim)
   do mm=1,nm
      opch => opno%subops(mm)%p

      ! 2) apply half the uterms to A
      ! First, copy A.
      allocate(upsi(philen, wfno%navecs, 1))
      do i=1,wfno%navecs
         z0 = wfno%z(i,1)
         zf = wfno%z(i,2)
         upsi(:,i,1) = psi(z0:zf)
      enddo
      ! Loop over (selected) siblings.
      m1 = 0
      m2 = 0
      do m=1,nm
         opsib => opno%subops(m)%p
         if (opch%early_wtm_sibling(m)) then
            ! Sum_{j_m} A[K,i] * U[j_m k_m b_m]
            call u_expand(opsib%uterms, m, upsi, wfno%ndim)
            m1 = m1+1
            b1dim(m1) = opsib%nops
         else
            ! Store the sizes of the uterms that we skip here.
            m2 = m2+1
            b2dim(m2) = opsib%nops
         endif
      enddo
      nm1 = m1
      nm2 = m2

      ! 3) contract with VW
      ! Example: nm = 4, early_wtm_sibling = T T F F
      ! UVWA[J,l,b3 b4] = Sum_{b1 b2 i} UA[J,i,b1 b2] VW[il,B]
      b1len = product(b1dim(1:nm1))
      b2len = product(b2dim(1:nm2))
      allocate(uvwpsi(philen, wfno%navecs, b2len))
      call vw_contract(uvwpsi, opno%vw, upsi, philen, b1len, b2len, &
           wfno%navecs, nm, bdim, opch%early_wtm_sibling)
      deallocate(upsi)

      ! 4) apply remaining uterms
      m2 = 1
      do m=1,nm
         if (opch%early_wtm_sibling(m)) cycle
         if (m /= mm) then
            ! Sum_{j_m b_m} UVWA*[J,i,b3 b4] * U[j_m k_m b_m]
            opsib => opno%subops(m)%p
            call u_contract(opsib%uterms, m, m2, uvwpsi, wfno%ndim, b2dim(1:nm2))
         endif
         m2 = m2+1
      enddo
      !!! DEBUG
      if (size(uvwpsi,3) /= opch%nops) &
         call stop_code("mlop_eom_m", "compute_wterms_tree", "size(upsi,3) /= opch%nops")
      !!! END DEBUG

      ! 5) contract with A*
      z0 = wfno%z(1,1)
      zf = wfno%z(wfno%navecs,2)
      do c=1,opch%nops
         call a_contract(opch%wterms(c)%p%matrix, uvwpsi(:,:,c), psi(z0:zf), mm, wfno%ndim)
      enddo

      ! Done.
      deallocate(uvwpsi)
   enddo

   ! Clean up.
   call stop_timer(wfno%tid_wterms)

   contains !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   subroutine addvw(vw,w,v,dim1,dim2)
      integer                     :: dim1,dim2
      complex(cdbl),intent(inout) :: vw(dim1,dim2)
      complex(cdbl),intent(in)    :: w(dim1)
      real(dbl),intent(in)        :: v(dim2)
      integer                     :: i,j
      do j=1,dim2
         do i=1,dim1
            vw(i,j) = vw(i,j) + w(i)*v(j)
         enddo
      enddo
   end subroutine addvw

   subroutine a_contract(wterm, upsi, psi, m, phidim)
      complex(cdbl),intent(out) :: wterm(:,:)
      complex(cdbl),intent(in)  :: upsi(:,:)
      complex(cdbl),intent(in)  :: psi(:)
      integer,intent(in)        :: m
      integer,intent(in)        :: phidim(:)
      integer                   :: nm
      integer                   :: psidim(size(phidim)+1)
      integer                   :: ijk(3)
      nm = size(phidim)
      psidim(1:nm) = phidim
      psidim(nm+1) = size(upsi,2) ! navecs
      ijk = ijk_shape(m,psidim)
      call dmat_tensor_tensor_compl(wterm,psi,upsi,ijk)
   end subroutine a_contract

end subroutine compute_wterms_node


!-----------------------------------------------------------------------
subroutine mlop_mfsum_node(opno,wfno,phi,mfsum)
!-----------------------------------------------------------------------
   type(opnode_t),intent(in)    :: opno
   type(wfnode_t),intent(in)    :: wfno
   complex(cdbl),intent(in)     :: phi(:)       ! all CONE
   complex(cdbl),intent(inout)  :: mfsum(:,:,:)
   integer :: philen,c,k,i,j
   complex(cdbl),pointer :: wterm(:,:)
   complex(cdbl),allocatable :: hphi(:)

   philen = product(wfno%ndim)
   allocate(hphi(philen))
   do c=1,opno%nops
      ! Skip non-diagonal operators.
      if (opno%ktypes(c) < 0) cycle
      ! Apply primitive mode operator (encoded in kterm).
      k = opno%ks(c)
      call hcall(wfno%m, k, hphi, phi, philen)
      ! Multiply inverse density matrix with wterm.
      wterm => opno%wterms(c)%p%matrix
      ! Add to the mean-field-sum.
      do j=1,wfno%navecs
         do i=1,wfno%navecs
            mfsum(:,i,j) = mfsum(:,i,j) + wterm(i,j)*hphi(:)
         enddo
      enddo
   enddo
   deallocate(hphi)
end subroutine mlop_mfsum_node


!-----------------------------------------------------------------------
subroutine mlop_rhs_node(opno,wfno,psi,dtpsi)
!-----------------------------------------------------------------------
! Eqs.(86-87) in my notes.
!-----------------------------------------------------------------------
   type(opnode_t),intent(in)    :: opno
   type(wfnode_t),intent(in)    :: wfno
   complex(cdbl),intent(in)     :: psi(:)
   complex(cdbl),intent(inout)  :: dtpsi(:)
   type(opnode_t),pointer    :: opch
   complex(cdbl),pointer     :: wterm(:,:)
   integer                   :: philen,psilen,b1len,b2len
   integer                   :: c,k,i,j,z0,zf,z00,zff
   integer                   :: nm,nm1,nm2,m,m1,m2
   complex(cdbl),allocatable :: work(:)
   complex(cdbl),pointer     :: upsi(:,:,:)
   complex(cdbl),pointer     :: uvwpsi(:,:,:)
   integer                   :: bdim(wfno%nmodes)
   integer                   :: b1dim(wfno%nmodes)
   integer                   :: b2dim(wfno%nmodes)
   logical                   :: lcmf
!  integer,save              :: tid = 0

!  call get_timer(tid,"rhs_mlop")
!  call start_timer(tid)

   philen = product(wfno%ndim)
   psilen = philen*wfno%navecs
   lcmf = (wfno%topnode%cmftyp /= 0)

   if (wfno%isprimitive) then ! -- LEAF NODE

      allocate(work(psilen))
      do c=1,opno%nops
         ! For CMF, skip summed-up terms -- they are included in wfno%mfsum,
         ! which gets applied in compute_rhs_node.
         if (lcmf) then
            if (opno%ktypes(c) >= 0) cycle
         endif
         k = opno%ks(c)
         do j=1,wfno%navecs
            z0 = wfno%zr(j,1)
            zf = wfno%zr(j,2)
            ! Apply primitive mode operator (encoded in kterm) to j-th SPF.
            call hcall(wfno%m, k, work(z0:zf), psi(z0:zf) , philen)
         enddo
         ! Apply wterm. Accumulate result in dtpsi.
         wterm => opno%wterms(c)%p%matrix
         do j=1,wfno%navecs
            z00 = wfno%zr(j,1)
            zff = wfno%zr(j,2)
            do i=1,wfno%navecs
               z0 = wfno%zr(i,1)
               zf = wfno%zr(i,2)
               dtpsi(z0:zf) = dtpsi(z0:zf) + wterm(i,j)*work(z00:zff)
            enddo
         enddo
      enddo
      deallocate(work)

   else ! -- NODE WITH CHILDREN

      nm = wfno%nmodes
      bdim = opno%tt%dims(1:nm)

      ! Copy A.
      allocate(upsi(philen, wfno%navecs, 1))
      do j=1,wfno%navecs
         z0 = wfno%zr(j,1)
         zf = wfno%zr(j,2)
         upsi(:,j,1) = psi(z0:zf)
      enddo

      ! Apply first part of the uterms to A.
      m1 = 0
      m2 = 0
      do m=1,nm
         opch => opno%subops(m)%p
         if (opno%early_utm_child(m)) then
            call u_expand(opch%uterms, m, upsi, wfno%ndim)
            m1 = m1+1
            b1dim(m1) = opch%nops
         else
            m2 = m2+1
            b2dim(m2) = opch%nops
         endif
      enddo
      nm1 = m1
      nm2 = m2

      ! Contract with VW.
      b1len = product(b1dim(1:nm1))
      b2len = product(b2dim(1:nm2))
      allocate(uvwpsi(philen, wfno%navecs, b2len))
      call vw_contract(uvwpsi, opno%vw, upsi, philen, b1len, b2len, &
           wfno%navecs, nm, bdim, opno%early_utm_child)
      deallocate(upsi)

      ! Apply remaining uterms.
      m2 = 1
      do m=1,nm
         if (opno%early_utm_child(m)) cycle
         opch => opno%subops(m)%p
         call u_contract(opch%uterms, m, m2, uvwpsi, wfno%ndim, b2dim(1:nm2))
         m2 = m2+1
      enddo
      !!! DEBUG
      if (size(uvwpsi,3) /= 1) &
         call stop_code("mlop_eom_m", "mlop_rhs_node", "size(uvwpsi,3) /= 1")
      !!! END DEBUG

      ! Add result to dtpsi.
      do i=1,wfno%navecs
         z0 = wfno%zr(i,1)
         zf = wfno%zr(i,2)
         dtpsi(z0:zf) = dtpsi(z0:zf) + uvwpsi(:,i,1)
      enddo

      ! Done.
      deallocate(uvwpsi)

   endif

!  call stop_timer(tid)

end subroutine mlop_rhs_node


!-----------------------------------------------------------------------
subroutine u_expand(uterms,f,psi,phidim,tp)
!-----------------------------------------------------------------------
! psi(L,j,ab) := uterms(b) * psi(L,j,a)  ; L = l1 l2 l3 ...
! where "*" means "apply to f-th index (of L)"
!-----------------------------------------------------------------------
   type(hterm_tp),intent(in)   :: uterms(:)  ! list of hterms to be applied
   integer,intent(in)          :: f          ! to the f-th index
   complex(cdbl),pointer       :: psi(:,:,:) ! see above; overwritten/reallocated here
   integer,intent(in)          :: phidim(:)  ! = l1 l2 l3 ...
   logical,intent(in),optional :: tp         ! use transpose of hterms or not
   complex(cdbl),pointer       :: upsi(:,:,:)
   integer                     :: psidim(size(phidim)+1)
   integer                     :: plen,na,nb,nm,b,ab
   logical                     :: tp1
   tp1 = .false.
   if (present(tp)) tp1=tp

   plen = size(psi)
   nb = size(uterms)
   na = size(psi,3)
   nm = size(phidim)
   psidim(1:nm) = phidim
   psidim(nm+1) = size(psi,2)*na

   if (nb>1) then
      allocate(upsi(size(psi,1), size(psi,2), na*nb))
!$OMP parallel do default(shared) private(b,ab)
      do b=1,nb
         ab = (b-1)*na
         upsi(:,:,ab+1:ab+na) = psi
         call apply_hterm(uterms(b)%p, f, plen, upsi(:,:,ab+1:ab+na), psidim, tp1)
      enddo
!$OMP end parallel do
      deallocate(psi)
      psi => upsi
   else
      call apply_hterm(uterms(1)%p, f, plen, psi, psidim, tp1)
   endif
end subroutine u_expand


!-----------------------------------------------------------------------
subroutine u_contract(uterms,f,g,psi,phidim,bdim,tp)
!-----------------------------------------------------------------------
! psi(L,j,B\bg) := Sum_{bg} uterms(bg)*psi(L,j,B)  ; L = l1 l2 l3 ... ; B = b1 b2 ...
! where "*" means "apply to f-th index (of L)"
!-----------------------------------------------------------------------
   type(hterm_tp),intent(in)   :: uterms(:)  ! list of hterms to be applied
   integer,intent(in)          :: f          ! to the f-th index of L
   integer,intent(in)          :: g          ! and to the g-th index of B
   complex(cdbl),pointer       :: psi(:,:,:) ! see above; overwritten/reallocated here
   integer,intent(in)          :: phidim(:)  ! = l1 l2 l3 ...
   integer,intent(inout)       :: bdim(:)    ! = b1 b2 ...
   logical,intent(in),optional :: tp         ! use transpose of hterms or not
   integer                     :: nm,philen,navecs,blen,nu,blen1,bb,bb1,bg
   integer                     :: psidim(size(phidim)+1)
   integer                     :: bidx(size(bdim))
   logical                     :: lmask(size(bdim))
   complex(cdbl),pointer       :: upsi(:,:,:)
   logical                     :: tp1
   tp1 = .false.
   if (present(tp)) tp1=tp

   nm     = size(phidim)
   philen = size(psi,1)
   navecs = size(psi,2)
   blen   = size(psi,3)
   nu     = size(uterms)
   !!! DEBUG
   if (nu /= bdim(g)) &
      call stop_code("mlop_eom_m","u_contract","nb /= bdim(g)")
   if (blen /= product(bdim)) &
      call stop_code("mlop_eom_m","u_contract","blen /= product(bdim)")
   if (philen /= product(phidim)) &
      call stop_code("mlop_eom_m","u_contract","plen /= product(phidim)")
   !!! END DEBUG

   if (nu==1) then
      psidim(1:nm) = phidim
      psidim(nm+1) = navecs*blen
      call apply_hterm(uterms(1)%p, f, size(psi), psi, psidim, tp1)
   else
      lmask    = .true.
      lmask(g) = .false.
      blen1 = product(bdim,mask=lmask)
      psidim(1:nm) = phidim
      psidim(nm+1) = navecs
      allocate(upsi(philen,navecs,blen1))
      upsi = CZERO

!$OMP parallel default(shared) private(bb1,bidx,bg,bb)
!$OMP do schedule(static)
      do bb1=1,blen1
         call get_indices(bidx,bdim,lmask,bb1)
         do bg=1,bdim(g)
            bidx(g) = bg
            bb = flat_index(bidx,bdim)
            call apply_hterm(uterms(bg)%p, f, philen*navecs, psi(:,:,bb), psidim, tp1)
            upsi(:,:,bb1) = upsi(:,:,bb1) + psi(:,:,bb)
         enddo
      enddo
!$OMP end do
!$OMP end parallel

!     bidx = 1
!     do
!        bb  = flat_index(bidx,bdim)
!        bb1 = flat_index(bidx,bdim,lmask)
!        call apply_hterm(uterms(bidx(g))%p, f, philen*navecs, psi(:,:,bb), psidim, tp1)
!        upsi(:,:,bb1) = upsi(:,:,bb1) + psi(:,:,bb)
!        if (.not.next_point(bidx,bdim)) exit
!     enddo

      deallocate(psi)
      psi => upsi
      bdim(g) = 1
   endif
end subroutine u_contract


!-----------------------------------------------------------------------
subroutine apply_hterm(hterm,f,philen,phi,phidim,tp)
!-----------------------------------------------------------------------
   type(hterm_t),intent(in)    :: hterm       ! the operator
   integer,intent(in)          :: f           ! apply hterm to f-th dimension of phi
   integer,intent(in)          :: philen
   complex(cdbl),intent(inout) :: phi(philen) ! the complex vector
   integer,intent(in)          :: phidim(:)   ! shape of phi
   logical,intent(in),optional :: tp          ! use transpose of hterm or not
   logical                     :: tp1
   integer                     :: ijk(3)
   tp1 = .false.
   if (present(tp)) tp1=tp
   select case (hterm%htype)
      case(1)
         continue
      case(2)
         ijk = ijk_shape(f,phidim)
         if (tp1) then
            call matrix_tensor_compl_tp(hterm%matrix, phi, ijk)
         else
            call matrix_tensor_compl(hterm%matrix, phi, ijk)
         endif
      case(3)
         ijk = ijk_shape(f,phidim)
         call vector_tensor_compl(hterm%vector, phi, ijk)
      case default
         write(*,'(a,i0)') "hterm type = ",hterm%htype
         call stop_code("mlop_eom_m", "apply_hterm", "cannot handle hterm type")
   end select
end subroutine apply_hterm


!-----------------------------------------------------------------------
subroutine vw_contract(uvwpsi, vw, upsi, nn, mm1, mm2, n1, nmode, bdim, lmask)
!-----------------------------------------------------------------------
   integer,intent(in)        :: nn,mm1,mm2,n1,nmode
   complex(cdbl),intent(in)  :: vw(n1,n1,mm1*mm2)
   complex(cdbl),intent(in)  :: upsi(nn,n1,mm1)
   complex(cdbl),intent(out) :: uvwpsi(nn,n1,mm2)
   integer,intent(in)        :: bdim(nmode)
   logical,intent(in)        :: lmask(nmode)
   integer                   :: bb,bb1,bb2,l,i,jj
   integer                   :: b(nmode)
   complex(cdbl)             :: work(nn,n1)
   !!! DEBUG
   if (product(bdim,mask=lmask) /= mm1) &
      call stop_code("mlop_eom_m","apply_vw","product(bdim,mask=lmask) /= mm1")
   if (product(bdim,mask=.not.lmask) /= mm2) &
      call stop_code("mlop_eom_m","apply_vw","product(bdim,mask=.not.lmask) /= mm2")
   !!! END DEBUG
!$OMP parallel default(shared) private(bb,bb1,bb2,l,i,jj,b,work)
!$OMP do schedule(static)
   do bb2=1,mm2
      work = CZERO
      call get_indices(b,bdim,.not.lmask,bb2)
      do bb1=1,mm1
         call get_indices(b,bdim,lmask,bb1)
         bb = flat_index(b,bdim)
         do l=1,n1
            do i=1,n1
               do jj=1,nn
                  work(jj,i) = work(jj,i) + vw(i,l,bb)*upsi(jj,l,bb1)
               enddo
            enddo
         enddo
      enddo
      uvwpsi(:,:,bb2) = work
   enddo
!$OMP end do
!$OMP end parallel
end subroutine vw_contract

!-----------------------------------------------------------------------
   subroutine if_koeff(time,k,ktype,cz)
!-----------------------------------------------------------------------
      ! Return coefficient of k-term k

      real(dbl),  intent(in)   :: time
      integer,    intent(in)   :: k
      integer,    intent(out)  :: ktype
      complex(cdbl), intent(out)  :: cz
      integer                  :: h,idum
      real(dbl)                :: c,rdum(1),gpoint(maxdim)

      gpoint = 0.d0
      c = 0.d0

      h=hamilton(fkoe,k) ! Inquire type of coefficient
      ktype=htmtype(h)
      if (ktype .eq. 101) then  ! TD coefficient
         gpoint(fkoe)=1.0_dop
         gpoint(ftime)=time
         call poteval(h,c,gpoint,cz,rdum,idum)
         koeff(k) = cz
      else
         cz = koeff(k)        ! Read fixed koeff value
      end if
   end subroutine if_koeff


!=======================================================================
end module mlop_eom
