!=======================================================================
module optree
!=======================================================================
!
! This module implements the structure of a hierarchical operator (MLOp)
!
! FO 2013
!
!=======================================================================

use mlbase, only: dbl,cdbl,stop_code
use mltypes
use ilist
use tensors
use if_grid
use htermmod
implicit none
private

public :: make_opleaf, make_opnode, make_optree, &
          dispose_optree, dispose_opnode_nonrec, &
          set_opleaf_kterms, set_opnode_tt, &
          mlop_modenums, mlop_expand, &
          dump_optree_data, load_optree_data, skip_optree_data, &
          optree2dot, optree2txt


!-----------------------------------------------------------------------
! An MLOp is either recursively expressed as a linear combination of
! outer products of sub-MLOps
!
!                          p
!  ^(z)   ---    (z)   ,--. z   ^(z,f)
!  U   =  \     V      |\/|     U          ; b = 1...m
!   b     /      b;C   |/\|      c                    z
!         ---          `--'       f
!          C               f=1
!
! or, if the node z is a leaf, it's an operator acting on a primitive
! MCTDH-mode. For each b we have a different MLOp, but they only differ
! in the transfer tensor V; their sub-structure is the same. Hence we
! store the full list of MLOps at node z in the following structures.
!-----------------------------------------------------------------------


contains !==============================================================


!-----------------------------------------------------------------------
function make_opleaf(m) result(leaf)
!-----------------------------------------------------------------------
   type(opnode_t),pointer :: leaf  ! created here
   integer,intent(in)     :: m     ! MCTDH mode number
   ! Create the leaf node.
   allocate(leaf)
   leaf%isprimitive = .true.
   ! Store the MCTDH mode.
   leaf%m = m
end function make_opleaf


!-----------------------------------------------------------------------
subroutine set_opleaf_kterms(leaf,ks)
!-----------------------------------------------------------------------
   type(opnode_t),intent(inout) :: leaf
   integer,intent(in)           :: ks(:) ! list of k-terms
   ! Store the k-terms.
   leaf%nops = size(ks)
   allocate(leaf%ks(leaf%nops))
   allocate(leaf%ktypes(leaf%nops)) ! for memory locality
   leaf%ks = ks
end subroutine set_opleaf_kterms


!-----------------------------------------------------------------------
function make_opnode(children) result(node)
!-----------------------------------------------------------------------
   type(opnode_t),pointer        :: node ! created here
   type(opnode_tp),intent(inout) :: children(:)
   integer                       :: nmodes,m
   nmodes = size(children)
   ! Create non-leaf node.
   allocate(node)
   node%isprimitive = .false.
   ! Link the children into the node, and set their parent.
   node%nmodes = nmodes
   allocate(node%subops(nmodes))
   do m=1,nmodes
      node%subops(m)%p => children(m)%p
      children(m)%p%parent => node
   enddo
end function make_opnode


!-----------------------------------------------------------------------
subroutine set_opnode_tt(node,tt)
!-----------------------------------------------------------------------
   type(opnode_t),intent(inout) :: node
   type(tensor_t),intent(in)    :: tt
   ! Check if the given tensor has the right number of indices
   if (tt%ndim /= node%nmodes+1) &
      call stop_code("optree_m", "set_opnode_tt", &
         "attempt to store incompatible transfer tensor in opnode")
   ! The extent of the last index is the number of entries.
   node%nops = tt%dims(tt%ndim)
   ! Store the whole thing.
   node%tt = tt
end subroutine set_opnode_tt


!-----------------------------------------------------------------------
function make_optree(topnode) result(tree)
!-----------------------------------------------------------------------
   type(optree_t),pointer              :: tree
   type(opnode_t),target,intent(inout) :: topnode
   integer                             :: numnodes,numleaves,numlayers
   integer                             :: idx,m,f
   type(opnode_t),pointer              :: node
   allocate(tree)
   ! Set the top node.
   tree%topnode => topnode
   ! Count number of nodes and dofs in whole tree.
   call count_nodes(topnode,numnodes,numleaves,numlayers)
   tree%numnodes  = numnodes
   tree%numleaves = numleaves
   tree%numlayers = numlayers
   ! Record the pre-/post-order sequence of the nodes.
   allocate(tree%preorder(numnodes))
   idx = 1
   call set_preorder(topnode,tree%preorder,idx)
   allocate(tree%postorder(numnodes))
   idx = 1
   call set_postorder(topnode,tree%postorder,idx)
   ! Record leaves.
   allocate(tree%leaves(numleaves))
   f = 1
   do m=1,numnodes
      node => tree%preorder(m)%p
      if (node%isprimitive) then
         tree%leaves(f)%p => node
         f = f+1
      endif
   enddo
   ! Set tree-related data in nodes.
   do m=1,numnodes
      node => tree%preorder(m)%p
      node%tree => tree
   enddo
   call set_layer(topnode,1)
end function make_optree


!-----------------------------------------------------------------------
recursive subroutine count_nodes(node,numnodes,numleaves,numlayers)
!-----------------------------------------------------------------------
! Counts the total number of nodes, leaves, and layers below
! a certain node, itself included.
!-----------------------------------------------------------------------
   type(opnode_t),intent(in) :: node
   integer,intent(out)       :: numnodes,numleaves,numlayers
   type(opnode_t),pointer    :: child
   integer                   :: m,chnodes,chleaves,chlayers
   if (node%isprimitive) then
      numnodes  = 1
      numleaves = 1
      numlayers = 1
   else
      numnodes  = 1
      numleaves = 0
      numlayers = 1
      do m=1,node%nmodes
         child => node%subops(m)%p
         call count_nodes(child,chnodes,chleaves,chlayers)
         numnodes  = numnodes  + chnodes
         numleaves = numleaves + chleaves
         numlayers = max(numlayers,chlayers+1)
      enddo
   endif
end subroutine count_nodes


!-----------------------------------------------------------------------
recursive subroutine set_preorder(node,seq,idx)
!-----------------------------------------------------------------------
   type(opnode_t),target,intent(inout) :: node
   type(opnode_tp),intent(inout)       :: seq(:)
   integer,intent(inout)               :: idx
   type(opnode_t),pointer              :: child
   integer                             :: m
   ! First record this node.
   seq(idx)%p => node
   node%num = idx
   idx = idx+1
   ! Then record all children, if any.
   if (.not. node%isprimitive) then
      do m=1,node%nmodes
         child => node%subops(m)%p
         call set_preorder(child,seq,idx)
      enddo
   endif
end subroutine set_preorder


!-----------------------------------------------------------------------
recursive subroutine set_postorder(node,seq,idx)
!-----------------------------------------------------------------------
   type(opnode_t),target,intent(inout) :: node
   type(opnode_tp),intent(inout)       :: seq(:)
   integer,intent(inout)               :: idx
   type(opnode_t),pointer              :: child
   integer                            :: m
   ! First record all children, if any.
   if (.not. node%isprimitive) then
      do m=1,node%nmodes
         child => node%subops(m)%p
         call set_postorder(child,seq,idx)
      enddo
   endif
   ! Then record this node.
   seq(idx)%p => node
   node%postnum = idx
   idx = idx+1
end subroutine set_postorder


!-----------------------------------------------------------------------
recursive subroutine set_layer(node,layer)
!-----------------------------------------------------------------------
   type(opnode_t),intent(inout) :: node
   integer,intent(in)           :: layer
   type(opnode_t),pointer       :: child
   integer                      :: m
   ! Set the layer of this node.
   node%layer = layer
   ! Set the layer of all its children.
   if (.not. node%isprimitive) then
      do m=1,node%nmodes
         child => node%subops(m)%p
         call set_layer(child,layer+1)
      enddo
   endif
end subroutine set_layer


!-----------------------------------------------------------------------
recursive subroutine dispose_opnode(node)
!-----------------------------------------------------------------------
   type(opnode_t),pointer :: node
   integer                :: m
   if (node%isprimitive) then
      if (associated(node%ks)) deallocate(node%ks)
      if (associated(node%ktypes)) deallocate(node%ktypes)
   else
      do m=1,node%nmodes
         call dispose_opnode(node%subops(m)%p)
      enddo
      deallocate(node%subops)
      call dispose_tensor(node%tt)
   endif
   if (associated(node%uterms)) deallocate(node%uterms)
   if (associated(node%wterms)) deallocate(node%wterms)
   if (associated(node%vw)) deallocate(node%vw)
   if (associated(node%early_utm_child)) deallocate(node%early_utm_child)
   if (associated(node%early_wtm_sibling)) deallocate(node%early_wtm_sibling)
   deallocate(node)
end subroutine dispose_opnode


!-----------------------------------------------------------------------
subroutine dispose_opnode_nonrec(node)
!-----------------------------------------------------------------------
   type(opnode_t),pointer :: node
   if (node%isprimitive) then
      if (associated(node%ks)) deallocate(node%ks)
   else
      deallocate(node%subops)
      call dispose_tensor(node%tt)
   endif
   if (associated(node%uterms)) deallocate(node%uterms)
   if (associated(node%wterms)) deallocate(node%wterms)
   if (associated(node%vw)) deallocate(node%vw)
   if (associated(node%early_utm_child)) deallocate(node%early_utm_child)
   if (associated(node%early_wtm_sibling)) deallocate(node%early_wtm_sibling)
   deallocate(node)
end subroutine dispose_opnode_nonrec


!-----------------------------------------------------------------------
subroutine dispose_optree(tree)
!-----------------------------------------------------------------------
   type(optree_t),pointer :: tree
   deallocate(tree%preorder)
   deallocate(tree%postorder)
   deallocate(tree%leaves)
   call dispose_opnode(tree%topnode)
   deallocate(tree)
end subroutine dispose_optree


!-----------------------------------------------------------------------
function mlop_modenums(node) result(mlist)
!-----------------------------------------------------------------------
   type(opnode_t),intent(in) :: node
   type(ilist_t),pointer     :: mlist
   integer                   :: n,idx
   n = mlop_modecount(node)
   mlist => new_ilist(n)
   idx = 1
   call record_modenum(node,mlist%i,idx)
end function mlop_modenums


!-----------------------------------------------------------------------
recursive function mlop_modecount(node) result(c)
!-----------------------------------------------------------------------
   type(opnode_t),intent(in) :: node
   integer                   :: c
   integer                   :: m
   if (node%isprimitive) then
      c = 1
   else
      c = 0
      do m=1,node%nmodes
         c = c + mlop_modecount(node%subops(m)%p)
      enddo
   endif
end function mlop_modecount


!-----------------------------------------------------------------------
recursive subroutine record_modenum(node,seq,idx)
!-----------------------------------------------------------------------
   type(opnode_t),intent(in) :: node
   integer,intent(inout)     :: seq(:)
   integer,intent(inout)     :: idx
   integer                   :: m
   if (node%isprimitive) then
      seq(idx) = node%m
      idx = idx+1
   else
      do m=1,node%nmodes
         call record_modenum(node%subops(m)%p,seq,idx)
      enddo
   endif
end subroutine record_modenum


!-----------------------------------------------------------------------
function mlop_expand(node) result(flatnode)
!-----------------------------------------------------------------------
! Returns an MLOp with the same mathematical content as the given one,
! but eliminating all intermediate layers, so that all children are
! primitive. This is done by multiplying out the transfer tensors.
! NB: The resulting node will link to the same primitive nodes as the
! original one. To avoid double deallocation of these primitive nodes,
! use dispose_node_nonrec to deallocate the returned node.
!-----------------------------------------------------------------------
   type(opnode_t),pointer           :: flatnode ! created here
   type(opnode_t),intent(in),target :: node
   integer                          :: nmodes,nmodes2,nnotprim,m,m2,mm
   type(opnode_tp),pointer          :: subops(:),subops2(:)
   type(tensor_t)                   :: tt,ttmp
   type(opnode_t),pointer           :: subop

   if (node%nops /= 1) &
      call stop_code("opnode_m", "mlop_expand", "nops /= 1")
   allocate(flatnode)
   flatnode%isprimitive = .false.
   flatnode%nops = 1

   if (node%isprimitive) then
      ! Node is already primitive: insert a dummy node with the
      ! primitive node as only child.
      flatnode%nmodes = 1
      allocate(flatnode%subops(1))
      flatnode%subops(1)%p => node
      flatnode%tt = unit_tensor(2)
      return
   endif

   ! Start with the child-structure and transfer tensor of the given node.
   nmodes = node%nmodes
   allocate(subops(nmodes))
   do m=1,nmodes
      subops(m)%p => node%subops(m)%p
   enddo
   tt = clone_tensor(node%tt)

   do ! until all children are primitive

      ! Count the number of children that we would get if we flattened
      ! all non-primitive children (nmodes2), and how many of the current
      ! children are non-primitive (nnotprim).
      nnotprim = 0
      nmodes2 = 0
      do m=1,nmodes
         subop => subops(m)%p
         if (subop%isprimitive) then
            nmodes2 = nmodes2 + 1
         else
            nmodes2 = nmodes2 + subop%nmodes
            nnotprim = nnotprim + 1
         endif
      enddo
      ! If all children are already primitive, we are finished.
      if (nnotprim == 0) exit

      ! Flatten all direct non-primitive children. Keep all
      ! primitive children.
      !   subops2: resulting child-structure
      allocate(subops2(nmodes2))
      m2 = 0
      do m=1,nmodes
         subop => subops(m)%p
         if (subop%isprimitive) then
            m2 = m2 + 1
            subops2(m2)%p => subop
         else
            ! Multiply the child's transfer tensor onto the
            ! current transfer tensor.
            ttmp = tt
            tt = mul_matrix_tensor(subop%tt,ttmp,m)
            call dispose_tensor(ttmp)
            do mm=1,subop%nmodes
               m2 = m2 + 1
               subops2(m2)%p => subop%subops(mm)%p
            enddo
         endif
      enddo
      ! Fix the shape of the transfer tensor.
      tt%ndim = nmodes2 + 1
      deallocate(tt%dims)
      allocate(tt%dims(tt%ndim))
      do m2=1,nmodes2
         tt%dims(m2) = subops2(m2)%p%nops
      enddo
      tt%dims(nmodes2+1) = 1
      ! Check for consistency.
      if (product(tt%dims) /= tt%vlen) &
         call stop_code("opnode_m","mlop_expand","internal error: product(dims) /= vlen")

      ! Prepare data for next round.
      nmodes = nmodes2
      deallocate(subops)
      subops => subops2
   enddo

   ! Store the final data.
   flatnode%nmodes = nmodes
   flatnode%subops => subops
   flatnode%tt = tt

end function mlop_expand


!-----------------------------------------------------------------------
subroutine optree2dot(lun,t,edgectrl)
!-----------------------------------------------------------------------
   integer,intent(in)          :: lun
   type(optree_t),intent(in)   :: t
   integer,intent(in),optional :: edgectrl
   type(opnode_t),pointer      :: cnode
   integer                     :: ectl,i
   ectl = 0
   if (present(edgectrl)) ectl=edgectrl
   ! output header
   write(lun,'(5(a,/),a)') &
      'digraph G {', &
      '// customize the following three lines to your liking', &
      'graph [layout=dot rankdir=BT nodesep=0.1 ranksep=0.3]', &
      'node  [fontname=Arial width=0.25 height=0.25 margin=0.05]', &
      'edge  [fontname=Arial]', &
      '// the rest describes the ML tree'
   ! loop over nodes
   do i=1,t%numnodes
      cnode => t%preorder(i)%p
      ! output this node (label by number)
      write (lun,'(a,i0,a,i0,a)') &
         'n', cnode%num, &
         ' [shape=circle label="', cnode%num, &
         '" margin=0]'
      if (cnode%isprimitive) then
         ! output mode
         write (lun,'(a,i0,a,i0,a)') 'f', cnode%m, &
            ' [shape=box,label="', cnode%m, '"]'
         ! output edge
         write (lun,'(a,i0,a,i0,a)') &
            'f', cnode%m, ' -> n', cnode%num, ' [dir=none]'
      endif
      ! output edge to parent
      if (associated(cnode%parent)) then
         if (ectl==0) then
            write (lun,'(a,i0,a,i0,a)') &
               'n', cnode%num, ' -> n', cnode%parent%num, ' [dir=none]'
         else
            write (lun,'(a,i0,a,i0,a,i0,a)') &
               'n', cnode%num, ' -> n', cnode%parent%num, &
               ' [dir=none,label=', cnode%nops, ']'
         endif
      endif
   end do
   ! output footer
   write(lun,'(a)') '}'
end subroutine optree2dot


!-----------------------------------------------------------------------
subroutine optree2txt(lun,t,head)
!-----------------------------------------------------------------------
   integer,intent(in)          :: lun
   type(optree_t),pointer      :: t
   character(len=*),intent(in) :: head
   character(len=80)           :: empty
   empty = ""
   call opnode2txt(lun, t%topnode, head, empty(1:len(head)))
end subroutine optree2txt


!-----------------------------------------------------------------------
recursive subroutine opnode2txt(lun,no,myhead,chldhead)
!-----------------------------------------------------------------------
   integer,intent(in)          :: lun
   type(opnode_t),pointer      :: no
   character(len=*),intent(in) :: myhead,chldhead
   integer                     :: m
   type(opnode_t),pointer      :: chld
   write (lun,'(a,i0)',advance="no") myhead,no%num
   if (no%isprimitive) then
      write (lun,'(2x,a,i0)') "["//trim(if_modestring(no%m))//"] * ",no%nops
   else
      write (lun,'(2x,a,i0)')  "* ",no%nops
      do m=1,no%nmodes
         chld => no%subops(m)%p
         if (m < no%nmodes) then
            call opnode2txt(lun, chld, chldhead//"├──", chldhead//"│  ")
         else
            call opnode2txt(lun, chld, chldhead//"└──", chldhead//"   ")
         endif
      enddo
   endif
end subroutine opnode2txt


!-----------------------------------------------------------------------
subroutine dump_optree_data(op,lun)
!-----------------------------------------------------------------------
   type(optree_t),intent(in) :: op
   integer,intent(in)        :: lun
   type(opnode_t),pointer    :: no
   integer                   :: m,i
   do m=1,op%numnodes
      no => op%postorder(m)%p
      write(lun) no%nops
      if (no%isprimitive) then
         write(lun) (no%ks(i), i=1,no%nops)
      else
         call dump_tensor(no%tt,lun)
      endif
   enddo
end subroutine dump_optree_data


!-----------------------------------------------------------------------
subroutine skip_optree_data(op,lun)
!-----------------------------------------------------------------------
   type(optree_t),intent(in) :: op
   integer,intent(in)        :: lun
   type(opnode_t),pointer    :: no
   integer                   :: m
   do m=1,op%numnodes
      no => op%postorder(m)%p
      read(lun) ! nops
      if (no%isprimitive) then
         read(lun) ! ks
      else
         call skip_tensor(lun) ! tt
      endif
   enddo
end subroutine skip_optree_data


!-----------------------------------------------------------------------
subroutine load_optree_data(op,lun)
!-----------------------------------------------------------------------
   type(optree_t),intent(in) :: op
   integer,intent(in)        :: lun
   type(opnode_t),pointer    :: no
   integer                   :: m,i
   logical                   :: lerr
   do m=1,op%numnodes
      no => op%postorder(m)%p
      read(lun) no%nops
      if (no%isprimitive) then
         allocate(no%ks(no%nops))
         allocate(no%ktypes(no%nops))
         read(lun) (no%ks(i), i=1,no%nops)
      else
         call load_tensor(no%tt,lun)
         ! Check consistency.
         lerr = .false.
         if (no%tt%ndim /= no%nmodes+1) then
            write(*,'(a,i0)') "tt%ndim        = ",no%tt%ndim
            write(*,'(a,i0)') "nmodes         = ",no%nmodes
            lerr = .true.
         endif
         if (no%tt%dims(no%nmodes+1) /= no%nops) then
            write(*,'(a,i0,a,i0)') "tt%dims(",no%nmodes+1,")     = ",no%tt%dims(no%nmodes+1)
            write(*,'(a,i0)') "nops           = ",no%nops
            lerr = .true.
         endif
         do i=1,no%nmodes
            if (no%tt%dims(i) /= no%subops(i)%p%nops) then
               write(*,'(a,i0)') "tt%dims(",i,")     = ",no%tt%dims(i)
               write(*,'(a,i0)') "subops(",i,")%nops = ",no%subops(i)%p%nops
               lerr = .true.
            endif
         enddo
         if (lerr) then
            write(*,'(a,i0)') "opnode%num     = ",no%num
            call stop_code("optree_m", "load_optree_data", &
               "MLOp read from file inconsistent with system structure")
         endif
      endif
   enddo
end subroutine load_optree_data


end module optree
! vim: set ts=3 sw=3 expandtab:
