!=======================================================================
module tensors
!=======================================================================

use mltypes, only: tensor_t, tensor_tp
use gridutil
use mlbase
implicit none
private

public :: unit_tensor, krondelta_tensor, clone_tensor, dispose_tensor, &
          tensor_extend, mul_matrix_tensor, compare_tensors, &
          dump_tensor, load_tensor, skip_tensor


contains !==============================================================


!-----------------------------------------------------------------------
function unit_tensor(ndim) result(u)
!-----------------------------------------------------------------------
   integer,intent(in) :: ndim
   type(tensor_t)     :: u
   u%ndim = ndim
   allocate(u%dims(ndim))
   u%dims = 1
   u%vlen = 1
   allocate(u%vr(1))
   u%vr = 1.d0
end function unit_tensor


!-----------------------------------------------------------------------
function krondelta_tensor(sz) result(t)
!-----------------------------------------------------------------------
   integer,intent(in) :: sz
   type(tensor_t)     :: t
   integer            :: n,i(2),j
   t%ndim = 2
   allocate(t%dims(2))
   t%dims = sz
   t%vlen = sz**2
   allocate(t%vr(t%vlen))
   t%vr = 0.d0
   do n=1,sz
      i = [n,n]
      j = flat_index(i,[sz,sz])
      t%vr(j) = 1.d0
   enddo
end function krondelta_tensor


!-----------------------------------------------------------------------
function clone_tensor(t) result(ct)
!-----------------------------------------------------------------------
   type(tensor_t),intent(in) :: t
   type(tensor_t)            :: ct
   ct%ndim = t%ndim
   allocate(ct%dims(t%ndim))
   ct%dims = t%dims
   ct%vlen = t%vlen
   allocate(ct%vr(t%vlen))
   ct%vr = t%vr
end function clone_tensor


!-----------------------------------------------------------------------
subroutine dispose_tensor(t)
!-----------------------------------------------------------------------
   type(tensor_t),intent(inout) :: t
   if (associated(t%dims)) deallocate(t%dims)
   if (associated(t%vr))   deallocate(t%vr)
end subroutine dispose_tensor


!-----------------------------------------------------------------------
function tensor_extend(t1,t2,map) result(t)
!-----------------------------------------------------------------------
! Extends tensor t1 by multiplying t2 onto it, where dimension d of t2
! corresponds to dimension map(d) of tensor t1.  Result is tensor t.
! Special cases:
! - If t2 is unit_tensor, then t is a copy of t1.
! - If t1 is unit_tensor, then t is a reshaped version of t2.
!-----------------------------------------------------------------------
   type(tensor_t)             :: t
   type(tensor_t),intent(in)  :: t1,t2
   integer,intent(in)         :: map(:)
   integer,parameter          :: maxndimdim = 8
   integer,allocatable        :: ndimdim(:),dimdims(:,:),map2(:)
   integer                    :: d,d1,d2,j,j1,j2
   integer,allocatable        :: i(:),i1(:),i2(:),ii(:,:)

   ! The resulting tensor t has as many indices as t1. But for t, each
   ! index is actually a multi-index. Each multi-index has one index
   ! from t1 and zero or more indices from t2.
   !   ndimdim(d)  : length of multi-index d
   !   dimdims(k,d): extent of k-th index of multi-index d
   !   d=map(d2)   : index d2 of t2 maps to the k-th index of...
   !   k=map2(d2)  : ...the multi-index d
   allocate(ndimdim(t1%ndim))
   allocate(dimdims(maxndimdim,t1%ndim))
   allocate(map2(t2%ndim))
   dimdims = 0
   ! index d1 of t1 becomes the 1st index of multi-index d1
   do d1=1,t1%ndim
      ndimdim(d1) = 1
      dimdims(1,d1) = t1%dims(d1)
   enddo
   ! index d2 of t2 becomes the next index of the multi-index it maps to
   do d2=1,t2%ndim
      d1 = map(d2)
      ndimdim(d1) = ndimdim(d1) + 1
      dimdims(ndimdim(d1),d1) = t2%dims(d2)
      map2(d2) = ndimdim(d1)
   enddo

   ! In the end, all the multi-indices will be flattened. This determines
   ! the shape of tensor t. Set this now.
   t%ndim = size(ndimdim)
   allocate(t%dims(t%ndim))
   do d=1,t%ndim
      t%dims(d) = product(dimdims(1:ndimdim(d),d))
   enddo
   ! Now we can allocate the tensor data.
   t%vlen = product(t%dims)
   allocate(t%vr(t%vlen))

   ! Now we can build the tensor data.
   !   i1(d1) : index d1 (of t1)
   !   i2(d2) : index d2 (of t2)
   !   ii(k,d): k-th index of multi-index d (of t)
   !   i(d)   : flattened multi-index d (of t)
   allocate(i1(t1%ndim))
   allocate(i2(t2%ndim))
   allocate(ii(maxndimdim,t%ndim))
   allocate(i(t%ndim))

   ! Loop over the indices of t2.
   i2 = 1
   do
      ! Loop over the indices of t1.
      i1 = 1
      do
         ! Map the indices of t1 and t2 to the corresponding multi-index.
         do d1=1,t1%ndim
            ii(1,d1) = i1(d1)
         enddo
         do d2=1,t2%ndim
            ii(map2(d2),map(d2)) = i2(d2)
         enddo
         ! Flatten all multi-indices.
         do d=1,t%ndim
            i(d) = flat_index(ii(1:ndimdim(d),d), dimdims(1:ndimdim(d),d))
         enddo
         ! Get the (flat) offsets into the tensor data.
         j  = flat_index(i,t%dims)
         j1 = flat_index(i1,t1%dims)
         j2 = flat_index(i2,t2%dims)
         ! Multiply the elements of t1 and t2 to get the element of t.
         t%vr(j) = t1%vr(j1) * t2%vr(j2)
         ! Advance to next element of t1.
         if (.not.next_point(i1,t1%dims)) exit
      enddo
      ! Advance to next element of t2.
      if (.not.next_point(i2,t2%dims)) exit
   enddo

   ! Clean up.
   deallocate(i)
   deallocate(ii)
   deallocate(i2)
   deallocate(i1)
   deallocate(map2)
   deallocate(dimdims)
   deallocate(ndimdim)

end function tensor_extend


!-----------------------------------------------------------------------
function mul_matrix_tensor(s,t,m) result(st)
!-----------------------------------------------------------------------
! Interpret tensor _s_ as a rectangular matrix (using its last
! dimension for columns) and multiply it onto the m-th dimension of
! tensor _t_.
! st(A,b,C) = sum_b' s(b,b')*t(A,b',C)
!-----------------------------------------------------------------------
   type(tensor_t)            :: st  ! created here
   type(tensor_t),intent(in) :: s,t
   integer,intent(in)        :: m
   integer                   :: d1,d2,d3,dd
   d1 = product(t%dims(1:m-1))
   d2 = t%dims(m)
   d3 = product(t%dims(m+1:t%ndim))
   if (d2 /= s%dims(s%ndim)) &
      call stop_code('tensor_m','mul_matrix_tensor','size mismatch')
   dd = product(s%dims(1:s%ndim-1))
   st%ndim = t%ndim
   allocate(st%dims(t%ndim))
   st%dims = t%dims
   st%dims(m) = dd
   st%vlen = product(st%dims)
   allocate(st%vr(st%vlen))
   call mul_mat_t3d_rr(s%vr,t%vr,st%vr)
   return

   contains

   subroutine mul_mat_t3d_rr(x,y,xy)
      real(dbl) :: x(dd,d2)
      real(dbl) :: y(d1,d2,d3)
      real(dbl) :: xy(d1,dd,d3)
      integer   :: i1,i2,i3,ii
      xy = 0.d0
      do i3=1,d3
         do i2=1,d2
            do ii=1,dd
               do i1=1,d1
                  xy(i1,ii,i3) = xy(i1,ii,i3) + x(ii,i2)*y(i1,i2,i3)
               enddo
            enddo
         enddo
      enddo
   end subroutine mul_mat_t3d_rr

end function mul_matrix_tensor


!-----------------------------------------------------------------------
subroutine compare_tensors(t1,t2,emax,eavg,norm)
!-----------------------------------------------------------------------
   type(tensor_t),intent(in) :: t1,t2
   real(dbl),intent(out)     :: emax,eavg,norm
   integer                   :: i
   real(dbl)                 :: e
   if (t1%vlen /= t2%vlen) then
      print *, 't1%vlen = ', t1%vlen
      print *, 't2%vlen = ', t2%vlen
      print *, 't1%dims = ', t1%dims
      print *, 't2%dims = ', t2%dims
      call stop_code("tensor_m","compare_tensors","tensor sizes don't match")
   endif
   emax = 0.d0
   eavg = 0.d0
   norm = 0.d0
   do i=1,t1%vlen
      e = abs(t1%vr(i) - t2%vr(i))
      norm = norm + t1%vr(i)**2
      if (e > emax) emax = e
      eavg = eavg + e**2
   enddo
   eavg = sqrt(eavg / t1%vlen)
   norm = sqrt(norm / t1%vlen)
end subroutine compare_tensors


!-----------------------------------------------------------------------
subroutine dump_tensor(t,lun)
!-----------------------------------------------------------------------
   type(tensor_t),intent(in) :: t
   integer,intent(in)        :: lun
   integer                   :: i
   write(lun) t%ndim
   write(lun) (t%dims(i), i=1,t%ndim)
   write(lun) (t%vr(i), i=1,t%vlen)
end subroutine dump_tensor


!-----------------------------------------------------------------------
subroutine skip_tensor(lun)
!-----------------------------------------------------------------------
   integer,intent(in) :: lun
   read(lun) ! ndim
   read(lun) ! dims
   read(lun) ! vr
end subroutine skip_tensor


!-----------------------------------------------------------------------
subroutine load_tensor(t,lun)
!-----------------------------------------------------------------------
   type(tensor_t),intent(inout) :: t
   integer,intent(in)           :: lun
   integer                      :: i
   read(lun) t%ndim
   allocate(t%dims(t%ndim))
   read(lun) (t%dims(i), i=1,t%ndim)
   t%vlen = product(t%dims)
   allocate(t%vr(t%vlen))
   read(lun) (t%vr(i), i=1,t%vlen)
end subroutine load_tensor


!=======================================================================
end module tensors
