module mlpostproc


  use global
  use globmemmod
  use mlbase
  use mltypes
  use timing

  use griddatmod
  use psidef
  use runpropmod, only: currentnode
  use gwplib, only: normgwp,renormdgwp1
  use importhomod, only: importho1

private

  public  :: int_postprocess,int_postprocess_show,mlimportho,mlrenorma 

contains

  subroutine int_postprocess(y, ny, ipost)
  
    use mlbase
    
    implicit none
    complex(cdbl) :: y(ny)  ! vector of dynamic variables
    integer       :: ny     ! size of this vector
    integer       :: ipost  ! parameter for choosing what to do
    
    select case (ipost)
    case (1)
       ! improved-orthonormality for ML-CMF=split
       ! y = SPFs for one node
       call mlimportho(y,ny)
    case (2)
       ! renormalize A-vector for relaxation
       ! y = (single) SPF for top node
       call mlrenorma(y,ny)
    case default
       call stop_code('(none)', 'int_postprocess', 'unknown postprocessing parameter')
    end select
    return

  end subroutine int_postprocess


  subroutine int_postprocess_show(ipost, ppdesc)

    implicit none
    integer,intent(in)           :: ipost
    character(len=*),intent(out) :: ppdesc
    
    select case (ipost)
    case (1)
       ppdesc = "with improved orthonormality"
    case (2)
       ppdesc = "with renormalization"
    case default
       call stop_code('(none)', 'int_postprocess_show', 'unknown postprocessing parameter')
    end select
    return
    
  end subroutine int_postprocess_show


  subroutine mlimportho(phi,phidim)
    implicit none
    integer,intent(in)          :: phidim
    complex(cdbl),intent(inout) :: phi(phidim)
    integer                     :: nspf,ndim,m,pointer
    complex(cdbl),allocatable   :: work(:)
    integer,save                :: tid = 0
    call get_timer(tid,'mlimportho')
    call start_timer(tid)
    m = currentnode%m
    nspf = currentnode%navecs
    ndim = product(currentnode%ndim)
    if (currentnode%isprimitive .and. gwpm(m)) then
       pointer  = zdgwp(m,1)-zetf(m,1)+1

       call renormdgwp1(phi(pointer),gs2(gs2mat(m,1)),dicht4(dmat(m,1)),nspf)

       call normgwp(nspfdof(m),nspf,agmat(zagmat(m,1)),agmat(zxivec(m,1)),&
            agmat(zphase(m,1)))
    else
       ! check consistency
       if (nspf*ndim /= phidim) then
          write (*, '(a,/,4(a,i0,3x))') &
               '*** INTERNAL ERROR DETECTED (mlimportho) ***', &
               'currentnode = ',currentnode%num, &
               'nspf = ',nspf,'ndim = ',ndim,'phidim = ',phidim
          stop 1
       endif
       ! improve orthonormality of vectors
       allocate(work(nspf*(nspf+1)))
       call importho1(ndim,nspf,phi)
       deallocate(work)
    endif
    call stop_timer(tid)
  end subroutine mlimportho


  subroutine mlrenorma(phi,phidim)
    implicit none
    integer,intent(in)          :: phidim
    complex(cdbl),intent(inout) :: phi(phidim)
    integer                     :: nspf,ndim
    real(dbl)                   :: invnorm
    integer,save                :: tid = 0
    call get_timer(tid,'mlrenorma')
    call start_timer(tid)
    nspf = currentnode%navecs
    ndim = product(currentnode%ndim)
    ! check consistency
    if (.not.currentnode%istop .or. nspf /= 1 .or. ndim /= phidim) then
       write (*, '(a,/,2(a,i0,3x))') &
            '*** INTERNAL ERROR DETECTED (mlrenorma) ***', &
            'currentnode = ',currentnode%num, &
            'nspf = ',nspf,'ndim = ',ndim,'phidim = ',phidim
       stop 1
    endif
    ! renormalize A-vector
    invnorm = ONER/sqrt(dble(dot_product(phi,phi)))
    phi = invnorm*phi
    call stop_timer(tid)
  end subroutine mlrenorma

end module mlpostproc
