!#############################################################################
!                 --- Heidelberg MCTDH ---                                   #
!                     Module: multilayer_m                                   #
!                     Author: Oriol Vendrell (oriol.vendrell@gmail.com)      #
!                     Year  : 2008                                           #
!#############################################################################

! Driver on top of the ML machinery. Together with the files interface.F and
! mlservices.F, it provides the necessary infrastructure for the interaction of
! the Heidelberg MCTDH code with the ML modules.

    module crossov_m
!.....This module is used only by the subroutines setoverlap and mloverlap

      use mltypes

      implicit none
      logical                :: ldiffer
      integer                :: cross_dgldim
      type(wfnode_t),pointer :: externtree
      type(wfnode_t),pointer :: crosstree
    end module crossov_m

module multilayer

    use decimal
    use global
    use globmemmod
    use wftree
    use wfprop
    use wftrafo
    use mlanalysis
    use mlop_eom
    use integrator
    use timing
    use if_grid
    use ctensor
    use htucker
    use evaluate, only: getprefactor

    use mltypes
    use gridutil
    use linear
    use eqofmotion1, only: compute_dmat_tree

    use inputmlmod
    use griddatmod, only: tree,tz,treeaux1
    use operdef, only: hamilton,htmtype
    use psidef
    use psiutil, only: spfovl
    use runpropmod, only: currentnode

    implicit none

    private

    ! Public module methods
    public  :: setoperml,      &
               newoperml,      &
               setpropml,      &
               setoverlap,     &
               mloverlap,      &
               automl,         &
               normml,         &
               traceml,        &
               ztraceml,       &
               outputml,       &
               logintml,       &
               outlnp,         &
               mkdotml,        &
               copywfnml,      &
               wim_ml,         &
               errortotml,     &
               reorthoml,      &
               get_moduletree,   &
               set_moduletree,   &
               swap_moduletree,  &
               swap_moduleauxtrees, &
               allocate_auxwftreearray, &
               swap_moduleauxtreesarray

   ! Private module variables
    type(wfnode_tp),pointer     :: treeauxl(:) => null() ! list to pointers of auxiliary trees

    contains

    subroutine setoperml(nham,lfields,useunit)

        implicit none
        integer(long), intent(in)    :: nham
        logical, intent(in)          :: lfields,useunit

        if(.not.ml_input_ready) then
            write(6,*) 'startml: no tree has been set'
            stop
        end if

        if (nham .gt. 0) then
           ! Allocate space for h-terms, mean fields
           call allocate_wftree_h(tree,lfields,useunit)
           ! MLOps need additional allocations.
           call alloc_all_mlop_tree(tree,nham)
        endif

        ! Multilayer is ready to operate
        ml_oper_ready = .true.
        if(ml_input_ready.and.&
           ml_oper_ready.and.&
           ml_prop_ready) then
           ml_ready=.true.
        end if

    end subroutine setoperml

    subroutine newoperml(newtree,nham,lf,useunit)
        implicit none
        integer,intent(in) :: nham    ! Hamiltonian number
        logical,intent(in) :: lf      ! whether to allocate space for mfields etc.
        logical,intent(in) :: useunit ! whether to use shortcut for unit operators
        type(wfnode_t)     :: newtree

        if (nham .gt. 0) then
           ! Allocate space for h-terms, mean fields
           call allocate_wftree_h(newtree,lf,useunit)
           ! MLOps need additional allocations.
           !call alloc_all_mlop_tree(newtree,nham)
           newtree%isallocated = .true.
        endif
    end subroutine newoperml


    subroutine setpropml(nham)
        implicit none

        integer,intent(in) :: nham

        if(.not.ml_input_ready) then
            write(6,*) 'startml: no tree has been set'
            stop
        end if

        ! Allocate space needed only in propagation
        call alloc_prop_wftree(tree,.true.)
        ! Now that cmftyp has been set, revisit MLOp allocation.
        call alloc_all_mlop_tree(tree,nham)

        call set_prefactor(tree,getprefactor())        ! get prefactor
        call set_isrelax(tree,lrlx)          ! get computation type (prop/rlx)
        call set_isrlxt(tree,lrlxt)    ! get computation type 
                                                 !(lrlxt, thermal)
        if (tree%cmftyp == 2) then
           call init_integrators(tree)
        endif

        ! Multilayer is ready to operate
        ml_prop_ready = .true.
        if(ml_input_ready.and.&
           ml_oper_ready.and.&
           ml_prop_ready) then
           ml_ready=.true.
        end if
    end subroutine setpropml

    subroutine setoverlap(tapecross,ldiff,dgldim1)
        use crossov_m
        implicit none
        integer,intent(in)          :: tapecross(maxtape)
        integer,intent(in)          :: dgldim1
        logical,intent(in)          :: ldiff

        call get_moduletree(externtree)
        ldiffer=ldiff
        if(ldiffer) then
           crosstree  => new_wftree(tapecross)
           cross_dgldim=dgldim1
        else
           cross_dgldim=dgldim
        endif
    end subroutine setoverlap

    subroutine mloverlap(psicross,psi,ovl,ovlmat)
        use crossov_m
        implicit none
        complex(cdbl),intent(in)  :: psicross(cross_dgldim),psi(dgldim)
        complex(cdbl),intent(in)  :: ovlmat(dmatdim)
        complex(cdbl),intent(out) :: ovl

        if(ldiffer) then
           call overlap_2(psicross,psi,crosstree,externtree,ovl,ovlmat)
        else
           call overlap_1(psicross,psi,externtree,ovl,ovlmat)
        endif
    end subroutine mloverlap

    subroutine normml(psi,norm2,ovlmat)
        implicit none
        complex(cdbl),intent(in)     :: psi(dgldim)
        complex(cdbl),intent(in)     :: ovlmat(dmatdim)
        complex(dbl),intent(out)     :: norm2
        call overlap_1(psi,psi,tree,norm2,ovlmat)
    end subroutine normml

    subroutine traceml(psi,trace)
        implicit none
        integer                       :: i
        complex(cdbl),intent(in)      :: psi(adim)
        real(dbl),intent(out)         :: trace
        complex(cdbl)                 :: work(adim)

        work(1:adim) = psi(1:adim)
        call hprod_phi(tree%mtraces,work)

        trace = 0.0_dop
        do i = 1,adim
           trace = trace + dble(work(i))
        enddo

    end subroutine traceml

    subroutine ztraceml(psi,trace)
        implicit none
        integer                       :: i
        complex(cdbl),intent(in)      :: psi(adim)
        complex(cdbl),intent(out)     :: trace
        complex(cdbl)                 :: work(adim)

        work(1:adim) = psi(1:adim)
        call hprod_phi(tree%mtraces,work)

        trace = 0.0_dop
        do i = 1,adim
           trace = trace + work(i)
        enddo

    end subroutine ztraceml


    subroutine automl(auto,psi1,psi2,ovlmat)
        implicit none
        complex(cdbl),intent(out)    :: auto
        complex(cdbl),intent(in)     :: psi1(dgldim)
        complex(cdbl),intent(in)     :: psi2(dgldim)
        complex(cdbl),intent(in)     :: ovlmat(dmatdim)
        call overlap_1(psi1,psi2,tree,auto,ovlmat,1)
    end subroutine automl


    subroutine errortotml(err,psi1,psi2,agmat1,agmat2)
        use timing
        implicit none
        real(dbl), intent(out)       :: err
        complex(cdbl), intent(in)    :: psi1(dgldim)
        complex(cdbl), intent(in)    :: psi2(dgldim)
        complex(cdbl), intent(in)    :: agmat1(agmatdim)
        complex(cdbl), intent(in)    :: agmat2(agmatdim)
        complex(cdbl)                :: ovlmat(dmatdim)
        complex(cdbl)                :: p2,q2,pq
        integer, save                :: tid = 0 ! timer ID
        call get_timer(tid, 'errortotml')
        call start_timer(tid)
        call spfovl(ovlmat,psi1,psi1,agmat1,agmat1,0)
        call overlap_1(psi1,psi1,tree,p2,ovlmat)
        call spfovl(ovlmat,psi2,psi2,agmat2,agmat2,0)
        call overlap_1(psi2,psi2,tree,q2,ovlmat)
        call spfovl(ovlmat,psi1,psi2,agmat1,agmat2,0)
        call overlap_1(psi1,psi2,tree,pq,ovlmat)
        err = dble(p2 + q2 - 2.d0*pq)
        ! sanity check necessary due to numerical noise
        if (.not. lgwp .and. err < 0.d0) then
            if (abs(err) > 1.d-12) then
                call stop_code('multilayer','errortotml','non-real error')
            else
                err = 0.d0
            endif
        else if (lgwp .and. err < 0.d0) then
            err = 0.d0
        endif
        call stop_timer(tid)
    end subroutine errortotml

    subroutine outputml(iout)
        implicit none
        integer :: iout
        call outpoptree(tree,iout)
    end subroutine outputml


    subroutine logintml(ilog)
       implicit none
       integer,intent(in)     :: ilog
       type(wfnode_t),pointer :: no
       integer                :: i
       character(len=240)     :: desc
       write (ilog, '(/,a,/)') '***** Split Propagation: Integrator Details *****'
       do i=1,tree%tot_nodes
          no => tree%swap_down(i)%p
          call show_integrator(no%integrator,desc)
          write (ilog, '(a,i3,2a)') 'Node ',no%num,': ',trim(desc)
       enddo
       write (ilog, *)
    end subroutine logintml


    subroutine outlnp(ich,time,lhead)
        implicit none
        integer,intent(in)   :: ich
        real(dbl),intent(in) :: time
        logical,intent(in)   :: lhead
        call outlowpop(tree,ich,time,lhead)
    end subroutine outlnp


!#######################################################################

    subroutine mkdotml(idot)
        ! creates an input file for graphviz, for visualising the ML tree
        implicit none
        integer,intent(in)      :: idot      ! output channel
        type(wfnode_t),pointer  :: cnode
        integer                 :: i,f
        character(len=16)       :: label
        ! output header
        write(idot,'(5(a,/),a)') &
            'digraph G {', &
            '// customize the following three lines to your liking', &
            'graph [layout=dot rankdir=BT nodesep=0.1 ranksep=0.3]', &
            'node  [fontname=Arial width=0.25 height=0.25 margin=0.05]', &
            'edge  [fontname=Arial]', &
            '// the rest describes the ML tree'
        ! loop over nodes
        do i=1,tree%tot_nodes
            cnode => tree%swap_down(i)%p
            ! output this node (label by number)
            write (idot,'(a,i0,a,i0,a)') &
               'n', cnode%num, &
               ' [shape=circle label="', cnode%num, &
               '" margin=0]'
            if (cnode%isprimitive) then
                ! loop over primitive subnodes
                do f=1,cnode%nmodes
                    ! output subnode
                    call if_modelabel(cnode%degf(f),label)
                    write (idot,'(a,i0,3a)') 'f', cnode%degf(f), &
                        ' [shape=box,label="', trim(label), '"]'
                    ! output edge
                    write (idot,'(a,i0,a,i0,a,i0,a)') &
                        'f', cnode%degf(f), ' -> n', cnode%num, &
                        ' [dir=none,label=', cnode%ndim(f), ']'
                end do
            endif
            ! output edge to parent
            if (.not.cnode%istop) then
                write (idot,'(a,i0,a,i0,a,i0,a)') &
                    'n', cnode%num, ' -> n', cnode%parent%num, &
                    ' [dir=none,label=', cnode%navecs, ']'
            endif
        end do
        ! output footer
        write(idot,'(a)') '}'
    end subroutine mkdotml

!#######################################################################

    subroutine wim_ml(iout)
       implicit none
       integer,intent(in)     :: iout
       type(wfnode_t),pointer :: no
       integer                :: i
       do i=1,tree%tot_nodes
          no => tree%swap_down(i)%p
          write (iout,'(a,i0,a)',advance='no') 'Node ',no%num,': '
          call stats_integrator(no%integrator,iout)
       enddo
       write (iout,*)
    end subroutine wim_ml


    subroutine copywfnml(oldpsi,oldtape,psi)
        ! Only the various layers of coefficients are copied by this
        ! routine, SPF's at the lowest layers are not transferred.
      use maxv, only: maxtape
        implicit none
        complex(cdbl)    :: oldpsi(*)
        integer          :: oldtape(maxtape)
        complex(cdbl)    :: psi(dgldim)

        !call compare_tapes(tree%tape,oldtape)

        treeaux1 => new_wftree(oldtape)
        call copy_wftree(oldpsi,treeaux1,psi,tree)
    end subroutine copywfnml

    subroutine reorthoml(psi)
       implicit none
       complex(cdbl),intent(inout) :: psi(dgldim)
       call compute_dmat_tree(tree,psi)  ! Dieter 29.03.20
       call natorb_tree(tree,psi)        ! Dieter 29.03.20
! Careful - reortho_tree does a radical reorthonormalisation rather than
! the simple Gram-Schmidt
       call reortho_tree(tree,psi)
    end subroutine reorthoml

!-----------------------------------------------------------------------------

    subroutine get_moduletree(externtree)
      implicit none
      type(wfnode_t), pointer, intent(INOUT) :: externtree
      externtree => tree
    end subroutine get_moduletree

    subroutine set_moduletree(externtree)
      implicit none
      type(wfnode_t), pointer, intent(INOUT) :: externtree
      tree => externtree
    end subroutine set_moduletree

    subroutine swap_moduletree(externtree)
      implicit none
      type(wfnode_t), pointer, intent(INOUT) :: externtree
      type(wfnode_t), pointer :: tmptree
      tmptree  => externtree
      externtree => tree
      tree  => tmptree
    end subroutine swap_moduletree

    subroutine swap_moduleauxtrees()
      implicit none
      type(wfnode_t), pointer :: tmptree
      tmptree  =>  treeaux1
      treeaux1 => tree
      tree  => tmptree
    end subroutine swap_moduleauxtrees

    subroutine allocate_auxwftreearray(ntree)
      implicit none
      integer, intent(IN) :: ntree
      integer             :: stat
      integer,save        :: ntreeold=0

      if(ntree.le.0) then
         print*, 'allocate_auxwftreearray: ntree must be .gt. 0 ,  STOP'
         stop 1
      elseif(ntreeold.eq.ntree) then
         return
      elseif(ntreeold.gt.0) then
         print*, 'treeauxl(ntree) is allready allocated, but with a different ntree.'
         print*, 'ntree =',ntree,', ntreeold =',ntreeold, '  STOP'
         stop 1
      else
         allocate(treeauxl(ntree), stat=stat)
         ntreeold=ntree
      endif

      if (stat .ne. 0) then
         print*, "allocating treeauxl(ntree) failed"
         stop 1
      end if
    end subroutine allocate_auxwftreearray

    subroutine swap_moduleauxtreesarray(n)
      implicit none
      integer, intent(in)     :: n
      type(wfnode_t), pointer :: tmptree

      if (.not. associated(treeauxl)) then
         print*, "treeauxl(ntree) not allocated"
         stop 1
      end if
      if (size(treeauxl) .lt. n) then
         print*, "treeauxl(ntree) not long enough"
         stop 1
      end if

      tmptree  =>  treeauxl(n)%p
      treeauxl(n)%p => tree
      tree  => tmptree
    end subroutine swap_moduleauxtreesarray

end module multilayer
