program cutcpd
  use cutcpd_opts_m
  use dvrtype_m
  use eximporttools
  use cpdtype_m
  use cpdtype_s
  use channels, only: ilog
  use mlbase
  use pftools
  implicit none

  type(ccopts_t)       :: opts
  type(dvr_t)          :: dvr_in, dvr_out
  type(cpd_t)          :: cpd_in, cpd_out
  integer, allocatable :: frem(:), refidx(:)


  call default

  ! read input, make name and opern log
  call cc_inp(opts,ilog)

  ! load input data
  call loaddvr(dvr_in, opts%dvrfile_in)
  call importMCTDH(cpd_in, dvr_in, opts%cpdfile_in)

  ! create a list of reference points form input
  allocate(frem(opts%nref))
  allocate(refidx(opts%nref))
  call mkremlist(dvr_in,opts,frem,refidx,ilog)

  ! remove some grids from the DVR
  call copy_dvr_t(dvr_in, dvr_out)
  call remove_dof_dvr_t(dvr_out, frem, opts%nref)
  call savedvr(dvr_out,opts%dvrfile_out,opts%lovwr)

  ! cut the cpd file and save it
  call cc_cutcpd(cpd_in, cpd_out, dvr_out, frem, refidx, opts%nref, ilog)
  call exportMCTDH(cpd_out, dvr_out, opts%cpdfile_out, opts%lovwr)

  ! clean up
  call deallocate_cpd_t(cpd_in)
  call deallocate_cpd_t(cpd_out)
  call deallocate_dvr_t(dvr_in)
  call deallocate_dvr_t(dvr_out)
  deallocate(refidx)
  deallocate(frem)

contains

subroutine mkremlist(dvr,opts,frem,refidx,ilog)
  implicit none
  type(dvr_t), intent(IN)    :: dvr
  type(ccopts_t), intent(IN) :: opts
  integer, intent(INOUT)     :: frem(opts%nref), refidx(opts%nref)
  integer, intent(IN)        :: ilog

  type(refcoord_t), pointer  :: ptr
  integer                    :: n,f, z,z1,g
  character(len=c5)          :: rout, mdl, msg
  double precision           :: delta, deltamin
  logical                    :: lmatch

  rout = "mkremlist (cutcpd.f90)"
  mdl = "cutcpd"

  write(ilog, '(a)') "Final values of reference points:"
  write(ilog, '(a)') "           Label   Index    Value"

  do n=1,opts%nref
     ptr =>  get_refpoint(opts,n)
     lmatch = .false.
     do f=1,dvr%ndof
        if (lmatch) cycle
        if (trim(ptr%label) .eq. trim(dvr%modelabel(f))) then
           lmatch = .true.
           frem(n) = f

           select case (ptr%type)
           case (REFTYPE_INDEX)
              if (ptr%idx .gt. dvr%gdim(f) .or. ptr%idx .le. 0) then
                 msg = "Reference index is out of bounds for DOF: " &
                      //trim(dvr%modelabel(f))
                 call stop_code(mdl,rout,msg)
              end if

              refidx(n) = ptr%idx
              g = refidx(n)
              z =  dvr%zort(f) + g -1

              write(ilog, '(a,i8,f14.8)') ADJUSTR(ptr%label(1:16)), refidx(n),   dvr%ort(z)


           case (REFTYPE_COORD)
              z =  dvr%zort(f)
              z1 = z + dvr%gdim(f) - 1

              if (ptr%refpoint .gt. dvr%ort(z1) .or. ptr%refpoint .lt.  dvr%ort(z)) then
                 msg = "Reference point is out of DVR grid for DOF: " &
                      //trim(dvr%modelabel(f))
                 call stop_code(mdl,rout,msg)
              end if

              deltamin = 1.d99
              refidx(n) = -1
              do g=1,dvr%gdim(f)
                 z =  dvr%zort(f) + g -1
                 delta = abs( ptr%refpoint - dvr%ort(z))
                 if (delta .lt. deltamin) then
                    refidx(n) = g
                    deltamin = delta
                 end if
              end do

              g = refidx(n)
              z =  dvr%zort(f) + g -1

              write(ilog, '(a,i8,f14.8)') ADJUSTR(ptr%label(1:16)), refidx(n),   dvr%ort(z)

           end select

        end if

     end do
     if (.not. lmatch) then
        msg = "Reference coordinate label not found in input dvr: "//trim(ptr%label)
        call stop_code(mdl,rout,msg)
     end if

  end do

  write(ilog, *)

  if (opts%nref .eq. dvr%ndof) then
     msg = "It seems that all DOF shall be removed from the CPD...??"
     call stop_code(mdl,rout,msg)
  end if


end subroutine mkremlist



subroutine cc_cutcpd(cpd_in, cpd_out, dvr_out, frem, refidx, nref, ilog)
  implicit none

  type(cpd_t), intent(IN)    :: cpd_in
  type(cpd_t), intent(INOUT) :: cpd_out
  type(dvr_t), intent(IN)    :: dvr_out
  integer, intent(IN)        :: nref
  integer, intent(IN)        :: frem(nref),refidx(nref)
  integer, intent(IN)        :: ilog

  logical                    :: lremmode(cpd_in%nmode)
  integer                    :: m,m1,n,n1,f,f1,fr(1),zi,zf,zi1,zf1,g,g1,r
  integer                    :: nmode1
  logical                    :: loned1(dvr_out%ndof)
  integer                    :: potdofmap1(2,dvr_out%ndof)
  integer                    :: dofspf1(dvr_out%ndof)
  integer                    :: newm(cpd_in%nmode), newf(cpd_in%ndof)
  integer, allocatable       :: gdim1(:), gindex(:), gindex1(:)
  type(cpd_t)                :: cpd_tmp

  ! check if we need to remove complete modes
  do m=1, cpd_in%nmode
     lremmode(m) = .true.
     do n=1, cpd_in%nspfdof(m)
        f = cpd_in%spfdof(n,m)
        lremmode(m) = lremmode(m) .and. any(frem .eq. f)
     end do
  end do

  ! mapping of the modes
  nmode1 = 0
  do m=1, cpd_in%nmode
     if (.not. lremmode(m)) then
        nmode1 = nmode1 + 1
        newm(m) = nmode1
     else
        newm(m) = -1
     end if
  end do

  !mapping of the DOF
  f1 = 0
  do f=1,cpd_in%ndof
     if (.not. any(f .eq. frem)) then
        f1 = f1 + 1
        newf(f) = f1
     else
        newf(f) = -1
     end if
  end do

  write(ilog,'(a)') "DOF re-mapping"
  write(ilog,'(a)') "     f    f1  label"
  do f=1,cpd_in%ndof
     if (newf(f).gt.0) then
        write(ilog,'(2i6,2x,a)') f, newf(f), trim(cpd_in%modelabel(f))
     else
        write(ilog,'(i6,a6,2x,a)') f, "    --", cpd_in%modelabel(f)//"  REMOVED"
     end if
  end do
  write(ilog,*)
  write(ilog,'(a)') "MODE re-mapping"
  write(ilog,'(a)') "     m     m1 "
  do m=1,cpd_in%nmode
     if (newm(m).gt.0) then
        write(ilog,'(2i6)') m, newm(m)
     else
        write(ilog,'(i6,a6,a)') m, "    --", "  REMOVED"
     end if
  end do
  write(ilog,*)

  ! now set it the structural information of the new CPD
  do f=1, cpd_in%ndof
     m = cpd_in%dofspf(f)
     m1 = newm(m)
     f1 = newf(f)
     if (m1.ne.-1 .and. f1.ne.-1) then
        dofspf1(f1) = m1
        loned1(f1) = cpd_in%loned(f)
        potdofmap1(1,f1)=m1
     end if
  end do
  do m=1,cpd_in%nmode
     n1 = 0
     do n=1,cpd_in%nspfdof(m)
        f = cpd_in%spfdof(n,m)
        f1 = newf(f)
        m1 = newm(m)
        if (m1.ne.-1 .and. f1.ne.-1) then
           n1 = n1 + 1
           potdofmap1(2,f1) = n1
        end if
     end do
  end do


  ! allocate
  call allocate_cpd_t(cpd_out,cpd_in%rank,dvr_out%ndof,nmode1, &
       dvr_out%gdim,dofspf1,potdofmap1, &
       dvr_out%modelabel,dvr_out%ort,dvr_out%ortdim, loned1)

  write(ilog,'(a)') "New DOF arangement"
  write(ilog,'(a)') "     f     m     n   label"
  do f1=1,dvr_out%ndof
     write(ilog,'(3i6,3x,a)')  f1, potdofmap1(1,f1), potdofmap1(2,f1),  dvr_out%modelabel(f1)
  end do
  write(ilog,*)


  ! copy data
  cpd_out%coeff = cpd_in%coeff

  do f=1, cpd_in%ndof
     f1 = newf(f)
     if (f1 .gt. 0) then
        if (loned1(f1)) then
           zi = cpd_in%zort(f)
           zf = zi +  cpd_in%gdim(f) - 1
           zi1 = cpd_out%zort(f1)
           zf1 = zi1 +  cpd_out%gdim(f1) - 1
           cpd_out%oned(zi1:zf1) = cpd_in%oned(zi:zf)
        end if
     end if
  end do


  do m=1, cpd_in%nmode
     m1 = newm(m)

     if (m1.lt.0) then
        ! the complete mode is removed, We need to
        ! multiply the coefficient
        ! with the spp value on the reference point.
        allocate(gdim(cpd_in%nspfdof(m)))
        allocate(gindex(cpd_in%nspfdof(m)))

        do n=1, cpd_in%nspfdof(m)
           f = cpd_in%spfdof(n,m)
           gdim(n) = cpd_in%gdim(f)
           fr(1) = findloc_int(frem,f)
           gindex(n) = refidx(fr(1))
        end do

        call sindex2cindex(g,gindex,gdim,cpd_in%nspfdof(m))
        cpd_out%coeff = cpd_out%coeff*cpd_in%spp(m)%htm(g,:)

        deallocate(gdim)
        deallocate(gindex)
     else
        ! mode is untouched
        if (cpd_in%nspfdof(m) .eq. cpd_out%nspfdof(m1)) then
           cpd_out%spp(m1)%htm(:,:) = cpd_in%spp(m)%htm(:,:)
        else
           ! some DOF must be removed
           allocate(gdim(cpd_in%nspfdof(m)))
           allocate(gdim1(cpd_out%nspfdof(m1)))
           allocate(gindex(cpd_in%nspfdof(m)))
           allocate(gindex1(cpd_out%nspfdof(m1)))

           n1 = 0
           do n=1, cpd_in%nspfdof(m)
              f = cpd_in%spfdof(n,m)
              f1 = newf(f)
              gdim(n) = cpd_in%gdim(f)
              if (f1.gt.0) then
                 n1 = n1 + 1
                 gdim1(n1) =  cpd_out%gdim(f1)
              end if
           end do

           do g1=1, cpd_out%subdim(m1)
              call cindex2sindex(g1,gindex1,gdim1,cpd_out%nspfdof(m1))

              n1 = 0
              do n=1, cpd_in%nspfdof(m)
                 f = cpd_in%spfdof(n,m)
                 f1 = newf(f)
                 if (f1.gt.0) then

                    n1 = n1 + 1
                    gindex(n) = gindex1(n1)

                 else
                    fr(1) = findloc_int(frem,f)
                    gindex(n) = refidx(fr(1))
                 end if

              end do

              call sindex2cindex(g,gindex,gdim,cpd_in%nspfdof(m))

              cpd_out%spp(m1)%htm(g1,:) = cpd_in%spp(m)%htm(g,:)

           end do

           deallocate(gdim)
           deallocate(gdim1)
           deallocate(gindex)
           deallocate(gindex1)
        end if
     end if
  end do


  ! if only one mode is left
  ! we can sum all the terms into
  ! a single one.
  if (cpd_out%nmode .eq. 1) then

     call copy_cpd(cpd_out, cpd_tmp)
     call deallocate_cpd_t(cpd_out)
     ! allocate

     call allocate_cpd_t(cpd_out,1,dvr_out%ndof,1, &
          dvr_out%gdim,dofspf1,potdofmap1, &
          dvr_out%modelabel,dvr_out%ort,dvr_out%ortdim, loned1)

     cpd_out%coeff(1) = 1.d0

     do r=1,cpd_tmp%rank
        cpd_out%spp(1)%htm(:,1) = cpd_out%spp(1)%htm(:,1) &
             + cpd_tmp%spp(1)%htm(:,r)*cpd_tmp%coeff(r)
     end do

     if (any(loned1)) then
        cpd_out%oned = cpd_tmp%oned
     end if

     call deallocate_cpd_t(cpd_tmp)

  end if



  call normspp(cpd_out)



  write(ilog, '(a)') " OLD CPD INFO "
  write(ilog, '(a)') " ------------ "
  call wrcpdproperties(cpd_in, ilog)
  write(ilog,*)
  write(ilog, '(a)') " NEW CPD INFO "
  write(ilog, '(a)') " ------------ "
  call wrcpdproperties(cpd_out, ilog)

end subroutine cc_cutcpd

  ! substitute for intrinsic findloc()
  ! which is only supported in gcc9
  ! or later.
  pure integer function findloc_int(arr,elem)
    implicit none
    integer, intent(in) :: arr(:), elem
    integer :: i

    findloc_int = 0
    do i=1,size(arr)
       if (elem .eq. arr(i)) then
          findloc_int = i
          return
       end if
    end do

  end function findloc_int


end program cutcpd
