

module cpd_als_m
  use cpdtype_m
  implicit none

  type :: alsmat_t
     double precision, allocatable :: mat(:,:)
     double precision, allocatable :: B(:,:), BT(:,:)
  end type alsmat_t

  type(alsmat_t),allocatable  :: alsmats(:)
  integer                     :: alsndof, alsrank
  logical, save               :: alsinitialized = .false.

  double precision, allocatable :: ovlp(:,:)

  integer                       :: isolver
  ! isolver == 1: LAPACK (P)DPOSV with equal regularization
  ! isolver == 2: LAPACK (P)DSYEV with regularizazion of small Eval

contains

  ! free memory
  subroutine alsfree()
    implicit none
    integer :: m


    if (allocated(alsmats)) then

       do m=1,alsndof
          if (allocated(alsmats(m)%mat)) deallocate(alsmats(m)%mat)
          if (allocated(alsmats(m)%B)) deallocate(alsmats(m)%B)
          if (allocated(alsmats(m)%BT)) deallocate(alsmats(m)%BT)
       end do

       deallocate(alsmats)
       alsndof = 0
       alsrank = 0

    end if

    if (allocated(ovlp)) deallocate(ovlp)

    isolver = 0
    alsinitialized = .false.

  end subroutine alsfree

  ! allocate memory and populate
  subroutine alsinit(cpd)
    implicit none
    type(cpd_t), intent(IN)  :: cpd
    integer                  :: m,r,r1
    double precision         :: ddot

    call alsfree()

    alsndof = cpd%nmode
    alsrank = cpd%rank
    allocate(alsmats(cpd%nmode))

    allocate(ovlp(cpd%rank,cpd%rank))

    do m=1,alsndof

       allocate(alsmats(m)%mat(alsrank,alsrank))
       allocate(alsmats(m)%B(cpd%subdim(m),cpd%rank))
       allocate(alsmats(m)%BT(cpd%rank,cpd%subdim(m)))


       !$omp parallel do private(r1)
       do r=1,alsrank
          do r1=r,alsrank
             alsmats(m)%mat(r1,r) = ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r1),1,cpd%spp(m)%htm(:,r),1 )
          end do
       end do
       !$omp end parallel do

       !$omp parallel do private(r1) schedule(dynamic)
       do r=1,alsrank
          do r1=r+1,alsrank
             alsmats(m)%mat(r,r1) = alsmats(m)%mat(r1,r)
          end do
       end do
       !$omp end parallel do

    end do

    alsinitialized = .true.

    isolver = 1

  end subroutine alsinit

end module cpd_als_m



module cpd_als_s

use cpd_als_m
use pglobal
use timing

private 

public :: alsiter1m, alssetsolver

contains

  subroutine alssetsolver(solver)
    implicit none
    integer, intent(IN) :: solver

    if (solver .le. 0 .or. solver .gt. 2 ) then
       write(*,*) "alssetsolver: solver must be on of 1, or 2. Not:", solver
       stop 1
    end if

    isolver = solver

  end subroutine alssetsolver



  ! do one  ALS iteration for one mode m:
  ! input:
  ! func -    external subroutine. calculates
  !           The right hand side of the ALS equations
  !           that is "B(gdim,rank)"
  !           Inteface: func(cpd,m,B), B: double precision
  !           m, (integer, input) mode for which B is to be computed
  ! cpd -     The candecomp to be optimized.
  !           on exit, optimized candecomp
  ! m -       Mode in which an ALS iteration is done
  ! alpha -   regularization parameter
  !
  ! output:
  ! cpd -     Optimized CPD
  ! ierr -    Nonzero on error, 0 otherwise.
  subroutine alsiter1m(func, cpd, m,alpha, ierr, fname)

    implicit none
    type(cpd_t), intent(INOUT)    :: cpd
    integer, intent(IN)           :: m
    double precision, intent(IN)  :: alpha
    integer, intent(INOUT)        :: ierr
    character(len=*), intent(IN)  :: fname
    integer                       :: m1,r,r1,n
    double precision              :: norm2
    logical                       :: lstart
    integer                       :: tid = 0

    ! BLAS
    double precision              :: ddot

    external func

    if (.not. alsinitialized) then
       write(*,*) "ALS for CPD (Candecomp) is not initialized."
       write(*,*) "Call 'alsinit' first."
       stop 1
    end if

    ierr = 0

    ! right hand side
    call func(cpd,m,alsmats(m)%B)

    lstart = .true.
    do m1=1,cpd%nmode
       if (m1.ne.m) then
          if (lstart) then
             !$omp parallel do private(r) schedule(guided)
             do r1=1,cpd%rank
                do r=r1,cpd%rank
                   ovlp(r,r1) =  alsmats(m1)%mat(r,r1)
                end do
             end do
             !$omp end parallel do
             lstart = .false.
          else
             !$omp parallel do private(r) schedule(guided)
             do r1=1,cpd%rank
                do r=r1,cpd%rank
                   ovlp(r,r1) =  ovlp(r,r1)*alsmats(m1)%mat(r,r1)
                end do
             end do
             !$omp end parallel do
          end if
       end if
    end do

    !$omp parallel do private(r) schedule(guided)
    do r1=1,cpd%rank
       do r=r1+1,cpd%rank
          ovlp(r1,r) =  ovlp(r,r1)
       end do
    end do
    !$omp end parallel do

    !$omp parallel do private(n)
    do r=1,cpd%rank
       do n=1,cpd%subdim(m)
          alsmats(m)%BT(r,n) = alsmats(m)%B(n,r)
       end do
       if (isolver.eq.1) then
          ovlp(r,r) =  ovlp(r,r) + alpha
       end if
    end do
    !$omp end parallel do

    call get_timer(tid, 'ptsolve')
    call start_timer(tid)

    if (isolver.eq.1) then
       call mcsolve_spd(ovlp,alsmats(m)%BT,cpd%rank,cpd%subdim(m),ierr)
    else if (isolver.eq.2) then

       call mcsolve_eigen(ovlp,alsmats(m)%BT,alpha,.false.,.true., &
            cpd%rank,cpd%subdim(m),fname,ierr)
    else
       write(*,*) "ALS: unknown solver:", isolver
       return
    end if
    call stop_timer(tid)

    if (ierr .ne. 0) then
       write(*,*) " solver error code:", ierr
       return
    end if

    !$omp parallel do private(n)
    do r=1,cpd%rank
       do n=1,cpd%subdim(m)
          alsmats(m)%B(n,r) = alsmats(m)%BT(r,n)
       end do
    end do
    !$omp end parallel do


    !$omp parallel do private(norm2)
    do r=1,cpd%rank

       norm2 = ddot(cpd%subdim(m), alsmats(m)%B(:,r),1, alsmats(m)%B(:,r),1)

       if (abs(norm2) .gt. 1.d-12) then

          cpd%coeff(r) = sqrt(norm2)
          cpd%spp(m)%htm(:,r) =  alsmats(m)%B(:,r)/cpd%coeff(r)

       else

          cpd%coeff(r) = 0.d0
          call random_number(cpd%spp(m)%htm(:,r))
          cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r) - 0.5d0
          norm2 = ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r),1)
          cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r) / sqrt(norm2)

       end if
    end do
    !$omp end parallel do

    !$omp parallel do private(r) schedule(guided)
    do r1=1,cpd%rank

       do r=r1,cpd%rank

          alsmats(m)%mat(r,r1) = ddot(cpd%subdim(m), &
               cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1),1)

       end do
    end do
    !$omp end parallel do

    !$omp parallel do private(r) schedule(guided)
    do r1=1,cpd%rank
       do r=r1+1,cpd%rank
          alsmats(m)%mat(r1,r) =  alsmats(m)%mat(r,r1)
       end do
    end do
    !$omp end parallel do


  end subroutine alsiter1m

end module cpd_als_s
