

module co_symmap_m
  use mlbase
  use svdbasis_m
  use linear, only: dgramschmidt2
  implicit none

 ! type :: symmap_t
 !    ! 1:herm, -1: anti-herm, 0: other
 !    ! 0 should not occur
 !    integer, allocatable      :: optype(:)
 ! end type symmap_t


contains

  ! extend SVD basis to be hermitian
  ! build symmetry maps at the same time
  !
  ! classifies all operators as hermitian, anti-hermitian or other
  ! All operators classified as 'other' are separated into
  ! a hermitian and an anti-hermitian part. then all hermitian
  ! and anti-hermitian operators are grouped together and
  ! are Schmidt-orthogonalized.



  subroutine extend_svdbasis_hermitian_1m(basis1m, symmap1m)
    type(svdbasis_t), intent(INOUT) :: basis1m
    type(symmap_t), intent(INOUT)   :: symmap1m
    real(dbl), allocatable          :: work(:,:,:), herm(:,:,:), aherm(:,:,:)
    real(dbl), allocatable          :: tmpT(:,:)
    real(dbl)                       :: norm
    real(dbl)                       :: dnrm2 ! BLAS Euklidian norm
    integer                         :: n,n1,subdim,nherm,naherm,nother,ierr
    integer, allocatable            :: optype(:)

    if (.not. allocated(basis1m%basis)) return


    subdim = nint(sqrt(dble(basis1m%subdim)))
    if (subdim**2 .ne. basis1m%subdim) then
       print*, "Basis apperently not containing matrices"
       stop 1
    end if

    allocate(work(subdim,subdim,basis1m%nbasis))
    allocate(tmpT(subdim,subdim))

    do n=1,basis1m%nbasis
       work(:,:,n) = reshape(basis1m%basis(:,n),(/subdim,subdim/))
    end do

    allocate(optype(basis1m%nbasis))
    optype = 0

    nherm = 0
    naherm = 0
    do n=1,basis1m%nbasis
       tmpT = transpose(work(:,:,n))
       ! check if this basis is hermitian
       if ( all( abs(tmpT  - work(:,:,n)) .le. 1.d-8)) then
          optype(n) = 1
          nherm = nherm + 1
          ! check if its hermitian conjugate is present
       else  if ( all( abs(tmpT  + work(:,:,n)) .le. 1.d-8)) then
          optype(n) = -1
          naherm = naherm + 1
       end if
    end do

    deallocate(tmpT)

    nother = basis1m%nbasis - nherm - naherm
    nherm = nherm + nother
    naherm = naherm + nother

    allocate(herm(subdim,subdim,nherm))
    allocate(aherm(subdim,subdim,naherm))

    n1 = 0
    do n=1,basis1m%nbasis
       if (optype(n) .eq. 1 .or. optype(n) .eq. 0) then
          n1 = n1 + 1
          herm(:,:,n1) = work(:,:,n) + transpose(work(:,:,n))
          norm = dnrm2(subdim**2, herm(:,:,n1),1)
          herm(:,:,n1) = herm(:,:,n1)/norm
       end if
    end do

    n1 = 0
    do n=1,basis1m%nbasis
       if (optype(n) .eq. -1 .or. optype(n) .eq. 0) then
          n1 = n1 + 1
          aherm(:,:,n1) = work(:,:,n) - transpose(work(:,:,n))
          norm = dnrm2(subdim**2, aherm(:,:,n1),1)
          aherm(:,:,n1) = aherm(:,:,n1)/norm
       end if
    end do

    call dgramschmidt2(herm,subdim**2,nherm,ierr)
    if (ierr .ne. 0) then
       print*, "Hermitian terms: Schmidt orthoginalization failed for vector ", ierr
    end if

    call dgramschmidt2(aherm,subdim**2,naherm,ierr)
    if (ierr .ne. 0) then
       print*, "Anti-Hermitian terms: Schmidt orthoginalization failed for vector ", ierr
    end if


    basis1m%nbasis = nherm + naherm


    if (allocated(basis1m%basis)) deallocate(basis1m%basis)
    if (allocated(basis1m%sigma)) deallocate(basis1m%sigma) !this makes no sense any more

    allocate(basis1m%basis(subdim**2,basis1m%nbasis))

    do n=1,nherm
       basis1m%basis(:,n) = reshape(herm(:,:,n), (/subdim**2/))
    end do
    do n=1,naherm
       basis1m%basis(:,n+nherm) = reshape(aherm(:,:,n), (/subdim**2/))
    end do

    if (allocated(symmap1m%symtype)) deallocate(symmap1m%symtype)
    allocate(symmap1m%symtype(basis1m%nbasis))

    do n=1,nherm
       symmap1m%symtype(n) = 1
    end do
    do n=1,naherm
       symmap1m%symtype(n+nherm) = -1
    end do

!!$    open(newunit=lun, file="basisovlp")
!!$    do n=1,basis1m%nbasis
!!$       do n1=1,basis1m%nbasis
!!$          write(lun,*) n1,n,ddot(subdim**2,basis1m%basis(:,n),1,basis1m%basis(:,n1),1 )
!!$       end do
!!$        write(lun,*)  ""
!!$     end do
!!$    close(lun)
!!$    stop


    deallocate(work)
    deallocate(herm)
    deallocate(aherm)

  end subroutine extend_svdbasis_hermitian_1m


end module co_symmap_m
