module cutcpd_opts_m
  use decimal
  use global
  use mlbase
  use versions

  implicit none

  integer, parameter :: REFTYPE_INDEX = 0
  integer, parameter :: REFTYPE_COORD = 1

  ! linked list of reference points
  type refcoord_t
     character(c2)             :: label
     integer                   :: type, idx
     real(kind=dbl)            :: refpoint
     type(refcoord_t), pointer :: next => Null()

  end type refcoord_t

  ! typie holding all input information
  type ccopts_t
     character(len=MAXFNAME)   :: name
     character(len=MAXFNAME)   :: inpfile
     character(len=MAXFNAME)   :: dvrfile_in
     character(len=MAXFNAME)   :: dvrfile_out
     character(len=MAXFNAME)   :: cpdfile_in
     character(len=MAXFNAME)   :: cpdfile_out
     character(len=MAXFNAME)   :: logfile

     logical                   :: lovwr
     logical                   :: lmnd

     type(refcoord_t), pointer :: first_ref => Null()
     integer                   :: nref

   contains

     procedure, pass(self)     :: get_refpoint


  end type ccopts_t

contains

  ! append a reference point to the list of reference points
  ! current must be a pointer to the last reference pount
  ! in the list
  subroutine append_refpoint(current)
    implicit none
    type(refcoord_t), pointer, intent(INOUT) :: current
    type(refcoord_t), pointer       :: next

    allocate(next)
    nullify(next%next)

    next%next => Null()
    current%next => next
    current => next

  end subroutine append_refpoint


  ! get a reference point object by its number form the options object
  function get_refpoint(self,n) result(ptr)
    implicit none
    integer, intent(in)          :: n
    class(ccopts_t),  intent(IN) :: self
    type(refcoord_t), pointer    :: ptr
    integer                      :: f

    ptr => self%first_ref

    if (n .eq. 1) return

    do f=1,n-1
       ptr => ptr%next
    end do

  end function get_refpoint


  ! set options defaults
  subroutine ccopts_default(opts)
    implicit none

    type(ccopts_t), intent(INOUT) :: opts

    opts%name = " "
    opts%inpfile = " "
    opts%dvrfile_in = " "
    opts%dvrfile_out = " "
    opts%cpdfile_in = " "
    opts%cpdfile_out = " "
    opts%logfile = " "

    opts%lovwr = .false.
    opts%lmnd = .false.
    opts%nref = 0
    opts%first_ref => Null()

  end subroutine ccopts_default

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! prnt help
  subroutine cc_help()
    implicit none

    write(6,*)
    write(6,'(a)')  'Purpose: read a cpd file and produce a cut through a'
    write(6,'(a)')  '         pre-defined reference point. Produces a new'
    write(6,'(a)')  '         cpd file that only depends on coordinates'
    write(6,'(a)')  '         that are not part of the reference geometry.'
    write(6,'(a)')  '         Modes that only contain reference points'
    write(6,'(a)')  '         (i.e., which are constant) are removed from'
    write(6,'(a)')  '         the cpd file.'

    write(6,*)
    write(6,'(a)')  '  Usage:  cpdcut<ver><d/D> [-h|-?|-ver] [-w -mnd -D] <inp>'
    write(6,'(a)')  '          where <inp> is an input file.'
    write(6,*)
    write(6,'(a)')  'Options:'

    write(6,'(a)')  '  -h      : Print this help-text.'
    write(6,'(a)')  '  -?      : Print this help-text.'
    write(6,'(a)')  '-ver      : Version information.'
    write(6,'(a)')  '  -w      : Allow overwrite.'
    write(6,'(2a)') '-mnd      : Make Name Directory. ', &
         'Creates the name directory, if not existent.'
    write(6,'(2a)') '-D <name> : Write output to <name>.', &
         ' (Ignore <name> in input file.)'
    write(6,'(a)')  ''
    write(6,'(a)')  ' Remark:  It may be beneficial to run compactoper'
    write(6,'(a)')  '          on the newly created cpd file.'
    write(6,'(a)')  ''
    stop 0

  end subroutine cc_help


  ! read input and command line options
  ! make name and open log
  subroutine cc_inp(opts,ilog)
    implicit none

    type(ccopts_t), intent(INOUT) :: opts
    integer, intent(OUT)          :: ilog

    integer                       :: i, iarg, myiargc, ierr
    integer                       :: lenname
    character(len=c5)             :: buf, options
    character(len=c5)             :: rout, mdl, msg
    logical                       :: lname, lexists
    character(len=c5)             :: inptit

    rout = "cc_inp"
    mdl  = "cpdcut_opts_m"

    lname = .false.


    call ccopts_default(opts)

    iarg = myiargc()

    call mygetarg(0,options)
    do i = 1, iarg
       call mygetarg(i,buf)
       options = trim(options)//' '//trim(buf)
    end do

    do i=1,iarg
       call mygetarg(i,buf)
       if( buf(1:1) .eq. '-' ) then
          if ( buf .eq. '-h' .or. &
               buf .eq. '--help' .or. &
               buf .eq. '-?' ) then
             call cc_help()

          elseif  ( buf .eq. '-ver') then
             call wrversion(6)
             stop 0
          end if

       end if
    end do

    i = 0
    do while (i.lt. iarg-1)
       i=i+1

       call mygetarg(i,buf)
       if( buf(1:1) .eq. '-' ) then

          if( trim(buf) .eq. '-mnd') then
             opts%lmnd = .true.
          elseif( buf .eq. '-w') then
             opts%lovwr = .true.
          elseif( trim(buf) .eq. '-D') then
             lname = .true.
             i = i+1
             if (i.eq. iarg) then
                msg = "Expecting <name> after -D option"
                call stop_code(mdl,rout,msg)
             end if
             call mygetarg(i,buf)
             opts%name = trim(buf)
          else
             msg = "Unknown option: "//trim(buf)
             call stop_code(mdl,rout,msg)
          end if
       else
          msg = "Unknown option: "//trim(buf)
          call stop_code(mdl,rout,msg)
       end if
    end do

    ! the last keyword is the input file
    i = i+1
    call mygetarg(i,buf)
    opts%inpfile = trim(buf)

    ! parse the input file
    ! make sure not to overwrite <name>
    call cc_parse_inpfile(opts, lname)

    ! make name if requered and start logging
    if (len_trim(opts%name) .eq. 0) then
       msg = "No name directory given."
       call stop_code(mdl,rout,msg)
    end if

    if (len_trim(opts%cpdfile_in) .eq. 0) then
       msg = "No cpd file given (readcpd=...)"
       call stop_code(mdl,rout,msg)
    end if

    if (len_trim(opts%dvrfile_in) .eq. 0) then
       i = len_trim(opts%cpdfile_in)-4
       opts%dvrfile_in = opts%cpdfile_in(1:i)//"/dvr"
    end if



    lenname = len_trim(opts%name)
    call abspath(opts%name,lenname)
    if (opts%lmnd) then
       inquire (file=trim( opts%name), exist=lexists)
       if (lexists) then
          write(6,'(a)') "Name-directory already exists: "//trim(opts%name)
       else
          inptit='mkdir '//trim(opts%name)
          call excmd2(ierr, inptit)
          if (ierr.ne.0) then
             msg = '### ERROR: Invalid path for name-directory (-mnd option). ###'
             call stop_code(mdl,rout,msg)
          endif
          write(6,'(2a)') 'Created name-directory: ',trim(opts%name)
       end if
    end if

    opts%cpdfile_out = trim(opts%name)//"/cpd"
    opts%dvrfile_out = trim(opts%name)//"/dvr"
    opts%logfile = trim(opts%name)//    "/log"

    inquire (file=opts%cpdfile_in, exist=lexists)
    if (.not. lexists ) then
       msg = "File "//trim(opts%cpdfile_in)//" not found."
       call stop_code(mdl,rout,msg)
    end if
    inquire (file=opts%cpdfile_out, exist=lexists)
    if (lexists .and. .not. opts%lovwr) then
       msg = "No permission to overwrite "//trim(opts%cpdfile_out)//". Use -w option?"
       call stop_code(mdl,rout,msg)
    end if
    inquire (file=opts%dvrfile_out, exist=lexists)
    if (lexists .and. .not. opts%lovwr) then
       msg = "No permission to overwrite "//trim(opts%dvrfile_out)//". Use -w option?"
       call stop_code(mdl,rout,msg)
    end if


    inquire (file=opts%logfile, exist=lexists)
    if (lexists .and. .not. opts%lovwr) then
       msg = "No permission to overwrite "//trim(opts%logfile)//". Use -w option?"
       call stop_code(mdl,rout,msg)
    end if

    open(newunit=ilog, file=trim(opts%logfile), action="write")
    call cc_log_initial(opts, ilog)




  end subroutine cc_inp

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! after the input is read we log the information gathered
  subroutine cc_log_initial(opts,ilog)
    integer, intent(IN)        :: ilog
    type(ccopts_t), intent(IN) :: opts
    character(len=100)         :: buffer
    type(refcoord_t), pointer  :: ptr
    integer                    :: n

    call fpstartup(buffer)
    write(ilog,*)
    write(ilog,'(a)') trim(buffer)
    write(ilog,'(1x,a)') trim(opts%name)

! --- WRITE PROJECT VERSION NUMBERS TO LOG FILE ---
    call wrversion(ilog)

    write(ilog,'(a)') "Options:"


    write(ilog,'(2x,2a)') "name:               ", trim(opts%name)
    write(ilog,'(2x,2a)') "Input file:         ", trim(opts%inpfile)
    write(ilog,'(2x,2a)') "DVR-file (input):   ", trim(opts%dvrfile_in)
    write(ilog,'(2x,2a)') "DVR-file (output):  ", trim(opts%dvrfile_out)
    write(ilog,'(2x,2a)') "CPD-file (input):   ", trim(opts%cpdfile_in)
    write(ilog,'(2x,2a)') "CPD-file (output):  ", trim(opts%cpdfile_out)
    write(ilog,'(2x,2a)') "Log (this file)     ", trim(opts%logfile)
    write(ilog,*)
    if (opts%lovwr) then
       write(ilog,'(2x,2a)') "Overwrite enabled:  ", "yes"
    else
       write(ilog,'(2x,2a)') "Overwrite enabled:  ", "no"
    end if
    if (opts%lmnd) then
       write(ilog,'(2x,2a)') "Make name enabled:  ", "yes"
    else
       write(ilog,'(2x,2a)') "Make name enabled:  ", "no"
    end if

    write(ilog,*)
    write(ilog,'(a)') "Reference points:"

    do n=1,opts%nref
       ptr => get_refpoint(opts,n)
       if (ptr%type .eq.  REFTYPE_INDEX) then
          write(ilog,'(i4,2x,a16,a,i4)') n, ptr%label, " DVR-Point: ", ptr%idx
       else if (ptr%type .eq.  REFTYPE_COORD) then
          write(ilog,'(i4,2x,a16,a,f12.8)') n, ptr%label, " Value:     ", ptr%refpoint
       end if
    end do
    write(ilog,*)


  end subroutine cc_log_initial

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! read through the input
  subroutine cc_parse_inpfile(opts, lname)
    type(ccopts_t), intent(INOUT) :: opts
    logical, intent(IN)           :: lname

    integer                       :: iin,ierr
    character(len=c5)             :: rout, mdl, msg
    logical                       :: lexists

    rout = "cc_parse_inpfile"
    mdl  = "cpdcut_opts_m"

    inquire (file=trim(opts%inpfile), exist=lexists)
    if (.not. lexists) then
       msg = "Input file "//trim(opts%cpdfile_out)//". not found."
       call stop_code(mdl,rout,msg)
    end if

    open(newunit=iin, file=trim(opts%inpfile), action='read',status='old',iostat=ierr)
    if (ierr .ne. 0) then
       msg = "Could not open input file for reading: "//trim(opts%inpfile)
       call stop_code(mdl,rout,msg)
    end if

    call cc_parse_run(iin, opts, lname)
    call cc_parse_ref(iin, opts)

    close(iin)

  end subroutine cc_parse_inpfile

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !read RUN-SECTION
  subroutine cc_parse_run(iin, opts, lname)
    integer, intent(IN)           :: iin
    type(ccopts_t), intent(INOUT) :: opts
    logical, intent(IN)           :: lname

    integer, parameter            :: maxkey=48, maxkeylen=200
    integer                       :: iz,ic,ierr,lc(maxkey)
    character(len=maxkeylen)      :: keyword(maxkey),keyorig(maxkey)
    character(len=c5)             :: inptit
    character(len=c5)             :: rout, mdl, msg

    integer                       :: i

    rout = "cc_parse_run"
    mdl  = "cpdcut_opts_m"

    rewind(iin)
    iz=0
10  continue

    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
    if (ierr .eq. 1) then
       msg = trim(inptit)
       call stop_code(mdl,rout,msg)
    endif

    if (keyword(1) .eq. 'run-section') then
       goto 20
    else if (keyword(1) .eq. 'end-input') then
       msg='No RUN-SECTION in input file'
       call stop_code(mdl,rout,msg)
    else
       go to 10
    endif

20  continue
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

    i=1
30  continue

    if (keyword(i).eq.'end-run-section') then
       return

    else if (keyword(1).eq.'end-input') then
       msg='No END-RUN-SECTION keyword in input file'
       call stop_code(mdl,rout,msg)

    elseif( keyword(i) .eq. 'name') then
       if (keyword(i+1).ne.'=') then
          msg = 'No argument specified with name keyword'
          call stop_code(mdl,rout,msg)
       end if
       i = i+2
       if (.not. lname) then
          opts%name = keyorig(i)
       end if

    elseif (keyword(i).eq.'overwrite') then
       opts%lovwr = .true.

    else if (keyword(i) .eq. 'readdvr') then
       if (keyword(i+1) .eq. '=') then
          i=i+2
          opts%dvrfile_in=trim(keyorig(i))//"/dvr"
       else
          msg = 'expecting "=" after "readdvr".'
          call stop_code(mdl,rout,msg)
       end if
    else if (keyword(i) .eq. 'readcpd') then
       if (keyword(i+1) .eq. '=') then
          i=i+2
          opts%cpdfile_in=trim(keyorig(i))//"/cpd"
       else
          msg = 'expecting "=" after "readcpd".'
          call stop_code(mdl,rout,msg)
       end if

    else
       msg = ' unknown keyword : '//trim(keyword(i))//" Line:"//trim(inptit)
       call stop_code(mdl,rout,msg)
    endif

    if (i.lt.ic) then
       i=i+1
       go to 30
    else
       go to 20
    endif


  end subroutine cc_parse_run

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! readd REFERENCE-SECTION
  subroutine cc_parse_ref(iin, opts)
    integer, intent(IN)           :: iin
    type(ccopts_t), intent(INOUT) :: opts

    integer, parameter            :: maxkey=48, maxkeylen=200
    integer                       :: iz,ic,ierr,lc(maxkey),n,n1
    character(len=maxkeylen)      :: keyword(maxkey),keyorig(maxkey)
    character(len=c5)             :: inptit
    character(len=c5)             :: rout, mdl, msg
    character(len=c1)             :: label
    logical                       :: lstart
    type(refcoord_t), pointer     :: current

    rout = "cc_parse_ref"
    mdl  = "cpdcut_opts_m"

    lstart = .true.

    rewind(iin)
    iz=0
100  continue

    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

    if (ierr .eq. 1) then
       msg = trim(inptit)
       call stop_code(mdl,rout,msg)
    endif

    if (keyword(1) .eq. 'reference-section') then
       goto 200
    else if (keyword(1) .eq. 'end-input') then
       msg='No REFERENCE-SECTION in input file'
       call stop_code(mdl,rout,msg)
    else
       go to 100
    endif

200 continue
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

    if (keyword(1) .eq. 'begin-reference') then
       goto 300
    else if  (keyword(1) .eq. 'end-reference-section') then
       msg='REFERENCE-SECTION: expecting "begin-reference"'
       call stop_code(mdl,rout,msg)
    else if  (keyword(1) .eq. 'end-reference') then
       msg='REFERENCE-SECTION: found "end-reference" but no "begin-reference"'
       call stop_code(mdl,rout,msg)
    else if  (keyword(1) .eq. 'end-input') then
       return
    else
       goto 200
    end if

300 continue
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

    if (keyword(1) .eq. 'end-reference') then
       goto 400
    else if (keyword(2) .eq. "=") then ! coord based
       if (lstart) then
          allocate(opts%first_ref)
          current => opts%first_ref
          lstart = .false.
       else
          call append_refpoint(current)
       end if

       current%label = keyorig(1)
       current%type =  REFTYPE_COORD
       if (trim(keyorig(2)) .ne. "=") then
          msg = "REFERENCE-SECTION: expecting either '=' or '[' as second key in reference definition"
          call stop_code(mdl,rout,msg)
       end if
       read(keyword(3),*,iostat=ierr) current%refpoint
       if (ierr .ne. 0) then
          msg = "REFERENCE-SECTION: expecting real after 'coord_label =', not"//trim(keyword(3))
          call stop_code(mdl,rout,msg)
       end if
       opts%nref = opts%nref + 1

       goto 300
    else if  (keyword(2) .eq. "[") then ! index based
       if (lstart) then
          allocate(opts%first_ref)
          current => opts%first_ref
          nullify(current%next)
          lstart = .false.
       else
          call append_refpoint(current)
       end if
       current%label = trim(keyorig(1))
       current%type =  REFTYPE_INDEX
       if (trim(keyorig(2)) .ne. "[" .or. trim(keyorig(4)) .ne. ']' ) then
          msg = "REFERENCE-SECTION: expecting either 'coord_label =' or 'coord_label[int]', not "&
               //trim(inptit)
          call stop_code(mdl,rout,msg)
       end if
       read(keyword(3),*,iostat=ierr) current%idx
       if (ierr .ne. 0) then
          msg = "REFERENCE-SECTION: expecting integer in square brackets, not"//trim(keyword(3))
          call stop_code(mdl,rout,msg)
       end if

       opts%nref = opts%nref + 1

       goto 300
    else if (ic .eq. 0) then ! empty
       goto 300
    else ! other, not valid
       msg = "REFERENCE-SECTION: did not understand line: "//trim(inptit)
       call stop_code(mdl,rout,msg)
    end if


400 continue


    if (opts%nref .eq. 0) then
       msg = "REFERENCE-SECTION: did not find any reference point."
       call stop_code(mdl,rout,msg)
    end if

    ! check for duplicate DOF labels
    do n=1, opts%nref
       current=> opts%get_refpoint(n)
       label = current%label
       do n1=n+1, opts%nref
          current=> opts%get_refpoint(n1)
          if (trim(current%label) .eq. trim(label)) then
             msg = "REFERENCE-SECTION: found duplicate reference point, DOF: "//trim(label)
             call stop_code(mdl,rout,msg)
          end if
       end do
    end do


  end subroutine cc_parse_ref


end module cutcpd_opts_m
