module hals_m
  ! lib90
  use global
  use timing
  ! mcpot
  use tensortype
  use hcpdtype_m
  use hcpdtype_s
  use cpdtype_m
  use cpdtype_s
  use tensortools
  ! opertools
  use co_opts_m
  use svdbasis_m, only : symmap_t
  use diistype_m

  ! other
  use ISO_FORTRAN_ENV ! defines real128 if supported by platform

  implicit none

  real(dbl), allocatable :: crosssigma(:,:,:)
  real(dbl), allocatable :: selfsigma(:,:,:)


  ! list of compilers that support real128
#if defined(__GFORTRAN__) || defined(__INTEL_COMPILER)
  real(kind=real128)   :: ovlp_orig_orig
#else
  ! fall back to double precision
  real(dbl)            :: ovlp_orig_orig
#endif


contains


  ! dot_product with quad presision result
  ! if available. otherwise fall back to
  ! blas ddot
  function qddot(N,arr1,arr2) result(dot)

    integer, intent(IN)     :: N
    real(dbl), intent(IN)   :: arr1(N), arr2(N)

#if defined(__GFORTRAN__) || defined(__INTEL_COMPILER)
    real(kind=real128)      :: dot
    integer                 :: i

    dot = 0.d0
    !!$omp parallel do reduction(+:dot) schedule(static) if(N .gt. 4096)
    do i=1,N
       dot = dot + arr1(i)*arr2(i)
    end do
    !!$omp end parallel do

#else
    real(dbl)               :: dot, ddot

    ! real*8 blas
    dot = ddot(N,arr1,1,arr2,1)

#endif

  end function qddot



  ! check is the stop file exists
  ! and if it says 'continue'
  ! if not, then stop iterations.
  subroutine chkstop(stopfile, lstop)
    implicit none
    character(len=*), intent(IN) :: stopfile
    logical, intent(OUT)         :: lstop
    logical           :: lexists
    integer           :: istop
    character(len=c4) :: content
    integer           :: ilbl
    character(len=c5) :: rout, mdl, msg

    inquire(file=trim(stopfile), exist=lexists)

    if (.not. lexists) then
       lstop = .true.
    else
       open(newunit=istop, file=stopfile, err=398)
       read(istop,*,err=399) content
       close(istop)
       call cncntrblnk(content)
       content = adjustl(content)
       ilbl = index(content,' ')
       if (content(1:ilbl) .eq. "continue") then
          lstop = .false.
       else
          lstop = .true.
       end if
    end if

    return

398 continue
    ! open-error
    rout = "chkstop"
    mdl  = "hcpd_als"
    msg  = "Could not open file: "//trim(stopfile)
    call warn_msg(mdl,rout,msg)
    lstop = .true.
    return
399 continue
    ! read or open-error
    rout = "chkstop"
    mdl  = "hcpd_als"
    msg  = "Could not read file: "//trim(stopfile)
    call warn_msg(mdl,rout,msg)
    lstop = .true.
    return

  end subroutine chkstop


  ! fast evaluation of RMS error
  ! uses things that have been calculated before
  ! if the compiler supports it (gfortran/intel)
  ! the result is in quadruple precision
  function fast_rms(opercoeff,fitcoeff,  &
       operrank,fitrank,nmode,rtotgdim)  result(RMS)

    implicit none

    integer, intent(IN)    :: operrank, fitrank, nmode
    real(dbl), intent(IN)  :: rtotgdim
    real(dbl), intent(IN)  :: opercoeff(operrank)
    real(dbl), intent(IN)  :: fitcoeff(fitrank)


#if defined(__GFORTRAN__) || defined(__INTEL_COMPILER)
    real(kind=real128)     :: RMS, cross, self
#else
    real(dbl)              :: RMS, cross, self
#endif

    real(dbl), allocatable :: rtmp(:,:)
    integer                :: r,r1,m
    integer, save          :: tid = 0


    call get_timer(tid, 'fast_rms')
    call start_timer(tid)

    allocate(rtmp(operrank,fitrank))
    rtmp = 0.d0

    call DGER(operrank,fitrank,1.d0,opercoeff,1,fitcoeff,1,rtmp,operrank)

    do m=1,nmode
       rtmp(:,:) = rtmp(:,:) * crosssigma(:,:,m)
    end do

    cross = sum(rtmp(:,:))

    deallocate(rtmp)
    allocate(rtmp(fitrank,fitrank))
    rtmp = 0.d0

    call DSYR('L',fitrank,1.d0,fitcoeff,1,rtmp,fitrank)

    do m=1,nmode
       rtmp(:,:) = rtmp(:,:) * selfsigma(:,:,m)
    end do


    self = 0.d0
    do r1=1, fitrank
       self = self + rtmp(r1,r1)
       do r=r1+1, fitrank
          self = self + rtmp(r,r1)*2.d0
       end do
    end do


    deallocate(rtmp)

    rms = abs(ovlp_orig_orig - 2.d0*cross + self)
    rms = sqrt(rms/rtotgdim)

    call stop_timer(tid)

  end function fast_rms



  subroutine init_hals(toper,thcpd,lupdateoperovlp,luseblas)
    implicit none
    class(tensor_t), intent(IN),target  :: toper
    class(tensor_t), intent(IN),target  :: thcpd
    logical, intent(IN)                 :: lupdateoperovlp
    logical, intent(IN)                 :: luseblas

    type(hcpd_t), pointer               :: oper => Null()
    type(hcpd_t), pointer               :: hcpd => Null()
    type(cpd_t), pointer                :: diagoper => Null()
    type(cpd_t), pointer                :: cpd => Null()

    real(dbl), allocatable              :: rtmp(:,:)
    real(dbl), allocatable              :: rtmpm(:,:)

    real(dbl)                           :: ddot
    integer                             :: m,N,r,r1
    integer                             :: operrank, fitrank, nmode
    integer, save                       :: tid = 0

    external ddot

    call get_timer(tid, 'init_als')
    call start_timer(tid)

    call free_hals()

    select type(toper)
    type is(hcpd_t)
       oper => toper
       operrank = oper%rank
       nmode = oper%nmode
    type is(cpd_t)
       diagoper => toper
       operrank = diagoper%rank
       nmode = diagoper%nmode
    class default
       print*, "OPER UNDEFINED"
       stop 1
    end select

    select type(thcpd)
    type is(hcpd_t)
       hcpd => thcpd
       fitrank = hcpd%rank
    type is(cpd_t)
       cpd => thcpd
       fitrank = cpd%rank
    class default
       print*, "FIT UNDEFINED"
       stop 1
    end select


    allocate(crosssigma(operrank, fitrank, nmode))
    allocate(selfsigma(fitrank, fitrank, nmode))


    do m=1,nmode
       if (associated(oper) .and. associated(hcpd)) then
          call update_als_hcpd(oper, hcpd, m, luseblas)
       else if (associated(diagoper) .and. associated(cpd)) then
          call update_als_cpd(diagoper, cpd, m,luseblas)
       else
          print*, "subroutine init_hals(toper,thcpd,lupdateoperovlp)"
          print*, "Invalid mix operators."
          print*, "Please file a bug report."
          stop 1
       end if
    end do


    if (.not. lupdateoperovlp) then
       call stop_timer(tid)
       return
    end if

    if (associated(cpd)) then

       if (luseblas) then
          allocate(rtmp(diagoper%rank,diagoper%rank))

          R = diagoper%rank
          rtmp = 0.d0
          call dsyr('L',R,1.d0,diagoper%coeff,1,rtmp, R)

          allocate(rtmpm(diagoper%rank,diagoper%rank))
          do m=1,diagoper%nmode

             rtmpm = 0.d0
             N = diagoper%subdim(m)

             call DSYRK("L","T",R,N,1.d0,diagoper%spp(m)%htm,N,0.d0,rtmpm(:,:),R)
             rtmp(:,:) = rtmp(:,:)*rtmpm(:,:)
          end do
          deallocate(rtmpm)


          ovlp_orig_orig = 0.d0
          !$omp parallel do private(r) reduction(+:ovlp_orig_orig)
          do r1=1, diagoper%rank
             ovlp_orig_orig = ovlp_orig_orig + rtmp(r1,r1)
             do r=r1+1, diagoper%rank
                ovlp_orig_orig = ovlp_orig_orig + rtmp(r,r1)*2.d0
             end do
          end do
          !$omp end parallel do
          deallocate(rtmp)

       else

          ovlp_orig_orig =  toverlap(diagoper,diagoper)

       end if
       call stop_timer(tid)
       return
    end if


    allocate(rtmp(oper%rank,oper%rank))


    if (luseblas) then

       R = oper%rank
       rtmp = 0.d0
       call dsyr('L',R,1.d0,oper%coeff,1,rtmp, R)

       allocate(rtmpm(oper%rank,oper%rank))


       do m=1,oper%nmode

          rtmpm = 0.d0
          N = oper%subdim2(m)

          call DSYRK("L","T",R,N,1.d0,oper%sph(m)%htm,N,0.d0,rtmpm(:,:),R)
          rtmp(:,:) = rtmp(:,:)*rtmpm(:,:)
       end do


       deallocate(rtmpm)

    else


       !!$omp parallel do private(r)
       do r1=1, oper%rank
          do r=r1, oper%rank
             rtmp(r,r1) = oper%coeff(r)*oper%coeff(r1)
          end do
       end do
       !!$omp end parallel do

       do m=1,hcpd%nmode
          N = oper%subdim2(m)
          !$omp parallel do private(r)
          do r1=1, oper%rank
             do r=r1, oper%rank
                rtmp(r,r1) = rtmp(r,r1) * ddot(N,oper%sph(m)%htm(:,:,r),1, &
                     oper%sph(m)%htm(:,:,r1), 1)
             end do
          end do
          !$omp end parallel do
       end do

    end if ! noblas



    ovlp_orig_orig = 0.d0

    !$omp parallel do private(r) reduction(+:ovlp_orig_orig)
    do r1=1, oper%rank
       ovlp_orig_orig = ovlp_orig_orig + rtmp(r1,r1)
       do r=r1+1, oper%rank
          ovlp_orig_orig = ovlp_orig_orig + rtmp(r,r1)*2.d0
       end do
    end do
    !$omp end parallel do

    deallocate(rtmp)

    call stop_timer(tid)

  end subroutine init_hals



  subroutine free_hals()

    implicit none

    if (allocated(crosssigma)) deallocate(crosssigma)
    if (allocated(selfsigma)) deallocate(selfsigma)

  end subroutine free_hals


  subroutine update_als_hcpd(oper, hcpd, m, luseblas)
    implicit none

    type(hcpd_t), intent(IN) :: oper
    type(hcpd_t), intent(IN) :: hcpd
    integer, intent(IN)      :: m
    logical, intent(IN)      :: luseblas
    real(dbl)                :: ddot
    integer                  :: r,r1,N
    integer, save            :: tid = 0

    external ddot


    call get_timer(tid, 'update_hcpd_1m')
    call start_timer(tid)


    if (luseblas) then

       R = oper%rank
       R1 = hcpd%rank
       N = oper%subdim2(m)

       call dgemm('T','N',R,R1,N,1.d0,oper%sph(m)%htm,N,hcpd%sph(m)%htm,N,0.d0,crosssigma(:,:,m),R)
       call DSYRK("L","T",R1,N,1.d0,hcpd%sph(m)%htm,N,0.d0,selfsigma(:,:,m),R1)

    else

       N = oper%subdim2(m)

       !$omp parallel do collapse(2) schedule(static)
       do r1=1, hcpd%rank
          do r=1, oper%rank
             crosssigma(r,r1,m) = ddot(N,oper%sph(m)%htm(:,:,r),1,hcpd%sph(m)%htm(:,:,r1), 1)
          end do
       end do
       !$omp end parallel do

       !$omp parallel do private(r)
       do r1=1, hcpd%rank
          do r=r1, hcpd%rank
             selfsigma(r,r1,m) = ddot(N,hcpd%sph(m)%htm(:,:,r),1,hcpd%sph(m)%htm(:,:,r1), 1)
          end do
       end do
       !$omp end parallel do

    end if

    do r1=1, hcpd%rank
       do r=r1+1, hcpd%rank
          selfsigma(r1,r,m) = selfsigma(r,r1,m)
       end do
    end do

    call stop_timer(tid)

  end subroutine update_als_hcpd


  subroutine update_als_cpd(oper, cpd, m, luseblas)
    implicit none

    type(cpd_t), intent(IN)  :: oper
    type(cpd_t), intent(IN)  :: cpd
    integer, intent(IN)      :: m
    logical, intent(IN)      :: luseblas
    real(dbl)                :: ddot
    integer                  :: r,r1,N
    integer, save            :: tid = 0

    external ddot

    call get_timer(tid, 'update_cpd_1m')
    call start_timer(tid)

    N = oper%subdim(m)

    if (luseblas) then

       R = oper%rank
       R1 = cpd%rank
       N = oper%subdim(m)

       call dgemm('T','N',R,R1,N,1.d0,oper%spp(m)%htm,N,cpd%spp(m)%htm,N,0.d0,crosssigma(:,:,m),R)
       call DSYRK("L","T",R1,N,1.d0,cpd%spp(m)%htm,N,0.d0,selfsigma(:,:,m),R1)

    else

       !$omp parallel do collapse(2) schedule(static)
       do r1=1, cpd%rank
          do r=1, oper%rank
             crosssigma(r,r1,m) = ddot(N,oper%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1), 1)
          end do
       end do
       !$omp end parallel do

       !$omp parallel do private(r) schedule(dynamic,1) if(cpd%rank .gt. 2048)
       do r1=1, cpd%rank
          do r=r1, cpd%rank
             selfsigma(r,r1,m) = ddot(N,cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1),1)
          end do
       end do
       !$omp end parallel do

    end if

    do r1=1, cpd%rank
       do r=r1+1, cpd%rank
          selfsigma(r1,r,m) = selfsigma(r,r1,m)
       end do
    end do

    call stop_timer(tid)

  end subroutine update_als_cpd



  ! once we have set up a HCPD we can start optimizing it
  subroutine optimize_hcpd(toper,thcpd, opts, ilog, symmap)

    implicit none

    class(tensor_t), intent(IN), target    :: toper ! original tensor
    class(tensor_t), intent(INOUT), target :: thcpd ! fitted tensoer
    type(coopts_t), intent(IN)             :: opts  ! program options
    integer, intent(IN)                    :: ilog  ! log unit
    type(symmap_t), pointer, intent(IN)    :: symmap(:)

    type(hcpd_t), pointer                  :: oper => Null()
    type(hcpd_t), pointer                  :: hcpd => Null()
    type(cpd_t), pointer                   :: diagoper => Null()
    type(cpd_t), pointer                   :: cpd => Null()
    type(diis_t)                           :: diis

    type(cpd_t)                            :: best_cpd
    type(hcpd_t)                           :: best_hcpd
    integer                                :: best_iter


    real(dbl), pointer                     :: opercoeff(:), fitcoeff(:)
    real(dbl)                              :: rtotgdim
    integer                                :: operrank,fitrank,newrank

    integer                                :: iiter, ierr, nmode, nline
    integer                                :: i, m, pot10, s

    logical                                :: lout, lstop
    character(len=c5)                      :: line, efile
    character(len=c5)                      :: mdl, rout, msg

    integer                                :: iit ! unit for iterations file
    integer, save                          :: tid = 0

#if defined(__GFORTRAN__) || defined(__INTEL_COMPILER)
    real(kind=real128)                     :: rms, cond, jfunc, covlp, brms, arms, best_rms
    logical, parameter                     :: lreal16 = .true.
#else
    real(dbl)                              :: rms, cond, jfunc, covlp, brms, arms, best_rms
    logical, parameter                     :: lreal16 = .false.
#endif


    mdl = "hcpd_als"
    rout = "optimize_hcpd"

    call get_timer(tid, 'ALS_main')
    call start_timer(tid)

    select type(toper)
    type is(hcpd_t)
       oper => toper
       operrank = oper%rank
       nmode = oper%nmode
       opercoeff => oper%coeff
       rtotgdim = oper%rvpotdim2
    type is(cpd_t)
       diagoper => toper
       operrank = diagoper%rank
       nmode = diagoper%nmode
       opercoeff => diagoper%coeff
       rtotgdim = diagoper%rvpotdim
    class default
       msg = "Operator object is of unknown type " &
            //"Please file a bug report."
       call stop_code(mdl,rout,msg)
    end select

    select type(thcpd)
    type is(hcpd_t)
       hcpd => thcpd
       fitrank = hcpd%rank
       fitcoeff =>  hcpd%coeff
       call copy_hcpd(hcpd, best_hcpd)
       best_iter = -1
    type is(cpd_t)
       cpd => thcpd
       fitrank = cpd%rank
       fitcoeff =>  cpd%coeff
       call copy_cpd(cpd, best_cpd)
       best_iter = -1
    class default
       msg = "Fit object is of unknown type " &
            //"Please file a bug report."
       call stop_code(mdl,rout,msg)
    end select


    ! build auxilliary quantities
    call init_hals(toper, thcpd, .true., opts%luseblas)

    if (lreal16) then
       if (opts%isolver .ne. 1) then
          nline = 118
       else
          nline = 145
       end if
    else
       if (opts%isolver .ne. 1) then
          nline = 72
       else
          nline = 87
       end if
    end if

    line = " "
    do i=1,nline
       line(i:i) = "-"
    end do

    if (opts%lcontinue .and. .not. opts%lresetiter) then
       open(newunit=iit,file=trim(opts%iterationsfile),iostat=ierr,position='append')
    else
       open(newunit=iit,file=trim(opts%iterationsfile),iostat=ierr)
    end if
    if (ierr.ne.0) then
       msg = "Cannot open file "//trim(opts%iterationsfile)
       call stop_code(mdl,rout,msg)
    end if

    lout = .true.
    iiter = opts%startiter

    write(ilog,'(a)') "Starting ALS iterations. For details see file:"
    write(ilog,'(a)') trim(opts%iterationsfile)


    ! haeder
    if (opts%isolver .eq. 1) then
       ! Cholesky
       write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond","Jfunc"
       write(*, '(a)') trim(line)
       if (.not. opts%lcontinue) then
          write(iit, 197) "#","iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond", "Jfunc"
          line(1:1) = "#"
          write(iit, '(a)') trim(line)
       end if
    else
       ! eigenvalue
       write(*, 198) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
       write(*, '(a)') trim(line)
       if (.not. opts%lcontinue) then
          write(iit, 199) "#","iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
          line(1:1) = "#"
          write(iit, '(a)') trim(line)
       end if
    end if


    ! start-error
    rms = fast_rms(opercoeff,fitcoeff,  &
         operrank,fitrank,nmode,rtotgdim)

    best_rms = rms

    covlp = qddot(fitrank,fitcoeff,fitcoeff)
    cond = sqrt(covlp / ovlp_orig_orig)
    jfunc = rms**2 + opts%alpha*covlp/rtotgdim

    if (.not. opts%lcontinue) then
       ! this is already the last entry in the file for continuation
       if (opts%isolver .eq. 1) then
          write(iit,200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
       else
          write(iit,201) iiter, rms*au2ic, rms*au2mev, rms, cond
       end if
    end if

    ! but its nice to have in screen
    if (opts%isolver .eq. 1) then
       write(*,200) iiter, rms*au2ic,rms*au2mev, rms, cond, jfunc
    else
       write(*,201) iiter, rms*au2ic, rms*au2mev, rms, cond
    end if
    call myflush(iit)

    !! init DIIS by SUDIP
    if (opts%ldiis) then
       if (associated(oper) .and. associated(hcpd)) then
          call allocate_diis_t(diis,opts%ndiis,opts%maxrank,nmode,hcpd%subdim(:nmode),.true.)
       else if (associated(diagoper) .and. associated(cpd)) then
          call allocate_diis_t(diis,opts%ndiis,opts%maxrank,nmode,cpd%subdim(:nmode),.false.)
       endif
    endif

    do iiter=opts%startiter+1,opts%maxiter

!       call strttim(ptiteration)

       do m=1, nmode

          if (iiter.eq.opts%maxiter .and. opts%isolver .eq.2) then
             write(efile,'(a,i4)') trim(opts%name)//"/als_eigenval_mode_", m
             call rmblnk(efile)
          else
             efile = ""
          end if

          if (associated(oper) .and. associated(hcpd)) then

             ! DIIS
             if ((opts%ldiis .eqv. .true.) .and. (fitrank .eq. opts%maxrank)) then
                call diis_saveold1m_hcpd(diis,hcpd%rank,m,hcpd%sph(m)%htm(:,:,:hcpd%rank))
             endif

             call halsiter1m_hcpd(oper, hcpd, m, opts%alpha, efile, opts%isolver, &
                  ierr, opts%lhermitian, opts%luseblas)
             call update_als_hcpd(oper, hcpd, m, opts%luseblas)

          else if (associated(diagoper) .and. associated(cpd)) then

             ! DIIS
             if ((opts%ldiis .eqv. .true.) .and. (fitrank .eq. opts%maxrank)) then
                call diis_saveold1m_cpd(diis,cpd%rank,m,cpd%spp(m)%htm(:,:cpd%rank))
             endif

             call halsiter1m_cpd(diagoper, cpd, m, opts%alpha, efile, opts%isolver,&
                  ierr, symmap, opts%luseblas)
             call update_als_cpd(diagoper, cpd, m, opts%luseblas)

          end if


          if (ierr .ne. 0) then
             write(msg,'(a,i6)') "ALS failed with error code", ierr
             call stop_code(mdl,rout,msg)
          end if

       end do ! end of mode => m loop



       !! For DIIS: SUDIP
       if ((opts%ldiis .eqv. .true.) .and. (fitrank .eq. opts%maxrank)) then
          !! error matrix calculation
          if (associated(oper) .and. associated(hcpd)) then
             call diis_error_hcpd(diis,hcpd,iiter)
          else if (associated(diagoper) .and. associated(cpd)) then
             call diis_error_cpd(diis,cpd,iiter)
          endif
          !!
          s = modulo(iiter,opts%ndiis)
          if (s .eq. 0) then
             print*, "DIIS extrapolation ..."
             if (associated(oper) .and. associated(hcpd)) then
                call diis_step_hcpd(diis,hcpd,oper,opts)
             else if (associated(diagoper) .and. associated(cpd)) then
                call diis_step_cpd(diis,cpd,diagoper,opts,symmap)
             endif
          endif  ! --> s=0
       endif  ! --> opts%ldiis = true


       !call stoptim(ptiteration)

       pot10 = ceiling(log10(dble(iiter)))
       if (pot10.gt.2) then
          pot10 = pot10-1
          if(pot10.ge.2 .and. int(iiter/10**pot10).lt.2) then
             i=5
          elseif(pot10.ge.2 .and. int(iiter/10**pot10).lt.5) then
             i=2
          else
             i=1
          endif
          pot10 = pot10-1
          lout = (modulo(iiter,10**pot10/i) .eq. 0)
       else
          lout = .true.
       end if

       if (lout .or. iiter .ge. opts%maxiter) then

          rms = fast_rms(opercoeff,fitcoeff,  &
               operrank,fitrank,nmode,rtotgdim)

          covlp = qddot(fitrank,fitcoeff,fitcoeff)
          cond = sqrt(covlp / ovlp_orig_orig)
          jfunc = rms**2 + opts%alpha*covlp/rtotgdim

          if (opts%isolver .eq. 1) then
             write(*,  200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
             write(iit,200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
          else
             write(*,  201) iiter, rms*au2ic, rms*au2mev, rms, cond
             write(iit,201) iiter, rms*au2ic, rms*au2mev, rms, cond
          end if
          call myflush(iit)


         if (rms .lt. best_rms) then
             if (associated(hcpd)) then
                call copy_hcpd(hcpd, best_hcpd)
             else if (associated(cpd)) then
                call copy_cpd(cpd, best_cpd)
             end if
             best_rms = rms
             best_iter = iiter
          end if


          if (rms .le. opts%deltastop) then
             lstop = .true.
             write(*,'(a)') "RMS dropped below threshold. Stopping iterations."
             write(ilog,'(a)') "RMS dropped below threshold. Stopping iterations."
             exit ! iiter-loop
          end if

          ! check stop file
          call chkstop(opts%stopfile, lstop)
          if (lstop) then
             write(*,'(a)') "Iterations stopped by stop-file"
             write(ilog,'(a)')  "Iterations stopped by stop-file"
             exit ! iiter-loop
          end if



       end if

       if (opts%lgrow .and. fitrank .lt. opts%maxrank) then
          if (modulo(iiter,opts%giter) .eq.0 .and. iiter .lt. opts%maxiter) then
             newrank = min( opts%maxrank,fitrank + opts%grank)

             write(msg,'(a,i8,a,i8,a,i9)') "Changing rank ", &
                  fitrank, " --> ", newrank, "  after iteration", iiter
             call cncntrblnk(msg)
             write(*,'(a)') trim(msg)
             write(ilog,'(a)') trim(msg)
             call myflush(ilog)

             fitrank = newrank

             if (associated(hcpd)) then
                if (opts%lhermitian) then
                   call changehrank(hcpd, newrank, opts%inittype, 2)
                else
                   call changehrank(hcpd, newrank, opts%inittype, 1)
                end if
                fitcoeff =>  hcpd%coeff
             elseif (associated(cpd)) then
                if (opts%lhermitian) then
                   call changerank(cpd, newrank, .true.,2)
                   call co_symcpd(cpd, symmap)
                else
                   call changerank(cpd, newrank, .true.,1)
                end if
                fitcoeff =>  cpd%coeff
             end if

             call init_hals(toper, thcpd, .false.,opts%luseblas)
          end if
       end if

    end do !iiter
    iiter=min(iiter,opts%maxiter)


    ! this is not guqranteed to oimprove the RMS error.
    ! it will improve the functional, though.

!!$    write(*,*) " finalizing ALS..."
!!$    write(iit,*) "# finalizing ALS..."
!!$    if (associated(hcpd)) then
!!$       call als_finalize(oper%coeff, oper%rank, hcpd%coeff, hcpd%rank, hcpd%nmode, &
!!$            opts%alpha, opts%isolver, opts%lhermitian)
!!$    else
!!$       call als_finalize(diagoper%coeff, diagoper%rank, cpd%coeff, cpd%rank, cpd%nmode, &
!!$            opts%alpha, opts%isolver, opts%lhermitian)
!!$    end if
!!$    iiter = iiter + 1
!!$
!!$    rms = fast_rms(opercoeff,fitcoeff,  &
!!$         operrank,fitrank,nmode,rtotgdim)
!!$
!!$    covlp = qddot(fitrank,fitcoeff,fitcoeff)
!!$    cond = sqrt(covlp / ovlp_orig_orig)
!!$    jfunc = rms**2 + opts%alpha*covlp/rtotgdim
!!$
!!$    if (opts%isolver .eq. 1) then
!!$       write(*,  200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
!!$       write(iit,200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
!!$    else
!!$       write(*,  201) iiter, rms*au2ic, rms*au2mev, rms, cond
!!$       write(iit,201) iiter, rms*au2ic, rms*au2mev, rms, cond
!!$    end if



    ! print the header again on screen
    line(1:1) = "-"
    write(*, '(a)') trim(line)
    if (opts%isolver .eq. 1) then
       ! Cholesky
       write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond","Jfunc"
    else
       ! eigenvalue
       write(*, 198) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
    end if
    write(*, '(a)') trim(line)

    !copy back the best cpd/hcpd we saw during the iterations:
    if (associated(hcpd)) then
       call copy_hcpd(best_hcpd, hcpd)
    else if (associated(cpd)) then
       call copy_cpd(best_cpd, cpd)
    end if
    write(*,'(/,a)') 'Taking (h)cpd with smallest RMS:'
    write(*, 296) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]"
    write(iit,'(/,a)') '# Taking (h)cpd with smallest RMS:'
    write(iit, 296) "# iiter","RMS[cm-1]","RMS[meV]","RMS[au]"
    write(*,  220) best_iter, best_rms*au2ic, best_rms*au2mev, best_rms
    write(iit,220) best_iter, best_rms*au2ic, best_rms*au2mev, best_rms

    call myflush(iit)

    close(iit)



    if (opts%lwrovlp) then
       call wr_basis_ovlp(opts%ovlpfile)
    end if
    if (opts%lwrcoeff) then
       if (associated(hcpd)) then
          call wr_fit_coeff(opts%coefffile, hcpd%coeff)
       else
          call wr_fit_coeff(opts%coefffile, cpd%coeff)
       end if
    end if


    ! free DIIS
    if (opts%ldiis) then
       call deallocate_diis_t(diis)
    endif
    call free_hals()
    call stop_timer(tid)


#if defined(__GFORTRAN__) || defined(__INTEL_COMPILER)

196 FORMAT (3x,a,17x,a,19x,a,22x,a,24x,a,22x,a)
197 FORMAT (a,2x,a,17x,a,19x,a,22x,a,24x,a,22x,a)

198 FORMAT (3x,a,17x,a,19x,a,22x,a,24x,a)
199 FORMAT (a,2x,a,17x,a,19x,a,22x,a,24x,a)

200 FORMAT (i8,2x,5(ES24.16E3,4x))
201 FORMAT (i8,2x,4(ES24.16E3,4x))

220 FORMAT (i8,2x,3(ES24.16,4x))
296 FORMAT (3x,a,17x,a,19x,a,22x,a)

#else

196 FORMAT (4x,a,6x,a,7x,a,12x,a,10x,a,10x,a)
197 FORMAT (a,2x,a,7x,a,7x,a,11x,a,10x,a,11x,a)

198 FORMAT (4x,a,6x,a,6x,a,12x,a,11x,a)
199 FORMAT (a,2x,a,7x,a,7x,a,11x,a,11x,a)

200 FORMAT (i8,2x,3(ES14.8,4x),4x,ES14.8,14x,ES14.8)
201 FORMAT (i8,2x,3(ES14.8,4x),4x,ES14.8)

220 FORMAT (i8,2x,3(ES14.8,4x))
296 FORMAT (4x,a,6x,a,6x,a,12x,a)

#endif

  end subroutine optimize_hcpd





  subroutine righthandside_hcpd(oper,hcpd,mtarget,BT,luseblas)

    implicit none
    type(hcpd_t), intent(IN)  :: oper
    type(hcpd_t), intent(IN)  :: hcpd
    integer, intent(IN)       :: mtarget
    logical, intent(IN)       :: luseblas

    real(dbl), intent(OUT)    :: BT(hcpd%subdim(mtarget),hcpd%subdim(mtarget),hcpd%rank)
    real(dbl), allocatable    :: sigmam(:,:)
    integer                   :: r,r1,m,n
    integer, save             :: tid = 0

    call get_timer(tid, 'righthandside_hcpd')
    call start_timer(tid)


    allocate(sigmam(oper%rank,hcpd%rank))

    if (mtarget .eq. 1) then


       do r1=1,hcpd%rank
          sigmam(:,r1) = crosssigma(:,r1,2)*oper%coeff(:)
       end do


       do m=3,oper%nmode
          sigmam(:,:) = sigmam(:,:)*crosssigma(:,:,m)
       end do

    else

       do r1=1,hcpd%rank
          sigmam(:,r1) = crosssigma(:,r1,1)*oper%coeff(:)
       end do

       do m=2,oper%nmode
          if (m.ne.mtarget) then
             sigmam(:,:) = sigmam(:,:)*crosssigma(:,:,m)
          end if
       end do

    end if


    if (luseblas) then

       R = oper%rank
       R1 = hcpd%rank
       N = hcpd%subdim2(mtarget)
       call dgemm('N','N',N,R1,R,1.d0,oper%sph(mtarget)%htm,N,sigmam,R, 0.d0,BT,N)

    else

       !$omp parallel do private(r) schedule(static)
       do r1=1,hcpd%rank
          BT(:,:,r1) = oper%sph(mtarget)%htm(:,:,1) * sigmam(1,r1)
          do r=2,oper%rank
             BT(:,:,r1) =  BT(:,:,r1) + sigmam(r,r1) * oper%sph(mtarget)%htm(:,:,r)
          end do
       end do
       !$omp end parallel do

    end if

    call stop_timer(tid)

  end subroutine righthandside_hcpd



  subroutine righthandside_cpd(oper,cpd,mtarget,BT, luseblas)
    implicit none
    type(cpd_t), intent(IN)       :: oper
    type(cpd_t), intent(IN)       :: cpd
    integer, intent(IN)           :: mtarget
    logical, intent(IN)           :: luseblas
    real(dbl), intent(OUT)        :: BT(cpd%subdim(mtarget),cpd%rank)

    real(dbl), allocatable        :: sigmam(:,:)
    integer                       :: r,r1,m,n
    integer, save                 :: tid = 0

    call get_timer(tid, 'righthandside_cpd')
    call start_timer(tid)


    allocate(sigmam(oper%rank,cpd%rank))

    if (mtarget .eq. 1) then

       !$omp parallel do simd collapse(2) schedule(simd:static)
       do r1=1,cpd%rank
          do r=1,oper%rank
             sigmam(r,r1) = crosssigma(r,r1,2)*oper%coeff(r)
          end do
       end do
       !$omp end parallel do simd


       do m=3,oper%nmode
          !$omp parallel do simd collapse(2) schedule(simd:static)
          do r1=1,cpd%rank
             do r=1,oper%rank
                sigmam(r,r1) = sigmam(r,r1)*crosssigma(r,r1,m)
             end do
          end do
          !$omp end parallel do simd
       end do

    else
       !$omp parallel do simd collapse(2) schedule(simd:static)
       do r1=1,cpd%rank
          do r=1,oper%rank
             sigmam(r,r1) = crosssigma(r,r1,1)*oper%coeff(r)
          end do
       end do
       !$omp end parallel do simd

       do m=2,oper%nmode
          if (m.ne.mtarget) then
             !$omp parallel do simd collapse(2) schedule(simd:static)
             do r1=1,cpd%rank
                do r=1,oper%rank
                   sigmam(r,r1) = sigmam(r,r1)*crosssigma(r,r1,m)
                end do
             end do
             !$omp end parallel do simd

          end if
       end do

    end if


    if (luseblas) then

       BT = 0.d0

       R = oper%rank
       R1 = cpd%rank
       N = cpd%subdim(mtarget)
       call dgemm('N','N',N,R1,R,1.d0,oper%spp(mtarget)%htm,N,sigmam,R, 0.d0,BT,N)

    else

       !$omp parallel do simd private(r) schedule(static)
       do r1=1,cpd%rank

          BT(:,r1) = oper%spp(mtarget)%htm(:,1) * sigmam(1,r1)

          do r=2,oper%rank
             BT(:,r1) =  BT(:,r1) + sigmam(r,r1) * oper%spp(mtarget)%htm(:,r)
          end do

       end do
       !$omp end parallel do simd

    end if

     call stop_timer(tid)
  end subroutine righthandside_cpd






  subroutine lefthandside(mtarget,nmode,rank,sigmam)
    implicit none
    integer, intent(IN)    :: mtarget, nmode,rank
    real(dbl), intent(OUT) :: sigmam(rank,rank)

    integer                :: r,r1,m
    integer, save          :: tid = 0

    call get_timer(tid, 'lefthandside')
    call start_timer(tid)


    ! this seems to not to parallelize efficiently

    if (mtarget .eq. 1) then

       sigmam(:,:) = selfsigma(:,:,2)
       do m=3,nmode
          do r1=1,rank
             do r=r1,rank
                sigmam(r,r1) = sigmam(r,r1)*selfsigma(r,r1,m)
             end do
          end do
       end do

    else

       sigmam(:,:) = selfsigma(:,:,1)
       do m=2,nmode
          if (m.ne.mtarget) then
             do r1=1,rank
                do r=r1,rank
                   sigmam(r,r1) = sigmam(r,r1)*selfsigma(r,r1,m)
                end do
             end do
          end if
       end do
    end if

    ! symmetric comletion
    do r1=1,rank
       do r=r1+1,rank
          sigmam(r1,r) = sigmam(r,r1)
       end do
    end do

    call stop_timer(tid)


  end subroutine lefthandside



  subroutine halsiter1m_hcpd(oper, hcpd, m, alpha, fname, isolver, ierr, lhermite, luseblas)
    implicit none
    type(hcpd_t), intent(IN)      :: oper
    type(hcpd_t), intent(INOUT)   :: hcpd
    integer, intent(IN)           :: m, isolver
    real(dbl), intent(IN)         :: alpha
    character(len=*), intent(IN)  :: fname
    integer, intent(OUT)          :: ierr
    logical, intent(IN)           :: lhermite, luseblas


    real(dbl), allocatable        :: B(:,:,:), BT(:,:,:), work(:,:,:), workT(:,:,:)
    real(dbl), allocatable        :: sigma(:,:)

    real(dbl)                     :: norm2, ddot
    integer                       :: subdim, subdim2, rank, r,r1
    character(len=c5)             :: rout, mdl, msg
    integer, save                 :: tid = 0

    external ddot

    call get_timer(tid, 'ALS_hcpd_iter_1m')
    call start_timer(tid)

    rank = hcpd%rank
    subdim = hcpd%subdim(m)
    subdim2 = hcpd%subdim2(m)

    allocate(BT(subdim,subdim, rank), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_hcpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate left-hand-side vectors"
       call stop_code(mdl,rout,msg)
    end if

    allocate(sigma(rank, rank), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_hcpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate right-hand-side overlap"
       call stop_code(mdl,rout,msg)
    end if


    if (lhermite) then

       allocate(work(subdim,subdim,rank), stat=ierr)
       if (ierr .ne. 0) then
          rout = "halsiter1m_hcpd"
          mdl  = "hcpd_als"
          msg  = "Could not allocate workspace"
          call stop_code(mdl,rout,msg)
       end if
       allocate(workT(subdim,subdim,rank), stat=ierr)
       if (ierr .ne. 0) then
          rout = "halsiter1m_hcpd"
          mdl  = "hcpd_als"
          msg  = "Could not allocate workspaceT"
          call stop_code(mdl,rout,msg)
       end if



       call righthandside_hcpd(oper,hcpd,m,work, luseblas)

       ! this seems to not efficiently parallelize:

       do r=1,rank
          BT(:,:,r) = work(:,:,r)
       end do

       do r=1,rank
          workT(:,:,r) = transpose(work(:,:,r))
       end do

       do r=1,rank/2
          r1 = r + rank/2
          BT(:,:,r) = BT(:,:,r) +  workT(:,:,r1)
       end do

       do r=1,rank/2
          r1 = r + rank/2
          BT(:,:,r1) = BT(:,:,r1) + workT(:,:,r)
       end do


       deallocate(work, workT)

    else

       call righthandside_hcpd(oper,hcpd,m,BT, luseblas)

    end if




    if (lhermite) then

       allocate(work(rank,rank,1), stat=ierr)
       if (ierr .ne. 0) then
          rout = "halsiter1m_hcpd"
          mdl  = "hcpd_als"
          msg  = "Could not allocate workspace1"
          call stop_code(mdl,rout,msg)
       end if

       call lefthandside(m,hcpd%nmode,rank,work(:,:,1))

       sigma(:,:) = work(:,:,1)


       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1,r) =  sigma(r1,r) + work(r1+rank/2,r+rank/2,1)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1+rank/2,r+rank/2) =  sigma(r1+rank/2,r+rank/2) + work(r1,r,1)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1+rank/2,r) =  sigma(r1+rank/2,r) + work(r1,r+rank/2,1)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1,r+rank/2) =  sigma(r1,r+rank/2) + work(r1+rank/2,r,1)
          end do
       end do

       deallocate(work)

    else
       call lefthandside(m,hcpd%nmode,rank,sigma)

    end if

    allocate(B(rank,subdim,subdim), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_hcpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate workspace B"
       call stop_code(mdl,rout,msg)
    end if


    do r=1,rank
       B(r,:,:) = BT(:,:,r)
    end do

    !  call strttim(ptsolve)
    if (isolver.eq.1) then
       do r=1,rank
          sigma(r,r) = sigma(r,r) + alpha
       end do

       call mcsolve_spd(sigma,B,rank,subdim**2,ierr)
    else if (isolver.eq.2) then
       call mcsolve_eigen(sigma,B,alpha,.false.,.true., &
            rank,subdim**2,fname,ierr)
    else
       write(*,*) "ALS: unknown solver:", isolver
       return
    end if
    ! call stoptim(ptsolve)


    if (ierr .ne. 0) then
       rout = "halsiter1m_hcpd"
       mdl  = "hcpd_als"
       write(msg,'(a,i8)')   "Solver error code:", ierr
       call stop_code(mdl,rout,msg)
    end if


    do r=1,rank
       BT(:,:,r) = B(r,:,:)
    end do


    deallocate(sigma)
    deallocate(B)


    do r=1,hcpd%rank

       norm2 = ddot(subdim2,BT(:,:,r),1,BT(:,:,r),1)

       if (abs(norm2) .gt. 0.d0) then
          hcpd%coeff(r) = sqrt(norm2)!/subdim
          hcpd%sph(m)%htm(:,:,r) =  BT(:,:,r)/hcpd%coeff(r)
       else
          hcpd%coeff(r) = 0.d0
          call random_number(hcpd%sph(m)%htm(:,:,r))
          hcpd%sph(m)%htm(:,:,r) = hcpd%sph(m)%htm(:,:,r) - 0.5d0
          norm2 = ddot(subdim2, hcpd%sph(m)%htm(:,:,r),1,hcpd%sph(m)%htm(:,:,r),1)
          hcpd%sph(m)%htm(:,:,r) =  hcpd%sph(m)%htm(:,:,r) / sqrt(norm2)

       end if
    end do

    deallocate(BT)
    call stop_timer(tid)

  end subroutine halsiter1m_hcpd








  subroutine halsiter1m_cpd(oper, cpd, m, alpha, fname, isolver, ierr, symmap, luseblas)
    implicit none
    type(cpd_t), intent(IN)            :: oper
    type(cpd_t), intent(INOUT)         :: cpd
    integer, intent(IN)                :: m, isolver
    real(dbl), intent(IN)              :: alpha
    character(len=*), intent(IN)       :: fname
    integer, intent(OUT)               :: ierr
    type(symmap_t), pointer,intent(IN) :: symmap(:)
    logical, intent(IN)                :: luseblas


    real(dbl), allocatable             :: B(:,:), BT(:,:)
    real(dbl), allocatable             :: sigma(:,:), work(:,:)

    real(dbl)                          :: norm2, dnrm2
    integer                            :: subdim, rank, r,r1,g
    character(len=c5)                  :: rout, mdl, msg
    integer, save                      :: tid = 0

    external ddot

    call get_timer(tid, 'ALS_cpd_iter_1m')
    call start_timer(tid)


    rank = cpd%rank
    subdim = cpd%subdim(m)

    allocate(BT(subdim, rank), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_cpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate BT"
       call stop_code(mdl,rout,msg)
    end if

    allocate(sigma(rank, rank), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_cpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate sigma"
       call stop_code(mdl,rout,msg)
    end if


    if (associated(symmap)) then

       allocate(work(rank,rank),stat=ierr)
       if (ierr .ne. 0) then
          rout = "halsiter1m_cpd"
          mdl  = "hcpd_als"
          msg  = "Could not allocate workspace sigma"
          call stop_code(mdl,rout,msg)
       end if

       call lefthandside(m,cpd%nmode,rank,work )

       sigma(:,:) = work

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1,r) =  sigma(r1,r) + work(r1+rank/2,r+rank/2)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1+rank/2,r+rank/2) =  sigma(r1+rank/2,r+rank/2) + work(r1,r)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1+rank/2,r) =  sigma(r1+rank/2,r) + work(r1,r+rank/2)
          end do
       end do

       do r=1,rank/2
          do r1=1,rank/2
             sigma(r1,r+rank/2) =  sigma(r1,r+rank/2) + work(r1+rank/2,r)
          end do
       end do

       deallocate(work)

       allocate(work(subdim,rank), stat=ierr)
       if (ierr .ne. 0) then
          rout = "halsiter1m_cpd"
          mdl  = "hcpd_als"
          msg  = "Could not allocate workspace BT"
          call stop_code(mdl,rout,msg)
       end if

       call righthandside_cpd(oper,cpd,m,work,luseblas)

       BT(:,:) = work(:,:)


       do r=1,rank/2
          do g=1,cpd%subdim(m)
             BT(g,r) =  BT(g,r)  + work(g,r+rank/2)*symmap(m)%symtype(g)
          end do
       end do

       do r=1,rank/2
          do g=1,cpd%subdim(m)
             BT(g,r+rank/2) =  BT(g,r+ rank/2) +  work(g,r)*symmap(m)%symtype(g)
          end do
       end do

       deallocate(work)


    else

       call righthandside_cpd(oper,cpd,m,BT, luseblas)
       call lefthandside(m,cpd%nmode,rank,sigma)

    end if


    allocate(B(rank,subdim), stat=ierr)
    if (ierr .ne. 0) then
       rout = "halsiter1m_cpd"
       mdl  = "hcpd_als"
       msg  = "Could not allocate workspace B"
       call stop_code(mdl,rout,msg)
    end if


    do r=1,rank
       B(r,:) = BT(:,r)
    end do


    if (isolver.eq.1) then
       do r=1,rank
          sigma(r,r) = sigma(r,r) + alpha
       end do

       call mcsolve_spd(sigma,B,rank,subdim,ierr)
    else if (isolver.eq.2) then
       call mcsolve_eigen(sigma,B,alpha,.false.,.true., &
            rank,subdim,fname,ierr)
    else
       write(*,*) "ALS: unknown solver:", isolver
       return
    end if
!    call stoptim(ptsolve)


    if (ierr .ne. 0) then
       rout = "halsiter1m_hcpd"
       mdl  = "hcpd_als"
       write(msg,'(a,i8)')   "Solver error code:", ierr
       call stop_code(mdl,rout,msg)
    end if

    do r=1,rank
       BT(:,r) = B(r,:)
    end do

    deallocate(sigma)
    deallocate(B)

    do r=1,cpd%rank

       norm2 = dnrm2(subdim,BT(:,r),1)

       if (abs(norm2) .gt. 0.d0) then
          cpd%coeff(r) = norm2 !/subdim
          cpd%spp(m)%htm(:,r) =  BT(:,r)/cpd%coeff(r)
       else
          cpd%coeff(r) = 0.d0
          call random_number(cpd%spp(m)%htm(:,r))
          cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r) - 0.5d0
          norm2 = dnrm2(subdim, cpd%spp(m)%htm(:,r),1)
          cpd%spp(m)%htm(:,r) =  cpd%spp(m)%htm(:,r) / norm2

       end if
    end do


    deallocate(BT)

    call stop_timer(tid)

  end subroutine halsiter1m_cpd




  subroutine co_symcpd(cpd,symmap)
    type(cpd_t), intent(INOUT)          :: cpd
    type(symmap_t), intent(IN), pointer :: symmap(:)
    integer :: r,m,g

    if (.not. associated(symmap)) return

    ! this is called after changerank
    ! so it is sufficient to only
    ! do one half of the symmetrization

    do r=1,cpd%rank/2
       cpd%coeff(r+cpd%rank/2) = cpd%coeff(r)
       do m=1,cpd%nmode
          do g=1,cpd%subdim(m)
             cpd%spp(m)%htm(g,r+cpd%rank/2) = cpd%spp(m)%htm(g,r)*symmap(m)%symtype(g)
          end do
       end do
    end do


  end subroutine co_symcpd


  subroutine wr_basis_ovlp(ovlpfile)
    character(len=*), intent(in) :: ovlpfile
    integer                      :: r,r1,rank,m,nmode,lun
    real(dbl), allocatable       :: sigma(:,:)

    rank = size(selfsigma, dim=1)
    nmode = size(selfsigma, dim=3)

    allocate(sigma(rank,rank))
    sigma = selfsigma(:,:,1)
    do m=2,nmode
       sigma = sigma *selfsigma(:,:,m)
    end do

    open(newunit=lun,file=trim(ovlpfile))
    write(lun,'(a)') "# r,r1, ovlp(r,r1)"
    do r=1,rank
       do r1=1,rank
          write(lun,'(2i8,es18.10)') r1,r,sigma(r1,r)
       end do
       write(lun,'(a)') ""
    end do
    close(lun)

  end subroutine wr_basis_ovlp


  subroutine wr_fit_coeff(coefffile, coeff)
    character(len=*), intent(in) :: coefffile
    real(dbl), intent(IN)        :: coeff(:)
    integer                      :: r,rank,lun

    rank = size(coeff)
    open(newunit=lun,file=trim(coefffile))
    write(lun,'(a)') "# r, coeff(r)"
    do r=1,rank
       write(lun,'(i8,es18.10)') r,coeff(r)
    end do
    close(lun)

  end subroutine wr_fit_coeff



  subroutine als_finalize(opercoeff, operrank, cpcoeff, cprank, nmode,alpha, isolver, lhermite)
    implicit none
    integer, intent(IN)      :: operrank, cprank, nmode, isolver
    real(dbl), intent(IN)    :: opercoeff(operrank)
    real(dbl), intent(INOUT) :: cpcoeff(cprank)
    real(dbl), intent(IN)    :: alpha
    logical, intent(IN)      :: lhermite


    real(dbl), allocatable   :: B(:), work(:), work2(:,:)
    real(dbl), allocatable   :: sigma(:,:)

    integer                  :: m,r,r1, ierr
    character(len=c5)        :: rout, mdl, msg
    integer, save            :: tid = 0
    character(len=c1)        :: fname = ""

    call get_timer(tid, 'ALS_finalize')
    call start_timer(tid)

    allocate(B(cprank))
    allocate(sigma(cprank, operrank))

    sigma(:,:) = transpose(crosssigma(:,:,1))
    do m=2,nmode
       sigma(:,:) = sigma(:,:)*transpose(crosssigma(:,:,m))
    end do

    if (lhermite) then

       allocate(work(cprank))

       work = matmul(sigma,opercoeff)
       B = work

       do r=1,cprank/2
          B(r) =  B(r) + work(r+cprank/2)
       end do
       do r=1,cprank/2
          B(r+cprank/2) =  B(r+cprank/2) + work(r)
       end do

       deallocate(work)

    else

       B = matmul(sigma,opercoeff)

    end if

    deallocate(sigma)
    allocate(sigma(cprank,cprank))

    sigma(:,:) = selfsigma(:,:,1)
    do m=2,nmode
       do r1=1,cprank
          do r=r1,cprank
             sigma(r,r1) = sigma(r,r1)*selfsigma(r,r1,m)
          end do
       end do
    end do

    ! symmetric comletion
    do r1=1,cprank
       do r=r1+1,cprank
          sigma(r1,r) = sigma(r,r1)
       end do
    end do


    if (lhermite) then

       allocate(work2(cprank,cprank))
       work2(:,:) = sigma

       do r=1,cprank/2
          do r1=1,cprank/2
             sigma(r1,r) =  sigma(r1,r) + work2(r1+cprank/2,r+cprank/2)
          end do
       end do

       do r=1,cprank/2
          do r1=1,cprank/2
             sigma(r1+cprank/2,r+cprank/2) =  sigma(r1+cprank/2,r+cprank/2) + work2(r1,r)
          end do
       end do

       do r=1,cprank/2
          do r1=1,cprank/2
             sigma(r1+cprank/2,r) =  sigma(r1+cprank/2,r) + work2(r1,r+cprank/2)
          end do
       end do

       do r=1,cprank/2
          do r1=1,cprank/2
             sigma(r1,r+cprank/2) =  sigma(r1,r+cprank/2) + work2(r1+cprank/2,r)
          end do
       end do

       deallocate(work2)

    end if


    if (isolver.eq.1) then
       do r=1,cprank
          sigma(r,r) = sigma(r,r) + alpha
       end do

       call mcsolve_spd(sigma,B,cprank,1,ierr)
    else if (isolver.eq.2) then
       call mcsolve_eigen(sigma,B,alpha,.false.,.true., &
            cprank,1,fname,ierr)
    else
       write(*,*) "ALS: unknown solver:", isolver
       return
    end if


    if (ierr .ne. 0) then
       rout = "als_finalize_hcpd"
       mdl  = "hcpd_als"
       write(msg,'(a,i8)')   "Solver error code:", ierr
       call stop_code(mdl,rout,msg)
    end if


    cpcoeff(:) = B


    call stop_timer(tid)

  end subroutine als_finalize




  !DIIS routines below

  subroutine diis_error_hcpd(diis,hcpd,iiter)
    implicit none
    type(diis_t), intent(INOUT)   :: diis
    type(hcpd_t), intent(IN)      :: hcpd
    integer, intent(IN)           :: iiter
    integer                       :: s,m
    double precision, allocatable :: temp(:)
    s = modulo(iiter,diis%nspace)
    if(s.eq.0) s = diis%nspace
    do m=1,diis%nmode
       allocate(temp(diis%subdim2(m)*hcpd%rank))
       !temp = reshape(hcpd%sph(m)%htm(:diis%subdim(m),:diis%subdim(m),:hcpd%rank),(/diis%subdim2(m)*hcpd%rank/))
       call diis_pack_hcpd(hcpd%sph(m)%htm(:hcpd%subdim(m),:hcpd%subdim(m),:hcpd%rank),diis%subdim(m),hcpd%rank,temp)
       diis%spt(m)%espo(s,:diis%subdim2(m)*hcpd%rank) = temp(:)-diis%spt(m)%ospo(:diis%subdim2(m)*hcpd%rank)
       diis%spt(m)%aspo(s,:diis%subdim2(m)*hcpd%rank) = temp(:)
       deallocate(temp)
    enddo
  end subroutine diis_error_hcpd

  subroutine diis_error_cpd(diis,cpd,iiter)
    implicit none
    type(diis_t), intent(INOUT)   :: diis
    type(cpd_t), intent(IN)       :: cpd
    integer, intent(IN)           :: iiter
    integer                       :: s,m
    double precision, allocatable :: temp(:)
    s = modulo(iiter,diis%nspace)
    if(s.eq.0) s = diis%nspace
    do m=1,diis%nmode
       allocate(temp(diis%subdim(m)*cpd%rank))
       !temp = reshape(cpd%spp(m)%htm(:diis%subdim(m),:cpd%rank),(/diis%subdim(m)*cpd%rank/))
       call diis_pack_cpd(cpd%spp(m)%htm(:diis%subdim(m),:cpd%rank),diis%subdim(m),cpd%rank,temp)
       diis%spt(m)%espo(s,:diis%subdim(m)*cpd%rank) = temp(:)-diis%spt(m)%ospo(:diis%subdim(m)*cpd%rank)
       diis%spt(m)%aspo(s,:diis%subdim(m)*cpd%rank) = temp(:)
       deallocate(temp)
    enddo
  end subroutine diis_error_cpd

  subroutine diis_step_hcpd(diis,hcpd,oper,opts)
    implicit none
    type(diis_t), intent(INOUT)      :: diis
    type(hcpd_t), intent(INOUT)      :: hcpd
    type(hcpd_t), intent(IN)         :: oper
    type(coopts_t), intent(IN)       :: opts
    integer           :: m,ierr
    double precision  :: brms,arms
    double precision  :: ocoef(hcpd%rank)

    !! stores coeff in case DIIS fails
    ocoef(:) = hcpd%coeff(:hcpd%rank)
    !! error before DIIS extrapolation
    brms = fast_rms(oper%coeff,hcpd%coeff,oper%rank, &
         hcpd%rank,oper%nmode,oper%rvpotdim2)
    !! DIIS extrapolation of all mode
    call diis_extrapolation_allmode(diis,hcpd%rank)
    !! Normalize SPO afte DIIS
    call afterdiis_normalization(diis,hcpd%rank)
    !! !! Save to hcpd
    do m=1,diis%nmode
       call diis_retnew1m_hcpd(diis,hcpd%rank,m,hcpd%sph(m)%htm(:hcpd%subdim(m),:hcpd%subdim(m),:hcpd%rank))
    enddo
    !! Update before taking one more steps
    do m=1,diis%nmode
       call update_als_hcpd(oper, hcpd, m, opts%luseblas)
    enddo
!!!! Calculate new coeff
    !!call als_finalize(oper%coeff, oper%rank, hcpd%coeff, hcpd%rank, hcpd%nmode, &
    !!                  opts%alpha, opts%isolver, opts%lhermitian)
    !! One more steps
    do m=1,hcpd%nmode
       call halsiter1m_hcpd(oper, hcpd, m, opts%alpha, "", opts%isolver, &
            ierr, opts%lhermitian, opts%luseblas)
       call update_als_hcpd(oper, hcpd, m, opts%luseblas)
    enddo
    !! error after DIIS extrapolation
    arms = fast_rms(oper%coeff,hcpd%coeff,oper%rank, &
         hcpd%rank,oper%nmode,oper%rvpotdim2)
    !! If DIIS fails, fall back to old
    if (brms .le. arms) then
       print*, "   but DIIS fails to improve the SPOs. So fall back to old SPOs ..."
       call diis_fallback_hcpd(diis,hcpd,oper,opts,ocoef)
    endif

  end subroutine diis_step_hcpd


  subroutine diis_step_cpd(diis,cpd,diagoper,opts,symmap)
    implicit none
    type(diis_t), intent(INOUT)      :: diis
    type(cpd_t), intent(INOUT)       :: cpd
    type(cpd_t), intent(IN)          :: diagoper
    type(coopts_t), intent(IN)       :: opts
    type(symmap_t), pointer, intent(IN) :: symmap(:)
    integer           :: m,ierr
    double precision  :: brms,arms
    double precision  :: ocoef(cpd%rank)

    !! stores coeff in case DIIS fails
    ocoef(:) = cpd%coeff(:cpd%rank)
    !! error before DIIS extrapolation
    brms = fast_rms(diagoper%coeff,cpd%coeff,diagoper%rank, &
         cpd%rank,diagoper%nmode,diagoper%rvpotdim)
    !! DIIS extrapolation of all mode
    call diis_extrapolation_allmode(diis,cpd%rank)
    !! Normalize SPP afte DIIS
    call afterdiis_normalization(diis,cpd%rank)
    !! Save to cpd
    do m=1,diis%nmode
       call diis_retnew1m_cpd(diis,cpd%rank,m,cpd%spp(m)%htm(:cpd%subdim(m),:cpd%rank))
    enddo
    !! Update before taking one more steps
    do m=1,cpd%nmode
       call update_als_cpd(diagoper, cpd, m, opts%luseblas)
    enddo
!!! Calculate new coeff
    !call als_finalize(diagoper%coeff, diagoper%rank, cpd%coeff, cpd%rank, cpd%nmode, &
    !                  opts%alpha, opts%isolver, opts%lhermitian)
    do m=1,cpd%nmode
       call halsiter1m_cpd(diagoper, cpd, m, opts%alpha, "", opts%isolver,&
            ierr, symmap, opts%luseblas)
       call update_als_cpd(diagoper, cpd, m, opts%luseblas)
    enddo
    !! error after DIIS extrapolation
    arms = fast_rms(diagoper%coeff,cpd%coeff,diagoper%rank, &
         cpd%rank,diagoper%nmode,diagoper%rvpotdim)
    !! If DIIS fails, fall back to old
    if (brms .le. arms) then
       print*, "   but DIIS fails to improve the SPOs. So fall back to old SPOs ..."
       call diis_fallback_cpd(diis,cpd,diagoper,opts,ocoef)
    endif

  end subroutine diis_step_cpd


  subroutine diis_fallback_hcpd(diis,hcpd,oper,opts,ocoef)
    type(diis_t), intent(INOUT)      :: diis
    type(hcpd_t), intent(INOUT)      :: hcpd
    type(hcpd_t), intent(IN)         :: oper
    type(coopts_t), intent(IN)       :: opts
    double precision, intent(IN)     :: ocoef(hcpd%rank)
    integer           :: m
    hcpd%coeff(:hcpd%rank) = ocoef(:hcpd%rank)
    do m=1,diis%nmode
       !hcpd%sph(m)%htm(:hcpd%subdim(m),:hcpd%subdim(m),:hcpd%rank) = &
       !   reshape(diis%spt(m)%aspo(diis%nspace,:hcpd%subdim2(m)*hcpd%rank), &
       !           (/hcpd%subdim(m),hcpd%subdim(m),hcpd%rank/))
       call diis_unpack_hcpd(diis%spt(m)%aspo(diis%nspace,:hcpd%subdim2(m)*hcpd%rank), &
            hcpd%subdim(m),hcpd%rank, &
            hcpd%sph(m)%htm(:hcpd%subdim(m),:hcpd%subdim(m),:hcpd%rank))
       call update_als_hcpd(oper, hcpd, m, opts%luseblas)
    enddo
  end subroutine diis_fallback_hcpd

  subroutine diis_fallback_cpd(diis,cpd,diagoper,opts,ocoef)
    type(diis_t), intent(INOUT)      :: diis
    type(cpd_t), intent(INOUT)       :: cpd
    type(cpd_t), intent(IN)          :: diagoper
    type(coopts_t), intent(IN)       :: opts
    double precision, intent(IN)     :: ocoef(cpd%rank)
    integer           :: m
    cpd%coeff(:cpd%rank) = ocoef(:cpd%rank)
    do m=1,diis%nmode
       !cpd%spp(m)%htm(:cpd%subdim(m),:cpd%rank) = &
       !   reshape(diis%spt(m)%aspo(diis%nspace,:cpd%subdim(m)*cpd%rank), &
       !           (/cpd%subdim(m),cpd%rank/))
       call diis_unpack_cpd(diis%spt(m)%aspo(diis%nspace,:cpd%subdim(m)*cpd%rank), &
            cpd%subdim(m),cpd%rank, &
            cpd%spp(m)%htm(:cpd%subdim(m),:cpd%rank))
       call update_als_cpd(diagoper, cpd, m, opts%luseblas)
    enddo
  end subroutine diis_fallback_cpd


end module hals_m


