!
! contents:
!  - rdpes2cpd: read a PES file into a cpd format
!  - rdoper2hcpd: read an OPER file into a hcpd format
!  - chkimplemented: only allow certain types of operators
!                    (real diagonal and real matrices)




! Read MCTDH PES file into a cpd type.
! All operators are expected to be diagonal.

subroutine rdpes2cpd(cpd,dvr,ham,nham,filename,loglun)
  use cpdtype_m
  use dvrtype_m
  use global
  use globmemmod, only: mrhops
  use operdef
  use memhopsmod
  use griddatmod
  use channels
  use logdat
  use hphimod, only: hphi1mk, hunphi1ms
  use rdopermod

  implicit none

  type(cpd_t), intent(INOUT)    :: cpd
  type(dvr_t), intent(IN)       :: dvr
  character(len=*), intent(IN)  :: filename, ham
  integer, intent(OUT)          :: nham, loglun
  double precision, allocatable :: hops(:)
  integer                       :: chkdvr,chkgrd, ierr
  logical                       :: lerr
  logical, allocatable          :: loned(:)
  double complex, allocatable   :: psi(:), hpsi(:), workc(:,:)
  integer                       :: hktype, nuncorr, nterms,ndiag,nmat


  ! dummy: these are not used at the moment:
  double complex    :: hc(1)
  integer           :: hi(1)
  logical           :: hl(1)

  ! locals
  integer           :: m,n,s,k1,k2,k,f,ff,r
  double precision  :: time


  external ddot

  !IMPORTANT: only one state
  s = 1
  !IMPORTANT: time independent
  time = 0.d0

  call default

  ilog = loglun
  operfile=trim(filename)
  open(unit=ioper,file=trim(filename),form='unformatted', &
       status='old',iostat=ierr)
  if (ierr .ne. 0) then
     routine = "rdpes2cpd"
     message = "Unable to open file: "//trim(operfile)
     call errormsg
  end if

  chkdvr=1
  chkgrd=1
  call operinfo(lerr,chkdvr,chkgrd)
  if (lerr)  then
     routine = "rdpes2cpd"
     message = "Error reading oper-info from: "//trim(operfile)
     call errormsg
  end if

  call memhops()

  allocate(hops(hopsdim), stat=ierr)
  if (ierr .ne. 0) then
     routine = "rdpes2cpd"
     message = "Unable to allocate memory for: "//trim(operfile)
     call errormsg
  end if
  close(ioper)
  call rdoper(hops,chkdvr,chkgrd)
  close(ioper)


  nham = -1
  do n=1,hamdim
     write(ilog,'(i4,a)') n, " :  "//hamlab(n)
     if (hamlab(n) .eq. ham) nham=n
  enddo
  write(ilog,'(a)') ''
  call myflush(ilog)

  if (nham .eq. -1) then
     routine = "rdpes2cpd"
     message = "Hamiltonian "//trim(ham)//" not found. See log for available Hamiltonians."
     call errormsg
  else
     write(ilog,'(a,i4,a)') "Using Hamiltonian ",nham,": "//trim(hamlab(nham))
     write(ilog,'(a)') ''
     write(*,'(a,i4,a)') "Using Hamiltonian ",nham,": "//trim(hamlab(nham))
  end if
  call myflush(ilog)

  call chkimplemented(nham,ndiag,nmat)

  nuncorr = 0

  write(ilog,'(a)') 'Operator info:'
  do m=1,nmode
     nuncorr = nuncorr + khzahl(m,s,nham)
  end do
  nterms = nuncorr + kzahl(nham)

  allocate(potdofmap(2,ndof))
  do m=1,nmode
     ff=nspfdof(m)
     do n=1,ff
        f = spfdof(n,m)
        potdofmap(1,f) = m
        potdofmap(2,f) = n
     end do
  end do

  allocate(loned(ndof))
  loned(:) = .false.

  call allocate_cpd_t(cpd,nterms,ndof,nmode,gdim,dofspf,potdofmap,modelabel,dvr%ort,ortdim, loned)
  deallocate(loned)


  do m=1,nmode
     cpd%spp(m)%htm(:,:) = 1.d0
  end do

  cpd%coeff(:) = 0.d0

  hktype = 1 !no CAPs
  r = 0
  mrhops = 1

  ! uncorrelated
  do m=1, nmode
     if(khzahl(m,s,nham) .ne. 0) then

        r = r+1


        allocate(psi(cpd%subdim(m)))
        psi = 1.d0


        allocate(hpsi(cpd%subdim(m)))
        hpsi = 0.d0
        allocate(workc(cpd%subdim(m), 1))


        call hunphi1ms (psi,hpsi,m,s,1,cpd%subdim(m),hktype,nham)


        cpd%spp(m)%htm(:,r) = dble(hpsi(:))

        ! hunphi1ms multiplies coeff into psi
        cpd%coeff(r)= 1.d0

        deallocate(psi, hpsi, workc)

     end if
  end do

  ! correlated

  k1=zham(nham)
  k2=zham(nham)+kzahl(nham)-1

  !$omp parallel do private(m,r,psi,hpsi)
  do k=k1,k2

     r = nuncorr + k
     cpd%coeff(r)= dble(koeff(k))

     do m=1, nmode

        allocate(psi(cpd%subdim(m)))
        psi = 1.d0

        allocate(hpsi(cpd%subdim(m)))
        hpsi = 0.d0

        call hphi1mk(psi,hpsi,m,k,1,cpd%subdim(m))

        cpd%spp(m)%htm(:,r) = dble(hpsi)

        deallocate(psi, hpsi)

     end do
  end do
  !$omp end parallel do

  cpd%initialized = .true.

end subroutine rdpes2cpd






! Read MCTDH SOP oper into a hcpd type.
! All operators are transformed into a matrix representation
! even if they are diagonal or of other form (e.g.,fft)
! returns number of diagonal and matrix operators ( ndiag, nmat)
subroutine rdoper2hcpd(oper,dvr,ham,nham,filename,loglun, ndiag, nmat)
  use hcpdtype_m
  use dvrtype_m
  use global
  use globmemmod, only: mrhops
  use operdef
  use memhopsmod
  use griddatmod
  use channels
  use logdat
  use hphimod, only: hphi1mk, hunphi1ms
  use rdopermod

  use op1lib, only: unitqxz, zeromxz, unitqxd

  implicit none

  type(hcpd_t), intent(INOUT)   :: oper
  type(dvr_t), intent(IN)       :: dvr
  character(len=*), intent(IN)  :: filename, ham
  integer, intent(IN)           :: loglun
  integer, intent(OUT)          :: nham, ndiag, nmat
  double precision, allocatable :: hops(:)
  integer                       :: chkdvr,chkgrd, ierr
  logical                       :: lerr
  logical, allocatable          :: loned(:)
  double complex, allocatable   :: psi(:,:), hpsi(:,:), workc(:,:)
  integer                       :: hktype, nuncorr, nterms


  ! locals
  integer           :: m,n,s,k1,k2,k,f,ff,r
  double precision  :: time

  external ddot

  !IMPORTANT: only one state
  s = 1
  !IMPORTANT: time independent
  time = 0.d0

  ! counters for types
  ndiag = 0
  nmat  = 0
  call default

  ilog = loglun
  operfile=trim(filename)
  open(unit=ioper,file=trim(filename),form='unformatted', &
       status='old',iostat=ierr)
  if (ierr .ne. 0) then
     routine = "rdmctdhoper"
     message = "Unable to open file: "//trim(operfile)
     call errormsg
  end if

  chkdvr=1
  chkgrd=1
  call operinfo(lerr,chkdvr,chkgrd)
  if (lerr)  then
     routine = "rdmctdhoper"
     message = "Error reading oper-info from: "//trim(operfile)
     call errormsg
  end if

  call memhops()

  allocate(hops(hopsdim), stat=ierr)
  if (ierr .ne. 0) then
     routine = "rdmctdhoper"
     message = "Unable to allocate memory for: "//trim(operfile)
     call errormsg
  end if
  close(ioper)
  call rdoper(hops,chkdvr,chkgrd)
  close(ioper)


  nham = -1
  do n=1,hamdim
     write(ilog,'(i4,a)') n, " :  "//hamlab(n)
     if (hamlab(n) .eq. ham) nham=n
  enddo
  write(ilog,'(a)') ''
  call myflush(ilog)

  if (nham .eq. -1) then
     routine = "rdmctdhoper"
     message = "Hamiltonian "//trim(ham)//" not found. See log for available Hamiltonians."
     call errormsg
  else
     write(ilog,'(a,i4,a)') "Using Hamiltonian ",nham,": "//trim(hamlab(nham))
     write(ilog,'(a)') ''
     write(*,'(a,i4,a)') "Using Hamiltonian ",nham,": "//trim(hamlab(nham))
  end if
  call myflush(ilog)

  call chkimplemented(nham,ndiag,nmat)

  nuncorr = 0

  write(ilog,'(a)') 'Operator info:'
  do m=1,nmode
     nuncorr = nuncorr + khzahl(m,s,nham)
  end do
  nterms = nuncorr + kzahl(nham)

  allocate(potdofmap(2,ndof))
  do m=1,nmode
     ff=nspfdof(m)
     do n=1,ff
        f = spfdof(n,m)
        potdofmap(1,f) = m
        potdofmap(2,f) = n
     end do
  end do

  allocate(loned(ndof))
  loned(:) = .false.

  call allocate_hcpd_t(oper,nterms,ndof,nmode,gdim,dofspf,potdofmap,modelabel,dvr%ort,ortdim, loned)
  deallocate(loned)


  do m=1,nmode
     do r=1,oper%rank
        call unitqxd(oper%sph(m)%htm(:,:,r), oper%subdim(m))
     end do
  end do
  do r=1,oper%rank
     oper%coeff(r) = 0.d0
  end do


  hktype = 1 !no CAPs
  r = 0
  mrhops = 1

  ! uncorrelated
  do m=1, nmode
     if(khzahl(m,s,nham) .ne. 0) then

        r = r+1


        allocate(psi(oper%subdim(m), oper%subdim(m)))
        call unitqxz(psi, oper%subdim(m))


        allocate(hpsi(oper%subdim(m), oper%subdim(m)))
        call zeromxz(hpsi, oper%subdim(m),oper%subdim(m))
        allocate(workc(oper%subdim(m), oper%subdim(m)))


        call hunphi1ms (psi,hpsi,m,s,oper%subdim(m),oper%subdim(m), &
             hktype,nham)


        oper%sph(m)%htm(:,:,r) = dble(hpsi(:,:))

        ! hunphi1ms multiplies coeff into psi
        oper%coeff(r)= 1.d0

        deallocate(psi, hpsi, workc)

     end if
  end do

  ! correlated

  k1=zham(nham)
  k2=zham(nham)+kzahl(nham)-1

  !$omp parallel do private(m,r,psi,hpsi)
  do k=k1,k2

     r = nuncorr + k
     oper%coeff(r)= dble(koeff(k))

     do m=1, nmode

        allocate(psi(oper%subdim(m), oper%subdim(m)))
        call unitqxz(psi, oper%subdim(m))

        allocate(hpsi(oper%subdim(m), oper%subdim(m)))

        call hphi1mk(psi,hpsi,m,k,oper%subdim(m),oper%subdim(m))

        oper%sph(m)%htm(:,:,r) = dble(hpsi)

        deallocate(psi, hpsi)

     end do
  end do
  !$omp end parallel do

  oper%initialized = .true.
end subroutine rdoper2hcpd













subroutine chkimplemented(nham,ndiag,nmat)
  use global
  use griddatmod
  use operdef
  use logdat

  implicit none

  integer, intent(IN)  :: nham
  integer, intent(OUT) :: ndiag,nmat
  integer              :: s,k,k1,k2,h,ht,f,m,n
  integer, parameter   :: maxerr=500
  logical              :: lerr(-1:maxerr) ! htmtypes


  lerr = .false.
  ndiag = 0
  nmat = 0

  ! check for states
  if (nstate .ne. 1) then
     message = "Compacting operators for electronic" &
          //" states is not implemented."
     routine = "rdmctdhoper"
     call errormsg
  end if

  !check for not implemented operators
  do m = 1,nmode
     if(m.ne.mpb) then
        s=1
        k1=zhun(m,s,nham)
        k2=zhun(m,s,nham)+khzahl(m,s,nham)-1
        do k=k1,k2
           n=1
           f=spfdof(n,m)
           h=hamilton(f,k)

           if (h .lt. 0) then
              lerr(0) =  .true.
              nmat = nmat + 1
           else
              ht = htmtype(h)
              lerr(ht) = .true.
              if ( ht .eq. 0 .or. &
                   ht .eq. 1 .or. &
                   ht .eq. 2  &
              ) then
              ndiag = ndiag + 1
           else
              nmat = nmat + 1
           end if

           end if
        end do

        k1=zham(nham)
        k2=zham(nham)+kzahl(nham)-1
        do k = k1,k2
           n = 1
           f = spfdof(n,m)
           h = hamilton(f,k)
           if (h .lt. 0) then
              lerr(0) =  .true.
              nmat = nmat + 1
           else
              ht = htmtype(h)
              lerr(ht) = .true.
              if ( ht .eq. 0 .or. &
                   ht .eq. 1 .or. &
                   ht .eq. 2  &
                   ) then
                 ndiag = ndiag + 1
              else
                 nmat = nmat + 1
              end if
           end if
        end do
     end if
  end do

  ! implemented htmtypes:
  lerr(0) = .false.
  lerr(1) = .false.
  lerr(2) = .false.
  lerr(8) = .false.
  lerr(12) = .false.
  lerr(25) = .false.

  if (lerr(-1)) then
     write(*,'(a)') "Reading LMR operators is not implemented"
  end if
  do ht=1,maxerr
     if (lerr(ht)) then
        write(*,'(a,i4)') "* htmtype is not implemented: ", ht
     end if
  end do

  if (any(lerr)) then
     routine = "oper2hcpd.f90: ckhimplemented"
     message = "Found unsupported htmtypes in oper. Feel free to implement."
     call errormsg
  end if


end subroutine chkimplemented
