! find a minimal basis of a CPD or a HCPD object


module svdbasis_m

   use decimal

  implicit none

  integer, parameter :: HTM_TYPE_NONE = 0
  integer, parameter :: HTM_TYPE_HERMITIAN = 1
  integer, parameter :: HTM_TYPE_ANTI_HERMITIAN = -1


  type :: symmap_t
     integer                :: gdim
     integer, allocatable   :: symtype(:)
  end type symmap_t


  type :: svdbasis_t
     real(dbl), allocatable :: basis(:,:)
     real(dbl), allocatable :: sigma(:)
     integer, allocatable   :: htype(:)
     integer                :: nherm, naherm, nother
     integer                :: subdim, nbasis
     real(dbl)              :: l2err
  end type svdbasis_t


contains


  subroutine dispose_symmap_t(symmap)
    implicit none
    type(symmap_t), intent(INOUT) :: symmap

    if (allocated(symmap%symtype)) deallocate(symmap%symtype)
    symmap%gdim = 0

  end subroutine dispose_symmap_t


  subroutine dispose_svd_basis_t(svdbasis)
    implicit none
    type(svdbasis_t), intent(INOUT) :: svdbasis

    if (allocated(svdbasis%basis)) deallocate(svdbasis%basis)
    if (allocated(svdbasis%sigma)) deallocate(svdbasis%sigma)
    if (allocated(svdbasis%htype)) deallocate(svdbasis%htype)
    svdbasis%nbasis = 0
    svdbasis%nherm = 0
    svdbasis%naherm = 0
    svdbasis%nother = 0
    svdbasis%subdim = 0
    svdbasis%l2err = 0.d0

  end subroutine dispose_svd_basis_t

end module svdbasis_m

module svdbasis_s

  use global
  use decimal
  use svdbasis_m
  use tensortype
  use hcpdtype_m
  use cpdtype_m
  use cpdtype_s
  use hcpdtype_s
  use timing

implicit none

contains

  subroutine svd_basis_1m_from_sop(htm, rank, basism, lhermitian, svderr, lsvdabs)
    class(htm_t), intent(IN), target    :: htm     ! CPD or HCPD basis object
    integer, intent(IN)                 :: rank    ! mode to find basis for
    type(svdbasis_t), intent(INOUT)     :: basism  ! object containing the
                                                   ! basis as vectors
    real(dbl), intent(IN)               :: svderr  ! SVD cut-off

    type(sph_t), pointer                :: sph => Null()
    type(spp_t), pointer                :: spp => Null()
    logical,intent(IN), optional        :: lhermitian
    logical,intent(IN), optional        :: lsvdabs

    real(dbl), allocatable :: hmat(:,:), work(:,:)
    type(svdbasis_t)                    :: hbasis, ahbasis

    integer :: basdim, r, hrank, ahrank
    logical :: labs, lherm



    labs = .false.
    if (present(lsvdabs)) labs = lsvdabs
    lherm = .false.
    if (present(lhermitian)) lherm = lhermitian

    call dispose_svd_basis_t(basism)

    select type (htm)
    type is (sph_t)
       sph => htm
    type is (spp_t)
       spp => htm
       ! no need for special hermitian treatments
       ! real diagonal operators are all hermitian
       lherm = .false.
    end select

    if (associated(sph)) then
       basdim = size(sph%htm,1) * size(sph%htm,2)
    elseif (associated(spp)) then
       basdim = size(spp%htm,1)
    end if

    allocate(hmat(basdim,rank))

    if (lherm) then
       ! here we only have matrices

       allocate(work( size(sph%htm,1),  size(sph%htm,2) ))

       ! hermitian part
       hrank = 0
       do r=1,rank
          work = sph%htm(:,:,r) + transpose(sph%htm(:,:,r))
          if (any(abs(work) .gt. 1.d-8)) then
             hrank = hrank + 1
             hmat(:,hrank) = reshape(work,(/basdim/))/2.d0
          end if
       end do

 !      print*, "Number of Hermitian SPO-components in original operator:", hrank,'/',rank

       call svd_basis_1m_from_mat(hmat, basdim, hrank, hbasis, svderr, labs)

       allocate( hbasis%htype(hbasis%nbasis) )
       hbasis%htype(:) = HTM_TYPE_HERMITIAN

       ! anti-hermitian part
       ahrank = 0
       do r=1,rank
          work = sph%htm(:,:,r) - transpose(sph%htm(:,:,r))
          hmat(:,r) = reshape(work,(/basdim/))/2.d0
          if (any(abs(work) .gt. 1.d-8)) then
             ahrank = ahrank + 1
             hmat(:,ahrank) = reshape(work,(/basdim/))/2.d0
          end if
       end do

 !      print*, "Number of anti-Hermitian SPO-components in original operator:", ahrank,'/',rank

       call svd_basis_1m_from_mat(hmat, basdim, ahrank, ahbasis, svderr, labs)
       allocate( ahbasis%htype(ahbasis%nbasis) )

       ahbasis%htype(:) = HTM_TYPE_ANTI_HERMITIAN

       deallocate(work)

       ! join the two basis sets into one
       basism%nbasis = hbasis%nbasis + ahbasis%nbasis
       basism%nherm =  hbasis%nbasis
       basism%naherm = ahbasis%nbasis
       basism%nother = 0
       basism%subdim = basdim
       basism%l2err = hbasis%l2err + ahbasis%l2err
       basism%subdim = hbasis%subdim

       allocate(basism%basis(basdim,basism%nbasis))
       allocate(basism%htype(basism%nbasis))
       allocate(basism%sigma(basism%nbasis))

       basism%basis(:,1:hbasis%nbasis) = hbasis%basis(:,:)
       basism%basis(:,1+hbasis%nbasis:basism%nbasis) = ahbasis%basis(:,:)

       basism%sigma(1:hbasis%nbasis) = hbasis%sigma(:)
       basism%sigma(1+hbasis%nbasis:basism%nbasis) = ahbasis%sigma(:)

       basism%htype(1:hbasis%nbasis) = hbasis%htype(:)
       basism%htype(1+hbasis%nbasis:basism%nbasis) = ahbasis%htype(:)

       call dispose_svd_basis_t(hbasis)
       call dispose_svd_basis_t(ahbasis)

    else
       if (associated(sph)) then
          hmat = reshape(sph%htm,  (/basdim, rank/))
       elseif (associated(spp)) then
          hmat = reshape(spp%htm,  (/basdim, rank/))
       end if

       call svd_basis_1m_from_mat(hmat, basdim, rank, basism, svderr, labs)

    end if

  end subroutine svd_basis_1m_from_sop




  subroutine svd_basis_1m_from_mat(mat, subdim, rank, basism, svderr, lsvdabs)
    implicit none
    integer, intent(IN)                 :: subdim, rank
    ! Matrix containing vectors (operators) to transform:
    real(dbl), intent(IN)               :: mat(subdim,rank)
    ! object containing the basis as vectors
    type(svdbasis_t), intent(INOUT)     :: basism
    real(dbl), intent(IN)               :: svderr   ! SVD cut-off
    logical, intent(IN), optional       :: lsvdabs

    real(dbl), allocatable :: hmat(:,:), sigma(:), sigma2(:), work(:)
    real(dbl)              :: fakeU(1,1), fakeVT(1,1), sumsigma
    integer                :: k, nsigma, ierr, LWORK, nrelevant
    logical                :: labs
    integer, save          :: tid = 0


    call get_timer(tid, "svd_basis")
    call start_timer(tid)

    labs = .false.
    if (present(lsvdabs)) labs = lsvdabs

    call dispose_svd_basis_t(basism)

    allocate(hmat(subdim,rank))
    hmat = mat
    nsigma = min(subdim,rank)
    allocate(sigma(nsigma))

    ! workspace query
    LWORK = -1
    call dgesvd('O','N',subdim,rank,hmat,subdim,sigma,fakeU,1, &
         fakeVT,1,fakeU,LWORK,ierr)
    LWORK = int(fakeU(1,1))+1

    ! for large operators the workspace must be reduced
    if (LWORK .le. 0) then
       LWORK = max(3*min(subdim,rank) + max(subdim,rank), &
            5*min(subdim,rank))*10
    end if
    if (LWORK .le. 0) then
       LWORK = max(3*min(subdim,rank) + max(subdim,rank), &
            5*min(subdim,rank))
    end if


    allocate(work(LWORK))
    call dgesvd('O','N',subdim,rank,hmat,subdim,sigma,fakeU,1, &
         fakeVT,1,work,LWORK,ierr)
    if (ierr .ne. 0) then
       write(*,*) "ERROR"
       write(*,*) "Compression: SVD failed"
       write(*,*) "DGESVD error code: ", ierr
       stop 1
    end if

    deallocate(WORK)
    allocate(sigma2(size(sigma)))

    sigma2 = sigma**2
    if (.not.labs) then
       sigma2 = sigma2/sum(sigma2)
    end if

    nrelevant = 0
    sumsigma = 0.d0
    basism%subdim = subdim

    do k=nsigma,1,-1

       sumsigma = sumsigma + sigma2(k)

       if (sumsigma .gt. svderr) then
          nrelevant = min(k+1,nsigma)

          basism%nbasis = nrelevant

          if (nrelevant .lt. nsigma) then
             basism%l2err = sum(sigma2(nrelevant+1:nsigma))
          else
             basism%l2err = 0.d0
          end if

          allocate(basism%basis(subdim,nrelevant))

          basism%basis(:,:) = hmat(:,1:nrelevant)

          allocate(basism%sigma(nrelevant))
          basism%sigma = sigma2(1:nrelevant)

          exit !k-loop
       end if

    end do

    deallocate(hmat, sigma, sigma2)
    call stop_timer(tid)
  end subroutine svd_basis_1m_from_mat




  subroutine trafo_cp_grid_2_svd_basis(thcpd,zcpd,svdbasis,nmode)
    implicit none

    class(tensor_t), intent(IN), target :: thcpd
    type(cpd_t), intent(INOUT)          :: zcpd
    integer, intent(IN)                 :: nmode
    type(svdbasis_t), intent(IN)        :: svdbasis(nmode)


    type(cpd_t), pointer                :: cpd => NULL()
    type(hcpd_t), pointer               :: hcpd => NULL()

    ! for compressed cpd
    integer, allocatable           :: gdim(:),dofspf(:), potdofmap(:,:)
    real(dbl), allocatable         :: ort(:)
    character(len=16), allocatable :: modelabel(:)
    logical, allocatable           :: loned(:)
    integer                        :: ortdim


    integer :: m, rank
    logical :: lerr

    select type(thcpd)
    type is (hcpd_t)
       hcpd => thcpd
       rank = hcpd%rank
    type is (cpd_t)
       cpd => thcpd
       rank = cpd%rank
    end select

    if (nmode .ne. thcpd%nmode) then
       print*, "ERROR: subroutine trafo2svdbasis"
       print*, "--> Number of modes are not consistent"
       print*, "svdbasis, nmode", nmode
       print*, "(h)cpd%nmode", thcpd%nmode
       stop 1
    end if
    lerr = .false.
    if (associated(hcpd)) then
       do m=1,nmode
          if (svdbasis(m)%subdim .ne. hcpd%subdim2(m)) then
             print*, "ERROR: trafo2svdbasis: incosistent operator " &
                  //"sizes in mode ", m
             lerr = .true.
          end if
       end do
    elseif (associated(cpd)) then
       do m=1,nmode
          if (svdbasis(m)%subdim .ne. cpd%subdim(m)) then
             print*, "ERROR: trafo2svdbasis: incosistent operator " &
                  //"sizes in mode ", m
             lerr = .true.
          end if
       end do
    end if
    if (lerr) stop 1


    allocate(gdim(nmode))
    allocate(dofspf(nmode))
    allocate(potdofmap(2,nmode))
    allocate(modelabel(nmode))
    allocate(loned(nmode))
    ortdim = 0
    do m=1,nmode
       gdim(m) = svdbasis(m)%nbasis
       ortdim = ortdim + gdim(m)
       dofspf(m) = m
       potdofmap(1,m) = m
       potdofmap(2,m) = 1
       write(modelabel(m),'(a,i4)') "mode_", m
       call rmblnk(modelabel(m))
       loned(m) = .false.
    end do
    allocate(ort(ortdim))
    ort = 0.d0


    call allocate_cpd_t(zcpd,rank,nmode,nmode,gdim,dofspf,potdofmap, &
         modelabel,ort,ortdim, loned)

    if (associated(hcpd)) then
       zcpd%coeff = hcpd%coeff

       do m=1,nmode
          zcpd%spp(m)%htm(:,:) = 0.d0
          call dgemm('T','N', svdbasis(m)%nbasis,rank,svdbasis(m)%subdim,1.d0, &
               svdbasis(m)%basis, svdbasis(m)%subdim, hcpd%sph(m)%htm, &
               svdbasis(m)%subdim, 0.d0, zcpd%spp(m)%htm,svdbasis(m)%nbasis)

       end do

    elseif (associated(cpd)) then
       zcpd%coeff = cpd%coeff

       do m=1,nmode
          zcpd%spp(m)%htm(:,:) = 0.d0

          call dgemm('T','N', svdbasis(m)%nbasis,rank,svdbasis(m)%subdim,1.d0, &
               svdbasis(m)%basis,svdbasis(m)%subdim, cpd%spp(m)%htm, &
               svdbasis(m)%subdim, 0.d0, zcpd%spp(m)%htm,svdbasis(m)%nbasis)
       end do

    end if



  end subroutine trafo_cp_grid_2_svd_basis



  subroutine trafo_cp_svd_2_grid_basis(zcpd,thcpd,svdbasis,nmode)
    implicit none

    type(cpd_t), intent(in)               :: zcpd
    class(tensor_t), intent(INOUT),target :: thcpd
    integer, intent(IN)                   :: nmode
    type(svdbasis_t), intent(IN)          :: svdbasis(nmode)


    type(hcpd_t), pointer                 :: hcpd => NULL()
    type(cpd_t), pointer                  :: cpd => NULL()

    integer :: m, rank

    select type(thcpd)
    type is (hcpd_t)

       hcpd => thcpd
       rank = zcpd%rank
       if (rank.ne.hcpd%rank) then
          call changehrank(hcpd,rank)
       end if

       hcpd%coeff = zcpd%coeff

       do m=1,hcpd%nmode
          hcpd%sph(m)%htm(:,:,:) = 0.d0
          call dgemm('N','N', svdbasis(m)%subdim,rank,svdbasis(m)%nbasis,1.d0 &
               ,svdbasis(m)%basis, svdbasis(m)%subdim, zcpd%spp(m)%htm, &
               svdbasis(m)%nbasis, 0.d0, hcpd%sph(m)%htm, hcpd%subdim2(m))
       end do

    type is (cpd_t)

       cpd => thcpd
       rank = zcpd%rank

       if (rank.ne.cpd%rank) then
          call changerank(cpd,rank)
       end if

       cpd%coeff = zcpd%coeff

       do m=1,cpd%nmode
          cpd%spp(m)%htm(:,:) = 0.d0
          call dgemm('N','N', svdbasis(m)%subdim,rank,svdbasis(m)%nbasis, &
               1.d0,svdbasis(m)%basis, svdbasis(m)%subdim, zcpd%spp(m)%htm, &
               svdbasis(m)%nbasis, 0.d0,cpd%spp(m)%htm, cpd%subdim(m))
       end do


    end select

  end subroutine trafo_cp_svd_2_grid_basis



end module svdbasis_s
