!   -*- f90 -*-

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! program natpot2cpd
!
! Transform a natpot file to a canonical polyadic
! decomposition (CPD)
!
! The program actually transforms the core tensor
! of the natpot to CPD format and subsequently exands
! the now CPD-formatted core tensor into the SPP basis
! of the natpot (i.e., from the SPP basis to the
! primitive basis).
!
! 11/18 MS
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



! module with parameters from input etc
module natpot2cpd_m
  use natpottype
  use tensortools

  implicit none
  integer            :: maxiter, startiter
  integer            :: rank, maxrank
  double precision   :: alpha, deltastop, errorestimate
  logical            :: linitrandom, lcontinue, lrank
  logical            :: lstopfile, ldeltastop
  logical            :: lresetiter
  logical            :: lwrovlp,lwrcoeff,lvpot,lcheck,lgrow,lcc,lddnp
  integer            :: deltastopunit
  integer            :: grank, giter

  character(len=512) :: inpdir
  character(len=512) :: dvrfile
  character(len=512) :: npotfile
  character(len=512) :: cpdfile
  character(len=512) :: logfile
  character(len=512) :: iterationsfile
  character(len=512) :: stopfile
  character(len=512) :: timingfile
  character(len=512) :: ovlpfile
  character(len=512) :: coefffile
  character(len=512) :: vpotfile
  character(len=512) :: ccfile

  type (natpot_t), pointer    :: pnatpot => NULL()

  double precision, parameter :: au2ic = 2.1947463137d+5
  double precision, parameter :: au2ev = 27.21138386d0
  double precision, parameter :: au2mev = 27211.38386d0

  double precision            :: coreovlp      ! for the fit-error
  integer, allocatable        :: controrder(:) ! integration order


end module natpot2cpd_m



! main program
program natpot2cpd
  use natpot2cpd_m
  use dvrtype_m
  use tensortype
  use natpottype
  use cpdtype_m
  use cpdtype_s
  use tensortools
  use eximporttools
  use mccomplete
  use tensortools
  use cpd_als_m
  use cpd_als_s
  use pftools

  use channels
  use global
  use pglobal
  use timing
  use pdefaultmod

  use mtlib, only: mttxdd
  use op1lib, only: cuttxd1
  use op2lib, only: subtxxzz

  implicit none

  type(dvr_t)                    :: dvr
  type(natpot_t), target         :: natpot
  type(cpd_t)                    :: cpd,cpd_grid
  double precision               :: ovlp1,ovlp2,ovlp3,err
  integer                        :: m,nspp,nspp1,maxnspp,ierr,iiter
  double precision, allocatable  :: ort(:)
  integer                        :: t1,t2,orig_modc,tstart,tend,ibl
  double precision               :: fcputime,fsystime,tcpu,ttot,cpdcore2,ddot
  character(len=c4)              :: string
  logical, allocatable           :: loned(:)


  external  fcputime,fsystime

  call unixzeit(tstart)

  coreovlp = -999.d9

  ! set defaults and get input
  call default()
  call pdefault()


  ! get input params and log
  call np2cpd_input()
  call np2cpd_log()

  ! load natpot file
  call loaddvr(dvr,dvrfile)
  call logdvrtype(dvr, ilog)
  call importMCTDH(natpot,dvr,npotfile)
  call wrnpotproperties(natpot, ilog)
  orig_modc = natpot%modc

  ! check if we need t optimize
  lexact = .false.
  if (natpot%modcdim .lt. rank) then
     routine = "natpot2cpd"
     write(message,'(a,i8)') "CPD rank is larger than or equal to the " &
          //"number of terms in the natpot. I will adjust the " &
          //"rank to ", natpot%modcdim
     call warnmsg
     lwarn = .false.
     write(message, '(a,i8,a,i8)') "Adjusting rank from", &
          rank, " to ", natpot%modcdim
     call cncntrblnk(message)
     write(ilog,'(a)') message
     rank =  natpot%modcdim
     lexact = .true.
  end if

  if (natpot%modcdim .eq. rank) then
     ! this is an exact re-write
     message = "CPD rank is equal to natpot core tensor dimension. Skipping "// &
          "optimization."
     call warnmsg
     lwarn = .false.

     lexact = .true.
  end if

  call myflush(ilog)

  ! further down we we assume that the SPP
  ! being orhonormal (which may not be the
  ! case, for instance for ext-grid
  ! natpots from mcpotfit) so we just
  ! trafo here
  call orthonormSPP(natpot)

  ! if this is an exact run, we skip the optimization
  ! and jump to the end to just re-write the
  ! natpot to cpd format
  if (lexact) goto 111



  ! we remove the contracted mode for optimization to get
  ! a more compact core tensor. SPP which contribute less than
  ! 1.d-10 to the trace of the density are neglected
  ! (so we do not lose too much accuracy)
  if (natpot%modc .ne. 0) then
     m = natpot%modc
     if( errorestimate .gt. 0.d0) then
        err=errorestimate**2
     else
        err = sum(natpot%coeff**2) / natpot%rvpotdim * 1.d-10
     endif
     nspp1 = natpot%subdim(m)

     write(ilog,'(a)') "Removing contracted mode to reduce memory costs"
     call getModeLabelsStr(natpot,orig_modc,string)
     write(ilog,'(a,i3,a)') "Contacted mode:",orig_modc, ", coord: "//trim(string)

     call remove_contracted(natpot, err, nspp)

     write(ilog,'(/a)') "New natpot properties:"
     call wrnpotproperties(natpot, ilog)

     maxnspp = min(nspp1, natpot%potvdim(orig_modc)*natpot%potndim(orig_modc))
     err = dsqrt(err)
     write(ilog,'(a)')          "Impact of removing contracted mode on global natpot error:"
     write(message,'(a,3(i6,a))')  "New No SPP | max possible No. SPP | gridpoints  : ", &
          nspp, " | ",  maxnspp, " | ", natpot%subdim(orig_modc)
     call cncntrblnk(message)
     write(ilog,'(a)') trim(message)
     write(*,'(/a,i6,a,i6)')  "Transforming contracted mode to SPP space..."
     write(*,'(a,i3,a)')  "Contacted mode:",orig_modc, ", coord: "//trim(string)
     write(*,'(a)')  "Impact of removing contracted mode on global natpot error:"
     write(*,'(a)') trim(message)

     write(ilog,'(a,ES16.8,a)') "RMS Error (estimated): ", err,       "   au"
     write(*,'(a,ES16.8,a)')    "RMS Error (estimated): ", err,       "   au"
     if(err*au2mev .ge. 1000.d0 .or. err*au2mev .le. 0.0001d0) then
        write(ilog,'(a,ES16.8,a)')  "RMS Error (estimated): ", err*au2mev,"  meV"
        write(ilog,'(a,ES16.8,a/)') "RMS Error (estimated): ", err*au2ic, " cm-1"
        write(*,'(a,ES16.8,a)')     "RMS Error (estimated): ", err*au2mev,"  meV"
        write(*,'(a,ES16.8,a/)')    "RMS Error (estimated): ", err*au2ic, " cm-1"
     else
        write(ilog,'(a,f16.9,a)')  "RMS Error (estimated): ", err*au2mev,"  meV"
        write(ilog,'(a,f16.9,a/)') "RMS Error (estimated): ", err*au2ic, " cm-1"
        write(*,'(a,f16.9,a)')     "RMS Error (estimated): ", err*au2mev,"  meV"
        write(*,'(a,f16.9,a/)')    "RMS Error (estimated): ", err*au2ic, " cm-1"
     endif
     call myflush(ilog)
  end if

  ! if this is a continuation run
  ! we need to load an old cpd file
  ! as initial guess
  if (lcontinue) then
     ! load an old cpd-file
     ! and compare to natpot
     ! if compatible transform to SPP basis
     if(lcc) then
        call importMCTDH(cpd_grid,dvr,ccfile)
     else
        call importMCTDH(cpd_grid,dvr,cpdfile)
     endif
     if (.not. iscompatible(cpd_grid,natpot)) then
        routine = "natpot2cpd, loading cpd for continuation"
        message = "Incompatible grids or mode combitions with natpot."
        call errormsg
     end if
  end  if

  if (rank .eq. -1) then
     ! The user has not pre-defined a rank
     ! we guess (a small) one or use the same as
     ! restart if present
     if (lcontinue) then
        rank = cpd_grid%rank
        write(ilog,'(a,i8)') "No rank supplied. Using rank:", rank
        write(ilog,'(a)'   ) "of old file "//trim(cpdfile)
     else
        rank = min(sum(natpot%potdim), product(natpot%potdim))
        write(ilog,'(a,i8)') "No rank supplied. Using guessed rank: ", rank
     end if
     write(ilog,*)
     call myflush(ilog)

  else if (lgrow) then
     rank = max(grank,cpd_grid%rank)
  end if


  ! integrate over the largerst modes
  ! first so we redude the strain on the
  ! memory bus
  call get_controrder(natpot)


  ! allocate space for the cpd.
  ! we do not work on the primitive grid,
  ! but only on the core tensor of the
  ! natpot in the SPP basis.
  ! therefore we allocate a cpd
  ! which has grid dimensions matching
  ! the SPP basis dimension of the natpot
  ! hence the dimensions of the core
  ! tensor. This is sort of
  ! an artificial setup so we also
  ! fill in some artificial information.

  ! Arrays with dof/mode info:
  allocate(dofspf(natpot%nmode))
  allocate(potdofmap(2,natpot%nmode))
  allocate(modelabel(natpot%nmode))
  ortdim=sum(natpot%potdim)
  allocate(ort(ortdim))
  ! some fake mode-labels
  do m=1,natpot%nmode
     write(modelabel(m), '(a,i4)') "mode_",m !artificial mode labels
     call rmblnk(modelabel(m))
     potdofmap(1,m)=m
     potdofmap(2,m)=1
     dofspf(m)=m
  end do
  ! some fake ort
  ort = 0.d0
  allocate(loned(natpot%nmode))
  loned = .false.

  call allocate_cpd_t(cpd,rank,natpot%nmode,natpot%nmode, &
       natpot%potdim,dofspf,potdofmap, &
       modelabel,ort,ortdim,loned)

  ! free things we do not need any more
  deallocate(dofspf,potdofmap,modelabel,ort,loned)

  ! make an initial guess for the cpd.
  ! if this is a continuation run
  ! we load an old cpd file
  if (lcontinue) then
     ! transform read-in cpd to SPP basis
     ! and get restart iteration number
     call cpd2spp(cpd_grid, cpd, natpot)
     if (.not. lresetiter) then
        call getstartiter()
     else
        startiter = 0
     end if
  else
     ! create an initial guess from natpot
     call natpot2cpd_init(cpd, natpot, orig_modc)
  end if

  ! this is for getrms - we just need this once -
  ! it statys constant over the iterations

  coreovlp = ddot(natpot%cdim,natpot%coeff,1,natpot%coeff,1)

  ! optimize the cpd against the natpot
  ! if the core tensor size is .eq. cpd%rank
  ! we are exact
  if ( natpot%cdim .gt. cpd%rank) then
     pnatpot => natpot
     call optimize_cpd(cpd, pnatpot, iiter)
  end if

  ! we are done. sort the cpd so the
  ! largest coeff come first
  call sortcpd(cpd)


  ! transform the CPD to the primitive grid
  ! and write to file.
  call cpd2grid(cpd, cpd_grid, natpot)

111 continue
  if (lexact) then
     ! we are exact. Jest re-write
     call transscribenp2cpd(cpd_grid,natpot)
     call sortcpd(cpd_grid)
  end if


 ! export CPD to mctdh-readable format
  call exportMCTDH(cpd_grid, dvr, cpdfile, lovwr)

  ! a final test on the cpd and natpot in grid basis
  ! this should give the same as in the SPP basis
  ! within the last iteration
  ovlp1 = toverlap(natpot,natpot)/natpot%rvpotdim
  ovlp2 = toverlap(cpd_grid,natpot)/natpot%rvpotdim
  ovlp3 = toverlap(cpd_grid,cpd_grid)/natpot%rvpotdim

  err = abs((ovlp1 + ovlp3 - 2.d0*ovlp2))
  err = dsqrt(err)
  write(message,'(a,i8,a,i7)') "Iterations =",iiter,",   rank =",cpd%rank
  call cncntrblnk(message)

  write(*,'(/a)') "Final error estimate of cpd against natpot: "
  write(*,'(a,F16.6,a)') "RMS = ", err, " au"
  write(*,'(a,F16.6,a)') "RMS = ", err*au2mev, " meV"
  write(*,'(a,F16.6,a)') "RMS = ", err*au2ic,  " cm-1"
  write(*,'(/a)') trim(message)
  write(ilog,'(/a)') "Final error estimate of cpd against natpot: "
  write(ilog,'(a,F16.6,a)') "RMS = ", err, " au"
  write(ilog,'(a,F16.6,a)') "RMS = ", err*au2mev, " meV"
  write(ilog,'(a,F16.6,a)') "RMS = ", err*au2ic,  " cm-1"
  write(ilog,'(/a)') trim(message)


  if (lvpot) then
     call vpoterr(cpd_grid, dvr, err)

     write(*,'(a)') ""
     write(*,'(a)') "Error estimate of cpd against vpot: "
     write(*,'(a,F16.6,a)') "RMS = ", err, " au"
     write(*,'(a,F16.6,a)') "RMS = ", err*au2mev, " meV"
     write(*,'(a,F16.6,a)') "RMS = ", err*au2ic,  " cm-1"
     write(ilog,'(a)') ""
     write(ilog,'(a)') "Error estimate of cpd against vpot: "
     write(ilog,'(a,F16.6,a)') "RMS = ", err, " au"
     write(ilog,'(a,F16.6,a)') "RMS = ", err*au2mev, " meV"
     write(ilog,'(a,F16.6,a)') "RMS = ", err*au2ic,  " cm-1"

  end if




  cpdcore2 = ddot(cpd_grid%rank,cpd_grid%coeff,1,cpd_grid%coeff,1)
  if (coreovlp .lt. -1.d0) then
     coreovlp = ddot(natpot%cdim,natpot%coeff,1,natpot%coeff,1)
  end if

  write(ilog,'(/a)') "Natpot and CPD properties:"
  write(ilog,'(a,ES14.8)') "CPD core tensor norm    CTN = sqrt(sum_R C_R^2) = ", &
       dsqrt(cpdcore2)
  write(ilog,'(a,ES14.8)') "Natpot core tensor norm NTN = sqrt(sum_I C_I^2) = ", &
       dsqrt(coreovlp)
  write(ilog,'(a,ES14.8/)') "Condition number                        CTN/NTN = ", &
       dsqrt(cpdcore2/coreovlp)

  write(*,'(/a)') "Natpot and CPD properties:"
  write(*,'(a,ES14.8)') "CPD core tensor norm    CTN = sqrt(sum_R C_R^2) = ", &
       dsqrt(cpdcore2)
  write(*,'(a,ES14.8)') "Natpot core tensor norm NTN = sqrt(sum_I C_I^2) = ", &
       dsqrt(coreovlp)
  write(*,'(a,ES14.8/)') "Condition number                        CTN/NTN = ", &
       dsqrt(cpdcore2/coreovlp)


  ! write output of requested
  if (lwrcoeff) then
     if (.not. lexact) then
        call np2cpd_wrcoeff(cpd)
     else
        call np2cpd_wrcoeff(cpd_grid)
     end if
  end if
  if (lwrovlp) then
     if (.not. lexact) then
        call np2cpd_wrovlp(cpd)
     else
        call np2cpd_wrovlp(cpd_grid)
     end if
  end if


  ! free memory
  call deallocate_cpd_t(cpd)
  call deallocate_cpd_t(cpd_grid)
  call deallocate_natpot_t(natpot)

  ! remove stop file
  inquire (file=stopfile, exist=lcheck)
  if (lcheck) then
     open (istop,file=stopfile,form='formatted',status='old')
     close(istop,status='delete')
  endif

  ! write timing information
  open(itim, file=timingfile,iostat=ierr)
  if(ierr .ne. 0) then
     message = "Cannot open file "//trim(timingfile)
     routine = "natpot2cpd"
     call errormsg
  end if
  call timout

  call unixzeit(tend)


  ttot = dble(tend - tstart)
  t1   = int(ttot/3600)
  ttot = ttot - dble(3600*t1)
  t2   = int(ttot/60)
  ttot = ttot - dble(60*t2)
  write(itim,900) 'Walltime       [h:m:s] : ',t1,t2,ttot
  write(message,900) 'Walltime       [h:m:s] : ',t1,t2,ttot
  close(itim)


  ! final timing
  tcpu=fcputime(0)+fsystime(0)
  ttot = tcpu
  t1   = int(ttot/3600)
  ttot = ttot - dble(3600*t1)
  t2   = int(ttot/60)
  ttot = ttot - dble(60*t2)

  call fzeit(string(1:80))
  ibl=index(string(20:80),'   ')+16
  write(ilog,'(a,/)') '***** Overall Timing *****'
  write(ilog,'(a)') string(1:ibl)
  write(ilog,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
  write(ilog, 900) 'Total time     [h:m:s] : ',t1,t2,ttot
  write(ilog,'(a)') trim(message)
  write(ilog,'(/,a,/)') '***** God''s final message *****'
  if (lstopfile) then
     write(ilog,'(a,/)') 'Program was stopped by stop file.'
  else if (ldeltastop) then
     write(ilog,'(a)') 'Program was stopped because the'
     write(ilog,'(a,/)') 'RMS error droped below the required threashold'
  else
     write(ilog,'(a,/)') 'Program terminated normally.'
  end if
  close(ilog)

  write(6,'(a)') string(1:ibl)
  write(6, 900) 'CPU-time       [h:m:s] : ',t1,t2,ttot
  write(6,'(a)') trim(message)
  write(6,'(a)') trim(options)
  write(6,'(a)') '++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++'
900 format( a,i4,' :',i3,' :',f5.1)


contains

subroutine vpoterr(cpd, dvr, err)

  implicit none

  type(cpd_t), intent(IN)       :: cpd
  type(dvr_t), intent(IN)       :: dvr
  double precision, intent(OUT) :: err


  double precision              :: vv, vcpd, cc , ddot
  integer                       :: vdim1(cpd%ndof)
  integer                       :: gdim1(cpd%ndof)
  integer                       :: ndim1(cpd%ndof)
  integer                       :: fmap(cpd%ndof)
  double precision, allocatable, target :: work1(:), work2(:)
  double precision, pointer     :: pwork(:), pcoeff(:)
  integer                       :: m,r,f,z,z1, ierr, f1,m1,n
  integer                       :: dim1,dim2,dim3,dim4
  logical                       :: lfound


  external ddot

  call loadmcvpot(dvr,cpd,vpotfile,ierr)
  if (ierr.ne.0) then
     message = "Error while loading file: "//trim(vpotfile)
     routine = "vpoterr"
     call errormsg
  end if

  ! vpot is now allocated and loaded in module
  ! mccomplete. before further processing,
  ! subtract any 1-d potentials which may be
  ! contained in the cpd

  ! other then in potfit, loadmcvpot stores vpot in mode-order
  ! not in dof-order. If 1d terms are present we need to
  ! map this into the mode order and subtract there


  if (any(cpd%loned)) then

     ! build the f-map
     ! f1 = fmap(f) is the position of
     ! f in the mode-ordered vpot

     do f=1,cpd%ndof
        f1 = 0
        m = cpd%dofspf(f)
        do m1=1,m-1
           f1 = f1 + cpd%nspfdof(m1)
        end do
        lfound = .false.
        do n=1,cpd%nspfdof(m)
           if (cpd%spfdof(n,m).eq.f) then
              f1 = f1 + 1
              lfound = .true.
           else if (.not. lfound) then
              f1 = f1 + 1
           end if

        end do
        fmap(f) = f1
     end do



     ! build v,g,n with respect to f1

     f1 = fmap(1)
     vdim1(1) = 1
     gdim1(1) = cpd%gdim(f1)
     do f=2,cpd%ndof
        f1 =  fmap(f)
        vdim1(f) = vdim1(f-1)*cpd%gdim(f1-1) ! << cpd%gdim is f-order
        gdim1(f) = cpd%gdim(f1)
     end do

     ndim1(cpd%ndof)=1
     do f=cpd%ndof-1,1,-1
        ndim1(f) = ndim1(f+1)*gdim1(f+1) ! << gdim is f1-order
     end do

     do f=1,cpd%ndof
        if (cpd%loned(f)) then
           f1 = fmap(f)

           z = cpd%zort(f)       ! oned is in f-order
           z1 = z+cpd%gdim(f)-1

           dim1 = vdim1(f1)       ! dims are in f1-order
           dim2 = gdim1(f1)
           dim3 = ndim1(f1)
           call subtxxzz (cpd%oned(z:z1),vpot,dim1,dim2,dim3)
        end if
     end do

  end if

  ! vpot*vpot
  vv = ddot(cpd%vpotdim,vpot,1,vpot,1)

  !cpd*cpd
  cc = toverlap(cpd,cpd)


  !cpd*vpot
  vcpd = 0.d0

  ! re-define dims for modes
  vdim1(1) = 1
  gdim1(1) = cpd%subdim(1)
  do m=2,cpd%nmode
     vdim1(m) = vdim1(m-1)*cpd%subdim(m-1)
     gdim1(m) = cpd%subdim(m)
  end do
  ndim1(cpd%nmode)=1
  do m=cpd%nmode-1,1,-1
     ndim1(m) = ndim1(m+1)*cpd%subdim(m+1)
  end do


  !$omp parallel private(r,work1,work2,pwork,pcoeff) &
  !$omp private(dim1,dim2,dim3,dim4,m)

  allocate(work1(ndim1(2)))
  allocate(work2(ndim1(1)))

  !$omp do reduction(+:vcpd)
  do r=1,cpd%rank

     dim1 = 1
     dim4 = 1

     m = 1
     dim2 = gdim1(m)
     dim3 = ndim1(m)

     call mttxdd (cpd%spp(m)%htm(:,r),vpot,work2,dim1,dim2,dim3,dim4)

     pcoeff => work2
     pwork => work1

     do m=2,cpd%nmode

        dim2 = gdim1(m)
        dim3 = ndim1(m)

        call mttxdd (cpd%spp(m)%htm(:,r),pcoeff,pwork,dim1,dim2,dim3,dim4)

        if (associated(pcoeff, work1)) then
           pcoeff => work2
           pwork => work1
        else
           pcoeff => work1
           pwork => work2
        end if

     end do

     vcpd =  vcpd + pcoeff(1)*cpd%coeff(r)

  end do
  !$omp end do

  deallocate(work1,work2)

  !$omp end parallel


  err = abs(vv + cc - 2.d0*vcpd)/cpd%rvpotdim
  err = sqrt(err)


end subroutine vpoterr



subroutine np2cpd_wrcoeff(cpd)

  implicit none

  type(cpd_t), intent(IN) :: cpd

  integer :: ierr,r

  open(idump,file=trim(coefffile), status="unknown",iostat=ierr)
  if (ierr.ne.0) then
     routine = "np2cpd_wrcoeff"
     message = "Could not open file "//trim(coefffile)
     call errormsg
  end if

  write(idump,'(a)') "# No.      Coeff"

  if (ierr.ne.0) then
     routine = "np2cpd_wrcoeff"
     message = "Could not write to file "//trim(coefffile)
     call errormsg
  end if

  do r=1,cpd%rank
     write(idump, '(i8,4x,ES14.8)', iostat=ierr) r, cpd%coeff(r)
     if (ierr .ne. 0) exit
  end do

  close(idump)

  if (ierr.ne.0) then
     routine = "np2cpd_wrcoeff"
     message = "Could not write to file "//trim(coefffile)
     call errormsg
  end if


end subroutine np2cpd_wrcoeff



subroutine np2cpd_wrovlp(cpd)

  implicit none

  type(cpd_t), intent(IN)       :: cpd
  double precision, allocatable :: ovlp(:,:)
  double precision              :: oij, ddot
  character(len=24)             :: FORMAT

  integer :: ierr,r,r1,m,rank

  rank = min( cpd%rank, 1000)
  if (rank .lt. cpd%rank) then
     message = "Truncating overlap matrix to the first 1000x1000 terms"
     routine = "np2cpd_wrovlp"
     call warnmsg
  end if


  allocate(ovlp(rank, rank), stat=ierr)
  if (ierr.ne. 0) then
     message = "Could not allocate overlap matrix. Skipping."
     routine = "np2cpd_wrovlp"
     call warnmsg
     return
  end if

  !$omp parallel do private(r1,oij,m)
  do r=1,rank
     do r1=r,rank

        m = 1
        oij = ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1),1)

        do m=2,cpd%nmode
           oij = oij * ddot(cpd%subdim(m), cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r1),1)
        end do

        ovlp(r1,r) = oij

     end do
  end do
  !$omp end parallel do


  !$omp parallel do private(r1)
  do r=1,rank
     do r1=r,rank
        ovlp(r,r1) = ovlp(r1,r)
     end do
  end do
  !$omp end parallel do

  open(idump,file=trim(ovlpfile), status="unknown",iostat=ierr)
  if (ierr.ne.0) then
     routine = "np2cpd_wrcoeff"
     message = "Could not open file "//trim(ovlpfile)
     call errormsg
  end if

  FORMAT='(1000ES11.3,a,i4)'
  write(FORMAT(2:5),'(i4)') rank

  if (rank .lt. cpd%rank) then
     write(idump,'(a)') "# overlap matrix (trunkated)"
  else
     write(idump,'(a)') "# overlap matrix"
  end if
  write(idump,'(a,1000(i5,6x))') "#", (r,r=1,rank)

  if (ierr.ne.0) then
     routine = "np2cpd_wrcoeff"
     message = "Could not write to file "//trim(ovlpfile)
     call errormsg
  end if

  do r=1,rank
     write(idump,FORMAT,iostat=ierr) (ovlp(r,r1), r1=1,rank),'  #',r
     if (ierr.ne.0) exit
  end do

  close(idump)

  if (ierr.ne.0) then
     routine = "np2cpd_wrovlp"
     message = "Could not write to file "//trim(ovlpfile)
     call errormsg
  end if


end subroutine np2cpd_wrovlp





! calculate  sqrt( sum(natpot - cpd)^2 )
! using binomial formula
double precision function getrms(natpot, cpd)

  implicit none

  type(cpd_t), intent(IN)     :: cpd
  type(natpot_t), intent(IN)  :: natpot

  double precision    :: rnc,rcc, diff
  integer, save       :: tid

  call get_timer(tid,'pfiterr')
  call start_timer(tid)

  call ovlp_np_cpd(natpot,cpd,rnc)
  rcc = toverlap (cpd,cpd)

  diff = abs(coreovlp + rcc - 2.d0*rnc)/natpot%rvpotdim
  diff = dsqrt(diff)
  getrms = diff

  call stop_timer(tid)

end function getrms





! get a good contraction order
! so that the largest modes are contracted
! first (to reduce trafic for memory bound
! jobs)
subroutine get_controrder(natpot)

  implicit none
  type(natpot_t), intent(IN) :: natpot
  integer                    :: m,m1,m2,pdim
  integer                    :: potdim(natpot%nmode)

  if (allocated(controrder)) deallocate(controrder)
  allocate(controrder(natpot%nmode))

  if (natpot%cdim .le. 2**19) then !4 mb
     ! we are probably not memory bound
     ! so we go by the first indices
     do m=1,natpot%nmode
        controrder(m) = m
     end do
     return
  end if

  ! we are likely memory bound
  ! and moving the core tensor back
  ! and forth between ram and cache
  ! is more very time-consuming so we
  ! reduce its size as quickly as possible
  controrder = 0
  potdim(1) = natpot%potdim(1)
  controrder(1) = 1

  do m=2,natpot%nmode

     if (natpot%potdim(m) .le. potdim(m-1)) then
        potdim(m) = natpot%potdim(m)
        controrder(m) = m
     else
        pdim = natpot%potdim(m)
        m2 = 1
        do m1=m,2,-1
           if (potdim(m1-1) .lt. pdim) then
              potdim(m1) = potdim(m1-1)
              controrder(m1) = controrder(m1-1)
           else
              m2=m1
              exit
           end if

        end do
        potdim(m2) = pdim
        controrder(m2) = m
     end if
  end do

end subroutine get_controrder



! this is a special routine as we have
! expanded the natpot core-tensor into cpd form
! so we should not use the toverlap routine
! from tensortools but operate directly on the
! core tensor and spare the SPP part
subroutine ovlp_np_cpd(natpot, cpd, ovlp)

  implicit none

  type(cpd_t), intent(IN)       :: cpd
  type(natpot_t), intent(IN)    :: natpot
  double precision, intent(OUT) :: ovlp

  double precision, allocatable, target  :: work1(:)
  double precision, allocatable, target  :: work2(:)
  double precision, pointer     :: pwork(:), pcoeff(:)

  integer                       :: dim1,dim2,dim3,dim4
  integer                       :: r,m

  ovlp = 0.d0

  !$omp parallel private(r,work1,work2,pwork,pcoeff) &
  !$omp private(dim1,dim2,dim3,dim4,m)

  allocate(work1(natpot%potndim(2)))
  allocate(work2(natpot%potndim(1)))

  !$omp do reduction(+:ovlp)
  do r=1,cpd%rank

     dim1 = 1
     dim4 = 1

     m = 1
     dim2 = natpot%potdim(m)
     dim3 = natpot%potndim(m)

     call mttxdd (cpd%spp(m)%htm(:,r),natpot%coeff,work2,dim1,dim2,dim3,dim4)

     pcoeff => work2
     pwork => work1

     do m=2,natpot%nmode

        dim2 = natpot%potdim(m)
        dim3 = natpot%potndim(m)

        call mttxdd (cpd%spp(m)%htm(:,r),pcoeff,pwork,dim1,dim2,dim3,dim4)

        if (associated(pcoeff, work1)) then
           pcoeff => work2
           pwork => work1
        else
           pcoeff => work1
           pwork => work2
        end if

     end do

     ovlp = ovlp + pcoeff(1)*cpd%coeff(r)

  end do
  !$omp end do

  deallocate(work1,work2)

  !$omp end parallel

end subroutine ovlp_np_cpd



! once we have set up a CPD we can start optimizing it
subroutine optimize_cpd(cpd, natpot, iiter)

  implicit none

  type(cpd_t), intent(INOUT)       :: cpd
  type(natpot_t), intent(IN)       :: natpot

  integer                          :: iiter, ierr, nline, newrank
  integer                          :: i, m, pot10
  double precision                 :: rms, cond, ddot, jfunc, covlp
  logical                          :: lout
  character(len=87)                :: line, efile
  integer, save                    :: tid

  nline = 87
  if (isolver .ne. 1) nline = 72
  line = " "
  do i=1,nline
     line(i:i) = "-"
  end do

  if (lcontinue .and. .not. lresetiter) then
     open(iit,file=trim(iterationsfile),iostat=ierr,position='append')
  else
     open(iit,file=trim(iterationsfile),iostat=ierr)
  end if
  if (ierr.ne.0) then
     message = "Cannot open file "//trim(iterationsfile)
     routine = "optimize_cpd (natpot2cpd.F90)"
     call errormsg
  end if

  lout = .true.
  iiter = startiter

  write(ilog,'(a)') "Starting ALS iterations. For details see file:"
  write(ilog,'(a)') trim(iterationsfile)


  ! haeder
  if (isolver .eq. 1) then
     ! Cholesky
     write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond","Jfunc"
     write(*, '(a)') trim(line)
     if (.not. lcontinue) then
        write(iit, 197) "#","iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond", "Jfunc"
        line(1:1) = "#"
        write(iit, '(a)') trim(line)
     end if
  else
     ! eigenvalue
     write(*, 198) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
     write(*, '(a)') trim(line)
     if (.not. lcontinue) then
        write(iit, 199) "#","iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
        line(1:1) = "#"
        write(iit, '(a)') trim(line)
     end if
  end if


  ! start-error
  rms = getrms(natpot, cpd)
  covlp = ddot(cpd%rank,cpd%coeff,1,cpd%coeff,1)
  jfunc = rms**2*natpot%rvpotdim + alpha*covlp
  cond = dsqrt(covlp / coreovlp)

  if (.not. lcontinue) then
     ! this is already the last entry in the file for continuation
     if (isolver .eq. 1) then
        write(iit,200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
     else
        write(iit,201) iiter, rms*au2ic, rms*au2mev, rms, cond
     end if
  end if
  ! but its nice to have in screen
  if (isolver .eq. 1) then
     write(*,200) iiter, rms*au2ic,rms*au2mev, rms, cond, jfunc
  else
     write(*,201) iiter, rms*au2ic, rms*au2mev, rms, cond
  end if
  call myflush(iit)

  ! init ALS
  call alsinit(cpd)
  call alssetsolver(isolver)

  lwarn = .false.

  do iiter=startiter+1,maxiter

     call get_timer(tid,'ptiteration')
     call start_timer(tid)

     do m=1, cpd%nmode

        if (iiter.eq.maxiter .and. isolver .eq.2) then
           write(efile,'(a,i4)') "/als_eigenval_mode_", m
           call rmblnk(efile)
        else
           efile = ""
        end if

        call alsiter1m(righthandside, cpd, m, alpha, ierr, efile)

        if (ierr .ne. 0) then
           routine = "optimize_cpd(cpd, natpot)"
           write(message,'(a,i6)') "ALS failed with error code", ierr
           call errormsg
        end if

     end do

     call stop_timer(tid)

     pot10 = ceiling(log10(dble(iiter)))
     if (pot10.gt.2) then
        pot10 = pot10-1
        if(pot10.ge.2 .and. int(iiter/10**pot10).lt.2) then
           i=5
        elseif(pot10.ge.2 .and. int(iiter/10**pot10).lt.5) then
           i=2
        else
           i=1
        endif
        pot10 = pot10-1
        lout = (modulo(iiter,10**pot10/i) .eq. 0)
     else
        lout = .true.
     end if

     if (lout .or. iiter .ge. maxiter .or. lwarn) then

        rms = getrms(natpot, cpd)
        covlp = ddot(cpd%rank,cpd%coeff,1,cpd%coeff,1)
        jfunc = rms**2*natpot%rvpotdim + alpha*covlp
        cond = dsqrt(covlp / coreovlp)

        if (isolver .eq. 1) then
           write(*,  200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
           write(iit,200) iiter, rms*au2ic, rms*au2mev, rms, cond, jfunc
        else
           write(*,  201) iiter, rms*au2ic, rms*au2mev, rms, cond
           write(iit,201) iiter, rms*au2ic, rms*au2mev, rms, cond
        end if
        call myflush(iit)


        if (rms .le. deltastop) then
           ldeltastop = .true.
           write(*,'(a)') "RMS dropped below threshold. Stopping iterations."
           write(ilog,'(a)') "RMS dropped below threshold. Stopping iterations."
           exit ! iiter-loop
        end if



        ! check stop file
        if (np2cpdstop()) then
           write(*,'(a)') "Iterations stopped by stop-file"
           write(ilog,'(a)')  "Iterations stopped by stop-file"
           exit ! iiter-loop
        end if

     end if

     if (lwarn) exit ! loop

     if (lgrow .and. cpd%rank .lt. maxrank) then
        if (modulo(iiter,giter) .eq.0 .and. iiter .lt. maxiter) then
           newrank = min(maxrank,cpd%rank + grank)
           write(message,'(a,i8,a,i8,a,i9)') "Changing rank ", &
               cpd%rank, " --> ", newrank, "  after iteration", iiter
           call cncntrblnk(message)
           write(*,'(a)') trim(message)
           write(ilog,'(a)') trim(message)
           call myflush(ilog)
           call changerank(cpd, newrank, .true.)
           call alsfree()
           call alsinit(cpd)
        end if
     end if

  end do !iiter
  iiter=min(iiter,maxiter)

  close(iit)


  ! print the header again on screen
  line(1:1) = "-"
  write(*, '(a)') trim(line)
  if (isolver .eq. 1) then
     ! Cholesky
     write(*, 196) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond","Jfunc"
  else
     ! eigenvalue
     write(*, 198) "iiter","RMS[cm-1]","RMS[meV]","RMS[au]","cond"
  end if
  write(*, '(a)') trim(line)

  call alsfree()




196 FORMAT (4x,a,6x,a,7x,a,12x,a,10x,a,10x,a)
197 FORMAT (a,2x,a,7x,a,7x,a,11x,a,10x,a,11x,a)

198 FORMAT (4x,a,6x,a,6x,a,12x,a,11x,a)
199 FORMAT (a,2x,a,7x,a,7x,a,11x,a,11x,a)

200 FORMAT (i8,2x,2(f14.4,1x),4x,ES14.8,1x,ES14.4,1x,ES14.4)
201 FORMAT (i8,1x,2(f14.4,1x),4x,ES14.8,1x,ES14.4)

end subroutine optimize_cpd


subroutine righthandside(cpd,mtarget,B)

  implicit none
  type(cpd_t), intent(IN)       :: cpd
  integer, intent(IN)           :: mtarget
  double precision, intent(OUT) :: B(cpd%subdim(mtarget),cpd%rank)

  double precision, allocatable, target :: work1(:), work2(:)
  double precision, pointer     :: pwork(:), pcoeff(:)
  integer                       :: workdim

  integer                       :: r,m,m1,m2,mctr
  integer                       :: dim1,dim2,dim3,dim4

  logical                       :: ldone(pnatpot%nmode), lstart
  integer                       :: vgn(3,pnatpot%nmode)
  integer, save                 :: tid

  call get_timer(tid,'ptrhs')
  call start_timer(tid)

  ! optimal workdims
  ! integration is done over largest mode first
  ! unless the largest mode is mtarget
  ! then the second largest mode is contracted
  ! first

  m1 = controrder(1)
  m2 = controrder(2)
  if (mtarget.eq.m1) then
     mctr = m2 ! second largest first
  else
     mctr = m1 ! largest first
   end if

  workdim = pnatpot%potvdim(mctr)*pnatpot%potndim(mctr)


  ! create local vdim1, gdim1, ndim1
  ! for the current contraction scheme
  ldone = .false.
  do m=1,pnatpot%nmode
     mctr = controrder(m)
     if (mctr .eq. mtarget) cycle
     dim1 = 1
     do m1=1,mctr-1
        if (.not. ldone(m1)) then
           dim1 = dim1 * pnatpot%potdim(m1)
        end if
     end do

     dim2 = pnatpot%potdim(mctr)

     dim3 = 1
     do m1=mctr+1,pnatpot%nmode
        if (.not. ldone(m1)) then
           dim3 = dim3 * pnatpot%potdim(m1)
        end if
     end do
     vgn(1,m) = dim1
     vgn(2,m) = dim2
     vgn(3,m) = dim3
     ldone(mctr) = .true.
  end do


  !$omp parallel private(r,work1,work2,pwork,pcoeff) &
  !$omp private(dim1,dim2,dim3,dim4,ldone) &
  !$omp private(m,m1,m2,mctr,lstart)


  allocate(work1(workdim))
  allocate(work2(workdim))

  !$omp do
  do r=1,cpd%rank

     lstart = .true.

     pwork => work2
     pcoeff => work1

     do m=1,pnatpot%nmode

        mctr = controrder(m)

        if (mctr .eq. mtarget) cycle

        dim1 = vgn(1,m)
        dim2 = vgn(2,m)
        dim3 = vgn(3,m)
        dim4 = 1

        if (lstart) then
           lstart = .false.
           call mttxdd(cpd%spp(mctr)%htm(:,r),pnatpot%coeff,pwork,dim1,dim2,dim3,dim4)
        else
           call mttxdd(cpd%spp(mctr)%htm(:,r),pcoeff,pwork,dim1,dim2,dim3,dim4)
        end if

        if (associated(pcoeff, work1)) then
           pcoeff => work2
           pwork => work1
        else
           pcoeff => work1
           pwork => work2
        end if

     end do

     B(:,r) = pcoeff(1:pnatpot%potdim(mtarget))

  end do
  !$omp end do

  deallocate(work1)
  deallocate(work2)

  !$omp end parallel

  call stop_timer(tid)

end subroutine righthandside



! expand a cpd in SPP basis of the natpot to the
! primitive grid basis of the natpot
subroutine cpd2grid(cpd, cpd_grid, natpot)

  implicit none

  type(cpd_t), intent(IN)    :: cpd
  type(cpd_t), intent(INOUT) :: cpd_grid
  type(natpot_t), intent(IN) :: natpot

  integer                    :: m,r,n,zn,zn1

  ! this deletes any earlier versions of cpd_grid
  call allocate_cpd_t(cpd_grid,cpd%rank,natpot%ndof,natpot%nmode, &
       natpot%gdim,natpot%dofspf,natpot%potdofmap, &
       natpot%modelabel,natpot%ort,natpot%ortdim, natpot%loned)

  cpd_grid%coeff =  cpd%coeff


  do m=1,cpd%nmode
     !$omp parallel do private(n,zn,zn1)

     do r=1,cpd%rank

        n = 1
        zn = natpot%zevec(m,n)
        zn1 = zn + natpot%subdim(m) - 1

        cpd_grid%spp(m)%htm(:,r) = cpd%spp(m)%htm(n,r) * natpot%evec(zn:zn1)

        do n=2, natpot%potdim(m)
           zn = natpot%zevec(m,n)
           zn1 = zn + natpot%subdim(m) - 1

           cpd_grid%spp(m)%htm(:,r) =  cpd_grid%spp(m)%htm(:,r)  &
                +  cpd%spp(m)%htm(n,r) * natpot%evec(zn:zn1)

        end do

     end do
     !$omp end parallel do

  end do


  if (any(natpot%loned)) cpd_grid%oned = natpot%oned


end subroutine cpd2grid



subroutine transscribenp2cpd(cpd,natpot)

  implicit none
  type(natpot_t), intent(IN)  :: natpot
  type(cpd_t), intent(INOUT)  :: cpd

  integer      :: idx(natpot%nmode), gdim1(natpot%nmode)
  integer      :: vnidx(2), vndim(2)
  integer      :: rank, r
  integer      :: zn1,zn2,n,m

  rank = natpot%modcdim

  ! this deletes any earlier versions of cpd
  call allocate_cpd_t(cpd,rank,natpot%ndof,natpot%nmode, &
       natpot%gdim,natpot%dofspf,natpot%potdofmap, &
       natpot%modelabel,natpot%ort,natpot%ortdim,natpot%loned)

  gdim1 = natpot%potdim

  if (natpot%modc .ne. 0) then
     vndim(1) = natpot%potvdim(natpot%modc)
     vndim(2) = natpot%potndim(natpot%modc)
     gdim1(natpot%modc) = 1
  end if

  !$omp parallel do private(vnidx,idx,m,n,zn1,zn2)
  do r=1,rank

     if (natpot%modc .ne. 0) then
        call cindex2sindex(r,vnidx,vndim,2)
     end if

     call cindex2sindex(r,idx,gdim1,natpot%nmode)

     do m=1,natpot%nmode

        if (m .eq. natpot%modc) then
           call cuttxd1(natpot%coeff,cpd%spp(m)%htm(:,r), &
                vndim(1),natpot%subdim(natpot%modc),vndim(2), &
                vnidx(1),vnidx(2))
        else
           n = idx(m)
           zn1 = natpot%zevec(m,n)
           zn2 = zn1 + natpot%subdim(m) - 1
           cpd%spp(m)%htm(:,r) = natpot%evec(zn1:zn2)
        end if

     end do

     cpd%coeff(r) = 1.d0
  end do
  !$omp end parallel do

  if (any(natpot%loned)) cpd%oned = natpot%oned

  call normspp(cpd)


end subroutine transscribenp2cpd




! contract a cpd in primitive grid basis
! to the SPP basis of the natpot
! We rely on the cpd being allocated and
! compatible with the natpot potdims.
! the ranks of the two cpd's may differ.
! if  cpd%rank is larger than cpd_grid%rank
! (restart with different rank),
! the extra coeff are set to zero and the
! extra SPP are set to random
! if  cpd%rank is smaller than cpd_grid%rank
! the surplus terms are thrown away
subroutine cpd2spp(cpd_grid, cpd, natpot)

  implicit none

  type(cpd_t), intent(INOUT) :: cpd
  type(cpd_t), intent(IN)    :: cpd_grid
  type(natpot_t), intent(IN) :: natpot

  double precision :: norm, ddot
  integer          :: m,r,n,zn,zn1
  integer          :: rank, subdim

  ! both cpds are allocated at this point.
  ! we just need to transform the SPP of the
  ! cpd living on the primitive grid into the
  ! SPP basis of the natpot

  rank = min(cpd%rank,cpd_grid%rank)

  ! coeff are the same
  cpd%coeff(1:rank) =  cpd_grid%coeff(1:rank)



  do m=1,cpd_grid%nmode

     subdim = cpd%subdim(m)

     !$omp parallel do private(n,zn,zn1)
     do r=1,rank

        do n=1, natpot%potdim(m)
           zn = natpot%zevec(m,n)
           zn1 = zn + natpot%subdim(m) - 1
           cpd%spp(m)%htm(n,r) = ddot(subdim,cpd_grid%spp(m)%htm(:,r),1, &
                natpot%evec(zn:zn1),1)
        end do
     end do
     !$omp end parallel do

  end do


  if (cpd%rank .le. cpd_grid%rank) return

  ! add some random SPP with zero coeff

  do m=1,cpd%nmode
     subdim = cpd%subdim(m)

     !$omp parallel do private(norm)
     do r=rank+1,cpd%rank
        call random_number(cpd%spp(m)%htm(:,r))
        cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r) - 0.5d0
        norm = ddot(subdim,cpd%spp(m)%htm(:,r),1,cpd%spp(m)%htm(:,r),1)
        cpd%spp(m)%htm(:,r) = cpd%spp(m)%htm(:,r)/sqrt(norm)

        cpd%coeff(r) = 0.d0
     end do
     !$omp end parallel do
  end do



end subroutine cpd2spp



subroutine np2cpd_input()

  implicit none

  integer            :: i,j
  character(len=256) :: buf, buf1
  integer            :: myiargc,iarg,ierr,stat,leninpdir,lencc
  logical            :: lname,lexists,lcomma
  logical            :: lspd,leigen,linpdir,lmnd
  double precision   :: DLAMCH

  external myiargc
  external DLAMCH ! lapack


  maxiter = 1000
  alpha =  DLAMCH("E")
  alpha = sqrt(alpha)

  rank = -1
  maxrank = -1
  linitrandom = .false.
  deltastop = 0.d0
  deltastopunit = 1 ! au
  errorestimate=-1.d0
  ldeltastop = .false.
  isolver = 1
  leigen = .false.
  lspd = .false.
  lcontinue = .false.
  lrank = .false.
  lstopfile = .false.
  lresetiter = .false.
  lwrovlp = .false.
  lwrcoeff = .false.
  lmnd  = .false.
  lvpot = .false.
  lgrow = .false.
  grank = 0
  giter = 0

  call mygetarg(1,buf)
  if( buf(1:1) .eq. ' ' ) then
     ! No arguments: print help and exit
     call np2cpd_wrphelptext()
  end if

  lname = .false.
  linpdir = .false.
  lcc = .false.

  iarg = myiargc()

  call mygetarg(0,options)
  do i = 1, iarg
     call mygetarg(i,buf1)
     options = trim(options)//' '//trim(buf1)
  end do

  i = 0
  do while (i.lt. iarg)
     i=i+1

     call mygetarg(i,buf)

     if( buf(1:1) .eq. '-' ) then

        if ( buf .eq. '-h' .or. &
             buf .eq. '-?' ) then
           call np2cpd_wrphelptext()
        elseif  ( buf .eq. '-ver') then
           call wrversion(6)
           stop 0
        elseif  ( buf .eq. '-mnd') then
           lmnd  = .true.
        elseif  ( buf .eq. '-w') then
           lovwr = .true.
        elseif ( buf .eq. '-c') then
           lcontinue = .true.
        elseif ( buf .eq. '-C') then
           lcontinue = .true.
           lresetiter = .true.
        elseif ( buf .eq. '-CC') then
           lcontinue = .true.
           lresetiter = .true.
           lcc = .true.
           i=i+1
           call mygetarg(i,buf)
           ccfile = trim(buf)
           if (len_trim(name) .eq. 0) then
              message = "Expecting string after -CC option"
              routine = "np2cpd_input"
              call errormsg
           end if

        elseif ( buf .eq. '-o') then
           lwrovlp = .true.
        elseif ( buf .eq. '-l') then
           lwrcoeff = .true.
        elseif  ( buf .eq. '-r') then
           lrank = .true.
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) rank
           if (ierr .ne. 0) then
              message = "Cannot read rank after -r option. " &
                   // "Expecting integer, not: "//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           end if
           maxrank = rank

        elseif  ( buf .eq. '-d') then
           buf1 = " "
           i=i+1
           call mygetarg(i,buf)
           j = index(buf,",")
           lcomma = j.ne.0

           if (lcomma) then
              call rmblnk(buf(j+1:))
              if (index(buf(j+1:), ' ') .gt.1) then
                 !---
                 buf1=buf(j+1:)
                 call keylowc(buf1)
              else
                 !-- -
                 i=i+1
                 call mygetarg(i,buf1)
                 call keylowc(buf1)
              end if
              buf(j:) = " "
           else if( i.lt.iarg) then
              ! - -- or - - -
              i = i+1
              call mygetarg(i,buf1)
              if(buf1(1:1) .eq. "-") then
                 i=i-1
                 buf1 = " "
              elseif (buf1 .eq. ",") then
                 ! - - -
                 i = i+1
                 call mygetarg(i,buf1)
              elseif(buf1(1:1) .eq. ",") then
                 ! - --
                 buf1(1:1) = " "
                 call rmblnk(buf1)
              end if
           end if
           read(buf,*, iostat=ierr) errorestimate
           if (ierr .ne. 0  .or.  errorestimate .le. 0.d0 ) then
               message = "Cannot read error-estimate after -r option. " &
                  // "Expecting positive real, not: "//trim(buf)
               routine = "np2cpd_input"
               call errormsg
            endif
           if (buf1 .eq. "au" .or. buf1 .eq. " ") then
              continue
           elseif (buf1 .eq. "ev") then
              errorestimate=errorestimate/au2ev
           elseif (buf1 .eq. "mev") then
              errorestimate=errorestimate/(1.d3*au2ev)
           elseif (buf1 .eq. "cm-1") then
              errorestimate=errorestimate/au2ic
           else
              message = "Expecting one of 'au', 'eV', 'meV',  or 'cm-1' " &
                   //"after '-s <real>,', not "//trim(buf1)
              routine = "np2cpd_input"
              call errormsg
           end if

        elseif  ( buf .eq. '-R') then
           linitrandom = .true.
        elseif  ( buf .eq. '-g') then
           lgrow = .true.
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) grank
           if (ierr .ne. 0) then
              message = "Cannot read first value after -g option. " &
                   // "Expecting integer, not: "//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           end if
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) giter
           if (ierr .ne. 0) then
              message = "Cannot read second value after -g option. " &
                   // "Expecting integer, not: "//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           end if

        elseif  ( buf .eq. '-s') then
           buf1 = " "
           i=i+1
           call mygetarg(i,buf)
           j = index(buf,",")
           lcomma = j.ne.0

           if (lcomma) then
              call rmblnk(buf(j+1:))
              if (index(buf(j+1:), ' ') .gt.1) then
                 !---
                 buf1=buf(j+1:)
                 call keylowc(buf1)
              else
                 !-- -
                 i=i+1
                 call mygetarg(i,buf1)
                 call keylowc(buf1)
              end if
              buf(j:) = " "
           else if( i.lt.iarg) then
              ! - -- or - - -
              i = i+1
              call mygetarg(i,buf1)
              if(buf1(1:1) .eq. "-") then
                 i=i-1
                 buf1 = " "
              elseif (buf1 .eq. ",") then
                 ! - - -
                 i = i+1
                 call mygetarg(i,buf1)
              elseif(buf1(1:1) .eq. ",") then
                 ! - --
                 buf1(1:1) = " "
                 call rmblnk(buf1)
              end if
           end if

           read(buf,*, iostat=ierr) deltastop
           if (ierr.ne.0) then
              message = "expecting real after -s option, not:"//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           end if

           if (buf1 .eq. "au" .or. buf1 .eq. " ") then
              deltastopunit = 1
           elseif (buf1 .eq. "ev") then
              deltastopunit = 2
           elseif (buf1 .eq. "cm-1") then
              deltastopunit = 3
           elseif (buf1 .eq. "mev") then
              deltastopunit = 4
           else
              message = "Expecting one of 'au', 'eV', or 'cm-1' " &
                   //"after '-s <real>,', not "//trim(buf1)
              routine = "np2cpd_input"
              call errormsg
           end if

        elseif  ( buf .eq. '-n') then
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) maxiter
           if (ierr .ne. 0) then
              message = "Cannot read iterations after -n option. " &
                   // "Expecting integer, not: "//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           end if

        elseif  ( buf .eq. '-spd') then
           lspd = .true.
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) alpha
           if (ierr .ne. 0) then
              message = "Cannot read regularization parameter after " &
                     // "-spd option. Expecting real, not: "//trim(buf)
              routine = "np2cpd_input"
              call errormsg
           endif

        elseif  ( buf .eq. '-ev') then
           leigen = .true.
           i=i+1
           call mygetarg(i,buf)
           read(buf,*, iostat=ierr) alpha
           if (ierr .ne. 0) then
              message = "Cannot read regularization parameter after " &
                     // "-ev option. Expecting real, not: "//trim(buf)
              routine = "np2cpd_input"
             call errormsg
          endif

        elseif  ( buf .eq. '-v') then
           lvpot = .true.
        elseif  ( buf .eq. '-D'  .or.  buf .eq. '-DD') then
           if(buf.eq.'-DD') lddnp=.true.
           i=i+1
           call mygetarg(i,buf)
           name = trim(buf)
           if (len_trim(name) .eq. 0) then
              message = "Expecting string after -D option"
              routine = "np2cpd_input"
              call errormsg
           end if
           lname = .true.

        elseif  ( buf .eq. '-I') then
           i=i+1
           call mygetarg(i,buf)
           inpdir = trim(buf)
           if (len_trim(inpdir) .eq. 0) then
              message = "Expecting string after -I option"
              routine = "np2cpd_input"
              call errormsg
           end if
           linpdir = .true.
        else
           message = "Unknown option: "//trim(buf)
           routine = "np2cpd_input"
           call errormsg
        end if
     else
        message = "Unknown option: "//trim(buf)
        routine = "np2cpd_input"
        call errormsg
     end if
  end do ! arg-loop

  if (lspd .and. leigen) then
     routine = "np2cpd_input"
     message = "Only one of '-spd' and '-ev' can be used."
     call errormsg()
  end if

  if (lcontinue .and. linitrandom) then
     routine = "np2cpd_input"
     message = "Only one of '-R' and '-c' can be used."
     call errormsg()
  end if

  if (lspd) then
     isolver = 1
  end if
  if (leigen) then
     isolver = 2
  end if

!  if (lgrow .and. lcontinue) then
!     message = "-g option is not implemented to use together with -c or -C option."
!     routine = "np2cpd_input"
!     call errormsg
!  end  if

  if (lgrow .and. rank .le. 0) then
     message = "-g option requires -r option."
     routine = "np2cpd_input"
     call errormsg
  end  if

  ! check if name exists and is directory
  if (.not. lname) name = "."
  laenge = len_trim(name)
  call abspath(name,laenge)

  if (.not. linpdir) inpdir = "."
  leninpdir = len_trim(inpdir)
  call abspath(inpdir,leninpdir)

  if(lcc) then
     lencc = len_trim(ccfile)
     call abspath(ccfile,lencc)
     ccfile=trim(ccfile)//"/cpd"
  endif

  dvrfile = trim(inpdir)//       "/dvr"
  npotfile = trim(inpdir)//      "/natpot"

  if (lmnd) then
     inquire (file=trim(name), exist=lexists)
     if (lexists) then
        write(6,'(a)') "Name-directory already exists: "//trim(name)
     else
        inptit='mkdir '//trim(name)
        call excmd2(ierr, inptit)
        if (ierr.ne.0) then
           write(6,'(2a)') '### ERROR: ', &
                'Invalid path for name-directory (-mnd option). ###'
           call stopnow
        endif
        write(6,'(2a)') 'Created name-directory: ',trim(name)
     end if
  end if


  cpdfile = trim(name)//       "/cpd"
  if(lddnp) then
     logfile = trim(name)//       "/log"
     stopfile =  trim(name)//     "/stop"
     iterationsfile = trim(name)//"/iterations"
     timingfile = trim(name)//    "/timing"
  else
     logfile = trim(name)//       "/np2cpd.log"
     stopfile =  trim(name)//     "/np2cpd_stop"
     iterationsfile = trim(name)//"/np2cpd_iterations"
     timingfile = trim(name)//    "/np2cpd_timing"
  endif

  inquire (file=dvrfile, exist=lexists)
  if (.not. lexists ) then
     routine = "np2cpd_input"
     message = "File "//trim(dvrfile)//" not found."
     call errormsg
  end if

  inquire (file=npotfile, exist=lexists)
  if (.not. lexists ) then
     routine = "np2cpd_input"
     message = "File "//trim(npotfile)//" not found."
     call errormsg
  end if

  inquire (file=timingfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "np2cpd_input"
     message = "Cannot overwrite "//trim(timingfile)//". Use -w option."
     call errormsg
  end if

  inquire (file=cpdfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "np2cpd_input"
     message = "Cannot overwrite "//trim(cpdfile)//". Use -w option."
     call errormsg
  end if

  inquire (file=logfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "np2cpd_input"
     message = "Cannot overwrite "//trim(logfile)//". Use -w option."
     call errormsg
  end if

  inquire (file=iterationsfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "np2cpd_input"
     message = "Cannot overwrite "//trim(iterationsfile)//". Use -w option."
     call errormsg
  end if

  if (lwrcoeff) then
     coefffile = trim(name)//       "/cpdcoeff"
     inquire (file=coefffile, exist=lexists)
     if (lexists  .and. .not. lovwr) then
        routine = "np2cpd_input"
        message = "Cannot overwrite "//trim(coefffile)//". Use -w option."
        call errormsg
     end if
  else
     coefffile = " "
  end if

  if (lwrovlp) then
     ovlpfile = trim(name)//       "/cpdovlp"
     inquire (file=ovlpfile, exist=lexists)
     if (lexists  .and. .not. lovwr) then
        routine = "np2cpd_input"
        message = "Cannot overwrite "//trim(ovlpfile)//". Use -w option."
        call errormsg
     end if
  else
     ovlpfile = " "
  end if

  if (lvpot) then
     vpotfile = trim(inpdir)//       "/vpot"
     inquire (file=vpotfile, exist=lexists)
     if (.not. lexists) then
        routine = "np2cpd_input"
        message = "Cannot find "//trim(vpotfile)//". Ignoring -v option."
        call warnmsg
        lwarn = .false.
        lvpot = .false.
     end if
  else
     vpotfile = " "
  end if

  inquire (file=stopfile, exist=lexists)
  if (lexists  .and. .not. lovwr) then
     routine = "np2cpd_input"
     message = "Cannot remove old stopfile "//trim(stopfile)//". Use -w option."
     call errormsg
  else
     open(unit=istop, iostat=stat, file=stopfile, status='unknown')
     if (stat .ne. 0) then
        routine = "np2cpd_input"
        message = "Unknown error while opening file "//trim(stopfile)//"."
        call errormsg
     end if
     write(istop,'(a)',iostat=stat) "continue"
     if (stat .ne. 0) then
        routine = "np2cpd_input"
        message = "Unknown error while writing file "//trim(stopfile)//"."
        call errormsg
     end if
     close(istop)
  end if

  open(ilog, file=trim(logfile), iostat=ierr, form="formatted", &
       status="unknown")
  if (ierr .ne. 0) then
     routine = "np2cpd_input"
     message = "Unknown error while opening file "//trim(logfile)//"."
     call errormsg
  end if

  logisopen=.true.
  call fpstartup(buf(1:100))
  i=index(buf(1:100),'   ')-2
  write(ilog,'(a)') buf(1:i)
  call wrversion(ilog)
  if (len_trim(options) .le. 60) then
     write(ilog,'(2a/)')   'Command line: ',trim(options)
  else
     write(ilog,'(a/,a/)') 'Command line: ',trim(options)
  endif
  call myflush(ilog)


  select case (deltastopunit)
  case (1)
     continue
  case(2)
     deltastop = deltastop/au2ev
  case (3)
     deltastop = deltastop/au2ic
  case (4)
     deltastop = deltastop/(1.d3*au2ev)
  case default
     routine = "np2cpd_input"
     message = "Delta-Stop unit unknown. Please file a bug report."
     call errormsg
  end select


end subroutine np2cpd_input


subroutine np2cpd_log()

  implicit none

  write(ilog,'(a)') "Program parameters:"
  if (lgrow) then
     write(ilog,'(a,i8)')  " Using 'grow-option'"
     write(ilog,'(a,i8)')  " CPD max rank:           ", maxrank
     write(ilog,'(a,i8)')  " CPD start rank:         ", grank
     write(ilog,'(a,i8)')  " N-iter to rank increase:", giter
  else
     write(ilog,'(a,i8)')  " CPD max rank:           ", rank
  end if

  write(ilog,'(a,i8)')     " ALS maxiter:            ", maxiter
  write(ilog,'(a,ES16.8)') " ALS regularization:     ", alpha

  if (deltastop .eq. 0.d0) then
    write(ilog,'(a)') " Stop ALS if RMS <=   (not set)"
  else
     write(ilog,'(a,ES16.8,a)') " Stop ALS if RMS <= ", deltastop, " au"
     write(ilog,'(a,ES16.8,a)') "                    ", deltastop*au2ev, " eV"
     write(ilog,'(a,ES16.8,a)') "                    ", deltastop*au2ic, " cm-1"
  end if
  if (isolver.eq.1) then
     write(ilog,'(a,ES16.8)') " ALS solver:          Cholesky decomposition"
     write(ilog,'(a,ES16.8)') " ALS solver:          (uniform regularization)"
  elseif (isolver.eq.2) then
     write(ilog,'(a,ES16.8)') " ALS solver:          eigen-decomposition"
     write(ilog,'(a,ES16.8)') " ALS solver:          (non-uniform regularization)"
  end if
  if (lcontinue) then
     write(ilog,'(a)') " Initial CPD is read from file."
  else
     if (linitrandom) then
        write(ilog,'(a)') " Initial guess for CPD is random"
     else
        write(ilog,'(a)') " Initial guess for CPD are largest SPP/Coeff of natpot"
     end if
  end if
  write(ilog,'(a)') ""


  write(ilog,'(a/)') "Files:"
  write(ilog,'(a)')  "Input:"
  write(ilog,'(a)') " natpot         "//trim(npotfile)
  write(ilog,'(a)') " dvr            "//trim(dvrfile)
  if(lcc) &
  write(ilog,'(a)') " CPD            "//trim(ccfile)
  if (lvpot) &
  write(ilog,'(a)') " vpot           "//trim(vpotfile)

  write(ilog,'(/a)') "Output:"
  write(ilog,'(a)') " CPD            "//trim(cpdfile)
  write(ilog,'(a)') " log            "//trim(logfile)
  write(ilog,'(a)') " ALS iterations "//trim(iterationsfile)
  write(ilog,'(a)') " stop           "//trim(stopfile)
  write(ilog,'(a)') " timing         "//trim(timingfile)
  if (lwrovlp)  write(ilog,'(a)') &
       " overlap        "//trim(ovlpfile)
  if (lwrcoeff)  write(ilog,'(a)') &
       " coefficients   "//trim(coefffile)

  write(ilog,'(a)') ""

end subroutine np2cpd_log




! initial guess for candecomp:
! since there is no contacted mode any more
! (removed in the beginning) we can simply
! take the largest coefficients. The formerly
! contracted mode is taken as a cut
! through the core tensor
subroutine natpot2cpd_init(cpd, natpot, orig_modc)
  implicit none

  type(natpot_t), intent(IN)    :: natpot
  type(cpd_t), intent(INOUT)    :: cpd
  integer, intent(IN)           :: orig_modc
  integer                       :: rank,r,k,r1,m
  integer                       :: maxrank, cmode, cgdim
  integer                       :: vnidx(2),vndim(2)
  double precision              :: c
  double precision, allocatable :: cut(:)
  integer                       :: midx(natpot%nmode), pdim(natpot%nmode)
  integer, allocatable          :: idx(:)
  integer, save                 :: tid

  call get_timer(tid,'ptinitguess')
  call start_timer(tid)


  ! if initial guess are random numbers...
  if (linitrandom) then
     cpd%coeff = 1.d0
     do m=1,cpd%nmode
        call random_number(cpd%spp(m)%htm)
     end do

     call normspp(cpd)
     cpd%coeff = 1.d0
     call stop_timer(tid)
     return
  end if

  ! extract from natpot:

  allocate(idx(cpd%rank))
  idx = 0

  if (orig_modc.ne.0) then
     cmode = orig_modc
     cgdim = natpot%potdim(cmode)
  else
     cmode = 0
     cgdim = 0
     do m=1,natpot%nmode
        if (natpot%potdim(m) .gt. cgdim) then
           cmode = m
           cgdim = natpot%potdim(m)
        end if
     end do
  end if
  allocate(cut(cgdim))

  vndim(1) = natpot%potvdim(cmode)
  vndim(2) = natpot%potndim(cmode)
  maxrank = natpot%potvdim(cmode)*natpot%potndim(cmode)

  k = 1
  call cindex2sindex(k,vnidx,vndim,2)
  call cuttxd1(natpot%coeff,cut, &
       vndim(1),cgdim,vndim(2), &
       vnidx(1),vnidx(2))

  rank = 1
  cpd%coeff(1) = dsqrt(dot_product(cut,cut))
  idx(1) = 1

  cpd%spp(cmode)%htm(:,rank)  = cut/cpd%coeff(1)

  do k=2,maxrank
     call cindex2sindex(k,vnidx,vndim,2)
     call cuttxd1(natpot%coeff,cut, &
          vndim(1),cgdim,vndim(2), &
          vnidx(1),vnidx(2))

     c = dsqrt(dot_product(cut,cut))


     if (abs(cpd%coeff(rank)) .lt. abs(c)) then
        ! insert new coeff in the middle

        if (rank .lt. cpd%rank) then
           rank=rank+1
           cpd%coeff(rank) = 0.d0
        end if

        ! where to insert
        r1=1
        do r=rank,1,-1
           if (abs(cpd%coeff(r)) .gt. abs(c)) then
              r1 = r+1
              exit
           end if
        end do
        if (r1 .gt. rank) then
           print*, "This part of the code should never be reached"
        end if

        ! copy smaller terms one position back
        do r=rank-1,r1,-1
           cpd%coeff(r+1) = cpd%coeff(r)

           cpd%spp(cmode)%htm(:,r+1) =  cpd%spp(m)%htm(:,r)

           idx(r+1) = idx(r)
        end do

        cpd%coeff(r1) = c

        cpd%spp(cmode)%htm(:,r1) = cut/c
        idx(r1) = k

     else if( rank .lt. cpd%rank) then
        ! append at the end if there is space
        rank = rank + 1
        cpd%coeff(rank) = c
        cpd%spp(cmode)%htm(:,rank) = cut/c
        idx(rank) = k
     end if

  end do

  deallocate(cut)

  ! we have the coefficients. now add basis
  ! -> they are just deltas on the spp grid
  ! where we did not use a cut

  pdim = natpot%potdim
  pdim(cmode) = 1
  do r=1,cpd%rank

     call cindex2sindex(idx(r),midx,pdim,natpot%nmode)

     do m=1,natpot%nmode
        if (m.ne.cmode) then
           cpd%spp(m)%htm(midx(m),r) = 1.d0
        end if
     end do
  end do


  call stop_timer(tid)

end subroutine natpot2cpd_init



logical function np2cpdstop()

  implicit none
  logical  :: lexists
  character(len=c4) :: content
  integer :: ilbl


  inquire(file=stopfile, exist=lexists)
  if (.not. lexists) then
     np2cpdstop = .true.
  else
     open(istop, file=stopfile, err=398)
     read(istop,*,err=399) content
     close(istop)
     call cncntrblnk(content)
     content = adjustl(content)
     ilbl = index(content,' ')
     if (content(1:ilbl) .eq. "continue") then
        np2cpdstop = .false.
     else
        np2cpdstop = .true.
     end if
  end if

  return

398 continue
  ! read or open-error
  message = "Could not open stop file: "//trim(stopfile)
  routine = "np2cpdstop"
  call warnmsg
  np2cpdstop = .true.
  return
399 continue
  ! read or open-error
  message = "Could not read stop file: "//trim(stopfile)
  routine = "np2cpdstop"
  call warnmsg
  np2cpdstop = .true.
  return

end function np2cpdstop



subroutine np2cpd_wrphelptext()
  implicit none

  write(6,*)
  write(6,'(a)')  'Purpose: Read a natpot file and convert to a'
  write(6,'(a)')  '         canonical polyadic decomposition (CPD)'
  write(6,'(a)')  '         which is subsequently optimized using ALS.'
  write(6,*)
  write(6,'(a)')  '  Usage:  natpot2cpd<vers><d/D> [-h|-?]  [-ver -o -l -c -C -CC'
  write(6,'(32x,a)')                   ' -r -R -s -n -g -w -spd -ev -d -v -D -DD]'
  write(6,'(a)')  '   vers:  Program version (e.g. 84 for version 8.4)'
  write(6,'(a)')  '    d/D:  Indicates a debug version'
  write(6,*)
  write(6,'(a)')  ' optn: -h           : Print this help-text.'
  write(6,'(a)')  '       -?           : Print this help-text.'
  write(6,'(a)')  '       -ver         : Version information.'
  write(6,*)
  write(6,'(2a)') '       -mnd         : Make Name Directory.', &
       'Creates the name directory, if not existent.'
  write(6,'(a)')  '       -w           : Allow overwrite.'
  write(6,'(a)')  '       -c           : Continue obtimizing ' &
                                     // 'an existing CPD file.'
  write(6,'(a)')  '       -C           : Same as -c but reset iteration counter.'
  write(6,'(a)')  '       -CC <dir>    : Same as -C but reads the initial CPD-file ' &
                                     // 'from directory <dir>.'
  write(6,'(2a)') '       -r <int>     : CPD rank,', &
                                       ' default: sum of SPP of natpot'
  write(6,'(a)')  '                      or rank of read-in CPD.'
  write(6,'(a)')  '       -R           : Initial guess of CPD are random numbers,'
  write(6,'(a)')  '                      default: pick largest coeff and SPP'
  write(6,'(a)')  '                      from natpot or use read-in CPD.'
  write(6,'(2a)') '       -a <real>    : Regularization, default: ', &
                                       ' sqrt(machine_precision).'
  write(6,'(2a)') '       -s <real>[,S]: Stop if |natpot-cpd| is ', &
                                        'less than <real>.'
  write(6,'(2a)') '                      Default: not set. <real> may bear a'
  write(6,'(a)')  '                      unit S [au, eV, meV, cm-1]. default a.u.'
  write(6,'(a)')  '       -o           : Write overlap matrix of the cpd on exit.'
  write(6,'(a)')  '                      Default: not set.'
  write(6,'(a)')  '       -l           : Write coefficients of the cpd on exit.'
  write(6,'(a)')  '                      Default: not set.'

  write(6,'(a)')  '       -spd <real>  : Solve ALS using Cholesky decomposition with'
  write(6,'(a)')  '                      uniform regularization on the diagonal.'
  write(6,'(a)')  '                      If neither -spd nor -ev is set, then -spd with'
  write(6,'(a)')  '                      regularization=sqrt(machine_precision) is assumed.'
  write(6,'(a)')  '       -ev <real>   : Solve ALS using eigen-decomposition with'
  write(6,'(a)')  '                      regularization of smallest eigenvalues.'
  write(6,'(a)')  '       -d <real>[,S]: Use <real> as error-estimate when de-contracting'
  write(6,'(a)')  '                      the natpot D-tensor. <real> may bear a'
  write(6,'(a)')  '                      unit S [au, eV, meV, cm-1]. default a.u.'
  write(6,'(2a)') '       -n <int>     : Maximum number of', &
       ' iterations of ALS optimizer, default 1000.'
  write(6,'(a)')  '       -g <int int1>: Do not start with full rank but with rank <int> '
  write(6,'(a)')  '                      and increase every <int1> iterations by the same amount.'
  write(6,'(a)')  '                      If -g is set, -r must be set too, to set a maximal rank.'
  write(6,'(a)')  '       -v           : If a vpot file is present, calculate the '
  write(6,'(a)')  '                      error of the cpd against the vpot in the end.'
  write(6,*)
  write(6,'(2a)') '       -I <dir>     : Read natpot and dvr', &
       ' from <dir>. Default: current directory'
  write(6,'(2a)') '       -D <name>    : Write output to <name>.', &
       ' Default: current directory.'
  write(6,'(a)') '       -DD <name>   : Similar to -D, but additionally drops'
  write(6,'(a)') '                      the prefix np2cpd from created files.'
  write(6,'(/a)') '        The program supports OpenMP (compile with -P option)'


  call stopnow()


end subroutine np2cpd_wrphelptext

end program natpot2cpd

