!***********************************************************************
!
!             Analyse program PEXPECT
!
! This program reads the pdensity-file and a specified operator and
! computes the expectation value o(t) = <psi(t)|operator|psi(t)>,
! where the operator must operate on one particle only.
! The program outputs or plots o(t).
!
!     10/00 HDM
!
!
!***********************************************************************
      
      program pexpect

use decimal, only: long,dop,sip
use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use griddatmod
use psidef
use operdef
use hpsimod
use channels
use datenmod

use xvlib, only: vvaxzz
use iofile, only: rdrhoinfo
use rddvrmod, only: dvrinfo
use rdopermod, only: operinfo
use zeigauswmod, only: zeigausw
use memhopsmod, only: memhops
use zeighpsimod, only: operflag,zeighpsi
use psiutil, only: maxmatwk

use module_interface

      implicit none

      integer(long), external           :: slen
      integer(long)                     :: chkdvr,chkgrd,chkpsi,chkdat,ilbl,n,nham,s,m
      character(len=c5)                 :: filename
      character(len=10)                 :: stat
      logical(kind=4)                   :: lerr,lexist

complex(dop), dimension(:), allocatable    :: mc
real(dop), dimension(:), allocatable       :: mr
integer(long), dimension(:), allocatable   :: mi
logical(kind=4), dimension(:), allocatable :: ml
complex(sip), dimension(:), allocatable    :: ms
real(sip), dimension(:), allocatable       :: mf


!-----------------------------------------------------------------------
! Initialize variables and set defaults
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.true.

!-----------------------------------------------------------------------
! Declare resources required (other than those needed to read a file)
!
! ncomp   : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi    : wavefunction
! lpsi1   : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat   : density matrices for psi
! ldmat1  : density matrices for psi1
! lort    : grid coordinates
! lgpop   : grid populations
! ltrafo  : FBR/DVR trafo matrices
!-----------------------------------------------------------------------
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat=.true.
      ldmat1=.false.
      lort=.true.
      lgpop=.false.
      ltrafo=.true.

!-----------------------------------------------------------------------
! read input parameters and options
! open files
!-----------------------------------------------------------------------
      call pexpinput(stat)

!-----------------------------------------------------------------------
! open input and output data files
!
! use standard channels if possible
!  ipden : pdensity file
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------
      filename=filein
      open(ipden,file=filename,form='unformatted',status='old',err=1000)
      if (fileout .ne. 'no') then
         filename=fileout
         open(iaus,file=filename,form='formatted',&
             status=stat,err=1000)
      else
         iaus=6
      endif

!-----------------------------------------------------------------------
! write gnuplot info to file
!-----------------------------------------------------------------------
      if (lgnu) then
         write(iaus,'(a,i2,3a)') 'set title "mode =',indx,&
                            ',  ',fileout(1:laeout),'"'
         write(iaus,'(a)')  'set style data lines'
         write(iaus,'(a)')  'set nologscale'
         write(iaus,'(3a)') 'set ylabel "',unitlab(1),'"'
         write(iaus,'(a)')  'set xlabel "Time [fs]"'
         write(iaus,'(3a)') 'plot "-" using 1:2 notitle'
         write(iaus,'(a)')  '   '
      end if

!-----------------------------------------------------------------------
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      filename=filein
      read(ipden,err=999) filever(ipden)
      call rdrhoinfo(ipden,chkdvr,chkgrd,chkdat)
      read(ipden) nmode, nstate, ((dim(m,s),m=1,nmode),s=1,nstate)
      if(indx.gt.nmode) then
         write(6,'(2a/,2i6)') ' The  first argument (mode)',&
              ' must not be larger than nmode.', indx, nmode
         stop
      end if
      
      dgldim = 1
      do s = 1, nstate
         do m = 1, nmode
           dgldim = max(dgldim,subdim(m)*dim(m,s)) 
        enddo
      enddo 

!-----------------------------------------------------------------------
! read information from system files
!-----------------------------------------------------------------------
      if (lrddvr) then
         chkdvr=2
         filename=dname(1:dlaenge)//'/dvr'
         open(idvr,file=filename,form='unformatted',status='old',&
              err=1000)
         call dvrinfo(lerr,chkdvr)
         close(idvr)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
      if (lrdoper) then
         chkdvr=2
         chkgrd=2
         filename=operfile
         open(ioper,file=filename,form='unformatted',status='old',&
              err=1000)
         call operinfo(lerr,chkdvr,chkgrd)
         close(ioper)
         call chksyserr(filename,lerr)
         if (lerr) go to 999
      endif
         
!-----------------------------------------------------------------------
! get operator name
!-----------------------------------------------------------------------
      if (operator .eq. ' ') then
         write(6,*) 'No operator specified! Use -h option to see usage.'
         write(6,'(1x,a)') 'Name of operator ? Choose from:'
         do n=1,hamdim
            write(6,'(2x,a)') hamlab(n)
         enddo
         read(5,'(a)') operator
         if (operator .eq. ' ') stop
      endif

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
 20   nham=0
      do n=1,hamdim
         if (hamlab(n) .eq. operator) nham=n
      enddo

      if (nham .eq. 0) then
         write(6,'(4a)')'Operator "',&
           operator(1:slen(operator)),'" not found in ',&
           operfile(1:slen(operfile))
         write(6,'(1x,a)') 'Name of operator ? Choose from:'
         do n=1,hamdim
            write(6,'(2x,a)') hamlab(n)
         enddo
         read(5,'(a)') operator
         if (operator .eq. ' ') stop
         goto 20
      endif

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

!-----------------------------------------------------------------------
! set up various flags and array information needed to use an operator
!-----------------------------------------------------------------------
      call operflag(nham)
      call zeighpsi(nham)

      mchunpsi=mcdim
      mcdim=mchunpsi+dgldim


!-----------------------------------------------------------------------
! assign work arrays
! NB work arrays must be at end of mc, mr, mi, ml arrays.
!-----------------------------------------------------------------------
      call maxmatwk(nham)

      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
!  assign memory for operator
!-----------------------------------------------------------------------
      if (lrdoper) call memhops

!-----------------------------------------------------------------------
! call analysis routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))

call receiver(mc,mr,mi,ml,ms,mf)

deallocate(mc,mr,mi,ml,ms,mf)


      stop
  999 continue
      write(6,'(2a)') 'Error reading :', filename
      stop
 1000 continue
      ilbl=index(filename,' ')-1
      inquire(file=filename(1:ilbl),exist=lexist)
      if(lexist) write(6,'(2a,/a)') &
          ' The output file: ',filename(1:ilbl),' already exists!'&
         ,' overwrite disabled? (Use -w option to enable overwrite)'
      write(6,'(2a)') ' Cannot open file: ',filename(1:ilbl)
      stop
      contains

!#######################################################################

      subroutine receiver (mc, mr, mi, ml, ms, mf)

use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use griddatmod
use psidef
use datenmod

      implicit none

      complex(dop), dimension(mcdim), intent(inout)    :: mc
      real(dop), dimension(mrdim), intent(inout)       :: mr
      integer(long), dimension(midim), intent(inout)   :: mi
      logical(kind=4), dimension(mldim), intent(inout) :: ml
      complex(sip), dimension(msdim), intent(inout)    :: ms
      real(sip), dimension(mfdim), intent(inout)       :: mf

!-----------------------------------------------------------------------
! provide arguments of pointers to arrays needed for analysis
!  e.g. mc(mcpsi) for psi array
!-----------------------------------------------------------------------
      call subpexpect(mc(mcpsi),mr(mrort),mc(mchin),mc(mcrueck),&
           mi(mifftfak),mr(mrhops),ms(mspsi),mi(mijindx),mc(mchunpsi),&
           mc(mcworkc),mc(mchc),mr(mrhr),mi(mihi),ml(mlhl))

      return
      end subroutine receiver
      
!***********************************************************************
!
!     HERE IS THE SUBROUTINE TO DO THE ANALYSIS
!
!***********************************************************************

      subroutine subpexpect (psi,ort,hin,rueck,fftfak,hops,spsi,jindx,&
                 hunpsi,workc,hc,hr,hi,hl)

use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use griddatmod
use psidef
use operdef
use hpsimod
use channels
use datenmod
use dirdyn

use rddvrmod, only: rddvr
use rdopermod, only: rdoper
use tdkoeffs, only: calctdk
use hphimod, only: hunphi1ms

      implicit none

      complex(dop), dimension(*), intent(inout) :: hc
      real(dop), dimension(*), intent(in)       :: hr
      integer(long), dimension(*), intent(in)   :: hi
      logical(kind=4), dimension(*), intent(in) :: hl

      integer(long), external                       :: slen
      integer(long)                                 :: chkdvr,chkgrd,n,hktype,nham,s,m,k,khz,zeig,j
      integer(long), dimension(1)                   :: idum
      integer(long), dimension(fftdim), intent(out) :: fftfak
      integer(long), dimension(adim), intent(out)   :: jindx
      real(dop)                                     :: dt,time,aovl
      real(dop), dimension(1)                       :: rdum
      real(dop), dimension(maxdim)                  :: avec
      real(dop), dimension(ortdim), intent(out)     :: ort
      real(dop), dimension(hopsdim), intent(out)    :: hops

      complex(dop)                                 :: expval,sum
      complex(dop), dimension(1)                   :: cdum
      complex(dop), dimension(1)                   :: zdum
      complex(sip), dimension(spsidim), intent(in) :: spsi
      complex(dop), dimension(dgldim), intent(out) :: psi
      complex(dop), dimension(fftdim), intent(out) :: hin,rueck
      complex(dop), dimension(dgldim), intent(out) :: hunpsi
      complex(dop), dimension(dgldim), intent(out) :: workc
      character(len=120)                           :: gnucmd
      logical(kind=4)                              :: lexist,limag
      logical(kind=4), dimension(1)                :: ldum

!-----------------------------------------------------------------------
! read data from dvr file if desired
! edit which data (set dvrdata(x) to .true. and substitute dummy array)
!-----------------------------------------------------------------------
      if (lrddvr) then
         dvrdata(1)=.true.
         dvrdata(2)=.false.
         dvrdata(4)=.false.
         dvrdata(5)=.true.
         dvrdata(6)=.true.
         dvrdata(7)=.true.
         dvrdata(8)=.false.
         dvrdata(9)=.false.
         dvrdata(11)=.false.
         dvrdata(12)=.false.
         chkdvr=2
         call rddvr(ort,rdum,rdum,rdum,hin,&
           rueck,fftfak,zdum,zdum,idum,idum,rdum,chkdvr)
      endif

!-----------------------------------------------------------------------
! read data from oper file.
!-----------------------------------------------------------------------
      if (lrdoper) then
         chkdvr=1
         chkgrd=1
         call rdoper(hops,chkdvr,chkgrd)
      endif

!-----------------------------------------------------------------------
! Get operator number
!-----------------------------------------------------------------------
      nham=0
      do n=1,hamdim
         if (hamlab(n) .eq. operator) nham=n
      enddo

      write(ilog,'(a,i3,a,l2,a,i3)')  ' nham =',nham,&
           '   nodiag =',nodiag(nham),'     m =',indx
      if(nodiag(nham)) then
         write(ilog,'(a)')' Operator is "nodiag" but must be "usediag".'
         write(6,'(a)')   ' Operator is "nodiag" but must be "usediag".'
         stop
      endif

      write(iaus,'(/,2a)') '# Expectation value of operator : ',&
            operator(1:slen(operator))
      write(iaus,'(2a)') '#    Taken from file :',&
            operfile(1:slen(operfile))
      write(iaus,'(2a)') '#    Units : ',unitlab(1)
      if(lnw) then
         write(iaus,'(a)') '#  ---  Not normalised!  ---'
         write(iaus,'(/,a11,2a22)')&
              '# Time[fs] ','  Real part','Imaginary part'
      else
         write(iaus,'(/,a11,2a22,a21)')&
              '# Time[fs] ','  Real part','Imaginary part','A-Norm^2'
      end if
      call myflush(iaus)

      dt     = out1/fs
      time   = tinit/fs
      limag  = .false.

      khz = 0
      do s = 1, nstate
         write(ilog,'(a,i3,a,i3,4x,a,i4,a,i4)')' s =',s,'   khzahl =',&
                       khzahl(indx,s,nham),'  k1 =',zhun(indx,s,nham),&
                     '   k2 =',zhun(indx,s,nham)+khzahl(indx,s,nham)-1
         khz = khz + khzahl(indx,s,nham)
      enddo
      if(khz.le.0) then
         write(ilog,'(a)')' WARNING: There is no uncorrelated operator!'
         write(6   ,'(a)')' WARNING: There is no uncorrelated operator!'
      endif 
!-----------------------------------------------------------------------
! loop over steps
!-----------------------------------------------------------------------
   10 continue
      expval = (0.0_dop,0.0_dop)
      aovl   = (0.0_dop,0.0_dop)
   11 read(ipden,end=77) s,m, (avec(k),k=1,dim(m,s))
      read(ipden,end=77) (psi(k),k=1,subdim(m)*dim(m,s))
      if( m .ne. indx ) goto 11

      hktype=0

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
      if (ltdh(nham)) call calctdk(time,nham,psi,cdum,rdum,idum,ldum)

      call hunphi1ms (psi,hunpsi,workc,m,s,dim(m,s),subdim(m),&
                       hc,hr,hi,hl,hktype,nham)

      do j = 1, dim(m,s)
         zeig=(j-1)*subdim(m)+1
         call vvaxzz(psi(zeig),hunpsi(zeig),sum,subdim(m))
         expval = expval + avec(j)*sum
         aovl   = aovl +  avec(j)
      end do
      if( s .ne. nstate ) goto 11

!-----------------------------------------------------------------------
! normalise expectation value by division through A-norm^2.
!-----------------------------------------------------------------------
      if(.not.lnw) then
         expval=expval/aovl
      end if

!-----------------------------------------------------------------------
! write values, continue or terminate the loop over data-sets.
!-----------------------------------------------------------------------

      if(lnw) then
         write(iaus,'(f10.3,1x,2f22.14)') &
              time,dble(expval)*eunit,dimag(expval)*eunit
      else
         write(iaus,'(f10.3,1x,2f22.14,6x,f15.12)') &
              time,dble(expval)*eunit,dimag(expval)*eunit,aovl 
      end if
      if(abs(dimag(expval)*eunit) .gt. 1.0e-12_dop ) limag = .true.
      time=time+dt

      do n=1,step-1
 12      read(ipden,end=77) s,m, (avec(k),k=1,dim(m,s))
         read(ipden,end=77) (psi(k),k=1,subdim(m)*dim(m,s))
         if( m .ne. indx ) goto 12
         time=time+dt
      enddo
      if (time .gt. tfin) go to 77

      go to 10


 77   continue
      close(ipden)
      close(iaus)
      if(limag) write(6,'(a)') &
               ' Warning: The expectation value is complex!'
      if( time .eq. tinit/fs ) then
         write(6,'(a,i3,a)') ' No data sets for mode',indx,&
                             ' on pdensity file.'
         stop
      endif

!-----------------------------------------------------------------------
! Plot and/or print the data.
!-----------------------------------------------------------------------
      if(lprint) then
         inquire(file="printplot",exist=lexist)
         open(iaus,file="printplot",form='formatted',status='unknown')
         if(.not.lexist) then
            open(iaus,file="printplot",form='formatted')
            write(iaus,'(a)') 'set terminal postscript'
            write(iaus,'(a)')'set output "| lpr" '
            write(6,'(a)')'File printplot created.'
         end if
         rewind iaus
         read(iaus,'(a1)') gnucmd
         read(iaus,'(a30)') gnucmd
         close(iaus)
         write(6,'(a)')'Press enter twice to print. Ctrl C to abort.'
         write(6,'(3a)')'Abort and edit printplot if you printcommand',&
                        ' is not: '//gnucmd(11:29)
      end if
      if(lagnu) then
! This may cause problems when running on Mac
!         gnucmd = 'gnuplot -geometry +0+0 -persist "'&
         gnucmd = 'gnuplot -geometry "'&
          //fileout(1:slen(fileout))//'"'
         call excmd(gnucmd)
      end if
      if(lprint) then
         gnucmd = 'gnuplot -geometry '&
         //'"printplot"  "'//fileout(1:slen(fileout))//'"'
         call excmd(gnucmd)
      end if

      return
      end subroutine subpexpect

!***********************************************************************
!-----------------------------------------------------------------------
! EXPINPUT reads the options and the argument.
!-----------------------------------------------------------------------
      subroutine pexpinput(stat)

use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use griddatmod
use psidef
use hpsimod
use channels
use datenmod

      implicit none

      integer(long), external       :: myiargc
      integer(long)                 :: iarg,i,ierr
      character(len=80)             :: buf
      character(len=*), intent(out) :: stat

!-----------------------------------------------------------------------
! provide default input and output filenames 
!-----------------------------------------------------------------------
      filein  = 'pdensity'
      operfile = 'oper'
      fileout = 'pexpect.pl'
      operator = ' '
      unitlab(1)='a.u.'

!-----------------------------------------------------------------------
! set default flags
!
! stat : status of output file   
!-----------------------------------------------------------------------
      stat   = 'new'
      lgnu   = .false.
      step   = 1
      eunit  = 1.0
      lgnu   = .false.
      lagnu  = .false.
      lprint = .false.
      lnw    = .false.
      tfin   = 1.0e+50_dop

!-----------------------------------------------------------------------
! read options 
!-----------------------------------------------------------------------
      iarg = myiargc() 
      if (iarg .lt. 1) goto 200 

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf(1:2) .eq. '-f') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if(buf(1:2) .eq. '-i') then
            i = i+1
            call mygetarg(i,buf)
            filein = buf
         else if( buf(1:2) .eq. '-o' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
         else if (buf(1:3) .eq. '-r' ) then
            lnw = .true.
         else if (buf(1:3) .eq. '-t' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*) tfin
         else if (buf(1:3) .eq. '-u' ) then
            i = i+1
            call mygetarg(i,buf)
            call keylowc(buf)
            eunit = 1.0
            call keyunits(eunit,buf,ierr)
            if (ierr .ne. 0) then
               write(6,'(a)') 'Unknown Units with -u :'//trim(buf)
               stop
            endif
            unitlab(1)=trim(buf)
            eunit=1.0/eunit
         else if( buf(1:2) .eq. '-n' ) then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (ierr .ne. 0  .or.  step .le. 0) then
               write(6,'(2a/,a)')&
                 ' The  argument following the option -n',&
                 ' must be a positive integer. Not : ', trim(buf)
               stop
            end if
         else if( buf(1:2) .eq. '-w' ) then
            lovwr=.true.
         else if( buf(1:2) .eq. '-g' ) then
            lgnu = .true.
         else if( buf(1:3) .eq. '-a ' ) then
            lgnu  = .true.
            lagnu = .true.
            stat  = 'unknown'
         else if( buf(1:3) .eq. '-p ' ) then
            lgnu   = .true.
            lprint = .true.
         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(78a1)')  ('-',i=1,78)
            write(6,'(1x,2a,/1x,3a)')&
       'Purpose: ',&
       'Calculates the time evolution of an expectation value.',&
       'Usage  : ',&
       'pexpect<ver> [-f -i -o -g -a -r -t -p -u -n -w -ver -h -?] ',&
       '[mode operator].'
            write(6,'(/,a)') 'Options : '
            write(6,'(2a,/,9x,a)') &
                     ' -f FILE : The operator is read '&
                    ,'from file FILE rather than from ./oper'&
            ,'  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a,/,9x,a)') &
                     ' -i FILE : The one particle density is read '&
                    ,'from file FILE rather'&
                    ,'  than from ./pdensity'&
            ,'  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,11x,a/,11x,a)') &
                     ' -o FILE : The output is written to file ',&
                     'FILE.pl rather than to ./pexpect.pl',&
             'The string FILE may be a relative or a full path-name.',&
             'If FILE=no, i.e. "-o no", no output file will be opened.'
             write(6,'(2a)') &
                     ' -g      : GNUplot comand lines are written to ',&
                     'the output file(s).'
            write(6,'(2a)')&
                     ' -a      : GNUplot is called automatically.',&
                     ' (sets -g and -w).'
            write(6,'(a)')&
                     ' -p      : Print GNUplot to lpr (sets -g). ' 
            write(6,'(2a)')&
                     ' -r      : Do not divide the expectation values',&
                     ' by norm^2.' 
            write(6,'(2a)')&
                     ' -t tfin : The expectation values are computed',&
                     ' only up to time=tfin.'
            write(6,'(2a)')&
                     ' -u UNIT : The unit used is UNITs',&
                     ' (Default is a.u.).'
            write(6,'(2a)') ' -n step : Compute the property', &
           ' only every step-th output.'
            write(6,'(2a)') &
                     ' -w      : An existing property file is ',&
                     'overwritten.'
            write(6,'(a)')&
                     ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(/2a/,2a)') ' The first argument "mode" is the',&
                                 ' (combined) mode for which the',&
                                 ' expectation value is calculated.' 
            write(6,'(2a/,2a)')  ' The second argument "operator" is',&
                      ' the name of the operator, as specified in an',&
                      ' HAMILTONIAN-SECTION_xxx, for which the',&
                      ' expectation value is required.' 
            write(6,'(2a/,a)')  ' If "system" is chosen as operator',&
                      ' the uncorrelated energy of the specified',&
                      ' mode is computed.'
            write(6,'(2a)') ' If no argument is given, ',&
                 'the user is prompted for the missing argument(s).'  
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         if (iarg .le. i) goto 200
         goto 100
      end if
      if (iarg .lt. i) goto 200
      
!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      open (unit=ilog,file='pexpect.log',form='formatted')
      logisopen=.true.

!-----------------------------------------------------------------------
! get compulsory argument.
!-----------------------------------------------------------------------
      if (iarg .ge. i) then
         call mygetarg(i,buf)
         read(buf,*,iostat=ierr) indx
         if(ierr.ne.0 .or. indx.le.0 ) then
            write(6,'(2a/,a)') ' The  first argument (mode)',&
                 ' must be a positive integer. Not : ', buf
            stop
         end if
      end if

      i=i+1
      if (iarg .eq. i) call mygetarg(i,operator)

!-----------------------------------------------------------------------
! End of loop over options and arguments.
!-----------------------------------------------------------------------
 200  continue

      if( indx .eq. 0 ) then
         write(6,'(a)') ' mode = ?'
         read(5,*,iostat=ierr) indx
         if(ierr.ne.0 .or. indx.le.0 ) then
            write(6,'(2a/,a)') ' The  first argument (mode)',&
                 ' must be a positive integer.'
            stop
         end if
      end if


!-----------------------------------------------------------------------
! set status variable to overwrite files
!-----------------------------------------------------------------------
      if (lovwr) stat='unknown'

!-----------------------------------------------------------------------
! convert filenames to absolute
!-----------------------------------------------------------------------
      laenge=index(filein,' ')-1
      call abspath(filein,laenge)

      opflaenge=index(operfile,' ')-1
      call abspath(operfile,opflaenge)

      if (fileout .ne. 'no') then
         laeout=index(fileout,'  ')-1
         call abspath(fileout,laeout)
      end if

      return 
      end subroutine pexpinput

!#######################################################################

      end program pexpect

