!***********************************************************************
!
!                         SUMRST
!
! This program can be used to calculate the superposition of
! up to 8 wavefunctions which are read from restart files. The
! program requires an input file with RUN-SECTION and INIT_WF-SECTION.
! 
! ALGORYTHM:
!
!  0 --  Scan input files and determine number of SPFs of output as 
!        dim_out(m,s)=max(dim_allinputs(m,s)).
!
!  1 --  Read first restart (psi2) and multiply with coefficient.
!  2 --  Read next restart  (psi1) and multiply with coefficient.
!  3 --  Construct a SPF-basis from SPFs of psi1 and psi2
!        while skipping linear dependent SPFs.
!  4 --  Transform psi1 and psi2 to the new SPF-basis.
!  5 --  Add psi1 and psi2  (spsi).
!  6 --  Transform spsi to natural orbitals.
!  7 --  Reduce SPF basis of spsi to min( dim_out(m,s),dim_spsi(m,s)) .
!  8 --  Copy spsi to psi2.
!  9 --  If files left then goto 2.
! 10 --  Check superposition and write restart.
! 11 --  Done.
!
!***********************************************************************
      
      program sumrst

use decimal, only: long,dop,sip
use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use griddatmod
use psidef
use channels
use datenmod
use overlapmod
use compdatmod
use compdat1mod
use sumrstmod
use runinwfmod

use op1lib, only: cpvxz, normvxz, zerovxz, zeromxz, cpmxi, zeromxi
use mmlib, only: mmaxzz
use mtlib, only: mtxxzz
use xvlib, only: xvixdzo, vvaxzz, xvxxzzr, xvxxzzo
use schmidtorthomod, only: schmidtortho
use logdat, only: errormsg, warnmsg, wrline, cntrl
use compdat, only: savecpsi, savecsys
use iopsidef, only: psidat
use iorst, only: rstinfo
use zeigauswmod, only: zeigausw

      implicit none

      integer(long) :: ilbl,i

complex(dop), dimension(:), allocatable    :: mc
real(dop), dimension(:), allocatable       :: mr
integer(long), dimension(:), allocatable   :: mi
logical(kind=4), dimension(:), allocatable :: ml
complex(sip), dimension(:), allocatable    :: ms
real(sip), dimension(:), allocatable       :: mf
     
!-----------------------------------------------------------------------
! Initialize variables and set defaults
!-----------------------------------------------------------------------
      ncomp=2
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file
!     lrdoper   OPER file
!-----------------------------------------------------------------------
      lrddvr=.false.
      lrdoper=.false.

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
! lsphfbr : jfbr, mfbr and kinfbr matrices.
!-----------------------------------------------------------------------
      ncomp=2
      lpsi=.true.
      lpsi1=.true.
      lpsigrd=.false.
      ldmat=.true.
      ldmat1=.false.
      lort=.false.
      lgpop=.false.
      ltrafo=.false.
      lsphfbr=.false.

!-----------------------------------------------------------------------
! read input parameters and options, open files
!-----------------------------------------------------------------------
      call suminput
    
!-----------------------------------------------------------------------
!     pointers and memory for constructing the superposition
!-----------------------------------------------------------------------
      call srdrstinfo

!-----------------------------------------------------------------------
!     memory for two input WFs
!     (the density matrices are for the temporary WF and not used for 
!     the input WFs)
!-----------------------------------------------------------------------
      call zeigausw
   
      dmatdim=sdmatdim   
      d3matdim=sd3matdim
      dgldim=maxdgldim
      call smemausw
      
!-----------------------------------------------------------------------
!     memory for a temporary larger WF
!-----------------------------------------------------------------------
      mcspsi=mcdim
      mcdim=mcdim+smaxdgldim

      mcdover=mcdim
      mcdim=mcdim+doverdim
      
      workcdim=max(workcdim,smaxdgldim,2*smaxblock,2*smaxspf)
      workrdim=workrdim

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! assign work arrays
! NB work arrays must be at end of mc, mr, mi, ml arrays.
!-----------------------------------------------------------------------
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

!-----------------------------------------------------------------------
! call routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
      call myflush(ilog)

maxdim=32
maxsta=6
allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(svdim(maxdim,maxsta))
allocate(spdim(maxdim,maxsta))
allocate(sndim(maxdim,maxsta))
allocate(szetf(maxdim,maxsta))
allocate(szpsi(maxsta))
allocate(sblock(maxsta))
allocate(mcgs2(gs2matdim))
allocate(mcdicht1(dmatdim))
allocate(mcdicht2(dmatdim))
allocate(mcdicht4(dmatdim))

call receiver(mc,mr,mi,ml,ms,mf,mcgs2,mcdicht1,mcdicht2,mcdicht4)

deallocate(mc,mr,mi,ml,ms,mf,mcgs2,mcdicht1,mcdicht2,mcdicht4)
deallocate(svdim, spdim, sndim, szetf)
deallocate(szpsi, sblock)

      write(ilog,*)
      if (.not. lwarn) then
         write(ilog,'(a)') "Program terminated normally"
      else
         write(ilog,'(a)') "Program terminated with warnmessage."
      end if

      stop

      contains

!#######################################################################

      subroutine receiver (mc, mr, mi, ml, ms, mf,&
                           mcgs2,mcdicht1,mcdicht2,mcdicht4)

      implicit none

      complex(dop), dimension(mcdim), intent(inout)  :: mc
      real(dop), dimension(mrdim), intent(inout)     :: mr
      integer(long), dimension(midim), intent(inout) :: mi
      logical(kind=4), dimension(mldim), intent(in)  :: ml
      complex(sip), dimension(msdim), intent(inout)  :: ms
      real(sip), dimension(mfdim), intent(in)        :: mf
      complex(dop), dimension(gs2matdim), intent(out)  :: mcgs2
      complex(dop), dimension(dmatdim), intent(out)  :: mcdicht1
      complex(dop), dimension(dmatdim), intent(out)  :: mcdicht2
      complex(dop), dimension(dmatdim), intent(out)  :: mcdicht4

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! provide arguments of pointers to arrays needed for analysis
!  e.g. mc(mcpsi) for psi array
!-----------------------------------------------------------------------
      call sumwf(mc(mcpsi),mc(mcpsi1),mc(mcspsi),mc(mcdover), &
           mcdicht1,mcdicht2,mr(mrdicht3),mcdicht4, &
           mc(mcworkc),mi(miworki),mi(mijindx),mi(mitrajst), &
           mr(mrddpes),mc(mcagmat),mcgs2)

      return
      end subroutine receiver 
      
!***********************************************************************
!
!     HERE IS THE SUBROUTINE TO DO THE ANALYSIS
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! change name of subroutine (or else mutil gets confused)
! provide dummy arguments of arrays needed for analysis
!  e.g. psi to match argument mc(mcpsi)
!
!***********************************************************************

      subroutine sumwf(psi2,psi1,spsi,dover,dicht1,dicht2,dicht3,dicht4, &
           workc,worki,jindx,trajst,ddpes,agmat,gs2)

use iopsidef, only: psidat
use iogrddef, only: griddat
use iorst, only: rdrst,wrrst
use compdat, only: savecsys, savecpsi, getcsys, getcpsi
use densitymod, only: density
use trafonatmod, only: trafonat
use getpsimod, only: akcopy

      implicit none
      
      integer(long) :: m,s,ierr, irst0,i,zeig1,zeig2,swapzeig, &
                       chkdvr,chkgrd,chkpsi,chkprp,idum,iflag,zeig, &
                       mblock,dim2
      integer(long), dimension(workidim), intent(out)  :: worki
      real(dop), dimension(smaxd3matdim), intent(out)  :: dicht3
      real(dop), dimension(1)                          :: rdum
      real(dop)                                        :: t1,norm1
      complex(dop), dimension(maxdgldim), intent(out)  :: psi2,psi1
      complex(dop), dimension(smaxdgldim), intent(out) :: spsi
      complex(dop), dimension(smaxdmatdim), intent(out) :: dicht1,dicht2,dicht4
      complex(dop), dimension(workcdim), intent(out)   :: workc
      complex(dop), dimension(smaxdovdim), intent(out) :: dover
      complex(dop), dimension(1)                       :: cdum
      logical(kind=4)                                  :: ldum
      integer(long), dimension(adim), intent(out)      :: jindx
      integer(long), dimension(gwpinfodim), intent(out):: trajst
      real(dop), dimension(*), intent(out)             :: ddpes
      complex(dop), dimension(agmatdim), intent(out)   :: agmat
      complex(dop), dimension(gs2matdim), intent(out)  :: gs2

      routine='sumwf'
    
!-----------------------------------------------------------------------
! read the first restart and store the data
!-----------------------------------------------------------------------
      irst = 591
      chkdvr=2
      chkgrd=2
      chkpsi=1
      chkprp=0
      idum=1
      iflag=0
      ldum=.true.

      call rdrst(t1,psi2,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
                 cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
      call psidat
 
!-----------------------------------------------------------------------
! special treatment for exact WF
!-----------------------------------------------------------------------
      if (lexact) then
         do i=1,dgldim
            psi2(i)= psi2(i)*scoeff(1)
         end do

         do irst = 592, irdwf

            chkdvr=2
            chkgrd=2
            chkpsi=1
            chkprp=0
            idum=1
            iflag=0
            
            call rdrst(t1,psi1,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
                       cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
            close(irst)
            
            do i=1,dgldim
               psi2(i)= psi2(i)+psi1(i)*scoeff(irst-590)
            end do
         end do

         if (lsnormalize) then
            do i=1,adim
               psi2(i)=0.0d0
            enddo
            call normvxz(psi2,norm1,dgldim)
            write(ilog,*)
            write(ilog,'(44a1)')  ('-',i=1,44)
            write(ilog,'(a,E16.8)') 'Norm before normalization: ', norm1
            norm1=norm1/snorm
            call xvixdzo (norm1,psi2,dgldim)
            call normvxz(psi2,norm1,dgldim)
            write(ilog,'(a,E16.8)') 'Norm after normalization:  ', norm1
            write(ilog,'(44a1)')  ('-',i=1,44)
         end if
         do i=1,adim  ! Set dummy A-vector
            psi2(i)=1.0d0
         enddo

         !restore data and exit
         ldum=.true.
         irst=22
         t1=0.d0
         call wrrst(t1,psi2,ldum,cdum,idum,rdum,rdum,jindx,agmat,ddpes,trajst)
         
         return
      end if

!-----------------------------------------------------------------------
! MCTDH WF
!-----------------------------------------------------------------------

!     re-orthogonalize the SPF 
      call sorthospf(psi2,dover,workc)
      call savecsys(2)
      call savecpsi(2)
      
!     Multiply the coefficient
      call xvxxzzo(scoeff(1),psi2,adim)
      
!-----------------------------------------------------------------------
! Do-Loop over all remaining restart files defined in read-inwf block
!-----------------------------------------------------------------------
      do irst = 592, irdwf

         chkdvr=2
         chkgrd=2
         chkpsi=1
         chkprp=0
         idum=1
         iflag=0
         
         call rdrst(t1,psi1,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
                    cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
         call psidat
         call sorthospf(psi1,dover,workc)
         call xvxxzzo(scoeff(irst-590),psi1,adim)
         
  

!     calculate the size of spsi (the temporary larger psi)
         do s=1,nstate
            do m=1,nmode
               spdim(m,s)=dim(m,s)+cdim(m,s,2)
            end do
         end do
!     set up some pointers of the large system 
         call zeigswf
         call zerovxz(spsi,smaxdgldim)

!     !onstruct a new SPF-basis out of the the SPF's of the
!     Two initial WF's (no check on idmode!!!)

         do s=1,nstate
            do m=1,nmode
               call zerovxz(workc,workcdim)

               call cpvxz(psi2(czetf(m,s,2)),workc,cdim(m,s,2)*subdim(m))
               
               zeig = cdim(m,s,2)*subdim(m)+1

               call cpvxz(psi1(zetf(m,s)),workc(zeig),dim(m,s)*subdim(m))

!     Here we leave out some SPFs if they are linear dependent
!     of the others. sortho resets spdim(m,s) to the new number of
!     SPFs. Hence, later we call concentratespf to remove holes on the 
!     memory block of spsi and to reset the pointers and array sizes.
               i=spdim(m,s)
               call sortho(workc,m,s,ierr)
 
               if (ierr.ne.0) then
                  routine='sumwf (sumrst)'
                  message="There hase been a problem with"// &
                    " orthogonalizing SPFs"
                  call errormsg
               end if
               call cpvxz(workc,spsi(szetf(m,s)),spdim(m,s)*subdim(m))
            end do
         end do

!     Reset pointers because some SPFs might have been skipped.
!     Also move unused memory blocks in spsi to the end
         call concentratespf(spsi,workc)
         
!     Express both functions in the new basis and add them: 
!     trafo of psi2
         do s=1,nstate
            do m=1,nmode
               call mmaxzz(spsi(szetf(m,s)),psi2(czetf(m,s,2)),dover(dovmat(m,s)), &
                           subdim(m),spdim(m,s),cdim(m,s,2))
            end do
         end do
         call zerovxz(workc,workcdim)
         do s=1,nstate
            zeig2=szpsi(s)
            zeig1=zeig2 + max(sblock(s),cblock(s,2))
            call cpvxz(psi2(czpsi(s,2)),workc(zeig2),cblock(s,2))
            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call mtxxzz(dover(dovmat(m,s)),workc(zeig1),workc(zeig2), &
                    svdim(m,s),cdim(m,s,2),cndim(m,s,2),spdim(m,s))
            enddo
            if(zeig1.lt.zeig2) call cpvxz(workc(zeig2),workc(zeig1),sblock(s))
         end do
         call cpvxz(workc,spsi,sadim)


!     trafo of psi1
         call zerovxz(workc,workcdim)
         call zerovxz(dover,smaxdovdim)
         do s=1,nstate
            do m=1,nmode
               call mmaxzz(spsi(szetf(m,s)),psi1(zetf(m,s)), &
                    dover(dovmat(m,s)),subdim(m),spdim(m,s),dim(m,s))
            end do
         end do
         do s=1,nstate
            zeig2=szpsi(s)
            zeig1=zeig2 + max(sblock(s), block(s))
            call cpvxz(psi1(zpsi(s)),workc(zeig2),block(s))
            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call mtxxzz(dover(dovmat(m,s)),workc(zeig1), &
                    workc(zeig2),svdim(m,s),dim(m,s),ndim(m,s),spdim(m,s))
            enddo
            if(zeig1.lt.zeig2) call cpvxz(workc(zeig2),workc(zeig1),sblock(s))
         end do



!     add the wavefunctions
         do i=1,sadim
            spsi(i)=workc(i)+spsi(i)
         end do
      
!     Make spsi pointers the system pointers
         call cpmxi(spdim,dim,maxdim,maxsta)
         call zeromxi(zetf,maxdim,maxsta)
         call psidat
         maxblock=smaxblock
  
!     Transform to natorbs
         call density(spsi,dicht1,dicht2,dicht3,dicht4,jindx,gs2)
         call trafonat(spsi,dicht4,jindx)

!     The natorbs are stored such, that the SPFs with
!     heighest population are in the front. Now, if there
!     are waves left to add to the superposition, reduce
!     spsi so that it matches specifications in compdat(1).
!     Otherwise store system in compdat(1).
!     If dim(m,s) happens to be smaller then cdim(m,s,1) 
!     (unlikely) then use dim instead. Result is stored in psi2          
         if (irst .eq. irdwf) goto 21

         do s=1,nstate
            do m=1,nmode
               if (dim(m,s) .gt. cdim(m,s,1)) then
                  dim(m,s)=cdim(m,s,1)
               end if
            end do
         end do
         call psidat
         call savecpsi(2)
         call savecsys(2)
         call cpmxi(spdim,dim,maxdim,maxsta)
         call psidat
!     Now compdat(2) contains the basis to trafo to  
!     system the basis to trafo from

         call zerovxz(psi2,maxdgldim)       
         do s=1,nstate

            mblock=1
            do m=1,nmode
               dim2=max(dim(m,s),cdim(m,s,2))
               mblock=mblock*dim2
            enddo
            
            zeig=1
            zeig1=zeig+mblock

            call cpvxz(spsi(zpsi(s)),workc(zeig),block(s))
          
            do m=1,nmode
               call akcopy(workc(zeig),workc(zeig1),cvdim(m,s,2), &
                           cdim(m,s,2),dim(m,s),ndim(m,s))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            end do
            call cpvxz(workc(zeig),psi2(czpsi(s,2)),cblock(s,2))
         end do   
         
         do s=1,nstate
            do m=1,nmode
                  call cpvxz(spsi(zetf(m,s)),psi2(czetf(m,s,2)), &
                       cdim(m,s,2)*subdim(m))
            end do
         end do

      end do !irst

!     Restore system we transformed to the last time and store 
!     in compdat(1). This is to be used in chkspos()
      call getcpsi(2)
      call getcsys(2)
      call cpvxz(psi2,spsi,dgldim)
 21   continue
      call savecpsi(1)
      call savecsys(1)
      
!     check quality of the superposition and write log
      call chkspos(spsi,psi1,psi2,dover,workc,jindx,trajst,ddpes,agmat)

      call getcpsi(1)
      call getcsys(1)

!     Normalize 

      if (lsnormalize) then
         call normvxz(spsi,norm1,adim)
         write(ilog,*)
         write(ilog,'(44a1)')  ('-',i=1,44)
         write(ilog,'(a,E16.8)') 'Norm before normalization:', norm1
         norm1=norm1/snorm
         call xvixdzo (norm1,spsi,adim)
         call normvxz(spsi,norm1,adim)
         write(ilog,'(a,E16.8)') 'Norm after normalization: ', norm1
         write(ilog,'(44a1)')  ('-',i=1,44)
      end if

!     write the restart of the superposition
      ldum=.true.
      irst=22
      t1=0.d0
      call wrrst(t1,spsi,ldum,cdum,idum,rdum,rdum,jindx,agmat,ddpes,trajst)

      write(ilog,*)
      write(ilog,'(a)')"***** PROPERTIES OF THE NEW RESTART WF *****"
      write(ilog,*)
      write(ilog,'(a,i8)')" dgldim =   ", dgldim
      write(ilog,'(a,i8)')" adim =     ", adim
      write(ilog,*)
      write(ilog,'(a,i8)')" Number of SPF:"
      write(ilog,*)
      write(ilog,'(a,i8)')"    State(s)   Mode(m)   dim(m,s)"
      write(ilog,'(44a1)')  ('-',i=1,44)
      do s=1,nstate
         do m=1,nmode
            write(ilog,'(3(i8,3x))') s,m,dim(m,s)
         end do
      end do
      write(ilog,*)
      write(ilog,'(a)')"********************************************"
 

!     we are done
      return

 998  write(6,*) ' ROUTINE: ', routine
      write(6,*) ' ERROR:   reading restart file '
      stop 1
 
      end subroutine sumwf

!#######################################################################
!
!               CHKSPOS
!
!     !heck the quality of the superposition by calculating the 
!     overlaps of the single wave functions with the superposition WF
!     and determining the difference to the result obtained using the
!     input functions
!
!     psi:  contains the superposition 
!           (info saved in compdat(1) on enter)
!     psi1: from file 
!     psi2: from file
!
!#######################################################################
 
      subroutine chkspos(psi,psi1,psi2,dover,workc,jindx,trajst,ddpes,agmat)

use iopsidef, only: psidat
use iorst, only: rdrst
use compdat, only: savecpsi

use compdat, only: savecsys

      implicit none
      
      complex(dop), dimension(dgldim), intent(out) :: psi
      complex(dop), dimension(maxdgldim), intent(out) :: psi1,psi2
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), dimension(smaxdovdim), intent(out) :: dover
      complex(dop), dimension(8,8)              :: ovlp
      complex(dop), dimension(8)              :: cross, cross1
      complex(dop), dimension(1)              :: cdum
      integer(long)                           :: chkdvr,chkgrd,chkpsi,chkprp,&
                                                 idum,iflag,swapzeig, zeig1,zeig2
      integer(long), dimension(adim), intent(out)       :: jindx
      integer(long), dimension(gwpinfodim), intent(out) :: trajst
      real(dop), dimension(*), intent(out)              :: ddpes
      complex(dop), dimension(agmatdim), intent(out)    :: agmat

      real(dop)               :: t1
      real(dop), dimension(1) :: rdum
      logical(kind=4)         :: ldum
      integer(long)           :: m,s,i,j, irst1, irst0, nfile

      routine="chkspos"
      ldum=.true.
      irst=22

      call zeromxz(ovlp,8,8) 
!----------------------------------------------------------------------
! !alculate overlap matrix of input wavefunctions
! ovlp(i,j)=<psi_i|psi_j>, i is bra, j is ket
!--
      do irst0=591, irdwf
         i=irst0-590

         chkdvr=2
         chkgrd=2
         chkpsi=1
         chkprp=0
         idum=1
         iflag=0
         irst=irst0
         ldum=.true.
         call rdrst(t1,psi1,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
              cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
 
         call psidat
         call sorthospf(psi1,dover,workc)
         call savecpsi(2)
         call savecsys(2)

         do irst1= irst0, irdwf
            j=irst1-590

            if (irst.eq.irst1) then
               call vvaxzz(psi1,psi1,ovlp(i,j),adim)
            else

               chkdvr=2
               chkgrd=2
               chkpsi=1
               chkprp=0
               idum=1
               iflag=0
               
               irst=irst1
               call rdrst(t1,psi2,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
                          cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
  
               call psidat
               call sorthospf(psi2,dover,workc)


! express psi2 in the SPFs of Psi1, psi1 is in compdat2, psi2 in system
               do s=1,nstate
                  do m=1,nmode
                     call mmaxzz(psi1(czetf(m,s,2)),psi2(zetf(m,s)), &
                          dover(dovmat(m,s)), &
                          subdim(m),cdim(m,s,2),dim(m,s))
                  end do
               end do
               
               call zerovxz(workc,workcdim)
               do s=1,nstate
                  zeig2=czpsi(s,2)
                  zeig1=zeig2 + max(cblock(s,2),block(s))
                  call cpvxz(psi2(zpsi(s)),workc(zeig2),block(s))
                  do m=1,nmode
                     swapzeig=zeig1
                     zeig1=zeig2
                     zeig2=swapzeig
                     call mtxxzz(dover(dovmat(m,s)),workc(zeig1), &
                          workc(zeig2),cvdim(m,s,2),dim(m,s),ndim(m,s), &
                          cdim(m,s,2))
                  enddo
         
                  call vvaxzz(psi1(czpsi(s,2)),workc(zeig2),cdum(1),cblock(s,2))
                  ovlp(i,j) = ovlp(i,j)+cdum(1)
               end do

               ovlp(j,i)=dconjg(ovlp(i,j))
            end if
            
         end do
         
      end do
    

!----------------------------------------------------------------------
! Calculate overlap vector of input and superposition wavefunctions
! cross(j)=<psi_superpos,psi_j> = sum_i conjg(C_i) ovlp(i,j)
!----------------------------------------------------------------------

      do irst= 591, irdwf
         
         chkdvr=2
         chkgrd=2
         chkpsi=1
         chkprp=0
         idum=1
         iflag=0
         ldum=.true.

         call rdrst(t1,psi2,ldum,chkdvr,chkgrd,chkpsi,chkprp, &
                    cdum,idum,rdum,iflag,jindx,agmat,ddpes,trajst)
         call psidat
         call sorthospf(psi2,dover,workc)

         do s=1,nstate
            do m=1,nmode
               call mmaxzz(psi(czetf(m,s,1)),psi2(zetf(m,s)), &
                    dover(dovmat(m,s)),subdim(m),cdim(m,s,1),dim(m,s))
            enddo
         end do
               
!     transform A-vec of psi2 to the SPFs of psi 
         call zerovxz(workc,workcdim)
         cross(irst-590) =(0.d0,0.d0)
         do s=1,nstate
            zeig2=czpsi(s,1)
            zeig1=zeig2 + max(cblock(s,1),block(s))
            call cpvxz(psi2(zpsi(s)),workc(zeig2),block(s))
            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call mtxxzz(dover(dovmat(m,s)),workc(zeig1), &
                    workc(zeig2),cvdim(m,s,1),dim(m,s),ndim(m,s),cdim(m,s,1))
            enddo
            
            call vvaxzz(psi(czpsi(s,1)),workc(zeig2),cdum(1),cblock(s,1))
            cross(irst-590) = cross(irst-590) +cdum(1)
         end do
                  
      end do

! Compare and log
      write(ilog,*)
      write(ilog,'(a)')"******* QUALITY OF THE SUPERPOSITION *******"
      write(ilog,*)
      write(ilog,'(a)') "Overlaps O_ij=<psi_i|psi_j> of initial WFs:"
      write(ilog,*)
      write(ilog,'(a)') "   i   j     Re[O_ij]        Im[O_ij]"
      write(ilog,'(a)')"--------------------------------------------"
      nfile=irdwf-590
      do i=1,nfile
         do j=1,nfile
            write(ilog,'(2i4,2E16.8)')j,i,dble(ovlp(j,i)),dimag(ovlp(i,j))
         end do
      end do   
      write(ilog,*)
      write(ilog,'(a)') "Overlaps S_i=<psi_superpos|psi_i>:"
      write(ilog,*)
      write(ilog,'(a)') "   i        Re[S_i]         Im[S_i]"
      write(ilog,'(a)')"--------------------------------------------"
      do i=1,nfile
         write(ilog,'(i4,4x,2E16.8)') i, dble(cross(i)), dimag(cross(i))
      end do
      write(ilog,*)
      write(ilog,'(a)') "Same using the coefficients C_i and over-"
      write(ilog,'(a)') "laps O_ij as S1_i = sum_j conjg(C_j)*O_ji"
      write(ilog,*)
      write(ilog,'(a)') "   i        Re[S1_i]        Im[S1_i]"
      write(ilog,'(a)')"--------------------------------------------"
      do i=1,nfile
         call vvaxzz(scoeff,ovlp(1,i),cross1(i),nfile)
         write(ilog,'(i4,4x,2E16.8)') i, dble(cross1(i)),dimag(cross1(i))
      end do
      write(ilog,*)
      write(ilog,'(a)') "Difference DS_i = S_i - S1_i:"
      write(ilog,*)
      write(ilog,'(2a)') "   i        Re[DS_i]        Im[DS_i]", &
           "         |DS_i|^2"
      write(ilog,'(a)') &
           "--------------------------------------------------------"
      do i=1,nfile
         cross(i)=cross(i)-cross1(i)
         write(ilog,'(i4,4x,3E16.8)') i,dble(cross(i)),dimag(cross(i)), &
              abs(cross(i))**2
      end do
      
      rdum(1)=0.d0
      do i=1,nfile
         rdum(1)=rdum(1)+abs(cross(i))**2
      end do
      rdum(1)=dsqrt(rdum(1))
      write(ilog,*)
      write(ilog,'(a)') "Total difference:"
      write(ilog,*) 
      write(ilog,'(a,E16.8)')"  sqrt(sum_i |DS_i|^2) =", rdum(1)
      write(ilog,*) 
      
      if (rdum(1).gt. 1.d-3 ) then
         write(ilog,'(a)') "+++ WARNING +++ WARNING +++ WARNING +++"// &
                " WARNING +++ WARNING +++ WARNING +++"
         write(ilog,*) 
         write(ilog,'(2a)') " The following error is raised if", &
           " sqrt(sum_i |DS_i|^2) > 1.E-3 :"

         routine="chkspos"
         message="The quality of the superposition seems to be low."// &
              " See logfile for dedails."
         call warnmsg
         write(ilog,*) 
         write(ilog,'(a)') " This is probably due to a limited "//"number of SPFs."
         write(ilog,'(a)') " The number of SPFs for a mode"// &
                " in the output wavefunction is"
         write(ilog,'(a)') " determined by the maximum number of "// &
                "SPFs for that mode in all input"
         write(ilog,'(a)') " wavefunctions. To improve the accuracy"// &
                 " try using one input"
         write(ilog,'(a)')" wavefunction with more SPFs."
         write(ilog,*) 
      end if

      write(ilog,'(a)')"********************************************"

      return
      end subroutine chkspos

!#######################################################################
!
!               SRDRSTINFO
!
!     Read the restart files and calculate size of output WF 
!
!#######################################################################

      subroutine srdrstinfo

      implicit none

      integer(long) :: chkdvr,chkgrd,chkpsi,chkprp,s,m
      
      logical(kind=4) :: linwf,lerr

      linwf=.true.
      lexact=.false.
      maxdgldim=0
      
      call zeromxi(spdim,maxdim,maxsta)

      do irst = 591, irdwf
         if(irst.eq.591)then
            chkdvr=1
            chkgrd=1
            chkpsi=1
            chkprp=0
         else
            chkdvr=2
            chkgrd=2
            chkpsi=1
            chkprp=0
         end if

         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
         call psidat
         write(ilog,'(a,3i11)')&
            ' adim, totphidim, dgldim:',adim,totphidim,dgldim
         if(psitype.eq.1) then
            lexact=.true.
            write(ilog,'(a)') ' Wavefunction is of EXACT type!'
         elseif(psitype.gt.1) then
            write(ilog,'(a,i2)') ' ###  psitype =',psitype
            routine='sumwf (sumrst)'
            message='sumrst can handle neiter selected-CI nor '//&
                    'ML-MCTDH wavefunctions.              '//&
                    'Only the types "MCTDH" or "exact" are allowed.'
            call errormsg
         endif

!     calculate the maximum size of dim in all WFs
!     (this is dim of the output WF)
         if (irst.eq.591) then
           
            maxdgldim = dgldim
            do s=1,nstate
               do m=1,nmode
                  spdim(m,s) = dim(m,s)
               end do
            end do

         else
            do s=1,nstate
               do m=1,nmode
                  spdim(m,s) = max(spdim(m,s),dim(m,s))
               end do
            end do
            maxdgldim = max(dgldim,maxdgldim)
         end if
      enddo                     ! irst=591,irdwf
      

!     this is the config the output WF will have
!     (if the input WFs have no linear dependent SPFs)
      do s=1,nstate
         do m=1,nmode
            dim(m,s) = spdim(m,s)
         end do
      end do   
      call psidat
      call savecpsi(1)
      call savecpsi(2)
      call savecsys(1)
      call savecsys(2)

!     this is the maximum config the temporary WF will have
!     (might be a bit more if WFs have different dim(m,s))
      do s=1,nstate
         do m=1,nmode
            spdim(m,s) = spdim(m,s)*2
         end do
      end do     

      call zeigswf
      smaxdgldim   = sdgldim
      smaxdovdim   = doverdim
      smaxdmatdim  = sdmatdim
      smaxd3matdim = sd3matdim
      smaxadim     = sadim
      
      return
      end subroutine srdrstinfo


!#######################################################################
!
!              ZEIGSWF
!
!     Sets up pointers and matrix sizes for the temporary WF   
!
!#######################################################################

      subroutine zeigswf

      implicit none

      integer(long) :: m,s,zeig

!-----------------------------------------------------------------------
! maxdim
!-----------------------------------------------------------------------
      smaxspf=0
      do s=1,nstate
         do m=1,nmode
            smaxspf=max(smaxspf,spdim(m,s))
         end do
      end do
      

!-----------------------------------------------------------------------
! vdim, ndim, block
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig = 1
         do m=1,nmode
            svdim(m,s)=zeig
            zeig=zeig*spdim(m,s)
         end do
         do m=1,nmode
            sndim(m,s)=zeig/(spdim(m,s)*vdim(m,s))
         enddo
         sblock(s)=zeig
      end do


!-----------------------------------------------------------------------
! maxblock, zpsi, adim
!-----------------------------------------------------------------------
      zeig=1
      smaxblock=0
      do s=1,nstate
         smaxblock = max(smaxblock,sblock(s))
         szpsi(s)=zeig
         zeig=zeig+sblock(s)
      end do
      sadim = zeig-1


!-----------------------------------------------------------------------
! zetf
!-----------------------------------------------------------------------
      call zeromxi(szetf,maxdim,maxsta)
      zeig=sadim+1
      do m=1,nmode
         do s=1,nstate
            szetf(m,s)=zeig
            zeig=zeig+spdim(m,s)*subdim(m)
         end do
      end do
      sdgldim=zeig-1
   
!-----------------------------------------------------------------------
! dmat
!-----------------------------------------------------------------------
      zeig = 1
      do s=1,nstate
         do m=1,nmode
           dmat(m,s)=zeig
           zeig = zeig+spdim(m,s)**2
         end do
      end do
      sdmatdim=zeig-1


!-----------------------------------------------------------------------
! dicht3
!-----------------------------------------------------------------------
      zeig=1
      do s=1,nstate
         do m=1,nmode
            d3mat(m,s)=zeig
            zeig=zeig+spdim(m,s)*2
         enddo
      enddo
      sd3matdim =zeig-1
      
!-----------------------------------------------------------------------
! dovmat
!-----------------------------------------------------------------------
      zeig=1
      do s=1,nstate
         do m=1,nmode
            dovmat(m,s)=zeig
            zeig=zeig+spdim(m,s)**2
         enddo
      enddo
      doverdim = zeig-1

      return
      end subroutine zeigswf

!#######################################################################
!
!              CONCENTRATEPSI
!
!     concentrate spsi (spsi has 'holes' in the data structure so far)
!
!#######################################################################

      subroutine  concentratespf(spsi,workc)

      implicit none

      complex(dop), dimension(smaxdgldim), intent(inout) :: spsi
      complex(dop), dimension(workcdim), intent(out)     ::  workc

      integer(long), dimension(maxdim,maxsta) :: szetf1
      integer(long)                           :: zeig,s,m,idum

      call cpvxz(spsi,workc,smaxdgldim)
      call zerovxz(spsi,smaxdgldim)
        
!-----------------------------------------------------------------------
! adim, dgldim
!-----------------------------------------------------------------------
      sdgldim=0
      sadim  =0
      do s=1,nstate
         idum=1
         do m=1,nmode
            sdgldim = sdgldim+spdim(m,s)*subdim(m)
            idum=idum*spdim(m,s)
         end do
         sadim=sadim+idum
      end do
      sdgldim=sdgldim+sadim
     
!-----------------------------------------------------------------------
! zetf
!-----------------------------------------------------------------------
      zeig=sadim+1
      do m=1,nmode
         do s=1,nstate
            szetf1(m,s)=zeig
            zeig=zeig+spdim(m,s)*subdim(m)
         end do
      end do
      
!-----------------------------------------------------------------------
! copy the blocks (remove 'holes' in memory)
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call cpvxz(workc(szetf(m,s)),spsi(szetf1(m,s)),spdim(m,s)*subdim(m))
         end do
      end do
      call cpmxi(szetf1,szetf,maxdim,maxsta)
      call zerovxz(workc,workcdim)

      call zeigswf
      
      return
      end subroutine concentratespf

!#######################################################################
!
!              SORTHOSPF
!
!     Re-orthogonalize SPFs of a wave function and transform A-vector. 
!
!#######################################################################

      subroutine sorthospf(psi,dover,workc)

      implicit none

      complex(dop), dimension(dgldim), intent(inout)   :: psi
      complex(dop), dimension(smaxdovdim), intent(out) :: dover
      complex(dop), dimension(workcdim), intent(out)   :: workc
      integer(long) :: m,s,zeig1,zeig2,swapzeig,ierr
      real(dop) :: norm,norm1

      call cpvxz(psi,workc,dgldim)
      call normvxz(psi,norm,adim)

      do s=1,nstate
         do m=1,nmode
            call schmidtortho(workc(zetf(m,s)),subdim(m),dim(m,s),ierr)
            if (ierr.ne.0) goto 110
         end do
      end do

      do s=1,nstate
         do m=1,nmode
            call mmaxzz(workc(zetf(m,s)),psi(zetf(m,s)),dover(dovmat(m,s)), &
                 subdim(m),dim(m,s),dim(m,s))
         end do
      end do

      do s=1,nstate
         do m=1,nmode
            call cpvxz(workc(zetf(m,s)),psi(zetf(m,s)),subdim(m)*dim(m,s))
         end do
      end do

      call zerovxz(workc,workcdim)
      do s=1,nstate
         zeig2=zpsi(s)
         zeig1=zeig2 + block(s)
         call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))
         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(dover(dovmat(m,s)),workc(zeig1),workc(zeig2), &
                 vdim(m,s),dim(m,s),ndim(m,s),dim(m,s))
         enddo
         
         call cpvxz(workc(zeig2), psi(zpsi(s)),block(s))
      end do

      call normvxz(psi,norm1,adim)
      
      if (abs( norm-norm1).gt. 1.d-4 ) then
         routine="sorthospf (sumrst)"
         message="Norm before and after re-orthogonalization of SPFs"// &
                "in an input restart file differs more than 1.E-4."
         call warnmsg
         lwarn=.false.
      end if
      norm1=norm1/norm
      call xvixdzo(norm1,psi,adim)

      return

 110  continue
      routine="sorthospf (sumrst)"
      message="SPFs of WF in one of the input files could not be"// &
           " re-orthogonalized."
      call errormsg

      return
      end subroutine sorthospf


!#######################################################################
!
!              SORTHO
!
!     Normalize combined SPFs (neglecting linear dependent SPFs)
!     This is a slightly modified version of schmidtortho
!
!#######################################################################

      subroutine sortho(psi,m,s,ierr)

      implicit none

      integer(long) :: npoints,nphi,e,e1,ierr,m,s
      complex(dop), dimension(subdim(m),spdim(m,s)), intent(inout) :: psi
      complex(dop)                                                 :: overlap
      real(dop) :: norm

!-----------------------------------------------------------------------
! Schmidt-orthogonalize the set of functions.
! The orthonormalization is made twice to remove numerical inaccuracies.
!-----------------------------------------------------------------------

      npoints=subdim(m)
      nphi=spdim(m,s)
      
      ierr = 0
      do e=1,nphi

 20      continue
         if (spdim(m,s).lt.e) goto 30
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,npoints)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),npoints)
         enddo
         call normvxz(psi(1,e),norm,npoints)
         
         if (norm .le. tiny ) then !skip this SPF

            if (e .lt. nphi) then
               do e1=e,nphi
                  call cpvxz(psi(1,e+1),psi(1,e),npoints )
               end do
               
            end if

            call zerovxz(psi(1,nphi),npoints)
            spdim(m,s)=spdim(m,s)-1
            goto 20
         else
            call xvixdzo(norm,psi(1,e),npoints)
         end if
      enddo

 30   continue

      nphi=spdim(m,s)

      do e=1,nphi
         do e1=1,e-1
            call vvaxzz(psi(1,e1),psi(1,e),overlap,npoints)
            call xvxxzzr(overlap,psi(1,e1),psi(1,e),npoints)
         enddo
         call normvxz(psi(1,e),norm,npoints)
         if( norm .le. 0.8d0 ) then
            write(*,*)'Abort due to orthoerr'
            ierr = e 
            return
         end if
         call xvixdzo(norm,psi(1,e),npoints)
      enddo
      
      return
      end subroutine sortho


!***********************************************************************
!
!          SMEMAUSW
!
!     This is a !opy of memaus except for loading compdat(1)
!     as that would overwrite dmatdim.
!***********************************************************************

      subroutine smemausw

      implicit none

!-----------------------------------------------------------------------
! with the parameter "pdum" the pointers to arrays that are not
! used are initialized with the value one in order to avoid an abortion
! of the program when array bounds are checked (compiler option "-c").
!-----------------------------------------------------------------------
      integer   pdum
      parameter (pdum = 1)

!-----------------------------------------------------------------------
! assign standard arrays to memory arrays
! mcdim, mrdim, midim and mldim are final addresses assigned in mc, 
!  mr, mi and ml. 
!-----------------------------------------------------------------------
! set pointers to complex arrays in memory array mc 
!-----------------------------------------------------------------------
      mcdim=1
      if (lpsi .and. .not. lpsigrd) then
         mcpsi=mcdim
         mcdim=mcdim+dgldim
      else if (lpsi .and. lpsigrd) then
         mcpsi=mcdim
         mcdim=mcdim+griddim
      else
         mcpsi=pdum
      endif
      if (lpsi1 .and. .not. lpsigrd) then
         mcpsi1=mcdim
         mcdim=mcdim+cdgldim(2)
      else if (lpsi1 .and. lpsigrd) then
         mcpsi1=mcdim
         mcdim=mcdim+max(cdgldim(2),griddim)
      else
         mcpsi1=pdum
      endif

      if (ldmat) then
!         mcdicht1=mcdim
!         mcdim=mcdicht1+dmatdim
!         mcdicht2=mcdim
!         mcdim=mcdicht2+dmatdim
!         mcdicht4=mcdim
!         mcdim=mcdicht4+dmatdim
      else
!         mcdicht1=pdum
!         mcdicht2=pdum
!         mcdicht4=pdum
      endif
      if (ldmat1) then
!         mcdicht11=mcdim
!         mcdim=mcdicht11+cdmatdim(2)
!         mcdicht21=mcdim
!         mcdim=mcdicht21+cdmatdim(2)
!         mcdicht41=mcdim
!         mcdim=mcdicht41+cdmatdim(2)
      else
!         mcdicht11=pdum
!         mcdicht21=pdum
!         mcdicht41=pdum
      endif
      if (lfft) then
         mcrueck=mcdim
         mcdim=mcrueck+fftdim
         mcexphin=mcdim
         mcdim=mcexphin+expdim
      else
         mcrueck=mcdim
         mcdim=mcrueck+2
         mcexphin=mcdim
         mcdim=mcexphin+2
      endif

!-----------------------------------------------------------------------
! set pointers to real arrays in memory array mr 
!-----------------------------------------------------------------------
      mrdim=1
      if (lort) then
         mrort=mrdim
         mrdim=mrort+ortdim
      else
         mrort=pdum
      endif
      if (lgpop) then
         mrgpop1=mrdim
         mrdim=mrgpop1+ortdim*nstate
         mrgpop2=mrdim
         mrdim=mrgpop2+ortdim*nstate
         mrweight=mrdim
         mrdim=mrweight+maxgdim
      else
         mrgpop1=pdum
         mrgpop2=pdum
         mrweight=pdum
      endif
      if (ltrafo) then
         mrtrafo=mrdim
         mrdim=mrtrafo+dvrdim
      else
         mrtrafo=pdum
      endif
      if (lsphfbr) then 
         mrkinsph=mrdim
         mrdim=mrkinsph+3*sphdim
      else
         mrkinsph=pdum
      endif

!-----------------------------------------------------------------------
! set pointers to integer arrays in memory array mi 
!-----------------------------------------------------------------------
      midim   = 1
      if (lpsi .and. lpsiopt(2)) then
         mijindx = midim
         midim = mijindx+adim
      else
         mijindx = pdum
      endif
      if (lpsi1 .and. clpsiopt(2,2)) then
         mijindx1 = midim
         midim = mijindx1+cadim(2)
      else
         mijindx1 = pdum
      endif
      if (lsphfbr) then 
         mijsph=midim
         midim=mijsph+sphdim
         mimsph=midim
         midim=mimsph+sphdim
      else
         mijsph=pdum
         mimsph=pdum
      endif
      if (lfft) then
         mifftfak=midim
         midim=mifftfak+fftdim
      else
         mifftfak=midim
         midim=mifftfak+2
      endif

!----------------------------------------------------------------------
! set pointers to logical arrays in memory array ml 
!-----------------------------------------------------------------------
      mldim   = 1

!----------------------------------------------------------------------
! set pointers to single precision complex arrays in memory array ms 
!-----------------------------------------------------------------------
      msdim   = 1
      if (lpsi) then
         mspsi   = msdim
         msdim   = mspsi+spsidim
      else
         mspsi   = pdum
      endif

!----------------------------------------------------------------------
! set pointers to single precision real arrays in memory array mf 
!-----------------------------------------------------------------------
      mfdim   = 1
      if (lgpop) then
         mfgpop = mfdim
         mfdim = mfgpop+2*maxgdim
      else
         mfgpop = pdum
      endif

      return
      end subroutine smemausw

!--------------------------------------------------------------
!              SUMINPUT
!     Input routine for sumrst
!     Read .inp and check for existing data and owerwrite
!***********************************************************************

      subroutine suminput

use global
use maxv, only: maxdim,maxsta
use griddatmod
use channels
use chnpotmod
use joinpsimod
use fluxmod

      implicit none

      logical(kind=4) ::  lmnd, linperr,lexist
      integer(long)   ::  ierr,i,slen



! --- READ THE COMAND LINE; I.E. OPTIONS AND PATH OF INPUT FILE
      call sumcmd(lmnd)

      routine='suminput'

! --- READ RUN-SECTION AND INWF-SECTION
! --- FROM THE INPUT FILE

      if (inputfile(1:1) .eq. ' ') then
         message='No input file is given'
         call errormsg
      endif

      open(iin,file=inputfile,status='old',iostat=ierr)
      if (ierr.ne.0) then
          message='Could not open input file '//trim(inputfile)
          call errormsg
      endif

      call einrunsum(linperr)
      if (linperr) then
          message = 'Error while reading RUN-SECTION.'
          call errormsg
      endif


! --- CHECK VARIABLES ---

      if (name(1:1) .eq. ' ') then
         message='No name (output) directory is given.'
         call errormsg
      endif


! --- CREATE THE NAME DIRECTORY IF NECESSARY ---

      open(itmpin,file=name(1:laenge)//'/scratch',err=113)
      close(itmpin,status='delete')
      if(lmnd) write(6,'(2a)') 'Name-directory already exist: ',name(1:laenge)
      goto 114


 113  if(lmnd) then
         inptit='mkdir '//name(1:laenge)
         call excmd(inptit)
         write(6,'(2a)') 'Created name-directory: ',name(1:laenge)
      else
         write(6,'(a)') &
         '---------------------------------------------------------'
         write(6,'(a)') ' ################   ERROR   ##################'
         write(6,'(a)') ' There is no name-directory! The directory:'
         write(6,'(1x,a)') name(1:laenge)
         write(6,'(a)') ' does not exist! (or has no write permission)'
         write(6,'(2a)')' One may use the option -mnd to create the', &
                        ' name-directory.'
         write(6,'(a)') &
           '-------------------------------------------------------'
         stop
      endif


 114  continue

! --- OPEN AND WRITE SOME INFORMATION TO THE LOG FILE ---


      logfile=name(1:laenge)//'/'//'sumrst.log'
      llog = index(logfile,' ')-1
      open (unit=ilog,file=logfile(1:llog),form='formatted',iostat=ierr)
      if (ierr.ne.0) then
          message='Cannot open file '//logfile(1:llog)
          call errormsg
      endif
      logisopen=.true.
      write(ilog,'(a)') trim(cstartup(1:80))
      write(ilog,'(1x,a)') name(1:laenge)
      call wrversion(ilog)

      rewind iin
      write(ilog,'(/,20x,a,/80a1)') 'INPUT FILE (copy)',('*',i=1,80)
 200  read(iin,'(a)',err=210,end=220) inptit
      if (inptit(1:1) .ne. '#'     .and. &
           inptit(1:5) .ne. '-----' .and. &
           inptit(1:5) .ne. '_____' .and. &
           slen(inptit) .gt. 0 ) &
           write(ilog,'(a)',err=230) inptit(1:slen(inptit))
      call keylowc(inptit)
      if (inptit(1:9) .ne. 'end-input' ) goto 200
      write(ilog,'(80a1,/)') ('*',i=1,80)


      write(ilog,'(2a/)')   ' Command line options : ',options(1:optlaenge)
      write(ilog,'(2a/)') ' The new restart file is : ',name(1:laenge)//'/restart'

! --- CHECK FOR EXISTENCE OF FILES ---

      if (.not.lovwr) then
         inquire(file=name(1:laenge)//'/restart',exist=lexist)
         if (lexist) then
            write(6,'(/a)') ' One may use -w  option.'
            message='File already exists: '//name(1:laenge)//'/restart'
            call errormsg
         endif
      endif

      call eininwfsum(linperr)
      if (linperr) then
          message = 'Error while reading INWF-SECTION.'
          call errormsg
      endif
      close(iin)
      return

 210  message='Error while reading input file.'
      call errormsg
 220  message= 'An end-input is missing in input file.'
      call errormsg
 230  message= 'Error while writing input to log file.'
      call errormsg

      end subroutine suminput

!***********************************************************************
!              EINRUNSUM
!
!     Read run-section from .inp for generating a wave function
!***********************************************************************

      subroutine einrunsum(linperr)

use global
use maxv, only: maxdim,maxsta
use griddatmod
use channels
use chnpotmod
use joinpsimod
use fluxmod
use sumrstmod

      implicit none

      integer(long)   :: iz0,iz,ierr,ic,i
      logical(kind=4) :: linperr

      routine="einrunsum"
! --- read run-section
      linperr=.false.

      rewind(iin)
      iz0=0
      iz=0

 100  continue

      call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
          message = 'Cannot read input file'
          call errormsg
      endif

      if (keyword(1).eq.'end-input' .or. &
          keyword(1).eq.'end-run-section' ) then
         message='No RUN-SECTION in input file'
         call errormsg
      elseif (keyword(1).eq.'run-section') then
         goto 16
      else
         goto 100
      end if

 16   continue

      call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

      if (keyword(1).eq.'end-run-section') goto 21
      message = inptit
      call cntrl(keyword(1),lc(1),ierr)

      i=1
 15   continue
      if (keyword(i) .eq. 'name') then
         if (keyword(i+1) .ne. '=') then
            message = 'No argument specified with name keyword'
            call warnmsg
         else
            i=i+2
            if (laenge.eq.0) then
               name=keyorig(i)
               laenge = lc(i)
               if (laenge .eq. 0) then
                  message = ' No path name specified'
                  call wrline(zeroi,iz,iz0)
               endif
               call abspath(name,laenge)
            endif
         endif
      elseif (keyword(i).eq.'overwrite') then
         lovwr=.true.
      elseif (keyword(i).eq.'normalize') then

         lsnormalize=.true.
         i=i+1
         if (keyword(i) .eq. '=') then
            i=i+1
            if (keyword(i).eq.'no') then
               lsnormalize=.false.
            else
               read(keyword(i),*,iostat=ierr) snorm
               if (ierr.ne.0) then
                  message="Expecting real number or 'No' after " &
                       //'"normalize=", not: '//trim(keyword(i))
                  call errormsg
               end if
            end if
         else
            snorm=1.d0
         end if
      else
         message='Unknown keyword : '
         call wrline(i,iz,iz0)
      endif


!     --  check if more keywords are to be read, else read new line
      if (i.lt.ic) then
         i=i+1
         go to 15
      else
         go to 16
      endif



 21   continue
      linperr=linperr.or.lwarn

      return


      end subroutine einrunsum


!***********************************************************************
!              EININWFSUM
!
!     Read file info and coefficiens for generating a wave function
!     as a superposition of several restart wavefunctions from input
!***********************************************************************

      subroutine eininwfsum(linperr)

use global
use maxv, only: maxdim,maxsta
use griddatmod
use datenmod
use runinwfmod
use channels
use sumrstmod

      implicit none

      integer(long)                :: iz,iz0,nfiles,ierr, ic,i
      real(dop)                    :: recoeff,imcoeff
      logical(kind=4), intent(out) :: linperr
      logical(kind=4)              :: lrdcoeff

      irdwf=590
      nfiles=0
      routine = "eininwfsum"
      iz=0
      iz0=0

      linperr=.false.
      rewind(iin)
      lrdcoeff=.false.



 10   continue

      call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      if (ierr .eq. 1) then
          message = 'Cannot read input file'
          call errormsg
      endif

      if (keyword(1).eq.'end-input' .or. &
          keyword(1).eq.'end-inwf-section' ) then
         message='No INIT_WF-SECTION in input file'
         call errormsg
      elseif (keyword(1).eq.'init_wf-section') then
         goto 100
      else
         goto 10
      end if



 100  call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)

      if(irdwf.eq.590  .and.  keyword(1) .ne. 'file') then
         routine='eininwfsum'
         message='The first keyword in INWF-SECTION' &
            //'must be  "file=...", not :'
         call wrline(1,iz,iz0)
         stop
      endif

      if(keyword(1) .eq. 'file') then
         if (lrdcoeff) then
            message="A 'file' statement has to be followed by a " &
                 //"'coeff' statement in a new line in INIT_WF-SECTION"
            call errormsg
         end if
         if(keyword(2) .eq. '=') then
            nfiles=nfiles+1
            scoeff(nfiles)=cmplx(1.d0,0.d0)
            rname = keyorig(3)
            rlaenge = lc(3)
            call abspath(rname,rlaenge)
            irdwf=irdwf+1
            if(nfiles.gt.8) then
               routine='eininwfsum'
               message='Too many (more then 8) files in INWF-SECTION'
               call errormsg
            endif
            rname=rname(1:rlaenge)//'/restart'
            rlaenge=rlaenge+8

            open(irdwf,file=rname,form='unformatted',status='old',err=999)
            write(ilog,'(a,i4,2a)') " Channel",irdwf, &
                 ' opened : ',rname(1:rlaenge)
         else
            routine='eininwfsum'
            message='Expecting "=" after "file", not :'//keyword(2)
            call wrline(2,iz,iz0)
         endif
         lrdcoeff=.true.

      elseif (keyword(1) .eq. 'coeff') then
         if (.not.lrdcoeff) then
            message="A 'coeff' statement has to follow  a " &
                 //"'file' statement in a new line in INIT_WF-SECTION"
            call errormsg
         end if
         imcoeff=0.d0
         recoeff=0.d0

         if(keyword(2).eq.'=') then
            read(keyword(3),*,iostat=ierr) recoeff

            if (ierr.ne.0) then
               routine='eininwfsum'
               message='Expecting real number after "coeff =", ' &
                    //'not: '//keyword(3)
               call errormsg
            endif
         else
            routine='eininwfsum'
            message='Expecting "=" after "coeff",'//' not: '//keyword(4)
            call errormsg
         end if


         if(keyword(4).eq.',') then
            read(keyword(5),*,iostat=ierr) imcoeff
            if (ierr.ne.0) then
               routine='eininwfsum'
               message='Expecting real number after "coeff =R,", ' &
                    //'not: '//keyword(3)
               call errormsg
            endif
         end if

         scoeff(nfiles)=dcmplx(recoeff,imcoeff)
         lrdcoeff=.false.
      elseif(keyword(1) .eq. 'end-init_wf-section') then
         goto 200
      else
         message="Unknown keyword: "//keyword(1)
      end if
      goto 100


 200  continue

      write(ilog,*)
      write(ilog,'(a)') 'Coefficients "C_i" for superposition:'
      write(ilog,*)
      write(ilog,'(a)') '   i       Re[C_i]       Im[C_i]'
      write(ilog,'(a)')"--------------------------------------------"
      do i=1,nfiles
         write(ilog,'(i4,2f16.8)') i, dble(scoeff(i)),imag(scoeff(i))
      end do
      write(ilog,*)
      if (lsnormalize) then
         write(ilog,'(/a,f16.8)') 'Output WF will be normalized to:',snorm
      else
         write(ilog,'(/a,f16.8)') 'Output WF will not be normalized.'
      end if
      write(ilog,*)

      return

 999  routine='eininwfsum'
      message='Cannot open file:'//rname(1:rlaenge)
      call errormsg

      end subroutine eininwfsum! of eininwfsum

!***********************************************************************
!              SUMCMD
!
!     Collect command line options
!***********************************************************************

      subroutine sumcmd(lmnd)

use global
use aglobal
use maxv, only: maxdim,maxsta
use griddatmod
use channels
use chnpotmod
use joinpsimod
use fluxmod
use sumrstmod

      implicit none

      integer(long)                :: iarg, myiargc, i,linput,ilbl
      character(len=80)            :: buf
      logical(kind=4), intent(out) :: lmnd

      routine = "sumcmd"

      filein = '  '
      fileout= '  '
      name   = './'
      logfile = 'sumrst.log'
      laenge = 0

      lgnu = .false.
      lmnd = .false.
      lsnormalize=.false.

      iarg = myiargc()
      if (iarg .lt. 1) goto 200

      do i = 1, iarg
         call mygetarg(i,buf)
         ilbl  = index(buf,' ')
         ilbl = min(ilbl,C5-optlaenge)
         if(ilbl.gt.0) options = options(1:optlaenge)//buf(1:ilbl)
         optlaenge = optlaenge + ilbl
      end do

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if( buf .eq. '-ver' ) then
            call wrversion(6)
            stop
         else if( buf .eq. '-mnd' ) then
            lmnd=.true.
         else if (buf(2:3) .eq. 'D ') then
            i=i+1
            call mygetarg(i,name)
            laenge=index(name,' ')-1
            call abspath(name,laenge)
         else if( buf(1:2) .eq. '-w' ) then
            lovwr=.true.
         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(78a1)')  ('-',i=1,78)

            write(6,'(1x,a,2(/11x,a))') &
       'Purpose : Calculate the superposition of up to 8 wavefunctions', &
       '(using restart files). The superposition is written to the', &
       'restart file in the name directory.'
            write(6,'(/1x,a)') &
       'Usage   : sumrst$ver [-mnd -D -w -ver -h -?] file.inp  .'
            write(6,'(/,a)') ' Options : '
            write(6,'(a)') ' -mnd    : Make name directory.'
            write(6,'(a/,11x,a)') &
                 " -D name : 'name' denotes the directory where files" &
                 //"are written to,","(name in inpf file ignored)."
            write(6,'(2a)') &
                 ' -w      : Existing data is overwritten.'  
            write(6,'(a)') &
                 ' -ver    : Version information about the program.'
            write(6,'(a,/a)') &
                 ' -h      : Print this help text. ', &
                 ' -?      : Print this help text. '
            write(6,'(78a1)') ('-',i=1,78)
            stop

         else
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      end if
      if (i .lt. iarg-1) go to 100

!***********************************************************************
! get any compulsory arguments if in argument list
!***********************************************************************
      call mygetarg(i,inputfile)
      linput=index(inputfile,' ')-1
      if(linput .gt. 3) then
         if(inputfile(linput-3:linput) .ne. '.inp' ) then
            inputfile=inputfile(1:linput)//'.inp'
            linput=linput+4
         endif
      elseif(linput.gt.0) then
         inputfile=inputfile(1:linput)//'.inp'
         linput=linput+4
      else
         goto 200
      endif
      call abspath(inputfile,linput)

      return

 200  continue
      write(6,'(a)') ' There is no input file given'
      stop

      end subroutine sumcmd

      end program sumrst
