!#######################################################################
!                    poteval  (genoper/poteval.F)
!
! Calculates the value of the potential at coordinates contained in
! gpoint.
! ppoint is the momentum at gpoint, needed for classical trajectories
! 
! If the value is real, both output variables v and vz contain the same
! number. If the value is complex, v is set to zero. 
!
! contains: poteval,poteval1d,dereval,dereval1da,derevalmd,dereval1dn,
!           fpoteval,calcdiab,calcpes
!           callmuld,callmulder,functhphi,vqchem,derqchem
!-----------------------------------------------------------------------
      module potevalmod

use decimal, only: dop, long
use logdat
use op1lib,  only: zerovxi, cpvxd, zerovxl, zeromxd, zeromxz, cpmxd, cpmxz

use global
use maxv, only: maxdim,maxsta
use dvrdatmod
use griddatmod
use operdef
use psidef

use psiutil, only: f2mdecoder
use operutil, only: getvgn

use callmuldmod, only:callmuld,callmulder
use callanlzmod, only:callanlz
use callanldmod, only:callanld,callderd

use dirdyn, only: lgausspot,ldbsmall
use hpsimod, only:fddx, dermethod, gwpintordm
use timing
use dirdyn, only: num_gp

use funcgapmod, only: potgap
use lalib, only: simtranbd

      implicit none
      private
      public :: poteval,dereval,calcdiab,calcpes,poteval1d,calcdiabder,&
                alloc_klha,dealloc_klha,klhacalc,klhaconfig,gengconfig1m,&
                genpconfig1m

! arrays needed for LHA of muld potentials (klhaconfig, klhacalc)
      public :: mom0,mom1,mom2,gconfig,pconfig,gconfder1
      real(dop), dimension(:,:,:), allocatable,save     :: mom0
      real(dop), dimension(:,:,:,:), allocatable,save   :: mom1
      real(dop), dimension(:,:,:,:,:), allocatable,save :: mom2
      real(dop), dimension(:,:,:), allocatable,save     :: gconfig
      real(dop), dimension(:,:,:), allocatable,save     :: pconfig
      real(dop), dimension(:,:,:,:), allocatable,save   :: gconfder1

      contains

!-----------------------------------------------------------------------
      subroutine poteval (h,v,gpoint,vz,ppoint,pottyp)

      implicit none

      integer(long)                            :: h1,nh
      integer(long), intent(in)                :: h
      integer(long), intent(out)               :: pottyp
      real(dop)                                :: rfac,rterm
      real(dop), intent(out)                   :: v
      real(dop), dimension(*), intent(in)      :: gpoint,ppoint
      complex(dop)                             :: zfac, zterm
      complex(dop), intent(out)                :: vz

! Do not time routine - very slow in OpenMP
   ! call get_timer(tid, 'poteval')
   ! call start_timer(tid)

!-----------------------------------------------------------------------
! htmform(h) = 0: operator function, evaluated in poteval1d
! htmform(h) = 1: Operator is product of functions. In this case 
!                 hterm(nh,h) points to another term (which is evaluated
!                 in poteval1d)
!-----------------------------------------------------------------------
      if (htmform(h) .eq. 0) then
         call poteval1d (h,v,gpoint,vz,ppoint,pottyp)
      else if (htmform(h) .eq. 1) then
         v=0.0_dop
         vz=0.0_dop
         rterm=1.0_dop
         zterm=(1.0_dop,0.0_dop)  

         do nh=1,htmfac(h)
            h1=hterm(nh,h)
            if (h1 .eq. 0) then
               v=v+rterm
               rterm=1.0_dop
               vz=vz+zterm
               zterm=(1.0_dop,0.0_dop)  
            else
               rfac=0.0_dop
               zfac=0.0_dop
               call poteval1d(h1,rfac,gpoint,zfac,ppoint,pottyp)
               rterm=rterm*rfac
               zterm=zterm*zfac
            endif
         enddo
         v  = v  + rterm
         vz = vz + zterm
      endif

!       call stop_timer(tid)

      return
      end subroutine poteval

!#######################################################################
!             poteval1d  (genoper/poteval.F)
!
! For a given term h, this routine evaluates the analytic 1D functions
! and multiplies (and sums) the factors hl.
! If the value is real, both output variables v and vz contain the same
! number. If the value is complex, v is set to zero. 
!-----------------------------------------------------------------------

      subroutine poteval1d (h,v,gpoint,vz,ppoint,pottyp)

      implicit none

      integer(long)                                :: hl,nh,f,m,muld,f1,gf,nf,n
      integer(long), intent(in)                    :: h
      integer(long), intent(out)                   :: pottyp
      integer(long), dimension(maxdim)             :: grid
      real(dop)                                    :: rfac,rterm
      real(dop), intent(out)                       :: v
      real(dop), dimension(maxdim)                 :: gpoint1
      real(dop), dimension(maxdim), intent(in)     :: gpoint,ppoint
      complex(dop)                                 :: zfac,zterm
      complex(dop), intent(out)                    :: vz

      v=0.0_dop
      vz=0.0_dop
      rterm=1.0_dop
      zterm=(1.0_dop,0.0_dop)  
      pottyp = 0

      f=htmdof(h)
      m=htmmode(h)
      muld=htmmuld(h)
 
      nf=-999
      f1=-999
      if (muld .ne. 0) then
         nf=nmulddof(muld)
         do n=1,nf
            grid(n)=mulddof2(n,muld)  ! Use the original order
         enddo
         f1=mulddof(1,muld)
      else if (m .ne. 0) then
         nf=nspfdof(m)
         do n=1,nf
            grid(n)=spfdof(n,m)
         enddo
         f1=spfdof(1,m)
      else if (f .ne. 0) then
         nf=1
         grid(1)=f
         f1=f
         if (f .eq. h) then
            vz=(1.0_dop,0.0_dop)
            v=1.0_dop
            return
         endif
      endif

      do nh=1,htmfac(h)
         hl=hterm(nh,h)
         if (hl .eq. 0) then
            v=v+rterm
            rterm=1.0_dop
            vz=vz+zterm
            zterm=(1.0_dop,0.0_dop)
         else
            rfac=0.0_dop
            zfac=0.0_dop
            if (hopfile(hl).gt.200 .and. hopfile(hl).le.300) then
               gpoint1 = 0.0_dop
               if (ldd) then
                  do gf=1,ndof
                     gpoint1(gf)=gpoint(gf)
                  enddo
               else
                  do gf=1,nf
                     f1=grid(gf)
                     gpoint1(gf)=gpoint(f1)
                  enddo
               endif

               call callmuld (rfac,gpoint1,ndof,hopilab(hl), &
                    hopfile(hl),hopipar(1,hl),hoprpar(1,hl), &
                    ddsi(h),ddsf(h),hoplab(hl))
               zfac = rfac
            else if (hopfile(hl).gt.300 .and. hopfile(hl).le.400) then
               call callanld(rfac,gpoint(f1),hopilab(hl),hopfile(hl), &
                       hopipar(1,hl),hoprpar(1,hl),ppoint(f1),pottyp, &
                       hoplab(hl),mass(f1))
               zfac = rfac
            else if (hopfile(hl).gt.100 .and. hopfile(hl).lt.200) then
               call callanlz(zfac,gpoint(f1),hopilab(hl),hopfile(hl), &
                       hopipar(1,hl),hoprpar(1,hl))
               rfac=0.0_dop
               rterm=0.0_dop
               v=0.0_dop
            else
               write(6,*)
               write(6,*) ' hl =',hl,' , hopfile(hl) =',hopfile(hl),  &
                          ' , hopilab(hl) =', hopilab(hl) 
               write(6,*) ' hoplab(hl) = ',hoplab(hl)
               write(2,*)
               write(2,*) ' hl =',hl,' , hopfile(hl) =',hopfile(hl), &
                          ' , hopilab(hl) =', hopilab(hl)
               write(2,*) ' hoplab(hl) = ',hoplab(hl)
               routine='poteval1d (poteval.F)'
               message='Analytic potential not in analytic files'
               call errormsg
            endif
            rterm=rterm*rfac
            zterm=zterm*zfac
         endif
      enddo

      v  = v  + rterm
      vz = vz + zterm

      return
      end subroutine poteval1d

!#######################################################################
!
! DEREVAL
! Returns the derivative of a potential term h
!
! dertyp: returned from library
!         0  derivative wrt coordinates
!         1  derivative wrt momentum
! dermethod: select numerical derivative routine (not actually used)
!            0 standard
!            1 uses Ridder's extrapolation
!
!#######################################################################

      subroutine dereval (h,deriv,gpoint,order,ppoint,dertyp,dermethod)

      implicit none

      integer(long), intent(in)                    :: h
      integer(long), intent(out)                   :: dertyp,dermethod
      integer(long), dimension(maxdim), intent(in) :: order
      integer(long)                                :: f
      real(dop), intent(out)                       :: deriv
      real(dop), dimension(maxdim), intent(in)     :: gpoint,ppoint

      deriv=0.0_dop
! if derivative is of a ghost / point / feb mode return zero
      do f = 1,ndof
         if (order(f) .gt. 0) then
            if (basis(f) .eq. 19 .or. dofspf(f) .eq. 0 &
                   .or. f .eq. feb) return
         endif
      enddo

!-----------------------------------------------------------------------
! htmderiv(h) = 0: numerical derivative
!               1: analytic derivative
!
! htmform(h) = 0: derivative of simple function
!              1: derivative of compound function
!-----------------------------------------------------------------------
      if (htmform(h) .eq. 0) then
         if (htmderiv(h) .eq. 0) then
            call dereval1dn (h,deriv,gpoint,order,ppoint,dertyp,&
                 dermethod)
         else if (htmderiv(h) .eq. 1) then
            call dereval1da (h,deriv,gpoint,order,ppoint,dertyp)
         endif

      else if (htmform(h) .eq. 1) then
         call derevalmd (h,deriv,gpoint,order,ppoint,dertyp,dermethod)

      endif

      return
      end subroutine dereval

!#######################################################################

      subroutine dereval1da (h,deriv,gpoint,order,ppoint,dertyp)

      implicit none

      integer(long)                                :: hl,nh,f,m,muld,f1, &
                                                      nfac,nconst,dertyp1
      integer(long), intent(in)                    :: h
      integer(long), intent(out)                   :: dertyp
      integer(long), dimension(maxdim), intent(in) :: order
      real(dop)                                    :: rfac,rterm
      real(dop), intent(out)                       :: deriv
      real(dop), dimension(maxdim), intent(in)     :: gpoint,ppoint

      deriv=0.0_dop
      rterm=1.0_dop

      f=htmdof(h)
      m=htmmode(h)
      muld=htmmuld(h)

      f1=-999
      if (muld .ne. 0) then
         f1=mulddof(1,muld)
      else if (m .ne. 0) then
         f1=spfdof(1,m)
      else if (f .ne. 0) then
         f1=f
      endif

!-----------------------------------------------------------------------
! need to account for the case when a term is a constant. As constants
! may be in hopfile 301, or may be hopilab=1,2 in hopfile 302 (1 and -1), 
! need to count factors (nfac) and constants (nconst)in a term. Term is 
! then set to zero if only constants present.
!-----------------------------------------------------------------------
      nfac=0
      nconst=0
      dertyp=0
      do nh=1,htmfac(h)
         hl=hterm(nh,h)
         if (hl .eq. 0) then
            if (order(f1) .gt. 0 .and. nconst .eq. nfac) rterm=0.0_dop
            deriv=deriv+rterm
            rterm=1.0_dop
            nfac=0
            nconst=0
            dertyp=0
         else
            nfac=nfac+1
            if (hopfile(hl) .eq. 301) then
               nconst=nconst+1
            else if (hopfile(hl) .eq. 302) then
               if (hopilab(hl) .eq. 1 .or. hopilab(hl) .eq. 2) then
                  nconst=nconst+1
               endif
            endif

            rfac=0.0_dop
            if (hopfile(hl).gt.200 .and. hopfile(hl).le.300) then
               call callmulder(rfac,gpoint(f1),ndof,hopilab(hl), &
                    hopfile(hl),hopipar(1,hl),order,ddsi(h),ddsf(h))
            else if (hopfile(hl).gt.300 .and. hopfile(hl).le.400) then
               call callderd(rfac,gpoint(f1),hopilab(hl),hopfile(hl), &
                    hopipar(1,hl),hoprpar(1,hl),order(f1),ppoint(f1), &
                    dertyp1,mass(f1))
               dertyp=max(dertyp,dertyp1)
            else
               routine='Dereval'
               message='Analytic potential not in analytic files'
               call errormsg
            endif
            rterm=rterm*rfac
         endif
      enddo

!-----------------------------------------------------------------------
! if only single term which is a constant, set to zero
!-----------------------------------------------------------------------
      if (order(f1) .gt. 0 .and. nconst .eq. htmfac(h)) rterm=0.0_dop

      deriv=deriv+rterm
      

      return
      end subroutine dereval1da

!#######################################################################

      subroutine derevalmd (h,deriv,gpoint,order,ppoint,dertyp, &
                 dermethod)

      implicit none

      integer(long)                                :: h1,nh,nh1,nh2,f,n,nderdof
      integer(long), intent(in)                    :: h
      integer(long), intent(out)                   :: dertyp,dermethod
      integer(long), dimension(maxdim), intent(in) :: order
      integer(long), dimension(maxdim)             :: derdof
      real(dop)                                    :: rfac,rterm
      real(dop), intent(out)                       :: deriv
      real(dop), dimension(maxdim), intent(in)     :: gpoint,ppoint
      logical(kind=4)                              :: lder
      logical(kind=4), dimension(maxdim)           :: found

!-----------------------------------------------------------------------
! set up array containing dofs wrt which derivative is taken
!-----------------------------------------------------------------------
      nderdof=0
      call zerovxi(derdof,ndof)
      do f=1,ndof
         if (order(f) .gt. 0) then
            nderdof=nderdof+1
            derdof(nderdof)=f
         endif
      enddo

      nh1=1
      nh2=1

      deriv=0.0_dop
   10 continue
! term is bounded by pointers nh1:nh2
      h1=hterm(nh2,h)
      if (h1 .ne. 0) then
         nh2=nh2+1
         if (nh2 .le. htmfac(h)) go to 10
      endif
      nh2=nh2-1

! see which dofs have a non-unit factor in term
      call zerovxl(found,ndof)
      do nh=nh1,nh2
         h1=hterm(nh,h)
         f=htmdof(h1)
         found(f)=.true.
      enddo

! see if dofs wrt which derivative is taken are functions
      lder = .true.
      do n=1,nderdof
         f=derdof(n)
         if (.not. found(f)) lder=.false.
      enddo

! if taking derivative of unit, derivative is zero
      if (.not. lder) go to 100
         
! now find derivative
      rterm=1.0_dop
      do nh=nh1,nh2
         h1=hterm(nh,h)
         rfac=0.0_dop
         if (htmderiv(h1) .eq. 0) then
            call dereval1dn (h1,rfac,gpoint,order,ppoint,dertyp, &
                 dermethod)
         else if (htmderiv(h1) .eq. 1) then
            call dereval1da (h1,rfac,gpoint,order,ppoint,dertyp)
         endif
         rterm=rterm*rfac
      enddo
      deriv=deriv+rterm

  100 continue
! move on to next term
      nh1=nh2+2
      nh2=nh1
      if (nh1 .le. htmfac(h)) go to 10

      return
      end subroutine derevalmd

!#######################################################################

      subroutine dereval1dn (h,deriv,gpoint,order,ppoint,dertyp, &
                 dermethod)

      implicit none

      integer(long)                              :: f,m,muld,f1,n,nderdof,pottyp
      integer(long), intent(in)                  :: h
      integer(long), intent(out)                 :: dertyp,dermethod
      integer(long),dimension(maxdim)            :: derdof
      integer(long),dimension(maxdim),intent(in) :: order
      real(dop)                                  :: fddxlocal,temp,temp1,vp,vm, &
                                                    diff,v0,vpp,vmm,vpm,vmp,err
      real(dop), intent(out)                     :: deriv
      real(dop), dimension(maxdim*2)             :: work
      real(dop), dimension(maxdim)               :: dq,qp,qm,pp,pm,qpm,qmp,&
                                                    qpp,qmm,ppm,pmp,ppp,pmm
      real(dop), dimension(maxdim), intent(in)   :: gpoint,ppoint
      complex(dop)                               :: vz
      real(dop), external                        :: numder1,numder1r

      fddxlocal=0.01
      deriv=0.0_dop
            
!-----------------------------------------------------------------------
! get step value for derivatives
!-----------------------------------------------------------------------
      if (fddx>0) then
        dq(1:ndof)=fddx
      else
        do f=1,ndof
          temp = 1/3.0_dop
          temp1 = macheps**temp
          temp = temp1*gpoint(f)
          dq(f) = max(temp,temp1)
        enddo
      endif

      f=htmdof(h)
      m=htmmode(h)
      muld=htmmuld(h)

      nderdof=0
      call zerovxi(derdof,ndof)
      if (muld .ne. 0) then
         do n=1,nmulddof(muld)
            f = mulddof(n,muld)
            if (f .eq. 0) cycle

            if (order(f) .gt. 0) then
               nderdof=nderdof+1
               derdof(nderdof)=f
            endif
         enddo
            
      else if (m .ne. 0) then
         do n=1,nspfdof(m)
            f = spfdof(n,m)
            if (order(f) .gt. 0) then
               nderdof=nderdof+1
               derdof(nderdof)=f
            endif
         enddo
            
      else if (f .ne. 0) then
         if (order(f) .gt. 0) then
            nderdof=1
            derdof(nderdof)=f
         endif

      endif

      if (nderdof .eq. 0) then
         call poteval (h,deriv,gpoint,vz,ppoint,pottyp)

      else if (nderdof .eq. 1) then
         f=derdof(1)

         if (order(f) .eq. 1) then
            dertyp = 0           
            dermethod=0
            if (dermethod .eq. 1) then
               deriv = numder1r(fpoteval,gpoint,ndof,f,fddxlocal,err,work,h)
            else 
               deriv = numder1(fpoteval,gpoint,ndof,f,work,macheps,h)
            endif

         else if (order(f) .eq. 2) then
            call cpvxd(gpoint,qp,ndof)
            call cpvxd(gpoint,qm,ndof)
            qp(f)  = gpoint(f) + dq(f)
            qm(f)  = gpoint(f) - dq(f)

            call cpvxd(ppoint,pp,ndof)
            call cpvxd(ppoint,pm,ndof)
            pp(f)  = ppoint(f) + dq(f)
            pm(f)  = ppoint(f) - dq(f)

            call poteval(h,v0,gpoint,vz,ppoint,dertyp)
            call poteval(h,vp,qp,vz,pp,dertyp)
            call poteval(h,vm,qm,vz,pm,dertyp)

            diff = vp - 2.0_dop*v0 + vm

            deriv = diff/(dq(f)**2)
         endif

      else if (nderdof .eq. 2) then
         f=derdof(1)
         f1=derdof(2)

         if (order(f) .eq. 1 .and. order(f1) .eq. 1) then
            call cpvxd(gpoint,qpp,ndof)
            call cpvxd(gpoint,qpm,ndof)
            call cpvxd(gpoint,qmp,ndof)
            call cpvxd(gpoint,qmm,ndof)
            qpp(f)  = gpoint(f)  + dq(f)
            qpp(f1) = gpoint(f1) + dq(f1)
            qpm(f)  = gpoint(f)  + dq(f)
            qpm(f1) = gpoint(f1) - dq(f1)
            qmp(f)  = gpoint(f)  - dq(f)
            qmp(f1) = gpoint(f1) + dq(f1)
            qmm(f)  = gpoint(f)  - dq(f)
            qmm(f1) = gpoint(f1) - dq(f1)

            call cpvxd(ppoint,ppp,ndof)
            call cpvxd(ppoint,ppm,ndof)
            call cpvxd(ppoint,pmp,ndof)
            call cpvxd(ppoint,pmm,ndof)
            ppp(f)  = ppoint(f)  + dq(f)
            ppp(f1) = ppoint(f1) + dq(f1)
            ppm(f)  = ppoint(f)  + dq(f)
            ppm(f1) = ppoint(f1) - dq(f1)
            pmp(f)  = ppoint(f)  - dq(f)
            pmp(f1) = ppoint(f1) + dq(f1)
            pmm(f)  = ppoint(f)  - dq(f)
            pmm(f1) = ppoint(f1) - dq(f1)

            call poteval(h,vpp,qpp,vz,ppp,dertyp)
            call poteval(h,vpm,qpm,vz,ppm,dertyp)
            call poteval(h,vmp,qmp,vz,pmp,dertyp)
            call poteval(h,vmm,qmm,vz,pmm,dertyp)

            diff = vpp - vpm - vmp + vmm

            deriv = diff/(4.0_dop*dq(f)*dq(f1))

         endif

      endif

      return
      end subroutine dereval1dn

!#######################################################################
!
! FPOTEVAL
!
! function required by numerical derivative library routines 
! to evaluate potential function
!
! It is assumed that routine is not used for complex potentials
! or derivatives of momentum (should be analytic functions)
!
!#######################################################################

      real(dop) function fpoteval(gpoint,h)

      implicit none

      integer(long)                           :: dertyp
      integer(long), intent(in)               :: h
      real(dop)                               :: tmp,v
      real(dop), dimension(*), intent(in)     :: gpoint
      real(dop), dimension(maxdim)            :: ppoint
      complex(dop)                            :: vz

! zero dummy quantities
      ppoint(1)=0.0_dop
      vz=0.0_dop
      dertyp=0

      call poteval(h,v,gpoint,vz,ppoint,dertyp)
      if (dertyp .gt. 1) then
         routine = "Fpoteval (poteval.F)"
         message = "Derivative of momentum in potential function"
         call errormsg
      endif
      tmp=dimag(vz)
      tmp=abs(tmp)
      if (tmp .gt. 64*macheps) then
         routine = "Fpoteval (poteval.F)"
         message =  &
       "Cannot form numerical derivative of complex potential"
         call errormsg
      endif

      fpoteval = v

      return
      end function fpoteval

!#######################################################################
!
! CALCDIAB
!
! calculates the diabatic - adiabatic transformation matrix (rotmat)
! at a point specified by the integer array point(f)
!
! called by various analyse programs
!
! rotmatz is the (complex) ADT rotation matrix 
! pesad(s) = dconjg(rotmatz(s1,s)) * pesdia(s1,s2) * rotmatz(s2,s)
!#######################################################################

      subroutine calcdiab(hops,pesad,pesdia,rotmatz,point,gpoint,nham)

use dirdyn, only: lgausspot

      implicit none

      integer(long)                                      :: error,izflag,s,s1
      integer(long)                                      :: i,j,info
      integer(long), intent(in)                          :: nham
      integer(long), dimension(maxdim), intent(in)       :: point
      real(dop), dimension(maxdim), intent(in)           :: gpoint
      real(dop), dimension(maxsta), intent(out)          :: pesad
      real(dop), dimension(3*maxsta)                     :: workr
      real(dop), dimension(hopsdim), intent(in)          :: hops
      real(dop), dimension(maxsta,maxsta)                :: rotmat,tmprot
      real(dop), dimension(maxsta)                       :: tmpe,tmpe2
      real(dop), dimension(1,1)                          :: rdum,rdum1
      complex(dop)                                       :: rotdet
      real(dop), dimension(maxsta,maxsta), intent(out)   :: pesdia
      complex(dop), dimension(2*maxsta)                  :: workz
      complex(dop), dimension(maxsta,maxsta)             :: pesdiaz
      complex(dop), dimension(maxsta,maxsta), intent(out):: rotmatz

      izflag=0
      if(lddmctdh.or.lgausspot)then
         call calcpesgap(pesdia,gpoint)        
      else
         call calcpes(hops,pesdia,point,gpoint,pesdiaz,izflag,nham)
      endif

      if (nddstate .eq. 1) then
         pesad(1)=pesdia(1,1)
         rotmat(1,1) = 1.0_dop
      else if (nddstate .gt. 1 .and. izflag .eq. 0) then
         call cpmxd(pesdia,rotmat,maxsta,maxsta)
         call dsyev('V','U',nddstate,rotmat,maxsta,pesad,workr, &
                    3*maxsta,error)
         if (error .ne. 0) then
            write(6,'(a)') &
                     'Error diagonalising diabatic PES matrix'
            stop
         endif

! Check all vectors have positive diagonal element.
! (first vector lies in upper half of rotation plane.)
          do i=1,nddstate
             if(rotmat(i,i).lt.0.0_dop) then
                do j = 1,nddstate
                   rotmat(j,i) = -rotmat(j,i)
                enddo
             endif
          enddo

! Determinant of the transformation matrix.
         tmprot = rotmat
         call dgeev('N','N',nddstate,tmprot,maxsta,tmpe,tmpe2,&
              rdum,1,rdum1,1,workr,3*maxsta,info)
         rotdet = (1.0_dop,0.0_dop)
         do i = 1, nddstate
            rotdet = rotdet*dcmplx(tmpe(i),tmpe2(i))
         enddo

! Modify the sign of the transformation matrix if needed.
         if(dble(rotdet).lt.0.0_dop)then
            do i = 1,nddstate
               rotmat(i,2) = -1.0_dop*rotmat(i,2)
            enddo
         endif

      else if (nddstate .gt. 1 .and. izflag .eq. 1) then
         call cpmxz(pesdiaz,rotmatz,maxsta,maxsta)
         call zheev('V','U',nddstate,rotmatz,maxsta,pesad,workz, &
                    2*maxsta,workr,error)
         if (error .ne. 0) then
            write(6,'(a)') &
                  'Error diagonalising complex diabatic PES matrix'
            stop
         endif

      endif

      if (izflag .eq. 0) then
         do s=1,nddstate
            do s1=1,nddstate
               rotmatz(s1,s) = rotmat(s1,s)
            enddo
         enddo
      endif

      return
      end subroutine calcdiab

!#######################################################################
!
! nham defines which operator is used
!
! Calculates the diabatic potential matrix at the point gpoint
!#######################################################################

      subroutine calcpes(hops,pesdia,point,gpoint,pesdiaz,izflag,nham)

      implicit none
      
      logical(kind=4)                                     :: lkin
      integer(long), intent(in)                           :: nham
      integer(long)                                       :: s,m,f,k1,k,h,s1,&
                                                             n,p,nm,f1,muld,&
                                                             zeig,idum
      integer(long)                                       :: vd,gd,nd,&
                                                             lhaorder_old
      integer(long), intent(out)                          :: izflag
      integer(long), dimension(maxdim)                    :: mdvgdim
      integer(long), dimension(*), intent(in)             :: point
      integer(long), dimension(maxdim)                    :: mpoint
      real(dop)                                           :: pescorr,vss,tmp,&
                                                             koelha
      real(dop), dimension(maxdim)                        :: rdum
      real(dop), dimension(maxdim), intent(in)            :: gpoint
      real(dop), dimension(hopsdim), intent(in)           :: hops
      real(dop), dimension(maxsta,maxsta), intent(out)    :: pesdia

      complex(dop)                                        :: vzz,pescorrz
      complex(dop), dimension(maxsta,maxsta), intent(out) :: pesdiaz
      logical(kind=4), dimension(maxkoe)                  :: kignore

      call zeromxd(pesdia,maxsta,maxsta)
      call zeromxz(pesdiaz,maxsta,maxsta)
      call f2mdecoder(point,mpoint)

      izflag = 0
      lhaorder_old = lhaorder
      lhaorder = 0   ! only want energy

!-----------------------------------------------------------------------
!  calculate coefficients if from LHA
!-----------------------------------------------------------------------
      kignore(:) = .false.
      do k=1,khtot(nham)
         h=hamilton(fkoe,k)
         if (htmtype(h) .eq. 104) then
            call klhacalc1(h,koelha,gpoint,kignore(k))
            koeff(k) = koelha
         endif
      enddo

!-----------------------------------------------------------------------
! Uncorrelated terms.
! for single-set calculations nstate = 1 and nddstate is the no. of
! electronic states
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (m .eq. meb) cycle
            f1=spfdof(1,m)
            do k1=1,khzahl(m,s,nham)
               k=zhun(m,s,nham)+k1-1
               h=hamilton(f1,k)
! ignore any KE terms
               if (kinop(m,k) .ne. 0) cycle
               
               if (htmtype(h) .ge. 101) then
                  call poteval (h,vss,gpoint,vzz,rdum,idum)
                  vss = dble(koeff(k))*vss
                  vzz = koeff(k)*vzz
                  
               else
                  vss=dble(koeff(k))
                  vzz=koeff(k)
                  do n=1,nspfdof(m)
                     f1=spfdof(n,m)
                     h=hamilton(f1,k)
                     
                     if (htmtype(h) .eq. 1) cycle
                     
                     call getvgn(h,vd,gd,nd)
                     p = (mpoint(m)-1)/vd
                     p = mod(p,gd) + zhop(h)
                     vss=vss*hops(p)
                     vzz=vzz*hops(p)
                  enddo
                  
               endif
               if (.not. lmult .and. nddstate .gt. 1) then
! single-set
                  do s1 = 1, nddstate
                     pesdia(s1,s1)=pesdia(s1,s1)+vss
                     pesdiaz(s1,s1)=pesdiaz(s1,s1)+vzz
                  enddo
               else
! multi-set
                  pesdia(s,s)=pesdia(s,s)+vss
                  pesdiaz(s,s)=pesdiaz(s,s)+vzz
               endif
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! add any uncorrelated electronic state terms - needed for single set
!-----------------------------------------------------------------------
      if (leb .and. nddstate .gt. 1) then
         do k1=1,khzahl(meb,1,nham)
            k=zhun(meb,1,nham)+k1-1
            h=hamilton(feb,k)
            pescorr=dble(koeff(k))
            pescorrz=koeff(k)
            if (htmtype(h) .eq. 1 .or. htmtype(h) .eq. 2) then
               call ssetpes2(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            else
               call ssetpes(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            endif
         enddo
      endif

!-----------------------------------------------------------------------
! add correlated terms
!-----------------------------------------------------------------------
outer:do k=1,kzahl(nham)

! ignore PES terms with order > 0
         if (kignore(k)) cycle

! ignore any KE terms
         lkin = .false.
         do m=1,nmode
            lkin = lkin .or. kinop(m,k).ne.0
         enddo
         if(lkin)cycle outer

         pescorr=dble(koeff(k))
         pescorrz=koeff(k)

         inner:do m=1,nmode
            if (m .eq. meb) cycle inner
            if(lgausspot .and. m.gt.1)cycle
            
            f1=spfdof(1,m)
            h=hamilton(f1,k)
            
            if (htmtype(h) .ge. 101 .or. lgausspot) then
               call poteval (h,vss,gpoint,vzz,rdum,idum)
               pescorr=pescorr*vss
               pescorrz=pescorrz*vzz
               
            else
               do n=1,nspfdof(m)
                  f1=spfdof(n,m)
                  h=hamilton(f1,k)
                  
                  if (htmtype(h) .eq. 1) cycle
                  
                  call getvgn(h,vd,gd,nd)
                  p = (mpoint(m)-1)/vd
                  p = mod(p,gd) + zhop(h)
                  pescorr=pescorr*hops(p)
                  pescorrz=pescorrz*hops(p)
               enddo
               
            endif
         enddo inner
! sum up potential in operator matrix
         if (lmult) then
            s1=ksi(k)
            s=ksf(k)
            pesdia(s,s1)=pesdia(s,s1)+pescorr
            pesdiaz(s,s1)=pesdiaz(s,s1)+pescorrz
         else
            h=hamilton(feb,k)
            if (htmtype(h) .eq. 1 .or. htmtype(h) .eq. 2) then
               call ssetpes2(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            else
               call ssetpes(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            endif
         endif
      enddo outer

      do nm=1,nmulpot(nham)
         k=kzahl(nham)+nm
         pescorr=dble(koeff(k))
         pescorrz=koeff(k)

         f=fterm(k)
         h=hamilton(f,k)
         muld=htmmuld(h)
         if (htmtype(h) .ge. 101) then
            call poteval (h,vss,gpoint,vzz,rdum,idum)
            pescorr=pescorr*vss
            pescorrz=pescorrz*vzz
         else
            zeig=1
            do n=1,nmulddof(muld)
               f=mulddof(n,muld)
               mdvgdim(f)=zeig
               zeig=zeig*gdim(f)
            enddo

            p=zhop(h)
            do n=1,nmulddof(muld)
               f=mulddof(n,muld)
               p=p+mdvgdim(f)*(point(f)-1)
            enddo
            pescorr=pescorr*hops(p)
            pescorrz=pescorrz*hops(p)
         endif

! sum up potential in operator matrix
         if (lmult) then
            s1=ksi(k)
            s=ksf(k)
            pesdia(s,s1)=pesdia(s,s1)+pescorr
            pesdiaz(s,s1)=pesdiaz(s,s1)+pescorrz
         else
            h=hamilton(feb,k)
            if (htmtype(h) .eq. 1 .or. htmtype(h) .eq. 2) then
               call ssetpes2(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            else
               call ssetpes(pesdia,pesdiaz,pescorr,pescorrz,hops(zhop(h)))
            endif
         endif

      enddo 

      do s=1,nddstate
         do s1=1,nddstate
            tmp = dimag(pesdiaz(s1,s))
            if (dabs(tmp) .gt. tiny) izflag=1
         enddo
      enddo

      lhaorder = lhaorder_old

      return
      end subroutine calcpes

!#######################################################################
!
! nham defines which operator is used
!
! Calculates the diabatic potential matrix at the point gpoint
!#######################################################################

      subroutine calcpesgap(pesdia,gpoint)

      implicit none
      
      real(dop), dimension(maxdim), intent(in)            :: gpoint
      real(dop), dimension(maxsta,maxsta), intent(out)    :: pesdia

      call potgap(gpoint,pesdia,1,0)

      return
      end subroutine calcpesgap

!-----------------------------------------------------------------------

      subroutine ssetpes(pesdia,pesdiaz,pescorr,pescorrz,hops)

      implicit none

      integer(long)            :: s,s1
      real(dop), intent(in)    :: pescorr
      complex(dop), intent(in) :: pescorrz
      real(dop), dimension(maxsta,maxsta), intent(inout)    :: pesdia
      complex(dop), dimension(maxsta,maxsta), intent(inout) :: pesdiaz
      real(dop), dimension(nddstate,nddstate), intent(in)   :: hops

      do s=1,nddstate
         do s1=1,nddstate
            pesdia(s1,s) = pesdia(s1,s) + hops(s1,s)*pescorr
            pesdiaz(s1,s) = pesdiaz(s1,s) + hops(s1,s)*pescorrz
         enddo
      enddo
      
      end subroutine ssetpes

!-----------------------------------------------------------------------

      subroutine ssetpes2(pesdia,pesdiaz,pescorr,pescorrz,hops)

      implicit none

      integer(long)            :: s
      real(dop), intent(in)    :: pescorr
      complex(dop), intent(in) :: pescorrz
      real(dop), dimension(maxsta,maxsta), intent(inout)    :: pesdia
      complex(dop), dimension(maxsta,maxsta), intent(inout) :: pesdiaz
      real(dop), dimension(nddstate), intent(in)            :: hops

      do s=1,nddstate
         pesdia(s,s) = pesdia(s,s) + hops(s)*pescorr
         pesdiaz(s,s) = pesdiaz(s,s) + hops(s)*pescorrz
      enddo
      
      end subroutine ssetpes2

!-----------------------------------------------------------------------

      subroutine ssetder(f,derdia,deriv1,hops)

      implicit none

      integer(long),intent(in) :: f
      integer(long)            :: s,s1
      real(dop), intent(in)    :: deriv1
      real(dop), dimension(maxsta,maxsta,maxdim), intent(inout)    :: derdia
      real(dop), dimension(nddstate,nddstate), intent(in)   :: hops

      do s=1,nddstate
         do s1=1,nddstate
            derdia(s1,s,f) = derdia(s1,s,f) + hops(s1,s)*deriv1
         enddo
      enddo
      
      end subroutine ssetder

!-----------------------------------------------------------------------

      subroutine ssetder2(f,derdia,deriv1,hops)

      implicit none

      integer(long),intent(in) :: f
      integer(long)            :: s
      real(dop), intent(in)    :: deriv1
      real(dop), dimension(maxsta,maxsta,maxdim), intent(inout)    :: derdia
      real(dop), dimension(nddstate), intent(in)            :: hops

      do s=1,nddstate
         derdia(s,s,f) = derdia(s,s,f) + hops(s)*deriv1
      enddo
      
      end subroutine ssetder2

!#######################################################################

      subroutine klhacalc1(h,koelha,gpoint1,kignore)

      implicit none

      integer(long), intent(in)                     :: h
      integer(long)                                 :: f,f1,hl
      integer(long)                                 :: ilbl,jlbl,lablen,iord
      integer(long)                                 :: dertyp,pottyp,dermethod
      integer(long),dimension(maxdim)               :: order
      complex(dop)                                  :: vz
      real(dop),dimension(maxdim)                   :: ppoint
      real(dop), intent(out)                        :: koelha
      real(dop),dimension(ndof),intent(in)          :: gpoint1
      real(dop),dimension(ndof)                     :: gpoint 
      character(len=c2)                             :: label
      logical(kind=4), intent(out)                  :: kignore

      real(dop)                       :: mom0
      real(dop), dimension(ndof)      :: mom1
      real(dop), dimension(ndof,ndof) :: mom2
      ! Energy, gradient and Hessian arrays
      real(dop)                                     :: v0
      real(dop), dimension(ndof)                    :: v1
      real(dop), dimension(ndof,ndof)               :: v2

      gpoint=gpoint1

!-----------------------------------------------------------------------
! get label and order of term 
!-----------------------------------------------------------------------
      hl = hterm(1,h)
      label = hoplab(hl)
      lablen = index(label,' ')-1

      ilbl = index(label,'%')+1
      read(label(ilbl:ilbl),*) iord

!-----------------------------------------------------------------------
! Ignore terms higher than the order of integrals required
!-----------------------------------------------------------------------
      if (iord .gt. lhaorder) then
         kignore=.true.
         return
      endif

!-----------------------------------------------------------------------
! calculate Hamiltonian moments (just once per state)
!-----------------------------------------------------------------------
      v0=0.0
      v1=0.0
      v2=0.0

      ! add coordinates of point DVRs
      if (lpntdvr) then
         do f=1,ndof
            if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
         enddo
      endif

      call poteval(h,v0,gpoint,vz,ppoint,pottyp)
      mom0 = v0
      mom1(:) = 0.0d0
      mom2(:,:) = 0.0d0
     
      dermethod=1
      if (lhaorder .ge. 1) then
         do f=1,ndof
            if (dofspf(f) .eq. 0 .or. f .eq. feb) cycle
            order = 0
            order(f)=1
            call dereval(h,v1(f),gpoint,order,ppoint,dertyp,dermethod)
         enddo
         do f=1,ndof
            if (dofspf(f) .eq. 0 .or. f .eq. feb) cycle

            mom0 = mom0 - v1(f)*gpoint(f) 
            mom1(f) = mom1(f) + v1(f) 
         enddo
      endif
      if (lhaorder .ge. 2) then
         do f=1,ndof
            if (dofspf(f) .eq. 0 .or. f .eq. feb) cycle
            do f1=f,ndof
               if (dofspf(f1) .eq. 0) cycle
               order = 0
               order(f)=1
               order(f1)=order(f1)+1
               call dereval(h,v2(f,f1),gpoint,order,ppoint,dertyp,dermethod)
            enddo
         enddo

         do f=1,ndof
            if (dofspf(f) .eq. 0 .or. f .eq. feb) cycle

            mom0 = mom0 + 0.5_dop*v2(f,f)*gpoint(f)*gpoint(f)
            mom1(f) = mom1(f) - v2(f,f)*gpoint(f)
            mom2(f,f) = 0.5d0*v2(f,f)
            do f1=f+1,ndof
               if (dofspf(f1) .eq. 0 .or. f1 .eq. feb) cycle
               mom0 = mom0 + v2(f,f1)*gpoint(f)*gpoint(f1)
               mom1(f) = mom1(f) - v2(f,f1)*gpoint(f1)
               mom1(f1) = mom1(f1) - v2(f,f1)*gpoint(f)
               mom2(f,f1)=0.5d0*v2(f,f1)
               mom2(f1,f)=mom2(f,f1)
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! get derivatives 
!-----------------------------------------------------------------------
      f = 0
      f1 = 0
      if (iord .eq. 0) then
         koelha = mom0

      else if (iord .eq. 1) then
         ilbl = index(label,'_')+1
         read(label(ilbl:lablen),*) f

         koelha = mom1(f)

      else if(iord .eq. 2) then
         ilbl = index(label,'_')+1
         jlbl = index(label(ilbl:lablen),'_') + ilbl-1
         read(label(ilbl:jlbl-1),*) f
         read(label(jlbl+1:lablen),*) f1

         koelha = mom2(f,f1)
         if (f .ne. f1) koelha = 2.0_dop*koelha

      endif
     
      return
      end subroutine klhacalc1

!#######################################################################
!
! CALCDIABDER
!
! calculates the potential derivatives
!
!#######################################################################

      subroutine calcdiabder(hops,derad,derdia,rotmatz,gpoint,nham)

      implicit none

      integer(long)                                      :: s,s1,f,m,n
      integer(long), intent(in)                          :: nham
      real(dop), dimension(maxdim), intent(in)           :: gpoint
      real(dop), dimension(maxsta,maxsta)                :: rotmat
      real(dop), dimension(maxsta,maxsta,maxdim), intent(out)   :: derad
      real(dop), dimension(maxsta,maxsta,maxdim), intent(out)   :: derdia
      complex(dop), dimension(maxsta,maxsta), intent(in) :: rotmatz
      real(dop), dimension(hopsdim), intent(in)           :: hops

      do s=1,nddstate
         do s1=1,nddstate
            rotmat(s1,s) = dble(rotmatz(s1,s))
         enddo
      enddo
      call calcpesder(hops,derdia,gpoint,nham)

      derad = 0.0_dop
      if (nddstate .eq. 1) then
         derad=derdia
      else
         do m=1,nmode
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               call simtranbd(derdia(1:nddstate,1:nddstate,f),&
                    rotmat(1:nddstate,1:nddstate),&
                    derad(1:nddstate,1:nddstate,f),nddstate)
            enddo
         enddo
      endif

      return
      end subroutine calcdiabder

!#######################################################################
!
! nham defines which operator is used
!
! Calculates the diabatic potential matrix derivatives at the point gpoint
!#######################################################################

      subroutine calcpesder(hops,derdia,gpoint,nham)

      implicit none
      
      logical(kind=4)                                     :: lkin
      integer(long), intent(in)                           :: nham
      integer(long)                                       :: m,f,k,h,k1,&
                                                             n,f1,s,s1
      integer(long)                                       :: dermethod,dertyp
      integer(long), dimension(maxdim)                    :: order
      real(dop)                                           :: koelha,deriv1,rfac
      real(dop), dimension(maxdim)                        :: p0
      real(dop), dimension(maxdim), intent(in)            :: gpoint
      real(dop), dimension(maxsta,maxsta,maxdim), intent(out)    :: derdia
      real(dop), dimension(hopsdim), intent(in)           :: hops
      logical(kind=4), dimension(maxkoe)                  :: kignore

      derdia(:,:,:) = 0.0_dop
      p0(:) = 0.0_dop

!-----------------------------------------------------------------------
!  calculate coefficients if from LHA
!-----------------------------------------------------------------------
      do k=1,khtot(nham)
         h=hamilton(fkoe,k)
         if (htmtype(h) .eq. 104) then
            call klhacalc1(h,koelha,gpoint,kignore(k))
            koeff(k) = koelha
         endif
      enddo

!-----------------------------------------------------------------------
! Uncorrelated terms.
! for single-set calculations nstate = 1 and nddstate is the no. of
! electronic states
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            if (m .eq. meb) cycle
            f1=spfdof(1,m)
            do k1=1,khzahl(m,s,nham)
               k=zhun(m,s,nham)+k1-1
               h=hamilton(f1,k)
! ignore any KE terms
               if (kinop(m,k) .ne. 0) cycle

               order = 0
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  order(f)=1
                  deriv1 = 0.0d0
                  dermethod=0
                  call dereval (h,deriv1,gpoint,order,p0,dertyp,dermethod)

                  deriv1 = dble(koeff(k))*deriv1

                  if (.not. lmult .and. nddstate .gt. 1) then
! single-set
                     do s1 = 1, nddstate
                        derdia(s1,s1,f)=derdia(s1,s1,f)+deriv1
                     enddo
                  else
! multi-set
                     derdia(s,s,f)=derdia(s,s,f)+deriv1
                  endif
                  order(f)=0
               enddo
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! loop over correlated Hamiltonian terms
!-----------------------------------------------------------------------
      do k=1,kzahl(nham)
! ignore any KE terms
         lkin = .false.
         do m=1,nmode
            lkin = lkin .or. kinop(m,k).ne.0
         enddo
         if (lkin) cycle 

!-----------------------------------------------------------------------
! outer loop is over DOF to take derivative wrt
! inner loop is over factors in Hamiltonian term
!-----------------------------------------------------------------------
         order = 0
         outer:do f1 = 1,ndof
            if (dofspf(f1) .eq. 0) cycle
            if (f1 .eq. feb) cycle

            order(f1) = 1
            deriv1=dble(koeff(k))
            inner:do m=1,nmode

               if (m .eq. meb) cycle inner
               if(lgausspot .and. m.gt.1)cycle

               f=spfdof(1,m)
               h=hamilton(f,k)

               dermethod=0
               rfac = 0.0_dop
               call dereval (h,rfac,gpoint,order,p0,dertyp,dermethod)

               deriv1 = deriv1*rfac
            enddo inner

            if (dertyp .eq. 1) then
               routine='Calcpesder'
               message='Kinetic energy term in potentials'
               call errormsg
            endif

            order(f1)=0
! sum up derivative in operator matrix
            if (lmult) then
               s1=ksi(k)
               s=ksf(k)
               derdia(s,s1,f1)=derdia(s,s1,f1)+deriv1
            else
               h=hamilton(feb,k)
               if (htmtype(h) .eq. 1 .or. htmtype(h) .eq. 2) then
                  call ssetder2(f1,derdia,deriv1,hops(zhop(h)))
               else
                  call ssetder(f1,derdia,deriv1,hops(zhop(h)))
               endif
            endif

         enddo outer

      enddo
!
! TODO: NEED TO ACCOUNT FOR MULD TERMS
!


      return
      end subroutine calcpesder

!#######################################################################
! Routines to generate LHA moments needed for mean-fields and calcha
! Used in mfields and calcha
!#######################################################################

      subroutine klhaconfig(h,np,psi,zetf1)

      implicit none

      integer(long), intent(in)                     :: h,zetf1,np
      integer(long)                                 :: s
      integer(long)                                 :: m,m1,f,f1,i
      integer(long)                                 :: dertyp,pottyp
      integer(long),dimension(maxdim)               :: order
      complex(dop)                                  :: vz
      real(dop),dimension(maxdim)                   :: ppoint
      complex(dop), dimension(zetf1:dgldim), intent(in) :: psi

      ! Energy, gradient and Hessian arrays
      real(dop)                                     :: v0
      real(dop), dimension(ndof)                    :: v1
      real(dop), dimension(ndof,ndof)               :: v2

!-----------------------------------------------------------------------
! calculate Hamiltonian moments for operator hterm(h)
!-----------------------------------------------------------------------
      do s=1,nstate
         v0=0.0_dop
         v1=0.0_dop
         v2=0.0_dop

         ! generate configurations
         gconfig=0.0_dop

         do m=1,nmode
            if(m.eq.meb)cycle
            call gengconfig1m(m,gconfig(1,1,s),ndof,nspfdof(m),vdimgp(m,s),&
                 dimgp(m,s),ndimgp(m,s),psi(zxi(m,s)))
         enddo

         ! add coordinates of point DVRs
         if (lpntdvr) then
            do f=1,ndof
               if (basis(f) .eq. 19) gconfig(f,:,s) = rpbaspar(1,f)
            enddo
         endif

         ! loop over all coordinate configurations
         ! nsgp(s) = block(s) for ms, no. GWP configs for ss.
         do i=1,nsgp(s)

! get GWP number
            num_gp = zgp(s) + i -1
            
            call poteval(h,v0,gconfig(1,i,s),vz,ppoint,pottyp)
        
            do f=1,ndof
               if (dofspf(f) .eq. 0 .or. basis(f) .eq. 19 .or. &
                      f .eq. feb) cycle   ! ignore ghost / frozen / feb modes
               order = 0
               order(f)=1
               call dereval(h,v1(f),gconfig(1,i,s),order,ppoint,dertyp,&
                    dermethod)

               gconfder1(f,i,s,np) = v1(f)   ! store derivatives
            enddo

            if (lhaorder .ge. 2) then
               do f=1,ndof
                  if (dofspf(f) .eq. 0 .or. basis(f) .eq. 19 .or. &
                         f .eq. feb) cycle   ! ignore ghost / frozen / feb modes
                  do f1=f,ndof
                     if (dofspf(f1) .eq. 0 .or. basis(f1) .eq. 19 .or. &
                            f1 .eq. feb) cycle   ! ignore ghost / frozen / feb modes
                     order = 0
                     order(f)=1
                     order(f1)=order(f1)+1
                     call dereval(h,v2(f,f1),gconfig(1,i,s),order,ppoint,&
                          dertyp,dermethod)
                  enddo
               enddo
            endif

            mom0(i,s,np) = v0
            mom1(:,i,s,np) = 0.0_dop
            mom2(:,:,i,s,np) = 0.0_dop
            do f=1,ndof
               m=dofspf(f)
               if (m .eq. 0 .or. m .eq. meb .or. m .eq. mkoe) cycle

               if (gwpintordm(m) .ge. 1) then
                  mom0(i,s,np) = mom0(i,s,np) - v1(f)*gconfig(f,i,s) 
                  mom1(f,i,s,np) = mom1(f,i,s,np) + v1(f) 
               endif

               if (gwpintordm(m) .ge. 2) then
                  mom0(i,s,np) = mom0(i,s,np) + &
                                 0.5_dop*v2(f,f)*gconfig(f,i,s)*gconfig(f,i,s)
                  mom1(f,i,s,np) = mom1(f,i,s,np) - v2(f,f)*gconfig(f,i,s)

                  mom2(f,f,i,s,np) = 0.5_dop*v2(f,f)

                  do f1=f+1,ndof
                     m1=dofspf(f1)
                     if (m1 .eq. 0 .or. f1 .eq. feb) cycle

                     if (gwpintordm(m1) .ge. 2) then
                        mom0(i,s,np) = mom0(i,s,np) + &
                               v2(f,f1)*gconfig(f,i,s)*gconfig(f1,i,s)
                        mom1(f,i,s,np) = mom1(f,i,s,np) - v2(f,f1)*gconfig(f1,i,s)
                        mom1(f1,i,s,np) = mom1(f1,i,s,np) - v2(f,f1)*gconfig(f,i,s)
                        mom2(f,f1,i,s,np)=0.5_dop*v2(f,f1)
                        mom2(f1,f,i,s,np)=mom2(f,f1,i,s,np)
                     endif
                  enddo
               endif
            enddo
         enddo
      enddo

      if (.not. lmult) then
         call klhaconfig_eb(np)
      endif

      return
      end subroutine klhaconfig

!#######################################################################
!
! Expands the mom arrays from GWP configurations to A-coeff space
!
!#######################################################################

      subroutine klhaconfig_eb(np)

      implicit none

      integer(long), intent(in) :: np
      integer(long)             :: i,e,n
      real(dop), dimension(maxgwp) :: mom0eb
      real(dop), dimension(ndof,maxgwp) :: mom1eb
      real(dop), dimension(ndof,ndof,maxgwp) :: mom2eb
      real(dop), dimension(ndof,maxgwp) :: der1eb

      do i=1,nsgp(1)
         mom0eb(i) = mom0(i,1,np) 
         mom1eb(:,i) = mom1(:,i,1,np) 
         mom2eb(:,:,i) = mom2(:,:,i,1,np) 
         der1eb(:,i) = gconfder1(:,i,1,np) 
      enddo

      i = 0
      do e=1,dim(meb,1)
         do n=1,vdim(meb,1)*ndim(meb,1)
            i = i+1
            mom0(i,1,np) = mom0eb(n) 
            mom1(:,i,1,np) = mom1eb(:,n) 
            mom2(:,:,i,1,np) = mom2eb(:,:,n) 
            gconfder1(:,i,1,np) = der1eb(:,n) 
         enddo
      enddo

      return
      end subroutine klhaconfig_eb

!#######################################################################

      subroutine klhacalc(koelha,s,h,np)

      implicit none

      integer(long), intent(in)                     :: h,s,np
      integer(long)                                 :: f,f1,hl,i
      integer(long)                                 :: ilbl,jlbl,lablen,iord
      character(len=c2)                             :: label
      real(dop), dimension(maxblock), intent(out)   :: koelha

!-----------------------------------------------------------------------
! get derivatives 
!-----------------------------------------------------------------------
      hl = hterm(1,h)
      label = hoplab(hl)
      lablen = index(label,' ')-1

      ilbl = index(label,'%')+1
      read(label(ilbl:ilbl),*) iord

      f = 0
      f1 = 0
      if (iord .eq. 0) then
         do i=1,block(s)
            koelha(i) = mom0(i,s,np)
         enddo

      else if (iord .eq. 1) then
         ilbl = index(label,'_')+1
         read(label(ilbl:lablen),*) f

         do i=1,block(s)
            koelha(i) = mom1(f,i,s,np)
         enddo

      else if(iord .eq. 2) then
         ilbl = index(label,'_')+1
         jlbl = index(label(ilbl:lablen),'_') + ilbl-1
         read(label(ilbl:jlbl-1),*) f
         read(label(jlbl+1:lablen),*) f1

         do i=1,block(s)
            koelha(i) = mom2(f,f1,i,s,np)
         enddo

         if (f .ne. f1) koelha = 2.0_dop*koelha

      endif
     
      return
      end subroutine klhacalc

!#######################################################################
! Generate coordinates at centre of GWP configurations

      subroutine gengconfig1m(m,gconfig1,ndof,gdof,vdim1,dim1,ndim1,xi)

      implicit none

      integer(long), intent(in)                   :: m,ndof,gdof, &
                                                     vdim1,dim1,ndim1
      real(dop), dimension(ndof,vdim1,dim1,ndim1), intent(out) :: gconfig1
      complex(dop), dimension(gdof,dim1), intent(in)           :: xi
      integer(long)                               :: n,e,v,g,f

      do n=1,ndim1
         do e=1,dim1
            do v=1,vdim1
               do g=1,gdof
                  f=spfdof(g,m)
                  gconfig1(f,v,e,n) = dble(xi(g,e))
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine gengconfig1m

!#######################################################################
! Generate momenta at centre of GWP configurations

      subroutine genpconfig1m(m,pconfig1,ndof,gdof,vdim1,dim1,ndim1,xi)

      implicit none

      integer(long), intent(in)                   :: m,ndof,gdof, &
                                                     vdim1,dim1,ndim1
      real(dop), dimension(ndof,vdim1,dim1,ndim1), intent(out) :: pconfig1
      complex(dop), dimension(gdof,dim1), intent(in)           :: xi
      integer(long)                               :: n,e,v,g,f

      do n=1,ndim1
         do e=1,dim1
            do v=1,vdim1
               do g=1,gdof
                  f=spfdof(g,m)
                  pconfig1(f,v,e,n) = dimag(xi(g,e))
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine genpconfig1m

!#######################################################################
!
! allocates arrays to hold multi-D LHA potentials
! nklha: no. of potentials to be treated
!
!#######################################################################

      subroutine alloc_klha(nklha)

      implicit none

      integer(long)     :: ierr
      integer(long), intent(in) :: nklha
      character(len=10) :: cerr
      integer(long), external   :: integersize, realsize, logicalsize

! check all arrays are not allocated
      call dealloc_klha

      cerr = 'mom0'
      allocate(mom0(maxblock,maxsta,nklha),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mom1'
      allocate(mom1(ndof,maxblock,maxsta,nklha),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mom2'
      allocate(mom2(ndof,ndof,maxblock,maxsta,nklha),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gconfig'
      allocate(gconfig(ndof,maxgwp,nstate),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'gconfder1'
      allocate(gconfder1(ndof,maxblock,nstate,nklha),stat=ierr)
      if (ierr > 0) go to 1000

      mom0=0.0_dop
      mom1=0.0_dop
      mom2=0.0_dop
      gconfig=0.0_dop
      gconfder1=0.0_dop

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_KLHA :',cerr
      write(ilog,'(a)') 'Error Code :',ierr
      stop
      end subroutine alloc_klha

!#######################################################################

      subroutine dealloc_klha

      if (allocated(mom0)) deallocate(mom0)
      if (allocated(mom1)) deallocate(mom1)
      if (allocated(mom2)) deallocate(mom2)
      if (allocated(gconfig)) deallocate(gconfig)
      if (allocated(gconfder1)) deallocate(gconfder1)

      return
      end subroutine dealloc_klha

! end of KLHA routines
!#######################################################################

      end module potevalmod

