
      module diagrsmod

use decimal, only:long,dop
use global
use pglobal
use channels

use op1lib, only: trqxd
use timing
use logdat, only: stopnow

      implicit none
      private
      public :: diagrs,wrdiagrsout
      contains

!***********************************************************************
!
!              POTFIT  Subroutine DIAGRS
!
! + calls LAPCK routine DSYEV to calculate all eigenvalues
!   and eigenvectors of a real symmetric matrix with
!   the eigenvalues in increasing order
! + eigenvectors are rearranged to correspond to eigenvalues in
!   decreasing order 
! + see 'diagch' for a similar subroutine for Hermitian matrices 
!
!   input: matrix  -- real symmetric matrix 
!          dim     -- order of real symm. matrix
!          neval   -- number of "relevant" eigenvectors
!          work    -- work array 
!
!   output: eval -- array of eigenvalues
!           evec -- array of eigenvectors
!           srnw -- sum of relevant natural weights
!
!  11/96 AJ
!  07/99 sw
!  12/17 MS
!
!***********************************************************************

      subroutine diagrs(eval,evec,matrix,dim,neval,srnw,&
           ioformat)

      implicit none

      integer(long)                                :: g,g1,g2,ier,neval1
      integer(long), intent(in)                    :: neval,ioformat,dim
      real(dop)                                    :: trace,x
      real(dop), intent(out)                       :: srnw
      real(dop), dimension(dim), intent(out)       :: eval
      real(dop), dimension(dim,dim), intent(out)   :: evec
      real(dop), dimension(dim,dim), intent(inout) :: matrix
      real(dop), dimension(dim+2,dim)              :: work
      integer(long), save                          :: tid=0 ! timer ID

      call get_timer(tid, 'Diagrs')
      call start_timer(tid)

!     return all eigenvectors
!     in mcpotfit

      if (lmcpf) then
         neval1 = dim
      else
         neval1 = neval
      end if

!-----------------------------------------------------------------------
! Get trace of real symmetric matrix 
!-----------------------------------------------------------------------
      call trqxd(matrix,trace,dim)
      if (ioformat.eq.0) then
         write (iout,'(a,1pe11.4,a,/a,f14.4,a)') &
         ' Trace of reduced density matrix : ',trace,' au', &
         ' Reduced trace :', trace*ev*ev/swsum2, ' eV**2'
      endif

!-----------------------------------------------------------------------
! Get all eigenvalues 'eval' and the corresponding normalized
! eigenvectors 'evec' of the real symmetric matrix 'matrix'.
! initially, the eigenvectors are returned via 'matrix'.
!-----------------------------------------------------------------------
      call dsyev('V','U',dim,matrix,dim,eval,work,(dim+2)*dim,ier)
      if (ier .ne. 0) then
        write(ilog,*) &
           'Error on diagonalising the Hamiltonian matrix in DIAGRS'
        write(ilog,*) 'Error code : ',ier
        call stopnow
      endif 

!-----------------------------------------------------------------------
! Change order of eigenvalues and eigenfunctions and choose phase
! such that the first component of each eigenvector is positive.
!-----------------------------------------------------------------------
      do g=1,dim/2
         x=eval(g)
         eval(g)=eval(dim+1-g)
         eval(dim+1-g)=x
      enddo

      do g1=1,neval1
         do g2=1,dim
            evec(g2,g1)=matrix(g2,dim+1-g1)
         enddo
      enddo

      do g1=1,neval1
         if (evec(1,g1).lt.0.0_dop) then
            do g2=1,dim
               evec(g2,g1)=-evec(g2,g1)
            enddo
         endif
      enddo

      call wrdiagrsout(ioformat,dim,eval,neval,srnw,trace)

      call stop_timer(tid)
      if (ioformat.eq.0) call myflush(iout)

      return
      end subroutine diagrs

!#######################################################################

      subroutine wrdiagrsout(ioformat,dim,eval,neval,srnw,trace)

      implicit none

      integer(long),intent(in) :: ioformat, dim, neval
      real(dop), dimension(dim), intent(in) ::  eval
      real(dop), dimension(dim)             ::  work
      real(dop), intent(out)                ::  srnw
      real(dop), intent(in)                 ::  trace
      real(dop)                             ::  rsum
      integer(long) :: e,i,g1,g2

!-----------------------------------------------------------------------
! Write eigenvalues.
!-----------------------------------------------------------------------
      if (ioformat.eq.0) then
         write (iout,'(a,i5,a,i5)')&
           ' Number of eigenvalues considered:', neval,'  /', dim
         write (iout,'(a)') &
         ' Reduced Eigenvalues (Natural Weights) [eV**2] :'
         do i = 1, min(dim,900), 6
            if(eval(i) .lt. 1.0e-10_dop ) goto 23 
            write (iout,'(i5,1x,6(e11.4e2,1x))') i,&
                  (eval(e)*ev*ev/swsum2,e=i,min(i+5,dim))
         enddo
 23      write (iout,*)
      endif

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
      do g1=1,dim
         work(g1)=0.0_dop
         do g2=g1+1,dim
            work(g1)=work(g1)+eval(g2)*ev*ev/swsum2
         enddo
      enddo
      if (ioformat.eq.0) then
         write (iout,*) &
         ' Trace - Sum of all preceding Natural Weights [eV**2] :'
         do i = 1, min(dim,900), 6
            if(work(i) .lt. 1.0e-10_dop ) goto 24 
            write (iout,'(i5,1x,6(e11.4e2,1x))') i,&
                  (work(e),e=i,min(i+5,dim))
         enddo
 24      write (iout,*)      
      endif
            
!-----------------------------------------------------------------------
! Get some measure of the relative accuracy of the "relevant"
! "eigenvectors"
!-----------------------------------------------------------------------
      rsum=0
      do e=1,neval
         rsum=rsum+eval(e)
      enddo
      srnw = abs(trace-rsum)*ev*ev/swsum2
      if (ioformat.eq.0 .or. ioformat.eq.1) then
         write (iout,'(a,1pe13.5e2)')&
         ' Abs(Trace - Sum of relevant Natural Weights) [eV**2] :',srnw
         if(srnw.ge.0.9994_dop) then
            write (iout,'(a,f10.4)')&
             ' Sqrt(Abs(Trace - Sum of rel. Nat. Weights))     [eV] :',&
               sqrt(srnw)
         else
            write (iout,'(a,f10.4)')&
             ' Sqrt(Abs(Trace - Sum of rel. Nat. Weights))    [meV] :',&
               sqrt(srnw)*1.0e3_dop
         endif 
         if (ioformat.eq.0) write (iout,*) 
      endif

      return
      end subroutine wrdiagrsout

      end module diagrsmod

