!***********************************************************************
      module pesfit

use decimal, only: long,dop,sip
use op1lib,  only: initvxi, cuttxd2
use xvlib,   only: mvtxdd,mvtxdd1

use global
use pglobal
use maxv, only:maxdim,maxsta,maxkoe
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use combined
use timing

use directutil, only:xevec2

      implicit none
      private
      public :: addtxvxdd,fittime,getvfit2,getvfit
      contains

!***********************************************************************
!
!     adds double precision vector b to the second dimension of 
!     double precision tensor a. pm specifies the sign of b.
!
!     a (i/o)
!     b (i)
!     pm (i)
!     
!***********************************************************************

      subroutine addtxvxdd(a,b,dim1,dim2,dim3,pm)

      implicit none

      integer(long)                                       :: i,j,k
      integer(long), intent(in)                           :: dim1,dim2,dim3,pm
      real(dop), dimension(dim1,dim2,dim3), intent(inout) :: a
      real(dop), dimension(dim2), intent(in)              :: b

      do k=1,dim3
         do j=1,dim2
            do i=1,dim1
               a(i,j,k)=a(i,j,k)+pm*b(j)
            enddo
         enddo
      enddo

      return
      end subroutine addtxvxdd

!***********************************************************************
!
!     Calculates the fit of the potential energy at the gridpoint
!     designated by the indices stored in igpoint. Does not include
!     1d potentials! This routine is called from showpot.
!
!     igpoint (i)
!        vector of ndof integers representing the indices of a grid point
!     coeff (i)
!        coefficient vector contracted over mode modc
!     evec (i)
!        array holding the eigenvectors of the potential density
!        matrices of all degrees of freedom
!     work2 (w)
!     vfit (o)
!        returns the fit
!     
!***********************************************************************

      subroutine getvfit(igpoint,coeff,evec,work2,vfit)


      implicit none

      integer(long), dimension(maxdim)            :: imgpoint
      integer(long), dimension(ndof), intent(in)  :: igpoint
      real(dop), dimension(coeffdim), intent(in)   :: coeff
      real(dop), intent(out)                      :: vfit
      real(dop), dimension(evecdim), intent(in)   :: evec
      real(dop), dimension(workdim2), intent(out) :: work2

      integer(long)       :: f,n,m,mm,offs_a,offs_b,offs_c

      do m=1,maxdim
         imgpoint(m)=0
      enddo
      do m=1,npmode
        imgpoint(m)=0
        do n=npotdof(m),1,-1
           f=potdof(n,m)
           imgpoint(m)=imgpoint(m)*gdim(f)+igpoint(f)-1
        enddo
        imgpoint(m)=imgpoint(m)+1
      enddo

!     first handle the contracted mode
      call cuttxd2(coeff,work2,dimbef,dimmodc,dimaft,imgpoint(modc))

      vfit=work2(1)

      offs_a=1
      offs_b=offs_a+c2v(1)
      offs_c=offs_b+c2v(2)
      mm=0

!     now handle the other modes
      do m=1,npmode
         if (m.ne.modc) then
            mm=mm+1
            call xevec2(imgpoint(m),evec(zevec(m)),psubdim(m),&
                 potdim(m),work2(offs_c))
            if (mod(mm,2).eq.1) then
               call mvtxdd(work2(offs_a),work2(offs_c),work2(offs_b),&
                    potdim(m),c2v(mm+1))
               vfit=work2(offs_b)
            else
               call mvtxdd(work2(offs_b),work2(offs_c),work2(offs_a),&
                    potdim(m),c2v(mm+1))
               vfit=work2(offs_a)
            endif
         endif
      enddo

      return
      end subroutine getvfit ! of getvfit



!***********************************************************************
!
!     Like getvfit, but the cut through the tensor is computed
!     outside, so we avoid computing it for every grid point.
!
!     igpoint (i)
!        vector of ndof integers representing the indices of a grid point
!     evec (i)
!        array holding the eigenvectors of the potential density
!        matrices of all degrees of freedom
!     cut (i)
!        array holding the cut through the tensor
!     work (w)
!     vfit (o)
!        returns the fit
!     
!***********************************************************************

      subroutine getvfit2(igpoint,evec,cut,work2,vfit)

      implicit none

      integer(long), dimension(maxdim)            :: imgpoint
      integer(long), dimension(ndof), intent(in)  :: igpoint
      real(dop), dimension(modcdim), intent(in)   :: cut
      real(dop), intent(out)                      :: vfit
      real(dop), dimension(evecdim), intent(in)   :: evec
      real(dop), dimension(workdim2), intent(out) :: work2

      integer(long)       :: f,n,m,mm,offs_a,offs_b,offs_c

      do m=1,maxdim
         imgpoint(m)=0
      enddo
      do m=1,npmode
        imgpoint(m)=0
        do n=npotdof(m),1,-1
           f=potdof(n,m)
           imgpoint(m)=imgpoint(m)*gdim(f)+igpoint(f)-1
        enddo
        imgpoint(m)=imgpoint(m)+1
      enddo

      vfit=cut(1)

      offs_a=1
      offs_b=offs_a+c2v(1)
      offs_c=offs_b+c2v(2)
      mm=0

      do m=1,npmode
         if (m.ne.modc) then
            mm=mm+1
            call xevec2(imgpoint(m),evec(zevec(m)),psubdim(m),&
                        potdim(m),work2(offs_c))
            if (mm.eq.1) then
               call mvtxdd(cut,work2(offs_c),work2(offs_b),&
                           potdim(m),c2v(mm+1))
               vfit=work2(offs_b)
            elseif (mod(mm,2).eq.1) then
               call mvtxdd(work2(offs_a),work2(offs_c),work2(offs_b),&
                           potdim(m),c2v(mm+1))
               vfit=work2(offs_b)
            else
               call mvtxdd(work2(offs_b),work2(offs_c),work2(offs_a),&
                           potdim(m),c2v(mm+1))
               vfit=work2(offs_a)
            endif
         endif
      enddo

      return
      end subroutine getvfit2 ! of getvfit2



!***********************************************************************
!
!     calculates the complete potential fit on all grid points.
!     this is done 'npest' times for timing purposes.
!
!     coeff (i)
!        contracted coefficients
!     evec (i)
!     work2 (w)
!     
!***********************************************************************

      subroutine fittime(coeff,evec,work2)

      implicit none

      real(dop), dimension(coeffdim), intent(in)  :: coeff
      real(dop), dimension(evecdim), intent(in)   :: evec
      real(dop), dimension(workdim2), intent(out) :: work2
      
      integer(long)                          :: f,i,n,m
      integer(long), dimension(maxdim)       :: igpoint
      real(dop)                              :: vfit
      integer(long), save                    :: tid=0 ! timer ID

      call get_timer(tid, 'Fit_time')
      call start_timer(tid)

      do i=1,npest
         call initvxi(igpoint,ndof)
 1       call getvfit(igpoint,coeff,evec,work2,vfit)
        do m=1,npmode
         do n=1,npotdof(m)
            f=potdof(n,m)
            if (igpoint(f).lt.gdim(f)) then
              igpoint(f)=igpoint(f)+1
              goto 1
             else
              igpoint(f)=1
             endif
         enddo
        enddo
      enddo 


      call stop_timer(tid)
      timers(tid)%ncalls=npest
      
      return
      end  subroutine fittime


      end module pesfit

