!
!                        POTFIT 
!
!  Program package to calculate natural potential fits.
!  Any similarities to the MCTDH package are NOT accidental.
!
!
!   11/96 AJ
!   1999 Stephan Wefing (several 'direct' versions added)
!
!
!***********************************************************************

      program potfit

use decimal, only:long,sip,dop
use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use runprojmod
use channels
use combined
use logdvrmod
use timing

use logdat
use peingutil, only: pclosedat
use pdefaultmod, only:pdefault,init_cmode
use logdvrmod, only:logdvr
use zeigdvrmod, only:zeigdvr
use rdopermod, only:operinfo
use zeigpfitmod, only:zeigpfit,mempfit
use memhopsmod, only:memhops

use mpidata
#ifdef MPI
   use mpi
#endif
      
      implicit none 

!-----------------------------------------------------------------------
! Define variables
!-----------------------------------------------------------------------
      real(dop), external  :: fcputime,fsystime
      real(dop)            :: tcpu,ttot
      logical(kind=4)      :: linperr

      integer(long)        :: t1,t2,tm,tm0

!-----------------------------------------------------------------------
! Get the time when the job is started
!-----------------------------------------------------------------------
      call unixzeit(tm0)

!-----------------------------------------------------------------------
! Define default values and initialize variables
!-----------------------------------------------------------------------
      call default
      call pdefault

!-----------------------------------------------------------------------
! Initialize MPI
!-----------------------------------------------------------------------
#ifdef MPI
      call mpi_init(mpierr)
      if (mpierr.ne.0) then
         routine = 'POTFIT (MPI_Init)'
         message = 'MPI initialization failed.'
         call errormsg
      endif
      call mpi_comm_size(MPI_COMM_WORLD,nmpi,mpierr)
      call mpi_comm_rank(MPI_COMM_WORLD,mpirank,mpierr)
      if (mpirank.gt.0) goto 5
#else
      nmpi=1
      mpirank=0
#endif

!-----------------------------------------------------------------------
! Open 'name/log' file; read input file.
!-----------------------------------------------------------------------
      linperr = .false.
      call peingabe(linperr)

!-----------------------------------------------------------------------
! MPI checks
!-----------------------------------------------------------------------
      if (lmpi) then
         if (.not.lwo) then
!        prevent wasting CPUs
            routine = "potfit"
            message = "usempi currently only works with write-only"
            call errormsg
         endif
      endif

!-----------------------------------------------------------------------
! pes timing makes only sense if timing is set
!-----------------------------------------------------------------------
      if (.not.ltime) npest=0
      
!-----------------------------------------------------------------------
! If dvr file exists, get dvr data required for rest of program.
! Else build dvr file
!-----------------------------------------------------------------------
      if (irundvr .eq. 1) then
         call rundvr
      else if (irundvr .eq. 2) then
         call logdvrdat
      endif

!-----------------------------------------------------------------------
! RUN POTFIT or PROJECTION
!-----------------------------------------------------------------------
#ifdef MPI
    5 continue
#endif
      if (lprojection) then
         call runproj
      else
         call runpfit
      endif
      if (mpirank.gt.0) goto 996

!-----------------------------------------------------------------------
! Get overall timing and write data to "log" file
!-----------------------------------------------------------------------
      call timout
      tcpu=fcputime(0)+fsystime(0)
      ttot = tcpu
      t1   = int(ttot)/3600
      ttot = ttot - 3600*t1
      t2   = int(ttot)/60
      ttot = ttot - 60*t2

      write(ilog,'(/,a,/)') '***** Overall Timing *****'
      write(iout,'(/,a,/)') '***** Overall Timing *****'
      write(ilog,'(a,1x,f12.2)') 'Total cpu time     [s] : ',tcpu
      write(ilog, 900) 'Total time     [h:m:s] : ',t1,t2,ttot
      write(iout, 900) 'Total time     [h:m:s] : ',t1,t2,ttot

      call unixzeit(tm)
      tm = tm-tm0
      t1 = tm/3600
      tm = tm - 3600*t1
      t2 = tm/60
      tm = tm - 60*t2
      write(ilog, 901) 'Wall  time     [h:m:s] : ',t1,t2,tm
      write(iout, 901) 'Wall  time     [h:m:s] : ',t1,t2,tm
      write(ilog,*)

      write(ilog,'(/,a,/)') '***** God''s final message *****'
      write(iout,'(/,a,/)') '***** God''s final message *****'
      if (.not. lstop) then
         write(ilog,'(a,/)') 'Program terminated normally.'
         write(iout,'(a,/)') 'Program terminated normally.'
      else
         write(ilog,'(a)') 'Program was stopped.'
         write(ilog,'(a)') 'The stop-command reads:'
         write(ilog,'(a,/)')  trim(stopcommand)
         write(iout,'(a)') 'Program was stopped.'
         write(iout,'(a)') 'The stop-command reads:'
         write(iout,'(a,/)')  trim(stopcommand)
      endif
      
!-----------------------------------------------------------------------
! Close remaining open files
!-----------------------------------------------------------------------
      call pclosedat

!-----------------------------------------------------------------------
! Shut down MPI.
!-----------------------------------------------------------------------
 996  continue
#ifdef MPI
      call mpi_finalize(mpierr)
#endif

  900 format( a,i4,' :',i3,' :',f5.1)
  901 format( a,i4,' :',i3,' :',i3)

      end program potfit

