!***********************************************************************
      module projcalc

use decimal, only:long,dop
use op1lib, only: initvxi

use maxv
use global
use pglobal
use dvrdatmod
use griddatmod
use runpfitmod
use channels
use logdat

      implicit none
      private
      public :: calcvproj,calcprojerr,calcprjfs,logprjerr
      contains

!***********************************************************************
!
! SUBROUTINE CALCPRJFS
!
! Calculate all projector functions, times DVR weight, for projector #p.
! In:  p        - number of projector
!      ort      - DVR grid points
!      switch   - .true.=normal, .false.=complementary projfunc.
! Out: prjf     - computed projector functions
!      zprjf    - zprjf(q) is the pointer to the projfunc. for the
!                 q-th projected DOF
!
! 07/2012 FO
!
!***********************************************************************

      subroutine calcprjfs(p,prjf,zprjf,ort,weights,switch)

      implicit none

      integer(long), intent(in)                     :: p
      integer(long), dimension(maxdim), intent(out) :: zprjf
      real(dop), dimension(maxprjfdim), intent(out) :: prjf
      real(dop), dimension(ortdim), intent(in)      :: ort, weights
      logical(kind=4), intent(in)                   :: switch
      integer(long)                                 :: zeig,q,f,i
      real(dop)                                     ::  w

!     compute pointers
      zeig = 1
      do q=1,nprojdof(p)
         zprjf(q) = zeig
         f = projdof(p,q)
         zeig = zeig + gdim(f)
      enddo

      do q=1,nprojdof(p)
         f = projdof(p,q)
!        calculate projector function
         call calcprjf(prjf(zprjf(q)),p,q,ort,switch)
         if (switch) then
!        multiply them, but...
         do i=1,gdim(f)
            w = weights(zort(f)+i-1)
!           ... if function values are read from a file they
!           might include the weights already
            if (projwi(p,q).eq.1) then
               w = 1.d0
            elseif (projwi(p,q).eq.2) then
               w = sqrt(w)
            endif
            prjf(zprjf(q)+i-1) = prjf(zprjf(q)+i-1)*w
         enddo
         endif
      enddo

      end subroutine calcprjfs

!***********************************************************************
!
! SUBROUTINE CALCVPROJ
!
! Calculates the projected potential for projector #p.
! In:  p        - number of projector
!      nrdof    - number of remaining DOFs
!      rdof     - array of remaining DOFs
!      vpot     - full potential
!      ort      - DVR grid points
!      weights  - DVR weights
!      work     - work space
! Out: vproj    - projected potential
!      vprojdim - size of vproj (number of points in reduced grid)
!
! 10/2005 FO
!
!***********************************************************************

      subroutine calcvproj(vproj,vprojdim,p,nrdof,rdof,&
                           vpot,ort,weights)


      implicit none

      real(dop), dimension(maxvprojdim), intent(out)       :: vproj
      real(dop), dimension(vpotdim), intent(in)            :: vpot
      real(dop), dimension(ortdim), intent(in)             :: ort
      real(dop), dimension(ortdim), intent(in)             :: weights
      real(dop), dimension(maxprjfdim)                     :: work
      integer(long), intent(in)                            :: p,nrdof
      integer(long), intent(out)                           :: vprojdim
      integer(long), dimension(maxdim), intent(in)         :: rdof

      integer(long), dimension(maxdim)       :: ipoint ! index vector of grid point (full grid)
      integer(long)                          :: ivproj         ! index into vproj
      integer(long)                          :: idvpot         ! index into vpot
      integer(long), dimension(maxdim)       :: zprjf  ! pointers for the projector functions
      integer(long)                          :: f,q,r
      real(dop)                              :: v,v1

! ----------------------------------------------------------------------
!     Calculate the values of the projector functions, times DVR weight
! ----------------------------------------------------------------------
      call calcprjfs(p,work,zprjf,ort,weights,.true.)

! ----------------------------------------------------------------------
!     Loop over all points of the remaining grid.
! ----------------------------------------------------------------------

!     start with first point
      call initvxi(ipoint,ndof)
      ivproj = 1
 10   continue

! ----------------------------------------------------------------------
!     Integration over the DOFs which are projected out
!     ("projected grid") starts here.
!     The integration method used here is a very simple sum rule,
!     which is accurate if the corresponding DOF grid is a DVR and
!     the projector function is "related" to this DVR. This can only
!     be ensured by the user...
! ----------------------------------------------------------------------

      v = 0.0_dop ! holds the integration result
 20   continue

!     convert ipoint (index vector) to idvpot (flat index)
      idvpot = 0
      do f=ndof,1,-1
         idvpot = idvpot*gdim(f)+ipoint(f)-1
      enddo
      idvpot = idvpot+1

!     multiply potential value with the projector functions
      v1 = vpot(idvpot)
      do q=1,nprojdof(p)
         f = projdof(p,q)
         v1 = v1 * work(zprjf(q)+ipoint(f)-1)
      enddo
!     add this to the integration result
      v = v + v1

!     go to next point in projected grid
      do q=1,nprojdof(p)
         f = projdof(p,q)
         if (ipoint(f).lt.gdim(f)) then
            ipoint(f) = ipoint(f)+1
            goto 20
         else
            ipoint(f) = 1
         endif
      enddo

!     apply scale
      v = v * projscale(p)

!     store final result
      vproj(ivproj) = v

! ----------------------------------------------------------------------
!     Integration finished for this point of the remaining grid,
!     go to next point.
! ----------------------------------------------------------------------
      ivproj = ivproj+1
      do r=1,nrdof
         f = rdof(r)
         if (ipoint(f).lt.gdim(f)) then
            ipoint(f) = ipoint(f)+1
            goto 10
         else
            ipoint(f) = 1
         endif
      enddo

! ----------------------------------------------------------------------
!     All done, store length of projected potential.
! ----------------------------------------------------------------------
      vprojdim = ivproj-1

      return
      end subroutine calcvproj ! of calcvproj

!***********************************************************************
!
! SUBROUTINE CALCPROJERR
!
! Calculates the projection error.
! Very similar to CALCVPROJ since the main part is an integration
! over the projected grid, inside a loop over the remaining grid.
! But the calculation inside is quite different, so this gets its
! own subroutine.
!
! IN:  vpot  - original potential on full grid
!      vproj - projected potentials
!      nrdof - number of remaining DOFs
!      rdof  - array of remaining DOFs
!      ort   - DVR grid points
!      weights - DVR weights
!      trafo - DVR trafo matrices
!      work  - work space
! Out: vperr - holds the result
!
! 11/2005 FO
! 07/2012 FO
!
!***********************************************************************

      subroutine calcprojerr(vperr,vpot,vproj,nrdof,rdof,&
                             ort,weights)


      implicit none

      real(dop), dimension(maxvprojdim), intent(out)        :: vperr
      real(dop), dimension(vpotdim), intent(in)             :: vpot
      real(dop), dimension(maxvprojdim,nproj), intent(in)   :: vproj
      real(dop), dimension(ortdim), intent(in)              :: ort,weights
      real(dop), dimension(maxprjfdim*(nproj+1))            :: work
      integer(long), intent(in)                             :: nrdof
      integer(long), dimension(maxdim), intent(in)          :: rdof

      integer(long), dimension(maxdim)       :: ipoint ! index vector of grid point (full grid)
      integer(long)                          :: ivproj          ! index into vproj
      integer(long)                          :: idvpot          ! index into vpot
      integer(long)                          :: z,z1,f,p,q,r
      integer(long), dimension(maxdim,nproj) :: zprjf
      real(dop)                              :: v,vp,v1,verr,verr1,meanverr,&
                                                minverr,maxverr,volume,dvol
      real(dop), dimension(maxdim)           :: minort,maxort
      integer(long), external                :: slen

!     work array is laid out as follows:
!     work(zprjf(q,p)) [p<=nproj] : q-th proj. function for proj. #p

!     N.B. here, nprojdof(p) and projdof(p,q) are independent of p!
!     Otherwise the definition of the error measure makes no sense.
!     This was already checked in PEINGABE.

! ----------------------------------------------------------------------
!     Calculate projector functions and DVR weights.
! ----------------------------------------------------------------------
      z = 1
      do p=1,nproj
         call calcprjfs(p,work(z),zprjf(1,p),ort,weights,.false.)
         z1 = z
         do q=1,nprojdof(p)
            zprjf(q,p) = zprjf(q,p)+z-1
            f = projdof(p,q)
            z1 = z1 + gdim(f)
         enddo
         z = z1
      enddo

! ----------------------------------------------------------------------
!     Loop over the remaining grid.
! ----------------------------------------------------------------------
      minverr  = 1.0e99_dop
      maxverr  = 0.0_dop
      meanverr = 0.0_dop

!     start with first point
      call initvxi(ipoint,ndof)
      ivproj = 1
 10   continue

! ----------------------------------------------------------------------
!     Integrate over the projected grid.
! ----------------------------------------------------------------------

      verr = 0.0_dop ! holds the integrated error measure
      volume = 0.0_dop
 20   continue

!     convert ipoint (index vector) to idvpot (flat index)
      idvpot = 0
      do f=ndof,1,-1
         idvpot = idvpot*gdim(f)+ipoint(f)-1
      enddo
      idvpot = idvpot+1

!     get value of full potential
      v = vpot(idvpot)

!     calculate sum of projected potentials times proj. function
      vp = 0
      do p=1,nproj
         v1 = vproj(ivproj,p)
         do q=1,nprojdof(p)
            f = projdof(p,q)
            v1 = v1*work(zprjf(q,p)+ipoint(f)-1)
         enddo
         vp = vp+v1
      enddo

!     error measure: difference squared
      verr1 = v-vp
      verr1 = verr1*verr1

!     multiply with DVR weights
      dvol = 1.0_dop
      do q=1,nprojdof(1)
         f = projdof(1,q)
         dvol = dvol*weights(zort(f)+ipoint(f)-1)
      enddo

!     add to integrated error measure
      verr = verr+verr1*dvol
      volume = volume+dvol

!     go to next point in projected grid
      do q=1,nprojdof(1)
         f = projdof(1,q)
         if (ipoint(f).lt.gdim(f)) then
            ipoint(f) = ipoint(f)+1
            goto 20
         else
            ipoint(f) = 1
         endif
      enddo

!     store final result
      verr = sqrt(verr/volume)
      vperr(ivproj) = verr

! ----------------------------------------------------------------------
!     Find mean/minimum/maximum
! ----------------------------------------------------------------------
      meanverr = meanverr+verr
      if (verr.lt.minverr) then
         minverr=verr
         do r=1,nrdof
            f = rdof(r)
            minort(r) = ort(zort(f)+ipoint(f)-1)
         enddo
      endif
      if (verr.gt.maxverr) then
         maxverr=verr
         do r=1,nrdof
            f = rdof(r)
            maxort(r) = ort(zort(f)+ipoint(f)-1)
         enddo
      endif

! ----------------------------------------------------------------------
!     Go to next point of remaining grid.
! ----------------------------------------------------------------------
      ivproj = ivproj+1
      do r=1,nrdof
         f = rdof(r)
         if (ipoint(f).lt.gdim(f)) then
            ipoint(f) = ipoint(f)+1
            goto 10
         else
            ipoint(f) = 1
         endif
      enddo

! ----------------------------------------------------------------------
!     Finished. Write info to logfile.
! ----------------------------------------------------------------------
      meanverr = meanverr/maxvprojdim
      call logprjerr(minverr,maxverr,meanverr,minort,maxort,nrdof,rdof)

      return
      end subroutine calcprojerr ! of wrprojerr

!***********************************************************************
!
! SUBROUTINE LOGPRJERR
!
! Logs some statistics about the projection error.
! Formerly part of CALCPROJERR.
!
!***********************************************************************

      subroutine logprjerr(minverr,maxverr,meanverr,minort,maxort,nrdof,rdof)

      implicit none

      real(dop), intent(in)                        :: minverr,maxverr,meanverr
      real(dop), dimension(maxdim), intent(in)     :: minort,maxort
      integer(long), intent(in)                    :: nrdof
      integer(long), dimension(maxdim), intent(in) :: rdof
      integer(long)                                :: r,f,sl,i

      integer, external :: slen

      write(ilog,'(/,76a1)') ('-',i=1,76)
      write(ilog,'(a)') ' Calculating projection error'
      write(ilog,'(76a1)')  ('-',i=1,76)

      write(ilog,'(a,2(e12.6,a))') &
         '   mean error = ',meanverr,' au = ',meanverr*ev,' ev'

      write(ilog,'(a,2(e12.6,a))') &
         '   min. error = ',minverr,' au = ',minverr*ev,' ev'
      write(ilog,'(a,$)') '   at ( '
      do r=1,nrdof
         f = rdof(r)
         sl = slen(modelabel(f))
         write(ilog,'(2a,f12.6,$)') modelabel(f)(1:sl),' = ',minort(r)
         if (r.lt.nrdof) write(ilog,'(a,$)') ' , '
      enddo
      write(ilog,'(a)') ' )'

      write(ilog,'(a,2(e12.6,a))') &
         '   max. error = ',maxverr,' au = ',maxverr*ev,' ev'
      write(ilog,'(a,$)') '   at ( '
      do r=1,nrdof
         f = rdof(r)
         sl = slen(modelabel(f))
         write(ilog,'(2a,f12.6,$)') modelabel(f)(1:sl),' = ',maxort(r)
         if (r.lt.nrdof) write(ilog,'(a,$)') ' , '
      enddo
      write(ilog,'(a)') ' )'

      return
      end subroutine logprjerr

!***********************************************************************
!
! SUBROUTINE CALCPRJF
!
! Calculates the values of the projector function for projector #p
! and its q-th projected DOF.
! Result is stored in array res.
! Array ort contains the DVR grid points.
! switch: .true. for "normal" projector function,
!         .false. for complementary projector function.
!         (integral over normal x complementary must be 1)
!
! 10/2005 FO
!
!***********************************************************************

      subroutine calcprjf(res,p,q,ort,switch)


      implicit none

      integer(long), intent(in)                  :: p,q
      real(dop), dimension(maxgdim), intent(out) :: res
      real(dop), dimension(ortdim), intent(in)   :: ort
      logical(kind=4), intent(in)                :: switch

      integer(long)           :: f,i,j,l,m
      real(dop)               :: x,y,fak,nf,x0,sigma
      character(len=c5)       :: filename
      logical(kind=4)         :: linit,lopen

      real(dop), external     :: plgndr ! external functions
      integer(long), external :: slen

      routine = 'calcprjf'

      f = projdof(p,q)
      linit = .false.
      nf = 1

!     loop over all DVR grid points
      do i=1,gdim(f)

!        calculate value of projector function at this grid point
         x = ort(zort(f)+i-1) ! coordinate value
!        jump to the required evaluation routine
         select case (projifunc(p,q))

        case default
!        still here? something went wrong!
         write(message,'(a,i4)')&
               'Unknown projector function index: ',&
               projifunc(p,q)
         call errormsg

        case (1)
! ----------------------------------------------------------------------
!        ifunc=1: associated Legendre polynomial of cos(x)
!        Normalization factor: (2l+1)/2 * (l-m)!/(l+m)!
! ----------------------------------------------------------------------
         l = projipar(p,q,1)
         m = abs(projipar(p,q,2))
         y = plgndr(l,m,cos(x))
         if (switch) then
            if (.not. linit) then
               fak = 1.0_dop
               do j=l-m+1,l+m
                  fak = fak*j
               enddo
               nf = 0.5_dop*dble(2*l+1)/fak
               linit = .true.
            endif
            y = nf*y
         endif

        case (2)
! ----------------------------------------------------------------------
!        ifunc=2: cos(mx)
!        Normalization factor: 1/2pi
! ----------------------------------------------------------------------
         m = projipar(p,q,1)
         y = cos(m*x)
         if (switch) then
            y = y / (2.0_dop*pi)
         else
            if (m.ne.0) y = 2.0_dop*y
         endif

        case (3)
! ----------------------------------------------------------------------
!        ifunc=3: Gaussian, exp(-1/4 (x-x0/sigma)**2)
!        Normalization: 1/(sigma*sqrt(2*pi))
! ----------------------------------------------------------------------
         x0 = projrpar(p,q,1)
         sigma = projrpar(p,q,2)
         y = (x-x0)/sigma
         y = exp(-0.25_dop*y*y)
         if (switch) then
            y = y/(sigma*sqrt(2.0_dop*pi))
         endif

        case (4)
! ----------------------------------------------------------------------
!        ifunc=4: read values from ASCII file
! ----------------------------------------------------------------------
         if (.not. linit) then
            filename = projcpar(p,q,1)
            l = slen(filename)
            call abspath(filename,l)
            open(idump,file=filename,form='formatted',status='old',&
                 err=1000)
            linit = .true.
         endif
         read (idump,*,err=1010,end=1020) y

        case (5)
! ----------------------------------------------------------------------
!        ifunc=5: sin(mx)
!        Normalization factor: 1/2pi
! ----------------------------------------------------------------------
         m = projipar(p,q,1)
         y = sin(m*x)
         if (switch) then
            y = y / (2.d0*pi)
         else
            y = 2.d0*y
         endif

       end select

! ----------------------------------------------------------------------
! Code for new functions goes before here!
! ----------------------------------------------------------------------

! ----------------------------------------------------------------------
!        store result
! ----------------------------------------------------------------------
         res(i) = y

      enddo

      inquire(unit=idump, opened=lopen)
      if (lopen) close(idump)

      return

 1000 l=slen(filename)
      message='Cannot open file: '//filename(1:l)
      call errormsg
 1010 l=slen(filename)
      message='Error while reading file: '//filename(1:l)
      call errormsg
 1020 l=slen(filename)
      message='File contains too few data: '//filename(1:l)
      call errormsg

      end subroutine calcprjf ! of calcprjf


      end module projcalc


