!***********************************************************************
      module projutil

use versions, only: progver
use decimal, only:long,dop,vlong
use globmemmod
use logdat
use timing

      implicit none
      private
      public :: zeigproj,projstupid,projrun,memproj
      contains

!***********************************************************************
!
! SUBROUTINE ZEIGPROJ
!
! Calculates memory consumption and pointer structures for
! projection runs in POTFIT.
!
! 10/2005 FO
!
!***********************************************************************

      subroutine zeigproj

use global
use pglobal
use dvrdatmod
use griddatmod
use runpfitmod
use mpidata
#ifdef MPI
   use mpi
#endif

      implicit none

      integer(long)  :: f,p,q,zeig,workrdim1,itmp1,innerdim1
      integer(vlong) :: itmp8,vpotdim8,innerdim8,TWO
#ifdef MPI
      integer(long), dimension(6) :: mpicommi
#endif


      TWO = 2

      if (mpirank.eq.0) then
!     calculate size of full grid
         vpotdim8= 1
         zeig    = 1
         do f=1,ndof
            vpotdim8= vpotdim8*gdim(f)
            zeig    = zeig+gdim(f)
         enddo
         oneddim = zeig-1

!     calculate maximum size of remaining/projected grid
         maxvprojdim = 1
         maxinnerdim = 1
         do p=1,nproj
            itmp8 = vpotdim8
            innerdim8 = 1
            do q=1,nprojdof(p)
               f=projdof(p,q)
               itmp8 = itmp8/gdim(f)
               innerdim8 = innerdim8*gdim(f)
            enddo
            if (itmp8 .ge. TWO**31) then
               routine = 'zeigproj'
               message = 'remaining grid too large (integer overflow)'
               call errormsg
            endif
            itmp1 = itmp8
            if (innerdim8 .ge. TWO**31) then
               routine = 'zeigproj'
               message = 'projected grid too large (integer overflow)'
               call errormsg
            endif
            innerdim1 = innerdim8
            maxvprojdim = max(maxvprojdim,itmp1)
            maxinnerdim = max(maxinnerdim,innerdim1)
         enddo

!     set storage size for original potential
         if (lprojlowmem) then
            vpotdim = maxinnerdim
         else
            if (vpotdim8 .ge. TWO**31) then
               routine = 'zeigproj'
               message = 'full grid too large (integer overflow)'
               call errormsg
            else
               vpotdim = vpotdim8
            endif
         endif

!     calculate size of buffer, niobuf
         if(vpotdim .le. 16*1024*1024) then
            niobuf = vpotdim
         else
            niobuf = 4*1024*1024
         endif

!     calculate size of work array

!     1) for calculating the projected potentials
!     we need it to store the values of the projector functions.
!     so the needed size is the sum of the grid lengths of those DOFs
!     which are projected out (maximised over all projectors).
      maxprjfdim = 0
      totprjfdim = 0
      do p=1,nproj
         itmp1 = 0
         do q=1,nprojdof(p)
            itmp1 = itmp1+gdim(projdof(p,q))
         enddo
         maxprjfdim = max(maxprjfdim,itmp1)
         totprjfdim = totprjfdim + itmp1
      enddo
      workrdim = max(workrdim,maxprjfdim)

!     2) for calculating the projection error
!     here we need to store the projector functions for all projectors
         if (lprojerr) then
            workrdim1 = nproj*maxprjfdim
            workrdim = max(workrdim,workrdim1)
         endif

!     3) In low-memory mode, we must store all the projector functions
!     in memory. For the projection error, additionally all the
!     complementary projector functions are needed
         if (lprojlowmem) then
            workrdim1 = totprjfdim
            if (lprojerr) then
               workrdim1 = workrdim1 + totprjfdim
            endif
            workrdim = max(workrdim,workrdim1)
         endif

         maxnodedim = maxvprojdim/nmpi
         if (mod(maxvprojdim,nmpi).ne.0) maxnodedim=maxnodedim+1

      endif ! mpi-master, or not mpi

#ifdef MPI
!     transmit data to nodes
      if (mpirank.eq.0) then
         mpicommi(1)=vpotdim
         mpicommi(2)=maxvprojdim
         mpicommi(3)=maxnodedim
         mpicommi(4)=maxprjfdim
         mpicommi(5)=totprjfdim
         mpicommi(6)=workrdim
         call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
      else
         call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         vpotdim    =mpicommi(1)
         maxvprojdim=mpicommi(2)
         maxnodedim =mpicommi(3)
         maxprjfdim =mpicommi(4)
         totprjfdim =mpicommi(5)
         workrdim   =mpicommi(6)
      endif
#endif


      end subroutine zeigproj ! of zeigproj





!***********************************************************************
!
! SUBROUTINE MEMPROJ
!
! Determines offsets into the different global arrays (mc,mr,...)
! used by POTFIT for projection runs.
!
! 10/2005 FO
!
!***********************************************************************

      subroutine memproj

use global
use pglobal
use dvrdatmod
use griddatmod
use runpfitmod
use mpidata
#ifdef MPI
   use mpi
#endif

      implicit none

      mrvvec = vpotdim  ! Note, mrvvec is not a pointer but a dimension
      mrdim   = 1
      mrort   = mrdim
      mrdim   = mrort+ortdim
      mroned  = mrdim
      mrdim   = mroned+ortdim
      mrtrafo = mrdim
      mrdim   = mrtrafo+dvrdim
      mrbuf   = mrdim
      if (mpirank.eq.0) then
         mrdim = mrdim+niobuf
      endif
      mrvproj = mrdim
      if (lprojerr.or.lprojlowmem) then
!        for calculating the projection error, and in low-memory mode,
!        we need to keep all projected potentials in memory
         mrdim   = mrdim+nproj*maxnodedim
         mrvperr = mrdim
!        and we also need to store the projection error itself
         if (lprojerr) then
            mrdim = mrdim+maxnodedim
         endif
      else
!        otherwise we can re-use the memory
         mrdim   = mrdim+maxvprojdim
         mrvperr = mrdim ! not needed
      endif
      mrvbuf = mrdim
      if (mpirank.eq.0 .and. lmpi) then
!        master needs to store one full projected potential
         mrdim = mrdim+maxvprojdim
      endif

      mrworkr = mrdim
      mrdim   = mrdim+workrdim


      end subroutine memproj ! of memproj

!***********************************************************************
!
! SUBROUTINE PROJSTUPID
!
! Sets up some variables which would be initialised in PEINBAS.
! However, for projection runs, the NATPOT-BASIS-SECTION is ignored,
! and can even be omitted. So to be on the safe side, we initialise
! these variables here, although we don't actually need them.
!
! 10/2005 FO
!
!***********************************************************************

      subroutine projstupid

use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use combined

      implicit none

      integer(long)       :: f

!     In projection runs, we only deal with the primitive grid.
!     Mode combination does not matter. Thus we simply assume that no
!     DOFs are combined. So we just need to map DOFs to modes 1:1 .
      npmode = ndof
      lpconm = .false.
      do f=1,ndof
         npotdof(f) = 1
         potdof(1,f)= f
         dofpot(f)  = f
         psubdim(f) = gdim(f)
      enddo

!     Information about contracted modes/natpots:
!     Really unnecessary. But to fool certain parts of the code that
!     rely on this information, we just pretend that the first mode
!     is contracted, and that each mode has only one natpot.
      modc = 1
      do f=1,ndof
         potdim(f) = 1
      enddo
      
      end subroutine projstupid ! of projstupid


!***********************************************************************
!
! SUBROUTINE PROJRUN
!
! Loops over all projectors, calculates the corresponding projected
! potentials and writes them to vpot files.
!
! In:  vpot  - potential on full grid
!      ort   - ort array which contains the grid points
!      trafo - DVR trafo matrices (needed for weights)
!      vproj - array where projected potential gets stored
!      work  - real*8 work space
!
! 10/2005 FO
!
!***********************************************************************

      subroutine projrun(vpot,ort,trafo,vproj,vperr)

use global
use pglobal
use maxv, only:maxdim
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use channels

use projcalc, only:calcprojerr,calcvproj

      implicit none

      real(dop), dimension(vpotdim), intent(in)  :: vpot
      real(dop), dimension(ortdim), intent(in)   :: ort
      real(dop), dimension(dvrdim), intent(in)   :: trafo
      real(dop), dimension(*), intent(out)       :: vproj
      real(dop), dimension(*), intent(out)       :: vperr

      integer(long)                          :: p,q,i,f,&
                                                vprojdim,i1,nrdof
      integer(long), dimension(maxdim)       :: rdof
      real(dop), dimension(ortdim)           :: weights
      integer(long), external                :: slen
      integer(long), save                    :: tid=0 ! timer ID

      call get_timer(tid, 'Calcvproj')
      call start_timer(tid)

      routine = 'projrun'

! ----------------------------------------------------------------------
!     Projected Potentials
! ----------------------------------------------------------------------

!     write some info to the logfile
      write(ilog,'(/,76a1)') ('-',i=1,76)
      write(ilog,'(a)') ' Calculating projected potentials'
      write(ilog,'(76a1)')  ('-',i=1,76)

!     compute DVR weights
      do f=1,ndof
         call dvrweights(trafo(zdvr(f)),ort(zort(f)),gdim(f),gdim(f), &
              weights(zort(f)),basis(f),rpbaspar(1,f), &
              ipbaspar(1,f),ipbaspar(1,f+1),.true.,.false.)
      enddo


!     loop over all projectors
      do p=1,nproj
!        construct the array rdof(1..nrdof) which holds the
!        DVR DOF numbers of the *remaining* degrees of freedom
         nrdof = 0 ! number of remaining DOFs
         do f=1,ndof
!           check if DOF f is projected away
            do q=1,nprojdof(p)
               if (projdof(p,q).eq.f) goto 10
            enddo
!           DOF not projected: store it as a remaining DOF
            nrdof = nrdof + 1
            rdof(nrdof) = f
 10         continue
         enddo

!        calculate projected potential
         if (lprojerr) then
            i1=(p-1)*maxvprojdim+1
         else
            i1=1
         endif
         call calcvproj(vproj(i1),vprojdim,p,nrdof,rdof,vpot,&
                        ort,weights)

!        write vpot file containing projected potential, and logs
         call outproj(vproj(i1),vprojdim,p,nrdof,rdof)

      enddo ! end loop over projectors

! ----------------------------------------------------------------------
!     Projection Error
! ----------------------------------------------------------------------

      if (lprojerr) then

!        calculate the error
         call calcprojerr(vperr,vpot,vproj,nrdof,rdof,ort,weights)

!        write it to a vpot file, and logs
         call outprjerr(vperr,vprojdim,nrdof,rdof)

      endif

      call stop_timer(tid) ! stop timer

      end subroutine projrun ! of projrun


!***********************************************************************
!
! SUBROUTINE OUTPROJ
!
! Write projected potential to vpot and ASCII files. Also some logs.
! Formerly part of PROJRUN.
!
!***********************************************************************

      subroutine outproj(vproj,vprojdim,p,nrdof,rdof)

use maxv
use global
use pglobal
use dvrdatmod
use griddatmod
use channels

      implicit none

      integer(long), intent(in)                 :: vprojdim,p,nrdof
      integer(long),dimension(maxdim), intent(in) :: rdof
      integer(long)                             :: lablen,sl,ierr,i,q
      character(len=c5)                         :: vprojname,srfname
      real(dop)                                 :: vpmin,vpmax
      real(dop),dimension(vprojdim), intent(in) :: vproj
      character(len=c2)                          :: wrtag

      integer slen

!     write some more info to the logfile
      lablen=index(projlab(p),' ')-1
      write(ilog,'(i3,2a)') p,' -- Label: ',projlab(p)(1:lablen)
      write(ilog,'(a,i3,a,$)') '   ',nprojdof(p),' DOFs projected: '
      do q=1,nprojdof(p)
         sl=index(modelabel(projdof(p,q)),' ')-1
         write(ilog,'(2a,$)') modelabel(projdof(p,q))(1:sl),' '
      enddo
      write(ilog,'(x)')

!     open vpot_LABEL file for writing
      vprojname=name(1:laenge)//'/vpot_'//projlab(p)(1:lablen)
      sl=slen(vprojname)
      open(unit   = idump, &
           file   = vprojname(1:sl), &
           form   = 'unformatted', &
           status = 'unknown', &
           iostat = ierr)
      if (ierr.ne.0) then
         routine = 'outproj'
         message = 'could not create file "'//vprojname(1:sl)//'"'
         call errormsg
      endif

!     write vpot file header
      call wrvpheader(idump,vprojname,p,nrdof,rdof)

!     now write the projected potential
      wrtag= '<vpot pes>'
      write(idump) wrtag
      call wrvproj(idump,vproj,vprojdim,niobuf)
      wrtag= '<end vpot pes>'
      write(idump) wrtag

!     also write an ASCII file if required
      if (lprojascii) then
         srfname=name(1:laenge)//'/srf_'//projlab(p)(1:lablen)//'.asc'
         sl=slen(srfname)
         open(unit   = iaus, &
              file   = srfname(1:sl), &
              form   = 'formatted', &
              status = 'unknown', &
              iostat = ierr)
         if (ierr.ne.0) then
            routine = 'outproj'
            message = 'could not create file "'//srfname(1:sl)//'"'
            call errormsg
         endif
         do i=1,vprojdim
            write (iaus,*) vproj(i)
         enddo
         close(iaus)
      endif

!     for informational purposes, we calculate the minimum
!     and maximum values of the projected potential
      vpmin=vproj(1)
      vpmax=vproj(1)
      do i=1,vprojdim
         if (vproj(i).lt.vpmin) vpmin=vproj(i)
         if (vproj(i).gt.vpmax) vpmax=vproj(i)
      enddo
!     write this to the log file
      write(ilog,'(a,2(e12.6,a))') '     Vmin = ',vpmin,' au = ',vpmin*ev,' ev'
      write(ilog,'(a,2(e12.6,a))') '     Vmax = ',vpmax,' au = ',vpmax*ev,' ev'
      write(ilog,'(5x,a,i10,/)') 'vprojdim =',vprojdim

!     close vpot_LABEL file
      close(idump)

      end subroutine outproj

!***********************************************************************
!
! SUBROUTINE OUTPRJERR
!
! Write projection error to vpot file. Also some logs.
! Formerly part of PROJRUN.
!
!***********************************************************************

      subroutine outprjerr(vperr,vprojdim,nrdof,rdof)

use global
use pglobal
use maxv, only: maxdim
use channels

      implicit none

      integer(long), intent(in)                    :: vprojdim,nrdof
      integer(long), dimension(maxdim), intent(in) :: rdof
      integer(long)                                :: lablen,sl,ierr
      real(dop), dimension(vprojdim), intent(in)   :: vperr
      character(len=c5)                            :: vprojname

      integer slen

!     open projerr file for writing
      lablen=slen(projerrfile)
      vprojname=name(1:laenge)//'/'//projerrfile(1:lablen)
       sl=slen(vprojname)
      open(unit   = idump, &
           file   = vprojname(1:sl), &
           form   = 'unformatted', &
           status = 'unknown', &
           iostat = ierr)
      if (ierr.ne.0) then
         routine = 'outprjerr'
         message = 'could not create file "'//vprojname(1:sl)//'"'
         call errormsg
      endif

!     The projerr file is another vpot file. First we write the header.
!     Here we just use the header data from the last projected potential,
!     since the (projected/remaining) grids are the same for all
!     projectors (checked in PEINGABE).
      call wrvpheader(idump,vprojname,-1,nrdof,rdof)

!     write error data
      call wrvproj(idump,vperr,vprojdim,niobuf)

!     close projerr file
      close(idump)

      end subroutine outprjerr

!***********************************************************************
!
! SUBROUTINE PROJRUN2
!
! Low-memory version of PROJRUN. Also contains parts of WRITE_VPOT and
! CALCPROJERR (and probably others).
!
! Here, we don't need to store the full-grid original potential. Instead,
! only the part needed to compute one point of the remaining grid is
! computed and kept in memory.
!
! Arg: vpot  - potential on partial grid (is computed here)
!      ort   - ort array which contains the grid points
!      trafo - DVR trafo matrices (needed for weights)
!      vproj - array where projected potential gets stored
!      vperr - array where projection error gets stored
!      prjf  - array for storing the projector functions
!      cprjf - array for storing the complementary projfunc.s
!
! 07/2012 FO
!
!***********************************************************************

      subroutine projrun2(vpot,ort,trafo,vproj,vperr,vbuf,prjf,cprjf)

use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use channels
use mpidata
#ifdef MPI
   use mpi
#endif

use projcalc, only:calcprojerr,calcvproj
use op1lib, only: initvxi
use directutil, only: index2coord
use projcalc, only: calcprjfs,logprjerr
use getpesmod, only: vcut
use callmuldmod, only: callmuld

      implicit none

      real(dop), dimension(*), intent(out)       :: vpot
      real(dop), dimension(ortdim), intent(in)   :: ort
      real(dop), dimension(dvrdim), intent(in)   :: trafo
      real(dop), dimension(maxnodedim,nproj), intent(out) :: vproj
      real(dop), dimension(maxnodedim), intent(out)       :: vperr
      real(dop), dimension(maxvprojdim), intent(out)      :: vbuf
      real(dop), dimension(totprjfdim), intent(out) :: prjf,cprjf

      integer(long) :: p,q,i,j,jj,f,r,z,z1
      integer(long) :: vprojdim,nodedim,nrdof,idum
      integer(long), dimension(maxdim) :: rdof, gpoint
      integer(long), dimension(maxdim,nproj) :: zprjf
      logical(kind=4), dimension(maxdim) :: isproj
      real(dop) :: v,v1
      real(dop), dimension(maxdim) :: gcoord,minort,maxort
      real(dop), dimension(ortdim) :: weights
      real(dop)                    :: vp,verr,verr1,meanverr,minverr,&
                                      maxverr,volume,dvol
      integer(long), save          :: tid=0 ! timer ID
#ifdef MPI
      integer(long) :: node
      integer(long), dimension(MPI_STATUS_SIZE) :: mpistat
      real(dop), dimension(2*maxdim+3)          :: mpicommr
#endif

      call get_timer(tid, 'Calcvproj2')
      call start_timer(tid)

      routine = 'projrun2'

!     write some info to the logfile
      if (mpirank.eq.0) then
         write(ilog,'(/,76a1)') ('-',i=1,76)
         write(ilog,'(a)') ' Calculating projected potentials'
         write(ilog,'(a)') ' using low-memory strategy'
         write(ilog,'(76a1)')  ('-',i=1,76)
      endif

!     construct the array rdof(1..nrdof) which holds the
!     DVR DOF numbers of the *remaining* degrees of freedom,
!     and isproj(1..ndof) telling whether a DOF is projected
      nrdof = 0 ! number of remaining DOFs
      do f=1,ndof
         isproj(f) = .true.
!        check if DOF f is projected away
         do q=1,nprojdof(1)
            if (projdof(1,q).eq.f) goto 10
         enddo
!        DOF not projected: store it as a remaining DOF
         nrdof = nrdof + 1
         rdof(nrdof) = f
         isproj(f) = .false.
 10      continue
      enddo

      if (mpirank.eq.0) then
!     compute weights
      do f=1,ndof
         call dvrweights(trafo(zdvr(f)),ort(zort(f)),gdim(f),gdim(f), &
              weights(zort(f)),basis(f),rpbaspar(1,f), &
              ipbaspar(1,f),ipbaspar(1,f+1),.true.,.false.)
      enddo
      endif ! mpi-master (or non-mpi)
#ifdef MPI
!     transmit ort/weights arrays to slaves
      call MPI_Bcast(ort,ortdim,MPI_DOUBLE_PRECISION,&
           0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(weights,ortdim,MPI_DOUBLE_PRECISION,&
           0,MPI_COMM_WORLD,mpierr)
#endif

!     compute projector functions (times weight) and pointers
      if (mpirank.eq.0) then
      z = 1
      do p=1,nproj
         call calcprjfs(p,prjf(z),zprjf(1,p),ort,weights,.true.)
         if (lprojerr) then
            call calcprjfs(p,cprjf(z),zprjf(1,p),ort,weights,.false.)
         endif
         z1 = z
         do q=1,nprojdof(p)
            zprjf(q,p) = zprjf(q,p)+z-1
            f = projdof(p,q)
            z1 = z1 + gdim(f)
         enddo
         z = z1
      enddo
      endif ! mpi-master
#ifdef MPI
!     transmit prjf/cprjf/zprjf to slaves
      call MPI_Bcast(zprjf,maxdim*nproj,MPI_INTEGER, &
           0,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(prjf,totprjfdim,MPI_DOUBLE_PRECISION, &
           0,MPI_COMM_WORLD,mpierr)
      if (lprojerr) then
         call MPI_Bcast(cprjf,totprjfdim,MPI_DOUBLE_PRECISION, &
              0,MPI_COMM_WORLD,mpierr)
      endif
#endif

!     start loop over remaining grid
      minverr  = 1.d99
      maxverr  = 0.d0
      meanverr = 0.d0
      call initvxi(gpoint,ndof)
      j = 1
      jj= 1

! ======================================================================
! Parallelisation is done over the remaining (outer grid)
! j  -> outer grid point (flat index)
! jj -> this node's outer grid point
! j = (jj-1)*nmpi + mpirank + 1
! ======================================================================

 20   continue
#ifdef MPI
      if (mod(j-1,nmpi).ne.mpirank) goto 21
#endif

! ----------------------------------------------------------------------
!     Compute original potential (partially)
! ----------------------------------------------------------------------

      i = 1
 30   continue
!     compute potential at gridpoint
      call index2coord(gpoint,ort,gcoord)
      call callmuld(v,gcoord,ndof,ippes,hopfile(1),hopipar(1,1),hoprpar(1,1),&
           idum,idum,hoplab(1))
      vpot(i) = v
!     advance to next point on projected grid
      i = i+1
      do f=1,ndof
         if (isproj(f)) then
            if (gpoint(f).lt.gdim(f)) then
               gpoint(f) = gpoint(f)+1
               goto 30
            else
               gpoint(f) = 1
            endif
         endif
      enddo
!     handle potential cut
      call vcut(vpot,vpotdim)

! TODO: write vpot? we don't have correct order here...

! ----------------------------------------------------------------------
!     Compute projected potentials
! ----------------------------------------------------------------------

!     loop over all projectors
      do p=1,nproj
!        perform simple integration over projected grid
         i = 1
         v = 0.d0 ! holds the integration result
 31      continue
!        multiply potential value with the projector functions
         v1 = vpot(i)
         do q=1,nprojdof(p)
            f = projdof(p,q)
            v1 = v1*prjf(zprjf(q,p)+gpoint(f)-1)
         enddo
!        add this to the integration result
         v = v + v1
!        advance to next point on projected grid
         i = i+1
         do f=1,ndof
            if (isproj(f)) then
               if (gpoint(f).lt.gdim(f)) then
                  gpoint(f) = gpoint(f)+1
                  goto 31
               else
                  gpoint(f) = 1
               endif
            endif
         enddo
!        apply scale
         v = v * projscale(p)
!        integration finished, store result
         vproj(jj,p) = v
      enddo

! ----------------------------------------------------------------------
!     Compute projection error
! ----------------------------------------------------------------------
      if (lprojerr) then
!        integrate over projected grid
         i = 1
         verr = 0.d0 ! holds the integrated error measure
         volume = 0.d0
 32      continue
         v = vpot(i) ! value of full potential
!        calculate sum of projected potentials times compl.proj.func.
         vp = 0.d0
         do p=1,nproj
            v1 = vproj(jj,p)
            do q=1,nprojdof(p)
               f = projdof(p,q)
               v1 = v1*cprjf(zprjf(q,p)+gpoint(f)-1)
            enddo
            vp = vp+v1
         enddo
!        error measure: difference squared
         verr1 = v-vp
         verr1 = verr1*verr1
!        calculate volume element
         dvol = 1.d0
         do q=1,nprojdof(1)
            f = projdof(1,q)
            dvol = dvol*weights(zort(f)+gpoint(f)-1)
         enddo
!        add to integrated error measure
         verr = verr + verr1*dvol
         volume = volume + dvol
!        advance to next point on projected grid
         i = i+1
         do f=1,ndof
            if (isproj(f)) then
               if (gpoint(f).lt.gdim(f)) then
                  gpoint(f) = gpoint(f)+1
                  goto 32
               else
                  gpoint(f) = 1
               endif
            endif
         enddo
!        store final result
         verr = sqrt(verr/volume)
         vperr(jj) = verr
!        find mean/minimum/maximum
         meanverr = meanverr+verr
         if (verr.lt.minverr) then
            minverr=verr
            do r=1,nrdof
               f = rdof(r)
               minort(r) = ort(zort(f)+gpoint(f)-1)
            enddo
         endif
         if (verr.gt.maxverr) then
            maxverr=verr
            do r=1,nrdof
               f = rdof(r)
               maxort(r) = ort(zort(f)+gpoint(f)-1)
            enddo
         endif
      endif ! lprojerr

      jj= jj+1
 21   continue
!     advance to next point on remaining grid
      j = j+1
      do f=1,ndof
         if (.not.isproj(f)) then
            if (gpoint(f).lt.gdim(f)) then
               gpoint(f) = gpoint(f)+1
               goto 20
            else
               gpoint(f) = 1
            endif
         endif
      enddo
!     finished loop over remaining grid

      vprojdim = j-1
      nodedim  = jj-1
      call stop_timer(tid) ! stop timer

! ----------------------------------------------------------------------
!     Write out data.
! ----------------------------------------------------------------------

!     write projected potential to files, and logs
      do p=1,nproj
#ifdef MPI
         if (lmpi) then
!           gather data from slaves
            if (mpirank.eq.0) then
               do node=0,nmpi-1
                  if (node.ne.0) call MPI_Recv(vproj(1,p), &
                     maxnodedim,MPI_DOUBLE_PRECISION, &
                     node,p,MPI_COMM_WORLD,mpistat,mpierr)
                  do jj=1,maxnodedim
                     j = (jj-1)*nmpi + node + 1
                     if (j.le.vprojdim) vbuf(j) = vproj(jj,p)
                  enddo
               enddo
               call outproj(vbuf,vprojdim,p,nrdof,rdof)
            else
               call MPI_Send(vproj(1,p),maxnodedim,MPI_DOUBLE_PRECISION, &
                    0,p,MPI_COMM_WORLD,mpierr)
            endif
         else
            call outproj(vproj(1,p),vprojdim,p,nrdof,rdof)
         endif
#else
         call outproj(vproj(1,p),vprojdim,p,nrdof,rdof)
#endif
      enddo ! loop over projectors


      if (lprojerr) then
#ifdef MPI
         if (lmpi) then
!        gather statistics from slaves
            if (mpirank.eq.0) then
               do node=1,nmpi-1
                  call MPI_Recv(mpicommr,2*nrdof+3,MPI_DOUBLE_PRECISION, &
                       node,0,MPI_COMM_WORLD,mpistat,mpierr)
                  if (mpicommr(1).lt.minverr) then
                     minverr=mpicommr(1)
                     do r=1,nrdof
                        minort(r)=mpicommr(3+r)
                     enddo
                  endif
                  if (mpicommr(2).gt.maxverr) then
                     maxverr=mpicommr(2)
                     do r=1,nrdof
                        maxort(r)=mpicommr(3+nrdof+r)
                     enddo
                  endif
                  meanverr = meanverr+mpicommr(3)
               enddo
            else
               mpicommr(1) = minverr
               mpicommr(2) = maxverr
               mpicommr(3) = meanverr
               do r=1,nrdof
                  mpicommr(3+r)=minort(r)
                  mpicommr(3+nrdof+r)=maxort(r)
               enddo
               call MPI_Send(mpicommr,2*nrdof+3,MPI_DOUBLE_PRECISION, &
                    0,0,MPI_COMM_WORLD,mpierr)
            endif
         endif
#endif
         if (mpirank.eq.0) then
            meanverr = meanverr/vprojdim
!           write logs
            call logprjerr(minverr,maxverr,meanverr,minort,maxort,nrdof,rdof)
         endif

!        write projection error to file, and logs
#ifdef MPI
         if (lmpi) then
!           gather data from slaves
            if (mpirank.eq.0) then
               do node=0,nmpi-1
                  if (node.ne.0) call MPI_Recv(vperr, &
                     maxnodedim,MPI_DOUBLE_PRECISION, &
                     node,nproj+1,MPI_COMM_WORLD,mpistat,mpierr)
                  do jj=1,maxnodedim
                     j = (jj-1)*nmpi + node + 1
                     if (j.le.vprojdim) vbuf(j) = vperr(jj)
                  enddo
               enddo
               call outprjerr(vbuf,vprojdim,nrdof,rdof)
            else
               call MPI_Send(vperr,maxnodedim,MPI_DOUBLE_PRECISION, &
     +              0,nproj+1,MPI_COMM_WORLD,mpierr)
            endif
         else
            call outprjerr(vperr,vprojdim,nrdof,rdof)
         endif
#else
         call outprjerr(vperr,vprojdim,nrdof,rdof)
#endif
      endif ! lprojerr

      end subroutine projrun2


!***********************************************************************
!
! SUBROUTINE WRVPHEADER
!
! Writes the vpot header to the vpot_LABEL file created during
! a projection run.
! The vpot_LABEL file must already be opened and connected to
! channel "unit". Its name is passed in "fname".
!
! If p1=-1, create the header for the projerr.vpot file.
!
! This code does basically the same thing as the routine sequence
! wrdvrdef--writevpotpar. However, these routines write
! their information from the wrong global variables, so we must
! duplicate some code.
!
! 10/2005 FO
! 05/2006 FO - update for new file format
!
!***********************************************************************

      subroutine wrvpheader(unit,fname,p1,nrdof,rdof)

use global
use pglobal
use maxv, only:maxdim
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use channels

use iodvrdef
use vpotpars, only:writevpotpars

      implicit none

      integer(long), intent(in)                      :: unit,p1,nrdof
      integer(long), dimension(maxdim), intent(in)   :: rdof
      character(len=*), intent(in)                   :: fname
      integer(long)                                  :: p,i,f,sl,ierr
      integer(long), external                        :: log2int
      integer(long), dimension(maxdim)               :: vpgdim,vpbasis
      integer(long), dimension(mbaspar,maxdim)       :: vpipbaspar
      real(dop), dimension(2,maxdim)                 :: vpxend
      real(dop), dimension(mbaspar,maxdim)           :: vprpbaspar
      character(len=c1), dimension(maxdim)           :: vpmodelabel
      logical(kind=4), dimension(maxdim)             :: vpldvr
      character(len=c2)                              :: wrtag


      if (p1.eq.-1) then
         p=nproj
      else
         p=p1
      endif

! ----------------------------------------------------------------------
!     construct the DVR info of the "reduced" DVR, i.e. a DVR where
!     all the DOFs that are projected out are omitted
! ----------------------------------------------------------------------

!     for all the remaining DOFs:
      do f=1,nrdof
!        copy the data from the system DVR
         vpmodelabel(f) = modelabel(rdof(f))
         vpgdim(f)      = gdim(rdof(f))
         vpbasis(f)     = basis(rdof(f))
         vpldvr(f)      = ldvr(rdof(f))
         do i=1,mbaspar
            vpipbaspar(i,f) = ipbaspar(i,rdof(f))
            vprpbaspar(i,f) = rpbaspar(i,rdof(f))
         enddo
         do i=1,2
            vpxend(i,f) = xend(i,rdof(f))
         enddo
      enddo

! ----------------------------------------------------------------------
!     now we have all information to write the headers
! ----------------------------------------------------------------------

!     write file version
      filever(unit)=progver
      sl=index(fname,' ')-1
      write(unit,iostat=ierr) progver
      if (ierr.ne.0) then
         routine='wrvpheader (potfit/projutil.F)'
         message='cannot write program version to '//fname(1:sl)
         call errormsg
      endif

      wrtag= '<start dvr definition>'
      write(unit) wrtag

      wrtag= '<ndof>'
      write(unit) wrtag
      write(unit) nrdof
      wrtag= '<modelabel>'
      write(unit) wrtag
      write(unit) (vpmodelabel(f),f=1,nrdof)
      wrtag= '<gdim>'
      write(unit) wrtag
      write(unit) (vpgdim(f),f=1,nrdof)

      wrtag= '<mbaspar>'
      write(unit) wrtag
      write(unit) mbaspar
      wrtag= '<basis>'
      write(unit) wrtag
      write(unit) (vpbasis(f),f=1,nrdof)
      wrtag= '<ldvr>'
      write(unit) wrtag
      write(unit) (log2int(vpldvr(f)),f=1,nrdof)
      wrtag= '<ipbaspar>'
      write(unit) wrtag
      write(unit) ((vpipbaspar(i,f),i=1,mbaspar),f=1,nrdof)
      wrtag= '<rpbaspar>'
      write(unit) wrtag
      write(unit) ((vprpbaspar(i,f),i=1,mbaspar),f=1,nrdof)
      wrtag= '<xend>'
      write(unit) wrtag
      write(unit) ((vpxend(i,f),i=1,2),f=1,nrdof)

      wrtag= '<end dvr definition>'
      write(unit) wrtag

!     write vpot parameters header
      call writevpotpars(unit,fname,p1)

!     write even more info to the logfile
      write(ilog,'(a,i3,a,$)') '   ',nrdof,' DOFs remaining: '
      do f=1,nrdof
         sl=index(vpmodelabel(f),' ')-1
         write(ilog,'(2a,$)') vpmodelabel(f)(1:sl),' '
      enddo
      write(ilog,'(x)')

      return
      end subroutine wrvpheader ! of wrvpheader


!***********************************************************************
!
! SUBROUTINE WRVPROJ
!
! Writes the projected potential (stored in array vproj, of size
! vprojdim) to the file connected to unit. Data has to be written
! in chunks of niobuf for compatibility reasons.
!
! 10/2005 FO
!
!***********************************************************************

      subroutine wrvproj(unit,vproj,vprojdim,niobuf)

      implicit none

      integer(long), intent(in)           :: unit,vprojdim,niobuf
      real(dop), dimension(*), intent(in) :: vproj
      integer(long)                       :: j,j1,j2

!     write size of reduced grid and chunksize
      write(unit) vprojdim,niobuf

!     write data points in chunks of niobuf
      j1 = 1
 10   j2 = min(j1+niobuf-1,vprojdim)
      write(unit) (vproj(j), j=j1,j2)
      j1 = j1+niobuf
      if (j1.le.vprojdim) goto 10

      return
      end subroutine wrvproj ! of wrvproj


      end module projutil

