!***********************************************************************
      module weightsmod

use decimal, only:long,dop
use logdat

      implicit none
      private
      public :: weights,getsepwei
      contains

!***********************************************************************
!
!                 POTFIT   subroutine  WEIGHTS
!
!   Defines depending on value of 'iwei(m)' one-dimensional
!   weight functions (and their inverse) for each degree of freedom.
!   Weights are written to channel 'iout' and 'ipw'.
!
!   input:  ort -- array with grid points for each dimension
!
!   output:  weight -- array with weights for each point
!            weighti -- array with inverse weights for 
!                       each point
!   02/97 AJ 
!
!
!***********************************************************************

      subroutine weights(weight,weighti,ort)

use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use combined

use callanldmod, only:defanld
use getpesmod, only:get1dpot

      implicit none

      integer(long), parameter :: maxwpotdim = 1024

      integer(long), external                      :: slen
      integer(long)                                :: g,f,hl,ierr,n,m
      real(dop), dimension(weightdim), intent(out) :: weight,weighti
      real(dop), dimension(ortdim), intent(in)     :: ort
      real(dop), dimension(maxwpotdim)             :: wpot
      real(dop)                                    :: w,q,v1dmin
      character(len=c4)                            :: ctmp
      character(len=16)                            :: qstring

!-----------------------------------------------------------------------
! Set name of subroutine for error mesages
!-----------------------------------------------------------------------
      routine = 'WEIGHTS'

!------------------------------------------------------------------------
! Define weight functions for each degree of freedom
!------------------------------------------------------------------------
      do m=1,npmode
         do n=1,npotdof(m)
            f=potdof(n,m) 

!------------------------------------------------------------------------
! Dependent on one-dimensional potentials.
!------------------------------------------------------------------------
            if (iwei(f).eq.2) then
               if (gdim(f).gt.maxwpotdim) then
                  message = 'Increase maxwpotdim'
                  call errormsg               
               endif
               hl=nint(wpar(1,f))
               call defanld(hopilab(hl),hoplab(hl),hopfile(hl),&
                    hopipar(1,hl),hoprpar(1,hl),maxhpar,hoppar,&
                    apar,rpar,npar,ierr,zpesopts,pesopts,hamlab,hamdim)
               if (hopilab(hl) .eq. 0) then
                  routine='weights'
                  message='1D PES for dof '//modelabel(f)
                  ctmp=message(1:slen(message))//' not found in library :' 
                  message=ctmp(1:slen(ctmp)+1)//hoplab(hl)
                  call errormsg
               endif

               call get1dpot(wpot,ort(zort(f)),gdim(f),&
                    hopilab(hl),hopfile(hl),hopipar(1,hl),hoprpar(1,hl),&
                    pesmax,pesmin,lcutpes,hoplab(hl),lsoftcutpes)

               v1dmin = wpot(1)
               do g=2,gdim(f)
                  v1dmin=min(wpot(g),v1dmin)
               enddo
               do g=1,gdim(f)
                  wpot(g)=wpot(g)-v1dmin
               enddo
            endif

            do g=0,gdim(f)-1
               q=ort(zort(f)+g)

!------------------------------------------------------------------------
! No weights.               
!------------------------------------------------------------------------
               if (iwei(f).eq.0) then
                  w=1

!------------------------------------------------------------------------
!   [ q/q0 / 1+q/q0 ] ** 4
!------------------------------------------------------------------------
               elseif (iwei(f).eq.1) then
                  w=(q/(wpar(1,f)+q))**4

!------------------------------------------------------------------------
! Dependent on one-dimensional potential. (For NOCl it was
!    w=1_dop/(1_dop+(oned(zort(m)+g)-v1dmin)/wpar(1,m))**wpar(2,m))
!------------------------------------------------------------------------
               elseif (iwei(f).eq.2) then
                  w=1.0_dop/(1.0_dop+wpot(g+1)/wpar(2,f))**wpar(3,f)

!------------------------------------------------------------------------
!   exp [ -w2(q-w1)**2 ]
!------------------------------------------------------------------------
               elseif (iwei(f).eq.3) then
                  w=exp(-(wpar(2,f)*(q-wpar(1,f))**2))

!------------------------------------------------------------------------
!   cos [ 0.5 (q-q0) ]
!------------------------------------------------------------------------
               elseif (iwei(f).eq.4) then
                  w=cos(0.5_dop*(q-wpar(1,f)))

!------------------------------------------------------------------------
!   0.5 + 1/Pi arctan [ (q-w1)/w2 ]
!------------------------------------------------------------------------
               elseif (iwei(f).eq.5) then
                  w=0.5_dop+1.0_dop/Pi*atan((q-wpar(1,f))/wpar(2,f))

!------------------------------------------------------------------------
!  Weight 6 actually does not act as a weight but as a projector (for adproj)
!------------------------------------------------------------------------
               elseif (iwei(f).eq.6) then
                  if( q.gt.wpar(1,f) .and. q.lt.wpar(2,f) ) then
                     w=1.d0
                  else
                     w=0.d0
                  endif

!------------------------------------------------------------------------
!   0.5 - 1/Pi arctan [(q-w1)*tan[(0.5-w3)Pi]/w2]
! 
!  The parameters w1,w2 and w3 are chosen such that the weights change
!  for points located in the interval [w1-w2, w1+w2] from w3 to 1-w3.
!------------------------------------------------------------------------
               elseif (iwei(f).eq.7) then
                  w=0.5_dop-1.0_dop/Pi*atan((q-wpar(1,f))&
                       *tan((0.5-wpar(3,f))*Pi)/wpar(2,f))

!-----------------------------------------------------------------------
!  1 - (1-w1)*sin[x]**2  0<=x<=Pi
!
!  The weight at the edges of the interval 0 and Pi is one and
!  decreases monotonically to the minimum value 0<=w1<=1 at the middle
!  of the interval at Pi/2.
!-----------------------------------------------------------------------
               elseif (iwei(f).eq.8) then
                  w=1.0_dop-(1.0_dop-wpar(1,f))*sin(q)**2

!-----------------------------------------------------------------------
!
!-----------------------------------------------------------------------
               elseif (iwei(f).eq.9) then
                  w=cos(q-wpar(1,f))**2+wpar(2,f)
               
!------------------------------------------------------------------------
! Invalid digit in iwei(m)
!------------------------------------------------------------------------
               else
                  w=1.0_dop
                  message = 'Invalid weight identification number'
                  call errormsg
               endif

!......avoid too small weights
               if(iwei(f) .ne. 6) w = max(w,1.0d-6)

               weight(zweight(f)+g)  = w

               if (w.eq.0.0_dop .and. iwei(f).ne.6) then
                  routine='weights'
                  write(qstring,'(e16.8)') q
                  message='vanishing separable weight: '//&
                     'mode='//modelabel(f)//' q='//qstring
                  call errormsg
               endif
               if(iwei(f) .eq. 6) then
                  weighti(zweight(f)+g) = 1.0_dop
               else
                  weighti(zweight(f)+g) = 1.0_dop/w     
               endif
  
            enddo
         enddo
      enddo

      return
      end subroutine weights

!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!
!     returns via sw the product of all separable weights which may 
!     be defined for the grid point of indices igpoint. if no
!     separable weights have been defined the routine returns 1.
!
!     igpoint (i)
!        vector of ndof integers representing the indeces of a grid
!        point.
!     weight (i)
!        vector containing the separable weights at the grid points of
!        every degree of freedom.
!     sw (o)
!        product of all separable weights defined at the grid point
!        described by igpoint.
!
!ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine getsepwei(igpoint,weight,sw)

use global
use pglobal
use dvrdatmod
use griddatmod
use operdef
use runpfitmod
use combined

      implicit none

      integer(long)                               :: f,n,m
      integer(long), dimension(ndof), intent(in)  :: igpoint
      real(dop), intent(out)                      :: sw
      real(dop), dimension(weightdim), intent(in) :: weight

      sw=1.0_dop
      if (lsw) then
         do m=1,npmode
            do n=1,npotdof(m)
               f=potdof(n,m)
            if (iwei(f).gt.0) sw=sw*weight(zweight(f)+igpoint(f)-1)
            enddo
         enddo
      endif
      
      return
      end subroutine getsepwei


      end module weightsmod


