!***********************************************************************
!
!                              AUTOCORR.F
!
! Contains all subroutines related to the output of the autocorrelation
! function.
!
! Subroutines:
!  d1autoout :  writes auto file at t=k*tout, k=0,1,...
!  d1autoout2 : writes auto file at t=(k+1/2)*tout
!  d1autolap  : computes overlap <psi1^*|psi2> (needed in auto routines)
!    
! V81 AR
!
!***********************************************************************
      module d1autocorr

use decimal, only: long, dop
use global
use maxv, only: maxsta
use psidef
use griddatmod
use channels
use multilayer, only: automl

use mmlib,   only: mmtxzz
use mtlib,   only: qtxxzz
use xvlib,   only: vvtxzz
use op1lib,  only: cpvxz

      implicit none
      private
      public :: d1autoout,d1autoout2
      contains

!***********************************************************************
!                               D1AUTOOUT
!
! Outputs autocorrelation function at time t to file auto
!
!***********************************************************************

      subroutine d1autoout(psi,atime)

      implicit none

      real(dop), intent(in)                               :: atime
      complex(dop), dimension(dgldim), intent(inout)      :: psi
      complex(dop)                                        :: auto
      complex(dop), dimension(maxsta,maxsta)              :: autos

!-----------------------------------------------------------------------
! Calculate autocorrelation function
!-----------------------------------------------------------------------
      call d1calcauto(psi,psi,auto,autos)

!-----------------------------------------------------------------------
! Write 'auto'-file output
!-----------------------------------------------------------------------
      write(iaut,'(1x,g18.9,3(1x,f18.14))')&
           atime,dble(auto),dimag(auto),abs(auto)
      call myflush(iaut)

      return
      end subroutine d1autoout

!***********************************************************************
!                                D1AUTOOUT2
!
! Outputs autocorrelation function at time t/2 to file auto. (Used in
! a CMF calculation with adaptive step size.)
! Whereas "autoout" writes the autocorrelation function c(2nt) =
! <psi^*(nt)|psi(nt)> to file auto, this routine writes c(2(n-1)t) =
! <psi^*((n-1)t)|psi(nt)>.
!
! V6.4 MB
!***********************************************************************

      subroutine d1autoout2(psiold,psinew,atime)

      implicit none

      real(dop), intent(in)                               :: atime
      complex(dop), dimension(dgldim), intent(in)         :: psiold,psinew
      complex(dop)                                        :: auto
      complex(dop), dimension(maxsta,maxsta)              :: autos

!-----------------------------------------------------------------------
! Calculate autocorrelation function
!-----------------------------------------------------------------------
      call d1calcauto(psiold,psinew,auto,autos)

!-----------------------------------------------------------------------
! Write 'auto'-file output
!-----------------------------------------------------------------------
      write(iaut,'(1x,g18.9,3(1x,f18.14))')&
           atime,dble(auto),dimag(auto),abs(auto)
      call myflush(iaut)
      
      return
      end subroutine d1autoout2

!***********************************************************************
!                               D1AUTOLAP
!
! Computes the overlap of two wave functions as needed for
! autocorrelation functions, i.e. auto = <psi1^*|psi2>. Autos contains
! the corresponding overlap for each state.
! It is assumed that the number of spfs for any mode is the same for
! psi1 and psi2.
!
!***********************************************************************

      subroutine d1calcauto(psi1,psi2,auto,autos)

      implicit none

      integer(long)                                       :: zeig1,zeig2,m,s,t
      complex(dop), intent(out)                           :: auto
      complex(dop), dimension(dgldim), intent(in)         :: psi1,psi2
      complex(dop), dimension(2*maxblock)                 :: workc
      complex(dop), dimension(maxsta,maxsta), intent(out) :: autos
      complex(dop), dimension(dmatdim)                    :: tmpovl
 
!-----------------------------------------------------------------------
! Compute overlap <psi1^*|psi2>
!-----------------------------------------------------------------------
      do s = 1,nstate
         do t=1,nstate
            do m = 1,nmode
               call mmtxzz(psi1(dzetf(m,s,t)),psi2(dzetf(m,s,t)),&
                    tmpovl(ddmat(m,s,t)),subdim(m),ddim(m,s,t),ddim(m,s,t))
            enddo
         enddo
      enddo

      auto = (0.0_dop,0.0_dop)
      if (lrunml) then
         call automl(autos(1,1),psi1,psi2,tmpovl)
         auto = autos(1,1)

      else
         do s = 1,nstate
            do t=1,nstate
               zeig1 = 1
               zeig2 = dblock(s,t)+1
               call cpvxz(psi2(dzpsi(s,t)),workc(zeig2),dblock(s,t))
               
               do m = 1,nmode
                  call iswap1(zeig1,zeig2)
                  call qtxxzz(tmpovl(ddmat(m,s,t)),workc(zeig1),workc(zeig2),&
                       vddim(m,s,t),ddim(m,s,t),nddim(m,s,t))
               enddo
               call vvtxzz(psi1(dzpsi(s,t)),workc(zeig2),autos(s,t),&
                    dblock(s,t))

               auto = auto+autos(s,t)
            enddo
         enddo
      endif

      return
      end subroutine d1calcauto

!***********************************************************************
!                                ISWAP1
!
! Swaps two integers (needed in autohigh). This routine has
! intentionally not been stored in the libraries to allow the optimizer
! to inline it.
!
! V7.1 MB
!***********************************************************************

      subroutine iswap1(int1,int2)

      implicit none

      integer(long)                :: aux
      integer(long), intent(inout) :: int1,int2

      aux = int1
      int1 = int2
      int2 = aux

      return
      end subroutine iswap1

      end module d1autocorr
