!#######################################################################
!
!                      DMCTDH module D1CALCDAT
!
!
! subroutines :
!    CALCSPOP1    : Calculates state populations 
!    CALCEUNCORR1 : Calculates uncorrelated energy per state
!    CALCECORR1   : Calculates correlated energy per state
!
!  V81 AR
!
!***********************************************************************
      module d1calcdatmod

use decimal, only: long, dop

use global
use maxv, only: maxsta
use dvrdatmod
use griddatmod
use psidef
use operdef
use runpropmod
use datenmod
use hpsimod

use xvlib, only: vvaxzz, xvxxzzo
use mmlib, only: mmaxzzh
use op1lib,  only: cpvxz, trhxz, normvxz
use op2lib,  only: sum2qxxz, addvxxzo
use d1lib, only: trspdo,trspdoc
use d1tens, only: ptr1,ptrc
use wftree1, only: allocate_mtraces_matrices,deallocate_mtraces_matrices,&
                  allocate_zmtraces_matrices
use mlanalysis, only: compute_mtraces_tree
use multilayer, only: traceml,ztraceml
use d1hphimod, only: d1hunphi1ms,d1hphi1mk
use expectval, only: calceuncorr1,calcecorr1,calcenrho,calcspop1

      implicit none
      private
      public :: d1calcdat
      contains

!***********************************************************************
!
!                  MCTDH subroutine d1CALCDAT
!
! calculates data to be written to various files.
!
! passed back to calling routine:
!   spop(s):     Population for each state.
!   detot(s):    Total energy for each state
!   decorr(s):   Correlated energy for each state
!   deuncorr(s): Uncorrelated energy for each state
!   detot2:      Total Tr(p^2 H)
!   trace:       Total trace.
!   intrace:     Initital trace. For relxation this is set to 1.0 as
!                trace is reported in output file as trace/intrace.
!
! 02/02 CC
!
!***********************************************************************

      subroutine d1calcdat(psi,dtpsi,hpsi,hunpsi,dicht3,spop,firstcall,&
                 dicht2,detot2,ecorr,etot,trace)

      implicit none

      integer(long)                                      :: s,m,d,s1,f
      real(dop), dimension(d3matdim), intent(in)         :: dicht3
      real(dop), dimension(maxsta), intent(out)          :: spop,etot,ecorr
      real(dop), intent(out)                             :: detot2,trace
      complex(dop), dimension(dgldim), intent(in)        :: psi,dtpsi
      complex(dop), dimension(dmatdim), intent(in)       :: dicht2
      complex(dop), dimension(hpsidim), intent(in)       :: hpsi
      complex(dop), dimension(adim+1:dgldim), intent(in) :: hunpsi
      complex(dop)                                       :: prefactor,sum
      complex(dop), dimension(maxspf*maxspf)             :: workc
      logical(kind=4), intent(in)                        :: firstcall

!-----------------------------------------------------------------------
! State populations
!-----------------------------------------------------------------------
      if (lexact) then
         do s=1,nstate
            call trhxz(psi(dzetf(1,s,s)),spop(s),supgdim(1))
         enddo

      else if (lquadpop) then
!    --- calculate the spops
         do s=1,nstate
            spop(s) = 0.0_dop
            do m=1,nmode
               do d=1,ddim(m,s,s)
                  spop(s)=spop(s) + dicht3(dd3mat(m,s,s)+d-1)
               enddo
            enddo
         enddo

      else
!        Use the old, i.e. linear, method. This is default.
         call calcspop1(psi,spop)
      endif

!-----------------------------------------------------------------------
! Trace
!-----------------------------------------------------------------------
      trace = 0.0_dop
      do s=1,nstate
         trace = trace + dble(spop(s))
      enddo

      if (firstcall) then
         if (lrlx) then
            intrace = 1.0_dop
         else
            intrace = trace
            if (intrace.le.tiny) intrace=1.0e99_dop
         endif
      endif

!    --- calculate euncorr and ecorr

      if (lexact) then
         call calcenrho(psi(dzetf(1,1,1)),decorr,deuncorr,1)
      else
         call calceuncorr1(psi,hunpsi,deuncorr,1)
         call calcecorr1(psi,hpsi,decorr,1)
      endif

      do s=1,nstate
         detot(s)=decorr(s)+deuncorr(s)
      enddo

!-----------------------------------------------------------------------
! Calculate expectation values
!-----------------------------------------------------------------------
      do s=1,nstate
         spop(s) = spop(s)/trace
         ecorr(s) = decorr(s) /trace
         etot(s) = detot(s)/trace
      enddo

!-----------------------------------------------------------------------
! Determine prefactor of dtpsi vector
!-----------------------------------------------------------------------
      if (lrlx) then
         prefactor = (-1.0_dop,0.0_dop)
         if(inttypea.eq.6) prefactor=(1.0_dop,0.0_dop)
      else
         prefactor = (0.0_dop,-1.0_dop)
      endif

!-----------------------------------------------------------------------
! calculate Tr(p^2 L) from the scalar product of <psi| H|psi>. 
! If run is with "proj-h" option, this is the total energy, otherwise 
! this is the correlated energy.
!
! for relaxation, dtpsi stores H-E|psi>
!-----------------------------------------------------------------------
      detot2=0.0_dop
      call vvaxzz(psi(dzpsi(1,1)),dtpsi(dzpsi(1,1)),sum,adim)
      if (lrlx) then
         detot2 = detot2 + abs((sum-rlxenergy)/prefactor)
      else
         detot2 = detot2 + abs(sum/prefactor)
      endif

!-----------------------------------------------------------------------
! calculate uncorrelated energy expectation value i.e. sum of
! uncorrelated matrix elements multiplied by the appropriate density
! matrix element. Sum of h|phi> terms are stored in hunpsi. 
!-----------------------------------------------------------------------
      if (lfunkl) then
         do s=1,nstate
            do s1=1,nstate
               do m=1,nmode
                  f=spfdof(1,m)
                  call mmaxzzh(psi(dzetf(m,s1,s)),hunpsi(dzetf(m,s1,s)),&
                       workc,subdim(m),ddim(m,s1,s))

                  call sum2qxxz(dicht2(ddmat(m,s1,s)),workc,sum,ddim(m,s1,s))
                  sum = sum - dconjg(sum)
                  detot2=detot2+abs(sum)

               enddo
            enddo
         enddo
      endif

      return
      end subroutine d1calcdat

      end module d1calcdatmod

