!***********************************************************************
!
! D1CROCOFLY.F
!
! Module to calculate the cross-correlation function on the fly
! for density matrices of type 1.
!
! Contains the following routines:
!   D1WKCROCO
!   D1ZEIGCROCO
!   D1CCOVERLAP
!   D1CCOUT
!
! Mar 2002, FO
!
!***********************************************************************
      module d1crocofly

use decimal, only: dop, long
use globmemmod
use mmlib,   only: mmaxzz
use mtlib,   only: mtxxzz
use xvlib,   only: vvaxzz
use op1lib,  only: cpvxz
use timing

      implicit none
      private
      public :: d1ccout,d1zeigcroco
      contains

!***********************************************************************
!
! subroutine D1ZEIGCROCO
!
! Calculates the size of the (static) memory needed in the calculation
! of the cross-correlation function.
!
! Output variables:
!   (to common blocks)
!   psirefdim : size of reference WF
!   ccjindxdim: size of JINDX array of ref. WF
!   ccwcdim: workc-space needed for this module
!   doverdim,ddovmat: needed later in D1CCOVERLAP
!
! Mar 2002, FO
!
!***********************************************************************

      subroutine d1zeigcroco

use griddatmod
use psidef
use crocoflymod
use compdatmod

use crocofly, only: rdccref1

      implicit none

      integer(long)       :: zeig,zeig1,m,s,t

!     obtain system definition of reference WF from "restart" file
      call rdccref1

!     calculate size of reference WF
      psirefdim = cdgldim(2)

!     memory needed for JINDX arrray
      if (ccitype(2) .gt. 0) then
         ccjindxdim = cadim(2)
      else
         ccjindxdim = 1
      endif

!     memory needed for the overlap calculation
!     [mainly cut'n'pasted from ANALYSE/CROSSCORR.F]
      ccwcdim=ccmaxblock
      zeig=1
      do s=1,nstate
         do t=1,nstate
            zeig1=dblock(s,t)
            do m=1,nmode
               ddovmat(m,s,t)=zeig
               zeig=zeig+cddim(m,s,t,2)*ddim(m,s,t)
               zeig1=zeig1*cddim(m,s,t,2)/ddim(m,s,t)
               ccwcdim=max(ccwcdim,zeig1)
            enddo
         enddo
      enddo
      doverdim=zeig-1

      end subroutine d1zeigcroco


!***********************************************************************
!
! subroutine D1CCOVERLAP
!
! Calculates the overlap between two MCTDH type 1 density matrices.
! [mainly cut'n'pasted from ANALYSE/CROSSCORR.F]
!
! In:
!   psiref : reference WF in MCTDH form
!   psi    : system WF
!   dover  : memory needed for overlap matrices [=mc(mcccdover)]
!   workc  : complex scratch array
! Out:
!   ovl = overlap of matrices
!
! Mar 2002, FO
!
!***********************************************************************

      subroutine d1ccoverlap(ovl,psiref,psi,dover)

use global
use griddatmod
use psidef
use crocoflymod
use compdatmod

      implicit none

      integer(long)                                  :: s,t,m,zeig1,zeig2,&
                                                        swapzeig
      integer(long)                                  :: tmpdim
      complex(dop)                                   :: zdum
      complex(dop), dimension(*), intent(in)         :: psiref
      complex(dop), dimension(*), intent(in)         :: psi
      complex(dop), intent(out)                      :: ovl
      complex(dop), dimension(doverdim), intent(out) :: dover
      complex(dop), dimension(:), allocatable        :: tempc
      integer(long), save                            :: tid=0 ! timer ID

      call get_timer(tid, 'Cross')
      call start_timer(tid)

!-----------------------------------------------------------------------
! work array
!-----------------------------------------------------------------------
      tmpdim = max(2*ccwcdim, 2*ccmaxblock)
      allocate(tempc(tmpdim))

!-----------------------------------------------------------------------
!     calculate density matrices for each degree of freedom defined as
!     overlap of the single-particle functions rho^m_ij=<phi_i,2|phi_j,1>
!-----------------------------------------------------------------------
      do s=nst1,nst2
         do t=nst1,nst2
            do m=1,nmode
               call mmaxzz(psiref(cdzetf(m,s,t,2)),psi(dzetf(m,s,t)),&
                    dover(ddovmat(m,s,t)),subdim(m),cddim(m,s,t,2),&
                    ddim(m,s,t))
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      ovl=(0.0_dop,0.0_dop)
      do s=nst1,nst2
         do t=nst1,nst2
            zeig1=1
            zeig2=ccwcdim+1
            call cpvxz(psi(dzpsi(s,t)),tempc(zeig2),dblock(s,t))

            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call mtxxzz(dover(ddovmat(m,s,t)),tempc(zeig1),&
                    tempc(zeig2),cvddim(m,s,t,2),ddim(m,s,t),&
                    nddim(m,s,t),cddim(m,s,t,2))
            enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* Sum_L <Phi_J,1|Phi_L,2>A_L,2 within each state
!-----------------------------------------------------------------------
            call vvaxzz(psiref(cdzpsi(s,t,2)),tempc(zeig2),zdum,&
                 cdblock(s,t,2))
            ovl=ovl+zdum
         enddo
      enddo

      deallocate(tempc)

      call stop_timer(tid)

      return
      end subroutine d1ccoverlap

!***********************************************************************
!
! subroutine D1CCOUT
!
! Writes the cross-correlation function to the ICROCO file.
!
! Mar 2002, FO
!
!***********************************************************************

      subroutine d1ccout(time,psi)

use global
use griddatmod
use psidef
use crocoflymod
use channels

      implicit none

      real(dop), intent(in)                     :: time
      complex(dop), dimension(*), intent(in)    :: psi
      complex(dop)                              :: ovl

!     calculate cross-correlation function
      call d1ccoverlap(ovl,mc(mcpsiref),psi,mc(mcccdover))

!     write result to ICROCO file
      write(icroco,'(1x,g18.9,3(1x,f18.14))') &
            time, dble(ovl), dimag(ovl), abs(ovl)
      call myflush(icroco)

      end subroutine d1ccout


      end module d1crocofly
