! **********************************************************************
! *                                                                    *
! *                         D1DICHT1PHI
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! * d1hlochphi1m:   Computes for one mode and all correlated           *
! *                 Hamiltonian terms the product of the mean-field    *
! *                 matrices and h|phi>.                               *
! * d1dicht1phi:    Multiplies for all modes and states the            *
! *                 single-particle functions with the inverse of the  *
! *                 density matrix.                                    *
! * d1dicht1phi1ms: Multiplies for one mode and state the              *
! *                 single-particle functions with the inverse of the  *
! *                 density matrix.                                    *
! *                                                                    *
! **********************************************************************
      module d1dicht1phimod

use decimal,  only: long, dop
use mmlib,    only: mqxtzz
use op1lib,   only: cpmxz


      implicit none
      private
      public :: d1dicht1phi,d1dicht1phi1ms
      contains

! *** V8.0 AR *************> SUBROUTINE D1DICHT1PHI <*********************
!                           ~~~~~~~~~~~~~~~~~~~~~~
! Multiplies for all modes and states the single-particle functions
! with the inverse of the density matrix.
! **********************************************************************

      subroutine d1dicht1phi (dtpsi,dicht1)

use global
use griddatmod
use psidef
use d1tens, only: makeaadj
use maxv,   only: maxdim,maxsta

      implicit none

      integer(long)                                  :: m,s,t
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi
      complex(dop), dimension(dmatdim), intent(in)   :: dicht1

! --- LOOP OVER EACH MODE AND STATE---

      do m = 1,nmode
         if (complete(m)) cycle
         do s=1,nstate
            call d1dicht1phi1ms(dtpsi(dzetf(m,s,s)),dicht1,m,s,s,&
                 ddim(m,s,s),subdim(m))
            do t=(s+1),nstate
               call d1dicht1phi1ms(dtpsi(dzetf(m,s,t)),dicht1,m,&
                    s,t,ddim(m,s,t),subdim(m))
               call makeaadj(dtpsi(dzetf(m,s,t)),dtpsi(dzetf(m,t,s)),&
                    supgdim(m),ddim(m,s,t))
            enddo
         enddo
      enddo

      return
      end subroutine d1dicht1phi

! *** V8.0 AR ************> SUBROUTINE DICHT1PHI 1MS <******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~~~

! Multiplies for one mode and one state the single-particle functions
! with the inverse of the density matrix.
! **********************************************************************

      subroutine d1dicht1phi1ms (dtpsi,dicht1,m,s,t,dim1,subdim1)

use global
use griddatmod
use psidef
use operdef
use d1tens, only: mkaherm
use d2tens, only: mkherm
use maxv,   only: maxdim,maxsta,maxkoe
use openmpmod
use runpropmod

      implicit none

      integer(long)                                         :: n
      integer(long), intent(in)                             :: m,s,t,dim1,subdim1
      complex(dop), dimension(subdim1,dim1), intent(inout)  :: dtpsi
      complex(dop), dimension(subdim1,dim1)                 :: workc
      complex(dop), dimension(dmatdim), intent(in)          :: dicht1

! --- IF BASIS IS COMPLETE, DO NOT CALCULATE ---

      if (complete(m)) return

! --- MULTIPLY SINGLE-PARTICLE FUNCTION WITH INVERSE DENSITY MATRIX ---

      call cpmxz(dtpsi,workc,subdim(m),ddim(m,s,t))
      call mqxtzz(workc,dicht1(ddmat(m,s,t)),dtpsi,subdim(m),&
           ddim(m,s,t))

!     Take care of initial conditions
      if (s.eq.t) then
         if (lrlx) then
            do n=1,ddim(m,s,s)
               call mkherm(dtpsi(1,n),supgdim(m))
            enddo
         else
            do n=1,ddim(m,s,s)
               call mkaherm(dtpsi(1,n),supgdim(m))
            enddo
         endif
      endif
      

      return
      end subroutine d1dicht1phi1ms

      end module d1dicht1phimod

