! **********************************************************************
! *                                                                    *
! *                         EVALUATION (evaluation.f)                  *
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! *   prefactor:    Multiplies the wavefunction derivative with -i for *
! *                 propagation or -1 for imaginary time relaxation.   *
! *   subenergy:    Subtracts the energy from the Hamiltonian, i. e.   *
! *                 H|psi> is transformed to (H-E)|psi> where E =      *
! *                 <psi|H|psi>/<psi|psi> (needed for relaxation).     *
! *                                                                    *
! **********************************************************************
      module d1evaluate

use decimal, only: long, dop
use globmemmod
use global
use griddatmod
use psidef
use xvlib,   only: xvxxdzr
use d1tens, only: mkaherm,mkaherm2
use d2tens, only: chkherm, mkherm, chkherm2, mkherm2, chkaherm, chkaherm2



      implicit none
      private
      public :: chkini,chkinib,chkini1m
      contains

!     *** V8.0 AR ************> SUBROUTINE CHKINI <**********************
!                             ~~~~~~~~~~~~~~~~~~
!
! Checks initial conditions

      subroutine chkini(psi)

      implicit none

      integer(long)                               :: b,m,n,s,t,z,z1,z2
      real(dop)                                   :: err
      complex(dop), dimension(dgldim), intent(in) :: psi

      err = 0.0_dop
      do s=1,nstate
         do b=1,dblock(s,s)
            err = max(err,abs(dimag(psi(dzpsi(s,s)+b-1))))
         enddo
!         psi(dzpsi(s,s)+b-1) = dble(psi(dzpsi(s,s)+b-1))
         do t=(s+1),nstate
            do b=1,dblock(s,t)
               err = max(err, abs(psi(dzpsi(s,t)+b-1) -  &
                    conjg(psi(dzpsi(t,s)+b-1))) )
!               psi(dzpsi(s,t)+b-1) = conjg(psi(dzpsi(t,s)+b-1))
            enddo
         enddo
      enddo
      write(6,*) 'B: ',err

      err = 0.0_dop
      do s=1,nstate
         do m=1,nmode
            z = dzetf(m,s,s)
            do n=1,ddim(m,s,s)
               call chkherm(psi(z),err,supgdim(m))
!               call mkherm(psi(z),supgdim(m))
               z = z + subdim(m)
            enddo
         enddo
         do t=(s+1),nstate
            z1 = dzetf(m,s,t)
            z2 = dzetf(m,t,s)
            do n=1,ddim(m,s,t)
               call chkherm2(psi(z1),psi(z2),err,supgdim(m),supgdim(m))
!               call mkherm2(psi(z1),psi(z2),supgdim(m),supgdim(m))
            enddo
            z1 = z1 + subdim(m)
            z2 = z2 + subdim(m)
         enddo
      enddo

      write(6,*) 'Psi: ',err
!      write(6,*)

      return
      end subroutine chkini

!     *** V8.0 AR ************> SUBROUTINE CHKINI1M <**********************
!                              ~~~~~~~~~~~~~~~~~~~~
! Checks initial conditions 
! **********************************************************************

      subroutine chkini1m(psi,phidim1,m,herm)

      implicit none

      integer(long)                                   :: n,s,t,z,z1,z2
      integer(long), intent(in)                       :: phidim1,m
      real(dop)                                       :: err
      complex(dop), dimension(phidim1), intent(inout) :: psi
      logical(kind=4), intent(in)                     :: herm

      err = 0.0_dop
      do s=1,nstate
         z = dzetf(m,s,s) - dzetf(m,1,1) + 1
         do n=1,ddim(m,s,s)
            if (herm) then
               call chkherm(psi(z),err,supgdim(m))
               call mkherm(psi(z),supgdim(m))
            else
               call chkaherm(psi(z),err,supgdim(m))
               call mkaherm(psi(z),supgdim(m))
            endif
            z = z + subdim(m)
         enddo
         do t=(s+1),nstate
            z1 = dzetf(m,s,t) - dzetf(m,1,1) + 1
            z2 = dzetf(m,t,s) - dzetf(m,1,1) + 1
            do n=1,ddim(m,s,t)
               if (herm) then
                  call chkherm2(psi(z1),psi(z2),err,supgdim(m), &
                       supgdim(m))
                  call mkherm2(psi(z1),psi(z2),supgdim(m),supgdim(m))
               else
                  call chkaherm2(psi(z1),psi(z2),err,supgdim(m),&
                       supgdim(m))
                  call mkaherm2(psi(z1),psi(z2),supgdim(m),supgdim(m))
               endif
            enddo
            z1 = z1 + subdim(m)
            z2 = z2 + subdim(m)
         enddo
      enddo

      write(6,*) err
!      if (err .gt. 1.0d-15) then
!         write(6,*) 'Warning: Initial conditions are not conserved for '
!     +            //' mode ',m,'.' 
!      endif

      return
      end subroutine chkini1m


!     *** V8.0 AR ************> SUBROUTINE CHKINIB <**********************
!                              ~~~~~~~~~~~~~~~~~~~

! Checks initial conditions for the coefficients
!-----------------------------------------------------------------------

      subroutine chkinib(psi,herm)

      implicit none

      integer(long)                                :: b,s,t
      real(dop)                                    :: err
      complex(dop), dimension(adim), intent(inout) :: psi
      logical(kind=4), intent(in)                  :: herm

      err = 0.0_dop
      do s=1,nstate
         if (herm) then
            do b=1,dblock(s,s)
               err = max(err,abs(dimag(psi(dzpsi(s,s)+b-1))))
               psi(dzpsi(s,s)+b-1) = dble(psi(dzpsi(s,s)+b-1))
            enddo
         else
            do b=1,dblock(s,s)
               err = max(err,abs(dble(psi(dzpsi(s,s)+b-1))))
               psi(dzpsi(s,s)+b-1) = (0.0_dop,1.0_dop) *&
                    dimag(psi(dzpsi(s,s)+b-1))
            enddo
         endif
         do t=(s+1),nstate
            if (herm) then
               do b=1,dblock(s,t)
                  err = max(err, abs(psi(dzpsi(s,t)+b-1) - &
                       conjg(psi(dzpsi(t,s)+b-1))) )
                  psi(dzpsi(s,t)+b-1) = 0.5_dop*( psi(dzpsi(s,t)+b-1)&
                       + conjg(psi(dzpsi(t,s)+b-1)) )
                  psi(dzpsi(t,s)+b-1) = conjg(psi(dzpsi(s,t)+b-1))
               enddo
            else
               do b=1,dblock(s,t)
                  err = max(err, abs(psi(dzpsi(s,t)+b-1) + &
                       conjg(psi(dzpsi(t,s)+b-1))) )
                  psi(dzpsi(s,t)+b-1) = 0.5_dop*( psi(dzpsi(s,t)+b-1)&
                       - conjg(psi(dzpsi(t,s)+b-1)) )
                  psi(dzpsi(t,s)+b-1) = -conjg(psi(dzpsi(s,t)+b-1))
               enddo
            endif
         enddo
      enddo
!      write(6,*) err

!      if (err .gt. 1.0d-15) then
!         write(6,*) 'Warning: Initial conditions are not conserved for '
!     +            //' the coefficients.' 
!      endif

      return
      end subroutine chkinib

      end module d1evaluate

