! **********************************************************************
! *                                                                    *
! *                         D1HLOCHPHI
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! * d1hlochphi:     Computes for all modes, states, and all correlated *
! *                 Hamiltonian terms the product of the mean-field    *
! *                 matrices and h|phi>.                               *
! * d1hlochphi1m:   Computes for one mode and all correlated           *
! *                 Hamiltonian terms the product of the mean-field    *
! *                 matrices and h|phi>.                               *
! **********************************************************************
      module d1hlochphimod

use decimal,    only: long, dop 
use mmlib,      only: mmxtzz, mmxtzza, mmxxzz
use op2lib,     only: addmxxzo, submxxzo1

      implicit none
      private
      public :: d1hlochphi,d1hlochphi1m
      contains

! *** V6.0 MB *************> SUBROUTINE HLOCHPHI <**********************
!                           ~~~~~~~~~~~~~~~~~~~~~

! Computes for all modes, all states and all Hamiltonian terms the
! product of the mean-field matrices and h|phi>.
! **********************************************************************

      subroutine d1hlochphi (dtpsi,hpsi,hloch)

use global
use griddatmod
use psidef
use hpsimod

      implicit none

      integer(long)                                       :: m
      complex(dop), dimension(dgldim), intent(inout)      :: dtpsi
      complex(dop), dimension(hpsidim), intent(in)        :: hpsi
      complex(dop), dimension(hlochdim), intent(in)       :: hloch

! --- LOOP OVER EACH MODE AND STATE---

      do m = 1,nmode
         call d1hlochphi1m(dtpsi(dzetf(m,1,1)),hpsi,hloch,&
              m,dzetf(m,1,1),phidim(m))
      enddo

      return
      end subroutine d1hlochphi

! *** V8.0 AR ************> SUBROUTINE HLOCHPHI 1M <*******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~
!
! Computes for one mode, and all Hamiltonian terms the product of the
! mean-field matrices and h|phi>. For density operators one has
! additionally to add hloch^dagger times hpsi^dagger to account for
! the Liouvillian action
! **********************************************************************

      subroutine d1hlochphi1m (dtpsi,hpsi,hloch,m,dzetf1,phidim1)

use global
use griddatmod
use psidef
use operdef
use runpropmod
use hpsimod
use d1tens, only: makeadj
use maxv,   only: maxdim,maxsta,maxkoe
use openmpmod

      implicit none

      integer(long)                   :: k,kt,s,t,s1,s2,t1,t2,z1,z2
      integer(long), intent(in)       :: m,dzetf1,phidim1
      complex(dop), dimension(dzetf1:dzetf1+phidim1-1), intent(inout) :: dtpsi
      complex(dop), dimension(hpsidim), intent(in)                    :: hpsi
      complex(dop), dimension(hlochdim), intent(in)                   :: hloch
      complex(dop), dimension(2*sdim*maxspf)                          :: workc

! --- IF BASIS IS COMPLETE DO NOT CALCULATE ---

      if (complete(m)) return

! --- COMPUTE MEAN-FIELD TIMES H|PHI> FOR ALL STATES AND TERMS ---

      do k = 1,kzahl(1)
         if (.not. diag(m,k)) then
!     Hamiltonian terms
            if (kact(k).eq.1) then
               if (.not.lconst .or. kinop(m,k).eq.1) then
                  s=kf(k)
                  s1=ki(k)
                  do t=1,nstate
                     call mmxtzz(hpsi(dzhpsi(m,k,t)), &
                          hloch(dphloch(m,k,t)),workc,subdim(m), &
                          ddim(m,s1,t),ddim(m,s,t))

!     ddim(m,s,t) = ddim(m,t,s)
                     call addmxxzo(workc,dtpsi(dzetf(m,s,t)), &
                          subdim(m),ddim(m,s,t))
                     z1 = 1
                     z2 = subdim(m)*ddim(m,s,t)+1
                     call makeadj(workc(z1),workc(z2),supgdim(m), &
                          ddim(m,s,t))
                     if (lrlx) then
                        call addmxxzo(workc(z2),dtpsi(dzetf(m,t,s)), &
                             subdim(m),ddim(m,t,s))               
                     else
                        call submxxzo1(dtpsi(dzetf(m,t,s)),workc(z2), &
                             subdim(m),ddim(m,t,s))
                     endif
                  enddo      
               endif
            endif
!     Purely left-acting dissipative terms
            if (kact(k).eq.2) then
               if (.not.lconst .or. kinop(m,k).eq.1) then
                  s=kf(k)
                  s1=ki(k)
                  do t=1,nstate
! Parallel?
                     call mmxtzza(hpsi(dzhpsi(m,k,t)), &
                          hloch(dphloch(m,k,t)),dtpsi(dzetf(m,s,t)), &
                          subdim(m),ddim(m,s1,t),ddim(m,s,t))
                  enddo
               endif
            endif
!     Purely right-acting dissipative terms
            if (kact(k).eq.3) then
               if (.not.lconst .or. kinop(m,k).eq.1) then
                  t1=ki(k)
                  t2=kf(k)
                  do t=1,nstate
! Parallel?
                     call mmxtzza(hpsi(dzhpsi(m,k,t)), &
                          hloch(dphloch(m,k,t)),dtpsi(dzetf(m,t,t1)), &
                          subdim(m),ddim(m,t,t2),ddim(m,t,t1))
                  enddo
               endif
            endif
!     left-right acting operators
            if (kact(k).eq.4) then
               kt = ktwin(k)            
               s1 = kf(k)
               s2 = ki(kt)
               t1 = ki(k)
               t2 = kf(kt)
! Parallel?
               call mmxtzza(hpsi(dzhpsi(m,k,1)), &
                    hloch(dphloch(m,k,1)),dtpsi(dzetf(m,s1,s2)), &
                    subdim(m),ddim(m,t1,t2),ddim(m,s1,s2))
            endif               
         endif
      enddo

! Call was commented out, so loop no longer served purpose. Perhaps it 
! does though. GWR 9/13
!      k=kzahl(1)
!      do k1=1,nmulpot(1)
!         k=k+1
!         if (.not. diag(m,k)) then
!            s=kf(k)
!!            call addmxxzo(hloch(dphloch(m,k)),dtpsi(dzetf(m,s)),
!!     +              subdim(m),ddim(m,s))
!         endif
!      enddo

      return
      end subroutine d1hlochphi1m

      end module d1hlochphimod
