!#######################################################################
!
!                      DMCTDH module D1OUTGPOP
!
!
! subroutines :
!    OUTGPOP1     : Calculates grid populations in coordinate and 
!                  momentum representations and outputs results to 
!                  dicht file
!    CALCGPOP1    : Calculates the grid pops gpop1 and gpop2 for one 
!                  mode and one state
!    EDTRAFO1     : Transformation of the density for exp-DVR
!    ADDGPOP1     : Adds up the gpops for dof f
!    CALCSHT1     : Calculates traces of Single-hole density operators
!
!  V81 AR
!
!***********************************************************************
      module d1outgpop

use decimal, only: long, dop
use global
use globmemmod
use dvrdatmod
use griddatmod
use operdef
use psidef
use hpsimod
use runpropmod
use datenmod
use channels

use d2tens, only: rm1_mf
use d1tens, only: ptr1,ptr2
use d1lib,  only: trspdo
use trafonatmod, only: d1trafonat
use daten, only: activcap

use mmlib,   only: qmxxdz, mqxtzd, qqxxzz
use xvlib,   only: xqxxdzo
use fftmod,  only: fft
use op1lib,  only: cpqxz, cpvxz, cpvxd, zerovxd
use dichtoutmod, only: stproj
use densitymod, only: density1
use eqofmotion1, only: compute_dmat_tree
use wftree1, only: allocate_mtraces_matrices,deallocate_mtraces_matrices, &
                  allocate_shtraces_matrices,deallocate_shtraces_matrices
use mlanalysis, only: compute_mtraces_tree,compute_shtraces_tree

      implicit none
      private
      public :: outgpop1
      contains

!***********************************************************************
!
! OUTGPOP1:  Calculates grid populations in coordinate and momentum
!            representations and outputs results to dicht file.
!
! Passed back to calling routine:
! gpop1(m,s): first set of grid populations 
! gpop2(m,s): second set of grid populations 
!
! gpop1 contains:
!   if time independent basis, basis function population
!   if DVR of FFT, coordinate representation populations
! gpop2 contains:
!   if time independent basis, basis function population (identical to 
!      gpop1)
!   if DVR, FBR basis function occupation numbers
!   if FFT, momentum representation populations
!
!
! V8.0  AR
!***********************************************************************

      subroutine outgpop1(psi,gpop1,gpop2,trafo,fftfak,hin,rueck,&
                 shtraces,time,trace)

      implicit none

      integer(long)                                    :: helpsize,m,s,d,z,n,f,k
      integer(long)                                    :: nst,s1,kk,ggdim
      integer(long), external                          :: complexsize
      integer(long), dimension(2)                      :: help
      integer(long), dimension(fftdim), intent(in)     :: fftfak
      real(dop), dimension(ortdim,nstate), intent(out) :: gpop1, gpop2
      real(dop), dimension(dvrdim), intent(in)         :: trafo
      real(dop), dimension(d3matdim), intent(out)      :: shtraces
      real(dop), dimension(2*maxgdim)                  :: gp
      real(dop), intent(in)                            :: time,trace
      real(dop)                                        :: tgpop
      real(dop), dimension(ortdim,nddstate*npacktot)   :: tmpgpop1,tmpgpop2
      complex(dop), dimension(dgldim), intent(inout)   :: psi
      complex(dop), dimension(dgldim)                  :: tmppsi,tmppsi1
      complex(dop), dimension(fftdim), intent(in)      :: hin,rueck
      complex(dop), dimension(sdim)                    :: workc
      character(len=c2)                                :: wrtag

      if (lmult .or. .not. lgpel) then
         nst = nstate

      else
         nst = nddstate*npacktot

      endif

      if (lquadpop) then
!        store the psi-array
         helpsize = dgldim*complexsize()
         call alloc(help,helpsize)
         call setblock(tmppsi, help, 0, helpsize )

         call d1trafonat(psi,dicht4)
!        abuse psi to store squared sigma matrices
         do s1=1,nst
! get state dependent densities for single-set calculations
            if (leb .and. lgpel .and. nst .gt. 1) then
               s = 1
               call stproj(s1,s1,psi,tmppsi1)

               if (lrunml) then
                  call compute_dmat_tree(tree,tmppsi)
               else
                  call density1(tmppsi,dicht1,dicht2,dicht3,dicht4)
               endif
            else
               s = s1
               tmppsi = psi
            endif

            do m=1,nmode
               call cpvxd(dicht3(dd3mat(m,s,s)),shtraces(dd3mat(m,s,s)),ddim(m,s,s))

               z=0
               do d=1,ddim(m,s,s)
                  call qqxxzz(tmppsi(dzetf(m,s,s)+z),tmppsi(dzetf(m,s,s)+z),&
                       workc,supgdim(m))
!                 subdim(m)=supgdim(m)^2
                  call cpvxz(workc,tmppsi(dzetf(m,s,s)+z),subdim(m))
                  z=z+subdim(m)
               enddo


               call calcgpop1(tmppsi(dzetf(m,s,s)),tmpgpop1(1,s),tmpgpop2(1,s),&
                    trafo,fftfak,hin,rueck,shtraces(dd3mat(m,1,1)),ddim(m,s,s), &
                    subdim(m),m,s)
            enddo
         enddo

!        restore data and dealloc memory
         call getblock(tmppsi, help, 0, helpsize )
         call dealloc(help, helpsize)

      else                      ! do the old method
         do s1=1,nst

!!get state dependent densities for single-set calculations
            if (leb .and. lgpel .and. nst .gt. 1) then
               s = 1
               call stproj(s1,s1,psi,tmppsi)

               if (lrunml) then
                  call compute_dmat_tree(tree,tmppsi)
               else
                  call density1(tmppsi,dicht1,dicht2,dicht3,dicht4)
               endif
            else
               s = s1
               tmppsi = psi
            endif
!        Calculate Single-hole traces
            call calcsht1(tmppsi,shtraces,s)
!        Calculate grid pops
            do m=1,nmode
               call calcgpop1(psi(dzetf(m,s,s)),tmpgpop1(1,s1),tmpgpop2(1,s1),&
                    trafo,fftfak,hin,rueck,shtraces(dd3mat(m,s,s)),ddim(m,s,s), &
                    subdim(m),m,s)
            enddo
         enddo

      endif

!-----------------------------------------------------------------------
! divide by the trace to get the correct expectation values
!-----------------------------------------------------------------------
      tmpgpop1(:,:) = tmpgpop1(:,:) / trace
      tmpgpop2(:,:) = tmpgpop2(:,:) / trace

!-----------------------------------------------------------------------
! Write Data Set
!-----------------------------------------------------------------------
!     Output grid populations in single precision to dicht file
      if (lgpop) then
         countrho=countrho+1

         write(wrtag,'(a9,i10,a1)') '<data set',countrho,'>'
         write(irho) wrtag

         wrtag='<gpop>'
         write(irho) wrtag
         if (lgpel) then
            do s=1,nst
               do m=1,nmode
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     ggdim = gdim(f)
                     if(basis(f).eq.12 .or.basis(f).eq.20 .or.  &
                        basis(f).eq.21)  &
                         ggdim = ipbaspar(3,f)
                     write(irho) (real(tmpgpop1(zort(f)+k,s)),k=0,gdim(f)-1), &
                                 (real(tmpgpop2(zort(f)+k,s)),k=0,ggdim-1)
                  enddo
               enddo
            enddo

         else
            do m=1,nmode
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  ggdim = gdim(f)
                  if(basis(f).eq.12 .or. basis(f).eq.20 .or.  &
                     basis(f).eq.21)  &
                     ggdim = ipbaspar(3,f)
                  do k = 1, gdim(f)+ggdim
                     gp(k) = 0.0_dop
                  enddo
                  do s=1,nst
                     do k = 1, gdim(f)
                        gp(k) = gp(k) + tmpgpop1(zort(f)+k-1,s)
                     enddo
                     do k = 1, ggdim
                        kk = k + gdim(f)
                        gp(kk) = gp(kk) + tmpgpop2(zort(f)+k-1,s)
                     enddo
                  enddo
                  write(irho) (real(gp(k)),k=1,gdim(f)), &
                              (real(gp(k)),k=gdim(f)+1,gdim(f)+ggdim)
               enddo
            enddo
         endif

         write(wrtag,'(a13,i10,a1)') '<end data set',countrho,'>'
         write(irho) wrtag
         call myflush(irho)
      endif

!-----------------------------------------------------------------------
! sum up grid populations into gpop arrays for single set
! or simply copy for muti-set
!-----------------------------------------------------------------------
      if (leb) then
         gpop1(:,:) = 0.0_dop
         gpop2(:,:) = 0.0_dop
         do s=1,nst
            do k=1,ortdim
               gpop1(k,1) = gpop1(k,1) + tmpgpop1(k,s)
               gpop2(k,1) = gpop2(k,1) + tmpgpop2(k,s)
            enddo
         enddo

      else
         gpop1(:,:) = tmpgpop1(:,:)
         gpop2(:,:) = tmpgpop2(:,:)

      endif

!-----------------------------------------------------------------------
! restore full density matrices for ML-MCTDH
!-----------------------------------------------------------------------
      if (leb .and. lgpel .and. nst .gt. 1) then
         if (lrunml) then
            call compute_dmat_tree(tree,psi)
         else
            call density1(psi,dicht1,dicht2,dicht3,dicht4)
         endif
      endif

!-----------------------------------------------------------------------
! turn on autocap if grid population has entered it
!-----------------------------------------------------------------------
      n=0
      do k=1,khtot(1)
         if (autocap(k) .gt. 0) then
            f=autocap(k)
            n=n+1
            tgpop=0.0_dop
            do s=1,nstate
               tgpop=tgpop+gpop1((zort(f)+indicap(f)-1),s)
            enddo
            call activcap(tgpop,autocap(k),gpopcap(f),timecap(n), &
                          time/fs,ilog,modelabel(f),countpsi,f,n)
         endif
      enddo

      return
      end subroutine outgpop1                      ! of outgpop1

!***********************************************************************
!     CALCGPOP1: Calculates the grid pops gpop1 and gpop2 for one mode 
!                and one state
!
!     07/99 AR
!***********************************************************************

      subroutine calcgpop1(psi,gpop1,gpop2,trafo,fftfak,hin,&
                 rueck,shtraces,dim1,subdim1,m,s)

      implicit none

      integer(long)                                     :: d,f,n
      integer(long), intent(in)                         :: m,s
      integer(long), intent(in)                         :: dim1,subdim1
      integer(long), dimension(fftdim), intent(in)      :: fftfak
      real(dop)                                         :: xx
      real(dop), dimension(ortdim), intent(out)         :: gpop1,gpop2
      real(dop), dimension(dim1), intent(in)            :: shtraces
      real(dop), dimension(dvrdim), intent(in)          :: trafo
      complex(dop), dimension(fftdim), intent(in)       :: hin,rueck
      complex(dop), dimension(subdim1,dim1), intent(in) :: psi
      complex(dop), dimension(maxgdim2,2)               :: workc
      character(len=20)                                 :: file

      file='dicht'

      do n=1,nspfdof(m)
         f=spfdof(n,m)
         call zerovxd(gpop1(zort(f)),gdim(f))
         call zerovxd(gpop2(zort(f)),gdim(f))
         do d=1,ddim(m,s,s)
!     Calculate density for dof f if necessary
            if (nspfdof(m) .gt. 1) then
               call rm1_mf(psi(1,d),workc(1,1),vgdim(f),gdim(f),ngdim(f))
            else
               call cpqxz(psi(1,d),workc(1,1),gdim(f))
            endif
!     Add up gpop1
            call addgpop1(gpop1(zort(f)),workc(1,1),shtraces(d), &
                 gdim(f))
!     Transform density for dof f
            if (basis(f).eq.0) then
               call cpvxd(gpop1(zort(f)),gpop2(zort(f)),gdim(f))
            elseif (basis(f).eq.4) then
               call fft(workc(1,1),workc(1,2),rueck(zfft(f)), &
                    fftfak(zfft(f)),gdim(f),1,gdim(f),.true.)
               call fft(workc(1,1),workc(1,2),hin(zfft(f)), &
                    fftfak(zfft(f)),gdim(f),gdim(f),1,.false.)
               xx = 1.0_dop/dble(gdim(f))
               call xqxxdzo(xx,workc(1,1),gdim(f))
            elseif (ldvr(f)) then
                  if (basis(f) .eq. 5)  then
                     call edtrafo1(workc(1,1),workc(1,2),gdim(f))
                  else
                     call qmxxdz(trafo(zdvr(f)),workc(1,1),workc(1,2), &
                          gdim(f),gdim(f))
                     call mqxtzd(workc(1,2),trafo(zdvr(f)),workc(1,1), &
                          gdim(f),gdim(f))
                  endif
            endif
!     Add up gpop2
            call addgpop1(gpop2(zort(f)),workc(1,1),shtraces(d), &
                 gdim(f))
         enddo
      enddo

      return
      end subroutine calcgpop1


!***********************************************************************
!     EDTRAFO1: Transformation of the density for exp-DVR, the result is
!               stored in a
!
!     07/99 AR
!***********************************************************************

      subroutine edtrafo1(a,b,dim)

      implicit none

      integer(long)                                   :: i,j,k,ds
      integer(long), intent(in)                       :: dim
      real(dop)                                       :: sd
      complex(dop)                                    :: z
      complex(dop), dimension(dim,dim), intent(inout) :: a
      complex(dop), dimension(dim,dim), intent(out)   :: b

      ds = (dim+1)/2
      sd = sqrt(dble(dim))
      z  = (0.0_dop,-8.0_dop)*atan(1.0_dop)/dble(dim)

      do j=1,dim
         do i=1,dim            
            b(i,j) = (exp(z*(i-ds))/sd) * a(1,j)
            do k=2,dim
               b(i,j) = b(i,j) + (exp(z*(i-ds)*k)/sd) * a(k,j)
            enddo
         enddo
      enddo
      do j=1,dim
         do i=1,dim            
            a(i,j) = b(i,1)*conjg(exp(z*(1-ds)*j)/sd)
            do k=2,dim
               a(i,j) = a(i,j) + b(i,k)*conjg(exp(z*(k-ds)*j)/sd)
            enddo
         enddo
      enddo

      return
      end subroutine edtrafo1


!***********************************************************************
!     ADDGPOP1: Adds up the gpops for dof f
!
!     07/99 AR
!***********************************************************************

      subroutine addgpop1(gpop,den,shtrace,gdim)

      implicit none

      integer(long)                                  :: i
      integer(long), intent(in)                      :: gdim
      real(dop), intent(in)                          :: shtrace
      real(dop), dimension(gdim), intent(inout)      :: gpop
      complex(dop), dimension(gdim,gdim), intent(in) :: den

      do i=1,gdim
         gpop(i) = gpop(i) + dble(den(i,i))*shtrace
      enddo

      return
      end subroutine addgpop1


!***********************************************************************
!     CALCSHT1: Calculates traces of Single-hole density operators
!
!     07/99 AR
!***********************************************************************

      subroutine calcsht1(psi,shtraces,s)

      implicit none

      integer(long)                                    :: i,m1,m2,z1,z2,z3,hz
      integer(long), intent(in)                        :: s
      real(dop), dimension(d3matdim), intent(out)      :: shtraces
      real(dop), dimension(d3matdim)                   :: trace
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(maxblock,3)              :: workc

      z1 = 1
      z2 = 2
      z3 = 3
      call cpvxz(psi(dzpsi(s,s)),workc(1,z1),dblock(s,s))

      do m1=1,nmode
         call trspdo(psi(dzetf(m1,s,s)),trace(dd3mat(m1,s,s)),&
              ddim(m1,s,s),subdim(m1),supgdim(m1)) 
      enddo

      if (lrunml) then
         call allocate_mtraces_matrices(tree,trace)
         call compute_mtraces_tree(psi,tree)

         call allocate_shtraces_matrices(tree,shtraces)
         call compute_shtraces_tree(psi,tree)

         call deallocate_mtraces_matrices(tree)
         call deallocate_shtraces_matrices(tree)

      else

         do m1=1,(nmode-1)
            call ptr1(workc(1,z1),trace(dd3mat(m1,s,s)),workc(1,z3),&
                 ddim(m1,s,s),nddim(m1,s,s))

            do m2=(m1+1),nmode
               call ptr2(workc(1,z1),trace(dd3mat(m2,s,s)),workc(1,z2),&
                    ddim(m1,s,s),ddim(m2,s,s),nddim(m2,s,s))
               hz = z1
               z1 = z2
               z2 = hz
            enddo

!     Get traces for mode m1
            do i=1,ddim(m1,s,s)
               shtraces(dd3mat(m1,s,s)-1+i) = dble(workc(i,z1)) 
            enddo
            hz = z1
            z1 = z3
            z3 = hz
         enddo

!     Get traces for mode nmode
         do i=1,ddim(nmode,s,s)
            shtraces(dd3mat(nmode,s,s)-1+i) = dble(workc(i,z1)) 
         enddo
      endif

      return
      end subroutine calcsht1

      end module d1outgpop

