! **********************************************************************
! *                                                                    *
! *                         EVALUATION (evaluation.f)                  *
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! *   prefactor:    Multiplies the wavefunction derivative with -i for *
! *                 propagation or -1 for imaginary time relaxation.   *
! *   subenergy:    Subtracts the energy from the Hamiltonian, i. e.   *
! *                 H|psi> is transformed to (H-E)|psi> where E =      *
! *                 <psi|H|psi>/<psi|psi> (needed for relaxation).     *
! *                                                                    *
! **********************************************************************
      module d1subenergymod

use decimal, only: long, dop
use globmemmod
use global
use maxv, only: maxsta
use griddatmod
use psidef
use runpropmod

use xvlib,   only: xvxxdzr
use expectval, only: calceuncorr1,calcecorr1,calcspop1

      implicit none
      private
      public :: d1subenergy
      contains

! *** V8.0 AR ************> SUBROUTINE SUBENERGY <**********************
!                          ~~~~~~~~~~~~~~~~~~~~~~
!
! Subtracts the energy from the Hamiltonian, i. e. H is transformed
! to (H-E) where E = tr(H rho)/tr(rho) (needed for relaxation).
! dot{B} = {H,B} - 2EB
! Flag lener is only relevant if inttypea .eq. 3
!-----------------------------------------------------------------------

      subroutine d1subenergy(psi,dtpsi,lener)

      implicit none

      integer(long)                                  :: s
      real(dop)                                      :: trace,energy,factor
      real(dop), dimension(maxsta)                   :: spop
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi
      logical(kind=4), intent(in)                    :: lener

!-----------------------------------------------------------------------
! Calculate energies and spops
!-----------------------------------------------------------------------
      if (inttypea .eq. 3) then
         if (lener) then
            call calceuncorr1(psi,mc(mchunpsi),deuncorr,1)
            call calcecorr1(psi,mc(mchpsi),decorr,1)
         endif
         do s=1,nstate
            spop(s) = 0.0_dop
         enddo
      else
         call calcspop1(psi,spop)
         call calceuncorr1(psi,mc(mchunpsi),deuncorr,1)
         call calcecorr1(psi,mc(mchpsi),decorr,1)
      endif

!-----------------------------------------------------------------------
! Calculate trace and energy
!-----------------------------------------------------------------------
      trace  = 0.0_dop
      energy = 0.0_dop
      do s=1,nstate
         if (lfunkr) then
            detot(s) = decorr(s)
            decorr(s) = detot(s)-deuncorr(s)
         else
            detot(s)=decorr(s)+deuncorr(s)
         endif
         trace  = trace + spop(s)
         energy = energy + detot(s)
      enddo

!-----------------------------------------------------------------------
! Substract 2EB, note that prefactor is called later
!-----------------------------------------------------------------------

      if (inttypea .eq. 3) then
         factor = 2.0_dop*energy
      else
         factor = 2.0_dop*energy/trace
      endif
      call xvxxdzr(factor,psi,dtpsi,adim)

      return
      end subroutine d1subenergy

      end module d1subenergymod

