! **********************************************************************
! *                                                                    *
! *                         EVALUATION (evaluation.f)                  *
! *                                                                    *
! * MCTDH module containing (auxiliary) subroutines to evaluate the    *
! * right hand side of the MCTDH working equations.                    *
! *                                                                    *
! * Contains:                                                          *
! * d1summf:        Multiplies for all modes and all states the sum    *
! *                 over the mean-field matrices times the             *
! *                 single-particle potential operator matrices with   *
! *                 |phi>.                                             *
! * d1mfsumphi:     Sums for all modes and all states the product of   *
! *                 the mean-field matrix times the single-particle    *
! *                 potential operator.                                *
! * d1mfsumphi1m:   Sums for one mode and all states the product of    *
! *                 the mean-field matrix times the single-particle    *
! *                 potential operator.                                *
! *                                                                    *
! **********************************************************************
      module d1summfmod

use decimal, only: long, dop
use op1lib,  only: zerovxz, initvxz
use op2lib,  only: addmxxzo, submxxzo1

      implicit none
      private
      public :: d1summf,d1mfsumphi,d1mfsumphi1m
      contains

! *** V6.0 MB ***************> SUBROUTINE SUMMF <***********************
!                             ~~~~~~~~~~~~~~~~~~
!
! Sums up for all modes and all states the product of the mean-field
! matrices times the single-particle potential operators.
! **********************************************************************

      subroutine d1summf (mfsum,hloch)

use global
use griddatmod
use operdef
use psidef
use runpropmod
use hpsimod
use maxv,      only: maxdim,maxsta,maxkoe
use d1hphimod, only: d1hcall
use lalib,     only: mktens

      implicit none

      integer(long)                                    :: m,k,sf,si,t
      complex(dop), dimension(mfsumdim,3), intent(out) :: mfsum
      complex(dop), dimension(hlochdim), intent(in)    :: hloch
      complex(dop), dimension(sgdim*sgdim,2)           :: workc
      logical(kind=4)                                  :: luncorr = .false.

! --- INITIALISE VARIABLES ---

      if (ldissin) then
         call zerovxz(mfsum,3*mfsumdim)
      else
         call zerovxz(mfsum,mfsumdim)
      endif

! --- LOOP OVER EACH MODE AND EACH CORRELATED HAMILTONIAN TERM ---
! --- SUM UP MEAN-FIELD TIMES SINGLE-PARTICLE OPERATOR ---
      do m = 1,nmode
         do k = 1,kzahl(1)
            if (kinop(m,k).eq.0 .and. .not. diag(m,k)) then
!     Hamiltonian action
               if (kact(k).eq.1) then
!     makes a copy of the diagonal hamiltonian matrix h_(m,k)
                  call initvxz(workc(1,1),supgdim(m))
                  call d1hcall(m,k,workc(1,2),workc(1,1),supgdim(m),1,luncorr)
                  si = ki(k)
                  sf = kf(k)
                  do t=1,nstate
                     call mktens(mfsum(dzmfsum(m,sf,si,t),1),workc(1,2), &
                          hloch(dphloch(m,k,t)),supgdim(m),ddim(m,sf,t), &
                          ddim(m,si,t),true)
                  enddo
               endif
!     Pure left action     
               if (kact(k).eq.2) then
!     makes a copy of the diagonal matrix h_(m,k)
                  call initvxz(workc(1,1),subdim(m))
                  call d1hcall(m,k,workc(1,2),workc(1,1),supgdim(m),1,luncorr)
                  si = ki(k)
                  sf = kf(k)
                  do t=1,nstate
                     call mktens(mfsum(dzmfsum(m,sf,si,t),2),workc(1,2), &
                          hloch(dphloch(m,k,t)),supgdim(m),ddim(m,sf,t), &
                          ddim(m,si,t),true)
                  enddo
               endif
!     Pure right action     
               if (kact(k).eq.3) then
!     makes a copy of the diagonal matrix h_(m,k)
                  call initvxz(workc(1,1),subdim(m))
                  call d1hcall(m,k,workc(1,2),workc(1,1),supgdim(m),1,luncorr)
                  si = ki(k)
                  sf = kf(k)
                  do t=1,nstate
                     call mktens(mfsum(dzmfsum(m,si,sf,t),3),workc(1,2), &
                          hloch(dphloch(m,k,t)),supgdim(m),ddim(m,t,si), &
                          ddim(m,t,sf),true)
                  enddo
               endif
            endif
         enddo
      enddo

      return
      end subroutine d1summf 

! *** V6.0 MB *************> SUBROUTINE MFSUMPHI <**********************
!                           ~~~~~~~~~~~~~~~~~~~~~

! Multiplies for all modes and all states the sum over the products of
! the mean-field matrices times the single-particle potential operator
! matrices with |phi>.

      subroutine d1mfsumphi (dtpsi,psi,mfsum)

use global
use griddatmod
use operdef
use psidef
use runpropmod
use hpsimod

      implicit none

      integer(long)                                       :: m
      complex(dop), dimension(dgldim), intent(inout)      :: dtpsi
      complex(dop), dimension(dgldim), intent(in)         :: psi
      complex(dop), dimension(mfsumdim), intent(in)       :: mfsum

! --- LOOP OVER EACH MODE AND STATE---

      do m = 1,nmode
         call d1mfsumphi1m(dtpsi(dzetf(m,1,1)),psi(dzetf(m,1,1)), &
              mfsum,m,dzetf(m,1,1),phidim(m))
      enddo

      return
      end subroutine d1mfsumphi

! *** V6.0 MB *************> SUBROUTINE MFSUMPHI 1M <*******************
!                           ~~~~~~~~~~~~~~~~~~~~~~~~

! Multiplies for one mode and all states the sum over the products of
! the mean-field matrices times the single-particle potential operator
! matrices with |phi>.

      subroutine d1mfsumphi1m (dtpsi,psi,mfsum,m,dzetf1,phidim1)

use global
use griddatmod
use operdef
use psidef
use runpropmod
use hpsimod
use d1tens, only: tenstens,makeadj,tenstens1,tenstens2
use maxv,   only: maxdim,maxsta,maxkoe

      implicit none

      integer(long)                                                   :: sf,si,t,z1,z2
      integer(long), intent(in)                                       :: m,dzetf1,phidim1
      complex(dop), dimension(dzetf1:dzetf1+phidim1-1), intent(inout) :: dtpsi
      complex(dop), dimension(dzetf1:dzetf1+phidim1-1), intent(in)    :: psi
      complex(dop), dimension(mfsumdim,3), intent(in)                 :: mfsum
      complex(dop), dimension(2*sdim*maxspf)                          :: workc

! --- IF BASIS IS COMPLETE, DO NOT CALCULATE ---

      if (complete(m)) return

! --- LOOP OVER EACH COMBINATION OF STATES ---

      do sf=1,nstate
         do si=1,nstate
            do t=1,nstate

!       --- MULTIPLY SUM OVER MEAN-FIELDS WITH |PHI> ---

!     Hamiltonian action
               call tenstens(workc,mfsum(dzmfsum(m,sf,si,t),1), &
                    psi(dzetf(m,si,t)),supgdim(m),ddim(m,si,t), &
                    ddim(m,sf,t))
               call addmxxzo(workc,dtpsi(dzetf(m,sf,t)),subdim(m), &
                    ddim(m,sf,t))
!     ddim(m,sf,t) = ddim(m,t,sf)
               z1 = 1
               z2 = subdim(m)*ddim(m,sf,t) + 1
               call makeadj(workc(z1),workc(z2),supgdim(m), &
                    ddim(m,sf,t))
               if (lrlx) then
                  call addmxxzo(workc(z2),dtpsi(dzetf(m,t,sf)), &
                       subdim(m),ddim(m,t,sf))               
               else
                  call submxxzo1(dtpsi(dzetf(m,t,sf)),workc(z2), &
                       subdim(m),ddim(m,t,sf))
               endif

               if (ldissin) then
!     Left action
                  call tenstens1(dtpsi(dzetf(m,sf,t)), &
                       mfsum(dzmfsum(m,sf,si,t),2),psi(dzetf(m,si,t)), &
                       supgdim(m),ddim(m,si,t),ddim(m,sf,t))
!     Right action
                  call tenstens2(dtpsi(dzetf(m,t,sf)), &
                       mfsum(dzmfsum(m,si,sf,t),3),psi(dzetf(m,t,si)), &
                       supgdim(m),ddim(m,t,si),ddim(m,t,sf))
               endif
            enddo
         enddo
      enddo

      return
      end subroutine d1mfsumphi1m

!#######################################################################

      end module d1summfmod

