!#######################################################################
!
!                      MCTDH module D2OUTGPOP
!
!
! subroutines :
!    OUTGPOP2 : Calculates grid populations in coordinate and 
!               momentum representations and outputs results 
!    
!
!  V81 AR
!
!#######################################################################

      module d2outgpop

use decimal, only: long, dop
use mmlib,   only: qmxxdz, mqxtzd
use xvlib,   only: xqxxdzo
use fftmod,  only: fft
use op1lib,  only: zerovxd, zeromxz, cpvxd

use global
use globmemmod
use dvrdatmod
use griddatmod 
use psidef
use operdef
use datenmod
use runpropmod
use channels
use maxv,   only: maxdim,maxsta,maxkoe
use daten,  only: activcap
use d2tens, only: rm1_mf
use hpsimod
use densitymod, only: density2
use eqofmotion1, only: compute_dmat_tree
use dichtoutmod, only: stproj,gwp2grid

      implicit none
      private
      public :: outgpop2
      contains

!#######################################################################
!
! OUTGPOP2:  Calculates grid populations in coordinate and momentum
!            representations and outputs results to dicht file.
!
! Passed back to calling routine:
! gpop1(m,s): first set of grid populations 
! gpop2(m,s): second set of grid populations 
!
! gpop1 contains:
!   if time independent basis, basis function population
!   if DVR of FFT, coordinate representation populations
! gpop2 contains:
!   if time independent basis, basis function population (identical to 
!      gpop1)
!   if DVR, FBR basis function occupation numbers
!   if FFT, momentum representation populations
!
!
! V80  AR
!***********************************************************************

      subroutine outgpop2(psi,gpop1,gpop2,trafo,fftfak,hin,rueck,time,&
                 ort,dicht3,agmat,trace)

      implicit none

      integer(long)                                    :: m,s,k,n,f
      integer(long)                                    :: nst,s1,kk,ggdim
      integer(long), dimension(fftdim), intent(in)     :: fftfak
      real(dop)                                        :: tgpop
      real(dop), intent(in)                            :: time,trace
      real(dop), dimension(ortdim,nstate), intent(out) :: gpop1,gpop2
      real(dop), dimension(ortdim,nddstate*npacktot)   :: tmpgpop1,tmpgpop2
      real(dop), dimension(dvrdim), intent(in)         :: trafo
      real(dop), dimension(ortdim), intent(in)         :: ort
      real(dop), dimension(d3matdim), intent(inout)    :: dicht3
      real(dop), dimension(2*maxgdim)                  :: gp
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim)                  :: tmppsi
      complex(dop), dimension(fftdim), intent(in)      :: hin,rueck
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      logical(kind=4)                                  :: lmf0
      character(len=c2)                                :: wrtag

      if (lmult .or. .not. lgpel) then
         nst = nstate

      else
         nst = nddstate*npacktot

      endif

! get "mean-field" density matrices for DF propagation
      if (.not. lmf .and. &
          .not. (leb .and. lgpel .and. nst .gt. 1)) then
         lmf = .true.
         if (lrunml) then
            call compute_dmat_tree(tree,psi)
         else
            call density2(psi,dicht1,dicht2,dicht3,dicht4,.false.)
         endif
         lmf = .false.
      endif

      tmpgpop1 = 0.0_dop
      tmpgpop2 = 0.0_dop
      do s1=1,nst
! get state dependent densities for single-set calculations
         if (leb .and. lgpel .and. nst .gt. 1) then
            s = 1
            call stproj(s1,s1,psi,tmppsi)

            lmf0 = lmf
            lmf = .true.
            if (lrunml) then
               call compute_dmat_tree(tree,tmppsi)
            else
               call density2(tmppsi,dicht1,dicht2,dicht3,dicht4,.false.)
            endif
            lmf = lmf0

         else
            s = s1

         endif

         do m=1,nmode
            if (gwpm(m)) then
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  call gwp2grid(n,f,agmat(zagmat(m,s)),nspfdof(m), &
                       agmat(zxivec(m,s)),agmat(zphase(m,s)),dim(m,s), &
                       gdim(f),ort(zort(f)),dicht2(dmat(m,s)), &
                       tmpgpop1(zort(f),s1),psi(zdgwp(m,s)))
               enddo

            else
               call d2calcgpop(psi(zetf(m,s)),tmpgpop1(1,s1),tmpgpop2(1,s1), &
                    trafo,fftfak,hin,rueck,dicht2(dmat(m,s)), &
                    dim(m,s),subdim(m),m,s)
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! divide by the trace to get the correct expectation values
!-----------------------------------------------------------------------
      tmpgpop1(:,:) = tmpgpop1(:,:) / trace
      tmpgpop2(:,:) = tmpgpop2(:,:) / trace

!-----------------------------------------------------------------------
! Write Data Set
!-----------------------------------------------------------------------
!     Output grid populations in single precision to dicht file
      if (lgpop) then
         countrho=countrho+1

         write(wrtag,'(a9,i10,a1)') '<data set',countrho,'>'
         write(irho) wrtag

         wrtag='<gpop>'
         write(irho) wrtag
         if (lgpel) then
            do s=1,nst
               do m=1,nmode
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     ggdim = gdim(f)
                     if(basis(f).eq.12 .or.basis(f).eq.20 .or.  &
                        basis(f).eq.21) ggdim = ipbaspar(3,f)
                     write(irho) (real(tmpgpop1(zort(f)+k,s)),k=0,gdim(f)-1), &
                                 (real(tmpgpop2(zort(f)+k,s)),k=0,ggdim-1)
                  enddo
               enddo
            enddo

         else
            do m=1,nmode
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  ggdim = gdim(f)
                  if(basis(f).eq.12 .or. basis(f).eq.20 .or.  &
                     basis(f).eq.21)  ggdim = ipbaspar(3,f)
                  do k = 1, gdim(f)+ggdim
                     gp(k) = 0.0_dop
                  enddo
                  do s=1,nst
                     do k = 1, gdim(f)
                        gp(k) = gp(k) + tmpgpop1(zort(f)+k-1,s)
                     enddo
                     do k = 1, ggdim
                        kk = k + gdim(f)
                        gp(kk) = gp(kk) + tmpgpop2(zort(f)+k-1,s)
                     enddo
                  enddo
                  write(irho) (real(gp(k)),k=1,gdim(f)), &
                              (real(gp(k)),k=gdim(f)+1,gdim(f)+ggdim)
               enddo
            enddo
         endif

         write(wrtag,'(a13,i10,a1)') '<end data set',countrho,'>'
         write(irho) wrtag
         call myflush(irho)
      endif

!-----------------------------------------------------------------------
! sum up grid populations into gpop arrays for single set
! or simply copy for muti-set
!-----------------------------------------------------------------------
      if (leb) then
         gpop1(:,:) = 0.0_dop
         gpop2(:,:) = 0.0_dop
         do s=1,nst
            do k=1,ortdim
               gpop1(k,1) = gpop1(k,1) + tmpgpop1(k,s)
               gpop2(k,1) = gpop2(k,1) + tmpgpop2(k,s)
            enddo
         enddo

      else
         gpop1(:,:) = tmpgpop1(:,:)
         gpop2(:,:) = tmpgpop2(:,:)

      endif

!-----------------------------------------------------------------------
! restore full density matrices for ML-MCTDH
!-----------------------------------------------------------------------
      if ((leb .and. lgpel .and. nst .gt. 1) .or. .not. lmf) then
         if (lrunml) then
            call compute_dmat_tree(tree,psi)
         else
            call density2(psi,dicht1,dicht2,dicht3,dicht4,.false.)
         endif
      endif

!-----------------------------------------------------------------------
! turn on autocap if grid population has entered it
!-----------------------------------------------------------------------
      n=0
      do k=1,khtot(1)
         if (autocap(k) .gt. 0) then
            f=autocap(k)
            n=n+1
            tgpop=0.0_dop
            do s=1,nstate
               tgpop=tgpop+gpop1((zort(f)+indicap(f)-1),s)
            enddo
            call activcap(tgpop,autocap(k),gpopcap(f),timecap(n), &
                          time/fs,ilog,modelabel(f),countpsi,f,n)
         endif
      enddo


      return
      end subroutine outgpop2


!***********************************************************************
!     D2CALCGPOP: Calculates the grid pops gpop1 and gpop2 for one mode 
!               and one state
!
!     07/99 AR
!***********************************************************************

      subroutine d2calcgpop(psi,gpop1,gpop2,trafo,fftfak,hin, &
                 rueck,den,dim1,subdim1,m,s)

      implicit none

      integer(long)                                     :: i,j,f,n
      integer(long), intent(in)                         :: m,s,dim1,subdim1
      integer(long), dimension(fftdim), intent(in)      :: fftfak
      real(dop)                                         :: xx
      real(dop), dimension(ortdim), intent(out)         :: gpop1,gpop2
      real(dop), dimension(dvrdim), intent(in)          :: trafo
      complex(dop), dimension(subdim1,dim1), intent(in) :: psi
      complex(dop), dimension(fftdim), intent(in)       :: hin,rueck
      complex(dop), dimension(dim1,dim1), intent(inout) :: den
      complex(dop), dimension(maxgdim2,2)               :: gden

      do n=1,nspfdof(m)

         f=spfdof(n,m)
         call zerovxd(gpop1(zort(f)),gdim(f))
         call zerovxd(gpop2(zort(f)),gdim(f))

!     Calculate grid density for dof f 
         call zeromxz(gden(1,1),gdim(f),gdim(f))
         do i=1,dim(m,s)
            do j=1,dim(m,s)
               den(i,j) = den(i,j)
               call addgden(psi(1,i),psi(1,j),gden(1,1),den(i,j), &
                    vgdim(f),gdim(f),ngdim(f),f)
            enddo
         enddo
!     Calculate gpop1
         call d2calcgpop1f(gpop1(zort(f)),gden(1,1),gdim(f))
!     Transform grid density for dof f and calculate gpop2
         if (basis(f).eq.0) then          ! el-state
            call cpvxd(gpop1(zort(f)),gpop2(zort(f)),gdim(f))
         elseif (basis(f).eq.4) then      ! FFT
            call fft(gden(1,1),gden(1,2),hin(zfft(f)), &
                 fftfak(zfft(f)),gdim(f),1,gdim(f),.false.)
            call fft(gden(1,1),gden(1,2),rueck(zfft(f)), &
                 fftfak(zfft(f)),gdim(f),gdim(f),1,.true.)
            xx = 1.0_dop/dble(gdim(f)) 
            call xqxxdzo(xx,gden(1,1),gdim(f))
            call d2calcgpop1f(gpop2(zort(f)),gden(1,1),gdim(f))
         elseif (ldvr(f)) then 
            if (basis(f) .eq. 5)  then    !  exp-DVR
               call edtrafo(gden(1,1),gden(1,2),gdim(f))
               call d2calcgpop1f(gpop2(zort(f)),gden(1,1),gdim(f))
            else
               call qmxxdz(trafo(zdvr(f)),gden(1,1),gden(1,2), &
                    gdim(f),gdim(f))
               call mqxtzd(gden(1,2),trafo(zdvr(f)),gden(1,1), &
                    gdim(f),gdim(f))
               call d2calcgpop1f(gpop2(zort(f)),gden(1,1),gdim(f))
            endif
         endif
      enddo


      return
      end subroutine d2calcgpop


!***********************************************************************
!     MKGDEN: Calculates 'grid density'
!
!     10/99 AR
!***********************************************************************

      subroutine addgden(psi1,psi2,gden,den,vgdim1,gdim1,ngdim1,f)

      implicit none

      integer(long)                                            :: i,j,n,v
      integer(long), intent(in)                                :: f,vgdim1,gdim1,ngdim1
      complex(dop)                                             :: hgden
      complex(dop), intent(in)                                 :: den
      complex(dop), dimension(gdim1,gdim1), intent(inout)      :: gden
      complex(dop), dimension(vgdim1,gdim1,ngdim1), intent(in) :: psi1
      complex(dop), dimension(vgdim1,gdim1,ngdim1), intent(in) :: psi2

      do j=1,gdim(f)
         do i=1,gdim(f)
            hgden = (0.0_dop,0.0_dop)
            do n=1,ngdim(f)
               do v=1,vgdim(f)
                  hgden = hgden + conjg(psi1(v,i,n))*psi2(v,j,n)
               enddo
            enddo
            gden(i,j) = gden(i,j) + hgden*den
         enddo
      enddo

      return
      end subroutine addgden
                  

!***********************************************************************
!     D2CALCGPOP1F: Adds up the gpops for dof f
!
!     10/99 AR
!***********************************************************************
      subroutine d2calcgpop1f(gpop,gden,gdim)

      implicit none

      integer(long)                                  :: i
      integer(long), intent(in)                      :: gdim
      real(dop), dimension(gdim), intent(inout)      :: gpop
      complex(dop), dimension(gdim,gdim), intent(in) :: gden

      do i=1,gdim
         gpop(i) = gpop(i) + dble(gden(i,i))
      enddo

      return
      end subroutine d2calcgpop1f


!***********************************************************************
!     EDTRAFO: Transformation of the density for exp-DVR, the result is
!              stored in a
!
!     07/99 AR
!***********************************************************************
      subroutine edtrafo(a,b,dim)

      implicit none

      integer(long)                                   :: i,j,k,ds
      integer(long), intent(in)                       :: dim
      real(dop)                                       :: sd
      complex(dop)                                    :: z
      complex(dop), dimension(dim,dim), intent(out)   :: b
      complex(dop), dimension(dim,dim), intent(inout) :: a

      ds = (dim+1)/2
      sd = sqrt(dble(dim))
      z  = dcmplx(0.0_dop,-8.0_dop*atan(1.0_dop)/dble(dim))

      do j=1,dim
         do i=1,dim            
            b(i,j) = (exp(z*(i-ds))/sd) * a(1,j)
            do k=2,dim
               b(i,j) = b(i,j) + (exp(z*(i-ds)*k)/sd) * a(k,j)
            enddo
         enddo
      enddo
      do j=1,dim
         do i=1,dim            
            a(i,j) = b(i,1)*conjg(exp(z*(1-ds)*j)/sd)
            do k=2,dim
               a(i,j) = a(i,j) + b(i,k)*conjg(exp(z*(k-ds)*j)/sd)
            enddo
         enddo
      enddo

      return
      end subroutine edtrafo

      end module d2outgpop

