! **********************************************************************
! *                                                                    *
! *                          function (function.f)                     *
! *                                                                    *
! * mctdh module that calls the routines that compute the right hand   *
! * side of the mctdh working equations 
!
!   when the mean fields are held 'constant':
!       funk0, funk1, funk2, funk3, funkphi, funka.
!   when the single-particle hamiltonian is rhs of the projector:
!       funkr
!   when the single-particle hamiltonian is lhs of the projector:
!       funk
!
! contains: funkphi2,funka2,funkex,qdot,qdot1m,xdot,xdot1m,deriv1,
! **********************************************************************
      module functionrw

use decimal, only: long, dop
use globmemmod
use logdat
use timing

      implicit none
      private
      public :: funkrwp,chebsqrt,hoprw
              
      contains

!***********************************************************************
! 
!                           subroutine funkrwp
!
! computes the time derivative of the complete wavefunction. the
! projector is applied before the uncorrelated hamiltonian terms are
! added.
!
! called from: propagatepsi, abm, bs
! calls to:    density, hphi, phihphi, mfields, hlochphi, project,
!              dicht1phi, hunphi
!              prefactor
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine funkrwp (psit1,psit3,mr)

use global
use griddatmod
use runpropmod
use operdef
use psidef
use tdkoeffs, only: calctdk

use op1lib, only: zerovxd

      implicit none

      real(dop), dimension(*), intent(in)          :: mr
      real(dop), dimension(totphidim), intent(in)  :: psit1 
      real(dop), dimension(totphidim), intent(out) :: psit3
      integer(long), save                             :: tid=0 ! timer ID

      call get_timer(tid, 'Funkrw')
      call start_timer(tid)

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call zerovxd(psit3,totphidim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
!      nham=1
!      if (ltdh(nham)) call calctdk(time,nham,psi,mc,mr,mi,ml)

! --- CALCULATE RIGHT HAND SIDE FOR HAMILTONIAN ---

!         hktype=0
!         call hunphi1rw(psit1,psit3,1,1,dim(1,1),subdim(1),mr(mrtrafo), &
!                        mr(mrhr),mi(mihi),ml(mlhl),hktype,1) 
         call hunphi1rw(psit1,psit3,1,1,dim(1,1),subdim(1),mr)

      call stop_timer(tid)

      return
      end subroutine funkrwp

!#######################################################################
! *** V6.6 GW *************> SUBROUTINE HUNPHI 1RW <********************
!                           ~~~~~~~~~~~~~~~~~~~~~~~
!
! Computes for one mode and one state the action of the uncorrelated
! Hamiltonian on the single-particle functions.
!
! hktype governs which terms are included:
! hktype = 0 : all terms
! hktype = 1 : exclude CAP terms
! hktype = 2 : only CAP terms
!
! nham      : Number of operator to be used
!
!#######################################################################

      subroutine hunphi1rw (psit1,psit3,m,s,dim1,sdm,mr)

use global
use maxv, only: maxkoe
use dvrdatmod
use griddatmod
use psidef
use operdef

use realwpmod
use op1lib, only: zeromxd
use op2lib, only: addmxxdo
!use xvlib, only: xmxxzza

      implicit none

      integer(long), parameter                       :: nham=1
      integer(long)                                  :: k,k1,k2,h
      integer(long), intent(in)                      :: m,s,dim1,sdm
      logical(kind=4), dimension(maxkoe)             :: kcalc

      real(dop), dimension(sdm,dim1), intent(in)    :: psit1
      real(dop), dimension(sdm,dim1), intent(out)   :: psit3
      real(dop), dimension(sdm,dim1)                :: workr

      real(dop), dimension(*), intent(in)            :: mr

      call zeromxd(workr,sdm,dim1) ! do we need this???

!-----------------------------------------------------------------------
! first find which terms need to be calculated
!-----------------------------------------------------------------------
      if(khzahl(m,s,nham).eq.0) return
      k1=zhun(m,s,nham)
      k2=zhun(m,s,nham)+khzahl(m,s,nham)-1
outer:do k=k1,k2
         kcalc(k)=.true.
         if ((autoabs(k) .gt. 0) .or. (autoflux(k) .gt. 0)) kcalc(k)=.false.
      enddo outer

! --- LOOP OVER EACH UNCORRELATED HAMILTONIAN TERM ---

      do k=k1,k2
         if (kcalc(k)) then
            call hphi1krw(psit1,workr,m,k,dim1,sdm,mr)

            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then ! do we need this??? yes!!!
               call addmxxdo(workr,psit3,sdm,dim1)
!            else
!               call xmxxzza(koeff(k),workr,psit3,sdm,dim1) ! now it is not working, if we need then we have to change to real mode!!!
            endif

         endif
      enddo

      return
      end subroutine hunphi1rw

!#######################################################################

! *** V6.6 GW **************> SUBROUTINE HPHI 1KRW <*********************
!                            ~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and ONE Hamiltonian term the action
! of the correlated Hamiltonian on the single-particle functions.
!#######################################################################

      subroutine hphi1krw (psit1,hpsi,m,k,dim,subdim,mr)

!use griddatmod

      implicit none

      integer(long)                                    :: e
      integer(long), intent(in)                        :: k,dim,subdim,m
      real(dop), dimension(subdim,dim), intent(in)     :: psit1
      real(dop), dimension(subdim,dim), intent(out)    :: hpsi
      real(dop), dimension(*), intent(in)              :: mr

! --- LOOP OVER EACH HAMILTONIAN TERM ---

      do e = 1,dim
         call hcallrw(m,k,hpsi(1,e),psit1(1,e),subdim,mr)
      enddo

      return
      end subroutine hphi1krw


!#######################################################################

      subroutine hcallrw(m,k,hpsi,psit1,subdim1,mr)

use global
use griddatmod
use psidef
use hpsimod
use operdef

use realwpmod

use op1lib, only: cpvxd

      implicit none

      integer(long)                                 :: f,n,n1,h 
      integer(long), intent(in)                     :: m,k,subdim1
      real(dop), dimension(subdim1), intent(in)     :: psit1
      real(dop), dimension(subdim1), intent(out)    :: hpsi
      real(dop), dimension(subdim1)                 :: workr
      real(dop), dimension(*), intent(in)       :: mr


      if (lrotcoup(k)) then
         do n1=1,nspfdof(m)
            f=spfdof(n1,m)
            h=hamilton(f,k)
            if (htmtype(h) /= 1) then
               call hoprw(f,h,k,hpsi,psit1,mr(mrhops),mr(mrtrafo),subdim1)
            endif
         enddo
      else
         n=1
         f=spfdof(n,m)
         h=hamilton(f,k)
         call hoprw(f,h,k,hpsi,psit1,mr(mrhops),mr(mrtrafo),subdim1)

!     if( htmdof(h) .eq. 0 ) return
!     NEWKLEG: the above is just an optimization,
!     for multi-mode operators the following loop won't actually
!     do anything because hamilton(f,k) is the unit operator for
!     all but the first DOF of the mode.
!     But if multiple KLegs are in one mode, this optimization
!     would skip the "later" KLeg operators.

         do n1=2,nspfdof(m)
            f=spfdof(n1,m)
            h=hamilton(f,k)
            if (htmtype(h) .ne. 1) then
               call cpvxd(hpsi,workr,subdim1)
               call hoprw(f,h,k,hpsi,workr,mr(mrhops),mr(mrtrafo),subdim1)
            endif
         enddo
      endif

      return
      end subroutine hcallrw

!***********************************************************************

      subroutine hoprw (f,h,k,hpsi,psi,hops,trafo,subdim1)

use global
use dvrdatmod
use griddatmod
use psidef
use operdef

use operutil, only: getvgn
use realwpmod

use op1lib, only: cpvxd
use xvlib, only: dvxxdd
use mtlib, only: dtxxdd,qtxxdd

      implicit none

      integer(long)                                 :: h,vd,gd,nd,n,f1,rdim
      integer(long), intent(in)                     :: subdim1,f,k
      real(dop), dimension(hopsdim), intent(in)     :: hops
      real(dop), dimension(subdim1), intent(in)     :: psi
      real(dop), dimension(dvrdim), intent(in)      :: trafo
      real(dop), dimension(subdim1), intent(out)    :: hpsi
      real(dop), dimension(subdim1)                 :: workr1,workr2
      integer(long)                                 :: tid=0 ! timer ID
                                                             
!-----------------------------------------------------------------------
! Go to selected operator
!-----------------------------------------------------------------------

     select case (htmtype(h))

     case default
      routine='Hoprw'
      write(message,'(a,i3,a,3i5)') 'Unknown htmtype :',htmtype(h), &
                                    ',  h, k, f', h,k,f
      call errormsg

     case (1)!     unit matrix
      call cpvxd(psi,hpsi,subdim1)
      return

     case (2)!     real vector (diagonal matrix)

      if(lrotvec(k)) then

         tid=0
         call get_timer(tid, 'Funk2j')
         call start_timer(tid)

outer:  do n=1,nspfdof(1)
           f1=spfdof(n,1)
           if (basis(f1) == 2) then
              rdim=totphidim/gdim(f1) ! set from basis condition!!!
! DVR to FBR
              call DGEMM('N','T',rdim,gdim(f1),gdim(f1),oner,psi,&
                   rdim,trafo(zdvr(f1)),gdim(f1),0.0_dop,workr1,rdim)

              call dvxxdd(hops(zhop(h)),workr1,workr2,totphidim) 

! FBR to DVR
              call DGEMM('N','N',rdim,gdim(f1),gdim(f1),oner,workr2,&
                   rdim,trafo(zdvr(f1)),gdim(f1),oner,hpsi,rdim)
              exit outer

           else if (basis(f1)==7 .or. basis(f1)==20) then
              call klegfbrdvr(gdim(f1),gdim(f1+1),psi,trafo(zdvr(f1)),workr1,onei)

              call dvxxdd(hops(zhop(h)),workr1,workr2,totphidim) 

! FBR to DVR
              call klegfbrdvr(gdim(f1),gdim(f1+1),workr2,trafo(zdvr(f1)),hpsi,twoi)

              exit outer
           endif
        enddo outer


         call stop_timer(tid)

      else if (lrotcoup(k)) then  

         tid=0
         call get_timer(tid, 'Funk2k')
         call start_timer(tid)

outer1: do n=1,nspfdof(1)
           f1=spfdof(n,1)
           if (basis(f1)==7 .or. basis(f1)==20) then
! DVR to FBR
              call klegfbrdvr(gdim(f1),gdim(f1+1),psi,trafo(zdvr(f1)),workr1,onei)
              if (nrotcoup(k) == 1) then
                 call jplusminuspsirw(hops(zhop(h)),workr1,workr2,vgdim(f1),gdim(f1),&
                                      gdim(f1+1),ngdim(f1+1),onei)
              else if (nrotcoup(k) == 2) then
                 call jpluspsirw(hops(zhop(h)),workr1,workr2,vgdim(f1),gdim(f1),&
                                      gdim(f1+1),ngdim(f1+1),onei)
              else if (nrotcoup(k) == 3) then
                 call jminuspsirw(hops(zhop(h)),workr1,workr2,vgdim(f1),gdim(f1),&
                                       gdim(f1+1),ngdim(f1+1),onei)
              else if (nrotcoup(k) == 4) then
                 call jpluspsirw(hops(zhop(h)),workr1,workr2,vgdim(f1),gdim(f1),&
                                       gdim(f1+1),ngdim(f1+1),twoi)
              else if (nrotcoup(k) == 5) then
                 call jminuspsirw(hops(zhop(h)),workr1,workr2,vgdim(f1),gdim(f1),&
                                       gdim(f1+1),ngdim(f1+1),twoi)
              endif
! FBR to DVR
              call klegfbrdvr(gdim(f1),gdim(f1+1),workr2,trafo(zdvr(f1)),hpsi,twoi)

              exit outer1
           endif
        enddo outer1

         call stop_timer(tid)

      else   

         tid=0
         call get_timer(tid, 'Funk2p')
         call start_timer(tid)

         call getvgn(h,vd,gd,nd)
         call dtxxdd(hops(zhop(h)),psi,hpsi,vd,gd,nd) 

         call stop_timer(tid)
      endif
         
      return

     case (8)!     real matrix 

      tid=0
      call get_timer(tid, 'Funkrw8')
      call start_timer(tid)

      call getvgn(h,vd,gd,nd)
      call qtxxdd(hops(zhop(h)),psi,hpsi,vd,gd,nd)

      call stop_timer(tid)

      return

      end select

      end subroutine hoprw

!#######################################################################
      subroutine klegfbrdvr(gdim1,kdim,work1,trafo,work2,ip)

use global
use psidef

use op1lib, only: zerovxd

      implicit none

      integer(long)                                     :: k,rdim,zeig,zeig1
      integer(long), intent(in)                         :: ip,gdim1,kdim
      real(dop), dimension(totphidim), intent(in)       :: work1
      real(dop), dimension(totphidim), intent(out)      :: work2
      real(dop), dimension(gdim1*gdim1*kdim), intent(in):: trafo

   call zerovxd(work2,totphidim) 

   rdim=totphidim/(gdim1*kdim)
   zeig=1
   zeig1=1
   if (ip == 1) then
      do k=1,kdim
         call DGEMM('N','T',rdim,gdim1,gdim1,oner,work1(zeig),& 
                    rdim,trafo(zeig1),gdim1,0.0_dop,work2(zeig),rdim)

         zeig=zeig+rdim*gdim1
         zeig1=zeig1+gdim1*gdim1
      enddo
   else if (ip == 2) then
      do k=1,kdim
         call DGEMM('N','N',rdim,gdim1,gdim1,oner,work1(zeig),&
                    rdim,trafo(zeig1),gdim1,oner,work2(zeig),rdim)

         zeig=zeig+rdim*gdim1
         zeig1=zeig1+gdim1*gdim1
      enddo

   endif 

   return

   end subroutine klegfbrdvr

!-----------------------------------------------------------------------

      subroutine jpluspsirw (jpl,psi,jplpsi,vgdim,gdim,kdim, &
                             ngdim,m)

      implicit none

      integer(long)                                            :: g,v,n,k
      integer(long), intent(in)                                :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: jpl
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jplpsi

      do n=1,ngdim
         do k=1,m
            do g=1,gdim
               do v=1,vgdim
                  jplpsi(v,g,k,n)=0.0_dop
               enddo
            enddo
         enddo
      enddo

      do n=1,ngdim
         do k=m+1,kdim
            do g=1,gdim
               do v=1,vgdim
                  jplpsi(v,g,k,n)=jpl(v,g,k-m,n)*psi(v,g,k-m,n)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine jpluspsirw

!-----------------------------------------------------------------------

      subroutine jminuspsirw (jmin,psi,jminpsi,vgdim,gdim,kdim, &
                              ngdim,m)

      implicit none

      integer(long)                                            :: g,v,n,k
      integer(long), intent(in)                                :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: jmin
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jminpsi


      do n=1,ngdim
         do k=1,kdim-m
            do g=1,gdim
               do v=1,vgdim
                  jminpsi(v,g,k,n)=jmin(v,g,k+m,n)*psi(v,g,k+m,n)
               enddo
            enddo
         enddo
      enddo

      do n=1,ngdim
         do k=kdim-m+1,kdim
            do g=1,gdim
               do v=1,vgdim
                  jminpsi(v,g,k,n)=0.0_dop
               enddo
            enddo
         enddo
      enddo


      return
      end subroutine jminuspsirw


!-----------------------------------------------------------------------


      subroutine jplusminuspsirw (jplmin,psi,jplminpsi,vgdim,gdim,kdim, &
                                  ngdim,m)

      implicit none

      integer(long)                                               :: g,v,n,k
      integer(long), intent(in)                                   :: m,gdim,kdim,vgdim,ngdim
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)            :: jplmin
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(in)  :: psi
      real(dop), dimension(vgdim,gdim,kdim,ngdim), intent(out) :: jplminpsi


!---- for k=1,...,m  and  k=kdim,...,kdim-m+1
       do n=1,ngdim
          do k=1,m
             do g=1,gdim
                do v=1,vgdim
                   jplminpsi(v,g,k,n)=jplmin(v,g,k,n)*psi(v,g,k+1,n) ! for j-
                   jplminpsi(v,g,kdim-k+1,n)=jplmin(v,g,kdim-k,n)*&  ! for j+
                                             psi(v,g,kdim-k,n)
                enddo
             enddo
          enddo
       enddo

!---- for k=m+1,...,kdim-m
      do n=1,ngdim
         do k=m+1,kdim-m
            do g=1,gdim
               do v=1,vgdim
                  jplminpsi(v,g,k,n)=jplmin(v,g,k-1,n)*psi(v,g,k-1,n) + & ! for j+
                                     jplmin(v,g,k,n)  *psi(v,g,k+1,n)     ! for j-
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine jplusminuspsirw


!#######################################################################
!
!  This subroutine calculates the action of sqrt(1-H^2_s)
!  on the imaginary part of the initial wave function.
!  Furthermore, it calculates q1, the first step in the
!  integration of the total wave function.
!
!#######################################################################
      subroutine chebsqrt(time,psit1,hpsit1,psit2,mr)

use global
use dvrdatmod
use griddatmod
use runpropmod
use operdef
use psidef
use channels

use realwpmod

use absorbrwmod, only:absorbrwp
use op1lib, only: cpvxd, zerovxd

      implicit none

      real(dop), dimension(*), intent(inout)      :: mr

      real(dop), intent(in)                          :: time
      real(dop), dimension(totphidim), intent(in)    :: psit1
      real(dop), dimension(totphidim), intent(inout) :: psit2
      real(dop), dimension(totphidim), intent(out)   :: hpsit1

      integer(long), parameter                       :: m = 1, s = 1, nham = 1
      integer(long)                                  :: i,ix,k,k1,k2,h,f,&
                                                        zeig,zeig1,zeig2
      real(dop)                                      :: fac,fac2,ss,normcheb
      real(dop), dimension(3*totphidim)              :: workr
      real(dop), external                            :: dnrm2,ddot

      real(dop)                                      :: flop,energ,normtest,&
                                                        normtest2
      real(dop), dimension(ndof)                     :: flopt
      logical(kind=4)                                :: lprob


!-----------------------------------------------------------------------
!  To scales the matrices that form the Hamiltonian
!  to get its eigenvalues between -1 and 1.
!-----------------------------------------------------------------------

      if(khzahl(m,s,nham).eq.0) return
      k1=zhun(m,s,nham)
      k2=zhun(m,s,nham)+khzahl(m,s,nham)-1
      if (time > 0.0_dop) return

      fac=2.0_dop/pi
      ix=1
!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call zerovxd(workr,3*totphidim)

!------------------------------------------------------------------------------------------
! try to manage 'funkrwp' some more general way by removing some unnecessary stuff!!
!-----------------------------------------------------------------------------------------
      call funkrwp(psit1,hpsit1,mr) ! Re(psi(t=0))---> hpsi 

      zeig=1
      zeig1=zeig+totphidim
      zeig2=zeig1+totphidim

      call daxpy(totphidim,oner,hpsit1,ix,workr(zeig),ix) ! do we need this???  hpsi ---> Re(psi(t=1)), half of the Eq. 15 in paper jcp 108, 950(1998) 
      call daxpy(totphidim,-fac,psit2,ix,workr(zeig),ix) ! Re(psi(t=1)) + Im(psi(t=0) ---> Re(psi(t=1))

!-----------------------------------------------------------------------
!  Do chebychev series using:
!  -sqrt(1-H_s^2)= -(2/pi)*(1-2*Sum_s=1 (4s^2-1)^-1 (H_s)^2s)
!-----------------------------------------------------------------------
      fac2=1.0_dop
!
      write(ilog,'(/,72a1)')  ('#',i=1,72)
      do i=1,norder,2

         normcheb=dnrm2(totphidim,workr(zeig),1)
         write (ilog,'(a,i3,1x,f13.10)')' Iter. no. & chebnorm =',i-1,normcheb
         if (normcheb > 1.0_dop) then
            print*,' Iter. no. & chebnorm =',i-1,normcheb
            print*,'Check Chebyshev iteration'
            stop
         endif

         call funkrwp(psit2,workr(zeig1),mr) ! Im(psi(t=0)) ---> hpsi

         call dscal(totphidim,-oner,workr(zeig2),ix)   
         call daxpy(totphidim,fac2,workr(zeig1),ix,workr(zeig2),ix)

         fac2=2.0_dop

!   NEED TO ABSORB Non reactive and reactive part
        if (i /= 1) then
           do k=k1,k2
              if (autoabs(k) > 0) then
                 f = autoabs(k)
                 h = hamilton(f,k)
                 lprob=.false.
                 call absorbrwp(lprob,f,h,mr(mrhops),workr(zeig2),psit2,flop,flopt)
              endif
           enddo
         endif

         call funkrwp(workr(zeig2),workr(zeig1),mr)
!
         call dscal(totphidim,-oner,psit2,ix) 
         call daxpy(totphidim,2.0_dop,workr(zeig1),ix,psit2,ix) ! Im(psi(t=0)) + hpsi ---> Im(psi(t=1)
!
         ss = 2.0_dop*fac/((4.0_dop*(((i+1)/2)**2))-1.0_dop)
         call daxpy(totphidim,ss,psit2,ix,workr(zeig),ix) ! Im(psi(t=1)) + Re(psi(t=1)) ---> psi(t=1)

      enddo ! i
      
      call cpvxd(workr(zeig),psit2,totphidim)

      energ=ddot(totphidim,psit1,onei,hpsit1,onei)/ddot(totphidim,psit1,onei,psit1,onei)
      timestep=aunds/dsqrt(1.0_dop-energ**2)
      write (ilog,'(a,f13.10)')' timestep =',timestep

      normtest=dnrm2(totphidim,psit1,1)
      normtest2=dnrm2(totphidim,psit2,1)
      normwav=normtest2

      write(ilog,'(/,72a1)')  ('#',i=1,72)
      write (ilog,'(2a,f12.6)')' The norm of the real part',&
                  ' of the wave function at t=0 is ',normtest
      write (ilog,'(2a,f12.6)')' The norm of the real part',&
                  ' of the wave function at t=1 is ',normtest2
      write (ilog,'(a)') ' '

      call myflush(ilog)


      return
      end subroutine chebsqrt

      end module functionrw
