! **********************************************************************
! *                                                                    *
! *                          function (function.f)                     *
! *                                                                    *
! * mctdh module that calls the routines that compute the right hand   *
! * side of the mctdh working equations 
!
!   when the mean fields are held 'constant':
!       funk0, funk1, funk2, funk3, funkphi, funka.
!   when the single-particle hamiltonian is rhs of the projector:
!       funkr
!   when the single-particle hamiltonian is lhs of the projector:
!       funk
!
! contains: funkphi2,funka2,funkex,qdot,qdot1m,xdot,xdot1m,deriv1,
! **********************************************************************
      module functiontr

use decimal, only: long, dop
use global
use globmemmod
use logdat
use op1lib,  only: zerovxi, zerovxd, zerovxz, zeromxz, zeromxd, cpqxd
use timing

      implicit none
      private
      public :: funktr,funktr1,deriv1
              
      contains

!***********************************************************************
! 
!                           subroutine funktr
!
! computes the time derivative of the trajectories and weights
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine funktr (time,psi,dtpsi)

use griddatmod
use runpropmod
use operdef
use psidef

use tdkoeffs

      implicit none

      integer(long)                                  :: nham
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(out)   :: dtpsi
      integer(long)                                   :: tid=0 ! timer ID

      call get_timer(tid, 'Funktr')
      call start_timer(tid)

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call zerovxz(dtpsi,dgldim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! swarm of trajectories 
!-----------------------------------------------------------------------
      call qdot(time,psi,dtpsi,mi(mitrajst),ml(1))
      
      call stop_timer(tid)

      return
      end subroutine funktr   

!***********************************************************************
! 
!                           subroutine funktr1 
!
! computes the time derivative of the trajectories
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine funktr1 (time,psi,dtpsi)

use dvrdatmod
use griddatmod
use runpropmod
use operdef
use psidef

use tdkoeffs, only: calctdk

      implicit none

      integer(long)                                  :: nham,nsta
      real(dop), intent(in)                          :: time
      complex(dop), dimension(*), intent(inout)      :: psi
      complex(dop), dimension(*), intent(out)        :: dtpsi
      integer(long)                                  :: tid=0 ! timer ID

      call get_timer(tid, 'Funktr1')
      call start_timer(tid)

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call zeromxz(dtpsi,nspfdof(1),dim(1,1))

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)

!-----------------------------------------------------------------------
! swarm of trajectories 
!-----------------------------------------------------------------------
      if (feb .gt. 0) then
         nsta=gdim(feb)
      else
         nsta=1
      endif
      call qdot1ms(psi,dtpsi,dim(1,1),nspfdof(1),nsta, &
             mi(mitrajst+zgwpinfo(1,1)-1))

      call stop_timer(tid)

      return
      end subroutine funktr1   

!#######################################################################
!
! If lpred = T call is from predictor step of ABM and so no hopping
! is wanted
!
! routine to calculate time-derivative of trajectories.
! including surface hopping
!#######################################################################

      subroutine qdot(time,psi,dtpsi,trajst,lpred)


use dvrdatmod
use griddatmod
use runpropmod
use hpsimod
use psidef

      implicit none

      integer(long)                                    :: s,m,nsta
      integer(long), dimension(gwpinfodim), intent(out):: trajst
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi
      real(dop), intent(in)                            :: time
      logical(kind=4), intent(in)                      :: lpred

      m=1
      s=1
      if (feb .gt. 0) then
         nsta = gdim(feb)
      else
         nsta = 1
      endif
      call qdot1m(time,psi(zetf(m,s)),dtpsi(zetf(m,s)),dim(m,s),     &
           nspfdof(m),nsta,trajst(zgwpinfo(m,s)),psi(zpsi(s)), &
           dtpsi(zpsi(s)),lpred)  

      return
      end subroutine qdot


!#######################################################################

      subroutine qdot1m(time,psi,dtpsi,dim1,gdof,nsta,trajst,psiel, &
                        dtpsiel,lpred)

use maxv, only: maxdim,maxsta
use dvrdatmod
use griddatmod
use runpropmod
use hpsimod
use operdef

use potevalmod, only: poteval,dereval
use openmpmod

      implicit none

      logical(kind=4), intent(in)                         :: lpred

      integer(long), intent(in)                           :: dim1,gdof,nsta
      integer(long), dimension(dim1), intent(out)         :: trajst

      real(dop), intent(in)                               :: time

      complex(dop), dimension(gdof,dim1), intent(in)      :: psi
      complex(dop), dimension(gdof,dim1), intent(inout)   :: dtpsi
      complex(dop), dimension(nstate), intent(in)         :: psiel
      complex(dop), dimension(nstate), intent(out)        :: dtpsiel
      complex(dop), dimension(maxsta,maxsta)              :: srfhp

      logical(kind=4)                                     :: lhop,ltry
      logical(kind=4), save                               :: initial = .true.
      integer(long)                                       :: f,m,s,nham,h,k,&
                                                             e,n,dertyp,n1,&
                                                             pottyp,f1,h1,si,&
                                                             sf,s1,s2,s3,ierr
      integer(long), dimension(ndof), target              :: order
      real(dop)                                           :: deriv1,rfac
      real(dop)                                           :: diffe,delt
      real(dop), save                                     :: lasttime
      real(dop), dimension(maxdim),target                 :: q0,p0,dp1,dq1,&
                                                             p0new
      real(dop), dimension(nsta), target                  :: vmat
      real(dop), dimension(nsta,nsta), target             :: wmat,rotmat
      real(dop), dimension(nsta,nsta,maxdim), target      :: wmat1,vmat1
      real(dop), dimension(3*maxsta), target              :: workr
      complex(dop)                                        :: vz


      m=1
      nham=1
      call zerovxi(order,ndof)
      lhop = .false.

!  time step
      if (initial) then
         lasttime = time
         initial = .false.
         delt = 0.0
         ltry = .false.
      else if (.not. lpred) then
         delt = time - lasttime
         lasttime = time
         ltry = .true.
      else
         ltry = .false.
         delt = 0.0
      endif

!-----------------------------------------------------------------------
! loop over trajectories
! dq1 = dH/dq, dp1 = dH/dp
!-----------------------------------------------------------------------
      do e=1,dim1
         call zerovxd(dq1,gdof)
         call zerovxd(dp1,gdof)
         call zeromxd(wmat,nsta,nsta)
         call zerovxd(vmat,nsta)
         call zeromxd(wmat1,nsta,nsta*gdof)
         call zeromxd(vmat1,nsta,nsta*gdof)
         call zeromxd(rotmat,nsta,nsta)

!-----------------------------------------------------------------------
! get coordinates
!-----------------------------------------------------------------------
         do n=1,gdof
            f=spfdof(n,m)
            q0(f)=dble(psi(n,e))
            p0(f)=dimag(psi(n,e))
         enddo

!-----------------------------------------------------------------------
! loop over Hamiltonian terms
!-----------------------------------------------------------------------
         do k=1,khtot(nham)
            if (kinop(m,k) .eq. 1) then
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  h=hamilton(f,k)
                  order(f)=1
                  deriv1 = 0.0d0
                  call dereval (h,deriv1,q0,order,p0,dertyp,dermethod)
                  do n1=1,nspfdof(m)
                     if (n1 .ne. n) then
                        f1=spfdof(n1,m)
                        h1=hamilton(f1,k)
                        call poteval(h1,rfac,q0,vz,p0,pottyp)
                        deriv1 = rfac*deriv1
                     endif
                  enddo
                  if (dertyp .eq. 0) then
                     dq1(n) = dq1(n) + dble(koeff(k))*deriv1
                  else if (dertyp .eq. 1) then
                     dp1(n) = dp1(n) + dble(koeff(k))*deriv1
                  endif
                  order(f)=0
               enddo

            else 
               f=spfdof(1,m)
               h=hamilton(f,k)

               si=ksi(k)
               sf=ksf(k)

               rfac=dble(koeff(k))
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  order(f)=1
                  deriv1 = 0.0d0
                  call dereval (h,deriv1,q0,order,p0,dertyp,dermethod)
                
                  if (dertyp .eq. 0) then
                     wmat1(sf,si,n) = wmat1(sf,si,n) + rfac*deriv1
                  else if (dertyp .eq. 1) then
                     routine='Qdot1m'
                     message='Kinetic energy term in potentials'
                     call errormsg
                  endif

                  order(f)=0
               enddo
            endif
         enddo

         if (nsta .gt. 1) then
! Get diabatic potential energy matrix
            do k=1,khtot(nham)
               if (kinop(m,k) .eq. 0) then
                  f=spfdof(1,m)
                  h=hamilton(f,k)
                  si=ksi(k)
                  sf=ksf(k)

                  call poteval (h,rfac,q0,vz,p0,pottyp)
                
                  wmat(sf,si) = wmat(sf,si) + dble(koeff(k))*rfac
                  if (sf .ne. si) wmat(si,sf) = wmat(sf,si) 
               endif
            enddo

! diagonalise wmat
            call cpqxd(wmat,rotmat,nsta)
            call dsyev('V','U',nsta,rotmat,nsta,vmat,workr,  &
                 3*nsta,ierr)
            if (ierr .ne. 0) then
               routine = 'Qdot1m'
               message = 'Cannot diagonalize Wmat matrix'
               call errormsg
            endif

            call zeromxd(vmat1,nsta*nsta,gdof)

            do n=1,gdof
               do s=1,nsta
                  do s1=1,nsta
                     do s2=1,nsta
                        do s3=1,nsta
                           vmat1(s1,s,n)=vmat1(s1,s,n) +    &
                             rotmat(s1,s2)*wmat1(s2,s3,n)*rotmat(s,s3)
                        enddo
                     enddo
                  enddo
               enddo
            enddo

!-----------------------------------------------------------------------
! time-derivative of electronic coefficients
!-----------------------------------------------------------------------
            call zeromxz(srfhp,maxsta,maxsta)
            call zerovxz(dtpsiel,nsta)

            do s1=1,nsta
               srfhp(s1,s1) = (0,-1.0d0)*vmat(s1) 
               do s2=s1+1,nsta
                  diffe = vmat(s1) - vmat(s2)
                  do n=1,gdof
                     f=spfdof(n,m)
                     srfhp(s1,s2) = srfhp(s1,s2)           &
                           - (p0(f) / (mass(f)*diffe)) * vmat1(s1,s2,n) 
                  enddo
                  srfhp(s2,s1) = -(srfhp(s1,s2))
               enddo
            enddo

            do s1=1,nsta
               do s2=1,nsta
                  dtpsiel(s1) = dtpsiel(s1) + srfhp(s1,s2) * psiel(s2)
               enddo
            enddo

!-----------------------------------------------------------------------
! surface hop
!-----------------------------------------------------------------------
            if (ltry) then
               call srfhop(trajst(e),psiel,nsta,p0,gdof,lhop,delt,srfhp,&
                    vmat,vmat1,p0new)
            endif

            s=trajst(e)

         else
            lhop = .false.
            do n=1,nspfdof(m)
               vmat1(1,1,n) = wmat1(1,1,n)
            enddo

            s = 1
         endif

!-----------------------------------------------------------------------
! add up forces
!-----------------------------------------------------------------------
         do n=1,nspfdof(m)
            dq1(n) = dq1(n) + vmat1(s,s,n)
         enddo

!-----------------------------------------------------------------------
! Correct momenta
!-----------------------------------------------------------------------
         if (lhop) then
            do n=1,nspfdof(m)
               f=spfdof(n,1)
               dp1(n) = dp1(n) + (p0new(f)-p0(f))/mass(f)
            enddo
         endif

!-----------------------------------------------------------------------
! add up time-derivatives
!-----------------------------------------------------------------------
         do n=1,nspfdof(m)
            dtpsi(n,e)=dtpsi(n,e) + dcmplx(dp1(n),-dq1(n))
         enddo
      enddo

      return
      end subroutine qdot1m

!#######################################################################

      subroutine qdot1ms(psi,dtpsi,dim1,gdof,nsta,trajst)

use maxv, only: maxdim,maxsta
use dvrdatmod
use griddatmod
use runpropmod
use hpsimod
use operdef

use potevalmod, only: poteval,dereval
use openmpmod

      implicit none

      integer(long), intent(in)                           :: dim1,gdof,nsta
      integer(long), dimension(dim1), intent(in)          :: trajst

      complex(dop), dimension(gdof,dim1), intent(in)      :: psi
      complex(dop), dimension(gdof,dim1), intent(inout)   :: dtpsi

      integer(long)                                       :: f,m,s,nham,h,k,&
                                                             e,n,dertyp,n1,&
                                                             pottyp,f1,h1,si,&
                                                             sf,s1,s2,s3,ierr
      integer(long), dimension(ndof), target              :: order
      real(dop)                                           :: deriv1,rfac
      real(dop), dimension(maxdim),target                 :: q0,p0,dp1,dq1
      real(dop), dimension(nsta), target                  :: vmat
      real(dop), dimension(nsta,nsta), target             :: wmat,rotmat
      real(dop), dimension(nsta,nsta,maxdim), target      :: wmat1,vmat1
      real(dop), dimension(3*maxsta), target              :: workr
      complex(dop)                                        :: vz


      m=1
      nham=1
      call zerovxi(order,ndof)

!-----------------------------------------------------------------------
! loop over trajectories
! dq1 = dH/dq, dp1 = dH/dp
!-----------------------------------------------------------------------
      do e=1,dim1
         call zerovxd(dq1,gdof)
         call zerovxd(dp1,gdof)
         call zeromxd(wmat,nsta,nsta)
         call zerovxd(vmat,nsta)
         call zeromxd(wmat1,nsta,nsta*gdof)
         call zeromxd(vmat1,nsta,nsta*gdof)
         call zeromxd(rotmat,nsta,nsta)

!-----------------------------------------------------------------------
! get coordinates
!-----------------------------------------------------------------------
         do n=1,gdof
            f=spfdof(n,m)
            q0(f)=dble(psi(n,e))
            p0(f)=dimag(psi(n,e))
         enddo

!-----------------------------------------------------------------------
! loop over Hamiltonian terms
!-----------------------------------------------------------------------
         do k=1,khtot(nham)
            if (kinop(m,k) .eq. 1) then
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  h=hamilton(f,k)
                  order(f)=1
                  deriv1 = 0.0d0
                  call dereval (h,deriv1,q0,order,p0,dertyp,dermethod)
                  do n1=1,nspfdof(m)
                     if (n1 .ne. n) then
                        f1=spfdof(n1,m)
                        h1=hamilton(f1,k)
                        call poteval(h1,rfac,q0,vz,p0,pottyp)
                        deriv1 = rfac*deriv1
                     endif
                  enddo
                  if (dertyp .eq. 0) then
                     dq1(n) = dq1(n) + dble(koeff(k))*deriv1
                  else if (dertyp .eq. 1) then
                     dp1(n) = dp1(n) + dble(koeff(k))*deriv1
                  endif
                  order(f)=0
               enddo

            else 
               f=spfdof(1,m)
               h=hamilton(f,k)

               si=ksi(k)
               sf=ksf(k)

               rfac=dble(koeff(k))
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  order(f)=1
                  deriv1 = 0.0d0
                  call dereval (h,deriv1,q0,order,p0,dertyp,dermethod)
                
                  if (dertyp .eq. 0) then
                     wmat1(sf,si,n) = wmat1(sf,si,n) + rfac*deriv1
                  else if (dertyp .eq. 1) then
                     routine='Qdot1m'
                     message='Kinetic energy term in potentials'
                     call errormsg
                  endif

                  order(f)=0
               enddo
            endif
         enddo

         if (nsta .gt. 1) then
! Get diabatic potential energy matrix
            do k=1,khtot(nham)
               if (kinop(m,k) .eq. 0) then
                  f=spfdof(1,m)
                  h=hamilton(f,k)
                  si=ksi(k)
                  sf=ksf(k)

                  call poteval (h,rfac,q0,vz,p0,pottyp)
                
                  wmat(sf,si) = wmat(sf,si) + dble(koeff(k))*rfac
                  if (sf .ne. si) wmat(si,sf) = wmat(sf,si) 
               endif
            enddo

! diagonalise wmat
            call cpqxd(wmat,rotmat,nsta)
            call dsyev('V','U',nsta,rotmat,nsta,vmat,workr,  &
                 3*nsta,ierr)
            if (ierr .ne. 0) then
               routine = 'Qdot1m'
               message = 'Cannot diagonalize Wmat matrix'
               call errormsg
            endif

            call zeromxd(vmat1,nsta*nsta,gdof)

            do n=1,gdof
               do s=1,nsta
                  do s1=1,nsta
                     do s2=1,nsta
                        do s3=1,nsta
                           vmat1(s1,s,n)=vmat1(s1,s,n) +    &
                             rotmat(s1,s2)*wmat1(s2,s3,n)*rotmat(s,s3)
                        enddo
                     enddo
                  enddo
               enddo
            enddo

            s=trajst(e)

         else
            do n=1,nspfdof(m)
               vmat1(1,1,n) = wmat1(1,1,n)
            enddo

            s = 1
         endif

!-----------------------------------------------------------------------
! add up forces
!-----------------------------------------------------------------------
         do n=1,nspfdof(m)
            dq1(n) = dq1(n) + vmat1(s,s,n)
         enddo

!-----------------------------------------------------------------------
! add up time-derivatives
!-----------------------------------------------------------------------
         do n=1,nspfdof(m)
            dtpsi(n,e)=dtpsi(n,e) + dcmplx(dp1(n),-dq1(n))
         enddo
      enddo

      return
      end subroutine qdot1ms

!#######################################################################
!
!
! nsta: no. of states
! psiel(e): coefficient of state e in electronic wavefunction
!
!#######################################################################

      subroutine srfhop(trajst,psiel,nsta,p0,gdof,lhop,   &
                 delt,srfhp,vmat,vmat1,p0new)


use maxv, only: maxdim,maxsta
use griddatmod
use operdef
use randommod, only: randuniform

      implicit none

      integer(long), intent(out)                         :: trajst
      integer(long), intent(in)                          :: nsta,gdof
      real(dop), dimension(maxdim), intent(in)           :: p0
      real(dop), dimension(maxdim), intent(out)          :: p0new
      real(dop), dimension(nsta,nsta,gdof), intent(in)   :: vmat1
      real(dop), dimension(nsta), intent(in)             :: vmat
      complex(dop), dimension(nsta), intent(in)          :: psiel
      complex(dop), dimension(maxsta,maxsta), intent(in) :: srfhp
      real(dop), intent(in)                              :: delt
      logical(kind=4), intent(out)                       :: lhop
      real(dop)                                          :: dele,delpara,dotp,norm
      real(dop), dimension(maxsta,maxsta,maxdim)         :: vmat1norm
      integer(long)                                      :: f,n,i,j
      real(dop)                                          :: p12,rx,rtmp
      complex(dop)                                       :: denii,denij

!     print*,'trajst',trajst

      lhop = .false.
!-------------------------------------------
! Hopping Probability
!-------------------------------------------
! if (trajst .eq. 1) then
      i = trajst
      rtmp=-999.0d0
      if (i .eq. 1) then
         j = 2
         denij = dconjg(psiel(j))*psiel(i)
         rtmp = dble(srfhp(j,i)*denij)
      else if (i .eq. 2) then
         j = 1
         denij = dconjg(psiel(i))*psiel(j)
         rtmp = dble(srfhp(i,j)*denij)
      endif
      denii = dconjg(psiel(i))*psiel(i)
      if (rtmp .gt. 0) then
         p12 = 2*delt*rtmp/dble(denii)
      else
         p12 = 0.0d0
      endif

!--------------------------------------------
! Generate random no.
!--------------------------------------------
      rx = randuniform()
!      print*, 'rx',rx

!--------------------------------------------
! Random No. Check - If random no. < p12, hop
!--------------------------------------------
      j=0
      if (rx .lt. p12) then
         lhop = .true.
         trajst = j
      else
         return
      endif

!--------------------------------------------
! Energy gap
!--------------------------------------------
      dele = vmat(j) - vmat(i)

!--------------------------------------------
! Energy - after hop has occurred
!--------------------------------------------
      norm=0.0
      do n=1,gdof
         norm=norm+dble(vmat1(i,j,n))*(vmat1(i,j,n))
      enddo
      norm=dsqrt(norm)
      do n=1,gdof
         vmat1norm(i,j,n)=vmat1(i,j,n)/norm
      enddo
!     print*,vmat1norm(i,j,1)
      dotp=0.0
      do n=1,gdof
         f=spfdof(n,1)
         dotp=dotp+(p0(f)/dsqrt(mass(f)))*vmat1norm(i,j,n)
      enddo

!-----------------------------------------------------------------------
! check HOP is allowed (should collect statustics)
! 0.5*dotp**2 is KE in direction of non-adiabatic coupling
!-----------------------------------------------------------------------
      if (dele .gt. 0.5d0*dotp*dotp) then
         lhop = .false.
         return
      endif

!-----------------------------------------------------------------------
! calculate new momentum
!-----------------------------------------------------------------------
      delpara = dsqrt(dotp*dotp - 2.0d0*dele)

      do n =1,gdof
         f=spfdof(n,1)
         norm = (delpara / dsqrt(dotp*dotp)) - 1.0d0
         norm = norm*dotp
         p0new(f) = p0(f) + norm*vmat1norm(i,j,n)
      enddo

      return
      end subroutine srfhop

!#######################################################################

      subroutine deriv1(dq1,h,q0,dq,f)

use maxv, only: maxdim

use potevalmod, only: poteval

      implicit none

      integer(long)                            :: idum
      integer(long), intent(in)                :: h,f
      real(dop)                                :: vp,vm,qp,qm,diff
      real(dop), intent(in)                    :: dq
      real(dop), intent(out)                   :: dq1
      real(dop), dimension(maxdim)             :: rdum
      real(dop), dimension(maxdim)             :: gpoint
      real(dop), dimension(maxdim), intent(in) :: q0
      complex(dop)                             :: vz

      qp  = q0(f) + dq/2.0_dop
      qm  = q0(f) - dq/2.0_dop

      gpoint(f)=qp
      call poteval (h,vp,gpoint,vz,rdum,idum)
      gpoint(f)=qm
      call poteval (h,vm,gpoint,vz,rdum,idum)

      diff = vp - vm

      dq1 = diff/dq

      return
      end subroutine deriv1


      end module functiontr
