! **********************************************************************
! *                                                                    *
! *                        INTEGRATION (integrat.f90)                  *
! *                                                                    *
! * MCTDH module containing all those routines needed for the          *
! * integration that are MCTDH specific and thus not stored in the     *
! * various integrator library modules.                                *
! *                                                                    *
! * Contains:                                                          *
! *   runabm:       Drives the ABM integration routine.                *
! *   runbs:        Drives the BS integration routine.                 *
! *   runrk5:       Drives the RK5 integration routine.                *
! *   runrk8:       Drives the RK8 integration routine.                *
! *                                                                    *
! **********************************************************************
      module integrattr

use decimal, only: dop, long
use global
use globmemmod
use logdat
use timing

      implicit none
      private
      public :: runabmtr,runbstr,runrk5tr,runrk8tr,runverlettr
      contains

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNABM                        *
! *                                                                    *
! * Drives the ABM integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Lenght of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   restartabm:    Restart flag; if false, a previous integration is *
! *                  continued, otherwise a new integration is         *
! *                  started.                                          *
! *   steps:         Overall number of integration steps made so far.  *
! *   repsteps:      Number of failed and repeated steps made so far.  *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *   steps:         Overall number of integration steps made so far.  *
! *   repsteps:      Number of failed and repeated steps made so far.  *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runabmtr (psi,dtpsi,intperiod,time,restartabm,steps, &
                         repsteps)

use griddatmod
use psidef
use datenmod
use runpropmod
use channels

use functiontr, only: funktr
use abmlib, only: abm,abmerrormsg,absabmerror

      implicit none

      real(dop), parameter                           :: abmaccucorr = 1.0_dop

      logical(kind=4), intent(in)                    :: restartabm
      integer(long)                                  :: error
      integer(long), intent(out)                     :: repsteps,steps
      real(dop)                                      :: abmaccuracy
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      integer(long)                                  :: tid=0 ! timer ID

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)') 'A: Adams-Bashforth-Moulton'
         write (istp,'(a7,9x,a6,6x,a13,11x,a5,11x,a9)') &
            'Step','Order','Stepsize [fs]','Error','Time [fs]'
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      abmaccuracy = abmaccucorr*accuracy

! --- CALL ADAMS-BASHFORTH-MOULTON INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'ABM')
      call start_timer(tid)

      call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt,&
               abmaccuracy,restartabm,steps,repsteps,error,&
               mc(mcintpsi),funktr,absabmerror)

      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'abm'
         call abmerrormsg(error,message)
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine runabmtr

! **********************************************************************
! *                                                                    *
! *                            SUBROUTINE RUNBS                        *
! *                                                                    *
! * Drives the Bulirsch-Stoer integration routine.                     *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   dtpsi:      Time derivative of the initial-value vector.         *
! *   intperiod:  Lenght of time interval to be integrated.            *
! *   abstime:    Absolute time, i. e. Psi_initial=Psi(AbsTime).       *
! *   restartbs:  Restart flag; if false, the previous stepsize is     *
! *               used as initial stepsize, otherwise IntPeriod is     *
! *               taken.                                               *
! *   goodsteps:  Number of succesful large integration steps made so  *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed large integration steps   *
! *               made so far.                                         *
! *   smallsteps: Number of small integration steps made so far.       *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *   dtpsi:      Dtpsi is undetermined.                               *
! *   goodsteps:  Number of succesful large integration steps made so  *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed large integration steps   *
! *               made so far.                                         *
! *   smallsteps: Number of small integration steps made so far.       *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:         Complex*16 data needed in the Func (and possibly     *
! *               CalcError) routine.                                  *
! *   mr:         Real*8 data needed in the Func (and possibly         *
! *               CalcError) routine.                                  *
! *   mi:         Integer data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *   ml:         Logical data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runbstr (psi,dtpsi,intperiod,abstime,restartbs,goodsteps, &
                        badsteps,smallsteps)

use psidef
use datenmod
use runpropmod
use channels

use functiontr, only: funktr
use bslib, only: bsstep,bserrormsg,polyextrapol,absbserror

      implicit none

      real(dop), parameter                          :: bsaccucorr = 1.0_dop, &
                                                       enlargefactor = 1.7_dop

      logical(kind=4), intent(in)                   :: restartbs
      integer(long)                                 :: error
      integer(long), intent(inout)                  :: goodsteps,badsteps,smallsteps
      real(dop)                                     :: bsaccuracy,time,largestep,nextstep,actualstep
      real(dop), save                               :: lastperiod
      real(dop), save                               :: laststep
      real(dop), intent(in)                         :: abstime,intperiod
      complex(dop), dimension(dgldim), intent(inout):: psi,dtpsi
      integer(long)                                 :: tid=0 ! timer ID

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)')&
               'Bulirsch-Stoer'
         write(istp,'(a11,2x,a11,4x,a13,11x,a5,11x,a9)') 'Iteration',&
                    'Small steps', 'Stepsize [fs]', 'Error','Time [fs]'
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      bsaccuracy = bsaccucorr*accuracy

! --- INITIALIZE VARIABLES ---

      time = abstime
      if (restartbs) then
         largestep = min(dt,intperiod)
         laststep=0.0
         lastperiod=0.0
      else
         if (abs(laststep-lastperiod) .lt. tiny*lastperiod) then
            largestep = intperiod
         else
            largestep = min(laststep,intperiod)
         endif
      endif

! --- INTEGRATION LOOP ---

 100  continue

! --- PERFORM ONE BULIRSCH-STOER INTEGRATION STEP ---

      call get_timer(tid, 'BS')
      call start_timer(tid)
      call bsstep(psi,dtpsi,dgldim,intperiod,time,intord, &
                  largestep,bsaccuracy,actualstep,nextstep, &
                  smallsteps,error,mc(mcintpsi),funktr,absbserror, &
                  polyextrapol)
      call stop_timer(tid)

! --- UPDATE TIME VARIABLE ---

      time = time+actualstep

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'bsstep'
         call bserrormsg(error,message)
         call errormsg
      endif

! --- COUNT GOOD AND BAD STEPS ---

      if (largestep .eq. actualstep) then
         goodsteps = goodsteps+1
      else
         badsteps = badsteps+1
      endif

! --- SAVE STEPSIZE AS INITIAL VALUE FOR NEXT CALL ---

      lastperiod = intperiod
      if (abs(time-abstime-intperiod) .ge. tiny*intperiod) then
         laststep = actualstep
      else
         laststep = min(nextstep,intperiod)
      endif

! --- LET STEPSIZE BE A DIVISOR OF UPDATE-TIME ---

      if (laststep .lt. intperiod) then
         laststep = intperiod/nint(intperiod/laststep)
      endif

! --- RETURN TO CALLING ROUTINE WHEN FINISHED ---

      if (abs(time-abstime-intperiod) .lt. tiny*intperiod) then
         return
      endif

! --- EVALUATE FUNCTION WITH NEW PSI ---

      call funktr(time,psi,dtpsi)

! --- FIT STEPSIZE TO REST OF INTEGRATION INTERVAL ---

      if (abstime+intperiod-time .lt. enlargefactor*nextstep) then
         largestep = abstime+intperiod-time
      else
         largestep = nextstep
      endif

! --- CONTINUE INTEGRATION ---

      goto 100

      end subroutine runbstr

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNRK5                        *
! *                                                                    *
! * Drives the RK5 integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runrk5tr (psi,dtpsi,intperiod,time)

use maxv, only: maxdim
use dvrdatmod
use griddatmod
use psidef
use datenmod
use runpropmod
use channels

use importhomod, only: importhovmf,importho,dummypost
use functiontr, only: funktr

use rklib, only: rk5

      implicit none

      real(dop), parameter                           :: rk5accucorr = 0.1_dop

      integer(long)                                  :: error
      real(dop)                                      :: rk5accuracy
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      logical(kind=4), save                          :: initialized = .false.
!      external                                       :: funk,funkr,funkphi,funka,funkex,&
!                                                        importho,importhovmf,dummypost,funkphi2
      integer(long)                                  :: tid=0 ! timer ID


      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- INITIALIZATION ---

!     is this the first call?
      if (.not. initialized) then
!        yes: initialize the stepsizes
         laststep(0) = dt
         initialized = .true.
      endif
      laststep(1)=0.0_dop

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)') 'Runge-Kutta order 5'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
              'Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)',&
              'Time [fs]'
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      rk5accuracy = rk5accucorr*accuracy

! --- CALL RK5 INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'RK5')
      call start_timer(tid)
      call rk5(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
               rk5accuracy,error,mc(mcintpsi),funktr,&
               limportho, importhovmf,.false.,dummypost,.false.,dummypost)
      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'RunRK5'
         message = 'Integration failed.'
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine runrk5tr

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNRK8                        *
! *                                                                    *
! * Drives the RK8 integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runrk8tr (psi,dtpsi,intperiod,time)

use griddatmod
use psidef
use datenmod
use runpropmod
use channels

use importhomod, only: importhovmf,importho,dummypost
use functiontr, only: funktr

use rklib, only: rk8

      implicit none

      real(dop), parameter                           :: rk8accucorr = 1.0_dop

      integer(long)                                  :: error
      real(dop)                                      :: rk8accuracy
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      logical(kind=4), save                          :: initialized = .false.
!      external                                       :: funk,funkr,funkphi,funka,funkex,&
!                                                        importho,importhovmf,dummypost,funkphi2
      integer(long)                                  :: tid=0 ! timer ID

      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- INITIALIZATION ---

!     is this the first call?
      if (.not. initialized) then
!        yes: initialize the stepsizes
         laststep(0) = dt
         initialized = .true.
      endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)') 'Runge-Kutta order 8'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)') &
              'Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)', &
              'Time [fs]'
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      rk8accuracy = rk8accucorr*accuracy

! --- CALL RK8 INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'RK8')
      call start_timer(tid)
      if (limportho) then
         call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0), &
                 rk8accuracy,error,mc(mcintpsi),funktr, &
                 .true., importhovmf,.false.,dummypost,.false.,dummypost)
      else
         call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0), &
                 rk8accuracy,error,mc(mcintpsi),funktr, &
                 .false.,dummypost,.false.,dummypost,.false.,dummypost)
      endif

      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'RunRK8'
         message = 'Integration failed.'
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine runrk8tr


! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNVERLET                     *
! *                                                                    *
! * Drives the verlet integration routine.                             *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runverlettr (psi,dtpsi,intperiod,time)

use griddatmod
use psidef
use datenmod
use runpropmod
use channels
use rklib

use importhomod, only: importhovmf,importho,dummypost
use functiontr, only: funktr1

use verletlib, only: verlet

      implicit none

      integer(long)                                  :: m,s
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      integer(long)                                  :: tid=0 ! timer ID

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)') 'Verlet'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
              'Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)',&
              'Time [fs]'
         call myflush(istp)
      endif


! --- CALL VVERLET INTEGRATOR WITH DESIRED FUNCTIONS ---
      m=1
      s=1
      call get_timer(tid, 'Verlet')
      call start_timer(tid)
      call verlet(psi(zetf(m,s)),dtpsi(zetf(m,s)),nspfdof(m),dim(m,s),intperiod,&
                  time,dt,funktr1)
      call stop_timer(tid)

! --- RETURN TO CALLING ROUTINE ---

      return
      end subroutine runverlettr

      end module integrattr
