! **********************************************************************
!                                                                      
!                   module PROP
!                                                                      
!  MCTDH module controlling the propagation of the wavefunction,       
!  i. e. calling the required evaluation and integration routines.     
!                                                                      
! **********************************************************************
      module proptr

use decimal, only: long, dop
use globmemmod

      implicit none
      private
      public :: propagatetr
      contains

! **********************************************************************
! *** V6.2 MB *********> SUBROUTINE PROPAGATE PSI <*********************
!                       ~~~~~~~~~~~~~~~~~~~~~~~~~~
! Controls the propagation of the wavefunction.
! **********************************************************************

      subroutine propagatetr (psi,dtpsi)

use global
use griddatmod
use operdef
use psidef
use runpropmod
use dirdyn
use channels
use openmpmod

use expectval, only: trexpectvalue
use functiontr, only: funktr
use outputtrmod, only: outputtr
use mempropmod, only: checkmemprop
use proptr1, only: proptraj,writeintmethodtr
use prop, only: zeitstop,nextday,oldhour
use datenmod

      implicit none

      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(out)   :: dtpsi

      real(dop)                     :: time,timesinceoutput,timestep,&
                                       sumupdate,sum2update

      integer(long)     :: abmsteps,repabmsteps,smallbssteps,goodbssteps,&
                           badbssteps,goodsilsteps,badsilsteps,updatesteps,&
                           badupdatesteps,badbackw,i,nfin
      logical(kind=4)   :: restartabm,restartbs,restartbs2,restartsil,relaxev
                           
      character(len=72) :: stopcommand

! --- CHECK ---
      if( tfinal .le. tinit + tiny ) then
         write(ilog,'(a)') '*******************************'
         write(iout,'(/a)') 'Nothing left to do!'
         write(ilog,'(/a)') 'Nothing left to do!'
         write(ilog,'(a,f15.6,a/,a,f15.6,a/)') &
         't-init =', tinit/fs,' fs','tfinal =',tfinal/fs,' fs'
         lstop = .false.
         goto 201
      end if

! --- INITIALIZE VARIABLES ---
      time = tinit
      timesinceoutput = 0.0_dop
      abmsteps = 0
      repabmsteps = 0
      smallbssteps = 0
      goodbssteps = 0
      badbssteps = 0
      goodsilsteps = 0
      badsilsteps = 0
      restartabm = .true.
      restartbs = .true.
      restartbs2 = .true.
      restartsil = .true.
      relaxev = .false.
      sumupdate = 0.0_dop
      sum2update = 0.0_dop
      updatesteps = 0
      badupdatesteps = 0
      badbackw = 0
      nfin = 0

! --- CALCULATE EXPECTATION VALUE FOR t=t0 IF REQUIRED ---
      do i=1,nexpect
         call trexpectvalue(psi(zetf(1,1)),expectlab(i),expvalue(i), &
              mi(mitrajst),dim(1,1),nspfdof(1))
      enddo
      do i=1,nexpect1
         call trexpectvalue(psi(zetf(1,1)),expectlab1(i),expvalue1(i), &
              mi(mitrajst),dim(1,1),nspfdof(1))
      enddo

! --- EVALUATE RIGHT HAND SIDE OF ODE WITH INITIAL WAVEFUNCTION ---
      
      call funktr(time,psi,dtpsi)

! --- WRITE INITIAL DATA AND WAVEFUNCTION ---
      if (.not. lcont) call outputtr(time,psi)

! --- CHECK WHETHER RUN IS TEST RUN ---

      if (test) then
         write(iout,'(/,a,/)') 'Test run was successful.'
         write(ilog,'(/,a,/)') '***** Test run *****'
         write(ilog,'(a,/)')   'Test run completed.'
         goto 200
      endif


! --- PROPAGATION LOOP ---

 100  continue

!     --- check consistency of mc,mr,mi,ml arrays, 
#ifdef DEBUG
      call checkmemprop(.true.,0)
#endif


! --- WRITE TIME TO INTSTEP FILE ---
! --- The stopcommand String is used when lstop is set true in integrat.F --
      stopcommand = 'stop  by internal check'
      if (lintstep) write(istp,'(a,f12.5)') 'Time in fs: ', time/fs


      call proptraj(psi,dtpsi,time,out1,timestep,restartabm,restartbs,&
                   abmsteps,repabmsteps,goodbssteps,badbssteps,&
                   smallbssteps)

! --- UPDATE TIME VARIABLES ---
      time = time+timestep
      timesinceoutput = timesinceoutput+timestep
      if(louteverystep) timesinceoutput = out1 

! --- EVALUATE RIGHT HAND SIDE OF ODE ---

      call funktr(time,psi,dtpsi)

! --- CALL OUTPUT ROUTINES AND LOOK FOR STOP COMMAND ---
      if (abs(out1-timesinceoutput) .lt. tiny*out1) then
         timesinceoutput = 0.0_dop

! --- CALCULATE EXPECTATION VALUE FOR ACTUAL TIME, IF REQUIRED ---
         do i=1,nexpect
            call trexpectvalue(psi(zetf(1,1)),expectlab(i),expvalue(i), &
                 mi(mitrajst),dim(1,1),nspfdof(1))
         enddo
         do i=1,nexpect1
            call trexpectvalue(psi(zetf(1,1)),expectlab1(i),expvalue1(i), &
                 mi(mitrajst),dim(1,1),nspfdof(1))
         enddo

! --- WRITE DATA (including RESTART file) 
         call outputtr(time,psi)

!........dummy write to ilog in order to touch the log-file.
!        (do not write if psi=compact)
         if (.not. lpsiopt(2)) then
!           backspace ilog
            write(ilog,'(a,f12.3)') ' time[fs] ', time/fs
!           endfile ilog
!           backspace ilog
         endif

!........flush data
         if(lupdate) call myflush(iupd)
         if(lspeed)  call myflush(ispeed)
         if(lpdens)  call myflush(ipden)
         if(nexpect.gt.0) call myflush(iexpect)
         if(nexpect1.gt.0)call myflush(iexpect1)

         if(lstopf.and.time.lt.(tfinal-0.1_dop*out1).and. .not.lstop) then
            open(istop,file=name(1:laenge)//'/stop', &
            form='formatted',status='old',err=666)
            read(istop,'(a)') stopcommand
!...........Go to the end of the stop-file and write the propagation
!...........time to it in order to 'touch' the stop-file.
            do i = 1, 21
                read(istop,'(a)') 
            end do
            write(istop,'(f12.3)') time/fs
            close(istop)
            goto 667
! --- If the stop-file cannot be opened stop the calculation ---
 666        stopcommand = 'stop      Cannot open stop-file'  
 667        if( stopcommand(1:4) .eq. 'cont' ) then
               nextday = 0
               oldhour = 0
            else if( stopcommand(1:4) .eq. 'stop' ) then
               lstop = .true.
            else if( stopcommand(3:3).eq.':' .or.  &
                     stopcommand(10:10).eq.':' ) then
               call zeitstop(stopcommand)
            end if
         end if
      endif

! --- CONTINUE PROPAGATION ---
      if (((tfinal-time).ge.tiny*tfinal) .and. .not.lstop) goto 100


! --- WRITE TIME TO INTSTEP FILE ---
      if (lintstep) write(istp,'(a,f12.5)') 'Time in fs: ', time/fs

! --- WRITE INTEGRATION METHOD TO LOG FILE ---
      call writeintmethodtr(abmsteps,repabmsteps,smallbssteps,goodbssteps, &
                          badbssteps)

! --- WRITE FINAL MESSAGES TO FILES ---
  201 continue
      write(iout,*)
      if (.not. lstop) then
         write(iout,'(a)') 'Propagation was successful.'
      else
         write(iout,'(a)') 'Propagation was stopped.'
         write(iout,'(a)') 'The stop-command reads:'
         write(iout,'(a)')  stopcommand
         write(iout,*)
      endif
      write(iout,*)
      call myflush(iout)

 200  continue

      return
      end subroutine propagatetr

      end module proptr

