!***********************************************************************
!
!                     module AUTOCORR.F
!
! Contains all subroutines related to the output of the autocorrelation
! function.
!
! Subroutines:
!    autoout :  writes auto file at t=k*tout, k=0,1,...
!    autoout2 : writes auto file at t=(k+1/2)*tout
!    autolap  : computes overlap <psi1^*|psi2> (needed in auto routines)
!    autoouthigh: writes higher order auto files at t=k*tout, k=0,1,...
!    autoout2high: writes higher order auto files t=(k+1/2)*tout
!    autohigh:  computes <psi1^*|H|psi2> and <psi1^*|H^2|psi2>
!               
!***********************************************************************
      module autocorr

use decimal, only: dop, long
use logdat
use mmlib,   only: mmtxzz,mqxczz,qqtxzz,qqxxzz,mmaxzz
use mtlib,   only: qtxxzz, qttxzz
use xvlib,   only: vvtxzza, vvtxzz
use op1lib,  only: cpvxz, zerovxl, cpvxz_s, cpvxi, cpqxz, overmxz
use op2lib,  only: swapsxxi

use maxv, only: maxdim,maxsta
use global
use globmemmod
use dvrdatmod
use griddatmod
use psidef
use datenmod
use channels
use runpropmod
use multilayer, only: automl
use gwplib, only: param_from_psi, diagonal_moments
use ciselect, only: venpsi
use lalib,  only: simtran2tz

      implicit none
      private
      public :: autoout,autoout2,autoouthigh,autoout2high
      contains

!***********************************************************************
!                               AUTOOUT
!
! Outputs autocorrelation function at time t to file auto
!
! V6.3 GW
!***********************************************************************

      subroutine autoout(psi,atime,jindx,agmat)

      implicit none

      integer(long)                                    :: p1,p2
      integer(long), dimension(jindxdim), intent(in)   :: jindx
      integer(long)                                    :: npack
      real(dop), intent(in)                            :: atime
      complex(dop), dimension(dgldim), intent(in)      :: psi
      complex(dop), dimension(1:maxsta,1:maxsta)       :: auto
      complex(dop), dimension(maxsta)                  :: autos
      complex(dop), dimension(agmatdim), intent(in)    :: agmat

!-----------------------------------------------------------------------
! Loop over all combinations of packets (to produce cross-correlation)
!-----------------------------------------------------------------------
      npack=max(npacket,npackts) ! Do the loop for single- and multi-set
      if(npack.gt.maxsta) then
         write(6,'(/a,2i6)') ' npack, maxsta', npack, maxsta
         routine='autoout'
         message='ERROR: The dimension of auto must be at least'// &
                 ' the number of packets, or do not set "auto" in'// &
                 ' in Run-Section. (This needs to be fixed.)'
         call errormsg
      endif

      do p2 = 1,npacket
         do p1 = p2,npacket

!-----------------------------------------------------------------------
! Calculate autocorrelation function
!-----------------------------------------------------------------------
            call autolap(psi,psi,auto(p1,p2),autos,p1,p2,jindx,agmat,agmat)
            if (p1 .ne. p2) auto(p2,p1) = auto(p1,p2)

         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate error estimate of autocorrelation function
!-----------------------------------------------------------------------
      if (lautoe) then
         if (lmulpack .or. mpb.gt.0) then
            routine = 'autoout'
            message = 'Error estimate of c(t) not implemented '// &
                      'for multi-packet run'
            call errormsg
         endif
         call autooute(psi,psi,dicht3,dicht4,dicht3,dicht4,autos)
         call autooutea(psi,psi,dicht4,dicht4,auto(1,1),true)
      endif

!-----------------------------------------------------------------------
! Write 'auto'-file output
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = 1,npacket
            write(iaut,'(1x,g18.9,3(1x,f18.14))') atime, &
                  dble(auto(p1,p2)),dimag(auto(p1,p2)), &
                  abs(auto(p1,p2))
         enddo
      enddo
      call myflush(iaut)

      return
      end subroutine autoout

!***********************************************************************
!                                AUTOOUT2
!
! Outputs autocorrelation function at time t/2 to file auto. (Used in
! a CMF calculation with adaptive step size.)
! Whereas "autoout" writes the autocorrelation function c(2nt) =
! <psi^*(nt)|psi(nt)> to file auto, this routine writes c(2(n-1)t) =
! <psi^*((n-1)t)|psi(nt)>.
!
! V6.4 MB
!***********************************************************************

      subroutine autoout2(psi1,psi2,atime,jindx,agmat2)

      implicit none

      integer(long)                                    :: p1,p2,s,m
      integer(long), dimension(jindxdim), intent(in)   :: jindx
      real(dop), intent(in)                            :: atime
      complex(dop), dimension(dgldim), intent(inout)   :: psi1,psi2
      complex(dop), dimension(maxsta)                  :: autos
      complex(dop), dimension(1:maxsta,1:maxsta)       :: auto
      complex(dop), dimension(agmatdim), intent(in)    :: agmat2
      complex(dop), dimension(agmatdim)                :: agmat1

!-----------------------------------------------------------------------
! get agmat for old wavefunction. If variable width get from psi,
! if fixed width copy agmat.
!-----------------------------------------------------------------------
      if (lgwp) then
         if (gwptype .le. 2) then
            do m=1,nmode
               if (.not. gwpm(m)) cycle
               do s=1,nstate
                  call param_from_psi(nspfdof(m),gwpdim(m,s),&
                       agmat1(zagmat(m,s)),psi1(zpg(m,s)),psi1(zpg(m,s)),&
                       psi1(zphi(m,s)),agmat1(zphase(m,s)),&
                       agmat1(zxivec(m,s)),1,s,m)
               enddo
            enddo
         else
            call cpvxz(agmat2,agmat1,agmatdim)
         endif
      endif

!-----------------------------------------------------------------------
! Loop over all combinations of packets (to produce cross-correlation)
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = p2,npacket

!-----------------------------------------------------------------------
! Calculate autocorrelation function
!-----------------------------------------------------------------------
            call autolap(psi1,psi2,auto(p1,p2),autos,p1,p2,jindx,&
                 agmat1,agmat2)
            if (p1 .ne. p2) auto(p2,p1) = auto(p1,p2)

         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate error estimate of autocorrelation function
!-----------------------------------------------------------------------
      if (lautoe) then
         if (lmulpack) then
            routine = 'autoout2'
            message = 'Error estimate of c(t) not implemented '//&
                      'for multi-packet run'
            call errormsg
         endif
         call autooute(psi1,psi2,dicht3o,dicht4o,dicht3,dicht4,autos)
         call autooutea(psi1,psi2,dicht4o,dicht4,auto(1,1),false)
      endif

!-----------------------------------------------------------------------
! Write 'auto'-file output
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = 1,npacket
            write(iaut,'(1x,g18.9,3(1x,f18.14))') &
                  atime,dble(auto(p1,p2)),dimag(auto(p1,p2)), &
                  abs(auto(p1,p2))
         enddo
      enddo
      call myflush(iaut)
      
      return
      end subroutine autoout2

!***********************************************************************
!                               AUTOLAP
!
! Computes the overlap of two wave functions as needed for
! autocorrelation functions, i.e. auto = <psi1^*|psi2>. Autos contains
! the corresponding overlap for each state.
! It is assumed that the number of spfs for any mode is the same for
! psi1 and psi2.
!
! V7.1 MB
!***********************************************************************

      subroutine autolap (psi1,psi2,auto,autos,p1,p2,jindx,agmat1,agmat2)

      implicit none

      integer(long)                                    :: zeig1,zeig2,&
                                                          m,s,s1,s2,&
                                                          block1,block2,blk
      integer(long), intent(in)                        :: p1,p2
      integer(long), dimension(jindxdim), intent(in)   :: jindx
      integer(long), dimension(2*fmblock)              :: tmpindx
      complex(dop), intent(out)                        :: auto
      complex(dop), dimension(dgldim), intent(in)      :: psi1,psi2
      complex(dop), dimension(2*maxblock)              :: tmppsi
      complex(dop), dimension(dmatdim)                 :: tmpovl
      complex(dop), dimension(maxsta), intent(out)     :: autos
      complex(dop), dimension(agmatdim), intent(in)    :: agmat1
      complex(dop), dimension(agmatdim), intent(in)    :: agmat2
      complex(dop), dimension(gfgwpdim)                :: gfgwp
      complex(dop), dimension(aigwpdim)                :: aigwp
      complex(dop), dimension(zgwpdim)                 :: zgwp
 
      if(lmulpack .and. leb ) goto 30  ! single-set multi-packet WF

!-----------------------------------------------------------------------
! Compute overlap <psi1^*|psi2>
!-----------------------------------------------------------------------
      auto = (0.0_dop,0.0_dop)
      tmppsi=0.0_dop
      tmpindx=0
      tmpovl=0.0_dop
      do s = 1,nstate/npacket
         s1 = nstate/npacket*(p1-1)+s
         s2 = nstate/npacket*(p2-1)+s

         if (citype .gt. 0) then
            zeig1 = 1
            zeig2 = fmblock+1 
            block2=block(s)
            call cpvxi(jindx(zpsi(s2)),tmpindx(zeig2),block2)
            call cpvxz(psi2(zpsi(s2)),tmppsi(zeig2),block2)

            do m = 1,nmode
               call mmtxzz(psi1(zetf(m,s1)),psi2(zetf(m,s2)), &
                    tmpovl(dmat(m,s)),subdim(m),dim(m,s),dim(m,s))
               block1=vdim(m,s)*dim(m,s)*ndim(m,s)
               call venpsi(m,jindx(zpsi(s)),tmpindx(zeig1), &
                    tmpindx(zeig2),block(s),block1,block2, &
                    nmode,vdim(m,s),dim(m,s),ndim(m,s),jvdim, &
                    tmppsi(zeig1),tmppsi(zeig2))
               block2=block1
               call cpvxi(tmpindx(zeig1),tmpindx(zeig2),block1)

               call qtxxzz(tmpovl(dmat(m,s)),tmppsi(zeig1),tmppsi(zeig2), &
                           vdim(m,s),dim(m,s),ndim(m,s))
            enddo
            call cpvxz_s(tmppsi(zeig2),tmppsi(zeig1),tmpindx(zeig2), &
                    jindx(zpsi(s)),block2,block(s))
            call swapsxxi(zeig1,zeig2)
            call vvtxzz(psi1(zpsi(s1)),tmppsi(zeig2),autos(s),block(s))

         else
! calculate overlaps
            do m = 1,nmode
               if (gwpm(m)) then
                  call diagonal_moments(m,s1,s2,nspfdof(m),gwpdim(m,s1),&
                       gwpdim(m,s2),agmat1(zaGmat(m,s1)),agmat2(zaGmat(m,s2)),&
                       agmat1(zxivec(m,s1)),agmat2(zxivec(m,s2)),&
                       agmat1(zphase(m,s1)),agmat2(zphase(m,s2)),& 
                       gfgwp(zgfgwp(m,s1,s2)),aigwp(zaigwp(m,s1,s2)),&
                       zgwp(zzgwp(m,s1,s2)),1,0)
! gfgwp contains overlap
                  if (psitype .eq. 0 .or. psitype .eq. 5) then
                     call simtran2tz(psi1(zdgwp(m,s1)),gfgwp(zgfgwp(m,s1,s2)),&
                          psi2(zdgwp(m,s2)),tmpovl(dmat(m,s)),gwpdim(m,s),&
                          gwpdim(m,s),dim(m,s),dim(m,s))
                  else
                     call cpqxz(gfgwp(zgfgwp(m,s1,s2)),tmpovl(dmat(m,s)),&
                          gwpdim(m,s))
                  endif
               else
                  call mmtxzz(psi1(zetf(m,s1)),psi2(zetf(m,s2)),&
                       tmpovl(dmat(m,s)),subdim(m),dim(m,s),dim(m,s))
               endif
            enddo

            if(.not.lrunml) then
               zeig1 = 1
               zeig2 = block(s)+1
               call cpvxz(psi2(zpsi(s2)),tmppsi(zeig2),block(s))
               do m = 1,nmode
                  call swapsxxi(zeig1,zeig2)
                  call qtxxzz(tmpovl(dmat(m,s)),tmppsi(zeig1),tmppsi(zeig2),&
                       vdim(m,s),dim(m,s),ndim(m,s))
               enddo
               call vvtxzz(psi1(zpsi(s1)),tmppsi(zeig2),autos(s),block(s))
            else
               call automl(autos(s),psi1,psi2,tmpovl)
            endif
         endif

         auto = auto+autos(s)
      enddo
      return

!-----------------------------------------------------------------------
! Compute overlap <psi1^*|psi2>  single-set multi-packet
!-----------------------------------------------------------------------
  30  blk  = adim/npacket
      zeig1 = 1
      zeig2 = blk+1
      s1 = (p1-1)*blk + 1
      s2 = (p2-1)*blk + 1

      call cpvxz(psi2(s2),tmppsi(zeig2),blk)
      do m = 1,nmode-1  ! Do not run over the electronic mode.
         call mmtxzz(psi1(zetf(m,1)),psi2(zetf(m,1)),tmpovl,subdim(m),&
              dim(m,1),dim(m,1))
         call swapsxxi(zeig1,zeig2)
         call qtxxzz(tmpovl,tmppsi(zeig1),tmppsi(zeig2),&
                           vdim(m,1),dim(m,1),ndim(m,1)/npacket)
      enddo

      call vvtxzz(psi1(s1),tmppsi(zeig2),auto,blk)


    return
    end subroutine autolap

!***********************************************************************
!                                AUTOOUTHIGH
!
! Outputs autocorrelation function of order 1 and 2 at time t to files
! auto1 and auto2, respectively.
! Note that the computation of the second order autocorrelation function
! requires (in its current form) a symmetric Hamiltonian!
!
! V7.1 MB
!***********************************************************************

      subroutine autoouthigh (psi,dtpsi,atime)

      implicit none

      real(dop), intent(in)                            :: atime
      complex(dop), dimension(dgldim), intent(in)      :: psi,dtpsi
      complex(dop), dimension(1:maxsta,1:maxsta)       :: auto1,auto2

      logical(kind=4)                                  :: order2
      integer(long)                                    :: p1, p2
      complex(dop)                                     :: ctemp,dummy

!-----------------------------------------------------------------------
! Loop over all combinations of packets (to produce cross-correlation)
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = p2,npacket

!-----------------------------------------------------------------------
! Compute first and second order autocorrelation functions
! <psi^*|H|psi> = i<psi^*|dpsi/dt> and <psi^*|H^2|psi> =
! -<dpsi/dt^*|dpsi/dt> (the latter equation only holds if H is
! symmetric).
!-----------------------------------------------------------------------
            if (lautoord2) then
               call autohigh(psi,dtpsi,psi,dtpsi, &
                             auto1(p1,p2),auto2(p1,p2),p1,p2,lautoord2)
               if (p1 .ne. p2) auto2(p2,p1) = auto2(p1,p2)
            else
               call autohigh(psi,dtpsi,psi,dtpsi, &
                             auto1(p1,p2),auto2(p1,p2),p1,p2,lautoord2)
            endif

!-----------------------------------------------------------------------
! For first order cross-correlation explicitly compute upper right part.
! This was found numerically more precise than mirroring the matrix.
!-----------------------------------------------------------------------
            if (p1 .ne. p2) then
               order2 = .false.
               call autohigh(psi,dtpsi,psi,dtpsi, &
                             auto1(p2,p1),dummy,p2,p1,order2)
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! Symmetrise first order cross-correlation matrix
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = p2+1,npacket
            ctemp        = 0.5_dop*(auto1(p1,p2)+auto1(p2,p1))
            auto1(p1,p2) = ctemp
            auto1(p2,p1) = ctemp
         enddo
      enddo

!-----------------------------------------------------------------------
! Write data to auto files
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = 1,npacket
            if(abs(auto1(p1,p2)) .gt. 9.9_dop) then
               write(iaut1,'(1x,g18.9,3(1x,f18.10))',err=10) &
                    atime,dble(auto1(p1,p2)),dimag(auto1(p1,p2)), &
                    abs(auto1(p1,p2))
            else
               write(iaut1,'(1x,g18.9,3(1x,f18.14))',err=10) &
                    atime,dble(auto1(p1,p2)),dimag(auto1(p1,p2)), &
                    abs(auto1(p1,p2))
            endif
            if (lautoord2) then
               if(abs(auto2(p1,p2)) .gt. 9.9_dop) then
                  write(iaut2,'(1x,g18.9,3(1x,f18.10))',err=10) &
                        atime,dble(auto2(p1,p2)),dimag(auto2(p1,p2)), &
                        abs(auto2(p1,p2))
               else
                  write(iaut2,'(1x,g18.9,3(1x,f18.14))',err=10) &
                        atime,dble(auto2(p1,p2)),dimag(auto2(p1,p2)), &
                        abs(auto2(p1,p2))
               endif
            endif
         enddo
      enddo
      call myflush(iaut1)
      if (lautoord2) call myflush(iaut2)

      return

!-----------------------------------------------------------------------
! Write error message
!-----------------------------------------------------------------------
 10   continue
      routine = 'autoouthigh'
      message = 'Cannot write data to file auto1 or auto2'
      call errormsg

      end subroutine autoouthigh

!***********************************************************************
!                             AUTOOUT2HIGH
!
! Outputs autocorrelation function of order 1 and 2 at time t/2 to files
! auto1 and auto2, respectively. (Used in a CMF calculation with
! adaptive step size.)
! Whereas "autoouthigh" writes the autocorrelation function c(2nt) =
! <psi^*(nt)|H^1,2|psi(nt)> to file auto1,2, this routine writes
! c(2(n-1)t) = <psi^*((n-1)t)|H^1,2|psi(nt)>.
! Note that the computation of the second order autocorrelation function
! requires (in its current form) a symmetric Hamiltonian!
!
! V7.1 MB
!***********************************************************************

      subroutine autoout2high (psi1,dtpsi1,psi2,dtpsi2,atime)

      real(dop), intent(in)                            :: atime
      complex(dop), dimension(dgldim), intent(in)      :: psi1, dtpsi1, &
                                                          psi2, dtpsi2
      complex(dop), dimension(1:maxsta,1:maxsta)       :: auto1,auto2

      logical(kind=4)                                  :: order2
      integer(long)                                    :: p1, p2
      complex(dop)                                     :: ctemp,dummy

!-----------------------------------------------------------------------
! Loop over all combinations of packets (to produce cross-correlation)
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = p2,npacket

!-----------------------------------------------------------------------
! Compute first and second order autocorrelation function
!-----------------------------------------------------------------------
            if (lautoord2) then
               call autohigh(psi1,dtpsi1,psi2,dtpsi2, &
                             auto1(p1,p2),auto2(p1,p2),p1,p2,lautoord2)
               if (p1 .ne. p2) auto2(p2,p1) = auto2(p1,p2)
            else
               call autohigh(psi1,dtpsi1,psi2,dtpsi2, &
                             auto1(p1,p2),auto2(p1,p2),p1,p2,lautoord2)
            endif
      
!-----------------------------------------------------------------------
! For first order cross-correlation explicitly compute upper right part.
! This was found numerically more precise than mirroring the matrix.
!-----------------------------------------------------------------------
            if (p1 .ne. p2) then
               order2 = .false.
               call autohigh(psi1,dtpsi1,psi2,dtpsi2, &
                             auto1(p2,p1),dummy,p2,p1,order2)
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! Symmetrise first order cross-correlation matrix
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = p2+1,npacket
            ctemp        = 0.5_dop*(auto1(p1,p2)+auto1(p2,p1))
            auto1(p1,p2) = ctemp
            auto1(p2,p1) = ctemp
         enddo
      enddo

!-----------------------------------------------------------------------
! Write data to auto files
!-----------------------------------------------------------------------
      do p2 = 1,npacket
         do p1 = 1,npacket
            write(iaut1,'(1x,g18.9,3(1x,f18.14))',err=10) &
                  atime,dble(auto1(p1,p2)),dimag(auto1(p1,p2)), &
                  abs(auto1(p1,p2))
            if (lautoord2) then
               write(iaut2,'(1x,g18.9,3(1x,f18.14))',err=10) &
                     atime,dble(auto2(p1,p2)),dimag(auto2(p1,p2)), &
                     abs(auto2(p1,p2))
            endif
         enddo
      enddo
      call myflush(iaut1)
      if (lautoord2) call myflush(iaut2)

      return

!-----------------------------------------------------------------------
! Write error message
!-----------------------------------------------------------------------
 10   continue
      routine = 'autoout2high'
      message = 'Cannot write data to file auto1 or auto2'
      call errormsg

      end subroutine autoout2high

!***********************************************************************
!                               AUTOHIGH
!
! Computes autocorrelation function of order 1 and 2, i.e. 
! auto1,2 = <psi1^*|H^1,2|psi2>.
! It is assumed that the number of spfs for any mode is the same for
! psi1 and psi2.
!
! We use the notations A1,2'=dA1,2/dt, <k>=<phi1^*(k)|phi2(k)>,
! <k'>=<phi1^*(k)|dtphi2(k)>, and <'k>=<dtphi1^*(k)|phi2(k)>, where k
! denotes the degree of freedom. Further we omit all indices and sums,
! e.g. <3> A = Sum_i3 <phi1^*_j3(3)|phi2_i3(3)> A_i1...if.
!
! V7.1 MB
!***********************************************************************

      subroutine autohigh (psi1,dtpsi1,psi2,dtpsi2,auto1,auto2,p1,p2,order2)

      implicit none

      logical(kind=4), intent(in)                      :: order2
      integer(long), intent(in)                        :: p1,p2
      complex(dop), intent(out)                        :: auto1,auto2
      complex(dop), dimension(dgldim), intent(in)      :: psi1,dtpsi1,psi2,dtpsi2
      complex(dop), dimension(3*maxblock)              :: aaux
      complex(dop), dimension(spfovrdim)               :: ovr1,ovr2,ovr3

      integer(long)                                    :: zeig1,zeig2,zeig3,&
                                                          s,m,m1,m2,s1,s2
      logical(kind=4)                                  :: first1,first2
      logical(kind=4), dimension(maxdim)               :: modeincl

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      auto1 = (0.0_dop,0.0_dop)
      auto2 = (0.0_dop,0.0_dop)

!-----------------------------------------------------------------------
! Compute spf overlap matrices <k>, <k'>, and <'k> for all k. See above
! for notation.
!-----------------------------------------------------------------------
      do s = 1,nstate/npacket
         s1 = nstate/npacket*(p1-1)+s
         s2 = nstate/npacket*(p2-1)+s
         do m = 1,nmode
            call mmtxzz(psi1(zetf(m,s1)),psi2(zetf(m,s2)), &
                        ovr1(zspfovr(m,s)),subdim(m),dim(m,s),dim(m,s))
            call mmtxzz(psi1(zetf(m,s1)),dtpsi2(zetf(m,s2)), &
                        ovr2(zspfovr(m,s)),subdim(m),dim(m,s),dim(m,s))
            if (order2) then
               if (lexact) then
                  call mmtxzz(dtpsi1(zetf(m,s1)),dtpsi2(zetf(m,s2)), &
                              ovr3(zspfovr(m,s)),subdim(m),dim(m,s), &
                              dim(m,s))
               else
                  call mmtxzz(dtpsi1(zetf(m,s1)),psi2(zetf(m,s2)), &
                              ovr3(zspfovr(m,s)),subdim(m),dim(m,s), &
                              dim(m,s))
               endif
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! For exact wavepacket we have all we need.
!-----------------------------------------------------------------------
      if (lexact) then
         do s = 1,nstate/npacket
            do m = 1,nmode
               auto1 = auto1+ovr2(zspfovr(m,s))
               if (order2) auto2 = auto2+ovr3(zspfovr(m,s))
            enddo
         enddo
         goto 100
      endif

!-----------------------------------------------------------------------
! Compute A1<f>...<2><1'>A2 and A1'<f>...<2><1'>A2 ...
! A1<f'><f-1>...<1>A2 and A1'<f'><f-1>...<1>A2. Compare the calculation
! of the mean-fields.
! For each state s and mode m do:
! 1. Compute <f><f-1>...<m+1><m-1>...<1>A2. Store <m-1>...<1>A2 in
!    aaux(zeig3) 
! 2. Multiply with <m'>
! 3. Multiply with A1 or A1'
!-----------------------------------------------------------------------
      do s = 1,nstate/npacket
         s1 = nstate/npacket*(p1-1)+s
         s2 = nstate/npacket*(p2-1)+s
         zeig1 = 1
         zeig2 = block(s)+1
         zeig3 = 2*block(s)+1
         do m = 1,nmode
            if (m .eq. 1) then
               call cpvxz(psi2(zpsi(s2)),aaux(zeig2),block(s))
            elseif (m .eq. 2) then
               call cpvxz(psi2(zpsi(s2)),aaux(zeig1),block(s))
            else
               call cpvxz(aaux(zeig3),aaux(zeig1),block(s))
            endif
            if (m .ge. 2) then
               call qtxxzz(ovr1(zspfovr(m-1,s)),aaux(zeig1),aaux(zeig2), &
                           vdim(m-1,s),dim(m-1,s),ndim(m-1,s))
               call cpvxz(aaux(zeig2),aaux(zeig3),block(s))
            endif
            do m1 = m+1,nmode
               call swapsxxi(zeig1,zeig2)
               call qtxxzz(ovr1(zspfovr(m1,s)),aaux(zeig1),aaux(zeig2), &
                           vdim(m1,s),dim(m1,s),ndim(m1,s))
            enddo
            call swapsxxi(zeig1,zeig2)
            call qtxxzz(ovr2(zspfovr(m,s)),aaux(zeig1),aaux(zeig2), &
                        vdim(m,s),dim(m,s),ndim(m,s))
            call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto1,block(s))
            if (order2) then
               call vvtxzza(dtpsi1(zpsi(s1)),aaux(zeig2),auto2,block(s))
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! If 2nd order autocorrelation is desired compute A1<f>...<2><'1>A2' ...
! A1<'f><f-1>...<1>A2' and (as a by-product) A1<f>...<2><1>A2' and
! A1'<f>...<2><1>A2', else directly compute A1<f>...<2><1>A2'.
! Procedure is the same as above.
!-----------------------------------------------------------------------
      if (order2) then
         do s = 1,nstate/npacket
            s1 = nstate/npacket*(p1-1)+s
            s2 = nstate/npacket*(p2-1)+s
            zeig1 = 1
            zeig2 = block(s)+1
            zeig3 = 2*block(s)+1
            do m = 1,nmode
               if (m .eq. 1) then
                  call cpvxz(dtpsi2(zpsi(s2)),aaux(zeig2),block(s))
               elseif (m .eq. 2) then
                  call cpvxz(dtpsi2(zpsi(s2)),aaux(zeig1),block(s))
               else
                  call cpvxz(aaux(zeig3),aaux(zeig1),block(s))
               endif
               if (m .ge. 2) then
                  call qtxxzz(ovr1(zspfovr(m-1,s)),aaux(zeig1), &
                              aaux(zeig2),vdim(m-1,s),dim(m-1,s), &
                              ndim(m-1,s))
                  call cpvxz(aaux(zeig2),aaux(zeig3),block(s))
               endif
               do m1 = m+1,nmode
                  call swapsxxi(zeig1,zeig2)
                  call qtxxzz(ovr1(zspfovr(m1,s)),aaux(zeig1), &
                              aaux(zeig2),vdim(m1,s),dim(m1,s), &
                              ndim(m1,s))
               enddo
               call swapsxxi(zeig1,zeig2)
               call qtxxzz(ovr3(zspfovr(m,s)),aaux(zeig1),aaux(zeig2), &
                           vdim(m,s),dim(m,s),ndim(m,s))
               call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto2,block(s))
            enddo
            call qtxxzz(ovr1(zspfovr(nmode,s)),aaux(zeig3),aaux(zeig2), &
                        vdim(nmode,s),dim(nmode,s),ndim(nmode,s))
            call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto1,block(s))
            call vvtxzza(dtpsi1(zpsi(s1)),aaux(zeig2),auto2,block(s))
         enddo
      else
         do s = 1,nstate/npacket
            s1 = nstate/npacket*(p1-1)+s
            s2 = nstate/npacket*(p2-1)+s
            zeig1 = 1
            zeig2 = block(s)+1
            call cpvxz(dtpsi2(zpsi(s2)),aaux(zeig2),block(s))
            do m = 1,nmode
               call swapsxxi(zeig1,zeig2)
               call qtxxzz(ovr1(zspfovr(m,s)),aaux(zeig1),aaux(zeig2), &
                           vdim(m,s),dim(m,s),ndim(m,s))
            enddo
            call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto1,block(s))
         enddo
      endif

!-----------------------------------------------------------------------
! If 2nd order autocorrelation is desired compute for all m1 != m2:
! A1<f><f-1>...<m1+1><'m1><m1-1>...<m2+1><m2'><m2-1>...<2><1>A2
!-----------------------------------------------------------------------
      if (order2) then
         do s = 1,nstate/npacket
            s1 = nstate/npacket*(p1-1)+s
            s2 = nstate/npacket*(p2-1)+s
            zeig1 = 1
            zeig2 = block(s)+1
            zeig3 = 2*block(s)+1
            do m = 1,nmode
               call zerovxl(modeincl,nmode)
               call qtxxzz(ovr3(zspfovr(m,s)),psi2(zpsi(s2)), &
                           aaux(zeig3),vdim(m,s),dim(m,s),ndim(m,s))
               modeincl(m) = .true.
               first1 = .true.
               do m1 = 1,nmode
                  if (m1 .ne. m) then
                     call cpvxz(aaux(zeig3),aaux(zeig2),block(s))
                     first2 = .true.
                     do m2 = 1,nmode
                        if (.not. modeincl(m2)) then
                           call swapsxxi(zeig1,zeig2)
                           if (m2 .ne. m1) then 
                              call qtxxzz(ovr1(zspfovr(m2,s)), &
                                          aaux(zeig1),aaux(zeig2), &
                                          vdim(m2,s),dim(m2,s), &
                                          ndim(m2,s))
                           else
                              call qtxxzz(ovr2(zspfovr(m2,s)),&
                                          aaux(zeig1),aaux(zeig2),&
                                          vdim(m2,s),dim(m2,s),&
                                          ndim(m2,s))
                           endif
                           if (.not. first1) then
                              if (first2) then
                                 call cpvxz(aaux(zeig2),aaux(zeig3),&
                                            block(s))
                                 modeincl(m2) = .true.
                              endif
                           endif
                           first2 = .false.
                        endif
                     enddo
                     call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto2,&
                                  block(s))
                     first1 = .false.
                  endif
               enddo
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! If 2nd order autocorrelation is desired compute for all m:
! A1<f><f-1>...<m+1><'m'><m-1>...<2><1>A2
!-----------------------------------------------------------------------
      if (order2) then
         do s = 1,nstate/npacket
            s1 = nstate/npacket*(p1-1)+s
            s2 = nstate/npacket*(p2-1)+s
            zeig1 = 1
            zeig2 = block(s)+1
            zeig3 = 2*block(s)+1
            do m = 1,nmode
               call mmtxzz(dtpsi1(zetf(m,s1)),dtpsi2(zetf(m,s2)), &
                           ovr2(zspfovr(m,s)),subdim(m),dim(m,s), &
                           dim(m,s))
               if (m .eq. 1) then
                  call cpvxz(psi2(zpsi(s2)),aaux(zeig2),block(s))
               elseif (m .eq. 2) then
                  call cpvxz(psi2(zpsi(s2)),aaux(zeig1),block(s))
               else
                  call cpvxz(aaux(zeig3),aaux(zeig1),block(s))
               endif
               if (m .ge. 2) then
                  call qtxxzz(ovr1(zspfovr(m-1,s)),aaux(zeig1), &
                              aaux(zeig2),vdim(m-1,s),dim(m-1,s), &
                              ndim(m-1,s))
                  call cpvxz(aaux(zeig2),aaux(zeig3),block(s))
               endif
               do m1 = m+1,nmode
                  call swapsxxi(zeig1,zeig2)
                  call qtxxzz(ovr1(zspfovr(m1,s)),aaux(zeig1), &
                              aaux(zeig2),vdim(m1,s),dim(m1,s), &
                              ndim(m1,s))
               enddo
               call swapsxxi(zeig1,zeig2)
               call qtxxzz(ovr2(zspfovr(m,s)),aaux(zeig1),aaux(zeig2), &
                           vdim(m,s),dim(m,s),ndim(m,s))
               call vvtxzza(psi1(zpsi(s1)),aaux(zeig2),auto2,block(s))
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! This routine assumes "dtpsi" to contain H|psi>, as was the case in an
! older program version. Now "dtpsi" contains -iH|psi>, so correct for
! the prefactor.
!-----------------------------------------------------------------------
 100  continue
      auto1 = (0.0_dop,1.0_dop)*auto1
      auto2 = -auto2

      return
      end subroutine autohigh

!***********************************************************************
!                              AUTOOUTE
!
! Calculate autocorrelation without one hightest natural orbital:
!
!     1. Calculate overlap matrices and store in ovm.
!     2. Calculate mean-field matrices and store in mfm.
!     3. Transform both to natural orbital picture (if necessary).
!     4. Calculate estimate of changed autocorrelation.
!
!     AR, jan 98
!***********************************************************************

      subroutine autooute(psi1,psi2,dicht31,dicht41,dicht32,dicht42, &
                 autos)

      implicit none

      integer(long)             :: zeig,m,m1,s
      real(dop), intent(in)     :: dicht31(d3matdim),dicht32(d3matdim)
      complex(dop), intent(in)  :: psi1(dgldim),psi2(dgldim) 
      complex(dop), intent(in)  :: dicht41(dmatdim),dicht42(dmatdim)
      complex(dop), intent(in)  :: autos(maxsta)
      complex(dop)              :: autoe,auto,autoes
      complex(dop)              :: workc(2*maxblock),ovm(dmatdim),&
                                   mfm(dmatdim)

!-----------------------------------------------------------------------
! Calculate overlap matrices and store in ovm
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call mmtxzz(psi1(zetf(m,s)),psi2(zetf(m,s)),ovm(dmat(m,s)),&
                        subdim(m),dim(m,s),dim(m,s))
         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate mean-field matrices and store in mfm
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
               call cpvxz(psi2(zpsi(s)),workc,block(s))
               zeig=block(s)+1
               do m1=1,nmode
                  if (m1 .ne. m) then
                     call qtxxzz(ovm(dmat(m1,s)),workc,workc(zeig), &
                                 vdim(m1,s),dim(m1,s),ndim(m1,s))
                     call cpvxz(workc(zeig),workc,block(s))
                  endif
               enddo
               call calcmfs(psi1(zpsi(s)),workc,mfm(dmat(m,s)), &
                            vdim(m,s),dim(m,s),ndim(m,s))
         enddo
      enddo

!-----------------------------------------------------------------------
! Transform to natural orbitals (if necessary)
!-----------------------------------------------------------------------
      if (lnaturcmf) then
         do s=1,nstate
            do m=1,nmode
               call trafoov(ovm(dmat(m,s)),dicht41(dmat(m,s)), &
                            dicht42(dmat(m,s)),dim(m,s))
               call trafomf(mfm(dmat(m,s)),dicht41(dmat(m,s)), &
                            dicht42(dmat(m,s)),dim(m,s))
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! Calculate autoe for single natural orbital (autoes) and write to
! 'autoe'-file
!-----------------------------------------------------------------------
      auto=(0.0d0,0.0d0)
      do s=1,nstate
         auto = auto + autos(s)
      enddo
      do s=1,nstate
         do m=1,nmode
            call auterrs(autoes,ovm(dmat(m,s)),mfm(dmat(m,s)), &
                         dicht31(d3mat(m,s)),dicht32(d3mat(m,s)), &
                         dim(m,s))
            autoe = auto - autos(s) + autoes
            write(iaute) autoe
         enddo
      enddo

      return
      end subroutine autooute

!***********************************************************************
!                              AUTOOUTEA
!
! Calculate autocorrelation neglegting all hightest natural orbitals:
!
!     1. Calculate overlap matrices and store in dicht1.
!     2. Calculate mean-field matrices and store in dicht2.
!     3. Transform both to natural orbital picture (if necessary).
!     4. Calculate estimate of changed autocorrelation.
!
!     AR, jan 98
!***********************************************************************

      subroutine autooutea(psi1,psi2,dicht41,dicht42,auto,lpsiequ)

      implicit none

      integer(long)               :: zeig,m,s,b
      logical(kind=4), intent(in) :: lpsiequ
      real(dop)                   :: nor1,nor2
      complex(dop), intent(in)    :: psi1(dgldim),psi2(dgldim),auto, &
                                     dicht41(dmatdim),dicht42(dmatdim)
      complex(dop)                :: workc(2*maxblock),ovm(dmatdim), &
                                     autoe

!-----------------------------------------------------------------------
! Calculate changed norms:
!
!     1. Calculate changed overlap for psi1
!     2. Calculate norm of psi1
!     3. Do the same for psi2 if necessary
!-----------------------------------------------------------------------
      nor1=0.0d0
      do s=1,nstate
         do m=1,nmode
            call overmxz(psi1(zetf(m,s)),ovm(dmat(m,s)),subdim(m),dim(m,s))
            if (lnaturcmf) then
               call chovms2(ovm(dmat(m,s)),dim(m,s))
            else
               call chovm(ovm(dmat(m,s)),dicht41(dmat(m,s)),dim(m,s))
            endif
         enddo
      enddo
      do s=1,nstate
         call cpvxz(psi1(zpsi(s)),workc,block(s))
         zeig=block(s)+1
         do m=1,nmode
            call qtxxzz(ovm(dmat(m,s)),workc,workc(zeig),vdim(m,s),&
                 dim(m,s),ndim(m,s))
            call cpvxz(workc(zeig),workc,block(s))
         enddo
         do b=1,block(s)
            nor1 = nor1 + dble(dconjg(psi1(zpsi(s)+b-1))*workc(b))
         enddo
      enddo

      if (.not. lpsiequ) then
         nor2=0.0d0
         do s=1,nstate
            do m=1,nmode
               call overmxz(psi2(zetf(m,s)),ovm(dmat(m,s)),subdim(m),dim(m,s))
               if (lnaturcmf) then
                  call chovms2(ovm(dmat(m,s)),dim(m,s))
               else
                  call chovm(ovm(dmat(m,s)),dicht42(dmat(m,s)),dim(m,s))
               endif
            enddo
         enddo
         do s=1,nstate
            call cpvxz(psi2(zpsi(s)),workc,block(s))
            zeig=block(s)+1
            do m=1,nmode
               call qtxxzz(ovm(dmat(m,s)),workc,workc(zeig),vdim(m,s),&
                    dim(m,s),ndim(m,s))
               call cpvxz(workc(zeig),workc,block(s))
            enddo
            do b=1,block(s)
               nor2 = nor2 + dble(dconjg(psi2(zpsi(s)+b-1))*workc(b))
            enddo
         enddo
      else
         nor2 = nor1
      endif

!-----------------------------------------------------------------------
! Calculate changed autocorrelation:
!
!     1. Calculate overlap matrices and store in ovm
!     2. Calculate changed overlaps
!     3. Calculate autocorrelation
!-----------------------------------------------------------------------
      do s=1,nstate
         do m=1,nmode
            call mmtxzz(psi1(zetf(m,s)),psi2(zetf(m,s)), &
                    ovm(dmat(m,s)),subdim(m),dim(m,s),dim(m,s))

            if (lnaturcmf) then
               call chovms2(ovm(dmat(m,s)),dim(m,s))
            else
               call chovms1(ovm(dmat(m,s)),dicht41(dmat(m,s)), &
                            dicht42(dmat(m,s)),dim(m,s))
            endif
         enddo
      enddo

      autoe=(0.0d0,0.0d0)
      do s=1,nstate
         call cpvxz(psi1(zpsi(s)),workc,block(s))
         zeig=block(s)+1
         do m=1,nmode
            call qtxxzz(ovm(dmat(m,s)),workc,workc(zeig),vdim(m,s), &
                        dim(m,s),ndim(m,s))
            call cpvxz(workc(zeig),workc,block(s))
         enddo
         do b=1,block(s)
            autoe = autoe + psi2(zpsi(s)+b-1)*workc(b)
         enddo
      enddo
      autoe = autoe / sqrt(nor1*nor2)

!-----------------------------------------------------------------------
! Write 'autoe'-file output
!-----------------------------------------------------------------------
      write(iaute) autoe, auto
      call myflush(iaute)

      return
      end subroutine autooutea

!***********************************************************************
!                              CALCMFS
!
!     AR, oct 97
!***********************************************************************

      subroutine calcmfs(psi1,psi2,mfield,dim1,dim2,dim3)

      implicit none

      integer(long), intent(in) :: dim1,dim2,dim3
      integer(long)             :: i,j,k,l
      complex(dop), intent(in)  :: psi1(dim1,dim2,dim3), &
                                   psi2(dim1,dim2,dim3)
      complex(dop), intent(out) :: mfield(dim2,dim2)

      do i=1,dim2
         do j=1,dim2
            mfield(i,j)=(0.0d0,0.0d0)
            do l=1,dim3
               do k=1,dim1
                  mfield(i,j) = mfield(i,j) + psi1(k,i,l) * psi2(k,j,l)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine calcmfs

!***********************************************************************
!                              TRAFOOV
!
!     AR, oct 97
!***********************************************************************

      subroutine trafoov(ovm,trafom1,trafom2,dim)

      implicit none

      integer(long), intent(in)   :: dim
      complex(dop), intent(inout) :: ovm(dim,dim)
      complex(dop), intent(in)    :: trafom1(dim,dim),trafom2(dim,dim)
      complex(dop)                :: tovm(dim,dim)

      call mmaxzz(trafom1,ovm,tovm,dim,dim,dim)
      call mqxczz(tovm,trafom2,ovm,dim,dim)

      return
      end subroutine trafoov

!***********************************************************************
!                               TRAFOMF
!
!     AR, oct 97
!***********************************************************************

      subroutine trafomf(mfm,trafom1,trafom2,dim)

      implicit none

      integer(long), intent(in)   :: dim
      complex(dop), intent(inout) :: mfm(dim,dim)
      complex(dop), intent(in)    :: trafom1(dim,dim),trafom2(dim,dim)
      complex(dop)                :: tmfm(dim,dim)


      call qqtxzz(trafom1,mfm,tmfm,dim)
      call qqxxzz(tmfm,trafom2,mfm,dim)

      return
      end subroutine trafomf

!***********************************************************************
!                               AUTERRS
!
!     AR, sep 97
!***********************************************************************

      subroutine auterrs(err,ovm,mfm,dicht31,dicht32,dim)

      implicit none

      integer(long), intent(in) :: dim
      integer(long)             :: d1,d2
      real(dop), intent(in)     :: dicht31(dim),dicht32(dim)
      complex(dop), intent(in)  :: ovm(dim,dim),mfm(dim,dim)
      complex(dop), intent(out) :: err

      err=(0.0d0,0.0d0)
      do d1=2,dim
         do d2=2,dim
            err=err+mfm(d2,d1)*ovm(d2,d1)
         enddo
      enddo
      err=err/sqrt( (1.0d0-dicht31(1))*(1.0d0-dicht32(1)) )

      return
      end subroutine auterrs

!***********************************************************************
!                               CHOVM
!
!     Calculate changed overlap matrix for psi':
!        1. Transform to natural orbitals
!        2. Transform back and consider change
!
!     AR, jan 98
!***********************************************************************

      subroutine chovm(ovm,evecs,dim)

      implicit none

      integer(long), intent(in)   :: dim
      integer(long)               :: j,k,m,n
      complex(dop), intent(inout) :: ovm(dim,dim)
      complex(dop), intent(in)    :: evecs(dim,dim)
      complex(dop)                :: ovm2(dim,dim)

      call mqxczz(ovm,evecs,ovm2,dim,dim)
      call qqtxzz(evecs,ovm2,ovm,dim)

      do k=1,dim
         do n=2,dim
            ovm2(n,k) = ovm(n,2)*evecs(k,2)
         enddo
      enddo
      do m=3,dim
         do k=1,dim
            do n=2,dim
               ovm2(n,k) = ovm2(n,k) + ovm(n,m)*evecs(k,m)
            enddo
         enddo
      enddo

      do k=1,dim
         do j=1,dim
            ovm(j,k) = dconjg(evecs(j,2))*ovm2(2,k)
            do n=3,dim
               ovm(j,k) = ovm(j,k) + dconjg(evecs(j,n))*ovm2(n,k)
            enddo
         enddo
      enddo

      return
      end subroutine chovm

!***********************************************************************
!                              CHOVMS1
!
!     Calculate changed overlap matrix <psi1'*,psi2'>:
!        1. Transform to natural orbitals
!        2. Transform back and consider change
!
!     AR, jan 98
!***********************************************************************

      subroutine chovms1(ovm,evecs1,evecs2,dim)

      implicit none

      integer(long), intent(in)   :: dim
      integer(long)               :: j,k,l,m
      complex(dop), intent(inout) :: ovm(dim,dim)
      complex(dop), intent(in)    :: evecs1(dim,dim), evecs2(dim,dim)
      complex(dop)                :: ovm2(dim,dim)

      call trafoov(ovm,evecs1,evecs2,dim)

      do k=1,dim
         do l=2,dim
            ovm2(l,k)=ovm(l,2)*evecs2(k,2)
         enddo
      enddo
      do m=3,dim
         do k=1,dim
            do l=2,dim
               ovm2(l,k) = ovm2(l,k) + ovm(l,m)*evecs2(k,m)
            enddo
         enddo
      enddo

      do k=1,dim
         do j=1,dim
            ovm(j,k) = evecs1(j,2)*ovm2(2,k)
            do l=3,dim
               ovm(j,k) = ovm(j,k) + evecs1(j,l)*ovm2(l,k)
            enddo
         enddo
      enddo

      return
      end subroutine chovms1

!***********************************************************************
!                              CHOVMS2
!
!     Calculate changed overlap matrix <psi1'*,psi2'> for lnat=.true.
!
!     AR, jan 98
!***********************************************************************

      subroutine chovms2(ovm,dim)

      implicit none

      integer(long), intent(in)   :: dim
      integer(long)               :: j
      complex(dop), intent(inout) :: ovm(dim,dim)

      ovm(1,1) = (0.0d0,0.0d0)
      do j=2,dim
         ovm(j,1) = (0.0d0,0.0d0)
         ovm(1,j) = (0.0d0,0.0d0)
      enddo

      return
      end subroutine chovms2

      end module autocorr

