!***********************************************************************
      module calcdatmod

use decimal
use mmlib,  only: mmaxzzh, mmaxzz
use mtlib,  only: qtxxzz
use xvlib,  only: vvaxzz
use op1lib, only: sum2vxz, cpvxz, cpqxz, zeromxz, unitqxz, zerovxz
use op2lib, only: sum2mxxz, sum2qxxz

use global
use globmemmod
use dvrdatmod
use griddatmod
use psidef
use runpropmod
use operdef
use hpsimod
use gwplib, only: ovlgspf

use xlocphihphimod, only: xlocphihunphi
use multilayer
use eqofmotion1, only: ecorrml, euncorrml
use psiutil, only: spfovl
use gwptrj
use gwptrjmod
use datenmod, only: lgentrj
use mlop_eom, only: compute_all_mlop_etot
use potevalmod
use gh_elementsmod, only: kefac
use evaluate, only: calcenergy_gs,calcenergy

      implicit none
      private
      public :: calcdat

! Not used
      public :: classen
      contains

!***********************************************************************
!
!                  MCTDH module CALCDAT
!
! calculates data to be written to various files.
!
! passed back to calling routine:
!   spop(s): population (sum of squares of A-coefficients) for each 
!            state.
!   etot(s): Total energy for each state
!   ecorr(s): Correlated energy for each state
!   euncorr(s): Uncorrelated energy for each state
!
! V6.4  GW
!
!***********************************************************************

      subroutine calcdat(psi,dtpsi,dicht2,hunpsi,spop,ecorr,etot,euncorr, &
                 spoptr,time)

      implicit none

      integer(long) :: m,s,f,e,nham,zeig,zeig1,zeig2
      real(dop), intent(in)                     :: time
      real(dop), dimension(nstate), intent(out) :: spop,ecorr,euncorr,etot
      real(dop), dimension(ngwptrjst), intent(out)       :: spoptr
      real(dop)                                          :: energy
      complex(dop)                                       :: norm2
      complex(dop), dimension(dgldim), intent(in)        :: psi,dtpsi
      complex(dop), dimension(2*maxblock)                :: tmppsi
      complex(dop), dimension(maxspf*maxspf)             :: workc
      complex(dop)                                       :: sum, prefactor,cetot
      complex(dop), dimension(adim+1:dgldim), intent(in) :: hunpsi
      complex(dop), dimension(dmatdim), intent(in)       :: dicht2
      complex(dop), dimension(dmatdim)                   :: overlap

!-----------------------------------------------------------------------
! Determine prefactor of dtpsi vector
! for GWP prefactor is 1.0 as energy is determined directly 
!-----------------------------------------------------------------------
      if (lrlx) then
         prefactor = (-1.0_dop,0.0_dop)
         if(inttypea.eq.6) prefactor=(1.0_dop,0.0_dop)
      else
         prefactor = (0.0_dop,-1.0_dop)
      endif

!-----------------------------------------------------------------------
! calculate SPF overlap matrices
!-----------------------------------------------------------------------
      if (.not. lexact) then
         if (.not. lBEL) call spfovl(overlap,psi,psi,agmat,agmat,0)
! iMCG overlaps taken as unit. Need to do this after spfovl as overlaps
! for  meb need to be calculated.
         if (lgwp .and. (lgentrj .or. gwpmodel .eq. 2)) &
               call unitqxz(overlap,gwpdim(1,1))
      endif

!-----------------------------------------------------------------------
! calculate state populations:
!    for exact calculations sum grid populations
!    for MCTDH calculate A^conjg * S * A :
!    for trajectory GWP calculations include electronic states
!-----------------------------------------------------------------------
      if (lrunml) then
         call normml(psi,norm2,overlap)
         spop(1)=dble(norm2)

      else if (lexact) then
         do s=1,nstate
            call sum2vxz(psi(zetf(1,s)),sum,subdim(1))
            spop(s)=dble(sum)
         enddo

      else if (lgwptrj) then
! gwptrj is single set
         spop(1)=0.0_dop
         do s=1,ngwptrjst
            do e=1,gwpdim(1,1)
               tmppsi(e) = psi(e)*gwptrjci(s,e)
            enddo

            zeig=1
            zeig1=gwpdim(1,1)+1
            call qtxxzz(overlap(dmat(1,1)),tmppsi(zeig),tmppsi(zeig1), &
                 vdim(1,1),dim(1,1),ndim(1,1))

            call vvaxzz(tmppsi(zeig),tmppsi(zeig1),sum,gwpdim(1,1))
            spoptr(s)=dble(sum)

            spop(1)=spop(1)+dble(sum)
         enddo
      
      else if (lBEL) then
         do s=1,nstate
            zeig=1
            zeig1=block(s)+1
            call cpvxz(psi(zpsi(s)),tmppsi(zeig),block(s))
            do m=1,nmode
              if (gwpm(m)) then
                call cpqxz(gs2(gs2mat(m,s)),overlap(dmat(m,s)),&
                           gwpdim(m,s))
             else
                call mmaxzz(psi(zetf(m,s)),psi(zetf(m,s)),&
                            overlap(dmat(m,s)),subdim(m),dim(m,s),dim(m,s))
             endif
               call qtxxzz(overlap(dmat(m,s)),tmppsi(zeig),tmppsi(zeig1), &
                    vdim(m,s),dim(m,s),ndim(m,s))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            enddo
            call vvaxzz(psi(zpsi(s)),tmppsi(zeig),sum,block(s))
            spop(s)=dble(sum)
         enddo

      else
         do s=1,nstate
            zeig=1
            zeig1=block(s)+1
            call cpvxz(psi(zpsi(s)),tmppsi(zeig),block(s))

            do m=1,nmode
               call qtxxzz(overlap(dmat(m,s)),tmppsi(zeig),tmppsi(zeig1), &
                    vdim(m,s),dim(m,s),ndim(m,s))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            enddo

            call vvaxzz(psi(zpsi(s)),tmppsi(zeig),sum,block(s))
            spop(s)=dble(sum)
         enddo

      endif

!-----------------------------------------------------------------------
! calculate uncorrelated energy expectation value i.e. sum of
! uncorrelated matrix elements multiplied by the appropriate density
! matrix element. Sum of h|phi> terms are stored in hunpsi. 
!-----------------------------------------------------------------------
      if (lexact) then
         do s=1,nstate
            call vvaxzz(psi(zetf(1,s)),dtpsi(zetf(1,s)),sum, &
                 subdim(1))

            if (lrlx) then
               etot(s)=dble((sum-rlxenergy*spop(s))/prefactor)
            else
               etot(s)=dble(sum/prefactor)
            endif
            euncorr(s)=etot(s)
         enddo

      else if (.not.lrunml) then
         do s=1,nstate
            euncorr(s)=0.0_dop
            do m=1,nmode
               f=spfdof(1,m)
               if (gwpm(m)) then
                  if (khzahl(m,s,1) .gt. 0) then
                     call cpqxz(hteil(pmat(m,kzahl(1)+s)),workc,dim(m,s))
                  else
                     call zeromxz(workc,dim(m,s),dim(m,s))
                  endif
             !     if (lgwptrj) call addelen(workc,dim(m,s))
                  
               else if (basis(f) .eq. 27) then
                  nham=1
                  call xlocphihunphi(psi(zetf(m,1)),workc,m,f,3,dim(m,1),  &
                       nham)
                  
               else
                  call mmaxzzh(psi(zetf(m,s)),hunpsi(zetf(m,s)),workc, &
                       subdim(m),dim(m,s))
                
               endif
               call sum2qxxz(dicht2(dmat(m,s)),workc,sum,dim(m,s))
               euncorr(s)=euncorr(s)+dble(sum)
            enddo
         enddo

!-----------------------------------------------------------------------
! calculate energy expectation value from the scalar product
! of |psi> with H|psi> (stored in dtpsi coefficients). If run is
! with "proj-h" option, this is the total energy, otherwise this is the
! correlated energy.
!
! for relaxation, dtpsi stores H-E|psi>
!-----------------------------------------------------------------------
         do s=1,nstate
            if (abs(psitype) .eq. 4) then
               call calcenergy_gs(psi(zpsi(s)),dtpsi(zpsi(s)),energy,&
                    s,block(s),1)
              
            else
               call vvaxzz(psi(zpsi(s)),dtpsi(zpsi(s)),sum,block(s))
               energy = dble(sum/prefactor)
            endif
            
            if (lfunkr) then
               if (lrlx) then
                  etot(s)=energy-dble(rlxenergy*spop(s)/prefactor)
               else
                  etot(s)=energy
               endif
            else
               if (lrlx) then
                  ecorr(s)=energy-dble(rlxenergy*spop(s)/prefactor)
               else
                  ecorr(s)=energy
               endif
            endif
         enddo

      else
         call ecorrml(time,tree,psi,cetot)
         ecorr(1)=dble(cetot)

         call euncorrml(tree,psi,cetot)
         euncorr(1)=dble(cetot)
      end if

!-----------------------------------------------------------------------
! sum up total energy, or calculate ecorr, as required
!-----------------------------------------------------------------------
      do s=1,nstate
         if (lrunml) then
            etot(1)=euncorr(1) + ecorr(1)
         else
            if (lfunkr) then
               ecorr(s) = etot(s)-euncorr(s)
            else
               etot(s) = ecorr(s)+euncorr(s)
            endif
         endif
      enddo

!      if (lvmcg) then
!         call classen(eclass,peclass,psi,mr(mrhops))
!      endif

      return
      end subroutine calcdat

!#######################################################################

      subroutine classen(eclass,peclass,psi,hops)

      implicit none

      integer(long)                                     :: m,s,nham,f,i,n,&
                                                           v,n1,h
      integer(long)                                     :: zeig1,zeig2,zeig3
      integer(long)                                     :: lhaorder0,ierr,&
                                                           izflag
      integer(long)                                     :: k,k1,k2
      integer(long), dimension(1)                       :: point
      integer(long), dimension(maxdim)                  :: ng
      integer(long)                                     :: icalcv
      real(dop), intent(out)                            :: eclass,peclass
      real(dop)                                         :: keclass
      real(dop)                                         :: ketmp,coeff
      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(maxsta*maxsta)               :: pesdia
      complex(dop), dimension(maxsta*maxsta)            :: pesdiaz
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(2*maxblock)               :: workc
      complex(dop), dimension(maxblock)                 :: psig,psigl
      complex(dop)                                      :: ztmp

      character(len=10) :: cerr
 
! Store initial value of lhaorder and set to only calculate energy
      lhaorder0 = lhaorder
      lhaorder = 0

      nham = 1
      point(1) = 1
      k1=zham(nham)+kzahl(nham)+nmulpot(nham)
      k2=zham(nham)+khtot(nham)-1

      cerr = 'gconfig'
      allocate(gconfig(ndof,maxgwp,nstate),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'pconfig'
      allocate(pconfig(ndof,maxgwp,nstate),stat=ierr)
      if (ierr > 0) go to 1000

      peclass = 0.0_dop
      keclass = 0.0_dop
      eclass = 0.0_dop
      do s=1,nstate
         zeig1 = block(s)+1
         zeig2 = 1
         zeig3 = 2*block(s)+1
         call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))

! transform A-coeffs to GWP basis
! and store psi for RHS
!        if (psitype .eq. 0 .or. psitype .eq. 5) then
!           do m = 1,nmode
!              if (.not. gwpm(m)) cycle

!              zeig3 = zeig1
!              zeig1 = zeig2
!              zeig2 = zeig3
!              call qtxxzz(psi(zdgwp(m,s)),workc(zeig1), &
!                   workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
!           enddo

!           psig(1:block(s)) = workc(zeig2:zeig2-1+block(s))

!        else
!           psig(1:block(s)) = psi(zpsi(s):zpsi(s)-1+block(s))
!        endif
         psig(1:block(s)) = 1.0_dop

! multiply by GWP overlaps to form psi for LHS
     !    do m = 1,nmode
     !       if (.not. gwpm(m)) cycle

     !       zeig3 = zeig1
     !       zeig1 = zeig2
     !       zeig2 = zeig3
     !       call qtxxzz(gs2(gsmat(m,s)),workc(zeig1), &
     !            workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
     !    enddo
     !    psigl(1:block(s)) = workc(zeig2:zeig2-1+block(s))
          psigl(1:block(s)) = psig(1:block(s))

         ! generate configurations
         gconfig=0.0_dop

         do m=1,nmode
            if(m.eq.meb)cycle
            call gengconfig1m(m,gconfig(1,1,s),ndof,nspfdof(m),vdimgp(m,s),&
                 dimgp(m,s),ndimgp(m,s),psi(zxi(m,s)))
         enddo
         do m=1,nmode
            if(m.eq.meb)cycle
            call genpconfig1m(m,pconfig(1,1,s),ndof,nspfdof(m),vdimgp(m,s),&
                 dimgp(m,s),ndimgp(m,s),psi(zxi(m,s)))
         enddo

         ! add coordinates of point DVRs
         if (lpntdvr) then
            do f=1,ndof
               if (basis(f) .eq. 19) gconfig(f,:,s) = rpbaspar(1,f)
            enddo
         endif

         ! loop over all coordinate configurations and calculate PES value
         ! nsgp(s) = block(s) for ms, no. GWP configs for ss.
         if (leb) then
            i = 0
            do n=1,ndim(meb,1)
               do v=1,vdim(meb,1)
                  i=i+1
! get GWP number
                  num_gp = zgp(s) + i -1
                  call calcpes(hops,pesdia,point,gconfig(1,i,s),pesdiaz,izflag,nham)

                  call pescoeffg(pesdia,psig,workc,vdim(meb,s),&
                       dim(meb,s),ndim(meb,s),v,n)
               enddo
            enddo
         else if (nstate .eq. 1) then
            zeig3 = zeig1
            zeig1 = zeig2
            zeig2 = zeig3
            do i=1,nsgp(s)
! get GWP number
               num_gp = zgp(s) + i -1
               call calcpes(hops,pesdia,point,gconfig(1,i,s),pesdiaz,izflag,nham)

               workc(zeig1-1+i) = pesdia(1)*psig(i)

               ketmp = 0.0_dop
               do m=1,nmode
               do k=k1,k2  ! loop over uncorrelated operators
! find DOF with operator information. For KE or flux term this is
! in n rather than 1. For muld operators this is in fterm(k).
                  n=1
                  if (fterm(k) .gt. 0) then
                     f = fterm(k)
                  else
                     do n1=1,nspfdof(m)
                        f=spfdof(n1,m)
                        if (hamilton(f,k) .ne. f) n=n1
                     enddo
                     f = spfdof(n,m)
                  endif
                  h = hamilton(f,k)

!-----------------------------------------------------------------------
! get factor associated with KE operator (e.g. -1/2M)
! ng(maxdim) is an array whose entries are the derivative orders of the
! individual dof's as entries
!-----------------------------------------------------------------------
                  if (htmtype(h) .eq. 102) then
                     icalcv=1
                     call kefac(k,m,coeff,ng,icalcv)
                     coeff = -coeff
 
                     ketmp = ketmp+coeff*pconfig(f,i,s)*pconfig(f,i,s)
                  endif
               enddo
               enddo

               workc(zeig2-1+i) = ketmp*psig(i)

            enddo

         else
  print*,'HELP'
         endif

         call vvaxzz(psigl,workc(zeig1),ztmp,block(s))
         peclass = peclass+dble(ztmp)

         call vvaxzz(psigl,workc(zeig2),ztmp,block(s))
         keclass = keclass+dble(ztmp)
      enddo

      eclass = keclass + peclass

      print*,'ECLASS',peclass*ev,keclass*ev,eclass*ev

      deallocate(gconfig,pconfig)


! Restore flags
      lhaorder = lhaorder0

      return
 1000 write(ilog,'(a)') 'Error allocating memory in CLASSEN :',cerr
      write(ilog,'(a)') 'Error Code :',ierr
      stop
      end subroutine classen

!#######################################################################

      subroutine pescoeffg(pes,psi1,psi2,vdim1,dim1,ndim1,v1,n1)

      implicit none

      integer(long), intent(in)   :: vdim1,dim1,ndim1,v1,n1
      integer(long)               :: s,s1
      real(dop), dimension(dim1,dim1), intent(in)            :: pes
      complex(dop), dimension(vdim1,dim1,ndim1), intent(in)  :: psi1
      complex(dop), dimension(vdim1,dim1,ndim1), intent(out) :: psi2

      do s=1,dim1
         psi2(v1,s,n1) = 0.0_dop
         do s1=1,dim1
            psi2(v1,s,n1) = psi2(v1,s,n1) + pes(s,s1)*psi1(v1,s1,n1)
         enddo
      enddo

      return
      end subroutine pescoeffg

!#######################################################################
 
      end module calcdatmod
