!***********************************************************************
!
! CROCOFLY.F
!
! Module to calculate the cross-correlation function on the fly.
!
! Contains the following routines:
!   WKCROCO
!   ZEIGCROCO
!   RDCCREF1
!   RDCCREF2
!   CCOVERLAP
!   CCOUT
!
!***********************************************************************

      module crocofly

use decimal, only: long, dop
use globmemmod
use logdat
use psiutil, only: spfovl
use multilayer, only: setoverlap,mloverlap
use wftree, only: compare_tapes

use mmlib,   only: mmaxzz
use mtlib,   only: mtxxzz
use xvlib,   only: vvaxzz
use op1lib,  only: cpvxz
use timing

      implicit none
      private
      public :: zeigcroco,rdccref1,rdccref2,ccout
      contains

!***********************************************************************
!
! subroutine ZEIGCROCO
!
! Calculates the size of the (static) memory needed in the calculation
! of the cross-correlation function.
!
! Output variables:
!   (to common blocks)
!   psirefdim : size of reference WF
!   ccjindxdim: size of JINDX array of ref. WF
!   ccwcdim: workc-space needed for this module
!   doverdim,dovmat: needed later in CCOVERLAP
!
! Dec 2001, FO
!
!***********************************************************************

      subroutine zeigcroco

use global
use griddatmod
use psidef
use crocoflymod
use compdatmod

      implicit none

      integer(long) :: zeig,zeig1,m,s

!     obtain system definition of reference WF from "restart" file
      call rdccref1

!     calculate size of reference WF
      psirefdim = cdgldim(2)

!     memory needed for JINDX arrray
      if (ccitype(2) .gt. 0) then
         ccjindxdim = cadim(2)
      else
         ccjindxdim = 1
      endif

!     memory needed for the overlap calculation
!     [mainly cut'n'pasted from ANALYSE/CROSSCORR.F]
      ccwcdim=ccmaxblock
      zeig=1
      do s=1,nstate
         zeig1=block(s)
         do m=1,nmode
            dovmat(m,s)=zeig
            zeig=zeig+cdim(m,s,2)*dim(m,s)
            zeig1=zeig1*cdim(m,s,2)/dim(m,s)
            ccwcdim=max(ccwcdim,zeig1)
         enddo
      enddo
      doverdim=zeig-1

      return
      end subroutine zeigcroco

!***********************************************************************
!
! subroutine RDCCREF1
!
! Opens the file containing the reference wave function.
! Checks if this WF's system definition matches the current system.
! System def. of reference WF is stored into comparison data set 2.
!
! Output variables:
!   (to common blocks)
!   ovtype,ebtype (see "include/crocofly.inc")
!   + everything declared in "include/compdat.inc"
!
! Dec 2001, FO
!
!***********************************************************************

      subroutine rdccref1

use global
use griddatmod
use psidef
use crocoflymod
use compdatmod
use channels

use compdat, only: savecsys,savecpsi,getcsys,getcpsi
use iorst, only: rstinfo
use iopsidef, only: psidat

      implicit none

      logical(kind=4)          :: linwf,lerr,readsys,ltape,ltree
      integer(long)            :: chkdvr,chkgrd,chkpsi,chkprp,&
                                  ilbl,s,t,i,j,lentp,lentp2
      character(len=240)       :: filename

!     save system definition
      call savecsys(1)
      call savecpsi(1)

!     tell the LOG file what's going to happen
      write(ilog,'(/a)') &
           '***** CROSS: reading system definition for '// &
           'reference wavefunction *****'

      if(test0 .and. crossname(1:crosslng).eq.name(1:laenge)) goto 100
!     build filename of restart file
      filename = crossname(1:crosslng)//'/restart'
!     determine length of filename
      ilbl=index(filename,' ')-1
!     open the restart file (contains reference WF)
      open(irst, file=filename, form='unformatted', &
           status='old', err=666)

!     check system definition
!     DVR and GRID definitions should match current system (CHK=2)
!     PSI definition can differ and will be read (CHK=1)
!     PRP definition is read but ignored (CHK=0)
      chkdvr=2
      chkgrd=2
      chkpsi=1
      chkprp=0
      call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)

!     check if some checks failed
      readsys = .false.
      if (lerr) then
         routine = 'RDCCREF1(cross)'
!        investigate which checks failed
         if (chkerr(1) .ne. 0) then
!           I/O error: fatal
            message = 'Error reading '//filename(1:ilbl)
            call errormsg
         endif
         if (chkerr(2) .ne. 0) then
!           Definitions of DOFs differ: fatal
            message = 'DOF definitions from '//&
                 filename(1:ilbl)//' don''t match system.'
            call errormsg
         endif
         if (chkerr(3) .ne. 0) then
!           Definitions of DVR basis differ: fatal
            message = 'DVR basis definitions from '//&
                 filename(1:ilbl)//' don''t match system.'
            call errormsg
         endif
         if (chkerr(4) .ne. 0) then
!           Mode combination definitions differ: non-fatal
            ovtype = 1
            readsys = .true.
            message = 'Definiton of mode combinations in '//&
                 filename(1:ilbl)//' differs from system.'
!           not implemented yet, so quit instead of only warn
            call errormsg
         endif
         if (chkerr(5) .ne. 0) then
!           unclear; see RDGRDDEF for details. non-fatal
            ebtype = 1
            readsys = .true.
            message = 'chkerr(5) != 0 -- aborting'
!           not implemented yet, so quit instead of only warn
            call errormsg
         endif
!        chkerr(6:9) were not affected by the RSTINFO call
         if (chkerr(10) .ne. 0) then
!           Density types differ: fatal
            message = 'Wavefunction from '//filename(1:ilbl)//&
                 ' uses a density type different from system.'
            call errormsg
         endif
      endif

!     check in case of a ML-run if tapes are identical,
!     then call setoverlap, which creates the trees
      if(lrunml) then
         call compare_tapes(tape,ctape(1,1),ltape,ltree,lentp,lentp2)
         if(ltree) then
            do j=1,int(max(lentp,lentp2)),20
               write(ilog,'(/a)') '    j/cross-tape(j)/system-tape(j)'
               write(ilog,'(20i4)') (      i   ,i=j,min(lentp ,19+j))
               write(ilog,'(20i4)') ( tape(i  ),i=j,min(lentp ,19+j))
               write(ilog,'(20i4)') (ctape(i,1),i=j,min(lentp2,19+j))
            enddo
            routine='RDCCREF1 on crocofly.F'
            message='The topologies of the trees cross/system differ.'
            call errormsg
         elseif(ltape) then
            write(ilog,'(2a,2i9)') 'The tapes cross/system are not ', &
                       'identical. dgldim:',dgldim,cdgldim(1)
            write(ilog,'(2a,i5/)') 'But the topologies of the trees ', &
                       'cross/system are identical. lentape =',lentp
         else
            write(ilog,'(2a,i5/)') 'The tapes cross/system are ', &
                                   'identical. lentape =',lentp
         endif
         call setoverlap(tape,ltape,dgldim)
      endif


!     since ovtype=1 and ebtype=1 are not implemented yet,
!     the following part is commented out

! $$$*     if needed, repeat the RSTINFO call, but this time also read
! $$$*     the GRID definition for later use
! $$$*     [ cut'n'pasted from analyse/crosscorr.F ]
! $$$      if (readsys) then
! $$$         rewind(irst)
! $$$         chkdvr=2
! $$$         chkgrd=1
! $$$         chkpsi=1
! $$$         chkprp=0
! $$$         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
! $$$         if (nmode .eq. 1) then
! $$$            do s=1,nstate
! $$$               dim(1,s)=1
! $$$               do t=1,nstate
! $$$                  ddim(1,s,t)=1
! $$$               enddo
! $$$            enddo
! $$$         endif
! $$$      endif

!     we have all info we need, now close restart file
      close(irst)

!     store the system definition of the reference WF
 100  call savecsys(2)
      call savecpsi(2)

!     restore original system definition
      call getcsys(1)
      call getcpsi(1)
      call psidat

!     calculate biggest block
      ccmaxblock = 1
      if (dentype .eq. 1) then
         do s = 1,nstate
            do t = 1, nstate
               ccmaxblock = max(ccmaxblock,dblock(s,t),cdblock(s,t,2))
            enddo
         enddo
      else
         do s = 1,nstate
            ccmaxblock = max(ccmaxblock,block(s),cblock(s,2))
         enddo
      endif

      return

!     things to do in case of error: bailing out
 666  continue
      routine = 'RDCCREF1(cross)'
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg

      end subroutine rdccref1


!***********************************************************************
!
! subroutine RDCCREF2
!
! Module to calculate the cross-correlation function on the fly.
!
! Called from RUNPROP.
! Reads the reference wavefunction from a "restart" file into memory.
!
! Input variables:
!   mc,mr,mi,ml : the well-known memory arrays
! Output variables:
!   the reference WF is stored in mc(mcpsiref)
!
! Dec 2001, FO
!
!***********************************************************************

      subroutine rdccref2

use global
use channels

use iopsidef, only: psidat
use iorst, only: rdrst
use compdat

      implicit none

      complex(dop), dimension(1)                  :: cdum,cdum1
      integer(long)                               :: chkdvr,chkgrd,chkpsi,&
                                                     chkprp,iflag, idum, ilbl
      integer(long), dimension(1)                 :: idum1
      real(dop), dimension(1)                     :: rdum
      logical(kind=4)                             :: ldum
      character(len=240)                          :: filename

!     save current system definition
      call savecsys(1)
      call savecpsi(1)

!     get system definiton of reference WF
      call getcsys(2)
      call getcpsi(2)
      call psidat

!     tell the LOG file what's going to happen
      write(ilog,'(/a)')&
           '***** CROSS: reading reference wavefunction '//&
           'into memory *****'

!     open "restart" file
      filename = crossname(1:crosslng)//'/restart'
      write(ilog,'(a)') filename(1:crosslng+8)
      open(irst, file=filename, form='unformatted',&
           status='old', err=667)

!     Read reference WF:
!     Since the system definiton was already read (during RDCCREF1),
!     we can ignore it this time (CHK=0). Neither are we interested
!     in the lanczos vectors (IFLAG=0). Most parameters returned from
!     RDRST are discarded, therefore the ?DUM variables. The last
!     parameter for RDRST is for the JINDX array.
      chkdvr=0
      chkgrd=0
      chkpsi=0
      chkprp=0
      iflag=0
      call rdrst(rdum(1), mc(mcpsiref), ldum, chkdvr, chkgrd, chkpsi,&
           chkprp,cdum,idum,rdum,iflag,mi(miccjindx),cdum1,idum1)

!     close "restart" file
      close(irst)

!     restore original system definiton
      call getcsys(1)
      call getcpsi(1)
      call psidat

      return

!     things to do in case of error: bailing out
 667  continue
      routine = 'RDCCREF2'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg

      end subroutine rdccref2


!***********************************************************************
!
! subroutine CCOVERLAP
!
! Calculates the overlap between two wavefucntions in MCTDH form.
! [mainly cut'n'pasted from ANALYSE/CROSSCORR.F]
!
! In:
!   psiref : reference WF in MCTDH form
!   psi    : system WF
!   dover  : memory needed for overlap matrices [=mc(mcccdover)]
!   workc  : complex scratch array
! Out:
!   ovl = <psiref|psi>
!
! Jan 2002, FO
!
!***********************************************************************

      subroutine ccoverlap(ovl,psiref,psi,dover)

use global
use griddatmod
use psidef
use crocoflymod
use compdatmod
use runpropmod

      implicit none

      integer(long)                                  :: s,m,zeig1,zeig2,swapzeig
      integer(long)                                  :: tmpdim
      complex(dop)                                   :: zdum
      complex(dop), intent(out)                      :: ovl
      complex(dop), dimension(*), intent(in)         :: psiref,psi
      complex(dop), dimension(doverdim), intent(out) :: dover
      complex(dop), dimension(:), allocatable        :: tempc 
      integer(long), save                            :: tid=0 ! timer ID

      call get_timer(tid, 'Cross')
      call start_timer(tid)

!-----------------------------------------------------------------------
! work array
!-----------------------------------------------------------------------
      tmpdim = max(2*ccwcdim, 2*ccmaxblock)
!.....Improved relaxation, DAV, lock.
      if(mod(rlaxnum,10000).eq.930) tmpdim = max(tmpdim,2*adim)
      allocate(tempc(tmpdim))

!-----------------------------------------------------------------------
!   calculate density matrices for each degree of freedom defined as 
!   overlap of the single-particle functions rho^m_ij=<phi_i,ref|phi_j>
!-----------------------------------------------------------------------
      do s=nst1,nst2
         do m=1,nmode
            call mmaxzz(psiref(czetf(m,s,2)),psi(zetf(m,s)), &
                 dover(dovmat(m,s)),subdim(m),cdim(m,s,2),dim(m,s))
         enddo
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,ref|Phi_L>A_L within each state
!-----------------------------------------------------------------------
      ovl=(0.0_dop,0.0_dop)
      do s=nst1,nst2
         zeig1=1
         zeig2=ccwcdim+1
         call cpvxz(psi(zpsi(s)),tempc(zeig2),block(s))

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(dover(dovmat(m,s)),tempc(zeig1), &
                    tempc(zeig2),cvdim(m,s,2),dim(m,s),ndim(m,s), &
                    cdim(m,s,2))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J^*,ref [ Sum_L <Phi_J,ref|Phi_L>A_L ] within each state
!-----------------------------------------------------------------------
         call vvaxzz(psiref(czpsi(s,2)),tempc(zeig2),zdum,cblock(s,2))
         ovl=ovl+zdum
      enddo

      if (allocated(tempc)) deallocate(tempc)

      call stop_timer(tid)

      return
      end subroutine ccoverlap

!***********************************************************************
!
! subroutine CCOUT
!
! Writes the cross-correlation function to the ICROCO file.
!
! Jan 2002, FO
!
!***********************************************************************

      subroutine ccout(time,psi)

use global
use griddatmod
use psidef
use crocoflymod
use channels

      implicit none

      real(dop), intent(in)                     :: time
      complex(dop)                              :: ovl
      complex(dop), dimension(*), intent(in)    :: psi
      complex(dop), dimension(:), allocatable   :: ovlmat

     if(lrunml) then
         allocate(ovlmat(dmatdim))
         call spfovl(ovlmat,psi,psi,agmat,agmat,0)
         call mloverlap(mc(mcpsiref),psi,ovl,ovlmat)
         deallocate(ovlmat)
      else
!        calculate cross-correlation function
         call ccoverlap(ovl, mc(mcpsiref), psi, mc(mcccdover))
      endif

!     write result to ICROCO file
      write(icroco,'(1x,g18.9,3(1x,f18.14))')&
            time, dble(ovl), dimag(ovl), abs(ovl)
      call myflush(icroco)

      end subroutine ccout


      end module crocofly

