!#######################################################################
!
!                      MCTDH module DATEN
!
! various small routines used to output data.
!
! subroutines :
!    diagchk  : calculates the largest off diagonal matrix element
!    activcap : Automatic CAP, writes to log file when switched on.
!    wrchck   : writes check file
!    speedout : writes time spent between two outputs to file "speed".
!    writestep: Writes the stepsize and error of the current integration
!               step to a file.
!    WriteStepDav : wites steps for DAV
!    WriteStepCDav: wites steps for CDAV
!    WriteStepBDav: wites steps for BDAV
!    realdaverror : writes error messages to log file if A-vector not real.
!
!  V6.4 GW
!
!#######################################################################

      module daten

use decimal, only: long, dop
use datenmod
use global
use dglobal
use dvrdatmod
use griddatmod
use psidef
use channels
use runpropmod
use iopsidef, only: tapelength

      implicit none
      private
      public :: diagchk,activcap,speedout,wrchck,timespend
      contains

!#######################################################################
! DIAGCHK
!
! called from datout
!
! for natural orbital propagations, at each output step it is checked
!  that the density matrix dicht2 is diagonal. This subroutine
!  calculates the largest absolute value of the off diagonal elements
!  which is tested in datout against an error criterion
!
! for a single mode:
!  dicht2: density matrix
!  dim: number of single particle functions
!  diagerr: largest absolute off-diagonal element
!
! V4.0 
! not in use!!!
!#######################################################################

      subroutine diagchk(dicht2,dim,diagerr)

      implicit none

      integer(long)                                :: e,e1
      integer(long), intent(in)                    :: dim
      complex(dop), dimension(dim,dim), intent(in) :: dicht2
      real(dop), intent(out)                       :: diagerr

      diagerr=0.0
      do e=1,dim
        do e1=1,dim
          if (e1 .ne. e) then
            diagerr=max(abs(dicht2(e1,e)),diagerr)
          endif
        enddo
      enddo

      return
      end subroutine diagchk

          
!#######################################################################
! ACTIVCAP
!
!  Called by dichtout.
!  Test on the grid population of the first grid point of the CAP
!  to decide whether or not the CAP has to be activated. The CAP is 
!  activated when the grid-population at the beginning of the CAP increases.
!  Write the step number to the log file so that the analysis
!  of the reactive flux can be done properly.
!
!  mai 98  MCH
!#######################################################################

      subroutine activcap(gpop1,autocap,gpopcap,timecap,time, &
                          ilog,modelabel,countpsi,f,n)

      implicit none

      integer(long), intent(in)     :: ilog,countpsi,f,n
      integer(long), intent(inout)  :: autocap
      real(dop), intent(in)         :: gpop1,timecap,time
      real(dop), intent(inout)      :: gpopcap
      character(len=16), intent(in) :: modelabel

      if (gpop1.gt.gpopcap .and. time.gt.timecap) then
         autocap=-autocap
         write(ilog,'(/,3a,/a)')'******* CAP is activated for mode ', &
                         modelabel(1:9),' *************************', &
                    ' the first psi with CAP is stored at position :'
         write(ilog,'(a,i6)') ' step-number ', countpsi+1
         write(ilog,'(a,1p2e11.3)') ' gridpop:', gpop1, gpopcap
         write(ilog,'(a,i4,a,i4)')  ' DOF =', f, ' ,  n =', n
         write(ilog,'(a,f9.2,a,f9.2)') ' timecap =',timecap, &
                                       ',  time =',time
         write(ilog,'(a,/)')  &
         '***********************************************************'
      else
          gpopcap=gpop1 + 1.0e-11_dop
      endif

      return
      end subroutine activcap

!#######################################################################
! WRCHCK
!
! writes data to check file
!
! V6.0 GW
!#######################################################################

      subroutine wrchck(spop,sspop,ecorr,etot,dicht3,qdq,norm,ggp,&
                 trace)

      implicit none

      integer(long)                                   :: s,m,i,f,t
      real(dop), dimension(nstate), intent(in)        :: spop,ecorr,etot
      real(dop), dimension(nddstate), intent(in)      :: sspop
      real(dop), dimension(d3matdim), intent(in)      :: dicht3
      real(dop), dimension(4,ndof,nstate), intent(in) :: qdq
      real(dop), dimension(gwpinfodim), intent(in)    :: ggp
      real(dop), intent(in)                           :: norm,trace
      character(len=c2)                               :: wrtag

!-----------------------------------------------------------------------
! Write Data Set
!-----------------------------------------------------------------------
      countchk=countchk+1

      write(wrtag,'(a9,i10,a1)') '<data set',countchk,'>'
      write(ichk) wrtag

!-----------------------------------------------------------------------
! If dynamical WF: write extra info about the wavefunction
!-----------------------------------------------------------------------
      if (ldynspf.or.lBEL) then
         wrtag= '<dimtape>'
         write(ichk) wrtag
         write (ichk) ((dim(m,s),m=1,nmode),s=1,nstate)
         if (lrunml) then
            write(ichk) tapelength()
            write(ichk) (tape(m),m=1,tapelength())
         endif
      endif

! for single-set calculations state populations are on diagonal of density
! matrix for mode meb. For multi-set they are stored in spop array
      if (leb) then
         wrtag='<sspop>'
         write(ichk) wrtag
         write(ichk) (sspop(i),i=1,nddstate)
      else
         wrtag='<spop>'
         write(ichk) wrtag
         write(ichk) (spop(s),s=1,nstate)
      endif
      if (dentype .eq. 1 .or. dentype .eq. 2) then
         wrtag='<decorr>'
         write(ichk) wrtag
         write(ichk) (ecorr(s),s=1,nstate)
         wrtag='<detot>'
         write(ichk) wrtag
         write(ichk) (etot(s),s=1,nstate)
      else
         wrtag='<ecorr>'
         write(ichk) wrtag
         write(ichk) (ecorr(s),s=1,nstate)
         wrtag='<etot>'
         write(ichk) wrtag
         write(ichk) (etot(s),s=1,nstate)
      endif

      if (dentype .eq. 1) then
         wrtag='<ddicht3>'
         write(ichk) wrtag
         write(ichk) ((((dicht3(dd3mat(m,s,t)+i),i=0,ddim(m,s,t)-1),m=1,nmode), &
               s=1,nstate),t=1,nstate)
      else
         wrtag='<dicht3>'
         write(ichk) wrtag
         write(ichk) (((dicht3(d3mat(m,s)+i),i=0,dim(m,s)-1),m=1,nmode),s=1,nstate)
      endif

      wrtag='<qdq>'
      write(ichk) wrtag
      do s=1,nstate
         do f=1,ndof
            if (dofspf(f) .eq. 0) cycle
            write(ichk) (qdq(i,f,s),i=1,4)
         enddo
      enddo

      if (lgwp) then
         wrtag='<ggp>'
         write(ichk) wrtag
         do s=1,nstate
            do m=1,nmode
               if (.not. gwpm(m)) cycle
               if (m .eq. meb) cycle
               write(ichk) (ggp(zgwpinfo(m,s)+i),i=0,gwpdim(m,s)-1)
            enddo
         enddo
      endif

      wrtag='<norm>'
      write(ichk) wrtag
      write(ichk) norm

      if (dentype .gt. 0) then
         wrtag='<trace>'
         write(ichk) wrtag
         write(ichk) trace
      endif

      write(wrtag,'(a13,i10,a1)') '<end data set',countchk,'>'
      write(ichk) wrtag

      call myflush(ichk)

      return
      end subroutine wrchck

!***********************************************************************
!                                 SPEEDOUT
!
! Writes the CPU-time spent since last output to file "speed".
!
! V6.4 MB
!***********************************************************************

      subroutine speedout (time)

      implicit none

      real(dop)               :: now,spent
      real(dop), save         :: old
      real(dop), intent(in)   :: time
      integer(long)           :: i,tm
      integer(long), save     :: tm0
      character(len=80)       :: actual
      real(dop), external     :: fcputime,fsystime

      logical(kind=4), save   :: firstcall = .true.
      real(dop), save         :: previous  = 0.0_dop

!----------------------------------------------------------------------
! determine actual time and CPU-time spent since last call
!----------------------------------------------------------------------
      call fzeit2(tm,actual)
      i = index(actual,':')
      actual=actual(i-9:i+6)

      now=fcputime(0)+fsystime(0)
      spent=now-previous
      previous=now

!----------------------------------------------------------------------
! write data to speed-file
!----------------------------------------------------------------------
      if (firstcall) then
         firstcall=.false.
         if( time .le. -1.d40 ) then    ! continuation run
            backspace ispeed
            read(ispeed,'(13x,f14.1,34x,i9)',end=34,err=34) old, tm0
            goto 35
 34         old=0.d0
            tm0=0
 35         tm0=tm-tm0
            write(ispeed,'(2a)') '# -------------------------------', &
                     ' continuation -------------------------------'
            return
         else
            tm0=tm
            old=0.d0
         endif
         write(ispeed,'(2a)') '#    Start: ',trim(cstartup)
         if(fs .lt. 2.0) then
            write(ispeed,'(a,3x,a,5x,a,4x,a,8x,a)') &
                      '#    Time[au]','CPU-time[s]','Delta-CPU', &
                      'Real time','RealT[s]     [h]'
         else
            write(ispeed,'(a,3x,a,5x,a,4x,a,8x,a)') &
                      '#    Time[fs]','CPU-time[s]','Delta-CPU', &
                      'Real time','RealT[s]     [h]'
         endif
         write(ispeed,'(a,f12.2,f14.1,18x,a16,i9,f9.2)') &
                      '#', time/fs,now+old,actual, &
                      tm-tm0,dble(tm-tm0)/3600.d0
      else
         write(ispeed,'(f13.2,f14.1,f14.3,4x,a16,i9,f9.2)') &
                      time/fs,now+old,spent,actual, &
                      tm-tm0,dble(tm-tm0)/3600.d0
      end if

      call myflush(ispeed)

      return
      end subroutine speedout

!***********************************************************************
!                  TIMESPEND                          
!
! Writes the CPU-time and real time to channel chn.
!
! 01/07 HDM
!***********************************************************************

      subroutine timespend(chn,string)

      implicit none

      real(dop)                     :: now,spent
      integer(long)                 :: i
      integer(long), intent(in)     :: chn
      character(len=80)             :: actual
      character(len=80), intent(in) :: string
      real(dop), external           :: fcputime,fsystime
!
      logical(kind=4), save         :: firstcall = .true. 
      real(dop), save               :: previous  = 0.0_dop

!----------------------------------------------------------------------
!! determine actual time and CPU-time spent since last call
!----------------------------------------------------------------------
!      call fzeit(actual)
      i = index(actual,':')
      actual=actual(i-9:i+6)

      now=fcputime(0)+fsystime(0)
      spent=now-previous
      previous=now

!----------------------------------------------------------------------
! write data to file (chn)
!----------------------------------------------------------------------
      if (firstcall) then
         firstcall=.false.
         write(chn,'(a,8x,a,19x,a,4x,a,5x,a)') &
              '%','string','CPU-time[s]','Delta-CPU','Real time'
         write(chn,'(a2,a30,f13.1,18x,a16)') '% ',string, now, actual
      else
         write(chn,'(a2,a30,f13.1,f13.3,5x,a16)') &
                     '% ',string,now,spent,actual
      end if

      call myflush(chn)
!
      return
      end subroutine timespend

! **********************************************************************


      end module daten
