! **********************************************************************
! *                                                                    *
! *                          function (function.F90)                   *
! *                                                                    *
! * mctdh module that calls the routines that compute the right hand   *
! * side of the mctdh working equations 
!
!   when the mean fields are held 'constant':
!       funk0, funk1, funk2, funk3, funkphi, funka.
!   when the single-particle hamiltonian is rhs of the projector:
!       funkr
!   when the single-particle hamiltonian is lhs of the projector:
!       funk
!
! contains: funkphi2,funka2,funkex,qdot,qdot1m,xdot,xdot1m,deriv1x
! **********************************************************************
      module function

!use decimal,  only: long, dop
use decimal,  only: long, dop
use maxv, only: maxdim
use global
use dglobal
use globmemmod
use logdat
use xvlib,    only: xvxxzza, xvxxdzr, xvxxdda, xvxxddr, xvxxzzo, vvaxzz
use mmlib,    only: qqxxzz
use mtlib,    only: qtxxzzr, qtxxzd, qtxxzda, mtxxzd, dtxxdz
use op1lib,   only: zerovxz, zerovxd, cpvxd, cpvxz
use op2lib,   only: addvxxzo, addvxxdo, addvcxzo, addmxxzo, submxxzo1, &
                    subvcxzo
use d1tens,   only: makeadj
use timing
use dvrdatmod
use eqofmotion
use eqofmotion1
use griddatmod
use runpropmod
use operdef
use psidef
use channels
use datenmod
use svdregmod

use tdkoeffs
use hphimod
use hpsimod
use phihphimod
use mfieldsmod
use gh_elementsmod
use densitymod
use d2mfieldsmod, only: d2mfields
use d2summfmod, only: d2summf
use d2calchamod
use summfmod
use multilayer
use mlop_eom, only: compute_all_uterms_tree, compute_all_wterms_tree
use addhunphimod
use evaluate
use hlochphimod
use calchamod
use gaussian
use gwptrjmod
use gwplib
use projectmod
use dicht1phimod
use propmod_interface
use potevalmod, only: poteval

use d1hphimod
use d1phihphimod, only: d1phihphi,d1phihunphi
use d1mfieldsmod, only: d1mfields
use d1hlochphimod
use d1dicht1phimod
use d1addhunphimod
use d1calchamod
use d1summfmod
use d1tens, only: makeaadj
use d1evaluate, only: chkinib

#ifdef MPI
use mpi
use mpidata
use hlochphimod, only: hlochphi1m,d2hlochphi1m,hlochphi1mpar
use mpifunctionmod, only: mpifunka2
#endif
use openmpmod

      implicit none
      private
      public :: funk,funk0,funk1,funk2,funk3,funkr,funkphi,funkphi2,funka,&
                funka2,funkex,wkfunction,funkrreal,funkphinode,funkml_dt,&
                funkgbf,ddmfunk,ddmfunk0,ddmfunk1,ddmfunk2,ddmfunk3,ddmfunka,&
                ddmfunkphi,ddmfunkex,ddmfunkr,constener,funkphiml

      contains

! **********************************************************************
! *                                                                    *
! *                          subroutine funk0                          *
! *                                                                    *
! * calculates the mean-field and inverse density matrices and the     *
! * time derivative of the mctdh-coefficients without computing the    *
! * time derivative of the single-particle functions.                  *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, phihphi, hunphi, phihunphi                      *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   dtpsi: time derivative of the mctdh-coefficients                 *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.2 mb                                                            *
! *                                                                    *
! * if modus = 1, for ML-MCTDH the time-derivative of the A-vector     *
! * is NOT calculated (for ada2)                                       *
! **********************************************************************

      subroutine funk0 (time,psi,dtpsi)

      implicit none

      integer(long)                                  :: hktype,ktype,nham,modgh
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(:), allocatable        :: psi1
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi
      integer(long)                                  :: i

      integer(long), save                            :: tid = 0 ! timer ID
      logical(kind=4)                                :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funk0 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

!-----------------------------------------------------------------------
! allocate memory
!-----------------------------------------------------------------------
      if (lsvd) allocate(psi1(nmode*adim))

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      call get_timer(tid, 'Funk0')
      call start_timer(tid)

      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! compute gaussian matrix elements
!-----------------------------------------------------------------------
      if (lgwp) then
         if(lrlx)then
            call gwp_param_a(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         else
            call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         endif
         call gwp_overlap(gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,0)
! only zero order elements needed for mean-fields
         modgh = 4
         call gh_elements(gh,hung,ghderiv1,agmat,gfgwp,aigwp,zgwp,&
              mr(mrqcent),1,modgh,gs2,time)
      endif

! --- compute density and inverse density matrices ---

      if (lsvd) then
         call svdreg(psi,psi1,dicht1,dicht2,dicht3,dicht4)
      else
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
      endif

! --- compute action of correlated hamiltonian terms on spf's ---
! --- determine correlated matrix elements <phi|h|phi> ---

      ktype=0
      hktype=0
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hphi(psi,mc(mchpsi),ktype,1)
         call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
      else if (dentype .eq. 1) then
         call d1hphi(psi,mc(mchpsi),ktype,1)
         call d1phihphi(psi,mc(mchpsi),hteil,1)
      endif
      if (lrunml) then
! Also need uncorrelated matrix elements for ML-CMF
         call hunphi(psi,mc(mchunpsi),hktype,1)    
         call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,nham)

         call compute_hterms_treer(time,tree,psi,1)
         call compute_all_uterms_tree(time,1,tree,psi)
         call compute_all_wterms_tree(time,1,tree,psi)
      endif


! --- determine mean-field matrices and time derivative of a-vector ---
! --- compute sum over mean-field times single-particle operator ---
      if (lrunml) then
         call compute_mfields_tree(tree,psi)
         call kmfields(mc(mchloch))
         if (dentype .eq. 1) then
            call d1summf(mfsum,mc(mchloch))
         else
            call compute_mfsum_tree(time,tree)
         endif
      else if (dentype .eq. 0) then
         if (lsvd) then
            call mfields(psi,dtpsi,1,psi1,0,usehsym)
         else
            call mfields(psi,dtpsi,1,psi,0,usehsym)
         endif
         call summf(mfsum,mc(mchloch),1)
      else if (dentype .eq. 1) then
         call d1mfields(psi,dtpsi)
         call d1summf(mfsum,mc(mchloch))
      else if (dentype .eq. 2) then
         call d2mfields(psi)
         call d2summf(mfsum,mc(mchloch))
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,dgldim)

      if (allocated(psi1)) deallocate(psi1)

      call stop_timer(tid)

      return
      end subroutine funk0   

! **********************************************************************
! *                                                                    *
! *                          subroutine funk1                          *
! *                                                                    *
! * computes the time derivative of the single-particle functions      *
! * without recomputing the density and mean-field matrices.           *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, hunphi, hlochphi, mfsumphi, dicht1phi,          *
! *              addhunphi, project, prefactor                         *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   dtpsi: time derivative of the wavefunction psi                   *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.2 mb                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funk1 (time,psi,dtpsi)

      implicit none

      integer(long)                                :: hktype, ktype,&
                                                      nham, i,modgh
      real(dop), intent(in)                        :: time
      complex(dop), dimension(dgldim), intent(inout):: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi

      integer(long), save                          :: tid = 0 ! timer ID
      logical(kind=4)                              :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK (only SPFs are used)
!-----------------------------------------------------------------------
      lcheck = .true.
      if (dentype .eq. 1) then
         do i=dzetf(1,1,1)+1,dgldim
            if (psi(i) .ne. psi(i)) lcheck = .false.
         enddo
      else
         do i=zetf(1,1)+1,dgldim
            if (psi(i) .ne. psi(i)) lcheck = .false.
         enddo
      endif
      if (.not. lcheck) then
         routine = 'Funk1 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

! --- initialise variables --- 

      call get_timer(tid, 'Funk1')
      call start_timer(tid)
      call zerovxz(dtpsi(adim+1),totphidim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- compute action of correlated hamiltonian terms on spf's ---
      ktype=1
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hphi(psi,mc(mchpsi),ktype,1)
      else if (dentype .eq. 1) then
         call d1hphi(psi,mc(mchpsi),ktype,1)
      endif

!    --- compute action of uncorrelated hamiltonian terms on spf's ---

      hktype=0
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hunphi(psi,mc(mchunpsi),hktype,1)
      else if (dentype .eq. 1) then
         call d1hunphi(psi,mc(mchunpsi),hktype,1)
      endif

! --- multiply h|phi> with mean-field matrix (kinetic terms) ---

      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hlochphi(psi,dtpsi,mc(mchpsi),mc(mchloch),1,1)
      else if (dentype .eq. 1) then
         call d1hlochphi(dtpsi,mc(mchpsi),mc(mchloch))
      endif

! --- multiply |phi> with sum over mean-fields (potential terms) ---

      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call mfsumphi(dtpsi,psi,mfsum)
      else if (dentype .eq. 1) then
         call d1mfsumphi(dtpsi,psi,mfsum)
      endif

! --- multiply with inverse of density matrix ---

      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call dicht1phi(dtpsi,dicht1)
      else if (dentype .eq. 1) then
         call d1dicht1phi(dtpsi,dicht1)
      endif

!    --- add uncorrelated motion to time derivative vector ---
      
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call addhunphi(dtpsi,mc(mchunpsi))
      else if (dentype .eq. 1) then
         call d1addhunphi(dtpsi,mc(mchunpsi))
      endif

!    --- apply projection operator on time derivative vector ---

      call project(psi,dtpsi)

      if (lgwp) then
         if(lrlx)then
            call gwp_param_a(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         else
            call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         endif
         call gwp_overlap(gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,0)

         modgh = 0
         call gh_elements(gh,hung,ghderiv1,agmat,gfgwp,aigwp,zgwp,&
              mr(mrqcent),1,modgh,gs2,time)

         call gdot(gs2,gsinv,hung,gsdot,psi,dtpsi,dicht2,&
              mi(migwpdyn),mi(migwpdep),mc(mchloch),gh,agmat,&
              ghderiv1)
         call dgwpdot(dtpsi,psi,gsinv,gsdot,hteil,gs2)
      endif

      if (lrunml) call compute_rhs_allphi(tree,psi,dtpsi)

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,dgldim)

! --- multiply time derivative with -i (or -1 for relaxation) ---

      call prefactor(dtpsi(adim+1),totphidim)
      call stop_timer(tid)

! save time-derivative of GWPs
      if (lgwp) mc(mcdtgwp:mcdtgwp+totphidim-1) = dtpsi(adim+1:dgldim)

      return
      end  subroutine funk1   

! **********************************************************************
! *                                                                    *
! *                          subroutine funk2                          *
! *                                                                    *
! * computes - after having propagated the single-particle functions - *
! * the matrix elements needed for the propagation of the              *
! * a-coefficients.                                                    *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, phihphi, hunphi, phihunphi                      *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   none                                                             *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.0 mb                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funk2 (time,psi)

      implicit none

      integer(long)                                  :: hktype,ktype,nham,&
                                                        i,modgh
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi

      integer(long), save                            :: tid = 0 ! timer ID
      logical(kind=4)                                :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funk2 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

! --- initialise variables --- 

      call get_timer(tid, 'Funk2')
      call start_timer(tid)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! compute gaussian matrix elements
!-----------------------------------------------------------------------
      if (lgwp) then
         if(lrlx)then
            call gwp_param_a(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         else
            call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
         endif
         call gwp_overlap(gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,0)
         modgh=0
         call gh_elements(gh,hung,ghderiv1,agmat,gfgwp,aigwp,zgwp,&
              mr(mrqcent),1,modgh,gs2,time)
      endif          

! --- compute action of correlated hamiltonian terms on spf's ---
! --- compute action of uncorrelated hamiltonian terms on spf's ---
! --- determine correlated matrix elements <phi|h|phi> ---
! --- determine uncorrelated matrix elements <phi|h_uncorr|phi> ---

      ktype=0
      hktype=0
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hphi(psi,mc(mchpsi),ktype,1)
         call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
         call hunphi(psi,mc(mchunpsi),hktype,1)
         call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,1)
      else if (dentype .eq. 1) then
         call d1hphi(psi,mc(mchpsi),ktype,1)
         call d1phihphi(psi,mc(mchpsi),hteil,1)
         call d1hunphi(psi,mc(mchunpsi),hktype,1)
         call d1phihunphi(psi,mc(mchunpsi),hteil,1)
      endif

      if (lrunml) then
         call compute_hterms_treer(time,tree,psi,1)
         call compute_all_uterms_tree(time,1,tree,psi)
      endif

      call stop_timer(tid)

      return
      end subroutine funk2 

! **********************************************************************
! *                                                                    *
! *                          SUBROUTINE FUNK3                          *
! *                                                                    *
! * Updates the mean fields, etc. after the A-coefficients have been   *
! * propagated. Note that the matrix elements must have already been   *
! * computed in "funk2". Note that the "intpsi" array is (mis)used as  *
! * complex work array!                                                *
! * Funk3 computes the time derivative dtpsi for both A and phi.       *
! *                                                                    *
! * Called from: propagatepsi                                          *
! * Calls to:    density, mfields, summf, summfphi, hlochphi, project, *
! *              hunphi, phihunphi, dicht1phi, addhunphi, calchuna     *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: action of hamiltonian on the wavefunction psi             *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funk3 (time,psi,dtpsi)

      implicit none

      integer(long)                                  :: i
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi
      complex(dop), dimension(:), allocatable        :: psi1

      integer(long), save                            :: tid = 0 ! timer ID
      logical(kind=4)                                :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funk3 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

!-----------------------------------------------------------------------
! allocate memory
!-----------------------------------------------------------------------
      if (lsvd) allocate(psi1(nmode*adim))

! --- initialise variables --- 

      call get_timer(tid, 'Funk3')
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

! --- compute density and inverse density matrices ---
! --- determine mean-field matrices and time derivative of a-vector ---
! --- compute sum over mean-field times single-particle operator ---
      if (lsvd) then
         call svdreg(psi,psi1,dicht1,dicht2,dicht3,dicht4)
      else
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
      endif

      if (lrunml) then
         call compute_mfields_tree(tree,psi)
         call kmfields(mc(mchloch))
         call compute_all_wterms_tree(time,1,tree,psi)

         if (dentype .eq. 1) then
            call d1summf(mfsum,mc(mchloch))
         else
            call compute_mfsum_tree(time,tree)
          endif
      else if (dentype .eq. 0) then
         if (lsvd) then
            call mfields(psi,dtpsi,1,psi1,0,usehsym) 
         else
            call mfields(psi,dtpsi,1,psi,0,usehsym) 
         endif
         call summf(mfsum,mc(mchloch),1)
      else if (dentype .eq. 1) then
         call d1mfields(psi,dtpsi)
         call d1summf(mfsum,mc(mchloch))
      else if (dentype .eq. 2) then
         call d2mfields(psi)
         if (lsil) then
            call d2calcha2(psi,dtpsi,hteil,1)
         else
            call d2calcha(psi,dtpsi,hteil,1)
         endif
         if (ldproj) then
            call d2calcdissa2(psi,dtpsi,hteil)
         else
            call d2calcdissa1(psi,dtpsi,hteil)
         endif
         call d2summf(mfsum,mc(mchloch))
      endif

! --- multiply h|phi> with mean-field matrix (kinetic terms) ---
! --- multiply |phi> with sum over mean-fields (potential terms) ---
! --- multiply with inverse of density matrix ---

      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call hlochphi(psi,dtpsi,mc(mchpsi),mc(mchloch),1,1)
         call mfsumphi(dtpsi,psi,mfsum)
         if (lrunml) call compute_mlop_noder(tree,psi,dtpsi)
         call dicht1phi(dtpsi,dicht1)
      else if (dentype .eq. 1) then
         call d1hlochphi(dtpsi,mc(mchpsi),mc(mchloch))
         call d1mfsumphi(dtpsi,psi,mfsum)
         call d1dicht1phi(dtpsi,dicht1)
      endif

!    --- add uncorrelated motion to time derivative vector ---    
      if (dentype .eq. 0 .or. dentype .eq. 2) then
         call addhunphi(dtpsi,mc(mchunpsi))
      else if (dentype .eq. 1) then
         call d1addhunphi(dtpsi,mc(mchunpsi))
      endif

!    --- apply projection operator on time derivative vector ---

      call project(psi,dtpsi)

      if (lgwp) then
         call gdot(gs2,gsinv,hung,gsdot,psi,dtpsi,dicht2,&
              mi(migwpdyn),mi(migwpdep),mc(mchloch),gh,agmat,&
              ghderiv1)
         call dgwpdot(dtpsi,psi,gsinv,gsdot,hteil,gs2)
      endif                                                          
      if (lrunml) then
          call compute_rhs_treer(tree,psi,dtpsi)

      else
! --- compute action of uncorrelated hamiltonian on a-vector ---

         if (dentype .eq. 0) then
            if (abs(psitype) .eq. 4) then
               call adot(psi,dtpsi,gsinv,gsdot,hteil)
               call calchuna_nogwp(psi,dtpsi,hteil,1,mi(mijindx))
            else
               call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
            endif
         else if (dentype .eq. 1) then
            call d1calchuna(psi,dtpsi,hteil)
         else if (dentype .eq. 2) then
            if (lsil) then
               call d2calchuna2(psi,dtpsi,hteil,1)
            else
               call d2calchuna(psi,dtpsi,hteil,1)
            endif
         endif
      endif

! --- SUBTRACT <psi|psidot> (ENERGY) FROM HAMILTONIAN IF RELAXATION TO RETAIN NORM ---
      if (lrlx .and. dentype .eq. 0 .and. rlaxnum .lt. 0 .and. &
               .not. lrlxt) then
         if (dentype .eq. 0 .and. psitype .eq. 4) then
            call calcenergy_g(psi,dtpsi,rlxenergy,0)
         else if (dentype .gt. 0 .and. psitype .eq. 4) then
            routine = 'Funk3'
            message = 'Relaxation for rho-GMCTDH not implemented'
            call errormsg
         else
            call calcenergy(psi,dtpsi,rlxenergy,0)
         endif
         if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
            call xvxxdzr(rlxenergy,psi,dtpsi,adim)
         else
            call subenergy(rlxenergy,psi,dtpsi)
         endif

      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,dgldim)

!print*,'Time 3',time/fs
!do i=1,dgldim
!  print*,i,dble(psi(i)),dimag(psi(i)),dble(dtpsi(i)),dimag(dtpsi(i))
!  print*,i,dble(dtpsi(i)),dimag(dtpsi(i))
!enddo
! --- MULTIPLY TIME DERIVATIVE WITH -I (OR -1 FOR RELAXATION) ---
! --- NO MULTIPLICATION (i.e. prefactor=1) FOR A-VECTOR IF DAVIDSON ---
 
      if(inttypea.eq.6) then
         call prefactor(dtpsi(adim+1),totphidim)
      else
         call prefactor(dtpsi,dgldim)
      endif
      if(lnumberproj) call numberproj(psi,dtpsi,0)

      if (allocated(psi1)) deallocate(psi1)

      call stop_timer(tid)

      return
      end subroutine funk3   

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE FUNKPHI                         *
! *                                                                    *
! * Computes the time derivative of the single-particle functions of   *
! * one mode with mean-field and inverse density matrices held         *
! * constant. The projector is applied after the uncorrelated          *
! * Hamiltonian terms have been added.                                 *
! *                                                                    *
! * Called from: abm, bs, RK5, RK8                                     *
! * Calls to:    hphi1m, hlochphi1m, project1m, dicht1phi1m, hunphi1m, *
! *              addhunphi1m, prefactor, summfphi1m                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  single-particle function vector of one mode                *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files. In particular, the actual mode is "currentmode".    *
! * This variable is in the module runpropmod.F90                      *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funkphi (time,psi,dtpsi)

      implicit none

      integer(long)                                   :: s,pointer,pointer1, &
                                                         hktype,ktype,f,t,i, &
                                                         modgh
      real(dop), intent(in)                           :: time
      real(dop)                                       :: time1
      complex(dop), dimension(*), intent(inout)       :: psi
      complex(dop), dimension(*), intent(out)         :: dtpsi

      integer(long)                                   :: tid,tid1 ! timer IDs
                                                             ! do not save
                                                             ! as name changes
      character(len=15)                               :: string
      logical(kind=4)                                 :: lcheck 

      time1=time   ! cp to prevent unused variable

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,phidim(currentmode)
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funkphi (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

! --- initialise variables --- 

      tid=0
      f=spfdof(1,currentmode)
      string='Funkphi-'//modelabel(f)(1:6)
      call get_timer(tid, string)
      call start_timer(tid)

      call zerovxz(dtpsi,phidim(currentmode))

      if (gwpm(currentmode)) then
         call gwp_param1m(currentmode,psi,agmat,gfgwp,aigwp,zgwp,&
              mr(mrqcent),zetf(currentmode,1),phidim(currentmode),0)

         call gwp_overlap1m (gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,&
              currentmode,0)

         tid1=0
         string='Gh_elems-'//modelabel(f)(1:6)
         call get_timer(tid1, string)
         call start_timer(tid1)
         modgh=0
         call gh_elements1m(currentmode,gh,hung,ghderiv1,agmat, &
                      gfgwp,aigwp,zgwp,mr(mrqcent),1, &
                      modgh,gs2,nspfdof(currentmode))
         call stop_timer(tid1)

         do s=1,nstate
            call gdot1ms(currentmode,s,psi,dtpsi,zetf(currentmode,s),&
                 phidim(currentmode),gs2,gsinv,hung,gsdot, &
                 dicht2,mi(migwpdyn),mi(migwpdep),&
                 mc(mchloch),gh,agmat,ghderiv1)

            call dgwpdot1ms(currentmode,s,dtpsi,psi,gsinv,&
                 gsdot,hteil,gs2,zetf(currentmode,1),phidim(currentmode))
         enddo

      else
! --- calculate action of correlated hamiltonian on spf's ---

         ktype=1
         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call hphi1m(psi,mc(mchpsi),currentmode, &
                     zetf(currentmode,1),phidim(currentmode),ktype,1)
         else if (dentype .eq. 1) then
            call d1hphi1m(psi,mc(mchpsi),currentmode, &
                 dzetf(currentmode,1,1),phidim(currentmode),ktype,1)
         endif

! --- multiply h|phi> with (constant) mean-field (kinetic terms) ---

! --- Use parallel hlochphi1m routine.
#ifdef MPI
          if ( lmpi .and. lmpihlochphi.and.lmpistarted) then
             if (dentype .eq. 0) then
                call hlochphi1mpar(psi,dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,1,1,zetf(currentmode,1), &
                     phidim(currentmode))
             else if (dentype .eq. 1) then
                call d1hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,dzetf(currentmode,1,1),phidim(currentmode))
             else if (dentype .eq. 2) then
                call d2hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                     zetf(currentmode,1),phidim(currentmode),1)
             endif
          else
! --- Use standard hlochphi1m routine.
#endif
            if (dentype .eq. 0) then
               call hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                    zetf(currentmode,1),phidim(currentmode),1,1)
            else if (dentype .eq. 1) then
                call d1hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,dzetf(currentmode,1,1),phidim(currentmode))
            else if (dentype .eq. 2) then
               call d2hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                    zetf(currentmode,1),phidim(currentmode),1)
            endif
#ifdef MPI
          endif
#endif

! --- multiply |phi> with sum over mean-fields (potential terms) ---

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call mfsumphi1m(dtpsi,psi,mfsum,currentmode, &
                         zetf(currentmode,1),phidim(currentmode))
         else if (dentype .eq. 1) then
            call d1mfsumphi1m(dtpsi,psi,mfsum,currentmode, &
                 dzetf(currentmode,1,1),phidim(currentmode))
         endif

! --- loop over each state ---

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            do s = 1,nstate
               pointer = zetf(currentmode,s)-zetf(currentmode,1)+1
               pointer1 = zetf(currentmode,s)-adim

!    --- multiply with (constant) inverse density matrix ---

               call dicht1phi1ms(dtpsi(pointer),dicht1, &
                                 currentmode,s,dim(currentmode,s), &
                                 subdim(currentmode))

!       --- calculate action of uncorrelated hamiltonian on spf's ---

               hktype=0
               call hunphi1ms(psi(pointer),mc(mchunpsi+pointer1-1), &
                              currentmode,s,dim(currentmode,s), &
                              subdim(currentmode),hktype,1)

!       --- add uncorrelated motion to time derivative vector ---

               call addhunphi1ms(dtpsi(pointer),mc(mchunpsi),currentmode,&
                    s,dim(currentmode,s),subdim(currentmode))

!       --- apply projection operator on time derivative vector ---

               call project1ms(psi(pointer),dtpsi(pointer), &
                               currentmode,dim(currentmode,s),subdim(currentmode))
            enddo

         else if (dentype .eq. 1) then

!    --- MULTIPLY WITH (CONSTANT) INVERSE DENSITY MATRIX ---

            do s = 1,nstate
               pointer = dzetf(currentmode,s,s)-dzetf(currentmode,1,1)+1
               call d1dicht1phi1ms(dtpsi(pointer),dicht1,&
                    currentmode,s,s,ddim(currentmode,s,s),subdim(currentmode))
               do t = (s+1),nstate
                  pointer = dzetf(currentmode,s,t)-dzetf(currentmode,1,1)+1
                  pointer1 = dzetf(currentmode,t,s)-dzetf(currentmode,1,1)+1
                  call d1dicht1phi1ms(dtpsi(pointer),dicht1,&
                       currentmode,s,t,ddim(currentmode,s,t),subdim(currentmode))
                  call makeaadj(dtpsi(pointer),dtpsi(pointer1),&
                       supgdim(currentmode),ddim(currentmode,s,t))
               enddo
            enddo

!     --- ADD UNCORRELATED TERMS BEFORE PROJECTION

            do s = 1,nstate
               do t = 1,nstate
                  pointer = dzetf(currentmode,s,t)-dzetf(currentmode,1,1)+1
                  pointer1 = dzetf(currentmode,s,t)-adim

!     --- SELECT ORDER OF OPERATORS: P-h OR h-P


!     --- CALCULATE ACTION OF UNCORRELATED HAMILTONIAN ON SPFs ---

                  hktype=0
                  call d1hunphi1ms(psi(pointer),mc(mchunpsi+pointer1-1),&
                       currentmode,s,t,ddim(currentmode,s,t),&
                       subdim(currentmode),hktype,1)

!     --- ADD UNCORRELATED MOTION TO TIME DERIVATIVE VECTOR ---

                  call d1addhunphi1ms(dtpsi,mc(mchunpsi),currentmode,&
                       s,t,ddim(currentmode,s,t),subdim(currentmode),&
                       dzetf(currentmode,1,1),phidim(currentmode))
               enddo
            enddo


!     --- APPLY PROJECTION OPERATOR ON TIME DERIVATIVE VECTOR ---

            do s=1,nstate
               do t=1,nstate
                  pointer = dzetf(currentmode,s,t)-dzetf(currentmode,1,1)+1
                  call project1ms(psi(pointer),dtpsi(pointer),&
                       currentmode,ddim(currentmode,s,t),subdim(currentmode))
               enddo
            enddo

         endif
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,phidim(currentmode))

!print*,'Time Phi1',currentmode,time/fs
!do i=1,phidim(currentmode)
!  print*,i,dble(psi(i)),dimag(psi(i)),dble(dtpsi(i)),dimag(dtpsi(i))
!  print*,i,dble(dtpsi(i)),dimag(dtpsi(i))
!enddo
! --- multiply time derivative with -i (or -1 for relaxation) ---

      call prefactor(dtpsi,phidim(currentmode))
      if(lnumberproj) call numberproj(psi,dtpsi,currentmode)
      call stop_timer(tid)

      return
      end subroutine funkphi

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE FUNKPHIML                       *
! *                                                                    *
! * Computes the time derivative of the single-particle functions of   *
! * one mode with mean-field and inverse density matrices held         *
! * constant. The projector is applied after the uncorrelated          *
! * Hamiltonian terms have been added.                                 *
! *                                                                    *
! * Version for ML-MCTDH Split-CMF  integrator                         *
! * single-set only                                                    *
! *                                                                    *
! * Called from: abm, bs, RK5, RK8                                     *
! * Calls to:    hphi1m, hlochphi1m, project1m, dicht1phi1m, hunphi1m, *
! *              addhunphi1m, prefactor, summfphi1m                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  single-particle function vector of one mode                *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files. In particular, the actual mode is "currentmode".    *
! * This variable is in a common block in runprop.inc.                 *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! * This routine is as funkphi but for single-set (prevents problems   *
! * pointers)                                                          *
! **********************************************************************

      subroutine funkphiml (time,psi,dtpsi)

      implicit none

      integer(long)                                   :: pointer1, modgh,&
                                                         hktype,ktype,f,i,s
      real(dop), intent(in)                           :: time
      real(dop)                                       :: time1
      complex(dop), dimension(:), intent(inout)       :: psi
      complex(dop), dimension(:), intent(out)         :: dtpsi

      integer(long)                                   :: tid, tid1 ! timer ID
                                                             ! do not save
                                                             ! as name changes
      character(len=15)                               :: string
      logical(kind=4)                                 :: lcheck 

      time1=time   ! cp to prevent unused variable

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,phidim(currentmode)
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funkphiml (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

! --- initialise variables --- 

      tid=0
      if (nspfdof(currentmode) .eq. 1) then
         f=spfdof(1,currentmode)
         string='Funk-'//modelabel(f)(1:10)
      else if (nspfdof(currentmode) .gt. 1) then
         write(string,'(a8,i3)') 'Funkphi-',currentmode
      endif
      call get_timer(tid, string)
      call start_timer(tid)

      call zerovxz(dtpsi,phidim(currentmode))

      if (gwpm(currentmode)) then
        call gwp_param1m(currentmode,psi,agmat,gfgwp,aigwp,zgwp,&
             mr(mrqcent),zetf(currentmode,1),phidim(currentmode),0)

         call gwp_overlap1m (gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,&
              currentmode,0)

         tid1=0
         f=spfdof(1,currentmode)
         string='Gh_elems-'//modelabel(f)(1:6)
         call get_timer(tid1, string)
         call start_timer(tid1)
         modgh=0
         call gh_elements1m(currentmode,gh,hung,ghderiv1,agmat, &
                      gfgwp,aigwp,zgwp,mr(mrqcent),1, &
                      modgh,gs2,nspfdof(currentmode))
         call stop_timer(tid1)

         do s=1,nstate
            call gdot1ms(currentmode,s,psi,dtpsi,zetf(currentmode,s),&
                 phidim(currentmode),gs2,gsinv,hung,gsdot, &
                 dicht2,mi(migwpdyn),mi(migwpdep),&
                 mc(mchloch),gh,agmat,ghderiv1)

            call dgwpdot1ms(currentmode,s,dtpsi,psi,gsinv,&
                 gsdot,hteil,gs2,zetf(currentmode,1),phidim(currentmode))
         enddo

      else
! --- calculate action of correlated hamiltonian on spf's ---

         ktype=1
         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call hphi1m(psi,mc(mchpsi),currentmode, &
                     zetf(currentmode,1),phidim(currentmode),ktype,1)
         else if (dentype .eq. 1) then
            call d1hphi1m(psi,mc(mchpsi),currentmode, &
                 dzetf(currentmode,1,1),phidim(currentmode),ktype,1)
         endif

! --- multiply h|phi> with (constant) mean-field (kinetic terms) ---

! --- Use parallel hlochphi1m routine.
#ifdef MPI
          if ( lmpi .and. lmpihlochphi.and.lmpistarted) then
             if (dentype .eq. 0) then
                call hlochphi1mpar(psi,dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,1,1,zetf(currentmode,1), &
                     phidim(currentmode))
             else if (dentype .eq. 1) then
                call d1hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,dzetf(currentmode,1,1),phidim(currentmode))
             else if (dentype .eq. 2) then
                call d2hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                     zetf(currentmode,1),phidim(currentmode),1)
             endif
          else
! --- Use standard hlochphi1m routine.
#endif
            if (dentype .eq. 0) then
               call hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                    zetf(currentmode,1),phidim(currentmode),1,1)
            else if (dentype .eq. 1) then
                call d1hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch), &
                     currentmode,dzetf(currentmode,1,1),phidim(currentmode))
            else if (dentype .eq. 2) then
               call d2hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
                    zetf(currentmode,1),phidim(currentmode),1)
            endif
#ifdef MPI
          endif
#endif

! --- multiply |phi> with sum over mean-fields (potential terms) ---

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call mfsumphi1m(dtpsi,psi,mfsum,currentmode, &
                         zetf(currentmode,1),phidim(currentmode))
         else if (dentype .eq. 1) then
            call d1mfsumphi1m(dtpsi,psi,mfsum,currentmode, &
                 dzetf(currentmode,1,1),phidim(currentmode))
         endif

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            pointer1 = zetf(currentmode,1)-adim

!    --- multiply with (constant) inverse density matrix ---

            call dicht1phi1ms(dtpsi,dicht1,currentmode,1,dim(currentmode,1), &
                              subdim(currentmode))

!       --- calculate action of uncorrelated hamiltonian on spf's ---

            hktype=0
            call hunphi1ms(psi,mc(mchunpsi+pointer1-1), &
                           currentmode,1,dim(currentmode,1), &
                           subdim(currentmode),hktype,1)

!       --- add uncorrelated motion to time derivative vector ---

            call addhunphi1ms(dtpsi,mc(mchunpsi),currentmode,1, &
                              dim(currentmode,1),subdim(currentmode))

!       --- apply projection operator on time derivative vector ---

            call project1ms(psi,dtpsi, &
                            currentmode,dim(currentmode,1),subdim(currentmode))

         else if (dentype .eq. 1) then

!    --- MULTIPLY WITH (CONSTANT) INVERSE DENSITY MATRIX ---

            call d1dicht1phi1ms(dtpsi,dicht1,currentmode,1,1, &
                 ddim(currentmode,1,1),subdim(currentmode))

!     --- ADD UNCORRELATED TERMS BEFORE PROJECTION

            pointer1 = dzetf(currentmode,1,1)-adim

!     --- SELECT ORDER OF OPERATORS: P-h OR h-P


!     --- CALCULATE ACTION OF UNCORRELATED HAMILTONIAN ON SPFs ---

            hktype=0
            call d1hunphi1ms(psi,mc(mchunpsi+pointer1-1),&
                 currentmode,1,1,ddim(currentmode,1,1),&
                 subdim(currentmode),hktype,1)

!     --- ADD UNCORRELATED MOTION TO TIME DERIVATIVE VECTOR ---

            call d1addhunphi1ms(dtpsi,&
                 mc(mchunpsi),currentmode,1,1,ddim(currentmode,1,1),&
                 subdim(currentmode),dzetf(currentmode,1,1),&
                 phidim(currentmode))

!     --- APPLY PROJECTION OPERATOR ON TIME DERIVATIVE VECTOR ---

            call project1ms(psi,dtpsi,currentmode,&
                 ddim(currentmode,1,1),subdim(currentmode))

         endif
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,phidim(currentmode))

! --- multiply time derivative with -i (or -1 for relaxation) ---

      call prefactor(dtpsi,phidim(currentmode))
      call stop_timer(tid)

      return
      end subroutine funkphiml

! **********************************************************************
!                     SUBROUTINE FUNKPHI2
!  applies funkphi twice (improved relaxation)     
! **********************************************************************

      subroutine funkphi2 (time,psi,dtpsi)

      implicit none

      integer(long)                                     :: s,pointer,pointer1, &
                                                           hktype,ktype,nham,f,i
      real(dop), intent(in)                             :: time
      complex(dop), dimension(maxphidim), intent(inout) :: psi
      complex(dop), dimension(maxphidim), intent(inout)   :: dtpsi
      complex(dop), dimension(128000)                   :: dthelp
      integer(long)                                     :: tid ! timer ID
                                                               ! do not save
                                                               ! as name changes
      character(len=15)                                 :: string
      logical(kind=4)                                   :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,maxphidim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funkphi2 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

      if (dentype .eq. 1) then
         routine = 'Funkphi2'
         message = 'Dentype 1 not implemented'
         call errormsg
      endif

      if(maxphidim.ge.128000) then
         write(6,'(2a,i7,a)')  'Enlarge dimension of dthelp to:', &
                               ' maxphidim =',maxphidim,'  STOP'
      endif


! --- INITIALISE VARIABLES --- 

      tid=0
      if (nspfdof(currentmode) .eq. 1) then
         f=spfdof(1,currentmode)
         string='Funk-'//modelabel(f)(1:10)
      else if (nspfdof(currentmode) .gt. 1) then
         write(string,'(a8,i3)') 'Funkphi-',currentmode
      endif
      call get_timer(tid, string)
      call start_timer(tid)
      call zerovxz(dtpsi,phidim(currentmode))
      call zerovxz(dthelp,phidim(currentmode))

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- CALCULATE ACTION OF CORRELATED HAMILTONIAN ON SPFs ---

      ktype=1
      call hphi1m(psi,mc(mchpsi),currentmode, &
                  zetf(currentmode,1),phidim(currentmode),ktype,1)

! --- MULTIPLY H|PHI> WITH (CONSTANT) MEAN-FIELD (KINETIC TERMS) ---
#ifdef MPI
       if ( lmpi.and.lmpihlochphi.and.lmpistarted) then
! --- Use parallel hlochphi1m routine.
          call hlochphi1mpar(psi,dtpsi,mc(mchpsi),mc(mchloch), &
                             currentmode,1,1,zetf(currentmode,1),phidim(currentmode))
      else
! --- Use standard hlochphi1m routine.
#endif
         call hlochphi1m(dthelp,mc(mchpsi),mc(mchloch),currentmode, &
                         zetf(currentmode,1),phidim(currentmode),1,1)
#ifdef MPI
      endif
#endif

! --- MULTIPLY |PHI> WITH SUM OVER MEAN-FIELDS (POTENTIAL TERMS) ---

      call mfsumphi1m(dthelp,psi,mfsum,currentmode, &
                      zetf(currentmode,1),phidim(currentmode))

! --- LOOP OVER EACH STATE ---

      do s = 1,nstate
         pointer = zetf(currentmode,s)-zetf(currentmode,1)+1
         pointer1 = zetf(currentmode,s)-adim

!    --- MULTIPLY WITH (CONSTANT) INVERSE DENSITY MATRIX ---

         call dicht1phi1ms(dthelp(pointer),dicht1, &
                           currentmode,s,dim(currentmode,s), &
                           subdim(currentmode))

!    --- CALCULATE ACTION OF UNCORRELATED HAMILTONIAN ON SPFs ---

         hktype=0
         call hunphi1ms(psi(pointer),mc(mchunpsi+pointer1-1), &
                        currentmode,s,dim(currentmode,s), &
                        subdim(currentmode),hktype,1)

!    --- ADD UNCORRELATED MOTION TO TIME DERIVATIVE VECTOR ---

         call addhunphi1ms(dthelp(pointer),mc(mchunpsi),currentmode,s, &
                           dim(currentmode,s),subdim(currentmode))

!    --- ADD -E*dthelp TO TIME DERIVATIVE VECTOR ---
!         call xvxxdza(-1._dop*edav,psi(pointer),dthelp(pointer),
!     +                dim(currentmode,s)*subdim(currentmode))

!    --- APPLY PROJECTION OPERATOR ON TIME DERIVATIVE VECTOR ---

!         call project1ms(psi(pointer),dthelp(pointer),
!     +                   currentmode,dim(currentmode,s),subdim(currentmode))
      enddo

! --  SECOND ROUND ------------------------------------------
! --- CALCULATE ACTION OF CORRELATED HAMILTONIAN ON SPFs  ---

      ktype=1
      call hphi1m(dthelp,mc(mchpsi),currentmode, &
                  zetf(currentmode,1),phidim(currentmode),ktype,1)

! --- MULTIPLY H|PHI> WITH (CONSTANT) MEAN-FIELD (KINETIC TERMS) ---
#ifdef MPI
       if (lmpi.and.lmpihlochphi.and.lmpistarted) then
! --- Use parallel hlochphi1m routine.
          call hlochphi1mpar(psi,dtpsi,mc(mchpsi),mc(mchloch), &
                             currentmode,1,1,zetf(currentmode,1),phidim(currentmode))
       else
! --- Use standard hlochphi1m routine.
#endif

         call hlochphi1m(dtpsi,mc(mchpsi),mc(mchloch),currentmode, &
              zetf(currentmode,1),phidim(currentmode),1,1)

#ifdef MPI
       endif
#endif

! --- MULTIPLY |PHI> WITH SUM OVER MEAN-FIELDS (POTENTIAL TERMS) ---

      call mfsumphi1m(dtpsi,dthelp,mfsum,currentmode, &
                      zetf(currentmode,1),phidim(currentmode))

! --- LOOP OVER EACH STATE ---

      do s = 1,nstate
         pointer = zetf(currentmode,s)-zetf(currentmode,1)+1
         pointer1 = zetf(currentmode,s)-adim

!    --- MULTIPLY WITH (CONSTANT) INVERSE DENSITY MATRIX ---

         call dicht1phi1ms(dtpsi(pointer),dicht1, &
                           currentmode,s,dim(currentmode,s), &
                           subdim(currentmode))


!       --- CALCULATE ACTION OF UNCORRELATED HAMILTONIAN ON SPFs ---

         hktype=0
         call hunphi1ms(dthelp(pointer),mc(mchunpsi+pointer1-1), &
                        currentmode,s,dim(currentmode,s), &
                        subdim(currentmode),hktype,1)

!    --- ADD UNCORRELATED MOTION TO TIME DERIVATIVE VECTOR ---

         call addhunphi1ms(dtpsi(pointer),mc(mchunpsi),currentmode,s, &
                           dim(currentmode,s),subdim(currentmode))


!    --- ADD -2E*dthelp TO TIME DERIVATIVE VECTOR ---
!         call xvxxdza(-2._dop*edav,dthelp(pointer),dtpsi(pointer),
!     +                dim(currentmode,s)*subdim(currentmode))

!    --- APPLY PROJECTION OPERATOR ON TIME DERIVATIVE VECTOR ---

         call project1ms(psi(pointer),dtpsi(pointer), &
                         currentmode,dim(currentmode,s),subdim(currentmode))
      enddo

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,maxphidim)

! --- MULTIPLY TIME DERIVATIVE WITH -I (OR -1 FOR RELAXATION) ---

      call prefactor(dtpsi,phidim(currentmode))
      call stop_timer(tid)
 
      return
      end subroutine funkphi2

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE FUNKGBF                         *
! *                                                                    *
! * Copies the time derivative of the GBFs from a tempory array        *
! * to dtpsi.                                                          *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  single-particle function vector of one mode                *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * The interface must be same as funkphi. Hence unused variables.     *
! *                                                                    *
! **********************************************************************

      subroutine funkgbf (time,psi,dtpsi)

      implicit none

      integer(long)                                   :: m,zi,zf
      real(dop), intent(in)                           :: time
      real(dop)                                       :: time1
      complex(dop), dimension(maxgbfdim), intent(inout):: psi
      complex(dop)                                     :: psi1
      complex(dop), dimension(maxgbfdim), intent(inout) :: dtpsi

      time1=time   ! cp to prevent unused variable
      psi1=psi(1)   ! cp to prevent unused variable

      m=currentmode
      zi = mcdtgwp+(zphi(m,1)-(adim+1))
      zf = gbfdim(m)
      dtpsi(1:zf) = mc(zi:zi+zf-1)

      return
      end subroutine funkgbf


! **********************************************************************
! *                                                                    *
! *                          SUBROUTINE FUNKA2                         *
! *                                                                    *
! * Performs an evaluation of the right hand side of the MCTDH working *
! * equation for the MCTDH-coefficients. Used when only the            *
! * A-coefficients are propagated.                                     *
! * Special routine using real arithmetic throughout.                  * 
! * no denstity, no el. states, no muld, only real operators (no p).   *
! *                                                                    *
! * Called from: rdavstep  (rrDAV)                                     *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  MCTDH-coefficient vector (real!)                           *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi (real!)                        *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * HDM 10/05                                                          *
! *                                                                    *
! * Only used in relaxation so no need to calculate time-dependent     *
! * Hamiltonian coefficients 
! **********************************************************************

      subroutine funka2 (time,psi,dtpsi)

      implicit none

      integer(long)                                 :: h,k,s,s1,k0,koffset,m,&
                                                       zeig1,zeig2,zeig3,&
                                                       vdim1,i
      real(dop), intent(in)                         :: time
      real(dop)                                     :: time1
      real(dop), dimension(adim), intent(in)        :: psi
      real(dop), dimension(adim), intent(inout)       :: dtpsi
      integer(long), save                           :: tid=0 ! timer ID
      logical(kind=4)                               :: lcheck 

      time1=time   ! cp to prevent unused variable

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,adim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funka2 (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

      if (dentype .eq. 1) then
         routine = 'Funkphi2'
         message = 'Dentype 1 not implemented'
         call errormsg
      endif

! --- INITIALISE VARIABLES --- 

      call get_timer(tid, 'Funka')
      call start_timer(tid)
      call zerovxd(dtpsi,adim)

! --- CALCULATE RIGHT HAND SIDE FOR HAMILTONIAN ---

      koffset=zham(1)+kzahl(1)-1
      k0=zham(1)

#ifdef MPI
      if(lmpi.and.lmpifunka2.and.lmpistarted) then
         if(.not.lmpidavrun) then
            mpitask=1
! --- Determine Job
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         endif
         call mpifunka2(psi,dtpsi)
      else
#endif
!-----------------------------------------------------------------------
!     If no parallelisation is used then the standard calculation is 
!     performed.
!-----------------------------------------------------------------------
! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

         do k=k0,koffset
            if(kskip.eq.0)  goto 88
            if(diag(meb,k)) goto 88 ! next line: reflex
            if(abs(hteil(pmat(meb,k)-3+kskip*3)).le.1.d-8) goto 99
 88         s = kf(k)
            s1= ki(k)
            if(s.eq.s1) then
               zeig2 = maxblock + 1
               zeig1 = 1
               call cpvxd(psi(zpsi(s)),mr(mrworkr+zeig2-1),block(s))
               do m = 1,nmode
                  if (.not. diag(m,k)) then

!       --- SWAP POINTERS OR SELECT COEFFICIENTS ---

                     zeig3 = zeig1
                     zeig1 = zeig2
                     zeig2 = zeig3

!       --- MULTIPLY WITH MATRIX ELEMENTS ---

              !       if(.not.lompthread) then
                       call qtxxzd(hteil(pmat(m,k)), &
                            mr(mrworkr+zeig1-1),mr(mrworkr+zeig2-1), &
                            vdim(m,s),dim(m,s),ndim(m,s))
              !       else
              !          call qtxxzdomp(hteil(pmat(m,k)), &
              !               mr(mrworkr+zeig1-1),mr(mrworkr+zeig2-1), &
              !               vdim(m,s),dim(m,s),ndim(m,s),ompthread)
              !       endif

                  endif
               enddo
            else
               zeig2 = maxblock+1
               zeig1 = 1
               call cpvxd(psi(zpsi(s1)),mr(mrworkr+zeig2-1),block(s1))

               vdim1 = vdim(1,s1)
               do m = 1,nmode

!    --- SWAP POINTERS ---

                  zeig3 = zeig1
                  zeig1 = zeig2
                  zeig2 = zeig3

!    --- MULTIPLY WITH MATRIX ELEMENTS ---

!                  call zerovxd(mr(mrworkr+zeig2-1),maxblock)
                !     if(.not.lompthread) then
                        call mtxxzd(hteil(pmat(m,k)), &
                             mr(mrworkr+zeig1-1),mr(mrworkr+zeig2-1), &
                             vdim1,dim(m,s1),ndim(m,s1),dim(m,s))
                !     else
                !        call mtxxzdomp(hteil(pmat(m,k)), &
                !             mr(mrworkr+zeig1-1),mr(mrworkr+zeig2-1), &
                !             vdim1,dim(m,s1),ndim(m,s1),dim(m,s),ompthread)
                !     endif
                  vdim1 = vdim1*dim(m,s)
               enddo
            endif

!    --- STORE RIGHT HAND SIDE IN DTPSI ---

            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then
               call addvxxdo(mr(mrworkr+zeig2-1),dtpsi(zpsi(s)),block(s))
            else
               call xvxxdda(dble(koeff(k)),mr(mrworkr+zeig2-1),dtpsi(zpsi(s)), &
                            block(s))
            endif
 99      continue
         enddo
#ifdef MPI
      endif
! --- COMPUTE ACTION OF UNCORRELATED HAMILTONIAN ---
      if (mpirank.eq.0) then
#endif
         if (lfunkr .or. lcap) then
            k=zham(1)+kzahl(1)-1
            do s=1,nstate
               do m=1,nmode
                  if(khzahl(m,s,1) .gt. 0) &
                       call qtxxzda(hteil(pmat(m,k+s)),psi(zpsi(s)), &
                       dtpsi(zpsi(s)),vdim(m,s),dim(m,s),ndim(m,s))
               enddo
            enddo
         endif
         
! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
         call xvxxddr(rlxenergy,psi,dtpsi,adim)
#ifdef MPI
      endif
#endif

      call stop_timer(tid)

      return
      end subroutine funka2

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE FUNKA                         *
! *                                                                    *
! * Performs an evaluation of the right hand side of the MCTDH working *
! * equation for the MCTDH-coefficients. Used when only the            *
! * A-coefficients are propagated.                                     *
! *                                                                    *
! * Called from: abm, bs, lanczos, davstep, sil                        *
! * Calls to:    calcha, calchuna, prefactor, subenergy                *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  MCTDH-coefficient vector                                   *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funka (time,psi,dtpsi)

      implicit none


      real(dop), intent(in)                        :: time
      real(dop)                                    :: time1
      complex(dop), dimension(adim), intent(in)    :: psi
      complex(dop), dimension(adim), intent(inout)   :: dtpsi
      integer(long)                                :: z0,zf,i

      integer(long), save                          :: tid = 0 ! timer ID
      logical(kind=4)                              :: lcheck 

      time1=time   ! cp to prevent unused variable

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,adim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funka (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

! --- initialise variables --- 

      call get_timer(tid, 'Funka')
      call start_timer(tid)
      call zerovxz(dtpsi,adim)

! --- calculate right hand side for hamiltonian ---

      if (psitype .eq. 5) then
         tz => tree%swap_down(1)%p
         z0 = tz%zr(1,1)
         zf = tz%zr(1,2)
         call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))
      else if (dentype .eq. 0) then
         call calcha(psi,dtpsi,hteil,mr(mrhops),mc(mcphi),1,&
              mc(mcphi),mi(mijindx),mr(mrort))
      else if (dentype .eq. 1) then
         if (lsil) then
            call d1calcha2(psi,dtpsi,hteil,1)
         else
            call d1calcha(psi,dtpsi,hteil,1)
         endif
         call d1calcdissa(psi,dtpsi,hteil)
         if (ldissin .and. .not. lsil) call chkinib(dtpsi,.false.)
      else if (dentype .eq. 2) then
         if (lsil) then
            call d2calcha2(psi,dtpsi,hteil,1)
         else
            call d2calcha(psi,dtpsi,hteil,1)
         endif
         if (ldproj) then
            call d2calcdissa2(psi,dtpsi,hteil)
         else
            call d2calcdissa1(psi,dtpsi,hteil)
         endif
      endif

#ifdef MPI
      if (mpirank.eq.0) then
#endif

! --- compute action of uncorrelated hamiltonian ---
      if (psitype .ne. 5) then
         if (dentype .eq. 0) then
            call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
         else if (dentype .eq. 1) then
            if (lsil) then
               call d1calchuna2(psi,dtpsi,hteil)
            else
               call d1calchuna(psi,dtpsi,hteil)
            endif
         else if (dentype .eq. 2) then
            if (lsil) then
               call d2calchuna2(psi,dtpsi,hteil,1)
            else
               call d2calchuna(psi,dtpsi,hteil,1)
            endif
         endif
      endif

! --- subtract energy from hamiltonian if relaxation ---
      if (lrlx .and. .not.lrlxt) then
         if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
            call xvxxdzr(rlxenergy,psi,dtpsi,adim)
         else
            call subenergy(rlxenergy,psi,dtpsi)
         endif
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,adim)

! --- multiply time derivative with -i (or -1 for relaxation) ---

      if(inttypea.ne.6) call prefactor(dtpsi,adim)
#ifdef MPI
      endif
#endif

      call stop_timer(tid)

      return
      end subroutine funka

!***********************************************************************
! 
!                           subroutine funk
!
! computes the time derivative of the complete wavefunction. the
! projector is applied before the uncorrelated hamiltonian terms are
! added.
!
! called from: propagatepsi, abm, bs
! calls to:    density, hphi, phihphi, mfields, hlochphi, project,
!              dicht1phi, hunphi, addhunphi, 
!              prefactor
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine funk (time,psi,dtpsi)

      implicit none

      integer(long)                                  :: hktype,ktype,nham
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi
      complex(dop), dimension(:), allocatable        :: psi1
      integer(long)                                  :: m,s,zeig2,i

      integer(long), save                            :: tid = 0 ! timer ID
      logical(kind=4)                                :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) lcheck = .false.
      enddo
      if (.not. lcheck) then
         routine = 'Funk (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

!-----------------------------------------------------------------------
! Multilayer run should use Funkr
!-----------------------------------------------------------------------
      if (lrunml) then
         routine = 'Funk (function.F90)'
         message = 'ML-MCTDH must use funkr - should not be here'
         call errormsg
      end if

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call get_timer(tid, 'Funk')
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

!-----------------------------------------------------------------------
! allocate memory
!-----------------------------------------------------------------------
      if (lsvd) allocate(psi1(nmode*adim))

      zeig2=0
      if (lBEL .and. (BELflags(7)/=0)) then
         do s=1,nstate
            do m = 1,nmode
               if (gwpm(m)) then
                  zeig2=zeig2+dim(m,s)*dim(m,s)
               end if
            end do
         end do
         if (allocated(saveworkc)) then
           if (size(saveworkc)<zeig2) deallocate(saveworkc)
         end if
      end if


      if ((.not. lBEL) .or. (BELflags(7)==0) .or.(reCompDtPsi==1) &
           .or. (.not. allocated(saveworkc))) then

         if (lBEL .and. (BELflags(7)/=0) .and. &
              (.not. allocated(saveworkc))) then
              ! allocate memory to store S^-1 H for subsequent calls 
            allocate(saveworkc(zeig2))
         endif


!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
         nham=1
         if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
         if (ltdp) then
            if (lrlx) then
               call calctdp(0.0d0,psi,mr(mrort))
            else
               call calctdp(time,psi,mr(mrort))
            endif
         endif

!-----------------------------------------------------------------------
! compute gaussian matrix elements
!-----------------------------------------------------------------------
         if (lgwp) then
            if(lrlx)then
               call gwp_param_a(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
            else
               call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
            endif
            call gwp_overlap(gs2,gsinv,mr(mrgsval),gsvek,gfgwp,aigwp,zgwp,&
                 0)
         endif

!-----------------------------------------------------------------------
! Compute density and inverse density matrices
!-----------------------------------------------------------------------
         if (lsvd) then
            call svdreg(psi,psi1,dicht1,dicht2,dicht3,dicht4)
         else
            call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
         endif

!-----------------------------------------------------------------------
! Loop over correlated Hamiltonian terms
! 1. compute hpsi array
! 1a. calculate matrix elements for GWP modes
! 2. compute hteil matrices
! 3. compute hloch matrices and dtpsi for A coefficients
! 4. multiply hloch and hpsi and store in dtpsi
!-----------------------------------------------------------------------
         if (lgwp) call gh_elements(gh,hung,ghderiv1,agmat, &
                     gfgwp,aigwp,zgwp,mr(mrqcent),1,0,gs2,time)
      
         ktype=0
         if (dentype .eq. 0) then
            call hphi(psi,mc(mchpsi),ktype,1)
            call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
            if (lsvd) then
               call mfields(psi,dtpsi,1,psi1,0,usehsym)
            else
               call mfields(psi,dtpsi,1,psi,0,usehsym)
            endif
         else if (dentype .eq. 1) then
            call d1hphi(psi,mc(mchpsi),ktype,1)
            call d1phihphi(psi,mc(mchpsi),hteil,1)
            call d1mfields(psi,dtpsi)
         else if (dentype .eq. 2) then
            ktype=0
            call hphi(psi,mc(mchpsi),ktype,1)
            call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
            call d2mfields(psi)
            call d2calcha(psi,dtpsi,hteil,1)
            if (ldproj) then
               call d2calcdissa2(psi,dtpsi,hteil)
            else
               call d2calcdissa1(psi,dtpsi,hteil)
            endif
         endif

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call hlochphi(psi,dtpsi,mc(mchpsi),mc(mchloch),1,0)
         else if (dentype .eq. 1) then
            call d1hlochphi(dtpsi,mc(mchpsi),mc(mchloch))
         endif

!-----------------------------------------------------------------------
! remaining action of hamiltonian on single-particle functions
! 1. multiply with inverse density matrix
! 2. determine and add any cap uncorrelated hamiltonian terms
! 3. apply projection operator
! 4. determine and add action of "normal" uncorrelated hamiltonian terms
!    and correct coefficient motion for natural orbitals if required
! 5. multiply with -i (or -1 for relaxation)
!-----------------------------------------------------------------------

         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call dicht1phi(dtpsi,dicht1)
         else if (dentype .eq. 1) then
            call d1dicht1phi(dtpsi,dicht1)
         endif
         
         if (lcap) then
            hktype=2
            if (dentype .eq. 0 .or. dentype .eq. 2) then
               call hunphi(psi,mc(mchunpsi),hktype,1)
               call addhunphi(dtpsi,mc(mchunpsi))
               call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,nham)
            else if (dentype .eq. 1) then
               call d1hunphi(psi,mc(mchunpsi),hktype,1)
               call d1addhunphi(dtpsi,mc(mchunpsi))
               call d1phihunphi(psi,mc(mchunpsi),hteil,1)
            endif

            if (dentype .eq. 0 .and. .not. (psitype .eq. 4)) then
               call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
            else if (dentype .eq. 1) then
               call d1calchuna(psi,dtpsi,hteil)
            else if (dentype .eq. 2) then
               call d2calchuna(psi,dtpsi,hteil,1)
            endif
         endif

         call project(psi,dtpsi)

         hktype=1
         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call hunphi(psi,mc(mchunpsi),hktype,1)
            call addhunphi(dtpsi,mc(mchunpsi))
         else if (dentype .eq. 1) then
            call d1hunphi(psi,mc(mchunpsi),hktype,1)
            call d1addhunphi(dtpsi,mc(mchunpsi))
         endif
         
         if (lxloc) then
            call xdot(psi,dtpsi,dicht1,mc(mchloch),mc(mcpcons))
         endif

         if (lgwp) then
            if (.not. lcap) then
               call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,nham)
            endif
            
            call gdot(gs2,gsinv,hung,gsdot,psi,dtpsi,dicht2,&
                 mi(migwpdyn),mi(migwpdep),mc(mchloch),gh,agmat,&
                 ghderiv1)
            if (psitype .eq. 0 .or. psitype .eq. 5) then
               call dgwpdot(dtpsi,psi,gsinv,gsdot,hteil,gs2)
            endif
         endif

      endif

      if (psitype .eq. 4) then
         call adot(psi,dtpsi,gsinv,gsdot,hteil)
      end if

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,dgldim)

      call prefactor(dtpsi,dgldim)
      if(lnumberproj) call numberproj(psi,dtpsi,0)

      if (allocated(psi1)) deallocate(psi1)

      call stop_timer(tid)

      if ((.not. lBEL) .or. (BELflags(7)==0)) then
         ! Please don't just deallocate this. The whole point of saveworkc is to keep computed
         ! values between calls of funk!!

         if(allocated(saveworkc)) deallocate(saveworkc)
      end if

      return
      end subroutine funk

!***********************************************************************
!
!                         mctdh subroutine funkr
!
! calculates the derivative wrt time of the mctdh wavefunction.
! single-particle operators are on the right of the projector.
! caps can thus be included as single-particle operators as long as the
! complex cap coefficients are defined as the corresponding uncorrelated
! expansion coefficient koeffh(k_cap,m,s) to ensure hermiticity of the
! remaining real potential matrix.
!
! called from: propagatepsi, abm, bs
! calls to:    density, hphi, phihphi, mfields, hlochphi, project,
!              hunphi, phihunphi, dicht1phi, addhunphi, calchuna,
!              prefactor, subenergy
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine funkr (time,psi,dtpsi)

      implicit none

      integer(long)                                  :: hktype,ktype,nham
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi
      complex(dop), dimension(:), allocatable        :: psi1

      integer(long)                                  :: m,s,zeig2,i

      integer(long), save                            :: tid = 0 ! timer ID

      logical(kind=4)                                :: lcheck 

!-----------------------------------------------------------------------
! test psi is OK
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         if (psi(i) .ne. psi(i)) then
            lcheck = .false.
         endif
      enddo
      if (.not. lcheck) then
         routine = 'Funkr (function.F90)'
         message = 'Psi has become corrupted '
         call errormsg
      endif

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      if(lrunml) then
         call get_timer(tid, 'funkml')
      else
         call get_timer(tid, 'Funkr')
      endif
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

!-----------------------------------------------------------------------
! allocate memory
!-----------------------------------------------------------------------
      if (lsvd) allocate(psi1(nmode*adim))

      zeig2=0
      if (lBEL .and. (BELflags(7)/=0)) then
         do s=1,nstate
            do m = 1,nmode
               if (gwpm(m)) then
                  zeig2=zeig2+dim(m,s)*dim(m,s)
               end if
            end do
         end do
         if (allocated(saveworkc)) then
           if (size(saveworkc)<zeig2) deallocate(saveworkc)
         end if
      end if

      if ((.not. lBEL) .or. (BELflags(7)==0) .or.(reCompDtPsi==1) &
           .or. (.not. allocated(saveworkc))) then

         if (lBEL .and. (BELflags(7)/=0).and.(.not. allocated(saveworkc))) then
         ! allocate memory to store S^-1 H for subsequent calls 
            allocate(saveworkc(zeig2))
         endif

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
         nham=1
         if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
         if (ltdp) then
            if (lrlx) then
               call calctdp(0.0_dop,psi,mr(mrort))
            else
               call calctdp(time,psi,mr(mrort))
            endif
         endif

!-----------------------------------------------------------------------
! Check periodic GWPs are inside range
!-----------------------------------------------------------------------
         if (lpergwp) then
            call chkpergwp(psi)
         endif

!-----------------------------------------------------------------------
! compute gaussian matrix elements
!-----------------------------------------------------------------------
         if (lgwp) then
            if(lrlx)then
               call gwp_param_a(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
            else if (lgwptrj) then
               call gwptrj_param(m,time,psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent))
            else
               call gwp_param(psi,agmat,gfgwp,aigwp,zgwp,mr(mrqcent),2)
            endif
            if (lBEL) then
               call gwp_overlap(gs2,gsinv,mr(mrgsval), &
                    gsvek,gfgwp,aigwp,zgwp,1)
            else
               call gwp_overlap(gs2,gsinv,mr(mrgsval), &
                    gsvek,gfgwp,aigwp,zgwp,0)
            endif

            if (lsheptraj) write(itmpin,*)'|A|',real(abs(psi(1:adim)))

            if (lBEL) then
               call gh_elements(gh,hung,ghderiv1,agmat, &
                    gfgwp,aigwp,zgwp,mr(mrqcent),1,1,gs2,time)       
            else
               call gh_elements(gh,hung,ghderiv1,agmat, &
                    gfgwp,aigwp,zgwp,mr(mrqcent),1,0,gs2,time)       
            endif
         endif

!-----------------------------------------------------------------------
! compute density and inverse density matrices
!-----------------------------------------------------------------------
         if (lsvd) then
            call svdreg(psi,psi1,dicht1,dicht2,dicht3,dicht4)
         else
            call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)
         endif

!-----------------------------------------------------------------------
! loop over correlated hamiltonian terms
! 1. compute hpsi array
! 2. compute hteil matrices
! 3. compute hloch matrices and dtpsi for a coefficients
!-----------------------------------------------------------------------
         ktype=0
         hktype=0
         if (dentype .eq. 0 .or. dentype .eq. 2 ) then
            call hphi(psi,mc(mchpsi),ktype,1)
            call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
  
            call hunphi(psi,mc(mchunpsi),hktype,1)
            call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,nham)

         else if (dentype .eq. 1) then
            call d1hphi(psi,mc(mchpsi),ktype,1)  
            call d1phihphi(psi,mc(mchpsi),hteil,1)
            call d1hunphi(psi,mc(mchunpsi),hktype,1)
            call d1phihunphi(psi,mc(mchunpsi),hteil,1)

         end if

         if (lrunml) then
           call compute_hterms_treer(time,tree,psi,1)
           call compute_all_uterms_tree(time,1,tree,psi)
           call compute_all_wterms_tree(time,1,tree,psi)
         endif

      end if

!-----------------------------------------------------------------------
! Calculate mean fields
!-----------------------------------------------------------------------
      if (lrunml) then
         call compute_mfields_tree(tree,psi)
         call kmfields(mc(mchloch))
      else if (dentype .eq. 0) then
         if (lsvd) then
            call mfields(psi,dtpsi,1,psi1,0,usehsym)
         else
            call mfields(psi,dtpsi,1,psi,0,usehsym)
         endif
      else if (dentype .eq. 1) then
         call d1mfields(psi,dtpsi)
      else if (dentype .eq. 2) then
         call d2mfields(psi)
      endif

!-----------------------------------------------------------------------
! Add on MLop contribution to lowest layer
!-----------------------------------------------------------------------
      if (lrunml) call compute_mlop_noder(tree,psi,dtpsi)

      if ((.not. lBEL) .or. (BELflags(7)==0) .or.(reCompDtPsi==1) &
           .or. (.not. allocated(saveworkc))) then
!-----------------------------------------------------------------------
! 4. multiply hloch and hpsi and store in dtpsi
! remaining action of hamiltonian on single-particle functions
! 1. multiply with inverse density matrix
! 2. determine action of uncorrelated hamiltonian terms
! 3. add uncorrelated motion to time derivative
! 4. compute matrix elements <phi|h_uncorr|phi>
! 5. apply projection operator
!-----------------------------------------------------------------------
         if (dentype .eq. 0 .or. dentype .eq. 2) then
            call hlochphi(psi,dtpsi,mc(mchpsi),mc(mchloch),1,0)
            if (.not.lBEL) then
               call dicht1phi(dtpsi,dicht1)
            endif
            call addhunphi(dtpsi,mc(mchunpsi))
            call project(psi,dtpsi)
         else if (dentype .eq. 1) then 
            call d1hlochphi(dtpsi,mc(mchpsi),mc(mchloch))
            call d1dicht1phi(dtpsi,dicht1) 
            call d1addhunphi(dtpsi,mc(mchunpsi))
            call project(psi,dtpsi)
         end if

!if (TerryFlag) stop

!-----------------------------------------------------------------------
!     now calculate parameter dynamics (gdot) and modified time
!     evolution of a coefficients (adot)
!-----------------------------------------------------------------------
         if (lgwptrj) then
            call gdottrj(time,dtpsi(zphi(1,1)),gs2,gsdot,&
                 agmat,hung,psi(zxi(1,1)),nspfdof(1),gwpdim(1,1),npg(1),gh)

         else if (lgwp) then
            if (.not.lBEL) then
              call gdot(gs2,gsinv,hung,gsdot,psi,dtpsi,dicht2,&
                   mi(migwpdyn),mi(migwpdep),mc(mchloch),gh,agmat,&
                   ghderiv1)
            endif
            if (psitype .eq. 0 .or. psitype .eq. 5) then
               call dgwpdot(dtpsi,psi,gsinv,gsdot,hteil,gs2)
            endif
         end if

!-----------------------------------------------------------------------
! Compute RHS for ML-MCTDH, excluding primitive nodes
!-----------------------------------------------------------------------
         if (lrunml) call compute_rhs_treer(tree,psi,dtpsi)
      end if

!-----------------------------------------------------------------------
! 1. compute action of uncorrelated hamiltonian on a-vector
!    note that the action of the correlated hamiltonian on the a-vector
!    for WPs has already been computed (as a by-product) in "mfields".
! 2. multiply with -i (or -1 for relaxation). for relaxation, transform
!    h|psi> to (h-e)|psi> before.
!-----------------------------------------------------------------------
      if (.not. lrunml) then
         if (dentype .eq. 0) then
            if (abs(psitype) .eq. 4) then
               call adot(psi,dtpsi,gsinv,gsdot,hteil)
               call calchuna_nogwp(psi,dtpsi,hteil,1,mi(mijindx))

            else
               call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
            endif
!if (TerryFlag) write(*,*)3,dtpsi(1)

         else if (dentype .eq. 2) then
            call d2calchuna(psi,dtpsi,hteil,1)

            call d2calcha(psi,dtpsi,hteil,1)
            if (ldproj) then
               call d2calcdissa2(psi,dtpsi,hteil)
            else
               call d2calcdissa1(psi,dtpsi,hteil)
            endif

         else if (dentype .eq. 1) then
            call d1calchuna(psi,dtpsi,hteil)

         endif
      end if

! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
      if (lrunml) then
         ! do top layer projector;
         ! but not for ml-cmf=split where we renormalize instead
         if (tree%isrelax .and. tree%cmftyp /= 2) then
            call projectout_psi(tree,psi,dtpsi)
         endif

      else
         if (lrlx .and. .not. lrlxt) then
            if (dentype .eq. 0 .and. psitype .eq. 4) then 
               call subengwp(dicht2,psi,mc(mchunpsi), &
                    gs2,mc(mchloch),hteil, &
                    mc(mcpcons),dtpsi)

            else if (dentype .gt. 0 .and. psitype .eq. 4) then
               routine = 'Funkr'
               message = 'Relaxation for rho-GMCTDH not implemented'
               call errormsg

            else 
               call calcenergy(psi,dtpsi,rlxenergy,0)
               if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
                  call xvxxdzr(rlxenergy,psi,dtpsi,adim)
               else
                  call subenergy(rlxenergy,psi,dtpsi)
               endif
            endif
         endif
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,dgldim)

!print*,'Time ',time/fs
!do m=1,dgldim
!  print*,m,dble(psi(m)),dimag(psi(m)),dble(dtpsi(m)),dimag(dtpsi(m))
!   print*,m,dble(dtpsi(m)),dimag(dtpsi(m))
!   print*,m,dble(psi(m)),dimag(psi(m))
!enddo

      call prefactor(dtpsi,dgldim)
      if(lnumberproj) call numberproj(psi,dtpsi,0)

      if (allocated(psi1)) deallocate(psi1)

      call stop_timer(tid)

      if ((.not. lBEL) .or. (BELflags(7)==0)) then
         ! Please don't just deallocate this. The whole point of saveworkc is to keep computed
         ! values between calls of funkr!!

         if(allocated(saveworkc)) deallocate(saveworkc)
      end if

      return
      end subroutine funkr

!#######################################################################

      subroutine constener(gs2,gsdot,dtpsi,psi,dicht2,lconsten_iter)

use xvlib, only: qvxxzz,vvaxzz,xvxxdzo
use mtlib, only: qtxxzz
use op2lib, only: sum2qxxz

      implicit none

      integer(long)             :: s
      integer(long)                                  :: zeig1,zeig2,zeig3,m

      complex(dop), dimension(gs2matdim), intent(in) :: gs2
      complex(dop), dimension(gsmatdim), intent(in)  :: gsdot
      complex(dop), dimension(dmatdim), intent(in)   :: dicht2
      complex(dop)                                   :: ztmp1
      complex(dop)                                   :: etot
      complex(dop), dimension(dgldim), intent(in)    :: dtpsi
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(2*maxblock)            :: ztmp
      real(dop)                                      :: dscale,etotd,dke,dele
      logical(kind=4), intent(inout)                   :: lconsten_iter

      lconsten_iter = .false.

! At this point dtpsi = -i H psi, but gsdot is still <g| gdot>. Hence
! need to take i <g|gdot>

      ztmp=0.0_dop
      ztmp1=0.0_dop
      etot=0.0_dop
      do s=1,nstate
         
         zeig2 = maxblock+1    
         zeig1 = 1
         call cpvxz(dtpsi(zpsi(s)),ztmp(zeig2),block(s))

         do m=1,nmode

            if (gwpm(m)) then
               zeig3 = zeig1
               zeig1 = zeig2
               zeig2 = zeig3

               call qtxxzz(gs2(gs2mat(m,s)),ztmp(zeig1), &
                    ztmp(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo
         call vvaxzz(psi(zpsi(s)),ztmp(zeig2),ztmp1,block(s))
         etot = etot+ztmp1
      enddo

      ztmp=0.0_dop
      ztmp1=0.0_dop
      do s=1,nstate
         
         zeig2 = maxblock+1    
         zeig1 = 1
         call cpvxz(psi(zpsi(s)),ztmp(zeig2),block(s))

         do m=1,nmode

            if (gwpm(m)) then
               zeig3 = zeig1
               zeig1 = zeig2
               zeig2 = zeig3

               call qtxxzz(gsdot(gsmat(m,s)),ztmp(zeig1), &
                    ztmp(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo
         call vvaxzz(psi(zpsi(s)),ztmp(zeig2),ztmp1,block(s))
         etot = etot-(0,1)*ztmp1
      enddo
   !   ztmp1=0.0_dop
   !   do s=1,nstate
   !      do m=1,nmode
   !         if (gwpm(m)) then
   !            call sum2qxxz(gsdot(gsmat(m,s)),dicht2(dmat(m,s)),ztmp1,dim(m,s))
   !         endif
   !      enddo
   !   enddo
   !   etot = etot+ztmp1
      etotd = -1.0_dop*dimag(etot)


      dke = 0.0_dop
      do s=1,nstate
         do m=1,nmode
            call diagke(m,psi(zxi(m,s)),dicht2(dmat(m,s)),nspfdof(m),&
                 gwpdim(m,s),dke)
         enddo
      enddo
      print*,'DETOT',etotd*ev,estart*ev,etotd-estart,dke
   
      if (abs(dke) .lt. tiny) return

      dele = etotd - estart

      if (dke .lt. 10.0*abs(dele)) return

      if (abs(dele) .lt. 1.0d-4) then
         return
      else
         lconsten_iter = .true.
      endif

      dscale = dke - dele
      dscale = dscale / dke 
      dscale = dsqrt(dscale)

      do s=1,nstate
         do m=1,nmode
            call scalep_en(psi(zxi(m,s)),nspfdof(m),gwpdim(m,s),dscale)
         enddo
      enddo

      lha_enerr = lha_enerr + dele
      print*,'LHA_ENERR',lha_enerr*ev,dscale

      return
      end subroutine constener

!#######################################################################

      subroutine diagke(m1,xi,dicht2,gdof,gwpdim1,dke)

      implicit none

      integer(long)                                        :: e,g,f
      integer(long), intent(in)                            :: gdof,gwpdim1,m1
      complex(dop), dimension(gdof,gwpdim1), intent(in)    :: xi
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in) :: dicht2
      real(dop), intent(inout)                               :: dke
      real(dop)                                            :: mom,weight

      do e=1,gwpdim1
         weight = dble(dicht2(e,e))
         do g=1,gdof
            f = spfdof(g,m1)

            mom = dimag(xi(g,e))
            dke = dke + weight*mom*mom / (2.0_dop*mass(f))

         enddo
      enddo

      return
      end subroutine diagke

!#######################################################################

      subroutine scalep_en(xi,gdof,gwpdim1,dscale)

      implicit none

      integer(long)                                        :: e,g
      integer(long), intent(in)                            :: gdof,gwpdim1
      complex(dop), dimension(gdof,gwpdim1), intent(inout) :: xi
      real(dop), intent(in)                                :: dscale
      real(dop)                                            :: pos,mom

      do e=1,gwpdim1
         do g=1,gdof

            pos = dble(xi(g,e))
            mom = dimag(xi(g,e))
            mom = dscale*mom 
            xi(g,e) = dcmplx(pos,mom)

         enddo
      enddo

      return
      end subroutine scalep_en

!--------------------------------------------------------------------
!
! funkrreal : a wrapper for funkr that expands/compresses the 
!       complex vectors into real storage
!
! I use the TRANSFER function to convert between real and complex
! variables.  This only works if the Fortran compiler uses "sane"
! memory layout for complex arrays.  If it fails, then the transfers
! need to be replaced with real/aimag/cmplx intrinsics.  See the 
! commented out lines in runlsoda of propwf/integrat.F90 for an example.
!
!--------------------------------------------------------------------

subroutine funkrreal (n,time,psireal,dpsireal)

use psidef
use datenmod
use channels

integer,intent(in)::n
real(kind=dop),intent(in)::time
real(kind=dop),dimension(n),intent(in)::psireal
real(kind=dop),dimension(n),intent(inout)::dpsireal
complex(dop), dimension(dgldim):: psi
complex(dop), dimension(dgldim):: dtpsi
real(kind=dop),save::lasttime=-1
psi=transfer(psireal,psi)
call funkr (time,psi,dtpsi)
dpsireal=transfer(dtpsi,dpsireal)
if (lintstep) then
  if (abs(lasttime-time)>1.0e-15_dop) then
    write(istp,'(999g16.8)')time,dtpsi(9),psi(9)
    lasttime=time
  endif
endif
end subroutine funkrreal


! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE FUNKEX                        *
! *                                                                    *
! * Performs an evaluation of the right hand side of the MCTDH working *
! * equation for the MCTDH-coefficients. Used when a numerically       *
! * exact wavepacket is propagated.                                    *
! *                                                                    *
! * Called from: abm, bs, lanczos                                      *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  MCTDH-coefficient vector                                   *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine funkex (time,psi,dtpsi)

      implicit none

      integer(long)                                     :: pointer,s,s1,k, &
                                                           hktype,h,nham,&
                                                           t,t1,kt
      real(dop), intent(in)                             :: time
      complex(dop), dimension(totphidim), intent(inout) :: psi
      complex(dop), dimension(totphidim), intent(inout) :: dtpsi
      complex(dop), dimension(sdim)                     :: tmphpsi,tmphpsi1
      logical(kind=4)                                   :: luncorr=.false.
      integer(long), save                               :: tid=0 ! timer ID

! --- INITIALISE VARIABLES ---

      call get_timer(tid, 'Funkex')
      call start_timer(tid)
      call zerovxz(dtpsi,totphidim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! --- CALCULATE RIGHT HAND SIDE FOR HAMILTONIAN ---
!-----------------------------------------------------------------------
      if (dentype .eq. 0) then

         hktype=0
         do s=1,nstate
            pointer = zetf(1,s)-zetf(1,1)+1
            call hunphi1ms(psi(pointer),dtpsi(pointer), &
                        1,s,dim(1,s),subdim(1),hktype,1)
         enddo

!-----------------------------------------------------------------------
! If multi-state calculation, correlation hamiltonian couples states
!-----------------------------------------------------------------------
         do k=1,kzahl(1)
            if (autocap(k) .gt. 0) cycle
            s1=ki(k)
            pointer = zetf(1,s1)-zetf(1,1)+1
            call hphi1mk(psi(pointer),tmphpsi,1,k,dim(1,s1),subdim(1))

            s=kf(k)
            pointer = zetf(1,s)-zetf(1,1)+1
            h=hamilton(fkoe,k)
            if (htmtype(h) .eq. 1) then
               call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))
            else
               call xvxxzza(koeff(k),tmphpsi,dtpsi(pointer),subdim(1))
            endif

         enddo

!-----------------------------------------------------------------------
! --- CALCULATE RIGHT HAND SIDE FOR LIOUVILLIAN ---
!-----------------------------------------------------------------------
      else if (dentype .eq. 1) then

         hktype=0
         do t=1,nstate
            do s=1,nstate
               pointer = dzetf(1,s,t)-dzetf(1,1,1)+1
               call d1hunphi1ms(psi(pointer),tmphpsi, &
                        1,s,t,ddim(1,s,t),subdim(1),hktype,1)

               call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))

               call makeadj(tmphpsi,tmphpsi1,supgdim(1),ddim(1,s,t))
               pointer = dzetf(1,t,s)-dzetf(1,1,1)+1
               if (lrlx) then
                  call addmxxzo(tmphpsi1,dtpsi(pointer),&
                                subdim(1),ddim(1,t,s))
               else
                  call submxxzo1(dtpsi(pointer),tmphpsi1,&
                                 subdim(1),ddim(1,t,s))
               endif
            enddo
         enddo

!-----------------------------------------------------------------------
! If multi-state calculation, correlation hamiltonian couples states
!-----------------------------------------------------------------------
         do k=1,kzahl(1)
            if (autocap(k) .gt. 0) cycle

!     system operators
            if (kact(k).eq.1) then
               s1=ki(k)
               do t=1,nstate
                  pointer = dzetf(1,s1,t)-dzetf(1,1,1)+1
                  call d1hphi1mk(psi(pointer),tmphpsi,1,k,&
                       ddim(1,s1,t),subdim(1),supgdim(1),luncorr)

                  h=hamilton(fkoe,k)
                  if (htmtype(h) .ne. 1) &
                      call xvxxzzo(koeff(k),tmphpsi,subdim(1))
                  s=kf(k)
                  pointer = dzetf(1,s,t)-dzetf(1,1,1)+1
                  call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))

                  call makeadj(tmphpsi,tmphpsi1,supgdim(1),ddim(1,s,t))
                  pointer = dzetf(1,t,s)-dzetf(1,1,1)+1
                  if (lrlx) then
                     call addmxxzo(tmphpsi1,dtpsi(pointer),&
                                   subdim(1),ddim(1,t,s))
                  else
                     call submxxzo1(dtpsi(pointer),tmphpsi1,&
                                    subdim(1),ddim(1,t,s))
                  endif
               enddo

!     pure left acting operators
            else if (kact(k).eq.2) then
               s1=ki(k)
               do t=1,nstate
                  pointer = dzetf(1,s1,t)-dzetf(1,1,1)+1
                  call d1hphi1mk(psi(pointer),tmphpsi,1,k,&
                       ddim(1,s1,t),subdim(1),supgdim(1),luncorr)

                  s=kf(k)
                  pointer = dzetf(1,s,t)-dzetf(1,1,1)+1
                  h=hamilton(fkoe,k)
                  if (htmtype(h) .eq. 1) then
                     call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))
                  else
                     call xvxxzza(koeff(k),tmphpsi,dtpsi(pointer),subdim(1))
                  endif
               enddo

!     pure right acting operators
            else if (kact(k).eq.3) then
               s1=kf(k)
               do t=1,nstate
                  pointer = dzetf(1,t,s1)-dzetf(1,1,1)+1
                  call d1hphi1mk(psi(pointer),tmphpsi,1,k,&
                       ddim(1,t,s1),subdim(1),supgdim(1),luncorr)

                  s=ki(k)
                  pointer = dzetf(1,t,s)-dzetf(1,1,1)+1
                  h=hamilton(fkoe,k)
                  if (htmtype(h) .eq. 1) then
                     call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))
                  else
                     call xvxxzza(koeff(k),tmphpsi,dtpsi(pointer),subdim(1))
                  endif
               enddo

!     left-right acting operators
            else if ((kact(k).eq.4)) then
               kt = ktwin(k)
               s1 = ki(k)
               t  = kf(kt)
               pointer = dzetf(1,s1,t)-dzetf(1,1,1)+1
               call d1hphi1mk(psi(pointer),tmphpsi,1,k,&
                    ddim(1,s1,t),subdim(1),supgdim(1),luncorr)

               s  = kf(k)
               t1 = ki(kt)
               pointer = dzetf(1,s,t1)-dzetf(1,1,1)+1
               h=hamilton(fkoe,k)
               if (htmtype(h) .eq. 1) then
                  call addvxxzo(tmphpsi,dtpsi(pointer),subdim(1))
               else
                  call xvxxzza(koeff(k),tmphpsi,dtpsi(pointer),subdim(1))
               endif
            endif

         enddo
      endif

! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
! lgetenergy ensures that rlxenergy is not calculated, e.g. during SIL
! iterations

      if (lrlx .and. .not. lrlxt) then
         if (lgetenergy) call calcenergy_ex(psi,dtpsi,rlxenergy)
         if (inttypephi .eq. 3) then
            call xvxxdzr(rlxenergy,psi,dtpsi,totphidim)
         else
            call subenergy_ex(rlxenergy,psi,dtpsi)
         endif
      endif

! remove spurious small nubers in dtpsi (noise) 
      if (lrmnoise) call rm_noise(dtpsi,totphidim)

!print*,'Time ',time/fs
!do s=1,totphidim
!  print*,s,dble(psi(s)),dimag(psi(s)),dble(dtpsi(s)),dimag(dtpsi(s))
  !print*,m,dble(dtpsi(m)),dimag(dtpsi(m))
!enddo

! --- MULTIPLY TIME DERIVATIVE WITH -I (OR -1 FOR RELAXATION) ---

      call prefactor(dtpsi,totphidim)
      call stop_timer(tid)

      return
      end subroutine funkex

!#######################################################################

    subroutine funkphinode(time,psi,dtpsi)

        implicit none
        real(dbl),intent(in)        :: time
        real(dop)                   :: time1
        complex(cdbl),intent(inout) :: psi(:)
        complex(cdbl),intent(out)   :: dtpsi(:)

        time1=time   ! cp to prevent unused variable
        call compute_rhs_node(currentnode, psi, dtpsi)
        dtpsi = tree%prefactor*dtpsi

    end subroutine funkphinode

!#######################################################################

    subroutine funkml_dt(time,psi,dtpsi)
! interface from integration routine bstep to funktionml
          implicit none
          real(dbl)                 :: time  ! actual time (for TD operators)
          complex(cdbl)             :: psi(dgldim)
          complex(cdbl)             :: dtpsi(dgldim)
  
          call funk1(time,psi,dtpsi)
          call funka(time,psi,dtpsi)

    end subroutine funkml_dt


!***********************************************************************
! 
!                           subroutine ddmfunk
!
! computes the time derivative of the complete wavefunction. the
! projector is applied before the uncorrelated hamiltonian terms are
! added. This is a version of routine funk used for DD-MCTDH.
!
! called from: propagatepsi, abm, bs
! calls to:    density, hphi, phihphi, mfields, hlochphi, project,
!              dicht1phi, hunphi, addhunphi, 
!              prefactor
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine ddmfunk (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: gapdrv,hphigap,phihphigap,mfieldsgap,hlochphigap,&
                       hunphigap

      implicit none

      integer(long)                                  :: hktype,nham,i
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi

      integer(long), save                            :: tid = 0 ! timer ID

      logical(kind=4)                                :: lcheck

!-----------------------------------------------------------------------
! Check that psi is okay.
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         lcheck = lcheck .and. psi(i).eq.psi(i)
      enddo
      if(.not.lcheck)then
         routine = 'ddmfunk'
         message = 'Psi is corrupted; it has NaN entries.'
         call errormsg
      endif

!-----------------------------------------------------------------------
! Multilayer run should use Funkr
!-----------------------------------------------------------------------
      if (lrunml) then
         routine = 'DDMFunk (function.F90)'
         message = 'ML-MCTDH must use funkr - should not be here'
         call errormsg
      end if

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call get_timer(tid, 'DDMFunk')
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! Compute density and inverse density matrices
!-----------------------------------------------------------------------
      call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)

!-----------------------------------------------------------------------
! Deal with the machinery of calculating the GAP weights.
!-----------------------------------------------------------------------
!      call gapdrv(psi,mr(mrort))
!-----------------------------------------------------------------------
! Calculate components of the EOMs involving the GAP.
!-----------------------------------------------------------------------
! test mem
!      call hphigap(psi,hpsigap)
! test mem
!      call phihphigap(psi,hpsigap,hteilgap)
      call phihphigap(psi,hteilgap)
      call mfieldsgap(psi,dtpsi,usehsym)
!      call hlochphigap(psi,dtpsi,hpsigap,hlochgap,0)
      call hlochphigap(psi,dtpsi,hlochgap,0)

!-----------------------------------------------------------------------
! Loop over correlated Hamiltonian terms
! 1. compute hpsi array
! 1a. calculate matrix elements for GWP modes
! 2. compute hteil matrices
! 3. compute hloch matrices and dtpsi for A coefficients
! 4. multiply hloch and hpsi and store in dtpsi
!-----------------------------------------------------------------------
!      ktype=0
!      call hphi(psi,mc(mchpsi),ktype,1)
!      call phihphi(psi,gh,mc(mchpsi),hteil,1,usehsym)
!      call mfields(psi,dtpsi,1,psi,0,usehsym)
!      call hlochphi(psi,dtpsi,mc(mchpsi),mc(mchloch),1,0)

!-----------------------------------------------------------------------
! remaining action of hamiltonian on single-particle functions
! 1. multiply with inverse density matrix
! 2. determine and add any cap uncorrelated hamiltonian terms
! 3. apply projection operator
! 4. determine and add action of "normal" uncorrelated hamiltonian terms
!    and correct coefficient motion for natural orbitals if required
! 5. multiply with -i (or -1 for relaxation)
!-----------------------------------------------------------------------
      call dicht1phi(dtpsi,dicht1)
         
      if (lcap) then
         hktype=2
         call hunphi(psi,mc(mchunpsi),hktype,1)         
         call addhunphi(dtpsi,mc(mchunpsi))        
         call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,nham)
         call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
      endif

      call project(psi,dtpsi)

      hktype=1
      call hunphigap(psi,mc(mchunpsi),hktype,1)
!      call hunphi(psi,mc(mchunpsi),hktype,1)
      call addhunphi(dtpsi,mc(mchunpsi))
      call prefactor(dtpsi,dgldim)
      call stop_timer(tid)

!print*,'Time ',time/fs
!do i=1,adim
!  print*,i,dble(psi(i)),dimag(psi(i)),dble(dtpsi(i)),dimag(dtpsi(i))
!enddo

      return
      end subroutine ddmfunk

!***********************************************************************
!
!                         mctdh subroutine ddmfunkr
!
! calculates the derivative wrt time of the mctdh wavefunction.
! single-particle operators are on the right of the projector.
! caps can thus be included as single-particle operators as long as the
! complex cap coefficients are defined as the corresponding uncorrelated
! expansion coefficient koeffh(k_cap,m,s) to ensure hermiticity of the
! remaining real potential matrix.
!
! called from: propagatepsi, abm, bs
! calls to:    density, hphi, phihphi, mfields, hlochphi, project,
!              hunphi, phihunphi, dicht1phi, addhunphi, calchuna,
!              prefactor, subenergy
!
! input parameters:
!     time: absolute time (dummy parameter for time-independent h)
!     psi:  wavefunction vector
!
! output parameters:
!     dtpsi: time derivative of the wavefunction psi
!
! other parameters are passed through the common blocks in the include
! files.
!
! v6.0 mb
!
!***********************************************************************

      subroutine ddmfunkr (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: phihphigap,mfieldsgap,hlochphigap,phihunphigap,&
                       hunphigap,calchunagap

      implicit none

      integer(long)                                  :: hktype,ktype,nham,i
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi

!      integer(long)                                  :: zeig2

      integer(long), save                            :: tid = 0 ! timer ID

      logical(kind=4)                                :: lcheck

!-----------------------------------------------------------------------
! Check that psi is okay.
!-----------------------------------------------------------------------
      lcheck = .true.
      do i=1,dgldim
         lcheck = lcheck .and. psi(i).eq.psi(i)
      enddo
      if(.not.lcheck)then
         routine = 'ddmfunkr'
         message = 'Psi is corrupted; it has NaN entries.'
         call errormsg
      endif

!-----------------------------------------------------------------------
! initialise arrays
!-----------------------------------------------------------------------
      call get_timer(tid, 'DDMFunkr')
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

!      zeig2=0

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

!-----------------------------------------------------------------------
! compute density and inverse density matrices
!-----------------------------------------------------------------------
      call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)

!-----------------------------------------------------------------------
! loop over correlated hamiltonian terms
! 1. compute hpsi array
! 2. compute hteil matrices
! 3. compute hloch matrices and dtpsi for a coefficients
!-----------------------------------------------------------------------
      ktype=0
      call phihphigap(psi,hteilgap)
      hktype=0
      call hunphigap(psi,mc(mchunpsi),hktype,1)
      call phihunphigap(psi,mc(mchunpsi),hteilgap,nham)

!-----------------------------------------------------------------------
! Calculate mean fields
!-----------------------------------------------------------------------
      call mfieldsgap(psi,dtpsi,usehsym)

!-----------------------------------------------------------------------
! 4. multiply hloch and hpsi and store in dtpsi
! remaining action of hamiltonian on single-particle functions
! 1. multiply with inverse density matrix
! 2. determine action of uncorrelated hamiltonian terms
! 3. add uncorrelated motion to time derivative
! 4. compute matrix elements <phi|h_uncorr|phi>
! 5. apply projection operator
!-----------------------------------------------------------------------
      call hlochphigap(psi,dtpsi,hlochgap,0)
      call dicht1phi(dtpsi,dicht1)
      call addhunphi(dtpsi,mc(mchunpsi))
      call project(psi,dtpsi)

!-----------------------------------------------------------------------
! 1. compute action of uncorrelated hamiltonian on a-vector
!    note that the action of the correlated hamiltonian on the a-vector
!    has already been computed (as a by-product) in "mfields".
! 2. multiply with -i (or -1 for relaxation). for relaxation, transform
!    h|psi> to (h-e)|psi> before.
!-----------------------------------------------------------------------
      call calchunagap(psi,dtpsi,hteilgap,1)

! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
      if (lrlx .and. .not. lrlxt) then
         call calcenergy(psi,dtpsi,rlxenergy,0)
         if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
            call xvxxdzr(rlxenergy,psi,dtpsi,adim)
         else
            call subenergy(rlxenergy,psi,dtpsi)           
         endif
      endif

      call prefactor(dtpsi,dgldim)
      call stop_timer(tid)

      return
      end subroutine ddmfunkr

! **********************************************************************
! *                                                                    *
! *                          subroutine ddmfunk0                       *
! *                                                                    *
! * calculates the mean-field and inverse density matrices and the     *
! * time derivative of the mctdh-coefficients without computing the    *
! * time derivative of the single-particle functions.                  *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, phihphi, hunphi, phihunphi                      *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   dtpsi: time derivative of the mctdh-coefficients                 *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.2 mb                                                            *
! *                                                                    *
! * if modus = 1, for ML-MCTDH the time-derivative of the A-vector     *
! * is NOT calculated (for ada2)                                       *
! **********************************************************************

      subroutine ddmfunk0 (time,psi,dtpsi,modus)

use gapmod
use gapfuncsmod, only: hphigap,phihphigap,mfieldsgap,summfgap

      implicit none

      integer(long)                                  :: ktype,nham
      real(dop), intent(in)                          :: time
      integer(long), intent(in)                      :: modus
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi
      integer(long)                                  :: z0,zf

      integer(long), save                            :: tid = 0 ! timer ID

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      call get_timer(tid, 'DDMFunk0')
      call start_timer(tid)

      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0d0,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- compute density and inverse density matrices ---

      call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)

! --- compute action of correlated hamiltonian terms on spf's ---
! --- determine correlated matrix elements <phi|h|phi> ---

      ktype=0
!      call hphigap(psi,hpsigap)
!      call phihphigap(psi,hpsigap,hteilgap)
      call phihphigap(psi,hteilgap)
      if (lrunml) call compute_hterms_treer(time,tree,psi,1)

! --- determine mean-field matrices and time derivative of a-vector ---
! --- compute sum over mean-field times single-particle operator ---
      call mfieldsgap(psi,dtpsi,usehsym)
      call summfgap(mfsum,hlochgap,1)

! ML-MCTDH needs extra call to calculate dtA
      if (lrunml .and. modus .eq. 1) then
         tz => tree%swap_down(1)%p
         z0 = tz%zr(1,1)
         zf = tz%zr(1,2)
         call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))
      endif

      call stop_timer(tid)

      return
      end subroutine ddmfunk0   

! **********************************************************************
! *                                                                    *
! *                          subroutine ddmfunk1                       *
! *                                                                    *
! * computes the time derivative of the single-particle functions      *
! * without recomputing the density and mean-field matrices.           *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, hunphi, hlochphi, mfsumphi, dicht1phi,          *
! *              addhunphi, project, prefactor                         *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   dtpsi: time derivative of the wavefunction psi                   *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.2 mb                                                            *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunk1 (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: hphigap,hlochphigap

      implicit none

      integer(long)                                :: hktype, ktype
      real(dop), intent(in)                        :: time
      real(dop)                                    :: time1
      complex(dop), dimension(dgldim), intent(inout):: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi

      integer(long), save                          :: tid = 0 ! timer ID

      time1=time   ! cp to prevent unused variable

! --- initialise variables --- 

      call get_timer(tid, 'DDMFunk1')
      call start_timer(tid)
      call zerovxz(dtpsi(adim+1),totphidim)

! --- compute action of correlated hamiltonian terms on spf's ---
      ktype=1
!      call hphigap(psi,hpsigap)

!    --- compute action of uncorrelated hamiltonian terms on spf's ---

      hktype=0
      call hunphi(psi,mc(mchunpsi),hktype,1)

! --- multiply h|phi> with mean-field matrix (kinetic terms) ---

!      call hlochphigap(psi,dtpsi,hpsigap,hlochgap,1)
      call hlochphigap(psi,dtpsi,hlochgap,1)

! --- multiply |phi> with sum over mean-fields (potential terms) ---

      call mfsumphi(dtpsi,psi,mfsum)

! --- multiply with inverse of density matrix ---

      call dicht1phi(dtpsi,dicht1)

!    --- add uncorrelated motion to time derivative vector ---
      
      call addhunphi(dtpsi,mc(mchunpsi))

!    --- apply projection operator on time derivative vector ---

      call project(psi,dtpsi)

      if (lrunml) call compute_rhs_allphi(tree,psi,dtpsi)

! --- multiply time derivative with -i (or -1 for relaxation) ---

      call prefactor(dtpsi(adim+1),totphidim)
      call stop_timer(tid)

      return
      end  subroutine ddmfunk1   

! **********************************************************************
! *                                                                    *
! *                          subroutine ddmfunk2                       *
! *                                                                    *
! * computes - after having propagated the single-particle functions - *
! * the matrix elements needed for the propagation of the              *
! * a-coefficients.                                                    *
! * This is a version of routine funk2 used for DD-MCTDH.              *
! *                                                                    *
! * called from: propagatepsi                                          *
! * calls to:    hphi, phihphi, hunphi, phihunphi                      *
! *                                                                    *
! * input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent h)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * output parameters:                                                 *
! *   none                                                             *
! *                                                                    *
! * other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * v6.0 mb                                                            *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunk2 (time,psi)

use gapmod
use gapfuncsmod, only: gapdrv,hphigap,phihphigap

      implicit none

      integer(long)                                  :: hktype,nham
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi

      integer(long), save                            :: tid = 0 ! timer ID

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      call get_timer(tid, 'DDMFunk2')
      call start_timer(tid)

      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- compute action of correlated hamiltonian terms on spf's ---
! --- compute action of uncorrelated hamiltonian terms on spf's ---
! --- determine correlated matrix elements <phi|h|phi> ---
! --- determine uncorrelated matrix elements <phi|h_uncorr|phi> ---

!-----------------------------------------------------------------------
! Deal with the machinery of calculating the GAP weights.
!-----------------------------------------------------------------------
!      call gapdrv(psi,mr(mrort))
!-----------------------------------------------------------------------
! Calculate components of the EOMs involving the GAP.
!-----------------------------------------------------------------------
!      call hphigap(psi,hpsigap)
!      call phihphigap(psi,hpsigap,hteilgap)
      call phihphigap(psi,hteilgap)

!-----------------------------------------------------------------------
! Calculate components of the EOMs involving the KE operators.
!-----------------------------------------------------------------------
      hktype=0
      call hunphi(psi,mc(mchunpsi),hktype,1)
      call phihunphi(psi,mc(mchunpsi),hung,gh,hteil,1)
      if (lrunml) call compute_hterms_treer(time,tree,psi,1)

      call stop_timer(tid)

      return
      end subroutine ddmfunk2 

! **********************************************************************
! *                                                                    *
! *                          SUBROUTINE DDMFUNK3                       *
! *                                                                    *
! * Updates the mean fields, etc. after the A-coefficients have been   *
! * propagated. Note that the matrix elements must have already been   *
! * computed in "funk2". Note that the "intpsi" array is (mis)used as  *
! * complex work array!                                                *
! * Funk3 computes the time derivative dtpsi for both A and phi.       *
! * This is a version of routine funk3 used for DD-MCTDH.              *
! *                                                                    *
! * Called from: propagatepsi                                          *
! * Calls to:    density, mfields, summf, summfphi, hlochphi, project, *
! *              hunphi, phihunphi, dicht1phi, addhunphi, calchuna     *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  wavefunction vector                                        *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: action of hamiltonian on the wavefunction psi             *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunk3 (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: mfieldsgap,hlochphigap,summfgap

      implicit none

      real(dop), intent(in)                          :: time
      real(dop)                                      :: time1
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi

      integer(long), save                            :: tid = 0 ! timer ID

      time1=time   ! cp to prevent unused variable

! --- initialise variables --- 

      call get_timer(tid, 'DDMFunk3')
      call start_timer(tid)
      call zerovxz(dtpsi,dgldim)

! --- compute density and inverse density matrices ---
! --- determine mean-field matrices and time derivative of a-vector ---
! --- compute sum over mean-field times single-particle operator ---
      call density(psi,tree,dicht1,dicht2,dicht3,dicht4,mi(mijindx),gs2)

      call mfieldsgap(psi,dtpsi,usehsym)
      call summfgap(mfsum,hlochgap,1)

! --- multiply h|phi> with mean-field matrix (kinetic terms) ---
! --- multiply |phi> with sum over mean-fields (potential terms) ---
! --- multiply with inverse of density matrix ---

!      call hlochphigap(psi,dtpsi,hpsigap,hlochgap,1)
      call hlochphigap(psi,dtpsi,hlochgap,1)
      call mfsumphi(dtpsi,psi,mfsum)
      call dicht1phi(dtpsi,dicht1)

!    --- add uncorrelated motion to time derivative vector ---    
      call addhunphi(dtpsi,mc(mchunpsi))

!    --- apply projection operator on time derivative vector ---

      call project(psi,dtpsi)

      if (lrunml) then
          call compute_rhs_treer(tree,psi,dtpsi)

      else
! --- compute action of uncorrelated hamiltonian on a-vector ---
         call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
      endif

! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
      if (lrlx .and. rlaxnum .lt. 0 .and. .not. lrlxt) then
         call calcenergy(psi,dtpsi,rlxenergy,0)
         if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
            call xvxxdzr(rlxenergy,psi,dtpsi,adim)
         else
            call subenergy(rlxenergy,psi,dtpsi)
         endif
      endif

! --- MULTIPLY TIME DERIVATIVE WITH -I (OR -1 FOR RELAXATION) ---
! --- NO MULTIPLICATION (i.e. prefactor=1) FOR A-VECTOR IF DAVIDSON ---
 
      if(inttypea.eq.6) then
         call prefactor(dtpsi(adim+1),totphidim)
      else
         call prefactor(dtpsi,dgldim)
      endif

      call stop_timer(tid)

      return
      end subroutine ddmfunk3   

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE DDMFUNKA                      *
! *                                                                    *
! * Performs an evaluation of the right hand side of the MCTDH working *
! * equation for the MCTDH-coefficients. Used when only the            *
! * A-coefficients are propagated.                                     *
! *                                                                    *
! * Called from: abm, bs, lanczos, davstep, sil                        *
! * Calls to:    calcha, calchuna, prefactor, subenergy                *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  MCTDH-coefficient vector                                   *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunka (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: calchagap

      implicit none


      real(dop), intent(in)                        :: time
      real(dop)                                    :: time1
      complex(dop), dimension(adim), intent(in)    :: psi
      complex(dop), dimension(adim), intent(inout)   :: dtpsi
      integer(long)                                :: z0,zf

      integer(long), save                          :: tid = 0 ! timer ID

      time1=time   ! cp to prevent unused variable

! --- initialise variables --- 

      call get_timer(tid, 'DDMFunka')
      call start_timer(tid)
      call zerovxz(dtpsi,adim)

! --- calculate right hand side for hamiltonian ---

      if (psitype .eq. 5) then
         tz => tree%swap_down(1)%p
         z0 = tz%zr(1,1)
         zf = tz%zr(1,2)
         call compute_rhs_node(tz,psi(z0:zf),dtpsi(z0:zf))
      else if (dentype .eq. 0) then
         call calchagap(psi,dtpsi,hteilgap,1)
      endif

#ifdef MPI
      if (mpirank.eq.0) then
#endif

! --- compute action of uncorrelated hamiltonian ---
      if (psitype .ne. 5) then
         call calchuna(psi,dtpsi,hteil,1,mi(mijindx))
      endif

! --- subtract energy from hamiltonian if relaxation ---
!     note that sil-integration of the a-matrix is assumed for dentype=2
      if (lrlx .and. lrlxt) then
         if (inttypea.eq.3 .or. inttypea.eq.6) then  ! SIL or Davidson
            call xvxxdzr(rlxenergy,psi,dtpsi,adim)
         else
            call subenergy(rlxenergy,psi,dtpsi)
         endif
      endif

! --- multiply time derivative with -i (or -1 for relaxation) ---

      if(inttypea.ne.6) call prefactor(dtpsi,adim)
#ifdef MPI
      endif
#endif
      call stop_timer(tid)

      return
      end subroutine ddmfunka

! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE DDMFUNKPHI                      *
! *                                                                    *
! * Computes the time derivative of the single-particle functions of   *
! * one mode with mean-field and inverse density matrices held         *
! * constant. The projector is applied after the uncorrelated          *
! * Hamiltonian terms have been added.                                 *
! *                                                                    *
! * Called from: abm, bs, RK5, RK8                                     *
! * Calls to:    hphi1m, hlochphi1m, project1m, dicht1phi1m, hunphi1m, *
! *              addhunphi1m, prefactor, summfphi1m                    *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  single-particle function vector of one mode                *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files. In particular, the actual mode is "currentmode".    *
! * This variable is in a common block in runprop.inc.                 *
! *                                                                    *
! * V6.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunkphi (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: hphigap1m,hlochphi1mgap

      implicit none

      integer(long)                                   :: s,pointer,pointer1, &
                                                         hktype,ktype,f
      real(dop), intent(in)                           :: time
      real(dop)                                       :: time1
      complex(dop), dimension(maxphidim), intent(inout):: psi
      complex(dop), dimension(maxphidim), intent(inout) :: dtpsi

      integer(long)                                   :: tid ! timer ID
                                                             ! do not save
                                                             ! as name changes
      character(len=18)                               :: string

      time1=time   ! cp to prevent unused variable

! --- initialise variables --- 
      tid=0
      if (nspfdof(currentmode) .eq. 1) then
         f=spfdof(1,currentmode)
         string='DDMFunk-'//modelabel(f)(1:10)
      else if (nspfdof(currentmode) .gt. 1) then
         write(string,'(a11,i3)') 'DDMFunkphi-',currentmode
      endif
      call get_timer(tid, string)
      call start_timer(tid)

      call zerovxz(dtpsi,phidim(currentmode))

! --- calculate action of correlated hamiltonian on spf's ---

      ktype=1
!      call hphigap1m(psi,hpsigap,currentmode, &
!                  zetf(currentmode,1),phidim(currentmode))

! --- multiply h|phi> with (constant) mean-field (kinetic terms) ---

      call hlochphi1mgap(psi,dtpsi,hlochgap,currentmode, &
           zetf(currentmode,1),phidim(currentmode),1)

! --- multiply |phi> with sum over mean-fields (potential terms) ---

      call mfsumphi1m(dtpsi,psi,mfsum,currentmode, &
                      zetf(currentmode,1),phidim(currentmode))

! --- loop over each state ---

      do s = 1,nstate
         pointer = zetf(currentmode,s)-zetf(currentmode,1)+1
         pointer1 = zetf(currentmode,s)-adim

!    --- multiply with (constant) inverse density matrix ---

         call dicht1phi1ms(dtpsi(pointer),dicht1, &
                           currentmode,s,dim(currentmode,s), &
                           subdim(currentmode))

!    --- SELECT ORDER OF OPERATORS: Proj-h OR h-Proj
!    --- calculate action of uncorrelated hamiltonian on spf's ---

         hktype=0
         call hunphi1ms(psi(pointer),mc(mchunpsi+pointer1-1), &
              currentmode,s,dim(currentmode,s), &
              subdim(currentmode),hktype,1)

!       --- add uncorrelated motion to time derivative vector ---

         call addhunphi1ms(dtpsi(pointer),mc(mchunpsi), &
              currentmode,s, &
              dim(currentmode,s),subdim(currentmode))

!       --- apply projection operator on time derivative vector ---

         call project1ms(psi(pointer),dtpsi(pointer), &
              currentmode,dim(currentmode,s),subdim(currentmode))

      enddo

! --- multiply time derivative with -i (or -1 for relaxation) ---

      call prefactor(dtpsi,phidim(currentmode))
      call stop_timer(tid)

      return
      end subroutine ddmfunkphi

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE DDMFUNKEX                     *
! *                                                                    *
! * Performs an evaluation of the right hand side of the standard      *
! * method working equation for the coefficients. Used when a          *
! * numerically exact wavepacket is propagated using the direct 
! * dynamics version of the code.                                      *
! *                                                                    *
! * Called from: abm, bs, lanczos                                      *
! *                                                                    *
! * Input parameters:                                                  *
! *   time: absolute time (dummy parameter for time-independent H)     *
! *   psi:  MCTDH-coefficient vector                                   *
! *                                                                    *
! * Output parameters:                                                 *
! *   dtpsi: time derivative of the psi                                *
! *                                                                    *
! * Other parameters are passed through the common blocks in the       *
! * include files.                                                     *
! *                                                                    *
! * Modified from funkex (V6.0 MB) GWR 10/16                           *
! *                                                                    *
! **********************************************************************

      subroutine ddmfunkex (time,psi,dtpsi)

use gapmod
use gapfuncsmod, only: hunphigap1ms,hphigap1mk

      implicit none

      integer(long)                                     :: pointi,s,s1,k, &
                                                           hktype,h,nham,&
                                                           pointf
      integer(long)                                     :: kdum=1
      real(dop), intent(in)                             :: time
      complex(dop), dimension(totphidim), intent(inout) :: psi
      complex(dop), dimension(totphidim), intent(inout)   :: dtpsi
      complex(dop), dimension(totphidim)                :: tmphpsi
      integer(long), save                               :: tid=0 ! timer ID

! --- INITIALISE VARIABLES ---

      call get_timer(tid, 'DDMFunkex')
      call start_timer(tid)
      call zerovxz(dtpsi,totphidim)

!-----------------------------------------------------------------------
! calculate time-dependent coefficients and points
!-----------------------------------------------------------------------
      nham=1
      if (ltdh(nham)) call calctdk(time,nham,psi)
! no driving field change during relaxation
      if (ltdp) then
         if (lrlx) then
            call calctdp(0.0_dop,psi,mr(mrort))
         else
            call calctdp(time,psi,mr(mrort))
         endif
      endif

! --- CALCULATE RIGHT HAND SIDE FOR HAMILTONIAN ---

       do s=1,nstate
          pointi = zetf(1,s)-zetf(1,1)+1
          hktype=0
          call hunphigap1ms(psi(pointi),dtpsi(pointi), &
                      1,s,dim(1,s),subdim(1),hktype,1)
       enddo

!-----------------------------------------------------------------------
! If multi-state calculation, correlated hamiltonian couples states
!-----------------------------------------------------------------------
      do k=1,kzahl(1)
         s1=ki(k)
         s=kf(k)
         pointi = zetf(1,s1)-zetf(1,1)+1
         pointf = zetf(1,s)-zetf(1,1)+1
         if(lddmopt .or. lddgbtendec)then
            call ogparrew(ogpar)
            do while (associated(ogpar%next))
               if(ogpar%k.eq.k)then
                  call hphigap1mk(psi(pointi),tmphpsi,1,k,kdum,ogpar%kidx,&
                       dim(1,s1),subdim(1))
                  h=hamilton(fkoe,k)
                  if (htmtype(h) .eq. 1) then
                     call addvxxzo(tmphpsi,dtpsi(pointf),subdim(1))
                  else
                     call xvxxzza(koeff(k),tmphpsi,dtpsi(pointf),subdim(1))
                  endif
               endif
               ogpar => ogpar%next
            enddo
         else
               call hphigap1mk(psi(pointi),tmphpsi,1,k,1,kdum,dim(1,s1),&
                    subdim(1))
               h=hamilton(fkoe,k)
               if (htmtype(h) .eq. 1) then
                  call addvxxzo(tmphpsi,dtpsi(pointf),subdim(1))
               else
                  call xvxxzza(koeff(k),tmphpsi,dtpsi(pointf),subdim(1))
               endif
         endif
      enddo

! --- SUBTRACT ENERGY FROM HAMILTONIAN IF RELAXATION ---
! lgetenergy ensures that rlxenergy is not calculated, e.g. during SIL 
! iterations
      if (lrlx .and. .not. lrlxt) then
         if (lgetenergy) then
            call calcenergy_ex(psi,dtpsi,rlxenergy)
         endif
         if (inttypephi .eq. 3) then
            call xvxxdzr(rlxenergy,psi,dtpsi,totphidim)
         else
            call subenergy_ex(rlxenergy,psi,dtpsi)
         endif
      endif

!print*,'Time ',time/fs
!do s=1,totphidim
!  print*,s,dble(psi(s)),dimag(psi(s)),dble(dtpsi(s)),dimag(dtpsi(s))
!  print*,s,dble(dtpsi(s)),dimag(dtpsi(s))
!enddo

! --- MULTIPLY TIME DERIVATIVE WITH -I (OR -1 FOR RELAXATION) ---

      call prefactor(dtpsi,totphidim)
      call stop_timer(tid)

      return
      end subroutine ddmfunkex

!#######################################################################

      subroutine xdot(psi,dtpsi,dicht1,hloch,pcons)

      implicit none

      integer(long)                                  :: s,m,f
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(dgldim), intent(inout)   :: dtpsi
      complex(dop), dimension(dmatdim), intent(in)   :: dicht1
      complex(dop), dimension(hlochdim), intent(in)  :: hloch
      complex(dop), dimension(pconsdim), intent(inout) :: pcons

      do s=1,nstate
         do m=1,nmode
            f=spfdof(1,m)
            if (basis(f) .eq. 26 .or. basis(f) .eq. 27) then
               call xdot1m(psi(zetf(m,s)),dtpsi(zetf(m,s)),subdim(m), &
                    dim(m,s),f,m,s,dicht1(dmat(m,s)),hloch,pcons(zpcons(f)))
            endif
            if (basis(f) .eq. 26) then
               call qtxxzzr(pcons(zpcons(f)),psi(zpsi(1)), &
                    dtpsi(zpsi(s)),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo
      enddo

      return
      end subroutine xdot

!#######################################################################

      subroutine xdot1m(psi,dtpsi,subdim1,dim1,f,m,s,dicht1,hloch,pcons)

      implicit none

      real(dop), parameter      :: para1=1.0_dop,para2=1.0_dop,para3=0.02_dop

      integer(long)             :: nham,h,k,k1,k2,kun,e,kke,hl,nh,e1,idum
      integer(long), intent(in) :: subdim1,dim1,f,m,s

      real(dop)                                          :: dq,d1,coeff,v,mom,gaa,diff,&
                                                            width,x1,mindiff
      real(dop), dimension(maxdim)                       :: rdum
      real(dop), dimension(maxdim)                       :: q0,gpoint
      complex(dop)                                       :: pab2,vz
      complex(dop), dimension(subdim1,dim1), intent(in)  :: psi
      complex(dop), dimension(subdim1,dim1), intent(inout) :: dtpsi
      complex(dop), dimension(hlochdim), intent(in)      :: hloch
      complex(dop), dimension(dim1,dim1), intent(in)     :: dicht1
      complex(dop), dimension(dim1,dim1)                 :: workc
      complex(dop), dimension(dim1,dim1), intent(inout)    :: pcons

! find uncorrelated potential and KE terms
      nham=1
      k1=zhun(m,s,nham)
      k2=zhun(m,s,nham)+khzahl(m,s,nham)-1
      kun=0
      kke=0
      do k=k1,k2
         h=hamilton(f,k)
         if (htmtype(h) .eq. 101) kun=k
         if (htmtype(h) .eq. 102) kke=k
      enddo

! xdot
      coeff=1.0_dop
      if (kke .gt. 0) then
         coeff=dble(koeff(kke))
         h=hamilton(f,kke)
         if (htmfac(h) .gt. 1) then
            do nh=1,htmfac(h)
               hl=hterm(nh,h)
               if (hopfile(hl) .eq. 301) then
                  coeff=coeff*hoprpar(1,hl)
               endif
            enddo
         endif
      
         do e=1,dim1
            mom=dble(psi(2,e))
            width=dimag(psi(2,e))
            dtpsi(1,e)=coeff*mom/mass(f)
            dtpsi(3,e)=coeff*(mom*mom)/(2.0_dop*mass(f))
         enddo
      endif

! uncorrelated force
      if (kun .gt. 0) then
         h=hamilton(f,kun)
         dq=fddx
         do e=1,dim1
            q0(f)=dble(psi(1,e))
            call deriv1x(d1,h,q0,dq,f)
            dtpsi(2,e)=-d1
         enddo

         do e=1,dim1-1
            do e1=e+1,dim1
!           e1=e+1
            diff=dble(psi(1,e1)-psi(1,e))
!           diff=diff**(-7)
!           diff=0.000001*6*diff
!           dtpsi(2,e)=dtpsi(2,e)-diff
            diff=dabs(diff)
            diff=diff/para3
            diff=para2*(1.0_dop-diff)
            diff=(para1/para3)*dexp(diff)
!           dtpsi(2,e)=dtpsi(2,e)-diff
            enddo
         enddo
         do e=2,dim1
            do e1=1,e-1
!           e1=e-1
            diff=dble(psi(1,e1)-psi(1,e))
!           diff=diff**(-7)
!           diff=0.000001*6*diff
!           dtpsi(2,e)=dtpsi(2,e)-diff
            diff=dabs(diff)
            diff=diff/para3
            diff=para2*(1.0_dop-diff)
            diff=(para1/para3)*dexp(diff)
!           dtpsi(2,e)=dtpsi(2,e)+diff
            enddo
         enddo

      endif

! now correlated terms
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      do k=k1,k2
         if (.not. diag(m,k)) then
            call qqxxzz(dicht1,hloch(phloch(m,k)),workc,dim1)

            h=hamilton(f,k)
            do e=1,dim1
               q0(f)=dble(psi(1,e))
               call deriv1x(d1,h,q0,dq,f)
               dtpsi(2,e)=dtpsi(2,e)-dble(workc(e,e))*d1
            enddo
         endif
      enddo
!
! calculate momentum constraint term
!
      if (basis(f) .eq. 26) then
!        call xlocmom(psi,workc,subdim(m),dim1)
!        if (abs(coeff - 1.0_dop) .lt. tiny) then
!           call xqxxdzo(coeff,workc,dim1)
!        endif

         mindiff=9999.0_dop
         do e=1,dim1
            pcons(e,e)=0.0_dop
            do e1=e+1,dim1
               diff=dble(psi(1,e)-psi(1,e1))
               diff=mass(f)*diff
               mindiff=min(mindiff,abs(diff))
               pcons(e1,e)=(0,1)*coeff*workc(e1,e)/diff 
            enddo

            if (e .lt. dim1) then
            do e1=e+1,dim1
!              e1=e+1
               diff=dble(psi(1,e1)-psi(1,e))
!              diff=diff**(-6)
!              pcons(e,e)=pcons(e,e)+0.000001*diff
            diff=dabs(diff)
            diff=diff/para3
            diff=para2*(1.0_dop-diff)
            diff=para1*dexp(diff)
!           pcons(e,e)=pcons(e,e)+diff
            enddo
            endif
            if (e .gt. 1) then
            do e1=1,e-1
!              e1=e-1
               diff=dble(psi(1,e1)-psi(1,e))
!              diff=diff**(-6)
!              pcons(e,e)=pcons(e,e)+0.000001*diff
            diff=dabs(diff)
            diff=diff/para3
            diff=para2*(1.0_dop-diff)
            diff=para1*dexp(diff)
!           pcons(e,e)=pcons(e,e)+diff
            enddo
            endif

         enddo

         do e=2,dim1
            do e1=1,e-1
               pcons(e1,e)=dconjg(pcons(e,e1))
            enddo
         enddo

         do e=1,dim1
            gaa=0.0_dop
            x1=dble(psi(1,e))
            do e1=1,dim1
               if (e1 .ne. e) then
                  pab2=workc(e,e1)*pcons(e1,e)
                  gaa=gaa+dimag(pab2)
               endif
            enddo
            gaa=2.0_dop*gaa
            dtpsi(2,e)=dtpsi(2,e)+gaa
         enddo
      endif

!
! take potential from KE to form phase derivative
!
      if (kun .gt. 0) then
         h=hamilton(f,kun)
         do e=1,dim1
            gpoint(f)=dble(psi(1,e))
            call poteval (h,v,gpoint,vz,rdum,idum)
            dtpsi(3,e)=dtpsi(3,e)-v
            x1=dble(psi(1,e))
            mom=dble(dtpsi(2,e))
!           dtpsi(3,e)=0.0

         enddo
      endif

      return
      end subroutine xdot1m

!#######################################################################

      subroutine deriv1x(dq1,h,q0,dq,f)

      implicit none

      integer(long)                            :: idum
      integer(long), intent(in)                :: h,f
      real(dop)                                :: vp,vm,qp,qm,diff
      real(dop), intent(in)                    :: dq
      real(dop), intent(inout)                   :: dq1
      real(dop), dimension(maxdim)             :: rdum
      real(dop), dimension(maxdim)             :: gpoint
      real(dop), dimension(maxdim), intent(in) :: q0
      complex(dop)                             :: vz

      qp  = q0(f) + dq/2.0_dop
      qm  = q0(f) - dq/2.0_dop

      gpoint(f)=qp
      call poteval (h,vp,gpoint,vz,rdum,idum)
      gpoint(f)=qm
      call poteval (h,vm,gpoint,vz,rdum,idum)

      diff = vp - vm

      dq1 = diff/dq

      return
      end subroutine deriv1x

!#######################################################################
!
! Remove any number < eps from a complex vector
!
!#######################################################################

      subroutine rm_noise(vec,dim1)

      implicit none

      integer(long), intent(in)                    :: dim1
      integer(long)                                :: i
      real(dop)                                    :: dvec,zvec
      complex(dop), dimension(dim1), intent(inout) :: vec(dim1)

      do i=1,dim1
         dvec = dble(vec(i))
         zvec = dimag(vec(i))
         if (dabs(dvec) .lt. 100.0*macheps) dvec = 0.0_dop
         if (dabs(zvec) .lt. 100.0*macheps) zvec = 0.0_dop
         vec(i) = dcmplx(dvec,zvec)
      enddo


      return
      end subroutine rm_noise
   
!#######################################################################
!
!               WKXDOT
!
! Here the workspace required for the module is assigned
!
! workcdim, workrdim, workidim and workldim are passed via a common
! block in global.inc, and are checked that they are large enough for
! the module.
!
!#######################################################################

      subroutine wkfunction

      implicit none

!-----------------------------------------------------------------------
! assign workspace used in module
!-----------------------------------------------------------------------
      if (lxloc) then
         workcdim=max(maxspf*maxspf,workcdim)
      else if (lgwp .and. lfunkl) then
! for gwp4dot1m
         workcdim=max(3*maxspf*maxspf+maxspf*maxspf*maxdim*maxdim, &
                   workcdim)
      else if (lgwp) then
! for funka
         if (lconst) then
            workcdim = workcdim+adim
            workcdim = max(workcdim,maxspf*maxspf)
         endif
      else
         workcdim=workcdim
      endif
      if (lgwp .and. lfunkl) then
         workrdim=max(workrdim,maxdim*maxspf)
      else
         workrdim=workrdim
      endif
      workidim=workidim
      workldim=workldim

      return
      end subroutine wkfunction

      end module function
