! **********************************************************************
      module hlochphimod

use decimal, only: dop,long
use global
use globmemmod
use timing
use logdat
use mmlib,   only: mmxxzza, mmxtzza, mmxtzz
use op1lib,  only: zerovxl,zeromxz,cpvxz,zerovxz
use op2lib,  only: addmxxzo 

use maxv, only: maxkoe
use dvrdatmod
use griddatmod
use operdef
use psidef
use memhopsmodmod
use hpsimod
use openmpmod
use runpropmod
use hphimod, only: hphi1mk

use mpidata
#ifdef MPI
use mpi
use mpiutilsmod, only: mpe_decomp1d
#endif
use mpidata

      implicit none
      private
      public :: hlochphi,hlochphi1m,d2hlochphi1m
#ifdef MPI
      public:: hlochphi1mpar, mpihlochphi
#endif
      contains

! **********************************************************************
!                                                                       
!                           HLOCHPHI
!                                                                       
!   MCTDH module containing (auxiliary) subroutines to evaluate the     
!   right hand side of the MCTDH working equations.                     
!                                                                       
!   Contains:                                                           
!     hlochphi:     Computes for all modes, states, and all correlated  
!                   Hamiltonian terms the product of the mean-field     
!                   matrices and h|phi>.                                
!     hlochphi1m:   Computes for one mode and all correlated            
!                   Hamiltonian terms the product of the mean-field     
!                   matrices and h|phi>.                                
!   d2hlochphi1m:   Computes for one mode and all correlated           
!                   Hamiltonian terms the product of the mean-field    
!                   matrices and h|phi>.                               
!                                                                       
! **********************************************************************

! *** V6.0 MB *************> SUBROUTINE HLOCHPHI <**********************
!                           ~~~~~~~~~~~~~~~~~~~~~

! Computes for all modes, all states and all Hamiltonian terms the
! product of the mean-field matrices and h|phi>.

      subroutine hlochphi (psi,dtpsi,hpsi,hloch,nham,ktype)

      implicit none

      integer(long)                                  :: m
      integer(long), save                            :: tid=0
      integer(long), intent(in)                      :: nham,ktype
      complex(dop), dimension(hlochdim), intent(in)  :: hloch
      complex(dop), dimension(hpsidim), intent(inout):: hpsi
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(dgldim), intent(out)   :: dtpsi
      complex(dop)                                   :: ztmp

      call get_timer(tid,'Hlochphi')
      call start_timer(tid)

! dummy cp for unused variable
      ztmp = psi(1)

! --- LOOP OVER EACH MODE AND STATE---

      if(dentype.eq.0)then
          if ((lmpi .and. lmpihlochphi .and. lmpistarted)) then
#ifdef MPI
! --- Use parallel hlochphi1m routine.
             do m = 1,nmode
                call hlochphi1mpar(psi(zetf(m,1)),dtpsi(zetf(m,1)),hpsi,&
                     hloch,m,nham,ktype,zetf(m,1),phidim(m))
             enddo
#else
             ztmp = psi(1) ! dummy cp to prevent unused variable in serial code

             routine="Hlochphi"
             message="Wants MPI code but not complied MPI"
             call errormsg
#endif
          else
            do m = 1,nmode
! --- Use standard hlochphi1m routine.
               call hlochphi1m(dtpsi(zetf(m,1)),hpsi,hloch,m,zetf(m,1), &
                    phidim(m),nham,ktype)
            enddo
          endif
      else if(dentype.eq.2)then
! --- Use standard hlochphi1m routine.
         do m = 1,nmode
            call d2hlochphi1m(dtpsi(zetf(m,1)),hpsi,hloch,m,zetf(m,1), &
                 phidim(m),nham)
         enddo
      endif

      call stop_timer(tid)

      return
      end subroutine hlochphi

! *** V6.0 MB ************> SUBROUTINE HLOCHPHI 1M <*******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and all Hamiltonian terms the product of the
! mean-field matrices and h|phi>.
!
! ktype governs which terms are included:
! ktype = 0 : all terms
! ktype = 1 : Potential terms are excluded (i.e. only kinop terms
!             calculated)
! nham      : Number of operator to be used


      subroutine hlochphi1m (dtpsi,hpsi,hloch,m,zetf1,phidim1,nham,ktype)

!$ use omp_lib

      implicit none

      integer(long)                                               :: k,k1,k2, &
                                                                     s1,s,f
      integer(long), intent(in)                                   :: nham,m, &
                                                                     zetf1, &
                                                                     phidim1, &
                                                                     ktype
      complex(dop), dimension(hlochdim), intent(in)               :: hloch
      complex(dop), dimension(hpsidim), intent(in)                :: hpsi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(out) :: dtpsi
      logical(kind=4), dimension(maxkoe)                          :: kcalc

! --- IF BASIS IS COMPLETE DO NOT CALCULATE ---

      if (complete(m)) return

! --- IF GWP or X-local mode, do not calculate

      f = spfdof(1,m)
      if(basis(f).eq.25 .or. basis(f) .eq. 26 .or. basis(f) .eq. 27) &
          return

!-----------------------------------------------------------------------
! first find which terms need to be calculated
!-----------------------------------------------------------------------
      call zerovxl(kcalc,ktot)
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      if (ktype .eq. 0) then
         do k=k1,k2
            if (.not. diag(m,k)) kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
         enddo
      else if (ktype .eq. 1) then
         do k=k1,k2
            if (kinop(m,k) == 1) kcalc(k)=.true.
            if (autocap(k) > 0) kcalc(k)=.false.
         enddo
      endif

! --- COMPUTE MEAN-FIELD TIMES H|PHI> FOR ALL STATES AND TERMS ---

!      if(.not.lompthread.or.(k2-k1+1).le.30)then
         do k = k1,k2
            if (kcalc(k)) then
               s=kf(k)
               s1=ki(k)
               call mmxtzza(hpsi(zhpsi(m,k)),hloch(phloch(m,k)), &
                    dtpsi(zetf(m,s)),subdim(m),dim(m,s1),dim(m,s))
            endif
         enddo
!      else
!         do k = k1,k2
!            if (kcalc(k)) then
!               s=kf(k)
!               s1=ki(k)
!               call mmxtzzaomp(hpsi(zhpsi(m,k)),hloch(phloch(m,k)), &
!                    dtpsi(zetf(m,s)),subdim(m),dim(m,s1),dim(m,s),ompthread)
!            endif
!         enddo
!      endif

      k=zham(nham)+kzahl(nham)-1
      do k1=1,nmulpot(nham)
         k=k+1
         if (.not. diag(m,k)) then
            s=kf(k)
            call addmxxzo(hloch(phloch(m,k)),dtpsi(zetf(m,s)), &
                    subdim(m),dim(m,s))
         endif
      enddo

      return
      end subroutine hlochphi1m

! *** V6.0 MB ************> SUBROUTINE D2HLOCHPHI1M <******************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and all Hamiltonian terms the product of the
! mean-field matrices and h|phi>.

      subroutine d2hlochphi1m (dtpsi,hpsi,hloch,m,zetf1,phidim1,nham)

     implicit none

      integer(long)                                               :: k,k1,k2,kt,sf,s,si,f
      integer(long), intent(in)                                   :: nham,m,zetf1,phidim1
      complex(dop), dimension(hlochdim), intent(in)               :: hloch
      complex(dop), dimension(hpsidim), intent(in)                :: hpsi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(out) :: dtpsi
      logical(kind=4)                                             :: lcalc
      logical(kind=4), dimension(maxkoe)                          :: kcalc

! --- IF BASIS IS COMPLETE DO NOT CALCULATE ---

      if (complete(m)) return

! --- IF GWP or X-local mode, do not calculate

      f = spfdof(1,m)
      if(basis(f).eq.25 .or. basis(f) .eq. 26 .or. basis(f) .eq. 27) &
          return

! --- COMPUTE MEAN-FIELD TIMES H|PHI> FOR ALL STATES AND TERMS ---
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      do k=k1,k2
         if (.not. diag(m,k)) kcalc(k)=.true.
         if (autocap(k) .gt. 0) kcalc(k)=.false.
      enddo

      do k = k1,k2
         if (.not. kcalc(k)) cycle

!     kact = 1
         if ((kact(k).eq.1) .and. (.not. diag(m,k))) then
            if (.not.lconst .or. kinop(m,k).eq.1) then
               sf=kf(k)
               si=ki(k)
               call mmxtzza(hpsi(zhpsi(m,k)),hloch(phloch(m,k)), &
                    dtpsi(zetf(m,sf)),subdim(m),dim(m,si),dim(m,sf))
            endif
         endif
!     kact = 2
         if ((kact(k).eq.2) .and. (k.lt.ktwin(k))) then
            kt = ktwin(k)
            if ((.not.diag(m,k)) .or. (.not.diag(m,kt))) then
               if ((.not.lconst)       .or.&
                   (kinop(m,k).eq.1)   .or.&
                   (kinop(m,kt).eq.1)) then
                  sf=kf(k)
                  si=ki(kt)
                  call mmxtzza(hpsi(zhpsi2(m,k)),hloch(phloch(m,k)),&
                       dtpsi(zetf(m,sf)),subdim(m),dim(m,si),&
                       dim(m,sf))
               endif
            endif
         endif
!     kact = 4
         if ((kact(k).eq.4) .and. (k.lt.ktwin(k))) then
            if (.not. diag(m,k)) then
               if (.not.lconst .or. kinop(m,k).eq.1) then
                  kt = ktwin(k)
                  lcalc = .false.
                  if (lmf) then
                     if (ki(kt) .eq. kf(k)) lcalc = .true.
                  else
                     lcalc = .true.
                  endif
                  if (lcalc) then
                     sf=kf(k)
                     si=ki(k)
                     call mmxtzza(hpsi(zhpsi(m,k)),hloch(phloch(m,k)),&
                          dtpsi(zetf(m,sf)),subdim(m),dim(m,si),&
                          dim(m,sf))
                  endif
               endif
            endif
         endif
      enddo

      k=zham(nham)+kzahl(nham)-1
      do k1=1,nmulpot(nham)
!         message = 'Mulpot is not yet implemented.'
!         call errormsg
         k=k+1
         if (.not. diag(m,k)) then
            s=kf(k)
            call addmxxzo(hloch(phloch(m,k)),dtpsi(zetf(m,s)),&
                    subdim(m),dim(m,s))
         endif
      enddo

      return
      end subroutine d2hlochphi1m

! *** V6.0 MB ************> SUBROUTINE HLOCHPHI 1M PAR<****************
!                          ~~~~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and all Hamiltonian terms the product of the
! mean-field matrices and h|phi> in parallel.
!
! ktype governs which terms are included:
! ktype = 0 : all terms
! ktype = 1 : Potential terms are excluded (i.e. only kinop terms
!             calculated)
! nham      : Number of operator to be used

#ifdef MPI
      subroutine hlochphi1mpar(psi,dtpsi,hpsi,hloch,m,nham,ktype,zetf1,phidim1)

      implicit none

      integer(long)                               :: k1,k2,f
      integer(long), intent(in)                   :: nham,m,zetf1,phidim1,ktype
      complex(dop), dimension(hlochdim), intent(in)               :: hloch
      complex(dop), dimension(hpsidim), intent(inout)             :: hpsi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(in)  :: psi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(out) :: dtpsi

! --- IF BASIS IS COMPLETE DO NOT CALCULATE ---

      if (complete(m)) return

! --- IF GWP or X-local mode, do not calculate

      f = spfdof(1,m)
      if(basis(f).eq.25 .or. basis(f) .eq. 26 .or. basis(f) .eq. 27) &
          return

      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

      if(lmpi.and.lmpistarted.and.lmpihlochphi) then
         mpitask=9
! --- Determine Job
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(lmpiupdt,nmpiupdtlen,MPI_LOGICAL,master, &
                        MPI_COMM_WORLD,mpierr)
         if (lmpiupdt(2)) then
            call MPI_Bcast(phidim1,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(zetf1,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(m,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(nham,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(ktype,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
            lmpiupdt(2) = lmpiupdt(1)
         endif
         call mpihlochphi(psi(zetf1),dtpsi(zetf1),hpsi,hloch,m,nham,ktype, &
                          zetf1,phidim1)
      endif

      return
      end subroutine hlochphi1mpar
#endif

!#######################################################################
!
!   mpihlochphi
!-----------------------------------------------------------------------

#ifdef MPI
      subroutine mpihlochphi(psi,dtpsi,hpsi,hloch,m,nham,ktype, &
                             zetf1,phidim1)

      implicit none

      integer(long)                                   :: i1,i2,k,k1,k2,k3,s,s1
      integer(long), intent(in)                       :: nham,m,zetf1, &
                                                         phidim1,ktype
      complex(dop), dimension(hlochdim), intent(in)               :: hloch
      complex(dop), dimension(hpsidim), intent(out)               :: hpsi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(in)  :: psi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(out) :: dtpsi
      complex(dop), dimension(phidim1)                            :: tmpdtpsi

! --- Get Data
      call MPI_Bcast(psi(zetf1),phidim1,MPI_DOUBLE_COMPLEX,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(hloch,hlochdim,MPI_DOUBLE_COMPLEX,0,MPI_COMM_WORLD,mpierr)

      i1=zham(nham)
      i2=zham(nham)+kzahl(nham)-1
      if (mpirank.gt.0) call zerovxz(dtpsi(zetf1),phidim1)

! Divide up the work.
      call mpe_decomp1d(i1,i2,k1,k2)

      do k = k1,k2
         s =kf(k)
         s1=ki(k)
         if (ktype .eq. 0) then
            if (.not. diag(m,k)) then
               if (autocap(k) .le. 0) then
                  call hphi1mk(psi(zetf(m,s1)),hpsi,m,k,dim(m,s1),subdim(m))
                  call mmxtzza(hpsi,hloch(phloch(m,k)), &
                       dtpsi(zetf(m,s)),subdim(m),dim(m,s1),dim(m,s))
               endif
            endif
         else if (ktype .eq. 1) then
            if (kinop(m,k) .eq. 1) then
               if (autocap(k) .le. 0) then
                  call hphi1mk(psi(zetf(m,s1)),hpsi,m,k,dim(m,s1),subdim(m))
                  call mmxtzza(hpsi,hloch(phloch(m,k)), &
                       dtpsi(zetf(m,s)),subdim(m),dim(m,s1),dim(m,s))
               endif
            endif
         endif
      enddo

      k=zham(nham)+kzahl(nham)-1
      do k3=zham(nham)+kzahl(nham)+mpirank,k+nmulpot(nham),nmpi
         if (.not. diag(m,k3)) then
            s=kf(k3)
            call addmxxzo(hloch(phloch(m,k3)),dtpsi(zetf(m,s)), &
                 subdim(m),dim(m,s))
         endif
      enddo

      call MPI_Reduce(dtpsi(zetf1),tmpdtpsi,phidim1, &
                   MPI_DOUBLE_COMPLEX,MPI_SUM,master,MPI_COMM_WORLD,mpierr)
      if(mpirank.eq.master) call cpvxz(tmpdtpsi,dtpsi(zetf1),phidim1)

      return
      end subroutine mpihlochphi
#endif

      end module hlochphimod


