!***********************************************************************
      module importhomod

use global
use globmemmod, only: agmat
use decimal, only: long,dop
use globmemmod
use op1lib, only: overmxz
use timing

use runpropmod
use psidef
use griddatmod

use gwplib, only: normgwp,renormdgwp1


      implicit none
      private
      public :: importho,importho1,importhovmf,dummypost,&
                rhonrm,dummyerrfn
      contains

!***********************************************************************
!
! SUBROUTINE IMPORTHO
!
! This is the post-processing functions called from the RK integrators
! (maybe others in the future, too) during CMF propagations.
! Calls IMPORTHO1 (see below) for all states in one mode.
! 
!***********************************************************************
!
! INPUT:
!     psi: where the SPFs for the current mode are stored
!     currentmode (global): specifies the mode
!
! OUTPUT:
!     psi: new SPFs with improved orthonormality
!     
! 10/2002 FO
!
!***********************************************************************

      subroutine importho(psi)

use global
use runpropmod
use psidef
use griddatmod

use gwplib, only: normgwp,renormdgwp1

      implicit none

      integer(long)                             :: s,t,dim1,pointer
      complex(dop), dimension(*), intent(inout) :: psi
      integer(long), save                             :: tid=0 ! timer ID

!     start timer
      call get_timer(tid, 'Importho')
      call start_timer(tid)

!     loop over all states
      if (gwpm(currentmode)) then
         do s=1,nstate
            if (abs(psitype) .ne. 4) then
               pointer = zdgwp(currentmode,s)-zetf(currentmode,1)+1
               dim1 = dim(currentmode,s)
               call renormdgwp1(psi(pointer),gs2(gs2mat(currentmode,s)),&
                    dicht4(dmat(currentmode,s)),dim1)
            endif

            call normgwp(nspfdof(currentmode),dim1,&
                 agmat(zagmat(currentmode,s)), &
                 agmat(zxivec(currentmode,s)),agmat(zphase(currentmode,s)))
         enddo
      else if (dentype .eq. 1) then
         do s=1,nstate
            do t=1,nstate
               pointer = dzetf(currentmode,s,t)-dzetf(currentmode,1,1)+1
               dim1 = ddim(currentmode,s,t)
               call importho1(subdim(currentmode), dim1, psi(pointer))
            enddo
         enddo
      else
         do s=1,nstate
            pointer = zetf(currentmode,s)-zetf(currentmode,1)+1
            dim1 = dim(currentmode,s)
            call importho1(subdim(currentmode), dim1, psi(pointer))
         enddo
      endif

!     stop timer
      call stop_timer(tid)

      end subroutine importho ! of importho


!***********************************************************************
!
! SUBROUTINE IMPORTHOVMF
!
! Post-processing function called from the RK integratoes (maybe others
! in the future, too) during VMF propagations.
! Calls IMPORTHO1 (see below) for all modes and states.
!
!***********************************************************************
!
! INPUT:
!     psi: complete MCTDH wavefunction
! OUTPUT:
!     psi: MCTDH wavefunction, with SPF orthonormality improved
!
!***********************************************************************

      subroutine importhovmf(psi)

      implicit none

      integer(long)                             :: m,s,t,dim1
      complex(dop), dimension(*), intent(inout) :: psi
      integer(long), save                       :: tid=0 ! timer ID

!     start timer
      call get_timer(tid, 'Importho')
      call start_timer(tid)

      if (dentype .eq. 1) then
!     loop over all modes ...
         do m=1,nmode
!        ... and states
            do s=1,nstate
               do t=1,nstate
                  dim1 = ddim(m,s,t)
                  call importho1(subdim(m), dim1, psi(dzetf(m,s,t)))
               enddo
            enddo
         enddo
      else
!     loop over all modes ...
         do m=1,nmode
!        ... and states
            do s=1,nstate
               if (gwpm(m)) then
                  if (abs(psitype) .ne. 4) then
                     call renormdgwp1(psi(zdgwp(m,s)),gs2(gs2mat(m,s)),&
                          dicht4(dmat(m,s)),dim(m,s))
                  endif
                  call normgwp(nspfdof(m), gwpdim(m,s), agmat(zagmat(m,s)), &
                       agmat(zxivec(m,s)),agmat(zphase(m,s)))
               else
                  dim1 = dim(m,s)
                  call importho1(subdim(m), dim1, psi(zetf(m,s)))
               endif
            enddo
         enddo
      endif

!     stop timer
      call stop_timer(tid)

      end subroutine importhovmf ! of importhovmf


!***********************************************************************
!
! SUBROUTINE IMPORTHO1
!
! Improves the orthonormality of a set of vectors
! (dim vectors of dimension subdim).
! The linear space spanned by these vectors stays the same.
!
! This is done by calculating
!    phi~ = phi * (1 + .5 * H), where
!    H    = 1 - adj(phi) * phi
!
! (Ref: SIAM J. Num. Anal. 7 (1970), 386-389)
!
!***********************************************************************
!
! INPUT:
!     phi: set of vectors, stored as [subdim x dim] matrix
!     work: workspace, dimension [dim x dim]
!     tmp: more workspace, dimension [dim]
!
! OUTPUT:
!     phi: set of vectors with improved orthonormality
!
! 10/2002 FO
!
!***********************************************************************

      subroutine importho1(subdim,dim,phi)

      implicit none

      integer(long)                                      :: i,j,k
      integer(long), intent(in)                          :: subdim,dim
      complex(dop), dimension(dim)                       :: tmp
      complex(dop), dimension(dim,dim)                   :: work
      complex(dop), dimension(subdim,dim), intent(inout) :: phi

!     calculate the Gram matrix (adj(phi)*phi)
      call overmxz(phi,work,subdim,dim)

!     calculate 3/2 - 1/2 Gram
!     first the diagonal elements
      do i=1,dim
         work(i,i) = 1.5_dop - 0.5_dop*work(i,i)
      enddo
!     now the off-diagonal elements
      do i=1,dim-1
         do j=i+1,dim
            work(i,j) = -(0.5_dop*work(i,j))
            work(j,i) = dconjg(work(i,j))
         enddo
      enddo

!     now apply this matrix to the phi's
      do i=1,subdim
         do j=1,dim
            tmp(j) = phi(i,1)*work(1,j)
            do k=2,dim
               tmp(j) = tmp(j) + phi(i,k)*work(k,j)
            enddo
         enddo
!        now phi(i,*) is no longer needed
         do j=1,dim
            phi(i,j) = tmp(j)
         enddo
      enddo

      end subroutine importho1 ! of importho1


!***********************************************************************
!
! SUBROUTINE DUMMYPOST
!
! A dummy post-processing subroutine that does nothing 
!
! 10/2002 FO
!***********************************************************************

      subroutine dummypost(psi)

      implicit none

      complex(dop), dimension(*), intent(in)  :: psi
      complex(dop)                            :: cdum

!     do nothing
      cdum=psi(1)
      return

      end subroutine dummypost


!***********************************************************************
!
! RHONRM
! computes the rho-weighted L2-norm of a vector
!
!***********************************************************************
!
! INPUT:
! ------
! ndim    INTEGER     dimension of the vectors
! vec     COMPLEX*16  vector, whose norm is to be computed
! rho     COMPLEX*16  MCTDH density matrix (dicht2)
! subdim, dim, dmat, zetf: usual MCTDHpointers (see phelp)
! mxdm = maxdim
!
! OUTPUT :
! --------
! norm  REAL*8     function value
!
! Functions  RHONRM2, RHONRM
!
!***********************************************************************

      SUBROUTINE rhonrm(norm,vec)

      IMPLICIT NONE

      REAL(dop), intent(out)   :: norm
      COMPLEX(dop), intent(in) :: vec(*)
      INTEGER(long) :: m,m1,s,zeig

      m = currentmode
      if(m.gt.0) then ! CMF, one specific mode.
         norm = 0.d0
         do s=1,nstate
            zeig=zetf(m,s)-zetf(m,1)+1
            norm = norm + &
                   rhonrm2(dim(m,s),subdim(m),vec(zeig),dicht2(dmat(m,s)))
         enddo
         norm = sqrt(norm/dble(nstate))*0.5d0
      else
!........VMF
         norm = 0.d0
         do m1 = 1, adim
            norm = norm + abs(vec(m1))**2
         enddo
         do m1=1,nmode
            if(m1.eq.meb  .or.  m1.eq.mpb) cycle
            do s=1,nstate
               norm = norm + rhonrm2(dim(m1,s),subdim(m1), &
                                 vec(zetf(m1,s)),dicht2(dmat(m1,s)))
            enddo
         enddo
         norm = sqrt(norm/dble(nmode*nstate+1))*0.5d0
      endif
!.....The 0.5 is added to adjust the error-measure similar to wgtnrm
!.....nmode*nstate+1 is the number of rhonrm2 calls + 1 for the A-vector.

      RETURN
      END SUBROUTINE RHONRM 

!***********************************************************************

      REAL(dop) FUNCTION rhonrm3(dim,subdim,vec,rho)

      IMPLICIT NONE

      INTEGER(long) ::  i,j,k,dim,subdim
      COMPLEX(dop)  ::  vec(subdim,dim),rho(dim,dim),tmp

      tmp = 0.d0
      DO j = 1, dim
         DO k = 1, dim
            DO i = 1, subdim
               tmp = tmp + dconjg(vec(i,j))*rho(j,k)*vec(i,k)
            ENDDO
         ENDDO
      ENDDO
      rhonrm3 = SQRT(abs(tmp))

      RETURN
      END FUNCTION RHONRM3

!***********************************************************************
!!!  RHONRM3 and RHONRM2 are practically identical, but RHONRM2 makes use of the
!!!  hermicity of rho and of the overlap matrix. In addition it allows to add
!!!  a constant to the diagonal.
!***********************************************************************

      REAL(dop) FUNCTION rhonrm2(dim,subdim,vec,rho)

      IMPLICIT NONE

      INTEGER(long) :: i,j,k,dim,subdim
      COMPLEX(dop)  :: vec(subdim,dim),rho(dim,dim),ovl
      REAL(dop)     :: tmp

      tmp = 0.d0
      DO j = 2, dim
         DO k = 1, j-1
            ovl = 0.d0
            DO i = 1, subdim
               ovl = ovl + dconjg(vec(i,j))*vec(i,k)
            ENDDO
            tmp = tmp + 2.d0*dble(rho(j,k)*ovl)
         ENDDO
      ENDDO

      DO j = 1, dim
         ovl = 0.d0
         DO i = 1, subdim
            ovl = ovl + dconjg(vec(i,j))*vec(i,j)
         ENDDO
         tmp = tmp +  (dble(rho(j,j))+1.d-8)*dble(ovl)
      ENDDO

      rhonrm2 = tmp

      RETURN
      END FUNCTION RHONRM2


!***********************************************************************
!
! SUBROUTINE DUMMYERRFN
!
! A dummy error function that does nothing 
!
! 10/2002 FO
!***********************************************************************

      subroutine dummyerrfn(err,psi)

      implicit none

      real(dop)                               :: err
      complex(dop), dimension(*), intent(in)  :: psi
      complex(dop)                            :: cdum

!     do nothing
      cdum=psi(1)
      err=0.0_dop
      return

      end subroutine dummyerrfn

      end module importhomod

