! **********************************************************************
! *                                                                    *
! *                        INTEGRATION (integrat.f90)                  *
! *                                                                    *
! * MCTDH module containing all those routines needed for the          *
! * integration that are MCTDH specific and thus not stored in the     *
! * various integrator library modules.                                *
! *                                                                    *
! * Contains:                                                          *
! *   runabm:       Drives the ABM integration routine.                *
! *   runbs:        Drives the BS integration routine.                 *
! *   runsil:       Drives the SIL integration routine.                *
! *   rundav:       Drives the DAV 'integrators'.                      *
! *   runrk5:       Drives the RK5 integration routine.                *
! *   runrk8:       Drives the RK8 integration routine.                *
! *   runzvode:     Drives the ZVODE integration routine.              *
! *   runsod:       Drives the SOD integration routine.                *
! *   runspo:       Drives the SPO integration routine.                *
! *   runcheb:      Drives the Chebyshev integration routine.          *
! *   renorma:      Renormalises the A-vector (in relaxation           *
! *                 calculations).                       
! *   runrk5f:      Drives the RK5 fixed step integration routine      *
! * For the routines evaltype takes a value according to what is being 
! * propagated and which function routine:
! * evaltype = 0 All (funkl)
! *            1 All (funkr)
! *            4 SPFs (funkphi or funkex)
! *            5 A-coefficients (funka)
! *            6 ML-MCTDH (propphiml)
! *            7 ML-MCTDH mid layer SPFs (funkphinode)
! *           -1 Optimising GWPs (gwpoptimmod)
! **********************************************************************

      module integrat

use decimal, only: dop, long
use maxv, only: maxdim
use global
use dglobal
use globmemmod
use logdat
use xvlib,   only: vvaxzz
use mmlib,   only: mmaxzzh, mmaxzz
use mtlib,   only: qtxxzz, mtxxzz
use op1lib,  only: zerovxd, cpvxz,normvxz
use timing

use dvrdatmod
use griddatmod
use operdef
use psidef
use datenmod
use runpropmod
use channels
use inputmlmod
use wrintegrat

use importhomod, only: importhovmf,importho,dummypost,&
                       rhonrm,dummyerrfn
use gwpdepdynmod, only: gwpodetest,gwpodetest1m
use BEL, only : resetGwp
use gwplib

use function
use abmlib
use bslib
use rklib
use csillib
use sillib
use sodlib
use spolib
use cheblib
use davlib
use lsodalib
use eulerlib
use odesim
#ifdef MPI
use mpi
use davlibpar
use hpsimod
use mpidata
#endif
use psiutil
use evaluate, only: prefactor

      implicit none
      private
      public :: runabm,runsil,rundav,runbs,runrk5,runrk8,runlsoda,&
                runzvode,zvodeout,gwperror1,wferror,runsod,runspo,runcheb,&
                runverlet,runrk5f,runeuler
save
real(dop),dimension(:),allocatable::LSODArw
integer,dimension(:),allocatable::LSODAiw
integer::liw,lrw

      contains

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNABM                        *
! *                                                                    *
! * Drives the ABM integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Lenght of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   restartabm:    Restart flag; if false, a previous integration is *
! *                  continued, otherwise a new integration is         *
! *                  started.                                          *
! *   steps:         Overall number of integration steps made so far.  *
! *   repsteps:      Number of failed and repeated steps made so far.  *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *   steps:         Overall number of integration steps made so far.  *
! *   repsteps:      Number of failed and repeated steps made so far.  *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runabm (psi,dtpsi,intperiod,time,restartabm,steps, &
                         repsteps,evaltype,mode)

      implicit none

      real(dop), parameter                             :: abmaccucorr = 1.0_dop

      logical(kind=4), intent(in)                           :: restartabm
      integer(long)                                         :: error,z0,zf
      integer(long), intent(in)                             :: evaltype,mode
      integer(long), intent(out)                            :: repsteps,steps
      real(dop)                                             :: abmaccuracy
      real(dop), intent(in)                                 :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout)        :: psi,dtpsi
      integer(long), save                             :: tid=0 ! timer ID
      character(len=c2) :: string


! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1) .or. &
              (evaltype .eq. 6)) then
            write(istp,'(a)') &
               'Complete equation: Adams-Bashforth-Moulton'
         elseif (evaltype .eq. 4 .or. evaltype .eq. 7) then
            if (lexact) then
               write(istp,'(a)') &
                    'Exact wavepacket: Adams-Bashforth-Moulton'
            else
               write(istp,'(a,i2,a)') 'Mode', mode, &
                    ': Adams-Bashforth-Moulton'
            endif
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: Adams-Bashforth-Moulton'
         else
            goto 10
         endif
         if (restartabm) write(istp,'(a)') 'Integrator initialized'
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write (istp,'(a7,9x,a6,6x,a13,11x,a5,11x,a9)') &
            'Step','Order','Stepsize [fs]','Error',trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .le. 1) then
         abmaccuracy = abmaccucorr*accuracy
      elseif (evaltype .eq. 4) then
         abmaccuracy = abmaccucorr*accphi
      elseif (evaltype .eq. 5) then
         abmaccuracy = abmaccucorr*accavec
      elseif (evaltype .eq. 6) then
         abmaccuracy = abmaccucorr*accphi
      else
         goto 10
      endif

! --- CALL ADAMS-BASHFORTH-MOULTON INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'ABM')
      call start_timer(tid)

      if (evaltype .eq. 0) then
         if (lwferror) then
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt,&
                     abmaccuracy,restartabm,steps,repsteps,error,&
                     mc(mcintpsi),funk,wferror)
         else if (lddmctdh) then
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt,&
                     abmaccuracy,restartabm,steps,repsteps,error,&
                     mc(mcintpsi),ddmfunk,absabmerror)
         else
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt,&
                     abmaccuracy,restartabm,steps,repsteps,error,&
                     mc(mcintpsi),funk,absabmerror)
         endif
      elseif (evaltype .eq. 1) then
         if (lwferror) then
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt, &
                     abmaccuracy,restartabm,steps,repsteps,error, &
                     mc(mcintpsi),funk,wferror)
         else if(lddmctdh) then
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt, &
                     abmaccuracy,restartabm,steps,repsteps,error, &
                     mc(mcintpsi),ddmfunkr,absabmerror)
         else
            call abm(psi,dtpsi,dgldim,intperiod,time,intord,dt, &
                     abmaccuracy,restartabm,steps,repsteps,error, &
                     mc(mcintpsi),funkr,absabmerror)
         endif
      elseif (evaltype .eq. 4) then
         if (lexact) then
            if(lddmctdh)then
               call abm(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                    intperiod,time,intorderphi,dtphi,abmaccuracy, &
                    restartabm,steps,repsteps,error,mc(mcintpsi),ddmfunkex, &
                    absabmerror)
            else
               call abm(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                    intperiod,time,intorderphi,dtphi,abmaccuracy, &
                    restartabm,steps,repsteps,error,mc(mcintpsi),funkex, &
                    absabmerror)
            endif
         else if (lddmctdh) then
            call abm(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                  intperiod,time,intorderphi,dtphi,abmaccuracy, &
                  restartabm,steps,repsteps,error,mc(mcintpsi),ddmfunkphi, &
                  absabmerror)
         else
            call abm(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                  intperiod,time,intorderphi,dtphi,abmaccuracy, &
                  restartabm,steps,repsteps,error,mc(mcintpsi),funkphi, &
                  absabmerror)
         endif
      elseif (evaltype .eq. 5) then
         if (lddmctdh) then
            call abm(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                 intordera,dta,abmaccuracy,restartabm,steps,repsteps, &
                 error,mc(mcintpsi),ddmfunka,absabmerror)
         else
            call abm(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                 intordera,dta,abmaccuracy,restartabm,steps,repsteps, &
                 error,mc(mcintpsi),funka,absabmerror)
         endif
      elseif (evaltype .eq. 6) then
         call abm(psi,dtpsi,dgldim,intperiod,time, &
              intordera,dtphi,abmaccuracy,restartabm,steps,repsteps, &
              error,mc(mcintpsi),funkml_dt,absabmerror)
      elseif (evaltype .eq. 7) then
! ML-MCTDH prop of one SPF from middle layers
         z0 = currentnode%z(1,1)
         zf = currentnode%z(currentnode%navecs,2)
         call abm(psi(z0),dtpsi(z0),zf-z0+1,intperiod,time, &
              intordera,dtphi,abmaccuracy,restartabm,steps,repsteps, &
              error,mc(mcintpsi),funkphinode,absabmerror)
      else
         goto 10
      endif
      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'abm'
         call abmerrormsg(error,message)
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunABM'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runabm

! **********************************************************************
! *                                                                    *
! *                            SUBROUTINE RUNBS                        *
! *                                                                    *
! * Drives the Bulirsch-Stoer integration routine.                     *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   dtpsi:      Time derivative of the initial-value vector.         *
! *   intperiod:  Lenght of time interval to be integrated.            *
! *   abstime:    Absolute time, i. e. Psi_initial=Psi(AbsTime).       *
! *   restartbs:  Restart flag; if false, the previous stepsize is     *
! *               used as initial stepsize, otherwise IntPeriod is     *
! *               taken.                                               *
! *   goodsteps:  Number of succesful large integration steps made so  *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed large integration steps   *
! *               made so far.                                         *
! *   smallsteps: Number of small integration steps made so far.       *
! *   evaltype:   Specifies the evaluation type, i. e. the "funk"      *
! *               routine.                                             *
! *   mode:       The mode to be integrated when only single-particle  *
! *               functions are propagated.                            *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *   dtpsi:      Dtpsi is undetermined.                               *
! *   goodsteps:  Number of succesful large integration steps made so  *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed large integration steps   *
! *               made so far.                                         *
! *   smallsteps: Number of small integration steps made so far.       *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:         Complex*16 data needed in the Func (and possibly     *
! *               CalcError) routine.                                  *
! *   mr:         Real*8 data needed in the Func (and possibly         *
! *               CalcError) routine.                                  *
! *   mi:         Integer data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *   ml:         Logical data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runbs (psi,dtpsi,intperiod,abstime,restartbs,goodsteps, &
                        badsteps,smallsteps,evaltype,mode)

      implicit none

      real(dop), parameter                          :: bsaccucorr = 1.0_dop, &
                                                       enlargefactor = 1.7_dop

      logical(kind=4), intent(in)                   :: restartbs
      integer(long)                                 :: error,z0,zf
      integer(long), intent(in)                     :: evaltype,mode
      integer(long), intent(inout)                  :: goodsteps,badsteps, &
                                                       smallsteps
      real(dop)                                     :: bsaccuracy,time,&
                                                       largestep,nextstep, &
                                                       actualstep
      real(dop), save                               :: lastperiod
      real(dop), dimension(:), allocatable, save    :: laststep
      real(dop), intent(in)                         :: abstime,intperiod
      complex(dop), dimension(dgldim), intent(inout):: psi,dtpsi
      integer(long), save                           :: tid=0 ! timer ID
      character(len=c2) :: string
   
      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1) .or. &
             (evaltype .eq. 6)) then
            write(istp,'(a)')&
               'Complete equation: Bulirsch-Stoer'
         elseif (evaltype .eq. 4 .or. evaltype .eq. 7) then
            if (lexact) then
               write(istp,'(a)') 'Exact wavepacket: Bulirsch-Stoer'
            else
               write(istp,'(a,i2,a)') 'Mode', mode, ': Bulirsch-Stoer'
            endif
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: Bulirsch-Stoer'
         else
            goto 10
         endif
         if (restartbs) write(istp,'(a)') 'Integrator initialized'
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write(istp,'(a11,2x,a11,4x,a13,11x,a5,11x,a9)') 'Iteration',&
                    'Small steps', 'Stepsize [fs]', 'Error',trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if ((evaltype .eq. 0) .or. (evaltype .eq. 1)) then
         bsaccuracy = bsaccucorr*accuracy
      elseif (evaltype .eq. 4) then
         bsaccuracy = bsaccucorr*accphi
      elseif (evaltype .eq. 5) then
         bsaccuracy = bsaccucorr*accavec
      elseif (evaltype .eq. 6) then
         bsaccuracy = bsaccucorr*accphi
      else
         goto 10
      endif

! --- INITIALIZE VARIABLES ---

      time = abstime
      if (restartbs) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1)) then
            largestep = min(dt,intperiod)
         elseif (evaltype .eq. 4) then
            largestep = min(dtphi,intperiod)
         elseif (evaltype .eq. 5) then
            largestep = min(dta,intperiod)
         elseif (evaltype .eq. 6) then
            largestep = min(dtphi,intperiod)
         else
            goto 10
         endif
         if (mode .ne. 0) laststep(mode)=0.0
         lastperiod=0.0
      else
         if (abs(laststep(mode)-lastperiod) .lt. tiny*lastperiod) then
            largestep = intperiod
         else
            largestep = min(laststep(mode),intperiod)
         endif
      endif

! --- INTEGRATION LOOP ---

 100  continue

! --- PERFORM ONE BULIRSCH-STOER INTEGRATION STEP ---

      call get_timer(tid, 'BS')
      call start_timer(tid)

      if (evaltype .eq. 0) then
         if(lddmctdh)then
            call bsstep(psi,dtpsi,dgldim,intperiod,time,intord, &
                        largestep,bsaccuracy,actualstep,nextstep, &
                        smallsteps,error,mc(mcintpsi),ddmfunk,absbserror, &
                        polyextrapol)
         else
            call bsstep(psi,dtpsi,dgldim,intperiod,time,intord, &
                        largestep,bsaccuracy,actualstep,nextstep, &
                        smallsteps,error,mc(mcintpsi),funk,absbserror, &
                        polyextrapol)
         endif
      elseif (evaltype .eq. 1) then
         call bsstep(psi,dtpsi,dgldim,intperiod,time,intord,largestep, &
                     bsaccuracy,actualstep,nextstep,smallsteps,error, &
                     mc(mcintpsi),funkr,absbserror,polyextrapol)
      elseif (evaltype .eq. 4) then
         if (lexact) then
            if(lddmctdh)then
               call bsstep(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                           intperiod,time,intorderphi,largestep,bsaccuracy, &
                           actualstep,nextstep,smallsteps,error,mc(mcintpsi), &
                           ddmfunkex,absbserror,polyextrapol)
            else
               call bsstep(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                           intperiod,time,intorderphi,largestep,bsaccuracy, &
                           actualstep,nextstep,smallsteps,error,mc(mcintpsi), &
                           funkex,absbserror,polyextrapol)
            endif
         else if (lddmctdh) then
            call bsstep(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                        intperiod,time,intorderphi,largestep,bsaccuracy, &
                        actualstep,nextstep,smallsteps,error,mc(mcintpsi), &
                        ddmfunkphi,absbserror,polyextrapol)
         else
            call bsstep(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                        intperiod,time,intorderphi,largestep,bsaccuracy, &
                        actualstep,nextstep,smallsteps,error,mc(mcintpsi), &
                        funkphi,absbserror,polyextrapol)
         endif
      elseif (evaltype .eq. 5) then
         if(lddmctdh) then
            call bsstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                 intordera,largestep,bsaccuracy,actualstep,nextstep, &
                 smallsteps,error,mc(mcintpsi),ddmfunka,absbserror, &
                 polyextrapol)
         else
            call bsstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                 intordera,largestep,bsaccuracy,actualstep,nextstep, &
                 smallsteps,error,mc(mcintpsi),funka,absbserror, &
                 polyextrapol)
         endif
      elseif (evaltype .eq. 6) then
!     Multi-Layer dtpsi
         call bsstep(psi, dtpsi, dgldim, intperiod, time, intorderphi, &
              largestep, bsaccuracy, actualstep, nextstep, smallsteps, &
              error, mc(mcintpsi), funkml_dt, absbserror, polyextrapol)

      elseif (evaltype .eq. 7) then
! ML-MCTDH prop of one SPF from middle layers
         z0 = currentnode%z(1,1)
         zf = currentnode%z(currentnode%navecs,2)
         call bsstep(psi(z0),dtpsi(z0),zf-z0+1,intperiod,time,intorderphi, &
              largestep,bsaccuracy,actualstep,nextstep,smallsteps, &
              error,mc(mcintpsi),funkphinode,absbserror,polyextrapol)
      else
         goto 10
      endif
      call stop_timer(tid)

! --- UPDATE TIME VARIABLE ---

      time = time+actualstep

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'bsstep'
         call bserrormsg(error,message)
         call errormsg
      endif

! --- COUNT GOOD AND BAD STEPS ---

      if (largestep .eq. actualstep) then
         goodsteps = goodsteps+1
      else
         badsteps = badsteps+1
      endif

! --- SAVE STEPSIZE AS INITIAL VALUE FOR NEXT CALL ---

      lastperiod = intperiod
      if (abs(time-abstime-intperiod) .ge. tiny*intperiod) then
         laststep(mode) = actualstep
      else
         laststep(mode) = min(nextstep,intperiod)
      endif

! --- LET STEPSIZE BE A DIVISOR OF UPDATE-TIME ---

      if (laststep(mode) .lt. intperiod) then
         laststep(mode) = intperiod/nint(intperiod/laststep(mode))
      endif

! --- RETURN TO CALLING ROUTINE WHEN FINISHED ---

      if (abs(time-abstime-intperiod) .lt. tiny*intperiod) then
         return
      endif

! --- EVALUATE FUNCTION WITH NEW PSI ---

      if (evaltype .eq. 0) then
         if(lddmctdh)then
            call ddmfunk(time,psi,dtpsi)
         else
            call funk(time,psi,dtpsi)
         endif
      elseif (evaltype .eq. 1) then
         call funkr(time,psi,dtpsi)
      elseif (evaltype .eq. 4) then
         if(lexact) then
            if(lddmctdh)then
               call ddmfunkex(time,psi(zetf(mode,1)),dtpsi(zetf(mode,1)))
            else
               call funkex(time,psi(zetf(mode,1)),dtpsi(zetf(mode,1)))
            endif
         else if (lddmctdh) then
            call ddmfunkphi(time,psi(zetf(mode,1)),dtpsi(zetf(mode,1)))
         else
            call funkphi(time,psi(zetf(mode,1)),dtpsi(zetf(mode,1)))
         endif
      elseif (evaltype .eq. 5) then
         if(lddmctdh)then
            call ddmfunka(time,psi(zpsi(1)),dtpsi(zpsi(1)))
         else
            call funka(time,psi(zpsi(1)),dtpsi(zpsi(1)))
         endif
      elseif (evaltype .eq. 6) then
         call funk1(time,psi,dtpsi)
         call funka(time,psi,dtpsi)
      else
         goto 10
      endif

! --- FIT STEPSIZE TO REST OF INTEGRATION INTERVAL ---

      if (abstime+intperiod-time .lt. enlargefactor*nextstep) then
         largestep = abstime+intperiod-time
      else
         largestep = nextstep
      endif

! --- CONTINUE INTEGRATION ---

      goto 100

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunBS'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runbs


! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNSIL                        *
! *                                                                    *
! * Drives the short iterative Lanczos integration routines,           *
! * (Lanczos/Anoldi-Lanczos).                                           *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   dtpsi:      Action of Hamiltonian on the initial-value vector.   *
! *   intperiod:  Lenght of time interval to be integrated.            *
! *   restartsil: If true, the Krylov space is built up before         *
! *               propagation, else the old Krylov vectors are used.   *
! *   goodsteps:  Number of succesful integration steps made so far.   *
! *   badsteps:   Number of failed but fixed integration steps made so *
! *               far.                                                 *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *   goodsteps:  Number of succesful integration steps made so far.   *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed integration steps made so *
! *               far.                                                 *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:         Complex*16 data needed in the Func (and possibly     *
! *               CalcError) routine.                                  *
! *   mr:         Real*8 data needed in the Func (and possibly         *
! *               CalcError) routine.                                  *
! *   mi:         Integer data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *   ml:         Logical data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runsil (psi,dtpsi,intperiod,time,restartsil, &
                         goodsteps,badsteps,evaltype)

      implicit none

      real(dop), parameter                          :: silaccucorr = 1.0_dop
      logical(kind=4), parameter                    :: stdform = .true.
 
      logical(kind=4), intent(in)                   :: restartsil
      integer(long)                                 :: steps,error,p1,p2,p3,&
                                                       maxorder
      integer(long), intent(in)                     :: evaltype
      integer(long), intent(inout)                  :: goodsteps,badsteps
      real(dop)                                     :: silaccuracy,inttime,stepsize,&
                                                       truestepsize
      real(dop), intent(in)                         :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout):: psi,dtpsi
!      external                                      :: funka,funkex,funka2
      integer(long), save                           :: tid=0 ! timer ID
      character(len=c2) :: string

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         if (evaltype .eq. 4) then
            if(lcap) then
               write(istp,'(a)') 'Exact wavepacket: Lanczos (CSIL)'
               write(istp,'(a7,9x,a6,6x,3a,11x,a5,11x,a9)') &
                 'Step', 'Order', 'Stepsize [',trim(rlxtime),']', &
                 'Error',trim(string)
            else
               write(istp,'(a)') 'Exact wavepacket: Lanczos (SIL)'
               write(istp,'(a7,9x,a6,6x,3a,11x,a5,11x,a9)') &
                 'Step', 'Order', 'Stepsize [',trim(rlxtime),']', &
                 'Error',trim(string)
            endif
            call myflush(istp)
         else if (evaltype .eq. 5) then
            if(lcap) then
               write(istp,'(a)') 'A: Lanczos (CSIL)'
               write(istp,'(a7,9x,a6,6x,3a,11x,a5,11x,a9)')   &
                 'Step', 'Order', 'Stepsize [',trim(rlxtime),']', &
                 'Error',trim(string)
            else
               write(istp,'(a)') 'A: Lanczos (SIL)'
               write(istp,'(a7,9x,a6,6x,3a,11x,a5,11x,a9)')   &
                 'Step', 'Order', 'Stepsize [',trim(rlxtime),']', &
                 'Error',trim(string)
            endif
            call myflush(istp)
         else
            goto 10
         endif
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .eq. 4) then
         silaccuracy = silaccucorr*accphi
      else if (evaltype .eq. 5) then
         silaccuracy = silaccucorr*accavec
      else
         goto 10
      endif

! --- INITIALIZE VARIABLES ---

      inttime = 0.0_dop

! --- CALL LANCZOS ROUTINE UNTIL TIME INTERVAL IS INTEGRATED ---

 100  continue
! for a relaxation do not recalculate the energy during the SIL iterations
      if (lexact) lgetenergy=.false.

      lsil = .true.
      steps = goodsteps+badsteps+1
      stepsize = intperiod-inttime

      call get_timer(tid, 'SIL')
      call start_timer(tid)

! --- PERFORM ONE LANCZOS INTEGRATION STEP ---
      if (lcap .or. ldissin) then
         if (lexact) then
            p1 = (intorderphi+1)**2
            if(lddmctdh)then
               call csilstep(psi(zetf(1,1)),dtpsi(zetf(1,1)),totphidim, &
                             stepsize,intorderphi,silaccuracy,lrlx, &
                             restartsil,stdform,loldsilerr,steps, &
                             truestepsize,maxorder,error,time,macheps, &
                             ddmfunkex,mc(mcintpsi),mc(mcintpsi+p1), &
                             mc(mcintpsi+2*p1))
            else
               call csilstep(psi(zetf(1,1)),dtpsi(zetf(1,1)),totphidim, &
                             stepsize,intorderphi,silaccuracy,lrlx, &
                             restartsil,stdform,loldsilerr,steps, &
                             truestepsize,maxorder,error,time,macheps, &
                             funkex,mc(mcintpsi),mc(mcintpsi+p1), &
                             mc(mcintpsi+2*p1))
            endif
         else 
            p1 = (intordera+1)**2
            if(lddmctdh)then
               call csilstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,stepsize, &
                             intordera,silaccuracy,lrlx,restartsil, &
                             stdform,loldsilerr,steps,truestepsize, &
                             maxorder,error,time,macheps,ddmfunka, &
                             mc(mcintpsi),mc(mcintpsi+p1), &
                             mc(mcintpsi+2*p1))
            else
               call csilstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,stepsize, &
                             intordera,silaccuracy,lrlx,restartsil, &
                             stdform,loldsilerr,steps,truestepsize, &
                             maxorder,error,time,macheps,funka, &
                             mc(mcintpsi),mc(mcintpsi+p1), &
                             mc(mcintpsi+2*p1))
            endif
         endif
      else
         if (lexact) then
            p1 = mrintr + intorderphi + 1 
            p2 = mrintr + 2*(intorderphi+1) 
            p3 = mrintr + 3*(intorderphi+1) 
            if(lddmctdh)then
               call silstep(psi(zetf(1,1)),dtpsi(zetf(1,1)),totphidim, &
                            stepsize,intorderphi,silaccuracy,lrlx, &
                            restartsil,stdform,steps,mc(mcintpsi), &
                            truestepsize,maxorder,error,time, &
                            ddmfunkex,mr(mrworkr),mr(mrintr),mr(p1), &
                            mr(p2),mr(p3))
            else
               call silstep(psi(zetf(1,1)),dtpsi(zetf(1,1)),totphidim, &
                            stepsize,intorderphi,silaccuracy,lrlx, &
                            restartsil,stdform,steps,mc(mcintpsi), &
                            truestepsize,maxorder,error,time, &
                            funkex,mr(mrworkr),mr(mrintr),mr(p1), &
                            mr(p2),mr(p3))
            endif
         else if (rlaxnum.ge.0) then
            p1 = mrintr + intordera + 1
            p2 = mrintr + 2*(intordera+1)
            p3 = mrintr + 3*(intordera+1)
            truestepsize=stepsize
            call sildiag(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intordera, &
                       silaccuracy,stdform,steps,mc(mcintpsi),maxorder, &
                       error,rlaxnum,time,lortho,funka,mr(mrworkr), &
                       mr(mrintr),mr(p1),mr(p2),mr(p3))

         else 
            p1 = mrintr + intordera + 1 
            p2 = mrintr + 2*(intordera+1) 
            p3 = mrintr + 3*(intordera+1) 
            if(lddmctdh)then
               call silstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,stepsize, &
                            intordera,silaccuracy,lrlx,restartsil,stdform, &
                            steps,mc(mcintpsi),truestepsize,maxorder, &
                            error,time,ddmfunka,mr(mrworkr),mr(mrintr), &
                            mr(p1),mr(p2),mr(p3))
            else
               call silstep(psi(zpsi(1)),dtpsi(zpsi(1)),adim,stepsize, &
                            intordera,silaccuracy,lrlx,restartsil,stdform, &
                            steps,mc(mcintpsi),truestepsize,maxorder, &
                            error,time,funka,mr(mrworkr),mr(mrintr), &
                            mr(p1),mr(p2),mr(p3))
            endif
         endif
      endif
      call stop_timer(tid)
      lsil = .false.

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         if (inttypea.eq.6) then
            routine = 'davstep or rdavstep'
            call daverrormsg(error,message)
         elseif (lcap) then
            routine = 'csilstep'
            call csilerrormsg(error,message)
         else
            routine = 'silstep'
            call silerrormsg(error,message)
         endif
         call errormsg
      endif

! --- COUNT GOOD AND BAD STEPS AND DETERMINE MAXIMUM ORDER ---

      if (abs(stepsize-truestepsize) .lt. abs(tiny*stepsize)) then
         goodsteps = goodsteps+1
      else
         badsteps  = badsteps +1
      endif
      maxsilorder = max(maxsilorder,maxorder)


! --- RENORMALISE A-VECTOR if relaxation ---

      if (lrlx .and. .not. lrlxt) then
         if (lexact) then
            call renormphi(psi(zetf(1,1)))
         else 
            call renorma(psi)
         endif
      endif

! --- CHECK WHETHER RELAXATION-EVEC IS COMPLETE (internal check)---
      if(rlaxnum.ge.0 .and. maxorder.le.1+sqrt(0.2_dop*intordera) &
         .and. time .gt. out2+40.0_dop .and. inttypea.lt.6 ) then 
         if( abs( time+truestepsize -  &
             out2*nint((time+truestepsize)/out2)) &
             .lt. 0.01_dop*out2 ) lstop=.true.
      endif

! --- CHECK WHETHER INTEGRATION IS COMPLETE ---
      inttime = inttime+truestepsize
      if (abs(intperiod-inttime) .gt. abs(tiny*intperiod)) then

!    --- EVALUATE FUNCTION ---

         if (lexact) then
            lgetenergy = .true.
            if(lddmctdh)then
               call ddmfunkex(inttime,psi(zetf(1,1)),dtpsi(zetf(1,1))) 
            else
               call funkex(inttime,psi(zetf(1,1)),dtpsi(zetf(1,1))) 
            endif
         else 
            if(lddmctdh)then
               call ddmfunka(time,psi(zpsi(1)),dtpsi(zpsi(1)))
            else
               call funka(time,psi(zpsi(1)),dtpsi(zpsi(1)))
            endif
         endif

!    --- CONTINUE INTEGRATION ---

         goto 100
      endif

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunSIL'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runsil

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNDAV                        *
! *                                                                    *
! * Drives the Davidson "integration" routines, DAv, rDAV, cDAV,       *
! * BrDAV, and BDAV.                                                   *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   dtpsi:      Action of Hamiltonian on the initial-value vector.   *
! *   intperiod:  Length of time interval to be integrated.            *
! *   goodsteps:  Number of succesful integration steps made so far.   *
! *   badsteps:   Number of failed but fixed integration steps made so *
! *               far.                                                 *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *   goodsteps:  Number of succesful integration steps made so far.   *
! *               far.                                                 *
! *   badsteps:   Number of failed but fixed integration steps made so *
! *               far.                                                 *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:         Complex*16 data needed in the Func (and possibly     *
! *               CalcError) routine.                                  *
! *   mr:         Real*8 data needed in the Func (and possibly         *
! *               CalcError) routine.                                  *
! *   mi:         Integer data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *   ml:         Logical data needed in the Func (and possibly        *
! *               CalcError) routine.                                  *
! *                                                                    *
! * 03.10 HDM                                                          *
! *                                                                    *
! **********************************************************************

      subroutine rundav (psi,dtpsi,intperiod,time,goodsteps,  &
                         badsteps,evaltype,Krylov,HKrylov,Kryc,HKryc)

      implicit none

      real(dop), parameter             :: silaccucorr = 1.0_dop
      logical(kind=4), parameter       :: stdform=.true.

      integer(long)                                         :: error
      integer(long), intent(in)                             :: evaltype
      real(dop), intent(in)                                 :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout)        :: psi,dtpsi
      complex(dop), dimension(ikryc), intent(inout)         :: Kryc,HKryc
      real(dop), dimension(ikrylov), intent(inout)          :: Krylov,HKrylov
      character(len=c2) :: string

      logical(kind=4)       :: lf
      integer(long), intent(out) ::    goodsteps,badsteps
      integer(long) ::    steps,p1,p2,p3,p4,p5,maxorder,icr,nmod,j
      real(dop) ::     silaccuracy,inttime,stepsize,truestepsize 
      integer(long), save                           :: tid=0 ! timer ID

#ifdef MPI
      integer(long), dimension(6) :: mpicommi
      real(dop), dimension(6)     :: mpicommr 
#endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if (mpb.gt.0) then
            write(string,'(3a)') 'Q   QQ  Time [',trim(rlxtime),']'

            if(lreal) then
               write(istp,'(a)') 'A: Block-Davidson (real)'
               write(istp,'(a,2x,a,9x,a,15x,a,8x,a,7x,a)')   &
                  'Step','Order','E','beta','Error',trim(string)
            else
               write(istp,'(a)') 'A: Block-Davidson (Hermitian)'
               write(istp,'(a,2x,a,9x,a,15x,a,8x,a,7x,a)')   &
                  'Step','Order','E','beta','Error',trim(string)
            endif
         else 
            write(string,'(3a)') 'Q   Time [',trim(rlxtime),']'

            if(lcap) then
               write(istp,'(a)') 'A: Davidson (complex)'
               write(istp,'(a,2x,a,7x,a,9x,a,8x,a,7x,a5,7x,a)')  &
               'Step','Order','Re(E)','Im(E)','beta','Error', &
                trim(string) 
            elseif(lreal) then
               write(istp,'(a)') 'A: Davidson (real)'
               write(istp,'(a,2x,a,9x,a,14x,a,9x,a5,8x,a)')      &
               'Step','Order','E','beta','Error',trim(string)
            else
               write(istp,'(a)') 'A: Davidson (Hermitian)'
               write(istp,'(a,2x,a,9x,a,14x,a,9x,a5,8x,a)')      &
               'Step','Order','E','beta','Error',trim(string)
            endif
            call myflush(istp)
         endif
      endif

! --- USE SPECIAL ROUTINE FOR REFLEX ---

      if(lreflex) then
         call rundav2(psi,dtpsi,intperiod,time,goodsteps, &
                      badsteps,evaltype,Krylov,HKrylov)
         return
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .eq. 4) then
         silaccuracy = silaccucorr*accphi
      else if (evaltype .eq. 5) then
         silaccuracy = silaccucorr*accavec
      else
         routine = 'RunDAV'
         message = 'Illegal evaluation type'
         call errormsg
      endif

! --- INITIALIZE VARIABLES ---

      inttime = 0.0d0
      kskip = 0

! --- CALL DAV ROUTINE  ---

      lsil = .true.
      steps = goodsteps+badsteps+1
      stepsize = intperiod-inttime

! --- PERFORM ONE DAVIDSON DIAGONALISATION ---

      icr = 1
      if(mod(rlaxnum,10000).eq.930) icr = 2               ! lock

      call get_timer(tid, 'DAV')
      call start_timer(tid)
!
! ------------ BRDAV (block Davidson real)    -----------------------
      if (lreal .and. mpb.gt.0) then ! B-R-DAV
         nmod=0
         do j = 1, 2*maxdim ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 22
! ....      116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode-1
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or. msymmtr(j).eq.-116736) then
               nmod=nmode-1
               lf=.true.
            endif
         enddo
 22      continue
         truestepsize=stepsize
         p2 = (intordera+1)*(intordera+1)
         p3 = p2 + 2*(intordera+1)
         p4 = p3 + adim/npackts + precon*(precon+1)
#ifdef MPI
!         if (lmpi.and.lmpidav.and.lmpistarted) then
!            mpitask=4
!            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!            mpicommi(1)=steps
!            mpicommi(2)=p2
!            mpicommi(3)=p3
!            mpicommi(4)=p4
!            mpicommi(5)=nmod
!            call MPI_Bcast(mpicommi,5,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!            mpicommr(1)=silaccuracy
!            mpicommr(2)=time
!            mpicommr(3)=rlxemin
!            mpicommr(4)=rlxemax
!            call MPI_Bcast(mpicommr,4,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
!            call MPI_Bcast(lf,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
!            lmpidavrun=.true.
!            call mpibrdavstep(psi,dtpsi, &
!                 adim/npackts,intordera,silaccuracy,steps,rlaxnum, &
!                 Krylov,HKrylov,               &   ! Krylov, HKrylov
!                 mc(mcintpsi),mr(mrintr+p4),   &   ! HDiagonal, Tmat
!                 mr(mrintr),mr(mrintr+p2),     &   ! EigenVector, EigenVal
!                 mr(mrintr+p2+intordera+1),    &   ! ovlps
!                 mr(mrworkr),mr(mrintr+p3),    &   ! workr, Rdiagonal
!                 mr(mrintr+p3+adim/npackts),   &   ! davmat
!                 maxorder,error,time,  &
!                 precon,lolsen,funka,funka2,lrreal, &
!                 rlxemin,rlxemax,npackts,nmod,lf, &
!                 mi(miworki),mi(miworki+precon), &  ! indxdav,converged 
!                 mi(miworki+precon+8*npackts+10),&  ! qmap
!                 mc,mr,mi,ml,mi(miworki),mc(mcworkc))  ! workc
!            lmpidavrun=.false.
!         else
#endif
         call brdavstep(psi,dtpsi,    &
                 adim/npackts,intordera,silaccuracy,steps,rlaxnum,  &
                 Krylov,HKrylov, &                 ! Krylov, HKrylov
                 mc(mcintpsi),               &     ! HDiagonal
                 mr(mrintr),mr(mrintr+p2),   &     ! EigenVector, EigenVal
                 mr(mrintr+p2+intordera+1),  &     ! ovlps
                 mr(mrintr+p3),  &                 ! Rdiagonal
                 mr(mrintr+p3+adim/npackts), &     ! davmat
                 maxorder,error,time,        &
                 precon,lolsen,funka,funka2,lrreal,   &
                 rlxemin,rlxemax,npackts,adim,nmod,lf,   &
                 lmult)
#ifdef MPI
!         endif
#endif
!
! ------------ BDAV (block Davidson hermitian)    -------------------
       elseif (.not.lreal .and. mpb.gt.0) then ! B-H-DAV
         nmod=0
         do j = 1, 2*maxdim     ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 24
! ....      116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode-1
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or. msymmtr(j).eq.-116736) then
               nmod=nmode-1
               lf=.true.
            endif
         enddo
 24      continue
         truestepsize=stepsize
         p2 = (intordera+1)*(intordera+1)
         p3 = p2 + 2*(intordera+1)
         p4 = p3 + ((intordera+1)*(intordera+2))/2
#ifdef MPI
!         if (lmpi.and.lmpidav.and.lmpistarted) then
!            mpitask=41
!            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!            mpicommi(1)=steps
!            mpicommi(2)=p2
!            mpicommi(3)=p3
!            mpicommi(4)=p4
!            mpicommi(5)=p5
!            mpicommi(6)=nmod
!            call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!            mpicommr(1)=silaccuracy
!            mpicommr(2)=time
!            mpicommr(3)=rlxemin
!            mpicommr(4)=rlxemax
!            call MPI_Bcast(mpicommr,4,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
!            call MPI_Bcast(lf,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
!            lmpidavrun=.true.
!            call mpibdavstep(psi,dtpsi, &
!                 adim/npackts,intordera,silaccuracy,steps,rlaxnum, &
!                 Kryc,HKryc,                   &   ! Krylov, HKrylov
!                 mc(mcintpsi),mc(mcintpsi+p4), &   ! HDiagonal, Tmat
!                 mc(mcintpsi+p5),mr(mrintr+p2),&   ! EigenVector, EigenVal
!                 mr(mrintr+p2+intordera+1),    &   ! ovlps
!                 mr(mrworkr),mr(mrintr+p3),    &   ! workr, Rdiagonal
!                 mr(mrintr+p3+adim/npackts),   &   ! davmat
!                 maxorder,error,time, &
!                 precon,lolsen,funka, &
!                 rlxemin,rlxemax,npackts,nmod,lf, &
!                 mi(miworki),mi(miworki+precon), &  ! indxdav,converged 
!                 mi(miworki+precon+8*npackts+10), & ! qmap
!                 mc,mr,mi,ml,mi(miworki),mc(mcworkc))  ! workc
!            lmpidavrun=.false.
!         else
#endif
         call bdavstep(psi,dtpsi,adim/npackts,  &
                 intordera,silaccuracy,steps,rlaxnum,   &
                 Kryc,HKryc,                   &    ! Krylov, HKrylov
                 mc(mcintpsi),                 &    ! HDiagonal
                 mc(mcintpsi+p4),mr(mrintr+p2),&    ! EigenVector, EigenVal
                 mr(mrintr+p2+intordera+1),    &    ! ovlps
                 mr(mrintr+p3),                &    ! Rdiagonal
                 maxorder,error,time,          &
                 precon,lolsen,funka,          &
                 rlxemin,rlxemax,npackts,adim,nmod,lf,   &
                 lmult)
#ifdef MPI
!         endif
#endif
!
! ------------ RDAV (Davidson real)    ------------------------------
      elseif (lreal) then       ! R-DAV
         nmod=0
         do j = 1, 2*maxdim     ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 23
! ....      116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or. msymmtr(j).eq.-116736) then
               nmod=nmode
               lf=.true.
            endif
         enddo
 23      continue
         truestepsize=stepsize
         p2 = (intordera+1)*(intordera+1)
         p3 = p2 + 2*(intordera+1) 
         p4 = p3 + adim + precon*(precon+1)
         p5 = p4 + ((intordera+1)*(intordera+2))/2
#ifdef MPI
         if (lmpi.and.lmpidav.and.lmpistarted) then
            mpitask=5
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            mpicommi(1)=steps
            mpicommi(2)=p2
            mpicommi(3)=p3
            mpicommi(4)=p4
            mpicommi(5)=nmod
            mpicommi(6)=p5
            call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            mpicommr(1)=silaccuracy
            mpicommr(2)=time
            mpicommr(3)=rlxemin
            mpicommr(4)=rlxemax
            call MPI_Bcast(mpicommr,4,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
            call MPI_Bcast(lf,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
            lmpidavrun=.true.
            call mpirdavstep(psi,dtpsi,adim, &
                 intordera,silaccuracy,steps,rlaxnum, &
                 Krylov,HKrylov,           &  ! Krylov, HKrylov
                 mc(mcintpsi),mr(mrintr+p4),& ! HDiagonal, Tmat
                 mr(mrintr+p5),             & ! Qmat
                 mr(mrintr),mr(mrintr+p2),  & ! EigenVector, EigenVal
                 mr(mrintr+p2+intordera+1), & ! ovlps
                 mr(mrworkr),mr(mrintr+p3), & ! workr, Rdiagonal
                 mr(mrintr+p3+adim),mi(miworki),& ! davmat,indxdav 
                 maxorder,error,time, &
                 rlxemin,rlxemax,precon,lquad,lolsen, &
                 funka,funka2,lrreal,nmod,lf,mi(miworki),mc(mcworkc)) 
            lmpidavrun=.false.
         else
#endif
         call rdavstep(psi,dtpsi,adim,   &
              intordera,silaccuracy,steps,rlaxnum,   &
              Krylov,HKrylov,              &   ! Krylov, HKrylov
              mc(mcintpsi),mr(mrintr+p4),  &   ! HDiagonal, Tmat
              mr(mrintr+p5),               &   ! Qmat
              mr(mrintr),mr(mrintr+p2),    &   ! EigenVector, EigenVal
              mr(mrintr+p2+intordera+1),   &   ! ovlps
              mr(mrintr+p3),               &   ! Rdiagonal
              mr(mrintr+p3+adim),          &   ! davmat
              maxorder,error,time,         &
              rlxemin,rlxemax,precon,lquad,lolsen,  &
              funka,funka2,lrreal,nmod,lf)
#ifdef MPI
         endif
#endif
!
! ------------   DAV, (hermitian Davidson) --------------------------
      elseif (.not.lreal .and. .not.lcap) then ! DAV
         nmod=0
         do j = 1, 2*maxdim     ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 25
! ....      116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or.msymmtr(j).eq.-116736) then
               nmod=nmode
               lf=.true.
            endif
         enddo
 25      continue
         truestepsize=stepsize
         p2 = icr*adim
         p3 = p2 + (intordera+1)*(intordera+1)
         call davstep(psi,dtpsi,adim,   &
              intordera,silaccuracy,steps,rlaxnum,   &
              Kryc,HKryc,                       &    ! Krylov, HKrylov
              mc(mcintpsi),mc(mcintpsi+p2),     &    ! HDiagonal, Tmat
              mc(mcintpsi+p3),mr(mrintr),       &    ! EigenVector, EigenVal
              mr(mrintr+intordera+1),           &    ! ovlps 
              maxorder,error,time,rlxemin,rlxemax,  &
              precon,lquad,lolsen,funka,nmod,lf)
!
! ------------  cDAV (complex Davidson (resonances)) ---------
      elseif (lcap) then        ! C-DAV
         truestepsize=stepsize
         p1 = 0
         p2 = icr*adim
         p3 = p2 + (intordera+1)*(intordera+1)
         p4 = 16*intordera+18
         p5 = p4 + (intordera+1)*(intordera+1)
#ifdef MPI
         if (lmpi.and.lmpidav.and.lmpistarted) then
            p1 = 0
            p2 = icr*adim
            p3 = p2 + (intordera+1)*(intordera+1)
            p4 = 16*intordera+18
            p5 = p4 + (intordera+1)*(intordera+1)
            mpitask=12
            call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            mpicommi(1)=steps
            mpicommi(2)=p1
            mpicommi(3)=p2
            mpicommi(4)=p3
            mpicommi(5)=p4
            mpicommi(6)=p5
            call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
            mpicommr(1)=silaccuracy
            mpicommr(2)=time
            mpicommr(3)=rlxemin
            mpicommr(4)=rlxemax
            mpicommr(5)=stepsize
            mpicommr(6)=rlximin
            call MPI_Bcast(mpicommr,6,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD,mpierr)
            lmpidavrun=.true.
            call mpicdavstep(psi,dtpsi,adim, &
                 intordera,silaccuracy,steps,rlaxnum, &
                 Kryc,HKryc,                     &  ! Krylov, HKrylov
                 mc(mcintpsi),mc(mcintpsi+p2),   &  ! HDiagonal, Tmat
                 mc(mcworkc+p4),                 &  ! TTmat
                 mc(mcintpsi+p3),mc(mcworkc+p5), & ! EigenVector, EigenVal
                 mc(mcworkc),mr(mrworkr),mr(mrintr), & ! workc, workr, ovlps
                 stepsize,maxorder,error,time,rlxemin,rlxemax,rlximin, &
                 precon,lquad,lolsen,funka)
            lmpidavrun=.false.
         else
#endif
         call cdavstep(psi,dtpsi,adim,   &
              intordera,silaccuracy,steps,rlaxnum,   &
              Kryc,HKryc,                     &   ! Krylov, HKrylov
              mc(mcintpsi),mc(mcintpsi+p2),   &   ! HDiagonal, Tmat
              mc(mcintpsi+p3),mc(mcworkc+p5), &   ! EigenVector,EigenVal
              mr(mrintr),                     &   ! ovlps
              maxorder,error,time,rlxemin,rlxemax, &
              rlximin,precon,lquad,lolsen,funka)
#ifdef MPI
         endif
#endif

      endif      ! Different DAV routines

      call stop_timer(tid)
      lsil = .false.

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         if (lreal) then
            routine = 'rdavstep'
         elseif (lcap) then
            routine = 'cdavstep'
         else
            routine = 'davstep'
         endif
         call daverrormsg(error,message)
         call errormsg
      endif

! --- COUNT GOOD AND BAD STEPS AND DETERMINE MAXIMUM ORDER ---

      if (abs(stepsize-truestepsize) .lt. abs(tiny*stepsize)) then
         goodsteps = goodsteps+1
      else
         badsteps  = badsteps +1
      endif
      maxsilorder = max(maxsilorder,maxorder)


! --- CHECK WHETHER RELAXATION-EVEC IS COMPLETE (internal check)---
      if(rlaxnum.ge.0 .and. maxorder.le.1+sqrt(0.2d0*intordera)   &
         .and. time .gt. out2+40.d0 .and. inttypea.lt.6 ) then 
         if( abs( time+truestepsize -  out2*nint((time+truestepsize)/out2))  &
             .lt. 0.01d0*out2 ) lstop=.true.
      endif

      return
      end subroutine rundav

! **********************************************************************

      subroutine rundav2(psi,dtpsi,intperiod,time,goodsteps,   &
                         badsteps,evaltype,Krylov,HKrylov)

      implicit none

      real(dop), parameter             :: silaccucorr = 1.0_dop
      logical(kind=4), parameter       :: stdform=.true.

      integer(long)                                         :: error
      integer(long), intent(in)                             :: evaltype
      real(dop), intent(in)                                 :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout)        :: psi,dtpsi
      real(dop), dimension(ikrylov), intent(inout)          :: Krylov,HKrylov

      logical(kind=4)       :: lf
      integer(long), intent(out) ::    goodsteps,badsteps
      integer(long) ::    steps,p2,p3,p4,p5,maxorder,icr,nmod,j,kk,m
      real(dop) ::     silaccuracy,inttime,stepsize,truestepsize 
      integer(long), save                           :: tid=0 ! timer ID

#ifdef MPI
      integer(long), dimension(6) :: mpicommi
      real(dop), dimension(4)     :: mpicommr 
#endif

! --- INITIALIZE VARIABLES ---

      inttime = 0.0d0

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .eq. 5) then
         silaccuracy = silaccucorr*accavec
      else
         routine = 'RunDAV2'
         message = 'Illegal evaluation type'
         call errormsg
      endif

! --- CALL DAV ROUTINE ---

      lsil = .true.
      steps = goodsteps+badsteps+1
      stepsize = intperiod-inttime

      icr = 1
      if(mod(rlaxnum,10000).eq.930) icr = 2               ! lock

      call get_timer(tid, 'DAV2')
      call start_timer(tid)

      nmode    = nmode-1      ! ignore packet mode
      adim     = adim/2       ! adjust adim
      maxblock = maxblock/2   ! adjust maxblock
      block(1) = block(1)/2   ! adjust block
      do m = 1, nmode         ! adjust ndim
         ndim(m,1) = ndim(m,1)/2
      enddo


! ------------ BRDAV (block Davidson real)    -----------------------
      if (lreal .and. mpb.gt.0) then ! B-R-DAV
         nmod=0
         do j = 1, 2*maxdim ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 22
! ....      116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode-1
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or. msymmtr(j).eq.-116736) then
               nmod=nmode-1
               lf=.true.
            endif
         enddo
 22      continue
         truestepsize=stepsize
         p2 = (intordera+1)*(intordera+1)
         p3 = p2 + 2*(intordera+1)

! --- PERFORM TWO DAVIDSON DIAGONALISATION, ONE FOR EACH PACKET ---

         do kskip = 1, 2
            kk = (kskip-1)*adim + 1

#ifdef MPI
!            if (lmpi.and.lmpidav.and.lmpistarted) then
!               mpitask=4
!               call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!               mpicommi(1)=steps
!               mpicommi(2)=p2
!               mpicommi(3)=p3
!               mpicommi(4)=p4
!               mpicommi(5)=nmod
!               call MPI_Bcast(mpicommi,5,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!               mpicommr(1)=silaccuracy
!               mpicommr(2)=time
!               mpicommr(3)=rlxemin
!               mpicommr(4)=rlxemax
!               call MPI_Bcast(mpicommr,4,MPI_DOUBLE_PRECISION,0,MPI_COMM_WORLD, &
!                              mpierr)
!               call MPI_Bcast(lf,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
!               lmpidavrun=.true.
!               call mpibrdavstep(psi(kk),dtpsi(kk), &
!                    adim/npackts,intordera,silaccuracy,steps,rlaxnum, &
!                    Krylov,HKrylov,                &  ! Krylov, HKrylov
!                    mc(mcintpsi),mr(mrintr+p4),    &  ! HDiagonal, Tmat
!                    mr(mrintr),mr(mrintr+p2),      &  ! EigenVector,EigenVal
!                    mr(mrintr+p2+intordera+1),     &  ! ovlps
!                    mr(mrworkr),mr(mrintr+p3),     &  ! workr, Rdiagonal
!                    mr(mrintr+p3+adim/npackts),    &  ! davmat
!                    maxorder,error,time, &
!                    precon,lolsen,funka,funka2,lrreal, &
!                    rlxemin,rlxemax,npackts,nmod,lf, &
!                    mi(miworki),mi(miworki+precon), & ! indxdav,converged
!                    mi(miworki+precon+8*npackts+10), &! qmap
!                    mc,mr,mi,ml,mi(miworki),mc(mcworkc))  ! workc
!               lmpidavrun=.false.
!            else
#endif
            call brdavstep(psi(kk),dtpsi(kk),   &
                 adim/npackts,intordera,silaccuracy,steps,rlaxnum,   &
                 Krylov,HKrylov,                &   ! Krylov, HKrylov
                 mc(mcintpsi),                  &   ! HDiagonal, Tmat
                 mr(mrintr),mr(mrintr+p2),      &   ! EigenVector,EigenVal
                 mr(mrintr+p2+intordera+1),     &   ! ovlps
                 mr(mrintr+p3),                 &   ! Rdiagonal
                 mr(mrintr+p3+adim/npackts),    &   ! davmat
                 maxorder,error,time,           &  
                 precon,lolsen,funka,funka2,lrreal,  &
                 rlxemin,rlxemax,npackts,adim,nmod,lf,  &
                 lmult)
#ifdef MPI
!            endif
#endif
            if(lintstep.and.kskip.eq.1) write(istp,'(78a1)') ('-',j=1,78)
         enddo    ! kskip = 1, 2


      else

! ------------ RDAV (Davidson real)    ------------------------------
         nmod=0
         do j = 1, 2*maxdim ! Setting nmod, i.e. with or without symmetrization
            if( msymmtr(j) .eq. 0 ) then
               goto 23
! ....   116736=51200+65536, 182272=51200+2*65536 (without/with persist)
            elseif(msymmtr(j).eq.182272 .or. msymmtr(j).eq.116736) then
               nmod=nmode
               lf=.false.
            elseif(msymmtr(j).eq.-182272 .or. msymmtr(j).eq.-116736) then
               nmod=nmode
               lf=.true.
            endif
         enddo
 23      continue
         truestepsize=stepsize
         p2 = (intordera+1)*(intordera+1)
         p3 = p2 + 2*(intordera+1) 
         p4 = p3 + adim + precon*(precon+1)
         p5 = p4 + ((intordera+1)*(intordera+2))/2

! --- PERFORM TWO DAVIDSON DIAGONALISATION, ONE FOR EACH PACKET ---

         do kskip = 1, 2
            kk = (kskip-1)*adim + 1
#ifdef MPI
!            if (lmpi.and.lmpidav.and.lmpistarted) then
!               mpitask=5
!               call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!               mpicommi(1)=steps
!               mpicommi(2)=p2
!               mpicommi(3)=p3
!               mpicommi(4)=p4
!               mpicommi(5)=nmod
!               mpicommi(6)=p5
!               call MPI_Bcast(mpicommi,6,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
!               mpicommr(1)=silaccuracy
!               mpicommr(2)=time
!               mpicommr(3)=rlxemin
!               mpicommr(4)=rlxemax
!               call MPI_Bcast(mpicommr,4,MPI_DOUBLE_PRECISION,0, &
!                                                  MPI_COMM_WORLD,mpierr)
!               call MPI_Bcast(lf,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
!               lmpidavrun=.true.
!               call mpirdavstep(psi(kk),dtpsi(kk),adim, &
!                    intordera,silaccuracy,steps,rlaxnum, &
!                    Krylov,HKrylov,             & ! Krylov, HKrylov
!                    mc(mcintpsi),mr(mrintr+p4), & ! HDiagonal, Tmat
!                    mr(mrintr+p5),              & ! Qmat
!                    mr(mrintr),mr(mrintr+p2),   & ! EigenVector, EigenVal
!                    mr(mrintr+p2+intordera+1),  & ! ovlps
!                    mr(mrworkr),mr(mrintr+p3),  & ! workr, Rdiagonal
!                    mr(mrintr+p3+adim),mi(miworki), & ! davmat,indxdav 
!                    maxorder,error,time, &
!                    rlxemin,rlxemax,precon,lquad,lolsen, &
!                    funka,funka2,lrreal,nmod,lf, &
!                    mc,mr,mi,ml,mi(miworki),mc(mcworkc))
!               lmpidavrun=.false.
!            else
#endif
            call rdavstep(psi(kk),dtpsi(kk),adim,   &
                 intordera,silaccuracy,steps,rlaxnum,   &
                 Krylov,HKrylov,              &   ! Krylov, HKrylov
                 mc(mcintpsi),mr(mrintr+p4),  &   ! HDiagonal, Tmat
                 mr(mrintr+p5),               &   ! Qmat
                 mr(mrintr),mr(mrintr+p2),    &   ! EigenVector, EigenVal
                 mr(mrintr+p2+intordera+1),   &   ! ovlps
                 mr(mrintr+p3),               &   ! Rdiagonal
                 mr(mrintr+p3+adim),          &   ! davmat
                 maxorder,error,time,         &
                 rlxemin,rlxemax,precon,lquad,lolsen,  &
                 funka,funka2,lrreal,nmod,lf)
#ifdef MPI
!            endif
#endif
 
            if(lintstep.and.kskip.eq.1) write(istp,'(78a1)') ('-',j=1,78)
         enddo    ! kskip = 1, 2

      endif ! BRDAV or RDAV

      call stop_timer(tid)
      lsil  = .false.
      kskip = 0
      do m = 1, nmode        ! restore ndim
         ndim(m,1) = ndim(m,1)*2
      enddo
      adim     = adim*2      ! restore adim
      maxblock = maxblock*2  ! restore maxblock
      block(1) = block(1)*2  ! restore block
      nmode    = nmode+1     ! restore nmode 

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'rdavstep'
         call daverrormsg(error,message)
         call errormsg
      endif

! --- COUNT GOOD AND BAD STEPS AND DETERMINE MAXIMUM ORDER ---

      if (abs(stepsize-truestepsize) .lt. abs(tiny*stepsize)) then
         goodsteps = goodsteps+1
      else
         badsteps  = badsteps +1
      endif
      maxsilorder = max(maxsilorder,maxorder)


! --- CHECK WHETHER RELAXATION-EVEC IS COMPLETE (internal check)---
      if(rlaxnum.ge.0 .and. maxorder.le.1+sqrt(0.2d0*intordera)   &
         .and. time .gt. out2+40.d0 .and. inttypea.lt.6 ) then
         if( abs( time+truestepsize - out2*nint((time+truestepsize)/out2))  &
             .lt. 0.01d0*out2 ) lstop=.true.
      endif

      return
      end subroutine rundav2

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNRK5                        *
! *                                                                    *
! * Drives the RK5 integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runrk5 (psi,dtpsi,intperiod,time,evaltype,mode,restartrk5)

      implicit none

      real(dop), parameter                           :: rk5accucorr = 1.0_dop

      integer(long)                                  :: error,m,z0,zf
      integer(long), intent(in)                      :: evaltype,mode
      real(dop)                                      :: rk5accuracy
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      logical(kind=4), save                          :: initialized = .false.
      logical(kind=4), intent(inout)                 :: restartrk5
      integer(long), save                            :: tid=0 ! timer ID
      character(len=c2) :: string
     
      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- INITIALIZATION ---

      if (restartrk5) then
         initialized = .false.
         dt = 0.0_dop
      endif
      restartrk5 = .false.

!     is this the first call?
!     Initialize integrator the first call or after spawning
      if ((.not. initialized) .or. lnewspf) then
         write(istp,'(a)') 'Integrator initialized'
!        yes: initialize the stepsizes
         if (evaltype .le. 1) then
!           VMF integrations
            laststep(0) = dt
         elseif (evaltype .ge. 4 .or. evaltype.eq.5) then
!           CMF integrations
            laststep(0) = dta
            do m=1,nmode
               laststep(m) = dtphi
            enddo
         elseif (evaltype.eq.6) then
!           ML-CMF integration, no A-vector
            laststep(0) = dtPhi
         else
            goto 10
         endif
         initialized = .true.
         lnewspf = .false.
      endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1) .or. &
             (evaltype .eq. -1) .or. (evaltype .eq. 6)) then
            write(istp,'(a)')&
               'Complete equation: Runge-Kutta order 5'
         elseif (evaltype .eq. 4) then
            if (lexact) then
               write(istp,'(a)')&
                    'Exact wavepacket: Runge-Kutta order 5'
            else
               write(istp,'(a,i2,a)') 'Mode', mode,&
                    ': Runge-Kutta order 5'
            endif
         elseif (evaltype .eq. 7) then
            write(istp,'(a,i2,a)') 'Node', currentnode%num,&
                    ': Runge-Kutta order 5'
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: Runge-Kutta order 5'
         else
            goto 10
         endif
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write(istp,'(a7,8x,a9,4x,3a,6x,a14,7x,a9)')&
              'Step', 'Accepted?', 'Stepsize [',trim(rlxtime),']', &
              'Error (scaled)',trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .le. 1) then
         rk5accuracy = rk5accucorr*accuracy
      elseif (evaltype .eq. 4) then
         rk5accuracy = rk5accucorr*accphi
      elseif (evaltype .eq. 5) then
         rk5accuracy = rk5accucorr*accavec
      elseif (evaltype .eq. 6) then
         rk5accuracy = rk5accucorr*accphi
      elseif (evaltype .eq. 7) then
         rk5accuracy = rk5accucorr*accphi
      else
         goto 10
      endif

! --- CALL RK5 INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'RK5')
      call start_timer(tid)

      if (evaltype .eq. 0) then        ! VMF propagation, funkl, h-proj
         if(lddmctdh)then
            call rk5(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                     rk5accuracy,error,mc(mcintpsi),ddmfunk,&
                     liork, importhovmf,lrktest,gwpodetest,lrhonrm,rhonrm)
         else
            call rk5(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                     rk5accuracy,error,mc(mcintpsi),funk,&
                     liork, importhovmf,lrktest,gwpodetest,lrhonrm,rhonrm)
         endif
      elseif (evaltype .eq. 1) then    ! VMF propagation, funkr, proj-h
         call rk5(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),funkr,&
                 liork, importhovmf,lrktest,gwpodetest,lrhonrm,rhonrm)
      elseif (evaltype .eq. -1) then
         call rk5(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),funkr,&
                 liork, importhovmf,lrktest,gwpodetest,lrhonrm,rhonrm)

      elseif (evaltype .eq. 4) then    ! CMF, Phi-propagation (one SPF), or exact
         if (lexact) then
            if(lddmctdh)then
               call rk5(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                        intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                        mc(mcintpsi),ddmfunkex,.false.,dummypost,&
                        .false.,dummypost,.false.,dummyerrfn)
            else
               call rk5(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                        intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                        mc(mcintpsi),funkex,.false.,dummypost,&
                        .false.,dummypost,.false.,dummyerrfn)
            endif
         else if (lddmctdh) then
            call rk5(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                    intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                    mc(mcintpsi),ddmfunkphi,liork,importho,lrktest,gwpodetest1m,&
                    lrhonrm,rhonrm)
         else 
            call rk5(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                    intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                    mc(mcintpsi),funkphi,liork,importho,lrktest,gwpodetest1m,&
                    lrhonrm,rhonrm)
         endif
      elseif (evaltype .eq. 5) then    ! CMF, A-propagation
         if (lddmctdh) then
            call rk5(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                     ddmfunka,.false.,dummypost,.false.,dummypost,&
                     .false.,dummyerrfn)
         else
            call rk5(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),funka, &
                     .false.,dummypost,.false.,dummypost,.false.,dummyerrfn)
         endif
      elseif (evaltype .eq. 6) then  ! ML-MCTDH prop of whole WF tree
         call rk5(psi,dtpsi,dgldim,intperiod,time, &
                 laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                 funkml_dt,liork, importhovmf,lrktest,gwpodetest,&
                 .false.,dummyerrfn)
      elseif (evaltype .eq. 7) then  ! ML-MCTDH prop of one SPF from middle layers
         z0 = currentnode%z(1,1)
         zf = currentnode%z(currentnode%navecs,2)
 
         call rk5(psi(z0),dtpsi(z0),zf-z0+1,intperiod,time, &
                 laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                 funkphinode,liork, importho,lrktest,gwpodetest,&
                 .false.,dummyerrfn)
      else
         goto 10
      endif
      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'RunRK5'
         message = 'Integration failed.'
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunRK5'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runrk5

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNRK8                        *
! *                                                                    *
! * Drives the RK8 integration routine.                                *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runrk8 (psi,dtpsi,intperiod,time,evaltype,mode,restartrk5)

      implicit none

      real(dop), parameter                           :: rk8accucorr = 1.0_dop

      integer(long)                                  :: error,m,z0,zf
      integer(long), intent(in)                      :: evaltype,mode
      real(dop)                                      :: rk8accuracy
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      logical(kind=4), save                          :: initialized = .false.
      logical(kind=4), intent(inout)                 :: restartrk5
      integer(long), save                            :: tid=0 ! timer ID
      character(len=c2) :: string

      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- INITIALIZATION ---

      if (restartrk5) then
         initialized = .false.
         dt = 0.0_dop
      endif
      restartrk5=.false.
!     is this the first call?
!     Initialize integrator the first call or after spawning
      if ((.not. initialized) .or. lnewspf) then
         write(istp,'(a)') 'Integrator initialized'
!        yes: initialize the stepsizes
         if (evaltype .le. 1) then
!           VMF integrations
            laststep(0) = dt
         elseif (evaltype .ge. 4 .or. evaltype .eq. 5) then
!           CMF integrations
            laststep(0) = dta
            do m=1,nmode
               laststep(m) = dtphi
            enddo
         elseif (evaltype.eq.6) then
!           ML-CMF integration, no A-vector
            laststep(0) = dtPhi
         else
            goto 10
         endif
         initialized = .true.
         lnewspf = .false.
      endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1) .or. &
               evaltype .eq. 6) then
            write(istp,'(a)') &
               'Complete equation: Runge-Kutta order 8'
         elseif (evaltype .eq. 4 .or. evaltype .eq. 7) then
            if (lexact) then
               write(istp,'(a)') &
                    'Exact wavepacket: Runge-Kutta order 8'
            else
               write(istp,'(a,i2,a)') 'Mode', mode, &
                    ': Runge-Kutta order 8'
            endif
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: Runge-Kutta order 8'
         else
            goto 10
         endif
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write(istp,'(a7,8x,a9,4x,3a,6x,a14,7x,a9)') &
              'Step', 'Accepted?', 'Stepsize [',trim(rlxtime),']', &
              'Error (scaled)', trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .le. 1) then
         rk8accuracy = rk8accucorr*accuracy
      elseif (evaltype .eq. 4) then
         rk8accuracy = rk8accucorr*accphi
      elseif (evaltype .eq. 5) then
         rk8accuracy = rk8accucorr*accavec
      elseif (evaltype .eq. 6) then
         rk8accuracy = rk8accucorr*accphi
      else
         goto 10
      endif

! --- CALL RK8 INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'RK8')
      call start_timer(tid)

      if (evaltype .eq. 0) then    ! VMF propagation, funkl, h-proj
         if(lddmctdh)then
            call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0), &
                     rk8accuracy,error,mc(mcintpsi),ddmfunk,liork,importhovmf,&
                     lrktest,gwpodetest,lrhonrm,rhonrm)
         else
            call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0), &
                     rk8accuracy,error,mc(mcintpsi),funk,liork,importhovmf,&
                     lrktest,gwpodetest,lrhonrm,rhonrm)
         endif
      elseif (evaltype .eq. 1) then  ! VMF propagation, funkr, proj-h
         call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0), &
                 rk8accuracy,error,mc(mcintpsi),funkr,liork,importhovmf,&
                 lrktest,gwpodetest,lrhonrm,rhonrm)
      elseif (evaltype .eq. 4) then   ! CMF, Phi-propagation (one SPF), or exact
         if (lexact) then
            if(lddmctdh)then
               call rk8(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                        intperiod,time,laststep(mode),rk8accuracy,error, &
                        mc(mcintpsi),ddmfunkex,.false.,dummypost,&
                        .false.,dummypost,.false.,dummyerrfn)
            else
               call rk8(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                        intperiod,time,laststep(mode),rk8accuracy,error, &
                        mc(mcintpsi),funkex,.false.,dummypost,&
                        .false.,dummypost,.false.,dummyerrfn)
            endif
! ------ NEW STUFF DAV
         else if(lquadphi) then
            call rk8(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                    intperiod,time,laststep(mode),rk8accuracy,error, &
                    mc(mcintpsi),funkphi2,.false.,dummypost,&
                    .false.,dummypost,.false.,dummyerrfn)
         else if (lddmctdh) then
            call rk8(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                    intperiod,time,laststep(mode),rk8accuracy,error, &
                    mc(mcintpsi),ddmfunkphi,liork,importho,&
                    lrktest,gwpodetest,lrhonrm,rhonrm)
         else
            call rk8(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode), &
                    intperiod,time,laststep(mode),rk8accuracy,error, &
                    mc(mcintpsi),funkphi,liork,importho,&
                    lrktest,gwpodetest,lrhonrm,rhonrm)
         endif
      elseif (evaltype .eq. 5) then   ! CMF, A-propagation
         if (lddmctdh) then
            call rk8(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),rk8accuracy,error,mc(mcintpsi),ddmfunka, &
                     .false.,dummypost,.false.,dummypost,.false.,dummyerrfn)
         else
            call rk8(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),rk8accuracy,error,mc(mcintpsi),funka, &
                     .false.,dummypost,.false.,dummypost,.false.,dummyerrfn)
         endif
      elseif (evaltype .eq. 6) then   ! ML-MCTDH prop of whole WF tree
         call rk8(psi,dtpsi,dgldim,intperiod,time,laststep(0),&
                 rk8accuracy,error,mc(mcintpsi),funkml_dt,&
                 .false.,dummypost,.false.,dummypost,.false.,dummyerrfn)
      elseif (evaltype .eq. 7) then  ! ML-MCTDH prop of one SPF from middle layers
         z0 = currentnode%z(1,1)
         zf = currentnode%z(currentnode%navecs,2)
 
         call rk8(psi(z0),dtpsi(z0),zf-z0+1,intperiod,time,laststep(0),&
                 rk8accuracy,error,mc(mcintpsi),funkphinode,&
                 .false.,dummypost,.false.,dummypost,.false.,dummyerrfn)
      else
         goto 10
      endif
      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'RunRK8'
         message = 'Integration failed.'
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunRK8'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runrk8

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNLSODA                      *
! *                                                                    *
! * Drives the LSODA integration routine.                              *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *                                                                    *
! **********************************************************************

      subroutine runlsoda (psi,intperiod,time)

      implicit none

      complex(dop), dimension(dgldim), intent(inout) :: psi
      real(dop), intent(in)                          :: intperiod,time
      integer(long), save                            :: tid=0 ! timer ID

real(dop),dimension(1),parameter::aint=1.0e-15_dop
real(dop),dimension(2*dgldim)::realpsi
real(dop),dimension(1),save::rtol
real(dop)::t0,t1
integer,dimension(1),save::neq
integer::istate
logical(kind=4), save          :: initialized = .false.

! --- INITIALIZATION ---

if (.not.initialized) then
  istate=1
  rtol=accuracy
  neq=2*dgldim
  lrw=22+32*dgldim+4*dgldim**2
  liw=20+2*dgldim
  allocate(LSODArw(lrw))
  allocate(LSODAiw(liw))
  initialized = .true.
else
  istate=2
endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---
!      if (lintstep) then
!!         if ((evaltype .eq. 0) .or. (evaltype .eq. 1)) then
!            write(istp,'(a)') &
!               'Complete equation: LSODA adaptive integrator'
!         call myflush(istp)
!      endif

!      realpsi(1:2*dgldim:2)=real(psi)
!      realpsi(2:2*dgldim:2)=aimag(psi)
      realpsi=transfer(psi,realpsi)

! --- CALL LSODA INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'LSODA')
      call start_timer(tid)

      LSODArw(5:10)=0
      LSODAiw(5:10)=0
      LSODAiw(5)=1     ! print method switch
      LSODAiw(6)=huge(1)-5   ! max steps
      t0=time
      t1=time+intperiod
write(*,*)time,intperiod,t1
       ! Only implemented for VMF at the moment.  Hence lfunkr and evaltype==1
      call dlsoda(funkrreal,neq,realpsi,t0,t1, &
                1,rtol,aint,1,istate,1,LSODArw,lrw,LSODAiw,liw,dummypost,2)
      call stop_timer(tid)

      if (istate<0) then
        routine = 'RunLSODA'
        write(message,'(''LSODA failed:'',i4)')istate
        call errormsg
      endif

!      psi=cmplx(realpsi(1:2*dgldim:2),realpsi(2:2*dgldim:2))
      psi=transfer(realpsi,psi)

      end subroutine runlsoda

! **********************************************************************
! *                           SUBROUTINE RUNZVODE                      *
! *                                                                    *
! * Drives the ZVODE integration routine.                              *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runzvode (psi,intperiod,time,evaltype,mode,restartzvode)

       use decimal, only: dop, long
       use BELglobals, only : errorBELprop
      implicit none

      real(dop), parameter                           :: zvodeaccucorr = 1e-1_dop

      integer(long),save                             :: error,m,s
      integer(long),save                             :: lastngwpd
      integer(long), intent(in)                      :: evaltype,mode
      logical(kind=4), intent(inout)                    :: restartzvode
      logical(kind=4)                                :: localrestart
      real(dop)                                      :: zvodeaccuracy,t0,t1
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(inout)                       :: intperiod
      real(dop), intent(in)                          :: time
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim)                :: psisave
      complex(dop), dimension(agmatdim)              :: asave
      logical(kind=4), save                          :: initialized = .false.
      logical(kind=4), save                          :: errmsg

      integer(long),save                             :: liwork,lrwork,lastN=0
      integer(long),save                             :: lcwork,N

      complex(dop), dimension(:), allocatable,save   :: cwork
      real(dop),dimension(:), allocatable,save       :: rwork, deviation
      integer(long), dimension(:), allocatable,save  :: iwork

      integer(long)                                  :: gwpdynloop,&
                                                        recoverymethod

      integer(long)                                  :: maxdi,maxdm,maxds,itask
      ! integer(long),save                             :: lmaxdim      
      character(len=c2) :: string

      ! complex(dop), dimension(:,:), allocatable,save :: ovr,ovr1
      real(dop),save :: lastRA
      integer(long), save                              :: tid=0 ! timer ID

!write(*,*) "start runzvode"
!write(*,*) "adim",adim
!write(*,*) "dgldim",dgldim
!write(*,*) "N",N
!
      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

      ! --- INITIALIZATION ---

      localrestart=restartzvode
      restartzvode=.false.

!      N=max(adim,dgldim)!,lmaxdim*lmaxdim*2+lmaxdim*2)
!      if (mode>0) N=max(N,phidim(mode))
!write(*,*) "N",N

!         lastN=N
!         if (allocated(iwork)) deallocate(iwork)
!         if (allocated(rwork)) deallocate(rwork)
!         if (allocated(cwork)) deallocate(cwork)
!         if (allocated(deviation)) deallocate(deviation)
!         liwork=N+30
!         allocate(iwork(liwork))
!         lrwork=N+20
!         allocate(rwork(lrwork))
!         lcwork=8*N+2*N**2
!         allocate(cwork(lcwork))
!         allocate(deviation(N))

      !     is this the first call?
     if (.not. initialized) then
         ! initialize ZVODE work arrays
         ! the minimum size of these is specified at the beginning of zvode.f
         error=1
         lastngwpd=sum(ngwpdyn)

         lastN=N
         if (allocated(iwork)) deallocate(iwork)
         if (allocated(rwork)) deallocate(rwork)
         if (allocated(cwork)) deallocate(cwork)
         if (allocated(deviation)) deallocate(deviation)
         liwork=N+30
         allocate(iwork(liwork))
         lrwork=N+20
         allocate(rwork(lrwork))
         lcwork=8*N+2*N**2
         if (N>0.25_dop*sqrt(64+8.0_dop*huge(1_long))-2) then
           routine = 'RunZVODE'
           message = 'Problem too large to declare cwork'
           call errormsg
         endif
         allocate(cwork(lcwork))
         allocate(deviation(N))

!         rwork(5:10)=0
!         iwork(5:10)=0
! --- set maximum iterations and minimum stepsize
!         iwork(6)=abs(zvodemaxit)
!         rwork(5)=zvode1ststp
!         rwork(6)=zvodemaxstp
!         rwork(7)=zvodeminstp

!  initialize the stepsizes
         if (evaltype .le. 1) then
!           VMF integrations
            laststep(0) = dt
         elseif (evaltype .ge. 4) then
!           CMF integrations
            laststep(0) = dta
            do m=1,nmode
               laststep(m) = dtphi
            enddo
         else
            goto 10
         endif
         errmsg = zvodemaxit>=0
         initialized = .true.
         recoverymethod=1
         lastRA=0
      endif

!         rwork(5:10)=0
!         iwork(5:10)=0
!! --- set maximum iterations and minimum stepsize
!         iwork(6)=abs(zvodemaxit)
!         rwork(5)=zvode1ststp
!         rwork(6)=zvodemaxstp
!         rwork(7)=zvodeminstp

!      if (N>lastN) stop 'Need to reimplement workspace growing in runzvode'

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

!      if (lintstep) then
!         if ((evaltype .eq. 0) .or. (evaltype .eq. 1)) then
!            write(istp,'(a)')&
!               'Complete equation: ZVODE'
!         elseif (evaltype .eq. 4) then
!            if (lexact) then
!               write(istp,'(a)')&
!                    'Exact wavepacket: ZVODE'
!            else
!               write(istp,'(a,i2,a)') 'Mode', mode,&
!                    ': ZVODE'
!            endif
!         elseif (evaltype .eq. 5) then
!            write(istp,'(a)') 'A: ZVODE'
!         else
!            goto 10
!         endif
!         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
!              'Step', '#GWP DoFs', 'Stepsize [fs]', 'Error (scaled)',&
!              'Time [fs]'
!         call myflush(istp)
!      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

!      if (evaltype .le. 1) then
!         zvodeaccuracy = zvodeaccucorr*accuracy
!      elseif (evaltype .eq. 4) then
!         zvodeaccuracy = zvodeaccucorr*accphi
!      elseif (evaltype .eq. 5) then
!         zvodeaccuracy = zvodeaccucorr*accavec
!      else
!         goto 10
!      endif

! --- CALL ZVODE INTEGRATOR WITH DESIRED FUNCTIONS ---

!      if (lzvodecrit) then
!        rwork(1)=time+intperiod
!        itask=4
!      else
!        itask=1
!      endif


!      call get_timer(tid, 'ZVODE')
!      call start_timer(tid)

      ! save current data for restoration if integration fails
      psisave(:)=psi(:)
      asave(1:agmatdim)=agmat(1:agmatdim)
!write(*,*) "psisave init", psisave
!write(*,*) "asave init", asave
      gwpdynloop=1
      do while (gwpdynloop .eq. 1)
!write(*,*) "start runzvode"
!write(*,*) "adim",adim
!write(*,*) "dgldim",dgldim
!write(*,*) "N",N

         error=1
         lastngwpd=sum(ngwpdyn)

      N=max(adim,dgldim)!,lmaxdim*lmaxdim*2+lmaxdim*2)
      if (mode>0) N=max(N,phidim(mode))
!write(*,*) "N",N

         lastN=N
         if (allocated(iwork)) deallocate(iwork)
         if (allocated(rwork)) deallocate(rwork)
         if (allocated(cwork)) deallocate(cwork)
         if (allocated(deviation)) deallocate(deviation)
         liwork=N+30
         allocate(iwork(liwork))
         lrwork=N+20
         allocate(rwork(lrwork))
         lcwork=8*N+2*N**2
         if (lcwork < 0) then
           write(1000,*) lcwork,N 
           stop "lcwork digit problem"
         endif
         allocate(cwork(lcwork))
         allocate(deviation(N))

         rwork(5:10)=0
         iwork(5:10)=0
! --- set maximum iterations and minimum stepsize
         iwork(6)=abs(zvodemaxit)
         rwork(5)=zvode1ststp
         rwork(6)=zvodemaxstp
         rwork(7)=zvodeminstp

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1)) then
            write(istp,'(a)')&
               'Complete equation: ZVODE'
         elseif (evaltype .eq. 4) then
            if (lexact) then
               write(istp,'(a)')&
                    'Exact wavepacket: ZVODE'
            else
               write(istp,'(a,i2,a)') 'Mode', mode,&
                    ': ZVODE'
            endif
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: ZVODE'
         else
            goto 10
         endif
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
              'Step', '#GWP DoFs', 'Stepsize [fs]', 'Error (scaled)',&
              trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .le. 1) then
         zvodeaccuracy = zvodeaccucorr*accuracy
      elseif (evaltype .eq. 4) then
         zvodeaccuracy = zvodeaccucorr*accphi
      elseif (evaltype .eq. 5) then
         zvodeaccuracy = zvodeaccucorr*accavec
      else
         goto 10
      endif

! --- CALL ZVODE INTEGRATOR WITH DESIRED FUNCTIONS ---

      if (lzvodecrit) then
        rwork(1)=time+intperiod
        itask=4
      else
        itask=1
      endif


      call get_timer(tid, 'ZVODE')
      call start_timer(tid)

         ! this check is a bit sloppy i.e. only works as long as there
         ! are no mutually cancelling increases and decreases
         ! if (sum(ngwpdyn) .ne. lastngwpd) then
         !    error=1
         !    lastngwpd=sum(ngwpdyn)
         !    rwork(5)=zvodeminstp/10.0_dop
         ! end if
         if (localrestart) then
            error=1
            if (evaltype .le. 1) then
               laststep(0) = dt
            elseif (evaltype .ge. 4) then
               laststep(0) = dta
               do m=1,nmode
                  laststep(m) = dtphi
               enddo
            endif
            localrestart=.false.
         endif
!         if (evaltype .eq. 4) then
!            rwork(5)=laststep(mode)
!         else
!            rwork(5)=laststep(0)
!         end if

         t0=time
         t1=time+intperiod

         call zerovxd(deviation,N)
         ! warning: no postprocessing, yet. limportho is ignored.
!write(*,*)"evaltype",evaltype
!write(*,*)"limportho",limportho
         if (evaltype .eq. 0) then
            if (limportho) then
               call zvode(funk,dgldim,psi,t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else if(lddmctdh)then
               call zvode(ddmfunk,dgldim,psi,t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else
               call zvode(funk,dgldim,psi,t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            endif
         elseif (evaltype .eq. 1) then
            if (limportho) then
               call zvode(funkr,dgldim,psi,t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else
!write(*,*) " "
!write(*,*) "restartzvode before zvode in source/propwf/integrat.f90", restartzvode
!write(*,*) "N before zvode in source/propwf/integrat.f90", N
!write(*,*) "zvode1ststp before zvode in source/propwf/integrat.f90", zvode1ststp
!write(*,*) "dgldim before zvode in source/propwf/integrat.f90", dgldim
!write(*,*) "psi before zvode in source/propwf/integrat.f90", psi
!write(1000,*) "t0 before zvode in source/propwf/integrat.f90", t0
!write(1000,*) "t1 before zvode in source/propwf/integrat.f90", t1
!write(*,*) "zvodeaccuracy before zvode in source/propwf/integrat.f90", zvodeaccuracy
!write(*,*) "itask before zvode in source/propwf/integrat.f90", itask
!write(*,*) "error before zvode in source/propwf/integrat.f90", error
!write(*,*) "cwork before zvode in source/propwf/integrat.f90", cwork
!write(*,*) "lcwork before zvode in source/propwf/integrat.f90", lcwork
!write(*,*) "rwork before zvode in source/propwf/integrat.f90", rwork
!write(*,*) "lrwork before zvode in source/propwf/integrat.f90", lrwork
!write(*,*) "iwork before zvode in source/propwf/integrat.f90", iwork
!write(*,*) "liwork before zvode in source/propwf/integrat.f90", liwork
!write(*,*) "iflagzvode before zvode in source/propwf/integrat.f90", iflagzvode
!write(*,*) "deviation before zvode in source/propwf/integrat.f90", deviation
!write(*,*) "errmsg before zvode in source/propwf/integrat.f90", errmsg
!               write(1000,*)time/fs,lcwork,N,"before call zvode"
               call zvode(funkr,dgldim,psi,t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,gwpodetest)
!write(1000,*) "t0 after zvode in source/propwf/integrat.f90", t0
!write(1000,*) "t1 after zvode in source/propwf/integrat.f90", t1
!write(*,*) "psi after zvode in source/propwf/integrat.f90", psi
            endif
         elseif (evaltype .eq. 4) then
            if (lexact) then
               if(lddmctdh)then
                  call zvode(ddmfunkex,phidim(mode),psi(zetf(mode,1)),t0,&
                       t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                       cwork,lcwork,rwork,lrwork,iwork,liwork,&
                       0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
               else
                  call zvode(funkex,phidim(mode),psi(zetf(mode,1)),t0,&
                       t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                       cwork,lcwork,rwork,lrwork,iwork,liwork,&
                       0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
               endif
            else if (limportho) then
               call zvode(funkphi,phidim(mode),psi(zetf(mode,1)),t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else if (lddmctdh) then
               call zvode(ddmfunkphi,phidim(mode),psi(zetf(mode,1)),t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else
               call zvode(funkphi,phidim(mode),psi(zetf(mode,1)),t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            endif
         elseif (evaltype .eq. 5) then
            if (lddmctdh) then
               call zvode(ddmfunka,adim,psi(zpsi(1)),t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            else
               call zvode(funka,adim,psi(zpsi(1)),t0,&
                    t1,1,zvodeaccuracy,zvodeaccuracy,itask,error,1,&
                    cwork,lcwork,rwork,lrwork,iwork,liwork,&
                    0,iflagzvode,zvodeout,1,deviation,errmsg,dummypost)
            endif
         else
            goto 10
         endif
         call stop_timer(tid)
         zvodesteps=zvodesteps+IWORK(11)
         ! zvodeaccpt=zvodeaccpt+IWORK(17)
         zvoderejct=zvoderejct+IWORK(23)

         if (evaltype == 4) then
            laststep(mode)=rwork(12)
         else
            laststep(0)=rwork(12)
         end if

!write(*,*)evaltype,zvodemaxit,error,recoverymethod,any(ngwpdyn(:,:)>0),"source/propwf/integrat.F90"
         ! check for ZVODE failure and remove a GWP from dynamics if any are left
         if ((evaltype/=5) .and. (zvodemaxit<0) .and. (error < 0) .and. &
              ((recoverymethod==1) .or. (any(ngwpdyn(:,:) > 0)))) then
!write(*,*)evaltype,zvodemaxit,error,recoverymethod,any(ngwpdyn(:,:)>0),"source/propwf/integrat.F90"
            psi(:)=psisave(:)
            agmat(1:agmatdim)=asave(1:agmatdim)
!write(*,*) "psisave" , psisave
!write(*,*) "asave" , asave
            
!            if ((lBEL) .and. (recoverymethod==1)) then
            if ((lBEL) .and. (error<0) ) then
!               intperiod=0.0_dop
!               if (error == -1) goto 888
               errorBELprop=1
               write(1000,*) 'Integration failed. error=',error
!               write(1000,*)evaltype,zvodemaxit,error, &
!                            recoverymethod,any(ngwpdyn(:,:)>0),"source/propwf/integrat.F90"
               write(1000,*)lcwork,N 
!               write(1000,*) 'Try again in shorter time propagation'
               error=1
               goto 999
!               write(*,*)'i',time
!               call resetGwp(time,psi,agmat,iflag=iflag,cinteg=.true.)
!               call resetGwp(time,psi,agmat,iflag=iflag)
!            dtpsi=(0.0,0.0)
!write(*,*)"dtpsi before funkr in integrat",dtpsi
!            call funkr(time,psi,dtpsi)
!write(*,*)"dtpsi after funkr in integrat",dtpsi
               gwpdynloop=1
               maxdm=1
               maxds=1
               maxdi=1
               CALL WriteStep(zvodesteps,ngwpdyn(maxdm,maxds),time-rwork(13),&
                    deviation(zphi(maxdm,maxds)+maxdi-1),rwork(13),.false.)
               localrestart=.true.
               recoverymethod=2
!write(*,*)'gwpdynloop',gwpdynloop,"source/propwf/integrat.F90"
!intperiod=0.5d0*intperiod
!write(*,*)'intperiod',intperiod,"source/propwf/integrat.F90"
            else
               gwpdynloop=0
               if (evaltype <= 1) then
                  maxdm=1
                  maxds=1
                  do m=1,nmode
                     if (gwpm(m)) then
                        do s=1,nstate
                           call zvodeGwpDynHandle(0,mi(migwpdyn+zgwpdyn(m,s)-1),npg(m),dim(m,s),&
                                deviation(zphi(m,s)),deviation(zphi(maxdm,maxds)),&
                                maxdm,maxds,maxdi,m,s,gwpdynloop)
                        end do
                     end if
                  end do
               else
                  if (gwpm(mode)) then
                     maxdm=mode
                     maxds=1
                     do s=1,nstate
                        call zvodeGwpDynHandle(0,mi(migwpdyn+zgwpdyn(mode,s)-1),npg(mode),dim(mode,s),&
                             deviation(zphi(mode,s)-zetf(mode,1)+1),deviation(zphi(maxdm,maxds)-zetf(mode,1)+1),&
                             maxdm,maxds,maxdi,mode,s,gwpdynloop)
                     end do
                  end if
               end if
               if (gwpdynloop /= 0) then
                  call zvodeGwpDynHandle(1,mi(migwpdyn+zgwpdyn(maxdm,maxds)-1),npg(maxdm),dim(maxdm,maxds),&
                       deviation(zphi(maxdm,maxds)),deviation(zphi(maxdm,maxds)),&
                       maxdm,maxds,maxdi,mode,s,gwpdynloop)
                  if (gwpdynloop/=-1) then
                     CALL WriteStep(zvodesteps,ngwpdyn(maxdm,maxds),time-rwork(13),&
                          deviation(zphi(maxdm,maxds)+maxdi-1),rwork(13),.false.)
                     localrestart=.true.
                  end if
               end if
               recoverymethod=1
            end if
            gwpdynloop=1
            error=1
         else
            gwpdynloop=0
         end if
      end do

      intperiod=rwork(13)-time
      if (lgwp .and. lBEL .and. (error == -1)) then
         error=1
      end if

      ! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---
      if (error < 0) then
         routine = 'RunZVODE'
         message = 'Integration failed.'
         call errormsg
      else
         CALL WriteStep(zvodesteps,ngwpdyn(1,1),rwork(11),0.0_dop,rwork(13),.false.)
      endif

! Deallocation
!      if (allocated(iwork)) deallocate(iwork)
!      if (allocated(rwork)) deallocate(rwork)
!      if (allocated(cwork)) deallocate(cwork)
!      if (allocated(deviation)) deallocate(deviation)


! --- RETURN TO CALLING ROUTINE ---

999   continue !TM
!write(*,*)'intperiod',intperiod
      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunZVODE'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runzvode


! ***************************************************************************
! *     zvodeGwpDynHandle
! *  Wrapper used in runzvode for finding (job==0) or removing (job==1) the
! *  GWP with the highest total contribution to ZVODE integration error.
! *  In case of job==0 (finding), dynloop has to be set to 0 on first
! *     invocation such that maxdm, maxds, and maxdi can be initialized.
! *  In case of job==1, dynloop returns 1 if a GWP was disabled or -1 if not.
! *
! ***************************************************************************

subroutine zvodeGwpDynHandle(job,gwpdyn,npg1,dim1,dev,maxdev,maxdm,maxds,maxdi,m,s,dynloop)

  implicit none

  integer(long),intent(in)                            :: job
  integer(long),intent(in)                            :: npg1,dim1
  integer(long),dimension(npg1-1,dim1), intent(inout) :: gwpdyn
  real(dop),dimension(dim1),intent(in)                :: dev,maxdev
  integer(long),intent(inout)                         :: maxdm,maxds,maxdi
  integer(long),intent(in)                            :: m,s
  integer(long),intent(inout)                         :: dynloop

  integer(long) :: i

  if (job == 0) then
     do i=1,dim1
        if (((dynloop == 0) .or. (dev(i) > maxdev(maxdi))) &
             .and. (any(gwpdyn(:,i) == 0))) then
           dynloop=1
           maxdm=m
           maxds=s
           maxdi=i
        end if
     end do
  else
     i=count(gwpdyn(:,maxdi)==0)
     if (i /= 0) then
        ngwpdyn(maxdm,maxds)=ngwpdyn(maxdm,maxds)-i
        gwpdyn(:,maxdi)=1
        dynloop=1
     else
        dynloop=-1
     end if
  end if
end subroutine zvodeGwpDynHandle


subroutine zvodeout(nst,n,told,t,y,error)
use wrintegrat, only : writestep
use runpropmod, only : ngwpdyn,accuracy
use channels

  integer(long) :: nst,n
  real(dop)     :: t,told
  complex(dop),dimension(n) :: y
  real(dop),dimension(n)    :: error
  complex(dop)              :: dy
  real(dop)                 :: derror

! dummy code to prevent unused variable warning
     dy = y(1)
     derror = error(1)

!write(*,*)"nst",nst
!write(*,*)"n",n
!write(*,*)"told",told
!write(*,*)"t",t
!write(*,*)"y",y
!write(*,*)"error",error
  CALL WriteStep(nst,ngwpdyn(1,1),t-told,accuracy,t,.false.)

  ! if ((t>24/0.02418) .and. (t<28/0.02418)) then
  !    write(6,'(A1,999g16.8)')'P',t,y
  ! end if

end subroutine zvodeout

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNSOD                        *
! *                                                                    *
! * Drives the second order differencing integration routines,         *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector at t.             *
! *   dtpsi:      Action of Hamiltonian on the initial-value vector.   *
! *   oldpsi:     Wavefunction at t-dt                                 *
! *   time:       Present time at start of step                        *
! *   intperiod:  Length of time interval to be integrated.            *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *                                                                    *
! * Eryn Spinlove 7/16
! * SOD - CALL TO THE ACTUAL INTEGRATOR(SODLIB) MADE HERE              * 
! **********************************************************************

      subroutine runsod (psi,dtpsi,oldpsi,time,intperiod) 
                
      implicit none

      real(dop), parameter                           :: sodaccucorr = 1.0_dop

      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi,oldpsi
      real(dop), save                                :: stepsize,stepsafety
      real(dop), intent(in)                          :: intperiod
      real(dop), intent(in)                          :: time
      real(dop)                                      :: norm
      integer(long)                                  :: order,error

      logical(kind=4), save                          :: linit=.true.
      integer(long), save                            :: tid=0 ! timer ID

      call get_timer(tid, 'SOD')
      call start_timer(tid)

!-----------------------------------------------------------------------
! dtpsi is -iH|psi>. For SOD need H|psi>
! convert initial dtpsi to Hpsi using ldtpsihpsi flag
!-----------------------------------------------------------------------
      ldtpsi2hpsi=.true.
      call prefactor(dtpsi,dgldim)
      ldtpsi2hpsi=.false.

!-----------------------------------------------------------------------
! Setting lfuncnopre means funkex routine returns Hpsi rather than dtpsi. 
! This is required for SOD integrator.
!-----------------------------------------------------------------------
      lfuncnopre = .true.

!-----------------------------------------------------------------------
! In call to SODSTEP
!
!      psi(zetf(1,1))   - wavefunction at t - returns wf @ t+dt
!      dtpsi(zetf(1,1)) - HPsi @ t
!      oldpsi(zetf(1,1))- Psi @ t-dt
!      time             - time at start of step
!      totphidim        - grid length
!      intperiod        - time length of integration period
!      stepsize         - stepsize for SOD integrator
!      funkex           - calcs Hpsi @ new time
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! initial call 
!-----------------------------------------------------------------------
      if (linit) then
! 1. Compute a stepsize for which the SOD-integrator is stable.
         stepsafety = dtphi
         call sodsize(intperiod,stepsafety,maxeigval,stepsize)
         write(ilog,'(a)') 'SOD integrator :'
         write(ilog,'(a,f12.6)')   'Stepsafety         : ',stepsafety
         write(ilog,'(a,e16.6,a)') 'Maximum eigenvalue : ',maxeigval*ev,' eV'
         write(ilog,'(a,f12.6,a)') 'Stepsize           : ',stepsize/fs,' fs'

! 2. psi (t0 - delta t) to provide the (two-step) SOD method with a 
!    second initial wavefunction
! accphi and intorderphi can be changed using SOD keyword
         order=intorderphi
         call initsod(psi(zetf(1,1)),phidim(1),time,stepsize,accphi,funkex,&
               normvxz,oldpsi(zetf(1,1)),dtpsi(zetf(1,1)),order,norm,error)
         write(ilog,'(a,i5)')    'Initial step converged order : ',order
         write(ilog,'(a,f12.8)') 'Norm                         : ',norm
         if (error .eq. 1) then
            write(ilog,'(a)') &
                'Failed to converge. Try a smaller safetystep size'
         endif

! 3. integrate remaining interval from time+stepsize to time+intperiod
         call sodstep(psi(zetf(1,1)),oldpsi(zetf(1,1)),dtpsi(zetf(1,1)),&
              phidim(1),time+stepsize,intperiod-stepsize,stepsize,funkex)


         linit = .false.

      else

! SOD -----------------------------------------------------------------------
! --- CALL SOD ROUTINE TO INTEGRATE OVER INTPERIOD  ---
! --------------------------------------------------------------------------
          call sodstep(psi(zetf(1,1)),oldpsi(zetf(1,1)),dtpsi(zetf(1,1)),&
               phidim(1),time,intperiod,stepsize,funkex)

      endif

      lfuncnopre = .false.

      call stop_timer(tid)
      return

     end subroutine runsod

!#######################################################################
! *                                                                    *
! *                           SUBROUTINE RUNSPO                        *
! * Drives the split operator integration routine                      *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   hops:       The Hamiltonian operator                             *
! *   fftp:       Momentum on the momentum grid                        *
! *   time:       Time at start of step                                *
! *   hin, rueck: Forward and backward FFT tansfornations              *
! *   fftfak:     Factors used in FFT decomposition                    *
! *   intperiod:  Length of time interval to be integrated.            *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *                                                                    *
! * Marcus Taylor 7/16                                                 *
! *                                                                    *
! SPO - CALL TO THE ACTUAL INTEGRATOR(SPOLIB) MADE HERE 
! **********************************************************************

      subroutine runspo (psi,hops,fftp,intperiod,hin,rueck,fftfak) 
                
      implicit none

      real(dop), parameter                           :: sodaccucorr = 1.0_dop

      complex(dop), dimension(dgldim), intent(inout) :: psi
      real(dop), save                                :: stepsize,stepsafety
      real(dop), intent(in)                          :: intperiod
      integer(long), dimension(fftdim), intent(in)   :: fftfak
      real(dop), dimension(hopsdim), intent(in)      :: hops
      real(dop), dimension(fftdim), intent(in)       :: fftp
      complex(dop), dimension(fftdim), intent(in)    :: hin,rueck

      integer(long)                                  :: h,k

      logical(kind=4), save                          :: linit=.true.
      integer(long), save                            :: tid=0 ! timer ID

      call get_timer(tid, 'SPO')
      call start_timer(tid)

!-----------------------------------------------------------------------
! initial call 
!-----------------------------------------------------------------------
      if (linit) then
! 1. Compute a stepsize for which the SPO-integrator is stable?
        stepsafety = dtphi
! find PE
         h=0
         do k=1,khtot(1)
            if (kinop(1,k) .eq. 0) h=hamilton(1,k)
         enddo
         call spoinit(stepsize,fftp,hops(zhop(h)),mass,ndof,gdim,zfft,&
              fftdim,phidim(1),maxeigval,intperiod,stepsafety)
         write(ilog,'(a)') 'SPO integrator :'
         write(ilog,'(a,f12.6,a)') 'Stepsize           : ',stepsize/fs,' fs'

         linit = .false.

      endif

! --------------------------------------------------------------------------
!
! --- CALL SPO ROUTINE TO INTEGRATE iOVER INTPERIOD  ---
!
!      SO kind of IN THE ABOVE (TO BE REWRITTEN) WE HAVE
!      psi(zetf(1,1))   - wavefunction at t - returns wf @ t+dt
!      dtpsi(zetf(1,1)) - time derivative of wf @ t
!      totphidim        - grid length
!      stepsize         - SPO stepsize
!      intperiod        - time period for integration step
! --------------------------------------------------------------------------
      call spostep(psi(zetf(1,1)),phidim(1),intperiod,stepsize,vgdim,&
           gdim,ngdim,hin,rueck,fftfak,zfft,fftdim,ndof)

      call stop_timer(tid)
      return

     end subroutine runspo

!#######################################################################
! *                                                                    *
! *                           SUBROUTINE RUNCHEB                       *
! *                                                                    *
! *    Runs the Chebyshev Integration scheme                           *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:        The (complex) initial-value vector.                  *
! *   dtpsi:      Action of Hamiltonian on the initial-value vector.   *
! *   intperiod:  Length of time interval to be integrated.            *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:        Propagated wavefunction.                             *
! *                                                                    *
! * inttime: integrated time in step. Starts at 0. Ends at intperiod   *
! CHEB - CALL TO THE ACTUAL INTEGRATOR(CHEBLIB) MADE HERE 
! **********************************************************************

      subroutine runcheb (psi,dtpsi,intperiod)

      implicit none

!      real(dop), parameter                          :: sodaccucorr = 1.0_dop
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      real(dop), intent(in)                          :: intperiod
      integer(long), save                            :: tid=0 ! timer ID


      call get_timer(tid, 'CHEB')
      call start_timer(tid)

!-----------------------------------------------------------------------
! dtpsi is -iH|psi>. For CHEB need H|psi>
! convert dtpsi to Hpsi using ldtpsihpsi flag
!-----------------------------------------------------------------------
      ldtpsi2hpsi=.true.
      call prefactor(dtpsi,dgldim)
      ldtpsi2hpsi=.false.

!-----------------------------------------------------------------------
! Setting lfuncnopre means funkex routine returns Hpsi rather than dtpsi. 
! This is required for CHEB integrator.
!-----------------------------------------------------------------------

      lfuncnopre = .true.

! CHEB-----------------------------------------------------------------------
!
! --- CALL CHEB ROUTINE TO INTEGRATE iOVER INTPERIOD  ---
!
!      SO kind of IN THE ABOVE (TO BE REWRITTEN) WE HAVE
!      psi(zetf(1,1))   - wavefunction at t - returns wf @ t+dt
!      dtpsi(zetf(1,1)) - time derivative of wf @ t
!      totphidim        - grid length
!      stepsize/intorderphi 
!                       - variables specifying integration parameters
!      funkex           - calcs dtpsi @ new time
! --------------------------------------------------------------------------

      call chebstep(psi(zetf(1,1)),dtpsi(zetf(1,1)),intperiod,phidim(1),&
           maxeigval,mineigval,funkex,accphi)

      lfuncnopre = .false.

      call stop_timer(tid)
      return

     end subroutine runcheb

! End of CAsh edit

! **********************************************************************
! *                                                                    *
! *                           SUBROUTINE RUNVERLET                     *
! *                                                                    *
! * Drives the verlet integration routine.                             *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runverlet (psi,dtpsi,intperiod,time,evaltype,mode)

use griddatmod
use psidef
use datenmod
use runpropmod
use channels
use rklib

use importhomod, only: importhovmf,importho,dummypost
use functiontr, only: funktr1

use verletlib, only: verlet

      implicit none

      integer(long), intent(in)                      :: evaltype,mode
      integer(long)                                  :: m,s
      real(dop), intent(in)                          :: intperiod,time
      real(dop)                                      :: timestep
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      integer(long)                                  :: tid=0 ! timer ID

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         write(istp,'(a)') 'Verlet'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
              'Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)',&
              'Time [fs]'
         call myflush(istp)
      endif


! --- CALL VERLET INTEGRATOR WITH DESIRED FUNCTIONS ---
      if (evaltype .eq. 8) then
         m=mode
         call get_timer(tid, 'Verlet')
         call start_timer(tid)
         timestep = intperiod
         do s=1,nstate
            call verlet(psi(zphi(m,s)),dtpsi(zphi(m,s)),nspfdof(m),gwpdim(m,s),intperiod,&
                        time,timestep,funkgbf)
         enddo
         call stop_timer(tid)
      else
         go to 10
      endif

! --- RETURN TO CALLING ROUTINE ---

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunVerlet'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runverlet

! **********************************************************************
! *   HEREHEREHERE 15/11/17                                            *
! *                           SUBROUTINE RUNRK5F                       *
! *                                                                    *
! * Drives the RK5F (fixed stepsize) integration routine.              *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! *   time:          Absolute time, i. e. Psi_initial=Psi(AbsTime).    *
! *   evaltype:      Specifies the evaluation type, i. e. the "funk"   *
! *                  routine.                                          *
! *   mode:          The mode to be integrated when only               *
! *                  single-particle functions are propagated.         *
! *                                                                    *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
! *   dtpsi:         dtpsi is undetermined.                            *
! *                                                                    *
! * Other parameters:                                                  *
! *   mc:            Complex*16 data needed in the Func (and possibly  *
! *                  CalcError) routine.                               *
! *   mr:            Real*8 data needed in the Func (and possibly      *
! *                  CalcError) routine.                               *
! *   mi:            Integer data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *   ml:            Logical data needed in the Func (and possibly     *
! *                  CalcError) routine.                               *
! *                                                                    *
! * V7.0 MB                                                            *
! *                                                                    *
! **********************************************************************

      subroutine runrk5f (psi,dtpsi,intperiod,time,evaltype,mode,restartrk5)

      implicit none

      real(dop), parameter                           :: rk5accucorr = 1.0_dop

      integer(long)                                  :: error,m,z0,zf
      integer(long), intent(in)                      :: evaltype,mode
      real(dop)                                      :: rk5accuracy
      real(dop), dimension(:), allocatable, save     :: laststep
      real(dop), intent(in)                          :: intperiod,time
      complex(dop), dimension(dgldim), intent(inout) :: psi,dtpsi
      logical(kind=4), save                          :: initialized = .false.
      logical(kind=4), intent(in)                    :: restartrk5
      integer(long), save                            :: tid=0 ! timer ID
      character(len=c2) :: string
     
      if (.not. allocated(laststep)) allocate(laststep(0:maxdim))

! --- INITIALIZATION ---

      if (restartrk5) then
         initialized = .false.
      !   dt = 0.0_dop
      endif
!     is this the first call?
      if (.not. initialized) then
!        yes: initialize the stepsizes
         if (evaltype .le. 1) then
!           VMF integrations
            laststep(0) = dt
         elseif (evaltype .ge. 4) then
!           CMF integrations
            laststep(0) = dta
            do m=1,nmode
               laststep(m) = dtphi
            enddo
         else
            goto 10
         endif
         initialized = .true.
      endif

! --- CREATE HEADLINE IN INTSTEP FILE IF DESIRED ---

      if (lintstep) then
         if ((evaltype .eq. 0) .or. (evaltype .eq. 1) .or. &
             (evaltype .eq. -1) .or. (evaltype .eq. 6)) then
            write(istp,'(a)')&
               'Complete equation: Runge-Kutta order 5'
         elseif (evaltype .eq. 4) then
            if (lexact) then
               write(istp,'(a)')&
                    'Exact wavepacket: Runge-Kutta order 5'
            else
               write(istp,'(a,i2,a)') 'Mode', mode,&
                    ': Runge-Kutta order 5'
            endif
         elseif (evaltype .eq. 7) then
            write(istp,'(a,i2,a)') 'Node', currentnode%num,&
                    ': Runge-Kutta order 5'
         elseif (evaltype .eq. 5) then
            write(istp,'(a)') 'A: Runge-Kutta order 5'
         else
            goto 10
         endif
         write(string,'(3a)') 'Time [',trim(rlxtime),']'
         write(istp,'(a7,8x,a9,4x,a13,6x,a14,7x,a9)')&
              'Step', 'Accepted?', 'Stepsize [fs]', 'Error (scaled)',&
              trim(string)
         call myflush(istp)
      endif

! --- CORRECT THE ACCURACY WITH RESPECT TO OTHER INTEGRATORS ---

      if (evaltype .le. 1) then
         rk5accuracy = rk5accucorr*accuracy
      elseif (evaltype .eq. 4) then
         rk5accuracy = rk5accucorr*accphi
      elseif (evaltype .eq. 5) then
         rk5accuracy = rk5accucorr*accavec
      elseif (evaltype .eq. 6) then
         rk5accuracy = rk5accucorr*accphi
      elseif (evaltype .eq. 7) then
         rk5accuracy = rk5accucorr*accphi
      else
         goto 10
      endif

! --- CALL RK5 INTEGRATOR WITH DESIRED FUNCTIONS ---

      call get_timer(tid, 'RK5')
      call start_timer(tid)

      if (evaltype .eq. 0) then
         if(lddmctdh)then
            call rk5f(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),ddmfunk,&
                 liork, importhovmf,lrktest,gwpodetest)
         else
            call rk5f(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),funk,&
                 liork, importhovmf,lrktest,gwpodetest)
         endif
      elseif (evaltype .eq. 1) then
         call rk5f(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),funkr,&
                 liork, importhovmf,lrktest,gwpodetest)
      elseif (evaltype .eq. -1) then
         call rk5f(psi,dtpsi,dgldim,intperiod,time,laststep(0),minrk5,&
                 rk5accuracy,error,mc(mcintpsi),funkr,&
                 liork, importhovmf,lrktest,gwpodetest)

      elseif (evaltype .eq. 4) then
         if (lexact) then
            if(lddmctdh)then
               call rk5f(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                    intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                    mc(mcintpsi),ddmfunkex,.false.,dummypost,&
                    .false.,dummypost)
            else
               call rk5f(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                    intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                    mc(mcintpsi),funkex,.false.,dummypost,&
                    .false.,dummypost)
            endif
         else if(lddmctdh)then
            call rk5f(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                 intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                 mc(mcintpsi),ddmfunkphi,liork,importho,lrktest,gwpodetest1m)
         else
            call rk5f(psi(zetf(mode,1)),dtpsi(zetf(mode,1)),phidim(mode),&
                 intperiod,time,laststep(mode),minrk5,rk5accuracy,error,&
                 mc(mcintpsi),funkphi,liork,importho,lrktest,gwpodetest1m)
         endif
      elseif (evaltype .eq. 5) then
         if (lddmctdh) then
            call rk5f(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                     ddmfunka,.false.,dummypost,.false.,dummypost)
         else
            call rk5f(psi(zpsi(1)),dtpsi(zpsi(1)),adim,intperiod,time, &
                     laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),funka, &
                     .false.,dummypost,.false.,dummypost)
         endif
      elseif (evaltype .eq. 6) then
! ML-MCTDH prop of whole WF tree
         call rk5f(psi,dtpsi,dgldim,intperiod,time, &
                 laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                 funkml_dt,liork, importhovmf,lrktest,gwpodetest)
      elseif (evaltype .eq. 7) then
! ML-MCTDH prop of one SPF from middle layers
         z0 = currentnode%z(1,1)
         zf = currentnode%z(currentnode%navecs,2)
 
         call rk5f(psi(z0),dtpsi(z0),zf-z0+1,intperiod,time, &
                 laststep(0),minrk5,rk5accuracy,error,mc(mcintpsi),&
                 funkphinode,liork, importho,lrktest,gwpodetest)
      else
         goto 10
      endif
      call stop_timer(tid)

! --- WRITE ERROR MESSAGE IF INTEGRATION FAILED ---

      if (error .ne. 0) then
         routine = 'RunRK5F'
         message = 'Integration failed.'
         call errormsg
      endif

! --- RETURN TO CALLING ROUTINE ---

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'RunRK5F'
      message = 'Illegal evaluation type'
      call errormsg

      end subroutine runrk5f

!***********************************************************************
! One-step Euler integrator for use in testing aspects of the exact 
! method. Not to be used for serious calculations.
! GWR 02/19
! * Input parameters:
! *   psi:           The (complex) initial-value vector.               *
! *   dtpsi:         Time derivative of the initial-value vector.      *
! *   intperiod:     Length of time interval to be integrated.         *
! * Output parameters:                                                 *
! *   psi:           Propagated wavefunction.                          *
!***********************************************************************
      subroutine runeuler(psi,dtpsi,intperiod)

      implicit none

      real(dop), intent(in)                          :: intperiod
      complex(dop), dimension(dgldim), intent(inout) :: psi
      complex(dop), dimension(dgldim), intent(in)    :: dtpsi

      if(intperiod.ne.out1)then
         routine = 'Euler'
         message = 'Time-step should be equal to output time.'
         call errormsg
      endif

      call euler(psi,dtpsi,dgldim,intperiod)

! Renormalise the wavefunction as the integrator does not ensure this.
      call renormphi(psi(zetf(1,1)))

      return
      end subroutine runeuler

! **********************************************************************
! * Input parameters:                                                  *
! *   psi1:   For ABM the predicted psi-vector, for BS a dummy value.  *
! *   psi2:   For ABM the corrected psi-vector, for BS a vector        *
! *           containing the error of each psi-component.              *
!
! error is sqrt( || psi1 - psi2||^2)
! not in use!
!#######################################################################

      subroutine wferror (psi1,psi2,psidim,error)

      implicit none

      integer(long), intent(in)                   :: psidim
      real(dop)                                   :: normsq1,normsq2
      real(dop), intent(out)                      :: error
      complex(dop), dimension(psidim), intent(inout) :: psi1,psi2
      complex(dop), dimension(agmatdim)           :: agmat2
      complex(dop)                                :: ovl,ovl1,ovl2

      call cpvxz(agmat,agmat2,agmatdim)
      call overlap_err(ovl,ovl1,ovl2,psi1,psi2,agmat,agmat2)

      normsq1=dble(ovl1)
      normsq2=dble(ovl2)

      error=normsq1+normsq2-2.0_dop*dble(ovl)
      error=max(error,1.0e-16_dop)
      if (error .lt. 0.0_dop) then
         if ((error*(-1.0_dop)).gt.tiny) &
              write(6,*) ' WARNING: Error is negative ! '
         error=0.0_dop
      endif

      error = dsqrt(error)

      return
      end subroutine wferror

!#######################################################################

      subroutine overlap_err(ovl,ovl1,ovl2,psi1,psi2,agmat1,agmat2)

      implicit none

      integer(long)                                  :: s,m,zeig1,zeig2,swapzeig
      complex(dop)                                   :: zdum
      complex(dop), intent(out)                      :: ovl1,ovl2,ovl
      complex(dop), dimension(1)                     :: cdum,cdum1
      complex(dop), dimension(dgldim), intent(inout) :: psi1,psi2
      complex(dop), dimension(dmatdim)               :: dover,ovr1,ovr2
      complex(dop), dimension(agmatdim), intent(out) :: agmat1,agmat2
      complex(dop), dimension(2*maxblock)            :: workc

!-----------------------------------------------------------------------
!     calculate density matrices for each degree of freedom defined as
!     overlap of the single-particle functions 
!     rho^m_ij=<phi_i,2|phi_j,1>
!-----------------------------------------------------------------------

      do s=1,nstate
         do m=1,nmode
            if (gwpm(m)) then
! for GWPs, get agmat for both sets of GWPs the ncalculate overlaps
               call param_from_psi(nspfdof(m),gwpdim(m,s), &
                    agmat1(zagmat(m,s)),psi1(zpg(m,s)),psi1(zpg(m,s)),&
                    psi1(zphi(m,s)),agmat1(zphase(m,s)),agmat1(zxivec(m,s)),&
                    1,s,m)

               call param_from_psi(nspfdof(m),gwpdim(m,s), &
                    agmat2(zagmat(m,s)),psi2(zpg(m,s)),psi2(zpg(m,s)),&
                    psi2(zphi(m,s)),agmat2(zphase(m,s)),agmat2(zxivec(m,s)),&
                    1,s,m)

               call diagonal_moments(m,s,s,nspfdof(m),gwpdim(m,s),gwpdim(m,s), &
                    agmat1(zagmat(m,s)),agmat1(zagmat(m,s)), &
                    agmat1(zxivec(m,s)),agmat1(zxivec(m,s)), &
                    agmat1(zphase(m,s)),agmat1(zphase(m,s)), & ! phase mod yapol
                    ovr1(dmat(m,s)),cdum,cdum1,0,0)

               call diagonal_moments(m,s,s,nspfdof(m),gwpdim(m,s),gwpdim(m,s), &
                    agmat2(zagmat(m,s)),agmat2(zagmat(m,s)), &
                    agmat2(zxivec(m,s)),agmat2(zxivec(m,s)), &
                    agmat2(zphase(m,s)),agmat2(zphase(m,s)), & ! phase mod yapol
                    ovr2(dmat(m,s)),cdum,cdum1,0,0)

               call diagonal_moments(m,s,s,nspfdof(m),gwpdim(m,s),gwpdim(m,s), &
                    agmat1(zagmat(m,s)),agmat2(zagmat(m,s)), &
                    agmat1(zxivec(m,s)),agmat2(zxivec(m,s)), &
                    agmat1(zphase(m,s)),agmat2(zphase(m,s)), & ! phase mod yapol
                    dover(dmat(m,s)),cdum,cdum1,2,0)

            else
               call mmaxzzh(psi1(zetf(m,s)),psi1(zetf(m,s)), &
                    ovr1(dmat(m,s)),subdim(m),dim(m,s))
               call mmaxzzh(psi2(zetf(m,s)),psi2(zetf(m,s)), &
                    ovr2(dmat(m,s)),subdim(m),dim(m,s))
               call mmaxzz(psi1(zetf(m,s)),psi2(zetf(m,s)), &
                    dover(dmat(m,s)),subdim(m),dim(m,s), &
                    dim(m,s))
            endif
         enddo
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J|Phi_L>A_L within each state for psi1
!-----------------------------------------------------------------------
      ovl1=(0.0_dop,0.0_dop)
      do s=1,nstate
         zeig1=1
         zeig2=block(s)+1
         call cpvxz(psi1(zpsi(s)),workc(zeig2),block(s))

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call qtxxzz(ovr1(dmat(m,s)),workc(zeig1),workc(zeig2), &
                        vdim(m,s),dim(m,s),ndim(m,s))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J^* Sum_L <Phi_J|Phi_L>A_L within each state (psi1)
!-----------------------------------------------------------------------
         call vvaxzz(psi1(zpsi(s)),workc(zeig2),zdum,block(s))
         ovl1=ovl1+zdum
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J|Phi_L>A_L within each state (for psi2)
!-----------------------------------------------------------------------
      ovl2=(0.0_dop,0.0_dop)
      do s=1,nstate
         zeig1=1
         zeig2=block(s)+1
         call cpvxz(psi2(zpsi(s)),workc(zeig2),block(s))
         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call qtxxzz(ovr2(dmat(m,s)),workc(zeig1),workc(zeig2),&
                        vdim(m,s),dim(m,s),ndim(m,s))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J^* Sum_L <Phi_J|Phi_L>A_L within each state (psi2)
!-----------------------------------------------------------------------
         call vvaxzz(psi2(zpsi(s)),workc(zeig2),zdum,block(s))
         ovl2=ovl2+zdum
      enddo

!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      ovl=(0.0_dop,0.0_dop)
      do s=1,nstate
         zeig1=1
         zeig2=maxblock+1
         call cpvxz(psi2(zpsi(s)),workc(zeig2),block(s))

         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            call mtxxzz(dover(dmat(m,s)),workc(zeig1),&
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s),&
                    dim(m,s))
         enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* Sum_L <Phi_J,1|Phi_L,2>A_L,2 within each state
!-----------------------------------------------------------------------
         call vvaxzz(psi1(zpsi(s)),workc(zeig2),zdum,block(s))
         ovl=ovl+zdum
      enddo

      return
      end subroutine overlap_err


! **********************************************************************
! *                                                                    *
! *                         SUBROUTINE GWPERROR                        *
! *                                                                    *
! * Estimates for both ABM and BS the error of the current integration *
! * step including a weighting to improve the efficiency for GWP       *
! * propagation.                                                       *
! *                                                                    *
! * Input parameters:                                                  *
! *   psi1:   For ABM the predicted psi-vector, for BS a dummy value.  *
! *   psi2:   For ABM the corrected psi-vector, for BS a vector        *
! *           containing the error of each psi-component.              *
! *   psidim: Length of psi1 and psi2.                                 *
! *   mc    : Complex data containing dicht2; needed for weighting     *
! *           of error.                                                *
! *                                                                    *
! * Output parameters:                                                 *
! *   error:  The estimated error of the current integration step.     *
! *                                                                    *
! * Other parameters:                                                  *
! *   rdata:  Real data array (not used in this error routine).        *
! *                                                                    *
! * V6.0 MB                                                            *
! * V7.0 GW addition of data arrays                                    *
! *                                                                    *
! not in use!
! **********************************************************************

      subroutine gwperror1 (psi1,psi2,psidim,error)

      implicit none

      integer(long)                               :: s,b,m,e,g
      integer(long), intent(in)                   :: psidim
      real(dop)                                   :: sum,error1,error2
      real(dop), intent(out)                      :: error
      complex(dop)                                :: diff
      complex(dop), dimension(psidim), intent(in) :: psi1,psi2
      integer(long), save                         :: cnt

! --- INITIALISE VARIABLES ---
 
      error = 0
      cnt=0

! --- LOOP OVER EACH STATE ---

      do s = 1,nstate

!    --- ERROR OF MCTDH-COEFFICIENTS ---

         do b = 1,block(s)
            if (inttyp .eq. 1) then
               diff = psi1(zpsi(s)+b-1)-psi2(zpsi(s)+b-1)
            elseif (inttyp .eq. 2) then
               diff = psi2(zpsi(s)+b-1)
            else
               goto 10
            endif
            error = error+dble(diff*dconjg(diff))
         enddo
         error1=error
         error2=error

!    --- ERROR OF SINGLE-PARTICLE FUNCTIONS ---

         do m = 1,nmode
            do e = 1,dim(m,s)
               sum = 0
               do g = 1,subdim(m)
                  if (inttyp .eq. 1) then
                     diff = psi1(zetf(m,s)+(e-1)*subdim(m)+g-1)&
                           -psi2(zetf(m,s)+(e-1)*subdim(m)+g-1)
                  elseif (inttyp .eq. 2) then
                     diff = psi2(zetf(m,s)+(e-1)*subdim(m)+g-1)
                  else
                     goto 10
                  endif
                  sum = sum+dble(diff*dconjg(diff))
               enddo
               error = error+sum
               error1 = error1+sum&
                       *dble(dicht2(dmat(m,s)+(e-1)*dim(m,s)+e-1))
               error2 = error2+sum&
                      *(1.0_dop-dble(dicht2(dmat(m,s)+(e-1)*dim(m,s)+e-1)))
            enddo
         enddo
      enddo
      error = sqrt(error)
      error1 = sqrt(error1)
      error1=error/error1
      error2 = sqrt(error2)

      return

! --- WRITE ERROR MESSAGE ---

 10   continue
      routine = 'gwperror1'
      message = 'illegal integration type'
      call errormsg

      end subroutine gwperror1

! **********************************************************************
! *                      SUBROUTINE VERLET                             *
! *                    
! **********************************************************************
      

    end module integrat

