!--------------------------------------------------------------------------
!                        MODULE ORBEN.F
! Computes orbital energies and, if energyorb is set, transforms the
! WF to energy-orbitals.
! 
! Contains:
! orben, e_funkphi, orbenergies, e_hlochphi1m, dwtehlochphi1mpar
! e_calcha, dwtecalcha, wkorben
!
! **********************************************************************
      module orbenmod

use decimal, only: long, dop
use global
use globmemmod
use logdat
use xvlib,   only: xvxxzza, xmxxzza, vvaxzz
use mmlib,   only: mmaxzz, mmaxzzh, dmxxzza
use op1lib,  only: zerovxz, zerovxl, zerovxd, trqxz, trtxz, cpqxz
use op2lib,  only: sum2qxxz, addmxxzo, addvxxzo
use timing
use openmpmod

      implicit none
      
      private
      public :: orben,e_funkphi,wkorben
      contains

! **********************************************************************
! * Driver routine                                                     *
! **********************************************************************

      subroutine orben(time,psi)

      use griddatmod
      use psidef
    
      use runpropmod
      use channels

      use trafonatmod, only: trafonat
      use function, only: funk2

      implicit none

      complex(dop), dimension(dgldim), intent(out) :: psi

      integer(long)         :: mode
      real(dop), intent(in) :: time
      integer(long), save                                  :: tid=0 ! timer ID

      if(test0) return
      call get_timer(tid, 'Orben')
      call start_timer(tid)

      do mode = 1, nmode
         if(idmode(mode).eq.0 .and. mode.ne.mpb .and. mode.ne.meb) then
            call e_funkphi(time,psi,mode)
         endif
      enddo
      call myflush(iorb)
      if(lenorb) then
         call trafonat(psi,tree,mc(mcintpsi),mi(mijindx))
         call funk2(time,psi)
      endif
      call stop_timer(tid)

      return
      end subroutine orben

! **********************************************************************
! *                                                                    *
! *                        SUBROUTINE E_FUNKPHI                        *
! *                                                                    *
! * Computes the orbital energies.                                     *
! * Computes for one mode the matrix <phi(i)| H-av | phi(j)>,          *
! * where H-av denotes the trace over the mean fields.                 *
! *                                                                    *
! *                                                                    *
! * Called from: output                                                *
! * Calls to:    e_calcha, e_hlochphi1m, orbenergies                   *
! *                                                                    *
! *                                                                    *
! * 12/01 HDM                                                          *
! *                                                                    *
! **********************************************************************
      subroutine e_funkphi(time,psi,mode)

use griddatmod
use psidef
use operdef
use runpropmod
use mpidata

      implicit none

      integer(long)                                              :: ktype,m,s,pointer,pointer1
      integer(long)                                              :: ldim
      integer(long), intent(in)                                  :: mode
      real(dop)                                                  :: sume,summ
      real(dop), intent(in)                                      :: time
      complex(dop)                                               :: sum,sumc
      complex(dop), dimension(dgldim), intent(in)                :: psi
      complex(dop), dimension(:), allocatable                    :: dtpsi
      complex(dop), dimension(:), allocatable                    :: workc

      ldim = max(sdim,maxspf*maxspf)
      allocate(workc(ldim))
      ldim = max(adim,maxphidim)
      allocate(dtpsi(ldim))
      workc=0.0_dop
      dtpsi=0.0_dop

! --- INITIALISE VARIABLES --- 

      call zerovxz(dtpsi,maxphidim)
      if (dentype .gt. 0) then
         message='Orbital energies not implemented for densities'
         call errormsg
      endif

      if(lconst) then
         ktype=1  
      else
         ktype=0  
      endif

! --- COMPUTE ENERGY TERMS IGNORED IN THE MEAN FIELDS BECAUSE OF diag ---

      call e_calcha (psi,dtpsi,hteil,1,mode,psi(zetf(1,1)),mi(mijindx))

      call vvaxzz(psi,dtpsi,sum,adim)
      sume = dble(sum)

! --- MULTIPLY H|PHI> WITH (CONSTANT) MEAN-FIELD (KINETIC TERMS) ---

#ifdef MPI
      if (lmpi .and. lmpihlochphi .and. lmpistarted) then
         call e_hlochphi1mpar(psi,dtpsi,mc(mchpsi),mc(mchloch),mode, &
                              phidim(mode),1,ktype)
      else
#endif
         call e_hlochphi1m(psi,dtpsi,mc(mchpsi),mc(mchloch),mode, &
                           phidim(mode),1,ktype)
#ifdef MPI
      endif
#endif

! --- MULTIPLY |PHI> WITH SUM OVER MEAN-FIELDS (POTENTIAL TERMS) ---

      if(ktype.eq.1) then
         do s = 1,nstate
            pointer = zetf(mode,s)-zetf(mode,1)+1
            call trtxz(mfsum(zmfsum(mode,s,s)), &
                 workc,subdim(mode),dim(mode,s))
            call dmxxzza (workc,psi(zetf(mode,s)),dtpsi(pointer), &
                 subdim(mode),dim(mode,s))
         enddo
      endif

! --- LOOP OVER EACH STATE ---

      do s = 1,nstate

! --- Calculate uncorrelated energies for all but the mode 'mode':
         summ=0.0_dop
         do m=1,nmode
            pointer1 = zetf(m,s)-adim
            if(khzahl(m,s,1).gt.0) then
               call mmaxzzh(psi(zetf(m,s)),mc(mchunpsi+pointer1-1),  &
                    workc,subdim(m),dim(m,s))
               call sum2qxxz(dicht2(dmat(m,s)),workc,sum,dim(m,s))
               if(m.eq.mode) then
                  summ=dble(sum)
               else
                  sume=sume+dble(sum)
               endif
            endif
         enddo

!     --- CALCULATE ACTION OF UNCORRELATED HAMILTONIAN ON SPF'S ---

         pointer = zetf(mode,s)-zetf(mode,1)+1
         pointer1 = zetf(mode,s)-adim
         
!     --- ADD UNCORRELATED MOTION TO TIME DERIVATIVE VECTOR ---
 
         if( khzahl(mode,s,1) .gt. 0 ) then
            if(lreflex) then
               sumc=2.d0
               call xmxxzza(sumc,mc(mchunpsi+pointer1-1),dtpsi(pointer),&
                    subdim(mode),dim(mode,s))
            elseif(npackts.eq.1) then
               call addmxxzo(mc(mchunpsi+pointer1-1),dtpsi(pointer),&
                    subdim(mode),dim(mode,s))
            else
               sumc=dble(npackts)
               call xmxxzza(sumc,mc(mchunpsi+pointer1-1),&
                    dtpsi(pointer),subdim(mode),dim(mode,s))
            endif
         endif

!     --- DIAGONALIZE THE MATRIX ---
         
         call orbenergies(psi(zetf(mode,s)),dtpsi(pointer), &
              workc,sume,summ,dicht2(dmat(mode,s)), &
              time,mode,s,dim(mode,s),subdim(mode))

!........Copy orben-eigenvectors to intpsi
         if(lenorb)  call cpqxz(workc, &
               mc(mcintpsi+dmat(mode,s)-1),dim(mode,s))

      enddo
      deallocate(workc,dtpsi)
      
      return
      end subroutine e_funkphi


! ------------------------------------------------------------------------


      subroutine orbenergies(psi,dtpsi,workc2,sume,summ, &
                             dicht2,time,m,s,dim1,subdim1)

use psidef
use channels

      implicit none

      integer(long)                                     :: i,j,k
      integer(long), intent(in)                         :: m,s,dim1,subdim1
      real(dop)                                         :: trace,en
      real(dop), intent(in)                             :: time,sume,summ
      real(dop), dimension(maxspf)                      :: eval
      real(dop), dimension(maxspf)                      :: weight
      real(dop), dimension(3*maxspf)                    :: workr
      complex(dop), dimension(dim1,dim1), intent(in)    :: dicht2
      complex(dop), dimension(maxspf*maxspf)            :: evek
      complex(dop), dimension(dim1,dim1), intent(out)   :: workc2
      complex(dop), dimension(subdim1,dim1), intent(in) :: psi,dtpsi

!.....Compute the matrix <phi_j|dtphi_k> = <phi_j|H_av|phi_k>
      call mmaxzz(psi,dtpsi,workc2,subdim1,dim1,dim1)
      trace = 0.0_dop
      do i = 1, dim1
         workc2(i,i) = workc2(i,i) + sume
         trace = trace + dble(workc2(i,i))
      enddo

!.....Diagonalise the matrix <phi_j|dtphi_k>. Eigenvectors in workc2.
      call zheev ('V', 'L', dim1, workc2, dim1, eval, evek, &
           dim1*dim1, workr, i )
      if(i.ne.0) then
         routine = 'orbenergies'
         message = 'Diagonalisation failed (ZHEEV)'
         call errormsg
      endif

!.....Compute weights. <EV|dicht2|EV>. Dicht2 is the MCTDH-density.
      call zerovxd(weight,dim1)
      do i = 1, dim1
         do j = 1, dim1
            do k = 1, dim1
               weight(k) = weight(k) +  &
                   dble( dconjg(workc2(i,k))*dicht2(i,j)*workc2(j,k) )
            enddo
         enddo
      enddo

      en = 0.0_dop
      do k = 1, dim1
         en = en + weight(k)*eval(k)
      enddo

!.....Output to file orben.
      if(m.eq.1 .and. s.eq.1) write(iorb,'(80a1,/a,f9.2,4x,a)')  &
           ('-',i=1,80),' Time[fs] =', time/fs, &
           '(eigen values [eV] / weights*1000)'
      write(iorb,'(a,i3,a,i3,a,f13.6,a,f13.6,a)') & 
           ' Mode :',m,',  State :',s,',  Trace =', &
            trace*ev,' eV,  E-uncorr =', summ*ev,' eV'

      write(iorb,'(8f10.4)')  (eval(i)*ev,i=1,min(dim1,8)) 
      write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=1,min(dim1,8)) 
      if(dim1.gt.8) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=9,min(dim1,16)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=9,min(dim1,16)) 
      endif
      if(dim1.gt.16) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=17,min(dim1,24)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=17,min(dim1,24)) 
      endif
      if(dim1.gt.24) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=25,min(dim1,32)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=25,min(dim1,32)) 
      endif
      if(dim1.gt.32) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=33,min(dim1,40)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=33,min(dim1,40)) 
      endif
      if(dim1.gt.40) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=41,min(dim1,48)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=41,min(dim1,48)) 
      endif
      if(dim1.gt.48) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=49,min(dim1,56)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=49,min(dim1,56)) 
      endif
      if(dim1.gt.56) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=57,min(dim1,64)) 
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=57,min(dim1,64)) 
      endif
      if(dim1.gt.56) then
         write(iorb,*)
         write(iorb,'(8f10.4)')  (eval(i)*ev,i=65,dim1)
         write(iorb,'(8f10.4)')  (weight(i)*1.0e3_dop,i=65,dim1)
      endif
      write(iorb,'(a,f16.9,a)')  &
           ' Average energy: <psi|H-av|psi> =', en*ev,' eV'

      return
      end subroutine orbenergies


! ------------------------------------------------------------------------
! *** ************> SUBROUTINE E_HLOCHPHI 1M <*******************
!                   ~~~~~~~~~~~~~~~~~~~~~~~~

! Computes for one mode, and all Hamiltonian terms the product of the
! trace of the mean-field matrices and h|phi>.
!
! ktype governs which terms are included:
! ktype = 0 : all terms
! ktype = 1 : Potential terms are excluded (i.e. only kinop terms
!             calculated)
! nham      : Number of operator to be used
! HDM 12/01
! *** ************************************************************
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in e_hlochphi1m, please do them in
! e_hlochphi1mpar and dwtehlochphi1mpar in dwtorbenmod too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine e_hlochphi1m (psi,dtpsi,hpsi,hloch,m,phidim1,nham,ktype)

use griddatmod
use psidef
use operdef
use hpsimod
use hphimod, only: hphi1mk

      implicit none

      integer(long)                                 :: k,s,k1,k2,pointer
      integer(long), intent(in)                     :: m,phidim1,nham,ktype
      complex(dop)                                  :: trace
      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(phidim1), intent(out) :: dtpsi
      complex(dop), dimension(hpsidim), intent(out) :: hpsi
      complex(dop), dimension(hlochdim), intent(in) :: hloch
      logical(kind=4), dimension(maxkoe)            :: kcalc


!-----------------------------------------------------------------------
! first find which terms need to be calculated
!-----------------------------------------------------------------------
      call zerovxl(kcalc,ktot)
      call zerovxz(dtpsi,phidim1)
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1
      if (ktype .eq. 0) then
         do k=k1,k2
            if (.not. diag(m,k))   kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
            if (ki(k).ne.kf(k))    kcalc(k)=.false.
         enddo
      else if (ktype .eq. 1) then
         do k=k1,k2
            if (kinop(m,k).eq.1)   kcalc(k)=.true.
            if (autocap(k) .gt. 0) kcalc(k)=.false.
            if (ki(k).ne.kf(k))    kcalc(k)=.false.
         enddo
      endif

! --- COMPUTE TRACE OF MEAN-FIELD TIMES H|PHI> FOR ALL STATES AND TERMS ---

      do k = k1,k2
         if (kcalc(k)) then
            s=kf(k)
            pointer = zetf(m,s)-zetf(m,1)+1
            call trqxz(hloch(phloch(m,k)),trace,dim(m,s))  
            call hphi1mk(psi(zetf(m,s)),hpsi,m,k,dim(m,s),subdim(m))
            call xmxxzza(trace,hpsi(zhpsi(m,k)),dtpsi(pointer),&
                          subdim(m),dim(m,s))
         endif
      enddo

      if( nmulpot(1) .gt. 0 ) then
         message='orben cannot use muld-potentials'
         routine='e_hlochphi1m'
         call errormsg
      endif
!      k=zham(nham)+kzahl(nham)-1
!      do k1=1,nmulpot(1)
!         k=k+1
!         if (.not. diag(m,k)) then
!            s=kf(k)
!            call addmxxzo(hloch(phloch(m,k)),dtpsi(zetf(m,s)),
!     +              subdim(m),dim(m,s))
!         endif
!      enddo

      return
      end subroutine e_hlochphi1m

! ------------------------------------------------------------------------
! *** ************> SUBROUTINE E_HLOCHPHI 1M PAR <****************
!                   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ifdef MPI
      subroutine e_hlochphi1mpar (psi,dtpsi,hpsi,hloch,m,phidim1,nham,ktype)

use mpi

use griddatmod
use psidef
use operdef
use hpsimod
use mpidata

use op1lib, only: cpvxz
use mpipropmod, only: mpiehlochphi

      implicit none

      integer(long), intent(in) :: phidim1,nham,ktype,m
      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(dgldim), intent(out)  :: dtpsi
      complex(dop), dimension(hpsidim), intent(out) :: hpsi
      complex(dop), dimension(hlochdim), intent(in) :: hloch

      integer(long) :: k1,k2

!-----------------------------------------------------------------------
! first find which terms need to be calculated
!-----------------------------------------------------------------------
      call zerovxz(dtpsi,phidim1)
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

! --- COMPUTE TRACE OF MEAN-FIELD TIMES H|PHI> FOR ALL STATES AND TERMS ---

! --- IF BASIS IS COMPLETE DO NOT CALCULATE ---

      if(lmpi.and.lmpistarted.and.lmpihlochphi) then
        mpitask=11
! --- Determine Job
         call MPI_Bcast(mpitask,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
         call MPI_Bcast(phidim1,1,MPI_LOGICAL,0,MPI_COMM_WORLD,mpierr)
         call mpiehlochphi(psi,dtpsi,hpsi,hloch,m,nham,ktype,phidim1)
      endif

      if( nmulpot(1) .gt. 0 ) then
         message='orben cannot use muld-potentials'
         routine='e_hlochphi1m'
         call errormsg
      endif

      return
      end subroutine e_hlochphi1mpar 
#endif

!#######################################################################
! *** HDM 01/02 **********> SUBROUTINE e_CALCHA   <*********************
!                          ~~~~~~~~~~~~~~~~~~~~~
!
! Calculates the action of the correlated Hamiltonian on the
! MCTDH-coefficients for the intra-state terms.
! Only terms wich are diagonal for mode 'mode' are considered.
! Result is returned as a sum over all terms in dtpsi.
!
! nham denotes which operator is to be used. 

      subroutine e_calcha (psi,dtpsi,hteil,nham,mode,phi,jindx)

#ifdef MPI
use mpi
#endif

use dvrdatmod
use griddatmod
use psidef
use operdef
use hpsimod
use distribmod
use mpidata

use calchamod, only: calcha1k
use op1lib, only: cpvxz
#ifdef MPI
use mpipropmod, only: mpiecalcha
#endif

      implicit none

      integer(long)                                       :: s,k,k0,koffset,ithr
      integer(long), intent(in)                           :: nham,mode
      integer(long), dimension(jindxdim), intent(in)      :: jindx
      complex(dop), dimension(adim), intent(in)           :: psi
      complex(dop), dimension(adim), intent(out)          :: dtpsi
      complex(dop), dimension(adim+1,dgldim), intent(in)  :: phi
      complex(dop), dimension(pmatdim), intent(in)        :: hteil
      complex(dop), dimension(:,:), allocatable           :: tmpdtpsi
#ifdef OMP
      integer(long), external  :: omp_get_thread_num, omp_get_num_threads
#endif

      koffset=zham(nham)+kzahl(nham)-1
      k0=zham(nham)
      call zerovxz(dtpsi,adim)

#ifdef MPI
      if(lmpi.and.lmpicalcha.and.lmpistarted) then
         mpitask=10
! --- Determine Job
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
!         call MPI_Bcast(fmblock,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call mpiecalcha(psi,dtpsi,hteil,nham,mode,koffset,k0,jindx)
      else
#endif

!-----------------------------------------------------------------------
! allocate memory for omp threads
!-----------------------------------------------------------------------
#ifdef OMP
      if (lompthread) then
          allocate(tmpdtpsi(adim,ompthread-1))
          tmpdtpsi=0.0_dop
      endif
#endif

! --- LOOP OVER EACH CORRELATED HAMILTONIAN TERM ---

!$omp parallel num_threads(ompthread) private(k,s,ithr) if(lompthread .and. lompcalcha)
!$omp do schedule(dynamic)
         do k=k0,koffset
#ifdef OMP
            ithr = omp_get_thread_num()
#else
            ithr = 0
#endif
            s = kf(k)
            if (s.eq.ki(k) .and. diag(mode,k)) then
               if (ithr .eq. 0) then
                  call calcha1k(psi,phi,hteil,dtpsi,k,s,jindx)
               else
                  call calcha1k(psi,phi,hteil,tmpdtpsi(1,ithr),k,s,jindx)
               endif
            endif

         enddo
!$omp enddo
!$omp end parallel 

#ifdef OMP
         if (lompthread) then
            do ithr = 1,ompthread-1
               call addvxxzo(tmpdtpsi(1,ithr),dtpsi,adim)
            enddo
         endif
#endif


#ifdef MPI
      endif
#endif

      return
      end subroutine e_calcha

!#######################################################################

      subroutine wkorben

use psidef

      implicit none

!-----------------------------------------------------------------------
! assign workspace used in module ORBEN
!-----------------------------------------------------------------------

      workrdim=max(workrdim,5*maxspf)       ! used in orbenergies
      workcdim=max(workcdim,max(adim,maxphidim)+2*maxspf+3*maxblock)

      return
      end subroutine wkorben


!***********************************************************************

      end module orbenmod

