!#######################################################################
!                                                                       
!                    module PROJECT
!                                                                       
!   MCTDH module containing (auxiliary) subroutines to evaluate the     
!   effect of the spf projector 
!                                                                       
!   Contains:                                                           
!     project:      Determines for all modes and states the action of   
!                   the (1-P)-projection operator on the                
!                   single-particle functions.                          
!     project1ms:   Determines for one mode and state the action of     
!                   the (1-P)-projection operator on the                
!                   single-particle functions.                          
!     opro:         Calculates for one mode and state the projection    
!                   on the space orthogonal to the space spanned by     
!                   the single-particle functions (new projector        
!                   routine).                                           
!     projector:    Computes for one mode and state the action of the   
!                   (1-P)-projection operator on the single-particle    
!                   functions (old projector routine).                  
!                                                                       
!#######################################################################

      module projectmod

use decimal,  only: dop, long
use global
use dglobal
use dvrdatmod
use griddatmod
use psidef
use runpropmod
use timing

use logdat
use mmlib,    only: mmaxzz, qqxxzz, mqxxzz
use op1lib,   only: zeromxz, overmxz
use op2lib,   only: submxxzo1

use openmpmod, only: lompthread,ompthread
use d2tens, only: filldens


      implicit none
      private
      public :: project,projector,projectorT,project1ms,opro,chkovl,numberproj

      contains

! *** V6.0 MB **************> SUBROUTINE PROJECT <**********************
!                            ~~~~~~~~~~~~~~~~~~~~

! Determines for all modes and states the action of the
! (1-P)-projection operator on the single-particle functions.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in project, please do them in projectpar too!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine project (psi,dtpsi)

      implicit none

      integer(long)                                  :: m,s,t
      integer(long), save                            :: tid=0
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(dgldim), intent(inout) :: dtpsi

      call get_timer(tid,'Project')
      call start_timer(tid)

! --- LOOP OVER EACH MODE AND STATE ---

      if (dentype .eq. 1) then
         do m = 1,nmode
            do s = 1,nstate
               do t = 1,nstate
                  call project1ms(psi(dzetf(m,s,t)),dtpsi(dzetf(m,s,t)), &
                       m,ddim(m,s,t),subdim(m))
               enddo
            enddo
         enddo
      else
         do m = 1,nmode
            do s = 1,nstate
               call project1ms(psi(zetf(m,s)),dtpsi(zetf(m,s)),m, &
                    dim(m,s),subdim(m))
            enddo
         enddo
      endif

      call stop_timer(tid)

      return
      end subroutine project  

! *** V6.0 MB *************> SUBROUTINE PROJECT 1MS <*******************
!                           ~~~~~~~~~~~~~~~~~~~~~~~~
! Determines for one mode and state the action of the (1-P)-projection
! operator on the single-particle functions.
!#######################################################################

      subroutine project1ms (psi,dtpsi,m,dim1,subdim1)

      implicit none

      integer(long)                                        :: f
      integer(long), intent(in)                            :: m,dim1,subdim1
      complex(dop), dimension(subdim1,dim1), intent(in)    :: psi
      complex(dop), dimension(subdim1,dim1), intent(inout) :: dtpsi

! --- IF BASIS IS COMPLETE SET time derivative to zero ---

      if (complete(m)) then
         dtpsi(:,:) = 0.0_dop
         return
      endif

! --- IF GWP or X-local mode, do not calculate

      f = spfdof(1,m)
      if(basis(f).eq.25 .or. basis(f) .eq. 26 .or. basis(f) .eq. 27)&
         return

! --- COMPUTE ACTION OF PROJECTOR ON SPF ---
      if (lopro) then
         call opro(psi,dtpsi,subdim1,dim1,m)
      else
         call projector(psi,dtpsi,dim1,subdim1,dim1)
      endif
      
      return
      end  subroutine project1ms

! *** V6.0 MB ***************> SUBROUTINE OPRO <************************
!                             ~~~~~~~~~~~~~~~~~
! Calculates for one mode the projection on the space orthogonal to the
! space spanned by the single-particle functions.

! The standard MCTDH projection operator P = sum_l |phi_l><phi_l|
! is now replaced by P = sum_lm |phi_l> (ovl^-1)_lm <phi_m|
! where (ovl^-1)_lm is a matrix element of the inverse of the Hermitian
! spf overlap matrix ovl_lm = <phi_l|phi_m>.
! Psi contains the spf as input and is not altered, dtpsi contains the
! functions on which the (1-P)-projection is employed as input and the
! result of the projection as output
!#######################################################################

      subroutine opro(psi,dtpsi,subdim,dim,m)

      implicit none

      integer(long)                                       :: error
      integer(long), intent(in)                           :: dim,subdim,m
      complex(dop), dimension(dim*dim,3)                  :: tmpovl
      complex(dop), dimension(subdim,dim), intent(in)     :: psi
      complex(dop), dimension(subdim,dim), intent(inout)  :: dtpsi
      complex(dop), dimension(subdim,dim)                 :: help


! --- CALCULATE OVERLAP MATRIX OF SINGLE-PARTICLE FUNCTIONS ---

      call overmxz(psi,tmpovl(1,2),subdim,dim)

!      call chkovl(tmpovl(1,2),dim)
! --- INVERT OVERLAP MATRIX ---

!      call matinvzhp(tmpovl(1,1),tmpovl(1,2),dim,dim,tiny,error)
      tmpovl(:,1)=tmpovl(:,2)
      call zpotrf('L',dim,tmpovl(:,1),dim,error)
      call zpotri('L',dim,tmpovl(:,1),dim,error)
      if (error .ne. 0) then
         routine = 'opro'
         write(message,'(a,i4,25x,2a)') &
          'Overlap matrix not positive definite for mode', m,&
          '(This error could be caused by a too large initial ',&
          'integration step)'
         call errormsg
      endif
      call filldens(tmpovl(:,1),dim)

! --- APPLY (1-P)-PROJECTOR ---

!      if(.not.lompthread)then
         call mmaxzz(psi,dtpsi,tmpovl(1,2),subdim,dim,dim)
         call qqxxzz(tmpovl(1,1),tmpovl(1,2),tmpovl(1,3),dim)
         call mqxxzz(psi,tmpovl(1,3),help,subdim,dim)
!      else
!         call mmaxzzomp(psi,dtpsi,tmpovl(1,2),subdim,dim,dim,ompthread)
!         call qqxxzzomp(tmpovl(1,1),tmpovl(1,2),tmpovl(1,3),dim,ompthread)
!         call mqxxzzomp(psi,tmpovl(1,3),help,subdim,dim,ompthread)
!      endif
      call submxxzo1(dtpsi,help,subdim,dim)

      return
      end subroutine opro

      subroutine chkovl(ovl,dim)

      integer(long)                                       :: i,j
      complex(dop)                                        :: sum
      integer(long),intent(in)                            :: dim
      complex(dop), dimension(dim,dim), intent(in)        :: ovl

sum=0.0_dop
do i=1,dim
   sum = sum+ovl(i,i)
enddo
print*,'On-diag ',dim,sum

sum=0.0_dop
do i=1,dim
do j=1,dim
   if (i .eq. j) cycle
   sum = sum+ovl(j,i)
enddo
enddo
print*,'Off-diag ',sum


      return
      end subroutine chkovl

! *** V6.0 MB ***********> SUBROUTINE PROJECTOR <***********************
!                         ~~~~~~~~~~~~~~~~~~~~~~
! Computes the action of the (1-P) projection operator on the single-
! particle functions.
!#######################################################################

      subroutine projector (psi,dtpsi,dim,subdim,dim2)

!$ use omp_lib

      implicit none

      integer(long)                                       :: e,e1,g
      integer(long), intent(in)                           :: dim,subdim,dim2
      complex(dop)                                        :: overlap
      complex(dop), dimension(subdim,dim), intent(in)     :: psi
      complex(dop), dimension(subdim,dim2), intent(inout) :: dtpsi

      complex(dop), dimension(dim*dim,3)                  :: tmpovl

! --- CALCULATE OVERLAP MATRIX OF SINGLE-PARTICLE FUNCTIONS ---

      call overmxz(psi,tmpovl(1,2),subdim,dim)

!      call chkovl(tmpovl(1,2),dim)

! --- COMPUTE ACTION OF PROJECTOR ON SINGLE-PARTICLE FUNCTIONS ---
!     (do it twice to improve numerics)

! $omp parallel num_threads(ompthread) private(e,e1,g,overlap) if(lompthread)
! $omp do schedule(dynamic)
      do e = 1,dim2
         do e1 = 1,dim
            overlap=(0.0_dop,0.0_dop)
            do g = 1,subdim
               overlap=overlap+dconjg(psi(g,e1))*dtpsi(g,e)
            enddo
            do g = 1,subdim
               dtpsi(g,e) = dtpsi(g,e) - overlap * psi(g,e1)
            enddo
         enddo
      enddo
      do e = 1,dim2
         do e1 = 1,dim
            overlap=(0.0_dop,0.0_dop)
            do g = 1,subdim
               overlap=overlap+dconjg(psi(g,e1))*dtpsi(g,e)
            enddo
            do g = 1,subdim
               dtpsi(g,e) = dtpsi(g,e) - overlap * psi(g,e1)
            enddo
         enddo
      enddo
! $omp end do
! $omp end parallel
      return
      end subroutine projector

!#######################################################################

      subroutine projectorT (psi,dtpsi,dim,subdim,dim2)

      implicit none

      integer(long), intent(in) :: dim2,dim,subdim
      integer(long)             :: e,e1,g
      complex(dop), dimension(subdim,dim), intent(in)   :: psi
      complex(dop), dimension(subdim,dim2), intent(out) :: dtpsi
      complex(dop)                                      :: overlap

! --- COMPUTE ACTION OF PROJECTOR ON SINGLE-PARTICLE FUNCTIONS ---

      do e = 1,dim2
         do e1 = 1,dim
            overlap=(0.0d0,0.0d0)
            do g = 1,subdim
               overlap=overlap+psi(g,e1)*dtpsi(e,g)
            enddo
            do g = 1,subdim
               dtpsi(e,g) = dtpsi(e,g) - overlap * dconjg(psi(g,e1))
            enddo
         enddo
      enddo
      do e = 1,dim2
         do e1 = 1,dim
            overlap=(0.0d0,0.0d0)
            do g = 1,subdim
               overlap=overlap+psi(g,e1)*dtpsi(e,g)
            enddo
            do g = 1,subdim
               dtpsi(e,g) = dtpsi(e,g) - overlap * dconjg(psi(g,e1))
            enddo
         enddo
      enddo

      return
      end subroutine projectorT 

!#######################################################################
! ********************> SUBROUTINE NUMBERPROJ <*************************
!                       ~~~~~~~~~~~~~~~~~~~~~
! This subroutine does the number projection of SPFs to keep the
! expectation value of the Number Operator constant.
! Input parameters:
!     psi  : wavefunction
!     dtpsi: time-derivative of the of the psi
!     currm: current mode
!     if currm = 0, then psi(dgldim),dtpsi(dgldim)
!     else               psi(dgldim),dtpsi(maxphidim)
! Output parameters:
!     dtpsi: time-derivative of the of the psi after number projection
!            on SPFs.
! Intermediate array:
!     npgrid: This array contains the information about the
!             grid points of each SPF that belongs to
!             a fixed particle number.
!             For given SPF 'e' of mode 'm', only those grid points
!             that are .ge. npgrid(m,e,1) .or. .le. npgrid(m,e,2)
!             can have non-zero amplitude.
!             This array is created in the first call and then saved
!             for the whole propagation time.
!-----------------------------------------------------------------------

      subroutine numberproj(psi,dtpsi,currm)

      implicit none

      integer(long)             :: s,m,e,g,ipt
      integer(long), intent(in) :: currm
      complex(dop), dimension(*), intent(in)  :: psi
      complex(dop), dimension(*), intent(out) :: dtpsi
      integer(long), dimension(:,:,:),allocatable, save :: npgrid

      if (.not. allocated(npgrid)) allocate(npgrid(nmode,maxspf,2))

      if(lrdnumproj) call numspfgrid(psi,currm,npgrid)

      if(currm.eq.0) then
         do s = 1,nstate
            do m = 1,nmode
               do e = 1,dim(m,s)
                  do g = 1,subdim(m)
                     if((g.lt.npgrid(m,e,1)).or. (g.gt.npgrid(m,e,2)))then
                        ipt = zetf(m,s)+subdim(m)*(e-1)+g-1
                        dtpsi(ipt) = (0.d0,0.d0)
                     endif
                  enddo
               enddo
            enddo
         enddo
      else
         m=currm
         do e = 1,dim(m,1)
            do g = 1,subdim(m)
               if((g.lt.npgrid(m,e,1)).or.(g.gt.npgrid(m,e,2)))then
                  ipt = subdim(m)*(e-1)+g
                  dtpsi(ipt) = (0.d0,0.d0)
               endif
            enddo
         enddo
      endif

      return
      end subroutine numberproj

!-----------------------------------------------------------------------
!#######################################################################
! ********************> SUBROUTINE NUMSPFGRID <***********************
!                       ~~~~~~~~~~~~~~~~~~~~~~~
! This subroutine reads Hilbert space information (1st grid point
! of different Hilbert space) of each DOF and then depending on the
! psi(t=0), generates the 'npgrid' array that contains the grid points
! information of each SPF that belongs to a fixed particle number.
!-----------------------------------------------------------------------
! Called from numberproj
! Input parameters:
!     psi  : wavefunction
!     currm: current mode
!     if currm = 0, then psi(dgldim)
!     else               psi(dgldim)
! Input parameters coming from common block:
!     numprojfile: Name of the input file
!     npflaenge:   character length of the input file
!     lrdnumproj:  should be .true. at the start of the subroutine
!                  and .false. when leaving the subroutine.
! Output parameters:
!     npgrid: dimension is ==> npgrid(nmode,maxspf,2)
!             For given SPF 'e' of mode 'm', only those grid points
!             that are .ge. npgrid(m,e,1) .or. .le. npgrid(m,e,2)
!             can have non-zero amplitude.
!-----------------------------------------------------------------------

      subroutine numspfgrid(psi,currm,npgrid)

      implicit none

      integer(long), intent(in)                             :: currm
      integer(long), dimension(nmode,maxspf,2), intent(out) :: npgrid
      integer(long)                              :: m,e,g,i,ipt,maxhil
      integer(long),dimension(nmode)             :: nhil
      complex(dop), dimension(*), intent(in)     :: psi
      integer(long), dimension(:,:), allocatable :: initgp,fingp

      open(81,file=numprojfile(1:npflaenge),status='old')
      read(81,*)(nhil(m), m=1,nmode)
      maxhil=maxval(nhil)
      allocate(initgp(nmode,maxhil))
      do m=1,nmode
         read(81,*)(initgp(m,g), g=1,nhil(m))
      enddo
      close(81)

      allocate(fingp(nmode,maxhil))
      do m=1,nmode
         do i=1,nhil(m)-1
           fingp(m,i)=initgp(m,i+1)-1
         enddo
         fingp(m,nhil(m))=subdim(m)
      enddo

      if(currm.eq.0) then
         do m = 1,nmode
            do e = 1,dim(m,1)
               do g = 1,subdim(m)
                  ipt = zetf(m,1)+subdim(m)*(e-1)+g-1
                  if(cdabs(psi(ipt)).ge.1.D-8) then
                     do i=1,nhil(m)
                        if((initgp(m,i).le.g).and. (g.le.fingp(m,i))) then
                           npgrid(m,e,1)=initgp(m,i)
                           npgrid(m,e,2)=fingp(m,i)
                           goto 100
                        endif
                     enddo
                  endif
               enddo
 100           continue
            enddo
         enddo
      else
         m=currm
         do e = 1,dim(m,1)
            do g = 1,subdim(m)
               ipt = subdim(m)*(e-1)+g
               if(cdabs(psi(ipt)).ge.0.99d0) then
                  do i=1,nhil(m)
                     if((initgp(m,i).le.g).and. (g .le.fingp(m,i))) then
                        npgrid(m,e,1)=initgp(m,i)
                        npgrid(m,e,2)=fingp(m,i)
                        goto 200
                     endif
                  enddo
               endif
            enddo
 200        continue
         enddo
      endif

      deallocate(initgp,fingp)

      lrdnumproj=.false.

      return
      end subroutine numspfgrid

! **********************************************************************

      end module projectmod

