! **********************************************************************
!                                                                       
!                           EVALUATION (evaluation.f)                   
!                                                                       
!   MCTDH module containing (auxiliary) subroutines to evaluate the     
!   right hand side of the MCTDH working equations.                     
!                                                                       
!   Contains:                                                           
!     summf:        Multiplies for all modes and all states the sum     
!                   over the mean-field matrices times the              
!                   single-particle potential operator matrices with    
!                   |phi>.                                              
!     mfsumphi:     Sums for all modes and all states the product of    
!                   the mean-field matrix times the single-particle     
!                   potential operator.                                 
!     mfsumphi1m:   Sums for one mode and all states the product of     
!                   the mean-field matrix times the single-particle     
!                   potential operator.                                 
!     dwtsummf:     Calculates the same like the original summf
!                   routine, but the loop over the k-terms is done in
!                   parallel                        
!                                                                       
! **********************************************************************
      module summfmod

use decimal, only: long, dop
use global
use logdat
use lalib,   only: mattens, mktens
use op1lib,  only: zerovxz, initvxz
use sdlib,   only: omattens

#ifdef MPI
use mpi
#endif
use operdef
use runpropmod
use hpsimod
use hphimod, only: hcall
use mpidata
use griddatmod
use operdef
use psidef

use timing

use op1lib, only: initvxz, zerovxz,cpvxz

!$ use omp_lib
use openmpmod

      implicit none
      private
      public :: summf,mfsumphi1m,mfsumphi
#ifdef MPI
      public :: mpisummf
#endif
      contains

! *** V6.0 MB ***************> SUBROUTINE SUMMF <***********************
!                             ~~~~~~~~~~~~~~~~~~
!
! Sums up for all modes and all states the product of the mean-field
! matrices times the single-particle potential operators.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in summf, please do them in dwtsummf too!
!
! NOTE: summf now serves merely as a case selector to call the different
! parallel or serial subroutines that actually do the summation.
! MS 10/2012
!
! F90 version: GWR 12/12
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine summf (mfsum,hloch,nham)

      implicit none

      integer(long), intent(in)     :: nham
      integer(long)                 ::    k1,k2,kstep
      complex(dop), dimension(mfsumdim), intent(out) :: mfsum
      complex(dop), dimension(hlochdim), intent(in)  :: hloch
      complex(dop), dimension(3*sdim)                :: workc
      integer(long), save                            :: tid=0 ! timer ID

      
! --- INITIALISE VARIABLES ---
      kstep = 1 ! kstep > 1 in MPI routines, but set there
      
      call get_timer(tid, 'Summf')
      call start_timer(tid)

      call zerovxz(mfsum,mfsumdim)

! --- LOOP OVER EACH MODE AND EACH CORRELATED HAMILTONIAN TERM ---
! --- SUM UP MEAN-FIELD TIMES SINGLE-PARTICLE OPERATOR ---
      k1=zham(nham)
      k2=zham(nham)+kzahl(nham)-1

!-----------------------------------------------------------------------
!     If the shared memory parallelisation is used and the global
!     threads have been started all values are passed via common blocks
!     and globally known arrays to the threads. The control is passed to
!     the threads by increasing the Semaphore 1. The threads then
!     return the control by increasing Semaphore 2. Additionally the
!     mutex-array is initialized and later destroyed. In this case an
!     array of mutex is used to prevent the threads to lock them
!     mutually because a huge part of the time is spent writing to the
!     memory.
!-----------------------------------------------------------------------
#ifdef MPI
      if (lmpi.and.lmpistarted.and.lmpisummf) then
         mpitask = 13
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         call mpisummf(mfsum,hloch,k1,k2)
      else
!-----------------------------------------------------------------------
!     If no parallelisation is used then a usual calculation is 
!     performed.
!-----------------------------------------------------------------------
#endif
                      !k1,k2,kstep,...
         call summfser(k1,k2,1,mfsum,hloch,workc)

#ifdef MPI
      endif

      if (lmpi .and. lmpistarted .and. lmpisummf) then
         mpitask = 23
         call MPI_Bcast(mpitask,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
         ! send mfsum out to the world... 
         call MPI_Bcast(mfsum,mfsumdim,MPI_DOUBLE_COMPLEX,master, &
                        MPI_COMM_WORLD,mpierr)
      end if
#endif

      call stop_timer(tid)

      return
      end subroutine summf

!---------------------------------------------------------------------
! --- summfser: execute serial summf
      subroutine summfser(k1,k2,kstep,mfsum,hloch,workc)

      implicit none 

      integer(long), intent(in)                      :: k1,k2,kstep
      complex(dop), dimension(mfsumdim), intent(out) :: mfsum
      complex(dop), dimension(hlochdim), intent(in)  :: hloch
      complex(dop), dimension(3*sdim), intent(out)   :: workc

      integer(long) ::    m,k,zeig,zeig1,s,s1,f,h,shift,shiftidx
      
      do m = 1,nmode
         if (gwpm(m)) cycle
         if(idmode(m).eq.0) then
            do k = k1,k2,kstep
               if (kinop(m,k).le.0 .and. .not.diag(m,k)) then
                  zeig=1
                  zeig1=zeig+subdim(m)
                  s1=ki(k)
                  s=kf(k)
                  if (kinop(m,k).lt.0) then
                     f=spfdof(1,m)
                     h=hamilton(f,k)
                     shift=htmshift(h)
                     call ilookup(shiftidx,ihtmshift(1,m), &
                          nhtmshift(m),shift)
                     if (shiftidx.le.0) then
                        routine='summf'
                        message='FIXME: '&
                             //'called with unexpected htmshift'
                        call errormsg
                     endif
                  else
                     shiftidx=1
                  endif
                  call initvxz(workc(zeig1),subdim(m))
                  call hcall(m,k,workc(zeig),workc(zeig1),subdim(m))
                  call mktens(mfsum(zmfsum1(m,s1,s,shiftidx)), &
                       workc(zeig),hloch(phloch(m,k)),subdim(m), &
                       dim(m,s),dim(m,s1),true)
               endif
            enddo
         endif
      enddo

      return

      end subroutine summfser

! *** V6.0 MB *************> SUBROUTINE MFSUMPHI <**********************
!                           ~~~~~~~~~~~~~~~~~~~~~

! Multiplies for all modes and all states the sum over the products of
! the mean-field matrices times the single-particle (side-)diagonal operator
! matrices with |phi>.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! If you do some changes in mfsumphi, please also do the same changes
! in mfsumphipar!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine mfsumphi (dtpsi,psi,mfsum)

      implicit none

      complex(dop), dimension(dgldim), intent(in)   :: psi
      complex(dop), dimension(dgldim), intent(out)  :: dtpsi
      complex(dop), dimension(mfsumdim), intent(in) :: mfsum

      integer(long) :: m

      if (kzahl(1) .eq. 0) return

! --- LOOP OVER EACH MODE AND STATE---

! omp parallel do num_threads(ompthread) private(m) 
      do m = 1,nmode
         if (complete(m)) cycle

         call mfsumphi1m(dtpsi(zetf(m,1)),psi(zetf(m,1)),mfsum,m, &
              zetf(m,1),phidim(m))
      enddo
! omp end parallel do

      return

      end subroutine mfsumphi

! **********************************************************************
! *** V6.0 MB *************> SUBROUTINE MFSUMPHI 1M <*******************
!                           ~~~~~~~~~~~~~~~~~~~~~~~~
! Multiplies for one mode and all states the sum over the products of
! the mean-field matrices times the single-particle (side-)diagonal operator
! matrices with |phi>.
! **********************************************************************

      subroutine mfsumphi1m (dtpsi,psi,mfsum,m,zetf1,phidim1)

      implicit none

      integer(long)                                               :: s,s1,p, &
                                                                     idx, &
                                                                     shift
      integer(long), intent(in)                                   :: m,zetf1, &
                                                                     phidim1
      complex(dop), dimension(mfsumdim), intent(in)               :: mfsum
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(in)  :: psi
      complex(dop), dimension(zetf1:zetf1+phidim1-1), intent(out) :: dtpsi

      if (kzahl(1) .eq. 0) return

! --- LOOP OVER EACH COMBINATION OF STATES WITHIN EACH PACKET ---

      do p = 1,npacket
         do s = (p-1)*nstate/npacket+1,p*nstate/npacket
            do s1 = (p-1)*nstate/npacket+1,p*nstate/npacket

!          --- MULTIPLY SUM OVER MEAN-FIELDS WITH |PHI> ---

               do idx=1,nhtmshift(m)
                  shift=ihtmshift(idx,m)
                  if (shift.eq.0) then
                     call mattens(dtpsi(zetf(m,s)),psi(zetf(m,s1)), &
                          mfsum(zmfsum(m,s1,s)),subdim(m),dim(m,s), &
                          dim(m,s1),.true.)
                  else
                     call omattens(dtpsi(zetf(m,s)),psi(zetf(m,s1)), &
                          mfsum(zmfsum1(m,s1,s,idx)),shift,subdim(m), &
                          dim(m,s),dim(m,s1),.true.)
                  endif
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine mfsumphi1m

!-----------------------------------------------------------------------
!   mpisummf
!-----------------------------------------------------------------------
#ifdef MPI
      subroutine mpisummf(mfsum,hloch,k1,k2)  

use mpi
use distribmod
use mpidata

      implicit none

      complex(dop), dimension(mfsumdim), intent(out):: mfsum
      complex(dop), dimension(hlochdim), intent(in) :: hloch
      complex(dop), dimension(mfsumdim)             :: tmpsum
      integer(long), intent(inout)                  :: k1,k2

      integer    kstep
    
      call zerovxz(mfsum,mfsumdim)

      call MPI_Bcast(k1,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(k2,1,MPI_INTEGER,master,MPI_COMM_WORLD,mpierr)
      call MPI_Bcast(hloch,hlochdim,MPI_DOUBLE_COMPLEX,master, &
                     MPI_COMM_WORLD,mpierr)

      ! MPI parellelization is only over the k's 
    
      k1 = k1+mpirank
      kstep = nmpi

      call summfser(k1,k2,kstep,mfsum,hloch,tmpsum)
  
      call MPI_Reduce(mfsum,tmpsum,mfsumdim,MPI_DOUBLE_COMPLEX, &
           MPI_SUM,master,MPI_COMM_WORLD,mpierr)

      if (mpirank.eq.master) mfsum = tmpsum

      return

      end subroutine mpisummf
#endif 


      end module summfmod
