!***********************************************************************
!
!                  MCTDH module SVDRED
!
! Performs a singular value decomposition (SVD) of the matricised A-vector
! Called from funk0, funk3, funk and funkr routines on propwf/function.F
!
! passed back to calling routine:
!   psi1  : U matrix ( A=U*lambda*V^dagger ) (for all modes!)
!   dicht4: V matrix ( A=U*lambda*V^dagger )
!   dicht3(d3mat(m,s),1): lambda(m,s)**2, i.e. the natural populations
!   dicht3(d3mat(m,s),2): max(lambda(m,s),sqrt(epssvd)); regularized
!   dicht2: density matrix rho=V*lambda**2 * V^dagger
!   dicht1: V*lambda_reg^{-1}  (after zgesvd: dicht1 = V^dagger)
! 03/2021  HDM
!
!***********************************************************************

      module svdregmod

use decimal
use global
use maxv
use dvrdatmod
use griddatmod
use operdef
use psidef
use runpropmod
use timing
use op1lib, only: cpvxz

      implicit none
      private

      public :: svdreg

      contains

!#######################################################################

      subroutine svdreg(psi,psi1,dicht1,dicht2,dicht3,dicht4)

      implicit none

      integer(long) ::  m,s,info,vdnd,pntr
      real(dop)     ::  workr(5*maxspf)
      real(dop), intent(out)    ::  dicht3(d3matdim)
      complex(dop), intent(in)  :: psi(adim)
      complex(dop), intent(out) :: psi1(nmode*adim)
      complex(dop), intent(out) ::  dicht1(dmatdim),dicht2(dmatdim),dicht4(dmatdim)
      complex(dop)  ::  wk(2*maxblock+4),udum(1)

      integer(long), save :: tid=0 ! timer ID

      call get_timer(tid, 'SVD Reg')
      call start_timer(tid)

!-----------------------------------------------------------------------
! Matricisation: unfold A-tensor (psi->psi1)
! zgesvd: compute the singular value decomposition
!-----------------------------------------------------------------------
      do m=1,nmode
         do s=1,nstate
            pntr=zpsi(s) + (m-1)*adim
            if(ndim(m,s).eq.1) then
!.......... If ndim=1 one does not need to reshape the A-vector
               call cpvxz(psi(zpsi(s)),psi1(pntr),block(s))
            else
               call unfold(psi(zpsi(s)),psi1(pntr),vdim(m,s), &
                           dim(m,s),ndim(m,s),.true.)
            endif

            vdnd=vdim(m,s)*ndim(m,s)
!                     JOBU, JOBVT, M,   N,       A,        LDA,
            call zgesvd('O', 'A',vdnd,dim(m,s),psi1(pntr),vdnd, &
              dicht3(d3mat(m,s)),udum,1,dicht1(dmat(m,s)),dim(m,s),wk, &
              2*maxblock+4,workr,info)
!                 S               U  LDU      VT,          LDVT,   WORK,
!                LWORK,   RWORK, INFO
            if(info.ne.0) then
               write(6,'(a,i5)') 'zgesvd: info =', info
               write(2,'(a,i5)') 'zgesvd: info =', info
               stop 1
            endif

!-----------------------------------------------------------------------
! Compute natural populations lambda**2 and regularized inverse of lamba,
! map V^dagger --> V and compute V*lambda_reg^{-1}
!-----------------------------------------------------------------------
            call vlambda(dicht4(dmat(m,s)),dicht3(d3mat(m,s)), &
              dicht2(dmat(m,s)),dicht1(dmat(m,s)),dim(m,s),sqrt(epssvd))

            if(ndim(m,s).gt.1) then
               call cpvxz(psi1(pntr),wk,block(s))
               call unfold(psi1(pntr),wk,vdim(m,s),dim(m,s),ndim(m,s), &
                                                              .false.)
            endif
         enddo  ! s=1,nstate
      enddo     ! m=1,nmode

      call stop_timer(tid)

      return
      end subroutine svdreg

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine unfold(psi,psi1,vdim,dim,ndim,switch)

      implicit none

      integer(long)   :: vdim,dim,ndim,v,d,n
      logical(kind=4) :: switch
      complex(dop)    :: psi(vdim,dim,ndim),psi1(vdim,ndim,dim)

      if(switch) then
         do n=1,ndim
            do d=1,dim
               do v=1,vdim
                  psi1(v,n,d)=psi(v,d,n)
               enddo
             enddo
         enddo
      else
         do n=1,ndim
            do d=1,dim
               do v=1,vdim
                  psi(v,d,n)=psi1(v,n,d)
               enddo
             enddo
         enddo
      endif
      return
      end subroutine unfold

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      subroutine vlambda(dicht4,dicht3,dicht2,dicht1,dim,epssqrt)

!     input:
!       dicht1      = V^dagger
!       dicht3(*,1) = lambda (singular values)
!       epssqrt     = sqrt(eps_SVD)
!     output:
!       dicht1      = V*lambda_reg^{-1}
!       dicht2      = rho = V*lambda^2*V^dagger
!       dicht3(*,1) = lambda**2 (natural populations)
!       dicht3(*,2) = lambda_reg^{-1}
!       dicht4      = V  (eigenvector matrix of density)

      implicit none

      integer(long) :: i,j,k,dim
      real(dop)     :: dicht3(dim,2),epssqrt
      complex(dop)  :: dicht1(dim,dim),dicht2(dim,dim),dicht4(dim,dim)

      do i=1,dim
         dicht3(i,2) = 1.d0/max(dicht3(i,1),epssqrt) ! lambda_reg^{-1}
         dicht3(i,1) = dicht3(i,1)**2           ! natural populations
      enddo

      do j=1,dim
         do i=1,dim
            dicht4(i,j)=dconjg(dicht1(j,i))     ! V^dagger --> V
         enddo
      enddo

      do j=1,dim
         do i=1,dim
            dicht2(i,j) = dicht4(i,1)*dicht3(1,1)*dicht1(1,j)
         enddo
      enddo
      do j=1,dim
         do k=2,dim
            do i=1,dim
               dicht2(i,j) = dicht2(i,j) + &
                 dicht4(i,k)*dicht3(k,1)*dicht1(k,j) ! rho=V*lambda^2*V^dagger
            enddo
         enddo
      enddo

      do j=1,dim
         do i=1,dim
            dicht1(i,j)=dicht4(i,j)*dicht3(j,2) ! V*lambda_reg^{-1}
         enddo
      enddo

      return
      end subroutine vlambda

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      end module svdregmod
