!#######################################################################
!
!                              Module  TULLY
!
!  Parameters taken from:
!  Tully et al., J. Vac. Sci. Technol. A 11(4), Jul/Aug 1993, 1914-1920
!
!
!   09/02 CC
!
!#######################################################################


!***********************************************************************
!
!                  Subroutine    tully
!
!
!***********************************************************************

      subroutine tully(x,y,z,r,theta,phi,epot,N_xy,N_z)

      implicit none
      real*8 x,y,z,r,theta,phi,epot
      integer N_xy,N_z

!     masses in a.u.
      real*8 mc,mo,M
      parameter(mc=12.0d0*1822.89d0)
      parameter(mo=16.0d0*1822.89d0)
      parameter(M=mc+mo)

!     potential parameters, all parameter are finally in a.u.
!     values taken from:
!     Tully et al., J. Vac. Sci. Technol. A 11(4), Jul/Aug 1993, 1914-1920
      real*8 A,B,F,alpha,beta,gamma,re_0,rCO_0,rCU_0
      parameter(A=84300.00d0*3.80881d-4) ! [kJoule/mol] -> [a.u.]
      parameter(B=   55.86d0*3.80881d-4) ! [kJoule/mol] -> [a.u.]
      parameter(F= 1070.00d0*3.80881d-4) ! [kJoule/mol] -> [a.u.]
      parameter(alpha=3.36d0 * 0.52917725d0) ! [1/Angstroem] -> [1/a.u.]
      parameter(beta =3.00d0 * 0.52917725d0) ! [1/Angstroem] -> [1/a.u.]
      parameter(gamma=2.30d0 * 0.52917725d0) ! [1/Angstroem] -> [1/a.u.]
      parameter(re_0 =1.900d0 * 1.889726d0) ! [Angstroem] -> [a.u.]
      parameter(rCO_0=1.125d0 * 1.889726d0) ! [Angstroem] -> [a.u.]
      parameter(rCU_0=3.614d0 * 1.889726d0) ! [Angstroem] -> [a.u.]


!     variables for the routine
      real*8 R_O(3),R_C(3),R_CU(3), &
           R_cyl(3),R_CO_vec(3),R_CUO_vec(3),R_CUC_vec(3), &
           rCO,rCUO,rCUC,coseta,help

      integer nx,ny,nz,k


      R_cyl(1) = (r/M) * cos(phi)*sin(theta)
      R_cyl(2) = (r/M) * sin(phi)*sin(theta)
      R_cyl(3) = (r/M) *    cos(theta)

      R_O(1) = x + mc*R_cyl(1)
      R_O(2) = y + mc*R_cyl(2)
      R_O(3) = z + mc*R_cyl(3)

      R_C(1) = x - mo*R_cyl(1)
      R_C(2) = y - mo*R_cyl(2)
      R_C(3) = z - mo*R_cyl(3)

      do k=1,3
         R_CO_vec(k) = R_C(k) - R_O(k)
!         print*,R_CO_vec(k)
      enddo
      call normvxd(R_CO_vec,rCO,3)
!      print*,'---',rCO

      epot = 0.0d0
!     sum over the copper atoms
      do nx=-N_xy+1 , N_xy-1
         do ny=-N_xy+1 , N_xy-1
            do nz=-N_z+1,0

!              coordinate of the i-th copper atom
               R_CU(1) = nx * rCU_0
               R_CU(2) = ny * rCU_0
               R_CU(3) = nz * rCU_0

!              distance between O and copper atom
               do k=1,3
                  R_CUO_vec(k) = R_CU(k) - R_O(k)
               enddo
               call normvxd(R_CUO_vec,rCUO,3)

!              distance between C and copper atom
               do k=1,3
                  R_CUC_vec(k) = R_CU(k) - R_C(k)
               enddo
               call normvxd(R_CUC_vec,rCUC,3)


!              calculates eta
               call vvtxdd(R_CUC_vec,R_CO_vec,help,3)
               coseta = -help/(rCUC*rCO)

               epot = epot + A * exp(-alpha*rCUO) &
                    + B * exp(-2.0d0*beta*(rCUC-re_0)) &
                    - 2.0d0*B*(coseta**2)*exp(-beta*(rCUC-re_0))

            enddo
         enddo
      enddo

      epot = epot + F * (exp(-2.0d0*gamma*(rCO-rCO_0)) &
                       -2.0d0*exp(-gamma*(rCO-rCO_0)))

      epot = epot + F

      return
      end

!-----------------------------------------------------------------------

      subroutine normvxd(vec,norm,dim)

      implicit none

      integer                         :: i
      integer, intent(in)             :: dim
      real*8, intent(out)                :: norm
      real*8, dimension(dim), intent(in) :: vec

      norm=vec(1)*vec(1)
      do i=2,dim
         norm=norm+vec(i)*vec(i)
      enddo
      norm=dsqrt(norm)

      return
      end subroutine

!-----------------------------------------------------------------------

      subroutine vvtxdd (u,v,s,dim)

      implicit none

      integer                         :: i
      integer, intent(in)             :: dim
      real*8, dimension(dim), intent(in) :: u
      real*8, dimension(dim), intent(in) :: v
      real*8, intent(out)                :: s

      s = u(1)*v(1)
      do i = 2,dim
         s = s+u(i)*v(i)
      enddo

      return
      end subroutine




