#!/usr/bin/python

# For a given integer number, find the closest numbers with a restricted
# set of prime factors. If the factors in question are 2-3-5, the sought
# numbers are known as "regular numbers" or "Hamming numbers". The relation
# to QUANTICS is that our FFT algorithm only accepts numbers (of grid points)
# which have this form.
#
# The solution implemented here is inspired by Chapter 6.3 of the book
# "Higher Order Perl" by M.J. Dominus, see:
#   http://hop.perl.plover.com/
# Apparently the idea for the algorithm goes back to E. Dijkstra.
#
# The algorithm builds a sequence of all the numbers which can be made
# of the given prime factors, in ascending order. This is done by
# multiplying the sequence (while it is being built) element-wise by
# each of the prime factors, and then merging the resulting sequences
# together.
#
# In Haskell, the special case of factors 2-3-5 can be concisely expressed
# by the following code: (it runs ~10x faster than the Python code)
#
# nums = 1:(merge3
#            [ 2*x | x <- nums ]
#            [ 3*x | x <- nums, mod x 2 /= 0 ]
#            [ 5*x | x <- nums, mod x 2 /= 0, mod x 3 /= 0 ])
#        where merge3 xs ys zs = merge xs (merge ys zs)
#              merge (x:xs) (y:ys) = if x<y then x:(merge xs (y:ys)) else y:(merge (x:xs) ys)
#              merge xs ys = xs ++ ys
#
# For the specific problem of finding regular numbers in a given range,
# there might be more efficient algorithms, but the current code is already
# more than fast enough for any reasonable number of grid points.
#
# FO 07/2012

import sys


def usage():
    print '''\
Usage: find235 <number> [ <list-of-primes> ]
Purpose: Finds numbers of the form 2^x*3^y*5^z which are closest to the
    given one.  If <list-of-primes> is given, then these primes are used
    instead of 2,3,5.  (Actually, it suffices if the numbers in the list
    are relatively prime to each other.)
Example: find235.py 12345 2 3
'''


def prettystr(n, facs):
    counts = []
    m = n
    for f in facs:
        c = 0
        while (m%f == 0):
            m //= f
            c += 1
        counts.append(c)
    if m != 1:
        raise ValueError("%d cannot be factored with %s" % (n,facs))
    str = "%d =" % n
    for (f,c) in zip(facs,counts):
        str += " %d^%d" % (f,c)
    return str


class GenMult:
    def __init__(self, fac):
        self.fac = fac
        self.queue = []
    def append(self,n):
        self.queue.append(self.fac*n)
    def head(self):
        return self.queue[0]
    def pop(self):
        n = self.queue[0]
        self.queue[0:1] = []
        return n


class MergeGenMult:
    def __init__(self, qs):
        self.qs = qs
        self.queue = []
    def append(self,n):
        self.queue.append(n)
    def pop(self):
        # find q with smallest head
        (q0,min) = (None,None)
        for q in self.qs:
            qh = q.head()
            if (min is None) or (qh < min):
                (q0,min) = (q,qh)
        # enqueue multiples of min, but prevent redundancy
        for q in self.qs:
            if q.fac <= q0.fac:
                q.append(min)
        # store next and return current element
        self.append(q0.pop())
        cur = self.queue[0]
        self.queue[0:1] = []
        return cur


if __name__ == "__main__":
    try:
        input = sys.argv[1]
        target = int(input)
        facs = []
        for input in sys.argv[2:]:
            f = int(input)
            facs.append(f)
        if facs == []:
            facs = [2,3,5]
    except:
        usage()
        sys.exit(1)

    qs = []
    for f in facs:
        q = GenMult(f)
        q.append(1)
        qs.append(q)
    mq = MergeGenMult(qs)
    mq.append(1)

    (a,b,c) = (1,1,1)
    while c<=target:
        (a,b,c) = (b,c,mq.pop())
    for n in [a,b,c]:
        print prettystr(n, facs)

