#!/usr/bin/perl

# default values
$ny = 4;
$nx = 4;
$Delta_0=1.0;
$delta = 1.0;
$J = 0.05;
$swap = 0;
$broad = 0;
$num_flips = 0;
$name = "sgqc";
$combo_type = "subgrid";
$combo_spec = "2x2";
$sub_ny = 2;
$sub_nx = 2;
$sub_nspf = 4;

# print usage info
sub usage {
    print <<"EOF";
Usage: $0 [ options ]
Options:
    <n1>x<n2>    : specify grid dimensions (default: n1=n2=4)
    delta=<value>: specify value of delta (default: delta=1.0)
    J=<value>    : specify value of J (default: J=0.05); for J=0, the
                   operator-file won't contain the interaction section.
    swap         : swap x- and z-coordinate in Hamiltonian
    name=<string>: use filenames "name.op" and "name.inp"; set the
                   name-directory to "name"; default = "sgqc"
    combo=<spec> : specify mode-combination; possible specs:
                   <n1>x<n2>:<n3>  combine subgrids of the given dimension
                                   and use n3 SPFs for these modes
    flip=<value> : flip this number of states; can be given as a
                   percentage (i.e. flip=50%)
    broad        : use (1,1)/sqrt(2) as initial WF instead of (1,0)
    seed=<value> : seed random number generator
EOF
    exit 1;
}

# subroutine for parsing mode-combination specifications
sub parse_spec {
    my $spec = shift;
    if ($spec =~ /^(\d+)x(\d+):(\d+)$/) {
        $combo_type = "subgrid";
        $combo_spec = $1."x".$2;
        $sub_ny = $1;
        $sub_nx = $2;
        $sub_nspf = $3;
    } else {
        print STDERR "mode-combination spec '$spec' not implemented yet\n";
        exit 1;
    }
}

# parse command line
while ($opt = shift) {
    if ($opt =~ /^(\d+)x(\d+)$/) { $ny = $1; $nx = $2; }
    elsif ($opt =~ /^delta=([0-9.]+)$/) { $delta = $1*$Delta_0 }
    elsif ($opt =~ /^[jJ]=([0-9.]+)$/) { $J = $1*$Delta_0 }
    elsif ($opt =~ /^name=(.+)$/) { $name = $1 }
    elsif ($opt =~ /^swap$/) { $swap = 1 }
    elsif ($opt =~ /^combo=(.+)$/) { parse_spec($1) }
    elsif ($opt =~ /^seed=(\d+)$/) { srand($1) }
    elsif ($opt =~ /^broad$/) { $broad = 1 }
    elsif ($opt =~ /^flip=(\d+)$/) { $num_flips = $1 }
    elsif ($opt =~ /^flip=(.+)%$/) { $num_flips = int($nx*$ny*$1/100+.5) }
    else { usage }
}

$opfilename = $name . ".op";
$inpfilename = $name . ".inp";

$sigma_x = "q";
$sigma_z = "dq^2";
if (1-$swap) {
    ($sigma_x, $sigma_z) = ($sigma_z, $sigma_x);
}

########################################################################
# generation of operator-file
########################################################################

print "Generating operator-file '$opfilename'... ";
open (OPFILE, ">$opfilename") or die "$!";

$date = `date`;
chomp $date;
print OPFILE <<"EOF";
########################################################################
# SGQC (standard generic quantum computer) Hamiltonian
# file generated by gen_sgqc.plx on ${date}
########################################################################

OP_DEFINE-SECTION
title
Standard Generic Quantum Computer Model
[ ${nx}x${ny} grid, delta=${delta}, J=${J} ]
[ sigma_z = ${sigma_z} , sigma_x = ${sigma_x} ]
end-title
end-op_define-section

PARAMETER-SECTION
EOF

for ($y=1; $y<=$ny; $y++) {
    for ($x=1; $x<=$nx; $x++) {
            $gamma_tmp = $Delta_0 - $delta/2 + rand($delta);
        print OPFILE "gamma_${y}_${x} = ${gamma_tmp}\n";
    }
}
unless ($J == 0) {
    print OPFILE "\n";
    for ($y1=1; $y1<=$ny; $y1++) {
        for ($x1=1; $x1<=$nx; $x1++) {
            # interaction with left neighbor
            $J_tmp = -$J + rand(2*$J);
            $x2 = $x1-1; if ($x2==0) { $x2 = $nx }
            $y2 = $y1;
            print OPFILE "j_${y1}_${x1}_${y2}_${x2} = ${J_tmp}\n";
            # interaction with upper neighbor
            $J_tmp = -$J + rand(2*$J);
            $x2 = $x1;
            $y2 = $y1-1; if ($y2==0) { $y2 = $ny }
            print OPFILE "j_${y1}_${x1}_${y2}_${x2} = ${J_tmp}\n";
        }
    }
}
print OPFILE <<'EOF';
end-parameter-section

HAMILTONIAN-SECTION
EOF

# degrees of freedom
for ($y=1; $y<=$ny; $y++) {
    print OPFILE "modes ";
    for ($x=1; $x<=$nx; $x++) {
        print OPFILE "| s_${y}_${x} ";
        #$tmp = ($y-1)*$nx + $x;
        #print OPFILE "| s${tmp} ";
    }
    print OPFILE "\n";
}
print OPFILE "\n";

# non-interactive part
for ($y=1; $y<=$ny; $y++) {
    for ($x=1; $x<=$nx; $x++) {
        $dof = ($y-1)*$nx + $x;
        print OPFILE "2.0*gamma_${y}_${x} |${dof} ${sigma_z}\n";
    }
}
print OPFILE "\n";

# interactive part
unless ($J == 0) {
    for ($y1=1; $y1<=$ny; $y1++) {
        for ($x1=1; $x1<=$nx; $x1++) {
            # left neighbor
            $x2 = $x1-1; if ($x2==0) { $x2 = $nx }
            $y2 = $y1;
            $dof1= ($y1-1)*$nx + $x1;
            $dof2= ($y2-1)*$nx + $x2;
            if ($dof1 > $dof2) { ($dof1,$dof2) = ($dof2,$dof1) }
            print OPFILE
                "4.0*j_${y1}_${x1}_${y2}_${x2} ".
                "|${dof1} ${sigma_x} |${dof2} ${sigma_x}\n";
            # upper neighbor
            $x2 = $x1;
            $y2 = $y1-1; if ($y2==0) { $y2 = $ny }
            $dof1= ($y1-1)*$nx + $x1;
            $dof2= ($y2-1)*$nx + $x2;
            if ($dof1 > $dof2) { ($dof1,$dof2) = ($dof2,$dof1) }
            print OPFILE
                "4.0*j_${y1}_${x1}_${y2}_${x2} ".
                "|${dof1} ${sigma_x} |${dof2} ${sigma_x}\n";
        }
    }
}

print OPFILE <<'EOF';
end-hamiltonian-section

EOF

########################################################################
# operators for components of total spin
########################################################################

if ($swap) {
    print "\n Warning: spin operators might be wrong for `swap'\n";
    print " Now: sigma_z=dq^2 ,  sigma_x=q\n";
}
$operator{"x"} = { factor => "1.0", op => "dq^2" };
$operator{"y"} = { factor => "-1.0",  op => "I*dq" };
$operator{"z"} = { factor => "1.0", op => "q" };

foreach $coord (qw(x y z)) {
    print OPFILE "HAMILTONIAN-SECTION_S${coord}\n";
    for ($y=1; $y<=$ny; $y++) {
        print OPFILE "modes ";
        for ($x=1; $x<=$nx; $x++) {
            print OPFILE "| s_${y}_${x} ";
        }
        print OPFILE "\n";
    }
    print OPFILE "\n";
    for ($y=1; $y<=$ny; $y++) {
        for ($x=1; $x<=$nx; $x++) {
            $dof = ($y-1)*$nx + $x;
            print OPFILE $operator{$coord}->{"factor"}.
                  "  |${dof} ".$operator{$coord}->{"op"}."\n";
        }
    }
    print OPFILE "end-hamiltonian-section\n\n";
}

print OPFILE <<'EOF';
end-operator
EOF

close OPFILE;
print "done.\n";


########################################################################
# generation of input-file
########################################################################

print "Genrating input-file '$inpfilename'... ";
open (INPFILE, ">$inpfilename") or die "$!";

print INPFILE <<"EOF";
########################################################################
# SGQC (standard generic quantum computer) Input File
# File generated by gen_sgqc.plx on ${date}
########################################################################

RUN-SECTION
  name = ${name}
  title = SGQC on ${ny}x${nx}-grid, combinations of ${combo_spec}
  propagation
  tfinal = 100.0
  tout = 0.5
  auto = twice
  steps
  energy-not-ev
  time-not-fs
  expect = Sx, Sy, Sz
end-run-section

OPERATOR-SECTION
  opname = ${name}
end-operator-section

SPF-BASIS-SECTION
EOF

if ($combo_type eq "subgrid") {
    for ($y=1; $y<=$ny; $y++) {
        $iy = int(($y-1)/$sub_ny);
        for ($x=1; $x<=$nx; $x++) {
            $ix = int(($x-1)/$sub_nx);
            $coord = $iy."__".$ix;
            # print "Debug: y=$y x=$x iy=$iy ix=$ix coord=$coord\n";
            $mode{$coord} = [] if ($mode{$coord} == undef);
            push @{$mode{$coord}}, "s_${y}_${x}";
        }
    }
} else { die "can't be" }

foreach $key (sort keys %mode) {
    $first = 1;
    while ($dof = shift @{$mode{$key}}) {
        if ($first) { print INPFILE "  " }
        else { print INPFILE ", " }
        print INPFILE $dof;
        $first = 0;
    }
    print INPFILE " = ${sub_nspf}\n";
}

print INPFILE <<"EOF";
end-spf-basis-section

PRIMITIVE-BASIS-SECTION
# Label  DVR  N   Parameter
EOF

for ($y=1; $y<=$ny; $y++) {
    for ($x=1; $x<=$nx; $x++) {
        print INPFILE "  s_${y}_${x}  sin  2   -0.5  0.5  spin\n";
    }
}

print INPFILE <<"EOF";
end-primitive-basis-section

INTEGRATOR-SECTION
  CMF/var = 0.05 , 2.0d-6
  BS/spf  =   9  , 1.0d-6
  SIL/A   =  12  , 1.0d-6
end-integrator-section

INIT_WF-SECTION
  BUILD
EOF

# flip spins
for ($i=1; $i<=$num_flips; $i++) {
    $x = int(rand($nx))+1;
    $y = int(rand($ny))+1;
    $key = $y."__".$x;
    redo if ($flipped{$key});
    $flipped{$key} = 1;
}

for ($y=1; $y<=$ny; $y++) {
    for ($x=1; $x<=$nx; $x++) {
        print INPFILE "    s_${y}_${x}  gauss  ";
        if ($broad) { print INPFILE "0.0  0.0  1000  " }
        else { print INPFILE "0.5  0.0  0.001  " }
        $key = $y."__".$x;
        if ($flipped{$key}) { print INPFILE "pop=2" }
        else { print INPFILE "pop=1" }
        print INPFILE "\n";
    }
}

print INPFILE <<"EOF";
  end-build
end-init_wf-section

end-input
EOF

close INPFILE;
print "done.\n";

exit 0;
