#!/usr/bin/env python3

# extracts the wavepacket into plottable gnuplot format for every time frame. Needs the mctdh output file and the corresponding psi file. (Any questions or feedback to Moritz Hendl)

import os

import sys

import copy

import colorsys

import subprocess as sp


def readfile(filename):

    try:

        f = open(filename)

        out = f.readlines()

        f.close()

    except IOError:

        print("File %s does not exist!" % (filename))

        sys.exit(12)

    return out


# ======================================================================= #


def writefile(filename, content):

    # content can be either a string or a list of strings

    try:

        f = open(filename, "w")

        if isinstance(content, list):

            for line in content:

                f.write(line)

        elif isinstance(content, str):

            f.write(content)

        else:

            print("Content %s cannot be written to file!" % (content))

        f.close()

    except IOError:

        print("Could not write to file %s!" % (filename))

        sys.exit(13)


# ======================================================================= #


def get_nrmodes(data):

    # do a short run through the first timestep to get the number of

    # modes and states

    done = False

    labels = []

    nrstates = 0

    for i in range(len(data)):

        if "state =" in data[i] and "pop.: " in data[i]:

            nrstates += 1

        if "Mode expectation" in data[i]:

            line = data[i].split()

            nrmodes = 0

            sub_not_modes = 0

            while len(line) != 0:

                line = data[i + nrmodes + 1].split()

                if len(line) == 0:

                    break

                nrmodes += 1

                if not "<q>=" in line:

                    sub_not_modes += 1

                else:

                    labels.append(line[0][1:])

                done = True

        if done:

            break

    return nrstates, nrmodes - sub_not_modes, labels


# ======================================================================= #


def do_showsys_input(nrstates, nrmodes, labels):

    # create inputs to extract the wavepacket for all states for every mode

    string = "270\n"

    string += set_states(nrstates)

    for target in range(nrmodes):

        string += set_modes(target, labels)

        string += "\n10\n1\n2\n5\nadia_v%s_pes.dat\n" % labels[target]

        string += "\n10\n2\n2\n5\ndia_v%s_pes.dat\n" % labels[target]

        string += "\n10\n7\n"

        string += "\n2\n5\nv%s.dat\n" % labels[target]

    writefile("showsys.inputs", string)


# ======================================================================= #


def set_modes(target, labels):

    string = "20\n"

    for i in range(len(labels)):

        if i != target:

            string += "v%s 0\n" % labels[i]

        else:

            string += "v%s x\n" % labels[i]

    return string


# ======================================================================= #


def set_states(nrstates):

    string = "60\n%i\nall\n" % nrstates

    return string


# ======================================================================= #


def run_showsys():

    string = "showsys < showsys.inputs"

    stdoutfile = open(os.path.join("showsys.out"), "w")

    stderrfile = open(os.path.join("showsys.err"), "w")

    runerror = sp.call(string, shell=True)

    runerror = sp.call(string, shell=True, stdout=stdoutfile, stderr=stderrfile)

    stdoutfile.close()

    stderrfile.close()


# ======================================================================= #


def combine_states(label, nrstates, scaling_factor, skipframes):

    print("Extracting data for mode %s" % label)

    data = readfile("dia_v%s_pes.dat" % label)

    nrpoints = 0

    for i in range(len(data)):

        if data[i].strip() == "":

            break

        nrpoints += 1

    point = 0

    state = 0

    pes = [[0.0 for x in range(nrstates + 1)] for x in range(nrpoints)]

    for i in range(len(data)):

        if data[i].strip() == "":

            point = 0

            state += 1

            if state == nrstates:

                break

            continue

        pes[point][state + 1] = float(data[i].split()[1])

        if state == 0:

            pes[point][0] = data[i].split()[0]

        point += 1

    s = ""

    for entry in pes:

        for ent in entry:

            s += "%s " % ent

        s += "\n"

    writefile("v%s_pes" % (label), s)

    data = readfile("v%s.dat" % label)

    point = 0

    state = 0

    all_junks = []

    junk = [[0.0 for x in range(nrstates + 1)] for x in range(nrpoints)]

    for i in range(len(data)):

        if data[i].strip() == "":

            point = 0

            state += 1

            if state == nrstates:

                state = 0

                all_junks.append(copy.deepcopy(junk))

                junk = [[0.0 for x in range(nrstates + 1)] for x in range(nrpoints)]

            continue

        # junk[point][state+1] = data[i].split()[1]

        junk[point][state + 1] = (
            float(data[i].split()[1]) * scaling_factor + pes[point][state + 1]
        )

        if state == 0:

            junk[point][0] = data[i].split()[0]

        point += 1

    for i, junk in enumerate(all_junks):

        if i % skipframes != 0:

            continue

        s = ""

        for entry in junk:

            for ent in entry:

                s += "%s " % ent

            s += "\n"

        writefile("v%s_%04i" % (label, i), s)

    return len(all_junks)


# ======================================================================= #


class rgbcolor:

    """an object which you initialize with a list of integers

and whose hexcolor() routine returns a hex-coded color for a given pair (index,state)

initialize: [6,0,3]

- each non-empty group is allocated the same space on the colorwheel

- each group space is divided equally between the elements of this group


=> first group gets 180deg of the colorwheel, each element gets 30deg

=> second group is empty, does not get space

=> third group gets 180deg of the colorwheel, each element gets 60deg


- the script also allows to eliminate certain colors

- if the Index,Element pair is invalid (e.g. (2,1) for the above input), it returns white (#FFFFFF)


Example of usage:

a=[6,0,3]

R=rgbcolor(a)

for index,num in enumerate(a):

  for el in range(num):

    print index,el,R.hexcolor(index+1,el+1)

print 2,1,R.hexcolor(2,1)


                            .--------- running index over the groups

                            | .------- number of groups

                            | | .----- running index over elements

                            | | | .--- number of groups

                            | | | | .- number of elements in group

                            | | | | |

Output:                     v v v v v

1     1         #FF0000  =  0/2+0/2/6

1     2         #FF7F00  =  0/2+1/2/6

1     3         #FFFF00  =  0/2+2/2/6

1     4         #7FFF00  =  0/2+3/2/6

1     5         #00FF00  =  0/2+4/2/6

1     6         #00FF7F  =  0/2+5/2/6

3     1         #00FFFF  =  1/2+1/2/3

3     2         #0000FF  =  1/2+2/2/3

3     2         #FF00FF  =  1/2+3/2/3

2     1         #FFFFFF #invalid, hence white

"""

    def __init__(self, initlist):

        # Hues:

        # 0.0     0.15       ...

        # Red     Yellow     ...

        excluded = [[0.12, 0.22]]  # exclude yellow hues from the colorwheel

        # excluded-list must be sorted, for each element x[0]<=x[1]

        # each excluded[i][0]<=excluded[i+1][0]

        # everything between 0 and 1

        # sort the pairs

        for i, el in enumerate(excluded):

            excluded[i] = [min(el), max(el)]

        # sort the list

        excluded.sort(key=lambda x: x[0])

        # make all elements between 0 and 1

        temp1 = []

        for i, el in enumerate(excluded):

            temp1.append([min(1.0, max(0.0, el[0])), max(0.0, min(1.0, el[1]))])

        # resolve overlapping ranges

        temp2 = [[0.0, 0.0]]

        for i, el in enumerate(temp1):

            if el[0] >= temp2[-1][1]:

                temp2.append(el)

            else:

                temp2[-1][1] = el[1]

        self.excluded = temp2

        # number of non-empty groups

        self.initlist = initlist

        n = 0

        for index, el in enumerate(initlist):

            # negative numbers in initlist are not allowed, make them to zero

            self.initlist[index] = max(0, el)

            if el > 0:

                n += 1

        self.n = n  # number of non-empty groups

        self.m = len(initlist)  # number of groups

        # available colorspace self.a and shifts to skip excluded regions self.ex

        self.a = 1.0

        self.ex = [0.0 for i in self.excluded]

        for i, el in enumerate(self.excluded):

            self.a -= el[1] - el[0]

            self.ex[i] = el[1] - el[0]

        # list of starting values

        self.startlist = [0.0 for i in range(self.m)]

        for index, el in enumerate(initlist):

            if index == 0:

                continue

            if el > 0:

                self.startlist[index] = self.a / self.n + self.startlist[index - 1]

            else:

                self.startlist[index] = self.startlist[index - 1]

        # list of increments

        self.incrlist = [0.0 for i in range(self.m)]

        for index, el in enumerate(initlist):

            if el > 0:

                self.incrlist[index] = self.a / self.n / el

    def rgb_to_hex(self, rgb):

        # rgb is a list with three elements, which are floats 0<=x<=1

        color = 0

        for i in range(3):

            a = max(min(rgb[i], 1.0), 0.0)

            color += int(255 * a) * 256 ** (2 - i)

        # color=int(255*triple[0])*256**2+int(255*triple[1])*256+int(255*triple[2])

        string = hex(color)[2:].upper()

        string = "#" + "0" * (6 - len(string)) + string

        return string

    def hexcolor(self, index, el):

        if not 1 <= index <= self.m:

            return "#FFFFFF"

        if not 1 <= el <= self.initlist[index - 1]:

            return "#FFFFFF"

        deg = self.startlist[index - 1] + self.incrlist[index - 1] * (el - 1)

        for i, el in enumerate(self.excluded):

            if deg > el[0]:

                deg += self.ex[i]

        rgbtriple = colorsys.hsv_to_rgb(deg, 1, 1)

        return self.rgb_to_hex(rgbtriple)


# ======================================================================= #


def make_gnuplot(label, nrstates, skipframes, tstep):

    print("Creating gnuplot file 'plot_v%s.gp' " % label)

    R = rgbcolor([nrstates])

    s = ""

    s += "set xrange [-6.000000:6.000000]\nset yrange [0.000000:0.300000]\n"

    s += 'set xlabel "Normal mode coordinate"\nset ylabel ""\n'

    s += "set key at screen 0.30,0.85\nset border lw 2.0\n"

    s += 'set term pngcairo dashed size 640,480 font "Helvetica,12"\n'

    s += "do for [i=0:%i:%i]{\n" % (int(nrtsteps) - 1, skipframes)

    s += '  set out "pic_v%s_".i\n' % label

    s += '  set title sprintf("v8 t=%% .1f fs",i*%f)\n' % tstep

    for i in range(1, nrstates + 1):

        if i == 1:

            s += '  plot "v%s_pes"' % label

        else:

            s += '       ""       '

        s += 'u 1:%i w l tit "" lw 1.0 lc rgbcolor "%s"' % (i + 1, R.hexcolor(1, i))

        s += ", \\\n"

    for i in range(1, nrstates + 1):

        if i == 1:

            s += 'sprintf("v%s_%%04.0f",i)' % label

        else:

            s += '       ""       '

        s += 'u 1:%i w l tit "State %i" lw 2.0 lc rgbcolor "%s"' % (
            i + 1,
            i,
            R.hexcolor(1, i),
        )

        s += ", \\\n"

    s += "}"

    writefile("plot_v%s.gp" % label, s)


#  run_gnuplot('plot_v%s.gp' % label)


# ======================================================================= #


def run_gnuplot(infile):

    print("Plotting gnuplot file '%s' to create a set of PNGs" % infile)

    string = "gnuplot < %s" % infile

    stdoutfile = open(os.path.join("gnuplot.out"), "w")

    stderrfile = open(os.path.join("gnuplot.err"), "w")

    runerror = sp.call(string, shell=True)

    runerror = sp.call(string, shell=True, stdout=stdoutfile, stderr=stderrfile)

    stdoutfile.close()

    stderrfile.close()


# ======================================================================= #


data = readfile("output")

nrstates, nrmodes, labels = get_nrmodes(data)


do_showsys_input(nrstates, nrmodes, labels)

run_showsys()


# scale the amplitude of the considered populations

scaling_factor = 1.0 / 4.0

# include only every skipframeth step

skipframes = 1

# how big is each timestep #TODO get this already

tstep = 0.5


for label in labels:

    nrtsteps = combine_states(label, nrstates, scaling_factor, skipframes)

    make_gnuplot(label, nrstates, skipframes, tstep)


print(
    "Writing command line script %s to use convert to creat final gifs"
    % "create_gifs.sh"
)

s = "#!/bin/bash\n\n"

for label in labels:

    s += "convert -delay 10 -loop 0 pic_v%s_{0..%i..%i} v%s.gif\n" % (
        label,
        nrtsteps - 1,
        skipframes,
        label,
    )


writefile("create_gifs.sh", s)
