##############################################################################
#             --- Python extensions and utilities for MCTDH ---              #
#                 Module: readers                                            #
#                 Author: Oriol Vendrell                                     #
#                 Year  : 2006                                               #
##############################################################################

"""
Define functions to read/write information in/to  MCTDH files

By convection in this module the functions called (read|write)UPPERCASE
act on the MCTDH file called UPPERCASE. These routines return/expect
a dictionary with the file information. This dictionary
should containg *decoded* data, which means that these routines are
responsible for such decoding/coding.
"""
import sys
from struct import pack,unpack
from array import array
import IOFortran

from numpy import zeros




def decodeModelabels(data):
    """Split the raw modelabels record into a list of strings

    data -- raw tuple of strings (the read record) with the modelabels
    on return, list of strings
    """
    return ''.join(data).split()


##############################################################################
# Routines for interaction with input files
##############################################################################
def readInputSection(fpath,section,commentChars=['#','!']):
    """Read an input section of an input file

    f -- path to the file to read (string)
    section -- section to read (string)
    commentChars -- list of comment characters, lines starting with
                    one of those are ignored

    on return, a list of strings with the lines of the given section.
    If the file could not be open, section is not found or the end
    of file is reached before the end of the section prints a warning
    message to stdout and returns an empty string.
    """
    commentChars.append('\n')
    flist = []
    retlist = []
    startString = section + '-section'
    endString = 'end-' + startString
    emptyString = ''
    try:
        f = open(fpath,'r')
    except IOError:
        warning = "file %s could not be open\n" % (fpath)
        sys.stdout.write(warning)
        return emptyString
    istart = -1
    iend = -1
    i = 0
    try:
        flist = f.readlines()
        tmp = []
        for line in flist:
            iscomment = False
            for c in commentChars:
                if line[0] == c:
                    iscomment = True
            if not iscomment:
                l=line[:-1]
                tmp.append(l)
        flist = tmp
        for line in flist:
            if line.lower().find(startString.lower()) > -1 and not \
               line.lower().find(endString.lower()) > -1:
                istart = i
            if line.lower().find(endString.lower()) > -1:
                iend = i
                break
            i=i+1
        if istart >= 0 and iend >= 1:
            return flist[istart+1:iend]
        else:
            warning = "section %s could not be read, check input file\n" % \
                        (section)
            sys.stdout.write(warning)
            return emptyString
    finally:
        f.close()

def getOptionFromLine(s):
    """From a string return what's on the right side of a possible '=' sign

    s -- string
    If there is no '=', return ['s','']
    """
    ss = s.split('=')
    if len(ss) == 1:
        ss.append('')
    return ss

def clearLine(s):
    """Remove all the spaces from a string

    s -- string
    """
    return ''.join(s.split())

def splitWords(s):
    """Split a string at the spaces and return the resulting list of words

    s -- string
    """
    return s.split()


##############################################################################
# Routines for interaction with the autocorrelation file
##############################################################################
def readAUTO(f):
    """read the autocorrelation file

    f -- file object
    on return, two lists, one with the times, one with the complex
    autocorrelation
    """
    time = []
    auto = []
    flist= []
    if isinstance(f,file):
        flist = f.readlines()
    elif isinstance(f,str):
        ff = open(f)
        flist = f.readlines()
        ff.close
    else:
        print("f must be a strinf with the path to a file or a file object")
    for line in flist:
        if line[0] != '#':
            ss = line.split()
            time.append(float(ss[0]))
            auto.append(complex(float(ss[1]),float(ss[2])))
    return time,auto


def writeAUTO(time,auto,f):
    """write an autocorrelation file

    time -- list of times
    auto -- list of complex numbers
    f -- output file object
    """
    ini = '#    time[fs]         Re(autocorrel)     Im(autocorrel)     Abs(autocorrel)\n'
    f.write(ini)
    for i in range(len(time)):
        t = time[i]
        re = auto[i].real
        im = auto[i].imag
        ab = (re**2 + im**2)**0.5
        line = " %18.9f %18.14f %18.14f  %14.10f\n" % (t,re,im,ab)
        f.write(line)

##############################################################################
# Routines for interaction with the spec and crosspec files
##############################################################################
def readSPEC(f):
    """read a spectrum.pl file coming from the FT of an autocorrelation

    these files have 4 columns: energy  g0  g1  g2

    f -- file object
    on return, 4 lists, one with the energies, and 3 with the convoluted spectra
    """
    ener = []
    spec1 = []
    spec2 = []
    spec3 = []
    flist= []
    if isinstance(f,file):
        flist = f.readlines()
    elif isinstance(f,str):
        ff = open(f)
        flist = ff.readlines()
        ff.close
    else:
        print("f must be a strinf with the path to a file or a file object")
    for line in flist:
        if line[0] != '#':
            ss = line.split()
            ener.append(float(ss[0]))
            spec1.append(float(ss[1]))
            spec2.append(float(ss[2]))
            spec3.append(float(ss[3]))
    return ener,spec1,spec2,spec3

def readCROSSSPEC(f):
    """read a cspec.pl file coming from FT of a cross file

    these files have 4 columns: energy  re im abs

    f -- file object
    on return, 3 lists, energies,
                        complex spectrum,
                        absolute spectrum
    """
    ener = []
    specCom = []
    specAbs = []
    flist= []
    if isinstance(f,file):
        flist = f.readlines()
    elif isinstance(f,str):
        ff = open(f)
        flist = ff.readlines()
        ff.close
    else:
        print("f must be a strinf with the path to a file or a file object")
    for line in flist:
        if line[0] != '#':
            ss = line.split()
            ener.append(float(ss[0]))
            specCom.append(complex(float(ss[1]),float(ss[2])))
            specAbs.append(float(ss[3]))
    return ener,specCom,specAbs

def writeSPEC(l1,l2,f):
    """write a two columns file

    l1 -- first list of floats
    l2 -- second list of floats
    f -- output file object or string
    """
    ff=None
    closeFile=False
    if isinstance(f,file):
        ff = f
    elif isinstance(f,str):
        ff = open(f,'w')
        closeFile=True
    for i in range(len(l1)):
        il1 = l1[i]
        il2 = l2[i]
        line = " %18.9f %18.14f\n" % (il1,il2)
        ff.write(line)
    if closeFile:
        ff.close


##############################################################################
# Routines for interaction with the dvr file
##############################################################################
def readDVR(f):
    """read a DVR file

    f -- path to the file to be read
    on return, dictionary with the DVR file information

    the returned dictionary should contain decoded data

    dvr file records
       record num.  description                   format
       10.          startup string                100c
       11.          array lengths                 5i
                    ortdim,dvrdim,fftdim,expdim,sphdim
       12.          zort, pointers to ort array   [N]i
       13.          zdvr, pointers to dvr array   [N]i
       14.          zfft, pointers to fft array   [N]i
       15.          zexp, pointers to exp array   [N]i
       16.          zsph, pointers to sph array   [N]i
       17.          ort, coordinates in each dof  [N*..]d

    """
    records,data = IOFortran.readFortranBinFile2List(f)
    dvr = {}
    # Decode binary file
    _readFileVer(dvr,data)
    _readSysDef(dvr,data)
    _readStartUp(dvr,data)
    fmtNi = str(dvr["ndof"]) + 'i'
    dvr["totalDims"] = unpack('5i',data[11])
    ortdim = dvr["totalDims"][0]
    dvrdim = dvr["totalDims"][1]
    fftdim = dvr["totalDims"][2]
    expdim = dvr["totalDims"][3]
    sphdim = dvr["totalDims"][4]
    dvr["zort"] = unpack(fmtNi,data[12])
    dvr["zdvr"] = unpack(fmtNi,data[13])
    dvr["zfft"] = unpack(fmtNi,data[14])
    dvr["zexp"] = unpack(fmtNi,data[15])
    dvr["zsph"] = unpack(fmtNi,data[16])
    dvr["ort"] = unpack(str(ortdim)+'d',data[17])
    dvr["trafo"] = unpack(str(dvrdim)+'d',data[18])
    dvr["dvrord"] = unpack('i',data[19])[0]
    dvrord = dvr["dvrord"]
    dvr["dvrmat"] = unpack(str(dvrdim*dvrord)+'d',data[20])
    dvr["fftp"] = unpack(str(fftdim)+'d',data[21])
    dvr["hin"] = unpack(str(2*fftdim)+'d',data[22])
    dvr["rueck"] = unpack(str(2*fftdim)+'d',data[23])
    dvr["fftfak"] = unpack(str(fftdim)+'i',data[24])
    dvr["exphin"] = unpack(str(2*expdim)+'d',data[25])
    dvr["exprueck"] = unpack(str(2*expdim)+'d',data[26])
    dvr["jsph"] = unpack(str(sphdim)+'i',data[27])
    dvr["msph"] = unpack(str(sphdim)+'i',data[28])
    dvr["kinsph"] = unpack(str(3*sphdim)+'d',data[29])
    return dvr


def dvrGetModeLabels(filename):
    """
    dvrGetModeLabels(filename) -> list of modelabes

    Returns a list of modelabels from a dvr file, where
    'filename' is the path to the dvr file
    """

    dvr = readDVR(filename)
    return dvr["modelabel"].split()


def dvrCoordsDegF(f,degf):
    """
    dvrCoordsDegF(f,degf) -> list of coordinate values | None

    f - path to the DVR file
    degf - number of the degree of freedom
    on return, list with the coordinates
    returns None in case of not success
    """

    dvr = readDVR(f)
    return ortFromNum(degf,dvr)


def ortFromLabel(label,dvr):
    """Return list with the coordinates of the degf given by a modelabel

    label -- modelabel of the deg. of f. for which coordinates are wanted
    dvrdict -- dictionary with info of a dvr file

    on return, list of coordinates
    """
    ml = decodeModelabels(dvr["modelabel"])
    i = 0
    for m in ml:
        if label == m:
            break
        i=i+1
    i=i+1
    return ortFromNum(i,dvr)


def ortFromNum(degf,dvr):
    """Return list with the coordinates of the degf referred by its number

    num -- number of the degf
    dvr -- dictionary with info of a dvr file

    on return, list of coordinates
    """
    try:
        ipointer = dvr["zort"][degf-1] - 1
        fpointer = ipointer + dvr["gdim"][degf-1]
        return list(dvr["ort"][ipointer:fpointer])
    except IndexError:
        return None


##############################################################################
# Routines for interaction with a binary dump of the psi vector
##############################################################################
def readPSIdump(fpath,alen,modlen):
    """
    Read a psi vector dump

    fpath  -- path to file with the dump
    alen   -- length of A vector (how many terms)
    modlen -- list, lenght of each mode (how many terms)

    array with the a vector
    list of arrays, each with a vector of 1 mode
      - avec
      - modes
          - mode 1
          - mode 2
          - ...
    """
    l,d = IOFortran.readFortranBinFile2List(fpath)
    totlen = alen
    for n in modlen:
        totlen = totlen + n
    totlen2 = totlen*2
    fmt = str(totlen2)+'d'
    try:
        p = unpack(fmt,d[0])
    except:
        msg = """
        psidump could not be unpacked, check that the
        supplied legths correspond to the psi vector.
        Just look into the log file of the corresponding
        MCTDH calculation.
        """
        sys.exit(msg)
    avec = zeros(alen,complex)
    modes = []
    for m in modlen:
        tmp = zeros(m,complex)
        modes.append(tmp)
    i = 0
    for k in range(alen):
        avec[k] = p[2*i] + complex(str(p[2*i+1])+'j')
        i+=1
    m = 0
    for l in modlen:
        for n in range(l):
            modes[m][n] = p[2*i] + complex(str(p[2*i+1])+'j')
            i+=1
        m+=1
    return avec,modes





##############################################################################
# Routines for interaction with the natpot file
##############################################################################
def readNATPOT(f):
    """read a NATPOT file

    f -- path to the file to be read
    on return, dictionary with the NATPOT file information

    the returned dictionary should contain decoded data

    natpot file records
       record num.  description                   format

    """
    records,data = IOFortran.readFortranBinFile2List(f)
    dict = {}
    _readFileVer(dict,data)
    _readSysDef(dict,data,2)
    _readStartUp(dict,data,11)
    k = _readNatFit1(dict,data,12)
    k = _readGridDef(dict,data,k+1)
    k = _readNatFit2(dict,data,k+1)
    k = _readNatFitOut(dict,data,k+1)
    _readVpotPars(dict,data,k+1)
    return dict


def writeNATPOT(dict,fpath):
    """Code the dictionary dict with natpot information and write the
       binary records to fpath

    dict -- dictionary with the natpot file information
    fpath -- path to new natpot file
    """
    dbin = []
    dbin.append(pack('d',dict["fileversion"]))
    dbin.append(dict["liftp"])  # liftp is assumed to be already coded
    _writeSysDef(dict,dbin)
    dbin.append(dict["startup"])
    _writeNatFit1(dict,dbin)
    _writeGridDef(dict,dbin)
    _writeNatFit2(dict,dbin)
    _writeNatFitOut(dict,dbin)
    _writeVpotPars(dict,dbin)
    IOFortran.writeList2FortranBinFile(fpath,dbin)



def _readNatFit1(dict,data,startrecord=1):
    """Read general records of a natpot file

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, since this routine does not read a fix ammount
    of records, return the index of the last record read

    Read the general records of a natpot file which are
    written in the "wrfit" routine BEFORE "call wrgrddef"
    """
    j = startrecord
    dict["liftp"] = data[1] # keep coded, is obsolete info
    dict["lcount"] = unpack('i',data[j])[0]
    for i in range(dict["lcount"]):
        dict["docstring"+str(i)] = data[j+1+i]
    j = j + 1 + dict["lcount"]
    tmp = unpack(str(dict["ndof"]+1) + 'i',data[j])
    dict["modc"] = tmp[0]
    dict["onedipot"] = tmp[1:]
    dict["lpconm"] = unpack('i',data[j+1])[0]
    return j+1

def _writeNatFit1(dict,data):
    try:
        data.append(pack('i',dict["lcount"]))
        for i in range(dict["lcount"]):
            data.append(dict["docstring"+str(i)])
        tmp = (dict["modc"],)+dict["onedipot"]
        data.append(_listToBin(tmp,'i'))
        data.append(pack('i',dict["lpconm"]))
    except KeyError:
        print("in _writeNatFit1, key not found")


def _readNatFit2(dict,data,startrecord=1):
    """Read general records of a natpot file

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, since this routine does not read a fix ammount
    of records, return the index of the last record read

    Read the general records of a natpot file which are
    written in the "wrfit" routine AFTER "call wrgrddef"
    """
    j = startrecord
    fmtNModi = str(dict["ndofnmode"][1]) + 'i'
    dict["pdim"] = unpack(fmtNModi,data[j])
    i = 0
    dict["oned"] = []
    for r in range(dict["ndof"]):
        if dict["onedipot"][i]:
            fmt = str(dict["gdim"][i]) + 'i'
            dict["oned"].append(unpack(fmt,data[j+1+i]))
            i=i+1
    return j+i

def _writeNatFit2(dict,data):
    try:
        data.append(_listToBin(dict["pdim"],'i'))
        i = 0
        for r in range(dict["ndof"]):
            if dict["onedipot"][i]:
                data.append(_listToBin(dict["pdim"][i],'i'))
                i=i+1
    except KeyError:
        print("in _writeNatFit2, key not found")


def _readNatFitOut(dict,data,startrecord=1):
    """Read fitt records of a natpot file

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, since this routine does not read a fix ammount
    of records, return the index of the last record read

    Read records of a natpot file which were
    written in "fitout" routine
    All the records are stored in a macro structure that contains
    the whole set of coefficients and vectors
    """
    dict["natpotfit"] = []
    j = startrecord
    dimbef,dimaft,psubdim = _dimbefdimaftpsubdim(dict)
    for i in range(dict["ndofnmode"][1]):
        # loop over modes
        if i == dict["modc"]-1:
            # this is the contracted mode, read coefficients
            fmt = str(psubdim[i]) + 'd'
            for k in range(dimaft):
                for l in range(dimbef):
                    dict["natpotfit"][i].append(unpack(fmt,data[j]))
                    j=j+1
        else:
            # non contracted, read evecs
            fmt = str(psubdim[i]) + 'd'
            for k in range(dict["pdim"][i]):
                dict["natpotfit"][i].append(unpack(fmt,data[j]))
                j=j+1
    return j-1

def _writeNatFitOut(dict,data):
    try:
        dimbef,dimaft,psubdim = _dimbefdimaftpsubdim(dict)
        for i in range(dict["ndofnmode"][1]):
            if i == dict["modc"]-1:
                j=0
                for k in range(dimaft):
                    for l in range(dimbef):
                        data.append(_listToBin(dict["natpotfit"][i][j],'d'))
                        j=j+1
            else:
                j=0
                for k in range(dict["pdim"][i]):
                    data.append(_listToBin(dict["natpotfit"][i][j],'d'))
                    j=j+1
    except KeyError:
        print("in _writeNatFitOut, key not found")


def _readVpotPars(dict,data,startrecord=1):
    """Read vpot parameters from the natpot file

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, nothing

    Read records of a natpot file which were
    written in "writevpotpars" routine
    By now just return the unformatted records, it is not clear
    if it is of utility to decode them...
    """
    dict["vpotpars"] = []
    j = startrecord
    for record in data[j:]:
        dict["vpotpars"].append(record)

def _writeVpotPars(dict,data):
    try:
        for record in dict["vpotpars"]:
            data.append(record)
    except KeyError:
        print("in _writeVpotPars, key not found")



##############################################################################
# General read/write routines
##############################################################################

def _readFileVer(dict,data,startrecord=0):
    """Read the file version from the record data list

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, this routine has no return value, only adds entries to dict
    """
    dict["fileversion"] = unpack('d',data[startrecord])[0]

def _readStartUp(dict,data,startrecord=10):
    """Read the start up from the record data list

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, this routine has no return value, only adds entries to dict
    """
    dict["startup"] = data[startrecord]

def _readSysDef(dict,data,startrecord=1):
    """Read the system definition from the record data list

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, this routine has no return value, only adds entries to dict

    The records processed by this routine are the ones written by
    the mctdh routine "wrdvrdef"

    SysDef file records
       record num.  description                   format
       1.           number of degs of freedom     i
       2.           modelabels                    [N*16]c
       3.           num grid points               [N]i
       4.           mbaspar                       i
       5.           basis type code               [N]i
       6.           log2int                       [N]i
       7.           integer basis parameters      [N*mbaspar]i
       8.           real basis parameters         [N*mbaspar]d
       9.           start and end points          [N*2]d
    """
    j = startrecord
    dict["ndof"] = unpack('i',data[j])[0]
    ndof = dict["ndof"]
    fmtNi = str(ndof) + 'i'
    dict["modelabel"] = data[j+1]
    dict["gdim"] = unpack(fmtNi,data[j+2])
    dict["mbaspar"] = unpack('i',data[j+3])[0]
    fmtMNi = str(ndof*dict["mbaspar"]) + 'i'
    fmtMNd = str(ndof*dict["mbaspar"]) + 'd'
    dict["basis"] = unpack(fmtNi,data[j+4])
    dict["log2int"] = unpack(fmtNi,data[j+5])
    dict["ipbaspar"] = unpack(fmtMNi,data[j+6])
    dict["rpbaspar"] = unpack(fmtMNd,data[j+7])
    fmt2Nd = str(ndof*2) + 'd'
    dict["xend"] = unpack(fmt2Nd,data[j+8])

def _writeSysDef(dict,data):
    try:
        data.append(pack('i',dict["ndof"]))
        data.append(dict["modelabel"])
        data.append(_listToBin(dict["gdim"],'i'))
        data.append(pack('i',dict["mbaspar"]))
        data.append(_listToBin(dict["basis"],'i'))
        data.append(_listToBin(dict["log2int"],'i'))
        data.append(_listToBin(dict["ipbaspar"],'i'))
        data.append(_listToBin(dict["rpbaspar"],'d'))
        data.append(_listToBin(dict["xend"],'d'))
    except KeyError:
        print("in _writeVpotPars, key not found")


def _readGridDef(dict,data,startrecord=1):
    """Read the grid definition from the record data list

    dict -- dictionary in use, where the results are stored
    data -- the data records
    on return, since this routine does not read a fix ammount
    of records, return the index of the last record read

    The records processed by this routine are the ones written by
    the mctdh routine "wrgrddef"

    SysDef file records
       record num.  description                   format
    """
    j = startrecord
    dict["dentype"] = unpack('i',data[j])[0]
    dict["ndofnmode"] = unpack('2i',data[j+1])
    ndof = dict["ndofnmode"][0]
    nmode = dict["ndofnmode"][1]
    fmtNModi = str(nmode) + 'i'
    dict["nspfdof"] = unpack(fmtNModi,data[j+2])
    dict["spfdof"] = []
    for i in range(nmode):
        fmt = str(dict["nspfdof"][i]) + 'i'
        dict["spfdof"].append(unpack(fmt,data[j+3+i]))
    k = j+3+i
    fmtNDofi = str(ndof) + 'i'
    dict["dofspf"] = unpack(fmtNDofi,data[k+1])
    dict["logicgrid"] = unpack('3i',data[k+2])
    return k+2

def _writeGridDef(dict,data):
    try:
        nmode = nmode = dict["ndofnmode"][1]
        data.append(pack('i',dict["dentype"]))
        data.append(_listToBin(dict["ndofnmode"],'i'))
        data.append(_listToBin(dict["nspfdof"],'i'))
        for i in range(nmode):
            data.append(_listToBin(dict["spfdof"][i],'i'))
        data.append(_listToBin(dict["dofspf"],'i'))
        data.append(_listToBin(dict["logicgrid"],'i'))
    except KeyError:
        print("in _writeGridDef, key not found")


def _dimbefdimaftpsubdim(dict):
    """calculate dimbef, dimaft and psubdim
    """
    dimbef = 1
    dimaft = 1
    psubdim = []
    for i in range(dict["ndofnmode"][1]):
        dict["natpotfit"].append([])
        if i < dict["modc"]-1:
            dimbef = dimbef * dict["pdim"][i]
        if i > dict["modc"]-1:
            dimaft = dimaft * dict["pdim"][i]
        ptmp = 1
        for k in dict["spfdof"][i]:
            ptmp = ptmp * dict["gdim"][k-1]
        psubdim.append(ptmp)
    return dimbef,dimaft,psubdim


def _listToBin(l,fmt):
    a = array(fmt)
    a.fromlist(list(l))
    return a.tostring()
