#!/usr/bin/env python3
"""
-------------------------------------------------------------------------------
"""
# standard modules
import sys
import os
import math
import re
import subprocess
import time

# third party modules
import scipy
import numpy

# local modules
from parsingtools import *
from mtools import importUserSource
from mtools import *
from mmath import metropolisStepGrid
from mjob import MJob, JobError
from IOFiles import dvrGetModeLabels, dvrCoordsDegF
from parsingtools import absExpandPath


# This class requires numpy version 1.0.1 or higher
_REQUIRED_NUMPY_VERSION = '1.0.1'

# default file names
_FILENAME_ENERGIES = "energies"
_FILENAME_PES = "pes"
_FILENAME_INDEX = "dvrindex"
_FILENAME_COORDINATES = "coordinates"
_FILENAME_LOG = "checkpes.log"
_FILENAME_INPUTCOPY = "input_checkpes"
_FILENAME_RATIO = "ratio"
_FILENAME_DVR = "dvr"
_DIRNAME_DVR = "dvr"

# update log every _NUPDATE samples
_NUPDATE = 100000
# if Monte carlo step is not accepted for _TRYCOUNT_KILL
# tries in a row then terminate
_TRYCOUNT_KILL = 100000

# default executables
_VMINMAX_EXE = "vminmax84"

# the potential routine
potential = None


class CheckPesJob(MJob):

    def _parseCommandLine(self, opts=[]):
        """
        Parsing of command line options
        """
        self.steps = None
        self.ldebug = False
        self.loverwrite = False
        self.lmnd = False
        self.name = None
        self.trajectorypath = None
        self.trajectory_done = False

        for o, a in opts:
            if o == "-D":
                self.name = a
            elif o == '-n':
                try:
                    self.steps = int(a)
                except ValueError:
                    message = "Option -n must be followed by an integer."
                    raise ValueError(message)
            elif o == "-m" and a == "nd":
                self.lmnd = True
            elif o == "-d" and a == "eb":
                self.ldebug = True
            elif o == '-w':
                self.loverwrite = True
            elif o == "-t":
                self.trajectorypath = absExpandPath(a)
                if not os.path.exists(self.trajectorypath) or \
                       not os.path.isfile(self.trajectorypath):
                    message = "Object does not exist or is not a file: "
                    message += str(self.trajectorypath)
                    raise JobError(message)
                self.trajectory_done = True
            else:
                message = "Unknown option: " + str(o) + str(a)
                raise JobError(message)


    def __init__(self, inp, args=[]):

        # init base class
        MJob.__init__(self)

        # check for numpy version
        checkNumpyVersion(_REQUIRED_NUMPY_VERSION)

        self.input = inp

        # command line arguments
        self._parseCommandLine(args)

        # check for owerwrite permission and name
        self.loverwrite = self.loverwrite or \
                          bool(self.input.runsection.loverwrite)
        if self.name is None:
            self.name = self.input.runsection.name
        self.outdir = os.path.join(os.getcwd(),self.name)

        # make name if required
        self.mkNameDir(self.lmnd)
        self.checkNameDir()

        # start logging
        self.logfilename = os.path.join(self.outdir,_FILENAME_LOG)
        self.logSetup(self.logfilename)

        # save a copy of the input
        self.saveInputCopy()

        # copy rest of the job parameters
        if self.trajectorypath is None:
            if self.input.runsection.trajectorypath is not None:
                self.trajectorypath = \
                    absExpandPath(self.input.runsection.trajectorypath)
                if not os.path.exists(self.trajectorypath) or \
                       not os.path.isfile(self.trajectorypath):
                    message = "Object does not exist or is not a file: "
                    message += str(self.trajectorypath)
                    raise JobError(message)
                self.trajectory_done = True
        else:
            message = "Trajectory is obtained from " + str(self.trajectorypath)
            self.log.debug(message)

        self.temperature = self.input.runsection.temperature
        if self.steps < 0:
            self.steps = self.input.runsection.steps

        self.lwalkonly = self.input.runsection.lwalkonly
        if self.input.runsection.dvrdir is not None:
            self.dvrdir = absExpandPath(self.input.runsection.dvrdir)
        else:
            self.dvrdir = None
        if self.input.runsection.pesdir is not None:
            self.pesdir = absExpandPath(self.input.runsection.pesdir)
        else:
            self.pesdir = None

        if self.dvrdir is None:
            message = "Setting dvrdir = pesdir: "
            message += str(self.pesdir)
            self.log.info(message)
            self.dvrdir = self.pesdir

        if self.input.runsection.nham is not None:
            self.nham = int(self.input.runsection.nham)
        else:
            self.nham = 1

        message = "Number of the Hamiltonian is "
        message += str(self.nham) + "."
        self.log.debug(message)

        if self.input.runsection.vminmax_exe is not None:
            self.vminmax_exe = self.input.runsection.vminmax_exe
        else:
            self.vminmax_exe = _VMINMAX_EXE

        self.lwrcoord = self.input.runsection.lwrcoord

        self.input.checkConsistency()
        if self.input.message != "":
            self.log.error(self.input.message)
            sys.exit()

        # read the DVR
        dvrfilename = os.path.join(self.dvrdir,"dvr")
        message = "Reading DVR file " + dvrfilename
        self.log.debug(message)
        try:
            self.coord_labels = dvrGetModeLabels(dvrfilename)
        except:
            message = "Could not read DVR file: " + dvrfilename
            raise JobError(message)
        self.ndof = len(self.coord_labels)
        self.grids = []
        self.gdim = numpy.zeros(len(self.coord_labels), dtype=int)
        self.refindex = numpy.zeros(len(self.coord_labels), dtype=int)
        self.fix = numpy.zeros(len(self.coord_labels), dtype=bool)

        for n,lab in enumerate(self.coord_labels):
            try:
                self.grids.append(numpy.array(dvrCoordsDegF(dvrfilename,n+1)))
            except:
                message = "Could not retrieve grid for coordinate " + lab
                message += " in DVR file " + dvrfilename
                raise JobError(message)

            self.gdim[n] = len(self.grids[n])
            idx = self.input.initsection.coord_labels.index(lab)
            self.refindex[idx] =  self.input.initsection.refindex[idx]
            self.fix[n]  =  self.input.initsection.fix[idx]
        self.refindex -= 1 # zero-based index here



        # initialize the job
        self.curr_index = self.refindex.copy()
        self.curr_position = self._idx2coord(self.curr_index)
        self.count_accepted = 1 # we count the initial point
        self.count_tried = 0
        self.curr_energy = 0.e0
        self.ratio = 0.e0

        # import the potential energy surface etc
        self.user_source = str(self.input.runsection.user_source)
        if self.input.runsection.potential_routine is None:
            potential_routine = "potential"
        else:
            potential_routine = self.input.runsection.potential_routine


        global potential
        potential = importUserSource(self.user_source, potential_routine)

        self.curr_pval = self._probFunc(self.curr_position)
        self.log.info("Init complete.")


    def _idx2coord(self,idx):
        """
        Convert the grid index to coordinates

        _idx2coord(idx) --> Q

        idx:   integer list
        Q:     numpy.array(float),

        """

        q = numpy.zeros( self.ndof , dtype=numpy.float64)
        for n,g in enumerate(idx):
            q[n] = self.grids[n][g]
        return q

    ####################################################
    # I/O routines
    ####################################################
    def _openFiles(self):
        "open output files"

        wmode = "w"
        lreplace = False

        if not self.trajectory_done:
            fname = os.path.join(self.outdir,_FILENAME_INDEX)
            self.files.index = openFile(fname,wmode,self.loverwrite)

        if self.lwrcoord:
            fname = os.path.join(self.outdir,_FILENAME_COORDINATES)
            self.files.coordinates = openFile(fname,wmode,self.loverwrite)

        fname = os.path.join(self.outdir,_FILENAME_ENERGIES)
        self.files.energies = openFile(fname,wmode,self.loverwrite)

        fname = os.path.join(self.outdir, _FILENAME_RATIO)
        self.files.ratio =  openFile(fname,wmode,self.loverwrite)


    def _writeOutput(self):
        "Write output files."

        if self.lwrcoord:
            message = ""
            for elem in self.curr_position:
                message += "%20.12f  " % elem
            message += "\n"
            self.files.coordinates.write(message)

        if not self.trajectory_done:
            message = ""
            for elem in self.curr_index:
                message += "%4i " % (elem+1)
            message += "\n"
            self.files.index.write(message)

        message = "%20.12f" % self.curr_energy
        message += "\n"
        self.files.energies.write(message)

        message = "%8.4f" % (self.ratio) + "\n"
        self.files.ratio.write(message)

        self.files.flushWriteables()


    def load(self):
        """
        CHECKPES DOES NOT SUPPORT RESTARTS. DO NOT USE!
        """
        raise JobError("Program checkpes does not support retarts.")


    def _probFunc(self, Q):
        """
        Get the probability associated with a point given
        as a coordinate vector.

        _probFunc(Q) --> p

        Q:     float list
        q:     float

        """
        pot=numpy.float64(potential(Q))
        return numpy.exp(-pot/self.temperature)



    def _probFuncGrid(self, idx):
        """
        Get the probability associated with a point given
        as a DVR index.

        _probFuncGrid(idx) --> p

        idx:   integer list
        q:     float

        """
        Q = self._index2Q(idx)
        return self._probFunc(Q)

    def _index2Q(self,idx):
        """
        Convert the DVR index to a coordinate vector

        _index2Q(idx) --> q

        idx: list of integers
        q:   numpy.array()

        """
        q = numpy.zeros(self.ndof , dtype=numpy.float64)
        for i in range(self.ndof):
            q[i] = self.grids[i][idx[i]]
        return q


    def stepMetropolis(self):
        "Try until one metropolis step has been accepted."
        accepted = False
        idx = self.curr_index.copy()
        pval = self.curr_pval
        trycount = 0
        imin = numpy.zeros(self.ndof, dtype=int)
        imax = self.gdim.copy()-1
        while not accepted:
            accepted, idx, pval = metropolisStepGrid(idx,pval,
                                                     self._probFuncGrid,
                                                     imin, imax,self.fix)
            self.count_tried += 1
            trycount += 1
            if trycount > _TRYCOUNT_KILL:
                message = "Tried %d times but no Metropolis step " % _TRYCOUNT_KILL
                message += "has been accepted. Terminating."
                raise JobError(message)
            if (trycount % _NUPDATE) == 0:
                message = "Tried %d times but no Metropolis step " % trycount
                message += "has been accepted. Temperature too low? "
                message += "Keep trying..."
                self.log.warn(message)

        self.count_accepted += 1
        self.curr_index = idx
        self.curr_pval = pval
        self.ratio = float(self.count_accepted)/self.count_tried
        self.curr_position = self._index2Q(self.curr_index)
        self.curr_energy = float(potential(self.curr_position))


    def doStep(self):
        self.stepMetropolis()


    def save(self):
        pass


    def execute(self):
        "Execute the job"

        # build the tracectory
        if not self.trajectory_done:
            message = "Obtaining trajectory with %i samples..." % self.steps
            self.log.info(message)
            self.trajectorypath = os.path.join(self.outdir,_FILENAME_INDEX)

            self._openFiles()

            message = "# Metropolis, npoints="+ str(self.steps) +\
                      ", temperature=" + str(self.temperature) + " a.u." +\
                      "(no skipped steps or burn-in)\n"
            self.files.index.write(message)
            self.files.energies.write(message)

            self.curr_position = self._index2Q(self.curr_index)
            self.curr_energy = float(potential(self.curr_position))
            self.ratio = 0.e0

            self._writeOutput()

            while self.count_accepted < self.steps-1:
                self.doStep()
                self._writeOutput()
            self.trajectory_done = True
            self.files.closeAll()
            self.files.popClosed()

            if self.ratio < 0.4e0 or self.ratio > 0.9e0:
                message = "Acceptance rate of Nonte-Carlo steps is"
                message += " %4.1f" % (self.ratio*100.e0) + " %."
                self.log.warn(message)


        if not self.lwalkonly:
            message = "Starting vminmax..."
            self.log.info(message)

            args = [self.vminmax_exe,
                    "-c", os.path.join(self.outdir, _FILENAME_ENERGIES),
                    "-t", self.trajectorypath,
                    "-o", os.path.join(self.outdir, "vmm_checkpes"),
                    "-i", str(self.pesdir),
                    "-ham", str(self.nham)]
            if self.loverwrite is True:
                args.append("-w")

            cmd=re.sub('[,\'\[\]]',' ',str(args))
            cmd=re.sub(' +',' ',cmd)
            message = "Command line: " + cmd
            self.log.debug(message)

            #start vminmax
            cwd = os.getcwd()
            os.chdir(self.outdir)
            vmm_proc=subprocess.Popen(args, stdout=subprocess.PIPE)
            os.chdir(cwd)

            returncode=vmm_proc.poll()
            pollcount = 0
            while returncode is None:
                time.sleep(1) # sleep a second if not done yet
                returncode=vmm_proc.poll()
                pollcount += 1
                if (pollcount % 6) == 0:
                    message = "Waiting for " + str(self.vminmax_exe)
                    message += ". Time elapsed: %d minutes" %  (pollcount / 60)
                    self.log.debug(message)

            data=vmm_proc.communicate() #read from PIPE
            if returncode != 0:
                message="Error while running " + str(self.vminmax_exe)
                message+='. Program returned a non-zero exit code.'
                message+='Message: \n'+ str(data)
                raise JobError(message)

            self.log.info("-"*60)
            lfound = False
            i = 1
            for line in data[0].split("\n"):
                if line.startswith(" ***** Statistics *****"):
                    lfound = True
                if lfound and i < 21:
                    if line.startswith("External file"):
                        line1 = "Potential routine:"
                    elif line.startswith("Difference (PES - External)"):
                        line1 = "Difference (PES - Potential routine)"
                    else:
                        line1 = line
                    self.log.info(line1)
                    i += 1

            self.log.info("-"*60)


    def cleanup(self):
        "Flush output and close files"
        self.files.closeAll()


