#!/usr/bin/env python3

from minput import MSection, InputError
from parsingtools import *

import numpy

class RunSectionCheckPes(MSection):

    def __init__(self, lines):
        MSection.__init__(self,lines)
        if lines[0].upper() != "RUN-SECTION":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. First line reads: " \
                     + lines[0]
            raise InputError(message)
        elif lines[-1].lower() != "end-run-section":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. Last line reads: " \
                     + lines[-1]
            raise InputError(message)

        self.openstr = lines.pop(0)
        self.closestr = lines.pop(-1)

        key = "name"
        strval, lines = find_param(key, lines)
        if strval:
            self.name = strval
        else:
            self.missing_key_error(key)

        key = "user-source"
        strval, lines = find_param(key, lines)
        if strval:
            self.user_source = strval
        else:
            self.missing_key_error(key)

        key = "temperature"
        strval, lines = find_param(key, lines)
        if not strval:
            self.missing_key_error(key)
        temp = strval.split(',')
        if len(temp) == 1:
            self.temperature = float(temp[0])
            if self.temperature < 0.e0:
                raise InputError("temperature cannot be below zero.")
        elif len(temp) == 2:
            unit = getunit(temp[1])
            if unit:
                self.temperature = numpy.float64(temp[0])*unit
            else:
                message = "Temperature: Unknown unit: " + str(temp[1])
                raise InputError(message)
        else:
            message = "temperature: Too many keywords: " + str(temp)
            raise InputError(message)

        key = "steps"
        strval, lines = find_param(key, lines)
        if strval:
            self.steps = int(strval)
        else:
            self.steps = None

        # optional key-value-pairs

        key = "potential-routine"
        strval, lines = find_param(key, lines)
        if strval:
            self.potential_routine = strval
        else:
            self.potential_routine = None


        key = "dvrdir"
        strval, lines = find_param(key, lines)
        if strval:
            self.dvrdir = str(strval)
        else:
            self.dvrdir = None

        key = "pesdir"
        strval, lines = find_param(key, lines)
        if strval:
            self.pesdir = str(strval)
        else:
            self.pesdir = None

        key = "vminmax-exe"
        strval, lines = find_param(key, lines)
        if strval:
            self.vminmax_exe = str(strval)
        else:
            self.vminmax_exe = None

        key = "nham"
        strval, lines = find_param(key, lines)
        if strval:
            self.nham = int(strval)
        else:
            self.nham = None

        key = "tracectory"
        strval, lines = find_param(key, lines)
        if strval:
            self.trajectorypath = str(strval)
        else:
            self.trajectorypath = None

        # flags
        self.loverwrite, lines = find_flag("overwrite", lines)
        self.lwalkonly, lines = find_flag("walk-only", lines)
        self.lwrcoord, lines = find_flag("wrcoord", lines)

        if self.pesdir is None and self.dvrdir is None:
            message = "RUN-SECTION: at least one of 'pesdir' "
            message += "or 'dvrdir' has to be set."
            raise InputError(message)

        # consistency checks:
        if self.pesdir is None and self.lwalkonly is False:
            message = "RUN-SECTION: pesdir is not given. You "
            message += "may set the keyword 'walk-only' if "
            message += "an evaluation of the pes-file is not "
            message += "needed. In this case only a "
            message += "trajectory is obtained."
            raise InputError(message)

        if self.steps is None and self.trajectorypath is None:
            message = "RUN-SECTION: either the number of steps "
            message += "or an external travectory file are needed."
            raise InputError(message)

        if len(lines):
            message="Input contains unrecognized data: " + str(lines)
            raise InputError(message)


    def  __str__(self):
        string =  str(self.openstr)

        # key-value-pairs
        string += "\n name = " + str(self.name)
        string += "\n temperature = " + str(self.temperature)
        string += "\n user-source = " + str(self.user_source)
        if self.potential_routine is not None:
            string += "\n potential-routine = " + str(self.potential_routine)
        if self.steps is not None:
            string += "\n steps = " + str(self.steps)
        if self.pesdir is not None:
            string += "\n pesdir = " + str(self.pesdir)
        if self.dvrdir is not None:
            string += "\n dvrdir = " + str(self.dvrdir)
        if self.vminmax_exe is not None:
            string += "\n vminmax-exe = " + str(self.vminmax_exe)
        if self.nham is not None:
            string += "\n nham = " + str(self.nham)
        if self.trajectorypath is not None:
            string += "\n trajectory = "+str(self.trajectorypath)
        #flags
        if self.loverwrite is True:
            string += "\n overwrite"
        if self.lwalkonly is True:
            string += "\n walk-only"
        if self.lwrcoord is True:
            string += "\n wrcoord"

        string += "\n" + self.closestr + "\n"
        return string





