#!/usr/bin/env python3
"""----------------------------------------------------------------------------
Purpose:     Use a Monte-Carlo sampling to check a sampled PES file against
             an analytic potential routine.

Usage:       chkpes.py [-D -n -mnd -deb -w -t -ver -h -? ] inputfile.inp

  -D <dir>   Denotes the directory where files are written to
             (name in .inp file ignored).
  -n <int>   Use <int> Monte-Carlo steps (value in .inp is ignored).
  -mnd       Make name directory.
  -deb       Run in debug mode (more detailed logging).
  -w         Allow to overwrite existing data.
  -t <file>  Run with a given trajectory stored in <file>.
  -ver       Print version info.
  -h -?      Print this help text.
----------------------------------------------------------------------------"""

# builtin modules
import getopt
import sys
import os

# local modules
import checkpes_inp
import checkpes_job
from mjob import JobError
from minput import InputError
from parsingtools import absExpandPath

#version
from mversion import __version__, __status__, __author__
from mversion import __copyright__, __maintainer__, __email__
from mversion import versionInfo

if __name__=="__main__":


    # command line options
    caller_name = sys.argv.pop(0)
    if "-h" in sys.argv or "-?" in sys.argv:
        sys.exit(__doc__)
    if "-ver" in sys.argv:
        versionInfo(progname = caller_name,
                version = __version__,
                status = __status__,
                author = __author__,
                copyright = __copyright__,
                maintainer = __maintainer__,
                support = __email__)
        sys.exit()
    if len(sys.argv) > 0:
        inpfilename = sys.argv.pop(-1)
    else:
        sys.exit(__doc__)
    opts, args = getopt.getopt(sys.argv, 'D:n:m:d:wt:')


    # read imput file
    if inpfilename:
        lininp = len(inpfilename)
        if inpfilename[lininp-4:lininp] == ".inp":
            inpfilename = absExpandPath(inpfilename)
            try:
                inp = checkpes_inp.CheckPesInput(inpfilename)
                inp.checkConsistency()
            except InputError as I:
                message = "ERROR while reading input: " +str(I) + "\n"
                sys.exit(message)
            except Exception as E:
                message = "ERROR while reading input: " +str(E) + "\n"
                sys.exit(message)
            except:
                sys.stderr.write("ERROR: Unknown.")
                sys.exit("ERROR: Unknown error while reading input")
        else:
            message = "Input file name must the extension '.inp'. "
            message += 'Filename given: ' + str(inpfilename)
            sys.exit(message)
    else:
        message = "No input file."
        sys.exit(message)



    # create a job instance
    try:
        job = checkpes_job.CheckPesJob(inp, opts)
    except Exception as E:
        message = "Error while creating job: "
        message += str(E)
        sys.exit(message)
    except:
        message = "Unknown error while creating job."
        sys.exit(message)

    # execute the job instance
    try:
        job.execute()
    except KeyboardInterrupt as K:
        message = "KeyboardInterrupt: Exiting unfinished job. "
        message += "Restart is not supported."
        job.log.info(message)
        sys.exit('KeyboardInterrupt: exit.')
    except JobError as J:
        message = "Job failed. Reason: " + str(J)
        job.log.error(message)
    except Exception as E:
        job.log.exception(E)
        sys.exit("A serious error has occured. Aborting.")
    except:
        message = "A serious error has occured. Reason: unknown"
        job.log.error(message)
        sys.exit("Aborting.")

    # close files etc.
    job.cleanup()
    sys.exit()





