#!/usr/bin/env python3

# generic modules
import os

# third party modules
import numpy

# local modules
import mtools
from clustertools import str_mc2int_mc

# global
potential = None


class ClusterError(RuntimeError):
    pass


class ClusterBase(object):
    """
    Base class for Cluster objecs. Stores DOF, grids, order and modecombinations


    """

    def __init__(self, dof, coord_labels, mode_combinations, qref,
                 potfunc):

        global potential
        potential = potfunc

        # check consistency of input:
        for d in dof:
            if not d in coord_labels:
                message = "DOF " + str(d) + " missing in coord_labels."
                raise ClusterError(message)

        # reference point
        self.qref = numpy.array(qref).copy()

        # degrees of freedom and system info
        self.str_dof = [d for d in dof]
        self.coord_labels =  tuple([c for c in coord_labels])


        self.int_dof = [list(coord_labels).index(d) for d in self.str_dof]


        self.mode_combinations = tuple([tuple(m) for m in mode_combinations])
        self.int_mode_combinations = str_mc2int_mc(coord_labels, mode_combinations)

        # check on which modes we are operating, what order and if
        # we have 'holes' (holes need a DOF-Potfit)
        modes = []
        for m, mode in enumerate(self.mode_combinations):
            for d in self.str_dof:
                if d in mode and not m in modes:
                    modes.append(m)
        self.modes = tuple(modes)


        self.order = len(self.modes)

        # state =  0 -> not initialised, empty
        # state =  1 -> PES cut has been calculated
        # state =  2 -> cluster has been calculated
        self.state = 0

        # dependency links
        self.dependencies = []
        self.dependents = []
        self.dependencies_checked = False
        self.isroot = None

    def __str__(self):

        string = "Cluster of order " + str(self.order)
        string += " containing DOF " + str(self.str_dof)
        string += " in modes " + str([m+1 for m in self.modes])
        if self.isroot:
            string += " [Root-Cluster]"
        return string



    def intersectDOF(self,other):
        """
        intersectDOF(other) -> list of common DOF.
        """
        intersection = []
        for d in self.int_dof:
            if d in other.int_dof:
                intersection.append(d)
        return tuple(intersection)


    def intersectModes(self,other):
        """
        intersectModes(other) -> list of common modes.
        """
        intersection = []
        for m in self.modes:
            if m in other.modes:
                intersection.append(m)
        return tuple(intersection)

    	
    def isContainedIn(self,other):
        """
        isContainedIn(other) --> bool

        Result is True if instance has identical DOFs
        as other or a subset.
        """
    	
        for d in self.int_dof:
            if not d in other.int_dof:
                return False
        return True

    def isContainedInModes(self,other):
        """
        isContainedInModes(other) --> bool

        Result is True if instance has identical modes
        as other or a subset.
        """
        for m in self.modes:
            if not m in other.modes:
                return False
        return True



    def isIdenticalTo(self, other):
        """
        isIdenticaTo(other) --> bool

        Result is True if instance has identical
        set of DOFs as other.
        """

        if len(self.int_dof) != len(other.int_dof):
            return False
        else:
            return self.isContainedIn(other)

    	

    def checkHoles(self):
        """
        chekHoles() --> list of modes containing 'holes'.
        """
        mc = [numpy.zeros(len(m), dtype=int) for m in self.int_mode_combinations]
        holes = []


        for n,m in enumerate(self.int_mode_combinations):
            for d in self.int_dof:
                if d in m:
                    i = list(m).index(d)
                    mc[n][i] = 1

        for n,m in enumerate(mc):
            start = m[0]
            nchange = 0
            for i in m:
                if i != start:
                    start = i
                    nchange += 1
            if nchange > 1:
                holes.append(n)

        return holes

    @property
    def isSingleMode(self):
        if len(self.modes)  == 0 or len(self.modes)  == 1:
            return True
        else:
            return False

    def initDependencies(self, clusterlist=[]):
        """
        initDependencies(clusterlist=[]) --> None

        Searches through a list of other clusters and finds
        the clusters it depends on, i.e., those that are needed
        to calculate the current cluster from the PES-cut.
        Also dependents, i.e., clusters that require this
        node are identified.

        Pointers to the instances are stored in the
        local dependnencies and dependnents lists.
        """

        self.dependencies = []
        self.dependents = []

        for cl in clusterlist:

            if cl.order < self.order:

                if cl.isContainedInModes(self) or \
                       len(cl.modes) == 0:
                    self.dependencies.append(cl)

            elif cl.order == self.order \
                     and not id(cl) == id(self):
                if self.isContainedIn(cl):
                    message = "Clusterlist contains redundant "
                    message += "clusters: "
                    message += str(self) + " and " + str(cl)
                    message += "."
                    raise ClusterError(message)
            elif cl.order > self.order:
                if self.isContainedInModes(cl):
                    self.dependents.append(cl)

        if len(self.dependencies) == 0:
            self.isroot = True
        else:
            self.isroot = False

        self.dependencies_checked = True



