#!/usr/bin/env python3

# generic modules
import os

# third party modules
import numpy

# local modules
from mtools import openFile, multiIndexIterator
from parsingtools import absExpandPath
from cluster_base import ClusterBase, ClusterError

# float to string formatting:
# exponential notation with 10 decimal places
FORMAT = "% .10e"
# I/O buffer size for PES data
# 16777216= 2**24 -> 128 MB real*8
BUFSIZE = 100000


potential = None



class ClusterB(ClusterBase):
    """
    ClusterB for 'Build' jobs.

    Initialization:
      ClusterB(dof, modecombinations, ref,
               potfunc, pgrids, allowholes=True)

    dof:              list of coordinate labels
                      contained in this cluster.
    coord_labels:     list of ALL coordinate labels
                      defined in input file
    modecombinations: List of tuples of integers.
                      integers: number of DOF.
                      tuples group dof in combined modes.
    qrefidx           DVR grid index if the reference point
    potfunc:          function that returns the PES value
                      for a given coordinate vector
    pgrids:           list of arrays with primitive grids points
                      ('ort' in MCTDH)

    """

    def __init__(self, dof, coord_labels,
                 modecombinations, qrefidx,
                 potfunc, pgrids,
                 pes_min=None, pes_max=None,
                 cluster_min=None, cluster_max=None):



        global potential
        potential = potfunc

        # qref must live on a gridpoint. extract qref from index
        self.qrefidx = numpy.array(qrefidx).copy()

        self.qref = numpy.zeros(len(self.qrefidx), dtype=float)
        for n, i in enumerate(self.qrefidx):
            self.qref[n] = pgrids[n][i]


        # Init base class now that we have qref
        ClusterBase.__init__(self, dof, coord_labels,
                             modecombinations, self.qref,
                             potfunc)


        # primitive grids: make a local copy
        self.pgrids = [numpy.array(grid).copy() for grid in pgrids]
        self.gdim = numpy.zeros(len(dof), dtype=int)
        for n,d in enumerate(dof):
            idx = coord_labels.index(d)
            self.gdim[n] = len(pgrids[idx])

        if len(self.gdim) > 0:
            self.totgdim = self.gdim.prod()
        else:
            self.totgdim = 1

        self.dof = [d for d in dof]

        self.ndof = len(self.gdim)

        self._cumdim = numpy.array([], dtype=int)
        if self.ndof > 0:
            self._cumdim = self.gdim.cumprod()

        self.dofrefindex = numpy.zeros(len(self.dof),dtype=int)
        for n,d in enumerate(self.dof):
            i = coord_labels.index(d)
            self.dofrefindex[n] = self.qrefidx[i]

        # cluster data.
        self.data = None

        # files for data to be stored to
        self.cutfilename = None
        self.clusterfilename = None

        # energy cut-offs
        self.pes_min = pes_min
        self.pes_max = pes_max
        self.cluster_min = cluster_min
        self.cluster_max = cluster_max

    def _initRestart(self, potfunc):

        global potential
        if potential is None:
            potential = potfunc


    def initDependencies(self, clusterlist):
        """
        initDependencies(clusterlist=[]) --> None

        Searches through a list of other clusters and finds
        the clusters it depends on, i.e., those that are needed
        to calculate the current cluster from the PES-cut and those
        this cluster is needed by.

        Pointers to the needed instances are stored in the
        local dependnencies and dependnents lists.
        """

        # first call the base class initDependencies
        # to initialize the pointers.
        ClusterBase.initDependencies(self,clusterlist)


        # for the dependents there might be the need
        # to calculate a PES cut that does not contain
        # all DOF as we cannot substract the complete cluster
        # when some of the DOF in self are set to qref in a dependent.
        # we use a dictionary to map all needed cuts with a DOF tuple
        # and initialize it to None

        containsroot = False
        for cl in self.dependencies:
            containsroot = containsroot or cl.isroot
        if not containsroot and not self.isroot:
            message = "Initialization of dependencies failed: "
            message += "Root cluster is missing for "
            message += str(self)
            raise ClusterError(message)



    def combIndex2GridIndex(self,i):
        """
        combIndex2GridIndwx(i) --> tuple(integers)

        Maps the combined index onto the tuple of indices
        """
        if i < 0:
            raise IndexError("Negative indices are not supportet.")
        if i >= self.totgdim:
            raise IndexError("Index overflow.")

        idx = numpy.zeros(self.ndof, dtype=int)
        # single point :
        if len(idx) == 0:
            return idx

        for n in range(self.ndof-1,0,-1):
            idx[n] = int(i/self._cumdim[n-1])
            i -= idx[n]*self._cumdim[n-1]
        idx[0] = i
        return idx


    def qFromGridIndex(self,idx):
        """
        qFromGridIndex(i) --> Q

        From the local index of grid points
        of this node calculate the position Q in
        physical coordinates. Q is a numpy array of type float
        and contains all pysical DOF of the system.
        """

        q = self.qref.copy()
        for n,f in enumerate(self.int_dof):
            q[f] = self.pgrids[f][idx[n]]
        return q


    def qFromCombIndex(self,i):
        """
        qFromCombIndex(i) --> Q

        From the combined index calculate the position Q in
        physical coordinates. Q is a numpy array of type float
        and contains all pysical DOF of the system.
        """
        idx = combIndex2GridIndex
        return self.qFromGridIndex(idx)


    def gridIndex2CombIndex(self,idx):

        i = 0
        for n in range(self.ndof-1,0,-1):
            i += idx[n]*self._cumdim[n-1]
        i += idx[0]
        return i


    def free(self):
        """
        free() --> None

        Frees RAM by detatching the cluster/PES data from self.
        This relies on the garbage collection of Python, i.e.,
        it does not work if self.data is referenced elsewehre.
        """
        self.data = None




    def setCutFileName(self, filename):
        """
        setCutFile(filename) --> None

        Set the name of the file to which the PES-cut is to be written.
        File name should be relative to name dir.
        """
        self.cutfilename = filename


    def setClusterFileName(self, filename):
        """
        setClusterFile(filename) --> None

        Set the name of the file to which the cluster is to be written.
        File name should be relative to name dir.
        """
        self.clusterfilename = filename



    def calculateCut(self, name=None, loverwrite=False):
        """
        CalculateCut(loverwrite=False) --> None

        Calculates the PES cut on the grid
        using the PES routine and writes to file.
        No data is stored in RAM.

        Existing data from previous or current runs is overwriten if
        'loverwrite' is True.

        Note: 'cutfilename' has to be set before this routine is called.
              See 'setCutFileName'.
        """


        if name is not None:
            self.setCutFileName(name)

        cutfile = openFile(self.cutfilename, "w", loverwrite)

        if self.totgdim <= 0:
            message = "Internal error: Cluster not initialized."
            raise ClusterError(message)

        # just a single point
        if self.totgdim == 1:
            message = FORMAT % float(potential(self.qref)) + "\n"
            cutfile.write(message)
            cutfile.close()
            self.state=1
            return

        # total grid
        index = numpy.zeros(len(self.dof),int)
        q = self.qref.copy()
        bufcount = 0



        #        for i,idx in enumerate(counter):
        for i in range(self.totgdim):

            idx = self.combIndex2GridIndex(i)

            if bufcount == 0:
                buf = numpy.zeros(BUFSIZE,dtype=float)

            for n,f in enumerate(self.int_dof):
                q[f] = self.pgrids[f][idx[n]]

            buf[bufcount] =  float(potential(q))
            bufcount += 1

            if bufcount == BUFSIZE or i == self.totgdim - 1:

                # cutoffs:
                if self.pes_max is not None:
                    buf = numpy.where(buf < self.pes_max, buf,
                                      self.pes_max)

                if self.pes_min is not None:
                    buf = numpy.where(buf > self.pes_min, buf,
                                      self.pes_min)


                for j in range(bufcount):
                    cutfile.write(FORMAT % buf[j] + "\n")
                bufcount = 0
                cutfile.flush()

        cutfile.close()
        self.state = 1
        return





    def calculateCluster(self, clusterfilename="", loverwrite=False,
                         cutfilename = ""):

        """
        calculateCluster(self, clusterfilename="", loverwrite=False,
                         cutfilename = "") --> None

        Calculate the cluster for this node and write it to file.

        Requires that the PES of this node and clusters of higher nodes
        (lower order in coordinates) are already calculated.

        if cutfilename or clusterfilename are given, the default file names
        are reset and the PES will be read
        from the file specified with cutfilename and the cluster will be
        written to the file specified with clusterfilename. If not given,
        the default file names must be set.

        If loverwrite is True, existing data will be replaced.


        """


        if not self.dependencies_checked:
            message = "Cluster " + self.idxStr() + ":"
            message += " Dependencies not initialized."
            raise ClusterError(message)

        if cutfilename != "":
            self.cutfilename = cutfilename

        if bool(self.cutfilename) is False:
            message = "calculateCluster: No filename for PES cut defined."
            raise ClusterError(message)
        if clusterfilename != "":
            self.clusterfilename = clusterfilename
        if bool(self.clusterfilename) is False:
            message = "calculateCluster: No filename for cluster defined."
            raise ClusterError(message)

        if self.state == 0:
            message = str(self) + "\n"
            message += "No PES-cut. Call 'calculateCut' first."
            raise ClusterError(message)

        if self.state != 1 and self.state != 2 and  self.state != 0:
            message = "Unknown state of node " + str(selfstate)
            raise ClusterError(message)

        # here state is 1 or 2: we have a PES-cut but not
        # yet substracted the lower order clusters.
        # The cut is on file an needs to be loaded first
        self.loadCut()

        # run over all lower order clusters we depend on
        for node in self.dependencies:

            # check if those nodes contain clusters
            if node.state != 2:
                message = "Dependensies must contain a cluster. "
                message += "Failed: " + node.idxStr()
                raise ClusterError(message)

            # make sure that node has data

            if node.data is None:
                node.loadCluster()

            nodeslc = []
            for d in node.int_dof:
                if d in self.int_dof:
                    nodeslc.append(slice(None,None,None))
                else:
                    nodeslc.append(node.qrefidx[d])

            # usually the dependencies will be orders of
            # magnitude smaller then self. so we use use something like
            # self.data[:,:,i,:,j,k] -= other[i,j,k]
            # and let numpy do the numerics on a slice of the PES.
            # while we just extract a single number of the node.
            # Build an index generator for this:

            if len(node.int_dof) > 0:
                # node contains more then one point of data
                start = []
                nsteps = []
                for n,d in enumerate(self.int_dof):
                    if not d in node.int_dof:
                        start.append(0)
                        nsteps.append(self.gdim[n])
                    else:
                        start.append(self.qrefidx[d])
                        nsteps.append(0)

                counter = multiIndexIterator(start,nsteps,delta=1)
            else:
                # node just contains the data for qref
                counter = None


            # build the slice for the first time. later
            # the slice objects remain unchanged, just the indices are
            # changed using an index map.
            slc = []
            for d in self.int_dof:
                if d in node.int_dof:
                    slc.append(slice(None,None,None))
                else:
                    slc.append(0)

            if counter is not None:
                for idx in counter:

                    # build the slice of self from which the
                    # node value is to be substracted
                    for n,d in enumerate(self.int_dof):
                        if d not in node.int_dof:
                            slc[n] = idx[n]


                    self.data[slc] -= node.data[nodeslc]

            # node.data is just a single point
            else:
                self.data -= node.data


            # deallocate memory of the node data
            node.data = None


        # cutoffs:
        if self.cluster_max is not None:
            self.data = numpy.where(self.data < self.cluster_max,
                                    self.data, self.cluster_max)

        if self.cluster_min is not None:
            self.data = numpy.where(self.data > self.cluster_min,
                                    self.data, self.cluster_min)



        self.state = 2
        # save immediately
        self.writeCluster(loverwrite = loverwrite)






    def loadCut(self, filename = ""):
        """
        loadCut(filename) -> None

        Load the PES-cut from file <filename>.
        """

        if filename == "":
            filename = self.cutfilename
        if bool(filename) is False:
            message = "loadCut: No filename defined."
            raise ClusterError(message)

        self.data = numpy.loadtxt(filename)

        if len(self.data.shape) != 0:

            if len(self.data) != self.totgdim:
                message = str(self)
                message += ": data does not match the size "
                message += "of the primitive grid "
                message += "of this cluster. Size should be: "
                message += str(self.totgdim)
                message += " but is: " + str(len(data))
                message += ". Aborting."
                self.state = 0
                self.data = None
                raise ClusterError(message)


        self.data = self.data.reshape(self.gdim, order="F")
        self.state = 1


    def writeCluster(self, filename="", loverwrite=False):
        """
        writeCluster(filename="", loverwrite=False) -> None

        Write the PES-cluster to file. If loverwrite is True,
        replace existing file.
        If 'filename' is given, the default file name is reset.

        Exception: ClusterError if Cluster does not contain a cluster.
        """



        if self.state != 2:
            message = "Node does not contain a cluster-data. State is " + str(self.state)
            if state == 0:
                message += " (not initialized)."
            elif state == 1:
                message += " (contains PES-Cut)."
            else:
                message += " (Undefined. Something went terribly wrong)."
            raise ClusterError(message)


        if os.path.exists(filename) and not loverwrite:
            message = "The file " + filename + " exists. No "
            message += "permission to overwrite."
            raise ClusterError(message)

        if filename != "":
            self.clusterfilename = filename

        self.data = self.data.reshape(self.totgdim, order="F")
        numpy.savetxt(self.clusterfilename,self.data, fmt=FORMAT)
        self.data = self.data.reshape(self.gdim, order="F")


    def loadCluster(self, filename=""):
        """
        loadCluster(filename="") -> None

        Load the cluster-data from file.
        """

        if filename == "":
            filename = self.clusterfilename
        if bool(filename) is False:
            message = "loadCluster: No filename defined."
            raise ClusterError(message)

        self.data = numpy.loadtxt(filename)
        if len(self.data.shape) != 0:

            if len(self.data) != self.totgdim:
                message = str(self)
                message += ": data does not match the size "
                message += "of the primitive grid "
                message += "of this cluster. Size should be: "
                message += str(self.totgdim)
                message += " but is: " + str(len(data))
                message += ". Aborting"
                self.state = 0
                self.data = None
                raise ClusterError(message)

        self.data = self.data.reshape(self.gdim, order="F")
        self.state = 2



    def idxStr(self):
        """
        idxStr() --> string

        Return the index string of the cluster as it
        would be found in the input file.
        """

        idx = []
        for n,mc in enumerate(self.mode_combinations):
            complete = True
            for d in mc:
                complete = complete and d in self.str_dof
            if complete:
                idx.append(str(n+1))
            else:
                for d in mc:
                    if d in self.str_dof:
                        idx.append(str(d))
        string = "(" + ", ".join(idx) + ")"
        return string
