#!/usr/bin/env python3

# generic modules
import os

# third party modules
import numpy

# local modules
from cluster_base import ClusterBase, ClusterError
from mtools import intersection
# CONSTANTS:

# float to string formatting:
# exponential notation with 10 decimal places
FORMAT = "%.10e\n"


potential = None

class ClusterA(ClusterBase):
    """
    ClusterA for 'Analysis' jobs.
    """

    def __init__(self, dof, coord_labels, modecombinations, qref,
                 potfunc):



        # Init base class
        ClusterBase.__init__(self, dof, coord_labels, modecombinations, qref,
                             potfunc)

        global potential
        potential = potfunc

        self.state = 0

        self.cuts = {}
        self.clusters = {}
        tpl = tuple(self.int_dof)
        self.cuts[tpl] = None
        self.clusters[tpl] = None


    def _initRestart(self, potfunc):

        global potential
        if potential is None:
            potential = potfunc


    def initDependencies(self, clusterlist):
        """
        initDependencies(clusterlist=[]) --> None

        Searches through a list of other clusters and finds
        the clusters it depends on, i.e., those that are needed
        to calculate the current cluster from the PES-cut and those
        this cluster is needed by.

        Pointers to the needed instances are stored in the
        local dependnencies and dependnents lists.
        """

        # first call the base class initDependencies
        # to initialize the pointers.
        ClusterBase.initDependencies(self,clusterlist)


        # for the dependents there might be the need
        # to calculate a PES cut that does not contain
        # all DOF as we cannot substract the complete cluster
        # when some of the DOF in self are set to qref in a dependent.
        # we use a dictionary to map all needed cuts with a DOF tuple
        # and initialize it to None

        containsroot = False
        for cl in self.dependencies:
            containsroot = containsroot or cl.isroot
        if not containsroot and not self.isroot:
            message = "Initialization of dependencies failed: "
            message += "Root cluster is missing for "
            message += str(self)
            raise ClusterError(message)

        self.cuts = {}
        self.clusters = {}

        # first is self
        tpl = tuple(self.int_dof)
        self.cuts[tpl] = None
        self.clusters[tpl] = None

        # then come the others
        for cl in self.dependents:
            tpl = self.intersectDOF(cl)
            if not tpl in self.cuts:
                self.cuts[tpl] = None
                self.clusters[tpl] = None


    def calculateCut(self, qin):

        for tpl in self.cuts:
            q = self.qref.copy()
            for d in tpl:
                q[d] = qin[d]

            self.cuts[tpl] = float(potential(q))
        self.state = 1


    def calculateCluster(self):

        if self.state != 1:
            message = "Node does not contain PES cuts: "
            message += str(self)
            raise ClusterError(message)

        for tpl in self.cuts:
            self.clusters[tpl] = self.cuts[tpl]

            for cl in self.dependencies:

                if cl.state != 2:
                    message = "Node does not contain clusters: "
                    message += str(cl)
                    raise ClusterError(message)

                tpl2 = tuple(intersection(tpl,cl.int_dof))
                #corrupted dependencies if this raises an exception:
                self.clusters[tpl] -= cl.clusters[tpl2]

        self.state = 2


    def idxStr(self):
        """
        idxStr() --> string

        Return the index string of the cluster as it
        would be found in the input file.
        """

        idx = []
        for n,mc in enumerate(self.mode_combinations):
            complete = True
            for d in mc:
                complete = complete and d in self.str_dof
            if complete:
                idx.append(str(n+1))
            else:
                for d in mc:
                    if d in self.str_dof:
                        idx.append(str(d))
        string = "(" + ", ".join(idx) + ")"
        return string




