#!/usr/bin/env python3
"""
-------------------------------------------------------------------------------
Purpose:     Statistical analysis for the cluster expansion
             of a potential surface.

Usage:       clusterana [-deb -ver -h -? ] inputfile.inp

  -deb       Run in debug mode (more detailed logging).
  -D <name>  The name directory of a clusterstat run.
  -ver       Print version info.
  -h -?      Print this help text.

Forther analysis of data produced by statistical analysis program
'clusterstat'. This requires an (at least partly) completed run
of 'clusterstat'. The system information is obtained from the restart
file of 'clusterstat'.

-------------------------------------------------------------------------------
"""

# standard modules
import getopt
import sys
import os

# local modules
import clusterana_inp
import clusterana_job
from mjob import JobError
from minput import InputError
from parsingtools import absExpandPath

from mversion import __version__, __status__, __author__
from mversion import __copyright__, __maintainer__, __email__
from mversion import versionInfo

if __name__=="__main__":

    # command line options
    caller_name = sys.argv.pop(0)

    if "-h" in sys.argv or "-?" in sys.argv:
        sys.exit(__doc__)
    if "-ver" in sys.argv:
        versionInfo(progname = caller_name,
                version = __version__,
                status = __status__,
                author = __author__,
                copyright = __copyright__,
                maintainer = __maintainer__,
                support = __email__)
        sys.exit()
    if len(sys.argv) > 0:
        inpfilename = sys.argv.pop(-1)
    else:
        sys.exit(__doc__)

    opts, args = getopt.getopt(sys.argv, 'D:d:w')

    # imput file
    if inpfilename:
        if inpfilename.endswith(".inp"):
            inpfilename = absExpandPath(inpfilename)
            inp = clusterana_inp.ClusteranaInput(inpfilename)
            inp.checkConsistency()
        else:
            message = "Input file name must the extension '.inp'. "
            message += 'Filename given: ' + str(inpfilename)
            sys.exit(message)
    else:
        message = "No input file."
        sys.exit(message)

    # create a job instance and execute
    job = clusterana_job.ClusteranaJob(inp,opts)

    try:
        job.execute()
    except KeyboardInterrupt as K:
        message = 'KeyboardInterrupt: saving files. This may take a while...'
        job.log.info(message)
        sys.exit('KeyboardInterrupt: exit.')
    except Exception as E:
        job.log.exception(E)
        message = "Exit after exception."
        job.log.error(message)
        sys.exit(message)
    except:
        message = "Exit after unknown error."
        job.log.error(message)
        sys.exit(message)

    job.cleanUp()





