#!/usr/bin/env python3

# standard modules
import os
import sys
import getopt
import pickle
# third party modules
import numpy
import scipy

# local modles
from mjob import MJob, FileContainer, JobError

from parsingtools import getunit
from parsingtools import absExpandPath
from clustertools import expandWildCards
from clustertools import idx2doflist
from clustertools import doftpl2modetpl
from mtools import importUserSource
from mtools import checkNumpyVersion
from mtools import getNonUnique
from mtools import openFile
from mtools import substractseq
from mtools import incrAllInt
from mjob import MJob, JobError


# This class requires numpy version 1.1.0 or higher
_REQUIRED_NUMPY_VERSION = '1.1.0'

# update log every _NUPDATE samples
_NUPDATE = 10000

# filenames that occur in this program:
# input
_FILENAME_ENERGIES = "energies"
_FILENAME_CLUSTERS = "clusters"
_FILENAME_COORDINATES = "coordinates"
_FILENAME_WEIGHTS = "switch"
_FINENAME_RESTART = "restart"
_FILENAME_TRAJWEIGHT = "trajweight"

# output
_FILENAME_MEAN_ERR_ORD = "mean_error_order_ana"
_FILENAME_RMS_ERR_ORD = "rms_error_order_ana"
_FILENAME_CURR_ERR_ORD = "curr_error_order_ana"

_FILENAME_MEAN_ERR_REF = "mean_error_ref_ana"
_FILENAME_RMS_ERR_REF = "rms_error_ref_ana"
_FILENAME_CURR_ERR_REF = "curr_error_ref_ana"

_FILENAME_MEAN_ORD = "mean_order_ana"
_FILENAME_RMS_ORD = "rms_order_ana"
_FILENAME_CURR_ORD = "curr_order_ana"

_FILENAME_MEAN_REF = "mean_ref_ana"
_FILENAME_RMS_REF = "rms_ref_ana"
_FILENAME_CURR_REF = "curr_ref_ana"

_FILENAME_AUTO = "auto"
_FILENAME_DEN1D = "density1d"
_FILENAME_DEN2D = "density2d"
_FILENAME_DELTA = "deltaE"
_FILENAME_INPUTCOPY = "input_analysis"
_FILENAME_LOG = "analysis.log"
_FILENAME_STATISTICS = "statistics"
_FILENAME_LARGEST_DELTA = "samples_largest_delta"
_FILENAME_LARGEST_EXACT = "samples_largest_exact"
_FILENAME_LARGEST_APPROX = "samples_largest_approx"
_FILENAME_LOWEST_DELTA = "samples_lowest_delta"
_FILENAME_LOWEST_EXACT = "samples_lowest_exact"
_FILENAME_LOWEST_APPROX = "samples_lowest_approx"


# global user provided function (optional)
# to check wether coordinates are in a
# given region
_isInRange = None


# The job class
class ClusteranaJob(MJob):

    def __parseCommandLine(self, opts=[]):

        self.name = None
        self.ldebug = False
        self.loverwrite = False

        for o, a in opts:
            if o == "-D":
                self.name = a
            elif o == "-d" and a == "eb":
                self.ldebug = True
            elif o == '-w':
                self.loverwrite = True
            else:
                message = "Unknown option: " + str(o) + str(a)
                raise JobError(message)



    def __init__(self, inp, args=[]):


        MJob.__init__(self)

        # check numpy version
        checkNumpyVersion(_REQUIRED_NUMPY_VERSION)

        # command line arguments

        self.__parseCommandLine(args)

        inp.checkConsistency()
        self.input = inp



        if self.name is None:
            self.name = self.input.runsection.name
        self.loverwrite = bool(self.loverwrite) or \
                          bool(self.input.runsection.loverwrite)

        # output directory
        self.outdir = os.path.join(os.getcwd(),self.name)
        self.outdir = absExpandPath(self.outdir)
        self.checkNameDir()

        # dump a copy of the input

        self.saveInputCopy(_FILENAME_INPUTCOPY)


        # Start the logging system
        self.logfilename = os.path.join(self.outdir,_FILENAME_LOG)
        self.logSetup(self.logfilename)

        # extract data from input (needs system data)
        self.lcluster_statistics = self.input.runsection.lcluster_statistics
        self.nsteps = self.input.runsection.nsteps
        if self.input.runsection.nskip is not None:
            self.nskip = int(self.input.runsection.nskip)
        else:
           self.nskip = 0
        if  self.input.runsection.nlog is not None:
            self.nlog = self.input.runsection.nlog
        else:
            self.nlog =0
        if self.input.runsection.nskiplog is not None:
            self.nskiplog = self.input.runsection.nskiplog
        else:
            self.nskiplog = 0

        self.countlow =  self.input.runsection.countlow
        self.counthigh = self.input.runsection.counthigh

        if self.input.runsection.countunit is not None:
            unit = getunit(self.input.runsection.countunit)
            if unit is None:
                message = "Error while importing Input: "
                message += "unit not recognized: "
                message += str(self.input.runsection.countunit)
                message += "."
                raise JobError(message)
            if self.countlow is not None:
                self.countlow *= unit
            if self.counthigh is not None:
                self.counthigh *= unit


        self.lden1d = inp.runsection.lden1d
        self.lden2d = inp.runsection.lden2d
        self.nbins = self.input.runsection.nbins
        self.lauto = inp.runsection.lauto
        self.lrms = inp.runsection.lrms
        self.lmean = inp.runsection.lmean
        self.outunit = inp.runsection.outunit
        if self.outunit is None:
            self.outunit = "au"
        self.outunit_float = 1.e0/getunit(self.outunit)


        global _isInRange
        if inp.coordsection is not None:
            self.routine = self.input.coordsection.routine
            self.source = self.input.coordsection.source
            if self.routine is not None and self.source is not None:
                _isInRange = importUserSource(self.source, self.routine)
            self.coord_labels = [str(lab) for lab in inp.coordsection.coord_labels]
            self.ranges_start = numpy.array(inp.coordsection.ranges_start)
            self.ranges_end = numpy.array(inp.coordsection.ranges_end)
            if _isInRange is not None:
                if len(self.coord_labels) > 0\
                   or len(self.ranges_start) > 0\
                   or len(self.ranges_end) > 0:
                    message = "Error: user-provided routine and coordinates range are given. Choose one."
                    raise JobError(message)
        else:
            self.routine = None
            self.source = None
            self.coord_labels =  []
            self.ranges_start = None
            self.ranges_end = None


        # read restart file
        message = "Reading restart file: "
        fname = os.path.join(self.outdir,_FINENAME_RESTART)
        message += fname
        self.log.info(message)
        f=open(fname,mode="rb")
        self.restart = pickle.load(f)
        f.close()

        # set recursion limit is necessary
        if sys.getrecursionlimit() < self.restart.ncluster:
            sys.setrecursionlimit(self.restart.ncluster)

        self.mode_combinations = [mc for mc in self.restart.mode_combinations]
        if len(self.coord_labels) == 0:
            self.coord_labels = [lab for lab in self.restart.coord_labels]
        else:
            # check identical coordinaes
            seq = substractseq(self.coord_labels,self.restart.coord_labels)
            if seq:
                message = "Coordinate labels not found in restart: "
                message += ", ".join(seq)
                raise JobError(message)
            seq = substractseq(self.restart.coord_labels,self.coord_labels)
            if seq:
                message = "Coordinate in restart but not in input: "
                message += ", ".join(seq)
                raise JobError(message)
            # sort coordinates of input as in restart
            ranges_start = numpy.zeros(len(self.restart.coord_labels),dtype=float)
            ranges_end = numpy.zeros(len(self.restart.coord_labels),dtype=float)
            for n,lab in enumerate(self.restart.coord_labels):
                i = self.coord_labels.index(lab)
                ranges_start[n] = self.ranges_start[i]
                ranges_end[n] = self.ranges_end[i]
            self.ranges_start = ranges_start
            self.ranges_end = ranges_end
            self.coord_labels = [lab for lab in self.restart.coord_labels]


            for n in range(len(self.coord_labels)):
                if self.ranges_start[n] < self.restart.ranges_start[n]:
                    message = "Coordinate " + str(self.coord_labels[n])
                    message += ": start range is less then in restart: "
                    message += "Resetting to " + str(self.restart.ranges_start[n])
                    message += "."
                    self.ranges_start[n] = self.restart.ranges_start[n]
                    self.log.warn(message)
                if self.ranges_end[n] > self.restart.ranges_end[n]:
                    message = "Coordinate " + str(self.coord_labels[n])
                    message += ": end range is larger then in restart: "
                    message += "Resetting to " + str(self.restart.ranges_end[n])
                    message += "."
                    self.ranges_end[n] = self.restart.ranges_end[n]
                    self.log.warn(message)
                if self.ranges_start[n] > self.restart.ranges_start[n]:
                    message = "Coordinate " + str(self.coord_labels[n])
                    message += ": start range is different from restart. "
                    message += "Was: " + str(self.restart.ranges_start[n])
                    message += ". Is now: " + str(self.ranges_start[n])
                    self.log.debug(message)
                if self.ranges_end[n] < self.restart.ranges_end[n]:
                    message = "Coordinate " + str(self.coord_labels[n])
                    message += ": start range is different from restart. "
                    message += "Was " + str(self.restart.ranges_end[n])
                    message += ". Is now: " + str(self.ranges_end[n])
                    self.log.debug(message)



        if self.restart.nref > 1:
            if self.restart.switch_function is None:
                self.weights = numpy.array(self.restart.weights).copy()
        else:
            self.weights = numpy.array([1.e0,],dtype=float)
        self.nref = self.restart.nref

        if self.nsteps is None:
            self.nsteps = self.restart.nsteps
        else:
            if self.restart.nsteps < self.nsteps:
                message = "Restart only contains " + str(self.restart.nsteps)
                message += " samples. Resetting in input to "
                message += str(self.restart.nsteps) + "."
                self.log.warn(message)
                self.nsteps = self.restart.nsteps

                if self.nskip > self.nsteps:
                    message = "Cannot skip " + str(self.nskip)
                    message += " samples after resetting steps."
                    raise JobError(message)



        # potential expansion - can be different from restart
        # but needs to be a subset
        # expand the wildcards
        if self.input.expansionsection is not None:
            clusterlist = [cl for cl in self.input.expansionsection.clusterlist]
            expclusterlist = []
            for cluster in clusterlist:
                cluster1 = []
                for idx in cluster:
                    try:
                        n = int(idx)
                        cluster1.append(n-1)
                    except:
                        cluster1.append(idx)
                clist = expandWildCards(cluster1, self.mode_combinations)
                expclusterlist += clist
                self.clusterlist = expclusterlist
            # clusters are generated through DOF, not modes.
            # and the DOF list serves as an ID and to find
            # dependencies
            self.dofclusterlist = []
            for cl in self.clusterlist:
                dofs = idx2doflist(cl, self.coord_labels, self.mode_combinations )
                self.dofclusterlist.append(tuple(dofs))

            # the clusterlist has to be unique
            doubles = getNonUnique(self.dofclusterlist)
            if len(doubles) > 0:
                message = "Multiply defined clusters have been found:\n"
                for cl in doubles:
                    mtpl = incrAllInt(doftpl2modetpl(cl,self.mode_combinations))
                    message += "Cluster "
                    message += str(tuple(mtpl)) + "\n"
                raise JobError(message)

            #check if clusters are in restart
            missing = []
            for cluster in self.dofclusterlist:
                if not cluster in self.restart.dofclusterlist:
                    missing.append(cluster)
            if len(missing) > 0:
                message = "Clusters not in restart:\n"
                for cl in missing:
                    mtpl = incrAllInt(doftpl2modetpl(cl,self.mode_combinations))
                    message += "Cluster "
                    message += str(tuple(mtpl)) + "\n"
                raise JobError(message)



        else:
            self.dofclusterlist = self.restart.dofclusterlist
            self.clusterlist = self.restart.clusterlist



        # check wich clusters are used
        self.ncluster = len(self.restart.clusters[0])
        self.usemap = numpy.zeros(self.ncluster,dtype=bool)
        self.order = 0
        for n,cluster in enumerate(self.restart.clusters[0]):
            if tuple(cluster.str_dof) in self.dofclusterlist:
                self.usemap[n] = True
                self.order = max(self.order,cluster.order)



        self.ordermap = numpy.zeros((self.ncluster,self.order+1),dtype=float, order="F")

        for n,cluster in enumerate(self.restart.clusters[0]):

            if tuple(cluster.str_dof) in self.dofclusterlist:
                i = cluster.order
                self.ordermap[n,i] = 1.e0


        if self.nskip > self.restart.nsteps:
            message = "Cannot skip " + str(self.nskip) + " steps: "
            message += "restart contains only "
            message += str(self.restart.steps) + " steps."
            raise JobError(message)
        if self.restart.nsteps  < self.nsteps:
            message = "Cannot process " + str(self.nsteps)
            message += " steps: restart contains only "
            message += str(self.restart.steps) + " steps."
            message += "Setting number of steps to "
            message += str(self.nsteps) + "."
            self.log.warn(message)
            self.nsteps = self.restart.nsteps

        if hasattr(self.restart,"lusetrajweights"):
            self.lusetrajweights = bool(self.restart.lusetrajweights)
            self.mean_trajweight = float(self.restart.mean_trajweight)
            self.rms_trajweight = float(self.restart.rms_trajweight)
        else:
            self.lusetrajweights = False
            self.mean_trajweight = 1.e0
            self.rms_trajweight = 1.e0

        # set initial values of members that do not come directly
        # from inp or restart
        self.eof = False

        self.count_accepted = 0
        self.count_read = 0

        self.resetNumericalData()


    def resetNumericalData(self):
        """
        resetNumericalData() --> None

        Resets all statistical data to zero end empties log-lists.
        System data is kept, files are still open and not repositioned.
        """

        self.curr_energy_exact = 0.e0
        self.curr_energy_approx = 0.e0
        self.curr_energy_delta = 0.e0


        self.sum_error = 0.e0
        self.sum2_error = 0.e0
        #self.mean_error = 0.e0
        #self.rms_error = 0.e0

        self.curr_order = numpy.zeros(self.order+1, float)
        self.sum_order = numpy.zeros(self.order+1, float)
        self.sum2_order = numpy.zeros(self.order+1, float)
        self.mean_order = numpy.zeros(self.order+1, float)
        self.rms_order = numpy.zeros(self.order+1, float)

        self.curr_error_order = numpy.zeros(self.order+1, float)
        self.sum_error_order = numpy.zeros(self.order+1, float)
        self.sum2_error_order = numpy.zeros(self.order+1, float)
        self.mean_error_order = numpy.zeros(self.order+1, float)
        self.rms_error_order = numpy.zeros(self.order+1, float)

        self.curr_ref = numpy.zeros(self.nref, float)
        self.sum_ref = numpy.zeros(self.nref, float)
        self.sum2_ref = numpy.zeros(self.nref, float)
        self.mean_ref = numpy.zeros(self.nref, float)
        self.rms_ref = numpy.zeros(self.nref, float)

        self.curr_error_ref = numpy.zeros(self.nref, float)
        self.sum_error_ref = numpy.zeros(self.nref, float)
        self.sum2_error_ref = numpy.zeros(self.nref, float)
        self.mean_error_ref = numpy.zeros(self.nref, float)
        self.rms_error_ref = numpy.zeros(self.nref, float)


        # nref index: 0 for all, other for refs
        self.curr_clusters = numpy.zeros((self.ncluster, self.nref), float, order="F")
        self.mean_clusters = numpy.zeros((self.ncluster, self.nref+1), float, order="F")
        self.rms_clusters = numpy.zeros((self.ncluster, self.nref+1), float, order="F")
        self.sum_clusters = numpy.zeros((self.ncluster, self.nref+1), float, order="F")
        self.sum2_clusters = numpy.zeros((self.ncluster, self.nref+1), float, order="F")
        self.ncount_low = numpy.zeros((self.ncluster, self.nref+1), int, order="F")
        self.ncount_high = numpy.zeros((self.ncluster, self.nref+1), int, order="F")

        self.samples_max_delta = []
        self.samples_low_delta = []
        self.samples_max_exact = []
        self.samples_low_exact = []
        self.samples_max_approx = []
        self.samples_low_approx = []


        self.nsince_log_max_delta = self.nskiplog
        self.nsince_log_low_delta = self.nskiplog
        self.nsince_log_max_exact = self.nskiplog
        self.nsince_log_low_exact = self.nskiplog
        self.nsince_log_max_approx = self.nskiplog
        self.nsince_log_low_approx = self.nskiplog


    def __repr__(self):
        string = "Statistical analysis data evaluation job.\n"
        string += "Working Directory:" + str(self.outdir) + "\n"
        return string


    def _parseCommandLine(self, opts=[]):
        "Parse command line options"
        self.ldebug = False
        opts, args = getopt.getopt(sys.argv, 'd:')
        for o, a in opts:
            if o == "-d" and a == "eb":
                self.ldebug = True
            else:
                message = "Unknown option: " + str(o) + str(a)
                raise JobError(message)






    def mkHistograms(self):
        "Read coordinates file and write 1D and 2D densities."

        if self.lden1d:
            self.log.info("Calculatung 1D densities." +
                          " This may take a while.")
            for n,coord in enumerate(self.coord_labels):

                self.log.info(" 1D density: "+self.coord_labels[n])

                nbins = self.nbins
                frange = (self.restart.ranges_start[n], self.restart.ranges_end[n])
                lnormed = True
                lnew = True

                fname = _FILENAME_DEN1D + "_" + self.coord_labels[n]
                fname += ".dat"
                fname = os.path.join(self.outdir,fname)

                self.files.coordinates.seek(0)
                f = self.files.coordinates
                coordinates = numpy.loadtxt(f, usecols=(n,),
                                            skiprows=self.nskip)[:self.nsteps]

                hist, edges = numpy.histogram(coordinates, bins=nbins,range=frange,
                                              normed=lnormed, new=lnew)
                f = open(fname,mode="w")
                for i in range(self.nbins):
                    x = (edges[i] + edges[i+1])/2.e0
                    line = "%f16.8 " % x + "%f16.8 \n" % hist[i]
                    f.write(line)
                f.close()

        if self.lden2d:
            self.log.info("Calculatung 2D densities." +
                          " This may take a while.")
            for n1,coord in enumerate(self.coord_labels):
                for n2 in range(n1+1,len(self.coord_labels)):

                    message = " 2D density: "+self.coord_labels[n1]
                    message += ", " + self.coord_labels[n2]
                    self.log.info(message)


                    nbins = self.nbins
                    frange = [(self.restart.ranges_start[n1], self.restart.ranges_end[n1])]
                    frange += [(self.restart.ranges_start[n2], self.restart.ranges_end[n2])]
                    lnormed = True

                    fname = _FILENAME_DEN2D + "_" + self.coord_labels[n1]
                    fname += "_" + self.coord_labels[n2]
                    fname += ".dat"
                    fname = os.path.join(self.outdir,fname)

                    self.files.coordinates.seek(0)
                    f = self.files.coordinates

                    coordinates = numpy.loadtxt(f, usecols=(n1,n2,),
                                                skiprows=self.nskip)[:self.nsteps,:]

                    hist, edges1, edges2 =  numpy.histogram2d(coordinates[:,0],
                                                              coordinates[:,1],
                                                              bins=nbins,
                                                              range=frange,
                                                              normed=lnormed)

                    f = open(fname,mode="w")
                    for i in range(self.nbins):
                        x = (edges1[i] + edges1[i+1])/2.e0
                        for j in range(self.nbins):
                            y = (edges2[j] + edges2[j+1])/2.e0
                            line = "%f16.8 " % x
                            line += "%f16.8 " % y
                            line += "%f16.8 \n" % hist[i,j]
                            f.write(line)
                        f.write("\n")
                    f.close()





    def wrOutput(self):
        "Write output files."

        def mkmsg(arr, unit):
            message = ""
            for elem in arr:
                message += "%16.8f" % (elem*unit)
            message += "\n"
            return message

        if self.lmean:
            message = mkmsg(self.mean_error_order, self.outunit_float)
            self.files.mean_error_ord.write(message)
            message = mkmsg(self.mean_order, self.outunit_float)
            self.files.mean_ord.write(message)
            if self.nref > 1:
                message = mkmsg(self.mean_error_ref, self.outunit_float)
                self.files.mean_error_ref.write(message)
                message = mkmsg(self.mean_ref, self.outunit_float)
                self.files.mean_ref.write(message)


        if self.lrms:
            message = mkmsg(self.rms_error_order, self.outunit_float)
            self.files.rms_error_ord.write(message)
            message = mkmsg(self.rms_order, self.outunit_float)
            self.files.rms_ord.write(message)
            if self.nref > 1:
                message = mkmsg(self.rms_error_ref, self.outunit_float)
                self.files.rms_error_ref.write(message)
                message = mkmsg(self.rms_ref, self.outunit_float)
                self.files.rms_ref.write(message)


        message = mkmsg(self.curr_error_order, self.outunit_float)
        self.files.curr_error_ord.write(message)
        message = mkmsg(self.curr_order, self.outunit_float)
        self.files.curr_ord.write(message)
        if self.nref > 1:
            message = mkmsg(self.curr_ref, self.outunit_float)
            self.files.curr_ref.write(message)
            message = mkmsg(self.curr_error_ref, self.outunit_float)
            self.files.curr_error_ref.write(message)



        message = mkmsg([self.curr_energy_delta,
                         self.curr_energy_exact,
                         self.curr_energy_approx],
                        self.outunit_float)
        self.files.delta.write(message)

        self.files.flushWriteables()


    def openStatFiles(self):

        if self.restart.switch_function is not None\
               and self.restart.nref > 1:
            fname = os.path.join(self.outdir, _FILENAME_WEIGHTS)
            self.files.weights = openFile(fname,"r")


        if self.lusetrajweights:
            fname = os.path.join(self.outdir, _FILENAME_TRAJWEIGHT)
            self.files.trajweight = openFile(fname,"r")


        for n in range(self.nref):
            fname = _FILENAME_CLUSTERS+ "_ref_" + str(n+1)
            fpath = os.path.join(self.outdir,fname)
            f = openFile(fpath,'r')
            self.files.appendFile(fname,f)


        if self.restart.trajectorypath is None:
            fname = os.path.join(self.outdir, _FILENAME_COORDINATES)
        else:
            fname = self.restart.trajectorypath
        try:
            self.files.coordinates = openFile(fname,"r")
        except:
            message = "Unable to open coordinates file: "
            message += fname + ". "
            if self.restart.trajectorypath is not None:
                message += "This fiel has been used with 'clusterstat'"
                message += " and must not be moved or deleted."
            else:
                message += "The name directory seems to be corrupted."
            raise JobError(message)

        buf = self.files.coordinates.readline()
        lhash = buf.startswith("#")
        if lhash:
            pos = self.files.coordinates.tell()
            while lhash:
                pos = self.files.coordinates.tell()
                buf = self.files.coordinates.readline()
                lhash = buf.startswith("#")
            self.files.coordinates.seek(pos)
        else:
            self.files.coordinates.seek(0)




        fname = os.path.join(self.outdir, _FILENAME_ENERGIES)
        self.files.energies = openFile(fname,"r")



        if self.lmean:
            fname = os.path.join(self.outdir, _FILENAME_MEAN_ORD)
            self.files.mean_ord = openFile(fname,"w",self.loverwrite)
            fname = os.path.join(self.outdir, _FILENAME_MEAN_ERR_ORD)
            self.files.mean_error_ord = openFile(fname,"w",self.loverwrite)

            if self.nref > 1:
                fname = os.path.join(self.outdir, _FILENAME_MEAN_REF)
                self.files.mean_ref = openFile(fname,"w",self.loverwrite)
                fname = os.path.join(self.outdir, _FILENAME_MEAN_ERR_REF)
                self.files.mean_error_ref = openFile(fname,"w",self.loverwrite)

        if self.lrms:
            fname = os.path.join(self.outdir, _FILENAME_RMS_ORD)
            self.files.rms_ord = openFile(fname,"w",self.loverwrite)
            fname = os.path.join(self.outdir, _FILENAME_RMS_ERR_ORD)
            self.files.rms_error_ord = openFile(fname,"w",self.loverwrite)
            if self.nref > 1:
                fname = os.path.join(self.outdir, _FILENAME_RMS_REF)
                self.files.rms_ref = openFile(fname,"w",self.loverwrite)
                fname = os.path.join(self.outdir, _FILENAME_RMS_ERR_REF)
                self.files.rms_error_ref = openFile(fname,"w",self.loverwrite)


        fname = os.path.join(self.outdir, _FILENAME_CURR_ORD)
        self.files.curr_ord = openFile(fname,"w",self.loverwrite)
        fname = os.path.join(self.outdir, _FILENAME_CURR_ERR_ORD)
        self.files.curr_error_ord = openFile(fname,"w",self.loverwrite)
        if self.nref > 1:
            fname = os.path.join(self.outdir, _FILENAME_CURR_REF)
            self.files.curr_ref = openFile(fname,"w",self.loverwrite)
            fname = os.path.join(self.outdir, _FILENAME_CURR_ERR_REF)
            self.files.curr_error_ref = openFile(fname,"w",self.loverwrite)

        fname = os.path.join(self.outdir, _FILENAME_DELTA)
        self.files.delta = openFile(fname,"w",self.loverwrite)


    def readDataSets(self):
        "Read one data set"

        if self.eof:
            self.log.error("Call to readDataSets after EOF was detected.")
            return

        # read values of switching function
        if self.restart.switch_function is not None\
               and self.restart.nref > 1:
            line = self.files.weights.readline()
            if line == "":
                self.eof = True
                self.log.warning("Reached end of weights file.")
            else:
                self.weights = numpy.array(map(float,line.split()),dtype=float)


        # read clusters
        for n in range(self.nref):
            fname = _FILENAME_CLUSTERS+ "_ref_" + str(n+1)
            f = self.files[fname]
            line = f.readline()
            if line == "":
                self.eof = True
                self.log.warning("Reached end of file " + fname)
            else:
                self.curr_clusters[:,n] = numpy.array(map(float,line.split()),dtype=float)
                self.curr_clusters[:,n] /= self.restart.outunit_float

        # read coordinates
        line = self.files.coordinates.readline()
        if line == "":
            self.eof = True
            self.log.warning("Reached end of coordinates file.")
        else:
            pos = map(float,line.split())
            self.curr_position = numpy.array(pos,dtype=float)

        # read exact energy
        line = self.files.energies.readline()
        if line == "":
            self.eof = True
            self.log.warning("Reached end of energies file.")
        else:
            self.curr_energy_exact = float(line.split()[0])
            self.curr_energy_exact /= self.restart.outunit_float

        if not self.eof:
            self.count_read += 1

        # read weights from trajectory
        if self.lusetrajweights:
            line = self.files.trajweight.readline()
            if line == "":
                self.eof = True
                self.log.warning("Reached end of trajweigt file.")
            else:
                self.curr_trajweight = float(line.split()[0])
        else:
            self.curr_trajweight = 1.e0


    def calcAuto(self):
        "Calculate the auto-correlation function from the coordinates file."

        self.log.info("Calculatung auto correlation " +
                      "function of coordinates. This may take a while.")

        if (self.restart.nsteps - self.nskip) * self.restart.ndof > 10E+8:
            message = "More then 10E+8 data points are to be read. Trying anyways..."
            self.log.warn(message)


        self.files.coordinates.seek(0)
        f = self.files.coordinates
        self.log.debug("Loading coordinates...")

        #this is sub-optimal if self.nsteps << self.restart.count_accepted
        try:
            coordinates = numpy.loadtxt(f,skiprows=self.nskip,dtype=float)[0:self.nsteps-self.nskip,:]

        except MemoryError as M:
            self.log.exception(M)
            message = "Not enough memory to load coordinates file. "
            message += "Calculation auf auto-correlation function is skipped."
            self.log.error(message)
            return
        try:
            auto = numpy.zeros((coordinates.shape[0],coordinates.shape[1]+1),dtype=float,order="F")
        except MemoryError as M:
            self.log.exception(M)
            message = "Not enough memory to calculate the auto-correlation "
            message += "function. Skipping this task."
            self.log.error(message)
            return

        for i in range(coordinates.shape[1]):
            coordinates[:,i] -= coordinates[:,i].mean()
            auto[:,i+1] = numpy.correlate(coordinates[:,i],coordinates[:,i],
                                    "full")[coordinates.shape[0]-1:]
        for i in range(coordinates.shape[0]):
            auto[i,0] = auto[i,1:].sum()


        fname = os.path.join(self.outdir, _FILENAME_AUTO)
        numpy.savetxt(fname, auto)


    def wrStatistics(self):
        "Write statistical data about all clusters"


        def mkstats(i,stringlen):

            if i == 0:
                fname = os.path.join(self.outdir,_FILENAME_STATISTICS)
            else:
                fname = os.path.join(self.outdir,_FILENAME_STATISTICS + "_ref_" + str(i))
            f = openFile(fname,"w",self.loverwrite)

            message = ""

            if self.nref > 1:
                if i == 0:
                    message += "#Statistics of the sum of the weighted clusters.\n"
                else:
                    message += "#Statistics of the clusters obtained around ref-point " + str(i) + ".\n"
                    message += "#These values do not include weights.\n"

            message += "# Cluster" + " "*(stringlen+5)
            message += "mean                "
            message += "rms     "
            if self.countlow is not None and self.counthigh is not None:
                message += "< %6.e" %  (self.countlow*self.outunit_float)
                message += " " + self.outunit
                message += ", > %6.e" %  (self.counthigh*self.outunit_float)
                message += " " + self.outunit
            message += "\n"
            f.write(message)

            lunused = False
            for n, cluster in enumerate(self.restart.clusters[0]):
                message = cluster.idxStr().rjust(stringlen)
                message += "  %16.8f " % (self.mean_clusters[n,i]*self.outunit_float)
                message += "  %16.8f " % (self.rms_clusters[n,i]*self.outunit_float)
                if self.countlow is not None and self.counthigh is not None:
                    message += "  %10d " %  self.ncount_low[n,i]
                    message += "  %10d " %  self.ncount_high[n,i]
                if tuple(cluster.str_dof) not in self.dofclusterlist:
                    message +=  " <-- Not used"
                    lunused = True
                message += "\n"
                f.write(message)

            if i == 0:
                if lunused:
                    message = "\n\nTOTAL ERROR (w/o the clusters marked as 'not used' above)\n"
                else:
                    message = "\n\nTOTAL ERROR\n"
                message += " Mean (exact-approx) %16.8f  " % (self.mean_error_order[-1]*self.outunit_float,)
                message += self.outunit + "\n"
                message += " RMS  (exact-approx) %16.8f  " % (self.rms_error_order[-1]*self.outunit_float,)
                message += self.outunit + "\n"
                f.write(message)

            f.close()


        stringlen = 15
        for cluster in self.restart.clusters[0]:
            stringlen = max(stringlen, len(cluster.idxStr()))

        mkstats(0,stringlen)

        if self.nref > 1:
            for i in range(1,self.nref+1):
                mkstats(i,stringlen)




    def logSamples(self):
        "Logging of samples"

        maxcoordlen = max([len(s) for s in self.restart.coord_labels])
        maxclusterlen = max([len(cl.idxStr()) for cl in self.restart.clusters[0]]  )


        def mkmsg(tpl):
            "Local string formatting function"
            # tpl = [V, Vappr, deltaV, count, pos, clusters, weights ]
            message = "Sample No. " + str(tpl[7]) + "\n"
            message += "True potential V: %16.8f" % (tpl[0]*self.outunit_float)
            message += " " + self.outunit + " \n"
            message += "Approx(V):        %16.8f" % (tpl[1]*self.outunit_float)
            message += " " + self.outunit + " \n"
            message += "V - Approx(V):    %16.8f" % (tpl[2]*self.outunit_float)
            message += " " + self.outunit + " \n\n"

            w = tpl[8][0]
            m = tpl[8][1]
            r = tpl[8][2]
            message += "Weight of this sample\n"
            message += "Absolute weight    :  %16.4f \n" % w
            message += "Mean of all weights:  %16.4f \n" % m
            message += "RMS of all weights :  %16.4f \n" % r
            message += "Relative weight:      %16.4f (Used when calculating Mean and RMS)" % (w/m)
            message += "\n\n"

            weights = tpl[5]

            if self.nref > 1:
                message += "No. references: %d " % self.nref
                message += "Weights:        " + ", ".join([" %8.4f" % w for w in weights])

            if self.countlow is not None and self.counthigh is not None:

                message += "\nRelevant Clusters (>= %6.3e" % (self.counthigh*self.outunit_float)
                message += " or <= %6.3e " % (self.countlow*self.outunit_float)
                message +=  self.outunit + "):\n"
                if self.nref > 1:
                    message += "# cluster, sum over references, (reference(i), i=1, nref), [comment]\n"
                else:
                    message += "# cluster, contribution, [comment]\n"

                clustervalues = tpl[6]
                for n, cluster in enumerate(self.restart.clusters[0]):

                    lrelevant = False
                    if self.nref > 1:
                        val = clustervalues[n,:].sum()
                        if val >= self.counthigh or val <= self.countlow:
                            lrelevant = True

                        for i in range(self.nref):
                            if clustervalues[n,i] >= self.counthigh \
                                   or clustervalues[n,i] <= self.countlow:
                                lrelevant = True


                        if lrelevant:
                            message += cluster.idxStr().rjust(maxclusterlen+4)
                            message += "%16.8f" % (val*self.outunit_float)
                            message += " | "
                            for i in range(self.nref):
                                message += "   %16.8f" % (clustervalues[n,i]*self.outunit_float)

                            if tuple(cluster.str_dof) not in self.dofclusterlist:
                                message +=  " <-- Not used"
                            message += "\n"
                    else:

                        val = clustervalues[n,0]
                        if val >= self.counthigh or val <= self.countlow:
                            lrelevant = True
                        if lrelevant:
                            message += cluster.idxStr().rjust(maxclusterlen+4)
                            message += "%16.8f" % (val*self.outunit_float)
                            if tuple(cluster.str_dof) not in self.dofclusterlist:
                                message +=  " <-- Not used"
                            message += "\n"

            message += "\nCoordinates: \n"

            pos = tpl[4]
            for n in range(len(self.coord_labels)):
                string = self.coord_labels[n].rjust(maxcoordlen+4)
                message += string + "    =%16.8f" % pos[n]
                message += "\n"
            message += "-"*50 + "\n\n"
            return message


        if self.nlog is not None and self.nlog > 0:
            # tpl = [V, Vappr, deltaV, count, pos, clusters, weights ]

            fname = os.path.join(self.outdir,_FILENAME_LARGEST_EXACT)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_max_exact.sort(key = lambda tpl: abs(tpl[0]), reverse=True)
            for tpl in self.samples_max_exact:
                message = mkmsg(tpl)
                f.write(message)
            f.close()

            fname = os.path.join(self.outdir,_FILENAME_LOWEST_EXACT)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_low_exact.sort(key = lambda tpl: abs(tpl[0]))
            for tpl in self.samples_low_exact:
                message = mkmsg(tpl)
                f.write(message)
            f.close()

            fname = os.path.join(self.outdir,_FILENAME_LARGEST_APPROX)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_max_approx.sort(key = lambda tpl: abs(tpl[1]), reverse=True)
            for tpl in self.samples_max_approx:
                message = mkmsg(tpl)
                f.write(message)
            f.close()

            fname = os.path.join(self.outdir,_FILENAME_LOWEST_APPROX)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_low_approx.sort(key = lambda tpl: abs(tpl[1]))
            for tpl in self.samples_low_approx:
                message = mkmsg(tpl)
                f.write(message)
            f.close()

            fname = os.path.join(self.outdir,_FILENAME_LARGEST_DELTA)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_max_delta.sort(key = lambda tpl: abs(tpl[2]), reverse=True)
            for tpl in self.samples_max_delta:
                message = mkmsg(tpl)
                f.write(message)
            f.close()

            fname = os.path.join(self.outdir,_FILENAME_LOWEST_DELTA)
            f = openFile(fname,"w",self.loverwrite)
            self.samples_low_delta.sort(key = lambda tpl: abs(tpl[2]))
            for tpl in self.samples_low_delta:
                message = mkmsg(tpl)
                f.write(message)
            f.close()


    def updateStatistics(self):
        """
        The current sample is checked if it is to be logged.
        """

        self.curr_order = numpy.zeros(self.order+1,dtype=float)
        self.curr_ref = numpy.zeros(self.nref,dtype=float)


        weighted_curr_clusters =  self.curr_clusters.copy()
        for nc in range(self.ncluster):
            weighted_curr_clusters[nc,:] *= self.weights


        for i in range(self.order+1):
            for n in range(self.nref):
                val =  numpy.dot(self.ordermap[:,i],weighted_curr_clusters[:,n])
                self.curr_order[i] += val
                self.curr_ref[n] += val

        wscale =1.e0
        if self.lusetrajweights:
            wscale = self.curr_trajweight/self.mean_trajweight


        self.sum_order += self.curr_order*wscale
        self.sum2_order += self.curr_order**2*wscale

        self.mean_order = self.sum_order/self.count_accepted
        self.rms_order = numpy.sqrt(self.sum2_order/self.count_accepted)


        curr_energy_approx_ref =  self.curr_ref.sum()
        curr_energy_approx_order =  self.curr_order.sum()
        if abs(curr_energy_approx_ref -  curr_energy_approx_order) > 1.e-7:
            message = "Internal error while calculating statistics."
            raise JobError(message)

        self.curr_energy_approx = curr_energy_approx_order
        self.curr_energy_delta = self.curr_energy_exact - self.curr_energy_approx

        self.sum_error += self.curr_energy_delta
        self.sum2_error += self.curr_energy_delta**2
        self.mean_error = self.sum_error/self.count_accepted
        self.rms_error = numpy.sqrt(self.sum2_error/self.count_accepted)

        self.curr_error_order = self.curr_energy_exact - self.curr_order.cumsum()
        self.sum_error_order += self.curr_error_order*wscale
        self.sum2_error_order += self.curr_error_order**2*wscale
        self.mean_error_order =  self.sum_error_order/self.count_accepted
        self.rms_error_order =  numpy.sqrt(self.sum2_error_order/self.count_accepted)

        self.sum_ref += self.curr_ref*wscale
        self.sum2_ref += self.curr_ref**2*wscale
        self.mean_ref = self.sum_ref/self.count_accepted
        self.rms_ref = numpy.sqrt(self.sum2_ref/self.count_accepted)

        self.curr_error_ref = self.curr_energy_exact*self.weights - self.curr_ref
        self.sum_error_ref += self.curr_error_ref *wscale
        self.sum2_error_ref += self.curr_error_ref**2*wscale
        self.mean_error_ref = self.sum_error_ref/self.count_accepted
        self.rms_error_ref = numpy.sqrt(self.sum2_error_ref/self.count_accepted)


        self.sum_clusters[:,1:self.nref+1] += self.curr_clusters*wscale
        for i in range(self.ncluster):
            self.sum_clusters[i,0] += weighted_curr_clusters[i,:].sum()*wscale

        self.sum2_clusters[:,1:self.nref+1] += self.curr_clusters**2*wscale
        for i in range(self.ncluster):
            self.sum2_clusters[i,0] += (weighted_curr_clusters[i,:].sum())**2*wscale

        self.mean_clusters = self.sum_clusters/self.count_accepted
        self.rms_clusters = numpy.sqrt(self.sum2_clusters/self.count_accepted)


        tpl = [self.curr_energy_exact, self.curr_energy_approx, self.curr_energy_delta]
        tpl += [self.count_read, self.curr_position.copy(), self.weights.copy()]
        tpl.append(weighted_curr_clusters.copy())
        tpl.append(self.count_read)
        tpl.append((self.curr_trajweight, self.mean_trajweight, self.rms_trajweight))


        def chkSample(tpl, nkey,
                      container,
                      count,mincount,
                      maxlog, func, lreverse = False):



            val = func(tpl[nkey])
            if count >= mincount:
                if len(container) < maxlog:
                    container.append(tpl)
                    container.sort(key = lambda x: func(x[nkey]), reverse=lreverse)
                    count = 0
                else:
                    if lreverse is False:
                        idx = 0
                    else:
                        idx = len(container)-1

                    if val > func(container[idx][nkey]):
                        container.pop(0)
                        container.append(tpl)
                        container.sort(key = lambda x: func(x[nkey]), reverse=lreverse)
                        count = 0
            else:
                count += 1

            return count

        if self.nlog > 0:


            self.nsince_log_max_exact = chkSample(tpl, 0,
                                                  self.samples_max_exact,
                                                  self.nsince_log_max_exact,
                                                  self.nskiplog, self.nlog, abs, False)
            self.nsince_log_low_exact = chkSample(tpl, 0,
                                                  self.samples_low_exact,
                                                  self.nsince_log_low_exact,
                                                  self.nskiplog,self.nlog,abs,True)

            self.nsince_log_max_approx = chkSample(tpl, 1,
                                                   self.samples_max_approx,
                                                   self.nsince_log_max_approx,
                                                   self.nskiplog, self.nlog, abs, False)
            self.nsince_log_low_approx = chkSample(tpl, 1,
                                                   self.samples_low_approx,
                                                   self.nsince_log_low_approx,
                                                   self.nskiplog, self.nlog,abs,True)

            self.nsince_log_max_delta = chkSample(tpl, 2,
                                                  self.samples_max_delta,
                                                  self.nsince_log_max_delta,
                                                  self.nskiplog, self.nlog,abs, False)
            self.nsince_log_low_delta = chkSample(tpl, 2,
                                                  self.samples_low_delta,
                                                  self.nsince_log_low_delta,
                                                  self.nskiplog, self.nlog,abs, True)



        # count how many times a cluster contributed more then
        # 'counthigh' or less then 'countlow'
        for n in range(self.nref):
            for m,val in enumerate(self.curr_clusters[:,n]):
                if val >= self.counthigh:
                    self.ncount_high[m,n+1] += 1
                elif val <= self.countlow:
                    self.ncount_low[m,n+1] += 1
        for n in range(self.ncluster):
            val = self.curr_clusters[n,:].sum()
            if val >= self.counthigh:
                self.ncount_high[n,0] += 1
            if val <= self.countlow:
                self.ncount_low[n,0] += 1


    def useThisSample(self):
        """
        useThisSample() --> bool

        Check if Sample is in specified
        coordinate range
        """

        # ranges are specified
        if self.ranges_start is not None \
           and self.ranges_end is not None \
           and _isInRange is None:

            luse = True
            for n,start in enumerate(self.ranges_start):
                end = self.ranges_end[n]
                if self.curr_position[n] <= start or \
                   self.curr_position[n] >= end:
                    return False

            return luse

        # user function is specified
        elif _isInRange is not None:
            return bool(_isInRange(self.curr_position.copy()))

        # no restrictions
        else:
            return True

    def execute(self):
        """
        Execute the job. Perform analysis on all samples.
        This sould be the routine to call after initialization.
        """
        self.log.info("Starting analysis.")

        if self.lcluster_statistics:

            self.openStatFiles()
            self.eof = False

            # position files
            for i in range(self.nskip):
                self.readDataSets()
            self.resetNumericalData()
            self.count_accepted = 0

            for i in range(self.nsteps-self.nskip):
                self.readDataSets()
                if self.useThisSample() and not self.eof:
                    self.count_accepted += 1
                    self.updateStatistics()
                    self.wrOutput()

                    # write statistics every once in a while
                    if self.count_accepted % _NUPDATE == 0 or\
                           self.eof:
                        message = "Samples read %12d, " % self.count_read
                        message += "used %12d, "  % self.count_accepted
                        message += "Mean Err: %12.3f " % (self.mean_error_order[-1]*self.outunit_float)
                        message += str(self.outunit) + ",  "
                        message += "RMS Err: %12.3f " % (self.rms_error_order[-1]*self.outunit_float)
                        message += str(self.outunit)
                        self.log.info(message)

                if self.eof is True:
                    break

            self.wrStatistics()
            if self.nlog is not None and self.nlog > 0:
                self.logSamples()

        if self.lauto:
            self.calcAuto()
        if self.lden1d or self.lden2d:
            self.mkHistograms()

        message = "Job has Comleted."
        self.log.info(message)
        message = "Samples read: %d" % (self.count_read - self.nskip)
        self.log.info(message)
        message = "Samples used: %d" % (self.count_accepted)
        self.log.info(message)


    def cleanUp(self):
        """
        Close files, write statistics, etc.
        Should be called before exit.
        """
        self.log.info('Cleaning up...')
        self.files.closeAll()




