#!/usr/bin/env python3

from minput import MSection, InputError
from parsingtools import *

class ClusteranaRunSection(MSection):

    def __init__(self, lines):
        MSection.__init__(self,lines)
        if lines[0].upper() !=  "RUN-SECTION":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. First line reads: " \
                     + lines[0]
            raise Exception(message)
        elif lines[-1].lower() != "end-run-section":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. Last line reads: " \
                     + lines[-1]
            raise Exception(message)

        self.openstr = lines.pop(0)
        self.closestr = lines.pop(-1)


        key = "name"
        strval, lines = find_param(key, lines)
        if strval:
            self.name = strval
        else:
             self.name = None

        # optional keywords
        key = "steps"
        strval, lines = find_param(key, lines)
        if strval:
            vlist = strval.split(",")
            if len(vlist) == 1:
                self.nsteps = int(strval)
                self.nskip = None
                if self.nsteps < 0:
                    message = "RUN-SECTION: 'steps = I1(,I2)': "
                    message += "I1 and I2 must be non-negative integers."
                    raise ValueError(message)
            elif len(vlist) == 2:
                self.nskip = int(vlist[0])
                self.nsteps = int(vlist[1])
                if self.nsteps < 0 or self.nskip < 0:
                    message = "RUN-SECTION: 'steps = I1(,I2)': "
                    message += "I1 and I2 must be non-negative integers."
                    raise ValueError(message)
                if self.nsteps <= self.nskip:
                    message = "RUN-SECTION: 'steps = I1(,I2)': "
                    message += "I1 must be smaller then I2."
                    raise ValueError(message)
            else:
                message = "RUN-SECTION: 'steps = I1(,I2)': "
                message += "too many keywords: (" + strval + ")."
                raise InputError(message)
        else:
            self.nsteps = None
            self.nskip = 0

        key = "logminmax"
        strval, lines = find_param(key, lines)
        if strval:
            vlist = strval.split(",")
            if len(vlist) == 1:
                self.nlog = int(strval)
                self.nskiplog = None
            elif len(vlist) == 2:
                self.nlog = int(vlist[0])
                self.nskiplog = int(vlist[1])
            else:
                message = "RUN-SECTION: 'logminmax = I1(,I2)': "
                message += "wrong number of keywords: "
                message += "(" + strval + ")."
                raise InputError(message)
        else:
            self.nlog = None
            self.nskiplog = None

        key = "count-samples"
        strval, lines = find_param(key, lines)
        if strval:
            vlist = strval.split(",")
            if len(vlist) == 2:
                self.countlow = float(vlist[0])
                self.counthigh = float(vlist[1])
                self.countunit = None
            elif len(vlist) == 3:
                self.countlow = float(vlist[0])
                self.counthigh = float(vlist[1])
                self.countunit =  vlist[2]

            else:
                message = "RUN-SECTION: count-samples: wrong number of "
                message += "comma separated keywords: (" + strval + "). "
                message += "Expecting 'count-samples = R1,R2(,S)'."
                raise InputError(message)
        else:
            self.countlow = None
            self.counthigh = None
            self.countunit = None


        # check countlow and counthigh
        if self.counthigh is not None and self.countlow is not None:
            if self.countlow >= self.counthigh:
                message = "RUN-SECTION: 'count-samples = R1,R2(,S)': "
                message += "R1 must be smaller than R2."
                raise InputError(message)
        # check unit
        if self.countunit is not None:
            if getunit(self.countunit) is None:
                message = "RUN-SECTION: 'count-samples = R1,R2(,S)': "
                message += "S must be a unit ('eV','au',...)."
                raise InputError(message)




        key = "densities"
        strval, lines = find_param(key, lines)
        if strval:
            vlist = strval.split(",")
            if len(vlist) > 2:
                message = "RUN-SECTION: 'densities = S1(,S2)': "
                message += "Detected more the 2 arguments "
                messate += str(tuple(vlist)) + "."
                raise InputError(message)
            self.lden1d = False
            self.lden2d = False
            for v in vlist:
                if v.lower() == "1d":
                    self.lden1d = True
                elif  v.lower() == "2d":
                    self.lden2d = True
                else:
                    message = "RUN-SECTION: 'densities = S1(,S2)': "
                    message += "expecting '1D' or '2D', not "
                    message += "'" + v + "'."
                    raise InputError(message)
        else:
            self.lden1d = False
            self.lden2d = False


        key = "bins"
        strval, lines = find_param(key, lines)
        if strval:
            self.nbins = int(strval)
        else:
            self.nbins = None

        key = "outunit"
        strval, lines = find_param(key, lines)
        if strval:
            self.outunit = strval
            if getunit(self.outunit) is None:
                message = "RUN-SECTION: 'outunit = S': "
                message += "S must be a unit ('eV','au',...)."
        else:
            self.outunit = None

        # flags
        self.loverwrite, lines = find_flag("overwrite", lines)
        self.lauto, lines = find_flag("auto", lines)
        self.lmean, lines = find_flag("mean", lines)
        self.lrms, lines = find_flag("rms", lines)
        self.lcluster_statistics, lines \
               = find_flag("cluster-statistics", lines)

        # consistency checks
        if self.lden1d and not self.nbins:
            message = "RUN-SECTION: The Keywords 'densities' "
            message += "requires 'bins = N'."
            raise InputError(message)

        if self.lden2d and not self.nbins:
            message = "RUN-SECTION: The Keywords 'densities' "
            message += "requires 'bins = N'."
            raise InputError(message)


        if len(lines):
            message = "Input contains unrecognized keywords:\n"
            message += " \n".join(lines)
            raise InputError(message)


    def  __str__(self):
        string =  str(self.openstr)

        if self.name is not None:
            string += "\n name = " + str(self.name)

        if self.lcluster_statistics is True:
            string += "\n cluster-statistics"
        if self.countlow is not None \
               and self.counthigh is not None:
            string += "\n count-samples = "
            string += str( self.countlow) + ", "
            string += str( self.counthigh)
            if self.countunit is not None:
                string += ", " + str(self.countunit)

        # key-value-pairs
        if self.nsteps is not None:
            if self.nskip is None:
                string += "\n steps = " + str(self.nsteps)
            else:
                string += "\n steps = " + str(self.nskip)
                string += ", " + str(self.nsteps)

        if self.nlog  is not None:
            string += "\n logminmax = " + str(self.nlog)
            if self.nskiplog is not None:
                string += ", " + str(self.nskiplog)
        if self.countlow is not None and self.counthigh is not None:
            string += "\n count-samples = " + str(self.countlow)
            string += ", " + str(self.counthigh)
            if self.countunit is not None:
                string += ", " + str(self.countunit)


        if self.lden1d is True or self.lden2d is True:
            tlist = []
            if self.lden1d:
                tlist += ["1D"]
            if self.lden2d:
                tlist += ["2D"]
            string += "\n densities = " + ", ".join(tlist)
        if self.nbins is not None:
            string += "\n bins = "+str(self.nbins)

        if self.outunit is not None:
            string += "\n outunit = " + str(self.outunit)

        #flags
        if self.lmean is True:
            string += "\n mean"
        if self.lrms is True:
            string += "\n rms"
        if self.lauto is True:
            string += "\n auto"
        if self.loverwrite is True:
            string += "\n overwrite"

        string += "\n" + self.closestr + "\n"
        return string


