#!/usr/bin/env python3
"""
-------------------------------------------------------------------------------
Purpose:     Calculate the cluster expansion of a PES.

Usage:       clusterbuild [-mnd -c -C -D -deb -w -ver -h -? ] inputfile.inp

  -mnd       Make name directory
  -c         Continuation run
  -C         Continuation run; do not read restart file
  -D <dir>   Denotes the directory where files are written to
             (name in .inp file ignored).
  -deb       Run in debug mode (more detailed logging)
  -w         Allow to overwriting of existing data.
  -ver       Print version info
  -h -?      Print this help text.

Note: The name directory should be located on a local disk as the program may
cause massive network traffic and be very slow if the name directory
is located in a network file system.

For continuation runs (-c), the program reads the restart file to determine
which clusters have already been created and loads them into memory,
if required. Additional clusters can be defined in the input file.

For continuation runs (-C), the program does not read the restart file but
re-calculates all tasks of the job from the given input. Existing data files in
the name directory are checked for consistency with the input and will be
re-calculated if considered corrupted or inconsistent. This usually requires
using the -w option.

-------------------------------------------------------------------------------
"""

import getopt
import sys
import os

import clusterbuild_inp
import clusterbuild_job
from mjob import JobError
from minput import InputError
from parsingtools import absExpandPath

from mversion import __version__, __status__, __author__
from mversion import __copyright__, __maintainer__, __email__
from mversion import versionInfo


if __name__=="__main__":

    # command line options
    caller_name = sys.argv.pop(0)

    if "-h" in sys.argv or "-?" in sys.argv:
        sys.exit(__doc__)
    if "-ver" in sys.argv:
        versionInfo(progname = caller_name,
                version = __version__,
                status = __status__,
                author = __author__,
                copyright = __copyright__,
                maintainer = __maintainer__,
                support = __email__)
        sys.exit()
    if len(sys.argv) > 0:
        inpfilename = sys.argv.pop(-1)
    else:
        sys.exit(__doc__)


    opts, args = getopt.getopt(sys.argv, 'm:cCD:d:iw')
    continuation = False
    for o, a in opts:
        if o == "-c":
            continuation = True
    opts = [(o,a) for (o,a) in opts if not o == "-c"]

    # imput file
    if inpfilename:
        lininp = len(inpfilename)
        if inpfilename[lininp-4:lininp] == ".inp":
            inpfilename = absExpandPath(inpfilename)
            try:
                inp = clusterbuild_inp.ClusterbuildInput(inpfilename)
            except InputError as I:
                message = "ERROR while reading input: " +str(I) + "\n"
                sys.exit(message)

            inp.checkConsistency()
        else:
            message = "Input file name must the extension '.inp'. "
            message += 'Filename given: ' + str(inpfilename)
            sys.exit(message)
    else:
        message = "No input file."
        sys.exit(message)

    # create a job instance and execute
    if continuation:
        fname = os.path.join(inp.runsection.name, 'restart')
        job = clusterbuild_job.ClusterbuildJob.load(fname,inp, opts)
        job.log.info('Job imported from restart')
    else:
        job = clusterbuild_job.ClusterbuildJob(inp, opts)
        job.log.info('Job created from input file')
    try:
        job.execute()
    except KeyboardInterrupt as K:
        message = 'KeyboardInterrupt: saving files. This may take a while...'
        job.log.info(message)
        job.save()
        sys.exit('KeyboardInterrupt: exit.')


    job.cleanup()





