#!/usr/bin/env python3



import numpy

from minput import MInput, InputError
from clusterbuild_runsec import ClusterbuildRunSection
from mctdh_pbassec import PBasSection
from clusterbuild_refsec import ClusterbuildReferenceSection
from clusterbuild_expansionsec import ClusterbuildExpansionSection
from clusterbuild_modessec import ClusterbuildModesSection

from parsingtools import absExpandPath, find_section
from parsingtools import preparse_input, chop_end
from mtools import file2lines

class ClusterbuildInput(MInput):
    """Parser for input to generate clusters."""


    def __init__(self, filename=""):
        self.runsection = None
        self.referencesection = None
        self.expansionsection = None
        self.pbassection = None
        self.modessection = None
        if filename:
            self.import_file(filename)


    def __str__(self):
        string = "# INPUT (PARSED COPY)\n\n"
        if self.runsection is not None:
            string += str(self.runsection)
            string += "\n"
        if self.referencesection is not None:
            string += str(self.referencesection)
            string += "\n"
        if self.pbassection is not None:
            string += str(self.pbassection)
            string += "\n"
        if self.expansionsection is not None:
            string += str(self.expansionsection)
            string += "\n"
        if self.modessection is not None:
            string += str(self.modessection)
        string += "\nend-input\n"
        return string


    def import_file(self, filename):
        """
        Import an input file.
        """
        self.filename = absExpandPath(filename)
        # raw inport
        lines = file2lines(self.filename)
        lines = preparse_input(lines)
        lines = chop_end(lines)

        tags = ("RUN-SECTION", "end-run-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.runsection = ClusterbuildRunSection(seclines)

        tags = ("PBASIS-SECTION", "end-pbasis-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            tags = ("PRIMITIVE-BASIS-SECTION", "end-pbasis-section")
            seclines, lines = find_section(lines, tags)
        if not seclines:
            tags = ("PBASIS-SECTION", "end-primitive-basis-section")
            seclines, lines = find_section(lines, tags)
        if not seclines:
            tags = ("PRIMITIVE-BASIS-SECTION", "end-primitive-basis-section")
            seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.pbassection = PBasSection(seclines)

        tags = ("REFERENCE-SECTION", "end-reference-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.referencesection = ClusterbuildReferenceSection(seclines)

        tags = ("EXPANSION-SECTION", "end-expansion-section")
        seclines, lines = find_section(lines, tags)
        if seclines:
            self.expansionsection = ClusterbuildExpansionSection(seclines)


        tags = ("MODES-SECTION", "end-modes-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.modessection = ClusterbuildModesSection(seclines)


        if 'end-input' in lines:
            lines.pop(lines.index('end-input'))
        else:
            self.message = "Mising 'end-input' in file " + filename
            raise InputError(self.message)

        if len(lines):
            self.message = "Input file " + filename
            self.message += " contains extra lines: "
            self.message += str(lines)
            raise InputError(self.message)

        self.checkConsistency()



    def checkConsistency(self):
        "Check if the Data in the input file is complete and consistent"

        ndof = len(self.pbassection.coord_labels)
        maxorder = len(self.modessection.mode_combinations)


        nlabs = 0
        for mc in self.modessection.mode_combinations:
            for lab in mc:
                nlabs += 1
                if not lab in self.pbassection.coord_labels:
                    self.message = 'MODES-SECTION: label '
                    self.message +=  str(lab) + ' not '
                    self.message += 'found in PRIMITIVE-BASIS-SECTION.'
                    raise InputError(self.message)
        if nlabs != ndof:
            self.message = "MODES-SECTION: Inconsistent number of labels. "
            self.message += "Number of labels in PRIMITIVE-BASIS-SECTION: "
            self.message +=  str(ndof)
            self.message += ". Number of labels in MODES-SECTION: "
            self.message += str(nlabs)
            self.message += "."
            raise InputError(self.message)

        nlabs = 0
        for lab in self.referencesection.coord_labels:
            if not lab in self.pbassection.coord_labels:
                self.message = 'REFERENCE-SECTION: label '
                self.message +=  str(lab) + ' not '
                self.message += 'found in PRIMITIVE-BASIS-SECTION.'
                raise ValueError(self.message)
            nlabs += 1
        if nlabs != ndof:
            self.message = "REFERENCE-SECTION: Inconsistent "
            self.message += "number of labels. "
            self.message += "Number of labels in PRIMITIVE-BASIS-SECTION: "
            self.message +=  str(ndof)
            self.message += ". Number of labels in REFERENCE-SECTION: "
            self.message += str(nlans) + "."
            raise InputError(self.message)

        nref = 0
        for ref in self.referencesection.references:
            nref += 1
            if len(ref) != ndof:
                self.message = "REFERENCE-SECTION: reference " + str(nref)
                self.message += "Number of coordinates is diferent from "
                self.message += str(ndof)
                self.message += " - the number of labels in "
                self.message += "PRIMITIVE-BASIS-SECTION: "
                self.message += str(len(ref)) + "."
                raise ValueError(self.message)

        if self.referencesection.weights is not None:
            sumw = numpy.array(self.referencesection.weights).sum()
            if abs(sumw - 1.e0) > 1.e-8:
                self.message = "REFERENCE-SECTION: weights: "
                self.message += "The sum of the weights must be one. "
                self.message += "Weights are: "
                self.message += str(self.referencesection.weights)
                self.message += ". Sum = " + str(sumw)
                self.message +=  " Renormalized. New weights: "
                self.referencesection.weights = \
                    numpy.array(self.referencesection.weights)/sumw
                self.message += str(self.referencesection.weights)

            for n, weight in enumerate(self.referencesection.weights):
                if weight < 0.e0:
                    self.message = "REFERENCE-SECTION: weights: "
                    self.message += "Weights must be positive real numbers. "
                    self.message += "Wrong weight: No. " + str(n+1) + ", "
                    self.message += str(weight) + '.'
                    raise ValueError(self.message)

        if len(self.referencesection.sum_coords) > 0:
            if len(self.referencesection.references) <= 1:
                message = "EXPANSION-SECTION: 'join = S1(,S2,...)': "
                message += "'join' is only useful when more then "
                message += " one reference is used."
                raise InputError(message)
            missing = []
            for q in self.referencesection.sum_coords:
                if not q in self.pbassection.coord_labels:
                    missing.append(q)
            if len(missing) > 0:
                message = "EXPANSION-SECTION: 'join = S1(,S2,...)'"
                message += "Cordinates not found in PRIMITIVE-BASIS-SECTION: "
                message += ", ".join(missing)
                raise InputError(message)

        # Sort values in REFERENCE-SECTION
        # as in PRIMITIVE-BASIS-SECTION
        cl_labels = [lab for lab in self.referencesection.coord_labels]
        cl_refs = [numpy.array(ref).copy() \
                   for ref in self.referencesection.references]
        for nlab, lab in enumerate(cl_labels):
            idx = self.pbassection.coord_labels.index(lab)
            self.referencesection.coord_labels[idx] = lab
            for nref in range(len(cl_refs)):
                self.referencesection.references[nref][idx] \
                    = cl_refs[nref][nlab]





