#!/usr/bin/env python3
"""
 class ClusterbuildJob

 This class is used to generate a job object that generated the clusters.


"""


# standard modules
import sys
import os
import math
import re
import pickle
import random

# third party modules
import numpy

# local modules
from mjob import MJob, JobError
from clustertools import idx2doflist
from clustertools import expandWildCards
from mtools import openFile
from mtools import checkNumpyVersion
from mtools import getNonUnique
from mtools import multiIndexIterator
from mtools import importUserSource
from mtools import isInPath
from mtools import uniqifyList

from parsingtools import absExpandPath, getunit
from IOFiles import dvrCoordsDegF
from cluster_build import ClusterB

from mversion import __email__

# user provided functions are global
_DEFAULT_NAME_POTENTIAL = "potential"
potential = None
switch = None


# This class requires numpy version 1.0.0 or higher
_REQUIRED_NUMPY_VERSION = '1.0.0'

# File and directory names
_FILENAME_CLUSTER = "cluster"
_FILENAME_CLUSTERDIR = "clusters"
_FILENAME_CUT = "cut"
_FILENAME_CUTDIR = "cuts"
_FILENAME_POTFITDIR = "potfits"
_FILENAME_LOG = "clusterbuild.log"
_FILENAME_INPUTCOPY = "input"
_FILENAME_DVRDIR = 'dvr'
_FILENAME_DVR = 'dvr'
_FILENAME_DVRINPUT = "dvr.inp"
_FILENAME_TESTRESULTS = "testresults.txt"

# default commands
_EXE_MCTDH = 'mctdh84'
_EXE_POTFIT = 'potfit84'


# CONSTANTS (just for parsing and setup)
_TINY = 1.e-8
_AU2IC = 219474.7e0
_AU2EV = 27.2113961318e0


class ClusterbuildJob(MJob):

    def __parseCommandLine(self, opts=[]):
        "Elavuate the command line options."

        self.lmnd = False
        self.ldebug = False
        self.loverwrite = False
        self.name = None
        self.lretry = False

        for o, a in opts:
            if o == "-D":
                self.name = a
            elif o == "-m" and a == "nd":
                self.lmnd = True
            elif o == "-d" and a == "eb":
                self.ldebug = True
            elif o == '-w':
                self.loverwrite = True
            elif o == '-C':
                self.lretry = True
            else:
                message = "Unknown option: " + str(o) + str(a)
                raise NotImplementedError(message)




    def __init__(self, inp, args=[]):

        # init base class
        MJob.__init__(self)

        # check numpy version
        checkNumpyVersion(_REQUIRED_NUMPY_VERSION)

        # retry:
        self.lwarnrestart = False

        # command line arguments
        self.__parseCommandLine(args)

        # basic checks of input
        inp.checkConsistency()
        self.input = inp

        # if __init__ is called this is
        # a fresh job and not
        # a continuation job
        self.lcontinuation = False


        # import input file
        # run-section
        if self.name is None:
            self.name = str(self.input.runsection.name)
        self.title = str(self.input.runsection.title)

        # as soon as we have the name we create the name dir
        # if we have permission and start the logging system
        self.outdir = self.name
        self.loverwrite = bool(self.loverwrite) or \
                          bool(self.input.runsection.loverwrite)

        self.pes_min = self.input.runsection.pes_min
        self.pes_max = self.input.runsection.pes_max
        self.cluster_min = self.input.runsection.cluster_min
        self.cluster_max = self.input.runsection.cluster_max


        if self.input.runsection.pes_min_unit is not None:
            self.pes_min *= getunit(self.input.runsection.pes_min_unit)
        if self.input.runsection.pes_max_unit is not None:
            self.pes_max *= getunit(self.input.runsection.pes_max_unit)

        if self.input.runsection.cluster_min_unit is not None:
            self.cluster_min *= getunit(self.input.runsection.cluster_min_unit)
        if self.input.runsection.cluster_max_unit is not None:
            self.cluster_max *= getunit(self.input.runsection.cluster_max_unit)

        if self.pes_min is not None and self.pes_max is not None:
            if self.pes_min > self.pes_max:
                message = "RUN-SECTION: pes-min > pes-max."
                raise JobError(message)

        if self.cluster_min is not None and self.cluster_max is not None:
            if self.cluster_min > self.cluster_max:
                message = "RUN-SECTION: cluster-min > cluster-max."
                raise JobError(message)


        self.mkNameDir(self.lmnd)
        self.saveInputCopy()


        # set up the logging system
        self.logfilename = os.path.join(self.outdir,_FILENAME_LOG)
        self.logSetup(self.logfilename)
        if self.title is not None:
            self.log.info("Job-title: " + str(self.title))
        self.log.info("Starting job...")

        # transfer all other information from input to self

        self.title = str(self.input.runsection.title)
        if self.input.runsection.mctdh_exe is not None:
            prog = str(self.input.runsection.mctdh_exe)
            if (os.path.exists(prog) and os.path.isfile(prog)) or isInPath(prog):
                self.mctdh_exe = prog
            else:
                message = "RUN-SECTION: The executable " + prog
                message += " does not exist and is not in the $PATH."
                raise JobError(message)
        elif isInPath(_EXE_MCTDH):
            message = "Using '" + str(_EXE_MCTDH) + "' as MCTDH executable"
            self.log.debug(message)
            self.mctdh_exe = _EXE_MCTDH
        else:
            message = "Executable '" + str(_EXE_MCTDH) + "' not found in $PATH. "
            message += "Try to recompile MCTDHor specify an executable in the RUN-SECTION."
            raise JobError(message)

        self.lrunpotfit = bool(self.input.runsection.lrunpotfit)
        self.lmkpotfitinp = bool(self.input.runsection.lmkpotfitinp)
        if self.lrunpotfit:
            if self.input.runsection.potfit_exe is not None:
                prog = str(self.input.runsection.potfit_exe)
                if (os.path.exists(prog) and os.path.isfile(prog)) or isInPath(prog):
                    self.potfit_exe = prog
                else:
                    message = "RUN-SECTION: The executable " + prog
                    message += " does not exist and is not in the $PATH."
                    raise JobError(message)
            elif isInPath(_EXE_POTFIT):
                message = "Using '" + str(_EXE_POTFIT) + "' as POTFIT executable"
                self.log.debug(message)
                self.potfit_exe = _EXE_POTFIT
            else:
                message = "Executable '" + str(_EXE_POTFIT) + "' not found in $PATH. "
                message += "Try to recompile or specify an executable in the RUN-SECTION."
                raise JobError(message)






        # Mode combinations and primitive basis
        self.coord_labels = [str(lab) for lab in self.input.pbassection.coord_labels]
        self.ndof = len(self.coord_labels)
        self.mode_combinations = [mc for mc in self.input.modessection.mode_combinations]
        self.int_mode_combinations = []
        for mc in self.mode_combinations:
            int_mc = []
            for lab in mc:
                int_mc.append(self.coord_labels.index(lab))
            self.int_mode_combinations.append(tuple(int_mc))
        self.nmode = len(self.int_mode_combinations)

        # reference points and weights or switching function
        self.references = []
        for ref in self.input.referencesection.references:
            self.references.append(numpy.array(ref, dtype=float).copy())
        self.nref = len(self.references)
        self.switch_function = self.input.referencesection.switch_function
        self.weights = self.input.referencesection.weights

        if self.nref == 1 and self.weights is None and self.switch_function is None:
            self.weights=numpy.array([1.e0,], dtype=float)
        elif self.nref > 1 and self.weights is None and self.switch_function is None:
            self.weights=numpy.array([1.e0,]*self.nref, dtype=float)
            self.weights /= self.nref
            self.log.info("Set weights of clusters to 1/"+str(self.nref)+"." )

        if self.weights is not None and self.switch_function is None:
            self.weights = numpy.array(self.weights, dtype=float).copy()
        elif self.weights is not None and self.switch_function is not None:
            message = "REFERNCE-SECTION: "
            message += "Only one of 'switch-function' or 'weights' can be used."
            raise JobError(message)

        self.sum_coords = [q for q in self.input.referencesection.sum_coords]
        self.join_index = numpy.zeros(len(self.sum_coords))
        for n,q in enumerate(self.sum_coords):
            self.join_index[n] = self.coord_labels.index(q)


        #create sub-directories if required
        self.clusterdir = os.path.join(self.outdir,_FILENAME_CLUSTERDIR)
        self.cutdir = os.path.join(self.outdir,_FILENAME_CUTDIR)
        self.potfitdir = os.path.join(self.outdir,_FILENAME_POTFITDIR)
        if not os.path.exists(self.cutdir):
            os.mkdir(self.cutdir)
        elif not os.path.isdir(self.cutdir):
            message = "Object exists but is not a directory: "
            message += os.path.join(os.getcwd(),self.cutdir)
            raise JobError(message)
        if not os.path.exists(self.clusterdir):
            os.mkdir(self.clusterdir)
        elif not os.path.isdir(self.clusterdir):
            message = "Object exists but is not a directory: "
            message += os.path.join(os.getcwd(),self.clusterdir)
            raise JobError(message)
        if not os.path.exists(self.potfitdir) and (self.lrunpotfit or self.lmkpotfitinp):
            os.mkdir(self.potfitdir)
        elif self.lrunpotfit and not os.path.isdir(self.potfitdir):
            message = "Object exists but is not a directory: "
            message += os.path.join(os.getcwd(),self.potfitdir)
            raise JobError(message)


        # potential - expansion
        # expand the wildcards
        clusterlist = [cl for cl in self.input.expansionsection.clusterlist]
        expclusterlist = []
        for cluster in clusterlist:
            cluster1 = []
            for idx in cluster:
                try:
                    n = int(idx)
                    cluster1.append(n-1)
                except:
                    cluster1.append(idx)
            clist = expandWildCards(cluster1, self.mode_combinations)
            expclusterlist += clist
        self.clusterlist = expclusterlist
        self.ncluster = len(self.clusterlist)

        if sys.getrecursionlimit() < self.ncluster:
            message = "Large number of clusters: " + str(self.ncluster)
            message += ". This may result in many files, "
            message += "a considerable amount of CPU time "
            message += "and massive disk traffic."
            self.log.warn(message)
            sys.setrecursionlimit(self.ncluster)



        # clusters are generated through DOF, not modes.
        # and the DOF list serves as an ID and to find
        # dependencies
        self.dofclusterlist = []
        for cl in self.clusterlist:
            dofs = idx2doflist(cl, self.coord_labels, self.mode_combinations )
            self.dofclusterlist.append(tuple(dofs))

        # the clusterlist has to be unique
        doubles = getNonUnique(self.dofclusterlist)
        if len(doubles) > 0:
            message = "Multiply defined clusters have been found:\n"
            for cl in doubles:
                message += "DOF: " + ", ".join([self.coordlabels[i] for i in cl])
                message += "; Cluster: "
                message += str(doftpl2modetpl(cl,self.mode_combinations)) + "\n"
            raise JobError(message)


        # The clusters are generated and
        # stored in a list
        self.clusters = [ ]
        for n in range(self.nref):
            # this creates a new empty list every time append is called
            self.clusters.append([])


        # primitive grid
        self.grids = []
        self.mkDVR()
        self.readDVR()

        # import user-provided routines
        self.user_source = self.input.runsection.user_source
        self.potential_routine = self.input.runsection.potential_routine
        self.switch_function = self.input.referencesection.switch_function
        self._importSource()

        self.refidx = []

        for ref in range(self.nref):
            ridx = numpy.zeros(len(self.coord_labels), dtype=int)
            for i in range(len(self.coord_labels)):
                lfound = False
                closest = 0
                for j in range(len(self.grids[i])):
                    if abs(self.grids[i][j] - self.references[ref][i]) <= _TINY\
                           and lfound is False:
                        lfound = True
                        ridx[i] = j
                    else:
                        if abs(self.grids[i][closest] - self.references[ref][i]) >\
                           abs(self.grids[i][j] - self.references[ref][i]):
                            closest = j


                if lfound is False:
                    message = "Reference " + str(ref+1) + ": Reference points must "
                    message += "lie on a DVR grid point. Wrong coordinate: "
                    message += self.coord_labels[i] + ", given value "
                    message += str(self.references[ref][i])
                    message += ", closest point: "
                    message += str(self.grids[i][closest]) + "."
                    raise JobError(message)

            self.refidx.append(ridx.copy())

            for dof in self.dofclusterlist:
                cluster = ClusterB(dof, self.coord_labels,
                                   self.mode_combinations,
                                   ridx.copy(),potential,
                                   [q.copy() for q in self.grids],
                                   self.pes_min, self.pes_max,
                                   self.cluster_min, self.cluster_max)

                self.clusters[ref].append(cluster)

            for cluster in self.clusters[ref]:
                cluster.initDependencies(self.clusters[ref])
            roots = []
            for cluster in self.clusters[ref]:
                if cluster.isroot is True:
                    roots.append(cluster.idxStr())
            if len(roots) != 1:
                message = "The hierarchy of clusters must have one root. Detected "
                message += str(len(roots)) + " roots:\n"
                message += "\n".join(roots)
                raise JobError(message)


        self.completed_sum = []
        self.completed_potfit = []

        self.log.info("Init complete.")




    def _importSource(self):
        "Import user provided source code"

        global potential, switch

        if self.potential_routine is None:
            self.potential_routine = _DEFAULT_NAME_POTENTIAL
            message = "Setting potential-routine to "
            message += _DEFAULT_NAME_POTENTIAL + "."
            self.log.debug(message)
        message = "Impoting '" + str(self.potential_routine)
        message += "' from module " + self.user_source + "."
        self.log.debug(message)
        potential = importUserSource(self.user_source,self.potential_routine)

        if self.switch_function is not None:
            message = "Inpoting '" + self.switch_function + "' from file " + self.user_source
            self.log.debug(message)
            switch = importUserSource(self.user_source,self.switch_function)

        self.log.debug("Import sucessful.")
        self.log.info("Testing imported source: " + str(self.potential_routine))
        for n,ref in enumerate(self.references):
            q = ref.copy()
            try:
                energy = float(potential(q))
            except:
                energy = potential(q)
                message = "Testing reference " + str(n+1) + ": "
                message += "Potential routine '" +  str(self.potential_routine) + "' "
                message += "does not return a float but " + str(type(energy))
                message += "."
                raise JobError(message)
            message = "Potential routine '" +  str(self.potential_routine) + "': "
            message += "energy of reference point %d: "  % (n+1)
            message += "% .6e au = % .6e eV = % .6e cm-1" % (energy, energy*_AU2EV, energy*_AU2IC)
            self.log.info(message)



        if self.switch_function is not None:

            for n,ref in enumerate(self.references):
                q = ref.copy()
                weights = numpy.array(switch(q))
                if len(weights) != self.nref:
                    message = "Testing reference " + str(n+1) + ": "
                    message += "switch function '" + str(self.switch_function) + "' "
                    message += "returns " + str(len(weights)) + " values. "
                    message += "Should be " + str(self.nref)
                    message += " (number of reference points)."
                    raise JobError(message)
                if abs(weights.sum() - 1.e0) > _TINY:
                    message = "Testing reference " + str(n+1) + ": "
                    message += "switch function '" +  str(self.switch_function) + "': "
                    message += "sum of weights is not one but % .10e" % weights.sum()
                    self.log.warn(massage)

                message = "Switch function '" +  str(self.switch_function) + "': "
                message += " weights for reference point %d: "  % (n+1)
                sweights = []
                for w in weights:
                    sweights.append( "% .6e" % w)
                message += ", ".join(sweights)
                self.log.info(message)

            # we check 10 random points with the sum_coordinates set to the
            # ones of ref 0 to test if the switch function returns the same
            # values for all points
            comparew = []
            compareq = []
            for i in range(10):
                q = numpy.zeros(self.ndof, dtype=float)
                for n in range(self.ndof):
                    q[n] = random.random()*(self.grids[n][-1] - self.grids[n][0])
                    q[n] = q[n] - self.grids[n][-1]
                for d in self.sum_coords:
                    n = self.coord_labels.index(d)
                    q[n] = self.references[0][n]

                weights = numpy.array(switch(q))
                comparew.append(weights.copy())
                compareq.append(q.copy())
            lwarn = False
            for weights in comparew:
                diff = weights - comparew[0]
                if (diff.sum() - 1.e0) > _TINY:
                    lwarn = True

            if lwarn:
                message = "Testet 10 random coordinate sets with arguments of 'sum' set"
                message += " to reference point 1: The results were different but should"
                message += " be equal. See file " + _FILENAME_TESTRESULTS
                self.log.warn(message)

                fname = os.path.join(self.outdir, _FILENAME_TESTRESULTS)
                f = openFile(fname,"w",self.loverwrite)
                for n,w in enumerate(comparew):
                    message = "Coordinates:\n"
                    for i,d in enumerate(self.coord_labels):
                        message += " " + d + " % .6e" % (compareq[n][i])
                        message += "\n"
                    message += "\nWeights:\n"
                    for i,d in enumerate(w):
                        message += " % 2d " % (i+1)
                        message += " % .6e" % (d)
                        message += "\n"
                    message += "-"*30 + "\n\n"
                    f.write(message)
                f.close()


    def _mkIdxStr(self,idxstr):
        name = idxstr
        name = re.sub(",","_",name)
        name = re.sub("\(","",name)
        name = re.sub("\)","",name)
        name = re.sub(" +","",name)
        return name


    def mkDVR(self):
        string = "# MCTDH input file for generating the DVR\n\n"
        string += "RUN-SECTION\n"
        string += " name = " + _FILENAME_DVRDIR + "\n"
        string += " gendvr\n"
        string += "end-run-section\n\n"

        string += str(self.input.pbassection)
        string += "\nend-input\n"

        filename = os.path.join(self.outdir, _FILENAME_DVRINPUT)
        f = openFile(filename,'w',self.loverwrite)
        f.write(string)
        f.close()

        commandline = [self.mctdh_exe," -mnd"]
        if self.loverwrite is True:
            commandline.append("-w")

        commandline.append(_FILENAME_DVRINPUT)
        cwd = os.getcwd()
        os.chdir(self.outdir)
        self.log.debug("System call: " + " ".join(commandline))

        status=os.system(" ".join(commandline))

        os.chdir(cwd)

        if status:
            message = "MCDTH returned a nonzero exit code "
            message += "while generating the DVR file. "
            message += "Command line: " + " ".join(commandline)
            self.log.warn(message)
            raise JobError(message)


    def readDVR(self):

        # Now read the dvr from file
        filename = os.path.join(self.outdir, _FILENAME_DVRDIR, _FILENAME_DVR)
        for n,g in enumerate(self.coord_labels):
            self.grids.append(numpy.array(dvrCoordsDegF(filename,n+1)))




    def generateCuts(self):
        """
        generateCuts() --> None

        Writes the n-dimensional cuts through the PES to an ASCII file.
        """

        for n, clusterlist in enumerate(self.clusters):
            clusterlist.sort(key = lambda cl: cl.order )
            for cluster in clusterlist:
                if cluster.state == 0:
                    name1 = cluster.idxStr()
                    name = self._mkIdxStr(name1)
                    if self.nref > 1:
                        name = "cut_" + name + "_ref_" + str(n+1)
                    else:
                        name = "cut_" + name

                    cutfilename = os.path.join(self.cutdir, name)

                    lerr = False
                    if self.lretry:
                        if os.path.exists(cutfilename):
                            message = "Checking file " + cutfilename + "..."
                            self.log.info( message )
                            cfile = open(cutfilename,mode="r")
                            nlines = 0

                            for line in cfile:
                                nlines += 1
                                try:
                                    rdum = float(line)
                                except:
                                    lerr = True
                                    message = "   --> line in file is not a float."
                                    self.log.info( message )
                                if nlines > cluster.totgdim:
                                    lerr = True
                                    message = "   --> too many lines in file."
                                    self.log.info( message )
                                if lerr:
                                    cfile.close()
                                    break
                            if nlines != cluster.totgdim and not lerr:
                                lerr = True
                                cfile.close()
                                message = "   --> inconsistent number of lines in file."
                                self.log.info( message )


                    if not lerr and self.lretry:
                        cluster.setCutFileName(cutfilename)
                        cluster.state = 1
                    elif (lerr or not self.lretry):
                        message = "Ref " +str(n+1) + ": generating PES cut " + name1
                        message += " --> " + str(cluster.totgdim) + " points."
                        self.log.info( message )

                        cluster.calculateCut(cutfilename, self.loverwrite)
                        cluster.free()
                        self.save()
                    else:
                        message = "Uh, something went terribly wrong."
                        message += " Please fix me and/or send a bug report to "
                        message += __email__
                        self.log.error(message)



    def calculateClusters(self):
        """
        calculateClusters() --> None

        Generates the clusters from previously calculated PES cuts.
        Clusters are stored as ASCII files.
        """



        for n, clusterlist in enumerate(self.clusters):
            clusterlist.sort(key = lambda cl: cl.order )
            for cluster in clusterlist:
                if cluster.state == 1:
                    name1 = cluster.idxStr()
                    name = self._mkIdxStr(name1)

                    if self.nref > 1:
                        name = "cluster_" + name + "_ref_" + str(n+1)
                    else:
                        name = "cluster_" + name

                    clusterfilename = os.path.join(self.clusterdir,name)


                    lerr = False
                    if self.lretry:
                        if os.path.exists(clusterfilename):
                            message = "Checking file " + clusterfilename + "..."
                            self.log.info( message )
                            cfile = open(clusterfilename,mode="r")
                            nlines = 0

                            for line in cfile:
                                nlines += 1
                                try:
                                    rdum = float(line)
                                except:
                                    lerr = True
                                    message = "   --> line in file is not a float."
                                    self.log.info( message )
                                if nlines > cluster.totgdim:
                                    lerr = True
                                    message = "   --> too many lines in file."
                                    self.log.info( message )
                                if lerr:
                                    cfile.close()
                                    break
                            if nlines != cluster.totgdim and not lerr:
                                lerr = True
                                cfile.close()
                                message = "   --> inconsistent number of lines in file."
                                self.log.info( message )

                    if not lerr and self.lretry:
                        cluster.setClusterFileName(clusterfilename)
                        cluster.state = 2
                    elif (lerr or not self.lretry):
                        message = "Ref " +str(n+1) + ": calculating CLUSTER " + name1
                        message += " --> " + str(cluster.totgdim) + " points."
                        self.log.info( message )

                        cluster.calculateCluster(clusterfilename, self.loverwrite)
                        cluster.free() # deallocate before writing restart
                        self.save()

                    else:
                        message = "Uh, something went terribly wrong."
                        message += " Please fix me and/or send a bug report to "
                        message += __email__
                        self.log.error(message)


                elif cluster.state == 0:
                    message = str(cluster) + " --> Cannot generate cluster: "
                    message += "PES cut has not been calculated"
                    raise JobError(message)



    def sumClusters(self):

        if len(self.references) == 1:
            return

        FORMAT = "% .10e"

        lwarned = False

        # loop over all clusters
        for ncl,cluster in enumerate(self.clusters[0]):

            # is a switch function is used, only certain clusters
            # are summed
            if self.switch_function is not None:
                lsum = False
                for d in self.sum_coords:
                    if d in cluster.str_dof:
                        lsum = True
            else: # otherwise, all are summed
                lsum = True

            if lsum:

                if tuple(cluster.str_dof) in self.completed_sum:
                    continue

                # if we have to sum: open files
                sumclusters = []
                for n in range(self.nref):
                    sumclusters.append(self.clusters[n][ncl])
                sumfiles = []
                for cl in sumclusters:
                    sumfiles.append(openFile(cl.clusterfilename,"r"))

                index = numpy.zeros(len(cluster.str_dof), dtype=int)
                if len(cluster.gdim) < 1:
                    # one datapoint only
                    counter = [True,]
                else:
                    counter = multiIndexIterator(index, cluster.gdim)
                name1 = cluster.idxStr()
                name = self._mkIdxStr(name1)
                clusterfilename = "cluster_" + name + "_sum"
                clusterfilename = os.path.join(self.clusterdir,clusterfilename)
                message = "Summing clusters " + name1 + " --> "
                message += str(cluster.totgdim) + " points."
                self.log.info( message )
                f = openFile(clusterfilename, "w", self.loverwrite)

                sumarr = numpy.zeros(self.nref, dtype=float)
                weights = numpy.zeros(self.nref, dtype=float)

                # loop over all values and sum
                for index in counter:
                    for n in range(self.nref):
                        sumarr[n] = float(sumfiles[n].readline())

                    if self.switch_function is not None:
                        # this relies on
                        # that the
                        #
                        Q = numpy.array(sumclusters[0].qFromGridIndex(index))
                        weights = numpy.array(switch(Q))

                    else:
                        weights = self.weights

                    if lwarned is False:
                        if (weights.sum() - 1.e0) > _TINY:
                            message = "Sum of weights / switch is not 1."
                            self.log.warn(message)
                            lwarned = True

                    val = numpy.dot(weights,sumarr)
                    f.write(FORMAT % val + "\n")

                # close files
                f.close()
                for f in sumfiles:
                    f.close()

                self.completed_sum.append(tuple(cluster.str_dof))
            self.save()


    def runPotfit(self):
        "Make input files for POTFIT"

        def mkInpFile(name,vfile,pbas, modecomb, loverwrite):


            grids = {}
            for line in pbas:
                grids[line.split()[0]] = int( line.split()[2])

            modeflags =  [numpy.zeros(len(mc), dtype=int) for mc in modecomb]

            # determine which of the modes is contracted
            size = [1,]*len(modecomb)
            for n,mc in enumerate(modecomb):
                for d in grids.keys():
                    if d in mc:
                        i = list(mc).index(d)
                        modeflags[n][i] = 1
                        size[n] *=  grids[d]

            hole = [False,]*len(modecomb)

            for n,flgs in enumerate(modeflags):
                start = flgs[0]
                nchange = 0
                for fl in flgs:
                    if fl != start:
                        start = fl
                        nchange += 1
                if nchange > 1:
                    hole[n] = True
            for n,s in enumerate(size):
                if s is 1 or hole[n] is True:
                    size[n] = 0

            scontr = max(size)
            if s == 0:
                ncontr = -1
            else:
                ncontr = size.index(scontr)


            string = "# POTFIT INPUT \n"
            string += "# Coordinates: " + ",".join(grids.keys())
            string += "\n\nRUN-SECTION"
            string += "\n name = " + name
            string += "\nend-run-section\n\n"
            string += "OPERATOR-SECTION"
            string += "\n pes = readsrf{" + vfile
            string += " ascii}\nend-operator-section\n\n"
            string += "PRIMITIVE-BASIS-SECTION\n"
            for line in pbas:
                string += " " + line + "\n"
            string += "end-primitive-basis-section\n\n"
            string += "NATPOT-BASIS-SECTION\n"

            lcontrdone = False

            # modes with "holes" are different: they must be treated DOF wise
            for n,mode in enumerate(modeflags):
                if hole[n] is True:
                    for i,d in enumerate(mode):
                        if d == 1:
                            q = modecomb[n][i]
                            if lcontrdone is False and (ncontr == n or  ncontr == -1):
                                string += q  + " = contr\n"
                                lcontrdone = True
                            else:
                                string += q  + " = " + str(int(numpy.sqrt(grids[q]+1e0))) + "\n"
                else:
                    qs = []
                    pts = 0
                    for i,d in enumerate(mode):
                        if d == 1:
                            q = modecomb[n][i]
                            qs.append(modecomb[n][i])
                            pts += int(numpy.sqrt(grids[q]+1e0))
                    if len(qs) > 0:
                        if lcontrdone is False and (ncontr == n or  ncontr == -1):
                            string  += ",".join(qs) +  " = contr\n"
                            lcontrdone = True
                        else:
                            string  += ",".join(qs) +  " = " + str(pts*2) + "\n"

            string += "end-natpot-basis-section\n\n"
            string += "end-input\n"

            f = openFile(name+".inp","w",loverwrite)
            f.write(string)
            f.close()


        if os.path.exists(self.potfitdir) and not os.path.isdir(self.potfitdir):
            message = "Not a directory: " + str(self.potfitdir)
            raise JobError(message)


        cwd = os.getcwd()
        os.chdir(self.potfitdir)

        for cl in self.clusters[0]:

            name1 = cl.idxStr()
            name = self._mkIdxStr(name1)

            clusterfilename = "cluster_" + name
            if (self.nref > 1):
                clusterfilename += "_sum"
            lpotfit = True
            if cl.isSingleMode:
                lpotfit = False

            if self.lmkpotfitinp and lpotfit:

                cluster_id = []
                for d in cl.str_dof:
                    cluster_id.append(d)
                cluster_id = tuple(cluster_id)

                if cluster_id in self.completed_potfit:
                    continue

                potfitname = re.sub("cluster", "potfit",clusterfilename )
                clusterpath = os.path.join("../",_FILENAME_CLUSTERDIR,clusterfilename )

                pbas = []
                for q in cl.str_dof:
                    pbas.append(self.input.pbassection.strByLabel(q))


                mkInpFile(potfitname,clusterpath, pbas,
                          self.mode_combinations, self.loverwrite)


            if self.lrunpotfit and lpotfit:

                cmdline = self.potfit_exe
                if self.loverwrite:
                    cmdline += " -w"
                if not os.path.exists(potfitname):
                    cmdline += " -mnd"
                cmdline += " " + potfitname + ".inp"

                self.log.info("Calling POTFIT. Commandline: " + cmdline)
                exitcode = os.system(cmdline)
                if exitcode != 0:
                    message = "Something went wrong while calling POTFIT "
                    message += "for cluster " + clusterfilename
                    self.log.warning(message)
                else:
                    self.completed_potfit.append(cluster_id)


        os.chdir(cwd)

    def dump(self, filename="restart"):
        """
        Write the restart file to the name directory.
        This is done using pickle after setting not
        picklable members (files, log) to none.
        """
        # these things cannot be pickled:
        log = self.log
        self.log = None
        log_handler_stdout = self.log_handler_stdout
        self.log_handler_stdout = None
        log_handler_stderr = self.log_handler_stderr
        self.log_handler_stderr = None

        lwarn = False
        message = ""
        try:
            # now pickle
            f = open(os.path.join(self.outdir,filename), mode='wb')
            pickle.dump(self,f)
            f.close()
            if self.lwarnrestart:
                message = "The restart file has been fixed."
                self.log.info(message)
                self.lwarnrestart = False
        except Exception as E:
            lwarn = True
            message = str(E)

        # and restore
        self.log = log
        self.log_handler_stdout = log_handler_stdout
        self.log_handler_stderr = log_handler_stderr

        if lwarn:
            message = "The restart file may be currupted: " + message
            self.log.warn(message)
            self.lwarnrestart = True


    def load(cls, filename, inp=None, opts=[]):
        """
        load(filename, inp=None, opts = []) -> job

        return a ClusterJobAnalyse object from a pickle
        file. If call_potinit is True, then also
        user_potential_init from cluster_userfunc.py is called
        """

        message = "Restart file: " + str(filename)
        message1 = None
        job = MJob.load(filename)
        job.__parseCommandLine(opts)
        job.continuation = True
        job.logSetup(job.logfilename, append=True)
        job.log.info(message)

        job._initRestart(inp)

        return job
    load = classmethod(load)


    def _initRestart(self,inp):
        """
        Check the restart file against new input. load already calculated
        clusters into memory and perform additional tasks is needed.
        """
        global potential, switch

        if potential is None:
            self._importSource()

        for clist in self.clusters:
            for cluster in clist:
                cluster._initRestart(potential)

        if inp is None:
            return

        self.log.info("Reading EXPANSION-SECTION of input. Ignoring other sections.")


        # potential - expansion
        # expand the wildcards in input
        clusterlist = [cl for cl in inp.expansionsection.clusterlist]
        expclusterlist = []
        for cluster in clusterlist:
            cluster1 = []
            for idx in cluster:
                try:
                    n = int(idx)
                    cluster1.append(n-1)
                except:
                    cluster1.append(idx)
            clist = expandWildCards(cluster1, self.mode_combinations)
            expclusterlist += clist

        missing = []
        for cluster in self.clusterlist:
            if cluster not in expclusterlist:
                missing.append(str(cluster))
        if len(missing) > 0:
            message = "Removal of clusters in a restart calculation is not supported."
            message += " Missing clusters are: \n  " + "  \n".join(missing)
            raise JobError(message)

        for cl in expclusterlist:
            if cl not in self.clusterlist:
                self.clusterlist.append(cl)

        for cl in self.clusterlist:
            dofs = idx2doflist(cl, self.coord_labels, self.mode_combinations )

            if dofs not in self.dofclusterlist:
                message = "Adding cluster with DOF " + ",".join(dofs)
                self.dofclusterlist.append(tuple(dofs))
                self.log.info(message)

                for ref in range(self.nref):
                    cluster = ClusterB(dofs, self.coord_labels,
                                       self.mode_combinations,
                                       self.refidx[ref].copy(),potential,
                                       [q.copy() for q in self.grids],
                                       self.pes_min, self.pes_max,
                                       self.cluster_min, self.cluster_max)

                    self.clusters[ref].append(cluster)

        for ref in range(self.nref):
            for cluster in self.clusters[ref]:
                cluster.initDependencies(self.clusters[ref])
            roots = []
            for cluster in self.clusters[ref]:
                if cluster.isroot:
                    roots.append(cluster.idxStr())
            if len(roots) != 1:
                message = "The hierarchy of clusters must have one root. Detected "
                message += str(len(roots)) + " roots:\n"
                message += "\n".join(roots)
                raise JobError(message)



        self.ncluster = len(self.clusters[0])

        if sys.getrecursionlimit() < self.ncluster:
            message = "Large number of clusters: " + str(self.ncluster)
            message += ". This may result in many files, "
            message += "a considerable amount of CPU time "
            message += "and massive disk traffic."
            self.log.warn(message)
            sys.setrecursionlimit(self.ncluster)



    def doStep(self):
        "No meaning in this class"
        self.log.warn("Step has no meaning. Check your script.")


    def save(self):
        "Save all files"
        self.dump()


    def execute(self):
        "Execute the job"

        self.generateCuts()
        self.calculateClusters()
        if self.nref > 1:
            self.sumClusters()

        if self.lrunpotfit is True or self.lmkpotfitinp is True:
            self.runPotfit()



    def cleanup(self):
        "Write otput and close files"
        self.log.info('Cleaning up...')
        self.dump()

