#!/usr/bin/env python3

import numpy

from clusterstat_refsec import ClusterstatReferenceSection
from parsingtools import find_param, find_section
from minput import InputError

_TINY = 1.e-8

class ClusterbuildReferenceSection(ClusterstatReferenceSection):
    """
    ClusterbuildReferenceSection(ClusterstatReferenceSection)

    A simple wrapper for the build job as
    there is no difference to the reference
    section for statistical analysis.
    Feel free to change.
    """

    def __init__(self, lines):

        key = "sum"
        strval, lines = find_param(key, lines)
        if strval:
            self.sum_coords = strval.split(",")
        else:
            self.sum_coords = []

        ClusterstatReferenceSection.__init__(self, lines)

        if len(self.references) == 1 and len(self.sum_coords) > 0:
            message = "REFERENCE-SECTION: keyword 'sum' requires "
            message += "at least 2 reference points."
            raise InputError(message)
        if  len(self.sum_coords) > 0 and self.switch_function is None:
            message = "REFERENCE-SECTION: keyword 'sum' requires "
            message += "definition of a switching function."
            raise InputError(message)

        if self.switch_function is not None and len(self.sum_coords) == 0:
            message = "REFERENCE-SECTION: keyword 'switch-function' requires "
            message += " 'sum' keyword."
            raise InputError(message)

        for coord in self.sum_coords:
            missing = []
            if coord not in self.coord_labels:
                missing.append(coord)
            if len(missing) > 0:
                message = "Coordinate not found in reference "
                message += "definitions: " + ", ".join(missing)
                raise InputError(message)

        idx = [self.coord_labels.index(q) for q in self.coord_labels \
               if q not in self.sum_coords]

        if self.switch_function is not None:
            for n,ref in enumerate(self.references):
                if n == 0:
                    ref0 = numpy.array([float(ref[i]) for i in idx],
                                       dtype=float).copy()
                else:
                    refn = numpy.array([float(ref[i]) for i in idx],
                                   dtype=float).copy()
                    refn -= ref0
                    for i in idx:
                        if abs(refn[i]) > _TINY:
                            message = "Reference points must be equal "
                            message += "in the coordinates not used in the "
                            message += "switching function. Detected "
                            message += "difference at coordinate "
                            message += self.coord_labels[i] + " between "
                            message += "reference 1 and " + str(n+1) + "."
                            raise InputError(message)

    def __str__(self):

        string = str(self.openstr)
        if self.references:
            string += "\n begin-reference\n"
            for idx in range(len(self.coord_labels)):
                string += "  " + self.coord_labels[idx]
                for ref in range(len(self.references)):
                        string += "  " + str(self.references[ref][idx])
                string += "\n"
            string += " end-reference\n"
        if self.weights is not None:
            string += " weights = "
            string += ', '.join([str(w) for w in self.weights]) + '\n'
        if self.switch_function is not None:
            string += ' switch-function = '
            string += str(self.switch_function) + "\n"
        if len(self.sum_coords) > 0:
            string += " sum = "
            string += ", ".join(self.sum_coords)
            string += "\n"

        string += self.closestr + '\n'
        return string
