#!/usr/bin/env python3

from minput import MSection, InputError
from parsingtools import find_param, find_title, find_flag, getunit


class ClusterbuildRunSection(MSection):

    def __init__(self, lines):
        MSection.__init__(self,lines)
        if lines[0].upper() != "RUN-SECTION":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. First line reads: " \
                     + lines[0]
            raise Exception(message)
        elif lines[-1].lower() != "end-run-section":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. Last line reads: " \
                     + lines[-1]
            raise Exception(message)

        self.openstr = lines.pop(0)
        self.closestr = lines.pop(-1)

        # title first, as it matches a complete line
        # title may contain other keywords which could
        # confuse the imput parser later on
        strval, lines = find_title(lines)
        if strval:
            self.title = str(strval)
        else:
            self.title = None

        # obligatory key-value-pairs
        key = "name"
        strval, lines = find_param(key, lines)
        if strval:
            self.name = strval
        else:
            self.missing_key_error(key)

        key = "user-source"
        strval, lines = find_param(key, lines)
        if strval:
            self.user_source = strval
        else:
            self.missing_key_error(key)


        key = "potential-routine"
        strval, lines = find_param(key, lines)
        if strval:
            self.potential_routine = str(strval)
        else:
            self.potential_routine = None


        key = "mctdh-exe"
        strval, lines = find_param(key, lines)
        if strval:
            self.mctdh_exe = str(strval)
        else:
            self.mctdh_exe = None

        key = "potfit-exe"
        strval, lines = find_param(key, lines)
        if strval:
            self.potfit_exe = str(strval)
        else:
            self.potfit_exe = None

        key = "pes-min"
        strval, lines = find_param(key, lines)
        if strval:
            temp = strval.split(',')
            if len(temp) == 1:
                self.pes_min = float(temp[0])
                self.pes_min_unit = None
            elif len(temp) == 2:
                self.pes_min = float(temp[0])
                self.pes_min_unit = temp[1]
                unit = getunit(temp[1])
                if not unit:
                    message = "pes-min: Unknown unit: " + str(temp[1])
                    raise InputError(message)
            else:
                message = "pes-min: Too many keywords: " + str(temp)
                raise InputError(message)
        else:
            self.pes_min = None
            self.pes_min_unit = None


        key = "pes-max"
        strval, lines = find_param(key, lines)
        if strval:
            temp = strval.split(',')
            if len(temp) == 1:
                self.pes_max = float(temp[0])
                self.pes_max_unit = None
            elif len(temp) == 2:
                self.pes_max = float(temp[0])
                self.pes_max_unit = temp[1]
                unit = getunit(temp[1])
                if not unit:
                    message = "pes-max: Unknown unit: " + str(temp[1])
                    raise InputError(message)
            else:
                message = "pes-max: Too many keywords: " + str(temp)
                raise InputError(message)
        else:
            self.pes_max = None
            self.pes_max_unit = None


        key = "cluster-min"
        strval, lines = find_param(key, lines)
        if strval:
            temp = strval.split(',')
            if len(temp) == 1:
                self.cluster_min = float(temp[0])
                self.cluster_min_unit = None
            elif len(temp) == 2:
                self.cluster_min = float(temp[0])
                self.cluster_min_unit = temp[1]
                unit = getunit(temp[1])
                if not unit:
                    message = "cluster-min: Unknown unit: " + str(temp[1])
                    raise InputError(message)
            else:
                message = "cluster-min: Too many keywords: " + str(temp)
                raise InputError(message)
        else:
            self.cluster_min = None
            self.cluster_min_unit = None

        key = "cluster-max"
        strval, lines = find_param(key, lines)
        if strval:
            temp = strval.split(',')
            if len(temp) == 1:
                self.cluster_max = float(temp[0])
                self.cluster_max_unit = None
            elif len(temp) == 2:
                self.cluster_max = float(temp[0])
                self.cluster_max_unit = temp[1]
                unit = getunit(temp[1])
                if not unit:
                    message = "cluster-max: Unknown unit: " + str(temp[1])
                    raise InputError(message)
            else:
                message = "cluster-max: Too many keywords: " + str(temp)
                raise InputError(message)
        else:
            self.cluster_max = None
            self.cluster_max_unit = None


        # flags
        self.loverwrite, lines = find_flag("overwrite", lines)
        self.lrunpotfit, lines = find_flag("run-potfit", lines)
        self.lmkpotfitinp, lines = find_flag("potfit-inp", lines)
        if self.lrunpotfit is True:
            self.lmkpotfitinp = True

        if len(lines):
            message = "Input containes unrecognized lines: \n"
            message += "\n".join(lines)
            raise InputError(message)




    def  __str__(self):
        string =  str(self.openstr)

        # key-value-pairs
        if self.title is not None:
            string += "\n title = " + str(self.title)
        if self.name is not None:
            string += "\n name = " + str(self.name)
        if self.user_source is not None:
            string += "\n user-source = " + str(self.user_source)
        if self.potential_routine is not None:
            string += "\n potential-routine = "
            string += str(self.potential_routine)
        if self.mctdh_exe:
            string += "\n mctdh-exe = " + str(self.mctdh_exe)
        if self.potfit_exe:
            string += "\n potfit-exe = " + str(self.potfit_exe)

        if self.pes_min:
            string += "\n pes-min = " + str(self.pes_min)
            if self.pes_min_unit:
                string += ", "+ str(self.pes_min_unit)

        if self.pes_max:
            string += "\n pes-max = " + str(self.pes_max)
            if self.pes_max_unit:
                string += ", "+ str(self.pes_max_unit)

        if self.cluster_min:
            string += "\n cluster-min = " + str(self.cluster_min)
            if self.cluster_min_unit:
                string += ", "+ str(self.cluster_min_unit)

        if self.cluster_max:
            string += "\n cluster-max = " + str(self.cluster_max)
            if self.cluster_max_unit:
                string += ", "+ str(self.cluster_max_unit)


        #flags
        if self.loverwrite is True:
            string += "\n overwrite"
        if self.lrunpotfit is True:
            string += "\n run-potfit"
        if self.lmkpotfitinp is True:
            string += "\n potfit-inp"

        string += "\n" + self.closestr + "\n"
        return string





