#!/usr/bin/env python3
"""
-------------------------------------------------------------------------------
Purpose:     Statistical analysis for the cluster expansion
             of a potential surface.

Usage:       clusterstat [-mnd -c -D -deb -w -ver -h -? ] inputfile.inp

  -mnd       Make name directory
  -c         Continuation run
  -n <int>   Perform <int> steps (ignore value in input file)
  -t <file>  Use a trajectory stored in <file>
  -D <dir>   Denotes the directory where files are written to
             (name in .inp file ignored).
  -deb       Run in debug mode (more detailed logging)
  -w         Allow to overwrite existing data.
  -ver       Print version info
  -h -?      Print this help text.

-------------------------------------------------------------------------------
"""

import getopt
import sys
import os

import clusterstat_inp
import clusterstat_job
from mjob import JobError
from minput import InputError
from parsingtools import absExpandPath

from mversion import __version__, __status__, __author__
from mversion import __copyright__, __maintainer__, __email__
from mversion import versionInfo


if __name__=="__main__":

    # check command line options
    caller_name = sys.argv.pop(0)

    if "-h" in sys.argv or "-?" in sys.argv:
        sys.exit(__doc__)
    if "-ver" in sys.argv:
        versionInfo(progname = caller_name,
                version = __version__,
                status = __status__,
                author = __author__,
                copyright = __copyright__,
                maintainer = __maintainer__,
                support = __email__)
        sys.exit()
    if len(sys.argv) > 0:
        inpfilename = sys.argv.pop(-1)
    else:
        sys.exit(__doc__)


    opts, args = getopt.getopt(sys.argv, 'm:cn:t:D:d:w')


    continuation = False
    for o, a in opts:
        if o == "-c":
            continuation = True

    # read imput file
    if inpfilename:
        lininp = len(inpfilename)
        if inpfilename[lininp-4:lininp] == ".inp":
            inpfilename = absExpandPath(inpfilename)
            try:
                inp = clusterstat_inp.ClusterstatInput(inpfilename)
                inp.checkConsistency()
            except InputError as I:
                message = "ERROR while reading input: " +str(I) + "\n"
                sys.exit(message)

        else:
            message = "Input file name must the extension '.inp'. "
            message += 'Filename given: ' + str(inpfilename)
            sys.exit(message)
    else:
        message = "No input file."
        sys.exit(message)


    # create a job instance
    if continuation:
        cwd = os.getcwd()
        fname = os.path.join(cwd, inp.runsection.name, 'restart')
        if not os.path.exists(fname):
            message = "Error: File does not exist: " + fname
            sys.exit(message)
        if not os.path.isfile(fname):
            message = "Error: Not a file: " + fname
            sys.exit(message)

        job = clusterstat_job.ClusterstatJob.load(fname)

        # does not get here if an error has occured
        job.log.info('Job imported from restart')

    else:

        job = clusterstat_job.ClusterstatJob(inp, opts)

        # does not get here if an error has occured
        job.log.info('Job created from input file')


    # job is ready for execution
    try:
        job.execute()
    except KeyboardInterrupt as K:
        message = 'KeyboardInterrupt: saving files. This may take a while...'
        job.log.info(message)
        job.save()
        sys.exit('KeyboardInterrupt: exit.')
#    except:
#        message = "An unknown error has occured. Trying to Save files. "
#        message += "This may take a while..."
#        try:
#            job.save()
#        except:
#            message = "Saving failed. Exit."
#            sys.exit(message)
#        message = "Exit after unknown error."
#        job.log.error(message)
#        sys.exit()

    job.cleanup()

# done

