#!/usr/bin/env python3

# built in modules
import os

# third party modules
import numpy

# local nodules
from minput import MInput, InputError
from clusterstat_runsec import ClusterstatRunSection
from clusterstat_coordsec import ClusterstatCoordSection
from clusterstat_refsec import ClusterstatReferenceSection
from clusterstat_expansionsec import ClusterstatExpansionSection
from clusterstat_modessec import ClusterstatModesSection
from parsingtools import *
from mtools import file2lines


class ClusterstatInput(MInput):
    """Input """


    def __init__(self, filename=""):
        self.runsection = None
        self.referencesection = None
        self.expansionsection = None
        self.coordsection = None
        self.modessection = None

        if filename != "":
            self.import_file(filename)


    def __str__(self):
        string = "# Input for 'clusterstat'.\n\n"
        if self.runsection is not None:
            string += str(self.runsection)
            string += "\n"
        if self.referencesection is not None:
            string += str(self.referencesection)
            string += "\n"
        if self.coordsection is not None:
            string += str(self.coordsection)
            string += "\n"
        if self.expansionsection is not None:
            string += str(self.expansionsection)
            string += "\n"
        if self.modessection is not None:
            string += str(self.modessection)
        string += "\nend-input\n"
        return string


    def import_file(self, filename):
        """
        Import an input file.
        """
        self.filename = absExpandPath(filename)
        # raw inport
        lines = file2lines(self.filename)
        lines = preparse_input(lines)
        lines = chop_end(lines)

        # import sections
        tags = ("RUN-SECTION", "end-run-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.runsection = ClusterstatRunSection(seclines)

        tags = ("COORDINATES-SECTION", "end-coordinates-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.coordsection = ClusterstatCoordSection(seclines)

        tags = ("EXPANSION-SECTION", "end-expansion-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.expansionsection = ClusterstatExpansionSection(seclines)

        tags = ("REFERENCE-SECTION", "end-reference-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.referencesection = ClusterstatReferenceSection(seclines)


        tags = ("MODES-SECTION", "end-modes-section")
        seclines, lines = find_section(lines, tags)
        if not seclines:
            self.missing_section_error(tags[0], filename)
        self.modessection = ClusterstatModesSection(seclines)

        # check for more
        if 'end-input' in lines:
            lines.pop(lines.index('end-input'))
        else:
            self.message = "Mising 'end-input' in file " + filename
            raise InputError(self.message)

        if len(lines):
            self.message = "Input file " + filename
            self.message += " contains unrecognized data: \n"
            self.message += "\n".join(lines)
            raise InputError(self.message)

        # check consistency of input
        self.checkConsistency()


    def checkConsistency(self):
        """
        Check if the Data in the input file is complete and consistent
        This is for inter-section consistency. Intra-section consistency
        should be checkted in section objects.
        """
        ndof = len(self.coordsection.coord_labels)
        if ndof != len(self.coordsection.ranges_start):
            message = 'File ' + str(self.filename) + ': '
            message += 'Number of coordinate labels: ' + str(self.ndof)
            message += ' Number of start values: '
            message += str(len(self.coordsection.ranges_start))
            message += '. Numbers must be equal.'
            raise InputError(message)
        if ndof != len(self.coordsection.ranges_end):
            message = 'File ' + str(self.filename) + ': '
            message += 'Number of coordinate labels: ' + str(self.ndof)
            message += ' Number of end values: '
            message += str(len(self.coordsection.ranges_end))
            message += '. Numbers must be equal.'
            raise InputError(message)
        if ndof != len(self.coordsection.step_scale):
            message = 'Internal error while parsing input: '
            message += 'Number of coordinate labels: ' + str(self.ndof)
            message += ' Number of end step scale: '
            message += str(len(self.coordsection.step_scale))
            message += '. Numbers must be equal.'
            raise InputError(message)

        missing = []
        labs = []
        for mc in self.modessection.mode_combinations:
            for lab in mc:
                labs.append(lab)
                if not lab in self.coordsection.coord_labels:
                    missing.append(lab)
        if len(missing) > 0:
            message = 'MODES-SECTION: labels '
            message += ", ".join(missing)
            message += ' not found in COORDINATES-SECTION.'
            raise InputError(message)
        missing = []
        for lab in self.coordsection.coord_labels:
            if not lab in labs:
                missing.append(lab)
        if len(missing) > 0:
            message = 'COORDINATE-SECTION: labels '
            message += ", ".join(missing)
            message += ' not found in MODES-SECTION.'
            raise InputError(message)

        missing = []
        for lab in labs:
            if not lab in self.coordsection.coord_labels:
                missing.append(lab)
        if len(missing) > 0:
            message = "MODES-SECTION: coordinate labels "
            message += ", ".join(labs)
            message += " have not been assigned to a mode."
            raise InputError(message)
        missing = []
        for lab in self.referencesection.coord_labels:
            if not lab in self.coordsection.coord_labels:
                missing.apend(lab)
        if len(missing) > 0:
            message = 'REFERENCE-SECTION: labels '
            message +=  ", ".join(missing)
            message += ' not found in COORDINATES-SECTION.'
            raise InputError(self.message)
        missing = []
        for lab in self.coordsection.coord_labels:
            if not lab in self.referencesection.coord_labels:
                missing.apend(lab)
        if len(missing) > 0:
            message = 'REFERENCE-SECTION: for labels '
            message +=  ", ".join(missing)
            message += ' there no reference point.'
            raise InputError(message)



        missing = []
        # Sort values in REFERENCE-SECTION
        # as in COORDINATES-SECTION
        cl_labels = [lab for lab in self.referencesection.coord_labels]
        cl_refs = [numpy.array(ref).copy() \
                   for ref in self.referencesection.references]
        for nlab, lab in enumerate(cl_labels):
            idx = self.coordsection.coord_labels.index(lab)
            self.referencesection.coord_labels[idx] = lab
            for nref in range(len(cl_refs)):
                self.referencesection.references[nref][idx] = cl_refs[nref][nlab]





