#!/usr/bin/env python3
"""
-------------------------------------------------------------------------------
"""
# standard modules
import sys
import os
import math
import time
import pickle

# third party modules
import scipy
import numpy

# local modules

from parsingtools import getunit
from clustertools import expandWildCards
from clustertools import idx2doflist
from mtools import importUserSource
from mtools import checkNumpyVersion
from mtools import getNonUnique
from mtools import openFile
from mmath import metropolisStep
from mjob import MJob, JobError
from cluster_stat import ClusterA
from parsingtools import absExpandPath


# user provided functions are global
potential = None
probe = None
switch = None

# This class requires numpy version 1.0.0 or higher
_REQUIRED_NUMPY_VERSION = '1.0.0'

# default file names
_FILENAME_ENERGIES = "energies"
_FILENAME_PROBE = "probe"
_FILENAME_CLUSTERS = "clusters"
_FILENAME_COORDINATES = "coordinates"
_FILENAME_LOG = "clusterstat.log"
_FILENAME_ERROR = "curr_error_order"
_FILENAME_MEAN = "mean_error_order"
_FILENAME_RMS = "rms_error_order"
_FILENAME_INPUTCOPY = "input"
_FILENAME_RATIO = "ratio"
_FILENAME_SWITCH = "switch"
_FILENAME_TRAJWEIGHT = "trajweight"

# update log every _NUPDATE samples
_NUPDATE = 10000

# just small number
_TINY = 1.e-9

class ClusterstatJob(MJob):

    def __parseCommandLine(self, opts=[]):
        """
        Parse command line options:
        opt       var                  meaning (default)
        ---------------------------------------------------------------
        -mnd      lmnd                 Make name dir (False)
        -n <int>  nsteps               No of steps (-1)
        -t <file> trajfile             File holding a trajectory (None)
        -D <dir>  name                 Name directory ("")
        -deb      ldebug               Debug mode (False)
        -w        loverwrite           Allow owerwrite (False)
        ---------------------------------------------------------------

        Note: -ver -h -? are handled in __main__

        """

        self.nsteps = None
        self.lmnd = False
        self.ldebug = False
        self.loverwrite = False
        self.name = None
        self.trajectory_done = False
        self.trajectorypath = None

        for o, a in opts:
            if o == "-m" and a == "nd":
                self.lmnd = True
            elif o == '-n':
                try:
                    self.nsteps = int(a)
                except ValueError:
                    message = "Option -n must be followed by an integer."
                    raise ValueError(message)
            elif o == "-t":
                self.trajectorypath = absExpandPath(a)
                if not os.path.exists(self.trajectorypath) or \
                       not os.path.isfile(self.trajectorypath):
                    message = "Object does not exist or is not a file: "
                    message += str(self.trajectorypath)
                    raise JobError(message)
                self.trajectory_done = True
            elif o == "-D":
                self.name = a
            elif o == "-d" and a == "eb":
                self.ldebug = True
            elif o == '-w':
                self.loverwrite = True
            else:
                message = "Unknown option: " + str(o) + str(a)
                raise JobError(message)



    def __init__(self, inp, args=[]):



        # init base class
        MJob.__init__(self)

        # check numpy version
        checkNumpyVersion(_REQUIRED_NUMPY_VERSION)

        # command line arguments
        self.__parseCommandLine(args)


        # basic checks of input
        inp.checkConsistency()
        self.input = inp

        # if __init__ is called this is
        # a fresh job and not
        # a continuation job
        self.lcontinuation = False

        # import input file
        # run-section
        if self.name is None:
            self.name = str(self.input.runsection.name)
        self.title = str(self.input.runsection.title)

        # as soon as we have the name we create the name dir
        # if we have permission and start the logging system
        self.outdir = os.path.join(os.getcwd(),self.name)
        self.loverwrite = bool(self.loverwrite) or \
                          bool(self.input.runsection.loverwrite)
        self.mkNameDir(self.lmnd)
        self.saveInputCopy()

        # set up the logging system
        self.logfilename = os.path.join(self.outdir,_FILENAME_LOG)
        self.logSetup(self.logfilename)
        if self.title is not None:
            self.log.info("Job-title: " + str(self.title))
        self.log.info("Starting job...")

        # continue to import input
        if self.trajectorypath is None:
            if self.input.runsection.trajectorypath is not None:
                self.trajectorypath = absExpandPath(self.input.runsection.trajectorypath)
                if not os.path.exists(self.trajectorypath) or \
                       not os.path.isfile(self.trajectorypath):
                    message = "Object does not exist or is not a file: "
                    message += str(self.trajectorypath)
                    raise JobError(message)
                self.trajectory_done = True
        self.temperature = self.input.runsection.temperature
        if self.input.runsection.temperature_unit is not None:
            unit = getunit(self.input.runsection.temperature_unit)
            self.temperature *= unit

        self.lweightedtrajectory = False
        self.lusetrajweights = False


        if self.nsteps is None:
            self.nsteps = self.input.runsection.nsteps
        self.stepnorm = self.input.runsection.stepnorm
        self.outunit = self.input.runsection.outunit
        if self.outunit is None:
            self.outunit = 'au'
        self.outunit_float = 1.e0/getunit(self.outunit)



        self.lrms = bool(self.input.runsection.lrms)
        self.lmean = bool(self.input.runsection.lmean)
        self.lwalkonly = bool(self.input.runsection.lwalkonly)
        self.lignoreweights =  bool(self.input.runsection.lignoreweights)


        # coordinate-section
        self.coord_labels = [str(lab) for lab in self.input.coordsection.coord_labels]
        self.ndof = len(self.coord_labels)
        self.ranges_start = numpy.array(self.input.coordsection.ranges_start,dtype=float).copy()
        self.ranges_end = numpy.array(self.input.coordsection.ranges_end,dtype=float).copy()
        self.step_scale = numpy.array(self.input.coordsection.step_scale,dtype=float).copy()

        # modes-section
        self.mode_combinations = [mc for mc in self.input.modessection.mode_combinations]
        self.int_mode_combinations = []
        for mc in self.mode_combinations:
            int_mc = []
            for lab in mc:
                int_mc.append(self.coord_labels.index(lab))
            self.int_mode_combinations.append(tuple(int_mc))
        self.nmode = len(self.int_mode_combinations)

        # references
        self.references = []
        for ref in self.input.referencesection.references:
            self.references.append(numpy.array(ref,dtype=float))
        self.nref = len(self.references)

        if self.input.referencesection.weights is not None\
               and self.input.referencesection.switch_function is None:
            self.weights = numpy.array(self.input.referencesection.weights,dtype=float).copy()
            self.switch_function = None
        elif self.input.referencesection.switch_function is not None\
                 and self.input.referencesection.weights is None:
            self.switch_function = self.input.referencesection.switch_function
            self.weights = None
        else:
            self.weights = numpy.array([1.e0,]*self.nref, dtype=float)/self.nref
            if self.nref > 1:
                self.log.warn("Set all weights of references to %6.3f" % (1.e0/self.nref))
            self.switch_function = None
        # other cases should be ruled out by inp.checkConsistency()


        # import user-provided routines
        self.user_source = absExpandPath(self.input.runsection.user_source)
        self.potential_routine = self.input.runsection.potential_routine
        self.probe_routine = self.input.runsection.probe_routine
        self._importSource()


        # potential - expansion
        # expand the wildcards
        clusterlist = [cl for cl in self.input.expansionsection.clusterlist]
        expclusterlist = []
        for cluster in clusterlist:
            cluster1 = []
            for idx in cluster:
                try:
                    n = int(idx)
                    cluster1.append(n-1)
                except:
                    cluster1.append(idx)
            clist = expandWildCards(cluster1, self.mode_combinations)
            expclusterlist += clist
        self.clusterlist = expclusterlist
        self.ncluster = len(self.clusterlist)


        if self.ncluster > 500 and self.ncluster <= 2000:
            message = "Large number of clusters: " + str(self.ncluster)
            message += ". This may result in huge files and "
            message += "a considerable amount of CPU time."
            self.log.warn(message)
        elif self.ncluster > 2000:
            message = "Number of clusters is too large: " + str(self.ncluster)
            message += ". Aborting"
            raise ClusterError(message)
        if sys.getrecursionlimit() < self.ncluster*2:
            # the recursion limit is exceeded in some cases
            # when the job instance is sent through pickle
            # when the number of clusters is too large. this is
            # due to the pointer structure that inter-connects the clusters
            # in most cases one would not have larger numbers of clusters
            # then a few hundred so simply increasing the recursion limit
            # is not a problem here. in the future we might have to consider
            # destroying the pointer structure before saving and
            # rebuilding it after loading.
            #
            # recursion limit default in most cases: 1000
            # maximum here: 4000 (see code above)
            sys.setrecursionlimit(self.ncluster*2)

        # clusters are generated through DOF, not modes.
        # and the DOF list serves as an ID and to find
        # dependencies
        self.dofclusterlist = []
        for cl in self.clusterlist:
            dofs = idx2doflist(cl, self.coord_labels, self.mode_combinations )
            self.dofclusterlist.append(tuple(dofs))

        # the clusterlist has to be unique
        doubles = getNonUnique(self.dofclusterlist)
        if len(doubles) > 0:
            message = "Multiply defined clusters have been found:\n"
            for cl in doubles:
                message += "DOF: " + ", ".join([self.coordlabels[i] for i in cl])
                message += "; Cluster: "
                message += str(doftpl2modetpl(cl,self.mode_combinations)) + "."
                raise JobError(message)



        # we have read all input.
        # now set up internal data


        # the initial position of the random walker
        # is the first reference point
        # this might be changed
        self.curr_position = self.references[0].copy()

        # first call to potential routine
        # this should also initialize the PES routine
        # if necessary
        self.curr_pval = float(potential(self.curr_position))

        # The clusters are renerated and
        # stored in a list
        self.clusters = [ ]
        for n in range(self.nref):
            # this creates a new list every time append is called
            self.clusters.append([])

        for ref in range(self.nref):
            for dofs in self.dofclusterlist:
                cluster = ClusterA(dofs, self.coord_labels,
                                   self.mode_combinations,
                                   self.references[ref].copy(),potential)
                self.clusters[ref].append(cluster)


        # the list of clusters is sorted by order
        # this makes sure that if we walk
        # through the list to calculate
        # the cluster values, all lower order cluster are
        # holding valid data
        # the order is given by the number of modes
        # a cluster is defined on.
        for ref in range(self.nref):
            self.clusters[ref].sort(key = lambda cl: cl.order)

        # initialize dependencies of clusters once they are created
        # The clusters seek through the complete list and
        # identify those they depend on by their DOF list
        for ref in range(self.nref):
            tree = self.clusters[ref]
            for cluster in tree:
                cluster.initDependencies(tree)
            roots = []
            for cluster in tree:
                if cluster.isroot is True:
                    roots.append(cluster.idxStr())
            if len(roots) != 1:
                message = "The hierarchy of clusters must have one root. Detected "
                message += str(len(roots)) + " roots:\n"
                message += "\n".join(roots)
                raise JobError(message)

        # this is the maximum order of clusters
        self.order = self.clusters[0][-1].order

        # statistics over the clusters of certain orders
        ordercount = numpy.zeros(self.order+1, dtype=int)
        totcount = 0
        for cluster in self.clusters[0]:
            ordercount[cluster.order] += 1
            totcount += 1
        for i in range(self.order+1):
            message = "No. of Clusters of order %4i" % i
            message += ": %6i" % ordercount[i]
            self.log.info(message)
        message = "Total no. of Clusters: %14i" % totcount
        self.log.info(message)

        # statistical data is initialized
        self.count_accepted = 0
        self.count_skipped = 0
        self.count_tried = 0
        self.count_read = 0
        self.curr_energy = 0.e0
        self.sum_energy = 0.e0
        self.sum2_energy = 0.e0
        self.mean_energy = 0.e0
        self.rms_energy = 0.e0
        self.ratio = 0.e0

        self.curr_error = numpy.zeros(self.order+1, float)
        self.sum_error = numpy.zeros(self.order+1, float)
        self.sum2_error = numpy.zeros(self.order+1, float)
        self.mean_error = numpy.zeros(self.order+1, float)
        self.rms_error = numpy.zeros(self.order+1, float)

        self.curr_trajweight = 1.e0
        self.sum_trajweight = 0.e0
        self.mean_trajweight = 1.e0
        self.rms_trajweight = 1.e0

        self.log.info("Init complete.")


    def _importSource(self):
        "Import user provided source code"

        global potential, probe, switch

        if self.potential_routine is None:
            self.potential_routine = "potential"
            message = "Setting potential-routine to 'potential'. "
            self.log.debug(message)
        message = "Impoting '" + self.potential_routine
        message += "' from module " + self.user_source + "."
        self.log.debug(message)
        potential = importUserSource(self.user_source,self.potential_routine)

        if self.probe_routine is not None \
           and self.probe_routine != self.potential_routine:
            probe = importUserSource(self.user_source,self.probe_routine)
        else:
            self.probe_routine = self.potential_routine
            probe = potential
            message = "Setting probe-routine to '"
            message += self.potential_routine + "'."
            self.log.debug(message)

        if self.switch_function is not None:
            message = "Inpoting '" + self.switch_function + "' from file " + self.user_source
            self.log.debug(message)
            switch = importUserSource(self.user_source,self.switch_function)


        self.log.debug("Import sucessful.")
        self.log.info("Testing imported source: " + str(self.potential_routine))
        for n,ref in enumerate(self.references):
            q = ref.copy()
            energy = float(potential(q))
            message = "Potential routine '" +  str(self.potential_routine) + "': "
            message += "energy of reference point %d: "  % (n+1)
            message += "% .10e au." % energy
            self.log.info(message)

        for n,ref in enumerate(self.references):
            if self.probe_routine !=  self.potential_routine:
                q = ref.copy()
                energy = float(probe(q))
                message = "Probe routine '" +  str(self.probe_routine) + "': "
                message += "energy of reference point %d: "  % (n+1)
                message += "% .10e au." % energy
                self.log.info(message)



        if self.switch_function is not None:

            for n,ref in enumerate(self.references):
                q = ref.copy()
                weights = numpy.array(switch(q),dtype=float)
                if len(weights) != self.nref:
                    message = "Testing reference " + str(n+1) + ": "
                    message += "switch function '" + str(self.switch_function) + "' "
                    message += "returns " + str(len(weights)) + " values. "
                    message += "Should be " + str(self.nref)
                    message += " (number of reference points)."
                    raise JobError(message)
                if abs(weights.sum() - 1.e0) > _TINY:
                    message = "Testing reference " + str(n+1) + ": "
                    message += "switch function '" +  str(self.switch_function) + "': "
                    message += "sum of weights is not one but % .10e" % weights.sum()
                    self.log.warn(massage)

                message = "Switch function '" +  str(self.switch_function) + "': "
                message += " weights for reference point %d: "  % (n+1)
                sweights = []
                for w in weights:
                    sweights.append( "% .4e" % w)
                message += ", ".join(sweights)
                self.log.info(message)

            # we check 5 random points with the sum_coordinates set to the
            # ones of ref 0 to test if the switch function returns the same
            # values for all points
            comparew = []
            compareq = []
            for i in range(5):
                q = numpy.zeros(self.ndof, dtype=float)
                for n in range(self.ndof):
                    q[n] = random.random()*(self.ranges_end[n] - self.ranges_start[n])
                    q[n] = q[n] - self.ranges_start[n]
                for d in self.sum_coords:
                    n = self.coord_labels.index(d)
                    q[n] = self.references[0][n]

                weights = numpy.array(switch(q))
                comparew.append(weights.copy())
                compareq.append(q.copy())
            lwarn = False
            for weights in comparew:
                diff = weights - comparew[0]
                if (diff.sum() - 1.e0) > _TINY:
                    lwarn = True

            if lwarn:
                message = "Testet 5 random coordinate sets with arguments of 'sum' set"
                message += " to reference point 1: The results were different but should"
                message += " be equal. See file " + _FILENAME_TESTRESULTS
                self.log.warn(message)

                fname = os.path.join(self.outdir, _FILENAME_TESTRESULTS)
                f = openFile(fname,"w",self.loverwrite)
                for n,w in enumerate(comparew):
                    message = "Coordinates:\n"
                    for i,d in enumerate(self.coord_labels):
                        message += " " + d + " % .4e" % (compareq[n][i])
                        message += "\n"
                    message += "\nWeights:\n"
                    for i,d in enumerate(w):
                        message += " % 2d " % (i+1)
                        message += " % .4e" % (d)
                        message += "\n"
                    message += "-"*30 + "\n\n"
                    f.write(message)
                f.close()



    ####################################################
    # I/O routines
    ####################################################
    def _openFiles(self):
        "open output files"


        if self.lcontinuation is True:
            wmode = "a"
            lreplace = True
        else:
            wmode = "w"
            lreplace = False


        fname = os.path.join(self.outdir,_FILENAME_ERROR)
        self.files.error = openFile(fname,wmode,self.loverwrite)
        fname =  os.path.join(self.outdir,_FILENAME_ENERGIES)
        self.files.energies = openFile(fname,wmode,self.loverwrite)

        if self.lmean:
            fname = os.path.join(self.outdir,_FILENAME_MEAN)
            self.files.mean = openFile(fname,wmode,self.loverwrite)
        if self.lrms:
            fname = os.path.join(self.outdir,_FILENAME_RMS)
            self.files.rms = openFile(fname,wmode,self.loverwrite)


        for n in range(self.nref):
            ref = "_ref_" + str(n+1)
            name = _FILENAME_CLUSTERS + ref
            fname = os.path.join(self.outdir, name)
            f = openFile(fname,wmode,self.loverwrite)
            self.files.appendFile(name,f,lreplace)

        # Trajectory or Psi file
        fname = self.trajectorypath
        self.files.trajectory = openFile(fname,'r')

        # check format of file
        buf = self.files.trajectory.readline()
        coords = buf.split()

        if len(coords) < self.ndof:
            message = "Trajectoy file " + self.files.trajectory.name
            message += " seems corrupted near line 1"
            message += ". Expected " + str(self.ndof) + " coordinate values "
            message += "but found " + str(len(coords)) + "."
            raise JobError(message)

        elif len(coords) == self.ndof:
            self.lweightedtrajectory = False
            message = "Trajectory file contains coordinate data only."
            self.log.info(message)

        elif len(coords) == self.ndof + 1:
            self.lweightedtrajectory = True
            if not self.lignoreweights:
                message = "Trajectory file " + self.files.trajectory.name
                message += " contains more than coordinate data. "
                message += "Taking last column as weights."
                self.log.info(message)

                message = "Scanning file " + self.files.trajectory.name


            else:
                message = "Trajectory file contains more than coordinate data. "
                message += "Ignoring last column."
                self.log.info(message)

        elif len(coords) > self.ndof + 1:
            message = "Trajectoy file " + self.files.trajectory.name
            message += " contains more data then expected. "
            message += "Expected " + str(self.ndof) + " coordinate values "
            message += " and possibly a weighting factor "
            message += "but found " + str(len(coords)) + " values. "
            message += "Will use the first " + str(self.ndof)
            message += " columns as coordinate values"
            if not self.lignoreweights:
                message += " and the " + str(self.ndof+1) + "th as weights."
            else:
                message += "."
            self.log.warn(message)

        self.lusetrajweights = self.lweightedtrajectory and not self.lignoreweights
        self.files.trajectory.seek(0)

        self.mean_trajweight = 1.e0
        self.rms_trajweight = 1.e0

        message = "Scanning file " + self.files.trajectory.name
        if self.lusetrajweights:
            message += " for the mean weight."
            self.mean_trajweight = 0.e0
            self.rms_trajweight = 0.e0

        self.log.info(message)

        self.files.trajectory.seek(0)
        buf = self.files.trajectory.readline()


        nread = 0
        while buf != "" and nread < self.nsteps:
            arr = buf.split()
            if self.lusetrajweights:
                try:
                    self.mean_trajweight += float(arr[self.ndof])
                    self.rms_trajweight += float(arr[self.ndof])**2
                except:
                    message = "File " + self.trajectorypath + " seems to be corrupted near"
                    message += " line " + str(nread +1) + "."
                    raise JobError(message)

            buf = self.files.trajectory.readline()
            nread += 1

        if self.nsteps > nread:
            message = "Coordinates file only contains " + str(nread) + " data sets."
            message += " Resetting 'steps' to "+ str(nread) + "."
            self.log.warn(message)
            self.nsteps = nread


        if self.lusetrajweights:
            self.mean_trajweight /= nread
            self.rms_trajweight = numpy.sqrt(self.rms_trajweight/nread)

            message = "Mean weight is %16.4f." % self.mean_trajweight
            self.log.info(message)
            message = "RMS weight is  %16.4f." % self.rms_trajweight
            self.log.info(message)

            # append a wile containing weights of the trajectory
            fname = os.path.join(self.outdir,_FILENAME_TRAJWEIGHT)
            self.files.trajweight = openFile(fname,wmode,self.loverwrite)

        self.files.trajectory.seek(0)

        if self.switch_function is not None:
            fname = os.path.join(self.outdir,_FILENAME_SWITCH)
            self.files.switch = openFile(fname,wmode,self.loverwrite)

        # todo: seek forward for continuation


    def _writeOutput(self):
        "Write output files."


        message = "%16.8f " % (self.curr_energy * self.outunit_float)
        message += "%16.8f " % (self.mean_energy * self.outunit_float)
        message += "%16.8f " % (self.rms_energy  * self.outunit_float)
        message += "\n"
        self.files.energies.write(message)

        message = ""
        for elem in self.curr_error:
            message += "%16.8f " % (elem*self.outunit_float)
        message += "\n"
        self.files.error.write(message)


        if self.lmean:
            message = ""
            for elem in self.mean_error:
                message += "%16.8f " % (elem*self.outunit_float)
            message += "\n"
            self.files.mean.write(message)
        if self.lrms:
            message = ""
            for elem in self.rms_error:
                message += "%16.8f " % (elem*self.outunit_float)
            message += "\n"
            self.files.rms.write(message)



        for n in range(self.nref):
            ref = "_ref_" + str(n+1)
            name = _FILENAME_CLUSTERS + ref
            cfile = self.files[name]
            message = ""
            for elem in self.clusters[n]:
                key = tuple(elem.int_dof)
                message +=  "%16.8f " % (elem.clusters[key]*self.outunit_float)
            message += "\n"
            cfile.write(message)

        if self.switch_function is not None:
            message = ""
            for n in range(self.nref):
                message += "%16.8f "  % self.weights[n]
            message += "\n"
            self.files.switch.write(message)

        if self.lusetrajweights:
            message = "%16.8f "  % self.curr_trajweight
            message += "\n"
            self.files.trajweight.write(message)

        self.files.flushWriteables()



    def load(cls, filename):
        """
        load(filename) -> job

        return a ClusterstatJob object from a pickle
        file.

        Arguments:
        filename  - name of the pickle file

        """

        message = "Restart file: " + str(filename)
        message1 = None

        f = open(filename, mode='rb')
        job= pickle.load(f)
        job.lcontinuation = True

        job.logSetup(logfilename=job.logfilename, append=True)

        message = "Job status: "
        if job.trajectory_done is False:
            message += "Trajectory has not been completed. "
            message += "Current sampe:  %d of %d." % (job.count_accepted,job.nsteps)
            job.log.info(message)
        else:
            message += "Trajectory is completed. Calculating clusters."
            message += "Current sampe:  %d of %d." % (job.count_accepted,job.nsteps)
            job.log.info(message)

        global potential, probe, switch
        if potential is None:
            job._importSource()

        for clist in job.clusters:
            for cluster in clist:
                cluster._initRestart(potential)


        message = "Restart stopped at sample %d " % (job.count_accepted +1)
        job.log.info(message)
        return job
    load = classmethod(load)


    def calculateCuts(self):
        "Calculate the potential cuts for the current position."

        for n in range(self.nref):
            for cluster in self.clusters[n]:
                cluster.calculateCut(self.curr_position)

    def calculateClusters(self):
        "Calculate the potential clusters from the current PES-cuts."
        for n in range(self.nref):
            for cluster in self.clusters[n]:
                cluster.calculateCluster()


    def probFunc(self, Q):
        "Get the probability associated with a point"
        pot=float(probe(Q))
        return math.exp(-pot/self.temperature)



    ####################################################
    # interface
    ####################################################
    def stepMetropolis(self):
        "Try until one metropolis step has been accepted."
        accepted = False
        q = self.curr_position.copy()
        pval = self.curr_pval
        trycount = 0
        while not accepted:
            accepted,q,pval = metropolisStep(q,pval,self.probFunc,
                                             xmin=self.ranges_start.copy(),
                                             xmax=self.ranges_end.copy(),
                                             stepweight=self.step_scale.copy(),
                                             stepnorm=self.stepnorm)

            self.count_tried += 1
            trycount += 1
            if trycount > _NUPDATE:
                message = "Tried %d times but no Metropolis step " % _NUPDATE
                message += "had been accepted. Temperature too low "
                message += "or step-norm too large? I keep trying..."
                self.log.warn(message)
                trycount = 0

        self.count_accepted += 1
        self.curr_position = q.copy()
        self.curr_pval = pval


    def randomWalk(self):
        "Perform the random walk"

        if self.lcontinuation:
            mode = "a"
        else:
            mode = "w"

        fname = os.path.join(self.outdir,_FILENAME_COORDINATES)
        self.files.coordinates = openFile(fname,mode,self.loverwrite)
        self.trajectorypath = fname

        fname = os.path.join(self.outdir,_FILENAME_PROBE)
        self.files.probe = openFile(fname,mode,self.loverwrite)

        fname = os.path.join(self.outdir,_FILENAME_RATIO)
        self.files.ratio = openFile(fname,mode,self.loverwrite)

        for i in range(self.nsteps):
            # do the random walk intill the
            # trajectory has reqired length
            self.stepMetropolis()

            # write the trajectory
            message = ""
            for n in range(self.ndof):
                message += "%16.8f " % self.curr_position[n]
            message += "\n"
            self.files.coordinates.write(message)

            # this statistics stuff as well as
            # re-calculating the potential
            # slows down the calculation of the trajectory
            # approximately by a factor of 2.
            # however, obtaining the ttajectory is not the
            # demanding part.
            self.curr_energy = probe(self.curr_position)
            self.sum_energy += self.curr_energy
            self.sum2_energy += self.curr_energy**2
            self.mean_energy = self.sum_energy/self.count_accepted
            self.rms_energy = numpy.sqrt(self.sum2_energy/self.count_accepted)

            message = "%16.8f " % (self.curr_energy * self.outunit_float)
            message += "%16.8f " % (self.mean_energy * self.outunit_float)
            message += "%16.8f " % (self.rms_energy  * self.outunit_float)
            message += "\n"
            self.files.probe.write(message)

            self.ratio = float(self.count_accepted)/self.count_tried
            message = "%8.4f " % (self.ratio*100.e0) + "%\n"
            self.files.ratio.write(message)


            # write statistics every once in a while
            if self.count_accepted % _NUPDATE == 0 or\
                   self.count_accepted == self.nsteps:
                message = "Steps: %12d, "  % self.count_accepted
                message += "Mean: %12.3f " % (self.mean_energy*self.outunit_float)
                message += str(self.outunit) + ",  "
                message += "RMS: %12.3f " % (self.rms_energy*self.outunit_float)
                message += str(self.outunit) + ",  "
                message += "Accept: %6.1f " % (self.ratio*100) + "%."
                self.log.info(message)


        # Cleanup, log and reset:
        self.files.coordinates.close()
        self.files.popFile(_FILENAME_COORDINATES)

        self.files.probe.close()
        self.files.popFile(_FILENAME_PROBE)

        self.files.ratio.close()
        self.files.popFile(_FILENAME_RATIO)

        self.trajectory_done = True
        self.curr_energy = 0.e0
        self.sum_energy = 0.e0
        self.sum2_energy = 0.e0
        self.mean_energy = 0.e0
        self.rms_energy = 0.e0


    def doStep(self):
        """
        doStep() --> isValid

        Read one line from the potential and calculate the clusters
        Requires files to be open and positioned correctly.

        isValid is True if the sample is to be used, False otherwise

        """

        if self.trajectory_done is False:
            raise JobError("Trajectory not specified or not created.")

        # load a set of coordinates from trajectory
        def readQ(f):
            line = f.readline()
            vals = [float(s) for s in line.split()]
            return numpy.array(vals, dtype=float)


        # check if coordinates are in given range
        def isInRange(q,ranges_start, ranges_end):
            for n, qq  in enumerate(q):
                if qq < ranges_start[n] or qq > ranges_end[n]:
                    return False
            return True

        # check if the coordinates are valid
        q = readQ(self.files.trajectory)
        self.count_read += 1

        # some checks on trajectory file
        ndof = len(q)
        if ndof == 0:
            self.eof = True
            self.count_read -= 1
            return False
        elif ndof < self.ndof:
            message = "Trajectoy file " + self.files.trajectory.name
            message += " seems corrupted near line " + str(self.count_read)
            message += ". Expected " + str(self.ndof) + " coordinate values "
            message += "but found" + str(ndof) + "."
            raise JobError(message)

        if self.lusetrajweights:
            if  ndof < self.ndof + 1:
                message = "Trajectoy file " + self.files.trajectory.name
                message += " seems corrupted near line " + str(self.count_read)
                message += ". Expected " + str(self.ndof+1) + " columns of data "
                message += "but found" + str(ndof) + "."
                raise JobError(message)

            self.curr_trajweight = float(q[self.ndof])
            q = q[0:self.ndof].copy()
        else:
            self.curr_trajweight = 1.e0
            q = q[0:self.ndof].copy()

        if not isInRange(q,self.ranges_start, self.ranges_end):
            self.count_skipped += 1
            return False

        # q is checked - now do the statistics
        self.curr_position = q
        self.count_accepted += 1
        self.calculateCuts()
        self.calculateClusters()


        self.curr_energy = float(potential(q))


        if self.nref > 1:
            if self.switch_function is not None:
                self.weights = numpy.array(switch(q),dtype=float)

        contrib_order = numpy.zeros(self.order+1, dtype=float)

        for n in range(self.nref):
            contrib_order_ref = numpy.zeros(self.order+1, dtype=float)
            for cluster in self.clusters[n]:
                tpl = tuple(cluster.int_dof)
                contrib_order_ref[cluster.order] += cluster.clusters[tpl]
            contrib_order += (self.weights[n]) * contrib_order_ref

        self.curr_error = self.curr_energy - contrib_order.cumsum()

        if self.lusetrajweights:
            wscale = self.curr_trajweight/self.mean_trajweight
            self.sum_energy += self.curr_energy*wscale
            self.sum2_energy += self.curr_energy**2*wscale
        else:
            self.sum_energy += self.curr_energy
            self.sum2_energy += self.curr_energy**2

        self.mean_energy = self.sum_energy/self.count_accepted
        self.rms_energy = numpy.sqrt(self.sum2_energy/self.count_accepted)
        return True

    def save(self):
        "Save a restart file to disk."
        self.wrRestart()


    def execute(self):
        "Execute the job."

        if self.trajectory_done is False:
            message = "Obtaining Monte-Carlo trajectory with "
            message += str(self.nsteps) + " steps."
            self.log.info(message)
            self.randomWalk()
            self.wrRestart()
            message = "Obtaining trajectory is complete."
            self.log.info(message)
        elif self.trajectorypath is not None:
            message = "Reading trajectory from file "
            message += str(self.trajectorypath)
            self.log.info(message)
        else:
            message = "No trajectory given. Specify trajectory file "
            message += "or perform a random walk."
            raise JobError(message)

        if self.lwalkonly is True:
            self.log.info("Nothing left to do.")
            return

        message = "Calculating clusters. This may take a while..."
        self.log.info(message)

        # prepare for main job
        self._openFiles()
        if self.lcontinuation:
            message = "Repositioning files..."
            self.log.debug(message)
            for i in range(self.count_read):
                self.files.trajectory.readline()
            self.log.debug("Ready.")
        else:
            self.count_read = 0
            self.count_accepted  = 0
            self.count_skipped = 0

        self.eof = False
        self.sum_error = numpy.zeros(self.order+1, dtype=float)
        self.sum2_error = numpy.zeros(self.order+1, dtype=float)
        self.mean_error = numpy.zeros(self.order+1, dtype=float)
        self.rms_error = numpy.zeros(self.order+1, dtype=float)

        nt = 100
        # loop through all sampling points
        for i in range(self.nsteps):

            if i == 0:
                starttime = time.time()
            if self.nsteps > nt and i == nt:
                timent =  time.time()
                dt = timent - starttime
                duration = dt*(self.nsteps-nt)/nt
                endtime = starttime + duration
                tpl = time.localtime(endtime)
                message = "Estimated completion: "
                message += time.asctime(tpl)
                self.log.info(message)

            # doStep returns True on success
            lskip = not self.doStep()
            # stop loop if files are at end
            if self.eof:
                break

            if not lskip:

                if ((i+1) % _NUPDATE)  == 0:
                    message = "Processed samples: %12d / %12d." % (i+1,self.nsteps)
                    self.log.info(message)
                    self.wrRestart()

                # update statistics

                if self.lusetrajweights:
                    wscale = self.curr_trajweight/self.mean_trajweight
                    self.sum_error += self.curr_error.copy()*wscale
                    self.sum2_error += self.curr_error.copy()**2*wscale
                else:
                    self.sum_error += self.curr_error.copy()
                    self.sum2_error += self.curr_error.copy()**2

                self.mean_error = self.sum_error.copy()/self.count_accepted
                self.rms_error = numpy.sqrt(self.sum2_error.copy()/self.count_accepted)
                self._writeOutput()

        # save
        self.wrRestart()

        # we are done
        self.log.info("Nothing left to do.")
        self.log.info("SUMMARY:")
        message = "Points read from trajectory file: %16d" % self.count_read
        self.log.info(message)
        message = "Of those points were skipped:     %16d" % self.count_skipped
        self.log.info(message)
        if self.eof:
            message = "Stopped because coordinates file '" + str(self.trajectorypath) + "' ended."
            self.log.warn(message)
        if self.count_read <= self.count_skipped:
            message = "All data points were skipped. No data was produced."
            self.log.warn(message)



    def cleanup(self):
        "Write otput and close files"
        self.log.info('Cleaning up...')
        self.wrRestart()
        self.files.closeAll()
        self.log.info('Done.')

