#!/usr/bin/env python3

import re
from minput import MSection, InputError
from parsingtools import find_param, find_section

class ClusterstatReferenceSection(MSection):


    def __init__(self, lines):

        MSection.__init__(self,lines)
        if lines[0].upper() != "REFERENCE-SECTION":
            message="Input seems not to contain a " \
                     + "REFERENCE-SECTION. First line reads: " \
                     + lines[0]
            raise InputError(message)
        elif lines[-1].lower() != "end-reference-section":
            message="Input seems not to contain a " \
                     + "REFERENCE-SECTION. Last line reads: " \
                     + lines[-1]
            raise InputError(message)

        self.openstr = lines.pop(0)
        self.closestr = lines.pop(-1)
        self.references = []
        self.coord_labels = []

        nweights = 0
        nref = 0

        tags = ('begin-reference','end-reference')
        seclines, lines = find_section(lines, tags)
        if not seclines or len(seclines) < 2:
            message = "REFERENCE-SECTION: No begin-reference -- "
            message += "end-reference block found"
            raise InputError(message)
        seclines.pop(-1)
        seclines.pop(0)
        startup = True
        for line in seclines:
            refs = line.split()
            self.coord_labels.append(refs.pop(0))
            try:
                refs = [float(re.sub('d','e',ref)) for ref in refs]
            except ValueError as V:
                message = "REFERENCE-SECTION: begin-reference: "
                message += "Coordinate labels must "
                message += "be followed by real numbers. "
                message += "Error for coordinate "
                message += self.coord_labels[-1] + "."
                message += "Wrong line reads: " + str(line)
                raise InputError(message)
            if startup:
                nref = len(refs)
                for ref in refs:
                    self.references.append([ref,])
                startup = False
            else:
                if len(refs) != nref:
                    message = "REFERENCE-SECTION: begin-reference: "
                    message += "Inconsistent number of references for "
                    message += "coordinate "+str(self.coord_labels[-1])
                    message += ". There are " + str(len(refs))
                    message += " entries but there should be "
                    message += str( nref) + "."
                    raise InputError(message)
                for n in range(nref):
                    self.references[n].append(refs[n])



        # key-value pairs
        key = "weights"
        strval, lines = find_param(key, lines)
        if strval:
            weights = strval.split(',')
            try:
                weights = [float(re.sub('d','e',w)) for w in weights]
                nweights = len(weights)
            except ValueError as V:
                message = "REFERENCE-SECTION: Keyword 'weights' must "
                message += "be followed by real number(s)."
                raise InputError(message)
            self.weights = tuple(weights)
        else:
            self.weights = None
            nweights = 0

        key = "switch-function"
        strval, lines = find_param(key, lines)
        if strval:
            self.switch_function = strval
        else:
            self.switch_function = None


        if len(lines) > 0:
            message = "REFERENCE-SECTION: found unknown keywords: "
            message += " ".join(lines)
            raise InputError(lines)


        if self.switch_function is not None and self.weights is not None:
            message = "REFERENCE-SECTION: Keywords 'switch-function' and "
            message += "'weights' must not be used together."
            raise InputError(message)

        # consistency checks
        if nref != nweights and not self.switch_function \
               and nweights != 0:
            message = "REFERENCE-SECTION: number of weights does not "
            message += "match the number of reference points: there are "
            message += str(nref) + " reference points but "
            message += str(nweights) + " weights."
            raise InputError(message)
        if nref == 1 and self.switch_function:
            message = "REFERENCE-SECTION: There is no need for a "
            message += "switching function if there is only one reference"
            message += " point."
            raise InputError(message)


        if self.weights is not None:
            sumw = 0.e0
            for w in self.weights:
                sumw += float(w)
            if abs(sumw - 1.e0) > 1.e-8:
                message = "REFERENCE-SECTION: weights: "
                message += "The sum of the weights must be one. "
                message += "Weights are: "
                wts = [str(w) for w in self.weights]
                message += ", ".join(wts)
                message += "; Sum = " + str(sumw) + "."
                raise InputError(message)

            tosmall = []
            for n, w in enumerate(self.weights):
                if w < 0.e0:
                    tosmall.append("No. " + str(n) + " weight=" + str(w))
            if len(tosmall) > 0:
                message = "REFERENCE-SECTION: weights: "
                message += "Weights must not be less then zero. "
                message += "Wrong weights: "
                message += ", ".join(tosmall) + '.'
                raise InputError(self.message)


    def __str__(self):
        string = str(self.openstr)
        if self.references:
            string += "\n begin-reference\n"
            for idx in range(len(self.coord_labels)):
                string += "  " + self.coord_labels[idx]
                for ref in range(len(self.references)):
                    try:
                        string += "  " + str(self.references[ref][idx])
                    except IndexError as I:
                        message = " Index of labels cannot access references: "
                        message += I.args[0]
                        raise IndexError(message)

                string += "\n"
            string += " end-reference\n"
        if self.weights is not None:
            string += " weights = "
            string += ', '.join([str(w) for w in self.weights]) + '\n'
        if self.switch_function is not None:
            string += ' switch-function = '
            string += str(self.switch_function) + "\n"
        string += self.closestr + '\n'
        return string


