#!/usr/bin/env python3

from minput import MSection, InputError
from parsingtools import *


class ClusterstatRunSection(MSection):

    def __init__(self, lines):
        MSection.__init__(self,lines)
        if lines[0].upper() != "RUN-SECTION":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. First line reads: " \
                     + lines[0]
            raise Exception(message)
        elif lines[-1].lower() != "end-run-section":
            message="Input seems not to contain a " \
                     + "RUN-SECTION. Last line reads: " \
                     + lines[-1]
            raise Exception(message)

        self.openstr = lines.pop(0)
        self.closestr = lines.pop(-1)

        # title first, as it matches a complete line
        # title may contain other keywords which could
        # confuse the imput parser later on
        strval, lines = find_title(lines)
        if strval:
            self.title = str(strval)
        else:
            self.title = None

        # obligatory key-value-pairs
        key = "name"
        strval, lines = find_param(key, lines)
        if strval:
            self.name = strval
        else:
            self.missing_key_error(key)

        key = "user-source"
        strval, lines = find_param(key, lines)
        if strval:
            self.user_source = strval
        else:
            self.missing_key_error(key)


        self.temperature_unit = None
        key = "temperature"
        strval, lines = find_param(key, lines)
        if strval:
            temp = strval.split(',')
            if len(temp) == 1:
                self.temperature = float(temp[0])
            elif len(temp) == 2:
                self.temperature = float(temp[0])
                self.temperature_unit = temp[1]
                unit = getunit(temp[1])
                if not unit:
                    message = "Temperature: Unknown unit: " + str(temp[1])
                    raise InputError(message)
            else:
                message = "Temperature: Too many keywords: " + str(temp)
                raise InputError(message)
        else:
            self.temperature = None


        key = "steps"
        strval, lines = find_param(key, lines)
        if strval:
            self.nsteps = int(strval)
        else:
            self.missing_key_error(key)

        key = "stepnorm"
        strval, lines = find_param(key, lines)
        if strval:
            self.stepnorm = float(strval)
        else:
            self.missing_key_error(key)

        # optional key-value-pairs

        key = "potential-routine"
        strval, lines = find_param(key, lines)
        if strval:
            self.potential_routine = str(strval)
        else:
            self.potential_routine = None

        key = "probe-routine"
        strval, lines = find_param(key, lines)
        if strval:
            self.probe_routine = str(strval)
        else:
            self.probe_routine = None


        key = "outunit"
        strval, lines = find_param(key, lines)
        if strval:
            self.outunit = str(strval)
        else:
            self.outunit = None

        key = "trajectory"
        strval, lines = find_param(key, lines)
        if strval:
            self.trajectorypath = str(strval)
        else:
            self.trajectorypath = None


        # flags
        self.loverwrite, lines = find_flag("overwrite", lines)
        self.lmean, lines = find_flag("mean", lines)
        self.lrms, lines = find_flag("rms", lines)
        self.lwalkonly, lines = find_flag("walk-only", lines)
        self.lignoreweights, lines = find_flag("ignore-weights", lines)

        if self.trajectorypath is None and self.temperature is None:
            message = "RUN-SECTION: either 'temperature' or 'trajectory'"
            message += " has to be set."
            raise InputError(message)
        if self.lwalkonly is True and self.trajectorypath is not None:
            message = "RUN-SECTION: 'walk-only' cannot be used with 'trajectory'"
            raise InputError(message)

        if len(lines):
            message = "Input containes unrecognized lines: \n"
            message += "\n".join(lines)
            raise InputError(message)


    def  __str__(self):
        string =  str(self.openstr)

        # key-value-pairs
        if self.title is not None:
            string += "\n title = " + str(self.title)
        if self.name is not None:
            string += "\n name = " + str(self.name)
        if self.temperature is not None:
            string += "\n temperature = " + str(self.temperature)
            if self.temperature_unit is not None:
                string += ", " + str(self.temperature_unit)
        if self.nsteps is not None:
            string += "\n steps = " + str(self.nsteps)
        if self.stepnorm is not None:
            string += "\n stepnorm = " + str(self.stepnorm)
        if self.outunit is not None:
            string += "\n outunit = " + str(self.outunit)
        if self.trajectorypath is not None:
            string += "\n trajectory = " + str(self.trajectorypath)
        if self.user_source is not None:
            string += "\n user-source = " + str(self.user_source)
        if self.potential_routine is not None:
            string += "\n potential-routine = " + str(self.potential_routine)
        if self.probe_routine is not None:
            string += "\n probe-routine = " + str(self.probe_routine)


        #flags
        if self.lwalkonly is True:
            string += "\n walk-only"
        if self.loverwrite is True:
            string += "\n overwrite"
        if self.lmean is True:
            string += "\n mean"
        if self.lrms is True:
            string += "\n rms"
        if self.lignoreweights is True:
            string += "\n ignore-weights"

        string += "\n" + self.closestr + "\n"
        return string

    def copy(self, other):
        if not isinstance(other,ClusterstatRunSection):
            message = 'ClusterstatRunSection.copy(other): other is not '
            message += 'an instance of RunSectionClu.'
            raise Exception(message)

        lines = str(other).split('\n')
        lines = preparse_input(lines)
        self.__init__(lines)




