#!/usr/bin/env python3
"""
Tools specific to cluster expansion
"""

import re
import sys
import os

from mtools import uniqifyList, getNonUnique, combIterator, incrAllInt
from parsingtools import str2int_tupel, absExpandPath




def chkClusterDef(cluster,modecombinations):
    """
    chkClusterDef(cluster,modecombinations) --> None

    Check ordering of modes and validity of labels.
    Raises ValueError if error is found.

    """


    nmodes = len(modecombinations)

    # check if the labels are valid
    for idx in cluster:
        mode = -1
        if idx != "*":
            try:
                mode = int(idx)
            except:
                for m,mc in enumerate(modecombinations):
                    if idx in mc:
                        mode = m
            if mode == -1:
                message = "Not a mode integer or a coordinate label: "
                message += str(idx) + "."
                raise ValueError(message)
            if mode >= nmodes:
                message = "Mode integers must be zero based, i.e., smaller "
                message += "than the number of modes. Found integer: "
                message += str(mode) + ", number of modes: "
                message += str(nmodes) + "."
                raise ValueError(message)



    tmpcl = [None,]*len(cluster)
    minmode = -1
    prevIsLabel = False
    isLabel = False
    seenLabels = {}
    for n,idx in enumerate(cluster):
        prevIsLabel = isLabel
        if idx == "*":
            mode = minmode +1
            isLabel = False
        else:
            try:
                mode = int(idx)
                isLabel = False
            except:
                for m,mc in enumerate(modecombinations):
                    if idx in mc:
                        mode = m
                        isLabel=True
                        if idx not in seenLabels:
                            seenLabels[idx] = True
                        else:
                            message = "Multiply defined label: " + str(idx)
                            message += " in cluster " + str(cluster) + "."
                            raise ValueError(message)

        if (mode <= minmode and not isLabel) or \
               (mode < minmode and isLabel and prevIsLabel):
            message = "Modes in in cluster definition must increase and "
            message += "modes integers and coordinate labels muts not mix "
            message += "for the same mode. "
            message += "Wrong cluster reads: " + str(cluster) + "."
            raise ValueError(message)
        else:
            minval = mode
        minmode = mode


def expandLastWildCard(cluster,modecombinations):
    """
    expandLastWildCard(cluster,modecombinations) --> list of clusters

    Expands the last wildcard in a tuple containing a cluster definition.

    cluster can be any sequence of increasing integers, coordinate labels or
    wildcard characters '*'. Integers are taken as zero based
    mode indices, i.e., integer larger or equal to the number of modes are
    not allowed.

    modecombinations must be a sequence of mode combinations. Each mode
    combination must be a sequence of coordinate labels, e.g.,
    modecombinations = (('Q1','Q2'),('Q3','Q4','Q5'),('Q6',),('Q7','Q8'))

    Examples:

    The cluster (1,'*') will be ecpanded into
    [(1,2), (1,3)]
    Meanung clusters containing coordinates  'Q3','Q4','Q5','Q6' and
    'Q3','Q4','Q5','Q7','Q8'

    The  cluster ('Q6','*') will be expanded into
    [('Q6',3)] meaning a cluster containing coordinates 'Q6','Q7','Q8'


    """

    #checks first
    chkClusterDef(cluster,modecombinations)

    nmodes = len(modecombinations)

    # first find index of last wildcard
    last = -1
    for n,idx in enumerate(cluster):
        if idx == "*":
            last = n
    if last == -1:
        return [cluster,]

    # determine the largest integer the
    # wildcard represents
    upperbound = -1
    if last == len(cluster)-1:
        upperbound = nmodes
    else:

        idx = -1
        try:
            idx = int(cluster[last+1])
        except:
            for n,mc in enumerate(modecombinations):
                if cluster[last+1] in mc:
                    idx = n
                    break
        if idx == -1:
            message = "Unable to expand wildcards in cluster definition. "
            message += "Error uccured during determination of upper bound "
            message += "of index " + str(last+1) + " of cluster "
            message += str(cluster) + ". The following index is not an "
            message += "integer nor a mode label."
            raise RuntimeError(message)
        else:
            upperbound = idx


    # determine the lowest integer the
    # wildcard represents
    lowerbound = -1
    if last == 0:
        lowerbound = 0
    else:
        wildcardcount = 0
        nonwildcardidx = -1
        for n in range(last):
            if cluster[n] == "*":
                wildcardcount += 1
            else:
                wildcardcount = 0
                nonwildcardidx = n

        if nonwildcardidx >= 0:
            idx = -1
            try:
                idx = int(cluster[nonwildcardidx])
            except:
                for n,mc in enumerate(modecombinations):
                    if cluster[nonwildcardidx] in mc:
                        idx = n
                        break
            if idx == -1:
                message = "Unable to expand wildcards in cluster definition. "
                message += "Error uccured during determination of lower bound "
                message += "of index " + str(last+1) + " of cluster "
                message += str(cluster) + ". one of the preceeding indices is "
                message += "not an integer nor a mode label."
                raise RuntimeError(message)

            lowerbound = idx + wildcardcount + 1

        else:
            lowerbound =  wildcardcount

    if lowerbound >= len(modecombinations):
        message = "Too many indices in Cluster " + str(cluster) + "."
        raise RuntimeError(message)
    if upperbound <= lowerbound:
        message = "Invalid Cluster " + str(cluster) + "."
        raise RuntimeError(message)




    clusters = []
    for i in range(lowerbound, upperbound):
        newcluster = [idx for idx in cluster]
        newcluster[last] = i
        clusters.append(tuple(newcluster))

    return clusters


def expandWildCards(cluster, modecombinations):

    """
    expandWildCards(cluster,modecombinations) --> list of clusters

    Expands the all wildcards in a tuple containing a cluster definition.

    cluster can be any sequence of increasing integers, coordinate labels or
    wildcard characters '*'. Integers are taken as zero based
    mode indices, i.e., integer larger or equal to the number of modes are
    not allowed.

    modecombinations must be a sequence of mode combinations. Each mode
    combination must be a sequence of coordinate labels, e.g.,
    modecombinations = (('Q1','Q2'),('Q3','Q4','Q5'),('Q6',),('Q7','Q8'))

    Examples (using the modecombinations above):

    The cluster (1,'*') will be ecpanded into
    [(1,2), (1,3)]
    Meanung clusters containing coordinates  'Q3','Q4','Q5','Q6' and
    'Q3','Q4','Q5','Q7','Q8'

    The  cluster ('Q6','*') will be expanded into
    [('Q6',3)] meaning a cluster containing coordinates 'Q6','Q7','Q8'

    the cluster ('*','*') will be expandet into
    [(0,1),(0,2),(0,3),(1,2),(1,3),(2,3)]

    expandWildCards recursively calls expandLastWildCard

    """

    clusters = expandLastWildCard(cluster, modecombinations)

    count = 0
    while "*" in clusters[0]:
        count += 1
        tmpclusters = []
        for cl in clusters:
            tmpclusters += expandLastWildCard(cl, modecombinations)

        clusters = tmpclusters

        if count > 30:
            message = "Maximim recursion exceeded."
            raise RuntimeError(message)

    clusters.sort()
    return clusters



def idx2doflist(idx,sysdof,modecombinations):
    """
    idx2doflist(idx,sysdof,modecombinations)
          --> list of integers with all DOF in (mode-)index

    idx:              A sequence containing the index of a cluster,
                      either mode integers or coordinate label
                      strings
    sysdof:           A sequence of all DOF labels in the system
    modecombinations: A sequence of sequences. the integer
                      sub-sequences define the mode combinations


    """

    doflist = []
    for i in idx:
        isint = True
        try:
            n = int(i)
        except:
            isint = False

        if isint:
            if n < 0:
                raise IndexError("Index must be >= 0")
            mci = [d for d in modecombinations[n]]
            for d in mci:
                doflist.append(d)
        else:
            if i in sysdof:
                doflist.append(i)
            else:
                message = "Label not found in system DOF: "
                message += str(i) + "."
                raise IndexError(message)

    return tuple(doflist)


def doftpl2modetpl(dof,modecombinations):
    """
    doftpl2modetpl(dof,modecombinations) --> integer tuple

    Return the modes (integer) a sequence of DOF (integer), belongs to.

    Example:
    dof = (0,1,5)
    mc = ((0,1),(2,3),(4,5))
    doftpl2modetpl(dof,mc) --> (0,2)
    """

    modes = []
    for n,mc in enumerate(modecombinations):
        for d in mc:
            if d in dof and not n in modes:
                modes.append(n)
    return tuple(modes)


def str_mc2int_mc(sysdof, modes):
    """
    str_mc2int_mc(sysdof, modes) --> list of integer tuples

    Maps a sequence of sequences of coordinate labels
    (mode combinations) into a list of integer tuples

    Raises ValueError if mapping fails.

    Example:

    sysdof = ('q1','q2','q3','q4','q5','q6')
    modes = (('q1','q2'),('q3','q5'),('q4','q6'))

    str_mc2int_mc(sysdof,modes) --> ((0,1),(2,4),(3,5))
    """


    mdof = []
    for mc in modes:
        for d in mc:
            mdof.append(d)

    doubles = getNonUnique(mdof)
    if len(doubles) != 0:
        message = "Labels in in modes not unique: " + str(doubles)
        raise ValueError(message)
    doubles = getNonUnique(sysdof)
    if len(doubles) != 0:
        message = "Labels in in sysdof not unique: " + str(doubles)
        raise ValueError(message)

    for d in mdof:
        if not d in sysdof:
            message = "Label " + str(d) + " not in sysdof."
            raise ValueError(message)
    for d in sysdof:
        if not d in mdof:
            message = "Label " + str(d) + " not in modes."
            raise ValueError(message)

    imc = [[None,]*len(mc) for mc in modes]
    for nmc, mc in enumerate(modes):
        for nd, d in enumerate(mc):
            imc[nmc][nd] = sysdof.index(d)
    return imc
