#!/usr/bin/env python3

import sys
import os
import logging
import pickle

from mcontainers import FileContainer
from mfilter import *
from parsingtools import absExpandPath


class JobError(RuntimeError):
    "class JobError(RuntimeError)"
    pass




class MJob(object):
    """
    Base class for python jobs of the MCTDH package.
    This is not meant to be instantiated.
    """

    def __init__(self):
        self._job_finished = False
        self.log = None
        self.outdir = None
        self.input = None
        self.name = ""

        # containers to group data types
        self.files = FileContainer() # empty file container

        self.loverwrite = False
        self.ldebug = False


    def __repr__(self):
        string = "Class MJob. Base Class for job objects."
        return string

    def __call__(self):
        self.execute()
        self.cleanUp()


    def logSetup(self, logfilename, append=False):
        "Basic logging setup for all jobs."
        logfilename = absExpandPath(logfilename)
        if os.path.exists(logfilename):
            if self.loverwrite is False and append is False:
                message = "Cannot open logfile " + str(logfilename) + ". "
                message += "Use option -w or set 'overwrite' in input file."
                raise JobError(message)

        # logfile goes extra and not into the file container
        self.log = logging.getLogger(logfilename)

        if self.ldebug:
            # Some more detailed info if we are in debug mode.
            loglevel=logging.DEBUG
            logformat='%(asctime)s %(levelname)-8s file= %(filename)-25s '+\
                       'line= %(lineno)-6s: %(message)s'
        else:
            # Just pass infos and errors if we are in normal mode.
            # No need to be detailed here.
            loglevel=logging.INFO
            logformat='%(asctime)s %(levelname)-8s : %(message)s'

        if append is True:
            mode = "a"
        else:
            mode = "w"

        # Basic logging config
        logging.basicConfig(
            level=loglevel,
            format=logformat,
            filename = logfilename,
            filemode = mode,
            datefmt = "[%d.%m.%Y %H:%M:%S]"
            )

        # Stream to stdout
        self.log_handler_stdout = logging.StreamHandler(sys.stdout)
        frm = logging.Formatter("%(message)s")
        self.log_handler_stdout.setFormatter(frm)
        if self.ldebug:
            self.log_handler_stdout.addFilter(OpenStream(logging.DEBUG,\
                                                         logging.INFO))
        else:
            self.log_handler_stdout.addFilter(OpenStream(logging.INFO,\
                                                         logging.INFO))
        self.log.addHandler(self.log_handler_stdout)

        # Stream to stderr
        self.log_handler_stderr = logging.StreamHandler(sys.stderr)
        frm = logging.Formatter("\n %(levelname)s: \n %(message)s\n")
        self.log_handler_stderr.setFormatter(frm)
        self.log_handler_stderr.addFilter(BlockStream(logging.DEBUG,\
                                                      logging.INFO))
        self.log.addHandler(self.log_handler_stderr)


    def mkNameDir(self,lmnd=False):
        """
        mknamedir(self,lmnd=False)
        Make name directory (-mnd option) if required.
        """
        lmnd = bool(lmnd)
        if not lmnd:
            return

        if self.outdir is None or self.outdir == "":
            raise JobError("Name directory is not specified.")

        if not os.path.exists(self.outdir) and not lmnd:
            namepath = os.path.jaoin(os.getcwd(), self.name)
            message = "Unable to create name directory: " + namepath
            message += " File o directory exists."
            message += " Use -mnd option"
            raise JobError(message)
        elif not os.path.exists(self.outdir) and lmnd:
            os.mkdir(self.outdir)
        elif os.path.exists(self.outdir) and os.path.isdir(self.outdir) \
                 and lmnd:
            message = "Name directory already exists:\n"
            message += str(self.outdir) + "\n"
            sys.stderr.write(message)
        elif os.path.exists(self.outdir) and not os.path.isdir(self.outdir) \
                 and lmnd:
            message = "Name directory already exists:\n"
            message += str(self.outdir) + "\n"
            message += "but is not a directory. Aborting."
            raise JobError(message)

    def checkNameDir(self):
        """
        check it the name directory exists and is a directory.
        raise JobError if not.
        """

        if self.outdir is None:
            raise JobError("Name directory is not specified.")

        if not os.path.exists(self.outdir):
            message = "Name directory " + str(self.outdir)
            message += " does not exist or could not be created."
            raise JobError(message)
        elif not os.path.isdir(self.outdir):
            message = "Name directory " + str(self.outdir)
            message += " does exist but is no directory."
            raise JobError(message)


    def saveInputCopy(self,filename="input"):
        """
        saveInputCopy(self,filename='input')

        Save a copy of input using the str() method for an inpot object.
        Needs self.outdir, self.overwrite and self.inp to be set.

        """
        if self.outdir is not None:
            outfilepath = os.path.join(self.outdir, filename)
        else:
            outfilepath = filename

        # save a copy of input (processed)
        if os.path.isfile(outfilepath) and self.loverwrite is not True:
            message = "Cannot open file " + outfilepath + ". Use "
            message += "option -w."
            raise JobError(message)

        f=open(outfilepath,mode="w")
        f.write("# INPUT FILE (COPY)\n\n")
        f.write(str(self.input))
        f.close()


    def wrRestart(self, filename="restart"):
        """
        Write the restart file to the name directory.
        This is done using pickle after detaching not
        picklable members (files, log). Files are reattached
        after picle is done.

        Note: This method should be overloaded in
        a derived type.
        """

        # these things cannot be pickled:
        log = self.log
        self.log = None
        log_handler_stdout = self.log_handler_stdout
        self.log_handler_stdout = None
        log_handler_stderr = self.log_handler_stderr
        self.log_handler_stderr = None
        files = self.files.detachAll()

        # now pickle
        f = open(os.path.join(self.outdir,filename), mode='wb')
        pickle.dump(self,f)
        f.close()

        # and restore
        self.log = log
        self.log_handler_stdout = log_handler_stdout
        self.log_handler_stderr = log_handler_stderr
        self.files.reAttach(files)


    def load(cls, filename="restart"):
        """
        load(filename="restart") -> job

        return a Job object from a pickled restart
        file.

        Note: This will usually need overloading in derived types.
        """
        f = open(filename, mode='rb')
        job = pickle.load(f)
        f.close()
        return job
    load = classmethod(load)



    # Dummy routines that sould be overloaded in
    # derived types.
    def doStep(self):
        "This is a dummy."
        pass


    def execute(self):
        "This is a dummy"
        pass


    def cleanUp(self):
        "This is a dummy"
        pass







