#!/usr/bin/env python3
"""
Mathematics tools
"""


import random
import numpy


##############################################################################
#       --- Sampling & Monte-Carlo functions ---
##############################################################################

def randomWalkStep(xvec, xmin=None, xmax=None,
                     stepweight=None, stepnorm=1.e0):
    """
    Make a random movement on a vector of real (float) coordinates

    input
        xvec - current state of the system in coordinate space
        min - minimum value for each coordinate
        max - maximum value for each coordinate
        stepweight - weight (for step size) for each coordinate
        stepnorm - length of random steps to be performed

    return
        xnext - next random position
    """
    lenx = len(xvec)
    rstep = numpy.random.random(lenx) - 0.5
    if not stepweight is None:
        rstep *= stepweight
    norm = (numpy.dot(rstep,rstep))**0.5
    rstep = (stepnorm/norm)*rstep
    xvecnew = xvec + rstep
    if not xmax is None:
        delta = xmax - xvecnew
        hit = (delta < 0)
        if hit.any():
            for i in range(lenx):
                if hit[i]: xvecnew[i] = xmax[i] + delta[i]

    if not xmin is None:
        delta = xmin - xvecnew
        hit = (delta > 0)
        if hit.any():
            for i in range(lenx):
                if hit[i]: xvecnew[i] = xmin[i] + delta[i]


    return xvecnew


def randomWalkStepGrid(ivec, imin, imax):
    """
    Make a random movement on a vector of integer coordinates
    (grid-points)

    input
        ivec - current index of the system
        min - minimum index for each coordinate
        max - maximum index for each coordinate

    return
        inext - next random gridpoint index
    """
    # new:
    lenr = float(len(ivec))
    icoord = numpy.array(numpy.random.random(1)*lenr,dtype=int)[0]
    idir = numpy.array(numpy.random.random(1)*2.e0,dtype=int)[0]*2-1

    ivecnew = ivec.copy()
    if ivecnew[icoord] + idir > imax[icoord] or ivecnew[icoord] + idir < imin[icoord]:
        idir = -idir

    ivecnew[icoord] += idir

    return ivecnew


def metropolisStep(xvec, pval, pfunc, pfuncargs=None,
                   xmin=None, xmax=None, stepweight=None, stepnorm=1.0):
    """
    Make a random movement, accept or reject according to metropolis alg.

    xvec - xvec state of the vector of integers (list, tuple or array)
    pval - probability value associated to the xvec position
    pfunc - maps the xvec position in N-Dim space to a probability
       pfuncargs - passed to pfunc, put pfunc arguments there

    return
       1. True if accepted, False if rejected
       2. accepted or old xvec position
       3. probability accepted point, pval retuned if not acc.
    """
    xnext = randomWalkStep(xvec,xmin=xmin,xmax=xmax,
                           stepweight=stepweight, stepnorm=stepnorm)
    if pfuncargs is None:
        pnext = pfunc(xnext)
    else:
        pnext = pfunc(xnext,pfuncargs)
    if pnext > pval:
        return True,xnext,pnext
    else:
        p = pnext/pval
        r = random.random()
        if r < p:
            return True,xnext,pnext
        else:
            return False,xvec,pval



def metropolisStepGrid(ivec, pval, pfunc,
                   imin, imax, fix=None,pfuncargs=None):
    """
    Make a random movement, accept or reject according to metropolis alg.

    ivec - integer vector of grid points (ND-Array)
    pval - probability value associated to the ivec position
    pfunc - maps the ivec position in N-Dim space to a probability
       pfuncargs - passed to pfunc, put pfunc arguments there
    imin - minimum index of gridpoints (ND-Array)
    imax - minimum index of gridpoints (ND-Array)
    fix  - grid points to fix (bool array)
           fix[i] = True -> do not change ivec[i]


    return
       1. True if accepted, False if rejected
       2. accepted or old xvec position
       3. probability accepted point, pval retuned if not acc.
    """

    # map on non-fixed

    if fix is not None:
        ivec1 = []
        imin1 = []
        imax1 = []
        for n,i in enumerate(ivec):
            if bool(fix[n]) is False:
                ivec1.append(ivec[n])
                imin1.append(imin[n])
                imax1.append(imax[n])
        ivec1 = numpy.array(ivec1,dtype=int)
        imin1 = numpy.array(imin1,dtype=int)
        imax1 = numpy.array(imax1,dtype=int)
    else:
        ivec1 = ivec
        imin1 = imin
        imax1 = imax

    inext = randomWalkStepGrid(ivec1,imin=imin1,imax=imax1)

    # map inext back
    if fix is not None:
        inext1 = []
        count = 0
        for n,i in enumerate(ivec):
            if bool(fix[n]) is False:
                inext1.append(inext[count])
                count += 1
            else:
                inext1.append(i)
        inext = numpy.array(inext1,dtype=int)


    if pfuncargs is None:
        pnext = pfunc(inext)
    else:
        pnext = pfunc(inext,pfuncargs)
    if pnext > pval:
        return True,inext,pnext
    else:
        p = pnext/pval
        r = random.random()
        if r <= p:
            return True,inext,pnext
        else:
            return False,ivec,pval
