#!/usr/bin/env python3
"""
Some general tools
"""

# standard modules
import os
import re
import sys
import imp

# third party modules
import numpy
import scipy



def importUserSource(sourcepath, objname=None):
    """
    Import an object from a user defined source file.

    Usage:
    importUserSource(sourcepath, objname=None) --> object

    "sourcepath" is the full path of a source file or a
    module/package name.
    If the source is a file name and the file already
    is in $PYTHONPATH or resides current directory a
    module/package name is sufficient.

    "objname" is the name of a Python object defined in the
    specified module or None. If None, the module as a
    whole is returnd.

    """
    sourcepath = str(sourcepath)
    sdir = os.path.split(sourcepath)[0]
    sfile = os.path.split(sourcepath)[1]

    # check if module is supposedly in the current dir
    if sfile == "":
        raise ImportError("Not a file or module name: " + str(sourcepath))

    # see if the module is already loaded


    obj = None
    n= len(sfile)
    if sfile.endswith(".so") or sfile.endswith(".py"):
        mname = sfile[0:n-3]
    elif sfile.endswith(".pyc"):
        mname = sfile[0:n-4]
    else:
        mname = sfile

    try:
        obj = sys.modules[nname]
    except:
        pass

    err = None
    if sourcepath.endswith(".so") and obj is None:
        try:
            obj = imp.load_dynamic(mname,sourcepath)
        except Exception as E:
            err = E
    elif sourcepath.endswith(".py") and obj is None:
        try:
            obj = imp.load_source(mname,sourcepath)
        except Exception as E:
            err = E
    elif sourcepath.endswith(".pyc") and obj is None:
        try:
            obj = imp.load_compiled(mname,sourcepath)
        except Exception as E:
            err = E
    elif obj is None:
        tpl = None
        try:
            tpl = imp.find_module(mname)
        except:
            pass
        if tpl is not None:
            try:
                obj = imp.load_module(mname, tpl[0], tpl[1], tpl[2])
            except:
                pass

    if obj is not None:

        objname = str(objname)
        if objname is not None:
            if objname in obj.__dict__:
                return obj.__dict__[objname]
            elif "." in objname:
                    subobjts = objname.split(".")
                    tmpobj = obj #call by ref
                    for subobj in subobjts:
                        if subobj in tmpobj.__dict__:
                            tmpobj =  tmpobj.__dict__[subobj]
                        else:
                            message = "Module '" + sourcepath
                            message += "' does not define an object called '"
                            message += objname + "'.\n"
                            message += "Objet NOT found: " +  subobj
                            raise ImportError(message)
                    return tmpobj
            else:
                message = "Module '" + sourcepath
                message += "' does not define an object called '"
                message += objname + "'."
                message += "Objets found: " + str(dir(obj))
                raise ImportError(message)
        else:
            return obj
    else:
        message = "Import of module failed: " + sourcepath + "."
        if err is not None:
            message += " Message: " + str(err)
        raise ImportError(message)



def checkexepaths(progs, mctdh_ver, suffix):
    """
    Check if executables are in $PATH. If not raise Exception.
    """

    prognames=[prog+str(mctdh_ver)+str(suffix) for prog in progs]
    paths=os.environ['PATH'].split(os.pathsep)

    for i in range(len(prognames)):
        lex=False
        for p in paths:
            if os.path.exists(os.path.join(p,prognames[i])): lex=True
        if not lex:
            raise Exception("Program "+prognames[i]+" is not in the $PATH")

    return prognames


def checkNumpyVersion(required):
    if numpy.__version__ < required:
        message = "\nThe NumPy version you are using is too old: "
        message += str(numpy.__version__) + ". "
        message += "\nPlease update at least to version "
        message += required + '. '
        message += "\nNumPy and SciPy can be obtained from "
        message += "http://www.scipy.org/."
        raise RuntimeError(message)


def isInPath(prog):
    """
    isInPath(prog) --> bool
    Check if a file called prog is in $PATH.
    """

    paths=os.environ['PATH'].split(os.pathsep)
    lex=False
    for p in paths:
        fname = os.path.join(p,prog)
        if os.path.exists(fname) and os.path.isfile(fname):
            lex=True
    return lex



def intersection(seq1, seq2):
    """
    intersection(seq1, seq2) -> seq

    Intersection of two sequences.

    Call by reference
    """

    isect = []
    for s in seq1:
        if s in seq2:
            isect.append(s)
    return isect


def diffseq(seq1,seq2):
    """
    diffseq(seq1,seq2) -> seq

    Return a sequence with elements wich are only
    in one of seq1 or seq2.

    Call by reference
    """

    seq = []
    for s in seq1:
        if not s in seq2:
            seq.append(s)
    for s in seq2:
        if not s in seq1:
            seq.append(s)
    return seq



def substractseq(seq1,seq2):
    """
    substractseq(seq1,seq2) -> seq

    Return a sequence containing elements
    of seq1 without those in seq2

    Call by reference
    """

    seq = []
    for s in seq1:
        if not s in seq2:
            seq.append(s)
    return seq


def uniqifyList(seq):
    "Remove dublicates from list. Order preserving, fast"
    seen = {}
    result = []
    for item in seq:
        if item in seen: continue
        seen[item] = 1
        result.append(item)
    return result


def getNonUnique(seq):
    "Return dublicates (only once) in a sequence."
    dublicates = []
    seen = {}
    for item in seq:
        if item in seen:
            if not item in dublicates:
                dublicates.append(item)
        else:
            seen[item] = 1
    return dublicates


def containesDoublicates(seq):
    "Determine, if there are dublicates in a sequence."

    seen = {}
    for item in seq:
        if item in seen:
            return True
        else:
            seen[item] = 1
    return False


def incrAllInt(seq):
    """
    Return a sequence with all integers in it incremented by one.
    """
    seq1 = list(seq) #make it a list
    for n,i in enumerate(seq1):
        if isinstance(i,int):
            seq1[n]+= 1
    return seq1


def tupleStr(seq):
    """
    Return a string representation of a sequence in tuple form
    but remove quotation marks
    """
    seq = tuple(seq) # make it a tuple
    s = str(seq)
    return re.sub("'","",s)


def intTupleStr(seq, strlen=None, ndigit=None,
                leftbound = True, sep=","):
    """
    incrTupleStr(seq, strlen=None, ndigit=None,
                 leftbound = True, sep=",") --> string

    Return a formated string representation of a sequence of integers.
    Note: This will fail for non-numbers and convert to integer for
          floats.

    Arguments:
    seq:         Sequence of integers
    strlen:      total number of characters (fill remeinng with space)
                 if string is too long return '*' times strlen
    ndigit:      format integer with n digits fixed
    leftbound:   Align text to left, right otherwise
    sep:         searator for tuple entries
    """


    string = "("
    if ndigit is not None:
        if ndigit < 1:
            raise ValueError("ndigit must not be < 1.")
        sdigit = str(ndigit)
    else:
        sdigit = ""
    for i in seq:
        try:
            string1 = "%" + sdigit + "d"
            string1 = string1 % int(i)
        except:
            string1 = str(i)
        if ndigit is not None and len(string1) > ndigit:
            string1 = "*" * ndigit
        string += string1
        string += sep

    if len(string) > 1:
        string = string[0:len(string)-len(sep)]
    string += ")"
    if strlen is None:
        return string
    if len(string) > strlen:
        string = "*"*strlen
    else:
        string1 = " "*(strlen - len(string))
        if leftbound:
            string += string1
        else:
            string = string1 + string

    return string



def getLastLine(f):
    """
    getLastLine(fileobj) --> string

    Emulate 'tail -n 1' on fileobj
    fileobj must be readable

    On exit, fileobj is positioned at EOF

    Note: There are better tools for Python 2.6 or later.

    """

    f.seek(0,2)
    filesize = f.tell()
    if filesize == 0:
        return ''
    if filesize == 1:
        f.seek(-1,2)
        return f.read(1)

    found = False
    pos = filesize - 2
    while not found:
        f.seek(pos)
        character = f.read(1)
        if character == "\n":
            found = True
        elif pos < 1:
            #there is no newline
            found = True
            pos = -1
        else:
            pos = pos - 1

    f.seek(pos+1)
    return f.readline()


def file2lines(filename, maxsize=1048576):
    """
    file2lines(filename, maxsize=1048576) --> list of lines

    Reads a file and retuns a list of lines. maxsize is the
    maximum size of the file in bytes. Default: 1 MB.

    Raises IOError if file size is > maxsize.
    """

    size=os.path.getsize(filename)
    if size > maxsize:
        message = "File " + str(filename) + " is too big: "
        message += "size = " + str(size) + "; "
        message += "maxsize = " + str(maxsize) + "."
        raise IOError(message)

    f = open(filename, mode="r")
    lines = f.readlines()
    f.close()
    return lines


def openFile(filename, xmode="r", loverwrite=False):
    "openFile(filename, xmode="r", loverwrite=False) --> file"

    if os.path.exists(filename) and not loverwrite \
           and xmode.startswith("w"):
        message = "No permission to overwrite file: "
        message += filename
        message += ". Set option for overwriting."
        raise RuntimeError(message)

    return open(filename, mode=xmode)



def buildSlices(tpl):
    """
    buildSlices(tpl) -> tuple of integers and slice objects

    Raises RuntimeError on error.

    tpl is a sequence of integers or 'None'. None will be translated
    into slice ':' while integer stays the same.

    Example:
    calling x[0,1,:,3] is equivalent to x[buildSlices((0,1,None,3))]
    What buildSlices((0,1,None,3)) actually returns is
    (0, 1, slice(None,None,None), 3)

    This routine is especially ment to build index tuples with slices
    for numpy arrays at runtime.

    No steps, ellypsis or ranges supported.
    """

    slc = [None,]*len(tpl)

    for n, idx in enumerate(tpl):
        if idx is None:
            slc[n] = slice(None, None, None)
        else:
            try:
                idx1 = int(idx)
            except:
                idx1 = None
            # check for error and if idx is
            # equivalent to integer
            if idx1 is not None and idx1 == idx:
                slc[n] = idx1
            else:
                message = "buildSlices expects integers "
                message += "and None's, not: '" + str(idx) + "'."
                raise RuntimeError(message)

    return tuple(slc)


def multiIndexIterator(start,nsteps,delta=1):
    """
    multiIndexIterator(start,nsteps,delta=1) -> Index

    Generator function for a multi-index counter.

    Usage:

    counter = multiIndexIterator(start,nsteps,delta=1)
    for index in counter:
       # do something with index

    On calling 'next()', the following index sequence is retuned.
    First index runs fastest.

    'start' is a sequence of integer start values, 'nsteps' is a sequence of
    integer values defining the number of steps to take for each index.
    'delta' can be either integer or a sequence
    of integers and defines the incrementation.
    All sequences must be of same length.

    if 'nsteps[i]' == 0 or delta[i] == 0 the index i will de skipped
    four counting.

    """

    # make local copies and convert to int if necessary
    isiterable = lambda obj: getattr(obj, '__iter__', False)
    if not isiterable(delta):
        delta = [delta,]*len(start)
    else:
        delta = [int(d) for d in delta]

    start = [int(s) for s in start]
    nsteps = [int(n) for n in nsteps]

    if len(start) != len(nsteps):
        message = "'start' and 'nsteps' must be sequences of same length. "
        message += "len(start) = " + str(len(start)) + ", "
        message += "len(nsteps) = " + str(len(nsteps)) + "."
        raise RuntimeError(message)


    if len(start) != len(delta):
        message = "'start', 'nsteps' and 'delta' must be sequences of "
        message += "same length. "
        message += "len(start) = " + str(len(start)) + ", "
        message += "len(bsteps) = " + str(nsteps(end)) + "."
        message += "len(delta) = " + str(len(delta)) + "."
        raise RuntimeError(message)

    for n in nsteps:
        if n < 0:
            message = "number of steps cennot be negative."
            raise RuntimeError(message)

    idx = [s for s in start]
    lidx = len(idx)
    const = [nsteps[i]==0 or delta[i] == 0 for i in range(lidx)]
    for i in range(lidx):
        if const[i]:
            nsteps[i] = 0
            delta[i] = 0

    totsteps = 1
    lzerosteps = True
    for n in range(lidx):
        if nsteps[n] != 0:
            totsteps *= nsteps[n]
            lzerosteps = False

    if lzerosteps:
        message = "Not iterable: all steps are zero."
        raise IndexError(message)

    count = [0,]*lidx


    yield idx


    for ts in range(totsteps-1):
        ic = const.index(False)
        idx[ic] += delta[ic]
        count[ic] += 1
        for n in range(lidx):
            for j in range(n,lidx-1):
                if count[j] == nsteps[j] and nsteps[j] != 0:
                    jc = const[j+1:].index(False) + j + 1
                    idx[j] = start[j]
                    idx[jc] += delta[jc]
                    count[j] = 0
                    count[jc] += 1
                else:
                    break
        yield idx



def combIterator(N,k):
    """
    combIterator(N,k) --> integer tuples

    Iterator to return all combinations of N things
    taking k at the same time.

    Usage:
    counter = combIterator(N,k)
    for index in counter:
       # do something with index

    Requires scipy.misc.comb

    Note: In newer Python versions this can be replaced
          with itertools.combinations
    """

    if k > N:
        yield []

    idx = range(k)
    stop = range(N-k,N)

    totcount = scipy.misc.comb(N,k,exact=1)

    for i in range(totcount):
        yield idx
        if i == totcount - 1:
            raise StopIteration()
        idx[-1] += 1
        for i in range(k-1):
            if idx[-i - 1] == stop[-i - 1] + 1:
                idx[-i - 2] += 1
                for j in range(i+1):
                    idx[-i - 1 + j] = idx[-i - 2 ] + 1 + j


