#!/usr/bin/env python3
"""
Tools used for parsing input files.
"""

# standard modules
import re
import os

import numpy

def build_re(param):
    pattern=[ int(s.isalpha()) *('['+s+s.swapcase()+']')+ \
              (1-int(s.isalpha()))*(s) for s in param]

    return ''.join(pattern)


def preparse_input(lines):
    " Concentrate strings and remove empty lines"

    lines1 = []

    for line in lines:

        pat=("#.*$","^[ \t]*$","[ \t]*\\n",'^[ \t]*','[ \t]*$')

        for i in range(len(pat)):
            line= re.subn(pat[i],'',line)[0]

        line = line.lstrip().rstrip()
        pat=('[ \t]*,[ \t]*',
             '[ \t]*=[ \t]*',
             '[ \t][ \t]*',
             '^[ \t]*',
             '[ \t]\)',
             '\([ \t]')

        rep=(',','=',' ','',')','(')

        for i in range(len(pat)):
            line= re.subn(pat[i],rep[i],line)[0]

        if line:
            lines1.append(line)

    return lines1


def find_section(lines, tags):
    """
    findsec(lines,tags) -> list1, list2
    Returns list1 of lines between tags[0] and tags[1] in lines.
    list2 contains orignal lines without list1
    """

    startstr=tags[0].upper()
    closestr=tags[1].upper()

    startline=-1
    endline=-1
    for j in range(len(lines)):
        if re.search("^"+startstr+'[ \t]*$',lines[j].upper()):
            startline=j
            break


    for j in range(len(lines))[startline:]:
        if re.search("^"+closestr+'[ \t]*$',lines[j].upper()):
            endline=j+1
            break


    if startline>endline \
           or startline==endline \
           or startline<0 \
           or endline<0:
        return None, lines

    lines1 = lines[startline:endline]
    lines2 = lines[0:startline] + lines[endline:]

    return lines1, lines2


def chop_end(lines, endline="end-input"):
    """
    chop_end(lines, endline="end-input") --> lines
    If existing, remove lines after endline="end-input".
    """
    idx = None
    try:
        idx = lines.index(endline)
    except ValueError as V:
        # catch the "not in list" error
        pass
    if idx:
        return lines[:idx+1]
    else:
        return lines


def str2tupel(string, separator=","):
    """
    str2tupel(string) --> tuple
    Turns string '(a, b, c, ...)' into tuple('a', 'b', 'c', ...)
    """
    if string == "":
        raise ValueError("Empty string cannot be turned into a tuple.")
    if string == "()":
        return tuple()
    lenstring = len(string)
    if string[0] != "(" or  string[lenstring-1] != ")":
        message="No parenthesis: String is no tuple: " + string
        raise ValueError(message)
    string = string[1:lenstring - 1]
    slist = string.split(separator)
    return tuple(slist)


def str2int_tupel(string, separator=","):
    """
    str2int_tupel(string) --> tuple
    Turns string '(1, 2, 3, ...)' into tuple(1, 2, 3, ...)
    """
    if string == "":
        raise ValueError("Empty string cannot be turned into a tuple.")
    if string == "()":
        return tuple()
    stupel = str2tupel(string, separator)
    ilist = []
    for i in stupel:
        isint = True
        try:
            n = int(i)
        except:
            isint = False
        if isint:
            ilist.append(n)
        else:
            ilist.append(i)
    return tuple(ilist)


def str2float_tupel(string, separator=","):
    """
    str2float_tupel(string) --> tuple
    Turns string '(1.0, 2.0, 3.0,...)' into tuple(1.0, 2.0, 3.0, ...)
    """
    if string == "":
        raise ValueError("Empty string cannot be turned into a tuple.")
    if string == "()":
        return tuple()
    stupel = self._str2tupel(string,separator)
    flist = []
    for f in stupel:
        isfloat = True
        try:
            d = float(f)
        except:
            isfloat = False
        if isfloat:
            flist.append(d)
        else:
            flist.append(f)
    return tuple(flist)


def find_title(lines):
    """ find_title(lines) -> title, lines
    return title, lines if string "title = some title" is contained in lines
    "title = some title" is removed from lines.
    If title is not in lines, return None, lines
        """

    pattern = build_re('title')
    pattern += "=.*"
    findpat = re.compile(pattern)

    title = None
    lines1 = []
    for line in lines:
        m = findpat.match(line)
        if m:
            title = m.group().split("=")[1]
            line = re.sub(pattern,'',line)
        lines1.append(line.lstrip().rstrip())
    lines = [line for line in lines1 if line]

    return title, lines


def find_param(param, lines):
    """
    find_param(param, liness) -> val, lines

    Return val, lines if string "param = val" is contained in lines
    param = val is removed from lines.
    If param is not in line, return None, lines
    """

    pattern = build_re(param)
    pattern += "=\S+"
    findpat = re.compile(pattern)

    parval = None
    lines1 = []
    for line in lines:
        m = findpat.match(line)
        if m:
            parval = m.group().split("=")[1]
            line = re.sub(pattern,'',line)
        lines1.append(line.lstrip().rstrip())
    lines = [line for line in lines1 if line != ""]

    return parval, lines

def find_flag(flag, lines):
    """
    find_flag(param, line) -> bool, lines

    Return True, lines if string "flag" is contained in lines
    flag is remofed from lines.
    If param is not in line, return False, lines
    Use with care: this is context-independent.
    """

    lowflag = flag.lower()
    hasflag = False
    for j in range(len(lines)):
        words_lower = lines[j].lower().split()
        words_orig = lines[j].split()
        if lowflag in words_lower:
            hasflag = True
            words_orig.pop(words_lower.index(lowflag))
            lines[j] = ' '.join(words_orig)

    lines = [line for line in lines if line != ""]
    return hasflag, lines


def getunit(string):
    """
    getunit(string) --> float or None

    Returns the conversion factor to atomic units
    Returns None if unit is not recognized.

    string can be one of
    'cm-1'     -->   4.556333828E-06
    'eV'       -->   3.67493088248E-2
    'meV'      -->   3.67493088248E-5
    'au'       -->   1.E0
    'kcal/mol' -->   1.5936145299079E-3
    'kJ/mol'   -->   3.80880861452E-4

    """
    string = preparse_input([string,])
    string = string[0].lower()

    if string == "cm-1":
        return numpy.float64(4.556333828E-06)

    elif string == "ev":
        return numpy.float64(3.67493088248E-2)

    elif string == "mev":
        return numpy.float64(3.67493088248E-5)

    elif string == "au":
        return numpy.float64(1.E0)

    elif string == "kcal/mol":
        return  numpy.float64(1.5936145299079E-3)

    elif string == "kj/mol":
        return numpy.float64(3.80880861452E-4)

    else:
        return None


def absExpandPath(path = ""):
    """
    absExpandPath(path = "") --> path

    Expands user homes, variables etc. in a path and returns the
    absolute path. No checks if path exists. If path
    is empty on enter, the current directory is returned.
    """

    if not path:
        return os.getcwd()

    path = os.path.expanduser(path)
    path = os.path.expandvars(path)
    path = os.path.abspath(path)
    return path


def int2bin(n, count=30, reverse=False):
    """
    int2bin(n, count=30) --> string
    returns the binary code of integer n, using count number of digits
    """
    if n < 0:
        raise ValueError("n must be a positive or zero integer")
    if count < 1:
        raise ValueError("count < 1 makes no sense.")
    if n > 2**count:
        raise RuntimeError("n > 2^count: String is too short.")
    if reverse:
        return "".join([str((n >> y) & 1) for y in range(count)])
    else:
        return "".join([str((n >> y) & 1) for y in range(count-1, -1, -1)])
