"The Krotov OCT algorithm"


#########################################################################
#
# FUNCTIONS IN THIS MODULE
#
# - anabw                     Analysis after backward propagation
# - anafw                     Analysis after forward propagation
# - backward                  Perform a backward propagation
# - cleanup                   A dummy
# - forward                   Perform a forward propagation
# - mkinputfiles              Create all temporary input files
# - mkkrotov                  Generator function for the algorithm
# - parseinput                Parse input files
# - restartinfo               Collect data for restart runs
# - startup                   Create initial fields, states and perform
#                             the first forward propagation
#
# 04/2009 MS
#########################################################################

from octtools import *

def parseinput(jobopts,inp,op):
    "Extract information from input files"

    import os,re
    from messages import log
    from octtools import chksections, parserundefault
    from octtools import parsetardefault, setdefaultfiles

    log.info("parsing input")
    # check if all needed sections are there
    log.debug('Checking that all sections are there.')
    # Exception handling in calling module
    chksections(inp)

    if inp.sections.has_key('filtersection'):
        lfiltersection=True
    else:
        lfiltersection=False

    jobopts['lfilter'] = False

    parserundefault(inp,jobopts)
    parsetardefault(inp,jobopts)
    IsMoreToParse = parseoctdefault(inp,jobopts)

    if IsMoreToParse:
        message="Unknown keywords found in OCT-SECTION:\n"
        message+=str(inp['octsection'])
        raise Exception(message)

    if jobopts['lfilter'] and not jobopts['lwrrawflds']:
        message='OCT-SECTION: filtering requires'
        message+=' "write=rawfield" or "write=all"'
        message+=' if a FILTER-SECTION is present in .inp file.'
        raise Exception(message)


    #reference function
    if jobopts['lreffnc']:
        message="The reference-function within the Krotov-algorithm "
        message+="is pre-defined as the previous field. Remove keyword "
        message+="'reference' from OCT-SECTION.\n\n"
        raise Exception(message)


    setdefaultfiles(jobopts)
    os.chdir(jobopts['outdir'])


##############################################################################
# Create temporary input files
# Same as in defailtoct.py, except that siprop option is not implemented
##############################################################################
def mkinputfiles(jobopts,inp,op):
    "Write temporary files, assign filenames, create temporary directories"

    import mctdhinp, os, re
    from messages import log


    #most is the same as in default algorithm
    from defaultoct import  mkinputfiles as mkinputdefault
    mkinputdefault(jobopts,inp,op)


############################ FILTER INPUT FILES #############################
    if jobopts['lfilter']:
        sec=mctdhinp.mctdhsection()
        sec.copy( inp.sections['filtersection'] )
        inp.delsection('filtersection')

        sec.openstr='begin-fldop'
        sec.closestr='end-fldop'

        message=str(sec)
        dumpfile('filterf.inp',message)
        jobopts['filterf']='filterf.inp'
        del sec


##############################################################################

def startup(jobopts,inp,op):
    """Generate fields, initial and target states,
    do first forward propagarion"""

    import shutil, os
    from messages import log


    # Write the header of the jfunc file
    if jobopts['lfilter']:
        message=\
            """
# Control functional: J1=population, J2=field-gradient (negative), J3=filter-part (signed)
# no.       J1                    J2                J3                J1+J2-ABS(J3)
"""
    else:
        message=\
            """
# Control functional: J1=population, J2=field-gradient (negative)
# no.       J1                    J2                J1+J2
"""

    dumpfile(jobopts['jfile'],message,jobopts['overwrite'],'w')

    # prepare overlaps file
    message= "# Overlaps with target state or operator. tfinal = "+\
        str(jobopts['tfinal'])+' tout = '+str(jobopts['tout'])+'\n'
    message+="# Format:  No.,  ( Re, Im,  Abs**2, i=1, npacket), total_target_population\n"
    dumpfile(jobopts['overlaps'],message,jobopts['overwrite'],'w')

    #for the field generator (uses au)
    fs=float(41.3414)
    tfinal=jobopts['tfinal']*fs
    tout  =jobopts['tout']*fs

    log.info("Generating initial functions:")

    #guess field
    if jobopts['lguessfnc']:
        # A guess function has been given
        genfldcmd=jobopts['genfld_exe']+\
            ' -a 2 -O '+jobopts['operfile']+\
            ' -w '*jobopts['overwrite']+\
            ' -o '+outpath('fieldf.dat')+\
            ' 0.0 '+ str(tfinal)+' '+str(tout)+' '+jobopts['guessfnc']

        log.info("Generating Guess-field: "+jobopts['guessfnc'])
        log.debug('Command line: '+genfldcmd)
        status=os.system(genfldcmd)
        if status:
            message= "genfld returned a nonzero exit code while generating the guess field."
            log.warn(message)
	
        os.rename('genfld.log', 'genfld_'+jobopts['guessfnc']+'.log')

        #reverse guess field
        genfldcmd=jobopts['genfld_exe']+\
            ' -a 2 -O '+jobopts['operfile']+\
            ' -R '+\
            ' -w '*jobopts['overwrite']+\
            ' -o '+outpath('fieldb.dat')+\
            ' 0.0 '+ str(tfinal)+' '+str(tout)+' '+jobopts['guessfnc']

        log.info("Generating Guess-field: "+jobopts['guessfnc'] + " (backwards)")
        log.debug('Command line: '+genfldcmd)
        status=os.system(genfldcmd)
        if status:
            message= "genfld returned a nonzero exit code while generating the guess field."
            log.warn(message)

        if jobopts['lsvflds']:
            shutil.copyfile(outpath('fieldf.dat'), outpath('fieldf.0'))

        os.rename('genfld.log', 'genfld_'+jobopts['guessfnc']+'b.log')

    else:
        # No guess field was given, use zero
        genfldcmd=jobopts['genfld_exe']+\
            ' -a 2 -O '+jobopts['operfile']+\
            ' -w '*jobopts['overwrite']+\
            ' -o '+outpath('fieldf.dat')+\
            ' 0.0 '+ str(tfinal)+' '+str(tout)+' zero'

        log.info("Generating zero guess-field:")
        log.debug('Command line: '+genfldcmd)
        status=os.system(genfldcmd)
        if status:
            message= "genfld returned a nonzero exit code while generating the guess field."
            log.warn(message)
        shutil.copyfile(outpath('fieldf.dat'), outpath('fieldb.dat'))


    if jobopts['lsimprop']:
        # copy some files is we are in parallel mode
        shutil.copyfile(outpath('fieldf.dat'), outpath('fieldf_targ.dat'))
        shutil.move(outpath('fieldf.dat'), outpath('fieldf_init.dat'))
        shutil.copyfile(outpath('fieldb.dat'), outpath('fieldb_targ.dat'))
        shutil.move(outpath('fieldb.dat'), outpath('fieldb_init.dat'))


    #reference funcion (zero for the first propagation)
    genfldcmd=jobopts['genfld_exe']+\
        ' -a 2 '+\
        ' -w '*jobopts['overwrite']+\
        ' -o '+outpath('reffnc.dat')+\
        ' 0.0 '+ str(tfinal)+' '+str(tout)+' zero'
    log.debug('Command line: '+genfldcmd)
    status=os.system(genfldcmd)
    if status:
        message="genfld returned a nonzero exit code while generating"+\
                 " the first reference function (zero)."
        log.warn(message)
    # krotov: there is reference function by definition
    jobopts['lreffnc']=True


   #first lagrange function (zero)
    if jobopts['lfilter']:
        genfldcmd=jobopts['genfld_exe']+\
            ' -w '*jobopts['overwrite']+\
            ' -o '+outpath('lagrfnc.dat')+\
            ' 0.0 '+ str(tfinal)+' '+str(tout)+' zero'

        log.debug('Generating the very first Lagrange function (zero)')
        log.debug('Command line: '+genfldcmd)
        status=os.system(genfldcmd)
        if status:
            message="genfld returned a nonzero exit code while generating"+\
                     " the first Lagrange function (zero)."
            log.warn(message)


    #penalty function (envelope)
    if jobopts['lpnltyfnc']:
        log.info("Generating Penalty-function: "+jobopts['pnltyfnc']  )
        genfldcmd=jobopts['genfld_exe']+\
        ' -a 2 -O '+jobopts['operfile']+\
        ' -w '*jobopts['overwrite']+\
        ' -o '+outpath('pnltyfnc.dat')+\
        ' 0.0 '+ str(tfinal)+' '+str(tout)+' '+jobopts['pnltyfnc']
        log.debug('Command line: '+genfldcmd)
        status=os.system(genfldcmd)
        if status:
            message="genfld returned a nonzero exit code while generating"+\
                " the penalty function."
            log.warn(message)

        os.rename('genfld.log', 'genfld_'+jobopts['pnltyfnc']+'.log')


    del genfldcmd


    log.info("Generating initial state")
    # generate initial state
    mctdhcmd=jobopts['quantics_exe']+' '+int(jobopts['overwrite'])*\
        ' -w '+ jobopts['initstate_inp']
    log.debug('Command line: '+mctdhcmd )
    status=os.system(mctdhcmd)
    if status:
        message= "QUANTICS returned a nonzero exit code while"+\
            " generating the initial state"
        log.warn(message)

    # generate target state
    if not jobopts['ltaroper']:
        log.info("Generating target state")
        log.debug('Command line: '+mctdhcmd )
        mctdhcmd=jobopts['quantics_exe']+' '+int(jobopts['overwrite'])*\
            ' -w '+ jobopts['targstate_inp']
        status=os.system(mctdhcmd)
        if status:
            message= "QUANTICS returned a nonzero exit code while"+\
                " generating the target state"
            log.warn(message)

    # first forward propagation
    log.info("Performing the first forward propagation with guessed field.")
    mctdhcmd=jobopts['quantics_exe']+' '+int(jobopts['overwrite'])*\
        ' -w '+ jobopts['forward1_inp']
    log.debug('Command line: '+mctdhcmd )
    status=os.system(mctdhcmd)
    if status:
        message= "QUANTICS returned a nonzero exit code while "+\
            "performing the first forward propagation"
        log.error(message)
        raise Exception(message)
    jobopts['curriter']=0

    log.info("Setup is complete.")

    message='jobopts after setup has completed:\n'
    for k in jobopts.keys(): message+= str(k)+' = '+str(jobopts[k])+'\n'
    dumpfile(outpath('jobopts'),message,jobopts['overwrite'])

##############################################################################
##############################################################################




def forward(jobopts):
    """Perform the forward propagation"""
    from messages import log
    from mctdhinp import cleanline
    import os, re, subprocess,shutil

    if os.path.exists(outpath('stop')):
        log.info('Found stop-file.')
        jobopts['continue']=False
        os.remove(outpath('stop'))
        return

    log.info('Iteration '+ str(jobopts['curriter'])+' (forward)')


    #operate fldop on the raw field (produce the lagrangien)
    if jobopts['lfilter']:
        args=( jobopts['fldop_exe'],
               '-f ' +os.path.join(jobopts['outdir'],'rawfieldb.dat'),
               ' -o '+os.path.join(jobopts['outdir'],'lagrfnc.dat'),
               '-O '+jobopts['operfile'],
               '-C '*jobopts['lpassfilter'],
               '-R ',
               '-w',
               jobopts['filterf'] )
        oargs =[ str(arg) for arg in args if arg]
        cmd=re.sub('[,\'\[\]]',' ',str(oargs))
        cmd=cleanline(cmd)
        log.debug("Calculating Lagrange multiplier:")
        log.debug("Command line: "+ cmd )
        status=os.system(cmd)
        if status:
            message= "fldop returned a nonzero exit code while"+\
                " operating on rawfieldb.dat."
            raise Exception(message)



    # operate fldop on the field:
    # produce the filtered reference function if required
    if jobopts['lreffilter'] and jobopts['lfilter']:

        if jobopts['lsimprop']:
            shutil.copyfile(outpath('fieldf_init.dat'), outpath('fieldf.dat'))

        args=( jobopts['fldop_exe'],
               '-f ' +os.path.join(jobopts['outdir'],'fieldf.dat'),
               ' -o '+os.path.join(jobopts['outdir'],'pref.dat'),
               '-O '+jobopts['operfile'],
               '-C '*(not jobopts['lpassfilter']),
               '-w',
               jobopts['filterf'])

        oargs =[ str(arg) for arg in args if arg]
        cmd=re.sub('[,\'\[\]]',' ',str(oargs))
        cmd=cleanline(cmd)
        log.debug("Calculating reference function:")
        log.debug("Command line: "+ cmd )
        status=os.system(cmd)
        if status:
            message= "fldop returned a nonzero exit code while"+\
                " operating the filter."
            raise Exception(message)

        shutil.copyfile(outpath('pref.dat'), outpath('reffnc.dat'))


    # operate fldop on the field:
    # produce the filtered reference function for lagrange multiplyer
    # to be used in anafw later
        if jobopts['lreffilter']:
            args=( jobopts['fldop_exe'],
                   '-f ' +os.path.join(jobopts['outdir'],'fieldf.dat'),
                   ' -o '+os.path.join(jobopts['outdir'],'qref.dat'),
                   '-O '+jobopts['operfile'],
                   '-C '*jobopts['lpassfilter'],
                   '-w',
                   jobopts['filterf'])

            oargs =[ str(arg) for arg in args if arg]
            cmd=re.sub('[,\'\[\]]',' ',str(oargs))
            cmd=cleanline(cmd)
            log.debug("Calculating reference function:")
            log.debug("Command line: "+ cmd )
            status=os.system(cmd)
            if status:
                message= "fldop returned a nonzero exit code while"+\
                    " operating the filter."
                raise Exception(message)
        else:
            #set qref=0
            genfldcmd=jobopts['genfld_exe']+\
                ' -a 2 -O '+jobopts['operfile']+\
                ' -w '*jobopts['overwrite']+\
                ' -o '+outpath('fieldf.dat')+\
                ' 0.0 '+ str(tfinal)+' '+str(tout)+' zero'

            log.debug('Generating qref.dat (zero)')
            log.debug('Command line: '+genfldcmd)
            status=os.system(genfldcmd)
            if status:
                message= "genfld returned a nonzero exit code while generating the qref.dat (zero)."
                log.warn(message)


    else:
        if jobopts['lsimprop']:
            shutil.copyfile(outpath('fieldf_init.dat'), outpath('reffnc.dat'))
        else:
            shutil.copyfile(outpath('fieldf.dat'), outpath('reffnc.dat'))


    if jobopts.has_key('taroper'):
        taroper=jobopts['taroper']+' '
    else:
        taroper=''

    # start efield in the background
    args=( jobopts['efield_exe'],
           '-pc'*jobopts['lpc'],
           '-F',str(jobopts['penalty-factor']),
           '-S'*jobopts['lpnltyfnc'],
           '-R',
           '-O',jobopts['dipoleoperator'],
           '-T'*jobopts['ltaroper'],taroper*jobopts['ltaroper'],
           '-i',str(jobopts['curriter']),
           '-p'*jobopts['lsimprop'],
           '-A'*jobopts['lalignphs'],
           '-d'*jobopts['ltdoverlap'],
           '-r'*jobopts['lwrrawflds'],
           '-m'*jobopts['lwrtddipmtx'],
           '-o'*jobopts['lwrtdoverlap'],
           '-t'*jobopts['lwrtdtargetpop'],
           '-L'*jobopts['lfilter'],
           '-K')

    eargs =[ str(arg) for arg in args if arg]
    cmd=re.sub('[,\'\[\]]',' ',str(eargs))
    cmd=cleanline(cmd)
    log.debug("Starting efield in background:")
    log.debug("Command line: "+ cmd )

    # BTW: avoid spaces in arguments list, Popen, spawnv etc. might fail
    # for os.system spaces seem to be OK
    efield_proc=subprocess.Popen(eargs)


    if jobopts['lsimprop']:
       #start one instance of mctdh
       args=( jobopts['quantics_exe'], 
              '-w', 
              os.path.split(jobopts['forward_init_inp'])[1] ) #
       margs1 =[ str(arg) for arg in args if arg]

       #just debug-logging
       cmd=re.sub('[,\'\[\]]',' ',str(margs1))
       cmd=cleanline(cmd)
       log.debug("Starting first QUANTICS")
       log.debug("Command line: "+ cmd )
       
       mctdh_proc1=subprocess.Popen(margs1)

       #start another instance of mctdh
       args=( jobopts['quantics_exe'], 
              '-w', 
              os.path.split(jobopts['forward_targ_inp'])[1] ) 

       margs =[ str(arg) for arg in args if arg]

       #just debug-logging
       cmd=re.sub('[,\'\[\]]',' ',str(margs))
       cmd=cleanline(cmd)
       log.debug("Starting second QUANTICS:")
       log.debug("Command line: "+ cmd )

       mctdh_proc=subprocess.Popen(margs)

    else:

        #start one instance of mctdh
        args=( jobopts['quantics_exe'], 
               '-w', 
               os.path.split(jobopts['forwardi_inp'])[1] ) 
        
        margs =[ str(arg) for arg in args if arg]

        #just debug-logging
        cmd=re.sub('[,\'\[\]]',' ',str(margs))
        cmd=cleanline(cmd)
        log.debug("Starting QUANTICS")
        log.debug("Command line: "+cmd )
        
        mctdh_proc=subprocess.Popen(margs)

    # whatch the child processes. kill them if anything goes wrong
    if jobopts['lsimprop']:
        whatchprogs(efield_proc,mctdh_proc,mctdh_proc1)
    else:
        whatchprogs(efield_proc,mctdh_proc)





##############################################################################


def anafw(jobopts):
    "Analyse after forward propagation"
    from messages import log
    from mctdheingabe import cleanline
    import shutil, time, re, subprocess, os


    #load overlaps
    log.debug('Reading overlaps file')
    f=ope('overlaps','r')
    lastline=f.readlines()[-1]
    f.close()
    del f
    lastline=cleanline(lastline)
    J1=float(lastline.split(' ')[-1])

    #the field part is done using FORTRAN
    if jobopts['curriter'] == 0:
        #there is no reference field and
        #no lagrange multiplicator during
        #the very first propagation
        J2=0.0
        J3=0.0
    else:

        if jobopts['lsimprop']:
            shutil.copyfile(outpath('fieldf_init.dat'), outpath('fieldf.dat'))


        if jobopts['lfilter']:
            args=( jobopts['fldop_exe'],
                   '-f ' +os.path.join(jobopts['outdir'],'fieldf.dat'),
                   ' -o '+os.path.join(jobopts['outdir'],'qfld.dat'),
                   '-O '+jobopts['operfile'],
                   '-C '*(jobopts['lpassfilter']),
                   '-w',
                   jobopts['filterf'])

            oargs =[ str(arg) for arg in args if arg]
            cmd=re.sub('[,\'\[\]]',' ',str(oargs))
            cmd=cleanline(cmd)
            log.debug("Operating filter on field:")
            log.debug("Command line: "+ cmd )
            status=os.system(cmd)
            if status:
                message= "fldop returned a nonzero exit code while"+\
                         " operating the filter."
                raise Exception(message)



        args=( jobopts['jfunc_exe'],
               '-S'*jobopts['lpnltyfnc'],
               '-F',str(jobopts['penalty-factor']),
               '-L'*(jobopts['lfilter'] or jobopts['lreffilter'] ),
               '-R')

        args =[ str(arg) for arg in args if arg]

        cmd=re.sub('[,\'\[\]]',' ',str(args))
        cmd=cleanline(cmd)
        log.debug("Starting jfunc:")
        log.debug("Command line: "+ cmd )

   	args =[ str(arg) for arg in args if arg]

        cmd=re.sub('[,\'\[\]]',' ',str(args))
        cmd=cleanline(cmd)
        log.debug("Starting jfunc:")
        log.debug("Command line: "+ cmd )

        #start jfunc
        jfunc_proc=subprocess.Popen(args, stdout=subprocess.PIPE)

        returncode=jfunc_proc.poll()
        while returncode is None:
            time.sleep(1) # sleep a second if not done yet
            returncode=jfunc_proc.poll()


        data=jfunc_proc.communicate()
        if returncode != 0:
            message="Error while calculating control functional.\n"
            message+=str(jobopts['jfunc_exe'])+ ' returned a non-zero exit code.'
            message+='Message: \n'+ str(data)
            raise Exception(message)

        #extract data (first line in pipe)
        data=cleanline(data[0]).split(' ')
        try:
            J2=float(data[0])
            J3=float(data[1])
        except:
            message="Error while calculating control functional.\n"
            message+=str(jobopts['jfunc_exe'])+ ' returned non-numbers:\n'
            message+=str(data)
            raise Exception(message)

    log.debug('Target populaton= '+str(J1))
    log.debug('Field gradient= '+str(-J2))
    log.debug('Filter penalty= '+str(-abs(J3)))
    if jobopts['curriter']==0:
        log.info('Target population obtained with guess-field: '+str(J1))
    else:
        log.info('Total cost-functional J= '+str(J1-J2-abs(J3)))

    if jobopts['lfilter']:
        line=str(jobopts['curriter'])+'   '+floatstr(J1) +'   '+floatstr(-J2)+'   '+floatstr(J3)+\
            '   '+floatstr(J1-J2-abs(J3))+' \n'
    else:
        line=str(jobopts['curriter'])+'   '+floatstr(J1) +'   '+floatstr(-J2)+\
            '   '+floatstr(J1-J2)+' \n'

    if jobopts['curriter'] == 0: line=str(jobopts['curriter'])+'   '+floatstr(J1)+' \n'

    dumpfile(jobopts['jfile'],line,True,'a')
    del line

    #save the output
    saveoutputf(jobopts)


    if jobopts['curriter']==jobopts['niterations']: jobopts['continue']=False




##############################################################################
def anabw(jobopts):
    "Analyse after forward propagation"

    #save the output
    saveoutputb(jobopts)

    # incrment after backward analysis
    jobopts['curriter']+=1

##############################################################################
def backward(jobopts):

    import os,re,time,subprocess,shutil
    from messages import log
    from mctdhinp import cleanline

    if jobopts.has_key('skipbw'):
        jobopts.pop('skipbw')
        return

    if os.path.exists(outpath('stop')):
        log.info('Found stop-file.')
        jobopts['continue']=False
        os.remove(outpath('stop'))
        return

    log.info('Iteration '+ str(jobopts['curriter'])+' (backward)')


    if jobopts.has_key('taroper'):
        taroper=jobopts['taroper']+' '
    else:
        taroper=''

    # start efield in the background
    args=( jobopts['efield_exe'],
#           '-pc'*jobopts['lpc'],  #useless
           '-b',
           '-F',str(jobopts['penalty-factor']),
           '-S'*jobopts['lpnltyfnc'],
           '-O',jobopts['dipoleoperator'],
           '-T'*jobopts['ltaroper'],taroper*jobopts['ltaroper'],
           '-i',str(jobopts['curriter']),
           '-p'*jobopts['lsimprop'],
           '-A'*jobopts['lalignphs'],
           '-d'*jobopts['ltdoverlap'],
           '-r'*jobopts['lwrrawflds'],
           '-m'*jobopts['lwrtddipmtx'],
           '-o'*jobopts['lwrtdoverlap'],
           '-L','-K')

    eargs =[ str(arg) for arg in args if arg]

    #just debug-logging
    cmd=re.sub('[,\'\[\]]',' ',str(eargs))
    cmd=cleanline(cmd)
    log.debug("Starting efield:")
    log.debug("Command line: "+ cmd )

    efield_proc=subprocess.Popen(eargs)


    if jobopts['lsimprop']:
        
        #start one instance of mctdh 
        args=( jobopts['quantics_exe'], 
               '-w',
               os.path.split(jobopts['backward_targ_inp'])[1])
        margs1 =[ str(arg) for arg in args if arg]

        #just debug-logging
        cmd=re.sub('[,\'\[\]]',' ',str(margs1))
        cmd=cleanline(cmd)
        log.debug("Starting QUANTICS in background:")
        log.debug("Command line: "+ cmd )

        mctdh_proc1=subprocess.Popen(margs1)

        #start another instance of mctdh 
        args=( jobopts['quantics_exe'], 
               '-w',
               os.path.split(jobopts['backward_init_inp'])[1])

        margs =[ str(arg) for arg in args if arg]

        #just debug-logging
        cmd=re.sub('[,\'\[\]]',' ',str(margs))
        cmd=cleanline(cmd)
        log.debug("Starting QUANTICS:")
        log.debug("Command line: "+ cmd )

        mctdh_proc=subprocess.Popen(margs)

    else:
      
        args=( jobopts['quantics_exe'], 
               '-w',
               os.path.split(jobopts['backwardi_inp'])[1])
        margs =[ str(arg) for arg in args if arg]

        cmd=re.sub('[,\'\[\]]',' ',str(margs))
        cmd=cleanline(cmd)
        log.debug("Starting QUANTICS:")
        log.debug("Command line: "+ cmd )

        mctdh_proc=subprocess.Popen(margs)

    # whatch the child processes. kill them if anything goes wrong
    if jobopts['lsimprop']:
        whatchprogs(efield_proc,mctdh_proc,mctdh_proc1)
    else:
        whatchprogs(efield_proc,mctdh_proc)

    if jobopts['curriter'] == 0: anafw(jobopts)

##############################################################################

def cleanup(jobopts):
    pass

##############################################################################


def restartinfo(jobopts,inp,op):
    "Gather information for restarting an optimal control job."

    #this is the same as in the rabitz scheme
    from defaultoct import restartinfo as defaultrestartinfo
    defaultrestartinfo(jobopts,inp,op)
    if jobopts['lfilter']:
        jobopts['filterf']='filterf.inp'

##############################################################################


def mkkrotov(jobopts,inp,op):
    "Select the functions for propagation and analysis"

    import shutil

    parseinput(jobopts,inp,op)

    if not jobopts['restartb'] and not jobopts['restartf']:

        if jobopts['lclear']:
            clearoldoutput(jobopts['outdir'])

        mkinputfiles(jobopts,inp,op)
        startup(jobopts,inp,op)
    else:
        restartinfo(jobopts,inp,op)
        shutil.copyfile(outpath('fieldf.dat'), outpath('reffnc.dat'))
        # for anafw
        jobopts['lreffnc']=True

    return (forward, anafw, backward, anabw, cleanup)
