"""
Eingabe routines for mctdhinp
"""

#########################################################################
#
# FUNCTIONS IN THIS MODULE
#
# - build_cire                Buird a case-independent regular expression
# - cleanline                 Remove unused parts from a string
# - concntrt                  Concentrate whitespaces
# - einfilter                 Read the FILTER-SECTION
# - eininitwf                 Read the INIT_WF-SECTION
# - einintegrator             Read the INTEGRATOR-SECTION
# - einoct                    Read the OCT-SECTION
# - einoper                   Read the OPER-SECTION
# - einpbasis                 Read the PBASIS-SECTION
# - einrun                    Read the RUN-SECTION
# - einsection                Read a general section
# - einspf                    Read the SBASIS-SECTION
# - eintarget                 Read the TARGET-SECTION
# - findparam                 Find a key=value pair
# - findsec                   Find a section
# - secnames                  Extract names (e.g. of Hamiltonians)
# - stripblkcmts              Remove comments
#
# To do additional parsing of sections modify the specialized
# functions. All, except einrun, are based on function einsection.
# einsection provides automatic key=value extraction and
# seaches for boolean flags. More complex structures remain inparsed.
#
# 04/2009 MS
#########################################################################


__all__=['build_cire', 'eininitwf', 'einintegrator', 'einoct',
         'einoper', 'einpbasis', 'einrun', 'einspf', 'einsection',
         'einfilter','eintarget', 'secnames','cleanline', 'concntrt',
         'findparam', 'findsec', 'stripblkcmts']

##############################################################################
# Some tools first
##############################################################################

def build_cire(str):
    """
    build_cire(str) -> str1:
    Build a regular expression string str1 that finds str
    regardles of upper or lower case letters in str
    """
    pattern=[ int(s.isalpha()) *('['+s+s.swapcase()+']')+ \
                    (1-int(s.isalpha()))*(s) for s in str]
    #evaluates * 0 or 1 ----^
    return ''.join(pattern)


##############################################################################


def stripblkcmts(line):
    """
    stripblkcmts(string) -> str
    - Remove comments of form '# commenttext' from end of str
    - Make str only containing whitespaces an empty string
    - remove all '\n' escape characters
    - remove leading whitespaces
    """
    import re
    pat=("#.*$","^[ \t]*$","[ \t]*\\n",'^[ \t]*','[ \t]*$')
    for i in range(len(pat)): line= re.subn(pat[i],'',line)[0]
    return line

##############################################################################

def concntrt(line):
    """
    concntrt(string) -> str
    - concentrate comma separated text
    (remove blanks, e.g.: "a, b , c" -> "a,b,c")
    - concentrate assignments
    (remove blanks, e.g.: "a  =  b" -> "a=b")
    - make multiple blanks a single space
    """
    import re

    pat=('[ \t]*,[ \t]*',
         '[ \t]*=[ \t]*',
         '[ \t][ \t]*',
        '^[ \t]*' )

    rep=(',','=',' ','')


    for i in range(len(pat)): line= re.subn(pat[i],rep[i],line)[0]
    return line


##############################################################################


def cleanline(line):
    """
    cleanline(string) -> str
    - Remove comments of form '# commenttext' from end of str
    - Make str only containing whitespaces an empty string
    - remove all '\\n' escape characters
    - concentrate comma separated text
      (remove blanks, e.g.: "a, b , c" -> "a,b,c")
    - concentrate assignments
      (remove blanks, e.g.: "a  =  b" -> "a=b")
    - make multiple blanks a single space
    """
    line=stripblkcmts(line)
    line=concntrt(line)

    return line

##############################################################################

def findsec(startstr,closestr,lines):
    """
    findsec(startstr,closestr,lines) -> list[]

    Returns a list of lines between startstr and closestr in lines
    list[] does not include the lines containing startstr and closestr
    """

    startstr=startstr.upper()
    closestr=closestr.upper()
    import re
    startline=-1
    endline=-1
    for j in range(len(lines)):
        if re.search("^"+startstr+'.*$',lines[j].upper()):
            startline=j+1
            break


    for j in range(len(lines))[startline:]:
        if re.search("^"+closestr+'.*$',lines[j].upper()):
            endline=j
            break


    if startline>endline or startline==endline or startline<0 or endline<0:
        return None

    return lines[startline:endline] #return section

##############################################################################

def findparam(param, line):
    """ findparam(param, line) -> val,line
    return val,line if string "param=val" is contained in line
    remove param=val from line. If param is not in line, return None,line"""
    import re

    pattern=build_cire(param)

    pattern+="=.*"
    findpat=re.compile(pattern)
    m=findpat.match(line)
    if m:
        parval = m.group().split("=")[1]
        line=re.sub(pattern,'',line)
        return (parval,line)

    return (None,line)


#############################################################################
# Input routines for sections
##############################################################################

##############################################################################

# RUN-SECTION is sort of special - it gets its own routine:
# Except from the title, there are only flags and key=value
# pairs in a RUN-SECTION. So we go by formal patterns instead
# of keylists.
def einrun(f):
    """ Read a run-section from file and return a section object"""
    import re
    from mctdhinp import mctdhsection

    startstr='RUN-SECTION'
    closestr='END-RUN-SECTION'

    #create a section object
    sec=mctdhsection(startstr,closestr)


    f.seek(0)
    #read lines and do some cleanup (remove comments etc)
    lines=map( cleanline , f.readlines())
    # throw away empty lines
    lines=[line for line in lines if line]

    # see if we can find the section
    lines=findsec(startstr,closestr,lines)
    if lines is None:
        # Section not found
        return None

    #find title (given as rest of line)
    findttl=re.compile(build_cire('title')+"[ \t]*=")
    for j in range(len(lines)):
        m=findttl.match(lines[j])
        if m:
            sec["title"]= lines[j][m.end():]
            lines[j]=re.sub(build_cire('title')+'.*$','',lines[j])
            break
    #title is an entire line
    findttl=re.compile( build_cire('title') )
    for j in range(len(lines)):
        m=findttl.match(lines[j])
        if m:
            sec['title']=lines[j+1]
            lines[j]=re.sub(build_cire('title'),'',lines[j])
            lines.pop(j+1)
            break

    #find parameters (structures like: a=b,c,d.... )
    findpar=re.compile("[\S]*=[\S\.\,]*")

    for j in range(len(lines)):
        m=findpar.findall(lines[j])
        if m:
            for i in range(len(m)):
                li=m[i].split("=")
                sec[li[0].lower()]=li[1]
                lines[j]=re.sub(m[i],'',lines[j])
    del findpar, m

    #everything else is flags
    str=" ".join(lines)
    str=cleanline(str)
    flags=str.split(' ')
    for i in range(len(flags)): sec.setflag(flags[i].lower())

    return sec


##############################################################################

# all sections wich are not RUN-SECTION are wrapped

def einoper(f):
    """ Read a oper-section from file and return a section object"""

    startstr='OPERATOR-SECTION'
    closestr='END-OPERATOR-SECTION'

    keylist=("opname","oppath")

    return einsection(f,startstr,closestr,keylist)


##############################################################################


def einspf(f):
    """ Read a SPF-section from file and return a section object"""

    startstr='SPF-BASIS-SECTION'
    closestr='END-SPF-BASIS-SECTION'

    keys=('packets',)
    flags=('multi-set',"single-set")

    sec=einsection(f,startstr,closestr,keys,flags)

    if sec is None:

        startstr='SBASIS-SECTION'
        closestr='END-SPF-BASIS-SECTION'

        sec=einsection(f,startstr,closestr,keys,flags)

    if sec is None:

        startstr='SPF-BASIS-SECTION'
        closestr='END-SBASIS-SECTION'

        sec=einsection(f,startstr,closestr,keys,flags)


    if sec is None:

        startstr='SBASIS-SECTION'
        closestr='END-SBASIS-SECTION'

        sec=einsection(f,startstr,closestr,keys,flags)

    if 'multi-set' in sec.flags and "single-set" in sec.flags:
        message = startstr + ": either 'multi-set' or 'single-set' "
        message += "may be used. Not both."
        raise RuntimeError(message)

    if len(sec.data) < 2:
        message = "\n\n"
        message += "Found only one mode in " + sec.openstr + ".\n"
        message += "'optcntrl' does not (yet) support for one-mode\n"
        message += "problems. As a workaround you may add a dummy\n"
        message += "mode in an eigenstate of its single-particle\n"
        message += "Hamiltonian.\n"
        raise RuntimeError(message)


    return sec

##############################################################################

def einpbasis(f):
    """ Read a SPF-section from file and return a section object"""

    startstr='PBASIS-SECTION'
    closestr='END-PBASIS-SECTION'

    sec= einsection(f,startstr,closestr)

    if sec is None:
        startstr='PRIMITIVE-BASIS-SECTION'
        closestr='END-PBASIS-SECTION'

        sec= einsection(f,startstr,closestr)

    if sec is None:
        startstr='PBASIS-SECTION'
        closestr='END-PRIMITIVE-BASIS-SECTION'

        sec= einsection(f,startstr,closestr)

    if sec is None:
        startstr='PRIMITIVE-BASIS-SECTION'
        closestr='END-PRIMITIVE-BASIS-SECTION'

        sec= einsection(f,startstr,closestr)


    return sec



##############################################################################

def einoct(f):
    """ Read a oct-section from file and return a section object"""

    startstr='OCT-SECTION'
    closestr='END-OCT-SECTION'

    keylist=( "algorithm",'functype',"iterations",
              "operator", "guess","penalty-function",
              "penalty-factor","reference","write","save",
              "filter")

    flaglist=('tdoverlap','alignphs','no-reffilter','filter')

    return einsection(f,startstr,closestr,keylist,flaglist)




##############################################################################

def einintegrator(f):
    """ Read a integrator-section from file and return a section object"""

    startstr='INTEGRATOR-SECTION'
    closestr='END-INTEGRATOR-SECTION'

    return einsection(f,startstr,closestr)


##############################################################################

def eintarget(f):
    """ Read a target-section from file and return a section object"""

    startstr='TARGET-SECTION'
    closestr='END-TARGET-SECTION'

    keylist=("operator",)
    return einsection(f,startstr,closestr,keylist)


##############################################################################
def eininitwf(f):
    """ Read a init_wf-section from file and return a section object"""

    startstr='INIT_WF-SECTION'
    closestr='END-INIT_WF-SECTION'

    return einsection(f,startstr,closestr)

##############################################################################
def einfilter(f):
    """ Read a init_wf-section from file and return a section object"""

    startstr='FILTER-SECTION'
    closestr='END-FILTER-SECTION'

    return einsection(f,startstr,closestr)


##############################################################################
##############################################################################

def einsection(f,startstr,closestr, keys=[],flags=[],readrest=True):
    """
    einsection(f,startstr,closestr,keys=[],flags=[],readrest=True) -> section

    - Create section object with startstr,closestr
    - reads file f, searches for textblock between lines startstr
      and closestr.
    - If textblock not found, return None
      Otherwise:
    - Extracts parameters of form key=value if
      key is in keys and sets section[key]=value.
    - Removes key=value from textblock.
    - Looks if the textblock contains the word
      flags[i], and calls section.setflag[flags[i]] if the word is
      in the text
    - removes flags[i] from textblock.
    - Reads the textblock into section.data if readrest is True
    - Returns section
    """
    import re
    from mctdhinp import mctdhsection

    sec=mctdhsection(startstr,closestr)

    f.seek(0)
    lines=f.readlines()

    #if we need to process, do the basic string conversions
    if len(keys)>0:
        lines=map( cleanline ,lines)
    else:
        #otherwise remove only comments and whitespaces
        #(data goes into 'data')
        lines=map( stripblkcmts ,lines)

    #remove empty lines
    lines=[line for line in lines if line]

    # see if we can find our section
    lines=findsec(startstr,closestr,lines)
    if lines is None:
        return None


    #read parameters list
    for j in range(len(lines)):
        for i in range(len(keys)):
            val,lines[j]=findparam(keys[i], lines[j])
            if val :
                sec[keys[i].lower()]=val
                break

    #read flags ist
    for j in range(len(lines)):
        for i in range(len(flags)):
            words= lines[j].lower().split(' ')
            WORDS= lines[j].split(' ')
            if flags[i].lower() in words:
                sec.setflag(flags[i].lower())
                WORDS.pop(words.index(flags[i].lower()) )
                lines[j]=' '.join(WORDS)

    # cleaning up
    lines=[line.lstrip() for line in lines if line.lstrip()]
    del line

    #everything else is 'data'
    if readrest: map(sec.adddataitem,lines)

    # done! section is ready
    return sec



##############################################################################


def secnames(f,secname,delim='_'):
    """
    secname(f,secname,delim) -> list
    Returns a list of names of sections in file f where sections are labeled
    "section_name", e.g, the line "HAMILTONIAN-SECTION_dipole" will result in
    ['dipole'] if secname is "HAMILTONIAN-SECTION" and delim is "_"
    """
    if f.closed: f.open()
    f.seek(0)
    lines=map( cleanline , f.readlines())

    import re
    pat = "^[ \t]*"+build_cire(secname)+delim+"[a-zA-Z0-9]+[ \t]*$"


    newlines = [ line.split(delim)[1] for line in lines if re.match(pat,line)]

    return newlines






##############################################################################
