"""
Module to wrap input data in classes mctdhsection, mctdhinput and mctdhoper

Sections have the following structure:
-- They are enclosed in an opening string and a closing string, e.g.
   "RUN-SECTION" and "END-RUN-SECTION"
-- They may contain parameters: variable names and assigned value, e.g.,
   "tfinal=100". These are stored in a dictionary 'parameters'
-- Boolean Flags may be set, e.g., "auto" and "geninwf"
   This is stored as in the 'flags' list
-- There may as well be other data of a different type, e.g.,
   specification of the primitive grid, the integrator etc.
   for which there is no need to parse it at this point
   this is just treated as stringsdata and is not processed
   (except concentration of blanks)

mctdhinput contains a dictionary with sections imported from file.

mctdhoper contains three dictionaries, one for sections, one for
   Hamiltonians and one for functions

No exception handling in this module. Exceptions have to be catched
in calling module

"""

#########################################################################
#
# CLASSES IN THIS MODULE
#
# - mctdhsection              Wrapping of a section object
# - mctdhinput                Wrapping of a complete input file
# - mctdhoper                 Wrapping of a complete operator file
#
# 04/2009 MS
#########################################################################




__all__=['mctdhsection', 'mctdhinput', 'mctdhoper']


##############################################################################


class mctdhsection:
    """ Definition of a section object to wrap all
    XXX-SECTION patterns of MCTDH imput files"""

    def __init__(self,ostr="",cstr=""):
        self.parameters={}    # all data of form "key=value1[,value2[,...]]"
        self.flags=[]         # flags that evaluate to True of False
        self.openstr=ostr     # opening 'tag' of a text section
        self.closestr=cstr    # closing 'tag' of a text section
        self.data=[]          # any other data


    def __str__(self):
        """ Return a string representation of section """
        strng=self.openstr+"\n"
        if len(self.parameters):
            strng+=" \n".join([" %s = %s " % (key,value) for key, value \
                                   in self.parameters.items() ])
            strng+=" \n"



        flagslen=len(self.flags)
        for i in range(flagslen):
            strng+=" "+str(self.flags[i])
            if (i+1)%3==0: strng+="\n"

        if (flagslen)%3!=0: strng+="\n"
        if flagslen: strng+=" \n"
        del flagslen


        datalen=len(self.data)
        for i in range(datalen):
            strng+=" "+str(self.data[i])+"\n"
        del datalen

        strng+=self.closestr+'\n'

        return strng

    def __setitem__(self,key,item):
        """ Set a key-value-pair in parameters{} """
        self.parameters[key]=item

    def __getitem__(self,key):
        """ Return value of key in parameters{} """
        if self.parameters.has_key(key):
            return self.parameters[key]
        else:
            return None

    def setparam(self,key,item):
        """ Set a key-value-pair in parameters{} """
        self.parameters[key]=item

    def getparam(self,key):
        """ Return value of key in parameters{} """
        return self.parameters[key]

    def delparam(self,key):
        """ Return value of key in parameters{} """
        if self.parameters.has_key(key):
            return self.parameters.pop(key)

    def hasparam(self,key):
        """ Return true if key in parameters{} """
        return self.parameters.has_key(key)

    def setflag(self,flag=''):
        """ Set a flag in flags[] """
        if not self.hasflag(flag):
            self.flags.append(flag)


    def getflag(self,index):
        """ get a flag from flags[] by index """
        return self.flags[index]


    def hasflag(self,flag=''):
        """ Check for flag being in flags[] """
        return self.flags.__contains__(flag)


    def delflag(self,flag=''):
        """ If flags[] contains flag """
        if  self.flags.__contains__(flag):
            self.flags.remove(flag)
        else:
            return


    def adddataitem(self,data):
        self.data.append(data)

    def deldataitem(self,item=''):
        if  self.data__contains__(item):
            self.data.remove(item)
        else:
            return
		
    def popdataitem(self,index):
        return self.data.pop(index)

    def isempty(self):
        if len(self.parameters) or len(self.flags) or len(self.data):
            return False
        else:
            return True


    def copy(self,other):
        """ Copy a section object """
        self.clear()
        self.openstr=other.openstr
        self.closestr=other.closestr
        self.parameters=other.parameters.copy()
        self.flags=[]
        for i in range(len(other.flags)):
            self.flags.append(other.flags[i])
        self.data=[]
        for i in range(len(other.data)):
            self.data.append(other.data[i])


    def clear(self):
        """ Clear flags parameters and data,
        set openstr and closestr to empty strings """
        self.parameters.clear()
        self.flags=[]
        self.data=[]
        self.openstr=''
        self.closestr=''

    def clearcontent(self):
        """ Clear flags parameters and data,
        but leave openstr and closestr unchanged """
        self.parameters.clear()
        self.flags=[]
        self.data=[]




from mctdheingabe import *

class mctdhinput:
    """ Wrapping of an MCTDH input file. class MCTDHInput
    contains a dictionary holding the sections of an input file """

# this could be extended with other input sections
    secimportmap={"runsection":        einrun,
                  "opersection":       einoper,
                  "octsection":        einoct,
                  "pbassection":       einpbasis,
                  "spfbassection":     einspf,
                  "initwfsection":     eininitwf,
                  "intsection":        einintegrator,
                  "targetsection":     eintarget,
                  "filtersection":     einfilter}


    def __init__(self,filename=''):
      self.sections={}
      if filename:
          self.importfile(filename)

    def __str__(self):
        " retuns a string representing an MCTDH input file "

        strng="## MCTDH input file:\n\n"
        for key in self.sections.keys():
            strng+=str(self.sections[key])
            strng+="\n"
        strng+="end-input\n "
        return strng

    def __setitem__(self,key,item):
        """ Set a key-value-pair in parameters{} """
        self.sections[key]=item

    def __getitem__(self,key):
        """ Return value of key in parameters{} """
        return self.sections[key]

    def hasseckey(self,secname,seckey):
        """
        hasseckey(secname,seckey) -> bool
        Return True if section secname exists and contains key seckey.
        Return False otherwise
        """
        if self.sections.has_key(secname):
            return self.sections[secname].parameters.has_key(seckey)
        else:
            return False


    def hassecflag(self,secname,flag):
        """
        hassecflag(secname,secflag) -> bool
        Return True if section secname exists and contains flag flag.
        Return False otherwise
        """
        if self.sections.has_key(secname):
            try:
                i=self.sections[secname].flags.index(flag)
            except ValueError:
                return False
            return True

        else:
            return False


    def copy(self,other):
        "Copy other instance to self"
        self.clear()
        for name in other.sections.keys():
            self.sections[name]=mctdhsection()
            self.sections[name].copy(other.sections[name])


    def clear(self):
        "Delete all data"
        self.sections.clear()

    def addsection(self,name,sec):
        """
        addsection(secname,section)
        Copy section and add the copy with name secname
        to the sections dictionary
        """
        s=mctdhsection()
        s.copy(sec)
        self.sections[name]=s

    def delsection(self,name):
        """
        delsection(secname)
        If exists: delete the section with name secname
        from the sections dictionary. Otherwise do nothing.
        """
        if self.sections.has_key(name): return self.sections.pop(name)


    def sectionnames(self):
        """Return a list of all section names"""
        return self.sections.keys()


    def importfile(self,filename):
        """ importfile(filename): Import input from file filename"""

        f=open(filename,'r')
        for name,importfunc in self.secimportmap.items():
            sec=importfunc(f)
            if sec: self[name]=sec


    def exportfile(self,filename,overwrite=True,mode='w'):
        """
        ExportFile(filename,owerwrite=True,mode='w'):
        Save input to ASCII file.
        """
        import os
        if os.path.exists(filename) and not overwrite:
            message="\nNo permition to overwrite file: "+filename+'\n'
            message+="Try setting the overwrite flag."
            raise Exception(message)
        else:
            f=open(filename,mode)
            f.write(str(self))
            f.close()
            #catch possible ecxeption in calling module

##############################################################################

class mctdhoper(mctdhinput):
    """ Wrapping of an MCTDH oper file. class mctdhoper
    contains a dictionaries holding the sections, Hamiltonians
    and functions of an operator file """

    #              name            --------------------arguments-------------------  function
    secimportmap={"opdefine"    : ("OP_DEFINE-SECTION","END-OP_DEFINE-SECTION",
                                   einsection),
                  "parameter"   : ("PARAMETER-SECTION","END-PARAMETER-SECTION",
                                   einsection),
                  "dissipative" : ("DISSIPATIVE-SECTION","END-DISSIPATIVE-SECTION",
                                   einsection),
                  "labels"      : ("LABELS-SECTION","END-LABELS-SECTION",
                                   einsection),
                  }


    def __init__(self,filename=''):
        self.sections={}  # Sections
        self.operators={} # hamiltonians
        self.functions={} # functions as used sometimes for OCT-MCTDH
        if filename:
            self.importfile(filename)


    def __str__(self):
        " retuns a tring representing an MCTDH operator file "

        strng="## MCTDH operator file:\n\n"
        for key in self.sections.keys():
            strng+=str(self.sections[key])+"\n"
            strng+="\n\n"

        # write system operator first
        syskey = "system"
        if syskey in self.operators.keys():
            strng+=str(self.operators[syskey])+"\n"
            strng+="\n\n"
        for key in self.operators.keys():
            if not key == syskey:
                strng+=str(self.operators[key])+"\n"
                strng+="\n\n"

        for key in self.functions.keys():
            strng+=str(self.functions[key])+"\n"
            strng+="\n\n"

        strng+="end-operator\n "
        return strng


    def switchsign(self,opname='system'):
        """
        switchsign(opname='system') : multiply operator opname with -1
        Raise Exception of operator does not exist.
        """

        if self.operators.has_key(opname):

            def swsgn(line):
                import re
                if line[0:5]=='-----': return line
                if re.search('^[ \t]*[mM][oO][dD][eE][sS]',line) is not None:
                    return line
                pat='^[ \t]*-[a-zA-Z0-9\.]+'
                if re.search(pat,line) is not None:
                    return re.sub('^[ \t]*-','',line)
                pat='^[ \t]*\+[ \t]*[a-zA-Z0-9\.]+'
                if re.search(pat,line)is not None:
                    return re.sub('^[ \t]*\+','-',line)
                pat='^[ \t]*[a-zA-Z0-9\.]+'
                if re.search(pat,line) is not None:
                    return re.sub('^[ \t]*','-',line)

            self.operators[opname].data= \
                [ swsgn(line) for line in self.operators[opname].data ]

        else:
            raise Exception("No operator with name: "+opname)


    def copy(self,other):
        "Copy from other to self"
        self.clear()
        for name in other.sections.keys():
            self.sections[name]=mctdhsection()
            self.sections[name].copy(other.sections[name])
        for name in other.operators.keys():
            self.operators[name]=mctdhsection()
            self.operators[name].copy(other.operators[name])
        for name in other.functions.keys():
            self.functions[name]=mctdhsection()
            self.functions[name].copy(other.functions[name])



    def clear(self):
        "Delete all data"
        self.sections.clear()
        self.operators.clear()
        self.functions.clear()



    def addoper(self,name,sec):
        """
        AddOper(name,sec): Add section object sec to operators dictionary
        with key name
        """
        s=mctdhsection()
        s.copy(sec)
        self.operators[name]=s

    def deloper(self,name):
        """
        deloper(name):
        if self.operators.has_key(name): return self.operators.pop(name)
        """
        if self.operators.has_key(name): return self.operators.pop(name)



    def addfunction(self,name,sec):
        """
        addfunction(name,sec): Add section object sec to functionss dictionary
        with key name
        """
        s=mctdhsection()
        s.copy(sec)
        self.functions[name]=s


    def delfunction(self,name):
        """
        delfunction(name):
        if self.functionss.has_key(name): return self.functions.pop(name)
        """
        if self.functions.has_key(name): return self.functions.pop(name)

    def delallfunctions(self):
        """
        delallfunctions(name):
        Clear the functions dictionary.
        """
        self.functions.clear()

    def importfile(self,filename):
        """importfile(filename): Import operator file filename where filename
        can be a relative or absolute path of the file"""

        f=open(filename,'r')

        # import all sections
        for name,(arg0,arg1,importfunc) in self.secimportmap.items():
            sec=importfunc(f,arg0,arg1)
            if sec:self.sections[name]=sec


        #append system hamiltonian
        sstr='HAMILTONIAN-SECTION'
        estr='END-HAMILTONIAN-SECTION'
        delim='_'
        sec=einsection(f,sstr,estr)
        if sec:self.operators['system']=sec



        #append other hamiltonians
        hamsecs = secnames(f,sstr,delim)

        for name in hamsecs:
            self.operators[name]= einsection(f, sstr+str(delim)+str(name),estr )


        #append 1d functions

        sstr='FUNCTION-SECTION'
        estr='END-FUNCTION-SECTION'
        delim='_'

        hamsecs = secnames(f,sstr,delim)

        for name in hamsecs:
            self.functions[name]= einsection(f, sstr+str(delim)+str(name),estr )




    def exportfile(self,filename,overwrite=True,mode='w'):
        """
        exportfile(filename,overwrite=True,mode='w'):
        Save operator file as ASCII to file filename.
        Raise Exception if file exists and overwrite if False.
        """
        import os
        if os.path.exists(filename) and not overwrite:
            message="\nNo permition to overwrite file: "+filename+'\n'
            message+="Try setting the overwrite flag."
            raise Exception(message)
        else:
            f=open(filename,mode)
            f.write(str(self))
            f.close()
            #catch possible ecxeption in calling module


