"Tools used in the OCT scripts."

#########################################################################
#
# FUNCTIONS IN THIS MODULE
#
# - chksections:              Check that all sections are in .inp file
# - clearoldoutput:           Delete old output files
# - createfifos:              Make FiFo's
# - dumpfile:                 Dump a string to a file
# - emptytree:                Create an empty directory, by creating
#                             it or deleting all contents
# - floatstr:                 Convert float to string
# - outpath:                  Join outpath and a filename
# - parseoctdefault:          Basic parsing of the OCT-SECTION
# - parserundefault:          Basic parsing of the RUN-SECTION
# - parsetardefault:          Basic parsing of the TARGET-SECTION
# - saveoutputb:              Save output files after backward propagation
# - saveoutputf:              Save output files after forward propagation
# - setdefaultfiles:          Setup of filenames for temporary files
# - whatchprogs:              Monitoring of background precesses
#
# GLOBAL VARIABLES IN THIS MODULE
#
# - fs                        One femtosecond in atomic units
# - tiny                      A small number
#
#
# 04/2009 MS
#########################################################################

fs=float(41.3414)
tiny=1.e-8

# Check if the process are still running
# procs have to be instances of subprocess.Popen
#
# This only works on UNIX like platforms.
#
# NOTE: This relies on the programs to return a non-zero
#       exit code on error.

def whatchprogs(efproc, mctdhproc,mctdhproc1=None):
    "Monitoring of child processes "
    import time, os

    keepwhatching=True

    # check every time_wait seconds
    time_wait=1

    while keepwhatching:
        # Popen.wait() is not always reliable.

        efpoll    =efproc.poll()
        mctdhpoll =mctdhproc.poll()
        if mctdhproc1 is not None:
            mctdhpoll1=mctdhproc1.poll()
        else:
            mctdhpoll1=0

        # If anything goes wrong: kill the children!
        if efpoll != 0 and efpoll is not None:
            pid=mctdhproc.pid
            os.kill(pid,9)
            if mctdhproc1 is not None:
                pid=mctdhproc1.pid
                os.kill(pid,9)
            raise Exception('efield returned a non-zero exit code. '+\
                                'Sent SIG_KILL to MCTDH.')

        if mctdhpoll != 0 and mctdhpoll is not None:
            pid=efproc.pid
            os.kill(pid,9)
            if mctdhproc1 is not None:
                pid=mctdhproc1.pid
                os.kill(pid,9)

            raise Exception('MCTDH returned a non-zero exit code. '+\
                                'Sent SIG_KILL to efield.')

        if mctdhpoll1 != 0 and mctdhpoll1 is not None:
            pid=efproc.pid
            os.kill(pid,9)
            pid=mctdhproc.pid
            os.kill(pid,9)

            raise Exception('MCTDH returned a non-zero exit code. '+\
                                'Sent SIG_KILL to efield.')

        # has everything gone well? If yes, stop whatching.
        if efpoll == 0 and mctdhpoll == 0 and mctdhpoll1 == 0:
            keepwhatching=False

        #check once a second
        try:
            time.sleep(time_wait)
        except:
            # This should be executed when ctrl+c is pressed:
            # Kill child-processes and exit.
            pid=efproc.pid
            os.kill(pid,9)
            pid=mctdhproc.pid
            os.kill(pid,9)
            if mctdhproc1 is not None:
                pid=mctdhproc1.pid
                os.kill(pid,9)
            raise Exception("whatchprogs() has been interrupted. "+\
                                "Sent SIG_KILL to efield and MCTDH.")



##########################################################################
def outpath(str):
    import os
    from octjob import jobopts
    return os.path.join(jobopts['outdir'],str)

##########################################################################
def dumpfile(filename,content='',writepermit=True,mode='w'):
    """
    dumpfile(filename,content='',writepermit=True,mode='w'):
    Save str(content) to file filename.
    """
    import os
    if os.path.exists(filename) and not writepermit:
        raise Exception("ERROR: in subroutine dumpfile:\n Writepermit"+\
                            " not passed for file: "+filename)
    else:
        f=open(filename,mode)
        f.write(str(content))
        f.close()
        #catch possible ecxeption in calling module
##########################################################################

def emptytree(dirname):
    """
    emptytree(dirname): recursively delete all files and directories in dirname
    leaving dirname an empty directory. Create directory dirname if nonexistent.
    """
    import shutil,os
    try:
        if os.path.exists(dirname):shutil.rmtree(dirname)
        os.mkdir(dirname)
    except:
        message='Error while cleaning directory'+dirname
        raise Exception(message)

##########################################################################
def chksections(inp):
    """
    chksections(mctdhinput):
    Check that all sections required for OCT are in the input file.
    Raise exception if a section is missing
    """
    if not inp.sections.has_key('runsection'):
        raise Exception('RUN-SECTION not found in input file')

    if not inp.sections.has_key('opersection'):
        raise Exception('OPERATOR-SECTION not found in input file')

    if not inp.sections.has_key('octsection'):
        raise Exception('OCT-SECTION not found in input file')

    if not inp.sections.has_key('pbassection'):
        raise Exception('PBASIS-SECTION not found in input file')

    if not inp.sections.has_key('spfbassection'):
        raise Exception('SPF-BASIS-SECTION not found in input file')

    if not inp.sections.has_key('initwfsection'):
        raise Exception('INIT_WF-SECTION not found in input file')

    if not inp.sections.has_key('targetsection'):
        raise Exception('TARGET-SECTION not found in input file')

    if not inp.sections.has_key('intsection'):
        raise Exception('INTEGRATOR-SECTION not found in input file')

##########################################################################
def clearoldoutput(dir='.'):
    """
    clearoldoutput(dir='.'): remove all old output files from dir:
    rm -f  jfunc fieldf* fieldb* targetpop.* dipmtx* rawfield* overlap*
    """
    from messages import log

    import os
    filelist= os.listdir(dir)
    dellist=[file for file in filelist if \
                 file[0:6]=='fieldf' or\
                 file[0:6]=='fieldb' or\
                 file[0:9]=='targetpop' or\
                 file[0:8]=='rawfield' or\
                 file[0:7]=='overlap' or\
                 file[0:6]=='dipmtx' or\
                 file[0:6]=='reffnc' or\
                 file[0:7]=='lagrfnc']

    for i in range(len(dellist)):
        log.debug("Deleting file: "+str(dellist[i]))
        os.remove(os.path.join(dir,dellist[i]))

    try:
        os.remove(os.path.join(dir,'jfunc'))
    except:
        pass


##########################################################################
#  Default files and filenames
##########################################################################


################################ FIFO files ##############################
def createfifos(jobopts):
    "Create the FIFO files required for OCT in the current directory"

    import os

    try:
        fifos=('prop.fifo','efield.fifo','prop1.fifo','efield1.fifo',)
        fifos1=map(outpath,fifos)
        rmfifo=lambda f: os.remove(f)
        map(rmfifo,fifos1)
    except:
        pass

    if not os.path.exists(outpath('prop.fifo')):   os.mkfifo(outpath('prop.fifo'))
    if not os.path.exists(outpath('efield.fifo')): os.mkfifo(outpath("efield.fifo"))
    if jobopts['lsimprop']:
        if not os.path.exists(outpath('prop1.fifo'))  : os.mkfifo(outpath("prop1.fifo"))
        if not os.path.exists(outpath('efield1.fifo')): os.mkfifo(outpath("efield1.fifo"))



################################ Filenames ##############################
def setdefaultfiles(jobopts):
    "Set filenames of temporary and output files"

    jobopts['jfile']=outpath('jfunc')
    jobopts['overlaps']=outpath('overlaps')
    jobopts['initstate_inp']=outpath('initialstate.inp')
    jobopts['targstate_inp']=outpath('targetstate.inp')

    jobopts['forward1_inp']=outpath('forward1.inp')
    if jobopts['lsimprop']:
        jobopts['forward_init_inp']=outpath('forward_init.inp')
        jobopts['backward_init_inp']=outpath('backward_init.inp')
        jobopts['forward_targ_inp']=outpath('forward_targ.inp')
        jobopts['backward_targ_inp']=outpath('backward_targ.inp')

        jobopts['forward_init_op']=outpath('forward_init.op')
        jobopts['backward_init_op']=outpath('backward_init.op')
        jobopts['forward_targ_op']=outpath('forward_targ.op')
        jobopts['backward_targ_op']=outpath('backward_targ.op')

    else:
        jobopts['forwardi_inp']=outpath('forwardi.inp')
        jobopts['backwardi_inp']=outpath('backwardi.inp')

        jobopts['forward_op']=outpath('forward.op')
        jobopts['backward_op']=outpath('backward.op')


##########################################################################
#  Default parsing functions
##########################################################################

############################## RUN-SECTION ###############################
def parserundefault(inp,jobopts):
    "Default parsing of the RUN-SECTION"


    from messages import log
    import re

    # exact?
    log.debug('Parsing RUN-SECTION')
    log.debug('Making sure that no exact propagation is used.')
    if inp.hassecflag('runsection','exact'):
        raise Exception('OCT with exact propagation is not implemented')


    #MPI? not supported
    log.debug('Making sure that MPI is not used.')
    if inp.hassecflag('runsection','usempi') or \
           inp.hasseckey('runsection','usempi'):
        message='Distributed memory parallelization (MPI) is not supported'
        message+=' by optcntrl.'
        raise Exception(message)


    # check times
    log.debug('Checking time steps')
    if inp.hasseckey('runsection','tout'):
        tout=float( re.sub('[dD]','e',inp['runsection']['tout']) )
    if inp.hasseckey('runsection','tpsi'):
        tpsi=float( re.sub('[dD]','e',inp['runsection']['tpsi']) )
    else:
        tpsi=tout
    if inp.hasseckey('runsection','tfinal'):
        tfinal=float( re.sub('[dD]','e',inp['runsection']['tfinal']) )

    if abs(tpsi-tout) > tiny:
        message="In RUN-SECTION: tpsi (if given) has to be the same as tout."
        raise Exception(message)

    log.info('tfinal='+str(tfinal))
    log.info('tout='+str(tout))
    jobopts['tfinal']=tfinal
    jobopts['tout']  =tout



    # propagation mode

    jobopts['lsimprop']=False
    jobopts['lpc']=False
    if inp.hasseckey('runsection','optcntrl'):
        octparams=inp['runsection']['optcntrl'].lower().split(',')

        idx = -1
        try:
            idx=octparams.index('simprop')
        except:
            pass
        if idx>=0:
            jobopts['lsimprop']=True
            octparams.pop(idx)
            log.info('Found flag for simultaneous propagations')
        idx = -1
        try:
            idx=octparams.index('pc')
        except:
            pass
        if idx>=0:
            jobopts['lpc']=True
            octparams.pop(idx)
            log.info('Found flag for predictor-corrector mode')

        del idx

        if len(octparams):
            message='RUN-SECTION: unknown keyword(s)'
            message+=' following "optcntrl": '
            for i in range(len(octparams)):
                message+=str(octparams[i])+', '
            raise Exception(message)

    else:
        inp['runsection'].delparam('optcntrl')
        inp['runsection'].setflag('optcntrl')




    if jobopts['lsimprop'] and jobopts['lpc']:
        message='"simprop" and "pc" cannot be used together '\
            +'as argument for "optcntrl".'
        raise Exception(message)

    # psi=double set?
    if inp.hasseckey('runsection','psi'):
        if inp['runsection']['psi'].rfind('double') < 0:
            if not jobopts['lsimprop']:
                message='psi=double needs to be set in RUN-SECTION'
                log.error(message)
                raise Exception(message)
    else:
        if not jobopts['lsimprop']:
            message='psi=double needs to be set in RUN-SECTION'
            log.error(message)
            raise Exception(message)



############################## TARGET-SECTION ############################
def parsetardefault(inp,jobopts):
    "Default parsing of the TARGET-SECTION"

    from messages import log

    log.debug('Parsing TARGET-SECTION')
    # Operator or wavefunction?
    jobopts['ltaroper'] = inp.hasseckey('targetsection','operator')
    if jobopts['ltaroper']:
        jobopts['taroper']=inp['targetsection']['operator']
        log.info('Target is the operator: '+jobopts['taroper'] )
    else:
        log.info('Target is a wavefunction')


############################## OCT-SECTION ###############################
def parseoctdefault(inp,jobopts):
    "Default parsing of the OCT-SECTION"

    from messages import log
    import re

    log.debug('Parsing OCT-SECTION')

    # parameters of the control functional

    if not jobopts.has_key('niterations'):
        if inp['octsection'].hasparam('iterations'):
            jobopts['niterations']=int(inp['octsection']['iterations'])
    if not jobopts.has_key('niterations'):
        raise Exception("OCT-SECTION: number of iterations not given.")
    inp['octsection'].delparam('iterations')

    log.info('Number of iterations: '+str(jobopts['niterations']) )

    #penalty-factor
    if inp['octsection'].hasparam('penalty-factor'):
        jobopts['penalty-factor']=float(re.sub('d','e',inp['octsection']['penalty-factor']))
        log.info('Pernalty factor: '+str(jobopts['penalty-factor']))
        inp['octsection'].delparam('penalty-factor')
    else:
        jobopts['penalty-factor']=1.0
        log.info('Pernalty factor: '+str(jobopts['penalty-factor']))


    #reference function
    jobopts['lreffnc']=inp.hasseckey('octsection','reference')
    if jobopts['lreffnc']:
        jobopts['reffnc']=inp['octsection']['reference']
        inp['octsection'].delparam('reference')
        log.info('Reference-function is: '+jobopts['reffnc'])
    else:
        log.debug('No reference-function given')

    #penalty function
    jobopts['lpnltyfnc']=inp.hasseckey('octsection','penalty-function')
    if jobopts['lpnltyfnc']:
        jobopts['pnltyfnc']=inp['octsection']['penalty-function']
        inp['octsection'].delparam('penalty-function')
        log.info('Penalty-function is: '+jobopts['pnltyfnc'])
    else:
        log.info('No penalty-function given. Using unity.')

    #guess field
    jobopts['lguessfnc']=inp.hasseckey('octsection','guess')
    if jobopts['lguessfnc']:
        jobopts['guessfnc']=inp['octsection']['guess']
        inp['octsection'].delparam('guess')
        log.info('Guess-function is: '+jobopts['guessfnc'])
    else:
        jobopts['lguessfnc'] = True
        jobopts['guessfnc']  = 'zero'
        log.info('No guess-field given. Using zero.')

    #dipole operator
    if not inp.hasseckey('octsection','operator'):
        raise Exception('OCT-SECTION: operator not given')

    jobopts['dipoleoperator']=inp['octsection']['operator']
    log.info('Dipole operator is: '+jobopts['dipoleoperator'])
    inp['octsection'].delparam('operator')

    if inp.hassecflag('octsection','alignphs'):
        jobopts['lalignphs']=True
        inp['octsection'].delflag('alignphs')
        log.info('Overlap type is: alignphs' )
    else:
        jobopts['lalignphs']=False
        log.info('overlap type is: default' )


    if jobopts['ltaroper'] and jobopts['lalignphs']:
        raise Exception('Keyword \'alignphs\' in OCT-SECTION cannot be used'+
                    'together with \'operator\' keyword in TARGET-SECTION')

    if inp.hassecflag('octsection','tdoverlap'):
        jobopts['ltdoverlap']=True
        inp['octsection'].delflag('tdoverlap')
        log.info('Using time-dependent overlap to calculate the field.' )
    else:
        jobopts['ltdoverlap']=False
        log.info('Using time-independent overlap to calculate the field.' )


    # output
    jobopts['lwrtddipmtx']=False
    jobopts['lwrtdoverlap']=False
    jobopts['lwrtdtargetpop']=False
    jobopts['lwrrawflds']=False
    if inp.hasseckey('octsection','write'):
        listwrite=inp['octsection']['write'].lower().split(',')
        inp['octsection'].delparam('write')
        for j in range(len(listwrite)):
            if   listwrite[j]=='tddipmtx'   : jobopts['lwrtddipmtx']=True
            elif listwrite[j]=='tdoverlap'  : jobopts['lwrtdoverlap']=True
            elif listwrite[j]=='tdtargetpop': jobopts['lwrtdtargetpop']=True
            elif listwrite[j]=='rawfield'   : jobopts['lwrrawflds']=True
            elif listwrite[j]=='all':
                jobopts['lwrtddipmtx']=True
                jobopts['lwrtdoverlap']=True
                jobopts['lwrtdtargetpop']=True
                jobopts['lwrrawflds']=True
                break
            else:
                message='OCT-SECTION: "write" does not have the argument: '+listwrite[j]
                raise Exception(message)

        listwrite=[]

        if jobopts['lwrtddipmtx']:
            listwrite.append('Time-dependent dipole matrix\n')
        if jobopts['lwrtdoverlap']:
            listwrite.append('Time-dependent overlap\n')
        if jobopts['lwrtdtargetpop']:
            listwrite.append('Time-dependent target population\n')
        if jobopts['lwrrawflds']:
            listwrite.append('Raw fields\n')

        log.info("Writing output: \n   "+'   '.join(listwrite) )
        del listwrite


    #special parsing for Krotov
    jobopts['lfilter']=False
    if jobopts['algorithm']=='krotov':
        if inp.hassecflag('octsection','filter'):
            log.info('Filtering method: stop')
            jobopts['lpassfilter']=False
            inp['octsection'].delflag('filter')
            jobopts['lfilter'] = True

        elif inp['octsection'].hasparam('filter'):
            if not inp.sections.has_key('filtersection'):
                message="OCT-SECTION: keyword 'filter' requires"
                message+=" a FILTER-SECTION."
                raise Exception(message)
            if inp['octsection']['filter']=='pass':
                jobopts['lpassfilter']=True
                jobopts['lfilter'] = True
                log.info('Filtering method: pass')
                inp['octsection'].delparam('filter')
            elif inp['octsection']['filter']=='stop':
                log.info('Filtering method: stop')
                jobopts['lfilter'] = True
                jobopts['lpassfilter']=False
                inp['octsection'].delparam('filter')
            elif inp['octsection']['filter']=='none':
                log.info('No filtering used.')
                jobopts['lfilter'] = False
                inp['octsection'].delparam('filter')
            else:
                message="OCT-SECTION: 'filter' does not support"
                message+=" keyword '"+ inp['octsection']['filter']+"'"
                raise Exception(message)
        else:
            log.info('No filtering used.')
            jobopts['lfilter'] = False

        jobopts['lreffilter']=jobopts['lfilter']
        if inp.hassecflag('octsection','no-reffilter'):
            if not jobopts['lfilter']:
                log.warn("OCT-SECTION: Filtering not defined. Ignoring 'no-reffilter'.")
            else:
                jobopts['lreffilter']=False
                log.info('Reference function remains unfiltered')

            inp['octsection'].delflag('no-reffilter')
        elif jobopts['lfilter']:
            log.info('Reference function is filtered.')



    jobopts['lsvflds']=False
    jobopts['lsvtddipmtx']=False
    jobopts['lsvtdoverlap']=False
    jobopts['lsvtdtargetpop']=False
    jobopts['lsvrawflds']=False
    jobopts['lsvfilter']=False
    if inp.hasseckey('octsection','save'):
        listsave=inp['octsection']['save'].lower().split(',')
        inp['octsection'].delparam('save')
        for j in range(len(listsave)):
            if   listsave[j]=='field'      : jobopts['lsvflds']=True
            elif listsave[j]=='tddipmtx'   : jobopts['lsvtddipmtx']=True
            elif listsave[j]=='tdoverlap'  : jobopts['lsvtdoverlap']=True
            elif listsave[j]=='tdtargetpop': jobopts['lsvtdtargetpop']=True
            elif listsave[j]=='rawfield'   : jobopts['lsvrawflds']=True
            elif listsave[j]=='filter'     : jobopts['lsvfilter']=True

            elif listsave[j]=='all':
                jobopts['lsvflds']=True
                jobopts['lsvtddipmtx']=True
                jobopts['lsvtdoverlap']=True
                jobopts['lsvtdtargetpop']=True
                jobopts['lsvrawflds']=True
                if jobopts['lfilter']: jobopts['lsvfilter']=True
                break
            else:
                message='OCT-SECTIOIN: "save" does not have the argument: '+listsave[j]
                raise Exception(message)

        if jobopts['algorithm']!='krotov' and jobopts['lsvfilter']:
            raise Exception('OCT-SECTION: "save=filter" requires "algorithm=krotov".')
        if not jobopts['lfilter'] and jobopts['lsvfilter']:
            raise Exception('OCT-SECTION: "save=filter" requires "filter (=S)".')
        if jobopts['lsvtddipmtx'] and not jobopts['lwrtddipmtx']:
            raise Exception('OCT-SECTION: "save=tddipmtx" requires "write=tddipmtx" or "write=all".')
        if jobopts['lsvtdoverlap'] and not jobopts['lwrtdoverlap']:
            raise Exception('OCT-SECTION: "save=tdoverlap" requires "write=tdoverlap" or "write=all".')
        if jobopts['lsvtdtargetpop'] and not jobopts['lwrtdtargetpop']:
            raise Exception('OCT-SECTION: "save=tdtargetpop" requires "write=tdtargetpop" or "write=all".')
        if jobopts['lsvrawflds'] and not jobopts['lwrrawflds']:
            raise Exception('OCT-SECTION: "save=rawfield" requires "write=rawfield" or "write=all".')

        listsave=[]
        if jobopts['lsvflds']: listsave.append('Fields\n')
        if jobopts['lsvrawflds']: listsave.append('Raw fields\n')
        if jobopts['lsvtddipmtx']:
            listsave.append('Time-dependent dipole matrix\n')
        if jobopts['lsvtdoverlap']:
            listsave.append('Time-dependent overlap\n')
        if jobopts['lsvtdtargetpop']:
            listsave.append('Time-dependent target population\n')
        if jobopts['lsvfilter']:
            listsave.append('Filtered temporary fields (filter algorithm)\n')
        log.info("Saving output: \n   "+'   '.join(listsave) )
        del listsave



    # return True if there are more keywords to parse
    # otherwise return False
    return not inp['octsection'].isempty()



##########################################################################
def floatstr(r,format="%.16f"):
    """
    floatstr(r,format="%.16f") -> string
    Return a string for float r with format format.
    """
    return format % r


##########################################################################
def saveoutputf(jobopts):
    "Save output files after forward iteration."

    from messages import log
    import shutil

    try:
        if jobopts['lsvtdtargetpop']:
            log.debug('Saving target population')
            shutil.copyfile('targetpop.dat', 'targetpop.'+str(jobopts['curriter']))
        if jobopts['lsvflds']:
            log.debug('Saving fields')
            if jobopts['lsimprop']:
                shutil.copyfile('fieldf_init.dat', 'fieldf.'+str(jobopts['curriter']))
            else:
                shutil.copyfile('fieldf.dat', 'fieldf.'+str(jobopts['curriter']))
        if jobopts['lsvtddipmtx']:
            log.debug('Saving dipole matrix')
            shutil.copyfile('dipmtxf.dat', 'dipmtxf.'+str(jobopts['curriter']))
        if jobopts['lsvrawflds']:
            log.debug('Saving raw field')
            shutil.copyfile('rawfieldf.dat', 'rawfieldf.'+str(jobopts['curriter']))
        if jobopts['lsvtdoverlap']:
            log.debug('Saving overlap')
            shutil.copyfile('overlapf.dat', 'overlapf.'+str(jobopts['curriter']))

        if jobopts['lsvfilter']:
            log.debug('Saving filtered fields')
            shutil.copyfile('lagrfnc.dat', 'lagrfnc.'+str(jobopts['curriter']))
            shutil.copyfile('reffnc.dat', 'preffnc.'+str(jobopts['curriter']))


    except Exception as E:
        if jobopts["curriter"] == 0:
            pass
        else:
            raise E


##########################################################################
def saveoutputb(jobopts):
    "Save output files after backward iteration."
    from messages import log
    import shutil

    if jobopts['lsvtddipmtx']:
        log.debug('Saving dipole matrix')
        shutil.copyfile('dipmtxb.dat', 'dipmtxb.'+str(jobopts['curriter']))
    if jobopts['lsvrawflds']:
        log.debug('Saving raw field')
        shutil.copyfile('rawfieldb.dat', 'rawfieldb.'+str(jobopts['curriter']))
    if jobopts['lsvtdoverlap']:
        log.debug('Saving overlap')
        shutil.copyfile('overlapb.dat', 'overlapb.'+str(jobopts['curriter']))
    if jobopts['lsvflds']:
        log.debug('Saving fields')
        if jobopts['lsimprop']:
            shutil.copyfile('fieldb_targ.dat', 'fieldb.'+str(jobopts['curriter']))
        else:
            shutil.copyfile('fieldb.dat', 'fieldb.'+str(jobopts['curriter']))



