#!/usr/bin/env python2
"""
--------------------------------------------------------------------------------
Purpose    : Optimal control calculation by runnig QUANTICS iteratively.

Usage      : optcntrl [ -mnd -c -D -n -r -R -w -x -deb -ver -h -? ] inputfile

  -mnd       Make name directory.
  -nd        Do not delete old output files in name directory before starting
             a new calculation.
  -D <dir>   Denotes the directory where files are written to,
             (name in .inp file ignored).
  -n <niter> Perform niter iterations (ignore value in input file)
  -r         Restart calculation. Perform additional iterations.
  -R         Restart calculation similar to -r, but start with a
             forward calculation (if crashed during forward).
  -w         Allow to overwrite existing data.
  -x <suf>   Use QUANTICS/analyse executables with ending "suf",
             e.g., "-x .g95" to use binaries compiled with g95.
  -deb       Run in debug mode (more detailed logging)
  -ver       Print version info
  -h -?      Print this help text.

For restart-runs, the input file is parsed but no temporary input
files are created. The current iteration number is obtained from
file 'overlaps'
To stop the calculation, create an empty file 'stop'.
I.e. type 'touch stop' while being in the working directory.
--------------------------------------------------------------------------------"""
# This script has been tested with Python versions 2.4.4 and 2.5.2


if __name__ == "__main__":

    import os, mctdhinp
    from sys import argv

    # import quit for compatibility with versions earlier than python 2.5
    from sys import exit as quit
    from octjob import jobopts, parsecmdln, checkexepaths
    import datetime
    datetm = datetime.datetime.now()

    # avoid 'print' command
    def write(msg=""):
        from sys import stdout

        stdout.write(str(msg) + "\n")
        stdout.flush()

    jobopts["mctdh_ver"] = "1.0"

    #######################################################################
    #   Read the command line options and store in jobopts{}
    #######################################################################
    try:
        cont = parsecmdln(argv)
    except Exception as E:
        write(" ERROR while parsing the command line options:")
        write("       " + str(E))
        quit()
    except:
        write(" ERROR while parsing the command line options:")
        write("       Unexpected error")
        quit()
    if not cont:
        quit()

    write("-" * 78)

    #######################################################################
    #   Extract input filename
    #######################################################################

    filename = os.path.expanduser(jobopts["inpfile"])
    filename = os.path.expandvars(filename)
    filename = os.path.abspath(filename)
    jobopts["inpfile"] = filename

    #######################################################################
    # Load the input file
    #######################################################################
    try:
        inp = mctdhinp.mctdhinput(filename)
    except Exception as E:
        write(" ERROR while reading input file:")
        write("       " + str(E))
        quit()
    except:
        write(" ERROR while reading input file.")
        write("       Unexpected error")
        quit()

    del filename

    #######################################################################
    #   Get basic informaton from input file (name,overwrite) and
    #   create name directory if required
    #######################################################################

    # overwrite?
    jobopts["overwrite"] = (
        inp.hassecflag("runsection", "overwrite") or jobopts["overwrite"]
    )  # from command line

    # name
    if "name" not in jobopts:
        jobopts["name"] = inp["runsection"]["name"]
    if not jobopts["name"]:
        write(" ERROR: No name directory given.:")
        write("        Use -D option")
        quit()

    jobopts["outdir"] = os.path.join(os.getcwd(), jobopts["name"])

    if not os.path.exists(jobopts["outdir"]) and not jobopts["mnd"]:
        write(" ERROR: Unable to create name directory: " + jobopts["name"])
        write("        Use -mnd option")
        quit()

    if not os.path.isdir(jobopts["outdir"]) and os.path.exists(jobopts["outdir"]):
        write(" ERROR: Unable to create name directory: " + jobopts["name"])
        write("        File exists but is not a directory.")
        quit()

    if jobopts["mnd"] and not os.path.exists(jobopts["outdir"]):
        try:
            os.mkdir(jobopts["outdir"])
        except Exception as E:
            write(" ERROR: Unable to create name directory:")
            write("       " + str(E))
            quit()
        except:
            write(" ERROR: Unable to create name directory:")
            write("        Unexpected error.")
            quit()
        write("Created name directory: " + jobopts["outdir"])

    #######################################################################
    #   We are ready to start the logging system.
    #######################################################################

    from messages import log, logsetup

    jobopts["logfile"] = os.path.join(jobopts["outdir"], "optcntrl.log")

    # append to existing log if this is a restart
    mode = "w"
    if jobopts["restartf"] or jobopts["restartb"]:
        mode = "a"

    if (not os.path.exists(jobopts["logfile"])) or jobopts["overwrite"] or mode == "a":
        logsetup(jobopts["logfile"], jobopts["ldebug"], mode)
    else:
        print(" ERROR: unable to write to logfile. Use -w option.")
        quit()
    del mode
    log.info( datetm )

    #######################################################################
    #   save a copy of the input file
    #######################################################################

    inpcopy = jobopts["outdir"] + "/input"
    if (not os.path.exists(inpcopy)) or jobopts["overwrite"]:
        f = open(inpcopy, mode="w")
        f.write(str(inp))
        f.close()
        del f
    elif os.path.exists(inpcopy) and not jobopts["overwrite"]:
        log.error("Cannot overwrite " + inpcopy + ": use -w option.")
        quit()

    #######################################################################
    #   Check if QUANTICS-executables are in the path (add others if neeed
    #   in new algorithm).
    #   Filenames are stored in jobopts['efield_exe'],jobopts['quantics_exe'] etc.
    #   Later on we will rely on the executables being in the $PATH.
    #   checkexepaths() will append version number and suffixes if given
    #   in the command line
    #######################################################################

    log.info("OCT-QUANTICS: Starting optimal control job")

    # list of QUANTICS programs needed
    programs = ("efield", "quantics", "genfld", "fldop", "jfunc")
    try:
        log.debug("Checking if programs are in the PATH")
        checkexepaths(programs, jobopts)
    except Exception as E:
        log.exception(E)
        quit()
    except:
        log.error("Unexpected error while checking $PATH")
        quit()

    #   Set executables manually if this fails. Use something like:
    #   jobopts['program_exe']= "program"

    #######################################################################
    #   Read the OCT operator file
    #######################################################################

    if "opersection" not in inp.sections:
        log.error("No OPERATOR-SECTION found in input file. Typo?")
        quit()

    if inp.hasseckey("opersection", "oppath"):
        try:
            oppath = os.path.expanduser(inp["opersection"]["oppath"])
            oppath = os.path.expandvars(oppath)
            oppath = os.path.abspath(oppath)
        except KeyError as E:
            log.exception(E)
            quit()
        except:
            log.error("Unexpected error while parsing path to operator.")
            quit()
    else:
        oppath = os.getcwd()

    try:
        opname = inp["opersection"]["opname"]
    except KeyError as E:
        log.exception(E)
        quit()
    except:
        log.error("Unexpected error trying to retrieve Operator name.")
        quit()

    if opname[-3:] != ".op":
        opname += ".op"
        inp["opersection"]["opname"] = opname

    filename = os.path.join(oppath, opname)
    del opname, oppath

    jobopts["operfile"] = filename

    log.debug("Reading operator file: " + jobopts["operfile"])
    try:
        op = mctdhinp.mctdhoper(filename)
    except Exception as E:
        log.exception(E)
        quit()
    except:
        log.error("Unexpected error while reading oper file")
        quit()

    f = open(inpcopy, mode="a")
    f.write("\n\n####################################\n\n")
    f.write(str(op))
    f.close()

    #######################################################################
    #   We have all information, Now load the algorithm
    #######################################################################
    jobopts["continue"] = True

    from setup import setup

    try:
        # setup prepares all that is needed for the algorithm,
        # i.e., it generates temporary files,
        # performs a first propagation with a guess-field and the
        # very first anafw if needed etc.
        # It also handles the restart option and takes appropriate
        # precautions.
        # setup() returns the functions to be used in the iterations
        forw, anafw, backw, anabw, cleanup = setup(jobopts, inp, op)

    except Exception as E:
        log.exception(E)
        quit()
    except:
        log.error("Unexpected error during generation of algorithm")
        quit()

    #######################################################################
    #   Iterate the algorithm until jobopts['continue'] is False
    #######################################################################

    try:
        while jobopts["continue"]:
            backw(jobopts)
            anabw(jobopts)
            if not jobopts["continue"]:
                break
            forw(jobopts)
            anafw(jobopts)
            if not jobopts["continue"]:
                break
    except Exception as E:
        log.exception(E)
        quit()
    except:
        log.error("Unexpected error during iterations")
        quit()

    #######################################################################
    #   Done! Whats left is some final analysis and garbage collection
    #######################################################################
    cleanup(jobopts)

    log.info("Program terminated normally")
    datetm = datetime.datetime.now()
    log.info( datetm )
    write("-" * 78)
    quit()
#######################################################################
