#!/bin/bash

########################################################################
# Name: g09_scf.sh
# Syntax: ./g09_scf.sh
# Author: David mendive-Tapia
########################################################################
# Description
# Made to work on schroedinger
########################################################################
# Route followed by the script
# 1) Initialise variables
# 2) Consistency checks
# 3) Default values
# 4) Functions
# 5) Create points if they do not exist already
# 6) Create database
# 7) Fitting
########################################################################
# Created: 17/10/2020
# Last modification: 14/11/2020
########################################################################

#-----------------------------------------------------------------------
# 1) Initialise variables
#-----------------------------------------------------------------------
path=./templates
molec=pyr
method=uhf
ljcp='F' # Flag to reproduce the numerical results from JCP-96(7)-1992

#-----------------------------------
# STO-3G
#-----------------------------------
filefreq=pyr_RHF_sto3g_FreqHPModes_S0
modes=(3 8 12 15)
# Only one option should be uncommented:
# 1) VALENCE (RMSD=0.88340953E-01 with Model 1)
ljcp='T'
states=(0 1 2)
mo1=(0 20 21)
mo2=(0 22 22)
# Uncomment to have a better agreement with JCP-96(7)-1992
# (RMSD=0.11793138E+00 with Model 1)
#steps[3]=1
#steps[12]=1
#steps[15]=1
# 2) X-RAY (RMSD=0.40989512E-01 with Model 2)
#ljcp='F'
#states=(1 2 3 4 5 6)
#mo1=(0  1  1  1  2  2  2)
#mo2=(0 22 23 24 22 23 24)
# 3) VALENCE + X-RAY (RMSD=0.56723410E-01 with Model 2)
#ljcp='T'
#states=(0 1 2 3 4 5 6 7 8)
#mo1=(0 20 21  1  1  1  2  2  2)
#mo2=(0 22 22 22 23 24 22 23 24)

#-----------------------------------
# D95**
#-----------------------------------
#filefreq=pyr_RHF_D95s2_FreqHPModes_S0
#modes=(3 7 12 15)
# Only one option should be uncommented:
# 1) VALENCE (RMSD=0.78115448E-01 with Model 1)
#            (RMSD=0.69849792E-01 with Model 2)
#ljcp='T'
#states=(0 1 2)
#mo1=(0 20 21)
#mo2=(0 22 22)
# 2) X-RAY (RMSD=0.36706244E-01 with Model 2)
#ljcp='F'
#states=(1 2 3 4 5 6)
#mo1=(0  1  1  1  2  2  2)
#mo2=(0 22 23 29 22 23 29)
# 3) VALENCE + X-RAY (RMSD=0.53681923E-01 with Model 2)
#ljcp='T'
#states=(0 1 2 3 4 5 6 7 8)
#mo1=(0 20 21  1  1  1  2  2  2)
#mo2=(0 22 22 22 23 29 22 23 29)

#-----------------------------------------------------------------------
# 2) Consistency checks
#-----------------------------------------------------------------------
if [ ! -d $path ]; then
   echo "Directory $path does not exist!"
   exit
elif [ ! -f $path/${filefreq}.log ]; then
   echo "File ${filefreq}.log does not exist!"
   exit
fi

#-----------------------------------------------------------------------
# 3) Default values
#-----------------------------------------------------------------------
# Total number of modes
nmodes=$(grep 'Frequencies ---' $path/${filefreq}.log | tr -s ' ' | cut -d ' ' -f4-8 | xargs -n1 | nl | tail -1 | awk '{printf("%s\n",$1)}')
# Steps and increment set as 10 and 0.5 for all modes (if they have not already been defined)
first=0
last=${modes[-1]}
for ((i=$first;i<=$last;i++)); do
   [ -z "${steps[$i]}" ] && steps[$i]=10
   [ -z "${increment[$i]}" ] && increment[$i]=0.5
done

#-----------------------------------------------------------------------
# 4) Functions
#-----------------------------------------------------------------------
function vcpnt_scf {
   for s in "${states[@]}"; do
      echo ''
      echo "state: $s"
#     Open/close files
      fileopen=${molec}_s${s}_open
      fileclose=${molec}_s${s}_close
      if [ $s -eq 0 ] ; then
         cp $path/${method}_open.dat ${fileopen}.dat
         cp $path/${method}_close.dat ${fileclose}.dat
      else
         cp $path/${method}_alter_open.dat ${fileopen}.dat
         cp $path/${method}_alter_close.dat ${fileclose}.dat
         sed -i "s/{MO1}/${mo1[$s]}/g" ${fileclose}.dat
         sed -i "s/{MO2}/${mo2[$s]}/g" ${fileclose}.dat
      fi
#     Create and run VCPNT input files
      for m in "${modes[@]}"; do
         input=${molec}_s${s}_m${m}_pnt
         echo "mode: $m => ${input}.inp"
#        Copy template
         cp $path/1_pnt.inp ${input}.inp
#        Fixed string substitutions
         sed -i "s/{filefreq}/${filefreq}/g" ${input}.inp
         sed -i "s/{nmodes}/${nmodes}/g" ${input}.inp
         sed -i "s/{fileopen}/${fileopen}/g" ${input}.inp
         sed -i "s/{fileclose}/${fileclose}/g" ${input}.inp
#        Loop-dependent string substitutions
         sed -i "s/{mode}/${m}/g" ${input}.inp
         sed -i "s/{steps}/${steps[$m]}/g" ${input}.inp
         sed -i "s/{increment}/${increment[$m]}/g" ${input}.inp
         sed -i "s/{molecule}/${molec}_s${s}/g" ${input}.inp
#        Call vcpnt85
         vcpnt85 ${input}.inp >> g09_scf.out
      done
   done
}

function sed_states_energy0_fclog_scf {
#  Handle special case S0 (i.e. No Guess=alter)
   if [ ${states[0]} -eq 0 ] ; then
      first=0
      last=${states[-1]}
      string=$(echo "${states[@]}" | for ((i=$first;i<=$last;i++)); do echo "$((${states[$i]}+1))"; done | xargs | tr -s ' ' | tr ' ' ',')
   else
      string=$(echo ${states[@]} | tr -s ' ' | tr ' ' ',')
   fi
#  Write "states = 1,2,etc"
   sed -i "s/{states}/${string}/g" ${input}.inp
#  Ground state energy at FC geometry
   energy0=$(grep 'SCF Done:  E(RHF) =' ${filefreq}.log | awk '{printf("%s\n",$5)}')
   sed -i "s/{energy0}/${energy0}/g" ${input}.inp
#  Excited state energies at FC geometry
   grep 'SCF Done:' ./points/${molec}_s?_${modes[0]}_000.log | cut -d '/' -f3 | sed 's/://g' | awk '{printf("%s %s\n",$1,$6)}' > temp1.dat
   cat temp1.dat | sort -n -k2 | nl | awk '{printf("%s %s\n",$2,$1)}' > temp2.dat
   string='{files:fclog}'
   sed -i -e "/${string}/r temp2.dat" -e "/${string}/d" ${input}.inp
}

function setfiles_scf {
   for m in "${modes[@]}"; do
      echo -n > ${molec}_${m}l.set
      echo -n > ${molec}_${m}r.set
      first=1
      last=${steps[$m]}
      for ((i=$first;i<=$last;i++)); do
         echo -n > temp1.dat
         echo -n > temp2.dat
         for s in "${states[@]}"; do
            string=$(echo "$i" | awk '{printf("%02d\n",$0)}')
            # Points to the left
            string2=$(grep 'The electronic state is' ./points/${molec}_s${s}_${m}_${string}l.log | awk '{printf("%s\n",$5)}')
            string3=$(grep 'SCF Done:' ./points/${molec}_s${s}_${m}_${string}l.log | awk '{printf("%s\n",$5)}')
            echo "${molec}_s${s}_${m}_${string}l.log # ${string2} ${string3}" >> temp1.dat
            # Points to the right
            string2=$(grep 'The electronic state is' ./points/${molec}_s${s}_${m}_${string}r.log | awk '{printf("%s\n",$5)}')
            string3=$(grep 'SCF Done:' ./points/${molec}_s${s}_${m}_${string}r.log | awk '{printf("%s\n",$5)}')
            echo "${molec}_s${s}_${m}_${string}r.log # ${string2} ${string3}" >> temp2.dat
         done
         cat temp1.dat | sort -n -k4 | nl | awk '{printf("%s %s %s %s %s\n",$2,$1,$3,$4,$5)}' >> ${molec}_${m}l.set
         cat temp2.dat | sort -n -k4 | nl | awk '{printf("%s %s %s %s %s\n",$2,$1,$3,$4,$5)}' >> ${molec}_${m}r.set
      done
      echo 'end-files' >> ${molec}_${m}l.set
      echo 'end-files' >> ${molec}_${m}r.set
      echo "${molec}_${m}l.set ${molec}_${m}r.set"
   done
   ls -1 *.set | sort -n -t '_' -k2 > temp1.dat
   string='{files:set}'
   sed -i -e "/${string}/r temp1.dat" -e "/${string}/d" ${input}.inp
}

#  Put limits
#  Comment all the energy readings for a given state and mode outside the FC region
function limitset {
   # $1: mode
   # $2: state
   sed -i "s/pyr_s${2}/#&/" pyr_${1}l.set
   sed -i "s/pyr_s${2}/#&/" pyr_${1}r.set
}

function sed_liststates_scf {
   grep 'SCF Done:' ./points/${molec}_s?_${modes[0]}_000.log | sed 's/://g' | sort -n -k6 | awk '{printf("%s\n",$1)}' > temp1.dat
   cat temp1.dat | xargs grep 'The electronic state is' | awk '{printf("%s\n",$NF)}' | sed 's/\.//g' | cut -d '-' -f2 | nl | awk '{printf("%s %s\n",$1,$2)}' > temp2.dat
   string='{list:states}'
   sed -i -e "/${string}/r temp2.dat" -e "/${string}/d" ${input}.inp
}

#-----------------------------------------------------------------------
# 5) Create points if they do not exist already
#-----------------------------------------------------------------------
echo -n > g09_scf.out
if [ ! -d ./points ] ; then

   echo 'Generating points...'
#  Copy Freq output
   cp $path/${filefreq}.log .
#  Create points
   mkdir points
   vcpnt_scf
#  Store results
   mv go.com ./points
   mv ${molec}_*.com ./points
   rm -rf 1_VCPNT
   mkdir 1_VCPNT
   mv ${molec}_*.inp ./1_VCPNT
   mv ${molec}_*.dat ./1_VCPNT
   mv ${molec}_*.log ./1_VCPNT

else

#-----------------------------------------------------------------------
#  5) Create database
#-----------------------------------------------------------------------
   echo ''
   echo 'Generating database...'
   echo ''
   input=${molec}_trans
#  Copy Freq output
   cp $path/${filefreq}.log .
#  Copy template
   cp $path/2_trans.inp ${input}.inp
#  Fixed string substitutions
   sed -i "s/{filefreq}/${filefreq}/g" ${input}.inp
   sed -i "s/{nmodes}/${nmodes}/g" ${input}.inp
   sed -i "s/{method}/${method}/g" ${input}.inp
#  More complicated substitutions dependent on the ab initio method (e.g. SCF)
   sed_states_energy0_fclog_scf
#  Make .set files
   setfiles_scf
   [ ${ljcp} == 'T' ] && limitset ${modes[1]} 2
#  Call vctrans85
   vctrans85 ${input}.inp >> g09_scf.out
#  Store results
   rm -rf temp1.dat
   rm -rf temp2.dat
   rm -rf 2_VCTRANS
   mkdir 2_VCTRANS
   mv *.set ./2_VCTRANS
   mv *.inp ./2_VCTRANS
   mv *.info ./2_VCTRANS
   mv *.log ./2_VCTRANS

#-----------------------------------------------------------------------
#  7) Fitting
#-----------------------------------------------------------------------
   echo ''
   echo 'Fitting of parameters...'
   echo ''
   input=${molec}_vchfit
#  Copy template
   cp $path/3_vchfit.inp ${input}.inp
#  Copy database
   cp ./2_VCTRANS/${molec}_trans.info .
#  Fixed string substitutions:
#  1) Database
   sed -i "s/{name.info}/${molec}_trans.info/g" ${input}.inp
#  2) List of states {list:states}
   sed_liststates_scf
#  3) Modes to fit {fit_select}
   string=$(echo ${modes[@]} | tr -s ' ' | tr ' ' ',')
   sed -i "s/{fit_select}/${string}/g" ${input}.inp
#  Call vchfit85
   vchfit85 ${input}.inp >> g09_scf.out
#  Store results
   rm -rf temp1.dat
   rm -rf temp2.dat
   rm -rf 3_VCHFIT
   mkdir 3_VCHFIT
   mv *.inp ./3_VCHFIT
   mv *.vcham ./3_VCHFIT
   mv *.info ./3_VCHFIT
   mv *.log ./3_VCHFIT
   mv *.op ./3_VCHFIT

fi
#-----------------------------------------------------------------------

exit
