!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! An example module with a test PES
! for cluster tools
!
! Contains:
!
!  setup     Preparations: allocate temporary
!            memory and initialize it
!
!  free      Deallocate temporary memory
!
!  potential Calculate the PES (V) from a
!            coordinate  vector (Q)
!
!            V = Q**T A Q
!
!            (Quadratic Form with matrix A)
!
!            Calls setup it required.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!




module exampleroutine
  implicit none
!  private
!  public :: setup, potential, free

  integer :: ndof=0
  logical :: isInitialized = .false.
  double precision, allocatable :: Amat(:,:)


contains

  ! allocate memory for Amat
  subroutine setup(N, rseed)
    integer, intent(IN)           :: N
    integer, intent(IN), optional :: rseed(:)

    integer, allocatable          :: rseed1(:)
    integer                       :: slen, slen1
    integer                       :: i

    call mfree()

    call random_seed(size=slen1)
    allocate(rseed1(slen1))
    rseed1 = 0

    if (present(rseed)) then
       slen = size(rseed,1)

       if (slen .gt. slen1) then
          rseed1(1:slen1) = rseed(1:slen1)
       else
          rseed1(1:slen) =  rseed(1:slen)
       end if

    else
       call random_seed(get=rseed1(1:slen1))
    end if

    call random_seed(put=rseed1(1:slen1))
    deallocate(rseed1)

    ndof = N
    allocate(Amat(ndof,ndof))
    call random_number(Amat)
    Amat = Amat-0.5d0
    Amat = Amat + transpose(Amat)
    Amat = Amat/4

    do i=1,ndof
       Amat(i,i) =  abs(Amat(i,i)) + 1.d0
    end do

    isInitialized = .true.

  end subroutine setup



  ! free memory of amat
  ! NOTE: memory leak if not called after
  !       potential is not needed any more!
  subroutine mfree()

    if (allocated(Amat)) deallocate(Amat)
    isInitialized = .false.
    ndof = 0

  end subroutine mfree



  ! create a potential - this is just a
  ! quadratic form.
  double precision function potential(Q,N)
    integer, intent(IN)          :: N
    double precision, intent(IN) :: Q(N)

    double precision             :: tmp(N) ! auto alloc
    integer                      :: i,j

    ! call a default setup if not done manually
    if ( .not. allocated(Amat) .or. &
          .not. isInitialized ) then
       call setup(N)
    end if
    if (N .ne. ndof) then
       write(0,*) "Error in 'potential(Q,N)':"
       write(0,*) "N .ne. ndof."
       stop 1
    end if

    do j=1,ndof
       tmp(j) = 0.d0
       do i=1,ndof
           tmp(j) = tmp(j) + Amat(i,j)*q(i)
       end do
    end do
    potential = dot_product(Q,tmp)

  end function potential


end module exampleroutine
