/****************************************************************
 * CREMA - convert record markers of unformatted Fortran output *
 ****************************************************************
 *
 * Written by Frank Otto, 04/2006.
 * I hereby place this code in the public domain.
 *
 ****************************************************************/



/* includes */
#define _GNU_SOURCE
#include <getopt.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>



/* constants */
#define CONV_NONE 0
#define CONV_8TO4 1
#define CONV_4TO8 2

#define STATUS_OK 0
#define STATUS_EOF 1
#define STATUS_INVALID 2
#define STATUS_ERROR 3

#define CHUNKSIZE (16*1024)



/* prototypes */
void usage (FILE*, int);
void convert_record_markers(const char*, int);
int read_record_marker(FILE*, int, int32_t*);
int write_record_marker(FILE*, int, int32_t);



/* main program */
int main(int argc, char* argv[]) {

    int convert_mode = CONV_NONE;
    int opt, i, nfiles;
    const char** files;

    struct option opts[] = {
	{ "8to4", no_argument, &convert_mode, CONV_8TO4 },
	{ "4to8", no_argument, &convert_mode, CONV_4TO8 },
	{ "h", no_argument },
	{ 0, 0, 0, 0 }
    };

    /* parse options */
    while (1) {
	opt = getopt_long_only(argc, argv, "", opts, NULL);
	if (opt == -1) break; /* no more options */
	switch(opt) {
	  case 0:
	    /* ok */
	    break;
	  case 'h':
	    usage(stderr, 0);
	    break;
	  default:
	    /* unknown option */
	    usage(stderr, 1);
	    break;
	}
    }

    /* specifying the conversion mode is mandatory */
    if (convert_mode == CONV_NONE) {
	usage(stderr, 1);
    }

    /* rest of the arguments are filenames */
    nfiles = argc-optind;
    files  = (const char**) malloc( sizeof(char*) * nfiles );
    for (i=0; i<nfiles; i++) {
	files[i] = argv[optind+i];
    }

    /* process the files */
    for (i=0; i<nfiles; i++) {
	convert_record_markers(files[i], convert_mode);
    }

    return 0;
}



/* print usage information to file and exit with given exitcode */
void usage (FILE* f, int exitcode) {
    fprintf(f, "%s",
      "--------------------------------------------------------------\n"
      "Usage  : crema { -h | -8to4 | -4to8 } file [ file ... ]\n"
	"Options:\n"
      "  -h   : print this help text\n"
	"  -8to4: change 8 byte FORTRAN record markers to 4 byte\n"
	"  -4to8: change 4 byte FORTRAN record markers to 8 byte\n"
	"For each file, the record marker will be changed as specified.\n"
	"The original files will be renamed with the ending \".bak\".\n"
      "--------------------------------------------------------------\n"
	);
    exit(exitcode);
}



/* conversion routine */
void convert_record_markers(const char* fname, int mode) {
    char* fname_new; /* temporary filename of new file */
    char* fname_bak; /* filename of backup file */
    FILE* origfile;  /* original file */
    FILE* newfile;   /* new file */
    char  buffer[CHUNKSIZE];
    int len, ret, ok, nread, nwritten, ntodo, nchunk;
    int32_t rm_pre, rm_post;

    /* create filenames for new/backup files */
    len = strlen(fname);

    fname_new = (char*) malloc(len+5);
    strncpy(fname_new, fname, len+1);
    strncat(fname_new, ".new", 4);
    
    fname_bak = (char*) malloc(len+5);
    strncpy(fname_bak, fname, len+1);
    strncat(fname_bak, ".bak", 4);

    /* open old file for reading */
    origfile = fopen(fname, "rb");
    if (origfile == NULL) {
	perror(fname);
	return;
    }

    /* open new file for writing */
    newfile = fopen(fname_new, "wb");
    if (newfile == NULL) {
	perror(fname_new);
	fclose(origfile);
	return;
    }

    /* read-write loop */
    ok = 0;
    while (1) {
	/* read pre record marker */
	ret = read_record_marker(origfile, mode, &rm_pre);
	if (ret == STATUS_EOF) {
	    /* ok, original file is finished */
	    ok = 1;
	    break;
	} else if (ret == STATUS_INVALID) {
	    /* invalid record marker */
	    fprintf(stderr,
		    "%s: encountered invalid record marker, aborting\n",
		    fname);
	    break;
	}

	/* write corresponding record marker to new file */
	ret = write_record_marker(newfile, mode, rm_pre);
	if (ret != STATUS_OK) {
	    fprintf(stderr,
		    "%s: error while writing to file\n",
		    fname_new);
	    break;
	}

	ntodo = rm_pre; /* number of bytes we still need to write */
	while (ntodo>0) {
	    /* don't process more than CHUNKSIZE bytes at once */
	    nchunk = (ntodo<CHUNKSIZE) ? ntodo : CHUNKSIZE;

	    /* read data into buffer */
	    nread = fread(buffer, 1, nchunk, origfile);
	    if (nread < nchunk) {
		fprintf(stderr,
			"%s: premature end, aborting\n",
			fname);
		break;
	    }

	    /* write data to new file */
	    nwritten = fwrite(buffer, 1, nchunk, newfile);
	    if (nwritten < nchunk) {
		fprintf(stderr,
			"%s: error while writing to file\n",
			fname_new);
		break;
	    }
			  
	    /* decrease number of bytes that need to be written */
	    ntodo -= nchunk;
	}

	if (ntodo == 0) {
	    /* the above loop finished without error */

	    /* read post record marker */
	    ret = read_record_marker(origfile, mode, &rm_post);
	    if (ret != STATUS_OK) {
		/* original file is malformed */
		fprintf(stderr,
		    "%s: error while reading post record marker, aborting\n",
		    fname);
		break;
	    }
	    /* pre and post record markers must match */
	    if (rm_post != rm_pre) {
		fprintf(stderr,
		    "%s: inconsistent pre and post record markers, aborting\n",
		    fname);
		break;
	    }

	    /* write post record marker */
	    ret = write_record_marker(newfile, mode, rm_post);
	    if (ret != STATUS_OK) {
		fprintf(stderr,
		    "%s: error while writing to file\n",
		    fname_new);
		break;
	    }

	} else {
	    /* there were errors when reading/writing the data */
	    break;
	}
    }

    /* if something went wrong, remove the new file;
     * else rename the files */
    if (ok) {
        ret = rename(fname, fname_bak);
        if (ret) {
            /* couldn't rename original file */
            perror(fname);
            fprintf(stderr, "%s: cannot make backup, aborting\n", fname);
            ret = unlink(fname_new);
            if (ret) perror(fname_new);
        } else {
            ret = rename(fname_new, fname);
            if (ret) {
                /* couldn't rename the new file */
                perror(fname_new);
                fprintf(stderr, "%s: cannot rename new file -- YOU ARE SCREWED!\n", fname_new);
                /* let's keep it anyway... */
            }
        }
    } else {
	ret = unlink(fname_new);
	if (ret) perror(fname_new);
    }

    fclose(newfile);
    fclose(origfile);
    return;
}



/* read a single record marker from file */
int read_record_marker(FILE* f, int mode, int32_t* rm) {
    int32_t rm_4;
    int64_t rm_8;
    size_t  nread;
    int     status;

    switch(mode) {
      case CONV_4TO8:
	/* read 4 bytes */
	nread = fread(&rm_4, 4, 1, f);
	if (nread != 1) {
	    *rm = 1;
	    status = STATUS_EOF;
	} else {
	    *rm = rm_4;
	    status = STATUS_OK;
	}
	break;

      case CONV_8TO4:
	/* read 8 bytes */
	nread = fread(&rm_8, 8, 1, f);
	if (nread != 1) {
	    *rm = 1;
	    status = STATUS_EOF; 
	} else {
	    if (rm_8 > 2147483647L || rm_8 < 0) {
		/* record too big or nonsense */
		*rm = 1;
		status = STATUS_INVALID;
	    } else {
		*rm = rm_8;
		status = STATUS_OK;
	    }
	}
	break;

      default:
	/* huh? TODO */
	*rm = 1;
	status = STATUS_INVALID;
    }

    if (*rm < 0) {
	status = STATUS_INVALID;
    }

    return status;
}



/* write a single record marker to file */
int write_record_marker(FILE* f, int mode, int32_t rm) {
    int32_t rm_4;
    int64_t rm_8;
    size_t nwritten;
    int status;

    switch(mode) {
      case CONV_4TO8:
	/* write 8 bytes */
	rm_8 = rm;
	nwritten = fwrite(&rm_8, 8, 1, f);
	if (nwritten != 1)
	    status = STATUS_ERROR;
	else
	    status = STATUS_OK;
	break;

      case CONV_8TO4:
	/* write 8 bytes */
	rm_4 = rm;
	nwritten = fwrite(&rm_4, 4, 1, f);
	if (nwritten != 1)
	    status = STATUS_ERROR;
	else
	    status = STATUS_OK;
	break;

      default:
	/* huh? TODO */
        status = STATUS_ERROR;
	break;
    }

    return status;
}
