#! /usr/bin/env python2
# vim: set ts=4 sw=4 expandtab :

# Read (from stdin) the auto-generated dependencies from the "compile -g" run
# and write (to stdout) a list of Makefile rules which can be included in the
# Quantics compile run.
# -- FO 06/2018

import os
import sys
from subprocess import Popen,PIPE

quantics_dir = os.environ["QUANTICS_DIR"]
quantics_platform = os.environ["QUANTICS_PLATFORM"]

quantics_src = quantics_dir + "/HDmctdh"                                 # PATH_SRC
quantics_obj = quantics_dir + "/object/" + quantics_platform + "/mkdep"  # PATH_OBJ
quantics_mod = quantics_obj + "/include"                                 # PATH_MOD

suffix_o = ".o"
suffix_a = ".a"

class Rule:
    def cleandeps(self):
        """Remove dependencies on self and system."""
        newdeps = []
        for dep in self.deps:
            # skip dependencies on self
            if dep in self.targets: continue
            # skip dependencies on source file, covered by other rules
            if is_src(dep): continue
            # skip dependencies on system files
            if dep.endswith("/mpi.mod"): continue
            newdeps.append(dep)
        self.deps = newdeps
    def modtargets(self):
        """Return list of module targets."""
        return filter(lambda s: s.endswith(".mod"), self.targets)
    def prn_deps(self):
        """Print the list of processed dependencies."""
        deps = map(sub_mod, self.deps)
        deps = map(sub_src, deps)
        print "  " + " \\\n  ".join(deps)
    def prn_modrule(self):
        """Print the dummy rule for the module targets."""
        # cf https://gcc.gnu.org/bugzilla/show_bug.cgi?id=47495
        print " ".join(map(sub_mod, self.mods)) + ": " + self.obj()
        print "\t@true"
    def prn(self):
        """Print Makefile recipe for this rule."""
        self.mods = self.modtargets()
        if self.deps:
            print self.obj() + ": \\"
            self.prn_deps()
        if self.mods:
            self.prn_modrule()
        if self.deps or self.mods:
            print


class ArchiveRule(Rule):
    def __init__(self, archive, path, member, targets, deps):
        self.archive = archive
        self.path = path
        self.member = member
        self.targets = targets
        self.deps = rmdups(deps)
        self.cleandeps()
    def obj(self):
        return sub_ver(sub_obj(self.archive)) + "(" + sub_src(self.path) + "/" + self.member + ")"


class ObjectRule(Rule):
    def __init__(self, obj, targets, deps):
        self.object = obj
        self.targets = targets
        self.deps = rmdups(deps)
        self.cleandeps()
    def obj(self):
        return sub_obj(sub_ver(self.object)) 



def is_src(x):
    for ending in [".f", ".F", ".f90", ".F90"]:
        if x.endswith(ending):
            return True
    return False

def add_mod(x):
    return "$(PATH_MOD)/" + x

def sub_src(x):
    return x.replace(quantics_src, "$(PATH_SRC)")

def sub_obj(x):
    return x.replace(quantics_obj, "$(PATH_OBJ)")

def sub_mod(x):
    return x.replace(quantics_mod, "$(PATH_MOD)")

def sub_ver(x):
    if x.endswith(suffix_o):
        return x[:-len(suffix_o)] + "$(VER).o"
    elif x.endswith(suffix_a):
        return x[:-len(suffix_a)] + "$(VER).a"
    else:
        prnerr("sub_ver called on '%s' which is neither .o nor .a!" % x)
        sys.exit(1)


def prnerr(msg):
    """Print an error message to stdout."""
    sys.stderr.write(msg+"\n")


def rmdups(xs):
    """Removes duplicates from a list."""
    ys = []
    last = None
    for x in sorted(xs):
        if x != last:
            ys.append(x)
        last = x
    return ys


def split2(line):
    try:
        tag, value = line.split(None, 1)
    except ValueError:
        tag, value = line, None
    return (tag, value)


def ppdeps(io):
    """Post-process dependencies, read from file _io_."""
    rules = []
    group = []
    for line in io:
        line = line.rstrip()
        if line == "":
            rule = ppgroup(group)
            rules.append(rule)
            group = []
        else:
            group.append(line)
    return rules


def ppgroup(lines):
    """Post-process one rule, stored in _lines_ (list of strings)."""
    tag1, target = split2(lines[0])
    tag2, path   = split2(lines[1])
    tag3, member = split2(lines[2])
    # check for properly formed input
    if tag1 != "TARGET" or tag2 != "PATH" or tag3 != "MEMBER":
        prnerr("Malformed input!")
        sys.exit(1)
    # remove line continuation markers
    rule = " ".join(lines[3:]).replace("\\", "")
    # split rule into targets and dependencies
    targets = []
    deps = []
    intargets = True
    for item in rule.split():
        if intargets:
            if item.endswith(":"):
                # last target
                item = item[:-1] # remove trailing colon
                intargets = False
            targets.append(item)
        else:
            deps.append(item)
    # create a Rule object
    if member is None:
        # double-check
        if not path is None:
            prnerr("Encountered PATH without MEMBER!")
            sys.exit(1)
        rule = ObjectRule(target, targets, deps)
    else:
        rule = ArchiveRule(target, path, member, targets, deps)
    return rule


if __name__ == "__main__":
    # first argument is VER
    argc = len(sys.argv)
    if argc > 1:
        ver = sys.argv[1]
        suffix_o = "%s.o" % ver
        suffix_a = "%s.a" % ver
        quantics_mod = quantics_obj + "/include%s" % ver
    # process rules from stdin to stdout
    rules = ppdeps(sys.stdin)
    for rule in rules:
        rule.prn()

