
C **********************************************************************
C *                                                                    *
C *                          DAMPING (damping.f)                       *
C *                                                                    *
C * Program package containing various damping functions.              *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                           FUNCTION CDAMPFUNC
C
C PURPOSE:
C     Selects and computes a damp function (p in the PhD thesis).
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     damptype: selector for the damp function having the meaning:
C             1: cosine damping
C             2: squared cosine damping
C             3: box damping
C             4: sine damping
C             5: squared sine damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function cdampfunc (time, damptime, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime

      logical initial
      real*8  pi, halfpi

      save initial, pi, halfpi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      cdampfunc = 0.0d0
      if (initial) then
         pi      = 4.0d0*atan(1.0d0)
         halfpi  = 0.5d0*pi
         initial = .false.
      endif

C --- SELECT AND COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 1) then
         if (time .lt. 0.0d0 .or. time .ge. damptime) return
         cdampfunc = cos(halfpi*time/damptime)
      elseif (damptype .eq. 2) then
         if (time .lt. 0.0d0 .or. time .ge. damptime) return
         cdampfunc = (cos(halfpi*time/damptime))**2
      elseif (damptype .eq. 3) then
         if (time .lt. 0.0d0 .or. time .ge. damptime) return
         cdampfunc = 1.0d0
      elseif (damptype .eq. 4) then
         if (time .lt. 0.0d0 .or. time .ge. damptime) return
         cdampfunc = sin(pi*time/damptime)
      elseif (damptype .eq. 5) then
         if (time .lt. 0.0d0 .or. time .ge. damptime) return
         cdampfunc = sin(pi*time/damptime)**2
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                           FUNCTION CDAMPFOUR
C
C PURPOSE:
C     Selects and computes the Fourier transform of a damp function
C     (\hat{p} in the PhD thesis).
C INPUT PARAMETERS:
C     energy: argument for the Fourier transform of the damp function
C     damptime: time defining the width of the damp function
C     damptype: selector for the damp function having the meaning:
C             1: cosine damping
C             2: squared cosine damping
C             3: box damping
C             4: sine damping
C             5: squared sine damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen function at energy "energy".
C ----------------------------------------------------------------------

      complex*16 function cdampfour (energy, damptime, damptype, error)

      implicit none

      integer damptype, error
      real*8  energy, damptime

      logical initial
      real*8  arg, halfarg, arg2, pi, pi2, dsinc

      save initial, pi, pi2
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error    = 0
      cdampfour = (0.0d0,0.0d0)
      arg      = energy*damptime
      halfarg  = 0.5d0*arg
      arg2     = arg**2
      if (initial) then
         pi      = 4.0d0*atan(1.0d0)
         pi2     = pi**2
         initial = .false.
      endif

C --- SELECT DAMP FUNCTION AND COMPUTE ITS FOURIER TRANSFORM ---

      if (damptype .eq. 1) then
         cdampfour =
     +      dcmplx(-0.5d0*pi*damptime*cos(arg)/(arg2-0.25d0*pi2),
     +             (arg*damptime-0.5d0*pi*damptime*sin(arg))/
     +             (arg2-0.25d0*pi2))
      elseif (damptype .eq. 2) then
         cdampfour = dcmplx(-0.5d0*pi2*damptime*dsinc(arg)/(arg2-pi2),
     +                     (arg*damptime-0.5d0*pi2*damptime*sin(halfarg)
     +                      *dsinc(halfarg))/(arg2-pi2))
      elseif (damptype .eq. 3) then
         cdampfour = dcmplx(damptime*dsinc(arg),
     +                     damptime*sin(halfarg)*dsinc(halfarg))
      elseif (damptype .eq. 4) then
         cdampfour =
     +      dcmplx(-pi*damptime*(cos(arg)+1.0d0)/(arg2-pi2),
     +             -pi*damptime*sin(arg)/(arg2-pi2))
      elseif (damptype .eq. 5) then
         cdampfour = dcmplx(-2.0d0*pi2*damptime*dsinc(arg)
     +                      /(arg2-4.0d0*pi2),
     +                     -2.0d0*pi2*damptime*sin(halfarg)
     +                      *dsinc(halfarg)/(arg2-4.0d0*pi2))
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION CDAMP
C
C PURPOSE:
C     Selects and computes a (transformed) damp function (P in the PhD
C     thesis). Needed for the representation where all autocorrelation
C     functions c0(t), c1(t), and c2(t) are used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C            4: sine damping
C            5: squared sine damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function cdamp (time, damptime, deltae, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime, deltae
      
      logical initial
      real*8  dsinc, arg1, arg2, arg3, arg4, arg5, pi, omegat, ratio,
     +        lambda, cosi

      save initial, pi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi = 4.0d0*atan(1.0d0)
         initial = .false.
      endif
      cdamp = 0.0d0
      ratio = 0.5d0*time/damptime
      lambda = min(ratio,1.0d0-ratio)
      omegat = deltae*damptime

C --- SELECT DAMP FUNCTION ---

      if (damptype .eq. 1) then
         if (lambda .lt. 0.0d0) return
         arg1 = omegat*lambda
         arg2 = (pi-omegat)*lambda
         arg3 = (pi+omegat)*lambda
         cdamp = 0.5d0*lambda*damptime*(2.0d0*cos(pi*ratio)
     +          *dsinc(arg1)+dsinc(arg2)+dsinc(arg3))
      elseif (damptype .eq. 2) then
         if (lambda .lt. 0.0d0) return
         arg1 = omegat*lambda
         arg2 = (pi-omegat)*lambda
         arg3 = (pi+omegat)*lambda
         arg4 = (2.0d0*pi-omegat)*lambda
         arg5 = (2.0d0*pi+omegat)*lambda
         cosi = cos(pi*ratio)
         cdamp = 0.5d0*lambda*damptime*((cosi**2+0.5d0)*dsinc(arg1)
     +          +cosi*(dsinc(arg2)+dsinc(arg3))
     +          +0.25d0*(dsinc(arg4)+dsinc(arg5)))
      elseif (damptype .eq. 3) then
         if (lambda .lt. 0.0d0) return
         arg1 = omegat*lambda
         cdamp = 2.0d0*lambda*damptime*dsinc(arg1)
      elseif (damptype .eq. 4) then
         if (lambda .lt. 0.0d0) return
         arg1 = omegat*lambda
         arg2 = (2.0d0*pi-omegat)*lambda
         arg3 = (2.0d0*pi+omegat)*lambda
         cdamp = 0.5d0*lambda*damptime*(-2.0d0*cos(2.0d0*pi*ratio)
     +          *dsinc(arg1)+dsinc(arg2)+dsinc(arg3))
      elseif (damptype .eq. 5) then
         if (lambda .lt. 0.0d0) return
         arg1 = omegat*lambda
         arg2 = (2.0d0*pi-omegat)*lambda
         arg3 = (2.0d0*pi+omegat)*lambda
         arg4 = (4.0d0*pi-omegat)*lambda
         arg5 = (4.0d0*pi+omegat)*lambda
         cosi = cos(2.0d0*pi*ratio)
         cdamp = 0.5d0*lambda*damptime*((cosi**2+0.5d0)*dsinc(arg1)
     +          -cosi*(dsinc(arg2)+dsinc(arg3))
     +          +0.25d0*(dsinc(arg4)+dsinc(arg5)))
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION CKAMP1
C
C PURPOSE:
C     Computes the first (transformed) damp function (F1 in the PhD
C     thesis). Needed for the representation where only the
C     autocorrelation function c0(t) is used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function ckamp1 (time, damptime, deltae, damptype, error)

      implicit none

      real*8 tiny
      parameter (tiny = 1.0d-12)

      integer damptype, error
      real*8  time, damptime, deltae

      logical initial
      real*8  tbar, ratio, pitime, dsinc, arg1, arg2, arg3, pi, halfpi,
     +        kappa

      save initial, pi, halfpi
      data initial /.true./

C!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
C THIS ROUTINE HAS NOT BEEN REWRITTEN FOR COMPLEX SYMMETRIC HAMILTONIANS
C YET!
C!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi      = 4.0d0*atan(1.0d0)
         halfpi  = 0.5d0*pi
         initial = .false.
      endif
      ckamp1  = 0.0d0
      ratio  = time/damptime
      pitime = pi/damptime
      tbar   = damptime-0.5d0*time
      kappa  = tbar/damptime

C --- COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 1) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         ckamp1 = -pi*kappa*sin(halfpi*ratio)*dsinc(arg1)
      elseif (damptype .eq. 2) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         ckamp1 = -halfpi*kappa*(sin(pi*ratio)*dsinc(arg1)
     +           +sin(halfpi*ratio)*(dsinc(arg2)+dsinc(arg3)))
      elseif (damptype .eq. 3) then
         if ((ratio .ge. 2.0d0) .or. (ratio .le. tiny)) return
         arg1  = deltae*tbar
         ckamp1 = -2.0d0*cos(arg1)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION CKAMP2
C
C PURPOSE:
C     Computes the second (transformed) damp function (F2 in the PhD
C     thesis). Needed for the representation where only the
C     autocorrelation function c0(t) is used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function ckamp2 (time, damptime, deltae, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime, deltae

      logical initial
      real*8  tbar, ratio, pitime, dsinc, arg1, arg2, arg3, arg4, arg5,
     +        pi, halfpi, kappa

      save initial, pi, halfpi
      data initial /.true./

C!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
C THIS ROUTINE HAS NOT BEEN REWRITTEN FOR COMPLEX SYMMETRIC HAMILTONIANS
C YET!
C!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi = 4.0d0*atan(1.0d0)
         halfpi  = 0.5d0*pi
         initial = .false.
      endif
      ckamp2  = 0.0d0
      ratio  = time/damptime
      pitime = pi/damptime
      tbar   = damptime-0.5d0*time
      kappa  = tbar/damptime

C --- COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 1) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         ckamp2 = 
     +           halfpi/damptime*kappa*(pi*cos(halfpi*ratio)*dsinc(arg1)
     +           +(deltae*damptime-halfpi)*dsinc(arg2)
     +           -(deltae*damptime+halfpi)*dsinc(arg3))
      elseif (damptype .eq. 2) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         arg4  = (2.0d0*pitime-deltae)*tbar
         arg5  = (2.0d0*pitime+deltae)*tbar
         ckamp2 = 0.25d0*pi/damptime*kappa*(
     +             2.0d0*pi*cos(pi*ratio)*dsinc(arg1)
     +            +2.0d0*deltae*damptime*cos(halfpi*ratio)
     +             *(dsinc(arg2)-dsinc(arg3))
     +            +(deltae*damptime-pi)*dsinc(arg4)
     +            -(deltae*damptime+pi)*dsinc(arg5))
      elseif (damptype .eq. 3) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         ckamp2 = 2.0d0*deltae*sin(arg1)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end
