
C **********************************************************************
C *                                                                    *
C *                          OUTPUT (output.f)                         *
C *                                                                    *
C * Program package containing the subroutines that write the data to  *
C * files.                                                             *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                          SUBROUTINE WRITEPARAM
C
C PURPOSE:
C     Writes the input parameters to the output file.
C INPUT PARAMETERS:
C     outch: channel connected to output file
C     path: path to directory containing the output files
C     outname: name of the output files
C     hdim: size of reduced Hamiltonian matrix
C     edim: number of energy points
C     autodim: length of autocorrelation function
C     eminin: lower boundary of energy window
C     emaxin: upper boundary of energy window
C     damptype: selector for damping function
C     autocorr: 0: only zero order autocorrelation function is used
C               1: also higher order autocorrelation functions are used
C     damptime: damping time
C     proptime: propagation period
C     inttime: integration period
C     hamshift: energy shift of the Hamiltonian
C     vartype: selects the variational principle
C     sigma: energy shift used for the variational principle
C     sigma2: same as "sigma", but for second variational principle
C     errortype: selector for the different internal error checks
C     vartype2: variational principle used for internal error control
C               (must be different from "vartype"; 0 means none)
C     epscut: cut-off parameter for eigenvalues of overlap matrix
C     usedcmemory: length of "giant" complex*16 memory array
C     usedrmemory: length of "giant" real*8     memory array
C     usedimemory: length of "giant" integer    memory array
C     econv: conversion for energy unit (output is divided by econv)
C     tconv: conversion for time unit (output is divided by tconv)
C     usepack: number of packets to be employed
C     deltat: time gap of autocorrelation function
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C     intenspack: number of the packet for which the intensity is
C                 computed.
C     imagshift: shift for the imaginary part of the eigenvalues, i.e.
C                c(t) is replaced by e^(imagshift/2*t)*c(t)
C     inittime: initial time of c(t) being used
C OUTPUT PARAMETERS:
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open log file
C            3: illegal damp function
C            4: illegal variational principle
C ----------------------------------------------------------------------

      subroutine cwriteparam (outch, path, outname, hdim, edim, autodim,
     +                       eminin, emaxin, damptype, autocorr,
     +                       damptime, proptime, inttime, hamshift,
     +                       vartype, sigma, sigma2, errortype,
     +                       vartype2, epscut, usedcmemory, usedrmemory,
     +                       usedimemory, econv, tconv, usepack, deltat,
     +                       refener, intenspack, imagshift, inittime,
     +                       error)

      implicit none

      real*8  kilobyte
      parameter (kilobyte = 1024.0d0)

      integer      outch, edim, hdim, autodim, damptype, autocorr,
     +             vartype, errortype, vartype2, usedcmemory,
     +             usedrmemory, usedimemory, usepack, intenspack, error
      real*8       eminin, emaxin, damptime, proptime, inttime,
     +             hamshift, sigma, sigma2, epscut, econv, tconv,
     +             deltat, refener, imagshift, inittime
      character*80 path, outname

      integer pathlength, namelength, complexsize, realsize, integersize

C --- INITIALISE VARIABLES ---

      error = 0
      pathlength = index(path,' ')-1
      namelength = index(outname,' ')-1

C --- OPEN LOG FILE ---

      open(outch,
     +     file=path(1:pathlength)//'/'//outname(1:namelength)//'.log',
     +     form='formatted',status='unknown',err=20)
      write(outch,*,err=10) 
      write(outch,'(a54)',err=10) '-------- FDC OUTPUT FILE --------'
      write(outch,*,err=10) 

C --- GET WRITE VERSION TO LOG FILE ---

      call wrversionf77(outch)     

C --- WRITE INPUT DATA TO OUTPUT FILE ---

      write(outch,'(a40,i11)',err=10)
     +   'Length of complex*16 array: ', usedcmemory
      write(outch,'(a40,i11)',err=10)
     +   'Length of real*8 array: ', usedrmemory
      write(outch,'(a40,i11)',err=10)
     +   'Length of integer array: ', usedimemory
      write(outch,'(a40,i9)',err=10) 'Size of memory (in KByte): ',
     +   nint((complexsize()*usedcmemory+realsize()*usedrmemory
     +        +integersize()*usedimemory)/kilobyte)
      write(outch,*,err=10)
      write(outch,'(a40,f17.10)',err=10)
     +   'Lower boundary of energy window: ', (eminin-refener)/econv
      write(outch,'(a40,f17.10)',err=10)
     +    'Upper boundary of energy window: ', (emaxin-refener)/econv
      write(outch,'(a43,i3)',err=10) 'Number of energy points:    ',
     +   edim
      write(outch,'(a40,f17.10)',err=10) 'Reference energy: ',
     +   refener/econv
      write(outch,*,err=10)
      write(outch,'(a43,i3)',err=10) 'Number of packets:    ',
     +   usepack
      write(outch,'(a43,i3)',err=10) 'Packet of intensity:    ',
     +   intenspack
      write(outch,'(a40,f17.10)',err=10) 'Propagation period: ',
     +   proptime/tconv
      write(outch,'(a40,f17.10)',err=10) 'Initial time: ',
     +   inittime/tconv
      write(outch,'(a40,f17.10)',err=10) 'Integration period: ',
     +   inttime/tconv
      write(outch,'(a40,f17.10)',err=10) 'Output interval: ',
     +   deltat/tconv
      write(outch,'(a43,i5)',err=10) 'Length of autocorrelation:    ',
     +   autodim
      write(outch,'(a40,f17.10)',err=10) 'Hamiltonian energy shift: ',
     +   hamshift/econv
      write(outch,'(a40,f17.10)',err=10) 'Hamiltonian width shift: ',
     +   imagshift/econv
      write(outch,*,err=10)
      if (damptype .eq. 1) then
         write(outch,'(a40,a9)',err=10) 'Weight function: ', 'cosine'
         write(outch,'(a40,f17.10)',err=10)
     +      'Weight function equals zero at time: ',
     +      2.0d0*damptime/tconv
      elseif (damptype .eq. 2) then
         write(outch,'(a40,a17)',err=10) 'Weight function: ',
     +      'squared cosine'
         write(outch,'(a40,f17.10)',err=10)
     +      'Weight function equals zero at time: ',
     +      2.0d0*damptime/tconv
      elseif (damptype .eq. 3) then
         write(outch,'(a40,a6)',err=10) 'Weight function: ', 'box'
         write(outch,'(a40,f17.10)',err=10)
     +      'Weight function equals zero at time: ',
     +      2.0d0*damptime/tconv
      elseif (damptype .eq. 4) then
         write(outch,'(a40,a9)',err=10) 'Weight function: ', 'sine'
         write(outch,'(a40,f17.10)',err=10)
     +      'Weight function equals zero at time: ',
     +      2.0d0*damptime/tconv
      elseif (damptype .eq. 5) then
         write(outch,'(a40,a17)',err=10) 'Weight function: ',
     +      'squared sine'
         write(outch,'(a40,f17.10)',err=10)
     +      'Weight function equals zero at time: ',
     +      2.0d0*damptime/tconv
      else
         error = 3
         return
      endif
      write(outch,*,err=10) 
      if (autocorr .eq. 1) then
         write(outch,'(a40,a26)',err=10)
     +      'Functions used: ','c0(t), c1(t), and c2(t)'
      elseif (autocorr .eq. 0) then
         write(outch,'(a40,a13)',err=10)
     +      'Functions used: ','c0(t) only'
      else
         error = 3
         return
      endif
      write(outch,*,err=10) 
      write(outch,'(a43,i3)',err=10)
     +   'Size of full Hamiltonian:    ', edim*usepack
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Cut-off for overlap eigenvalues: ', epscut
      write(outch,'(a43,i3)',err=10)
     +   'Size of reduced Hamiltonian:    ', hdim
      write(outch,*,err=10) 
      if (vartype .eq. 1) then
         write(outch,'(a40,a26)',err=10) 'Variational principle: ',
     +      '<psi|H|psi> / <psi|psi>'
      elseif (vartype .eq. 2) then
         write(outch,'(a40,a32)',err=10) 'Variational principle: ',
     +      '<psi|(H-s)^2|psi> / <psi|psi>'
         write(outch,'(a40,f17.10)',err=10)
     +      'Shift for variational principle: ', (sigma-refener)/econv
      elseif (vartype .eq. 3) then
         write(outch,'(a40,a36)',err=10) 'Variational principle: ',
     +      '<psi|H-s|psi> / <psi|(H-s)^2|psi>'
         write(outch,'(a40,f17.10)',err=10)
     +      'Shift for variational principle: ', (sigma-refener)/econv
      else
         error = 4
         return
      endif
      if (errortype .eq. 0) then
         write(outch,'(a40,a26)',err=10) 'Internal error: ',
     +     'sqrt(<psi|(H-E)^2|psi>)'
      elseif (errortype .eq. 1) then
         if (vartype2 .eq. 1) then
            write(outch,'(a40,a26)',err=10)
     +        'Second variational principle: ','<psi|H|psi> / <psi|psi>'
         elseif (vartype2 .eq. 2) then
            write(outch,'(a40,a32)',err=10)
     +        'Second variational principle: ',
     +        '<psi|(H-s)^2|psi> / <psi|psi>'
            write(outch,'(a40,f17.10)',err=10)
     +         'Shift for 2nd variational principle: ',
     +         (sigma2-refener)/econv
         elseif (vartype2 .eq. 3) then
            write(outch,'(a40,a36)',err=10)
     +        'Second variational principle: ',
     +        '<psi|H-s|psi> / <psi|(H-s)^2|psi>'
            write(outch,'(a40,f17.10)',err=10)
     +         'Shift for 2nd variational principle: ',
     +         (sigma2-refener)/econv
         else
            error = 4
            return
         endif
      endif
      write(outch,*,err=10)

C --- RETURN TO CALLING ROUTINE ---

      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                          SUBROUTINE WRITEENERGY
C
C PURPOSE:
C     Writes the results (i.e. the energies and intensities) to the
C     output file.
C INPUT PARAMETERS:
C     outch: channel connected to output file
C     path: path to directory containing the output files
C     outname: name of the output files
C     eigenval: eigenvalue vector (is sorted and energy-converted on
C               output)
C     eigenerror: vector containing error estimates for the eigenvalues
C                 (is sorted and energy-converted on output)
C     intensity: vector containing the intensities of the eigenvalues
C     exactenergy: vector containing exact energies (sorted on output)
C     exactintensity: vector containing exact intensities
C     hdim: size of reduced Hamiltonian matrix
C     maxauxdim: leading dimension of auxiliary matrix "auxmat"
C     exactdim: number of exact energies inside window
C     eminin: lower boundary of energy window
C     emaxin: upper boundary of energy window
C     errortype: selector for the different internal error checks
C     exactflag: 0: no file with exact results have been read
C                1: the exact results have been read from file
C     spurcut: cut-off for spurious eigenvalues (all eigenvalues with
C              internal error I greater than spurcut*I_max are deleted)
C     econv: conversion for energy unit (output is divided by econv)
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C     imagshift: shift for the imaginary part of the eigenvalues, i.e.
C                c(t) is replaced by e^(imagshift/2*t)*c(t)
C OUTPUT PARAMETERS:
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open '.eig' file
C OTHER PARAMETERS:
C     auxmat,auxvec: auxiliary arrays needed to sort the eigenvalues
C     assignforw: integer array with assignments from computed to exact
C                 eigenvalues
C     assignback: integer array with assignments from exact to computed
C                 eigenvalues
C ----------------------------------------------------------------------

      subroutine cwriteenergy (outch, path, outname, eigenval,
     +                        eigenerror, intensity, intenserr,
     +                        exactenergy, exactintensity, hdim,
     +                        maxauxdim, exactdim, eminin, emaxin,
     +                        errortype, exactflag, spurcut, econv,
     +                        refener, imagshift, error, auxmat, auxvec,
     +                        assignforw, assignback)

      use op1lib

      implicit none

      integer   eigch
      real*8    large
      parameter (eigch = 1, large = 1.0d99)

      integer      outch, hdim, maxauxdim, exactdim, errortype,
     +             exactflag, error, assignforw(1:hdim),
     +             assignback(1:exactdim)
      real*8       eminin, emaxin, spurcut, econv, refener, imagshift
      complex*16   eigenval(1:hdim), eigenerror(1:hdim),
     +             intensity(1:hdim), intenserr(1:hdim),
     +             exactenergy(1:exactdim), exactintensity(1:exactdim),
     +             auxmat(1:maxauxdim,1:6), auxvec(1:hdim)
      character*80 path, outname

      character*1 label
      integer     elower, eupper, eaccu, einteraccu, eunfound,
     +            eigalldim, pathlength, namelength, i, j
      real*8      meanintererr, meanexacterr, medeigerr, medexacterr,
     +            minintererr, maxintererr, minexacterr, maxexacterr,
     +            minintenserr, maxintenserr, minintensexact,
     +            maxintensexact

C --- INITIALISE VARIABLES ---

      error = 0

C --- SORT DATA ---

      do i = 1,hdim
         auxmat(i,1) = eigenval(i)+dcmplx(0.0d0,-0.5d0*imagshift)
         auxmat(i,2) = eigenerror(i)
         auxmat(i,3) = intensity(i)
         if (errortype .ge. 1) auxmat(i,4) = intenserr(i)
      enddo
      if (errortype .ge. 1) then
         call zbubble1(auxmat(1,1),auxmat(1,2),hdim,maxauxdim,3)
      else
         call zbubble1(auxmat(1,1),auxmat(1,2),hdim,maxauxdim,2)
      endif

C --- IF NO EXACT RESULTS ARE KNOWN, THEN OUTPUT ---

      if (exactflag .eq. 0) then
         eigalldim = hdim
         goto 100
      endif

C --- COPY SORTED RESULTS BACK ---

      call cpvxz(auxmat(1,1),eigenval,hdim)
      call cpvxz(auxmat(1,2),eigenerror,hdim)
      call cpvxz(auxmat(1,3),intensity,hdim)
      if (errortype .ge. 1) call cpvxz(auxmat(1,4),intenserr,hdim)

C --- FIND BEST MATCHES BETWEEN THE TWO LISTS OF EIGENVALUES ---
      
      call zmatch(eigenval,exactenergy,hdim,exactdim,assignforw,
     +            assignback)

C --- STORE EIGENVALUES THAT HAVE BEEN MATCHED ---
      
      eigalldim = 0
      do i = 1,hdim
         if (assignforw(i) .ne. 0) then
            eigalldim = eigalldim+1
            auxmat(eigalldim,1) = eigenval(i)
            auxmat(eigalldim,2) = eigenerror(i)
            auxmat(eigalldim,3) =
     +         dcmplx(abs(dble(eigenval(i))
     +                   -dble(exactenergy(assignforw(i)))),0.0d0)
            auxmat(eigalldim,4) = intensity(i)
            auxmat(eigalldim,5) =
     +         dcmplx(abs(dble(intensity(i))
     +                   -dble(exactintensity(assignforw(i)))),0.0d0)
            if (errortype .ge. 1) auxmat(eigalldim,6) = intenserr(i)
         endif
      enddo

C --- STORE SPURIOUS EIGENVALUES ---

      do i = 1,hdim
         if (assignforw(i) .eq. 0) then
            eigalldim = eigalldim+1
            auxmat(eigalldim,1) = eigenval(i)
            auxmat(eigalldim,2) = eigenerror(i)
            auxmat(eigalldim,3) = (0.0d0,0.0d0)
            auxmat(eigalldim,4) = intensity(i)
            auxmat(eigalldim,5) = (0.0d0,0.0d0)
            if (errortype .ge. 1) auxmat(eigalldim,6) = intenserr(i)
         endif
      enddo

C --- STORE MISSING EXACT EIGENVALUES ---

      do j = 1,exactdim
         if (assignback(j) .eq. 0) then
            eigalldim = eigalldim+1
            auxmat(eigalldim,1) = exactenergy(j)
            auxmat(eigalldim,2) = (0.0d0,0.0d0)
            auxmat(eigalldim,3) = (0.0d0,0.0d0)
            auxmat(eigalldim,4) = exactintensity(j)
            auxmat(eigalldim,5) = (0.0d0,0.0d0)
            if (errortype .ge. 1) auxmat(eigalldim,6) = (0.0d0,0.0d0)
         endif
      enddo

C --- ENSURE THAT ENERGIES OUTSIDE WINDOW HAVE EXACT ERROR OF ZERO ---

      do i = 1,eigalldim
         if ((dble(auxmat(i,1)) .lt. eminin) .or.
     +       (dble(auxmat(i,1)) .gt. emaxin)) then
            auxmat(i,3) = (0.0d0,0.0d0)
         endif
      enddo

C --- SORT THE DATA ---

      if (errortype .ge. 1) then
         call zbubble1(auxmat(1,1),auxmat(1,2),eigalldim,maxauxdim,5)
      else
         call zbubble1(auxmat(1,1),auxmat(1,2),eigalldim,maxauxdim,4)
      endif

C --- COUNT ENERGIES THAT LIE OUTSIDE THE WINDOW ---

 100  continue
      elower = 0
      eupper = 0
      do i = 1,eigalldim
         if (dble(auxmat(i,1)) .lt. eminin) elower = elower+1
         if (dble(auxmat(i,1)) .gt. emaxin) eupper = eupper+1
      enddo

C --- COUNT ACCURATE AND UNFOUND ENERGIES INSIDE WINDOW ---

      eaccu      = 0
      einteraccu = 0
      eunfound   = 0
      do i = 1,eigalldim
         if ((dble(auxmat(i,1)) .ge. eminin) .and.
     +       (dble(auxmat(i,1)) .le. emaxin)) then
            if (exactflag .eq. 1) then
               if ((auxmat(i,2) .eq. (0.0d0,0.0d0)) .and.
     +             (auxmat(i,3) .eq. (0.0d0,0.0d0)))
     +            eunfound = eunfound+1
               if (auxmat(i,3) .ne. (0.0d0,0.0d0)) eaccu = eaccu+1
               if (dble(auxmat(i,2)) .gt. 0.0d0)
     +            einteraccu = einteraccu+1
            else
               if (errortype .lt. 1) then
                  if (dble(auxmat(i,2)) .lt. spurcut) eaccu = eaccu+1
               else
                  if (dble(auxmat(i,2)) .gt. 0.0d0) eaccu = eaccu+1
               endif
               einteraccu = eaccu
            endif
         endif
      enddo

C --- COMPUTE MEAN AND EXTREME ERRORS ---

      meanintererr = 0.0d0
      meanexacterr = 0.0d0
      medexacterr = 0.0d0
      minintererr = large
      minexacterr = large
      minintenserr = large
      minintensexact = large
      maxintererr = -large
      maxexacterr = -large
      maxintenserr = -large
      maxintensexact = -large
      do i = 1,eigalldim
         if ((dble(auxmat(i,1)) .ge. eminin) .and.
     +       (dble(auxmat(i,1)) .le. emaxin)) then
            if (exactflag .eq. 0) then
               if (errortype .eq. 0) then
                  if (dble(auxmat(i,2)) .lt. spurcut) then
                     meanintererr = meanintererr+dble(auxmat(i,2))
                     minintererr = min(minintererr,dble(auxmat(i,2)))
                     maxintererr = max(maxintererr,dble(auxmat(i,2)))
                  endif
               else
                  if (dble(auxmat(i,2)) .gt. 0.0d0) then
                     meanintererr = meanintererr+dble(auxmat(i,2))
                     minintererr = min(minintererr,dble(auxmat(i,2)))
                     maxintererr = max(maxintererr,dble(auxmat(i,2)))
                     minintenserr = min(minintenserr,dble(auxmat(i,4)))
                     maxintenserr = max(maxintenserr,dble(auxmat(i,4)))
                  endif
               endif
            else
               if (auxmat(i,3) .ne. (0.0d0,0.0d0)) then
                  if (dble(auxmat(i,2)) .gt. 0.0d0) then
                     meanintererr = meanintererr+dble(auxmat(i,2))
                     minintererr  = min(minintererr,dble(auxmat(i,2)))
                     maxintererr  = max(maxintererr,dble(auxmat(i,2)))
                     if (errortype .ge. 1) then
                        minintenserr =
     +                     min(minintenserr,dble(auxmat(i,6)))
                        maxintenserr =
     +                     max(maxintenserr,dble(auxmat(i,6)))
                     endif
                  endif
                  meanexacterr = meanexacterr+dble(auxmat(i,3))
                  minexacterr = min(minexacterr,dble(auxmat(i,3)))
                  maxexacterr = max(maxexacterr,dble(auxmat(i,3)))
                  minintensexact = min(minintensexact,dble(auxmat(i,5)))
                  maxintensexact = max(maxintensexact,dble(auxmat(i,5)))
               endif
            endif
         endif
      enddo
      if (einteraccu .gt. 0) then
         meanintererr = meanintererr/einteraccu
      else
         meanintererr = large
      endif
      if (exactflag .eq. 1) then
         if (eaccu .gt. 0) then
            meanexacterr = meanexacterr/eaccu
         else
            meanexacterr = large
         endif
      endif

C --- COMPUTE MEDIAN OF INTERNAL ERROR ---

      j = 0
      do i = 1,eigalldim
         if ((dble(auxmat(i,1)) .ge. eminin) .and.
     +       (dble(auxmat(i,1)) .le. emaxin)) then
            if (exactflag .eq. 0) then
               if (errortype .lt. 1) then
                  if (dble(auxmat(i,2)) .lt. spurcut) then
                     j = j+1
                     auxvec(j) = auxmat(i,2)
                  endif
               else
                  if (dble(auxmat(i,2)) .gt. 0.0d0) then
                     j = j+1
                     auxvec(j) = auxmat(i,2)
                  endif
               endif
            else
               if (auxmat(i,3) .ne. (0.0d0,0.0d0)) then
                  j = j+1
                  auxvec(j) = auxmat(i,2)
               endif
            endif
         endif
      enddo
      call zbubble(auxvec,auxmat,j,0)
      if (j .eq. 0) then
         medexacterr = 0.0d0
      elseif (mod(j,2) .eq. 1) then
         medeigerr = dble(auxvec((j+1)/2))
      else
         medeigerr = 0.5d0*(dble(auxvec(j/2))+dble(auxvec(j/2+1)))
      endif

C --- COMPUTE MEDIAN OF EXACT ERROR ---

      if (exactflag .eq. 1) then
         j = 0
         do i = 1,eigalldim
            if (auxmat(i,3) .ne. (0.0d0,0.0d0)) then
               j = j+1
               auxvec(j) = auxmat(i,3)
            endif
         enddo
         call zbubble(auxvec,auxmat,j,0)
         if (j .eq. 0) then
            medexacterr = 0.0d0
         elseif (mod(j,2) .eq. 1) then
            medexacterr = dble(auxvec((j+1)/2))
         else
            medexacterr = 0.5d0*(dble(auxvec(j/2))+dble(auxvec(j/2+1)))
         endif
      endif

C --- WRITE STATISTICS TO OUTPUT FILE ---

      write(outch,'(a43,i3)',err=10) 'Number of energies found:    ',
     +   hdim
      write(outch,'(a43,i3)',err=10)
     +   'Number of energies left of window:    ', elower
      write(outch,'(a43,i3)',err=10)
     +   'Number of energies right of window:    ', eupper
      write(outch,'(a43,i3)',err=10)
     +   'Number of energies inside window:    ',
     +   eigalldim-elower-eupper-eunfound
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Cut-off for spurious eigenvalues: ', spurcut/econv
      write(outch,'(a43,i3)',err=10)
     +   'Number of accurate energies in window:    ', eaccu
      if (exactflag .eq. 1) then
         write(outch,'(a43,i3)',err=10)
     +      'Number of energies not found:    ', eunfound
      endif
      write(outch,*,err=10)
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Average of internal error: ', meanintererr/econv
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Median of internal error: ', medeigerr/econv
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Minimum of internal error: ', minintererr/econv
      write(outch,'(a40,1pe16.6e3)',err=10)
     +   'Maximum of internal error: ', maxintererr/econv
      if (exactflag .eq. 1) then
         write(outch,*,err=10)
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Average of exact error: ', meanexacterr/econv
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Median of exact error: ', medexacterr/econv
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Minimum of exact error: ', minexacterr/econv
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Maximum of exact error: ', maxexacterr/econv
      endif
      if (errortype .ge. 1) then
         write(outch,*,err=10)
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Minimum of internal intensity error: ', minintenserr
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Maximum of internal intensity error: ', maxintenserr
      endif
      if (exactflag .eq. 1) then
         write(outch,*,err=10)
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Minimum of exact intensity error: ', minintensexact
         write(outch,'(a40,1pe16.6e3)',err=10)
     +      'Maximum of exact intensity error: ', maxintensexact
      endif
      write(outch,*,err=10)

C --- OPEN FILE WITH PURE SPECTRUM ---

      pathlength = index(path,' ')-1
      namelength = index(outname,' ')-1
      open(eigch,
     +     file=path(1:pathlength)//'/'//outname(1:namelength)//'.eig',
     +     form='formatted',status='unknown',err=20)
      
C --- GENERATE TABLE OF ENERGIES ---

      if (exactflag .eq. 0) then
         if (errortype .eq. 0) then
            write(outch,'(x,a7,8x,a6,11x,a5,15x,a5,9x,a9)',err=10)
     +         'Number', 'Energy', 'Width', 'Error', 'Intensity'
         else
            write(outch,'(x,a6,7x,a6,11x,a5,7x,a8,5x,a9,4x,a8)',
     +         err=10) 'Number', 'Energy', 'Width', 'Internal',
     +                 'Intensity', 'Internal'
            write(outch,'(44x,a5,21x,a5)',err=10) 'error', 'error'
         endif
      else
         if (errortype .eq. 0) then
            write(outch,'(2x,a3,5x,a6,8x,a5,6x,a8,5x,a5,5x,a9,5x,a5)',
     +         err=10) 'Nr.', 'Energy', 'Width', 'Internal', 'Exact',
     +         'Intensity', 'Exact'
            write(outch,'(37x,a5,6x,a5,19x,a5)',err=10)
     +         'error', 'error', 'error'
         else
            write(outch,
     +         '(x,a6,4x,a6,6x,a5,5x,a8,3x,a5,3x,a9,2x,a8,2x,a5)',
     +          err=10) 'Number', 'Energy', 'Width', 'Internal',
     +          'Exact', 'Intensity', 'Internal', 'Exact'
            write(outch,'(35x,a5,4x,a5,16x,a5,3x,a5)',err=10)
     +         'error', 'error', 'error', 'error'
         endif
      endif
      do i = 1,eigalldim
         if (i .gt. 1) then
            if (((dble(auxmat(i-1,1)) .lt. eminin) .and.
     +           (dble(auxmat(i,1)) .ge. eminin)) .or.
     +          ((dble(auxmat(i-1,1)) .le. emaxin) .and.
     +           (dble(auxmat(i,1)) .gt. emaxin)))
     +         write(outch,'(79a)') ('-', j=1,79)
         else
            if (dble(auxmat(i,1)) .ge. eminin)
     +         write(outch,'(79a)') ('-', j=1,79)
         endif
         label = ' '
         if ((dble(auxmat(i,1)) .ge. eminin) .and.
     +       (dble(auxmat(i,1)) .le. emaxin)) then
            if (errortype .lt. 1) then
               if (dble(auxmat(i,2)) .gt. spurcut) label = '*'
            else
               if (dble(auxmat(i,2)) .le. 0.0d0) label = '*'
            endif
            if (exactflag .eq. 1) then
               if (auxmat(i,3) .eq. (0.0d0,0.0d0)) then
                  if (errortype .lt. 1) then
                     if (dble(auxmat(i,2)) .gt. spurcut) then
                        label = '+'
                     else
                        label = '#'
                     endif
                  else
                     if (dble(auxmat(i,2)) .le. 0.0d0) then
                        label = '+'
                     else
                        label = '#'
                     endif
                  endif
               endif
               if ((auxmat(i,2) .eq. (0.0d0,0.0d0)) .and.
     +             (auxmat(i,3) .eq. (0.0d0,0.0d0))) label = '-'
            endif
            if (label .eq. ' ') then
               write(eigch,'(1p3e25.15e3)')
     +            (dble(auxmat(i,1))-refener)/econv,
     +            -2.0d0*dimag(auxmat(i,1))/econv,
     +            dble(auxmat(i,3+exactflag))
            endif
         endif
         if (exactflag .eq. 0) then
            if (errortype .eq. 0) then
               write(outch,'(a1,x,i4,2x,f18.10,2x,1pe15.7e3,2x,'//
     +            'f18.10,2x,1pe13.7e2)',err=10)
     +            label, i, (dble(auxmat(i,1))-refener)/econv,
     +            -2.0d0*dimag(auxmat(i,1))/econv,
     +            dble(auxmat(i,2))/econv, dble(auxmat(i,3))
            else
               write(outch,'(a1,x,i3,2x,f17.10,2x,1pe14.7e2,2x,'//
     +           '1pe10.3e2,2x,1pe13.7e2,2x,1pe9.3e2)',err=10)
     +           label, i, (dble(auxmat(i,1))-refener)/econv,
     +           -2.0d0*dimag(auxmat(i,1))/econv,
     +           dble(auxmat(i,2))/econv, dble(auxmat(i,3)),
     +           dble(auxmat(i,4))
            endif
         else
            if (errortype .eq. 0) then
               write(outch,'(a1,x,i3,2x,f12.6,2x,1pe12.5e2,'//
     +            '2(x,1pe10.3e2),x,1pe12.5e2,x,1pe10.3e2)',err=10)
     +            label, i, (dble(auxmat(i,1))-refener)/econv,
     +            -2.0d0*dimag(auxmat(i,1))/econv,
     +            dble(auxmat(i,2))/econv, dble(auxmat(i,3))/econv,
     +            dble(auxmat(i,4)), dble(auxmat(i,5))
            else
               write(outch,'(a1,x,i3,2x,f11.5,2x,1pe11.4e2,2x,'//
     +            '1pe8.1e2,2x,1pe7.1e2,2x,1pe9.3e2,2x,1pe7.1e2,'//
     +            '2x,1pe7.1e2)',err=10)
     +            label, i, (dble(auxmat(i,1))-refener)/econv,
     +            -2.0d0*dimag(auxmat(i,1))/econv,
     +            dble(auxmat(i,2))/econv, dble(auxmat(i,3))/econv,
     +            dble(auxmat(i,4)), dble(auxmat(i,6)),
     +            dble(auxmat(i,5))
            endif
         endif
      enddo
      if (dble(auxmat(eigalldim,1)) .lt. eminin)
     +   write(outch,'(79a)') ('-', j=1,79)
      write(outch,*,err=10)
      write(outch,'(a)')
     +   '*: eigenvalue is spurious (according to internal error)'
      if (exactflag .eq. 1) then
         write(outch,'(a)')
     +      '#: eigenvalue is spurious (according to exact results)'
         write(outch,'(a)')
     +      '+: eigenvalue is spurious (according to both criterions)'
         write(outch,'(a)')
     +      '-: eigenvalue is missing (compared with exact results)'
      endif
      write(outch,*,err=10)

C --- CLOSE FILES AND RETURN TO CALLING ROUTINE ---

      close(eigch,err=10)
      close(outch,err=10)
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                         SUBROUTINE WRITEFAUTO
C
C PURPOSE:
C     Computes and writes the Fourier transform of the autocorrelation
C     function to a file.
C INPUT PARAMETERS:
C     reauto0, imauto0: real and imaginary part of 1st autocorrelation
C                       function <dconjg(psi(t))|psi(t)>, respectively
C     autodim: length of autocorrelation vectors
C     path: path to directory the results are written to
C     outname: name of output files
C     damptype: selector for damping function
C     damptime: time defining the width of the damp function
C     deltat: time gap of autocorrelation function
C     eminin, emaxin: lower and upper boundary of energy range (as given
C                     in input file)
C     hamshift: energy shift of the Hamiltonian
C     econv: conversion for energy unit (output is devided by econv)
C     sodflag: 1 if propagation has been done with SOD, else 0
C     damppoints: number of energy grid points
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C OUTPUT PARAMETERS:
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open auto file
C            3: illegal damp function selected
C ----------------------------------------------------------------------

      subroutine cwritefauto (reauto0, imauto0, autodim, path, outname,
     +                       damptype, damptime, deltat, eminin, emaxin,
     +                       hamshift, econv, sodflag, damppoints,
     +                       refener, error)

      implicit none

      integer ftch
      parameter (ftch = 1)

      integer      autodim, damptype, damppoints, sodflag, error
      real*8       reauto0(0:autodim), imauto0(0:autodim), damptime,
     +             deltat, eminin, emaxin, hamshift, econv, refener
      character*80 path, outname

      real*8     time, damp, weight, energyx, energyy, eeffx, eeffy,
     +           cdampfunc
      complex*16 fourier, arg, cdampfour
      integer    pathlength, namelength, i, j, k

C --- INITIALISE VARIABLES ---

      error = 0
      pathlength = index(path,' ')-1
      namelength = index(outname,' ')-1

C --- OPEN AUTOCORRELATION FILE ---

      open(ftch,
     +     file=path(1:pathlength)//'/'//outname(1:namelength)//'.ft',
     +     form='formatted',status='unknown',err=20)

C --- LOOP OVER ALL ENERGY POINTS ---

C The factor 2 ensures that the full available autocorrelation is used.

      damp = 2.0d0*damptime
      do i = 1,damppoints
         do j = 1,damppoints

C       --- DETERMINE ENERGY ---

            energyx = eminin+(i-1)*(emaxin-eminin)/(damppoints-1)
            energyy = eminin+(j-1)*(emaxin-eminin)/(damppoints-1)
            if (sodflag .eq. 1) then
               eeffx = asin(0.5d0*(energyx-hamshift)*deltat)/
     +                 (0.5d0*deltat)
               eeffy = asin(0.5d0*energyy*deltat)/(0.5d0*deltat)
            else
               eeffx = energyx-hamshift
               eeffy = energyy
            endif

C       --- COMPUTE AND OUTPUT FOURIER TRANSFORM ---

            fourier = (0.0d0,0.0d0)
            do k = 0,autodim
               time   = k*deltat
               arg    = dcmplx(-eeffy*time,eeffx*time)
               weight = cdampfunc(time,damp,damptype,error)
               if (error .ne. 0) goto 30
               if ((k .eq. 0) .or. (k .eq. autodim))
     +            weight = 0.5d0*weight
               fourier = fourier
     +            +weight*dcmplx(reauto0(k),imauto0(k))*exp(arg)
            enddo
            fourier = deltat*fourier
     +                /cdampfour(0.0d0,damptime,damptype,error)
            if (error .ne. 0) goto 30
            write(ftch,'(1p5e15.5e2)',err=10)
     +         (energyx-refener)/econv, -2.0d0*energyy/econv,
     +         dble(fourier), dimag(fourier), abs(fourier)
         enddo
      enddo

C --- CLOSE FILE AND RETURN TO CALLING ROUTINE ---

      close(ftch)
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

 30   continue
      error = 3
      return

      end

C ----------------------------------------------------------------------
C                          SUBROUTINE ERRORMSG
C
C PURPOSE:
C     Sends an error message to the screen and stops the program.
C INPUT PARAMETERS:
C     routine: number specifying the subroutine where the error occured
C     error: number describing the error
C OUTPUT PARAMETERS:
C     none
C ----------------------------------------------------------------------

      subroutine fcerrormsg (routine, error)

      implicit none

      integer maxroutine, maxerror
      parameter (maxroutine = 20, maxerror = 20)

      integer routine, error

      integer      i, j
      character*64 message(maxroutine,maxerror)

C --- INITIALISE ERROR MESSAGES ---

      do j = 1,maxerror
         do i = 1,maxroutine
            message(i,j) = 'unknown error occured (errormsg)'
         enddo
      enddo

C --- DEFINE ERROR MESSAGES ---

      message( 1,1) = 'I/O-error (readinput)'
      message( 1,2) = 'input file not found (readinput)'
      message( 1,3) = 'input file too long (readinput)'
      message( 1,4) = 'too many entries in input file (readinput)'
      message( 1,5) = 'too few energy points (readinput)'
      message( 1,6) = 'illegal VP for internal error (readinput)'
      message( 1,7) = 'I/O-error (fccpinput)'
      message( 1,8) = 'cannot open file (fccpinput)'
      message( 1,9) = 'I/O-error (getautolen)'
      message( 1,10)= 'autocorrelation file not found (getautolen)'
      message( 1,11)= 'I/O-error (readdata)'
      message( 1,12)= 'data file not found (readdata)'
      message( 1,14)=
     +   'integration time exceeds propagation time (readinput)'
      message( 1,15)= 'too many packets specified (readinput)'
      message( 1,16)= 'illegal packet for intensity (readinput)'
      message( 1,17)= 'usage of only c^0 not implemented (readinput)'
      message( 1,18)=
     +   'initial time exceeds propagation time (readinput)'
      message( 2,1) = 'I/O-error (readauto)'
      message( 2,2) = 'autocorrelation file not found (readauto)'
      message( 2,3) = 'insufficient size of correlation file (readauto)'
      message( 3,1) = 'illegal filter function (calchamil)'
      message( 4,1) = 'cannot diagonalise matrix (diagoverlap)'
      message( 4,2) = 'matrix not positive semidefinite (diagoverlap)'
      message( 5,1) = 'cannot diagonalise matrix B (variation)'
      message( 5,2) = 'matrix B has no positive eigenvalues (variation)'
      message( 5,3) = 'cannot diagonalise truncated matrix (variation)'
      message( 5,4) = 'not enough real workspace (variation)'
      message( 5,5) = 'illegal variational principle (variation)'
      message( 6,1) = 'I/O-error (writeparam)'
      message( 6,2) = 'cannot open log file (writeparam)'
      message( 6,3) = 'illegal filter function (writeparam)'
      message( 6,4) = 'illegal variational principle (writeparam)'
      message( 7,1) = 'I/O-error (getexactlen)'
      message( 7,2) = 'file with exact results not found (getexactlen)'
      message( 8,1) = 'illegal filter function (calcintensity)'
      message( 9,1) = 'I/O-error (writeenergy)'
      message( 9,2) = 'cannot open eig file (writeenergy)'
      message(10,1) = 'I/O-error (writeauto)'
      message(10,2) = 'cannot open auto file (writeauto)'
      message(10,3) = 'illegal filter function (writeauto)'
      message(11,1) = 'illegal filter function (calckamil)'
      message(12,1) = 'I/O-error (readexact)'
      message(12,2) = 'file with exact results not found (readexact)'
      message(12,3) = 'file with exact results is too long (readexact)'
      message(13,1) = 'illegal error estimate (calcerror/calcerror2)'
      
C --- SEND ERROR MESSAGE TO THE SCREEN ---

      write(6,*)
      write(6,'(2a)') 'Error message: ', message(routine,error)
      write(6,*)

C --- STOP THE PROGRAM ---

      stop
      end
