
C **********************************************************************
C *                                                                    *
C *                        VARIATION (variation.f)                     *
C *                                                                    *
C * Program package containing various variational principles.         *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                          SUBROUTINE VARIATION
C
C PURPOSE:
C     Selects the variational principle.
C INPUT PARAMETERS:
C     h0,h1,h2: Hamiltonian matrices H^(0), H^(1), and H^(2)
C     hamdim: size of Hamiltonian matrices
C     vartype: selects the variational principle:
C              1: <psi|H|psi>/<psi|psi>
C              2: <psi|(H-sigma)^2|psi>/<psi|psi>
C              3: <psi|H-sigma|psi>/<psi|(H-sigma)^2|psi>
C     sigma: energy shift used for the variational principle
C     epscut: cut-off parameter for eigenvalues in SVD
C     workdim: size of complex workspace
C OUTPUT PARAMETERS:
C     hdim: size of reduced Hamiltonian matrices (= no. of eigenvalues)
C     eigenvec: eigenvector matrix of Hamiltonian
C     eigenval: vector containing eigenenergies of Hamiltonian
C     error: error code having the following meaning:
C            0: everything was o. k.
C            1: cannot diagonalise matrix B
C            2: all eigenvalues of matrix B are negative
C            3: cannot diagonalise truncated matrix A
C            4: not enough complex workspace
C            5: illegal variational principle
C OTHER PARAMETERS:
C     work: complex workspace of length workdim
C ----------------------------------------------------------------------

      subroutine cvariation (h0, h1, h2, hamdim, vartype, sigma, epscut,
     +                      workdim, hdim, eigenvec, eigenval, error,
     +                      work)

      implicit none

      integer    hamdim, vartype, workdim, hdim, error
      real*8     sigma, epscut
      complex*16 h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +           h2(1:hamdim,1:hamdim), eigenvec(1:hamdim,1:hamdim),
     +           eigenval(1:hamdim), work(1:workdim)

C --- SELECT THE VARIATIONAL PRINCIPLE ---

      if (vartype .eq. 1) then
         call cvariation1(h0,h1,hamdim,epscut,workdim,hdim,eigenvec,
     +                   eigenval,error,work)
      elseif (vartype .eq. 2) then
         call cvariation2(h0,h1,h2,hamdim,sigma,epscut,workdim,hdim,
     +                   eigenvec,eigenval,error,work)
      elseif (vartype .eq. 3) then
         call cvariation3(h0,h1,h2,hamdim,sigma,epscut,workdim,hdim,
     +                   eigenvec,eigenval,error,work)
      else
         error = 5
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                          SUBROUTINE VARIATION 1
C
C PURPOSE:
C     Employs the variational principle <psi|H|psi>/<psi|psi>.
C INPUT PARAMETERS:
C     h0,h1: Hamiltonian matrices H^(0) and H^(1)
C     hamdim: size of Hamiltonian matrices
C     epscut: cut-off parameter for eigenvalues in SVD
C     workdim: size of complex workspace
C OUTPUT PARAMETERS:
C     hdim: size of reduced Hamiltonian matrices (= no. of eigenvalues)
C     eigenvec: eigenvector matrix of Hamiltonian
C     eigenval: vector containing eigenenergies of Hamiltonian
C     error: error code having the meaning as defined above
C OTHER PARAMETERS:
C     work: complex workspace of length workdim
C ----------------------------------------------------------------------

      subroutine cvariation1 (h0, h1, hamdim, epscut, workdim, hdim,
     +                       eigenvec, eigenval, error, work)

      use svdmod, only: svdz
      use op1lib

      implicit none

      integer    hamdim, workdim, hdim, error
      real*8     epscut
      complex*16 h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +           eigenvec(1:hamdim,1:hamdim), eigenval(1:hamdim),
     +           work(1:workdim)

      integer wrk, wrkdim

C --- INITIALISE VARIABLES ---

      error = 0
      wrk = 1+hamdim**2
      wrkdim = workdim-wrk+1
      if (wrkdim .lt. hamdim**2+4*hamdim) then
         error = 4
         return
      endif

C --- SAVE THE HAMILTONIAN MATRICES ---

      call cpqxz(h0,work(1),hamdim)
      call cpqxz(h1,eigenvec,hamdim)

C --- SOLVE GENERALISED EIGENVALUE PROBLEM ---

      call svdz(eigenvec,work(1),hamdim,epscut,wrkdim,eigenval,hdim,
     +          error,work(wrk))
      if (error .ne. 0) return

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                          SUBROUTINE VARIATION 2
C
C PURPOSE:
C     Employs the variational principle <psi|(H-sigma)^2|psi>/<psi|psi>.
C INPUT PARAMETERS:
C     h0,h1,h2: Hamiltonian matrices H^(0), H^(1), and H^(2)
C     hamdim: size of Hamiltonian matrices
C     sigma: energy shift used for the variational principle
C     epscut: cut-off parameter for eigenvalues in SVD
C     workdim: size of complex workspace
C OUTPUT PARAMETERS:
C     hdim: size of reduced Hamiltonian matrices (= no. of eigenvalues)
C     eigenvec: eigenvector matrix of Hamiltonian
C     eigenval: vector containing eigenenergies of Hamiltonian
C     error: error code having the meaning as defined above
C OTHER PARAMETERS:
C     work: complex workspace of length workdim
C ----------------------------------------------------------------------

      subroutine cvariation2 (h0, h1, h2, hamdim, sigma, epscut, 
     +                       workdim, hdim, eigenvec, eigenval, error, 
     +                       work)

      use svdmod, only: svdz
      use op1lib
      use xvlib

      implicit none

      integer    hamdim, workdim, hdim, error
      real*8     sigma, epscut
      complex*16 h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +           h2(1:hamdim,1:hamdim), eigenvec(1:hamdim,1:hamdim),
     +           eigenval(1:hamdim), work(1:workdim)

      integer    wrk, wrkdim, i
      complex*16 factor1, factor2

C --- INITIALISE VARIABLES ---

      error = 0
      wrk = 1+hamdim**2
      wrkdim = workdim-wrk+1

C --- SAVE H^(0) - MATRIX ---

      if (workdim .lt. hamdim**2) then
         error = 4
         return
      endif
      call cpqxz(h0,work(1),hamdim)

C --- SET UP (H-SIGMA)^2 - MATRIX ---

      factor1 = dcmplx(-2.0d0*sigma,0.0d0)
      factor2 = dcmplx(sigma**2,0.0d0)
      call cpqxz(h2,eigenvec,hamdim)
      call xqxxzza(factor1,h1,eigenvec,hamdim)
      call xqxxzza(factor2,h0,eigenvec,hamdim)

C --- SOLVE GENERALISED EIGENVALUE PROBLEM ---

      call svdz(eigenvec,work(1),hamdim,epscut,wrkdim,eigenval,hdim,
     +          error,work(wrk))
      if (error .ne. 0) return

C --- COMPUTE EIGENENERGIES THROUGH RAYLEIGH-QUOTIENT ---

      do i = 1,hdim
         call qvxxzz(h1,eigenvec(1,i),work(1),hamdim)
         call vvtxzz(eigenvec(1,i),work(1),factor1,hamdim)
         call qvxxzz(h0,eigenvec(1,i),work(1),hamdim)
         call vvtxzz(eigenvec(1,i),work(1),factor2,hamdim)
         eigenval(i) = factor1/factor2
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                          SUBROUTINE VARIATION 3
C
C PURPOSE:
C     Employs the variational principle
C    <psi|H-sigma|psi> / <psi|(H-sigma)^2|psi>.
C INPUT PARAMETERS:
C     h0,h1,h2: Hamiltonian matrices H^(0), H^(1), and H^(2)
C     hamdim: size of Hamiltonian matrices
C     sigma: energy shift used for the variational principle
C     epscut: cut-off parameter for eigenvalues in SVD
C     workdim: size of complex workspace
C OUTPUT PARAMETERS:
C     hdim: size of reduced Hamiltonian matrices (= no. of eigenvalues)
C     eigenvec: eigenvector matrix of Hamiltonian
C     eigenval: vector containing eigenenergies of Hamiltonian
C     error: error code having the meaning as defined above
C OTHER PARAMETERS:
C     work: complex workspace of length workdim
C ----------------------------------------------------------------------

      subroutine cvariation3 (h0, h1, h2, hamdim, sigma, epscut, 
     +                       workdim, hdim, eigenvec, eigenval, error, 
     +                       work)

      use svdmod, only: svdz
      use op1lib
      use xvlib

      implicit none

      integer    hamdim, workdim, hdim, error
      real*8     sigma, epscut
      complex*16 h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +           h2(1:hamdim,1:hamdim), eigenvec(1:hamdim,1:hamdim),
     +           eigenval(1:hamdim), work(1:workdim)

      integer    wrk, wrkdim, i, j, k
      complex*16 factor1, factor2, norm

C --- INITIALISE VARIABLES ---

      error = 0
      wrk = 1+hamdim**2
      wrkdim = workdim-wrk+1

C --- SET UP (H-SIGMA)^2 MATRIX ---

      if (workdim .lt. hamdim**2) then
         error = 4
         return
      endif
      factor1 = dcmplx(-2.0d0*sigma,0.0d0)
      factor2 = dcmplx(sigma**2,0.0d0)
      call cpqxz(h2,work(1),hamdim)
      call xqxxzza(factor1,h1,work(1),hamdim)
      call xqxxzza(factor2,h0,work(1),hamdim)

C --- SET UP (H-SIGMA) MATRIX ---

      factor1 = dcmplx(-sigma,0.0d0)
      call cpqxz(h1,eigenvec,hamdim)
      call xqxxzza(factor1,h0,eigenvec,hamdim)

C --- SOLVE GENERALISED EIGENVALUE PROBLEM ---

      call svdz(eigenvec,work(1),hamdim,epscut,wrkdim,eigenval,hdim,
     +          error,work(wrk))
      if (error .ne. 0) return

C --- COMPUTE EIGENENERGIES ---

      do i = 1,hdim
         eigenval(i) = sigma+1.0d0/eigenval(i)
      enddo

C --- NORMALISE COLUMNS OF EIGENVECTOR MATRIX ---

      do i = 1,hdim
         norm = dcmplx(0.0d0,0.0d0)
         do j = 1,hamdim
            do k = 1,hamdim
               norm = norm+eigenvec(j,i)*h0(j,k)*eigenvec(k,i)
            enddo
         enddo
         norm = sqrt(norm)
         call xvixzzo(norm,eigenvec(1,i),hamdim)
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end
