
C **********************************************************************
C *                                                                    *
C *                          ALLOCATE (allocate.f)                     *
C *                                                                    *
C * Program package containing the subroutine for memory allocation.   *
C *                                                                    *
C * Michael Beck                                                  7/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                          SUBROUTINE ALLOCATE
C
C PURPOSE:
C     Allocates memory, i. e. computes pointers to the addresses of the
C     "giant" memory array.
C INPUT PARAMETERS:
C     exactflag: flag denoting wether or not the exact results are read
C     maxexactdim: maximum number of exact energies to be read
C     edim: number of energy points
C     hamdim: size of Hamiltonian matrices
C     inttime: time interval for integrals
C     deltat: after every period of "deltat" the autocorrelation
C             function has been written to disk
C     autocorr: 0: only zero order autocorrelation function is used
C               1: also higher order autocorrelation functions are used
C     errortype: selector for different internal error checks
C     usepack: number of packets to be employed.
C OUTPUT PARAMETERS:
C     inttime: (corrected) time interval for integrals
C     autodim: length of autocorrelation file to be read
C     workdim: size of real workspace
C     mrdim: length of "giant" real*8 memory array actually used
C     midim: length of "giant" integer memory array actually used
C     maxauxdim: leading dimension of auxiliary matrix "auxmat"
C OTHER PARAMETERS:
C     Other parameters are passed via common blocks.
C ----------------------------------------------------------------------

      subroutine allocate (exactflag, maxexactdim, edim, hamdim,
     +                     inttime, deltat, autocorr, errortype,
     +                     usepack, autodim, workdim, mrdim, midim,
     +                     maxauxdim)

      implicit none

#include "fmemory.inc"

      integer dummy
      real*8  tiny
      parameter (dummy = 1, tiny = 1.0d-8)

      integer exactflag, maxexactdim, edim, autocorr, errortype,
     +        usepack, autodim, workdim, mrdim, midim, maxauxdim, hamdim
      real*8  inttime, deltat

C --- DETERMINE LENGTH OF AUTOCORRELATION FILE TO BE READ ---

      autodim = int(inttime/deltat+tiny*deltat)
      inttime = autodim*deltat

C --- SET POINTERS TO MEMORY ADDRESSES OF INTEGER ARRAYS ---

      midim = 1
      if ((exactflag .eq. 1) .or. (errortype .eq. 1)) then
         assignforw = midim
         midim = midim+2*hamdim
         assignback = midim
         midim = midim+max(maxexactdim,hamdim)
      else
         assignforw = dummy
         assignback = dummy
      endif
      
C --- SET POINTERS TO MEMORY ADDRESSES OF NON-AUXILIARY ARRAYS ---

      mrdim = 1
      h0 = mrdim
      mrdim = mrdim+hamdim*hamdim
      h1 = mrdim
      mrdim = mrdim+hamdim*hamdim
      h2 = mrdim
      mrdim = mrdim+hamdim*hamdim
      reauto0 = mrdim
      mrdim = mrdim+(autodim+1)*usepack**2
      imauto0 = mrdim
      mrdim = mrdim+(autodim+1)*usepack**2
      if (autocorr .ne. 0) then
         reauto1 = mrdim
         mrdim = mrdim+(autodim+1)*usepack**2
         imauto1 = mrdim
         mrdim = mrdim+(autodim+1)*usepack**2
         reauto2 = mrdim
         mrdim = mrdim+(autodim+1)*usepack**2
         imauto2 = mrdim
         mrdim = mrdim+(autodim+1)*usepack**2
      else
         reauto1 = dummy
         imauto1 = dummy
         reauto2 = dummy
         imauto2 = dummy
      endif
      intensity = mrdim
      if (errortype .eq. 0) then
         mrdim = mrdim+hamdim
      else
         mrdim = mrdim+2*hamdim
      endif
      eigenvec = mrdim
      mrdim = mrdim+hamdim*hamdim
      eigenval = mrdim
      if (errortype .eq. 0) then
         mrdim = mrdim+hamdim
      else
         mrdim = mrdim+2*hamdim
      endif
      eigenerror = mrdim
      if (errortype .eq. 0) then
         mrdim = mrdim+hamdim
      else
         mrdim = mrdim+2*hamdim
      endif
      if (exactflag .eq. 1) then
         exactenergy = mrdim
         mrdim = mrdim+hamdim
         exactintensity = mrdim
         mrdim = mrdim+hamdim
      else
         exactenergy = dummy
         exactintensity = dummy
      endif
      if (errortype .eq. 1) then
         intensity2 = mrdim
         mrdim = mrdim+hamdim
         intenserr = mrdim
         mrdim = mrdim+2*hamdim
         eigenvec2 = mrdim
         mrdim = mrdim+hamdim*hamdim
         eigenval2 = mrdim
         mrdim = mrdim+hamdim
      else
         intenserr = dummy
         intensity2 = dummy
         eigenvec2 = dummy
         eigenval2 = dummy
      endif
      dampvalue = mrdim
      mrdim = mrdim+edim
      if (autocorr .eq. 0) then
         dampvalue2 = mrdim
         mrdim = mrdim+edim
      else
         dampvalue2 = dummy
      endif

C --- DETERMINE SIZE OF AND POINTER TO REAL WORKSPACE ---

      if (exactflag .eq. 1) then
         maxauxdim = 2*hamdim+maxexactdim
      else
         maxauxdim = 2*hamdim
      endif
      work = mrdim
      workdim = 1
C for subroutine variation
      workdim = max(workdim,2*hamdim**2+4*hamdim-1)
C for subroutine calcintensity
      workdim = max(workdim,hamdim)
C for subroutine calcerror
      workdim = max(workdim,hamdim**2)
C for subroutine calcerror2
      workdim = max(workdim,8*hamdim)
C for subroutine writeenergy
      if (exactflag .eq. 1) then
         workdim = max(workdim,7*maxauxdim)
      else
         workdim = max(workdim,5*maxauxdim)
      endif
      mrdim = mrdim+workdim

C --- RETURN TO CALLING ROUTINE ---

      return
      end
