
C **********************************************************************
C *                                                                    *
C *                        CALCULATE (calculate.f)                     *
C *                                                                    *
C * Program package containing the subroutines that perform the        *
C * various calculations needed by the filtering method.               *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                        SUBROUTINE CALC INTENSVEC
C
C PURPOSE:
C     Computes the intensity vector (d in the PhD thesis).
C INPUT PARAMETERS:
C     reauto0: contains the real part of the autocorrelation function
C     imauto0: contains the imaginary part of the autocorrel. function
C     autodim: denotes the length (plus 1) of the autocorrel. function
C     hamdim: size of Hamiltonian matrices
C     edim: number of energy points
C     emin, emax: lower and upper boundary of energy range
C     deltat: time gap of autocorrelation function
C     damptime: time defining the width of the damp function
C     damptype: specifies the damp function to be used
C     usepack: number of packets to be employed.
C     intenspack: number of the packet for which the intensity is
C                 computed.
C OUTPUT PARAMETERS:
C     intensity: vector containing the intensities
C     error: error flag passed to damp function (see there for details)
C ----------------------------------------------------------------------

      subroutine calcintensvec (reauto0, imauto0, autodim, hamdim, edim,
     +                          emin, emax, deltat, damptime, damptype,
     +                          usepack, intenspack, intensity, error)

      use op1lib

      implicit none

      integer autodim, hamdim, edim, damptype, usepack, intenspack,
     +        error
      real*8  reauto0(1:usepack,1:usepack,0:autodim),
     +        imauto0(1:usepack,1:usepack,0:autodim), emin, emax,
     +        deltat, damptime, intensity(1:hamdim)

      integer dim, n, i, p, j
      real*8  time, ediff, energy, dampvalue, dampfunc

C --- INITIALISE VARIABLES ---

      error = 0
      ediff = (emax-emin)/(edim-1)
      if (damptype .eq. 0) then
         dim = autodim
      else
         dim = autodim/2+1
      endif
      call zerovxd(intensity,hamdim)

C --- LOOP OVER EACH PACKET AND ENERGY ---

      do p = 1,usepack
         do j = 1,edim
            energy = emin+(j-1)*ediff
            n = j+(p-1)*edim

C       --- COMPUTE INTENSITY VECTOR (d IN THE PhD THESES) ---

C The factor 0.5 accounts for the trapezoidal rule.

            do i = 0,dim
               time = i*deltat
               dampvalue = dampfunc(time,damptime,damptype,error)
               if (error .ne. 0) return
               if ((i .eq. 0) .or. (i .eq. dim)) then
                  dampvalue = 0.5d0*dampvalue
               endif
               intensity(n) = intensity(n)+2.0d0*deltat*dampvalue
     +                        *(reauto0(intenspack,p,i)*cos(energy*time)
     +                        -imauto0(intenspack,p,i)*sin(energy*time))
            enddo
         enddo
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                        SUBROUTINE CALC INTENSITY
C
C PURPOSE:
C     Computes the intensities.
C INPUT PARAMETERS:
C     intensity: intensity vector as determined by "calcintensvec"
C     coeff: coefficient matrix (B in the PhD thesis)
C     hamdim: size of Hamiltonian matrices
C     hdim: second dimension of coefficient matrix
C OUTPUT PARAMETERS:
C     intensity: vector containing the intensities
C OTHER PARAMETERS:
C     work: real workspace; must be >= hdim
C ----------------------------------------------------------------------

      subroutine calcintensity (intensity, coeff, hamdim, hdim, work)

      use xvlib

      implicit none

      integer hamdim, hdim
      real*8  intensity(1:hamdim), coeff(1:hamdim,1:hdim), work(1:hdim)

      integer i

C --- DETERMINE OVERLAPS OF INITIAL WAVEFUNCTION AND EIGENFUNCTIONS ---

      call mvtxdd(coeff,intensity,work,hamdim,hdim)

C --- COMPUTE SPECTRAL INTENSITIES ---

      do i = 1,hdim
         intensity(i) = work(i)**2
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                         SUBROUTINE TRAFO ENERGY
C
C PURPOSE:
C     Transforms the energy, i. e. computes E = 1/t arcsin((E-S)t)
C     where S is the energy shift of the Hamiltonian. Note that this
C     formula is only valid if the wavefunction has been propagated
C     employing the SOD-method where the effective Hamiltonian equals
C     H_eff = 1/t arcsin(H t).
C INPUT PARAMETERS:
C     eminin, emaxin: lower and upper boundary of energy range as given
C                     in input file
C     sigmain: energy shift used for the variational principle as given
C              in input file
C     sigmain2: same as "sigmain", but for second variational principle
C     deltat: after every period of "deltat" the autocorrelation
C             function has been written to disk
C     hamshift: energy shift of the Hamiltonian
C     sodflag: 1 if propagation has been done with SOD, else 0
C OUTPUT PARAMETERS:
C     emin, emax: transformed lower and upper boundary of energy range
C     sigma: transformed energy shift used for the variational principle
C     sigma2: same as "sigma", but for second variational principle
C ----------------------------------------------------------------------

      subroutine trafoenergy (eminin, emaxin, sigmain, sigmain2, deltat,
     +                        hamshift, sodflag, emin, emax, sigma,
     +                        sigma2)

      implicit none

      integer sodflag
      real*8  eminin, emaxin, sigmain, sigmain2, deltat, hamshift, emin,
     +        emax, sigma, sigma2

      real*8  stepsize

C --- CORRECT THE ENERGIES ---

      if (sodflag .eq. 1) then
         stepsize = 0.5d0*deltat
         emin   = asin((eminin  -hamshift)*stepsize)/stepsize
         emax   = asin((emaxin  -hamshift)*stepsize)/stepsize
         sigma  = asin((sigmain -hamshift)*stepsize)/stepsize
         sigma2 = asin((sigmain2-hamshift)*stepsize)/stepsize
      else
         emin   = eminin  -hamshift
         emax   = emaxin  -hamshift
         sigma  = sigmain -hamshift
         sigma2 = sigmain2-hamshift
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                       SUBROUTINE CORRECT ENERGY
C
C PURPOSE:
C     Corrects for the energy shift S of the Hamiltonian. Note that the
C     formula E = S + 1/t sin((E-S)t) is only valid if the wavefunction
C     has been propagated employing the SOD-method where the effective
C     Hamiltonian equals H_eff = 1/t arcsin(H t).
C INPUT PARAMETERS:
C     energy: array containing the computed energies.
C     hdim: size of reduced Hamiltonian matrices and size of "energy"
C     deltat: after every period of "deltat" the autocorrelation
C             function has been written to disk
C     hamshift: energy shift of the Hamiltonian
C     autocorr: 0: only zero order autocorrelation function is used
C               1: also higher order autocorrelation functions are used
C     sodflag: 1 if propagation has been done with SOD, else 0
C OUTPUT PARAMETERS:
C     energy: array containing the corrected energies.
C ----------------------------------------------------------------------

      subroutine correctenergy (energy, hdim, deltat, hamshift,
     +                          autocorr, sodflag)

      implicit none

      integer hdim, autocorr, sodflag
      real*8  energy(1:hdim), deltat, hamshift

      integer i
      real*8  stepsize

C --- CORRECT THE ENERGIES ---

      if ((autocorr .eq. 0) .and. (sodflag .eq. 1)) then
         stepsize = 0.5d0*deltat
         do i = 1,hdim
            energy(i) = sin(energy(i)*stepsize)/stepsize+hamshift
         enddo
      else
         do i = 1,hdim
            energy(i) = energy(i)+hamshift
         enddo
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                         SUBROUTINE CALCERROR
C
C PURPOSE:
C     Estimates the error of the eigenvalues.
C INPUT PARAMETERS:
C     h0,h1,h2: Hamiltonian matrices H^(0), H^(1), and H^(2)
C     coeff: coefficient matrix (B in the PhD thesis)
C     hamdim: size of Hamiltonian matrices
C     hdim: second dimension of coefficient matrix
C     eigenval: eigenvalue vector
C OUTPUT PARAMETERS:
C     eigenerror: array containing estimated error for each eigenvalue
C     hdimall: number of eigenvalues (equals hdim for this estimate)
C OTHER PARAMETERS:
C     work: real workspace; must be >= hamdim**2+hamdim
C ----------------------------------------------------------------------

      subroutine calcerror (h0, h1, h2, coeff, hamdim, hdim, eigenval,
     +                      eigenerror, hdimall, work)

      use op1lib
      use xvlib

      implicit none

      integer hamdim, hdim, hdimall
      real*8  h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +        h2(1:hamdim,1:hamdim), coeff(1:hamdim,1:hdim),
     +        eigenval(1:hdim), eigenerror(1:hdim),
     +        work(1:hamdim**2+hamdim)

      integer vec, i
      real*8  factor1, factor2

C --- INITIALISE VARIABLES ---

      hdimall = hdim
      vec = 1+hamdim**2

C --- LOOP OVER EACH EIGENVALUE ---

      do i = 1,hdim

C    --- SET UP (H-E)^2 - MATRIX ---

         factor1 = -2.0d0*eigenval(i)
         factor2 = eigenval(i)**2
         call cpqxd(h2,work(1),hamdim)
         call xqxxdda(factor1,h1,work(1),hamdim)
         call xqxxdda(factor2,h0,work(1),hamdim)

C    --- COMPUTE SQUARE OF NORM OF (H-E)|PSI> ---

         call qvxxdd(work(1),coeff(1,i),work(vec),hamdim)
         call vvtxdd(coeff(1,i),work(vec),eigenerror(i),hamdim)

C    --- COMPUTE NORM ---

         eigenerror(i) = sqrt(abs(eigenerror(i)))
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                         SUBROUTINE CALCERROR 2
C
C PURPOSE:
C     Estimates the error of the eigenvalues that have been determined
C     employing two different variational principles.
C INPUT PARAMETERS:
C     eigenval: first eigenvalue vector
C     eigenval2: second eigenvalue vector
C     intensity: first intensity vector
C     intensity2: second intensity vector
C     hdim: size of first eigenvalue/intensity vector
C     hdim2: size of second eigenvalue/intensity vector
C     econv: conversion for energy unit
C OUTPUT PARAMETERS:
C     eigenerror: array containing estimated error for each eigenvalue
C     intenserr: array containing estimated error for each intensity
C     hdimall: size of combined energy vectors
C OTHER PARAMETERS:
C     work: real workspace; must be >= 4*(hdim1+hdim2)
C     assignforw: integer array with forward assignments between
C                 computed eigenvalues
C     assignback: integer array with backward assignments between
C                 computed eigenvalues
C ----------------------------------------------------------------------

      subroutine calcerror2 (eigenval, eigenval2, intensity, intensity2,
     +                       hdim, hdim2, eigenerror, intenserr,
     +                       hdimall, work, assignforw, assignback,
     +                       econv)

      use op1lib
      use xvlib

      implicit none

      integer hdim, hdim2, hdimall, assignforw(1:hdim),
     +        assignback(1:hdim2)
      real*8  econv, eigenval(1:hdim+hdim2), eigenval2(1:hdim2),
     +        intensity(1:hdim+hdim2), intensity2(1:hdim2),
     +        eigenerror(1:hdim+hdim2), intenserr(1:hdim+hdim2),
     +        work(1:hdim+hdim2,1:4)

      integer i, j

C --- SORT RESULTS OF FIRST VARIATIONAL PRINCIPLE ---

      call cpvxd(eigenval,work(1,1),hdim)
      call cpvxd(intensity,work(1,2),hdim)
      call dbubble1(work(1,1),work(1,2),hdim,hdim+hdim2,1)
      call cpvxd(work(1,1),eigenval,hdim)
      call cpvxd(work(1,2),intensity,hdim)

C --- SORT RESULTS OF SECOND VARIATIONAL PRINCIPLE ---

      call cpvxd(eigenval2,work(1,1),hdim2)
      call cpvxd(intensity2,work(1,2),hdim2)
      call dbubble1(work(1,1),work(1,2),hdim2,hdim+hdim2,1)
      call cpvxd(work(1,1),eigenval2,hdim2)
      call cpvxd(work(1,2),intensity2,hdim2)

C --- FIND BEST MATCHES BETWEEN THE TWO LISTS OF EIGENVALUES ---
      
      call dmatch(eigenval,eigenval2,hdim,hdim2,assignforw,assignback)

C --- STORE EIGENVALUES THAT HAVE BEEN MATCHED ---

      hdimall = 0
      do i = 1,hdim
         if (assignforw(i) .ne. 0) then
            hdimall = hdimall+1
            work(hdimall,1) = eigenval(i)
            work(hdimall,2) =
     +         abs(eigenval(i)-eigenval2(assignforw(i)))
            work(hdimall,3) = intensity(i)
            work(hdimall,4) =
     +         abs(intensity(i)-intensity2(assignforw(i)))
         endif
      enddo

C --- STORE SPURIOUS EIGENVALUES FROM FIRST CALCULATION ---

      do i = 1,hdim
         if (assignforw(i) .eq. 0) then
            hdimall = hdimall+1
            work(hdimall,1) = eigenval(i)
            work(hdimall,2) = -1.0d0*econv
            work(hdimall,3) = intensity(i)
            work(hdimall,4) = 0.0d0
         endif
      enddo

C --- STORE SPURIOUS EIGENVALUES FROM SECOND CALCULATION ---

      do j = 1,hdim2
         if (assignback(j) .eq. 0) then
            hdimall = hdimall+1
            work(hdimall,1) = eigenval2(j)
            work(hdimall,2) = -1.0d0*econv
            work(hdimall,3) = intensity2(j)
            work(hdimall,4) = 0.0d0
         endif
      enddo

C --- COPY RESULTS BACK ---

      call cpvxd(work(1,1),eigenval,hdimall)
      call cpvxd(work(1,2),eigenerror,hdimall)
      call cpvxd(work(1,3),intensity,hdimall)
      call cpvxd(work(1,4),intenserr,hdimall)

C --- RETURN TO CALLING ROUTINE ---

      return
      end
