
C **********************************************************************
C *                                                                    *
C *                          DAMPING (damping.f)                       *
C *                                                                    *
C * Program package containing various damping functions.              *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                           FUNCTION DAMPFUNC
C
C PURPOSE:
C     Selects and computes a damp function (g in the PhD thesis).
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     damptype: selector for the damp function having the meaning:
C             0: Gaussian damping
C             1: cosine damping
C             2: squared cosine damping
C             3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function dampfunc (time, damptime, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime

      logical initial
      real*8  halfpi

      save initial, halfpi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      dampfunc = 0.0d0
      if (initial) then
         halfpi = 0.5d0*4.0d0*atan(1.0d0)
         initial = .false.
      endif

C --- SELECT AND COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 0) then
         dampfunc = exp(-((time/damptime)**2))
      elseif (damptype .eq. 1) then
         if (time .ge. damptime) return
         dampfunc = cos(halfpi*time/damptime)
      elseif (damptype .eq. 2) then
         if (time .ge. damptime) return
         dampfunc = (cos(halfpi*time/damptime))**2
      elseif (damptype .eq. 3) then
         if (time .ge. damptime) return
         dampfunc = 1.0d0
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                           FUNCTION DAMPFOUR
C
C PURPOSE:
C     Selects and computes the Fourier transform of a damp function
C     (\hat{g} in the PhD thesis).
C INPUT PARAMETERS:
C     energy: argument for the Fourier transform of the damp function
C     damptime: time defining the width of the damp function
C     damptype: selector for the damp function having the meaning:
C             0: Gaussian damping
C             1: cosine damping
C             2: squared cosine damping
C             3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen function at energy "energy".
C ----------------------------------------------------------------------

      real*8 function dampfour (energy, damptime, damptype, error)

      implicit none

      integer damptype, error
      real*8  energy, damptime

      logical initial
      real*8  arg, pi, sqrtpi, pi2, dsinc

      save initial, pi, sqrtpi, pi2
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error    = 0
      dampfour = 0.0d0
      arg      = energy*damptime
      if (initial) then
         pi      = 4.0d0*atan(1.0d0)
         sqrtpi  = sqrt(pi)
         pi2     = pi**2
         initial = .false.
      endif

C --- SELECT DAMP FUNCTION AND COMPUTE ITS FOURIER TRANSFORM ---

      if (damptype .eq. 0) then
         dampfour = sqrt(pi)*damptime*exp(-0.25d0*arg**2)
      elseif (damptype .eq. 1) then
         dampfour = -pi*damptime*cos(arg)/(arg**2-0.25d0*pi2)
      elseif (damptype .eq. 2) then
         dampfour = -pi2*damptime*dsinc(arg)/(arg**2-pi2)
      elseif (damptype .eq. 3) then
         dampfour = 2.0d0*damptime*dsinc(arg)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION DAMP
C
C PURPOSE:
C     Selects and computes a (transformed) damp function (G in the PhD
C     thesis). Needed for the representation where all autocorrelation
C     functions c0(t), c1(t), and c2(t) are used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            0: Gaussian damping
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function damp (time, damptime, deltae, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime, deltae
      
      logical initial
      real*8  tbar, ratio, pitime, dsinc, arg1, arg2, arg3, arg4, arg5,
     +        pi, halfpi

      save initial, pi, halfpi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi = 4.0d0*atan(1.0d0)
         halfpi = 0.5d0*pi
         initial = .false.
      endif
      damp = 0.0d0
      ratio = time/damptime
      pitime = pi/damptime
      tbar = damptime-0.5d0*time

C --- SELECT DAMP FUNCTION ---

      if (damptype .eq. 0) then
         damp = sqrt(2.0d0*pi)*damptime
     +          *exp((-0.125d0)*(damptime*deltae)**2-0.5d0*ratio**2)
      elseif (damptype .eq. 1) then
         if (ratio .ge. 2.0d0) return
         arg1 = deltae*tbar
         arg2 = (pitime-deltae)*tbar
         arg3 = (pitime+deltae)*tbar
         damp = tbar*(2.0d0*cos(halfpi*ratio)*dsinc(arg1)
     +                +dsinc(arg2)+dsinc(arg3))
      elseif (damptype .eq. 2) then
         if (ratio .ge. 2.0d0) return
         arg1 = deltae*tbar
         arg2 = (pitime-deltae)*tbar
         arg3 = (pitime+deltae)*tbar
         arg4 = (2.0d0*pitime-deltae)*tbar
         arg5 = (2.0d0*pitime+deltae)*tbar
         damp = tbar*(((cos(halfpi*ratio))**2+0.5d0)*dsinc(arg1)
     +                +cos(halfpi*ratio)*(dsinc(arg2)+dsinc(arg3))
     +                +0.25d0*(dsinc(arg4)+dsinc(arg5)))
      elseif (damptype .eq. 3) then
         if (ratio .ge. 2.0d0) return
         arg1 = deltae*tbar
         damp = 4.0d0*tbar*dsinc(arg1)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION KAMP1
C
C PURPOSE:
C     Computes the first (transformed) damp function (F1 in the PhD
C     thesis). Needed for the representation where only the
C     autocorrelation function c0(t) is used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            0: Gaussian damping
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function kamp1 (time, damptime, deltae, damptype, error)

      implicit none

      real*8 tiny
      parameter (tiny = 1.0d-12)

      integer damptype, error
      real*8  time, damptime, deltae

      logical initial
      real*8  tbar, ratio, pitime, dsinc, arg1, arg2, arg3, pi, halfpi,
     +        kappa

      save initial, pi, halfpi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi      = 4.0d0*atan(1.0d0)
         halfpi  = 0.5d0*pi
         initial = .false.
      endif
      kamp1  = 0.0d0
      ratio  = time/damptime
      pitime = pi/damptime
      tbar   = damptime-0.5d0*time
      kappa  = tbar/damptime

C --- COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 0) then
         kamp1 = -ratio*sqrt(2.0d0*pi)*exp((-0.125d0)
     +           *(damptime*deltae)**2-0.5d0*ratio**2)
      elseif (damptype .eq. 1) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         kamp1 = -pi*kappa*sin(halfpi*ratio)*dsinc(arg1)
      elseif (damptype .eq. 2) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         kamp1 = -halfpi*kappa*(sin(pi*ratio)*dsinc(arg1)
     +           +sin(halfpi*ratio)*(dsinc(arg2)+dsinc(arg3)))
      elseif (damptype .eq. 3) then
         if ((ratio .ge. 2.0d0) .or. (ratio .le. tiny)) return
         arg1  = deltae*tbar
         kamp1 = -2.0d0*cos(arg1)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                             FUNCTION KAMP2
C
C PURPOSE:
C     Computes the second (transformed) damp function (F2 in the PhD
C     thesis). Needed for the representation where only the
C     autocorrelation function c0(t) is used.
C INPUT PARAMETERS:
C     time: argument for the damp function
C     damptime: time defining the width of the damp function
C     deltae: energy difference E-E'
C     damptype: selector for the damp function having the meaning:
C            0: Gaussian damping
C            1: cosine damping
C            2: squared cosine damping
C            3: box damping
C OUTPUT PARAMETERS:
C     error: error flag having the following meaning:
C            0: everything was o. k.
C            1: an illegal damp function has been specified
C FUNCTION VALUE:
C     value of the chosen damp function at time "time"
C ----------------------------------------------------------------------

      real*8 function kamp2 (time, damptime, deltae, damptype, error)

      implicit none

      integer damptype, error
      real*8  time, damptime, deltae

      logical initial
      real*8  tbar, ratio, pitime, dsinc, arg1, arg2, arg3, arg4, arg5,
     +        pi, halfpi, kappa

      save initial, pi, halfpi
      data initial /.true./

C --- INITIALISE VARIABLES ---

      error = 0
      if (initial) then
         pi = 4.0d0*atan(1.0d0)
         halfpi  = 0.5d0*pi
         initial = .false.
      endif
      kamp2  = 0.0d0
      ratio  = time/damptime
      pitime = pi/damptime
      tbar   = damptime-0.5d0*time
      kappa  = tbar/damptime

C --- COMPUTE DAMP FUNCTION ---

      if (damptype .eq. 0) then
         kamp2 = (1.0d0-ratio**2+0.25d0*(damptime*deltae)**2)/damptime
     +           *sqrt(2.0d0*pi)
     +           *exp((-0.125d0)*(damptime*deltae)**2-0.5d0*ratio**2)
      elseif (damptype .eq. 1) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         kamp2 = halfpi/damptime*kappa*(pi*cos(halfpi*ratio)*dsinc(arg1)
     +           +(deltae*damptime-halfpi)*dsinc(arg2)
     +           -(deltae*damptime+halfpi)*dsinc(arg3))
      elseif (damptype .eq. 2) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         arg2  = (pitime-deltae)*tbar
         arg3  = (pitime+deltae)*tbar
         arg4  = (2.0d0*pitime-deltae)*tbar
         arg5  = (2.0d0*pitime+deltae)*tbar
         kamp2 = 0.25d0*pi/damptime*kappa*(
     +             2.0d0*pi*cos(pi*ratio)*dsinc(arg1)
     +            +2.0d0*deltae*damptime*cos(halfpi*ratio)
     +             *(dsinc(arg2)-dsinc(arg3))
     +            +(deltae*damptime-pi)*dsinc(arg4)
     +            -(deltae*damptime+pi)*dsinc(arg5))
      elseif (damptype .eq. 3) then
         if (ratio .ge. 2.0d0) return
         arg1  = deltae*tbar
         kamp2 = 2.0d0*deltae*sin(arg1)
      else
         error = 1
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end
