
C **********************************************************************
C *                                                                    *
C *                         GENHAMIL (genhamil.f)                      *
C *                                                                    *
C * Program package containing the subroutines that generate the       *
C * Hamiltonian matrices needed by the filtering method.               *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                         SUBROUTINE CALC HAMIL
C
C PURPOSE:
C     Computes the overlap matrix and - when all autocorrelation
C     functions c0, c1, and c2 are used - also the hamiltonian matrices.
C INPUT PARAMETERS:
C     reauto0,1,2: contains the real part of the autocorrelation
C                  function of order 0, 1, and 2, respectively
C     imauto0,1,2: contains the imaginary part of the autocorrelation
C                  function of order 0, 1, and 2, respectively
C     autodim: denotes the length (plus 1) of the autocorrel. function
C     edim: number of energy points
C     hamdim: size of Hamiltonian matrices
C     emin, emax: lower and upper boundary of energy range
C     deltat: time gap of autocorrelation function
C     damptime: time defining the width of the damp function
C     damptype: specifies the damp function to be used
C     autocorr: 0: only zero order autocorrelation function is used
C               1: also higher order autocorrelation functions are used
C     usepack: number of packets to be employed.
C OUTPUT PARAMETERS:
C     h0,h1,h2: hamiltonian matrices of order 0, 1, and 2, respectively
C     error: error flag passed to damp function (see there for details)
C OTHER PARAMETERS:
C     dampvalue: auxiliary array for storing some values of the damp
C                function during the calculation
C ----------------------------------------------------------------------

      subroutine calchamil (reauto0, imauto0, reauto1, imauto1, reauto2,
     +                      imauto2, autodim, edim, hamdim, emin, emax,
     +                      deltat, damptime, damptype, autocorr,
     +                      usepack, dampvalue, h0, h1, h2, error)

      use op1lib, only: zeromxd

      implicit none

      integer autodim, edim, hamdim, damptype, autocorr, usepack, error
      real*8  reauto0(1:usepack,1:usepack,0:autodim),
     +        imauto0(1:usepack,1:usepack,0:autodim),
     +        reauto1(1:usepack,1:usepack,0:autodim),
     +        imauto1(1:usepack,1:usepack,0:autodim),
     +        reauto2(1:usepack,1:usepack,0:autodim),
     +        imauto2(1:usepack,1:usepack,0:autodim), emin, emax,
     +        deltat, damptime, dampvalue(0:edim-1),
     +        h0(1:hamdim,1:hamdim), h1(1:hamdim,1:hamdim),
     +        h2(1:hamdim,1:hamdim)

      integer m, n, i, j, k, p, q
      real*8  deltae, ediff, time, sinus, cosinus, si, co, fourier0,
     +        fourier1, fourier2, tmp, damp, mean

C --- INITIALIZE VARIABLES ---

      error = 0
      ediff = (emax-emin)/(edim-1)
      call zeromxd(h0,hamdim,hamdim)
      call zeromxd(h1,hamdim,hamdim)
      call zeromxd(h2,hamdim,hamdim)

C --- SYMMETRISE CROSS-CORRELATION FUNCTIONS ---

      do i = 0,autodim
         do q = 1,usepack
            do p = q+1,usepack
               mean = 0.5d0*(reauto0(p,q,i)+reauto0(q,p,i))
               reauto0(p,q,i) = mean
               reauto0(q,p,i) = mean
               mean = 0.5d0*(imauto0(p,q,i)+imauto0(q,p,i))
               imauto0(p,q,i) = mean
               imauto0(q,p,i) = mean
               if (autocorr .eq. 1) then
                  mean = 0.5d0*(reauto1(p,q,i)+reauto1(q,p,i))
                  reauto1(p,q,i) = mean
                  reauto1(q,p,i) = mean
                  mean = 0.5d0*(imauto1(p,q,i)+imauto1(q,p,i))
                  imauto1(p,q,i) = mean
                  imauto1(q,p,i) = mean
                  mean = 0.5d0*(reauto2(p,q,i)+reauto2(q,p,i))
                  reauto2(p,q,i) = mean
                  reauto2(q,p,i) = mean
                  mean = 0.5d0*(imauto2(p,q,i)+imauto2(q,p,i))
                  imauto2(p,q,i) = mean
                  imauto2(q,p,i) = mean
               endif
            enddo
         enddo
      enddo
            
C --- LOOP OVER AUTO-/CROSS-CORRELATION FUNCTION ---

      do i = 0,autodim
         time = i*deltat

C    --- EVALUATE DAMP FUNCTION FOR ALL DELTA_E ---

C The factor 0.5 accounts for the trapezoidal rule.

         do m = 0,edim-1
            deltae = m*ediff
            dampvalue(m) = damp(time,damptime,deltae,damptype,error)
            if (error .ne. 0) return
            if ((i .eq. 0) .or. (i .eq. autodim))
     +         dampvalue(m) = 0.5d0*dampvalue(m)
         enddo

C    --- LOOP OVER ALL PACKETS ---

         do q = 1,usepack
            do p = q,usepack

C          --- COMPUTE Re(c(t) exp(i ebar t)) FOR ALL E_BAR ---

               sinus   = sin(time*emin)
               cosinus = cos(time*emin)
               si      = sin(0.5d0*time*ediff)
               co      = cos(0.5d0*time*ediff)
               do n = 2,2*edim
                  fourier0 = reauto0(p,q,i)*cosinus-imauto0(p,q,i)*sinus
                  if (autocorr .eq. 1) then
                     fourier1 =
     +                  reauto1(p,q,i)*cosinus-imauto1(p,q,i)*sinus
                     fourier2 =
     +                  reauto2(p,q,i)*cosinus-imauto2(p,q,i)*sinus
                  endif
                  tmp     = sinus*co+cosinus*si
                  cosinus = cosinus*co-sinus*si
                  sinus   = tmp

C             --- DETERMINE LOWER LEFT HALF OF HAMILTONIAN MATRIX ---

                  do m = mod(n,2),min(n-2,2*edim-n),2
                     j = (n+m)/2
                     k = (n-m)/2
                     h0(j+(p-1)*edim,k+(q-1)*edim) =
     +                  h0(j+(p-1)*edim,k+(q-1)*edim)+
     +                  deltat*fourier0*dampvalue(m)
                     if (autocorr .eq. 1) then
                        h1(j+(p-1)*edim,k+(q-1)*edim) =
     +                     h1(j+(p-1)*edim,k+(q-1)*edim)+
     +                     deltat*fourier1*dampvalue(m)
                        h2(j+(p-1)*edim,k+(q-1)*edim) =
     +                     h2(j+(p-1)*edim,k+(q-1)*edim)+
     +                     deltat*fourier2*dampvalue(m)
                     endif
                  enddo
               enddo
            enddo
         enddo
      enddo

C --- MIRROR HAMILTON-MATRIX FOR NON-DIAGONAL BLOCKS ---

      do q = 1,usepack
         do p = q+1,usepack
            do k = 1,edim
               do j = k+1,edim
                  h0(k+(p-1)*edim,j+(q-1)*edim) =
     +               h0(j+(p-1)*edim,k+(q-1)*edim)
                  if (autocorr .eq. 1) then
                     h1(k+(p-1)*edim,j+(q-1)*edim) =
     +                  h1(j+(p-1)*edim,k+(q-1)*edim)
                     h2(k+(p-1)*edim,j+(q-1)*edim) =
     +                  h2(j+(p-1)*edim,k+(q-1)*edim)
                  endif
               enddo
            enddo
         enddo
      enddo

C --- CALCULATE SECOND HALF OF (SYMMETRIC) HAMILTONIAN ---

      do j = 1,hamdim
         do k = j+1,hamdim
            h0(j,k) = h0(k,j)
         enddo
      enddo
      if (autocorr .eq. 1) then
         do j = 1,hamdim
            do k = j+1,hamdim
               h1(j,k) = h1(k,j)
               h2(j,k) = h2(k,j)
            enddo
         enddo
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                         SUBROUTINE CALC KAMIL
C
C PURPOSE:
C     Computes the hamiltonian matrices K1 and K2 (needed when only the
C     autocorrelation function c0 is used).
C INPUT PARAMETERS:
C     reauto: contains the real part of the autocorrelation function
C     imauto: contains the imaginary part of the autocorrelat. function
C     autodim: denotes the length (plus 1) of the autocorrel. function
C     h0: overlap matrix
C     edim: number of energy points
C     hamdim: size of Hamiltonian matrices
C     emin, emax: lower and upper boundary of energy range
C     deltat: time gap of autocorrelation function
C     damptime: time defining the width of the damp function
C     damptype: specifies the damp function to be used
C     usepack: number of packets to be employed.
C OUTPUT PARAMETERS:
C     k1: first hamiltonian matrix
C     k2: second hamiltonian matrix
C     error: error flag passed to damp function (see there for details)
C OTHER PARAMETERS:
C     dampvalue1: auxiliary array for storing some values of the damp
C                 function during the calculation
C     dampvalue2: auxiliary array for storing some values of the damp
C                 function during the calculation
C ----------------------------------------------------------------------

      subroutine calckamil (reauto, imauto, autodim, h0, edim, hamdim,
     +                      emin, emax, deltat, damptime, damptype,
     +                      usepack, k1, k2, error, dampvalue1,
     +                      dampvalue2)

      use op1lib, only: zeromxd

      implicit none

      integer autodim, edim, hamdim, damptype, usepack, error
      real*8  reauto(1:usepack,1:usepack,0:autodim),
     +        imauto(1:usepack,1:usepack,0:autodim),
     +        h0(1:hamdim,1:hamdim), emin, emax, deltat, damptime,
     +        k1(1:hamdim,1:hamdim), k2(1:hamdim,1:hamdim),
     +        dampvalue1(0:edim-1), dampvalue2(0:edim-1)

      integer m, n, i, j, k, p, q
      real*8  deltae, ediff, time, tmp, fourier1, fourier2, integral1,
     +        integral2, ebar, kamp1, kamp2, sinus, cosinus, si, co,
     +        factor

C --- INITIALISE VARIABLES ---

      error = 0
      ediff = (emax-emin)/(edim-1)
      call zeromxd(k1,hamdim,hamdim)
      call zeromxd(k2,hamdim,hamdim)

C --- LOOP OVER AUTOCORRELATION FUNCTION ---

      do i = 0,autodim
         time = i*deltat

C    --- COMPUTE VALUES OF DAMP FUNCTION (FIRST PART) ---

         do m = 0,edim-1
            deltae = m*ediff
            dampvalue1(m) = kamp1(time,damptime,deltae,damptype,error)
            if (error .ne. 0) return
            dampvalue2(m) = kamp2(time,damptime,deltae,damptype,error)
            if (error .ne. 0) return
C The factor 0.5 accounts for the trapezoidal rule.
            if ((i .eq. 0) .or. (i .eq. autodim)) then
               dampvalue1(m) = 0.5d0*dampvalue1(m)
               dampvalue2(m) = 0.5d0*dampvalue2(m)
            endif
C For a box filter, add the contribution of the delta functions.
C The factor 4.0 below should actually be a factor 2.0, because only one
C half of the delta function at t=0 lies within the integration
C interval. However, one then obtains completely wrong results, while
C the formula below works perfectly.
            if (damptype .eq. 3) then
               if (i .eq. 0) then
                  dampvalue2(m) = dampvalue2(m)
     +                           +4.0d0*cos(deltae*damptime)/deltat
               elseif (i .eq. autodim) then
                  dampvalue2(m) = dampvalue2(m)-2.0d0/deltat
               endif
            endif
         enddo

C    --- LOOP OVER ALL PACKAGES ---

         do q = 1,usepack
            do p = q,usepack

C          --- COMPUTE Re AND Im (c(t) exp(i ebar t)) FOR ALL E_BAR ---

               sinus   = sin(time*emin)
               cosinus = cos(time*emin)
               si      = sin(0.5d0*time*ediff)
               co      = cos(0.5d0*time*ediff)
               do n = 2,2*edim
                  fourier1 = reauto(p,q,i)*sinus+imauto(p,q,i)*cosinus
                  fourier2 = reauto(p,q,i)*cosinus-imauto(p,q,i)*sinus
                  tmp      = sinus*co+cosinus*si
                  cosinus  = cosinus*co-sinus*si
                  sinus    = tmp
                  
C             --- DETERMINE LOWER LEFT HALF OF K-MATRIZES ---

                  do m = mod(n,2),min(n-2,2*edim-n),2
                     j = (n+m)/2
                     k = (n-m)/2
                     integral1 = deltat*fourier1*dampvalue1(m)
                     integral2 = deltat*fourier2*dampvalue2(m)
                     k1(j+(p-1)*edim,k+(q-1)*edim) =
     +                    k1(j+(p-1)*edim,k+(q-1)*edim)+integral1
                     k2(j+(p-1)*edim,k+(q-1)*edim) =
     +                    k2(j+(p-1)*edim,k+(q-1)*edim)+integral2
                  enddo
               enddo
            enddo
         enddo
      enddo

C --- MIRROR HAMILTON-MATRICES FOR NON-DIAGONAL BLOCKS ---

      do q = 1,usepack
         do p = q+1,usepack
            do k = 1,edim
               do j = k+1,edim
                  k1(k+(p-1)*edim,j+(q-1)*edim) =
     +               k1(j+(p-1)*edim,k+(q-1)*edim)
                  k2(k+(p-1)*edim,j+(q-1)*edim) =
     +               k2(j+(p-1)*edim,k+(q-1)*edim)
               enddo
            enddo
         enddo
      enddo

C --- CALCULATE UPPER RIGHT HALF OF K-MATRICES ---

      do k = 1,hamdim
         do j = k+1,hamdim
            k1(k,j) = k1(j,k)
            k2(k,j) = k2(j,k)
         enddo
      enddo

C --- COMPUTE SECOND HAMILTON MATRIX K2 ---

      do j = 1,edim
         do k = 1,edim
            ebar   = emin+0.5d0*(k+j-2)*ediff
            deltae = (j-k)*ediff
            factor = ebar**2-(0.5d0*deltae)**2
            do q = 1,usepack
               do p = 1,usepack
                  k2(k+(p-1)*edim,j+(q-1)*edim) =
     +                k2(k+(p-1)*edim,j+(q-1)*edim)
     +               +factor*h0(k+(p-1)*edim,j+(q-1)*edim)
     +               +2.0d0*ebar*k1(k+(p-1)*edim,j+(q-1)*edim)
               enddo
            enddo
         enddo
      enddo

C --- COMPUTE FIRST HAMILTON MATRIX K1 ---

      do j = 1,edim
         do k = 1,edim
            ebar = emin+0.5d0*(k+j-2)*ediff
            do q = 1,usepack
               do p = 1,usepack
                  k1(k+(p-1)*edim,j+(q-1)*edim) =
     +               k1(k+(p-1)*edim,j+(q-1)*edim)
     +              +ebar*h0(k+(p-1)*edim,j+(q-1)*edim)
               enddo
            enddo
         enddo
      enddo

C --- RETURN TO CALLING ROUTINE ---

      return
      end
