
C **********************************************************************
C *                                                                    *
C *                              INPUT (input.f)                       *
C *                                                                    *
C * Program package containing the subroutines that read the input and *
C * autocorrelation files.                                             *
C *                                                                    *
C * Michael Beck                                                  5/96 *
C *                                                                    *
C **********************************************************************


C ----------------------------------------------------------------------
C                         SUBROUTINE READINPUT
C
C PURPOSE:
C     Reads the input file that must reside in the same directory than
C     the program itself.
C INPUT PARAMETERS:
C     none
C OUTPUT PARAMETERS:
C     path: path to directory containing the autocorrelation files
C     asciiflag: 0: autocorrelation is given as binary file
C                1: autocorrelation is given as ascii file
C     exactflag: 0: no file with exact results is read
C                1: the exact results are read from file "exactname"
C     exactname: name of file the exact results are stored in
C     dampautoflag: 0: no file with weighted autocorrelation function is
C                      generated
C                   1: the weighted autocorrelation function is written
C                      to file "dampautoname"
C     damppoints: number of energy grid points
C     outname: name of the output files
C     econv: conversion for energy unit (input is multiplied with econv)
C     tconv: conversion for time unit (input is multiplied with tconv)
C     inttime: time interval for integrals
C     emin, emax: lower and upper boundary of energy range
C     edim: number of energy points (and length of Hamiltonian matrices)
C     dampconst: damping constant 
C     damptype: selector for damping function
C     autocorr: defines which autocorrelation functions to be used
C     errortype: selector for the different internal error checks
C     vartype: selector for variational principle
C     sigma: energy shift used for the variational principle
C     vartype2: variational principle used for internal error control
C               (must be different from "vartype"; 0 means none)
C     sigma2: same as "sigma", but for second variational principle
C     epscut: cut-off parameter for eigenvalues of overlap matrix
C     spurcut: cut-off for spurious eigenvalues (all eigenvalues with
C              internal error I greater than spurcut*I_max are deleted)
C     deltat: after every period of "deltat" the autocorrelation
C             function has been written to disk
C     hamshift: energy shift of the Hamiltonian
C     sodflag: 1 if propagation has been done with SOD, else 0
C     proptime: period for which the autocorrelation function is known
C     npacket: number of packets stored in cross-correlation function.
C     usepack: number of packets to be employed.
C     skipauto: only every "skipauto"th value of "auto" will be used.
C     hamdim: size of Hamiltonian matrices.
C     damptime: time defining the width of the damp function
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C     intenspack: number of the packet for which the intensity is
C                 computed.
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open input file
C            3: input file too long
C            4: too many entries in input file
C            5: too few energy points
C            6: illegal variational principle for internal error control
C            7: I/O-error (fdcpinput)
C            8: cannot open file (fdcpinput)
C           13: no cross-correlation implemented for use of psi
C           14: integration time exceeds propagation time 
C           15: too many packets specified
C ----------------------------------------------------------------------

      subroutine readinput (path, asciiflag, exactflag, exactname,
     +                      dampautoflag, damppoints, outname, econv,
     +                      tconv, inttime, emin, emax, edim, dampconst,
     +                      damptype, autocorr, errortype, vartype,
     +                      sigma, vartype2, sigma2, epscut, spurcut,
     +                      deltat, hamshift, sodflag, proptime,
     +                      npacket, usepack, skipauto, hamdim,
     +                      damptime, refener, intenspack, error)

      implicit none

#include "fkeyword.inc"

      integer       asciiflag, exactflag, dampautoflag, damppoints,
     +              edim, damptype, autocorr, errortype, vartype,
     +              vartype2, sodflag, npacket, usepack, skipauto,
     +              hamdim, intenspack, error
      real*8        econv, tconv, inttime, emin, emax, dampconst, sigma,
     +              sigma2, epscut, spurcut, deltat, hamshift, proptime,
     +              damptime, refener
      character*(*) path, exactname, outname

      integer entries, cmdentries, idummy
      real*8  deltae, rdummy

C --- INITIALISE VARIABLES ---

      initial = .true.
      error = 0

C --- GET INPUT FILE AND COUNT ENTRIES ---

      call getinput(entries,cmdentries,error)
      if (error .ne. 0) return

C --- READ PARAMETERS ---
      
      call getrpara('prop_outputinterval',deltat,'0.0d0')
      call getrpara('prop_timeconversion',tconv,'41.3414d0')
      call getrpara('prop_energyshift',hamshift,'0.0d0')
      call getlpara('prop_sodintegrator',sodflag,'no,yes','no')
      call getipara('prop_skipauto',skipauto,'1')
      call getipara('prop_wavepackets',npacket,'1')
      call getipara('prop_usepackets',usepack,'0')
      call getspara('file_autopath',path,'.')
      call getlpara('file_autoformat',asciiflag,'binary,ascii,mctdh',
     +              'mctdh')
      call getipara('file_psilength',idummy,'0')
      call getspara('file_exactname',exactname,' ')
      call getspara('file_outputname',outname,'filter')
      call getlpara('file_fourier',dampautoflag,'no,yes','no')
      call getipara('file_fourierresolution',damppoints,'512')
      call getrpara('window_minimumenergy',emin,'0.0d0')
      call getrpara('window_energywidth',deltae,'0.0d0')
      call getipara('window_energypoints',edim,'0')
      call getrpara('window_energyconversion',econv,'0.0367493d0')
      call getrpara('window_referenceenergy',refener,'0.0d0')
      call getipara('window_intensitypacket',intenspack,'1')
      call getrpara('filter_timeinterval',inttime,'0.0d0')
      call getlpara('filter_function',damptype,'gauss,cos,cos^2,box',
     +              'cos')
      call getrpara('filter_gausswidth',dampconst,'0.3d0')
      call getlpara('filter_autoorder',autocorr,'zero,two','two')
      call getrpara('filter_cutoff',epscut,'1.0d-10')
      call getlpara('vp_principle',vartype,'H,H^2,1/H','1/H')
      call getrpara('vp_energyshift',sigma,'0.0d0')
      call getlpara('vp_iterative',idummy,'no,yes','no')
      call getrpara('vp_accuracy',rdummy,'1.0d-9')
      call getlpara('error_estimate',errortype,'stddeviation,variation',
     +              'variation')
      call getrpara('error_spurious',spurcut,'0.1d0')
      call getlpara('error_principle',vartype2,'H,H^2,1/H','H')
      call getrpara('error_energyshift',sigma2,'0.0d0')
      call getlpara('error_iterative',idummy,'no,yes','no')

C --- DETERMINE PERIOD FOR WHICH AUTOCORRELATION IS KNOWN ---

      if (sodflag .eq. 0) then
         call getautolen(path,asciiflag,deltat,proptime,npacket,error)
         if (error .ne. 0) then
            error = error+8
            return
         endif
      else
         call readdata(path,proptime,deltat,hamshift,error)
         if (error .ne. 0) then
            error = error+10
            return
         endif
         proptime = proptime/tconv
         deltat   = deltat/tconv
         hamshift = hamshift/econv
      endif

C --- CHANGE SOME DEFAULTS IN DEPENDENCE OF OTHER PARAMETERS ---

      if ((sodflag .eq. 0) .and. (asciiflag .eq. 2)) then
         call ignorekey('prop_outputinterval')
         call chrdef('prop_outputinterval',deltat,deltat)
         call ignorekey('prop_wavepackets')
         call chidef('prop_wavepackets',npacket,npacket)
      endif
      if (sodflag .eq. 1) then
         call ignorekey('prop_outputinterval')
         call chrdef('prop_outputinterval',deltat,deltat)
         call ignorekey('prop_energyshift')
         call chrdef('prop_energyshift',hamshift,hamshift)
      endif
      call chrdef('filter_timeinterval',inttime,proptime)
      call chrdef('vp_energyshift',sigma,emin+0.5d0*deltae)
      call chrdef('error_energyshift',sigma2,emin+0.5d0*deltae)
      if (vartype+1 .eq. 3) then
         call chldef('error_principle',vartype2,'H,H^2,1/H','H')
      else
         call chldef('error_principle',vartype2,'H,H^2,1/H','1/H')
      endif
      if (usepack .eq. 0) then
         call chidef('prop_usepackets',usepack,npacket)
         usepack = npacket
      endif

C --- SET ADDITIONAL PARAMETERS ---

      if (exactname .eq. ' ') then
         exactflag = 0
      else
         exactflag = 1
      endif
      vartype   = vartype+1
      vartype2  = vartype2+1
      emin      = emin+refener
      emax      = emin+deltae
      sigma     = sigma+refener
      sigma2    = sigma2+refener
      spurcut   = 0.5d0*abs(emin+emax)*spurcut
      if (usepack .gt. 1) edim = edim/usepack
      hamdim    = edim*usepack
      if (skipauto .gt. 1) deltat = dble(skipauto)*deltat
      if (damptype .eq. 0) then
         damptime = 0.5d0*dampconst*inttime
      else
         damptime = 0.5d0*inttime
      endif

C --- CHECK THE INPUT ---

      if (edim .le. 1) then
         error = 5
         return
      endif
      if ((errortype .eq. 1) .and. (vartype2 .eq. 1) .and.
     +    (vartype .eq. 1)) then
         error = 6
         return
      endif
      if (inttime .gt. proptime) then
         error = 14
         return
      endif
      if (usepack .gt. npacket) then
         error = 15
         return
      endif
      if (intenspack .gt. npacket .or. intenspack .le. 0) then
         error = 16
         return
      endif

C --- CONVERT TO DESIRED UNIT SYSTEM ---

      inttime  = tconv*inttime
      proptime = tconv*proptime
      deltat   = tconv*deltat
      damptime = tconv*damptime
      hamshift = econv*hamshift
      refener  = econv*refener
      emin     = econv*emin
      emax     = econv*emax
      deltae   = econv*deltae
      sigma    = econv*sigma
      sigma2   = econv*sigma2
      spurcut  = econv*spurcut

C --- LABEL REDUNDANT PARAMETERS AS IGNORED ---

      if (dampautoflag .eq. 0) call ignorekey('file_fourierresolution')
      if (damptype     .ne. 0) call ignorekey('filter_gausswidth')
      if (vartype      .eq. 1) call ignorekey('vp_energyshift')
      if (errortype    .eq. 1) call ignorekey('error_spurious')
      if (errortype    .ne. 1) call ignorekey('error_energyshift')
      if (errortype    .ne. 1) call ignorekey('error_principle')
      call ignorekey('file_psilength')
      call ignorekey('vp_accuracy')
      call ignorekey('vp_iterative')
      call ignorekey('error_iterative')

C --- ENSURE THAT COMPULSORY PARAMETERS ARE PRESENT ---

      if ((asciiflag .ne. 2) .and. (sodflag .ne. 1))
     +   call compulsory('prop_outputinterval')
      call compulsory('window_minimumenergy')
      call compulsory('window_energywidth')
      call compulsory('window_energypoints')

C --- CHECK FOR UNKNOWN KEYWORDS IN THE INPUT FILE ---

      call checkinput(entries,cmdentries)

C --- WRITE INPUT PARAMETERS TO FILE ---

      call fdcpinput(path,outname,error)
      if (error .ne. 0) then
         error = error+6
         return
      endif

C --- RETURN TO CALLING ROUTINE ---

      return
      end

C ----------------------------------------------------------------------
C                          SUBROUTINE GETINPUT
C
C PURPOSE:
C     Reads a file from the current directory and removes all
C     blanks and comments.
C INPUT PARAMETERS:
C     input, inplength: see "fkeyword.inc"
C OUTPUT PARAMETERS:
C     entries: number of entries in input file
C     cmdentries: number of entries in command line
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open input file
C            3: input file too long
C            4: too many entries in input file
C ----------------------------------------------------------------------

      subroutine getinput (entries, cmdentries, error)

      implicit none

#include "fkeyword.inc"

      integer inputch
      parameter (inputch = 1)

      integer entries, cmdentries, error

      integer        args, myiargc, length, i
      character*24   buf
      character*160  inpfile
      character*9999 inpline

C --- INITIALISE VARIABLES ---

      error      = 0
      inplength  = 0
      cmdlength  = 0
      entries    = 0
      cmdentries = 0
      args       = myiargc()

C --- GET NAME OF INPUT FILE ---

      length = 0
      if(args .eq. 0 ) goto 50
      call mygetarg(args,inpfile)
      length = index(inpfile,' ')-1

      do i = 1, args
         call mygetarg(i,buf)
         if(buf .eq. '-h' ) then
            length=0
         elseif(buf .eq. '-ver' ) then
            call wrversionf77(6)
            stop
         endif
      enddo   
      

C --- CHECK WHETHER PARAMETER IS MISSING ---

 50   if ((length .eq. 0) .or. (args .gt. 2)) then
         write(6,*)
         write(6,'(a)')
     +      'purpose: computes spectra employing the FD method'
         write(6,'(a)') 'usage: filter<ver> [options] [<cmd>] <inp>'
         write(6,'(a)') '  ver: version number'
         write(6,'(a)') '  cmd: command line keywords'
         write(6,'(a)') '  inp: input file'
         write(6,'(/a)')'options:'
         write(6,'(a)') '  -h   :  print this help text.'
         write(6,'(a)')'  -ver : Version information about the program.' 
         write(6,*)
         stop
      endif

C --- OPEN INPUT FILE ---

      if (inpfile(length-3:length) .eq. '.inp') then
         open(inputch,file=inpfile(1:length),status='old',
     +        form='formatted',err=20)
      else
         open(inputch,file=inpfile(1:length)//'.inp',status='old',
     +        form='formatted',err=20)
      endif

C --- LOOP OVER ALL LINES ---

 100  continue
      read(inputch,'(a)',err=10,end=200) inpline
      
C --- REMOVE BLANKS AND COMMENTS ---

      call convertline(input,inplength,inpline,error)
      if (error .eq. 1) then
         error = 3
         return
      endif
      goto 100

C --- CLOSE FILE ---

 200  continue
      close(inputch)

C --- COUNT ENTRIES ---

      call countentries(input,inplength,entries)
      if (entries .gt. maxkeys) then
         error = 4
         return
      endif

C --- READ INPUT PARAMETERS FROM COMMAND LINE ---

      if (args .gt. 1) then
         call mygetarg(1,inpline)
         call convertline(cmdline,cmdlength,inpline,error)
         if (error .eq. 1) then
            error = 3
            return
         endif
         call countentries(cmdline,cmdlength,cmdentries)
         if (cmdentries .gt. maxkeys) then
            error = 4
            return
         endif
      endif

C --- RETURN TO CALLING ROUTINE ---

      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                         SUBROUTINE READDATA
C
C PURPOSE:
C     Reads the data file (all quantities must be given in a.u.).
C INPUT PARAMETERS:
C     path: path to directory containing the data file
C OUTPUT PARAMETERS:
C     proptime: propagation time
C     deltat: after every period of "deltat" the autocorrelation
C             function has been written to disk
C     hamshift: energy shift of the Hamiltonian
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open data file
C ----------------------------------------------------------------------

      subroutine readdata (path, proptime, deltat, hamshift, error)

      implicit none

      integer datach
      parameter (datach = 1)

      integer      error
      real*8       proptime, deltat, hamshift
      character*80 path

      integer pathlength

C --- INITIALISE VARIABLES ---

      error = 0
      pathlength = index(path,' ')-1

C --- GET DATA ---

      open(datach,file=path(1:pathlength)//'/data',form='formatted',
     +     status='old',err=20)
      read(datach,*,err=10) proptime
      read(datach,*,err=10) deltat
      read(datach,*,err=10) hamshift
      close(datach)

C --- RETURN TO CALLING ROUTINE ---

      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                        SUBROUTINE GETAUTOLEN
C
C PURPOSE:
C     Determines the length of the autocorrelation file and computes the
C     propagation time. It also determines the output interval and the
C     number of packets.
C INPUT PARAMETERS:
C     path: path to directory containing the autocorrelation files
C     asciiflag: 0: autocorrelation is given as binary file
C                1: autocorrelation is given as ascii file
C     deltat: output interval of the autocorrelation function (input
C             parameter only for file_autoformat=binary,ascii)
C OUTPUT PARAMETERS:
C     proptime: period for which the autocorrelation function is known
C     deltat: output interval of the autocorrelation function (output
C             parameter only for file_autoformat=mctdh)
C     npacket: number of packets stored in the cross-correlation
C              function
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open autocorrelation file
C ----------------------------------------------------------------------

      subroutine getautolen (path, asciiflag, deltat, proptime, npacket,
     +                       error)

      implicit none

      integer autoch
      real*8  large, tiny
      parameter (autoch = 1, large = 1.0d99, tiny = 1.0d-12)

      integer      asciiflag, npacket, error
      real*8       deltat, proptime
      character*80 path

      logical       ready
      integer       pathlength, i
      real*8        time,oldtime,dummy
      character*160 string

C --- INITIALISE VARIABLES ---

      ready = .false.
      error = 0
      pathlength = index(path,' ')-1
      if (asciiflag .eq. 2) npacket = 0
      oldtime = large

C --- SCAN AUTOCORRELATION FILE ---

      if (asciiflag .eq. 0) then
         open(autoch,file=path(1:pathlength)//'/auto',
     +        form='unformatted',status='old',err=20)
      else
         open(autoch,file=path(1:pathlength)//'/auto',
     +        form='formatted',status='old',err=20)
      endif
      i = 0
 100  continue
      if (asciiflag .eq. 0) then
         read(autoch,err=10,end=200) dummy
         read(autoch,err=10)         dummy
      elseif (asciiflag .eq. 1) then
         read(autoch,*,err=10,end=200) dummy, dummy
      else
         read(autoch,'(a)',err=10,end=200) string
         if (string(1:1) .eq. '#') goto 100
         if (string(1:30).eq.'                              ') goto 100
         read(string,*,err=10) time, dummy, dummy, dummy
         if (.not. ready) then
            if (time-oldtime .gt. tiny) then
               deltat = time-oldtime
               npacket = nint(sqrt(dble(npacket)))
               ready = .true.
            else
               oldtime = time
               npacket = npacket+1
            endif
         endif
      endif
      i = i+1
      goto 100
 200  continue
      close(autoch)

C --- COMPUTE PROPAGATION TIME AND RETURN TO CALLING ROUTINE ---

      proptime = dble(((i/npacket**2)-1))*deltat
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                         SUBROUTINE READAUTO
C
C PURPOSE:
C     Reads the autocorrelation files.
C INPUT PARAMETERS:
C     autodim: length of autocorrelation vectors
C     path: path to directory containing the autocorrelation files
C     asciiflag: 0: autocorrelation is given as binary file
C                1: autocorrelation is given as ascii file
C     autocorr: 0: read only zero order autocorrelation function
C               1: read also higher order autocorrelation functions
C     npacket: number of packets stored in cross-correlation function.
C     usepack: number of packets to be employed.
C     skipauto: only every "skipauto"th value of "auto" will be used.
C OUTPUT PARAMETERS:
C     reauto0, imauto0: real and imaginary part of 1st autocorrelation
C                       function <dconjg(psi(t))|psi(t)>, respectively
C     reauto1, imauto1: real and imaginary part of 2nd autocorrelation
C                       function <dconjg(psi(t))|H|psi(t)>, respectively
C                       (only if "damptype" has corresponding value)
C     reauto2, imauto2: real and imaginary part of 3rd autocorrelation
C                       function <dconjg(psi(t))|H^2|psi(t)>,
C                       respectively (only if "damptype" has
C                       corresponding value)
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open autocorrelation file
C            3: insufficient length of autocorrelation file
C ----------------------------------------------------------------------

      subroutine readauto (autodim, path, asciiflag, autocorr, npacket,
     +                     usepack, skipauto, reauto0, imauto0, reauto1,
     +                     imauto1, reauto2, imauto2, error)

      implicit none

      integer autoch1, autoch2, autoch3
      parameter (autoch1 = 1, autoch2 = 2, autoch3 = 3)

      integer      autodim, asciiflag, autocorr, npacket, usepack,
     +             skipauto, error
      real*8       reauto0(1:(autodim+1)*usepack**2),
     +             imauto0(1:(autodim+1)*usepack**2),
     +             reauto1(1:(autodim+1)*usepack**2),
     +             imauto1(1:(autodim+1)*usepack**2),
     +             reauto2(1:(autodim+1)*usepack**2),
     +             imauto2(1:(autodim+1)*usepack**2)
      character*80 path

      integer pathlength

C --- INITIALISE VARIABLES ---

      error      = 0
      pathlength = index(path,' ')-1

C --- OPEN AUTOCORRELATION FILES ---

      if (asciiflag .eq. 0) then
         open(autoch1,file=path(1:pathlength)//'/auto',
     +        form='unformatted',status='old',err=20)
      else
         open(autoch1,file=path(1:pathlength)//'/auto',
     +        form='formatted',status='old',err=20)
      endif
      if (autocorr .ne. 0) then
         if (asciiflag .eq. 0) then
            open(autoch2,file=path(1:pathlength)//'/auto1',
     +           form='unformatted',status='old',err=20)
         else
            open(autoch2,file=path(1:pathlength)//'/auto1',
     +           form='formatted',status='old',err=20)
         endif
         if (asciiflag .eq. 0) then
            open(autoch3,file=path(1:pathlength)//'/auto2',
     +           form='unformatted',status='old',err=20)
         else
            open(autoch3,file=path(1:pathlength)//'/auto2',
     +           form='formatted',status='old',err=20)
         endif
      endif

C --- READ AUTOCORRELATION FILES ---

      call getauto(autoch1,autodim,asciiflag,npacket,usepack,skipauto,
     +             reauto0,imauto0,error)
      if (error .ne. 0) return
      if (autocorr .ne. 0) then
         call getauto(autoch2,autodim,asciiflag,npacket,usepack,
     +                skipauto,reauto1,imauto1,error)
         if (error .ne. 0) return
         call getauto(autoch3,autodim,asciiflag,npacket,usepack,
     +                skipauto,reauto2,imauto2,error)
         if (error .ne. 0) return
      endif

C --- CLOSE AUTOCORRELATION FIlES ---

      close(autoch1)
      if (autocorr .ne. 0) then
         close(autoch2)
         close(autoch3)
      endif

C --- RETURN TO CALLING ROUTINE ---

      return

C --- SET ERROR FLAG ---

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                         SUBROUTINE GETAUTO
C
C PURPOSE:
C     Reads a single autocorrelation file.
C INPUT PARAMETERS:
C     autoch: channel linked to the autocorrelation file
C     autodim: length of autocorrelation vectors
C     asciiflag: 0: autocorrelation is given as binary file
C                1: autocorrelation is given as ascii file
C     npacket: number of packets stored in cross-correlation function.
C     usepack: number of packets to be employed.
C     skipauto: only every "skipauto"th value of "auto" will be used.
C OUTPUT PARAMETERS:
C     reauto, imauto: real and imaginary part of autocorrelation
C                     function <dconjg(psi(t))|psi(t)>
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            3: insufficient length of autocorrelation file
C ----------------------------------------------------------------------

      subroutine getauto (autoch, autodim, asciiflag, npacket, usepack,
     +                    skipauto, reauto, imauto, error)

      implicit none

      integer      autoch, autodim, asciiflag, npacket, usepack,
     +             skipauto, error
      real*8       reauto(1:(autodim+1)*usepack**2),
     +             imauto(1:(autodim+1)*usepack**2)

      integer       i, j, k, l
      real*8        dummy
      character*160 string

C --- INITIALISE VARIABLES ---

      error = 0
      i     = 1
      j     = 0
      k     = 0
      l     = 1

C --- READ AUTOCORRELATION FILE ---

 100  continue
      if (l .gt. (autodim+1)*usepack**2) goto 200
      if (asciiflag .eq. 0) then
         read(autoch,err=10,end=30) reauto(l)
         read(autoch,err=10)        imauto(l)
      elseif (asciiflag .eq. 1) then
         read(autoch,*,err=10,end=30) reauto(l), imauto(l)
      else
         read(autoch,'(a)',err=10,end=30) string
         if (string(1:1) .eq. '#') goto 100
         read(string,*,err=10) dummy, reauto(l), imauto(l), dummy
      endif
      j = mod(j,npacket)+1
      if (j .eq. 1) k = mod(k,npacket)+1
      if ((j .le. usepack) .and. (k .le. usepack) .and.
     +    (mod((i-1)/npacket**2,skipauto) .eq. 0)) l = l+1
      i = i+1
      goto 100

C --- RETURN TO CALLING ROUTINE ---

 200  continue
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 30   continue
      error = 3
      return

      end

C ----------------------------------------------------------------------
C                        SUBROUTINE GETEXACTLEN
C
C PURPOSE:
C     Determines the number of exact energies and intensities.
C INPUT PARAMETERS:
C     exactname: name of file the exact results are stored in
C     eminin, emaxin: lower/upper boundary of energy window
C     econv: conversion for energy unit (energies are divided by econv)
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C OUTPUT PARAMETERS:
C     maxexactdim: number of exact energies
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open exact file
C ----------------------------------------------------------------------

      subroutine getexactlen (exactname, eminin, emaxin, econv, refener,
     +                        maxexactdim, error)

      implicit none

      integer exactch
      parameter (exactch = 1)

      integer      maxexactdim, error
      real*8       eminin, emaxin, econv, refener
      character*80 exactname

      integer exactlength
      real*8  energy, rdummy

C --- INITIALISE VARIABLES ---

      error = 0
      maxexactdim = 0
      exactlength = index(exactname,' ')-1

C --- OPEN "EXACT" FILE ---

      open(exactch,file=exactname(1:exactlength),form='formatted',
     +     status='old',err=20)

C --- READ "EXACT" FILE ---

 100  continue
      read(exactch,*,err=10,end=200) energy,rdummy
      energy = energy+refener/econv
      if ((energy .ge. eminin/econv) .and. (energy .le. emaxin/econv))
     +  then
         maxexactdim = maxexactdim+1
      endif
      goto 100

C --- ENLARGE LIST FOR FIRST VALUE BELOW AND ABOVE ENERGY WINDOW ---

 200  continue
      maxexactdim = maxexactdim+2

C --- CLOSE FILE AND RETURN TO CALLING ROUTINE ---

      close(exactch)
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

      end

C ----------------------------------------------------------------------
C                         SUBROUTINE READEXACT
C
C PURPOSE:
C     Reads the exact energies and intensities from file.
C INPUT PARAMETERS:
C     exactname: name of file the exact results are stored in
C     eminin, emaxin: lower/upper boundary of energy window
C     econv: conversion for energy unit (energies are divided by econv)
C     maxexactdim: allocated length of "exactenergy" and
C                  "exactintensity"
C     refener: reference energy that is subtracted from the
C              eigenenergies.
C OUTPUT PARAMETERS:
C     exactdim: number of exact energies
C     exactenergy: vector containing exact energies
C     exactintensity: vector containing exact intensities
C     error: error code having the following meaning:
C            0: everything is o. k.
C            1: an I/O-error occured
C            2: can't open exact file
C            3: "exact" file is too long
C ----------------------------------------------------------------------

      subroutine readexact (exactname, eminin, emaxin, econv,
     +                      maxexactdim, refener, exactdim, exactenergy,
     +                      exactintensity, error)

      implicit none

      integer exactch
      parameter (exactch = 1)

      integer      maxexactdim, exactdim, error
      real*8       eminin, emaxin, econv, refener,
     +             exactenergy(1:maxexactdim),
     +             exactintensity(1:maxexactdim)
      character*80 exactname

      integer exactlength
      real*8  energy, intensity

C --- INITIALISE VARIABLES ---

      error = 0
      exactdim = 0
      exactlength = index(exactname,' ')-1

C --- OPEN "EXACT" FILE ---

      open(exactch,file=exactname(1:exactlength),form='formatted',
     +     status='old',err=20)

C --- READ "EXACT" FILE ---

C Note that the first element below and above the window is also stored.

 100  continue
      read(exactch,*,err=10,end=200) energy, intensity
      energy = energy+refener/econv
      if (econv*energy .lt. eminin) then
         exactdim = 1
      else
         exactdim = exactdim+1
         if (exactdim .gt. maxexactdim) goto 30
      endif
      exactenergy(exactdim)    = econv*energy
      exactintensity(exactdim) = intensity
      if (econv*energy .gt. emaxin) then
         goto 200
      endif
      goto 100

C --- CLOSE FILE AND RETURN TO CALLING ROUTINE ---

 200  continue
      close(exactch)
      return

C --- SET ERROR FLAG ---

 10   continue
      error = 1
      return

 20   continue
      error = 2
      return

 30   continue
      error = 3
      return

      end
