!***********************************************************************
! 
!                         ANALYSE program EFIELD
!
!     This program calculates the overlap between the wave functions
!     on the restart file of the current directory and a WF on a psi
!     file. 
!     From these overlaps it computes the electric field for optimal
!     control. 
!
!     The wave functions to be overlapped are forward and backward 
!     propageted WFs. Both WFs must have two electronic states,
!     same grids and even the same numbers of SPFs.
!
!     04/05 HDM and LW
!     05/06 IA and HDM,  fifo implemented, trig. extrapol.
!     09/07 MS and AB,   general operators implemented.
!
!
!***********************************************************************

      program efield

use decimal, only:long,dop,sip
use global
use globmemmod
use versions
use aglobal
use maxv
use dvrdatmod
use griddatmod
use psidef
use datenmod
use operdef
use hpsimod
use oct
use channels
use runpropmod
use compdatmod

use op1lib, only: cpvcz, zerovxz, cpvxz
use logdat
use compdat, only: savecsys,savecpsi,getcpsi
use iorst, only: rstinfo
use iofile, only: rdpsiinfo
use rddvrmod, only: dvrinfo
use rdopermod, only: operinfo
use rdfilesmod, only: rdpsi
use auswutil, only: savecdat
use zeigauswmod, only: zeigausw
use zeighpsimod, only: operflag,zeighpsi
use iopsidef, only: psidat
use efinput
use efroutines
use efoutput
use memhopsmod, only: memhops

use module_interface

      implicit none

      integer(long), external              :: slen
      integer(long)                        :: nham
      integer(long)                        :: zeig,zeig1,m,ilbl,chkdvr,chkgrd, &
                                              chkpsi,chkdat,chkprp,n,s,i
      real(dop)                            :: time,totmem
      character(len=c5)                    :: filename
      logical(kind=4)                      :: lerr,linwf,check

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault
      call ef_input
      call ef_handshake
      tfinal=dble(nfin)*out2

!-----------------------------------------------------------------------
! get basic dimensions (e.g. maxdim, maxsta) from restart file
!-----------------------------------------------------------------------
      inquire(irst,opened=check)
      if (check) close(irst)
      filename='initialstate/restart'
      ilbl=index(filename,' ')-1
      open(irst,file=filename(1:ilbl),form='unformatted',status='old',err=900)
      call rdmemdim(irst)
      close(irst)
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_hpsimod
      call alloc_psidef
      call alloc_compdat
      allocate(dovmat(maxdim,maxsta))

! --- Set dummies for GWP arrays. 
! --- GWP OCT not implemented yet.
      ghmatdim = 1
      hungdim  = 1  

!-----------------------------------------------------------------------
! set no. of comparison data sets
!-----------------------------------------------------------------------
      ncomp=2
      lwrlog = .true.
 
! --- open the files psi, psi1 and ilog --
      call ef_open 
      
!-----------------------------------------------------------------------
! read data from initialstate file (ipsi)
!-----------------------------------------------------------------------
      chkdvr=1
      chkgrd=1
      chkpsi=1
      chkdat=1
      chkprp=0

      if (lback .and. .not. lsimprop) then
!     read data from initialstate file (ipsi, psi) if backward
!     if simprop get data from restart
         read(ipsi) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
         filename=filein
         if (chkerr(1) .ne. 0) go to 999
         call chksyserr(filename,lerr)
      else
!        read data from initialstate file (ipsi, restart) if forward
         irst = ipsi
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
         filename=filein
         if (lerr) goto 999
! NOT USED
!         read(irst,err=999) time
         close(irst)
      end if
      
!-----------------------------------------------------------------------
! Get operator number 
! The operator will be used to calculate the field
!-----------------------------------------------------------------------
      if (lrdoper) then

         chkdvr=2
         chkgrd=2
         call dvrinfo(lerr,chkdvr)
         if (lerr) goto 996
         call operinfo(lerr,chkdvr,chkgrd)
         if (lerr) goto 997

         nham=0
         ntarop = 0
         do n=1,hamdim
            if (hamlab(n) .eq. operator) nham=n
            if (ltarop) then
               if (hamlab(n) .eq. tarop) ntarop=n
            end if
         enddo
     
!     Don't ask the user if operators are not found because
!     the program is likely to be run by a script.
!     Insted qiot with an error message.

         routine="efield (efield.F)"
         if (nham .eq. 0) then
            message='(Dipole) operator "'//operator(1:slen(operator)) &
                 //'" not found in '//operfile(1:slen(operfile))
            call errormsg
         endif
         if (ntarop .eq. 0 .and. ltarop) then
            message='(Target) operator "'//tarop(1:slen(tarop)) &
                 //'" not found in '//operfile(1:slen(operfile))
            call errormsg
         endif
         if(.not.nodiag(nham)) then
            ilbl=index(hamlab(nham),' ')-1
            message='The Hamiltonian '//hamlab(nham)(1:ilbl)// &
                 ' must be of "nodiag" type.'
            call errormsg
         endif
         if (ltarop) then
            if(.not.nodiag(ntarop)) then
               ilbl=index(hamlab(ntarop),' ')-1
               message='The Hamiltonian '//hamlab(nham)(1:ilbl)// &
                 ' must be of "nodiag" type.'
               call errormsg
            endif
         endif
      endif

      !save system in both comparison sets (just for memausw)
      ncomp=2
      call savecsys(1)
      call savecpsi(1)
      call savecdat(1)

!-----------------------------------------------------------------------
! read data from tagetstate file (ipsi1)
!-----------------------------------------------------------------------
      chkdvr=2
      chkgrd=2
      chkpsi=2
      chkdat=1
      chkprp=0
      if (.not. lback .and. .not. lsimprop) then
!     read data from targetstate file (ipsi1, psi) if forward
!     if simprop get data from restart
         read(ipsi1) filever(ipsi1)
         call rdpsiinfo(ipsi1,chkdvr,chkgrd,chkpsi,chkdat)
         filename=filein1
         if (chkerr(1) .ne. 0) go to 999
         call chksyserr(filename,lerr)
      else
!        read data from targetstat file (ipsi1, restart) if backward
         irst = ipsi1
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
         filename=filein1
         if (lerr) goto 999
! NOT USED
!         read(irst,err=999) time
         close(irst)
      end if
      
      call savecsys(2)
      call savecpsi(2)
      call savecdat(2)

!-----------------------------------------------------------------------
! ERROR checks
!-----------------------------------------------------------------------

      ovtype = 0
      ebtype = 0
      if (chkerr(2).ne.0 .or. chkerr(3).ne.0) then
         routine='efield (efield.F)'
         message='Different primitive bases'
         call errormsg
      endif
      if (chkerr(4) .ne. 0) then
         ovtype = 1
         routine='efield (efield.F)'
         message='ovtype=1 not implemented'
         call errormsg
      endif
      if (chkerr(5) .ne. 0) then
         ebtype = 1
         routine='efield (efield.F)'
         message='ebtype=1 not implemented'
         call errormsg
      endif
      if (dentype .ne. 0 ) then
         routine='efield'
         message='Not implemented for density matrices'
         call errormsg
      endif
      
      if (npacket.gt.8) then
         routine='efield (efield.F)'
         message="Format of some output files may screw up for" &
               //" more then 8 targets"
         call warnmsg
         lwarn=.false.
      endif
      
!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
!-----------------------------------------------------------------------

      lpsi    = .true.
      lpsi1   = .true.
      lort    = lrdoper
      lsphfbr = lrdoper
      ltrafo  = .false.
      ldmat   = .true.
      ldmat1  = .false.
      lgpop   = .false.
      lpsigrd=ovtype.eq.1 !false

      call zeigausw

!     compute size of overlap matrices 
!     (assume same structure of target and initial waves) 

      ovmaxblock=0
      zeig=1
 
      ! ovtype = 0, dentype = 0
      do s=1,nstate
         ovmaxblock=max(ovmaxblock,block(s),cblock(s,2))
         zeig1=cblock(s,2)
         do m=1,nmode
            dovmat(m,s)=zeig
            zeig=zeig+dim(m,s)*cdim(m,s,2)
            zeig1=zeig1*dim(m,s)/cdim(m,s,2)
            ovmaxblock=max(ovmaxblock,zeig1)
         enddo
      enddo
          
      doverdim=zeig-1
      griddim=1

      call memausw

      mcovr=mcdim
      mcdim=mcovr+cdmatdim(1)
      mcovr1=mcdim
      mcdim=mcovr1+cdmatdim(2)
      
      mcdover=mcdim
      mcdim=mcdover+doverdim
      

!     get sizes of arrays for using operators
      if(lrdoper)  then
         call operflag(nham)
         call zeighpsi(nham)
         if (ltarop) then
            call operflag(ntarop)
            call zeighpsi(ntarop)
         end if
         mchpsi  = mcdim
         mcdim   = mchpsi+hpsidim
         mcdtpsi = mcdim
         mcdim   = mcdtpsi+dgldim
!         mchteil = mcdim
!         mcdim   = mchteil+pmatdim
         workcdim=max(workcdim,dgldim,3*ovmaxblock)
      else
         mcdtpsi = 1
         mchpsi  = 1
!         mchteil = 1
      end if
     
      workcdim=max(workcdim,2*ovmaxblock)
  
! ->  store the electric field here
      nfin = nfin+5

      mrfld = mrdim
      mrdim = mrdim + 3*nfin

      mrpnlty = mrdim
      mrdim = mrdim + nfin

      mrref = mrdim
      mrdim = mrdim + nfin

      mrlagr = mrdim
      mrdim = mrdim + nfin

      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      mrworkr = mrdim
      mrdim   = mrworkr+workrdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim 
      msdim   = msworks+worksdim

      mfworkf = mfdim 
      mfdim   = mfworkf+workfdim
      
      if(lrdoper) call memhops

!-----------------------------------------------------------------------
! call routine
!-----------------------------------------------------------------------

      call wrversion(ilog)
      write(ilog,'(a)') '  *** Memory allocated dynamically ***'
      write(ilog,'(6(a,i12,/),a,f16.0,a,f11.2,a)') &
       ' complex*16 :', mcdim, &
       ' real*8     :', mrdim, &
       ' integer    :', midim, &
       ' logical    :', mldim, &
       ' complex*8  :', msdim, &
       ' real*4     :', mfdim, &
       ' Total   :',totmem,' Bytes ,',totmem/1048576.d0,' MB' 
       write(ilog,'(80a1,/)') ('-',i=1,80)
       call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(hteil))deallocate(hteil)
if(allocated(agmat))deallocate(agmat)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(hteil(pmatdim))
allocate(agmat(agmatdim))

      call getcpsi(2)
      dgldim1=dgldim
      adim1=adim
      dmatdim1=dmatdim
     
      if(lrdoper) call ef_readoperator

      call ef_ovrlp(mc(mcpsi),mc(mcpsi1),ms(mspsi), mc(mcdover), &
           agmat,mc(mcworkc),mi(mijindx),mi(mijindx1),mr(mrfld), &
           mr(mrpnlty), mr(mrref),mr(mrlagr))

deallocate(mc,mr,mi,ml,ms,mf,hteil,agmat)
      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_operdef
      call dealloc_hpsimod
      call dealloc_psidef
      call dealloc_compdat
deallocate(dovmat)

      write(ilog,'(/a)') "Efield is terminating normally."
      close(ilog)

      stop 

 900  continue
      routine='efield'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg
      stop

 999  m=index(filename,' ')-1
      routine='efield'
      message= 'Error reading: '//filename(1:m)
      call errormsg
      stop
 
 997  continue
      routine='efield'
      message= 'Error reading oper-info from:'//operfile(1:opflaenge)
      call errormsg
      stop

 996  continue
      routine='efield'
      message= 'Error reading dvr-info from:'//dname(1:dlaenge)//'/dvr'
      call errormsg
      stop

      contains

!#######################################################################
!     EF_OVRLP:
!     Main analyse routine 
!#######################################################################

      subroutine ef_ovrlp (psi,psi1,spsi,dover,agmat,workc,jindx,jindx1, &
                 fld, pnltyfnc,reffnc,lagrfnc)

      implicit none

      integer(long), dimension(jindxdim), intent(out) :: jindx
      integer(long), dimension(adim1), intent(in)     :: jindx1
      integer(long) ::    i,itime,itime1,itime2, ierr
      real(dop) ::     rdum,val1,time
      real(dop), dimension(3,3)                       :: oldfld
      complex(sip), dimension(spsidim), intent(out)   :: spsi
      complex(dop), dimension(dgldim), intent(out)    :: psi
      complex(dop), dimension(dgldim), intent(out)    :: psi1
      complex(dop), dimension(workcdim), intent(out)  :: workc
      complex(dop), dimension(doverdim), intent(out)  :: dover
      complex(dop), dimension(agmatdim), intent(out)  :: agmat
      complex(dop), dimension(maxsta)                 :: ovlp1,ovlp,dipmtx,ctmp
      character(len=c5)                               :: filename
      logical(kind=4) :: lcalcnewovlp,lskipwr
      real(dop), dimension(nfin) :: pnltyfnc,reffnc,lagrfnc
      real(dop), dimension(nfin,3) :: fld
      real(dop), dimension(3) :: Y

      lskipwr=.false.
      lcalcnewovlp=.false.
      
      call ef_loadfncs(fld,pnltyfnc,reffnc,lagrfnc)
      do i=1,nfin
         pnltyfnc(i)=pnltyfnc(i)/pnlty 
      end do
      

      !read the overlap from last fw-propagation
      !after initial forward iteration lcalcnewovlp=true
      call ef_rdovlps(ovlp,npacket,lcalcnewovlp)
      ! conjg for krotov 
!      if (lkrotov) call cpvcz(ovlp,ovlp,npacket)

      if (lsimprop) then
!     Load the old field from last propagation. this has to bee fed 
!     into mctdh in simultaneous mode to 'backdate' the field
!     as mctdh works with the predicted field. This prediction
!     has to be reversed.(not in pc-mode)      
         if (lback) then
            filename='fieldb_init.dat'
         else
            filename='fieldf_targ.dat'
         endif
         open(iaus,file=filename,status='old')
         read(iaus,*) rdum,  (oldfld(1,i), i=1,3)
         read(iaus,*) rdum,  (oldfld(2,i), i=1,3)
         read(iaus,*) rdum,  (oldfld(3,i), i=1,3)
      endif
      
!     open output files
      call ef_openout()

!---------- The main loop ----------------------------------------------
      
      itime1 = -1
 100  continue
! - IA - Wait for the signal from prop.F that the restart file ----
! ------ is updated and  get the current time from fifo.   --------
         routine='ef_ovrlp (efield.F)'
         read(ipfifo,*) itime
         if (lsimprop) then
            read(ip1fifo,*) itime2
            if(itime .ne. itime2 ) then
               message='The instances of MCTDH are not synchronized'
               call errormsg
            end if
         endif
         if(itime.lt.0) goto 120        ! MCTDH is finished 
         if(itime.eq.itime1) goto 60    ! predictor-corrector
         itime1 = itime
         
!        read reference wave (old)
!         call ef_getrefwf(psi,spsi,jindx,itime)

         ! initial state is on psi, target on psi1
         if (lback) then
            call ef_getrefwf(psi,spsi,jindx,itime,agmat,mi(mitrajst), &
                 mc(mcadgwp),mi(migwpdep))
         else
            call ef_getrefwf(psi1,spsi,jindx,itime,agmat,mi(mitrajst), &
                 mc(mcadgwp),mi(migwpdep))
         end if

         if (loctpc.and.itime.ne.0) lskipwr=.true.
         goto 61

         !PC mode - corrector step
 60      continue
         fld(itime+3,3) = fld(itime+3,1)
         lskipwr=.false.
         if (.not.loctpc) then
            routine='ef_ovrlp (efield.F)'
            message='MCTDH is running in PC-mode, efield is not.'
            call errormsg
         endif

 61      continue
!        read restart wave (new)
!         call ef_getrstwf(psi1)
         ! initial state is on psi, target on psi1
         if (lback) then
            call ef_getrstwf(psi1)
         else
            call ef_getrstwf(psi)
         end if

         !very first backward propagation: need to calculate overlap
         if (itime.eq.0 .and. lcalcnewovlp .and. lback) then
            call ef_povlp(psi,psi1,dover,workc,ovlp)

            lcalcnewovlp=.false.
            call zerovxz(ovlp1,maxsta)
            if (ltarop) then
               !we are writing the target population at the end of 
               !the overlaps file - need to calculate it first. 
               time=0.0
               call ef_mtrxelement(ovlp1,psi,psi,workc,time, &
                    mc(mchpsi),hteil,mr(mrhops),mc(mcdtpsi), &
                    mr(mrort),ntarop,mi(mijindx))
!               call ef_mtrxelement(ovlp1,psi,psi,workc,time, &
!                    mc(mchpsi),mc(mchteil),mr(mrhops),mc(mcdtpsi), &
!                    mr(mrort),ntarop,mi(mijindx))
            end if
            call ef_wrovlpfile(ovlp,ovlp1,npacket)

!            if (lkrotov) call cpvcz(ovlp,ovlp,npacket)

         end if
         
!-----------------------------------------------------------------------
! calculate the matrix element of dipole operator first set time 
! (important if dipole is time-dependent, as used in pump-
! probe scenarios with perturbation theory)
!----------------------------------------------------------------------- 
         if (lback) then
            time=tfinal-dble(itime)*out2 
         else
            time=dble(itime)*out2
         end if
        
!     dipole matrixelement
         call ef_mtrxelement(dipmtx,psi1,psi,workc,time, &
              mc(mchpsi),hteil,mr(mrhops),mc(mcdtpsi), &
              mr(mrort),nham,mi(mijindx))
!         call ef_mtrxelement(dipmtx,psi1,psi,workc,time, &
!              mc(mchpsi),mc(mchteil),mr(mrhops),mc(mcdtpsi), &
!              mr(mrort),nham,mi(mijindx))
!         if (lkrotov) call cpvcz(dipmtx,dipmtx,npacket)

!     time-dependent overlap
         if ((ltdovlp .and. .not. ltarop) .or. lwrovlp) then
            call ef_povlp(psi,psi1,dover,workc,ovlp1)
!            if (lkrotov) call cpvcz(ovlp1,ovlp1,npacket)
            if (ltdovlp) call cpvxz(ovlp1,ovlp,npacket)
         end if

!     the field (without penalty, envelope and reference function)
         call ef_mkfield(dipmtx,ovlp,val1) 

!     if backwards: skip sending the field in krotov algorithm
         if (lkrotov .and. lback) goto 62
        
!     the field (with penalty, envelope and reference)
         if (lkrotov) then      !Krotov
            fld(itime+3,1) = reffnc(itime+3) + &
                 (val1-lagrfnc(itime+3) )*pnltyfnc(itime+3)/2.d0
         else !Rabitz
            fld(itime+3,1) = reffnc(itime+3) + val1*pnltyfnc(itime+3)
         end if

         Y(1)=fld(itime+3,1) 
         Y(2)=fld(itime+2,1) 
         Y(3)=fld(itime+1,1) 
         call ef_extrpltefld(Y,extorder,ierr)
         if (ierr.ne.0)then
            message="Extrapolation of field failed."
            routine="ef_ovrlp (efield.F)"
            call errormsg
         end if
         fld(itime+4,2) =Y(1)
         fld(itime+5,3) =Y(2)

 62      continue

!     send the new field (and old field if lsimprop) to mctdh
         call ef_sendfld(fld,oldfld,itime)
         
!     pc-mode: we are doing the prediction - don't write output
         if (lskipwr) goto 100

!     reset the time
         time=dble(itime)*out2

!     write time-dependent output (except field) on the fly 
         if (.not. lback .and. lwrpop) then
!     NOTE: ef_wrtargetpop overwrites psi1 with target
!           if target is a WF
            call ef_wrtargetpop(psi,psi1,time,workc)
         end if
         if (lwrrawfld)  call ef_wrrawfld(time,val1)
         if (lwrdipmtx) then
            if (lback) then 
               call cpvcz(dipmtx,ctmp,npacket)
               call ef_wrmtx(time,ctmp,npacket)
            else
               call ef_wrmtx(time,dipmtx,npacket)
            endif
         end if

         if (lwrovlp.and..not.ltarop) call ef_wrovlp(time,ovlp1,npacket)
        
! --- End of Main Loop
      goto 100         

 120  continue 
! --- Finalize and exit

! --- close channels
      if (.not. lback .and. lwrpop) close(itarfile)
      if (lwrdipmtx)                close(idipfile)
      if (lwrrawfld)                close(irawfldfile)
      if (lwrovlp.and..not.ltarop)  close(iovlpfile)

      close(ipfifo)
      close(ip1fifo)

! --- Final output of the field 
      call ef_wrfld(fld) 
 
! --- write overlaps file after forward propagation
      if (.not.lback) then
         call zerovxz(ovlp1,maxsta)
         call ef_povlp(psi,psi1,dover,workc,ovlp)
         if (ltarop) then
            call ef_mtrxelement(ovlp1,psi,psi,workc,time, &
              mc(mchpsi),hteil,mr(mrhops),mc(mcdtpsi), &
              mr(mrort),ntarop,mi(mijindx))
!            call ef_mtrxelement(ovlp1,psi,psi,workc,time, &
!              mc(mchpsi),mc(mchteil),mr(mrhops),mc(mcdtpsi), &
!              mr(mrort),ntarop,mi(mijindx))
         end if
         call ef_wrovlpfile(ovlp,ovlp1,npacket)
      end if
      
! --- Tell mctdh that efield has finished (dummy write)
      open(iefifo,file='efield.fifo',status='old')
      write(iefifo,'(i3)') irst
      close(iefifo)

      if (lsimprop) then
         open(ie1fifo,file='efield1.fifo',status='old')
         write(ie1fifo,'(i3)') irst
         close(ie1fifo)
      end if

      return

      end subroutine ef_ovrlp 

      end program efield


