      module efroutines

use decimal, only: long, dop
use op1lib, only: cpvcz, zerovxz, cpvxz
use mmlib, only: mmaxzz
use mtlib, only: mtxxzz
use xvlib, only: vvaxzz

      implicit None

      private
      public :: ef_mtrxelement,ef_sovlp,ef_povlp,ef_mkfield,ef_extrpltefld, &
                ef_handshake,ef_sendfld

      contains


!#######################################################################
!
! ef_routines.F: contains subroutines that perform calculations for 
!                efield and communication with mctdh.
!
! ef_mtrxelement: Calculates mtx(p) = <psi_p|oper|psi1_p> 
! ef_sovlp:       Calculates ovlp   = <psi_s|psi1_s1> of states s,s1
! ef_povlp:       Calculates ovlp(p)= <psi_p|psi1_p> of packets p
! ef_mkfield:     Calculates (part of) the field from mtx and ovlp
! ef_extrpltefld: Extrapolates the field from the last 3 values
! ef_handshake:   Initial communication with MCTDH 
! ef_sendfld:     Send the newly calculated field to mctdh
!#######################################################################

!#######################################################################
!
! MTRXELEMENT
!
! Calculates the matrix-element of oper between two wavefunctions in 
! MCTDH form with identical mode combinations.
!
!  mtx(p) = <psi_p|oper|psi1_p> 
! 
! psi_p and psi1_p can be multi-pack (npacket > 1) wave functions 
! psi and psi1 must to have the same structure, i,e, same mode 
! compinations, same no. of packeges etc. 
!
! Hamiltonian is assumed to consist of correlated terms only 
!
!#######################################################################

      subroutine ef_mtrxelement(mtx,psi,psi1,workc,time, &
                 hpsi,hteil,hops,dtpsi,ort,nop,jindx)
 
use global
use aglobal
use maxv, only: maxdim,maxsta,maxkoe
use dvrdatmod
use griddatmod
use psidef
use datenmod
use fluxmod
use hpsimod
use operdef
use compdatmod
use compdat1mod
use oct

use hphimod, only: hphi,hunphi
use phihphimod, only: phihphi,phihunphi
use calchamod, only: calcha,calchuna

      implicit none

      integer(long) :: s,s1,ktype,p,npackstates,hktype
      integer(long), intent(in) :: nop
      integer(long), dimension(*), intent(in) ::  jindx
      complex(dop), dimension(dgldim1), intent(in) :: psi1
      complex(dop), dimension(dgldim), intent(in) :: psi
      complex(dop), dimension(dgldim), intent(out) :: dtpsi
      complex(dop), dimension(adim+1:dgldim) :: hunpsi
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), dimension(maxsta), intent(out) :: mtx
      complex(dop) :: csum
      complex(dop), dimension(1) :: cdum,cdum1
      complex(dop), dimension(hpsidim), intent(out) :: hpsi
      complex(dop), dimension(pmatdim), intent(out) :: hteil
      real(dop), dimension(hopsdim), intent(in) :: hops
      real(dop), dimension(ortdim), intent(in) :: ort
      real(dop), intent(in) :: time

! --- NOTE: This all works only, if psi and psi1 have the same
! ---       structure.

      call zerovxz(hteil,pmatdim)
      call zerovxz(dtpsi,dgldim)
      cdum = (1.0_dop,0.0_dop)

      ktype=0
      call hphi(psi1,hpsi,ktype,nop)    
      call phihphi(psi,cdum,hpsi,hteil,nop,.false.)
      call calcha(psi1,dtpsi,hteil,hops,psi1(zetf(1,1)),nop, &
           psi(zetf(1,1)),jindx,ort)
      
      hktype=0        
      call hunphi(psi1,hunpsi,hktype,nop)
      call phihunphi(psi,hunpsi,cdum,cdum1,hteil,nop)
      call calchuna(psi1,dtpsi,hteil,nop,jindx)
   
! ->  walk through packtes p and calculate mtx(p)
      npackstates = nstate/npacket  !electronic states per packet
      do p=1,npacket
         mtx(p)=(0.0_dop,0.0_dop)
         do s=1,npackstates
            s1 = (p-1)*npackstates + s
            call vvaxzz(psi(zpsi(s1)),dtpsi(zpsi(s1)),csum,block(s1))
            mtx(p) = mtx(p) + csum
          enddo
      enddo
      
      return
      end subroutine ef_mtrxelement

!#######################################################################
!
! EF_MKOVLP
!
! Similar to OVERLAP0s of overlap.F
! Calculates the overlap between two wavefucntions in MCTDH form
! Special routine which performs the overlap <psi(s=s2)|psi1(s=s1).
!
!  ovlp = <psi_s|psi1_s1>
!
!
! NOTE: this does not destinguish betwenn multi-pack and multi-set
! as multi-packs are just treated a collection of uncoupled electronic
! surfases so with this function it is possible to calculate overlaps 
! between different packets by just choosing s1 and s2 in blocks
! of different wave funktions
!#######################################################################

      subroutine ef_sovlp(ovlp,psi,psi1,dover,workc,s2,s1)

use global
use maxv, only: maxdim,maxsta
use griddatmod
use psidef
use oct

      implicit none

      integer(long), intent(in) :: s1,s2
      integer(long) :: m,zeig1,zeig2,swapzeig
      complex(dop), dimension(dgldim1), intent(in) :: psi1
      complex(dop), dimension(dgldim), intent(in) :: psi
      complex(dop), dimension(doverdim), intent(out) :: dover
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), intent(out) :: ovlp

!-----------------------------------------------------------------------
!     calculate density matrices for each degree of freedom defined as
!     overlap of the single-particle functions rho^m_ij=<phi_i,2|phi_j
!     ,1>
!-----------------------------------------------------------------------
      do m=1,nmode
         call mmaxzz(psi(zetf(m,s2)),psi1(zetf(m,s1)), &
              dover(dovmat(m,1)),subdim(m),dim(m,s2),dim(m,s1))
      enddo


!-----------------------------------------------------------------------
! calculate Sum_L <Phi_J,2|Phi_L,1>A_L,1 within each state
!-----------------------------------------------------------------------
      ovlp=(0d0,0d0)
      zeig1=1
      zeig2=maxblock+1
      call cpvxz(psi1(zpsi(s1)),workc(zeig2),block(s1))

      do m=1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         call mtxxzz(dover(dovmat(m,1)),workc(zeig1),workc(zeig2),vdim(m,s2), &
              dim(m,s1),ndim(m,s1),dim(m,s2))
      enddo

!-----------------------------------------------------------------------
! calculate Sum_J A_J,2^* Sum_L <Phi_J,1|Phi_L,2>A_L,2 within each state
!-----------------------------------------------------------------------
      call vvaxzz(psi(zpsi(s2)),workc(zeig2),ovlp,block(s2))

      return
      end subroutine ef_sovlp

!#######################################################################
!
! EF_POVLP
!
! Calculates the overlap between two wavefucntions in MCTDH form
!
! ovlp(p) = <psi_p|psi1_p>
!
! for packets p (sums over states s in p) 
!####################################################################

      subroutine ef_povlp(psi,psi1,dover,workc,ovlp)

use global
use maxv, only: maxdim,maxsta
use griddatmod
use psidef
use oct

      implicit none

      integer(long) :: s1,s,npackstates,p
      complex(dop), dimension(dgldim), intent(in) :: psi1
      complex(dop), dimension(dgldim), intent(in) :: psi
      complex(dop), dimension(doverdim), intent(out) :: dover
      complex(dop), dimension(workcdim), intent(out) :: workc
      complex(dop), dimension(maxsta), intent(out) :: ovlp
      complex(dop) :: csum

!     Number of states per packet
      npackstates = nstate/npacket 
      
      do p=1,npacket
         ovlp(p)=dcmplx(0.d0,0.d0)
         do s=1,npackstates
            s1 = (p-1)*npackstates + s
            call ef_sovlp(csum,psi,psi1,dover,workc,s1,s1)
            ovlp(p) = ovlp(p) + csum
         enddo
      enddo

      return
      end subroutine ef_povlp

!#######################################################################
!
!     EF_MKFIELD
!
! This subroutine computes (makes) the electric field.
! dipmtx(p) = <psi_p|\mu|psi1_p>
! ovp(p)    = <psi1_p|psi_p>
! val1:       output
! if(ltarop): ovp is ignored 
!
!#######################################################################

      subroutine ef_mkfield(dipmtx,ovp,val1)
    
use global
use maxv, only: maxdim,maxsta
use griddatmod
use psidef
use oct

      implicit none

      real(dop) :: remtx,immtx,reovp,imovp
      real(dop), intent(out) :: val1
      complex(dop), dimension(maxsta), intent(in) :: dipmtx,ovp
      integer(long) :: i
   

      if(ltarop) then              ! genaral target operator  
         !single operator can be used for 
         !multiple initial states. (not extensively tested)

         val1 = 0.d0
         do i=1,npacket
            immtx  = -dimag(dipmtx(i))
!            if(lback) immtx=-immtx
            val1 = val1 + immtx
         enddo
         val1=val1/dble(npacket)

      else if (.not. lalgnphs) then   ! Oper = |Psi_tar><Psi_tar| 
                                      ! no phase align
         val1 = 0.d0
         do i=1,npacket

            remtx  = dble(dipmtx(i))
            immtx  = dimag(dipmtx(i))
            reovp = dble(ovp(i))
            imovp = dimag(ovp(i))

!            if(lback) immtx=-immtx

!            val1 = val1 - (reovp*immtx+imovp*remtx)

            val1=val1 -dimag(ovp(i)*dipmtx(i))

         enddo
         val1=val1/dble(npacket)
!     ---  The second line is often necessary to enforce convergence.  
!       val1 = val1/sqrt(recross*recross+imcross*imcross) (modify this!)

      else                      ! Oper = |Psi_tar><Psi_tar| 
                                ! with phase align (not extensively tested)
         remtx = 0.d0          
         immtx = 0.d0
         reovp =  0.d0
         imovp =  0.d0
         
         do i=1,npacket
            
            remtx  = remtx + dble(dipmtx(i))
            immtx  = immtx + dimag(dipmtx(i))
            reovp  = reovp + dble(ovp(i))
            imovp  = imovp + dimag(ovp(i))
         enddo
            
!         if(lback) immtx=-immtx
         
         val1 = -(reovp*immtx+imovp*remtx)/dble(npacket**2)
                  
!     ---  The second line is often necessary to enforce convergence.  
!          val1 = val1/sqrt(recross*recross+imcross*imcross)
      endif

! --- extrapolation in subroutine ef_extrpltefld

      return 
      end subroutine ef_mkfield


!#######################################################################
!     
!     EF_EXTRPLTEFLD
!
! Extrapolate the field using the last 3 points of the field
! where Y(1)=E(t), Y(2)=E(t-dt), Y(3)=E(t-2dt)
!
! NOTE: The points are assumed to be equidistant and Y(1) being the
!       latest value! 
!
! order has to be 0,1 or 2 for constant, linear and quadratic 
! extrapolation, respectively
!
! Output is: Y(1) = E(t+dt), Y(2)=E(t+2dt)
!
!#######################################################################

      subroutine ef_extrpltefld(Y,order,ierr)

      implicit none
      
      integer(long), intent(in) :: order
      integer(long), intent(out) :: ierr
      real(dop), dimension(3), intent(inout) :: Y
      real(dop) :: a,b,c
      
      ierr = 0
      if (order.eq.2) then
!     quadratic extrapolation: y = a x**2 + b x + c
         c=Y(1) !assume x=0
         a=5.d-1*(Y(3)+c) - Y(2)
         b=a+c-Y(2)
         Y(1)=a+b+c
         Y(2)=4.d0*a + 2.d0*b + c 
!         Y(3)=9.d0*a + 3.d0*b + c 

      elseif (order.eq.1)then
!     linear extrapolation: y = a x + b
         b = Y(1) !assume x=0
         a = Y(1) - Y(2)
         Y(1) = a+b
         Y(2) = 2.d0*a+b
!         Y(3) = 3.d0*a+b
      elseif (order.eq.0)then
!     constant extrapolation y = const
    
         Y(2)=Y(1)
!         Y(3)=Y1(1)
        
!      elseif (order .eq. -1) then
!     trigonometric extrapolation: y = a cos(w x + phi)
         
      else
         ierr=1
      end if

      close(12)

      end subroutine ef_extrpltefld

!#######################################################################
!
!     EF_HANDSHAKE
!
!  Initial handshake with MCTDH, get the number of timesteps from fifo.
!
!#######################################################################

      subroutine ef_handshake()

use global
use channels
use oct
use maxv, only: maxdim,maxsta
use datenmod
use runpropmod

use logdat

      implicit none
      
      open(ipfifo,file='prop.fifo',status='old')
      open(iefifo,file='efield.fifo',status='old')
      write(iefifo,*) -irst     ! dummy write for synchronization
      close(iefifo)
      read(ipfifo,*) nfin
      open(iefifo,file='efield.fifo',status='old')
      write(iefifo,*)  irst     ! dummy write for synchronization
      close(iefifo)
      if(nfin.lt.0) read(ipfifo,*) nfin, out1,out2
      if(out2 .lt. 1.d-7) goto 993
!     do the same with second instance of M!TDH if simprop is true
      if (lsimprop) then
         
         open(ip1fifo,file='prop1.fifo',status='old')
         open(ie1fifo,file='efield1.fifo',status='old')
         write(ie1fifo,*) -irst ! dummy write for synchronization
         close(ie1fifo)
         read(ip1fifo,*) nfin1
         open(ie1fifo,file='efield1.fifo',status='old')
         write(ie1fifo,*)  irst ! dummy write for synchronization
         close(ie1fifo)
         if(nfin1.lt.0) read(ip1fifo,*) nfin1
         if (nfin1.ne.nfin) goto 995
         
      endif

      return
      
 993  continue
      routine='efield'
      message='Time step is zero. This happens sometimes.'//' Try to restart.'
      call errormsg

 995  continue
      routine='efield'
      message='Simult. propagation: inconsistent number of time steps.'
      call errormsg

      end subroutine ef_handshake


!#######################################################################
!
!     EF_SENDFLD
!     
!     Send the field through the fifos to mctdh
!     
!#######################################################################

      subroutine ef_sendfld(fld, oldfld,itime)

use global
use maxv, only: maxdim,maxsta
use channels
use oct
use runpropmod

use logdat

      implicit none

      integer(long) :: i
      integer(long), intent(in) :: itime
      real(dop), dimension(nfin,3), intent(in) :: fld
      real(dop), dimension(3,3), intent(inout) :: oldfld
      real(dop) :: rdum

!     release the field to fifo 
      open(iefifo,file='efield.fifo',status='old', err=994)
      write(iefifo,'(3e24.12)',err=994) fld(itime+3,1), fld(itime+4,2), &
                                        fld(itime+5,3)
      close(iefifo)
      
!     synchronize with second MCTDH (if exists)
!     (the second propagation using the old field)
      if (lsimprop) then 
         do i=1,3
            oldfld(1,i)=oldfld(2,i)
            oldfld(2,i)=oldfld(3,i)
         enddo
         read(iaus,*) rdum, (oldfld(3,i), i=1,3)
         open(ie1fifo,file='efield1.fifo',status='old',err=994)
         if (iiter.eq.0) then
            ! the guess field does not have a prediction
            write(ie1fifo,'(3e24.12)',err=994) oldfld(3,1),oldfld(3,1),oldfld(3,1) 
         else
            write(ie1fifo,'(3e24.12)',err=994) oldfld(3,1),oldfld(2,2),oldfld(1,3) 
         end if
         close(ie1fifo)
      endif

      return

 994  continue
      routine='sendfld (efield.F)'
      message='Unable to send field through fifo.'
      call errormsg

      end subroutine ef_sendfld

      end module efroutines
