module test_readqc
!Module to test the ioqc functions

    use testdrive, only : new_unittest, unittest_type, error_type, check
    !Load readqc to test it
    use readqc
    implicit none
    private

    public :: collect_readqc

    contains

    ! Collect all exported unit tests
    subroutine collect_readqc(testsuite)
        ! Collection of tests in ioqc test suite
        type(unittest_type), allocatable, intent(out) :: testsuite(:)

        !Collect the tests into a suite to export to quantics_test
        testsuite = [ &
            new_unittest("Test Readqc: Allocation", test_rdqc_allocreadqc_valid), &
            ! new_unittest("invalid", test_rdc_alloc_fail,should_fail=.true.) &
            new_unittest("Test Readqc: Molpro", test_rd_qc_molpro_cas_valid)&
            ]

    end subroutine collect_readqc

    subroutine test_rdqc_allocreadqc_valid(error)
        !Will test that readqc alloc correctly when given dimensions
        !Should be a valid test
        type(error_type), allocatable, intent(out) :: error
        
        integer(long)             :: stdim

        !Get the values
        !File is this_schn_1_000.out
        !qc_ variables defined in readqc
        !Thiophene has 9 atoms
        !27 ncoo
        qc_ncoo = 27
        !Calc used 5 states
        qc_nst = 5
        stdim = qc_nst + 1
        !Have 3N-6 modes
        qc_nmode = 21
        qc_nbasis = 0
        !Now can call alloc_readqc
        call alloc_readqc(6)
        !Check arrays are sized correctly
        !And that they are also there
        !qcv is size of nstates
        call check(error,SIZE(qcv),stdim)
        !qcderiv1 is number of states times number of coords
        call check(error, SIZE(qcderiv1), stdim*qc_ncoo)
        !qcderiv2 is stdim*qc_ncoo*qc_ncoo
        call check(error, SIZE(qcderiv2),stdim*qc_ncoo*qc_ncoo)
        !qcmocoeff is qc_nbasis*qc_nbasis
        call check(error, SIZE(qcmocoeff),qc_nbasis*qc_nbasis)
        !qcnact is qc_ncoo*stdim*stdim
        call check(error, SIZE(qcnact),qc_ncoo*stdim*stdim)
        !qcdipole is 3*stdim
        call check(error, SIZE(qcdipole),3*stdim)
        !qctdipole is 4*stdim*stdim
        call check(error, SIZE(qctdipole),4*stdim*stdim)
        

        !etc...


        !Should dealloc too at end
        call dealloc_readqc()

    end subroutine test_rdqc_allocreadqc_valid
    subroutine test_rdc_alloc_fail(error)
        !This should produce an error message and fail
        !As no ncoo given
        type(error_type), allocatable, intent(out) :: error

        !Get the values
        !File is this_schn_1_000.out
        !qc_ variables defined in readqc
        !Thiophene has 9 atoms
        !27 ncoo
        qc_ncoo = 0
        !Calc used 5 states
        qc_nst = 5
        !Have 3N-6 modes
        qc_nmode = 21
        qc_nbasis = 0
        !Now can call alloc_readqc
        call alloc_readqc(6)

        !Should dealloc too at end
        call dealloc_readqc()
    end subroutine test_rdc_alloc_fail

    subroutine test_rd_qc_molpro_cas_valid(error)
        !This should be a valid test
        !Will check correct CAS values returned
        type(error_type), allocatable, intent(out) :: error
        !dummy variables for rdoutqc
        integer(long)                               :: imethod, istate, nroot, ifiletyp, ifunit
        logical(kind=4)                             :: lfound, lfail, ltest

        character(len=255) :: cwd, filename  !Get cwd for path - this will need to be sorted
        
        !Array to hold energies for comparision
        integer                                    :: i
        real(dop), dimension(:)  , allocatable     :: checkCASeng, checkPT2eng, checkMolGeom

        !qc_ variables defined in readqc
        ltest = .true.
        !Thiophene has 9 atoms
        !27 ncoo
        qc_ncoo = 27
        !Calc used 5 states
        qc_nst = 5
        !Have 3N-6 modes
        qc_nmode = 21
        qc_nbasis = 82         !Number of CAS orbitals
        !Alloc check arrays -D0 is mto make them dop
        allocate(checkCASeng(qc_nst))
        checkCASeng = [-551.086542797305D0,-551.077840557251D0,-550.959364052124D0,-550.9497670443431D0,-550.892168065962D0]
        allocate(checkPT2eng(qc_nst))
        checkPT2eng = [-551.605853516309D0,-551.596185797559D0,-551.484239142748D0,-551.490063124284D0,-551.440428971973D0]
        allocate(checkMolGeom(qc_ncoo))
        checkMolGeom = [0.000000000D0,-2.363631642D0,0.051684010D0,&
                        0.000000000D0,-1.355047015D0,2.418792748D0,&
                        0.000000000D0,1.355047015D0,2.418792748D0,&
                        0.000000000D0,2.363631642D0,0.051684010D0,&
                        0.000000000D0,0.000000000D0,-2.292993680D0,&
                        0.000000000D0,-4.317797427D0,-0.472091380D0,&
                        0.000000000D0,-2.480246641D0,4.106242588D0,&
                        0.000000000D0,2.480246641D0,4.106242588D0,&
                        0.000000000D0,4.317797427D0,-0.472091380D0]
        !Now can call alloc_readqc
        call alloc_readqc(6)

        call getcwd(cwd)
        filename = "thio_schn_1_000.out"
        !Open file, this is relative and needs fixing to absolute
        write(*,*) "            Testing MolPro with file: ", trim(filename)
        ifunit = 7
        open(ifunit,file=trim(cwd)//"/testsrc/read_qc/molpro/thio_schn_1_000.out",form='formatted',status='old')
        
        ifiletyp = 5  !Molpro
        imethod = 12    !CAS
        istate = 0
        nroot = 0
        qcread(:) = .false.
        qcread(1) = .true.
        lfound = .true.    ! Test coordinates found
        lfail = .false.

        qcorient = 'input' ! Gaussian input orientation
        
        !This is the function we are testing
        call rdqcout(ifunit,ifiletyp,imethod,6,lfound,lfail)

        
        !Check the geometry
        write(*,*) "            Checking Molpro Geometry"
        do i = 1, qc_ncoo
            call check(error, qcxcoo(i), checkMolGeom(i))
        end do
        !Check the energies that are returned are correct
        !Comparison is with energies in file
        write(*,*) "            Checking Molpro CAS Energies"
        do i = 1, qc_nst
            !Check energies returned agree within error
            call check(error, qcv(i), checkCASeng(i))
            !Check that all energies are flagged as found
            !call check(error, lqcfound3, ltest )  !This doesn't work, a type issue
        end do

        !Can check CAS-PT2
        imethod = 13    !CAS
        call rdqcout(ifunit,ifiletyp,imethod,6,lfound,lfail)
        write(*,*) "            Checking Molpro CAS-PT2 Energies"
        do i = 1, qc_nst
            call check(error, qcv(i), checkPT2eng(i))
        end do

        !Clear arrays
        call dealloc_readqc
        if (allocated(checkCASeng)) deallocate(checkCASeng)
        if (allocated(checkPT2eng)) deallocate(checkPT2eng)
        if (allocated(checkMolGeom)) deallocate(checkMolGeom)
        !Close open files
        close(ifunit)
    end subroutine test_rd_qc_molpro_cas_valid
end module test_readqc