!-----------------------------------------------------------------------
      module vchammod

use decimal, only: long,dop
use maxv
use vcglobal, only: ilog

      implicit none

!-----------------------------------------------------------------------
! parameters for arrays
! maxadd: maximum no. of general terms that can be added
!-----------------------------------------------------------------------
      integer(long), parameter :: maxcoe = 30, maxadd=500, maxswitch=2000

!-----------------------------------------------------------------------
! definition of system and coordinates
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save     :: iatsym
      real(dop), dimension(:), allocatable, save         :: xcoo0,q0,mass, &
                                                            freq0,xcom0, &
                                                            fragfreq0
      real(dop), dimension(:,:), allocatable, save       :: nmcoo,coonm
      character(len=16), dimension(:), allocatable, save :: atsym,nmlab,stalab
      logical(kind=4), save                              :: lortho,lrdq0, &
                                                            lalleig = .false.
      real(dop), save                                    :: e0

!-----------------------------------------------------------------------
! for ignoring modes
!-----------------------------------------------------------------------
      integer(long)  ,dimension(:), allocatable, save :: adof
      logical(kind=4),dimension(:), allocatable, save :: ignoredof
      integer(long), save                             :: i_ignore
      integer(long), dimension(300) :: cc

!-----------------------------------------------------------------------
! internal coords
!-----------------------------------------------------------------------
      integer(long), dimension(:), allocatable, save   :: cootyp
      integer(long), dimension(:,:), allocatable, save :: coodef
      real(dop), dimension(:), allocatable, save       :: intcoo0
      logical(kind=4), save                            :: linternal

!-----------------------------------------------------------------------
! parameters for Hamiltonian
!
! hamord  = 1 linear Hamiltonian (const + kappa + lambda)
!         = 2 adiabatic 2nd order (const + kappa + lambda + gamma)
!         = 3 diabatic 2nd order (const + kappa + lambda + gamma + mu)
! nfit: no. of modes in fit
! mfit(m) = 0 parameters for mode not fitted
!         = 1 parameters for mode fitted
! nsfit: no. of states in fit
! sfit(s) = 0 parameters for state not fitted
!         = 1 parameters for state fitted
! nhigher: no. of parameters with order >=3 in "higher" array
! highord: order included in higher. 3 or 4 allowed
! nadd: number of general terms added
! termdef(m,nadd): power of q for added term 
! termst(2,nadd): states for which added term apply
! forcepos: gives the order of modes/state coefficients that must be 
!           set positive in the hamiltonian (vchfit) 
!-----------------------------------------------------------------------
      integer(long), dimension(:,:), allocatable, save   :: ncoeff,ifdiab
      integer(long), save                                :: nhigher,highord,&
                                                            nadd
      integer(long), dimension(:,:), allocatable, save   :: termdef,stermdef
      integer(long), dimension(:,:), allocatable, save   :: termst,stermst
      integer(long), dimension(:,:,:),allocatable, save  :: forcepos
      logical(kind=4), save                              :: lforcepos,lmodesym
      integer(long), dimension(:,:,:), allocatable, save :: ncoeffcp,ifdiabcp
      integer(long), dimension(:), allocatable, save     :: cp2dtyp
      integer(long), dimension(:,:), allocatable, save   :: dofcp2d
      integer(long), dimension(:,:), allocatable, save   :: stacp2d
      integer(long), dimension(:), allocatable, save     :: ocp2d

      integer(long), save :: nfit,nsfit
      integer(long), dimension(:), allocatable, save :: mfit
      integer(long), dimension(:), allocatable, save :: sfit

!-----------------------------------------------------------------------
! parameters describing datasets
!
!
! asta(n): adiabatic state as defined in Hamiltonian (may be different
!          from actual electronic state number cf eigen(n) in 
!          vctrans.inc)
! asta1(nact,n): coupling information (dcp) included between asta(n) and 
!           asta1(nact,n). Asta1(nact,n) is adiabatic state as defined in 
!           Hamiltonian (may be different from actual electronic state 
!           number cf eigen1(n) in vctrans.inc)
! info(n,1) = 1 dataset(n) includes gradient
!           = 2 dataset(n) includes gradient + hessian
! info(n,2) = 1 dataset(n) includes linear coupling (+ gradient
!               difference)
! info(n,3) = 1 dataset(n) includes transition dipole from ground-state
! states(i) : which states are in database in ordered list
! edsup : upper bound for data set energies to be used in fit
! edslow : lower bound for data set energies to be used in fit
! xbnd(i,1): lower bound for coordinate i
! xbnd(i,2): upper bound for coordinate i
!-----------------------------------------------------------------------
       integer(long), dimension(:), allocatable, save   :: asta
       integer(long), dimension(:,:), allocatable, save :: asta1
       integer(long), dimension(:,:), allocatable, save :: info
       integer(long), dimension(:), allocatable, save   :: states
       real(dop), save                                  :: edsup,edslow
       real(dop), dimension(:,:), allocatable, save     :: qbnd

!------------------------------------------------------------------------
! replacement of normal modes by internal coords
!------------------------------------------------------------------------
        integer(long),save    :: rmode,rmode1,ratm1,ratm2,ratm3,ratm4
        character(len=8),save :: rdef,rdef1

!------------------------------------------------------------------------
! plotting of diabatic surfaces
!------------------------------------------------------------------------
        logical(kind=4), save :: ldiabret

!------------------------------------------------------------------------
! changes lambda*q to lambda*sin(q) when constructing the hamiltonian
!------------------------------------------------------------------------
        logical(kind=4), save :: lsincoup,lpoo

!------------------------------------------------------------------------
! changes lambda*q to lambda*cos(q) when constructing the hamiltonian
!------------------------------------------------------------------------
        logical,dimension(:), allocatable, save :: lcoscoup


!------------------------------------------------------------------------
! 2x2 rotation of normal modes
!------------------------------------------------------------------------
        logical(kind=4), save                           :: lrotmodes
        integer(long),dimension(:,:), allocatable, save :: rotm
        real(dop),dimension(:), allocatable, save       :: rotang
        real(dop),dimension(:), allocatable, save       :: rotmxfq
!       deprecated:
        integer(long), save   :: rotm1,rotm2
        real(dop), save       :: mixang
        real(dop)             :: mixfreq

!------------------------------------------------------------------------
! Symmetry information or use when lautosym=true
!
! irreplab: irrep labels
! chdim: order of the point group
! chel: elements of the character table
!------------------------------------------------------------------------
        character(len=16), dimension(:), allocatable, save :: irreplab
        integer(long), save                                :: chdim
        real(dop), dimension(:,:), allocatable, save       :: chel

!------------------------------------------------------------------------
! Jacobi coordinate information
!------------------------------------------------------------------------
        integer(long), dimension(:),allocatable, save :: jatm
        real(dop), dimension(:), allocatable, save    :: jcoo0

!------------------------------------------------------------------------
! Polyspherical coordinate information
!------------------------------------------------------------------------
        integer(long), dimension(:),allocatable, save :: patm,vcon
        real(dop), dimension(:), allocatable, save    :: pcoo0
        logical(kind=4), save                         :: lradau,ljacobi, &
                                                         lpoly,lvalence

!-----------------------------------------------------------------------
! Array to hold Q0 coordinates
!-----------------------------------------------------------------------
      real(dop), dimension(:), allocatable, save      :: q0val

!-----------------------------------------------------------------------
! Tolerance for defining if a point lies on a cut
!-----------------------------------------------------------------------
      real(dop), save :: cuttol = 0.001_dop

!------------------------------------------------------------------------
! for turning normal modes into curvilinear DOF's
!------------------------------------------------------------------------
!	determines what modes belong to some set of modes under some 
!       curvilinear transformation,
!	the order that the user gives in is important, 
!	eg: for type (1) (polar spherical) the first mode given will 
!       become R, second theta, third phi 
!	the user selects which type of curvilinear transformation the 
!       set nmcurvtype(i) has eg: 
!	 (1) spherical p[olar transformation
!	the user specifies the revference value of the curvilinear coordinates 
!	when it defines them in the input section
!
! dihedflags = flags used to specify a dihedral angle
! qtransform = flags if a transformed coordinate (e.g. cos q) is to be used 
!    in place of q
! nm2curv(m) : mode m is of type
!              1 = dihedral
!------------------------------------------------------------------------
        integer(long), dimension(:,:), allocatable, save :: nmcurv
        integer(long), dimension(:), allocatable, save   :: nmcurvtype, &
                                                            nm2curv
        integer(long), dimension(:), allocatable, save   :: qtransform
        real(dop), dimension(:,:), allocatable, save     :: c0
        logical(kind=4), save                            :: lnmcurv,&
                                                            lqtransform

        character(len=1), dimension(:), allocatable, save :: dihedflag

!-----------------------------------------------------------------------
! Switching functions
!-----------------------------------------------------------------------
! aswitch: character array containing information about each switching
!          function
! nswitch: number of switching functions
! nswitchcoe: number of coefficients for each switching function
!-----------------------------------------------------------------------
        character(len=80), dimension(:), allocatable, save :: aswitch
        integer(long), save                                :: nswitch
        integer(long), dimension(:), allocatable, save     :: nswitchcoe,swmd,&
                                                              swm,swm1,sws, &
                                                              sws1,swk,swpar,&
                                                              swfunc
        real(dop), dimension(:), allocatable, save         :: swini
        logical(kind=4), save                              :: lswitch

      contains

!#######################################################################
!
! VCHAMMOD_ALLOC
!
! memory required for arrays in vchammodmod allocated here.
!
! First check if allocated, then allocate. All arrays must be
! deallocated in
! vchammod_dealloc
!
!#######################################################################

      subroutine alloc_vchammod

      implicit none

      integer(long)     :: ierr
      character(len=10) :: cerr

! check all arrays are not allocated
      call dealloc_vchammod

      cerr = 'iatsym'
      allocate(iatsym(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xcoo0'
      allocate(xcoo0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'q0'
      allocate(q0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mass'
      allocate(mass(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'freq0'
      allocate(freq0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'xcom0'
      allocate(xcom0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'fragfreq0'
      allocate(fragfreq0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nmcoo'
      allocate(nmcoo(maxdim,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'coonm'
      allocate(coonm(maxdim,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'atsym'
      allocate(atsym(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nmlab'
      allocate(nmlab(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stalab'
      allocate(stalab(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'adof'
      allocate(adof(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ignoredof'
      allocate(ignoredof(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'cootyp'
      allocate(cootyp(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'coodef'
      allocate(coodef(4,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'intcoo0'
      allocate(intcoo0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ncoeff'
      allocate(ncoeff(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ifdiab'
      allocate(ifdiab(maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'termdef'
      allocate(termdef(maxdim,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stermdef'
      allocate(stermdef(maxdim,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'termst'
      allocate(termst(2,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stermst'
      allocate(stermst(2,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'forepos'
      allocate(forcepos(maxdim,maxdim,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ncoeffcp'
      allocate(ncoeffcp(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ifdiabcp'
      allocate(ifdiabcp(maxdim,maxsta,maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'cp2dtyp'
      allocate(cp2dtyp(maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dofcp2d'
      allocate(dofcp2d(2,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'stacp2d'
      allocate(stacp2d(2,maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'ocp2d'
      allocate(ocp2d(maxadd),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'asta'
      allocate(asta(maxset),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'asta1'
      allocate(asta1(maxsta,maxset),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'info'
      allocate(info(maxset,3),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'states'
      allocate(states(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qbnd'
      allocate(qbnd(maxdim,2),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'rotm'
      allocate(rotm(INT(maxdim/2)+1,1:2),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'rotang'
      allocate(rotang(INT(maxdim/2)+1),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'rotmxfq'
      allocate(rotmxfq(INT(maxdim/2)+1),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'irreplab'
      allocate(irreplab(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'lcoscoup'
      allocate(lcoscoup(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'chel'
      allocate(chel(maxdim,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'jatm'
      allocate(jatm(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'jcoo0'
      allocate(jcoo0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'patm'
      allocate(patm(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'vcon'
      allocate(vcon(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'pcoo0'
      allocate(pcoo0(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'aswitch'
      allocate(aswitch(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nswitchcoe'
      allocate(nswitchcoe(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swmd'
      allocate(swmd(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swm'
      allocate(swm(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swm1'
      allocate(swm1(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'sws'
      allocate(sws(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'sws1'
      allocate(sws1(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swk'
      allocate(swk(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swpar'
      allocate(swpar(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swfunc'
      allocate(swfunc(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'swini'
      allocate(swini(maxswitch),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nmcurv'
      allocate(nmcurv(INT(maxdim/2)+1,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nm2curv'
      allocate(nm2curv(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'nmcurvtype'
      allocate(nmcurvtype(INT(maxdim/2)+1),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'c0'
      cerr = 'qtransform'
      allocate(qtransform(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      allocate(c0(INT(maxdim/2)+1,maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'dihedflag'
      allocate(dihedflag(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'mfit'
      allocate(mfit(maxdim),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'sfit'
      allocate(sfit(maxsta),stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'q0val'
      allocate(q0val(maxdim),stat=ierr)
      if (ierr > 0) go to 1000

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_VCHAMMOD :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_vchammod

!#######################################################################

      subroutine dealloc_vchammod

      implicit none

      if (allocated(iatsym)) deallocate(iatsym)
      if (allocated(xcoo0)) deallocate(xcoo0)
      if (allocated(q0)) deallocate(q0)
      if (allocated(mass)) deallocate(mass)
      if (allocated(freq0)) deallocate(freq0)
      if (allocated(xcom0)) deallocate(xcom0)
      if (allocated(fragfreq0)) deallocate(fragfreq0)
      if (allocated(nmcoo)) deallocate(nmcoo)
      if (allocated(coonm)) deallocate(coonm)
      if (allocated(atsym)) deallocate(atsym)
      if (allocated(nmlab)) deallocate(nmlab)
      if (allocated(stalab)) deallocate(stalab)
      if (allocated(adof)) deallocate(adof)
      if (allocated(ignoredof)) deallocate(ignoredof)
      if (allocated(cootyp)) deallocate(cootyp)
      if (allocated(coodef)) deallocate(coodef)
      if (allocated(intcoo0)) deallocate(intcoo0)
      if (allocated(ncoeff)) deallocate(ncoeff)
      if (allocated(ifdiab)) deallocate(ifdiab)
      if (allocated(termdef)) deallocate(termdef)
      if (allocated(stermdef)) deallocate(stermdef)
      if (allocated(termst)) deallocate(termst)
      if (allocated(stermst)) deallocate(stermst)
      if (allocated(forcepos)) deallocate(forcepos)
      if (allocated(ncoeffcp)) deallocate(ncoeffcp)
      if (allocated(ifdiabcp)) deallocate(ifdiabcp)
      if (allocated(cp2dtyp)) deallocate(cp2dtyp)
      if (allocated(dofcp2d)) deallocate(dofcp2d)
      if (allocated(stacp2d)) deallocate(stacp2d)
      if (allocated(ocp2d)) deallocate(ocp2d)
      if (allocated(rotm)) deallocate(rotm)
      if (allocated(rotang)) deallocate(rotang)
      if (allocated(rotmxfq)) deallocate(rotmxfq)
      if (allocated(lcoscoup)) deallocate(lcoscoup)
      if (allocated(irreplab)) deallocate(irreplab)
      if (allocated(chel)) deallocate(chel)
      if (allocated(jatm)) deallocate(jatm)
      if (allocated(jcoo0)) deallocate(jcoo0)
      if (allocated(patm)) deallocate(patm)
      if (allocated(vcon)) deallocate(vcon)
      if (allocated(pcoo0)) deallocate(pcoo0)
      if (allocated(aswitch)) deallocate(aswitch)
      if (allocated(nswitchcoe)) deallocate(nswitchcoe)
      if (allocated(swmd)) deallocate(swmd)
      if (allocated(swm)) deallocate(swm)
      if (allocated(swm1)) deallocate(swm1)
      if (allocated(sws)) deallocate(sws)
      if (allocated(sws1)) deallocate(sws1)
      if (allocated(swk)) deallocate(swk)
      if (allocated(swpar)) deallocate(swpar)
      if (allocated(swfunc)) deallocate(swfunc)
      if (allocated(swini)) deallocate(swini)
      if (allocated(nmcurv)) deallocate(nmcurv)
      if (allocated(nm2curv)) deallocate(nm2curv)
      if (allocated(nmcurvtype)) deallocate(nmcurvtype)
      if (allocated(qtransform)) deallocate(qtransform)
      if (allocated(c0)) deallocate(c0)
      if (allocated(dihedflag)) deallocate(dihedflag)
      if (allocated(mfit)) deallocate(mfit)
      if (allocated(sfit)) deallocate(sfit)
      if (allocated(q0val)) deallocate(q0val)

      return
      end subroutine dealloc_vchammod

      end module vchammod
