module vcminmod

  use decimal, only: long,dop

  implicit none

!------------------------------------------------------------------------
! File names
!------------------------------------------------------------------------
  character(len=80), save :: ainp,aout,alog,aham,axyz
  integer(long), save     :: ixyz

!------------------------------------------------------------------------
! Geometry information
!------------------------------------------------------------------------
  real(dop), dimension(:), allocatable, save :: inxcoo,inqcoo
  logical(kind=4), save                      :: llin

!------------------------------------------------------------------------
! Geometry optimisation
!------------------------------------------------------------------------
  integer(long), save                        :: imin,sopt,iter
  real(dop), dimension(:), allocatable, save :: optqcoo,optxcoo

end module vcminmod

!#######################################################################

program vcmin
  
  use decimal,  only: long,dop,sip
  use vcglobal
  use global, only:pi, keyword,keyorig,maxkey,lc,lwarn
  use maxv,     only: maxdim,maxsta,maxset
  use vchammod
  use vcminmod
  use op1lib,   only: zerovxd,zerovxi,cpvxd
  use utils,    only: vchdefault,vccntrl,vcwrline,getfreq,x2q,x2poly,&
                      q2x,poly2x

  use iofiles,  only: readguess,rdfile0
  use symmetry, only: getnadd
  use ioqc,     only: nm2xmat
  use invchammod
  use utils,    only: vcerrormsg
  use logdat, only: message, routine, inptit
  use functions, only: pes


  implicit none

  real(dop), dimension(:), allocatable        :: molstates

  integer(long)                               :: ilbl,ierr,i,iau
  character(len=10)                           :: cerr
  logical(kind=4)                             :: lsocdegen

  complex(dop), dimension(:), allocatable     :: mc
  real(dop), dimension(:), allocatable        :: mr
  integer(long), dimension(:), allocatable    :: mi
  logical(kind=4), dimension(:), allocatable  :: ml
  complex(sip), dimension(:), allocatable     :: ms
  real(sip), dimension(:), allocatable        :: mf

!-----------------------------------------------------------------------
! Initialise calculation 
! get input file name, set maxdim and then allocate memory
!-----------------------------------------------------------------------
      call initvcham(1)

      if (maxdim .lt. 0) then
         routine='VCMIN'
         message='Error reading dimensions from input file'
         call vcerrormsg
      endif
      if (maxsta .lt. 0) then
         routine='VCMIN'
         message='Error reading no. states from input file'
         call vcerrormsg
      endif

      call alloc_vchammod

!-----------------------------------------------------------------------
! Set defaults
!-----------------------------------------------------------------------
  call vchdefault
  call vcmdefault

!-----------------------------------------------------------------------  
! Read input
!-----------------------------------------------------------------------  
  call vcminp

!-----------------------------------------------------------------------
! Determine the no. explicitely added terms and the no. switching
! functions
!-----------------------------------------------------------------------
  call getnadd(iguess)

!-----------------------------------------------------------------------
! Read file0
!-----------------------------------------------------------------------
  iau=1
  lsocdegen=.true.
  call rdfile0(.false.,0,iau,lsocdegen)

!-----------------------------------------------------------------------
! Construct transformation matrices
!-----------------------------------------------------------------------
  iau=1
  call nm2xmat(nmcoo,coonm,iatsym,freq0,iau,ncoo,maxdim,nmodes,ilog)

!-----------------------------------------------------------------------
! Assign memory
!-----------------------------------------------------------------------
  call memvcham
  
  ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
  write(ilog,'(/,a)') 'MEMORY ALLOCATED'
  write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)')&
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
  call myflush(ilog)
  
  if(allocated(mc))deallocate(mc)
  if(allocated(mr))deallocate(mr)
  if(allocated(mi))deallocate(mi)
  if(allocated(ml))deallocate(ml)
  if(allocated(ms))deallocate(ms)
  if(allocated(mf))deallocate(mf)
  if(allocated(molstates))deallocate(molstates)
  
  cerr = 'lmtrans_mc '
  allocate(mc(mcdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'lmtrans_mr '
  allocate(mr(mrdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'lmtrans_mi '
  allocate(mi(midim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'lmtrans_ml '
  allocate(ml(mldim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'lmtrans_ms '
  allocate(ms(msdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'lmtrans_mf '
  allocate(mf(mfdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'inxcoo '
  allocate(inxcoo(maxdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'inqcoo '
  allocate(inqcoo(maxdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'optqcoo '
  allocate(optqcoo(maxdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'optxcoo '
  allocate(optxcoo(maxdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'molstates'
  allocate(molstates(maxset),stat=ierr)
  if (ierr > 0) go to 2000


  call zerovxd(mr,mrdim)
  call zerovxi(mi,midim)

!-----------------------------------------------------------------------
! Read VCHAM file
!
! N.B. re-read frequencies from the VCHAM file as these can differ from
!       those in file0
!-----------------------------------------------------------------------
  call getfreq(iguess)
  call readguess(iguess,mr(mrconst),mr(mrkappa),&
       mr(mrlambda),mr(mrgamma),mr(mriota),mr(mrmu),mr(mrexpcoe),&
       mi(miexpcoemask),mr(mrhigher),mr(mrtermk),mr(mrcpcoe),&
       mr(mrswitch),mr(mrcp2d))

!-----------------------------------------------------------------------
! If polyspherical coordinates are being used, determine pcoo0
!-----------------------------------------------------------------------
  if (lpoly) call x2poly(xcoo0,pcoo0,ncoo,nmodes,patm,mass,vcon)
  
!-----------------------------------------------------------------------
! Determine the point in normal modes/polyspherical coords of the
! starting geometry
!-----------------------------------------------------------------------  
  if (lpoly) then
     call x2poly(inxcoo,inqcoo,ncoo,nmodes,patm,mass,vcon)
     do i=1,nmodes
        inqcoo(i)=inqcoo(i)-pcoo0(i)
     enddo
  else
     call x2q(inqcoo,inxcoo,xcoo0,coonm,nmodes,ncoo,maxdim)
  endif

!-----------------------------------------------------------------------
! Optimise the geometry
!-----------------------------------------------------------------------
  call optgeom(mr)

!-----------------------------------------------------------------------
! Output the optimised geometry
!-----------------------------------------------------------------------
  call outgeom

  deallocate(mc,mr,ml,mi,ms,mf)
  deallocate(inxcoo,inqcoo,optqcoo,optxcoo)

!-----------------------------------------------------------------------
! End
!-----------------------------------------------------------------------
  write(ilog,'(/,a,/)') 'Successful termination of VCMIN'
  STOP

2000 write (6,*) 'Error in Memory Allocation of ',&
          cerr,' Error code is ',ierr
  STOP

contains

!#######################################################################

  subroutine vcmdefault

    implicit none

    integer(long) :: i

! File names/channels
    aham=''
    ainp=''
    file0=''
    
    ixyz=66
    
! System information
    do i=1,maxdim
       inxcoo(i)=0.0
       inqcoo(i)=0.0
       optxcoo(i)=0.0
       optqcoo(i)=0.0
       patm(i)=0
       vcon(i)=0
    enddo

    llin=.false.

    ljacobi=.false.
    lradau=.false.
    lvalence=.false.
    lpoly=.false.

    file0orient='input'

    nsta=0

!-----------------------------------------------------------------------
! imin = 1: simplex
!        2: conjugate gradient
!
! Default optimiser: simplex
!-----------------------------------------------------------------------
    imin=1
    
    sopt=0
    iter=100

    return
  end subroutine vcmdefault

!#######################################################################

  subroutine vcminp

    implicit none
    
    integer(long) :: i,j,k,n,ic,ierr,iz,iz0,ilbl

!-----------------------------------------------------------------------
! Determine input file name and open files
!-----------------------------------------------------------------------
    ilbl=index(ainp,'.')-1
    if (ilbl .le. 0) then
       ilbl=index(ainp,' ')-1
       ainp=ainp(1:ilbl)//'.inp'
    endif

    aout=ainp(1:ilbl)//'.out'
    alog=ainp(1:ilbl)//'.log'
    axyz=ainp(1:ilbl)//'.xyz'

    open(iin,file=ainp,form='formatted',status='old')
    open(iout,file=aout,form='formatted',status='unknown')
    open(ilog,file=alog,form='formatted',status='unknown')
    open(ixyz,file=axyz,form='formatted',status='unknown')

!-----------------------------------------------------------------------
! Read input
!-----------------------------------------------------------------------
10  continue
    
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
    routine = 'vcminp (read VCMIN input)'
    call vccntrl(ierr)
    
    i=1
15  continue
    if (keyword(i) .eq. 'end-input') then
       goto 200

    else if (keyword(i).eq.'geometry_define') then
       n=0
20     continue
       call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
       routine = 'vcminp (read VCMIN input)'
       call vccntrl(ierr)
       if (keyword(1).ne.'end-geometry_define') then
          n=n+1
          read(keyword(2),*) inxcoo(n*3-2)
          read(keyword(3),*) inxcoo(n*3-1)
          read(keyword(4),*) inxcoo(n*3)
          goto 20
       endif
       ncoo=3*n

    else if (keyword(i).eq.'file0') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),'(a)') file0
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the file0 keyword '
          call vcwrline(i,iz,iz0)
       endif

    else if (keyword(i).eq.'vcham_file') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),'(a)') aham
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the vcham_file keyword '
          call vcwrline(i,iz,iz0)
       endif

    else if (keyword(i).eq.'jacobi_define') then
       lpoly=.true.
       ljacobi=.true.
       n=0
25     continue
       call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
       call vccntrl(ierr)
       if (keyword(i).ne.'end-jacobi_define') then
          n=n+1
          read (keyword(1),*) patm(n)
          goto 25
       endif
       
    else if (keyword(i).eq.'radau_define') then
       lpoly=.true.
       lradau=.true.
       n=0
30     continue
       call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
       call vccntrl(ierr)
       if (keyword(i).ne.'end-radau_define') then
          n=n+1
          read (keyword(1),*) patm(n)
          goto 30
       endif
       
    else if (keyword(i).eq.'valence_define') then
       lpoly=.true.
       lvalence=.true.
       n=0
35     continue
       call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
       call vccntrl(ierr)
       if (keyword(i).ne.'end-valence_define') then
          n=n+1
          read (keyword(1),*) patm(n+1)
          read (keyword(2),*) vcon(n)
          if (n.eq.1) read (keyword(2),*) patm(1)
          goto 35
       endif

    else if (keyword(i).eq.'linear') then
       llin=.true.

    else if (keyword(i).eq.'nstates') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) nsta
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the nstates keyword '
          call vcwrline(i,iz,iz0)
       endif

    else if (keyword(i).eq.'orient') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),'(a)') file0orient
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the orient keyword '
          call vcwrline(i,iz,iz0)
       endif
       
    else if (keyword(i).eq.'sopt') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) sopt
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the sopt keyword '
          call vcwrline(i,iz,iz0)
       endif

    else if (keyword(i).eq.'iter') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) iter
       else
          routine = 'vcminp (read VCMIN input)'
          message = 'No argument given with the iter keyword '
          call vcwrline(i,iz,iz0)
       endif

    else
       routine = 'vcminp (read VCMIN input)'
       message = ' unknown keyword : '
       call vcwrline(i,iz,iz0)
    endif

!-----------------------------------------------------------------------
! check if more keywords are to be read, else read new line
!-----------------------------------------------------------------------
    if (i.lt.ic) then
       i=i+1
       go to 15
    else
       go to 10
    endif

200 continue

    if (llin) then
       nmodes=ncoo-5
    else
       nmodes=ncoo-6
    endif

!------------------------------------------------------------------
! Check that all the required information has been given
!------------------------------------------------------------------
    if (aham.eq.'') then
       write(6,'(/,a,/)') 'VCHAM file has not been specified!'
       STOP
    endif

    if (file0.eq.'') then
       write(6,'(/,a,/)') 'file0 has not been specified!'
       STOP
    endif

    if (nsta.eq.0) then
       write(6,'(/,a,/)') 'Number of states has not been specified!'
       STOP
    endif

    if (ncoo.eq.0) then
       write(6,'(/,a,/)') 'Starting geometry has not been specified!'
       STOP
    endif

    if (sopt.eq.0) then
       write(6,'(/,2a,/)') 'State whose potential energy is to be ',&
            'minimised has not been specified!'
       STOP
    endif

!------------------------------------------------------------------
! Open VCHAM file
!------------------------------------------------------------------
    open(iguess,file=aham,form='formatted',status='old')

    return
  end subroutine vcminp

!#######################################################################

  subroutine optgeom(mr)

    implicit none
    
    integer(long)                   :: i,j

    real(dop), dimension(1000*1000) :: xi
    real(dop), dimension(1000,3)    :: work

    real(dop), dimension(mrdim)     :: mr
    real(dop)                       :: fpes


    write(6,'(/,a,2x,i2)') &
         'Minimisation of the adiabatic potential for state',sopt
    write(ilog,'(/,a,2x,i2)') &
         'Minimisation of the adiabatic potential for state',sopt

    fpes=getpes(mr,sopt,inqcoo)
    write(6,'(/,a,2x,F9.5)') 'Initial potential value:',fpes
    write(ilog,'(/,a,2x,F9.5)') 'Initial potential value:',fpes

!------------------------------------------------------------------
! Initialise arrays
!------------------------------------------------------------------
    do i=1,nmodes
       optqcoo(i)=inqcoo(i)
    enddo

!------------------------------------------------------------------
! Minimise on the specified adiabatic potential surface
!------------------------------------------------------------------
    if (imin.eq.1) then      
       write(6,'(/,a)') 'Optimiser used: Simplex'
       write(ilog,'(/,a)') 'Optimiser used: Simplex'
       call initialise(xi,work(1,1),optqcoo,nmodes,mr)
       call simplex(xi,work(1,1),nmodes,0.001_dop,iter,work(1,2),&
            work(1,3),mr)
    else
       write(6,'(/,a,/)') "Unknown optimiser"
       write(ilog,'(/,a,/)') "Unknown optimiser"
       STOP
    endif

!------------------------------------------------------------------
! Save optimised coordinates
!------------------------------------------------------------------
    fpes=getpes(mr,sopt,xi)
    write(6,'(/,a,2x,F9.5)') 'Final potential value:',fpes
    write(ilog,'(/,a,2x,F9.5)') 'Final potential value:',fpes

    if (lpoly) then 
       do i=1,nmodes
          optqcoo(i)=xi(i)+pcoo0(i)
       enddo
       call poly2x(optxcoo,xcoo0,optqcoo,ncoo,nmodes,patm,mass,vcon)
    else
       do i=1,nmodes          
          optqcoo(i)=xi(i)
       enddo
       call q2x(optxcoo,xcoo0,optqcoo,nmcoo,nmodes,ncoo,maxdim)
    endif

    return
  end subroutine optgeom

!#######################################################################

  real(dop) function getpes(mr,s,q)
    
    implicit none

    integer(long)                       :: s
    real(dop), dimension(mrdim)         :: mr
    real(dop), dimension(nmodes)        :: q
    real(dop), dimension(maxsta)        :: v
    logical(kind=4)                     :: leigvec
    real(dop), dimension(maxsta,maxsta) :: diabvec
    real(dop), dimension(maxsta,maxsta) :: diabv

    leigvec=.false.

    call pes(q,v,mr(mrconst),mr(mrkappa),mr(mrlambda),mr(mrgamma),&
         mr(mriota),mr(mrmu),mr(mrexpcoe),mr(mrhigher),mr(mrtermk),&
         mr(mrcpcoe),mr(mrswitch),mr(mrcp2d),leigvec,diabvec,diabv)
    
    getpes=v(s)

  end function getpes

!#######################################################################

  SUBROUTINE initialise(p,y,x,ndim,rdata)

      implicit none

      integer i,j,ndim
      real*8 rdata(*)
      real*8 p(ndim,ndim+1),y(ndim+1),x(ndim),h

      write(6,'(/,a)') "Initialising amoeba..."

      do i=1,ndim
         p(i,1)=x(i)
      enddo
      y(1)=getpes(rdata,sopt,p(1,1))
      do j=2,ndim+1         
         do i=1,ndim
            p(i,j)=x(i)
         enddo
         h=0.5
   10    continue
         p(j-1,j)=x(j-1)+h
         y(j)=getpes(rdata,sopt,p(1,j))
         if (y(j) .gt. y(1)*1000.0) then
            h=h*0.9
            go to 10
         endif
      enddo
      
      return
    end SUBROUTINE initialise

!#######################################################################
!
! p(ndim+1,ndim) = input points
! y(ndim+1) = input values
! mp=ndim+1
! psum(ndim),ptry(ndim) = work arrays
!
    SUBROUTINE simplex(p,y,ndim,ftol,iter,psum,ptry,rdata)

      INTEGER iter,ndim,ITMAX
      REAL*8 ftol,p(ndim,ndim+1),y(ndim+1),rdata(*)
      PARAMETER (ITMAX=100000)
      INTEGER i,ihi,ilo,inhi,j,m,n
      REAL*8 rtol,sum,swap,ysave,ytry,psum(ndim),ptry(ndim)

      write(6,'(/,a)') "Fitting Parameters..."

      iter=0
1     do 12 n=1,ndim
         sum=0.
         do 11 m=1,ndim+1
            sum=sum+p(n,m)
11          continue
            psum(n)=sum
12          continue
2           ilo=1
            if (y(1).gt.y(2)) then
               ihi=1
               inhi=2
            else
               ihi=2
               inhi=1
            endif
            do 13 i=1,ndim+1
        if(y(i).le.y(ilo)) ilo=i
        if(y(i).gt.y(ihi)) then
          inhi=ihi
          ihi=i
        else if(y(i).gt.y(inhi)) then
          if(i.ne.ihi) inhi=i
        endif
13    continue
      rtol=2.*abs(y(ihi)-y(ilo))/(abs(y(ihi))+abs(y(ilo)))
      if (rtol.lt.ftol) then
        swap=y(1)
        y(1)=y(ilo)
        y(ilo)=swap
        do 14 n=1,ndim
          swap=p(n,1)
          p(n,1)=p(n,ilo)
          p(n,ilo)=swap
14      continue
        return
      endif
      if (iter.ge.ITMAX) stop 'ITMAX exceeded in amoeba'
      iter=iter+2
      ytry=amotryi(p,y,psum,ndim,ihi,-1.0d0,ptry,rdata)
      if (ytry.le.y(ilo)) then
        ytry=amotryi(p,y,psum,ndim,ihi,2.0d0,ptry,rdata)
      else if (ytry.ge.y(inhi)) then
        ysave=y(ihi)
        ytry=amotryi(p,y,psum,ndim,ihi,0.5d0,ptry,rdata)
        if (ytry.ge.ysave) then
          do 16 i=1,ndim+1
            if(i.ne.ilo)then
              do 15 j=1,ndim
                psum(j)=0.5*(p(j,i)+p(j,ilo))
                p(j,i)=psum(j)
15            continue
              y(i)=getpes(rdata,sopt,psum)
           endif
16        continue
          iter=iter+ndim
          goto 1
        endif
      else
        iter=iter-1
      endif
      goto 2

   END SUBROUTINE simplex

!#######################################################################

   real(dop) FUNCTION amotryi(p,y,psum,ndim,ihi,fac,ptry,rdata)

     INTEGER ihi,ndim
     REAL*8 fac,p(ndim,ndim+1),psum(ndim),y(ndim+1),rdata(*)
     INTEGER j
     REAL*8 fac1,fac2,ytry,ptry(ndim)
     fac1=(1.-fac)/ndim
     fac2=fac1-fac
     do 11 j=1,ndim
        ptry(j)=psum(j)*fac1-p(j,ihi)*fac2
11      continue
        ytry=getpes(rdata,sopt,ptry)
        if (ytry.lt.y(ihi)) then
           y(ihi)=ytry
           do 12 j=1,ndim
              psum(j)=psum(j)-p(j,ihi)+ptry(j)
              p(j,ihi)=ptry(j)
12            continue
           endif
           amotryi=ytry
           return

         END FUNCTION amotryi

!#######################################################################

  subroutine outgeom

    implicit none

    integer(long) :: i,j
    
!-----------------------------------------------------------------------
! Output starting coordinates
!-----------------------------------------------------------------------
    write(iout,'(/,a,/)') &
         '*** Starting Geometry: Cartesian Coordinates ***'
    do i=1,ncoo,3
       write(iout,'(a2,3x,3(F10.7,3x))') atsym(i),(inxcoo(j), j=i,i+2)
    enddo

    if (lpoly) then
       write(iout,'(/,a,/)') &
            '*** Starting Geometry: Polyspherical Coordinates ***'
       do i=1,nmodes
          if (i.lt.10) then
             write(iout,'(a1,i1,5x,F10.7)') 'q',i,inqcoo(i)+pcoo0(i)
          else
             write(iout,'(a1,i2,4x,F10.7)') 'q',i,inqcoo(i)+pcoo0(i)
          endif
       enddo
    else
       write(iout,'(/,a,/)') &
            '*** Starting Geometry: Normal Mode Coordinates ***'
       do i=1,nmodes
          if (i.lt.10) then
             write(iout,'(a1,i1,5x,F10.7)') 'q',i,inqcoo(i)
          else
             write(iout,'(a1,i2,4x,F10.7)') 'q',i,inqcoo(i)
          endif
       enddo       
    endif

!-----------------------------------------------------------------------
! Output optimised coordinates
!-----------------------------------------------------------------------
    write(iout,'(/,a,/)') &
         '*** Optimised Geometry: Cartesian Coordinates ***'
    do i=1,ncoo,3
       write(iout,'(a2,3x,3(F10.7,3x))') atsym(i),(optxcoo(j), j=i,i+2)
    enddo

    if (lpoly) then
       write(iout,'(/,a,/)') &
            '*** Optimised Geometry: Polyspherical Coordinates ***'
       do i=1,nmodes
          if (i.lt.10) then
             write(iout,'(a1,i1,5x,F10.7)') 'q',i,optqcoo(i)
          else
             write(iout,'(a1,i2,4x,F10.7)') 'q',i,optqcoo(i)
          endif
       enddo
    else
       write(iout,'(/,a,/)') &
            '*** Optimised Geometry: Normal Mode Coordinates ***'
       do i=1,nmodes
          if (i.lt.10) then
             write(iout,'(a1,i1,5x,F10.7)') 'q',i,optqcoo(i)
          else
             write(iout,'(a1,i2,4x,F10.7)') 'q',i,optqcoo(i)
          endif
       enddo
    endif

!-----------------------------------------------------------------------
! Write xyz file
!-----------------------------------------------------------------------
    write(ixyz,'(i2,/)') ncoo/3
    do i=1,ncoo,3
       write(ixyz,'(a2,3x,3(F10.7,3x))') atsym(i),(optxcoo(j), j=i,i+2)
    enddo

    return
  end subroutine outgeom

!#######################################################################

end program vcmin
