!#####################################################################
!
! VCSRF: A program to calculate the values of model VCHAM diabatic
!        potential functions and projectors at the grid points to be
!        used in a subsequent wavepacket propagation.
!
!        Quantities that may be calculated:
!
!        - Elements of the full diabatic potential, with or without
!          block diagonalisation.
!
!        - Cut-HDMR component functions of the elements of the
!          diabatic potential, using either uncombined or combined
!          clusters, and with or without block diagonalisation.
!
!        - Elements of a projector onto an adiabatic state.
!
! SN 02/2013 
!
!#####################################################################

program vcsrf

  use decimal, only: long,dop,sip,vlong
  use vcglobal
  use global, only:pi, keyword,keyorig,maxkey,lc,lwarn
  use maxv, only: maxdim,maxsta
  use keyunitsmod
  use vchammod
  use vcsrfmod
  use versions
  use op1lib, only: zerovxd,zerovxi,cpvxd,zeromxi
  use utils, only: vcwarnmsg,vcerrormsg,vchdefault,getfreq,&
                   vccntrl,vcwrline
  use op1lib, only: zeromxd

  use iofiles, only: readguess
  use symmetry, only: rdexplicit,getnadd
  use invchammod
  use logdat, only: message, routine, inptit
  use functions, only: pes
  
  implicit none

  integer(long)                               :: i,ilbl,ierr
  character(len=10)                           :: cerr

  complex(dop), dimension(:), allocatable     :: mc
  real(dop), dimension(:), allocatable        :: mr
  integer(long), dimension(:), allocatable    :: mi
  logical(kind=4), dimension(:), allocatable  :: ml
  complex(sip), dimension(:), allocatable     :: ms
  real(sip), dimension(:), allocatable        :: mf

  integer(long), dimension(:,:), allocatable  :: ucmap
  integer(long), dimension(:), allocatable    :: ucpnt

!-----------------------------------------------------------------------
! Initialise calculation 
! get input file name, set maxdim and then allocate memory
!-----------------------------------------------------------------------
      call initvcham(1)

      if (maxdim .lt. 0) then
         routine='VCTRANS'
         message='Error reading dimensions from input file'
         call vcerrormsg
      endif
      if (maxsta .lt. 0) then
         routine='VCTRANS'
         message='Error reading no. states from input file'
         call vcerrormsg
      endif

      call alloc_vcsrfmod
      call alloc_vchammod

!-----------------------------------------------------------------------
! Set defaults
!-----------------------------------------------------------------------
  call vchdefault
  call vcsdefault

!-----------------------------------------------------------------------
! Read input
!-----------------------------------------------------------------------
  call vcsinp

!-----------------------------------------------------------------------
! Get grid points
!-----------------------------------------------------------------------
  call getgrid

!-----------------------------------------------------------------------
! Determine the no. explicitely added terms and the no. switching
! functions
!-----------------------------------------------------------------------
  call getnadd(iguess)

!-----------------------------------------------------------------------
! Initialise the Uncombined Cluster (UC) mapping array ucmap, the
! UC point no. array ucpnt, and the UC coordinate array uccoo
!-----------------------------------------------------------------------
  if(allocated(ucmap)) deallocate(ucmap)
  cerr = 'vcsrf_ucmap'
  allocate(ucmap(2**nmodes,nmodes),stat=ierr)
  if (ierr > 0) go to 2000
  call zeromxi(ucmap,2**nmodes,nmodes)

  if(allocated(ucpnt)) deallocate(ucpnt)
  cerr = 'vcsrf_ucpnt'
  allocate(ucpnt(2**nmodes),stat=ierr)
  if (ierr > 0) go to 2000
  call zerovxi(ucpnt,midim)

!-----------------------------------------------------------------------
! Determine cluster information
!-----------------------------------------------------------------------
  call cluster_info(ucmap,ucpnt)

!-----------------------------------------------------------------------
! Assign memory
!-----------------------------------------------------------------------
! VCHAM parameters
  call memvcham

! Values of the UCs and the CC at their grid points
  do i=1,nuc
     mruc(i)=mrdim
     mrdim=mrdim+ucpnt(i)
  enddo
  mrcc=mrdim
  mrdim=mrdim+ucpnt(nuc)

! Coordinates of the UC grid points
  do i=1,nuc
     mruccoo(i)=mrdim
     mrdim=mrdim+ucpnt(i)*nmodes
  enddo

  ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
  write(ilog,'(/,a)') 'MEMORY ALLOCATED'
  write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)')&
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
  call myflush(ilog)
  
  if(allocated(mc)) deallocate(mc)
  if(allocated(mr)) deallocate(mr)
  if(allocated(mi)) deallocate(mi)
  if(allocated(ml)) deallocate(ml)
  if(allocated(ms)) deallocate(ms)
  if(allocated(mf)) deallocate(mf)
  
  cerr = 'vcsrf_mc '
  allocate(mc(mcdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'vcsrf_mr '
  allocate(mr(mrdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'vcsrf_mi '
  allocate(mi(midim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'vcsrf_ml '
  allocate(ml(mldim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'vcsrf_ms '
  allocate(ms(msdim),stat=ierr)
  if (ierr > 0) go to 2000
  cerr = 'vcsrf_mf '
  allocate(mf(mfdim),stat=ierr)
  if (ierr > 0) go to 2000

  call zerovxd(mr,mrdim)
  call zerovxi(mi,midim)

!-----------------------------------------------------------------------
! Read VCHAM file
!-----------------------------------------------------------------------
  call getfreq(iguess)
  call readguess(iguess,mr(mrconst),mr(mrkappa),&
       mr(mrlambda),mr(mrgamma),mr(mriota),mr(mrmu),mr(mrexpcoe),&
       mi(miexpcoemask),mr(mrhigher),mr(mrtermk),mr(mrcpcoe),&
       mr(mrswitch),mr(mrcp2d))

!-----------------------------------------------------------------------
! Calculate the combined cluster
!-----------------------------------------------------------------------
  call combined_cluster(mr)

!-----------------------------------------------------------------------
! End... here
!-----------------------------------------------------------------------
  if (lblock) then
     write(6,'(/,5x,a,x,i6,/)') 'Number of regularised inversions:',nreg
     write(ilog,'(/,5x,a,x,i6,/)') 'Number of regularised inversions:',nreg
  endif
  write(ilog,'(/,5x,a,/)') 'Normal termination of VCSRF'
  write(6,'(/,5x,a,/)') 'Normal termination of VCSRF'

  STOP

2000 write (6,*) &
          'Error in Memory Allocation of ',cerr,' Error code is ',ierr
  STOP

contains

!#####################################################################
  
  subroutine vcsdefault

    implicit none

    integer(long) :: i,j,k

    do i=1,maxdim
       gcut(i)=0
       dcut(i)=0
       npnt(i)=0
       aidof(i)=''
       do j=1,maxpnt
          gpnt(i,j)=0.0d0
       enddo
       do j=1,maxdim
          particle(j,i)=0
          dcut2(j,i)=0
       enddo
       mruc(i)=0
       blockdim(i)=0
    enddo

    scut=0
    scut1=0

    vchamfile=''

    dvrfile=''
    
    ltrans=.false.
    itrans=0
    ltest=.false.
    lhdmr=.false.

    npart=0

    nuc=0

! Block diagonalisation
    lblock=.false.
    nblock=0
    lamtol=1.0e-6_dop
    lreg=.true.
    ireg=2
    nreg=0
    eps=1.0e-16_dop
    do i=1,maxdim
       blockdim(i)=0
    enddo

! Reresentation of the electronic Hamiltonian: by default we output the
! potential in the diabatic representation
    ladiab=.false.

! By default, output binary. Ascii or x,y,z formats may be set by input
    lbin=.true.
    lascii=.false.
    lplot=.false.

! Energy units (a.u. by default):
    ! 1: a.u.
    ! 2: eV
    ! 3: cm-1
    entype=1

! Adiabatic state projection matrix
    ladproj=.false.
    iadproj=0

    return
  end subroutine vcsdefault

!#####################################################################

  subroutine vcsinp

    implicit none

    integer(long)     :: i,iz,iz0,ic,ierr,ilbl,m,n
    character(len=80) :: filename,string

!-----------------------------------------------------------------------
! Get input file name and file stem
!-----------------------------------------------------------------------
    ilbl=index(filein,'.')-1
    if (ilbl .le. 0) then
       ilbl=index(filein,' ')-1
       filein=filein(1:ilbl)//'.inp'
    endif
    fileout=filein(1:ilbl)//'.dat'
    filelog  = filein(1:ilbl)//'.log'

    filename=filein
    open(iin,file=filename,form='formatted',status='old',iostat=ierr)

!-----------------------------------------------------------------------
! Read input file
!-----------------------------------------------------------------------
10  continue
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
    routine = 'vcsinp (read VCSRF input)'
    call vccntrl(ierr)

    i=1
15  continue
    if (keyword(i) .eq. 'end-input') then
       goto 20

    else if (keyword(i).eq.'vcham_file') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          vchamfile=keyword(i)
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with orient keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'dvr_file') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          dvrfile=keyword(i)
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with orient keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'nmodes') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) nmodes
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with nmodes keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'nstates') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) nsta
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with nstates keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'cut') then
       if (keyword(i+1).eq.'=') then
          i=i+2
30        continue
          if (keyword(i).eq.'{') then
             npart=npart+1
             i=i+1
33           continue
             read(keyword(i),*) n
             gcut(n)=1
             particle(npart,n)=1
             if (keyword(i+1).eq.',') then
                i=i+2
                goto 33
             else if (keyword(i+1).eq.'}'.and.keyword(i+2).eq.',') then
                i=i+3
                goto 30
             else
                i=i+1
             endif
          else
             read(keyword(i),*) n
             gcut(n)=1
             npart=npart+1
             particle(npart,n)=1
             if (keyword(i+1).eq.',') then
                i=i+2
                goto 30
             endif             
          endif
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with cut keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'dcut') then
       if (keyword(i+1).eq.'=') then
          i=i+2
31        continue
          read(keyword(i),*) n
          dcut(n)=1
          if (keyword(i+1).eq.',') then
             i=i+2
             goto 31
          endif
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with dcut keyword'
          call vcwrline(0,iz,iz0)
       endif

    else if (keyword(i).eq.'scut') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) scut
          if (keyword(i+1).eq.',') then
             i=i+2
             read(keyword(i),*) scut1
          endif
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with cut keyword'
          call vcwrline(0,iz,iz0)
       endif
       
    else if (keyword(i).eq.'coord_trans') then
       if (keyword(i+1).eq.'=') then
          ltrans=.true.
          i=i+2
          if (keyword(i).eq.'costh1_costh2') then
             itrans=1
          else if (keyword(i).eq.'costh1') then
             itrans=2
          else if (keyword(i).eq.'costh2') then
             itrans=3
          else
             write(6,'(/,x,a,x,a/)') &
                  'Unknown coord. transformation type:',&
                  keyword(i)(1:len_trim(keyword(i)))
             STOP
          endif
       else
          routine = 'vcsinp (read VCSRF input)'
          message = 'No argument specified with coord_trans keyword'
          call vcwrline(0,iz,iz0)
       endif
          
    else if (keyword(i).eq.'test') then
       ltest=.true.

    else if (keyword(i).eq.'hdmr') then
       lhdmr=.true.

    else if (keyword(i).eq.'out') then
        if (keyword(i+1).eq.'=') then
           i=i+2
           if (keyword(i).eq.'ascii') then
              lascii=.true.
              lbin=.false.
           else if (keyword(i).eq.'plot') then
              lplot=.true.
              lbin=.false.
           else
              write(6,'(/,3(1x,a),/)') 'Output format',&
                   keyword(i)(1:len_trim(keyword(i))),'is unrecognised.'
              STOP
           endif
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the out keyword'
           call vcwrline(0,iz,iz0)
        endif

     else if (keyword(i).eq.'units') then
        if (keyword(i+1).eq.'=') then
           i=i+2
           if (keyword(i).eq.'au') then
              entype=1
           else if (keyword(i).eq.'ev'.or.keyword(i).eq.'eV') then
              entype=2
           elseif (keyword(i).eq.'cm-1') then
              entype=3
           else
              write(6,'(/,2(1x,a),/)') 'Unknown energy unit',&
                   keyword(i)(1:len_trim(keyword(i)))
              STOP
           endif
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the units keyword'
           call vcwrline(0,iz,iz0)
        endif

     else if (keyword(i).eq.'block_diag') then
        lblock=.true.
        if (keyword(i+1).eq.'=') then
           i=i+2
32         continue
           nblock=nblock+1
           read(keyword(i),*) n
           blockdim(nblock)=n
           if (keyword(i+1).eq.',') then
              i=i+2
              goto 32
           endif
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the block_diag keyword'
           call vcwrline(0,iz,iz0)
        endif

     else if (keyword(i).eq.'adiabatic') then
        ladiab=.true.

     else if (keyword(i).eq.'adproj') then
        ladproj=.true.
        if (keyword(i+1).eq.'=') then
           i=i+2
           read(keyword(i),*) iadproj
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the adproj keyword'
           call vcwrline(0,iz,iz0)
        endif

     else if (keyword(i).eq.'lambda_tol') then
        if (keyword(i+1).eq.'=') then
           i=i+2
           read(keyword(i),*) lamtol
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the lambda_tol keyword'
           call vcwrline(0,iz,iz0)
        endif
        
     else if (keyword(i).eq.'no_reg') then
        lreg=.false.

     else if (keyword(i).eq.'reg_type') then
        if (keyword(i+1).eq.'=') then
           i=i+2
           read(keyword(i),*) string
           if (string.eq.'constant') then
              ireg=1
           else if (string.eq.'exp') then
              ireg=2
           else
              routine = 'vcsinp (read VCSRF input)'
              message = 'Unknown argument given with the reg_type keyword'
           call vcwrline(0,iz,iz0)
           endif
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the reg_type keyword'
           call vcwrline(0,iz,iz0)
        endif

     else if (keyword(i).eq.'eps') then
        if (keyword(i+1).eq.'=') then
           i=i+2
           read(keyword(i),*) eps
        else
           routine = 'vcsinp (read VCSRF input)'
           message = 'No argument specified with the eps keyword'
           call vcwrline(0,iz,iz0)
        endif

    else
       routine = 'vcsinp (read VCSRF input)'
       message = ' unknown keyword : '
       call vcwrline(i,iz,iz0)
    endif
    
!-----------------------------------------------------------------------
! Check if more keywords are to be read, else read new line
!-----------------------------------------------------------------------
    if (i.lt.ic) then
       i=i+1
       go to 15
    else
       go to 10
    endif

20  continue
    close(iin)

!-----------------------------------------------------------------------
! Check that all required information has been specified
!-----------------------------------------------------------------------
    if (vchamfile.eq.'') then
       write(6,'(/,a,/)') 'The VCHAM file has not been specified!'
       STOP
    endif

    if (dvrfile.eq.'') then
       ! Assume the path to the dvr file is ./dvr
       dvrfile='dvr'
    endif

    if (scut.eq.0.and.scut1.eq.0) then
       write(6,'(/,a,/)') 'Enter state 1:'
       read(5,*) scut       
       write(6,'(/,a,/)') 'Enter state 2:' 
       write(6,'(3a,/)')'(enter 0 if you are making a cut through a',&
            'surface corresponding to an ON-DIAGONAL element of the',&
            'diabatic potential)'
       read(5,*) scut1
       if (scut.eq.0.and.scut1.eq.0) then
          write(6,'(/,2a,/)') 'The element of the potential matrix of',&
               'interest has not been specified!'
          STOP
       endif
    endif
    if (scut.eq.0) scut=scut1
    if (scut1.eq.0) scut1=scut
    
    if (nmodes.eq.0) then
       write(6,'(/,a,/)') 'Enter number of modes:'
       read(5,*) nmodes
       if (nmodes.eq.0) then
          write(6,'(/,a,/)') 'The number of modes has not been specified!'
          STOP
       endif
    endif

    if (lblock) then
       if (nblock.eq.0) then
          write(6,'(/,a,/)') 'Block structure has not been specified!'
          STOP
       endif
       n=0
       do i=1,nblock
          n=n+blockdim(i)
       enddo
       if (n.lt.nsta) then
          write(6,'(/,2a,/)') &
               'Sum of block dimensions is LESS than the dimension of ',&
               'the potential matrix!'
          STOP
       endif
       if (n.gt.nsta) then
          write(6,'(/,2a,/)') &
               'Sum of block dimensions is GREATER than the dimension ',&
               'the potential matrix!'
          STOP
      endif
   endif

!------------------------------------------------------------------
! Open VCHAM file
!------------------------------------------------------------------
    open(iguess,file=vchamfile,form='formatted',status='old')

!------------------------------------------------------------------
! Open output file
!------------------------------------------------------------------
    if (lbin) then
       open(iout,file=fileout,form='unformatted',status='unknown')
    else
       open(iout,file=fileout,form='formatted',status='unknown')
    endif

!------------------------------------------------------------------
! Set energy unit conversion factor
!
! N.B. if we are calculating a projector, then set conv to 1. This
!      is necessary as the uc array is multiplied by conv even 
!      though it here contains a projector element that is
!      dimensionless
!------------------------------------------------------------------
    if (ladproj) entype=2

    if (entype.eq.1) then
       ! a.u.
       conv=1.0d0/uev
    else if (entype.eq.2) then
       ! eV
       conv=1.0d0
    else if (entype.eq.3) then
       ! cm-1
       conv=1.0d0/invcm2ev
    endif

!------------------------------------------------------------------
! If we are dealing with a Cut-HDMR component function, copy the
! gcut array to the dcut array
!------------------------------------------------------------------
    if (lhdmr) then
       do n=1,nmodes
          dcut(n)=gcut(n)
       enddo
    endif

!------------------------------------------------------------------
! If an adiabatic potential is requested, ensure that an 
! on-diagonal element has been requested
!------------------------------------------------------------------
    if (ladiab) then
       if (scut.ne.scut1) then
          write(6,'(/,2a,/)')'Adiabatic potential requested, but',&
               'state1 and state2 are not equal!'
          STOP
       endif
    endif

!------------------------------------------------------------------
! If block-diagonalisation and an adiabatic potential have been
! requested, then exit with a useful message
!------------------------------------------------------------------
    if (ladiab.and.lblock) then
       write(6,'(/,4a,/)') 'Block-diagonalisation and then ',& 
            'diagonalisation of the resulting potential is ',&
            'redundant. If you really want an adiabatic surface, ',&
            'remove the block_diag keyword.'
       STOP
    endif

    return
  end subroutine vcsinp

!#####################################################################
! getgrid: Rather crudely, calls the rdgrid program for each mode and 
!          reads the grid points from the resulting output
!#####################################################################

  subroutine getgrid

    implicit none

    integer(long)     :: ierr,unit,ver,i,j,k,l

    integer(long)     :: iz,iz0,ic

    real(dop)         :: ftmp
    character(len=20) :: prog,fmat
    character(len=80) :: string,atmp,adat
    
    unit=30
    
    adat=''
    k=index(filein,'.')-1
    write(adat(1:k),'(a)') filein(1:k)    
    write(adat(k+1:k+4),'(a)') '.tmp'

!-----------------------------------------------------------------------
! Determine version number
!-----------------------------------------------------------------------
    ver=progver*10
    prog=''
    write(prog(1:6),'(a6)') 'rdgrid'
    write(prog(7:9),'(i3)') ver

!-----------------------------------------------------------------------
! Determine grid points
!-----------------------------------------------------------------------
    open(unit,file=dvrfile,form='unformatted',status='old',iostat=ierr)
    if (ierr.ne.0) then
       write(6,'(/,a,/)') 'dvr file does not exist!'
       STOP
    endif

    string=''
    do i=1,nmodes
! execute rdgrid for current mode
       k=len_trim(prog)
       j=len_trim(adat)
       write(string(1:k),'(a)') prog(1:k)
       k=k+1
       fmat=''       
       write(fmat(1:10),'(a10)') '(a1,i2,a2,'
       if (j.ge.10) then
          write(fmat(11:13),'(a1,i2)') 'a',j
          write(fmat(14:14),'(a)') ')'
       else
          write(fmat(11:12),'(a1,i1)') 'a',j
          write(fmat(13:13),'(a)') ')'
       endif
       l=k+5+j
       write(string(k:l),fmat) ' ',i,' >',adat(1:j)
       k=len_trim(string)
       call excmd(string(1:k))

! read rdgrid output
       open(unit,file=adat,form='formatted',status='old',iostat=ierr)
       rewind(unit)

! read mode labels
       if (i.eq.1) then
5         continue
          read(unit,'(a)') string
          if (string(8:12).ne.'label') goto 5
          read(unit,*)
          do j=1,nmodes
             call rdinpf(unit,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
             read(keyword(2),*) aidof(j)
          enddo
          rewind(unit)
       endif
10     continue
       read(unit,'(a)') string
       if (string(3:7).ne.'dof =') goto 10
       read(unit,*)
20     continue
       read(unit,'(a)',end=30) string
       if (string(1:2).ne.'--') then
          read(string,'(2x,i3,32x,i3)') j,k
          if (j.gt.maxpnt.or.k.gt.maxpnt) then
             write(6,'(/,a,/)') 're-dimension maxpnt in vcsrfmod'
             STOP
          endif
          if (k.ne.0) then
             read(string,'(6x,F15.10,20x,F15.10)') gpnt(i,j),gpnt(i,k)
          else
             read(string,'(6x,F15.10)') gpnt(i,j)
          endif
          read(string(22:22),'(a)') atmp
          if (atmp.eq.'E') then
             read(string(23:25),'(F3.0)') ftmp
             gpnt(i,j)=gpnt(i,j)*(10.0**ftmp)
          endif
          if (k.ne.0) then
             read(string(57:57),'(a)') atmp
             if (atmp.eq.'E') then
                read(string(58:60),'(F3.0)') ftmp
                gpnt(i,k)=gpnt(i,k)*(10.0**ftmp)
             endif
          endif
          goto 20
       endif
       if (k.ne.0) then
          npnt(i)=k
       else
          npnt(i)=j
       endif
       close(unit)
    enddo

! clean-up
    string=''
    write(string(1:3),'(a3)') 'rm '
    k=len_trim(adat)
    write(string(4:3+k),'(a)') adat(1:k)
    call excmd(string)

    return

30  continue
    write(6,'(/,a,/)') 'Problem reading the output of rdgrid'
    STOP

  end subroutine getgrid

!#####################################################################

  subroutine cluster_info(ucmap,ucpnt)
    
    implicit none

    integer(long)                              :: i,k,m,n,u
    integer(long)                              :: ncomb,maxord
    integer(long), dimension(nmodes)           :: combconf,itmp,dof
    integer(vlong), dimension(nmodes)          :: ncombchange,combchk
    integer(long), dimension(2**nmodes,nmodes) :: ucmap
    integer(long), dimension(2**nmodes)        :: ucpnt
    logical(kind=4)                            :: luc

!-----------------------------------------------------------------------
! Determine DOF mapping
!-----------------------------------------------------------------------
    n=0
    do m=1,nmodes
       dof(m)=0
       if (gcut(m).gt.0) then
          n=n+1
          dof(n)=m
       endif
    enddo

!-----------------------------------------------------------------------
! Determine the uncombined clusters (UCs) that are contained within the 
! combined cluster (CC)
!-----------------------------------------------------------------------
    maxord=0
    do m=1,npart
       k=0
       do n=1,nmodes
          if (particle(m,n).eq.1) k=k+1
       enddo
       maxord=maxord+k
    enddo

    ncomb=2**maxord

    do m=1,maxord
       ncombchange(m)=2**(m-1)
       combchk(m)=0
       combconf(m)=1
    enddo

! Loop over all combinations of the DOFs, ignore those that do not contain
! contributions from all particles and save those that do
    nuc=0
    do k=1,ncomb
       luc=.true.
       do m=1,maxord
          combchk(m)=combchk(m)+1
          if (combchk(m).gt.ncombchange(m)) then
             combchk(m)=1
             combconf(m)=combconf(m)+1
             if (combconf(m).gt.2) combconf(m)=1
          endif
       enddo

       call zerovxi(itmp,nmodes)
       do m=1,maxord
          if (combconf(m)-1.eq.1) then
             do n=1,npart
                if (particle(n,dof(m)).eq.1) itmp(n)=itmp(n)+1
             enddo
          endif
       enddo

       do m=1,npart
          if (itmp(m).eq.0) luc=.false.
       enddo
       
       if (luc) then
          nuc=nuc+1
          do m=1,maxord
             if (combconf(m)-1.eq.1) ucmap(nuc,dof(m))=1
          enddo
       endif
    enddo

!-----------------------------------------------------------------------
! Determine the number of points on the full grid for each UC
!-----------------------------------------------------------------------
    do k=1,nuc
       ucpnt(k)=1
       do m=1,nmodes
          if (ucmap(k,m).eq.1) then
             ucpnt(k)=ucpnt(k)*npnt(m)
          endif
       enddo
    enddo

    return
  end subroutine cluster_info

!#####################################################################

  subroutine combined_cluster(mr)

    implicit none

    integer(long)                              :: i,k,m,ilbl,s,s1
    real(dop), dimension(mrdim), intent(inout) :: mr
    character(len=80)                          :: string

    if (ladproj) then
!-----------------------------------------------------------------------
! Matrix representation of the projector onto an adiabatic state
!-----------------------------------------------------------------------
       write(ilog,'(/,70a)') ('+',i=1,70)
       write(6,'(/,70a)') ('+',i=1,70)
       
       write(ilog,'(/,2x,a,i2,a,i2,1x,a,/,2x,a,i2)') &
            'Calculating the element',scut,',',scut1,&
            'of the projector onto the adiabatic','state',iadproj
       write(6,'(/,2x,a,i2,a,i2,1x,a,/,2x,a,i2)') &
            'Calculating the element',scut,',',scut1,&
            'of the projector onto the adiabatic','state',iadproj

       do m=1,nmodes
          gcut(m)=ucmap(1,m)
          if (lhdmr) dcut(m)=gcut(m)
          dcut2(1,m)=gcut(m)
       enddo

       call makecuts(mr,mr(mrconst),mr(mruc(1)),1,mr(mruccoo(1)))

    else
!-----------------------------------------------------------------------
! Potential matrix elements
!-----------------------------------------------------------------------
! Loop over the uncombined clusters that are contained in the combined
! cluster
!-----------------------------------------------------------------------
       if (lhdmr) then
          if (nuc.eq.1) then
             string='Calculating the Combined Cluster'
             write(ilog,'(/,2x,a)') &
                  'The Combined Cluster contains one Uncombined Cluster'
             write(6,'(/,2x,a)') &
                  'The Combined Cluster contains one Uncombined Cluster'
          else
             string='Uncombined Cluster'
             write(ilog,'(/,x,i2,x,2a)') nuc,&
                  'Uncombined Clusters will be summed to create the',& 
             ' Combined Cluster'
             write(6,'(/,x,i2,x,2a)') nuc,&
                  'Uncombined Clusters will be summed to create the',&
                  ' Combined Cluster'
          endif
       endif
       ilbl=len_trim(string)
       
       do k=1,nuc
          if (lhdmr) then
             write(ilog,'(/,70a)') ('+',i=1,70)
             write(6,'(/,70a)') ('+',i=1,70)
             if (nuc.eq.1) then
                write(ilog,'(5x,a/)') string(1:ilbl)
                write(6,'(5x,a/)') string(1:ilbl)
             else
                write(ilog,'(5x,a,x,i1,/)') string(1:ilbl),k
                write(6,'(5x,a,x,i1,/)') string(1:ilbl),k
             endif
          endif
          do m=1,nmodes
             gcut(m)=ucmap(k,m)
             if (lhdmr) dcut(m)=gcut(m)
             dcut2(k,m)=gcut(m)
          enddo
          call makecuts(mr,mr(mrconst),mr(mruc(k)),k,mr(mruccoo(k)))
       enddo

    endif

!-----------------------------------------------------------------------
! Sum the UCs to yield the CC and output the CC
!
! N.B. if ladproj=.true., then we are actually outputting an element
!      of the projector onto an adiabatic state here.
!-----------------------------------------------------------------------
    if (nuc.gt.1) then
       write(ilog,'(/,5x,a)') 'Calculating the Combined Cluster...'
       write(6,'(/,5x,a)') 'Calculating the Combined Cluster...'
    endif
    
    if (nuc.eq.1) then
       call outcc(mr(mruc(1)),ucpnt(nuc),mr(mruccoo(nuc)))
    else
       do k=1,nuc-1
          call sum_ucs(mr(mruccoo(k)),ucpnt(k),mr(mruccoo(nuc)),&
               ucpnt(nuc),k,mr(mruc(k)),mr(mruc(nuc)),mr(mrcc))
       enddo
       call outcc(mr(mrcc),ucpnt(nuc),mr(mruccoo(nuc)))
    endif
    
    return
  end subroutine combined_cluster

!#####################################################################

  subroutine sum_ucs(coo1,n1,coo2,n2,k,ucl,uch,cc)

    implicit none
    
    integer(long)                   :: i,j,k,m,n1,n2,itmp
    real(dop), dimension(n1,nmodes) :: coo1
    real(dop), dimension(n2,nmodes) :: coo2
    real(dop), dimension(n1)        :: ucl
    real(dop), dimension(n2)        :: uch,cc

!-----------------------------------------------------------------------
! Copy the highest-dimension UC to the CC
!-----------------------------------------------------------------------
    if (k.eq.1) then
       do i=1,n2
          cc(i)=uch(i)
       enddo
    endif

!-----------------------------------------------------------------------
! Loop over the CC points and the UC points, adding the UC to the CC at
! the points at which the two correspond
!-----------------------------------------------------------------------
    do i=1,n2
       do j=1,n1
          itmp=0
          do m=1,nmodes
             if (dcut2(k,m).eq.1) then
                if (coo1(j,m).ne.coo2(i,m)) itmp=itmp+1
             endif
          enddo
          
! If the UC and CC coords map onto each other,add the current UC to the CC
          if (itmp.eq.0) then
             cc(i)=cc(i)+ucl(j)
          endif

       enddo
    enddo

    return
  end subroutine sum_ucs

!#####################################################################

  subroutine outcc(cc,n,q)

    implicit none

    integer(long)                  :: i,j,k,n,dim
    real(dop), dimension(n)        :: cc
    real(dop), dimension(n,nmodes) :: q,qtmp
    real(dop), parameter           :: tol=0.000001
    character(len=10)              :: fmat
    logical(kind=4)                :: lzero

    lzero=.true.

!-----------------------------------------------------------------------
! Copy q to qtmp such that the first N elements are the coords of the
! DOFs involved in the N-dimensional cut
!-----------------------------------------------------------------------
    call zeromxd(qtmp,n,nmodes)
    do j=1,n
       k=0
       do i=1,nmodes
          if (dcut2(nuc,i).eq.1) then
             k=k+1
             qtmp(j,k)=q(j,i)
          endif
       enddo
    enddo
    dim=k

!-----------------------------------------------------------------------
! Output the values of the CC at the grid points
!-----------------------------------------------------------------------
    do i=1,n
! Keep track of zero surfaces
       if (abs(cc(i)).gt.tol) lzero=.false.
       if (.not.ltest) then
          if (lbin) then
             write(iout) cc(i)
          else if (lascii) then
             write(iout,'(F15.6)') cc(i)
          else
             fmat=''
             write(fmat(1:8),'(a1,i1,a6)') '(',dim+1,'F15.6)'
             write(iout,fmat) (qtmp(i,k), k=1,dim), cc(i)
          endif
       endif
    enddo

! If surface is zero everywhere, flag this in the log file
    if (lzero) then
       write(ilog,'(/,5x,a)') 'Zero surface!'
       write(6,'(/,5x,a)') 'Zero surface!'
    endif

    return
  end subroutine outcc

!#####################################################################
! makecuts: calculates the cluster uc numbered ucnum
!#####################################################################

  subroutine makecuts(mr,const,uc,ucnum,uccoo)

    implicit none

    integer(long)                                 :: m,n,ncut,nd,ndcut,&
                                                     dim,ilbl,l,u,ucnum
    integer(long), dimension(nmodes)              :: dof,conf,dmap,dconf
    integer(vlong)                                :: ntot,i,k
    integer(vlong), dimension(nmodes)             :: nchange,chk,ndchange,&
                                                     dchk
    integer(long), dimension(maxcut)              :: istart
    real(dop), dimension(mrdim), intent(inout)    :: mr
    real(dop), dimension(nmodes)                  :: q,qtmp,qref
    real(dop), dimension(nsta,nsta)               :: const
    real(dop), dimension(maxsta,maxsta)           :: pot
    real(dop), parameter                          :: tol=0.0001
    real(dop), dimension(2**maxcut)               :: fcuts,ftmp
    integer(long), dimension(2**maxcut,nmodes)    :: mchk,tmp
    integer(long), dimension(2**maxcut)           :: pos
    integer(long), dimension(2**maxcut,2**maxcut) :: sub
    real(dop), dimension(ucpnt(ucnum))            :: uc
    real(dop), dimension(ucpnt(ucnum),nmodes)     :: uccoo
    character(len=20)                             :: atmp

    do m=1,2**maxcut
       fcuts(m)=0.0
       pos(m)=0.0
       do n=1,nmodes
          mchk(m,n)=0
          tmp(m,n)=0
       enddo
    enddo

!-----------------------------------------------------------------------
! In the following:
!-----------------------------------------------------------------------
!
! ncut:       no. DOFs in the subset considered
!
! dof(m):     mapping of DOF m of the subset to the full set of modes
!
! ntot:       total no. points on the full grid in the subspace of 
!             modes considered
!
! nchange(m): out of the total no. of increments (ntot), increment the
!             position on grid m cyclically every nchange(m) increments
!
! nd:         no. of DOFs defining the surfaces to be subtracted from
!             the element of the potential matrix of interest
!
! ndcut:      no. of cuts to be subtracted from the element of the 
!             potential matrix of interest (of maximum dimension 
!             nd-1)
!
! N.B. For use with POTFIT, the output file containing the potential
!      values has to be written with the index of the first specified 
!      DOF running fastest. Any changes must ensure this stays true.
!
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
! If a coordinate transformation is being used, determine the values
! of the grid coordinates corresponding to the reference values of the
! VCHAM coordinates
!-----------------------------------------------------------------------
    if (ltrans) then
       do m=1,nmodes
          qref(m)=0.0
       enddo
       call qv2qg(qref,nmodes)
    endif

!-----------------------------------------------------------------------
! Determine grid dimension, DOF mapping, etc
!-----------------------------------------------------------------------
    ncut=0
    do m=1,nmodes
       q(m)=0
       dof(m)=0
       if (gcut(m).gt.0) then
          ncut=ncut+1
          dof(ncut)=m
       endif
    enddo
    if (ncut.gt.maxcut) then
       write(6,'(/,x,a,/)') 'Re-dimension maxcut!'
       STOP
    endif
    ntot=1
    do m=1,ncut
       ntot=ntot*npnt(dof(m))
       chk(m)=0
       conf(m)=1
    enddo    
    nchange(1)=1
    k=1
    do m=2,ncut
       k=k*npnt(dof(m-1))
       nchange(m)=k
    enddo

!-----------------------------------------------------------------------
! Determine information defining any cuts to be subtracted from the
! element of the potential matrix of interest
!-----------------------------------------------------------------------
    nd=0
    do m=1,nmodes
       dmap(m)=0
       if (dcut(m).gt.0) then
          nd=nd+1
          dmap(nd)=m
       endif
    enddo
    ndcut=2**nd
    if (nd.gt.maxcut) then
       write(6,'(/,x,a,/)') 'Re-dimension maxcut!'
       STOP
    endif
    do m=1,nd
       ndchange(m)=2**(m-1)
       dchk(m)=0
       dconf(m)=1
    enddo

    if (ladproj) then
       write(ilog,'(/,2x,a,x,I9)') &
            'Number of grid points:',ntot
       write(6,'(/,2x,a,x,I9)') &
            'Number of grid points:',ntot
       write(ilog,'(/,70a)') ('+',i=1,70)
       write(6,'(/,70a)') ('+',i=1,70)
    else
       if (nd.eq.0) then
          atmp=''
          k=1
          do m=1,ncut
             write(atmp(k:k+2),'(i2,x)') dof(m)
             k=k+4
          enddo
          write(ilog,'(/,70a)') ('+',i=1,70)
          write(6,'(/,70a)') ('+',i=1,70)
          write(ilog,'(/,2x,a,/,2x,a,x,a)') &
               'Calculation of the full potential in the space spanned',&
               'by the modes:',atmp(1:len_trim(atmp))
          write(6,'(/,2x,a,/,2x,a,x,a)') &
               'Calculation of the full potential in the space spanned',&
               'by the modes:',atmp(1:len_trim(atmp))
          write(ilog,'(/,70a)') ('+',i=1,70)
          write(6,'(/,70a)') ('+',i=1,70)
          write(ilog,*)
       else
          atmp=''
          k=1
          do m=1,nd
             write(atmp(k:k+2),'(i2,x)') dmap(m)
             k=k+4
          enddo
          write(ilog,'(5x,i1,3a)') nd,&
               '-dimensional cluster defined wrt ',&
               'the modes: ',atmp(1:len_trim(atmp))
          write(6,'(5x,i1,3a)') nd,&
               '-dimensional cluster defined wrt ',&
               'the modes: ',atmp(1:len_trim(atmp))
          write(ilog,'(/,5x,a,10x,I9)') 'Total number of grid points:',ntot
          write(6,'(/,5x,a,10x,I9)') 'Total number of grid points:',ntot
          write(ilog,'(/,5x,a,I9)') &
               'Total number of potential evaluations:',ntot*ndcut
          write(6,'(/,5x,a,I9)') &
               'Total number of potential evaluations:',ntot*ndcut
          write(ilog,'(70a)') ('+',i=1,70)
          write(6,'(70a)') ('+',i=1,70)
       endif
    endif

!-----------------------------------------------------------------------
! Determine which cuts have to be subtracted from which to form the
! Cut-HDMR component functions: 
!
! pos(k)=l ===> position k in fcuts is taken by dcut l
!
! sub(k,n)=1 ===> dcut n is to be subtracted from dcut k (with the 
!                 integers n and k refering to to the cut numbers
!                 after fcuts is filled)
!-----------------------------------------------------------------------
    do k=1,ndcut
       call zerovxd(qtmp,nmodes)
       dim=0
       do m=1,nd
          mchk(k,dmap(m))=0
          dchk(m)=dchk(m)+1
          if (dchk(m).gt.ndchange(m)) then
             dchk(m)=1
             dconf(m)=dconf(m)+1
             if (dconf(m).gt.2) dconf(m)=1
          endif
          if (dconf(m)-1.eq.1) then
             dim=dim+1
             mchk(k,dmap(m))=1
          endif
       enddo

       ! ilbl:   start of current block
       ! ilbl+m: end of current block
       ilbl=1
       do l=0,dim-1
          n=factorial(nd)
          m=n
          n=factorial(nd-l)
          m=m/n
          n=factorial(l)
          m=m/n
          ilbl=ilbl+m
       enddo
       n=factorial(nd)
       m=n
       n=factorial(nd-dim)
       m=m/n
       n=factorial(dim)
       m=m/n
       m=m-1
       do l=ilbl,ilbl+m                
          if (pos(l).eq.0) then
             pos(l)=k
             goto 5
          endif
       enddo
5      continue
    enddo

! Re-arrange mchk so that mchk(k,m) corresponds to the kth position in the
! fcuts array instead of the kth cut
    do k=1,ndcut
       do m=1,nmodes
          tmp(k,m)=mchk(pos(k),m)
       enddo
    enddo
    do k=1,ndcut
       do m=1,nmodes
          mchk(k,m)=tmp(k,m)
       enddo
    enddo

! Loop over cuts, determining which other cuts to subtract to form
! the Cut-HDMR component functions
    do k=1,ndcut
       do n=1,k-1
          sub(k,n)=1
          do m=1,nmodes
             if (mchk(k,m).ne.1.and.mchk(n,m).eq.1) then
                sub(k,n)=0
             endif
          enddo          
       enddo
    enddo

!-----------------------------------------------------------------------
! Determine the requested potential function at all grid points: either
! the full potential (ndcut=0) or a Cut-HDMR component function of
! of dimension ndcut (ndcut>0)
!-----------------------------------------------------------------------
! Loop over grid points
    do i=1,ntot
! Loop over DOFs, cyclically incrementing the position on the grid for 
! DOF m when the limit nchange(m) is reached
       do m=1,ncut
          chk(m)=chk(m)+1
          if (chk(m).gt.nchange(m)) then           
             chk(m)=1
             conf(m)=conf(m)+1
             if (conf(m).gt.npnt(dof(m))) conf(m)=1 
          endif
          q(dof(m))=gpnt(dof(m),conf(m))
       enddo

! Ensure that any modes that the cluster is NOT defined wrt are at their
! reference values
       do k=1,nmodes
          if (gcut(k).ne.1) q(k)=qref(k)
       enddo

! Store coordinates of the ith grid point of the current cluster
       do k=1,nmodes
          uccoo(i,k)=q(k)
       enddo

! Perform any requested coordinate transformations
       call cpvxd(q,qtmp,nmodes)
       if (ltrans) call qg2qv(qtmp,nmodes)

       if (nd.eq.0) then
!-----------------------------------------------------------------------
! Full potential
!-----------------------------------------------------------------------
          call evalpotmat(qtmp(1:nmodes),mr,pot)
          if (ladproj) then
             uc(i)=pot(scut,scut1)
          else
             uc(i)=pot(scut,scut1)-const(scut,scut1)
          endif
       else
!-----------------------------------------------------------------------
! Cut-HDMR component functions
!
! dconf(m)=1 ===> mode dmap(m) takes its reference value
! dconf(m)=2 ===> mode dmap(m) takes the value of mode dmap(m) at 
!                 the point q
!-----------------------------------------------------------------------
          call zerovxd(fcuts,2**maxcut)
          do k=1,ndcut
             dim=0
             do m=1,nd
                dchk(m)=dchk(m)+1
                if (dchk(m).gt.ndchange(m)) then
                   dchk(m)=1
                   dconf(m)=dconf(m)+1
                   if (dconf(m).gt.2) dconf(m)=1
                endif
                if (dconf(m).eq.1) then
                   qtmp(dmap(m))=qref(dmap(m))
                else
                   qtmp(dmap(m))=q(dmap(m))
                endif
                if (dconf(m).eq.2) dim=dim+1
             enddo

! Ensure that any modes that the cluster is NOT defined wrt are at their
! reference values
             do l=1,nmodes
                if (gcut(l).ne.1) qtmp(l)=qref(l)
             enddo
             
             if (ltrans) call qg2qv(qtmp,nmodes)

             call evalpotmat(qtmp(1:nmodes),mr,pot)

! Store potential value in the fcuts array for use later
!
! The fcuts array is structured such that the reference value is
! stored 1st, then the one-dimensional cuts, then the two-dimensional cuts,
! and so-on
             ilbl=1
             do l=0,dim-1
                n=factorial(nd)
                m=n
                n=factorial(nd-l)
                m=m/n
                n=factorial(l)
                m=m/n
                ilbl=ilbl+m
             enddo
             n=factorial(nd)
             m=n
             n=factorial(nd-dim)
             m=m/n
             n=factorial(dim)
             m=m/n
             m=m-1
             
             do l=ilbl,ilbl+m                
                if (fcuts(l).eq.0) then
                   fcuts(l)=pot(scut,scut1) 
                   goto 10
                endif
             enddo
10           continue
          enddo

! Calculate the Cut-HDMR component functions
          do k=1,ndcut
             ftmp(k)=fcuts(k)
             do n=1,k-1
                if (sub(k,n).eq.1) then
                   ftmp(k)=ftmp(k)-fcuts(n)
                endif
             enddo
             fcuts(k)=ftmp(k)
          enddo
          uc(i)=ftmp(ndcut)
       endif

! Perform any energy unit conversions
       uc(i)=uc(i)*conv

    enddo

    return
  end subroutine makecuts

!#####################################################################
  
  subroutine qg2qv(q,nmodes)

    implicit none

    integer(long)                :: nmodes,ith1,ith2,m
    real(dop), dimension(nmodes) :: q
    real(dop)                    :: th1,th2,tmp,ref,pi

    pi=2.0*acos(0.0d0)

    select case (itrans)

    case (1) ! costh1, costh2
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       th1=q(ith1)
       th2=q(ith2)
       tmp=acos(th1)
       th1=tmp
       tmp=acos(th2)
       th2=tmp
       tmp=th1-ref
       th1=tmp
       tmp=th2-ref
       th2=tmp
       q(ith1)=(1.0/sqrt(2.0))*(th1+th2)
       q(ith2)=(1.0/sqrt(2.0))*(th1-th2)
       return
   
    case (2) ! costh1
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       th1=q(ith1)
       tmp=acos(th1)
       th1=tmp
       tmp=th1-ref
       th1=tmp
       th2=0.0
       q(ith1)=(1.0/sqrt(2.0))*(th1+th2)
       q(ith2)=(1.0/sqrt(2.0))*(th1-th2)
       return
       
    case (3) ! costh2
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       th2=q(ith2)
       tmp=acos(th2)
       th2=tmp
       tmp=th2-ref
       th2=tmp
       th1=0.0
       q(ith1)=(1.0/sqrt(2.0))*(th1+th2)
       q(ith2)=(1.0/sqrt(2.0))*(th1-th2)
       return
       
    end select

    return
  end subroutine qg2qv

!#####################################################################

  subroutine qv2qg(q,nmodes)

    implicit none

    integer(long)                :: nmodes,ith1,ith2,m
    real(dop), dimension(nmodes) :: q
    real(dop)                    :: tmp1,tmp2,ref,pi

    pi=2.0*acos(0.0d0)

    select case (itrans)

    case (1) ! costh1, costh2
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       tmp1=(sqrt(2.0)/2.0)*(q(ith1)+q(ith2))
       tmp2=(sqrt(2.0)/2.0)*(q(ith1)-q(ith2))
       tmp1=tmp1+ref
       tmp2=tmp2+ref
       tmp1=cos(tmp1)
       tmp2=cos(tmp2)
       q(ith1)=tmp1
       q(ith2)=tmp2
       return

    case (2) ! costh1
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       tmp1=(sqrt(2.0)/2.0)*(q(ith1)+q(ith2))
       tmp2=(sqrt(2.0)/2.0)*(q(ith1)-q(ith2))
       tmp1=tmp1+ref
       tmp2=tmp2+ref
       tmp1=cos(tmp1)
       tmp2=cos(tmp2)
       q(ith1)=tmp1
       q(ith2)=tmp2      
       return

    case (3) ! costh2
       ith1=0
       ith2=0
       do m=1,nmodes
          if (aidof(m).eq.'costh1') ith1=m
          if (aidof(m).eq.'costh2') ith2=m
       enddo
       ref=(120.0/180.0)*pi
       tmp1=(sqrt(2.0)/2.0)*(q(ith1)+q(ith2))
       tmp2=(sqrt(2.0)/2.0)*(q(ith1)-q(ith2))
       tmp1=tmp1+ref
       tmp2=tmp2+ref
       tmp1=cos(tmp1)
       tmp2=cos(tmp2)
       q(ith1)=tmp1
       q(ith2)=tmp2      
       return

    end select

    return
  end subroutine qv2qg

!#####################################################################

  function factorial(n)

    implicit none

    integer(long) :: n,i,fac,factorial

    fac=1
    do i=1,n
       fac=fac*i
    enddo

    factorial=fac

    return
  end function factorial

!#####################################################################

  subroutine evalpotmat(q,mr,pot)
    
    implicit none

    integer(long)                              :: e2,error
    real(dop), dimension(maxsta)               :: vadiab
    real(dop), dimension(3*maxsta)             :: work


    integer(long)                              :: i,j,k,l
    real(dop), dimension(nmodes)               :: q
    real(dop), dimension(mrdim), intent(inout) :: mr
    real(dop), dimension(maxsta,maxsta)        :: diabv,pot
    real(dop), dimension(maxsta,maxsta)        :: diabvec
    real(dop), dimension(maxsta)               :: v
    logical(kind=4)                            :: lvec

!-----------------------------------------------------------------------
! Zero potential array
!-----------------------------------------------------------------------
    do i=1,nsta
       do j=1,nsta
          pot(j,i)=0.0
       enddo
    enddo

!-----------------------------------------------------------------------
! If calculation is a block-diagonalisation or the calculation of a
! projector onto an adiabatic state, then calculate the eigenvectors of
! the potential matrix
!-----------------------------------------------------------------------
    if (lblock.or.ladproj) then
       lvec=.true.
    else
       lvec=.false.
    endif

!-----------------------------------------------------------------------
! Evaluate the diabatic potential matrix
!-----------------------------------------------------------------------
    call pes(q(1:nmodes),v,mr(mrconst),mr(mrkappa),&
         mr(mrlambda),mr(mrgamma),mr(mriota),mr(mrmu),mr(mrexpcoe),&
         mr(mrhigher),mr(mrtermk),mr(mrcpcoe),&
         mr(mrswitch),mr(mrcp2d),lvec,diabvec,diabv)

    do i=1,nsta
       do j=i,nsta
          if (nanchk(diabv(i,j))) then
             write(6,'(/,2x,a,/)') &
                  "Failure in subroutine pes, NaN returned."
             STOP
          endif
       enddo
    enddo

    if (ladproj) then
! Projector onto an adiabatic state
       if (lblock) call blockdiag(diabvec,diabv)
       do i=1,nsta
          do j=1,nsta
             pot(j,i)=diabvec(j,iadproj)*diabvec(i,iadproj)
          enddo
       enddo
    else if (ladiab) then
! Adiabatic potential
       do i=1,nsta
          pot(i,i)=v(i)
       enddo
    else
! Diabatic potential
       if (lblock) call blockdiag(diabvec,diabv)
       do i=1,nsta
          do j=i,nsta
             pot(i,j)=diabv(i,j)
          enddo
       enddo
    endif

  end subroutine evalpotmat

!#####################################################################
!
! blockdiag: block diagonalises the diabatic potential using the
!            transformation matrix T:
!
!            T = A*B
!            A = U*U_bd^T
!            B = 1/sqrt(U_bd*U_bd^T)
!
!            Here, U is the matrix of eigenvectors of the diabatic
!            potential and U_bd is the block diagonal part of U.
!
!#####################################################################
  subroutine blockdiag(vec,diabv)

    implicit none
    
    integer(long)                       :: i,j,k,l,n,dim,e2,error
    integer(long), dimension(nsta)      :: ipiv
    real(dop), dimension(maxsta,maxsta) :: vec,vecbd,diabv
    real(dop), dimension(nsta,nsta)     :: amat,bmat,tmp,tmat
    real(dop), dimension(3*nsta)        :: work
    real(dop), dimension(nsta)          :: lambda
    real(dop), dimension(nsta+1)        :: invwork
    
!-----------------------------------------------------------------------
! Construct U_bd
!-----------------------------------------------------------------------
    do i=1,nsta
       do j=1,nsta
          vecbd(j,i)=0.0
       enddo
    enddo

    k=1
    do n=1,nblock
       dim=blockdim(n)       
       do i=k,k+dim-1
          do j=k,k+dim-1
             vecbd(j,i)=vec(j,i)
          enddo
       enddo
       k=k+dim
    enddo

!-----------------------------------------------------------------------  
! Construct the A matrix
!-----------------------------------------------------------------------  
    do i=1,nsta
       do j=1,nsta
          amat(i,j)=0.0
          do k=1,nsta
             amat(i,j)=amat(i,j)+vec(i,k)*vecbd(j,k)
          enddo
       enddo
    enddo

!-----------------------------------------------------------------------
! Construct the B matrix
!-----------------------------------------------------------------------

! Calculate b=(U_bd*U_bd^T)^0.5
       do i=1,nsta
          do j=1,nsta
             tmp(i,j)=0.0
             do k=1,nsta
                tmp(i,j)=tmp(i,j)+vecbd(i,k)*vecbd(j,k)
             enddo
          enddo
       enddo

       e2=3*nsta
       call dsyev('V','U',nsta,tmp,nsta,lambda,work,e2,error)       
       if (error .ne. 0) then
          write(6,'(/,2a,/)') &
               'Diagonalisation of U_bd*U_bd^T failed in subroutine', &
               'blockdiag'
          STOP
       endif

       do i=1,nsta
          if (lambda(i).lt.0.and.abs(lambda(i)).lt.lamtol) lambda(i)=0.0
          lambda(i)=sqrt(lambda(i))
       enddo

       do i=1,nsta
          do j=1,nsta
             bmat(i,j)=0.0
             do k=1,nsta
                bmat(i,j)=bmat(i,j)+tmp(i,k)*lambda(k)*tmp(j,k)
             enddo
          enddo
       enddo

! Regularise and invert b=(U_bd*U_bd^T)^0.5
       call invert(bmat)

!-----------------------------------------------------------------------
! Construct the transformation matrix T = A*B
!-----------------------------------------------------------------------
       do i=1,nsta
          do j=1,nsta
             tmat(i,j)=0.0d0
             do k=1,nsta
                tmat(i,j)=tmat(i,j)+amat(i,k)*bmat(k,j)
             enddo
          enddo
       enddo

!-----------------------------------------------------------------------
! Transform the diabatic potential to block diagonal form
!-----------------------------------------------------------------------
       do i=1,nsta
          do j=1,nsta
             tmp(i,j)=0.0
             do k=1,nsta
                do l=1,nsta
                   tmp(i,j)=tmp(i,j)+tmat(k,i)*diabv(k,l)*tmat(l,j)
                enddo
             enddo
          enddo
       enddo

       do i=1,nsta
          do j=1,nsta
             diabv(i,j)=tmp(i,j)             
          enddo
       enddo

!-----------------------------------------------------------------------
! Check that all is OK
!-----------------------------------------------------------------------
       do i=1,nsta
          do j=i,nsta
             if (nanchk(diabv(i,j))) then
                write(6,'(/,2x,a,x,2(i2,x),a,/)') &
                     'Error in blockdiag: the element indexed',i,j,&
                     'is returned as NaN'
                write(ilog,'(/,2x,a,x,2(i2,x),a,/)') &
                     'Error in blockdiag: the element indexed',i,j,&
                     'is returned as NaN'
                STOP
             endif
          enddo
       enddo

!-----------------------------------------------------------------------
! If a projector onto an adiabatic state is to be calculated, then
! calculate the eigenvectors of the block-diagonalised potential and
! save these to the diabvec array (called vec here)
!
! N.B. For W_BD = S^T W S,
!      and V = U^T W U = X^T W_BD X,
!      we have X = S^T U
!
! Here: S --> tmat and U --> vec
!-----------------------------------------------------------------------
       if (ladproj) then
          do i=1,nsta
             do j=1,nsta
                tmp(j,i)=0.0
                do k=1,nsta
                   tmp(j,i)=tmp(j,i)+tmat(k,j)*vec(k,i)
                enddo
             enddo
          enddo
          do i=1,nsta
             do j=1,nsta
                vec(j,i)=tmp(j,i)
             enddo
          enddo
       endif

       return

998    continue
       write(6,'(/,2x,a)') 'Inversion failed in subroutine blockdiag'
       write(6,'(2x,a,/)') 'Failure in dgetrf'
       STOP
       
999    continue
       write(6,'(/,2x,a)') 'Inversion failed in subroutine blockdiag'
       write(6,'(2x,a,/)') 'Failure in dgetri'
       STOP
       
  end subroutine blockdiag

!#####################################################################

  logical function nanchk(f)

    real(dop) :: f

    if (f.ne.f) then
       nanchk=.true.
    else
       nanchk=.false.
    endif

    return
  end function nanchk

!#####################################################################

  subroutine invert(mat)

    implicit none

    integer(long)                   :: dim,i,j,k,e2,error
    real(dop)                       :: trace,eps1
    real(dop), dimension(nsta,nsta) :: mat,tmp,omat
    real(dop), dimension(3*nsta)    :: work
    real(dop), dimension(nsta)      :: lambda
    real(dop), parameter            :: tol=1.0e-10_dop
    logical(kind=4)                 :: lsing

! Flag for a singular matrix
    lsing=.false.

!-----------------------------------------------------------------------
! Diagonalise the matrix
!-----------------------------------------------------------------------
    do i=1,nsta
       do j=1,nsta
          tmp(j,i)=mat(j,i)
       enddo
    enddo

    e2=3*nsta
    call dsyev('V','U',nsta,tmp,nsta,lambda,work,e2,error)       
    if (error .ne. 0) then
       write(6,'(/,a,/)') &
            'Diagonalisation failed in subroutine blockdiag'
       STOP
    endif

!-----------------------------------------------------------------------
! Regularise
!-----------------------------------------------------------------------
    if (lreg) then

       select case(ireg)

       case(1) ! Constant
          do i=1,nsta
             if (lambda(i).lt.tol) then
                lsing=.true.
                do j=1,nsta
                   lambda(i)=lambda(i)+eps
                   goto 10
                enddo
             endif
          enddo
10        continue
          
       case(2) ! Exponential
          trace=0.0
          do i=1,nsta
             trace=trace+mat(i,i)
          enddo

          eps1=eps*trace

          do i=1,nsta
             if(lambda(i).lt.64.0_dop*eps1) then
                lsing=.true.
                lambda(i)=lambda(i)+eps1*exp(-lambda(i)/eps1)
             endif
          enddo
          
       end select
    
    endif

!-----------------------------------------------------------------------
! Calculate the inverse of the matrix
!-----------------------------------------------------------------------
    do i=1,nsta
       do j=1,nsta
          mat(i,j)=0.0d0
          do k=1,nsta
             mat(i,j)=mat(i,j)+tmp(i,k)*tmp(j,k)/lambda(k)
          enddo
       enddo
    enddo

!-----------------------------------------------------------------------
! Keep track of the number of regularisations performed
!-----------------------------------------------------------------------
    if (lsing) nreg=nreg+1

    return

  end subroutine invert

!#####################################################################

end program vcsrf
