!#######################################################################
!
! XCURL: a program to evaluate the extended Curl equation for a given
!        set of states using the output of given derivative coupling 
!        calculations.
!
! SN 02/2013
!
!#######################################################################

program xcurl

  use decimal, only: long,dop
  use vcglobal
  use global, only:pi, keyword,keyorig,maxkey,lc,lwarn
  use maxv, only: maxdim,maxsta,maxset
  use keyunitsmod
  use vchammod
  use xcurlmod
  
  use ioqc, only: getintyp, getxcoo
  use utils, only: vchdefault,vccntrl,vcwrline,uppercase

  use logdat, only: message, routine, inptit
  
  use op1lib, only: zerovxd

  implicit none

!-----------------------------------------------------------------------
! Set default values
!-----------------------------------------------------------------------
  call vchdefault
  call xcdefault
  
!-----------------------------------------------------------------------
! Read input file
!-----------------------------------------------------------------------
  call xcinp

!-----------------------------------------------------------------------
! Create quantum chemistry input
!-----------------------------------------------------------------------
  if (lpnt) then 
     call qcinfo
     call mkpnts
  endif

!-----------------------------------------------------------------------
! Evaluate the extended Curl equation
!-----------------------------------------------------------------------
  if (lcalc) then
     call xceq
  endif

!-----------------------------------------------------------------------
! Output the elements of the F-tensor
!-----------------------------------------------------------------------
  if (lcalc) call outf

!-----------------------------------------------------------------------
! End
!-----------------------------------------------------------------------

  STOP

contains

!#######################################################################

  subroutine xcdefault
    
    implicit none
    
    integer(long) :: i,j,s,s1

    ndat=0
    natm=0
    step=0
    nsta=0
    formout=0
    nqc=0
    igeom=0
    igeompos=0
    igeomneg=0
    step=0.0d0
    nstep=0.0d0
    astem=''
    qcfile=''
    aset=''
    aout=''
    alog=''
    lcalc=.false.
    lpnt=.false.
    lnum=.false.
    l2pnt=.false.
    l3pnt=.false.
    lddr=.false.

    do i=1,maxdim
       aatm=''
       xcoo0(i)=0.0d0
       xcoo(i)=0.0d0
       do s=1,maxsta
          do s1=s+1,maxsta
             tau(i,s,s1)=0.0d0
             do j=1,maxdim 
                postau(i,j,s,s1)=0.0d0
                negtau(i,j,s,s1)=0.0d0
                dtaudq(i,j,s,s1)=0.0d0
                delxtau(i,j,s,s1)=0.0d0
                tauxtau(i,j,s,s1)=0.0d0
                ftensor(i,j,s,s1)=0.0d0
             enddo
          enddo
       enddo
    enddo

    do i=1,maxstring
       aqc(i)=''
       stachk(i)=0
    enddo

    do i=1,maxset
       adat(i)=''
       geom(i)=0
    enddo

    lssel=.false.
    do i=1,maxsta
       issel(i)=0
    enddo

    return
  end subroutine xcdefault

!#######################################################################

  subroutine xcinp

    implicit none
    
    integer(long) :: i,j,k,ic,ierr,iz,iz0,ilbl

!-----------------------------------------------------------------------
! Determine input file name and open files
!-----------------------------------------------------------------------
    ainp=''
    call getarg(1,ainp)
    if (ainp.eq.'') then
       write(6,'(/,2x,a,/)') 'Input file not specified!'
       STOP
    endif

    ilbl=index(ainp,'.')-1
    if (ilbl .le. 0) then
       ilbl=index(ainp,' ')-1
       ainp=ainp(1:ilbl)//'.inp'
    endif

    aout=ainp(1:ilbl)//'.dat'
    alog=ainp(1:ilbl)//'.log'
    
    open(iin,file=ainp,form='formatted',status='old')
    open(iout,file=aout,form='formatted',status='unknown')
    open(ilog,file=alog,form='formatted',status='unknown')

!-----------------------------------------------------------------------
! Read input
!-----------------------------------------------------------------------
10  continue
    call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
    routine = 'xcinp (read XCURL input)'
    call vccntrl(ierr)
    
    i=1
15  continue
    if (keyword(i) .eq. 'end-input') then
       goto 200
       
    else if (keyword(i).eq.'reference_geometry') then
       natm=0
20     continue
       call rdinpf(iin,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
       if (keyword(1).ne.'end-reference_geometry') then
          natm=natm+1            
          read(keyword(1),'(a)') aatm(natm)
! Convert to uppercase to avoid problems with GAMESS-US input
          call uppercase(aatm(natm),len(aatm(natm)))
          read(keyword(2),*) xcoo0(natm*3-2)
          read(keyword(3),*) xcoo0(natm*3-1)
          read(keyword(4),*) xcoo0(natm*3)
          goto 20
       endif
       ncoo=natm*3
       
    else if (keyword(i).eq.'step') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) step
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with step keyword'
          STOP
       endif

    else if (keyword(i).eq.'nact_step') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) nstep
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with nact_step keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'filestem') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),'(a)') astem
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with file_stem keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'run_type') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          if (keyword(i).eq.'fcalc') then
             lcalc=.true.
          else if (keyword(i).eq.'points') then
             lpnt=.true.
          else
             k=len_trim(keyword(i))
             write(6,'(/,2x,2(a,2x),/)') 'Unknown run type:',&
                  keyword(i)(1:k)
             STOP
          endif
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with run_type keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'qcinfo') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),'(a)') qcfile
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with qcinfo keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'nstates') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          read(keyword(i),*) nsta
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with nstates keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'formout') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          if (keyword(i).eq.'molpro') then
             formout=1
          else if (keyword(i).eq.'gaussian') then
             formout=2
          else if (keyword(i).eq.'gamess') then
             formout=3
          else
             k=len_trim(keyword(i))
             write(6,'(/,2x,2a,/)') 'Unknown file format:',&
                  keyword(i)(1:k)
          endif
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with formout keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'files') then
       if (keyword(i+1).eq.'=') then
          i=i+2
          aset=keyword(i)
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with files keyword'
          STOP
       endif
       
    else if (keyword(i).eq.'num_nact') then
       if (keyword(i+1).eq.'=') then
          lnum=.true.
          i=i+2
          read(keyword(i),'(i1)') k
          if (k.eq.2) then 
             l2pnt=.true.
          else if (k.eq.3) then
             l3pnt=.true.
          else
             write(6,'(/,2x,a,/)') &
                  'Unrecognised argument specified with num_nact keyword'
             STOP
          endif
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with num_nact keyword'
          STOP
       endif

    else if (keyword(i).eq.'ddr') then
       lddr=.true.
       lnum=.true.

    else if (keyword(i).eq.'state_select') then
       lssel=.true.
       if (keyword(i+1).eq.'=') then
          i=i+1
25        continue
          i=i+1
          read(keyword(i),*) k
          issel(k)=1
          if (i.lt.ic) goto 25
       else
          write(6,'(/,2x,a,/)') &
               'No argument specified with state_select keyword'
          STOP
       endif
       
    else
       routine = 'xcinp (read XCURL input)'
       message = ' unknown keyword : '
       call vcwrline(i,iz,iz0)
    endif
      
!-----------------------------------------------------------------------
! check if more keywords are to be read, else read new line
!-----------------------------------------------------------------------
    if (i.lt.ic) then
       i=i+1
       go to 15
    else
       go to 10
    endif
    
200 continue

!-----------------------------------------------------------------------
! Check that all required information has been given
!-----------------------------------------------------------------------
    if (.not.lcalc.and..not.lpnt) then
       write(6,'(/,2x,a,/)') 'Run type has not been specified!'
       STOP
    endif
    
    if (natm.eq.0) then
       write(6,'(/,2x,a,/)') 'Reference geometry has not been specified!'
       STOP
    endif
    
    if (nsta.eq.0) then
       write(6,'(/,2x,a,/)') 'No. states has not been specified!'
       STOP
    endif
    
    if (lpnt) then
       if (step.eq.0) then
          write(6,'(/,2x,a,/)') 'Step size has not been specified!'
          STOP
       endif
       if (formout.eq.0) then
          write(6,'(/,2x,a,/)') 'Output format has not been specified!'
          STOP
       endif
       if (astem.eq.'') then
          write(6,'(/,2x,a,/)') 'File stem has not been specified!'
          STOP
       endif
       if (qcfile.eq.'') then
          write(6,'(/,2x,a,/)') 'QC info has not been specified!'
          STOP
       endif
    endif
       
    if (lcalc) then
       if (aset.eq.'') then
          write(6,'(/,2x,a,/)') 'Dataset file has not been specified!'
          STOP
       endif
    endif
    
    return
  end subroutine xcinp

!#######################################################################

  subroutine qcinfo

    implicit none
    
    integer(long)     :: i,j,k,unit
    character(len=80) :: string
    
    unit=30
      
!-----------------------------------------------------------------------
! Read QC information file
!-----------------------------------------------------------------------
    open(unit,file=qcfile,form='formatted',status='old')
    rewind(unit)
30  continue
    read(unit,'(a)') string
    if (string.ne.'end-file') then
       nqc=nqc+1
       aqc(nqc)=string
       goto 30
    endif
    close(unit)

    do i=1,nqc
       k=len_trim(aqc(i))
!-----------------------------------------------------------------------
! Determine which strings contain state numbers to be changed
!-----------------------------------------------------------------------
       do j=1,k
          if (aqc(i)(j:j).eq.'$') then
             if (aqc(i)(j:j+2).eq.'$s1'.or.&
                  aqc(i)(j:j+2).eq.'$s2') stachk(i)=1
          endif
       enddo
!-----------------------------------------------------------------------
! Determine which strings are to be replaced by geometries
!-----------------------------------------------------------------------
       if (aqc(i)(1:k).eq.'$geom') igeom=i
       if (aqc(i)(1:k).eq.'$geom_pos') igeompos=i
       if (aqc(i)(1:k).eq.'$geom_neg') igeomneg=i
     enddo

!-----------------------------------------------------------------------
! Check that all required information has been given
!-----------------------------------------------------------------------
     if (igeom.eq.0) then
        write(6,'(/,2x,a,/)') '$geom keyword not given in qcinfo file.'
          STOP
     endif

!-----------------------------------------------------------------------
! Flag whether or not the NACTs are being calculated by finite
! differences
!-----------------------------------------------------------------------
     if (igeompos.gt.0) lnum=.true.

    return
  end subroutine qcinfo

!#######################################################################

  subroutine mkpnts

    implicit none

    integer(long) :: s,s1

!-----------------------------------------------------------------------
! Determine whether the NACTs are being calculated by finite differences
!-----------------------------------------------------------------------
    if (igeompos.ne.0) lnum=.true.

    if (lnum) then
       if (lddr) then
          do s=1,nsta
             do s1=s+1,nsta
! Check whether the elements F_p,q,s,s1 are to be calculated
                if (lssel) then
                   if (issel(s).eq.0.or.issel(s1).eq.0) cycle
                endif                
                call num_pnts(s,s1)
             enddo
          enddo
       else
          s=0
          s1=0
          call num_pnts(s,s1)
       endif
    else
       call ana_pnts
    endif
    
    return
  end subroutine mkpnts

!#######################################################################

  subroutine num_pnts(s,s1)

    implicit none

    integer(long)                :: s,s1,n,a,ilbl,k,dir,d,nn
    character(len=100)           :: afile
    character(len=80)            :: atmp

!-----------------------------------------------------------------------
! Reference geometry
!-----------------------------------------------------------------------
    dir=0
    a=0
    n=0
! Loop over Cartestian coordinates
    do nn=1,natm*3
! Write filename
       call wrout(afile,s,s1,n,nn,a,dir,lnum)
! Write to output file
       open(idat,file=afile,form='formatted',status='unknown')
       rewind(idat)
       do k=1,nqc
! Write geometries
          call zerovxd(diff,maxdim)
          if (k.eq.igeom) then
             ! No displacements
             call calccoord
             call wrcoord
          else if (k.eq.igeompos) then
             ! Positive displacement (NACT calc)
             diff(nn)=diff(nn)+nstep             
             call calccoord
             call wrcoord
          else if (k.eq.igeomneg) then
             ! Negative displacement (NACT calc)
             diff(nn)=diff(nn)-nstep
             call calccoord
             call wrcoord
          else
! Write character strings
             if (k.ne.igeom) then
                if (stachk(k).eq.1) then
                   call replace(s,s1,aqc(k),atmp)
                   ilbl=len_trim(atmp)
                   write(idat,'(a)') atmp(1:ilbl)
                else
                   ilbl=len_trim(aqc(k))
                   write(idat,'(a)') aqc(k)(1:ilbl)
                endif
             endif
          endif
       enddo
       close(idat)
    enddo

!-----------------------------------------------------------------------
! Displaced geometries
!-----------------------------------------------------------------------
! Loop over components of the NACT (Cartesian coords)
    do nn=1,natm*3
! Loop over atoms (displacements for the calculation of dtaudq)
       do n=1,natm
! Loop over x, y, and z coords (displacements for the calculation of dtaudq)
          do a=1,3
! Loop over +ve and -ve directions (displacements for the calculation of dtaudq)
             do d=1,2
                dir=(-1)**d
! Write filename
                call wrout(afile,s,s1,n,nn,a,dir,lnum)
! Write to output file
                open(idat,file=afile,form='formatted',status='unknown')
                rewind(idat)
                do k=1,nqc
! Write geometries
                   call zerovxd(diff,maxdim)
                   ! Displacement for dtau/dq calc
                   diff(n*3-3+a)=diff(n*3-3+a)+(dir*step)
                   if (k.eq.igeom) then
                      ! No displacements
                      call calccoord
                      call wrcoord
                   else if (k.eq.igeompos) then
                      ! Positive displacement (NACT calc)
                      diff(nn)=diff(nn)+nstep
                      call calccoord
                      call wrcoord
                   else if (k.eq.igeomneg) then
                      ! Negative displacement (NACT calc)
                      diff(nn)=diff(nn)-nstep
                      call calccoord
                      call wrcoord
                   else
! Write character strings
                      if (k.ne.igeom) then
                         if (stachk(k).eq.1) then
                            call replace(s,s1,aqc(k),atmp)
                            ilbl=len_trim(atmp)
                            write(idat,'(a)') atmp(1:ilbl)
                         else
                            ilbl=len_trim(aqc(k))
                            write(idat,'(a)') aqc(k)(1:ilbl)
                         endif
                      endif
                   endif
                enddo
                close(idat)
             enddo
          enddo
       enddo
    enddo
   
    return
  end subroutine num_pnts

!#######################################################################

  subroutine ana_pnts

    implicit none

    integer(long)      :: s,s1,n,nn,a,ilbl,k,dir,d,i,j
    character(len=100) :: afile
    character(len=80)  :: atmp

!-----------------------------------------------------------------------
! Reference geometry
!-----------------------------------------------------------------------
    dir=0
    n=0
    nn=0
    a=0
    call zerovxd(diff,maxdim)
! Loop over pairs of states
    do s=1,nsta
       do s1=s+1,nsta
! Check whether the NACTs coupling states s and s1 are to be calculated
          if (lssel) then
             if (issel(s).eq.0.and.issel(s1).eq.0) cycle
          endif
! Write filename
          call wrout(afile,s,s1,n,nn,a,dir,lnum)
! Write to output file
          open(idat,file=afile,form='formatted',status='unknown')
          rewind(idat)
          do k=1,nqc
             if (k.ne.igeom) then
                if (stachk(k).eq.1) then
                   call replace(s,s1,aqc(k),atmp)
                   ilbl=len_trim(atmp)
                   write(idat,'(a)') atmp(1:ilbl)
                else
                   ilbl=len_trim(aqc(k))
                   write(idat,'(a)') aqc(k)(1:ilbl)
                endif
             else
                call calccoord
                call wrcoord
             endif
          enddo
          close(idat)
       enddo
    enddo
    
!-----------------------------------------------------------------------
! Displaced geometries
!-----------------------------------------------------------------------
! Loop over pairs of states
    do s=1,nsta
       do s1=s+1,nsta
! Check to se whether the gradients of the NACTs coupling states s and s1
! are to be calculated
          if (lssel) then
             if (issel(s).eq.0.or.issel(s1).eq.0) cycle
          endif
! Loop over atoms
          do n=1,natm
! Loop over x, y, and z coords
             do a=1,3
! Loop over zero, +ve and -ve directions
                do d=1,2
                   dir=(-1)**d
! Write filename                     
                   call wrout(afile,s,s1,n,nn,a,dir,lnum)
! Write to output file
                   open(idat,file=afile,form='formatted',status='unknown')
                   rewind(idat)
                   do k=1,nqc
                      if (k.ne.igeom) then
                         if (stachk(k).eq.1) then
                            call replace(s,s1,aqc(k),atmp)
                            ilbl=len_trim(atmp)
                            write(idat,'(a)') atmp(1:ilbl)
                         else
                            ilbl=len_trim(aqc(k))
                            write(idat,'(a)') aqc(k)(1:ilbl)
                         endif
                      else
                         call zerovxd(diff,maxdim)
                         diff(n*3-3+a)=diff(n*3-3+a)+step*dir
                         call calccoord
                         call wrcoord
                      endif
                   enddo
                   close(idat)
                enddo
             enddo
          enddo
       enddo
    enddo

    return
  end subroutine ana_pnts

!#######################################################################

  subroutine wrout(afile,s,s1,n,nn,a,dir,lnum)
    
    implicit none
    
    integer(long)      :: s,s1,n,a,ilbl,jlbl,k,dir,d,nn
    character(len=100) :: afile
    character(len=80)  :: atmp
    logical(kind=4)    :: lnum

    afile=''
    ilbl=len_trim(astem)
    write(afile(1:ilbl),'(a)') astem(1:ilbl)
    ilbl=ilbl+1
    write(afile(ilbl:ilbl),'(a)') '_'
    ilbl=ilbl+1
    atmp=''

    if (n.lt.10) then
       write(atmp(1:4),'(a2,i1)') 'a0',n
    else
       write(atmp(1:3),'(a1,i2)') 'a',n
    endif
    write(afile(ilbl:ilbl+2),'(a3)') atmp(1:3)
    ilbl=ilbl+3
    
    if (a.eq.1) then
       write(afile(ilbl:ilbl+1),'(a2)') 'x_'
    else if (a.eq.2) then
       write(afile(ilbl:ilbl+1),'(a2)') 'y_'
    else if (a.eq.3) then
       write(afile(ilbl:ilbl+1),'(a2)') 'z_'
    else
       write(afile(ilbl:ilbl+1),'(a2)') '0_'
    endif
    ilbl=ilbl+2
    
    if (lnum) then
       if (nn.lt.10) then
          write(afile(ilbl:ilbl+3),'(a2,i1)') 'x0',nn  
       else
          write(afile(ilbl:ilbl+3),'(a1,i2)') 'x',nn
       endif
       ilbl=ilbl+3
    endif

    if (lddr) then
       write(afile(ilbl:ilbl),'(a1)') '_'
       ilbl=ilbl+1
    endif

    if (.not.lnum.or.lddr) then
       if (s.lt.10.and.s1.lt.10) then
          write(afile(ilbl:ilbl+6),'(a2,i1,a3,i1)') &
               's0',s,'_s0',s1
       else if (s.ge.10.and.s1.lt.10) then
          write(afile(ilbl:ilbl+6),'(a1,i2,a3,i1)') &
               's',s,'_s0',s1
       else if (s.lt.10.and.s1.ge.10) then
          write(afile(ilbl:ilbl+6),'(a2,i1,a2,i2)') &
               's0',s,'_s',s1
       else
          write(afile(ilbl:ilbl+6),'(a1,i2,a2,i2)') &
               's',s,'_s',s1
       endif
       ilbl=ilbl+7
    endif

    if (dir.lt.0) then
       write(afile(ilbl:ilbl+3),'(a4)') '_neg'
    else if (dir.gt.0) then
       write(afile(ilbl:ilbl+3),'(a4)') '_pos'
    else
       write(afile(ilbl:ilbl+3),'(a4)') '_ref'
    endif
    ilbl=ilbl+4

    if (formout.eq.1.or.formout.eq.3) then
       ! Molpro or Gamess
       atmp='.inp'
    else if (formout.eq.2) then
       ! Gaussian
       atmp='.com'
    endif
    jlbl=len_trim(atmp)-1
    write(afile(ilbl:ilbl+jlbl),'(a)') atmp(1:jlbl+1)
    
    return
  end subroutine wrout

!#######################################################################

  subroutine replace(s,s1,string,string1)
      
    implicit none
      
    integer(long)     :: s,s1,i,j,k,n,ignore
    character(len=80) :: string,string1
    
    ignore=0
    j=0
    string1=''
    n=len_trim(string)

    do i=1,n
       if (ignore.gt.0) then
          ignore=ignore-1
       else
          if (string(i:i+2).eq.'$s1') then
             if (s.lt.10) then
                j=j+1
                write(string1(j:j),'(i1)') s
             else
                j=j+2
                write(string1(j-1:j),'(i2)') s
             endif
             ignore=2            
          else if (string(i:i+2).eq.'$s2') then
             if (s1.lt.10) then
                j=j+1
                write(string1(j:j),'(i1)') s1
             else
                j=j+2
                write(string1(j-1:j),'(i2)') s1
             endif
             ignore=2
          else
             j=j+1
             string1(j:j)=string(i:i)
          endif
       endif
    enddo
    
    return
  end subroutine replace

!#######################################################################

  subroutine calccoord

    implicit none

    integer(long)                :: n

! Displace reference coordinates
    do n=1,natm*3
       xcoo(n)=xcoo0(n)+diff(n)
    enddo

    return
  end subroutine calccoord

!#######################################################################

  subroutine wrcoord

    implicit none
      
    integer(long)     :: n,len,fmat
    real(dop)         :: atnum
    character(len=80) :: string

    if (formout.eq.1.or.formout.eq.2) then
! Molpro and Gaussian output (xyz format)
       do n=1,natm
          string=''
          write(string(1:2),'(a2)') aatm(n)(1:2)
          write(string(3:51),'(3(2x,F14.10))') &
               xcoo(n*3-2),xcoo(n*3-1),xcoo(n*3)
          len=len_trim(string)
          write(idat,'(a)') string(1:len)
       enddo

    else if (formout.eq.3) then
! Gamess output       
       do n=1,natm
          call getatnum(n,atnum)
          string=''
          if (len_trim(aatm(n)).eq.1) then
             write(idat,100) &
                  aatm(n),',',atnum,',',xcoo(n*3-2),',',xcoo(n*3-1),',',&
                  xcoo(n*3)
          else
             write(idat,101) &
                  aatm(n),',',atnum,',',xcoo(n*3-2),',',xcoo(n*3-1),',',&
                  xcoo(n*3)
          endif          
       enddo

    endif
    
100 format (2a1,3x,F4.2,a1,2(2x,F13.10,a1),2x,F13.10)
101 format (a2,a1,3x,F4.2,a1,2(2x,F13.10,a1),2x,F13.10)

    return
  end subroutine wrcoord

!#######################################################################

  subroutine getatnum(n,atnum)

    implicit none

    integer(long) :: n
    real(dop)     :: atnum

    if (aatm(n).eq.'H') then
       atnum=1.0d0
    else if (aatm(n).eq.'C') then
       atnum=6.0d0
    else if (aatm(n).eq.'N') then
       atnum=7.0d0
    else
       write(6,'(/,4(a,x),/)') &
            'The atom type',aatm(1)(1:2),'is not yet supported.',&
            'See subroutine getatnum to change this.'
       STOP
    endif

    return
  end subroutine getatnum

!#######################################################################

    subroutine xceq

      implicit none
      
      integer(long)              :: unit,i,ifiletyp,j,s,s1,k,len
      real(dop), dimension(ncoo) :: tautmp
      real(dop)                  :: diff

      unit=30

!-----------------------------------------------------------------------
! Determine QC output file names
!-----------------------------------------------------------------------
      call rdset

!-----------------------------------------------------------------------
! Loop over QC output files
!-----------------------------------------------------------------------   
      do i=1,ndat

!-----------------------------------------------------------------------   
! Open file and determine filetype
!-----------------------------------------------------------------------   
         len=len_trim(adat(i))
         write(ilog,'(2a,/)') 'Reading data from file: ',adat(i)(1:len)
         open(unit,file=adat(i),form='formatted',status='unknown')
         rewind(unit)
         call getintyp(unit,ifiletyp,adat(i),ilog)

!-----------------------------------------------------------------------
! Read coordinates
!-----------------------------------------------------------------------   
         call qc_xcoo(unit,ifiletyp,geom(i))

!-----------------------------------------------------------------------
! Read NACTs
!-----------------------------------------------------------------------
         call qc_nact(unit,ifiletyp,diff,geom(i))

      enddo

!-----------------------------------------------------------------------
! Calculate derivative of the NACTs wrt the Cartesian coordinates
!-----------------------------------------------------------------------
      diff=2*step/angstroem
      do s1=1,nsta
         do s=1,nsta
            do j=1,ncoo
               do i=1,ncoo
                  dtaudq(i,j,s,s1)=(postau(i,j,s,s1)-negtau(i,j,s,s1))/diff
               enddo
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate the Curl of the NACT tensor (H-tensor)
!-----------------------------------------------------------------------
      do s1=1,nsta
         do s=1,nsta
            do j=1,ncoo
               do i=1,ncoo
                  delxtau(i,j,s,s1)=dtaudq(j,i,s,s1)-dtaudq(i,j,s,s1)
               enddo
            enddo
         enddo
      enddo

!-----------------------------------------------------------------------
! Calculate the cross product of the NACT tensor with itself (T-tensor)
!-----------------------------------------------------------------------
      if (nsta.gt.2) then
         do s1=1,nsta
            do s=1,nsta
               do j=1,ncoo
                  do i=1,ncoo
                     tauxtau(i,j,s,s1)=0.0d0
                     do k=1,nsta
                        tauxtau(i,j,s,s1)=tauxtau(i,j,s,s1)+&
                             tau(i,s,k)*tau(j,k,s1)-tau(j,s,k)*tau(i,k,s1)
                     enddo
                  enddo
               enddo
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! Calculate the F-tensor:
!
! Abelian case:     F = H
!
! Non-Abelian case: F = H - T
!-----------------------------------------------------------------------
      do s1=1,nsta
         do s=1,nsta
            do j=1,ncoo
               do i=1,ncoo
                  if (nsta.gt.2) then
                     ! Non-Abelian case                     
                     ftensor(i,j,s,s1)=delxtau(i,j,s,s1)-tauxtau(i,j,s,s1)
                  else
                     ! Abelian case
                     ftensor(i,j,s,s1)=delxtau(i,j,s,s1)
                  endif
               enddo
            enddo
         enddo
      enddo

      return
    end subroutine xceq

!#######################################################################

    subroutine rdset

      implicit none

      integer(long) :: unit,ic,ierr,iz,iz0,i

      unit=30
    
      open(unit,file=aset,form='formatted',status='unknown')
      rewind(unit)

10    continue
      call rdinpf(unit,keyword,keyorig,inptit,lc,ic,iz,ierr,maxkey)
      routine = 'rdset (read dataset)'
      call vccntrl(ierr)

      if (keyword(1).ne.'end-files') then
         ndat=ndat+1
         i=1
         adat(ndat)=keyword(i)
! Read any additional arguments
         if (ic.gt.1) then
15          continue
            i=i+1
            if (keyword(i).eq.'geom') then
               i=i+2
               read(keyword(i),*) geom(ndat)
            endif
            if (i.lt.ic) goto 15
         endif
         goto 10
      endif

      close(unit)

      return
    end subroutine rdset

!#######################################################################
    
    subroutine qc_xcoo(unit,ifiletyp,currgeom)

      implicit none

      integer(long) :: unit,ifiletyp,currgeom,iau
      integer(long), dimension(ncoo) :: iatiso

!-----------------------------------------------------------------------
! Read Cartesian coordinates from the current QC output file
!-----------------------------------------------------------------------
      if (lnum) then
         if (ifiletyp.eq.5) then
            call rdxcoo_mol(currgeom,unit)
         else
            write(6,'(/,a,x,i2,x,a,/)') &
                 'Reading of the file type',ifiletyp,&
                 'is not yet supported.'
            STOP
         endif
      else
         iau=1
         call getxcoo(unit,xcoo,ifiletyp,iau,'standard',iatsym,&
              iatiso,ncoo,nsta,nmodes,ilog)
      endif

      return
    end subroutine qc_xcoo
      
!#######################################################################

    subroutine rdxcoo_mol(currgeom,unit)

      implicit none

      integer(long)           :: i,j,currgeom,unit,n,m
      real(dop)               :: isym
      real(dop), dimension(3) :: coo
      character(len=80)       :: string

!-----------------------------------------------------------------------
! Read to the requested geometry
!-----------------------------------------------------------------------
      if (currgeom.eq.0) then
         n=1
      else
         n=currgeom
      endif

      m=0
10    continue
      read(unit,'(a)',end=999) string
      if (string(2:19).eq.'ATOMIC COORDINATES') then
         m=m+1
         if (m.eq.n) then
            goto 20
         else
            goto 10
         endif
      else
         goto 10
      endif

20    continue
      do n=1,3
         read(unit,*)
      enddo

      n=0
      do i=1,ncoo/3
         read(unit,100) isym,coo(1),coo(2),coo(3)
         do j=1,3
            n=n+1
            iatsym(n)=isym
            xcoo(n)=coo(j)*angstroem
         enddo
      enddo

100   format(14x,f4.2,2x,f13.9,2x,f13.9,2x,f13.9)

      return
      
999   continue
      write(6,'(/,a,/)') 'Geometry not found!'
      STOP

    end subroutine rdxcoo_mol

!#######################################################################

    subroutine qc_nact(unit,ifiletyp,diff,currgeom)

      implicit none

      integer(long)        :: unit,ifiletyp,itmp,j,currgeom
      real(dop)            :: ftmp,diff
      real(dop), parameter :: tol=1.0e-7_dop
      logical(kind=4)      :: ldpos,ldneg,ldref

!-----------------------------------------------------------------------
! Determine displacements (for the numerical calculation of dtau_p/dq,
! i.e., q)
!-----------------------------------------------------------------------
      ldref=.false.
      ldpos=.false.
      ldneg=.false.

      diff=0.0d0
      itmp=0
      do j=1,ncoo
         ftmp=xcoo(j)-xcoo0(j)
         if (abs(ftmp).gt.tol) then
            diff=ftmp
            itmp=j
            step=abs(diff)
         endif
      enddo

      diff=diff/angstroem

      if (abs(diff).gt.tol) then
         if (diff.gt.0) then
            ldpos=.true.
         else
            ldneg=.true.
         endif
      else
         ldref=.true.
      endif

!-----------------------------------------------------------------------
! Read the NACTs
!-----------------------------------------------------------------------
      if (ifiletyp.eq.5) then
         if (lnum) then
            if (lddr) then
               call nactmolpro_ddr(unit,ldref,ldpos,ldneg,itmp,currgeom)
            else
               call nactmolpro_num(unit,ldref,ldpos,ldneg,itmp,currgeom)
            endif
         else
            call nactmolpro(unit,ldref,ldpos,ldneg,itmp)
         endif
      else
         write(6,'(/,a,x,i2,x,a,/)') &
              'Reading of NACTs from file type',ifiletyp,&
              'is not yet supported.'
         STOP
      endif

      return
    end subroutine qc_nact

!#######################################################################

    subroutine nactmolpro(unit,ldref,ldpos,ldneg,itmp)

      implicit none
      
      integer(long)              :: unit,s,s1,i,j,n,itmp
      real(dop), dimension(ncoo) :: tautmp
      real(dop), dimension(3)    :: dum
      character(len=80)          :: string
      logical(kind=4)            :: ldpos,ldneg,ldref

      do i=1,natm
         tautmp(i)=0.0d0
      enddo
      
      rewind(unit)

10    continue
      read(unit,'(a)',end=999) string
      if (string(2:12).ne.'SA-MC NACME') goto 10

!-----------------------------------------------------------------------
! See which states the NACTs have been calculated for
!-----------------------------------------------------------------------
      read(string(25:25),'(i1)') s
      read(string(31:31),'(i1)') s1

!-----------------------------------------------------------------------
! Read NACTs
!-----------------------------------------------------------------------
      do i=1,3
         read(unit,*)
      enddo

      n=0
      do i=1,ncoo,3
         read(unit,20) (dum(j),j=1,3)
         do j=1,3
            n=n+1
            tautmp(n)=dum(j)
         enddo
      enddo
!20    format (5x,3(F19.9,1x))
20    format (4x,3(F20.9))

!-----------------------------------------------------------------------
! Fill the appropriate NACT array
!-----------------------------------------------------------------------
      if (ldpos) then
         do j=1,ncoo
            postau(itmp,j,s,s1)=tautmp(j)
            postau(itmp,j,s1,s)=-postau(itmp,j,s,s1)
         enddo
      else if (ldneg) then
         do j=1,ncoo
            negtau(itmp,j,s,s1)=tautmp(j)
            negtau(itmp,j,s1,s)=-negtau(itmp,j,s,s1)
         enddo
      else
         do j=1,ncoo
            tau(j,s,s1)=tautmp(j)
            tau(j,s1,s)=-tau(j,s,s1)
         enddo
      endif

999   continue

      return
    end subroutine nactmolpro

!#######################################################################

    subroutine nactmolpro_num(unit,ldref,ldpos,ldneg,itmp,currgeom)

      implicit none

      integer(long)                   :: unit,itmp,currgeom,i,j,m,n,itmp2,&
                                         ilbl,jlbl,klbl,s,s1
      real(dop), dimension(nsta,nsta) :: tautmp
      real(dop), dimension(ncoo)      :: tmpxcoo
      real(dop), dimension(3)         :: coo
      real(dop)                       :: ftmp,diff
      real(dop), parameter            :: tol=1.0e-7_dop
      logical(kind=4)                 :: ldref,ldpos,ldneg
      character(len=80)               :: string
      character(len=12)               :: atmp,atmp1,atmp2

      rewind(unit)

      do i=1,nsta
         do j=1,nsta
            tautmp(j,i)=0.0d0
         enddo
      enddo

!-----------------------------------------------------------------------
! Determine which nuclear coordinate component of the NACT we are 
! dealing with (for the numerical calculation of dtau_p/dq,
! i.e., p)
!-----------------------------------------------------------------------
      do i=1,3
         if (i.ne.currgeom) n=i         
      enddo

      m=0
10    continue
      read(unit,'(a)',end=999) string
      if (string(2:19).eq.'ATOMIC COORDINATES') then
         m=m+1
         if (m.eq.n) then
            goto 20
         else
            goto 10
         endif
      else
         goto 10
      endif

20    continue
      do n=1,3
         read(unit,*)
      enddo
      
      n=0
      do i=1,ncoo/3
         read(unit,100) coo(1),coo(2),coo(3)
         do j=1,3
            n=n+1
            tmpxcoo(n)=coo(j)*angstroem
         enddo
      enddo
100   format(18x,2x,f13.9,2x,f13.9,2x,f13.9)

      do n=1,ncoo
         ftmp=xcoo(n)-tmpxcoo(n)
         ftmp=abs(ftmp)
         if (ftmp.gt.tol) then
            itmp2=n
            diff=ftmp/angstroem
         endif
      enddo

!-----------------------------------------------------------------------
! Read NACTs
!-----------------------------------------------------------------------
      rewind(unit)
30    continue
      read(unit,'(a)',end=999) string
      if (string(4:32).ne.'Transition moment calculation') goto 30

40    continue
      read(unit,'(a)') string
      if (string(2:8).ne.'*******') then
         if (string(3:14).eq.'MRCI overlap') then
            read(string,'(23x,a12)') atmp
            ilbl=index(atmp,'<')
            jlbl=index(atmp,'>')
            klbl=index(atmp,'|')
            atmp1=''
            j=klbl-ilbl-1
            write(atmp1(1:j),'(a)') atmp(ilbl+1:klbl-1)
            atmp2=''
            j=jlbl-klbl-2
            write(atmp2(1:j),'(a)') atmp(klbl+2:jlbl-1)
            ilbl=index(atmp1,'.')
            read(atmp1(1:ilbl-1),*) s
            ilbl=index(atmp2,'.')
            read(atmp2(1:ilbl-1),*) s1
            read(string,'(38x,F16.12)') tautmp(s,s1)
            tautmp(s,s1)=-tautmp(s,s1)/(2*diff)

!-----------------------------------------------------------------------
! Different orbitals exist for the bra and ket, so take the average of
! tautmp(s,s1) and tautmp(s1,s), taking into account the antisymmetry
! of the NACT
!-----------------------------------------------------------------------
            do s=1,nsta-1
               do s1=s+1,nsta
                  if (tautmp(s,s1).lt.0.and.tautmp(s1,s).gt.0) then
                     ilbl=s
                     jlbl=s1
                  endif
                  if (tautmp(s1,s).lt.0.and.tautmp(s,s1).gt.0) then
                     ilbl=s1
                     jlbl=s
                  endif
                  ftmp=abs(tautmp(s,s1))+abs(tautmp(s1,s))
                  ftmp=ftmp/2.0d0
                  tautmp(ilbl,jlbl)=-ftmp
                  tautmp(jlbl,ilbl)=ftmp
               enddo
            enddo

!-----------------------------------------------------------------------
! Assign to the correct array
!-----------------------------------------------------------------------
            if (ldpos) then
               do s=1,nsta
                  do s1=1,nsta
                     postau(itmp,itmp2,s,s1)=tautmp(s,s1)
                     postau(itmp,itmp2,s1,s)=-postau(itmp,itmp2,s,s1)
                  enddo
               enddo
            else if (ldneg) then
               do s=1,nsta
                  do s1=1,nsta
                     negtau(itmp,itmp2,s,s1)=tautmp(s,s1)
                     negtau(itmp,itmp2,s1,s)=-negtau(itmp,itmp2,s,s1)
                  enddo
               enddo
            else
               do s=1,nsta
                  do s1=1,nsta
                     tau(itmp2,s,s1)=tautmp(s,s1)
                     tau(itmp2,s1,s)=-tau(itmp2,s,s1)
                  enddo
               enddo
            endif

         endif
         goto 40
      endif

      return

999   continue
      write(6,'(/,a,/)') "Wavefunction overlaps not found."
      STOP

    end subroutine nactmolpro_num

!#######################################################################

    subroutine nactmolpro_ddr(unit,ldref,ldpos,ldneg,itmp,currgeom)

      implicit none

      integer(long)                   :: unit,itmp,currgeom,i,j,m,n,itmp2,&
                                         ilbl,jlbl,klbl,s,s1,k
      real(dop)                       :: tautmp
      real(dop), dimension(ncoo)      :: tmpxcoo
      real(dop), dimension(3)         :: coo
      real(dop)                       :: ftmp,diff
      real(dop), parameter            :: tol=1.0e-7_dop
      logical(kind=4)                 :: ldref,ldpos,ldneg
      character(len=80)               :: string
      character(len=12)               :: atmp,atmp1,atmp2

      rewind(unit)

!-----------------------------------------------------------------------
! Determine which nuclear coordinate component of the NACT we are 
! dealing with (for the numerical calculation of dtau_p/dq,
! i.e., p)
!-----------------------------------------------------------------------
      do i=1,3
         if (i.ne.currgeom) n=i
      enddo

      m=0
10    continue
      read(unit,'(a)',end=999) string
      if (string(2:19).eq.'ATOMIC COORDINATES') then
         m=m+1
         if (m.eq.n) then
            goto 20
         else
            goto 10
         endif
      else
         goto 10
      endif

20    continue
      do n=1,3
         read(unit,*)
      enddo
      
      n=0
      do i=1,ncoo/3
         read(unit,100) coo(1),coo(2),coo(3)
         do j=1,3
            n=n+1
            tmpxcoo(n)=coo(j)*angstroem
         enddo
      enddo
100   format(18x,2x,f13.9,2x,f13.9,2x,f13.9)

      do n=1,ncoo
         ftmp=xcoo(n)-tmpxcoo(n)
         ftmp=abs(ftmp)
         if (ftmp.gt.tol) then
            itmp2=n
            diff=ftmp/angstroem
         endif
      enddo      

!-----------------------------------------------------------------------
! Read NACT
!-----------------------------------------------------------------------
30    continue
      read(unit,'(a)',end=999) string
      if (string(2:33).ne.'Construct non-adiabatic coupling') goto 30
      
40    continue
      read(unit,'(a)',end=999) string
      if (string(2:19).ne.'Transition density') goto 40

      ilbl=index(string,'states')+6
      k=len_trim(string)

      n=0
      do i=ilbl,k
         if (string(i:i).eq.'.') then
            n=n+1
            if (n.eq.1) then
               read(string(i-1:i-1),'(i1)') s
            else
               read(string(i-1:i-1),'(i1)') s1
            endif
         endif
      enddo

50    continue
      read(unit,'(a)',end=999) string
      if (string(3:13).ne.'Total NACME') goto 50
      
      read(string,'(30x,F14.8)') tautmp

!-----------------------------------------------------------------------
! Assign to the correct array
!-----------------------------------------------------------------------
      if (ldpos) then
         postau(itmp,itmp2,s,s1)=tautmp
         postau(itmp,itmp2,s1,s)=-postau(itmp,itmp2,s,s1)
      else if (ldneg) then
         negtau(itmp,itmp2,s,s1)=tautmp
         negtau(itmp,itmp2,s1,s)=-negtau(itmp,itmp2,s,s1)
      else
         tau(itmp2,s,s1)=tautmp
         tau(itmp2,s1,s)=-tau(itmp2,s,s1)
      endif
      
      return
      
999   continue
      write(6,'(/,a,/)') "NACT not found."
      STOP      

    end subroutine nactmolpro_ddr

!#######################################################################

    subroutine outf

      implicit none
      
      integer(long) :: nblck,i,j,s,s1,n,j1,j2,j3,stmp,stmp1,itmp,itmp1
      real(dop)     :: ftmp,ftmp1

      ftmp=0.0d0
      do s=1,nsta-1
         do s1=s+1,nsta
            do i=1,ncoo-1
               do j=i+1,ncoo
                  ftmp1=abs(ftensor(i,j,s,s1))
                  ftmp1=sqrt(ftmp1)
                  if (ftmp1.gt.ftmp) then
                     ftmp=ftmp1
                     stmp=s
                     stmp1=s1
                     itmp=i
                     itmp1=j
                  endif
               enddo
            enddo
         enddo
      enddo

      write(6,'(/,3x,a33,x,E13.5,x,a4/)') &
           'Largest (|F_i,j,s,s1|)^0.5 value:',ftmp,'a.u.'
      write(6,'(3x,a3,4(2x,a4,i2),/)') 'For','s =',stmp,'s1 =',stmp1,'i =',itmp,'j =',itmp1

      write(iout,'(/,5(a,/))') &
           '# In the following, atomic units are used.',&
           '# The square roots of the absolute values of the elements of',&
           '# the F-tensor give a measure of the order of magnitude of',&
           '# the NACTs coupling the P-space and Q-space states. If they',&
           '# are large, you do not have a Hilbert subspace.'
     
      write(iout,'(/,71(a1))') ('*',j=1,71)
      write(iout,'(a1,23x,a,22x,a1)') '*','Elements of the F-tensor','*'
      write(iout,'(71(a1),/)') ('*',j=1,71)
      nblck=((ncoo-1)/4)+1
      do s=1,nsta-1
         do s1=s+1,nsta
            if (lssel) then
               if (issel(s).eq.0.or.issel(s1).eq.0) cycle
            endif
            write(iout,'(70(a1))') ('-',j=1,70)
            write(iout,10) 'States',mod(s,10),'and',mod(s1,10)
            write(iout,'(70(a1))') ('-',j=1,70)
            do n=1,nblck
               j1=(n-1)*4+1
               j2=min(j1+3,ncoo)
               write(iout,'(11x,4(4x,i2,9x))') (j,j=j1,j2)
               do i=j1,ncoo
                  j3=min(i,j2)
                  write(iout,20) i,(ftensor(i,j,s,s1),j=j1,j3)
               enddo
            enddo
            write(iout,*)
         enddo
      enddo
      
      write(iout,'(2/,71(a1))') ('*',j=1,71)
      write(iout,'(2a)') &
           '* Square roots of the absolute values of the ',&
      'elements of the F-tensor *'
      write(iout,'(71(a1),/)') ('*',j=1,71)
      do s=1,nsta-1
         do s1=s+1,nsta
            if (lssel) then
               if (issel(s).eq.0.or.issel(s1).eq.0) cycle
            endif
            write(iout,'(70(a1))') ('-',j=1,70)
            write(iout,10) 'States',mod(s,10),'and',mod(s1,10)
            write(iout,'(70(a1))') ('-',j=1,70)
            do n=1,nblck
               j1=(n-1)*4+1
               j2=min(j1+3,ncoo)
               write(iout,'(11x,4(4x,i2,9x))') (j,j=j1,j2)
               do i=j1,ncoo
                  j3=min(i,j2)
                  write(iout,20) i,(sqrt(abs(ftensor(i,j,s,s1))),j=j1,j3)
               enddo
            enddo
            write(iout,*)
         enddo
      enddo

10    format (27x,a6,1x,i2,1x,a3,1x,i2)
20    format(i3,7x,4(E13.5,2x))

      return
    end subroutine outf

!#######################################################################

  end program xcurl
