!***********************************************************************
!
!                     ANALYSE ROUTINE ADPOP
!
!     This routine calculates the population (norm of WF) on the  
!     adiabatic surfaces.
!     It reads the pes file and the psi file. At each grid point the
!     transformation matrix will be calculated and the diabatic WF 
!     will be transformed into the adiabatic one. Then the norm will
!     be calculated.
!     Also one or two dimensional densities can be calculated.
!
! MRB Feb 05
!
!***********************************************************************
      module adpopmod


      use randommod

use decimal, only: long,dop,sip
    
! Density array declarations and sizes
      integer(long), save                  :: dofcount,dof2dimcount
      integer(long), save                  :: zeigdim,zeigidim,zeigweight
      integer(long), save                  :: den1dim,den2dim
      integer(long), dimension(:), allocatable, save   :: dofnumber,zeigden
      integer(long), dimension(:), allocatable, save   :: zeig2dimden
      integer(long), dimension(:,:), allocatable, save :: dof2dimnumber
! Options
      integer(long), save                  :: mcran = 1, &
                                              maxit = 50000
      real(dop), save                      :: qtol=1.0d-6, &
                                              tcut = -999.0, &
                                              mctol=1.0d-3, &
                                              sadgap=0.5/27.2114
      logical(kind=4), save                :: lquick = .true.

      integer(long), save :: ompthread=1

!-----------------------------------------------------------------------
! flags for integration method
! ladpint : Use full grid (exact) integration
! lcdvrint : Use CDVR grid integration
! ltddvrint : Use TDDVR grid integration
! lmcint  : Monte-Carlo integration
! lmcint1 : Monte-Carlo integration using 1D density as sampling weight
! lsaddle : Saddle point approximation (for vMCG)
! lsaddle0 : RHS Saddle point approximation (for vMCG)
!
! lgwpint : for GWP only calculations, integrate GWP matrix elements
!           rather than expectation value.
!-----------------------------------------------------------------------
      logical(kind=4), save                :: lmcint = .false., &
                                              lmcint1 = .false., &
                                              lsaddle = .false.,&
                                              lsaddle0 = .false.,&
                                              lcdvrint = .false.,&
                                              ltddvrint = .false.,&
                                              lcdvr1d = .false.,&
                                              ladpint = .true.

      logical(kind=4), save                :: lgwpint = .false.

! Run as a batch job (no output to screen)
      logical(kind=4), save                :: lbatch = .false.

! Read restart file rather than psi
      logical(kind=4), save                :: lrst = .false.

! Calculate adiabatic densities
      logical(kind=4), save                :: lden = .false.

! Density files
      integer(long), dimension(:), allocatable, save :: outden,out2dimden

      end module adpopmod

!#######################################################################

      program adpop

use decimal, only: long,dop,sip
use maxv, only: maxdim,maxsta
use adpopmod
use versions
use logdat
use global
use aglobal
use dglobal
use globmemmod
use aglobal
use dvrdatmod
use griddatmod
use operdef
use hpsimod
use psidef
use channels
use datenmod
use compdatmod
use memdimmod
use db_sql
use dirdyn
use gapmod

use rdopermod, only: rdoper
use rddvrmod, only: rddvr
use rdfilesmod, only: rdpsi
use gwplib, only: gwp_param, gwp_overlap, gwptophi1
use densitymod, only: density
use dichtoutmod, only: dichtout,gwpgpopwei
use datenmod
use runpropmod
use mfieldsmod, only: getcdvr
use plotutils, only: find_nearest_gpoint
use compdat, only: savecpsi,getcpsi

use op1lib, only: cpvxz, zerovxd, cpvxi, zerovxi, cpmxz, cpqxz, &
                  unitqxz, cuttxz1, sum2vxz
use mmlib, only: mmxxzz,mmvxzz,qqxxzz
use xvlib, only: qvaxzz1,qvxxzz,xvxxdzo,mvtxzz,qvtxzz1,vvaxzz,mvxxzz
use mtlib, only: qtxxzz,qtaxzz,mtaxzz,mtxxzz,mtvxzz,mtaxzz
use rmlib, only: rmhxxxzz,rmqxxxzz
use iorst, only: rstinfo
use iofile, only: rdpsiinfo
use zeigauswmod, only: zeigausw
use memhopsmod, only: memhops
use potevalmod, only: calcdiab
use rdopermod, only: operinfo
use rddvrmod, only: dvrinfo, dvrdata
use dd_db, only: gpdbupdate,getdbnrec
use randommod, only: randuniform,sobseq

use psiutil, only: adecoder,f2mdecoder,vendecoder,decoder,decoderm,adecoder_nomeb

use module_interface

#ifdef OMP
use omp_lib
#endif

      implicit none

      integer(long)                        :: chkdvr,chkgrd,chkdat,chkpsi,&
                                              chkprp,ilbl,f,n,m,s,zeig
      logical(kind=4)                      :: lerr,linwf,lexist
      character(len=c5)                    :: filename
      character(len=c5)                    :: dbname1

!-----------------------------------------------------------------------
! Initialize variables
!-----------------------------------------------------------------------
      call default
      call adefault

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare files to be read
! set flag to .true. if file is to be read
!
!     lrddvr    DVR file. Set in adpopinfo.
!     lrdoper   OPER file.
!-----------------------------------------------------------------------
      lrddvr=.true.
      lrdoper=.true.
      
!-----------------------------------------------------------------------
! read input options
!-----------------------------------------------------------------------
      call adpopinput(n)
      
!-----------------------------------------------------------------------
! get memory sizes 
!-----------------------------------------------------------------------
      filename=operfile
      ilbl=index(filename,' ')-1
      open(ioper,file=filename,form='unformatted',status='old',err=900)
      call rdmemdim(ioper)
      close(ioper)

      call rdmemdim(ipsi)

!-----------------------------------------------------------------------
! Allocate mememory 
!-----------------------------------------------------------------------
      allocmemory=0
!      if (ldd.or.lddmctdh) ldbsave = .true.
      ldbsave = .true.
! USe Local DB if lquick is set
      if (lquick) then
         ldbsmall = .true.
      else
         ldbsmall = .false.
      endif
      call alloc_dvrdat
      call alloc_grddat
      call alloc_operdef
      call alloc_dirdyn
      call alloc_psidef
      if (ldynspf) call alloc_dynwf
      call alloc_compdat
      call alloc_hpsimod
      call alloc_dercp
      call alloc_gapkernel

      allocate(dofnumber(maxdim))
      allocate(zeigden(maxdim)) 
      allocate(zeig2dimden(maxdim*2)) 
      allocate(dof2dimnumber(maxdim*2,2)) 
      allocate(outden(maxdim)) 
      allocate(out2dimden(maxdim*2)) 

!-----------------------------------------------------------------------
! Reading the DVR file
!-----------------------------------------------------------------------
      filename=dname
      open(idvr,file=filename,form='unformatted',status='old',err=900)
      ilbl=index(filename,' ')-1
      if (.not. lbatch) &
         write(6,'(2a)') ' The dvr file is : ',filename(1:ilbl)

      chkdvr=1
      call dvrinfo(lerr,chkdvr)
      call chksyserr(filename,lerr)
      if(lerr) goto 999
      close(idvr)

!-----------------------------------------------------------------------
!  Reading the oper file.
!-----------------------------------------------------------------------
      filename=operfile
      open(ioper,file=filename,form='unformatted',status='old',err=900)
      ilbl=index(filename,' ')-1
      if (.not. lbatch) &
         write(6,'(2a)') ' The operator file is : ',filename(1:ilbl)

      chkdvr=2
      chkgrd=1
      call operinfo(lerr,chkdvr,chkgrd)
      if (lerr) go to 999
      close(ioper)

      if (ldd .or. lddmctdh) then
         lexist=.false.
         inquire(file=ddname(1:ddlaenge)//'/geo.dba',exist=lexist)
         if (lexist) then
            write(6,'(2a)') ' The DD DB is in ascii format.&
      &        First use convdbtosql'
            stop
         endif
         inquire(file=ddname(1:ddlaenge)//'/database.sql',exist=lexist)
         if (lexist) then
            if (.not. lbatch) &
               write(6,'(2a)') ' The DD DB is in : ',ddname(1:ddlaenge)&
               //'/database.sql'
         else
            write(6,'(2a)') ' Error: The DD DB is not found'
            write(6,'(2a)') ' DB asked for is : ',ddname(1:ddlaenge)&
               //'/database.sql'
            write(6,'(a)') ' Use -DB keyord to specify path?'
            stop
         endif

         call abspath(ddname,ddlaenge)
         dbname1 = trim(ddname) // '/database.sql'

         if (.not.the_db%db_open) then
            !Needed for NEWSQL
            !call makeflagarray(flagarray)
            !call makeinfoarray(nddstate, infoarray)
            call dbopen(dbname1,the_db,sqlvers,quanticsver,infoarray,&
                 flagarray,0)  
         endif

! Get size of DB and allocate memory
         call getdbnrec(dbnrec)
         dbmemdim = dbnrec
         call alloc_dddb
      endif

!-----------------------------------------------------------------------
! read the densities to be calculated
!-----------------------------------------------------------------------
      call adpopdeninp(n)
      
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Declare resources required (other than those needed to read a file)
!
! ncomp  : no. of comparison data sets to be used.
!          See auswutil.F for details of use of comparison sets
! lpsi   : wavefunction
! lpsi1  : second wavefunction
! lpsigrd : wavefunction to be used in grid representation (rather
!           than spf)
! ldmat  : density matrices for psi
! ldmat1 : density matrices for psi1
! lort   : grid coordinates
! lgpop  : grid populations
! ltrafo : FBR/DVR trafo matrices
!-----------------------------------------------------------------------
      ncomp=0
      lpsi=.true.
      lpsi1=.false.
      lpsigrd=.false.
      ldmat1=.false.
      lort=.true.
      ltrafo=.true.
      if (lquick .or. lmcint) then
         ldmat=.true.
         lgpop=.true.
      else
         ldmat=.false.
         lgpop=.false.
      endif
      
!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! open input and output data files
!
! use standard channels if possible
!  irho : gridpop file
!  ipsi : psi file
!  iaut : auto file
!  ichk : check file
!  iaus : output file
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Read system information from property file(s)
! chkflags = 0   read information and ignore
!            1   read information and store
!            2   read information and check
! First read file version number
!-----------------------------------------------------------------------
      chkdvr=2
      chkgrd=2
      chkpsi=1
      chkdat=1
      chkprp=0      

      if(lrst) then
         irst=ipsi
         call rstinfo(linwf,lerr,chkdvr,chkgrd,chkpsi,chkprp)
      else
         read(ipsi,err=999) filever(ipsi)
         call rdpsiinfo(ipsi,chkdvr,chkgrd,chkpsi,chkdat)
      endif
      if(psitype.eq.5) then
         routine='adpop'
         message='Sorry, adpop cannot work with ML-wavefunctions.'
         call errormsg
      endif
      if (ltddvrint .and. lmult) then
         routine='adpop'
         message='Sorry, TDDVR adpop cannot work with Multi-set wavefunctions.'
         call errormsg
      endif
      call chksyserr(filein,lerr)
      if (lerr) go to 999

!-----------------------------------------------------------------------
! assign pointers and sizes for required arrays
! assign memory arrays
!-----------------------------------------------------------------------
      call zeigausw
      call memausw

      if(lrddvr) then
         zeigweight = mrdim
         mrdim = mrdim + maxgdim*ndof
      else
         zeigweight = 1
      endif
      
      if (ltddvrint) then
         zeig=1
         do s=1,nstate
            do m=1,nmode
               if (m .eq. meb) cycle
               if(idmode(m).eq.0) then
                  c3mat(m,s)=zeig
                  zeig=zeig+dim(m,s)*nspfdof(m)
               else
                  c3mat(m,s)=c3mat(idmode(m),s)
               endif
            enddo
         enddo
         c3matdim=zeig-1
         mrcdvr3=mrdim
         mrdim=mrcdvr3+c3matdim + 1
      else
         c3matdim=1
      endif

      if(lden) then
         zeigdim = mrdim
         do f=1,dofcount
            mrdim = mrdim + gdim(dofnumber(f))*nddstate
         enddo 
         den1dim=mrdim-zeigdim
      
         do f=1,dof2dimcount
            mrdim = mrdim + gdim(dof2dimnumber(f,1)) &
                            *gdim(dof2dimnumber(f,2))*nddstate
         enddo
         den2dim=mrdim-zeigdim-den1dim

         if (lmcint .or. lmcint1) then
            zeigidim = midim
            midim = midim+den1dim+den2dim
         else
            zeigidim = 1
         endif

      else
         zeigdim = 1
         zeigidim = 1
      endif
      
!-----------------------------------------------------------------------
! assign work arrays
!-----------------------------------------------------------------------
      mcworkc = mcdim
      mcdim = mcworkc+workcdim

      miworki = midim
      midim   = miworki+workidim

      mlworkl = mldim
      mldim   = mlworkl+workldim

      msworks = msdim
      msdim   = msworks+worksdim

      mfworkf = mfdim
      mfdim   = mfworkf+workfdim

      mrworkr = mrdim

!-----------------------------------------------------------------------
! assign memory for operator
!-----------------------------------------------------------------------
      call memhops

!-----------------------------------------------------------------------
! call routine
!-----------------------------------------------------------------------
      call wrversion(ilog)
      ilbl = 16*mcdim+8*mrdim+4*midim+4*mldim+8*msdim+4*mfdim
      write(ilog,'( 6(a,i10,/),a,i13,a,f10.2,a)') &
       ' complex*16 : ', mcdim,&
       ' real*8     : ', mrdim,&
       ' integer    : ', midim,&
       ' logical    : ', mldim,&
       ' complex*8  : ', msdim,&
       ' real*4     : ', mfdim,&
       ' Total   : ', ilbl, ' Bytes ,', dble(ilbl)/1048576.0_dop,' MB'
#ifdef OMP
      write(ilog,'(/,a,i5)') ' No. of threads : ',ompthread
#endif
      call myflush(ilog)

if(allocated(mc))deallocate(mc)
if(allocated(mr))deallocate(mr)
if(allocated(mi))deallocate(mi)
if(allocated(ml))deallocate(ml)
if(allocated(ms))deallocate(ms)
if(allocated(mf))deallocate(mf)
if(allocated(gfgwp))deallocate(gfgwp)
if(allocated(aigwp))deallocate(aigwp)
if(allocated(zgwp))deallocate(zgwp)
if(allocated(gs2))deallocate(gs2)
if(allocated(gsinv))deallocate(gsinv)
if(allocated(gsvek))deallocate(gsvek)
if(allocated(dicht1))deallocate(dicht1)
if(allocated(dicht2))deallocate(dicht2)
if(allocated(dicht3))deallocate(dicht3)
if(allocated(dicht4))deallocate(dicht4)
if(allocated(agmat))deallocate(agmat)

allocate(mc(mcdim))
allocate(mr(mrdim))
allocate(mi(midim))
allocate(ml(mldim))
allocate(ms(msdim))
allocate(mf(mfdim))
allocate(gfgwp(gfgwpdim))
allocate(aigwp(aigwpdim))
allocate(zgwp(zgwpdim))
allocate(gs2(gs2matdim))
allocate(gsinv(gsmatdim))
allocate(gsvek(gsmatdim))
allocate(dicht1(dmatdim))
allocate(dicht2(dmatdim))
allocate(dicht3(d3matdim))
allocate(dicht4(dmatdim))
allocate(agmat(agmatdim))

      call population(mr(mrhops),mc(mcpsi),ms(mspsi),&
           mi(mijindx),mr(mrort),mr(mrtrafo),mr(zeigdim),mr(zeigweight),&
           dicht3,dicht1,dicht2,dicht4,&
           mr(mrgpop1),mi(zeigidim),agmat,gs2,&
           gsinv,gfgwp,aigwp,zgwp,mr(mrqcent),&
           mc(mcadgwp),mr(mrgsval),gsvek,mi(migwpdep))

      deallocate(mc,mr,mi,ml,ms,mf,gfgwp,aigwp,zgwp,gs2,gsinv,gsvek,&
                 dicht1,dicht2,dicht3,dicht4,agmat)
      deallocate(dofnumber)
      deallocate(zeigden) 
      deallocate(zeig2dimden) 
      deallocate(dof2dimnumber) 
      deallocate(outden) 
      deallocate(out2dimden) 

      call dealloc_dvrdat
      call dealloc_grddat
      call dealloc_operdef
      call dealloc_dirdyn
      call dealloc_dddb
      call dealloc_psidef
      call dealloc_dynwf
      call dealloc_compdat
      call dealloc_dercp
      call dealloc_gapkernel

      stop
  900 continue
      routine='adpop'
      ilbl=index(filename,' ')-1
      message = 'Cannot open file: '//filename(1:ilbl)
      call errormsg

  999 continue
      routine='adpop'
      message= '1: Error reading :'//filein(1:laein)
      call errormsg
      stop

      contains

!#######################################################################

      subroutine population(hops,psi,spsi,jindx,ort,trafo,&
                 adiabden,weight,dicht3,dicht1,dicht2,dicht4,gpop1,&
                 adiabnsum,agmat,gs2,gsinv,gfgwp,aigwp,zgwp,&
                 qcenter,adgwp,gsval,gsvek,gwpdep)

      implicit none

      integer(long)       :: f,m,i,chkdvr,chkgrd,totgrid,s,&
                             nham,g,h,platzspar1,n,platzspar2,t1,t2,kmin2,dk,&
                             kmin,kmax,dof,gm,dj,p1,j0,whereiam,ggdim,gdim1,&
                             sphtyp,jmax,gj,mmax,mincr,mmin,wfnum,mmmin,&
                             mmmax,mmmin2,dm,p2,ctotgrd,itmp

      integer(long), dimension(adim), intent(out)           :: jindx
      integer(long), dimension(1)                           :: idum,idum1,idum2,&
                                                               idum3
      integer(long), dimension(maxdim)                      :: subdimo,gdimo,&
                                                               gqi,gqf,&
                                                               subdimn,gdimn,&
                                                               vgdimo,vgdimn,&
                                                               ngdimo,ngdimn
      integer(long), dimension(maxdim)                      :: subdimgwp
      integer(long),dimension(den1dim+den2dim),intent(inout):: adiabnsum
      real(dop)                                             :: tcpu,tsys,ttot,&
                                                               atotgrd,time
      real(dop), dimension(1)                               :: rdum,rdum1,rdum2
      real(dop), dimension(maxsta)                          :: diapop,adiabpop
      complex(dop)                                          :: diapop12,adiabpop12
      integer(long), dimension(maxdim)                      :: imgpop1
      real(dop), dimension(maxdim)                          :: maxgpop1
      real(dop), dimension(qcentdim), intent(out)           :: qcenter
      real(dop), dimension(hopsdim), intent(out)            :: hops
      real(dop), dimension(ortdim), intent(out)             :: ort
      real(dop), dimension(dvrdim), intent(out)             :: trafo
      real(dop), dimension(d3matdim), intent(out)           :: dicht3
      real(dop), dimension(den1dim+den2dim), intent(out)    :: adiabden
      real(dop), dimension(ortdim,nddstate*npacktot), intent(out)    :: gpop1
      real(dop), dimension(maxgdim,ndof), intent(out)       :: weight
      complex(sip), dimension(spsidim), intent(out)         :: spsi
      complex(dop), dimension(1)                            :: zdum,zdum1,&
                                                               zdum2,zdum3
      complex(dop), dimension(dgldim), intent(out)          :: psi
      complex(dop), dimension(dmatdim), intent(inout)       :: dicht4
      complex(dop), dimension(dmatdim), intent(out)         :: dicht1,dicht2
      complex(dop), dimension(agmatdim), intent(out)        :: agmat
      complex(dop), dimension(gsmatdim), intent(out)        :: gsinv,adgwp
      complex(dop), dimension(gs2matdim), intent(out)       :: gs2
      complex(dop), dimension(gfgwpdim), intent(out)        :: gfgwp
      complex(dop), dimension(aigwpdim), intent(out)        :: aigwp
      complex(dop), dimension(zgwpdim), intent(out)         :: zgwp
      complex(dop), dimension(gsmatdim), intent(out)        :: gsvek
      real(dop), dimension(gwpinfodim), intent(out)         :: gsval
      integer(long), dimension(gwpinfodim),intent(out)      :: gwpdep
      character(len=80)                                     :: buffer
      real(dop), external                                   :: fcputime,fsystime

      nham = 1
      wfnum = 0

!-----------------------------------------------------------------------
! for GWP integrals use symmetric LHA as default and energy only
!-----------------------------------------------------------------------
      if (lgwp) then
         gwpintm(:)=2
         gwpintordm(:) = 0
         gwpcxm(:) = 0
      endif

!-----------------------------------------------------------------------
! set regularisation parameter needed in density routine
!-----------------------------------------------------------------------
      epsmat=1.0d-8

!-----------------------------------------------------------------------
! read DVR file
!-----------------------------------------------------------------------      
      if(lrddvr) then
         chkdvr=1
         dvrdata(1)=.true.
         dvrdata(2)=.true.
         call rddvr(ort,trafo,rdum,rdum1,zdum,zdum1,idum,zdum2,zdum3,&
              idum1,idum2,rdum2,chkdvr)      
      endif

!-----------------------------------------------------------------------
! read operator file
!-----------------------------------------------------------------------
      rewind(ioper)
      chkdvr=1
      chkgrd=1
      call rdoper(hops,chkdvr,chkgrd)
      
!-----------------------------------------------------------------------
! Make operator analytical for CDVR or TDDVR integration
!-----------------------------------------------------------------------
      if (ltddvrint) then
         do h=1,htmdim
            if (htmdof(h) .eq. feb) cycle
            if (htmmode(h) .eq. meb) cycle
            if (htmtype(h) .eq. 1) htmtype(h)=103
            if (htmtype(h) .eq. 2) htmtype(h)=101
         enddo
      endif

!-----------------------------------------------------------------------
! For GWPs only evaluate PES at a point - no LHA
!-----------------------------------------------------------------------
      if (lgwp) gwpintordm(:)=0
      
!-----------------------------------------------------------------------
! change subdim for GWPs to be grid dimension rather than
! number of parameters 
!-----------------------------------------------------------------------
      do m = 1, nmode
         if (gwpm(m)) then
            subdimgwp(m) = subdim(m)
            subdim(m) = 1
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               subdim(m) = subdim(m)*gdim(f)
            enddo
         endif
      enddo

!-----------------------------------------------------------------------
! Log method
!-----------------------------------------------------------------------
      if (ladpint) then
         write(ilog,'(a)') 'Full Integration'
      else if (ltddvrint .and. lcdvrint) then
         write(ilog,'(a)') 'CDVR Integration'
      else if (ltddvrint) then
         write(ilog,'(a)') 'TDDVR Integration'
      else if (lmcint .or. lmcint1) then
         write(ilog,'(a)') 'Monte-Carlo Integration '
         if (lgwpint) then
            write(ilog,'(a)') 'Integrating matrix elements using GWP distribution'
! get no. of integrals
            if (leb) s = adim/nddstate
            i = (s*s - s)/2
            i = i+s
            write(ilog,'(a,i7)') 'No. of integrals : ',i
            write(ilog,'(a,i7)') 'Max iterations   : ',maxit
         else if (mcran .eq. 1) then
            write(ilog,'(a)') 'Using Sobol Sequence '
         endif
         write(ilog,'(a,e20.12)') 'Error :',mctol
      endif
      if (lquick) then
         write(ilog,'(a)') 'Quick Algorithm '
         write(ilog,'(a,e20.12)') 'Tolerance :',qtol
      endif

!-----------------------------------------------------------------------
! determine total grid dimension. 
! If single-set ignore the electronic coordinate.
!-----------------------------------------------------------------------
      totgrid = 1
      atotgrd = 1.0_dop
            
      do m = 1, nmode
         if (m .eq. meb) cycle
         if (gwpm(m)) then
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               atotgrd = atotgrd*gdim(f)
            enddo
         else
            atotgrd = atotgrd*subdim(m)
         endif 
      enddo
! If vector longer than 2GB cannot do full integration
      if (atotgrd .gt. 2147483647.0_dop) then
         if (ladpint) then
            write(ilog,'(a/,a/,a)')  'The primitive grid is too large!',&
              'Integer overflow. More than 2,147,483,647 grid points.',&
              'Use a coarser grid. '
            write(ilog,'(a,f18.0,es14.4)') ' Total griddim =',atotgrd,atotgrd
            stop
         else
            totgrid = -999
         endif
      else
         do m = 1, nmode
            if (m .eq. meb) cycle
            if (gwpm(m)) then
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  totgrid = totgrid*gdim(f)
               enddo
            else
               totgrid = totgrid*subdim(m)
            endif 
         enddo
      endif

      write(ilog,'(a,i11,e14.4)') ' Total griddim = ',totgrid,atotgrd

      if (ltddvrint) then
         write(ilog,'(a,i11)')  ' TDDVR griddim = ',block(1)/dim(meb,1)
         if (lcdvrint) then
            atotgrd = 1.0_dop
            ctotgrd = 1

            do m = 1, nmode
               if (m .eq. meb) cycle
               itmp=block(1)/(dim(m,1)*dim(meb,1))

               atotgrd = max(atotgrd,dble(itmp*subdim(m)))
               ctotgrd = max(ctotgrd,itmp*subdim(m))
            enddo

            if (atotgrd .gt. 2147483647.0_dop) then
               write(ilog,'(a)') 'CDVR Grid too large .....'
               write(ilog,'(a,f18.0)')     ' CDVR griddim  = ',atotgrd
            else
               atotgrd = atotgrd/(1024*1024)
               write(ilog,'(a,i11,4x,f10.3,a3)') ' CDVR griddim  = ',ctotgrd,atotgrd,' MB'
            endif

         endif
      endif

      write(iaus,'(a,i2,a,a,i2,a)') '# The first ',nddstate,' columns',&
           ' are the adiabatic populations. The second ',nddstate,&
           ' columns are the diabatic populations.'
      write(iaus,'(/a,4x,99(6x,i2,6x))') '#    Time[fs]',&
                            (s,s=1,nddstate), (s,s=1,nddstate)
      if(nstate.eq.2 .and. .not.lmcint) then
        write(iaus,'(2a)') '# The last 4 colums show the complex ', &
          'off-diagonal terms of adiabatic and diabatic el. density.'
        write(iaus,'(/a,4x,4(6x,i2,6x),17x,a,23x,a)') &
          '#    Time[fs]',(s,s=1,2), (s,s=1,2),'adiabatic','diabatic'
      else
         write(iaus,'(/a,4x,99(6x,i2,6x))') '#    Time[fs]', &
                            (s,s=1,nddstate), (s,s=1,nddstate)
      endif
      call myflush(iaus)

      if (lden) then
         do f=1,dofcount
            write(outden(f),'(a)') &
               '# 1D Density. First Nstate columns include the DVR weights.'
         enddo
         do f=1,dof2dimcount
            write(out2dimden(f),'(a)') &
               '# 2D Density. First Nstate columns include the DVR weights.'
         enddo
      endif

      if (lmcint .or. lmcint1) then
         if (lgwpint) then
            write(ilog,'(a)') ' Time      total /  av per integral  '
         else
            write(ilog,'(a)') ' Time      npnts /  grid  '
         endif
      else if (lquick) then
         write(ilog,'(a)') ' Time     included / grid  '
      endif

!-----------------------------------------------------------------------
! save grid sizes
!-----------------------------------------------------------------------
      if (ltddvrint .or. lquick) then
         do m=1,nmode
            subdimo(m) = subdim(m)
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               gdimo(f)=gdim(f)
               vgdimo(f)=vgdim(f)
               ngdimo(f)=ngdim(f)
            enddo
         enddo
      endif
      
!-----------------------------------------------------------------------
! calculate time step for data
!-----------------------------------------------------------------------
      dt     = out2/fs
      time   = tinit/fs
 
!-----------------------------------------------------------------------
! find point for tcut. 
! ncut marks wanted data set, nskip marks data set before
!-----------------------------------------------------------------------
      if (tcut .gt. 0) then
         ncut=0
    5    continue
         if (abs(tcut-time) .lt. 1.0d-4) then
            ncut = ncut + 1
            nskip = ncut -1
         else if (tcut .lt. time) then
            write(6,*) ' '
            write(6,*) 'ERROR: Time specified not found', tcut
            write(6,*) ' '
            stop
         else 
            time = time+dt
            ncut = ncut + 1
            go to 5
         endif
            
         time   = tinit/fs
      endif

!-----------------------------------------------------------------------
! skip first nskip WFs
!-----------------------------------------------------------------------
      if (.not.lrst) then
         do i=1,nskip
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) goto 77
            time = time + dt
            wfnum=wfnum+1
         enddo
      endif
      
!-----------------------------------------------------------------------
! Determine DVR weights
!-----------------------------------------------------------------------
      if(lrddvr) then
        do f=1,ndof
           dof=f
           ggdim = gdim(dof)
           gdim1 = gdim(dof)
           if(basis(dof) .eq. 12 .or. basis(dof).eq.20 .or. basis(dof)&
                         .eq. 21) ggdim = ipbaspar(3,dof)

!-----------------------------------------------------------------------
! Special handling for sphFBR
!-----------------------------------------------------------------------
           if( basis(f) .eq. 6 .or. basis(f) .eq. -1 ) then
              if(basis(f) .eq. -1 ) dof = f-1
              sphtyp = ipbaspar(1,dof)
              jmax   = ipbaspar(2,dof)
              gj     = ipbaspar(4,dof)
              mmax   = ipbaspar(1,dof+1)
              mincr  = ipbaspar(2,dof+1)
              mmin   = ipbaspar(3,dof+1)
              gm = (mmax-mmin)/mincr + 1
              dj = 1
              if( sphtyp.eq.2  .or. sphtyp.eq.3 ) dj = 2 
              j0 = mod(jmax,dj)
              if(basis(f) .eq. -1 ) then
                 gdim1 = gm
                 ggdim = gm
                 do g = 1, gdim1
                    ort(zort(dof)-1+g) = mmin + (g-1)*mincr
                 end do
              else
                 gdim1 = gj
                 ggdim = gj
                 do g = 1, gdim1
                    ort(zort(dof)-1+g) = j0 + (g-1)*dj
                 end do
              end if
           end if

!...Determine the appropriate 'column' of trafo(gdim,gdim,kdim) for KLeg/PLeg
!...and trafo(gdim,gdim,kdim,mmdim) for Wigner
           p1=0
           if( basis(f) .eq. 7 .or. basis(f) .eq. 20 ) then ! KLeg
              kmin=ipbaspar(1,dof+1)
              kmax=ipbaspar(2,dof+1)
              kmin2=ipbaspar(3,dof+1)
              p1=(min(kmax,kmin2)-kmin)
              dk=max(ipbaspar(5,dof+1),1)
              p1=(p1+dk-1)/dk
              p1=p1*ggdim**2
           elseif( basis(dof) .eq. 9 .or. basis(dof) .eq.21 ) then           ! PLeg

              kmin=ipbaspar(1,dof+1)
              kmax=ipbaspar(2,dof+1)
              kmin2=ipbaspar(4,dof+1)
              p1=(min(kmax,kmin2)-kmin)*ggdim**2
           elseif( basis(dof) .eq. 33 ) then        ! Wigner

              kmin=ipbaspar(2,dof)
              kmax=ipbaspar(3,dof)
              kmin2=ipbaspar(5,dof)
              p2=min(kmax,kmin2)-kmin
              if (basis(dof+1).eq.8) then
                 dk=ipbaspar(5,dof+1)
                 p2=(p2+dk-1)/dk
              endif
              p2=p2*gdim(dof)**2
              mmmin=ipbaspar(6,dof)
              mmmax=ipbaspar(7,dof)
              mmmin2=ipbaspar(9,dof)
              p1=min(mmmax,mmmin2)-mmmin
              if (basis(dof+2).eq.8) then
                 dm=ipbaspar(5,dof+2)
                 p1=(p1+dm-1)/dm
              endif
              p1=p1*ipbaspar(4,dof)*gdim(dof)**2 + p2
           
           endif

           call dvrweights(trafo(zdvr(dof)+p1),ort(zort(dof)),gdim1,&
                   ggdim,weight(1,f),basis(dof),rpbaspar(1,dof),&
                   ipbaspar(1,dof),ipbaspar(1,dof+1),.true.,.false.)
        enddo
      endif

!-----------------------------------------------------------------------
! loop over all time steps
!-----------------------------------------------------------------------
   10 continue
   
!-----------------------------------------------------------------------
! read WF
!----------------------------------------------------------------------- 
      if (lrst) then
         irst=ipsi
         read (irst,err=999) time
         time = time/fs
         read (irst,err=999) (psi(g),g=1,dgldim)
         if (lgwp .and. gwptype .ge. 3) then
            read(irst,err=999) (agmat(g),g=1,agmatdim)
         endif

      else
         call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
         if (lend) goto 77
      endif

      cpsitype(1)=0
      if (lgwp) then
         call gwp_param(psi,aGmat,gfgwp,aigwp,zgwp,qcenter,1)
         if (ltddvrint) then
            call gwp_overlap(gs2,gsinv,gsval,gsvek,gfgwp,aigwp,zgwp,3) 
         else
            call gwp_overlap(gs2,gsinv,gsval,gsvek,gfgwp,aigwp,zgwp,3) 
         endif
         if (ldd .and. ldbsmall) call gpdbupdate(qcenter)
      endif

!-----------------------------------------------------------------------
! calculate 1D grid densities. Needed for quick algorithm. 
! and for MC using gpop as the weight
!
! leave population summed in the gpop1(g,1)
!-----------------------------------------------------------------------
      if (lquick) then
         call density(psi,tree,dicht1,dicht2,dicht3,dicht4,jindx,gs2)

! turn off projection of densities onto electronic states for single-set
         lgpel = .false.
         call dichtout(psi,dicht3,gpop1,rdum,trafo,idum,zdum,&
              zdum1,zdum2,idum1,idum2,time,ort,.false.,agmat,&
              jindx,idum3)

! multi-set populations are returned divided into states. Sum up and
! store as s=1
         if (lmult) then
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  do s=2,nstate
                     do g=1,gdim(f)
                        gpop1(zort(f)-1+g,1) = gpop1(zort(f)-1+g,1)&
                                             + gpop1(zort(f)-1+g,s)
                     enddo
                  enddo
               enddo
            enddo
         endif

!-----------------------------------------------------------------------
! Get maximum grid population and expectation values
!-----------------------------------------------------------------------
         do m=1,nmode
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               maxgpop1(f) = 0.0_dop
               imgpop1(f) = 0
               do g=1,gdim(f)
                  if (gpop1(zort(f)-1+g,1) .gt. maxgpop1(f)) then
                     maxgpop1(f) = gpop1(zort(f)-1+g,1)
                     imgpop1(f) = g
                  endif
               enddo
            enddo
         enddo

!-----------------------------------------------------------------------
! get new grid size
!
! For GWPs, mode grid may be very large 
!-----------------------------------------------------------------------
         do m=1,nmode
            subdimn(m) = 1
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f=spfdof(n,m)

               g=1
   20          continue
               if (gpop1(zort(f)-1+g,1) .lt. qtol) then
                  g=g+1
                  if(g.lt.gdim(f)) goto 20
               endif
               gqi(f) = max(1,g-1)

               g=gdim(f)
   22          continue
               if (gpop1(zort(f)-1+g,1) .lt. qtol) then
                  g=g-1
                  if(g.gt.1) goto 22
               endif
               gqf(f) = min(gdim(f),g+1)

               gdimn(f) = gqf(f) - gqi(f) + 1
! If no population gdimn(f) is negative so set grid size to 0
               gdimn(f) = max(gdimn(f),0)
               subdimn(m) = subdimn(m)*gdimn(f)
! check for overflow
               if (subdimn(m) .gt. 999999) subdimn(m) =0
            enddo
            zeig=1
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               vgdimn(f) = zeig
! following is catch for numerical problems when grid too large
               if (vgdimn(f) .lt. 0) vgdimn(f) = 1 
               zeig=zeig*gdimn(f)
            enddo
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               if (vgdimn(f)*gdimn(f) .le. 0) then
                  ngdimn(f) = 0
               else if (gdimn(f) .gt. 0) then
                  ngdimn(f) = subdimn(m)/(gdimn(f)*vgdimn(f))
               else
                  ngdimn(f) = 0
               endif
            enddo
         enddo

         totgrid = 1
         atotgrd = 1.0_dop
         do m = 1, nmode
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               atotgrd = atotgrd*gdimn(f)
            enddo
         enddo
         if (atotgrd .gt. 2147483647.0_dop) then
            if (ladpint) then
               write(6,'(a/,a/,a)') &  
            'The primitive grid is too large!', &
            'Integer overflow. More than 2,147,483,647 grid points.', &
            'Use MC integration or a coarser qtol. '
               write(6,'(a,e11.4)') ' Total griddim =',atotgrd
               stop
            else
               totgrid=-999
            endif
         else
            do m = 1, nmode
               totgrid = totgrid*subdimn(m)
            enddo
         endif

      else
! Using full grid - set initial and final grid points
         do m=1,nmode
            subdimn(m) = subdim(m)
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               gqi(f) = 1
               gqf(f) = gdim(f)
               gdimn(f) = gdim(f)
            enddo
         enddo
      endif
      if (.not. lbatch) &
         write(6,'(a,f10.3,a,i11)') 'time =',time,'  totgrid =',totgrid

!-----------------------------------------------------------------------
! initialize density and population arrays
!-----------------------------------------------------------------------       
      do s=1,nddstate
         adiabpop(s)=0.0_dop
         diapop(s)=0.0_dop
      enddo
      adiabpop12 = (0.d0,0.d0)
      diapop12   = (0.d0,0.d0)
      
      if(lden) then
         call zerovxd(adiabden,den1dim+den2dim)
         if (lmcint .or. lmcint1) call zerovxi(adiabnsum,den1dim+den2dim)
      endif
 
      if (lmcint) then
         if (lgwpint .and. .not. lmult) then
            call mcinteggwp(psi,adiabpop,diapop,qcenter,hops,time)
         else
            call mcintegrate(psi,adiabpop,diapop,adiabden,adiabnsum,&
                 subdimo,gdimo,subdimn,gdimn,totgrid,hops,ort,time,gqi)
         endif
      else if (lmcint1) then
         call mcintegrate1(psi,adiabpop,diapop,gdimn,gpop1,&
              totgrid,hops,ort,time,gqi,maxgpop1,imgpop1)
      else if (lsaddle) then
         call sadintegrate(psi,adiabpop,diapop,qcenter,hops)
      else if (lsaddle0) then
         call sadintegrate0(psi,adiabpop,diapop,qcenter,hops)
      else if (ladpint) then
         call adpintegrate(psi,adiabpop,diapop,adiabden,&
              subdimo,gdimo,subdimn,gdimn,gpop1,&
              totgrid,hops,ort,time,gqi,adiabpop12,diapop12)
      else if (ltddvrint .and. (lgwp .or. lcdvr1d)) then
         call cdvrintegrate(psi,adiabpop,diapop,subdimo,gdimo,&
              subdimn,gdimn,gqi,totgrid,hops,ort,qcenter,time)
      else if (ltddvrint) then
         call cdvrintegrate0(psi,adiabpop,diapop,subdimo,gdimo,&
              subdimn,gdimn,gqi,totgrid,hops,ort,qcenter,time)
      endif
      
!-----------------------------------------------------------------------
! write results (with and without weights)
!-----------------------------------------------------------------------
      if(nstate.eq.2 .and. .not.lmcint) then
         write(iaus,'(6x,f8.3,6x,4(f11.9,3x),4x,2(2f14.10,4x))') &
           time,(adiabpop(s),s=1,2), (diapop(s),s=1,2), &
           adiabpop12,diapop12
      else
         write(iaus,'(6x,f8.3,6x,99(f11.9,3x))')  time, &
           (adiabpop(s),s=1,nddstate), (diapop(s),s=1,nddstate)
      endif

      call myflush(iaus)

      if (lden) then
         do f=1,dofcount
            write(outden(f),'(a12,4x,f8.3)') '# Time step: ', time
            write(outden(f),'(a2,2x,a14,99(i2,12x),99(i2,12x))') '# ', &
                  modelabel(dofnumber(f)),(s,s=1,nddstate),(s,s=1,nddstate)
            do g=1,gdim(dofnumber(f))
               write(outden(f),'(f10.5,4x,99(f10.8,4x),99(f10.8,4x))') &
                  ort(zort(dofnumber(f))+g-1),(adiabden(zeigden(f)+&
                  (s-1)*gdim(dofnumber(f))+g-1),s=1,nddstate),&
                  (adiabden(zeigden(f)+(s-1)*gdim(dofnumber(f))+g-1)&
                  /weight(g,dofnumber(f)),s=1,nddstate)
            enddo
            write(outden(f),'(a)') 'e'
            call myflush(outden(f))
         enddo
 
         do f=1,dof2dimcount
            write(out2dimden(f),'(a12,4x,f8.3)') '# Time step: ', time
            write(out2dimden(f),'(a2,2x,a14,a14,99(i2,12x),99(i2,12x))')&
                 '# ', &
            modelabel(dof2dimnumber(f,1)),modelabel(dof2dimnumber(f,2)),&
            (s,s=1,nddstate), (s,s=1,nddstate)
            do g=1,gdim(dof2dimnumber(f,2))
               do h=1,gdim(dof2dimnumber(f,1))
                 whereiam=zeig2dimden(f)+g+(h-1)*&
                 gdim(dof2dimnumber(f,2))-1+den1dim
                 platzspar1=zort(dof2dimnumber(f,2))
                 platzspar2=gdim(dof2dimnumber(f,1))*&
                           gdim(dof2dimnumber(f,2))
                 write(out2dimden(f),'(f10.5,4x,f10.5,4x,99(f10.8,4x))')&
                 ort(zort(dof2dimnumber(f,1))+h-1),ort(platzspar1+g-1)&
                 ,(adiabden(whereiam+(s-1)*platzspar2),s=1,nddstate)&
                 ,(adiabden(whereiam+(s-1)*platzspar2)/(weight(h,&
                 dof2dimnumber(f,1))*weight(g,dof2dimnumber(f,2))),s=1,&
                 nddstate)
               enddo
               write(out2dimden(f),'(a)') ' '
            enddo
            write(out2dimden(f),'(a)') 'e'
            call myflush(out2dimden(f))
         enddo
      endif

!-----------------------------------------------------------------------
! increment of time step
!-----------------------------------------------------------------------
      time=time + dt
      wfnum=wfnum+1
      
!-----------------------------------------------------------------------
! go out of loop if enough WFs read
!-----------------------------------------------------------------------
      if (ncut .ne. 0) then
         if (wfnum .ge. ncut) goto 77
      endif

!-----------------------------------------------------------------------
! skip WFs
!-----------------------------------------------------------------------
      if (.not.lrst) then
         do i=2,step
            call rdpsi(ipsi,psi,spsi,jindx,agmat,idum,adgwp,gwpdep)
            if (lend) goto 77
            time=time + dt
            wfnum=wfnum+1
         enddo
      endif

!-----------------------------------------------------------------------
! go to the beginning of loop
!-----------------------------------------------------------------------      
      if (.not.lrst) goto 10

 77   continue
      close(ipsi)
      close(ioper)
      close(idvr)

!-----------------------------------------------------------------------
! Get overall timing 
!-----------------------------------------------------------------------
      tcpu = fcputime(0)
      tsys = fsystime(0)
      ttot = tcpu+tsys
      t1   = int(ttot/3600)
      ttot = ttot - 3600*t1
      t2   = int(ttot/60)
      ttot = ttot - 60*t2

      write(ilog,'(//,a,/)') '# ***** Overall Timing *****'
      write(ilog,'(a,1x,f12.2)') '# Total cpu time     [s] : ',tcpu
      write(ilog,'(a,1x,f12.2)') '# Total system time  [s] : ',tsys
      write(ilog, 900) '# Total time     [h:m:s] : ',t1,t2,ttot
      write(ilog,*)

! --- WRITE DATE AND MACHINE  ---
      call fpstartup(buffer)
      write(iaus,*)
      write(iaus,'(a2,a78)') '# ',buffer
      if (.not. lbatch) then
         write(6,*)
         write(6,'(a80)') buffer
         write(6,'(1x,a,/)') name(1:laenge)
      endif
      write(ilog,*)
      write(ilog,'(a80)') buffer
      write(ilog,'(1x,a,/)') name(1:laenge)
      

  900 format( a,i4,' :',i3,' :',f5.1)
      stop
      
  999 continue
      routine='population'
      message= '2: Error reading :'//filein(1:laein)
      call errormsg

      return
      end subroutine population

!-----------------------------------------------------------------------
!***********************************************************************
! i is no. of arguments read, excluding DOFs for densities
!-----------------------------------------------------------------------

      subroutine adpopinput(i)

      implicit none

      integer(long), intent(out)                     :: i
      integer(long), external                        :: myiargc
      integer(long)                                  :: j,iarg,ilbl,ierr
      character(len=80)                              :: buf,host
      character(len=c5)                              :: filename,dvrfile
      logical(kind=4)                                :: lname = .false.

      operfile = ' '
      fileout  = ' '
      filein   = ' '
      name     = './'
      iarg     = myiargc()
      step     = 1
      nskip    = 0
      ncut     = 0
      dvrfile  = ' '

      ddpath = '..'

!-----------------------------------------------------------------------
! if lgwpint = .true. can use routines for GWP integrals (integrate 
! matrix elements separately and sum rather than evaluating expectation
! value directly). Presently only with MC.
!-----------------------------------------------------------------------
      lgwpint = .true.
      do m=1,nmode
         if (m .eq. meb) cycle
         if (.not. gwpm(m)) lgwpint = .false.
      enddo

!***********************************************************************

      i = 0
 100  i=i+1
      call mygetarg(i,buf)
      if( buf(1:1) .eq. '-' ) then
         if(buf .eq. '-f') then
            i = i+1
            call mygetarg(i,buf)
            filein = buf
         else if( buf .eq. '-i' ) then
            i=i+1
            call mygetarg(i,buf)
            name = buf
            lname=.true.       
         else if( buf .eq. '-p') then
            i = i+1
            call mygetarg(i,buf)
            operfile = buf
         else if( buf .eq. '-d' ) then
            i = i+1
            call mygetarg(i,buf)
            dvrfile = buf
         else if( buf .eq. '-o' ) then
            i = i+1
            call mygetarg(i,buf)
            fileout = buf
         elseif(buf .eq. '-r') then
            lrst = .true.
            filein = 'restart'
         elseif(buf .eq. '-w') then
            lovwr = .true.
         elseif(buf .eq. '-q') then
            lquick = .true.
         elseif(buf .eq. '-qtol') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) qtol
            lquick = .true.
         elseif(buf .eq. '-mc') then
            lmcint = .true.
            ladpint = .false.
         elseif(buf .eq. '-mc1') then
            lmcint1 = .true.
            ladpint = .false.
            mcran=0
         elseif(buf .eq. '-nosob') then
            mcran=0
         elseif(buf .eq. '-mctol') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) mctol
            if (ierr .ne. 0 .or. mctol.lt.0) then
               if (.not. lbatch) &
                  write(6,'(2a,/a)') 'Expecting a positive number ',&
                       'after -mctol, not: ', buf
               stop
            endif
         elseif(buf .eq. '-maxit') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) maxit
         elseif(buf .eq. '-nogwp') then
            lgwpint = .false.
         elseif(buf .eq. '-noq') then
            lquick = .false.
         elseif(buf .eq. '-saddle') then
            lsaddle = .true.
            ladpint = .false.
         elseif(buf .eq. '-saddle0') then
            lsaddle0 = .true.
            ladpint = .false.
         elseif(buf .eq. '-loose') then
            mctol=5.0d-3
            qtol=1.0d-4
         elseif(buf .eq. '-tight') then
            mctol=5.0d-4
            qtol=1.0d-8
         elseif(buf .eq. '-cdvr') then
            ltddvrint = .true.
            lcdvrint = .true.
            ladpint = .false.
         elseif(buf .eq. '-cdvr1d') then
            ltddvrint = .true.
            lcdvrint = .true.
            lcdvr1d = .true.
            ladpint = .false.
         elseif(buf .eq. '-tddvr') then
            ltddvrint = .true.
            ladpint = .false.
         elseif(buf .eq. '-n') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) ncut
            if (ierr .ne. 0 .or. ncut.le.0) then
               if (.not. lbatch) &
                  write(6,'(2a,/a)') 'Expecting a positive integer ',&
                       'after -n, not: ', buf
               stop
            endif
         elseif(buf .eq. '-skip') then
            i = i+1         
            call mygetarg(i,buf)         
            read(buf,*,iostat=ierr) nskip 
            if (ierr .ne. 0 .or. nskip.le.0) then
               if (.not. lbatch) &
                  write(6,'(2a,/a)') 'Expecting a positive integer ',&
                       'after -skip, not: ', buf
               stop         
            endif         
         elseif(buf .eq. '-step') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) step
            if (ierr .ne. 0 .or. step.le.0) then
               if (.not. lbatch) &
                  write(6,'(2a,/a)') 'Expecting a positive integer ',&
                       'after -step, not: ', buf
               stop
            endif
         elseif(buf .eq. '-tcut') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) tcut
            if (ierr .ne. 0 .or. tcut.le.0) then
               if (.not. lbatch) &
                  write(6,'(2a,/a)') 'Expecting a positive number ',&
                       'after -tcut, not: ', buf
               stop
            endif

         else if( buf(1:4) .eq. '-DB ' ) then
            i = i+1
            call mygetarg(i,buf)
            ddpath = buf

         else if( buf .eq. '-dblocal ' ) then
            ldblocal=.true.

         elseif(buf .eq. '-batch') then
            lbatch = .true.

         elseif(buf .eq. '-omp') then
            i = i+1
            call mygetarg(i,buf)
            read(buf,*,iostat=ierr) ompthread

         else if( buf(1:2) .eq. '-h' .or. buf(1:2) .eq. '-?' ) then
            write(6,'(1x,a,/a)')&
       ' Purpose: Calculates the diabatic and adiabatic population.',&
       '          And also one or two dimensional adiabatic densities.'
            write(6,'(1x,a,3(/11x,a))')&
       'Usage:   adpop [-i -f -d -p -o -r -n -skip -step -tcut -w ',&
                       ' -q -mc -mcd -mcs -mcsob -noq -qtol -mctol -nogwp ', &
                       ' -loose -tight -saddle -saddle0 -saddle1 ', &
#ifdef OMP
                       ' -cdvr -tddvr -omp -DB -batch -ver -h -?] [Degrees of Freedom]'
#else
                       ' -cdvr -tddvr -DB -dblocal -batch -ver -h -?] [Degrees of Freedom]'
#endif
            write(6,'(/,a)') 'Options  : '
            write(6,'(a)') ' -i DIR  : data is stored in directory DIR'
            write(6,'(2a,/,9x,a)') &
                     ' -f FILE : the PSI is read from',&
                     ' FILE rather than from ./psi',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a)') &
                     ' -d FILE : the DVR is read from',&
                     ' FILE rather than from ./dvr',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a)') &
                     ' -p FILE : the PES is read '&
                    ,'from FILE rather than from ./pes',&
            '  The string FILE may be a relative or a full path-name.'
            write(6,'(2a,/,9x,a,/,9x,a)') &
                     ' -o FILE : The ouput is written to file ',&
                     'FILE rather than to ./adp',&
            '  The string FILE may be a relative or a full path-name.',&
            '  If FILE=no, i.e. -o no, no output file will be written.'
            write(6,'(2a)') ' -r      : The PSI is read from',&
                             ' restart file rather than from ./psi'
            write(6,'(2a)')&
                     ' -n num  : only the first num WFs are read '&
                    ,' from psi-file.'
            write(6,'(a)')&
                     ' -skip m : the first m WFs are skipped.'
            write(6,'(2a)') ' -step step : Compute the population',&
                            ' only every step-th WF.'
            write(6,'(2a)') ' -tcut tcut : Compute the population',&
                            ' only at time tcut.'
            write(6,'(2a)') &
                     ' -w      : Overwrite the output-file adp.'
            write(6,'(2a)') &
                     ' -q      : The quick algorithm is used.'
            write(6,'(2a)') &
                     ' -mc     : Monte-Carlo integration is used.'
            write(6,'(2a)') &
                     ' -nosob  : Turn off Sobol sampling.'
            write(6,'(2a)') &
                     ' -noq    : Quick algorithm not used.'
            write(6,'(2a)') ' -qtol qtol   : The quick algorithm cutoff'&
                     ,' set to qtol.'
            write(6,'(2a)') ' -mctol mctol : The Monte-Carlo'&
                     ,' convergence set to mctol.'
            write(6,'(2a)') ' -nogwp : GWP integration turned off'
            write(6,'(5x,a)') ' Defaults: qtol=1.0d-4, mctol=1.0d-2'
            write(6,'(2a)') ' -sleazy : qtol and mctol relaxed   to:',&
                            '  qtol=1.0d-3, mctol=1.0d-1'
            write(6,'(2a)') ' -tight  : qtol and mctol tightened to:',&
                            '  qtol=1.0d-8, mctol=1.0d-3'
#ifdef OMP
            write(6,'(a)')&
                 ' -omp N : Use N threads '
#endif
            write(6,'(a)')&
                 ' -DB PATH : The path to a direct dynamics DB.'
            write(6,'(a)')&
                 ' -dblocal : Take nearest point in DB - no interpolation.'
            write(6,'(a)')&
                 ' -saddle : Use saddle point approximation for vMCG.'
            write(6,'(a)')&
                 ' -saddle0 : Uses RHS saddle point approximation for vMCG.'
            write(6,'(a,/,a)')&
                 ' -saddle1 sadtol : Uses saddle point approximation for vMCG', &
                 '          with MC for points with energy gap less than sadgap (in eV).'
            write(6,'(a)')&
                 ' -cdvr : Uses CDVR integration. Only works for single-set WFs'
            write(6,'(a)')&
                 ' -tddvr : Uses TDDVR integration. Only works for single-set WFs'
            write(6,'(a)')&
                 ' -batch : Do not output anything to the screen'
            write(6,'(a)')&
                     ' -ver    : Version information about the program '
            write(6,'(a,/a)') &
                     ' -h      : Print this help text. ',&
                     ' -?      : Print this help text. '
            write(6,'(/2a)')&
             ' If an exact WF is used then the pes file must be in',&
             ' exact representation too!'
            write(6,'(/2a,/2a)') &
             ' If there are two electronic states and if -mc is not', &
             ' set, then adpop',' computes the off-diagonal element of', &
             ' the electronic density matrix as well.'
            write(6,'(/2a,/2a,/2a)') &
             ' The arguments "Degrees of Freedom" have to be replaced',&
             ' by the names of the',&
             ' degrees of freedom for which densities are to be',&
             ' calculated.',&
             ' To calculate 2-dim densities separate the two DOF with',&
             ' a comma.'
            write(6,'(a)')  ' EXAMPLES: adpop dof1 '
            write(6,'(a)')  '           adpop dof3,dof4'
            write(6,'(a)')  '           adpop dof1 dof2 dof3,dof4' 
            write(6,'(2a)') ' The reduced densities can be plotted',&
                            ' with pladpop.'
            write(6,'(78a1)')  ('-',i=1,78)
            stop
         else if( buf(1:4) .eq. '-ver' ) then
            call wrversion(6)
            stop
         else 
            write(6,'(2a)') ' unknown option : ', buf
            stop
         end if
         goto 100
      endif
      i = i-1  ! No. of options read
      
!-----------------------------------------------------------------------
! set flags if DVR file needs to be read
!-----------------------------------------------------------------------
      laenge=index(name,' ')-1
      call abspath(name,laenge)

      if(operfile .eq. ' ') operfile='oper'
      opflaenge=index(operfile,' ')-1
      if(filein .eq. ' ') filein='psi'
      laein=index(filein,' ')-1
      if(dvrfile .eq. ' ') dvrfile='dvr'
      dlaenge=index(dvrfile,' ')-1

      if (laenge .gt. 0) then
         operfile = name(1:laenge)//'/'//operfile(1:opflaenge)
         opflaenge  = opflaenge+laenge+1
         filein = name(1:laenge)//'/'//filein(1:laein)
         laein  = laein+laenge+1
         dvrfile = name(1:laenge)//'/'//dvrfile(1:dlaenge)
         dlaenge  = dlaenge+laenge+1
      else
         call abspath(operfile,opflaenge)
         call abspath(filein,laein)
         call abspath(dvrfile,dlaenge)
      endif

      dname=dvrfile(1:dlaenge)
      
      if (fileout .eq. 'no') then
         continue
      else
         if (fileout .eq. ' ') fileout=name(1:laenge)//'/adp'
         laeout=index(fileout,' ')-1
         call abspath(fileout,laeout)
      endif

!-----------------------------------------------------------------------
! Open log file channel
!-----------------------------------------------------------------------
      if (fileout .eq. 'no') then
         filename=name(1:laenge)//'/adp.log'
      else
         filename=fileout(1:laeout)//'.log'
      endif
      open (unit=ilog,file=filename,form='formatted')
      logisopen=.true.
      if(lname) write(6,'(2a)') 'Output directed to: ',name(1:laenge)

!-----------------------------------------------------------------------
! --- WRITE DATE AND MACHINE TO LOG FILE ---
!-----------------------------------------------------------------------
      call fpstartup(host)
      write(ilog,*)
      write(ilog,'(a)') trim(host(1:80))
      write(ilog,'(1x,a)') name(1:laenge)
      
      if (.not. lbatch) &
         write(6,'(78a1)')  ('-',j=1,78)

!-----------------------------------------------------------------------
! Opening the PSI or RESTART file.
!-----------------------------------------------------------------------
      filename=filein
      if (lrst) irst=ipsi
      open(ipsi,file=filename,form='unformatted',status='old',err=1000)
      if (lrst) then
         if (.not. lbatch) &
            write(6,'(2a)') ' The restart file is  : '//filein(1:laein)
      else 
         if (.not. lbatch) &
            write(6,'(2a)') ' The psi file is  : '//filein(1:laein)
      endif

!-----------------------------------------------------------------------
!  Opening the output file
!-----------------------------------------------------------------------
      if (fileout .eq. 'no') then
         laeout=2
         open(iaus,file='/dev/null',form='formatted',status='unknown')
      else
         filename=fileout
         if(lovwr) then
            open(iaus,file=filename,form='formatted',status='unknown',&
                 err=1030)
         else
            open(iaus,file=filename,form='formatted',status='new',&
                 err=1020)
         endif
      endif

      write(iaus,'(/a2,a78)') '# ',host

!-----------------------------------------------------------------------
! Set lcdvr for memory allocation in hpsimod
!-----------------------------------------------------------------------
      lcdvr = lcdvrint .or. ltddvrint

!-----------------------------------------------------------------------
! set MC integrator flags
! Default is to use quick algorithm, unless using GWP integration
! or -noq flag was used.
!-----------------------------------------------------------------------
      if (lmcint1 .and. lgwpint) then
         write(ilog,'(/,a)') &
            'WARNING: MC with 1D sampling asked for but GWP integrals are possible'
         write(ilog,'(a)') &
            '         Defaulting to latter. To force MC with 1D sampling, use -nogwp'
         lmcint = .true.
         lmcint1 = .false.
         lquick = .false.
      endif

      return 
      
 1000 write(6,*)' Cannot open file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
 1020 write(6,*)' Cannot open the output file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      write(6,*) 'Use option -w to overwrite the output file.'
      stop
 1030 write(6,*)'Cannot open the output file: '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
      
      end subroutine adpopinput
      

!-----------------------------------------------------------------------
!***********************************************************************
!-----------------------------------------------------------------------

      subroutine adpopdeninp(i0)

      implicit none

      integer(long), intent(in)                      :: i0
      integer(long), external                        :: myiargc
      integer(long)                                  :: i,iarg,ilbl,endloop,&
                                                        s,f,k,outlaenge
      character(len=80)                              :: buf
      character(len=c5)                              :: filename,densityfile
      character(len=10), dimension(maxdim)           :: dofstring
      character(len=10), dimension(maxdim*2,2)       :: dof2dimstring
      logical :: lfound,lfound1

      iarg     = myiargc()
      zeigden(1)=1
      zeig2dimden(1)=1
      
      do f=1,maxdim
         dofnumber(f)=0
         dof2dimnumber(f,1)=0
         dof2dimnumber(f+maxdim,1)=0
         dof2dimnumber(f,2)=0
         dof2dimnumber(f+maxdim,2)=0
         outden(f)=50+f
         out2dimden(f)=50+f+maxdim
         out2dimden(f+maxdim)=50+f+2*maxdim
      enddo
      
!-----------------------------------------------------------------------
! Further arguments are for modes for which densities are to be 
! calculated
!-----------------------------------------------------------------------
      i=i0
      dofcount=0
      dof2dimcount=0
      dofstring=' '
      dof2dimstring=' '

 150  continue
      i=i+1
      if (iarg .lt. i) goto 200
      call mygetarg(i,buf)
      if(index(buf,',').ne.0)then
         dof2dimcount=dof2dimcount+1
         dof2dimstring(dof2dimcount,1)=buf(1:index(buf,',')-1)
         dof2dimstring(dof2dimcount,2)=&
                     buf(index(buf,',')+1:index(buf,' ')-1)
      else
         dofcount=dofcount+1
         dofstring(dofcount)=buf(1:index(buf,' ')-1)
      endif
      if (i .le. iarg) goto 150
      
  200 continue

      if(dofcount.gt.maxdim) then
         write(6,*) ' Too many degrees of freedom! '
         stop
      elseif(dof2dimcount.gt.maxdim*2) then
         write(6,*) ' Too many 2-dim densities! '
         stop
      endif

! check DOFs exist
      if (dofcount .gt. 0) then
         do i = 1,dofcount
            lfound = .false.
            do f=1,ndof
               if (dofstring(i) .eq. modelabel(f)) lfound = .true.
            enddo
            if (.not. lfound) then
               write(6,*) &
          'Incorrect modelabel specified for density calculation :',dofstring(i)
               write(6,'(a)') "STOP"
               write(6,*)
               stop
            endif
         enddo
         lden = .true.
      endif
      if (dof2dimcount .gt. 0) then
         do i = 1,dof2dimcount
            lfound = .false.
            lfound1 = .false.
            do f=1,ndof
               if (dof2dimstring(i,1) .eq. modelabel(f)) lfound = .true.
               if (dof2dimstring(i,2) .eq. modelabel(f)) lfound1 = .true.
            enddo
            if (.not. lfound .and. lfound1) then
               write(6,*) &
          'Incorrect modelabel specified for 2D density calculation :',&
                  dof2dimstring(i,1), dof2dimstring(i,2)
               write(6,'(a)') "STOP"
               write(6,*)
               stop
            endif
         enddo
         lden = .true.
      endif

      if (lden .and. .not. ladpint) then
         write(6,*)
         write(6,*) "Input Error: "
         write(6,'(a)') &
           "Can only calculate adiabatic densities using full grid integration"
         write(6,'(a)') "STOP"
         write(6,*)
         stop
      endif
      
!-----------------------------------------------------------------------
!  determine the degrees of freedom used for the densities
!  and open the density output files
!-----------------------------------------------------------------------
      do k=1,dof2dimcount
         do f=1,ndof
            if(modelabel(f).eq.dof2dimstring(k,1))&
               dof2dimnumber(k,1)=f
            if(modelabel(f).eq.dof2dimstring(k,2))&
               dof2dimnumber(k,2)=f
         enddo
         if(k.le.dof2dimcount) then
            zeig2dimden(k+1)=zeig2dimden(k)+gdim(dof2dimnumber(k,1))*&
                                        gdim(dof2dimnumber(k,2))*nddstate
         endif
         if(dof2dimnumber(k,1).eq.0) then
            write(6,'(2a)') 'unknown DOF : ', dof2dimstring(k,1)
            stop
         elseif(dof2dimnumber(k,2).eq.0) then
            write(6,'(2a)') 'unknown DOF : ', dof2dimstring(k,2)
            stop
         elseif(dof2dimnumber(k,1).eq.dof2dimnumber(k,2)) then
            write(6,'(4a)') 'twice the same DOF makes no sense: ',&
                             dof2dimstring(k,1),',',dof2dimstring(k,2)
            stop
         endif
         endloop=index(densityfile,' ')
         do s=1,endloop
            densityfile(s:s)=' '
         enddo
         densityfile(1:4)='adp_'
         densityfile(5:4+index(dof2dimstring(k,1),' '))=&
                  dof2dimstring(k,1)(1:index(dof2dimstring(k,1),' ')-1)
         densityfile(4+index(dof2dimstring(k,1),' '):&
                     4+index(dof2dimstring(k,1),' '))='_'
         densityfile(5+index(dof2dimstring(k,1),' '): &
                     4+index(dof2dimstring(k,1),' ')&
                      +index(dof2dimstring(k,2),' '))=&
                     dof2dimstring(k,2)(1:&
                                    index(dof2dimstring(k,2),' ')-1)
         outlaenge=index(densityfile,' ')-1
         call abspath(densityfile,outlaenge)
         filename=densityfile
         if(lovwr) then
            open(out2dimden(k),file=filename,form='formatted',&
                 status='unknown',err=1030)
         else
            open(out2dimden(k),file=filename,form='formatted',&
                 status='new',err=1020)
         endif
      enddo
    
      do k=1,dofcount
         do f=1,ndof
            if(modelabel(f).eq.dofstring(k))&
               dofnumber(k)=f
         enddo
         if(k.le.dofcount) then
            zeigden(k+1)=zeigden(k)+gdim(dofnumber(k))*nddstate
         endif
         if(dofnumber(k).eq.0) then
            write(6,'(2a)') 'unknown DOF: ', dofstring(k)
            stop
         endif
         endloop=index(densityfile,' ')
         do s=1,endloop
            densityfile(s:s)=' '
         enddo
         densityfile(1:4)='adp_'
         densityfile(5:4+index(dofstring(k),' '))=&
                     dofstring(k)(1:index(dofstring(k),' ')-1)
         outlaenge=index(densityfile,' ')-1
         call abspath(densityfile,outlaenge)
         filename=densityfile
         if(lovwr) then
            open(outden(k),file=filename,form='formatted',&
                 status='unknown',err=1030)
         else
            open(outden(k),file=filename,form='formatted',&
                 status='new',err=1020)
         endif
      enddo
      
      if(lrddvr) then
         maxgdim=1
         do f=1,ndof
            maxgdim=max(maxgdim,gdim(f))
         enddo
      endif
      
      return 
      
 1020 write(6,*)' Cannot open the output file : '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      write(6,*) 'Use option -w to overwrite the output file.'
      stop
 1030 write(6,*)'Cannot open the output density file: '
      ilbl=index(filename,' ')-1
      write(6,*) filename(1:ilbl)
      stop
      
      end subroutine adpopdeninp
      
!#######################################################################
! 
! Simple linear MC integration
!
!#######################################################################

      subroutine mcintegrate(psi,adpop,dipop,adiabden,adiabnsum, &
                 subdimo,gdimo,subdimn,gdimn,totgrid,&
                 hops,ort,time,gqi)

      implicit none

      integer(long)                                  :: f,m,i,s,n,&
                                                        nsum,ierr,&
                                                        idum,n1,f1
      integer(long)                                  :: whereiam
      integer(long), dimension(maxdim)               :: mpoint,fpoint
      integer(long), intent(in)                      :: totgrid
      integer(long), dimension(maxdim), intent(in)   :: subdimo,gdimo,&
                                                        gqi,subdimn,&
                                                        gdimn
      integer(long), dimension(den1dim+den2dim), intent(out) :: adiabnsum
      real(dop)                                      :: maxerr,pnt,&
                                                        xi1,xi2
      real(dop)                                      :: sumpop,norm,&
                                                        normsum,normvar,&
                                                        normsumvar,&
                                                        normerr
      real(sip), dimension(2)                        :: ran
      real(dop)                                      :: err
      real(dop), dimension(maxsta)                   :: adsum,disum,&
                                                        adpopvar,adsumvar,&
                                                        pop
      real(dop), intent(in)                          :: time
      real(dop), dimension(maxsta), intent(out)      :: dipop,adpop
      real(dop), dimension(den1dim+den2dim), intent(out) :: adiabden
      real(dop), dimension(hopsdim), intent(in)      :: hops
      real(dop), dimension(ortdim), intent(in)       :: ort
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(maxsta)                :: psidia,psiad
      real(dop), dimension(ortdim)                   :: dumweight

      dumweight(:) = 1.0_dop

!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      do s=1,nddstate
         adsum(s)=0.0_dop
         adsumvar(s)=0.0_dop
         disum(s)=0.0_dop
         dipop(s)=0.0_dop
         adpop(s)=0.0_dop
         adpopvar(s)=0.0_dop
      enddo
      normsum = 0.0_dop
      normsumvar = 0.0_dop

!-----------------------------------------------------------------------
! mcran sets the random no. generator used in MC. mcran = 2 is Sobol
! sequence
! uniform random generator is using the default seed, which may vary for
! diferent runs, so results may differ in detail
!-----------------------------------------------------------------------
      if (mcran .eq. 1) then
         call sobseq(-2,ran,ierr)
         if (ierr .eq. 1) then
            write(ilog,'(a)') &
                'MAXBIT too small. Changing to uniform random generator'
            mcran = 0
         endif
      endif

! start MC loop
      nsum=0
  100 continue

!-----------------------------------------------------------------------
! If totgrid small enough, use that to sample grid points.
! ELSE
! Sample particle grids
!-----------------------------------------------------------------------
      if (totgrid .gt. 0) then
! get 2 random nos.
  102    continue
         if (mcran .eq. 1) then
            call sobseq(2,ran,ierr)
            if (ierr .eq. 1) then
               write(ilog,'(a)') 'MAXBIT too small. Changing to uniform random generator'
               mcran = 0
               go to 102
            endif
            xi1=ran(1)
            xi2=ran(2)
         else
            xi1=randuniform()
            xi2=randuniform()
         endif

! get grid point
         pnt = 1 + xi1*(totgrid-1)
         i=nint(pnt)
         if (lquick) then
            call cpvxi(subdimn,subdim,nmode)
            call cpvxi(gdimn,gdim,ndof)
         endif

         call decoder(i,mpoint,fpoint)

      else
         if (lquick) then
            call cpvxi(subdimn,subdim,nmode)
            call cpvxi(gdimn,gdim,ndof)
         endif
         do m=1,nmode
            if (m .eq. meb) cycle

            do n=1,nspfdof(m)
               f = spfdof(n,m)
  103          continue
               if (mcran .eq. 1) then
                  call sobseq(1,ran,ierr)
                  if (ierr .eq. 1) then
                     write(ilog,'(a)') &
                         'MAXBIT too small. Changing to uniform random generator'
                     mcran = 0
                     go to 103
                  endif
                  xi1=ran(1)
               else
                  xi1=randuniform()
               endif

! get grid point
               pnt = 1 + xi1*(gdimn(f)-1)
               i=nint(pnt)
               fpoint(f) = i

            enddo
         enddo

      endif

!-----------------------------------------------------------------------
! correct grid positions for quick algorithm
!-----------------------------------------------------------------------
      if (lquick) then
         call cpvxi(subdimo,subdim,nmode)
         call cpvxi(gdimo,gdim,ndof)
         do m=1,nmode
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f=spfdof(n,m)
               fpoint(f) = fpoint(f) + gqi(f) - 1
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! get mpoint, i.e. point on mode grid
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle

         f=spfdof(1,m)
         mpoint(m) = fpoint(f)
         do n=2,nspfdof(m)
            f=spfdof(n,m)
            idum=1
            do n1=1,n-1
               f1=spfdof(n1,m)
               idum=idum*gdim(f1)
            enddo
            mpoint(m) = mpoint(m) + idum*(fpoint(f)-1)
         enddo
      enddo
!-----------------------------------------------------------------------
! Calculation of WF at point
!-----------------------------------------------------------------------
      if (psitype .eq. 1) then
         call expop(hops,psi(zetf(1,1)),subdim(1),nddstate,&
              psidia,psiad,i,fpoint,ort)
      else
         call wfpop(hops,psi,dgldim,nddstate,psidia,&
              psiad,fpoint,mpoint,ort,agmat,dumweight)
      endif

!-----------------------------------------------------------------------
! get density at point
!-----------------------------------------------------------------------
      sumpop = 0.0_dop
      do s=1,nddstate
         pop(s) = dble(psidia(s)*dconjg(psidia(s)))
         sumpop = sumpop + pop(s)
      enddo

!-----------------------------------------------------------------------
! sum up function values
!-----------------------------------------------------------------------
      nsum=nsum+1
      do s=1,nddstate
         disum(s) = disum(s)+pop(s)

         pop(s) = dble(psiad(s)*dconjg(psiad(s)))
         adsum(s) = adsum(s)+pop(s)
         adsumvar(s) = adsumvar(s)+pop(s)*pop(s)
      enddo
      normsum = normsum + sumpop
      normsumvar = normsumvar + sumpop*sumpop

!-----------------------------------------------------------------------
! make averages
!-----------------------------------------------------------------------
      do s=1,nddstate
         dipop(s) =  disum(s)/dble(nsum)
         adpop(s) =  adsum(s)/dble(nsum)
         adpopvar(s) = adsumvar(s)/dble(nsum)
      enddo
      norm = normsum/dble(nsum)
      normvar = normsumvar/dble(nsum)

!-----------------------------------------------------------------------
! sum the densities
!-----------------------------------------------------------------------
      if(lden) then
         do f=1,dofcount
            do s=1,nstate
               whereiam = zeigden(f)+(s-1)*gdim(dofnumber(f))+ &
                                     fpoint(dofnumber(f))-1
               adiabden(whereiam) = adiabden(whereiam) + pop(s)
               adiabnsum(whereiam) = adiabnsum(whereiam) + 1
            enddo
         enddo

         do f=1,dof2dimcount
            do s=1,nstate
               whereiam=zeig2dimden(f)+(s-1)* &
                         gdim(dof2dimnumber(f,1))* &
                         gdim(dof2dimnumber(f,2))+ &
                         gdim(dof2dimnumber(f,2))* &
                        (fpoint(dof2dimnumber(f,1))-1)+ &
                         fpoint(dof2dimnumber(f,2))+den1dim

               adiabden(whereiam)=adiabden(whereiam) + pop(s)
               adiabnsum(whereiam)=adiabnsum(whereiam) + 1
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! Force at least 100 samples
!-----------------------------------------------------------------------
      if (nsum .lt. 100) go to 100

!-----------------------------------------------------------------------
! calculate the error and stop if finished
!-----------------------------------------------------------------------
      normerr =(normvar - norm*norm)/dble(nsum)
      if (normerr .lt. 0.0_dop) normerr=0.0_dop
      normerr= normerr/dble(nsum)
      normerr= normerr/norm

      if (normerr .gt. mctol) go to 100

      maxerr =0.0_dop
      do s=1,nddstate
         err = (adpopvar(s) - adpop(s)*adpop(s))
         if (err .lt. 0.0_dop) err=0.0_dop
         err= err/dble(nsum)
         err= dsqrt(err)/norm

         maxerr = max(maxerr,err)
      enddo

      if (maxerr .gt. mctol) go to 100

! Normalise (multiply by integration volume)
      do s=1,nddstate
         dipop(s) = dipop(s)/norm
         adpop(s) = adpop(s)/norm
      enddo

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      write(ilog,'(f8.3,i10,a3,i10)') time,nsum,' / ',totgrid
      call myflush(ilog)

      return
      end subroutine mcintegrate

!#######################################################################
!
! MC integration using 1D densities as a sampling probablitiy
!
!#######################################################################

      subroutine mcintegrate1(psi,adpop,dipop,&
                 gdimn,gpop1,totgrid,hops,&
                 ort,time,gqi,maxgpop1,imgpop1)

      implicit none

      integer(long)                                  :: f,f1,m,i,s,n,g,&
                                                        nsum,&
                                                        idum,n1
      integer(long), dimension(maxdim)               :: mpoint,fpoint
      integer(long)                                  :: nadof,ntry
      integer(long), dimension(maxdim)               :: adof
      integer(long), intent(in)                      :: totgrid
      integer(long), dimension(maxdim), intent(in)   :: gqi,gdimn
      real(dop)                                      :: maxerr,pnt,&
                                                        xi1,xi2,xi3,&
                                                        pnttol,pnttol0
      real(dop)                                      :: sumpop,gpopnorm
      real(dop)                                      :: err,dtmp,normsum
      real(dop), dimension(maxsta)                   :: adsum,disum,&
                                                        adpopvar,adsumvar,&
                                                        pop
      real(dop), intent(in)                          :: time
      real(dop), dimension(maxsta), intent(out)      :: dipop,adpop
      real(dop), dimension(hopsdim), intent(in)      :: hops
      real(dop), dimension(ortdim), intent(in)       :: ort
      real(dop), dimension(ortdim,1), intent(in)     :: gpop1
      real(dop), dimension(ortdim)                   :: pntweight,gpop1mx
      integer(long), dimension(maxdim), intent(in)   :: imgpop1
      real(dop), dimension(maxdim), intent(in)       :: maxgpop1
      complex(dop), dimension(dgldim), intent(in)    :: psi
      complex(dop), dimension(maxsta)                :: psidia,psiad
   
!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      do s=1,nddstate
         adsum(s)=0.0_dop
         adsumvar(s)=0.0_dop
         disum(s)=0.0_dop
         dipop(s)=0.0_dop
         adpop(s)=0.0_dop
         adpopvar(s)=0.0_dop
      enddo
      normsum = 0.0_dop

! get active DOF
      i=0
      do m=1,nmode
         if (m .eq. meb) cycle
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            i=i+1
            adof(i)=f
         enddo
      enddo
      nadof=i
      
!-----------------------------------------------------------------------
! Get sqrt of 1D densities for weights at each point
!-----------------------------------------------------------------------
      pntweight(:) = 0.0_dop
      gpopnorm = 1.0_dop
      do m=1,nmode
         if (m .eq. meb) cycle
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gpopnorm = gpopnorm*maxgpop1(f)
            do g=1,gdim(f)
               dtmp = dsqrt(gpop1(zort(f)+g-1,1))
               if (dtmp .lt. tiny) then
                  pntweight(zort(f)+g-1)= 9999.0
               else
                  pntweight(zort(f)+g-1)= 1.0d0/dtmp
               endif

               gpop1mx(zort(f)+g-1)= gpop1(zort(f)+g-1,1)/(maxgpop1(f))
            enddo
         enddo
      enddo

! start MC loop
      nsum=0
  100 continue

      if (nsum .eq. 0) then
! first point at equilibrium values
         do m=1,nmode
            if (m .eq. meb) cycle
            do n=1,nspfdof(m)
               f = spfdof(n,m)
               fpoint(f) = imgpop1(f) 
            enddo
         enddo

! set guess density
         pnttol = 1.0_dop
         do m=1,nmode
            if (m .eq. meb) cycle
          
            do n=1,nspfdof(m)
               f = spfdof(n,m)
               i = fpoint(f)
               pnttol = pnttol*gpop1mx(zort(f)+i-1)
            enddo
         enddo
         pnttol0 = pnttol

      else
         
         ntry = 0

! move to new dof+grid point. 
         xi1=randuniform()
         xi1 = 1 + xi1*(nadof-1)
         i=nint(xi1)
         f1 = adof(i)

   103   continue

         xi2=randuniform()
         pnt = gqi(f1) + xi2*(gdimn(f1)-1)
         n=nint(pnt)

         ntry = ntry+1
         pnttol=1.0d0
         do i=1,nadof
            f = adof(i)
            g = fpoint(f)
            if (f .eq. f1) g=n
            pnttol = pnttol*gpop1mx(zort(f)+g-1)
         enddo

         xi3=randuniform()
         xi3 = xi3*pnttol0
         if (xi3 .gt. pnttol) then
            if (ntry .ge. 1000) then
              go to 103
            endif
         endif

         fpoint(f1)=n

      endif

!-----------------------------------------------------------------------
! get mpoint, i.e. point on mode grid
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle

         f=spfdof(1,m)
         mpoint(m) = fpoint(f)
         do n=2,nspfdof(m)
            f=spfdof(n,m)
            idum=1
            do n1=1,n-1
               f1=spfdof(n1,m)
               idum=idum*gdim(f1)
            enddo
            mpoint(m) = mpoint(m) + idum*(fpoint(f)-1)
         enddo
      enddo

!-----------------------------------------------------------------------
! Calculation of WF at point
!-----------------------------------------------------------------------
      if (psitype .eq. 1) then
         call expop(hops,psi(zetf(1,1)),subdim(1),nddstate,&
              psidia,psiad,i,fpoint,ort)
      else
         call wfpop(hops,psi,dgldim,nddstate,psidia,&
              psiad,fpoint,mpoint,ort,agmat,pntweight)
      endif

!-----------------------------------------------------------------------
! get density at point
!-----------------------------------------------------------------------
      sumpop = 0.0_dop
      do s=1,nddstate
         pop(s) = dble(psidia(s)*conjg(psidia(s)))
         sumpop = sumpop + pop(s)
      enddo

!-----------------------------------------------------------------------
! sum up function values
!-----------------------------------------------------------------------
      nsum=nsum+1
      do s=1,nddstate
         pop(s) = dble(psidia(s)*conjg(psidia(s)))
         disum(s) = disum(s)+pop(s)

         pop(s) = dble(psiad(s)*conjg(psiad(s)))
         adsum(s) = adsum(s)+pop(s)
         adsumvar(s) = adsumvar(s)+pop(s)*pop(s)
      enddo

!-----------------------------------------------------------------------
! make averages
!-----------------------------------------------------------------------
      do s=1,nddstate
         dipop(s) =  disum(s)/dble(nsum)
         adpop(s) =  adsum(s)/dble(nsum)
         adpopvar(s) = adsumvar(s)/dble(nsum)
      enddo

      if (nsum .lt. 100) go to 100

!-----------------------------------------------------------------------
! calculate the error and stop if finished
!-----------------------------------------------------------------------
      maxerr =0.0_dop
      do s=1,nddstate
         err = (adpopvar(s) - adpop(s)*adpop(s))
         if (err .lt. 0.0_dop) err=0.0_dop
         err= err/dble(nsum)
         err= dsqrt(err)

         maxerr = max(maxerr,err)
      enddo

      if (maxerr .gt. mctol) go to 100
   
!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      write(ilog,'(f8.3,i10,a3,i10)') time,nsum,' / ',totgrid
      call myflush(ilog)

      return
      end subroutine mcintegrate1

!#######################################################################
! ONLY FOR FG
!#######################################################################

      subroutine mcinteggwp(psi,adiabpop,diapop,qcenter,hops,time)

      implicit none

      integer(long)                                     :: m,i,s,s1,i1,&
                                                           zeig,zeig1,zeig2,sa
      integer(long)                                     :: n,block1
      integer(long)                                     :: v1,e1,n1,dim1,nsum,&
                                                           v2,e2,n2,totiter,totint
      integer(long), dimension(:), allocatable          :: tmpiter,nint
      integer(long)                                     :: nham
      integer(long)                                     :: ithr,iit
      integer(long), dimension(maxdim)                  :: mpoint,fpoint,mpoint1
      real(dop), dimension(maxdim)                      :: gpoint,gpoint1,gpoint2
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(:,:), allocatable            :: tmpadiabpop
      real(dop), dimension(hopsdim), intent(in)         :: hops
      complex(dop), dimension(dgldim), intent(in)       :: psi
      real(dop), dimension(maxsta)                      :: adiab,popsum,popvar,&
                                                           adpop,adpopvar
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat
      complex(dop), dimension(nddstate,nddstate)        :: pops

      real(dop), intent(in)                             :: time
      real(dop), dimension(qcentdim), intent(in)        :: qcenter
      real(dop), dimension(maxdim)                      :: qcoo,qsigma2,qmom
      real(dop)                                         :: norm0,pop,err,maxerr,&
                                                           totphase,imphase,&
                                                           alpha,delq,maxmaxerr

      complex(dop), dimension(adim)                     :: dipsi
      complex(dop), dimension(2*maxblock)               :: tmppsi
      complex(dop), dimension(maxsta)                   :: dipsig,adpsig,&
                                                           dipsig1,adpsig1
      complex(dop)                                      :: gpair
      complex(dop), dimension(maxdim)                   :: A,B,A1,B1
      complex(dop)                                      :: C,C1,ztmp,zpop
      complex(dop)                                      :: phase,gstmp,gsgwp

      iit = 0
      fpoint(:) = 0.0_dop

      allocate(tmpiter(ompthread))
      allocate(nint(ompthread))
      allocate(tmpadiabpop(maxsta,ompthread))

!-----------------------------------------------------------------------
! transform A coefficients to correct basis (electronic state + GWP)
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         zeig2=0
         call cpvxz(psi(zpsi(s)),tmppsi(zeig),block(s))
         do m=1,nmode
            if (m .eq. meb) then
! if meb then transform from SPF basis to electronic state basis.
               call mtxxzz(psi(zetf(m,s)),tmppsi(zeig),tmppsi(zeig1),&
                           vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            else if (gwpm(m) .and. psitype .eq. 0) then
! if gwp-spf then transform from SPF basis to GWP basis.
               call qtxxzz(psi(zdgwp(m,s)),tmppsi(zeig), &
                    tmppsi(zeig1),vdim(m,s),dim(m,s),ndim(m,s))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            endif
         enddo
      enddo
      dipsi(1:adim)=tmppsi(zeig:zeig-1+adim)

!-----------------------------------------------------------------------
! multiply DIA by overlaps and get diabatic pops
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         call cpvxz(dipsi(zpsi(s)),tmppsi(zeig),block(s))

         do m=1,nmode
            if (m .eq. meb) cycle

            call qtxxzz(gs2(gs2mat(m,s)),tmppsi(zeig),tmppsi(zeig1), &
                 vdim(m,s),dim(m,s),ndim(m,s))
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo

         if (meb .gt. 0) then
            call rmhxxxzz(dipsi(zpsi(s)),tmppsi(zeig),pops,vdim(meb,s),&
                 dim(meb,s),ndim(meb,s))
            do s1=1,gdim(feb)
               diapop(s1) = dble(pops(s1,s1))
            enddo
         else
            diapop(s) = 0.0_dop
            do i=1,block(s)
               diapop(s) = diapop(s) + &
                   dble(dconjg(dipsi(zpsi(s)+i-1))*tmppsi(zeig+i-1))
            enddo
         endif
      enddo

!-----------------------------------------------------------------------
! build operator matrix.
! i / mpoint GWP on RHS
! i1 / mpoint1 GWP on LHS
!-----------------------------------------------------------------------
      s = 1
      block1 = block(s)
      if (leb) block1 = block1/dim(meb,1)
      
! calculate no. of integrals and set a threshold for calculating integrals

      tmpiter(:) = 0
      tmpadiabpop(:,:) = 0.0
      nint(:) = 0
      maxmaxerr = 0.0
!$omp parallel default(shared) num_threads(ompthread) &
!$omp private(ithr,mpoint,mpoint1,gpoint,gpoint1,gpoint2,v1,e1,n1,v2,e2,n2) &
!$omp private(dipsig,dipsig1,adpsig,adpsig1,A,B,C,A1,B1,C1,qsigma2,nham) &
!$omp private(m,f,dim1,popsum,popvar,gpair,qcoo,rotmat,adiab,diab,fpoint) &
!$omp private(nsum,adpop,adpopvar,err,maxerr,i,i1,norm0,alpha,delq,imphase) &
!$omp private(ztmp,qmom,totphase,phase,zpop,pop,gstmp)
!$omp do schedule(dynamic)
      do i=1,block1

#ifdef OMP
         ithr = omp_get_thread_num()+1
#else
         ithr = 1
#endif
         call adecoder_nomeb(i,mpoint)

         dim1=1
         call vendecoder(v1,e1,n1,i,vdim(meb,s),dim1)

         call di2dig(dipsi,dipsig,vdim(meb,s),dim(meb,s),ndim(meb,s),v1,n1)
         
!-----------------------------------------------------------------------
! Get centre of GWP (no. of function for each mode in mpoint(m))
!-----------------------------------------------------------------------
         gpoint = 0.0_dop
         C = 0.0_dop   ! C is additive for particles
         do m=1,nmode
            if (m .eq. meb) cycle

            call getqcent(m,gpoint,qcenter(zcent(m,s)),nspfdof(m),&
                 gwpdim(m,s),mpoint(m))
            call getgwppar(m,nspfdof(m),gwpdim(m,s),A,B,C,agmat(zagmat(m,s)),&
                 agmat(zxivec(m,s)),agmat(zphase(m,s)),mpoint(m))
         enddo

!-----------------------------------------------------------------------
! Test for density. If small, ignore integral
!-----------------------------------------------------------------------
         if (lquick) then
            call sum2vxz(dipsig,ztmp,dim(meb,s))
            if (dble(ztmp) .lt. qtol) go to 100  ! skip integral
         endif

         nint(ithr) = nint(ithr)+1   ! count number of integrals done

!-----------------------------------------------------------------------
! variance.
!-----------------------------------------------------------------------
         qsigma2=0.0_dop
         do m=1,nmode
            if (m .eq. meb) cycle

            do n=1,nspfdof(m)
               f=spfdof(n,m)
               qsigma2(f)=dsqrt(-4.0d0*dble(A(f)))
               qsigma2(f)=1.0/qsigma2(f)
            enddo
         enddo

!-----------------------------------------------------------------------
! Now do MC integration
!-----------------------------------------------------------------------
         popsum(:) = 0.0_dop
         popvar(:) = 0.0_dop
         nsum = 0
         do 
! get random point
            qcoo(:) = gpoint(:)
            do m=1,nmode
               if (m .eq. meb) cycle

               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  gpair=randgaussian()

                  qcoo(f)=qcoo(f) + qsigma2(f)*real(gpair)
               enddo
            enddo

! get ADT at qcoo
            nham=1
            num_gp =i
            call calcdiab(hops,adiab,diab,rotmat,fpoint,qcoo,nham)

! now store integrated ADT matrix
            nsum = nsum+1
            call qvtxzz1(rotmat,dipsig,adpsig,maxsta,nddstate)
            do sa=1,nddstate
               pop = dble(dconjg(adpsig(sa))*adpsig(sa))
               popsum(sa) = popsum(sa) + pop
               popvar(sa) = popvar(sa) + pop*pop
            enddo

            if (nsum .lt. 100) cycle
            do sa=1,nddstate
               adpop(sa) = popsum(sa) / dble(nsum)
               adpopvar(sa) = popvar(sa) / dble(nsum)
            enddo

            maxerr = 0.0_dop
            do sa=1,nddstate
               err = adpopvar(sa) - adpop(sa)*adpop(sa)
               if (err .lt. 0.0_dop) err=0.0_dop
               err= err/dble(nsum)
               err= dsqrt(err)

               maxerr = max(maxerr,err)
            enddo
            if (maxerr .lt. mctol) exit
            if (nsum .gt. maxit) then
               iit = 1
               maxmaxerr = max(maxerr,maxmaxerr)
               exit
            endif

         enddo  ! end of MC iterations

! sum no. of MC steps
         tmpiter(ithr) = tmpiter(ithr) + nsum

! Add term to total population
         do sa=1,nddstate
            tmpadiabpop(sa,ithr) = tmpadiabpop(sa,ithr) + adpop(sa)
         enddo

  100    continue
!-----------------------------------------------------------------------
! Start of loop over off-diagonal elements
!-----------------------------------------------------------------------
         do i1=i+1,block1
            call adecoder_nomeb(i1,mpoint1)

            dim1=1
            call vendecoder(v2,e2,n2,i1,vdim(meb,s),dim1)

            call di2dig(dipsi,dipsig1,vdim(meb,s),dim(meb,s),ndim(meb,s),v2,n2)

!-----------------------------------------------------------------------
! Test for density. If small, ignore integral
!-----------------------------------------------------------------------
            if (lquick) then
               call vvaxzz(dipsig1,dipsig,ztmp,dim(meb,s))
               ztmp = 1.0_dop
               do m=1,nmode
                  if (m .eq. meb) cycle
                  call getgsmat(gstmp,gs2(gs2mat(m,s)),gwpdim(m,s),&
                                mpoint1(m),mpoint(m))
                  ztmp = ztmp*gstmp
               enddo
               if (zabs(ztmp) .lt. qtol) cycle  ! skip integral
            endif 

            nint(ithr) = nint(ithr)+1

            gpoint1 = 0.0_dop
            C1 = 0.0_dop   ! C is additive for particles
            do m=1,nmode
               if (m .eq. meb) cycle

               call getqcent(m,gpoint1,qcenter(zcent(m,s)),nspfdof(m),&
                    gwpdim(m,s),mpoint1(m))
               call getgwppar(m,nspfdof(m),gwpdim(m,s),A1,B1,C1,agmat(zagmat(m,s)),&
                    agmat(zxivec(m,s)),agmat(zphase(m,s)),mpoint1(m))
            enddo

            qsigma2=0.0_dop
            do m=1,nmode
               if (m .eq. meb) cycle

               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  qsigma2(f)=dsqrt(-2.0*dble(A(f) +A1(f)))
                  qsigma2(f)=1.0/qsigma2(f)
               enddo
            enddo

            gpoint2(:) = 0.5_dop*(gpoint(:) + gpoint1(:))

            norm0 = 0.0_dop
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  alpha = -0.25_dop*dble(A(f) + A1(f))
                  delq = gpoint1(f) - gpoint(f)
                  norm0 = norm0 - alpha*delq*delq
               enddo
            enddo
            norm0 = dexp(norm0)

            ztmp = dconjg(C1) + C
            imphase=dimag(ztmp)
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  ztmp = dconjg(B1(f)) + B(f)
                  qmom(f)=dimag(ztmp)
               enddo
            enddo

!-----------------------------------------------------------------------
! Now do MC integration
!-----------------------------------------------------------------------
            popsum(:) = 0.0_dop
            popvar(:) = 0.0_dop
            nsum = 0
            do
! get random point
               qcoo(:) = gpoint2(:)
               do m=1,nmode
                  if (m .eq. meb) cycle

                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     gpair=randgaussian()

                     qcoo(f)=qcoo(f) + qsigma2(f)*real(gpair)
                  enddo
               enddo

               totphase = imphase
               do m=1,nmode
                  if (m .eq. meb) cycle

                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     totphase = totphase + qmom(f)*qcoo(f) 
                  enddo
               enddo
               phase = cos(totphase) + (0,1)*sin(totphase)

! get ADT at qcoo
               nham=1
               num_gp=i
               call calcdiab(hops,adiab,diab,rotmat,fpoint,qcoo,nham)

! now store integrated ADT matrix
               nsum = nsum+1
               call qvtxzz1(rotmat,dipsig,adpsig,maxsta,nddstate)
               call qvtxzz1(rotmat,dipsig1,adpsig1,maxsta,nddstate)
               do sa=1,nddstate
                  zpop = dconjg(adpsig1(sa))*adpsig(sa) 
                  pop = 2.0_dop*norm0*dble(zpop*phase)
                  popsum(sa) = popsum(sa) + pop
                  popvar(sa) = popvar(sa) + pop*pop
               enddo

               if (nsum .lt. 100) cycle
               do sa=1,nddstate
                  adpop(sa) = popsum(sa) / dble(nsum)
                  adpopvar(sa) = popvar(sa) / dble(nsum)
               enddo

               maxerr = 0.0_dop
               do sa=1,nddstate
                  err = adpopvar(sa) - adpop(sa)*adpop(sa)
                  if (err .lt. 0.0_dop) err=0.0_dop
                  err= err/dble(nsum)
                  err= dsqrt(err)

                  maxerr = max(maxerr,err)
               enddo
               if (maxerr .lt. mctol) exit
               if (nsum .gt. maxit) then
                  iit = 1
                  maxmaxerr = max(maxerr,maxmaxerr)
                  exit
               endif

            enddo  ! end of MC iterations

! sum no. of MC steps
            tmpiter(ithr) = tmpiter(ithr) + nsum

! Add term to total population
            do sa=1,nddstate
               tmpadiabpop(sa,ithr) = tmpadiabpop(sa,ithr) + adpop(sa)
            enddo

         enddo  ! end of off-diagonal loop (GWP on LHS)

      enddo ! end of loop over GWP on RHS
!$omp end do
!$omp end parallel

      totiter = 0
      totint = 0
      adiabpop(:) = 0
      do ithr=1,ompthread
         totiter = totiter + tmpiter(ithr)
         adiabpop(1:nddstate) = adiabpop(1:nddstate) + &
             tmpadiabpop(1:nddstate,ithr)
         totint = totint + nint(ithr)
      enddo

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      if (iit .eq. 0) then
        write(ilog,'(f8.3,i10,a3,i10)') &
           time,totiter,' / ',totiter/totint
      else
        write(ilog,'(f8.3,i10,a3,i10,a,e10.2)') &
           time,totiter,' / ',totiter/totint, &
          ' Not converged. Maximum error: ',maxmaxerr
      endif
      call myflush(ilog)

      deallocate(tmpiter,tmpadiabpop,nint)

      return
      end subroutine mcinteggwp

!#######################################################################
!
! Extracts parameters for a GWP. Note that C is additive for partitioned
! GWPs.
!
!#######################################################################

      subroutine getgwppar(m,gdof,gwpdim1,A,B,C,agmat,xigmat,phigmat,e)

      integer(long), intent(in) :: m,e,gdof,gwpdim1
      integer(long)             :: n,f
      complex(dop), dimension(gdof,gdof,gwpdim1), intent(in) :: agmat
      complex(dop), dimension(gdof,gwpdim1), intent(in)      :: xigmat
      complex(dop), dimension(gwpdim1), intent(in)           :: phigmat
      complex(dop), dimension(maxdim), intent(out)           :: A,B
      complex(dop), intent(inout)                            :: C

      do n=1,gdof
         f=spfdof(n,m)
         A(f) = agmat(n,n,e)
         B(f) = xigmat(n,e)
      enddo
      C = C + phigmat(e)

      return
      end subroutine getgwppar

!#######################################################################

      subroutine di2dig(dipsi,dipsig,vdim1,dim1,ndim1,v1,n1)

      implicit none

      integer(long), intent(in) :: vdim1,dim1,ndim1,v1,n1
      integer(long)             :: s
      complex(dop), dimension(vdim1,dim1,ndim1), intent(in)  :: dipsi
      complex(dop), dimension(dim1), intent(out)             :: dipsig

      do s=1,dim1
         dipsig(s) = dipsi(v1,s,n1)
      enddo

      end subroutine di2dig

!#######################################################################
!
      subroutine sadintegrate(psi,adiabpop,diapop,qcenter,hops)

      implicit none

      integer(long)                                     :: m,i,s,s1,i1,si,sf,&
                                                           zeig,zeig1,zeig2,sa
      integer(long)                                     :: nham
      integer(long), dimension(maxdim)                  :: mpoint,fpoint,mpoint1
      real(dop), dimension(maxdim)                      :: gpoint,gpoint1,gpoint2
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(hopsdim), intent(in)         :: hops
      complex(dop), dimension(dgldim), intent(in)       :: psi
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat
      complex(dop), dimension(nddstate,nddstate)        :: pops

      real(dop), dimension(qcentdim), intent(in)        :: qcenter

      complex(dop), dimension(adim)                     :: dipsi,adpsi
      complex(dop), dimension(2*maxblock)               :: tmppsi
      complex(dop)                                      :: gstmp

      complex(dop), dimension(:,:,:), allocatable         :: opmat

!-----------------------------------------------------------------------
! transform A coefficients to correct basis (electronic state + GWP)
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         zeig2=0
         call cpvxz(psi(zpsi(s)),tmppsi(zeig),block(s))
         do m=1,nmode
            if (m .eq. meb) then
! if meb then transform from SPF basis to electronic state basis.
               call mtxxzz(psi(zetf(m,s)),tmppsi(zeig),tmppsi(zeig1),&
                           vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            else if (gwpm(m) .and. psitype .eq. 0) then
! if gwp-spf then transform from SPF basis to GWP basis.
               call qtxxzz(psi(zdgwp(m,s)),tmppsi(zeig), &
                    tmppsi(zeig1),vdim(m,s),dim(m,s),ndim(m,s))
               zeig2=zeig
               zeig=zeig1
               zeig1=zeig2
            endif
         enddo
      enddo
      dipsi(1:adim)=tmppsi(zeig:zeig-1+adim)

!-----------------------------------------------------------------------
! matrix operator
!-----------------------------------------------------------------------
      allocate(opmat(adim,adim,nddstate))
      opmat(:,:,:)=0.0_dop

!-----------------------------------------------------------------------
! build operator matrix. 
! i / mpoint GWP on RHS
! i1 / mpoint1 GWP on LHS
!-----------------------------------------------------------------------
      do i=1,adim
         call adecoder(i,mpoint,s)
         num_gp = mpoint(1)
         do m=2,nmode
            if (m .eq. meb) cycle
            num_gp = num_gp*mpoint(m)
         enddo


         gpoint = 0.0_dop
         do m=1,nmode
            if (m .eq. meb) cycle

            call getqcent(m,gpoint,qcenter(zcent(m,s)),nspfdof(m),&
                 gwpdim(m,s),mpoint(m))
         enddo

         do i1=i,adim
            if (i1 .ne. i) then
               call adecoder(i1,mpoint1,s1)

               gpoint1 = 0.0_dop
               do m=1,nmode
                  if (m .eq. meb) cycle

                  call getqcent(m,gpoint1,qcenter(zcent(m,s1)),nspfdof(m),&
                       gwpdim(m,s1),mpoint1(m))

                  gpoint2(:) = 0.5_dop*(gpoint(:) + gpoint1(:))
               enddo
            else
               s1 = s
               mpoint1=mpoint
               gpoint2 = gpoint
            endif

            if (meb .gt. 0) then
               si = mpoint(meb)
               sf = mpoint1(meb)
            else
               si = s
               sf = s1
            endif

! get ADT at gpoint
            nham=1
            call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint2,nham)

! now build operator matrices
            do sa=1,nddstate
               opmat(i1,i,sa) = rotmat(sf,sa)*dconjg(rotmat(si,sa))

               do m=1,nmode
                  if (m .eq. meb) cycle

                  call getgsmat(gstmp,gs2(gs2mat(m,s)),gwpdim(m,s),&
                       mpoint1(m),mpoint(m))

                  opmat(i1,i,sa) = opmat(i1,i,sa)*gstmp
               enddo

               if (i .eq. i1) then
                  opmat(i,i,sa) = dble(opmat(i,i,sa))
               else
                  opmat(i,i1,sa) = dconjg(opmat(i1,i,sa))
               endif
            enddo

         enddo

      enddo

!-----------------------------------------------------------------------
! multiply DIA by overlap and get diabatic pops
!-----------------------------------------------------------------------
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         call cpvxz(dipsi(zpsi(s)),tmppsi(zeig),block(s))
         do m=1,nmode
            if (m .eq. meb) cycle
       
            call qtxxzz(gs2(gs2mat(m,s)),tmppsi(zeig),tmppsi(zeig1), &
                 vdim(m,s),dim(m,s),ndim(m,s))
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo

         if (meb .gt. 0) then
            call rmhxxxzz(dipsi(zpsi(s)),tmppsi(zeig),pops,vdim(meb,s),&
                 dim(meb,s),ndim(meb,s))
            do s1=1,gdim(feb)
               diapop(s1) = dble(pops(s1,s1))
            enddo
         else
            diapop(s) = 0.0_dop
            do i=1,block(s)
               diapop(s) = diapop(s) + &
                   dble(dconjg(dipsi(zpsi(s)+i-1))*tmppsi(zeig+i-1))
            enddo
         endif
      enddo

!-----------------------------------------------------------------------
! evaluate projection operators to get adiabatic populations
!-----------------------------------------------------------------------
      do sa=1,nddstate
         adiabpop(sa) = 0.0_dop

         call qvxxzz(opmat(1,1,sa),dipsi,adpsi,adim)

         do i=1,adim
            adiabpop(sa) = adiabpop(sa) + &
                   dble(dconjg(dipsi(i))*adpsi(i))
         enddo

      enddo

      deallocate(opmat)

      return
      end subroutine sadintegrate

!#######################################################################
!
      subroutine sadintegrate0(psi,adiabpop,diapop,qcenter,hops)

      implicit none

      integer(long)                                     :: m,i,s,si,sf,&
                                                           zeig,zeig1,zeig2,&
                                                           swapzeig,sa
      integer(long)                                     :: nham
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(hopsdim), intent(in)         :: hops
      complex(dop), dimension(dgldim), intent(in)       :: psi
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(qcentdim), intent(in)        :: qcenter

      complex(dop), dimension(adim)                     :: dipsi
      complex(dop), dimension(2*maxblock)               :: tmppsi

      complex(dop), dimension(:,:), allocatable         :: opmat
      real(dop), dimension(:,:,:), allocatable          :: ggp
      complex(dop), dimension(:), allocatable     :: workc

! Note that for single set gdim(feb) = dim(meb,1) = nddstate
      if (leb) then
         ngwp = adim/gdim(feb)
      else
         ngwp = maxblock
      endif

!-----------------------------------------------------------------------
! matrix operator
!-----------------------------------------------------------------------
      allocate(opmat(ngwp,nddstate))
      allocate(ggp(ngwp,nddstate,nddstate))
      allocate(workc(2*maxblock))

!-----------------------------------------------------------------------
! transform A coefficients from spf to GWP basis
!-----------------------------------------------------------------------
      zeig1=maxblock+1
      zeig2=1
      do s=1,nstate
         call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))
         do m=1,nmode
            if (m .eq. meb) then
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
 !if meb then transform from SPF basis to electronic state basis.
               call mtxxzz(psi(zetf(m,s)),workc(zeig1),workc(zeig2),&
                              vdim(m,s),dim(m,s),ndim(m,s),subdim(m))
            else if (psitype .eq. 0) then
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call qtxxzz(psi(zdgwp(m,s)),workc(zeig1), &
                    workc(zeig2),vdim(m,s),dim(m,s),ndim(m,s))
            endif
         enddo
         dipsi(zpsi(s):zpsi(s)+block(s)-1)=workc(zeig2:zeig2+block(s)-1)
      enddo

!-----------------------------------------------------------------------
! Calculate GGPs
! multiply DIA by overlap 
!-----------------------------------------------------------------------
      zeig=0
      do s=1,nstate
         zeig=1
         zeig1=block(s)+1
         call cpvxz(dipsi(zpsi(s)),tmppsi(zeig),block(s))
         do m=1,nmode
            if (m .eq. meb) cycle
       
            call qtxxzz(gs2(gs2mat(m,s)),tmppsi(zeig),tmppsi(zeig1), &
                 vdim(m,s),dim(m,s),ndim(m,s))
            zeig2=zeig
            zeig=zeig1
            zeig1=zeig2
         enddo
      enddo

      if (leb) then
         call getggps(dipsi,tmppsi(zeig),ggp,vdim(meb,1),&
              dim(meb,1),ndim(meb,1))
      else
         if (.not. lbatch) &
            write(6,'(a)') 'GGP integrals Not Yet For multi-set'
         stop
      endif

!-----------------------------------------------------------------------
! build operator matrix. 
!-----------------------------------------------------------------------
      if (leb) then

         opmat(:,:) = 0.0_dop
         do i = 1,ngwp
            call adecoder(i,mpoint,s)

            gpoint = 0.0_dop
            do m=1,nmode
               if (m .eq. meb) cycle

               call getqcent(m,gpoint,qcenter(zcent(m,s)),nspfdof(m),&
                    gwpdim(m,s),mpoint(m))
            enddo

! get ADT at gpoint
            nham=1
            num_gp =i
            call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

! now build operator matrices
            do sa=1,nddstate
               do si=1,gdim(feb)
                  do sf=1,gdim(feb)
                     opmat(i,sa) = opmat(i,sa) + &
                         rotmat(sf,sa)*dconjg(rotmat(si,sa))*ggp(i,sf,si)
                 enddo
               enddo
            enddo
         enddo

      endif

!-----------------------------------------------------------------------
! sum up ggps to get diabatic populations
!-----------------------------------------------------------------------
      if (leb) then
         diapop(:) = 0.0_dop
         do s=1,nddstate
            do i=1,ngwp
               diapop(s) = diapop(s) + ggp(i,s,s)
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! evaluate projection operators to get adiabatic populations
!-----------------------------------------------------------------------
      adiabpop(:) = 0.0_dop
      do sa=1,nddstate
         do i=1,ngwp
            adiabpop(sa) = adiabpop(sa) + dble(opmat(i,sa))
         enddo
      enddo

      deallocate(opmat,ggp,workc)

      return
      end subroutine sadintegrate0

!#######################################################################

      subroutine getqcent(m,gpoint,qcenter,gdof,gwpdim1,mpoint)

      implicit none

      integer(long), intent(in) :: m,gdof,gwpdim1
      integer(long), intent(in) :: mpoint
      integer(long) :: n,f
      real(dop), dimension(maxdim), intent(out) :: gpoint
      real(dop), dimension(gdof,gwpdim1), intent(in) :: qcenter

      do n=1,nspfdof(m)
         f=spfdof(n,m)
         gpoint(f) = qcenter(n,mpoint)
      enddo

!
! add coordinates of point DVRs 
!
      if (lpntdvr) then
         do f=1,ndof
            if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
         enddo
      endif

      return
      end subroutine getqcent

!#######################################################################

      subroutine getgsmat(gstmp,gs2,gwpdim1,mpoint1,mpoint)

      implicit none

      integer(long), intent(in) :: gwpdim1,mpoint,mpoint1
      complex(dop), dimension(gwpdim1,gwpdim1), intent(in) :: gs2
      complex(dop), intent(out) :: gstmp

      gstmp = gs2(mpoint1,mpoint)

      return
      end subroutine getgsmat

!#######################################################################

      subroutine getggps(psi,tmppsi,ggp,vdim1,dim1,ndim1)

      implicit none

      integer(long), intent(in) :: vdim1,dim1,ndim1
      real(dop), dimension(vdim1,ndim1,dim1,dim1), intent(out) :: ggp
      complex(dop), intent(in), dimension(vdim1,dim1,ndim1)    :: psi,tmppsi
      integer(long) :: s,s1,v,n
      complex(dop)  :: ztmp

      do n=1,ndim1
         do s=1,dim1
            do s1=1,dim1
               do v=1,vdim1
                  ztmp = dconjg(psi(v,s1,n))*tmppsi(v,s,n)
                  ggp(v,n,s1,s) = dble(ztmp)
               enddo
            enddo
         enddo
      enddo

      return
      end subroutine getggps

!#######################################################################

      subroutine adpintegrate(psi,adiabpop,diapop,adiabden,&
                 subdimo,gdimo,subdimn,gdimn,gpop1,&
                 totgrid,hops,ort,time,gqi,adiabpop12,diapop12)

      implicit none

      integer(long)                                     :: f,m,i,s,n,whereiam,&
                                                           idum,n1,f1,nquick
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), intent(in)                         :: totgrid
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      real(dop), intent(in)                             :: time
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      real(dop), dimension(ortdim)                      :: dumweight
      real(dop),dimension(den1dim+den2dim),intent(inout):: adiabden
      real(dop), dimension(ortdim,nddstate*npacktot), intent(in) :: gpop1
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), intent(out)                         :: diapop12,adiabpop12


!-----------------------------------------------------------------------
! no weight for points
!-----------------------------------------------------------------------
      dumweight(:) = 1.0_dop

!-----------------------------------------------------------------------
! loop over grid. Get grid point.
!-----------------------------------------------------------------------
      nquick = 0
outer:do i = 1, totgrid
         if (lquick) then
            call cpvxi(subdimn,subdim,nmode)
            call cpvxi(gdimn,gdim,ndof)
         endif

         call decoder(i,mpoint,fpoint)

         if (lquick) then
            call cpvxi(subdimo,subdim,nmode)
            call cpvxi(gdimo,gdim,ndof)
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  fpoint(f) = fpoint(f) + gqi(f) - 1
               enddo
               f=spfdof(1,m)
               mpoint(m) = fpoint(f)
               do n=2,nspfdof(m)
                  f=spfdof(n,m)
                  idum=1
                  do n1=1,n-1
                     f1=spfdof(n1,m)
                     idum=idum*gdim(f1)
                  enddo
                  mpoint(m) = mpoint(m) + idum*(fpoint(f)-1)
               enddo
            enddo

            call guessden(gpop1,fpoint,lignore,qtol)
            if (lignore) cycle outer
            nquick=nquick+1
         endif

!-----------------------------------------------------------------------
! Calculation of WF at point
!-----------------------------------------------------------------------
         if (psitype .eq. 1) then
            call expop(hops,psi(zetf(1,1)),subdim(1),nddstate,&
                 psidia,psiad,mpoint(1),fpoint,ort)
         else if (psitype .eq. 0) then
            call wfpop(hops,psi,dgldim,nddstate,psidia,&
                 psiad,fpoint,mpoint,ort,agmat,dumweight)
         else    
            write(6,'(a)') ' psitype =',psitype,' not implemented!'
            write(2,'(a)') ' psitype =',psitype,' not implemented!'
            stop 1
         endif
     
!-----------------------------------------------------------------------
! calculate the state population
!-----------------------------------------------------------------------
         do s=1,nddstate
            adiabpop(s)=adiabpop(s)+dble(psiad(s)*dconjg(psiad(s)))
            diapop(s)=diapop(s)+dble(psidia(s)*dconjg(psidia(s)))
         enddo
         adiabpop12= adiabpop12 + psiad(2)*dconjg(psiad(1))
         diapop12  = diapop12   + psidia(2)*dconjg(psidia(1))

!----------------------------------------------------------------------- 
!  calculate the densities
!----------------------------------------------------------------------- 
         if(lden) then
            do f=1,dofcount
               do s=1,nddstate
                  adiabden(zeigden(f)+(s-1)*gdim(dofnumber(f))+&
                       fpoint(dofnumber(f))-1)=&
                       adiabden(zeigden(f)+(s-1)*gdim(dofnumber(f))+&
                       fpoint(dofnumber(f))-1)+&
                       dble(psiad(s)*dconjg(psiad(s)))
               enddo
            enddo
            
            do f=1,dof2dimcount
               do s=1,nddstate
                  whereiam=zeig2dimden(f)+(s-1)*&
                         gdim(dof2dimnumber(f,1))*&
                         gdim(dof2dimnumber(f,2))+&
                         gdim(dof2dimnumber(f,2))*&
                        (fpoint(dof2dimnumber(f,1))-1)+&
                         fpoint(dof2dimnumber(f,2))+den1dim
   
                  adiabden(whereiam)=adiabden(whereiam)+&
                         dble(psiad(s)*dconjg(psiad(s)))
               enddo
            enddo
         endif

! 99     continue
      enddo outer    ! i=1,totgrid

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      if (lquick) then
         write(ilog,'(f8.3,i10,a3,i10)')&
                   time,nquick,' / ',totgrid
      endif
      call myflush(ilog)

      return
      end subroutine adpintegrate

!#######################################################################

      subroutine cdvrintegrate0(psi,adiabpop,diapop,subdimo,gdimo,&
                 subdimn,gdimn,gqi,totgrid,hops,ort,&
                 qcenter,time)

      implicit none

      integer(long)                                     :: f,m,i,s,n,n1,e1,&
                                                           g,nham,v1,ierr,m1
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long) :: swapzeig,zeig,zeig1,zeig2,mblock1
      integer(long), dimension(:), allocatable          :: nignore
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      integer(long), intent(in)                         :: totgrid
      integer(long)                                     :: tdvrgrid,cdvrgrid
      integer(long)                                     :: ithr
      real(dop), intent(in)                             :: time 
      real(dop), dimension(qcentdim), intent(in)        :: qcenter
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(maxsta)                      :: diapop0,adiabpop0
      real(dop), dimension(:,:), allocatable            :: tmpadiabpop, &
                                                           tmpdiapop
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      real(dop), dimension(maxsta,maxdim)               :: adiabpopm
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(inout)       :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(:), allocatable           :: workc,psi1

      complex(dop), dimension(dmatdim)        :: cdvr2,cdvr4
      real(dop), dimension(c3matdim)          :: cdvr3
      real(dop)                               :: denpnt,norm

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock1 = block(1)/dim(meb,1)
      allocate(workc(2*block(1)))
      workc(:) = 0.0_dop

      allocate(tmpadiabpop(maxsta,ompthread))
      tmpadiabpop = 0.0_dop
      allocate(tmpdiapop(maxsta,ompthread))
      tmpdiapop = 0.0_dop
      allocate(nignore(ompthread))
      nignore(:) = 0

!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      adiabpop(:) = 0.0_dop
      diapop(:) = 0.0_dop
      adiabpop0(:) = 0.0_dop
      diapop0(:) = 0.0_dop

      fpoint(:)=0
      mpoint(:)=0

!-----------------------------------------------------------------------
! get TDDVR points
!-----------------------------------------------------------------------
      call getcdvr(psi,cdvr2,cdvr4,cdvr3,ort,1)

!-----------------------------------------------------------------------
! transform A coefficients from spf to TDVR/GWP basis
!   and save in zeig2
!-----------------------------------------------------------------------
      zeig=0
      zeig1=block(1)+1
      zeig2=1
      call cpvxz(psi(zpsi(1)),workc(zeig2),block(1))
      do m=1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         if (m .eq. meb) then
! if meb then transform from SPF basis to electronic state basis.
            call mtxxzz(psi(zetf(m,1)),workc(zeig1),workc(zeig2),&
                          vdim(m,1),dim(m,1),ndim(m,1),subdim(m))
         else 
! transform from SPF basis to TDDVR basis.
            call qtaxzz(cdvr4(dmat(m,1)),workc(zeig1), &
                 workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
         endif
      enddo

!-----------------------------------------------------------------------
! save PSI in TDDVR basis as psi1
!-----------------------------------------------------------------------
      allocate(psi1(block(1)))
      psi1(1:block(1)) = workc(zeig2:zeig2-1+block(1))

!-----------------------------------------------------------------------
! loop over TDDVR grid. Get grid point and transform
!-----------------------------------------------------------------------
      vdim1 = vdim(meb,1)
      dim1 = dim(meb,1)
      ndim1 = ndim(meb,1)
      nham=1

!$omp parallel default(shared) num_threads(ompthread) &
!$omp private(n1,v1,e1,i,denpnt,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(m,n,g,f,adiab,diab,ithr,fpoint)
!$omp do 
      do i=1,vdim1*ndim1
         call vendecoder(v1,e1,n1,i,vdim1,1)

#ifdef OMP
         ithr = omp_get_thread_num()+1
#else
         ithr = 1
#endif
         call adecoder(i,mpoint,s)

         call getwfpnt(psidia,psi1,v1,n1,vdim1,dim1,ndim1)

         if (lquick) then
            denpnt = 0.0
            do s=1,dim(meb,1)
               denpnt = denpnt + dble(dconjg(psidia(s))*psidia(s))
            enddo
            if (denpnt .lt. qtol) then
               nignore(ithr) = nignore(ithr) + 1
               cycle
            endif
         endif

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!
! for mD spfs, TDDVR grid point is the same for each DOF (simultaneous
! diagonalisation of X)
!-----------------------------------------------------------------------
         do m=1,nmode
            if (m .eq. meb) cycle

            if (gwpm(m)) then
               call getqcent(m,gpoint,qcenter(zcent(m,1)),nspfdof(m),&
                    dim(m,1),mpoint(m))

            else
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  g=mpoint(m)
                  zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                  gpoint(f) = cdvr3(zeig+g)
               enddo
            endif
         enddo

         fpoint(:)=0
         call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

!-----------------------------------------------------------------------
! calculate adiabatic WF
!-----------------------------------------------------------------------
         psiad(:)=0.0_dop
         call qvaxzz1(rotmat,psidia,psiad,maxsta,nddstate)

!-----------------------------------------------------------------------
! calculate the state population
!-----------------------------------------------------------------------
         do s=1,nddstate
            tmpadiabpop(s,ithr)=tmpadiabpop(s,ithr)+ &
                    dble(psiad(s)*conjg(psiad(s)))
            tmpdiapop(s,ithr)=tmpdiapop(s,ithr)+ &
                    dble(psidia(s)*conjg(psidia(s)))
         enddo

      enddo     
!$omp end do
!$omp end parallel

      norm = 0.0_dop
      do s=1,nddstate
         do ithr=1,ompthread
            adiabpop(s)=adiabpop(s)+tmpadiabpop(s,ithr)
            diapop(s)=diapop(s)+tmpdiapop(s,ithr)
         enddo
         norm = norm+dble(diapop(s))
      enddo
      do s=1,nddstate
         adiabpop(s)=adiabpop(s)/norm
         diapop(s)=diapop(s)/norm
      enddo

      deallocate(workc,tmpadiabpop,tmpdiapop)

      do ithr=2,ompthread
         nignore(1) = nignore(1) + nignore(ithr)
      enddo
      if (lquick) write(ilog,'(2(a,i10))') &
                'Ignored : ',nignore(1),' / ',vdim1*ndim1
      write(ilog,'(a,f10.6)') 'TDDVR norm : ',norm

!-----------------------------------------------------------------------
! CDVR correction for adiabatic populations (TDDVR exact for diabatic)
!-----------------------------------------------------------------------
      if (lcdvrint) then
         cdvrgrid=1

!-----------------------------------------------------------------------
! Store TDDVR result in adiabpop0
!-----------------------------------------------------------------------
         do s=1,nddstate
            adiabpop0(s) = adiabpop(s)
         enddo

!----------------------------------------------------------------------- 
! calculate correction one mode at a time
! contains Psi multiplied by GWP overlaps
!----------------------------------------------------------------------- 
         do m1=1,nmode
            if (m1 .eq. meb) cycle
            call cdvrintm0(m1,psi,psi1,cdvr3,cdvr4,&
                    adiabpopm(1,m1),subdimo,gdimo,subdimn,gdimn,gqi,&
                    hops,ort,mblock1,dim(meb,1),cdvrgrid,ierr)
         enddo

!-----------------------------------------------------------------------
! Add CDVR correction to populations
!-----------------------------------------------------------------------
         do s=1,nddstate
            do m=1,nmode
               if (m .eq. meb) cycle
               adiabpop(s)=adiabpop(s)+(adiabpopm(s,m) - adiabpop0(s))
            enddo
         enddo

      endif  ! End of CDVR

!-----------------------------------------------------------------------
! Restore dimensions
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle
         subdim(m) = subdimo(m)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f) = gdimo(f)
         enddo
      enddo

      deallocate(psi1)

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      write(ilog,'(f8.3,i10,2(a3,i10))')&
                   time, tdvrgrid,' / ',cdvrgrid,' / ',totgrid
      call myflush(ilog)

      return
      end subroutine cdvrintegrate0

!#######################################################################
!
! It is assumed that meb is the last mode (standard Quantics)
!
!#######################################################################

      subroutine cdvrintm0(m1,psi,psi1,cdvr3,cdvr4,adiabpopm,subdimo,&
                 gdimo,subdimn,gdimn,gqi,hops,ort,mblock1,eldim,ierr,cdvrgrid)

      implicit none

      integer(long), intent(in)                         :: mblock1,m1,eldim
      integer(long)                                     :: f,m,i,n,n1,&
                                                           g,nham,v1,e,ns,&
                                                           gg,vg,ng
      integer(long)                                     :: vdim1,dim1,ndim1,&
                                                           subdimm1
      integer(long)                                     :: ithr
      integer(long), intent(inout)                      :: ierr
      integer(long)                                     :: zeig
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      integer(long)                                     :: cdvrgrid
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(out)         :: adiabpopm
      real(dop), dimension(:,:), allocatable            :: tmpadiabpopm, &
                                                           tmpdiapop 
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(mblock1,eldim), intent(in) :: psi1
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(maxspf,nddstate)          :: psipnt
      complex(dop), dimension(sdim*maxspf)              :: tmpspf,psi4
      complex(dop), dimension(sdim,nddstate)            :: psi2

      complex(dop), dimension(dmatdim), intent(in) :: cdvr4
      real(dop), dimension(c3matdim), intent(in)   :: cdvr3
      real(dop)                                    :: denpnt,norm
      logical(kind=4) :: lignore

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      allocate(tmpadiabpopm(maxsta,ompthread),stat=ierr)
      if (ierr .gt. 0) return
      allocate(tmpdiapop(maxsta,ompthread))
      if (ierr .gt. 0) return

      psi2(:,:)=0.0_dop
      psi4(:)=0.0_dop
      tmpspf(:)=0.0_dop

      tmpadiabpopm = 0.0_dop
      tmpdiapop = 0.0_dop
      adiabpopm = 0.0_dop

!-----------------------------------------------------------------------
! Get effective grid sizes in CDVR rep
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif
      do m=1,nmode 
         if (m .eq. m1) cycle
         subdim(m)=dim(m,1)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f)=dim(m,1)
         enddo
      enddo

      ns = dim(meb,1)

!-----------------------------------------------------------------------
! Set up CDVR functions on grid
!-----------------------------------------------------------------------
      if (lquick) then
         call cut_spf_grid(m1,psi(zetf(m1,1)),tmpspf,subdimo(m1),&
              subdimn(m1),dim(m1,1),gdimo,gdimn,gqi)
      else
         call cpmxz(psi(zetf(m1,1)),tmpspf,subdimo(m1),dim(m1,1))
      endif
      call mmxxzz(tmpspf,cdvr4(dmat(m1,1)),psi4,&
           subdim(m1),dim(m1,1),dim(m1,1))

!-----------------------------------------------------------------------
! now operate on TDDVR vector by ADT, expanding each point onto the
! primitive grid for mode m1.
!-----------------------------------------------------------------------
      vdim1 = vdim(m1,1)
      dim1 = dim(m1,1)
      ndim1 = ndim(m1,1)
      ndim1 = (ndim1/ns)
      subdimm1 = subdim(m1)
      subdim(m1) = 1
      nham=1

!$omp parallel default(shared) num_threads(ompthread) &
!$omp private(n1,v1,i,denpnt,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(m,n,g,f,adiab,diab,ithr,psipnt,psi2,lignore,gg,zeig) &
!$omp private(fpoint)
!$omp do schedule(dynamic)
      do i=1,vdim1*ndim1
         call vendecoder(v1,e,n1,i,vdim1,1)

#ifdef OMP
         ithr = omp_get_thread_num()+1
#else
         ithr = 1
#endif
!-----------------------------------------------------------------------
! Extract WF at single point 
!-----------------------------------------------------------------------
         do s=1,ns
            call cuttxz1(psi1(1,s),psipnt(1,s),vdim1,dim1,ndim1,v1,n1)
         enddo

         if (lquick) then
            lignore = .true.
            do e=1,dim1
               denpnt = 0.0_dop
               do s=1,ns
                  denpnt = denpnt + dble(dconjg(psipnt(e,s))*psipnt(e,s))
                  if (denpnt .lt. qtol) then
                     psipnt(e,s) = 0.0_dop
                  else
                     lignore = .false.
                  endif
               enddo
            enddo
            if (lignore) cycle
         endif

!-----------------------------------------------------------------------
! Transform from TDDVR to CDVR at point
!-----------------------------------------------------------------------
         do s=1,ns
            call mvxxzz(psi4,psipnt(1,s),psi2(1,s),subdimm1,dim1)
         enddo

!-----------------------------------------------------------------------
! get grid point coordinates for TDDVR modes
!-----------------------------------------------------------------------
         call decoder(i,mpoint,fpoint)
         do m=1,nmode
            if (m .eq. meb .or. m .eq. m1) cycle

            do n=1,nspfdof(m)
               f=spfdof(n,m)
               g=mpoint(m)
               zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
               gpoint(f) = cdvr3(zeig+g)
            enddo
         enddo

!-----------------------------------------------------------------------
! Loop over grid of CDVR mode and transform Psi
!-----------------------------------------------------------------------
         do gg = 1,subdimm1
            psidia(1:ns) = psi2(gg,1:ns)
         !   if (lquick) then
         !      denpnt = 0.0
         !      do s=1,ns
         !         denpnt = denpnt + dble(dconjg(psidia(s))*psidia(s))
         !      enddo
         !      if (denpnt .lt. 0.001*qtol) cycle
         !   endif

!-----------------------------------------------------------------------
! get grid point coordinates for m1 and calculate transformation matrix
!-----------------------------------------------------------------------
            do n=1,nspfdof(m1)
               f=spfdof(n,m1)
               call vendecoder(vg,g,ng,gg,vgdim(f),gdim(f))
               if (lquick) g = g + gqi(f) - 1
               gpoint(f) = ort(zort(f)-1+g)
            enddo
            call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

!-----------------------------------------------------------------------
! calculate adiabatic WF (stored in psi3)
!-----------------------------------------------------------------------
            psiad(:)=0.0_dop
            call qvaxzz1(rotmat,psidia,psiad,maxsta,nddstate)

            do s=1,nddstate
               tmpadiabpopm(s,ithr)=tmpadiabpopm(s,ithr)+&
                       dble(psiad(s)*conjg(psiad(s)))
               tmpdiapop(s,ithr)=tmpdiapop(s,ithr)+&
                       dble(psidia(s)*conjg(psidia(s)))
            enddo
         enddo
      enddo
!$omp end do
!$omp end parallel
      subdim(m1) = subdimm1

      norm=0.0
      do s=1,nddstate
         do ithr=1,ompthread
            adiabpopm(s)=adiabpopm(s)+tmpadiabpopm(s,ithr)
            norm=norm+tmpdiapop(s,ithr)
         enddo
      enddo
      if (norm .lt. 1.0d-8) then
         if (.not. lbatch) &
            write(6,'(a)') 'No density found in CDVR (norm ~0). Use smaller qtol?'
         stop
      endif

      do s=1,nddstate
         adiabpopm(s)=adiabpopm(s)/norm
      enddo

      deallocate(tmpadiabpopm)

      return
      end subroutine cdvrintm0 

!#######################################################################

      subroutine cdvrintm0_old(m1,psi,psi1,cdvr3,cdvr4,adiabpopm,subdimo,&
                 gdimo,subdimn,gdimn,gqi,hops,ort,mblock1,ierr,cdvrgrid)

      implicit none

      integer(long), intent(in)                         :: mblock1,m1
      integer(long)                                     :: dimpsi4
      integer(long)                                     :: f,m,i,n,n1,&
                                                           g,nham,v1,e1
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long)                                     :: ithr
      integer(long), intent(inout)                      :: ierr
      integer(long) :: zeig,mblock2
      integer(long)                                     :: effgrid
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      integer(long)                                     :: cdvrgrid
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(out)         :: adiabpopm
      real(dop), dimension(:,:), allocatable            :: tmpadiabpopm
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(mblock1), intent(in)      :: psi1
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(:), allocatable     :: tmpspf,psi2,psi4

      complex(dop), dimension(dmatdim), intent(in) :: cdvr4
      real(dop), dimension(c3matdim), intent(in)   :: cdvr3
      real(dop)                                    :: denpnt

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock2 = block(1)
      effgrid=block(1)/dim(m1,1)
      if (lquick) then
         effgrid=effgrid*subdimn(m1)
      else 
         effgrid=effgrid*subdimo(m1)
      endif
      mblock2=max(mblock2,effgrid)
      cdvrgrid = max(cdvrgrid,effgrid)
      allocate(psi2(mblock2),stat=ierr)
      if (ierr .gt. 0) return

      if (lquick) then
         dimpsi4= subdimn(m1)*dim(m1,1)
      else 
         dimpsi4= subdimo(m1)*dim(m1,1)
      endif
      allocate(psi4(dimpsi4),stat=ierr)
      if (ierr .gt. 0) return

      allocate(tmpspf(sdim*maxspf),stat=ierr)
      if (ierr .gt. 0) return

      allocate(tmpadiabpopm(maxsta,ompthread))
      tmpadiabpopm = 0.0_dop

      psi2(:)=0.0_dop
      psi4(:)=0.0_dop
      tmpspf(:)=0.0_dop

      adiabpopm = 0.0_dop

!-----------------------------------------------------------------------
! Get grid sizes
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif

!-----------------------------------------------------------------------
! Set up CDVR functions on grid
!-----------------------------------------------------------------------
      if (lquick) then
         call cut_spf_grid(m1,psi(zetf(m1,1)),tmpspf,subdimo(m1),&
              subdimn(m1),dim(m1,1),gdimo,gdimn,gqi)
      else
         call cpmxz(psi(zetf(m1,1)),tmpspf,subdimo(m1),dim(m1,1))
      endif
      call mmxxzz(tmpspf,cdvr4(dmat(m1,1)),psi4,&
           subdim(m1),dim(m1,1),dim(m1,1))

!-----------------------------------------------------------------------
! transform from TDVR basis to DVR points for mode m1
! block1 is the length of the transformed array
!-----------------------------------------------------------------------
      call mtxxzz(psi4,psi1,psi2,vdim(m1,1),dim(m1,1),ndim(m1,1),&
           subdim(m1))

!-----------------------------------------------------------------------
! Get effective grid sizes in CDVR rep
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif
      do m=1,nmode 
         if (m .eq. m1) cycle
         subdim(m)=dim(m,1)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f)=dim(m,1)
         enddo
      enddo

!-----------------------------------------------------------------------
! now operate on vector by V(q1,.xm...qf)
! by recursively dividing out the number of grid points for each
! mode, the gridpoint gpnt is transformed to the mode grid point mgpnt
!-----------------------------------------------------------------------
      vdim1 = vdim(meb,1)
      dim1 = dim(meb,1)
      ndim1 = ndim(meb,1)
      if (m1 .lt. meb) then
         vdim1 = (vdim1/dim(m1,1))*subdim(m1)
      else
         ndim1 = (ndim1/dim(m1,1))*subdim(m1)
      endif

!$omp parallel num_threads(ompthread) &
!$omp private(n1,v1,i,denpnt,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(m,n,g,f,adiab,diab,ithr)
!$omp do 
      do i=1,vdim1*ndim1
         call vendecoder(v1,e1,n1,i,vdim1,1)
#ifdef OMP
         ithr = omp_get_thread_num()+1
#else
         ithr = 1
#endif
         call decoder(i,mpoint,fpoint)

         call getwfpnt(psidia,psi2,v1,n1,vdim1,dim1,ndim1)

         if (lquick) then
            denpnt = 0.0
            do s=1,dim(meb,1)
               denpnt = denpnt + dble(dconjg(psidia(s))*psidia(s))
            enddo
            if (denpnt .lt. qtol) cycle
         endif

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!-----------------------------------------------------------------------
         do m=1,nmode
            if (m .eq. meb) cycle

            if (m .eq. m1) then
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  g=fpoint(f)
                  if (lquick) g = g + gqi(f) - 1
                  gpoint(f) = ort(zort(f)-1+g)
               enddo
            else
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  g=mpoint(m)
                  zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                  gpoint(f) = cdvr3(zeig+g)
               enddo
            endif
         enddo
         nham=1
         call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

!-----------------------------------------------------------------------
! calculate adiabatic WF (stored in psi3)
!-----------------------------------------------------------------------
         psiad(:)=0.0_dop
         call qvaxzz1(rotmat,psidia,psiad,maxsta,nddstate)

         do s=1,nddstate
            tmpadiabpopm(s,ithr)=tmpadiabpopm(s,ithr)+&
                       dble(psiad(s)*conjg(psiad(s)))
         enddo
      enddo
!$omp end do
!$omp end parallel

      do s=1,nddstate
         do ithr=1,ompthread
            adiabpopm(s)=adiabpopm(s)+tmpadiabpopm(s,ithr)
         enddo
      enddo

      deallocate(psi2,psi4,tmpspf,tmpadiabpopm)

      return
      end subroutine cdvrintm0_old

!#######################################################################

      subroutine cdvrintegrate(psi,adiabpop,diapop,subdimo,gdimo,&
                 subdimn,gdimn,gqi,totgrid,hops,ort,&
                 qcenter,time)

      implicit none

      integer(long)                                     :: f,m,i,s,n,n1,e1,&
                                                           g,nham,v1,ierr,&
                                                           s1,si,sf,m1
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long) :: swapzeig,zeig,zeig1,zeig2,mblock1
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      integer(long), intent(in)                         :: totgrid
      integer(long)                                     :: tdvrgrid,cdvrgrid
      real(dop), intent(in)                             :: time 
      real(dop), dimension(qcentdim), intent(in)        :: qcenter
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(maxsta)                      :: diapop0,adiabpop0
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      real(dop), dimension(maxsta,maxdim)               :: adiabpopm
      complex(dop), dimension(nddstate,nddstate)        :: dimat
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(inout)       :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(:), allocatable     :: workc,psi1,psi2
      complex(dop), dimension(maxsta,maxsta)      :: opmat
      complex(dop)                                :: ztmp

      complex(dop), dimension(dmatdim)        :: cdvr2,cdvr4
      real(dop), dimension(c3matdim)          :: cdvr3

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock1 = block(1)
      allocate(workc(2*mblock1))
      workc(:) = 0.0_dop

!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      adiabpop(:) = 0.0_dop
      diapop(:) = 0.0_dop
      adiabpop0(:) = 0.0_dop
      diapop0(:) = 0.0_dop

      fpoint(:)=0
      mpoint(:)=0

!-----------------------------------------------------------------------
! get CDVR points
!-----------------------------------------------------------------------
      call getcdvr(psi,cdvr2,cdvr4,cdvr3,ort,1)

!-----------------------------------------------------------------------
! transform A coefficients from spf to TDVR/GWP basis
!   and save in zeig2
!-----------------------------------------------------------------------
      zeig=0
      zeig1=mblock1+1
      zeig2=1
      call cpvxz(psi(zpsi(1)),workc(zeig2),block(1))
      do m=1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         if (m .eq. meb) then
! if meb then transform from SPF basis to electronic state basis.
            call mtxxzz(psi(zetf(m,1)),workc(zeig1),workc(zeig2),&
                          vdim(m,1),dim(m,1),ndim(m,1),subdim(m))
         else if (gwpm(m) .and. psitype .eq. 0) then
! if gwp-spf then transform from SPF basis to GWP basis.
            call qtxxzz(psi(zdgwp(m,1)),workc(zeig1), &
                 workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
         else if (.not. gwpm(m)) then
! transform from SPF basis to TDDVR basis.
            call qtaxzz(cdvr4(dmat(m,1)),workc(zeig1), &
                 workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
         else
! do nothing
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
         endif
      enddo

!-----------------------------------------------------------------------
! save PSI in TDDVR basis as psi1
! psi2 is PSI in adiabatic basis
!-----------------------------------------------------------------------
      allocate(psi1(mblock1))
      allocate(psi2(mblock1))
      psi1(1:mblock1) = workc(zeig2:zeig2-1+mblock1)

!-----------------------------------------------------------------------
! loop over adiabatic states
!-----------------------------------------------------------------------
      do s1=1,nddstate

!-----------------------------------------------------------------------
! loop over TDDVR grid. Get grid point and transform
!-----------------------------------------------------------------------
         vdim1 = vdim(meb,1)
         dim1 = dim(meb,1)
         ndim1 = ndim(meb,1)
!$omp parallel num_threads(ompthread) &
!$omp private(n1,v1,e1,i,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(opmat,m,n,g,f,adiab,diab,sf,si)
!$omp do
         do i=1,vdim1*ndim1
            call vendecoder(v1,e1,n1,i,vdim1,1)

            call adecoder(i,mpoint,s)

            call getwfpnt(psidia,psi1,v1,n1,vdim1,dim1,ndim1)

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!
! for mD spfs, TDDVR grid point is the same for each DOF (simultaneous
! diagonalisation of X)
!-----------------------------------------------------------------------
            do m=1,nmode
               if (m .eq. meb) cycle

               if (gwpm(m)) then
                  call getqcent(m,gpoint,qcenter(zcent(m,1)),nspfdof(m),&
                       dim(m,1),mpoint(m))

               else
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     g=mpoint(m)
                     zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                     gpoint(f) = cdvr3(zeig+g)
                  enddo
               endif
            enddo
            if (lpntdvr) then
               do f=1,ndof
                  if (basis(f) .eq. 19) gpoint(f) = rpbaspar(1,f)
               enddo
            endif

            nham=1
            fpoint(:)=0
            call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

            do si=1,dim(meb,1)
               do sf=1,dim(meb,1)
                   opmat(sf,si) = rotmat(sf,s1)*dconjg(rotmat(si,s1))
               enddo
            enddo
     
!-----------------------------------------------------------------------
! calculate adiabatic WF
!-----------------------------------------------------------------------
            psiad(:)=0.0_dop
            call qvtxzz1(opmat,psidia,psiad,maxsta,nddstate)

            call putwfpnt(psiad,psi2,v1,n1,vdim1,dim1,ndim1)
         enddo     
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! calculate the state populations
!-----------------------------------------------------------------------
         zeig=0
         zeig1=mblock1+1
         zeig2=1
         call cpvxz(psi1(zpsi(1)),workc(zeig2),block(1))
         if (lgwp) then
            do m=1,nmode
               if (.not. gwpm(m)) cycle
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call qtxxzz(gs2(gs2mat(m,1)),workc(zeig1), &
                    workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
            enddo
         endif

! adiabatic populations
         call vvaxzz(workc(zeig2),psi2,ztmp,block(1))
         adiabpop(s1) = dble(ztmp)

      enddo
! diabatic populations
      call rmqxxxzz(workc(zeig2),psi1,dimat,vdim1,dim1,ndim1)
      do s=1,nddstate
         diapop(s) = dble(dimat(s,s))
      enddo

!-----------------------------------------------------------------------
! CDVR correction for adiabatic populations (TDDVR exact for diabatic)
!-----------------------------------------------------------------------
      if (lcdvrint) then
         cdvrgrid=1

         if (lgwp) then
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  gdim(f) = gdimo(f)
               enddo
            enddo
            call getgwpgrids(ort)
         endif
       
!----------------------------------------------------------------------- 
! calculate correction one mode at a time
! workc(zeig2) contains Psi multiplied by GWP overlaps
!----------------------------------------------------------------------- 
         do m1=1,nmode
            if (m1 .eq. meb) cycle
            do s=1,nddstate
               if (gwpm(m1)) then
                  workc(:)=0.0_dop
                  zeig=0
                  zeig1=mblock1+1
                  zeig2=1
                  call cpvxz(psi1(zpsi(1)),workc(zeig2),block(1))
                  do m=1,nmode
                     if (.not. gwpm(m)) cycle
                     if (m .eq. m1) cycle
                     swapzeig=zeig1
                     zeig1=zeig2
                     zeig2=swapzeig
                     call qtxxzz(gs2(gs2mat(m,1)),workc(zeig1), &
                          workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
                  enddo
                  call cdvrintg(m1,s,psi1,workc(zeig2),cdvr3,adiabpopm(s,m1),&
                       gdimo,gdimn,gqi,hops,ort,mblock1,ierr,&
                       qcenter)
               else       
                  call cdvrintm(m1,s,psi,psi1,workc(zeig2),cdvr3,cdvr4,&
                       adiabpopm(s,m1),&
                       subdimo,gdimo,subdimn,gdimn,gqi,hops,ort,mblock1,&
                       cdvrgrid,ierr)
               endif
            enddo
         enddo

!-----------------------------------------------------------------------
! Add CDVR correction to populations
!
! Store TDDVR result in adiabpop0
!-----------------------------------------------------------------------
         do s=1,nddstate
            adiabpop0(s) = adiabpop(s)
         enddo

         do s=1,nddstate
            do m=1,nmode
               if (m .eq. meb) cycle
               if (gwpm(m)) then
! For GWPs there is a correction for each DOF
                  adiabpop(s)=adiabpop(s)+ &
                      (adiabpopm(s,m) - nspfdof(m)*adiabpop0(s))
               else
                  adiabpop(s)=adiabpop(s)+(adiabpopm(s,m) - adiabpop0(s))
               endif
            enddo
         enddo

      endif  ! End of CDVR

!-----------------------------------------------------------------------
! Restore dimensions
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle
         subdim(m) = subdimo(m)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f) = gdimo(f)
         enddo
      enddo

      deallocate(psi1,workc)

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      write(ilog,'(f8.3,i10,2(a3,i10))')&
                   time, tdvrgrid,' / ',cdvrgrid,' / ',totgrid
      call myflush(ilog)

      return
      end subroutine cdvrintegrate

!#######################################################################

      subroutine cdvrintm(m1,sa,psi,psi1,psig,cdvr3,cdvr4,adiabpopm,subdimo,&
                 gdimo,subdimn,gdimn,gqi,hops,ort,mblock1,ierr,cdvrgrid)

      implicit none

      integer(long), intent(in)                         :: mblock1,m1,sa
      integer(long)                                     :: dimpsi4
      integer(long)                                     :: f,m,i,n,n1,e1,&
                                                           g,nham,v1,si,sf
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long), intent(inout)                      :: ierr
      integer(long) :: zeig,mblock2
      integer(long)                                     :: effgrid
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn
      integer(long)                                     :: cdvrgrid
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), intent(out)                            :: adiabpopm
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat,opmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(mblock1), intent(in)      :: psi1,psig
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(:), allocatable     :: tmpspf,psi2,psi3,psi4
      complex(dop)                                :: ztmp

      complex(dop), dimension(dmatdim), intent(in) :: cdvr4
      real(dop), dimension(c3matdim), intent(in)   :: cdvr3

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock2 = block(1)
      effgrid=block(1)/dim(m1,1)
      if (lquick) then
         effgrid=effgrid*subdimn(m1)
      else 
         effgrid=effgrid*subdimo(m1)
      endif
      mblock2=max(mblock2,effgrid)
      cdvrgrid = max(cdvrgrid,effgrid)
      allocate(psi2(mblock2),stat=ierr)
      if (ierr .gt. 0) return
      allocate(psi3(mblock2),stat=ierr)
      if (ierr .gt. 0) return

      if (lquick) then
         dimpsi4= subdimn(m1)*dim(m1,1)
      else 
         dimpsi4= subdimo(m1)*dim(m1,1)
      endif
      allocate(psi4(dimpsi4),stat=ierr)
      if (ierr .gt. 0) return

      allocate(tmpspf(sdim*maxspf),stat=ierr)
      if (ierr .gt. 0) return

      psi2(:)=0.0_dop
      psi3(:)=0.0_dop
      psi4(:)=0.0_dop
      tmpspf(:)=0.0_dop

      adiabpopm = 0.0_dop

!-----------------------------------------------------------------------
! Get grid sizes
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif

!-----------------------------------------------------------------------
! Set up CDVR functions on grid
!-----------------------------------------------------------------------
      if (lquick) then
         call cut_spf_grid(m1,psi(zetf(m1,1)),tmpspf,subdimo(m1),&
              subdimn(m1),dim(m1,1),gdimo,gdimn,gqi)
         call mmxxzz(tmpspf,cdvr4(dmat(m1,1)),psi4,&
              subdimn(m1),dim(m1,1),dim(m1,1))
      else
         call cpmxz(psi(zetf(m1,1)),tmpspf,subdimo(m1),dim(m1,1))
      endif
      call mmxxzz(psi(zetf(m1,1)),cdvr4(dmat(m1,1)),psi4,&
           subdim(m1),dim(m1,1),dim(m1,1))

!-----------------------------------------------------------------------
! transform from TDVR basis to DVR points for mode m1
! block1 is the length of the transformed array
!-----------------------------------------------------------------------
      call mtxxzz(psi4,psi1,psi2,vdim(m1,1),dim(m1,1),ndim(m1,1),&
           subdim(m1))

!-----------------------------------------------------------------------
! Get effective grid sizes in CDVR rep
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif
      do m=1,nmode 
         if (m .eq. m1) cycle
         subdim(m)=dim(m,1)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f)=dim(m,1)
         enddo
      enddo

!-----------------------------------------------------------------------
! now operate on vector by V(q1,.xm...qf)
! by recursively dividing out the number of grid points for each
! mode, the gridpoint gpnt is transformed to the mode grid point mgpnt
!-----------------------------------------------------------------------
      vdim1 = vdim(meb,1)
      dim1 = dim(meb,1)
      ndim1 = ndim(meb,1)
      if (m1 .lt. meb) then
         vdim1 = (vdim1/dim(m1,1))*subdim(m1)
      else
         ndim1 = (ndim1/dim(m1,1))*subdim(m1)
      endif

!$omp parallel num_threads(ompthread) &
!$omp private(n1,v1,e1,i,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(opmat,m,n,g,f,adiab,diab)
!$omp do
      do i=1,vdim1*ndim1
            call vendecoder(v1,e1,n1,i,vdim1,1)

            call decoder(i,mpoint,fpoint)

            call getwfpnt(psidia,psi2,v1,n1,vdim1,dim1,ndim1)

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!-----------------------------------------------------------------------
            do m=1,nmode
               if (m .eq. meb) cycle

               if (m .eq. m1) then
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     g=fpoint(f)
                     if (lquick) g = g + gqi(f) - 1
                     gpoint(f) = ort(zort(f)-1+g)
                  enddo
               else
                  do n=1,nspfdof(m)
                     f=spfdof(n,m)
                     g=mpoint(m)
                     zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                     gpoint(f) = cdvr3(zeig+g)
                  enddo
               endif
            enddo
            nham=1
            call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

            do si=1,dim(meb,1)
               do sf=1,dim(meb,1)
                   opmat(sf,si) = rotmat(sf,sa)*dconjg(rotmat(si,sa))
               enddo
            enddo

!-----------------------------------------------------------------------
! calculate adiabatic WF (stored in psi3)
!-----------------------------------------------------------------------
            psiad(:)=0.0_dop
            call qvtxzz1(opmat,psidia,psiad,maxsta,nddstate)

            call putwfpnt(psiad,psi3,v1,n1,vdim1,dim1,ndim1)
      enddo
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! Transform back to TDDVR basis
!-----------------------------------------------------------------------
      psi2(:) = 0.0_dop
      call mtaxzz(psi4,psi3,psi2,vdim(m1,1),subdim(m1),ndim(m1,1),&
           dim(m1,1))

!-----------------------------------------------------------------------
! calculate the state populations
!-----------------------------------------------------------------------
      call vvaxzz(psig,psi2,ztmp,block(1))
      adiabpopm = dble(ztmp)

      deallocate(psi2,psi3,psi4,tmpspf)

      return
      end subroutine cdvrintm 

!#######################################################################

      subroutine cdvrintg(m1,sa,psi1,psig,cdvr3,adiabpopm,&
                 gdimo,gdimn,gqi,hops,ort,mblock1,ierr,qcenter)

      implicit none

      integer(long), intent(in)                         :: mblock1,m1,sa
      integer(long), intent(inout)                      :: ierr
      integer(long)                                     :: f,m,i,n,n1,j,&
                                                           g,nham,v1,f1,n2,e1
      integer(long)                                     :: g1,ve,ee,ne,e2,si,sf,gg
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long) :: zeig
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,gdimo,gdimn
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), intent(out)                            :: adiabpopm
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      real(dop), dimension(qcentdim), intent(in)        :: qcenter
      complex(dop)                                      :: ztmp
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      complex(dop), dimension(mblock1), intent(in)      :: psi1,psig
      complex(dop), dimension(maxsta,maxspf)            :: psidia,psiad
      complex(dop), dimension(:), allocatable           :: psi4,psi2,psigd
      complex(dop), dimension(:,:,:,:), allocatable   :: tddvrmat
      complex(dop), dimension(:,:,:,:), allocatable  :: opmat

      real(dop), dimension(c3matdim), intent(in)   :: cdvr3

!-----------------------------------------------------------------------
! Allocate work arrays
!
! psi1: Psi in TDDVR basis
!
! psi4: SPFs in TDDVR rep
!-----------------------------------------------------------------------
      allocate(psi4(maxspf*maxspf),stat=ierr)
      if (ierr .gt. 0) return
      allocate(psigd(maxgdim*maxspf),stat=ierr)
      if (ierr .gt. 0) return
      if (ierr .gt. 0) return
      allocate(psi2(mblock1))
      psi2(:)=0.0_dop
      psi4(:)=0.0_dop

      adiabpopm = 0.0_dop
      do m=1,nmode
         if (m .eq. meb) cycle
         subdim(m)=dim(m,1)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f) = dim(m,1)
         enddo
      enddo

      allocate(tddvrmat(dim(m1,1),dim(m1,1),dim(meb,1),dim(meb,1)))
!-----------------------------------------------------------------------
! Loop over DOFs
!-----------------------------------------------------------------------
      do n1=1,nspfdof(m1)

         f1=spfdof(n1,m1)
         if (lquick) then
            gdim(f1)=gdimn(f1)
         else
            gdim(f1)=gdimo(f1)
         endif
         allocate(opmat(gdim(f1),dim(m1,1),dim(meb,1),dim(meb,1)))

         call gwpf_on_grid(n1,m1,psigd,gdim(f1),dim(m1,1),ort,gqi)

!-----------------------------------------------------------------------
! now operate on vector by V(q1,.xm...qf)
! by recursively dividing out the number of grid points for each
! mode, the gridpoint gpnt is transformed to the mode grid point mgpnt
!-----------------------------------------------------------------------
         vdim1 = vdim(m1,1)
         dim1 = dim(m1,1)
         ndim1 = ndim(m1,1)
         if (m1 .lt. meb) then
            ndim1 = ndim(m1,1)/dim(meb,1)
         else
            vdim1 = vdim(m1,1)/dim(meb,1)
         endif
!$omp parallel num_threads(ompthread) &
!$omp private(n2,v1,e1,i,j,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(tddvrmat,opmat,m,n,g,f,adiab,diab,si,sf,g1,gg,ve,ee,ne)
!$omp do
         do i=1,vdim1*ndim1
            call vendecoder(v1,e1,n2,i,vdim1,1)

               call adecoder(i,mpoint,si)

               do e1=1,dim1   ! Loop over functions of m1 to get state dependent WF 
                  mpoint(m1) = e1
                  j = v1 + (e1-1)*vdim1 + (n2-1)*vdim1*dim1
                  call vendecoder(ve,ee,ne,j,vdim(meb,1),dim(meb,1))
                  call getwfpnt(psidia(1,e1),psi1,ve,ne,vdim(meb,1),&
                       dim(meb,1),ndim(meb,1))
               enddo

!-----------------------------------------------------------------------
! get TDDVR coordinates 
!-----------------------------------------------------------------------
               tddvrmat=0.0_dop
               opmat=0.0_dop
               do e1=1,dim1   ! Loop over functions of m1
                  mpoint(m1) = e1

                  do m=1,nmode
                     if (m .eq. meb) cycle

                     if (gwpm(m)) then
                        call getqcent(m,gpoint,qcenter(zcent(m,1)),nspfdof(m),&
                             dim(m,1),mpoint(m))

                     else
                        do n=1,nspfdof(m)
                           f=spfdof(n,m)
                           g=mpoint(m)
                           zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                           gpoint(f) = cdvr3(zeig+g)
                        enddo
                     endif
                  enddo

                  do g1 = 1,gdim(f1)
                     gg=g1
                     if (lquick) gg = gg + gqi(f1) - 1
                     gpoint(f1) = ort(zort(f1)-1+gg) 
                     nham=1
                     call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)
 
! build operators at each point
                     do si=1,dim(meb,1)
                        do sf=1,dim(meb,1)
                            opmat(gg,e1,sf,si) = &
                                   rotmat(sf,sa)*dconjg(rotmat(si,sa))
                        enddo
                     enddo
                  enddo
               enddo

! now build operator matrices
               do si=1,dim(meb,1)
                  do sf=1,dim(meb,1)
                     call buildopmat(tddvrmat(1,1,sf,si),&
                          opmat(1,1,sf,si),psigd,gs2(gs2mat(m1,1)),&
                          gdim(f1),dim(m1,1))
                  enddo
               enddo

               psiad(:,:)=0.0_dop
               do e1=1,dim1   ! Loop over functions of m1
                  do e2=1,dim1
                     do si=1,dim(meb,1)
                        do sf=1,dim(meb,1)
                           psiad(sf,e2) = psiad(sf,e2) + &
                                 tddvrmat(e2,e1,sf,si)*psidia(si,e1)
                        enddo
                     enddo
                  enddo
               enddo

               do e1=1,dim1   ! Loop over functions of m1
                  j = v1 + (e1-1)*vdim1 + (n2-1)*vdim1*dim1
                  call vendecoder(ve,ee,ne,j,vdim(meb,1),dim(meb,1))
                  call putwfpnt(psiad(1,e1),psi2,ve,ne,vdim(meb,1),dim(meb,1),ndim(meb,1))
               enddo
            
         enddo 
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! calculate the state populations
! (sum over DOF contributions)
!-----------------------------------------------------------------------
         call vvaxzz(psig,psi2,ztmp,block(1))
         adiabpopm = adiabpopm + dble(ztmp)

         deallocate(opmat)
      enddo

      deallocate(tddvrmat)
      deallocate(psi4,psi2,psigd)

      return
      end subroutine cdvrintg

!#######################################################################     

      subroutine buildopmat(tddvrmat,opmat,psig,gsm,gdim1,dim1)

      implicit none

      integer(long), intent(in) :: gdim1,dim1
      integer(long)             :: e,e1,g
      complex(dop), dimension(dim1,dim1), intent(out)  :: tddvrmat
      complex(dop), dimension(gdim1,dim1), intent(out) :: opmat
      complex(dop), dimension(dim1,dim1), intent(in)   :: gsm
      complex(dop), dimension(gdim1,dim1), intent(in)  :: psig
      complex(dop), dimension(dim1,dim1)               :: ovl

      ovl=0.0_dop
      do e=1,dim1
         do e1=1,dim1
            do g=1,gdim1
               ovl(e1,e) = ovl(e1,e) + dconjg(psig(g,e1))*psig(g,e)
            enddo
            ovl(e1,e) = gsm(e1,e)/ovl(e1,e)
         enddo
      enddo

      do e=1,dim1
         do e1=1,dim1
            do g=1,gdim1
                tddvrmat(e1,e) = tddvrmat(e1,e) + &
                          dconjg(psig(g,e1))*psig(g,e)*&
                          ovl(e1,e)*opmat(g,e)
            enddo
         enddo
      enddo

      return
      end subroutine buildopmat

!#######################################################################     

      subroutine getwfpnt(psipnt,psi,v1,n1,vdim1,dim1,ndim1)

      implicit none

      integer(long), intent(in) :: v1,n1,vdim1,dim1,ndim1
      integer(long)             :: s
      complex(dop), dimension(vdim1,dim1,ndim1), intent(in) :: psi
      complex(dop), dimension(dim1), intent(out)            :: psipnt

      do s=1,dim1
         psipnt(s) = psi(v1,s,n1)
      enddo

      return
      end subroutine getwfpnt

!#######################################################################     

      subroutine putwfpnt(psipnt,psi,v1,n1,vdim1,dim1,ndim1)

      implicit none

      integer(long), intent(in) :: v1,n1,vdim1,dim1,ndim1
      integer(long)             :: s
      complex(dop), dimension(vdim1,dim1,ndim1), intent(out) :: psi
      complex(dop), dimension(dim1), intent(in)              :: psipnt

      do s=1,dim1
         psi(v1,s,n1) = psipnt(s) 
      enddo

      return
      end subroutine putwfpnt

!#######################################################################
!
! Expresses GWPs for DOF n1 of SPF m1 on a grid stored in 
! psig(gdim1,dim1)
!
!#######################################################################

      subroutine gwpf_on_grid(n1,m1,psig1,gdim1,dim1,ort,gqi)

      implicit none

      integer(long), intent(in) :: n1,m1,gdim1,dim1
      integer(long)             :: e,f,g,g1
      integer(long), dimension(maxdim)                 :: gqi
      real(dop), dimension(ortdim), intent(in)         :: ort
      real(dop), dimension(maxdim)                     :: gpoint,dvrweight
      complex(dop), dimension(gdim1,dim1), intent(out) :: psig1
      complex(dop), dimension(gdim1,dim1)              :: tmppsi
      complex(dop), dimension(maxspf)                  :: tmpc

      f=spfdof(n1,m1)
      dvrweight(f) = dsqrt(ort(zort(f)+1) - ort(zort(f)))
      do g=1,gdim1
         g1=g
         if (lquick) g1 = gqi(f)-1+g
         gpoint(f) = ort(zort(f)-1+g1)

         call gwppnt1d(m1,n1,agmat(zagmat(m1,1)),agmat(zxivec(m1,1)),&
              tmpc,gpoint,dim1,nspfdof(m1),dvrweight)
         do e=1,dim1
            tmppsi(g,e) = tmpc(e)
         enddo
      enddo

      psig1(:,:) = tmppsi(:,:)

      return
      end subroutine gwpf_on_grid

!#######################################################################     

      subroutine cut_spf_grid(m1,psi,cutspf,subdimo1,subdimn1,dim1,gdimo,&
                 gdimn,gqi)
    
      implicit none

      integer(long), intent(in) :: subdimo1,subdimn1,dim1,m1
      integer(long)             :: n,f,e,vgdim1
      integer(long)             :: zeig1,zeig2,swapzeig
      integer(long), dimension(maxdim), intent(in)      :: gdimo,gdimn,gqi
      complex(dop), dimension(subdimo1,dim1), intent(in)  :: psi
      complex(dop), dimension(subdimn1,dim1), intent(out) :: cutspf
      complex(dop), dimension(2*subdimo1)                 :: workc

      zeig1=1
      zeig2=subdimo1+1
      do e=1,dim1
         call cpvxz(psi(1,e),workc(zeig2),subdimo1)
         vgdim1 = 1
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)

            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig

            call cut_spf_grid1(workc(zeig1),workc(zeig2),&
                 gdimo(f),ngdim(f),vgdim1,gdimn(f),gqi(f))
        
             vgdim1 = vgdim1*gdimn(f)
         enddo
         call cpvxz(workc(zeig2),cutspf(1,e),subdimn1)
      enddo

      return
      end subroutine cut_spf_grid

!#######################################################################     

      subroutine cut_spf_grid1(cutspf1,cutspf2,gdim2,ngdim2,&
                 vgdim1,gdim1,gqi)
    
      implicit none

      integer(long), intent(in) :: gdim2,ngdim2,vgdim1,gdim1,gqi
      integer(long)             :: vg,g,ng
      complex(dop), dimension(vgdim1,gdim2,ngdim2), intent(in)  :: cutspf1
      complex(dop), dimension(vgdim1,gdim1,ngdim2), intent(out) :: cutspf2

      do ng = 1,ngdim2
         do g = 1,gdim1
            do vg = 1,vgdim1
               cutspf2(vg,g,ng) = cutspf1(vg,gqi-1+g,ng)
            enddo
         enddo
      enddo

      return
      end subroutine cut_spf_grid1

!#######################################################################
! code for 1D CDVR even for combined modes. Not complete
!#######################################################################

      subroutine cdvrintegrate_1d(psi,adiabpop,diapop,subdimo,vgdimo,gdimo,&
                 ngdimo,subdimn,vgdimn,gdimn,ngdimn,gqi,totgrid,hops,ort,&
                 qcenter,time)

      implicit none

      integer(long)                                     :: f,m,i,s,n,n1,e1,&
                                                           g,nham,v1,ierr,&
                                                           s1,si,sf,m1
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long) :: swapzeig,zeig,zeig1,zeig2,mblock1
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn,&
                                                           vgdimo,vgdimn,&
                                                           ngdimo,ngdimn
      integer(long), intent(in)                         :: totgrid
      integer(long)                                     :: tdvrgrid,cdvrgrid
      real(dop), intent(in)                             :: time 
      real(dop), dimension(qcentdim), intent(in)        :: qcenter
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), dimension(maxsta), intent(inout)       :: diapop,adiabpop
      real(dop), dimension(maxsta)                      :: diapop0,adiabpop0
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      real(dop), dimension(maxsta,maxdim)               :: adiabpopm
      complex(dop), dimension(nddstate,nddstate)        :: dimat
      complex(dop), dimension(maxsta,maxsta)            :: rotmat

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(inout)       :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(maxsta)                   :: psidia,psiad
      complex(dop), dimension(:), allocatable     :: workc,psi1,psi2
      complex(dop), dimension(maxsta,maxsta)      :: opmat
      complex(dop)                                :: ztmp

      complex(dop), dimension(dmatdim)        :: cdvr2,cdvr4
      real(dop), dimension(c3matdim)          :: cdvr3

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock1 = block(1)
      allocate(workc(2*mblock1))
      workc(:) = 0.0_dop

!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      adiabpop(:) = 0.0_dop
      diapop(:) = 0.0_dop
      adiabpop0(:) = 0.0_dop
      diapop0(:) = 0.0_dop

      fpoint(:)=0
      mpoint(:)=0

!-----------------------------------------------------------------------
! get CDVR points
!-----------------------------------------------------------------------
      call getcdvr(psi,cdvr2,cdvr4,cdvr3,ort,1)

!-----------------------------------------------------------------------
! transform A coefficients from spf to TDVR/GWP basis
!   and save in zeig2
!-----------------------------------------------------------------------
      zeig=0
      zeig1=mblock1+1
      zeig2=1
      call cpvxz(psi(zpsi(1)),workc(zeig2),block(1))
      do m=1,nmode
         swapzeig=zeig1
         zeig1=zeig2
         zeig2=swapzeig
         if (m .eq. meb) then
! if meb then transform from SPF basis to electronic state basis.
            call mtxxzz(psi(zetf(m,1)),workc(zeig1),workc(zeig2),&
                          vdim(m,1),dim(m,1),ndim(m,1),subdim(m))
         else if (gwpm(m) .and. psitype .eq. 0) then
! if gwp-spf then transform from SPF basis to GWP basis.
            call qtxxzz(psi(zdgwp(m,1)),workc(zeig1), &
                 workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
         else if (.not. gwpm(m)) then
! transform from SPF basis to TDDVR basis.
            call qtaxzz(cdvr4(dmat(m,1)),workc(zeig1), &
                 workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
         else
! do nothing
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
         endif
      enddo

!-----------------------------------------------------------------------
! save PSI in TDDVR basis as psi1
! psi2 is PSI in adiabatic basis
!-----------------------------------------------------------------------
      allocate(psi1(mblock1))
      allocate(psi2(mblock1))
      psi1(1:mblock1) = workc(zeig2:zeig2-1+mblock1)

!-----------------------------------------------------------------------
! loop over adiabatic states
!-----------------------------------------------------------------------
      do s1=1,nddstate

!-----------------------------------------------------------------------
! loop over TDDVR grid. Get grid point and transform
!-----------------------------------------------------------------------
         vdim1 = vdim(meb,1)
         dim1 = dim(meb,1)
         ndim1 = ndim(meb,1)
!$omp parallel num_threads(ompthread) &
!$omp private(n1,v1,e1,i,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(opmat,m,n,g,f,adiab,diab,sf,si)
!$omp do
         do i=1,vdim1*ndim1
            call vendecoder(v1,e1,n1,i,vdim1,1)

               call adecoder(i,mpoint,s)

               call getwfpnt(psidia,psi1,v1,n1,vdim1,dim1,ndim1)

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!
! for mD spfs, TDDVR grid point is the same for each DOF (simultaneous
! diagonalisation of X)
!-----------------------------------------------------------------------
               do m=1,nmode
                  if (m .eq. meb) cycle

                  if (gwpm(m)) then
                     call getqcent(m,gpoint,qcenter(zcent(m,1)),nspfdof(m),&
                          dim(m,1),mpoint(m))

                  else
                     do n=1,nspfdof(m)
                        f=spfdof(n,m)
                        g=mpoint(m)
                        zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                        gpoint(f) = cdvr3(zeig+g)
                     enddo
                  endif
               enddo

               nham=1
               fpoint(:)=0
               call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

               do si=1,dim(meb,1)
                  do sf=1,dim(meb,1)
                      opmat(sf,si) = rotmat(sf,s1)*dconjg(rotmat(si,s1))
                  enddo
               enddo
     
!-----------------------------------------------------------------------
! calculate adiabatic WF
!-----------------------------------------------------------------------
               psiad(:)=0.0_dop
               call qvtxzz1(opmat,psidia,psiad,maxsta,nddstate)

               call putwfpnt(psiad,psi2,v1,n1,vdim1,dim1,ndim1)
         enddo     
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! calculate the state populations
!-----------------------------------------------------------------------
         zeig=0
         zeig1=mblock1+1
         zeig2=1
         call cpvxz(psi1(zpsi(1)),workc(zeig2),block(1))
         if (lgwp) then
            do m=1,nmode
               if (.not. gwpm(m)) cycle
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               call qtxxzz(gs2(gs2mat(m,1)),workc(zeig1), &
                    workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
            enddo
         endif

! adiabatic populations
         call vvaxzz(workc(zeig2),psi2,ztmp,block(1))
         adiabpop(s1) = dble(ztmp)

      enddo
! diabatic populations
      call rmqxxxzz(workc(zeig2),psi1,dimat,vdim1,dim1,ndim1)
      do s=1,nddstate
         diapop(s) = dble(dimat(s,s))
      enddo

!-----------------------------------------------------------------------
! CDVR correction for adiabatic populations (TDDVR exact for diabatic)
!-----------------------------------------------------------------------
      if (lcdvrint) then
         cdvrgrid=1

         if (lgwp) then
            do m=1,nmode
               if (m .eq. meb) cycle
               do n=1,nspfdof(m)
                  f=spfdof(n,m)
                  gdim(f) = gdimo(f)
               enddo
            enddo
            call getgwpgrids(ort)
         endif
       
!----------------------------------------------------------------------- 
! calculate correction one mode at a time
! workc(zeig2) contains Psi multiplied by GWP overlaps
!----------------------------------------------------------------------- 
         do m1=1,nmode
            if (m1 .eq. meb) cycle
            do s=1,nddstate
               if (gwpm(m1)) then
                  workc(:)=0.0_dop
                  zeig=0
                  zeig1=mblock1+1
                  zeig2=1
                  call cpvxz(psi1(zpsi(1)),workc(zeig2),block(1))
                  do m=1,nmode
                     if (.not. gwpm(m)) cycle
                     if (m .eq. m1) cycle
                     swapzeig=zeig1
                     zeig1=zeig2
                     zeig2=swapzeig
                     call qtxxzz(gs2(gs2mat(m,1)),workc(zeig1), &
                          workc(zeig2),vdim(m,1),dim(m,1),ndim(m,1))
                  enddo
                  call cdvrintg(m1,s,psi1,workc(zeig2),cdvr3,adiabpopm(s,m1),gdimo,&
                       gdimn,gqi,hops,ort,mblock1,ierr,&
                       qcenter)
               else       
                  call cdvrintm_1d(m1,s,psi,psi1,workc(zeig2),cdvr3,cdvr4,&
                       adiabpopm(s,m1),qcenter,subdimo,vgdimo,gdimo,ngdimo,&
                       subdimn,vgdimn,gdimn,ngdimn,gqi,hops,ort,mblock1,&
                       cdvrgrid,ierr)
               endif
            enddo
         enddo

!-----------------------------------------------------------------------
! Add CDVR correction to populations
!
! Store TDDVR result in adiabpop0
!-----------------------------------------------------------------------
         do s=1,nddstate
            adiabpop0(s) = adiabpop(s)
         enddo

         do s=1,nddstate
            do m=1,nmode
               if (m .eq. meb) cycle
             !  if (gwpm(m)) then
! For GWPs there is a correction for each DOF
                  adiabpop(s)=adiabpop(s)+ &
                      (adiabpopm(s,m) - nspfdof(m)*adiabpop0(s))
             !  else
             !     adiabpop(s)=adiabpop(s)+(adiabpopm(s,m) - adiabpop0(s))
             !  endif
            enddo
         enddo

      endif  ! End of CDVR

!-----------------------------------------------------------------------
! Restore dimensions
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle
         subdim(m) = subdimo(m)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f) = gdimo(f)
         enddo
      enddo

      deallocate(psi1,workc)

!-----------------------------------------------------------------------
! write out no. of included points in quick calculation
!-----------------------------------------------------------------------
      write(ilog,'(f8.3,i10,2(a3,i10))')&
                   time, tdvrgrid,' / ',cdvrgrid,' / ',totgrid
      call myflush(ilog)

      return
      end subroutine cdvrintegrate_1d

!#######################################################################

      subroutine cdvrintm_1d(m1,sa,psi,psi1,psig,cdvr3,cdvr4,adiabpopm,qcenter,&
                 subdimo,vgdimo,gdimo,ngdimo,subdimn,vgdimn,gdimn,ngdimn,&
                 gqi,hops,ort,mblock1,ierr,cdvrgrid)

      implicit none

      integer(long), intent(in)                         :: mblock1,m1,sa
      integer(long)                                     :: dimpsi4
      integer(long)                                     :: f,m,i,n,n1,e1,g1,gg,&
                                                           g,nham,v1,si,sf
      integer(long)                                     :: f1,j,n2,ve,ee,ne
      integer(long)                                     :: vdim1,dim1,ndim1
      integer(long), intent(inout)                      :: ierr
      integer(long) :: zeig,mblock2
      integer(long)                                     :: effgrid
      integer(long), dimension(maxdim)                  :: mpoint,fpoint
      integer(long), dimension(maxdim), intent(in)      :: gqi,subdimo,&
                                                           gdimo,subdimn,gdimn,&
                                                           vgdimo,vgdimn,&
                                                           ngdimo,ngdimn
      integer(long)                                     :: cdvrgrid
      real(dop), dimension(maxdim)                      :: gpoint
      real(dop), intent(out)                            :: adiabpopm
      real(dop), dimension(maxsta)                      :: adiab
      real(dop), dimension(maxsta,maxsta)               :: diab
      complex(dop), dimension(maxsta,maxsta)            :: rotmat
      real(dop), dimension(qcentdim), intent(in)        :: qcenter

      real(dop), dimension(hopsdim), intent(in)         :: hops
      real(dop), dimension(ortdim), intent(in)          :: ort
      complex(dop), dimension(dgldim), intent(in)       :: psi
      complex(dop), dimension(mblock1), intent(in)      :: psi1,psig
      complex(dop), dimension(maxsta,maxspf)            :: psidia,psiad
      complex(dop), dimension(:), allocatable     :: tmpspf,psi2,psi4
      complex(dop), dimension(:), allocatable     :: covl
      complex(dop)                                :: ztmp
      complex(dop), dimension(:,:,:), allocatable   :: tddvrmat
      complex(dop), dimension(:,:,:,:), allocatable  :: opmat

      complex(dop), dimension(dmatdim), intent(in) :: cdvr4
      real(dop), dimension(c3matdim), intent(in)   :: cdvr3

!-----------------------------------------------------------------------
! Allocate work arrays
!-----------------------------------------------------------------------
      mblock2 = block(1)
      cdvrgrid = max(cdvrgrid,effgrid)
      allocate(psi2(mblock2),stat=ierr)
      if (ierr .gt. 0) return

      if (lquick) then
         dimpsi4= subdimn(m1)*dim(m1,1)
      else 
         dimpsi4= subdimo(m1)*dim(m1,1)
      endif
      allocate(psi4(dimpsi4),stat=ierr)
      if (ierr .gt. 0) return

      allocate(tmpspf(sdim*maxspf),stat=ierr)
      allocate(covl(maxgdim*maxspf),stat=ierr)
      if (ierr .gt. 0) return


      psi2(:)=0.0_dop
      psi4(:)=0.0_dop
      tmpspf(:)=0.0_dop
      covl(:)=0.0_dop

      adiabpopm = 0.0_dop

!-----------------------------------------------------------------------
! Get grid sizes
!-----------------------------------------------------------------------
      if (lquick) then
         subdim(m1)=subdimn(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimn(f)
         enddo
      else
         subdim(m1)=subdimo(m1)
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            gdim(f)=gdimo(f)
         enddo
      endif

!-----------------------------------------------------------------------
! Set up CDVR functions on grid
!-----------------------------------------------------------------------
      if (lquick) then
         call cut_spf_grid(m1,psi(zetf(m1,1)),tmpspf,subdimo(m1),&
              subdimn(m1),dim(m1,1),gdimo,gdimn,gqi)
      else
         call cpmxz(psi(zetf(m1,1)),tmpspf,subdimo(m1),dim(m1,1))
      endif
      call mmxxzz(tmpspf,cdvr4(dmat(m1,1)),psi4,&
           subdim(m1),dim(m1,1),dim(m1,1))

!-----------------------------------------------------------------------
! Get effective grid sizes in CDVR rep
!-----------------------------------------------------------------------
      do m=1,nmode 
         subdim(m)=dim(m,1)
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            gdim(f)=dim(m,1)
         enddo
      enddo

      allocate(tddvrmat(dim(m1,1),dim(meb,1),dim(meb,1)))

!-----------------------------------------------------------------------
! now operate on vector by V(q1,.xm...qf)
! by recursively dividing out the number of grid points for each
! mode, the gridpoint gpnt is transformed to the mode grid point mgpnt
!-----------------------------------------------------------------------
      vdim1 = vdim(m1,1)
      dim1 = dim(m1,1)
      ndim1 = ndim(m1,1)
      if (m1 .lt. meb) then
         ndim1 = ndim1/dim(meb,1)
      else
         vdim1 = vdim1/dim(meb,1)
      endif

!-----------------------------------------------------------------------
! Change SPF pointers for modes after m1 to set e1=1 in decomposition of index
!-----------------------------------------------------------------------
      do m=m1+1,nmode
         if (m .eq. meb) cycle
         vdim(m,1) = vdim(m,1) /dim(m1,1)
      enddo

      do n1=1,nspfdof(m1)
         f1=spfdof(n1,m1)

         if (lquick) then
            vgdim(f1)=vgdimn(f1)
            gdim(f1)=gdimn(f1)
            ngdim(f1)=ngdimn(f1)
         else
            vgdim(f1)=vgdimo(f1)
            gdim(f1)=gdimo(f1)
            ngdim(f1)=ngdimo(f1)
         endif
         allocate(opmat(gdim(f1),dim(m1,1),dim(meb,1),dim(meb,1)))

         call intcovr(psi4,covl,vgdim(f1),gdim(f1),ngdim(f1),dim(m1,1))

!$omp parallel num_threads(ompthread) &
!$omp private(n1,v1,e1,i,psidia,mpoint,gpoint,psiad,rotmat) &
!$omp private(opmat,m,n,g,f,adiab,diab)
!$omp do
         do i=1,vdim1*ndim1
            call vendecoder(v1,e1,n2,i,vdim1,1)

            call adecoder(i,mpoint,si)

            do e1=1,dim1   ! Loop over functions of m1 to get state dependent WF 
               mpoint(m1) = e1
               j = v1 + (e1-1)*vdim1 + (n2-1)*vdim1*dim1
               call vendecoder(ve,ee,ne,j,vdim(meb,1),dim(meb,1))
               call getwfpnt(psidia(1,e1),psi1,ve,ne,vdim(meb,1),&
                    dim(meb,1),ndim(meb,1))
            enddo

!-----------------------------------------------------------------------
! get TDDVR coordinates 
!-----------------------------------------------------------------------
            tddvrmat=0.0_dop
            opmat=0.0_dop
            do e1=1,dim1   ! Loop over functions of m1
               mpoint(m1) = e1

               do m=1,nmode
                  if (m .eq. meb) cycle

                  if (gwpm(m)) then
                     call getqcent(m,gpoint,qcenter(zcent(m,1)),nspfdof(m),&
                          dim(m,1),mpoint(m))

                  else
                     do n=1,nspfdof(m)
                        f=spfdof(n,m)
                        g=mpoint(m)
                        zeig = c3mat(m,1)-1+dim(m,1)*(n-1)
                        gpoint(f) = cdvr3(zeig+g)
                     enddo
                  endif
               enddo

               do g1 = 1,gdim(f1)
                  gg=g1
                  if (lquick) gg = gg + gqi(f1) - 1
                  gpoint(f1) = ort(zort(f1)-1+gg)
                  nham=1
                  call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

! build operators at each point
                  do si=1,dim(meb,1)
                     do sf=1,dim(meb,1)
                         opmat(gg,e1,sf,si) = &
                                rotmat(sf,sa)*dconjg(rotmat(si,sa))
                     enddo
                  enddo
               enddo
            enddo

! now build operator matrices
            do si=1,dim(meb,1)
               do sf=1,dim(meb,1)
                  call buildopmatm(tddvrmat(1,sf,si),&
                       opmat(1,1,sf,si),covl,gdim(f1),dim(m1,1))
               enddo
            enddo

            psiad(:,:)=0.0_dop
            do e1=1,dim1   ! Loop over functions of m1
               do si=1,dim(meb,1)
                  do sf=1,dim(meb,1)
                     psiad(sf,e1) = psiad(sf,e1) + &
                           tddvrmat(e1,sf,si)*psidia(si,e1)
                  enddo
               enddo
            enddo

            do e1=1,dim1   ! Loop over functions of m1
               j = v1 + (e1-1)*vdim1 + (n2-1)*vdim1*dim1
               call vendecoder(ve,ee,ne,j,vdim(meb,1),dim(meb,1))
               call putwfpnt(psiad(1,e1),psi2,ve,ne,vdim(meb,1),dim(meb,1),ndim(meb,1))
            enddo

         enddo
!$omp enddo
!$omp end parallel

!-----------------------------------------------------------------------
! calculate the state populations
! (sum over DOF contributions)
!-----------------------------------------------------------------------
         call vvaxzz(psig,psi2,ztmp,block(1))
         adiabpopm = adiabpopm + dble(ztmp)

         deallocate(opmat)

         gdim(f1) = dim(m1,1)

      enddo

! Restore dimensions
      do m=m1+1,nmode
         if (m .eq. meb) cycle
         vdim(m,1) = vdim(m,1)*dim(m1,1)
      enddo

      deallocate(tddvrmat)
      deallocate(psi4,psi2)

      return
      end subroutine cdvrintm_1d

!#######################################################################

      subroutine intcovr(psi4,covl,vgdim1,gdim1,ngdim1,dim1)

      integer(long), intent(in) :: vgdim1,gdim1,ngdim1,dim1
      integer(long)             :: e1,vg,g,ng
      complex(dop), dimension(vgdim1,gdim1,ngdim1,dim1), intent(in) :: psi4
      complex(dop), dimension(gdim1,dim1), intent(out) :: covl

      covl=0.0
      do e1=1,dim1
         do ng=1,ngdim1
            do g=1,gdim1
               do vg=1,vgdim1
                  covl(g,e1) = covl(g,e1) + &
                        dconjg(psi4(vg,g,ng,e1))*psi4(vg,g,ng,e1)
               enddo
             enddo
          enddo
      enddo

      return
      end subroutine intcovr

!#######################################################################     

      subroutine buildopmatm(tddvrmat,opmat,covl,gdim1,dim1)

      implicit none

      integer(long), intent(in) :: gdim1,dim1
      integer(long)             :: e,g
      complex(dop), dimension(dim1), intent(out)  :: tddvrmat
      complex(dop), dimension(gdim1,dim1), intent(in)  :: opmat
      complex(dop), dimension(gdim1,dim1), intent(in)  :: covl

      tddvrmat=0.0
      do e=1,dim1
         do g=1,gdim1
             tddvrmat(e) = tddvrmat(e) + &
                          covl(g,e)*opmat(g,e)
         enddo
      enddo

      return
      end subroutine buildopmatm

!#######################################################################
! END code for 1D CDVR even for combined modes. Not complete
!#######################################################################

!#######################################################################     
      
      subroutine expop(hops,psi,subdim1,nstate1,psidia,psiad,p,fpoint,&
                 ort)

      implicit none

      integer(long)                                        :: s,s1,g,f,m,n,nham
      integer(long), intent(in)                            :: p,nstate1,subdim1
      integer(long), dimension(maxdim), intent(in)         :: fpoint
      real(dop), dimension(hopsdim), intent(in)            :: hops
      real(dop), dimension(ortdim), intent(in)             :: ort
      real(dop), dimension(maxdim)                         :: gpoint
      real(dop), dimension(maxsta)                         :: adiab
      real(dop), dimension(maxsta,maxsta)                  :: diab
      complex(dop), dimension(subdim1,nstate1), intent(in) :: psi
      complex(dop), dimension(nstate1), intent(out)        :: psiad,psidia
      complex(dop), dimension(maxsta,maxsta)               :: rotmat


!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!-----------------------------------------------------------------------
      gpoint=0.0_dop
      do m=1,nmode
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            g=fpoint(f)
            gpoint(f) = ort(zort(f)-1+g)
         enddo
      enddo

      nham=1

      call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

      do s=1,nstate1
         psidia(s)=psi(p,s)
         psiad(s)=0.0_dop
         do s1=1,nstate1
            psiad(s) = psiad(s) + dconjg(rotmat(s1,s))*psi(p,s1)
         enddo
      enddo

      return
      end subroutine expop

!#######################################################################     
      
      subroutine wfpop(hops,psi,dgldim1,nstate1,psidia,psiad,fpoint,&
                 mpoint,ort,agmat,pntweight)

      implicit none

      integer(long)                                    :: s,m,zeig1,zeig2,&
                                                          swapzeig,vdim1,&
                                                          f,g,n,nham,s1
      integer(long), intent(in)                        :: nstate1,dgldim1
      integer(long), dimension(maxdim), intent(in)     :: fpoint,mpoint
      real(dop), dimension(hopsdim), intent(in)        :: hops
      real(dop), dimension(ortdim), intent(in)         :: ort,pntweight
      real(dop), dimension(maxdim)                     :: gpoint,dvrweight,&
                                                          gweight
      real(dop), dimension(maxsta)                     :: adiab
      real(dop), dimension(maxsta,maxsta)              :: diab
      complex(dop), dimension(dgldim1), intent(in)     :: psi
      complex(dop), dimension(nstate1), intent(out)    :: psiad,psidia
      complex(dop), dimension(agmatdim), intent(in)    :: agmat
      complex(dop), dimension(maxspf)                  :: tmpc,tmpc1
      complex(dop), dimension(maxsta,maxsta)           :: rotmat
      complex(dop), dimension(:), allocatable          :: workc

      if (lmult) then
         allocate(workc(2*maxblock))
      else
         allocate(workc(2*maxblock*nddstate))
      endif

!-----------------------------------------------------------------------
! get "weight" required for GWP integration 
!-----------------------------------------------------------------------
      do m=1,nmode
         if (.not. gwpm(m)) cycle
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            dvrweight(f) = dsqrt(ort(zort(f)+1) - ort(zort(f)))
         enddo
      enddo

!-----------------------------------------------------------------------
! get grid point coordinates and calculate transformation matrix
!-----------------------------------------------------------------------
      do m=1,nmode
         if (m .eq. meb) cycle

         do n=1,nspfdof(m)
            f=spfdof(n,m)
            g=fpoint(f)
            gpoint(f) = ort(zort(f)-1+g)
            gweight(f) = pntweight(zort(f)-1+g)
         enddo
      enddo

      nham=1
      call calcdiab(hops,adiab,diab,rotmat,fpoint,gpoint,nham)

!-----------------------------------------------------------------------
! calculate diabatic WF
!-----------------------------------------------------------------------
      if (lmult) then
         do s=1,nstate
            zeig1=1
            zeig2=block(s)+zeig1
            call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))
            do m=1,nmode
               swapzeig=zeig1
               zeig1=zeig2
               zeig2=swapzeig
               if (gwpm(m)) then
                  call gwppnt(m,agmat(zagmat(m,s)),agmat(zxivec(m,s)),&
                       agmat(zphase(m,s)),tmpc,gpoint,dim(m,s),nspfdof(m),&  
                       dvrweight,gweight)

                  call mmvxzz(tmpc,workc(zeig1),workc(zeig2),&
                              1,dim(m,s),ndim(m,s),1)
               else
                  call extrpnt(m,psi(zetf(m,s)),tmpc,gweight,subdim(m),&
                       dim(m,s),mpoint(m))

                  call mmvxzz(tmpc,workc(zeig1),workc(zeig2),&
                              1,dim(m,s),ndim(m,s),1)
               endif
            enddo
            psidia(s)=workc(zeig2)
         enddo
      else
         s=1
         zeig1=1
         zeig2=block(s)+zeig1
         call cpvxz(psi(zpsi(s)),workc(zeig2),block(s))
         vdim1 = 1
         do m=1,nmode
            swapzeig=zeig1
            zeig1=zeig2
            zeig2=swapzeig
            if (m .eq. meb) then
               call mtxxzz(psi(zetf(m,s)),workc(zeig1),workc(zeig2),&
                           vdim1,dim(m,s),ndim(m,s),subdim(m))
! if meb then transform from SPF basis to electronic state basis.
! make vdim=nddstate so that at end of loop summed WP is in a vector 
! length nddstate starting at workc(zeig2).
               vdim1=nddstate
            else
               if (gwpm(m)) then
                  call gwppnt(m,agmat(zagmat(m,s)),agmat(zxivec(m,s)),&
                       agmat(zphase(m,s)),tmpc,gpoint,dim(m,s),nspfdof(m),&  
                       dvrweight,gweight)
                  if (psitype .eq. 0) then
                     call mvtxzz(psi(zdgwp(m,s)),tmpc,tmpc1,dim(m,s),dim(m,s))
                     tmpc(:)=tmpc1(:)
                  endif

                  call mtvxzz(tmpc,workc(zeig1),workc(zeig2),&
                                 vdim1,dim(m,s),ndim(m,s),1,1)
               else
                  call extrpnt(m,psi(zetf(m,s)),tmpc,gweight,subdim(m),&
                       dim(m,s),mpoint(m))

                  call mtvxzz(tmpc,workc(zeig1),workc(zeig2),&
                              vdim1,dim(m,s),ndim(m,s),1,1)

               endif
            endif
            do s1=1,nddstate
               psidia(s1)=workc(zeig2+s1-1)
            enddo
         enddo
      endif

!-----------------------------------------------------------------------
! calculate adiabatic WF
!-----------------------------------------------------------------------
      do s=1,nddstate
         psiad(s)=0.0_dop
      enddo

      call qvtxzz1(rotmat,psidia,psiad,maxsta,nddstate)

      deallocate(workc)

      return
      end subroutine wfpop 

!#######################################################################

      subroutine extrpnt(m1,psi1,wfpnt,gweight,subdim1,dim1,p1)

      implicit none

      integer(long), intent(in) :: subdim1,dim1,p1,m1
      integer(long)             :: e,n,f
      real(dop), dimension(maxdim), intent(in)      :: gweight
      real(dop)                                     :: wght
      complex(dop), dimension(subdim1,dim1), intent(in) :: psi1
      complex(dop), dimension(dim1), intent(out)        :: wfpnt

! multiply weights at point for DOFs in mode m1
      wght = 1.0_dop
      do n=1,nspfdof(m1)
         f=spfdof(n,m1)
         wght = wght * gweight(f)
      enddo

! extract wieghted WF at point p1 
      do e=1,dim1
         wfpnt(e) = psi1(p1,e)*wght
      enddo

      return
      end subroutine extrpnt
   
!#######################################################################
!
! guess the grid density by Prod rho(m), i.e. assume separable 
! wavepacket. Ignore gridpoint if density < qtol
!
!#######################################################################

      subroutine guessden(gpop1,fpoint,lignore,qtol)

      implicit none

      integer(long)                                    :: m,n,f,g,s
      integer(long), dimension(maxdim), intent(in)     :: fpoint
      real(dop)                                        :: sepden
      real(dop), intent(in)                            :: qtol
      real(dop), dimension(ortdim,nddstate*npacktot), intent(in)  :: gpop1
      logical(kind=4), intent(out)                     :: lignore

      lignore=.false.

      sepden=1.0_dop
      do m=1,nmode
         if (m .eq. meb) cycle
         do n=1,nspfdof(m)
            f=spfdof(n,m)
            g=fpoint(f)
            sepden=sepden*gpop1(zort(f)-1+g,1)
         enddo
      enddo

      if (sepden .lt. qtol) lignore=.true.

      return
      end subroutine guessden

!#######################################################################

      subroutine gwppnt(m1,agmat,xi,phi,wfpnt,gpoint,dim1,gdof,&
                 dvrweight,gweight)

      implicit none

      integer(long)                                       :: e,n,f
      integer(long), intent(in)                           :: dim1,gdof,m1
      real(dop)                                           :: zarg,zmod
      real(dop), dimension(maxdim), intent(in)            :: gpoint,dvrweight
      real(dop), dimension(maxdim), intent(in)            :: gweight
      complex(dop), dimension(dim1), intent(in)           :: phi
      complex(dop), dimension(dim1), intent(out)          :: wfpnt
      complex(dop), dimension(gdof,dim1), intent(in)      :: xi
      complex(dop), dimension(gdof,gdof,dim1), intent(in) :: agmat

      do e=1,dim1
         wfpnt(e) = phi(e) 
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            wfpnt(e) = wfpnt(e) + agmat(n,n,e)*gpoint(f)*gpoint(f) +&
                       xi(n,e)*gpoint(f) 
            wfpnt(e) = wfpnt(e) + log(gweight(f))
         enddo
      enddo

      do e=1,dim1
         zmod = dble(wfpnt(e))
         zarg = dimag(wfpnt(e))
         wfpnt(e) = dexp(zmod)*dcmplx(cos(zarg),sin(zarg))
      enddo

      do e=1,dim1
         do n=1,nspfdof(m1)
            f=spfdof(n,m1)
            wfpnt(e) = wfpnt(e) * dvrweight(f) 
         enddo
      enddo

      return
      end subroutine gwppnt

!#######################################################################
!
! gets the value of GWPs for DOF n1 of SPF m1 on the grid point gpoint
!
!#######################################################################

      subroutine gwppnt1d(m1,n1,agmat,xi,wfpnt,gpoint,dim1,gdof,dvrweight)

      implicit none

      integer(long)                                       :: e,f
      integer(long), intent(in)                           :: dim1,gdof,m1,n1
      real(dop)                                           :: zarg,zmod
      real(dop), dimension(maxdim), intent(in)            :: gpoint,dvrweight
      real(dop)                                           :: a,p,q
      complex(dop), dimension(dim1)                       :: phi
      complex(dop), dimension(dim1), intent(out)          :: wfpnt
      complex(dop), dimension(gdof,dim1), intent(in)      :: xi
      complex(dop), dimension(gdof,gdof,dim1), intent(in) :: agmat

! get scalar parameters
      do e=1,dim1
         a = -dble(agmat(n1,n1,e))
         p = dimag(xi(n1,e))
         q = dble(xi(n1,e))
         q = 0.5*q/a
         phi(e) = - a*q**2 &
                   - (0.0_dop,1.0_dop)*p*q &
                   - 0.25_dop*dlog(pi/(2.0_dop*dble(a)))
      enddo

      do e=1,dim1
         wfpnt(e) = phi(e) 
         f=spfdof(n1,m1)
         wfpnt(e) = wfpnt(e) + agmat(n1,n1,e)*gpoint(f)*gpoint(f) +&
                    xi(n1,e)*gpoint(f) 
         wfpnt(e) = wfpnt(e) 
      enddo

      do e=1,dim1
         zmod = dble(wfpnt(e))
         zarg = dimag(wfpnt(e))
         wfpnt(e) = dexp(zmod)*dcmplx(cos(zarg),sin(zarg))
      enddo

      do e=1,dim1
         f=spfdof(n1,m1)
         wfpnt(e) = wfpnt(e) * dvrweight(f) 
      enddo

      return
      end subroutine gwppnt1d

!#######################################################################
!
! Calculates the grid that supports the GWP basis
!
! On Out: ort(f) contains grid points
!
!#######################################################################

      subroutine getgwpgrids(ort)

      implicit none

      integer(long) :: m,s
      real(dop), dimension(maxdim)                :: qmax,qmin
      real(dop), dimension(ortdim), intent(out)   :: ort

      do m=1,nmode
         if (.not. gwpm(m)) cycle

         do s=1,nstate
            call gwpgrids(m,agmat(zagmat(m,s)),agmat(zxivec(m,s)),&
                          nspfdof(m),dim(m,s),qmin,qmax)
         enddo

         do n=1,nspfdof(m)
            f=spfdof(n,m)
            call setort(ort(zort(f)),gdim(f),qmax(n),qmin(n))
         enddo
      enddo

      return
      end subroutine getgwpgrids

!#######################################################################

      subroutine gwpgrids(m1,agmat,xi,gdof,dim1,qmin,qmax)

      implicit none

      integer(long)                                       :: e,n,f
      integer(long), intent(in)                           :: m1,dim1,gdof
      real(dop)                                           :: a,q,sigma
      real(dop), dimension(gdof), intent(out)             :: qmax,qmin
      complex(dop), dimension(gdof,dim1), intent(in)      :: xi
      complex(dop), dimension(gdof,gdof,dim1), intent(in) :: agmat

      do n=1,gdof
         f=spfdof(n,m1)
! get parameters
         qmax=-999.0
         qmin=999.0
         do e=1,dim1
            a = -dble(agmat(n,n,e))
            q = dble(xi(n,e))
            q = 0.5*q/a
            sigma = 1.0/(2.0*a)
            sigma=dsqrt(sigma)
            qmax = max(qmax,q+3.0*sigma)
            qmin = min(qmin,q-3.0*sigma)
         enddo
      enddo

      return
      end subroutine gwpgrids

!#######################################################################

      subroutine setort(ort,gdim1,qmax,qmin)

      implicit none

      integer(long), intent(in) :: gdim1
      integer(long)             :: g
      real(dop), intent(in)     :: qmax,qmin
      real(dop),dimension(gdim1), intent(out) :: ort
      real(dop)                 :: dx,dlen

      dlen = qmax-qmin
      dx=dlen/dble(gdim1-1)
      do g=1,gdim1
         ort(g) = qmin + (g-1)*dx
      enddo

      return
      end subroutine setort
 
      end program adpop
! vi: set sw=3 ts=3:
