!#######################################################################
!
! READQC
! Module for reading QC output files to extract data
!
! All output in atomic units
!
! Normal modes are for atoms in Cartesian space, i.e. not the 
! mass-weighted eigenvectors, but the eigenvalues corrected
! for the mass-weighting. This is the standard QC output 
! required to plot the nm vectors.
!
! ifiletyp: 
!     1   Gaussian
!     2   Gamess-uk
!     3   QOMMMA
!     4   Bagel
!     5   Molpro
!     6   DFT/MRCI
!     7   Turbomole
!     8   Gamess-US
!     10  Q-Chem
!     11  Molcas
!     12  Orca
!
! imethod specifies the QC method:
!     1  : RHF
!     2  : MP2
!     3  : MP3
!     4  : MP4
!     5  : UHF
!     6  : ROHF
!     9  : ADC
!     10 : OVGF
!     11 : CIS
!     12 : CAS
!     13 : CASPT2
!     14 : MRCI
!     15 : CAS/MRCI  (energies MRCI, grads + hessians CAS)
!     16 : 
!     17 : EOM-CCSD
!     18 : DFT/MRCI
!     19 : DFT 
!     20 : CR-EOM-CCSD(T)
!     21 : Delta-CR-EOM-CCSD(T)
!     22 : CCSD
!     23 : RI-CC2
!     24 : TDDFT with PCM
!     25 : TDDFT with pt-ss-PCM
!     26 : local-ADC(2)
!     27 : MS-CASPT2
!     28 : RMS-CASPT2
!     29 : QoMMMa
!     30 : Ehrenfest
!     31 : External
!     32 : CAS/PT2  (energies PT2, grads + hessians CAS. SOC MRCI (Molpro))
!     33 : TDDFT
!     34 : local-CC2
!     36 : XDW-CASPT2
!     37 : CAS/XDW
!     38 : ONIOM model system CAS (highest layer)
!
! OLD NUMBERS DD-VMCG
!            1  RHF
!            2  CAS (single-state)
!            3  SA-CAS
!            4  DFT
!            5  QoMMMa
!            6  UHF
!            7  TDDFT
!            8  ROHF
!            9  Ehrenfest
!           10  External
!           11  CAS/PT2
!           12  EOM-CCSD
!
! 1. Set values specifying array sizes
! 2. Allocate arrays
! 3. Call rdqcout
! 4. Deallocate arrays
!
! State numbering:
! Output is ordered by state as found in the output file. If SOC is
! included then the SOC data is returned in qcsoc array with numbering
! of states including the multiplet components in blocks of different 
! ms. This is following the Molpro convention.
! For example, if a calculation includes 3S + 2T (S0, S1, S2, T1, T2)
! then 
! S = 1  2  3  4  5
!     S0 S1 S2 T1 T2
! S_soc = 1   2   3   4   5    6   7   8    9
!         S0  S1  S2  T1+ T1_0 T1- T2+ T2_0 T2-
!#######################################################################

      module readqc

use decimal, only: sip, dop, long
use keyunitsmod

      implicit none

      public 

!-----------------------------------------------------------------------
! If set to true, routines can be probed for what is implemented
!-----------------------------------------------------------------------
      logical(kind=4),save :: qc_enquire=.false.

!-----------------------------------------------------------------------
! Arrays specifying sizes
! qc_ncoo   : number of coordinates
! qc_nmode  : number of normal coordinates
! qc_nst    : number of states
! qc_nbasis : no. of basis functions
! qc_natm   : number of atoms     (defined in rdqcout as qc_ncoo/3)
! qc_ncivec : maximum no. of CI vector configurations to be read
!
! qc_nstdeg : no. of spin states (no. of states including "degeneracies")
! qc_nmult(sp)  : number of states`of multiplicity sp
! qc_nspblk     : no. of blocks of different multiplicities
! qc_spboff(i)  : index offest for spin block i
!
!-----------------------------------------------------------------------
      integer(long), save :: qc_natm=0,&
                             qc_nst=0,&
                             qc_ncoo=0,&
                             qc_nmode=0,&
                             qc_nbasis=0,&
                             qc_ncivec=0

      integer(long), save               :: qc_nstdeg=0,&
                                           qc_nspblk = 0
      integer(long), dimension(7), save :: qc_nmult=0
      integer(long), dimension(8), save :: qc_spboff = 0

!-----------------------------------------------------------------------
! qc_nirrep     : no. of symmetries
! qc_nsymorb(i) : no. orbs of different symmetries stored 
!
! If not set by external program, nirrep=1 and qc_nsymorb(1) = qc_nbasis
!-----------------------------------------------------------------------
      integer(long), save               :: qc_nirrep = 1
      integer(long), dimension(8), save :: qc_nsymorb = 0

!-----------------------------------------------------------------------
! Arrays to define data required so check can be made that data has 
! been read (data in a.u.)
! 1  coordinates (Bohr)
! 2  normal modes (frequencies in Hartree and normal mode in 
!    normalised mass-scaled displacement)
! 3  energy (Hartree)
! 4  first derivatives (Hartree/Bohr)
! 5  second derivatives (Hartree/Bohr**2)
! 6  MO coefficients
! 7  non-adiabatic coupling vector 
! 8  dipole moment (e*Bohr. 1D=0.393430307)
! 9  transition dipole
! 10 Isotopes
! 11 SOC
! 12 CI vector
!
! qcfound(i) : information i found
! lqcfoundX  : details of which information (e.g. state) found
!
! qcread(i) : information i asked for
! 
!-----------------------------------------------------------------------
      logical(kind=4), dimension(12), save :: qcread

      logical(kind=4), dimension(12), save :: qcfound
      logical(kind=4), dimension(:), allocatable, save   :: lqcfound3
      logical(kind=4), dimension(:), allocatable, save   :: lqcfound4
      logical(kind=4), dimension(:), allocatable, save   :: lqcfound5
      logical(kind=4), dimension(:,:), allocatable, save :: lqcfound7
      logical(kind=4), dimension(:), allocatable, save   :: lqcfound8
      logical(kind=4), dimension(:,:), allocatable, save :: lqcfound9
      logical(kind=4), dimension(:,:), allocatable, save :: lqcfound11
      logical(kind=4), dimension(:,:), allocatable, save :: lqcfound12

!-----------------------------------------------------------------------
! Arrays for data to be read
! qcv(nst)                     : Energies
! qcderiv1(ncoo,nst)           : First derivatives
! qcderiv2(ncoo,ncoo,nst)      : Second derivatives
! qcnact(ncoo,nst,nst)         : Non-adiabatic coupling vectors. 
! qcmocoeff(nbasis,nbasis)     : MO coefficients. qcmocoeff(i,j) i=MO, j=AO
! qcdipole(3,nst)              : Dipole moments
! qctdipole(4,nst,nst)         : Transition Dipole moments
! qcsoc(nstdeg,nstdeg)         : Spin orbit couplings  (dimensions include 
!                                spin components
!
! qciatsym(ncoo)      : Atomic symbol
! qciatiso(ncoo)      : Isotopic variant
! qcxcoo(ncoo)        : Cartesian coordinates   
! qcfreq(ncoo)        : Frequencies
! qcnmcoo(ncoo,nmcoo) : Normal mode coordinates
! qcnmlab(ncoo)       : Normal mode labels
! qcnciconf(ns)       : No. of CI vector configurations for multiplicity ns
! qcciconfig(ncivec,ns) : CI vector configurations for multiplicity ns
! qccicoeff(ncivec,nst) : CI vector coefficients for state nst
!
! NB qcmocoeff and are  sip, all other data dop
!-----------------------------------------------------------------------
      real(dop), dimension(:)  , allocatable, save         :: qcv
      real(dop), dimension(:,:), allocatable, save         :: qcderiv1
      real(dop), dimension(:,:,:), allocatable, save       :: qcderiv2
      real(sip), dimension(:,:), allocatable, save         :: qcmocoeff
      real(dop), dimension(:,:,:), allocatable, save       :: qcnact
      real(dop), dimension(:,:), allocatable, save         :: qcdipole
      real(dop), dimension(:,:,:), allocatable, save       :: qctdipole
      complex(dop), dimension(:,:), allocatable, save      :: qcsoc

      integer(long), dimension(:), allocatable, save       :: qciatsym
      integer(long), dimension(:), allocatable, save       :: qciatiso
      real(dop), dimension(:), allocatable, save           :: qcxcoo
      real(dop), dimension(:), allocatable, save           :: qcfreq
      real(dop), dimension(:,:), allocatable, save         :: qcnmcoo
      character(len=16), dimension(:), allocatable, save   :: qcnmlab

      integer(long), dimension(7), save                    :: qcnciconf
      character(len=30), dimension(:,:), allocatable, save :: qcciconfig
      real(dop), dimension(:,:), allocatable, save         :: qccicoeff

!-----------------------------------------------------------------------
! Control Variables 
! qcorient: coordinate orientation to be read (input / standard / ...)
! 
! for Gaussian, qcorient selects orientation to be read. 
! qcorient = standard, z-matrix or input (default)
! for transition dipole
! lqctdveloc = .true. , read the velocity gauge
!-----------------------------------------------------------------------
      character(len=32), save :: qcorient = 'input'
      logical(kind=4), save   :: lqctdveloc = .false.

      contains

!#######################################################################
!
! Allocate arrays used to read QC output
!
!#######################################################################

      subroutine alloc_readqc(iqclog)

      implicit none

      integer(long), intent(in) :: iqclog
      integer(long)             :: ierr
      character(len=12)         :: cerr
      logical(kind=4)           :: lcheck
      integer(long)             :: stdim,socdim,sp

!-----------------------------------------------------------------------
! Check dimensions have been set
! qc_nbasis may be 0 if basis not read.
!-----------------------------------------------------------------------
      lcheck = .false.
      if (qc_ncoo .eq. 0) then
         write(iqclog,'(/,a)') 'Routine Alloc_readqc (readqc.F90)'
         write(iqclog,'(a)') &
            'ERROR: Asked for QC Output but qc_ncoo=0'
         write(iqclog,'(a)') ' '
         lcheck = .true.
      endif
      if (qc_nmode .eq. 0) then
         write(iqclog,'(/,a)') 'Routine Alloc_readqc (readqc.F90)'
         write(iqclog,'(a)') &
            'ERROR: Asked for QC Output but qc_nmode=0'
         write(iqclog,'(a)') ' '
         lcheck = .true.
      endif
      if (qc_nst .eq. 0) then
         write(iqclog,'(/,a)') 'Routine Alloc_readqc (readqc.F90)'
         write(iqclog,'(a)') &
            'ERROR: Asked for QC Output but qc_nst=0'
         write(iqclog,'(a)') ' '
         lcheck = .true.
      endif
      qc_natm=qc_ncoo/3

      if (lcheck) then
         write(iqclog,'(/,a)') 'Routine Alloc_readqc (readqc.F90)'
         write(iqclog,'(a)') &
            'ERROR: Stopped with Dimensioning Error.'
         write(iqclog,'(a)') ' '
         stop
      endif

!-----------------------------------------------------------------------
! Set up multiplicity mapping
!-----------------------------------------------------------------------
      qc_nspblk = count(qc_nmult > 0)
      if (qc_nspblk .gt. 0) then
! if multiplicities defined, get total no. of SOC states
         qc_nstdeg = 0
         do sp=1,7
            qc_nstdeg = qc_nstdeg + qc_nmult(sp)*sp
         enddo
      else if (qc_nspblk .eq. 0) then
! if multiplicities not defined, assume all are singlets.
         qc_nspblk = 1
         qc_nmult(1) = qc_nst
         qc_nstdeg = qc_nst
      endif
! calulate offset for each spin block in the list of states
      qc_spboff(1)=0
      do sp =2,7
         qc_spboff(sp) = qc_spboff(sp-1)+qc_nmult(sp-1)
      enddo

!-----------------------------------------------------------------------
! set dimension for state arrays to qc_nst+1 to give space for 
! information about states with Nst > qc_nst
!-----------------------------------------------------------------------
      stdim = qc_nst+1
      socdim = qc_nstdeg+1

! check all arrays are not allocated
      call dealloc_readqc

      cerr = 'qcv'
      allocate(qcv(stdim), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qcderiv1'
      allocate(qcderiv1(qc_ncoo,stdim), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qcderiv2'
      allocate(qcderiv2(qc_ncoo,qc_ncoo,stdim), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qcmocoeff'
      allocate(qcmocoeff(qc_nbasis,qc_nbasis), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qcnact'
      allocate(qcnact(qc_ncoo,stdim,stdim), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qcdipole'
      allocate(qcdipole(3,stdim), stat=ierr)
      if (ierr > 0) go to 1000
      cerr = 'qctdipole'
      allocate(qctdipole(4,stdim,stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qcsoc'
      allocate(qcsoc(socdim,socdim), stat=ierr) 
      if (ierr > 0) go to 1000

      cerr = 'qciatsym'
      allocate(qciatsym(qc_ncoo), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qciatiso'
      allocate(qciatiso(qc_ncoo), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qcxcoo'
      allocate(qcxcoo(qc_ncoo), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qcfreq'
      allocate(qcfreq(qc_nmode), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qcnmcoo'
      allocate(qcnmcoo(qc_ncoo,qc_nmode), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qcnmlab'
      allocate(qcnmlab(qc_nmode), stat=ierr) 
      if (ierr > 0) go to 1000

      cerr = 'qcciconfig'
      allocate(qcciconfig(qc_ncivec,7), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'qccicoeff'
      allocate(qccicoeff(qc_ncivec,stdim), stat=ierr) 
      if (ierr > 0) go to 1000

      cerr = 'lqcfound3'
      allocate(lqcfound3(stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound4'
      allocate(lqcfound4(stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound5'
      allocate(lqcfound5(stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound7'
      allocate(lqcfound7(stdim,qc_nst), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound8'
      allocate(lqcfound8(stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound9'
      allocate(lqcfound9(stdim,stdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound11'
      allocate(lqcfound11(socdim,socdim), stat=ierr) 
      if (ierr > 0) go to 1000
      cerr = 'lqcfound12'
      allocate(lqcfound12(qc_ncivec,stdim), stat=ierr) 
      if (ierr > 0) go to 1000

      return

 1000 write(iqclog,'(a)') 'Error allocating memory in ALLOC_READQC :',cerr
      write(iqclog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine alloc_readqc

!#######################################################################
!
! Deallocate arrays used to read QC output
!
!#######################################################################

      subroutine dealloc_readqc

      if (allocated(qcv)) deallocate(qcv)
      if (allocated(qcderiv1)) deallocate(qcderiv1)
      if (allocated(qcderiv2)) deallocate(qcderiv2)
      if (allocated(qcmocoeff)) deallocate(qcmocoeff)
      if (allocated(qcnact)) deallocate(qcnact)
      if (allocated(qcdipole)) deallocate(qcdipole)
      if (allocated(qctdipole)) deallocate(qctdipole)
      if (allocated(qcsoc)) deallocate(qcsoc)

      if (allocated(qciatsym)) deallocate(qciatsym)
      if (allocated(qciatiso)) deallocate(qciatiso)
      if (allocated(qcxcoo)) deallocate(qcxcoo)
      if (allocated(qcfreq)) deallocate(qcfreq)
      if (allocated(qcnmcoo)) deallocate(qcnmcoo)
      if (allocated(qcnmlab)) deallocate(qcnmlab)

      if (allocated(qcciconfig)) deallocate(qcciconfig)
      if (allocated(qccicoeff)) deallocate(qccicoeff)

      if (allocated(lqcfound3)) deallocate(lqcfound3)
      if (allocated(lqcfound4)) deallocate(lqcfound4)
      if (allocated(lqcfound5)) deallocate(lqcfound5)
      if (allocated(lqcfound7)) deallocate(lqcfound7)
      if (allocated(lqcfound8)) deallocate(lqcfound8)
      if (allocated(lqcfound9)) deallocate(lqcfound9)
      if (allocated(lqcfound11)) deallocate(lqcfound11)
      if (allocated(lqcfound12)) deallocate(lqcfound12)

      return
      end subroutine dealloc_readqc

!#######################################################################
! RDQCOUT
!
! Driver routine for reading QC output files
! iunit    : unit connected to log file
! ifiletyp : Type of output file to be read
! imethod  : type of calculation to be read
! iqclog   : unit for log messages
! lfound   : If true, error if all information requested isn't found
! lfail    : If set to true, QC had failed.
!
! Calls specific routines according to file type
!
! Control of reading:
! imethod = 0 : energies, gradients + dcp not read
! qc_nbasis=0 : MO coeffs not read
!
! lfail : If set true while reading output, a failure in the QC has been
!         found
!#######################################################################

      subroutine rdqcout(iunit,ifiletyp,imethod,iqclog,lfound,lfail)

      implicit none

      integer(long), intent(in)                           :: iunit,iqclog
      integer(long), intent(in)                           :: imethod,ifiletyp
      logical(kind=4), intent(in)                         :: lfound
      logical(kind=4), intent(out)                        :: lfail
      logical(kind=4)                                     :: lfailqc,lcheck

!-----------------------------------------------------------------------
! check file is at the start
!-----------------------------------------------------------------------
      if (.not. qc_enquire) then
         inquire(iunit,opened=lcheck)
         if (.not. lcheck) then
            write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
            write(iqclog,'(a)') &
               'ERROR: File to be read has not been opened.'
            write(iqclog,'(a)') ' '
            stop
         endif
         rewind(iunit)
      endif

      lfailqc = .false.
!-----------------------------------------------------------------------
! Call correct routine to read output
!-----------------------------------------------------------------------
      if (ifiletyp .eq. 1) then
         call rdgaussian(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 2) then
         call rdgamess_uk(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 3) then
         call rdqommma(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 4) then
         call rdbagel(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 5) then
         call rdmolpro(iunit,imethod,iqclog,lfailqc)

      else if (ifiletyp .eq. 6) then
         call rddftmrci(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 7) then
         call rdturbomole(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 8) then
         call rdgamess_us(iunit,imethod,iqclog)

      else if (ifiletyp .eq. 10) then
         call rdqchem(iunit,imethod,iqclog,lfailqc)

      else if (ifiletyp .eq. 11) then
         call rdmolcas(iunit,imethod,iqclog,lfailqc)

      else if (ifiletyp .eq. 12) then
         call rdorca(iunit,imethod,iqclog)

      else if (qc_enquire) then
         return

      else
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(a,i0)') 'ERROR: Unknown File type :',ifiletyp
         write(iqclog,'(a)') ' '
         stop

      endif

!-----------------------------------------------------------------------
! Check output for failed calculation
! If lfound set by calling routine, set lfail
!-----------------------------------------------------------------------
      if (lfailqc) then
         if (lfound) then
            write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
            write(iqclog,'(a)') 'ERROR: QC Failed- check output.'
            write(iqclog,'(a)') ' '
            lfail = .true.
         else
            write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
            write(iqclog,'(a)') 'WARNING: QC Failed- check output.'
            write(iqclog,'(a)') ' '
         endif
      endif

      return
      end subroutine rdqcout

!#######################################################################
! INITQCOUT
!
! Routine for initializing the global variable used in rdqcout
!
!#######################################################################

      subroutine initqcout(lfail)

      implicit none

      logical(kind=4), intent(out)  :: lfail

!-----------------------------------------------------------------------
! zero arrays
!-----------------------------------------------------------------------
      qcv = 0.0_dop
      qcderiv1 = 0.0_dop
      qcderiv2 = 0.0_dop
      qcmocoeff = 0.0_dop
      qcnact = 0.0_dop
      qcdipole = 0.0_dop
      qctdipole = 0.0_dop
      qcsoc = 0.0_dop

      qciatsym = 0 
      qciatiso = 0 
      qcxcoo = 0.0_dop
      qcfreq = 0.0_dop
      qcnmcoo = 0.0_dop
      qcnmlab = ' '

!-----------------------------------------------------------------------
! zero found flags and define useful variables
!-----------------------------------------------------------------------
      qcfound(:) = .false.
      lqcfound3(:) = .false.
      lqcfound4(:) = .false.
      lqcfound5(:) = .false.
      lqcfound7(:,:) = .false.
      lqcfound8(:) = .false.
      lqcfound9(:,:) = .false.
      lqcfound11(:,:) = .false.

      lfail = .false.

      return
      end subroutine initqcout

!#######################################################################
! CHKQCOUT
!
! Routine for checking out the content of the QC output files and if
! data requested has been found
!
! lfound   : If true, error if all information requested isn't found
! lfail : If set true while reading output, a failure in the QC has been
!         found
!#######################################################################

      subroutine chkqcout(ifiletyp,iqclog,lfound,lfail)

      implicit none

      integer(long), intent(in)      :: ifiletyp,iqclog 
      logical(kind=4), intent(in)    :: lfound
      logical(kind=4), intent(out)   :: lfail
      character(len=10)              :: string

      lfail = .false.
!-----------------------------------------------------------------------
! Check output requested has been found
!
! Common errors for different codes can be added 
!-----------------------------------------------------------------------
      if (lfound) then
         string='Error: '
      else
         string='Warning: '
      endif

      if (qcread(1) .and. .not. qcfound(1)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Coordinates asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(2) .and. .not. qcfound(2)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Normal Modes asked for but not found.'
         if (ifiletyp .eq. 1) write(iqclog,'(a)') &   ! Gaussian 
               'In Gaussian, check freq=hpmodes is specified.'
         if (ifiletyp .eq. 10) write(iqclog,'(a)') &   ! QChem
               'In QCHem, check "vibman_print 6" is specified.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(3) .and. .not. qcfound(3)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Energy asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(4) .and. .not. qcfound(4)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' First Derivatives asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(5) .and. .not. qcfound(5)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Second Derivatives asked for but not found.'
         if (ifiletyp .eq. 1) write(iqclog,'(a)') &   ! Gaussian 
           'In Gaussian, for large molecules check if IOp(7/33=1) is specified.'
         if (ifiletyp .eq. 5) write(iqclog,'(a)') &   ! Molpro
             'In Molpro, check "print, hessian" is specified.'
         if (ifiletyp .eq. 10) write(iqclog,'(a)') &   ! QChem
               'In QCHem, check "vibman_print 6" is specified.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(6) .and. .not. qcfound(6)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' MO Coeffs asked for but not found.'
         if (ifiletyp .eq. 1) write(iqclog,'(a)') &   ! Gaussian 
            'In Gaussian, check pop=full is specified.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(7) .and. .not. qcfound(7)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Derivative Couplings asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(8) .and. .not. qcfound(8)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Dipole asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(9) .and. .not. qcfound(9)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Transition Dipole asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(10) .and. .not. qcfound(10)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' Isotopic data asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(11) .and. .not. qcfound(11)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' SOC data asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      else if (qcread(12) .and. .not. qcfound(12)) then
         write(iqclog,'(/,a)') 'Routine Rdqcout (readqc.F90)'
         write(iqclog,'(2a)') trim(string),&
               ' CI Vectors asked for but not found.'
         write(iqclog,'(a)') ' '
         lfail = .true.
      endif

      lfail = (lfail .and. lfound)

      return
      end subroutine chkqcout

!#######################################################################
!
! RDGAUSS
! Read a Gaussian Log file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdgaussian(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,iatnum,i1,n,n1,n2,nm,&
                                     j,k,koff,nrd,nx,s
      integer(long)               :: ilbl,jlbl,klbl,clbl,neq,&
                                     tdlbl,gslbl
      integer(long)               :: idum1,idum2,iorb
      integer(long)               :: istate,nroot,ihomo

      real(dop), dimension(3)     :: vec
      real(dop), dimension(5)     :: force
      real(dop)                   :: dtmp,egrd,energy,mag
      real(dop), dimension(qc_ncoo)         :: dtmpvec
      real(dop), dimension(qc_ncoo,qc_ncoo) :: dtmpmat
      character(len=80)           :: string
      character(len=40)           :: substring
      character(len=40)           :: eigstring
      character(len=40)           :: egsstring
      character(len=40)           :: coostring
      character(len=60)           :: tdstring
      character(len=10)           :: dummy
      logical(kind=4)             :: lswapci

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
            qcfound(10)=.true.
         else if ((imethod .ge. 1 .and. imethod .le. 6) .or. &
              imethod .eq. 10 .or. imethod .eq. 11 .or. &
              imethod .eq. 12 .or. imethod .eq. 17 .or. &
              imethod .eq. 19 .or. imethod .eq. 22 .or. &
              imethod .eq. 24 .or. imethod .eq. 25 .or. &
              imethod .eq. 33 .or. imethod .eq. 38) then
            qcfound(3)=.true.
            qcfound(4)=.true.
            qcfound(5)=.true.
            qcfound(8)=.true.
         endif
         if (imethod .eq. 11) qcfound(9)=.true.
         if (imethod .eq. 12) qcfound(7)=.true.

         return
      endif

      lswapci = .false.   ! if true, CI vectors N and N-1 swapped 
      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. For CAS, this is changed to highest state

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      ilbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY STRING"
         ilbl = 13
      else if (imethod.eq.1 .or. imethod.eq.5 .or. imethod.eq.6 .or. &
               imethod.eq.19 .or. imethod .eq. 24 .or. &
               imethod .eq. 25) then   ! SCF, DFT
         write(eigstring,'(a)') "SCF Done:  E("
         ilbl = 13
      else if (imethod .eq. 2) then   ! MP2
         write(eigstring,'(a)') "EUMP2"
         ilbl = 5 
      else if (imethod .eq. 3) then   ! MP3
         write(eigstring,'(a)') "EUMP3"
         ilbl = 5 
      else if (imethod .eq. 4) then   ! MP4
         write(eigstring,'(a)') "UMP4(SDTQ)"
         ilbl = 10 
      else if (imethod .eq. 10) then   ! OVGF
         write(eigstring,'(a)') "GREENY: Electron Propagator"
         ilbl = 27 
      else if (imethod .eq. 11) then   ! CIS
         write(eigstring,'(a)') "Excitation energies and oscillator"
         ilbl = 34 
      else if (imethod .eq. 12) then   ! CASSCF
         write(eigstring,'(a)') ")     EIGENVALUE"
         ilbl = 16
      else if (imethod .eq. 17) then   ! EOMCCSD
         write(eigstring, '(a)') "Excited State"
         ilbl = 13
      else if (imethod.eq.22) then
         write(eigstring,'(a)') "Wavefunction amplitudes converged. E(Corr)"
         ilbl = 42
      else if (imethod .eq. 30) then                       ! Ehrenfest
         write(eigstring,'(a)') "The total energy Time dep state is "
         ilbl = 35
      else if (imethod.eq.33) then                         !TDDFT
         write(eigstring,'(a)') "Excited State"
         ilbl = 13
      else if (imethod.eq.38) then        ! ONIOM model system CAS (highest layer)
         write(eigstring,'(a)') ") Final Eigenvalue"
         ilbl = 18
      else 
         write(iqclog,'(/,a)') 'Routine Rdgauss (readqc.F90)'
         write(iqclog,'(a,i5)') &
            'ERROR: QC Method not implemented :',imethod
         write(iqclog,'(a)') ' '
      endif
!-----------------------------------------------------------------------
! some methods have a different string for the GS energy
!-----------------------------------------------------------------------
      if (imethod.eq.10 .or. imethod .eq. 11 .or. imethod .eq. 33) then
         write(egsstring,'(a)') "SCF Done:  E("
         gslbl = 13
      else if (imethod.eq.17) then
         write(egsstring,'(a)') "E(CORR)"
         gslbl = 7
      else
         write(egsstring,'(a)') "NO EGS STRING"
         gslbl = 13
      endif
!-----------------------------------------------------------------------
! string for finding coordinates depends on orientation
!-----------------------------------------------------------------------
      clbl=0
      if (qcorient(1:8).eq.'standard') then
         coostring = 'Standard orientation:'
         clbl = 21
      else if (qcorient(1:8).eq.'z-matrix') then
         coostring = 'Z-Matrix orientation:'
         clbl = 21
      else 
         coostring = 'Input orientation:'
         clbl = 18
      endif

!-----------------------------------------------------------------------
! string for finding transition dipole depends on options
!-----------------------------------------------------------------------
      tdlbl=0
      if (lqctdveloc) then
         tdstring = 'Ground to excited state transition velocity dipole'
         tdlbl = 50
      else
         tdstring = 'Ground to excited state transition electric dipole'
         tdlbl = 50
      endif

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
  100    continue
         read(iunit,'(a)',end=900,err=1000) string

         if (index(string,'Swap CI vectors ivec and ivec-1') .ne. 0) then 
            lswapci = .true.

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         else if (index(string,coostring(1:clbl)) .ne. 0) then 

            do i=1,4
               read(iunit,*)
            enddo
            do i=1,qc_ncoo/3
               read(iunit,*,err=1001) idum1,qciatsym((i-1)*3+1),&
                    idum2,(qcxcoo((i-1)*3+k),k=1,3)
            enddo
! fill holes in iatsym
            qciatsym(2:qc_ncoo:3)=qciatsym(1:qc_ncoo:3)
            qciatsym(3:qc_ncoo:3)=qciatsym(1:qc_ncoo:3)

! Convert to Bohr
            qcxcoo(1:qc_ncoo) = qcxcoo(1:qc_ncoo)/angstroem

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (index(string,"Harmonic frequencies (cm**-1)") .ne. 0) then

            nrd = ((qc_nmode-1)/5) + 1  ! No. of blocks of normal modes to read

            do
               read(iunit,'(a)',err=1002) string
               if (index(string,'Frequencies --') .ne. 0) exit
            enddo

            backspace(iunit)
            backspace(iunit)
            backspace(iunit)

            nm=0
            do n=1,nrd
               nx = min((qc_nmode-5*(n-1)),5) ! no. of columns to be read
               read(iunit,*,err=1002)
               read(iunit,'(26x,5a10)') (qcnmlab(nm+j),j=1,nx)
               read(iunit,'(23x,5f10.4)',err=1002) (qcfreq(nm+j),j=1,nx)

               do
                  read(iunit,'(a)',err=1002) string
                  if (string(2:20) .eq. 'Coord Atom Element:') then
                     exit
                  else if (string(3:10) .eq. 'Atom  AN') then
                     go to 100     ! ignore frequencies - wrong format
                  endif
               enddo

               do j=1,qc_ncoo
                  read(iunit,'(23x,5f10.5)',err=1002) (qcnmcoo(j,nm+i),i=1,nx)
               enddo

               nm=nm+5 ! go to next block of normal modes
            enddo
            do n=1,4  ! read past header of non-hpmode frequencies
               read(iunit,*,err=1002)
            enddo

! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.
 
!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
! Also get root number for forces etc read later. If root number is 
! higher than qc_nst (no. of states asked for) store informatio in 
! qc_nst+1 to prevent memory overflow. Store in qc_nst+1.
!-----------------------------------------------------------------------
         else if (index(string,eigstring(1:ilbl)) .ne. 0) then 

            if (imethod .eq. 0) cycle   ! do not read energy

            if (imethod.eq.1 .or. imethod.eq.5 .or. imethod.eq.6 .or. imethod.eq.19) then
               neq=index(string,'=')+1
               read(string(neq:80),'(f16.9)',err=1003) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.2) then
               neq=index(string,'EUMP2 =')+7
               read(string(neq:80),*,err=1003) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.3) then
               neq=index(string,'EUMP3=')+6
               read(string(neq:80),*,err=1003) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.4) then
               neq=index(string,'UMP4(SDTQ)=')+11
               read(string(neq:80),*,err=1003) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.10) then
               do     ! read down to ovgf data
                  read(iunit,'(a)',err=1003) string
                  if (index(string,'HOMO') .ne. 0) then
                     read (string(27:29),'(i3)') ihomo

! Read Energies
                  else if (index(string,'recommended value') .ne. 0) then
                     read(iunit,*)
                     read(iunit,*) iorb,dtmp,energy

! Get state number (HOMO is 1, HOMO-1 is 2, ....)
                     istate = (ihomo-iorb)+1
                     istate = min(istate,qc_nst+1)  ! trap unwanted states
                     if (istate .gt. 0) then
! Distinguish between ionisation energies and electron affinities,
! as Gaussian uses an annoying sign convention
                        if (iorb.gt.0) energy=abs(energy)
                        qcv(istate)=(energy/uev)+egrd   ! energy in Hartree
                        qcfound(3)=.true.
                        lqcfound3(istate)=.true.
                     endif
                  else if (index(string,'are degenerate') .ne. 0) then
                     read (string(55:57),'(i3)') iorb
                     istate = (ihomo-iorb)+1
                     istate = min(istate,qc_nst+1)  ! trap unwanted states
                     if (istate .gt. 0) then
                        qcv(istate)=qcv(istate+1)   ! energy in Hartree
                        qcfound(3)=.true.
                        lqcfound3(istate)=.true.
                     endif

                  else if (index(string,'Leave Link') .ne. 0) then
                     exit
                  endif
               enddo

            else if (imethod.eq.11) then
               do
                  read(iunit,'(a)',err=1003) string
                  if ((index(string,'Excited State') .ne. 0) .and. &
                        (index(string,'eV') .ne. 0)) then

                     jlbl = index(string,'State')+5
                     klbl = index(string,':')-1
                     read(string(jlbl:klbl),*) istate
                     istate = istate+1    ! CIS counting root=1 is S1
                     istate = min(istate,qc_nst+1)  ! trap unwanted states
                     jlbl = index(string,'eV')-10
                     klbl = index(string,'eV')-1
                     read(string(jlbl:klbl),*) energy
                     qcv(istate)=(energy/uev)+egrd   ! energy in Hartree
                     qcfound(3)=.true.
                     lqcfound3(istate)=.true.

                  else if (index(string,'This state for optimization').ne.0) then
                     nroot = istate    ! save state for forces etc.

                  else if (index(string,'Leave Link  914').ne.0) then
                     exit
                  endif
               enddo

            else if (imethod.eq.12) then
               jlbl=index(string,'(')+1
               klbl=index(string,')')-1
               read(string(jlbl:klbl),*,err=1003) istate  ! State
               istate = min(istate,qc_nst+1)  ! trap unwanted states
               nroot = max(nroot,istate)                       ! get highest state

               neq=23
               read(string(neq:80),'(f19.10)',err=1003) qcv(istate)
               qcfound(3)=.true.
               lqcfound3(istate)=.true.

            else if (imethod.eq.17) then
               klbl=index(string,':')-1
               jlbl=klbl-2
               read(string(jlbl:klbl),'(i3)') istate
               istate = istate+1
               istate = min(istate,qc_nst+1)  ! trap unwanted states
               klbl=index(string,'eV')-2
               jlbl=klbl-9
               read(string(jlbl:klbl),*) energy
               qcv(istate)=(energy/uev)+egrd   ! energy in Hartree
               qcfound(3)=.true.
               lqcfound3(istate)=.true.

            else if (imethod.eq.22) then
               klbl=index(string,'Delta')-1
               jlbl=klbl-20
               read(string(jlbl:klbl),*) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.30) then
               neq=37
               read(string(neq:80),'(f13.5)',err=1003) qcv(1)
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.33) then
               jlbl = index(string,'State')+6
               klbl = index(string,':')-1
               read(string(jlbl:klbl),*) istate
               istate = istate+1
               istate = min(istate,qc_nst+1) 
               jlbl = index(string,'eV')-10
               klbl = index(string,'eV')-1
               read(string(jlbl:klbl),*) energy
               qcv(istate)=(energy/uev)+egrd   ! energy in Hartree
               qcfound(3)=.true.
               lqcfound3(istate)=.true.

            else if (imethod.eq.38) then
               jlbl=index(string,'(')+1
               klbl=index(string,')')-1
               read(string(jlbl:klbl),*,err=1003) istate  ! State
               istate = min(istate,qc_nst+1)  ! trap unwanted states
               nroot = max(nroot,istate)                       ! get highest state

               neq=33
               read(string(neq:80),'(f15.7)',err=1003) qcv(istate)
               qcfound(3)=.true.
               lqcfound3(istate)=.true.

            else
               write(iqclog,'(/,a,i5)') &
                   'ERROR: Unknown ab initio type for Gaussian output :',imethod
               qcfound(3)=.false.

            endif

!-----------------------------------------------------------------------
! GS ENERGY (required by e.g. OVGF, CIS)
!-----------------------------------------------------------------------
         else if (index(string,egsstring(1:gslbl)) .ne. 0) then 

            if (imethod .eq. 0) cycle   ! do not read energy

            if (imethod.eq.11 .or. imethod.eq.33) then ! for CIS & TDDFT, GS energy is S0
               neq=index(string,'=')+1
               read(string(neq:80),'(f16.9)',err=1003) egrd

               qcv(1) = egrd
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            else if (imethod.eq.17) then  ! read CCSD for EOM-CCSD
               klbl=index(string,'Delta')-1
               jlbl=index(string,egsstring(1:gslbl))+8
               read(string(jlbl:klbl),*) egrd

               qcv(1) = egrd
               qcfound(3)=.true.
               lqcfound3(1)=.true.

            endif

!-----------------------------------------------------------------------
! FIRST DERIVATIVES
!-----------------------------------------------------------------------
         elseif (index(string, "Forces (Hartrees/Bohr)").ne.0) then

            if (imethod .eq. 0) cycle   ! do not read gradients

            read(iunit,*)
            read(iunit,*)
            koff = 0
            do n=1,qc_natm
               read(iunit,*,err=1004) n1,iatnum,(force(i),i=1,3)
               do i=1,3
                  qcderiv1(koff+i,nroot)=-force(i)
               enddo
               koff=koff+3
            enddo
 
            qcfound(4)=.true.
            lqcfound4(nroot)=.true.

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
!-----------------------------------------------------------------------
         elseif (index(string, "Force constants in Cartesian").ne.0) then

            if (imethod .eq. 0) cycle   ! do not read gradients

            do i1 =1,qc_ncoo,5 !read different chunks of the hessian
               read(iunit,*)
               do n=i1,qc_ncoo
                  n1 = min(5,n-i1+1)
! Hessian calculations from Gaussian are not converged to the decimal places 
! displayed this can cause reproducibility issues, and differences in the 
! dynamics 
!DEBUG: read 4 digits from hessian only
!              read(iunit,'(5x,i2,5(x,f7.4,3x,i3))',err=1005) &
!                           iatnum,(forcenum(i),forcemag(i),i=1,n1)
!              force(:)=forcenum(:)*10.d0**forcemag(:)
!end DEBUG
                   read(iunit,*,err=1005) iatnum,(force(i),i=1,n1)
                   do i = 1,n1
                      n2 = i1+i-1
                      qcderiv2(n,n2,nroot) = force(i)
                      if (n .ne. n2) qcderiv2(n2,n,nroot) = force(i)
                   enddo
                enddo
            enddo

            qcfound(5)=.true.
            lqcfound5(nroot)=.true.

!-----------------------------------------------------------------------
! MO COEFFICIENTS
!-----------------------------------------------------------------------
         elseif (index(string,"Molecular Orbital Coefficients") .ne. 0) then

            if (qc_nbasis .le. 0) cycle    !  Do not read MOs

            koff = 0
            do i = 1,qc_nbasis/5
               read(iunit,*)
               read(iunit,*)
               read(iunit,*)
               do j =1, qc_nbasis
                  read(iunit,'(a21,5(1x,f9.5))',err=1006) &
                       substring,(qcmocoeff(k+koff,j),k=1,5)
               enddo
               koff = koff + 5
            enddo
 
            read(iunit,*)
            read(iunit,*)
            read(iunit,*)
            do j =1, qc_nbasis
               read(iunit,fmt='(a21,5(1x,f9.5))',err=1006) &
                    substring,(qcmocoeff(k+koff,j),k=1,mod(qc_nbasis,5))
            enddo

            qcfound(6)=.true.

!-----------------------------------------------------------------------
! DERIVATIVE COUPLINGS
!-----------------------------------------------------------------------
         elseif (string(2:20) .eq. "Derivative Coupling") then

            if (imethod .eq. 0) cycle   ! do not read gradients

! see which states are coupled
            backspace(iunit)
            backspace(iunit)
            read(iunit,'(a)') string
            if (string(1:22) .eq. '     Energy difference') then
               backspace(iunit)
               backspace(iunit)
               read(iunit,'(10x,i5,8x,i5)') i,j
            elseif (string(1:18) .eq. '  Crude BS vectors') then
               backspace(iunit)
               backspace(iunit)
               backspace(iunit)
               read(iunit,'(10x,i5,8x,i5)') i,j
               read(iunit,'(a)') string
            endif

            read(iunit,'(a)') string
            read(iunit,'(a)') string

            koff = 0
            do n = 1,qc_natm
               read(iunit,fmt='(x,3(7x,f13.10))',err=1007) &
                     (qcnact(k+koff,i,j),k=1,3)
               koff = koff + 3
            enddo
            qcnact(:,j,i) = qcnact(:,i,j)

            lqcfound7(i,j)=.true.
            lqcfound7(j,i)=.true.
            qcfound(7)=.true.

!-----------------------------------------------------------------------
! DIPOLE MOMENT
!-----------------------------------------------------------------------
         else if (index(string,"Dipole moment (field-") .ne. 0) then
            read(iunit,*,err=1008) (dummy,qcdipole(n,1),n=1,3)

! Convert dipole from debye to au
            qcdipole(:,1) = qcdipole(:,1)*0.393430307_dop

            qcfound(8)=.true.
            lqcfound8(1)=.true.

!-----------------------------------------------------------------------
! TRANSITION DIPOLE
!-----------------------------------------------------------------------
         else if (index(string,tdstring(1:tdlbl)) .ne. 0) then

            if (imethod .eq. 0) cycle   ! do not read 
            if (imethod .ne. 11) cycle  ! Only for CIS

            read(iunit,'(a)',err=1009) string
            do 
               read(iunit,'(a)',err=1009) string
               if (index(string,'Ground to excited') .ne. 0) exit

               read(string,*,err=1009) s,(vec(i),i=1,3)
               s = s+1

               mag=0.0_dop
               do i=1,3
                  mag=mag + vec(i)*vec(i)
               enddo
               if (mag .ge. 1.0e-8_dop) then
! tdip(4) is magnitude of dipole
                  mag=dsqrt(mag)
                  qctdipole(4,1,s)=mag

! tdip(i), i=1-3 is unit vector along dipole
                  do i=1,3
                     vec(i) = vec(i) / mag
                  enddo
! as convention take vector direction with positive dot product with (1,1,1)
                  mag=0.0_dop
                  do i=1,3
                     mag = mag + vec(i)
                  enddo
                  if (mag .lt. 0.0_dop) then
                     do i=1,3
                        vec(i)=-vec(i)
                     enddo
                  endif
                  do i=1,3
                     qctdipole(i,1,s) = vec(i)
                  enddo

                  qctdipole(:,s,1) = qctdipole(:,1,s) 

                  qcfound(9)=.true.
                  lqcfound9(1,s)=.true.
                  lqcfound9(s,1)=.true.

               else if (s .gt. 1) then ! TDIP=0
                  qcfound(9)=.true.
                  lqcfound9(1,s)=.true.
                  lqcfound9(s,1)=.true.
               endif
            enddo

!-----------------------------------------------------------------------
! ISOTOPES
!-----------------------------------------------------------------------
         else if (index(string,"Isotopes and Nuclear Properties:") .ne. 0) then

            read(iunit,*)
            read(iunit,*)
            nrd = ((qc_natm-1)/10) + 1  ! No. of blocks of isotopes to read
            nm = 0
            do n=1,nrd
               nx = min((qc_natm-10*(n-1)),10) ! no. of columns to be read
               read(iunit,*)
               read(iunit,*)
               read(iunit,*,err=1010) dummy,(qciatiso((nm+k-1)*3+1),k=1,nx)
               do j=1,5
                  read(iunit,*)
               enddo
               nm = nm+10
             enddo
!fill holes in iatiso
             qciatiso(2:qc_ncoo:3)=qciatiso(1:qc_ncoo:3)
             qciatiso(3:qc_ncoo:3)=qciatiso(1:qc_ncoo:3)

             qcfound(10)=.true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file
      if (.not. qcfound(3)) lswapci=.false.   ! If energies not read, 
                                              ! turn off swapping

!-----------------------------------------------------------------------
! Swap roots N and N-1
!-----------------------------------------------------------------------
      if (lswapci) then
         if (qcfound(3)) then
            dtmp = qcv(nroot)
            qcv(nroot) = qcv(nroot-1)
            qcv(nroot-1) = dtmp
         endif

         if (lqcfound4(nroot-1)) then
            dtmpvec(1:qc_ncoo) = qcderiv1(1:qc_ncoo,nroot)
            qcderiv1(:,nroot) = qcderiv1(:,nroot-1)
            qcderiv1(1:qc_ncoo,nroot-1) = dtmpvec(1:qc_ncoo)
         else
            qcderiv1(:,nroot-1) = qcderiv1(:,nroot)
            qcderiv1(:,nroot) = 0.0_dop
            lqcfound4(nroot-1) = .true.
            lqcfound4(nroot) = .false.
         endif

         if (lqcfound5(nroot-1)) then
            dtmpmat(1:qc_ncoo,1:qc_ncoo) = qcderiv2(1:qc_ncoo,1:qc_ncoo,nroot)
            qcderiv2(:,:,nroot) = qcderiv2(:,:,nroot-1)
            qcderiv2(1:qc_ncoo,1:qc_ncoo,nroot-1) = dtmpmat(1:qc_ncoo,1:qc_ncoo)
         else
            qcderiv2(:,:,nroot-1) = qcderiv2(:,:,nroot)
            qcderiv2(:,:,nroot) = 0.0_dop
            lqcfound5(nroot-1) = .true.
            lqcfound5(nroot) = .false.
         endif
      endif

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Energy)'
      stop
 1004 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (1st derivatives)'
      stop
 1005 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (2nd derivatives)'
      stop
 1006 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (MO coeffs.)'
      stop
 1007 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (NACTs)'
      stop
 1008 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Dipoles)'
      stop
 1009 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Transition Dipoles)'
      stop
 1010 write(iqclog,'(/,a)') 'ERROR reading GAUSSIAN output (Isotopes)'
      stop
      end subroutine rdgaussian


!#######################################################################
!
! RDGAMESS_UK
! Read a GAMESS-UK Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdgamess_uk(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,j,k,n,isym,nrd,zero,nmold
      integer(long)               :: nroot,nm

      real(dop), dimension(3)     :: coo
      real(dop), dimension(10)    :: xdum,xdum1,xdum2
      character(len=80)           :: string

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
         else if (imethod .eq. 1) then
            qcfound(3)=.true.
            qcfound(4)=.true.
         endif

         return
      endif

      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. 

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         if (string(31:59) .eq. 'molecular geometry (angstroem)') then

            read(iunit,*,err=1001)
            read(iunit,*,err=1001)
            read(iunit,*,err=1001)
            read(iunit,*,err=1001)
            n=0
            do i=1,qc_ncoo,3
               read(iunit,20,err=1001) (coo(k),k=1,3),isym
               do j=1,3
                  n=n+1
                  qciatsym(n)=isym
                  qcxcoo(n)=coo(j)
               enddo
            enddo

! Convert to Bohr
            qcxcoo(1:qc_ncoo) = qcxcoo(1:qc_ncoo)/angstroem

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (string(21:79) .eq. &
          'cartesians to normal mode coordinates transformation matrix') then

            read(iunit,*,err=1002)
            nrd = (((qc_nmode+6)-1)/8) + 1

            nm=0
            do n=1,nrd
               zero=0
               do i=1,5
                  read(iunit,*,err=1002)
               enddo
               read(iunit,'(16x,8f11.2)',err=1002) (xdum1(j),j=1,8)
               do j=1,8
                  if (abs(xdum1(j)).ge.(0.001)) then
                     nm=nm+1
                     qcfreq(nm)=xdum1(j)
                     write(qcnmlab(nm),'(a,i0)') 'Q',nm  ! no labels in file
                  else
                     zero=zero+1
                  endif
               enddo
               do j=1,2
                  read(iunit,*,err=1002)
               enddo
               nmold=nm-8+zero
               do j=1,qc_ncoo
                  read(iunit,'(16x,8f11.6)',err=1002) (xdum2(i),i=1,8)
                  nm=nmold
                  do i=1,8
                     if (abs(xdum1(i)).ge.(0.001)) then
                        nm=nm+1
                        qcnmcoo(j,nm)=xdum2(i)
                     endif
                  enddo
               enddo
            enddo

! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         else if(string(21:32).eq.'total energy') then
            if (imethod .ne. 1) cycle   ! only for HF

            read(string(41:56),'(f16.9)',err=1003) qcv(nroot)

            qcfound(3)=.true.
            lqcfound3(nroot) = .true.

         else if(string(11:22).eq.'total energy') then
            if (imethod .ne. 1) cycle   ! only for HF

            read(string(31:46),'(f16.9)',err=1003) qcv(nroot)

            qcfound(3)=.true.
            lqcfound3(nroot) = .true.

!-----------------------------------------------------------------------
! FIRST DERIVATIVES
!-----------------------------------------------------------------------
         else if (string(2:23) .eq. 'gradient of the energy') then

            nrd=(qc_ncoo-1)/24+1
            n=0
            do i=1,nrd
               do j=1,6
                  read(iunit,*,err=1004)
               enddo
               do j=1,3
                  read(iunit,30,err=1004) (xdum(k),k=1,8)
                  do k=0,7
                     qcderiv1(j+3*k,nroot)=xdum(k+1)
                  enddo
               enddo
            enddo

            qcfound(4)=.true.
            lqcfound4(nroot)=.true.


! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

   20 format(14x,3(f20.10),5x,i2)
   30 format(7x,8f15.7)

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading GAMESS-UK output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading GAMESS-UK output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading GAMESS-UK output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading GAMESS-UK output (Energy)'
      stop
 1004 write(iqclog,'(/,a)') 'ERROR reading GAMESS-UK output (1st derivatives)'
      stop
      end subroutine rdgamess_uk

!#######################################################################
!
! RDQOMMMA
! Read a QOMMMA Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdqommma(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,iatnum,n,n2,frow,nfullrws
      integer(long)               :: nroot

      real(dop), dimension(5)     :: force
      character(len=80)           :: string
      character(len=4)            :: atmname

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 29) then
            qcfound(3)=.true.
            qcfound(4)=.true.
            qcfound(5)=.true.
         endif

         return
      endif

      if (imethod .ne. 29) return   ! only for QOMMMA

      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. For CAS, this is changed to highest state

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         if (string(1:25) .eq. "Present, previous energy:") then
            if (imethod .eq. 0) cycle   ! do not read energies

            read(string,'(28x,f16.9)',err=1003) qcv(nroot)

            qcfound(3)=.true.
            lqcfound3(nroot) = .true.

!-----------------------------------------------------------------------
! FIRST DERIVATIVES
!-----------------------------------------------------------------------
         else if (string(1:32) .eq. "Total Gradient (no constraints):") then
            if (imethod .eq. 0) cycle   ! do not read gradients

            do n=1,qc_natm
               read(iunit,*,err=1004) atmname,(qcderiv1((n-1)*3+i,nroot),i=1,3)
            enddo

            qcfound(4)=.true.
            lqcfound4(nroot)=.true.

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
!-----------------------------------------------------------------------

         else if (string(1:13) .eq. "Total Hessian") then
            do n=1,qc_ncoo
               nfullrws=n/4
               do frow=1,nfullrws
               read(iunit,*,err=1005) iatnum,(force(i),i=1,4)
                  do i = 1,4
                     n2 = (frow*4)-(4-i)
                     qcderiv2(n,n2,nroot) = force(i)
                     if (n .ne. n2) qcderiv2(n2,n,nroot) = force(i)
                  enddo
               enddo
               if (mod(n,4) .ne. 0) then
               read(iunit,*,err=1005) iatnum,(force(i),i=1,mod(n,4))
                  do i = 1,mod(n,4)
                     n2 = (nfullrws*4)+i
                     qcderiv2(n,n2,nroot) = force(i)
                     if (n .ne. n2) qcderiv2(n2,n,nroot) = force(i)
                  enddo
               endif
            enddo

            qcfound(5)=.true.
            lqcfound5(nroot)=.true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file


      return
 1000 write(iqclog,'(/,a)') 'ERROR reading QOMMMA output'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading QOMMMA output (Energy)'
      stop
 1004 write(iqclog,'(/,a)') 'ERROR reading QOMMMA output (1st derivatives)'
      stop
 1005 write(iqclog,'(/,a)') 'ERROR reading QOMMMA output (2nd derivatives)'
      stop
      end subroutine rdqommma

      

!#######################################################################
!
! RDMOLPRO
! Read a Molpro Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
! lfail   : Set to true if QC has failed
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdmolpro(iunit,imethod,iqclog,lfail)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,i1,n,n1,n2,nm,sdx,&
                                     j,k,koff,nrd,nblk,nimag,&
                                     il,jl,kl,ioffs,joffs,noffs,s,s1
      integer(long)               :: gslbl,ilbl,jlbl,klbl,flbl,neq,&
                                     nmlbl,i1lbl,soclbl
      integer(long)               :: istate,jstate,issym
      integer(long)               :: imcespin,nsocs,nst
      integer(long)               :: idum1,niso,isoidx,nbaschk,ierr
      integer(long), dimension(qc_nst) :: ipt2map
      integer(long), dimension(:,:), allocatable  :: ispmatlab
      integer(long)               :: nci,soff

      real(dop)                   :: maxci
      real(dop)                   :: isym,rdum,rdum1,rdum2
      real(dop)                   :: vdiff,vtmp
      real(dop)                   :: egrd
      real(dop)                   :: energy
      real(dop), dimension(5)     :: force
      real(dop), dimension(qc_nst)                    :: qcv1,tmpvec
      real(dop), dimension(qc_ncoo,qc_nst+1)          :: tmpderiv1
      real(dop), dimension(qc_ncoo,qc_ncoo,qc_nst+1)  :: tmpderiv2
      real(dop), dimension(qc_ncoo,qc_nst+1,qc_nst+1) :: tmpnact
      real(dop), dimension(:,:,:), allocatable        :: tmpdsoc
      real(dop), dimension(qc_nst,qc_nst)             :: sci
      real(dop), dimension(qc_ncivec,qc_nst)          :: tconf
      character(len=120)          :: string
      character(len=80)           :: mostring
      character(len=40)           :: eigstring, grastring, egsstring
      character(len=50)           :: nmstring
      character(len=40)           :: eig1string
      character(len=10)           :: ststring
      character(len=60)           :: socstring, matstring
      character(len=5)            :: ccoord
      character(len=80)           :: fmttyp,stringtmp
      character(len=7), dimension(qc_natm) :: isonam,atomname
      real(dop), dimension(qc_natm)        :: isomass
      logical(kind=4), dimension(qc_nst)   :: lfounde
      logical(kind=4)              :: liso,lorbtr,lorbtrchk,lfoundmap
      logical(kind=4), intent(out) :: lfail

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
            qcfound(10)=.true.
         endif
         if (imethod .eq. 1 .or. imethod .eq. 2 .or. &
            (imethod .ge. 11 .and. imethod .le. 15) .or. &
            imethod .eq. 17 .or. imethod .eq. 22 .or. &
            imethod .eq. 26 .or. imethod .eq. 32 .or. &
            imethod .eq. 34) then
            qcfound(3)=.true.
         endif
         if (imethod .eq. 1 .or. imethod .eq. 2 .or. &
            imethod .eq. 12 .or. imethod .eq. 13 .or. &
            imethod .eq. 15 .or. imethod .eq. 32) then
            qcfound(4)=.true.
            qcfound(5)=.true.
            qcfound(6)=.true.
         endif
         if (imethod .eq. 12 .or. imethod .eq. 15 .or. &
             imethod .eq. 32) then
            qcfound(7)=.true.
         endif
         if (imethod .eq. 1) qcfound(8)=.true.
         if (imethod .eq. 14 .or. imethod .eq. 15) qcfound(11)=.true.

         return
      endif

! By default on GS
      istate = 1                  

! imcespin : Spin of wavefunction for next read MCSCF info
      imcespin = 0

! MO matrix transpose if .true.
      lorbtr=.false.
      lorbtrchk=.false.

! Flag that energies have been found for calc of NACTs
      lfounde(:) = .false.

! Flag that CSA <-> PT2  mapping has been found 
      lfoundmap = .false.

! If qc_nirrep = 1, make qc_nsymorb(1) it equal to nbasis
      if (qc_nirrep .eq. 1) qc_nsymorb(1) = qc_nbasis

! no. isotopes
      niso=0
      liso=.false.

! Keep track of number of states' energies read.
      sdx = 0

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      ilbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY NULL STRING"
         ilbl = 18
      else if (imethod.eq.1) then                        ! SCF
         write(eigstring,'(a)') "!RHF STATE 1.1 Energy"
         ilbl = 21
      else if (imethod.eq.2) then                        ! MP2
         write(eigstring,'(a)') "!MP2 total energy"
         ilbl = 17
      else if (imethod.eq.11) then                       ! CIS
         write(eigstring,'(a)') "CIS-Results"
         ilbl = 11
      else if (imethod.eq.12) then                       ! CAS
         write(eigstring,'(a)') "!MCSCF STATE"
         ilbl = 12
      else if (imethod.eq.13 .or. imethod .eq. 32) then  ! CASPT2
         write(eigstring,'(a)') "!RSPT2 STATE"
         ilbl = 12
      else if (imethod.eq.14 .or. imethod .eq. 15) then  ! MRCI
         write(eigstring,'(a)') "!MRCI STATE"
         ilbl = 11
      else if (imethod.eq.22) then  ! CCSD
         write(eigstring,'(a)') "!CCSD total energy"
         ilbl = 11
      else if (imethod.eq.26) then  ! LADC
         write(eigstring,'(a)') "Final LT-DF-LADC(2)-Results"
         ilbl = 27
      else if (imethod.eq.34) then  ! LCC2
         write(eigstring,'(a)') "Final LT-DF-LCC2-LR-Results"
         ilbl = 27
      else 
         write(iqclog,'(/,a)') 'Routine Rdmolpro (readqc.F90)'
         write(iqclog,'(a,i5)') &
            'ERROR: QC Method not implemented for energy:',imethod
         write(iqclog,'(a)') ' '
      endif

!-----------------------------------------------------------------------
! some methods (CAS/PT2) require a second energy
!-----------------------------------------------------------------------
      if (imethod .eq. 13 .or. imethod .eq. 15 .or. imethod.eq.32) then
         write(eig1string,'(a)') "!MCSCF STATE"
         i1lbl = 12
      else
         write(eig1string,'(a)') "NO EIG1 STRING"
         i1lbl = 13
      endif

!-----------------------------------------------------------------------
! some methods (CIS, LADC, ...) require a gs energy
!-----------------------------------------------------------------------
      if (imethod .eq. 11) then
         write(egsstring,'(a)') "!RHF STATE 1.1 Energy"
         gslbl = 21
      else if (imethod .eq. 26) then        !LADC
         write(egsstring,'(a)') " MP2 total energy:"
         gslbl = 18
      else if (imethod .eq. 34) then   !LCC2
         write(egsstring,'(a)') "!LCC2 total energy"
         gslbl = 18
      else
         write(egsstring,'(a)') "NO EGS STRING"
         gslbl = 13
      endif

!-----------------------------------------------------------------------
! string for finding gradient depends on method
!-----------------------------------------------------------------------
      flbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(grastring,'(a)') "GRADIENT NULL STRING"
         flbl = 20
      else if (imethod.eq.1 .or. imethod .eq. 2) then            
         write(grastring,'(a)') "GRADIENT FOR STATE 1.1"
         flbl = 22
      else if (imethod.eq.12 .or. imethod .eq. 15 .or. &
               imethod .eq. 32) then            
         write(grastring,'(a)') "SA-MC GRADIENT FOR STATE"
         flbl = 24
      else if (imethod.eq.13) then            
         write(grastring,'(a)') "RSPT2 GRADIENT FOR STATE"
         flbl = 24
      else
         write(grastring,'(a)') "NO GRA STRING"
         flbl = 13
      endif

!-----------------------------------------------------------------------
! string for finding SOC matrix
!-----------------------------------------------------------------------
      if (imethod.eq.14 .or. imethod .eq. 15 .or. imethod .eq. 32) then
         socstring = "Spin-orbit calculation in the basis of zeroth order"
         soclbl = 51
      else
         socstring = "NO SOC STRING"
         soclbl = 13
      endif

!-----------------------------------------------------------------------
! string for finding normal modes
!-----------------------------------------------------------------------
      nmstring = "FREQUENCIES * CALCULATION OF NORMAL MODES"
      nmlbl = 41

!-----------------------------------------------------------------------
! string for finding MO matrix. Empty at start.
!-----------------------------------------------------------------------
      matstring = 'EMPTY STRING'

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         if ((index(string,'ATOMIC COORDINATES') .ne. 0) .or. &
            (index(string,'Atomic Coordinates') .ne. 0))  then 

!           if (index(string,'ATOMIC COORDINATES') .ne. 0) then
               fmttyp = '(6x,a7,f5.2,3(2x,f13.9))'
            if (index(string,'Atomic Coordinates') .ne. 0)then
               fmttyp = '(5x,a7,f5.2,3(2x,f13.9))'
            endif

            do n=1,3
               read(iunit,*,err=1001)
            enddo

            do i=1,qc_natm
               read(iunit,fmttyp,err=1001) &
                     atomname(i)(1:7),isym,(qcxcoo((i-1)*3+k),k=1,3)

               call keylowc(atomname(i))
               atomname(i) = adjustl(atomname(i))
               isoidx = 0
               do k = 1, niso
                  if(trim(atomname(i)).eq.trim(isonam(k))) then
                     isoidx = int(isomass(k))
                     exit
                  endif
               enddo

               do k=1,3
                  qciatsym((i-1)*3+k)=int(isym)
                  if (isoidx.ne.0) then
                     qciatiso((i-1)*3+k) = isoidx
                     qcfound(10) = .true.   ! isotope info found.
                  endif
               enddo
            enddo

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (index(string,nmstring(1:nmlbl)) .ne. 0) then

!-----------------------------------------------------------------------
! First determine no. imaginary freqs (compare no. positive frequencies
! to no. of modes)
!-----------------------------------------------------------------------
            do 
               read(iunit,'(a)',err=1002) string
               if (index(string,'Vibration') .ne. 0) exit
               backspace(iunit)
               backspace(iunit)
            enddo
            read(iunit,*)
            nimag=0     ! read no. of real vibrations
            do
              read(iunit,'(a)',err=1002) string
              if (string(1:10) .eq. '          ') exit
              nimag=nimag+1 
            enddo

            if (nimag .gt. qc_nmode) then
               write(iqclog,'(/,a,i5)') &
                   'ERROR: Too many positive frequencies Found. '
               write(iqclog,'(2(a,i5))') &
                   ' Expected : ',qc_nmode,'  Found : ',nimag
               stop
            endif

            nimag = qc_nmode - nimag 

!-----------------------------------------------------------------------
! read real frequencies
!-----------------------------------------------------------------------
            nrd = qc_nmode-nimag          ! No. positive freqs to read
            nblk = ((nrd-1)/5) + 1      ! No. data blocks

            do 
               read(iunit,'(a)',err=1002) string
               if (string(2:19) .eq. 'Wavenumbers [cm-1]') exit
            enddo

            backspace(iunit)
            backspace(iunit)
            backspace(iunit)

            do n=1,nblk
               nm = nrd - (n-1)*5    ! No. of freqs still to read
               nm = min(5,nm)

               read(iunit,*)

               koff = nimag + (n-1)*5 ! leave space at start for im freqs

               read(iunit,'(23x,5(9x,a3))') (qcnmlab(koff+j),j=1,nm)
               read(iunit,'(23x,5f12.2)',err=1002) (qcfreq(koff+j),j=1,nm)

               do
                  read(iunit,'(a)',err=1003) string
                  if (string(2:23) .eq. 'Intensities [relative]') exit
               enddo

               do k=1,qc_ncoo
                  read(iunit,'(23x,5f12.5)') (qcnmcoo(k,koff+j),j=1,nm)
               enddo
            enddo

!-----------------------------------------------------------------------
! read imaginary frequencies
!-----------------------------------------------------------------------
            if (nimag.gt.0) then

               do
                  read(iunit,'(a)',err=1002) string
                  if (string(2:38).eq. &
                         'Normal Modes of imaginary frequencies') exit
               enddo

               nrd = nimag                 ! No. imaginary freqs to read
               nblk = ((nrd-1)/5) + 1      ! No. data blocks

               do 
                  read(iunit,'(a)',err=1002) string
                  if (string(2:19) .eq. 'Wavenumbers [cm-1]') exit
               enddo

               backspace(iunit)
               backspace(iunit)
               backspace(iunit)

               do n=1,nblk
                  nm = nrd - (n-1)*5    ! No. of freqs stil to read
                  nm = min(5,nm)

                  read(iunit,*)
                  koff = (n-1)*5 

                  read(iunit,'(23x,5(9x,a3))') (qcnmlab(koff+j),j=1,nm)
                  read(iunit,'(23x,5f12.2)',err=1002) (qcfreq(koff+j),j=1,nm)
                  do j=1,nm
                     qcfreq((n-1)*5+j) = -qcfreq((n-1)*5+j)   ! negative freq
                  enddo

                  do
                     read(iunit,'(a)',err=1003) string
                     if (string(2:23) .eq. 'Intensities [relative]') exit
                  enddo

                  do k=1,qc_ncoo
                     read(iunit,'(23x,5f12.5)') (qcnmcoo(k,koff+j),j=1,nm)
                  enddo
               enddo

            endif
! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.

!-----------------------------------------------------------------------
! ENERGY  
!-----------------------------------------------------------------------
         else if (index(string,eigstring(1:ilbl)) .ne. 0) then 

            if (imethod .eq. 0) cycle   ! do not read energy
            if ((index(string,'Energy') .eq. 0) .and. &
                   (index(string,'energy') .eq. 0)) then
               cycle  ! This is not the energy
            endif

            if (imethod.eq.1) then
               neq=index(string,"Energy")+6
               read(string(neq:80),*,err=1003) qcv(1)
               lqcfound3(1)=.true.

            else if (imethod .eq. 2 .or. imethod .eq. 22) then
               neq=index(string,"energy")+6
               read(string(neq:80),*,err=1003) qcv(1)
               lqcfound3(1)=.true.

            else if (imethod .eq. 11) then
               il = index(string,"state")+5
               jl = index(string,".")
               kl = index(string,":")-1
               read(string(il:jl-1),*) istate
               read(string(jl+1:kl),*) issym

               il=index(string,"energy")+6
               jl=index(string,"au")-1
               read(string(il:jl),*,err=1003) energy
               qcv(istate)=energy+egrd   ! energy in Hartree
               lqcfound3(istate)=.true.

            else if ((imethod .ge. 12 .and. imethod .le. 15) .or. &
                     imethod .eq. 32) then
               il = index(string,"STATE")+5
               jl = index(string,".")
               kl = index(string,"Energy")-1
               read(string(il:jl-1),*) istate
               read(string(jl+1:kl),*) issym

! Get spin block offset
               if (imcespin .eq. 0) then
                  write(iqclog,'(a)') 'Multiplicity not found for MCSCF'
                  stop
               endif
               if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle

               istate = istate + qc_spboff(imcespin)

               neq = kl+7
               read(string(neq:80),*) qcv(istate)

               lqcfound3(istate)=.true.
               lfounde(istate)=.true.  ! to be checked if reading NACTs

            else if (imethod .eq. 26 .or. imethod .eq. 34) then
               il = index(string,"state")+5
               jl = index(string,".")
               kl = index(string,":")-1
               read(string(il:jl-1),*) istate
               read(string(jl+1:kl),*) issym
               if (istate .le. qc_nst) then
                  klbl=index(string,'eV')-1
                  jlbl=klbl-7
                  read(string(jlbl:klbl),*) energy
                  qcv(istate)=(energy/uev)+egrd   ! energy in Hartree
                  qcfound(3)=.true.
                  lqcfound3(istate)=.true.
               endif

            else
               write(iqclog,'(/,a,i5)') &
                   'ERROR: Unknown ab initio type for Molpro output :',imethod
               cycle
            endif

            qcfound(3)=.true.

!-----------------------------------------------------------------------
! SECOND ENERGY (required by e.g. CAS/PT2)
!-----------------------------------------------------------------------
         else if (index(string,eig1string(1:i1lbl)) .ne. 0) then

            if (imethod .eq. 0) cycle   ! do not read energy
            if (index(string,'Energy') .eq. 0) cycle  ! This is not the energy

            if (imethod .eq. 13 .or. imethod .eq. 15 .or. imethod .eq. 32) then   
                                               ! CAS energies for CAS/PT2 NACT
               il = index(string,"STATE")+5
               jl = index(string,".")
               kl = index(string,"Energy")-1
               read(string(il:jl-1),*) istate
               read(string(jl+1:kl),*) issym

! Get spin block offset
               if (imcespin .eq. 0) then
                  write(iqclog,'(a)') 'Multiplicity not found for MCSCF'
                  stop
               endif
               if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle

               istate = istate + qc_spboff(imcespin)

               neq = kl+7
               read(string(neq:80),*) qcv1(istate)
               lfounde(istate)=.true.
            endif

!-----------------------------------------------------------------------
! GS ENERGY (required by e.g. CIS, LADC, ...)
!-----------------------------------------------------------------------
         else if (index(string,egsstring(1:gslbl)) .ne. 0) then
            if (imethod.eq.11) then  ! read HF for CIS
               neq=index(string,"Energy")+6
               read(string(neq:80),*) egrd
            else if (imethod.eq.26) then  ! read MP2 for LADC
               read (string(25:50),'(f26.8)') egrd
            else if (imethod.eq.34) then  ! read LCC2 GS for LCC2
               read (string(29:54),'(f26.8)') egrd
            endif

            qcv(1) = egrd
            qcfound(3)=.true.
            lqcfound3(1)=.true.

!-----------------------------------------------------------------------
! FIRST DERIVATIVES
! Molpro provides gradient directly
!-----------------------------------------------------------------------
         else if (index(string,grastring(1:flbl)).ne.0) then

            if (imethod .eq. 0) cycle   ! do not read gradients

            if (imethod .eq. 1 .or. imethod .eq. 2) then
               istate=1
               issym=1
            else if (imethod .eq. 12 .or. imethod .eq. 13 .or. &
                     imethod .eq. 15 .or. imethod .eq. 32) then
               il = index(string,"STATE")+5
               jl = index(string,".")
               kl = jl+5
               read(string(il:jl-1),*) istate
               read(string(jl+1:kl),*) issym

! Get spin block offset
               if (imcespin .eq. 0) then
                  write(iqclog,'(a)') 'Multiplicity not found for MCSCF'
                  stop
               endif
               if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle

               istate = istate + qc_spboff(imcespin)
            endif

            do n=1,3
               read(iunit,*)
            enddo
            koff = 0
            do n=1,qc_natm
               read(iunit,*,err=1004) n1,(force(i),i=1,3)
               do i=1,3
                  qcderiv1(koff+i,istate)=force(i)   
               enddo
               koff=koff+3
            enddo

            qcfound(4)=.true.
            lqcfound4(istate)=.true.

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
! istate taken from GRADIENT reading
!-----------------------------------------------------------------------
         elseif (index(string, "Force Constants (Second Derivatives").ne.0) then

            if (imethod .eq. 0) cycle   ! do not read gradients

            if (.not. lqcfound4(istate)) then
               write(iqclog,'(a)') &
                   'Need to calculate Force before Force Constants'
               stop
            endif

            i1=1
   20       continue
            read(iunit,*)
            do n=i1,qc_ncoo
               n1 = min(5,n-i1+1)
               read(iunit,*,err=1005) ccoord,(force(i),i=1,n1)
               do i = 1,n1
                  n2 = i1+i-1
                  qcderiv2(n,n2,istate) = force(i)
                  if (n .ne. n2) qcderiv2(n2,n,istate) = force(i)
               enddo
            enddo
            if (i1+4 .lt. qc_ncoo) then
               i1 = i1+5
               go to 20
            endif

            qcfound(5)=.true.
            lqcfound5(istate)=.true.

!-----------------------------------------------------------------------
! MO COEFFICIENTS
! Up to molpro 2021 can read from orbital print. After 2021 must
! read MATROP output. 
!-----------------------------------------------------------------------
         else if((index(string,'NATURAL ORBITALS') .ne. 0) .or. &
                (index(string,'PSEUDO CANONICAL ORBITALS') .ne. 0))then
                                      ! From orbprint so not transpose.

            if (qc_nbasis .le. 0) cycle    !  Do not read MOs

! Check if this is 2015 or earlier format - otherwise cannot read
            read(iunit,'(a)',err=1006) string
            read(iunit,'(a)',err=1006) string
            read(iunit,'(a)',err=1006) string
            if (string(1:28) .ne. "   Orb     Occ        Energy") cycle

! check nbasis set correctly
            nbaschk=0
            do i = 1,qc_nirrep
               nbaschk=nbaschk+qc_nsymorb(i)
            enddo
            if (nbaschk .ne. qc_nbasis) then
               write(iqclog,'(a,i5)') 'Nbasis incorrect. Should be: ',nbaschk
               cycle
            endif

            noffs=0
            do i = 1,qc_nirrep
               write(ststring,'(a,i1)')'1.',i
               do
                  read(iunit,'(a)',err=1006)mostring

                  if(index(mostring,trim(ststring)).ne.0)then
                     backspace(iunit)
                     backspace(iunit)
                     exit
                  endif
               enddo
               do n = 1, qc_nsymorb(i) ! loop over MOs
                  read(iunit,*,iostat=ierr)
                  ioffs = noffs
                  do n1 = 1, qc_nsymorb(i)/10
                     read(iunit,'(31x,10(f10.6))',iostat=ierr) &
                          (qcmocoeff(n+noffs,i1+ioffs),i1=1,10)
                     ioffs = ioffs + 10
                  enddo
                  if(mod(qc_nsymorb(i),10).ne.0) &
                       read(iunit,'(31x,10(f10.6))',iostat=ierr) &
                       (qcmocoeff(n+noffs,i1+ioffs),i1=1,mod(qc_nsymorb(i),10))
               enddo
               noffs = noffs+qc_nsymorb(i)
            enddo
            if (ierr .ne. 0) cycle ! error reading MO data

            qcfound(6)=.true.

         else if (index(string,'Orbitals') .ne. 0) then 
                        ! Set up string to be read for MOs from Matrop
                        ! "Orbitals NAME read from"   Want NAME.

            jlbl=index(string,'read') 
            if (jlbl .le. 0) cycle   ! Not reading orbitals

            ilbl=index(string,'Orbitals')+9
            jlbl=jlbl-1

            read(string(ilbl:jlbl),'(a)',err=1006) matstring
            jlbl = index(matstring,' ')-1
            matstring = 'MATRIX '//matstring(1:jlbl)
            
         else if (index(string,matstring) .ne. 0) then 
                                            ! From matrop 

            if (qc_nbasis .le. 0) cycle    !  Do not read MOs

! Reads transpose in scientific notation (PRIC)
            noffs=0
            do i = 1,qc_nirrep
               write(ststring,'(a,i1,a)')'BLOCK ',i,'.1'
               do
                  read(iunit,'(a)',err=1006) mostring
                  if(index(mostring,ststring) .ne. 0) exit
               enddo
               do n = 1, qc_nsymorb(i) ! loop over MOs
                  read(iunit,*)
                  read(iunit,*,err=1006) &
                             (qcmocoeff(n+noffs,i1),i1=1,qc_nbasis)
               enddo
               noffs = noffs+qc_nsymorb(i)
            enddo

            qcfound(6) = .true.

!-----------------------------------------------------------------------
! DERIVATIVE COUPLINGS
! In Molpro NACTS include 1/DeltaE
!-----------------------------------------------------------------------
         else if (string(2:23) .eq. "SA-MC NACME FOR STATES") then

            if (imethod .eq. 0) cycle   ! do not read DCP

            read(string(25:25),'(i1)') istate
            read(string(31:31),'(i1)') jstate

            if(istate.le.0 .or. jstate.le.0) cycle

            if (.not. lfounde(istate) .and. .not. lfounde(jstate)) then
               write(iqclog,'(a)')'ERROR reading Molpro NACT output'
               write(iqclog,'(a)')'Have not read energies '
               stop
            endif

! Get spin block offset
            if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle
            if (jstate.le.0 .or. jstate .gt. qc_nmult(imcespin)) cycle
            istate = istate + qc_spboff(imcespin)
            jstate = jstate + qc_spboff(imcespin)

! NACTS are at CAS level so need qcv1 rather than qcv for DeltaE
            vdiff=-999.0
            if (imethod .eq. 13 .or. imethod .eq. 15 .or. imethod .eq. 32) then
               vdiff = qcv1(jstate) - qcv1(istate)
            else if (imethod .eq. 12) then
               vdiff = qcv(jstate) - qcv(istate)
            endif

            read(iunit,*)
            read(iunit,*)
            read(iunit,*)
            ioffs = 0
            do n=1,qc_natm
               read(iunit,*,err=1007) idum1, &
                       (qcnact(i+ioffs,istate,jstate),i=1,3)
               ioffs = ioffs + 3
            enddo
            qcnact(1:qc_ncoo,istate,jstate) = &
                      qcnact(1:qc_ncoo,istate,jstate)*vdiff

            qcnact(:,jstate,istate) = qcnact(:,istate,jstate)

            lqcfound7(istate,jstate)=.true.
            lqcfound7(jstate,istate)=.true.
            qcfound(7)=.true.

!-----------------------------------------------------------------------
! DIPOLE MOMENT
!-----------------------------------------------------------------------
         else if (index(string,"!RHF STATE 1.1 Dipole moment") .ne. 0) then

            if (imethod .ne. 1) cycle

            read(iunit,'(a)',err=1008) string
            read(string,'(35x,3(3x,f12.8))',err=1008) (qcdipole(n,1),n=1,3)

            qcfound(8)=.true.
            lqcfound8(1)=.true.

!-----------------------------------------------------------------------
! ISOTOPES
!-----------------------------------------------------------------------
         else if (index(string,"SETTING MASS") .ne. 0) then

            isonam = ' '
            isomass = 0.0_dop
            stringtmp= ' '
            il=index(string,'FOR')+3
            jl=index(string,'(')-1
            kl=index(string,'TO')+2
          
            niso = niso + 1

            read(string(il:jl),'(a)',err=1010) isonam(niso)
            call keylowc(isonam(niso))
            isonam(niso)=adjustl(isonam(niso))
            read(string(kl:kl+10),*,err=1010) isomass(niso)

            qcfound(10)=.true.

!-----------------------------------------------------------------------
! SPIN ORBIT COUPLINGS
!-----------------------------------------------------------------------
         else if (index(string,socstring(1:soclbl)) .ne. 0) then

            if (imethod .eq. 0) cycle

! Read down to SO Matrix
            do
               read(iunit,'(a)',err=1011) string
               if (index(string,'Spin-Orbit Matrix') .ne. 0) exit
            enddo

! Get total no. of SOC states and allocate arrays to read data. Must 
! account for degeneracies.
            nsocs = qc_nstdeg
            allocate(tmpdsoc(nsocs,nsocs,2))
            allocate(ispmatlab(nsocs,3))   ! Labels: state,spin,SZ
            tmpdsoc=0.0_dop
            ispmatlab=0

            nblk = ((nsocs-1)/10) + 1      ! No. data blocks

            joffs = 0
            do n=1,nblk
               nm = nsocs - (n-1)*10   ! No. of SOCs still to read
               nm = min(10,nm)

               read(iunit,'(a)',err=1011) string
               read(iunit,'(a)',err=1011) string
               read(iunit,'(a)',err=1011) string
               do i = 1,nsocs
                  read(iunit,'(a)') string
                  read(iunit,*) idum1,rdum,rdum1,rdum2,&
                                (tmpdsoc(i,j,1),j=joffs+1,joffs+nm)
                  read(iunit,*)(tmpdsoc(i,j,2),j=joffs+1,joffs+nm)
               enddo
               joffs = joffs + nm
            enddo
            do i = 1,nsocs-1
               do j = 2,nsocs
                  qcsoc(i,j) = tmpdsoc(i,j,1) + (0,1)*tmpdsoc(i,j,2)
                  qcsoc(j,i) = dconjg(qcsoc(i,j))
               enddo
            enddo
            
! Convert SOCs from cm-1 to au.
            qcsoc = qcsoc/invcm

            deallocate(tmpdsoc,ispmatlab)

            qcfound(11) = .true.
            lqcfound11  = .true.

!-----------------------------------------------------------------------
! Get multiplicity of next energy set
!-----------------------------------------------------------------------
! Single block of results has multiplicity at start
         else if ((index(string,'PROGRAM * MULTI') .ne. 0) .or. &
                  (index(string,'PROGRAM * RS2C') .ne. 0) .or. &
                  (index(string,'PROGRAM * CI') .ne. 0)) then

            if (qc_nspblk .eq. 1) then   ! Spins not read if only single
               imcespin = 1              ! multiplicity in calculation
               cycle
            endif

            imcespin = 0
            do
               read(iunit,'(a)',err=1000) string
               if (index(string,'Singlet').ne.0)then
                  imcespin = 1
               else if (index(string,'Doublet').ne.0)then
                  imcespin = 2
               else if (index(string,'Triplet').ne.0)then
                  imcespin = 3
               else if (index(string,'Quartet').ne.0)then
                  imcespin = 4
               else if (index(string,'Quintet').ne.0)then
                  imcespin = 5
               else if (index(string,'Sextet').ne.0)then
                  imcespin = 6
               else if (index(string,'Septet').ne.0)then
                  imcespin = 7
               endif

               if (index(string,'Molecular orbitals read').ne.0) exit
               if (index(string,'Spin-orbit calculation').ne.0) exit
            enddo

! MCSCF with multiple multiplicities needs definition for each energy
         else if (index(string,'Results for state') .ne. 0) then

            if (qc_nspblk .eq. 1) then   ! Spins not read if only single
               imcespin = 1              ! multiplicity in calculation
               cycle
            endif

            if (index(string,'Singlet').ne.0)then
               imcespin = 1
            else if (index(string,'Doublet').ne.0)then
               imcespin = 2
            else if (index(string,'Triplet').ne.0)then
               imcespin = 3
            else if (index(string,'Quartet').ne.0)then
               imcespin = 4
            else if (index(string,'Quintet').ne.0)then
               imcespin = 5
            else if (index(string,'Sextet').ne.0)then
               imcespin = 6
            else if (index(string,'Septet').ne.0)then
               imcespin = 7
            endif

!-----------------------------------------------------------------------
! Get PT2 <-> CAS mapping for NACTs
! i=ipt2map(j)   CASPT2 state i = CASSCF state j
!-----------------------------------------------------------------------
         else if(index(string,'Performing an extended multi-state').ne.0)then
            if (imethod .eq. 0) cycle

            if (imethod .ne. 32) cycle

            read(iunit,*,err=1100)
            read(iunit,'(a)',err=1100) string
            if (string(2:5) .eq. 'Feff') then
               read(iunit,*,err=1100)
            else
               backspace(iunit)
            endif           
            do i=1,qc_nst
               read(iunit,'(a)',err=1100) string
               il = index(string,'Eigenvector')+12
               read(string(il:120),*,err=1011) &
                      (tmpvec(j),j=1,qc_nst)  ! Read zero-order eigenvector

               ipt2map(i) = 1
               do j=2,qc_nst    ! get largest coeff of eigenvector
                  k = ipt2map(i)
                  if (abs(tmpvec(j)) .gt. abs(tmpvec(k))) ipt2map(i) = j
               enddo
            enddo

            lfoundmap = .true.

!-----------------------------------------------------------------------
! Read the CI Vector
!-----------------------------------------------------------------------
         else if(index(string,'CI vector').ne.0)then
            if (imethod .eq. 0) cycle

            if (.not. qcread(12)) cycle

            nst = qc_nmult(imcespin)
            soff = qc_spboff(imcespin)

            read(iunit,*)
            read(iunit,*)
            i=0
            do
               read(iunit,'(a)',err=1012) string
               if (string .eq. ' ') exit
               i=i+1
               read(string,*) qcciconfig(i,imcespin),&
                       (qccicoeff(i,j),j=soff+1,soff+nst)
            enddo
            qcnciconf(imcespin) = i
            if (i .gt. 0) then
               qcfound(12) = .true.
               do j=soff+1,soff+nst
                  lqcfound12(j,imcespin) = .true.
               enddo
            endif

! Must ensure the sign of the CI vector follows the Molpro convention of 
! the largest element for each state being >0. Molpro sometimes prints the 
! CI vectors before fixing the sign.
            k=0
            maxci = 0.0_dop
            nci = qcnciconf(imcespin)
            do j = soff+1,soff+nst
               do i = 1,nci
                  if (abs(qccicoeff(i,j)).gt.maxci) then
                     k=i
                     maxci = abs(qccicoeff(i,j))
                   endif
               enddo

! Fix the signs if needed.
               if(qccicoeff(k,j).lt.0.0_dop)then
                  do i = 1,nci
                     qccicoeff(i,j) = -qccicoeff(i,j)
                  enddo
               endif
            enddo

! Orthonormalise the vectors (due to precision of output).
         do i = 1,nst
            do j = 1,nst
               sci(j,i) = dot_product(qccicoeff(1:nci,soff+j),&
                    qccicoeff(1:nci,soff+i))
            enddo
         enddo

         do i = 1,nst
            sci(i,i) = 1.5_dop-0.5_dop*sci(i,i)
            do j = i+1,nst
               sci(j,i) = -0.5_dop*sci(j,i)
            enddo
            do j = 1,i-1
               sci(j,i) = -0.5_dop*sci(j,i)
            enddo
         enddo

         tconf = 0.0_dop
         do i = 1,nst
            do j = 1,nci
               do k = 1,nst
                  tconf(j,i) = tconf(j,i) + sci(i,k)*qccicoeff(j,k)
               enddo
            enddo
         enddo
         do i = 1,nst
            qccicoeff(1:nci,soff+i) = tconf(1:nci,i)
         enddo

!-----------------------------------------------------------------------
! Flag any error messages from QC
!-----------------------------------------------------------------------
         else if(index(string,'MAXIMUM NUMBER OF ITERATIONS REACHED').ne.0)then
            write(iqclog,'(a)') 'CAS not converged '
            lfail = .true.
         else if(index(string,'NO CONVERGENCE OF CP-MCSCF').ne.0)then
            write(iqclog,'(a)') 'CPMCSCF not converged '
            lfail = .true.

         else if(index(string,'NO CONVERGENCE').ne.0)then
            write(iqclog,'(a)') 'CI calculation not converged '
            lfail = .true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

!-----------------------------------------------------------------------
! Map CAS vectors to correct state in XMS-PT2 calculations
!-----------------------------------------------------------------------
      if (lfoundmap) then
         if (qcfound(4)) then
            tmpderiv1(:,:) = qcderiv1(:,:)
            qcderiv1 = 0.0_dop

            do s=1,qc_nst
               i=ipt2map(s)
               qcderiv1(:,i) = tmpderiv1(:,s)
            enddo
         endif
         if (qcfound(5)) then
            tmpderiv2(:,:,:) = qcderiv2(:,:,:)
            qcderiv2 = 0.0_dop

            do s=1,qc_nst
               i=ipt2map(s)
               qcderiv2(:,:,i) = tmpderiv2(:,:,s)
            enddo
         endif
         if (qcfound(7)) then
            tmpnact(:,:,:) = qcnact(:,:,:)
            qcnact = 0.0_dop

            do s=1,qc_nst
               do s1=s+1,qc_nst
                  i=ipt2map(s)
                  j=ipt2map(s1)
                  qcnact(:,j,i) = tmpnact(:,s1,s)
                  qcnact(:,i,j) = qcnact(:,j,i)
               enddo
            enddo
         endif

      endif

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (Energy)'
      stop
 1004 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (1st derivatives)'
      stop
 1005 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (2nd derivatives)'
      stop
 1006 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (MO coeffs.)'
      stop
 1007 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (NACTs)'
      stop
 1008 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (Dipoles)'
      stop
 1010 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (Isotopes)'
      stop
 1011 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (SOC matrix)'
      stop
 1012 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (CI Vector)'
      stop
 1100 write(iqclog,'(/,a)') 'ERROR reading MOLPRO output (CAS/PT2 Mapping)'
      stop
 1101 write(iqclog,'(/a,i5)') 'ERROR reading MOLPRO output &
          &number of states in multiplicity: ',imcespin
      stop
      end subroutine rdmolpro

!#######################################################################
!
! RDDFTMRCI
! Read a DFTMRCI Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rddftmrci(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,j,n,s
      integer(long)               :: nroot,isym,isym1

      real(dop)                   :: mag
      real(dop), dimension(3)     :: coo,vec
      character(len=80)           :: string
      character(len=20)           :: atmp

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.false.
         else if (imethod .eq. 18) then
            qcfound(3)=.true.
            qcfound(9)=.true.
         endif

         return
      endif

      if (imethod .ne. 18) return   ! only for DFT/MRCI

      nroot=0    !   state of interest for energies

!-----------------------------------------------------------------------
! COORDINATES (in Bohr)
! Geometery is at top of file?
!-----------------------------------------------------------------------
      n=0
      read(iunit,*,err=1001)
      do i=1,qc_ncoo/3
         read(iunit,10,err=1001) coo(1),coo(2),coo(3),atmp
         call getinam(atmp,isym,isym1,iqclog)  ! convert atom name to atomic symbol
         do j=1,3
            n=n+1
            qciatsym(n)=isym
            qciatiso(n)=isym1
            qcxcoo(n)=coo(j)
         enddo
      enddo

      qcfound(1)=.true.

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         if (string(3:7).ne.'state') then

            nroot = nroot+1

            read(iunit,*,err=1003)
            read(iunit,'(a)',err=1003) string
            read(string(33:42),'(F11.6)',err=1003) qcv(nroot)

            qcfound(3)=.true.
            lqcfound3(nroot) = .true.

!-----------------------------------------------------------------------
! TRANSITION DIPOLE
!-----------------------------------------------------------------------
         else if (string(22:52).eq. 'T - D E N S I T Y - O U T P U T') then
 
            do

               if (string(22:48).eq.'D E N S I T Y - O U T P U T') then
                  exit

               else if (string(1:10).eq.'transition') then
                  read(string(18:18),*) s    ! state for transition

                  vec(:)=0.0_dop
                  read(iunit,*,err=1009)
                  read(iunit,*,err=1009)
! Velocity and Length gauge results are separated by one line
                  if (lqctdveloc) read(iunit,*,err=1009)

                  read(iunit,'(a)',end=1009) string
                  read(string(19:28),*) vec(1)
                  read(string(31:40),*) vec(2)
                  read(string(43:52),*) vec(3)

                  mag=0.0_dop
                  do i=1,3
                     mag=mag + vec(i)*vec(i)
                  enddo
                  if (mag .ge. 1.0e-8_dop) then
! tdip(4) is magnitude of dipole
                     mag=dsqrt(mag)
                     qctdipole(4,1,s)=mag

! tdip(i), i=1-3 is unit vector along dipole
                     do i=1,3
                        vec(i) = vec(i) / mag
                     enddo
! as convention take vector direction with positive dot product with (1,1,1)
                     mag=0.0_dop
                     do i=1,3
                        mag = mag + vec(i)
                     enddo
                     if (mag .lt. 0.0_dop) then
                        do i=1,3
                           vec(i)=-vec(i)
                        enddo
                     endif
                     do i=1,3
                        qctdipole(i,1,s) = vec(i)
                     enddo

                  endif

               endif
            enddo
            qctdipole(:,s,1) = qctdipole(:,1,s) 

            qcfound(9)=.true.
            lqcfound9(1,s)=.true.
            lqcfound9(s,1)=.true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

   10 format(2x,F18.15,4x,F18.15,4x,F18.15,6x,a5)

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading DFTMRCI output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading DFTMRCI output (Coordinates)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading DFTMRCI output (Energy)'
      stop
 1009 write(iqclog,'(/,a)') 'ERROR reading DFTMRCI output (Transition Dipole)'
      stop
      end subroutine rddftmrci

!#######################################################################
!
! RDTURBOMOLE
! Read a TURBOMOLE Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!
! THIS ROUTINE IS UNTESTED
!#######################################################################

      subroutine rdturbomole(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,j,k,n,isym,isym1,nrd
      integer(long)               :: nroot,nm,nimag,c
      integer(long)               :: ilbl,idx,jdx

      real(dop)                   :: rswap
      real(dop), dimension(3)     :: coo
      character(len=80)           :: string
      character(len=16)           :: atmp
      character(len=40)           :: eigstring
      character(len=16), dimension(:), allocatable :: atmpv
      character(len=16)                            :: aswap


!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
         else if (imethod .eq. 19 .or. imethod .eq. 23) then
            qcfound(3)=.true.
         endif

         return
      endif

      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. 

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      ilbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY NULL STRING"
         ilbl = 18
      else if (imethod .eq. 19) then
         write(eigstring,'(a)') "|  total energy"
         ilbl = 15
      else if (imethod .eq. 23) then
         write(eigstring,'(a)') "Final CC2 energy"
         ilbl = 16
      endif

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         if (index(string,'atomic coordinates') .ne. 0) then

            n=0
            do i=1,qc_natm
               read(iunit,20,err=1001) coo(1),coo(2),coo(3),atmp
               call getinam(atmp,isym,isym1,iqclog)  
                                           ! convert atom name to atomic symbol
               do j=1,3
                  n=n+1
                  qciatsym(n)=isym
                  qcxcoo(n)=coo(j)
               enddo
            enddo

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (index(string,'NORMAL MODES and VIBRATIONAL FREQUENCIES') &
              .ne. 0) then

            nrd = ((qc_ncoo-1)/6) + 1
            allocate(atmpv(qc_ncoo+6-mod(qc_ncoo,6)))

            do
               read(iunit,'(a)',err=1002) string
               if (index(string,'mode') .ne. 0) exit
            enddo

            read(iunit,*,err=1002)

! First determine the number of imaginary frequencies present
! Run through all normal modes, reading in the frequencies. Then, check
! which ones start with an "i". Increment nimag for each one found.
! Turbomole lists these at the start of the list.
            nm=0
            nimag=0
            c=qc_ncoo+16
            do n=1,nrd
               read(iunit,'(20x,6a9)',err=1002) (atmpv(j),j=nm+1,nm+6)
               do i=1,6
                  do j=1,9
                     if (atmpv(nm+i)(j:j).eq.'i') nimag=nimag+1
                  enddo
               enddo
               do i=1,c
                  read(iunit,*,err=1002)
               enddo
               nm=nm+6
            enddo
!    Once imaginary modes have been found, step back to the start of the
!    list of normal modes.
            c=qc_ncoo+17
            do n=1,nrd
               do i=1,c
                  backspace(iunit)
               enddo
            enddo

! Loop over each block of (up to) 6 modes.
            nm = 0
            do n=1,nrd

! Get the actual frequencies and save the numbers in array "freq".
               read(iunit,'(20x,6a9)',err=1002) (atmpv(j),j=nm+1,nm+6)
               do i=1,6
                  if ((nm+i).le.nimag) then
                     read(atmpv(nm+i)(3:9),'(F7.2)',err=900) qcfreq(nm+i)
                  else
                     read(atmpv(nm+i)(1:9),'(F9.2)',err=900) qcfreq(nm+i)
                  endif
               enddo

! Get irreps of the modes and save in "nmlab".
               read(iunit,*,err=1002)
               read(iunit,'(18x,6(7x,a2))',err=1002) (qcnmlab(j),j=nm+1,nm+6)

! Jump past the IR info.
               do i=1,8
                  read(iunit,*,err=1002)
               enddo

! Read in Cartesian coordinates of the modes to array "matrix". The columns
! represent the modes.
               do i=1,qc_ncoo
                  read(iunit,'(21x,6(F8.5,1x))',err=1002) &
                                  (qcnmcoo(i,j), j=nm+1,nm+6)
               enddo

               nm=nm+6
               do i=1,6
                  read(iunit,*,err=1002)
               enddo

            enddo

! Sort nmlab, matrix, and freq arrays by frequencies.
            do i=1,qc_ncoo
               if (qcfreq(i).eq.0.0_dop) qcfreq(i)=999999.9
            enddo

            do i=1,qc_ncoo-1
               do j=i+1,qc_ncoo

                  if (qcfreq(i).gt.qcfreq(j)) then

                     rswap=qcfreq(i)
                     qcfreq(i)=qcfreq(j)
                     qcfreq(j)=rswap

                     aswap=qcnmlab(i)
                     qcnmlab(i)=qcnmlab(j)
                     qcnmlab(j)=aswap

                     do k=1,qc_ncoo
                        rswap=qcnmcoo(k,i)
                        qcnmcoo(k,i)=qcnmcoo(k,j)
                        qcnmcoo(k,j)=rswap
                     enddo

                  endif
               enddo
            enddo

! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         else if(index(string,eigstring(1:ilbl)).ne.0) then
            if (imethod .eq. 0) cycle   

            if (imethod .eq. 19) then
               read(string(42:58),'(f17.11)',err=1003) qcv(nroot)
               lqcfound3(nroot) = .true.

            else if (imethod .eq. 23) then
               idx=index(string,':')
               read(string(idx+1:idx+18),'(f18.10)',err=1003) qcv(1)
               lqcfound3(1) = .true.

! Look for excited state energy...
               do
                  read(iunit,'(a)',err=1003) string
                  idx = index(string,'CC2 excitation energies')
                  jdx = index(string,'||T1||')
                  if(idx.ne.0 .or. jdx .ne. 0) exit
               enddo

! Read down to the line of interest, accounting for the extra lines of
! information before the energies.
               read(unit=iunit,fmt='(a)',err=1003) string
               read(unit=iunit,fmt='(a)',err=1003) string
               do nroot = 2, qc_nst   ! loop over excited states
                  read(iunit,'(a)',err=1003) string

                  do j = 1, 4
                     idx = index(string,'|')
                     string = string(idx+1:160)
                  enddo

                  read(string(1:13),'(f13.7)') qcv(nroot)

! Turbomole prints out an excitation energy, so add this to the GS energy
! to get the total energy of the state in Hartree.
                  qcv(nroot) = qcv(nroot) + qcv(1)
                  lqcfound3(nroot) = .true.
               enddo

            endif

            qcfound(3)=.true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

   20 format(1x,3(3x,F11.8),4x,a2)

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading TURBOMOLE output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading TURBOMOLE output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading TURBOMOLE output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading TURBOMOLE output (Energy)'
      stop
      end subroutine rdturbomole
!#######################################################################
!
! RDGAMESS_US
! Read a GAMESS-US Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdgamess_us(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,j,n,isym,isym1,ierr
      integer(long)               :: nroot,mxsta
      integer(long)               :: ilbl,gslbl
      integer(long), dimension(10) :: dgen

      real(dop)                   :: etmp,egrd,small
      real(dop), dimension(3)     :: coo
      real(dop), dimension(10)    :: entmp
      character(len=80)           :: string
      character(len=40)           :: eigstring
      character(len=40)           :: egsstring
      character(len=20)           :: atmp

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
         else if (imethod .eq. 12 .or. imethod .eq. 13 .or. &
                  imethod .eq. 20 .or. imethod .eq. 21) then
            qcfound(3)=.true.
         endif

         return
      endif

      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. 

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      ilbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY NULL STRING"
         ilbl = 18
      else if (imethod.eq.12) then                        ! CAS
         write(eigstring,'(a)') "-MCCI- BASED ON OPTIMIZED ORBITALS"
         ilbl = 34
      else if (imethod.eq.13) then                        ! CASPT2
         write(eigstring,'(a)') "MCQDPT2 ENERGIES"
         ilbl = 16
      else if (imethod.eq.20) then                        ! CR-EOM-CCSD
         write(eigstring,'(a)') "CR-EOMCCSD(T)IIC"
         ilbl = 16
      else if (imethod.eq.21) then                        ! DELTA-CR-EOM-CCSD
         write(eigstring,'(a)') "DELTA-CR-EOMCCSD(T)"
         ilbl = 19
      endif

!-----------------------------------------------------------------------
! some methods have a different string for the GS energy
!-----------------------------------------------------------------------
      if (imethod.eq.20 .or. imethod .eq. 21) then
         write(egsstring,'(a)') "CR-CCSD(T)IIC"
         gslbl = 13
      else
         write(egsstring,'(a)') "NO EGS STRING"
         gslbl = 13
      endif
      
      nroot = 0 ! counter for CR-EOM-CCSD
      small=0.0011

      entmp(:)=0.0_dop
      dgen(:)=0

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         if (index(string,'COORDINATES') .ne. 0) then

            read(iunit,'(a)',err=1001) string

            n = 0
            do i=1,qc_ncoo/3
               read(iunit,20,err=1001) atmp,coo(1),coo(2),coo(3)
               call getinam(atmp,isym,isym1,iqclog)  ! convert atom name to atomic symbol
               do j=1,3
                  n=n+1
                  qciatsym(n)=isym
                  qcxcoo(n)=coo(j)
               enddo
            enddo

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         else if(index(string,eigstring(1:ilbl)) .ne. 0) then
            if (imethod .eq. 0) cycle   

            n = 0
            if (imethod .eq. 12) then
               do
                  read(iunit,'(a)',err=1003) string
                  if (string(2:6).eq.'STATE') then
                     read(string(12:13),'(i2)') nroot
                     read(string(28:41),'(F14.9)') qcv(nroot)
                     qcfound(3)=.true.
                     lqcfound3(nroot) = .true.
                  else if (string(2:8).eq.'DENSITY') then
                     exit
                  endif
               enddo

            else if (imethod .eq. 13) then
               read(iunit,'(a)',err=1003) string
               do i=1,qc_nst
                  read(iunit,'(a)',err=1003) string
                  read(string(28:41),'(F14.9)',iostat=ierr) etmp
                  if (ierr .eq. 0) then
                     qcv(i) = etmp
                     qcfound(3)=.true.
                     lqcfound3(i) = .true.
                  endif
               enddo

            else if (imethod .eq. 20) then
               n=n+1
               read(iunit,'(a)',err=1003) string
               read(string,'(36x,F15.10)',iostat=ierr) etmp
               if (ierr .eq. 0) then
                  qcv(n+1) = etmp
                  qcfound(3)=.true.
                  lqcfound3(n+1) = .true.
               endif

            else if (imethod .eq. 21) then
               read(iunit,*,err=1003)
               read(iunit,*,err=1003)
               read(iunit,*,err=1003)
               do i=2,mxsta
                  if (dgen(i).ne.0) cycle

                  read(iunit,*,err=1003)
                  read(iunit,'(63x,F6.3)',err=1003) entmp(i)
                  entmp(i)=(entmp(i)/uev) + egrd
               enddo
               entmp(1) = egrd

               do i=2,mxsta
                  if (dgen(i).ne.0) entmp(i)=entmp(dgen(i))
               enddo

               call bubble(entmp,mxsta)

               do i=1,mxsta
                  qcv(i) = entmp(i)
                  qcfound(3)=.true.
                  lqcfound3(i) = .true.
               enddo

            endif

!-----------------------------------------------------------------------
! GS ENERGY (required by e.g. EOMCCSD)
!-----------------------------------------------------------------------
         else if (index(string,egsstring(1:gslbl)) .ne. 0) then

            if (imethod .eq. 0) cycle   ! do not read energy

            read(string,'(63x,F15.10)',err=1003) egrd

!-----------------------------------------------------------------------
! If two or more states are degenerate at the EOM-CCSD level GAMESS-US
! will only print the CR-EOM-CCSD(T) energy of one of these states.
! So, first determine whether any states are degenerate.
!-----------------------------------------------------------------------
         else if (string(21:39).eq.'SUMMARY OF EOM-CCSD') then

            if (imethod .ne. 21) cycle   ! do not read energy
 
            read(iunit,*,err=1003)
            read(iunit,*,err=1003)
            read(iunit,*,err=1003)

            mxsta=0
            do
               read(iunit,'(a)',err=1003) string
               if (string(1:13).ne.'            ') exit

               mxsta=mxsta+1
               read(string(35:41),'(F7.3)') entmp(mxsta)
            enddo

            do i=2,mxsta
               if (entmp(i).lt.(entmp(i-1)+small) .and. &
                    entmp(i).gt.(entmp(i-1)-small)) dgen(i)=i-1
            enddo

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

   20 format(x,a2,9x,3(6x,F14.10))

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading GAMESS-US output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading GAMESS-US output (Coordinates)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading GAMESS-US output (Energy)'
      stop
      end subroutine rdgamess_us

!#######################################################################
!
! RDQCHEM
! Read a QChem Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
! lfail   : Set to true if QC has failed
!
! Other variables and data arrays are defined above in module readqc 
!#######################################################################

      subroutine rdqchem(iunit,imethod,iqclog,lfail)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,i1,n,n1,n2,nm,sdx,&
                                     j,k,koff,nrd,nblk,nimag,&
                                     il,jl,kl,ioffs,joffs,noffs,s,s1
      integer(long)               :: ilbl,jlbl,klbl
      integer(long)               :: eiglbl,flbl,gslbl,fgslbl,&
                                     nmlbl,nmlbl1,naclbl,heslbl,&
                                     hesgslbl,idum,isym1,isym,&
                                     eiglbl1,eiglbl2
      integer(long)               :: bid,eid,neq,ntr
      integer(long)               :: istate,jstate,issym,&
                                     stread,stread1
      integer(long)               :: imcespin,nsocs
      integer(long)               :: idum1,niso,isoidx,nbaschk,ierr,iau
      integer(long), dimension(:,:), allocatable  :: ispmatlab
      integer(long), dimension(qc_ncoo) :: idxfreq

      real(dop), dimension(qc_ncoo)  :: mass
      real(dop)                   :: rdum,rdum1,rdum2,norm
      real(dop)                   :: vdiff,vtmp,s2,egrd,ediff
      real(dop), dimension(3)     :: coo
      real(dop), dimension(5)     :: force
      real(dop), dimension(qc_nst) :: qcv1,tmpvec
      real(dop), dimension(3,qc_natm)        :: tmpderiv1
      real(dop), dimension(qc_ncoo,qc_ncoo)  :: tmpderiv2
      real(dop), dimension(qc_ncoo)          :: tmpfreq,tmpfreq1
      real(dop), dimension(qc_ncoo,qc_nst+1,qc_nst+1) :: tmpnact
      real(dop), dimension(:,:,:), allocatable        :: tmpdsoc
      character(len=80)           :: atmp
      character(len=120)          :: string
      character(len=80)           :: mostring
      character(len=40)           :: eigstring, egsstring
      character(len=40)           :: eigstring1, eigstring2
      character(len=60)           :: grastring,gragsstring
      character(len=60)           :: hesstring,hesgsstring
      character(len=50)           :: nmstring, nmstring1
      character(len=10)           :: ststring
      character(len=60)           :: nacstring, matstring
      character(len=5)            :: ccoord
      character(len=80)           :: stringtmp,forma
      character(len=7), dimension(qc_natm) :: isonam
      logical(kind=4), dimension(qc_nst)   :: lfounde
      logical(kind=4), dimension(qc_ncoo)  :: lignore
      logical(kind=4)              :: liso,lorbtr,lorbtrchk,found,lnmeigvec
      logical(kind=4), intent(out) :: lfail
      logical(kind=4)              :: lforce,lfreq
      logical(kind=4)              :: lback = .true.

      interface
      function dnrm2(n,v,i)
         use decimal, only:dop
         real(dop) :: dnrm2
         integer,intent(in) :: n,i
         real(dop),dimension(n),intent(in) :: v
      end function dnrm2
      end interface

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
! TO TEST: 33 TDDFT 
!          Hessians
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
         endif
         if (imethod .eq. 1 .or. imethod .eq. 11 .or. &
             imethod .eq. 17 .or. imethod .eq. 19 .or. &
             imethod .eq. 22 .or. imethod .eq. 33) then
            qcfound(3)=.true.
            qcfound(4)=.true.
            qcfound(5)=.true.
         endif
         if (imethod .eq. 11 .or. imethod .eq. 17 .or. &
             imethod .eq. 33) then
            qcfound(7)=.true.
         endif

         return
      endif

      istate = 1            ! By default on GS     
      lnmeigvec = .true.    ! search for Nmodes as eigenvectors
                            ! If set to false could use low precision
                            ! standard nmode output

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      eiglbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY NULL STRING"
         eiglbl = 18
      else if (imethod .eq. 1 .or. imethod.eq.19) then   ! SCF or DFT
         write(eigstring,'(a)') "Total energy in the final basis set"
         eiglbl = 35
      else if (imethod .eq. 11 .or. imethod.eq.33) then  ! TDDFT/CIS
         write(eigstring,'(a)') "Total energy for state"
         eiglbl = 22
      else if (imethod.eq.17) then                       ! EOMCCSD is
         write(eigstring,'(a)') "ENERGY NULL STRING"     ! done separately
         eiglbl = 18
      else if (imethod.eq.22) then                       ! CCSD
         write(eigstring,'(a)') "CCSD total energy"
         eiglbl = 17
      else 
         write(iqclog,'(/,a)') 'Routine Rdqchem (readqc.F90)'
         write(iqclog,'(a,i5)') &
            'ERROR: QC Method not implemented for energy:',imethod
         write(iqclog,'(a)') ' '
      endif

!-----------------------------------------------------------------------
! some methods (TDDFT, ...) require a gs energy with a different string
!-----------------------------------------------------------------------
      gslbl=0
      if (imethod.eq.17) then                       ! CCSD
         write(egsstring,'(a)') "CCSD total energy"
         gslbl = 17
      else if (imethod .eq. 11 .or. imethod .eq. 33) then
         write(egsstring,'(a)') "Total energy in the final basis set"
         gslbl = 35
      else
         write(egsstring,'(a)') "NO EGS STRING"
         gslbl = 13
      endif

!-----------------------------------------------------------------------
! string for finding gradient depends on method
!-----------------------------------------------------------------------
      flbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(grastring,'(a)') "GRADIENT NULL STRING"
         flbl = 20
      else if (imethod .eq. 1 .or. imethod .eq. 19) then
         write(grastring,'(a)') "Gradient of SCF Energy"
         flbl = 24
      else if (imethod .eq. 11 .or. imethod .eq. 33) then
         write(grastring,'(a)') &
             "Gradient of the state energy (including CIS Excitation"
         flbl = 54
      else if (imethod .eq. 17 .or. imethod .eq. 22) then   ! CCSD, EOMCCSD
         write(grastring,'(a)') "Full Analytical Gradient"
         flbl = 24
      else
         write(grastring,'(a)') "NO GRA STRING"
         flbl = 13
      endif

!-----------------------------------------------------------------------
! some methods (TDDFT, ...) have a gs gradient with a different string
!-----------------------------------------------------------------------
      if (imethod .eq. 11 .or. imethod.eq.33) then
         write(gragsstring,'(a)') "Gradient of SCF Energy"
         fgslbl = 22
      else
         write(gragsstring,'(a)') "NO GRA STRING"
         fgslbl = 13
      endif 

!-----------------------------------------------------------------------
! string for finding Hessian
! string for GS Hessian needed for TDDFT, CIS, ....
!-----------------------------------------------------------------------
      if (imethod .eq. 1 .or. imethod .eq. 19) then
          hesstring = "Hessian of the SCF Energy"
          heslbl = 25
          hesgsstring = "NO GSHES STRING"
          hesgslbl = 15
      else if (imethod .eq. 17 .or. imethod .eq. 22) then
          hesstring = "Final Hessian"
          heslbl = 13
          hesgsstring = "NO GSHES STRING"
          hesgslbl = 15
      else if (imethod .eq. 11 .or. imethod .eq. 33) then
          hesstring = "Hessian of the CIS Energy"
          heslbl = 25
          hesgsstring = "Hessian of the SCF Energy"
          hesgslbl = 25
      else
         write(hesstring,'(a)') "NO HES STRING"
         heslbl = 13
         write(hesgsstring,'(a)') "NO GSHES STRING"
         hesgslbl = 15
      endif

!-----------------------------------------------------------------------
! string for finding normal modes
!-----------------------------------------------------------------------
      nmstring = "Eigenvectors of Proj. Mass-Weighted Hessian"
      nmlbl = 43
      nmstring1 = "VIBRATIONAL ANALYSIS"
      nmlbl1 = 20

!-----------------------------------------------------------------------
! string for finding NAC vectors
!-----------------------------------------------------------------------
      naclbl=0
      if (imethod .eq. 11 .or. imethod .eq. 33) then
         write(nacstring,'(a)') "CIS Force Matrix"
         naclbl = 16
      else if (imethod .eq. 17) then
         write(nacstring,'(a)') "NAC Force h^x"
         naclbl = 13
      else
         write(nacstring,'(a)') "NO NAC STRING"
         naclbl = 13
      endif

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
!-----------------------------------------------------------------------
         if (index(string,'Standard Nuclear Orientation') .ne. 0) then

            read(iunit,*)
            read(iunit,*)

            n = 0
            do i=1,qc_ncoo/3
               read(iunit,22) idum,atmp,(coo(j),j=1,3)
               call getinam(atmp,isym,isym1,iqclog)  ! convert atom name to atomic symbol
               do j=1,3
                  n=n+1
                  qciatsym(n)=isym
                  qcxcoo(n)=coo(j)
               enddo
            enddo

! Convert to Bohr
            qcxcoo(1:qc_ncoo) = qcxcoo(1:qc_ncoo)/angstroem

 22         format(2x,i3,6x,a2,2x,3(3x,F14.10))

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (index(string,nmstring(1:nmlbl)) .ne. 0) then
            lnmeigvec = .true.   ! read normal modes as eigenvectors

            do j = 0,qc_ncoo/6-1
               read(iunit,*)
               read(iunit,*)
               do i=1,qc_ncoo
                  read(iunit,*,err=1002) (tmpderiv2(i,(6*j)+k),k=1,6)
               enddo
            enddo
            if (mod(qc_ncoo,6) > 0) then
               read(iunit,*)
               read(iunit,*)
               do i=1,qc_ncoo
                  read(iunit,*,err=1002) &
                           (tmpderiv2(i,k),k=(qc_ncoo/6)*6+1,qc_ncoo)
               enddo
            endif

            read(iunit,*)
            read(iunit,*)
            do i = 0,qc_ncoo/6-1
               read(iunit,*,err=1002) (tmpfreq((i*6)+j),j=1,6)
            enddo
            if (mod(qc_ncoo,6) > 0) then
               read(iunit,*,err=1002) (tmpfreq(j),j=(qc_ncoo/6)*6+1,qc_ncoo)
            endif

! Order absolute values of freqs 
            do i = i,qc_ncoo
               tmpfreq1(i) = dabs(tmpfreq(i))
               idxfreq(i) = i
            enddo
            call ibubble_up(tmpfreq1,idxfreq,qc_ncoo)
            do i = i,qc_ncoo
               tmpfreq1(i) = dabs(tmpfreq(i))
               idxfreq(i) = i
            enddo
            call ibubble_up(tmpfreq1,idxfreq,qc_ncoo)

! Ignore lowest ntr = 6(5) modes.
            ntr = qc_ncoo - qc_nmode
            lignore(:) = .false.
            do i=1,ntr
               j=idxfreq(i)
               lignore(j) = .true.
            enddo

! Store freqs and nmodes ignoring trans + rot
            j=0
            do i=1,qc_ncoo
               if (lignore(i)) cycle

               j=j+1
               qcfreq(j) = tmpfreq(i)
               do k=1,qc_ncoo
                  qcnmcoo(k,j) = tmpderiv2(k,i)
               enddo
            enddo

! The QChem output says "Vibrational Frequencies in atomic units"
! but these are actually the eigenvalues of the mass-weighted Hessian
! in amu. Convert.
            do j=1,qc_nmode
               qcfreq(j) = dsqrt(qcfreq(j)/atomicmassunit)
            enddo

! remove mass-weighting in nm coords and re-normalise
            do i= 1,qc_ncoo
               call getmass(mass(i),qciatsym(i),iau,iqclog)
               qcnmcoo(i,1:qc_nmode) = qcnmcoo(i,1:qc_nmode)/sqrt(mass(i))
            enddo
            do i= 1,qc_nmode
               norm = dnrm2(qc_ncoo,qcnmcoo(1:qc_ncoo,i),1)
               qcnmcoo(1:qc_ncoo,i) = qcnmcoo(1:qc_ncoo,i)/norm
            enddo

            qcnmlab(1:qc_nmode) = 'A'    ! labels not given in QChem

            qcfound(2)=.true.

         else if (index(string,nmstring1(1:nmlbl1)) .ne. 0) then
            if (lnmeigvec) cycle   ! normal modes read as eigenvectors

            do i=1,7
               read(iunit,'(a)',end=1002) string
            enddo

            nrd = qc_nmode/3

            nm=0
            do n=1,nrd
               do i=1,3
                  read(iunit,*)
               enddo
               read(iunit,'(3x,3(13x,f10.4))',err=1002) &
                           qcfreq(nm+1),qcfreq(nm+2),qcfreq(nm+3)

               do i=1,6
                  read(iunit,*)
               enddo

               do i=0,qc_ncoo/3-1
                  read(iunit,"(8x,3(2x,3(1x,f6.3)))",err=1002) &
                          ((qcnmcoo(3*i+j,nm+k),j=1,3),k=1,3)
               enddo

               nm = nm + 3
            enddo

            qcnmlab(1:qc_nmode) = 'A'    ! labels not given in QChem

! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.

!-----------------------------------------------------------------------
! ENERGY  
!-----------------------------------------------------------------------
         else if (index(string,eigstring(1:eiglbl)) .ne. 0) then 


            if (imethod .eq. 0) cycle   ! do not read energy
            if (lfreq) cycle  ! ignore energy if part of freq calculation

            if (imethod .eq. 1 .or. imethod .eq. 22 .or. &
                imethod.eq.19) then
               neq=index(string,"=")+1
               read(string(neq:80),*,err=1003) qcv(1)
               lqcfound3(1)=.true.

            else if (imethod .eq. 11 .or. imethod.eq.33) then
               bid = index(string,eigstring(1:eiglbl))+eiglbl
               eid = index(string,":")-1
               read(string(bid:eid),*,err=1003) istate
               istate = istate+1 ! state index starts at 0
               istate = min(istate,qc_nst+1)  ! trap unwanted states
               bid = eid+2
               eid = index(string,'au')-1
               read(string(bid:eid),*,err=1003) qcv(istate)
               lqcfound3(istate)=.true.

            else
               write(iqclog,'(/,a,i5)') &
                   'ERROR: Unknown ab initio type for Qchem output :',imethod
               cycle
            endif

            qcfound(3)=.true.

         else if ((index(string,'EOM') .ne. 0) .and. &
                  (index(string,'transition') .ne. 0)) then  ! EOMCCSD varieties

            if (imethod .ne. 17) cycle   ! do not read energy
            if (lfreq) cycle  ! ignore energy if part of freq calculation

            bid = index(string,eigstring(1:eiglbl))+eiglbl
            eid = index(string,"/")-1
            read(string(bid:eid),*,err=1003) istate
            istate = istate+1 ! state index starts at 0
            istate = min(istate,qc_nst+1)  ! trap unwanted states

            read(iunit,'(a)') string
            bid = index(string,"=")+1
            eid = index(string,"a.u.")-1

            read(string(bid:eid),*,err=1003) qcv(istate)
            lqcfound3(istate)=.true.

            qcfound(3)=.true.

!-----------------------------------------------------------------------
! GS ENERGY (required by e.g. TDDFT, ...)
!-----------------------------------------------------------------------
         else if (index(string,egsstring(1:gslbl)) .ne. 0) then

            if (imethod .eq. 0) cycle   ! do not read energy
            if (lfreq) cycle  ! ignore energy if part of freq calculation

            if (imethod .eq. 11 .or. imethod .eq. 17. .or. &
                imethod.eq.33) then  ! read DFT for TDDFT, CCSD for EOMCCSD etc.
               ilbl=index(string,"=")+1                   
               jlbl = 80
               read (string(ilbl:jlbl),*) egrd      
            endif

            qcv(1) = egrd
            qcfound(3)=.true.
            lqcfound3(1)=.true.

!-----------------------------------------------------------------------
! From REM block, Read calculation type. Force or freq?
! Get states treated in next calculation 
! OR
! get states in following coupling matrix elements
!-----------------------------------------------------------------------
         else if (string(1:4)=="$rem") then
             stread=0
             lforce=.false.
             lfreq=.false.
             do 
                read(iunit,'(a)',end=900,err=1000) string
                if ((index(string,"FORCE") .ne. 0) .or. &
                       (index(string,"force") .ne. 0)) then
                   lforce = .true.
                else if ((index(string,"FREQ") .ne. 0) .or. &
                       (index(string,"freq") .ne. 0)) then
                   lfreq = .true.
                else if ((index(string,"cis_state_deriv") .ne. 0) .or. &
                       (index(string,"CIS_STATE_DERIV") .ne. 0)) then 
                   read(string(16:),*) stread 
                else if ((index(string,"cc_state_to_opt") .ne. 0) .or. &
                       (index(string,"CC_STATE_TO_OPT") .ne. 0)) then 
                   bid = index(string, '[')+1
                   eid = index(string, ']')-1
                   bid = index(string(bid:eid),',')+bid
                   read (string(bid:eid),*, err=1000) stread

                else if (string(1:4)=="$end") then
                    exit
                endif
             enddo
             stread=stread+1  ! indexing starts at 0
             stread1=0  ! indexing starts at 0

         else if (index(string,"between states") .ne. 0) then
             bid = index(string,"between states")+15
             eid = index(string,"and")-1
             read(string(bid:eid),*,err=1000) stread
             bid = eid+4
             eid = bid+10
             read(string(bid:eid),*,err=1000) stread1
             stread=stread+1 ! state labels start at 0
             stread1=stread1+1
             stread=min(stread,qc_nst+1)
             stread1=min(stread1,qc_nst+1)

         else if (index(string,"State A") .ne. 0) then
             bid = index(string,":",lback)+1
             eid = index(string,"/",lback)-1
             read(string(bid:eid),*,err=1000) stread
             stread=stread+1 ! state labels start at 0
             stread=min(stread,qc_nst+1)

         else if (index(string,"State B") .ne. 0) then
             bid = index(string,":",lback)+1
             eid = index(string,"/",lback)-1
             read(string(bid:eid),*,err=1000) stread1
             stread1=stread1+1 ! state labels start at 0
             stread1=min(stread1,qc_nst+1)

!-----------------------------------------------------------------------
! FIRST DERIVATIVES
!-----------------------------------------------------------------------
         else if ((index(string,grastring(1:flbl)).ne.0) .or. &
                 (index(string,gragsstring(1:fgslbl)).ne.0)) then

            if (imethod .eq. 0) cycle   ! do not read gradients
            if (lfreq) cycle  ! ignore gradient if part of freq calculation

            found=.false.
            if (imethod .eq. 1 .or. imethod .eq. 11 .or. &
                imethod .eq. 17 .or.  imethod .eq. 19 .or. &
                imethod .eq. 22 .or. imethod .eq. 33) then

               if (imethod .eq. 11 .or. imethod .eq. 17 .or. &
                   imethod .eq. 33) then
                  istate = stread
               else
                  istate = 1
               endif
               do i = 0,qc_natm/6-1
                  read(iunit,*) 
                  do j = 1,3
                     read(iunit,*,err=1004) &
                       idum,(tmpderiv1(j,i*6+k),k=1,6)           
                  enddo
               enddo  
                         
               if (mod(qc_natm,6) .ge. 0) then
                  read(iunit,*)
                  do j = 1,3
                     read(iunit,*,err=1004) &
                       idum,(tmpderiv1(j,k),k=(qc_natm/6)*6+1,qc_natm)
                  enddo
               endif     

               found=.true.
            endif

            if (found) then
               istate = min(istate,qc_nst+1)  ! trap unwanted states
               k=0
               do i=1,qc_natm
                  do j =1,3
                     k = k+1
                     qcderiv1(k,istate) =tmpderiv1(j,i)
                  enddo
               enddo

               qcfound(4)=.true.
               lqcfound4(istate)=.true.
            else
               write(iqclog,'(/,a,i5)') &
                'ERROR: Unknown ab initio type for Qchem derivatives :',imethod
               cycle
            endif

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
!-----------------------------------------------------------------------
         else if ((index(string,hesstring(1:heslbl)) .ne. 0) .or. &
                 (index(string,hesgsstring(1:hesgslbl)) .ne. 0)) then
            if (imethod .eq. 0) cycle

            if (imethod .eq. 1 .or. imethod .eq. 19 .or. imethod .eq. 22) then
               istate = 1
            else if (imethod .eq. 11 .or. imethod .eq. 17 .or. &
                     imethod .eq. 33) then
               istate = stread
            endif
            istate = min(istate,qc_nst+1)  ! trap unwanted states

            do i = 0,qc_ncoo/6-1
               read(iunit,*)
               do j = 1,qc_ncoo
                  read(iunit,*,err=1005) idum,(qcderiv2(j,(i*6)+k,istate),k=1,6)
               enddo
            enddo
            if (mod(qc_ncoo,6) > 0) then
               read(iunit,*)
               do j = 1,qc_ncoo
                  read(iunit,*,err=1005) &
                          idum,(qcderiv2(j,k,istate),k=(qc_ncoo/6)*6+1,qc_ncoo)
               enddo
            endif

            qcfound(5)=.true.
            lqcfound5(istate)=.true.

!-----------------------------------------------------------------------
! DERIVATIVE COUPLINGS
! CIS Force Matrix = coupling with ETF * Ediff
!-----------------------------------------------------------------------
         else if (index(string,nacstring(1:naclbl)) .ne. 0) then
            istate = stread
            jstate = stread1
            read(iunit,*)
            read(iunit,*)

            n=0
            do i=1,qc_ncoo,3
               read(iunit,*) idum,(coo(j),j=1,3)
               do j=1,3
                  n=n+1
                  qcnact(n,istate,jstate)=coo(j)
               enddo
            enddo

            qcnact(:,jstate,istate) = qcnact(:,istate,jstate)

            lqcfound7(istate,jstate)=.true.
            lqcfound7(jstate,istate)=.true.
            qcfound(7)=.true.

!-----------------------------------------------------------------------
! DIPOLE MOMENT
!-----------------------------------------------------------------------
!        else if (index(string,"!RHF STATE 1.1 Dipole moment") .ne. 0) then

!           if (imethod .ne. 1) cycle

!           read(iunit,'(a)',err=1008) string
!           read(string,'(35x,3(3x,f12.8))',err=1008) (qcdipole(n,1),n=1,3)

!           qcfound(8)=.true.
!           lqcfound8(1)=.true.

!-----------------------------------------------------------------------
! Flag any error messages from QC
!-----------------------------------------------------------------------
!        else if(index(string,'MAXIMUM NUMBER OF ITERATIONS REACHED').ne.0)then
!           write(iqclog,'(a)') 'CAS not converged '
!           lfail = .true.
!        else if(index(string,'NO CONVERGENCE OF CP-MCSCF').ne.0)then
!           write(iqclog,'(a)') 'CPMCSCF not converged '
!           lfail = .true.

!!END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading QChem output'
      write(iqclog,'(a)') string
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading QChem output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading QChem output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading QChem output (Energy)'
      stop
 1004 write(iqclog,'(/,a)') 'ERROR reading QChem output (1st derivatives)'
      stop
 1005 write(iqclog,'(/,a)') 'ERROR reading QChem output (2nd derivatives)'
      stop
 1007 write(iqclog,'(/,a)') 'ERROR reading QChem output (NACTs)'
      stop
 1008 write(iqclog,'(/,a)') 'ERROR reading QChem output (Dipoles)'
      stop
      end subroutine rdqchem

!#######################################################################
!
! RDMOLCAS
! Read a MOLCAS Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!
! THIS ROUTINE IS UNTESTED
!#######################################################################

      subroutine rdmolcas(iunit,imethod,iqclog,lfail)

         implicit none

         integer(long), intent(in)   :: iunit,iqclog
         integer(long), intent(in)   :: imethod
   
         integer(long)               :: i,n,n1,n2,&
                                        j,k,nrd,iau
         integer(long)               :: ilbl,jlbl,klbl,clbl,&
                                        eiglbl,molbl,nmlbl,&
                                        der1lbl,der2lbl,naclbl,diplbl,&
                                        der1stalbl,nacstalbl,soclbl
         integer(long)               :: idum1,nbasis0
         integer(long)               :: istate,inact,jnact, &
                                        nsgra,nsnact,nshess,nvib,nsdip
         integer(long)               :: imcespin
         integer(long)               :: isym,isym1,iatpos ! Atom number
         real(dop)                   :: dtmp,ztmp,norm
         real(dop), dimension(qc_ncoo)         :: freqdum, mass
         real(dop), dimension(qc_ncoo,qc_nmode):: der2tmp
         real(dop), dimension(qc_ncoo,qc_ncoo) :: xdum
         character(len=80)           :: string
         character(len=60)           :: eigstring,coostring,&
                                        nmstring,der1string,&
                                        der2string,nacstring,dipstring,&
                                        mostring,der1stastring,nacstastring,&
                                        socstring
         character(len=10)           :: dummy, atmp
         logical(kind=4)             :: lder2

         logical(kind=4), dimension(qc_ncoo)    :: vibmode
         logical(kind=4), intent(out)           :: lfail

         interface
         function streq(str1,str2)
            logical :: streq
            character(len=*),intent(in) :: str1,str2
         end function streq
         function slen(str)
            integer :: slen
            character(len=*),intent(in) :: str
         end function slen
         function dnrm2(n,v,i)
            use decimal, only:dop
            real(dop) :: dnrm2
            integer,intent(in) :: n,i
            real(dop),dimension(n),intent(in) :: v
         end function dnrm2
         end interface
   !-----------------------------------------------------------------------
   ! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
   !-----------------------------------------------------------------------
         if (qc_enquire) then
            if (imethod .eq. 0) then
               qcfound(1)=.true.
               qcfound(2)=.true.
            else if (imethod .eq. 1 .or. imethod .eq. 5 .or. &
                  imethod .eq. 6) then
               qcfound(3)=.true.
               qcfound(4)=.true.
               qcfound(5)=.true.
            else if (imethod .eq. 12 .or. imethod .eq. 13 .or. &
                  imethod .eq. 27 .or. imethod .eq. 28 .or. &
                  imethod .eq. 32 .or. imethod .eq. 37) then
               qcfound(3)=.true.
               qcfound(4)=.true.
               qcfound(5)=.true.
               qcfound(6)=.true.
               qcfound(7)=.true.
               
            endif
            if (imethod .eq. 12 .or. imethod .eq. 27) then
               qcfound(8)=.true.
            endif
            if (imethod .eq. 12 .or. imethod .eq. 32) then
               qcfound(11)=.true.
            endif
            return
         endif

! By default on GS
      istate = 1

! imcespin : Spin of wavefunction for next read MCSCF info
      imcespin = 0

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
         ilbl=0
         if (imethod.eq.0) then      ! Do not read energy
            write(eigstring,'(a)') "ENERGY STRING"
            eiglbl = 13
         else if (imethod .eq. 1 .or. imethod .eq. 5 .or. imethod .eq. 6) then
            write(eigstring,'(a)') "::    Total SCF energy"
            eiglbl = 22
         else if (imethod .eq. 12) then   ! CASSCF + RASSCF
            write(eigstring,'(a)') "RASSCF root"
            eiglbl = 11
         else if (imethod .eq. 13 .or. imethod .eq. 32) then   ! CASPT2 or CAS/PT2
            write(eigstring,'(a)') "CASPT2 Root"
            eiglbl = 11 
         else if (imethod .eq. 27) then   ! (X)MS-CASPT2
            write(eigstring,'(a)') "MS-CASPT2 Root"
            eiglbl = 14 
         else if (imethod .eq. 28) then   ! RMS-CASPT2
            write(eigstring,'(a)') "RMS-CASPT2 Root"
            eiglbl = 15 
         else if (imethod .eq. 36 .or. imethod .eq. 37) then   ! XDW-CASPT2 or CAS/XDW
            write(eigstring,'(a)') "XDW-CASPT2 Root"
            eiglbl = 15 
         else 
            write(iqclog,'(/,a)') 'Routine Rdmolcas (readqc.F90)'
            write(iqclog,'(a,i5)') &
               'ERROR: QC Method not implemented :',imethod
            write(iqclog,'(a)') ' '
         endif
      
!-----------------------------------------------------------------------
! string for finding coordinates depends on orientation
!-----------------------------------------------------------------------
         clbl=0
         if (qcorient(1:8).eq.'standard') then
            coostring = 'Standard orientation:'
            clbl = 21
         else if (qcorient(1:8).eq.'z-matrix') then
            coostring = 'Z-Matrix orientation:'
            clbl = 21
         else 
            coostring = 'Input orientation:'
            clbl = 18
         endif
!-----------------------------------------------------------------------
! string for finding gradient
!-----------------------------------------------------------------------
         der1string = 'Molecular gradients'
         der1lbl = 19
         if (imethod .eq. 12 .or. imethod .eq. 32 .or. imethod .eq. 37) then    ! CASSCF, RASSCF, CAS/PT2, CAS/XDW
            der1stastring="multipliers are calculated for state no."
            der1stalbl=40
         else
            der1stastring="NONE MULTIPLIER FOR DER1STRING"
            der1stalbl=30
         endif
!-----------------------------------------------------------------------
! string for finding derivative coupling
!-----------------------------------------------------------------------
         nacstring = 'CI derivative coupling'
         naclbl = 22
         if (imethod .eq. 12 .or. imethod .eq. 32 .or. imethod .eq. 37) then    ! CASSCF, RASSCF, CAS/PT2, CAS/XDW
            nacstastring= "multipliers are calculated for states no."
            nacstalbl=41
         else
            nacstastring="NONE MULTIPLIER FOR NACTSTRING"
            nacstalbl=30
         endif
!-----------------------------------------------------------------------
! string for finding 2nd derivative (hessian)
! Only available if analytical 2nd derivative
!-----------------------------------------------------------------------
         der2string = '*Complete static Hessian*'
         der2lbl = 25
!-----------------------------------------------------------------------
! string for finding normal modes and frequencies (assume numerical
! unless analytical 2nd derivative)
!-----------------------------------------------------------------------
         nmstring = 'Harmonic frequencies in cm-1' 
         nmlbl = 28
!-----------------------------------------------------------------------
! string for finding SOCs
!-----------------------------------------------------------------------
         if (imethod .eq. 12 .or. imethod .eq. 32 .or. imethod .eq. 27 .or. imethod .eq. 37) then    ! CASSCF, RASSCF, CAS/PT2, CAS/XDW, MS-CASPT2
            socstring = 'Complex SO-Hamiltonian matrix elements'
            soclbl=38
         else
            socstring="NO SOCSTRING"
            soclbl=12
         endif
!-----------------------------------------------------------------------
! string for finding molecular orbitals in RasOrb
!-----------------------------------------------------------------------        
         mostring = '#ORB'
         molbl = 4
!-----------------------------------------------------------------------
! string for finding dipole moment
!-----------------------------------------------------------------------  
         if (imethod.eq.0) then 
            dipstring = 'DIPOLE NULL STRING'
            diplbl = 18
         else if (imethod .eq. 12 .or. imethod .eq. 32 .or. imethod .eq. 37) then   ! CASSCF, RASSCF, CAS/PT2, CAS/XDW
            dipstring = &
               'Expectation values of various properties for root number:'
            diplbl = 57
         else if (imethod .eq. 13 .or. imethod .eq. 27) then   ! ((X)MS)-CASPT2
            dipstring = 'Expectation values of various properties:'
            diplbl = 41
         endif
!-----------------------------------------------------------------------
! Initialize variable
!----------------------------------------------------------------------- 
         nsnact = 0
         nsgra = 0
         nshess = 0
         nsdip = 0
         inact = 1
         jnact = 1
         lder2 = .false. ! flag to determine if 2nd derivative available
                         ! only present with analytical hessian 
!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string
       
!-----------------------------------------------------------------------
! COORDINATES !Read in the geometry in Bohr
!-----------------------------------------------------------------------
         if (index(string,'**** Cartesian Coordinates / Bohr, Angstrom ****') .ne. 0)  then 

            do i=1,3
               read(iunit,*)
            enddo
            do i=1,qc_natm
               read(iunit,*,err=1001) idum1,atmp,(qcxcoo((i-1)*3+k),k=1,3)
               iatpos=numberindex(atmp) ! find the number position in atom labelling
               if (iatpos .eq. 0) then
                  write(atmp,'(a)') atmp
               else
                  write(atmp,'(a)') atmp(1:iatpos-1)
               endif
               call getinam(atmp,isym,isym1,iqclog)

               qciatsym((i-1)*3 + 1)=isym
               qciatsym((i-1)*3 + 2)=isym
               qciatsym((i-1)*3 + 3)=isym

            enddo
           
           
           qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
! 2nd derivative obtained from Normal modes if numerical 2nd derivatives
!-----------------------------------------------------------------------
         else if (index(string,nmstring(1:nmlbl)) .ne. 0) then
            qcnmlab(:) = "A"
            vibmode(:) = .false.
            if (lder2) then! Analytical evaluation. 3N normal modes
               nrd = qc_ncoo
            else! Numerical evaluation. 3N-6 or 3N-5 modes
               nrd = qc_nmode
            endif
            vibmode(1:nrd) = .true.
            nvib = 0
            do while (nvib .lt. nrd) !read all frequency and nm
               read(iunit,'(a)',end=900,err=1002) string
               if (index(string,"Frequency:") .ne. 0) then
                  ilbl = index(string,":")+5
                  jlbl = ilbl
                  n1 = min(6,nrd-nvib)
                  do i =1,n1
                     klbl = jlbl+index(string(jlbl:),'.')+2
                     read(string(jlbl:klbl),'(a)',err=1002) dummy
                     ilbl=index(dummy,"i") !check for imaginary frequency
                     read(dummy(ilbl+1:),'(f10.2)',err=1002) freqdum(nvib+i)
                     if (ilbl .ne. 0) then
                        freqdum(nvib+i) = -freqdum(nvib+i)
                     endif
                     jlbl = klbl+1
                  enddo
                  do i =1,4 !skip 4 lines
                     read(iunit,*)
                  enddo
                  do n =1,qc_ncoo !read matrix with row =cart and column = nm
                     read(iunit,'(20x,6f10.5)',err=1002) (xdum(n,nvib+k),k=1,n1)
                  enddo
                  nvib = nvib+6
               endif
            enddo

            if (lder2) then !Need to find translation and rotation
               nvib = 0
               do while (nvib .lt. nrd)
                  read(iunit,'(a)',err=1002) string
                  if (index(string,"Mode") .ne. 0) then
                     nvib = nvib+1
                     do i =1,2 !skip 2 lines
                        read(iunit,*)
                     enddo
                     read(iunit,'(a)',err=1002) string
                     if (index(string,"T") .ne. 0 .or. index(string,"R") .ne. 0) then
                        vibmode(nvib) = .false.
                     endif
                  endif
               enddo
            endif

            nvib = 0
            qcfreq(:) = 0.0_dop
            qcnmcoo(:,:) = 0.0_dop 
            do n =1,qc_ncoo
               if (vibmode(n)) then
                  nvib = nvib+1
                  !convert frequencies to au
                  qcfreq(nvib) = freqdum(n)/invcm
                  !Add the nm to matrix ignoring translation/rotation
                  qcnmcoo(1:qc_ncoo,nvib) = xdum(1:qc_ncoo,n)
               endif
            enddo

            qcfound(2) = .true.

            if (.not. lder2 .and. qcread(5)) then
               nshess = nshess+1
               iau = 1 !Get mass in amu
               der2tmp(:,:) = qcnmcoo(:,:)
               do i= 1,qc_ncoo
                  call getmass(mass(i),qciatsym(i),iau,iqclog)
                  der2tmp(i,1:qc_nmode) = der2tmp(i,1:qc_nmode)*sqrt(mass(i))
               enddo
               do n = 1,qc_nmode
                  !call nmortho(qcnmcoo,qc_ncoo,qc_ncoo,qc_nmode,iqclog)
                  norm = dnrm2(qc_ncoo,der2tmp(1:qc_ncoo,n),1)
                  der2tmp(1:qc_ncoo,n) = der2tmp(1:qc_ncoo,n)/norm
               enddo
               !Need mass in au for construction of deriv2
               iau = 0 !All input in au so no need for conversion
               call gender2(qcderiv2(:,:,nshess),der2tmp,qcfreq,qciatsym,iau,&
                        qc_ncoo,qc_ncoo,qc_nmode,iqclog)

               qcfound(5) = .true.
               lqcfound5(nshess) = .true.

            endif

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
! Note: Program wil read the LAST set of energies in the output file
!-----------------------------------------------------------------------

         else if (index(string,eigstring(1:eiglbl)) .ne. 0) then 

            if (imethod .eq. 0) cycle   ! do not read energy

            if (imethod .eq. 1 .or. imethod .eq. 5 .or. &
                  imethod .eq. 6) then
               read(string,'(51x,f17.9)',err=1000) qcv(1)
               lqcfound3(1)=.true.

            else if (imethod .eq. 12) then   ! CASSCF and RASSCF
               jlbl = index(string,"number")+7
               klbl = index(string,"Total")-2
               read(string(jlbl:klbl),'(i4)') istate

! Get spin block offset
               if (imcespin .eq. 0) then
                  write(iqclog,'(a)') 'Multiplicity not found'
                  stop
               endif
               if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle

               istate = istate + qc_spboff(imcespin)

               ilbl = index(string,":",.true.)+1 !Find last occurence of :
               read(string(ilbl:80),'(f16.8)',err=1003) qcv(istate)
               lqcfound3(istate)=.true.
   
            else if (imethod .eq. 13 .or. imethod .eq. 27 .or. &
                     imethod .eq. 28 .or. imethod .eq. 32 .or. &
                     imethod .eq. 36 .or. imethod .eq. 37) then   ! XXX-CASPT2
               jlbl = index(string,"Root")+5
               klbl = index(string,"Total")-6
               read(string(jlbl:klbl),'(i4)') istate

! Get spin block offset
               if (imcespin .eq. 0) then
                  write(iqclog,'(a)') 'Multiplicity not found'
                  stop
               endif
               if (istate.le.0 .or. istate .gt. qc_nmult(imcespin)) cycle

               istate = istate + qc_spboff(imcespin)

               ilbl = index(string,":",.true.)+1 !Find last occurence of :
               read(string(ilbl:80),'(f16.8)',err=1003) qcv(istate)
               lqcfound3(istate)=.true.

            endif
            qcfound(3)=.true.
         
!-----------------------------------------------------------------------
! FIRST DERIVATIVES
!
!  Grade state reading for CASSCF
!  For CASPT2 and others, just increment the nsgra
!-----------------------------------------------------------------------  
         else if (index(string,der1stastring(1:der1stalbl)) .ne. 0) then

            read(string(57:80),*,err=1004) nsgra
! Get spin block offset
            if (imcespin .eq. 0) then
               write(iqclog,'(a)') 'Multiplicity not found'
               stop
            endif
            if (nsgra.le.0 .or. nsgra .gt. qc_nmult(imcespin)) cycle

            nsgra = nsgra + qc_spboff(imcespin)

         else if (index(string,der1string(1:der1lbl)) .ne. 0) then
            if (imethod .eq. 0) cycle   ! do not read gradients

            if (imethod .eq. 13 .or. imethod .eq. 27 &
                                  .or. imethod .eq. 28) then ! increase nsgra 
                                                             ! as there is no 
                                                             ! info on state in MCLR
               nsgra = nsgra+1
            endif
            
            do i=1,7 !Skip 7 lines
               read(iunit,*,err=1004)
            enddo

            do n=1,qc_natm
               read(iunit,*,err=1004) dummy,&
                     (qcderiv1((n-1)*3+k,nsgra),k=1,3)
            enddo

            qcfound(4) = .true.
            lqcfound4(nsgra) = .true.

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
! Only available with analytical evaluation of 2nd derivatives
!-----------------------------------------------------------------------
         else if (index(string,der2string(1:der2lbl)) .ne. 0) then!WIP
            lder2 = .true.
            if (imethod .eq. 0) cycle ! do not read hessian
            nshess = nshess+1

!-----------------------------------------------------------------------
! MOLECULAR ORBITALS
! Format for the RasOrb file
!-----------------------------------------------------------------------
         else if (index(string,mostring(1:molbl)) .ne. 0) then

            rewind(iunit)

            do 
               read(iunit,'(a)',end=900,err=1006) string
               if (index(string,'RASSCF') .ne. 0) exit
            enddo            
            read(iunit,'(a)',end=900,err=1006) string
            read(iunit,*,end=900,err=1006) nbasis0
            if (nbasis0 .ne. qc_nbasis) then
                write(iqclog,'(/,a,i0)') 'ERROR reading MOLCAS output (MO coeffs.) &
                   & Check no. of basis functions. Nbasis =',nbasis0
                stop
            endif
            do 
               read(iunit,'(a)',end=900,err=1006) string
               if (index(string,mostring(1:molbl)) .ne. 0) exit
            enddo            
            do i = 1,qc_nbasis !loop over MO
               read(iunit,'(a)',end=900,err=1006) string !Skip ORBITAL line
               read(iunit,*,err=1006) (qcmocoeff(i,j),j=1,qc_nbasis)
            enddo            
            qcfound(6) = .true.

!-----------------------------------------------------------------------
! DERIVATIVES COUPLING
!-----------------------------------------------------------------------
         ! Read pair of state for CASSCF
         else if (index(string,nacstastring(1:nacstalbl)) .ne. 0) then 
            read(string(59:61),*,err=1007) inact
            read(string(63:65),*,err=1007) jnact
            if(inact.le.0 .or. jnact.le.0) cycle

! Get spin block offset
            if (inact.le.0 .or. inact .gt. qc_nmult(imcespin)) cycle
            if (jnact.le.0 .or. jnact .gt. qc_nmult(imcespin)) cycle
            inact = inact + qc_spboff(imcespin)
            jnact = jnact + qc_spboff(imcespin)

         else if (index(string,nacstring(1:naclbl)) .ne. 0) then
            if (imethod .eq. 0) cycle ! do not read derivative coupling

            if (imethod .eq. 13 .or. imethod .eq. 27 & 
                            .or. imethod .eq. 28) then ! increase nact pair of state as there is no info on state in MCLR
               jnact = jnact+1  !need to change for different spin mult.
               if (jnact .gt. qc_nst) then
                  inact = inact+1
                  jnact = inact+1
               endif
            endif

            do i =1,7 !skip 7 lines
               read(iunit,*,err=1007)
            enddo
            do n =1,qc_natm
               read(iunit,*,err=1007) dummy,&
                  (qcnact((n-1)*3+k,inact,jnact),k=1,3)
            enddo
            qcnact(:,jnact,inact) = qcnact(:,inact,jnact)

            qcfound(7) = .true.
            lqcfound7(inact,jnact) = .true.
            lqcfound7(jnact,inact) = .true.

!-----------------------------------------------------------------------
! DIPOLE MOMENT
!-----------------------------------------------------------------------
         else if (index(string,dipstring(1:diplbl)) .ne. 0) then
            if (imethod .eq. 0) then !do not read dipole
               cycle
            else if (imethod .eq. 12 .or. imethod .eq. 32 .or. imethod .eq. 37) then ! CASSCF, RASSCF, CAS/PT2, CAS/XDW
               ilbl = index(string,':')+1
               read(string(ilbl:),'(i4)') nsdip
            else if (imethod .eq. 13 .or. imethod .eq. 27) then ! ((X)MS)-CASPT2
               nsdip = nsdip+1
            endif

            if (nsdip .gt. qc_nst) cycle

            do !read files until dipole moment is found
               read(iunit,'(a)',end=900,err=1008) string
               if (index(string,'Dipole Moment') .ne. 0) exit
            enddo
            read(iunit,*) !skip 1 line
            ilbl = 1
            read(iunit,'(a)',end=900,err=1008) string
            do i =1,3
               ilbl = index(string(ilbl:),'=')+1
               read(string(ilbl:ilbl+12),'(E12.4E2)') qcdipole(i,nsdip) 
            enddo

            qcfound(8) = .true.
            lqcfound8(nsdip) = .true.

!-----------------------------------------------------------------------
! SOCs
! NB SOCs use opposite ordering of components to Molpro. I.e. negative 
!    first. This corresponds to different definition of z-axis
!-----------------------------------------------------------------------
         else if (index(string,socstring(1:soclbl)) .ne. 0) then
            if (imethod .eq. 0) then !do not read SOCs
               cycle

            else if (imethod .eq. 12 .or. imethod .eq. 32 .or. &
                     imethod .eq. 27 .or. imethod .eq. 37) then ! CASSCF, RASSCF, CAS/PT2, CAS/XDW, MS-CASPT2
            ! read and store SOCS
               do i=1,5
                  read(iunit,*)
               enddo
               n = (qc_nstdeg*(qc_nstdeg+1))/2 
               do i = 1,n
                  read(iunit,'(1x,2(i5,10x),2(6x,f8.3))',err=1011) &
                       n1,n2,dtmp,ztmp
                  qcsoc(n1,n2) = dcmplx(dtmp,ztmp)
                  qcsoc(n2,n1) = dconjg(qcsoc(n1,n2))
                  lqcfound11(n1,n2) = .true.
                  lqcfound11(n2,n1) = .true.
               enddo

! Convert SOCs from cm-1 to au.
              qcsoc = qcsoc/invcm

              qcfound(11) = .true.
           endif

!-----------------------------------------------------------------------
! Get multiplicity and no. of states of next energy set
!-----------------------------------------------------------------------
         else if (index(string,'Wave function specifications:') .ne. 0) then

            if (qc_nspblk .eq. 1) then   ! Spins not read if only single
               imcespin = 1              ! multiplicity in calculation
               cycle
            endif

            imcespin = 0
            if (imethod .eq. 1 .or. imethod .eq. 5 .or. imethod .eq. 6) &
                     go to 500 ! ignore for HF

            do
               read(iunit,'(a)',err=1000) string
               if (index(string,'Spin quantum number').ne.0)then
                  ilbl = index(string,'number')+7
                  read(string(ilbl:60),*) dtmp
                  imcespin = nint(2.0*dtmp+1.0)

               else if (index(string,'Number of roots').ne.0)then   ! ALASKA
                  exit

               else if (index(string,'Number of root(s)').ne.0)then ! CASPT2
                  exit
               else if (index(string,'highest root included in the CI').ne.0)then
                                                                    ! RASSCF
                  exit    

               endif
            enddo

   500      continue 

!-----------------------------------------------------------------------
! Flag any error messages from QC
!-----------------------------------------------------------------------
         else if (index(string,'rc= _NOT_CONVERGED_') .ne. 0) then
            write(iqclog,'(a)') 'Calculation not converged'
            lfail = .true.

! END OF IF-BLOCK
         endif

      enddo   ! end of read loop
   
   900 continue  ! End of file

        return
         1000 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output'
              stop
         1001 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Coordinates)'
              stop
         1002 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Normal Modes)'
              stop
         1003 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Energy)'
              stop
         1004 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (1st derivatives)'
              stop
         1005 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (2nd derivatives)'
              stop
         1006 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (MO coeffs.)'
              stop
         1007 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (NACTs)'
              stop
         1008 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Dipoles)'
              stop
         1009 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Transition Dipoles)'
              stop
         1010 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (Isotopes)'
              stop        
         1011 write(iqclog,'(/,a)') 'ERROR reading MOLCAS output (SOCs)'
              stop        
      end subroutine rdmolcas

!#######################################################################
! Re-order spin states from Molcas to readqc standard (molpro order)
!#######################################################################

      subroutine molcas_mapsoc(socmap,s,spblk)

      implicit none

      integer(long), intent(inout) :: s,spblk
      integer(long), dimension(spblk), intent(out) :: socmap
      integer(long)                :: s1,s2

! s is no. of first state in multiplicity set
! s2 is no. of last state in multiplicity set
      s2 = s+spblk
      s = s-1
      do s1 = 1,spblk
         s2 = s2-1
         s = s+1
         socmap(s) = s2
      enddo

      return
      end subroutine molcas_mapsoc

!#######################################################################
!
! RDORCA
! Read a ORCA Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!
! THIS ROUTINE IS UNTESTED
!#######################################################################

      subroutine rdorca(iunit,imethod,iqclog)

      implicit none

      integer(long), intent(in)   :: iunit,iqclog
      integer(long), intent(in)   :: imethod

      integer(long)               :: i,j,n,isym,isym1,nm,nmoffs,nrd
      integer(long)               :: nroot,ist,jst,ioffs
      integer(long)               :: ilbl,gslbl
      integer(long), dimension(10) :: dgen

      real(dop)                   :: vtmp,small,tmp,vdiff
      real(dop), dimension(:), allocatable :: tmpvec
      real(dop), dimension(3)     :: coo
      real(dop), dimension(6)     :: xdum
      real(dop), dimension(10)    :: entmp
      character(len=80)           :: string
      character(len=40)           :: eigstring
      character(len=40)           :: egsstring
      character(len=20)           :: atmp

!-----------------------------------------------------------------------
! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
!-----------------------------------------------------------------------
      if (qc_enquire) then
         if (imethod .eq. 0) then
            qcfound(1)=.true.
            qcfound(2)=.true.
         else if (imethod .eq. 1 .or. imethod .eq. 9) then
            qcfound(3)=.true.
         else if (imethod .eq. 33) then
            qcfound(3)=.true.
            qcfound(7)=.true.
         endif

         return
      endif

      nroot=1    !   state of interest for derivatives etc. By default
                 !   this is GS. 

!-----------------------------------------------------------------------
! string for finding energy depends on method
!-----------------------------------------------------------------------
      ilbl=0
      if (imethod.eq.0) then      ! Do not read energy
         write(eigstring,'(a)') "ENERGY NULL STRING"
         ilbl = 18
      else if (imethod.eq.1 .or. imethod .eq. 9) then     ! HF/DFT
         write(eigstring,'(a)') "Total Energy"
         ilbl = 12
      else if (imethod.eq.33) then                        ! TDDFT
         write(eigstring,'(a)') "TD-DFT/TDA EXCITED STATES" 
         ilbl = 5 
      endif

!-----------------------------------------------------------------------
! some methods have a different string for the GS energy
!-----------------------------------------------------------------------
      if (imethod.eq.33) then
         write(egsstring,'(a)') "Total Energy"
         gslbl = 12
      else
         write(egsstring,'(a)') "NO EGS STRING"
         gslbl = 13
      endif
      
      nroot = 0 ! counter for CR-EOM-CCSD
      small=0.0011

      entmp(:)=0.0_dop
      dgen(:)=0

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES
! (returns last in file)
!-----------------------------------------------------------------------
         if (index(string,'CARTESIAN COORDINATES (ANGSTROEM)') .ne. 0) then

            read(iunit,*,err=1001)
            n=0

            do i=1,qc_ncoo/3
               read(iunit,'(2x,a3,3f12.6)') atmp,coo(1),coo(2),coo(3)

! convert atom name to atomic symbol
               call getinam(atmp,isym,isym1,iqclog)  
               do j=1,3
                  n=n+1
                  qciatsym(n)=isym
                  qcxcoo(n)=coo(j)
               enddo

            enddo

! Convert to Bohr
            qcxcoo(1:qc_ncoo) = qcxcoo(1:qc_ncoo)/angstroem

            qcfound(1)=.true.

!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         else if (index(string,"VIBRATIONAL FREQUENCIES") .ne. 0) then

            do i=1,4
               read(iunit,*,err=1002)
            enddo
            do i = 1,qc_ncoo
               read(iunit,'(10x,f8.2)',err=1002) qcfreq(i)
            enddo

            nrd = (qc_ncoo-1)/6 + 1

! read down to Nmodes
            do
               read(iunit,'(a)',err=1002) string
               if (index(string,'NORMAL MODES') .ne. 0) exit 
            enddo
            do i=1,6
               read(iunit,*,err=1002)
            enddo

            nm=0
            nmoffs=0
            do n=1,nrd
               read(iunit,'(a)',err=1002) string

               if(n.eq.nrd .and. mod(qc_ncoo,6).ne.0)then
                  nm = mod(qc_ncoo,6)
               else
                  nm = 6
               endif

               do j=1,qc_ncoo
                  read(iunit,'(a)',err=1002) string
                  read(string,20) (xdum(i),i=1,6)
                  do i=1,nm
                     qcnmcoo(j,i+nmoffs)=xdum(i)
                  enddo
               enddo
               nmoffs = nmoffs+6
            enddo

!--------------------------------------------------------------------
! reorder so that the indices of imaginary frequencies and
! corresponding matrix elements are within the range 1,2,...,nmodes
! i.e. get rid of the zero frequency modes
!--------------------------------------------------------------------
            allocate(tmpvec(qc_ncoo))
            do while (abs(qcfreq(1)).lt.1.0e-6_dop)
               tmp = qcfreq(1)
               tmpvec(1:qc_ncoo) = qcnmcoo(1:qc_ncoo,1)
               do j = 2,qc_ncoo
                  qcfreq(j-1) = qcfreq(j)
                  qcnmcoo(1:qc_ncoo,j-1) = qcnmcoo(1:qc_ncoo,j)
               enddo
               qcnmcoo(1:qc_ncoo,qc_ncoo) = tmpvec(1:qc_ncoo)
               qcfreq(qc_ncoo) = tmp
            enddo
            deallocate(tmpvec)

! convert frequencies to au
            qcfreq(:)=qcfreq(:)/invcm

            qcfound(2)=.true.

!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         else if(index(string,eigstring(1:ilbl)) .ne. 0) then
            if (imethod .eq. 0) cycle   

            if (imethod .eq. 1 .or. imethod .eq.9) then
               read(string,'(28x,f14.8)',err=1003) qcv(1)
               lqcfound3(1)=.true.
               qcfound(3)=.true.

            else if (imethod .eq. 33) then
               ist = 1
               do
                  read(iunit,'(a)',err=1003) string
                  if (string(1:5).eq."STATE") then
                     read(string,'(15x,f9.6)',err=1000) vtmp
                     ist = ist + 1
                     ist = min(ist,qc_nst+1)    ! Trap unwated states
                     qcv(ist) = qcv(1) + vtmp   ! Add Egs
                     lqcfound3(ist) = .true.
                     qcfound(3)=.true.
                     if(ist.ge.qc_nst) exit
                  endif
               enddo

            endif

!-----------------------------------------------------------------------
! GS ENERGY (required by e.g. TDDFT)
!-----------------------------------------------------------------------
         else if (index(string,egsstring(1:gslbl)) .ne. 0) then

            if (imethod .eq. 0) cycle   ! do not read energy

            read(string,'(63x,F15.10)',err=1003) qcv(1)
            lqcfound3(1) = .true.

!-----------------------------------------------------------------------
! DERIVATIVE COUPLINGS
!
! TO BE CHECKED. Only works for 2 states?
!-----------------------------------------------------------------------
         elseif (string(1:33) .eq. "CARTESIAN NON-ADIABATIC COUPLINGS") then

            if (imethod .eq. 0) cycle   ! do not read gradients

            do
               read(iunit,'(a)',err=1007) string
               if(string(1:5) .eq. '-----') exit
            enddo

            ist = 1
            jst = qc_nst
            read(iunit,'(a)',err=1007) string
            ioffs = 0
            do n=1,qc_natm
               read(iunit,'(a)',err=1007) string
               read(string,fmt='(12x,3(f15.9))',err=1007) &
                    (qcnact(i+ioffs,ist,jst),i=1,3)
               ioffs = ioffs + 3
            enddo

! Multiply by energy gap.
            vdiff = qcv(jst)-qcv(ist)
            qcnact(:,ist,jst) = qcnact(:,ist,jst)*vdiff
            qcnact(:,jst,ist) = qcnact(:,ist,jst)

            lqcfound7(ist,jst)=.true.
            lqcfound7(jst,ist)=.true.
            qcfound(7)=.true.

! END OF IF-BLOCK
         endif

      enddo

  900 continue  ! End of file

   20 format(11x,6f11.6)

      return
 1000 write(iqclog,'(/,a)') 'ERROR reading ORCA output'
      stop
 1001 write(iqclog,'(/,a)') 'ERROR reading ORCA output (Coordinates)'
      stop
 1002 write(iqclog,'(/,a)') 'ERROR reading ORCA output (Normal Modes)'
      stop
 1003 write(iqclog,'(/,a)') 'ERROR reading ORCA output (Energy)'
      stop
 1007 write(iqclog,'(/,a)') 'ERROR reading ORCA output (NACTs)'
      stop
      end subroutine rdorca

!#######################################################################
!
! RDBAGEL
! Read a BAGEL Output file
!
! iunit   : unit connected to log file
! imethod : type of calculation to be read
! iqclog  : unit for log messages
!
! Other variables and data arrays are defined above in module readqc 
!
!#######################################################################

      subroutine rdbagel(iunit,imethod,iqclog)

         implicit none

         integer(long), intent(in)  :: iunit,iqclog
         integer(long), intent(in)  :: imethod
         !Local variable
         character(len=64)          :: coostring,eigstring,grastring,&
                                       nacstring,der1string,nmstring,&
                                       der2string,dipstring,tdstring,&
                                       mostring,freqstring
         integer(long)              :: coolbl,eilbl,gralbl,naclbl,der1lbl,&
                                       nmlbl,der2lbl,diplbl,tdlbl,molbl,&
                                       ilbl,jlbl,freqlbl
         integer(long)              :: ns,ntr,n,&
                                       i,j,k,koff,istate,jstate,neq,n1
         integer(long)              :: nsgra,nshess,nsdip
         integer(long)              :: isym,isym1 ! Atom symbol (temp)
         integer(long)              :: icoo,iau1,hi,lo

         real(dop)                  :: mag
         real(dop), dimension(3)    :: vec
         real(dop), dimension(qc_ncoo) :: freqdum,force,atmass
         real(dop), dimension(qc_ncoo,qc_ncoo) :: nmcoodum
         logical(kind=4)            :: lgra,lnact,nmcoofound,rtfound,ldeigen
         character(len=8)           :: atmp
         character(len=150)         :: string

         integer(long), dimension(qc_nst) :: mapgra,maphess


   !-----------------------------------------------------------------------
   ! TEST BLOCK FOR IMPLEMENTED CAPABILITIES
   !-----------------------------------------------------------------------
         if (qc_enquire) then
            if (imethod .eq. 0) then
               qcfound(1)=.true.
               qcfound(2)=.true.
            else if (imethod .eq. 12 .or. imethod .eq. 27) then
               qcfound(3)=.true.
               qcfound(4)=.true.
               qcfound(5)=.true.
               qcfound(6)=.true.
               qcfound(7)=.true.
               qcfound(8)=.true.
            endif
   
            return
         endif       

         if (imethod .ne. 0) call rdbagelinp(iunit,iqclog,mapgra,maphess)

!-----------------------------------------------------------------------
! string for finding coordinate
!-----------------------------------------------------------------------
         write(coostring,'(a)') "*** Geometry ***"
         coolbl = 16

!-----------------------------------------------------------------------
! string for finding energy 
!-----------------------------------------------------------------------
         if (imethod .eq. 0) then !Do not read energy
            write(eigstring,'(a)') "ENERGY NULL STRING"
            eilbl = 18
         else if (imethod .eq. 12) then ! CASSCF
            write(eigstring,'(a)') &
               "* Second-order optimization converged. *"
            eilbl = 40
         else if (imethod .eq. 27) then !(X)MS-CASPT2
            write(eigstring,'(a)') "* MS-CASPT2 energy"
            eilbl = 18
!         else if (imethod .eq. 32) then !CAS/PT2
!            write(eigstring,'(a)') "* MS-CASPT2 energy"
!            eilbl = 18
         endif
!-----------------------------------------------------------------------
! string for finding gradient
!-----------------------------------------------------------------------
         if (imethod .eq. 0) then !Do not read gradient
            write(grastring,'(a)') "GRADIENT NULL STRING"
            gralbl = 20
         else if (imethod .eq. 12) then ! CASSCF
            write(grastring,'(a)') "* Permanent dipole moment: Unrelaxed"
            gralbl = 36
         else if (imethod .eq. 27) then !(X)MS-CASPT2
            write(grastring,'(a)') &
               "* CASPT2 energy:"
            gralbl = 17

!         else if (imethod .eq. 32) then !CAS/PT2
!            write(grastring,'(a)') &
!               "* Permanent dipole moment: CASPT2 unrelaxed"
!            gralbl = 43
         endif
!-----------------------------------------------------------------------
! string for finding derivative coupling
!-----------------------------------------------------------------------
         write(nacstring,'(a)') "* NACME Target states:"
         naclbl = 22
!-----------------------------------------------------------------------
! string for finding the 1st derivatives
!-----------------------------------------------------------------------         
         write(der1string,'(a)') "* Nuclear energy gradient"
         der1lbl = 25
!-----------------------------------------------------------------------
! string for finding frequencies
!-----------------------------------------------------------------------
         write(freqstring,'(a)') "* Vibrational frequencies, IR intensities,"
         freqlbl = 42
!-----------------------------------------------------------------------
! string for finding normal modes
!-----------------------------------------------------------------------
         write(nmstring,'(a)') "++ Mass Weighted Hessian Eigenvectors ++"
         nmlbl = 40
!-----------------------------------------------------------------------
! string for finding Hessian matrix (deriv2)
!-----------------------------------------------------------------------
         write(der2string,'(a)') "++ Hessian ++"
         der2lbl = 13
!-----------------------------------------------------------------------
! string for finding dipole moment
!-----------------------------------------------------------------------
         if (imethod .eq. 0) then !No reading of diple
            dipstring = "NULL DIPOLE STRING"
            diplbl = 18
         else if (imethod .eq. 12) then ! CASSCF
            dipstring = "* Permanent dipole moment: Relaxed"
            diplbl = 34
         else if (imethod .eq. 27) then !(X)MS-CASPT2
            dipstring = "* Permanent dipole moment: CASPT2 relaxed"
            diplbl = 41
!         else if (imethod .eq. 32) then !CAS/PT2
!            dipstring = "* Permanent dipole moment: CASPT2 relaxed"
!            diplbl = 41
         endif
!-----------------------------------------------------------------------
! string for finding transition dipole moment
!-----------------------------------------------------------------------         
         tdstring = &
            "* Permanent dipole moment: Transition dipole moment between"
         tdlbl = 59
!-----------------------------------------------------------------------
! string for finding MO
!-----------------------------------------------------------------------
         write(mostring,'(a)') "[MO]"
         molbl = 4
!-----------------------------------------------------------------------
! Initialize variable
!-----------------------------------------------------------------------
      ns = 0
      nsgra= 0
      nshess= 0
      nsdip= 0
      lgra = .false.
      lnact = .false.
      nmcoofound = .false.
      ldeigen = .false.
!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------
      do      
         read(iunit,'(a)',end=900,err=1000) string

!-----------------------------------------------------------------------
! COORDINATES !Read geometry in Bohr
!-----------------------------------------------------------------------
         if (stringcontains(coostring,string)) then
            read(iunit,*)
            do i=1,qc_ncoo/3
               read(iunit,'(a)',err=1001) string
               atmp=string(15:16)
               !Check if atom name with 1 or 2 letters
               if (atmp(2:2) .eq. '"') then
                  read(string,20) atmp,(qcxcoo((i-1)*3+j),j=1,3)
               else 
                  read(string,30) atmp,(qcxcoo((i-1)*3+j),j=1,3)
               endif
               ! Convert atom names to atomic numbers
               call getinam(atmp,isym,isym1,iqclog)
               ! Store in global results vectors
               do j=1,3
                  n=(i-1)*3+j
                  qciatsym(n)=isym
               enddo
            enddo
            qcfound(1)=.true.

            20  format(14x,a1,11x,3(1x,F14.6)) !coordinate format for 1 letter atom label
            30  format(14x,a2,11x,3(1x,F14.6)) !coordinate format for 2 letter atom label
      
!-----------------------------------------------------------------------
! NORMAL MODES AND FREQUENCIES
!-----------------------------------------------------------------------
         
         else if (stringcontains(nmstring,string)) then
            do i=1,qc_ncoo,6 ! Read NM by block of 6 (including rot and trans)
               read(iunit,*)
               read(iunit,*)

               if (i+5 > qc_ncoo) then
                  n=mod(qc_ncoo,6)
               else
                  n=6
               endif
               do j=1,qc_ncoo
                  read(iunit,*,err=1002) icoo,(nmcoodum(j,i-1+k),k=1,n)
               enddo
            enddo
               
            ! Get atomic masses (in amu)
            iau1=1
            do j=1,qc_ncoo
               call getmass(atmass(j),qciatsym(j),iau1,iqclog)
            enddo
            ! The NM need to be in cartesian coordinate (no mass weighting)
            ! Remove mass-weighting 
            do i=1,qc_ncoo
               do j=1,qc_ncoo
                  nmcoodum(j,i)=nmcoodum(j,i) / sqrt(atmass(j))
               enddo
            enddo


            nmcoofound=.true.

         else if (stringcontains(freqstring,string)) then
            do i=1,qc_ncoo,6 !read frequency blocks (includes rot and trans modes)
               read(iunit,*)
               read(iunit,*)

               if (i+5 > qc_ncoo) then
                  n=mod(qc_ncoo,6)
               else
                  n=6
               endif

               read(iunit,'(a)',err=1002) string
               read(string(18:),*) (freqdum(i-1+k),k=1,n)

               do j=1,4 + qc_ncoo
                  read(iunit,*)
               enddo
            enddo

! Remove rotational and translational modes
            if (.not. nmcoofound) then
               write(iqclog,'(/,a)') &
                 'ERROR reading BAGEL output (Normal modes and frequencies): &
                 & Mass Weighted Hessian Eigenvectors were not found before &
                 & frequencies were read (Routine rdbagel (readqc.F90))'
               stop
            endif
            ntr=qc_ncoo-qc_nmode
            lo=1
            hi=ntr
            rtfound=.false.
            do while (hi .le. qc_ncoo .and. .not. rtfound)
               if (abs(freqdum(hi+1)) < abs(freqdum(lo))) then
                  lo=lo+1
                  hi=hi+1
               else
                  rtfound=.true.
               endif
            enddo

            write(iqclog,1011) lo, hi, freqdum(lo:hi)
            1011 format ('Modes ', I3, ' - ', I3, ' with frequencies ', 6F10.3,3X, &
            ' are assumed to be rotations and translations and have been discarded.')
            
            qcfreq(:lo-1)=freqdum(:lo-1)
            qcfreq(lo:)=freqdum(hi+1:)
            qcnmcoo(:,:lo-1)=nmcoodum(:,:lo-1)
            qcnmcoo(:,lo:)=nmcoodum(:,hi+1:)
            n=min(6,qc_ncoo-ntr)

            write(iqclog,1012) qcfreq(:n)
            1012 format ('The lowest vibrational frequencies that have been kept are ', 6F10.3,3X)

            qcfreq(:)=qcfreq(:)/invcm ! Convert frequencies from cm^-1 to au
            qcnmlab(:)='A'   ! Give a symmetry label

            qcfound(2)=.true. 
!-----------------------------------------------------------------------
! ENERGY
!-----------------------------------------------------------------------
         else if (stringcontains(eigstring,string)) then

            if (imethod .eq. 0) cycle !Skip energy
            
            if (imethod .eq. 12) then !CASSCF
               neq=8
               backspace(iunit)
               do i=1,qc_nst ! Go back to first value of energy
                  backspace(iunit)
               enddo
               read(iunit,'(a)',err=1003) string
               read(string(neq:neq+2),'(i2)') istate
               istate=istate+1

               if (istate .le. qc_nst) then
                  neq=11
                  read(string(neq:neq+18),'(f18.8)') qcv(istate)
                  qcfound(3)=.true.
                  lqcfound3(istate)=.true.
               endif

            else if (imethod .eq. 27 .or. imethod .eq. 32) then !(X)MS-CASPT2
               jlbl=index(string,"state")+6
               read(string(jlbl:jlbl+2),'(i2)') istate
               istate=istate+1

               if(istate .le. qc_nst) then
                  neq=36
                  read(string(neq:80),'(f18.10)') qcv(istate)
                  qcfound(3)=.true.
                  lqcfound3(istate)=.true.
               endif

            endif
!-----------------------------------------------------------------------
! GRADIENT (set flag that nuclear gradient is a gradient of state nsgra)
! Assume gradient printing in increasing order of electronic states
!-----------------------------------------------------------------------
         else if (stringcontains(grastring,string)) then  
            lgra = .true. 
            nsgra = nsgra+1
            
!-----------------------------------------------------------------------
! DERIVATIVE COUPLINGS (set flag that nuclear gradient is a noweight 
! bagel derivative couplings)
!-----------------------------------------------------------------------
         else if (stringcontains(nacstring,string)) then
            !Get the pair of state for the derivative coupling
            ilbl = index(string,":")+1
            jlbl = index(string,"-")
            read(string(ilbl+1:jlbl-1),*) istate
            read(string(jlbl+1:jlbl+4),*) jstate
            istate = istate+1
            jstate = jstate+1
            lnact = .true.

!-----------------------------------------------------------------------
! FIRST DERIVATIVES (either gradient or derivative coupling,
! based on previous logical flag set to true)
!-----------------------------------------------------------------------
         else if (stringcontains(der1string,string)) then
            if (imethod .eq. 0) cycle
            read(iunit,'(a)',err=1004) string

            do n=1,qc_natm
               read(iunit,'(a)',err=1004) string
               do j=1,3
                  read(iunit,'(13x,f18.10)',err=1004) force((n-1)*3+j)
               enddo
            enddo

            if (lgra .and. .not. lnact) then
               qcderiv1(:,mapgra(nsgra)) = force(:)
               qcfound(4)=.true.
               lqcfound4(mapgra(nsgra))=.true.
               lgra = .false.
            else if (lnact .and. .not. lgra) then
               qcnact(:,jstate,istate) = force(:)
               qcnact(:,istate,jstate) = force(:)
               qcfound(7)=.true.
               lqcfound7(istate,jstate)=.true.
               lqcfound7(jstate,istate)=.true.
               lnact = .false.
            else
               write(iqclog,*) &
                  "WARNING: 1st derivative reading not associated gradient or nact"
            end if

!-----------------------------------------------------------------------
! SECOND DERIVATIVES
!-----------------------------------------------------------------------
         else if (stringcontains(der2string,string)) then
            if (imethod .eq. 0) cycle
            nshess = nshess+1 ! assume incremental order of state for hessian
            !Read hessian by block of 6
            koff = 0
            do n = 1,qc_ncoo,6
               !Skip 1 lines
               read(iunit,*,err=1005) string
               n1 = min(6,qc_ncoo-(n-1))
               do j = 1,qc_ncoo 
                  read(iunit,'(i6,6(2x,f18.10))',err=1005) i,(qcderiv2(n+k,j,maphess(nshess)),k=0,n1-1)
               enddo
            enddo

            !Symmetrize
            do i=1,qc_ncoo
               do j=i+1,qc_ncoo
                  qcderiv2(i,j,maphess(nshess))=(qcderiv2(i,j,maphess(nshess))+qcderiv2(j,i,maphess(nshess)))/2.0d0
                  qcderiv2(j,i,maphess(nshess))=qcderiv2(i,j,maphess(nshess))
               enddo
            enddo


            qcfound(5)=.true.
            lqcfound5(maphess(nshess))=.true.
!-----------------------------------------------------------------------
! DIPOLE MOMENT (Assume nst dipole before transition dipole printing)
!-----------------------------------------------------------------------
         else if (stringcontains(dipstring,string)) then
            if (imethod .eq. 0) cycle

            nsdip = nsdip+1
            if (nsdip .gt. qc_nst) cycle

            read(iunit,'(a)',err=1008) string
            ilbl = index(string,"(")+1
            do n =1,3
               jlbl = index(string(ilbl:),",")
               if (jlbl .eq. 0) then
                  jlbl = index(string(ilbl:),")")
               endif
               read(string(ilbl:ilbl+jlbl-1),'(f12.6)') qcdipole(n,mapgra(nsdip)) !Read in a.u.
               ilbl = ilbl+jlbl+1
            enddo
            qcfound(8)=.true.
            lqcfound8(mapgra(nsdip))=.true.
!-----------------------------------------------------------------------
! TRANSITION DIPOLE
!-----------------------------------------------------------------------
         else if (stringcontains(tdstring,string)) then
            if (imethod .eq. 0) cycle
            !Read the pair of state for the transition dipole moment
            ilbl = index(string,"-")
            read(string(64:ilbl-1),*,err=1009) i
            read(string(ilbl+1:ilbl+4),*,err=1009) j
            i = i+1
            j = j+1

            read(iunit,'(a)',err=1009) string
            ilbl = index(string,"(")+1
            vec(:) = 0.0_dop
            do n=1,3
               jlbl = index(string(ilbl:),",")
               if (jlbl .eq. 0) then
                  jlbl = index(string(ilbl:),")")
               endif
               read(string(ilbl:ilbl+jlbl-1),'(f12.6)') vec(n) !Read in a.u.
               ilbl = ilbl+jlbl+1
            enddo
            mag = 0.0_dop
            do n =1,3
               mag = mag + vec(n)*vec(n)
            enddo
            if (mag .ge. 1.0e-8_dop) then
! tdip(4) is magnitude of dipole
               mag = dsqrt(mag)
               qctdipole(4,i,j) = mag
! tdip(i), i=1-3 is unit vector along dipole
               do n =1,3
                  vec(n) = vec(n)/mag
               enddo
! as convention take vector direction with positive dot product with (1,1,1)
               mag=0.0_dop
               do n =1,3
                  mag = mag + vec(n)
               enddo               
               if (mag .lt. 0.0_dop) then
                  do n =1,3
                     vec(n)=-vec(n)
                  enddo
               endif
               qctdipole(1:3,i,j) = vec(:)
            endif

            qctdipole(:,j,i) = qctdipole(:,i,j)

            qcfound(9)=.true.
            lqcfound9(i,j)=.true.
            lqcfound9(j,i)=.true.
!-----------------------------------------------------------------------
! MO COEFFICIENTS (from molden file)
!-----------------------------------------------------------------------
         else if (stringcontains(mostring,string))  then
            !Loop over the MO
            do i=1,qc_nbasis
               read(iunit,*,err=1006)
               read(iunit,*,err=1006)
               read(iunit,*,err=1006)
               !Loop over AO coefficients
               do j=1,qc_nbasis
                  read(iunit,50,err=1006) qcmocoeff(i,j)
               enddo
            enddo            
            
            qcfound(6)=.true.
            !Set the other flags to true as the file contains only the MO
            !May need a better way for checking output
            qcfound(1:5)=.true.
            qcfound(7:9)=.true.
         endif
      enddo ! end of read loop
  900 continue  ! End of file



  50  format(4x,1f23.16)


      return
      1000 write(iqclog,'(/,a)') 'ERROR reading BAGEL output'
           stop
      1001 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (Coordinates)'
           stop
      1002 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (Normal Modes)'
           stop
      1003 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (Energy)'
           stop
      1004 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (1st derivatives)'
           stop
      1005 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (2nd derivatives)'
           stop
      1006 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (MO coeffs.)'
           stop
      1008 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (Dipoles)'
           stop
      1009 write(iqclog,'(/,a)') 'ERROR reading BAGEL output (Transition Dipoles)'
           stop

!-----------------------------------------------------------------------
! read through the file
!-----------------------------------------------------------------------

      end subroutine rdbagel

!#######################################################################

      subroutine rdbagelinp(iunit,iqclog,mapgra,maphess)


      integer(long), intent(in)  :: iunit,iqclog

      integer(long)              :: iunit1=24

      integer(long)              :: ilbl,jlbl,klbl, cix
      integer(long)              :: sublevel, gra_start, hes_start, na_start
      integer(long)              :: nsgra,nshess
      character(len=150)         :: inpfile
      character(len=150)         :: string, substring

      logical(kind=4)            :: linpgra,linphess,linpfile
      logical(kind=4)            :: gra_sec, hes_sec, na_sec
      logical(kind=4)            :: target_found, natype_found
      integer(long), dimension(qc_nst), intent(out) :: mapgra,maphess
      !integer(long), dimension(qc_nst*(qc_nst-1)/2,2) :: mapnact

!-----------------------------------------------------------------------
! Get input file name
!-----------------------------------------------------------------------
      inquire(iunit,name=string)                  ! get output file name
      ilbl= index(string,'.',back=.true.)-1       ! look for an extension
      if (ilbl .le. 0) then        ! Not an output file
         write(iqclog,'(/,a)') 'Output file name had no type'
         stop
      endif

      inpfile = string(1:ilbl)//'.json  '
      ilbl=ilbl+5

!-----------------------------------------------------------------------
! Check if file exist
!-----------------------------------------------------------------------
      inquire(file=inpfile(1:ilbl),exist=linpfile)
      if (.not. linpfile) then
         write(iqclog,'(/,a)') 'ERROR. No BAGEL input file found at'
         write(iqclog,'(a)') inpfile
         stop
      endif

!-----------------------------------------------------------------------
! Open input file
!-----------------------------------------------------------------------
      open(iunit1,file=inpfile(1:ilbl),form='formatted',status='old',err=2000)

!-----------------------------------------------------------------------
! Initialise variable to read through input file
!-----------------------------------------------------------------------
      nsgra = 0
      nshess = 0
      linpgra=.false.
      linphess=.false.
      sublevel = 0
      gra_start = 0
      hes_start = 0
      na_start = 0
      gra_sec = .false.
      hes_sec = .false.
      na_sec = .false.
      target_found = .false.
      natype_found = .false.

!-----------------------------------------------------------------------
! Loop to read input file
!-----------------------------------------------------------------------
      do
         read(iunit1,'(a)',end=910,err=2000) string
         do
            !iteratively split string at comma
            cix = index(string,',')           
            if (cix==0) then
               substring = string
            else 
               substring = string(1:cix)
               string = string(cix + 1:)
            endif
            

            sublevel = checklevel(substring, sublevel)

            if (stringcontains('"angstrom"',substring) .and. stringcontains('true',substring)) then ! Check units
               write(iqclog,'(/,a)') 'ERROR. Bagel calculation must use bohr not angstrom.'
               stop
            endif


            if (stringcontains('"force"',substring)) then ! Gradient command found
               gra_sec = .true.
               gra_start = sublevel
            else if (stringcontains('"hessian"',substring)) then ! Hessian command found
               hes_sec = .true.
               hes_start = sublevel
            else if (stringcontains('"nacme"',substring)) then ! NA coupling command found
               na_sec = .true.
               na_start = sublevel
            endif

            ilbl=index(substring,'"target"')+8 ! Check whether substring contains 'target'

            if (ilbl > 8) then ! Target keyword found
               target_found = .true.
               jlbl=index(substring(ilbl:),":")+1
               klbl=index(substring(ilbl+jlbl:),",")-1

               if (gra_sec) then ! Read gradient target
                  nsgra=nsgra+1
                  read(substring(ilbl+jlbl:ilbl+jlbl+klbl),*,err=2004) mapgra(nsgra)
                  mapgra(nsgra)=mapgra(nsgra)+1 !add 1 as bagel start index from 0
                  linpgra=.true.
               else if (hes_sec) then ! Read hessian target
                  nshess=nshess+1
                  read(substring(ilbl+jlbl:ilbl+jlbl+klbl),*,err=2005) maphess(nshess)       
                  maphess(nshess)=maphess(nshess)+1 !add 1 as bagel start index from 0
                  linphess=.true.
               endif                  
            endif

            if (stringcontains('"nacmtype"',substring) .and. na_sec) then
               natype_found = .true.
               if (index(substring, '"noweight"') == 0) then
                  write(iqclog,'(/,a)') 'ERROR. Bagel Nact requested is not the "noweight" type.'
                  stop
               endif
            endif

            if (gra_sec .and. sublevel < gra_start) then
               if (.not. target_found) then
                  write(iqclog,'(/,a)') 'Bagel json input file had no "target" for "force"'
                  stop
               endif
               gra_sec = .false.
               target_found = .false.
            else if (hes_sec .and. sublevel < hes_start) then
               if (.not. target_found) then
                  write(iqclog,'(/,a)') 'Bagel json input file had no "target" for "hessian"'
                  stop
               endif
               hes_sec = .false.
               target_found = .false.
            else if (na_sec .and. sublevel < na_start) then
               if (.not. natype_found) then
                  write(iqclog,'(/,a)') 'ERROR. "noweight" type was not requested for Bagel Nact.'
                  stop
               endif
               na_sec = .false.
               natype_found = .false.  
            endif

            if (substring == string) then
               exit
            endif

         enddo
      enddo

 910  continue ! End of input file
!-----------------------------------------------------------------------
! close input file
!-----------------------------------------------------------------------
      close(iunit1)

      if (qcread(4) .and. .not. linpgra) then
         write(iqclog,'(/,a)') "ERROR. 1st derivative requested but not found in BAGEL input file"
         write(iqclog,'(a)') inpfile
         stop 
      elseif (qcread(5) .and. .not. linphess) then
         write(iqclog,'(/,a)') "ERROR. 2nd derivative requested but not found in BAGEL input file"
         write(iqclog,'(a)') inpfile
         stop 
      endif
      return

      2000 write(iqclog,'(/,a)') 'ERROR opening BAGEL input file'
           write(iqclog,'(a)') inpfile
           stop 
      2004 write(iqclog,'(/,a)') 'ERROR reading BAGEL input file (1st derivative)'
           stop
      2005 write(iqclog,'(/,a)') 'ERROR reading BAGEL input file (2nd derivative)'
           stop
      end subroutine rdbagelinp

!#######################################################################

!-----------------------------------------------------------------------
! Function to count open and close curly braces to track which sublevel of 
! a json dictionary you are in after reading a line of json input.
!-----------------------------------------------------------------------

      integer(long) function checklevel(str,sublevel)

      character(len=*), intent(in) :: str ! Line of input that has been read.
      integer(long)              :: sublevel ! Dictionary sublevel before the current line was read.
      integer(long)              :: nopen, nclose
      integer(long)              :: ix, i

      
      nopen = 0
      nclose = 0

      ix = 1
      do
         i = index(str(ix:),'{')
         if (i==0) exit
         nopen = nopen + 1
         ix = ix + i
      end do

      ix = 1
      do
         i = index(str(ix:),'}')
         if (i==0) exit
         nclose = nclose + 1
         ix = ix + i
      end do

         
      checklevel = sublevel + nopen - nclose

   end function checklevel

!#######################################################################

!-----------------------------------------------------------------------
! Function to separate the Atom symbol from the Label (e.g. C2 -> C)
!-----------------------------------------------------------------------

      integer(long) function getiatsym_readqc(atomlab,iqclog)

      implicit none

      character(len=*), intent(in) :: atomlab
      integer(long), intent(in)    :: iqclog

      if (atomlab(1:2) .eq. 'Cr'.or.atomlab(1:2) .eq. 'cr') then
         getiatsym_readqc = 24
      else if (atomlab(1:2) .eq. 'Cl'.or. atomlab(1:2) .eq. 'cl' &
            .or. atomlab(1:2) .eq. 'CL') then
         getiatsym_readqc = 17
      else if (atomlab(1:1) .eq. 'H'.or.atomlab(1:1) .eq. 'h') then
         getiatsym_readqc = 1
      elseif (atomlab(1:1) .eq. 'C'.or.atomlab(1:1) .eq. 'c') then
         getiatsym_readqc = 6
      elseif (atomlab(1:1) .eq. 'N'.or.atomlab(1:1) .eq. 'n') then
         getiatsym_readqc = 7
      elseif (atomlab(1:1) .eq. 'O'.or.atomlab(1:1) .eq. 'o') then
         getiatsym_readqc = 8
      elseif (atomlab(1:1) .eq. 'F'.or.atomlab(1:1) .eq. 'f') then
         getiatsym_readqc = 9
      elseif (atomlab(1:1) .eq. 'S'.or.atomlab(1:1) .eq. 's') then
         getiatsym_readqc = 16
      elseif (atomlab(1:1) .eq. 'I'.or.atomlab(1:1) .eq. 'i') then
         getiatsym_readqc = 53
      else
         getiatsym_readqc=-999
         write(iqclog,'(2a)') 'Unknown label in getiatsym_readqc :',atomlab
      endif

      return
      end function getiatsym_readqc

!#######################################################################
!
! GETMASS
! define relative atomic masses for a given atomic number
! Value returned in amu
!
! iau = 0 mass in au
!     = 1 mass in amu
! 
! If any atoms are added to list, also add to getisym and getatnam below
! and getinam in readqc.f90
!#######################################################################

      subroutine getmass(mass,iatnum,iau,ilog)

      integer(long), intent(in)     :: iatnum,ilog,iau
      real(dop), intent(out)        :: mass
         
      if (iatnum .eq. 1) then
         mass=1.0079_dop
      elseif (iatnum .eq. 2) then
         mass=4.0026_dop
      elseif (iatnum .eq. 3) then
         mass=6.941_dop
      elseif (iatnum .eq. 4) then
         mass=9.0122_dop
      elseif (iatnum .eq. 5) then
         mass=10.811_dop
      elseif (iatnum .eq. 6) then
         mass=12.0107_dop
      elseif (iatnum .eq. 7) then
         mass=14.0067_dop
      elseif (iatnum .eq. 8) then
         mass=15.9994_dop
      elseif (iatnum .eq. 9) then
         mass=18.9984_dop
      elseif (iatnum .eq. 10) then
         mass=20.1797_dop
      elseif (iatnum .eq. 11) then
         mass=22.9897_dop
      elseif (iatnum .eq. 12) then
         mass=24.305_dop
      elseif (iatnum .eq. 13) then
         mass=26.9815_dop
      elseif (iatnum .eq. 14) then
         mass=28.0855_dop
      elseif (iatnum .eq. 15) then
         mass=30.9738_dop
      elseif (iatnum .eq. 16) then
         mass=32.065_dop
      elseif (iatnum .eq. 17) then
         mass=35.453_dop
      elseif (iatnum .eq. 18) then
         mass=39.948_dop
      elseif (iatnum .eq. 24) then
         mass=51.9961_dop
      elseif (iatnum .eq. 35) then
         mass=79.904_dop
      elseif (iatnum .eq. 53) then
         mass=126.9045_dop
      else if (iatnum .eq. 74) then
         mass=183.84_dop
      else if (iatnum .eq. 80) then
         mass=200.59_dop
      else
         write(ilog,'(/,a)') 'Routine Getmass (readqc.f90)'
         write(ilog,'(a,i5)') 'Mass not implemented for atomic symbol:',iatnum
         stop
      endif

!-----------------------------------------------------------------------
! convert to au
!-----------------------------------------------------------------------
      if (iau .eq. 0) mass = mass*atomicmassunit

      return
      end subroutine getmass

!#######################################################################
!
! GETINAM
!
! For a given atom name, assign atomic number
! (convert string to integer)
!
! If any atoms are added to list, also add to getmass, getatnam and
! getisym (above).
!
!#######################################################################

      subroutine getinam(atnam,iatnam,iatiso,ilog)

      integer(long), intent(in)     :: ilog
      integer(long), intent(out)    :: iatnam,iatiso
      character(len=*), intent(in)  :: atnam

      iatiso=0
! ensure name is lower case 
      call keylowc(atnam)
!
      if (atnam(1:2) .eq. 'h ' .or. atnam(1:2) .eq. '1 ') then
         iatnam=1
      elseif ((atnam(1:8) .eq. 'h(iso=2)') .or. &
                       (atnam(1:2) .eq. 'd ')) then
         iatnam=1
         iatiso=2
      elseif (atnam(1:3) .eq. 'he ' .or. atnam(1:2) .eq. '2 ') then
         iatnam=2
      elseif (atnam(1:3) .eq. 'li ' .or. atnam(1:2) .eq. '3 ') then
         iatnam=3
      elseif (atnam(1:3) .eq. 'be ' .or. atnam(1:2) .eq. '4 ') then
         iatnam=4
      elseif (atnam(1:2) .eq. 'b ' .or. atnam(1:2) .eq. '5 ') then
         iatnam=5
      elseif (atnam(1:2) .eq. 'c ' .or. atnam(1:2) .eq. '6 ') then
         iatnam=6
      elseif (atnam(1:2) .eq. 'n ' .or. atnam(1:2) .eq. '7 ') then
         iatnam=7
      elseif (atnam(1:2) .eq. 'o ' .or. atnam(1:2) .eq. '8 ') then
         iatnam=8
      elseif (atnam(1:2) .eq. 'f ' .or. atnam(1:2) .eq. '9 ') then
         iatnam=9
      elseif (atnam(1:3) .eq. 'ne ' .or. atnam(1:3) .eq. '10 ') then
         iatnam=10
      elseif (atnam(1:3) .eq. 'na ' .or. atnam(1:3) .eq. '11 ') then
         iatnam=11
      elseif (atnam(1:3) .eq. 'mg ' .or. atnam(1:3) .eq. '12 ') then
         iatnam=12
      elseif (atnam(1:3) .eq. 'al ' .or. atnam(1:3) .eq. '13 ') then
         iatnam=13
      elseif (atnam(1:3) .eq. 'si ' .or. atnam(1:3) .eq. '14 ') then
         iatnam=14
      elseif (atnam(1:2) .eq. 'p ' .or. atnam(1:3) .eq. '15 ') then
         iatnam=15
      elseif (atnam(1:2) .eq. 's ' .or. atnam(1:3) .eq. '16 ') then
         iatnam=16
      elseif (atnam(1:3) .eq. 'cl ' .or. atnam(1:3) .eq. '17 ') then
         iatnam=17
      elseif (atnam(1:3) .eq. 'ar ' .or. atnam(1:3) .eq. '18 ') then
         iatnam=18
      elseif (atnam(1:3) .eq. 'cr ' .or. atnam(1:3) .eq. '24 ') then
         iatnam=24
      elseif (atnam(1:3) .eq. 'br ' .or. atnam(1:3) .eq. '35 ') then
         iatnam=35
      elseif (atnam(1:2) .eq. 'i ' .or. atnam(1:3) .eq. '53 ') then
         iatnam=53
      elseif (atnam(1:2) .eq. 'w ' .or. atnam(1:3) .eq. '74 ') then
         iatnam=74
      elseif (atnam(1:2) .eq. 'hg' .or. atnam(1:3) .eq. '80 ') then
         iatnam=80
      else
         write(ilog,'(/,a)') 'Routine Getinam (readqc.f90)'
         write(ilog,'(a,a5)') &
             'Atom number not implemented for atomic name / no.:',atnam
         stop
      endif
      
      return
      end subroutine getinam
      
   !#######################################################################
   !
   ! GETATNAM
   !
   ! Give the atom name for a given atomic number. Atomic number
   ! is passed as a string as it can be read as the atom name in
   ! QC input / output files
   !
   ! If any atoms are added to list, also add to getmass (above) and 
   ! getisym, and getinam (below).
   !
   !#######################################################################
   
         subroutine getatnam(atnam,ilog)
   
         integer(long), intent(in)         :: ilog
         character(len=*), intent(inout)  :: atnam
   
         if(atnam(1:3) .eq. '1  ')then
            atnam(1:3) = 'H  '
         else if(atnam(1:3) .eq. '2  ')then
            atnam(1:3) = 'He '
         elseif (atnam(1:3) .eq. '3  ') then
            atnam(1:3) = 'Li '
         elseif (atnam(1:3) .eq. '4  ') then
            atnam(1:3) = 'Be '
         elseif (atnam(1:3) .eq. '5  ') then
            atnam(1:3) = 'B  '
         elseif (atnam(1:3) .eq. '6  ') then
            atnam(1:3) = 'C  '
         elseif (atnam(1:3) .eq. '7  ') then
            atnam(1:3) = 'N  '
         elseif (atnam(1:3) .eq. '8  ') then
            atnam(1:3) = 'O  '
         elseif (atnam(1:3) .eq. '9  ') then
            atnam(1:3) = 'F  '
         elseif (atnam(1:3) .eq. '10 ') then
            atnam(1:3) = 'Ne '
         elseif (atnam(1:3) .eq. '11 ') then
            atnam(1:3) = 'Na '
         elseif (atnam(1:3) .eq. '12 ') then
            atnam(1:3) = 'Mg '
         elseif (atnam(1:3) .eq. '13 ') then
            atnam(1:3) = 'Al '
         elseif (atnam(1:3) .eq. '14 ') then
            atnam(1:3) = 'Si '
         elseif (atnam(1:3) .eq. '15 ') then
            atnam(1:3) = 'P  '
         elseif (atnam(1:3) .eq. '16 ') then
            atnam(1:3) = 'S  '
         elseif (atnam(1:3) .eq. '17 ') then
            atnam(1:3) = 'Cl '
         elseif (atnam(1:3) .eq. '18 ') then
            atnam(1:3) = 'Ar '
         elseif (atnam(1:3) .eq. '24 ') then
            atnam(1:3) = 'Cr '
         elseif (atnam(1:3) .eq. '35 ') then
            atnam(1:3) = 'Br '
         elseif (atnam(1:3) .eq. '53 ') then
            atnam(1:3) = 'I  '
         elseif (atnam(1:3) .eq. '74 ') then
            atnam(1:3) = 'W  '
         elseif (atnam(1:3) .eq. '80 ') then
            atnam(1:3) = 'Hg '
         endif
   
         return
         end subroutine getatnam
   
   !#######################################################################
   !
   ! GETISYM
   ! get atomic number from relative atomic mass (in amu)
   ! 
   ! If any atoms are added to list, also add to getmass and getatnam 
   ! (above), and getinam (below).
   !
   !#######################################################################
   
         subroutine getisym(iatnum,mass,ilog)
   
         integer(long), intent(in)     :: ilog
         integer(long), intent(out)    :: iatnum
         real(dop), intent(in)         :: mass
   
         if (nint(mass) .eq. 1) then
            iatnum=1
         else if (nint(mass) .eq. 4) then
            iatnum=2
         else if (nint(mass) .eq. 7) then
            iatnum=3
         else if (nint(mass) .eq. 9) then
            iatnum=4
         else if (nint(mass) .eq. 11 ) then
            iatnum=5
         else if (nint(mass) .eq. 12) then
            iatnum=6
         else if (nint(mass) .eq. 14) then
            iatnum=7
         else if (nint(mass) .eq. 16) then
            iatnum=8
         else if (nint(mass) .eq. 19) then
            iatnum=9
         else if (nint(mass) .eq. 20) then
            iatnum=10
         else if (nint(mass) .eq. 23) then
            iatnum=11
         else if (nint(mass) .eq. 24) then
            iatnum=12
         else if (nint(mass) .eq. 27 ) then
            iatnum=13
         else if (nint(mass) .eq. 28) then
            iatnum=14
         else if (nint(mass) .eq. 31) then
            iatnum=15
         else if (nint(mass) .eq. 32) then
            iatnum=16
         else if (nint(mass) .eq. 35 .or. nint(mass) .eq. 36) then
            iatnum=17
         else if (nint(mass) .eq. 40) then
            iatnum=18
         else if (nint(mass) .eq. 52) then
            iatnum=24
         else if (nint(mass) .eq. 80) then
            iatnum=35
         else if (nint(mass) .eq. 127) then
            iatnum=53
         else if (nint(mass) .eq. 184) then
            iatnum=74
         else if (nint(mass) .eq. 201) then
            iatnum=80
         else
            write(ilog,'(/,a)') 'Routine Getisym (readqc.f90)'
            write(ilog,'(a,f10.4)') 'Atomic symbol not implemented for mass:',mass
            stop
         endif
   
         return
         end subroutine getisym

!#######################################################################
!
! GENDER2
!
! Generates 2nd derivatives in cartesian coordinate from frequencies 
! and normal mode coordinates
!
! freq1: frequency input in Hartree
! mass: mass input in AU (~1822AU for 1AMU)
! der2: 2nd derivatives in Hartree/Bohr**2
!#######################################################################

      subroutine gender2(der2,amat,freq1,iatsym,iau,ncoo,&
                 maxdim,nmodes,ilog)

      implicit none

      integer(long), intent(in) :: ncoo,maxdim,nmodes,ilog,iau
      integer(long)             :: i,j,k
      integer(long), dimension(maxdim), intent(in)     :: iatsym
      real(dop), dimension(maxdim,maxdim), intent(in)  :: amat
      real(dop), dimension(maxdim), intent(in)         :: freq1
      real(dop), dimension(maxdim)                     :: mass
      real(dop), dimension(maxdim,maxdim), intent(out) :: der2
      real(dop)                                        :: fact,sign

!-----------------------------------------------------------------------
! Get atomic masses and units
!-----------------------------------------------------------------------
      do j=1,ncoo
         call getmass(mass(j),iatsym(j),iau,ilog)
      enddo
      if (iau .eq. 0) then
         fact = 1.0_dop
      else if (iau .eq. 1) then
         fact = dsqrt(atomicmassunit/uev)
         fact = fact/angstroem
         fact=fact*fact
      endif

!-----------------------------------------------------------------------
! 2nd derivatives in x
!-----------------------------------------------------------------------
      der2(:,:) = 0.0_dop
      do i=1,ncoo
         do j=1,ncoo
            do k=1,nmodes
               sign=1.0_dop
               if (freq1(k) < 0) sign=-1.0_dop
               der2(i,j)=der2(i,j) + sign*amat(i,k)*freq1(k)*freq1(k)*amat(j,k)
            enddo
            der2(i,j)=der2(i,j)*fact*sqrt(mass(i)*mass(j))
         enddo
      enddo

      return
      end subroutine gender2

!#######################################################################
! Routine to Gram-Schmidt orthonormalise normal modes.
!#######################################################################

      subroutine nmortho(matrix,maxdim,ncoo,nmodes,ilog)

      implicit none

      integer(long), intent(in) :: maxdim,ncoo,nmodes,ilog
      integer(long)             :: i,j,k,i1
      real(dop), dimension(maxdim,maxdim), intent(inout) :: matrix
      real(dop), dimension(maxdim,maxdim)                :: temp
      real(dop)                                          :: off,on,x

! normalise nm coordinates
      do i=1,nmodes
         x=0.0_dop
         do j=1,ncoo
            x=x+matrix(j,i)*matrix(j,i)
         enddo

         do j=1,ncoo
            matrix(j,i)=matrix(j,i)/sqrt(x)
         enddo
      enddo

! Form matrix of overlaps of the normal modes.
      do i=1,ncoo
         do j=1,ncoo
            temp(j,i)=0.0_dop
            do k=1,ncoo
               temp(j,i)=temp(j,i)+matrix(k,j)*matrix(k,i)
            enddo
         enddo
      enddo
! Get the totals of the diagonal and off-diagonal elements.
      off=0.0_dop
      on=0.0_dop
      do i=1,ncoo
         do j=1,ncoo
            if (i .ne. j) then
               off=off+temp(i,j)
            else
               on=on+temp(i,i)
            endif
         enddo
      enddo

! Get the mean of the diagonal terms, then output both sets.      
      on=on/dble(nmodes)
      write(ilog,'(/,a)') 'Before orthonormalisation'
      write(ilog,'(a,1pe20.12)') 'off diagonal sum = ',off
      write(ilog,'(a,1pe20.12)') 'on diagonal sum  = ',on
      
      do i=1,nmodes
         do i1=1,i-1
            x=0.0_dop
            do j=1,ncoo
               x=x+matrix(j,i)*matrix(j,i1)
            enddo
            do j=1,ncoo
               matrix(j,i)=matrix(j,i)-x*matrix(j,i1)
            enddo
         enddo
         x=0.0_dop
         do j=1,ncoo
            x=x+matrix(j,i)*matrix(j,i)
         enddo
         x=1.0_dop/sqrt(x)
         do j=1,ncoo
            matrix(j,i)=x*matrix(j,i)
         enddo
      enddo
    
      do i=1,ncoo
         do j=1,ncoo
            temp(j,i)=0.0_dop
            do k=1,ncoo
               temp(j,i)=temp(j,i)+matrix(k,j)*matrix(k,i)
            enddo
         enddo
      enddo
      off=0.0_dop
      on=0.0_dop
      do i=1,ncoo
         do j=1,ncoo
            if (i .ne. j) then
               off=off+temp(i,j)
            else
               on=on+temp(i,i)
            endif
         enddo
      enddo

      on=on/dble(nmodes)
      write(ilog,'(a)') 'After orthonormalisation'
      write(ilog,'(a,1pe20.12)') 'off diagonal sum = ',off
      write(ilog,'(a,1pe20.12)') 'on diagonal sum  = ',on
      write(ilog,*)

      return
      end subroutine nmortho

!#######################################################################

      subroutine nmorthof(matrix,maxdim,ncoo,ilog)

      implicit none

      integer(long), intent(in) :: maxdim,ncoo,ilog
      integer(long)             :: i,j,k,i1
      real(dop), dimension(maxdim,maxdim), intent(inout) :: matrix
      real(dop), dimension(maxdim,maxdim)                :: temp
      real(dop)                                          :: off,on,x

      do i=1,ncoo
         do j=1,ncoo
            temp(j,i)=0.0_dop
            do k=1,ncoo
               temp(j,i)=temp(j,i)+matrix(k,j)*matrix(k,i)
            enddo
         enddo
      enddo
      off=0.0_dop
      on=0.0_dop
      do i=1,ncoo
         do j=1,ncoo
            if (i .ne. j) then
               off=off+temp(i,j)
            else
               on=on+temp(i,i)
            endif
         enddo
      enddo
      
      on=on/ncoo
      write(ilog,'(/,a)') 'Before orthonormalisation'
      write(ilog,'(a,1pe20.12)') 'off diagonal sum = ',off
      write(ilog,'(a,1pe20.12)') 'on diagonal sum  = ',on
      
      do i=1,ncoo
         do i1=1,i-1
            x=0.0_dop
            do j=1,ncoo
               x=x+matrix(j,i)*matrix(j,i1)
            enddo
            do j=1,ncoo
               matrix(j,i)=matrix(j,i)-x*matrix(j,i1)
            enddo
         enddo
         x=0.0_dop
         do j=1,ncoo
            x=x+matrix(j,i)*matrix(j,i)
         enddo
         x=1.0_dop/sqrt(x)
         do j=1,ncoo
            matrix(j,i)=x*matrix(j,i)
         enddo
      enddo
    
      do i=1,ncoo
         do j=1,ncoo
            temp(j,i)=0.0_dop
            do k=1,ncoo
               temp(j,i)=temp(j,i)+matrix(k,j)*matrix(k,i)
            enddo
         enddo
      enddo
      off=0.0_dop
      on=0.0_dop
      do i=1,ncoo
         do j=1,ncoo
            if (i .ne. j) then
               off=off+temp(i,j)
            else
               on=on+temp(i,i)
            endif
         enddo
      enddo

      on=on/ncoo
      write(ilog,'(a)') 'After orthonormalisation'
      write(ilog,'(a,1pe20.12)') 'off diagonal sum = ',off
      write(ilog,'(a,1pe20.12)') 'on diagonal sum  = ',on
      write(ilog,*)

      return
    end subroutine nmorthof

!#######################################################################

      subroutine bubble (vector, vecdim)

      implicit none

      integer(long), intent(in) :: vecdim
      integer(long)             :: i,j
      real(dop), dimension(vecdim), intent(inout)        ::  vector
      real(dop)                                          ::  swap

! --- SORT THE DATA ---

      do i = 1,vecdim-1
         do j = i+1,vecdim
            if (vector(i) .gt. vector(j)) then
               swap = vector(i)
               vector(i) = vector(j)
               vector(j) = swap
            endif
         enddo
      enddo

      return
      end subroutine bubble

! ----------------------------------------------------------------------
!                           SUBROUTINE IBUBBLE_UP
! 
! PURPOSE:
!     Bubble-sort algorithm for double precision numbers.              
! INPUT PARAMETERS:
!     vector: real vector containing the data to be sorted.            
!     ivector: integer vector 
!     vecdim: length of vectors 
! 
! OUTPUT PARAMETERS:
!     vector: contains the sorted data (from smallest to largest).     
!     ivector: contains the indixes of the original vector
! (See also "bsort" in propwf/getdavmat.F)
! ----------------------------------------------------------------------

      subroutine ibubble_up (vector, ivector, vecdim)

      implicit none

      integer(long)             :: i,j
      integer(long), intent(in) :: vecdim
      integer(long)                                 :: iswap   
      integer(long), dimension(vecdim), intent(out) :: ivector
      real(dop)                                     ::  swap
      real(dop), dimension(vecdim), intent(out)     :: vector

      do i = 1,vecdim
         ivector(i)=i 
      enddo

! --- SORT THE DATA ---

      do i = 1,vecdim-1
         do j = i+1,vecdim
            if (vector(i) .gt. vector(j)) then
               swap = vector(i)
               vector(i) = vector(j)
               vector(j) = swap

               iswap = ivector(i)
               ivector(i) = ivector(j)
               ivector(j) = iswap
            endif
         enddo
      enddo

      return
      end subroutine ibubble_up 

!#######################################################################

      logical(kind=4) function stringcontains (searchterm, str)

         implicit none  

         character(len=*), intent(in) :: searchterm,str

         stringcontains = (index(str,trim(searchterm)) .ne. 0)
      end function stringcontains

!####################################################################### 

      integer(long) function numberindex (instr)

         implicit none

         character(len=*), intent(in) :: instr
         character(len=1)             :: teststr

         integer(long) :: i,lowind,tempind

         lowind=0
         do i=0,9
            write(teststr,'(i1)') i
            tempind=index(instr,teststr)
            if (lowind .eq. 0) then
               lowind=tempind
            elseif (tempind .ne. 0 .and. tempind .le. lowind) then
               lowind=tempind
            endif
         enddo
         numberindex=lowind
      
      end function numberindex

      end module readqc
