!>#######################################################################
!> This routine interprets the options of the interface to the
!> external quantum chemistry program (V = qchem{...} or V = external{...}
!> in .op file) for direct dynamics calculations.
!> It is called from defmuld, which is called from assignhlab.
!> In the latter, all relevant information passed in arguments comes from
!> and goes to common blocks that are defined in operdef.inc.
!>
!> Information is extracted from label (<-- hoplab) and pesopts.
!> 1. A number identifying the program is assigned to ifunc (--> hopilab)
!> 2. The external script to run the quantum chemistry program is
!>    assigned to ddextcmdddga
!> 3. Other options are assigned to hopipar, lddwrdb, ddmindb,
!>    dde0,lfchk,fchkcmd (see below)
!>    Warning: hopipar(3)-hopipar(6) must be left blank as used to pass
!>    information on function and state of interest
!> 4. A flag for V = qchem is assigned to ifile (--> hopfile).
!>
!> To add a new quantum chemistry program, add a new possible value to
!> the keyword qcprog.
!> To add a new quantum chemistry method, add a new possible value to
!> the keyword method.
!>
!> ifile=205 for potentials
!> ifile=208 for GAPs
!>
!>#######################################################################

      module funcqchemmod

use decimal, only: dop, long
use global
use logdat
use timing
use channels
use griddatmod
use openmpmod
use dirdyn
use operdef
use dd_db
use datenmod, only: tinit
use op1lib, only: zerovxd,zerovxz
use readqc
#ifdef MPI
use mpidata
#endif

use omp_lib

      implicit none

      private

      public :: defqchem,vqchem,derqchem,qcheminfo,lhaqchem,wrmolpro_geom,&
                rdmolpro_ao,ddvmcgpestest,rdmolpro_ao_1,rdmolpro_ci,&
                wrmolpro_orbfile,molpro_moover,molpro_wfover,citrans

!-----------------------------------------------------------------------
! Derived type for holding the expansion of a CSF into Slater 
! determinants. 
!-----------------------------------------------------------------------
      type csfdet
         character(30) :: config
         real(dop) :: coeff
         type(csfdet), pointer :: prev,next
      endtype csfdet

! Array used in sorting CSF decompositions. 
      integer(long), parameter :: L(0:43) = (/ 1, 1, 3, 5, &
           9, 15, 25, 41, 67, 109, 177, 287, 465, 753, &
           1219, 1973, 3193, 5167, 8361, &
           13529, 21891, 35421, 57313, 92735, 150049, &
           242785, 392835, 635621, 1028457, 1664079, 2692537, &
           4356617, 7049155, 11405773, 18454929, 29860703, &
           48315633, 78176337, 126491971, 204668309, 331160281, &
           535828591, 866988873, 1402817465 /)

      contains


      subroutine defqchem(ifunc,label,ifile,pesopts,zpesopts,hopipar,ierr)

      implicit none

      integer(long)                                     :: ilbl,jlbl,klbl,np,iopt
      integer(long), intent(out)                        :: ifunc,ifile,ierr
      integer(long), dimension(*), intent(out)          :: hopipar
      integer(long), dimension(2*maxhpar,*), intent(in) :: zpesopts 
      character(len=*), intent(in)                      :: label
      character(len=*), dimension(*), intent(in)        :: pesopts
      character(len=64)                                 :: option

      ifunc=0
      ifile=0
      ierr=0

      ilbl=index(label,'{')-1
      if (ilbl .lt. 0) ilbl=index(label,' ')-1

      if (label(1:5) .ne. 'qchem') return

! ignore these labels as they are dealt with in funcqch1
      if (label(1:8).eq.'qchemdip' .or. &
          label(1:8) .eq. 'qchemtrj') return
!
! if option passed from .op file as option list qchem{np}, 
! extract option list number and read options
!
      hopipar(1)=-999
      ilbl=index(label,'{')+1
      jlbl=index(label,'}')-1
      if (ilbl .gt. 1) then
         read(label(ilbl:jlbl),*) np
         hopipar(1)=np
!
! extract quantum chemistry program from pesopts string (1st option)
!
         ilbl=zpesopts(1,np)
         read(pesopts(np)(1:ilbl),'(a)',err=1000,iostat=ierr) qcprog
         call keylowc(qcprog)
!    
! extract command to run external quantum chemistry program (2nd option)
!        
         ilbl=zpesopts(1,np)+1
         jlbl=zpesopts(2,np)
         read(pesopts(np)(ilbl:jlbl),'(a)',err=1000,iostat=ierr)  &
             ddextcmd
!
! extract potential energy origin in hartree (3rd option)
!
         ilbl=zpesopts(2,np)+1
         jlbl=zpesopts(3,np)
         read(pesopts(np)(ilbl:jlbl),*,err=1000,iostat=ierr) dde0
!     
! extract quantum chemistry method (4th option)
!     
         ilbl=zpesopts(3,np)+1
         jlbl=zpesopts(4,np)
         read(pesopts(np)(ilbl:jlbl),'(a)',err=1000,iostat=ierr)  &
             qcmethod
         call keylowc(qcmethod)
!
! if specified, extract non-compulsory options:
! - highest electronic state to nroot
! - command to format GAUSSIAN chk files
! - database information
!
         iopt = 0
  100    continue
         ilbl=zpesopts(4+iopt,np)+1
         jlbl=zpesopts(5+iopt,np)
         klbl=jlbl-ilbl+1
         if (jlbl .gt. 0) then
            read(pesopts(np)(ilbl:jlbl),'(a)',err=1000,iostat=ierr)  &
                 option
            if (option(1:6) .eq. 'NRoot:') then
               if (klbl .gt. 6) then
                  read(option(7:klbl),*,err=1000,iostat=ierr) nroot
               else
                  routine = 'Defqchem'
                  message = 'You must give an integer after NRoot'
                  call errormsg
               endif
            else if (option(1:5) .eq. 'fchk:') then
               lfchk = .true.
               if (klbl .gt. 5) then
                  read(option(6:klbl),'(a)',err=1000,iostat=ierr)  &
                     fchkcmd
               else
                  routine = 'Defqchem'
                  message = 'You must give a path after fchk'
                  call errormsg
               endif
            else if (option(1:2) .eq. 'DB') then
               if (option(3:6) .eq. 'rdwr') then
                  lddwrdb = .true.
                  ddmindb = 0.01_dop
               else if (option(3:4) .eq. 'wr') then
                  lddwrdb = .true.
               else
                  routine = 'Defqchem'
                  message = 'For DB field, '// &
                         'choose among DBrdwr/wr or leave empty'
                  call errormsg
               endif
               ilbl=index(option,':')
               if (ilbl .gt. 0) then
                  if (lddwrdb) then
                     if (klbl .gt. ilbl) then
                        read(option(ilbl+1:klbl),*,err=1000,iostat=ierr) &
                            ddmindb
                     else
                        routine = 'Defqchem'
                        message =  &
                               'Give a value after DBrdwr: '// &
                               'or remove colon (default is 0.01)'
                        call errormsg
                     endif
                  else
                     routine = 'Defqchem'
                     message =  &
                   'Do not give a value after DBwr '
                     call errormsg
                  endif
               endif
            else
               routine = 'Defqchem'
               message = 'Further options not implemented yet'
               call errormsg
            endif
            iopt=iopt+1
            go to 100
         endif

! end of loop over options
      endif

!-----------------------------------------------------------------------
! assign ifunc by quantum chemistry program
!            1  Gaussian
!            2  QoMMMa
!            3  Molpro
!            4  Q-Chem
!            5  Molcas
!            6  ORCA
!            7 External
!            8  Bagel
!          101-104 Dipole operators (assigned above)
!-----------------------------------------------------------------------
      ilbl=index(qcprog,' ')-1
      if (ilbl .le. 0) then
         routine = 'Defqchem'
         message =  &
       'Quantum chemistry program must be defined using qcprogram'
         call errormsg
      endif
      if (qcprog(1:ilbl) .eq. 'gaussian') then
         ifunc = 1
      else if (qcprog(1:ilbl) .eq. 'qommma') then
         ifunc = 2
      else if (qcprog(1:ilbl) .eq. 'molpro') then
         ifunc = 3
      else if (qcprog(1:ilbl) .eq. 'q-chem') then
         ifunc = 4
      else if (qcprog(1:ilbl) .eq. 'molcas') then
         ifunc = 5
      else if (qcprog(1:ilbl) .eq. 'orca') then
         ifunc = 6
      else if (qcprog(1:ilbl) .eq. 'external') then !adding external program when calculating potential, first and second derivatives from external programs 
         ifunc = 7    
      else if (qcprog(1:ilbl) .eq. 'bagel') then
         ifunc = 8
      else if (qcprog(1:ilbl) .eq. 'none') then
         ifunc = 9
      else
         routine = 'Defqchem'
         message = 'Quantum chemistry program not implemented yet'
         call errormsg
      endif

!-----------------------------------------------------------------------
! If running Q-Chem, deal with the scratch folder title
!-----------------------------------------------------------------------
     !ilbl=index(qcscr,' ')-1
     ! if (ifunc == 4 .and. ilbl .le. 0) then
     !    routine = 'Defqchem'
     !    message =  &
     !  'Scratch folder name must be defined using qcscratch if running Q-Chem'
     !    call errormsg
     ! endif

!-----------------------------------------------------------------------
! Assign options
!
! hopipar(2) = flag for diabatisation by QChem
! hopipar(7) quantum chemistry method. See utils/readqc.f90 for list
!
! hopipar(10) continuum solvent method
!             1 PCM
!
! if dddiab = 0 : adiabatic picture
!                 adia PES, adia WP - no NAC
!                 (Born-Oppenheimer approximation)
!             1 : regularised diabatic picture
!                 adia PES,  dia WP - adia --> dia PES
!                 (GAW scheme)
!             2 : local diabatic picture
!                 dia PES, adia WP - dia  --> adia WP
!                 (BL scheme)
!             4 : global diabatic picture
!                 dia PES,  dia WP - global gauge (GWR scheme)
!             5 : projection operator scheme (Robertson)
!             6 : Procrustes diabatisation scheme (GWR/SH)
!-----------------------------------------------------------------------
      if (qcmethod .eq. 'hf ' .or. qcmethod .eq. 'rhf ') then
         hopipar(7) = 1
         nroot = 1
         dddiab = 0

      else if (qcmethod .eq. 'cas') then
         if (nroot .le. 0) nroot = nddstate

         hopipar(7) = 12
         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      else if (qcmethod .eq. 'cas/mrci') then
         hopipar(7) = 15
         if (nroot .le. 0) nroot = nddstate

         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      else if (qcmethod .eq. 'dft') then
         hopipar(7) = 19
         nroot = 1
         dddiab = 0

      else if (qcmethod .eq. 'qommma') then
         hopipar(7) = 29
         nroot = 1
         dddiab = 0

      elseif (qcmethod .eq. 'uhf ') then
         hopipar(7) = 5
         nroot = 1
         dddiab = 0

      else if (qcmethod .eq. 'tddft') then
         hopipar(7) = 33
         nroot = nddstate
         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      elseif (qcmethod .eq. 'rohf ') then
         hopipar(7) = 6
         nroot = 1
         dddiab = 0

      elseif (qcmethod .eq. 'ms-cas' .or. qcmethod .eq. 'xms-cas') then
         hopipar(7) = 27
         if (nroot .le. 0) nroot = nddstate

         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      elseif (qcmethod .eq. 'ehrenfest ') then
         hopipar(7) = 30
         if (nroot .le. 0) nroot = nddstate
         dddiab = 0

      elseif (qcmethod .eq. 'ext') then
         hopipar(7) = 31
         nroot = nddstate   

      else if (qcmethod .eq. 'cas/pt2') then
         hopipar(7) = 32
         if (nroot .le. 0) nroot = nddstate

         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      else if (qcmethod .eq. 'cas/xdw') then
         hopipar(7) = 37
         if (nroot .le. 0) nroot = nddstate

         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      else if (qcmethod .eq. 'eom-ccsd-ee') then
         hopipar(7) = 17
         nroot = nddstate
         if (nroot .eq. 1) then
            dddiab = 0
         else
            if (dddiab .eq. -999) dddiab = 4
         endif

      else if (qcmethod .eq. 'none') then
         if (nroot .le. 0) nroot = nddstate
         dddiab = 0
         continue
      
      else
         routine = 'Defqchem'
         message =  &
           'Quantum chemistry method not implemented yet: '//qcmethod
         call errormsg
      endif

!      hopipar(10) = 0
!      if(lsolvent)then
!         if (solmethod .eq. 'pcm')then
!            hopipar(10) = 1
!         else
!            routine = 'Defqchem'
!            message =  &
!                 'Continuum solvent method not implemented yet: '//solmethod
!            call errormsg
!         endif
!      endif

!-----------------------------------------------------------------------
! Make sure ldiabdb flag set for diabatic representations
!-----------------------------------------------------------------------
       if (dddiab .ge. 2) ldiabdb = .true.

!-----------------------------------------------------------------------
! Check whether the method has been implemented for the program
!-----------------------------------------------------------------------
       if (ifunc .eq. 1) then
! gaussian
          if (hopipar(7) .ne. 1 .and. &
              hopipar(7) .ne. 5 .and. &
              hopipar(7) .ne. 6 .and. &
              hopipar(7) .ne. 12 .and. &
              hopipar(7) .ne. 19 .and. &
              hopipar(7) .ne. 30 .and. &
              hopipar(7) .ne. 33) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Gaussian '
             call errormsg
          endif

       else if (ifunc .eq. 2) then
! qommma
          if (hopipar(7) .ne. 29) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'Method not yet implemented for Qommma '
             call errormsg
          endif

       else if (ifunc .eq. 3) then
! molpro
          if (hopipar(7) .ne. 1 .and. &
              hopipar(7) .ne. 12 .and. &
              hopipar(7) .ne. 15 .and. &
              hopipar(7) .ne. 32) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Molpro '
             call errormsg
          endif

       else if (ifunc .eq. 4) then
! qchem
          if (hopipar(7) .ne. 1 .and. &
              hopipar(7) .ne. 17 .and. &
              hopipar(7) .ne. 19 .and. &
              hopipar(7) .ne. 33) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Qchem '
             call errormsg
          endif

       else if (ifunc .eq. 5) then
! molcas
         ! if (hopipar(7) .ne. 1 .and. & !WIP
              !hopipar(7) .ne. 5 .and. & !WIP
              !hopipar(7) .ne. 6 .and. & !WIP
          if( hopipar(7) .ne. 12 .and. &
              hopipar(7) .ne. 13 .and. &
              hopipar(7) .ne. 27 .and. &
              hopipar(7) .ne. 28 .and. &
              hopipar(7) .ne. 32 .and. &
              hopipar(7) .ne. 37) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Molcas '
             call errormsg
          endif

       else if (ifunc .eq. 6) then
! orca
          if (hopipar(7) .ne. 1 .and. &
              hopipar(7) .ne. 19 .and. &
              hopipar(7) .ne. 33) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Orca '
             call errormsg
          endif

       else if (ifunc .eq. 7) then
! External
          if (hopipar(7) .ne. 10) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'Error in Method for External '
             call errormsg
          endif

       else if (ifunc .eq. 8) then
! Bagel
          if (hopipar(7) .ne. 12 .and. &
              hopipar(7) .ne. 27) then
             routine = 'Defqchem (funcqchem.f90)'
             message = 'QC Method not yet implemented for Bagel '
             call errormsg
          endif

       endif

!-----------------------------------------------------------------------
! check compatibility of options
!-----------------------------------------------------------------------
      if (lfchk .and. (ifunc .ne. 1)) then
         routine = 'Defqchem'
         message = 'Specify fchk for a GAUSSIAN calculation only'
         call errormsg
      endif

      if (hopipar(7) .eq. 1 .or. hopipar(7) .eq. 5 .or. hopipar(7) .eq. 33) then
         if (lsacas) then
            routine = 'Defqchem'
            message = &
           'Hartree-Fock cannot be used with SA-CAS option '
            call errormsg
         endif
      endif

      if (nroot .lt. nddstate) then
         routine = 'Defqchem'
         message = 'The value given to NRoot must be at least '// &
                      'equal to the number of states'
         call errormsg
      endif

!      if(lsolvent .and. (hopipar(7) .ne. 12 .and. ifunc.ne. 1)) then
!         routine = 'Defqchem'
!         message = 'Continuum solvent methods for Gaussian CAS only so far.'
!         call errormsg
!      endif

!
! check correct diabatisation asked for (extra tests could be added here)
!
      if (ifunc .eq. 1) then
         if (dddiab .lt. 0 .or. dddiab .gt. 4) then
            routine='Defqchem'
            message='GAUSSIAN DD surfaces need an adiabatic '// &
                       'or a regularised/local/global diabatic picture'
            call errormsg
         else if (dddiab .eq. 2 .and. hopipar(7) .ne. 3) then
            routine='Defqchem'
            message='GAUSSIAN DD local diabatic surfaces need a two-'// &
                       'electronic-state CASSCF calculation'
            call errormsg
         else if (dddiab .eq. 2 .and. (.not. lsacas)) then
            routine='Defqchem'
            message='GAUSSIAN DD local diabatic surfaces need an '// &
                    'SA-CASSCF calculation'
            call errormsg
         endif
      elseif(ifunc.eq.3)then
        if (dddiab .lt. 0 .and. dddiab .gt. 6) then
            routine='Defqchem'
            message='Molpro DD surfaces need an adiabatic, '// &
             'or regularised/propagated/projected/Procrustes diabatic picture'
            call errormsg
         endif
      elseif(ifunc.eq.5)then
         if (dddiab .ne. 0 .and. dddiab .ne. 3 .and. dddiab .ne. 4) then
            routine='Defqchem'
            message='Molcas DD surfaces need an adiabatic, '// &
                       'or propagated/projected diabatic picture'
            call errormsg
         endif
      elseif(ifunc.eq.6)then
         if (dddiab .ne. 0 .and. dddiab .ne. 4) then
            routine='Defqchem'
            message='Orca DD surfaces need an adiabatic or propagated diabatic picture'
            call errormsg
         endif
      elseif(ifunc.eq.8)then !Bagel: Based on molcas diabatization approach
         if (dddiab .ne. 0 .and. dddiab .ne. 3 .and. dddiab .ne. 4) then
            routine='Defqchem'
            message='Bagel DD surfaces need an adiabatic, '// &
                       'or propagated/projected diabatic picture'
            call errormsg
         endif
      endif

! To sort out later                                   /!\ temporary (Benjamin)
      if (lfchk) then                               ! /!\ temporary (Benjamin)
         routine='Defqchem'                         ! /!\ temporary (Benjamin)
         message='FCHK FILES: WORK IN PROGRESS!!'   ! /!\ temporary (Benjamin)
         call errormsg                              ! /!\ temporary (Benjamin)
      endif                                         ! /!\ temporary (Benjamin)
!
! if quantum chemistry program is assigned, assign ifile no.
!
      if (ifunc .ne. 0)then
         if(lddmctdh)then
            ifile=208
            if(lddsoc) ifile=209  
         else
            ifile=205
            if(lddsoc .and. .not. lsocdegen) ifile=210   ! complex DD PES
         endif
      endif

      return
 1000 routine = 'Defqchem'
      write(message,'(a,i5)') &
           'Wrong type of argument in one of the fields. IOSTAT:',ierr
      call errormsg
      end subroutine defqchem


!>#######################################################################
!>
!> This routine gets the value of a potential at a grid point from a
!> DB of energies.
!>
!> The point is contained in gpoint. 
!>
!>#######################################################################

      subroutine vqchem (v,gpoint,si,sf)

      implicit none

      integer(long)                           :: ga,gb
      integer(long), intent(in)               :: si,sf

      real(dop), intent(out)                                  :: v
      real(dop), dimension(ndofddpes), intent(in)             :: gpoint
      real(dop), dimension(ndofddpes)                         :: xgpoint
      real(dop), dimension(nddstate,nddstate)                 :: ddv
      real(dop), dimension(ndofddpes,nddstate,nddstate)       :: ddder1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate) :: ddder2
      real(dop), dimension(3,nddstate,nddstate)               :: dddip
      integer(long), save                                     :: tid=0 ! timer ID

!-----------------------------------------------------------------------
! Read DB to get LHA
!-----------------------------------------------------------------------
      if(lddtrans)then
         call ddq2x(gpoint,xgpoint)
      else
         xgpoint = gpoint
      endif
      call dddb_rd(ddv,ddder1,ddder2,dddip,xgpoint,.true.,.false.,.false.)   

!-----------------------------------------------------------------------
! Select value of potential to return
!-----------------------------------------------------------------------
      v = ddv(sf,si)

!-----------------------------------------------------------------------
! Select and save derivatives - ddderiv1 and 2 passed to derqchem
!-----------------------------------------------------------------------
      if (ldbsmall) then  
         do ga=1,ndofdd
            ddderiv1_gp(ga,sf,si,num_gp) = ddder1(ga,sf,si)
           do gb=1,ndofdd
             ddderiv2_gp(gb,ga,sf,si,num_gp) = ddder2(gb,ga,sf,si)
           enddo
         enddo
      else
         do ga=1,ndofdd
            ddderiv1(ga,sf,si) = ddder1(ga,sf,si)
           do gb=1,ndofdd
             ddderiv2(gb,ga,sf,si) = ddder2(gb,ga,sf,si)
           enddo
         enddo
      endif   

      if (sf .ne. si) then
        if (ldbsmall) then
           ddderiv1_gp(1:ndofdd,si,sf,num_gp) = &
           ddderiv1_gp(1:ndofdd,sf,si,num_gp)
           ddderiv2_gp(1:ndofdd,1:ndofdd,si,sf,num_gp) = &
           ddderiv2_gp(1:ndofdd,1:ndofdd,sf,si,num_gp)
         else
            ddderiv1(1:ndofdd,si,sf) = ddderiv1(1:ndofdd,sf,si)
            ddderiv2(1:ndofdd,1:ndofdd,si,sf) = ddderiv2(1:ndofdd,1:ndofdd,sf,si)  
        endif   
      endif

      return
      end subroutine vqchem

!>#######################################################################
!>
!> This routine gets the derivatives of a potential at a grid point from a
!> DB of energies.
!>
!> The point is contained in gpoint. 
!>
!>#######################################################################

      subroutine derqchem (der,order,si,sf)

      implicit none

      integer(long), intent(in)                       :: si,sf
      integer(long)                                   :: f,f1,totord,i,&
                                                         n,m,gdof
      integer(long), dimension(ndofddpes), intent(in) :: order
      real(dop), intent(out)                          :: der

     gdof = 0
     do m=1,nmode
        if(.not.gwpm(m))cycle
        gdof = gdof + nspfdof(m)
     enddo

      totord = 0
      do m=1,nmode
        if(.not.gwpm(m))cycle
         do n=1,nspfdof(m)
            i = spfdof(n,m)
            totord=totord+order(i)
         enddo
      enddo

      f=0
      f1=0
      
      if (totord .eq. 1) then
         do m=1,nmode
            if(.not.gwpm(m))cycle
            do n=1,nspfdof(m)
               i = spfdof(n,m)
               if (order(i) .gt. 0) f=i
            enddo
         enddo


         if (ldbsmall) then
            der = ddderiv1_gp(f,sf,si,num_gp)
         else
            der = ddderiv1(f,sf,si)
         endif  

      else if (totord .eq. 2) then
         n=0
         m=1
         do while (f .eq. 0) 
            n=n+1
            if (n > nspfdof(m)) then
               m = m + 1
               n = 1
            endif
            if ( m > nmode ) then
               routine='Derqchem (funcqchem)'
               message='No Dof defined for derivative'
               call errormsg
            endif
            i = spfdof(n,m)
            if (order(i) .gt. 0) f=i
         enddo

         if (order(f) .eq. 1) then
            do while (f1 .eq. 0)
               n = n + 1
               if (n > nspfdof(m)) then
                  m = m + 1
                  n = 1
               endif
               if ( m > nmode ) then
                  routine='Derqchem (funcqchem)'
                  message='No Second Dof defined for derivative'
                  call errormsg
               endif
               i = spfdof(n,m)
               if (order(i) .gt. 0) f1=i
            enddo
         else
            f1=f
         endif

         if (ldbsmall) then
            der = ddderiv2_gp(f,f1,sf,si,num_gp)
          else   
            der = ddderiv2(f,f1,sf,si)
          endif
      endif

      return
      end subroutine derqchem

!#######################################################################
! Set of routines to calculate the LHA of the PES(s) at a grid point,
! calling a quantum chemistry program (direct dynamics interface).
! The grid point is contained in gpoint, the LHA in v, deriv1, deriv2.
!
! Representations of another order than 0 or 2 yet to be implemented.
!
! LHAqchem is called by getddpes.
!
! For GAUSSIAN calculations, it calls ddgaussian.
!
! To add a new interface, add a new hopilab1 number, and then any code
! needed to calculate and extract the quantum chemistry data.
!
! nst : no. of states in system (nddstate)
! nst1: no. of states (sets of spfs) in dynamics 
!       (single-set = 1, multi-set > 1)
!
! v(s1,s) returned with adiabatic energies as diagonal elements
! deriv1(n,s1,s) returned with state gradient on diagonal (s,s).
!                in ddgausian, nact vectors stored on off-diagonal
! deriv2(n1,n,s1,s) returned with state Hessian on digaonal (s,s)
! nact(n,nactdim) Nact vectors
! soc(s,s1) SOCs
!#######################################################################

      subroutine lhaqchem (time,hopilab1,xgpoint,v,deriv1,deriv2, &
                 dipole,mocoeff,nact,soc,ndofddpes,hopipar1,lfail, &
                 ciconfig,cicoeff)

      implicit none

      logical(kind=4), intent(out)                        :: lfail
      integer(long), intent(in)                           :: hopilab1,&
                                                             ndofddpes
      integer(long), dimension(*), intent(in)             :: hopipar1
      real(dop), dimension(*), intent(in)                 :: xgpoint
      real(dop), intent(in)                               :: time
      real(dop), dimension(nddstate,nddstate), intent(out):: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out) :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), &
                                                         intent(out) :: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)             :: mocoeff
      real(dop), dimension(ndofddpes,nactdim), intent(out)           :: nact
      complex(dop), dimension(nsocstate,nsocstate), intent(out)      :: soc

      character(len=*), dimension(maxcivec,nsmult), intent(out) :: ciconfig
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(out) :: cicoeff

      integer(long)      :: ilbl,s,s1,locval,idx,t,t1,sm,sp,i
      integer(long)      :: ifiletyp,imethod
      integer(long)      :: noutfile,ifile
      character(len=c5)  :: outfileall,outfile1
      character(len=c5), dimension(:), allocatable :: outfile
      character(len=c2)  :: loclogold,lochead
      character(len=400) :: extcmd
      logical(kind=4)    :: lfound

!-----------------------------------------------------------------------
! Different programs generate a different number of output files
!-----------------------------------------------------------------------
      if (allocated(outfile)) deallocate(outfile)
      if (hopilab1 .eq. 1) then
         allocate(outfile(nddstate))    ! Gaussian may have 1 file / state
         noutfile = maxsta

      else if (hopilab1 .eq. 2) then
         allocate(outfile(3))           ! QOMMMA generates files for energy /
                                        ! gradient / hessian
         noutfile = 3
      else if (hopilab1 .eq. 5) then
         allocate(outfile(2))           ! MOLCAS 1 output file and 1 orbital file

         noutfile =2
      else if (hopilab1 .eq. 8) then
         allocate(outfile(2))           ! BAGEL 1 output file and 1 orbital file
                                        ! Alternative is using the files for ENERGY
         noutfile = 2                   ! gradient / NAC and hessian in output 
      else
         allocate(outfile(1))    
         noutfile = 1
      endif

      ifiletyp = 0
      outfile(:)=' '

! Run Quantum Chemistry
      select case (hopilab1)

      case default

         routine = 'lhaqchem'
         write(message,'(a,i5)') &
              'Quantum chemistry program not found :',hopilab1
         call errormsg

      case (1)
         call ddgaussian(time,xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,outfile)
         ifiletyp = 1

      case (2)
         call ddqommma(xgpoint,v,deriv1,deriv2,hopipar1,nddstate,lfail,outfile)
         ifiletyp = 13

      case (3)
         call ddmolpro(time,xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,dbnrec,outfile(1),ciconfig,cicoeff)
         ifiletyp = 5

      case (4)
         call ddqchem(xgpoint,v,deriv1,deriv2,hopipar1,nddstate,nstate,nact,lfail)
         ifiletyp = 10
         
      case (5)
         call ddmolcas(xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,lfail,outfile)
         ifiletyp = 11

      case (6)
         call ddorca(xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,lfail)
         ifiletyp = 12

      case (7)
         call ddexternal(xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,nddstate,lfail)

      case (8)
         call ddbagel(xgpoint,v,deriv1,deriv2,dipole,mocoeff,nact,&
              hopipar1,lfail,outfile)
         ifiletyp = 4

      case (9) ! case 9 is no QC program (i.e. internal quantics potential)
         continue
         return

      end select

!-----------------------------------------------------------------------
! Read Output
!-----------------------------------------------------------------------
      if (ifiletyp .gt. 0) then    
! Initialise Interface to QC output
         qc_nst = nddstate
         qc_ncoo = ndofddpes
         qc_nbasis = nbasis
         qc_nmode = ndofdd
         qc_nmult(1:7) = nddmult(1:7)
         qc_ncivec = maxcivec
         call alloc_readqc(ilog)

!-----------------------------------------------------------------------
! specify what needs to be read and initialize logical variable
!-----------------------------------------------------------------------
         imethod = hopipar1(7)
         qcread(:)=.false.
         qcread(3)=.true.                            ! Energy
         qcread(4)=.not.lddmctdh                     ! 1st Derivs
         if (hopipar1(6) .gt. 1) qcread(5)=.true.    ! 2nd Derivs if order > 1 
         if (imethod .eq. 12 .or. imethod .eq. 13 .or. &
               imethod .eq. 15 .or. &
               imethod .eq. 27 .or. imethod .eq. 28 .or. &
               imethod .eq. 32 .or. imethod .eq. 37) qcread(6)=.true.     ! MOs needed for CAS
         if (ldiabdb .and. dddiab.lt.5) qcread(7) = .true.  ! NACTs
         if (lcasocdb) qcread(11) = .true.                  ! SOCs
         if (dddiab .eq. 6) qcread(12) = .true.             ! CI Vectors
         
         lfound=.true.
         lfail=.false.

         call initqcout(lfail)

         do ifile=1,noutfile

            if (outfile(ifile) .eq. ' ') cycle
! Open output file 
            open(iddin,file=outfile(ifile),status='old',err=1003)

! Read output file
            call rdqcout(iddin,ifiletyp,imethod,ilog,lfound,lfail)
            
            close(iddin)

         end do ! end loop over files
         
         if (.not. lfail) call chkqcout(ifiletyp,ilog,lfound,lfail)

!-----------------------------------------------------------------------
! Load data from QC output into local arrays
! NB For nsmult > 1 if SOC components not degenerate, istatemap takes 
!    care of mapping
!-----------------------------------------------------------------------
         do s=1,nddstate
            t=istatemap(s)    ! map QC states to dynamics 
            if (lqcfound3(t)) v(s,s) = qcv(t)-dde0
            if (lqcfound4(t)) deriv1(:,s,s) = qcderiv1(:,t)
            if (lqcfound5(t)) deriv2(:,:,s,s) = qcderiv2(:,:,t)

            do s1=s+1,nddstate
               t1=istatemap(s1)    ! map QC states to dynamics 

               if (lqcfound7(t,t1)) then
                  idx = (s1-1)*(s1-2)/2 + s
                  nact(:,idx) = qcnact(:,t,t1)
               endif

            enddo
         enddo

! MOs
         if (qcread(6)) mocoeff(:,:) = qcmocoeff(:,:)

! SOCs 
         if (qcread(11)) then
            soc = 0.0_dop

            do s=1,nsocstate-1
               do s1=s+1,nsocstate
                  soc(s1,s) = qcsoc(s1,s)  
                  soc(s,s1) = qcsoc(s,s1)  
               enddo
            enddo
         endif

! CI Vector
         if (qcread(12)) then
            do sm = 1,nsmult
               sp = idxmult(sm)
               do i = 1,nciconf(sm)
                  ciconfig(i,sm) = qcciconfig(i,sp)
               enddo
               do s=1,imultblk(sm)
                  t = imultoff(sm) + s
                  do i = 1,nciconf(sm)
                     cicoeff(i,s,sm) = qccicoeff(i,t)
                  enddo
               enddo
            enddo
         endif

         call dealloc_readqc
      endif

!-----------------------------------------------------------------------
! concatenate output files. Distinguish between codes that produce
! one outfile per state and one outfile with everything
!-----------------------------------------------------------------------
#ifdef MPI
      locval=mpirank
#else
      locval=0
#endif
write(lochead,'(a,i3)') 'header',locval
call rmblnk(lochead)
write(loclogold,'(a,i3)') 'logold',locval
call rmblnk(loclogold)

!$omp critical
      if (lqcsaveall .and. hopilab1 .eq. 1) then
        do s=1,nddstate
           ilbl = index(outfile(s),' ')-1
           outfileall = outfile(s)(1:ilbl)//'all'
           extcmd='touch '//outfileall
           call excmd(extcmd)

! write a header to go between logs
           write(extcmd,'(2a)') 'echo "            " >>',lochead
           call excmd(extcmd)
           write(extcmd,'(a,f12.8,2a)') 'echo "**** Time : "',time/fs,&
                 '"   ****" >> ',lochead
           call excmd(extcmd)
           write(extcmd,'(2a)') 'echo "            " >>',lochead
           call excmd(extcmd)

! concatenate old logs and new log
           write(extcmd,'(4a)') 'mv ',outfileall(1:ilbl+3),' ',loclogold
           call excmd(extcmd)
           
           write(extcmd,'(8a)') 'cat ',loclogold,' ',lochead,' ',&
                  outfile(s)(1:ilbl),' > ',outfileall(1:ilbl+3)
           call excmd(extcmd)
           write(extcmd,'(2a)') 'rm ',loclogold
           call excmd(extcmd)
           write(extcmd,'(2a)') 'rm ',lochead 
           call excmd(extcmd)
        enddo

      else if (lqcsaveall .and. hopilab1 .eq. 3) then
        ilbl = index(outfile1,' ')-1
        outfileall = outfile1(1:ilbl)//'all'
        extcmd='touch '//outfileall
        call excmd(extcmd)

! write a header to go between logs
        write(extcmd,'(2a)') 'echo "            " >>',lochead
        call excmd(extcmd)
        write(extcmd,'(a,f12.8,2a)') 'echo "**** Time : "',time/fs,&
              '"   ****" >> ',lochead
        call excmd(extcmd)
        write(extcmd,'(2a)') 'echo "            " >>',lochead
        call excmd(extcmd)

! concatenate old logs and new log
        write(extcmd,'(4a)') 'mv ',outfileall(1:ilbl+3),' ',loclogold
        call excmd(extcmd)

        write(extcmd,'(8a)') 'cat ',loclogold,' ',lochead,' ',&
               outfile1(1:ilbl),' > ',outfileall(1:ilbl+3)
        call excmd(extcmd)
        write(extcmd,'(2a)') 'rm ',loclogold
        call excmd(extcmd)
        write(extcmd,'(2a)') 'rm ',lochead
        call excmd(extcmd)

      endif
!$omp end critical

      if (allocated(outfile)) deallocate(outfile)

      return
 1003 routine = 'Lhaqchem (funcqchem.F90)'
      message(1:80) = 'ERROR opening QC output file :'
      ilbl = index(outfile(ifile),' ')-1
      message(81:80+ilbl) = outfile(ifile)(1:ilbl)
      call errormsg

      end subroutine lhaqchem

!#######################################################################
!
! read input file
!
! energies in eV, coordinates in bohr
!
!#######################################################################

      subroutine ddqommma (gpoint,v,deriv1,deriv2,hopipar1,nst,lfail,&
                 outfile)

!$    use OMP_LIB

      implicit none

      logical(kind=4), intent(out) :: lfail
      integer(long) :: idiab,imethod,jlbl,fnum,fstate,iorder,cstate,&
                       istate,ilbl,mlbl
      integer(long), intent(in)                           :: nst
      integer(long), dimension(*), intent(in)             :: hopipar1
      integer(long), external                             :: slen
      real(dop), dimension(*), intent(in)                 :: gpoint
      real(dop), dimension(nst,nst), intent(out)          :: v
      real(dop), dimension(ndofddpes,nst,nst), intent(out)          :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nst,nst), intent(out):: deriv2
      character(len=120)                                  :: templdatfile,&
                                                             comfile,filnam
      character(len=120)                                  :: qjobnam,    &
                                                             qjobfile
      character(len=240)                                  :: extcmd
      character(len=*), dimension(3)                      :: outfile

      lfail = .false.

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic (GAUSSIAN), = 1 quasi-diabatic (GAUSSIAN+ROT)
! fnum    : no. of function
! fstate  : state to which function belongs (multi-set formalism)
! cstate  : state being calculated (running index further below)
! istate  : index of cstate without considering nroot
! iorder  : local order (or 0 for propagating CAS when DB is read)
! imethod : flag for the quantum chemistry method
! nroot   : highest state
!-----------------------------------------------------------------------
      idiab = hopipar1(2)
      fnum = hopipar1(3)
      fstate = hopipar1(4)
      iorder = hopipar1(6)
      imethod = hopipar1(7)

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      templdatfile=ddname(1:ddlaenge)//'/template.dat'

!-----------------------------------------------------------------------
! highest electronic state
!-----------------------------------------------------------------------
      cstate = nroot
      istate = nst

!-----------------------------------------------------------------------
! generate name (full path) of init_geom1.xyz qommma input file
!-----------------------------------------------------------------------
      write(filnam,'(a,2(a,i5),2a)')&
             ddname(1:ddlaenge),'/init_geom1'
      write(qjobnam,'(a,2(a,i5),2a)')&
             ddname(1:ddlaenge),'/qomjob'
             

      ilbl=index(filnam,' ')-1
      mlbl=index(qjobnam,' ')-1

!-----------------------------------------------------------------------
! need to tidy up wrqommma - too many unnecessary variables passed.
!-----------------------------------------------------------------------
      comfile=filnam(1:ilbl)//'.xyz'
      qjobfile=qjobnam(1:mlbl)//'.in'
!$omp critical (wr)
      call wrqommma(templdatfile,comfile,gpoint,natm)
!$omp end critical (wr)

!-----------------------------------------------------------------------
! run QOMMMA
!-----------------------------------------------------------------------
      jlbl=index(ddextcmd,' ')
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//qjobnam(1:mlbl+1)//char(zeroi)
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//qjobnam(ddlaenge+2:mlbl+1)// &
              ddname(1:ddlaenge+1)//char(zeroi)
      endif
      if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! Results file always has the same name: report1 (for energy only).
! Gradient file always has the same name: jobfiles/gradientsbohr1.
! Hessian file always has the same name: jobfiles/image1/freq1/hessian_input.
!-----------------------------------------------------------------------
      outfile(1)=ddname(1:ddlaenge)//'/report1'
      outfile(2)=ddname(1:ddlaenge)//'/jobfiles/gradientsbohr1'
      outfile(3)=ddname(1:ddlaenge)//'/jobfiles/image1/freq1/hessian_input'

      return

      end subroutine ddqommma

!#######################################################################

      subroutine wrqommma(templdatfile,comfile,xcoo,natm)

      implicit none

      integer(long)                                  :: iline,i,llbl,n
      integer(long), external                        :: slen
      integer(long), intent(in)                      :: natm
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: templdatfile,comfile
      character(len=120)                             :: filename

      real(dop), parameter                           :: angstroem = 0.52917725_dop

      character(LEN=80)::title
      character(LEN=6)::atmnum
      character(LEN=5)::atmtyp,con1,con2,con3,con4,con5
      integer::natms
      character(LEN=4)::atmname

      open(itmpin,file=templdatfile)
      open(idump,file=comfile)

      read(itmpin,'(i6,a)')natms,title
      write(idump,'(i6,a)')natms,title
      n=0
      iline=1
   10   continue
     read(itmpin,'(a6,2x,a4,36x,a5,x,a5,x,a5,x,a5,x,a5,x,a5)',err=999,end=20)&
            atmnum,atmname,atmtyp,con1,con2,con3,con4,con5
      n=n+1
      iline=iline+1
     write(idump,'(a6,2x,a4,3f15.6,x,a5,x,a5,x,a5,x,a5,x,a5,x,a5)')&
     atmnum,atmname,(xcoo(i,n)*angstroem,i=1,3),atmtyp,con1,con2,con3,con4,con5
     goto 10
   20   continue

      close(itmpin)
      close(idump)

      return
  999 routine='WRQOMMMA'
      llbl=slen(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline+1
      call errormsg
      end subroutine wrqommma

!####################################################################### 

      subroutine ddgaussian (time,gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                             hopipar1,outfile)

!$ use OMP_LIB
use openmpmod

      implicit none

      real(dop), intent(in)                               :: time
      integer(long) :: idiab,imethod,jlbl,klbl,fnum,fstate,iorder,cstate,&
                       istate,ilbl,iunit,isolvent,&
                       nthreads,sthreads
      integer(long), dimension(nddstate) :: ddlaenge1
      integer(long), dimension(5)                         :: found
      integer(long), dimension(*), intent(in)             :: hopipar1
      integer(long), external                             :: slen
      real(dop), dimension(*), intent(in)                 :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)                    :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out)          :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), intent(out):: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)  :: mocoeff
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), &
           intent(out):: nact

      character(len=c5)  :: templdatfile,keywddatfile,comfile, &
                             filnam,loctempldatfile
      character(len=c5), dimension(nddstate)               :: ddname1
      character(len=c5), dimension(nddstate), intent(out)  :: outfile
      character(len=c6)                                    :: extcmd
      logical          :: lstartchk, lexist
      logical, save                                         :: linitial = .true.

      found(:) = 0

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic, = 1 regularised diabatisation,
!           = 2 local diabatisation, = 3 propagation diabatisation
! fnum    : no. of function
! fstate  : state to which function belongs (multi-set formalism)
! cstate  : state being calculated in quantum chemistry (running index 
!           further below)
! istate  : index of state without considering nroot, i.e. index in space
!           of states in dynamics. Thus istate .ne. cstate if there is a
!           state mapping between QC and QD.
! iorder  : local order 
! imethod : flag for the quantum chemistry method
! isolvent: flag for continuum solvent method if requested      
! nroot   : highest state
! nthreads: no. of cpu threads to use in Gaussian calculations
! sthreads: no. of cpu threads to use in calculating different electronic states
!-----------------------------------------------------------------------
      idiab = hopipar1(2)
      fnum = hopipar1(3)
      fstate = hopipar1(4)
      iorder = hopipar1(6)
      imethod = hopipar1(7)
      isolvent = hopipar1(10)
      nthreads=1 !will be changed for OMP runs        
      sthreads=0 !will be changed for OMP runs

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      keywddatfile=ddname(1:ddlaenge)//'/keywd.dat'
      templdatfile=ddname(1:ddlaenge)//'/template.dat'

!-----------------------------------------------------------------------
! check for existence of directories for files
!-----------------------------------------------------------------------
      do istate=1,nddstate
         lexist=.false.
         write(filnam,'(a,2(a,i5))')&
                    ddname(1:ddlaenge),'/gwp',fnum,'_V',istate
         call rmblnk(filnam)
         ilbl=index(filnam,' ')-1
         iunit = iddin 
         open(iunit,file=filnam(1:ilbl)//'/scratch',err=666)
         close(iunit,status='delete')
         lexist=.true.
  666    continue
         if (linitial .and. lexist .and. .not. lddwrdb) then
            if (.not. lqcsimulate) then
               if (lovwr) then
                  extcmd='rm -rf '//filnam(1:ilbl)//'/*'
                  call excmd(extcmd)
               else
                  routine='DDgaussian (funcqchem.f90)'
                  write(message(1:69),'(a)') &
       'Calculation does not use DB yet DB Gaussian directories (and files?) '
                  write(message(70:122),'(a)') &
       'already exist. Delete DB/gwp* directories and restart'
                  call errormsg
               endif
            endif

         else if (.not. lexist) then
            extcmd='mkdir '//filnam(1:ilbl)
            call excmd(extcmd)
         endif
! save path as ddname1(1:ddlaenge1)
         ddname1(istate)=filnam
         ddlaenge1(istate)=ilbl
      enddo
      
!-----------------------------------------------------------------------
! does a start.chk file exist?
! (ignore if no_orbread set
!-----------------------------------------------------------------------
      lstartchk=.false.
      if (lorbread0) then
         inquire(file=ddname(1:ddlaenge)//'/start.chk',exist=lstartchk) 
      endif

!-----------------------------------------------------------------------
! highest electronic state calculated first
!-----------------------------------------------------------------------
! Zero output arrays.
      v=0.d0
      deriv1=0.d0
      deriv2=0.d0

! determine the number of threads to spawn for different states,        
! and how many will be left for quantum chemistry program       
! (the following is not commented out, it is read by the OMP compiler)  
!$    sthreads=min(ompthread/omp_get_num_threads(),nddstate)    
!$    nthreads=ompthread/(sthreads*omp_get_num_threads())       

! set nested parallelization    
!$    call omp_set_nested(.true.)       
         
!$OMP PARALLEL DO num_threads(sthreads) default(shared) &      
!$OMP firstprivate(templdatfile) &     
!$OMP private(istate,cstate,filnam,ilbl,jlbl,klbl,llbl,tmpfile,comfile,outfile,&      
!$OMP    extcmd,found,lexist,loctempldatfile) &       
!$OMP if(lompthread .and. lompqc)
      do istate=nddstate,1,-1 !loop going from high to low

         cstate=nroot-nddstate+istate

!-----------------------------------------------------------------------
! get filestem of com file and log or fchk file
! add directory name to distinguish from potentially simultaneous calcs
!-----------------------------------------------------------------------
         klbl=ddlaenge1(istate)
         if (nstate .eq. 1) then
            write(filnam,'(a,2(a,i5),2a)')&
                ddname1(istate)(1:klbl),'/gwp',fnum,'_V',istate,&
                '_',ddname(ddlaenge-dbdlength+1:ddlaenge)
         else
            write(filnam,'(a,3(a,i5),2a)') &
                ddname1(istate)(1:klbl),'/gwp',fnum,'s',fstate,'_V',istate,&
                '_',ddname(ddlaenge-dbdlength+1:ddlaenge)
         endif
         call rmblnk(filnam)
         ilbl=index(filnam,' ')-1

         if (dbnrec .gt. 0) linitial = .false.

!-----------------------------------------------------------------------
! copy template file
!-----------------------------------------------------------------------
         write(templdatfile,'(a,a)') & 
            ddname(1:ddlaenge),'/template.dat'
         write(loctempldatfile,'(a,a)') &
            ddname1(istate)(1:klbl),'/template.dat'
         call rmblnk(loctempldatfile)
         extcmd='cp '//trim(templdatfile)//' '//&
            trim(loctempldatfile)//char(zeroi)
         call excmd(extcmd)

!-----------------------------------------------------------------------
! copy checkpoint file for CAS calculations. 3 cases
! 1. If multiple templates cp initial SP chk file
! 2. If not using DB, copy initial start.chk if needed
! 3. If using DB use start.chk if DB empty
!-----------------------------------------------------------------------
         if (lstartchk .and. .not. lddrddb) then
             lexist=.false.
             inquire(file=filnam(1:ilbl)//'.chk',exist=lexist) 
             if (.not. lexist) then
                extcmd='cp '//ddname(1:ddlaenge)//'/start.chk '// &
                     filnam(1:ilbl)//'.chk'//char(zeroi)
                if (.not. lqcsimulate) call excmd(extcmd)
             endif
         else if (lstartchk .and. linitial) then
             extcmd='cp '//ddname(1:ddlaenge)//'/start.chk '// &
                  filnam(1:ilbl)//'.chk'//char(zeroi)
             if (.not. lqcsimulate) call excmd(extcmd)
         endif

!-----------------------------------------------------------------------
! generate name of com file and write it
! this is OMP critical as all threads use same template file
!-----------------------------------------------------------------------
         comfile=filnam(1:ilbl)//'.com'
         call wrgaussian(loctempldatfile,comfile,gpoint,mocoeff,natm,cstate, &
              imethod,nroot,iorder,atnam,nthreads,time)
!!$OMP end critical (wr)

!-----------------------------------------------------------------------
! run GAUSSIAN
!-----------------------------------------------------------------------
          jlbl=index(ddextcmd,' ')
          klbl=ddlaenge1(istate)
          if (qcsubform .eq. 1) then
             extcmd=ddextcmd(1:jlbl)//filnam(1:ilbl+1)//char(zeroi)
          else if (qcsubform .eq. 2) then
             extcmd=ddextcmd(1:jlbl)//filnam(klbl+2:ilbl+1)// &
                  ddname1(istate)(1:klbl+1)//char(zeroi)
          endif
          if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! if specified, convert chk file into fchk file
!-----------------------------------------------------------------------
          if (lfchk) then
             call rmblnk(fchkcmd)
             klbl=index(fchkcmd,' ')
             extcmd=fchkcmd(1:klbl)//filnam(1:ilbl)//'.chk'//char(zeroi)
             if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! generate name of log or fchk file and read it
!-----------------------------------------------------------------------
             outfile(istate)=filnam(1:ilbl)//'.fchk'
          else
            outfile(istate)=filnam(1:ilbl)//'.log'
          endif
! Dipoles only implemented for HF at the moment.
          dipole(1:3,1:nddstate,1:nddstate)=0.0_dop

      end do ! end loop over states
!$OMP END PARALLEL DO

      linitial=.false.

      return
      end subroutine ddgaussian

!#######################################################################

      subroutine wrgaussian(loctempldatfile,comfile,xcoo,mocoeff,natm, &
                 cstate,imethod,nroot,iorder,atnam,nthreads,time)

      implicit none

!      logical(kind=4), intent(in)                    :: lhesscalc
      logical                                        :: lmultemplate = .false., &
                                                        lnolink1 = .false.,&
                                                        lstdroute=.true.
      logical, dimension(9)                          :: found
      integer                                        :: ilbl,klbl,jlbl,&
                                                        llbl,iline,i,n,nc,j, &
                                                        k,koff,bmem
      integer(long)                                  :: itempl1,iinput1,ithr
      integer(long), intent(in)                      :: natm,cstate,&
                                                        imethod,nroot,iorder,&
                                                        nthreads
      real(dop), intent(in)                          :: time
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      real(sip), dimension(nbasis,nbasis), intent(in):: mocoeff
      character(len=c5), intent(in)                   :: loctempldatfile,comfile
      character(len=c5)                             :: filestem,filename,string,&
                                                        substring,filestem0
      character(len=c1), dimension(natm), intent(in) :: atnam

      real(dop), parameter                           :: angstroem = 0.52917725_dop
      
      logical, external                              :: streq
      integer, external                              :: slen

      ithr = 0
!-----------------------------------------------------------------------
! see if template file is not template.dat
!-----------------------------------------------------------------------
      ilbl=index(loctempldatfile,' ')-1
      lmultemplate = (loctempldatfile(ilbl-11:ilbl) .ne. "template.dat") 

!$    ithr=omp_get_thread_num()


      itempl1=itempl+ithr
      iinput1=iinput+ithr

      open(itempl1,file=loctempldatfile)
      open(iinput1,file=comfile)

!-----------------------------------------------------------------------
! get file stem of com file
! comfile(ilbl:jlbl) is name of com file ignoring path
! comfile(klbl:klbl) is position of . in filestem.com
!-----------------------------------------------------------------------
      jlbl = slen(comfile)
      ilbl = jlbl
      do 
         if (comfile(ilbl:ilbl) .eq. '/' .or. ilbl .eq. 0) exit
         ilbl = ilbl - 1
      enddo
      ilbl=ilbl+1
      klbl=index(comfile(ilbl:jlbl),'.')-1
      klbl = ilbl + klbl -1
      if (klbl .le. 0) then
         routine = 'WRGAUSSIAN'
         message = 'Error in file name '//comfile(1:jlbl)
         call errormsg
      endif
      filestem=comfile(1:klbl)
      if (ilbl .gt. 0) filestem0 = filestem(ilbl:klbl)
 
!-----------------------------------------------------------------------
! write com file
! metrastrings are used for file stem, root number, substituted links,
! and special IOps
!-----------------------------------------------------------------------
      found(:)=.false.
      filename=loctempldatfile
      do !read through the template file 
         substring = ' '
         read(itempl1,'(a)',err=999,end=100) string
         if (string(1:1) .eq. '#') then
! check if using a non-standard route
            substring=string
            call rmblnk(substring)
            jlbl=slen(substring)
            if(streq(substring(jlbl-5:jlbl),'nonstd')) lstdroute=.false.
         endif


         if (string(1:9) .eq. "--Link1--" .and. lnolink1) then
! ignore link1 data if requestedi and exit loop
            exit

         else if(string .eq. '<geometry>') then
            found(9)=.true.
            do n=1,natm
               if(atnam(n)(1:2).eq.'D ')then
                  write(iinput1,'(a8,4x,3f15.6)') 'H(Iso=2)', &
                         (xcoo(i,n)*angstroem,i=1,3)
               else
                  write(iinput1,'(a12,3f15.6)') atnam(n),(xcoo(i,n)*angstroem,i=1,3)
               endif
            enddo

         else if(string .eq. '<mo-coefficients>') then
             found(7)=.true.
!-----------------------------------------------------------------------
! Write MO coefficients
!-----------------------------------------------------------------------
             if(found(6)) then
                write(iinput1,'(a)')'(5D15.8)'
                do i = 1, nbasis ! Loop over MOs
                   write(iinput1,'(i5,a,d15.8)')i,' Alpha MO OE=',0.0_dop
                   koff = 0
                   do j = 1, nbasis/5 ! Loop over AOs
                      write(iinput1,'(5d15.8)')(mocoeff(i,k+koff),k=1,5)
                      koff = koff + 5
                   enddo
                   if(mod(nbasis,5).ne.0)then
                      write(iinput1,'(5d15.8)')(mocoeff(i,k+koff),k=1,mod(nbasis,5))
                   endif
                enddo
! Some controversy whether a cards input should finish with a 0:
! http://www.gaussian.com/g_tech/g_ur/k_guess.htm
! or a blank line: http://www.gaussian.com/g_tech/g_ur/m_input.htm
! It seems irrelevant. Commenting the next line generates input with a blank line
!               write(itmp,'(i5)')0 
             else !if not writting orbital coefficients remove one blank line
                backspace(iinput1)
             endif

         else 
            ilbl=index(string,'$')
            if (ilbl .gt. 0) then
               klbl=slen(string)
               jlbl=index(string(ilbl+1:klbl),'$')
               if (jlbl .le. 0) then
                  llbl=slen(filename)
                  routine = 'WRGAUSSIAN'
                  write(message,'(a,i5)') &
                       'Missing second $ in file '//filename(1:llbl)//&
                       ', line ',iline
                  call errormsg
               endif
               jlbl=ilbl+jlbl
               read(string(ilbl+1:jlbl-1),'(a)') substring
               nc = jlbl-ilbl-1
               if (substring(1:nc) .eq. 'chk') then
                  llbl=slen(filestem)
                  string=string(1:ilbl-1)//filestem(1:llbl)//&
                         string(jlbl+1:klbl)//'.chk'
                  llbl=slen(string)
                  write(iinput1,'(a)') string(1:llbl)
                  found(1)=.true.
               else if (substring(1:nc) .eq. 'chk0') then
                  llbl=slen(filestem0)
                  string=string(1:ilbl-1)//filestem0(1:llbl)//&
                         string(jlbl+1:klbl)//'.chk'
                  llbl=slen(string)
                  write(iinput1,'(a)') string(1:llbl)
                  found(1)=.true.
               else if (substring(1:nc) .eq. 'time') then
                  write(string,'(f10.6)') time/fs
                  llbl=slen(string)
                  write(iinput1,'(a)') string(1:llbl)
               else if (substring(1:4) .eq. 'cpus') then
                  write(substring,'(i3)') nthreads
                  call rmlblnk(substring)
                  llbl=slen(substring)
                  string=string(1:ilbl-1)//substring(1:llbl)
                  llbl=slen(string)
                  write(iinput1,'(a)') string(1:llbl)
               else if (substring(1:5) .eq. 'bmem:') then
                  read(substring(6:nc),*) bmem 
                  write(substring,'(i5)') nthreads*bmem
                  call rmlblnk(substring)
                  llbl=slen(substring)
                  string=string(1:ilbl-1)//substring(1:llbl)//&
                         string(jlbl+1:klbl)
                  llbl=slen(string)
                  write(iinput1,'(a)') string(1:llbl)
               else if (substring(1:4) .eq. 'root') then
                  if (imethod .eq. 12) then
                     write(substring,'(i5)') nroot
                     call rmlblnk(substring)
                     llbl=slen(substring)
                     string=string(1:ilbl-1)//substring(1:llbl)//&
                            string(jlbl+1:klbl)
                     llbl=slen(string)
                     write(iinput1,'(a)') string(1:llbl)
                     found(2)=.true.
                  else
                     llbl=slen(filename)
                     routine = 'WRGAUSSIAN'
                     message = 'Only use $root$ metastring in CAS calculation '&
                               //filename(1:llbl)
                     call errormsg
                  endif
               else if (substring(1:5) .eq. 'Swap:') then
                  if (imethod .eq. 12) then
                     if (cstate .eq. nroot-1) then
                        llbl=slen(substring)
                        string=substring(6:llbl)
                        llbl=slen(string)
                        write(iinput1,'(a)') string(1:llbl)
                        found(3)=.true.
                     endif
                  else
                     llbl=slen(filename)
                     routine = 'WRGAUSSIAN'
                     message = 'Only use $Swap:...$ metastring in a CAS calculation '&
                               //filename(1:llbl)
                     call errormsg
                  endif
               else if (substring(1:5) .eq. 'Freq:') then
                  if (iorder .ge. 2) then
                     llbl=slen(substring)
                     string=substring(6:llbl)
                     llbl=slen(string)
                     write(iinput1,'(a)') string(1:llbl)
                     found(4)=.true.
                  endif
               else if(substring(1:7) .eq. 'Guess0:') then
                  if (lorbread0) then
                     if(dbnrec.eq.0 .and. (imethod.eq.12))then 
                        llbl=slen(substring)
                        string=substring(8:llbl)
                        llbl=slen(string)
                        write(iinput1,'(a)') string(1:llbl)
                        found(5)=.true.
                     endif
                  endif
               else if(substring(1:6) .eq. 'Guess:') then
                  if (lorbread) then
                     if(dbnrec.gt.0 .and. (imethod.eq.12))then 
                        llbl=slen(substring)
                        string=substring(7:llbl)
                        llbl=slen(string)
                        write(iinput1,'(a)') string(1:llbl)
                        if (substring(7:17) .eq. 'Guess=Cards') found(6)=.true.
                     endif
                  endif
               else if (substring(1:6) .eq. 'Force:') then
                  if (iorder .eq. 1) then
                     llbl=slen(substring)
                     string=substring(7:llbl)
                     llbl=slen(string)
                     if (string(1:llbl) .eq. 'nolink1') then
                        lnolink1 = .true.
                     else
                        write(iinput1,'(a)') string(1:llbl)
                        found(8)=.true.
                     endif
                  endif
               else
                  klbl=slen(substring)
                  llbl=slen(filename)
                  routine = 'WRGAUSSIAN'
                  message = 'Metastring '//substring(1:klbl)// &
                            ' not yet implemented in file '// &
                            filename(1:llbl)
                  call errormsg
               endif
            else ! reproduce template          
               llbl=slen(string)
               write(iinput1,'(a)') string(1:llbl)
            endif

         endif
      end do
 
  100 continue

!-----------------------------------------------------------------------
! check all information needed is in place
!-----------------------------------------------------------------------
      llbl=slen(filename)
      if (.not. found(1)) then
         routine = 'WRGAUSSIAN'
         message = 'Missing $chk$ metastring in file '//filename(1:llbl)
         call errormsg
      else if(.not. found(9)) then
         routine = 'WRGAUSSIAN'
         write(message,'(a)') 'No <geometry> keyword in '//filename(1:llbl)
         call errormsg
      endif

!-----------------------------------------------------------------------
! for multiple template files no checks needed for individual state
! calculations
! no checks if using non-standard route
!-----------------------------------------------------------------------
      if (.not.lmultemplate .and. lstdroute) then

      if(dbnrec.eq.0 .and. (imethod.eq.12)) then
         if(lorbread0 .and. .not.found(5)) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $Guess0$ metastring in file '//&
                      filename(1:llbl)
            call errormsg
         endif
      endif

      if(dbnrec.gt.0 .and. (imethod.eq.12)) then
         if(lorbread .and. .not.found(6)) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $Guess$ metastring in file '//&
                      filename(1:llbl)
            call errormsg
         else if(.not.found(7)) then
            routine = 'WRGAUSSIAN'
            message = 'Found $Guess$, but absent <mo-coefficients> metastring in file '//&
                      filename(1:llbl)
            call errormsg
         endif 
      endif

      if (imethod .eq. 12) then
         if (.not. found(2)) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $root$ metastring in file '//&
                      filename(1:llbl)
            call errormsg
         endif
         if (cstate .eq. nroot-1 .and. (.not. found(3))) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $Swap:...$ metastring in file '&
                         //filename(1:llbl)
            call errormsg
         endif
      endif

      if (iorder .ge. 2) then
         if (.not. found(4)) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $Freq:...$ meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif

      if (iorder .eq. 1) then
         if (.not. found(8)) then
            routine = 'WRGAUSSIAN'
            message = 'Missing $Force:...$ meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif 

! end of checks
      endif

!-----------------------------------------------------------------------
! Close files and exit.
!-----------------------------------------------------------------------
      close(itempl1)
      close(iinput1)

      return
  999 routine='WRGAUSSIAN'
      llbl=slen(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline
      call errormsg
      end subroutine wrgaussian

!#######################################################################

      subroutine ddmolcas (gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                             hopipar1,lfail,outfile)

      implicit none

      logical(kind=4), intent(out) :: lfail
      integer(long) :: idiab,imethod,jlbl,fnum,fstate,iorder,&
                       istate,ilbl,klbl,ithr,iunit
      integer(long)                                       :: ddlaenge1
      integer(long), dimension(*), intent(in)             :: hopipar1
      integer(long), external                             :: slen
      real(dop), dimension(*), intent(in)                 :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)                    :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out)          :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), intent(out):: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)  :: mocoeff
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), &
           intent(out):: nact
      character(len=240)  :: templdatfile,filnam,loctempldatfile
      character(len=240)                                  :: extcmd
      character(len=240)                                  :: ddname1
      logical                                             :: lmultemplate, lstartorb, lexist
      logical, dimension(5)                               :: found
      logical(kind=4), save                               :: linitial = .true.
      character(len=c5), dimension(2), intent(out)        :: outfile

      lfail = .false. !overall failure at this geometry
      found(:) = .false.
      if (dbnrec .gt. 0) linitial=.false.

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic, = 1 regularised diabatisation,
!           = 2 local diabatisation, = 3 propagation diabatisation
! fnum    : no. of function
! fstate  : state to which function belongs (multi-set formalism)
! istate  : index of state without considering nroot, i.e. index in space
!           of states in dynamics. Thus istate .ne. cstate if there is a
!           state mapping between QC and QD.
! iorder  : local order (or 0 for propagating CAS when DB is read)
! imethod : flag for the quantum chemistry method
!-----------------------------------------------------------------------
      idiab = hopipar1(2)
      fnum = hopipar1(3)
      fstate = hopipar1(4)
      iorder = hopipar1(6)
      imethod = hopipar1(7)

      ithr = 0
! following is read by OMP compiler and ithr rewritten
!$    ithr = omp_get_thread_num()

!-----------------------------------------------------------------------
! look to see if multiple template files are to be used
!-----------------------------------------------------------------------
      lmultemplate=.false.
      templdatfile=ddname(1:ddlaenge)//'/template_1.dat'
      inquire(file=templdatfile,exist=lmultemplate)
      if (lmultemplate) then
        message = 'DD in Molcas with multiple templates not implemented'
        call errormsg
      endif

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      templdatfile=ddname(1:ddlaenge)//'/template.dat'

! Zero output arrays.
      v=0.0_dop
      deriv1=0.0_dop
      deriv2=0.0_dop
      nact=0.0_dop
      dipole(:,:,:)=0.0_dop

!-----------------------------------------------------------------------
! check for existence of directories for files
!-----------------------------------------------------------------------
      istate=1
      lexist=.false.
      write(filnam,'(a,2(a,i5))')&
                 ddname(1:ddlaenge),'/gwp',fnum
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1
      iunit = iddin+ithr
      open(iunit,file=filnam(1:ilbl)//'/scratch',err=666)
      close(iunit,status='delete')
      lexist=.true.
666   continue
      if (lexist .and. .not. lddwrdb) then
         if (.not. lqcsimulate) then
            if (lovwr) then
               extcmd='rm -rf '//filnam(1:ilbl)//'/*'
               call excmd(extcmd)
            else
               routine='DDmolcas (funcqchem.f90)'
               write(message(1:69),'(a)') &
      'Calculation does not use DB yet DB Molcas directories (and files?) '
               write(message(70:122),'(a)') &
      'already exist. Delete DB/gwp* directories and restart'
               call errormsg
            endif
         endif
      else if (.not. lexist) then
         extcmd='mkdir '//filnam(1:ilbl)
         call excmd(extcmd)
      endif
      ! save path as ddname1(1:ddlaenge1)
      ddname1=filnam
      ddlaenge1=ilbl

!$omp critical (wr)

!-----------------------------------------------------------------------
! get filestem of input file and log file
!-----------------------------------------------------------------------
      klbl=ddlaenge1
      if (nstate .eq. 1) then
         write(filnam,'(2a,i5,2a)')&
             ddname1(1:klbl),'/gwp',fnum,&
             '_',ddname(ddlaenge-dbdlength+1:ddlaenge)
      else
         write(filnam,'(a,2(a,i5),2a)') &
             ddname1(1:klbl),'/gwp',fnum,'s',fstate,&
             '_',ddname(ddlaenge-dbdlength+1:ddlaenge)
      endif
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1

!-----------------------------------------------------------------------
! Copy template in each directory
!-----------------------------------------------------------------------
      write(loctempldatfile,'(a,a)')&
         ddname1(1:ddlaenge1),'/template.dat'
      extcmd='cp '//trim(templdatfile)//' '//&
         trim(loctempldatfile)//char(zeroi)
      call excmd(extcmd)

!-----------------------------------------------------------------------
! generate name of input file and write it
!-----------------------------------------------------------------------
      inputfile=filnam(1:ilbl)//'.input'
      call wrmolcas(loctempldatfile,inputfile,gpoint,mocoeff, & 
           natm,idiab,imethod,iorder,atnam)
!$omp end critical (wr)

!-----------------------------------------------------------------------
! Copy orbital file for CAS calculations at start
!-----------------------------------------------------------------------
      if (linitial .and. lorbread0) then
!$omp critical
         lstartorb=.false.
         inquire(file=ddname(1:ddlaenge)//'/start.RasOrb',exist=lstartorb)
         if ((imethod.eq.12 .or. imethod.eq.32 .or. imethod.eq.27 .or. imethod.eq.37) .and. .not. lstartorb) then
            routine = 'DDMOLCAS'
            message = 'No start.RasOrb file provided in DB directory'
            call errormsg
         else 
             extcmd='cp '//ddname(1:ddlaenge)//'/start.RasOrb '// &
               filnam(1:ilbl)//'.RasOrb'//char(zeroi)
             if (.not. lqcsimulate) call excmd(extcmd)
         endif
!$omp end critical 
      endif

!-----------------------------------------------------------------------
! run MOLCAS
!-----------------------------------------------------------------------
      jlbl=index(ddextcmd,' ')
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//filnam(1:ilbl+1)//char(zeroi)
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//filnam(klbl+2:ilbl+1)// &
              ddname1(1:klbl+1)//char(zeroi)
      endif
      if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! generate name of output files and read them
!-----------------------------------------------------------------------
      outfile(1)=filnam(1:ilbl)//'.out'
      outfile(2)=filnam(1:ilbl)//'.RasOrb'

      return

      end subroutine ddmolcas

!####################################################################### 

      subroutine wrmolcas(loctempldatfile,inpfile,xcoo,mocoeff,natm, &
                 idiab,imethod,iorder,xatnam)

!$    use OMP_LIB

      implicit none

      logical, dimension(7)                          :: found
      integer(long)                                  :: ilbl,klbl,jlbl,&
                                                        llbl,iline,i,n,&
                                                        j,k,koff,nthreads
      integer(long), intent(in)                      :: natm,idiab,&
                                                        imethod,iorder
      real(sip), dimension(nbasis,nbasis), intent(in):: mocoeff
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: loctempldatfile,inpfile
      character(len=120)                             :: filestem,filename,&
                                                        filestem0,orbfile
      character(len=360)                             :: string,substring,&
                                                        ssubstring
      character(len=c1), dimension(natm), intent(in) :: xatnam

      real(dop), parameter            :: angstroem =0.52917725_dop
      logical                                       :: lmultemplate,ltag
      logical                                       :: lcas=.false.

      routine = 'WRMOLCAS'

! number of threads for MOLCAS to use
      nthreads=1 !will be changed in OMP runs
! the following is only run by OMP compiler      
!$    nthreads=ompthread/omp_get_num_threads()

!-----------------------------------------------------------------------
! see if template file is not template.dat
!-----------------------------------------------------------------------
      ilbl=index(loctempldatfile,' ')-1
      lmultemplate = (loctempldatfile(ilbl-11:ilbl) .ne. "template.dat")

      open(itmpin,file=loctempldatfile)
      open(idump,file=inpfile)

!-----------------------------------------------------------------------
! get file stem of inp file
! inpfile(ilbl:jlbl) is name of inp file ignoring path
! inpfile(klbl:klbl) is position of . in filestem.inp
!-----------------------------------------------------------------------
      jlbl = len_trim(inpfile)
      ilbl = index(inpfile,'/',back=.true.)+1
      klbl = index(inpfile,'.',back=.true.)-1
      if (klbl .le. 0) then
         message = 'Error in file name '//inpfile(1:jlbl)
         call errormsg
      endif
      filestem=inpfile(1:klbl)
      if (ilbl .gt. 0) filestem0 = filestem(ilbl:klbl)

      if(imethod.eq.12 .or. imethod .eq. 13 .or. imethod .eq. 27 &
       & .or. imethod.eq.28 .or. imethod .eq. 32 .or. imethod .eq. 37) then
         lcas=.true.
      else if (imethod .eq. 1 .or. imethod .eq. 5 .or. imethod .eq. 6) then
         lcas=.false.
      else
         message = 'Method not implemented for Molcas'
         call errormsg
      endif

      if(lcas) then
         orbfile=filestem(1:klbl)//'.RasOrb'
      else 
         orbfile=filestem(1:klbl)//'.ScfOrb'
      endif

!-----------------------------------------------------------------------
! write inp file
!-----------------------------------------------------------------------
      found(:)=.false.
      filename=loctempldatfile
      do
        substring = ' '
        ltag=.false.
        read(itmpin,'(a)',err=999,end=100) string

! look for <geometry/>
        call parsexml(string,'geometry',substring,ilbl,ltag)
        if (ltag) then
           found(1)=.true.
           do n=1,natm
! molcas(?) doesn't recognise D, must use H1 and increase mass in the input file
              if(xatnam(n) .eq. 'D' .or. xatnam(n) .eq. 'd')then
                 write(idump,'(a12,3f15.6)') 'H1',&
                      (xcoo(i,n)*angstroem,i=1,3)
              else
                 write(idump,'(a12,3f15.6)') xatnam(n),&
                      (xcoo(i,n)*angstroem,i=1,3)
              endif
           enddo

           cycle
        endif

! look for <cpus/>
        call parsexml(string,'cpus',substring,ilbl,ltag)
        if (ltag) then
           write(ssubstring,'(i3)') nthreads
           jlbl=len_trim(string)
           substring=string(1:ilbl-1)//ssubstring(1:3)//string(ilbl+7:jlbl)
           write(idump,'(a)') trim(substring)

           cycle
        endif
        
! look for <Freq>,</Freq> pair
        call parsexml(string,'Freq',substring,ilbl,ltag)
        if (ltag ) then
           found(2)=.true.
           if (iorder.ge.2 .and. (.not.lupdhes .or. dbnrec.eq.0)) then
              write(idump,'(a)') trim(substring)
           endif

           cycle
        endif

! look for <Force>,</Force> pair
        call parsexml(string,'Force',substring,ilbl,ltag)
        if (ltag) then
           found(3)=.true.
           if (iorder.ge.1) then
              write(idump,'(a)') trim(substring)
           endif

           cycle
        endif

! look for <NACT>,</NACT> pair
        call parsexml(string,'NACT',substring,ilbl,ltag)
        if (ltag) then
           found(4)=.true.
           if (lcas .and. (idiab.eq.3 .or. idiab .eq. 4 .or. &
                       (idiab.eq.0 .and. nddstate .gt. 0)) ) then
              write(idump,'(a)') trim(substring)
           endif

           cycle
        endif

! look for <Orbread0>,</Orbread0> pair
        call parsexml(string,'Orbread0',substring,ilbl,ltag)
        if (ltag) then
           found(5)=.true.
           if (dbnrec.gt.0) cycle
           call parsexml(substring,'orbfile',ssubstring,jlbl,ltag)
           if (ltag) then
              klbl=len_trim(substring)
              ssubstring=substring(1:jlbl-1)//trim(orbfile)//substring(jlbl+10:klbl)
              write(idump,'(a)') trim(ssubstring)
           else
              write(idump,'(a)') trim(substring)
           endif

           cycle
        endif

! look for <Orbread>,</Orbread> pair
        call parsexml(string,'Orbread',substring,ilbl,ltag)
        if (ltag) then
           found(6)=.true.
           if (dbnrec .eq. 0) cycle
           call parsexml(substring,'orbfile',ssubstring,jlbl,ltag)
           if (.not. ltag) then
              message = 'Orbread: no <orbfile/> tag in template.'
              call errormsg
           endif
           klbl=len_trim(substring)
           ssubstring=substring(1:jlbl-1)//trim(orbfile)//substring(jlbl+10:klbl)
           write(idump,'(a)') trim(ssubstring)

           cycle
        endif

! look for <SOC>,</SOC> pair
        call parsexml(string,'SOC',substring,ilbl,ltag)
        if (ltag) then
           found(7)=.true.
           if (lcas .and. lddsoc) then
              write(idump,'(a)') trim(substring)
           endif

           cycle
        endif

! write any other line in template verbatim
        write(idump,'(a)') trim(string)

      end do
  100 continue

!-----------------------------------------------------------------------
! check all information needed is in place
!-----------------------------------------------------------------------
      llbl=len_trim(filename)
      jlbl=len_trim(orbfile)

      if(dbnrec.eq.0 .and. lcas) then
         if(.not.found(5)) then
            routine = 'WRMOLCAS'
            message = 'Missing <Orbread0>...</Orbread0> metastring in file '//&
             filename(1:llbl)//'. Program generated guess orbitals being used.'
            call warnmsg
         endif
      endif

      if(dbnrec.gt.0 .and. lcas) then
         if(.not.found(6)) then
            routine = 'WRMOLCAS'
            message = 'Missing <Orbread>...</Orbread> metastring in file '//&
                      filename(1:llbl)
            call errormsg
         endif
      endif

      if (iorder .ge. 2) then
         if (.not. found(2)) then
!            if(lhesscalc)then
               routine = 'WRMOLCAS'
               message = 'Missing <Freq>...</Freq> meta'//&
                    'string in file ' //filename(1:llbl)
               call errormsg
!            endif
         endif
      endif
      if (iorder .ge. 1) then
         if (.not. found(3)) then
            routine = 'WRMOLCAS'
            message = 'Missing <Force>...</Force> meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif
      if (lcas .and. (idiab.eq.3 .or. idiab .eq. 4)) then
         if (.not. found(4)) then
            routine = 'WRMOLCAS'
            message = 'Missing <NACT>...</NACT> meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif

!-----------------------------------------------------------------------
! Close files 
!-----------------------------------------------------------------------
      close(itmpin)
      close(idump)

!-----------------------------------------------------------------------
! Write orbfile with MO coefficients if db not empty
!-----------------------------------------------------------------------
      if(found(4) .and. dbnrec.gt.0) then
         open(idump,file=orbfile,status='replace')

         write(idump,'(a)')'#INPORB 2.0'
         write(idump,'(a)')'#INFO'
         write(idump,'(a)')'* RASSCF average (pseudo-natural) orbitals'
         write(idump,'(a)')'       0       1       0'
         write(idump,'(i8)') nbasis
         write(idump,'(i8)') nbasis
         write(idump,'(a)')'#ORB'
         do i = 1, nbasis
            koff = 0
            write(idump,'(a,i3)')'* ORBITAL    1  ',i
            do j = 1, nbasis/5
               write(idump,'(5(E22.14E2))')(mocoeff(i,k+koff),k=1,5)
               koff = koff + 5
            enddo
            if(mod(nbasis,5).ne.0)then
               write(idump,'(5(E22.14E2))')&
                    (mocoeff(i,k+koff),k=1,mod(nbasis,5))
            endif
         enddo
         write(idump,'(a)')'#OCC'
         write(idump,'(a)')'* OCCUPATION NUMBERS'
         do j = 1, nbasis/10
            write(idump,'(10(F8.4))') (0.0,k=1,10)
         enddo
         if(mod(nbasis,10).ne.0)then
            write(idump,'(10(F8.4))') (0.0,k=1,mod(nbasis,10))
         endif
         close(idump,status='keep')
      endif

      return
  999 routine='WRMOLCAS'
      llbl=len_trim(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline
      call errormsg
      end subroutine wrmolcas

!####################################################################### 

      subroutine ddmolpro (time,gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                           hopipar1,nrec,outfile,ciconfig,cicoeff)

!$    use OMP_LIB

      implicit none

      real(dop), intent(in)                               :: time
      integer(long) :: idiab,imethod,jlbl,fnum,fstate,iorder,cstate,&
                       istate,ilbl,iorder0,ithr,iunit,diabmod
      integer(long) :: ierr,ierrno
      integer(long), intent(in)                           :: nrec
      integer(long), dimension(*), intent(in)             :: hopipar1
      integer(long), external                             :: slen
      integer(long)                                       :: ddlaenge1
      real(dop), dimension(*), intent(in)                 :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)           :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out) :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), &
                                                          intent(out):: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)   :: mocoeff
      real(dop), dimension(ndofddpes,nactdim), intent(out) :: nact
      character(len=c5)  :: templdatfile,inpfile,filnam,filnam0,orbfile,&
                            wfufile
      character(len=c5), intent(out)                      :: outfile
      character(len=c5)                                   :: ddname1
      character(len=240)                                  :: extcmd
      logical(kind=4)                                     :: lmultemplate,&
                                                             lexist,lstartorb,&
                                                             lorbreadold
      logical(kind=4)                                     :: lcas
      logical(kind=4), save                               :: linitial = .true.
      character(30), dimension(maxcivec,nsmult), intent(out)        :: ciconfig
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(out) :: cicoeff

      routine = 'ddmolpro'

      if (dbnrec .gt. 0) linitial=.false.

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic (GAUSSIAN), = 1 diabatic (GAUSSIAN)
! idiab   : = 2 quasi-diabatic (GAUSSIAN+ROT), 
!         : = 4 propagation diabatisation
!         : = 5 Robertson projection
!         : = 6 Procrustes
! fnum    : no. of function eg. index of symmetry equiv. structure
! fstate  : state to which function belongs (multi-set formalism)
! cstate  : state being calculated (running index further below)
! istate  : index of cstate without considering nroot
! iorder  : local order (or 0 for propagating CAS when DB is read)
! imethod : flag for the quantum chemistry method
! nroot   : highest state
! diabmod : when using Robertson diabatisation, whether to calculate the 
!           reference wavefunction (=0) or not(=1).
!-----------------------------------------------------------------------
      idiab = hopipar1(2)
      fnum = hopipar1(3)
      fstate = hopipar1(4)
      diabmod = hopipar1(5)
      iorder = hopipar1(6)
      imethod = hopipar1(7)

      lcas = .false.
      lcas = (imethod .eq. 12 .or. imethod .eq. 15 .or. imethod .eq. 32)

#if OMP
   ithr = omp_get_thread_num()
#else
   ithr = 0
#endif

!-----------------------------------------------------------------------
! look to see if multiple template files are to be used
!-----------------------------------------------------------------------
      lmultemplate=.false.
      templdatfile=ddname(1:ddlaenge)//'/template_1.dat'
      inquire(file=templdatfile,exist=lmultemplate)
      if (lmultemplate) then
         message = 'CAS DD in Molpro with multiple templates not implemented'
         call errormsg
      endif

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      if(idiab.eq.5.and.diabmod.eq.0)then
         templdatfile=ddname(1:ddlaenge)//'/template_0.dat'
      else
         templdatfile=ddname(1:ddlaenge)//'/template.dat'
      endif

!-----------------------------------------------------------------------
! check for existence of directories for files
!-----------------------------------------------------------------------
      istate=nddstate

      lexist=.false.
      if((idiab.eq.5.and.diabmod.eq.0).or.dbnrec.eq.0)then
         write(filnam,'(2a)')&
              ddname(1:ddlaenge),'/reference'
      else
         write(filnam,'(a,2(a,i5))')&
              ddname(1:ddlaenge),'/gwp',fnum,'_V',istate
      endif
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1
      iunit = iddin + ithr
      open(iunit,file=filnam(1:ilbl)//'/scratch',err=666)
      close(iunit,status='delete')
      lexist=.true.

  666 continue

      if (.not. lexist) then
         extcmd='mkdir '//filnam(1:ilbl)
         if (.not. lqcsimulate) then
            call excmd3(ierr,ierrno,extcmd)
            if(ierr.ne.0)then
               write(*,*)'Error making directory: ',filnam(1:ilbl)
               write(*,*)'ierrno = ',ierrno
            endif
         endif
      endif

! save path as ddname1(1:ddlaenge1)
      ddname1=filnam
      ddlaenge1=ilbl

!-----------------------------------------------------------------------
! highest electronic state
!-----------------------------------------------------------------------
      cstate = nroot
      istate = nddstate

!-----------------------------------------------------------------------
! get filestem of inp file
!-----------------------------------------------------------------------
      iorder0 = 0
      if (nstate .eq. 1) then
         write(filnam,'(a,2(a,i5),2a)')&
             ddname1(1:ddlaenge1),'/gwp',fnum,'_v',istate
      else
         if(idiab.eq.5.and.diabmod.eq.0)then
            write(filnam,'(2a)') &
                 ddname1(1:ddlaenge1),'/reference'
         else
            write(filnam,'(a,3(a,i5),2a)') &
                 ddname1(1:ddlaenge1),'/gwp',fnum,'s',fstate,'_v',istate
         endif
      endif

      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1

!-----------------------------------------------------------------------
! set up name for SACAS or CAS/PT2 calc
!-----------------------------------------------------------------------
      if (lcas) then
         call rmblnk(filnam)
         ilbl=index(filnam,' ')-1
         write(filnam0,'(2a)')ddname(1:ddlaenge),'/reference'

!         filnam0=filnam

! Zero output arrays.
         call zerovxd(v,nddstate*nddstate)
         call zerovxd(deriv1,ndofddpes*nddstate*nddstate)
         call zerovxd(deriv2,ndofddpes*ndofddpes*nddstate*nddstate)
         call zerovxd(nact,ndofddpes*dercpdim)

      endif

!-----------------------------------------------------------------------
! Copy orbital file for CAS calculations at start
!-----------------------------------------------------------------------
      lorbreadold = lorbread
      if (linitial .and. lorbread0) then
!$omp critical
         lstartorb=.false.
         inquire(file=ddname(1:ddlaenge)//'/init-orbital.orb',exist=lstartorb)
         if (lcas .and. .not. lstartorb) then
            routine = 'DDMOLPRO'
            message = 'No init-orbital.orb file provided in DB directory'
            call errormsg
         endif

         if (lstartorb .and. linitial) then
             extcmd='cp '//ddname(1:ddlaenge)//'/init-orbital.orb '// &
               filnam(1:ilbl)//'.orb'//char(zeroi)
             if (.not. lqcsimulate) call excmd(extcmd)
         endif
!$omp end critical

      else if (linitial) then

! turn off orbital reading in wrmolpro
         lorbread = .false.

      endif

! save orbfile name
      orbfile=filnam(1:ilbl)//'.orb'

      if(idiab.eq.5.and.diabmod.gt.0)then
         wfufile = ddname(1:ddlaenge)//'/reference/reference.wfu'
         extcmd='cp '//wfufile//' '//ddname1(1:ddlaenge1)//'/. '
         if (.not. lqcsimulate) call excmd(extcmd)
      endif

!-----------------------------------------------------------------------
! generate name of com file and write it
!-----------------------------------------------------------------------
      inpfile=filnam(1:ilbl)//'.inp'

!$omp critical (wr)
      call wrmolpro(templdatfile,inpfile,orbfile,gpoint,mocoeff,natm, &
           idiab,imethod,iorder,xatnam,nrec,linitial,diabmod,wfufile)!,lhesscalc)

! restore orbread flag after initial calculation
      if (linitial) lorbread = lorbreadold
!$omp end critical (wr)

!-----------------------------------------------------------------------
! run MOLPRO
!-----------------------------------------------------------------------
      jlbl=index(ddextcmd,' ')
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//filnam(ddlaenge1+2:ilbl+1)//char(zeroi)
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//filnam(ddlaenge1+2:ilbl+1)// &
              ddname1(1:ddlaenge1+1)//char(zeroi)
      endif
      if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! generate name of log or fchk file and read it
!-----------------------------------------------------------------------
      outfile=filnam(1:ilbl)//'.out'

!-----------------------------------------------------------------------
! If using Procrustes diabatisation, read the 
! CI vector(s). The result is stored globally then written to file.
!
! TO SORT?
!-----------------------------------------------------------------------
!      if (idiab.eq.6) then
!         fnum=1
!         call rdmolpro_ci(ddname1,ddlaenge1,outfile,ciconfig,cicoeff,fnum)
!      endif

      return

      end subroutine ddmolpro

!#######################################################################
! Subroutine to read CI vectors from a Molpro punch file.
! GWR 05/19
!#######################################################################
      subroutine rdmolpro_ci(dname,dlaenge,outfileci,ciconfig,cicoeff,nsym)

      implicit none

      integer(long), intent(in) :: dlaenge,nsym
      character(len=dlaenge), intent(in) :: dname
      character(len=*), intent(in) :: outfileci

      logical(kind=4) :: fexist,lmrci
      integer(long)  :: sm,sm1,nst,nrow,n,soff,i,s,inpun,ncol,&
           msta,mconf,j,k,nci
      character(360) :: newpun,string
      character(50) :: form1,form2
      character(30), dimension(maxcivec,nsmult), intent(out)        :: ciconfig
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(out) :: cicoeff
      integer(long), dimension(nddstate) :: maxciidx
      real(dop), dimension(nddstate) :: maxci
      real(dop), allocatable, dimension(:,:) :: sci,tconf

      logical(kind=4), save :: lfirst=.true.

      inpun=800

! Check that the calculation converged.
      inquire(file=outfileci,exist=fexist)
      if(.not.fexist)then
         routine='rdmolpro_ci'
         message='Output file does not exist'
         call errormsg
      endif
      open(inpun,file=outfileci,err=100)
      rewind(inpun)
      string=' '
      do 
         read(inpun,'(a)',end=200)string
         if(index(string,'NO CONVERGENCE').ne.0)then
            routine='rdmolpro_ci'
            message='CI calculation failed to converge'
            call errormsg
         endif
         cycle
200      exit
      enddo
      close(inpun,status='keep')

! Open the punch file containing the CI vector.
      newpun = dname(1:dlaenge)//'/reference.pun'
      inquire(file=newpun,exist=fexist)
      if(.not.fexist)then
         routine='rdmolpro_ci'
         message='Punch file does not exist'
         call errormsg
      endif
      inpun=800
      open(inpun,file=newpun,err=101)
      rewind(inpun)

! Read through the punch file to get the number of configurations.
! Only do this once as the numbers should be fixed.  
! Also determine whther the punch file is from CASSCF or MRCI.
      lmrci = .false.
      if(lfirst)then
         do    
            string=' '
            read(inpun,'(a)',end=300)string
            if(index(string,'REFERENCE').ne.0 .or. index(string,'MRCI').ne.0) then
               lmrci=.true.
               exit
            else if (index(string,'MCSCF').ne.0)then
               lmrci =.false.
               exit
            endif

            cycle
300         exit
         enddo
         lfirst = .false.
      endif

! Set up the format for the Molpro punch file as it varies depending on
! the number of active orbitals.
      if(lmrci)then
         write(form1,'(a,i3,a)')'(a',ncas+9,',9f12.7)'
         call rmblnk(form1)
         write(form2,'(a,i3,a)')'(',ncas+9,'x,9f12.7)'
         call rmblnk(form2)
         ncol=9
      else
         write(form1,'(a,i3,a)')'(a',ncas+4,',7f12.7)'
         call rmblnk(form1)
         write(form2,'(a,i3,a)')'(',ncas+4,'x,7f12.7)'
         call rmblnk(form2)
         ncol=7
      endif

      msta = maxval(imultblk(1:nsmult))
      mconf = 1
      do sm = 1,nsmult
         if(nconf(sm,1)*nconf(sm,2).gt.mconf)&
              mconf = nconf(sm,1)*nconf(sm,2)
      enddo
      allocate(sci(msta,msta),tconf(mconf,msta))

! Read the CI vectors.
      do sm = 1,nsmult
! Loop to get to the start of the relevant spin block. 
         rewind(inpun)
         do sm1 = 1,sm
            do 
               read(inpun,'(a)',end=500)string
               if(index(string,'M').ne.0 .or. index(string,'***').ne.0 &
                    .or. index(string,'---').ne.0&
                    .or. index(string,'REFERENCE').ne.0&
                    .or. index(string,'MRCI').ne.0)then
                  cycle
               else if (sm1.eq.sm)then
                  backspace(inpun)
                  exit
               else
                  do 
                     read(inpun,'(a)',end=500)string
                     if(index(string,'M').ne.0 .or. index(string,'***').ne.0 &
                          .or. index(string,'---').ne.0&
                          .or. index(string,'REFERENCE').ne.0&
                          .or. index(string,'MRCI').ne.0) exit
                  enddo
                  exit
               endif
               cycle
500            exit
            enddo
         enddo

! Now read the actual CI vectors.
         nst = imultblk(sm)
         nci = nciconf(sm)
         nrow = (nst-1)/ncol+1
         maxci = 0.0_dop
         maxciidx = 0

         do i = 1,nci
            read(inpun,'(a)')string
            if(index(string,'MRCI').ne.0 .or. index(string,'---').ne.0)goto 1000
            read(string,form1,err=1000) ciconfig(i,sm),&
                 (cicoeff(i,s,sm),s=1,min(ncol,nst))
            call rmblnk(ciconfig(i,sm))
            soff=0
            do n = 2,nrow
               soff = soff+ncol
               if(n.ne.nrow)then
                  read(inpun,form2)&
                       (cicoeff(i,s,sm),s=soff+1,ncol+soff)
               else
                  read(inpun,form2)&
                       (cicoeff(i,s,sm),s=soff+1,nst)
               endif
            enddo
! Must ensure the sign of the CI vector follows the Molpro convention of the largest element
! for each state being >0. Molpro sometimes prints the CI vectors before fixing the sign.
            do s = 1,nst
               if(abs(cicoeff(i,s,sm)).gt.maxci(s))then
                  maxci(s) = abs(cicoeff(i,s,sm))
                  maxciidx(s) = i
               endif
            enddo
            cycle
! When reading MRCI coefficients, some with near 0 coefficients are usually left out 
! due to Molpro formatting. Also occurs when using symmetry.
1000        cicoeff(i:nci,1:nst,sm)=0.0_dop
            ciconfig(i:nci,sm) = ' '
            exit
         enddo

! Fix the signs if needed.
         do s = 1,nst
            if(cicoeff(maxciidx(s),s,sm).lt.0.0_dop)then
               do i = 1,nci
                  cicoeff(i,s,sm) = -1.0_dop*cicoeff(i,s,sm)
               enddo
            endif
         enddo

! Orthonormalise the vectors (due to precision of output).
         do i = 1,nst
            do j = 1,nst
               sci(i,j) = dot_product(cicoeff(1:nci,i,sm),&
                    cicoeff(1:nci,j,sm))
            enddo
         enddo

         do i = 1,nst
            sci(i,i) = 1.5_dop-0.5_dop*sci(i,i)
            do j = i+1,nst
               sci(j,i) = -0.5_dop*sci(j,i)
            enddo
            do j = 1,i-1
               sci(j,i) = -0.5_dop*sci(j,i)
            enddo
         enddo

         tconf = 0.0_dop
         do i = 1,nst
            do j = 1,nci
               do k = 1,nst
                  tconf(j,i) = tconf(j,i) + sci(i,k)*cicoeff(j,k,sm)
               enddo
            enddo
         enddo
         do i = 1,nst
            cicoeff(1:nci,i,sm) = tconf(1:nci,i)
         enddo
         
      enddo

      close(inpun,status='keep')

      deallocate(sci,tconf)

      return

100   routine='rdmolpro_ci'
      message='Error opening output file'
      call errormsg
101   routine='rdmolpro_ci'
      message='Error opening punch file'
      call errormsg

      end subroutine rdmolpro_ci

!#######################################################################
! Routine to read in the relevant parts of the AO overlap matrix 
! (between two different geometries).
! GWR 05/19
!#######################################################################
      subroutine rdmolpro_ao(outfile,sao,nbas,laofail)

      implicit none

      logical(kind=4), intent(out) :: laofail
      character(len=*), intent(in) :: outfile
      real(dop), dimension(nbasis,nbasis), intent(out) :: sao
      integer(long), intent(in)    :: nbas

      logical(kind=4)    :: fexist,ltransp
      character(len=360) :: string
      integer(long)      :: nrow,n,i,j,noff
      real(dop), dimension(2*nbas,2*nbas) :: saotmp

! Check that output file exists, then open it.
      inquire(file=outfile,exist=fexist)
      if(.not.fexist)then
         routine='rdmolpro_ao'
         message='ao_over.out file does not exist'
         call errormsg
      endif
      open(itmpin,file=outfile)
      rewind(itmpin)

! Read the output file to extract the AO overlap matrix between all atoms present.
! NB There are different formats; using "matrop,print" prints the overlap matrix
! as floating point values with 8 dp, whilst using "matrop,pric" prints the 
! TRANSPOSE of S in scientific notation with 6 sf.
      laofail = .true.
      ltransp = .false.

      string=' '
      do
         read(itmpin,'(a)',end=100)string

         if(index(string,'matrop,print').ne.0)then
            ltransp = .false.
         else if(index(string,'matrop,pric').ne.0)then
            ltransp = .true.

         elseif(index(string,'MATRIX S').ne.0)then
            read(itmpin,'(a)') string
            read(itmpin,'(a)') string

            nrow = nbasis/5

            if(ltransp)then
               do n = 1,2*nbasis
                  read(itmpin,*)
                  read(itmpin,*,err=999)(saotmp(j,n),j=1,2*nbasis)
               enddo
            else
               do n = 1,2*nbasis
                  read(itmpin,*)
                  read(itmpin,*,err=999)(saotmp(n,j),j=1,2*nbasis)
               enddo
            endif
         endif
      enddo

100   continue
      laofail = .false.

! Copy to relevant parts of the full overlap matrix to the matrix of overlaps between
! the basis functions at different geometries. sao(geom2,geom1)
      sao = 0.0_dop
      noff = nbasis
      do i = 1,nbasis
         do j = 1,nbasis
            sao(j,i) = saotmp(j+noff,i)
         enddo
      enddo

      close(itmpin)

      return
999   routine='Rdmolpro_ao (funcqchemmod.F90)'
      message='Error reading AO overlaps'
      call errormsg

      end subroutine rdmolpro_ao

!#######################################################################
! Routine to read in the relevant parts of the AO overlap matrix.
! GWR 05/19
!#######################################################################
      subroutine rdmolpro_ao_1(outfile,sao,nbas,laofail)

      implicit none

      logical(kind=4), intent(out) :: laofail
      character(len=*), intent(in) :: outfile
      real(dop), dimension(nbasis,nbasis), intent(out) :: sao
      integer(long), intent(in)    :: nbas

      logical(kind=4)    :: fexist
      character(len=360) :: string
      integer(long)      :: nrow,n,i,joff,j
      real(dop), dimension(nbas,nbas) :: saotmp

! Check that output file exists, then open it.
      inquire(file=outfile,exist=fexist)
      if(.not.fexist)then
         routine='rdmolpro_ao'
         message='ao_over.out file does not exist'
         call errormsg
      endif
      open(itmpin,file=outfile)
      rewind(itmpin)

! Read the output file to extract the AO overlap matrix between all atoms present.
      nrow = nbas/10
      laofail = .true.

      string=' '
      do
         read(itmpin,'(a)',end=100)string

         if(index(string,'MATRIX S').ne.0)then
            read(itmpin,*)
            read(itmpin,*)

            do n = 1,nbas
               read(itmpin,*)
               joff = 0
               do i = 1,nrow
                  read(itmpin,'(10(x,d13.6,x))',err=999)(saotmp(n,joff+j),j=1,10)
                  joff = joff+10
               enddo
               if(mod(nbas,10).ne.0)&
                  read(itmpin,'(10(x,d13.6,x))',err=999)(saotmp(n,joff+j),j=1,mod(nbas,10))
            enddo

            laofail = .false.
         endif

      enddo

100   continue

! Copy to relevant parts of the full overlap matrix to the matrix of overlaps between
! the basis functions at different geometries.
      do i = 1,nbas
         do j = 1,nbas
            sao(i,j) = saotmp(i,j)
         enddo
      enddo

999   close(itmpin)

      return

      end subroutine rdmolpro_ao_1

!#######################################################################
! Subroutine to write a Molpro input file containing up to two 
! geometries in one, so as to allow the creation of the matrix of 
! overlaps between the AO basis functions at the two geometries. Can
! also just input a single geometry if needed.
! GWR 05/19
!#######################################################################
      subroutine wrmolpro_geom(templdatfile,inpfile,xcoo,xcoo0,xatnam,natm,&
           ngeom)

      implicit none

      character(len=*), intent(in)                   :: templdatfile,inpfile
      character(len=c1), dimension(natm), intent(in) :: xatnam
      integer(long), intent(in)                      :: natm,ngeom
      real(dop), dimension(3,natm), intent(in)       :: xcoo,xcoo0

      logical(kind=4)                                :: fexist
      logical(kind=4), dimension(2)                  :: found
      character(len=360)                             :: string
      integer(long)                                  :: iline,llbl,n,i,imaxcoo
      integer(long), external                        :: slen
      real(dop)                                      :: atcomp,maxcoo
      real(dop), dimension(3,natm)                   :: xcootmp
      real(dop), parameter                           :: angstroem=0.52917725_dop

! Check that template file exists, then open it and the input file.
      inquire(file=templdatfile,exist=fexist)
      if(.not.fexist)then
         routine='wrmolpro_geom'
         write(message,'(2a)')'Template file does not exist: ',trim(templdatfile)
         call errormsg
      endif
      open(itmpin,file=templdatfile)
      open(idump,file=inpfile,status='replace')

      found=.false.

      iline = 0
      do 
         iline=iline+1
         read(itmpin,'(a)',err=999,end=100) string

         if(string.ne.'<geometry1>' .and. string.ne.'<geometry2>')then
            llbl=slen(string)
            write(idump,'(a)') string(1:llbl)
         else if(string .eq. '<geometry1>') then
            do n = 1,natm
               if(xatnam(n) .eq. 'D' .or. xatnam(n) .eq. 'd')then
                  write(idump,'(a12,3f15.6)') 'H1',&
                       (xcoo0(i,n)*angstroem,i=1,3)
               else
                  write(idump,'(a12,3f15.6)') xatnam(n),&
                       (xcoo0(i,n)*angstroem,i=1,3)
               endif
            enddo
            found(1) = .true.
         else if(string .eq. '<geometry2>') then
! Check for the presence of coincident atom centres, which Molpro doesn't like, and if
! there, slightly alter the geometry. This will introduce a small, but tolerable, error in 
! the overlap matrix.
            xcootmp = xcoo
            do n = 1,natm
               atcomp = 0.0_dop
               maxcoo = 0.0_dop
               do i = 1,3
                  atcomp = atcomp + abs(xcootmp(i,n)-xcoo0(i,n))*angstroem
                  if(abs(xcootmp(i,n)-xcoo0(i,n))*angstroem.ge.maxcoo)then
                     maxcoo = abs(xcootmp(i,n)-xcoo0(i,n))*angstroem
                     imaxcoo = i
                  endif
               enddo
               if(atcomp.lt.1.0e-6_dop)&
                    xcootmp(imaxcoo,n) = xcootmp(imaxcoo,n) &
                    + sign(1.0e-6_dop/angstroem,xcoo(imaxcoo,n))
            enddo
! Write the geometry.
            do n = 1,natm
               if(xatnam(n) .eq. 'D' .or. xatnam(n) .eq. 'd')then
                  write(idump,'(a12,3f15.6)') 'H1',&
                       (xcootmp(i,n)*angstroem,i=1,3)
               else
                  write(idump,'(a12,3f15.6)') xatnam(n),&
                       (xcootmp(i,n)*angstroem,i=1,3)
               endif
            enddo
            found(2) = .true.
         endif

      enddo

  100 continue

      close(itmpin)
      close(idump)

      if((.not.found(1) .or. .not.found(2)).and.ngeom.eq.2)then
         routine = 'wrmolpro_geom'
         message = 'Fewer than two geometry cards found in template.'
         call errormsg
      else if (.not.found(1).and.ngeom.eq.1)then
         routine = 'wrmolpro_geom'
         message = 'No geometry card found in template.'
         call errormsg
      endif

      return

  999 routine='wrmolpro_geom'
      llbl=slen(templdatfile)
      write(message,'(a,i5)')  &
           'Error reading file '//templdatfile(1:llbl)// &
           ', line ',iline
      call errormsg

      end subroutine wrmolpro_geom

!#######################################################################

      subroutine wrmolpro(templdatfile,inpfile,orbfile,xcoo,mocoeff,natm, &
                 idiab,imethod,iorder,xatnam,nrec,linitial,diabmod,wfufile)

      implicit none

!      logical(kind=4), intent(in)                    :: lhesscalc
      logical(kind=4), dimension(7)                  :: found
      integer(long)                                  :: ilbl,klbl,jlbl,&
                                                        llbl,iline,i,n,nc,&
                                                        orbflen,&
                                                        laenge 
      integer(long), external                        :: slen
      integer(long), intent(in)                      :: natm,idiab,nrec,&
                                                        imethod,iorder,diabmod
      real(sip), dimension(nbasis,nbasis), intent(in):: mocoeff
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: templdatfile,inpfile, &
                                                        orbfile,wfufile
      character(len=120)                             :: filestem,filename,&
                                                        filestem0
      character(len=360)                             :: string,substring!,&
!                                                        orbfile0
      character(len=c1), dimension(natm), intent(in) :: xatnam
      logical(kind=4), intent(in)                    :: linitial
      logical(kind=4)                                :: lorbfile,lcas

      integer, parameter                             :: tlines = 200 !max number of lines in template file
      real(dop), parameter                           :: angstroem = 0.52917725_dop

!      routine = 'WRMOLPRO'

      ilbl=index(templdatfile,' ')-1

      open(itmpin,file=templdatfile,status='old',err=990)
      open(idump,file=inpfile,status='unknown',err=990)

      lcas = .false.
      lcas = (imethod .eq. 12 .or. imethod .eq. 15 .or. imethod .eq. 32)

!-----------------------------------------------------------------------
! get file stem of inp file
! inpfile(ilbl:jlbl) is name of inp file ignoring path
! inpfile(klbl:klbl) is position of . in filestem.inp
!-----------------------------------------------------------------------
      jlbl = slen(inpfile)
      ilbl = jlbl
      do 
         if (inpfile(ilbl:ilbl) .eq. '/' .or. ilbl .eq. 0) exit
         ilbl = ilbl - 1
      enddo
      ilbl=ilbl+1
      klbl=index(inpfile(ilbl:jlbl),'.')-1
      klbl = ilbl + klbl -1
      if (klbl .le. 0) then
         message = 'Error in file name '//inpfile(1:jlbl)
         call errormsg
      endif
      filestem=inpfile(1:klbl)
      if (ilbl .gt. 0) filestem0 = filestem(ilbl:klbl)
 
      orbflen = slen(orbfile)
      if (orbflen .gt. 0) then
         laenge = index(orbfile,'/',back=.true.)+1
         orbflen = orbflen-laenge
      endif

!      wfuflen = slen(wfufile)
!      if(wfuflen.gt.0)then 
!         laenge = index(orbfile,'/',back=.true.)+1
!         orbflen = orbflen-laenge
!      endif

!-----------------------------------------------------------------------
! write inp file
! metrastrings are used for file stem, root number, substituted links,
! and special IOps
!-----------------------------------------------------------------------
      found(:)=.false.
      filename=templdatfile
      iline=0

      do 
         substring = ' '
         read(itmpin,'(a)',err=999,end=100) string
         iline=iline+1
         if (string .ne. '<geometry>') then
            ilbl=index(string,'$')
            if (ilbl .gt. 0) then
               klbl=slen(string)
               jlbl=index(string(ilbl+1:klbl),'$')
               if (jlbl .le. 0) then
                  llbl=slen(filename)
                  routine = 'WRMOLPRO'
                  write(message,'(a,i5)') &
                       'Missing second $ in file '//filename(1:llbl)//&
                       ', line ',iline
                  call errormsg
               endif
               jlbl=ilbl+jlbl
               read(string(ilbl+1:jlbl-1),'(a)') substring
               nc = jlbl-ilbl-1
               if (substring(1:7) .eq. 'Orbrot:'.or.substring(1:7) .eq. 'orbrot:') then
                  if (lcas) then
                     llbl=slen(substring)
                     string=substring(8:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                     found(1)=.true.
                  else
                     llbl=slen(filename)
                     routine = 'WRMOLPRO'
                     message = 'Do not use $Orbrot:...$ metastring in file ' &
                               //filename(1:llbl)
                     call errormsg
                  endif
               else if (substring(1:5) .eq. 'Freq:'.or.substring(1:5) .eq. 'freq:') then
                  if (iorder .ge. 2) then
                     if(.not. lupdhes .or. nrec .eq. 0)then
                        llbl=slen(substring)
                        string=substring(6:llbl)
                        llbl=slen(string)
                        write(idump,'(a)') string(1:llbl)
                        found(2)=.true.
                     endif
                  endif
               else if(substring(1:9) .eq. 'Orbread0:'.or.substring(1:9) .eq. 'orbread0:') then
                  routine='Wrmolpro (funcqchem.f90)'
                  message = 'Orbread0 no longer required. Remove from template'
                  call errormsg

               else if(substring(1:8) .eq. 'Orbread:'.or.substring(1:8) .eq. 'orbread:') then
                  if (lorbread .and. lcas) then
                     klbl=index(substring,'<')
                     llbl=index(substring,'>')
                     if (klbl.gt.0 .and. llbl.gt.0) then
! write orbfile name to .inp file
                        if(substring(klbl:llbl).ne.'<orbfile>')then
                           message = 'Orbread: no <orbfile> in template'
                           call errormsg
                        endif
                        string= ' '
                        string(1:klbl-9) = substring(9:klbl-1)
                        klbl=slen(string)
                        string(klbl+1:klbl+orbflen+1) = &
                               orbfile(laenge:laenge+orbflen)
                        klbl=slen(string)
                        jlbl=slen(substring)
                        string(klbl+1:klbl+jlbl-llbl) = substring(llbl+1:jlbl)
                        llbl=slen(string)
                        write(idump,'(a)') string(1:llbl)
                        found(4)=.true.
                     else
! write other keywords associated with orbfile reading to .inp file
                        llbl=slen(substring)
                        string=substring(9:llbl)
                        llbl=slen(string)
                        write(idump,'(a)') string(1:llbl)
                     endif
                  endif

               else if(substring(1:11) .eq. 'No_orbread:') then
                  if (.not. lorbread .and. lcas) then
                     llbl=slen(substring)
                     string=substring(12:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                  endif
               else if (substring(1:6) .eq. 'Force:'.or. &
                        substring(1:6) .eq. 'force:') then
                  if (iorder .ge. 1) then
                     llbl=slen(substring)
                     string=substring(7:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                     found(5)=.true.
                  endif
               else if (substring(1:5) .eq. 'NACT:'.or. &
                        substring(1:5) .eq. 'nact:') then
! Need NACTs for diabtisation or adiabatic prop with multiple states (e.g. in TSH)
                  if (lnactdb) then
                     llbl=slen(substring)
                     string=substring(6:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                     found(6)=.true.
                  endif
               else if (substring(1:4) .eq. 'SOC:'.or. &
                        substring(1:4) .eq. 'soc:') then
                  if (lcasocdb) then
                     llbl=slen(substring)
                     string=substring(5:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                     found(7)=.true.
                  endif
               else if (substring(1:7) .eq. 'Wfread:'.or. &
                        substring(1:7) .eq. 'wfread:') then
                  if (lcas .and. (idiab.eq.5))then
!                     klbl=index(substring,'<')
!                     llbl=index(substring,'>')
!                     if(klbl.le.0 .or. llbl.le.0 .or. llbl.lt.klbl)then
!                        message = 'Wfread </> position error in template'
!                        call errormsg
!                     endif
!                     if(substring(klbl:llbl).ne.'<wfufile>')then
!                        message = 'Wfread: no <wfufile> in template'
!                        call errormsg
!                     endif
!                     string= ' '
!                     string(1:klbl-8) = substring(8:klbl-1)
!                     klbl=slen(string)
!                     string(klbl+1:klbl+wfuflen+1) &
!                          = wfufile(1:wfuflen)
!                     klbl=slen(string)
!                     jlbl=slen(substring)
!                     string(klbl+1:klbl+jlbl-llbl) = substring(llbl+1:jlbl)
!                     llbl=slen(string)
                     llbl=slen(substring)
                     string=substring(8:llbl)
                     llbl=slen(string)
                     write(idump,'(a)') string(1:llbl)
                  endif
               else
                  klbl=slen(substring)
                  llbl=slen(filename)
                  routine = 'WRMOLPRO'
                  message = 'Metastring '//substring(1:klbl)// &
                            ' not yet implemented in file '// &
                            filename(1:llbl)
                  call errormsg
               endif
            else            
               llbl=slen(string)
               write(idump,'(a)') string(1:llbl)
            endif
         else !write new coordinates
            do n=1,natm
! molpro doesn't recognise D, must use H1 and increase mass in the input file
               if(xatnam(n) .eq. 'D' .or. xatnam(n) .eq. 'd')then
                  write(idump,'(a12,3f15.6)') 'H1',&
                       (xcoo(i,n)*angstroem,i=1,3)
               else
                  write(idump,'(a12,3f15.6)') xatnam(n),&
                       (xcoo(i,n)*angstroem,i=1,3)
               endif
            enddo
         endif
      end do
 
  100 continue

!-----------------------------------------------------------------------
! check all information needed is in place
!-----------------------------------------------------------------------
      llbl=slen(filename)

      if(imethod.eq.12 .or. imethod .eq. 32) then
         if(lorbread .and. .not.found(4)) then
            routine = 'WRMOLPRO'
            message = 'Missing $Orbread:...$ metastring in file '//&
                      filename(1:llbl)
            call errormsg
         endif
      endif

      if (iorder .ge. 2 .and. .not. (lddmctdh.or.lgausspot)) then
         if (.not. found(2)) then
!            if(lhesscalc)then
               routine = 'WRMOLPRO'
               message = 'Missing $Freq:...$ meta'//&
                    'string in file ' //filename(1:llbl)
               call errormsg
!            endif
         endif
      endif
      if (iorder .ge. 1) then
         if (.not. found(5)) then
            routine = 'WRMOLPRO'
            message = 'Missing $Force:...$ meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif 

      if(idiab .eq. 3 .or. idiab .eq. 4)then
         if(.not.found(6))then
            routine = 'WRMOLPRO'
            message = 'Missing $NACT:....$ meta'//&
                 'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif

      if(lcasocdb)then
         if(.not.found(7))then
            routine = 'WRMOLPRO'
            message = 'Missing $SOC:....$ meta'//&
                 'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif

!-----------------------------------------------------------------------
! Close files
!-----------------------------------------------------------------------
      close(itmpin)
      close(idump)

!-----------------------------------------------------------------------
! Write MO coefficients to an auxiliary file if required.
!-----------------------------------------------------------------------
      lorbfile=.false.
      if (found(4) .and. .not. linitial) then
         call wrmolpro_orbfile(orbfile,mocoeff)
!         inquire(file=orbfile,exist=lorbfile)
!         if (lorbfile) then
!            open(idump,file=orbfile,status='replace')
!         else
!            open(idump,file=orbfile,status='new')
!         endif
!         write(idump,'(a)')'BEGIN_DATA'
!         write(idump,'(a)')&
!              '# MATRIX OLDORBITALS        ORBITALS NATURAL  SYMMETRY=1'
!! Loop over AOs
!         do i = 1, nbasis
!! Loop over MOs
!            koff = 0
!            do j = 1, nbasis/5
!               write(idump,'(5(f15.8,","))')(mocoeff(k+koff,i),k=1,5)
!               koff = koff + 5
!            enddo
!            if(mod(nbasis,5).ne.0)then
!               write(idump,'(4(f15.8,","))')&
!                    (mocoeff(k+koff,i),k=1,mod(nbasis,5))
!            endif
!         enddo
!         write(idump,'(a)')'END_DATA,'
!         close(idump,status='keep')
      endif

      return
  990 routine='WRMOLPRO'
      llbl=slen(inpfile)
      write(message,'(a,i5)')  &
           'Error opening file '//inpfile(1:llbl)
      call errormsg
  999 routine='WRMOLPRO'
      llbl=slen(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline
      call errormsg
      end subroutine wrmolpro

!#######################################################################
! Routine to write a file containing MO coefficients for Molpro.
! Split from wrmolpro.
! GWR 10/19
!#######################################################################

      subroutine wrmolpro_orbfile(orbfile,mocoeff)

      implicit none

      real(sip), dimension(nbasis,nbasis), intent(in):: mocoeff
      character(len=*), intent(in)                   :: orbfile

      logical(kind=4) :: lorbfile
      integer(long)   :: i,j,k,koff

      inquire(file=orbfile,exist=lorbfile)
      if (lorbfile) then
         open(idump,file=orbfile,status='replace')
      else
         open(idump,file=orbfile,status='new')
      endif
      
      write(idump,'(a)')'BEGIN_DATA'
      write(idump,'(a)')&
              '# MATRIX OLDORBITALS        ORBITALS NATURAL  SYMMETRY=1'

! Loop over AOs
      do i = 1,nbasis
! Loop over MOs
         koff = 0
         do j = 1, nbasis/5
            write(idump,'(5(e16.8,","))')(mocoeff(k+koff,i),k=1,5)
            koff = koff + 5
         enddo
         if(mod(nbasis,5).ne.0)then
             write(idump,'(4(e16.8,","))')&
                    (mocoeff(k+koff,i),k=1,mod(nbasis,5))
         endif
      enddo
      write(idump,'(a)')'END_DATA,'
      close(idump,status='keep')

      return
      end subroutine wrmolpro_orbfile

!########################################################################
!
! Q-Chem interface
!
!########################################################################
!>-----------------------------------------------------------------------
!>
!> QChem Interface
!> ---------------
!> 
!> @brief
!> main subroutine for the qchem interface
!> 
!>------------------------------------------------------------------------
      subroutine ddqchem (gpoint,v,deriv1,deriv2,hopipar1,nst,nst1,nact,lfail)

!$    use OMP_LIB

      implicit none

      logical(kind=4), intent(out)                                 :: lfail
      integer(long)                                                :: imethod,iorder,ilbl,jlbl,&
                                                                      fnum,fstate,istate,cstate
      integer(long), dimension(5)                                  :: found
      integer(long), intent(in)                                    :: nst,nst1
      integer(long), dimension(*), intent(in)                      :: hopipar1
      integer(long), external                                      :: slen
      real(dop), dimension(*), intent(in)                          :: gpoint
      real(dop), dimension(nst,*), intent(out)                     :: v
      real(dop), dimension(ndofddpes,nst,*), intent(out)           :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nst,*), intent(out) :: deriv2
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), &
      intent(out)            :: nact
      !real(sip), dimension(nbasis,nbasis), intent(inout)                     :: mocoeff
      character(len=120)                                           :: templdatfile,inpfile,filnam,nproc !geofile,egfile,hfile
      character(len=240)                                           :: extcmd,outfile


     ! Assign necessary variables

      lfail = .false.
      found = 0

      fnum = hopipar1(3)    ! Gaussian function number
      fstate = hopipar1(4)  ! State on which the Gaussian is positioned
      iorder = hopipar1(6)  ! Taylor series order
      imethod = hopipar1(7) ! Quantum chemistry method identifier
      istate = nst          ! State for which to calculate properties
      cstate = nroot        ! Overall number of states

      ! Assign file names

      templdatfile=ddname(1:ddlaenge)//'/template.dat'

      if (nst1 .eq. 1) then  ! single-state or single-set calculation
         write(filnam,'(a,2(a,i5))')&
             ddname(1:ddlaenge),'/gwp',fnum,'_V',istate
      else                   ! multi-state, multi-set calculation
         write(filnam,'(a,3(a,i5))') &
             ddname(1:ddlaenge),'/gwp',fnum,'_s',fstate,'_V',istate
      endif
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1

      inpfile=filnam(1:ilbl)//'.inp'

      ! Write input file

!$omp critical (wr)
      call wrqchem(templdatfile,qcscr,inpfile,gpoint,natm,atnam,iorder)
!$omp end critical (wr)

      ! Run Q-Chem

      jlbl=index(ddextcmd,' ')
      write(nproc,'(i3)') qcnproc
      !extcmd = ddextcmd(1:jlbl)//filnam
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//filnam(1:ilbl+1)//char(zeroi)
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//filnam(ddlaenge+2:ilbl+1)// &
              ddname(1:ddlaenge+1)//char(zeroi)
      else if (qcsubform .eq. 3) then
         extcmd=ddextcmd(1:jlbl)//filnam(ddlaenge+2:ilbl+1)// &
              ddname(1:ddlaenge+1)//nproc(1:3)//char(zeroi)
      endif

      if (.not. lqcsimulate) call excmd(extcmd)

      return
      
      end subroutine ddqchem

!>-------------------------------------------------------------------------------
!>
!> subroutine to produce a QChem input file from the template
!>
!>-------------------------------------------------------------------------------

      subroutine wrqchem(templdatfile,qcscr,inpfile,xcoo,natm,atnam,iorder)

      implicit none

      integer(long)                                  :: i,n,ilbl,jlbl
      integer(long), intent(in)                      :: natm,iorder
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: templdatfile,qcscr,inpfile !geofile
      character(len=c1), dimension(natm), intent(in) :: atnam
      character(len=120)                             :: string
      logical                                        :: iscratch,irun
      real(dop), parameter                           :: angstroem=0.52917725_dop
      integer(long), external                        :: slen

      iscratch = .false.
      irun = .false.
      ! Get the last index of the path to the DB
      ilbl = slen(inpfile)
      do
         if (inpfile(ilbl:ilbl) .eq. '/' .or. ilbl .eq. 0) exit
         ilbl = ilbl - 1
      enddo
      ! Get the last index of the scratch folder name
      jlbl=index(qcscr,' ')
      
      ! Read through the template and create the new input file
      open(itmpin,file=templdatfile)
      open(idump,file=inpfile)
      do
         read(itmpin,'(a)',err=999,end=100) string
         string = trim(adjustl(string))
         
         if (string(1:9) .eq. "$molecule") then  ! Take care of geometry input
            write(idump,'(a)') trim(string)
            if (irun) then
               write(idump,'(a)') 'read'
               do ! roll down through template geometry (can be anything)
                  read(itmpin,'(a)',err=999,end=100) string
                  string = adjustl(string)
                  if (string(1:4) .eq. "$end") exit
               enddo
            else
               read(itmpin,'(a)',err=999,end=100) string ! Read in charge and multiplicity
               write(idump,'(a)') trim(string)
               do ! roll down through template geometry (can be anything)
                  read(itmpin,'(a)',err=999,end=100) string
                  string = adjustl(string)
                  if (string(1:4) .eq. "$end") exit
               enddo
               do n=1,natm  ! Write input geometry
                  write(idump,'(a12,3f15.6)') xatnam(n),(xcoo(i,n)*angstroem,i=1,3)
               enddo
               irun=.true.
            endif
            write(idump,'(a)') "$end"


        else if (string(1:9) .eq. 'SCF_GUESS' .or. string(1:9) .eq. 'scf_guess') then ! Check if sratch folder exists
            inquire(file=inpfile(1:ilbl)//'scratch/'//qcscr(1:jlbl-1)//'/molecule',exist=iscratch)
            if (iscratch .or. irun) write(idump,'(a)') string


         else if (string(1:8) .eq. "$comment" .or. string(1:8) .eq. "$COMMENT") then ! Decide whether to run frequency calculations 
            read(itmpin,'(a)',err=999,end=100) string
            if ((iorder .eq. 2 ) .and. (string(1:4)=="FREQ" .or. string(1:4) .eq. "freq")) then
               write(idump,'(a)') '$comment'
               write(idump,'(a)') trim(string)
               irun=.true.
           elseif((iorder < 2) .and. (string(1:4)=="FREQ" .or. string(1:4) .eq. "freq")) then
               do while (string(1:3)/="@@@")
                  read(itmpin,'(a)',err=999,end=100) string
               enddo
            else
                write(idump,'(a)') "$comment"
                write(idump,'(a)') trim(string)

            endif

         else
            write(idump,'(a)') trim(string)
         endif

      enddo

  100 continue

      close(itmpin)
      close(idump)

      return

  999 routine='WRQCHEM'
      write(message,'(a)')  &
           'Error reading file '//templdatfile
      call errormsg

      end subroutine wrqchem

!####################################################################### 

      subroutine ddorca (gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                         hopipar1,lfail)

!$    use OMP_LIB

      implicit none

      logical(kind=4), intent(out) :: lfail
      integer(long) :: idiab,imethod,jlbl,fnum,fstate,iorder,cstate,&
                       istate,ilbl,iorder0,ddlaenge1,iunit,ithr
      integer(long), dimension(*), intent(in)             :: hopipar1
      integer(long), external                             :: slen
      real(dop), dimension(*), intent(in)                 :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)          :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out)          :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), intent(out):: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)        :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)  :: mocoeff
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), intent(out):: nact
      character(len=120)  :: templdatfile,keywddatfile,inpfile,outfile, &
                             filnam,filnam0,orbfile,ddname1
      character(len=740)                                  :: extcmd
      logical(kind=4)                                     :: lmultemplate,&
                                                             lcasonly,lexist

      routine = 'ddorca'

      lfail = .false.

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic (GAUSSIAN), = 1 diabatic (GAUSSIAN)
! idiab   : = 2 quasi-diabatic (GAUSSIAN+ROT), = 3 propagation diabatisation
! fnum    : no. of function
! fstate  : state to which function belongs (multi-set formalism)
! cstate  : state being calculated (running index further below)
! istate  : index of cstate without considering nroot
! iorder  : local order (or 0 for propagating CAS when DB is read)
! imethod : flag for the quantum chemistry method
! nroot   : highest state
!-----------------------------------------------------------------------
      idiab = hopipar1(2)
      fnum = hopipar1(3)
      fstate = hopipar1(4)
      iorder = hopipar1(6)
      imethod = hopipar1(7)

#if OMP
   ithr = omp_get_thread_num()
#else
   ithr = 0
#endif

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      keywddatfile=ddname(1:ddlaenge)//'/keywd.dat'
      templdatfile=ddname(1:ddlaenge)//'/template.dat'

!-----------------------------------------------------------------------
! highest electronic state
!-----------------------------------------------------------------------
      cstate = nroot
      istate = nddstate

!-----------------------------------------------------------------------
! check for existence of directories for files
!-----------------------------------------------------------------------
      lexist=.false.
      write(filnam,'(a,2(a,i5))')&
           ddname(1:ddlaenge),'/gwp',fnum,'_V',istate
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1
      iunit = iddin + ithr
      open(iunit,file=filnam(1:ilbl)//'/scratch',err=666)
      close(iunit,status='delete')
      lexist=.true.

  666 continue

      if (.not. lexist) then
         extcmd='mkdir '//filnam(1:ilbl)
         if (.not. lqcsimulate) call excmd(extcmd)
      endif
! save path as ddname1(1:ddlaenge1)
      ddname1=filnam
      ddlaenge1=ilbl

!-----------------------------------------------------------------------
! get filestem of inp file
!-----------------------------------------------------------------------
      lcasonly = .false.
      iorder0 = 0
      if (lmultemplate) then
         message = 'CAS DD in Orca with multiple templates not implemented'
         call errormsg

!      else if (nst1 .eq. 1) then
      else
         write(filnam,'(a,2(a,i5),2a)')&
             ddname1(1:ddlaenge1),'/gwp',fnum,'_v',istate
!      else
!         write(filnam,'(a,3(a,i5),2a)') &
!             ddname(1:ddlaenge),'/gwp',fnum,'s',fstate,'_v',istate
      endif
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1

!
! set up name for SACAS calc
! Copy checkpoint file for CAS calculations at start
!
      if (imethod .eq. 12.or.imethod.eq.33) then
         call rmblnk(filnam)
         ilbl=index(filnam,' ')-1
         filnam0=filnam

! Zero output arrays.
         call zerovxd(v,nddstate*nddstate)
         call zerovxd(deriv1,ndofddpes*nddstate*nddstate)
         call zerovxd(deriv2,ndofddpes*ndofddpes*nddstate*nddstate)
         nact(:,:) = 0.0_dop

      endif

!-----------------------------------------------------------------------
! generate name of com file and write it
!-----------------------------------------------------------------------
      inpfile=filnam(1:ilbl)//'.inp'
      orbfile=filnam(1:ilbl)//'.orb'
      outfile=filnam(1:ilbl)//'.out'

!$OMP CRITICAL (wr)
      call wrorca(templdatfile,inpfile,orbfile,gpoint,mocoeff,natm, &
           iorder,xatnam)
!$OMP END CRITICAL (wr)

!-----------------------------------------------------------------------
! run ORCA
!-----------------------------------------------------------------------
      jlbl=index(ddextcmd,' ')
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//inpfile//' > '//outfile
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//filnam(ddlaenge+2:ilbl+1)// &
              ddname(1:ddlaenge+1)//' > '//outfile
      endif
      if (.not. lqcsimulate) call excmd(extcmd)

      return
      end subroutine ddorca

!#######################################################################

      subroutine wrorca(templdatfile,inpfile,orbfile,xcoo,mocoeff,natm, &
                 iorder,xatnam)

      implicit none

      logical, dimension(6)                          :: found
      integer(long)                                  :: ilbl,klbl,jlbl,&
                                                        llbl,iline,i,n,&
                                                        orbflen,j,k,koff 
      integer(long), external                        :: slen
      integer(long), intent(in)                      :: natm,iorder
      real(sip), dimension(nbasis,nbasis), intent(in):: mocoeff
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: templdatfile,inpfile, &
                                                        orbfile
      character(len=120)                             :: filestem,filename,&
                                                        filestem0
      character(len=360)                             :: string,substring

      character(len=c1), dimension(natm), intent(in) :: xatnam

      integer, parameter                             :: tlines = 200 !max number of lines in template file
      real(dop), parameter                           :: angstroem = 0.52917725_dop
      logical(kind=4)                                :: lmultemplate

      routine = 'WRORCA'

!-----------------------------------------------------------------------
! see if template file is not template.dat
!-----------------------------------------------------------------------
      ilbl=index(templdatfile,' ')-1
      lmultemplate = (templdatfile(ilbl-11:ilbl) .ne. "template.dat") 

      open(iin,file=templdatfile)
      open(itmpin,file=inpfile)

!-----------------------------------------------------------------------
! get file stem of inp file
! inpfile(ilbl:jlbl) is name of inp file ignoring path
! inpfile(klbl:klbl) is position of . in filestem.inp
!-----------------------------------------------------------------------
      jlbl = slen(inpfile)
      ilbl = jlbl
      do 
         if (inpfile(ilbl:ilbl) .eq. '/' .or. ilbl .eq. 0) exit
         ilbl = ilbl - 1
      enddo
      ilbl=ilbl+1
      klbl=index(inpfile(ilbl:jlbl),'.')-1
      klbl = ilbl + klbl -1
      if (klbl .le. 0) then
         message = 'Error in file name '//inpfile(1:jlbl)
         call errormsg
      endif
      filestem=inpfile(1:klbl)
      if (ilbl .gt. 0) filestem0 = filestem(ilbl:klbl)
 
      orbflen = slen(orbfile)

!-----------------------------------------------------------------------
! write inp file
! metrastrings are used for file stem, root number, substituted links,
! and special IOps
!-----------------------------------------------------------------------
      found(:)=.false.
      filename=templdatfile
      do 
         substring = ' '
         read(iin,'(a)',err=999,end=100) string
         if (string .ne. '<geometry>') then
!            ilbl=index(string,'$')
!            if (ilbl .gt. 0) then
!               klbl=slen(string)
!               jlbl=index(string(ilbl+1:klbl),'$')
!               if (jlbl .le. 0) then
!                  llbl=slen(filename)
!                  routine = 'WRORCA'
!                  write(message,'(a,i5)') &
!                       'Missing second $ in file '//filename(1:llbl)//&
!                       ', line ',iline
!                  call errormsg
!               endif
!               jlbl=ilbl+jlbl
!               read(string(ilbl+1:jlbl-1),'(a)') substring
!               nc = jlbl-ilbl-1
!               if (substring(1:7) .eq. 'Orbrot:') then
!                  if (imethod .eq. 12) then
!                     llbl=slen(substring)
!                     string=substring(8:llbl)
!                     llbl=slen(string)
!                     write(itmp,'(a)') string(1:llbl)
!                     found(1)=.true.
!                  else
!                     llbl=slen(filename)
!                     routine = 'WRORCA'
!                     message = 'Do not use $Orbrot:...$ metastring in file ' &
!                               //filename(1:llbl)
!                     call errormsg
!                  endif
!               else if (substring(1:5) .eq. 'Freq:') then
!                  if (iorder .ge. 2) then
!                     if(.not. lupdhes .or. nrec .eq. 0)then
!                        llbl=slen(substring)
!                        string=substring(6:llbl)
!                        llbl=slen(string)
!                        write(itmp,'(a)') string(1:llbl)
!                        found(2)=.true.
!                     endif
!                  endif
!               else if(substring(1:9) .eq. 'Orbread0:') then
!                  if((imethod.eq.12) .and. nrec.eq.0)then 
!                     llbl=slen(substring)
!                     string=substring(10:llbl)
!                     llbl=slen(string)
!                     write(itmp,'(a)') string(1:llbl)
!                     found(3)=.true.
!                  endif
!               else if(substring(1:8) .eq. 'Orbread:') then
!                  if((imethod.eq.12).and.nrec.gt.0)then 
!                     klbl=index(substring,'<')
!                     llbl=index(substring,'>')
!                     if(klbl.le.0 .or. llbl.le.0 .or. llbl.lt.klbl)then
!                        message = 'Orbread </> position error in template'
!                        call errormsg
!                     endif
!                     if(substring(klbl:llbl).ne.'<orbfile>')then
!                        message = 'Orbread: no <orbfile> in template'
!                        call errormsg
!                     endif
!                     string= ' '
!                     string(1:klbl-9) = substring(9:klbl-1)
!                     klbl=slen(string)
!                     string(klbl+1:klbl+orbflen-ddlaenge-1) = &
!                          orbfile(ddlaenge+2:orbflen)
!                     klbl=slen(string)
!                     jlbl=slen(substring)
!                     string(klbl+1:klbl+jlbl-llbl) = substring(llbl+1:jlbl)
!                     llbl=slen(string)
!                     write(itmp,'(a)') string(1:llbl)
!                     found(4)=.true.
!                  endif
!               else if (substring(1:6) .eq. 'Force:') then
!                  if (iorder .ge. 1) then
!                     llbl=slen(substring)
!                     string=substring(7:llbl)
!                     llbl=slen(string)
!                     write(itmp,'(a)') string(1:llbl)
!                     found(5)=.true.
!                  endif
!               else if (substring(1:5) .eq. 'NACT:') then
!                  if (imethod .eq. 12 .and. idiab.eq.3) then
!                     llbl=slen(substring)
!                     string=substring(6:llbl)
!                     llbl=slen(string)
!                     write(itmp,'(a)') string(1:llbl)
!                     found(6)=.true.
!                  endif
!               else
!                  klbl=slen(substring)
!                  llbl=slen(filename)
!                  routine = 'WRORCA'
!                  message = 'Metastring '//substring(1:klbl)// &
!                            ' not yet implemented in file '// &
!                            filename(1:llbl)
!                  call errormsg
!               endif
!            else            
               llbl=slen(string)
               write(itmpin,'(a)') string(1:llbl)
!            endif
         else !write new coordinates
            do n=1,natm
               write(itmpin,'(a12,3f15.6)') xatnam(n),&
                    (xcoo(i,n)*angstroem,i=1,3)
            enddo
         endif
      end do
 
  100 continue

!-----------------------------------------------------------------------
! check all information needed is in place
!-----------------------------------------------------------------------
      llbl=slen(filename)

!      if(nrec.eq.0 .and. (imethod.eq.12)) then
!         if(.not.found(3)) then
!            routine = 'WRORCA'
!            message = 'Missing $Orbread0:...$ metastring in file '//&
!                      filename(1:llbl)
!            call errormsg
!         endif
!      endif
!      if(nrec.gt.0 .and. (imethod.eq.12)) then
!         if(.not.found(4)) then
!            routine = 'WRORCA'
!            message = 'Missing $Orbread:...$ metastring in file '//&
!                      filename(1:llbl)
!            call errormsg
!         endif
!      endif

      if (iorder .ge. 2) then
         if (.not. found(2)) then
            routine = 'WRORCA'
            message = 'Missing $Freq:...$ meta'//&
                 'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif
      if (iorder .ge. 1) then
         if (.not. found(5)) then
            routine = 'WRORCA'
            message = 'Missing $Force:...$ meta'//&
                         'string in file ' //filename(1:llbl)
            call errormsg
         endif
      endif 

!-----------------------------------------------------------------------
! Write MO coefficients to an auxiliary file if required.
!-----------------------------------------------------------------------
      if(found(4)) then
         open(idump,file=orbfile,status='replace')

         write(idump,'(a)')'BEGIN_DATA'
         write(idump,'(a)')&
              '# MATRIX OLDORBITALS        ORBITALS NATURAL  SYMMETRY=1'
! Loop over AOs
         do i = 1, nbasis
! Loop over MOs
            koff = 0
            do j = 1, nbasis/5
               write(idump,'(5(f15.8,","))')(mocoeff(k+koff,i),k=1,5)
               koff = koff + 5
            enddo
            if(mod(nbasis,5).ne.0)then
               write(idump,'(4(f15.8,","))')&
                    (mocoeff(k+koff,i),k=1,mod(nbasis,5))
            endif
         enddo
         write(idump,'(a)')'END_DATA,'
         close(idump,status='keep')
      endif

! end of checks

!-----------------------------------------------------------------------
! Close files and exit.
!-----------------------------------------------------------------------
      close(iin)
!      close(irefdb) !Doesn't seem to be needed here
      close(itmpin)

      return
  999 routine='WRORCA'
      llbl=slen(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline
      call errormsg
      end subroutine wrorca

!>------------------------------------------------------------------------
!> 
!> subroutine to pass the geometry to an external prog and read the output
!> energies, 1st and 2nd derivatives
!>
!>-------------------------------------------------------------------------

      subroutine ddexternal (gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                             hopipar1,nst,lfail)

      implicit none

      logical(kind=4), intent(out)                   :: lfail
      integer(long), intent(in)                      :: nst
      integer(long)                                  :: jlbl,ilbl,i,st1,&
                                                        st2,j,a1,a2,a3,&
                                                        klbl,mlbl
      integer(long), dimension(*), intent(in)        :: hopipar1
      real(dop), dimension(*), intent(in)            :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)           :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out) :: deriv1
      real(dop), dimension(3,natm,nst,nst)                           :: guu
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), &
                                                        intent(out)  :: deriv2
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)             :: mocoeff
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), &
                                                      intent(out)     :: nact
      character(len=360)                            :: filename,filepath,exl,command

      lfail = .false. !overall failure at this geometry

      ! Zero output arrays.
      v=0.0_dop
      deriv1=0.0_dop
      deriv2=0.0_dop

      ilbl=index(ddextcmd,'/',.true.)
      mlbl=index(ddname,'/',.true.)
      filename= ddname(mlbl+1:ddlaenge)//'_in_out'
      klbl=index(filename,' ')
      filepath=ddextcmd(1:ilbl)//filename(1:klbl)

      open (unit=10,file=filepath, status="replace",action="write")
       do i=1,ndofddpes
         write(10,*) gpoint(i)
       enddo
      close(10)
      
      jlbl=index(ddextcmd,' ')
      command="cd ; ./"
      !exl (execute command line) Intrinsic Function
      exl=command(1:3)//ddextcmd(1:ilbl)//command(4:7)//ddextcmd(ilbl+1:jlbl)//filename(1:klbl)
      call execute_command_line(exl)
       
      open (unit=10,file=filepath, status="old",action="read")
       do j=1,nst
         read(10,*,end=100) (v(j,i), i=1,nst)
       enddo
      do st1=1,nst
          do st2=st1,nst
             a1=1
             a2=2
             a3=3
             do i=1,natm
              read(10,*,end=100) guu(1,i,st1,st2),guu(2,i,st1,st2),guu(3,i,st1,st2)
               deriv1(a1,st1,st2) = guu(1,i,st1,st2)
               deriv1(a2,st1,st2) = guu(2,i,st1,st2)
               deriv1(a3,st1,st2) = guu(3,i,st1,st2)
               a1=a1+3
               a2=a2+3
               a3=a3+3
            enddo
          enddo
      enddo

       do st1=1,nst
          do st2=st1,nst
            do i=1,ndofddpes
              read(10,*,end=100)(deriv2(i,j,st1,st2), j=1,ndofddpes)
             enddo
          enddo
       enddo

      100 close(10)

      return

      end subroutine ddexternal

!####################################################################### 

      subroutine ddbagel (gpoint,v,deriv1,deriv2,dipole,mocoeff,nact, &
                            hopipar1,lfail,outfile)

      implicit none

      logical(kind=4), intent(out)  :: lfail
      integer(long)  :: idiab,imethod,fnum,fstate,iorder,ithr,iunit,ilbl,jlbl,&
                        klbl,llbl,ddlaenge1
      integer(long), external                             :: slen
      integer(long), dimension(*), intent(in)             :: hopipar1
      real(dop), dimension(*), intent(in)                                     :: gpoint
      real(dop), dimension(nddstate,nddstate), intent(out)                    :: v
      real(dop), dimension(ndofddpes,nddstate,nddstate), intent(out)          :: deriv1
      real(dop), dimension(ndofddpes,ndofddpes,nddstate,nddstate), intent(out):: deriv2      
      real(dop), dimension(3,nddstate,nddstate), intent(out)         :: dipole
      real(sip), dimension(nbasis,nbasis), intent(inout)  :: mocoeff
      real(dop), dimension(ndofddpes,(nddstate*(nddstate-1)/2)), &
           intent(out):: nact
      character(len=240)  :: templdatfile,filnam,loctempldatfile,&
                             templorbfile,orbfile,loctemplorbfile
      character(len=*), dimension(2)                      :: outfile
      character(len=240)                                  :: extcmd
      character(len=240)                                  :: ddname1
      logical (kind=4)                                    :: lmultemplate,lexist,&
                                                             lstartorb,larchive
      logical, dimension(5)                               :: found
      logical (kind=4), save                              :: linitial = .true.

      lfail = .false. !set initially as not fail
      found(:) = .false.
      lstartorb=.false. !bool for orbitals and archive
      larchive=.false.
      if (dbnrec .gt. 0) linitial = .false.

!-----------------------------------------------------------------------
! flags to mark input and output files for different states/functions
! and other options
! idiab   : = 0 adiabatic, = 1 regularised diabatisation,
!           = 2 local diabatisation, = 3 propagation diabatisation (old)
!           = 4 propagation diabatisation (global)
!           only 0,3 and 4 for bagel 
! fnum    : no. of function
! fstate  : state to which function belongs (multi-set formalism)
! istate  : index of state without considering nroot, i.e. index in space
!           of states in dynamics. Thus istate .ne. cstate if there is a
!           state mapping between QC and QD.
! iorder  : local order (or 0 for propagating CAS when DB is read)
! imethod : flag for the quantum chemistry method
!-----------------------------------------------------------------------
      idiab   = hopipar1(2)
      fnum    = hopipar1(3)
      fstate  = hopipar1(4)
      iorder   = hopipar1(6)
      imethod = hopipar1(7)

      ithr = 0 !variable for omp
   
!-----------------------------------------------------------------------
! look to see if multiple template files are to be used
!-----------------------------------------------------------------------
      lmultemplate=.false.
      templdatfile=ddname(1:ddlaenge)//'/template_1.dat'
      inquire(file=templdatfile,exist=lmultemplate)
      if (lmultemplate) then
        message = 'DD in Bagel with multiple templates not implemented'
        call errormsg
      endif  

!-----------------------------------------------------------------------
! get full name of dat files
!-----------------------------------------------------------------------
      templdatfile=ddname(1:ddlaenge)//'/template.dat'
      templorbfile=ddname(1:ddlaenge)//'/template_orb.dat'

! Initialize output arrays with zero
      v(:,:)=0.0_dop
      deriv1(:,:,:)=0.0_dop
      deriv2(:,:,:,:)=0.0_dop
      nact(:,:)=0.0_dop
      dipole(:,:,:)=0.0_dop

!-----------------------------------------------------------------------
! check for existence of directories for files
!-----------------------------------------------------------------------      
      lexist=.false.
      write(filnam,'(a,2(a,i5))') ddname(1:ddlaenge),'/gwp',fnum
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1
      iunit=iddin+ithr
      open(iunit,file=filnam(1:ilbl)//'/scratch',err=666)
      close(iunit,status='delete')
      lexist=.true.
666   continue
      if (lexist .and. .not. lddwrdb) then
         if (.not. lqcsimulate) then
            if (lovwr) then
               extcmd='rm -rf '//filnam(1:ilbl)//'/*'
               call excmd(extcmd)
            else
               routine='DDbagel (funcqchem.f90)'
               write(message(1:69),'(a)') &
      'Calculation does not use DB yet DB Bagel directories (and files?) '
               write(message(70:122),'(a)') &
      'already exist. Delete DB/gwp* directories and restart'
               call errormsg
            endif
         endif
      else if (.not. lexist) then
         extcmd='mkdir '//filnam(1:ilbl)
         call excmd(extcmd)
      endif     
      ! save path as ddname1(1:ddlaenge1)
      ddname1=filnam
      ddlaenge1=ilbl

!-----------------------------------------------------------------------
! get filestem of input and output file
! add directory name to distinguish from potentially simultaneous calcs
!----------------------------------------------------------------------- 
      klbl=ddlaenge1
      write(filnam,'(2a,i5,2a)')&
         ddname1(1:klbl),'/gwp',fnum,'_',ddname(ddlaenge-dbdlength+1:ddlaenge)
      call rmblnk(filnam)
      ilbl=index(filnam,' ')-1

!-----------------------------------------------------------------------
! Copy template in each directory
!-----------------------------------------------------------------------
      write(loctempldatfile,'(a,a)')&
         ddname1(1:ddlaenge1),'/template.dat'
      extcmd='cp '//trim(templdatfile)//' '//&
         trim(loctempldatfile)//char(zeroi)
      call excmd(extcmd)
      write(loctemplorbfile,'(a,a)')&
         ddname1(1:ddlaenge1),'/template_orb.dat'
      extcmd='cp '//trim(templorbfile)//' '//&
         trim(loctemplorbfile)//char(zeroi)
      call excmd(extcmd)

!-----------------------------------------------------------------------
! Copy archive file for CAS calculations at start
!-----------------------------------------------------------------------
      if (linitial .and. lorbread0) then
         inquire(file=ddname(1:ddlaenge)//'/start.archive',exist=larchive)
         if (.not. larchive) inquire(file=ddname(1:ddlaenge)//'/start.molden',exist=lstartorb)
         if (.not. larchive .and. .not. lstartorb) then
            routine = 'DDBAGEL'
            message = 'No start.archive or start.molden file provided in DB directory'
            call errormsg
         else if (larchive) then
            extcmd='cp '//ddname(1:ddlaenge)//'/start.archive '// &
               ddname1(1:ddlaenge1)//'/start.archive'//char(zeroi)
            if (.not. lqcsimulate) call excmd(extcmd)
         else
            extcmd='cp '//ddname(1:ddlaenge)//'/start.molden '// &
               filnam(1:ilbl)//'.molden'//char(zeroi)
            if (.not. lqcsimulate) call excmd(extcmd)
         endif
      endif

!-----------------------------------------------------------------------
! generate name of input file and write it
!-----------------------------------------------------------------------
      inputfile=filnam(1:ilbl)//'.json'
      orbfile=filnam(1:ilbl)//'.molden'
      call wrbagel(loctempldatfile,inputfile,loctemplorbfile,orbfile,&
           gpoint,mocoeff,nbasis,natm,idiab,imethod,iorder,atnam,linitial,&
           larchive)

!-----------------------------------------------------------------------
! run BAGEL
!-----------------------------------------------------------------------
      jlbl=index(ddextcmd,' ')
      if (qcsubform .eq. 1) then
         extcmd=ddextcmd(1:jlbl)//filnam(1:ilbl+1)//char(zeroi)
      else if (qcsubform .eq. 2) then
         extcmd=ddextcmd(1:jlbl)//filnam(klbl+2:ilbl+1)// &
              ddname1(1:klbl+1)//char(zeroi)
      endif
      if (.not. lqcsimulate) call excmd(extcmd)

!-----------------------------------------------------------------------
! generate name of out file and read it
!-----------------------------------------------------------------------
      outfile(1)=filnam(1:ilbl)//'.out'
      klbl= index(outfile(1),'/',back=.true.)+1
      llbl= index(outfile(1),'.',back=.true.)-1
      outfile(2)=filnam(1:ilbl)//'/'//outfile(1)(klbl:llbl)//'.molden'
      !call rdbagel(v,deriv1,nddstate,ddname1,outfile,outorbfile,iorder,&
      !     natm,deriv2,dipole,ndofddpes,found,lfail,mocoeff,nbasis,nact,atnam)

!-----------------------------------------------------------------------
! check energy calculation ran O.K.
!-----------------------------------------------------------------------
      !if (.not.found(1)) then
      !   message = 'Energy not found on file '//filnam(1:ilbl)//&
      !             '/ENERGY.out'
      !   lfail = .true.
      !   write(ilog,'(a)')message
      !endif
      !if(.not.found(4)) then
       !  message = 'No orbitals found on file '//trim(outorbfile)
       !  lfail = .true.
      !   write(ilog,'(a)')message
      !endif
  
!-----------------------------------------------------------------------
! if derivatives to be calculated, check they were found
!-----------------------------------------------------------------------
      !llbl=slen(outfile)
      !if (iorder .ge. 1) then
      !   if (.not.found(2)) then
      !      lfail = .true.
      !      message = 'Gradient not found on file '//filnam(1:ilbl)//&
      !                '/FORCE_X.out'
      !      write(ilog,'(a)')message
      !   endif
      !   if (iorder .ge. 2 .and. .not.found(3)) then
      !      lfail = .true.
      !      message = 'Force constants not found on file '// &
      !                outfile(1:llbl)
      !      write(ilog,'(a)')message
      !   endif
      !endif
!-----------------------------------------------------------------------
! Check that deriative couplings are present if we want to use the 
! propagation diabatisation.
!-----------------------------------------------------------------------
      !if(.not.found(5) .and. ldiabdb) then
      !   lfail = .true.
      !   message = &
      !       'Global diabatisation requested but no deriv. couplings '// &
      !        filnam(1:ilbl)//'/NACME_X_Y.out'
      !   write(ilog,'(a)')message
      !endif

!-----------------------------------------------------------------------
! Return any failures.
!-----------------------------------------------------------------------
      !if(lfail)then
      !   message = 'Electronic structure failure in '//outfile
      !   write(ilog,'(a)')message
      !endif

!-----------------------------------------------------------------------
! add origin correction to potential energy
! and report if failure in calculation
!-----------------------------------------------------------------------
      !do istate=1,nddstate
      !   v(istate,istate)=v(istate,istate)/ev-dde0
      !enddo
      
      return

      end subroutine ddbagel

!##########################################################################

      subroutine wrbagel(loctempldatfile,inpfile,loctemplorbfile,orbfile,&
                 xcoo,mocoeff,nbasis,natm,idiab,imethod,iorder,xatnam,linitial,&
                 larchive)

      implicit none

      logical, dimension(7)                          :: found
      integer(long)                                  :: i,j,n,nthreads,&
                                                        ilbl,jlbl,klbl,llbl,iline
      integer(long), intent(in)                      :: natm,idiab,imethod,&
                                                        iorder,nbasis
      real(sip), dimension(nbasis,nbasis),intent(in) :: mocoeff
      real(dop), dimension(3,natm), intent(in)       :: xcoo
      character(len=*), intent(in)                   :: loctempldatfile,&
                                                        inpfile,loctemplorbfile,&
                                                        orbfile
      character(len=c1), dimension(natm), intent(in) :: xatnam
      logical(kind=4), intent(in)                    :: linitial,larchive                                                  
      character(len=120)                             :: filestem,filestem0,&
                                                        filename,orbname
      character(len=360)                             :: string,substring,ssubstring
      integer(long), dimension(natm)                 :: iatnam
      logical                                        :: ltag

      real(dop), parameter                           :: angst=0.52917725_dop

      routine = 'WRBAGEL'

      nthreads=1
!-----------------------------------------------------------------------
! Open the input and template file
!-----------------------------------------------------------------------
      ilbl=index(loctempldatfile,' ')-1
      open(itempl,file=loctempldatfile)
      open(iinput,file=inpfile)
      open(itmorb,file=loctemplorbfile)
      open(iiorb,file=orbfile)

!-----------------------------------------------------------------------
! get file stem of inp file
! inpfile(ilbl:jlbl) is name of inp file ignoring path
! inpfile(klbl:klbl) is position of . in filestem.inp
!-----------------------------------------------------------------------
      jlbl = LEN_TRIM(inpfile)
      ilbl = index(inpfile,'/',back=.true.)+1
      klbl = index(inpfile,'.',back=.true.)-1
      if (klbl .le. 0) then
         message = 'Error in file name'//inpfile(1:jlbl)
         call errormsg
      endif
      filestem=inpfile(1:klbl)
      if (ilbl .gt. 0) filestem0=filestem(ilbl:klbl)

      write(orbname,'(a,a)') filestem0(1:klbl-ilbl+1),'.molden'


!-----------------------------------------------------------------------
! write inp file
!-----------------------------------------------------------------------
      found(:)=.false.
      filename=loctempldatfile

      do
         substring = ' '
         ltag=.false.
         read(itempl,'(a)',err=999,end=100) string
! look for <geometry/>. Only write it in json input file if 1st calculation.
! the rest of the geometry section is encapsulated with orbread0 to make it easier
! Write in molden orbfile otherwise. Only write geo if archive
         call parsexml(string,'geometry',substring,ilbl,ltag)
         if (ltag ) then
            found(1)=.true.
            if (linitial .and. larchive) then   
               write(iinput,'(a)') '"angstrom":true,'
               write(iinput,'(a)') '"geometry":['
               do n=1,natm-1
                  write(iinput,'(3a,3(f15.6,a))') '{"atom":"',trim(xatnam(n)),&
                     '","xyz":[',xcoo(1,n)*angst,',',xcoo(2,n)*angst,',',xcoo(3,n)*angst,']},'
               enddo
               n=natm
               !Last line without the ,
               write(iinput,'(3a,3(f15.6,a))') '{"atom":"',trim(xatnam(n)),&
                     '","xyz":[',xcoo(1,n)*angst,',',xcoo(2,n)*angst,',',xcoo(3,n)*angst,']}'
               write(iinput,'(a)') ']'
            endif
            cycle
         endif
         
! look for <Freq>,</Freq> pair
         call parsexml(string,'Freq',substring,ilbl,ltag)
         if (ltag ) then
           found(2)=.true.
            if (iorder.ge.2 .and. (.not.lupdhes .or. dbnrec.eq.0)) then
              write(iinput,'(a)') trim(substring)
            endif

           cycle
         endif   

! look for <Force>,</Force> pair
         call parsexml(string,'Force',substring,ilbl,ltag)
         if (ltag) then
            found(3)=.true.
            if (iorder.ge.1) then
              write(iinput,'(a)') trim(substring)
            endif

           cycle
        endif   

! look for <NACT>,</NACT> pair
         call parsexml(string,'NACT',substring,ilbl,ltag)
         if (ltag) then
            found(4)=.true.
            if (idiab.eq.3 .or. idiab .eq. 4 .or. &
                       (idiab.eq.0 .and. nddstate .gt. 0)) then
               write(iinput,'(a)') trim(substring)
            endif

           cycle
        endif
      
! look for <Orbread0>,</Orbread0> pair (use for archive)
         call parsexml(string,'Orbread0',substring,ilbl,ltag)
         if (ltag) then
            found(5)=.true.
            if (dbnrec.eq.0 .and. larchive) then
               call parsexml(substring,'orbfile',ssubstring,jlbl,ltag)
               if (ltag) then
                  klbl=len_trim(substring)
                  ssubstring=substring(1:jlbl-1)//trim(orbname)//substring(jlbl+10:klbl)
                  write(iinput,'(a)') trim(ssubstring)
               else
                  write(iinput,'(a)') trim(substring)
               endif
            endif

            cycle
         endif

! look for <Orbread>,</Orbread> pair (use for molden)
         call parsexml(string,'Orbread',substring,ilbl,ltag)
         if (ltag) then
            found(6)=.true.
            if (.not. larchive) then
               call parsexml(substring,'orbfile',ssubstring,jlbl,ltag)
               if (ltag) then
                  klbl=len_trim(substring)
                  ssubstring=substring(1:jlbl-1)//trim(orbname)//substring(jlbl+10:klbl)
                  write(iinput,'(a)') trim(ssubstring)
               else
                  write(iinput,'(a)') trim(substring)
               endif
            endif

            cycle
         endif
         
! write any other line in template verbatin
         write(iinput,'(a)') trim(string)

      enddo
  100 continue
      close(iinput)
      close(itempl)
!-----------------------------------------------------------------------
! write molden input file
!-----------------------------------------------------------------------
      filename=loctemplorbfile
      if (.not. larchive) then
         !Get the atomic number for atoms
         do n=1,natm
            iatnam(n)=getiatsym_readqc(xatnam(n),ilog)
         enddo
         do
            substring = ' '
            ltag=.false.
            read(itmorb,'(a)',err=999,end=200) string
            ! look for <geometry></geometry> in orb template file
            call parsexml(string,'geometry',substring,ilbl,ltag)
            if (ltag) then
               found(1)=.true.
               do n=1,natm
                  write(iiorb,'(a2,i8,5x,i16,1x,3(E19.12E2,1x))') trim(xatnam(n)),n,&
                     iatnam(n),(xcoo(i,n)*angst,i=1,3)
               enddo
               cycle
            endif

            !look for <mo-coefficient></mo-coefficient> in orb template file
            call parsexml(string,'mo-coefficient',substring,ilbl,ltag)
            if (ltag) then
               found(7)=.true.
               do i=1,nbasis
                  write(iiorb,'(a,f12.6)') 'Ene=',0.0_dop
                  write(iiorb,'(a)') 'Spin=  Alpha'
                  write(iiorb,'(a,f12.6)') 'Occup=',0.0_dop
                  do j=1,nbasis
                     write(iiorb,'(i4,f23.16)') j,mocoeff(i,j)
                  enddo 
               enddo
               cycle
            endif

            ! write any other line in template verbatin
            write(iiorb,'(a)') trim(string) 
         enddo
      else
         !If using archive. No need to write mo-coefficient
         found(7)=.true.
      endif
  200 continue
      close(iiorb)
      close(itmorb)
      return
  999 routine='WRBAGEL'
      llbl=len_trim(filename)
      write(message,'(a,i5)')  &
           'Error reading file '//filename(1:llbl)// &
           ', line ',iline
      call errormsg
      end subroutine wrbagel

!##########################################################################
!
! output info to log file
!
!##########################################################################

      subroutine qcheminfo (hl)

      implicit none

      integer(long), intent(in)                  :: hl
!
! log information (more is written into op.log)
!
      if (lgwptrj) then
         write(ilog,'(a)') 'Precomputed trajectories provide potential'
      else
         write(ilog,'(a)') 'External QChem program used for potential'
      endif
      if (hopilab(hl) .eq. 1) &
         write(ilog,'(a)') 'Quantum chemistry program : GAUSSIAN'
      if (hopilab(hl) .eq. 2) &
         write(ilog,'(a)') 'Quantum chemistry program : QoMMMa'
      if (hopilab(hl) .eq. 3) &
         write(ilog,'(a)') 'Quantum chemistry program : Molpro'
      if (hopilab(hl) .eq. 4) &
         write(ilog,'(a)') 'Quantum chemistry program : QChem'
      if (hopilab(hl) .eq. 5) &
         write(ilog,'(a)') 'Quantum chemistry program : MOLCAS'
      if (hopilab(hl) .eq. 6) &
         write(ilog,'(a)') 'Quantum chemistry program : ORCA'
      if (hopilab(hl) .eq. 7) &
         write(ilog,'(a)') 'Quantum chemistry program : External'
      if (hopilab(hl) .eq. 8) &
         write(ilog,'(a)') 'Quantum chemistry program : Bagel'
      return

      if (hopfile(hl) .eq. 210) then
         write(ilog,'(a)') 'Complex PES including SOCs'
      endif

      end subroutine qcheminfo

!########################################################################
! Routine for testing DD-vMCG by calling an analytic potential function
! GWR 01/18
!########################################################################

      subroutine ddvmcgpestest(idtest,xgpoint,adv,adder1,adder2,&
           ndof1,nst,hopipar6)

      implicit none

      integer(long), intent(in)                        :: idtest,ndof1,&
                                                          hopipar6,nst
      real(dop), dimension(ndof1), intent(in)          :: xgpoint
      real(dop), dimension(nst,nst), intent(out)       :: adv
      real(dop), dimension(ndof1,nst,nst), intent(out) :: adder1
      real(dop), dimension(ndof1,ndof1,nst,nst), intent(out) :: adder2

      integer(long) :: f,f1,fdx

      real(dop), dimension(ndof1*(ndof1+1)/2) :: tmpder2

      routine = 'ddvmcgpestest'

      if(idtest.eq.1)then
         if(ndof1.ne.27)then
            message = 'Wrong number of Cartesians for malonaldehyde'
            call errormsg
         endif
         call malonpes(xgpoint,hopipar6,adv,adder1,tmpder2)   

         fdx = 0
         do f = 1,ndof1
            do f1 = 1,f
               fdx = fdx + 1
               adder2(f,f1,1,1) = tmpder2(fdx)
            enddo
         enddo
     
         do f = 1,ndof1
            do f1 = 1,f-1
               adder2(f1,f,1,1) = adder2(f,f1,1,1)
            enddo
         enddo
      else
         message = 'Unrecognised potential function'
         call errormsg
      endif

      return
      end subroutine ddvmcgpestest
         
!##########################################################################
! Routine to transform CASSCF CI vectors from those generated using natural
! orbitals to equivalent vectors expressed in terms of diabatised AS MOs.
! GWR 02/20
! Original version. 03/20
! Updated to decompose CSF wavefunctions to the SD basis before performing
! the transformation.
! GWR 05/20
! Parallelised.
! GWR 12/22 
!##########################################################################

      subroutine citrans(motransf,nciconf,ciconfig,cicoeff)

!$ use omp_lib

      implicit none

      real(dop), intent(in), dimension(ncas,ncas)   :: motransf

      character(2), parameter :: backsl='\\'
      logical(kind=4) :: lcsf
      logical(kind=4), dimension(:,:), allocatable :: lcalca,lcalcb
      integer(long) :: sm,nc,sp,nrow,ncol,idx,i,j,k,m,n,s,adxn,bdxn,adx,bdx,&
           ntota,ntotb,ierr,mdx,ndx,nst,nend,iel,nsd,jdx
      integer(long), dimension(2) :: confoff
      integer(long), dimension(:), allocatable :: occan,occbn,occa,occb
      integer(long), dimension(:,:), allocatable :: confstr
      integer(long), dimension(:,:,:), allocatable :: graph
      real(dop), allocatable, dimension(:,:) :: aminor,bminor
      real(dop), allocatable, dimension(:,:,:) :: asum
      real(dop), dimension(ncas,ncas) :: tmpdet
      real(dop), dimension(5*ncas) :: work
      real(dop), dimension(ncas) :: tau,wr,wi,z
      complex(dop) :: tmpc
      integer(long), dimension(nsmult), intent(in) :: nciconf
      character(len=*), dimension(maxcivec,nsmult),intent(inout)   :: ciconfig
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(inout) :: cicoeff
      type(civec) :: tmpcivec
      type(csfdet), pointer :: csfdetptr,icsfdetptr
      type(civec), dimension(:), allocatable :: civecs
      type(civec), pointer :: civecptr
      type(civec), target :: civectgt
      character(len=40) :: cerr

! Timing
      integer(long), save   :: tid = 0 ! timer ID
      call get_timer(tid, 'citrans')
      call start_timer(tid)

! Allocate and fill ci vector structure
      cerr = 'civecs'
      allocate(civecs(nsmult),stat=ierr)
      if (ierr > 0) go to 1000
      do n = 1,nsmult
         write(cerr,'(a,i6)') 'civecs(n)%config',n
         allocate(civecs(n)%config(maxcivec),stat=ierr)
         if (ierr > 0) go to 1000
         write(cerr,'(a,i6)') 'civecs(n)%coeff',n
         allocate(civecs(n)%coeff(maxcivec,maxsta),stat=ierr)
         if (ierr > 0) go to 1000
      enddo
      civecs(:)%nconf = 0
      do n = 1,nsmult
         civecs(n)%config(:) = ' '
         civecs(n)%coeff(:,:) = 0.0_dop
      enddo
      do n = 1,nsmult
         nc = nciconf(n)
         nst = imultblk(n)
         civecs(n)%nconf = nc
         civecs(n)%config(1:nc) = ciconfig(1:nc,n)
         civecs(n)%coeff(1:nc,1:nst) = cicoeff(1:nc,1:nst,n)
      enddo

! Allocate occupancy arrays.
      if(allocated(occa))deallocate(occa)
      allocate(occa(ncas))
      if(allocated(occb))deallocate(occb)
      allocate(occb(ncas))
      if(allocated(occan))deallocate(occan)
      allocate(occan(ncas))
      if(allocated(occbn))deallocate(occbn)
      allocate(occbn(ncas))

! Check whether we are dealing with CSFs or SDs just by checking how Molpro 
! outputs the configurations. May be different.
      lcsf = .false.
      csf_loop: do sm = 1,nsmult
         do i = 1,civecs(sm)%nconf
            do n = 1,ncas
               lcsf = lcsf .or. civecs(sm)%config(i)(n:n).eq.'/'&
                    .or. civecs(sm)%config(i)(n:n).eq.backsl &
                    .or. civecs(sm)%config(i)(n:n).eq.'\'
            enddo
            if(lcsf)exit csf_loop
         enddo
      enddo csf_loop

      if(associated(csfdetptr))nullify(csfdetptr)

! Loop over spin multiplicities.
      do sm = 1, nsmult
         tmpcivec%nconf = civecs(sm)%nconf
         nc = tmpcivec%nconf

! Allocate temporary CI vectors. Allocate an extra entry for use when reordering the
! new vectors.
         if(allocated(tmpcivec%config))deallocate(tmpcivec%config)
         allocate(tmpcivec%config(nc))
         if(allocated(tmpcivec%coeff))deallocate(tmpcivec%coeff)
         allocate(tmpcivec%coeff(nc,imultblk(sm)))
!         tmpcivec%config(1:nc) = civecs(sm)%config(1:nc)
!         tmpcivec%coeff = 0.0_dop

! Allocate logical and minor arrays.
         if(allocated(lcalca))deallocate(lcalca)
         allocate(lcalca(nconf(sm,1),nconf(sm,1)))
         lcalca = .true.
         if(allocated(lcalcb))deallocate(lcalcb)
         allocate(lcalcb(nconf(sm,2),nconf(sm,2)))
         lcalcb = .true.
         if(allocated(aminor))deallocate(aminor)
         allocate(aminor(nconf(sm,1),nconf(sm,1)))
         if(allocated(bminor))deallocate(bminor)
         allocate(bminor(nconf(sm,2),nconf(sm,2)))
! asum re-allocated below in OMP loop
         if(allocated(asum))deallocate(asum)
         allocate(asum(nconf(sm,1),nconf(sm,2),nddstate))

! Setup the graphs that will allow us to index configurations.
         nrow = ncas-minval(ncasel(sm,1:2))+1
         ncol = maxval(ncasel(sm,1:2))+1
         if(allocated(graph))deallocate(graph)
         allocate(graph(nrow,ncol,2))
         if(allocated(confstr))deallocate(confstr)
         allocate(confstr(ncas,nconf(sm,1)+nconf(sm,2)))
         confstr = 0
         confoff(1) = 1
         confoff(2) = nconf(sm,1)+1

         do sp = 1,2
            call setgraph(graph(1,1,sp),nrow,ncol,ncas,ncasel(sm,sp))
! We need to know all possible alpha and beta configurations so generate them here.
            nend = ncas-ncasel(sm,sp)+1
            idx = 1 
            do i = 1,nend
               iel = 1
               nst = i
               confstr(i,idx+confoff(sp)-1) = 1
               call confgen(confstr(1,confoff(sp)),ncas,ncasel(sm,sp),nconf(sm,sp),&
                    nst,nend,idx,iel)
            enddo
         enddo
         nsd = nconf(sm,1)*nconf(sm,2)

! Decompose the CI vectors for each geometry down into their component SDs if using CSFs.
         if(lcsf)then
! Allocate arrays.
            civecptr => civectgt
            civecptr%nconf = nsd
            if(allocated(civecptr%config))deallocate(civecptr%config)
            allocate(civecptr%config(nsd))
            civecptr%config(1:nsd) = ' '
            if(allocated(civecptr%coeff))deallocate(civecptr%coeff)
            allocate(civecptr%coeff(nsd,imultblk(sm)))
            civecptr%coeff(1:nsd,1:imultblk(sm)) = 0.0_dop
            
! Loop over the CSFs.
            do i = 1,tmpcivec%nconf
               if(civecs(sm)%config(i).eq.' ' &
                    .or. maxval(abs(civecs(sm)%coeff(i,1:imultblk(sm)))).lt.1.0e-16_dop) cycle
! Set up linked lists.
               nullify(icsfdetptr)
               allocate(icsfdetptr)
               icsfdetptr%config = ' '
               icsfdetptr%coeff = 999999.0_dop
               if(associated(csfdetptr))then
                  do while (associated(csfdetptr%next))
                     csfdetptr => csfdetptr%next
                  enddo
                  do while (associated(csfdetptr%prev))
                     csfdetptr => csfdetptr%prev
                     nullify(csfdetptr%next%prev)
                     deallocate(csfdetptr%next)
                     nullify(csfdetptr%next)
                  enddo
               endif
               csfdetptr => icsfdetptr
               nullify(csfdetptr%prev)
               nullify(csfdetptr%next)
! Decompose the CSF into its constituent SDs.
               call csftodet(civecs(sm)%config(i),confstr,nconf(sm,1)+nconf(sm,2),&
                    confoff,sm,csfdetptr)
               do while (associated(csfdetptr))
                  occa = 0
                  occb = 0
                  do sp = 1,2
! Which CAS orbitals are occupied in the configurations?
                     do n = 1,ncas
                        if(sp.eq.1)then
                           if(csfdetptr%config(n:n).eq.'2' &
                                .or. csfdetptr%config(n:n).eq.'a') then
                              occa(n) = 1
                           endif
                        else
                           if(csfdetptr%config(n:n).eq.'2' &
                                .or. csfdetptr%config(n:n).eq.'b') then
                              occb(n) = 1
                           endif
                        endif
                     enddo
                     if(sp.eq.1)then
                        idx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occa,graph(1,1,sp))
                     else
                        jdx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occb,graph(1,1,sp))
                     endif
                  enddo
                  idx = (idx-1)*nconf(sm,2)+jdx
                  civecptr%config(idx) = csfdetptr%config
                  do s = 1,imultblk(sm)
                     civecptr%coeff(idx,s) = civecptr%coeff(idx,s) &
                          + csfdetptr%coeff * civecs(sm)%coeff(i,s)
                  enddo
                  if(associated(csfdetptr%next))then
                     csfdetptr => csfdetptr%next
                  else
                     exit
                  endif
               enddo
            enddo
            
! Re-order the CI vectors in reducing maximum coefficient order, as output by Molpro.
            call smoothsort(civecptr%coeff,civecptr%config,imultblk(sm))
! Ensure that the phase is as Molpro, i.e. the largest coefficient for each state is positive.
            do s = 1,imultblk(sm)
               idx = maxloc(abs(civecptr%coeff(1:nsd,s)),1)
               if(civecptr%coeff(idx,s).lt.0.0_dop)&
                    civecptr%coeff(1:nsd,s) = -civecptr%coeff(1:nsd,s)
            enddo
            
            civecs(sm) = civecptr
!         else
!            tmpcivec%config(1:nc) = civecs(sm)%config(1:nc)
!            tmpcivec%coeff = 0.0_dop
!            civecptr => dbcivec(sm,iprox)
         endif
         tmpcivec%config(1:nc) = civecs(sm)%config(1:nc)
         tmpcivec%coeff = 0.0_dop

! test

! Loop to calculate minors of MO overlaps.
         do sp = 1,2
! Avoid repetition if the alpha and beta spaces are the same.
            if(ncasel(sm,1).eq.ncasel(sm,2).and.sp.eq.2)cycle

! Get the index of this configuration.
!            adx = confidx(nrow,ncol,ncas,ncasel(sm,1),occa,graph(1,1,1))
!            bdx = confidx(nrow,ncol,ncas,ncasel(sm,2),occb,graph(1,1,2))

            ntota = sum(confstr(1:ncas,confoff(sp)))

            do i = 1,nconf(sm,sp)
               adx = confidx(nrow,ncol,ncas,ncasel(sm,sp),confstr(1,confoff(sp)+i-1),graph(1,1,sp))
               do j = 1,nconf(sm,sp)
                  adxn = confidx(nrow,ncol,ncas,ncasel(sm,sp),confstr(1,confoff(sp)+j-1),graph(1,1,sp))
 
! If the necessary minor of the MO transformation matrix has not been calculated,
! calculate it by finding the eigenvalues and taking their product.
!                  if(lcalca(adx,adxn))then
                  tmpdet = 0.0_dop
                  mdx = 0
                  do m = 1,ncas
                     if(confstr(m,confoff(sp)+j-1).eq.0)cycle
                     mdx = mdx + 1
                     ndx = 0
                     do n = 1,ncas
                        if(confstr(n,confoff(sp)+i-1).eq.0)cycle
                        ndx = ndx + 1
                        tmpdet(ndx,mdx) = motransf(n,m)
                     enddo
                  enddo
                  call dgehrd(ntota,1,ntota,tmpdet,ncas,tau,work,5*ncas,ierr)
                  if(ierr.ne.0)then
                     routine='citrans'
                     write(message,'(a,i6)')'dgehrd error reducing minor:',ierr
                     call errormsg
                  endif
                  call dhseqr('E','N',ntota,1,ntota,tmpdet,ncas,wr,wi,z,1,work,5*ncas,ierr)
                  if(ierr.ne.0)then
                     routine='citrans'
                     write(message,'(a,i6)')'dhseqr error finding eigenvalues:',ierr
                     call errormsg
                  endif
            
                  tmpc = cmplx(wr(1),wi(1),dop)
                  do n = 2,ntota
                     tmpc = tmpc*cmplx(wr(n),wi(n),dop)
                  enddo
                  if(sp.eq.1)then
                     aminor(adx,adxn) = dble(tmpc)
                     lcalca(adx,adxn) = .false.
                     if(ncasel(sm,1).eq.ncasel(sm,2))then
                        bminor(adx,adxn) = dble(tmpc)
                        lcalcb(adx,adxn) = .false.
                     endif
                  else
                     bminor(adx,adxn) = dble(tmpc)
                     lcalcb(adx,adxn) = .false.
                  endif
               enddo
            enddo
         enddo

! test

! Loop over the configurations in the new vectors.
!$omp parallel num_threads(ompthread) if(lompthread) &
!$omp private(i,j,s,k,n,occa,occb,occan,occbn,ntota,ntotb,adx,bdx,adxn,bdxn,asum) 
         if(lompthread)then
            if(allocated(asum))deallocate(asum)
            allocate(asum(nconf(sm,1),nconf(sm,2),nddstate))
         endif
!$omp do
         do i = 1,nc
! What are the occupancies of this configuration, alpha and beta?
            occan = 0
            occbn = 0
            ntota = 0
            ntotb = 0
            do n = 1,ncas
               if(tmpcivec%config(i)(n:n).eq.'2' &
                    .or. tmpcivec%config(i)(n:n).eq.'a'&
                    .or. tmpcivec%config(i)(n:n).eq.'/') then
                  occan(n) = 1
                  ntota = ntota + 1
               endif
               if(tmpcivec%config(i)(n:n).eq.'2' &
                    .or. tmpcivec%config(i)(n:n).eq.'b'&
                    .or. tmpcivec%config(i)(n:n).eq.backsl&
                    .or. tmpcivec%config(i)(n:n).eq.'\') then
                  occbn(n) = 1
                  ntotb = ntotb + 1
               endif
            enddo

! If the configuration was not included in the Molpro output, all occupancies 
! will be 0, so cycle as the coefficient is likewise.
            if(maxval(occan).lt.1.or.maxval(occbn).lt.1)cycle

! Get the indices of the configurations.
            adxn = confidx(nrow,ncol,ncas,ncasel(sm,1),occan,graph(1,1,1))
            bdxn = confidx(nrow,ncol,ncas,ncasel(sm,2),occbn,graph(1,1,2))

! Loop over the alpha contributions and contract the indices with that of the 
! "old" coefficients.
            asum = 0.0_dop
            do j = 1,nc
               if(maxval(abs(civecs(sm)%coeff(j,1:imultblk(sm)))).lt.1.0e-8_dop)exit

! What are the occupancies of this alpha configuration?
               occa = 0
               occb = 0
               do n = 1,ncas
                  if(tmpcivec%config(j)(n:n).eq.'2' &
                       .or. tmpcivec%config(j)(n:n).eq.'a'&
                       .or. tmpcivec%config(j)(n:n).eq.'/') &
                       occa(n) = 1
                  if(tmpcivec%config(j)(n:n).eq.'2' &
                       .or. tmpcivec%config(j)(n:n).eq.'b'&
                       .or. tmpcivec%config(j)(n:n).eq.backsl &
                       .or. tmpcivec%config(j)(n:n).eq.'\') &
                       occb(n) = 1
               enddo

! If the configuration was not included in the Molpro output, all occupancies 
! will be 0, so cycle as the coefficient is likewise.
               if(maxval(occa).lt.1.or.maxval(occb).lt.1)cycle

! Get the index of this configuration.
               adx = confidx(nrow,ncol,ncas,ncasel(sm,1),occa,graph(1,1,1))
               bdx = confidx(nrow,ncol,ncas,ncasel(sm,2),occb,graph(1,1,2))
! Loop over the states.
               do s = 1,imultblk(sm)
                  asum(adxn,bdx,s) = asum(adxn,bdx,s) + aminor(adx,adxn)*civecs(sm)%coeff(j,s)
               enddo
            enddo ! j loop

! Loop over the beta minors.
            do k = 1,nconf(sm,2)
! What are the occupancies of this beta configuration?
               occb = confstr(1:ncas,k+confoff(2)-1)
! Get the index of this configuration.
               bdx = confidx(nrow,ncol,ncas,ncasel(sm,2),occb,graph(1,1,2))
! Multiply the beta parts by the alpha minor to get the contribution to the new coefficient.
               do s = 1,imultblk(sm)
                  tmpcivec%coeff(i,s) = tmpcivec%coeff(i,s) + asum(adxn,bdx,s)*bminor(bdx,bdxn)
               enddo
            enddo ! k loop
         enddo ! i loop
!$omp enddo
         if(lompthread .and. allocated(asum))deallocate(asum)
!$omp end parallel

! Re-order the CI vectors in reducing maximum coefficient order, as output by Molpro.
         call smoothsort(tmpcivec%coeff,tmpcivec%config,imultblk(sm))
! Copy the result back to the original array.
         civecs(sm)%config(1:nc) = tmpcivec%config(1:nc)
         civecs(sm)%coeff(1:nc,1:imultblk(sm)) = tmpcivec%coeff(1:nc,1:imultblk(sm))

         deallocate(tmpcivec%config)
         deallocate(tmpcivec%coeff)

      enddo ! nsmult loop

! copy final output
      do n = 1,nsmult
         nc = nciconf(n)
         nst = imultblk(n)
         ciconfig(1:nc,n) = civecs(n)%config(1:nc) 
         cicoeff(1:nc,1:nst,n) = civecs(n)%coeff(1:nc,1:nst) 
      enddo

      if(allocated(occa))deallocate(occa)
      if(allocated(occb))deallocate(occb)
      if(allocated(occan))deallocate(occan)
      if(allocated(occbn))deallocate(occbn)
      if(allocated(graph))deallocate(graph)
      if(allocated(lcalca))deallocate(lcalca)
      if(allocated(lcalcb))deallocate(lcalcb)
      if(allocated(aminor))deallocate(aminor)
      if(allocated(bminor))deallocate(bminor)
      if(allocated(asum))deallocate(asum)
      if(allocated(confstr))deallocate(confstr)
      if(allocated(tmpcivec%config))deallocate(tmpcivec%config)
      if(allocated(tmpcivec%coeff))deallocate(tmpcivec%coeff)
      do n=1,nsmult
         if(allocated(civecs(n)%config)) deallocate(civecs(n)%config)
         if(allocated(civecs(n)%coeff)) deallocate(civecs(n)%coeff)
      enddo
      if(allocated(civecs)) deallocate(civecs)

      if(lcsf)then
         if(associated(csfdetptr%next))then
            do while (associated(csfdetptr%next))
               csfdetptr => csfdetptr%next
            enddo
         endif
         do while (associated(csfdetptr%prev))
            csfdetptr => csfdetptr%prev
            nullify(csfdetptr%next%prev)
            deallocate(csfdetptr%next)
            nullify(csfdetptr%next)
         enddo
         nullify(csfdetptr%prev)
         nullify(csfdetptr%next)
         nullify(icsfdetptr)
         if(allocated(civecptr%config))deallocate(civecptr%config)
         if(allocated(civecptr%coeff))deallocate(civecptr%coeff)
      endif
      nullify(civecptr)

      call stop_timer(tid)

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_CIVEC :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop

      end subroutine citrans

!#######################################################################
! Subroutine to compare the phases of the newly calculated states with 
! those of the reference wavefunction.
! GWR 05/19
! Updated to deal with CSFs properly.
! GWR 04/20
!#######################################################################

      subroutine molpro_wfover(sao,smo,moco,mocoref,nbas,lsign,msta,&
                 stover,modus,nciconf,ciconfig_idb,cicoeff_idb,ciconfig_iprox,&
                 cicoeff_iprox)

      implicit none

      logical(kind=4), dimension(msta,nsmult), intent(out) :: lsign
      integer(long), intent(in) :: nbas,msta,modus
      real(dop), dimension(nbas,nbas), intent(in) :: sao
      real(dop), dimension(nocc,nocc), intent(in) :: smo
!      real(sip), dimension(nbas,nbas), intent(inout) :: moco,mocoref
      real(dop), dimension(nbas,nbas), intent(inout) :: moco,mocoref
      real(dop), dimension(nsmult*msta*msta), intent(out) :: stover

      logical(kind=4) :: lcsfipr,lcsfidb
      character(2), parameter :: backsl='\\'
      integer(long) :: i,j,k,sm,nst,nrow,s,n,s1,sdx,ncoff,&
           sp,ndx,ierr,smoff,maxst,idx,iel,ncol,nend,jdx,&
           ntot,rtot,kdx,nsd,nc
      integer(long), dimension(2) :: confoff
      integer(long), dimension(:), allocatable :: occr,occn
      integer(long), dimension(:,:), allocatable :: confstr
      integer(long), dimension(:,:,:), allocatable :: graph
      real(dop) :: dover,maxover,maxci,onorm,cover
      real(dop), dimension(nocc) :: tau,wr,wi,z
      real(dop), dimension(5*nocc) :: work
      real(dop), dimension(:,:), allocatable :: tmpmat,sdet,tmpdet,cicpr,vl,&
           vr,stmp,mataover,matbover
      complex(dop) :: cdet
      integer(long), dimension(nsmult), intent(in) :: nciconf
      character(len=*), dimension(maxcivec,nsmult),intent(inout)   :: ciconfig_idb
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(inout) :: cicoeff_idb
      character(len=*), dimension(maxcivec,nsmult),intent(inout)   :: ciconfig_iprox
      real(dop), dimension(maxcivec,maxmultblk,nsmult), intent(inout) :: cicoeff_iprox

      type(csfdet), pointer :: csfdet_iprox,icsfdet_iprox,csfdet_idb,icsfdet_idb

      integer(long), parameter :: idb = 1
      integer(long), parameter :: iprox = 2
      type(civec), pointer :: iprcivec,idbcivec
      type(civec), target :: tiprcivec,tidbcivec
      type(civec), dimension(:,:), allocatable, target :: civecs
      character(len=30) :: cerr

! Timing
      integer(long), save   :: tid = 0 ! timer ID
      call get_timer(tid, 'molpro_wfover')
      call start_timer(tid)

      allocate(tmpmat(nocc,nbas),sdet(nocc,nocc),tmpdet(nocc,nocc),&
           vl(nocc,nocc),vr(nocc,nocc),stmp(msta,msta))

! Allocate and fill ci vector structures
      cerr = 'civecs'
      allocate(civecs(nsmult,2),stat=ierr)
      if (ierr > 0) go to 1000
      do n = 1,nsmult
         write(cerr,'(a,i6)') 'civecs_idb(n)%config',n
         allocate(civecs(n,1)%config(maxcivec),stat=ierr)
         if (ierr > 0) go to 1000
         write(cerr,'(a,i6)') 'civecs(n)%coeff',n
         allocate(civecs(n,1)%coeff(maxcivec,maxsta),stat=ierr)
         if (ierr > 0) go to 1000
         write(cerr,'(a,i6)') 'civecs_idb(n)%config',n
         allocate(civecs(n,2)%config(maxcivec),stat=ierr)
         if (ierr > 0) go to 1000
         write(cerr,'(a,i6)') 'civecs(n)%coeff',n
         allocate(civecs(n,2)%coeff(maxcivec,maxsta),stat=ierr)
         if (ierr > 0) go to 1000
      enddo
      civecs(:,:)%nconf = 0
      do n = 1,nsmult
         civecs(n,1)%config(1:maxcivec) = ' '
         civecs(n,2)%config(1:maxcivec) = ' '
         civecs(n,1)%coeff(:,:) = 0.0_dop
         civecs(n,2)%coeff(:,:) = 0.0_dop
      enddo

      do n = 1,nsmult
         nc = nciconf(n)
         nst = imultblk(n)
         civecs(n,1)%nconf = nc
         civecs(n,1)%config(1:nc) = ciconfig_idb(1:nc,n)
         civecs(n,1)%coeff(1:nc,1:nst) = cicoeff_idb(1:nc,1:nst,n)
         civecs(n,2)%nconf = nc
         civecs(n,2)%config(1:nc) = ciconfig_iprox(1:nc,n)
         civecs(n,2)%coeff(1:nc,1:nst) = cicoeff_iprox(1:nc,1:nst,n)
      enddo

! First task is to construct a matrix of all possible occupied MO overlaps.
! SMO other way round from before!!!
!      call molpro_moover(sao,smo,mocoref,moco,nbas,0)

! Copy the (unchanging) overlaps of the closed orbitals to the 
! overlap determinant.
         sdet = 0.0_dop
         do j = 1,nclosed
            do i = 1,nclosed
               sdet(i,j) = smo(i,j)
            enddo
         enddo

! If decoupling the closed and AS orbitals, calculate the closed-closed overlap.
      if(ldecoup)then
         call dgehrd(nclosed,1,nclosed,sdet,nocc,tau,work,&
              5*nocc,ierr)
         if(ierr.ne.0)then
            routine='molpro_wfover'
            write(message,'(a,i6)')'dgehrd error reducing overlap:',ierr
            call errormsg
         endif
         call dhseqr('E','N',nclosed,1,nclosed,sdet,nocc,&
              wr,wi,z,1,work,5*nocc,ierr)
         if(ierr.ne.0)then
            routine='molpro_wfover'
            write(message,'(a,i6)')'dhseqr error finding eigenvalues:',ierr
            call errormsg
         endif
! Determinant is the product of the eigenvalues.
         cdet = cmplx(wr(1),wi(1),dop)
         do n = 2,nclosed
            cdet = cdet*cmplx(wr(n),wi(n),dop)
         enddo
         cover=dble(cdet)
         ncoff = 0
         sdet = 0.0_dop
      else
         cover = 1.0_dop
         ncoff = nclosed
      endif

! Allocate array to hold state overlaps, and arrays holding AS occupancies.
!      allocate(stover(nsmult*msta*msta))
      allocate(occn(ncas),occr(ncas))
      maxst = maxval(imultblk(1:nsmult))
      allocate(cicpr(maxst,maxst))

! Construct the overlaps of the states.
      stover = 0.0_dop  
      smoff = 0

! Check whether we are dealing with CSFs or SDs just by checking how Molpro 
! outputs the configurations. May be different.
      lcsfipr = .false.
      smipr_loop: do sm = 1,nsmult
         do i = 1,nciconf(sm)
            do n = 1,ncas
               lcsfipr = lcsfipr .or. dbciconfig(i,sm,iprox)(n:n).eq.'/'&
                    .or. dbciconfig(i,sm,iprox)(n:n).eq.backsl &
                    .or. dbciconfig(i,sm,iprox)(n:n).eq.'\' 
               lcsfipr = lcsfipr .or. civecs(sm,iprox)%config(i)(n:n).eq.'/'&
                    .or. civecs(sm,iprox)%config(i)(n:n).eq.backsl &
                    .or. civecs(sm,iprox)%config(i)(n:n).eq.'\' 
            enddo
            if(lcsfipr)exit smipr_loop
         enddo
      enddo smipr_loop
      lcsfidb = .false.
      smidb_loop: do sm = 1,nsmult
         do i = 1,nciconf(sm)
            do n = 1,ncas
               lcsfidb = lcsfidb .or. civecs(sm,idb)%config(i)(n:n).eq.'/'&
                    .or. civecs(sm,idb)%config(i)(n:n).eq.backsl &
                    .or. civecs(sm,idb)%config(i)(n:n).eq.'\' 
            enddo
            if(lcsfidb)exit smidb_loop
         enddo
      enddo smidb_loop

      if(associated(csfdet_iprox))nullify(csfdet_iprox)
      if(associated(csfdet_idb))nullify(csfdet_idb)

! Loop over the wavefunction spin multiplicities.
      do sm = 1,nsmult

! Evaluation and storage of all possible overlap determinants.
! First determine which configurations there should be and the indices 
! they will have in the list, then evaluate all possible overlap
! determinants and save them to arrays.
         nrow = ncas-minval(ncasel(sm,1:2))+1
         ncol = maxval(ncasel(sm,1:2))+1
         if(allocated(graph))deallocate(graph)
         allocate(graph(nrow,ncol,2))
         if(allocated(confstr))deallocate(confstr)
         allocate(confstr(ncas,nconf(sm,1)+nconf(sm,2)))
         confstr = 0
         confoff(1) = 1
         confoff(2) = nconf(sm,1)+1
         if(allocated(mataover))deallocate(mataover)
         allocate(mataover(nconf(sm,1),nconf(sm,1)))
         if(allocated(matbover))deallocate(matbover)
         allocate(matbover(nconf(sm,2),nconf(sm,2)))

         do sp = 1,2
! Set up the graphs needed to address alpha and beta configurations.
            call setgraph(graph(1,1,sp),nrow,ncol,ncas,ncasel(sm,sp))

! If overlapping CSFs we need to know all possible alpha and beta configurations so
! generate them here.
            nend = ncas-ncasel(sm,sp)+1
            idx = 1 
            do i = 1,nend
               iel = 1
               nst = i
               confstr(i,idx+confoff(sp)-1) = 1
               call confgen(confstr(1,confoff(sp)),ncas,ncasel(sm,sp),nconf(sm,sp),&
                    nst,nend,idx,iel)
            enddo
         enddo

         nsd = nconf(sm,1)*nconf(sm,2)

! Calculate the overlap determinants.
         do sp = 1,2
! Avoid repetition if the alpha and beta spaces are the same.
            if(ncasel(sm,1).eq.ncasel(sm,2).and.sp.eq.2)cycle

!$omp parallel do private(i,j,k,n,idx,jdx,kdx,ndx,tmpdet,tau,work,ierr,wr,wi,z,cdet,ntot)&
!$omp firstprivate(sdet) if(nconf(sm,sp).gt.100 .and. lompthread) num_threads(ompthread)
            do i = 1,nconf(sm,sp)
               idx = confidx(nrow,ncol,ncas,ncasel(sm,sp),confstr(1,confoff(sp)+i-1),graph(1,1,sp))
               do j = 1,nconf(sm,sp)
                  jdx = confidx(nrow,ncol,ncas,ncasel(sm,sp),confstr(1,confoff(sp)+j-1),graph(1,1,sp))
! Active-active.
                  kdx = 0                    
                  do k = 1,ncas
                     if(confstr(k,confoff(sp)+j-1).eq.0)cycle
                     kdx=kdx+1
                     ndx=0
                     do n = 1,ncas
                        if(confstr(n,confoff(sp)+i-1).eq.0)cycle
                        ndx=ndx+1
                        sdet(ncoff+ndx,ncoff+kdx) = smo(nclosed+n,nclosed+k)
                     enddo
                  enddo
                     
! If we are doing the full overlap then:
! Reference active-new closed.
                  if(.not.ldecoup)then
                     do k = 1,nclosed
                        ndx=0
                        do n = 1,ncas
                           if(confstr(n,confoff(sp)+i-1).eq.0)cycle
                           ndx=ndx+1
                           sdet(ncoff+ndx,k) = smo(nclosed+n,k)
                        enddo
                     enddo
                     
! Reference closed-new active.
                     ndx=0
                     do n = 1,ncas
                        if(confstr(n,confoff(sp)+j-1).eq.0)cycle
                        ndx=ndx+1
                        do k = 1,nclosed
                           sdet(k,ncoff+ndx) = smo(k,nclosed+n)
                        enddo
                     enddo
                  endif

                  ntot = sum(confstr(1:ncas,confoff(sp)+i-1),1)
                  tmpdet=sdet
                  call dgehrd(ncoff+ntot,1,ncoff+ntot,tmpdet,nocc,tau,work,&
                       5*nocc,ierr)
                  if(ierr.ne.0)then
                     routine='molpro_wfover'
                     write(message,'(a,i6)')'dgehrd error reducing overlap:',ierr
                     call errormsg
                  endif
                  call dhseqr('E','N',ncoff+ntot,1,ncoff+ntot,tmpdet,nocc,&
                       wr,wi,z,1,work,5*nocc,ierr)
                  if(ierr.ne.0)then
                     routine='molpro_wfover'
                     write(message,'(a,i6)')'dhseqr error finding eigenvalues:',ierr
                     call errormsg
                  endif
                  
! Determinant is the product of the eigenvalues.
                  cdet = cmplx(wr(1),wi(1),dop)
                  do n = 2,ncoff+ntot
                     cdet = cdet*cmplx(wr(n),wi(n),dop)
                  enddo

! Copy to the overlap arrays. Copy alpha result to beta if the spaces are the same.
                  if(sp.eq.1)then
                     mataover(idx,jdx) = dble(cdet)*cover
                     if(ncasel(sm,1).eq.ncasel(sm,2))then
                        matbover(idx,jdx) = mataover(idx,jdx)
                     endif
                  else
                     matbover(idx,jdx) = dble(cdet)*cover
                  endif
               enddo
            enddo
!$omp end parallel do
         enddo

! Decompose the CI vectors for each geometry down into their component SDs if using CSFs.
         if(lcsfipr)then
! Allocate arrays.
            iprcivec => tiprcivec
            iprcivec%nconf = nsd
            if(allocated(iprcivec%config))deallocate(iprcivec%config)
            allocate(iprcivec%config(nsd))
            iprcivec%config(1:nsd) = ' '
            if(allocated(iprcivec%coeff))deallocate(iprcivec%coeff)
            allocate(iprcivec%coeff(nsd,imultblk(sm)))
            iprcivec%coeff(1:nsd,1:imultblk(sm)) = 0.0_dop

! Loop over the CSFs.
            do i = 1,civecs(sm,iprox)%nconf
               if(civecs(sm,iprox)%config(i).eq.' ' &
                    .or. maxval(abs(civecs(sm,iprox)%coeff(i,1:imultblk(sm)))).lt.1.0e-16_dop) cycle
! Set up linked lists.
               nullify(icsfdet_iprox)
               allocate(icsfdet_iprox)
               icsfdet_iprox%config = ' '
               icsfdet_iprox%coeff = 999999.0_dop
               if(associated(csfdet_iprox))then
                  do while (associated(csfdet_iprox%next))
                     csfdet_iprox => csfdet_iprox%next
                  enddo
                  do while (associated(csfdet_iprox%prev))
                     csfdet_iprox => csfdet_iprox%prev
                     nullify(csfdet_iprox%next%prev)
                     deallocate(csfdet_iprox%next)
                     nullify(csfdet_iprox%next)
                  enddo
               endif
               csfdet_iprox => icsfdet_iprox
               nullify(csfdet_iprox%prev)
               nullify(csfdet_iprox%next)
! Decompose the CSF into its constituent SDs.
               call csftodet(civecs(sm,iprox)%config(i),confstr,nconf(sm,1)+nconf(sm,2),&
                    confoff,sm,csfdet_iprox)

               do while (associated(csfdet_iprox))
                  do sp = 1,2
! Which CAS orbitals are occupied in the configurations?
                     occr = 0
                     do n = 1,ncas
                        if(sp.eq.1)then
                           if(csfdet_iprox%config(n:n).eq.'2' &
                                .or. csfdet_iprox%config(n:n).eq.'a') then
                              occr(n) = 1
                           endif
                        else
                           if(csfdet_iprox%config(n:n).eq.'2' &
                                .or. csfdet_iprox%config(n:n).eq.'b') then
                              occr(n) = 1
                           endif
                        endif
                     enddo
                     if(sp.eq.1)then
                        idx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occr,graph(1,1,sp))
                     else
                        jdx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occr,graph(1,1,sp))
                     endif
                  enddo
                  idx = (idx-1)*nconf(sm,2)+jdx
                  iprcivec%config(idx) = csfdet_iprox%config
                  do s = 1,imultblk(sm)
                     iprcivec%coeff(idx,s) = iprcivec%coeff(idx,s) &
                          + csfdet_iprox%coeff * civecs(sm,iprox)%coeff(i,s)
                  enddo
                  if(associated(csfdet_iprox%next))then
                     csfdet_iprox => csfdet_iprox%next
                  else
                     exit
                  endif
               enddo
            enddo

! Re-order the CI vectors in reducing maximum coefficient order, as output by Molpro.
            call smoothsort(iprcivec%coeff,iprcivec%config,imultblk(sm))

! Ensure that the phase is as Molpro, i.e. the largest coefficient for each state is positive.
            do s = 1,imultblk(sm)
               idx = maxloc(abs(iprcivec%coeff(1:nsd,s)),1)
               if(iprcivec%coeff(idx,s).lt.0.0_dop)&
                    iprcivec%coeff(1:nsd,s) = -iprcivec%coeff(1:nsd,s)
            enddo
            civecs(sm,iprox) = iprcivec

         else
            iprcivec => civecs(sm,iprox)
         endif

         if(lcsfidb)then
! Allocate arrays.
            idbcivec => tidbcivec
            idbcivec%nconf = nsd
            if(allocated(idbcivec%config))deallocate(idbcivec%config)
            allocate(idbcivec%config(nsd))
            idbcivec%config(1:nsd) = ' '
            if(allocated(idbcivec%coeff))deallocate(idbcivec%coeff)
            allocate(idbcivec%coeff(nsd,imultblk(sm)))
            idbcivec%coeff(1:nsd,1:imultblk(sm)) = 0.0_dop

! Loop over the CSFs.
            do i = 1,civecs(sm,idb)%nconf
               if(civecs(sm,idb)%config(i).eq.' ' &
                    .or. maxval(abs(civecs(sm,idb)%coeff(i,1:imultblk(sm)))).lt.1.0e-16_dop) cycle
! Set up linked lists.
               nullify(icsfdet_idb)
               allocate(icsfdet_idb)
               icsfdet_idb%config = ' '
               icsfdet_idb%coeff = 999999.0_dop
               if(associated(csfdet_idb))then
                  do while (associated(csfdet_idb%next))
                     csfdet_idb => csfdet_idb%next
                  enddo
                  do while (associated(csfdet_idb%prev))
                     csfdet_idb => csfdet_idb%prev
                     nullify(csfdet_idb%next%prev)
                     deallocate(csfdet_idb%next)
                     nullify(csfdet_idb%next)
                  enddo
               endif
               csfdet_idb => icsfdet_idb
               nullify(csfdet_idb%prev)
               nullify(csfdet_idb%next)
! Decompose the CSF into its constituent SDs.
               call csftodet(civecs(sm,idb)%config(i),confstr,nconf(sm,1)+nconf(sm,2),&
                    confoff,sm,csfdet_idb)
               do while (associated(csfdet_idb))
                  do sp = 1,2
! Which CAS orbitals are occupied in the configurations?
                     occr = 0
                     do n = 1,ncas
                        if(sp.eq.1)then
                           if(csfdet_idb%config(n:n).eq.'2' &
                                .or. csfdet_idb%config(n:n).eq.'a') then
                              occr(n) = 1
                           endif
                        else
                           if(csfdet_idb%config(n:n).eq.'2' &
                                .or. csfdet_idb%config(n:n).eq.'b') then
                              occr(n) = 1
                           endif
                        endif
                     enddo
                     if(sp.eq.1)then
                        idx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occr,graph(1,1,sp))
                     else
                        jdx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occr,graph(1,1,sp))
                     endif
                  enddo
                  idx = (idx-1)*nconf(sm,2)+jdx
                  idbcivec%config(idx) = csfdet_idb%config
                  do s = 1,imultblk(sm)
                     idbcivec%coeff(idx,s) = idbcivec%coeff(idx,s) &
                          + csfdet_idb%coeff * civecs(sm,idb)%coeff(i,s)
                  enddo
                  if(associated(csfdet_idb%next))then
                     csfdet_idb => csfdet_idb%next
                  else
                     exit
                  endif
               enddo
            enddo

! Re-order the CI vectors in reducing maximum coefficient order, as output by Molpro.
            call smoothsort(idbcivec%coeff,idbcivec%config,imultblk(sm))

! Ensure that the phase is as Molpro, i.e. the largest coefficient for each state is positive.
            do s = 1,imultblk(sm)
               idx = maxloc(abs(idbcivec%coeff(1:nsd,s)),1)
               if(idbcivec%coeff(idx,s).lt.0.0_dop)&
                    idbcivec%coeff(1:nsd,s) = -idbcivec%coeff(1:nsd,s)
            enddo
            civecs(sm,idb) = idbcivec
               
         else
            idbcivec => civecs(sm,idb)
         endif

! Loops over the reference (i) and new (j) configurations
!$omp parallel do private(i,j,maxci,s,s1,cicpr,sp,occr,occn,ntot,rtot,n,idx,jdx,sdx,dover) &
!$omp reduction(+:stover) schedule(static,2) if(nsd.gt.100)
         iprox_loop: do i = 1,nsd

! Exit if we are into the small components. This assumes that the CI coefficients
! are in order of decreasing magnitude.
            maxci = 0.0_dop
            do s = 1,imultblk(sm)
               if(abs(iprcivec%coeff(i,s)).gt.maxci) &
                    maxci = abs(iprcivec%coeff(i,s))
            enddo

            if(maxci.lt.cicocut)cycle !exit iprox_loop

            idb_loop: do j = 1,nsd

! Construct vector of products of CI coefficients for each state. Used in 
! screening and during the overlap.
               maxci=0.0_dop
               do s = 1,imultblk(sm)
                  do s1 = 1,imultblk(sm)
                     cicpr(s1,s) = iprcivec%coeff(i,s1) &
                          * idbcivec%coeff(j,s)
                     if(abs(cicpr(s1,s)).gt.maxci) maxci = abs(cicpr(s1,s))
                 enddo
               enddo

! Don't bother if the product of the coefficients is insignificant.
               if(maxci.lt.cicocut) exit idb_loop

! Loop over alpha and beta parts.
               dover = 0.0
               do sp = 1,2
! Which CAS orbitals are occupied in the configurations?
                  occr = 0
                  occn = 0
                  ntot = 0
                  rtot = 0
                  do n = 1,ncas
                     if(sp.eq.1)then
                        if(idbcivec%config(j)(n:n).eq.'2' &
                             .or. idbcivec%config(j)(n:n).eq.'a') then
                           ntot = ntot + 1
                           occn(n) = 1
                        endif
                        if(iprcivec%config(i)(n:n).eq.'2' &
                             .or. iprcivec%config(i)(n:n).eq.'a') then
                           rtot = rtot + 1
                           occr(n) = 1
                        endif
                     else
                        if(idbcivec%config(j)(n:n).eq.'2' &
                             .or. idbcivec%config(j)(n:n).eq.'b') then
                           ntot = ntot + 1
                           occn(n) = 1
                        endif
                        if(iprcivec%config(i)(n:n).eq.'2' &
                             .or. iprcivec%config(i)(n:n).eq.'b') then
                           rtot = rtot + 1 
                           occr(n) = 1
                        endif
                     endif
                  enddo

                  if(ntot.ne.rtot)then
                     routine = 'molpro_wfover'
                     write(message,'(2a,4i6,3a)')'Inconsistent number of orbitals in new ', &
                          'and old configs.',ntot,rtot,i,j,iprcivec%config(i)(1:ncas),&
                          ' ',idbcivec%config(j)(1:ncas)
                     call errormsg
                  endif

! Combine the overlap of the determinants with the CI coefficients and add to the 
! overall overlap.
                  idx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occr,graph(1,1,sp))
                  jdx = confidx(nrow,ncol,ncas,ncasel(sm,sp),occn,graph(1,1,sp))

                  if(sp.eq.1)then
                     dover = mataover(idx,jdx)
                  else
                     dover = dover*matbover(idx,jdx)
                  endif
               enddo

               sdx = 0
               do s = 1,imultblk(sm)
                  do s1 = 1,imultblk(sm)
                     sdx = sdx + 1
                     stover(smoff+sdx) = stover(smoff+sdx) + dover * cicpr(s1,s)
                  enddo
               enddo

            enddo idb_loop
         enddo iprox_loop
!$omp end parallel do

         smoff = smoff + imultblk(sm)*imultblk(sm)

      enddo

! Scan down the columns to find the largest component, check its sign
! then, if negative, set lsign=true for that state.
! Only do this when fixing phases for propagation diabatisation, but we
! need the raw overlaps for the Procrustes diabatisation.
      if(modus.eq.1)then
         lsign = .false.
         sdx = 0
         do sm = 1,nsmult
            do s = 1,imultblk(sm)
               maxover = 0.0_dop
               onorm = 0.0_dop
               do s1 = 1,imultblk(sm)
                  sdx = sdx + 1
                  onorm = onorm + stover(sdx)*stover(sdx)
                  if(abs(stover(sdx)).gt.maxover)then
                     maxover = abs(stover(sdx))
                     lsign(s,sm) = stover(sdx).lt.0.0_dop
                  endif
               enddo
               onorm = sqrt(onorm)
               if(onorm.lt.0.25_dop)lsign(s,sm) = .false.
            enddo
         enddo
      else if(modus.eq.2 .and. idb.eq.iprox)then
! Normalise the overlaps by modifying the CI vectors.
         smoff = 0
         do sm = 1,nsmult
            sdx = 1
            do s = 1,imultblk(sm)
               do s1 = 1,imultblk(sm)
                  stmp(s1,s) = stover(sdx+smoff)
                  sdx = sdx + 1
               enddo
            enddo
            do s = 1,imultblk(sm)
               onorm = 1.0_dop/sqrt(stmp(s,s))
               do i = 1,civecs(sm,idb)%nconf
                  civecs(sm,idb)%coeff(i,s) = onorm*civecs(sm,idb)%coeff(i,s)
               enddo
            enddo
            smoff = smoff + imultblk(sm)*imultblk(sm)
         enddo        
      endif

      if(allocated(mataover))deallocate(mataover)
      if(allocated(matbover))deallocate(matbover)
      if(allocated(graph))deallocate(graph)
      if(allocated(confstr))deallocate(confstr)
      deallocate(tmpmat,sdet,tmpdet,occr,occn,cicpr,vl,vr,stmp)

      if(lcsfipr)then
         if(associated(csfdet_iprox%next))then
            do while (associated(csfdet_iprox%next))
               csfdet_iprox => csfdet_iprox%next
            enddo
         endif
         do while (associated(csfdet_iprox%prev))
            csfdet_iprox => csfdet_iprox%prev
            nullify(csfdet_iprox%next%prev)
            deallocate(csfdet_iprox%next)
            nullify(csfdet_iprox%next)
         enddo
         nullify(csfdet_iprox%prev)
         nullify(csfdet_iprox%next)
         nullify(icsfdet_iprox)
      endif
      iprcivec => tiprcivec
      if(allocated(iprcivec%config))deallocate(iprcivec%config)
      if(allocated(iprcivec%coeff))deallocate(iprcivec%coeff)
      
      if(lcsfidb)then
         do while (associated(csfdet_idb%next))
            csfdet_idb => csfdet_idb%next
         enddo
         do while (associated(csfdet_idb%prev))
            csfdet_idb => csfdet_idb%prev
            nullify(csfdet_idb%next%prev)
            deallocate(csfdet_idb%next)
            nullify(csfdet_idb%next)
         enddo
         nullify(csfdet_idb%prev)
         nullify(csfdet_idb%next)
         nullify(icsfdet_idb)
      endif
      idbcivec => tidbcivec
      if(allocated(idbcivec%config))deallocate(idbcivec%config)
      if(allocated(idbcivec%coeff))deallocate(idbcivec%coeff)

! Copy back final vectors
   !   do n = 1,nsmult
   !      nc = nciconf(n)
   !      nst = imultblk(n)
   !      ciconfig_idb(1:nc,n) = civecs(n,1)%config(1:nc) 
   !      cicoeff_idb(1:nc,1:nst,n) = civecs(n,1)%coeff(1:nc,1:nst) 
   !      ciconfig_iprox(1:nc,n) = civecs(n,2)%config(1:nc) 
   !      cicoeff_iprox(1:nc,1:nst,n) = civecs(n,2)%coeff(1:nc,1:nst) 
   !   enddo

      do n=1,nsmult
         if(allocated(civecs(n,1)%config)) deallocate(civecs(n,1)%config)
         if(allocated(civecs(n,1)%coeff)) deallocate(civecs(n,1)%coeff)
         if(allocated(civecs(n,2)%config)) deallocate(civecs(n,2)%config)
         if(allocated(civecs(n,2)%coeff)) deallocate(civecs(n,2)%coeff)
      enddo

      nullify(idbcivec)
      nullify(iprcivec)
      
      call stop_timer(tid)

      return
 1000 write(ilog,'(a)') 'Error allocating memory in ALLOC_CIVEC :',cerr
      write(ilog,'(a,i5)') 'Error Code :',ierr
      stop
      end subroutine molpro_wfover

!#######################################################################
! Subroutine to get the overlaps of sets of MOs from Molpro, given the 
! AO overlap matrix. 
! GWR 10/19
! If modus = 1, perform Lowdin orthonormalisation
!#######################################################################
      subroutine molpro_moover(sao,smo,moco1,moco2,nbas,modus)

      implicit none

      integer(long), intent(in) :: nbas, modus
      real(dop), dimension(nbas,nbas), intent(in) :: sao
!      real(sip), dimension(nbas,nbas), intent(inout) :: moco1,moco2
      real(dop), dimension(nbas,nbas), intent(inout) :: moco1,moco2
      real(dop), dimension(nocc,nocc), intent(out) :: smo

      integer(long)                   :: i,j,k
      real(dop), dimension(nbas,nbas) :: tmpmat
      real(dop), dimension(nocc,nocc) :: tmpsmo,sigmamo,umatmo,vmattmo
      integer(long)                        :: lwork,info
      real(dop), dimension(:), allocatable :: work

      smo = 0.0_dop
      tmpmat=0.0_dop
! Overlap new (moco1) with old (moco2)
      do k = 1,nbas
         do j = 1,nbas
            do i = 1,nocc
               tmpmat(i,j) = tmpmat(i,j) + dble(moco1(i,k))*sao(k,j)
            enddo
         enddo
      enddo
      do k = 1,nbas
         do j = 1,nocc
            do i = 1,nocc
               smo(i,j) = smo(i,j) + tmpmat(i,k)*dble(moco2(j,k))
            enddo
         enddo
      enddo

      if (modus .eq. 1) then
         lwork = 5*nocc
         allocate(work(lwork))

         tmpsmo = smo
         call dgesvd('S','S',nocc,nocc,tmpsmo,nocc,sigmamo,umatmo,nocc,vmattmo,&
              nocc,work,lwork,info)                                 
         if(info.ne.0)then
            routine = 'Molpro_moover (Funcqchemmod.F90)'
            write(message,'(a,i10,x,a,2i6)')'Error in SVD for AS MOs:',info
            call errormsg
         endif

         do i = 1,nocc
            do j = 1,nocc
               smo(j,i) = umatmo(j,1)*vmattmo(1,i)
               do k = 2,nocc
                  smo(j,i) = smo(j,i) + umatmo(j,k)*vmattmo(k,i)
               enddo
            enddo
         enddo
!        tmpmat = moco1
!        moco1(1:nocc,1:nbasis)=0.0_dop
!        do k = 1,nbasis
!           do j = 1,nocc
!              do i = 1,nocc
!                 moco1(i,k) = moco(i,k) + smo(j,i)*moco1(j,k)
!              enddo
!           enddo
!        enddo
         deallocate(work)
      endif

      return
      end subroutine molpro_moover

!#######################################################################
! Function to extract index of a given configuration using lexical 
! ordering.  
! GWR 10/19 
!#######################################################################
      integer*4 function confidx(nrow,ncol,norb,nel,conf,graph)

      implicit none

      integer(long), intent(in) :: norb,nel,nrow,ncol
      integer(long), dimension(norb), intent(in) :: conf
      integer(long), dimension(nrow,ncol), intent(in) :: graph

      integer(4) :: i,idx,ir,ic
      
      idx = 0 
      ir = 1     
      ic = 1     
      do i = 1,norb
         if(conf(i).eq.1)then
            ic = ic + 1
         elseif(conf(i).eq.0)then
            idx = idx + graph(ir,ic+1)
            ir = ir + 1
         else
            routine = 'confidx'
            write(message,'(a,2i6)')'Unrecognised config.',i,conf(i)
            stop
         endif
         if(ic.eq.nel+1)exit
      enddo
! Indices start at 0, so add 1 for Fortran.
      confidx = idx+1

      return
      end function confidx

!#########################################################################
! Subroutines to perform a smoothsort (smoothsort, smooth_heapify, 
! smooth_extract, sift_in, interheap_sift) on the list of CI vectors.
! Modified from www.mjr19.org.uk/IT/sorts/ to sort into decreasing order.
! GWR 04/20
!#########################################################################
      subroutine smoothsort(arrayr,arrayc,nst)

      implicit none

      character(30), intent(inout) :: arrayc(:)
      real(dop), intent(inout) :: arrayr(:,:)
      integer(long), intent(in) :: nst
      integer(vlong) :: mask
      integer(long) :: offset
        
      if (size(arrayr,1).le.1) return
                 
      call smooth_heapify(arrayr,arrayc,nst,mask,offset)
        
      call smooth_extract(arrayr,arrayc,nst,mask,offset)
      
      end subroutine smoothsort

!###########################################################################
! Recursive routine to generate all possible CAS configurations given the
! number of electrons and orbitals. The list is in lexical order.
! Needs to be initialised properly; see molpro_wfover routine.
! GWR 10/19      
!###########################################################################  
      recursive subroutine confgen(confstr,norb,nel,nconf1,nst,nend,idx,iel)
            
      implicit none
            
      integer(long), intent(in) :: norb,nel,nconf1
      integer(long), intent(inout) :: nst,nend,idx,iel
      integer(long), dimension(norb,nconf1), intent(inout) :: confstr

      integer(long), dimension(norb):: conf
      integer(long) :: i,nst1

      conf(1:norb) = confstr(1:norb,idx)
      nst = nst + 1
      iel = iel + 1
           
      nend = nend + 1
         
      if(nend.le.norb)then
         do i = nst,nend
            confstr(1:nst-1,idx) = conf(1:nst-1)
            confstr(i,idx) = 1
            if(iel.eq.nel)then
               idx = idx + 1
            else
               nst1 = i
               call confgen(confstr,norb,nel,nconf1,nst1,nend,idx,iel)
            endif
         enddo
      else
         idx = idx + 1
      endif

      nend = nend - 1
      iel = iel - 1

      return
      end subroutine confgen

!#######################################################################
! Subroutine to decompose a CSF to its component SDs. 
! Based on Molecular Electonic Structure Theory by Helgaker et al, 
! Chapter 2.6.
! GWR 04/20
!#######################################################################
      subroutine csftodet(csf,confstr,nconfstr,confoff,sm,csfdeta)

      implicit none

      character(len=ncas), intent(in) :: csf
      integer(long), dimension(ncas,nconfstr), intent(in) :: confstr
      integer(long), dimension(2), intent(in) :: confoff
      integer(long), intent(in) :: sm,nconfstr
      type(csfdet), pointer, intent(inout) :: csfdeta

      logical(kind=4) :: lfirst,lclosed
      integer(long) :: sp,idx,i,n,j,jdx
      integer(long), dimension(:), allocatable :: csfadet,csfbdet
      character(30) :: detocc
      real(dop) :: dcoeff
      real(dop), dimension(ncas) :: tcvec,tvec,pcvec,pvec

! Timing
!      integer(long), save   :: tid = 0 ! timer ID
!      call get_timer(tid, 'csftodet')
!      call start_timer(tid)

! If the CSF has no unpaired electrons, copy the CSF to the determinant 
! array and return.
      lclosed = .true.
      do i = 1,ncas
         lclosed = lclosed .and. (csf(i:i).eq.'2' .or. csf(i:i) .eq. '0')
      enddo
      if(lclosed)then
         csfdeta%coeff = 1.0_dop
         csfdeta%config(1:ncas) = csf(1:ncas)
!         call stop_timer(tid)
         return
      endif

! Get the total spin arrays corresponding to the CSF, T and t. 
      call get_csf_totspin(tcvec,tvec,csf)

! Check which alpha and beta configurations can contribute to the CSF 
      if(allocated(csfadet))deallocate(csfadet)
      allocate(csfadet(nconf(sm,1)))
      if(allocated(csfbdet))deallocate(csfbdet)
      allocate(csfbdet(nconf(sm,2)))

! Get the orbital indices of the unpaired electrons.
      csfadet = 0
      csfbdet = 0
      do sp = 1,2
! If the orbital in the SD is occupied and that in the CSF is not ignore; &
! likewise if the CSF orbital is doubly occupied and the SD orbital empty, &
! ignore. Do alpha and beta separately.  
         idx = 1
         det_loop: do i = 1,nconf(sm,sp)

            do n = 1,ncas
               if((tcvec(n).eq.-1024.0_dop .and. confstr(n,i+confoff(sp)-1).ne.0)&
                    .or.(tcvec(n).eq.1024.0_dop .and. confstr(n,i+confoff(sp)-1).eq.0))&
                    cycle det_loop
            enddo
            if(sp.eq.1)then
               csfadet(idx) = i
            else
               csfbdet(idx) = i
            endif
            idx = idx + 1
         enddo det_loop
      enddo

      lfirst = .true.
! Generate the full determinants by combining the alpha and beta parts of the SDs.
      iploop: do i = 1,nconf(sm,1)
         if(csfadet(i).eq.0)exit iploop
         idx = csfadet(i)
         jploop: do j = 1,nconf(sm,2)
            if(csfbdet(j).eq.0)exit jploop
            jdx = csfbdet(j)
            detocc = ' '
            do n = 1,ncas
               if(confstr(n,idx+confoff(1)-1).eq.1)then
                  if(confstr(n,jdx+confoff(2)-1).eq.1)then
                     detocc(n:n) = '2'
                  else
                     detocc(n:n) = 'a'
                  endif
               else
                  if(confstr(n,jdx+confoff(2)-1).eq.1)then
                     detocc(n:n) = 'b'
                  else
                     detocc(n:n) = '0'
                  endif
               endif
! Again check double/0 occupancies match between the CSF and SD.
               if((detocc(n:n).eq.'2' .and. csf(n:n).ne.'2') &
                    .or. (detocc(n:n).eq.'0' .and. csf(n:n).ne.'0'))&
                    cycle jploop
            enddo

! Now generate the coefficient of the SD in the CSF.
! First convert the SD to the projected spin vectors, P and p.
            call get_det_projspin(pcvec,pvec,detocc(1:ncas))

! Reject determinants where the projected spin is greater than the total spin at a given entry.
            do n = 1,ncas
               if(abs(pcvec(n)).lt.1024.0_dop .and. &
                    (abs(pcvec(n)).gt.tcvec(n)))cycle jploop
            enddo
! Finally, construct the coefficients.
            dcoeff = 1.0_dop
            do n = 1,ncas
               if(abs(pcvec(n)).eq.1024.0_dop)cycle
               if(tvec(n).gt.0.0_dop)then
                  dcoeff = dcoeff * sqrt((tcvec(n)+2.0_dop*pvec(n)*pcvec(n))&
                       /(2.0_dop*tcvec(n))) 
               else
                  dcoeff = dcoeff * (-2.0_dop*pvec(n)) &
                       * sqrt((tcvec(n)+1.0_dop-2.0_dop*pvec(n)*pcvec(n))&
                       /(2.0_dop*(tcvec(n)+1))) 
               endif
            enddo
            if(abs(dcoeff).gt.1.0e-8_dop)then
               if(.not.lfirst)then
                  allocate(csfdeta%next)
                  csfdeta%next%prev => csfdeta
                  csfdeta => csfdeta%next
                  nullify(csfdeta%next)
               else
                  lfirst = .false.
               endif
               csfdeta%config = ' '
               csfdeta%config(1:ncas) = detocc(1:ncas)
               csfdeta%coeff = dcoeff
            endif
         enddo jploop
      enddo iploop

! Rewind the linked list.
      do while (associated(csfdeta%prev))
         csfdeta => csfdeta%prev
      enddo

      if(allocated(csfadet))deallocate(csfadet)
      if(allocated(csfbdet))deallocate(csfbdet)

!      call stop_timer(tid)

      return
      end subroutine csftodet

!#######################################################################
! Subroutine to get the total spin vectors for a given CSF.
! GWR 04/20
!#######################################################################
      subroutine get_csf_totspin(tcvec,tvec,csf)

      implicit none

      real(dop), dimension(ncas), intent(out) :: tcvec,tvec
      character(*), intent(in) :: csf

      integer(long) :: i,iprev
      character(2), parameter :: backsl='\\'

      iprev = 0
      do i = 1,ncas
         if(csf(i:i).eq.'2')then
            tcvec(i) = 1024.0_dop
            tvec(i) = 1024.0_dop
            cycle
         else if(csf(i:i).eq.'0') then
            tcvec(i) = -1024.0_dop
            tvec(i) = -1024.0_dop
            cycle
         else if (csf(i:i).eq.'/' .or. csf(i:i).eq.'a')then
    !     else if (csf(i:i).eq.'/')then
            if(iprev.eq.0)then
               tcvec(i) = 0.5_dop
               tvec(i) = 0.5_dop
            else
               tcvec(i) = tcvec(iprev) + 0.5_dop
               tvec(i) = tcvec(i) - tcvec(iprev)
            endif
            iprev = i
         else if (csf(i:i) .eq. '\' .or. csf(i:i).eq. 'b' &
                     .or. csf(i:i).eq.backsl)then
   !      else if (csf(i:i) .eq. '\' .or. csf(i:i).eq.backsl)then
            if(iprev.eq.0)then
               tcvec(i) = 0.5_dop
               tvec(i) = 0.5_dop
            else
               tcvec(i) = tcvec(iprev) - 0.5_dop
               tvec(i) = tcvec(i) - tcvec(iprev)
            endif
            iprev = i
         else
            routine = 'get_csf_totspin'
            write(message,'(2a)')'Unrecognised CSF entry: ',csf(1:ncas)
            call errormsg
         endif
      enddo

      return

      end subroutine get_csf_totspin

!#######################################################################
! Subroutine to get the projected spin vectors for a given SD.
! GWR 04/20
!#######################################################################
      subroutine get_det_projspin(pcvec,pvec,det)

      implicit none

      real(dop), dimension(ncas), intent(out) :: pcvec,pvec
      character(*), intent(in) :: det

      integer(long) :: i,iprev

      iprev = 0
      do i = 1,ncas
         if(det(i:i).eq.'2')then
            pcvec(i) = 1024.0_dop
            pvec(i) = 1024.0_dop
            cycle
         else if(det(i:i).eq.'0') then
            pcvec(i) = -1024.0_dop
            pvec(i) = -1024.0_dop
            cycle
         else if (det(i:i).eq.'a')then
            if(iprev.eq.0)then
               pcvec(i) = 0.5_dop
            else
               pcvec(i) = pcvec(iprev) + 0.5_dop
            endif
            pvec(i) = 0.5_dop
            iprev = i
         else if (det(i:i).eq.'b')then
            if(iprev.eq.0)then
               pcvec(i) = -0.5_dop
            else
               pcvec(i) = pcvec(iprev) - 0.5_dop
            endif
            pvec(i) = -0.5_dop
            iprev = i
         else
            routine = 'get_det_projspin'
            message = 'Unrecognised SD entry.'
            call errormsg
         endif
      enddo

      return

      end subroutine get_det_projspin

!#######################################################################
! Define the configuration indexing graph needed for molpro_wfover.
! Based on Pascal's triangle using lexical ordering.
! GWR 10/19
!#######################################################################
      subroutine setgraph(graph,nrow,ncol,norb,nel)

      implicit none

      integer(long), intent(in) :: nrow,ncol,norb,nel
      integer(long), dimension(nrow,ncol), intent(out) :: graph

      integer(long) :: i,j,nrow1,ncol1
      integer(long), dimension(nrow+1,ncol+1) :: tmpgraph

      tmpgraph = 0
      nrow1 = norb-nel+1
      ncol1 = nel+1

      do j = ncol1,1,-1
         do i = nrow1,1,-1
            if(i.eq.nrow1.and.j.eq.ncol1)then
               tmpgraph(i,j) = 1
            else
               tmpgraph(i,j) = tmpgraph(i+1,j)+tmpgraph(i,j+1)
            endif
         enddo
      enddo

      do j = 1,nel+1
         do i = 1,norb-nel+1
            graph(i,j) = tmpgraph(i,j)
         enddo
      enddo

      return

      end subroutine setgraph

!#########################################################################
      subroutine smooth_heapify(arrayr,arrayc,nst,mask,offset)

      implicit none

      character(30), intent(inout) :: arrayc(0:)
      real(dop), intent(inout) :: arrayr(0:,:)
      integer(long), intent(in) :: nst
      integer(vlong), intent(out) :: mask
      integer(long), intent(out) :: offset
      integer(long) :: i,last
      logical(kind=4) :: wbf

      mask=1
      offset=1
      last=size(arrayr,1)

      do i=1,last-1
         if (iand(mask,2_vlong).eq.2) then
            mask=ior(ishft(mask,-2),1_vlong)
            offset=offset+2
         else if (offset.eq.1) then
            mask=ior(ishft(mask,1),1_vlong)
            offset=0
         else
            mask=ior(ishft(mask,offset-1),1_vlong)
            offset=1
         endif
! Following logical test rewritten to prevent attempt to access L(-1)
!         wbf=(((iand(mask,2_vlong).eq.2).and.(i+1.lt.last)).or. &
!              ((offset.gt.0).and.(1+i+L(offset-1).lt.last)))
         wbf=(iand(mask,2_vlong).eq.2).and.(i+1.lt.last)
         if(offset.gt.0)then
            wbf = wbf .or. (1+i+L(offset-1).lt.last)
         endif
         if (wbf) then
            call sift_in(arrayr,arrayc,nst,i,offset)
         else
            call interheap_sift(arrayr,arrayc,nst,i,mask,offset)
         endif
      enddo

      end subroutine smooth_heapify

!########################################################################
      subroutine smooth_extract(arrayr,arrayc,nst,mask,offset)

      implicit none
      
      character(30), intent(inout) :: arrayc(0:) 
      real(dop), intent(inout) :: arrayr(0:,:)              
      integer(long), intent(in) :: nst
      integer(long) :: i,j,last,ch(2)
      integer(vlong),value :: mask
      integer(long), value :: offset
      
      last=size(arrayr,1)
      
      do i=last-1,2,-1
         if (offset.lt.2) then
            do
               mask=ishft(mask,-1)
               offset=offset+1
               if (iand(mask,1_vlong).eq.1) exit
            enddo
         else
            ch(2)=i-1
            ch(1)=ch(2)-L(offset-2)
            mask=iand(mask,not(1_vlong))
            do j=1,2
               mask=ior(ishft(mask,1),1_vlong)
               offset=offset-1
               call interheap_sift(arrayr,arrayc,nst,ch(j),mask,offset)
            end do
         end if
      end do

      end subroutine smooth_extract

!#################################################################
      subroutine sift_in(arrayr,arrayc,nst,root,sz)

      implicit none

      character(30), intent(inout) :: arrayc(0:)
      character(30) :: tmpc
      real(dop), intent(inout) :: arrayr(0:,:)
      real(dop) :: tmpabs
      real(dop), dimension(nst) :: tmp
      integer(long), intent(in) :: nst
      integer(long) :: left,right,next
      integer(long),value :: root,sz

      if (sz.lt.2) return

      tmpabs=maxval(abs(arrayr(root,1:nst)))
      tmp = arrayr(root,1:nst)
      tmpc = arrayc(root)

      do
         right=root-1
         left=right-L(sz-2)

!         if (maxval(abs(arrayr(right,1:nst)))&
!              .lt.maxval(abs(arrayr(left,1:nst)))) then
         if (maxval(abs(arrayr(right,1:nst)))&
              .gt.maxval(abs(arrayr(left,1:nst)))) then
            next=left
            sz=sz-1
         else
            next=right
            sz=sz-2
         end if
         
!         if (maxval(abs(arrayr(next,1:nst))).le.tmpabs) exit
         if (maxval(abs(arrayr(next,1:nst))).ge.tmpabs) exit

         arrayr(root,1:nst)=arrayr(next,1:nst)
         arrayc(root)=arrayc(next)

         root=next
         if (sz.le.1) exit
      end do

      arrayr(root,1:nst)=tmp(1:nst)
      arrayc(root) = tmpc

      end subroutine sift_in

!################################################################
      subroutine interheap_sift(arrayr,arrayc,nst,root,mask,offset)

      implicit none

      character(30), intent(inout) :: arrayc(0:)
      character(30) :: tmpc
      real(dop) :: tmpabs
      real(dop), dimension(nst) :: tmp
      real(dop), intent(inout) :: arrayr(0:,:)
      real(dop) :: mx
      integer(long), intent(in) :: nst
      integer :: left,right,next
      integer(vlong),value :: mask
      integer, value :: root,offset

      tmpabs=maxval(abs(arrayr(root,1:nst)))
      tmp(1:nst)=arrayr(root,1:nst)
      tmpc = arrayc(root)

      do while (mask.ne.1)
         mx=tmpabs

         if (offset.gt.1) then
            right=root-1
            left=right-L(offset-2)
!            mx=max(mx,maxval(abs(arrayr(left,1:nst))),&
!                 maxval(abs(arrayr(right,1:nst))))
            mx=min(mx,maxval(abs(arrayr(left,1:nst))),&
                 maxval(abs(arrayr(right,1:nst))))
         end if

         next=root-L(offset)
       
!         if (maxval(abs(arrayr(next,1:nst))).le.mx) exit
         if (maxval(abs(arrayr(next,1:nst))).ge.mx) exit
         
         arrayr(root,1:nst) = arrayr(next,1:nst)
         arrayc(root) = arrayc(next)
         root=next

         do
            mask=ishft(mask,-1)
            offset=offset+1
            if (iand(mask,1_vlong).eq.1) exit
         end do
      end do

      arrayr(root,1:nst)=tmp(1:nst)
      arrayc(root) = tmpc

      call sift_in(arrayr,arrayc,nst,root,offset)

      end subroutine interheap_sift

!#########################################################################
      end module funcqchemmod

